/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: EditorRendererRegistry.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:41 $  $Locker:  $  Quest Software Inc.
package com.klg.jclass.cell;

import java.util.Vector;
import java.util.Hashtable;

/**
 * This is the central editor renderer registry for
 * JClass containers like HiGrid and LiveTable.  Both
 * HiGrid and LiveTable use a single instance of the
 * registry to get default editor and renderer types
 * based on data types.
 *
 * The EditorRendererRegistry is somewhat under the
 * covers in JClass.  In single container instances or
 * simple applications, you don't have to worry about
 * it - use the container (ie LiveTable or HiGrid) API
 * to control type-to-editor and type-to-renderer mappings.
 *
 * However, if you have multiple LiveTable and HiGrid
 * instances, and you want to control default type mappings,
 * or if you want to add your own default mappings, then
 * use this class.
 *
 * @see com.klg.jclass.higrid.HiGrid
 * @see com.klg.jclass.table.JCTable
 */
public class EditorRendererRegistry {

/**
 * The central registry instance
 */
static protected EditorRendererRegistry centralRegistry =
    new EditorRendererRegistry();

/**
 * Pointer to parent registry.  Allows registry chaining.
 */
protected EditorRendererRegistry parentRegistry = null;

/**
 * Flag indicating whether this is the central registry.
 */
protected boolean isCentralRegistry = false;

// Hashtables for storing the editors and renderers.
// Key is the class the editor/renderer is for.  The
// value is the editor/renderer class or an instance of
// the editor/renderer if it has been used atleast once.
protected Hashtable classCellsTable = new Hashtable();

/**
 * Private constructor used to set centralRegistry static variable.
 */
private EditorRendererRegistry()
{
    isCentralRegistry = true;
    initDefaultTable();
}

/**
 * Creates a new registry as a child of the provided registry.
 *
 * @param parentRegistry parent registry to use for additional
 * mappings.  Can be null.
 */
public EditorRendererRegistry(EditorRendererRegistry parentRegistry)
{
    this.parentRegistry = parentRegistry;
}

/**
 * Retrieves the central registry for use.  This is the
 * registry used by default in HiGrid and LiveTable.
 * @return The central registry
 */
static public EditorRendererRegistry getCentralRegistry()
{
    return(centralRegistry);
}

/**
 * Initializes the default mappings.  Override this method
 * if you want to create your own special mappings by
 * default.
 */
protected void initDefaultTable()
{
    // Note:  Any modifications can be made to this, but there MUST be a
    // "java.lang.Object" added with a valid renderer.  If not JCTable, Higrid
    // and others will not function properly.
    //
    // Every class added must have a valid Renderer.

    addClass("java.lang.Object",
             null,
             "com.klg.jclass.cell.editors.JCStringCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.math.BigDecimal",
             null,
             "com.klg.jclass.cell.editors.JCBigDecimalCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.lang.Boolean",
             null,
             "com.klg.jclass.cell.editors.JCCheckBoxCellEditor",
             "com.klg.jclass.cell.renderers.JCCheckBoxCellRenderer");
/*
    addClass("java.lang.Boolean",
             null,
             "com.klg.jclass.cell.editors.JCBooleanCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");
*/
    addClass("java.lang.Byte",
             null,
             "com.klg.jclass.cell.editors.JCByteCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.util.Date",
             null,
             "com.klg.jclass.cell.editors.JCDateCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.lang.Double",
             null,
             "com.klg.jclass.cell.editors.JCDoubleCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.lang.Float",
             null,
             "com.klg.jclass.cell.editors.JCFloatCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.awt.Image",
             null,
             "com.klg.jclass.cell.editors.JCImageCellEditor",
             "com.klg.jclass.cell.renderers.JCImageCellRenderer");

    addClass("java.lang.Integer",
             null,
             "com.klg.jclass.cell.editors.JCIntegerCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.lang.Long",
             null,
             "com.klg.jclass.cell.editors.JCLongCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.lang.String",
             null,
             "com.klg.jclass.cell.editors.JCStringCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.lang.String",
             "MultiLine",
             "com.klg.jclass.cell.editors.JCMultilineCellEditor",
             "com.klg.jclass.cell.renderers.JCWordWrapCellRenderer");

    addClass("java.lang.String",
             "WordWrap",
             "com.klg.jclass.cell.editors.JCWordWrapCellEditor",
             "com.klg.jclass.cell.renderers.JCWordWrapCellRenderer");

    addClass("java.lang.Short",
             null,
             "com.klg.jclass.cell.editors.JCShortCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.sql.Date",
             null,
             "com.klg.jclass.cell.editors.JCSqlDateCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.sql.Time",
             null,
             "com.klg.jclass.cell.editors.JCSqlTimeCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("java.sql.Timestamp",
             null,
             "com.klg.jclass.cell.editors.JCSqlTimestampCellEditor",
             "com.klg.jclass.cell.renderers.JCStringCellRenderer");

    addClass("com.klg.jclass.util.formulae.Expression",
             null,
             "com.klg.jclass.cell.editors.JCStringCellEditor",
             "com.klg.jclass.util.formulae.JCExpressionCellRenderer");

    try {
        addClass(byte[].class,
                 null,
                 Class.forName("com.klg.jclass.cell.editors.JCImageCellEditor"),
                 Class.forName("com.klg.jclass.cell.renderers.JCRawImageCellRenderer"));
    }
    catch (ClassNotFoundException e) {}
}

/**
 * Adds a type-to-editor/renderer mapping to the registry.
 *
 * @param cellClass Class instance representing the cell data type.
 * @param type modifier for mapping.
 * @param editor editor instance to use for editing specified cell data type.
 * @param renderer renderer instance to use for rendering cells containing
 * specified data type.
 * @return Whether the add succeeded
 */
public boolean addClass(Class cellClass, String type, JCCellEditor editor,
						JCCellRenderer renderer)
{
    if (cellClass == null) {
        return(false);
    }
    ClassCells classCells = (ClassCells)classCellsTable.get(cellClass);
    if (classCells == null) {
        classCellsTable.put(cellClass, new ClassCells(cellClass, editor,
		                                              renderer, type));
    }
    else {
        classCells.add(editor, renderer, type);
    }
    return(true);
}

/**
 * Adds a type-to-editor/renderer mapping to the registry.
 *
 * @param cellClass Class instance representing the cell data type.
 * @param type modifier for mapping.
 * @param editor editor Class to use for editing specified cell data type.
 * @param renderer renderer Class to use for rendering cells containing
 * specified data type.
 * @return Whether the add succeeded
 */
public boolean addClass(Class cellClass, String type, Class editor,
						Class renderer)
{
    if (cellClass == null) {
        return(false);
    }
    ClassCells classCells = (ClassCells)classCellsTable.get(cellClass);
    if (classCells == null) {
        classCellsTable.put(cellClass, new ClassCells(cellClass, editor,
		                                              renderer, type));
    }
    else {
        classCells.add(editor, renderer, type);
    }
    return(true);
}

/**
 * Adds a type-to-editor/renderer mapping to the registry.
 * Uses String instances and Class.forName() to load Class instances.
 *
 * @param cellClass String representing the cell data type.
 * @param type modifier for mapping.
 * @param editor editor class string to use for editing specified cell
 * data type.
 * @param renderer renderer class string to use for rendering cells containing
 * specified data type.
 * @return Whether the add succeeded
 */
public boolean addClass(String cellClass, String type, String editor,
						String renderer)
{
    boolean ret = true;
    try {
        ret = addClass(Class.forName(cellClass), type, Class.forName(editor),
		               Class.forName(renderer));
    }
    catch (ClassNotFoundException e) {
       ret = false;
    }
    return(ret);
}

/**
 *  Given a data type and a modifier, retrieve an editor.
 *
 * @param cellClass Class instance representing the cell data type.
 * @param type modifier for mapping.
 * @return cell editor.
 */
public JCCellEditor getCellEditorInstance(Class cellClass, String type)
{
	return(getCellEditorInstance(cellClass, type, true));
}

/**
 *  Given a data type and a modifier, retrieve an editor.
 *
 * @param cellClass Class instance representing the cell data type.
 * @param type modifier for mapping.
 * @param useCache
 * @return cell editor.
 */
public JCCellEditor getCellEditorInstance(Class cellClass, String type,
                                          boolean useCache)
{
    ClassCells classCells = findClass(cellClass);
    JCCellEditor editor = null;

    if (classCells != null) {
        editor = classCells.getEditor(type, useCache);
    }

    if ((parentRegistry != null) && (editor == null)) {
        editor = parentRegistry.getCellEditorInstance(cellClass, type,
		                                              useCache);
    }
    return(editor);
}

/**
 * Given a data type and a modifier, retrieve an editor.
 *
 * @param classString String representing the cell data type.
 * @param type Modifier for mapping.
 * @return The cell editor based on the classString and type
 */
public JCCellEditor getCellEditorInstance(String classString, String type)
{
    JCCellEditor editor = null;
    if (classString == null) {
        classString = "java.lang.Object";
    }
    try {
        editor = getCellEditorInstance(Class.forName(classString), type);
    }
    catch (ClassNotFoundException e) {
        editor = null;
    }
    return(editor);
}

/**
 *  Given a data type and a modifier, retrieve a renderer.
 *
 * @param cellClass Class instance representing the cell data type.
 * @param type Modifier for mapping.
 * @return The cell renderer based on the cellClass and type
 */
public JCCellRenderer getCellRendererInstance(Class cellClass, String type)
{
	return(getCellRendererInstance(cellClass, type, true));
}

/**
 *  Given a data type and a modifier, retrieve a renderer.
 *
 * @param cellClass Class instance representing the cell data type.
 * @param type Modifier for mapping.
 * @param useCache
 * @return The cell renderer based on the cellClass and type
 */
public JCCellRenderer getCellRendererInstance(Class cellClass, String type,
					                          boolean useCache)
{
    ClassCells classCells = findClass(cellClass);
    JCCellRenderer renderer = null;

    if (classCells != null) {
        renderer = classCells.getRenderer(type, useCache);
    }

    if ((parentRegistry != null) && (renderer == null)) {
        renderer = parentRegistry.getCellRendererInstance(cellClass, type,
		                                                  useCache);
    }

    // If this is the centralRegistry then guarantee a renderer is returned
    if ((renderer == null) && (isCentralRegistry)) {
        renderer = new com.klg.jclass.cell.renderers.JCStringCellRenderer();
    }
    return(renderer);
}

/**
 *  Given a data type and a modifier, retrieve a renderer.
 *
 * @param classString String representing the cell data type.
 * @param type Modifier for mapping.
 * @return The cell renderer based on the classString and type
 */
public JCCellRenderer getCellRendererInstance(String classString, String type)
{
    JCCellRenderer renderer = null;
    if (classString == null) {
        classString = "java.lang.Object";
    }
    try {
        renderer = getCellRendererInstance(Class.forName(classString),
		                                   type, true);
    }
    catch (ClassNotFoundException e) {
        renderer = null;
    }
    return(renderer);
}

/**
 * Given a data type and a modifier, gets the name of the associated
 * cell editor
 * @param cellClass Class instance representing the cell data type.
 * @param type Modifier for mapping.
 * @return The cell editor name as a string
 */
public String getCellEditorString(Class cellClass, String type)
{
    ClassCells classCells = findClass(cellClass);
    String editorString = null;

    if (classCells != null) {
        editorString = classCells.getEditorString(type);
    }

    if ((parentRegistry != null) && (editorString == null)) {
        editorString = parentRegistry.getCellEditorString(cellClass, type);
    }
    return(editorString);
}

/**
 * Given a data type and a modifier, gets the name of the associated
 * cell editor.
 * @param classString String representing the cell data type.
 * @param type Modifier for mapping.
 * @return The cell editor name as a string
 */
public String getCellEditorString(String classString, String type)
{
    String editorString = null;
    try {
        editorString = getCellEditorString(Class.forName(classString), type);
    }
    catch (ClassNotFoundException e) {
        editorString = null;
    }
    return(editorString);
}

/**
 * Given a data type and a modifier, gets the name of the associated
 * cell renderer
 * @param cellClass Class instance representing the cell data type.
 * @param type Modifier for mapping.
 * @return The cell renderer name as a string
 */
public String getCellRendererString(Class cellClass, String type)
{
    ClassCells classCells = findClass(cellClass);
    String rendererString = null;

    if (classCells != null) {
        rendererString = classCells.getRendererString(type);
    }

    if ((parentRegistry != null) && (rendererString == null)) {
        rendererString = parentRegistry.getCellRendererString(cellClass, type);
    }
    if ((rendererString == null) && (isCentralRegistry)) {
        rendererString = "com.klg.jclass.cell.renderers.JCStringCellRenderer";
    }
    return(rendererString);
}

/**
 * Given a data type and a modifier, gets the name of the associated
 * cell renderer
 * @param classString String representing the cell data type.
 * @param type modifier for mapping.
 * @return The cell renderer name as a string
 */
public String getCellRendererString(String classString, String type)
{
    String rendererString = null;
    try {
        rendererString = getCellRendererString(Class.forName(classString),
		                                       type);
    }
    catch (ClassNotFoundException e) {
        rendererString = null;
    }
    return(rendererString);
}

/**
 * Look for the class in the table.  If the class cannot be found search for
 * each super class/interface until one is found.
 * @param cellClass
 * @return The ClassCells object that matches the class
 */
protected ClassCells findClass(Class cellClass)
{
    ClassCells classCells = null;

    if (cellClass == null) {
        cellClass = Object.class;
    }

    do {
        classCells = (ClassCells)classCellsTable.get(cellClass);

        // Need to see if this class is indexed by any of its interfaces
        // rather than by its class
        if (classCells == null) {

        	Class interfaces[] = cellClass.getInterfaces();
            for (int i = 0; i < interfaces.length; i++) {
            	classCells = (ClassCells)classCellsTable.get(interfaces[i]);

                // If we found it, break
                if (classCells != null) {
                	break;
                }
                // If not it could be that the interface extends another
                // interface which is the indexed one we are looking for so
                // recursively traverse the inheritence tree. yuk.
                else {
                	classCells = findClass(interfaces[i]);
					// If we found it, break
                    if (classCells != null) {
                	    break;
                    }
                }
            }

        }

        // Check its super class then
        cellClass = cellClass.getSuperclass();
    }
    while((classCells == null) && (cellClass != null));
    return(classCells);
}


/**
 * Inner class used to store Cell objects
 */
protected class ClassCells extends Vector {
    Class cellClass;

    public ClassCells() {}

    public ClassCells(Class cellClass)
	{
        this.cellClass = cellClass;
    }

    public ClassCells(Class cellClass, JCCellEditor editor,
	                  JCCellRenderer renderer, String type)
	{
        this.cellClass = cellClass;
        add(new Cell(editor, renderer, type));
    }

    public ClassCells(Class cellClass, Class editor, Class renderer,
	                  String type)
	{
        this.cellClass = cellClass;
        add(new Cell(editor, renderer, type));
    }

    public Class getCellClass()
	{
        return(cellClass);
    }

    public void setCellClass(Class cellClass)
	{
        this.cellClass = cellClass;
    }

    /* Duplicate types are not allowed so the matching type will be removed from
       the list and the new one will be added. */
    public boolean add(Object editor, Object renderer, String type)
	{
        Cell cell = getCell(type);
        if (cell == null) {
            add(new Cell(editor, renderer, type));
        }
        else {
            remove(cell);
            cell.dispose();
            add(new Cell(editor, renderer, type));
        }
        return(true);
    }

    public Cell getCell(String type)
	{
        Cell cell;
        for (int i = 0; i < size(); i++) {
            cell = (Cell)get(i);
            if (type == null) {
                if (cell.type == null) {
                   return(cell);
                }
            }
            else {
                if ((cell.type != null) && (cell.type.compareTo(type) == 0)) {
                    return(cell);
                }
            }
        }
        return(null);
    }

	public JCCellEditor createEditor(Class c)
	{
		Object editor = null;
		if (c != null) {
            try {
                editor = c.newInstance();
            }
            catch (Exception e) {
                /* Unable to create new instance */
                e.printStackTrace();
            }
		}
		return((JCCellEditor)editor);
	}

    public JCCellEditor getEditor(String type)
	{
		return(getEditor(type, true));
	}

    public JCCellEditor getEditor(String type, boolean useCache)
	{
        Cell cell = getCell(type);

        if (cell == null) {
            return(null);
        }

		if (!useCache) {
			JCCellEditor editor = null;
			if (cell.editor instanceof Class) {
			    editor = createEditor((Class)cell.editor);
			}
			else if (cell.editor != null) {
				editor = createEditor(cell.editor.getClass());
			}
			// If we've created something, return it.  Otherwise fall back
			// to the cached value.
			if (editor != null) {
				return(editor);
			}
		}

		// Create a cached value
        if (useCache && cell.editor instanceof Class) {
			cell.editor = createEditor((Class)cell.editor);
        }
		// Return the cached value
        return((JCCellEditor) cell.editor);
    }

    public String getEditorString(String type)
	{
        String name;
        Cell cell = getCell(type);

        if (cell == null) {
            name = null;
        }
        else if (cell.editor instanceof Class) {
            name = ((Class)cell.editor).getName();
        }
        else {
            name = cell.editor.getClass().getName();
        }
        return(name);
    }

	public JCCellRenderer createRenderer(Class c)
	{
		Object renderer = null;
		if (c != null) {
            try {
                renderer = c.newInstance();
            }
            catch (Exception e) {
                /* Unable to create new instance */
                e.printStackTrace();
            }
		}
		return((JCCellRenderer)renderer);
	}

    public JCCellRenderer getRenderer(String type)
	{
		return(getRenderer(type, true));
    }

	public JCCellRenderer getRenderer(String type, boolean useCache)
	{
        Cell cell = getCell(type);

        if (cell == null) {
            return(null);
        }

		if (!useCache) {
			JCCellRenderer renderer = null;
			if (cell.renderer instanceof Class) {
			    renderer = createRenderer((Class)cell.renderer);
			}
			else if (cell.renderer != null) {
				renderer = createRenderer(cell.renderer.getClass());
			}
			// If we've created something, return it.  Otherwise fall back
			// to the cached value.
			if (renderer != null) {
				return(renderer);
			}
		}

		// Create a cached value
        if (useCache && cell.renderer instanceof Class) {
			cell.renderer = createRenderer((Class)cell.renderer);
        }
		// Return the cached value
        return((JCCellRenderer) cell.renderer);
    }

    public String getRendererString(String type)
	{
        String name;

        Cell cell = getCell(type);

        if (cell == null) {
            name = null;
        }
        else if (cell.renderer instanceof Class) {
            name = ((Class)cell.renderer).getName();
        }
        else {
            name = cell.renderer.getClass().getName();
        }
        return(name);
    }


    public class Cell {
        public Object editor;
        public Object renderer;
        public String type;

        public Cell() {}

        public Cell(Object editor, Object renderer, String type)
		{
            this.editor   = editor;
            this.renderer = renderer;
            this.type     = type;
        }

        public void dispose()
		{
            editor   = null;
            renderer = null;
            type     = null;
        }
    } /* class Items */
} /* class ClassEditors */

}
