/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Utilities.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:41 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.cell;

import java.awt.*;
import java.io.Serializable;

/**
 * Useful utilities for use with Cell Renderers and Editors
 */
public class Utilities implements Serializable {

/**
 * Creates a Rectangle to allow drawing over the whole cell
 * instead of being positioned inside the borders and margins.
 * @see #translateToWholeCell
 */
public static Rectangle getWholeCell(JCCellInfo info, 
                                     Rectangle drawing_area)
{
	Insets margins = info.getMarginInsets();
	Insets borders = info.getBorderInsets();
	return(new Rectangle(-(margins.right + borders.right),
				         -(margins.top + borders.top),
    					 drawing_area.width + margins.right + margins.left +
    					 borders.right + borders.left,
    					 drawing_area.height + margins.top + margins.bottom +
    					 borders.top + borders.bottom));
}

/**
 * Translates the Graphics object to allow drawing over the whole cell
 * instead of being positioned inside the borders and margins.
 * @see #restoreFromWholeCell
 */
public static void translateToWholeCell(Graphics gc, JCCellInfo info,
										Rectangle drawing_area)
{
    Rectangle rect = getWholeCell(info, drawing_area);
	gc.translate(rect.x, rect.y);
	drawing_area.setSize(rect.width, rect.height);
}

/**
 * Reverses any changes made by the <code>translateToWholeCell()</code> 
 * method call.
 * @see #translateToWholeCell
 */
public static void restoreFromWholeCell(Graphics gc, JCCellInfo info,
										Rectangle drawing_area)
{
	Insets margins = info.getMarginInsets();
	Insets borders = info.getBorderInsets();
	gc.translate(margins.right + borders.right,
				 margins.top + borders.top);
	drawing_area.setSize(drawing_area.width - margins.right - margins.left -
					     borders.right - borders.left,
					     drawing_area.height - margins.top - margins.bottom -
					     borders.top - borders.bottom);
}

/**
 * Translates the Graphics object to allow drawing inside the borders
 * instead of being positioned inside the borders and margins.
 * @param drawing_area rectangle representing current drawing area.
 *                     Will be modified to to the new extended drawing
 *                     width and height, but the x and y will not be changed
 *                     accordingly.
 * @see #restoreFromInsideBorders
 */
public static void translateToInsideBorders(Graphics gc, JCCellInfo info,
											Rectangle drawing_area)
{
	Insets margins = info.getMarginInsets();
	gc.translate(-margins.right, -margins.top);
	drawing_area.setSize(drawing_area.width + margins.right + margins.left,
					     drawing_area.height + margins.top + margins.bottom);
}

/**
 * Reverses any changes made by the <code>translateToInsideBorders()</code>
 * method call.
 * @param drawing_area rectangle representing current drawing area.
 *                     Will be modified to to the new reduced drawing
 *                     width and height, but the x and y will not be changed
 *                     accordingly.
 * @see #translateToInsideBorders
 */
public static void restoreFromInsideBorders(Graphics gc, JCCellInfo info,
											Rectangle drawing_area)
{
	Insets margins = info.getMarginInsets();
	gc.translate(margins.right, margins.top);
	drawing_area.setSize(
					 drawing_area.width - margins.right - margins.left,
					 drawing_area.height - margins.top - margins.bottom);
}

/** Draws the top and left lines. */
public static void drawTopLines(Graphics g, int size, int x, int y,
						        int w, int h, Color color)
{
	g.setColor(color);
	for (int i =0; i < size; i++) {
		g.drawLine(x + i, y + i,     x + w - (i + 1), y + i);
		g.drawLine(x + i, y + i + 1, x + i,           y + h - (i + 1));
	}
}
 
/** Draws the bottom and right lines. */
public static void drawBottomLines(Graphics g, int size, int x, int y,
							       int w, int h, Color color)
{
	g.setColor(color);
	for (int i = 1; i <= size; i++) {
		g.drawLine(x + i - 1, y + h - i, x + w - i, y + h - i);
		g.drawLine(x + w - i, y + i - 1, x + w - i, y + h - i);
	}
}

public static void drawNormal(Graphics g, int size, int x, int y, int w, int h,
					          Color top, Color bottom)
{
	drawTopLines(g, size, x, y, w, h, top);
	drawBottomLines(g, size, x, y, w, h, bottom);
}

private static final Color WHITE_BRIGHTER = new Color(200, 200, 200);
private static final Color WHITE_DARKER   = new Color(140, 140, 140);
private static final Color BLACK_BRIGHTER = new Color(125, 125, 125);
private static final Color BLACK_DARKER   = new Color(75, 75, 75);
private static final double FACTOR        = 0.5;

/**
 * Calculates the color even when it is saturated, for example,
 * when it is black or white.
 */
public static Color brighter(Color color)
{
	if (color.equals(Color.white)) {
		return (WHITE_BRIGHTER);
    }
	else if (color.equals(Color.black)) {
		return(BLACK_BRIGHTER);
    }
	else {
        // Instead of calling Color.brighter() I do my own calculation
        // for a lighter color.  This calc brightens up dark colors very
        // quickly while lightening up bright colors only a bit.
        int red = color.getRed();
        red += (int) ((255 - red) * FACTOR);
        int blue = color.getBlue();
        blue += (int) ((255 - blue) * FACTOR);
        int green = color.getGreen();
        green += (int) ((255 - green) * FACTOR);

        return(new Color(Math.min(red, 255),
						 Math.min(green, 255),
						 Math.min(blue, 255)));
	}
}

/**
 * Calculates the color even when it is saturated, for example,
 * when it is black or white.
 */
public static Color darker(Color color)
{
	if (color.equals(Color.white)) {
		return(WHITE_DARKER);
    }
	else if (color.equals(Color.black)) {
		return(BLACK_DARKER);
    }
	else {
		return(color.darker());
    }
}

// Border Styles

/** No border */
public static final int NONE        = 0;
/** Border drawn with double line; border appears inset */
public static final int ETCHED_IN   = 1;
/** Border drawn with double line; border appears raised */
public static final int ETCHED_OUT  = 2;
/** Border drawn appears inset */
public static final int IN          = 3;
/** Border drawn appears raised */
public static final int OUT         = 4;
/** Border drawn in foreground color */
public static final int PLAIN       = 5;
/** Border drawn with 1-pixel shadow-in at edge; border appears framed */
public static final int FRAME_IN    = 6;
/** Border drawn with 1-pixel shadow-out at edge; border appears framed */
public static final int FRAME_OUT   = 7;
/** Border drawn with MS-Windows style control shadows */
public static final int CONTROL_IN  = 8;
/** Border drawn with MS-Windows style control shadows */
public static final int CONTROL_OUT = 9;
/** Number of different border styles */
public static final int NUM_BORDER_STYLES = 10;

/**
 * Draws a border
 * @param type (see next <CODE>draw()</CODE> definition for full disclosure)
 * @param size the shadow thickness
 * @param bg the background color
 * @param plain the color used to draw the sides of PLAIN shadows
 */
public static void drawBorder(Graphics gc, int type, int size, 
							  int x, int y, int width, int height,
							  Color bg, Color plain)
{
	drawBorder(gc, type, size, x, y, width, height, 
		       brighter(bg), darker(bg), plain);
}

/**
 * Draws a border
 * @param type one of the following members:
 * <pre>
 *   ETCHED_IN   double line; border appears inset
 *   ETCHED_OUT  double line; border appears raised
 *   FRAME_IN    1-pixel shadow-in at edge; border appears framed
 *   FRAME_OUT   1-pixel shadow-out at edge; border appears framed
 *   IN          border appears inset
 *   OUT         border appears raised
 *   CONTROL_IN  MS-Windows control shadows
 *   CONTROL_OUT MS-Windows control shadows
 *   PLAIN       border drawn in foreground color
 *   NONE        no border drawn
 * </pre>
 * @param size the shadow thickness
 * @param bright bright shadow color
 * @param dark dark shadow color
 * @param plain the color used to draw the sides of PLAIN shadows
 */
public static void drawBorder(Graphics gc, int type, int size, 
						      int x, int y, int width, int height,
						      Color bright, Color dark, Color plain)
{
	Color old_color = gc.getColor();

    int r2, s2, s3;
	switch (type) {
	case PLAIN:
		drawNormal(gc, size, x, y, width, height, plain, plain);
		break;
	case IN:
		drawNormal(gc, size, x, y, width, height, dark, bright);
		break;
	case OUT:
		drawNormal(gc, size, x, y, width, height, bright, dark);
		break;
	case CONTROL_IN:
		if (size == 2) {
			drawNormal(gc, 1, x + 1, y + 1, width - 2, height - 2, Color.black,
                       bright);
			drawNormal(gc, 1, x, y, width, height, dark, brighter(bright));
		}
		else {
			drawNormal(gc, size, x, y, width, height, dark, bright);
		}
		break;
	case CONTROL_OUT:
		if (size == 2) {
			drawNormal(gc, 1, x, y, width, height, brighter(bright), dark);
			drawNormal(gc, 1, x + 1, y + 1, width - 2, height - 2, bright,
                       Color.black);
		}
		else {
			drawNormal(gc, size, x, y, width, height, bright, dark);
		}
		break;
	case ETCHED_IN:
		r2 = size % 2;
        s2 = (r2 == 0) ? size/2 : size/2 + 1;
        s3 = size-s2;
		drawNormal(gc, s2, x, y, width, height, dark, bright);
		drawNormal(gc, s3, x + s2, y + s2, width - size - r2,
                   height - size - r2, bright, dark);
	    break;
	case ETCHED_OUT:
		r2 = size % 2;
        s2 = (r2 == 0) ? size/2 : size/2+1;
        s3 = size-s2;
		drawNormal(gc, s2, x, y, width, height, bright, dark);
		drawNormal(gc, s3, x + s2, y + s2, width - size - r2,
                   height - size - r2, dark, bright);
	    break;
	case FRAME_OUT:
		// 1-pixel shadow
		drawNormal(gc, size, x, y, width, height, bright, dark);
		// plain
		drawNormal(gc, size - 1, x, y, width, height, plain, plain);
		break;
	case FRAME_IN:
		// 1-pixel shadow
		drawNormal(gc, size, x, y, width, height, dark, bright);
		// plain
		drawNormal(gc, size - 1, x, y, width, height, plain, plain);
		break;
	default:
		break;
	}
	gc.setColor(old_color);
}

/**
 * Default size of clip arrows used by cell renderers.
 */
public static final int DEFAULT_CLIP_ARROW_SIZE = 4;

/**
 * Draws horizontal and vertical clip arrows if requested to do so 
 * by the JCCellInfo object.
 * @param gc The graphics context
 * @param info The JCCellInfo object for this cell
 * @param preferredSize The size that determines whether arrows are needed
 * @param size The number of pixels along a side of the arrow
 * @param drawOutline Should an outline be drawn, as well as a plygon filled
 */
public static void drawClipArrows(Graphics gc, JCCellInfo info,
                                  Dimension preferredSize, int size,
                                  boolean drawOutline)
{
    int clipHints = info.getClipHints();
    Rectangle originalDrawingArea = info.getDrawingArea();
    Rectangle originalClipRect = gc.getClipBounds();

    Rectangle rect = new Rectangle(originalDrawingArea);

	// We're not interested in the drawing area x and y -- we're going from
	// the top corner of the clip
	rect.x = 0;
	rect.y = 0;
    translateToInsideBorders(gc, info, rect);

	// Check to see if original clip rect for drawing area was indeed clipped
	// (e.g. if the cell is scrolling off of the bottom of the table)!
	// If so, reproduce clipping here. otherwise arrows will try to draw
	// outside of cell.
	Rectangle newClipRect = new Rectangle(rect);
	if (originalClipRect.x > 0) {
		// Clipped on left hand side, use original clip x
		newClipRect.x = originalClipRect.x;
		if (originalClipRect.width + originalClipRect.x != 
									originalDrawingArea.width)
        {
		    // Clipped on the right, too? is so, use that width
			newClipRect.width = originalClipRect.width;
		}
        else {
		    // If only clipped on left, remove the clipped x
		    // value from the existing width
			newClipRect.width -= newClipRect.x;
		}
	}
    else if (originalClipRect.width != originalDrawingArea.width) {
		// Clipped on right hand side, add left margin to clipped width
		Insets margins = info.getMarginInsets();
		newClipRect.width = originalClipRect.width + margins.left;
	}
	if (originalClipRect.y > 0) {
		// clipped on top hand side, use original clip y
		newClipRect.y = originalClipRect.y;
		if (originalClipRect.height + originalClipRect.y !=
										originalDrawingArea.height)
        {
		    // Clipped on the bottom, too? is so, use that height
			newClipRect.height = originalClipRect.height;
		}
        else {
		    // If only clipped on top, remove the clipped y
		    // value from the existing height
			newClipRect.height -= newClipRect.y;
		}
	}
    else if (originalClipRect.height != originalDrawingArea.height) {
		// clipped on bottom hand side, add top margin to clipped height
		Insets margins = info.getMarginInsets();
		newClipRect.height = originalClipRect.height + margins.top;
	}
	
    gc.setClip(newClipRect);

    // Check if we need to draw horizontal arrows
    if ((clipHints == JCCellInfo.SHOW_HORIZONTAL) ||
        (clipHints == JCCellInfo.SHOW_ALL))
    {
        if (preferredSize.width > originalDrawingArea.width) {
            int y;
            switch(info.getVerticalAlignment()) {
            default:
            case JCCellInfo.TOP:
                y = rect.height - size;
                break;
            case JCCellInfo.CENTER:
                y = (rect.height + size) / 2;
                break;
            case JCCellInfo.BOTTOM:
                y = size * 2;
                break;
            }
            int alignment = info.getHorizontalAlignment();
            if (alignment == JCCellInfo.CENTER) {
                drawHorizontalArrow(gc, rect, y, JCCellInfo.LEFT, size,
                                    drawOutline);
                drawHorizontalArrow(gc, rect, y, JCCellInfo.RIGHT, size,
                                    drawOutline);
            }
            else {
                drawHorizontalArrow(gc, rect, y, alignment, size, drawOutline);
            }
        }
    }
    // Check if we need to draw vertical arrows
    if ((clipHints == JCCellInfo.SHOW_VERTICAL) ||
        (clipHints == JCCellInfo.SHOW_ALL))
    {
        if (preferredSize.height > originalDrawingArea.height) {
            int x;
            switch(info.getHorizontalAlignment()) {
            default:
            case JCCellInfo.LEFT:
                x = rect.width - size;
                break;
            case JCCellInfo.CENTER:
                x = (rect.width + size) / 2;
                break;
            case JCCellInfo.RIGHT:
                x = size * 2;
                break;
            }
            int alignment = info.getVerticalAlignment();
            if (alignment == JCCellInfo.CENTER) {
                drawVerticalArrow(gc, rect, x, JCCellInfo.TOP, size, drawOutline);
                drawVerticalArrow(gc, rect, x, JCCellInfo.BOTTOM, size, drawOutline);
            }
            else {
                drawVerticalArrow(gc, rect, x, alignment, size, drawOutline);
            }
        }
    }
    restoreFromInsideBorders(gc, info, rect);
    gc.setClip(originalClipRect);
}

static int xpoints[] = new int[3];
static int ypoints[] = new int[3];

/*
 * Draw a horizontal arrow based on the horizontal alignment
 */
static void drawHorizontalArrow(Graphics gc, Rectangle rect, int y,
                                int alignment, int size, boolean drawOutline)
{
    int x;
	switch (alignment) {
	case JCCellInfo.LEFT:
        // draw right arrow
        x = rect.width - 1;
    	xpoints[0] = x - size;
    	ypoints[0] = y - size;
		xpoints[1] = xpoints[0];
		ypoints[1] = ypoints[0] + size;
		xpoints[2] = x;
		ypoints[2] = y - size/2;
		break;
	case JCCellInfo.RIGHT:
        // draw left arrow
        x = rect.x;
    	xpoints[0] = x + size;
    	ypoints[0] = y - size;
		xpoints[1] = xpoints[0];
		ypoints[1] = ypoints[0] + size;
		xpoints[2] = x;
		ypoints[2] = y - size/2;
		break;
	default:
		return;
	}
	drawPolygon(gc, drawOutline);
}

/*
 * Draw a vertical arrow based on the vertical alignment
 */
static void drawVerticalArrow(Graphics gc, Rectangle rect, int x,
                              int alignment, int size, boolean drawOutline)
{
    int y;
	switch (alignment) {
	case JCCellInfo.TOP:
        // draw bottom arrow
        y = rect.height - 1;
    	xpoints[0] = x - size;
    	ypoints[0] = y - size;
		xpoints[1] = xpoints[0] + size;
		ypoints[1] = ypoints[0];
		xpoints[2] = x - size/2;
		ypoints[2] = y;
		break;
	case JCCellInfo.BOTTOM:
        // draw top arrow
        y = rect.y;
    	xpoints[0] = x - size;
    	ypoints[0] = y + size;
		xpoints[1] = xpoints[0] + size;
		ypoints[1] = ypoints[0];
		xpoints[2] = x - size/2;
		ypoints[2] = y;
		break;
	default:
		return;
	}
	drawPolygon(gc, drawOutline);
}

/*
 * Helper method to minimize code duplication
 */
static void drawPolygon(Graphics gc, boolean drawOutline)
{
	if (drawOutline) {
        gc.drawLine(xpoints[0], ypoints[0], xpoints[1], ypoints[1]);
        gc.drawLine(xpoints[1], ypoints[1], xpoints[2], ypoints[2]);
        gc.drawLine(xpoints[2], ypoints[2], xpoints[0], ypoints[0]);
    }
	gc.fillPolygon(xpoints, ypoints, 3);
}

/*
 * Method to add a new key to a list of keys and return a new (bigger) array
 */
public static JCKeyModifier[] addKey(JCKeyModifier keys[], 
									 JCKeyModifier new_key)
{
	JCKeyModifier new_keys[] = new JCKeyModifier[keys.length + 1];
	System.arraycopy(keys, 0, new_keys, 0, keys.length);
	new_keys[keys.length] = new_key;
	return(new_keys);
}

}
