/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseCellEditor.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.cell.editors;

import com.klg.jclass.cell.*;
import com.klg.jclass.cell.validate.JCValidateEvent;
import com.klg.jclass.cell.validate.JCValidateListener;
import com.klg.jclass.cell.validate.ValidateSupport;

import com.klg.jclass.util.JCEnvironment;

import javax.swing.FocusManager;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * <CODE>BaseCellEditor</CODE> provides a simple text editing component.
 */
public class BaseCellEditor extends JTextField
			                implements JCCellTextEditor,
									   JCValidateInterface,
									   KeyListener,
									   MouseListener,
									   FocusListener
{

/**
 * This editor can dynamically alter its reserved key list based upon
 * the key the user has last typed.  Thus, typing navigation keys will
 * cause the cell editor to traverse, but after typing a non-navigation
 * key, the user is placed into edit mode.  At that point, the reserved
 * keys are re-enabled and the traversal keys will traverse within the
 * editor itself.  The only way to get out of edit mode is to press
 * Enter, Escape, or use the mouse to click on another cell.
*/

protected Object               data;
protected int                  caretPosition;
protected AWTEvent             initialEvent;
protected JCCellEditorSupport  support = new JCCellEditorSupport();
protected boolean              selectAll = false;
protected JCKeyModifier[]      reservedKeys = null;
protected ValidateSupport      validate_support = new ValidateSupport();

/**
 * Default constructor
 */
public BaseCellEditor()
{
	addKeyListener(this);
	addMouseListener(this);
	addFocusListener(this);
	setDoubleBuffered(true);
	if (reservedKeys == null) {
		reservedKeys = new JCKeyModifier[4];
		reservedKeys[0] = new JCKeyModifier(KeyEvent.VK_LEFT);
		reservedKeys[1] = new JCKeyModifier(KeyEvent.VK_RIGHT);
		reservedKeys[2] = new JCKeyModifier(KeyEvent.VK_HOME);
		reservedKeys[3] = new JCKeyModifier(KeyEvent.VK_END);
	}
}


public void initialize(AWTEvent ev, JCCellInfo info, Object o)
{
	
	this.initialEvent = ev;

	data = o;
    selectAll = info.getSelectAll();
    setEditable(info.isEditable());
    if (!isEditable()) {
        selectAll = false;
    }

	if (ev instanceof KeyEvent && data != null) {
		caretPosition = data.toString().length();
	}
	else {
        // Will be handled by the show() method later on
		caretPosition = -1;
	}

    initData();
    selectAllText();

    setBackground(info.getBackground());
    setForeground(info.getForeground());
    if (info.isEditable() && (info.getBackground() != null)) {
    	setBackground(info.getBackground().brighter());
    }
    setSelectedTextColor(SystemColor.textHighlightText);
    setSelectionColor(SystemColor.textHighlight);
    setFont(info.getFont());
}

/**
 * Returns whether this editor will select the text on initialization
 * @return The value of selectAll
 */
public boolean getSelectAll()
{
	return(selectAll);
}

/**
 * Returns the edit component.
 * @return The current edit component
 */
public Component getComponent()
{
	return(this);
}

/**
 * Returns the object to be commited.
 * @return The current cell editor value
 */
public Object getCellEditorValue()
{
    return(getCellEditorText());
}

/**
 * Gets the text for the editor (as a String)
 * @return The current text in the editor
 */
public String getCellEditorText()
{
	return(getText());
}

/**
 * Sets the text for the editor
 * @param text The new cell editor text value
 */
public void setCellEditorText(String text)
{
	caretPosition = (text == null) ? 0 : text.length();
    setText(text);
}

/**
 * Returns true if the data has been modified.
 * @return Has the data been modified?
 */
public boolean isModified()
{
	

	Object currentValue = getCellEditorValue();
	if ((data == null) || (data.equals(""))) {
	    if ((currentValue == null) || (currentValue.equals(""))) {
	        return(false);
	    }
	    else {
	        return(true);
	    }
	}
	return(!data.equals(currentValue));
}

protected boolean fireValidated(JCValidateEvent ev)
{
	boolean validated = validate_support.fireValidateEvents(ev);
	if (validated) {
	    setVisible(true);
	}
	return(validated);
}

/**
 * Called when the edit process is stopped
 * @return true if current cell editor value passes validation
 */
public boolean stopCellEditing()
{
	
	String str = getText();
	JCValidateEvent ev = new JCValidateEvent(this, data, str, true);
	// All values are considered valid but someone registering
	// a ValidateListener can decide otherwise in the valueChangedEnd()
	// event
	return(fireValidated(ev));
}

/**
 * Called when the edit process is cancelled.
 */
public void cancelCellEditing()
{
	
    initData();
    caretPosition = (data == null) ? 0 : data.toString().length();
}

public Dimension getPreferredSize(Graphics gc, JCCellInfo cellInfo, Object o)
{
	return(getPreferredSize());
}

/**
 * This editor reserves the LEFT and RIGHT arrow keys for its own use.
 * However, after the edit comitted or cancelled, it temporarily
 * "unreserves" these keys.
 */
public JCKeyModifier[] getReservedKeys()
{
	return(reservedKeys);
}

/**
 * Overrides <CODE>setVisible</CODE> in order to control the caret
 * position.
 */
public void setVisible(boolean b)
{
	super.setVisible(b);
	if (!b) {
		return;
	}

    // This code attempts to position the caret based
    // on where the user clicked the mouse within this cell
	String text = getText();
	if (text == null) {
		return;
	}
	if (caretPosition == -1) {
        caretPosition = text.length();
        if (initialEvent != null && initialEvent instanceof MouseEvent) {
            MouseEvent me = (MouseEvent)initialEvent;
            Component source = (Component)me.getSource();
            Component comp = getComponent();
            Point offset = new Point(0, 0);
            if (comp != source) {
                Container parent = getParent();
                offset.setLocation(comp.getLocation());
                while (parent != null && parent != source) {
                    offset.translate(parent.getLocation().x,
                                     parent.getLocation().y);
                    comp = parent;
                    parent = comp.getParent();
                }
                if (parent == null) {
                    offset.setLocation(0, 0);
                }
            }
            caretPosition = viewToModel(new Point(me.getX() - offset.x,
                                                  me.getY() - offset.y));
        }
	}
    caretPosition = Math.min(caretPosition, text.length());
    select(caretPosition, caretPosition);
	selectAllText();
}

/**
 * Add a listener to the list that's notified when the editor starts,
 * stops, or cancels editing.
 *
 * @param l	The CellEditorListener
 */
public void addCellEditorListener(JCCellEditorListener l)
{
	support.addCellEditorListener(l);
}

/**
 * Remove a listener from the list that's notified
 *
 * @param l	The CellEditorListener
 */
public void removeCellEditorListener(JCCellEditorListener l)
{
	support.removeCellEditorListener(l);
}

/**
 * Override the key event processor.  This allows the parent to pass a key
 * event to this editor that it might otherwise not get
 * @param e The event to pass
 */
public void processKeyEvent(KeyEvent e)
{
	
	if (isVisible()) {
	    if (JCEnvironment.getJavaVersion() >= 140) {
		    super.processKeyEvent(e);
	    }
		else {
            // In Swing versions prior to JKD1.4, we have to disable the focus
            // manager in order to get the key processing we want
		    FocusManager fm = FocusManager.getCurrentManager();
		    FocusManager.disableSwingFocusManager();
		    super.processKeyEvent(e);
		    fm.setCurrentManager(fm);
	    }
	}
}

public void keyPressed(KeyEvent e)
{
	
    int keyCode = e.getKeyCode();
	switch(keyCode) {
	case KeyEvent.VK_ENTER:
		support.fireStopEditing(new JCCellEditorEvent(e));
		break;
	case KeyEvent.VK_ESCAPE:
		support.fireCancelEditing(new JCCellEditorEvent(e));
		break;
	default:
	    break;
	}
}

public void mouseClicked(MouseEvent e) {}

public void mouseEntered(MouseEvent e)
{
    setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
}

public void mouseExited(MouseEvent e) {}

public void mousePressed(MouseEvent e) {}

public void mouseReleased(MouseEvent e) {}

public void keyReleased(KeyEvent e) {}

public void keyTyped(KeyEvent e) {}

protected void initData()
{
	String text = (data == null) ? "" : data.toString();
    setText(text);
}

/**
 * Uses hint from CellInfo to select all the text in the component.
 */
public void selectAllText()
{
    if (getSelectAll() && isDisplayable()) {

        // Setting of the caret position is a hack to ensure that the
        // text is fully scrolled into view
        try {
            setCaretPosition(0);
            setCaretPosition((data == null) ? 0 : data.toString().length());
        }
        catch (Exception e) {
            // Checking for non-null peer above doesn't always work
        }

        selectAll();
    }
}

//
// ValidateInterface Implementation
//
/**
 * Add a new ValidateListener to the list.
 */
public void addValidateListener(JCValidateListener l)
{
	validate_support.addValidateListener(l);
}

/**
 * Remove a ValidateListener from the list.
 */
public void removeValidateListener(JCValidateListener l)
{
	validate_support.removeValidateListener(l);
}

public void requestFocus()
{
	
	super.requestFocus();
}

public void focusGained(FocusEvent event)
{
	
}

public void focusLost(FocusEvent event)
{
    
}

}
