/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCImageCellEditor.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.cell.editors;

import com.klg.jclass.cell.*;
import com.klg.jclass.util.swing.JCIconCreator;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.io.FileInputStream;

/**
 * <CODE>JCImageCellEditor</CODE> provides an editor using a simple text
 * field for <CODE>Image</CODE> objects.
 */
public class JCImageCellEditor extends JComponent
                               implements JCCellEditor,
                                          KeyListener,
                                          MouseListener,
                                          FocusListener
{

transient protected Image      data;
transient protected Image      originalData;
protected JCCellInfo           cellInfo;
protected JCCellEditorSupport  support = new JCCellEditorSupport();
protected boolean              isFullSize = true;
protected boolean              modified = false;
protected boolean              inDialog = false;
protected boolean              initialEntry = false;
protected boolean              contextPending = false;
protected JCKeyModifier[]      keys = null;
transient protected byte[]     dataBytes = null;
transient MediaTracker         tracker = null;

protected static final int BORDER_SIZE = 2;

/**
 * Default constructor.
 */
public JCImageCellEditor()
{
	super();
	addFocusListener(this);
	addKeyListener(this);
	addMouseListener(this);
	tracker = new MediaTracker(this);
}

public void addNotify()
{
	super.addNotify();
	getParent().addMouseListener(this);
}

public void removeNotify()
{
	getParent().removeMouseListener(this);
	super.removeNotify();
}

public void initialize(AWTEvent ev, JCCellInfo info,  Object o)
{
    modified = false;
    isFullSize = true;
    cellInfo = info;
    readData(o);
    originalData = data;
}

/**
 * Overridden to make the image full size.
 * @param height
 * @param width
 * @param x
 * @param y
 */
public void setBounds(int x, int y, int width, int height)
{
    Dimension d = getPreferredSize(getGraphics(), cellInfo, data);
    if ((width < (d.width + (2 * BORDER_SIZE))) ||
		(height < (d.height + (2 * BORDER_SIZE))))
    {
        super.setBounds(x, y, d.width + (2 * BORDER_SIZE),
						d.height + (2 * BORDER_SIZE));
    }
    else {
        super.setBounds(x, y, width, height);
    }
}

public void paintComponent(Graphics gc)
{
	super.paintComponent(gc);
    Rectangle rect = getBounds();
  	gc.setColor(cellInfo.isEditable() ? cellInfo.getBackground().brighter()
  	                                  : cellInfo.getBackground());
    if (data == null) {
        gc.fillRect(0, 0, rect.width, rect.height);
    }
    else {
    	if (isFullSize) {
        	gc.drawImage(data, BORDER_SIZE, BORDER_SIZE, gc.getColor(), this);
        }
        else {
            Rectangle drawingArea = cellInfo.getDrawingArea();
        	gc.drawImage(data, (rect.width - drawingArea.width) / 2,
        	                   (rect.height - drawingArea.height) / 2,
        	                   drawingArea.width,
        	                   drawingArea.height,
        	                   gc.getColor(), this);
        }
    }
    Utilities.drawBorder(gc, Utilities.CONTROL_IN, BORDER_SIZE, 0, 0,
						 rect.width, rect.height, Color.black, Color.white);
}

void readData(Object o)
{
	data = null;
    if ((o != null) && !(o instanceof Image)) {
		if (!(o instanceof byte[])) {
			return;
		}
        data = JCIconCreator.getIcon((byte[])o).getImage();
    }
    else {
        data = (Image)o;
    }
}

public boolean isFocusTraversable()
{
    return(true);
}

/**
 * Returns the object to be committed.
 * @return object to be committed
 */
public Object getCellEditorValue()
{
    if (isModified()) {
        return(dataBytes);
    }
    else {
        return(data);
    }
}

/**
 * Returns the edit component.
 * @return edit component
 */
public Component getComponent()
{
	return(this);
}

/**
 * Returns <code>true</code> if the data has been modified.
 * @return <code>true</code> if the data has been modified
 */
public boolean isModified()
{
    return(modified);
}

/**
 * Called when the edit process is stopped.
 * @return
 */
public boolean stopCellEditing()
{
	return(true);
}

/**
 * Called when the edit process is cancelled.
 */
public void cancelCellEditing()
{
    data = originalData;
    toggleSize(true);
    repaint();
    modified = false;
}

public Dimension getPreferredSize(Graphics gc, JCCellInfo cellInfo, Object o)
{
    if ((data == null) || (!isFullSize)) {
		if (cellInfo == null) {
            return(new Dimension(0,0));
        }
        Rectangle rect = cellInfo.getDrawingArea();
    	return(new Dimension(rect.width, rect.height));
    }
    else {
        waitForImage(data);
    	return(new Dimension(data.getWidth(null), data.getHeight(null)));
    }
}

/**
 * This editor reserves no keys for its own use.
 * @return
 */
public JCKeyModifier[] getReservedKeys()
{
	return(keys);
}


/**
 * Adds a listener to the list that's notified when the editor starts,
 * stops, or cancels editing.
 *
 * @param l the <code>JCCellEditorListener</code>
 */
public void addCellEditorListener(JCCellEditorListener l) {
	support.addCellEditorListener(l);
}

/**
 * Removes a listener from the list that's notified when the editor starts,
 * stops, or cancels editing.
 *
 * @param l the <code>JCCellEditorListener</code>
 */
public void removeCellEditorListener(JCCellEditorListener l)
{
	support.removeCellEditorListener(l);
}

boolean waitForImage(Image image)
{
	try {
		tracker.addImage(image, 0);
		tracker.waitForID(0);
	}
    catch(Exception e) {
		return(false);
	}
	return(!tracker.isErrorAny());
}

protected void toggleSize()
{
    isFullSize = !isFullSize;
    changeSize();
}

protected void toggleSize(boolean fullSize)
{
    isFullSize = fullSize;
    changeSize();
}

protected void changeSize()
{
	Rectangle rectWhole = Utilities.getWholeCell(cellInfo,
	                                             cellInfo.getDrawingArea());
    Rectangle rectBounds = getBounds();
    setBounds(rectBounds.x, rectBounds.y, rectWhole.width, rectWhole.height);
}

public void keyPressed(KeyEvent e)
{
	switch(e.getKeyCode()) {
	case KeyEvent.VK_ENTER:
        toggleSize();
		support.fireStopEditing(new JCCellEditorEvent(e));
		break;
	case KeyEvent.VK_ESCAPE:
        toggleSize(false);
		support.fireCancelEditing(new JCCellEditorEvent(e));
		break;
	case KeyEvent.VK_SPACE:
        toggleSize();
		break;
	case KeyEvent.VK_L: // LOAD
	case KeyEvent.VK_O: // OPEN
        loadNewImage();
		break;
	default:
//        if (!e.isActionKey() && Character.isLetterOrDigit(e.getKeyChar())) {
//            toggleSize();
//    	}
	    break;
	}
}

public void keyReleased(KeyEvent e) {}

public void keyTyped(KeyEvent e) {}

public void focusGained(FocusEvent e)
{
    if (!inDialog) {
        initialEntry = isFullSize;
        if (!isFullSize) {
            toggleSize(true);
        }
    }
}

public void focusLost(FocusEvent e)
{
    if (!inDialog) {
        toggleSize(false);
    }
}

static Frame getFrame(Component comp)
{
    while ((comp = comp.getParent()) != null) {
        if (comp instanceof Frame) {
            return((Frame)comp);
        }
    }
    return(null);
}

protected void loadNewImage()
{
    FileDialog f = new FileDialog(getFrame(this));
    f.setMode(FileDialog.LOAD);
    inDialog = true;
    f.show();
    requestFocus();
    inDialog = false;
    if (f.getFile() == null) {
        return;
    }
    String pathName = f.getDirectory() + f.getFile();
    dataBytes = null;
    try {
        FileInputStream in = new FileInputStream(new File(pathName));
        int count = in.available();
        if (count > 0) {
            dataBytes = new byte[count];
        }
        int bytesRead = in.read(dataBytes);
        in.close();
        if (bytesRead != count) {
            dataBytes = null;
        }
    }
    catch (Exception e) {
        dataBytes = null;
        System.out.println(e.toString());
    }
    if (dataBytes != null) {
        data = Toolkit.getDefaultToolkit().createImage(dataBytes);
        modified = true;
        readData(data);
        toggleSize(true);
        repaint();
    }
}

public void mousePressed(MouseEvent e)
{
    if (e.getSource() == this) {
        if (e.isPopupTrigger()) {
            contextPending = true;
        }
    }
    else {
        toggleSize(false);
        initialEntry = false;
    }
}

public void mouseReleased(MouseEvent e)
{
    handleMouseContext(e);
}

void handleMouseContext(MouseEvent e)
{
    if (e.getSource() == this) {
        if (contextPending || e.isPopupTrigger()) {
            loadNewImage();
            contextPending = false;
        }
        else {
            if (initialEntry) {
                initialEntry = false;
            }
            else {
                toggleSize();
            }
        }
    }
}

public void mouseClicked(MouseEvent e) {}

public void mouseEntered(MouseEvent e)
{
    setCursor(Cursor.getDefaultCursor());
}

public void mouseExited(MouseEvent e) {}

}
