/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCStringCellRenderer.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.cell.renderers;

import com.klg.jclass.cell.Utilities;
import com.klg.jclass.cell.JCCellInfo;
import com.klg.jclass.cell.JCLightCellRenderer;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.Graphics;

/**
 * A classic String cell renderer based on the lightweight
 * rendering model.  This renderer handles all text
 * rendering. Special features include multi-line detection
 * and clip arrows.
 * <p>
 * This is the default String cell renderer in JClass.
 * <p>
 * This renderer does not handle HTML rendering.  If you
 * want HTML rendering for String data, use <code>JCLabelCellRenderer</code>.
 * <code>JCLabelCellRenderer</code> is based on a <code>JLabel</code>, which handles
 * HTML rendering.
 *
 * @see com.klg.jclass.cell.renderers.JCLabelCellRenderer
 */
public class JCStringCellRenderer implements JCLightCellRenderer {

protected Utilities  utilities = new Utilities();
protected boolean    drawClipArrowOutline = true;

/**
 * Used to draw the cell. <br><B>Note:</b> The size of the cell comes from
 * <code>cellInfo.getDrawingArea()</code>.
 *
 * @param gc graphics context to use for drawing
 * @param cellInfo cell property information
 * @param o object to be rendered
 * @param selected
 */
public void draw(Graphics gc, JCCellInfo cellInfo, Object o, boolean selected)
{
	if (o == null) {
		return;
    }

	String display = o.toString();
	if (display == null || display.length() == 0) {
		return;
    }

	Rectangle cell_size = cellInfo.getDrawingArea();

 
	gc.setFont(cellInfo.getFont());
	FontMetrics fm = gc.getFontMetrics();
	gc.setColor(selected ? cellInfo.getSelectedForeground()
                         : cellInfo.getForeground());

	if (display.indexOf('\n') != -1) {
		int start = 0, end, h = fm.getHeight();
		int offset_x = 0;

		int height = getHeight(fm, display);
		int offset_y = getAlignedY(h - 4, cell_size.height, height,
                                   cellInfo.getVerticalAlignment());

		int horizAlign = cellInfo.getHorizontalAlignment();
		for (; (end = display.indexOf('\n', start)) != -1;
             start = end+1, offset_y += h)
        {
			String s = display.substring(start, end);
			offset_x = getAlignedX(0, cell_size.width, fm.stringWidth(s),
                                   horizAlign);
			gc.drawString(s, offset_x, offset_y);
		}
		String s = display.substring(start, display.length());
		offset_x = getAlignedX(0, cell_size.width, fm.stringWidth(s),
                               horizAlign);
		gc.drawString(s, offset_x, offset_y);
	}
    else {
		int offset_x = getAlignedX(0, cell_size.width, fm.stringWidth(display),
                                   cellInfo.getHorizontalAlignment());
		int offset_y = getAlignedY(fm.getHeight() - 4, cell_size.height,
                                   fm.getHeight(),
                                   cellInfo.getVerticalAlignment());
		gc.drawString(display, offset_x, offset_y);
	}
    // Arrows are drawn using the current foreground color
	Dimension ps = getPreferredSize(gc, cellInfo, display);
	if ((ps.width > cell_size.width || ps.height > cell_size.height) &&
		cellInfo.getClipHints() != JCCellInfo.SHOW_NONE)
    {
		utilities.drawClipArrows(gc, cellInfo, ps,
								 Utilities.DEFAULT_CLIP_ARROW_SIZE,
								 drawClipArrowOutline);
	}
}

/**
 * Returns an adjusted <i>X</i> position based on the String width and the
 * horizontal alignment.
 * @return adjusted <i>X</i> position based on the String width and the
 * horizontal alignment
 * @param align
 * @param textWidth
 * @param width
 * @param x
 */
public int getAlignedX(int x, int width, int textWidth, int align)
{
	if (align == JCCellInfo.CENTER) {
		return(x + (width - textWidth) / 2);
    }
	else if (align == JCCellInfo.RIGHT) {
		return(x + width - textWidth);
    }
	return(x);
}

/**
 * Returns an adjusted <i>Y</i> position based on the String height and the
 * vertical alignment.
 * @return an adjusted <i>Y</i> position based on the String height and the
 * vertical alignment
 * @param align
 * @param height
 * @param textHeight
 * @param y
 */
public int getAlignedY(int y, int height, int textHeight, int align)
{
	if (align == JCCellInfo.CENTER) {
		return(y + (height - textHeight) / 2);
    }
	else if (align == JCCellInfo.BOTTOM) {
		return(y + height - textHeight);
    }
	return(y);
}

/**
 * Returns the height required to paint the String.
 * @return height required to paing the String
 * @param fm
 * @param text
 */
public int getHeight(FontMetrics fm, String text)
{
	if (text == null || text.length() == 0) {
		return(0);
    }
	int lines = 1;
	for (int i = 0; i < text.length(); i++) {
		if (text.charAt(i) == '\n') {
            lines++;
        }
    }
	return(fm.getHeight() * lines);
}

/**
 * Returns the width required to paint a multi-line String.
 * @return width required to paint a multi-line String
 * @param fm
 * @param text
 */
public int getWidth(FontMetrics fm, String text)
{
	if (text == null || text.length() == 0) {
		return(0);
    }

	// Manually calculate width of multi-line string
	if (text.indexOf('\n') != -1) {
		int start = 0, end, width = 0;
		for (; (end = text.indexOf('\n', start)) != -1; start = end + 1) {
			width = Math.max(width, fm.stringWidth(text.substring(start, end)));
        }
		return(Math.max(width,
                        fm.stringWidth(text.substring(start, text.length()))));
	}
	else {
		return(fm.stringWidth(text));
    }
}

/**
 * Used to determine the preferred size of the cell.
 *
 * @param cellInfo cell property information
 * @return <code>java.awt.Dimension</code> instance containing preferred height and width.
 * @param gc
 * @param o
 */
public Dimension getPreferredSize(Graphics gc, JCCellInfo cellInfo, Object o)
{
	if (o == null) {
		return(new Dimension(0,0));
    }

	String display = o.toString();
	if (display == null || display.length() == 0) {
		return(new Dimension(0,0));
    }

	Font font = (cellInfo.getFont() != null ? cellInfo.getFont()
                                            : gc.getFont());
	FontMetrics fm = gc.getFontMetrics(font);

	if (display.indexOf('\n') != -1) {
		return(new Dimension(getWidth(fm, display), getHeight(fm, display)));
	}
	else {
		return(new Dimension(fm.stringWidth(display), fm.getHeight()));
	}
}

}
