/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: AreaChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:43 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;
import java.util.Vector;

/**
 * Drawable object for area charts.  This object
 * is used for rendering an area chart based on data
 * contained in the <code>dataObject</code>.
 */
public class AreaChartDraw extends ChartDraw {

// Local copies of data needed for draw(), pick(), and unpick().  They are
// set in recalc().
protected Point   firstLast = null;
JCLineStyle       outlineStyle = null;
protected MinMax  xlim = null;
protected MinMax  ylim = null;

// The list of polygons.  Each vector contains a number of AreaPolygon
// instances that make up that series.
protected Vector  polyList[];

final static int XINSIDE     = 1;   /* x lies between (x1, x2) */
final static int XBYX1       = 2;   /* x lies outside x1 */
final static int XBYX2       = 3;   /* x lies outside x2 */

final static int YINSIDE     = 10;  /* y lies within top and bottom line */
final static int YBYTOP      = 11;  /* y lies outside top line */
final static int YBYORIGIN   = 12;  /* y lies by origin */

/**
 * Default constructor.
 */
public AreaChartDraw()
{
	super();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc()
{
	recalc(-1, -1, -1);
}

public void recalc(int startPoint, int endPoint)
{
	recalc(startPoint, endPoint, -1);
}

public void recalc(int startPoint, int endPoint, int seriesIndex)
{
    super.recalc();

	// If fast update mode is set, always cache data
	fastUpdate = dataObject.getFastUpdate();
	if (fastUpdate) {
		cachePlotData = true;
	}
    else {
		cachePlotData = dataObject.getBufferPlotData();
	}

	boolean fu = false;
	// if start point and end point are specified, this is a fast
	// update case
	if (startPoint >= 0 && endPoint >= 0  &&
        endPoint >= startPoint            &&
        polyList != null)
    {
		fu = true;
	}

    // Make a list of points
	// If being called for a particular series, use previously stored series
	// list.  Otherwise, refresh series list.
	if (seriesIndex == -1) {
    	seriesList = dataObject.getSeries();
	}
	if (seriesList == null) {
        return;
    }

    inverted  = dataObject.getXAxis().i_vertical;
    holeValue = dataObject.getHoleValue();
	firstLast = dataObject.getFirstLast();
	xaxis     = dataObject.getXAxis();
	yaxis     = dataObject.getYAxis();

	if (dataObject.getXLimits() == null) {
		return;
	}
	if (dataObject.getYLimits() == null) {
		return;
	}

	xlim = new MinMax(dataObject.getXLimits().min,dataObject.getXLimits().max);
	ylim = new MinMax(dataObject.getYLimits().min,dataObject.getYLimits().max);
	dataObject.getDataBounds(xlim, ylim);

	boolean xlog = xaxis.isLogarithmic();
	boolean ylog = yaxis.isLogarithmic();

	int chartType = dataObject.getChartType();

	// carry on if caching or if not, a valid seriesIndex is specified
	if (!cachePlotData && seriesIndex == -1) {
		if (polyList == null) {
    		polyList = new Vector[seriesList.size()];
        }
		return;
	}

	// Current arrays are reused and expanded in the fast update case
	if (!fu) {
    	polyList = new Vector[seriesList.size()];
	}

   	// Setup parameters to loop thru the data series
	int firstSeries = (seriesIndex == -1) ? 0 : seriesIndex;
	int lastSeries = (seriesIndex == -1) ? seriesList.size() : (seriesIndex+1);
	int polyEndLine[] = null;
	for (int i = firstSeries; i < lastSeries; i++) {
        // Set up the basic information: series, style, x vals, y vals);
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);
		// Re-initialize arrays in normal (i.e. non-fast update) case
		if (!fu) {
			polyList[i] = null;
		}

        // Drop out if data is uninteresting
        if (series.maxXIndex() == 0 || series.maxYIndex() == 0) {
            continue;
        }

		// Re-initialize arrays in normal case; expand arrays in
		// fast update case
		if (!fu) {
			startPoint = series.getFirstPoint();
			endPoint = series.getLastPoint();
			polyList[i] = new Vector();
		}
        else {
			int tmpLength = (polyList[i] == null ? 0 : polyList[i].size());
			if (startPoint < tmpLength) {
                startPoint = tmpLength;
            }
		}
		double yvals[] = getYVals(series);

		// Calculate value of first point.  Use bogus first point if starting
		// a new polygon.
		double x1 = 0.0;
		double y1 = Double.MAX_VALUE;
		boolean inPoly = false;
		int polyNum = 0;
		if (fu && startPoint > 0) {
			x1 = series.getX(startPoint - 1);
			y1 = yvals[startPoint - 1];
			polyNum = findLastPoly(polyList[i]);
			// if previous point is part of last polygon, add new points
			// to existing polygon.
			if (polyNum >= 0) {
				AreaPolygon thisAP = (AreaPolygon)
								polyList[i].elementAt(polyNum);
				if (startPoint-1 >= thisAP.startPoint &&
					startPoint-1 <= thisAP.endPoint)
                {
					inPoly = true;
					thisAP.poly.npoints -= 2;
				}
			}
            else {
			// Otherwise, start a new polygon, either with this point
			// or with the previous point if it is not a hole value
				inPoly = false;
				polyNum++;
				if (x1 == holeValue || y1 == Double.MAX_VALUE) {
					x1 = 0.0;
					y1 = Double.MAX_VALUE;
				}
				AreaPolygon newAP = new AreaPolygon();
				newAP.series = i;
				polyList[i].addElement(newAP);
			}
		}
        else {
			AreaPolygon newAP = new AreaPolygon();
			newAP.series = i;
			polyList[i].addElement(newAP);
		}

		// The top line of this polygon will be stored in this array to reuse
		// later as the bottom line of next polygon.  First initialize with
		// values from previous polygon (in case of holes).  If no
		// previous polygon, initilalize with origin
		int polyTopLine[] = null;
		if (chartType == JCChart.STACKING_AREA) {
			polyTopLine = new int[endPoint - startPoint + 1];
			for (int j = 0; j < polyTopLine.length; j++) {
				if (polyEndLine != null && j < polyEndLine.length) {
					polyTopLine[j] = polyEndLine[j];
				}
                else {
					polyTopLine[j] = yaxis.toPixel(yaxis.origin.value,
                                                   drawFront);
				}
			}
		}

        for (int j = startPoint; j <= endPoint; j++) {
			if (j >= yvals.length) {
				continue;
            }
			double x2 = series.getX(j);
			if (xlog && x2 <= 0.0) {
				x2 = holeValue;
            }
			double y2 = yvals[j];
			if (ylog && y2 <= 0.0) {
				y2 = holeValue;
            }

			// add points of top line to area polygon
			inPoly =
                clipAreaPolygon(x1, y1, x2, y2, j,
							    (AreaPolygon) polyList[i].elementAt(polyNum),
							    inPoly, inverted);

			// we have a full polygon (perhaps due to encountering a hole value)
			// so finish it off and move on to the next polygon.
			if (!inPoly &&
                ((AreaPolygon)polyList[i].elementAt(polyNum)).poly.npoints > 0)
            {
				AreaPolygon thisPoly =
					(AreaPolygon) polyList[i].elementAt(polyNum);
				if (chartType == JCChart.STACKING_AREA) {
					finishAreaPolygon(thisPoly, polyEndLine, polyTopLine,
                                      inverted);
				}
                else {
					finishAreaPolygon(thisPoly.poly, inverted);
				}

				polyNum++;
				AreaPolygon newAP = new AreaPolygon();
				newAP.series = i;
				polyList[i].addElement(newAP);
			}
			x1 = x2;
			y1 = y2;
        }
		// finish off last polygon for this series
		AreaPolygon lastPoly =
			(AreaPolygon) polyList[i].elementAt(polyNum);
		if (lastPoly.poly.npoints > 0) {
			if (chartType == JCChart.STACKING_AREA) {
				finishAreaPolygon(lastPoly, polyEndLine, polyTopLine, inverted);
			} else {
				finishAreaPolygon(lastPoly.poly, inverted);
			}
		} else {
			// last polygon for this series is empty, so make it null
			polyList[i].setElementAt(null, polyNum);
		}
		// use the top line of this polygon as the bottom of the next
		if (chartType == JCChart.STACKING_AREA) {
			polyEndLine = polyTopLine;
		}
    }
}

/**
 * Gets the array of <i>Y</i> values to use.  This can be overridden to return custom
 * <i>Y</i> values (for example, stacking values).
 * Replaces hole values with <code>Double.MAX_VALUE</code>, an undrawable value, to
 * reduce confusion between hole values and valid percentage values used
 * in stacking area charts.
 * @param series the series from which to get the y values
 * @return an array of doubles containing the <i>y</i> values to use
 */
protected double[] getYVals(ChartDataViewSeries series)
{
    double oldyvals[] = series.getY();
    double yvals[] = new double[oldyvals.length];
    for (int i = 0; i < yvals.length; i++) {
        if (oldyvals[i] == holeValue ||
			(yaxis.isLogarithmic() && oldyvals[i] <= 0.0))
        {
            yvals[i] = Double.MAX_VALUE;
        }
        else {
            yvals[i] = oldyvals[i];
        }
    }
    return(yvals);
}

/**
 * Method used to draw related objects.
 * @param gc the graphics context
 */
public void draw(Graphics gc)
{
	draw(gc, -1, -1, true);
}

/**
 * Draw method with start and end points specified for fast update purposes.
 * @param gc the graphics context
 * @param startPoint the point to start drawing at
 * @param endPoint the point to end drawing at
 */
public void draw(Graphics gc, int startPoint, int endPoint)
{
	draw(gc, startPoint, endPoint, true);
}

/**
 * Draw method with direction of drawing specified to control stacking order
 * of polygons.
 * @param gc the graphics context
 * @param startPoint the point to start drawing at
 * @param endPoint the point to end drawing at
 * @param forward if <code>true</code>, draws by going forward through the polygon list;
 *          if <code>false</code>, draws by going backwards through the polygon list
 */
public void draw(Graphics gc, int startPoint, int endPoint, boolean forward)
{
	// Get list of data series to be drawn using this drawable
	if (seriesList == null) {
        return;
    }

    Rectangle clipBounds = gc.getClipBounds();
    if (clipBounds.width == 0 || clipBounds.height == 0) {
        return;
    }

	boolean fu = false;
	// if start point and end point are specified, this is a fast
	// update case
	if (startPoint >= 0 && endPoint >= 0 && endPoint >= startPoint) {
		fu = true;
	}

    if (polyList == null) {
        return;
    }

    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape) gc;
    }

	outlineStyle = dataObject.getOutlineStyle();

    // Must draw each data series in order
    for (int i = forward ? 0 : polyList.length - 1;
		 forward ? (i < polyList.length) : (i >= 0);
		 i += (forward ? 1 : -1) )
    {

		// If not caching plot data, get plot data
		if (!cachePlotData) {
			recalc(-1, -1, i);
		}

        if (polyList[i] == null) {
            continue;
        }
        // Set up the basic information: series, style, x vals, y vals);
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);
        if (!series.isVisible()) {
            continue;
        }
        JCChartStyle style = series.getStyle();
        if (style == null) {
            continue;
        }

        // Set up the line style
        JCFillStyle fstyle = style.getFillStyle();

        // Now loop over the data
		if (!fu) {
			AreaPolygon thisAP = null;
        	for (int j = 0; j < polyList[i].size(); j++) {
				if ((thisAP = (AreaPolygon) polyList[i].elementAt(j)) != null) {
					Rectangle bounding = thisAP.poly.getBounds();
					int fillOrientation =
                        calcFillOrientation(bounding, yaxis.origin.value,
                                            inverted, xaxis.reversed,
                                            yaxis.reversed);
					fstyle.updateFillOrientation(bounding, fillOrientation);
					fstyle.fillOutlinePolygon(gc, thisAP.poly, outlineStyle);
					fstyle.resetFillOrientation();

            		// if using DefineShape, define shapes for each data index
            		if (defShape != null) {
						defineAreaShapes(defShape, gc, fstyle, inverted,
														 thisAP, series);
            		}

				}
			}
		}
        else {
		// Fast Update case
			AreaPolygon thisAP = null;
        	for (int j = 0; j < polyList[i].size(); j++) {
				if ((thisAP = (AreaPolygon) polyList[i].elementAt(j)) != null) {
					// Old polygons, already drawn from before
					if (startPoint > thisAP.endPoint) {
						continue;
					}
                    else if (thisAP.startPoint >= startPoint - 1 &&
							 thisAP.endPoint <= endPoint)
                    {
					    // Complete polygon with new data.  Draw it.
						fstyle.fillOutlinePolygon(gc, thisAP.poly,
													outlineStyle);
					}
                    else if (startPoint - 1 > thisAP.startPoint) {
					    // Extending already drawn polygon with new data.  Only
					    // draw the new data.
						int realstart =
                            Math.max(startPoint - thisAP.startPoint, 0);
						int numpoints = thisAP.poly.npoints - realstart;
						int[] xpoints = new int[numpoints];
						int[] ypoints = new int[numpoints];
						for (int k = 0; k < numpoints; k++) {
							xpoints[k] =
								thisAP.poly.xpoints[realstart + k];
							ypoints[k] =
								thisAP.poly.ypoints[realstart + k];
						}
						if (inverted) {
							ypoints[numpoints - 1] = ypoints[0];
                        }
						else {
							xpoints[numpoints - 1] = xpoints[0];
                        }
						Polygon tmpPoly = new Polygon(xpoints, ypoints,
                                                      numpoints);
						fstyle.fillOutlinePartialPolygon(gc, tmpPoly,
                                                         outlineStyle);
					}
				}
			}
		}

		// Release plot data if not caching
		if (!cachePlotData) {
			polyList[i] = null;
		}
    }
}

/**
 * Adds points to the area polygon.  If both points are out in the
 * x-direction, then there can be no area underneath to draw.  Therefore, do not add
 * points to the polygon.  Also, if there is one visible point that is either surrounded
 * by holes or is an endpoint flanked by a hole, do not add to polygon.
 *
 * @param x1 x value of first point to add
 * @param y1 y value of first point to add
 * @param x2 x value of second point to add
 * @param y2 y value of second point to add
 * @param j index of x2, y2 point in series
 * @param poly polygon to add points to
 * @param inPoly <code>true</code> indicates that first point is already in polygon
 * @param inverted <code>true</code> if axes are inverted
 * @return new value of <code>inPoly</code>; <code>true</code> if second point was added to polygon
 */
private boolean clipAreaPolygon(double x1, double y1, double x2, double y2,
			                    int j, AreaPolygon poly, boolean inPoly,
                                boolean inverted)
{
	int oc1 = 0, oc2 = 0;

	if (!inPoly) {
        if (x1 == holeValue || y1 == Double.MAX_VALUE) {
			oc1 = oc2 = 1;
		}
        else {
			oc1 = JCChartUtil.locateX(x1, xaxis.min.value, xaxis.max.value);
		}
	}
	if (oc2 == 0) {
        if (x2 == holeValue || y2 == Double.MAX_VALUE) {
			oc1 = oc2 = 1;
		}
        else {
			oc2 = JCChartUtil.locateX(x2, xaxis.min.value, xaxis.max.value);
			if (inPoly && oc2 != 0) {
				/* Can assume no hole here since inPoly is true */
				oc1 = JCChartUtil.locateX(x1, xaxis.min.value, xaxis.max.value);
			}
		}
	}

	/* Check if both points out -- must be out on the same side */
	if (oc1 != 0 && oc2 != 0 && oc1 == oc2) {
		return(false);
	}

	if (inPoly) {
		// First point already added, add new point as next in area polygon
		if (inverted) {
			poly.poly.addPoint(yaxis.toPixel(y2, drawFront),
						       xaxis.toPixel(x2, drawFront));
		}
        else {
			poly.poly.addPoint(xaxis.toPixel(x2, drawFront),
						       yaxis.toPixel(y2, drawFront));
		}
		poly.endPoint = j;
	}
    else {
		// First point not added means this is a new polygon.  First add the
		// beginning bottom point, then add the new first and second points.
		if (inverted) {
			poly.poly.addPoint(yaxis.toPixel(yaxis.origin.value, drawFront),
				               xaxis.toPixel(x1, drawFront));
			poly.poly.addPoint(yaxis.toPixel(y1, drawFront),
						       xaxis.toPixel(x1, drawFront));
			poly.poly.addPoint(yaxis.toPixel(y2, drawFront),
						       xaxis.toPixel(x2, drawFront));
		}
        else {
			poly.poly.addPoint(xaxis.toPixel(x1, drawFront),
					           yaxis.toPixel(yaxis.origin.value, drawFront));
			poly.poly.addPoint(xaxis.toPixel(x1, drawFront),
						       yaxis.toPixel(y1, drawFront));
			poly.poly.addPoint(xaxis.toPixel(x2, drawFront),
						       yaxis.toPixel(y2, drawFront));
		}
		poly.startPoint = j - 1;
		poly.endPoint   = j;
	}

	return(true);
}

/**
 * Finishes up a polygon by adding the last points (the last bottom
 * point and a repeat of the first point to close of the polygon).
 * @param poly polygon to add point to
 * @param inverted <code>true</code> if axes are inverted
 */
private void finishAreaPolygon(Polygon poly, boolean inverted)
{
	if (inverted) {
		poly.addPoint(yaxis.toPixel(yaxis.origin.value, drawFront),
				      poly.ypoints[poly.npoints - 1]);
	}
    else {
		poly.addPoint(poly.xpoints[poly.npoints - 1],
				      yaxis.toPixel(yaxis.origin.value, drawFront));
	}
	poly.addPoint(poly.xpoints[0], poly.ypoints[0]);
}

/**
 * Finishes off the area polygon by constructing its bottom from a list
 * of points that represents the top of the previously constructed polygon.
 * This is used by <code>STACKING_AREA</code> charts.
 * @param ap          the <code>AreaPolygon</code> instance that is being finished
 * @param polyEndLine array of values representing the top line of
 *                    the previously calculated polygon, to be used
 *                    for the bottom line of this polygon
 * @param polyTopLine the top line of this polygon, saved to be used
 *                    by the next polygon
 * @param inverted    <code>true</code> if axes are inverted; <code>false</code> otherwise
 */
private void finishAreaPolygon(AreaPolygon ap, int polyEndLine[],
							   int polyTopLine[], boolean inverted)
{
	Polygon poly = ap.poly;

	// First construct the bottom of this polygon from the given
	// array of points.  If array is null, use the origin value.
	int npoints = poly.npoints;
	for (int j = ap.endPoint; j >= ap.startPoint; j--) {
		int pointIndex = (npoints - 1) - (ap.endPoint - j);
		int ypix;
		if (polyEndLine == null) {
			ypix = yaxis.toPixel(yaxis.origin.value, drawFront);
		}
        else {
			ypix = polyEndLine[j];
		}
		if (inverted) {
			poly.addPoint(ypix, poly.ypoints[pointIndex]);
		}
        else {
			poly.addPoint(poly.xpoints[pointIndex], ypix);
		}
	}

	// Even up the first and last points
	if (inverted) {
		poly.xpoints[0] = poly.xpoints[poly.npoints - 1];
	}
    else {
		poly.ypoints[0] = poly.ypoints[poly.npoints - 1];
	}

	// Now save the top of this polygon to be used by the next polygon
	for (int k = ap.startPoint; k <= ap.endPoint; k++) {
		int pointIndex = (k - ap.startPoint + 1);
		if (inverted) {
			polyTopLine[k] = ap.poly.xpoints[pointIndex];
		}
        else {
			polyTopLine[k] = ap.poly.ypoints[pointIndex];
		}
	}
}

/**
 * Finds the last polygon in the list of polygons.
 * @param polyList the list of polygons
 * @return the index of the last polygon
 */
private int findLastPoly(Vector polyList)
{
	for (int i = 0; i < polyList.size(); i++) {
		if (polyList.elementAt(i) == null) {
			return(i - 1);
        }
	}
	return(polyList.size() - 1);
}

/**
 * Calculates the correct orientation of the tiled image within the
 * rectangle to be filled.
 * @param bounding rectangle to be filled
 * @param origin the origin of the y axis
 * @param inverted whether the axes are inverted or not
 * @param x_reversed whether the x axis is reversed or not
 * @param y_reversed whether the y axis is reversed or not
 * @return the fill orientation
 */
protected int calcFillOrientation(Rectangle bounding, double origin,
					              boolean inverted, boolean x_reversed,
                                  boolean y_reversed)
{
    int originPix = yaxis.toPixel(origin, drawFront);
	if (!inverted) {
		if (bounding.y < originPix) {
			bounding.height = Math.abs(originPix - bounding.y);
			if (!x_reversed) {
				return(JCFillStyle.BOTTOM_LEFT);
			}
            else {
				return(JCFillStyle.BOTTOM_RIGHT);
			}
		}
        else {
			if (!x_reversed) {
				return(JCFillStyle.TOP_LEFT);
			}
            else {
				return(JCFillStyle.TOP_RIGHT);
			}
		}
	}
    else {
		if (bounding.x < originPix) {
			bounding.width = Math.abs(originPix - bounding.x);
			if (!x_reversed) {
				return(JCFillStyle.BOTTOM_RIGHT);
			}
            else {
				return(JCFillStyle.TOP_RIGHT);
			}
		}
        else {
			if (!x_reversed) {
				return(JCFillStyle.BOTTOM_LEFT);
			}
            else {
				return(JCFillStyle.TOP_LEFT);
			}
		}
	}
}

/**
 * Pick method for plot charts.  Given a point relative to the
 * chart area, this method will find the closest point.
 * @param pt (x,y) value relative to the chart area
 * @param focus pick focus used to determine the distance between pick point
 * and data point
 * @return <code>JCDataIndex</code> encapsulating the point index and data
 * series of the selected point
 */
public JCDataIndex pick(Point pt, int focus)
{
	return(pick(pt, focus, false));
}

/**
 * Pick method with direction of picking specified to control how stacking
 * order of polygons affects picking.
 * @param pt (x,y) value relative to the chart area
 * @param focus pick focus used to determine distance between pick point
 * and data point
 * @param forward if <code>true</code>, picks the series list in the forward direction;
 *                if <code>false</code>, picks the series list in the reverse direction
 * @return a <code>JCDataIndex</code> object which stores the indices of the selected point
 *    and the distance to the point
 */
public JCDataIndex pick(Point pt, int focus, boolean forward)
{
    Rectangle r = chartArea.getDrawingArea();
    Point p = new Point(pt.x - r.x, pt.y - r.y);
    int series = -1;
    int point = -1;
    JCDataIndex di = null;

	double xvalue1, yvalue1, xvalue2, yvalue2;

	int pixOrigin = yaxis.toPixel(yaxis.origin.value, drawFront);

    /*
     *  In all cases, the general idea is to slice the polygon up
     *  into trapezoids; by definition, these trapezoids have
     *  parallel sides which are perpindicular to the bottom,
     *  which is the line for the origin.  The top line consists
     *  of the line from the data.  Use function JCChartUtil.distTrap()
     *  to determine which area is the closest to the pixel;
     *  then within that area, use the focus to pick the point
     *  closest to the pixel.
     */

    int minPoint = -1;
    int minSeries = -1;
	double x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	long dist = 0;
	int j1 = 0, j2 = 0;
	boolean hole1 = false, hole2 = false;
	int oc1 = 0, oc2 = 0;

	long minDist = r.width + r.height;
    for (int i = forward ? 0 : seriesList.size() - 1;
		 (forward ? (i < seriesList.size()) : (i >= 0)) && minDist > 0;
		 i += (forward ? 1 : -1) )
    {
        // Set up the basic information: series, style, x vals, y vals);
		ChartDataViewSeries thisSeries =
            (ChartDataViewSeries) seriesList.get(i);
        if (thisSeries == null || !thisSeries.isVisible()) {
            continue;
        }

        if (firstLast == null         ||
            firstLast.x == holeValue  ||
            firstLast.y == holeValue)
        {
            continue;
        }

		double yvals[] = getYVals(thisSeries);
		xvalue1 = thisSeries.getX(firstLast.x);
		yvalue1 = yvals[firstLast.x];
		hole1 = (xvalue1 == holeValue || yvalue1 == Double.MAX_VALUE);
		oc1 = hole1 ? 1 : JCChartUtil.locateXY(xvalue1, yvalue1,
								               xlim.min, xlim.max,
                                               ylim.min, ylim.max);

		boolean firstPointDone = false;
		for (int j = firstLast.x + 1; j <= firstLast.y && minDist > 0; j++) {
			xvalue2 = thisSeries.getX(j);
			yvalue2 = yvals[j];
			hole2 = (xvalue2 == holeValue || yvalue2 == Double.MAX_VALUE);
			oc2 = hole2 ? 2 : JCChartUtil.locateXY(xvalue2, yvalue2,
								                   xlim.min, xlim.max,
                                                   ylim.min, ylim.max);
			boolean holes = hole1 || hole2;

			if (!(oc1 != 0 && oc2 != 0) && !holes) { // at least one visible
				if (inverted) {
					if (!firstPointDone) {
						y1 = xaxis.toPixel(xvalue1, drawFront);
						x1 = yaxis.toPixel(yvalue1, drawFront);
					}
					y2 = xaxis.toPixel(xvalue2, drawFront);
					x2 = yaxis.toPixel(yvalue2, drawFront);

					// if top line crosses origin, trapezoid is split in two
					if ((x2 < pixOrigin && x1 > pixOrigin) ||
						(x1 < pixOrigin && x2 > pixOrigin))
                    {
						double ycross = (y1 + ((pixOrigin - x1) *
													(y2 - y1) / (x2 - x1)));
						long tmpdist1 = JCChartUtil.distTrap(p.y, p.x,
                                                             pixOrigin, y1,
                                                             ycross, x1,
                                                             pixOrigin);
						long tmpdist2 = JCChartUtil.distTrap(p.y, p.x,
												             pixOrigin, ycross,
                                                             y2, pixOrigin, x2);
						if (tmpdist1 < tmpdist2) {
							dist = tmpdist1;
                        }
						else {
							dist = tmpdist2;
                        }
					}
                    else {
						dist = JCChartUtil.distTrap(p.y, p.x, pixOrigin,
													y1, y2, x1, x2);
					}
				}
                else {
					if (!firstPointDone) {
						x1 = xaxis.toPixel(xvalue1, drawFront);
						y1 = yaxis.toPixel(yvalue1, drawFront);
					}
					x2 = xaxis.toPixel(xvalue2, drawFront);
					y2 = yaxis.toPixel(yvalue2, drawFront);

					// if top line crosses origin, trapezoid is split in two
					if ((y2 < pixOrigin && y1 > pixOrigin) ||
						(y1 < pixOrigin && y2 > pixOrigin))
                    {
						double xcross = (x1 + ((pixOrigin - y1) *
													(x2 - x1) / (y2 - y1)));
						long tmpdist1 = JCChartUtil.distTrap(p.x, p.y,
												             pixOrigin, x1,
                                                             xcross, y1,
                                                             pixOrigin);
						long tmpdist2 = JCChartUtil.distTrap(p.x, p.y,
												             pixOrigin, xcross,
                                                             x2, pixOrigin, y2);
						if (tmpdist1 < tmpdist2) {
							dist = tmpdist1;
                        }
						else {
							dist = tmpdist2;
                        }
					}
                    else {
						dist = JCChartUtil.distTrap(p.x, p.y, pixOrigin,
                                                    x1, x2, y1, y2);
					}
				}
				if (dist < minDist) {
					minPoint = j-1;
					minSeries = i;
					minDist = dist;
				}
				firstPointDone = true;
				x1 = x2;
				y1 = y2;
			}
            else {
				firstPointDone = false;
			}
			xvalue1 = xvalue2;
			yvalue1 = yvalue2;
			hole1 = hole2;
			oc1 = oc2;
		}
    }

	if (minSeries == -1) {
		// All sets have at most one point -- degenerate case
		return(null);
	}

	int i = minSeries;
	j1 = minPoint;

	if (j1 < firstLast.x) {
		j1 = firstLast.x;
		j2 = firstLast.x + 1;
	}
    else if (j1 > firstLast.y) {
		j1 = firstLast.y - 1;
		j2 = firstLast.y;
	}
    else {
		j2 = j1 + 1;
		if (j2 > firstLast.y) {
			j1--;
			j2 = j1 + 1;
		}
	}
	if (j1 < firstLast.x) {
		j1 = j2 = firstLast.x; // special case -- one point
    }

	ChartDataViewSeries thisSeries = (ChartDataViewSeries) seriesList.get(i);
	double yvals[] = getYVals(thisSeries);
	xvalue1 = thisSeries.getX(j1);
	yvalue1 = yvals[j1];
	xvalue2 = thisSeries.getX(j2);
	yvalue2 = yvals[j2];

	oc1 = (xvalue1 == holeValue ||
		   yvalue1 == Double.MAX_VALUE) ? 1 : JCChartUtil.locateXY(xvalue1,
                                                                   yvalue1,
							                                       xlim.min,
                                                                   xlim.max,
                                                                   ylim.min,
                                                                   ylim.max);
	oc2 = (xvalue2 == holeValue ||
		   yvalue2 == Double.MAX_VALUE) ? 1 : JCChartUtil.locateXY(xvalue2,
                                                                   yvalue2,
							                                       xlim.min,
                                                                   xlim.max,
                                                                   ylim.min,
                                                                   ylim.max);

	if (inverted) {
		if (oc1 == 0) {
			y1 = xaxis.toPixel(xvalue1, drawFront);
			x1 = yaxis.toPixel(yvalue1, drawFront);
		}
		if (oc2 == 0) {
			y2 = xaxis.toPixel(xvalue2, drawFront);
			x2 = yaxis.toPixel(yvalue2, drawFront);
		}
	}
    else {
		if (oc1 == 0) {
			x1 = xaxis.toPixel(xvalue1, drawFront);
			y1 = yaxis.toPixel(yvalue1, drawFront);
		}
		if (oc2 == 0) {
			x2 = xaxis.toPixel(xvalue2, drawFront);
			y2 = yaxis.toPixel(yvalue2, drawFront);
		}
	}

	if (!(oc1 != 0 && oc2 != 0)) { // At least one visible
		if (!(oc1 != 0 || oc2 != 0)) { // Both visible
			switch (focus) {
            case ChartDataView.PICK_FOCUS_X :
                point = (Math.abs(p.x - x1) < Math.abs(p.x - x2)) ? j1 : j2;
                break;
            case ChartDataView.PICK_FOCUS_Y :
                point = (Math.abs(p.y - y1) < Math.abs(p.y - y2)) ? j1 : j2;
                break;
            case ChartDataView.PICK_FOCUS_XY :
            default:
                long dist1 = (long) (((p.x - x1) * (p.x - x1)) +
                                     ((p.y - y1) * (p.y - y1)));
                long dist2 = (long) (((p.x - x2) * (p.x - x2)) +
                                     ((p.y - y2) * (p.y - y2)));
                point = (dist1 < dist2) ? j1 : j2;
                break;
            }
		}
        else { // pick visible one
			point = (oc1 != 0) ? j2 : j1;
		}
		series = minSeries;
		if (series >= 0 && point >= 0) {
			di = new JCDataIndex(point,
				                 (ChartDataViewSeries)seriesList.get(series),
                                 series);
			di.distance = (int) minDist;
		}
	}
	return(di);
}

/**
 * Unpick method for plot charts.  Given a data series and a datapoint
 * within that series, the function returns pixel coordinates of
 * that point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing position in screen pixels
 */
public Point unpick(int pt, int seriesIndex)
{
    ChartDataViewSeries series =
        (ChartDataViewSeries) seriesList.get(seriesIndex);
	double yvals[] = getYVals(series);
    double yvalue = yvals[pt];
    double xvalue = series.getX(pt);
    int pointX = -1;
    int pointY = -1;

    if (xvalue != holeValue && yvalue != Double.MAX_VALUE) {
        if (inverted) {
            pointX = yaxis.toPixel(yvalue, drawFront);
            pointY = xaxis.toPixel(xvalue, drawFront);
        }
        else {
            pointX = xaxis.toPixel(xvalue, drawFront);
            pointY = yaxis.toPixel(yvalue, drawFront);
        }
        return(new Point(pointX, pointY));
    }

	return(null);
}

/** Location on the chart area of the point before the point being edited. */
Point prev = null;

/** Location on the chart area of the point after the point being edited. */
Point next = null;

/** Last position on the chart area of the edited point. */
Point last = null;

/** Short cut to drawing area for the chart area. */
Rectangle r = null;

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di)
{
    if (di == null || di.dataView != dataObject) {
        return;
    }
    ChartDataView dataView = di.dataView;
	ChartDataViewSeries series = di.series;
    boolean inverted = dataView.getXAxis().i_vertical;
    r = chartArea.getDrawingArea();
    prev = next = last = null;

	double prevDataX, prevDataY;
	double yvals[] = getYVals(series);
	// If prev point doesn't exist or is hole value, use current x at the
    // y origin (the left bottom point of the area polygon) for the prev point
    JCAxis xAxis = dataView.getXAxis();
    JCAxis yAxis = dataView.getYAxis();
    if (di.point - 1 >= 0                       &&
		di.point - 1 >= series.getFirstPoint()  &&
		di.point - 1 <= series.getLastPoint())
    {
		prevDataX = series.getX(di.point - 1);
		prevDataY = yvals[di.point - 1];
		if (prevDataX == dataView.getHoleValue() ||
			prevDataY == dataView.getHoleValue())
        {
			prevDataY = yAxis.origin.value;
			prevDataX = series.getX(di.point);
		}
	}
    else {
		prevDataX = series.getX(di.point);
		prevDataY = yAxis.origin.value;
	}
	if (inverted) {
    	prev = new Point(yAxis.toPixel(prevDataY, drawFront) + r.x,
						 xAxis.toPixel(prevDataX, drawFront) + r.y);
	}
    else {
    	prev = new Point(xAxis.toPixel(prevDataX, drawFront) + r.x,
						 yAxis.toPixel(prevDataY, drawFront) + r.y);
	}

	double nextDataX, nextDataY;
	// If next point doesn't exist or is hole value, use current x at the
    // y origin (the right bottom point of the area polygon) for the next point
    if (di.point + 1 < series.maxXIndex()       &&
		di.point + 1 >= series.getFirstPoint()  &&
		di.point + 1 <= series.getLastPoint())
    {
		nextDataX = series.getX(di.point + 1);
		nextDataY = yvals[di.point + 1];
		if (nextDataX == dataView.getHoleValue() ||
			nextDataY == dataView.getHoleValue())
        {
			nextDataY = yAxis.origin.value;
			nextDataX = series.getX(di.point);
		}
	}
    else {
		nextDataX = series.getX(di.point);
		nextDataY = yAxis.origin.value;
	}
	if (inverted) {
    	next = new Point(yAxis.toPixel(nextDataY, drawFront) + r.x,
						 xAxis.toPixel(nextDataX, drawFront) + r.y);
	}
    else {
    	next = new Point(xAxis.toPixel(nextDataX, drawFront) + r.x,
						 yAxis.toPixel(nextDataY, drawFront) + r.y);
	}
}

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x x location of the action, relative to the chart area
 * @param y y location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y)
{
    if (di == null || di.dataView != dataObject) {
        return;
    }
    chartArea.drawRubberLine(prev, last);
    chartArea.drawRubberLine(last, next);

	// Edit for area only works in y axis direction.
    JCAxis xAxis = di.dataView.getXAxis();
    if (last == null) {
		// Create new point at x axis value of picked point with varying
		// y axis value.
        double xValue = di.series.getX(di.point);
		if (xAxis.i_vertical) {
        	last = new Point(x, xAxis.toPixel(xValue, drawFront) + r.y);
        }
		else {
        	last = new Point(xAxis.toPixel(xValue, drawFront) + r.x, y);
        }
    }
    else {
		if (xAxis.i_vertical) {
			last.x = x;
        }
		else {
        	last.y = y;
        }
    }
    chartArea.drawRubberLine(prev, last);
    chartArea.drawRubberLine(last, next);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x x location of the action, relative to the chart area
 * @param y y location of the action, relative to the chart area
 * @return the x or y location relative to the chart area
 */
int endEdit(JCDataIndex di, int x, int y)
{
    if (di == null || di.dataView != dataObject) {
		return(-1);
    }
    chartArea.drawRubberLine(prev, last);
    chartArea.drawRubberLine(last, next);
    return(di.dataView.getXAxis().i_vertical ? x - r.x : y - r.y);
}

/**
 * Defines the shapes associated with areas of the chart for later lookup.
 * @param defShape a defined shape
 * @param gc the graphics context
 * @param fstyle the fill style
 * @param inverted <code>true</code> if axes are inverted; <code>false</code> otherwise
 * @param thisAP the current area polygon
 * @param series the series used
 */
protected void defineAreaShapes(DefineShape defShape, Graphics gc,
				                JCFillStyle fstyle, boolean inverted,
				                AreaPolygon thisAP, ChartDataViewSeries series)
{
	for (int k = thisAP.startPoint; k < thisAP.endPoint; k++) {
		if (!inverted) {
			int bottom = thisAP.poly.ypoints[0];
			int x1 = thisAP.poly.xpoints[k - thisAP.startPoint + 1];
			int y1 = thisAP.poly.ypoints[k - thisAP.startPoint + 1];
			int x2 = thisAP.poly.xpoints[k - thisAP.startPoint + 2];
			int y2 = thisAP.poly.ypoints[k - thisAP.startPoint + 2];
			double rise = (double) (y2 - y1) / (double) (x2 - x1);
			int xmid = x1 + (int) ((double)(x2 - x1) / 2.0);
			int ymid = y1 - (int) (rise * (double) (x1 - xmid));

			Polygon shapePoly = new Polygon();
			shapePoly.addPoint(x1, bottom);
			shapePoly.addPoint(x1, y1);
			shapePoly.addPoint(xmid, ymid);
			shapePoly.addPoint(xmid, bottom);

			defShape.beginShape(new JCDataIndex(k, series));
       		fstyle.fillPolygon(gc, shapePoly);
       		defShape.endShape(false);

			shapePoly = new Polygon();
			shapePoly.addPoint(xmid, bottom);
			shapePoly.addPoint(xmid, ymid);
			shapePoly.addPoint(x2, y2);
			shapePoly.addPoint(x2, bottom);

    		defShape.beginShape(new JCDataIndex(k+1, series));
        	fstyle.fillPolygon(gc, shapePoly);
        	defShape.endShape(false);
		}
		else {
			int bottom = thisAP.poly.xpoints[0];
			int x1 = thisAP.poly.xpoints[k - thisAP.startPoint + 1];
			int y1 = thisAP.poly.ypoints[k - thisAP.startPoint + 1];
			int x2 = thisAP.poly.xpoints[k - thisAP.startPoint + 2];
			int y2 = thisAP.poly.ypoints[k - thisAP.startPoint + 2];
			double rise = (double) (x2 - x1) / (double) (y2 - y1);
			int ymid = y1 + (int) ((double)(y2 - y1) / 2.0);
			int xmid = x1 - (int) (rise * (double) (y1 - ymid));

			Polygon shapePoly = new Polygon();
			shapePoly.addPoint(bottom, y1);
			shapePoly.addPoint(x1, y1);
			shapePoly.addPoint(xmid, ymid);
			shapePoly.addPoint(bottom, ymid);

			defShape.beginShape(new JCDataIndex(k, series));
       		fstyle.fillPolygon(gc, shapePoly);
       		defShape.endShape(false);

			shapePoly = new Polygon();
			shapePoly.addPoint(bottom, ymid);
			shapePoly.addPoint(xmid, ymid);
			shapePoly.addPoint(x2, y2);
			shapePoly.addPoint(bottom, y2);

    		defShape.beginShape(new JCDataIndex(k+1, series));
        	fstyle.fillPolygon(gc, shapePoly);
        	defShape.endShape(false);
		}
	}
}

}
