/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: AreaRadarChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;
import java.util.List;

/**
 * Drawable object for area radar charts.  This object
 * is used for rendering an area radar chart based on data
 * contained in the <code>dataObject</code>.
 */
public class AreaRadarChartDraw extends RadarChartDraw {

// Local copies of data needed for draw(), pick(), and unpick().  They are
// set in recalc().
protected Point        firstLast = null;
protected JCLineStyle  outlineStyle = null;
protected MinMax       xlim = null;
protected MinMax       ylim = null;

protected AreaPolygon areaPolygon[];

/**
 * Default constructor.
 */
public AreaRadarChartDraw()
{
  super();
}

/**
 * Method used to compute the polygons that make up the filled regions
 * of an area radar chart.  Stacked areas are currently not supported.
 */
public void recalc()
{
	recalc(-1, -1, -1);
}

/**
 * Method used to compute the polygons that make up the filled regions
 * of an area radar chart.  Stacked areas are currently not supported.
 *
 * @param startPoint the index of the first plotted point in a series
 * @param endPoint the index of the last plotted point in a series
 */
public void recalc(int startPoint, int endPoint)
{
	recalc(startPoint, endPoint, -1);
}

/**
 * Method used to compute the polygons that make up the filled regions
 * of an area radar chart.  Stacked areas are currently not supported.
 *
 * @param startPoint the index of the first plotted point in a series
 * @param endPoint the index of the last plotted point in a series
 * @param seriesIndex index into the series List
 */
public void recalc(int startPoint, int endPoint, int seriesIndex)
{
	cachePlotData = dataObject.getBufferPlotData();

    // Make a list of points
	// If being called for a particular series, use previously stored series
	// list.  Otherwise, refresh series list.
	if (seriesIndex == -1) {
    	seriesList = dataObject.getSeries();
	}
	if (seriesList == null) {
		return;
	}

	if (dataObject.getXLimits() == null) {
		return;
	}
	if (dataObject.getYLimits() == null) {
		return;
	}
	xlim = new MinMax(dataObject.getXLimits().min, dataObject.getXLimits().max);
	ylim = new MinMax(dataObject.getYLimits().min, dataObject.getYLimits().max);
	dataObject.getDataBounds(xlim, ylim);

    holeValue = dataObject.getHoleValue();
	firstLast = dataObject.getFirstLast();
	xaxis = dataObject.getXAxis();
	yaxis = dataObject.getYAxis();

	boolean ylog = yaxis.isLogarithmic();
	//  Have ylog only.  The theta axis is always linear.

	// carry on if caching or if not, a valid seriesIndex is specified
	if (!cachePlotData && seriesIndex == (-1)) {
		if (areaPolygon == null) {
    		areaPolygon = new AreaPolygon[seriesList.size()];
        }
		return;
	}

    areaPolygon = new AreaPolygon[seriesList.size()];

   	// Setup parameters to loop thru the data series
	int firstSeries = (seriesIndex == (-1)) ? 0 : seriesIndex;
	int lastSeries  = (seriesIndex == (-1)) ? seriesList.size()
                                            : (seriesIndex + 1);

	firstPoint = startPoint = getFirstPoint(seriesList, firstSeries,
                                            lastSeries);
	lastPoint = endPoint = getLastPoint(seriesList, firstSeries, lastSeries);

	for (int i = firstSeries; i < lastSeries; i++) {
        // Set up the basic information: series, style, x vals, y vals);
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);

        // Drop out if data is uninteresting
        if (series.maxXIndex() == 0 || series.maxYIndex() == 0) continue;
		int numPts;

		numPts = endPoint - startPoint + 1;
		areaPolygon[i] = new AreaPolygon();
		areaPolygon[i].series = i;

		double yvals[] = getYVals(series);

		// Calculate value of first point.
        double radarAngle1 = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
                                                       numPts, 0,
                                                       xaxis.isReversed());
        //  Note that we pass 0, not startPoint.
        double y1 = yvals[startPoint];
        if (ylog && y1 <= 0.0) {
            y1 = holeValue;
        }
        boolean inPoly = false;

        for (int j = startPoint + 1; j <= endPoint; j++) {
			if (j >= yvals.length) {
				continue;
			}
			double radarAngle2 = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
                                                           numPts,
                                                           j - startPoint,
                                                           false);
			// We pass false for the last parameter because theta axis
			// reversal is handled by the toPixelPolar() and toDataPolar()
			// methods.
			double y2 = yvals[j];
			if (ylog && y2 <= 0.0) {
				y2 = holeValue;
			}

			inPoly = areaPolygonAddPoints(radarAngle1, y1, radarAngle2, y2, j,
										  areaPolygon[i], inPoly);

			radarAngle1 = radarAngle2;
			y1 = y2;
        }

		if (areaPolygon[i].poly.npoints == 0) {
			//  The polygon for this series is empty, so make it null
			areaPolygon[i] = null;
		}
	}
}

protected int getFirstPoint(List seriesList, int firstSeries, int lastSeries)
{
	//  Search all of the series for the highest value of firstPoint
	int firstPoint = Integer.MIN_VALUE;
	int index;
	for (int i = firstSeries; i < lastSeries; i++) {
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);
        index = series.getFirstPoint();
        if (index > firstPoint) {
			firstPoint = index;
		}
	}
	return(firstPoint);
}

protected int getLastPoint(List seriesList, int firstSeries, int lastSeries)
{
	//  Search all of the series for the lowest value of lastPoint
	int lastPoint = Integer.MAX_VALUE;
	int index;
	for (int i = firstSeries; i < lastSeries; i++) {
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);
        index = series.getLastPoint();
        if (index < lastPoint) {
			lastPoint = index;
		}
	}
	return(lastPoint);
}

protected double[] getYVals(ChartDataViewSeries series)
{
    double oldyvals[] = series.getY();
    double yvals[] = new double[oldyvals.length];
    for (int i = 0; i < yvals.length; i++) {
        if (oldyvals[i] == holeValue ||
			(yaxis.isLogarithmic() && oldyvals[i] <= 0.0))
        {
            yvals[i] = Double.MAX_VALUE;
        }
        else {
            yvals[i] = oldyvals[i];
        }
    }
    return(yvals);
}


/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc)
{
	draw(gc, -1, -1, true);
}


/**
 * Draw method with direction of drawing specified to control the stacking order
 * of polygons.
 * @param startPoint the point to start drawing at
 * @param endPoint the point to end drawing at
 * @param forward if <code>true</code>, draws by going forward through the polygon list;
 *                if <code>false</code>, draws by going backwards through the polygon
 *                list
 * @param gc
 */
public void draw(Graphics gc, int startPoint, int endPoint, boolean forward)
{
	// Get list of data series to be drawn using this drawable
	if (seriesList == null) {
        return;
    }

    Rectangle clipBounds = gc.getClipBounds();
    if (clipBounds.width == 0 || clipBounds.height == 0) {
        return;
	}

    if (areaPolygon == null) {
        return;
    }

    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape) gc;
    }

	outlineStyle = dataObject.getOutlineStyle();

    // Must draw each data series in order
    for (int i = forward ? 0 : areaPolygon.length - 1;
			forward ? (i < areaPolygon.length) : (i >= 0);
			i+= (forward ? 1 : -1) )
    {
		// If not caching plot data, get plot data
		if (!cachePlotData) {
			recalc(-1, -1, i);
		}

        if (areaPolygon[i] == null) {
            continue;
        }
        // Set up the basic information: series, style, x vals, y vals);
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);
        if (!series.isVisible()) {
            continue;
        }
        JCChartStyle style = series.getStyle();
        if (style == null) {
            continue;
        }

        // Set up the line style
        JCFillStyle fstyle = style.getFillStyle();
		Rectangle bounding = areaPolygon[i].poly.getBounds();
		int fillOrientation = calcFillOrientation(bounding, yaxis.origin.value,
						                          xaxis.reversed,
                                                  yaxis.reversed);
		fstyle.updateFillOrientation(bounding, fillOrientation);
		fstyle.fillOutlinePolygon(gc, areaPolygon[i].poly, outlineStyle);
		fstyle.resetFillOrientation();
		// if using DefineShape, define shapes for each data index
		if (defShape != null) {
			defineAreaShapes(defShape, gc, fstyle, areaPolygon[i], series);
		}

		// Release plot data if not caching
		if (!cachePlotData) {
			areaPolygon[i] = null;
		}
    }
}

private Point getPixelPolar(double xValue, double rValue)
{
	return yaxis.toPixelPolar(xValue, rValue, true);
}

private boolean areaPolygonAddPoints(double x1, double y1, double x2, double y2,
								     int j, AreaPolygon poly, boolean inPoly)
{
	double r1 = y1;
	double r2 = y2;
	//  Hole values are plotted as points at the chart center:
	if (inPoly && r1 == holeValue) {
		r1 = (yaxis.isReversed()) ? yaxis.max.value : yaxis.min.value;
	}
	if (r2 == holeValue) {
		r2 = (yaxis.isReversed()) ? yaxis.max.value : yaxis.min.value;
	}

	Point pnt = null;
	if (inPoly) {
		// first point already added, add new point as next in area polygon
		pnt = getPixelPolar(x2, r2);
		poly.poly.addPoint(pnt.x, pnt.y);
		poly.endPoint = j;
	}
    else {
		// first point not added means this is a new polygon.
		pnt = getPixelPolar(x1, r1);
		poly.poly.addPoint(pnt.x, pnt.y);
		pnt = getPixelPolar(x2, r2);
		poly.poly.addPoint(pnt.x, pnt.y);

		poly.startPoint = j - 1;
		poly.endPoint = j;
	}
	return(true);
}

/**
 * Calculates the correct orientation of the tiled image within the
 * rectangle to be filled.
 * @param bounding rectangle to be filled
 * @param origin the origin of the y axis
 * @param x_reversed whether the x axis is reversed or not
 * @param y_reversed whether the y axis is reversed or not
 * @return the fill orientation
 */
protected int calcFillOrientation(Rectangle bounding, double origin,
								  boolean x_reversed, boolean y_reversed)
{
	return JCFillStyle.BOTTOM_LEFT;
}


/**
 * Defines the shapes associated with areas of the chart for later lookup.
 * @param ap
 * @param defShape
 * @param fstyle
 * @param gc
 * @param series
 */
protected void defineAreaShapes(DefineShape defShape, Graphics gc,
								JCFillStyle fstyle, AreaPolygon ap,
								ChartDataViewSeries series)
{

    Point center = null;
    if (!yaxis.isReversed()) {
        center = yaxis.toPixelPolar(0.0, yaxis.getMin(), true);
    }
    else {
        center = yaxis.toPixelPolar(0.0, yaxis.getMax(), true);
    }

	for (int k = ap.startPoint; k <= ap.endPoint; k++) {
		int thisIndex = k;
    	// area radar charts are circular -- if the next or previous point is
    	// off the beginning or end of the point array, wrap around to the
    	// other side

    	int prevIndex =  -1;
    	if (thisIndex > ap.startPoint) {
        	prevIndex = thisIndex - 1;
    	}
        else {
        	prevIndex = ap.endPoint;
    	}
    	int nextIndex =  -1;
    	if (thisIndex < ap.endPoint) {
        	nextIndex = thisIndex + 1;
    	}
        else {
        	nextIndex = ap.startPoint;
    	}

		Polygon shapePoly = null;
        int xvalue1 = ap.poly.xpoints[thisIndex];
        int yvalue1 = ap.poly.ypoints[thisIndex];
        // if the current point is the same as the series polygon center,
        // there is no shape to reference
        if (xvalue1 != center.x || yvalue1 != center.y) {
            shapePoly = new Polygon();
            shapePoly.addPoint(center.x, center.y);

            int xvalue0 = ap.poly.xpoints[prevIndex];
            int yvalue0 = ap.poly.ypoints[prevIndex];

            // if previous point is same as center of series polygon,
            // there's no need to worry about adding halfway point
            if (xvalue0 != center.x || yvalue0 != center.y) {
                xvalue0 = xvalue0 -
                    (int) ((double)(xvalue0 - xvalue1) / 2.0);
                yvalue0 = yvalue0 -
                    (int) ((double)(yvalue0 - yvalue1) / 2.0);

                shapePoly.addPoint(xvalue0, yvalue0);
            }

            // add current point
            shapePoly.addPoint(xvalue1, yvalue1);

            int xvalue2 = ap.poly.xpoints[nextIndex];
            int yvalue2 = ap.poly.ypoints[nextIndex];
            // if next point is same as center of series polygon,
            // there's no need to worry about adding halfway point
            if (xvalue2 != center.x || yvalue2 != center.y) {
                xvalue2 = xvalue2 +
                    (int) ((double)(xvalue1 - xvalue2) / 2.0);
                yvalue2 = yvalue2 +
                    (int) ((double)(yvalue1 - yvalue2) / 2.0);

                shapePoly.addPoint(xvalue2, yvalue2);
            }
        }

		if (shapePoly != null) {
			defShape.beginShape(new JCDataIndex(k, series));
       		fstyle.fillPolygon(gc, shapePoly);
       		defShape.endShape(false);
		}
	}
}

/**
 * Pick method for radar charts.  Given a point relative to the
 * chart area, this method finds the closest point.
 * @param pt (x,y) value relative to the chart area
 * @param focus pick focus used to determine distance between pick point
 * and data point
 * @return <code>JCDataIndex</code> encapsulating the point index and data
 * series of the selected point
 */
public JCDataIndex pick(Point pt, int focus)
{
	return pick(pt, focus, false);
}

/**
 *  Pick method with direction of picking specified to control how the stacking
 *  order of polygons affects picking.
 *  @param forward if <code>true</code>, picks the series list in the forward direction;
 *                 if <code>false</code>, will pick the series list in the reverse direction
 * @param pt
 * @param focus
 * @return
 */
public JCDataIndex pick(Point pt, int focus, boolean forward)
{
	JCDataIndex di = null;
	if (seriesList == null) {
		return(di);
	}

	Rectangle r = chartArea.getDrawingArea();
	//  Get coordinates relative to origin of chart area.
	int px = pt.x - r.x;
	int py = pt.y - r.y;
	JCDataCoord pickPoint = yaxis.toDataPolar(px, py);

	//  Check if pickPoint is outside the Radar chart:
	if (pickPoint.y > yaxis.max.value) {
		return di;   //  return null
	}

	//  Determine if the pick point is contained within a particular polygon.
	//  If so, set distance=0 and search the corresponding series for the
	//  closest data point (observe pick focus rule).
	int i = (forward) ? 0 : (areaPolygon.length - 1);
    for (i = forward ? 0 : areaPolygon.length - 1;
		 forward ? (i < areaPolygon.length) : (i >= 0);
		 i += (forward ? 1 : -1) )
    {
		// If not caching plot data, get plot data
		if (!cachePlotData) {
			recalc(-1, -1, i);
		}

		if (areaPolygon[i] == null) {
			continue;
		}

		if (areaPolygon[i].poly.contains(px, py)) {
			//  Now search the corresponding series for the data point
			//  closest to the pick point:
			di = findClosestPointInSeries(i, focus, px, py, pickPoint);
//			System.out.println("di.seriesIndex=" + di.seriesIndex +
//			", di.point=" + di.point);
			di.distance = 0;
			break;
		}
	}

	if (forward ? (i >= areaPolygon.length) : (i < 0)) {
		//  Couldn't find a polygon that contains the pick point.
		//  For PICK_FOCUS_X or PICK_FOCUS_Y, handle in the same way as for
		//  a non-area Radar chart.  For PICK_FOCUS_XY, iterate through each
		//  of the polygons, searching for the closest line segment.
		switch (focus) {
		case ChartDataView.PICK_FOCUS_X:   //  fall through to next case
		case ChartDataView.PICK_FOCUS_Y:
			di = super.pick(pt, focus);
			break;
		case ChartDataView.PICK_FOCUS_XY:
		default:
			di = pickPolygonXY(px, py, pickPoint);
		}
	}

//	System.out.println("di.seriesIndex=" + di.seriesIndex +
//					   ", di.point=" + di.point +
//					   ", di.distance=" + di.distance);

	return(di);
}

/*
 * Iterates through each of the polygons, searching for the closest
 * line segment.  Once we have identified the closest polygon,
 * searches the corresponding data series for the data point that
 * is closest to the pick point.
 */
private JCDataIndex pickPolygonXY(int px, int py, JCDataCoord pickPoint)
{
	JCDataIndex di = null;
	int series = -1;
	double length = Double.MAX_VALUE;
	double len;
	//  Direction of search doesn't matter here.
    for (int i = 0; i < areaPolygon.length; i++) {

		// If not caching plot data, get plot data
		if (!cachePlotData) {
			recalc(-1, -1, i);
		}

		if (areaPolygon[i] == null) {
			continue;
		}

		len = JCChartUtil.getPolygonDistance(px, py, areaPolygon[i].poly);
		if (len < length) {
			series = i;
			length = len;
		}

		if (len == 0) {
			// The pick point is inside the polygon.
			break;
		}
	}

	if (series >= 0) {
		//  Search that series for the data point closest
		//  to the pick point:
		di = findClosestPointInSeries(series, ChartDataView.PICK_FOCUS_XY,
									  px, py, pickPoint);
		di.distance = (int)length;
		//  We want the minimum distance to the polygon, not to the
		//  closest data point.
	}

	return(di);
}

}

