/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: AxisHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.util.Vector;

abstract class AxisHandler {

protected final static int ANNO_PADDING = 2;

JCAxis parent = null;

/** Internal property used to map data space to screen space */
double shift;

/**
 * Internal property used to map data space to screen space.
 * For Polar/Radar charts, this scale factor applies only to
 * the r Axis.
 */
double scale = 1.0;

AxisScaleHandler axisScaleHandler = null;

void setParent(JCAxis p)
{
    this.parent = p;
    axisScaleHandler.setParent(p);
}

static AxisHandler makeAxisHandler(JCAxis ax)
{
    AxisHandler rval = null;
    try {
        String name = null;
        if (ax.isPolar() || ax.isRadar()) {
            // circular chart
            if (ax.logarithmic && ax.i_vertical) {
                // Only the y axis of a circular chart can be logarithmic.
                name = "com.klg.jclass.chart.CircularChartLogAxisHandler";
            }
            else {
                name = "com.klg.jclass.chart.CircularChartLinearAxisHandler";
            }
        }
        else {
            // rectangular chart
            if (ax.logarithmic) {
                name = "com.klg.jclass.chart.RectangularChartLogAxisHandler";
            }
            else {
                name = "com.klg.jclass.chart.RectangularChartLinearAxisHandler";
            }
        }
        rval = (AxisHandler)Class.forName(name).newInstance();
    }
    catch (Exception e) {
//        e.printStackTrace();
        rval = null;
    }
    if (rval != null) {
        rval.setParent(ax);
    }
    return (rval);
}

/**
 * Get font metrics from current graphics object.
 */
protected FontMetrics getFontMetrics() {
    JCChart chart = parent.getParentChart();
    FontMetrics pFont = null;
    try {
        // we query the font metrics from the chart graphics object
        // in case a subsitute graphics object (like those used for
        // various encoders or the ServerReport/PageLayout products)
        // with its own font metrics is being used
        Graphics queryGC = chart.getQueryGraphics();
        if (queryGC != null) {
            pFont = queryGC.getFontMetrics(parent.getFont());
        }
        else {
            pFont = chart.getToolkit().getFontMetrics(parent.getFont());
        }
    }
    catch (Exception e) {
    }
    return pFont;
}

/**
 * This method returns a Dimension object containing the width and
 * height of a label.
 * @param value The value represented by the label.
 * @param precision The precision of the label.
 * @return A Dimension object containing width and height values. Null
 * if unable to calculate the label dimensions.
 */
Dimension getAnnoDimension(double value, int precision)
{
    FontMetrics pFont = getFontMetrics();
    if (pFont == null) {
        return (null);
    }

    Dimension dim;
    if (parent.labelGenerator != null) {
        Object o = parent.labelGenerator.makeLabel(value, precision);
        if (o == null) {
            return (null);
        }

        if (o instanceof ChartText) {
            ChartText c = (ChartText)o;
            c.setParentChart(parent.getParentChart());
            dim = new Dimension(c.getWidth(), c.getHeight());
        }
        else {
            String s = (o instanceof String) ? (String)o : o.toString();
            dim = new Dimension(pFont.stringWidth(s), pFont.getHeight());
        }
    }
    else {
        String s = axisScaleHandler.format(value, precision);
        dim = new Dimension(pFont.stringWidth(s), pFont.getHeight());
    }
    return (dim);
}

/**
 * Gets the width and height of an annotation generated for the
 * provided value and precision.  This method is applicable in cases
 * where annotationRotation is JCAxis.ROTATE_NONE, JCAxis.ROTATE_180,
 * JCAxis.ROTATE_90, or JCAxis.ROTATE_270.
 * @param p Dimension object in which to return the size
 * @param value Value to display
 * @param precision Precision to use to display the value
 */
void getAnnotationExtent(Dimension p, double value, int precision)
{
    Dimension dim = getAnnoDimension(value, precision);
    if (dim == null) {
        return;
    }

    if (parent.annotationRotation == JCAxis.ROTATE_90 ||
        parent.annotationRotation == JCAxis.ROTATE_270) {
        // Flip the width and height if the rotation is 90 or 270 degrees.
        if (p.width < dim.height) {
            p.width = dim.height;
        }
        if (p.height < dim.width) {
            p.height = dim.width;
        }
    }
    else {
        if (p.width < dim.width) {
            p.width = dim.width;
        }
        if (p.height < dim.height) {
            p.height = dim.height;
        }
    }
}

/**
 * Calculates the origin position
 */
double calcOrigin(double min, double max)
{
    double rval = 0.0;

    // Extra accounting for polar, stacking bars and other details.
    if (parent.origin.isDefault) {
        switch (parent.originPlacement.value) {
        case JCAxis.ZERO:
            rval = 0.0;
            break;
        case JCAxis.MIN:
            rval = min;
            break;
        case JCAxis.MAX:
            rval = max;
            break;
        case JCAxis.AUTOMATIC:
            if (min <= 0.0 && max >= 0.0)
                rval = 0.0;
            else
                rval = min;
        }
    }
    else {
        rval = parent.origin.value;
    }
    return rval;
}

/**
 * This method calculates the pixel spacing between major tick marks
 * needed to avoid overlap of rotated labels.
 * @param lmin The minimum label value.
 * @param lmax The maximum label value.
 * @return The required pixel spacing between major tick marks.
 */
int calcRotatedAnnoSpacing(double lmin, double lmax)
{
    Dimension d = null;
    int maxAnnoSpacing = 0;
    double theta = parent.getAnnotationRotationAngle(JCChartUtil.RADIANS);
    AnnotationHandler handler = parent.getAnnotationHandler();
    Vector userValueLabels = handler.userValueLabels;
    if (parent.getAnnotationMethod() != JCAxis.VALUE &&
        userValueLabels != null && userValueLabels.size() > 0) {
        int annoSpacing = 0;
        // Check all of the user provided annotations.
        for (int i = 0; i < userValueLabels.size(); i++) {
            JCValueLabel vl = (JCValueLabel)userValueLabels.elementAt(i);
            ChartText cl = vl.getChartText();
            annoSpacing = handler.calcRotatedAnnoSpacing(cl.width.value,
                                                         cl.height.value,
                                                         theta);
            maxAnnoSpacing = Math.max(annoSpacing, maxAnnoSpacing);
        }
    }
    else {
        int annoSpacing1 = 0;
        int annoSpacing2 = 0;
        d = getAnnoDimension(lmin, parent.getPrecision());
        if (d != null) {
            annoSpacing1 = handler.calcRotatedAnnoSpacing(d.width, d.height,
                                                          theta);
        }
        d = getAnnoDimension(lmax, parent.getPrecision());
        if (d != null) {
            annoSpacing2 = handler.calcRotatedAnnoSpacing(d.width, d.height,
                                                          theta);
        }
        maxAnnoSpacing = Math.max(annoSpacing1, annoSpacing2);
    }
    return(maxAnnoSpacing);
}

/**
 * Draw tick marks along the axis.
 * @param gc Graphics context to use for drawing.
 */
void drawTicks(Graphics gc)
{
    if (parent == null) {
        return;
    }
    int axisLoc = parent.linePosition();
    Vector ans = parent.getAnnotations();
    for (int i = 0; i < ans.size(); i++) {
        JCValueLabel vl = (JCValueLabel)ans.elementAt(i);
        if (!vl.isDrawTick()) {
            continue;
        }
        drawTickMark(gc, vl, axisLoc);
    }
}

/**
 * Called by the calcTransformation() methods of the linear and
 * logarithmic axis handler subclasses.
 */
protected void computeTransformation(double amin, double amax, int length)
{
    double err = JCChartUtil.calcError(parent.getPrecision());
    if (parent.reversed) {
        length = -length;
    }
    if (!parent.logarithmic && Math.abs(amax - amin) < err) {
        scale = 1.0;
        shift = 0.0;
        return;
    }
    double diff = amax - amin;
    scale = (diff != 0) ? length / diff : 0.0;
    int start = parent.pixelStart();
    if (length < 0) {
        shift = start - amax * scale;
    }
    else {
        shift = start - amin * scale;
    }
}

void calcPrecision()
{
    axisScaleHandler.calcPrecision();
}

double range()
{
    return axisScaleHandler.range();
}

void translate(double t)
{
    axisScaleHandler.translate(t);
}

/**
 * Converts each of the ticks specified by the JCAnno parameter
 * into a JCValueLabel.  The new JCValueLabels are added to the
 * AnnotationHandler's internal List of value labels.
 *
 * The calcParams() method of AxisHandler must be called first.
 */
protected void makeTickLabels(JCAnno tick)
{
    axisScaleHandler.makeTickLabels(tick);
}

/**
 * Returns the innerExtent property of the specified JCAnno
 * object.  Returns the default inner extent if the JCAnno
 * parameter is null.
 */
protected double getInnerExtent(JCAnno tick)
{
    return (tick == null) ?
        JCAnno.DEFAULT_INNER_EXTENT : tick.getInnerExtent();
}

/**
 * Returns the outerExtent property of the specified JCAnno
 * object.  Returns the default outer extent if the JCAnno
 * parameter is null.
 */
protected double getOuterExtent(JCAnno tick)
{
    return (tick == null) ?
        JCAnno.DEFAULT_OUTER_EXTENT : tick.getOuterExtent();
}

/**
 * Returns the tickColor property of the specified JCAnno
 * object.  Returns a default value for the tickColor if the
 * JCAnno parameter is null or if the axis foreground color
 * is null.
 */
protected Color getTickColor(JCAnno tick)
{
    Color tickColor = (tick == null) ? parent.getForeground()
                                     : tick.getTickColor();
    if (tickColor == null) {
        tickColor = Color.black;
    }
    return tickColor;
}

protected void updateAnnoObject(JCAnno anno) {
    axisScaleHandler.updateAnnoObject(anno);
}

abstract void calcParams();

abstract void calcTransformation();

abstract void drawTickMark(Graphics gc, JCValueLabel label, int axisLoc);

}
