/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BarChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;

/**
 * Drawable object for bar charts.  This object
 * is used for rendering a bar chart based on data
 * contained in the <code>dataObject</code>.
 */
public class BarChartDraw extends ChartDraw implements java.io.Serializable {

/** Used to specify sides of polygons to draw outlines for. */
public final static int SIDE_0	= 1;
public final static int SIDE_1	= 2;
public final static int SIDE_2	= 4;
public final static int SIDE_3	= 8;

/** Bar will be outlined if width and height are greater than this value. */
public final static int OUTLINE_SENSITIVITY	= 3;

// Local copies of data needed for draw(), pick(), and unpick().  They are
// set in recalc().
protected Point        firstLast = null;
protected JCLineStyle  outlineStyle = null;
protected int          x_reversed = 0;
protected int          y_reversed = 0;
protected int          delta_x = 0;
protected int          delta_y = 0;
protected double       xAxisPos = Double.MAX_VALUE;
protected double[]     xGridLineList = null;
protected JCLineStyle  gridLineStyle = null;
protected boolean      fastAction = false;
protected boolean      inAction = false;
protected double       yOrigin = 0.0;

/** Array of information representing each bar. */
protected transient BarPoint[][] pts = null;

/** Actual width of a bar. */
protected double bar_base;

/** Outer bounds of drawing area. */
protected int top, bottom, left, right;

/** <code>True</code> if drawing 3d bars. */
protected boolean is_3d = false;

/** <code>True</code> if drawing an outline around the bar. */
protected boolean drawOutline = true;

/**
 * Default constructor.
 */
public BarChartDraw()
{
	super();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc()
{
    super.recalc();
    JCBarChartFormat format = (JCBarChartFormat)dataObject.getChartFormat();

    // if no cluster width, null out pts array so that old bars won't
    // redraw on draw().
    if (format == null || format.clusterWidth == 0) {
        pts = null;
        return;
    }

    delta_x = chartArea.getDeltaX();
    delta_y = chartArea.getDeltaY();
    is_3d = (delta_x != 0 || delta_y != 0);

    // Do not allow bars to intersect if using 3d
    int clusterOverlap = format.clusterOverlap;
    if (is_3d && clusterOverlap > 0) {
        clusterOverlap = 0;
    }

    if (dataObject.getXLimits() == null) {
        return;
    }
    if (dataObject.getYLimits() == null) {
        return;
    }

    MinMax x = new MinMax(dataObject.getXLimits().min,
                          dataObject.getXLimits().max);
    MinMax y = new MinMax(dataObject.getYLimits().min,
                          dataObject.getYLimits().max);
    dataObject.getDataBounds(x, y);

    xaxis = dataObject.getXAxis();
    yaxis = dataObject.getYAxis();

    // Precompute frequently used values.
    // In the following:
    //	barf	 - bar factor - number of effective bars in a cluster
    //	bar_base - actual width of a bar (in data space)
    //    csize2	 - half of actual cluster size (in data space)
    //    binc	 - effective bar increment (in data space)
    //  Thus, csize would be equal to:  barf * bar_base

    double gapf = 1.0 - (clusterOverlap / 100.0);
    double barf = numActiveDataSeries;
    barf = (barf - 1) * gapf + 1;
    bar_base = ((dataObject.step * format.clusterWidth) / (barf * 100.0));
    double csize2 = barf * bar_base / 2;
    double binc = gapf * bar_base;
    inverted = (xaxis.i_vertical);
    //	If data_min/data_max use_defaults are FALSE, honor those
    //	values, otherwise see if we can extend the bounds to include
    //	half of a cluster.
    x.min = x.min - csize2;
    x.max = x.max + csize2;
    xaxis.setLimits(x);
    // Have to recalculate some axis parameters
    xaxis.calcTransformation();
    xaxis.layoutAnnotations();
    xaxis.checkAnnotations();

    y.min = JCChartUtil.min(yaxis.origin.value, y.min);
    y.max = JCChartUtil.max(yaxis.origin.value, y.max);
    yaxis.setLimits(y);
    yaxis.calcTransformation();

    x = xaxis.dataLimits;
    y = yaxis.dataLimits;

    boolean x_log = xaxis.isLogarithmic();
    boolean y_log = yaxis.isLogarithmic();

    // If width of bar is greater than sensitivity, draw an outline around
    // all bars.  The outline needs to be consistently drawn or not drawn
    // for all bars.  Check width from axis min for consistency when bars
    // are translated.
    drawOutline = (Math.abs(xaxis.toPixel(xaxis.min.value + bar_base,
                                          drawFront) -
                            xaxis.toPixel(xaxis.min.value, drawFront)) >
                   OUTLINE_SENSITIVITY);

    // Compute plotting area limits (this is the front plane in 3D)
    int width  = (inverted ? yaxis.getWidth() : xaxis.getHeight());
    int height = (inverted ? xaxis.getWidth() : yaxis.getHeight());

    left = (inverted ? yaxis.getLeft() : xaxis.getLeft());
    right = left + width - 1;

    top = (inverted ? xaxis.getTop() : yaxis.getTop());
    bottom = top + height - 1;

    firstLast = dataObject.getFirstLast();
    if (firstLast == null) return;
    seriesList = dataObject.getSeries();
    int nseries = seriesList.size();

    holeValue = dataObject.getHoleValue();

    // Transform data points to pixel coordinates.  The information
    // is stored in the array "pts".  beg_pix and end_pix are the
    // pixel values of beginning and ending of the bar in the x
    // direction and len_pix in the y direction (even if inverted).
    //
    // "x" and "y" store the x and y values of the bars in data space
    // to distinguish between valid and invalid values (invalid would
    // be holes and ones outside of data and/or axis bounds).

    pts = new BarPoint[firstLast.y - firstLast.x + 1][nseries];
    yOrigin = yaxis.origin.value;
    int pix_org = yaxis.toPixel(yOrigin, drawFront);

    x_reversed = xaxis.reversed ? 1 : 0;
    y_reversed = yaxis.reversed ? 1 : 0;
    int switch_order;
    if (inverted) {
        switch_order = (delta_y > 0) ? (1 ^ x_reversed) : (0 ^ x_reversed);
    }
    else {
        switch_order = (delta_x <= 0) ? (1 ^ x_reversed) : (0 ^ x_reversed);
    }

    fastAction = chartArea.getFastAction();
    inAction = chartArea.getInAction();

    if (!fastAction || !inAction) {
        xAxisPos = xaxis.getRealPlacementLocation();

        // Get list of grid line values for the x axis.  If null, make into an
        // empty array
        xGridLineList = xaxis.getGridLineList();
        if (xGridLineList == null) {
            xGridLineList = new double[0];
        }

        gridLineStyle = xaxis.getGridStyle().getLineStyle();
    }

    for (int i = firstLast.x; i <= firstLast.y; i++) {
        int bar_num = 0;
        for (int j = 0; j < seriesList.size(); j++) {
            BarPoint lp[] = pts[i - firstLast.x];
            ChartDataViewSeries thisSeries =
                (ChartDataViewSeries)seriesList.get(j);
            if (!thisSeries.isVisible()) continue;
            double sx = thisSeries.getX(i);
            double base = thisSeries.getX(i) - csize2;
            double xbeg_value = base + bar_num * binc;
            double xend_value = xbeg_value + bar_base;
            bar_num++;

            if ((xbeg_value < xaxis.min.value &&
                xend_value <= xaxis.min.value) ||
                (xbeg_value >= xaxis.max.value &&
                xend_value > xaxis.max.value)) {
                continue;
            }

            lp[j] = new BarPoint();
            /// Clip the bars to the data min and max
            lp[j].x = xbeg_value;
            if (x != null && (xend_value <= x.min || xbeg_value >= x.max)) {
                lp[j].x = holeValue;
            }

            // check for x value holes here since the x value that gets
            // saved in the BarPoint object is it's actual position on the
            // x axis, not the data x value
            if (sx == holeValue || (x_log && sx <= 0.0)) {
                lp[j].x = holeValue;
            }

            double yvalue = thisSeries.getY(i);
            if (yvalue == holeValue || (y_log && yvalue <= 0.0)) {
                lp[j].y = holeValue;
            }
            else {
                lp[j].y = yvalue;
            }

            BarPoint bp = lp[j];
            bp.front = null;
            bp.top = null;
            bp.side = null;

            int pix_x, pix_y;
            int bar_height, bar_width;

            if (inverted) {
                pix_x = yaxis.toPixel(bp.y, drawFront);
                pix_y = xaxis.toPixel(bp.x, drawFront);
                bar_height = pix_y - xaxis.toPixel(bp.x + bar_base, drawFront);
                bar_width = pix_x - pix_org;
            }
            else {
                pix_x = xaxis.toPixel(bp.x, drawFront);
                pix_y = yaxis.toPixel(bp.y, drawFront);
                bar_width = xaxis.toPixel(bp.x + bar_base, drawFront) - pix_x;
                bar_height = pix_org - pix_y;
            }

            /* Calc 3d shadows */
            if (is_3d) {
                if (bp.x != holeValue && bp.y != holeValue) {
                    Point[] top_points = new Point[4];
                    Point[] side_points = new Point[4];

                    top_points[0] = new Point(0, 0);
                    top_points[1] = new Point(delta_x, delta_y);
                    top_points[2] = new Point(0, 0);
                    top_points[3] = new Point(-delta_x, -delta_y);
                    side_points[0] = new Point(0, 0);
                    side_points[1] = new Point(delta_x, delta_y);
                    side_points[2] = new Point(0, 0);
                    side_points[3] = new Point(-delta_x, -delta_y);

                    if (inverted) {
                        if (bar_width >= 0) {
                            side_points[0].x = (delta_x > 0) ? pix_x : pix_org;
                        }
                        else {
                            side_points[0].x = (delta_x > 0) ? pix_org : pix_x;
                        }

                        int pix_top = (switch_order != 0) ?
                            pix_y : pix_y - bar_height;

                        top_points[0].x = pix_x - bar_width;
                        top_points[0].y = pix_top;
                        top_points[2].x = bar_width;
                        side_points[0].y = pix_y - bar_height;
                        side_points[2].y = bar_height;
                        bp.top = top_points;
                        bp.side = side_points;
                    }
                    else {

                        if (bar_height >= 0) {
                            top_points[0].y = (delta_y <= 0) ? pix_y : pix_org;
                        }
                        else {
                            top_points[0].y = (delta_y <= 0) ? pix_org : pix_y;
                        }

                        int pix_side = (switch_order != 0) ?
                            pix_x : pix_x + bar_width;
                        side_points[0].x = pix_side;
                        side_points[0].y = pix_y + bar_height;
                        side_points[2].y = -bar_height;
                        top_points[0].x = pix_x;
                        top_points[2].x = bar_width;
                        bp.top = top_points;
                        bp.side = side_points;
                    }
                }
            }

            if (inverted) {
                pix_x = pix_org;
                pix_y = xaxis.toPixel(bp.x, drawFront);
                bar_width = yaxis.toPixel(bp.y, drawFront) - pix_x;
                bar_height = pix_y - xaxis.toPixel(bp.x + bar_base, drawFront);
                pix_y -= bar_height;
            }

            if (bar_width < 0) {
                bar_width = -bar_width;
                pix_x -= bar_width;
            }

            if (bar_height < 0) {
                bar_height = -bar_height;
                pix_y -= bar_height;
            }

            // ensure bar displays even if it is too thin
            if (inverted) {
                bar_height = Math.max(1, bar_height);
            }
            else {
                bar_width = Math.max(1, bar_width);
            }
            bp.front = new Rectangle(pix_x, pix_y, bar_width, bar_height);
        }
    }
}

/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc)
{
    if (firstLast == null) {
        return;
    }

    Rectangle clipBounds = gc.getClipBounds();
    if (clipBounds.width == 0 || clipBounds.height == 0) {
        return;
    }

    // Handle deserialized case by calling recalc() if pts is null.
    if (pts == null) {
        try {
            recalc();
            if (pts == null) {
                return;
            }
        }
        catch (Exception e) {}
    }

    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape)gc;
    }

    outlineStyle = dataObject.getOutlineStyle();

    if (is_3d) {
        //	Draw the 3-D stuff, i.e. tops and sides of bars
        boolean yTroughGridDrawn = false;
        boolean finishTroughAfterBar = false;

        int switch_order;
        if (inverted) {
            switch_order = (delta_y > 0) ?
                (1 ^ x_reversed) : (0 ^ x_reversed);
        }
        else {
            switch_order = (delta_x <= 0) ?
                (1 ^ x_reversed) : (0 ^ x_reversed);
        }

        int xGridIndex = 0;

        int ibeg, jbeg, inc;
        boolean more_clusters;

        int first_set = 0;
        int last_set = seriesList.size() - 1;

        if (switch_order != 0) {
            ibeg = firstLast.y;
            jbeg = last_set;
            inc = -1;
            more_clusters = (ibeg >= firstLast.x);
            xGridIndex = xGridLineList.length - 1;
        }
        else {
            ibeg = firstLast.x;
            jbeg = first_set;
            inc = 1;
            more_clusters = (ibeg <= firstLast.y);
            xGridIndex = 0;
        }
        BarPoint lastBar = null;
        for (int i = ibeg; more_clusters; i += inc) {
            boolean more_bars = (switch_order != 0) ?
                (jbeg >= first_set) : (jbeg <= last_set);
            for (int j = jbeg; more_bars; j += inc) {
                BarPoint bp = null;

                // if you manually set wierd axis limits,
                // you may get array out of bounds, or pull
                // out a null ref

                try {
                    bp = pts[i - firstLast.x][j];
                }
                catch (ArrayIndexOutOfBoundsException e) {}

                if ((bp != null) && (!fastAction || !inAction)) {

                    // Axis trough grid lines need to be drawn here, when the
                    // bars near the axes are drawn, so that we can achieve
                    // the proper 3D look.

                    // the pixel values of the bar front along the x axis,
                    // used to check if the bar straddles the y axis or not.
                    int bar_x1 = 0;
                    int bar_x2 = 0;
                    if (inverted) {
                        bar_x2 = bp.front.y + bp.front.height;
                        bar_x1 = bp.front.y;
                    }
                    else {
                        bar_x1 = bp.front.x;
                        bar_x2 = bp.front.x + bp.front.width;
                    }

                    // if y axis trough lines not drawn yet, see if it is
                    // time to draw it.

                    // Don't check gridVisible here, check it inside
                    // tryDrawYAxisTroughGrid().  The reason?  We must
                    // check the trough ends the same way we check the
                    // grid lines, and the trough ends are always drawn.
                    if (!yTroughGridDrawn) {
                        int result = tryDrawYAxisTroughGrid(gc, delta_x,
                                                            delta_y, yaxis,
                                                            bar_x1, bar_x2);
                        if (result == TROUGH_DRAWN) {
                            yTroughGridDrawn = true;
                        }
                        else if (result == FINISH_TROUGH_LATER) {
                            finishTroughAfterBar = true;
                            yTroughGridDrawn = true;
                        }
                    }

                    // draw undrawn x axis trough lines up to current bar
                    if (xaxis.gridVisible) {
                        xGridIndex = drawXAxisTroughGrid(gc, delta_x, delta_y,
                                                         xaxis, y_reversed,
                                                         xAxisPos, bar_x1,
                                                         bar_x2, bp.y, yOrigin,
                                                         xGridLineList,
                                                         xGridIndex, inc,
                                                         gridLineStyle);
                    }
                }

                if (bp != null && bp.x != holeValue &&
                    bp.y != holeValue &&
                    !((ChartDataViewSeries)seriesList.get(j)).
                    getStyle().getFillStyle().isCustomStackable()) {

                    draw_3dbar(gc, i, j, bp.top, bp.side,
                               calcFillOrientation(bp.y, 0.0, inverted,
                                                   yaxis.reversed),
                               drawOutline);
                    lastBar = bp;

                    // Finish drawing y axis trough grid lines that draw over
                    // bar shadow
                    if (finishTroughAfterBar) {
                        finishDrawYAxisTroughGrid(gc, delta_x, delta_y,
                                                  yaxis, y_reversed, bp.y, yOrigin);
                        finishTroughAfterBar = false;
                    }
                }

                more_bars = (switch_order != 0) ?
                    (j > first_set) : (j < last_set);
            }
            more_clusters =
                (switch_order != 0) ? (i > firstLast.x) : (i < firstLast.y);
        }
        if (!fastAction || !inAction) {
            // draw the rest of the y axis grid lines

            // Don't draw last grid lines and re-outline trough
            // if last bar overhangs axis
            boolean drawOrNot = true;
            if (lastBar != null) {
                drawOrNot = isDrawFinalGridAndTrough(
                    lastBar,
                    isDrawOverBar(inverted, y_reversed,
                                  delta_y, delta_x, xAxisPos, lastBar.y, yOrigin),
                    inverted,
                    xaxis.toPixel(xaxis.max.value, drawFront),
                    xaxis.toPixel(xaxis.min.value, drawFront));
            }
            if (drawOrNot) {
                finalDrawAxisTroughGrid(gc,
                                        xaxis, yaxis, switch_order,
                                        yTroughGridDrawn, xGridLineList, xGridIndex,
                                        inc, gridLineStyle);
            }
        }
    }

    //    Draw the front faces of the bars.  Note that we first draw the
    //    all bar interiors and then all the bar outlines for a
    //	particular set.  This may seem inefficient at first.
    //	However, this is actually more efficient as it saves alot of
    //	data_style changes which is more expensive than recomputing
    //	the position, width, and height of the bar.

    for (int j = 0; j < seriesList.size(); j++) {
        ChartDataViewSeries thisSeries = (ChartDataViewSeries)seriesList.get(j);
        if (!thisSeries.isVisible()) {
            continue;
        }
        JCChartStyle cs = thisSeries.getStyle();

        //  Draw the bar interior.
        for (int i = firstLast.x; i <= firstLast.y; i++) {
            BarPoint bp = null;
            try {
                bp = pts[i - firstLast.x][j];
            }
            catch (ArrayIndexOutOfBoundsException e) {}

            if (bp == null) {
                continue;
            }
            if (bp.x != holeValue && bp.y != holeValue) {
                JCFillStyle fs = cs.getFillStyle();
                if (fs.isCustomStackable()) {
                    fs.fillImageRect(gc, bp.front.x, bp.front.y,
                                     bp.front.width, bp.front.height,
                                     (bp.y >= 0), inverted);
                }
                else {
                    fs.updateFillOrientation(
                        new Rectangle(bp.front.x, bp.front.y,
                                      bp.front.width, bp.front.height),
                        calcFillOrientation(bp.y, 0.0, inverted,
                                            yaxis.reversed));

                    // If using DefineShape, begin a shape definition for
                    // this index
                    if (defShape != null) {
                        defShape.beginShape(new JCDataIndex(i, thisSeries));
                    }

                    fs.fillRect(gc, bp.front.x, bp.front.y,
                                bp.front.width, bp.front.height);

                    // if using DefineShape, finish up the shape definition
                    if (defShape != null) {
                        defShape.endShape(true);
                    }

                    if (drawOutline) {
                        fs.drawOutlineRect(gc, bp.front.x, bp.front.y,
                                           bp.front.width, bp.front.height,
                                           outlineStyle);
                    }
                    fs.resetFillOrientation();
                }

            }
        }
    }
}

/**
 * Draws a 3d bar's top and side shadows.
 * @param gc graphics context for drawing anywhere in plot area
 * @param i index of current point within series
 * @param j index of current series
 * @param top array of points defining the top shadow
 * @param side array of points defining the side shadow
 * @param orientation orients the fill inside the polygon this way
 * @param drawOutline <code>true</code> if bar is drawn with an outline
 */
void draw_3dbar(Graphics gc, int i, int j, Point[] top,
                Point[] side, int orientation, boolean drawOutline)
{
    JCFillStyle fstyle = ((ChartDataViewSeries)seriesList.get(j)).
        getStyle().getFillStyle();
    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape)gc;
    }

    // Draw the top shadow
    // get positions of x and y axis for determining clip

    if (top != null) {
        Polygon top_poly = JCChartUtil.pointsToPolygon(top);
        fstyle.updateFillOrientation(top_poly.getBounds(), orientation);

        // if using DefineShape, begin a shape definition for this index
        if (defShape != null) {
            defShape.beginShape(
                new JCDataIndex(i, (ChartDataViewSeries)seriesList.get(j)));
        }

        if (delta_y <= 0) {
            fstyle.fillPolygonHighlight(gc, top_poly);
        }
        else {
            fstyle.fillPolygonShadow(gc, top_poly);
        }

        // if using DefineShape, finish up the shape definition
        if (defShape != null) {
            defShape.endShape(true);
        }

        if (drawOutline) {
            fstyle.drawOutlinePolygon(gc, top_poly, outlineStyle);
        }

        fstyle.resetFillOrientation();
    }

    // Draw the side shadow
    if (side != null) {
        Polygon side_poly = JCChartUtil.pointsToPolygon(side);
        fstyle.updateFillOrientation(side_poly.getBounds(), orientation);

        // if using DefineShape, begin a shape definition for this index
        if (defShape != null) {
            defShape.beginShape(
                new JCDataIndex(i, (ChartDataViewSeries)seriesList.get(j)));
        }

        if (delta_x <= 0) {
            fstyle.fillPolygonHighlight(gc, side_poly);
        }
        else {
            fstyle.fillPolygonShadow(gc, side_poly);
        }

        // if using DefineShape, finish up the shape definition
        if (defShape != null) {
            defShape.endShape(true);
        }

        if (drawOutline) {
            fstyle.drawOutlinePolygon(gc, side_poly, outlineStyle);
        }

        fstyle.resetFillOrientation();
    }
}

/**
 * Calculates the fill orientation for images based on the <code>yval</code> being
 * graphed and the inversion of the axes.  The following code is used
 * by bar and stacking bar charts.  Other drawables may override this
 * method as needed.
 * @param inverted
 * @param origin
 * @param reversed
 * @param yval
 * @return
 */
protected int calcFillOrientation(double yval, double origin,
                                  boolean inverted, boolean reversed)
{
    if (!inverted) {
        if ((!reversed && yval >= origin) || (reversed && yval < origin)) {
            return(JCFillStyle.BOTTOM);
        }
        else {
            return(JCFillStyle.TOP);
        }
    }
    else {
        if ((!reversed && yval >= origin) || (reversed && yval < origin)) {
            return(JCFillStyle.LEFT);
        }
        else {
            return(JCFillStyle.RIGHT);
        }
    }
}

/**
 * Picks the nearest bar to the given point.
 * @param pt point used to find the nearest bar
 * @param focus pick focus used to determine distance between point and bar
 * @return an index of the data point in the series that is represented by
 * the picked bar
 */
public JCDataIndex pick(Point pt, int focus)
{
    Rectangle r = chartArea.getDrawingArea();
    Point p = new Point(pt.x - r.x, pt.y - r.y);
    int series = -1;
    int point = -1;
    JCDataIndex di = null;
    int length = Integer.MAX_VALUE;

    if (firstLast == null || pts == null) {
        return(null);
    }

    // Check the 3d shadows first for pick
    // Note : the structure of this loop is the same as the loop for
    //        drawing the 3d shadows.
    boolean is_3d = (delta_x != 0 || delta_y != 0);
    if (is_3d) {
        int switch_order;
        if (inverted) {
            switch_order = (delta_y > 0) ? (1 ^ x_reversed) : (0 ^ x_reversed);
        }
        else {
            switch_order = (delta_x <= 0) ? (1 ^ x_reversed) : (0 ^ x_reversed);
        }
        int ibeg, jbeg, inc;
        boolean more_clusters;

        int first_set = 0;
        int last_set = seriesList.size() - 1;

        if (switch_order != 0) {
            ibeg = firstLast.y;
            jbeg = last_set;
            inc = -1;
            more_clusters = (ibeg >= firstLast.x);
        }
        else {
            ibeg = firstLast.x;
            jbeg = first_set;
            inc = 1;
            more_clusters = (ibeg <= firstLast.y);
        }
        for (int i = ibeg; more_clusters; i += inc) {
            boolean more_bars = (switch_order != 0) ? (jbeg >= first_set)
                                                    : (jbeg <= last_set);
            for (int j = jbeg; more_bars; j += inc) {
                BarPoint bp = null;
                try {
                    bp = pts[i - firstLast.x][j];
                }
                catch (ArrayIndexOutOfBoundsException e) {}
                if (bp != null) {
                    int len_top =
                        JCChartUtil.distanceTo3dPolygon(p.x, p.y, bp.top,
                                                        inverted, focus);
                    int len_side =
                        JCChartUtil.distanceTo3dPolygon(p.x, p.y, bp.side,
                                                        inverted, focus);
                    if ((len_top <= length) || (len_side <= length)) {
                        series = j;
                        point = i;
                        length = Math.min(len_side, len_top);
                    }
                }
                more_bars = (switch_order != 0) ? (j > first_set)
                                                : (j < last_set);
            }
            more_clusters =
                (switch_order != 0) ? (i > firstLast.x) : (i < firstLast.y);
        }
    }

    // Check the bar fronts
    for (int i = firstLast.x; i <= firstLast.y; i++) {
        int thisIndex = i - firstLast.x;
        if (pts[thisIndex] == null) {
            continue;
        }
        for (int j = 0; j < pts[thisIndex].length; j++) {
            if (!((ChartDataViewSeries)seriesList.get(j)).isVisible()) {
                continue;
            }
            BarPoint bp = null;
            try {
                bp = pts[thisIndex][j];
            }
            catch (ArrayIndexOutOfBoundsException e) {}
            if (bp == null || bp.x == holeValue || bp.y == holeValue) {
                continue;
            }
            int len = 0;

            if (bp.front != null) {
                len = JCChartUtil.distanceToFrontRect(p.x, p.y, bp.front.x,
                                                      bp.front.y,
                                                      bp.front.width,
                                                      bp.front.height, inverted,
                                                      focus);
                if (len <= length) {
                    series = j;
                    point = i;
                    length = len;
                }
            }
        }
    }
    if (series >= 0 && point >= 0) {
        di = new JCDataIndex(point,
                             (ChartDataViewSeries)seriesList.get(series), series);
        di.distance = length;
    }
    return di;
}

/**
 * Unpick method for bar charts.  Given a data series and a datapoint
 * within that series, the function returns pixel coordinates of
 * the point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return Point object representing position in screen pixels
 */
public Point unpick(int pt, int seriesIndex)
{
    BarPoint bp = null;
    try {
        bp = pts[pt - firstLast.x][seriesIndex];
    }
    catch (ArrayIndexOutOfBoundsException e) {}
    if (bp == null) {
        return(null);
    }
    double yvalue = bp.y;
    double xvalue = bp.x;
    int point_x = -1;
    int point_y = -1;

    if ((yvalue != holeValue) && (xvalue != holeValue)) {
        if (inverted) {
            point_x = yaxis.toPixel(yvalue, drawFront);
            point_y = bp.front.y + (bp.front.height / 2);
        }
        else {
            point_y = yaxis.toPixel(yvalue, drawFront);
            point_x = bp.front.x + (bp.front.width / 2);
        }
        return(new Point(point_x, point_y));
    }

    return(null);
}

/** Rectangle of drawing area. Used by edit routines.*/
Rectangle r = null;
/** <code>True</code> if axes are inverted. Used by edit routines.*/
boolean editInverted = false;

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di)
{
    r = chartArea.getDrawingArea();
    editInverted = di.dataView.getXAxis().i_vertical;
    start = end = null;
}

/** Start point of edit. */
Point start = null;
/** End point of edit. */
Point end = null;

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x x location of the action, relative to the chart area
 * @param y y location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y)
{
    if (di == null || di.dataView != dataObject) {
        return;
    }
    BarPoint bp = null;
    try {
        bp = pts[di.point - firstLast.x][di.series.getDrawingOrder()];
    }
    catch (ArrayIndexOutOfBoundsException e) {
    }
    if (bp == null) return;

    if (start == null || end == null) {
        start = new Point(0, 0);
        end = new Point(0, 0);
    }
    else {
        chartArea.drawRubberBand(end, start);
    }
    if (editInverted) {
        start.x = x;
        start.y = bp.front.y + r.y;
        end.x = bp.front.x + r.x;
        end.y = bp.front.y + r.y + bp.front.height;
    }
    else {
        start.x = bp.front.x + r.x;
        start.y = y;
        end.x = bp.front.x + r.x + bp.front.width;
        end.y = bp.front.height + bp.front.y + r.y;
    }
    chartArea.drawRubberBand(end, start);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x x location of the action, relative to the chart area
 * @param y y location of the action, relative to the chart area
 * @return
 */
int endEdit(JCDataIndex di, int x, int y)
{
    if (di == null || di.dataView != dataObject) {
        return(-1);
    }
    chartArea.drawRubberBand(end, start);
    int off = editInverted ? x - r.x : y - r.y;
    return(off);
}

}
