/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ChartDataView.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.chart.data.ChartDataSourceUtil;
import com.klg.jclass.util.JCName;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.io.OutputDataProperties;

import java.awt.*;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Vector;

/**
 * Class containing the internal representation of the chartable data.
 * Chart data is represented as a series of <code>ChartDataViewSeries</code>
 * objects, managed by a <code>ChartDataView</code> object.
 * <p><B>Note:</b> <code>ChartDataView</code> includes a mapping from
 * series number to series name.
 * <P>
 * <code>ChartDataView</code> also manages the
 * <code>JCChartStyles</code> for the data series. By
 * default, if no <code>JCChartStyle</code> is provided
 * by the developer, <code>ChartDataView</code> will
 * create a default <code>JCChartStyle</code> for the
 * data series.
 * <p>
 * <b>Note:</b> This means that <code>ChartDataView</code> is
 * <strong>not</strong> always a view -- it is
 * capable of being a view, but is not necessarily a view.
 */
public class ChartDataView implements Changeable,
                                      ChartDataListener,
                                      java.io.Serializable,
                                      JCName
{

protected Vector series = new Vector();

/** Enum values for pick focus. */
public final static int PICK_FOCUS_LOCAL = -1;
public final static int PICK_FOCUS_XY = 0;
public final static int PICK_FOCUS_X = 1;
public final static int PICK_FOCUS_Y = 2;

/** Error message for invalid series index. */
public final static String ERRMSG_INVALID_SERIES_INDEX =
                            "Invalid ChartDataViewSeries index";

private double[]                  xvalues = null;
protected boolean                 batched = false;
protected boolean                 fastUpdate = false;
protected boolean                 bufferPlotData = true;
protected double                  holeValue = Double.MAX_VALUE;
protected int                     chartType = JCChart.PLOT;
protected String                  name = null;

protected int                     drawingOrder=-1;
protected boolean                 drawFrontPlane = true;
protected ChartDataModel          dataSource=null;
protected JCAxis                  xaxis = null;
protected JCAxis                  yaxis = null;
protected List                    pointLabels = new Vector();
protected ImageMapInfo[]          clusterImageMapInfo;
protected ImageMapInfo            imageMapInfo;
protected boolean                 inverted = false;
protected int                     pickFocus = PICK_FOCUS_XY;
protected boolean                 visibleInLegend = true;
protected boolean                 visibleInLegendChanged = false;
private Vector                    styles = new Vector();
protected boolean                 autoLabel = false;
protected JCLineStyle             outlineStyle = null;
protected JCDrawableColorHandler  colorHandler;
protected OutputDataProperties    outputDataProperties;

transient protected DataHandler   dataHandler;
transient private Object          transientData = null;

protected boolean                 visible = true;
protected boolean                 changed = true;
protected int                     changedFlag = Changeable.NO_CHANGE;
protected boolean                 allowPick = true;
protected double                  step = 1.0;

protected JCChart                 chart = null;

protected MinMax                  xLimits;
protected MinMax                  yLimits;

protected Locale                  locale = Locale.getDefault();

/**
 * Class names of format classes. These names should correspond in position
 * in the array to the chart types in <code>JCChart</code>.
 */
public final static String[] format_class_names = {
    null,
    null,
    "com.klg.jclass.chart.JCPolarRadarChartFormat",
    "com.klg.jclass.chart.JCPolarRadarChartFormat",
    "com.klg.jclass.chart.JCPolarRadarChartFormat",
    "com.klg.jclass.chart.JCHiloChartFormat",
    "com.klg.jclass.chart.JCHLOCChartFormat",
    "com.klg.jclass.chart.JCCandleChartFormat",
    "com.klg.jclass.chart.JCAreaChartFormat",
    "com.klg.jclass.chart.JCBarChartFormat",
    "com.klg.jclass.chart.JCBarChartFormat",
    "com.klg.jclass.chart.JCPieChartFormat",
    "com.klg.jclass.chart.JCAreaChartFormat"};

// Hashtable of JCChartTypeFormat objects, keyed by chart type
protected Hashtable formatClassHashtable = new Hashtable();

protected static int dataViewNum = 0;

/**
 * Constructor with parent chart specified.
 * @param parent The parent chart
 */
ChartDataView(JCChart parent)
{
    this(parent.getLocale());
    setParent(parent);
}

/**
 * Constructor with locale specified
 * @param locale The locale to use
 */
ChartDataView(Locale locale)
{
    this.locale = locale;
    setChartType(JCChart.PLOT);
    name = JCChartBundle.string(JCChartBundle.key26, locale) + (++dataViewNum);
}

/**
 * Constructor. Default chart type of <code>PLOT</code>. Default name generated.
 */
public ChartDataView()
{
    this(Locale.getDefault());
}

/**
 * Sets the <code>JCChart</code> object with which a <code>ChartDataView</code>
 * instance is associated.
 * @param parent
 */
public void setParent(JCChart parent)
{
    chart = parent;
    locale = chart.getLocale();
    if (chart != null && chart.getChartArea() != null) {
        if (xaxis == null || !chart.getChartArea().getXAxes().contains(xaxis)) {
            xaxis = chart.getChartArea().getXAxis(0);
        }
        if (yaxis == null || !chart.getChartArea().getYAxes().contains(yaxis)) {
            yaxis = chart.getChartArea().getYAxis(0);
        }
    }
    // Make sure the "new" data view is marked as changed
    setChanged(true, Changeable.RECALC, false);
}


/**
 * Gets the <code>JCChart</code> object with which a <code>ChartDataView</code>
 * instance is associated.
 * @return <code>JCChart</code> object
 */
public JCChart getParent()
{
    return(chart);
}

/**
 * Set the locale for this axis.
 */
public void setLocale(Locale locale)
{
    if (locale == null) {
        locale = Locale.getDefault();
    }
    this.locale = locale;

    setChanged(true, Changeable.REDRAW);
}

/**
 * Retrieve the local for this axis.
 */
public Locale getLocale()
{
    return locale;
}

/**
 * Rendering method for the chart data. Using the given
 * Graphics context, this method draws the data in the space
 * provided. <br><B>Note:</B> It is up to the <code>ChartArea</code> to clip the
 * Graphics context appropriately. This method relies on the
 * clip rectangle to determine how to show the data.
 * @param gr graphics context for rendering
 */
void draw(java.awt.Graphics gr) {}

/**
 * This is not meant to be called from outside the package
 * User has editied a value, notify the data source.
 * @param ser the series that is being edited
 * @param pointIndex the index of the point in the series being edited
 * @param newValue the new value of the point being edited
 * @return a boolean indicating if the edit was accepted
 *
 */
boolean editPoint(ChartDataViewSeries ser, int pointIndex, double newValue)
{
    if (dataSource == null) {
        // no external data, allow the edit
        // find the appropriate series and change its data

        return(true);
    }

    if (!(dataSource instanceof EditableChartDataModel)) { // It is not editable
        return(false);
    }

    int seriesIndex = series.indexOf(ser);

    // was the series in this DataView?

    if (seriesIndex == -1) {
        throw new RuntimeException("Bad row in editPoint!");
    }

    return(getDataHandler().editPoint(seriesIndex, pointIndex, newValue));
}

/**
 * Implements the <code>ChartDataListener</code> interface.
 * This method will be called after chart data has changed.
 * @param evt the chart data event object
 */
public void chartDataChange(ChartDataEvent evt)
{
    if (evt == null) {
        return;
    }

    int seriesIndex = evt.getSeriesIndex();
    int pointIndex  = evt.getPointIndex();
    switch(evt.getType()) {
    case ChartDataEvent.RELOAD_POINT:
        if (seriesIndex == ChartDataEvent.ALL_SERIES &&
            pointIndex == ChartDataEvent.ALL_POINTS)
        {
            getDataHandler().reloadAllSeries();
        }
        else if (seriesIndex == ChartDataEvent.ALL_SERIES) {
            getDataHandler().reloadPoint(pointIndex);
        }
        else if (pointIndex == ChartDataEvent.ALL_POINTS) {
            getDataHandler().reloadSeries(seriesIndex);
        }
        else {
            getDataHandler().reloadPoint(seriesIndex, pointIndex);
        }
        break;
    case ChartDataEvent.RELOAD_SERIES:
        if (seriesIndex == ChartDataEvent.ALL_SERIES) {
            getDataHandler().reloadAllSeries();
       }
        else {
            getDataHandler().reloadSeries(seriesIndex);
        }
        break;
    case ChartDataEvent.ADD_SERIES:
        // ATTN GV Need series values!
        getDataHandler().addSeries();
        break;
    case ChartDataEvent.INSERT_SERIES:
        // ATTN GV Need series values!
        if (seriesIndex == ChartDataEvent.ALL_SERIES) {
            return;
        }
        getDataHandler().insertSeries(seriesIndex);
        break;
    case ChartDataEvent.REMOVE_SERIES:
        if (seriesIndex == ChartDataEvent.ALL_SERIES) {
            getDataHandler().removeAllSeries();
        }
        else {
            getDataHandler().removeSeries(seriesIndex);
        }
        break;
    case ChartDataEvent.RELOAD_ALL_POINT_LABELS:
        getDataHandler().reloadAllPointLabels();
        break;
    case ChartDataEvent.RELOAD_POINT_LABEL:
        if (pointIndex == ChartDataEvent.ALL_POINTS) {
            getDataHandler().reloadAllPointLabels();
        }
        else {
            getDataHandler().reloadPointLabel(pointIndex);
        }
        break;
    case ChartDataEvent.RELOAD_ALL_SERIES_LABELS:
        getDataHandler().reloadAllSeriesLabels();
        break;
    case ChartDataEvent.RELOAD_SERIES_LABEL:
        if (seriesIndex == ChartDataEvent.ALL_SERIES) {
            getDataHandler().reloadAllSeriesLabels();
        }
        else {
            getDataHandler().reloadSeriesLabel(seriesIndex);
        }
        break;
    case ChartDataEvent.RELOAD_DATA_SOURCE_NAME:
        getDataHandler().reloadDataSourceName();
        break;
    case ChartDataEvent.RELOAD_HOLE_VALUE:
        getDataHandler().reloadHoleValue();
        break;
    case ChartDataEvent.APPEND_DATA:
        getDataHandler().appendData();
        if (seriesIndex != ChartDataEvent.ALL_SERIES &&
            pointIndex != ChartDataEvent.ALL_POINTS)
        {
            int column  = pointIndex;  // Data column stored in pointIndex
            int nPoints = seriesIndex; // Number of points stored in seriesIndex

            // Perform fast update if fast update is turned on AND this is a
            // plot, scatter, or area chart AND the newly added points fit
            // inside the axis boundaries.
            int ct = getChartType();
            if (fastUpdate                      &&
                chart.isDisplayable()           &&
                (ct == JCChart.PLOT         ||
                 ct == JCChart.SCATTER_PLOT ||
                 ct == JCChart.AREA)            &&
                checkNewPoints(nPoints))
            {
                boolean success = doFastUpdate(column, nPoints);
                if (success) {
                    return;
                }
            }
        }
        break;
    case ChartDataEvent.RESET:
        getDataHandler().reset();
        break;
    case ChartDataEvent.RELOAD:
        getDataHandler().reload();
        break;
    case ChartDataEvent.CHANGE_CHART_TYPE:
        int type = seriesIndex;    // New chart type is passed in seriesIndex
        if (type < JCChart.PLOT || type > JCChart.LAST_PREDEFINED_CHART_TYPE) {
            throw new IllegalArgumentException("Invalid chart type");
        }
        getDataHandler().changeChartType(type);
        break;
    case ChartDataEvent.RELOAD_IMAGE_MAP_INFO:
        getDataHandler().loadImageMapInfo();
        break;
    default:
        throw new RuntimeException("Unkwown event type: " + evt.getType());
    }

    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Performs a fast update to the chart by just drawing over the current
 * chart without doing a lot of recalcing and redrawing of existing points.
 * Called internally; not intended to be called externally.
 * A call to this method is generated by sending an <code>APPEND_DATA</code>
 * event to chart when the <code>FastUpdate</code> property of the view has been
 * turned on.
 * <BR><b>NOTE:</b> This will cause problems with internal frames, drop down
 * menus, and other such items that might be in the way when a draw happens.
 * @param column column number of data array with new data
 * @param nPoints Number of points added
 * @return <code>true</code> if fast update drawn; <code>false</code> otherwise
 */
public boolean doFastUpdate(int column, int nPoints)
{
    boolean success = false;
    // Reproduce a simpler version of the recalc/redraw mechanism here
    ChartDraw d = chart.getChartArea().getDrawable(this);
    if (d != null) {
        // Fix up the gc and call the drawable's recalc() and draw()
        // method's directly.
        Graphics gc1 = null;
        // catch exceptions that sometimes occur with threads
        // when program is shutting down
        try {
            d.recalc(column, column + nPoints - 1);
            gc1 = chart.getChartArea().getGraphics();
            Rectangle rect;
            rect = chart.getChartArea().getDrawingArea();
            gc1.translate(rect.x, rect.y);

            Rectangle drawClipRect =
                chart.getChartArea().clipForDrawing();
            gc1.clipRect(drawClipRect.x, drawClipRect.y,
                         drawClipRect.width, drawClipRect.height);

            d.draw(gc1, column, column + nPoints - 1);

            success = true;
        }
        catch (Exception e) {}
        if (gc1 != null) {
            gc1.dispose();
        }
    }
    if (success) {
        chart.firePaint();
    }
    return(success);
}

/**
 * Checks newly added points to see if they fit inside current axis bounds.
 * @param npoints the number of points to check; it refers to the
 * number of points newly added to the end of all series
 * @return <code>true</code> if all points sit inside the current axis bounds,
 * which are set to the default values; <code>false</code> if any point sits
 * outside a current axis bound that is set to a default value
 */
public boolean checkNewPoints(int npoints)
{
    for (int i = 0; i < series.size(); i++) {
        ChartDataViewSeries s = getSeries(i);
        int maxj = s.maxIndex();
        for (int j = maxj - npoints + 1; j <= maxj; j++) {
            double x = s.getX(j);
            if (x != holeValue &&
                ((xaxis.getMaxIsDefault() && x > xaxis.getMax()) ||
                 (xaxis.getMinIsDefault() && x < xaxis.getMin())) )
            {
                return(false);
            }
            double y = s.getY(j);
            if (y != holeValue &&
                ((yaxis.getMaxIsDefault() && y > yaxis.getMax()) ||
                 (yaxis.getMinIsDefault() && y < yaxis.getMin())) )
            {
                return(false);
            }
        }
    }
    return(true);
}

/**
 * Returns the hole value for the chart data.  The hole value is
 * a floating point number used to represent a hole in the data.
 * Internally, <code>ChartData</code> places this value in the <TT>X</TT> and
 * <TT>Y</TT> arrays to represent a missing data value.
 * @return floating-point value used as a placeholder for missing
 * values
 */
public double getHoleValue()
{
    return(holeValue);
}

/**
 * Sets the hole value for the chart data. The hole value is
 * a floating point number used to represent a hole in the data.
 * Internally, <code>ChartData</code> places this value in the <TT>x</TT> and
 * <TT>y</TT> arrays to represent a missing data value.
 * The hole value should not be set to <code>Double.NAN</code>.
 *
 * <P>If the hole value is changed, previously designated hole values
 * will be drawn at their actual set value (that is, the old hole value).
 * If this old hole value is the default hole <code>Double.MAX_VALUE</code>,
 * chart will experience severe drawing slowness, as it will be trying to scale
 * its values to such a great range. If you are using this method
 * to change hole values, it is suggested that you first change the previous
 * hole values inside the data to something realistic or to the new
 * hole value you are about to set.
 * @param val floating-point value used as a placeholder for missing
 * values
 */
public void setHoleValue(double val)
{
    holeValue = val;
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Gets the <code>Batched</code> property. <code>ChartData</code> is a model for
 * the <code>ChartGroup</code> object. The <code>Batched</code> property
 * controls whether the <code>ChartGroup</code> is notified immediately of
 * <code>ChartData</code> changes, or if the changes are accumulated and sent
 * at a later date.
 * @return <code>true</code> if changes are accumulated; <code>false</code>
 * otherwise
 */
public boolean isBatched()
{
    return(batched);
}

/**
 * Sets the <code>Batched</code> property. <code>ChartData</code> is a model for
 * the <code>ChartGroup</code> object. The <code>Batched</code> property
 * controls whether the <code>ChartGroup</code> is notified immediately of
 * <code>ChartData</code> changes, or if the changes are accumulated and sent at
 * a later date.
 * @param isBatched <code>true</code> if changes are accumulated;
 * <code>false</code> otherwise
 */
public void setBatched(boolean isBatched)
{
    if (batched == isBatched) {
        return;
    }
    batched = isBatched;

    // Check batched because batched is not synchronized and we
    // want to call propagateChange() without a local lock.
    if (batched == false) {
        propagateChange();
    }
}

/**
 * Gets the <code>FastUpdate</code> property.  This property determines whether
 * column appends are done quickly, only recalculating and redrawing the new
 * data.
 * @return <code>true</code> if column appends are done quickly
 */
public boolean getFastUpdate()
{
    return(fastUpdate);
}

/**
 * Sets the <code>FastUpdate</code> property.  This property determines whether
 * column appends are done quickly, only recalculating and redrawing the new
 * data.
 * @param fast <code>true</code> if appends are to be done quickly
 */
public void setFastUpdate(boolean fast)
{
    if (fastUpdate == fast) {
        return;
    }
    fastUpdate = fast;
}

/**
 * Gets the <code>BufferPlotData</code> property.  This property determines
 * whether calculated plot data will be buffered to speed up drawing.
 * @return <code>true</code> if buffering is to be performed
 */
public boolean getBufferPlotData()
{
    return(bufferPlotData);
}

/**
 * Sets the <code>BufferPlotData</code> property.  This property determines whether
 * calculated plot data will be buffered to speed up drawing.
 * @param buffer if buffering is to be performed
 */
public void setBufferPlotData(boolean buffer)
{
    if (bufferPlotData == buffer) {
        return;
    }
    bufferPlotData = buffer;
}

/**
 * Gets the <code>ChartType</code> property of the <code>ChartData</code>
 * object, which determines how the chart data is plotted.
 * @return enumeration specifying the chart type
 */
public int getChartType()
{
    return(chartType);
}

private void updateAnnotationHandlers(List axes) {
    if (axes == null) {
        return;
    }
    JCAxis ax = null;
    for (int i=0; i < axes.size(); i++) {
        ax = (JCAxis)axes.get(i);
        if (ax != null) {
            ax.updateAnnoHandler(chartType);
        }
    }
}

private void updateAnnotationHandlers() {
    if (chart != null) {
        updateAnnotationHandlers(chart.getChartArea().getXAxes());
        updateAnnotationHandlers(chart.getChartArea().getYAxes());
    }
}

/**
 * Sets the <code>ChartType</code> property of the <code>ChartData</code>
 * object, which specifies the type of chart used to plot the data.
 * @param tp enumeration specifying the chart type; valid values are:
 * <code>JCChart.AREA</code>, <code>JCChart.AREA_RADAR</code>,
 * <code>JCChart.BAR</code>, <code>JCChart.CANDLE</code>,
 * <code>JCChart.HILO</code>, <code>JCChart.HILO_OPEN_CLOSE</code>,
 * <code>JCChart.PIE</code>, <code>JCChart.PLOT</code>,
 * <code>JCChart.POLAR</code>, <code>JCChart.RADAR</code>,
 * <code>JCChart.SCATTER_PLOT</code>, <code>JCChart.STACKING_AREA</code>,
 * and <code>JCChart.STACKING_BAR</code>.
 * @exception IllegalArgumentException If the new chart type is either
 * invalid or not currently supported.
 */
public void setChartType(int tp)
{
    if (chartType == tp) {
        return;
    }

    if (xaxis != null) {
        xaxis.annotationMethod = xaxis.savedXAxisAnnotationMethod;
        xaxis.updateAnnoDefaultPrecision(tp);
    }

    // Remember to update ChartType arrays in ChartConverter when new chart
    // types become supported
    switch (tp) {
    case JCChart.PLOT:
    case JCChart.SCATTER_PLOT:
    case JCChart.BAR:
    case JCChart.STACKING_BAR:
    case JCChart.PIE:
    case JCChart.HILO:
    case JCChart.HILO_OPEN_CLOSE:
    case JCChart.CANDLE:
    case JCChart.AREA:
    case JCChart.STACKING_AREA:
    case JCChart.POLAR:
    case JCChart.RADAR:
    case JCChart.AREA_RADAR:
        break;
    default:
        throw new IllegalArgumentException("Invalid chart type");
    }

    switch (chartType) {
    case JCChart.POLAR:
    case JCChart.RADAR:
    case JCChart.AREA_RADAR:
    case JCChart.PIE:
        if (yaxis != null) {
            yaxis.setVisible(true);
            yaxis.setChanged(true, Changeable.RECALC, false);
        }
        if (xaxis != null) {
            xaxis.setVisible(true);
            xaxis.setChanged(true, Changeable.RECALC, false);
        }
        break;
    case JCChart.BAR:
    case JCChart.STACKING_BAR:
        if (xaxis != null) {
            xaxis.setChanged(true, Changeable.RECALC, false);
        }
        break;
    }

    chartType = tp;
    updateAnnotationHandlers();
    getChartFormat(chartType);

    JCChartArea area = (chart == null) ? null : chart.getChartArea();
    switch (chartType) {
    case JCChart.PLOT:
    case JCChart.SCATTER_PLOT:
    case JCChart.HILO:
    case JCChart.HILO_OPEN_CLOSE:
    case JCChart.CANDLE:
        if (yaxis != null && xaxis != null) {
            boolean yaxisChanged = false;
            if (yaxis.getOriginPlacementIsDefault() &&
                yaxis.originPlacement.value != JCAxis.AUTOMATIC)
            {
                yaxis.originPlacement.value = JCAxis.AUTOMATIC;
                yaxisChanged = true;
            }
            if (yaxis.getPlacementIsDefault() &&
                yaxis.placement.value != JCAxis.AUTOMATIC)
            {
                yaxis.placement.value = JCAxis.AUTOMATIC;
                yaxisChanged = true;
            }
            if (yaxisChanged) {
                yaxis.setChanged(true, Changeable.RECALC, false);
            }
            if (xaxis.getPlacementIsDefault() &&
                xaxis.placement.value != JCAxis.AUTOMATIC)
            {
                xaxis.placement.value = JCAxis.AUTOMATIC;
                xaxis.setChanged(true, Changeable.RECALC, false);
            }
        }
        break;
    case JCChart.BAR:
    case JCChart.STACKING_BAR:
        if (yaxis != null && xaxis != null) {
            xaxis.savedXAxisAnnotationMethod = xaxis.annotationMethod;
            xaxis.placementAxis = yaxis;
            boolean allValueAnnotationMethods = true;
            if (area != null) {
                for (int i = 0; i < area.getXAxes().size(); i++) {
                    if (area.getXAxis(i).annotationMethod != JCAxis.VALUE) {
                        allValueAnnotationMethods = false;
                        break;
                    }
                }
            }
            // change annotation method for BAR like charts only if
            // all current x axes' annotation methods are by VALUE
            if (allValueAnnotationMethods) {
                xaxis.annotationMethod = JCAxis.POINT_LABELS;
            }
            // Unless the placement has been set explicitly,
            // default it to ORIGIN for x axis and MIN for y axis
            if (xaxis.getPlacementIsDefault()) {
                xaxis.placement.value = JCAxis.ORIGIN;
            }
            xaxis.setChanged(true, Changeable.RECALC, false);
            boolean yaxisChanged = false;
            if (yaxis.getPlacementIsDefault()) {
                yaxis.placement.value = JCAxis.MIN;
                yaxisChanged = true;
            }
            // Unless the origin placement has been set
            // explicitly, default it to ZERO
            if (yaxis.getOriginPlacementIsDefault()) {
                yaxis.originPlacement.value = JCAxis.ZERO;
                yaxisChanged = true;
            }
            if (yaxisChanged) {
                yaxis.setChanged(true, Changeable.RECALC, false);
            }
        }
        break;
    case JCChart.AREA:
    case JCChart.STACKING_AREA:
        if (yaxis != null && xaxis != null) {
            boolean yaxisChanged = false;
            if (yaxis.getPlacementIsDefault()) {
                yaxis.placement.value = JCAxis.MIN;
                yaxisChanged = true;
            }
            if (yaxis.getOriginPlacementIsDefault()) {
                yaxis.originPlacement.value = JCAxis.ZERO;
                yaxisChanged = true;
            }
            if (yaxisChanged) {
                yaxis.setChanged(true, Changeable.RECALC, false);
            }
            if (xaxis.getPlacementIsDefault()) {
                xaxis.placement.value = JCAxis.ORIGIN;
            }
            xaxis.placementAxis = yaxis;
            xaxis.setChanged(true, Changeable.RECALC, false);
        }
        break;
    case JCChart.POLAR:
    case JCChart.RADAR:
    case JCChart.AREA_RADAR:
    case JCChart.PIE:
    default:
        break;
    }

    // Chart Type has changed -- remove drawable and transient data
    if (area != null) {
        area.removeDrawable(this);
    }
    transientData = null;

    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the <code>Name</code> property of the <code>ChartData</code> object. The
 * <code>Name</code> property is used as an index for referencing particular
 * <code>ChartData</code> objects.
 * @return name of the <code>ChartData</code> object
 */
public String getName()
{
    return(name);
}

/**
 * Sets the <code>Name</code> property of the <code>ChartData</code> object. The
 * <code>Name</code> property is used as an index for referencing particular
 * <code>ChartData</code> objects.
 * @param nm name of the <code>ChartData</code> object
 */
public void setName(String nm)
{
    if (name != null && name.equals(nm)) {
        return;
    }

    if (nm != null) {
        name = nm;
    }

    setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the <code>DrawingOrder</code> property.
 * The <code>DrawingOrder</code> property determines the drawing order of items.
 * When the <code>DrawingOrder</code> property is changed, the order properties
 * of all <code>ChartDataView</code> instances managed by a single
 * <code>JCChart</code> object are normalized.
 * @return integer representing order
 */
public int getDrawingOrder()
{
    return(drawingOrder);
}

/**
 * Sets the <code>DrawingOrder</code> property of the <code>ChartDataView</code>
 * object. The <code>DrawingOrder</code> property determines the drawing order
 * of items.
 * When the <code>DrawingOrder</code> property is changed, the order properties
 * of all <code>ChartDataView</code> instances managed by a single
 * <code>JCChart</code> object are normalized.
 * @param neword new position of the <code>ChartDataView</code> object
 */
public void setDrawingOrder(int neword)
{
    if (neword < 0 || (chart != null &&
                       neword >= chart.getOriginalDataView().size()))
    {
        throw new IllegalArgumentException(
            "DrawingOrder property for ChartDataView is out of range");
    }
    drawingOrder = neword;
    setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Gets the <code>DrawFrontPlane</code> property.  <code>DrawFrontPlane</code>
 * determines whether a data view that has both axes on the front plane of a 3d
 * graph will draw on the front or back plane of that graph.  If
 * <code>true</code>, it will draw on the front plane; if <code>false</code>, it
 * will draw on the back plane.  If either axis associated with the data view is
 * on the back plane, this property will be ignored and the data view will
 * automatically be drawn on the back plane. This property is also ignored for
 * 3d chart types such as bar and stacking bar that automatically appear on the
 * front plane.
 * @return boolean representing whether to try to draw on the front or
 * back plane of a 3d chart
 */
public boolean getDrawFrontPlane()
{
    return(drawFrontPlane);
}

/**
 * Gets the <code>DrawFrontPlane</code> property.  <code>DrawFrontPlane</code>
 * determines whether a data view that has both axes on the front plane of a 3d
 * graph will draw on the front or back plane of that graph.  If
 * <code>true</code>, it will draw on the front plane; if <code>false</code>, it
 * will draw on the back plane.  If either axis associated with the data view is
 * on the back plane, this property will be ignored and the data view will
 * automatically be drawn on the back plane. This property is also ignored for
 * 3d chart types such as bar and stacking bar that automatically appear on the
 * front plane.
 * @param front <code>true</code> if trying to draw on front plane of 3d chart;
 * <code>false</code> otherwise
 */
public void setDrawFrontPlane(boolean front)
{
    drawFrontPlane = front;
    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the <code>DataSource</code> property for <code>ChartDataView</code>.
 * This <code>DataSource</code> property represents the class used to load data
 * into the <code>ChartDataView</code> class. The <code>DataSource</code>
 * property holds this populating class if it implements either
 * <code>EditableChartDataModel</code> or <code>ChartDataManageable</code>. If
 * the <code>DataSource</code> property is null, that means that the populating
 * data source class was discarded because it was no longer needed.
 *
 * @return object acting as a data source for <code>ChartDataView</code>;
 * this object implements either <code>EditableChartDataModel</code> or
 * <code>ChartDataManageable</code>
 */
public ChartDataModel getDataSource()
{
    return(dataSource);
}

/**
 * Sets the data source for this <code>ChartDataView</code>. The data source
 * class must implement either <code>ChartDataModel</code> or
 * <code>EditableChartDataModel</code>. JClass Chart will do the
 * correct operation based on the object provided.
 * If this data source class implements <code>ChartDataManageable</code>,
 * this data view will be registered as a listener for data changes.
 * The data source class will be saved in the <code>DataSource</code> property
 * for future reference if it implements either
 * <code>ChartDataManageable</code> or <code>EditableChartDataModel</code>.
 * Otherwise, since it is only needed to load data once,
 * the <code>DataSource</code> property will be set to null.
 * @param src object acting as a data source for <code>ChartDataView</code>;
 * this object must implement either <code>ChartDataModel</code> or
 * <code>EditableChartDataModel</code>
 */
public void setDataSource(ChartDataModel src)
{
    boolean oldBatched = false;

    // First, remove any listeners on the old data source
    if (dataSource instanceof ChartDataManageable) {
        ChartDataManageable manageable = (ChartDataManageable)dataSource;
        manageable.getChartDataManager().removeChartDataListener(this);
    }

    // Set the data source
    dataSource = src;

    if (dataSource == null) {
        // null data source : reset some properties to default values.
        // keep styles as they will be reused.
        series = new Vector();
        if (autoLabel && chart != null) {
            chart.getChartLabelManager().clearAutoLabels(this);
        }
        setX(null);
        setChanged(true, Changeable.RECALC               |
                         Changeable.NOTIFY_LEGEND_RECALC |
                         Changeable.NOTIFY_AXES_RECALC);
        return;
    }

    oldBatched = isBatched();
    setBatched(true);

    // Load the data
    getDataHandler().reset();
    getDataHandler().reloadAllPointLabels();
    getDataHandler().reloadAllSeriesLabels();
    getDataHandler().reloadDataSourceName();

    // If this data source implements ChartDataManageable, register
    // as a listener on data changes
    if (src instanceof ChartDataManageable) {
        ChartDataManageable manageable = (ChartDataManageable)src;
        manageable.getChartDataManager().addChartDataListener(this);
    }

    if (!(dataSource instanceof EditableChartDataModel) &&
        !(dataSource instanceof ChartDataManageable) )
    {
        // It is not editable or manageable, so we no longer need it
        dataSource = null;
    }

    if (autoLabel && chart != null) {
        chart.getChartLabelManager().generateAutoLabels(this);
    }

    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC |
               Changeable.NOTIFY_AXES_RECALC, false);
    setBatched(oldBatched);
}

/**
 * Gets a particular <code>PointLabel</code> from the <code>PointLabels</code>
 * property. The <code>PointLabels</code> property is an indexed property
 * consisting of a series of Strings representing the desired label for a
 * particular data point.
 * @param index index of the desired point
 * @return String representing the requested point label
 * @exception IllegalArgumentException If the point label index does
 * not address a valid point label.
 */
public String getPointLabel(int index)
{
    if (index < 0 || index >= pointLabels.size()) {
        throw new IllegalArgumentException("Invalid point label index");
    }
    return((String)pointLabels.get(index));
}

/**
 * Sets a particular <code>PointLabel</code> from the <code>PointLabels</code>
 * property. The <code>PointLabels</code> property is an indexed property
 * consisting of a series of Strings representing the desired label for a
 * particular data point.
 * @param index index of the desired point
 * @param label String representing the requested point label
 * @exception IllegalArgumentException If the point label index does
 * not address a valid point label.
 */
public void setPointLabel(int index, String label)
{
    if (index < 0 || index >= pointLabels.size()) {
        throw new IllegalArgumentException("Invalid point label index");
    }
    pointLabels.set(index, label);
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Gets the <code>PointLabels</code> property for the <code>ChartDataView</code>
 * object. The <code>PointLabels</code> property is an indexed property
 * consisting of a series of Strings representing the desired label for a
 * particular data point.
 * @return vector of String objects representing the point labels
 */
public List getPointLabels()
{
    return(pointLabels);
}

/**
 * Gets the number of point labels set on this view.
 * @return number of point labels set on this view
 */
public int getNumPointLabels()
{
    return(pointLabels.size());
}

/**
 * Sets the <code>PointLabels</code> property for the <code>ChartDataView</code>
 * object. The <code>PointLabels</code> property is an indexed property
 * consisting of a series of Strings representing the desired label for a
 * particular data point.
 * @param labels collection of String objects representing the point labels
 */
public void setPointLabels(List labels)
{
    pointLabels = labels;
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Retrieves an array of image map information, one for each point cluster.
 * @return The current cluster image map information array.
 */
public ImageMapInfo[] getClusterImageMapInfo()
{
    return(clusterImageMapInfo);
}

/**
 * Sets an array of image map information, one for each point cluster.
 * @param clusterImageMapInfo The new cluster image map information array.
 */
public void setClusterImageMapInfo(ImageMapInfo[] clusterImageMapInfo)
{
    this.clusterImageMapInfo = clusterImageMapInfo;
}

/**
 * Retrieve the image map information for this <code>ChartDataView</code>.  This
 * object is used to store the legend title image map information.
 * @return The current image map information.
 */
public ImageMapInfo getImageMapInfo()
{
    return(imageMapInfo);
}

/**
 * Set the image map information for this <code>ChartDataView</code>.  This
 * object is used to store the legend title image map information.
 * @param imageMapInfo The new image map information.
 */
public void setImageMapInfo(ImageMapInfo imageMapInfo)
{
    this.imageMapInfo = imageMapInfo;
}

/**
 * Gets the value of the <code>Series</code> property for this
 * <code>ChartDataView</code> instance. The <code>Series</code> property is an
 * indexed property that contains all data series for a particular
 * <code>ChartDataView</code>. The order of <code>ChartDataViewSeries</code>
 * objects in the series array corresponds to the drawing order.
 * @return list object containing <code>ChartDataViewSeries</code> instances
 * representing the requested data series
 */
public List getSeries()
{
    return(new Vector(series));
}

/**
 * Gets the value of the <code>Series</code> property at a particular index.
 * The <code>Series</code> property is an indexed property that
 * contains all data series for a particular <code>ChartDataView</code>.
 * The order of <code>ChartDataViewSeries</code> objects in the
 * <code>Series</code> array corresponds to the drawing order.
 * @param index index of the data series to retrieve
 * @exception IllegalArgumentException If the provided index does
 * not address an existing ChartDataViewSeries.
 * @return values of the <code>Series</code> property at a particular index
 */
public ChartDataViewSeries getSeries(int index)
{
    if (index < 0 || index >= series.size()) {
        throw new IllegalArgumentException(ERRMSG_INVALID_SERIES_INDEX);
    }
    return((ChartDataViewSeries)series.get(index));
}

/**
 * Sets the value of the <code>Series</code> property for this
 * <code>ChartDataView</code> instance. The <code>Series</code> property is an
 * indexed property that contains all data series for a particular
 * <code>ChartDataView</code>. The order of <code>ChartDataViewSeries</code>
 * objects in the <code>Series</code> array corresponds to the drawing order.
 * @param newseries list object containing <code>ChartDataViewSeries</code>
 * instances in the desired drawing order; the existing
 * series property will be replaced
 */
public void setSeries(List newseries)
{
    if (autoLabel && chart != null) {
        chart.getChartLabelManager().removeAutoLabels(this);
    }

    if (newseries == null) {
        series = new Vector();
    }
    else {
        series = new Vector(newseries);
    }

    if (autoLabel && chart != null) {
        chart.getChartLabelManager().generateAutoLabels(this);
    }

    // Allocate chart styles (which allocate default colors) to be
    // consistent between chart types.
    if (chart != null) {
        if (chartType == JCChart.PIE) {
            ((JCPieChartFormat)getChartFormat()).getOtherStyle();
        }
        for (int i = 0; i < series.size(); i++) {
            getChartStyle(i);
        }
    }
    setChanged(true, Changeable.RECALC               |
                     Changeable.NOTIFY_LEGEND_RECALC |
                     Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Sets the value of the <code>Series</code> property at a particular index.
 * The <code>Series</code> property is an indexed property that
 * contains all data series for a particular <code>ChartDataView</code>.
 * The order of <code>ChartDataViewSeries</code> objects in the
 * <code>Series</code> array corresponds to the drawing order.
 * @param index index of the data series to retrieve
 * @exception IllegalArgumentException If the provided index does
 * not address an existing ChartDataViewSeries.
 * @param s
 */
public void setSeries(int index, ChartDataViewSeries s)
{
    if (index < 0 || index >= series.size()) {
        throw new IllegalArgumentException(ERRMSG_INVALID_SERIES_INDEX);
    }
    series.set(index, s);
    if (autoLabel && chart != null) {
        for (int j = 0; j <= s.maxIndex(); j++) {
            chart.getChartLabelManager().setAutoLabelText(this, index, j);
        }
    }
    setChanged(true, Changeable.RECALC               |
                     Changeable.NOTIFY_LEGEND_RECALC |
                     Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Adds a data series to the <code>Series</code> property at the specified
 * index. <code>addSeries()</code> creates a new
 * <code>ChartDataViewSeries</code> instance and adds it to the
 * <code>ChartDataView</code> instance at the requested position if possible.
 * @param index desired position of the new data series
 * @return new <code>ChartDataViewSeries</code> instance representing
 * the new data series
 */
public ChartDataViewSeries addSeries(int index)
{
    ChartDataViewSeries s = null;
    if (index < 0) {
        index = 0;
    }
    if (index > series.size()) {
        index = series.size();
    }

    s = new ChartDataViewSeries();
    s.init(null, null, this);
    s.setDrawingOrder(index);
    series.add(index, s);
    if (autoLabel && chart != null) {
        chart.getChartLabelManager().addAutoLabels(this, index, -1);
    }
    setChanged(true, Changeable.RECALC);
    return(s);
}

/**
 * Adds a <code>ChartDataViewSeries</code> to the view at the end of
 * the list.
 * @param s
 */
public void addSeries(ChartDataViewSeries s)
{
    s.setDrawingOrder(series.size());
    series.add(s);
    if (autoLabel && chart != null) {
        chart.getChartLabelManager().addAutoLabels(this, series.size() - 1, -1);
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Removes a single data series from the <code>Series</code> property of
 * the <code>ChartDataView</code> instance. Once the item is removed, the
 * order properties for each <code>ChartDataViewSeries</code> are normalized
 * so that they correspond exactly to their index in the series
 * array.
 * @param index position of the <code>ChartDataViewSeries</code> instance
 * that is to be removed
 * @exception IllegalArgumentException If the index does not address
 * a valid <code>ChartDataViewSeries</code>.
 */
public void removeSeries(int index)
{
    if (index < 0 || index >= series.size()) {
        throw new IllegalArgumentException(ERRMSG_INVALID_SERIES_INDEX);
    }
    series.remove(index);
    if (autoLabel && chart != null) {
        chart.getChartLabelManager().removeAutoLabels(this, index, -1);
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * A convenience method to retrieve a data series index given the
 * name of the data series.
 * @param name name of the <code>ChartDataViewSeries</code> instance
 * @return index of the data series
 * name; -1 if none found
 */
public int getSeriesIndex(String name)
{
    try {
        for (int i = 0; i < series.size(); i++) {
            ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
            if (name.equals(s.getName())) {
                return(i);
            }
        }
    }
    catch (Exception e) {
    }
    return(-1);
}

/**
 * A convenience method to retrieve a data series index given data series
 * itself.
 * @param srs the <code>ChartDataViewSeries</code> instance
 * @return index of the data series; -1 if none found
 */
public int getSeriesIndex(ChartDataViewSeries srs)
{
    return(series.indexOf(srs));
}

/**
 * Gets the <code>Format</code> class for the current chart type.  If the
 * current chart type should have a format class but has not been created yet,
 * one will be created and returned.
 * @return the <code>Format</code> class for the current chart type; one of
 * <code>JCAreaChartFormat</code>, <code>JCBarChartFormat</code>,
 * <code>JCCandleChartFormat</code>, <code>JCHLOCChartFormat</code>,
 * <code>JCHiloChartFormat</code>, <code>JCPieChartFormat</code>, or
 * <code>JCPolarRadarChartFormat</code>
 * depending on the chart type, or <code>null</code> if the current chart type
 * should not have a format class
 */
public JCChartTypeFormat getChartFormat()
{
    return(getChartFormat(chartType));
}

/**
 * Gets the <code>Format</code> class for the specified chart type.  If the
 * specified chart type should have a format class but has not been created yet,
 * one will be created and returned.
 * @param type the chart type
 * @return the <code>Format</code> class for the specified chart type; one of
 * <code>JCAreaChartFormat</code>, <code>JCBarChartFormat</code>,
 * <code>JCCandleChartFormat</code>, <code>JCHLOCChartFormat</code>,
 * <code>JCHiloChartFormat</code>, <code>JCPieChartFormat</code>, or
 * <code>JCPolarRadarChartFormat</code>
 * depending on the chart type, or <code>null</code> if the specified chart type
 * should not have a format class
 */
public JCChartTypeFormat getChartFormat(int type)
{
    // First check if the format exists already.  If so return it.
    JCChartTypeFormat fmt = getExistingChartFormat(type);
    if (fmt != null) {
        return(fmt);
    }

    // BAR and STACKING_BAR share the same format class.  So do AREA and
    // STACKING_AREA.
    if (type == JCChart.STACKING_BAR) {
        type = JCChart.BAR;
    }
    else if (type == JCChart.STACKING_AREA) {
        type = JCChart.AREA;
    }
    if (type <= 0 || type >= format_class_names.length) {
        return(null);
    }

    String fmtName = format_class_names[type];
    if (fmtName != null) {
        try {
            fmt = (JCChartTypeFormat) Class.forName(fmtName).newInstance();
            fmt.setParent(this);
            formatClassHashtable.put(new Integer(type), fmt);
        }
        catch (Exception e) {}
    }
    return(fmt);
}

/**
 * Gets the <code>Format</code> class for the specified chart type if it exists.
 * Does not create a new format class if it doesn't exist.
 * @param type the chart type
 * @return the <code>Format</code> class for the specified chart type; one of
 * <code>JCAreaChartFormat</code>, <code>JCBarChartFormat</code>,
 * <code>JCCandleChartFormat</code>, <code>JCHLOCChartFormat</code>,
 * <code>JCHiloChartFormat</code>, <code>JCPieChartFormat</code>, or
 * <code>JCPolarRadarChartFormat</code>
 * depending on the chart type, or <code>null</code> if the specified chart type
 * should not have a format class
 */
public JCChartTypeFormat getExistingChartFormat(int type)
{
    // BAR and STACKING_BAR share the same format class.  So do AREA and
    // STACKING_AREA.
    if (type == JCChart.STACKING_BAR) {
        type = JCChart.BAR;
    }
    else if (type == JCChart.STACKING_AREA) {
        type = JCChart.AREA;
    }
    if (type < 0 || type >= format_class_names.length) {
        return(null);
    }

    if (format_class_names[type] != null) {
        return((JCChartTypeFormat)formatClassHashtable.get(new Integer(type)));
    }
    return(null);
}

/**
 * Gets the value of the <code>XAxis</code> property.  The <code>XAxis</code>
 * property determines the <TT>x</TT> axis against which the data in
 * <code>ChartDataView</code> is plotted.
 * @return <code>JCAxis</code> object representing the <TT>x</TT> axis for the
 * data
 */
public JCAxis getXAxis()
{
    return(xaxis);
}

/**
 * Sets the value of the <code>XAxis</code> property. The <code>XAxis</code>
 * property determines the <TT>x</TT> axis against which the data in
 * <code>ChartDataView</code> is plotted.
 * @param ax <code>JCAxis</code> object representing the <TT>x</TT> axis for the
 * data
 */
public void setXAxis(JCAxis ax)
{
    xaxis = ax;
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Gets the value of the <code>YAxis</code> property. The <code>YAxis</code>
 * property determines the <TT>y</TT> axis against which the data in
 * ChartDataView is plotted.
 * @return <code>JCAxis</code> object representing the <TT>y</TT> axis for the
 * data
 */
public JCAxis getYAxis()
{
    return(yaxis);
}

/**
 * Sets the value of the <code>YAxis</code> property. The <code>YAxis</code>
 * property determines the <TT>y</TT> axis against which the data in
 * ChartDataView is plotted.
 * @param ax <code>JCAxis</code> object representing the <TT>y</TT> axis for
 * the data
 */
public void setYAxis(JCAxis ax)
{
    yaxis = ax;
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Performs a map operation by mapping the provided screen pixel
 * information to data values along the <TT>x</TT> and <TT>y</TT> axes of the
 * <code>ChartDataView</code> instance.
 * @param x <i>X</i> value in screen pixels
 * @param y <i>Y</i> value in screen pixels
 * @return <code>JCDataCoord</code> instance
 * @see com.klg.jclass.chart.JCDataCoord
 */
public JCDataCoord map(int x, int y)
{
    // adjust co-ordinate to be withing chart area drawing area.
    Rectangle r1 = chart.getChartArea().getDrawingArea();
    Point     l  = chart.getChartArea().getLocation();
    x -= (r1.x + l.x);
    y -= (r1.y + l.y);

    JCDataCoord ret = new JCDataCoord(holeValue, holeValue);
    if (chartType == JCChart.POLAR ||
        chartType == JCChart.RADAR ||
        chartType == JCChart.AREA_RADAR)
    {
        if (yaxis != null) {
            JCDataCoord dataValue = yaxis.toDataPolar(x,y);
            ret.x = dataValue.getX();
            ret.y = dataValue.getY();
        }
    }
    else {
        if (xaxis != null) {
            ret.x = xaxis.i_vertical ? xaxis.toData(y) : xaxis.toData(x);
        }
        if (yaxis != null) {
            ret.y = yaxis.i_vertical ? yaxis.toData(y) : yaxis.toData(x);
        }
    }
    return(ret);
}

/**
 * Performs an unmap operation by taking floating-point <TT>x</TT> and
 * <TT>y</TT> values meant to represent a point along the <TT>x</TT>
 * and <TT>y</TT> axes of the data.
 * @param xval floating-point <TT>x</TT> value
 * @param yval floating-point <TT>y</TT> value
 * @return AWT Point object representing the location in screen
 * pixels (relative to the Chart component)
 */
public java.awt.Point unmap(double xval,double yval)
{
    Rectangle r1 = chart.getChartArea().getDrawingArea();
    Point l = chart.getChartArea().getLocation();
    Point p = new Point(r1.x + l.x, r1.y + l.y);
    // Resulting point is relative to chart area drawing area.  Adjust to
    // be relative to Chart component.
    if (chartType == JCChart.POLAR ||
        chartType == JCChart.RADAR ||
        chartType == JCChart.AREA_RADAR)
    {
        if (yaxis != null) {
            Point pnt = yaxis.toPixelPolar(
                JCChartUtil.convertAngle(chart.getChartArea().getAngleUnit(),
                                    JCChartUtil.RADIANS, xval), yval, true);
            p.x += pnt.x;
            p.y += pnt.y;
        }
    }
    else {
        if (xaxis != null) {
            if (xaxis.i_vertical) {
                p.y += xaxis.toPixel(xval);
            }
            else {
                p.x += xaxis.toPixel(xval);
            }
        }
        if (yaxis != null) {
            if (yaxis.i_vertical) {
                p.y += yaxis.toPixel(yval);
            }
            else {
                p.x += yaxis.toPixel(yval);
            }
        }
    }
    return(p);
}

/** Same as <code>map()</code>.  Converts pixel coordinates to data coordinates.
 * @param x <i>X</i> value in screen pixels
 * @param y <i>Y</i> value in screen pixels
 * @return <code>JCDataCoord</code> instance
 * @see com.klg.jclass.chart.JCDataCoord
*/
public JCDataCoord coordToDataCoord(int x, int y)
{
    return(map(x, y));
}

/**
 * Same as <code>unmap()</code>.  Converts data coordinates to pixel
 * coordinates.
 * @param xval floating-point <TT>x</TT> value
 * @param yval floating-point <TT>y</TT> value
 * @return AWT Point object representing the location in screen
 * pixels (relative to the Chart component)
 */
public java.awt.Point dataCoordToCoord(double xval,double yval)
{
    return(unmap(xval, yval));
}

/**
 * Similar to <code>unpick()</code> for a specific data view.  Converts a
 * <code>JCDataIndex</code> instance (containing a data view, series, point) to
 * pixel values relative to the Chart component.
 * @param di <code>JCDataIndex</code> instance containing data view, series, and
 * point
 * @return AWT Point object representing the location is screen pixels
 * relative to the Chart component
 */
public Point dataIndexToCoord(JCDataIndex di)
{
    Point p = chart.unpick(this, di.point, di.seriesIndex);
    return(p);
}

/**
 * Similar to <code>pick()</code> for a specific data view.  Converts pixel
 * values relative to the Chart component to a <code>JCDataIndex</code> instance
 * containing the data series and point closest to that location.
 * @param x <i>X</i> value of screen position
 * @param y <i>Y</i> value of screen position
 * @param focus pick focus value used to determine distance between point
 * and the data drawn by the chart; if equal to
 * <code>ChartDataView.PICK_FOCUS_LOCAL</code>, the <code>pickFocus</code>
 * property of the <code>ChartDataView</code> will be used
 * @return <code>JCDataIndex</code> instance containing data view, series, and
 * point
 */
public JCDataIndex coordToDataIndex(int x, int y, int focus)
{
    Point l = chart.getChartArea().getLocation();
    return(chart.getChartArea().pick(new Point(x - l.x, y - l.y), this, focus));
}

/**
 * Determines if chart type of this view is a 3D type.  Currently, this
 * is limited to either bar or stacking bar charts. <br><B>NOTE:</b> Pie charts
 * are technically 3D, but are handled differently.
 * @return <code>true</code> if the chart type of this view is either
 * <code>BAR</code> or <code>STACKING_BAR</code>
 */
public boolean is3DChartType()
{
    if (chartType == JCChart.BAR || chartType == JCChart.STACKING_BAR) {
        return(true);
    }
    return(false);
}

/**
 * Gets the value of the <code>NumSeries</code> parameter, which determines
 * how many data series there are in a <code>ChartDataView</code>. Used in
 * conjunction with <code>getSeries(int index)</code> to retrieve data series.
 * @return number of data series managed by <code>ChartDataView</code>
 */
public int getNumSeries()
{
    return((series == null) ? 0 : series.size());
}

/**
 * Output routine used for debugging.
 * @return
 */
public String toString()
{
    String s = getName();
    return((s == null) ?
        JCChartBundle.string(JCChartBundle.key142, locale) : s);
}

void calcXMinMax(MinMax x)
{
    // Calculate step for this data view
    step = Double.MAX_VALUE;
    for (int i = 0; i < series.size(); i++) {
        ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
        if (s.isIncluded()) {
            s.getXMinMax(x);
            s.calcStep(this);
        }
    }

    // If step less than that of xaxis, use it for x axis
    if (step != Double.MAX_VALUE && step != 0.0 && step < xaxis.step) {
        xaxis.step = step;
    }

    // If step calc didn't work, revert to default value
    if (xaxis.step == Double.MAX_VALUE) {
        xaxis.step = 1.0;
    }
}

void calcYMinMax(MinMax y)
{
    for (int i = 0; i < series.size(); i++) {
        ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
        if (s.isIncluded()) {
            s.getYMinMax(y);
        }
    }
}

private void calcYMinMaxOfTotals(MinMax y)
{
    Point firstLast = getFirstLast();
    if (firstLast == null) {
        return;
    }
    int nseries = series.size();

    MinMax newmm = new MinMax();
    for (int i = firstLast.x; i <= firstLast.y; i++) {
        MinMax totalmm = new MinMax(0.0, 0.0);
        for (int j = 0; j < nseries; j++) {
            ChartDataViewSeries s = (ChartDataViewSeries)series.get(j);
            if (s.isIncluded() == false) {
                continue;
            }
            double val = s.getY(i);
            if (val == holeValue) {
                continue;
            }
            if (val >= 0.0) {
                totalmm.max += val;
            }
            else {
                totalmm.min += val;
            }
        }
        newmm.union(totalmm);
    }
    y.union(newmm);
}

/**
 * Gets the data bounds for this data object.
 * @param xLimits
 * @param yLimits
 */
void getDataBounds(MinMax xLimits, MinMax yLimits)
{
    boolean is_polar, is_radar, polar_or_radar;
    boolean is_HLOC_or_candle;

    is_polar = (chartType == JCChart.POLAR);
    is_radar = (chartType == JCChart.RADAR ||
                chartType == JCChart.AREA_RADAR);
    polar_or_radar = (is_polar || is_radar);
    is_HLOC_or_candle = (chartType == JCChart.HILO_OPEN_CLOSE ||
                         chartType == JCChart.CANDLE);

    // if limits unitialized, initialize them
    if (xLimits.min == Double.MAX_VALUE && xLimits.max == -Double.MAX_VALUE) {
        calcXMinMax(xLimits);
    }
    if (yLimits.min == Double.MAX_VALUE && yLimits.max == -Double.MAX_VALUE) {
        calcYMinMax(yLimits);
    }
    // if still unitialized (probably no included series), set them to
    // a default value that will at least allow the lines to be clipped
    // and drawn.
    if (xLimits.min == Double.MAX_VALUE && xLimits.max == -Double.MAX_VALUE) {
        xLimits.min = -Double.MAX_VALUE;
        xLimits.max = Double.MAX_VALUE;
    }
    if (yLimits.min == Double.MAX_VALUE && yLimits.max == -Double.MAX_VALUE) {
        yLimits.min = -Double.MAX_VALUE;
        yLimits.max = Double.MAX_VALUE;
    }

    if (is_HLOC_or_candle) {
        double inc = JCChartUtil.abs(xaxis.getTickSpacing());
        // make allowances for ticks/candle width
        xLimits.min -= inc;
        xLimits.max += inc;
    }

    if ((!xaxis.min.isDefault && xaxis.min.value > xLimits.min) ||
        (xaxis.formula != null && xaxis.formula.getOriginator() != null &&
                                xaxis.min.value > xLimits.min) ||
        (polar_or_radar))
    {
        xLimits.min = xaxis.min.value;
    }
    if ((!xaxis.max.isDefault && xaxis.max.value < xLimits.max) ||
        (xaxis.formula != null && xaxis.formula.getOriginator() != null &&
                                xaxis.max.value < xLimits.max) ||
        (polar_or_radar))
    {
        xLimits.max = xaxis.max.value;
    }

    if ((!yaxis.min.isDefault && yaxis.min.value > yLimits.min) ||
        (yaxis.formula != null && yaxis.formula.getOriginator() != null &&
                                yaxis.min.value > yLimits.min) ||
        (polar_or_radar))
    {
        yLimits.min = yaxis.min.value;
    }

    if (is_polar && yLimits.min < 0.0) {
        if (JCChartUtil.abs(yLimits.min) > yLimits.max) {
            yLimits.max = JCChartUtil.abs(yLimits.min);
        }
        yLimits.min = 0.0;
    }

    if ((!yaxis.max.isDefault && yaxis.max.value < yLimits.max) ||
        (yaxis.formula != null && yaxis.formula.getOriginator() != null &&
                                yaxis.max.value < yLimits.max) ||
        (polar_or_radar))
    {
        yLimits.max = yaxis.max.value;
    }
}

DoubleCoord getOrigin(MinMax xLimits, MinMax yLimits, boolean primary)
{
    // Calculate the axes origins.
    DoubleCoord rval = new DoubleCoord(xLimits.min, yLimits.min);

    if (!primary) {
        return(rval);
    }

    int unit = chart.getChartArea().getAngleUnit();
    if (xaxis.origin.isDefault) {
        if (chartType == JCChart.POLAR ||
            chartType == JCChart.RADAR ||
            chartType  == JCChart.AREA_RADAR)
        {
            switch(unit) {
            case JCChartUtil.DEGREES:
                rval.x = 90.0;
                break;
            case JCChartUtil.RADIANS:
                rval.x = JCChartUtil.M_PI_2;
                break;
            case JCChartUtil.GRADS:
                rval.x = 100.0;
                break;
            }
        }
        else {
            switch(xaxis.originPlacement.value) {
            case JCAxis.ZERO:
                rval.x = 0.0;
                break;
            case JCAxis.MIN:
                rval.x = xLimits.min;
                break;
            case JCAxis.MAX:
                rval.x = xLimits.max;
                break;
            case JCAxis.AUTOMATIC:
                if (xLimits.min <= 0.0 && xLimits.max >= 0.0)    {
                    rval.x = 0.0;
                }
                else {
                    rval.x = xLimits.min;
                }
                break;
            }
        }
    }
    else {
        rval.x = JCChartUtil.boundAngle(unit, xaxis.origin.value);
    }

    if (chartType == JCChart.STACKING_BAR ||
        chartType == JCChart.STACKING_AREA)
    {
        if (yaxis.logarithmic) {
            rval.y = JCChartUtil.max(yLimits.min, 0.0);
        }
        else {
            rval.y = 0.0;
        }
    }
    else if (chartType == JCChart.POLAR ||
             chartType == JCChart.RADAR ||
             chartType == JCChart.AREA_RADAR)
    {
        if (yaxis.reversed) {
            rval.y = yLimits.min;
        }
        else {
            rval.y = yLimits.max;
        }
    }
    else {
        if (yaxis.origin.isDefault) {
            switch(yaxis.originPlacement.value) {
            case JCAxis.ZERO:
                rval.y = 0.0;
                break;
            case JCAxis.MIN:
                rval.y = yLimits.min;
                break;
            case JCAxis.MAX:
                rval.y = yLimits.max;
                break;
            case JCAxis.AUTOMATIC:
                if ((chartType == JCChart.BAR || chartType == JCChart.AREA) &&
                    !yaxis.logarithmic)
                {
                    rval.y = 0.0;
                }
                else {
                    if (yLimits.min <= 0.0 && yLimits.max >= 0.0) {
                        rval.y = 0.0;
                    }
                    else {
                        rval.y = yLimits.min;
                    }
                }
            }
        }
        else {
            rval.y = yaxis.origin.value;
        }
    }
    return(rval);
}


void clearLimits()
{
    if (xaxis != null) {
        xaxis.clearLimits();
        xaxis.step = Double.MAX_VALUE;
    }
    if (yaxis != null) {
        yaxis.clearLimits();
        yaxis.step = Double.MAX_VALUE;
    }
    xLimits = null;
    yLimits = null;
}

void calcLimits()
{
    if (xaxis != null) {
        xLimits = new MinMax();
        calcXMinMax(xLimits);
        xaxis.setLimits(xLimits);
    }
    if (yaxis != null) {
        yLimits = new MinMax();
        boolean is100Percent = false;
        switch (chartType) {
            case JCChart.STACKING_BAR :
                is100Percent =
                    ((JCBarChartFormat)getChartFormat()).is100Percent();
                break;
            case JCChart.STACKING_AREA :
                is100Percent =
                    ((JCAreaChartFormat)getChartFormat()).is100Percent();
                break;
        }

        if (chartType == JCChart.STACKING_BAR ||
            chartType == JCChart.STACKING_AREA)
        {
            calcYMinMaxOfTotals(yLimits);
            if (is100Percent) {
                if (yLimits.min < 0.0) {
                    yLimits.min = -100.0;
                }
                else {
                    yLimits.min = 0.0;
                }
                if (yLimits.max > 0.0) {
                    yLimits.max = 100.0;
                }
                else {
                    yLimits.max = 0.0;
                }
            }
        }
        else {
            calcYMinMax(yLimits);
        }
        yaxis.setLimits(yLimits);
    }
}

/**
 * Gets the the min and max <i>X</i> values in this data view.
 * @return min and max <i>X</i> values in this data view
 */
public MinMax getXLimits()
{
    return(xLimits);
}

/**
 * Gets the the min and max <i>Y</i> values in this data view.
 * @return min and max <i>Y</i> values in this data view
 */
public MinMax getYLimits()
{
    return(yLimits);
}

void getDataBoundValues(MinMax x, MinMax y)
{
    if (xLimits != null) {
        x.reset(xLimits);
    }
    if (yLimits != null) {
        y.reset(yLimits);
    }
}

/**
 * Determines if view has been marked as changed.
 * @return <code>true</code> if view has been marked as changed
 */
public boolean isChanged()
{
    return(getChanged());
}

/**
 * Determines if view has been marked for a particular type of change.
 * @param mask the type of change to check for; see Changeable for
 *             a list of masks
 * @return <code>true</code> if view has been marked as changed
 */
public boolean isChanged(int mask)
{
    return(getChanged(mask));
}

/**
 * Returns the value of the <code>Changed</code> property.
 * @return value of the <code>Changed</code> property
 */
public boolean getChanged()
{
    return(changed);
}

/**
 * Checks for the existence of the specified flag in the
 * <code>ChangedFlag</code> property.
 * @param mask the type of change to check for; see Changeable for
 *             a list of masks
 * @return
 */
public boolean getChanged(int mask)
{
    return((changedFlag & mask) == mask);
}

/**
 * Convenience method used to set the changed state
 * without update.
 * @param b
 */
protected void setChanged(boolean b)
{
    changed = b;
}

/**
 * Sets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property manages whether the data view
 * requires recalculation.  If set to <code>true</code>, a recalculation
 * may be triggered.
 * @param b
 * @param cf
 */
public void setChanged(boolean b, int cf)
{
    changed = b;

    // Only changed=true is interesting, as it can propagate elsewhere
    if (!changed) {
        changedFlag = cf;
        return;
    }
    else {
        changedFlag |= cf;
    }
    // Mark axes, chart area and legend as changed.
    markAsChanged(cf);

    // Possibly force update
    propagateChange();

}

/**
 * Utility routine used to force an update as a result of a
 * change inside <code>ChartDataView</code>.
 */
private void propagateChange()
{
    // Only send change if change is true
    if (!changed) {
        return;
    }

    // Don't send change if ChartDataView is batched
    if (isBatched()) {
        return;
    }

    // Don't send change if Chart is null or batched.
    if (chart == null) {
        return;
    }
    if (chart.isBatched()) {
        return;
    }

    // JCChart.update will invalidate, validate and repaint.
    chart.update();

    // Now that changes have been marked and propagated, clear the flags here
    // DAM unsure about this -- CA clears out data view changes flags
    // too.  Doing it here only ensures CDV isn't marked as changed when
    // a chart finally becomes visible.  I don't want to change it this
    // late in the game, though, for fear that it might cause other
    // problems.
    setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Utility routine used to mark appropriate objects as changed
 * as a result of a change in <code>ChartDataView</code>. <br><B>Note:</b> The
 * changes are not allowed to propagate up to the parent.
 * @param cf
 */
private void markAsChanged(int cf)
{
    // Axes must be marked as changed.
    // Note: Synchronization handled by setChanged() calls in
    // other objects.  No impact on ChartDataView
    int axisCF = Changeable.NO_CHANGE;
    if ((cf & Changeable.NOTIFY_AXES_RECALC) > 0) {
        axisCF |= Changeable.RECALC;
    }
    else if ((cf & Changeable.NOTIFY_AXES_REDRAW) > 0) {
        axisCF |= Changeable.REDRAW;
    }
    if (axisCF != Changeable.NO_CHANGE) {
        if (xaxis != null) {
            xaxis.setChanged(true, axisCF, false);
        }
        if (yaxis != null) {
            yaxis.setChanged(true, axisCF, false);
        }
    }

    // ChartArea should be marked as changed.
    if (chart != null && chart.getChartArea() != null) {
        chart.getChartArea().setChanged(true, cf & BASIC_CHANGE_MASK, false);
    }

    // Mark legend as changed if necessary
    int legendCF = Changeable.NO_CHANGE;
    if ((cf & Changeable.NOTIFY_LEGEND_RECALC) > 0) {
        legendCF |= Changeable.RECALC;
    }
    else if ((cf & Changeable.NOTIFY_LEGEND_REDRAW) > 0) {
        legendCF |= Changeable.REDRAW;
    }

    if (legendCF != Changeable.NO_CHANGE) {
        if (chart != null && chart.getLegend() != null) {
            chart.getLegend().setChanged(true, legendCF, false);
        }
    }
}

/**
 * Sets the value of the <code>Changed</code> property, with option to
 * prevent parent updating via the <code>updateParent</code> parameter.
 * <br><B>Note:</b> <code>updateParent</code> is independent of Batched
 * @param b
 * @param cf
 * @param updateParent
 */
public void setChanged(boolean b, int cf, boolean updateParent)
{
    if ((changed == b) && ((changedFlag & cf) == cf)) {
        return;
    }
    if (updateParent) {
        setChanged(b, cf);
    }
    else {
        changed = b;
        if (changed)  {
            changedFlag |= cf;
            markAsChanged(cf);
        }
        else {
            changedFlag = cf;
        }
    }
}

/**
 * Calculates the set of points to be covered by a bar chart by
 * finding the largest <code>FirstPoint</code> value and the smallest
 * <code>LastPoint</code> value. Both values will be tempered by the
 * smallest array of points.
 * @return
 */
public Point getFirstLast()
{
    if (series == null) {
        return(null);
    }
    int f = -1;
    int l = Integer.MAX_VALUE;
    int minSize = Integer.MAX_VALUE;
    for (int i = 0; i < series.size(); i++) {
        ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
        if (!s.isVisible()) {
            continue;
        }
        if (f < s.getFirstPoint()) {
            f = s.getFirstPoint();
        }
        if (l > s.getLastPoint()) {
            l = s.getLastPoint();
        }
        int indx = s.maxIndex();
        if (minSize > indx) {
            minSize = indx;
        }
    }
    // No visible data series
    if (minSize == Integer.MAX_VALUE) {
        return(null);
    }

    // If the first data point is bigger than the min size, forget it!
    if (f > minSize) {
        return(null);
    }

    if (l > minSize) {
        l = minSize;
    }

    // Nothing to plot.  Shouldn't run into this case, but double-check good
    if (f > l) {
        return(null);
    }

    return(new Point(f, l));
}

/**
 * Returns the number of children data series that are marked as visible.
 * @return number of children data series that are marked as visible
 */
public int getVisibleDataSeries()
{
    int rval = 0;
    for (int i = 0; i < series.size(); i++) {
        ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
        if (s.isVisible()) {
            rval++;
        }
    }
    return(rval);
}

/**
 * Calculates data associated with view data.  For example, pie charts
 * need percentage data to be calculated based on the x,y data specified
 * here.
 */
public void calcTransientData()
{
    boolean pieType = (getChartType() == JCChart.PIE);

    if (pieType) {
        ChartDraw pie = chart.getChartArea().getDrawable(this);
        if (pie != null) {
            pie.calcTransientData();
        }
    }
}

/**
 * Sets the <code>Inverted</code> property of the data view.  If the
 * <code>Inverted</code> property is set, the X-axis becomes vertical,
 * and the Y-axis becomes horizontal.
 * @param inv <code>true</code> if the data is to be inverted;
 * <code>false</code> otherwise
 */
public void setInverted(boolean inv)
{
    int change_value = 0;

    if (inverted == inv) {
        return;
    }
    inverted = inv;
    if (xaxis != null) {
        xaxis.vertical = inv;
        xaxis.i_vertical = inv;
    }
    if (yaxis != null) {
        yaxis.vertical = !inv;
        yaxis.i_vertical = !inv;
    }
    change_value = Changeable.RECALC | Changeable.NOTIFY_AXES_RECALC;
    if (chartType == JCChart.HILO ||
        chartType == JCChart.HILO_OPEN_CLOSE ||
        chartType == JCChart.CANDLE)
    {
        change_value |= Changeable.NOTIFY_LEGEND_RECALC;
    }
    setChanged(true, change_value);
}

/**
 * Sets the <code>Inverted</code> property of the data view.  If the
 * <code>Inverted</code> property is set, the X-axis becomes vertical,
 * and the Y-axis becomes horizontal.
 * @return <code>true</code> if the data is to be inverted; <code>false</code>
 * otherwise
 */
public boolean isInverted()
{
    return(inverted);
}

/**
 * Sets the <code>PickFocus</code> property of the data view, which specifies
 * how distance is determined for pick operations.  When set to
 * <code>PICK_FOCUS_XY</code>, a pick operation will use the actual distance
 * between the point and the drawn data.  When set to values of
 * <code>PICK_FOCUS_X</code> or <code>PICK_FOCUS_Y</code>, the distance only
 * along the X- or Y-axis is used.
 * @param foc the value to set the pick focus to
 */
public void setPickFocus(int foc)
{
    pickFocus = foc;
}

/**
 * Gets the <code>PickFocus</code> property of the data view.  Possible values
 * are <code>PICK_FOCUS_XY</code>, <code>PICK_FOCUS_X</code>, and
 * <code>PICK_FOCUS_Y</code>.
 * @return the <code>PickFocus</code> property
 */
public int getPickFocus()
{
    return(pickFocus);
}

/**
 * Gets the <code>VisibleInLegend</code> property. This property determines
 * whether or not the view name and its series will appear in the chart legend.
 * @return <code>true</code> if the name will appear in the legend
 */
public boolean isVisibleInLegend()
{
    return(visibleInLegend);
}

/**
 * Sets the <code>VisibleInLegend</code> property. This property determines
 * whether or not the view name and its series will appear in the chart legend.
 * @param shn
 */
public void setVisibleInLegend(boolean shn)
{
    if (shn == visibleInLegend) {
        return;
    }
    visibleInLegend = shn;
    visibleInLegendChanged = true;
    //setChanged(true, Changeable.NOTIFY_LEGEND_RECALC);
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Internal Use Only.  Return the boolean that tracks whether the user
 * has changed the VisibleInLagend property or whether it is still coupled
 * to the Visible property.
 */
public boolean hasVisibleInLegendChanged()
{
    return(visibleInLegendChanged);
}

/**
 * Internal Use Only.  Return the boolean that tracks whether the user
 * has changed the VisibleInLagend property or whether it is still coupled
 * to the Visible property.
 */
public void setVisibleInLegendChanged(boolean changed)
{
    visibleInLegendChanged = changed;
}

/**
 * Gets the <code>AutoLabel</code> property, which determines if a chart
 * automatically generates labels for each point in each series.  The default is
 * <code>false</code>. The labels are stored in the <code>AutoLabelList</code>
 * property.  They are created using the <code>Label</code> property of each
 * series.
 * @return the <code>AutoLabel</code> property
 */
public boolean getAutoLabel()
{
    return(autoLabel);
}

/**
 * Sets the <code>AutoLabel</code> property, which determines if a chart
 * automatically generates labels for each point in each series.  The labels
 * are stored in the <code>AutoLabelList</code> property.  They are created
 * using the <code>Label</code> property of each series.
 * @param auto the value to set the <code>AutoLabel</code> property to
 */
public void setAutoLabel(boolean auto)
{
    if (autoLabel == auto) {
        return;
    }
    autoLabel = auto;

    if (chart != null) {
        if (autoLabel) {
            chart.getChartLabelManager().generateAutoLabels(this);
        }
        else {
            chart.getChartLabelManager().removeAutoLabels(this);
        }
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>OutlineColor</code> property.  The
 * <code>OutlineColor</code> property determines the color with which to draw
 * the outline around a filled chart item (for example, bar, pie).
 * @return Color instance indicating the color that the outline will be drawn
 * in. If the property has not been set and a parent chart exists, the
 * default color of the chart's forground color will be returned; otherwise,
 * null is returned.
 */
public Color getOutlineColor()
{
    if (outlineStyle == null) {
        if (chart != null && chart.getChartArea() != null) {
            return(chart.getChartArea().getForeground());
        }
        return(null);
    }
    else {
        return(outlineStyle.getColor());
    }
}

/**
 * Sets the value of the <code>OutlineColor</code> property.  The
 * <code>OutlineColor</code> property determines the color with which to draw
 * the outline around a filled chart item (for example, bar, pie).
 * @param c the color to draw the outline in
 */
public void setOutlineColor(Color c)
{
    if (outlineStyle == null) {
        outlineStyle = new JCLineStyle(1, c, JCLineStyle.SOLID);
    }
    else if (outlineStyle.getColor() == c) {
        return;
    }
    else {
        outlineStyle.setColor(c);
    }
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the width of the outline around a filled chart item (for example, bar,
 * pie).
 * @return the pixel width of chart outlines
 */
public int getOutlineWidth()
{
    if (outlineStyle != null) {
        return(outlineStyle.getWidth());
    }
    return(1);
}

/**
 * Sets the width of the outline around a filled chart item (for example, bar,
 * pie).
 * @param width pixel width of the outline
 */
public void setOutlineWidth(int width)
{
    if (outlineStyle == null) {
        Color color = Color.black;
        if (chart != null && chart.getChartArea() != null) {
            color = chart.getChartArea().getForeground();
        }
        outlineStyle = new JCLineStyle(width, color, JCLineStyle.SOLID);
    }
    else if (outlineStyle.getWidth() == width) {
        return;
    }
    else {
        outlineStyle.setWidth(width);
    }
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the pattern of the outline around a filled chart item (for example,
 * bar, pie).
 * @return one of <code>JCLineStyle.NONE</code>, <code>JCLineStyle.SOLID</code>,
 * <code>JCLineStyle.LONG_DASH</code>, <code>JCLineStyle.SHORT_DASH</code>,
 * <code>JCLineStyle.LSL_DASH</code>, or <code>JCLineStyle.DASH_DOT</code>
 */
public int getOutlinePattern()
{
    if (outlineStyle != null) {
        return(outlineStyle.getPattern());
    }
    return(JCLineStyle.SOLID);
}

/**
 * Sets the pattern of the outline around a filled chart item (for example,
 * bar, pie).
 * @param pattern one of <code>JCLineStyle.NONE</code>,
 * <code>JCLineStyle.SOLID</code>, <code>JCLineStyle.LONG_DASH</code>,
 * <code>JCLineStyle.SHORT_DASH</code>, <code>JCLineStyle.LSL_DASH</code>, or
 * <code>JCLineStyle.DASH_DOT</code>
 */
public void setOutlinePattern(int pattern)
{
    if (outlineStyle == null) {
        Color color = Color.black;
        if (chart != null && chart.getChartArea() != null) {
            color = chart.getChartArea().getForeground();
        }
        outlineStyle = new JCLineStyle(1, color, pattern);
    }
    else if (outlineStyle.getPattern() == pattern) {
        return;
    }
    else {
        outlineStyle.setPattern(pattern);
    }
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the join attribute of the outline around a filled chart item
 * (for example, bar, pie).
 * @return one of <code>BasicStroke.JOIN_MITER</code>,
 * <code>BasicStroke.JOIN_BEVEL</code>, or <code>BasicStroke.JOIN_ROUND</code>
 */
public int getOutlineJoin()
{
    if (outlineStyle != null) {
        return(outlineStyle.getJoin());
    }
    return BasicStroke.JOIN_MITER;
}

/**
 * Sets the join attribute of the outline around a filled chart item
 * (for example, bar, pie).
 * @param join one of <code>BasicStroke.JOIN_MITER</code>,
 * <code>BasicStroke.JOIN_BEVEL</code>, or <code>BasicStroke.JOIN_ROUND</code>
 */
public void setOutlineJoin(int join)
{
    if (outlineStyle == null) {
        Color color = Color.black;
        if (chart != null && chart.getChartArea() != null) {
            color = chart.getChartArea().getForeground();
        }
        outlineStyle = new JCLineStyle(1, color, JCLineStyle.SOLID,
                                        BasicStroke.CAP_BUTT, join);
    }
    else if (outlineStyle.getJoin() == join) {
        return;
    }
    else {
        outlineStyle.setJoin(join);
    }
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the cap attribute of the outline around a filled chart item
 * (for example, bar, pie).
 * @return one of <code>BasicStroke.CAP_BUTT</code>,
 * <code>BasicStroke.CAP_ROUND</code>, or <code>BasicStroke.CAP_SQUARE</code>
 */
public int getOutlineCap()
{
    if (outlineStyle != null) {
        return(outlineStyle.getCap());
    }
    return(BasicStroke.CAP_BUTT);
}

/**
 * Sets the cap attribute of the outline around a filled chart item
 * (for example, bar, pie).
 * @param cap one of <code>BasicStroke.CAP_BUTT</code>,
 * <code>BasicStroke.CAP_ROUND</code>, or <code>BasicStroke.CAP_SQUARE</code>
 */
public void setOutlineCap(int cap)
{
    if (outlineStyle == null) {
        Color color = Color.black;
        if (chart != null && chart.getChartArea() != null) {
            color = chart.getChartArea().getForeground();
        }
        outlineStyle = new JCLineStyle(1, color, JCLineStyle.SOLID,
                                       cap, BasicStroke.JOIN_MITER);
    }
    else if (outlineStyle.getCap() == cap) {
        return;
    }
    else {
        outlineStyle.setCap(cap);
    }
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the <code>OutlineStyle</code> property.  The <code>OutlineStyle</code>
 * property determines the color, width, and pattern (<code>SOLID</code>,
 * <code>LONG_DASH</code>, <code>DASH_DOT</code>, etc.) of an outline around a
 * filled chart item (for example, bar, pie).
 * @return a <code>JCLineStyle</code> object
 */
public JCLineStyle getOutlineStyle()
{
    if (outlineStyle == null) {
        Color color = Color.black;
        if (chart != null && chart.getChartArea() != null) {
            color = chart.getChartArea().getForeground();
        }
        outlineStyle = new JCLineStyle(1, color, JCLineStyle.SOLID);
    }
    return(outlineStyle);
}

/**
 * Sets the <code>OutlineStyle</code> property.  The <code>OutlineStyle</code>
 * property determines the color, width, and pattern (<code>SOLID</code>,
 * <code>LONG_DASH</code>, <code>DASH_DOT</code>, etc.) of an outline around a
 * filled chart item (for example, bar, pie).
 * @param lineStyle a <code>JCLineStyle</code> object
 */
public void setOutlineStyle(JCLineStyle lineStyle)
{
    outlineStyle = lineStyle;
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Sets the value of the <code>ChartStyle</code> property.
 * The <code>ChartStyle</code> property contains all the
 * <code>ChartStyles</code> for the data series in this data view.
 * @param index index of the <code>ChartStyle</code> to set
 * @value <code>ChartStyle</code> object to place at specified index
 * @param value
 */
public void setChartStyle(int index, JCChartStyle value)
{
    if (index < 0 || index > styles.size())
        throw new IllegalArgumentException(
            "Invalid ChartStyle index in ChartDataView.");

    value.addParent(this);

    if (index < styles.size()) {
        styles.set(index, value);
    }
    else {
        styles.add(index, value);
    }

    setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the value of the <code>ChartStyle</code> property. The
 * <code>ChartStyle</code> property contains all of the <code>ChartStyles</code>
 * for the data series in this data view.
 * @param index index of the <code>ChartStyle</code> to retrieve
 * @return <code>ChartStyle</code> at the specified index
 */
public JCChartStyle getChartStyle(int index)
{
    if (index < 0) {
        throw new IllegalArgumentException("ChartStyle index must be positive");
    }
    if (index >= styles.size()) {
        // Must pad the array
        while (index >= styles.size()) {
            styles.add(null);
        }
    }

    // This is a workaround for a Netscape 3.01, JDK1.0.2., Solaris OS issue;
    // It was a ClassCastException raised by NS in the above conditions
    // the solution isn't too nice but is solving the problem. The exception
    // appeared at:
    //     JCChartStyle cs = (JCChartStyle)styles.get(index);
    // The cast from Object to JCChartStyle "bugs" Netscape 3.01 on Solaris
    // instead of a simple variable an array with a single element is used
    JCChartStyle[] cs = new JCChartStyle[1];

    cs[0] = (JCChartStyle)styles.get(index);
    if (cs[0] == null) {
        // Another workaround for a Netscape 3.01, JDK1.0.2., Solaris OS issue;
        // The problem seems to appear in Win'95 and NT, too
        // It was an ArrayStoreException raised by NS for the following
        // statement:
        //     cs[0] = JCChartStyle.makeDefault(chart, this);
        JCChartStyle temp_cs = new JCChartStyle();
        cs[0] = temp_cs.makeDefault(chart, this);
        styles.set(index, cs[0]);
        temp_cs = null;
    }
    return(cs[0]);
}

/**
 * Gets the value of the <code>ChartStyle</code> property.
 * The <code>ChartStyle</code> property contains all the
 * <code>ChartStyles</code> for the data series in this data view.
 * @return List object containing <code>JCChartStyle</code> instances used for
 * this data view
 */
public List getChartStyle()
{
    if (styles == null || styles.size() == 0) {
        return(null);
    }

    return(new Vector(styles));
}

/**
 * Sets the value of the <code>ChartStyle</code> property.
 * The <code>ChartStyle</code> property contains all the
 * <code>ChartStyles</code> for the data series in this data view.
 * @param cs List object containing <code>JCChartStyle</code> instances used
 * for this data view
 */
public void setChartStyle(List cs)
{
    styles = new Vector();
    if (cs == null) {
        return;
    }
    for (Iterator i = cs.iterator(); i.hasNext();) {
        JCChartStyle thisCS = (JCChartStyle) i.next();
        thisCS.addParent(this);
        styles.add(thisCS);
    }
    setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Sets the values of the <code>ColorHandler</code> property.
 * The <code>ColorHandler</code> property references a class that implements the
 * <code>JCDrawableColorHandler</code> interface to override the default color
 * for drawing chart data.  Currently, the color handler is only used
 * by Plot charts.  A null value for this property indicates that
 * the default series color is to be used.
 * @param dch a class implementing the <code>JCDrawableColorHandler</code>
 * interface
 */
public void setColorHandler(JCDrawableColorHandler dch)
{

    if (colorHandler == dch) {
        return;
    }
    colorHandler = dch;
    setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the values of the <code>ColorHandler</code> property.
 * The <code>ColorHandler</code> property references a class that implements the
 * <code>JCDrawableColorHandler</code> interface to override the default color
 * for drawing chart data.  Currently, the color handler is only used
 * by Plot charts.  A null value for this property indicates that
 * the default series color is to be used.
 * @return the color handler, a class implementing the
 * <code>JCDrawableColorHandler</code>
 * interface, or null
 */
public JCDrawableColorHandler getColorHandler()
{
    return(colorHandler);
}

/**
 * Get the current object which holds the output properties for writing data
 * to a file and to html/xml properties.
 * @return The current data output properties object
 */
public OutputDataProperties getOutputDataProperties()
{
    return(outputDataProperties);
}

/**
 * Set the current object which holds the output properties for writing data
 * to a file and to html/xml properties.
 * @param outputDataProperties The new ata output properties object
 */
public void setOutputDataProperties(OutputDataProperties outputDataProperties)
{
    this.outputDataProperties = outputDataProperties;
}

/**
 *  Gets a copy of the <i>X</i> values array currently used by this data view.
 *  @return a new array containing a copy of the <i>X</i> values array
 *  currently used by this data view
 */
public double[] getXArrayCopy()
{
    if (xvalues == null)
        return(null);

    double xarr[] = new double[xvalues.length];
    System.arraycopy(xvalues, 0, xarr, 0, xvalues.length);
    return(xarr);
}

double[] getX()
{
    return(xvalues);
}

void setX(double[] x)
{
    xvalues = x;
}

/**
 * This method returns true if the ChartDataView contains array data.
 */
public boolean isArrayData()
{
    return(xvalues != null);
}

private void readObject(java.io.ObjectInputStream in)
{
    try {
        in.defaultReadObject();
        if (dataSource != null && dataSource instanceof ChartDataManageable) {
            ((ChartDataManageable)dataSource).getChartDataManager().
                                                addChartDataListener(this);
        }
    }
    catch (Exception e) {
        e.printStackTrace(System.out);
    }
}

/**
 * Returns the transient data associated with this data view.
 * @return transient data associated with this data view
 */
Object getTransientData()
{
    return(transientData);
}

/**
 * Sets the transient data for this data view.
 * @param ntd
 */
void setTransientData(Object ntd)
{
    transientData = ntd;
    // Make sure legend is updated
    if (chart != null && chart.getLegend() != null) {
        chart.getLegend().setChanged(true, Changeable.RECALC, false);
    }
}

/**
 * Returns the number of actual series that comprise one item to be drawn.
 * For most chart types, this is a one to one relationship.  For Hi-Lo
 * charts, 2 series are needed for one item.  FOR Hi-Lo-Open-Close and
 * Candle charts, 4 series are needed.
 * @return number of acutal series that comprise one item to be drawn
 */
public int getNumSeriesPerData()
{
    int n = ChartDraw.NUM_SERIES_PER_DATA;
    switch (chartType) {
    case JCChart.HILO:
        n = FastHiloChartDraw.NUM_SERIES_PER_DATA;
        break;
    case JCChart.HILO_OPEN_CLOSE:
        n = FastHLOCChartDraw.NUM_SERIES_PER_DATA;
        break;
    case JCChart.CANDLE:
        n = FastCandleChartDraw.NUM_SERIES_PER_DATA;
        break;
    }
    return(n);
}

/**
 * Sets whether the dataview is showing or not.
 * <br><B>Note:</b> This will also set <code>VisibleInLegend</code> to the same
 * value.  If you wish <code>VisibleInLegend</code> to have a different value,
 * please call <code>setVisibleInLegend()</code> following
 * <code>setVisible()</code>.
 * @param showing
 */
public void setVisible(boolean showing)
{
    if (visible == showing) {
        return;
    }
    visible = showing;

    int cf = Changeable.RECALC;

    // if no one has yet set the visibleInLegend property, assume
    // its default value should follow the Visible property
    if (!visibleInLegendChanged) {
        // set it directly rather than using the set method, since
        // the set method will unset the boolean used to track
        // user changes to the property
        visibleInLegend = showing;
        cf |= Changeable.NOTIFY_LEGEND_RECALC;
    }

    setChanged(true, cf);
}

/**
 * Gets whether the data view is showing or not.
 * @return whether the data view is showing or not
 */
public boolean isVisible()
{
    return(visible);
}

/**
 * Saves the raw Data to the given output stream.
 * The data written should be readable by <code>InputStreamDataSource</code>.
 * @param stream the <code>OutputStream</code> to write the data to
 */
public void outputData(OutputStream stream)
{
    String dataString =
        ChartDataSourceUtil.buildDataSourceString(getDataSource());
    if (dataString != null) {
        try {
            DataOutputStream outputDataStream = new DataOutputStream(stream);
            outputDataStream.writeBytes(dataString);
            outputDataStream.flush();
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }
}

/**
 * Overrides the write object in order to change the state of the changed flags
 * being saved, making the chart recalc and draw when re-serialized.
 * @param out
 * @throws IOException
 */
private void writeObject(java.io.ObjectOutputStream out) throws IOException
{
  // save current state first
  boolean oldChanged = changed;
  changed = true;

  // save state the will generate a recalc
  int oldChangedFlag = changedFlag;
  changedFlag = Changeable.RECALC;
  out.defaultWriteObject();

  // restore current state
  changed = oldChanged;
  changedFlag = oldChangedFlag;
}

/**
 * Gets the current <code>DataHandler</code> object.
 * @return current <code>DataHandles</code> object
 */
protected DataHandler getDataHandler()
{
    if (dataHandler == null) {
        dataHandler = new DataHandler(this);
    }
    return(dataHandler);
}

/**
 * Used for Radar charts.
 *
 * This method searches all of the data series for the highest
 * value of <code>firstPoint</code> and lowest value of <code>lastPoint</code>.
 * If no error occurs, a two-element integer array is returned.  The
 * first array element is the value of <code>firstPoint</code>.  The second
 * array element is the value of <code>lastPoint</code>.  If an error occurs,
 * null is returned.
 * @return If no error occurs, a two-element integer array is returned.  The
 * first array element is the value of <code>firstPoint</code>.  The second
 * array element is the value of <code>lastPoint</code>.  If an error occurs,
 * null is returned.
 * @param dv
 */
protected static int[] getFirstLastPoints(ChartDataView dv)
{
    if (dv != null) {
        List seriesList = dv.getSeries();

        if (seriesList != null) {
            ChartDataViewSeries series = null;
            int firstPoint = Integer.MIN_VALUE;
            int lastPoint  = Integer.MAX_VALUE;
            int index;
            for (int i = 0; i < seriesList.size(); i++) {
                series = (ChartDataViewSeries)seriesList.get(i);
                if (series != null && series.isVisible()) {
                    index = series.getFirstPoint();
                    if (index > firstPoint) {
                        firstPoint = index;
                    }
                    index = series.getLastPoint();
                    if (index < lastPoint) {
                        lastPoint = index;
                    }
                }
            }
            int[] limits = new int[2];
            limits[0] = firstPoint;
            limits[1] = lastPoint;
            return(limits);
        }
    }
    return(null);
}

// Don't javadoc -- internal use only
public boolean isUsingDoublePixels()
{
    // Double-precision floating point pixels is currently
    // only implemented for plots and scatter plots.
    // All other chart types use integer-based pixels, even if
    // they are stored in a double.
    if (chart != null &&
        (chartType == JCChart.PLOT || chartType == JCChart.SCATTER_PLOT))
    {
        return(chart.isUsingDoublePixels());
    }
    else {
        return(false);
    }
}

}
