/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ChartDataViewSeries.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.util.Locale;

import com.klg.jclass.util.JCName;
import com.klg.jclass.util.ImageMapInfo;

/**
 * Class representing a single data series. While
 * <code>ChartDataView</code> manages multiple data series,
 * <code>ChartDataViewSeries</code> contains the data for a
 * single data series. The data is stored
 * as two arrays of <i>X</i> and <i>Y</i> values. <br><B>Note:</b>
 * If the <i>X</i> data array is null or empty,
 * the <i>X</i> data array in <code>ChartDataView</code> is used
 * (this interpretation used to be known as array format).
 */
public class ChartDataViewSeries implements java.io.Serializable, JCName {

protected ChartDataView     parent = null;
private double[]            xvalues;
private double[]            yvalues;
protected int               firstPoint;
protected DerivedInt        lastPoint = new DerivedInt(0,true);
protected String            name;
protected int               drawingOrder = -1;
protected boolean           visible = true;
protected boolean           visibleInLegend = true;
protected boolean           included = true;
protected String            label;
protected ImageMapInfo[]    pointImageMapInfo;
protected ImageMapInfo      seriesImageMapInfo;
protected ImageMapInfo      legendImageMapInfo;
protected Locale			locale = Locale.getDefault();

protected static int dataViewSeriesNum = 0;

void init(double[] x, double[] y, ChartDataView p)
{
	parent = p;
	setLocale(parent.getLocale());
	xvalues = x;
	yvalues = y;
	if (y == null) {
		firstPoint = Integer.MAX_VALUE;
		lastPoint.value = Integer.MAX_VALUE;
		lastPoint.isDefault = true;
	}
	else {
		firstPoint = 0;
		lastPoint.value = maxIndex();
		lastPoint.isDefault = true;
	}
	name = JCChartBundle.string(JCChartBundle.key27, locale)
												+ (++dataViewSeriesNum);
}

/**
 * Basic constructor. Parent and data arrays are set to null.
 */
public ChartDataViewSeries()
{
	parent = null;
	xvalues = null;
	yvalues = null;
}

/**
 * Internal constructor specifying parent data view and arrays holding data.
 * @param x the <i>X</i> array
 * @param y the <i>Y</i> array
 * @param org the originating parent
 */
ChartDataViewSeries(double[] x, double[] y, ChartDataView org)
{
	parent = org;
	setLocale(parent.getLocale());
	xvalues = x;
	yvalues = y;
	if (y == null) {
		firstPoint = Integer.MAX_VALUE;
		lastPoint.value = Integer.MAX_VALUE;
		lastPoint.isDefault = true;
	}
	else {
		firstPoint = 0;
		lastPoint.value = maxIndex();
		lastPoint.isDefault = true;
	}
}

/**
 * Set the locale for this axis.
 */
protected void setLocale(Locale locale)
{
    if (locale == null) {
        locale = Locale.getDefault();
    }
    this.locale = locale;
}

/**
 * Retrieve the local for this axis.
 */
protected Locale getLocale()
{
    return locale;
}

/**
 * Used to change the <i>Y</i> value of some point in the series. Informs the
 * <code>ChartDataView</code>, which in turn informs the external data object.
 * @param index the index with the <I>Y</I> array
 * @param newValue the new edited <I>Y</I> value
 */
public void editPoint(int index, double newValue)
{
	// Check to make sure the point is valid -- someday

	// Only change the data if the ChartDataView accepts the edit
	if (parent.editPoint(this, index, newValue)) {
		yvalues[index] = newValue;
		parent.setChanged(true, Changeable.RECALC);
	}
}


/**
 * Retrieves the index of the first point in the specified
 * series. <code>FirstPoint</code> and <code>LastPoint</code> can be used to
 * constrain the data that is displayed.
 * @return index of the first point of the series that will be
 * displayed
 */
public int getFirstPoint()
{
	// If firstPoint wasn't set due to no data, check to see if data now
	// exists.  If so, set first point to default value.
	if (firstPoint == Integer.MAX_VALUE) {
		if (getX() != null && getY() != null) {
			firstPoint = 0;
		}
	}
    return(firstPoint);
}

/**
 * Sets the index of the first point in the specified
 * series. <code>FirstPoint</code> and <code>LastPoint</code> can be used to
 * constrain the data that is displayed.
 * @param point new first point number
 * @exception IllegalArgumentException If the provided index does not
 * address a point in the data series.
 */
public void setFirstPoint(int point)
{
	if (point < 0 || point > maxIndex()) {
		throw new IllegalArgumentException("FirstPoint value does not address" +
		                                   " a point in the data series");
	}
	if (firstPoint == point) {
		return;
	}
	firstPoint = point;

	if (parent != null) {
		parent.setChanged(true, Changeable.RECALC |
		                        Changeable.NOTIFY_AXES_RECALC);
	}
}

/**
 * Retrieves the index of the last point in the specified
 * series. <code>FirstPoint</code> and <code>LastPoint</code> can be used to
 * constrain the data that is displayed.
 * @return index of the last point in the series that will be displayed
 */
public int getLastPoint()
{
	double xarr[] = getX();
	double yarr[] = getY();

	int ret = lastPoint.value;
	if (lastPoint.isDefault) {
		if (xarr != null && yarr != null) {
			ret = Math.min(yarr.length, xarr.length) - 1;
		}
		else {
			ret = -1;
		}
	}
	return(ret);
}

/**
 * Gets the <code>LastPointIsDefault</code> property of
 * <code>ChartDataViewSeries</code>. This property determines whether chart is
 * responsible for calculating the <code>LastPoint</code> property itself or if
 * the current value of the <code>LastPoint</code> property should be used.
 * @return value of <code>LastPointIsDefault</code> property; default is
 * <code>true</code>
 */
public boolean getLastPointIsDefault()
{
    return(lastPoint.isDefault);
}

/**
 * Sets the index of the last point in the specified
 * series. <code>FirstPoint</code> and <code>LastPoint</code> can be used to
 * constrain the data that is displayed.
 * @param point new last point index
 * @exception IllegalArgumentException If the provided index does not
 * address a point in the data series.
 */
public void setLastPoint(int point)
{
	if (point < 0 || point > maxIndex()) {
		throw new IllegalArgumentException("LastPoint value does not address " +
		                                   "a point in the data series");
	}
	if (lastPoint.value == point) {
		return;
	}
	lastPoint.value = point;
	lastPoint.isDefault = false;

	if (parent != null) {
		parent.setChanged(true, Changeable.RECALC |
		                        Changeable.NOTIFY_AXES_RECALC);
	}
}

/**
 * Sets the <code>LastPointIsDefault</code> property of
 * <code>ChartDataViewSeries</code>. It determines whether chart is responsible
 * for calculating the <code>LastPoint</code> property itself or if the current
 * value of the <code>LastPoint</code> property should be used.
 * @param b  if <code>true</code>, calculates the <code>LastPoint</code>
 * property; if <code>false</code>, uses the current value of the
 * <code>LastPoint</code> property
 */
public void setLastPointIsDefault(boolean b)
{
	lastPoint.isDefault = b;

	if (parent != null)
		parent.setChanged(true, Changeable.RECALC |
		                        Changeable.NOTIFY_AXES_RECALC);
}

/**
 * Gets the value of the <code>Name</code> property. The <code>Name</code>
 * property represents the name of the data series. In JClass Chart,
 * data series are named, and can be retrieved by name.
 * @return series name
 * @see com.klg.jclass.chart.ChartDataView
 */
public String getName()
{
    return(name);
}

/**
 * Sets the value of the <code>Name</code> property. The <code>Name</code>
 * property represents the name of the data series. In JClass Chart,
 * data series are named, and can be retrieved by name.
 * @param nm the new name
 * @see com.klg.jclass.chart.ChartDataView
 */
public void setName(String nm)
{
	if (name == nm) {
		return;
	}
	name = nm;
}

/**
 * Gets the value of the <code>Parent</code> property.
 * The <code>Parent</code> property points to the containing
 * <code>ChartDataView</code> for this <code>ChartDataView</code> series.
 * @return the current <code>ChartDataView</code> parent
 */
public ChartDataView getParent()
{
	return(parent);
}

/**
 * Gets the <code>Style</code> property for this
 * <code>ChartDataViewSeries</code>, which defines the rendering style for the
 * data series.
 * @return <code>JCChartStyle</code> instance that specifies how to render
 * the data series
 */
public JCChartStyle getStyle()
{
	if (parent == null) {
		return(null);
	}
	int i = parent.getSeriesIndex(this);
	if (i < 0) {
		return(null);
	}
    return(parent.getChartStyle(i));
}

/**
 * Sets the <code>Style</code> property for this
 * <code>ChartDataViewSeries</code>, which defines the rendering style for the
 * data series.
 * @param st <code>JCChartStyle</code> instance that specifies how to render
 * the data series
 */
public void setStyle(JCChartStyle st)
{
	if (parent == null) {
		return;
	}
	int i = parent.getSeriesIndex(this);
    parent.setChartStyle(i, st);
}

/**
 * Gets the value of the <code>DrawingOrder</code> property.  The
 * <code>DrawingOrder</code> property determines the order of display of data
 * series.  When the <code>DrawingOrder</code> property is changed,
 * <code>ChartDataView</code> will normalize the order properties of all of the
 * <code>ChartDataViewSeries</code> objects that it manages.
 * @return DrawingOrder for this ChartDataViewSeries item.
 */
public int getDrawingOrder()
{
    return(drawingOrder);
}

/**
 * Sets the value of the <code>DrawingOrder</code> property.  The
 * <code>DrawingOrder</code> property determines the order of display of data
 * series.  When the <code>DrawingOrder</code> property is changed,
 * <code>ChartDataView</code> will normalize the order properties of all the
 * <code>ChartDataViewSeries</code> objects that it manages.
 * @param neword <code>DrawingOrder</code> for this
 * <code>ChartDataViewSeries</code> item
 * @exception IllegalArgumentException If the order value provided
 * does not address a data series.
 */
public void setDrawingOrder(int neword) {
	/*
	 * This check is invalid because it is possible to create a
	 * parented data series that the data view does not yet have
	 * in its list of series.
	 *
	if (neword < 0 || (parent != null && neword > parent.getNumSeries()))
		throw new IllegalArgumentException("DrawingOrder property for " +
		            "ChartDataViewSeries is out of range");
		*/

	if (drawingOrder == neword) {
		return;
	}
	drawingOrder = neword;
	if (parent != null) {
		parent.setChanged(true, Changeable.REDRAW |
		                        Changeable.NOTIFY_LEGEND_REDRAW);
	}
}

/**
 * Gets the value of the <code>Visible</code> property.  The
 * <code>Visible</code> property determines whether the data series is showing
 * in the chart area. <br><B>Note:</b> Data series that are not showing are
 * still used in axis calculations.  See the <code>IsIncluded</code> property
 * for details on how to omit a data series from chart calculations.
 * @return <code>true</code> if the data series is showing; <code>false</code>
 * otherwise
 */
public boolean isVisible()
{
    return(visible);
}

/**
 * Sets the value of the <code>Visible</code> property.  The
 * <code>Visible</code> property determines whether the data series is showing
 * in the chart area. <br><B>Note:</b> Data series that are not showing are
 * still used in axis calculations.  See the <code>IsIncluded</code> property
 * for details on how to omit a data series from chart calculations.
 * @param shn <code>true</code> if the data series is showing;
 * <code>false</code> otherwise
 */
public void setVisible(boolean shn)
{
	if (shn == visible) {
		return;
	}
	visible = shn;

	if (parent != null) {
		parent.setChanged(true, Changeable.RECALC |
		                        Changeable.NOTIFY_LEGEND_RECALC |
								Changeable.NOTIFY_AXES_RECALC);
	}
}

/**
 * Gets the <code>VisibleInLegend</code> property. This property determines
 * whether or not this series will appear in the chart legend.
 * @return <code>true</code> if the series will appear in the legend
 */
public boolean isVisibleInLegend()
{
	return(visibleInLegend);
}

/**
 * Sets the <code>VisibleInLegend</code> property. This property determines
 * whether or not this series will appear in the chart legend.
 * @param vil the new value of <code>VisibleInLegend</code>
 */
public void setVisibleInLegend(boolean vil)
{
	if (vil == visibleInLegend) {
		return;
	}
	visibleInLegend = vil;

	if (parent != null) {
		parent.setChanged(true, Changeable.NOTIFY_LEGEND_RECALC);
	}
}

/**
 * Gets the value of the <code>IsIncluded</code> property. The
 * <code>IsIncluded</code> property determines whether a data series is included
 * in chart calculations (like axis bounds).
 * @return <code>true</code> if the data series is included; <code>false</code>
 * otherwise
 */
public boolean isIncluded()
{
    return(included);
}

/**
 * Sets the value of the <code>IsIncluded</code> property.  The
 * <code>IsIncluded</code> property determines whether a data series is included
 * in chart calculations (like axis bounds).
 * @param inc <code>true</code> if the data series is included;
 * <code>false</code> otherwise
 */
public void setIncluded(boolean inc)
{
	if (inc == included) {
		return;
	}
	included = inc;

	if (parent != null) {
		parent.setChanged(true, Changeable.RECALC |
                                Changeable.NOTIFY_AXES_RECALC);
	}
}

/**
 * Sets the <code>Label</code> property of the data series.  The
 * <code>Label</code> property controls the text that appears next to the data
 * series inside the legend.  It can be unparsed HTML text.
 * @param lab the new label
 */
public void setLabel(String lab)
{
	if (lab == label) {
		return;
	}
	label = lab;

	if (parent != null) {
		parent.setChanged(true, Changeable.NOTIFY_LEGEND_RECALC);
	}
}

/**
 * Gets the <code>Label</code> property of the data series.  The
 * <code>Label</code> property controls the text that appears next to the data
 * series inside the legend.
 * @return Label information
 */
public String getLabel()
{
	if (label == null) {
		label = new String(JCChartBundle.string(JCChartBundle.key127, locale) +
                           (drawingOrder + 1));
	}
	return(label);
}

/**
 * Retrieves an array of image map information, one for each point in the
 * series.
 * @return The current point image map information array.
 */
public ImageMapInfo[] getPointImageMapInfo()
{
    return(pointImageMapInfo);
}

/**
 * Sets an array of image map information, one for each point in the series.
 * @param pointImageMapInfo The new point image map information array.
 */
public void setPointImageMapInfo(ImageMapInfo[] pointImageMapInfo)
{
    this.pointImageMapInfo = pointImageMapInfo;
}

/**
 * Gets the image map information for this series.
 * @return The current image map information for this series.
 */
public ImageMapInfo getSeriesImageMapInfo()
{
    return(seriesImageMapInfo);
}

/**
 * Sets the image map information for this series.
 * @param seriesImageMapInfo The new image map information for this series.
 */
public void setSeriesImageMapInfo(ImageMapInfo seriesImageMapInfo)
{
    this.seriesImageMapInfo = seriesImageMapInfo;
}

/**
 * Gets the legend image map information for this series.
 * @return The current legend image map information for this series.
 */
public ImageMapInfo getLegendImageMapInfo()
{
    return(legendImageMapInfo);
}

/**
 * Sets the legend image map information for this series.
 * @param legendImageMapInfo The new legend image map information for this
 * series.
 */
public void setLegendImageMapInfo(ImageMapInfo legendImageMapInfo)
{
    this.legendImageMapInfo = legendImageMapInfo;
}

/**
 * Returns a String representation of this series.  If it has a name, it is
 * used.
 * @return a String representation of this series
 */
public String toString()
{
	String s = getName();
	return((s == null) ? JCChartBundle.string(JCChartBundle.key143, locale)
                       : s);
}

/**
 * Gets the maximum <i>X</i> index.
 * @return the current maximum <i>X</i> index
 */
public int maxXIndex()
{
	int pmax = 0;
	if (parent != null) {
	    double[] xarray = parent.getX();
		if (xarray != null) {
			pmax = xarray.length;
		}
	}
	return((xvalues == null) ? pmax : xvalues.length);
}

/**
 * Gets the maximum <i>Y</i> index.
 * @return the current maximum <i>Y</i> index
 */
public int maxYIndex()
{
	return((yvalues == null) ? 0 : yvalues.length);
}

/**
 * Gets the maximum of the <i>X</i> and <i>Y</i> indices.
 * @return the current maximum of the <i>X</i> and <i>Y</i> indices
 */
public int maxIndex()
{
	return(Math.max(0, Math.min(maxXIndex(), maxYIndex()) - 1));
}

protected void getXMinMax(MinMax x)
{
	if (x == null) {
        return;
    }
	JCAxis xaxis = parent.getXAxis();
	x.setMinMax((xvalues == null) ? parent.getX() : xvalues,
			    parent.getHoleValue(),
			    xaxis != null ? (xaxis.isLogarithmic() && !xaxis.isPolar() &&
								 !xaxis.isRadar())
							  : false);
}

protected void calcStep(ChartDataView cdv)
{
	double ar[] = (xvalues == null) ? parent.getX() : xvalues;
	if (ar == null || ar.length == 0) {
		return;
	}
	for (int i = 1; i < ar.length; i++) {
		double diff = ar[i]-ar[i-1];
		if (diff < 0.0) {
			diff = -diff;
		}
		if (diff < cdv.step && diff != 0) {
			cdv.step = diff;
		}
	}
	if (cdv.step == Double.MAX_VALUE) {
		cdv.step = 1.0;
	}
}

protected void getYMinMax(MinMax y)
{
	if (y == null) {
		return;
	}
	JCAxis yaxis = parent.getYAxis();
	y.setMinMax(yvalues, parent.getHoleValue(),
				yaxis != null ? yaxis.isLogarithmic() : false);
}

/**
 * Returns the <i>X</i> value at the specified index.
 * @param i the index of the value wanted
 * @return the <i>X</i> value at the specified index
 */
public double getX(int i)
{
	double xval = parent.getHoleValue();
	double[] xarray = (xvalues == null) ? parent.getX() : xvalues;
	if (i >= 0 && i < xarray.length) {
		xval = xarray[i];
	}
	return(xval);
}

/**
 * Returns the <i>Y</i> value at the specified index.
 * @param i the index of the value wanted
 * @return the <i>Y</i> value at the specified index
 */
public double getY(int i)
{
	double yval = parent.getHoleValue();
	if (i >= 0 && i < yvalues.length) {
		yval = yvalues[i];
	}
	return(yval);
}

protected double[] getX()
{
	return(xvalues == null ? parent.getX() : xvalues);
}

protected double[] getY()
{
	return(yvalues);
}

/**
 * Gets a copy of the <i>X</i> values array in this data series.
 * @return a new array containing a copy of the <i>X</i> values array
 */
public double[] getXArrayCopy()
{
	if (xvalues == null) {
		return(null);
	}
	double xarr[] = new double[xvalues.length];
	System.arraycopy(xvalues, 0, xarr, 0, xvalues.length);
	return(xarr);
}

/**
 * Gets a copy of the <i>Y</i> values array in this data series.
 * @return a new array containing a copy of the <i>Y</i> values array
 */
public double[] getYArrayCopy()
{
	if (yvalues == null) {
		return(null);
	}
	double yarr[] = new double[yvalues.length];
	System.arraycopy(yvalues, 0, yarr, 0, yvalues.length);
	return(yarr);
}

protected void setX(double[] x)
{
	xvalues = x;
}

protected void setY(double[] y)
{
	yvalues = y;
}

/**
 * Returns the index that contains the value closest to the specified
 * <i>X</i> value.
 * @param x the value to check against
 * @return the closest index to <i>X</i>
 */
public int getX(double x)
{
	return(getPoint(x, getX()));
}

/**
 * Returns the index that contains the value closest to the specified
 * <i>Y</i> value.
 * @param y the value to check against
 * @return the closest index to <i>Y</i>
 */
public int getY(double y)
{
	return(getPoint(y, yvalues));
}

/**
 * Finds the closest match to a value in an array.
 * @param v the value to check
 * @param arr the array to check against
 * @return the index of the closest match
 */
protected int getPoint(double v, double[] arr)
{
	if (arr == null) {
		return(0);
	}
	double dist = Double.MAX_VALUE;
	int match = 0;
	for (int i = 0; i < arr.length; i++) {
		double newDist = Math.abs(arr[i] - v);
		if (newDist < dist) {
			dist = newDist;
			match = i;
		}
	}
	return(match);
}

}
