/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.*;
import java.util.List;

/**
 * Base class for all drawables in <code>JCChart</code>.  The drawables
 * are objects instantiated to draw a chart of a particular
 * type.
 */
public abstract class ChartDraw extends TrackChange
                                implements Drawable, java.io.Serializable {

// Local copies of data needed for draw(), pick(), and unpick().  They are
// set by recalc().
protected List      seriesList = null;
protected boolean   inverted = false;
protected double    holeValue = Double.MAX_VALUE;
protected boolean   fastUpdate = false;
protected boolean   cachePlotData = true;
protected JCAxis    xaxis = null;
protected JCAxis    yaxis = null;

/** Enum value that holds the chart type.  Used in <code>JCChartArea</code>. */
protected int chartType = -1;

/** Reference to the data view object that is to be plotted. */
protected ChartDataView dataObject = null;

/** Reference to the chart area on which the data will be plotted. */
protected JCChartArea chartArea = null;

/** Number of active data series in the data object. */
protected int numActiveDataSeries;

/** <code>true</code> if drawing on the front plane; <code>false</code> otherwise. */
protected boolean drawFront = true;

/** <code>true</code> if we're using double pixels rather than ints to render data. */
protected boolean usingDoublePixels = false;

/** Name of derived drawable classes. */

public static String[] drawables = {
    "com.klg.jclass.chart.FastPlotChartDraw",
    "com.klg.jclass.chart.FastPlotChartDraw",
    "com.klg.jclass.chart.PolarChartDraw",
    "com.klg.jclass.chart.RadarChartDraw",
    "com.klg.jclass.chart.AreaRadarChartDraw",
    "com.klg.jclass.chart.FastHiloChartDraw",
    "com.klg.jclass.chart.FastHLOCChartDraw",
    "com.klg.jclass.chart.FastCandleChartDraw",
    "com.klg.jclass.chart.AreaChartDraw",
    "com.klg.jclass.chart.BarChartDraw",
    "com.klg.jclass.chart.SBarChartDraw",
    "com.klg.jclass.chart.PieChartDraw",
    "com.klg.jclass.chart.SAreaChartDraw"
};

public final static int BASE_MARGIN = 10;
public final static int NUM_SERIES_PER_DATA = 1;

public ChartDraw() {}

public void setDataView(ChartDataView data)
{
    dataObject = data;
    this.chartArea = data.getParent().getChartArea();
    if (dataObject != null) {
        chartType = dataObject.getChartType();
    }
}

/**
 * One and only constructor for <code>ChartDraw</code>.  Uses the
 * provided data view object to initialize the drawable.
 * @param data data view to be plotted
 */
public ChartDraw(ChartDataView data)
{
    setDataView(data);
}

/**
 * Given a data view object, creates a drawable of the appropriate
 * type.
 * @exception IllegalArgumentException If the data view provided is
 * of a type not currently supported by <code>ChartDraw</code>.
 * @return
 * @param dr
 */
public static ChartDraw makeDrawObject(ChartDataView dr)
{
    ChartDraw rval = null;
    try {
        String drawableClassName = drawables[dr.getChartType()];
        rval = (ChartDraw)Class.forName(drawableClassName).newInstance();
    }
    catch (Throwable e) {
        e.printStackTrace();
        rval = null;
        throw new RuntimeException("Could not find a drawable for chart type " +
                                   dr.getChartType());
    }
    if (rval != null) {
        rval.setDataView(dr);
    }
    return(rval);
}

/**
 * Sets the class to be used as the drawable for this chart type.
 * The class must subclass <code>ChartDraw</code>.
 * @param chartType  the chart type to set the drawable for
 * @param className  the class name to use as the drawable for this chart type
 */
public static void setDrawClass(int chartType, String className)
{
    drawables[chartType] = className;
}

/**
 * Gets the class being used as the drawable for this chart type.
 * @param chartType  the chart type to get the drawable for
 * @return the class name being used as the drawable for this chart type
 */
public static String getDrawClass(int chartType)
{
    return(drawables[chartType]);
}

/**
 * Sets the list of classes to be used as drawables.  Each entry in the
 * array must be placed at the index equal to the chart type value as defined
 * by the list of chart types in <code>JCChart</code>.  Each entry must be a String
 * containing the fully qualified class name (for example,
 * <i>com.klg.jclass.chart.BarChartDraw</i>).  The array must be at least big
 * enough to hold references to all currently existing chart types.  (The
 * last pre-existing chart type position is defined by
 * <code>JCChart.LAST_PREDEFINED_CHART_TYPE</code>.)  Each drawable class must subclass
 * <code>ChartDraw</code>.
 * @param drawablesArray  the new list of drawables
 */
public static void setDrawables(String[] drawablesArray)
{
    if (drawablesArray.length <= JCChart.LAST_PREDEFINED_CHART_TYPE) {
        throw new IllegalArgumentException("Array too small for " +
                                           "pre-existing chart types");
    }
    else {
        drawables = drawablesArray;
    }
}

/**
 * Gets the list of classes that are currently being used as drawables.  Each
 * entry in the array is placed at the index equal to the chart type value
 * as defined by the list of chart types in <code>JCChart</code>.
 * @return the list of drawables
 */
public static String[] getDrawables()
{
    return(drawables);
}

/**
 * Recalculates the drawable.  This default version calculates
 * the number of active data series in the data view object.
 */
public void recalc()
{
    numActiveDataSeries = 0;
    List s = dataObject.getSeries();
    if (s == null || s.size() == 0) {
        return;
    }

    for (int i = 0; i < s.size(); i++) {
        if (((ChartDataViewSeries)s.get(i)).isVisible()) {
            numActiveDataSeries++;
        }
    }

    // check if using double-precision floating point pixels
    // or plain old integer pixels -- this will affect some calculations
    usingDoublePixels = dataObject.isUsingDoublePixels();

    // 3d chart types always draw on front plane.
    // If both axes are on the front plane, draw on front plane unless
    // the data view DrawFrontPlane property is set to false.
    if (chartArea == null          ||
        !chartArea.getDraw3D()     ||
        dataObject.is3DChartType() ||
        (dataObject.getXAxis().frontPlane &&
         dataObject.getYAxis().frontPlane &&
         dataObject.getDrawFrontPlane()))
    {
        drawFront = true;
    }
    else {
        drawFront = false;
    }
}

public void recalc(int startPoint, int endPoint) {}

public void draw(Graphics gc, int startPoint, int endPoint) {}

/**
 * Sets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property controls whether the drawable will
 * be recalculated.
 * @param n new value for the <code>Changed</code> property
 * @param cf
 */
public void setChanged(boolean n, int cf)
{
    super.setChanged(n, cf);
    if (!changed && dataObject != null) {
        dataObject.setChanged(false, Changeable.NO_CHANGE, false);
    }
}

/**
 * Sets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property controls whether the drawable will
 * be recalculated.
 * @param n new value for the <code>Changed</code> property
 * @param updateParent not really used in this implementation of
 * <code>setChanged</code>, but required by the <code>TrackChange</code>
 * interface
 * @param cf change flag
 */
public void setChanged(boolean n, int cf, boolean updateParent)
{
    if (updateParent) {
        setChanged(n, cf);
    }
    else {
        changed = n;
        if (!changed && dataObject != null) {
            dataObject.setChanged(false, Changeable.NO_CHANGE, false);
        }
    }
}

protected void calcTransientData() {}

/**
 * Picks the data point closest to the provided point, given the specifies
 * pick focus.
 * @return
 * @param p
 * @param focus
 */
public JCDataIndex pick(Point p, int focus)
{
    return(null);
}

/**
 * Translates the series and datapoint within the series to pixel coordinates.
 * @return
 * @param pt
 * @param seriesIndex
 */
public Point unpick(int pt, int seriesIndex)
{
    return(null);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di) {}

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>X</i> location of the action, relative to the chart area
 * @param y <I>Y</i> location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y) {}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>X</i> location of the action, relative to the chart area
 * @param y <I>Y</I> location of the action, relative to the chart area
 * @return
 */
int endEdit(JCDataIndex di, int x, int y)
{
    return(-1);
}

public int getNumSeriesPerData()
{
    return(NUM_SERIES_PER_DATA);
}

/**
 * Clips the line between points (x1, y1) and (x2, y2), placing the resulting
 * points in the specified <code>PlotPoint</code> array.  This method is for use
 * only with financial types and assumes straight, either horizontal or
 * vertical, line segments consisting of two points.
 * @param x1 the <i>X</i> value of the first point to be clipped
 * @param y1 the <i>Y</i> value of the first point to be clipped
 * @param x2 the <i>X</i> value of the second point to be clipped
 * @param y2 the <i>Y</i> value of the second point to be clipped
 * @param xlim the <i>X</i> limits of the clip window
 * @param ylim the <i>Y</i> limits of the clip window
 * @param arr the array of <code>PlotPoint</code> instances to place the
 * resulting clipped points into
 * @param inverted <code>true</code> if axes are inverted
 * @param moveLine <code>true</code> if horizontal lines outside Y-axis bounds
 * should be moved to the bounds (this is used to create the effect of clipping
 * a rectangle where the horizontal line is one of the ends of the rectangle)
 */
void financialClip(double x1, double y1, double x2, double y2,
                   MinMax xlim, MinMax ylim, PlotPoint[] arr,
                   boolean inverted, boolean moveLine)
{
    double xMin  = xlim.min;
    double xMax  = xlim.max;
    double yMin  = ylim.min;
    double yMax  = ylim.max;

    boolean draw1 = true;
    boolean draw2 = true;

    if (y1 == dataObject.getHoleValue() || x1 == dataObject.getHoleValue() ||
        y2 == dataObject.getHoleValue() || x2 == dataObject.getHoleValue())
    {
        return;
    }

    boolean vertical = (x1 == x2);
    boolean horizontal = (y1 == y2);
    if (!vertical && !horizontal) {
        // Not a straight horizontal or vertical line! return.
        return;
    }

    if ((y1 < yMin && y2 < yMin) || (y1 > yMax && y2 > yMax)) {
        // Both points outside, return with no points added

        // Special case : stay in and clip the horizontal line
        // if moveLine is true
        if (!(horizontal && moveLine)) {
            return;
        }
    }

    // Still in?  Clip y values to limits.
    if (y1 < yMin) {
        y1 = yMin;
        draw1 = false;
    }
    else if (y1 > yMax) {
        y1 = yMax;
        draw1 = false;
    }
    if (y2 < yMin) {
        y2 = yMin;
        draw2 = false;
    }
    else if (y2 > yMax) {
        y2 = yMax;
        draw2 = false;
    }

    if ((x1 < xMin && x2 < xMin) || (x1 > xMax && x2 > xMax)) {
        // Both points outside, return with no points added
        return;
    }

    // Still in?  Clip x values to limits.
    if (x1 < xMin) {
        x1 = xMin;
        draw1 = false;
    }
    else if (x1 > xMax) {
        x1 = xMax;
        draw1 = false;
    }
    if (x2 < xMin)  {
        x2 = xMin;
        draw2 = false;
    }
    else if (x2 > xMax) {
        x2 = xMax;
        draw2 = false;
    }

    addClippedPoint(arr, 0, x1, y1, xaxis, yaxis, inverted, draw1);
    addClippedPoint(arr, 1, x2, y2, xaxis, yaxis, inverted, draw2);
}

/**
 * Adds a clipped point to the specified <code>PlotPoint</code> vector or array.
 * @param arr the array of <code>PlotPoint</code> instances to place the resulting
 * clipped points into
 * @param pointPos the current position in the <code>PlotPoint</code> array
 * @param x the <i>X</i> value of the point to be added
 * @param y the <i>Y</i> value of the point to be added
 * @param inverted <code>true</code> if axes are inverted
 * @param draw <code>true</code> if point not clipped; <code>false</code> otherwise
 * @param xAxis
 * @param yAxis
 */
void addClippedPoint(PlotPoint[] arr, int pointPos, double x, double y,
                     JCAxis xAxis, JCAxis yAxis, boolean inverted, boolean draw)
{
    if (xAxis.isPolarRadar() || yAxis.isPolarRadar()) {
        return;
    }
    PlotPoint pp = arr[pointPos];
    if (inverted) {
        if (usingDoublePixels) {
            pp.px = yAxis.toPixelDouble(y, drawFront);
            pp.py = xAxis.toPixelDouble(x, drawFront);
        }
        else {
            pp.px = (double) yAxis.toPixel(y, drawFront);
            pp.py = (double) xAxis.toPixel(x, drawFront);
        }
    }
    else {
        if (usingDoublePixels) {
            pp.px = xAxis.toPixelDouble(x, drawFront);
            pp.py = yAxis.toPixelDouble(y, drawFront);
        }
        else {
            pp.px = (double) xAxis.toPixel(x, drawFront);
            pp.py = (double) yAxis.toPixel(y, drawFront);
        }
    }
    pp.drawPoint = draw;
    pp.valid = true;
}

final static int TROUGH_NOT_DRAWN = 0;
final static int TROUGH_DRAWN = 1;
final static int FINISH_TROUGH_LATER = 2;

// Note that tryDrawYAxisTroughGrid(), finishDrawYAxisTroughGrid(),
// drawXAxisTroughGrid(), and finishDrawAxisTroughGrid() are meant to
// be used together.  See BarChartDraw or SBarChartDraw for examples of
// how they are used.

/** Attempts to draw the gridlines for the Y-axis trough.  If the axis
 * is before the current bar, the lines are drawn.  If the axis straddles
 * the current bar, part of it is drawn and the rest will be drawn later
 * (with <code>finishDrawYAxisTroughGrid</code>). If the axis is after the
 * current bar, the lines are not drawn.
 * This method is intended to be used in a loop, calling it before drawing
 * each bar if the lines have not yet been drawn.
 * @return one of <code>TROUGH_NOT_DRAWN</code>, <code>TROUGH_DRAWN</code>, or
 * <code>FINISH_TROUGH_LATER</code>, depending on the location of the <i>Y</i>
 * axis with respect to the current bar
 * @param bar_x1
 * @param bar_x2
 * @param delta_x
 * @param delta_y
 * @param gc
 * @param yaxis
 */
int tryDrawYAxisTroughGrid(Graphics gc, int delta_x, int delta_y,
                           JCAxis yaxis, int bar_x1, int bar_x2)
{
    // Y Axis Trough Grid Lines
    boolean yTroughGridDrawn = false;
    boolean finishTroughAfterBar = false;

    if ((!inverted && delta_x > 0) || (inverted && delta_y > 0)) {
        // bar after y trough -- draw y trough grid
        if (bar_x1 > yaxis.linePosition()) {
            if (yaxis.isVisible() && yaxis.gridVisible) {
                yaxis.drawGridInTrough(gc, Double.MAX_VALUE, Double.MAX_VALUE);
            }
            yaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_TOP |
                                            JCAxis.TROUGH_BOTTOM);
            yTroughGridDrawn = true;

        // Bar straddles y trough -- draw y trough grid
        // in pieces, first piece now, second piece after
        // bar shadows are drawn
        }
        else if (bar_x2 > yaxis.linePosition()) {
            if (yaxis.isVisible() && yaxis.gridVisible) {
                yaxis.drawGridInTrough(gc, Double.MAX_VALUE, Double.MAX_VALUE);
            }
            finishTroughAfterBar = true;
            yTroughGridDrawn = true;
        }
    }
    else {
        // Bar after y trough -- draw y trough grid
        if (bar_x2 < yaxis.linePosition()) {
            if (yaxis.isVisible() && yaxis.gridVisible) {
                yaxis.drawGridInTrough(gc, Double.MAX_VALUE, Double.MAX_VALUE);
            }
            yaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_TOP |
                                            JCAxis.TROUGH_BOTTOM);
            yTroughGridDrawn = true;

        // Bar straddles y trough -- draw y trough grid
        // in pieces, first piece now, second piece after
        // bar shadows are drawn
        }
        else if (bar_x1 < yaxis.linePosition()) {
            if (yaxis.isVisible() && yaxis.gridVisible) {
                yaxis.drawGridInTrough(gc, Double.MAX_VALUE, Double.MAX_VALUE);
            }
            finishTroughAfterBar = true;
            yTroughGridDrawn = true;
        }
    }
    return(finishTroughAfterBar ? FINISH_TROUGH_LATER
                                : (yTroughGridDrawn ? TROUGH_DRAWN
                                                    : TROUGH_NOT_DRAWN));
}

/** Finishes drawing the Y-axis trough gridlines that were started in
 * <code>tryDrawYAxisTroughGrid()</code>.
 * @param bar_bottom
 * @param bar_top
 * @param delta_x
 * @param delta_y
 * @param gc
 * @param y_reversed
 * @param yaxis
 */
void finishDrawYAxisTroughGrid(Graphics gc, int delta_x, int delta_y,
                               JCAxis yaxis, int y_reversed, double bar_top,
                               double bar_bottom)
{
    if ((!inverted && y_reversed == 0 && delta_y >= 0)||
        (!inverted && y_reversed == 1 && delta_y < 0) ||
        (inverted && y_reversed == 0  && delta_x < 0) ||
        (inverted && y_reversed == 1 && delta_x >= 0))
    {
        if (yaxis.isVisible() && yaxis.gridVisible) {
            yaxis.drawGridInTrough(gc, -Double.MAX_VALUE,
                                   Math.min(bar_top, bar_bottom));
        }
        yaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_BOTTOM);
    }
    else {
        if (yaxis.isVisible() && yaxis.gridVisible) {
            yaxis.drawGridInTrough(gc, Math.max(bar_bottom, bar_top),
                                   Double.MAX_VALUE);
        }
        yaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_TOP);
    }
}

/**
 * Draws the gridlines on the X-axis trough that are before the current
 * bar.  This is intended to be used in a loop, called before drawing each
 * bar.
 * @return the new index to the array of gridlines
 * @param axisPos
 * @param bar_bottom
 * @param bar_top
 * @param bar_x1
 * @param bar_x2
 * @param delta_x
 * @param delta_y
 * @param gc
 * @param gridLineStyle
 * @param inc
 * @param xaxis
 * @param xGridIndex
 * @param xGridLineList
 * @param y_reversed
 */
int drawXAxisTroughGrid(Graphics gc, int delta_x, int delta_y, JCAxis xaxis,
                        int y_reversed, double axisPos, int bar_x1, int bar_x2,
                        double bar_top, double bar_bottom,
                        double[] xGridLineList, int xGridIndex, int inc,
                        JCLineStyle gridLineStyle)
{
    boolean drawOverBar = isDrawOverBar(inverted, y_reversed, delta_y, delta_x,
                                        axisPos, bar_bottom, bar_top);

    // First and second point of bar base depends on drawing order
    int near, far;
    if ((!inverted && delta_x > 0) || (inverted && delta_y > 0)) {
        near = bar_x1;
        far = bar_x2;
    }
    else {
        near = bar_x2;
        far = bar_x1;
    }

    // Value trough begins at depends on drawing order (which is encapsulated
    // nicely by the incrememt of the grid line array).
    double troughBegin;
    if (inc == 1) {
        troughBegin = xaxis.getMin();
    }
    else {
        troughBegin = xaxis.getMax();
    }

    // ignore any grid lines at MAX_VALUE
    while (xGridIndex >= 0 &&
           (xGridIndex < xGridLineList.length) &&
           xGridLineList[xGridIndex] == Double.MAX_VALUE)
    {
        xGridIndex += inc;
    }

    boolean drawGrid = gridLineStyle.updateGraphics(gc);
    // draw undrawn x grid lines up to (and maybe including,
    // if it is to be obscured) the current bar
    if ((!inverted && delta_x > 0) || (inverted && delta_y > 0)) {
        while ((xGridIndex >= 0) &&
               (xGridIndex < xGridLineList.length) &&
               (xaxis.toPixel(xGridLineList[xGridIndex],
                              drawFront) <= (drawOverBar ? near : far)))
        {
            // Draw every grid line.  If grid line is at
            // beginning of the trough, redraw trough beginning.
            if (drawGrid) {
                xaxis.drawGridLineInTrough(gc, gridLineStyle,
                                           xGridLineList[xGridIndex]);
            }
            if (xGridLineList[xGridIndex] == troughBegin) {
                gridLineStyle.resetGraphics(gc);
                if (inc == 1) {
                    xaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_LEFT);
                }
                else {
                    xaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_RIGHT);
                }
                gridLineStyle.updateGraphics(gc);
            }
            xGridIndex += inc;
        }
    }
    else {
        while ((xGridIndex >= 0) &&
               (xGridIndex < xGridLineList.length) &&
               (xaxis.toPixel(xGridLineList[xGridIndex],
                              drawFront) > (drawOverBar ? near : far)))
        {
            // Draw every grid line.  If grid line is at
            // beginning of the trough, redraw trough beginning.
            if (drawGrid) {
                xaxis.drawGridLineInTrough(gc, gridLineStyle,
                                           xGridLineList[xGridIndex]);
            }
            if (xGridLineList[xGridIndex] == troughBegin) {
                gridLineStyle.resetGraphics(gc);
                if (inc == 1) {
                    xaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_LEFT);
                }
                else {
                    xaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_RIGHT);
                }
                gridLineStyle.updateGraphics(gc);
            }
            xGridIndex += inc;
        }
    }
    gridLineStyle.resetGraphics(gc);

    return(xGridIndex);
}

/**
 * Determines if we have to draw the gridline over
 * top of the bar (<code>true</code>) or whether the bar should obscure
 * the gridline (<code>false</code>).
 * @return
 * @param axisPos
 * @param bar_bottom
 * @param bar_top
 * @param delta_x
 * @param delta_y
 * @param inverted
 * @param y_reversed
 */
boolean isDrawOverBar(boolean inverted, int y_reversed,
                      int delta_y, int delta_x,
                      double axisPos, double bar_bottom, double bar_top)
{
    if ((!inverted && y_reversed == 0 && delta_y <= 0) ||
        (!inverted && y_reversed != 0 && delta_y > 0) ||
        (inverted && y_reversed == 0 && delta_x > 0) ||
        (inverted && y_reversed != 0 && delta_x <= 0))
    {
        if (axisPos >= bar_bottom && axisPos >= bar_top) {
            return(true);
        }
    }
    else {
        if (axisPos <= bar_bottom && axisPos <= bar_top) {
            return(true);
        }
    }
    return(false);
}

/**
 * Used to determine whether the last bar straddles the end of the
 * axis trough, hence preventing the trough from redrawing.
 * @return
 * @param drawOverBar
 * @param inverted
 * @param lastBar
 * @param maxAxisX
 * @param minAxisX
 */
boolean isDrawFinalGridAndTrough(BarPoint lastBar, boolean drawOverBar,
                                 boolean inverted, int maxAxisX, int minAxisX)
{
    if (drawOverBar) {
        return(true);
    }
    if (lastBar != null) {
        int bar_x1 = 0;
        int bar_x2 = 0;
        if (inverted) {
            bar_x2 = lastBar.front.y + lastBar.front.height;
            bar_x1 = lastBar.front.y;
        }
        else {
            bar_x1 = lastBar.front.x;
            bar_x2 = lastBar.front.x + lastBar.front.width;
        }
        int maxX = Math.max(maxAxisX, minAxisX);
        int minX = Math.min(maxAxisX, minAxisX);
        if (bar_x1 <= maxX && bar_x2 <= maxX &&
            bar_x1 >= minX && bar_x2 >= minX)
        {
            return(true);
        }
        else {
            return(false);
        }
    }
    return(true);
}


/** Draws the rest of the gridlines.  This should only be called after
 *  all bars are drawn.
 * @param gc
 * @param gridLineStyle
 * @param inc
 * @param switch_order
 * @param xaxis
 * @param xGridIndex
 * @param xGridLineList
 * @param yaxis
 * @param yTroughGridDrawn
 */
void finalDrawAxisTroughGrid(Graphics gc, JCAxis xaxis, JCAxis yaxis,
                             int switch_order, boolean yTroughGridDrawn,
                             double xGridLineList[], int xGridIndex, int inc,
                             JCLineStyle gridLineStyle)
{
    // Draw the rest of the y axis grid lines
    if (!yTroughGridDrawn && yaxis.isVisible() && yaxis.gridVisible) {
        yaxis.drawGridInTrough(gc, Double.MAX_VALUE, Double.MAX_VALUE);
    }

    // Draw the rest of the x axis grid lines
    if (xaxis.isVisible() && xaxis.gridVisible) {
        boolean drawGrid = gridLineStyle.updateGraphics(gc);
        // Ignore any x axis grid lines at MAX_VALUE
        while (xGridIndex >= 0 &&
               (xGridIndex < xGridLineList.length) &&
                xGridLineList[xGridIndex] == Double.MAX_VALUE)
        {
            xGridIndex += inc;
        }

        while ((xGridIndex >= 0) && (xGridIndex < xGridLineList.length)) {
            if (drawGrid) {
                xaxis.drawGridLineInTrough(gc, gridLineStyle,
                                           xGridLineList[xGridIndex]);
            }
            xGridIndex += inc;
        }
        gridLineStyle.resetGraphics(gc);
    }

    // Re-outline the axis troughs
    if (!yTroughGridDrawn) {
        yaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_TOP |
                                        JCAxis.TROUGH_BOTTOM);
    }

    if (switch_order == 0) {
        xaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_RIGHT);
    }
    else {
        xaxis.postDrawOutlineTrough(gc, JCAxis.TROUGH_LEFT);
    }
}

/**
 * Returns <code>true</code> if this drawable is drawing on the front plane,
 * <code>false</code> if it is drawing on the back plane.
 * @return <code>true</code> if this drawable is drawing on the front plane;
 * <code>false</code> if it is drawing on the back plane
 */
public boolean isDrawingOnFront()
{
    return(drawFront);
}

}
