/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: CircularChartAxisHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.*;

/**
 * A base class for Polar/Radar chart axis handlers.
 */
abstract class CircularChartAxisHandler extends AxisHandler {

/*
 * Returns a reference to the rAxis of a Polar/Radar chart.
 * Returns null if the chart type isn't POLAR, RADAR, or AREA_RADAR,
 * or if the r axis hasn't been created yet.
 */
protected JCAxis getRadialAxis()
{
    JCAxis rAxis = null;
    if (parent.isPolar() || parent.isRadar()) {
        if (parent.i_vertical) {
            rAxis = parent;
        }
        else if (parent.otherAxis != null) {
            rAxis = parent.otherAxis;
        }
    }
    return(rAxis);
}

/*
 * Called by the toDataPolar() methods of
 * CircularChartLinearAxisHandler and CircularChartLogAxisHandler.
 */
protected double calcThetaAngle(double dx, double dy)
{
    double originBase =
            parent.polarAxisParams.getOriginBase(JCChartUtil.RADIANS);
    double theta;
    if (dx > 0.0) {
        if (dy < 0.0) {
            theta = -Math.atan(dy/dx);
        }
        else {
            theta = JCChartUtil.M_2PI - Math.atan(dy/dx);
        }
    }
    else if (dx < 0.0) {
        theta = JCChartUtil.M_PI - Math.atan(dy/dx);
    }
    else {
        //  The pixel is directly above or below the center of the chart.
        if (dy < 0) {
            theta = JCChartUtil.M_PI_2;
        }
        else if (dy > 0) {
            theta = JCChartUtil.M_2PI - JCChartUtil.M_PI_2;
        }
        else {
            // If both dx and dy are zero, we are unable to determine theta.
            return(JCChartUtil.DEFAULT_FLOAT);
        }
    }
    theta = JCChartUtil.boundAngle(JCChartUtil.RADIANS, theta - originBase);
    theta = JCChartUtil.convertAngle(JCChartUtil.RADIANS,
                                     parent.chartArea.getAngleUnit(), theta);
    return(theta);
}

/**
 * Returns the chart area location of a point at polar coordinates (theta,radius).
 *
 * @param theta angle in radians
 * @param radius r value
 * @param convertDataToPixel false if radius is in pixel units.
 * @return (x,y) screen location in pixel units
 */
JCDataCoord toScreenLocationPolar(double theta, double radius,
                                  boolean convertDataToPixel)
{
    double angleRadians = -(theta + parent.polarAxisParams.originBase);
    double x = radius * Math.cos(angleRadians);
    double y = radius * Math.sin(angleRadians);
    if (convertDataToPixel) {
        JCAxis rAxis = (parent.i_vertical) ? parent : parent.otherAxis;
        double scaleFactor = scale;
        if ( (parent.isPolar() || parent.isRadar()) && rAxis.isReversed() ) {
            scaleFactor *= -1.0;
        }
        x = scaleFactor * x;
        y = scaleFactor * y;
    }

//  System.out.println("toScreenLocationPolar(): vertical=" + parent.vertical +
//                     ", xCenter=" + parent.xCenter + ", yCenter=" +
//                     parent.yCenter);

    x += parent.polarAxisParams.xCenter;
    y += parent.polarAxisParams.yCenter;
    return(new JCDataCoord(x,y));
}

/**
 * Returns the Polar coordinates for the endpoint of a tick mark.
 * @param tickExtent The inner or outer extent of the tick mark.
 * @param scaledValue The pixel position of the tick mark along the axis.
 * @return A DoubleCoord object containing the angle (radians) and radius
 * of the tick mark's endpoint.
 */
private DoubleCoord getTickEndPoint(double tickExtent, double scaledValue) {
    //  Pythagorean theorem:
    double r =  Math.sqrt(tickExtent * tickExtent +
                          scaledValue * scaledValue);
    double angle = 0.0;
    if (r != 0.0) {
        angle = Math.asin(tickExtent/r);  // radians
    }
    return new DoubleCoord(angle, r);
}

/**
 * Draws a tick mark on a Polar chart r-Axis or theta-Axis
 *
 * @param gc The graphics context.
 * @param label The value label associated with this tick mark.
 */
protected void drawTickPolar(Graphics gc, JCValueLabel label)
{
    if (label == null) {
        return;
    }
    JCAnno anno = label.getAnno();
    if (anno == null || !anno.getDrawTicks()) {
        return;
    }

    double value = label.getValue();
    Point startPoint = null;
    Point endPoint = null;
    double innerExtent = getInnerExtent(anno);
    double outerExtent = getOuterExtent(anno);

    Graphics localGC = gc.create();
    localGC.setColor(getTickColor(anno));

    if (parent.i_vertical) {
        if (value > parent.max.value ||
            value < parent.min.value) {
            return;
        }
        double scaledValue = getScaledValue(value);

        double annoAngle = parent.yAxisAngle;  //  radians
        DoubleCoord coord1 = getTickEndPoint(innerExtent, scaledValue);
        DoubleCoord coord2 = getTickEndPoint(outerExtent, scaledValue);
        startPoint = parent.toPixelPolar(
            annoAngle + coord1.x, coord1.y, false);
        //  false means don't convert from data to pixel units
        endPoint = parent.toPixelPolar(
            annoAngle - coord2.x, coord2.y, false);
        localGC.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);
//        System.out.println("startPoint.x=" + startPoint.x +
//                           ", startPoint.y=" + startPoint.y +
//                           ", endPoint.x=" + endPoint.x +
//                           ", endPoint.y=" + endPoint.y);
    }
    else {
        double angleRadians =
                JCChartUtil.convertAngle(parent.chartArea.getAngleUnit(),
                                         JCChartUtil.RADIANS, value);
        //  toPixelPolar() expects an angle value in RADIANS.
        startPoint = parent.toPixelPolar(angleRadians,
                                         parent.polarAxisParams.radius -
                                         innerExtent,
                                         false);
        endPoint = parent.toPixelPolar(angleRadians,
                                       parent.polarAxisParams.radius +
                                       outerExtent,
                                       false);
        localGC.drawLine(startPoint.x, startPoint.y, endPoint.x, endPoint.y);
    }
    localGC.dispose();
}

protected void drawTickMark(Graphics gc, JCValueLabel label, int axisLoc)
{
    if (parent.isPolar() || (parent.isRadar() && parent.i_vertical)) {
        drawTickPolar(gc, label);
    }
    //  No tick marks for Radar chart x axis
}

/**
 * Gets the width and height of an annotation generated for the
 * provided value and precision.  For Polar/Radar charts, the
 * JCAxis.annotationRotation property can have values of JCAxis.ROTATE_90,
 * JCAxis.ROTATE_180, JCAxis.ROTATE_270, or JCAxis.ROTATE_NONE.
 *
 * @param p Dimension object in which to return the size
 * @param value Value to display
 * @param precision Precision to use to display the value
 */
void getAnnoExtent(Dimension p, double value, int precision)
{
    getAnnotationExtent(p, value, precision);
}

/**
 * Override the base class method.  Don't generate tick labels for
 * the theta axis of a Radar chart.
 */
protected void makeTickLabels(JCAnno tick) {
    if (!(parent.isRadar() && !parent.i_vertical)) {
        axisScaleHandler.makeTickLabels(tick);
    }
}

abstract Point toPixelPolar(double theta, double r,
                            boolean convertDataToPixel);

abstract JCDataCoord toDataPolar(int x, int y);

abstract double getScaledValue(double value);

}
