/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.ImageMapInfo;

import java.util.Vector;
import java.util.Arrays;
import java.util.List;

/**
 * Data interpretation class (this class is used internally
 * by JClass Chart and should not be used by applications).
 * Ambitious users could attempt to create and plug
 * in their own data handlers for specific handling,
 * but this is not common use.
 */
public class DataHandler {

private ChartDataView view;

/**
 * Creates a data handler instance using the specified view.  A
 * <code>DataHandler</code> <i>must</i> have a <code>ChartDataView</code>
 * reference.
 *
 * @param view <code>ChartDataView</code> instance to which we are attached.
 */
public DataHandler(ChartDataView view)
{
    this.view = view;
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_POINT</code>.
 * @param seriesIndex The index of the series whose point we are reloading
 * @param pointIndex The index of the point to reload
 * @see ChartDataEvent
 */
public void reloadPoint(int seriesIndex, int pointIndex)
{
    ChartDataModel src = view.getDataSource();
    int num = (src == null) ? 0 : src.getNumSeries();

    if (src != null && seriesIndex >= 0 && seriesIndex < num) {
        ChartDataViewSeries series = view.getSeries(seriesIndex);
        double[] x = series.getX();
        double[] xNew = src.getXSeries(seriesIndex);
        if (pointIndex >= 0 && pointIndex < Math.min(x.length, xNew.length)) {
            x[pointIndex] = xNew[pointIndex];
        }
        double[] y = series.getY();
        double[] yNew = src.getYSeries(seriesIndex);
        if (pointIndex >= 0 && pointIndex < Math.min(y.length, yNew.length)) {
            y[pointIndex] = yNew[pointIndex];
        }
        view.setChanged(true);
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_POINT</code> when ALL_SERIES is
 * chosen
 * @param pointIndex The index of the point to reload
 */
public void reloadPoint(int pointIndex)
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }
    for (int i = 0; i < src.getNumSeries(); i++) {
        ChartDataViewSeries thisSeries = view.getSeries(i);
        double[] y = thisSeries.getY();
        double[] yNew = src.getYSeries(i);
        if (pointIndex >= 0 && pointIndex < Math.min(y.length, yNew.length)) {
            y[pointIndex] = yNew[pointIndex];
        }

        // Are we using a common X series?  If so, don't update x vals.
        if (view.getX() != null) {
            continue;
        }

        double[] x = thisSeries.getX();
        double[] xNew = src.getXSeries(i);
        if (pointIndex >= 0 && pointIndex < Math.min(x.length, xNew.length)) {
            x[pointIndex] = xNew[pointIndex];
        }
    }
    view.setChanged(true);
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_SERIES</code>.
 * @param seriesIndex The index of the series to reload
 * @see ChartDataEvent
 */
public void reloadSeries(int seriesIndex)
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }

    if (seriesIndex >= 0 && seriesIndex < src.getNumSeries()) {
        ChartDataViewSeries thisSeries = view.getSeries(seriesIndex);
        double newY[] = src.getYSeries(seriesIndex);
        thisSeries.setY(newY);

        // Are we using a common X series?  If so, don't update x vals.
        if (view.getX() == null) {
            double newX[] = src.getXSeries(seriesIndex);
            thisSeries.setX(newX);
        }

        view.setChanged(true);
    }
}

/**
 * Reloads all series
 */
public void reloadAllSeries()
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }
    for (int i = 0; i < src.getNumSeries(); i++) {
        ChartDataViewSeries thisSeries = view.getSeries(i);
        double newY[] = src.getYSeries(i);
        thisSeries.setY(newY);

        // Are we using a common X series?  If so, don't update x vals.
        if (view.getX() == null) {
            double newX[] = src.getXSeries(i);
            thisSeries.setX(newX);
        }
    }
    view.setChanged(true);
}

/**
 * Implements <code>ChartDataEvent&#46;ADD_SERIES</code>.
 *
 * @see ChartDataEvent
 */
public void addSeries()
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }
    int num = view.getNumSeries();
    int seriesIndex = num;

    double xvalues[] = src.getXSeries(seriesIndex);
    double yvalues[] = src.getYSeries(seriesIndex);

    boolean batched = view.isBatched();
    view.setBatched(true);

    ChartDataViewSeries newSeries = new ChartDataViewSeries(xvalues, yvalues,
                                                            view);
    view.addSeries(newSeries);

    reloadSeriesLabel(seriesIndex);

    view.setChanged(true);
    view.setBatched(batched);
}

/**
 * Implements <code>ChartDataEvent&#46;INSERT_SERIES</code>.
 * @param seriesIndex
 * @see ChartDataEvent
 */
public void insertSeries(int seriesIndex)
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }
    int num = view.getNumSeries();
    if (seriesIndex >= num) {
        return;
    }

    double xvalues[] = src.getXSeries(seriesIndex);
    double yvalues[] = src.getYSeries(seriesIndex);

    boolean batched = view.isBatched();
    view.setBatched(true);

    ChartDataViewSeries newSeries = view.addSeries(seriesIndex);
    newSeries.setX(xvalues);
    newSeries.setY(yvalues);

    reloadSeriesLabel(seriesIndex);

    view.setChanged(true);
    view.setBatched(batched);
}

/**
 * Implements <code>ChartDataEvent&#46;REMOVE_SERIES</code>.
 * @param seriesIndex
 * @see ChartDataEvent
 */
public void removeSeries(int seriesIndex)
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }
    int numSeries = view.getNumSeries();
    if (seriesIndex >= numSeries) {
        return;
    }
    view.removeSeries(seriesIndex);
}

/**
 * Removes all series
 */
public void removeAllSeries()
{
    ChartDataModel src = view.getDataSource();
    if (src == null) {
        return;
    }
    int numSeries = view.getNumSeries();
    for (int i = numSeries - 1; i >=0; i--) {
        view.removeSeries(i);
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_ALL_POINT_LABELS</code>.
 *
 * @see ChartDataEvent
 */
public void reloadAllPointLabels()
{
    if (view == null) {
        return;
    }
    ChartDataModel tempSrc = view.getDataSource();
    if (tempSrc == null || !(tempSrc instanceof LabelledChartDataModel)) {
        view.setPointLabels(new Vector());
        return;
    }
    LabelledChartDataModel src = (LabelledChartDataModel)tempSrc;
    String[] pl = src.getPointLabels();
    Vector nv;
    if (pl == null) {
        nv = new Vector();
    }
    else {
        nv = new Vector(Arrays.asList(pl));
    }
    view.setPointLabels(nv);
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_POINT_LABEL</code>.
 * @param pointIndex
 * @see ChartDataEvent
 */
public void reloadPointLabel(int pointIndex)
{
    if (view == null) {
        return;
    }
    ChartDataModel tempSrc = view.getDataSource();
    if (tempSrc == null || !(tempSrc instanceof LabelledChartDataModel)) {
        return;
    }

    LabelledChartDataModel src = (LabelledChartDataModel)tempSrc;
    String[] pl = src.getPointLabels();
    if (pl != null) {
        if (pointIndex == ChartDataEvent.ALL_POINTS) {
            for (int i = 0; i < pl.length; i++) {
                view.setPointLabel(i, pl[i]);
            }
        }
        else if (pointIndex >= 0 && pointIndex < pl.length) {
            view.setPointLabel(pointIndex, pl[pointIndex]);
        }
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_ALL_SERIES_LABELS</code>.
 *
 * @see ChartDataEvent
 */
public void reloadAllSeriesLabels()
{
    if (view == null) {
        return;
    }
    ChartDataModel tempSrc = view.getDataSource();
    if (tempSrc == null || !(tempSrc instanceof LabelledChartDataModel)) {
        return;
    }
    LabelledChartDataModel src = (LabelledChartDataModel)tempSrc;
    String[] sl = src.getSeriesLabels();
    if (sl == null) {
        return;
    }
    for (int i = 0; i < view.getNumSeries() && i < sl.length; i++) {
        ChartDataViewSeries sr = view.getSeries(i);
        sr.setLabel(sl[i]);
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_SERIES_LABEL</code>.
 * @param seriesIndex
 * @see ChartDataEvent
 */
public void reloadSeriesLabel(int seriesIndex)
{
    if (view == null) {
        return;
    }
    ChartDataModel tempSrc = view.getDataSource();
    if (tempSrc == null || !(tempSrc instanceof LabelledChartDataModel)) {
        return;
    }
    LabelledChartDataModel src = (LabelledChartDataModel)tempSrc;
    String[] sl = src.getSeriesLabels();
    if (sl != null && seriesIndex < sl.length &&
        seriesIndex < view.getNumSeries())
    {
        view.getSeries(seriesIndex).setLabel(sl[seriesIndex]);
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_DATA_SOURCE_NAME</code>.
 *
 * @see ChartDataEvent
 */
public void reloadDataSourceName()
{
    if (view == null) {
        return;
    }
    if (view.getDataSource() instanceof LabelledChartDataModel) {
        LabelledChartDataModel src =
            (LabelledChartDataModel)view.getDataSource();
        view.setName(src.getDataSourceName());
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_HOLE_VALUE</code>.
 *
 * @see ChartDataEvent
 */
public void reloadHoleValue()
{
    if (view == null) {
        return;
    }
    if (view.getDataSource() instanceof HoleValueChartDataModel) {
        HoleValueChartDataModel src =
            (HoleValueChartDataModel)view.getDataSource();
        view.setHoleValue(src.getHoleValue());
    }
}

/**
 * Implements <code>ChartDataEvent&#46;APPEND_DATA</code>.
 *
 * @see ChartDataEvent
 */
public void appendData()
{
    refreshData(view.getDataSource());
}

/**
 * Implements <code>ChartDataEvent&#46;RESET</code>.
 *
 * @see ChartDataEvent
 */
public void reset()
{
    loadData(view.getDataSource());
    reloadHoleValue();
    reloadAllPointLabels();
    reloadAllSeriesLabels();
    reloadDataSourceName();
    loadImageMapInfo();
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD</code>.
 *
 * @see ChartDataEvent
 */
public void reload()
{
    refreshData(view.getDataSource());
    reloadHoleValue();
    reloadAllSeriesLabels();
}

/**
 * Implements <code>ChartDataEvent&#46;CHANGE_CHART_TYPE</code>.
 * @param chartType
 * @see ChartDataEvent
 */
public void changeChartType(int chartType)
{
    view.setChartType(chartType);
}

/**
 * Loads data from the provided data source, attempting to re-use existing
 * data series.
 * @param src The data source from which to load.
 */
public void loadData(ChartDataModel src)
{
    int num = (src == null) ? 0 : src.getNumSeries();

    List s = null;
    if (src != null) {
        s = new Vector(num);
    }
    List oldSeries = view.getSeries();

    boolean batched = view.isBatched();
    view.setBatched(true);

    if (src != null && num != 0) {
        int longestY = 0;
        // Safe because we have at least one series
        view.setX(src.getXSeries(0));
        for (int i = 0; i < num; i++) {
            double newX[] = src.getXSeries(i);
            double newY[] = src.getYSeries(i);

            // Store length of longest y series
            if (newY != null && longestY < newY.length) {
                longestY = newY.length;
            }

            // Determine whether src provides a unique set of x
            // values for every set of y values.
            if (newX != view.getX()) {
                // Different x series values, blow away common x series
                view.setX(null);
            }
            ChartDataViewSeries newSeries;
            if (i < oldSeries.size()) {
                // Re-use old ChartDataViewSeries if possible
                // Keeps around old settings, which is important!
                newSeries = (ChartDataViewSeries)oldSeries.get(i);
                newSeries.setX(newX);
                newSeries.setY(newY);
            }
            else {
                newSeries = new ChartDataViewSeries(newX, newY, view);
                newSeries.setDrawingOrder(i);
            }
            s.add(newSeries);
        }
    }
    else {
        view.setX(null);
    }

    view.setSeries(s);
    view.setChanged(true);
    view.setBatched(batched);
}

/**
 * Loads point image map info into the chart.
 */
public void loadPointImageMapInfo()
{
    ChartDataModel src = view.getDataSource();
    int numSeries = (src == null) ? 0 : src.getNumSeries();
    if (numSeries == 0){
        return;
    }
    ImageMapChartDataModel imageMapSrc = null;
    if (src instanceof ImageMapChartDataModel) {
        imageMapSrc = (ImageMapChartDataModel)src;
    }
    for (int i = 0; i < numSeries; i++) {
        ChartDataViewSeries series = view.getSeries(i);
        ImageMapInfo[] pointImageMapInfo = null;
        if (imageMapSrc != null) {
            pointImageMapInfo = imageMapSrc.getPointImageMapInfo(i);
        }
        series.setPointImageMapInfo(pointImageMapInfo);
    }
}

/**
 * Loads cluster image map info into the chart.
 */
public void loadClusterImageMapInfo()
{
    ChartDataModel src = view.getDataSource();
    ImageMapChartDataModel imageMapSrc = null;
    if (src instanceof ImageMapChartDataModel) {
        imageMapSrc = (ImageMapChartDataModel)src;
    }
    ImageMapInfo[] clusterImageMapInfo = null;
    if (imageMapSrc != null) {
        clusterImageMapInfo = imageMapSrc.getClusterImageMapInfo();
    }
    view.setClusterImageMapInfo(clusterImageMapInfo);
}

/**
 * Loads series image map info into the chart.
 */
public void loadSeriesImageMapInfo()
{
    ChartDataModel src = view.getDataSource();
    int numSeries = (src == null) ? 0 : src.getNumSeries();
    if (numSeries == 0) {
        return;
    }
    ImageMapChartDataModel imageMapSrc = null;
    if (src instanceof ImageMapChartDataModel) {
        imageMapSrc = (ImageMapChartDataModel)src;
    }
    ImageMapInfo[] seriesInfo = null;
    if (imageMapSrc != null) {
        seriesInfo = imageMapSrc.getSeriesImageMapInfo();
    }
    for (int i = 0; i < numSeries; i++) {
        ChartDataViewSeries series = view.getSeries(i);
        ImageMapInfo info =
            (seriesInfo == null || i >= seriesInfo.length) ? null
                                                           : seriesInfo[i];
        series.setSeriesImageMapInfo(info);
    }
}

/**
 * Loads image image map info into the chart.
 */
public void loadLegendImageMapInfo()
{
    ChartDataModel src = view.getDataSource();
    int numSeries = (src == null) ? 0 : src.getNumSeries();
    if (numSeries == 0) {
        return;
    }
    ImageMapChartDataModel imageMapSrc = null;
    if (src instanceof ImageMapChartDataModel) {
        imageMapSrc = (ImageMapChartDataModel)src;
    }
    ImageMapInfo[] legendInfo = null;
    if (imageMapSrc != null) {
        legendInfo = imageMapSrc.getLegendImageMapInfo();
    }
    for (int i = 0; i < numSeries; i++) {
        ChartDataViewSeries series = view.getSeries(i);
        ImageMapInfo info =
            (legendInfo == null || i >= legendInfo.length) ? null
                                                           : legendInfo[i];
        series.setLegendImageMapInfo(info);
    }
}

/**
 * Implements <code>ChartDataEvent&#46;RELOAD_IMAGE_MAP_INFO</code>. Loads
 * image map info into the chart.
 */
public void loadImageMapInfo()
{
    loadPointImageMapInfo();
    loadClusterImageMapInfo();
    loadSeriesImageMapInfo();
    loadLegendImageMapInfo();
}

/**
 * Grabs the latest data from the data source for each existing series.
 * @param src
 */
public void refreshData(ChartDataModel src)
{
    int num = (view == null) ? 0 : view.getNumSeries();

    if (src != null && num != 0) {
        // Safe because we have at least one series
        view.setX(src.getXSeries(0));
        for (int i = 0; i < num; i++) {
            double newX[] = src.getXSeries(i);
            double newY[] = src.getYSeries(i);

            // Determine whether src provides a unique set of x
            // values for every set of y values.
            if (newX != view.getX()) {
                // Different x series values, blow away common x seres
                view.setX(null);
            }

            ChartDataViewSeries thisSeries = view.getSeries(i);
            thisSeries.setX(newX);
            thisSeries.setY(newY);
        }
    }
    else {
        view.setX(null);
    }
}

/**
 * Modifies data point in a <code>EditableChartDataModel</code> instance.
 * @param newValue
 * @param pointIndex
 * @param seriesIndex
 * @return
 */
public boolean editPoint(int seriesIndex, int pointIndex, double newValue)
{
    ChartDataModel tmpSrc = view.getDataSource();
    if (tmpSrc == null || !(tmpSrc instanceof EditableChartDataModel)) {
        return(false);
    }

    EditableChartDataModel editSrc = (EditableChartDataModel)tmpSrc;
    return(editSrc.setDataItem(seriesIndex, pointIndex, newValue));
}


}
