/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: EditHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.*;

/**
 * Handler object for edit actions within JCChart.  When
 * an edit action is started, JCChartArea will create an
 * instance of EditHandler and call start(), animate() and
 * end() at appropriate points.  The mapping between key
 * combinations and JCChart events is specified by EventTrigger.
 * @see com.klg.jclass.chart.ActionHandler
 * @see com.klg.jclass.chart.EventTrigger
 */
public class EditHandler extends ActionHandler {

/** Points to the data view that is being edited */
protected ChartDraw drawable = null;

/** (Point index, data series) pair pointing to the data point being edited */
protected JCDataIndex di = null;

/**
 * Default constructor
 */
protected EditHandler() {}

/**
 * Called by JCChartArea when the action is started.  For
 * the edit action, this method picks a point to be edited.
 */
protected void start(int x, int y)
{
    drawable = null;
    di = parent.pick(new Point(x, y), null, ChartDataView.PICK_FOCUS_LOCAL);
    if (di == null || di.dataView == null || di.series == null) {
        return;
    }

    ChartDataModel ds = di.dataView.getDataSource();
    if (ds == null || !(ds instanceof EditableChartDataModel)) {
        return;
    }

    drawable = parent.getDrawable(di.series.getParent());
    if (drawable == null) {
        return;
    }

    // Pie charts not allowed to edit.
    if (drawable.dataObject.getChartType() == JCChart.PIE) {
        drawable = null;
        return;
    }

    drawable.startEdit(di);
}

/**
 * Called by JCChartArea when the action is to be
 * animated.  For the edit action, this method calls the showEdit()
 * method on the drawable.  This technique allows bar charts and
 * line charts to show data editing differently.
 */
protected void animate(int x, int y)
{
    if (di == null || drawable == null) {
        return;
    }
    drawable.showEdit(di, x, y);
}


/**
 * Called by JCChartArea when the chart area has been
 * refreshed and the animation must be redrawn.
 */
protected void reanimate(int x, int y)
{
    animate(x, y);
}

/**
 * Called by JCChartArea when the action is over.  For the
 * edit action, this method calls endEdit() on the drawable,
 * then edits the point.  Note that currently, points can
 * only be edited in the y direction.
 */
protected void end(int x, int y)
{
    if (di == null || drawable == null) {
        return;
    }

    int chartType = drawable.dataObject.getChartType();

    boolean polarRadar = ((chartType == JCChart.POLAR) ||
                          (chartType == JCChart.RADAR) ||
                          (chartType == JCChart.AREA_RADAR));

    int point = drawable.endEdit(di, x, y);
    if (point == Integer.MIN_VALUE) {
        return;
    }

    JCChartTypeFormat chartFormat = drawable.dataObject.getChartFormat();
    if ((chartType == JCChart.STACKING_BAR) &&
        ((JCBarChartFormat)chartFormat).is100Percent())
    {
        // Special case for Stacking Bars as percentages.  Point is pixel
        // value of requested new percentage.  Convert to axis value (i.e.
        // percentage), then convert to data value by taking applying this
        // percentage to the total data value.  The resulting data value will
        // not be the exact percentage requested since changing the data
        // value will also change the total data value.
        double axis_value = drawable.dataObject.getYAxis().toData(point);
        double dat =
            (axis_value * ((SBarChartDraw)drawable).ytotal[di.point]) / 100.0;
        di.series.editPoint(di.point, dat);
    }
    else if (polarRadar) {
        JCAxis yaxis = di.series.getParent().getYAxis();
        double newY = Double.MAX_VALUE;

        double scaledPoint = yaxis.applyPixelToDataScaling(point);

        if (yaxis.isLogarithmic()) {
            if (yaxis.isReversed()) {
                newY = JCChartUtil.log10(yaxis.max.value) + scaledPoint;
            }
            //  Yes, add scaledPoint to log10(max.value).  For
            //  reversed y axis, scaledPoint is negative.
            else {
                newY = scaledPoint + JCChartUtil.log10(yaxis.min.value);
            }
            newY = Math.pow(10.0, newY);
        }
        else {
            if (yaxis.isReversed()) {
                newY = yaxis.max.value + scaledPoint;
            }
            //  Yes, add scaledPoint to the max value.  For
            //  reversed y axis, scaledPoint is negative.
            else {
                newY = scaledPoint + yaxis.min.value;
            }
        }
        di.series.editPoint(di.point, newY);
    }
    else {
        di.series.editPoint(di.point,
                            di.series.getParent().getYAxis().toData(point));
    }
    JCChart chart = getChart();
    if (chart != null) {
        chart.sendEvent(null);
    }
}

}
