/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: EventTrigger.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.event.InputEvent;

/**
 * Class responsible for defining a mapping from a user event to
 * a chart action. For example, <code>EventTrigger</code> can be used
 * to map from a <i>Shift+Left-mouse-button</i> set of events to the
 * rotate action. <br><B>Note:</b> By default, all event triggers are
 * mouse down-drag-up events (which occur as a single sequence of mouse
 * events) -- the only elements that can vary are the modifiers.
 */
public class EventTrigger implements java.io.Serializable {

protected final static int  USE_BUTTON1_ALWAYS=0;
protected final static int  USE_BUTTON1_EXCEPT_META=1;
protected final static int  USE_BUTTON1_NEVER=2;

protected int useButton1;
protected int  modifiers;
protected int  action;

public final static int  ROTATE=0;
public final static int  ZOOM=1;
public final static int  TRANSLATE=2;
public final static int  EDIT=3;
public final static int  PICK=4;
public final static int  DEPTH=5;
public final static int  CUSTOMIZE=6;
public final static int  NONE=-1;

EventTrigger() {
}

/**
 * A public constructor that creates a mapping from a series of
 * modifiers to a chart action.
 * @param mod Bitfield of modifiers from <code>Event</code>. Can include
 * <code>Event.SHIFT_MASK</code>, <code>Event.CTRL_MASK</code>, <code>Event.META_MASK</code>, and
 * <code>Event.ALT_MASK</code>. If <code>META_MASK</code> is not included (which is the same as the
 * third mouse button 3 mask), it is assumed that the bitfield is intended to
 * be applied to the button 1 (the left mouse button).
 *
 * @param act action enum; one of <code>ROTATE</code>, <code>ZOOM</code>, <code>TRANSLATE</code>,
 * <code>EDIT</code>, <code>PICK</code>, <code>DEPTH</code>, or <code>CUSTOMIZE</code>
 */
public EventTrigger(int mod, int act) {
	useButton1 = USE_BUTTON1_EXCEPT_META;
	setModifiers(mod);
	setAction(act);
}

/**
 * A public constructor that creates a mapping from a series of
 * modifiers to a chart action.
 * @param mod Bitfield of modifiers from <code>Event</code>. Can include
 * <code>Event.SHIFT_MASK</code>, <code>Event.CTRL_MASK</code>, <code>Event.META_MASK</code>, and
 * <code>Event.ALT_MASK</code>.
 * @param act action enum; one of <code>ROTATE</code>, <code>ZOOM</code>, <code>TRANSLATE</code>,
 * <code>EDIT</code>, <code>PICK</code>, <code>DEPTH</code>, or <code>CUSTOMIZE</code>
 * @param ub1 if <code>true</code>, the bitfield will be adjusted to include
 * mouse button 1 (the left mouse button); if <code>false</code>, the bitfield will
 * stand as set
 */
public EventTrigger(int mod, int act, boolean ub1) {
	if (ub1)
		useButton1 = USE_BUTTON1_ALWAYS;
	else
		useButton1 = USE_BUTTON1_NEVER;
	setModifiers(mod);
	setAction(act);
}

/**
 * Gets the value of the modifiers property for this <code>EventTrigger</code>
 * instance. The modifiers property specifies the modifier keys
 * that are pressed during mouse down-drag-up that trigger
 * the action.
 * @return Bitfields set using <code>Event.ALT_MASK</code>, <code>Event.CTRL_MASK</code>,
 * <code>Event.META_MASK</code>, and <code>Event.SHIFT_MASK</code>
 */
public int getModifiers() {
    return deAdjustModifier(modifiers, useButton1);
}

/**
 * Sets the value of the modifiers property for this <code>EventTrigger</code>
 * instance. The modifiers property specifies the modifier keys
 * that are pressed during mouse down-drag-up that trigger
 * the action. If the constructor has specified to NOT assume
 * that mouse button 1 is to be used, a modifier of <code>Event.ALT_MASK</code>
 * can be used to specify the use of the middle mouse button for the
 * trigger.  If the constructor has specified to NOT assume mouse
 * button 1 or if the default behaviour was specified, a modifier of
 * <code>Event.META_MASK</code> can be used to specify the use of the right mouse
 * button for the trigger.  If the constructor has specified that
 * mouse button 1 will always be used, any specified masks will be
 * adjusted to include the button 1 mask.
 * @param mod bit fields set using <code>Event.ALT_MASK</code>, <code>Event.CTRL_MASK</code>,
 * <code>Event.META_MASK</code>, and <code>Event.SHIFT_MASK</code>
 * @exception IllegalArgumentException If the modifier is not either 0,
 * Event.SHIFT_MASK, Event.ALT_MASK, Event.META_MASK or Event.CTRL_MASK
 */
public void setModifiers(int mod) {
	modifiers = adjustModifier(mod, useButton1);
}

/**
 * Gets the value of the <code>Action</code> property for this <code>EventTrigger</code> instance.
 * The <code>Action</code> property determines what action will be performed when
 * the specified AWT event occurs -- one of <code>ROTATE</code>, <code>ZOOM</code>,
 * <code>TRANSLATE</code>, <code>EDIT</code>, <code>PICK</code>, <code>DEPTH</code>, or <code>CUSTOMIZE</code>.
 * @return enum specifying chart action
 */
public int getAction() {
    return action;
}

/**
 * Default adjust method.  Calls main adjust method with default button
 * 1 rules.
 * @param mod
 * @return
 */
public static int adjustModifier(int mod)
{
	return adjustModifier(mod, USE_BUTTON1_EXCEPT_META);
}

/**
 * Adjusts the modifier as used by chart to be a real modifier akin to
 * that used by Swing. This might involve performing a bitwise "or" action on
 * the modifier (usually the Shift, Alt, or Ctrl mask) with the Swing
 * <code>InputEvent.BUTTON1_MASK</code> so that the correct combination
 * -- the user clicking on the left mouse button accompanied by the
 * appropriate key -- triggers the action.  Whether this is done or not
 * depends on the constructor used and whether the "use mouse button 1"
 * boolean was set.
 * @param mod
 * @param useButton1
 * @return
 */
public static int adjustModifier(int mod, int useButton1)
{
	if (useButton1 == USE_BUTTON1_NEVER) {
		return mod;
	} else if (useButton1 == USE_BUTTON1_ALWAYS) {
		return mod | InputEvent.BUTTON1_MASK;
	} else if (useButton1 == USE_BUTTON1_EXCEPT_META) {
		if ((mod & InputEvent.META_MASK) != 0) {
			return mod;
		} else {
			return mod | InputEvent.BUTTON1_MASK;
		}
	}
	return mod;
}

/**
 * Default de-adjust method.  Calls main de-adjust method with default button
 * 1 rules.
 * @param mod
 * @return
 */
public static int deAdjustModifier(int mod)
{
	return deAdjustModifier(mod, USE_BUTTON1_EXCEPT_META);
}

/**
 * De-adjusts the modifier as used by Swing, adjusting it back to a modifier
 * as used by chart. This generally involves removing the Swing
 * <code>InputEvent.BUTTON1_MASK</code>.  Whether this is done or not
 * depends on the constructor used and whether the "use mouse button 1"
 * boolean was set.
 * @param mod
 * @param useButton1
 * @return
 */
public static int deAdjustModifier(int mod, int useButton1)
{
	if (useButton1 == USE_BUTTON1_NEVER) {
		return mod;
	} else if (useButton1 == USE_BUTTON1_ALWAYS) {
		return mod ^ InputEvent.BUTTON1_MASK;
	} else if (useButton1 == USE_BUTTON1_EXCEPT_META) {
		if ((mod & InputEvent.META_MASK) != 0)
			return mod;
		else
			return mod ^ InputEvent.BUTTON1_MASK;
	}
	return mod;
}

/**
 * Sets the value of the action property for this <code>EventTrigger</code> instance.
 * The action property determines what action will be performed when
 * the specified AWT event occurs. One of <code>ROTATE</code>, <code>EDIT</code>, <code>ZOOM</code>,
 * <code>TRANSLATE</code>, <code>PICK</code>, <code>DEPTH</code>, or <code>CUSTOMIZE</code>.
 * @param act enum specifying chart action
 * @exception IllegalArgumentException If the action is not one of <code>ROTATE</code>, <code>ZOOM</code>,
 * <code>TRANSLATE</code>, <code>EDIT</code>, <code>PICK</code>, <code>DEPTH</code>, or <code>CUSTOMIZE</code>
 */
public void setAction(int act) {
	if (act == action) return;
	switch(act) {
	case ROTATE:
	case ZOOM:
	case TRANSLATE:
	case EDIT:
	case PICK:
	case DEPTH:
	case CUSTOMIZE:
	case NONE:
		action = act;
		break;
	default:
		throw new IllegalArgumentException("Invalid action specified for EventTrigger");
	}
}
}


