/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: FastCandleChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

import java.util.List;
import java.awt.Graphics2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;

import com.klg.jclass.util.graphics.DefineShape;

/**
 * Drawable object for candle charts.  This object
 * is used for rendering a candle chart based on data
 * contained in the <code>dataObject</code>.
 */
public class FastCandleChartDraw extends ChartDraw {

// number of chart data view series per candle data series
public final static int NUM_SERIES_PER_DATA = 4;
public final static int HI_SERIES_INDEX = 0;
public final static int LO_SERIES_INDEX = 1;
public final static int OPEN_SERIES_INDEX = 2;
public final static int CLOSE_SERIES_INDEX = 3;

// Local copy of data used by draw(), pick(), and unpick().  Set in recalc().
protected JCCandleChartFormat cf = null;
protected boolean isComplex = false;

protected MinMax xlim;
protected MinMax ylim;

/**
 * Default constructor.
 */
public FastCandleChartDraw() {
	super();
	xlim = new MinMax();
	ylim = new MinMax();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc() {
	recalc(-1);
}

public void recalc(int seriesIndex) {
	super.recalc();

	// Make a list of points
	// If recalcing for only one series, use the previous series list.
	// Otherwise, refresh the series list.
	if (seriesIndex == -1) {
		seriesList = dataObject.getSeries();
	}
	if (seriesList == null) return;

	cf = (JCCandleChartFormat) dataObject.getChartFormat();
	holeValue = dataObject.getHoleValue();
	xaxis = dataObject.getXAxis();
	yaxis = dataObject.getYAxis();
	inverted = xaxis.i_vertical;

	if (seriesIndex == (-1) || seriesIndex == 0) {
		// This may take awhile for large data set
		if (dataObject.getXLimits() == null) {
			return;
		}
		if (dataObject.getYLimits() == null) {
			return;
		}
		xlim = new MinMax(dataObject.getXLimits().min,
						  dataObject.getXLimits().max);
		ylim = new MinMax(dataObject.getYLimits().min,
						  dataObject.getYLimits().max);
		dataObject.getDataBounds(xlim, ylim);
	}
}

/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc) {
	try {
		// Get list of data series to be drawn using this drawable
		if (seriesList == null) return;

    	Rectangle clipBounds = gc.getClipBounds();
    	if (clipBounds.width == 0 || clipBounds.height == 0)
        	return;

		isComplex = cf.isComplex();

		JCChartStyle styleHilo;
		JCLineStyle lstyleHilo;
		JCChartStyle styleRising;
		JCChartStyle styleFalling;
		JCChartStyle styleOutline;

		boolean risingCandle = false;
		boolean hasOpenClose = true;
		boolean hasHiLo = true;

    	boolean xLog = xaxis.isLogarithmic();
     	boolean yLog = yaxis.isLogarithmic();

		Graphics candleGC = gc.create();

		// for graphics classes implementing DefineShape
    	DefineShape lineDefShape = null;
    	DefineShape candleDefShape = null;
    	if (gc instanceof DefineShape) {
        	lineDefShape = (DefineShape) gc;
    	}
    	if (candleGC instanceof DefineShape) {
        	candleDefShape = (DefineShape) candleGC;
    	}


		// Must draw each candle data series in order
		int numSeries = seriesList.size() / NUM_SERIES_PER_DATA;

		for (int i = 0; i < numSeries; i++) {

			boolean hasData = true;
			int firstPoint = Integer.MAX_VALUE;
			int lastPoint = Integer.MAX_VALUE;
			boolean showSeries = true;
			int k = 0;
			int count = 0;
			for (k = NUM_SERIES_PER_DATA * i, count = 0;
					count < NUM_SERIES_PER_DATA; count++, k++) {
				ChartDataViewSeries series =
						(ChartDataViewSeries) seriesList.get(k);

				if (!series.isVisible()) {
					showSeries = false;
					break;
				}
				// Drop out if data is uninteresting
				if (series.maxXIndex() == 0 || series.maxYIndex() == 0) {
					hasData = false;
					break;
				}

				// Determine first and last point in data series
				firstPoint = Math.min(firstPoint, series.getFirstPoint());
				lastPoint = Math.min(lastPoint, series.getLastPoint());
			}
			if (!showSeries) continue;
			if (!hasData) continue;

			// Set up the styles
			// Set up the basic information: series, style, x vals, y vals);
			int risingCandleWidth = 0;
			int fallingCandleWidth = 0;
			styleHilo = cf.getHiloStyle(i);
			if (styleHilo == null) continue;
			lstyleHilo = styleHilo.getLineStyle();

			if (isComplex) {
				styleRising = cf.getRisingCandleStyle(i);
				if (styleRising == null)
					continue;
				else
					risingCandleWidth = styleRising.getSymbolSize();
				styleFalling = cf.getFallingCandleStyle(i);
				if (styleFalling == null)
					continue;
				else
					fallingCandleWidth = styleFalling.getSymbolSize();
				styleOutline = cf.getCandleOutlineStyle(i);
				if (styleOutline == null) continue;
			}
			else {
				styleRising = styleHilo;
				styleFalling = styleHilo;
				styleOutline = styleHilo;
				risingCandleWidth = styleHilo.getSymbolSize();
				fallingCandleWidth = risingCandleWidth;
			}

			boolean drawHilo = lstyleHilo.updateGraphics(gc);

			PlotPoint newpts[] = new PlotPoint[2];
			newpts[0] = new PlotPoint();
			newpts[1] = new PlotPoint();

			for (int j = firstPoint; j <= lastPoint; j++) {
				double xhi, yhi, xlo, ylo, xopen, yopen, xclose, yclose;

				// get the candle's HLOC values
				int s = NUM_SERIES_PER_DATA * i;
				ChartDataViewSeries highSeries =
					(ChartDataViewSeries) seriesList.get(s++);
				xhi = highSeries.getX(j);
				yhi = highSeries.getY(j);
				ChartDataViewSeries lowSeries =
					(ChartDataViewSeries) seriesList.get(s++);
				xlo = lowSeries.getX(j);
				ylo = lowSeries.getY(j);
				if (yhi < ylo) {
					double xtmp = xhi;
					double ytmp = yhi;
					xhi = xlo;
					yhi = ylo;
					xlo = xtmp;
					ylo = ytmp;
				}
				ChartDataViewSeries openSeries =
					(ChartDataViewSeries) seriesList.get(s++);
				xopen = openSeries.getX(j);
				yopen = openSeries.getY(j);
				ChartDataViewSeries closeSeries =
					(ChartDataViewSeries) seriesList.get(s++);
				xclose = closeSeries.getX(j);
				yclose = closeSeries.getY(j);
				risingCandle = yopen < yclose;
				hasHiLo = !(yhi == holeValue || ylo == holeValue ||
								xhi == holeValue || xlo == holeValue ||
								(xLog && xhi <= 0.0) ||
								(yLog && yhi <= 0.0) ||
								(xLog && xlo <= 0.0) ||
								(yLog && ylo <= 0.0) );
				hasOpenClose = !(yclose == holeValue || yopen == holeValue ||
								xclose == holeValue || xopen == holeValue ||
								(xLog && xclose <= 0.0) ||
								(yLog && yclose <= 0.0) ||
								(xLog && xopen <= 0.0) ||
								(yLog && yopen <= 0.0) );
				int half_candle_width =
					(risingCandle ? risingCandleWidth : fallingCandleWidth) / 2;

				// Do the Hi points:

				double x1 = xhi;
				double y1 = yhi;

				double x2, y2;

    			int hlocX = Integer.MAX_VALUE;
    			int highY = Integer.MAX_VALUE;
    			int lowY = Integer.MAX_VALUE;

				if (hasHiLo && drawHilo) {
					if (hasOpenClose) {
						x2 = risingCandle ? xclose : xopen;
						y2 = risingCandle ? yclose : yopen;
					} else {
						x2 = xlo;
						y2 = ylo;
					}

					newpts[0].valid = false;
					newpts[1].valid = false;
					financialClip(x1, y1, x2, y2, xlim, ylim,
							   		newpts,
							   		inverted, false);

					if (newpts[0].valid && newpts[1].valid) {



						((Graphics2D)gc).draw(new Line2D.Double(
							Math.floor(newpts[0].px), Math.floor(newpts[0].py),
							Math.floor(newpts[1].px), Math.floor(newpts[1].py)));


                		// if using DefineShape, define high/low shapes
						if (lineDefShape != null) {
							defineHighLowShapes(lineDefShape, gc, inverted,
								true, !hasOpenClose, highSeries, lowSeries, j,
								newpts, half_candle_width);
						}
					}

					if (!hasOpenClose)
						continue;

					// Do the Lo points:

					x1 = xlo;
					y1 = ylo;

					x2 = risingCandle ? xopen : xclose;
					y2 = risingCandle ? yopen : yclose;

					newpts[0].valid = false;
					newpts[1].valid = false;
					financialClip(x1, y1, x2, y2, xlim, ylim,
							   		newpts,
							   		inverted, false);

					if (newpts[0].valid && newpts[1].valid) {



						((Graphics2D)gc).draw(new Line2D.Double(
							Math.floor(newpts[0].px), Math.floor(newpts[0].py),
							Math.floor(newpts[1].px), Math.floor(newpts[1].py)));

                		// if using DefineShape, define low shape
                		if (lineDefShape != null) {
							defineHighLowShapes(lineDefShape, gc, inverted,
								false, true, null, lowSeries, j,
								newpts, half_candle_width);
						}
					}
				}

				// Do the Open points:
				if (hasOpenClose) {
					boolean bypass = false;
					int left = Integer.MAX_VALUE;
					int top = Integer.MAX_VALUE;
					int right = Integer.MIN_VALUE;
					int bottom = Integer.MIN_VALUE;

					double x3 = xopen;
					double y3 = yopen;
					int point_x = xaxis.toPixel(x3, drawFront);

					x1 = xaxis.toData(point_x - half_candle_width, drawFront);
					y1 = y3;

					x2 = xaxis.toData(point_x + half_candle_width, drawFront);
					y2 = y3;

					// moveLine true indicates that we should move the horizontal line
					// that defines the open or close line down to the yaxis limits
					// so that the resulting rectangle will be clipped at the axis.
					// if false, the line will be clipped and the rectangle will
					// not appear.
					boolean moveLine =	!((yopen > ylim.max && yclose > ylim.max) ||
										(yopen < ylim.min && yclose < ylim.min));

					newpts[0].valid = false;
					newpts[1].valid = false;
					financialClip(x1, y1, x2, y2, xlim, ylim,
							   		newpts,
							   		inverted, moveLine);
                	// if using DefineShape, mark open series as null
					// if not drawn
                	if (candleDefShape != null) {
						if (!newpts[0].drawPoint || !newpts[1].drawPoint) {
							openSeries = null;
						}
					}
					for (k = 0; k < newpts.length && !bypass; k++) {
						if (newpts[k].valid) {
							left = Math.min(left, (int) newpts[k].px);
							right = Math.max(right, (int) newpts[k].px);
							top = Math.min(top, (int) newpts[k].py);
							bottom = Math.max(bottom, (int) newpts[k].py);
						} else {
							bypass = true;
						}
					}
					if (bypass) {
						continue;
					}

					// Do the Close points:

					x3 = xclose;
					y3 = yclose;
					point_x = xaxis.toPixel(x3, drawFront);

					x1 = xaxis.toData(point_x - half_candle_width, drawFront);
					y1 = y3;

					x2 = xaxis.toData(point_x + half_candle_width, drawFront);
					y2 = y3;

					newpts[0].valid = false;
					newpts[1].valid = false;
					financialClip(x1, y1, x2, y2, xlim, ylim,
							   		newpts,
							   		inverted, moveLine);
                	// if using DefineShape, mark close series as null
					// if not drawn
                	if (candleDefShape != null) {
						if (!newpts[0].drawPoint || !newpts[1].drawPoint) {
							closeSeries = null;
						}
					}
					for (k = 0; k < newpts.length && !bypass; k++) {
						if (newpts[k].valid) {
							left = Math.min(left, (int) newpts[k].px);
							right = Math.max(right, (int) newpts[k].px);
							top = Math.min(top, (int) newpts[k].py);
							bottom = Math.max(bottom, (int) newpts[k].py);
						} else {
							bypass = true;
						}
					}
					if (bypass) {
						continue;
					}

					if (risingCandle) {
						drawCandleRect(candleGC, styleRising.getFillStyle(),
									   styleOutline.getLineStyle(),
									   left, top, right, bottom,
									   isComplex, risingCandle);
					}
					else {
						drawCandleRect(candleGC, styleFalling.getFillStyle(),
									   styleOutline.getLineStyle(),
									   left, top, right, bottom,
									   isComplex, risingCandle);
					}
                	// if using DefineShape, note candle shape
                	if (candleDefShape != null) {
						defineOpenCloseShape(candleDefShape, candleGC,
							inverted, risingCandle, openSeries, closeSeries, j,
							left, top, right, bottom);

					}
				}
			}
			lstyleHilo.resetGraphics(gc);
		}
		candleGC.dispose();
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
}

/**
 * Defines shapes for high/low values.
 * @param defShape
 * @param doHigh
 * @param doLow
 * @param gc
 * @param half_candle_width
 * @param highSeries
 * @param inverted
 * @param j
 * @param lowSeries
 * @param newpts
 */
protected void defineHighLowShapes(DefineShape defShape, Graphics gc,
		boolean inverted, boolean doHigh, boolean doLow,
		ChartDataViewSeries highSeries, ChartDataViewSeries lowSeries, int j,
		PlotPoint[] newpts, int half_candle_width)
{
	if (doHigh && doLow) {
	 	if (newpts[0].drawPoint && newpts[1].drawPoint) {
			if (!inverted) {
				int midY = (int) newpts[0].py +
					(((int) newpts[1].py - (int) newpts[0].py) / 2);
				defineCandleShape(defShape, gc,
						highSeries, j,
						(int) newpts[0].px - half_candle_width,
						(int) newpts[0].py,
						(half_candle_width * 2) + 1,
						midY - (int) newpts[0].py);
				defineCandleShape(defShape, gc,
						lowSeries, j,
						(int) newpts[0].px - half_candle_width,
						midY,
						(half_candle_width * 2) + 1,
						(int) newpts[1].py - midY);
			} else {
				int midX = (int) newpts[0].px +
					(((int) newpts[1].px - (int) newpts[0].px) / 2);
				defineCandleShape(defShape, gc,
						highSeries, j,
						(int) newpts[0].px,
						(int) newpts[0].py - half_candle_width,
						midX - (int) newpts[0].px,
						(half_candle_width * 2) + 1);
				defineCandleShape(defShape, gc,
						lowSeries, j,
						midX,
						(int) newpts[0].py - half_candle_width,
						(int) newpts[1].px - midX,
						(half_candle_width * 2) + 1);
			}
		} else if (newpts[0].drawPoint) {
			if (!inverted) {
				defineCandleShape(defShape, gc,
					lowSeries, j,
					(int) newpts[0].px - half_candle_width,
					(int) newpts[0].py,
					(half_candle_width * 2) + 1,
					(int) newpts[1].py - (int) newpts[0].py);
			} else {
				defineCandleShape(defShape, gc,
					lowSeries, j,
					(int) newpts[0].px,
					(int) newpts[0].py - half_candle_width,
					(int) newpts[1].px - (int) newpts[0].px,
					(half_candle_width * 2) + 1);
			}
		} else if (newpts[1].drawPoint) {
			if (!inverted) {
				defineCandleShape(defShape, gc,
					highSeries, j,
					(int) newpts[0].px - half_candle_width,
					(int) newpts[0].py,
					(half_candle_width * 2) + 1,
					(int) newpts[1].py - (int) newpts[0].py);
			} else {
				defineCandleShape(defShape, gc,
					highSeries, j,
					(int) newpts[0].px,
					(int) newpts[0].py - half_candle_width,
					(int) newpts[1].px - (int) newpts[0].px,
					(half_candle_width * 2) + 1);
			}
		}
	} else {
	 	if (newpts[0].drawPoint) {
			if (!inverted) {
				defineCandleShape(defShape, gc,
					doHigh ? highSeries : lowSeries, j,
					(int) newpts[0].px - half_candle_width,
					(int) newpts[0].py,
					(half_candle_width * 2) + 1,
					(int) newpts[1].py - (int) newpts[0].py);
			} else {
				defineCandleShape(defShape, gc,
					doHigh ? highSeries : lowSeries, j,
					(int) newpts[0].px,
					(int) newpts[0].py - half_candle_width,
					(int) newpts[1].px - (int) newpts[0].px,
					(half_candle_width * 2) + 1);
			}
		}
	}
}

/**
 * Defines shapes for open/close values.
 * @param bottom
 * @param candleDefShape
 * @param candleGC
 * @param closeSeries
 * @param inverted
 * @param j
 * @param left
 * @param openSeries
 * @param right
 * @param risingCandle
 * @param top
 */
protected void defineOpenCloseShape(DefineShape candleDefShape,
	Graphics candleGC, boolean inverted, boolean risingCandle,
    ChartDataViewSeries openSeries, ChartDataViewSeries closeSeries,
	int j, int left, int top, int right, int bottom)
{
	if (closeSeries != null && openSeries != null) {
		if (!inverted) {
			int midY = top + ((bottom - top) / 2);
			defineCandleShape(candleDefShape, candleGC,
				risingCandle ? closeSeries : openSeries, j,
				left, top,
				right - left, midY - top);
			defineCandleShape(candleDefShape, candleGC,
				risingCandle ? openSeries : closeSeries, j,
				left, midY,
				right - left, bottom - midY);
		} else {
			int midX = left + ((right - left) / 2);
			defineCandleShape(candleDefShape, candleGC,
				risingCandle ? openSeries : closeSeries, j,
				left, top,
				midX - left, bottom - top);
			defineCandleShape(candleDefShape, candleGC,
				risingCandle ? closeSeries : openSeries, j,
				midX, top,
				right - midX, bottom - top);
		}
	} else if (closeSeries != null) {
		if (!inverted) {
			defineCandleShape(candleDefShape, candleGC,
				closeSeries, j,
				left, top,
				right - left, bottom - top);
		} else {
			defineCandleShape(candleDefShape, candleGC,
				closeSeries, j,
				left, top,
				right - left, bottom - top);
		}
	} else if (openSeries != null) {
		if (!inverted) {
			defineCandleShape(candleDefShape, candleGC,
				openSeries, j,
				left, top,
				right - left, bottom - top);
		} else {
			defineCandleShape(candleDefShape, candleGC,
				openSeries, j,
				left, top,
				right - left, bottom - top);
		}
	}
}

/**
 * Defines shapes for candle areas.
 * @param defSeries
 * @param defShape
 * @param gc
 * @param height
 * @param j
 * @param width
 * @param x
 * @param y
 */
protected void defineCandleShape(DefineShape defShape, Graphics gc,
	ChartDataViewSeries defSeries, int j,
	int x, int y, int width, int height)
{
	if (defSeries != null) {
		defShape.beginShape(new JCDataIndex(j, defSeries));
   		gc.fillRect(x, y, width, height);
 		defShape.endShape(false);
	}
}

final private void drawCandleRect(Graphics gc, JCFillStyle fs, JCLineStyle ls,
							int left, int top, int right, int bottom,
							boolean isComplex, boolean risingCandle) {

	int width = Math.max(right - left, 1);
	int height = Math.max(bottom - top, 1);

	if (isComplex) {
		fs.fillRect(gc, left, top, width == 1 ? width : width + 1,
									height == 1 ? height : height + 1);
		if (ls.updateGraphics(gc)) {



			((Graphics2D)gc).draw(new Rectangle2D.Double(
				(double)left, (double)top, (double)width, (double)height));

		}
	}
	else {
		if (ls.updateGraphics(gc)) {
			if (!risingCandle) {
				gc.fillRect(left, top, width == 1 ? width : width + 1,
									height == 1 ? height : height + 1);
			}



			((Graphics2D)gc).draw(new Rectangle2D.Double(
				(double)left, (double)top, (double)width, (double)height));

		}
	}
}

/**
 * Pick method for candle charts.  Given a point relative to the
 * chart area, this method finds the closest point.
 * @param pt (x,y) value relative to the chart area
 * @param focus pick focus used to determine distance between pick point
 * and data point
 * @return <code>JCDataIndex</code> encapsulating the point index and data
 * series of the selected point
 */
public JCDataIndex pick(Point pt, int focus) {
	Rectangle r = chartArea.getDrawingArea();
    int px = pt.x - r.x;
    int py = pt.y - r.y;

	int series = -1;
	int point = -1;
	JCDataIndex di = null;
	double length = Double.MAX_VALUE;
	int pix_x, pix_y;
	if (seriesList == null) return di;

   	boolean xLog = xaxis.isLogarithmic();
   	boolean yLog = yaxis.isLogarithmic();

    int numSeries = seriesList.size() / NUM_SERIES_PER_DATA;
	try {
    for (int i = 0; i < numSeries; i++) {

		boolean hasData = true;
		int firstPoint = Integer.MAX_VALUE;
		int lastPoint = Integer.MAX_VALUE;
		boolean showSeries = true;
		for (int s = NUM_SERIES_PER_DATA * i, count = 0;
				count < NUM_SERIES_PER_DATA; count++, s++) {
			ChartDataViewSeries thisSeries =
				(ChartDataViewSeries) seriesList.get(s);

			if (!thisSeries.isVisible()) {
				showSeries = false;
				break;
			}
			// Drop out if data is uninteresting
			if (thisSeries.maxXIndex() == 0 || thisSeries.maxYIndex() == 0) {
				hasData = false;
				break;
			}

			// Determine first and last point in data series
			firstPoint = Math.min(firstPoint, thisSeries.getFirstPoint());
			lastPoint = Math.min(lastPoint, thisSeries.getLastPoint());
		}
		if (!showSeries) continue;
		if (!hasData) continue;

		double realRisingCandleWidth = 0.0;
		double realFallingCandleWidth = 0.0;
		if (isComplex) {
			JCChartStyle style = cf.getRisingCandleStyle(i);
			if (style != null) {
				realRisingCandleWidth =
					Math.abs(xaxis.toData(0) -
						xaxis.toData(style.getSymbolSize()));
			}
			style = cf.getFallingCandleStyle(i);
			if (style != null) {
				realFallingCandleWidth =
					Math.abs(xaxis.toData(0) -
						xaxis.toData(style.getSymbolSize()));
			}
		}
		else {
			JCChartStyle style = cf.getHiloStyle(i);
			if (style != null) {
				realRisingCandleWidth =
					Math.abs(xaxis.toData(0) -
						xaxis.toData(style.getSymbolSize()));
				realFallingCandleWidth = realRisingCandleWidth;
			}
		}
        double highx[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+HI_SERIES_INDEX)).getX();
        double highy[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+HI_SERIES_INDEX)).getY();
        double lowx[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+LO_SERIES_INDEX)).getX();
        double lowy[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+LO_SERIES_INDEX)).getY();
        double openx[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+OPEN_SERIES_INDEX)).getX();
        double openy[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+OPEN_SERIES_INDEX)).getY();
        double closex[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+CLOSE_SERIES_INDEX)).getX();
        double closey[] = ((ChartDataViewSeries) seriesList.get(
				(NUM_SERIES_PER_DATA * i)+CLOSE_SERIES_INDEX)).getY();

        double thispx = 0.0;
        double thispy = 0.0;
        for (int j = firstPoint; j <= lastPoint; j++) {
			double xhi, yhi, xlo, ylo, xopen, yopen, xclose, yclose;

			// get the candle's HLOC values
			xhi = highx[j];
			yhi = highy[j];
			xlo = lowx[j];
			ylo = lowy[j];
			if (yhi < ylo) {
				double xtmp = xhi;
				double ytmp = yhi;
				xhi = xlo;
				yhi = ylo;
				xlo = xtmp;
				ylo = ytmp;
			}
			xopen = openx[j];
			yopen = openy[j];
			xclose = closex[j];
			yclose = closey[j];

			boolean hasOpenClose =
				!(yclose == holeValue || yopen == holeValue ||
				xclose == holeValue || xopen == holeValue ||
				(xLog && xclose <= 0.0) || (yLog && yclose <= 0.0) ||
				(xLog && xopen <= 0.0) || (yLog && yopen <= 0.0) );
           	double candleWidth = (yopen < yclose) ?
								realRisingCandleWidth : realFallingCandleWidth;

            boolean other = false;
			int startk, endk;
			// if has HiLo
			if (!(yhi == holeValue || ylo == holeValue ||
							xhi == holeValue || xlo == holeValue ||
							(xLog && xhi <= 0.0) ||
							(yLog && yhi <= 0.0) ||
							(xLog && xlo <= 0.0) ||
							(yLog && ylo <= 0.0) )) {
				if (hasOpenClose) {
					startk = 0;
					endk = NUM_SERIES_PER_DATA;
				} else {
					startk = 0;
					endk = OPEN_SERIES_INDEX;
				}
			} else {
				if (hasOpenClose) {
					startk = OPEN_SERIES_INDEX;
					endk = NUM_SERIES_PER_DATA;
				} else {
					startk = 0;
					endk = 0;
				}
			}

            for (int k = startk; k < endk; ) {
                switch (k) {
                    case HI_SERIES_INDEX:
                        thispx = xhi;
                        thispy = yhi;
                        break;
                    case LO_SERIES_INDEX:
                        thispx = xlo;
                        thispy = ylo;
                        break;
                    case OPEN_SERIES_INDEX:
						if (candleWidth == 0) {
							k++;
							continue;
						}
                        if (!other) {
                            thispx = xopen - candleWidth;
                            thispy = yopen;
                            other = true;
                        } else {
                            thispx = xopen + candleWidth;
                            other = false;
                        }
                        break;
                    case CLOSE_SERIES_INDEX:
						if (candleWidth == 0) {
                            k++;
                            continue;
                        }
                        if (!other) {
                            thispx = xclose - candleWidth;
                            thispy = yclose;
                            other = true;
                        } else {
                            thispx = xclose + candleWidth;
                            other = false;
                        }
                        break;
                }

                if (thispx > xlim.max ||
                    xlim.min > thispx ||
                    thispy > ylim.max ||
                    ylim.min > thispy) {
                     if (!other) k++;
                     continue;
                }

                double len;
                switch (focus) {
                case (ChartDataView.PICK_FOCUS_X) :
                    pix_x = xaxis.toPixel(thispx, drawFront);
                    len = inverted ?
                        Math.abs(pix_x - py) : Math.abs(pix_x - px);
                    break;
                case (ChartDataView.PICK_FOCUS_Y) :
                    pix_y = yaxis.toPixel(thispy, drawFront);
                    len = inverted ?
                        Math.abs(pix_y - px) : Math.abs(pix_y - py);
                    break;
                case (ChartDataView.PICK_FOCUS_XY) :
                default :
                    pix_x = xaxis.toPixel(thispx, drawFront);
                    pix_y = yaxis.toPixel(thispy, drawFront);
                    len = inverted ?
                        (Math.sqrt(((pix_y - px)*(pix_y - px)) +
                            ((pix_x - py)*(pix_x - py)) ) ) :
                        (Math.sqrt(((pix_x - px)*(pix_x - px)) +
                            ((pix_y - py)*(pix_y - py)) ) ) ;
                    break;
                }
                if (length > len) {
                    series = i * NUM_SERIES_PER_DATA + k;
                    point = j;
                    length = len;
                }

                if (!other) k++;
			}
		}
	}
	if (series >= 0 && point >= 0) {
		di = new JCDataIndex(point,
			(ChartDataViewSeries) seriesList.get(series), series);
		di.distance = (int) length;
	}
	return di;
	} catch (Exception e) {
		e.printStackTrace(System.out);
		return null;
	}
}

/**
 * Unpick method for candle charts.  Given a data series and a datapoint
 * within that series, the function will return pixel coordinates of
 * that point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing position in screen pixels
 */
public Point unpick(int pt, int seriesIndex) {

	ChartDataViewSeries series =
		(ChartDataViewSeries) seriesList.get(seriesIndex);
	if (series == null) return null;

	double yvalue = series.getY(pt);
	double xvalue = series.getX(pt);
	int point_x = -1;
	int point_y = -1;

	if (yvalue != holeValue && xvalue != holeValue) {
		if (inverted) {
			point_x = yaxis.toPixel(yvalue, drawFront);
			point_y = xaxis.toPixel(xvalue, drawFront);
		} else {
			point_x = xaxis.toPixel(xvalue, drawFront);
			point_y = yaxis.toPixel(yvalue, drawFront);
		}
		return new Point(point_x, point_y);
	}

	return null;
}

/** Location on the chart area of the point before the point being edited. */
Point start = null;

/** Last position on the chart area of the edited point. */
Point end = null;

/** Short cut to drawing area for the chart area. */
Rectangle r = null;

boolean useRubberBand = false;

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di) {
	if (di == null || di.dataView != dataObject) return;
	boolean inverted = di.dataView.getXAxis().i_vertical;
	int seriesIndex = di.series.getDrawingOrder() / NUM_SERIES_PER_DATA;
	r = chartArea.getDrawingArea();
	start = null;
	end = null;
	useRubberBand = false;

	int startInd = seriesIndex * NUM_SERIES_PER_DATA;
	int sInd = di.seriesIndex - startInd;

	int risingCandleWidth = 0;
	int fallingCandleWidth = 0;
	JCChartStyle style;
	if (isComplex) {
		style = cf.getRisingCandleStyle(seriesIndex);
		if (style != null)
			risingCandleWidth = style.getSymbolSize();
		style = cf.getFallingCandleStyle(seriesIndex);
		if (style != null)
			fallingCandleWidth = style.getSymbolSize();
	}
	else {
		style = cf.getHiloStyle(seriesIndex);
		if (style != null) {
			risingCandleWidth = style.getSymbolSize();
			fallingCandleWidth = risingCandleWidth;
		}
	}

	double yopen = ((ChartDataViewSeries) seriesList.get(
						startInd + OPEN_SERIES_INDEX)).getY(di.point);
	double yclose = ((ChartDataViewSeries) seriesList.get(
						startInd + CLOSE_SERIES_INDEX)).getY(di.point);
	boolean risingCandle = yopen < yclose;
	int candleWidth = risingCandle ? risingCandleWidth : fallingCandleWidth;

    double thispx = ((ChartDataViewSeries) seriesList.get(
						startInd + sInd)).getX(di.point);
    double thispy = ((ChartDataViewSeries) seriesList.get(
						startInd + sInd)).getY(di.point);
	if (sInd > LO_SERIES_INDEX) {
		useRubberBand = true;
    	if (inverted) {
        	start = new Point(yaxis.toPixel(thispy)+r.x,
                            	xaxis.toPixel(thispx)+r.y - (candleWidth / 2));
			end = new Point(start.x, start.y + candleWidth);
    	} else {
        	start = new Point(xaxis.toPixel(thispx)+r.x - (candleWidth / 2),
                            	yaxis.toPixel(thispy)+r.y);
			end = new Point(start.x + candleWidth, start.y);
		}
	} else {
    	if (inverted)
        	start = new Point(yaxis.toPixel(thispy)+r.x,
                            	xaxis.toPixel(thispx)+r.y);
    	else
       		start = new Point(xaxis.toPixel(thispx)+r.x,
                            	yaxis.toPixel(thispy)+r.y);
		end = new Point(start.x, start.y);
	}
}

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y) {
	if (di == null || di.dataView != dataObject) return;
	boolean inverted = di.dataView.getXAxis().i_vertical;
	if (useRubberBand) {
		chartArea.drawRubberBand(end, start);
	}
	else {
		chartArea.drawRubberLine(start, end);
	}

	// Edit for candle charts only edits in the y direction
	if (inverted) {
		end.x = x;
	}
	else {
		end.y = y;
	}
	if (useRubberBand) {
		chartArea.drawRubberBand(end, start);
	}
	else {
		chartArea.drawRubberLine(start, end);
	}
}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 * @return
 */
int endEdit(JCDataIndex di, int x, int y) {
	if (di == null || di.dataView != dataObject) return -1;
	if (useRubberBand) {
		chartArea.drawRubberBand(end, start);
	}
	else {
		chartArea.drawRubberLine(start, end);
	}
	return (di.dataView.getXAxis().i_vertical ? x - r.x : y - r.y);
}

public int getNumSeriesPerData() {
	return NUM_SERIES_PER_DATA;
}

}
