/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: FastHiloChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

import java.util.List;
import java.awt.geom.Line2D;
import java.awt.Graphics2D;

import com.klg.jclass.util.graphics.DefineShape;

/**
 * Drawable object for hilo charts.  This object
 * is used for rendering a hilo chart based on data
 * contained in the <code>dataObject</code>.
 */
public class FastHiloChartDraw extends ChartDraw {

// number of chart data view series per hilo data series
public final static int NUM_SERIES_PER_DATA = 2;
public final static int HI_SERIES_INDEX = 0;
public final static int LO_SERIES_INDEX = 1;

protected MinMax xlim;
protected MinMax ylim;

/**
 * Default constructor.
 */
public FastHiloChartDraw() {
	super();
	xlim = new MinMax();
	ylim = new MinMax();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc() {
	recalc(-1);
}

public void recalc(int seriesIndex) {
	super.recalc();

	// Make a list of points
	// If calling for one series, use existing series list.  Otherwise,
	// refresh the series list.
	if (seriesIndex == -1) {
		seriesList = dataObject.getSeries();
	}
	if (seriesList == null) return;

	xaxis = dataObject.getXAxis();
	yaxis = dataObject.getYAxis();
	holeValue = dataObject.getHoleValue();
	inverted = xaxis.i_vertical;

	int numSeries = seriesList.size() / NUM_SERIES_PER_DATA;

	if (seriesIndex == (-1) || seriesIndex == 0) {
		// This may take awhile for large data set
		if (dataObject.getXLimits() == null) {
			return;
		}
		if (dataObject.getYLimits() == null) {
			return;
		}
		xlim = new MinMax(dataObject.getXLimits().min,
						  dataObject.getXLimits().max);
		ylim = new MinMax(dataObject.getYLimits().min,
						  dataObject.getYLimits().max);
		dataObject.getDataBounds(xlim, ylim);
	}

}

/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc) {
	try {
	// Get list of data series to be drawn using this drawable
	if (seriesList == null) return;

   	Rectangle clipBounds = gc.getClipBounds();
   	if (clipBounds.width == 0 || clipBounds.height == 0)
       	return;

  	DefineShape defShape = null;
  	if (gc instanceof DefineShape) {
      	defShape = (DefineShape) gc;
  	}

    // Setup parameters to loop thru the data series
    int numSeries = seriesList.size() / NUM_SERIES_PER_DATA;
    int firstSeries = 0;
    int lastSeries = numSeries;

    boolean xLog = xaxis.isLogarithmic();
    boolean yLog = yaxis.isLogarithmic();

	// Must draw each hilo data series in order
    for (int i = firstSeries; i < lastSeries; i++) {

		// Set up the basic information: series, style, x vals, y vals);
		boolean showSeries = true;
		boolean hasData = true;
		int firstPoint = Integer.MAX_VALUE;
		int lastPoint = Integer.MAX_VALUE;

		for (int s = NUM_SERIES_PER_DATA * i, count = 0;
				count < NUM_SERIES_PER_DATA; count++, s++) {
            ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(s);

			if (!series.isVisible()) {
				showSeries = false;
				break;
			}
			// Drop out if data is uninteresting
			if (series.maxXIndex() == 0 || series.maxYIndex() == 0) {
				hasData = false;
				break;
			}

			// Determine first and last point in data series
			firstPoint = Math.min(firstPoint, series.getFirstPoint());
			lastPoint = Math.min(lastPoint, series.getLastPoint());
		}
		if (!hasData) continue;
		if (!showSeries) continue;

		JCChartStyle style = ((ChartDataViewSeries) seriesList.get(
				NUM_SERIES_PER_DATA * i)).getStyle();
		if (style == null) continue;
		// Set up the line style
		JCLineStyle lstyle = style.getLineStyle();
		if (!(lstyle.updateGraphics(gc)))
			continue;

        Line2D.Double lineSegment = new Line2D.Double();


        // Create a series of PlotPoint arrays holding the various lines that
        // make up this plot chart.
        PlotPoint newpts[] = new PlotPoint[2];
        newpts[0] = new PlotPoint();
        newpts[1] = new PlotPoint();

		// Now loop over the data
		for (int j = firstPoint; j <= lastPoint; j++) {

			int s = NUM_SERIES_PER_DATA * i;
			ChartDataViewSeries highSeries =
				(ChartDataViewSeries) seriesList.get(s++);

			double x1 = highSeries.getX(j);
            if (xLog && x1 <= 0.0)
                x1 = holeValue;
			double y1 = highSeries.getY(j);
            if (yLog && y1 <= 0.0)
                y1 = holeValue;

			ChartDataViewSeries lowSeries =
				(ChartDataViewSeries) seriesList.get(s++);
			double x2 = lowSeries.getX(j);
            if (xLog && x2 <= 0.0)
                x2 = holeValue;
			double y2 = lowSeries.getY(j);
            if (yLog && y2 <= 0.0)
                y2 = holeValue;

            newpts[0].valid = false;
            newpts[1].valid = false;
			financialClip(x1, y1, x2, y2, xlim, ylim,
						newpts, inverted, false);

            if (newpts[0].valid && newpts[1].valid) {



				lineSegment.setLine(
                	Math.floor(newpts[0].px), Math.floor(newpts[0].py),
                    Math.floor(newpts[1].px), Math.floor(newpts[1].py));
                ((Graphics2D)gc).draw(lineSegment);


			    // if using DefineShape, begin a shape definition for this index
                if (defShape != null) {
					defineHiloShapes(defShape, gc, inverted,
						j, highSeries, lowSeries, newpts);
                }
			}
		}
		lstyle.resetGraphics(gc);
	}
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
}

protected void defineHiloShapes(DefineShape defShape, Graphics gc,
	boolean inverted, int j, ChartDataViewSeries highSeries,
	ChartDataViewSeries lowSeries, PlotPoint[] newpts)
{
	if (newpts[0].drawPoint && newpts[1].drawPoint) {
    	defShape.beginShape(new JCDataIndex(j, highSeries));
		int halfHeight = 0;
		if (!inverted) {
			halfHeight = ((int) newpts[1].py - (int) newpts[0].py) / 2;
			gc.fillRect(
				(int) newpts[0].px - 5, (int) newpts[0].py,
				11, halfHeight);
		} else {
			halfHeight = ((int) newpts[1].px - (int) newpts[0].px) / 2;
			gc.fillRect(
				(int) newpts[0].px, (int) newpts[0].py - 5,
				halfHeight, 11);
		}
		defShape.endShape(false);

    	defShape.beginShape( new JCDataIndex(j, lowSeries));
		if (!inverted) {
			gc.fillRect(
				(int) newpts[0].px - 5, (int) newpts[0].py + halfHeight,
				11, ((int) newpts[1].py - (int) newpts[0].py) - halfHeight);
		} else {
			gc.fillRect(
				(int) newpts[0].px + halfHeight, (int) newpts[0].py - 5,
				((int) newpts[1].px - (int) newpts[0].px) - halfHeight, 11);
		}
		defShape.endShape(false);

	} else {
		if (newpts[0].drawPoint) {
    		defShape.beginShape(new JCDataIndex(j, highSeries));
		} else {
    		defShape.beginShape(new JCDataIndex(j, lowSeries));
		}
		if (!inverted) {
			gc.fillRect(
				(int) newpts[0].px - 5, (int) newpts[0].py,
				11, (int) newpts[1].py - (int) newpts[0].py);
		} else {
			gc.fillRect(
				(int) newpts[0].px, (int) newpts[0].py - 5,
				(int) newpts[1].px - (int) newpts[0].px, 11);
		}
		defShape.endShape(false);
	}

}

/**
 * Pick method for hilo charts.  Given a point relative to the
 * chart area, this method finds the closest point.
 * @param pt (x,y) value relative to the chart area
 * @param focus pick focus used to determine distance between pick point
 * and data point
 * @return <code>JCDataIndex</code> encapsulating the point index and data
 * series of the selected point
 */
public JCDataIndex pick(Point pt, int focus) {
	Rectangle r = chartArea.getDrawingArea();
    int px = pt.x - r.x;
    int py = pt.y - r.y;

	int series = -1;
	int point = -1;
	JCDataIndex di = null;
	double length = Double.MAX_VALUE;
	if (seriesList == null) return di;

    double holeValue = dataObject.getHoleValue();
    boolean xLog = xaxis.isLogarithmic();
    boolean yLog = yaxis.isLogarithmic();

    int numSeries = seriesList.size() / NUM_SERIES_PER_DATA;
    int firstSeries = 0;
    int lastSeries = numSeries;
	int pix_x, pix_y;
	try {
    for (int i = firstSeries; i < lastSeries; i++) {

        boolean hasData = true;
        boolean showSeries = true;
        int firstPoint = Integer.MAX_VALUE;
        int lastPoint = Integer.MAX_VALUE;
        for (int s = NUM_SERIES_PER_DATA * i, count = 0;
                count < NUM_SERIES_PER_DATA; count++, s++) {
            ChartDataViewSeries thisSeries =
				(ChartDataViewSeries) seriesList.get(s);

            if (!thisSeries.isVisible()) {
                showSeries = false;
                break;
            }
            // Drop out if data is uninteresting
            if (thisSeries.maxXIndex() == 0 || thisSeries.maxYIndex() == 0) {
                hasData = false;
                break;
            }

            // Determine first and last point in data series
            firstPoint = Math.min(firstPoint, thisSeries.getFirstPoint());
            lastPoint = Math.min(lastPoint, thisSeries.getLastPoint());

        }
        if (!hasData) continue;
        if (!showSeries) continue;

        int s = NUM_SERIES_PER_DATA * i;

        double highx[] =
			((ChartDataViewSeries) seriesList.get(s+HI_SERIES_INDEX)).getX();
        double highy[] =
			((ChartDataViewSeries) seriesList.get(s+HI_SERIES_INDEX)).getY();
        double lowx[] =
			((ChartDataViewSeries) seriesList.get(s+LO_SERIES_INDEX)).getX();
        double lowy[] =
			((ChartDataViewSeries) seriesList.get(s+LO_SERIES_INDEX)).getY();

        double thispx = 0.0;
        double thispy = 0.0;
        for (int j = firstPoint; j <= lastPoint; j++) {
            for (int k = 0; k < NUM_SERIES_PER_DATA; k++) {
                switch (k) {
                    case HI_SERIES_INDEX:
                        thispx = highx[j];
                        thispy = highy[j];
                        break;
                    case LO_SERIES_INDEX:
                        thispx = lowx[j];
                        thispy = lowy[j];
                        break;
				}
                if (thispx == holeValue ||
                    thispy == holeValue ||
                    thispx > xlim.max ||
                    xlim.min > thispx ||
                    thispy > ylim.max ||
                    ylim.min > thispy ||
					(xLog && thispx <= 0.0) ||
					(yLog && thispy <= 0.0) ) {
                     continue;
                }

                double len;
                switch(focus) {
                case (ChartDataView.PICK_FOCUS_X) :
                    pix_x = xaxis.toPixel(thispx, drawFront);
                    len = inverted ?
                        Math.abs(pix_x - py) : Math.abs(pix_x - px);
                    break;
                case (ChartDataView.PICK_FOCUS_Y) :
                    pix_y = yaxis.toPixel(thispy, drawFront);
                    len = inverted ?
                        Math.abs(pix_y - px) : Math.abs(pix_y - py);
                    break;
                case (ChartDataView.PICK_FOCUS_XY) :
                default :
                    pix_x = xaxis.toPixel(thispx, drawFront);
                    pix_y = yaxis.toPixel(thispy, drawFront);
                    len = inverted ?
                        (Math.sqrt(((pix_y - px)*(pix_y - px)) +
                            ((pix_x - py)*(pix_x - py)) ) ) :
                        (Math.sqrt(((pix_x - px)*(pix_x - px)) +
                            ((pix_y - py)*(pix_y - py)) ) ) ;
                    break;
                }
                if (length > len) {
                    series = i * NUM_SERIES_PER_DATA + k;
                    point = j;
                    length = len;
                }
            }
        }
	}
	if (series >= 0 && point >= 0) {
		di = new JCDataIndex(point,
				(ChartDataViewSeries) seriesList.get(series), series);
		di.distance = (int) length;
	}
	return di;
	} catch (Exception e) {
		e.printStackTrace(System.out);
		return null;
	}
}

/**
 * Unpick method for hilo charts.  Given a data series and a datapoint
 * within that series, the function returns pixel coordinates of
 * that point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing the position in screen pixels
 */
public Point unpick(int pt, int seriesIndex) {

	ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(seriesIndex);
	if (series == null) return null;

	double yvalue = series.getY(pt);
	double xvalue = series.getX(pt);
	int point_x = -1;
	int point_y = -1;

	if (yvalue != holeValue && xvalue != holeValue) {
		if (inverted) {
			point_x = yaxis.toPixel(yvalue, drawFront);
			point_y = xaxis.toPixel(xvalue, drawFront);
		} else {
			point_x = xaxis.toPixel(xvalue, drawFront);
			point_y = yaxis.toPixel(yvalue, drawFront);
		}
		return new Point(point_x, point_y);
	}

	return null;
}

/** Location on the chart area of the point before the point being edited. */
Point start = null;

/** Last position on the chart area of the edited point. */
Point end = null;

/** Short cut to drawing area for the chart area. */
Rectangle r = null;

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set that is plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di) {
	if (di == null || di.dataView != dataObject) return;
	int seriesIndex = di.series.getDrawingOrder() / NUM_SERIES_PER_DATA;
	r = chartArea.getDrawingArea();
	start = null;
	end = null;

	int startInd = seriesIndex * NUM_SERIES_PER_DATA;
    int sInd = di.seriesIndex - startInd;
    double thispx =
		((ChartDataViewSeries) seriesList.get(startInd + sInd)).getX(di.point);
    double thispy =
		((ChartDataViewSeries) seriesList.get(startInd + sInd)).getY(di.point);
    if (inverted)
        start = new Point(yaxis.toPixel(thispy)+r.x,
                            xaxis.toPixel(thispx)+r.y);
    else
        start = new Point(xaxis.toPixel(thispx)+r.x,
                            yaxis.toPixel(thispy)+r.y);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y) {
	if (di == null || di.dataView != dataObject) return;
	boolean inverted = di.dataView.getXAxis().i_vertical;
	chartArea.drawRubberLine(start, end);

	// Edit for hilo charts only edits in the y direction
	if (end == null)
		// Create new point at x axis value of piced point with varying
		// y axis value.
		if (inverted)
			end = new Point(x,
				di.dataView.getXAxis().toPixel(di.series.getX(di.point), drawFront) + r.y);
		else
			end = new Point(
				di.dataView.getXAxis().toPixel(di.series.getX(di.point), drawFront) + r.x, y);
	else {
		if (inverted)
			end.x = x;
		else
			end.y = y;
	}
	chartArea.drawRubberLine(start, end);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set that is plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 * @return
 */
int endEdit(JCDataIndex di, int x, int y) {
	if (di == null || di.dataView != dataObject) return -1;
	chartArea.drawRubberLine(start, end);
	return (di.dataView.getXAxis().i_vertical ? x - r.x : y - r.y);
}

public int getNumSeriesPerData() {
	return NUM_SERIES_PER_DATA;
}

}
