/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: FastPlotChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;
import java.awt.geom.Line2D;

/**
 * Drawable object for plot charts.  This object
 * is used for rendering a plot chart based on data
 * contained in the <code>dataObject</code>.
 */
public class FastPlotChartDraw extends ChartDraw {

public final static int TOP_EDGE    = (1 << 3);
public final static int BOTTOM_EDGE = (1 << 2);
public final static int RIGHT_EDGE  = (1 << 1);
public final static int LEFT_EDGE   = (1 << 0);

// Local copies of data needed for draw(), pick(), and unpick().  They are
// set in recalc().

protected MinMax xlim;
protected MinMax ylim;

/**
 * Default constructor.
 */
public FastPlotChartDraw()
{
	super();
	xlim = new MinMax();
	ylim = new MinMax();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc()
{
	recalc(-1, -1, -1);
}

public void recalc(int startPoint, int endPoint)
{
	recalc(startPoint, endPoint, -1);
}

public void recalc(int startPoint, int endPoint, int seriesIndex)
{
	super.recalc();

	boolean fu = false;
	// if start point and end point are specified, this is a fast
	// update case
	if (startPoint >= 0 && endPoint >= 0 && endPoint >= startPoint) {
		fu = true;
	}

	// Make a list of points
	// If being called for a particular series, use the previous series list.
	// If being called for all series, refresh the series list.
	if (seriesIndex == -1) {
		seriesList = dataObject.getSeries();
	}
	if (seriesList == null) {
		return;
	}

	// cache important info for later use by draw()
	xaxis = dataObject.getXAxis();
	yaxis = dataObject.getYAxis();
	inverted = xaxis.i_vertical;
	holeValue = dataObject.getHoleValue();

	// In fast update case, check the limits against the new points
	if (fu) {
		int firstSeries = 0;
		int lastSeries = seriesList.size();
		for (int i = firstSeries; i < lastSeries; i++) {
			ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(i);
			if (!series.isVisible()) {
                continue;
            }
			if (series.maxXIndex() == 0 || series.maxYIndex() == 0) {
                continue;
            }
			double xvals[] = series.getX();
			double yvals[] = series.getY();
			for (int j = startPoint; j <= endPoint; j++) {
				xlim.setMinMax(xvals[j], holeValue, xaxis.isLogarithmic());
				ylim.setMinMax(yvals[j], holeValue, yaxis.isLogarithmic());
			}
		}
		dataObject.getDataBounds(xlim, ylim);
	}
    else if (seriesIndex == -1 || seriesIndex == 0) {
		// This may take awhile for large data set
		if (dataObject.getXLimits() == null) {
			return;
		}
		if (dataObject.getYLimits() == null) {
			return;
		}
		xlim = new MinMax(dataObject.getXLimits().min,
						  dataObject.getXLimits().max);
		ylim = new MinMax(dataObject.getYLimits().min,
						  dataObject.getYLimits().max);
		dataObject.getDataBounds(xlim, ylim);
	}
}

/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc)
{
	draw(gc, -1, -1);
}

public void draw(Graphics gc, int startPoint, int endPoint)
{
	try {
        // Get list of data series to be drawn using this drawable
        if (seriesList == null) {
            return;
        }

        Rectangle clipBounds = gc.getClipBounds();
        if (clipBounds.width == 0 || clipBounds.height == 0) {
            return;
        }

        boolean fu = false;
        // If start point and end point are specified, this is a fast
        // update case
        if (startPoint >= 0 && endPoint >= 0 && endPoint >= startPoint) {
            fu = true;
        }

        // create a separate GC to use when drawing the lines
        Graphics lineGC = gc.create();

        DefineShape pointShapeDef = null;
        if (lineGC instanceof DefineShape) {
            pointShapeDef = (DefineShape) gc;
        }

        // Setup parameters to loop thru the data series
        boolean xlog = xaxis.isLogarithmic();
        boolean ylog = yaxis.isLogarithmic();
        int firstSeries = 0;
        int lastSeries = seriesList.size();
        for (int i = firstSeries; i < lastSeries; i++) {
            // Set up the basic information: series, style, x vals, y vals);
            ChartDataViewSeries series =
                (ChartDataViewSeries) seriesList.get(i);

            // Drop out if data is uninteresting
            if (!series.isVisible()) {
                continue;
            }
            if (series.maxXIndex() == 0 || series.maxYIndex() == 0) {
                continue;
            }

            // if not fast update, get start point and end point.
            if (!fu) {
                startPoint = series.getFirstPoint();
                endPoint = series.getLastPoint();
            }
            else {
                int boundaryPoint = series.getLastPoint();
                if (startPoint < boundaryPoint) {
                    startPoint = boundaryPoint;
                }
            }

            double xvals[] = series.getX();
            double yvals[] = series.getY();

            // Initialize first point to a false point
            double x1 = 0.0;
            double y1 = holeValue;
            boolean inArray = false;

            // If fast update case, first point is the last point in the
            // previously drawn.
            if (fu && startPoint > 0) {
                x1 = xvals[startPoint - 1];
                y1 = yvals[startPoint - 1];
            }

            JCChartStyle style = series.getStyle();
            if (style == null) continue;
            // Set up the line style
            JCLineStyle lstyle = style.getLineStyle();


            boolean noLines = false;
            if (!(lstyle.updateGraphics(lineGC))) {
                noLines = true;
            }

            Line2D.Double lineSegment = new Line2D.Double();

            // Set up the point style
            JCSymbolStyle pstyle = style.getSymbolStyle();
            boolean noShape = (pstyle.getShape() == JCSymbolStyle.NONE);

            // previous point
            PlotPoint prevpt = new PlotPoint();
            prevpt.valid = false;
            // array to hold newly clipped points
            PlotPoint newpts[] = new PlotPoint[2];
            newpts[0] = new PlotPoint();
            newpts[1] = new PlotPoint();

            for (int j = startPoint; j <= endPoint; j++) {
                double x2 = xvals[j];
                if (xlog && x2 <= 0.0) {
                    x2 = holeValue;
                }
                double y2 = yvals[j];
                if (ylog && y2 <= 0.0) {
                    y2 = holeValue;
                }

                newpts[0].valid = false;
                newpts[1].valid = false;
                //DAM I think there may be a bug here : xlim & ylim are data
                // limits, not axis limits.  This could cause things to clip
                // when they shouldn't
                inArray = plotClip(x1, y1, x2, y2, xlim, ylim,
                                   newpts, inArray, inverted);

                // Up to two new points may have been added in newpts.
                for (int k = 0; k < newpts.length; k++) {
                    // Draw line between this new point and previous
                    // point (if valid).
                    if (!noLines && newpts[k].valid) {
                        if (prevpt.valid && chartType == JCChart.PLOT) {
                            // Check for color handler and custom drawing color
                            Color col = null;
                            if (dataObject.getColorHandler() != null) {
                                col = dataObject.getColorHandler().getColor(
                                    x1, y1, x2, y2);
                                if (col == null) {
                                    col = lstyle.getColor();
                                }
                                if (!lineGC.getColor().equals(col)) {
                                    lineGC.setColor(col);
                                }
                            }
                            lineSegment.setLine(prevpt.px, prevpt.py,
                                                newpts[k].px, newpts[k].py);
                            ((Graphics2D)lineGC).draw(lineSegment);
                        }
                    }
                    // Always draw symbol at previous point (if valid and not
                    // clipped) the first time through.  Do it again if second
                    // point is valid.
                    if (k == 0 || newpts[k].valid) {
                        if (prevpt.valid && prevpt.drawPoint && !noShape) {
                            // Check for color handler and custom drawing color
                            Color col = null;
                            if (dataObject.getColorHandler() != null) {
                                col = dataObject.getColorHandler().getColor(x1, y1);
                            }

                            // Associate point shapes with data index
                            if (pointShapeDef != null) {
                                pointShapeDef.markShape(
                                    new JCDataIndex(j-1, series));
                            }

                            if (usingDoublePixels) {
                                pstyle.draw(gc, prevpt.px, prevpt.py, col);
                            }
                            else {
                                pstyle.draw(gc,
                                            (int) prevpt.px, (int) prevpt.py, col);
                            }

                            // De-associate point shapes with data index
                            if (pointShapeDef != null) {
                                pointShapeDef.unmarkShape();
                            }

                        }
                        // New Point becomes previous point
                        prevpt.px = newpts[k].px;
                        prevpt.py = newpts[k].py;
                        prevpt.drawPoint = newpts[k].drawPoint;
                        prevpt.valid = newpts[k].valid;
                    }
                }
                // if final point is clipped, mark new previout point as invalid
                if (!inArray) {
                    prevpt.valid = false;
                }

                x1 = x2;
                y1 = y2;
            }
            // Draw symbol for final point (if valid and not clipped);
            if (prevpt.valid && prevpt.drawPoint && !noShape) {
                // Check for color handler and custom drawing color
                Color col = null;
                if (dataObject.getColorHandler() != null) {
                    col = dataObject.getColorHandler().getColor(x1, y1);
                }

                // associate point shapes with data index
                if (pointShapeDef != null) {
                    pointShapeDef.markShape(new JCDataIndex(endPoint, series));
                }

                if (usingDoublePixels) {
                    pstyle.draw(gc, prevpt.px, prevpt.py, col);
                }
                else {
                    pstyle.draw(gc, (int) prevpt.px, (int) prevpt.py, col);
                }

                // de-associate point shapes with data index
                if (pointShapeDef != null) {
                    pointShapeDef.unmarkShape();
                }
            }
        }
        lineGC.dispose();
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
}

/**
 * Pick method for plot charts.  Given a point relative to the
 * chart area, this method will find the closest point.
 * @param pt (x,y) value relative to the chart area
 * @param focus pick focus used to determine distance between pick point
 * and data point
 * @return <code>JCDataIndex</code> encapsulating the point index and data
 * series of the selected point
 */
public JCDataIndex pick(Point pt, int focus)
{
	Rectangle r = chartArea.getDrawingArea();
	int px = pt.x - r.x;
	int py = pt.y - r.y;

	int series = -1;
	int point = -1;
	JCDataIndex di = null;

	double length = Double.MAX_VALUE;
	if (seriesList == null) return di;

	boolean xlog = xaxis.isLogarithmic();
	boolean ylog = yaxis.isLogarithmic();

	int firstSeries = 0;
	int lastSeries = seriesList.size();
	double thispx = 0.0;
	double thispy = 0.0;
	double len;
	int pix_x, pix_y;
	try {
        for (int i = firstSeries; i < lastSeries; i++) {
            ChartDataViewSeries thisSeries =
                (ChartDataViewSeries) seriesList.get(i);
            if (thisSeries == null || !thisSeries.isVisible()) {
                continue;
            }

            int startPoint = thisSeries.getFirstPoint();
            int endPoint = thisSeries.getLastPoint();
            if (startPoint == Integer.MAX_VALUE ||
                endPoint == Integer.MAX_VALUE)
            {
                continue;
            }
            double xvals[] = thisSeries.getX();
            double yvals[] = thisSeries.getY();

            for (int j = startPoint; j <= endPoint ; j++) {
                thispx = xvals[j];
                thispy = yvals[j];

                if (thispx == holeValue     ||
                    thispy == holeValue     ||
                    thispx > xlim.max       ||
                    thispx < xlim.min       ||
                    thispy > ylim.max       ||
                    thispy < ylim.min       ||
                    (xlog && thispx <= 0.0) ||
                    (ylog && thispy <= 0.0))
                {
                    continue;
                }

                len = Double.MAX_VALUE;
                switch(focus) {
                case (ChartDataView.PICK_FOCUS_X) :
                    pix_x = xaxis.toPixel(thispx, drawFront);
                    len = inverted ?
                        Math.abs(pix_x - py) : Math.abs(pix_x - px);
                    break;
                case (ChartDataView.PICK_FOCUS_Y) :
                    pix_y = yaxis.toPixel(thispy, drawFront);
                    len = inverted ?
                        Math.abs(pix_y - px) : Math.abs(pix_y - py);
                    break;
                case (ChartDataView.PICK_FOCUS_XY) :
                default :
                    pix_x = xaxis.toPixel(thispx, drawFront);
                    pix_y = yaxis.toPixel(thispy, drawFront);
                    len = inverted ?
                        (Math.sqrt(((pix_y - px)*(pix_y - px)) +
                                   ((pix_x - py)*(pix_x - py)) ) ) :
                        (Math.sqrt(((pix_x - px)*(pix_x - px)) +
                                   ((pix_y - py)*(pix_y - py)) ) ) ;
                    break;
                }
                if (len < length) {
                    series = i;
                    point = j;
                    length = len;
                }
            }
        }
        if (series >= 0 && point >= 0) {
            di = new JCDataIndex(point,
                                 (ChartDataViewSeries) seriesList.get(series),
                                 series);
            di.distance = (int) length;
        }
        return(di);
	}
    catch (Exception e) {
		e.printStackTrace(System.out);
		return(null);
	}
}

/**
 * Unpick method for plot charts.  Given a data series and a datapoint
 * within that series, the function will return pixel coordinates of
 * that point that are relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing the position in screen pixels
 */
public Point unpick(int pt, int seriesIndex)
{
	ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(seriesIndex);

	double yvalue = series.getY(pt);
	double xvalue = series.getX(pt);
	int point_x = -1;
	int point_y = -1;

	if (yvalue != holeValue && xvalue != holeValue) {
		if (inverted) {
			point_x = yaxis.toPixel(yvalue, drawFront);
			point_y = xaxis.toPixel(xvalue, drawFront);
		}
        else {
			point_x = xaxis.toPixel(xvalue, drawFront);
			point_y = yaxis.toPixel(yvalue, drawFront);
		}
		return(new Point(point_x, point_y));
	}

	return(null);
}

/** Location on the chart area of the point before the point being edited. */
Point prev = null;

/** Location on the chart area of the point after the point being edited. */
Point next = null;

/** Last position on the chart area of the edited point. */
Point last = null;

/** Short cut to drawing area for the chart area. */
Rectangle r = null;

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set that is plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di)
{
	if (di == null || di.dataView != dataObject || di.series == null) {
        return;
    }
	r = chartArea.getDrawingArea();
	prev = next = last = null;

	int j = di.point;
	if (j - 1 >= 0) {
		double tmppx = di.series.getX(j - 1);
		double tmppy = di.series.getY(j - 1);
		if (tmppx != Double.MAX_VALUE          &&
			tmppy != Double.MAX_VALUE          &&
			tmppx != dataObject.getHoleValue() &&
			tmppy != dataObject.getHoleValue() &&
			JCChartUtil.locateXY(tmppx, tmppy, xlim.min, xlim.max,
                                 ylim.min, ylim.max) == 0)
        {
			prev = inverted ?
				new Point(yaxis.toPixel(tmppy)+r.x, xaxis.toPixel(tmppx)+r.y) :
				new Point(xaxis.toPixel(tmppx)+r.x, yaxis.toPixel(tmppy)+r.y);
        }
	}
	if (j + 1 <= di.series.getLastPoint()) {
		double tmppx = di.series.getX(j + 1);
		double tmppy = di.series.getY(j + 1);
		if (tmppx != Double.MAX_VALUE          &&
			tmppy != Double.MAX_VALUE          &&
			tmppx != dataObject.getHoleValue() &&
			tmppy != dataObject.getHoleValue() &&
			JCChartUtil.locateXY(tmppx, tmppy, xlim.min, xlim.max,
                                 ylim.min, ylim.max) == 0)
        {
			next = inverted ?
				new Point(yaxis.toPixel(tmppy)+r.x, xaxis.toPixel(tmppx)+r.y) :
				new Point(xaxis.toPixel(tmppx)+r.x, yaxis.toPixel(tmppy)+r.y);
        }
	}
}

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y)
{
	if (di == null || di.dataView != dataObject) {
        return;
    }
    JCAxis xAxis = di.dataView.getXAxis();
	boolean inverted = xAxis.i_vertical;
	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);

	// Edit for plot charts only edits in the y direction
	if (last == null) {
		// Create new point at x axis value of piced point with varying
		// y axis value.

        int xPixel = xAxis.toPixel(di.series.getX(di.point), drawFront);
		if (inverted) {
			last = new Point(x, xPixel + r.y);
        }
		else {
			last = new Point(xPixel + r.x, y);
        }
    }
	else {
		if (inverted) {
			last.x = x;
        }
		else {
			last.y = y;
        }
	}
	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set that is plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited.
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 * @return
 */
int endEdit(JCDataIndex di, int x, int y)
{
	if (di == null || di.dataView != dataObject) {
        return(-1);
    }
	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);
	return(di.dataView.getXAxis().i_vertical ? x - r.x : y - r.y);
}

/**
 * Clips the vector between points (<i>x1</i>, <i>y1</i>) and (<i>x2</i>, <i>y2</i>),
 * placing the resulting points in the specified <code>PlotPoint</code> array.
 * This method may add two (if staring a new <code>PlotPoint</code> array), one
 * (if continuing an existing <code>PlotPoint</code> array), or zero (if the vector
 * defined by the two points is outside the clip window) points to the <code>PlotPoint</code> array.
 * @param x1 the <i>x</i> value of the first point to be clipped
 * @param y1 the <i>y</i> value of the first point to be clipped
 * @param x2 the <i>x</i> value of the second point to be clipped
 * @param y2 the <i>y</i> value of the second point to be clipped
 * @param xlim the <i>x</i> limits of the clip window
 * @param ylim the <i>y</i> limits of the clip window
 * @param arr the array of <code>PlotPoint</code> instances to place the resulting
 * clipped points into
 * @param inArray <code>false</code> if starting a new <code>PlotPoint</code> array;
 * <code>true</code> if using an existing <code>PlotPoint</code> array
 * @param inverted <code>true</code> if axes are inverted
 * @return new value of <code>inArray</code> (<code>false</code> if we've encountered a clipped point;
 * <code>true</code> if the last added points are inside the limits)
 */
boolean plotClip(double x1, double y1, double x2, double y2, MinMax xlim,
                 MinMax ylim, PlotPoint[] arr, boolean inArray,
                 boolean inverted)
{
    double x, y;

    double xMin  = xlim.min;
    double xMax  = xlim.max;
    double yMin  = ylim.min;
    double yMax  = ylim.max;

    /*
     * A vector between points (x1, y1) and (x2, y2) is clipped for every
     * iteration;  oc1 and oc2 are the original location codes, while
     * c1 and c2 are modified in the clipping portion below; keeping them
     * separate allows the code below the clipping to decide whether the
     * array of points should be drawn or not.  Each contiguous sequence
     * of vectors of which the first and last are partially visible (and
     * hence possibly clipped) and the intermediate vectors are totally
     * visible make up a list of vectors to be drawn.  As soon as oc2
     * determines that the second point is not visible, inArray is made
     * false which tells the calling routine to draw the list of vectors.
     */

	boolean draw1 = true;
	boolean draw2 = true;
    int oc1 = 0, oc2 = 0;
    boolean holes = false;
    /* If inArray = true the first point MUST be visible */

	boolean doClipping = true;
    if (!inArray) {    /* check first point only if inArray is false */
        if (y1 == holeValue||
			x1 == holeValue) {
            oc1 = 1;
            doClipping = false;
            holes = true;
        }
        else {
    		int ux = ((x1 > xMax ? 1 : 0) << 1);
    		int lx = ((x1 < xMin ? 1 : 0) << 0);
    		int uy = ((y1 > yMax ? 1 : 0) << 3);
    		int ly = ((y1 < yMin ? 1 : 0) << 2);
    		oc1 = (ux | lx | uy | ly);
        }
    }
    else {
    	int ux = ((x1 > xMax ? 1 : 0) << 1);
    	int lx = ((x1 < xMin ? 1 : 0) << 0);
    	int uy = ((y1 > yMax ? 1 : 0) << 3);
    	int ly = ((y1 < yMin ? 1 : 0) << 2);
    	oc1 = (ux | lx | uy | ly);
	}
    if (y2 == holeValue || x2 == holeValue) {
        oc2 = 1;
        doClipping = false;
        holes = true;
    }
    else {
		//Bug 6145 : Had to replace this line with the following two.
		//It was giving incorrect results under Solaris 2.6, jdk1.1.6
		//with native threads
    	//int ux = ((x2 > xMax ? 1 : 0) << 1);
    	int tmpux = (x2 > xMax ? 1 : 0);
    	int ux = (tmpux << 1);
    	int lx = ((x2 < xMin ? 1 : 0) << 0);
    	int uy = ((y2 > yMax ? 1 : 0) << 3);
    	int ly = ((y2 < yMin ? 1 : 0) << 2);
    	oc2 = (ux | lx | uy | ly);
    }

    int c1 = oc1;
    int c2 = oc2;
	/*
     * (c1 | c2)==0  means that points are accepted (both inside window)
     * (c1 & c2)!=0 means that the points are rejected (both outside same edge)
	 */
	int count = 0;
    while ( doClipping && !(((c1 | c2) == 0) || ((c1 & c2) != 0)) ) {
        /*
         * At this point, at least one point is outside the window;
         * c is the code of that point.
         */

        int c = (c1 != 0) ? c1 : c2;

        /*
         * Determine which edge it is outside of; once it has been
         * clipped against this edge, we only have to calculate if
         * the new point is outside one of the edges perpendicular
         * to the edge clipped against.
         */

		int ubound;
		int lbound;
        if ((c & LEFT_EDGE) != 0) {
            x = xMin;
           	y = y1 + (y2 - y1) * (xMin - x1) / (x2 - x1);
    		ubound = ((y > yMax ? 1 : 0) << 3);
    		lbound = ((y < yMin ? 1 : 0) << 2);
    		c = ubound | lbound;
        }
        else if ((c & RIGHT_EDGE) != 0) {
            y = y1 + (y2 - y1) * (xMax - x1) / (x2 - x1);
            x = xMax;
    		ubound = ((y > yMax ? 1 : 0) << 3);
    		lbound = ((y < yMin ? 1 : 0) << 2);
    		c = ubound | lbound;
        }
        else if ((c & BOTTOM_EDGE) != 0) {
            x = x1 + (x2 - x1) * (yMin - y1) / (y2 - y1);
            y = yMin;
    		ubound = ((x > xMax ? 1 : 0) << 1);
    		lbound = ((x < xMin ? 1 : 0) << 0);
    		c = ubound | lbound;
        }
        else {                  /* must be TOP_EDGE */
            x = x1 + (x2 - x1) * (yMax - y1) / (y2 - y1);
            y = yMax;
    		ubound = ((x > xMax ? 1 : 0) << 1);
    		lbound = ((x < xMin ? 1 : 0) << 0);
    		c = ubound | lbound;
        }

        /* set up next code for the clipped endpoint */
        if (c1 != 0) {
            c1 = c;
            x1 = x;
            y1 = y;
			draw1 = false;
			if (count > 4) {
				/* Too many iterations.  A point should never be clipped
					more than twice.  If it has, then it indicates a
					cycle of round-off errors that may go on endlessly.
					Better to just call it quits as we'll be close enough
					to the correct value for our purposes.
				*/
				c1 = 0;
			}
			count = (c1 == 0) ? 0 : count+1;
        }
        else {
            c2 = c;
            x2 = x;
            y2 = y;
			draw2 = false;
			if (count > 4) {
				/* Too many iterations.  A point should never be clipped
					more than twice.  If it has, then it indicates a
					cycle of round-off errors that may go on endlessly.
					Better to just call it quits as we'll be close enough
					to the correct value for our purposes.
				*/
				c2 = 0;
			}
			count = (c2 == 0) ? 0 : count+1;
        }
    }

    /*
     * If the vector is accepted (i.e. at least a portion is visible) then
     * either continue the current array or start a new one.  Then
     * determine whether to continue array or not (based on oc2 and
     * the buffer limit).
     */

    /* Need to check case where one point is visible and the other is a hole */
    boolean plotOnlyOnePoint = !inArray && holes && (oc1 == 0 || oc2 == 0);

	/*
     * (c1 | c2) == 0  means that points are accepted (i.e. both inside window)
	 */
    if (((c1 | c2) == 0) || plotOnlyOnePoint) {
        if (inArray) {             /* Continue array in most cases */
            if (!holes) {           /* Second point is not a hole */
				addClippedPoint(arr, 0, x2, y2,
								xaxis, yaxis, inverted, draw2);
                inArray = !(oc2 != 0);
            }
            else {                  /* 2nd point a hole -- end array */
                inArray = false;
            }
        }
        else {                                  /* Start a new array */
												/* Second point is visible */
            if (plotOnlyOnePoint && oc2 == 0) {
                x1 = x2;
                y1 = y2;
				draw1 = draw2;
			}

			addClippedPoint(arr, 0, x1, y1, xaxis, yaxis, inverted, draw1);

            if (!plotOnlyOnePoint) {
				addClippedPoint(arr, 1, x2, y2, xaxis, yaxis, inverted, draw2);
            }
            inArray = (oc2 == 0);           /* Is a next point possible? */
        }
    }
    else {
        inArray = false;
    }

    return(inArray);
}

}
