/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCAxisFormula.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

/**
 * Class specifying a relationship between the current {@link JCAxis} and
 * another axis. The relationship is defined in terms of the formula
 * <TT>y = mx + c</TT>, where <TT>x</TT> is the originating axis, <TT>m</TT> and
 *  <TT>c</TT> are constants, and <TT>y</TT> is the current axis.
 */
public class JCAxisFormula implements java.io.Serializable {

private JCAxis    originator = null;
private JCAxis    parent = null;

protected double  constant = 0.0;
protected double  multiplier = 1.0;

public JCAxisFormula() {}

public JCAxisFormula(JCAxis p)
{
    parent = p;
}

/**
 * Sets the <code>Constant</code> property of the <code>JCAxisFormula</code>,
 * which specified the "<TT>c</TT>" value in the relationship
 * <TT>y = mx + c</TT>.
 * @param val Double value representing "<TT>c</TT>" in <TT>y = mx + c</TT>
 */
public void setConstant(double val)
{
    if (constant == val) {
        return;
    }
    constant = val;
    if (parent != null) {
        parent.setChanged(true, Changeable.RECALC);
    }
}

/**
 * Gets the <code>Constant</code> property of the <code>JCAxisFormula</code>,
 * which specified the "<TT>c</TT>" value in the relationship
 * <TT>y = mx + c</TT>.
 * @return A double value representing "<TT>c</TT>" in <TT>y = mx + c</TT>.
 */
public double getConstant()
{
    return(constant);
}

/**
 * Sets the <code>Multiplier</code> property of the <code>JCAxisFormula</code>,
 * which specifies the "<TT>m</TT>" value in the relationship
 * <TT>y = mx + c</TT>.
 * @param val double value representing "<TT>m</TT>" in <TT>y = mx + c</TT>
 */
public void setMultiplier(double val)
{
    if (multiplier == val) {
        return;
    }
    multiplier = val;
    if (parent != null) {
        parent.setChanged(true, Changeable.RECALC);
    }
}

/**
 * Gets the <code>Multiplier</code> property of the <code>JCAxisFormula</code>,
 * which specifies the "<TT>m</TT>" value in the relationship
 * <TT>y = mx + c</TT>.
 * @return A double value representing "<TT>m</TT>" in <TT>y = mx + c</TT>
 */
public double getMultiplier()
{
    return(multiplier);
}

/**
 * Gets the <code>Originator</code> property of the <code>JCAxisFormula</code>,
 * which is an object representing the axis that is related to the current
 * axis by the formula <TT>y = mx + c</TT>. The originator is "<TT>x</TT>".
 * @return <code>JCAxis</code> instance representing the originating axis.
 */
public JCAxis getOriginator()
{
    return(originator);
}

/**
 * Sets the <code>Originator</code> property of the <code>JCAxisFormula</code>,
 * which is an object representing the axis that is related to the current
 * axis by the formula <TT>y = mx + c</TT>. The originator is "<TT>x</TT>".
 * @param org
 */
public void setOriginator(JCAxis org)
{
    if (originator == org) {
        return;
    }
    if (originator == parent) {
        throw new IllegalArgumentException("JCAxis cannot be related to itself");
    }
    originator = org;
    if (parent != null) {
        parent.setChanged(true, Changeable.RECALC);
    }
}

/**
 * Allows access to parent <code>JCAxis</code> object.
 * @return
 */
public JCAxis getParent()
{
    return(parent);
}

/**
 * Sets up parameters based on the related axis.
 */
void calcRelatedParams()
{
    if (originator == null) {
        return;
    }
    if (multiplier == 0.0) {
        return;
    }

    double absm = (multiplier < 0.0) ? -multiplier : multiplier;

    if (parent.i_vertical != originator.i_vertical) {
        parent.i_vertical = originator.i_vertical;
        // if changing the orientation of this related axis, reshape it
        // to the correct starting size.
        parent.setInitialSize();
    }
    parent.reversed = originator.reversed;

    // Figure out tick and num spacing
    if (parent.tickSpacing.isDefault && parent.numSpacing.isDefault) {
        parent.numSpacing.value = originator.numSpacing.value * absm;
        parent.numSpacing.value =
            JCChartUtil.precCorrect(parent.getPrecision(),
                                    parent.numSpacing.value);
        if (parent.tickSpacing.isDefault) {
            parent.tickSpacing.value = parent.numSpacing.value / 2.0;
        }
    }
    else if (parent.tickSpacing.isDefault) {
        parent.tickSpacing.value = parent.numSpacing.value / 2.0;
    }
    else if (parent.numSpacing.isDefault) {
        parent.numSpacing.value = parent.tickSpacing.value * 2.0;
    }

    // Figure out min and max
    if (multiplier > 0) {
        // Min
        if (parent.min.isDefault) {
            parent.min.value = originator.min.value * multiplier + constant;
        }
        else if (originator.min.isDefault) {
            originator.min.value = (parent.min.value - constant) / multiplier;
        }

        // Max
        if (parent.max.isDefault) {
            parent.max.value = originator.max.value * multiplier + constant;
        }
        else if (originator.max.isDefault) {
            originator.max.value = (parent.max.value - constant) / multiplier;
        }
    }
    else {
        // Max
        if (parent.max.isDefault) {
            parent.max.value = originator.min.value * multiplier + constant;
        }
        else if (originator.max.isDefault) {
            originator.max.value = (parent.min.value - constant) / multiplier;
        }

        // Min
        if (parent.min.isDefault) {
            parent.min.value = originator.max.value * multiplier + constant;
        }
        else if (originator.min.isDefault) {
            originator.min.value = (parent.max.value - constant) / multiplier;
        }
    }
}

}





