/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCCandleChartFormat.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

/**
 * Class that manages properties and convenience methods specific to candle
 * charts. <P>Candle charts take the <i>y</i> values of <EM>4</EM> data series
 * to  construct each "logical" candle series. The number of data points
 * defines the number of candles in each row.
 * <P><B>Note:</B> The methods in this class use the <EM>logical</EM> candle
 * series index instead of the actual series index, in order to make programming
 * candle styles more convenient.
 */
public class JCCandleChartFormat implements JCChartTypeFormat,
									        java.io.Serializable
{

protected final static int RISING_CANDLE_INDEX  = 1;
protected final static int FALLING_CANDLE_INDEX = 2;
protected final static int CANDLE_OUTLINE_INDEX = 3;

protected boolean        complex = false;
protected ChartDataView  parent = null;

public JCCandleChartFormat() {}

public JCCandleChartFormat(ChartDataView p)
{
    parent = p;
}

public void setParent(ChartDataView cdv)
{
    parent = cdv;
}

/**
 * Gets the <code>Complex</code> property of the candle chart; this
 * property determines whether the candle is a simple or complex candle.
 * @return <code>true</code> if complex.
 */
public boolean isComplex()
{
    return(complex);
}

/**
 * Sets the <code>Complex</code> property of the candle chart; this
 * property determines whether the candle is a simple or
 * complex candle.
 * @param b if <code>true</code>, the chart is a complex candle;
 * otherwise, the chart is a simple candle.
 */
public void setComplex(boolean b)
{
    if (complex == b) {
        return;
    }
    complex = b;
    if (parent != null) {
        parent.setChanged(true, Changeable.RECALC |
                                Changeable.NOTIFY_LEGEND_RECALC);
    }
}

/**
 * Gets the <code>JCChartStyle</code> that controls the candle style of the
 * simple candle or the HiLo line of the complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @return <code>JCChartStyle</code> object.
 */
public JCChartStyle getHiloStyle(int index)
{
    JCChartStyle cs = null;
    if (parent != null) {
        cs = parent.getChartStyle(index * parent.getNumSeriesPerData());
    }
    return(cs);
}

/**
 * Sets the candle style of the simple candle or the HiLo line of the
 * complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @param value <code>JCChartStyle</code> object containing the desired style
 */
public void setHiloStyle(int index, JCChartStyle value)
{
    if (parent != null) {
        parent.setChartStyle(index * parent.getNumSeriesPerData(), value);
    }
}

/**
 * Gets the <code>JCChartStyle</code> that controls the rising candle style of
 * the  complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @return <code>JCChartStyle</code> object.
 */
public JCChartStyle getRisingCandleStyle(int index)
{
    JCChartStyle cs = null;
    if (parent != null) {
        cs = parent.getChartStyle(index * parent.getNumSeriesPerData() +
                                  RISING_CANDLE_INDEX);
    }
    return(cs);
}

/**
 * Sets the candle style of the rising candle style of the
 * complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @param value <code>JCChartStyle</code> object containing the desired style
 */
public void setRisingCandleStyle(int index, JCChartStyle value)
{
    if (parent != null) {
        parent.setChartStyle(index * parent.getNumSeriesPerData() +
                             RISING_CANDLE_INDEX,
                             value);
    }
}

/**
 * Gets the <code>JCChartStyle</code> that controls the falling candle style of
 * the complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @return <code>JCChartStyle</code> object.
 */
public JCChartStyle getFallingCandleStyle(int index)
{
    JCChartStyle cs = null;
    if (parent != null) {
        cs = parent.getChartStyle(index * parent.getNumSeriesPerData() +
                                  FALLING_CANDLE_INDEX);
    }
    return(cs);
}

/**
 * Sets the candle style of the falling candle style of the
 * complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @param value <code>JCChartStyle</code> object containing the desired style
 */
public void setFallingCandleStyle(int index, JCChartStyle value)
{
    if (parent != null) {
        parent.setChartStyle(index * parent.getNumSeriesPerData() +
                             FALLING_CANDLE_INDEX,
                             value);
    }
}

/**
 * Gets the <code>JCChartStyle</code> that controls the candle outline style of
 * the  complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @return <code>JCChartStyle</code> object.
 */
public JCChartStyle getCandleOutlineStyle(int index)
{
    JCChartStyle cs = null;
    if (parent != null) {
        cs = parent.getChartStyle(index * parent.getNumSeriesPerData() +
                                  CANDLE_OUTLINE_INDEX);
    }
    return(cs);
}

/**
 * Sets the candle style of the candle outline style of the
 * complex candle chart.
 * @param index logical candle series (not the actual data series)
 * @param value <code>JCChartStyle</code> object containing the desired style
 */
public void setCandleOutlineStyle(int index, JCChartStyle value)
{
    if (parent != null) {
        parent.setChartStyle(index * parent.getNumSeriesPerData() +
                             CANDLE_OUTLINE_INDEX,
                             value);
    }
}

public ChartDataView getParent()
{
    return(parent);
}

}


