/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartArea.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.legend.JCLegend;
import com.klg.jclass.util.ImageMapInfo;

import java.awt.*;
import java.util.*;
import java.util.List;


 


 


 

/**
 * Class controlling the charting region in the Chart
 * instance. <code>JCChartArea</code> handles rendering
 * of the charts by using {@link JCAxis} and {@link ChartDataView}
 * object instances to determine how data is plotted.
 * A single <code>JCChartArea</code> can plot several
 * <code>ChartDataView</code> objects against a number of axes.
 *
 * <code>JCChartArea</code> is a Swing-based component and should behave accordingly.
 * It is transparent by default, meaning that the <code>Opaque</code> property must
 * be set to <code>true</code> if any background color is to be drawn.
 *
 */
public class JCChartArea extends ChartRegion {

/** Maximum elevation value. */
public final static int MAX_ELEVATION = 45;

/** Maximum rotation value. */
public final static int MAX_ROTATION = 45;

/** Maximum depth value. */
public final static int MAX_DEPTH = 500;

/** Error message for invalid x axis index. */
public final static String ERRMSG_INVALID_XAXIS_INDEX = "Invalid X axis index";
/** Error message for invalid y axis index. */
public final static String ERRMSG_INVALID_YAXIS_INDEX = "Invalid Y axis index";

protected final static int INVALID_INDEX = -1;

protected Vector xaxes = new Vector();
protected Vector yaxes = new Vector();

protected int angleUnit = JCChartUtil.RADIANS;

protected JCAxis horizActionAxis;
protected JCAxis vertActionAxis;

protected boolean axisBoundingBox = false;
protected boolean fastAction = false;
protected boolean inAction = false;
protected int depth = 0;
protected int elevation = 0;
protected int rotation = 0;

/** The image map information for the <code>JCChartArea</code>. */
protected ImageMapInfo imageMapInfo = null;

transient private Hashtable drawables = new Hashtable();

/** Hash table containing axes. */
transient public Hashtable axisHashtable;

private PlotArea plotArea = new PlotArea(this);
protected Rectangle chartAreaRect = new Rectangle(0, 0, 0, 0);
public Rectangle plotRect = new Rectangle(0, 0, 0, 0);

protected PolarAxisParams polarChartParams = new PolarAxisParams();

// Eventual 3D offsets
public int deltaX = 0;
public int deltaY = 0;
public boolean draw3D = false;
public boolean hasBar = false;

protected boolean axisDrawOkay = false;
public boolean pieFirst = false;
public boolean polarFirst = false;
public boolean radarFirst = false;

protected Rectangle backplane;

// the amount to pad around the plot area so that the largest symbol
// will not be clipped by the plot area
protected int symbolPad = 0;

protected transient ActionHandler handler = null;
protected transient boolean saveDBState = false;

/** Temporary Graphics object used to draw without showing to screen.*/
private transient Graphics prevSetGraphics = null;


 

private static boolean errorMessagesPrinted = false;

/**
 * Constructs a chart area for the provided chart.
 */
public JCChartArea()
{
    this(Locale.getDefault());
}

public JCChartArea(Locale locale)
{
    super();
    setOpaque(false);
    setLocale(locale);

    // Add one of each type...
    JCChart chart = (JCChart) getParent();

    JCAxis h = new JCAxis(this, false, JCAxis.VALUE);
    h.setParentChart(chart);
    h.setName(JCChartBundle.string(JCChartBundle.key28, locale));
    JCAxis v = new JCAxis(this, true, JCAxis.VALUE);
    v.setParentChart(chart);
    v.setName(JCChartBundle.string(JCChartBundle.key29, locale));
    xaxes.add(h);
    yaxes.add(v);
    setHorizActionAxis(h);
    setVertActionAxis(v);
}

/**
 * Sets the parent chart for all children axes. This used to be done in
 * <code>addNotify</code>, but the axes (especially in the beans) need to know their
 * parent chart before <code>addNotify()</code> is finally called. Hence, this is now
 * called manually in <code>JCChartArea</code> immediately following an add.
 * @param chart the parent <code>JCChart</code> instance
 */
protected void setParentOnAxes(JCChart chart)
{
    Iterator e = xaxes.iterator();
    while (e.hasNext()) {
        JCAxis axis = (JCAxis)e.next();
        axis.setParentChart(chart);
        if (chart != null) {
            axis.chartArea = chart.getChartArea();
        }
    }
    e = yaxes.iterator();
    while (e.hasNext()) {
        JCAxis axis = (JCAxis)e.next();
        axis.setParentChart(chart);
        if (chart != null) {
            axis.chartArea = chart.getChartArea();
        }
    }
}


/******************
 * Public property accessor methods
 */

/**
 * Overrides <code>setForeground</code> to make sure legend is marked for redraw.
 * (Needed for outline color of swatches, which is derived from
 * <code>ChartArea</code>.)
 * @param c the new foreground color
 */
public void setForeground(Color c)
{
    super.setForeground(c);
    setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Overrides <code>setBackground</code> to make sure entire chart is redrawn if
 * background is set on this subcomponent.<br>
 * Note that the <code>Opaque</code> property needs to be set to <code>true</code> for
 * the background color to be visible.  <code>JCChartArea</code> is transparent by
 * default.
 * @param bg the new background color
 */
public void setBackground(Color bg)
{
    super.setBackground(bg);
    invalidate();
    Container thisParent = getParent();
    if (thisParent != null) {
        thisParent.validate();
    }
}

/**
 * Overrides <code>setOpaque</code> to make sure entire chart is redrawn if
 * opacity is set on this subcomponent.
 * @param op the new value of the <code>opaque</code> property
 */
public void setOpaque(boolean op)
{
    super.setOpaque(op);
    invalidate();
    Container thisParent = getParent();
    if (thisParent != null) {
        thisParent.validate();
    }
}

/**
 * Gets the value of the <code>AngleUnit</code> property. The <code>AngleUnit</code> property
 * determines the unit of all angle values in this chart area --
 * one of <code>DEGREES</code>, <code>RADIANS</code>, or <code>GRADS</code>.
 * @return the current angle unit
 */
public int getAngleUnit()
{
    return(angleUnit);
}

/**
 * Sets the value of the <code>AngleUnit</code> property. The <code>AngleUnit</code> property
 * determines the unit of all angle values in this chart area --
 * one of <code>DEGREES</code>, <code>RADIANS</code>, or <code>GRADS</code>.
 * @param unit unit value
 * @exception IllegalArgumentException If the unit isn't one of
 * <code>DEGREES</code>, <code>RADIANS</code>, or <code>GRADS</code>.
 */
public void setAngleUnit(int unit)
{
    if (!JCChartUtil.validUnit(unit)) {
        throw new IllegalArgumentException("Invalid value for AngleUnit.  " +
                                           "Must be one of DEGREES, RADIANS " +
                                           "or GRADS");
    }
    angleUnit = unit;

    // For Polar charts, the default precision of the theta axis
    // depends on angleUnit.
    Iterator iterator = getDataViews();
    while (iterator.hasNext()) {
        ChartDataView dv = (ChartDataView) iterator.next();
        if (dv != null && dv.getChartType() == JCChart.POLAR) {
            if (dv.getXAxis() != null) {
                dv.getXAxis().updateAnnoDefaultPrecision(JCChart.POLAR);
            }
        }
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>XAxis</code> property at a specified
 * index. The <code>XAxis</code> property is an indexed property, and
 * contains all the <i>x</i> axes for the chart area.
 * @param index index of the <code>JCAxis</code> object
 * @return <code>JCAxis</code> object at the specified index
 */
public JCAxis getXAxis(int index)
{
    if (xaxes == null) {
        return null;
    }
    return((JCAxis)xaxes.get(index));
}

/**
 * Sets the value of the <code>XAxis</code> property at a specified
 * index. The <code>XAxis</code> property is an indexed property, and
 * contains all the <i>x</i> axes for the chart area.
 * @param index index of the <code>JCAxis</code> object
 * @param ax <code>JCAxis</code> object to place at the specified index
 * @exception IllegalArgumentException If the index does not address
 * a valid x axis object.
 */
public void setXAxis(int index, JCAxis ax)
{
    if (ax == null) {
        if (index < 0 || index >= xaxes.size()) {
            throw new IllegalArgumentException(ERRMSG_INVALID_XAXIS_INDEX);
        }
        removeAxis(index, xaxes);
    }
    else {
        if (index < 0 || index > xaxes.size()) {
            throw new IllegalArgumentException(ERRMSG_INVALID_XAXIS_INDEX);
        }
        if (index < xaxes.size()) {
            removeAxis(index, xaxes);
        }
        ax.setParentChart((JCChart) getParent());
        addAxis(index, xaxes, ax);
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Sets the value of the <code>XAxes</code> property.
 * The <code>XAxes</code> property is a collection of all the
 * <i>x</i> axes for the chart area.
 * @param ax collection of <code>JCAxis</code> objects to be used for the <i>x</i> axes
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this field's or method's return type
 * is based on the corresponding <code>com.sun.java.util</code> class.
 */
public void setXAxes(List ax)
{
    xaxes = new Vector();
    if (ax == null) {
        return;
    }
    for (int i = 0; i < ax.size(); i++) {
        Object temp = ax.get(i);
        if (temp == null || !(temp instanceof JCAxis)) {
            continue;
        }
        if (temp != null) {
            ((JCAxis)temp).setParentChart((JCChart)getParent());
            ((JCAxis)temp).i_vertical = false;
        }
        xaxes.add(temp);
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>XAxes</code> property.
 * The <code>XAxes</code> property is a collection of all the
 * <i>x</i> axes for the chart area.
 * @return collection of <code>JCAxis</code> objects to be used for the
 * <i>x</i> axes
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this field's or method's return type
 * is based on the corresponding <code>com.sun.java.util</code> class.
 */
public List getXAxes()
{
    return(xaxes);
}

/**
 * Gets the value of the <code>YAxis</code> property at a specified
 * index. The <code>YAxis</code> property is an indexed property, and
 * contains all the <i>y</i> axes for the chart area.
 * @param index index of the <code>JCAxis</code> object
 * @return <code>JCAxis</code> object at the specified index
 */
public JCAxis getYAxis(int index) {
    if (yaxes == null) {
        return(null);
    }
    return((JCAxis)yaxes.get(index));
}

/**
 * Sets the value of the <code>YAxis</code> property at a specified
 * index. The <code>YAxis</code> property is an indexed property, and
 * contains all the <i>y</i> axes for the chart area.
 * @param index index of the <code>JCAxis</code> object
 * @param ax <code>JCAxis</code> object to place at the specified index
 * @exception IllegalArgumentException If the index does not address
 * a valid y axis object.
 */
public void setYAxis(int index, JCAxis ax)
{
    if (ax == null) {
        if (index < 0 || index >= yaxes.size()) {
            throw new IllegalArgumentException(ERRMSG_INVALID_YAXIS_INDEX);
        }
        removeAxis(index, yaxes);
    }
    else {
        if (index < 0 || index > yaxes.size()) {
            throw new IllegalArgumentException(ERRMSG_INVALID_YAXIS_INDEX);
        }
        if (index < yaxes.size()) {
            removeAxis(index, yaxes);
        }
        ax.setParentChart((JCChart) getParent());
        addAxis(index, yaxes, ax);
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Sets the value of the <code>YAxes</code> property.
 * The <code>YAxes</code> property is a collection of all the
 * <i>y</i> axes for the chart area.
 * @param ay collection of <code>JCAxis</code> objects to be used for the
 * <i>y</i> axes
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this field's or method's return type
 * is based on the corresponding <code>com.sun.java.util</code> class.
 */
public void setYAxes(List ay)
{
    yaxes = new Vector();
    if (ay == null) {
        return;
    }
    for (int i = 0; i < ay.size(); i++) {
        Object temp = ay.get(i);
        if (temp == null || !(temp instanceof JCAxis)) {
            continue;
        }
        if (temp != null) {
            ((JCAxis)temp).setParentChart((JCChart)getParent());
            ((JCAxis)temp).i_vertical = true;
        }
        yaxes.add(temp);
    }
    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>YAxes</code> property.
 * The <code>YAxes</code> property is a collection of all the
 * <i>y</i> axes for the chart area.
 * @return collection of <code>JCAxis</code> objects to be used for the
 * <i>y</i> axes
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this field's or method's return type
 * is based on the corresponding <code>com.sun.java.util</code> class.
 */
public List getYAxes()
{
    return(yaxes);
}


/**
 * Internal method that removes an axis from a particular index.
 * @param index index at which to perform the replacement
 * @param axes array of <code>JCAxis</code> objects
 */
private void removeAxis(int index, List axes)
{
    if (index < axes.size()) {
        axes.remove(index);
    }
}

///**
// * Internal method that replaces the <code>JCAxis</code> object at a particular
// * index with a new (provided) <code>JCAxis</code> object.
// * @param index index at which to perform the replacement
// * @param axes array of <code>JCAxis</code> objects
// * @param ax <code>JCAxis</code> object to place at the specified index
// */
//private void replaceAxis(int index, List axes, JCAxis ax)
//{
//    axes.set(index, ax);
//}

/**
 * Internal method to insert a <code>JCAxis</code> object at a particular
 * index.
 * @param index index at which to perform the insertion
 * @param axes array of <code>JCAxis</code> objects
 * @param ax <code>JCAxis</code> object to insert at the specified index
 */
private void addAxis(int index, List axes, JCAxis ax)
{
    ax.setParentChart((JCChart) getParent());
    axes.add(index, ax);
}

/**
 * Sets the value of the <code>HorizActionAxis</code> property. The
 * <code>HorizActionAxis</code> property determines the axis used for actions
 * (zooming, translating) in the horizontal direction.
 * @param ax <code>JCAxis</code> object to use for chart actions in the
 * horizontal direction
 * @exception IllegalArgumentException If the provided axis is vertical or
 * if it isn't in the list of x or y axes.
 */
public void setHorizActionAxis(JCAxis ax)
{
    if (ax != null && (ax.i_vertical ||
                       (!xaxes.contains(ax) && !yaxes.contains(ax))))
    {
        throw new IllegalArgumentException("Invalid horizontal action axis");
    }
    horizActionAxis = ax;
}

/**
 * Gets the value of the <code>HorizActionAxis</code> property. The
 * <code>HorizActionAxis</code> property determines the axis used for actions
 * (zooming, translating) in the horizontal direction.
 * @return <code>JCAxis</code> object to use for chart actions in the horizontal
 * direction
 */
public JCAxis getHorizActionAxis()
{
    return(horizActionAxis);
}

/**
 * Sets the value of the <code>VertActionAxis</code> property. The
 * <code>VertActionAxis</code> property determines the axis used for actions
 * (zooming, translating) in the vertical direction.
 * @param ax <code>JCAxis</code> object to use for chart actions in the vertical
 * direction
 * @exception IllegalArgumentException If the provided axis is horizontal or
 * if it isn't in the list of x or y axes.
 */
public void setVertActionAxis(JCAxis ax)
{
    if (ax != null && (!ax.i_vertical ||
                       (!xaxes.contains(ax) && !yaxes.contains(ax))))
    {
        throw new IllegalArgumentException("Invalid vertical action axis");
    }
    vertActionAxis = ax;
}

/**
 * Gets the value of the <code>VertActionAxis</code> property. The
 * <code>VertActionAxis</code> property determines the axis used for actions
 * (zooming, translating) in the horizontal direction.
 * @return <code>JCAxis</code> object to use for chart actions in the vertical
 * direction
 */
public JCAxis getVertActionAxis()
{
    return(vertActionAxis);
}

/**
 * Sets the value of the <code>AxisBoundingBox</code> property. The
 * <code>AxisBoundingBox</code> property determines whether a box is drawn
 * around the area bound by the inner axes.
 * @param bb if <code>true</code>, bounding box is drawn; otherwise, box is not
 * drawn
 * @see com.klg.jclass.chart.PlotArea#setForeground
 */
public void setAxisBoundingBox(boolean bb)
{
    if (bb == axisBoundingBox) {
        return;
    }
    axisBoundingBox = bb;

    setChanged(true, Changeable.REDRAW);
}

/**
 * Gets the value of the <code>AxisBoundingBox</code> property. The
 * <code>AxisBoundingBox</code> property determines whether a box is drawn
 * around the area bound by the inner axes.
 * @return if <code>true</code>, bounding box is drawn; otherwise, box is not
 * drawn.
 */
public boolean getAxisBoundingBox()
{
    return(axisBoundingBox);
}

/**
 * Gets the value of the <code>FastAction</code> property.
 * The <code>FastAction</code> property determines whether chart actions will
 * use an optimized mode in which it does not bother to display axis
 * annotations or gridlines.
 * @return <code>true</code> if gridlines and axis annotations will be absent
 * during the action; otherwise, <code>false</code>
 */
public boolean getFastAction()
{
    return(fastAction);
}

/**
 * Sets the value of the <code>FastAction</code> property.
 * The <code>FastAction</code> property determines whether chart actions will
 * use an optimized mode in which it does not bother to display axis
 * annotations or gridlines.
 * @param fr <code>true</code> if gridlines and axis annotations will be absent
 * during the action; otherwise, <code>false</code>
 */
public void setFastAction(boolean fr)
{
    if (fr == fastAction) {
        return;
    }
    fastAction = fr;
}

/**
 * Gets the <code>Depth</code> property, which controls the apparent depth of a
 * graph.
 * @return the depth as a percentage of the width
 */
public int getDepth()
{
    return(depth);
}

/**
 * Sets the <code>Depth</code> property, which controls the apparent depth of a
 * graph.
 * @param newDepth the depth as a percentage of the width; valid values
 * are 0 - 500
 */
public void setDepth(int newDepth)
{
    int tdepth = (int) JCChartUtil.clamp((double) newDepth, 0.0,
                                         (double) MAX_DEPTH);
    if (depth == tdepth) {
        return;
    }
    depth = tdepth;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the <code>Elevation</code> property, which controls distance above
 * the <i>x</i> axis for the 3D effect.
 * @return the number of degrees above the <i>x</i> axis
 */
public int getElevation()
{
    return(elevation);
}

/**
 * Sets the <code>Elevation</code> property, which controls distance above
 * the <i>x</i> axis for the 3D effect.
 * @param newElevation the number of degrees above the <i>x</i> axis;
 * valid values are between -45 and 45
 */
public void setElevation(int newElevation)
{
    int televation = (int) JCChartUtil.clamp((double) newElevation,
                                             (double) -MAX_ELEVATION,
                                             (double) MAX_ELEVATION);
    if (televation == elevation) {
        return;
    }
    elevation = televation;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the <code>Rotation</code> property, which controls the position of
 * the eye relative to the <i>y</i> axis for the 3D effect.
 * @return the number of degrees to the right of the <i>y</i> axis
 */
public int getRotation()
{
    return(rotation);
}

/**
 * Sets the <code>Rotation</code> property, which controls the position of
 * the eye relative to the <i>y</i> axis for the 3D effect.
 * @param newRotation the number of degrees to the right of the Y-axis;
 * valid values are between -45 and 45
 */
public void setRotation(int newRotation)
{
    int trotation = (int) JCChartUtil.clamp((double)newRotation,
                                            (double)-MAX_ROTATION,
                                            (double)MAX_ROTATION);
    if (trotation == rotation) {
        return;
    }
    rotation = trotation;

    setChanged(true, Changeable.RECALC);
}

/**
 * Retrieve the image map information for this <code>JCChartArea</code>.
 * @return The current image map information.
 */
public ImageMapInfo getImageMapInfo()
{
    return(imageMapInfo);
}

/**
 * Set the image map information for this <code>JCChartArea</code>.
 * @param imageMapInfo The new image map information.
 */
public void setImageMapInfo(ImageMapInfo imageMapInfo)
{
    this.imageMapInfo = imageMapInfo;
}

/**
 * Gets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property determines whether the chart area requires
 * a recalculation.
 * @return the current value of the changed flag
 */
public boolean getChanged()
{
    return(super.getChanged() || isDirty());
}

/**
 * Sets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property determines whether the chart area requires
 * a recalculation.
 * @param n boolean value of the <code>Changed</code> property
 * @param cf the changed flag to add in
 */
public void setChanged(boolean n, int cf)
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }

    ensureDrawables();
    for (Enumeration e = drawables.keys(); e.hasMoreElements(); ) {
        Object key = e.nextElement();
        ChartDraw d = (ChartDraw)drawables.get(key);
        d.setChanged(true, cf, false);
        if (!n) {
            d.dataObject.setChanged(false, Changeable.NO_CHANGE, false);
        }
    }

    changed = n;
    if (changed) {
        changedFlag |= cf;
    }
    else {
        changedFlag = cf;
    }

    if (changed) {
        if (!chart.isBatched()) {
            JCLegend legend = chart.getLegend();
            if (legend != null && (cf & Changeable.NOTIFY_LEGEND_RECALC) > 0) {
                legend.setChanged(true, Changeable.RECALC, false);
                legend.recalc();
            }
            if (legend != null && (cf & Changeable.NOTIFY_LEGEND_REDRAW) > 0) {
                legend.setChanged(true, Changeable.REDRAW, false);
            }

            if ((cf & Changeable.RECALC) > 0) {
                recalc();
                chart.repaint();
            }
            else if ((cf & Changeable.REDRAW) > 0) {
                chart.repaint();
            }
        }
        return;
    }

    for (int i = 0; i < xaxes.size(); i++) {
        Object t = xaxes.get(i);
        if (t != null && t instanceof JCAxis) {
            ((JCAxis)t).setChanged(false, Changeable.NO_CHANGE, false);
        }
    }

    for (int i = 0; i < yaxes.size(); i++) {
        Object t = yaxes.get(i);
        if (t != null && t instanceof JCAxis) {
            ((JCAxis)t).setChanged(false, Changeable.NO_CHANGE, false);
        }
    }
}

/**
 * Gets the value of the <code>PlotArea</code> property, which represents the
 * region of the <code>ChartArea</code> that is inside the axes.
 * @return the current plot area
 */
public PlotArea getPlotArea()
{
    if (plotArea == null) {
        plotArea = new PlotArea(this);
    }
    return(plotArea);
}

protected void printErrorMessages(String text1, String text2, String[] messages)
{
    if (!errorMessagesPrinted) {
        if (text1 != null) {
            System.err.println(text1);
        }
        if (text2 != null) {
            System.err.println(text2);
        }
        if (messages != null) {
            for (int i = 0; i < messages.length; i++) {
                if (messages[i] != null) {
                    System.err.println(messages[i]);
                }
            }
        }
        errorMessagesPrinted = true;
    }
}

/**********
 * Non-property public methods.
 */

protected Graphics rbGC = null;

/**
 * Renders the <code>JCChartArea</code> contents.
 * @param gc graphics context used for drawing
 */
public void paintComponent(Graphics gc)
{
    if (getParent() == null || gc == null || !isVisible()) {
        return;
    }
    super.paintComponent(gc);

    JCChart chart = (JCChart)getParent();


 


 

    Graphics translatedGC = gc.create();
    Rectangle rect = getDrawingArea();
    translatedGC.translate(rect.x, rect.y);

    Rectangle defClipRect = translatedGC.getClipBounds();
    Graphics thisGC = translatedGC;

    Rectangle caClipRect = new Rectangle(0, 0, chartAreaRect.width,
                                         chartAreaRect.height);

    try {
        Rectangle tempClipRect = clipForDrawing();

        thisGC.clipRect(tempClipRect.x, tempClipRect.y,
                        tempClipRect.width, tempClipRect.height);
        drawBackplane(thisGC);
        thisGC.setClip(defClipRect.x, defClipRect.y,
                       defClipRect.width, defClipRect.height);


 
/* previous jprobe checking codes
        if (chart != null && !com.klg.jclass.chart.JCVersion.isJProbeRunning()) {
            drawExpired(gc);
            return;
        }
  */

        Rectangle localClipRect = clipToEntirePlotRect();

        // Lots of special rules for 3D drawing to get the layered look
        // just right.  The axes are meant to be sort of transparent.
        // If there is nothing behind them they have a colour, but
        // anything can be seen through them.

        // If a pie is in the first visible data view, draw only the pie!
        if (pieFirst) {
            ensureDrawables();
            Iterator iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                if (dv.getChartType() == JCChart.PIE) {
                    thisGC.clipRect(localClipRect.x, localClipRect.y,
                                    localClipRect.width, localClipRect.height);
                    d.draw(thisGC);
                    thisGC.setClip(defClipRect.x, defClipRect.y,
                                   defClipRect.width, defClipRect.height);
                    // Having drawn the the first pie chart, get out of the
                    // loop.  Otherwise, if another data view of type
                    // JCChart.PIE exists, multiple Pie charts will be
                    // displayed on top of each other.
                    break;
                }
            }
        }

        // If the first visible data view is of type RADAR or AREA_RADAR, don't
        // display any of the other data views.
        else if (radarFirst) {
            ensureDrawables();

            // Draw data views by chart type in order of creation
            // draw non-plot types first -- they should appear behind
            // the axes and annotations
            thisGC.clipRect(localClipRect.x, localClipRect.y,
                            localClipRect.width, localClipRect.height);

            Iterator iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                if (dv.getChartType() == JCChart.RADAR ||
                    dv.getChartType() == JCChart.AREA_RADAR)
                {
                    // Draw the axis grids first for this data view.
                    // Everything else is drawn over top of them.
                    preDrawGrids(thisGC, dv);
                    // Draw the chart.
                    d.draw(thisGC);
                    // Draw the axes for this data view.
                    drawAxes(thisGC, dv);
                    // Draw the annotations and titles of axes
                    // for this data view.
                    if (!fastAction || !inAction) {
                        drawAxisAnnotationsAndTitle(thisGC, dv);
                    }

                    break;
                    // Draw the first visible RADAR or AREA_RADAR
                    // dataview only.
                }
            }
            thisGC.setClip(defClipRect.x, defClipRect.y,
                           defClipRect.width, defClipRect.height);
        }
        // If the first visible data view is of type POLAR, display data
        // views of type POLAR only.
        else if (polarFirst) {
            ensureDrawables();

            thisGC.clipRect(localClipRect.x, localClipRect.y,
                            localClipRect.width, localClipRect.height);

            // Draw the grid lines for ALL Polar charts first.
            // Everything else is drawn over top.
            Iterator iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                if (dv.getChartType() == JCChart.POLAR) {
                    preDrawGrids(thisGC, dv);
                }
            }

            // Draw all POLAR charts in order of creation
            iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                if (dv.getChartType() == JCChart.POLAR) {
                    d.draw(thisGC);
                }
            }

            // Draw all the axes for POLAR data views
            iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                if (dv.getChartType() == JCChart.POLAR) {
                    // Draw the axes for this data view.
                    drawAxes(thisGC, dv);
                    // Draw the annotations and titles of axes
                    // for this data view.
                    if (!fastAction || !inAction) {
                        drawAxisAnnotationsAndTitle(thisGC, dv);
                    }
                }
            }
            thisGC.setClip(defClipRect.x, defClipRect.y,
                            defClipRect.width, defClipRect.height);
        }
        // Otherwise, if we're drawing 3D types, do that
        else if (draw3D) {
            ensureDrawables();

            // Find the front axis.  Since we are drawing in 3D, we
            // must have either a bar or stacking bar.  These chart
            // types must sit on the front axis.
            JCAxis frontXAxis = null;
            JCAxis frontYAxis = null;
            Iterator iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                if (is3DChartType(dv.getChartType())) {
                    frontXAxis = dv.getXAxis();
                    frontYAxis = dv.getYAxis();
                    break;
                }
            }

            if (axisDrawOkay) {
                // draw 3d axis troughs according to delta values
                thisGC.clipRect(localClipRect.x, localClipRect.y,
                                localClipRect.width, localClipRect.height);
                preDraw3DAxes(thisGC, frontXAxis, frontYAxis);
                thisGC.setClip(defClipRect.x, defClipRect.y,
                               defClipRect.width, defClipRect.height);

                // Now that the troughs are ok, draw the lines behind
                // the troughs, in order, to give a 3d appearance.

                // first the backplane outline
                if (axisBoundingBox) {
                    //  Polar/Radar charts are not currently drawn in 3D,
                    //  so no need to deal with circular or web-shaped
                    //  bounding boxes here.
                    thisGC.clipRect(tempClipRect.x, tempClipRect.y,
                                    tempClipRect.width, tempClipRect.height);
                    thisGC.setColor(plotArea.getForeground());
                    thisGC.drawRect(backplane.x, backplane.y,
                                    backplane.width, backplane.height);
                    thisGC.setClip(defClipRect.x, defClipRect.y,
                                defClipRect.width, defClipRect.height);
                }

                // next, the grids on backplane
                thisGC.clipRect(backplane.x + 1, backplane.y + 1,
                                backplane.width - 1, backplane.height - 1);
                if (!fastAction || !inAction) {
                    preDrawGrids(thisGC, xaxes);
                    preDrawGrids(thisGC, yaxes);
                }
                thisGC.setClip(defClipRect.x, defClipRect.y,
                               defClipRect.width, defClipRect.height);

                thisGC.clipRect(localClipRect.x, localClipRect.y,
                                localClipRect.width, localClipRect.height);
                // finally, the axis trough outlines
                if (frontXAxis != null && frontXAxis.visible) {
                    frontXAxis.preDrawOutlineTrough(thisGC);
                }
                if (frontYAxis != null && frontYAxis.visible) {
                    frontYAxis.preDrawOutlineTrough(thisGC);
                }
                thisGC.setClip(defClipRect.x, defClipRect.y,
                               defClipRect.width, defClipRect.height);
            }

            // Draw the backplane axes and annotations
            if (axisDrawOkay) {
                thisGC.clipRect(caClipRect.x, caClipRect.y,
                                caClipRect.width, caClipRect.height);
                drawBpAxes(thisGC, xaxes);
                drawBpAxes(thisGC, yaxes);

                if (!fastAction || !inAction) {
                    drawBpAxisAnnotationsAndTitle(thisGC, xaxes);
                    drawBpAxisAnnotationsAndTitle(thisGC, yaxes);
                }
                thisGC.setClip(defClipRect.x, defClipRect.y,
                               defClipRect.width, defClipRect.height);
            }

            // draw the data views on the back axes first, in
            // order of creation.
            iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                int chartType = dv.getChartType();
                if (chartType == JCChart.PIE ||
                    chartType == JCChart.POLAR ||
                    chartType == JCChart.RADAR ||
                    chartType == JCChart.AREA_RADAR)
                {
                    continue;
                }
                if (!d.drawFront && !dv.is3DChartType()) {
                    // Plots and scatters are clipped differently
                    // than other charts
                    if (chartType == JCChart.PLOT ||
                        chartType == JCChart.SCATTER_PLOT) {
                        thisGC.clipRect(tempClipRect.x, tempClipRect.y,
                                        tempClipRect.width, tempClipRect.height);
                        d.draw(thisGC);
                        thisGC.setClip(defClipRect.x, defClipRect.y,
                                       defClipRect.width, defClipRect.height);
                    }
                    else {
                        thisGC.clipRect(localClipRect.x, localClipRect.y,
                                        localClipRect.width, localClipRect.height);
                        d.draw(thisGC);
                        thisGC.setClip(defClipRect.x, defClipRect.y,
                                       defClipRect.width, defClipRect.height);
                    }
                }
            }

            // Draw bars and other 3d types before axes, plots and
            // other 2d types after axes.

            // Now draw the 3d data views on the front axis in order
            // of creation.
            iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                int chartType = dv.getChartType();
                if (chartType == JCChart.PIE ||
                    chartType == JCChart.POLAR ||
                    chartType == JCChart.RADAR ||
                    chartType == JCChart.AREA_RADAR)
                {
                    continue;
                }
                if (dv.is3DChartType()) {
                    // plots and scatters are clipped differently
                    // than other charts
                    if (chartType == JCChart.PLOT ||
                        chartType == JCChart.SCATTER_PLOT) {
                        thisGC.clipRect(tempClipRect.x, tempClipRect.y,
                                        tempClipRect.width, tempClipRect.height);
                        d.draw(thisGC);
                        thisGC.setClip(defClipRect.x, defClipRect.y,
                                       defClipRect.width, defClipRect.height);
                    }
                    else {
                        thisGC.clipRect(localClipRect.x, localClipRect.y,
                                        localClipRect.width, localClipRect.height);
                        d.draw(thisGC);
                        thisGC.setClip(defClipRect.x, defClipRect.y,
                                       defClipRect.width, defClipRect.height);
                    }
                }
            }

            // Draw Frontplane axes and annotations
            if (axisDrawOkay) {
                thisGC.clipRect(caClipRect.x, caClipRect.y,
                                caClipRect.width, caClipRect.height);
                drawFpAxes(thisGC, xaxes);
                drawFpAxes(thisGC, yaxes);

                if (!fastAction || !inAction) {
                    drawFpAxisAnnotationsAndTitle(thisGC, xaxes);
                    drawFpAxisAnnotationsAndTitle(thisGC, yaxes);
                }
                thisGC.setClip(defClipRect.x, defClipRect.y,
                               defClipRect.width, defClipRect.height);
            }

            // Now draw the 2d data views on the front axis in order
            // of creation.
            iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = getDrawable(dv);
                if (d == null) {
                    continue;
                }
                int chartType = dv.getChartType();
                if (chartType == JCChart.PIE ||
                    chartType == JCChart.POLAR ||
                    chartType == JCChart.RADAR ||
                    chartType == JCChart.AREA_RADAR)
                {
                    continue;
                }
                if (!dv.is3DChartType() && d.drawFront) {
                    // plots and scatters are clipped differently
                    // than other charts
                    if (chartType == JCChart.PLOT ||
                        chartType == JCChart.SCATTER_PLOT)
                    {
                        thisGC.clipRect(tempClipRect.x, tempClipRect.y,
                                        tempClipRect.width, tempClipRect.height);
                        d.draw(thisGC);
                        thisGC.setClip(defClipRect.x, defClipRect.y,
                                       defClipRect.width, defClipRect.height);
                    }
                    else {
                        thisGC.clipRect(localClipRect.x, localClipRect.y,
                                        localClipRect.width, localClipRect.height);
                        d.draw(thisGC);
                        thisGC.setClip(defClipRect.x, defClipRect.y,
                                       defClipRect.width, defClipRect.height);
                    }
                }
            }
        }
        // Normal 2d Drawing -- pretty simple.
        else {
            // Draw axis grids first, everything is over top of them
            if (axisDrawOkay) {
                if (!fastAction || !inAction) {
                    thisGC.clipRect(localClipRect.x, localClipRect.y,
                                    localClipRect.width, localClipRect.height);
                    preDrawGrids(thisGC, xaxes);
                    preDrawGrids(thisGC, yaxes);
                    thisGC.setClip(defClipRect.x, defClipRect.y,
                                   defClipRect.width, defClipRect.height);
                }
            }

            ensureDrawables();
            // Draw data views by chart type in order of creation
            // draw non-plot types first -- they should appear behind
            // the axes and annotations
            thisGC.clipRect(localClipRect.x, localClipRect.y,
                            localClipRect.width, localClipRect.height);
            int k = 0;
            Iterator iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = (ChartDraw) drawables.get(dv);
                if (radarFirst && k>0 && !(d instanceof RadarChartDraw)) {
                    continue;
                }
                k++;
                if (d == null) {
                    continue;
                }
                int chartType = dv.getChartType();
                if (chartType == JCChart.PIE ||
                    chartType == JCChart.POLAR ||
                    chartType == JCChart.RADAR ||
                    chartType == JCChart.AREA_RADAR)
                {
                    continue;
                }
                if (!(chartType == JCChart.PLOT ||
                      chartType == JCChart.SCATTER_PLOT))
                {
                    d.draw(thisGC);
                }
            }
            thisGC.setClip(defClipRect.x, defClipRect.y,
                           defClipRect.width, defClipRect.height);

            // Draw Axes!
            if (axisDrawOkay) {
                thisGC.clipRect(caClipRect.x, caClipRect.y,
                                caClipRect.width, caClipRect.height);
                drawAxes(thisGC, xaxes);
                drawAxes(thisGC, yaxes);

                if (!fastAction || !inAction) {
                    drawAxisAnnotationsAndTitle(thisGC, xaxes);
                    drawAxisAnnotationsAndTitle(thisGC, yaxes);
                }
                thisGC.setClip(defClipRect.x, defClipRect.y,
                               defClipRect.width, defClipRect.height);
            }

            // draw plot types first -- they should appear over top of
            // the axes and annotations
            thisGC.clipRect(tempClipRect.x, tempClipRect.y,
                            tempClipRect.width, tempClipRect.height);
            k = 0;
            iterator = chart.getOriginalDataView().iterator();
            while (iterator.hasNext()) {
                ChartDataView dv = (ChartDataView) iterator.next();
                ChartDraw d = (ChartDraw) drawables.get(dv);
                if (radarFirst && k > 0 && !(d instanceof RadarChartDraw)) {
                    continue;
                }
                k++;
                if (d == null) {
                    continue;
                }
                int chartType = dv.getChartType();
                if (chartType == JCChart.PIE ||
                    chartType == JCChart.POLAR ||
                    chartType == JCChart.RADAR ||
                    chartType == JCChart.AREA_RADAR)
                {
                    continue;
                }
                // Plots and scatters are clipped differently than other charts
                if (chartType == JCChart.PLOT ||
                    chartType == JCChart.SCATTER_PLOT)
                {
                    d.draw(thisGC);
                }
            }
            thisGC.setClip(defClipRect.x, defClipRect.y,
                           defClipRect.width, defClipRect.height);
        }


 

 

 

 

 

 

 

 

 

 

 

 

 

 

 
    }
    catch(Exception e) {
        e.printStackTrace(System.out);
    }
    if (translatedGC != null) {
        translatedGC.dispose();
    }

    rbGC = gc;
    if (handler != null) {
        handler.reanimate(last.x, last.y);
    }
    rbGC = null;

    // Now that we've drawn the chart area, the drawables have the
    // appropriate pixel values such that we can recalc the chart labels
    JCChart cht = (JCChart) getParent();
    if (cht.hasChartLabelManager()) {
        cht.getChartLabelManager().recalcChangedChartLabels();
    }

    setChanged(false, Changeable.NO_CHANGE);
}


 


 

protected Point last = new Point(0, 0);

void showAction(int x, int y)
{
    if (last.x == x && last.y == y) {
        return;
    }
    last.x = x;
    last.y = y;
    handler.animate(x, y);
}

void endAction(int x, int y)
{
    if (handler == null) {
        return;
    }
    ActionHandler tmp = handler;
    handler = null;
    //setDoubleBuffer(saveDBState);
    tmp.end(x, y);
}

void startAction(EventTrigger et, int x, int y)
{
    handler = ActionHandler.makeHandler(et.getAction(), this);
    if (handler != null) {
        handler.start(x, y);
    }
}


/**
 * Returns the preferred size for the chart area.
 * @return the current preferred size
 */
public Dimension getPreferredSize()
{

    // Call to ensure there's a drawable when legend preferred size is
    // calculated
    ensureDrawables();

    Dimension ps = super.getPreferredSize();
    if (ps == null) {
        ps = new Dimension(0, 0);
    }

 
/* previous jprobe checking codes
    if (chart != null && !com.klg.jclass.chart.JCVersion.isJProbeRunning()) {
        if (msg == null) {
            makeExpiredMessage();
        }
        Dimension sz = msg.getSize(chart, getFont());
        ps.width = sz.width;
        ps.height = sz.height;
    }
    else
 */
    if (ps.width <= 0 || ps.height <= 0) {
        ps.width = 100;
        ps.height = 100;
    }
    return(ps);
}

/**
 * Returns the minimum size for the chart area.
 * @return the current minimum size
 */
public Dimension getMinimumSize()
{
    return(getPreferredSize());
}

/**
 * Can't have inverted axes for Polar/Radar charts.  If
 * the chart type of the first data view is Polar/Radar,
 * set the i_vertical flags of the axes to appropriate values.
 */
protected void checkAxisVertical()
{
    JCChart chart = (JCChart)getParent();
    ChartDataView firstDataView = chart.getDataView(0);
    JCAxis xaxis = firstDataView.getXAxis();
    JCAxis yaxis = firstDataView.getYAxis();

    int chartType = firstDataView.getChartType();
    if (chartType == JCChart.POLAR ||
        chartType == JCChart.RADAR ||
        chartType == JCChart.AREA_RADAR)
    {
        xaxis.setVerticalInternal(false);
        yaxis.setVerticalInternal(true);
    }
    else {
        xaxis.setVerticalInternal(xaxis.vertical);
        yaxis.setVerticalInternal(yaxis.vertical);
    }
}

/**
 * If necessary, forces recalculation of the chart area.
 */
public void recalc()
{
    if (getParent() == null) {
        return;
    }
    if (!isChanged() && isValid() && !isPolarRadar()) {
        //  Get chart area layout problems for Polar/Radar
        //  if we exit here.
        return;
    }

    // Need to call ensureDrawables() here to ensure drawables have been
    // created so that legend recalc() will be able to get transient data.
    ensureDrawables();

    if (getSize().width <= 0 || getSize().height <= 0) {
        return;
    }

    try {
        resetPlotArea();
        // Can't have inverted axes for Polar/Radar charts.  If
        // the chart type of the first data view is Polar/Radar,
        // set the i_vertical flags of the axes to appropriate values.
        checkAxisVertical();
        calc3D();
        calcLimits();

        // When axis annotationRotation is JCAxis.ROTATE_OTHER
        // and the annotations are provided by the user,
        // we use ChartText width and height values in the
        // extents calculations.  So we have to make the JCValueLabels
        // before calling calcAxisParams().
        if (userProvidedAnnotations(xaxes) &&
            rotatedAxisAnnotations(xaxes)) {
            makeAnnotations(xaxes);
        }
        if (userProvidedAnnotations(yaxes) &&
            rotatedAxisAnnotations(yaxes)) {
            makeAnnotations(yaxes);
        }

        calcAxisParams(xaxes);
        calcAxisParams(yaxes);

        updateAnnotationsAndGrids(xaxes);
        updateAnnotationsAndGrids(yaxes);

        if (!inAction || !fastAction) {
            calcAnchors(xaxes);
            calcAnchors(yaxes);

            makeAnnotations(xaxes);
            makeAnnotations(yaxes);

            layoutAnnotations(xaxes);
            layoutAnnotations(yaxes);

            calcAnnotationExtents(xaxes);
            calcAnnotationExtents(yaxes);

            updateAnnotationsAndGrids(xaxes);
            updateAnnotationsAndGrids(yaxes);
        }
    }
    catch (Exception e) {
        e.printStackTrace(System.out);
    }
    layoutChartArea();
    if (checkDeltaValues()) {
        // once more if delta values got changed
        layoutChartArea();
    }
    calcGraphExtents();
    if (!inAction || !fastAction) {
        // Not quite sure this is needed, but it might be for cases where
        // the axis parameters have changed.
        JCChart chart = (JCChart) getParent();
        if (chart.hasChartLabelManager()) {
            chart.getChartLabelManager().markAllChartLabelsForRecalc();
        }
    }
    setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Positions axes and calculates the true plot area size based on
 * the position of the axes.
 */
public void layoutChartArea()
{
    resetPlotArea();

    // the symbol pad is half of the biggest symbol size rounded up
    symbolPad = (getBiggestSymbolSize() + 1) / 2;

    int plotWidth  = (chartAreaRect.width - plotArea.left.value -
                      plotArea.right.value);
    int plotHeight = (chartAreaRect.height - plotArea.top.value -
                      plotArea.bottom.value);
    plotRect = new Rectangle(plotArea.left.value, plotArea.top.value,
                             plotWidth, plotHeight);
    boolean fPolarRadar = isPolarRadar();

    if (!pieFirst) {  // Don't need to calc axis stuff if we have a pie chart

        List vertAxes = new Vector();
        List horizAxes = new Vector();
        getHorizVertAxes(horizAxes, vertAxes);

        if (!fPolarRadar) {
            adjustPlotRectForAxes(xaxes);
            adjustPlotRectForAxes(yaxes);
        }

        // Initial layout
        layoutAxes(xaxes, false);
        layoutAxes(yaxes, false);

        if (!fPolarRadar) {
            // Adjust for extents
            adjustForExtents(horizAxes, vertAxes);
            // Adjust for symbol size
            adjustForSymbols(horizAxes, vertAxes);

            if (rotatedAxisAnnotations(xaxes) ||
                rotatedAxisAnnotations(yaxes)) {
                for (int i=0; i < 3; i++) {
                    // Recalculate extents for adjusted axes.
                    calcAnnotationExtents(xaxes);
                    calcAnnotationExtents(yaxes);
                    adjustForExtents(horizAxes, vertAxes);
                }
            }

            // Re-layout as a result of extent changes
            layoutAxes(xaxes, true);
            layoutAxes(yaxes, true);

            // Account for origin-oriented axes
            layoutOrigins(xaxes);
            layoutOrigins(yaxes);
        }

        if (fPolarRadar) {
            ChartDataView dv1 = getFirstVisibleDataView();
            if (dv1 != null) {
                Deltas chartDeltas = new Deltas();
                getPolarChartDeltas(dv1, chartDeltas);
                adjustForPolarAnnoExtents(dv1.getYAxis(), chartDeltas);
                // Call adjustForPolarAnnoExtents() for the 1st data view
                // only, even when multiple data views exist.  The axes
                // in all of the data views share a common PolarAxisParams
                // object, and we don't want to apply radius and center
                // coordinate deltas multiple times.
            }
            if (isPolar()) {
                // We support multiple data views for Polar charts.
                // Update the scaling factor for the y axis of
                // each data view.
                Iterator dataViews = getDataViews();
                while (dataViews.hasNext()) {
                    ChartDataView dv = (ChartDataView)dataViews.next();
                    if (dv != dv1) {
                        dv.getYAxis().getAxisHandler().calcTransformation();
                    }
                }
            }
        }
        else {
            // Handle ortho extents on origin-anchored axes.
            adjustForOrthoExtents(horizAxes, vertAxes);
        }

        if (!inAction || !fastAction) {
            layoutAnnotations(xaxes);
            layoutAnnotations(yaxes);

            checkAnnotations(xaxes);
            checkAnnotations(yaxes);

            updateAnnotationsAndGrids(xaxes);
            updateAnnotationsAndGrids(yaxes);
        }
    }

    if (plotArea.top.isDefault) {
        plotArea.top.value = plotRect.y;
    }
    if (plotArea.left.isDefault) {
        plotArea.left.value = plotRect.x;
    }
    if (plotArea.bottom.isDefault) {
        plotArea.bottom.value = (chartAreaRect.height - plotRect.y -
                                 plotRect.height);
    }
    if (plotArea.right.isDefault) {
        plotArea.right.value = (chartAreaRect.width - plotRect.x -
                                plotRect.width);
    }
}

private boolean userProvidedAnnotations(List axes)
{
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.getAnnotationMethod() == JCAxis.VALUE_LABELS &&
            ax.annoHandler.userValueLabels != null &&
            ax.annoHandler.userValueLabels.size() > 0)
        {
            return(true);
        }
    }
    return(false);
}

/**
 * Returns <code>true</code> if any of the axes in the specified list
 * has annotations rotated at angles other than 0, 90, 180,
 * or 270 degrees.
 * @param axes the list of axes to check
 * @return whether any of the axes have angled annotation
 */
private boolean rotatedAxisAnnotations(List axes)
{
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.annotationRotation == JCAxis.ROTATE_OTHER) {
            return(true);
        }
    }
    return(false);
}

/**
 * Applies delta values to the center coordinates and radius of a
 * Polar or Radar chart.<br>
 * Called in <code>layoutChartArea()</code> to keep the axis annotations within view.
 *
 * @param ax the axis to which this applies
 * @param chartDeltas object containing delta values to be applied
 */
protected void adjustForPolarAnnoExtents(JCAxis ax, Deltas chartDeltas)
{
    ax.adjustCenterAndRadius(chartDeltas.dx,
                             chartDeltas.dy,
                             chartDeltas.dWidth);
                             //  dWidth=dHeight for Polar/Radar
}

protected Iterator getDataViews()
{
    Iterator iterator = null;
    JCChart chart = (JCChart) getParent();
    if (chart != null) {
        iterator = chart.getOriginalDataView().iterator();
    }
    return(iterator);
}

/**
 * This method returns the first visible data view.  If none exist,
 * the method returns null.
 * @return the first visible data view
 */
protected ChartDataView getFirstVisibleDataView()
{
    ChartDataView cdv = null;
    JCChart chart = (JCChart) getParent();
    if (chart != null) {
        Iterator dataViews = chart.getOriginalDataView().iterator();
        if (dataViews != null) {
            while (dataViews.hasNext()) {
                cdv = (ChartDataView)dataViews.next();
                if (cdv != null && cdv.isVisible()) {
                    return(cdv);
                }
            }
        }
    }
    return(null);
}

protected ChartDataView getDataView(int index)
{
    ChartDataView cdv = null;
    JCChart chart = (JCChart) getParent();
    if (chart != null) {
        List list = chart.getOriginalDataView();
        // Make sure the specified index is within range.
        if (list != null && index >=0 && index < list.size()) {
            cdv = (ChartDataView)list.get(index);
        }
    }
    return(cdv);
}

protected boolean isPolar()
{
    boolean fPolar = false;
    ChartDataView dv = getFirstVisibleDataView();
    if (dv != null) {
        fPolar = (dv.chartType == JCChart.POLAR);
    }
    return fPolar;
}

protected boolean isRadar()
{
    boolean fRadar = false;
    ChartDataView dv = getFirstVisibleDataView();
    if (dv != null) {
        fRadar = (dv.chartType == JCChart.RADAR ||
                    dv.chartType == JCChart.AREA_RADAR);
    }
    return(fRadar);
}

protected boolean isPolarRadar()
{
    return (isPolar() || isRadar());
}

/**
 * Finds the index of array member that matches the specified value.
 * Returns <code>INVALID_INDEX</code> if a matching value isn't found.
 */
private int getIndexOfMatchingValue(double[] arrayValues, double value,
                                    int startPoint, int lastPoint)
{
    boolean fMatchingValueFound = false;
    int i = 0;
    for (i = startPoint; i <= lastPoint; i++) {
        if (arrayValues[i] == value) {
            fMatchingValueFound = true;
            break;
        }
    }
    return (fMatchingValueFound ? i : INVALID_INDEX);
}

/**
 * Calculates the angle of a theta axis label in a Radar chart.
 *
 * @param values array of angle values
 * @param valueLabel the label for which the radar angle is to be calculated
 * @param isAxisReversed <code>true</code> if the direction of the theta axis is reversed
 * @param startPoint the index of the first plotted point
 * @param lastPoint the index of the last plotted point
 * @return the calculated angle of the theta axis label
 */
private double getRadarLabelAngle(double[] values, JCValueLabel valueLabel,
                                  boolean isAxisReversed, int startPoint,
                                  int lastPoint)
{
    int index = getIndexOfMatchingValue(values, valueLabel.value,
                                        startPoint, lastPoint);
    if (index == INVALID_INDEX) {
        return 0.;
    }
    int nPoints = lastPoint - startPoint + 1;
    return(JCChartUtil.getRadarAngle(JCChartUtil.RADIANS, nPoints,
                                     index - startPoint, isAxisReversed));
}

/**
 * Computes size and position offsets needed to keep the axis annotations of
 * a Polar or Radar chart in view.
 *
 * @param dv a <code>ChartDataView</code> object
 * @param chartDeltas object to contain the x, y, and radius delta values
 */
protected void getPolarChartDeltas(ChartDataView dv, Deltas chartDeltas)
{
    JCAxis thetaAxis = dv.getXAxis();

    AnnoBoundsInfo annoBoundsTheta = thetaAxis.getPolarAnnotationBounds();
    //  As in XRT chart, we will concern ourselves with theta axis
    //  annotation only.

    if (annoBoundsTheta == null) {
        chartDeltas.dx = 0;
        chartDeltas.dy = 0;
        chartDeltas.dWidth = 0;
        chartDeltas.dHeight = 0;
        return;
    }

    JCAxis rAxis = dv.getYAxis();
    final double MIN_RADIUS = 1.0;

    //  When the window is created, or resized, the radius is set to
    //  min(chartAreaRect.width, chartAreaRect.height)/2.0.  If the radius has
    //  the minimum value (as the result of a previous call to
    //  getPolarChartDeltas), don't recalculate the radius.  Just
    //  compute offsets to center the bounding rectangle for the chart
    //  annotations (i.e., we simply accept that fact that clipping is
    //  going to occur).

    if (rAxis.polarAxisParams.radius <= MIN_RADIUS && annoBoundsTheta.x > 0) {
        int dWidth = (annoBoundsTheta.width - chartAreaRect.width)/2;
        int dHeight = (annoBoundsTheta.height - chartAreaRect.height)/2;

        chartDeltas.dx = - annoBoundsTheta.x - dWidth;
        chartDeltas.dy = - annoBoundsTheta.y - dHeight;
        chartDeltas.dWidth = 0;
        chartDeltas.dHeight = 0;
        return;
    }

    //  Compute the angles and angle ratios of the labels that define the
    //  annotation bounds:
    boolean fRadar = (dv.chartType == JCChart.RADAR ||
                      dv.chartType == JCChart.AREA_RADAR);
    double angleMaxX = 0.;   //  angle of axis label having right-most extent
    double angleMinX = 0.;   //  angle of axis label having left-most extent
    double angleMaxY = 0.;   //  angle of axis label having top-most extent
    double angleMinY = 0.;   //  angle of axis label having bottom-most extent
    double angleOriginBaseMaxX = 0.;   //  angleMaxX + originBase
    double angleOriginBaseMinX = 0.;   //  angleMinX + originBase
    double angleOriginBaseMaxY = 0.;   //  angleMaxY + originBase
    double angleOriginBaseMinY = 0.;   //  angleMinY + originBase
    double xvals[] = null;
    if (fRadar && (thetaAxis.getAnnotationMethod() != JCAxis.VALUE)) {
        //  If annotation method for RADAR chart is VALUE_LABEL
        List seriesList = dv.getSeries();
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(0);
        xvals = series.getX();

        int[] limits = ChartDataView.getFirstLastPoints(dv);
        int startPoint = limits[0];
        int lastPoint = limits[1];

        //  For Radar chart VALUE_LABELs, need to compute the radar angle
        //  (ignore the theta value associated with the label):
        angleMaxX = getRadarLabelAngle(xvals,
                        annoBoundsTheta.valueLabelMaxBoundsX,
                        thetaAxis.isReversed(), startPoint, lastPoint);
        angleMaxY = getRadarLabelAngle(xvals,
                        annoBoundsTheta.valueLabelMaxBoundsY,
                        thetaAxis.isReversed(), startPoint, lastPoint);
        angleMinX = getRadarLabelAngle(xvals,
                        annoBoundsTheta.valueLabelMinBoundsX,
                        thetaAxis.isReversed(), startPoint, lastPoint);
        angleMinY = getRadarLabelAngle(xvals,
                        annoBoundsTheta.valueLabelMinBoundsY,
                        thetaAxis.isReversed(), startPoint, lastPoint);

//      System.out.println("angleMaxX=" + angleMaxX +
//                      ", angleMaxY=" + angleMaxY +
//                      ", angleMinX=" + angleMinX +
//                      ", angleMinY=" + angleMinY);
    }
    else {   //  POLAR
        int direction = thetaAxis.isReversed() ? -1 : 1 ;
        angleMaxX = JCChartUtil.convertAngle(getAngleUnit(),
                                JCChartUtil.RADIANS,
                                direction *
                                annoBoundsTheta.valueLabelMaxBoundsX.value);
        angleMaxY = JCChartUtil.convertAngle(getAngleUnit(),
                                JCChartUtil.RADIANS,
                                direction *
                                annoBoundsTheta.valueLabelMaxBoundsY.value);
        angleMinX = JCChartUtil.convertAngle(getAngleUnit(),
                                JCChartUtil.RADIANS,
                                direction *
                                annoBoundsTheta.valueLabelMinBoundsX.value);
        angleMinY = JCChartUtil.convertAngle(getAngleUnit(),
                                JCChartUtil.RADIANS,
                                direction *
                                annoBoundsTheta.valueLabelMinBoundsY.value);
    }

    angleOriginBaseMaxX = angleMaxX + rAxis.polarAxisParams.originBase;
    angleOriginBaseMaxY = angleMaxY + rAxis.polarAxisParams.originBase;
    angleOriginBaseMinX = angleMinX + rAxis.polarAxisParams.originBase;
    angleOriginBaseMinY = angleMinY + rAxis.polarAxisParams.originBase;

    double angleRatioMaxX = JCChartUtil.getAngleRatio(JCChartUtil.RADIANS,
                                        angleMaxX,
                                        thetaAxis.polarAxisParams.originBase,
                                        thetaAxis.isReversed());

    double angleRatioMinX = JCChartUtil.getAngleRatio(JCChartUtil.RADIANS,
                                        angleMinX,
                                        thetaAxis.polarAxisParams.originBase,
                                        thetaAxis.isReversed());

    double angleRatioMaxY = JCChartUtil.getAngleRatio(JCChartUtil.RADIANS,
                                        angleMaxY,
                                        thetaAxis.polarAxisParams.originBase,
                                        thetaAxis.isReversed());

    double angleRatioMinY = JCChartUtil.getAngleRatio(JCChartUtil.RADIANS,
                                        angleMinY,
                                        thetaAxis.polarAxisParams.originBase,
                                        thetaAxis.isReversed());

    double maxExtentX =
            (double)(annoBoundsTheta.valueLabelMaxBoundsX.label.left.value +
            annoBoundsTheta.valueLabelMaxBoundsX.label.width.value);
    double minExtentX =
            (double)(annoBoundsTheta.valueLabelMinBoundsX.label.left.value);
    double maxExtentY =
            (double)(annoBoundsTheta.valueLabelMaxBoundsY.label.top.value +
            annoBoundsTheta.valueLabelMaxBoundsY.label.height.value);
    double minExtentY =
            (double)(annoBoundsTheta.valueLabelMinBoundsY.label.top.value);

    if (!isWebShaped(rAxis)) {
        //  If not web shaped RADAR chart...
        maxExtentX = Math.max(maxExtentX,
                rAxis.polarAxisParams.xCenter + rAxis.polarAxisParams.radius);

        minExtentX = Math.min(minExtentX,
                rAxis.polarAxisParams.xCenter - rAxis.polarAxisParams.radius);

        maxExtentY = Math.max(maxExtentY,
                rAxis.polarAxisParams.yCenter + rAxis.polarAxisParams.radius);

        minExtentY = Math.min(minExtentY,
                rAxis.polarAxisParams.yCenter - rAxis.polarAxisParams.radius);
    }

    double dWidth = (maxExtentX - minExtentX) - chartAreaRect.width;
    double dHeight = (maxExtentY - minExtentY) - chartAreaRect.height;

    double newRadiusX = MIN_RADIUS;
    double newRadiusY = MIN_RADIUS;

    //  Compute a new radius based on maximum and minimum X bounds:
    double dxMax = 0.;
    double thetaMaxX = 0.;
    if (isWebShaped(rAxis) ||
        (maxExtentX > (rAxis.polarAxisParams.xCenter +
                            rAxis.polarAxisParams.radius))) {
        if (angleRatioMaxX < 0.25) {
            thetaMaxX = angleOriginBaseMaxX;
            dxMax = Math.cos(thetaMaxX) * rAxis.polarAxisParams.radius;
        }
        else if (angleRatioMaxX > 0.75) {
            thetaMaxX = JCChartUtil.M_2PI - angleOriginBaseMaxX;
            dxMax = Math.cos(thetaMaxX) * rAxis.polarAxisParams.radius;
        }
        else {
            //  90 or 270 degrees.  Leave dxMax = 0.
        }
    }
    else {
        //  thetaMaxX remains 0.  Math.cos(thetaMaxX) will be 1.0
        dxMax = rAxis.polarAxisParams.radius;
    }

    double dxMin = 0.;
    double thetaMinX = 0.;
    if (isWebShaped(rAxis) ||
        (minExtentX < (rAxis.polarAxisParams.xCenter -
                            rAxis.polarAxisParams.radius))) {
        if (angleRatioMinX < 0.5 && angleRatioMinX > 0.25) {
            thetaMinX = JCChartUtil.M_PI - angleOriginBaseMinX;
            dxMin = Math.cos(thetaMinX) * rAxis.polarAxisParams.radius;
        }
        else if (angleRatioMinX >= 0.5 && angleRatioMinX < 0.75) {
            thetaMinX = angleOriginBaseMinX - JCChartUtil.M_PI;
            dxMin = Math.cos(thetaMinX) * rAxis.polarAxisParams.radius;
        }
        else {
            //  90 or 270 degrees.  Leave dxMin = 0.
        }
    }
    else {
        //  thetaMinX remains 0.  Math.cos(thetaMinX) will be 1.0
        dxMin = rAxis.polarAxisParams.radius;
    }

    double denominator = dxMin + dxMax;
    double dRadius = 0.;
    if (dxMax > dxMin) {
        double dWidthFraction = dxMax / denominator;
        double cosineAngle = Math.cos(thetaMaxX);
        dRadius = dWidth * dWidthFraction;
        if (cosineAngle != 0.0) {
            dRadius = dRadius / cosineAngle;
        }
//      System.out.println("dxMin=" + dxMin + ", dxMax=" + dxMax +
//                      ", dWidthFraction=" + dWidthFraction +
//                      ", cosineAngle=" + cosineAngle + ", dWidth=" +
//                      dWidth + ", dRadius=" + dRadius);
    }
    else if (dxMin > dxMax) {
        double dWidthFraction = dxMin / denominator;
        double cosineAngle = Math.cos(thetaMinX);
        dRadius = dWidth * dWidthFraction;
        if (cosineAngle != 0.0) {
            dRadius = dRadius / cosineAngle;
        }
    }
    else {
        if (thetaMaxX != JCChartUtil.M_PI_2 &&
            thetaMaxX != (JCChartUtil.M_PI_2 * 3)) {
            dRadius = dWidth * 0.5 / Math.cos(thetaMaxX);
        }
        else {
            //  Labels with Max and Min x coordinates are at
            //  90 and/or 270 degrees.
            //  Decrease the radius by an arbitrary 10%
            dRadius = rAxis.polarAxisParams.radius * 0.1;
        }
    }
    newRadiusX = rAxis.polarAxisParams.radius - dRadius;

    //  Compute a new radius based on maximum and minimum Y bounds:
    double dyMax = 0.;
    double thetaMaxY = 0.;
    if (isWebShaped(rAxis) ||
        (maxExtentY > (rAxis.polarAxisParams.yCenter +
                            rAxis.polarAxisParams.radius)))
    {
        if (angleRatioMaxY > 0.75) {
            thetaMaxY = JCChartUtil.M_2PI - angleOriginBaseMaxY;
            dyMax = Math.sin(thetaMaxY) * rAxis.polarAxisParams.radius;
        }
        else if (angleRatioMaxY < 0.75 && angleRatioMaxY >= JCChartUtil.M_PI) {
            thetaMaxY = angleOriginBaseMaxY - JCChartUtil.M_PI;
            dyMax = Math.sin(thetaMaxY) * rAxis.polarAxisParams.radius;
        }
        else {   //  angleRatioMaxY == 0.75
            thetaMaxY = JCChartUtil.M_PI_2;  // NOT 3 * M_PI_2
            dyMax = rAxis.polarAxisParams.radius;
        }
    }
    else {
        thetaMaxY = JCChartUtil.M_PI_2;  // NOT 3 * M_PI_2
        dyMax = rAxis.polarAxisParams.radius;
    }

    double dyMin = 0.;
    double thetaMinY = 0.;

    if (isWebShaped(rAxis) ||
        (minExtentY < (rAxis.polarAxisParams.yCenter -
                            rAxis.polarAxisParams.radius)))
    {
        if (angleRatioMinY < 0.25) {
            thetaMinY = angleOriginBaseMinY;
            dyMin = Math.sin(thetaMinY) * rAxis.polarAxisParams.radius;
        }
        else if (angleRatioMinY <= 0.5 && angleRatioMinY > 0.25) {
            thetaMinY = JCChartUtil.M_PI - angleOriginBaseMinY;
            dyMin = Math.sin(thetaMinY) * rAxis.polarAxisParams.radius;
        }
        else {   //  angleRatioMinY == 0.25
            thetaMinY = JCChartUtil.M_PI_2;
            dyMin = rAxis.polarAxisParams.radius;
        }
    }
    else {
        thetaMinY = JCChartUtil.M_PI_2;
        dyMin = rAxis.polarAxisParams.radius;
    }

    denominator = dyMin + dyMax;
    if (dyMax > dyMin) {
        double dHeightFraction = dyMax / denominator;
        double sineAngle = Math.sin(thetaMaxY);
        dRadius = dHeight * dHeightFraction;
        if (sineAngle != 0.0) {
            dRadius = dRadius / sineAngle;
        }
    }
    else if (dyMin > dyMax) {
        double dHeightFraction = dyMin / denominator;
        double sineAngle = Math.sin(thetaMinY);
        dRadius = dHeight * dHeightFraction;
        if (sineAngle != 0.0) {
            dRadius = dRadius / sineAngle;
        }
    }
    else {
        if (thetaMaxY != 0 && thetaMaxY != JCChartUtil.M_PI) {
            dRadius = dHeight * 0.5 / Math.sin(thetaMaxY);
        }
        else {
            //  Labels with Max and Min y coordinates are at
            //  0 and/or 180 degrees.
            //  Decrease the radius by the text height.
            dRadius = annoBoundsTheta.valueLabelMaxBoundsY.label.height.value;
        }
    }
    newRadiusY = rAxis.polarAxisParams.radius - dRadius;

    double newRadius = Math.min(newRadiusX, newRadiusY);
    if (!isWebShaped(rAxis)) {
        //  If not a web shaped RADAR chart...
        //  Make sure theta axis or circular grid line isn't clipped.  This may
        //  happen when there are very few theta axis labels.
        newRadius = Math.min(newRadius, rAxis.polarAxisParams.radius * 0.95);
    }
    newRadius = Math.max(newRadius, 1.0);
    //  Allow only positive values for the radius

    Point pntMaxX = null;
    Point pntMaxY = null;
    Point pntMinX = null;
    Point pntMinY = null;

    if (fRadar) {
        //  toPixelPolar() handles theta axis reversal.  We need to
        //  "undo" angle reversal by the getRadarAngle() method.
        boolean fReversed = thetaAxis.isReversed();
        double radarAngleMinX = (fReversed) ? -angleMinX : angleMinX;
        double radarAngleMinY = (fReversed) ? -angleMinY : angleMinY;
        double radarAngleMaxX = (fReversed) ? -angleMaxX : angleMaxX;
        double radarAngleMaxY = (fReversed) ? -angleMaxY : angleMaxY;
        pntMinX = rAxis.toPixelPolar(radarAngleMinX,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
        pntMinY = rAxis.toPixelPolar(radarAngleMinY,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
        pntMaxX = rAxis.toPixelPolar(radarAngleMaxX,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
        pntMaxY = rAxis.toPixelPolar(radarAngleMaxY,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
    }
    else {   //  POLAR
        pntMinX = rAxis.toPixelPolar(angleMinX,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
        pntMinY = rAxis.toPixelPolar(angleMinY,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
        pntMaxX = rAxis.toPixelPolar(angleMaxX,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
        pntMaxY = rAxis.toPixelPolar(angleMaxY,
                    (int)newRadius +
                    thetaAxis.getThetaAxisLabelOffset(), false);
    }

    //  Need to compute new annotation extents in order to calculate margins:
    double newMaxExtentX = Double.MIN_VALUE;
    double newMinExtentX = Double.MAX_VALUE;
    double newMaxExtentY = Double.MIN_VALUE;
    double newMinExtentY = Double.MAX_VALUE;

    double newMaxExtentXRadius = 0.;
    double newMinExtentXRadius = 0.;
    double newMaxExtentYRadius = 0.;
    double newMinExtentYRadius = 0.;

    if (isWebShaped(rAxis) ||
        (maxExtentX > (rAxis.polarAxisParams.xCenter +
                                rAxis.polarAxisParams.radius)))
    {
        if (angleRatioMaxX < 0.25 || angleRatioMaxX > 0.75) {
            newMaxExtentX =
                    pntMaxX.x +
                    annoBoundsTheta.valueLabelMaxBoundsX.label.width.value;
        }
        else {  //  must be 90 degrees or 270 degrees
            newMaxExtentX =
                pntMaxX.x +
                annoBoundsTheta.valueLabelMaxBoundsX.label.width.value/2;
        }
    }

    if (isWebShaped(rAxis)) {
        newMaxExtentXRadius = 0.;
    }
    else {
        newMaxExtentXRadius = rAxis.polarAxisParams.xCenter + newRadius;
        newMaxExtentX = Math.max(newMaxExtentX, newMaxExtentXRadius);
    }

    if (isWebShaped(rAxis) ||
        (minExtentX < (rAxis.polarAxisParams.xCenter -
                                rAxis.polarAxisParams.radius)))
    {
        if (angleRatioMinX > 0.25 && angleRatioMinX < 0.75) {
            newMinExtentX =
                pntMinX.x -
                annoBoundsTheta.valueLabelMinBoundsX.label.width.value;
        }
        else {  //  must be either 90 degrees or 270 degrees
            newMinExtentX =
                pntMinX.x -
                annoBoundsTheta.valueLabelMinBoundsX.label.width.value/2;
        }
    }

    if (isWebShaped(rAxis)) {
        newMinExtentXRadius = Double.MAX_VALUE;
    }
    else {
        newMinExtentXRadius = rAxis.polarAxisParams.xCenter - newRadius;
        newMinExtentX = Math.min(newMinExtentX, newMinExtentXRadius);
    }

    double newWidth = newMaxExtentX - newMinExtentX;
    double deltaWidth = (chartAreaRect.width - newWidth)/2.0;

    if (angleRatioMaxY != 0. && angleRatioMaxY != 0.5) {
        if (isWebShaped(rAxis) ||
            (maxExtentY > (rAxis.polarAxisParams.yCenter +
                                rAxis.polarAxisParams.radius)))
        {
            newMaxExtentY =
                pntMaxY.y +
                annoBoundsTheta.valueLabelMaxBoundsY.label.height.value;
        }
    }
    else {
        //  label with maximum y extent is at 0 degrees or 180 degrees
        //  (probably a very large font!)
        newMaxExtentY =
            pntMaxY.y +
            annoBoundsTheta.valueLabelMaxBoundsY.label.height.value/2;
    }

    if (isWebShaped(rAxis)) {
        newMaxExtentYRadius = 0.;
    }
    else {
        newMaxExtentYRadius = rAxis.polarAxisParams.yCenter + newRadius;
        newMaxExtentY = Math.max(newMaxExtentY, newMaxExtentYRadius);
    }

    if (angleRatioMinY != 0. && angleRatioMinY != 0.5) {
        if (isWebShaped(rAxis) ||
            (minExtentY < (rAxis.polarAxisParams.yCenter -
                                rAxis.polarAxisParams.radius)))
        {
            newMinExtentY = pntMinY.y -
                annoBoundsTheta.valueLabelMinBoundsY.label.height.value;
        }
    }
    else {
        //  label with minimum y extent is at 0 degrees or 180 degrees
        //  (probably a very large font!)
        newMinExtentY = pntMinY.y -
            annoBoundsTheta.valueLabelMinBoundsY.label.height.value/2;
    }

    if (isWebShaped(rAxis)) {
        newMinExtentYRadius = Double.MAX_VALUE;
    }
    else {
        newMinExtentYRadius = rAxis.polarAxisParams.yCenter - newRadius;
        newMinExtentY = Math.min(newMinExtentY, newMinExtentYRadius);
    }

    double newHeight = newMaxExtentY - newMinExtentY;
    double deltaHeight = (chartAreaRect.height - newHeight)/2.0;

    double newCenterX = 0.;
    if (newMinExtentX < newMinExtentXRadius) {
        double dxNew = (rAxis.polarAxisParams.xCenter - pntMinX.x);
        if (angleRatioMinX != 0.25 && angleRatioMinX != 0.75)
        {
            newCenterX =
                annoBoundsTheta.valueLabelMinBoundsX.label.width.value +
                dxNew + deltaWidth;
        }
        else {
            //  label with minimum x coordinate is at either
            //  90 degrees or 270 degrees
            newCenterX =
                annoBoundsTheta.valueLabelMinBoundsX.label.width.value/2 +
                deltaWidth;
        }
    }
    else if (newMaxExtentX > newMaxExtentXRadius) {
        double dxNew = (pntMaxX.x - rAxis.polarAxisParams.xCenter);
        if (angleRatioMaxX != 0.25 && angleRatioMaxX != 0.75)
        {
            newCenterX = chartAreaRect.width -
                (annoBoundsTheta.valueLabelMaxBoundsX.label.width.value +
                dxNew + deltaWidth);
        }
        else {
            //  label with minimum x coordinate is at either
            //  90 degrees or 270 degrees
            newCenterX =
                annoBoundsTheta.valueLabelMaxBoundsX.label.width.value/2 +
                deltaWidth;
        }
    }
    else {
        newCenterX = chartAreaRect.width/2.0;
    }

    double newCenterY = 0.;
    if (newMinExtentY < newMinExtentYRadius) {
        double dyNew = (rAxis.polarAxisParams.yCenter - pntMinY.y);
        if (angleRatioMinY != 0. && angleRatioMinY != 0.5) {
            newCenterY =
                annoBoundsTheta.valueLabelMinBoundsY.label.height.value +
                dyNew + deltaHeight;
        }
        else {
            newCenterY =
                annoBoundsTheta.valueLabelMinBoundsY.label.height.value/2 +
                deltaHeight;
        }
    }
    else if (newMaxExtentY > newMaxExtentYRadius) {
        double dyNew = (pntMaxY.y - rAxis.polarAxisParams.yCenter);
        if (angleRatioMaxY != 0. && angleRatioMaxY != 0.5) {
            newCenterY =
                chartAreaRect.height -
                (annoBoundsTheta.valueLabelMaxBoundsY.label.height.value +
                dyNew + deltaHeight);
        }
        else {
            newCenterY =
                annoBoundsTheta.valueLabelMaxBoundsY.label.height.value/2 +
                deltaHeight;
        }
    }
    else {
        newCenterY = chartAreaRect.height/2.0;
    }

    //  Compute x, y, and radius offsets needed to keep all
    //  axis annotations within view:
    chartDeltas.dx = (int)(newCenterX - rAxis.polarAxisParams.xCenter);
                    //  x delta for chart center
    chartDeltas.dy = (int)(newCenterY - rAxis.polarAxisParams.yCenter);
                    //  y delta for chart center
    chartDeltas.dWidth = (int)(newRadius - rAxis.polarAxisParams.radius);
    chartDeltas.dHeight = chartDeltas.dWidth;
}

protected boolean isWebShaped(JCAxis ax)
{
    return(ax.isRadar() && !ax.polarAxisParams.isCircularGrid());
}

/**
 * Finds the biggest symbols size of all data views to be drawn in this
 * chart area.
 * @return the size of the biggest symbol
 */
protected int getBiggestSymbolSize()
{
    int biggest = 0;
    JCChart chart = (JCChart) getParent();
    if (chart != null) {
        List cdvList = chart.getDataView();
        for (int i = 0; i < cdvList.size(); i++) {
            ChartDataView cdv = (ChartDataView) cdvList.get(i);
            List seriesList = cdv.getSeries();
            for (int j = 0; j < seriesList.size(); j++) {
                ChartDataViewSeries cdvs =
                    (ChartDataViewSeries) seriesList.get(j);
                JCChartStyle style = cdvs.getStyle();
                biggest = Math.max(biggest, style.getSymbolSize());
            }
        }
    }

    return(biggest);
}

/**
 * Determines if the data is "dirty" (that is, if the data has changed
 * since the last redisplay).
 * @return if <code>true</code>, the data has changed; if <code>false</code>, the data has
 * not changed. Takes all <code>ChartDataView</code> objects into account.
 */
public boolean isDirty()
{
    ensureDrawables();
    for (Enumeration e = drawables.keys(); e.hasMoreElements(); ) {
        Object key = e.nextElement();
        ChartDraw d = (ChartDraw)drawables.get(key);
        if (d.dataObject.isChanged()) {
            return(true);
        }
    }
    for (int i = 0; i < xaxes.size(); i++) {
        if (((JCAxis)xaxes.get(i)).isChanged()) {
            return(true);
        }
    }
    for (int j = 0; j < yaxes.size(); j++) {
        if (((JCAxis)yaxes.get(j)).isChanged()) {
            return(true);
        }
    }
    return(false);
}

protected boolean inSuperReshape = false;

/**
 * An overridden <code>setBounds()</code> method, which makes sure that
 * the <code>Changed</code> property is set and that a recalc occurs as a result of the
 * reshape. This overridden reshape doesn't do a <code>repaint()</code>;
 * it is up to the caller to make a proper repaint call so that all of the
 * chart children get redrawn in the correct order.
 * @param x the new <i>x</i>-coordinate of this component
 * @param y the new <i>y</i>-coordinate of this component
 * @param w the new <code>width</code> of this component
 * @param h the new <code>height</code> of this component
 */
public void setBounds(int x, int y, int w, int h)
{
    if (x == getLocation().x &&
        y == getLocation().y &&
        w == getSize().width &&
        h == getSize().height)
    {
        return;
    }
    inSuperReshape = true;
    super.setBounds(x, y, w, h);
    inSuperReshape = false;
    changed = true;
    changedFlag |= Changeable.RECALC;
    recalc();
}

/**
 * Overrides the superclass's method in order to avoid repaint if
 * <code>super.setBounds()</code> is called.
 */
public void repaint()
{
    if (getSize().width <= 0 || getSize().height <= 0) {
        return;
    }
    if (inSuperReshape) {
        return;
    }
    super.repaint();
}

/**
 * Gets the value of the <code>DeltaX</code> property.
 * The <code>DeltaX</code> property controls the shift in the <i>x</i> direction
 * to create the 3D effect in 3D chart types. It is derived
 * from elevation, rotation, and depth, and should only be
 * set by drawables and other internal classes that need
 * to modify the delta in certain cases.
 * @return number of pixels to shift in the <i>x</i> direction
 */
public int getDeltaX()
{
    return(deltaX);
}

/**
 * Sets the value of the <code>DeltaX</code> property.
 * The <code>DeltaX</code> property controls the shift in the <i>x</i> direction
 * to create the 3D effect in 3D chart types. It is derived
 * from elevation, rotation, and depth, and should only be
 * set by drawables and other internal classes that need
 * to modify the delta in certain cases.
 * @param dx number of pixels to shift in the <i>x</i> direction
 */
public void setDeltaX(int dx)
{
    deltaX = dx;
}

/**
 * Gets the value of the <code>DeltaY</code> property.
 * The <code>DeltaY</code> property controls the shift in the <i>y</i> direction
 * to create the 3D effect in 3D chart types. It is derived
 * from elevation, rotation, and depth, and should only be
 * set by drawables and other internal classes that need
 * to modify the delta in certain cases.
 * @return number of pixels to shift in the <i>y</i> direction
 */
public int getDeltaY()
{
    return(deltaY);
}

/**
 * Sets the value of the <code>DeltaY</code> property.
 * The <code>DeltaY</code> property controls the shift in the <i>y</i> direction
 * to create the 3D effect in 3D chart types. It is derived
 * from elevation, rotation, and depth, and should only be
 * set by drawables and other internal classes that need
 * to modify the delta in certain cases.
 * @param dy number of pixels to shift in the <i>y</i> direction
 */
public void setDeltaY(int dy)
{
    deltaY = dy;
}

/**
 * Gets the value of the <code>Draw3D</code> property, which determines whether
 * the chart area will attempt to draw chart types as 3D chart types.
 * @return <code>true</code> if 3D drawing will be attempted
 */
public boolean getDraw3D()
{
    return(draw3D);
}

/**
 * Gets the value of the <code>HasBar</code> property.
 * The <code>HasBar</code> property determines whether the chart
 * area includes a bar chart (bar or stacking bar).
 * @return <code>true</code> if the chart area includes a bar chart type
 */
public boolean getHasBar()
{
    return(hasBar);
}

/**
 * Returns the current Rectangle representing the plotting region in chart area.
 * @return the plotting region rectangle
 */
public Rectangle getPlotRect()
{
     return(plotRect);
}

/**
 * Returns the current Rectangle representing the
 * chart area. This is the same as <code>getDrawingArea()</code>.
 * @return the chart area rectangle
 */
public Rectangle getChartAreaRect()
{
    return(chartAreaRect);
}

/**
 * Gets the action handler for this chart area. The action handler handles the
 * interactions for this chart area.
 * @return the current action handler
 */
public ActionHandler getActionHandler()
{
    return(handler);
}

/**
 * Gets the value of the <code>InAction</code> property, which is a flag that
 * determines whether an action is in process.
 * @return Is an action in progress?
 */
public boolean getInAction()
{
    return(inAction);
}

/**
 * Sets the value of the <code>InAction</code> property, which
 * is a flag that determines whether an action is in process.
 * @param ia if set to <code>true</code>, an action is now in process
 */
public void setInAction(boolean ia)
{
    inAction = ia;
}

/**
 * Finds an axis by searching for it using the specified name.
 * @param axisName the name to search for
 * @return the found <code>JCAxis</code> instance, or <code>null</code> if no
 * matching axis is found.
 */
public JCAxis findAxisByName(String axisName)
{
    JCAxis ax = findAxisByName(axisName, true);
    if (ax == null) {
        ax = findAxisByName(axisName, false);
    }
    return(ax);
}

/**
 * Searches for the axis with the specified name.
 * @param axisName the name of the axis to search for
 * @param x if <code>true</code>, will search the x axis list; if
 * <code>false</code>, will search the y axis list
 * @return the axis that was found; <code>null</code> if none is found
 */
protected JCAxis findAxisByName(String axisName, boolean x)
{
    List axes = x ? xaxes : yaxes;
    if (axes == null) {
        return(null);
    }

    for (ListIterator i = axes.listIterator(); i.hasNext();) {
        Object it = i.next();
        // Possibly useless check
        if (it instanceof JCAxis) {
            if (axisName.equals(((JCAxis)it).getName())) {
                return((JCAxis)it);
            }
        }
    }
    return(null);
}


/*********
 * Package-level methods.
 */

/**
 * Layout annotation for the given axes list.
 * @param axes the list of axes for which to layout annotations
 */
protected void layoutAnnotations(List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)(axes.get(i));
        ax.layoutAnnotations();
    }
}


/**
 * Creates a Drawable for the provided <code>ChartDataView</code> object,
 * removing the old Drawable if necessary.
 * @param dataObject <code>ChartDataView</code> for which the Drawable must be
 * created
 */
private void addDrawable(ChartDataView dataObject)
{
    if (dataObject == null) {
        return;
    }

    // Check for existing object
    if (hasMatchingDrawable(dataObject)) {
        removeDrawable(dataObject);
    }

    // Create a new Draw-based rendering object
    ChartDraw dr = ChartDraw.makeDrawObject(dataObject);
    drawables.put(dataObject, dr);
}

/**
 * Removes the Drawable created for the specified <code>ChartDataView</code> object.
 * @param dataObject <code>ChartDataView</code> object for which the Drawable is to be
 *     removed
 */
protected void removeDrawable(ChartDataView dataObject)
{
    if (drawables != null) {
        drawables.remove(dataObject);
    }
}

/**
 * Sets the drawable to <code>null</code> and resets hashtable.
 */
protected void clearDrawables()
{
    drawables = new Hashtable();
}

/**
 * Checks to see if the specified <code>ChartDataView</code> object has a
 * Drawable.
 * @param dataObject <code>ChartDataView</code> object to check
 * @return if <code>true</code>, a Drawable exists for the
 *     <code>dataObject</code>; if <code>false</code>, no
 *     Drawable exists for the <code>dataObject</code>
 */
protected boolean hasMatchingDrawable(ChartDataView dataObject)
{
    // Check for existing object
    if (drawables != null && drawables.containsKey(dataObject)) {
        ChartDraw d = (ChartDraw)drawables.get(dataObject);
        if (d.chartType == d.dataObject.getChartType()) {
            return(true);
        }
    }
    return(false);
}

/**
 * Given a data view object, finds the corresponding drawable.
 * @param dataObject data view object
 * @return drawable corresponding to the provided data view object
 */
public ChartDraw getDrawable(ChartDataView dataObject)
{
    ChartDraw ret = null;
    if (drawables != null && drawables.containsKey(dataObject)) {
        ChartDraw d = (ChartDraw)drawables.get(dataObject);
        if (d.chartType == d.dataObject.getChartType()) {
            ret = d;
        }
    }
    return(ret);
}

/**
 * Calculates the annotation anchor positions for the provided
 * axis objects.
 * @param axes vector of axis objects
 */
protected void calcAnchors(List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        ax.calcAnchors();
    }
}


/**
 * Calculates the data limits and axis limits.
 */
protected void calcLimits()
{
    clearLimits();
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    // If first data view is a pie chart, do only it
    if (pieFirst) {
        for (Iterator i=chart.getOriginalDataView().iterator(); i.hasNext();) {
            ChartDataView data = (ChartDataView)i.next();
            if (data.isVisible() && data.getChartType() == JCChart.PIE) {
                data.calcLimits();
                break;
            }
        }
    }
    // If first data view is a radar or area radar chart, do only it
    else if (radarFirst) {
        for (Iterator i=chart.getOriginalDataView().iterator(); i.hasNext();) {
            ChartDataView data = (ChartDataView)i.next();
            int chartType = data.getChartType();
            if (data.isVisible() &&
                (chartType == JCChart.RADAR ||
                 chartType == JCChart.AREA_RADAR))
            {
                data.calcLimits();
                break;
            }
        }
    }
    // If first data view is a polar chart, do only polar charts.
    else if (polarFirst) {
        for (Iterator i=chart.getOriginalDataView().iterator(); i.hasNext();) {
            ChartDataView data = (ChartDataView)i.next();
            if (data.isVisible() && (data.getChartType() == JCChart.POLAR)) {
                data.calcLimits();
                // Don't break out of the loop -- do ALL Polar charts.
            }
        }
    }
    // Otherwise, do everything else
    else {
        for (Iterator i=chart.getOriginalDataView().iterator(); i.hasNext();) {
            ChartDataView data = (ChartDataView)i.next();
            int chartType = data.getChartType();
            if (chartType != JCChart.PIE &&
                chartType != JCChart.RADAR &&
                chartType != JCChart.AREA_RADAR)
            {
                if (data.isVisible()) {
                    data.calcLimits();
                }
            }
        }
    }
}

/**
 * Clears the data limits.
 */
protected void clearLimits()
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    for (Iterator i = chart.getOriginalDataView().iterator(); i.hasNext();) {
        ChartDataView data = (ChartDataView)i.next();
        data.clearLimits();
    }
}

/**
 * Performs a cancel operation.
 */
protected void cancel()
{
    //setDoubleBuffer(saveDBState);
    handler = null;
    inAction = false;
    setChanged(true, Changeable.RECALC);
}

/**
 * Performs a reset operation.
 */
protected void reset()
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    boolean batched = chart.isBatched();
    chart.setBatched(true);
    if (horizActionAxis != null) {
        horizActionAxis.min.isDefault = true;
        horizActionAxis.max.isDefault = true;
        horizActionAxis.setChanged(true, Changeable.RECALC, false);
        chart.sendEvent(horizActionAxis);
    }
    if (vertActionAxis != null) {
        vertActionAxis.min.isDefault = true;
        vertActionAxis.max.isDefault = true;
        vertActionAxis.setChanged(true, Changeable.RECALC, false);
        chart.sendEvent(vertActionAxis);
    }
    // setChanged() called in cancel() - save an update
    cancel();
    chart.setBatched(batched);
}


/**
 * Method that draws a rubber band rectangle from the point in "start" to the
 * provided point.
 * @param p coordinates to use as the end point for the rectangle
 * @param start the start point
 */
public void drawRubberBand(Point p, Point start)
{
    if (p == null) {
        return;
    }
    if (p.x == start.x || p.y == start.y) {
        return;
    }
    int ulx = start.x;
    int uly = start.y;
    int lrx = p.x;
    int lry = p.y;
    if (rbGC == null) {
        if (getParent() == null) {
            return;
        }
    }
    Graphics gc = (rbGC == null) ? getGraphics() : rbGC.create();
    if (lrx-ulx < 0) {
        // Swap x values
        int tempx = lrx;
        lrx = ulx;
        ulx = tempx;
    }
    if (lry-uly < 0) {
        // Swap y values
        int tempy = lry;
        lry = uly;
        uly = tempy;
    }
    gc.setXORMode(Color.white);
    gc.drawRect(ulx, uly, lrx-ulx, lry-uly);
    gc.setXORMode(Color.black);
    gc.drawRect(ulx+1, uly+1, lrx-ulx, lry-uly);
    gc.dispose();
}

/**
 * Method used to draw a rubber band line on the chart area.
 * @param start the start point of the line
 * @param end the end point of the line
 */
public void drawRubberLine(Point start, Point end)
{
    if (start == null || end == null) {
        return;
    }
    if (getParent() == null) {
        return;
    }
    Graphics gc = getGraphics();

    gc.setXORMode(Color.white);
    gc.drawLine(start.x, start.y, end.x, end.y);
    gc.setXORMode(Color.black);
    gc.drawLine(start.x+1, start.y+1, end.x+1, end.y+1);
    gc.dispose();
}

/**
 * Chooses the point closest to the provided position.  If
 * a data view is supplied, picks a point on the provided
 * data view only.  Otherwise, tries to find the closest point
 * from any data view.
 * @param p pick point in pixels relative to <code>JCChartArea</code> object
 * @param data data view on which to perform pick.  If <code>null</code>,
 * all data views are used.
 * @param focus pick focus used to determine distance between point
 * and data point.  If equal to <code>ChartDataView.PICK_FOCUS_LOCAL</code>, the
 * pick focus specified in the appropriate <code>ChartDataView</code> object
 * will be used.
 * @return <code>JCDataIndex</code> object representing the picked point.
 */
protected JCDataIndex pick(Point p, ChartDataView data, int focus)
{
    JCDataIndex di = null;
    ChartDraw drawable = null;
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return(null);
    }

    if (data != null) {
        drawable = getDrawable(data);
        if (drawable != null) {
            int pickFocus =
                (focus == ChartDataView.PICK_FOCUS_LOCAL) ? data.getPickFocus()
                                                          : focus;
            di = drawable.pick(p, pickFocus);
        }
        if (di != null) {
            di.obj = this;
            di.dataView = data;
        }
        return(di);
    }

    // Check the data views in the reverse order that they were drawn.
    // Data Views on top should be given priority over those behind.

    List dtList = chart.getOriginalDataView();
    // First check the 2d data views on the front plane
    for (ListIterator i = dtList.listIterator(dtList.size());
         i.hasPrevious();) {
        ChartDataView dv = (ChartDataView) i.previous();
        ChartDraw d = getDrawable(dv);
        if (d == null) {
            continue;
        }
        if (!dv.is3DChartType() && d.drawFront) {
            di = pickOne(d, p, focus, di);
        }
    }
    // Now check the 3d data views on the front plane
    for (ListIterator i = dtList.listIterator(dtList.size()); i.hasPrevious();)
    {
        ChartDataView dv = (ChartDataView) i.previous();
        ChartDraw d = getDrawable(dv);
        if (d == null) {
            continue;
        }
        if (dv.is3DChartType()) {
            di = pickOne(d, p, focus, di);
        }
    }
    // Finally, check the data views on the back axes.
    for (ListIterator i = dtList.listIterator(dtList.size()); i.hasPrevious();)
    {
        ChartDataView dv = (ChartDataView) i.previous();
        ChartDraw d = getDrawable(dv);
        if (d == null) {
            continue;
        }
        if (!d.drawFront && !dv.is3DChartType()) {
            di = pickOne(d, p, focus, di);
        }
    }

    return(di);
}

/**
 * Performs a pick operation for one drawable; designed to be called a
 * number of times, once for each different drawable.
 * @param d the drawable to do the pick operation on
 * @param p pick point in pixels relative to <code>JCChartArea</code> object
 * @param focus pick focus used to determine distance between point
 * and data point.  If equal to <code>ChartDataView.PICK_FOCUS_LOCAL</code>, the
 * pick focus specified in the appropriate <code>ChartDataView</code> object
 * will be used.
 * @param di the <code>JCDataIndex</code> object that represents the point
 * picked so far. If a point in this drawable is closer, it will take
 * precedence; <code>null</code> if this is the first drawable being picked.
 * @return <code>JCDataIndex</code> object representing the picked point
 */
private JCDataIndex pickOne(ChartDraw d, Point p, int focus, JCDataIndex di)
{
    JCDataIndex nd = d.pick(p,
                        focus == ChartDataView.PICK_FOCUS_LOCAL ?
                        d.dataObject.getPickFocus() : focus);
    if (nd != null) {
        nd.obj = this;
        nd.dataView = d.dataObject;
    }
    if (di == null) {
        di = nd;
    }
    else if (nd != null && di.distance > nd.distance) {
        di = nd;
    }
    return(di);
}

boolean isXAxis(JCAxis ax)
{
    return(xaxes.contains(ax));
}

/********
 * Internal methods
 */

/**
 * Method that determines whether the chart drawn is 3d.
 * @param chartType the type of the chart
 * @return Is the chart 3d?
 */
private boolean is3DChartType(int chartType)
{
    if (chartType == JCChart.BAR || chartType == JCChart.STACKING_BAR) {
        return(true);
    }
    return(false);
}

private void getHorizVertAxes(List h, List v)
{
    for (int i = 0; i < xaxes.size(); i++) {
        JCAxis ax = (JCAxis)(xaxes.get(i));
        if (ax.i_vertical) {
            v.add(ax);
        }
        else {
            h.add(ax);
        }
    }
    for (int i = 0; i < yaxes.size(); i++) {
        JCAxis ax = (JCAxis)(yaxes.get(i));
        if (ax.i_vertical) {
            v.add(ax);
        }
        else {
            h.add(ax);
        }
    }
}

private void drawWebBoundingBox(Graphics gc, JCLineStyle ls, JCAxis xAxis)
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    List data = chart.getOriginalDataView();
    ChartDataView cdv = null;
    if (data != null) {
        cdv = (ChartDataView)data.get(0);
        int[] limits = ChartDataView.getFirstLastPoints(cdv);
        int firstPoint = limits[0];
        int lastPoint = limits[1];
        int nSpokes = lastPoint - firstPoint + 1;
        double radius = Math.rint(xAxis.polarAxisParams.radius);
        xAxis.drawWebLine(gc, ls, radius, nSpokes);
    }
}

private void fillWebRegion(Graphics gc, JCFillStyle fs, JCAxis xAxis)
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    List data = chart.getOriginalDataView();
    ChartDataView cdv = null;
    if (data != null) {
        cdv = (ChartDataView)data.get(0);
        int[] limits = ChartDataView.getFirstLastPoints(cdv);
        int firstPoint = limits[0];
        int lastPoint = limits[1];
        int nSpokes = lastPoint - firstPoint + 1;
        double radarAngle =
                JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
                                          nSpokes, 1, false);  // Radians
        Polygon poly = new Polygon();
        double radius = Math.rint(xAxis.polarAxisParams.radius);
        Point pt = null;
        for (int i = 0; i < nSpokes; i++) {
            pt = xAxis.toPixelPolar(radarAngle * (double)i, radius, false);
            //  toPixelPolar() adds originBase angle.
            poly.addPoint(pt.x, pt.y);
        }
        fs.fillPolygon(gc, poly);
    }
}

private void drawBackplane(Graphics gc)
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    List data = chart.getOriginalDataView();
    ChartDataView cdv = null;
    if (data != null) {
        cdv = (ChartDataView)data.get(0);
        if (data.size() == 1 && cdv != null &&
            cdv.getChartType() == JCChart.PIE) {
            return;
        }
    }

    JCFillStyle fs = null;
    // Opaque?  Initialize fill style for drawing backplane.
    if (isOpaque()) {
        fs = new JCFillStyle(plotArea.getBackground(), JCFillStyle.SOLID);
    }
    int x = plotRect.x;
    int y = plotRect.y;
    int width = plotRect.width;
    int height = plotRect.height;

    int chartType = (cdv == null) ? JCChart.NONE : cdv.getChartType();
    boolean isPolar = (chartType == JCChart.POLAR);
    boolean isRadar = (chartType == JCChart.RADAR ||
                       chartType == JCChart.AREA_RADAR);
    JCAxis ax = (cdv == null) ? null : cdv.getXAxis();
    //  For webbed Radar charts, we need to get the number of labels for
    //  the x axis.
    boolean fCircularGridLines = false;
    if ((isPolar || isRadar) && ax != null) {
        fCircularGridLines = ax.polarAxisParams.isCircularGrid();
    }

    if (isPolar || (isRadar && fCircularGridLines)) {
        //  Polar or Radar chart with circular grid lines.
        if (ax != null) {
            if (fs != null) {
                fs.fillCircle(gc,
                              (int)Math.round(ax.polarAxisParams.xCenter),
                              (int)Math.round(ax.polarAxisParams.yCenter),
                              (int)Math.round(ax.polarAxisParams.radius));
            }
            // Draw a bounding box if the axisBoundingBox flag is set
            //                       and
            // this is a Polar chart with the x axis hidden, or this
            // is a Radar chart.
            if (axisBoundingBox &&
                (isRadar || (isPolar && !ax.isVisible())))
            {
                Color color = plotArea.getForeground();
                JCLineStyle ls = new JCLineStyle(1, color, JCLineStyle.SOLID);
                ls.drawCircle(gc,
                              (int)Math.round(ax.polarAxisParams.xCenter),
                              (int)Math.round(ax.polarAxisParams.yCenter),
                              (int)Math.round(ax.polarAxisParams.radius));
            }
        }
    }
    else if (isRadar && !fCircularGridLines) {
        //  Radar chart with webbed grid lines.
        if (ax != null) {
            if (fs != null) {
                fillWebRegion(gc, fs, ax);
            }
            // Draw a bounding box if the axisBoundingBox flag is set.
            if (axisBoundingBox) {
                Color color = plotArea.getForeground();
                JCLineStyle ls = new JCLineStyle(1, color, JCLineStyle.SOLID);
                gc.setColor(color);
                drawWebBoundingBox(gc, ls, ax);
            }
        }
    }
    else if ((draw3D || axisBoundingBox) && !pieFirst) {
        if (draw3D && deltaX > 0) {
            x += deltaX;
        }
        if (draw3D && deltaY > 0) {
            y += deltaY;
        }
        int dx = (draw3D) ? JCChartUtil.abs(deltaX) : 0;
        int dy = (draw3D) ? JCChartUtil.abs(deltaY) : 0;
        width = Math.max(0, width-dx);
        height = Math.max(0, height-dy);
        if (fs != null) {
            fs.fillRect(gc, x, y, width, height);
        }
        if (axisBoundingBox) {
            gc.setColor(plotArea.getForeground());
            gc.drawRect(x, y, width, height);
        }
        backplane = new Rectangle(x, y, width, height);
    }
    else {
        if (fs != null)  {
            fs.fillOutlineRect(gc, x, y, width, height, fs.getColor());
        }
        backplane = new Rectangle(x, y, width, height);
    }
}

/**
 * Returns <code>true</code> if the specified axis is associated
 * only with a circular chart (<code>PIE</code>, <code>POLAR</code>,
 * <code>RADAR</code>, or <code>AREA_RADAR</code>).
 * <P>
 * <B>Note:</b>  A <code>PIE</code> chart doesn't normally have axes associated
 * with it, but we may have a case where the user switches
 * the chart type from <code>PLOT</code>, <code>BAR</code>, and so on (for
 * example, a chart type that has axes) to <code>PIE</code> in the customizer.
 * @return <code>true</code> if the specified axis is associated only
 * with a circular chart; <code>false</code> otherwise
 * @param ax
 */
private boolean isCircularChartAxisOnly(JCAxis ax)
{
    boolean fCircularChartAxisOnly = true;
    Iterator dataViews = getDataViews();
    ChartDataView dv = null;
    if (dataViews != null) {
        // Is the axis associated with any of the data views?
        // If not, return false;
        boolean fAssociatedWithDataView = false;
        while (dataViews.hasNext()) {
            dv = (ChartDataView)dataViews.next();
            if (dv != null && (dv.xaxis == ax || dv.yaxis == ax)) {
                int chartType = dv.getChartType();
                if (chartType != JCChart.PIE &&
                    chartType != JCChart.POLAR &&
                    chartType != JCChart.RADAR &&
                    chartType != JCChart.AREA_RADAR)
                {
                    fCircularChartAxisOnly = false;
                    break;
                }
                fAssociatedWithDataView = true;
            }
        }
        if (!fAssociatedWithDataView) {
            return(false);
        }
    }
    return(fCircularChartAxisOnly && (dataViews != null));
}

/**
 * Internal method that draws the axis objects contained in
 * a vector.
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be drawn
 */
private void drawAxes(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    boolean fPolar = isPolar();
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        // If the first data view is of type POLAR, then any subsequent
        // data view must also be POLAR if we are to display the associated
        // axes.
        if (fPolar && !ax.isPolar()) {
            continue;
        }
        // If an axis is uniquely associated with a circular
        // chart (PIE, POLAR, RADAR, or AREA_RADAR), don't
        // draw the axis.
        if (isCircularChartAxisOnly(ax)) {
            continue;
        }
        if (ax.isVisible()) {
            ax.draw(gc);
        }
    }
}

/**
 * Internal method that draws the axis objects associated with
 * a specified data view.  This method is called when the
 * first data view is of type <code>RADAR</code> or <code>AREA_RADAR</code>.
 * @param gc graphics context to use for drawing
 * @param cdv the data view for which the axes are to be drawn
 */
private void drawAxes(Graphics gc, ChartDataView cdv)
{
    if (cdv == null || !cdv.isVisible()) {
        return;
    }
    JCAxis ax = cdv.xaxis;
    if (ax != null && ax.isVisible()) {
        ax.draw(gc);
    }
    ax = cdv.yaxis;
    if (ax != null && ax.isVisible()) {
        ax.draw(gc);
    }
}

/**
 * Internal method that draws the axis objects, contained in
 * a vector, that are on the backplane ("on the backplane" refers to
 * a bar chart is not associated with the axis).
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be checked and possibly drawn
 */
private void drawBpAxes(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.visible && !ax.isBar()) {
            ax.draw(gc);
        }
    }
}

/**
 * Internal method that draws the axis objects, contained in
 * a vector, that are on the frontplane ("on the frontplane" refers to
 * a bar chart is associated with the axis).
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be checked and possibly drawn
 */
private void drawFpAxes(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.visible && ax.isBar()) {
            ax.draw(gc);
        }
    }
}

/**
 * Internal method that draws the axis annotations and title
 * for axes in a vector.
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be drawn
 */
private void drawAxisAnnotationsAndTitle(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    boolean fPolar = isPolar();
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (fPolar && !ax.isPolar()) {
            continue;
        }
        if (isCircularChartAxisOnly(ax)) {
            continue;
        }
        if (ax.visible) {
            ax.drawAnnotationsAndTitle(gc);
        }
    }
}

private void drawAxisAnnotationsAndTitle(Graphics gc, ChartDataView cdv)
{
    if (cdv == null || !cdv.isVisible()) {
        return;
    }
    JCAxis ax = cdv.xaxis;
    if (ax != null && ax.isVisible()) {
        ax.drawAnnotationsAndTitle(gc);
    }
    ax = cdv.yaxis;
    if (ax != null && ax.isVisible()) {
        ax.drawAnnotationsAndTitle(gc);
    }
}

/**
 * Internal method that draws the axis annotations and title for axes,
 * contained in a vector, that are on the backplane (the backplane is an
 * axis that does not have a bar chart associated with it).
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be checked and possibly drawn
 */
private void drawBpAxisAnnotationsAndTitle(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.visible && !ax.isBar()) {
            ax.drawAnnotationsAndTitle(gc);
        }
    }
}

/**
 * Internal method that draws the axis annotations and title for axes,
 * contained in a vector, that are on the frontplane (definition of the frontplane :
 * axis has a bar chart associated with it).
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be checked and possibly drawn
 */
private void drawFpAxisAnnotationsAndTitle(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.visible && ax.isBar()) {
            ax.drawAnnotationsAndTitle(gc);
        }
    }
}

/** Draw 3D Axis troughs.  Drawing order depends on delta values, axis
 *  inversion, and axis direction.
 *  @param gc  graphics context
 *  @param frontY  the 3d Y axis
 *  @param frontX  the 3d X axis
 */
private void preDraw3DAxes(Graphics gc, JCAxis frontX, JCAxis frontY)
{
    if (frontX == null || frontY == null) {
        return;
    }

    boolean inverted = frontX.i_vertical;
    int xPlacement = frontX.getPlacement();
    int yPlacement = frontY.getPlacement();

    if (xPlacement == JCAxis.ORIGIN) {
        if (frontY.getOrigin() == frontY.getMax()) {
            xPlacement = JCAxis.MAX;
        }
        else if (frontY.getOrigin() == frontY.getMin()) {
            xPlacement = JCAxis.MIN;
        }
    }
    else if (xPlacement == JCAxis.VALUE_ANCHORED) {
        if (frontX.getPlacementLocation() == frontY.getMax()) {
            xPlacement = JCAxis.MAX;
        }
        else if (frontX.getPlacementLocation() == frontY.getMin()) {
            xPlacement = JCAxis.MIN;
        }
    }

    if (yPlacement == JCAxis.ORIGIN) {
        if (frontX.getOrigin() == frontX.getMax()) {
            yPlacement = JCAxis.MAX;
        }
        else if (frontX.getOrigin() == frontX.getMin()) {
            yPlacement = JCAxis.MIN;
        }
    }
    else if (yPlacement == JCAxis.VALUE_ANCHORED) {
        if (frontY.getPlacementLocation() == frontX.getMax()) {
            yPlacement = JCAxis.MAX;
        }
        else if (frontY.getPlacementLocation() == frontX.getMin()) {
            yPlacement = JCAxis.MIN;
        }
    }

    if (frontX.isReversed()) {
        if (yPlacement == JCAxis.MIN) {
            yPlacement = JCAxis.MAX;
        }
        else if (yPlacement == JCAxis.MAX) {
            yPlacement = JCAxis.MIN;
        }
    }
    if (frontY.isReversed()) {
        if (xPlacement == JCAxis.MIN) {
            xPlacement = JCAxis.MAX;
        }
        else if (xPlacement == JCAxis.MAX) {
            xPlacement = JCAxis.MIN;
        }
    }
    if (frontX.visible) {
        frontX.preDrawCalc();
    }
    if (frontY.visible) {
        frontY.preDrawCalc();
    }

    // If one axis is in shadow and the other in highlight, must draw them
    // correctly for the proper 3d look
    if (deltaX >= 0 && deltaY < 0) {
        if (yPlacement != JCAxis.MIN && yPlacement != JCAxis.MAX &&
            xPlacement != JCAxis.MIN && xPlacement != JCAxis.MAX)
        {
            if (frontY.visible) {
                frontY.preDraw(gc, false, true);
            }
            if (frontX.visible) {
                frontX.preDraw(gc, true, false);
            }
            if (frontY.visible) {
                frontY.preDraw(gc, false, false);
            }
        }
        else if (yPlacement == JCAxis.MAX ||
                 (yPlacement != JCAxis.MAX &&
                  yPlacement != JCAxis.MIN &&
                  (inverted ? (xPlacement == JCAxis.MIN)
                            : (xPlacement != JCAxis.MAX))))
        {
            if (frontX.visible) {
                frontX.preDraw(gc, true, false);
            }
            if (frontY.visible) {
                frontY.preDraw(gc, true, false);
            }
        }
        else {
            if (frontY.visible) {
                frontY.preDraw(gc, true, false);
            }
            if (frontX.visible) {
                frontX.preDraw(gc, true, false);
            }
        }
    }
    else if (deltaX < 0 && deltaY >= 0) {
        if (yPlacement != JCAxis.MIN && yPlacement != JCAxis.MAX &&
            xPlacement != JCAxis.MIN && xPlacement != JCAxis.MAX)
        {
            if (frontY.visible) {
                frontX.preDraw(gc, false, false);
            }
            if (frontX.visible) {
                frontY.preDraw(gc, true, false);
            }
            if (frontY.visible) {
                frontX.preDraw(gc, false, true);
            }
        }
        else if (yPlacement == JCAxis.MIN ||
                 (yPlacement != JCAxis.MAX &&
                  yPlacement != JCAxis.MIN &&
                  (inverted ? (xPlacement != JCAxis.MIN)
                            : (xPlacement == JCAxis.MAX))))
        {
            if (frontX.visible) {
                frontX.preDraw(gc, true, false);
            }
            if (frontY.visible) {
                frontY.preDraw(gc, true, false);
            }
        }
        else {
            if (frontY.visible) {
                frontY.preDraw(gc, true, false);
            }
            if (frontX.visible) {
                frontX.preDraw(gc, true, false);
            }
        }
    }
    else {
        // If both are in the same color, just draw them
        if (frontX.visible) {
            frontX.preDraw(gc, true, false);
        }
        if (frontY.visible) {
            frontY.preDraw(gc, true, false);
        }
    }
}

/**
 * Internal method that pre-draws the grids in each axis object in
 * a vector. This method is used to make sure that the grids appear
 * on the backplane of a 3D chart.
 * @param gc graphics context to use for drawing
 * @param axes vector of axis objects that are to be drawn
 */
private void preDrawGrids(Graphics gc, List axes)
{
    if (axes == null) {
        return;
    }
    boolean fPolar = isPolar();
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (fPolar && !ax.isPolar()) {
            continue;
        }
        // If an axis is associated with a circular
        // chart only, don't draw that axis.
        if (isCircularChartAxisOnly(ax)) {
            continue;
        }
        if (ax.visible) {
            ax.drawGrid(gc);
        }
    }
}

/**
 * Internal method that pre-draws the grids in each axis object
 * in a specified data view.  This method is called when the
 * first data view is of type <CODE>RADAR</code> or <code>AREA_RADAR</code>.
 * @param gc graphics context to use for drawing
 * @param cdv the data view for which the grids are to be drawn
 */
private void preDrawGrids(Graphics gc, ChartDataView cdv)
{
    if (cdv == null || !cdv.isVisible()) {
        return;
    }
    JCAxis ax = cdv.xaxis;
    if (ax != null && ax.isGridVisible()) {
        ax.drawGrid(gc);
    }
    ax = cdv.yaxis;
    if (ax != null && ax.isGridVisible()) {
        ax.drawGrid(gc);
    }
}

/**
 * Internal method that adjusts the plot area for the
 * provided axes.
 * @param axes vector of axis objects that affect the plot area
 */
private void adjustPlotRectForAxes(List axes)
{
    if (axes == null) {
        return;
    }

    // See if any axis is visible.
    boolean anyAxisVisible = false;
    JCAxis ax = null;
    Iterator iterator = axes.iterator();
    while (iterator.hasNext()) {
        ax = (JCAxis)iterator.next();
        if (ax.visible) {
            anyAxisVisible = true;
            break;
        }
    }

    // Collect space to be left on each side of the plot area for the axes
    int leftSpace = 0;
    int rightSpace = 0;
    int topSpace = 0;
    int bottomSpace = 0;
    for (int i = 0; i < axes.size(); i++) {
        ax = (JCAxis)(axes.get(i));
        // Adjust the plot rect for visible axes only.  If none of the
        // axes are visible, make adjustments for the first axis only.
        if ((anyAxisVisible && !ax.visible) ||
            (!anyAxisVisible && i > 0) ||
            pieFirst || radarFirst)
        {
            continue;
        }

        if (isCircularChartAxisOnly(ax)) {
            continue;
        }

        // For 3d, axes are at same effective position, so space is
        // max of all axes at that position.  For non-3d, space is
        // the accumulation of all of the axes at that position

        if (ax.i_vertical) {
            if (ax.annoHandler.extents.ortho > ax.getWidth())  {
                ax.setSize(ax.annoHandler.extents.ortho, ax.getHeight());
            }

            if (ax.isLeftSide()) {
                if (draw3D) {
                    leftSpace = Math.max(leftSpace, ax.getWidth());
                }
                else {
                    leftSpace += ax.getWidth();
                }
            }
            else if (ax.isRightSide()) {
                if (draw3D) {
                    rightSpace = Math.max(rightSpace, ax.getWidth());
                }
                else {
                    rightSpace += ax.getWidth();
                }
            }
        }
        else {
            if (ax.annoHandler.extents.ortho > ax.getHeight()) {
                ax.setSize(ax.getWidth(), ax.annoHandler.extents.ortho);
            }
            if (ax.isTopSide()) {
                if (draw3D) {
                    topSpace = Math.max(topSpace, ax.getHeight());
                }
                else {
                    topSpace += ax.getHeight();
                }
            }
            else if (ax.isBottomSide()) {
                if (draw3D) {
                    bottomSpace = Math.max(bottomSpace, ax.getHeight());
                }
                else {
                    bottomSpace += ax.getHeight();
                }
            }
        }
    }

    // If plot margins are not default, then plot rect x and y are predetermined
    if (!getPlotArea().getLeftIsDefault()) {
        plotRect.x = getPlotArea().getLeft();
    }
    if (!getPlotArea().getTopIsDefault()) {
        plotRect.y = getPlotArea().getTop();
    }

    // start x and y's for positioning axes
    // if plot margins are defaults, these x's and y's are the edges of
    // the entire possible plot area.  If margins are specified, these values
    // are the edges of the specified plot area.
    int leftX = plotRect.x;
    int rightX =
        getPlotArea().getRightIsDefault() ? (plotRect.x + plotRect.width)
                                          : (chartAreaRect.width -
                                             getPlotArea().getRight());
    int topY = plotRect.y;
    int bottomY =
        getPlotArea().getBottomIsDefault() ? (plotRect.y + plotRect.height)
                                           : (chartAreaRect.height -
                                              getPlotArea().getBottom());

    // Position axes.
    // Edges of 3d axes are positioned at same spot on the plot area
    // Non-3d axes are position one after the other
    // If a margin is the default value, the axis is positioned inside
    //   the plot area and the plot area is later shrunk.
    // If a margin is not a default value, the axis is positioned wrt
    //   the plot area as defined by the margin value.

    for (int i = 0; i < axes.size(); i++) {
        ax = (JCAxis)(axes.get(i));
        // Adjust the plot rect for visible axes only.  If none of the
        // axes are visible, make adjustments for the first axis only.
        if ((anyAxisVisible && !ax.visible) ||
            (!anyAxisVisible && i > 0) ||
            pieFirst || radarFirst)
        {
            continue;
        }

        if (isCircularChartAxisOnly(ax)) {
            continue;
        }

        if (ax.i_vertical) {
            if (ax.isLeftSide()) {
                if (getPlotArea().getLeftIsDefault()) {
                    if (draw3D) {
                        ax.setLocation(leftX + leftSpace - ax.getWidth(),
                                       ax.getTop());
                    }
                    else {
                        ax.setLocation(leftX, ax.getTop());
                        leftX += ax.getWidth();
                    }
                }
                else {
                    if (draw3D) {
                        ax.setLocation(leftX - ax.getWidth(), ax.getTop());
                    }
                    else {
                        ax.setLocation(leftX - leftSpace, ax.getTop());
                        leftX += ax.getWidth();
                    }
                }
            }
            else if (ax.isRightSide()) {
                if (getPlotArea().getRightIsDefault()) {
                    if (draw3D) {
                        ax.setLocation(rightX - rightSpace, ax.getTop());
                    }
                    else {
                        ax.setLocation(rightX - rightSpace, ax.getTop());
                        rightX += ax.getWidth();
                    }
                }
                else {
                    if (draw3D) {
                        ax.setLocation(rightX, ax.getTop());
                    }
                    else {
                        ax.setLocation(rightX, ax.getTop());
                        rightX += ax.getWidth();
                    }
                }
            }
        }
        else {
            if (ax.isTopSide()) {
                if (getPlotArea().getTopIsDefault()) {
                    if (draw3D) {
                        ax.setLocation(ax.getLeft(),
                                       topY + topSpace - ax.getHeight());
                    }
                    else {
                        ax.setLocation(ax.getLeft(), topY);
                        topY += ax.getHeight();
                    }
                }
                else {
                    if (draw3D) {
                        ax.setLocation(ax.getLeft(), topY - ax.getHeight());
                    }
                    else {
                        ax.setLocation(ax.getLeft(), topY - topSpace);
                        topY += ax.getHeight();
                    }
                }
            }
            else if (ax.isBottomSide()) {
                if (getPlotArea().getBottomIsDefault()) {
                    if (draw3D) {
                        ax.setLocation(ax.getLeft(), bottomY - bottomSpace);
                    }
                    else {
                        ax.setLocation(ax.getLeft(), bottomY - bottomSpace);
                        bottomY += ax.getHeight();
                    }
                }
                else {
                    if (draw3D) {
                        ax.setLocation(ax.getLeft(), bottomY);
                    }
                    else {
                        ax.setLocation(ax.getLeft(), bottomY);
                        bottomY += ax.getHeight();
                    }
                }
            }
        }
    }

    // If plot margins are defaults, calculate them from the space used
    // by positioning axes.
    if (getPlotArea().getLeftIsDefault()) {
        plotRect.x += leftSpace;
        plotRect.width -= leftSpace;
    }
    if (getPlotArea().getRightIsDefault()) {
        plotRect.width -= rightSpace;
    }
    if (getPlotArea().getTopIsDefault()) {
        plotRect.y += topSpace;
        plotRect.height -= topSpace;
    }
    if (getPlotArea().getBottomIsDefault()) {
        plotRect.height -= bottomSpace;
    }
}

/**
 * Adjusts the plot rect and axis positions so that the biggest symbol
 * can be shown without it being clipped by the plot area.
 * @param horizAxes a list of horizontal axes
 * @param vertAxes a list of vertical axes
 */
private void adjustForSymbols(List horizAxes, List vertAxes)
{
    // if plot margins are defaults, then see if they need adjusting
    // to allow the largest possible symbol to be displayed without clipping
    if (getPlotArea().getLeftIsDefault()) {
        int diff = symbolPad - (plotRect.x - chartAreaRect.x);
        if (diff > 0) {
            plotRect.x += diff;
            plotRect.width -= diff;
            adjust(horizAxes, diff, 0, -diff, 0);
            adjust(vertAxes, diff, 0, 0, 0, JCAxis.MIN);
        }
    }
    if (getPlotArea().getRightIsDefault()) {
        int diff = symbolPad - (chartAreaRect.x + chartAreaRect.width -
                                plotRect.x - plotRect.width);
        if (diff > 0) {
            plotRect.width -= diff;
            adjust(horizAxes, 0, 0, -diff, 0);
            adjust(vertAxes, -diff, 0, 0, 0, JCAxis.MAX);
        }
    }
    if (getPlotArea().getTopIsDefault()) {
        int diff = symbolPad - (plotRect.y - chartAreaRect.y);
        if (diff > 0) {
            plotRect.y += diff;
            plotRect.height -= diff;
            adjust(vertAxes, 0, diff, 0, -diff);
            adjust(horizAxes, 0, diff, 0, 0, JCAxis.MAX);
        }
    }
    if (getPlotArea().getBottomIsDefault()) {
        int diff = symbolPad - (chartAreaRect.y + chartAreaRect.height -
                                plotRect.y - plotRect.height);
        if (diff > 0) {
            plotRect.height -= diff;
            adjust(vertAxes, 0, -diff, 0, -diff);
            adjust(horizAxes, 0, -diff, 0, 0, JCAxis.MIN);
        }
    }
}

/**
 * This method returns <code>true</code> if the specified axis is associated with
 * the first visible data view.
 * @return <code>true</code> if the specified axis is associated
 * with the first visible data view; <code>false</code> otherwise
 * @param ax
 */
private boolean inFirstVisibleDataView(JCAxis ax)
{
    JCChart chart = (JCChart) getParent();
    Iterator dataViews = chart.getOriginalDataView().iterator();
    if (dataViews != null) {
        ChartDataView cdv = null;
        while (dataViews.hasNext()) {
            cdv = (ChartDataView)dataViews.next();
            if (cdv != null && cdv.isVisible()) {
                // Having found the first visible data view...
                return(ax == cdv.getXAxis() || ax == cdv.getYAxis());
            }
        }
    }
    return(false);
}

/**
 * Internal method that positions the provided axis objects.
 * @param axes vector of axis objects to be positioned
 * @param last If <code>true</code>, this is the last axis layout.  This means that
 *     3D axes should be adjusted for the deltaX/Y values along their
 *     minor dimension (for example, move vertical axes by deltaX and move horizontal
 *     axes by deltaY).
 */
private void layoutAxes(List axes, boolean last)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)(axes.get(i));
        // Don't layout the axis if the associated data view is of type
        // POLAR, RADAR, or AREA_RADAR, and the data view isn't the
        // first visible one.  If the first visible data view is a rectangular
        // chart, and a subsequent data view is POLAR, RADAR, or AREA_RADAR,
        // we get trapped in an infinite loop (since ax.layoutPolarAxis()
        // calls chartArea.recalc()).
        if (!inFirstVisibleDataView(ax) && (ax.isPolar() || ax.isRadar())) {
            continue;
        }
        ax.layoutAxis(last);
    }
}

private void layoutOrigins(List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)(axes.get(i));
        if (!inFirstVisibleDataView(ax) && (ax.isPolar() || ax.isRadar())) {
            continue;
        }
        ax.layoutOrigin();
    }
}

/**
 * Internal method that adjusts the annotations on the
 * provided axis objects.
 * @param axes vector of axis objects to be examined
 */
void checkAnnotations(List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)(axes.get(i));
        ax.checkAnnotations();
    }
}

/**
 * Internal method that guaratees that a drawable exists for
 * all attached <code>ChartDataView</code> objects.
 */
private void ensureDrawables()
{
    JCChart chart = (JCChart) getParent();
    if (chart == null) {
        return;
    }
    if (drawables == null) {
        drawables = new Hashtable();
    }
    axisDrawOkay = true;
    pieFirst = false;
    polarFirst = false;
    radarFirst = false;
    int dvNumber = 0;
    for (Iterator i = chart.getOriginalDataView().iterator(); i.hasNext();) {
        ChartDataView dv = (ChartDataView)i.next();
        if (dv.isVisible()) {
            dvNumber++;
            if (dvNumber == 1) {
                if (dv.getChartType() == JCChart.PIE) {
                    pieFirst = true;
                    axisDrawOkay = false;
                }
                else if (dv.getChartType() == JCChart.POLAR) {
                    polarFirst = true;
                }
                else if (dv.getChartType() == JCChart.RADAR ||
                         dv.getChartType() == JCChart.AREA_RADAR)
                {
                    radarFirst = true;
                }
            }
        }
        if (!hasMatchingDrawable(dv) && dv.isVisible()) {
            addDrawable(dv);
        }
        else if (hasMatchingDrawable(dv) && !dv.isVisible()) {
            removeDrawable(dv);
        }
        //if (dv.getChartType() != JCChart.PIE ) axisDrawOkay = true;
    }
}

/**
 * Calculates all axis parameters for the provided axes. Equivalent
 * to <code>xrt_calc_axis_parms()</code> in Xrt/graph.
 * @param axes vector of axis objects
 */
private void calcAxisParams(List axes)
{
    if (axes == null) {
        return;
    }

    // Loop through data objects and calculate axis parameters for
    // the data object.
    boolean hasRelated = false;
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.formula != null) {
            hasRelated = true;
        }
        ax.calcAxisParams(true);
    }
    // Return early if no related axes were detected.
    if (!hasRelated) {
        return;
    }

    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax.formula != null) {
            ax.formula.calcRelatedParams();
        }
    }
}

/**
 * Creates annotations for the provided axis objects.
 * @param axes list of axis objects
 */
private void makeAnnotations(List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        ax.makeAnnotations();
    }
}

/**
 * Calculates the size of the axes.
 * @param axes the list of axes whose size is to be calculated
 */
private void calcAnnotationExtents(List axes)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        ax.calcAnnotationExtents();
    }
}

private void updateAnnotationsAndGrids(List axes) {
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (ax != null) {
            ax.updateAnnotations();
            ax.updateGrids();
        }
    }
}

/**
 * Calculates the position and size of the plot area.
 */
private void calcGraphExtents()
{
    if (drawables == null) {
        return;
    }

    // Calculate graph extents

    // If first data view is PIE, do only it.
    if (pieFirst) {
        Enumeration e = drawables.keys();
        while (e.hasMoreElements()) {
            ChartDataView key = (ChartDataView) e.nextElement();
            if (key.getChartType() == JCChart.PIE) {
                ChartDraw d = (ChartDraw)drawables.get(key);
                d.recalc();
                break;
            }
        }
    }
    else if (radarFirst) {
        // If first data view is RADAR, do only it.
        Enumeration e = drawables.keys();
        while (e.hasMoreElements()) {
            ChartDataView key = (ChartDataView) e.nextElement();
            int chartType = key.getChartType();
            if ((chartType == JCChart.RADAR) ||
                (chartType == JCChart.AREA_RADAR))
            {
                ChartDraw d = (ChartDraw)drawables.get(key);
                d.recalc();
                break;
            }
        }
    }
    else if (polarFirst) {
        // If first data view is POLAR, do all dataviews of type POLAR.
        Enumeration e = drawables.keys();
        while (e.hasMoreElements()) {
            ChartDataView key = (ChartDataView) e.nextElement();
            int chartType = key.getChartType();
            if (chartType == JCChart.POLAR)
            {
                ChartDraw d = (ChartDraw)drawables.get(key);
                d.recalc();
            }
        }
    }
    else {
        // Otherwise, do all rectangular charts
        Enumeration e = drawables.keys();
        while (e.hasMoreElements()) {
            ChartDataView key = (ChartDataView) e.nextElement();
            int chartType = key.getChartType();
            if (chartType != JCChart.PIE &&
                chartType != JCChart.POLAR &&
                chartType != JCChart.RADAR &&
                chartType != JCChart.AREA_RADAR)  // if not circular chart
            {
                ChartDraw d = (ChartDraw)drawables.get(key);
                d.recalc();
            }
        }
    }
}

///**
// * Recalculates the transformations for the axes.
// * @param axes the list of axes whose tranformation needs to be calculated
// */
//private void calcTransformation(List axes)
//{
//    if (axes == null) {
//      return;
//    }
//    for (int i = 0; i < axes.size(); i++) {
//        JCAxis ax = (JCAxis)axes.get(i);
//        ax.calcTransformation();
//    }
//}

/**
 * Adjusts the plot area and axes for axis extents.
 * @param horizAxes the list of horizontal axes
 * @param vertAxes the list of vertical axes
 */
private void adjustForExtents(List horizAxes, List vertAxes)
{
    Extents x = new Extents();
    Extents y = new Extents();
    findMaxExtents(horizAxes, x);
    findMaxExtents(vertAxes, y);

    // Adjust for x axis extents first
    int offset = x.minor - plotRect.x;
    if (offset > 0 && plotArea.left.isDefault) {
        if (offset > plotRect.width) {
            offset = plotRect.width;
        }
        plotRect.width -= offset;
        plotRect.x += offset;
        adjust(horizAxes, offset, 0, -offset, 0);
        adjust(vertAxes, offset, 0, 0, 0, JCAxis.MIN);
    }

    offset = x.major - (chartAreaRect.width - plotRect.x - plotRect.width);
    if (offset > 0 && plotArea.right.isDefault) {
        if (offset > plotRect.width) {
            offset = plotRect.width;
        }
        plotRect.width -= offset;
        adjust(horizAxes, 0, 0, -offset, 0);
        adjust(vertAxes, -offset, 0, 0, 0, JCAxis.MAX);
    }

    // Now adjust for y axis extents
    offset = y.major - (chartAreaRect.height - plotRect.y - plotRect.height);
    if (offset > 0 && plotArea.bottom.isDefault) {
        if (offset > plotRect.height) {
            offset = plotRect.height;
        }
        plotRect.height -= offset;
        adjust(vertAxes, 0, 0, 0, -offset);
        adjust(horizAxes, 0, -offset, 0, 0, JCAxis.MIN);
    }

    offset = y.minor - plotRect.y;
    if (offset > 0 && plotArea.top.isDefault) {
        if (offset > plotRect.height) {
            offset = plotRect.height;
        }
        plotRect.y += offset;
        plotRect.height -= offset;
        adjust(vertAxes, 0, offset, 0, -offset);
        adjust(horizAxes, 0, offset, 0, 0, JCAxis.MAX);
    }
}

private void adjustForOrthoExtents(List horizAxes, List vertAxes)
{
    Rectangle old = new Rectangle(plotRect.x, plotRect.y, plotRect.width,
                                  plotRect.height);
    for (int i = 0; i < horizAxes.size(); i++) {
        ((JCAxis)horizAxes.get(i)).adjustForOrthoExtent();
    }
    for (int i = 0; i < vertAxes.size(); i++) {
        ((JCAxis)vertAxes.get(i)).adjustForOrthoExtent();
    }
    if (plotRect.equals(old)) {
        return;
    }
    adjust(vertAxes, 0, plotRect.y-old.y, 0, plotRect.height-old.height);
    adjust(horizAxes, plotRect.x-old.x, 0, plotRect.width-old.width, 0);
    layoutOrigins(horizAxes);
    layoutOrigins(vertAxes);
}


private void findMaxExtents(List axes, Extents ex)
{
    if (axes == null) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        Extents axex = ax.getExtents();
        if (axex == null) {
            continue;
        }
        if (axex.major > ex.major) {
            ex.major = axex.major;
        }
        if (axex.minor > ex.minor) {
            ex.minor = axex.minor;
        }
        if (axex.ortho > ex.ortho) {
            ex.ortho = axex.ortho;
        }
    }
}

private void adjust(List axes, int x, int y, int w, int h)
{
    adjust(axes, x, y, w, h, JCAxis.AUTOMATIC);
}

private void adjust(List axes, int xoff, int yoff, int woff, int hoff,
                    int condition)
{
    if (axes == null) {
        return;
    }
    if (xoff == 0 && yoff == 0 && woff == 0 && hoff == 0) {
        return;
    }
    for (int i = 0; i < axes.size(); i++) {
        JCAxis ax = (JCAxis)axes.get(i);
        if (condition != JCAxis.AUTOMATIC &&
            condition != ax.getAdjustedRealPlacement())
        {
            continue;
        }
        int x = ax.getLeft() + xoff;
        int y = ax.getTop() + yoff;
        int w = ax.getWidth() + woff;
        int h = ax.getHeight() + hoff;
        ax.setBounds(x, y, w, h);
    }
}

/**
 * Calculates the delta x and y of the 3D shadow.
 */
private void calc3D()
{
    double nrotation = JCChartUtil.degToRad(rotation);
    double nelevation = JCChartUtil.degToRad(elevation);
    double sin_rotation = Math.sin(nrotation);
    double sin_elevation = Math.sin(nelevation);
    double cos_elevation = Math.cos(nelevation);
    double factor = depth / 100.0;
    double dwidth = JCChartUtil.max(1, getSize().width);
    dwidth /= (1 + JCChartUtil.abs(sin_rotation) * cos_elevation * factor);
    double depth = dwidth * factor;
    deltaX = (int) (depth * sin_rotation * cos_elevation);
    deltaY = (int) (-depth * sin_elevation);

    checkDeltaValues();

/*
 * Checks whether the chartArea should draw in 3d or not.  If there are only
 * Plot (or it's cousin Area) charts set, it will not draw in 3d no matter
 * what the values of deltaX * and deltaY.  Otherwise, 3d drawing depends
 * on deltaX and Y being != 0.
 * Hilo, HiloOpenClose, and Candle chart types do not support 3d drawing.
 */
    boolean onlyPlot = true;
    boolean firstView = true;

    JCChart chart = (JCChart) getParent();
    if (chart != null) {
        List cdvL = chart.getDataView();
        if (cdvL != null) {
            for (Iterator i = cdvL.iterator(); i.hasNext();) {
                ChartDataView cdv = (ChartDataView) i.next();
                if (!cdv.isVisible()) {
                    continue;
                }

                int type = cdv.getChartType();

                // If type is other than plot-based types,
                // then onlyPlot is false
                if (type != JCChart.PLOT &&
                    type != JCChart.SCATTER_PLOT &&
                    type != JCChart.STACKING_AREA &&
                    type != JCChart.AREA &&
                    type != JCChart.HILO &&
                    type != JCChart.HILO_OPEN_CLOSE &&
                    type != JCChart.CANDLE &&
                    type != JCChart.POLAR &&
                    type != JCChart.RADAR &&
                    type != JCChart.AREA_RADAR)
                {
                    // Ignore the case where a pie exists outside of
                    // the first data view
                    if (!(type == JCChart.PIE && !firstView)) {
                        onlyPlot = false;
                        break;
                    }
                }

                if (firstView) {
                    firstView = false;
                }
            }
        }
    }

    boolean deltas = (deltaX != 0 || deltaY != 0);

    hasBar = !onlyPlot;

    // 3d drawing depends on not being a plot-only chart and having deltas
    draw3D = (!onlyPlot && deltas);
}

private boolean checkDeltaValues()
{
    boolean chng = false;

    // no checking if plotRect is not initialized
    if (plotRect.width <= 0 || plotRect.height <= 0) {
        return(chng);
    }

    // ensure the delta values stay within limits
    int wh = plotRect.width;
    if (Math.abs(deltaX) > wh) {
        deltaX = (deltaX < 0) ? -wh : wh;
        chng = true;
    }
    wh = plotRect.height;
    if (Math.abs(deltaY) > wh) {
        deltaY = (deltaY < 0) ? -wh : wh;
        chng = true;
    }
    return(chng);

}

private Rectangle clipToEntirePlotRect()
{
    // If there is an axis bounding box (non-3d only), don't overwrite it.
    int x, y, w, h;
    if (axisBoundingBox && !draw3D) {
        x = plotRect.x + 1;
        y = plotRect.y + 1;
        w = plotRect.width - 1;
        h = plotRect.height - 1;
    }
    else {
        x = plotRect.x;
        y = plotRect.y;
        w = plotRect.width + 1;
        h = plotRect.height + 1;
    }
    x = Math.max(0, x);
    y = Math.max(0, y);
    w = Math.min(w, chartAreaRect.width-x);
    h = Math.min(h, chartAreaRect.height-y);
    return(new Rectangle(x, y, w, h));
}

Rectangle clipForDrawing()
{
    int ps = symbolPad * 2;
    int x = Math.max(0, plotRect.x-ps/2);
    int y = Math.max(0, plotRect.y-ps/2);
    int w = Math.min(plotRect.width+1+ps, chartAreaRect.width-x);
    int h = Math.min(plotRect.height+1+ps, chartAreaRect.height-y);
    return(new Rectangle(x, y, w, h));
}

/*
private Graphics clipToPlotRect(Graphics gc)
{
    Graphics tempGC = gc.create();
    if (vertical) {
        tempGC.clipRect(0, plotRect.y,
                        getDrawingArea().width+1,
                        plotRect.height+1);
    }
    else {
        tempGC.clipRect(plotRect.x, 0 ,
                        plotRect.width+1,
                        getDrawingArea().height+1);
    }
    return(tempGC);
}
*/


 

/* previous jprobe checking codes
private static JCString msg = null;

private final void makeExpiredMessage()
{
    JCChart chart = (JCChart) getParent();
    Image im = JClassInfo.makeKLImage(chart);
    msg = JCString.parse(chart, "This JClass Chart may only run inside JProbe Profiler.[NEWLINE]JClass Chart can be purchased from [HREF=http://www.quest.com]Quest Software[/HREF][NEWLINE]");
    msg.add(im);
    msg.add(JCString.VERT_SPACE);
    msg.add(20);
    msg.add(JCString.NEWLINE);
    msg.add(new String("The Tools to Use!"));
}

private final void drawExpired(Graphics gc)
{
    Rectangle r = getDrawingArea();
    gc.drawRect(r.x+1, r.y+1, r.width-2, r.height-2);
    if (msg == null) makeExpiredMessage();
    Dimension sz = msg.getSize(getParent(), getFont());
    int y = (r.y + (r.height - sz.height)/2);
    r.y = y;
    msg.draw(getParent(), gc, r, JCString.ALIGN_CENTER);
}
 */

private void resetPlotArea()
{
    // Calculate original plot area size
    chartAreaRect = getDrawingArea();

    if (plotArea.top.isDefault) {
        plotArea.top.value = 0;
    }
    if (plotArea.left.isDefault) {
        plotArea.left.value = 0;
    }
    if (plotArea.bottom.isDefault) {
        plotArea.bottom.value = 0;
    }
    if (plotArea.right.isDefault) {
        plotArea.right.value = 0;
    }
}


/**
 * Overrides parent <code>getGraphics()</code>. If another <code>Graphics</code> object has been
 * set to be the current one, it is returned; otherwise, the superclass's
 * <code>getGraphics()</code> method is called.
 * @return the <code>Graphics</code> object to use to draw this chart area
 */
public Graphics getGraphics()
{
    if (prevSetGraphics == null) {
        return(super.getGraphics());
    }
    // make sure to return a copy of the user-set graphics object
    // this is in keeping with the contract for getGraphics()
    // some Java code calls dispose() on the object returned from this method
    return(prevSetGraphics.create());
}

/**
 * Provides a way for the chart area to draw to a custom <code>Graphics</code> object.
 * Call with <code>null</code> to specify use of the default <code>Graphics</code> object used
 * to draw to the screen.
 * @param g <code>Graphics</code> object to use for drawing
 */
public void setGraphics(Graphics g)
{
    prevSetGraphics = g;
}

}
