/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartDateFormat.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.


package com.klg.jclass.chart;


import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.*;

/**
 * JCChartDateFormat is a collection of methods and variables
 * associated with time and date that are used throughout JCChart
 * This class provides capabilities for writing internationalized code.
 */
public class JCChartDateFormat {

//private final static String engEraSym = "G";
private final static String engYearSym = "y";
private final static String engMonthSym = "M";
private final static String engDaySym = "dEDFwW";
private final static String engHourSym = "ahHkKz";
private final static String engMinuteSym = "m";
private final static String engSecondSym = "s";
//private final static String engMillisecondSym = "S";

private String            secondSym = null;
private String            minuteSym = null;
private String            hourSym = null;
private String            daySym = null;
private String            monthSym = null;
private String            yearSym = null;

private Calendar          calendar = null;
private Locale            locale = Locale.getDefault();
private TimeZone          timeZone = TimeZone.getDefault();
private int               dateStyle = DateFormat.SHORT;
private int               timeStyle = DateFormat.SHORT;
private SimpleDateFormat  formatter = null;
private SimpleDateFormat  longFormatter = null;
private String            largestTimeLabel = null;
private String            currentPattern = "";

private static Date       minTime = null;
private static Date       maxTime = null;

public JCChartDateFormat()
{
    init();
}

/**
 * Set the locale for formatting.  This is called by JCAxis.setLocale().
 * @param locale The new locale
 * @see JCAxis#setLocale
 */
protected void setLocale(Locale locale)
{
    if (locale == null) {
        locale = Locale.getDefault();
    }
    this.locale = locale;
    init();
}

/**
 * Set the time zone for formatting.  This called by JCAxis.setTimeZone().
 * @param timeZone The new time zone
 * @see JCAxis#setTimeZone
 */
protected void setTimeZone(TimeZone timeZone)
{
    if (timeZone == null) {
        timeZone = TimeZone.getDefault();
    }
    this.timeZone = timeZone;
    init();
}

/**
 * Set the date style for formatted time label
 * @param dateStyle The new date style
 */
public void setDateStyle(int dateStyle)
{
    if (dateStyle < DateFormat.FULL || dateStyle > DateFormat.SHORT) {
        throw new IllegalArgumentException("Illegal date style " + dateStyle);
    }
    this.dateStyle = dateStyle;
    init();
}

/**
 * Retrieve the current date style for the formatted time label.
 * @return The current date style.
 */
public int getDateStyle()
{
    return(dateStyle);
}

/**
 * Set the time style for formatted time label
 * @param timeStyle The new time style
 */
public void setTimeStyle(int timeStyle)
{
    if (timeStyle < DateFormat.FULL || timeStyle > DateFormat.SHORT) {
        throw new IllegalArgumentException("Illegal time style " + timeStyle);
    }
    this.timeStyle = timeStyle;
    init();
}

/**
 * Retrieve the current time style for the formatted time label.
 * @return The current time style.
 */
public int getTimeStyle()
{
    return(timeStyle);
}

private void init()
{
    calendar = Calendar.getInstance(timeZone, locale);
    formatter = (SimpleDateFormat)DateFormat.getDateTimeInstance(dateStyle,
                                                                 timeStyle,
                                                                 locale);
    formatter.setCalendar(calendar);

    longFormatter =
        (SimpleDateFormat)DateFormat.getDateTimeInstance(DateFormat.FULL,
                                                         DateFormat.FULL,
                                                         locale);
    longFormatter.setCalendar(calendar);

    // Calculate the largest time label for this locale and timezone.  This has
    // the side effect of resetting the currentPattern.
    currentPattern = "";
    largestTimeLabel = getLargestLocaleTimeLabel();

    secondSym = getLocalizedPattern(engSecondSym);
    minuteSym = getLocalizedPattern(engMinuteSym);
    hourSym = getLocalizedPattern(engHourSym);
    daySym = getLocalizedPattern(engDaySym);
    monthSym = getLocalizedPattern(engMonthSym);
    yearSym = getLocalizedPattern(engYearSym);
}

/**
 * Given an instance of Date, returns a formatted time label
 * using the default locale, current dateStyle, and current timeStyle.
 * @param date the Date to be localized
 */
protected String timeLabel(Date date)
{
    return(formatter.format(date));
}

/**
 * Given a format style and an instance of Date,
 * returns a formatted time label using the default locale.
 * The style of formatting can have on of the followig values:
 * DateFormat.SHORT, DateFormat.MEDIUM, DateFormat.LONG or DateFormat.FULL
 * @param style the format style to be used in date/time localization and
 * can have one of the following values
 * @param date the Date to be localized
 */
protected static String timeLabel(int style, Date date, Locale l)
{
    String formattedTime = null;
    Calendar c = Calendar.getInstance(l);

    switch (style) {
    case DateFormat.SHORT:
    case DateFormat.MEDIUM:
    case DateFormat.LONG:
    case DateFormat.FULL:
        DateFormat styleFormatter = DateFormat.getDateTimeInstance(style,
                                                                   style, l);

        styleFormatter.setCalendar(c);
        formattedTime = styleFormatter.format(date);
        break;
    default:
        formattedTime = date.toString();
    }
    return(formattedTime);
}

/**
 * Given a format pattern and an instance of Date,
 * returns a formatted time label using the default locale<br>
 * <strong>Time Format Syntax:</strong><br>
 * To specify the time format use a time pattern string. In this pattern,
 * all ASCII letters are reserved as pattern letters, which are defined as the
 * following:
 * <br>
 * <TABLE BORDER=1 WIDTH=100%>
 * <TR>
 * <TH>Symbol</TH>
 * <TH>Meaning</TH>
 * <TH>Presentation</TH>
 * <TH>Example</TH>
 * </TR>
 * <TR>
 * <TD>G</TD><TD>era designator</TD><TD>(Text)</TD><TD>AD</TD></TR>
 * <TD>y</TD><TD>year</TD><TD>(Number)</TD><TD>1997</TD></TR>
 * <TD>M</TD><TD>month in year</TD><TD>(Text & Number)</TD>
 *           <TD>July & 07</TD></TR>
 * <TD>d</TD><TD>day in month</TD><TD>(Number)</TD><TD>10</TD></TR>
 * <TD>h</TD><TD>hour in am/pm (1~12)</TD><TD>(Number)</TD><TD>12</TD></TR>
 * <TD>H</TD><TD>hour in day (0~23)</TD><TD>(Number)</TD><TD>0</TD></TR>
 * <TD>m</TD><TD>minute in hour</TD><TD>(Number)</TD><TD>30</TD></TR>
 * <TD>s</TD><TD>second in minute</TD><TD>(Number)</TD><TD>55</TD></TR>
 * <TD>S</TD><TD>millisecond</TD><TD>(Number)</TD><TD>978</TD></TR>
 * <TD>E</TD><TD>day in week</TD><TD>(Text)</TD><TD>Tuesday</TD></TR>
 * <TD>D</TD><TD>day in year</TD><TD>(Number)</TD><TD>189</TD></TR>
 * <TD>F</TD><TD>day of week in month</TD><TD>(Number)</TD>
 *           <TD>2 (2nd Wed in July)</TD></TR>
 * <TD>w</TD><TD>week in year</TD><TD>(Number)</TD><TD>27</TD></TR>
 * <TD>W</TD><TD>week in month</TD><TD>(Number)</TD><TD>2</TD></TR>
 * <TD>a</TD><TD>am/pm marker</TD><TD>(Text)</TD><TD>PM</TD></TR>
 * <TD>k</TD><TD>hour in day (1~24)</TD><TD>(Number)</TD><TD>24</TD></TR>
 * <TD>K</TD><TD>hour in am/pm (0~11)</TD><TD>(Number)</TD><TD>0</TD></TR>
 * <TD>z</TD><TD>time zone</TD><TD>(Text)</TD>
 *           <TD>Pacific Standard Time</TD></TR>
 * <TD>'</TD><TD>escape for text</TD><TD>(delimiter)</TD><TD></TD></TR>
 * <TD>''</TD><TD>single quote</TD><TD>(Literal)</TD><TD></TD></TR>
 *</TABLE>
 */
protected String timeLabel(String pattern, Date date)
{
    String parsedDate = null;

    try {
        formatter.applyLocalizedPattern(getLocalizedPattern(pattern));
        parsedDate = formatter.format(date);
    }
    catch (Exception e) {
        e.printStackTrace();
    }
    return(parsedDate);
}

/*
static final int MIN_YEAR = 1969;
static final int MAX_YEAR = 2107;


static Date getMinTime() {
Calendar c = Calendar.getInstance();

c.set( MIN_YEAR, 0, 0, 0, 0, 0 );
return c.getTime();
}

static Date getMaxTime() {
Calendar c = Calendar.getInstance();

c.set( MAX_YEAR, 0, 0, 0, 0, 0 );
return c.getTime();
}
*/

/**
 * Returns the earliest date that can be represented in Chart
 */
protected static Date getMinTime()
{
    if (minTime == null) {

        Calendar c = Calendar.getInstance();

        c.set(c.getMinimum(Calendar.YEAR), c.getMinimum(Calendar.MONTH),
              c.getMinimum(Calendar.DATE), c.getMinimum(Calendar.HOUR_OF_DAY),
              c.getMinimum(Calendar.MINUTE), c.getMinimum(Calendar.SECOND));
        minTime = c.getTime();
    }
    return(minTime);
}

/**
 * Returns the latest date that can be represented in Chart
 */
protected static Date getMaxTime()
{
    if (maxTime == null) {
        Calendar c = Calendar.getInstance();
        c.set(c.getMaximum(Calendar.YEAR) - 1, c.getMaximum(Calendar.MONTH),
              c.getMaximum(Calendar.DATE), c.getMaximum(Calendar.HOUR_OF_DAY),
              c.getMaximum(Calendar.MINUTE), c.getMaximum(Calendar.SECOND));
        maxTime = c.getTime();
    }
    return(maxTime);
}

protected String getLargestTimeLabel(String pattern)
{
    calculateLargestTimeLabel(pattern);
    return(largestTimeLabel);
}

/**
 * Calculates and returns the largest time label based on a pattern
 * @param pattern Time label format
 */
protected void calculateLargestTimeLabel(String pattern)
{
    if (pattern.compareTo(currentPattern) == 0) {
        return;
    }

    currentPattern = pattern;

    String[] months = longFormatter.getDateFormatSymbols().getMonths();
    int longmonth = 0;
    int maxlen = 0;

    for (int m = 0; m < 12; m++) {
        int len = months[m].length();
        if (maxlen < len) {
            maxlen = len;
            longmonth = m;
        }
    }

    String[] weekdays = longFormatter.getDateFormatSymbols().getWeekdays();
    int longday = 0;

    maxlen = 0;
    for (int dy = 1; dy <= 7; dy++) {
        int len = weekdays[dy].length();
        if (maxlen < len) {
            maxlen = len;
            longday = dy;
        }
    }

    calendar.set(Calendar.MONTH, longmonth);
    calendar.set(Calendar.DAY_OF_MONTH, 1);
    for (;
         calendar.get(Calendar.DAY_OF_WEEK) != longday;
         calendar.roll(Calendar.DAY_OF_MONTH, true)) ;

    formatter.applyLocalizedPattern(getLocalizedPattern(pattern));
    largestTimeLabel = formatter.format(calendar.getTime());
    calendar.setTime(new Date()); //reseting the calendar
}

/**
 * Calculates and returns the largest time label based on the default locale
 */
protected String getLargestLocaleTimeLabel()
{
    return getLargestTimeLabel(longFormatter.toPattern());
}

/**
 * Round the given time down/up to the given unit.  For example, if the unit is
 * months, Feb 6 gets adjusted to Feb 1 00:00:00.
 * @param d The date to be rounded.
 * @param unit The unit of time (seconds, minutes, hours, days, etc)
 * @param adjust If equal to 0, round down.  If equal to 1, round up.
 */
protected void roundTime(Date d, long unit, boolean adjust)
{
    int a = adjust ? 1 : 0;

    calendar.setTime(d);
    if (unit < JCAxis.SECONDS) {
        // Rounding to the second
        int val = calendar.get(Calendar.MILLISECOND);
        calendar.set(Calendar.MILLISECOND, (((val / 1000) + a) * 1000));
    }
    else {
        // Set to be at beginning of the second
        calendar.set(Calendar.MILLISECOND, 0);
        if (unit < JCAxis.MINUTES) {
            // Rounding to seconds
            unit = unit / JCAxis.SECONDS;
            int val = calendar.get(Calendar.SECOND);
            calendar.set(Calendar.SECOND, (int)(((val / unit) + a) * unit));
        }
        else {
            // Set to be at beginning of the minute
            calendar.set(Calendar.SECOND, 0);
            if (unit < JCAxis.HOURS) {
                // Rounding to minutes
                unit = unit / JCAxis.MINUTES;
                int val = calendar.get(Calendar.MINUTE);
                calendar.set(Calendar.MINUTE, (int)(((val / unit) + a) * unit));
            }
            else {
                // Set to be at beginning of the hour
                calendar.set(Calendar.MINUTE, 0);
                if (unit < JCAxis.DAYS) {
                    // Rounding to hours
                    unit = unit / JCAxis.HOURS;
                    int val = calendar.get(Calendar.HOUR_OF_DAY);
                    calendar.set(Calendar.HOUR_OF_DAY,
                                 (int)(((val / unit) + a) * unit));
                }
                else {
                    // Set to be at beginning of the day
                    calendar.set(Calendar.HOUR_OF_DAY, 0);
                    if (unit < JCAxis.MONTHS) {
                        // Rounding to days
                        if (adjust) {
                            int val = calendar.get(Calendar.DAY_OF_MONTH);
                            calendar.set(Calendar.DAY_OF_MONTH,
                                         (int)(val + unit / JCAxis.DAYS));
                        }
                    }
                    else {
                        // First day of the month
                        calendar.set(Calendar.DAY_OF_MONTH, 1);
                        if (unit < JCAxis.YEARS) {
                            // Rounding to months
                            if (adjust) {
                                int val = calendar.get(Calendar.MONTH);
                                calendar.set(Calendar.MONTH,
                                             (int)(val + unit / JCAxis.MONTHS));
                            }
                        }
                        else {
                            // First month of the year
                            calendar.set(Calendar.MONTH, 0);
                            // Rounding to years
                            if (adjust) {
                                int val = calendar.get(Calendar.YEAR);
                                calendar.set(Calendar.YEAR,
                                             (int)(val + unit / JCAxis.YEARS));
                            }
                        }
                    }
                }
            }
        }
    }
    d.setTime(calendar.getTime().getTime());
    calendar.setTime(new Date()); //reseting the calendar
}

/**
 *  Return the value in seconds of the smallest "nice"
 *  time increment greater than 'incr'.  The increment
 *	should also be of the same magnitude as indicated
 *  by 'format' (eg. a "%m" format should not return an
 *	increment less than MONTH)
 */
protected long niceInc(long incr, String pattern)
{
    long mult = getMaxTime().getTime();

    String localizedPattern = getLocalizedPattern(pattern);

    if (matchPattern(localizedPattern, secondSym)) {
        mult = JCAxis.SECONDS;
    }
    else if (matchPattern(localizedPattern, minuteSym)) {
        mult = Math.min(mult, JCAxis.MINUTES);
    }
    else if (matchPattern(localizedPattern, hourSym)) {
        mult = Math.min(mult, JCAxis.HOURS);
    }
    else if (matchPattern(localizedPattern, daySym)) {
        mult = Math.min(mult, JCAxis.DAYS);
    }
    else if (matchPattern(localizedPattern, monthSym)) {
        mult = Math.min(mult, JCAxis.MONTHS);
    }
    else if (matchPattern(localizedPattern, yearSym)) {
        mult = Math.min(mult, JCAxis.YEARS);
    }

    if (mult == getMaxTime().getTime()) { /* unrecognized format */
        mult = 1;
    }

    /* 1, 5, 10, 15, 30 Seconds */

    if (mult < JCAxis.MINUTES) {
        if (incr <= 1 * JCAxis.SECONDS) {
            return(JCAxis.SECONDS);
        }
        if (incr <= 5 * JCAxis.SECONDS) {
            return(5 * JCAxis.SECONDS);
        }
        if (incr <= 10 * JCAxis.SECONDS) {
            return(10 * JCAxis.SECONDS);
        }
        if (incr <= 15 * JCAxis.SECONDS) {
            return(15 * JCAxis.SECONDS);
        }
        if (incr <= 30 * JCAxis.SECONDS) {
            return(30 * JCAxis.SECONDS);
        }
        mult = JCAxis.MINUTES;
    }

    /* 1, 5, 10, 15, 30 Minutes */
    if (mult < JCAxis.HOURS) {
        if (incr <= mult) {
            return(mult);
        }
        if (incr <= 5 * mult) {
            return(5 * mult);
        }
        if (incr <= 10 * mult) {
            return(10 * mult);
        }
        if (incr <= 15 * mult) {
            return(15 * mult);
        }
        if (incr <= 30 * mult) {
            return(30 * mult);
        }
        mult = JCAxis.HOURS;
    }

    /* 1, 3, 6, 12 Hours */
    if (mult < JCAxis.DAYS) {
        if (incr <= mult) {
            return(mult);
        }
        if (incr <= 3 * mult) {
            return(3 * mult);
        }
        if (incr <= 6 * mult) {
            return(6 * mult);
        }
        if (incr <= 12 * mult) {
            return(12 * mult);
        }
        mult = JCAxis.DAYS;
    }

    /* 1, 2, 7, 14 Days */
    if (mult < JCAxis.MONTHS) {
        if (incr <= mult) {
            return(mult);
        }
        if (incr <= 2 * mult) {
            return(2 * mult);
        }
        if (incr <= 7 * mult) {
            return(7 * mult);
        }
        if (incr <= 14 * mult) {
            return(14 * mult);
        }
        mult = JCAxis.MONTHS;
    }

    /* 1, 2, 3, 4, 6 Months */
    if (mult < JCAxis.YEARS) {
        if (incr <= mult) {
            return(mult);
        }
        if (incr <= 2 * mult) {
            return(2 * mult);
        }
        if (incr <= 3 * mult) {
            return(3 * mult);
        }
        if (incr <= 4 * mult) {
            return(4 * mult);
        }
        if (incr <= 6 * mult) {
            return(6 * mult);
        }
        mult = JCAxis.YEARS;
    }

    /* 1, 2, 5, 10, ... Years */
    if (incr <= mult) {
        return(mult);
    }
    if (incr <= 2 * mult) {
        return(2 * mult);
    }
    if (incr <= 5 * mult) {
        return(5 * mult);
    }
    if (incr <= 10 * mult) {
        return(10 * mult);
    }
    return(20 * mult);
}


protected static long timeIncUnits(long inc)
{
    if (inc < JCAxis.MINUTES) {
        return(JCAxis.SECONDS);
    }
    else if (inc < JCAxis.HOURS) {
        return(JCAxis.MINUTES);
    }
    else if (inc < JCAxis.DAYS) {
        return(JCAxis.HOURS);
    }
    else if (inc < JCAxis.MONTHS) {
        return(JCAxis.DAYS);
    }
    else if (inc < JCAxis.YEARS) {
        return(JCAxis.MONTHS);
    }
    return(JCAxis.YEARS);
}

/**
 * Adds a specified value to a particular unit in a Date instance.
 * @param d Date instance to be modified
 * @param unit Time unit to which to add
 * @param value Value to add to the specified unit
 */
protected void addToUnit(Date d, long unit, int value)
{
    calendar.setTime(d);
    if (unit < JCAxis.SECONDS) {
        calendar.add(Calendar.MILLISECOND, value);
    }
    else if (unit < JCAxis.MINUTES) {
        calendar.add(Calendar.SECOND, value);
    }
    else if (unit < JCAxis.HOURS) {
        calendar.add(Calendar.MINUTE, value);
    }
    else if (unit < JCAxis.DAYS) {
        calendar.add(Calendar.HOUR_OF_DAY, value);
    }
    else if (unit < JCAxis.MONTHS) {
        calendar.add(Calendar.DAY_OF_MONTH, value);
    }
    else if (unit < JCAxis.YEARS) {
        calendar.add(Calendar.MONTH, value);
    }
    else {
        calendar.add(Calendar.YEAR, value);
    }
    // And now the sanity checks
    if (calendar.getTime().getTime() < getMinTime().getTime()) {
        calendar.setTime(getMinTime());
    }
    if (calendar.getTime().getTime() > getMaxTime().getTime()) {
        calendar.setTime(getMaxTime());
    }
    d.setTime(calendar.getTime().getTime());
    calendar.setTime(new Date()); //reseting the calendar
}

/**
 * Internal method to return an appropriate time format
 * for the provided range
 */
protected static String getDefaultTimeFormat(double range)
{
    String format = null;

    if (range > 2 * JCAxis.YEARS) {
        format = "yyyy";
    }
    else if (range > JCAxis.YEARS) {
        format = "MMM yy";
    }
    else if (range > 3 * JCAxis.MONTHS) {
        format = "MMM";
    }
    else if (range > 2 * JCAxis.WEEKS) {
        format = "MMM dd";
    }
    else if (range > 2 * JCAxis.DAYS) {
        format = "EE dd";
    }
    else if (range > JCAxis.DAYS) {
        format = "EE HH:mm";
    }
    else if (range > JCAxis.HOURS) {
        format = "HH:mm";
    }
    else {
        format = "HH:mm:ss";
    }
    return(format);
}

/**
 * Internal method to returns a localized pattern based on the default locale
 * @param englishUSPattern the english pattern to be localized
 */
private String getLocalizedPattern(String englishUSPattern)
{
    formatter.applyPattern(englishUSPattern);
    return(formatter.toLocalizedPattern());
}

private boolean matchPattern(String pattern, String specificPattern)
{
    boolean match = false;

    for (int i = 0; i < specificPattern.length(); i++) {
        if (pattern.indexOf(specificPattern.charAt(i)) >= 0) {
            match = true;
            break;
        }
    }
    return(match);
}

/**
 * For testing purposes only
 * @param args Command line arguments
 */
public static void main(String args[])
{
    GregorianCalendar gcal = new GregorianCalendar(2004, 2, 29, 18, 22, 32);
    Date d = gcal.getTime();
    long time = d.getTime();
    System.out.println("Original Date:" + d.toString() + " msecs=" + time);
    JCChartDateFormat dateFormatter = new JCChartDateFormat();
    boolean doUp = true;

    d.setTime(time);
    dateFormatter.roundTime(d, JCAxis.SECONDS, doUp);
    System.out.println("Date(seconds):" + d.toString() +
                       " msecs=" + d.getTime());

    d.setTime(time);
    dateFormatter.roundTime(d, JCAxis.MINUTES, doUp);
    System.out.println("Date(minutes):" + d.toString() +
                       " msecs=" + d.getTime());

    d.setTime(time);
    dateFormatter.roundTime(d, JCAxis.HOURS, doUp);
    System.out.println("Date(hours):" + d.toString() +
                       " msecs=" + d.getTime());

    d.setTime(time);
    dateFormatter.roundTime(d, JCAxis.DAYS, doUp);
    System.out.println("Date(days):" + d.toString() +
                       " msecs=" + d.getTime());

    d.setTime(time);
    dateFormatter.roundTime(d, JCAxis.MONTHS, doUp);
    System.out.println("Date(months):" + d.toString() +
                       " msecs=" + d.getTime());

    d.setTime(time);
    dateFormatter.roundTime(d, JCAxis.YEARS, doUp);
    System.out.println("Date(years):" + d.toString() +
                       " msecs=" + d.getTime());
}

}

