/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartFactory.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.io.LoadProperties;
import com.klg.jclass.util.io.OutputDataProperties;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.io.JCParseException;
import com.klg.jclass.chart.property.*;
import com.klg.jclass.chart.property.html.JCHTMLFileAccessor;
import com.klg.jclass.chart.property.html.JCHTMLFilePersistor;
import com.klg.jclass.chart.property.html.JCHTMLStringAccessor;
import com.klg.jclass.chart.property.html.JCHTMLStringPersistor;
import com.klg.jclass.chart.data.*;

import java.io.*;
import java.util.Vector;

/**
 * Factory class that creates <code>JCChart</code> objects from various sources.
 */
public class JCChartFactory
{

/** Format (i.e. output file) types. */
public final static int HTML = 0;
//public final static int XML  = 1;

/** Ways to save data.  */
public final static int NO_DATA        = OutputDataProperties.NO_DATA;
public final static int EMBED_DATA     = OutputDataProperties.EMBED_DATA;
public final static int DATA_FILE_TEXT = OutputDataProperties.DATA_FILE_TEXT;
public final static int DATA_FILE_XML  = OutputDataProperties.DATA_FILE_XML;

/**
 * Updates a JCChart instance with new data coming from a file, an input stream,
 * or a reader.  This method always updates the first ChartDataView object.
 * @param chart The chart to update
 * @param dataType The type of data (must be <code>DATA_FILE_TEXT</code> or
 *     <code>DATA_FILE_XML</code> for anything to happen).
 * @param data the data object (can be a filename String, an
 *     <code>InputStream</code>, or a <code>Reader</code>)
 * @param lp a <code>LoadProperties</code> object.  This object
 *            contains properties relating to file loading and XML parsing.
 * @throws IOException If problems reading the data object
 */
public static void updateChartWithData(JCChart chart, int dataType,
                                       Object data, LoadProperties lp)
    throws IOException
{
    updateChartWithData(chart, dataType, data, 0, lp);
}

/**
 * Updates a JCChart instance with new data coming from a file, an input stream,
 * or a reader.
 * @param chart The chart to update
 * @param dataType The type of data (must be <code>DATA_FILE_TEXT</code> or
 *     <code>DATA_FILE_XML</code> for anything to happen).
 * @param data the data object (can be a filename String, an
 *     <code>InputStream</code>, or a <code>Reader</code>)
 * @param dataViewName The name of the ChartDataView object to update.  If the
 *     provided name does not match the name of any existing ChartDataView
 *     object, then nothing is done.
 * @param lp a <code>LoadProperties</code> object.  This object
 *            contains properties relating to file loading and XML parsing.
 * @throws IOException If problems reading the data object
 */
public static void updateChartWithData(JCChart chart, int dataType, Object data,
                                       String dataViewName, LoadProperties lp)
    throws IOException
{
    if (chart == null) {
        return;
    }
    Vector dataViews = (Vector)chart.getDataView();
    if (dataViews == null) {
        return;
    }
    int index = -1;
    for (int i = 0; i < dataViews.size(); i++) {
        ChartDataView dv = (ChartDataView)dataViews.get(i);
        if (dv != null) {
            String name = dv.getName();
            if (name != null && name.equals(dataViewName)) {
                index = i;
                break;
            }
        }
    }
    if (index >= 0) {
        updateChartWithData(chart, dataType, data, index, lp);
    }
}

/**
 * Updates a JCChart instance with new data coming from a file, an input stream,
 * or a reader.
 * @param chart The chart to update
 * @param dataType The type of data (must be <code>DATA_FILE_TEXT</code> or
 *     <code>DATA_FILE_XML</code> for anything to happen).
 * @param data the data object (can be a filename String, an
 *     <code>InputStream</code>, or a <code>Reader</code>)
 * @param dataViewIndex The index of the ChartDataView object to update.  If no
 *     ChartDataView exists for the provided index, this method will attempt to
 *     create a ChartDataView with that index and set the data on this new
 *     ChartDataView object.  This may result in the chart throwing an
 *     IllegalArgumentException if the index is more than one higher than the
 *     current number of dataviews.
 * @param lp a <code>LoadProperties</code> object.  This object
 *            contains properties relating to file loading and XML parsing.
 * @throws IOException If problems reading the data object
 */
public static void updateChartWithData(JCChart chart, int dataType, Object data,
                                       int dataViewIndex, LoadProperties lp)
    throws IOException
{
    // Check inputs
    if (chart == null || data == null) {
        return;
    }
    if (dataType != DATA_FILE_TEXT && dataType != DATA_FILE_XML) {
        return;
    }

    Vector dataViews = (Vector)chart.getDataView();
    if (dataViews == null || dataViewIndex < 0) {
        return;
    }

    if (lp == null) {
        lp = new LoadProperties();
    }

    if (dataViewIndex >= dataViews.size()) {
        ChartDataView newDataView = new ChartDataView();
        chart.setDataView(dataViewIndex, newDataView);
    }

    InputStream inputStream = null;
    Reader reader = null;
    if (data instanceof String) {
        inputStream = new FileInputStream((String)data);
    }
    else if (data instanceof InputStream) {
        inputStream = (InputStream)data;
    }
    else if (data instanceof Reader) {
        reader = (Reader)data;
    }
    if (inputStream == null && reader == null) {
        return;
    }

    ChartDataModel dataModel = null;
    JCDataInterpreter interpreter = null;
    switch (dataType) {
    case DATA_FILE_TEXT:
        interpreter = new JCDefaultDataInterpreter();
        break;
    case DATA_FILE_XML:
        interpreter = new JCXMLDataInterpreter(lp);
        break;
    }
    if (interpreter != null) {
        if (inputStream != null) {
            dataModel = new JCInputStreamDataSource(inputStream, interpreter);
        }
        else {
            dataModel = new JCReaderDataSource(reader, interpreter);
        }
    }
    if (dataModel != null) {
        chart.getDataView(dataViewIndex).setDataSource(dataModel);
    }
}

/**
 * Updates a JCChart instance by reading HTML-style chart parameters from an
 * input source. This object must either be a String which points to a filename,
 * an InputStream or a Reader. See the JClass Chart manual for more information
 * on the available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart The chart instance to be updated (if null, nothing is done)
 * @param inputSource The input source
 * @param name the name of the chart being read from parameters.  If non-null,
 *      only parameters of the form name.parameter are applied to this chart.
 * @param type The type of input. Currently this is always JCChartFactory.HTML.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.  If not
 *      needed, it should be null.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the input source
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
protected  static void updateChart(JCChart chart, String name, int type,
								   Object inputSource,
                                   LoadProperties loadProperties)
    throws IOException, JCIOException, JCParseException
{
    if (chart == null) {
        return;
    }
    chart.setName(name);

    if (loadProperties == null) {
        loadProperties = new LoadProperties();
    }

    PropertyAccessModel pam = null;
    switch (type) {
    case HTML:
        pam = new JCHTMLFileAccessor(chart, loadProperties);
        break;
//	case XML:
//      pam = new JCXMLFileAccessor(chart, loadProperties);
    default:
        throw new IllegalArgumentException("Bad chart file type");
    }
    if (pam == null) {
        return;
    }
    // Get the properties from the input source and store them
    pam.getProperties(inputSource);

    // Load the properties into the chart
    PropertyLoadModel loader = PropertyLoadFactory.makeLoader(chart, pam);
    if (loader != null) {
        loader.loadProperties(pam, null);
    }
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters
 * from an input stream.  See the JClass Chart manual for more information on
 * the available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param stream the input source
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save.  Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the input stream
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromStream(InputStream stream,
										  LoadProperties loadProperties,
										  String name,
                                          int type)
    throws IOException, JCIOException, JCParseException
{
	JCChart chart = new JCChart();
	updateChartFromStream(chart, stream, loadProperties, name, type);
	return(chart);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters
 * from an input stream.  See the JClass Chart manual for more information on
 * the available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param stream the input source
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save.  Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the input stream
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromStream(InputStream stream, String name,
                                          int type)
    throws IOException, JCIOException, JCParseException
{
	JCChart chart = new JCChart();
	updateChartFromStream(chart, stream, null, name, type);
	return(chart);
}

/**
 * Updates a <code>JCChart</code> instance by reading HTML-style chart
 * parameters from an input stream.  See the JClass Chart manual for more
 * information on the available parameters.  The parameters are expected to be
 * of the form: <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart the chart instance to be updated (if null, nothing is done)
 * @param stream the input source
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save. Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the input stream
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromStream(JCChart chart,
                                         InputStream stream,
                                         LoadProperties loadProperties,
                                         String name,
                                         int type)
    throws IOException, JCIOException, JCParseException
{
    updateChart(chart, name, type, stream, loadProperties);
}

/**
 * Updates a <code>JCChart</code> instance by reading HTML-style chart
 * parameters from an input stream.  See the JClass Chart manual for more
 * information on the available parameters.  The parameters are expected to be
 * of the form: <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart the chart instance to be updated (if null, nothing is done)
 * @param stream the input source
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save. Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the input stream
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromStream(JCChart chart, InputStream stream,
                                         String name, int type)
    throws IOException, JCIOException, JCParseException
{
    updateChart(chart, name, type, stream, null);
}

/**
 * Makes a JCChart instance by reading HTML-style chart parameters from a
 * reader.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <PARAM NAME=propertyName VALUE="value">
 *
 * @param reader The input source
 * @param name the name of the chart being read from parameters.  If non-null,
 *      only parameters of the form name.parameter are applied to this chart.
 * @param type The type of input. Currently this is always JCChartFactory.HTML.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @return JCChart instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the reader
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromReader(Reader reader,
                                          LoadProperties loadProperties,
                                          String name, int type)
    throws IOException, JCIOException, JCParseException
{
    JCChart chart = new JCChart();
    updateChartFromReader(chart, reader, loadProperties, name, type);
    return(chart);
}

/**
 * Makes a JCChart instance by reading HTML-style chart parameters from a
 * reader.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <PARAM NAME=propertyName VALUE="value">
 *
 * @param reader The input source
 * @param name the name of the chart being read from parameters.  If non-null,
 *      only parameters of the form name.parameter are applied to this chart.
 * @param type The type of input. Currently this is always JCChartFactory.HTML.
 * @return JCChart instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the reader
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromReader(Reader reader, String name, int type)
    throws IOException, JCIOException, JCParseException
{
    JCChart chart = new JCChart();
    updateChartFromReader(chart, reader, null, name, type);
    return(chart);
}

/**
 * Updates a JCChart instance by reading HTML-style chart parameters from a
 * reader.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <PARAM NAME=propertyName VALUE="value">
 *
 * @param chart The chart instance to be updated (if null, nothing is done)
 * @param reader The input source
 * @param name the name of the chart being read from parameters.  If non-null,
 *      only parameters of the form name.parameter are applied to this chart.
 * @param type The type of input. Currently this is always JCChartFactory.HTML.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the reader
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
*/
public static void updateChartFromReader(JCChart chart, Reader reader,
                                         LoadProperties loadProperties,
                                         String name, int type)
    throws IOException, JCIOException, JCParseException
{
    updateChart(chart, name, type, reader, loadProperties);
}

/**
 * Updates a JCChart instance by reading HTML-style chart parameters from a
 * reader.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form :
 * <PARAM NAME=propertyName VALUE="value">
 *
 * @param chart The chart instance to be updated (if null, nothing is done)
 * @param reader The input source
 * @param name the name of the chart being read from parameters.  If non-null,
 *      only parameters of the form name.parameter are applied to this chart.
 * @param type The type of input. Currently this is always JCChartFactory.HTML.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the reader
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromReader(JCChart chart, Reader reader,
                                         String name, int type)
    throws IOException, JCIOException, JCParseException
{
    updateChart(chart, name, type, reader, null);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters from a
 * file.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param fileName the file name of the input source containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save.  Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the file
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromFile(String fileName,
                                        LoadProperties loadProperties,
                                        String name, int type)
    throws IOException, JCIOException, JCParseException
{
    JCChart chart = new JCChart();
	updateChartFromFile(chart, fileName, loadProperties, name, type);
    return(chart);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters from a
 * file.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param fileName the file name of the input source containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save.  Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the file
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromFile(String fileName, String name, int type)
    throws IOException, JCIOException, JCParseException
{
    JCChart chart = new JCChart();
	updateChartFromFile(chart, fileName, null, name, type);
    return(chart);
}

/**
 * Updates a <code>JCChart</code> instance by reading HTML-style chart
 * parameters from a file.  See the JClass Chart manual for more information on
 * the available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart the chart instance to be updated (if null, nothing is done)
 * @param fileName the file name of the input source containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save.  Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the file
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromFile(JCChart chart, String fileName,
                                       LoadProperties loadProperties,
                                       String name, int type)
    throws IOException, JCIOException, JCParseException
{
	updateChart(chart, name, type, fileName, loadProperties);
}

/**
 * Updates a <code>JCChart</code> instance by reading HTML-style chart
 * parameters from a file.  See the JClass Chart manual for more information on
 * the available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart the chart instance to be updated (if null, nothing is done)
 * @param fileName the file name of the input source containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type The type of file to save.  Currently this is always
 *     <code>JCChartFactory.HTML</code>.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 * @throws IOException If problems reading from the file
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromFile(JCChart chart, String fileName,
                                       String name, int type)
    throws IOException, JCIOException, JCParseException
{
	updateChart(chart, name, type, fileName, null);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters
 * from a file.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param fileName the name of the file containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCFileAccessor
 */
public static JCChart makeChartFromFile(String fileName, String name)
{
	JCChart chart = null;
	try {
		chart = makeChartFromFile(fileName, name, HTML);
	}
	catch (Exception e) {
		e.printStackTrace();
	}
	return(chart);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters
 * from a String.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param params a String containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type the format of text within the String (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCStringAccessor
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromString(String params,
                                          LoadProperties loadProperties,
                                          String name, int type)
    throws JCIOException, JCParseException
{
	JCChart chart = new JCChart();
	updateChartFromString(chart, params, loadProperties, name, type);
	return(chart);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters
 * from a String.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param params a String containing chart parameters
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type the format of text within the String (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCStringAccessor
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static JCChart makeChartFromString(String params, String name, int type)
    throws JCIOException, JCParseException
{
	JCChart chart = new JCChart();
	updateChartFromString(chart, params, null, name, type);
	return(chart);
}

/**
 * Updates a <code>JCChart</code> instance by reading HTML-style chart parameters from
 * a String.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart the chart instance to be updated (if null, nothing is done)
 * @param params a String containing chart parameters.
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type the format of text within the String (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @param loadProperties instance of LoadProperties containing information
 *      that will be used during the load of chart properties.
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCStringAccessor
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromString(JCChart chart, String params,
                                         LoadProperties loadProperties,
                                         String name, int type)
    throws JCIOException, JCParseException
{
	if (chart == null) {
		return;
	}
	chart.setName(name);

	PropertyAccessModel pam = null;
	switch (type) {
	case HTML:
	    pam = new JCHTMLStringAccessor(chart, loadProperties);
		break;
//	case XML:
//	    pam = new JCXMLStringAccessor(chart, loadProperties);
//	    break;
	default:
	    throw new IllegalArgumentException("Bad chart file type");
	}
	if (pam == null) {
		return;
	}
    // Get the properties from the input source and store them
    try {
        pam.getProperties(params);
    }
    catch (JCIOException jcioe) {
        throw jcioe;
    }
    catch (JCParseException jcpe) {
        throw jcpe;
    }
    // This won't happen for strings but we need to catch it because the
    // PropertyAccessModel is used for other types
    catch (IOException e) {}

    // Load the properties into the chart
	PropertyLoadModel loader = PropertyLoadFactory.makeLoader(chart, pam);
	if (loader != null ) {
	    loader.loadProperties(pam, null);
	}
}

/**
 * Updates a <code>JCChart</code> instance by reading HTML-style chart parameters from
 * a String.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param chart the chart instance to be updated (if null, nothing is done)
 * @param params a String containing chart parameters.
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @param type the format of text within the String (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCStringAccessor
 * @throws JCIOException If an internal file referenced within the input source
 *     cannot be found.  These can be ignored by setting the appropriate option.
 * @throws JCParseException If errors in parsing the input source
 */
public static void updateChartFromString(JCChart chart, String params,
                                         String name, int type)
    throws JCIOException, JCParseException
{
    updateChartFromString(chart, params, null, name, type);
}

/**
 * Makes a <code>JCChart</code> instance by reading HTML-style chart parameters
 * from a String.  See the JClass Chart manual for more information on the
 * available parameters.  The parameters are expected to be of the form:
 * <code>&#60;PARAM NAME=propertyName VALUE="value"&#62;</code>
 *
 * @param params a String containing chart parameters.
 * @param name The name of the chart being read from parameters.  If non-null,
 *     only parameters of the form <code>name.parameter</code> are applied to
 *     this chart.
 * @return <code>JCChart</code> instance created from parameters
 * @see JCChart
 * @see com.klg.jclass.chart.property.JCStringAccessor
 */
public static JCChart makeChartFromString(String params, String name)
{
	JCChart chart = null;
	try {
		chart = makeChartFromString(params, name, HTML);
	}
	catch (Exception e) {
		e.printStackTrace();
	}
	return(chart);
}

/**
 * Set the dataSaveType and the outputDataFileName on all dataViews of the
 * chart.  The filename will be altered to be different for each dataView so
 * that all dataViews don't write to the same file.
 * @param chart The chart to be saved
 * @param dataSaveType How to write out the data.  This must be one of
 *     <code>NO_DATA</code>, <code>EMBED_DATA</code>,
 *     <code>DATA_FILE_TEXT</code>, or <code>DATA_FILE_XML</code>. Respectively,
 *     this means that either no data is saved, data is embeded in the HTML or
 *     XML file, data is saved as a separate text file, or data is saved as a
 *     separate XML file.  Note that if a separate text or xml file is written
 *     out, the corresponding property in the chart output is set to point to
 *     the absolute file name set in the outputDataFileName.
 * @param outputDataFileName the name of the data file if the data is being
 *     written to a separate file (a default name is used if NULL is passed in)
 */
public static void setOutputDataPropertiesOnChart(JCChart chart,
                                                  int dataSaveType,
                                                  String outputDataFileName)
{
    if (chart == null) {
        return;
    }
    Vector dataViews = (Vector)chart.getDataView();
    if (dataViews != null) {
        for (int i = 0; i < dataViews.size(); i++) {
            ChartDataView dv = (ChartDataView)dataViews.get(i);
            if (dv != null) {
                OutputDataProperties outProps = dv.getOutputDataProperties();
                if (outProps == null) {
                    outProps = new OutputDataProperties();
                }
                String numberString = (i > 0) ? "" + i : "";
                String ext = (dataSaveType == DATA_FILE_TEXT) ? ".dat" : ".xml";
                String fileName = outputDataFileName;
                if (fileName != null) {
                    // Allow user to use ".txt" endings for data files
                    if (dataSaveType == DATA_FILE_TEXT &&
                        fileName.endsWith(".txt"))
                    {
                        ext = ".txt";
                    }
                    // Add dataView string to the name if more than one dataView
                    if (fileName.endsWith(ext)) {
                        if (i > 0) {
                            int endIndex = fileName.length() - ext.length();
                            String baseName = fileName.substring(0, endIndex);
                            fileName = baseName + numberString + ext;
                        }
                    }
                    else {
                        fileName = fileName + numberString + ext;
                    }
                }
                else {
                    fileName = "jcchartdata" + numberString + ext;
                }
                outProps.setOutputFileName(fileName);
                outProps.setPropertyName(outputDataFileName);
                outProps.setSaveType(dataSaveType);
                dv.setOutputDataProperties(outProps);
            }
        }
    }
}

/**
 * Saves <code>JCChart</code> to an output stream in HTML format.
 * @param chart the chart to be saved
 * @param stream the output stream
 * @param type format type to save (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @param dataSaveType How to write out the data.  This must be one of
 *     <code>NO_DATA</code>, <code>EMBED_DATA</code>,
 *     <code>DATA_FILE_TEXT</code>, or <code>DATA_FILE_XML</code>.  Respectively
 *     this means that either no data is saved, data is embeded in the HTML or
 *     XML file, data is saved as a separate text file, or data is saved as a
 *     separate XML file.
 * @param dataFileName the name of the data file if the data is being written
 *     to a separate file (a default name is used if <code>NULL</code> is passed
 *     in)
 * @deprecated As of DesktopViews 6.2.  Set data output properties on each
 *     ChartDataView separately and use saveChartToStream(chart, stream, type)
 * @throws java.io.IOException
 */
public static void saveChartToStream(JCChart chart, OutputStream stream,
                                     int type, int dataSaveType,
									 String dataFileName)
	throws java.io.IOException
{
	setOutputDataPropertiesOnChart(chart, dataSaveType, dataFileName);
    saveChartToStream(chart, stream, type);
}

/**
 * Saves <code>JCChart</code> to an output stream in HTML format.
 * Output of data and images is specified by setting properties on the object
 * which holds the data/image.  For data, create an
 * <code>OutputDataProperties</code> object and set it on the corresponding
 * <code>ChartDataView</code>.  For images, create an
 * <code>OutputProperties</code> object and set in on the
 * <code>JCFillStyle</code> object (or other object which holds the image).
 * @param chart the chart to be saved
 * @param stream the output stream
 * @param type format type to save (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @throws IOException
 * @see com.klg.jclass.util.io.OutputProperties
 * @see com.klg.jclass.util.io.OutputDataProperties
 * @see com.klg.jclass.chart.ChartDataView
 * @see com.klg.jclass.chart.JCFillStyle
 */
public static void saveChartToStream(JCChart chart, OutputStream stream,
                                     int type)
    throws IOException
{
    if (chart == null || stream == null) {
        return;
    }

    PropertyPersistorModel ppm = null;
    switch (type) {
    case HTML:
        ppm = new JCHTMLFilePersistor(stream, chart);
        break;
//	case XML:
//	    ppm = new JCXMLFilePersistor(stream, chart);
//	    break;
    default:
        throw new IllegalArgumentException("Bad chart file type");
    }
    if (ppm == null) {
        return;
    }

    PropertySaveModel saver =
        PropertySaveFactory.makeSaver(chart, new JCChart(),
                                      PropertySaveFactory.getPackage(),
                                      ppm.getSubDirectory(), ppm.getType());
    if (saver != null) {
        saver.saveProperties(ppm, null, 0);
    }
}

/**
 * Saves <code>JCChart</code> to a file in HTML format.
 * @param chart the chart to be saved
 * @param fileName the name of the file
 * @param type format type to save (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @param dataSaveType How to write out the data.  This must be one of
 *     <code>NO_DATA</code>, <code>EMBED_DATA</code>,
 *     <code>DATA_FILE_TEXT</code>, or <code>DATA_FILE_XML</code>.  Respectively
 *     this means that either no data is saved, data is embeded in the HTML or
 *     XML file, data is saved as a separate text file, or data is saved as a
 *     separate XML file.
 * @param dataFileName the name of the data file if the data is being written
 *     to a separate file (a default name is used if <code>NULL</code> is passed
 *     in)
 * @deprecated As of DesktopViews 6.2.  Set data output properties on each
 *     ChartDataView separately and use saveChartToFile(chart, fileName, type)
 * @throws IOException
 */
public static void saveChartToFile(JCChart chart, String fileName, int type,
                                   int dataSaveType, String dataFileName)
	throws IOException
{
    setOutputDataPropertiesOnChart(chart, dataSaveType, dataFileName);
	saveChartToFile(chart, fileName, type);
}

/**
 * Saves <code>JCChart</code> to a file in HTML format.
 * Output of data and images is specified by setting properties on the object
 * which holds the data/image.  For data, create an
 * <code>OutputDataProperties</code> object and set it on the corresponding
 * <code>ChartDataView</code>.  For images, create an
 * <code>OutputProperties</code> object and set in on the
 * <code>JCFillStyle</code> object (or other object which holds the image).
 * @param chart the chart to be saved
 * @param fileName the name of the file
 * @param type format type to save (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @throws IOException
 * @see com.klg.jclass.util.io.OutputProperties
 * @see com.klg.jclass.util.io.OutputDataProperties
 * @see com.klg.jclass.chart.ChartDataView
 * @see com.klg.jclass.chart.JCFillStyle
 */
public static void saveChartToFile(JCChart chart, String fileName, int type)
    throws IOException
{
    saveChartToStream(chart, new FileOutputStream(fileName), type);
}

/**
 * Saves <code>JCChart</code> to a String in HTML format.
 * @param chart the chart to be saved
 * @param type format type to save (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @param dataSaveType How to write out the data.  This must be one of
 *     <code>NO_DATA</code>, <code>EMBED_DATA</code>,
 *     <code>DATA_FILE_TEXT</code>, or <code>DATA_FILE_XML</code>.  Respectively
 *     this means that either no data is saved, data is embeded in the HTML or
 *     XML file, data is saved as a separate text file, or data is saved as a
 *     separate XML file.
 * @param dataFileName the name of the data file if the data is being written
 *     to a separate file (a default name is used if <code>NULL</code> is passed
 *     in)
 * @deprecated As of DesktopViews 6.2.  Set data output properties on each
 *     ChartDataView separately and use saveChartToString(chart, type)
 * @return the String which contains the chart information in either HTML
 *     or XML format
 */
public static String saveChartToString(JCChart chart, int type,
                                       int dataSaveType, String dataFileName)
{
    setOutputDataPropertiesOnChart(chart, dataSaveType, dataFileName);
    String s = null;
    try {
        s = saveChartToString(chart, type);
    }
    catch (IOException e) {
        e.printStackTrace();
    }
	return(s);
}

/**
 * Saves <code>JCChart</code> to a String in HTML format.
 * Output of data and images is specified by setting properties on the object
 * which holds the data/image.  For data, create an
 * <code>OutputDataProperties</code> object and set it on the corresponding
 * <code>ChartDataView</code>.  For images, create an
 * <code>OutputProperties</code> object and set in on the
 * <code>JCFillStyle</code> object (or other object which holds the image).
 * @param chart the chart to be saved
 * @param type format type to save (currently this is always
 *     <code>JCChartFactory.HTML</code>)
 * @return the String which contains the chart information in either HTML
 *     or XML format
 * @throws JCIOException
 * @see com.klg.jclass.util.io.OutputProperties
 * @see com.klg.jclass.util.io.OutputDataProperties
 * @see com.klg.jclass.chart.ChartDataView
 * @see com.klg.jclass.chart.JCFillStyle
 */
public static String saveChartToString(JCChart chart, int type)
    throws JCIOException
{
    if (chart == null) {
        return (null);
    }

    PropertyPersistorModel ppm = null;
    StringBuffer sb = new StringBuffer();
    switch (type) {
    case HTML:
        ppm = new JCHTMLStringPersistor(sb, chart);
        break;
//	case XML:
//	    ppm = new JCXMLStringPersistor(sb, chart);
//	    break;
    default:
        throw new IllegalArgumentException("Bad chart file type");
    }
    if (ppm == null) {
        return (null);
    }

    PropertySaveModel saver =
        PropertySaveFactory.makeSaver(chart, new JCChart(),
                                      PropertySaveFactory.getPackage(),
                                      ppm.getSubDirectory(), ppm.getType());
    if (saver != null) {
        saver.saveProperties(ppm, null, 0);
    }
    return(sb.toString());
}

}
