/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartLegendManager.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.legend.JCLegend;
import com.klg.jclass.util.legend.JCLegendItem;
import com.klg.jclass.util.legend.JCLegendPopulator;
import com.klg.jclass.util.legend.JCLegendRenderer;
import com.klg.jclass.util.swing.TextRenderer;

import java.awt.*;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

/**
 * <code>JCChartLegend</code> manages the populating and rendering of legend
 * items in a JClass Chart.
 */
public class JCChartLegendManager implements JCLegendPopulator,
							                 JCLegendRenderer,
                                             java.io.Serializable
{

/**
 * Legend vector -- contains all items in the legend, organized into
 * subvectors representing one column (usually defined by one data view)
 * of the legend.
 */
protected List legendItems = null;

/**
 * Cached data views, series, and related data used by legend drawing. These
 * must be the same as those used in legend population.
 */
protected transient CachedViewProps cachedView[] = null;

/**
 * The <code>JCChart</code> instance for which this class is managing.
 */
protected JCChart chart;

/**
 * Internal constructor for legend.
 * @param c internal constructor
 */
public JCChartLegendManager(JCChart c)
{
	chart = c;
}

/**
 * This method returns <code>true</code> if the first visible data view is of
 * the specified chart type.
 * @param chartType
 * @return <Code>true</code> is the first visible data view is of the specified
 * chart type; <code>false</code> otherwise
 */
protected boolean isFirst(int chartType)
{
	boolean fFirst = false;
	if (chart != null) {
		Iterator dataViews = chart.getOriginalDataView().iterator();
		while (dataViews.hasNext()) {
			ChartDataView dv = (ChartDataView)dataViews.next();
			if (dv != null && dv.isVisible()) {
				fFirst = (chartType == dv.getChartType());
				break;
			}
		}
	}
	return(fFirst);
}

/**
 * Creates the list of legend items from the chart's data views and
 * populates the legend with it.
 * @param fm the <code>FontMetrics</code> for this legend
 * @return the list of legend items
 */
public List getLegendItems(FontMetrics fm)
{
    // If the first visible data view is of type PIE, RADAR,
    // or AREA_RADAR, display legend items for that data
    // view only.  If the first visible data view is of type
    // POLAR, display legend items for all data views of type
    // POLAR.
    // If the first visible data view isn't of type PIE, POLAR,
    // RADAR, or AREA_RADAR, don't display legend items for
    // data views of those types.
    boolean pieFirst   = isFirst(JCChart.PIE);
    boolean polarFirst = isFirst(JCChart.POLAR);
    boolean radarFirst = isFirst(JCChart.RADAR) || isFirst(JCChart.AREA_RADAR);
    legendItems = new Vector();
    Vector cachedViewList = new Vector();

    Iterator dataViews = chart.getOriginalDataView().iterator();
    while (dataViews.hasNext()) {
        ChartDataView dv = (ChartDataView)dataViews.next();
        if (dv != null && dv.isVisible()) {
            int chartType = dv.getChartType();
            boolean pieChart = (chartType == JCChart.PIE);
            boolean polarChart = (dv.getChartType() == JCChart.POLAR);
            boolean radarChart = (chartType == JCChart.RADAR ||
                                  chartType == JCChart.AREA_RADAR);

            // Check to see if chart type matches type of first dataView.  We
            // can't have a radar drawing on top of a bar chart, for example.
            boolean matchPie   = (pieFirst && pieChart);
            boolean matchRadar = (radarFirst && radarChart);
            boolean matchPolar = (polarFirst && polarChart);
            boolean matchOther = (!pieFirst && !radarFirst && !polarFirst &&
                                  !pieChart && !radarChart && !polarChart);
            if (matchPie || matchRadar || matchPolar || matchOther) {
                Vector viewItems = new Vector();
                legendItems.add(viewItems);
                CachedViewProps cachedViewProps = createLegendItems(dv,
                                                                    viewItems,
                                                                    fm);
                if (cachedViewProps != null) {
                    cachedViewList.add(cachedViewProps);
                }
                if (matchPie || matchRadar) {
                    // No multiple dataViews for pies or radar charts
                    break;
                }
            }
        }
    }

	cachedView = new CachedViewProps[cachedViewList.size()];
	cachedViewList.copyInto(cachedView);

	return(legendItems);
}

private CachedViewProps createLegendItems(ChartDataView dt, List viewItems,
										  FontMetrics fm)
{
    JCLegend legend = chart.getLegend();
    List series = dt.getSeries();
    CachedViewProps cachedViewProps = new CachedViewProps();
    cacheViewProperties(cachedViewProps, dt, series);

    dt.calcTransientData();
    if (series == null) {
        return null;
    }
    int seriesPerData = dt.getNumSeriesPerData();
    int series_len = series.size() / seriesPerData;
    if (series_len == 0) {
        return null;
    }

    int chartType = dt.getChartType();

    // Fill in the size of the data view title as the first item
    // in the size array (d[])
    TextRenderer textRenderer = legend.getTextRenderer();
    if (dt.isVisibleInLegend()) {
        String dataLabel = dt.getName();
        JCLegendItem titleItem = new JCLegendItem();
        if (TextRenderer.isHTML(dataLabel)) {
            titleItem.textDim = textRenderer.calcHTMLTextSize(dataLabel);
            titleItem.contents = dataLabel;
        }
        else {
            if ((dataLabel != null) && (dataLabel.length() > 0)) {
                titleItem.textDim = new Dimension(fm.stringWidth(dataLabel),
                                                  fm.getHeight());
                titleItem.contents = dataLabel;
            }
            else {
                titleItem.textDim = new Dimension(0, 0);
                titleItem.contents = null;
            }
        }
        titleItem.itemInfo = new JCDataIndex(dt, null, -1, -1);
        viewItems.add(titleItem);
    }
    else {
        return(null);
    }

    // If pie chart and we possibly have the other slice, we only show items
    // that are visible in one or more of the pies.  Thus we create legend items
    // for those series used in visible slices plus the other slice.  Each
    // position in the pieItems array represents one series.  If a series is not
    // used, the item is null, but the position is still kept.  Pie charts
    // are a special case this way.  The last position represents the
    // other slice.  Once we are done, all non-null items are added to the
    // viewItems list.
    JCPieChartFormat pieFormat =
        (JCPieChartFormat)dt.getChartFormat(JCChart.PIE);
    if (chartType == JCChart.PIE && pieFormat.getThresholdValue() > 0.0) {
        PieChartInfo pinfo = (PieChartInfo) dt.getTransientData();
        if (pinfo == null) {
            return(cachedViewProps);
        }
        JCLegendItem pieItems[] = new JCLegendItem[series.size() + 1];
        for (int k = 0; k < pinfo.num_pies; k++) {
            PieData pie_data = pinfo.data[k];
            for (int j = 0; j < pie_data.num_slices_pie; j++) {
                if (pie_data.other_flag != 0 && j == pie_data.other_slice) {
                    if (pieItems[series.size()] == null) {
                        JCLegendItem newItem = new JCLegendItem();
                        JCPieChartFormat format =
                            (JCPieChartFormat)dt.getChartFormat();
                        newItem.itemInfo =
                            new JCDataIndex(dt, null, format.OTHER_SLICE, -1);
                       String s = format.getOtherLabel();
                        if (TextRenderer.isHTML(s)) {
                            newItem.textDim = textRenderer.calcHTMLTextSize(s);
                            newItem.contents = s;
                        }
                        else {
                            newItem.textDim = new Dimension(fm.stringWidth(s),
                                                            fm.getHeight());
                            newItem.contents = s;
                        }
                        JCChartStyle style = format.getOtherStyle();
                        calcSymbolSize(newItem, style, chartType);

                        pieItems[series.size()] = newItem;
                    }
                }
                else {
                    int index = pie_data.percents[j].index;
                    if (index >= 0 && pieItems[index] == null) {
                        ChartDataViewSeries thisSeries =
                            (ChartDataViewSeries) series.get(index);
                        if (thisSeries.isVisibleInLegend()) {
                            JCLegendItem newItem = new JCLegendItem();
                            newItem.itemInfo = new JCDataIndex(dt, thisSeries,
                                                               index, -1);
                            String s = thisSeries.getLabel();
                            if (TextRenderer.isHTML(s)) {
                                newItem.textDim =
                                    textRenderer.calcHTMLTextSize(s);
                                newItem.contents = s;
                            }
                            else {
                                newItem.textDim =
                                    new Dimension(fm.stringWidth(s),
                                                  fm.getHeight());
                                newItem.contents = s;
                            }
                            calcSymbolSize(newItem, thisSeries.getStyle(),
                                           chartType);
                            pieItems[index] = newItem;
                        }
                    }
                }
            }
        }
        for (int j = 0; j < pieItems.length; j++) {
            if (pieItems[j] != null) {
                viewItems.add(pieItems[j]);
            }
        }
    }
    else {
        // For other chart types, fill in sizes of series showing.
        // Each position in the size array (d[]) represents part of one
        // row/column.

        for (int j = 0; j < series_len; j++) {
            int k = j * seriesPerData;

            ChartDataViewSeries thisSeries =
                (ChartDataViewSeries) series.get(k);
            JCLegendItem newItem = new JCLegendItem();
            newItem.itemInfo =
                new JCDataIndex(dt, thisSeries, k, -1);
            String s = null;
            int sHeight = 0;
            int sWidth = 0;
            s = thisSeries.getLabel();
            if (TextRenderer.isHTML(s)) {
                Dimension newDim = textRenderer.calcHTMLTextSize(s);
                newItem.contents = s;
                sWidth = newDim.width;
                sHeight = newDim.height;
            }
            else {
                sWidth = fm.stringWidth(s);
                sHeight = fm.getHeight();
                newItem.contents = s;
            }

            boolean showSeries = true;
            for (int si = k, count = 0; count < seriesPerData; count++, si++) {
                ChartDataViewSeries srs = (ChartDataViewSeries)series.get(si);
                if (!srs.isVisibleInLegend()) {
                    showSeries = false;
                    break;
                }
            }
            if (showSeries && thisSeries.isVisibleInLegend()) {
                newItem.textDim = new Dimension(sWidth, sHeight);
                calcSymbolSize(newItem, thisSeries.getStyle(), chartType);
                viewItems.add(newItem);
            }
        }
    }
    return(cachedViewProps);
}

/**
 * Calculates the appropriate symbol size to be used for a legend item.
 * Also notes the appropriate draw type.
 * @param newItem the legend item in question
 * @param style the style object used by this legend item
 * @param chartType the chart type associated with this legend item
 */
protected void calcSymbolSize(JCLegendItem newItem, JCChartStyle style,
							  int chartType)
{
    boolean barChart = (chartType == JCChart.BAR ||
                        chartType == JCChart.STACKING_BAR);
    boolean areaChart = (chartType == JCChart.AREA ||
                         chartType == JCChart.STACKING_AREA ||
                         chartType == JCChart.AREA_RADAR);
    boolean pieChart = (chartType == JCChart.PIE);

    JCFillStyle fs = style.getFillStyle();
    Image image = fs.getImage();
    int pattern = fs.getPattern();
    boolean customFill = (pattern == JCFillStyle.CUSTOM_FILL);
    boolean customStack = (pattern == JCFillStyle.CUSTOM_STACK);

    // If in a situation where we can use a custom image to
    // fill (i.e. if bar or stacking bar and either CUSTOM_FILL
    // or CUSTOM_STACK is set OR if area or stacking area or pie and
    // CUSTOM_FILL is set) AND we actually have an image,
    // mark down it's size
    if (image != null &&
        ((barChart && (customStack || customFill)) ||
		 ((areaChart || pieChart) && customFill)))
    {
        if (customStack) {
            newItem.drawType = JCLegend.IMAGE;
        }
        else if (customFill) {
            newItem.drawType = JCLegend.IMAGE_OUTLINED;
        }
        newItem.symbolDim = new Dimension(image.getWidth(null),
                                          image.getHeight(null));
 	}
    else {
	    // Otherwise, we use the standard box for many chart types and
        // draw custom symbols in other cases.
		if (barChart || areaChart || pieChart) {
			newItem.drawType = JCLegend.BOX;
		}
        else {
			newItem.drawType = JCLegend.CUSTOM_SYMBOL;
		}

		JCLegend legend = chart.getLegend();
    	newItem.symbolDim = new Dimension(legend.getSymbolSize(),
                                          legend.getSymbolSize());
   	}
    newItem.symbol = image;
}

/**
 * Caches properties of the <code>ChartDataView</code> and it's series for
 * later use by <code>draw()</code>. This caching is needed to make the chart
 * more threadsafe.
 * @param cachedView
 * @param dt
 * @param series
 */
void cacheViewProperties(CachedViewProps cachedView, ChartDataView dt,
                         List series)
{
    cachedView.view = dt;
    cachedView.visible = dt.isVisible();
    cachedView.visibleInLegend = dt.isVisibleInLegend();
    cachedView.chartType = dt.getChartType();
    cachedView.seriesPerData = dt.getNumSeriesPerData();
    cachedView.inverted = dt.getXAxis().i_vertical;
    cachedView.outlineColor = dt.getOutlineColor();

    if (series == null) {
        cachedView.seriesList = null;
        cachedView.seriesProps = null;
        return;
    }

    cachedView.seriesList = series;
    cachedView.seriesProps = new CachedSeriesProps[series.size()];
    for (int j = 0; j < series.size(); j++) {
		ChartDataViewSeries thisSeries = (ChartDataViewSeries) series.get(j);
        cachedView.seriesProps[j] = new CachedSeriesProps();
        cachedView.seriesProps[j].series = thisSeries;
        cachedView.seriesProps[j].visible = thisSeries.isVisible();
        cachedView.seriesProps[j].visibleInLegend =
            thisSeries.isVisibleInLegend();
    }
}

/**
 * <code>JCLegendRenderer</code> method that is not used for this
 * implementation.
 * Just return.
 * @param gc the graphics context
 * @param useFont the font used for this legend
 * @param thisItem the legend item to draw (a <code>JCLegendItem</code>
 * instance)
 */
public void drawLegendItem(Graphics gc, Font useFont, JCLegendItem thisItem)
{
	return;
}

/**
 * Draws one legend item based on the size and position given in it's
 * <code>JCLegendItem</code> instance.  The symbol to be drawn is derived from
 * the chart type and the symbol style for the series.
 * @param gc the graphics context
 * @param useFont the font used for this legend
 * @param thisItem the legend item to draw (a <code>JCLegendItem</code>
 * instance)
 */
public void drawLegendItemSymbol(Graphics gc, Font useFont,
                                 JCLegendItem thisItem)
{
    if (thisItem == null) {
        return;
    }
    CachedViewProps thisCachedView = null;
    boolean inverted = false;
    int chartType = 0;
    boolean showOpen = true;
    boolean showClose = true;
    boolean fullTickWidth = false;
    boolean isComplex = false;
	Color   outlineColor = null;

    // Find cached view for this item
	ChartDataView itemView = ((JCDataIndex)thisItem.itemInfo).getDataView();
    if (itemView != null) {
        for (int k = 0; k < cachedView.length; k++) {
            if (cachedView[k].view == itemView) {
                thisCachedView = cachedView[k];

                inverted = thisCachedView.inverted;
                chartType = thisCachedView.chartType;
                showOpen = true;
                showClose = true;
                fullTickWidth = false;
				outlineColor = thisCachedView.outlineColor;
                if (chartType == JCChart.HILO_OPEN_CLOSE) {
                    JCHLOCChartFormat fmt =
						(JCHLOCChartFormat)itemView.getChartFormat();
                    showOpen = fmt.isShowingOpen();
                    showClose = fmt.isShowingClose();
                    fullTickWidth = fmt.isOpenCloseFullWidth();
                }
                if (chartType == JCChart.CANDLE) {
                    isComplex = ((JCCandleChartFormat)itemView.
										getChartFormat()).isComplex();
                }
                break;
            }
        }
    }

	JCLegend legend = chart.getLegend();
    int width  = thisItem.symbolDim.width;
    int height = thisItem.symbolDim.height;

    if (gc.getColor() != legend.getForeground()) {
        gc.setColor(legend.getForeground());
    }
    if (gc.getFont() != useFont) {
        gc.setFont(useFont);
    }
	if (outlineColor == null) {
		outlineColor = chart.getChartArea().getForeground();
    }

    // Draw symbol
    int symX = thisItem.pos.x + thisItem.symbolPos.x;
    int symY = thisItem.pos.y + thisItem.symbolPos.y;

    // For HILO, CANDLE types
    int x1, x2, y1, y2;
    int h1 = Math.min(width, height);

    // Series determines symbol
    Graphics2D gc2 = (Graphics2D)gc;
	ChartDataViewSeries itemSeries =
        ((JCDataIndex)thisItem.itemInfo).getSeries();
    if (itemSeries != null) {
        JCChartStyle style = itemSeries.getStyle();
		JCLineStyle sourceLineStyle = style.getLineStyle();
		JCLineStyle useLineStyle =
            new JCLineStyle(1, sourceLineStyle.getColor(),
                            sourceLineStyle.getPattern(),
                            sourceLineStyle.getCap(),
                            sourceLineStyle.getJoin(),
                            sourceLineStyle.getCustomPatternArray(),
                            sourceLineStyle.getCustomLegendPatternArray());
        switch (itemView.getChartType()) {
        case JCChart.POLAR:
        case JCChart.RADAR:
        case JCChart.PLOT:
            if (useLineStyle.updateGraphics(gc2, width - 1)) {
                gc2.draw(new Line2D.Double((double) symX,
                                           (double) (symY + (height/2)),
                                           (double) (symX + width - 1),
                                           (double) (symY + (height/2))));
                useLineStyle.resetGraphics(gc2);
            }
            // No break.  Yes, this is intentional
        case JCChart.SCATTER_PLOT:
            style.getSymbolStyle().draw(gc2, symX + (width/2),
                                        symY + (height/2), width);
            break;
        case JCChart.BAR:
        case JCChart.STACKING_BAR:
        case JCChart.PIE:
        case JCChart.AREA_RADAR:
        case JCChart.AREA:
        case JCChart.STACKING_AREA:
            // These cases should be handled by default by JCLegend
            break;
        case JCChart.HILO:
        case JCChart.HILO_OPEN_CLOSE:
            if (useLineStyle.updateGraphics(gc2)) {
                if (inverted) {
                    x1 = symX;
                    x2 = symX + (width - 1);
                    y1 = y2 = symY + (h1 / 2);
                }
                else {
                    x1 = x2 = symX + (width/2);
                    y1 = symY;
                    y2 = symY + h1;
                }
                gc2.draw(new Line2D.Double((double) x1, (double) y1,
                                           (double) x2, (double) y2));
                if (chartType == JCChart.HILO) {
                    break;
                }
                if (showOpen) {
                    if (inverted) {
                        x1 = x2 = symX + 2;
                        y1 = symY + 1;
                        y2 = (fullTickWidth) ? symY + h1 - 2
                                             : symY + (h1 / 2);
                    }
                    else {
                        x1 = symX + 2;
                        x2 = (fullTickWidth) ? symX - 3 + width
                                             : symX + (width/2);
                        y1 = y2 = symY +
                            (int) Math.round(2 * (double) h1 / 3);
                    }
                    gc2.draw(new Line2D.Double((double) x1, (double) y1,
                                               (double) x2, (double) y2));
                }
                if (showClose) {
                    if (inverted) {
                        x1 = x2 = symX - 3 + width;
                        y1 = (fullTickWidth) ? symY + 1 : symY + (h1 / 2);
                        y2 = symY + h1 - 2;
                    }
                    else {
                        x1 = (fullTickWidth) ? symX + 2 : symX + (width/2);
                        x2 = symX - 3 + width;
                        y1 = y2 = symY + (h1 / 3);
                    }
                    gc2.draw(new Line2D.Double((double) x1, (double) y1,
                                               (double) x2, (double) y2));
                }
                useLineStyle.resetGraphics(gc2);
            }
            break;
        case JCChart.CANDLE:
            int candleWidth = ((width % 2) == 0) ? width : (width - 1);
            if (inverted) {
                x1 = symX;
                x2 = symX + 2;
                y1 = y2 = symY + (h1 / 2);
            }
            else {
                x1 = x2 = symX + (candleWidth/2);
                y1 = symY;
                y2 = symY + (h1 / 3) ;
            }
            if (useLineStyle.updateGraphics(gc2)) {
                gc2.draw(new Line2D.Double((double) x1, (double) y1,
                                           (double) x2, (double) y2));
                if (inverted) {
                    x1 = symX - 3 + candleWidth;
                    x2 = symX + (candleWidth - 1);
                    y1 = y2 = symY + (h1 / 2);
                }
                else {
                    x1 = x2 = symX + (candleWidth/2);
                    y1 = symY + (h1 * 2 / 3) ;
                    y2 = symY + h1;
                }
                gc2.draw(new Line2D.Double((double) x1, (double) y1,
                                           (double) x2, (double) y2));
                useLineStyle.resetGraphics(gc2);
            }
            if (inverted) {
                x1 = symX + 3;
                y1 = symY + 3;
                x2 = candleWidth - 7;
                y2 = h1 - 7;
            }
            else {
                x1 = symX + 3;
                y1 = symY + (h1 / 3);
                x2 = candleWidth - 6;
                y2 = (int) Math.round((double) h1 / 3);
            }
            JCChartStyle styleCandle = style;
            JCChartStyle styleOutline = style;
            if (isComplex) {
                JCCandleChartFormat cf =
                    (JCCandleChartFormat) itemView.getChartFormat();
                int itemSeriesIndex =
                    ((JCDataIndex)thisItem.itemInfo).getSeriesIndex();
                styleCandle = cf.getRisingCandleStyle(
                    itemSeriesIndex / thisCachedView.seriesPerData);
                styleOutline = cf.getCandleOutlineStyle(
                    itemSeriesIndex / thisCachedView.seriesPerData);
                styleCandle.getFillStyle().fillRect(gc, x1, y1, x2, y2);
            }

            JCLineStyle sourceOutlineStyle = styleOutline.getLineStyle();
            JCLineStyle useOutlineStyle =
                new JCLineStyle(1, sourceOutlineStyle.getColor(),
                                sourceOutlineStyle.getPattern(),
                                sourceOutlineStyle.getCap(),
                                sourceOutlineStyle.getJoin(),
                                sourceLineStyle.getCustomPatternArray(),
                                sourceLineStyle.getCustomLegendPatternArray());
            if (useOutlineStyle.updateGraphics(gc2)) {
                gc2.draw(new Rectangle2D.Double((double) x1, (double) y1,
                                                (double) x2, (double) y2));
                useOutlineStyle.resetGraphics(gc2);
            }
            break;
        default:
            if (useLineStyle.updateGraphics(gc)) {
                gc2.draw(new Line2D.Double((double) symX,
                                           (double) (symY - height/2),
                                           (double) (symX + width),
                                           (double) (symY - height/2) ));
                useLineStyle.resetGraphics(gc2);
            }
            break;
        }
    }
    // Special case : pie chart other slice
    else if (itemView != null) {
        switch (chartType) {
        case JCChart.PIE:
            ((JCPieChartFormat)itemView.getChartFormat()).getOtherStyle().
                getFillStyle().fillOutlineRect(gc, symX, symY, width - 1,
                                               height - 1, outlineColor);
            break;
        }
    }
}

/**
 * Used for implementing custom text draws for a legend item.  This method
 * is called when a legend item's contents field contains an object other
 * than a <code>String</code> object.
 *
 * @param gc <code>Graphics</code> object used to draw the legend
 * @param useFont the font being used in this legend
 * @param thisItem the <code>JCLegendItem</code> object representing the
 *     legend item to draw
 */
public void drawLegendItemText(Graphics gc, Font useFont, JCLegendItem thisItem)
{
	// No implementation needed
}

/**
 * Retrieves the appropriate outline color for this legend item's symbol.
 * In Chart, this is the chart area's foreground color.
 *
 * @param thisItem the legend item in question
 * @return the appropriate color to use to outline the legend symbol
 */
public Color getOutlineColor(JCLegendItem thisItem)
{
	return(chart.getChartArea().getForeground());
}

/**
 * Sets the supplied <code>Graphics</code> object with the appropriate drawing
 * properties for this legend item.  The actual draw will be done
 * by the legend class.
 * @param gc the <code>Graphics</code> object to change
 * @param thisItem the legend item in question
 */
public void setFillGraphics(Graphics gc, JCLegendItem thisItem)
{
    // Series determines symbol
    JCDataIndex dataIndex = (JCDataIndex)thisItem.itemInfo;
    ChartDataView itemView = dataIndex.getDataView();
	ChartDataViewSeries itemSeries = dataIndex.getSeries();

	// The chart data series this legend represents determines the
	// painting properties to be set on this graphics object
    if (itemSeries != null) {
        JCChartStyle style = itemSeries.getStyle();
        style.getFillStyle().updateGraphics(gc);
	}
    else if (itemView != null) {
		// Special case -- Pie chart other slice
        if (itemView.getChartType() == JCChart.PIE) {
            JCPieChartFormat fmt = (JCPieChartFormat)itemView.getChartFormat();
            fmt.getOtherStyle().getFillStyle().updateGraphics(gc);
		}
    }
}

/**
 * Checks if the legend item is a title.  In this implementation,
 * title items have a null series and no symbol attached.
 * @param item  the legend item in question
 * @return true if the legend item is a title.
 */
public boolean isTitleItem(JCLegendItem item)
{
	JCDataIndex di = (JCDataIndex) item.itemInfo;
	if (item.symbolDim == null) {
		if (di == null || di.getSeries() == null) {
        	return(true);
		}
	}
    return(false);
}

}
