/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartNumberFormat.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.text.NumberFormat;
import java.util.StringTokenizer;
import java.util.Locale;

/**
 * <code>JCChartNumberFormat</code> provides a number of static methods
 * used in numbers localization.
 */
public class JCChartNumberFormat {

private static final NumberFormat defaultFormatter =
    NumberFormat.getInstance(Locale.getDefault());
private NumberFormat formatter = null;
private ChartInteriorRegion parentRegion = null;

// Workaround for a bug in JDK; supports only 19 decimals
// Sun's bug id. filed against JDK1.1.4 is 4098741 and it was reported
// on Dec. 11. 1997
private static final int MAX_COUNT = 19;

public JCChartNumberFormat(ChartInteriorRegion parent)
{
    parentRegion = parent;
    formatter = defaultFormatter;
}

/**
 * The engine of the number localization.
 * The function is parsing a String and is looking
 * for numbers in the String.
 * @param s the String to be localized
 * @return
 */
private String parse(String s)
{
    String localized = new String();
    String delimiters = " \t\n\r[]()";
    StringTokenizer tokenizer = new StringTokenizer(s, delimiters, true);
    String splitToken = null;

    while (tokenizer.hasMoreTokens() || splitToken != null) {
        String token;
        if (splitToken != null) {
            token = splitToken;
            splitToken = null;
        }
        else {
            token = tokenizer.nextToken();
        }

        if (delimiters.indexOf(token) >= 0) {
            localized += token;
            continue;
        }
        try {
            // Check for exponential 'E' (scientific notation).  If present,
            // split string up at 'E', localize both bits, and reassemble
            int ePos = -1;
            if ((ePos = token.indexOf('E')) >= 0) {
                splitToken = token.substring(ePos + 1);
                token = token.substring(0, ePos);
            }

            Double d = Double.valueOf(token);

            int currentPrecision = formatter.getMaximumFractionDigits();

            // Don't apply precision to front of 'E' in exponential notation
            if (ePos >= 0) {
                formatter.setMaximumFractionDigits(MAX_COUNT);
            }
            boolean localize = true;

            String formattedValue = formatter.format(d);
            while ((formattedValue.compareTo("-0") == 0 ||
                formattedValue.compareTo("0") == 0) &&
                d.doubleValue() != 0) {
                int decimals =
                    formatter.getMaximumFractionDigits() <= MAX_COUNT ?
                    formatter.getMaximumFractionDigits() + 1 : MAX_COUNT;
                formatter.setMaximumFractionDigits(decimals);
                if (decimals == MAX_COUNT) {
                    localize = false;
                    break;
                }
                formattedValue = formatter.format(d);
            }
            if (localize) {
                localized += formattedValue;
            }
            else {
                localized += token;
            }

            // Add 'E' back into exponential notation
            if (ePos >= 0) {
                localized += "E";
            }

            formatter.setMaximumFractionDigits(currentPrecision);
        }
        catch (NumberFormatException e) {
            localized += token;
        }
    }
    return(localized);
}

/**
 * Localizes a String.
 * @param s the String to be localized
 * @return the localized String
 */
public String localize(String s)
{
    int prec = Integer.MAX_VALUE;
    // prec is used only if parentRegion is a JCAxis
    if (parentRegion instanceof JCAxis) {
        prec = ((JCAxis)parentRegion).getPrecision();
    }
    return localize(s, prec);
}

/**
 * Localizes a String.
 * @param s The String to be localized
 * @param prec The required precision if the String is for an axis annotation.
 * If not an axis annotation, the precision parameter is ignored.
 * @return The localized String
 */
public String localize(String s, int prec) {
    String localized = null;
    int oldPrec = formatter.getMaximumFractionDigits();
    int oldMinPrec = formatter.getMinimumFractionDigits();
    boolean isMicrosoftVM =
        System.getProperty("java.vendor").indexOf("Microsoft") >= 0;

    formatter.setGroupingUsed(parentRegion.getGroupingUsed());
    if (parentRegion instanceof JCAxis) {
        // Workaround to get around bug #11116 (aka Sun jdk bug #4252562)
        // Never let MaximumFractionDigits be less than or equal to zero.
        // If it is, all but the first attempt to format the number of zero
        // will come out as the string "1".
        if (prec <= 0) {
            formatter.setMaximumFractionDigits(1);
        }
        else {
            formatter.setMaximumFractionDigits(prec);
        }

        if (((JCAxis)parentRegion).isDecimalPadded()) {
            formatter.setMinimumFractionDigits(prec);
        }
    }
    if (!isMicrosoftVM && parentRegion.getNumberLocalization()) {
        localized = parse(s);
    }
    else {
        localized = s;
    }
    formatter.setMaximumFractionDigits(oldPrec);
    formatter.setMinimumFractionDigits(oldMinPrec);
    return(localized);
}

/**
 * Sets the number formatter to a given locale.
 * @param l
 */
public void setNumberFormatterLocale(Locale l)
{
    formatter = NumberFormat.getInstance(l);
}

}
