/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartUtil.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.JCNumberUtil;


import java.awt.*;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;

/**
 * <code>JCChartUtil</code> is a collection of static methods and static
 * variables that are used throughout JClass Chart.
 */
public class JCChartUtil extends JCNumberUtil {

/**
 * Finds the distance of a given point from the front of the
 * bar, taking focus into account.
 * @param x x value of point
 * @param y y value of point
 * @param left location of left side of bar
 * @param top location of top of bar
 * @param width width of bar
 * @param height height of bar
 * @param invert <code>true</code> if bar is inverted; <code>false</code>
 * otherwise
 * @param focus One of <code>PICK_FOCUS_X</code>, <code>PICK_FOCUS_Y</code>, or
 * <code>PICK_FOCUS_XY</code>. Determines the bias for the pick operation.
 * @return distance of a given point to the front of the bar
 */
static int distanceToFrontRect(int x, int y,
                               int left, int top,
                               int width, int height,
                               boolean invert, int focus)
{
    int right = left + width;
    int bottom = top + height;
    int xtol = 0;
    if (x > right) {
        xtol = x - right;
    }
    else if (x < left) {
        xtol = left - x;
    }
    int ytol = 0;
    if (y < top) {
        ytol = top - y;
    }
    else if (y > bottom) {
        ytol = y - bottom;
    }

    int tol;
    switch (focus) {
    case ChartDataView.PICK_FOCUS_X:
        tol = (invert) ? ytol : xtol;
        break;
    case ChartDataView.PICK_FOCUS_Y:
        tol = (invert) ? xtol : ytol;
        break;
    case ChartDataView.PICK_FOCUS_XY:
    default:
        tol = xtol + ytol;
        break;
    }
    return(tol);
}

/**
 * Finds the distance of a given point to the sides of a bar.  This routine
 * assumes that the pts array defines a rhombus and the mode
 * is always <code>CoorModePrevious</code>.
 * @param x x value of point
 * @param y y value of point
 * @param pts Array of points representing the polygon
 * @param invert <code>true</code> if bar is inverted; <code>false</code>
 * otherwise
 * @param focus One of <code>PICK_FOCUS_X</code>, <code>PICK_FOCUS_Y</code>, or
 * <code>PICK_FOCUS_XY</code>. Determines the bias for the pick operation.
 * @return distance from the point to the polygon
 */
static int distanceTo3dPolygon(int x, int y, Point pts[], boolean invert,
                               int focus)
{
    if (pts == null) {
        return(Integer.MAX_VALUE);
    }

    int left = pts[0].x;
    if (pts[1].x < 0) {         /* left side visible */
        left += pts[1].x;
    }

    int bottom = pts[0].y;
    if (pts[1].y > 0) {         /* bottom showing */
        bottom += pts[1].y;
    }

    int delta_x = Math.abs(pts[1].x);    /* deltas are always positive */
    int delta_y = Math.abs(pts[1].y);
    int width = Math.abs(pts[2].x);    /* height & width are always positive */
    int height = Math.abs(pts[2].y);
    int tol;
    if (pts[2].y == 0) { /* Checking the top or bottom polygon */
        if (pts[2].x < 0) {
            left += pts[2].x;       /* caused by negative chart values */
        }

        if (((pts[1].x > 0) && (pts[1].y > 0)) ||
            ((pts[1].x < 0) && (pts[1].y < 0)))
        {
            /* Along the secondary diagonal */
            tol = distanceTo2ndTop(x, y,
                                   left, bottom, width,
                                   delta_x, delta_y,
                                   invert, focus);
        }
        else {
            /* Along the main diagonal */
            tol = distanceToMainTop(x, y,
                                    left, bottom, width,
                                    delta_x, delta_y,
                                    invert, focus);
        }
    }
    else { /* Checking the left or right polygon */
        if (pts[2].y > 0) {
            bottom += pts[2].y;     /* caused by negative chart values */
        }

        if (((pts[1].x > 0) && (pts[1].y > 0)) ||
            ((pts[1].x < 0) && (pts[1].y < 0)))
        {
            /* Along the secondary diagonal */
            tol = distanceTo2ndSide(x, y,
                                    left, bottom, height,
                                    delta_x, delta_y,
                                    invert, focus);
        }
        else {
            /* Along the main diagonal */
            tol = distanceToMainSide(x, y,
                                     left, bottom, height,
                                     delta_x, delta_y,
                                     invert, focus);
        }
    }

    return(tol);
}

/**
 * Finds the distance of a given point from the side
 * of the bar, taking focus into account.
 * Expects that the bar is rotated through the main diagonal.
 * @param x x value of point
 * @param y y value of point
 * @param left location of left side of bar
 * @param bottom location of the bottom of the bar
 * @param height height of bar
 * @param delta_x 3D translation in the x direction
 * @param delta_y 3D translation in the y direction
 * @param invert <code>true</code> if bar is inverted; <code>false</code>
 * otherwise
 * @param focus One of <code>PICK_FOCUS_X</code>, <code>PICK_FOCUS_Y</code>, or
 * <code>PICK_FOCUS_XY</code>. Determines the bias for the pick operation.
 * @return distance from the point to the side of the bar
 */
static int distanceToMainSide(int x, int y,
                              int left, int bottom, int height,
                              int delta_x, int delta_y,
                              boolean invert, int focus)
{
    if (delta_x == 0) { /* Avoid divide-by-zero in slope calc'n */
        return(distanceToFrontRect(x, y, left, bottom - height - delta_y,
                                   0, height + delta_y, invert, focus));
    }

    int right = left + delta_x;
    int top = bottom - height - delta_y;
    int left_corner = bottom - height;
    int right_corner = bottom - delta_y;
    double m = (double)delta_y / (double)delta_x;

    int xtol = 0;
    int ytol = 0;
    if (y <= top) {         /* Closest point is (right,top) */
        xtol = Math.abs(x - right);
    }
    else if (y >= bottom) { /* Closest point is (left,bottom) */
        xtol = Math.abs(x - left);
    }
    else {
        /* Check if x is left of the rhombus */
        if (y < left_corner) {
            xtol = (int)Math.max(xtol, left +
                                       Math.round((left_corner - y) / m) - x);
        }
        else {
            xtol = Math.max(xtol, left - x);
        }
        /* Check if x is right of the rhombus */
        if (y <= right_corner) {
            xtol = Math.max(xtol, x - right);
        }
        else {
            xtol = (int)Math.max(xtol, x - (left +
                                            Math.round((bottom - y) / m)));
        }
    }
    if (left < x && x < right) {
        ytol = (int)Math.max(ytol, left_corner -
                                   Math.round((x - left) * m) - y);   /* up */
        ytol = (int)Math.max(ytol, y - (bottom -                      /* down */
                                        Math.round((x - left) * m)));
    }
    else {  /* Did not click within left and right bounds of the rhombus */
        ytol = Math.max(ytol, top - y);
        ytol = Math.max(ytol, y - bottom);
    }

    int tol;
    switch (focus) {
    case ChartDataView.PICK_FOCUS_X:
        tol = (invert) ? ytol : xtol;
        break;
    case ChartDataView.PICK_FOCUS_Y:
        tol = (invert) ? xtol : ytol;
        break;
    case ChartDataView.PICK_FOCUS_XY:
    default:
        tol = xtol + ytol;
        break;
    }
    return(tol);
}

/**
 * Finds the distance of a given point from the side
 * of the bar, taking focus into account.
 * Expects that the bar is rotated through the secondary diagonal.
 * @param x x value of point
 * @param y y value of point
 * @param left location of left side of bar
 * @param bottom location of the bottom of the bar
 * @param height height of bar
 * @param delta_x 3D translation in the x direction
 * @param delta_y 3D translation in the y direction
 * @param invert <code>true</code> if bar is inverted; <code>false</code>
 * otherwise
 * @param focus One of <code>PICK_FOCUS_X</code>, <code>PICK_FOCUS_Y</code>, or
 * <code>PICK_FOCUS_XY</code>. Determines the bias for the pick operation.
 * @return distance from the point to the side of the bar
 */
static int distanceTo2ndSide(int x, int y,
                             int left, int bottom, int height,
                             int delta_x, int delta_y,
                             boolean invert, int focus)
{
    if (delta_x == 0) { /* Avoid divide-by-zero in slope calc'n */
        return(distanceToFrontRect(x, y, left, bottom - height - delta_y,
                                   0, height + delta_y, invert, focus));
    }

    int right = left + delta_x;
    int top = bottom - height - delta_y;
    int left_corner = bottom - delta_y;
    int right_corner = bottom - height;
    double m = (double)delta_y / (double)delta_x;

    int xtol = 0;
    int ytol = 0;
    if (y <= top) {         /* Closest point is (left,top) */
        xtol = Math.abs(x - left);
    }
    else if (y >= bottom) { /* Closest point is (right,bottom) */
        xtol = Math.abs(x - right);
    }
    else {
        /* Check if x is right of the rhombus */
        if (y < right_corner) {
            xtol = (int)Math.max(xtol, x - right +
                                       Math.round((right_corner - y) / m));
        }
        else {
            xtol = Math.max(xtol, x - right);
        }

        /* Check if x is left of the rhombus */
        if (y <= left_corner) {
            xtol = Math.max(xtol, left - x);
        }
        else {
            xtol = (int)Math.max(xtol, right - x -
                                       Math.round((bottom - y) / m));
        }
    }
    if (left < x && x < right) {
        ytol = (int)Math.max(ytol, right_corner -
                                   Math.round((right - x) * m) - y);  /* up */
        ytol = (int)Math.max(ytol, y - (bottom -                      /* down */
                                        Math.round((right - x) * m)));
    }
    else {  /* Did not click within left and right bounds of the rhombus */
        ytol = Math.max(ytol, top - y);
        ytol = Math.max(ytol, y - bottom);
    }

    int tol;
    switch (focus) {
    case ChartDataView.PICK_FOCUS_X:
        tol = (invert) ? ytol : xtol;
        break;
    case ChartDataView.PICK_FOCUS_Y:
        tol = (invert) ? xtol : ytol;
        break;
    case ChartDataView.PICK_FOCUS_XY:
    default:
        tol = xtol + ytol;
        break;
    }
    return(tol);
}

/**
 * Finds the distance of a given point from the top
 * or bottom of the bar, taking focus into account.
 * Expects that the bar is rotated through the main diagonal.
 * @param x x value of point
 * @param y y value of point
 * @param left location of left side of bar
 * @param bottom location of the bottom of the bar
 * @param width width of the bar
 * @param delta_x 3D translation in the x direction
 * @param delta_y 3D translation in the y direction
 * @param invert <code>true</code> if bar is inverted; <code>false</code>
 * otherwise
 * @param focus One of <code>PICK_FOCUS_X</code>, <code>PICK_FOCUS_Y</code>, or
 * <code>PICK_FOCUS_XY</code>. Determines the bias for the pick operation.
 * @return distance from the point to the top or bottom of the bar
 */
static int distanceToMainTop(int x, int y,
                             int left, int bottom, int width,
                             int delta_x, int delta_y,
                             boolean invert, int focus)
{
    if (delta_x == 0) { // Avoid divide-by-zero in slope calc'n
        return(distanceToFrontRect(x, y, left, bottom - delta_y,
                                   width, delta_y, invert, focus));
    }

    int right = left + width + delta_x;
    int top = bottom - delta_y;
    int top_corner = left + delta_x;
    int bottom_corner = left + width;
    double m = (double)delta_y / (double)delta_x;

    int xtol = 0;
    int ytol = 0;
    if (top < y && y < bottom) {
        xtol = (int)Math.max(xtol, left + Math.round((bottom - y) / m) - x);
        xtol = (int)Math.max(xtol, x - (bottom_corner +
                                        Math.round((bottom - y) / m)));
    }
    else {
        // did not click within upper and lower bounds of the rhombus
        xtol = Math.max(xtol, left - x);
        xtol = Math.max(xtol, x - right);
    }

    if (x <= left) {
        // closest point is (left,bottom)
        ytol = Math.abs(y - bottom);
    }
    else if (x >= right) {
        // closest point is (right,top)
        ytol = Math.abs(y - top);
    }
    else {
        // Check if y is above the rhombus
        if (x < top_corner) {
            // check the lefthand triangular chunk
            ytol = (int)Math.max(ytol, bottom - Math.round((x - left) * m) - y);
        }
        else {
            ytol = Math.max(ytol, top - y);
        }
        /* Check if y is below the rhombus */
        if (x <= bottom_corner) {
            ytol = Math.max(ytol, y - bottom);
        }
        else {
            // Check the righthand triangular chunk
            ytol = (int)Math.max(ytol,
                                 y - (bottom -
                                      Math.round((x - bottom_corner) * m)));
        }
    }

    int tol;
    switch (focus) {
    case ChartDataView.PICK_FOCUS_X:
        tol = (invert) ? ytol : xtol;
        break;
    case ChartDataView.PICK_FOCUS_Y:
        tol = (invert) ? xtol : ytol;
        break;
    case ChartDataView.PICK_FOCUS_XY:
    default:
        tol = xtol + ytol;
        break;
    }
    return(tol);
}

/**
 * Finds the distance of a given point from the top
 * or bottom of the bar, taking focus into account.
 * Expects the bar is rotated through the secondary diagonal.
 * @param x x value of point
 * @param y y value of point
 * @param left location of left side of bar
 * @param bottom location of the bottom of the bar
 * @param width width of bar
 * @param delta_x 3D translation in the x direction
 * @param delta_y 3D translation in the y direction
 * @param invert <code>true</code> if bar is inverted; <code>false</code>
 * otherwise
 * @param focus One of <code>PICK_FOCUS_X</code>, <code>PICK_FOCUS_Y</code>, or
 * <code>PICK_FOCUS_XY</code>. Determines the bias for the pick operation.
 * @return distance from the point to the top or bottom of the bar
 */
static int distanceTo2ndTop(int x, int y,
                            int left, int bottom, int width,
                            int delta_x, int delta_y,
                            boolean invert, int focus)
{
    if (delta_x == 0) {
        // Avoids divide-by-zero in slope calc'n
        return(distanceToFrontRect(x, y, left, bottom - delta_y,
                                   width, delta_y, invert, focus));
    }

    int right = left + width + delta_x;
    int top = bottom - delta_y;
    int top_corner = left + width;
    int bottom_corner = left + delta_x;
    double m = (double)delta_y / (double)delta_x;

    int xtol = 0;
    int ytol = 0;
    if (top < y && y < bottom) {
        xtol = (int)Math.max(xtol, left + Math.round((y - top) / m) - x);
        xtol = (int)Math.max(xtol, x - (top_corner +
                                        Math.round((y - top) / m)));
    }
    else {
        // Did not click within upper and lower bounds of the rhombus
        xtol = Math.max(xtol, left - x);
        xtol = Math.max(xtol, x - right);
    }

    if (x <= left) {
        // Closest point is (left,top)
        ytol = Math.abs(y - top);
    }
    else if (x >= right) {
        // Closest point is (right,bottom)
        ytol = Math.abs(y - bottom);
    }
    else {
        // Check if y is below the rhombus
        if (x < bottom_corner) {
            // check the lefthand triangular chunk
            ytol = (int)Math.max(ytol, y - top - Math.round((x - left) * m));
        }
        else {
            ytol = Math.max(ytol, y - bottom);
        }
        // Check if y is above the rhombus
        if (x <= top_corner) {
            ytol = Math.max(ytol, top - y);
        }
        else {
            // Check the righthand triangular chunk
            ytol = (int)Math.max(ytol,
                                 bottom - y + Math.round((x - right) * m));
        }
    }

    int tol;
    switch (focus) {
    case ChartDataView.PICK_FOCUS_X:
        tol = (invert) ? ytol : xtol;
        break;
    case ChartDataView.PICK_FOCUS_Y:
        tol = (invert) ? xtol : ytol;
        break;
    case ChartDataView.PICK_FOCUS_XY:
    default:
        tol = xtol + ytol;
        break;
    }
    return(tol);
}

/**
 * Creates an array of line segments from
 * the vertex coordinates of the specified polygon.
 * @return array of line segments
 * @param poly the polygon
 */
public static Line2D.Double[] getLineSegments(Polygon poly)
{
    int nPoints = poly.npoints - 1;
    //  Note that Polygon.xpoints and Polygon.ypoints
    //  have the starting coordinate at the beginning and
    //  end of the arrays
    Line2D.Double[] lineSegments = new Line2D.Double[nPoints];
    for (int i = 0; i < nPoints; i++) {
        lineSegments[i] = new Line2D.Double(poly.xpoints[i],
                                            poly.ypoints[i],
                                            poly.xpoints[i + 1],
                                            poly.ypoints[i + 1]);
    }
    return(lineSegments);
}

/**
 * Returns the minimum pixel distance between a pick point
 * and a specified polygon. If the pick point is inside the polygon,
 * the return value is 0.
 * @param px the x pixel coordinate of the pick point
 * @param py the y pixel coordinate of the pick point
 * @param poly the Polygon
 * @return the pixel distance of the pick point from the polygon
 */
public static double getPolygonDistance(int px, int py, Polygon poly)
{
    if (poly.contains(px, py)) {
        return(0.0);
    }

    double length = Double.MAX_VALUE;
    Line2D.Double[] lineSegments = getLineSegments(poly);
    for (int i = 0; i < lineSegments.length; i++) {
        double len = lineSegments[i].ptSegDist((double)px, (double)py);
        if (len < length) {
            length = len;
        }
    }
    return(length);
}

/**
 * Returns the minimum pixel distance between a pick point
 * and a specified <code>Shape</code> object. If the pick point is inside the
 * <code>Shape</code>, the return value is 0.
 * @param pt the pick point
 * @param shape the Shape
 * @param flatness the maximum distance, from any point on the original curve,
 * that the line segments which are used to approximate curved segments are
 * allowed to deviate
 * @return the pixel distance of the pick point from the <code>Shape</code>
 */
public static double getShapeDistance(Point2D.Double pt, Shape shape,
                                      double flatness)
{
    if (shape.contains(pt)) {
        return(0.0);
    }

    Point2D.Double firstPoint = null;
    Point2D.Double lastPoint = null;
    Line2D.Double lineSegment = null;
    double distance = Double.MAX_VALUE;
    double minDist = Double.MAX_VALUE;
    PathIterator pathIterator = shape.getPathIterator(null, flatness);
    while (!pathIterator.isDone()) {
        double[] coords = new double[6];
        int type = pathIterator.currentSegment(coords);
        switch (type) {
        case PathIterator.SEG_MOVETO:
            firstPoint = lastPoint = new Point2D.Double(coords[0], coords[1]);
            break;
        case PathIterator.SEG_LINETO:
            Point2D.Double newPoint = new Point2D.Double(coords[0], coords[1]);
            lineSegment = new Line2D.Double(lastPoint, newPoint);
            lastPoint = newPoint;
            distance = lineSegment.ptSegDist(pt);
            if (distance < minDist) {
                minDist = distance;
            }
            break;
        case PathIterator.SEG_CLOSE:
            lineSegment = new Line2D.Double(lastPoint, firstPoint);
            distance = lineSegment.ptSegDist(pt);
            if (distance < minDist) {
                minDist = distance;
            }
            break;
        default:
        	
        }
        pathIterator.next();
    }

    return(minDist);
}

/**
 * Returns the minimum pixel distance between a pick point
 * and a specified line segment.
 *
 * @param endPoint1 first point of line segment
 * @param endPoint2 last point of line segment
 * @param point the pick point
 * @return pixel distance
 */
public static double getDistanceToVector(Point2D.Double endPoint1,
                                         Point2D.Double endPoint2,
                                         Point2D.Double point)
{
    Line2D.Double lineSegment = new Line2D.Double(endPoint1, endPoint2);
    return(lineSegment.ptSegDist(point.x, point.y));
}

/**
 * Calculates the tick and numbering increments by taking a prospective
 * increment size and gravitating it towards a 'nice' number, (for example,
 * one ending in 5 or 0).
 * @return
 * @param ts
 * @param prec
 */
public static DoubleCoord calcTik(double ts, int prec)
{
    DoubleCoord rval = new DoubleCoord(0.0, 0.0);
    int len;

    char[] buffer = internalFormat(ts, prec).toCharArray();
    if (prec > 0) {
        len = buffer.length - 1;
        switch (buffer[len]) {
        case '1':
        case '2':
        case '3':
        case '4':
            buffer[len] = '5';
            break;
        case '6':
        case '7':
        case '8':
        case '9':
            buffer[len] = '0';
            for (len--; len >= 0; len--) {
                if (buffer[len] == getDecimalPointChar()) {
                    continue;
                }

                if (buffer[len] != '9') {
                    buffer[len] += 1;
                    break;
                }
                else {
                    buffer[len] = '0';
                }
            }
            if (len < 0) {
                char oldbuff[] = buffer;
                buffer = new char[oldbuff.length + 1];
                len = buffer.length - 1;
                for (; len > 0; len--) {
                    buffer[len] = oldbuff[len - 1];
                }
                buffer[0] = '1';
            }
            break;
        }
    }
    double val = Double.valueOf(new String(buffer)).doubleValue();
    if (val == 0) {
        if (prec >= 0) {
            val = 1.0;
            for (len = 0; len < prec; len++) {
                val /= 10.0;
            }
        }
        else {
            val = 1.0;
            prec = -prec;
            for (len = 0; len < prec; len++) {
                val *= 10.0;
            }
        }
        rval = new DoubleCoord(val / 2, val);
    }
    else {
        rval = new DoubleCoord(val, 2 * val);
    }
    return(rval);
}

/**
 * Returns the radar angle for a data point.
 *
 * @param angleUnits either <code>DEGREES</code>, <code>GRADS</code>, or
 * <code>RADIANS</code>
 * @param numberPoints number of spokes in Radar chart
 * @param index array index of data point
 * @param isReversed <code>true</code> if the direction of the angle axis is reversed
 * @return The radar angle of the specified data point.  The returned value
 * does not include the origin base.
 */
public static double getRadarAngle(int angleUnits, int numberPoints,
                                   int index, boolean isReversed)
{
    double circum;
    switch (angleUnits) {
    case DEGREES:
        circum = 360.0;
        break;
    case GRADS:
        circum = 400.0;
        break;
    default:   //  RADIANS
        circum = M_2PI;
        break;
    }

    double radarAngle = 0.0;
    if (numberPoints != 0) {
        radarAngle = (double)index * circum / (double)numberPoints;
    }
    if (isReversed) {
        radarAngle *= -1.0;
    }

    return(radarAngle);
}

/**
 * Calculates the fraction of a full circle that a specified
 * angle represents.
 *
 * @param unit The units in which the angle is expressed.  Valid values are
 * <code>JCChartUtil.DEGREES</code>, <code>JCChartUtil.GRADS</code>, or
 * <code>JCChartUtil.RADIANS</code>.
 * @param angle the angle for which the angle ratio is to be calculated
 * @param angleOrigin the angle offset for the chart
 * @param clockwise <code>true</code> if the angle values increase in the clockwise
 * direction
 * @return the angle ratio
 */
protected static double getAngleRatio(int unit, double angle,
                                      double angleOrigin,
                                      boolean clockwise)
{
    double dAngle = (clockwise) ? (angleOrigin - angle) : (angleOrigin + angle);
    boolean negative = (dAngle < 0.0) ? true : false;

    double circum;
    switch (unit) {
    case DEGREES:
        circum = 360.0;
        break;
    case GRADS:
        circum = 400.0;
        break;
    default:   //  RADIANS
        circum = JCChartUtil.M_2PI;
        break;
    }

    dAngle = fmod(Math.abs(dAngle), circum);
    if (negative && dAngle != 0.0) {
        dAngle = circum - dAngle;
    }
    double ratio = dAngle / circum + 0.005;

    /* Round ratio off to two decimal places */
    int i_ratio = (int)(ratio * 100.0);
    ratio = ((double)i_ratio) / 100.0;

    ratio = Math.max(0.0, ratio);
    ratio = Math.min(1.0, ratio);

    return(ratio);
}

}
