/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCDataIndex.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.io.Serializable;

/**
 * <code>JCDataIndex</code> contains a unique index to a data point in Chart,
 * consisting of a data series instance and an index representing the
 * point within the data series. It is used by the chart <code>pick()</code>
 * methods, and also contains information related to these operations.
 */
public class JCDataIndex implements Serializable {

/** Chart object picked (for example, legend, chartArea). */
protected Object obj = null;
/** Data view value for the data point.  Null if no series. */
protected ChartDataView  dataView = null;
/** Data series value for the data point.  Null if no series. */
protected ChartDataViewSeries  series = null;
/**
 * Data series index for the data point.  -1 if no series. Equal to
 * <code>JCPieChartFormat.OTHER_SLICE</code> if the other slice in a pie chart
 * has been picked.
 */
protected int  seriesIndex = -1;
/** Index of the point within the data series.  -1 if no selected point.*/
protected int  point = -1;
/** Distance value, used during pick. */
protected int distance = Integer.MAX_VALUE;

/** Simple constructor for <code>JCDataIndex</code>. */
public JCDataIndex() {}

/**
 * Constructor that creates a data index from a series instance and a point
 * index within that series.  It fills in the <code>dataView</code> and
 * <code>seriesIndex</code> members bases on information from the series
 * instance.
 * @param p the point index within the series
 * @param s the data series instance that holds the point
 */
public JCDataIndex(int p, ChartDataViewSeries s)
{
    point = p;
    series = s;
    if (s != null) {
        seriesIndex = s.getParent().getSeriesIndex(s);
        dataView = s.getParent();
    }
}

/**
 * Constructor that creates a data index from the standard information needed
 * to create a data index: a data view, a series instance, the index of
 * that series instance, and a point index.
 * @param dv the data view that holds the series and point
 * @param s the data series instance that holds the point
 * @param i index of the data series within the series list
 * @param p the point index within the series
 */
public JCDataIndex(ChartDataView dv, ChartDataViewSeries s, int i, int p)
{
    dataView = dv;
    point = p;
    series = s;
    seriesIndex = i;
}

/**
 * Constructor that creates a data index from the following information:
 * a series instance, the index of that series, and a point index.
 * @param p the point index within the series
 * @param s the data series instance that holds the point
 * @param i index of the data series within the series list
 */
JCDataIndex(int p, ChartDataViewSeries s, int i)
{
    point = p;
    series = s;
    seriesIndex = i;
}

/**
 * Gets the value of the <code>Series</code> property.  It containts the
 * <code>ChartDataViewSeries</code> instance that holds the point that this data
 * index class is referring to.
 * @return the data series instance
 */
public ChartDataViewSeries getSeries()
{
    // Something wrong -- no series, but we have a data view and an index?
    // try to get a series, then.
    if (series == null && dataView != null && seriesIndex >= 0) {
        try {
            series = dataView.getSeries(seriesIndex);
        }
        catch (Exception e) {
            series = null;
        }
    }
    return(series);
}

/**
 * Gets the value of the <code>Series</code> property.  It containts the
 * <code>ChartDataViewSeries</code> instance that holds the point that this data
 * index class is referring to.
 * @param s the data series instance
 */
public void setSeries(ChartDataViewSeries s)
{
    series = s;
    seriesIndex = s.getParent().getSeriesIndex(s);
    dataView = s.getParent();
}

/**
 * Gets the value of the <code>DataView</code> property.  It contains the
 * <code>ChartDataView</code> instance that holds the point that this data index
 * class is referring to.
 * @return the data view
 */
public ChartDataView getDataView()
{
    return(dataView);
}

/**
 * Sets the value of the <code>DataView</code> property.  It contains the
 * <code>ChartDataView</code> instance that holds the point that this data index
 * class is referring to.
 * @param view the data view
 */
public void setDataView(ChartDataView view)
{
    dataView = view;
    if (dataView != null) {
        if (seriesIndex >= 0) {
            series = dataView.getSeries(seriesIndex);
        }
    }
}


/**
 * Gets the value of the <code>SeriesIndex</code> property.  It contains the
 * index of the data series (specified in the <code>Series</code> property) that
 * contains the point that this data index class is referring to.  Alternately,
 * it may be equal to <code>JCPieChartFormat.OTHER_SLICE</code> if this data
 * index class is referring to the other slice of a pie chart.
 * @return the series index
 */
public int getSeriesIndex()
{
    return(seriesIndex);
}

/**
 * Sets the value of the <code>SeriesIndex</code> property.  It contains the
 * index of the data series (specified in the <code>Series</code> property) that
 * contains the point that this data index class is referring to.  Alternately,
 * it may be equal to <code>JCPieChartFormat.OTHER_SLICE</code> if this data
 * index class is referring to the other slice of a pie chart.
 * @param index the series index
 */
public void setSeriesIndex(int index)
{
    seriesIndex = index;
    if (dataView != null && index >= 0) {
        series = dataView.getSeries(index);
    }
}

/**
 * Gets the value of the <code>Point</code> property.  It contains the index of
 * the point that this data index class refers to within the series (specified
 * by the <code>Series</code> property).
 * @return index of the selected point within the data series
 */
public int getPoint()
{
    return(point);
}

/**
 * Gets the value of the <code>Point</code> property.  It contains the index of
 * the point that this data index class refers to within the series (specified
 * by the <code>Series</code> property).
 * @param p index of the selected point within the data series
 */
public void setPoint(int p)
{
    point = p;
}

/**
 * Gets the value of the <code>Distance</code> property.  It is used by pick
 * operations and contains the distance between the point selected and the
 * nearest data point.
 * @return distance from the selected point
 */
public int getDistance()
{
    return(distance);
}

/**
 * Sets the value of the <code>Distance</code> property.  It is used by pick
 * operations and contains the distance between the point selected and the
 * nearest data point.
 * @param dist distance from the selected point
 */
public void setDistance(int dist)
{
    distance = dist;
}

/**
 * Gets the value of the <code>Obj</code> property.  It is used by pick
 * operations and contains the object picked (for example,
 * <code>JCChartArea</code> or <code>JCLegend</code> instance) by those
 * operations.
 * @return picked object
 */
public Object getObject()
{
    return(obj);
}

/**
 * Sets the value of the <code>Obj</code> property.  It is used by pick
 * operations and contains the object picked (for example,
 * <code>JCChartArea</code> or <code>JCLegend</code> instance) by those
 * operations.
 * @param object picked object
 */
public void setObject(Object object)
{
    obj = object;
}

/**
 * Determines if the specified object refers to the same data point as this
 * <code>JCDataIndex</code> instance.
 * @param o object to compare
 * @return if the passed object is a <code>JCDataIndex</code> instance and
 * refers to the same data point as this instance
 */
public boolean equals(Object o)
{
    if (o == null) {
        return(false);
    }

    if (o instanceof JCDataIndex) {
        JCDataIndex di = (JCDataIndex)o;
        return(dataView == di.dataView &&
               seriesIndex == di.seriesIndex &&
               point == di.point);
    }
    else {
        return(false);
    }
}

/**
 * Creates hash code, depending on the value of the data view hash code plus
 * series index and point value.  This should generate the same
 * <code>hashCode</code> for all equal <code>JCDataIndex</code> objects, and a
 * different <code>hashCode</code> for inequal <code>JCDataIndex</code> objects.
 * @return A hash code value for this object.
 * @see #equals
 * @see java.util.Hashtable
 */
public int hashCode()
{
    return(dataView.hashCode() + (10000 * seriesIndex) + point);
}

/**
 * Returns a string representation of this data index object.
 * @return A string representation of this object.
 */
public String toString()
{
    return("JCDataIndex: point " + point + " series " + seriesIndex +
           " dataview " + dataView);
}

}


