/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCDefaultChartLabelManager.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.util.Enumeration;



import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import java.util.Hashtable;

import java.awt.Point;
import java.awt.Graphics;

import javax.swing.JComponent;
import javax.swing.ToolTipManager;
import javax.swing.BorderFactory;

/**
 * <code>JCDefaultChartLabelManager</code> provides the default implementation
 * for managing ChartLabels on a chart.
 */
public class JCDefaultChartLabelManager implements JCChartLabelManager,
                                                   java.io.Serializable
{

public final static String ERRMSG_INVALID_LABEL_INDEX =
	"Invalid ChartLabel index";

// List of all JCChartLabel instances managed by this manager.  Included
// dwell labels.
protected Vector labels = new Vector();

// List of all dwell label JCChartLabel instanced managed by this manager.
protected Vector dwellLabels = null;

// Currently marked dwell label for popping up
protected JCChartLabel currentDwellLabel = null;

// Hashtable of autoLabel lists (of type JCChartLabel[][]) indexed by
// the ChartDataView they're associated with.
protected Hashtable autoLabelListHashtable = new Hashtable();

// Parent chart of this manager
protected JCChart parentChart = null;

/**
 * Sets the parent chart, whose chart labels are manged by this class.
 * @param chart parent <code>JCChart</code> instance
 */
public void setParentChart(JCChart chart)
{
    parentChart = chart;
}

/**
 * Gets the parent chart, whose chart labels are managed by this class.
 * @return parent <code>JCChart</code> instance
 */
public JCChart getParentChart()
{
    return(parentChart);
}

/**
 * Paints the connectors between the chart labels and the points they are
 * connected to. Only used for dwell labels.
 * @param gc <code>Graphics</code> object of parent chart; used for painting
 * connectors
 */
public void paintConnectors(Graphics gc)
{
    for (Enumeration e = labels.elements(); e.hasMoreElements();) {
        JCChartLabel l = (JCChartLabel)e.nextElement();
        if (l != null               &&
            l.component != null     &&
            l.component.isShowing() &&
            l.isConnected()         &&
            l.isDwellLabel())
        {
            l.drawConnector(gc);
        }
    }
}

/**
 * Adds a chart labelin the process, it forces a recalc and a redraw of the
 * parent chart.
 * @param label the <code>JCChartLabel</code> instance to add
 */
public void addChartLabel(JCChartLabel label)
{
    if (label == null) {
        return;
    }
    appendChartLabel(label);
    label.setChanged(true, Changeable.RECALC);
}

/**
 * Adds a chart label quietly.  If the label is a dwell label, it is also
 * added to the list of dwell labels.
 * @param label
 */
protected void appendChartLabel(JCChartLabel label)
{
    for (Enumeration e = labels.elements(); e.hasMoreElements();) {
        JCChartLabel curLabel = (JCChartLabel)e.nextElement();
        if (curLabel != null && curLabel == label) {
            return;
        }
    }

    labels.addElement(label);

    // If not a dwell label, add as a child of chart.
    // If it is a dwell label, don't add.  It'll be added to the
    // tool tip when necessary.
    JCChart chart = getParentChart();

    // set chart locale on new label
    if (chart != null) {
        label.component.setLocale(chart.getLocale());
    }

    if (!label.isDwellLabel() && chart != null) {
        if (label.isConnected() && label.line != null) {
            chart.add(label.line, 0);
        }
        chart.add(label.component, 0);
    }
    else {
        addToDwellLabels(label);
    }

    label.setParentManager(this);

    // Clean up possibly deadly attachment issues by specifying
    // a default data view where needed.  This case can arise when
    // creating chart labels before the chart (and hence the data view)
    // is known, as in the customizer.
    JCDataIndex di = label.getDataIndex();
    if (di != null && di.getDataView() == null) {
        di.setDataView(chart.getDataView(0));
    }
}

/**
 * Adds a new chart label to the list of dwell labels.
 * @param label the <code>JCChartLabel</code> instance being added to the dwell
 * label list
 */
public void addToDwellLabels(JCChartLabel label)
{
    if (dwellLabels != null && dwellLabels.contains(label)) {
        return;
    }

    if (label.isDwellLabel()) {
        if (dwellLabels == null) {
            dwellLabels = new Vector();

            // Register this component with the ToolTipManager for
            // dwell labels.
            JCChart chart = getParentChart();
            if (chart != null) {
                ToolTipManager ttm = ToolTipManager.sharedInstance();
                ttm.registerComponent(chart);
            }
        }
        dwellLabels.addElement(label);
    }
}

/**
 * Locates the current dwell label (setting the <code>currentDwellLabel</code>
 * variable
 * in the process) that is associated with the given coordinates.
 * @param x the x coordinate within the <code>JCChartArea</code>
 * @param y the y coordinate within the <code>JCChartArea</code>
 * @return <code>true</code> if dwell label found; <code>false</code> otherwise
 */
public boolean locateCurrentDwellLabel(int x, int y)
{
    JCChart chart = getParentChart();
    if (chart != null && dwellLabels != null && dwellLabels.size() != 0) {

        // Build list of possible data views that dwell labels might
        // be attached to and a list of the data indices that a
        // pick() on each data view returns given the current
        // mouse co-ordinates

        List cdvList = chart.getDataView();
        if (cdvList == null || cdvList.size() <= 0) {
            currentDwellLabel = null;
            return(false);
        }

        JCDataIndex[] diList = new JCDataIndex[cdvList.size()];
        Iterator iter = cdvList.iterator();
        for (int i = 0; iter.hasNext(); i++) {
            diList[i] = chart.pick(new Point(x, y), (ChartDataView)iter.next());
        }

        JCChartLabel potentialDwellLabel = null;
        int shortestDistance = Integer.MAX_VALUE;

        // Run through all of the dwell labels.
        // If the dwell label is attach
        for (Enumeration e = dwellLabels.elements(); e.hasMoreElements();) {
            JCChartLabel cl = (JCChartLabel)e.nextElement();
            if (cl == null) {
                continue;
            }
            // Get the data index (as calculated above) for the data view
            // that dwell label is associated with.
            JCDataIndex thisDI = null;

            int dvIndex = cdvList.indexOf(cl.getDataView());
            if (dvIndex >= 0) {
                thisDI = diList[dvIndex];
            }

            // If the dwell label is connected at the current data index and
            // the distance from the mouse pointer to the data index is
            // the smallest we've seen yet, mark it as a potential dwell
            // label.
            if (thisDI != null && cl.shouldPopupHere(thisDI)) {
                if (thisDI.getDistance() <= shortestDistance) {
                    potentialDwellLabel = cl;
                    shortestDistance = thisDI.getDistance();
                }
            }
        }

        // if found a dwell label, note it and return blank text.
        // Otherwise, return null and no tool tip will be brought up.
        if (potentialDwellLabel != null) {
            currentDwellLabel = potentialDwellLabel;
            return(true);
        }
        else {
            currentDwellLabel = null;
            return(false);
        }
    }
    currentDwellLabel = null;
    return(false);
}

/**
 * Gets the <code>AttachPoint</code> property of the current dwell label, if one
 * exists.
 * @return the <code>AttachPoint</code> property of the current dwell label, an
 * instance of the <code>Point</code> object, or null.
 */
public Point getCurrentDwellLabelAttachPoint()
{
    if (currentDwellLabel != null) {
        return(currentDwellLabel.getAttachPoint());
    }
    return(null);
}

/**
 * Gets the <code>Component</code> property of the current dwell label.
 * @return the component part of the current dwell label, an instance of
 * the <code>JComponent</code> class
 */
public JComponent getCurrentDwellLabelComponent()
{
    if (currentDwellLabel != null) {
        return(currentDwellLabel.getComponent());
    }
    return(null);
}

/**
 * Gets the current dwell label.
 * @return the current dwell label, an instance of the <code>JCChartLabel</code>
 * class
 */
public JCChartLabel getCurrentDwellLabel()
{
    return(currentDwellLabel);
}

/**
 * Removes a chart label, forcing a recalc of the label being removed and
 * a redraw of the parent chart.
 * @param label the <code>JCChartLabel</code> instance to remove from the chart
 * label list
 */
public void removeChartLabel(JCChartLabel label)
{
    if (label == null) {
        return;
    }
    deleteChartLabel(label);
    JCChart chart = getParentChart();
    if (chart != null) {
        chart.setChanged(true, Changeable.RECALC);
    }
}

/**
 * Removes a chart label with no update.  If the label is a dwell label, it
 * is also removed from the list of dwell labels.
 * @param label the <code>JCChartLabel</code> instance to remove from the chart
 * label list
 */
protected void deleteChartLabel(JCChartLabel label)
{
    if (label == null) {
        return;
    }
    JCChart chart = getParentChart();
    if (!label.isDwellLabel() && chart != null) {
        if (label.isConnected() && label.line != null) {
            chart.remove(label.line);
        }
        chart.remove(label.component);
    }
    else {
        removeFromDwellLabels(label);
    }

    labels.removeElement(label);
}

/**
 * Removes a chart label to the list of dwell labels.
 * @param label the <code>JCChartLabel</code> instance to remove from the dwell
 * label list
 */
public void removeFromDwellLabels(JCChartLabel label)
{
    if (dwellLabels == null || !dwellLabels.contains(label)) {
        return;
    }

    dwellLabels.removeElement(label);

    if (dwellLabels.size() == 0) {
        // No more dwell labels, so unregister this component with
        // the ToolTipManager for dwell labels.
        JCChart chart = getParentChart();
        if (chart != null) {
            ToolTipManager ttm = ToolTipManager.sharedInstance();
            ttm.unregisterComponent(chart);
        }
        dwellLabels = null;
    }
}

/**
 * Gets the current number of chart labels.
 * @return the number of chart labels this manager holds
 */
public int getNumChartLabels()
{
    int rval = 0;
    try {
        if (labels != null) {
            rval = labels.size();
        }
    }
    catch (Exception e) {}
    return(rval);
}

/**
 * Removes all of the chart labels.  If the <code>AutoLabel</code> property has
 * been set on any of the Chart's DataViews, the property will be set to
 * <code>false</code> and the list of auto labels nulled out.
 */
public void removeAllChartLabels()
{
    JCChart chart = getParentChart();
    if (chart != null) {
        for (Enumeration e = labels.elements(); e.hasMoreElements();) {
            JCChartLabel label = (JCChartLabel)e.nextElement();
            if (label.isConnected() && label.line != null) {
                chart.remove(label.line);
            }
            chart.remove(label.component);
        }
    }

    labels.removeAllElements();
    if (dwellLabels != null) {
        dwellLabels.removeAllElements();
        dwellLabels = null;
    }

    if (chart != null) {
        List data = chart.getDataView();
        for (Iterator i = data.iterator(); i.hasNext();) {
            ChartDataView cdv = (ChartDataView)i.next();
            autoLabelListHashtable.remove(cdv);
            cdv.setAutoLabel(false);
        }
        chart.setChanged(true, Changeable.REDRAW);
    }
}

/**
 * Gets a value of the <code>ChartLabels</code> indexed property.
 * The <code>ChartLabels</code> property controls a list of chart labels
 * managed by the chart.
 * @param index position at which to add the new object
 * @return requested chart label
 */
public JCChartLabel getChartLabels(int index)
{
    if (index < 0 || index >= labels.size()) {
        throw new IllegalArgumentException(ERRMSG_INVALID_LABEL_INDEX);
    }
    return((JCChartLabel)labels.elementAt(index));
}

/**
 * Sets a value of the <code>ChartLabels</code> indexed property.
 * The <code>ChartLabels</code> property controls a list of chart labels
 * managed by the chart.
 * @param index Position at which to add the new object.  To add a new
 * item at the end, set the position to the current size of the labels
 * array.
 * @param lbl new chart label to be added
 * @exception IllegalArgumentException Thrown if the index is larger than
 * the size of the array.
 */
public void setChartLabels(int index, JCChartLabel lbl)
{
    if (index < 0 || index > labels.size()) {
        throw new IllegalArgumentException(ERRMSG_INVALID_LABEL_INDEX);
    }
    JCChart chart = getParentChart();
    if (index < labels.size()) {
        if (lbl == null) {
            removeChartLabel((JCChartLabel)labels.elementAt(index));
        }
        else {
            if (!labels.contains(lbl)) {
                removeFromDwellLabels(((JCChartLabel)labels.elementAt(index)));
                if (chart != null) {
                    chart.remove(
                        ((JCChartLabel)labels.elementAt(index)).component);
                }

                labels.setElementAt(lbl, index);

                if (!lbl.isDwellLabel() && chart != null) {
                    if (lbl.isConnected() && lbl.line != null) {
                        chart.add(lbl.line, 0);
                    }
                    chart.add(lbl.component, 0);
                }
                else {
                    addToDwellLabels(lbl);
                }
            }
            lbl.setParentManager(this);
        }
    }
    else {
        addChartLabel(lbl);
    }
    if (chart != null) {
        chart.setChanged(true, Changeable.RECALC);
    }
}

/**
 * Gets the value of the <code>ChartLabels</code> indexed property.
 * The <code>ChartLabels</code> property controls a list of chart labels
 * managed by the chart.
 * @return List object containing all the chart labels for this JCChart
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this field's or method's return type
 * is based on the corresponding <code>com.sun.java.util</code> class.
 */
public List getChartLabels()
{
    if (labels == null || labels.size() == 0) {
        return(null);
    }
    return(new Vector(labels));
}

/**
 * Sets the value of the <code>ChartLabels</code> indexed property.
 * The <code>ChartLabels</code> property controls a list of chart labels
 * managed by the chart.
 * @param s list object containing chart labels to add to this JCChart
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this method's parameter must be
 * a <code>com.sun.java.util.collections.List</code>.
 */
public void setChartLabels(List s)
{
    JCChart chart = getParentChart();

    if (labels != null && chart != null) {
        for (Enumeration e = labels.elements(); e.hasMoreElements();) {
            chart.remove(((JCChartLabel)e.nextElement()).component);
        }
    }

    labels = new Vector();

    boolean isb = false;
    if (chart != null) {
        isb = chart.isBatched();
        chart.setBatched(false);
    }
    if (s != null && s.size() > 0) {
        for (Iterator iter = s.iterator(); iter.hasNext();) {
            JCChartLabel thisLabel = (JCChartLabel)iter.next();
            labels.addElement(thisLabel);

            if (!thisLabel.isDwellLabel() && chart != null) {
                if (thisLabel.isConnected() && thisLabel.line != null) {
                    chart.add(thisLabel.line, 0);
                }
                chart.add(thisLabel.component, 0);
            }
            else {
                addToDwellLabels(thisLabel);
            }
        }
        for (Iterator iter = s.iterator(); iter.hasNext();) {
            ((JCChartLabel)iter.next()).setParentManager(this);
        }
    }
    if (chart != null) {
        chart.setChanged(true, Changeable.RECALC);
        chart.setBatched(isb);
    }
}

/**
 * Marks all chart labels for later recalculation.
 */
public void markAllChartLabelsForRecalc()
{
    for (Enumeration e = labels.elements(); e.hasMoreElements();) {
        JCChartLabel lbl = (JCChartLabel)(e.nextElement());
        if (lbl != null) {
            lbl.setChanged(true, Changeable.RECALC, false);
        }
    }
}

/**
 * Forces recalculation of any chart labels marked for a recalc.  This
 * is necessary when chart label property changes force a chart layout.
 */
public void recalcChangedChartLabels()
{
    for (Enumeration e = labels.elements(); e.hasMoreElements();) {
        JCChartLabel lbl = (JCChartLabel)(e.nextElement());
        if (lbl != null && lbl.getChanged(Changeable.RECALC)) {
            lbl.recalc();
        }
    }
}

/**
 * Gets the <code>AutoLabelList</code> property for the
 * <code>ChartDataView</code> object. The <code>AutoLabelList</code> property is
 * a two-dimensional array of automatically-generated <code>JCChartLabel</code>
 * instances, one for every point and series.  The inner array is indexed by
 * point; the outer array by series.
 * @param cdv the <code>ChartDataView</code> instance that defines the auto
 * label list
 * @return array of <code>JCChartLabel</code> objects representing the
 * automatically generated chart labels
 */
public JCChartLabel[][] getAutoLabelList(ChartDataView cdv)
{
    if (cdv == null) {
        return(null);
    }

    return((JCChartLabel[][])autoLabelListHashtable.get(cdv));
}

/**
 * Automatically creates <code>JCChartLabels</code> for every series and point,
 * if requested.  The result is saved in the <code>autoLabelList</code>
 * hashtable.
 * @param cdv the <code>ChartDataView</code> instance that defines the auto
 * label list
 */
public void generateAutoLabels(ChartDataView cdv)
{
    JCChartLabel[][] oldAutoList =
        (JCChartLabel[][])autoLabelListHashtable.get(cdv);
    List series = cdv.getSeries();
    int maxi = series.size();
    JCChartLabel[][] autoLabelList = new JCChartLabel[maxi][];
    for (int i = 0; i < maxi; i++) {
        ChartDataViewSeries s;
        s = (ChartDataViewSeries)series.get(i);
        int maxj = s.maxIndex() + 1;
        if (s.getX() == null || s.getY() == null) {
            maxj = 0;
        }
        autoLabelList[i] = new JCChartLabel[maxj];
        for (int j = 0; j < maxj; j++) {
            // Remove old chart labels if they exist
            if (oldAutoList != null    &&
                i < oldAutoList.length &&
                oldAutoList[i] != null &&
                j < oldAutoList[i].length)
            {
                deleteChartLabel(oldAutoList[i][j]);
            }

            // Create and add new chart label for this series and point
            // Default type is dwell label.
            autoLabelList[i][j] =
                createAutoLabel(createAutoLabelText(cdv, i, j), s, j);
            //                createAutoLabel(null, s, j);
            appendChartLabel(autoLabelList[i][j]);
        }
    }
    autoLabelListHashtable.put(cdv, autoLabelList);
}

/**
 * Creates one automatically generated chart label.
 * @return an automatically generated chart label
 * @param srs
 * @param point
 * @param label
 */
protected JCChartLabel createAutoLabel(String label,
                                       ChartDataViewSeries srs, int point)
{
    JCChartLabel cl = new JCChartLabel(label);
    cl.setDataIndex(new JCDataIndex(point, srs));

    cl.setAttachMethod(JCChartLabel.ATTACH_DATAINDEX);
    cl.setAnchor(JCChartLabel.AUTO);
    cl.component.setBorder(BorderFactory.createRaisedBevelBorder());
    cl.setDataView(srs.getParent());
    cl.setOffset(new Point(0, -1));
    cl.setDwellLabel(true);
    return(cl);
}

/**
 * Creates the text for an auto label given the series and point.
 * @return text for an auto label
 * @param cdv the <code>ChartDataView</code> instance that defines the auto label list
 * @param series the series number
 * @param point the point number within the series
 */
public String createAutoLabelText(ChartDataView cdv, int series, int point)
{
    ChartDataViewSeries s = null;
    if (series >= 0) {
        s = cdv.getSeries(series);

        // point outside range of points in series -- return empty string
        if (point < 0 || point >= s.maxXIndex() || point >= s.maxYIndex()) {
            return " ";
        }
    }

    StringBuffer sb = new StringBuffer();
    double xv = 0.0;
    double yv = 0.0;

    // Create y label
    String ylabel = "";

    // Special case -- pie chart other slice
    // pull other slice label from format class
    JCAxis yAxis = cdv.getYAxis();
    if (series == JCPieChartFormat.OTHER_SLICE &&
        cdv.getChartType() == JCChart.PIE)
    {
        ylabel = ((JCPieChartFormat)cdv.getChartFormat()).getOtherLabel();
    }
    // Regular case, generate from y value
    else if (s != null) {
        xv = s.getX(point);
        yv = s.getY(point);
        ylabel = yAxis.getNumberFormatter().localize("" + yv);
    }

    // Use label generator if one is attached
    if (yAxis.getLabelGenerator() != null) {
        Object genlabel =
            yAxis.getLabelGenerator().makeLabel(yv, yAxis.getPrecision());
        if (genlabel instanceof String) {
            ylabel = (String)genlabel;
        }
        else {
            ylabel = genlabel.toString();
        }
    }

    switch (cdv.getChartType()) {
    case JCChart.PLOT:
    case JCChart.SCATTER_PLOT:
        // Create x label
        String xlabel = yAxis.getNumberFormatter().localize("" + xv);
        JCAxis xAxis = cdv.getXAxis();
        JCLabelGenerator labelGenerator = xAxis.getLabelGenerator();
        if (labelGenerator != null) {
            xlabel = (String)labelGenerator.makeLabel(xv, xAxis.getPrecision());
        }
        sb.append(" (");
        sb.append(xlabel);
        sb.append(",");
        sb.append(ylabel);
        sb.append(") ");
        break;
    case JCChart.POLAR:
    case JCChart.RADAR:
    case JCChart.AREA_RADAR:
    case JCChart.HILO:
    case JCChart.HILO_OPEN_CLOSE:
    case JCChart.CANDLE:
    case JCChart.AREA:
    case JCChart.STACKING_AREA:
    case JCChart.BAR:
    case JCChart.STACKING_BAR:
    case JCChart.PIE:
        sb.append(" ");
        sb.append(ylabel);
        sb.append(" ");
        break;
    default :
        sb.append(" ");
        break;
    }
    return(sb.toString());
}

/**
 * Sets the text on an auto label given the series and point.
 * @param cdv the <code>ChartDataView</code> instance that defines the auto
 * label list
 * @param series the series number
 * @param point the point number within the series
 */
public void setAutoLabelText(ChartDataView cdv, int series, int point)
{
    JCChartLabel[][] autoLabelList =
        (JCChartLabel[][])autoLabelListHashtable.get(cdv);
    JCChartLabel jcChartLabel = autoLabelList[series][point];
    if (jcChartLabel != null) {
        jcChartLabel.setText(createAutoLabelText(cdv, series, point));
    }
}

/**
 * Adds some automatically created <code>JCChartLabels</code>.
 * @param cdv the <code>ChartDataView</code> instance that defines the auto
 * label list
 * @param addSeries The series number to which auto labels are added.  If -1,
 * add applies to all series.
 * @param addPoint The point number within a series to which auto labels are
 * added. If -1, auto labels are added to all points within the given series.
 */
public void addAutoLabels(ChartDataView cdv, int addSeries, int addPoint)
{
    JCChartLabel[][] autoLabelList =
        (JCChartLabel[][])autoLabelListHashtable.get(cdv);

    if (autoLabelList == null) {
        return;
    }

    List series = cdv.getSeries();
    int maxi = series.size();

    // add chart labels for all points in a new series
    if (addSeries != -1 && addPoint == -1) {
        int i = addSeries;
        JCChartLabel[][] tmpArr = autoLabelList;
        autoLabelList = new JCChartLabel[maxi][];
        System.arraycopy(tmpArr, 0, autoLabelList, 0, i);
        if (tmpArr.length > i) {
            System.arraycopy(tmpArr, i + 1,
                             autoLabelList, i + 1, maxi - i - 1);
        }

        ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
        int maxj = s.maxIndex() + 1;
        autoLabelList[i] = new JCChartLabel[maxj];
        for (int j = 0; j < maxj; j++) {
            autoLabelList[i][j] =
                createAutoLabel(createAutoLabelText(cdv, i, j), s, j);
            appendChartLabel(autoLabelList[i][j]);
        }
        autoLabelListHashtable.put(cdv, autoLabelList);
    }
    else if (addPoint != -1) {
        int j = addPoint;

        int starti = 0;
        // add chart labels for one new point in one existing series
        if (addSeries != -1) {
            maxi = addSeries + 1;
            starti = addSeries;
        }
        // or add chart labels for one new point in all existing series
        for (int i = starti; i < maxi; i++) {
            ChartDataViewSeries s = (ChartDataViewSeries)series.get(i);
            int maxj = autoLabelList[i].length + 1;

            JCChartLabel[] tmpArr = autoLabelList[i];
            autoLabelList[i] = new JCChartLabel[maxj];
            if (j != 0) {
                System.arraycopy(tmpArr, 0, autoLabelList[i], 0, j);
            }
            if (j != maxj - 1) {
                System.arraycopy(tmpArr, j,
                                 autoLabelList[i], j + 1, maxj - j - 1);
            }

            autoLabelList[i][j] =
                createAutoLabel(createAutoLabelText(cdv, i, j), s, j);
            appendChartLabel(autoLabelList[i][j]);
        }
        autoLabelListHashtable.put(cdv, autoLabelList);
    }
}

/**
 * Removes some of the automatically created <code>JCChartLabels</code>.
 * @param cdv the <code>ChartDataView</code> instance that defines the auto
 * label list
 * @param remSeries The series number to which auto labels are removed.  If -1,
 * remove applies to all series.
 * @param remPoint The point number within a series for which auto labels are
 * removed. If -1, auto labels are removed from all points within the given
 * series.
 */
public void removeAutoLabels(ChartDataView cdv, int remSeries, int remPoint)
{
    JCChartLabel[][] autoLabelList =
        (JCChartLabel[][])autoLabelListHashtable.get(cdv);

    if (autoLabelList == null) {
        return;
    }

    List series = cdv.getSeries();
    int maxi = series.size();

    if (remSeries != -1 && remPoint == -1) {
        int i = remSeries;
        int maxj = autoLabelList[i].length;
        for (int j = 0; j < maxj; j++) {
            deleteChartLabel(autoLabelList[i][j]);
        }
        JCChartLabel[][] tmpArr = autoLabelList;
        autoLabelList = new JCChartLabel[maxi][];
        System.arraycopy(tmpArr, 0, autoLabelList, 0, i);
        if (tmpArr.length > i) {
            System.arraycopy(tmpArr, i + 1,
                             autoLabelList, i, maxi - i);
        }
        autoLabelListHashtable.put(cdv, autoLabelList);
    }
    else if (remPoint != -1) {
        int j = remPoint;
        int starti = 0;
        if (remSeries != -1) {
            maxi = remSeries + 1;
            starti = remSeries;
        }
        for (int i = starti; i < maxi; i++) {
            // Series already has point removed from it.
            int maxj = autoLabelList[i].length;

            deleteChartLabel(autoLabelList[i][j]);

            JCChartLabel[] tmpArr = autoLabelList[i];
            autoLabelList[i] = new JCChartLabel[maxj - 1];
            System.arraycopy(tmpArr, 0, autoLabelList[i], 0, j);
            System.arraycopy(tmpArr, j + 1,
                             autoLabelList[i], j, maxj - j - 1);

        }
        autoLabelListHashtable.put(cdv, autoLabelList);
    }
}

/**
 * Removes all of the automatically created <code>JCChartLabels</code>.
 * @param cdv the <code>ChartDataView</code> instance that defines the auto
 * label list
 */
public void removeAutoLabels(ChartDataView cdv)
{
    JCChartLabel oldList[][];

    JCChartLabel[][] autoLabelList = (JCChartLabel[][])
        autoLabelListHashtable.get(cdv);
    if (autoLabelList == null) return;
    oldList = autoLabelList;
    autoLabelListHashtable.remove(cdv);

    for (int i = 0; i < oldList.length; i++) {
        if (oldList[i] != null) {
            for (int j = 0; j < oldList[i].length; j++) {
                deleteChartLabel(oldList[i][j]);
            }
        }
    }
}

/**
 * Clears the auto labels list for this data view.  Simply removes the
 * list's reference in the internal hashtable.  <br> <B>Note:</b> It does not
 * actually remove the labels from the chart.
 * @param cdv <code>ChartDataView</code> instance from which the auto labels are
 * to be cleared
 */
public void clearAutoLabels(ChartDataView cdv)
{
    autoLabelListHashtable.remove(cdv);
}

/**
 * Updates the parent chart based on the type of change made to the child
 * chart label.
 * @param cf the changed flag
 */
public void updateParent(int cf)
{
    JCChart p = getParentChart();
    if (p != null) {
        if (((cf & Changeable.RECALC) > 0) ||
            ((cf & Changeable.NOTIFY_PARENT_LAYOUT) > 0))
        {
            p.invalidate();
            if (!p.isBatched()) {
                p.validate();
                // This repaint may be redundant in some cases.
                p.repaint();
            }
        }
        else if ((cf & Changeable.REDRAW) > 0) {
            if (!p.isBatched()) {
                p.repaint();
            }
        }
    }
}

}
