/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCSymbolStyle.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.*;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;

/**
 * <code>JCSymbolStyle</code> controls the appearance of points in a
 * plot graph by allowing the modification of
 * the symbol shape, symbol color and symbol size.
 */
public class JCSymbolStyle extends JCStyle {

protected int      shape = DOT;
protected int      colorIndex;
protected int      shapeIndex = 1;
protected Color    color = null;
protected int      size = 6;
protected JCShape  customShape = null;

public final static int  NONE       = 0;
public final static int  DOT        = 1;
public final static int  BOX        = 2;
public final static int  TRIANGLE   = 3;
public final static int  DIAMOND    = 4;
public final static int  STAR       = 5;
public final static int  VERT_LINE  = 6;
public final static int  HORIZ_LINE = 7;
public final static int  CROSS      = 8;
public final static int  CIRCLE     = 9;
public final static int  SQUARE     = 10;
public final static int  OTHER      = 11;
public final static int  LAST       = SQUARE;

/**
 * Constructor for symbol style objects.
 * @param tp symbol shape. One of <code>NONE</code>, <code>DOT</code>,
 * <code>BOX</code>, <code>TRIANGLE</code>, <code>DIAMOND</code>,
 * <code>STAR</code>, <code>VERT_LINE</code>, <code>HORIZ_LINE</code>,
 * <code>CROSS</code>, <code>CIRCLE</code>, or <code>SQUARE</code>.
 * @param c symbol color
 * @param sz symbol size
 */
public JCSymbolStyle(int tp, Color c, int sz)
{
	setShape(tp);
	color = c;
	size = sz;
}

/**
 * Creates a default symbol style.
 * @return <code>JCFillStyle</code> object with default color and symbol
 * style.  The defaults are cycled, so consecutive calls
 * will return different symbol styles.
 * @param c symbol color
 */
public static JCSymbolStyle makeDefault(JCChart c)
{
	JCSymbolStyle style = new JCSymbolStyle(getDefaultShape(c),
                                            getDefaultColor(c), 6);
	int sym = 1;
	int col = 0;
	if (c != null) {
		col = c.getSymbolColorIndex();
		if (col == 0) {
			col = defaultColors.length - 1;
        }
		else {
			col--;
        }

		sym = c.getSymbolShapeIndex();
		if (sym == 1) {
			sym = LAST;
        }
		else {
			sym--;
        }
	}
	style.colorIndex = col;
	style.shapeIndex = sym;
	return(style);
}


/**
 * Draws a symbol at the specified location, using the properties
 * of the <code>JCSymbolStyle</code> instance.
 * @param gc graphics context to use for drawing
 * @param x x position of symbol (in pixels)
 * @param y y position of symbol (in pixels)
 */
public void draw(Graphics gc, int x, int y)
{
	draw(gc, x, y, size, null);
}

/**
 * Draws a symbol at the specified location, using the properties
 * of the <code>JCSymbolStyle</code> instance.
 * @param gc graphics context to use for drawing
 * @param x x position of symbol (in pixels)
 * @param y y position of symbol (in pixels)
 * @param col the color used to draw
 */
public void draw(Graphics gc, int x, int y, Color col)
{
	draw(gc, x, y, size, col);
}

/**
 * Draws a symbol at the specified location, using the properties
 * of the <code>JCSymbolStyle</code> instance.
 * @param gc graphics context to use for drawing
 * @param x x position of symbol (in pixels)
 * @param y y position of symbol (in pixels)
 * @param col the color used to draw
 */
public void draw(Graphics gc, double x, double y, Color col)
{
	draw(gc, x, y, size, col);
}

/**
 * Draws a symbol at the specified location, using the properties
 * of the <code>JCSymbolStyle</code> instance.
 * @param gc graphics context to use for drawing
 * @param x x position of symbol (in pixels)
 * @param y y position of symbol (in pixels)
 * @param sz size of the symbol
 */
public void draw(Graphics gc, int x, int y, int sz)
{
	draw(gc, x, y, sz, null);
}

/**
 * Draws a symbol at the specified location, using the properties
 * of the <code>JCSymbolStyle</code> instance.
 * @param gc graphics context to use for drawing
 * @param x x position of symbol (in pixels)
 * @param y y position of symbol (in pixels)
 * @param sz size of the symbol
 * @param col the color used to draw
 */
public void draw(Graphics gc, int x, int y, int sz, Color col)
{
	Color thisColor = (col != null) ? col : color;
	gc.setColor(thisColor);
	if (customShape != null) { 
		customShape.resize(sz);
		customShape.draw(gc, x, y);
		return;
	}

    boolean even = (sz % 2 == 0);
    int hsize = sz/2;
    if (even) {
        hsize = (sz - 1)/2;
    }

	int delta = (int)(sz/(2 * JCChartUtil.M_SQRT2));
	switch (shape) {
	case NONE:
		break;
	case DOT:
		gc.fillOval(x - sz/2, y - sz/2, sz, sz);
		break;
	case BOX:
		gc.fillRect(x - sz/2, y - sz/2, sz, sz);
		break;
	case STAR:
		gc.drawLine(x - sz/2, y, x + hsize, y);
		gc.drawLine(x, y - sz/2, x, y + hsize);
		gc.drawLine(x - delta, y - delta, x + delta, y + delta);
		gc.drawLine(x - delta, y + delta, x + delta, y - delta);
		break;
	case VERT_LINE:
		gc.drawLine(x, y - sz/2, x, y + hsize);
		break;
	case HORIZ_LINE:
		gc.drawLine(x - sz/2, y, x + hsize, y);
		break;
	case CROSS:
		gc.drawLine(x, y - sz/2, x, y + hsize);
		gc.drawLine(x - sz/2, y, x + hsize, y);
		break;
	case CIRCLE:
		gc.drawArc(x - sz/2, y - sz/2, sz - 1, sz - 1, 0, 360);
		break;
	case SQUARE:
		gc.drawRect(x - sz/2, y - sz/2, sz - 1, sz - 1);
		break;
	}
}

/**
 * Draws a symbol at the specified location, using the properties
 * of the <code>JCSymbolStyle</code> instance.
 * @param gc graphics context to use for drawing
 * @param x x position of symbol (in pixels)
 * @param y y position of symbol (in pixels)
 * @param sz size of the symbol
 * @param col the color used to draw
 */
public void draw(Graphics gc, double x, double y, int sz, Color col)
{
    Color thisColor = (col != null) ? col : color;
    gc.setColor(thisColor);
    if (customShape != null) {
        customShape.resize(sz);
        customShape.draw(gc, x, y);
        return;
    }

    boolean even = (sz % 2 == 0);
    int hsize = sz / 2;
    if (even) {
        hsize = (sz - 1) / 2;
    }

    int delta = (int)(sz / (2 * JCChartUtil.M_SQRT2));

    Graphics2D g2d = (Graphics2D)gc;
    switch (shape) {
    case NONE:
        break;
    case DOT:
        g2d.fill(new Ellipse2D.Double(x - (double)sz / 2.0,
                                      y - (double)sz / 2.0,
                                      (double)sz, (double)sz));
        break;
    case BOX:
        g2d.fill(new Rectangle2D.Double(x - (double)sz / 2.0,
                                        y - (double)sz / 2.0,
                                        (double)sz, (double)sz));
        break;
    case STAR:
        g2d.draw(new Line2D.Double(x - (double)sz / 2.0, y,
                                   x + (double)hsize, y));
        g2d.draw(new Line2D.Double(x, y - (double)sz / 2.0,
                                   x, y + (double)hsize));
        g2d.draw(new Line2D.Double(x - (double)delta, y - (double)delta,
                                   x + (double)delta, y + (double)delta));
        g2d.draw(new Line2D.Double(x - (double)delta, y + (double)delta,
                                   x + (double)delta, y - (double)delta));
        break;
    case VERT_LINE:
        g2d.draw(new Line2D.Double(x, y - (double)sz / 2.0,
                                   x, y + (double)hsize));
        break;
    case HORIZ_LINE:
        g2d.draw(new Line2D.Double(x - (double)sz / 2.0, y,
                                   x + (double)hsize, y));
        break;
    case CROSS:
        g2d.draw(new Line2D.Double(x, y - (double)sz / 2.0,
                                   x, y + (double)hsize));
        g2d.draw(new Line2D.Double(x - (double)sz / 2.0, y,
                                   x + (double)hsize, y));
        break;
    case CIRCLE:
        g2d.draw(new Arc2D.Double(x - (double)sz / 2.0,
                                  y - (double)sz / 2.0,
                                  (double)(sz - 1), (double)(sz - 1),
                                  0.0, 360.0, Arc2D.OPEN));
        break;
    case SQUARE:
        g2d.draw(new Rectangle2D.Double(x - (double)sz / 2.0,
                                        y - (double)sz / 2.0,
                                        (double)(sz - 1), (double)(sz - 1)));
        break;
    }
}

/**
 * Gets the <code>Shape</code> property.<p> The <code>Shape</code> property
 * determines the shape of symbol that will be drawn.
 * @return One of <code>NONE</code>, <code>DOT</code>, <code>BOX</code>,
 * <code>TRIANGLE</code>, <code>DIAMOND</code>, <code>STAR</code>,
 * <code>VERT_LINE</code>, <code>HORIZ_LINE</code>, <code>CROSS</code>,
 * <code>CIRCLE</code>, or <code>SQUARE</code>.
 */
public int getShape()
{
    return(shape);
}

/**
 * Sets the <code>Shape</code> property.<p>  The <code>Shape</code> property
 * determines the shape of symbol that will be drawn.
 * @param tp one of <code>NONE</code>, <code>DOT</code>, <code>BOX</code>,
 * <code>TRIANGLE</code>, <code>DIAMOND</code>, <code>STAR</code>,
 * <code>VERT_LINE</code>, <code>HORIZ_LINE</code>, <code>CROSS</code>,
 * <code>CIRCLE</code>, or <code>SQUARE</code>.
 * @exception IllegalArgumentException If the shape value is not one of
 * <code>NONE</code>, <code>DOT</code>, <code>BOX</code>, <code>TRIANGLE</code>,
 * <code>DIAMOND</code>, <code>STAR</code>, <code>VERT_LINE</code>,
 * <code>HORIZ_LINE</code>, <code>CROSS</code>, <code>CIRCLE</code>, or
 * <code>SQUARE</code>.
 */
public void setShape(int tp)
{
	if (shape == tp) {
        return;
    }

	shapeIndex = tp;
	switch (tp) {
	case NONE:
	case DOT:
	case BOX:
	case STAR:
	case VERT_LINE:
	case HORIZ_LINE:
	case CROSS:
	case CIRCLE:
	case SQUARE:
	case OTHER:
		shape = tp;
		customShape = null;
		break;
	case TRIANGLE:
		shape = tp;
		customShape = new Triangle();
		break;
	case DIAMOND:
		shape = tp;
		customShape = new Diamond();
		break;
	default:
		throw new IllegalArgumentException("Invalid value for Shape in " +
                                           "JCSymbolStyle.");
	}
	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Gets the <code>Color</code> property of <code>JCSymbolStyle</code>, which
 * determines the color used to paint the symbols.
 * @return AWT Color class representing the color to be used
 * to paint the symbols. If null, the current color of the <code>Graphics</code>
 * object is used.
 */
public Color getColor()
{
    return(color);
}

/**
 * Sets the <code>Color</code> property of <code>JCSymbolStyle</code>, which
 * determines the color used to paint the symbols.
 * @param clr AWT Color class representing the color to be used
 * to paint the symbols. If null, the current color of the <code>Graphics</code>
 * object is used.
 */
public void setColor(Color clr)
{
	if (color == clr) {
        return;
    }
	color = clr;

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Gets the <code>Size</code> property.<p> The <code>Size</code> property
 * determines the size of the symbols. Note that a value of zero size means
 * the symbol will not be drawn.
 * @return positive integer representing size
 */
public int getSize()
{
    return(size);
}

/**
 * Sets the <code>Size</code> property of <code>JCSymbolStyle</code>, which
 * determines the size of the symbols. <br><b>Note:</b> A value of zero size
 * means the symbol will not be drawn.
 * @param sz positive integer representing size
 */
public void setSize(int sz)
{
	if (size == sz) {
        return;
    }
	size = sz;
	// HLOC and Candle chart types need a RECALC
	setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_REDRAW);
}

static int getDefaultShape(JCChart c)
{
	int rval;
	if (c == null) {
		rval = DOT;
	}
	else {
		int localIndex = c.getSymbolShapeIndex();
		rval = localIndex++;
		if (localIndex > LAST) {
            localIndex = DOT;
        }
		c.setSymbolShapeIndex(localIndex);
	}
	return(rval);
}

/**
 * Returns the index used to generate the color.  Only set
 * if default colors used.
 * @return index used to generate the color
 */
public int getColorIndex()
{
	return(colorIndex);
}

static Color getDefaultColor(JCChart c)
{
	Color rval;
	if (c == null) {
		rval = defaultColors[0];
	}
	else {
		int localIndex = c.getSymbolColorIndex();
		rval = defaultColors[localIndex++];
		if (localIndex >= defaultColors.length) {
			localIndex = 0;
		}
		if (c != null && c.getChartArea() != null && 
			JCChartUtil.colorsNear(rval, 
				c.getChartArea().getPlotArea().getBackground())) {
			rval = defaultColors[localIndex++];
			if (localIndex >= defaultColors.length) {
				localIndex = 0;
            }
		}
		c.setSymbolColorIndex(localIndex);
	}
	return(rval);
}

/**
 * Gets the value of the <code>CustomShape</code> property.<p>
 * The <code>CustomShape</code> property contains an object derived from
 * <code>JCShape</code> that is used to draw symbols.
 * @return object derived from <code>JCShape</code> that defined the shape
 * to be drawn for the symbols
 * @see com.klg.jclass.chart.JCShape
 */
public JCShape getCustomShape()
{
	return(customShape);
}

/**
 * Sets the value of the <code>CustomShape</code> property.<p>
 * The <code>CustomShape</code> property contains an object derived from
 * <code>JCShape</code> that is used to draw symbols.
 * @param ns object derived from <code>JCShape</code> that defined the shape
 * to be drawn for the symbols
 * @see com.klg.jclass.chart.JCShape
 */
public void setCustomShape(JCShape ns)
{
	if (customShape == ns) {
        return;
    }
    shape = (ns == null) ? NONE : OTHER;
	customShape = ns;

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

}


