/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCValueLabel.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

/**
 * A <code>JCValueLabel</code> is an association between
 * a floating-point axis value and a <code>ChartText</code>.
 * <code>JCValueLabels</code> are used to implement the
 * <code>VALUE_LABELS</code> annotation method, in
 * which axis labels are determined by the
 * axis values the user indicates should be
 * labelled.
 */
public class JCValueLabel implements java.io.Serializable {

protected double       value=0.0;
protected ChartText    label = new ChartText();
protected boolean      drawTick = true;
protected boolean      drawLabel = true;
protected boolean      drawGrid = true;
protected JCAnno       anno = null;
protected JCDataIndex  dataIndex = null;

/**
 * Retrieves the <code>ChartText</code> associated with this
 * <code>JCValueLabel</code>.
 * @return ChartText instance
 */
public ChartText getChartText()
{
    return(label);
}

/**
 * Retrieves the <code>JCAnno</code> associated with this
 * <code>JCValueLabel</code>.
 * @return JCAnno instance
 */
public JCAnno getAnno()
{
    return(anno);
}

/**
 * Sets the <code>JCAnno</code> property for this
 * <code>JCValueLabel</code>.
 */
public void setAnno(JCAnno anno)
{
    this.anno = anno;
    // Initialize drawTick, drawLabel, and label color to the
    // values in the anno object.
    if (anno != null) {
        drawTick = anno.getDrawTicks();
        drawLabel = anno.getDrawLabels();
        label.foreground = anno.getLabelColor();
    }
}

/**
 * Gets the value of the <code>Value</code> property.<p>
 * The <code>Value</code> property controls the position of a label in
 * data space along a particular axis.
 * @return data-space value.
 */
public double getValue()
{
    return(value);
}

/**
 * Sets the value of the <code>Value</code> property.<p>
 * The <code>Value</code> property controls the position of a label in
 * data space along a particular axis.
 * @param nv data-space value
 */
public void setValue(double nv)
{
    if (value == nv) {
        return;
    }
    value = nv;
}

/**
 * Gets the value of the <code>Text</code> property.<p>
 * The <code>Text</code> property specifies the text displayed inside the label.
 * @return String representing the text inside the label. This String can
 * be unparsed text.
 */
public String getText()
{
    return(label.getText());
}

/**
 * Sets the value of the <code>Text</code> property.<p>
 * The <code>Text</code> property specifies the text displayed inside the label.
 * @param t String representing the text inside the label. This String can
 * be unparsed text.
 */
public void setText(String t)
{
    label.setText(t);
}

/**
 * Default constructor required for JavaBeans.
 */
public JCValueLabel()
{
    value = 0.0;
}

/**
 * A smart constructor that creates a <code>ChartText</code> from a generic
 * object.
 * @param val floating-point axis value
 * @param lbl String to use as the value label
 * @param axis axis on which the value label will appear
 * @see com.klg.jclass.chart.ChartText
 */
public JCValueLabel(double val, String lbl, JCAxis axis)
{
    value = val;
    label = new ChartText(null, lbl, axis);
}

/**
 * Internal constructor.
 * @param val floating-point axis value
 * @param lbl String to use as the value label
 * @param axis axis on which the value label will appear
 * @param drawTick If true, a tick is drawn for this JCValueLabel.
 * @param drawLabel If true, the ChartText will be drawn.
 * @param drawGrid If true, a grid line will be drawn at this label.
 */
public JCValueLabel(double val, String lbl, JCAxis axis,
                    boolean drawTick, boolean drawLabel, boolean drawGrid) {
    this(val, lbl, axis);
    this.drawTick = drawTick;
    this.drawLabel = drawLabel;
    this.drawGrid = drawGrid;
}

/**
 * Internal constructor that creates a value label given a value
 * and a String.
 * @param val floating-point axis value
 * @param lbl string to use as the value label
 */
public JCValueLabel(double val, String lbl)
{
    value = val;
    label.updateParent = false;
    label.setText(lbl);
    label.updateParent = true;
}

/**
 * Internal constructor that creates a value Label from a JCAnno object.
 */
public JCValueLabel(JCAnno anno, double val)
{
    this.value = val;
    this.anno = anno;
    if (anno != null) {
        // Initialize drawTick, drawLabel, and label color to the
        // values in the anno object.
        drawTick = anno.getDrawTicks();
        drawLabel = anno.getDrawLabels();
        initValueLabel(anno.getParent(), val, anno.getPrecision());
        label.foreground = anno.getLabelColor();
    }
}

/**
 * Sets the parent for the <code>JCValueLabel</code> instance.  Internal only.
 * @param ax axis on which the value label will appear
 */
protected void setParent(JCAxis ax)
{
    if (label.parentChart == null) {
        label.setParentChart(ax.getParentChart());
    }

    label.setParentAxis(ax);
    if (label.getHeight() == 0) {
        label.setChanged(true, Changeable.RECALC, false);
        label.recalc();
    }
}

private void initValueLabel(JCAxis parent, double val, int prec)
{
    this.value = val;
    JCChart c = (parent == null) ? null : parent.getParentChart();
    if (parent != null && parent.labelGenerator != null) {
        setGeneratedLabel(c, val, prec, parent.labelGenerator, parent);
    }
    else {
        String s = JCChartUtil.format(val, prec);
        String text = null;
        // Localize our number if we can
        if (parent != null) {
            JCChartNumberFormat formatter = null;
            formatter = parent.getNumberFormatter();
            text = formatter.localize(s, prec);
            label = new ChartText(c, text, parent);
        }
        else {
            label = new ChartText(c, s, parent);
        }

        label.updateParent = false;
        if (parent != null) {
            label.setRotation(parent.annotationRotation);
        }
        label.updateParent = true;
    }
    setParent(parent);
}

/**
 * Convenience constructor for use within JClass Chart.  Uses the precision
 * to create a formatted String.
 * @param parent pointer to the parent axis object
 * @param val
 * @param prec precision to send to label generator
 */
protected JCValueLabel(JCAxis parent, double val, int prec)
{
    initValueLabel(parent, val, prec);
}

/**
 * Convenience constructor used by JClass Chart to create value labels for
 * log values.
 * @param parent pointer to the parent axis object
 * @param exponent power of 10 to be used to generate the label
 * @param prec precision to send to label generator
 */
protected JCValueLabel(JCAxis parent, int exponent, int prec)
{
    value = JCChartUtil.pow10(exponent);
    JCChart c = (parent == null) ? null : parent.getParentChart();
    if (parent != null && parent.labelGenerator != null) {
        setGeneratedLabel(c, (double)exponent, prec, parent.labelGenerator,
                          parent);
    }
    else {
        String s = JCChartUtil.logFormat(exponent);
        String text = null;
        // Localize our number if we can
        if (parent != null) {
            JCChartNumberFormat formatter = null;
            formatter = parent.getNumberFormatter();
            text = formatter.localize(s, prec);
            label = new ChartText(c, text, parent);
        }
        else {
            label = new ChartText(c, s, parent);
        }
        label.updateParent = false;
        if (parent != null) {
            label.setRotation(parent.annotationRotation);
        }
        label.updateParent = true;
    }
    setParent(parent);
}

private void setGeneratedLabel(JCChart c, double value, int prec,
                               JCLabelGenerator lg, JCAxis axis)
{
    if (lg == null) return;
    Object o = lg.makeLabel(value, prec);
    if (o instanceof ChartText) {
        label = (ChartText)o;
        label.setParentChart(c);
    }
    else {
        String s = (o instanceof String) ? (String)o : o.toString();
        label = new ChartText(c, s, axis);

        label.updateParent = false;
        if (axis != null) {
            label.setRotation(axis.annotationRotation);
        }
        label.updateParent = true;
    }
}

/**
 * Override toString to return the text
 * @return The text for this value label.
 */
public String toString()
{
    return(label.getText());
}

/**
 * Is the tick drawn for this value label?
 * @return true, if the tick is drawn.
 */
public boolean isDrawTick()
{
    return(drawTick);
}

/**
 * Set whether the tick is drawn for this value label.
 * @param drawTick The new value of drawTick.
 */
public void setDrawTick(boolean drawTick)
{
    this.drawTick = drawTick;
}

/**
 * Is a grid line drawn for this value label?
 * @return true, if a grid line is drawn.
 */
public boolean isDrawGrid()
{
    return(drawGrid);
}

/**
 * Set whether the grid line is drawn for this value label.
 * @param drawGrid The new value of drawGrid.
 */
public void setDrawGrid(boolean drawGrid)
{
    this.drawGrid = drawGrid;
}

/**
 * Ist the label drawn for this value label?
 * @return true, if the label is drawn.
 */
public boolean isDrawLabel()
{
    return(drawLabel);
}

/**
 * Set whether the label is drawn for this value label.
 * @param drawLabel The new value of drawLabel.
 */
public void setDrawLabel(boolean drawLabel)
{
    this.drawLabel = drawLabel;
}

/**
 * Get the stored data index for this value label.  Used to store the dataView,
 * series, point information to point back to where this label was generated.
 * @return The current data index object.
 */
protected JCDataIndex getDataIndex()
{
    return(dataIndex);
}

/**
 * Set the stored data index for this value label.  Used to store the dataView,
 * series, point information to point back to where this label was generated.
 * @param dataIndex The new data index object.
 */
protected void setDataIndex(JCDataIndex dataIndex)
{
    this.dataIndex = dataIndex;
}

}


