/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PieChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:47 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Drawable object for pie charts.  This object
 * is used for rendering an pie chart based on data
 * contained in the <code>dataObject</code>.
 */
public class PieChartDraw extends ChartDraw
{

/** Indicates pie coordinates have not been calculated yet. */
public final static int COORD_DEFAULT_VALUE = -2;

/** Indicates pie coordinates are of a zero slice. */
public final static int COORD_ZERO_SLICE = -1;

/** Indicates index of a data point with a hole value. */
public final static int INDEX_HOLE = -2;

/** Indicates index of a data point that will not be drawn. */
public final static int INDEX_NOT_DRAWN = -1;

/** Indicates where a point is relative to a pie.  Used by <code>pick()</code>. */
public final static int HIT_NOWHERE = 0;
public final static int HIT_TOP = 1;
public final static int HIT_EDGE = 2;

// Error Message -- In Theory, this should never appear if we have
// covered all of our bases correctly
public final static String ERRMSG_BIG_ERROR = "Big Error!";

// Local copies of data needed by draw(), pick(), and unpick().  Set
// in recalc.
protected PieChartInfo       pieInfo = null;
protected JCPieChartFormat   pcf = null;
protected int                delta_y = 0;
protected JCLineStyle        outlineStyle = null;
protected int                shapeOp = ProcessShapeModel.SHAPE_RENDER;
protected Picker             picker = null;
protected ProcessShapeModel  processShape = null;
protected Mapper             mapper = null;

/** Number of pie charts in this <code>ChartDataView</code>. */
protected int num_pies = 0;

/** Number of lines pie charts are displayed on. */
protected int num_lines = 0;

/** Number of pie charts per line. */
protected int num_pies_line = 0;

/** Width and height of bounding box each pie chart is drawn in. */
protected int box_width = 0, box_height = 0;

/** Offset of pie from the edge of the bounding box. */
protected int pie_offset = 0;

/** First point of data to be used. */
protected int first_point = 0;

/** Left and top positions of the overall graph area. */
protected int graph_left = 0, graph_top = 0;

/** Array of labels for each pie chart. */
transient protected ChartText[] cLabels;

/** Maximum height of chart labels. */
protected int label_height = 0;

/** Distance between the center of the pie and the exploded slices. */
protected int explode_offset = 0;

// Set to true when setRecalcFlag() is called in the update()
// method of JCChart.  Part of fix for PR 18172.
protected boolean recalcTransientData = false;

/**
 * Default constructor.
 */
public PieChartDraw()
{
    super();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc()
{
    super.recalc();

    // First calculate transient data
    calcTransientData();

    pieInfo = (PieChartInfo)dataObject.getTransientData();

    // Get list of data series to be drawn using this drawable
    Rectangle pa = chartArea.getPlotRect();

    graph_left = pa.x;
    graph_top = pa.y;
    int graph_width = Math.max(pa.width - 1, 0);
    int graph_height = Math.max(pa.height - 1, 0);

    // Get size of point Labels
    int numLabels = dataObject.getNumPointLabels();

    // See if we have a non-empty point label String.  If not,
    // set numLabels to zero.
    if (numLabels > 0) {
        boolean fNonEmptyLabelExists = false;
        for (int i = 0; i < numLabels; i++) {
            String pointLabel = dataObject.getPointLabel(i);
            if (pointLabel != null && !pointLabel.equals("")) {
                fNonEmptyLabelExists = true;
                break;
            }
        }
        if (!fNonEmptyLabelExists) {
            numLabels = 0;
        }
    }

    label_height = 0;
    cLabels = new ChartText[numLabels];
    for (int i = 0; i < numLabels; i++) {
        String pointLabel = dataObject.getPointLabel(i);
        if (pointLabel != null) {
            cLabels[i] = new ChartText();
            cLabels[i].setText(pointLabel);
            cLabels[i].setParentChart((JCChart)chartArea.getParent());
            cLabels[i].setParentRegion(chartArea);
            int cHeight = cLabels[i].getHeight();
            label_height = Math.max(label_height, cHeight);
        }
        else {
            cLabels[i] = null;
        }
    }

    if (num_pies == 0) {
        return;
    }

    /*
     * The dimension and position of each pie within its enclosing
     * rectangle/box is the same.  Compute this constant information
     * once, before drawing the pies.
     */
    int[] passback = pieLayout(num_pies, pa);
    num_lines = passback[0];
    num_pies_line = passback[1];

    box_width = graph_width / num_pies_line;
    box_height = graph_height / num_lines;

    explode_offset = pcf.getExplodeList() == null ? 0 : pcf.getExplodeOffset();

    int avail_width =
        Math.max(box_width - (label_height * 2) - (explode_offset * 2) - 2, 0);
    int avail_height =
        Math.max(box_height - (label_height * 2) - (explode_offset * 2) - 2, 0);

    double elevation =
        JCChartUtil.degToRad(Math.max(0, chartArea.getElevation()));
    double sin_elevation = Math.sin(elevation);
    double cos_elevation = Math.cos(elevation);
    double factor = (double)chartArea.getDepth() / 100.0;

    int pie_width = 0;
    if (avail_width * (cos_elevation + factor * sin_elevation) < avail_height) {
        pie_width = avail_width;
    }
    else {
        pie_width = (int)(avail_height / (cos_elevation +
                                          factor * sin_elevation));
    }
    int pie_height = (int)(pie_width * cos_elevation);

    // Override deltaX and deltaY with special pie chart version.
    // We can override this here because pie charts don't have axes.
    chartArea.setDeltaX(0);
    delta_y = (int)(pie_width * factor * sin_elevation);
    chartArea.setDeltaY(delta_y);

    // Pie width and height should always be odd.
    if (pie_width != (pie_width | 1)) {
        pie_width = (pie_width | 1) - 2;
        if (pie_width < 1) {
            pie_width = 1;
        }
    }
    if (pie_height != (pie_height | 1)) {
        pie_height = (pie_height | 1) - 2;
        if (pie_height < 1) {
            pie_height = 1;
        }
    }

    pieInfo.width = pie_width;
    pieInfo.height = pie_height;

    pie_offset = (int)((box_width - pie_width) / 2.0);
    if (pie_offset < 0) {
        pie_offset = 0;
    }
}

/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc)
{

    if (!okToProceed(gc)) {
        return;
    }

    /*
     *  Draw the pies -- draw the slices in their boxes taking into account
     *                   the depth and inclination -- save some information
     *                   in "PiePosition" structure to make picking easier
     *                -- draw any text underneath the pies
     */
    int pie_num = 0;
    int box_x = graph_left;
    int box_y = graph_top;
    for (int k = 0; k < num_lines; k++) {
        for (int l = 0; l < num_pies_line; l++) {
            if (pie_num < num_pies) {
                /*  Draw the pie taking into account depth and inclination */
                drawPie(gc, pie_num, box_x, box_y);
            }
            pie_num++;
            box_x += box_width;
        }
        box_x = graph_left;
        box_y += box_height;
    }
}

protected void drawSinglePie(Graphics gc, int pie_num, int box_x, int box_y)
{

    if (!okToProceed(gc)) {
        return;
    }
    drawPie(gc, pie_num, box_x, box_y);
}

private boolean okToProceed(Graphics gc)
{
    // gc is null when image mapping
    if (shapeOp != ProcessShapeModel.SHAPE_IMAGEMAP) {
        Rectangle clipBounds = gc.getClipBounds();
        if (clipBounds != null &&
            (clipBounds.width == 0 || clipBounds.height == 0))
        {
            return(false);
        }

        Rectangle pa = chartArea.getPlotRect();
        gc.translate(pa.x, pa.y);
    }

    if (pieInfo == null) {
        recalc();
    }

    if (num_pies == 0) {
        return(false);
    }

    return(true);
}

protected Point getPointLabelLocation(ChartText pie_label, int box_x, int box_y)
{
    Point p = new Point();
    int y = box_y + (box_height - pieInfo.height - delta_y - label_height) / 2;
    if (pie_label != null) {
        int cWidth = pie_label.getWidth();
        p.setLocation(box_x + ((box_width - cWidth) / 2),
                      y + pieInfo.height + delta_y + explode_offset);
    }
    return(p);
}

/*
* This private method draws a single pie contained in a bounding box
* with origin at (<i>x</i>,<i>y</i>).
*/
private void drawPie(Graphics gc, int pie_num, int box_x, int box_y)
{
    int x = box_x + pie_offset;
    int y = box_y + (box_height - pieInfo.height - delta_y - label_height) / 2;

    outlineStyle = dataObject.getOutlineStyle();

    switch (shapeOp) {
    case ProcessShapeModel.SHAPE_PICK:
        picker.setPieData(pieInfo.data[pie_num]);
        processShape = new PickShape(picker);
        break;
    case ProcessShapeModel.SHAPE_IMAGEMAP:
        mapper.setPieData(pieInfo.data[pie_num]);
        processShape = new ImageMapShape(mapper);
        break;
    default:
        processShape = new RenderShape(gc, outlineStyle, null);
    }

    if (!pieInfo.data[pie_num].zero_pie) {
        Rectangle fillOrientRect =
            new Rectangle(pieInfo.pos[pie_num].cx, pieInfo.pos[pie_num].cy,
                          pieInfo.width, pieInfo.height);
        /* Non-zero inclination or depth */
        if (delta_y != 0) { /* Non-zero inclin'n or depth */
            /*
             * Draw the backfacing plane.  This is the same as the
             * front plane (i.e. the top) except it is shifted down
             * (to match the bottom arc).  This needs to be done
             * because the bottom of the edge along the arc needs
             * to be filled in.
             * Almost all of what gets drawn here, gets drawn
             * over by the pie edge and front plane of the disc.
             */
            drawPiePlane(gc, seriesList, pieInfo.data[pie_num],
                         pieInfo, x, y + delta_y, pie_num, false,
                         fillOrientRect);

            /*
             * Now draw the edge of the disc (this will be drawn
             * over the back plane).
             */
            drawPieEdge(gc, pieInfo.data[pie_num], pieInfo, x, y, pie_num,
                        fillOrientRect);
        }

        /* Draw the front plane (or top of the pie) */
        drawPiePlane(gc, seriesList, pieInfo.data[pie_num],
                     pieInfo, x, y, pie_num, true, fillOrientRect);

        /* Draw the lines separating the slices */
        drawPieLines(pieInfo.data[pie_num], pieInfo, x, y, pie_num);
    }

    /*  Draw the pie label underneath the pie. */
    if (shapeOp == ProcessShapeModel.SHAPE_RENDER &&
        pie_num + first_point < cLabels.length)
    {
        ChartText pie_label = cLabels[pie_num + first_point];
        if (pie_label != null) {
            Point p = getPointLabelLocation(pie_label, box_x, box_y);
            pie_label.setLocation(p.x, p.y);
            pie_label.draw(gc);
        }
    }

    /* Draw a circle around the pie */
    if (!pieInfo.data[pie_num].zero_pie) {
        drawPieCircle(pieInfo.data[pie_num], pieInfo, x, y, pie_num);
    }
}

    /**
 * Lays out the pies in identically sized boxes within the given area.
 * @param num_pies number of pies
 * @param plotRect area to contain pies
 * @return integer array containing the number of lines followed by the number
 * of pies per line
 */
private int[] pieLayout(int num_pies, Rectangle plotRect)
{
    int num_pies_line, num_lines;

    // make a good first guess for the number of pies per line.
    if (plotRect.height == 0) {
        num_pies_line = num_pies;
    }
    else {
        num_pies_line = (int)
            Math.min((double)num_pies, Math.sqrt((double)num_pies) *
                                       plotRect.width / plotRect.height);
    }
    if (num_pies_line < 1) {
        num_pies_line = 1;
    }

    // Calc size of pie square with current number of pies per line.
    double v = Math.ceil((double)num_pies / num_pies_line);
    int pie_size = (int)Math.min(plotRect.width / num_pies_line,
                                 plotRect.height / v);

    // Now try neighbouring values for the number of pies per line to see
    // if we can do better (better == a bigger pie square).
    boolean done = false; // true if done loop
    boolean dir = true; // true if trying increasing values,
    // false if trying decreasing values.
    boolean upBound = false; // committed to increasing values.
    boolean downBound = false; // committed to decreasing values.

    // Here's how the loop works.  Try the next neighbouring value in an
    // upwards direction.  If it results in a bigger pie square, then commit
    // to looking in this direction for even bigger pies.  Otherwise, try
    // the next neighbouring value in a downwards direction.  If that value
    // results in a bigger pie square, commit to that direction.  Otherwise,
    // this is the value we want.
    while (!done) {
        int tmp_pies_line = num_pies_line;
        // Get next value for pies per line
        if (dir) {
            tmp_pies_line++;
        }
        else {
            tmp_pies_line--;
        }

        // check if we've hit boundary conditions
        if (tmp_pies_line > num_pies) {
            if (upBound || downBound) {
                done = true;
            }
            else if (dir) {
                dir = false;
            }
            else {
                done = true;
            }
            continue;
        }
        if (tmp_pies_line == 0) {
            done = true;
            continue;
        }

        // Calc the pie square size given this new value
        v = Math.ceil((double)num_pies / tmp_pies_line);
        int tmp_pie_size = (int)Math.min(plotRect.width / tmp_pies_line,
                                         plotRect.height / v);

        if (tmp_pie_size > pie_size) {
            // If this new pies per line value gives a bigger pie square,
            // mark this value and commit to going in this direction to
            // see if we can find even bigger squares.
            num_pies_line = tmp_pies_line;
            pie_size = tmp_pie_size;
            if (dir) {
                upBound = true;
            }
            else {
                downBound = true;
            }
        }
        else {
            // the new pies per line value results in a smaller pie square,
            // so ignore it.  If we'd already committed to a direction,
            // we are finished.  If not, and we haven't tried the other
            // direction yet, try it.  Otherwise, we are done.
            if (upBound || downBound) {
                done = true;
            }
            else if (dir) {
                dir = false;
            }
            else {
                done = true;
            }
        }
    }

    num_lines = num_pies / num_pies_line;
    if (num_lines != (double)(num_pies) / num_pies_line) {
        num_lines++;
    }
    int[] passback = new int[2];

    passback[0] = num_lines;
    passback[1] = num_pies_line;

    return(passback);
}

/**
 * Searches the explodeList array of the JCPieChartFormat object
 * for Points with x equal to JCPieChartFormat.ALL.
 * @return An ArrayList containing all Points from explodeList that
 * have x equal to JCPieChartFormat.ALL.
 */
private ArrayList getExplodeAllPoints()
{
    ArrayList allPoints = null;
    Point[] exList = pcf.getExplodeList();
    if (exList != null) {
        // First, check for the occurence of ALL
        // in the x values of the Point[] array.
        allPoints = new ArrayList();
        for (int i = 0; i < exList.length; i++) {
            if (exList[i].x == JCPieChartFormat.ALL) {
                allPoints.add(exList[i]);
            }
        }
    }
    return(allPoints);
}

/**
 * Sorts the data in the correct order, if required (for ascending and
 * descending pie orders).  Transfers the data into percentages.
 * Then computes the other slice based on the threshold value and
 * the threshold method (cutoff slice or percentile).  Stores the
 * resulting information in an <code>PieData</code> structure.
 * @param seriesList list of data series
 * @param pie_num number of current pie
 * @param num_slices number of slices in the current pie
 * @param pie_data structure to hold derived pie data
 * @param explodeAllPoints list of Points specifying the slices to be
 * drawn exploded on all pies.
 * @return 1 if the pie has an "other" slice; 0 if not
 */
private int computeOtherSlice(List seriesList, int pie_num, int num_slices,
                              PieData pie_data, List explodeAllPoints)
{
    int min_pie_slices = Math.min(num_slices, pcf.getMinSlices());
    double pie_threshold_value = pcf.getThresholdValue() / 100.0;

    /*
     *  Copy the required data into the sort array.  Check for zero pies!
     *  Also set the index = INDEX_HOLE for holes and non-displayed sets:
     *          --> this index is later stored in the order array
     *          --> an order = INDEX_HOLE means that the slice cannot be
     *              picked or unpicked, which is what we want for holes.
     */
    double holeValue = dataObject.getHoleValue();
    double ysum = 0.0;
    for (int i = 0; i < num_slices; i++) {
        ChartDataViewSeries thisSeries = (ChartDataViewSeries)seriesList.get(i);
        double x = thisSeries.getX(pie_num);
        double y = thisSeries.getY(pie_num);
        pie_data.sortPoints[i].index = i;
        if (y == holeValue || x == holeValue || !thisSeries.isVisible()) {
            y = -1.0;
            pie_data.sortPoints[i].index = INDEX_HOLE;
        }
        else if (y < 0.0) {
            y = 0.0;
        }
        else {
            ysum += y;
        }
        pie_data.sortPoints[i].y = y;
    }
    pie_data.zero_pie = (ysum == 0);

    // Mark slices to draw exploded
    Point[] exList = pcf.getExplodeList();
    if (exList != null) {
        // For point==pie_num, search for series==ALL
        boolean fAllSeries = false;
        for (int i = 0; i < exList.length; i++) {
            if (exList[i] != null &&
                (exList[i].x == pie_num ||
                 exList[i].x == JCPieChartFormat.ALL) &&
                exList[i].y == JCPieChartFormat.ALL)
            {
                fAllSeries = true;
                break;
            }
        }

        if (fAllSeries) {
            // Explode all slices (including "Other") for this pie:
            for (int i = 0; i < pie_data.exploded.length; i++) {
                pie_data.exploded[i] = true;
            }
        }
        else {
            for (int i = 0; i < exList.length; i++) {
                if (pie_num == exList[i].x) {
                    if (exList[i].y == JCPieChartFormat.OTHER_SLICE) {
                        int index = pie_data.exploded.length - 1;
                        pie_data.exploded[index] = true;
                    }
                    else if (exList[i].y < num_slices) {
                        pie_data.exploded[exList[i].y] = true;
                    }
                }
            }
            // If there are any Points in pcf.explodeList that have x equal to
            // JCPieChartFormat.ALL, mark as exploded the slices corresponding to
            // the associated y values.
            if (explodeAllPoints != null) {
                Point pt = null;
                int allPointsLength = explodeAllPoints.size();
                for (int i = 0; i < allPointsLength; i++) {
                    pt = (Point)explodeAllPoints.get(i);
                    if (pt != null) {
                        if (pt.y == JCPieChartFormat.OTHER_SLICE) {
                            int index = pie_data.exploded.length - 1;
                            pie_data.exploded[index] = true;
                        }
                        else if (pt.y >= 0 && pt.y < num_slices) {
                            pie_data.exploded[pt.y] = true;
                        }
                    }
                }
            }
        }
    }

    /* Do the sorting */
    switch (pcf.sortOrder) {
    case JCPieChartFormat.ASCENDING_ORDER:
        if (pie_data.sortPoints != null)
            Collections.sort(Arrays.asList(pie_data.sortPoints),
                             new PieSortAscending());
        break;
    case JCPieChartFormat.DESCENDING_ORDER:
        if (pie_data.sortPoints != null)
            Collections.sort(Arrays.asList(pie_data.sortPoints),
                             new PieSortDescending());
        break;
    default :
        break;
    }

    /* Convert the data into percentages. */
    for (int i = 0; i < num_slices; i++) {
        // calc pct if not a hole value
        if (pie_data.sortPoints[i].y >= 0.0) {
            pie_data.percents[i].pct =
                pie_data.zero_pie ? 0.0 : (pie_data.sortPoints[i].y / ysum);
        }
        else {
            pie_data.percents[i].pct = 0.0;
        }
        pie_data.percents[i].index = pie_data.sortPoints[i].index;
    }

    // Don't display the other slice if the minimum number of slices is equal
    // to the number of slices.
    if (num_slices == min_pie_slices) {
        pie_data.num_slices_pie = num_slices;
        pie_data.other_flag = 0;
        return(pie_data.other_flag);
    }

    /* Calc the other slice based on threshold method/value & sort order */
    pie_data.other_slice = -1;
    boolean other_flag = false;
    double pct_sum = 0.0;
    int m = 0;
    int num_slices_pie = num_slices;
    boolean incl_other = true;
    switch (pcf.sortOrder) {
    case JCPieChartFormat.ASCENDING_ORDER:
        /*
         * Keep the size of the other slice in the last contributing pie
         * slice. (The other slice is the first slice because of ascending
         * order).
         */
        for (int i = 0; i < num_slices; i++) {
            if (num_slices - i < min_pie_slices) {
                break;
            }
            // Don't compare hole values, they go right in the other slice
            if (pie_data.sortPoints[i].y >= 0.0) {
                double pct = pie_data.percents[i].pct;
                switch (pcf.thresholdMethod) {
                case JCPieChartFormat.PIE_PERCENTILE:
                    incl_other = pct_sum + pct < pie_threshold_value;
                    break;
                case JCPieChartFormat.SLICE_CUTOFF:
                    incl_other = pct < pie_threshold_value;
                    break;
                default :
                    System.out.println(ERRMSG_BIG_ERROR);
                    break;
                }
            }
            else {
                incl_other = true;
            }
            if (incl_other) {
                pie_data.percents[i].pct += pct_sum;
                pct_sum = pie_data.percents[i].pct;
                other_flag = true;
                m = i;
            }
        }
        // Adjust the array of percentages.  Make other slice the last slice.
        if (other_flag) {
            // Only have other slice if value is > 0
            if (pie_data.percents[m].pct > 0.0) {
                num_slices_pie = num_slices - m;
                pie_data.other_slice = num_slices_pie - 1;
                pie_data.percents[num_slices].pct = pie_data.percents[m].pct;
                pie_data.percents[num_slices].index = num_slices;
            }
            else {
                num_slices_pie = num_slices - m - 1;
                other_flag = false;
            }
            // Index of other slice needed for explode checking
            for (int i = m + 1; i <= num_slices; i++) {
                pie_data.percents[i - m - 1].pct =
                    pie_data.percents[i].pct;
                pie_data.percents[i - m - 1].index =
                    pie_data.percents[i].index;
            }

        }
        break;

    case JCPieChartFormat.DESCENDING_ORDER:
        /*
         * Keep the size of the other slice in the last contributing pie
         * slice.  (The other slice is the last slice because of descending
         * order).
         */
        m = num_slices - 1;
        for (int i = num_slices - 1; i >= 0; i--) {
            if (i + 1 < min_pie_slices) {
                break;
            }
            // Don't compare hole values, they go right in the other slice
            if (pie_data.sortPoints[i].y >= 0.0) {
                double pct = pie_data.percents[i].pct;
                switch (pcf.thresholdMethod) {
                case JCPieChartFormat.PIE_PERCENTILE:
                    incl_other = pct_sum + pct < pie_threshold_value;
                    break;
                case JCPieChartFormat.SLICE_CUTOFF:
                    incl_other = pct < pie_threshold_value;
                    break;
                default :
                    System.out.println(ERRMSG_BIG_ERROR);
                    break;
                }
            }
            else {
                incl_other = true;
            }
            if (incl_other) {
                pie_data.percents[i].pct += pct_sum;
                pct_sum = pie_data.percents[i].pct;
                m = i;
                other_flag = true;
            }
            else {
                break;
            }
        }
        // index of other slice needed for explode checking
        // only have other slice if value > 0
        if (other_flag) {
            if (pie_data.percents[m].pct > 0.0) {
                pie_data.percents[m].index = num_slices;
                num_slices_pie = m + 1;
                pie_data.other_slice = m;
            }
            else {
                num_slices_pie = m;
                other_flag = false;
            }
        }
        break;

    case JCPieChartFormat.DATA_ORDER:
        /*
         * Have to sort the points anyway to handle thresholding.  (Use
         * ascending order).  Note that the sort and percentage arrays will
         * now have different orders.
         */
        if (pie_data.sortPoints != null)
            Collections.sort(Arrays.asList(pie_data.sortPoints),
                             new PieSortAscending());
        pie_data.percents[num_slices].pct = 0.0;
        // index of other slice needed for explode checking
        pie_data.percents[num_slices].index = num_slices;
        for (int i = 0; i < num_slices; i++) {
            if (num_slices_pie < min_pie_slices) {
                break;
            }
            double tmp_pct = 0.0;
            // Don't compare hole values, they go right in the other slice
            if (pie_data.sortPoints[i].y >= 0.0) {
                tmp_pct = pie_data.zero_pie ? 0.0
                                            : (pie_data.sortPoints[i].y / ysum);
                switch (pcf.thresholdMethod) {
                case JCPieChartFormat.PIE_PERCENTILE:
                    incl_other = pct_sum + tmp_pct < pie_threshold_value;
                    break;
                case JCPieChartFormat.SLICE_CUTOFF:
                    incl_other = tmp_pct < pie_threshold_value;
                    break;
                default :
                    System.out.println(ERRMSG_BIG_ERROR);
                    break;
                }
            }
            else {
                incl_other = true;
            }
            if (incl_other) {
                pie_data.percents[num_slices_pie].pct += tmp_pct;
                pct_sum = pie_data.percents[num_slices_pie].pct;

                /* Search for matching percent in percent array */
                m = 0;
                while (m < num_slices_pie) {
                    if (pie_data.percents[m].pct == tmp_pct) {
                        m++;
                        break;
                    }
                    m++;
                }

                /* Remove found entry from percent array & readjust array */
                while (m <= num_slices_pie) {
                    pie_data.percents[m - 1].pct =
                        pie_data.percents[m].pct;
                    pie_data.percents[m - 1].index =
                        pie_data.percents[m].index;
                    m++;
                }

                // Only have other slice if value > 0
                if (tmp_pct > 0.0) {
                    other_flag = true;
                }
                num_slices_pie--;
            }
        }
        if (other_flag) {
            pie_data.other_slice = num_slices_pie;
            num_slices_pie++;
        }
        break;
    default :
        System.out.println(ERRMSG_BIG_ERROR);
        break;
    }
    pie_data.num_slices_pie = num_slices_pie;
    pie_data.other_flag = other_flag ? 1 : 0;

    return(pie_data.other_flag);
}

/**
 * This method is used to specify the type of operation to be
 * performed on shapes when the <code>draw()</code> method is invoked.  Shapes
 * may be rendered, picked, or image mapped.  The default operation
 * is <code>ProcessShapeModel.SHAPE_RENDER</code>.
 * @param operation possible values are
 * <code>ProcessShapeModel.SHAPE_RENDER</code>,
 * <code>ProcessShapeModel.SHAPE_PICK</code>, or
 * <code>ProcessShapeModel.SHAPE_IMAGEMAP</code>
 */
public void setShapeOperation(int operation)
{
    shapeOp = operation;
}

/**
 * Resets the shape operation to <code>ProcessShapeModel.SHAPE_RENDER</code>.
 */
public void resetShapeOperation()
{
    shapeOp = ProcessShapeModel.SHAPE_RENDER;
}

private void tagData(PieTagModel tagObj, int slice, int point)
{
    int seriesIndex = -1;
    ChartDataViewSeries series = null;
    PieData pieData = tagObj.getPieData();
    if (pieData.other_flag != 0 && slice == pieData.other_slice) {
        // series = null; For the 'other' slice.
        seriesIndex = JCPieChartFormat.OTHER_SLICE;
    }
    else {
        series = (ChartDataViewSeries)seriesList.get(slice);
        seriesIndex = pieData.percents[slice].index;
    }
    tagObj.setDataIndex(new JCDataIndex(dataObject, series, seriesIndex,
                                        point));
}

private void initShapeOperation(int sliceIndex, int pie_num,
                                JCFillStyle fstyle)
{
    if (shapeOp == ProcessShapeModel.SHAPE_PICK) {
        tagData(picker, sliceIndex, pie_num);
    }
    else if (shapeOp == ProcessShapeModel.SHAPE_IMAGEMAP) {
        tagData(mapper, sliceIndex, pie_num);
    }
    else if (shapeOp == ProcessShapeModel.SHAPE_RENDER) {
        ((RenderShape)processShape).setFillStyle(fstyle);
    }
}

/**
 * Draws the pie slices (front or back plane) centered at (<i>x</i>,<i>y</i>).
 * @param gc Graphics context used for drawing
 * @param seriesList list of data series
 * @param pie_data derived data for this pie
 * @param pPie transient data for this data view
 * @param x <i>x</i> position of pie
 * @param y <i>y</i> position of pie
 * @param pie_num number of pie being drawn
 * @param is_top <code>true</code> for drawing on front plane;
 * <code>false</code> for drawing on back plane
 * @param fillOrientRect
 */
private void drawPiePlane(Graphics gc, List seriesList,
                          PieData pie_data, PieChartInfo pPie, int x, int y,
                          int pie_num, boolean is_top, Rectangle fillOrientRect)
{
    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape)gc;
    }

    double start_angle = pcf.getStartAngle();
    double arc_sum = 0.0;

    for (int slice = 0; slice < pie_data.num_slices_pie; slice++) {
        JCFillStyle fstyle = getPieSliceFillStyle(pie_data, slice);

        initShapeOperation(slice, pie_num, fstyle);

        double arc_angle = pie_data.percents[slice].pct * 360;
        // Make sure the last slice draws up to the first slice
        if (slice == pie_data.lastDrawnSlice) {
            if (arc_sum >= 360.0)
                arc_angle = 0.0;
            else {
                arc_angle = start_angle - pcf.getStartAngle();
                if (arc_angle < 0.0 || (arc_angle == 0.0 && arc_sum == 0.0))
                    arc_angle += 360.0;
            }
        }

        if (arc_angle != 0.0) {
            if (pie_data.exploded[pie_data.percents[slice].index]) {
                double bisected_angle = arc_angle / 2.0;
                calcExplodeOffset(start_angle,
                                  bisected_angle,
                                  explode_offset,
                                  pPie.pos[pie_num].offset[slice]);
            }
            Point offset = pPie.pos[pie_num].offset[slice];
            processShape.updateFillOrientation(fillOrientRect,
                                               JCFillStyle.CENTER);

            // If using DefineShape, begin a shape definition for this index
            if (defShape != null) {
                JCDataIndex useIndex = getDataIndex(slice, pie_data, pie_num,
                                                    seriesList);
                defShape.beginShape(useIndex);
            }
            if (is_top) {
                processShape.fillArc(x + offset.x, y + offset.y,
                                     pPie.width, pPie.height,
                                     start_angle, -arc_angle);
            }
            else {
                processShape.fillArcShadow(x + offset.x, y + offset.y,
                                           pPie.width, pPie.height,
                                           start_angle, -arc_angle);
            }
            // If using DefineShape, finish up the shape definition
            if (defShape != null) {
                defShape.endShape(true);
            }
            processShape.resetFillOrientation();

            start_angle -= arc_angle;
            arc_sum += arc_angle;
            if (start_angle < 0.0) {
                start_angle += 360.0;
            }
        }
    }
}

private JCDataIndex getDataIndex(int slice, PieData pie_data, int pie_num,
                                 List seriesList)
{
    JCDataIndex dataIndex = null;
    if (slice == pie_data.other_slice) {
        dataIndex = new JCDataIndex(dataObject, null,
                                    JCPieChartFormat.OTHER_SLICE, pie_num);
    }
    else {
        int index = pie_data.percents[slice].index;
        ChartDataViewSeries s = (ChartDataViewSeries)seriesList.get(index);
        dataIndex = new JCDataIndex(pie_num, s);
    }
    return(dataIndex);
}

private boolean isSliceVisible()
{
    return(true);
}

private Point2D.Double createPoint(double xCenter, double yCenter, double angle,
                                   double xRadius, double yRadius)
{
    return new Point2D.Double(xCenter + Math.cos(angle) * xRadius,
                              yCenter - Math.sin(angle) * yRadius);
}

private Point computeExplodeOffset(PieChartInfo pPie, int pieNumber,
                                   int slice, double angle, double arcAngle)
{
    double bisectedAngle = arcAngle / 2.0;
    calcExplodeOffset(angle,
                      bisectedAngle,
                      explode_offset,
                      pPie.pos[pieNumber].offset[slice]);
    return((Point)pPie.pos[pieNumber].offset[slice].clone());
}

// Draws a filled parallelogram along a radial edge of a slice.
private void drawSidePanel(Graphics gc,
                           JCFillStyle fstyle,
                           Rectangle fillOrientRect,
                           int sliceIndex,
                           PieData pie_data,
                           PieChartInfo pPie,
                           int x, int y,
                           int pie_num,
                           double angle)
{

    int pie_width = pPie.width;
    int pie_height = pPie.height;
    double x_radius = pie_width / 2.0;
    double y_radius = pie_height / 2.0;
    double x_center = x + x_radius;
    double y_center = y + y_radius;

    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape)gc;
    }

    initShapeOperation(sliceIndex, pie_num, fstyle);

    Point2D.Double points[] = new Point2D.Double[4];
    //  Define the vertical line for the center:
    points[0] = new Point2D.Double(x_center, y_center);
    points[1] = new Point2D.Double(points[0].x, points[0].y + delta_y);

    //  Define the vertical line on the edge of the pie:
    angle = JCChartUtil.convertAngle(JCChartUtil.DEGREES, JCChartUtil.RADIANS,
                                     angle);
    points[3] = createPoint(x_center, y_center, angle, x_radius, y_radius);
    points[2] = new Point2D.Double(points[3].x, points[3].y + delta_y);

    Shape shape = null;
    if (shapeOp == ProcessShapeModel.SHAPE_IMAGEMAP) {
        // For image maps, we can't use GeneralPath.
        Polygon poly = new Polygon();
        for (int j = 0; j < points.length; j++) {
            poly.addPoint((int)Math.round(points[j].x),
                          (int)Math.round(points[j].y));
        }
        shape = poly;
    }
    else {   //  shapeOp is SHAPE_RENDER or SHAPE_PICK
        GeneralPath genPath = new GeneralPath();
        genPath.moveTo((float)points[0].x, (float)points[0].y);
        for (int j = 1; j < points.length; j++) {
            genPath.lineTo((float)points[j].x, (float)points[j].y);
        }
        genPath.closePath();
        shape = genPath;
    }

    processShape.updateFillOrientation(fillOrientRect, JCFillStyle.CENTER);

    //  If using DefineShape, begin a shape definition for this index.
    if (defShape != null) {
        JCDataIndex useIndex = getDataIndex(sliceIndex, pie_data, pie_num,
                                            seriesList);
        defShape.beginShape(useIndex);
    }

    processShape.fillPolygonShadow(shape);

    //  If using DefineShape, finish up the shape definition.
    if (defShape != null) {
        defShape.endShape(true);
    }

    processShape.resetFillOrientation();

    //  Draw a vertical line at the center:
    processShape.drawLine(points[0].x, points[0].y, points[1].x, points[1].y);
    //  Draw a vertical line at the edge of the pie:
    processShape.drawLine(points[2].x, points[2].y, points[3].x, points[3].y);
    //  Draw a line along the bottom of the panel:
    processShape.drawLine(points[1].x, points[1].y, points[2].x, points[2].y);
}

private Point2D.Double[] drawFilledParallelogram(Graphics gc,
                                                 JCFillStyle fstyle,
                                                 Rectangle fillOrientRect,
                                                 int sliceIndex,
                                                 PieData pie_data,
                                                 PieChartInfo pPie,
                                                 int x, int y,
                                                 int pie_num,
                                                 double start_angle,
                                                 double end_angle)
{
    Point2D.Double points[] = computePanelVertices(pPie, x, y, start_angle,
                                                   end_angle);
    initShapeOperation(sliceIndex, pie_num, fstyle);
    drawFilledParallelogram(gc, pie_data, sliceIndex, pie_num,
                            points, fillOrientRect);
    return(points);
}

private Point2D.Double[] computePanelVertices(PieChartInfo pPie,
                                              int x, int y,
                                              double start_angle,
                                              double end_angle)
{
    int pie_width = pPie.width;
    int pie_height = pPie.height;
    double x_radius = pie_width / 2.0;
    double y_radius = pie_height / 2.0;
    double x_center = x + x_radius;
    double y_center = y + y_radius;

    Point2D.Double points[] = new Point2D.Double[4];
    //  Define the vertical line that pertains to the start angle:
    double angle = JCChartUtil.convertAngle(JCChartUtil.DEGREES,
                                            JCChartUtil.RADIANS, start_angle);
    points[0] = createPoint(x_center, y_center, angle, x_radius, y_radius);
    points[1] = new Point2D.Double(points[0].x, points[0].y + delta_y);

    //  Define the vertical line that pertains to the end angle:
    angle = JCChartUtil.convertAngle(JCChartUtil.DEGREES,
                                     JCChartUtil.RADIANS, end_angle);
    points[3] = createPoint(x_center, y_center, angle, x_radius, y_radius);
    points[2] = new Point2D.Double(points[3].x, points[3].y + delta_y);

    return(points);
}

private void drawFilledParallelogram(Graphics gc,
                                     PieData pie_data,
                                     int sliceIndex,
                                     int pie_num,
                                     Point2D.Double[] points,
                                     Rectangle fillOrientRect)
{
    DefineShape defShape = null;
    if (gc instanceof DefineShape) {
        defShape = (DefineShape)gc;
    }

    //  Draw the current edge section by filling in the defined polygon
    //  and drawing the vertical line on both sides of the polygon.
    //  Note that the curved chord below the polygon is not filled here.
    //  That piece of the front panel is filled by the back plane.
    Shape shape = null;
    if (shapeOp == ProcessShapeModel.SHAPE_IMAGEMAP) {
        // For image maps, we can't use GeneralPath.
        Polygon poly = new Polygon();
        for (int j = 0; j < points.length; j++) {
            poly.addPoint((int)Math.round(points[j].x),
                          (int)Math.round(points[j].y));
        }
        shape = poly;
    }
    else {   //  shapeOp is SHAPE_RENDER or SHAPE_PICK
        GeneralPath genPath = new GeneralPath();
        genPath.moveTo((float)points[0].x, (float)points[0].y);
        for (int j = 1; j < points.length; j++) {
            genPath.lineTo((float)points[j].x, (float)points[j].y);
        }
        genPath.closePath();
        shape = genPath;
    }

    processShape.updateFillOrientation(fillOrientRect, JCFillStyle.CENTER);

    // If using DefineShape, begin a shape definition for this index
    if (defShape != null) {
        JCDataIndex useIndex = getDataIndex(sliceIndex, pie_data, pie_num,
                                            seriesList);
        defShape.beginShape(useIndex);
    }

    processShape.fillPolygonShadow(shape);

    //  If using DefineShape, finish up the shape definition.
    if (defShape != null) {
        defShape.endShape(true);
    }

    processShape.resetFillOrientation();
}

// Draws a filled parallelogram along the chord between start_angle and
// end_angle.
private void drawFrontPanel(Graphics gc,
                            JCFillStyle fstyle,
                            Rectangle fillOrientRect,
                            int sliceIndex,
                            PieData pie_data,
                            PieChartInfo pPie,
                            int x, int y,
                            int pie_num,
                            double start_angle, double end_angle)
{

    Point2D.Double points[] =
        drawFilledParallelogram(gc, fstyle, fillOrientRect,
                                sliceIndex, pie_data, pPie,
                                x, y, pie_num, start_angle,
                                end_angle);
    //  Draw the vertical lines at the start and end angles:
    processShape.drawLine(points[0].x, points[0].y, points[1].x, points[1].y);
    processShape.drawLine(points[2].x, points[2].y, points[3].x, points[3].y);

    //  Draw an arc along the bottom of the polygon.  The drawArc() method
    //  draws spurious lines when both start_angle and end_angle have
    //  extremely small values (JDK 1.4.1).  To avoid this problem, call
    //  precCorrect() to round-off the value of arcAngle.
    double arcAngle = JCChartUtil.precCorrect(8, end_angle - start_angle);
    processShape.drawArc(x, y + delta_y, pPie.width, pPie.height,
                         start_angle, arcAngle);
}

/**
 * Draws the curve around the outside of the pie slices (front plane)
 * centered at (<i>x</i>,<i>y</i>).
 * @param pie_data derived data for this pie
 * @param pPie transient data for this data view
 * @param x <i>x</i> position of pie
 * @param y <i>y</i> position of pie
 * @param pie_num number of pie being drawn
 */
private void drawPieCircle(PieData pie_data, PieChartInfo pPie, int x, int y,
                           int pie_num)
{
    double start_angle = pcf.getStartAngle();
    double arc_sum = 0.0;

    // Accumulate unexploded slices in an undrawn arc.  Draw this arc
    // when an exploded slice is found or when done looking through slices.
    double undrawn_angle = 0.0;

    for (int slice = 0; slice < pie_data.num_slices_pie; slice++) {
        double arc_angle = pie_data.percents[slice].pct * 360;
        // Make sure the last slice draws up to the first slice
        if (slice == pie_data.lastDrawnSlice) {
            // if last slice not exploded, undrawn angle will
            // already be part of below
            if (!pie_data.exploded[pie_data.percents[slice].index])
                undrawn_angle = 0.0;

            if (arc_sum >= 360.0) {
                arc_angle = 0.0;
            }
            else {
                arc_angle = start_angle - undrawn_angle - pcf.getStartAngle();
                if (arc_angle < 0.0 || (arc_angle == 0.0 && arc_sum == 0.0)) {
                    arc_angle += 360.0;
                }
            }
        }
        if (arc_angle != 0.0) {
            if (pie_data.exploded[pie_data.percents[slice].index]) {
                // First draw undrawn arc
                if (undrawn_angle != 0.0) {
                    processShape.drawArc(x, y, pPie.width, pPie.height,
                                         start_angle, -undrawn_angle);
                    start_angle -= undrawn_angle;
                    arc_sum += undrawn_angle;
                    if (start_angle < 0.0) {
                        start_angle += 360.0;
                    }
                    undrawn_angle = 0.0;
                }
                // Now draw exploded arc
                Point offset = pPie.pos[pie_num].offset[slice];
                processShape.drawArc(x + offset.x, y + offset.y,
                                     pPie.width, pPie.height,
                                     start_angle, -arc_angle);
                start_angle -= arc_angle;
                arc_sum += arc_angle;
                if (start_angle < 0.0) {
                    start_angle += 360.0;
                }
            }
            else {
                undrawn_angle += arc_angle;
            }
        }
    }

    // Draw remaining arc
    if (undrawn_angle != 0) {
        processShape.drawArc(x, y, pPie.width, pPie.height,
                             start_angle, -undrawn_angle);
    }
}

/**
 * Draws the edge of the pie disc (this will be drawn over the back plane).
 * Only draws an edge section for those slices which are on the bottom of
 * the disc (they are visible).  This means that at least one of the
 * start or end angles of a section must be negative for it to be drawn.
 * @param gc graphics context used for drawing
 * @param pie_data derived data for this pie
 * @param pPie transient data for this data view
 * @param x <i>x</i> position of pie
 * @param y <i>y</i> position of pie
 * @param pie_num
 * @param fillOrientRect
 */
private void drawPieEdge(Graphics gc, PieData pie_data, PieChartInfo pPie,
                         int x, int y, int pie_num, Rectangle fillOrientRect)
{
    double end_angle = Double.MAX_VALUE;
    double arc_angle = Double.MAX_VALUE;
    double front_start_angle;
    int slice;
    boolean fVisible;
    int straddlingSlice = Integer.MAX_VALUE;

    //  Calculate the start/end angle for each slice.
    double start_angle = pcf.getStartAngle();
    start_angle = JCChartUtil.boundAngle(JCChartUtil.DEGREES, start_angle);
    if (start_angle > 90.0) {
        start_angle -= 360.0;
    }

    //  Create and populate the PieSortPoint array.
    pie_data.sortPoints = new PieSortPoint[pie_data.num_slices_pie + 1];
    for (slice = 0; slice < pie_data.num_slices_pie; slice++) {
        arc_angle = pie_data.percents[slice].pct * 360;
        end_angle = start_angle - arc_angle;
        if (end_angle <= -270.0) {
            end_angle += 360.0;
        }
        pie_data.sortPoints[slice] = new PieSortPoint(start_angle, slice);

        //  If both side panels of a slice are visible, draw
        //  that slice first.

        if ((start_angle > 90.0 && start_angle < 270.0) ||
            (start_angle > -270.0 && start_angle < -90.0)) {

            if ((end_angle < 90.0 && end_angle > -90.0) ||
                (end_angle > 270.0)) {
                //  This slice straddles the 90 degree line and its
                //  side panels must be drawn first.
                straddlingSlice = slice;
            }
        }


        start_angle = end_angle;
    }
    pie_data.sortPoints[slice] = new PieSortPoint(start_angle, slice);

    //  If a slice straddles the 90 degree line, draw its side panels first.
    if (straddlingSlice != Integer.MAX_VALUE) {
        end_angle = pie_data.sortPoints[straddlingSlice].y;
        arc_angle = pie_data.percents[straddlingSlice].pct * 360;

        fVisible = isSliceVisible();

        if (fVisible && arc_angle != 0.0) {
            JCFillStyle fstyle = getPieSliceFillStyle(pie_data,
                                                      straddlingSlice);
            int xCoord = x;
            int yCoord = y;

            if (pie_data.exploded[pie_data.percents[straddlingSlice].index]) {
                Point offset = computeExplodeOffset(pPie, pie_num,
                                                    straddlingSlice,
                                                    end_angle - arc_angle,
                                                    -arc_angle);
                xCoord += offset.x;
                yCoord += offset.y;
            }
            drawSidePanel(gc, fstyle, fillOrientRect, straddlingSlice,
                          pie_data, pPie, xCoord, yCoord, pie_num,
                          end_angle - arc_angle);
            xCoord = x;
            yCoord = y;
            if (pie_data.exploded[pie_data.percents[straddlingSlice].index]) {
                Point offset = computeExplodeOffset(pPie, pie_num,
                                                    straddlingSlice,
                                                    end_angle,
                                                    arc_angle);
                xCoord += offset.x;
                yCoord += offset.y;
            }
            drawSidePanel(gc, fstyle, fillOrientRect, straddlingSlice,
                          pie_data, pPie, xCoord, yCoord, pie_num,
                          end_angle);
        }
    }

    //  Sort the end angles in descending order.
    //  First, replace the 0th element with Double.MAX_VALUE to
    //  ensure that it remains the first point in the array.
    PieSortPoint savedPoint = pie_data.sortPoints[0];
    pie_data.sortPoints[0] = new PieSortPoint(Double.MAX_VALUE, 0);

    Collections.sort(Arrays.asList(pie_data.sortPoints),
                     new PieSortDescending());
    //  Now restore the value of the 0th element:
    pie_data.sortPoints[0] = savedPoint;

    //  Draw the first set of front facing inner edges in z-order (i.e., the
    //  end angle edges between (-90, 90).  Draw the corresponding front edges.
    for (int i = 0; i < pie_data.num_slices_pie;) {
        i++;
        slice = pie_data.sortPoints[i].index - 1;
        end_angle = pie_data.sortPoints[i].y;
        if (end_angle < -90.0) {
            break;
        }

        arc_angle = pie_data.percents[slice].pct * 360;

        if (arc_angle == 0.0) {
            continue;
        }

        if (arc_angle < 0.0) {
            arc_angle += 360.0;
        }

        fVisible = isSliceVisible();

        if (fVisible) {
            JCFillStyle fstyle = getPieSliceFillStyle(pie_data, slice);

            int xCoord = x;
            int yCoord = y;
            if (pie_data.exploded[pie_data.percents[slice].index]) {
                Point offset = computeExplodeOffset(pPie, pie_num, slice,
                                                    end_angle + arc_angle,
                                                    arc_angle);
                xCoord += offset.x;
                yCoord += offset.y;
            }

            if (end_angle < 0.0) {
                //  The front panel is visible.
                start_angle = end_angle + arc_angle;
                if (start_angle > 0.0) {
                    //  The front panel is only partially visible.
                    start_angle = 0.0;
                }

                drawFrontPanel(gc, fstyle, fillOrientRect, slice, pie_data,
                               pPie, xCoord, yCoord, pie_num, start_angle,
                               end_angle);
            }

            if (slice != straddlingSlice) {
                drawSidePanel(gc, fstyle, fillOrientRect, slice, pie_data,
                              pPie, xCoord, yCoord, pie_num, end_angle);
            }
        }
    }

    //  Sort the start angles in ascending order.
    Collections.sort(Arrays.asList(pie_data.sortPoints),
                     new PieSortAscending());

    //  Draw the second set of front facing inner edges in z-order (i.e., the
    //  start angle edges between (-270, -90).  Draw the corresponding front
    //  edges.
    int i = 0;
    for (; i < (pie_data.num_slices_pie + 1); i++) {
        slice = pie_data.sortPoints[i].index;
        if (slice == pie_data.num_slices_pie) {
            //  Ignore the one that's only an end angle.
            continue;
        }

        start_angle = pie_data.sortPoints[i].y;
        if (start_angle > -90.0) {
            break;
        }

        arc_angle = pie_data.percents[slice].pct * 360;
        if (arc_angle == 0.0) {
            continue;
        }

        if (arc_angle < 0.0) {
            arc_angle += 360.0;
        }

        fVisible = isSliceVisible();

        if (fVisible) {
            JCFillStyle fstyle = getPieSliceFillStyle(pie_data, slice);

            int xCoord = x;
            int yCoord = y;
            if (pie_data.exploded[pie_data.percents[slice].index]) {
                Point offset = computeExplodeOffset(pPie, pie_num, slice,
                                                    start_angle, arc_angle);
                xCoord += offset.x;
                yCoord += offset.y;
            }

            if (start_angle > -180.0) {
                //  The front panel is visible.
                end_angle = start_angle - arc_angle;
                if (end_angle < -180.0) {
                    //  The front panel is only partially visible.
                    end_angle = -180.0;
                }

                drawFrontPanel(gc, fstyle, fillOrientRect, slice, pie_data,
                               pPie, xCoord, yCoord, pie_num, start_angle,
                               end_angle);
            }

            if (slice != straddlingSlice) {
                drawSidePanel(gc, fstyle, fillOrientRect, slice, pie_data,
                              pPie, xCoord, yCoord, pie_num, start_angle);
            }
        }
    }

    //  Draw the front panel that straddles the -90 line.
    if (pie_data.num_slices_pie > 0) {
        if (i < (pie_data.num_slices_pie + 1)) {
            //  Reached a start angle > -90.
            //  Note that slice and start_angle are already initialized
            //  from the preceeding loop.
            arc_angle = pie_data.percents[slice].pct * 360;
            end_angle = start_angle - arc_angle;
            front_start_angle = (start_angle > 0.0) ? 0.0 : start_angle;
        }
        else {
            //  All the start angles are between (-90, -270).
            //  Try to find a slice that straddles the -90 line.
            for (i = 0; i < (pie_data.num_slices_pie + 1); i++) {
                slice = pie_data.sortPoints[i].index;
                if (slice == pie_data.num_slices_pie) {
                    //  Ignore the one that's only an end angle.
                    continue;
                }

                start_angle = pie_data.sortPoints[i].y;
                arc_angle = pie_data.percents[slice].pct * 360;

                if (arc_angle == 0.0) {
                    continue;
                }
                end_angle = start_angle - arc_angle;
                if (end_angle <= -270.0) {
                    end_angle += 360.0;
                }

                if (arc_angle > 180) {
                    break;
                }
            }
            front_start_angle = 0.0;
        }

        if (i < (pie_data.num_slices_pie + 1)) {
            //  There is a slice which straddles the -90 line.

            fVisible = isSliceVisible();
            if (fVisible) {
                JCFillStyle fstyle = getPieSliceFillStyle(pie_data, slice);

                if (end_angle < -180.0) {
                    end_angle = -180.0;
                }

                int xCoord = x;
                int yCoord = y;
                if (pie_data.exploded[pie_data.percents[slice].index]) {
                    Point offset = computeExplodeOffset(pPie, pie_num, slice,
                                                        start_angle, arc_angle);
                    xCoord += offset.x;
                    yCoord += offset.y;
                }

                // If arc_angle is greater than or equal to 180 degrees, we
                // need to draw additional filled parallelograms (with no
                // outlines) to hide the slice edges when the start angle
                // is just below 0 degrees or when the end angle is just
                // above -180 degrees.
                if (arc_angle >= 180.0) {
                    if (front_start_angle <= 0 && front_start_angle > -90.0) {
                        // Draw a filled parallelogram from front_start_angle
                        // to -90.0
                        drawFilledParallelogram(gc, fstyle, fillOrientRect,
                                                slice, pie_data, pPie,
                                                xCoord, yCoord, pie_num,
                                                front_start_angle, -90.0);
                    }
                    if (end_angle < -90.0 && end_angle >= -180.0) {
                        // Draw a filled parallelogram from -90.0 to
                        // end_angle.
                        drawFilledParallelogram(gc, fstyle, fillOrientRect,
                                                slice, pie_data, pPie,
                                                xCoord, yCoord, pie_num,
                                                -90.0, end_angle);
                    }
                }

                drawFrontPanel(gc, fstyle, fillOrientRect, slice,
                               pie_data, pPie, xCoord, yCoord, pie_num,
                               front_start_angle, end_angle);
            }
        }
    }
}

/**
 * Draws the lines on the pie.  Also stores the centers and coords (coords
 * of the the slice/arc intersections) for use in picking/unpicking and
 * markers.  <p>Storing the coords array is tricky.  Since we allow unpicking
 * on zero slices, when we encounter a zero slice we must store the coords
 * of the NEXT non-zero slice (which could wrap around the array).  This is
 * done by updating previous zero slices (marked by
 * <code>COORD_ZERO_SLICE</code>) when a non-zero slice is encountered and
 * updating the end of the array to the first non-zero slice if there are zero
 * slices at the end of the array.
 * <p><B>Note:</b>  This routine assumes that it does not get a zero pie, but it
 *        should not crash if it does.  In that case, it should leave an
 *        array full of <code>COORD_ZERO_SLICE</code>'s.
 * @param pie_data derived data for this pie
 * @param pPie transient data for this data view
 * @param x <i>x</i> position of pie
 * @param y <i>y</i> position of pie
 * @param pie_num number of pie being drawn
 */

private void drawPieLines(PieData pie_data, PieChartInfo pPie, int x, int y,
                          int pie_num)
{
    /* Compute and store the center for use in picking/unpicking and markers */
    double x_radius = pPie.width / 2.0;
    double y_radius = pPie.height / 2.0;

    double x0 = 0;
    double y0 = 0;
    double x1 = x + x_radius;
    double y1 = y + y_radius;
    double x2;
    double y2;
    double x3 = 0;
    double y3 = 0;

    pPie.pos[pie_num].cx = (int)Math.round(x1);
    pPie.pos[pie_num].cy = (int)Math.round(y1);

    // Draw the lines (only for non-zero slices)
    int i, j;
    int lasti = -1;
    double start_angle = pcf.getStartAngle();
    double arc_sum = 0.0;
    int prevDrawnSlice = pie_data.lastDrawnSlice;
    boolean prevSliceExploded = false;
    boolean firstSliceExploded = false;

    for (i = 0; i < pie_data.num_slices_pie; i++) {
        x2 = y2 = COORD_ZERO_SLICE;
        int index = pie_data.percents[prevDrawnSlice].index;
        prevSliceExploded = (i == 0) ? false : pie_data.exploded[index];
        double arc_angle = pie_data.percents[i].pct * 360;
        if (i == pie_data.lastDrawnSlice) {
            if (arc_sum >= 360.0)
                arc_angle = 0.0;
            else {
                arc_angle = start_angle - pcf.getStartAngle();
                if (arc_angle < 0.0 || (arc_angle == 0.0 && arc_sum == 0.0)) {
                    arc_angle += 360.0;
                }
            }
        }
        if (arc_angle != 0.0) {
            // We have a non-zero slice
            double rad = 0.0;
            if (arc_sum == 0.0) {
                rad = JCChartUtil.degToRad(start_angle);
                x2 = x1 + Math.cos(rad) * x_radius;
                y2 = y1 - Math.sin(rad) * y_radius;

                x0 = x2;
                y0 = y2;
                firstSliceExploded =
                    pie_data.exploded[pie_data.percents[i].index];
            }
            else {
                x2 = x3;
                y2 = y3;
            }
            start_angle -= arc_angle;
            arc_sum += arc_angle;
            if (start_angle < 0.0) {
                start_angle += 360.0;
            }

            rad = JCChartUtil.degToRad(start_angle);
            x3 = x1 + Math.cos(rad) * x_radius;
            y3 = y1 - Math.sin(rad) * y_radius;

            // If this slice is exploded, draw the left exploded line
            if (pie_data.exploded[pie_data.percents[i].index]) {
                Point offset = pPie.pos[pie_num].offset[i];
                processShape.drawLine(x1 + offset.x, y1 + offset.y,
                                      x2 + offset.x, y2 + offset.y);
            }

            // If this slice is not exploded and previous slice is exploded,
            // draw the left non-exploded line
            else if (prevSliceExploded) {
                processShape.drawLine(x1, y1, x2, y2);
            }

            // Always draw the right side
            if (pie_data.exploded[pie_data.percents[i].index]) {
                Point offset = pPie.pos[pie_num].offset[i];
                processShape.drawLine(x1 + offset.x, y1 + offset.y,
                                      x3 + offset.x, y3 + offset.y);
            }
            else {
                processShape.drawLine(x1, y1, x3, y3);
            }

            prevDrawnSlice = i;
        }
        pPie.pos[pie_num].coord[i].x = (int)Math.round(x2);
        pPie.pos[pie_num].coord[i].y = (int)Math.round(y2);

        /* If the current slice is non-zero, fill in any coords corresponding
           to zero slices.  We now have the correct value for these slices. */
        if (x2 >= 0) {
            j = i - 1;
            while (j >= 0 && pPie.pos[pie_num].coord[j].x == COORD_ZERO_SLICE) {
                pPie.pos[pie_num].coord[j].x = (int)Math.round(x2);
                pPie.pos[pie_num].coord[j].y = (int)Math.round(y2);
                j--;
            }
            lasti = i;
        }
    }

    // If first slice is not exploded and the last draw slice is exploded,
    // draw the right line for the first slice.
    if (!firstSliceExploded &&
        pie_data.exploded[pie_data.percents[prevDrawnSlice].index])
    {
        processShape.drawLine(x1, y1, x0, y0);
    }

    /* If zero slices at end of array, update the (-1,-1)'s with the
       value at the beginning of array so as to complete the circle.
       --> lasti contains the index of the last non-zero slice.  */
    i = lasti + 1;
    while (i < pie_data.num_slices_pie) {
        pPie.pos[pie_num].coord[i].x = pPie.pos[pie_num].coord[0].x;
        pPie.pos[pie_num].coord[i].y = pPie.pos[pie_num].coord[0].y;
        i++;
    }
}

/**
 * Gets the appropriate fill style for the current pie slice.
 * @param pie_data derived data for this pie
 * @param slice slice number of this pie
 * @return fill style for this slice
 */
private JCFillStyle getPieSliceFillStyle(PieData pie_data, int slice)
{
    JCFillStyle fstyle = null;

    if ((slice == pie_data.other_slice) && (pie_data.other_flag != 0)) {
        fstyle = pcf.getOtherStyle().getFillStyle();
    }
    else {
        int index = pie_data.percents[slice].index;
        if (index >= 0) {
            ChartDataViewSeries series =
                ((ChartDataViewSeries)seriesList.get(index));
            fstyle = series.getStyle().getFillStyle();
        }
    }
    return(fstyle);
}

/**
 * Method that calculates data derived from the data in the data view, such
 * as percentage values of data points and the order of slices in each
 * pie chart.  This transient data is then stored in the <code>dataObject</code>.
 */
protected void calcTransientData()
{
    // If data has changed, recalc the pie slice data.  Otherwise, return.
    if (!dataObject.isChanged() &&
        dataObject.getTransientData() != null &&
        !recalcTransientData)
    {
        return;
    }
    recalcTransientData = false;

    // Get list of data series to be drawn using this drawable
    seriesList = dataObject.getSeries();
    pcf = (JCPieChartFormat)dataObject.getChartFormat();

    // Get the number of slices and pies
    int num_slices = dataObject.getNumSeries();
    Point firstLast = dataObject.getFirstLast();
    if (firstLast == null) {
        num_pies = 0;
        return;
    }
    num_pies = Math.max(firstLast.y - firstLast.x + 1, 0);
    if (num_pies == 0) {
        return;
    }

    PieChartInfo pPie = null;
    // If data has changed, allocate memory for new data calculations
    pPie = new PieChartInfo();
    pPie.pos = new PiePosition[num_pies];
    pPie.data = new PieData[num_pies];
    pPie.num_pies = num_pies;
    for (int i = 0; i < num_pies; i++) {
        pPie.pos[i] = new PiePosition();
        pPie.pos[i].coord = new Point[num_slices + 1];
        pPie.pos[i].offset = new Point[num_slices + 1];

        for (int j = 0; j <= num_slices; j++) {
            pPie.pos[i].coord[j] =
                new Point(COORD_DEFAULT_VALUE, COORD_DEFAULT_VALUE);
            pPie.pos[i].offset[j] = new Point(0, 0);
        }
        pPie.pos[i].order = new short[num_slices + 1];

        pPie.data[i] = new PieData();
        pPie.data[i].sortPoints = new PieSortPoint[num_slices];
        for (int j = 0; j < num_slices; j++) {
            pPie.data[i].sortPoints[j] = new PieSortPoint();
        }
        pPie.data[i].percents = new PiePercent[num_slices + 1];
        pPie.data[i].exploded = new boolean[num_slices + 1];
        for (int j = 0; j < num_slices + 1; j++) {
            pPie.data[i].percents[j] = new PiePercent();
            pPie.data[i].exploded[j] = false;
        }
    }

    // Calculate transient data for each pie chart
    first_point = ((ChartDataViewSeries)seriesList.get(0)).getFirstPoint();
    pPie.other_flag = 0;
    ArrayList explodeAllPoints = getExplodeAllPoints();
    for (int pie_num = 0; pie_num < num_pies; pie_num++) {
        /* Process data and decide if there is an "other" slice based
           on the threshold method/value and the sort order */
        int other_flag = computeOtherSlice(seriesList,
                                           pie_num + first_point,
                                           num_slices, pPie.data[pie_num],
                                           explodeAllPoints);
        int other_slice = pPie.data[pie_num].other_slice;
        int num_slices_pie = pPie.data[pie_num].num_slices_pie;

        int i = 0;
        for (; i < num_slices_pie; i++) {
            /* Save the order (used in pick/unpick) */
            pPie.pos[pie_num].order[i] =
                (short)pPie.data[pie_num].percents[i].index;

            // Last drawn slice is last non-zero slice
            if (pPie.data[pie_num].percents[i].pct > 0.0) {
                pPie.data[pie_num].lastDrawnSlice = i;
            }
        }
        for (; i <= num_slices; i++) {
            pPie.pos[pie_num].order[i] = (short)INDEX_NOT_DRAWN;
        }
        if (other_flag > 0) {
            pPie.pos[pie_num].order[other_slice] =
                (short)JCPieChartFormat.OTHER_SLICE;
        }
        pPie.other_flag |= other_flag;
    }

    // Store transient data in data object
    dataObject.setTransientData(pPie);

    dataObject.setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Called in the <code>update()</code> method of <code>JCChart</code> if the
 * data has changed and the Pie chart is not visible.  This method sets a flag
 * that forces the recalculation of transient data.
 */
protected void setRecalcFlag()
{
    recalcTransientData = true;
}

/**
 * Picks the nearest slice to the given point.
 * @param pt point used to find the nearest pie
 * @param focus pick focus (not used by this chart type)
 * @return an index to the data point in the series that is represented by
 * the picked slice
 */
public JCDataIndex pick(Point pt, int focus)
{
    //  Get the location and size of the drawing area.
    Rectangle rect = chartArea.getDrawingArea();
    //  The drawing area is surrounded by a border.  We need to translate
    //  the pick point by the thickness of that border.
    Point p = new Point(pt.x - rect.x, pt.y - rect.y);

    JCDataIndex di = null;

    PieData pie_data = new PieData();
    pie_data.sortPoints = null;
    pie_data.percents = null;

    // Get the number of slices and pies
    if (seriesList == null)
        return di;
    int num_slices = seriesList.size();
    if (num_slices == 0)
        return di;

    pie_data.sortPoints = new PieSortPoint[num_slices];
    for (int j = 0; j < num_slices; j++) {
        pie_data.sortPoints[j] = new PieSortPoint();
    }
    pie_data.percents = new PiePercent[num_slices + 1];
    pie_data.exploded = new boolean[num_slices + 1];
    for (int j = 0; j < num_slices + 1; j++) {
        pie_data.percents[j] = new PiePercent();
        pie_data.exploded[j] = false;
    }

    // The dimension and position of each pie within its enclosing
    // rectangle/box is the same.  Compute this constant information once.

    int pie_width = 0;
    int pie_height = 0;
    // Pie not actually drawn yet
    if (pieInfo == null) {
        return(di);
    }
    else {
        pie_width  = pieInfo.width;
        pie_height = pieInfo.height;
        if (pie_width == 0 || pie_height == 0) {
            return(di);
        }
    }

    int pie_offset = (box_width - pie_width) / 2;
    if (pie_offset < 0) {
        pie_offset = 0;
    }

    //  Determine the pie in whose enclosing rectangle the mouse_event
    //  took place.  If the mouse event was outside of all the enclosing
    //  rectangles (as can happen with the last row of pies having fewer
    //  pies than the other rows), then the event is treated as happening
    //  outside the graph area.

    int column = p.x / box_width;
    if (column >= num_pies_line) {
        column = num_pies_line - 1;
    }
    int row = p.y / box_height;
    if (row >= num_lines) {
        row = num_lines - 1;
    }
    int pie_num = row * num_pies_line + column +
        ((ChartDataViewSeries)seriesList.get(0)).getFirstPoint();

    int box_x = 0;
    int box_y = 0;

    box_x += column * box_width;
    box_y += row * box_height;

    ArrayList explodeAllPoints = getExplodeAllPoints();
    if ((pie_num < num_pies +
        ((ChartDataViewSeries)seriesList.get(0)).getFirstPoint()) &&
        (p.x >= box_x) && (p.x <= box_x + box_width) &&
        (p.y >= box_y) && (p.y <= box_y + box_height)) {
        int other_flag = computeOtherSlice(seriesList, pie_num, num_slices,
                                           pie_data, explodeAllPoints);
        setShapeOperation(ProcessShapeModel.SHAPE_PICK);
        picker = new Picker(p.x, p.y, pieInfo.data[pie_num]);
        Graphics gc = chartArea.getGraphics();
        if (gc != null) {
            gc.translate(rect.x, rect.y);
            drawSinglePie(gc, pie_num, box_x, box_y);
            gc.dispose();
        }
        resetShapeOperation();

        if (other_flag != 0 &&
            picker.set == JCPieChartFormat.OTHER_SLICE) {
            if (picker.point >= 0) {
                di = new JCDataIndex(picker.point, null,
                                     JCPieChartFormat.OTHER_SLICE);
                di.distance = picker.distance;
            }
        }
        else if (picker.set >= 0 && picker.point >= 0) {
            ChartDataViewSeries s =
                (ChartDataViewSeries)seriesList.get(picker.set);
            di = new JCDataIndex(picker.point, s, picker.set);
            di.distance = picker.distance;
        }
    }

    return (di);
}


/**
 * Unpick method for pie charts.  Given a data series and a datapoint
 * within that series, the function will return pixel coordinates of
 * that point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing position in screen pixels
 */
public Point unpick(int pt, int seriesIndex)
{
    PieData pie_data = pieInfo.data[pt];
    Point pixelCoords = new Point();
    double theta = find_point(seriesIndex, pt, pixelCoords);

    if (pixelCoords.x < 0 || pixelCoords.y < 0) {
        return(null);
    }

    //  If the slice is exploded, adjust the position by the offset.
    int sliceIndex = findSliceIndex(seriesIndex, pt);
    if (pie_data.exploded[pie_data.percents[sliceIndex].index]) {
        int xOffset = (int)Math.round(Math.cos(theta) * explode_offset);
        int yOffset = (int)Math.round(-Math.sin(theta) * explode_offset);
        pixelCoords.x += xOffset;
        pixelCoords.y += yOffset;
    }

    return(pixelCoords);
}

/**
 * Returns the slice index corresponding to the set and point parameters.
 * Returns -1 if unable to find the slice index.
 * @param point
 * @param set
 * @return -1 if unable to find the slice index
 */
private int findSliceIndex(int set, int point)
{
    int nsets = seriesList.size();
    for (int i = 0; i < nsets + 1; i++) {
        int pointIndex =
            point - ((ChartDataViewSeries)seriesList.get(0)).getFirstPoint();
        int order = pieInfo.pos[pointIndex].order[i];
        if (order == set) {
            return(i);
        }
    }
    return(-1);
}

/**
 * Given a set and point, finds the point on the pie arc which bisects
 * the corresponding slice.
 * @param set index of the data series
 * @param point index of the data point within the series
 * @param pixelCoords the found point.  The <i>x</i>,<i>y</i> coordinates will
 * be set to -1 if such a point is not feasible to find (it does not work for a
 * zero pie).
 * @return the angle of the found point, Double.MAX_VALUE if unable to find
 * such a point.
 */
private double find_point(int set, int point, Point pixelCoords)
{
    int nsets = seriesList.size();

    if (set != JCPieChartFormat.OTHER_SLICE) {
        if (!((ChartDataViewSeries)seriesList.get(set)).isVisible()) {
            pixelCoords.x = -1;
            pixelCoords.y = -1;
            return (Double.MAX_VALUE);
        }
    }

    double elevation =
        JCChartUtil.degToRad(Math.max(0, chartArea.getElevation()));

    double cos_inclination = Math.cos(elevation);

    int pix_x = -1;
    int pix_y = -1;
    double theta = Double.MAX_VALUE;

    for (int i = 0; i < nsets + 1; i++) {
        int pointIndex =
            point - ((ChartDataViewSeries)seriesList.get(0)).getFirstPoint();
        int order = pieInfo.pos[pointIndex].order[i];
        if (order == INDEX_NOT_DRAWN) {
            /* This (set, point) not drawn -- part of other slice */
            pixelCoords.x = -1;
            pixelCoords.y = -1;
            return(Double.MAX_VALUE);
        }
        else if (order == INDEX_HOLE) {
            /* Skip over data holes -- cannot unpick them */
            continue;
        }
        else if (order == set) {
            /*
             *  Get the coordinates of the endpoints of this slice.
             */
            double x1 = pieInfo.pos[pointIndex].coord[i].x;
            double y1 = pieInfo.pos[pointIndex].coord[i].y;
            double x2, y2;

            /* Is this a zero pie? */
            if (x1 == COORD_DEFAULT_VALUE) {
                pixelCoords.x = -1;
                pixelCoords.y = -1;
                return(Double.MAX_VALUE);
            }
            if (i == nsets || pieInfo.pos[pointIndex].order[i + 1] == -1) {
                x2 = pieInfo.pos[pointIndex].coord[0].x;
                y2 = pieInfo.pos[pointIndex].coord[0].y;
            }
            else {
                x2 = pieInfo.pos[pointIndex].coord[i + 1].x;
                y2 = pieInfo.pos[pointIndex].coord[i + 1].y;
            }

            /*
             *  We need to find the actual angles on the
             *  face of the pie, so rotate it back,
             *  yielding a circle again.
             */

            y1 = (y1 - pieInfo.pos[pointIndex].cy) / cos_inclination +
                 pieInfo.pos[pointIndex].cy;
            y2 = (y2 - pieInfo.pos[pointIndex].cy) / cos_inclination +
                 pieInfo.pos[pointIndex].cy;

            /*
             *  Find the two bounding angles, and bisect
             *  it, yielding theta.
             */

            double theta1 = Math.atan2(pieInfo.pos[pointIndex].cy - y1,
                                       x1 - pieInfo.pos[pointIndex].cx);
            double theta2 = Math.atan2(pieInfo.pos[pointIndex].cy - y2,
                                       x2 - pieInfo.pos[pointIndex].cx);
            theta = theta1 - theta2;
            if (theta < 0) {
                theta += 2 * JCChartUtil.M_PI;
            }
            else if (theta == 0) {
                // Fix for PR #31929.  Due to pixel round-off, the values of
                // theta1 and theta2 may be identical.  Look at the fractional
                // value of this slice to determine whether the desired point
                // is at theta or theta + 2PI.
                double percent = pieInfo.data[pointIndex].percents[order].pct;
                // "percent" is really a fractional value
                if (percent > 0.5) {
                    theta += 2 * JCChartUtil.M_PI;
                }
            }
            theta = theta1 - theta / 2.0;
            if (theta <= -JCChartUtil.M_PI) {
                theta += 2 * JCChartUtil.M_PI;
            }

            /*
             *  Now deal with the special cases first; in the general
             *  case, solve for the point on the circle with
             *          y = x * tan(theta)
             * and rotate it back to the viewing position.
             */
            if (theta == JCChartUtil.M_PI_2) {
                pix_x = pieInfo.pos[pointIndex].cx;
                pix_y = pieInfo.pos[pointIndex].cy - pieInfo.height / 2;
            }
            else if (theta == -JCChartUtil.M_PI_2) {
                pix_x = pieInfo.pos[pointIndex].cx;
                pix_y = pieInfo.pos[pointIndex].cy + pieInfo.height / 2;
            }
            else {
                double a = pieInfo.width / 2.0;
                double tant = Math.tan(theta);
                x1 = Math.sqrt((a * a) / (1 + tant * tant));
                if (theta > JCChartUtil.M_PI_2 ||
                    theta < -JCChartUtil.M_PI_2)
                {
                    x1 = -x1;   /* 3rd or 4th quadrant */
                }
                y1 = x1 * tant * cos_inclination;
                pix_x = (int)(pieInfo.pos[pointIndex].cx + x1);
                pix_y = (int)(pieInfo.pos[pointIndex].cy - y1);
            }
            break;
        }
    }
    pixelCoords.x = pix_x;
    pixelCoords.y = pix_y;

    return(theta);
}

private void calcExplodeOffset(double start_angle, double bisect_angle,
                               double dist, Point offset)
{
    double angle = start_angle - bisect_angle;
    if (angle < 0.0) {
        angle += 360.0;
    }

    double rad = 0.0;
    if (angle >= 0.0 && angle <= 90.0) {
        rad = JCChartUtil.degToRad(angle);
        offset.x = (int)Math.round(Math.cos(rad) * dist);
        offset.y = -(int)Math.round(Math.sin(rad) * dist);
    }
    else if (angle > 90.0 && angle <= 180.0) {
        angle -= 90.0;
        rad = JCChartUtil.degToRad(angle);
        offset.x = -(int)Math.round(Math.sin(rad) * dist);
        offset.y = -(int)Math.round(Math.cos(rad) * dist);
    }
    else if (angle > 180.0 && angle <= 270.0) {
        angle -= 180.0;
        rad = JCChartUtil.degToRad(angle);
        offset.x = -(int)Math.round(Math.cos(rad) * dist);
        offset.y = (int)Math.round(Math.sin(rad) * dist);
    }
    else if (angle > 270.0 && angle <= 360.0) {
        angle -= 270.0;
        rad = JCChartUtil.degToRad(angle);
        offset.x = (int)Math.round(Math.sin(rad) * dist);
        offset.y = (int)Math.round(Math.cos(rad) * dist);
    }
}

}
