/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PolarChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:47 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;
import java.awt.geom.Line2D;

/**
 * Drawable object for polar charts.  This object
 * is used for rendering a polar chart based on data
 * contained in the <code>dataObject</code>.
 */
public class PolarChartDraw extends ChartDraw {

// Local copies of data needed for draw(), pick(), and unpick().  They are
// set in recalc().
protected MinMax xlim;
protected MinMax ylim;
private boolean isRadar = false;

/**
 * Default constructor.
 */
public PolarChartDraw() {
	super();
	xlim = new MinMax();
	ylim = new MinMax();
}

/**
 * Method used to recalculate extents of related objects.
 */
public void recalc() {
	recalc(-1, -1, -1);
}

/**
 * Method used to recalculate extents of related objects.
 *
 * @param startPoint the index of the first plotted point
 * @param endPoint the index of the last plotted point
 */
public void recalc(int startPoint, int endPoint) {
	recalc(startPoint, endPoint, -1);
}

/**
 * Method used to recalculate extents of related objects.
 *
 * @param startPoint the index of the first plotted point
 * @param endPoint the index of the last plotted point
 * @param seriesIndex an index into the series List
 */
public void recalc(int startPoint, int endPoint, int seriesIndex) {
	if (!isChanged()) {
		return;
	}

	boolean fu = false;
	// if start point and end point are specified, this is a fast
	// update case
	if (startPoint >= 0 && endPoint >= 0 && endPoint >= startPoint) {
		fu = true;
	}

	// Make a list of points
	// If calling for one series, use existing series list.  Otherwise,
	// refresh the series list.
	if (seriesIndex == -1) {
		seriesList = dataObject.getSeries();
	}
	if (seriesList == null) {
		return;
	}

	// cache important info for later use by draw()
	xaxis = dataObject.getXAxis();
	yaxis = dataObject.getYAxis();
	holeValue = dataObject.getHoleValue();
	inverted = xaxis.i_vertical;

	// In fast update case, check the limits against the new points
	if (fu) {
		int firstSeries = 0;
		int lastSeries = seriesList.size();
		for (int i = firstSeries; i < lastSeries; i++) {
			ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(i);
			if (!series.isVisible()) continue;
			if (series.maxXIndex() == 0 || series.maxYIndex() == 0) continue;
			double yvals[] = series.getY();
			for (int j = startPoint; j <= endPoint; j++) {
				ylim.setMinMax(yvals[j], holeValue, yaxis.isLogarithmic());
			}
		}
		dataObject.getDataBounds(xlim, ylim);
	}
	else if (seriesIndex == (-1) || seriesIndex == 0) {

		if (dataObject.getXLimits() == null) {
			return;
		}
		if (dataObject.getYLimits() == null) {
			return;
		}

		// This may take awhile for large data set
		xlim = new MinMax(dataObject.getXLimits().min,
						  dataObject.getXLimits().max);
		ylim = new MinMax(dataObject.getYLimits().min,
						  dataObject.getYLimits().max);
		dataObject.getDataBounds(xlim, ylim);
	}
}

/**
 * Method used to draw related objects.
 * @param gc
 */
public void draw(Graphics gc) {
	draw(gc, -1, -1);
}

/**
 * Method used to draw related objects.
 * @param endPoint
 * @param gc
 * @param startPoint
 */
public void draw(Graphics gc, int startPoint, int endPoint) {
   	Rectangle clipBounds = gc.getClipBounds();
   	if (clipBounds.width == 0 || clipBounds.height == 0) {
       	return;
	}

	// Get list of data series to be drawn using this drawable
	if (seriesList == null) {
		return;
	}

	boolean fu = false;
	//  If both startPoint and endPoint are specified, then this
	//  is a fast update case.
	if (startPoint >= 0 && endPoint >= 0 && endPoint >= startPoint) {
		fu = true;
	}

	// create a separate GC to use when drawing the lines
	Graphics lineGC = gc.create();

	DefineShape lineShapeDef = null;
	DefineShape pointShapeDef = null;
	if (lineGC instanceof DefineShape) {
		lineShapeDef = (DefineShape) lineGC;
		pointShapeDef = (DefineShape) gc;
	}

	// Setup parameters to loop thru the data series
	int firstSeries = 0;
	int lastSeries = seriesList.size();
	for (int i = firstSeries; i < lastSeries; i++) {
		// Set up the basic information: series, style, x vals, y vals);
		ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(i);

		// Drop out if data is uninteresting
		if (!series.isVisible()) {
			continue;
		}
		if (series.maxXIndex() == 0 || series.maxYIndex() == 0) {
			continue;
		}

		// if not fast update, get start point and end point.
		if (!fu) {
			startPoint = getFirstPoint(series);
			endPoint = getLastPoint(series);
		} else {
			int boundaryPoint = series.getLastPoint();
			if (startPoint < boundaryPoint) startPoint = boundaryPoint;
		}

		double xvals[] = series.getX();
		double yvals[] = series.getY();

		// Initialize first point to a false point
		double x1 = 0.0;
		double y1 = holeValue;

		// If fast update case, first point is the last point in the
		// previously drawn.
		if (fu && startPoint > 0) {
			x1 = xvals[startPoint - 1];
			y1 = yvals[startPoint - 1];
		}

		JCChartStyle style = series.getStyle();
		if (style == null) continue;
		// Set up the line style
		JCLineStyle lstyle = style.getLineStyle();

		boolean noLines = false;
		if (!(lstyle.updateGraphics(lineGC))) {
			noLines = true;
		}

		Line2D.Double lineSegment = new Line2D.Double();

		// Set up the point style
		JCSymbolStyle pstyle = style.getSymbolStyle();
		boolean noShape = (pstyle.getShape() == JCSymbolStyle.NONE);

		// previous point
		PlotPoint prevPt = new PlotPoint();

		int j;
		int nPoints = endPoint - startPoint + 1;
		for (j = startPoint; j <= endPoint; j++) {
			double x2 = 0.0;
			double y2 = holeValue;

			boolean ylog = yaxis.isLogarithmic();
			//  xaxis is always linear for Polar/Radar charts
			x2 = xvals[j];
			y2 = yvals[j];
			if (ylog && y2 <= 0.0) {
				y2 = holeValue;
			}

			PlotPoint currPt = getPlotPoint(j-startPoint, nPoints, x2, y2);

			// Draw line between this new point and previous
			// point (if valid).
			if (!noLines && currPt.valid) {
				drawConnectingLine(gc, lineGC, currPt, prevPt, lstyle,
								   lineShapeDef, x1, y1, x2, y2, lineSegment);
			}

			if (prevPt.valid && prevPt.drawPoint && !noShape) {
				// Check for color handler and custom drawing color
				Color col = null;
				if (dataObject.getColorHandler() != null) {
					col = dataObject.getColorHandler().getColor(x1, y1);
				}

				// begin shape definition for this point
				if (pointShapeDef != null) {
					pointShapeDef.markShape(new JCDataIndex(j-1, series));
				}

				pstyle.draw(gc, prevPt.px, prevPt.py, col);

				// finish shape definition for this point
				if (pointShapeDef != null) {
					pointShapeDef.unmarkShape();
				}

			}

			// New Point becomes previous point
			prevPt.px = currPt.px;
			prevPt.py = currPt.py;
			prevPt.drawPoint = currPt.drawPoint;
			prevPt.valid = currPt.valid;

			x1 = x2;
			y1 = y2;
		}

		if (this instanceof RadarChartDraw) {
			// For Radar or Area Radar charts only:
			// Draw line between the final point and first
			// point (if valid) in the series.
			double x2 = 0.0;
			double y2 = holeValue;

			boolean ylog = yaxis.isLogarithmic();
			//  xaxis is always linear for Polar/Radar charts
			x2 = xvals[startPoint];
			y2 = yvals[startPoint];
			if (ylog && y2 <= 0.0) {
				y2 = holeValue;
			}

			PlotPoint currPt = getPlotPoint(0, nPoints, x2, y2);
			if (!noLines && currPt.valid) {
				drawConnectingLine(gc, lineGC, currPt, prevPt, lstyle,
								   lineShapeDef, x1, y1, x2, y2, lineSegment);
			}
		}

		// Draw symbol for final point (if valid);
		if (prevPt.valid && prevPt.drawPoint && !noShape) {
			// Check for color handler and custom drawing color
			Color col = null;
			if (dataObject.getColorHandler() != null) {
				col = dataObject.getColorHandler().getColor(x1, y1);
			}

			// begin shape definition for this point
			if (pointShapeDef != null) {
				pointShapeDef.markShape(new JCDataIndex(endPoint, series));
			}

			pstyle.draw(gc, prevPt.px, prevPt.py, col);

			// finish shape definition for this point
			if (pointShapeDef != null) {
				pointShapeDef.unmarkShape();
			}
		}
	}
	lineGC.dispose();
}

protected int getFirstPoint(ChartDataViewSeries series) {
	return series.getFirstPoint();
}

protected int getLastPoint(ChartDataViewSeries series) {
	return series.getLastPoint();
}

protected void drawConnectingLine(Graphics gc, Graphics lineGC, PlotPoint currPt, PlotPoint prevPt,
								  JCLineStyle lstyle, DefineShape lineShapeDef,
								  double x1, double y1, double x2, double y2, Line2D.Double lineSegment) {
	if (prevPt.valid) {
		// Check for color handler and custom drawing color
		Color col = null;
		if (dataObject.getColorHandler() != null) {
			col = dataObject.getColorHandler().getColor(
						x1, y1, x2, y2);
			if (col == null) {
				col = lstyle.getColor();
			}
			if (!lineGC.getColor().equals(col)) {
				lineGC.setColor(col);
			}
		}

		if (lineShapeDef != null) {
			lineShapeDef.beginShape();
		}

		lineSegment.setLine(prevPt.px, prevPt.py, currPt.px, currPt.py);
		((Graphics2D)lineGC).draw(lineSegment);
		// End shape definition for this line
		if (lineShapeDef != null) {
			lineShapeDef.endShape(true);
		}
	}
}

/**
 * Returns chart area (<i>x</i>,<i>y</i>) coordinates (pixel units) for the <i>j</i>th data point.
 *
 * @param j array index of data point
 * @param nPoints length of data series (used in <code>RadarChartDraw</code>)
 * @param theta the angle of the <i>j</i>th data point
 * @param r the r value of the <i>j</i>th data point
 * @return the chart area pixel coordinates of the point to be plotted
 */
PlotPoint getPlotPoint(int j, int nPoints, double theta, double r)
{
	PlotPoint pp = new PlotPoint();
	if (xaxis!= null && yaxis != null && r != holeValue) {
		Point point = null;
		int angleUnit = dataObject.getParent().chartArea.getAngleUnit();
		double angle = JCChartUtil.convertAngle(angleUnit, JCChartUtil.RADIANS, theta);
		point = yaxis.toPixelPolar(angle, r, true);
   		pp.px = point.x;
   		pp.py = point.y;
		pp.drawPoint = true;
		pp.valid = true;
	}
	else {
   		pp.px = Integer.MAX_VALUE;
   		pp.py = Integer.MAX_VALUE;
		pp.drawPoint = false;
		pp.valid = false;
	}

	return pp;
}

//  Override this method for Radar charts.
protected double getThetaValue(int index, int startPoint, int endPoint,
							   double[] xValues) {
	return xValues[index];
}

/**
 * Pick method for Polar charts.  Given a point relative to the
 * chart area, this method will find the closest point.
 * @param pt (<i>x</i>,<i>y</i>) value relative to the chart area
 * @param focus pick focus used to determine distance between pick point
 * and data point
 * @return <code>JCDataIndex</code> encapsulating the point index and data
 * series of the selected point
 */
public JCDataIndex pick(Point pt, int focus) {
	Rectangle r = chartArea.getDrawingArea();
	//  Get coordinates relative to origin of chart area.
	int px = pt.x - r.x;
	int py = pt.y - r.y;
	JCDataCoord pickPoint = yaxis.toDataPolar(px, py);
	JCDataIndex di = null;

	if (seriesList == null) return di;

	int firstSeries = 0;
	int lastSeries = seriesList.size();

	//  STAGE ONE:
	//
	//  Search all of the series for the data point closest to the pick
	//  point according to the pick focus.  Note that in the case of
	//  PICK_FOCUS_X or PICK_FOCUS_Y, there may be points having the same
	//  theta or r value.
	int series = -1;
	int point = -1;
	double length = Double.MAX_VALUE;   //  difference in theta or r values
	double len = Double.MAX_VALUE;
	int primaryDistance = Integer.MAX_VALUE;
	try {
	JCDataIndex diPrimary = new JCDataIndex();
	for (int i = firstSeries; i < lastSeries; i++) {
		len = findClosestPointPrimary(focus, i, diPrimary, px, py,
									  pickPoint);
		if (len < length) {
			length = len;
			series = i;
			point = diPrimary.point;
			primaryDistance = diPrimary.distance;
		}
	}
	diPrimary.seriesIndex = series;
	diPrimary.point = point;
	diPrimary.distance = primaryDistance;

	//  STAGE TWO:
	//
	//  For PICK_FOCUS_X or PICK_FOCUS_Y, two or more data points
	//  may have the same theta or r value.  Need to find the
	//  point closest (in pixels) to the pick point.
	int series2 = -1;
	int point2 = -1;
	int secondaryDistance = Integer.MAX_VALUE;
	if (focus != ChartDataView.PICK_FOCUS_XY) {
		JCDataIndex diSecondary = new JCDataIndex();
		boolean ok = false;
		for (int i = firstSeries; i < lastSeries; i++) {
			ok = findClosestPointSecondary(focus, i, diPrimary, diSecondary,
										   px, py, pickPoint);
			if ( ok &&
				 (diSecondary.distance < secondaryDistance) ) {
				secondaryDistance = diSecondary.distance;
				series2 = i;
				point2 = diSecondary.point;
			}
		}
	}

	if ( (series >= 0) && (point >= 0) ) {
		int closestSeries = series;
		int closestPoint = point;
		int closestDistance = primaryDistance;

		if ( (series2 >=0) && (point2 >= 0) &&
			 (secondaryDistance < primaryDistance) ) {
			//  A closer point was found in the secondary search.
			closestSeries = series2;
			closestPoint = point2;
			closestDistance = secondaryDistance;
		}

		di = new JCDataIndex(closestPoint,
						(ChartDataViewSeries) seriesList.get(closestSeries),
						closestSeries);
		di.distance = closestDistance;
	}

//	System.out.println("pick():  di.seriesIndex=" + di.seriesIndex +
//	", di.point=" + di.point + ", di.distance=" + di.distance);

	return di;
	} catch (Exception e) {
		e.printStackTrace(System.out);
		return null;
	}
}

/*
 * This method searches the specified series for the data point that
 * is closest to the pick point.  This method cannot resolve points
 * that have the same theta or r value.
 * <p>
 * The return value is the minimum theta or r separation, depending
 * on the pick focus.  If an error occurs during the search, this
 * method will return a negative value.
 */
private double findClosestPointPrimary(int focus, int seriesIndex,
									JCDataIndex di,
									int px, int py,
									JCDataCoord pickPoint) {

	ChartDataViewSeries thisSeries =
			(ChartDataViewSeries) seriesList.get(seriesIndex);
	if (thisSeries == null || !thisSeries.isVisible()) {
		return -1.0;
	}

	int startPoint = getFirstPoint(thisSeries);
	int endPoint = getLastPoint(thisSeries);
	if (startPoint == Integer.MAX_VALUE || endPoint == Integer.MAX_VALUE) {
		return -1.0;
	}
	double xvals[] = thisSeries.getX();
	double yvals[] = thisSeries.getY();

	isRadar = (this instanceof RadarChartDraw);

	double length = Double.MAX_VALUE;
	double len = Double.MAX_VALUE;
		//  since JCDataIndex.distance is an integer
	int point = -1;
	double thispx = 0.0;
	double thispy = 0.0;
	double pxClosest = holeValue;
	double pyClosest = holeValue;
	double pxBoundAngle = 0.0;
	boolean ylog = yaxis.isLogarithmic();
	int angleUnit = yaxis.chartArea.getAngleUnit();
	for (int j = startPoint; j <= endPoint ; j++) {
		thispx = getThetaValue(j, startPoint, endPoint, xvals);
		thispy = yvals[j];

		if (thispx == holeValue ||
			thispy == holeValue ||
			thispy > ylim.max ||
			(ylog && thispy <= 0.0) ) {
			continue;
		}

		if (!isRadar && !yaxis.isReversed() && (thispy < 0)) {
			//  In a Polar chart, data points with negative y values are
			//  displayed as positive y at (theta+180) degrees.
			thispy *= -1.0;
			thispx += JCChartUtil.convertAngle(JCChartUtil.DEGREES, angleUnit,
												180.0);
		}
		pxBoundAngle = JCChartUtil.boundAngle(angleUnit, thispx);

		len = Double.MAX_VALUE;
		switch(focus) {
		case (ChartDataView.PICK_FOCUS_X) :
			//  data space
			len = Math.abs(pxBoundAngle - pickPoint.x);
			len = Math.min(len, JCChartUtil.boundAngle(angleUnit, -len));
			break;
		case (ChartDataView.PICK_FOCUS_Y) :
			//  data space
			len = Math.abs(pickPoint.y - thispy);
			break;
		case (ChartDataView.PICK_FOCUS_XY) :
		default :
			//  screen space
			len = calcXYDistance(px, py, thispx, thispy);
		}

		//  Find the closest point according to pick rule:
		if (len < length) {
			point = j;
			pxClosest = thispx;
			pyClosest = thispy;
			length = len;
		}
	}

	di.seriesIndex = seriesIndex;
	di.point = point;
	if (focus == ChartDataView.PICK_FOCUS_XY) {
		di.distance = (int)Math.round(length);
	}
	else if ( (pxClosest != holeValue) && (pyClosest != holeValue) ) {
		di.distance =
			(int)Math.round(calcXYDistance(px, py, pxClosest, pyClosest));
	}
	else {
		di.distance = Integer.MAX_VALUE;
	}

	return length;
}

/*
 *  In case two or more data points have the same theta or r value,
 *  we need to call this method to find the data point closest to
 *  the pick point (in pixels).
 */
private boolean findClosestPointSecondary(int focus, int seriesIndex,
										JCDataIndex diPrimary,
										JCDataIndex di,
										int px, int py,
										JCDataCoord pickPoint) {

	if ( (focus != ChartDataView.PICK_FOCUS_X) &&
	     (focus != ChartDataView.PICK_FOCUS_Y) ) {
		return false;
	}

	ChartDataViewSeries primarySeries =
			(ChartDataViewSeries) seriesList.get(diPrimary.seriesIndex);
	if (primarySeries == null || !primarySeries.isVisible()) {
		return false;
	}
	double xvalsPrimary[] = primarySeries.getX();
	double yvalsPrimary[] = primarySeries.getY();
	double xPrimary = xvalsPrimary[diPrimary.point];
	double yPrimary = yvalsPrimary[diPrimary.point];

	ChartDataViewSeries thisSeries =
			(ChartDataViewSeries) seriesList.get(seriesIndex);
	if (thisSeries == null || !thisSeries.isVisible()) {
		return false;
	}

	int startPoint = getFirstPoint(thisSeries);
	int endPoint = getLastPoint(thisSeries);
	if (startPoint == Integer.MAX_VALUE ||
		endPoint == Integer.MAX_VALUE) {
		return false;
	}

	double xvals[] = thisSeries.getX();
	double yvals[] = thisSeries.getY();

	isRadar = (this instanceof RadarChartDraw);

	int point = -1;
	double length = Integer.MAX_VALUE;
	double len;
	double thispx = 0.0;
	double thispy = 0.0;
	double thispxClosest = holeValue;
	double thispyClosest = holeValue;
	double pxBoundAngle = 0.0;
	boolean ylog = yaxis.isLogarithmic();
	int angleUnit = yaxis.chartArea.getAngleUnit();

	for (int j = startPoint; j <= endPoint ; j++) {
		thispx = getThetaValue(j, startPoint, endPoint, xvals);
		thispy = yvals[j];

		if (thispx == holeValue ||
			thispy == holeValue ||
			thispy > ylim.max ||
			(ylog && thispy <= 0.0) ) {
			continue;
		}

		if (!isRadar && !yaxis.isReversed() && (thispy < 0)) {
			//  In a Polar chart, data points with negative y values are
			//  displayed as positive y at (theta+180) degrees.
			thispy *= -1.0;
			thispx += JCChartUtil.convertAngle(JCChartUtil.DEGREES, angleUnit,
												180.0);
		}
		pxBoundAngle = JCChartUtil.boundAngle(angleUnit, thispx);

		len = Double.MAX_VALUE;
		switch(focus) {
		case (ChartDataView.PICK_FOCUS_X) :
			if (thispx != xPrimary) {
				continue;
			}
			//  To break a tie, we look for the closest point
			//  along the radial axis.
			len = Math.abs(pickPoint.y - thispy);
			break;
		case (ChartDataView.PICK_FOCUS_Y) :
			if (thispy != yPrimary) {
				continue;
			}
			//  To break a tie, we look for the closest point
			//  along the theta axis.
			len = Math.abs(pxBoundAngle - pickPoint.x);
			len = Math.min(len,JCChartUtil.boundAngle(angleUnit,-len));
			break;
		default :
		}

		if (len < length) {
			point = j;
			length = len;
			thispxClosest = thispx;
			thispyClosest = thispy;
		}
	}

	di.seriesIndex = seriesIndex;
	di.point = point;
	if ((thispxClosest != holeValue) && (thispyClosest != holeValue)) {
		di.distance = (int)calcXYDistance(px, py, thispxClosest,
										  thispyClosest);
	}
	else {
		di.distance = Integer.MAX_VALUE;
	}

	return true;
}

/*
 *  Search a specified series for the closest point.
 */
protected JCDataIndex findClosestPointInSeries(int series, int focus,
											   int px, int py,
											   JCDataCoord pickPoint) {
	// Need to calculate ylim which is used in findClosestPointPrimary()
	// and findClosestPointSecondary().
	ylim = new MinMax(dataObject.getYLimits().min,
					  dataObject.getYLimits().max);
	JCDataIndex di = null;
	JCDataIndex diPrimary = new JCDataIndex();
	diPrimary.setSeries(dataObject.getSeries(series));
	double len = findClosestPointPrimary(focus, series, diPrimary, px, py,
										 pickPoint);
	//  In case two or more points have the same theta or r value,
	//  we call findClosestPointSecondary().  That can happen only
	//  if focus is PICK_FOCUS_X or PIC_FOCUS_Y.  We break a tie
	//  situation by taking the data point that is closest (in pixels)
	//  to the pick point.
	if (len >= 0) {
		if (focus != ChartDataView.PICK_FOCUS_XY) {
			JCDataIndex diSecondary = new JCDataIndex();
			diSecondary.setSeries(dataObject.getSeries(series));
			boolean ok = findClosestPointSecondary(focus, series, diPrimary,
												   diSecondary, px, py,
												   pickPoint);
			if (ok && diSecondary.distance < diPrimary.distance) {
				di = diSecondary;
			}
			else {
				di = diPrimary;
			}
		}
		else {
			di = diPrimary;
		}
	}

	return di;
}

private double calcXYDistance(int xPick, int yPick, double xData,
							  double yData) {
	//  toPixelPolar() requires theta to be in radians.
	int angleUnit = dataObject.getParent().chartArea.getAngleUnit();
	double angleRadians;
	if (isRadar) {
		//  Value returned from getRadarAngle() is already in radians.
		angleRadians = xData;
	}
	else {
		angleRadians = JCChartUtil.convertAngle(angleUnit,
										JCChartUtil.RADIANS, xData);
	}
	Point pnt = yaxis.toPixelPolar(angleRadians, yData, true);
	int dx = pnt.x - xPick;
	int dy = pnt.y - yPick;

//	System.out.println("xData=" + xData + ", yData=" + yData +
//						", angleRadians=" + angleRadians +
//						", pnt.x=" + pnt.x + ", pnt.y=" + pnt.y +
//						", xPick=" + xPick + ", yPick=" + yPick);

	return Math.sqrt(dx*dx + dy*dy);
}

/**
 * Unpick method for Polar charts.  Given a data series and a datapoint
 * within that series, the function will return pixel coordinates of
 * that point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing position in screen pixels
 */
public Point unpick(int pt, int seriesIndex) {
	ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(seriesIndex);

	double yvalue = series.getY(pt);
	double xvalue = series.getX(pt);

	if (yaxis != null && yvalue != holeValue && xvalue != holeValue) {
		xvalue = JCChartUtil.convertAngle(
						dataObject.getParent().chartArea.getAngleUnit(),
						JCChartUtil.RADIANS, xvalue);
		return (yaxis.toPixelPolar(xvalue, yvalue, true));
	}

	return null;
}

/** Used to determine sign of <code>radialDistance</code> in <code>showEdit()</code> and <code>endEdit()</code>. */
Line2D.Double perpendicular = null;

/** Location on the chart area of the point before the point being edited. */
Point prev = null;

/** Location on the chart area of the point after the point being edited. */
Point next = null;

/** Last position on the chart area of the edited point. */
Point last = null;

/**
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di) {
	if (di == null || di.dataView != dataObject || di.series == null) {
		return;
	}
	prev = next = last = null;

	int j = di.point;

	double theta = Double.MAX_VALUE;
	double tmppx = di.series.getX(j);
	double tmppy = di.series.getY(j);

	if (tmppx != Double.MAX_VALUE &&
		tmppy != Double.MAX_VALUE &&
		tmppx != dataObject.getHoleValue() &&
		tmppy != dataObject.getHoleValue()) {
		theta = JCChartUtil.convertAngle(chartArea.getAngleUnit(),
										 JCChartUtil.RADIANS,
										 tmppx);
		double rValue = (yaxis.isReversed()) ?
							yaxis.min.value : yaxis.max.value;

		Point pnt1 = yaxis.toPixelPolar(JCChartUtil.boundAngle(
										JCChartUtil.RADIANS,
										theta + JCChartUtil.M_PI_2),
										rValue, true);
		Point pnt2 = yaxis.toPixelPolar(JCChartUtil.boundAngle(
										JCChartUtil.RADIANS,
										theta - JCChartUtil.M_PI_2),
										rValue, true);
		perpendicular = new Line2D.Double(pnt1.x, pnt1.y, pnt2.x, pnt2.y);
//		System.out.println("pnt1.x=" + pnt1.x + ", pnt1.y=" + pnt1.y +
//							",  pnt2.x=" + pnt2.x + ", pnt2.y=" + pnt2.y +
//							",  xCenter=" + yaxis.polarAxisParams.xCenter +
//							", yCenter=" + yaxis.polarAxisParams.yCenter);
	}

	if (j - 1 >= di.series.getFirstPoint()) {
		tmppx = di.series.getX(j-1);
		tmppy = di.series.getY(j-1);
		if (tmppx != Double.MAX_VALUE &&
			tmppy != Double.MAX_VALUE &&
			tmppx != dataObject.getHoleValue() &&
			tmppy != dataObject.getHoleValue()) {
			theta = JCChartUtil.convertAngle(chartArea.getAngleUnit(),
											 JCChartUtil.RADIANS,
											 tmppx);
			prev = yaxis.toPixelPolar(theta, tmppy, true);
		}
	}
	if (j+1 <= di.series.getLastPoint()) {
		tmppx = di.series.getX(j+1);
		tmppy = di.series.getY(j+1);
		if (tmppx != Double.MAX_VALUE &&
			tmppy != Double.MAX_VALUE &&
			tmppx != dataObject.getHoleValue() &&
			tmppy != dataObject.getHoleValue()) {
			theta = JCChartUtil.convertAngle(chartArea.getAngleUnit(),
											 JCChartUtil.RADIANS,
											 tmppx);
			next = yaxis.toPixelPolar(theta, tmppy, true);
		}
	}
}

/**
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y) {
	if (di == null ||
		di.dataView != dataObject ||
		perpendicular == null) {
		return;
	}

	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);

	double theta = JCChartUtil.convertAngle(chartArea.getAngleUnit(),
											JCChartUtil.RADIANS,
											di.series.getX(di.point));

	double radialDistance = perpendicular.ptLineDist((double)x, (double)y);
	int relativePosition = perpendicular.relativeCCW((double)x, (double)y);
	if (relativePosition == -1) {
		radialDistance *= -1.0;
	}

	JCAxis ax = di.dataView.getYAxis();
	last = ax.toPixelPolar(theta, radialDistance, false);

	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);
}

/**
 * Method called by the <code>EditHandler</code> when an edit is finished
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 * @return pixel distance from the center of the chart; <code>Integer.MIN_VALUE</code>
 * if an error occurs
 */
int endEdit(JCDataIndex di, int x, int y) {
	if (di == null ||
		di.dataView != dataObject ||
		perpendicular == null) {
		return Integer.MIN_VALUE;
	}
	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);

	double radialDistance = perpendicular.ptLineDist((double)x, (double)y);
	int relativePosition = perpendicular.relativeCCW((double)x, (double)y);
	if (relativePosition == -1) {
		radialDistance *= -1.0;
	}

	return (int)Math.round(radialDistance);
}

}
