/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: RadarChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:47 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.*;
import java.awt.geom.Line2D;

/**
 * Drawable object for radar charts.  This object
 * is used for rendering a radar chart based on data
 * contained in the <code>dataObject</code>.
 */
public class RadarChartDraw extends PolarChartDraw {

protected int firstPoint = Integer.MIN_VALUE;
protected int lastPoint = Integer.MAX_VALUE;

/**
 * Default constructor.
 */
public RadarChartDraw() {
	super();
}

/**
 * Method used to recalculate extents of related objects.
 *
 * @param startPoint the index of the first plotted point
 * @param endPoint the index of the last plotted point
 * @param seriesIndex an index into the series List
 */
public void recalc(int startPoint, int endPoint, int seriesIndex) {
	super.recalc(startPoint, endPoint, seriesIndex);

    // Make a list of points
	// If being called for a particular series, use previously stored series
	// list.  Otherwise, refresh series list.
	if (seriesIndex == -1) {
    	seriesList = dataObject.getSeries();
	}
	if (seriesList == null) {
		return;
	}

	firstPoint = Integer.MIN_VALUE;
	lastPoint = Integer.MAX_VALUE;
   	// Setup parameters to loop thru the data series
	int firstSeries = (seriesIndex == (-1)) ? 0 : seriesIndex;
	int lastSeries = (seriesIndex == (-1)) ? seriesList.size() : (seriesIndex+1);
	int index;
	for (int i = firstSeries; i < lastSeries; i++) {
        ChartDataViewSeries series = (ChartDataViewSeries) seriesList.get(i);
        if (!series.isVisible()) {
			continue;
		}
        index = series.getFirstPoint();
        if (index > firstPoint) {
			firstPoint = index;
		}
		index = series.getLastPoint();
        if (index < lastPoint) {
			lastPoint = index;
		}
	}
}

/**
 * Returns chart area (<i>x</i>,<i>y</i>) coordinates (pixel units) for the <i>j</i>th data point.
 *
 * @param j array index of data point
 * @param nPoints length of data series
 * @param theta the angle of the <i>j</i>th data point
 * @param r the r value of the <i>j</i>th data point
 * @return the chart area pixel coordinates of the point to be plotted
 */
PlotPoint getPlotPoint(int j, int nPoints, double theta, double r)
{
	PlotPoint pp = new PlotPoint();
	if (xaxis != null && yaxis != null && r != holeValue) {
		Point point = null;
		double radarAngle = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
													  nPoints,
													  j, false);
		//  We pass false for last parameter because xaxis reversal
		//  is handled in toPixelPolar()

		point = yaxis.toPixelPolar(radarAngle, r, true);

   		pp.px = point.x;
   		pp.py = point.y;
		pp.drawPoint = true;
		pp.valid = true;
	}
	else {
   		pp.px = Integer.MAX_VALUE;
   		pp.py = Integer.MAX_VALUE;
		pp.drawPoint = false;
		pp.valid = false;
	}

	return pp;
}

//  Override the PolarChartDraw method.
protected double getThetaValue(int index, int startPoint, int endPoint,
							   double[] xValues) {
	int nPoints = endPoint - startPoint + 1;
	return JCChartUtil.getRadarAngle(JCChartUtil.RADIANS, nPoints,
									 index-startPoint, false);
	//  We pass false for the last parameter because theta axis reversal
	//  is handled by the toPixelPolar() and toDataPolar() methods.
}

protected int getFirstPoint(ChartDataViewSeries series) {
	return firstPoint;
}

protected int getLastPoint(ChartDataViewSeries series) {
	return lastPoint;
}

/**
 * Unpick method for Radar charts.  Given a data series and a datapoint
 * within that series, the function will return pixel coordinates of
 * that point relative to the chart area.
 * @param pt point index
 * @param seriesIndex data series
 * @return point object representing the position in screen pixels
 */
public Point unpick(int pt, int seriesIndex) {

	ChartDataViewSeries series =
				(ChartDataViewSeries) seriesList.get(seriesIndex);

	double yvalue = series.getY(pt);

	int[] limits = ChartDataView.getFirstLastPoints(dataObject);
	int startPoint = limits[0];
	int endPoint = limits[1];

	int nPoints = endPoint - startPoint + 1;
	double xvalue = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS, nPoints,
											  pt-startPoint, false);
	//  We pass false for last parameter because xaxis reversal
	//  is handled in toPixelPolar()

	if (yaxis != null && yvalue != holeValue && xvalue != holeValue) {
		return (yaxis.toPixelPolar(xvalue, yvalue, true));
	}

	return null;
}

double thetaEditPoint = 0.;

/**
 * Override method.
 * <p>
 * Method called by the <code>EditHandler</code> when an edit is started
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 */
void startEdit(JCDataIndex di) {
	if (di == null || di.dataView != dataObject || di.series == null) {
		return;
	}
	prev = next = last = null;

	int[] limits = ChartDataView.getFirstLastPoints(dataObject);
	int firstPoint = limits[0];
	int lastPoint = limits[1];

	int nSpokes = lastPoint - firstPoint + 1;
	int j = di.point;
	int k = -1;

	double theta = Double.MAX_VALUE;
	double tmppx = di.series.getX(j);
	double tmppy = di.series.getY(j);

	if (this instanceof AreaRadarChartDraw) {
		xlim = new MinMax(dataObject.getXLimits().min,
						  dataObject.getXLimits().max);
		ylim = new MinMax(dataObject.getYLimits().min,
						  dataObject.getYLimits().max);
	}

	// Note:  Don't call JCChartUtil.locateXY().  In a Radar chart,
	// we call getRadarAngle() instead of using the x data values.
	// Hence, the xlim min and max values are always 0 and PI,
	// respectively, and the x data values can exceed those limits.
	if (tmppx != Double.MAX_VALUE &&
		tmppy != Double.MAX_VALUE &&
		tmppx != dataObject.getHoleValue() &&
		tmppy != dataObject.getHoleValue()) {

		theta = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
										  nSpokes, j-firstPoint,
										  xaxis.isReversed());
		thetaEditPoint = theta;

		double rValue = (yaxis.isReversed()) ?
							yaxis.min.value : yaxis.max.value;

		Point pnt1 = yaxis.toPixelPolar(JCChartUtil.boundAngle(
										JCChartUtil.RADIANS,
										theta + JCChartUtil.M_PI_2),
										rValue, true);
		Point pnt2 = yaxis.toPixelPolar(JCChartUtil.boundAngle(
										JCChartUtil.RADIANS,
										theta - JCChartUtil.M_PI_2),
										rValue, true);
		perpendicular = new Line2D.Double(pnt1.x, pnt1.y, pnt2.x, pnt2.y);

//		System.out.println("pnt1.x=" + pnt1.x + ", pnt1.y=" + pnt1.y +
//							",  pnt2.x=" + pnt2.x + ", pnt2.y=" + pnt2.y +
//							",  xCenter=" + yaxis.polarAxisParams.xCenter +
//							", yCenter=" + yaxis.polarAxisParams.yCenter);
	}

	//  If (j-1) is less than firstPoint(), set prev to lastPoint.
	if (j - 1 < firstPoint) {
		k = lastPoint;
	}
	else {
		k = j-1;
	}
	tmppx = di.series.getX(k);
	tmppy = di.series.getY(k);
	if (tmppx != Double.MAX_VALUE &&
		tmppy != Double.MAX_VALUE &&
		tmppx != dataObject.getHoleValue() &&
		tmppy != dataObject.getHoleValue()) {
		theta = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
										  nSpokes, k-firstPoint,
										  xaxis.isReversed());

		prev = yaxis.toPixelPolar(theta, tmppy, true);
	}

	//  If (j+1) is greater than lastPoint, set next to firstPoint.
	if (j+1 > lastPoint) {
		k = firstPoint;
	}
	else {
		k = j+1;
	}
	tmppx = di.series.getX(k);
	tmppy = di.series.getY(k);
	if (tmppx != Double.MAX_VALUE &&
		tmppy != Double.MAX_VALUE &&
		tmppx != dataObject.getHoleValue() &&
		tmppy != dataObject.getHoleValue()) {
		theta = JCChartUtil.getRadarAngle(JCChartUtil.RADIANS,
										  nSpokes, k-firstPoint,
										  xaxis.isReversed());

		next = yaxis.toPixelPolar(theta, tmppy, true);
	}
}

/**
 * Override method.
 * <p>
 * Method called by the <code>EditHandler</code> when an edit is occuring
 * on the data set plotted by this drawable.
 * @param di <code>JCDataIndex</code> instance specifying the data series and
 * point values for the point being edited
 * @param x <i>x</i> location of the action, relative to the chart area
 * @param y <i>y</i> location of the action, relative to the chart area
 */
void showEdit(JCDataIndex di, int x, int y) {
	if (di == null ||
		di.dataView != dataObject ||
		perpendicular == null) {
		return;
	}

	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);

	double radialDistance = perpendicular.ptLineDist((double)x, (double)y);
	int relativePosition = perpendicular.relativeCCW((double)x, (double)y);
	if (relativePosition == -1) {
		radialDistance *= -1.0;
	}

	JCAxis ax = di.dataView.getYAxis();
	last = ax.toPixelPolar(thetaEditPoint, radialDistance, false);

	chartArea.drawRubberLine(prev, last);
	chartArea.drawRubberLine(last, next);
}

}
