/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SAreaChartDraw.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:47 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import com.klg.jclass.util.graphics.DefineShape;

import java.awt.*;

/**
 * Drawable object for stacking area charts.  This object
 * is used for rendering a stacking area chart based on data
 * contained in the <code>dataObject</code>.  It uses most of the functionality
 * of <code>AreaChartDraw</code>, only overriding a few methods to provide stacking
 * values.  Some Stacking Bar functionality is actually coded into
 * <code>AreaChartDraw</code> for convenience. <code>FastUpdate</code> mode will not work
 * with this chart type.
 */
public class SAreaChartDraw extends AreaChartDraw {

// Arrays to store stacking information
double[] ytotal;
double[] ysum;
double[] ysum_neg;
double[][] ywork;
JCAreaChartFormat areaFormat;

/**
 * Default constructor.
 */
public SAreaChartDraw() {
	super();
}
 
/**
 * Overridden <code>recalc()</code> method.  Calculates stacking information and then
 * calls superclass <code>recalc()</code> method.
 * @param endPoint
 * @param startPoint
 * @param seriesIndex
 */
public void recalc(int startPoint, int endPoint, int seriesIndex) {

	areaFormat = (JCAreaChartFormat) dataObject.getChartFormat();
    holeValue = dataObject.getHoleValue();
	firstLast = dataObject.getFirstLast();
	yaxis = dataObject.getYAxis();
    // Make a list of points
	// If being called for a particular series, use previously stored series
	// list.  Otherwise, refresh series list.
	if (seriesIndex == -1) {
    	seriesList = dataObject.getSeries();
	}
	if (seriesList == null) return;

	int nsets = seriesList.size();
	if ((nsets == 0) || (firstLast == null))
		return;
	int npoints = firstLast.y - firstLast.x + 1;

	ytotal = new double[npoints];
	ysum = new double[npoints];
	ysum_neg = new double[npoints];
	ywork = new double[nsets][npoints];

	ChartDrawUtil.calcStackingInfo(ysum, ysum_neg, ytotal, ywork,
				firstLast.x, firstLast.y, seriesList,
				areaFormat.is100Percent(), holeValue, yaxis.dataLimits, yaxis);
    super.recalc(startPoint, endPoint, seriesIndex);
}

/**
 * Overridden <code>getYVals()</code>.  Returns stacking values for <i>y</i> values.
 * Values not to be drawn are represented as <code>Double.MAX_VALUE</code> instead
 * of the hole value, allowing percentage values
 * that happen to be equal to the hole value to be drawn.
 * @param series
 * @return
 */
protected double[] getYVals(ChartDataViewSeries series)
{
    return ywork[series.getParent().getSeriesIndex(series)];
}

/**
 * Overridden <code>draw()</code> method.  Calls superclass <code>draw()</code> method
 * with directions to draw in the reverse order so that the polygons will have the correct
 * stacking order.
 * @param endPoint
 * @param forward
 * @param gc
 * @param startPoint
 */
public void draw(Graphics gc, int startPoint, int endPoint, boolean forward) {
	super.draw(gc, startPoint, endPoint, false);
}

/**
 * Overridden <code>pick()</code> method.  Calls superclass <code>pick()</code>
 * method with directions to do the picking in forward order so that the correct
 * layering of polygons is taken into account.
 * @param focus
 * @param pt
 * @return
 */
public JCDataIndex pick(Point pt, int focus) {
    return pick(pt, focus, true);
}

/** <code>true</code> if axes are inverted. */
boolean editInverted = false;

/**
 * Overridden <code>startEdit()</code> method.  Sets <code>editInverted</code> value before calling
 * superclass <code>startEdit()</code> method to do all of the real work.
 * @param di
 */
void startEdit(JCDataIndex di) {
    if (di == null || di.dataView != dataObject) return;
    editInverted = di.dataView.getXAxis().i_vertical;
	super.startEdit(di);
}

/** Overridden <code>endEdit()</code> method.  Un-draws rubber lines and calculates
 *  the correct value to pass back to the edit handler based on the
 *  difference between the new point and the old point added to the
 *  actual value (as opposed to the stacking value) of the old point.
 * @param di
 * @param x
 * @param y
 * @return
 */
int endEdit(JCDataIndex di, int x, int y) {
    if (di == null || di.dataView != dataObject) return -1;
    chartArea.drawRubberLine(prev, last);
    chartArea.drawRubberLine(last, next);

	int point_pos = yaxis.toPixel(di.series.getY(di.point), drawFront);
	int work_pos = yaxis.toPixel(
			ywork[di.series.getDrawingOrder()][di.point-firstLast.x], 
			drawFront);
    int newh = editInverted ? (x - r.x) - work_pos : work_pos - (y - r.y);

    // fake pixel value by pretending the bar starts at the axis
    // (this is converted to an appropriate data value by the edit handler)
    int off = editInverted ? point_pos + newh : 
									point_pos - newh;

    return off;
}

/**
 * Define the shapes associated with areas of the chart for later lookup.
 * @param defShape
 * @param fstyle
 * @param gc
 * @param inverted
 * @param series
 * @param thisAP
 */
protected void defineAreaShapes(DefineShape defShape, Graphics gc,
                JCFillStyle fstyle, boolean inverted,
                AreaPolygon thisAP, ChartDataViewSeries series)
{
	int endPolyIndex = ((thisAP.endPoint - thisAP.startPoint + 1) * 2);
    for (int k = thisAP.startPoint; k < thisAP.endPoint; k++) {

        if (!inverted) {
            int x1 = thisAP.poly.xpoints[k - thisAP.startPoint + 1];
            int y1 = thisAP.poly.ypoints[k - thisAP.startPoint + 1];
            int x2 = thisAP.poly.xpoints[k - thisAP.startPoint + 2];
            int y2 = thisAP.poly.ypoints[k - thisAP.startPoint + 2];
            int yb1 = thisAP.poly.ypoints[
				endPolyIndex - (k - thisAP.startPoint)];
            int yb2 = thisAP.poly.ypoints[
				endPolyIndex - (k - thisAP.startPoint) - 1];

            double rise = (double) (y2 - y1) / (double) (x2 - x1);
            int xmid = x1 + (int) ((double)(x2 - x1) / 2.0);
            int ymid = y1 - (int) (rise * (double) (x1 - xmid));

            double brise = (double) (yb2 - yb1) / (double) (x2 - x1);
            int ybmid = yb1 - (int) (brise * (double) (x1 - xmid));

            Polygon shapePoly = new Polygon();
            shapePoly.addPoint(x1, yb1);
            shapePoly.addPoint(x1, y1);
            shapePoly.addPoint(xmid, ymid);
            shapePoly.addPoint(xmid, ybmid);

            defShape.beginShape(new JCDataIndex(k, series));
            fstyle.fillPolygon(gc, shapePoly);
            defShape.endShape(false);

            shapePoly = new Polygon();
            shapePoly.addPoint(xmid, ybmid);
            shapePoly.addPoint(xmid, ymid);
            shapePoly.addPoint(x2, y2);
            shapePoly.addPoint(x2, yb2);

            defShape.beginShape(new JCDataIndex(k+1, series));
            fstyle.fillPolygon(gc, shapePoly);
            defShape.endShape(false);
        } else {

            int x1 = thisAP.poly.xpoints[k - thisAP.startPoint + 1];
            int y1 = thisAP.poly.ypoints[k - thisAP.startPoint + 1];
            int x2 = thisAP.poly.xpoints[k - thisAP.startPoint + 2];
            int y2 = thisAP.poly.ypoints[k - thisAP.startPoint + 2];
            int xb1 = thisAP.poly.xpoints[
				endPolyIndex - (k - thisAP.startPoint)];
            int xb2 = thisAP.poly.xpoints[
				endPolyIndex - (k - thisAP.startPoint) - 1];

            double rise = (double) (x2 - x1) / (double) (y2 - y1);
            int ymid = y1 + (int) ((double)(y2 - y1) / 2.0);
            int xmid = x1 - (int) (rise * (double) (y1 - ymid));

            double brise = (double) (xb2 - xb1) / (double) (y2 - y1);
            int xbmid = xb1 - (int) (brise * (double) (y1 - ymid));

            Polygon shapePoly = new Polygon();
            shapePoly.addPoint(xb1, y1);
            shapePoly.addPoint(x1, y1);
            shapePoly.addPoint(xmid, ymid);
            shapePoly.addPoint(xbmid, ymid);

            defShape.beginShape(new JCDataIndex(k, series));
            fstyle.fillPolygon(gc, shapePoly);
            defShape.endShape(false);

            shapePoly = new Polygon();
            shapePoly.addPoint(xbmid, ymid);
            shapePoly.addPoint(xmid, ymid);
            shapePoly.addPoint(x2, y2);
            shapePoly.addPoint(xb2, y2);

            defShape.beginShape(new JCDataIndex(k+1, series));
            fstyle.fillPolygon(gc, shapePoly);
            defShape.endShape(false);
        }

    }
}

}
