/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ZoomHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:47 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart;

import java.awt.Point;
import java.awt.Rectangle;

/**
 * Handler object for zoom actions within JCChart.  When
 * an zoom action is started, JCChartArea will create an
 * instance of ZoomHandler and call start(), animate() and
 * end() at appropriate points.  The mapping between key
 * combinations and JCChart events is specified by EventTrigger.
 * @see com.klg.jclass.chart.ActionHandler
 * @see com.klg.jclass.chart.EventTrigger
 */
public class ZoomHandler extends ActionHandler {

/** Point at which the zoom action started */
protected Point first = null;

/** Last point used to draw the bounding box */
protected Point last = null;

/**
 * Default constructor
 */
protected ZoomHandler() {}

/**
 * Called by JCChartArea when the action is started.
 * For the zoom action, this method saves the start point
 * as the anchor for the zoom rectangle.
 * @param x The x coordinate of the start point
 * @param y The y coordinate of the start point
 */
protected void start(int x, int y)
{
    first = new Point(x, y);
    last = null;
}

/**
 * Called by JCChartArea when the action is to be
 * animated.  For the zoom action, thid method
 * undraws the previous rubber-band rectangle and
 * draws a new rubber-band rectangle from the
 * start point to the current point.  The rectangle
 * represents the zoom area.
 * @param x The x coordinate of the animate point
 * @param y The y coordinate of the animate point
 */
protected void animate(int x, int y)
{
    if (first == null) return;
    if (last == null) {
        last = new Point(x, y);
        parent.drawRubberBand(last, first);
        return;
    }
    parent.drawRubberBand(last, first);
    last.x = x;
    last.y = y;
    parent.drawRubberBand(last, first);
}

/**
 * Called by JCChartArea when the chart area has been
 * refreshed and the animation must be redrawn.
 * @param x The x coordinate of the reanimate point
 * @param y The y coordinate of the reanimate point
 */
protected void reanimate(int x, int y)
{
    last = new Point(x, y);
    parent.drawRubberBand(last, first);
}

/**
 * Called by JCChartArea when the action is over.
 * For the zoom action, this method takes the start point
 * and the current point and uses them to zoom along the
 * horizontal and vertical action axes.
 * @param x The x coordinate of the end point
 * @param y The y coordinate of the end point
 */
protected void end(int x, int y)
{
    /** Drawing area of parent JCChartArea */
    Rectangle r = null;
    JCAxis xAxis = null;
    JCAxis yAxis = null;
    boolean changed = false;

    parent.drawRubberBand(last, first);
    r = parent.getDrawingArea();

    JCAxis horizActionAxis = parent.getHorizActionAxis();
    JCAxis vertActionAxis = parent.getVertActionAxis();
    if (horizActionAxis.isVertical() && !vertActionAxis.isVertical()) {
        // Inverted axes
        xAxis = vertActionAxis;
        yAxis = horizActionAxis;
    }
    else if (!horizActionAxis.isVertical() && vertActionAxis.isVertical()) {
        // Normal axes
        xAxis = horizActionAxis;
        yAxis = vertActionAxis;
    }
    else {
        // Error - both axes are going in the same direction!
        throw new IllegalStateException("ZoomHandler: Weird axes!");
    }

    int clickThreshold = 7;
    if ((Math.abs(x - first.x) + Math.abs(y - first.y)) < clickThreshold) {
        // This is a click on a target area.
        // Default the zoom region to a small area around the mouse click
        int zoomRadius = 10;
        first.x = Math.max(x - zoomRadius, r.x);
        x = Math.min(x + zoomRadius, r.x + r.width);
        first.y = Math.max(y - zoomRadius, r.y);
        y = Math.min(y + zoomRadius, r.y + r.height);
    }
    JCChart chart = getChart();
    if (xAxis != null && xAxis.editable && !xAxis.isPolarRadar()) {
        double startVal = xAxis.toData(first.x-r.x);
        double currVal = xAxis.toData(x-r.x);;
        if (chart != null) {
            chart.zoom(startVal, currVal, xAxis, false);
            changed = true;
        }
    }
    if (yAxis != null && yAxis.editable && !yAxis.isPolarRadar()) {
        double startVal = yAxis.toData(first.y-r.y);
        double currVal = yAxis.toData(y-r.y);
        if (chart != null) {
            chart.zoom(startVal, currVal, yAxis, false);
            changed = true;
        }
    }

    if (changed) {
        parent.setChanged(true, Changeable.RECALC);
    }
}

}
