/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:51 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;

import com.klg.jclass.chart.*;
import com.klg.jclass.util.ImageMapInfo;

import java.io.Serializable;

/**
 * This class is the base for any data source that chooses to
 * store data internally using arrays.
 */
public class BaseDataSource extends ChartDataSupport
			                implements ChartDataModel, LabelledChartDataModel,
									   HoleValueChartDataModel,
	                                   ChartDataManageable, Serializable,
                                       ImageMapChartDataModel
{

protected double          xvalues[][] = null;
protected double          yvalues[][] = null;
protected String          name = null;
protected String          seriesLabels[] = null;
protected String          pointLabels[] = null;
protected double          holeValue = Double.MAX_VALUE;

protected ImageMapInfo    pointImageMapInfo[][] = null;
protected ImageMapInfo    clusterImageMapInfo[] = null;
protected ImageMapInfo    seriesImageMapInfo[] = null;
protected ImageMapInfo    legendImageMapInfo[] = null;

protected boolean         cloneArrays = false;

/**
 * Retrieves the specified x-value series.
 * @param index data series index
 * @return array of double values representing x-value data
 */
public double[] getXSeries(int index)
{
	if (xvalues == null) {
		return(null);
	}
	if (xvalues.length <= 0) {
		return(null);
	}
	if (index >= xvalues.length) {
		return cloneCheck(xvalues[0]);
	}
	return(cloneCheck(xvalues[index]));
}

/**
 * Retrieves the specified y-value series.
 * @param index data series index
 * @return array of double values representing x-value data
 */
public double[] getYSeries(int index)
{
	if (yvalues == null || index >= yvalues.length) {
		return(null);
	}
	return(cloneCheck(yvalues[index]));
}

/**
 * Retrieves the number of data series.
 * @return The number of data series
 */
public int getNumSeries()
{
	if (yvalues == null) {
		return(0);
	}
	return(yvalues.length);
}

/*
 * Retrieves the labels to be used for each point in a
 * particular data series.
 * @return An array of point labels
 */
public String[] getPointLabels()
{
	return(cloneCheck(pointLabels));
}

/**
 * Retrieves the labels to be used for each data series.
 * @return An array of series labels
 */
public String[] getSeriesLabels()
{
	return(cloneCheck(seriesLabels));
}

/**
 * Retrieves the name for the data source.
 * @return The current name for the data source
 */
public String getDataSourceName()
{
	return(name);
}

/**
 * Retrieves the hole value for the data source.
 * @return The current hole value for the data source
 */
public double getHoleValue()
{
	return(holeValue);
}

/**
 * Retrieves a array of image map information indexed by series.
 * @param seriesIndex The series whose image map informaiton is required.
 * @return The current point image map information list for the given series.
 */
public ImageMapInfo[] getPointImageMapInfo(int seriesIndex)
{
    if (pointImageMapInfo == null || seriesIndex >= pointImageMapInfo.length) {
        return(null);
    }
    return(cloneCheck(pointImageMapInfo[seriesIndex]));
}

/**
 * Retrieves an array of image map information, one for each point cluster.
 * @return The current cluster image map information array.
 */
public ImageMapInfo[] getClusterImageMapInfo()
{
    return(cloneCheck(clusterImageMapInfo));
}

/**
 * Retrieves an array of image map information, one for each series.
 * @return The current series image map information array.
 */
public ImageMapInfo[] getSeriesImageMapInfo()
{
    return(cloneCheck(seriesImageMapInfo));
}

/**
 * Retrieves an array of legend image map information, one for each series.
 * @return The current legend image map information array.
 */
public ImageMapInfo[] getLegendImageMapInfo()
{
    return(cloneCheck(legendImageMapInfo));
}

/**
 * Retrieves the ChartDataManager instance.  Required by the
 * ChartDataManageable interface.
 * @return The ChartDataManager instance
 */
public ChartDataManager getChartDataManager()
{
    return(this);
}

/**
 * Sets the value of the CloneArrays property.
 * The CloneArrays property determines whether the
 * data source makes a copy of x and y series arrays
 * before sending them to the chart.
 * @param ca If true, make a copy of arrays
 */
public void setCloneArrays(boolean ca)
{
	cloneArrays = ca;
}

/**
 * Gets the value of the CloneArrays property.
 * The CloneArrays property determines whether the
 * data source makes a copy of x and y series arrays
 * before sending them to the chart.
 * @return The current value of the CloneArrays property
 */
public boolean getCloneArrays()
{
	return(cloneArrays);
}

/**
 * Convenience method used to optionally clone
 * a double array.  Override this method if you want
 * to change the way array copies are made.
 * @param a The array to clone
 * @return The array or its clone depending of the value of the CloneArrays
 *     property
 */
protected double[] cloneCheck(double[] a)
{
	if (cloneArrays) {
		return((double[])a.clone());
	}
	return(a);
}

/**
 * Convenience method used to optionally clone
 * a String array.  Override this method if you want
 * to change the way array copies are made.
 * @param a The array to clone
 * @return The array or its clone depending of the value of the CloneArrays
 *     property
 */
protected String[] cloneCheck(String[] a)
{
	if (cloneArrays && a != null) {
		return((String[])a.clone());
	}
	return(a);
}

/**
 * Convenience method used to optionally clone
 * an ImageMapInfo array.  Override this method if you want
 * to change the way array copies are made.
 * @param a The array to clone
 * @return The array or its clone depending of the value of the CloneArrays
 *     property
 */
protected ImageMapInfo[] cloneCheck(ImageMapInfo[] a)
{
    if (cloneArrays && a != null) {
        return((ImageMapInfo[])a.clone());
    }
    return(a);
}

}

