/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ChartDataSourceUtil.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:51 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;

import java.io.*;

import com.klg.jclass.chart.ChartDataModel;
import com.klg.jclass.chart.LabelledChartDataModel;
import com.klg.jclass.chart.HoleValueChartDataModel;
import com.klg.jclass.chart.JCChartBundle;

/**
 * ChartDataSourceUtil hosts a collection of useful
 * data source related methods.
 */
public class ChartDataSourceUtil {

private static int lowCharacterBound  = 32;
private static int highCharacterBound = 126;

private static final String QUOTE = "'";
private static final char   DOUBLE_QUOTE = '\"';
private static final String NL = "\n";
private static final String BLANK = " ";
private static final String COMMENT = "#";
private static final String commentLine1 =
    COMMENT + BLANK + JCChartBundle.string(JCChartBundle.key167) + NL;
private static final String commentLine2 =
    COMMENT + BLANK + JCChartBundle.string(JCChartBundle.key168) + NL;
private static final String commentLine3 =
    COMMENT + BLANK + JCChartBundle.string(JCChartBundle.key169) + NL;
private static final String commentLine4 =
    COMMENT + BLANK + JCChartBundle.string(JCChartBundle.key170) + NL;

/**
 * Set the lower and upper character bounds as integers.  By default this
 * range is set to be low=32 and high=126 which is the normal ascii printable
 * range.
 * @param lower The new lower bound
 * @param higher The new higher bound
 */
public static void setCharacterBounds(int lower, int higher)
{
    if (lower > higher || lower < 0) {
        throw new IllegalArgumentException("Illegal bounds - lower:" + lower +
                                           " higher:" + higher);
    }
    lowCharacterBound  = lower;
    highCharacterBound = higher;
}

/**
 * Get the low character bound.
 * @return The current low character bound
 */
public static int getLowCharacterBound()
{
    return(lowCharacterBound);
}

/**
 * Get the high character bound.
 * @return The current high character bound
 */
public static int getHighCharacterBound()
{
    return(highCharacterBound);
}

/**
 * Generates a single series of x data given the supplied array length.
 * @param length The length of the x series to generate
 * @return A single X series of the given length in a doubly subscripted array
 */
public static double[][] generateSingleXSeries(int length)
{
    // Create single x series
    double[][] x = new double[1][length];

    // Supply default values
    for (int i = 0; i < x[0].length; i++) {
        x[0][i] = i;
    }
    return(x);
}

/**
 * Generates multiple series of x data using the supplied y data.
 * @param y The y data
 * @return A multilple series of x data that mirrors the inputted y data
 */
public static double[][] generateMultipleXSeries(double[][] y)
{
    if (y == null || y.length < 0) {
        return(null);
    }

    // Create matching x series
    double[][] x = new double[y.length][];

    // Supply default values
    for (int i = 0; i < x.length; i++) {
        // Create x series for y series at series index i
        x[i] = new double[y[i].length];
        // Fill x series with default data
        for (int j = 0; j < y[i].length; j++) {
            x[i][j] = j;
        }
    }
    return(x);
}

/**
 * Convert this double to a string, optionally taking into account the hole
 * value.  Use "max"/"min" to write out Double.MAX_VALUE/Double.MIN_VALUE.
 * @param d The double to write out
 * @param hole The hole value
 * @param doHoles If true, write "hole" instead of the actual hole value
 * @return The double as a string
 */
protected static String doubleToString(double d, double hole, boolean doHoles)
{
    String doubleString = null;
    if (doHoles && d == hole) {
        doubleString = "hole";
    }
    else if (d == Double.MAX_VALUE) {
        doubleString = "max";
    }
    else if (d == Double.MIN_VALUE) {
        doubleString = "min";
    }
    else {
        doubleString = String.valueOf(d);
    }
    return(doubleString);
}

/**
 * Converts the provided data source into a string.  This string is compatible
 * with the format used to read in data from streams (file, URL, applet).
 *
 * @param src The data source
 * @return A string version of the data source
 */
public static String buildDataSourceString(ChartDataModel src)
{
    return(buildDataSourceString(src, true));
}

/**
 * Converts the provided data source into a string.  This string is compatible
 * with the format used to read in data from streams (file, URL, applet).
 *
 * @param src The data source
 * @param doExpand Replace characters outside of the character bounds range with
 *     html character entities
 * @see #setCharacterBounds
 * @return A string version of the data source
 */
public static String buildDataSourceString(ChartDataModel src, boolean doExpand)
{
    if (src == null || src.getNumSeries() <= 0) {
        return("");
    }
    int numSeries = src.getNumSeries();

    String[] pointLabels = null;
    String[] seriesLabels = null;
    String name = null;
    if (src instanceof LabelledChartDataModel) {
        LabelledChartDataModel lsrc = (LabelledChartDataModel)src;
        pointLabels = lsrc.getPointLabels();
        seriesLabels = lsrc.getSeriesLabels();
        name = lsrc.getDataSourceName();
    }
    boolean cvtHoles = false;
    double holeValue = Double.MAX_VALUE;
    // Convert hole values to word 'hole' if we can find out what the
    // hole value is
    if (src instanceof HoleValueChartDataModel) {
        holeValue = ((HoleValueChartDataModel)src).getHoleValue();
        cvtHoles = true;
    }

    double[][] xvalues = new double[numSeries][];
    double[][] yvalues = new double[numSeries][];
    for (int i = 0; i < numSeries; i++) {
        // Deal with x
        double[] temp = src.getXSeries(i);
        if (i == 0 || temp != xvalues[0]) {
            xvalues[i] = temp;
        }
        // Deal with y
        yvalues[i] = src.getYSeries(i);
    }
    if (xvalues.length > 1 && xvalues[1] == null) {
        double[] temp = xvalues[0];
        xvalues = new double[1][];
        xvalues[0] = temp;
    }

    StringBuffer s = new StringBuffer();
    if (xvalues.length == 1) {
        // Data format - array
        int rows = yvalues.length;
        int columns = yvalues[0].length;

        s.append(commentLine1);
        s.append(commentLine2);
        s.append(commentLine3);
        s.append(commentLine4);
        s.append("ARRAY "); // Setting the data format
        if (name != null ) {
            // Name of the data object
            String n = doExpand ? expandText(name) : name;
            s.append(QUOTE + n + QUOTE + BLANK);
        }
        else {
            s.append(QUOTE + QUOTE + BLANK);
        }
        // Rows and colums
        s.append(rows + BLANK + columns + BLANK + NL);

        if (cvtHoles && holeValue != Double.MAX_VALUE) {
            s.append("HOLE " + doubleToString(holeValue, holeValue, false) +
                     NL);
        }

        // Write out point labels
        writePointLabels(pointLabels, s, false, doExpand);

        // Spit out the x values
        for (int k = 0; k < columns; k++) {
            s.append(doubleToString(xvalues[0][k], holeValue, cvtHoles) +
                     BLANK);
        }
        s.append(NL);

        for (int i = 0; i < rows; i++ ) {
            // Series labels
            if (seriesLabels != null    &&
                i < seriesLabels.length &&
                seriesLabels[i] != null)
            {
                String label = doExpand ? expandText(seriesLabels[i])
                                        : seriesLabels[i];
                s.append(QUOTE + label + QUOTE + BLANK);
            }
            else {
                s.append(QUOTE + QUOTE  + BLANK);
            }
            // Y values
            for (int j = 0; j < columns; j++ ) {
                s.append(doubleToString(yvalues[i][j], holeValue, cvtHoles) +
                         BLANK);
            }
            s.append(NL);
        }
    }
    else {
        // Data format - general
        s.append( "GENERAL " );

        // Name of data source
        if (name != null ) {
            //name of the data object
            s.append(QUOTE + name + QUOTE + BLANK);
        }
        else {
            s.append(QUOTE + QUOTE + BLANK);
        }

        // Number of series
        s.append(yvalues.length + BLANK);

        // Calculate maximum # of elements per series
        int max = 0;
        for (int i = 0; i < yvalues.length; i++) {
            if (yvalues[i] != null && yvalues[i].length > max) {
                max = yvalues[i].length;
            }
        }

        // Write out max # of elements per series
        s.append(max + NL);

        if (cvtHoles && holeValue != Double.MAX_VALUE) {
            s.append("HOLE " + doubleToString(holeValue, holeValue, false) +
                     NL);
        }

        // Write out point labels
        writePointLabels(pointLabels, s, true, doExpand);

        // Write out the data
        for (int sr = 0; sr < yvalues.length; sr++) {
            // Series label
            if (seriesLabels != null     &&
                sr < seriesLabels.length &&
                seriesLabels[sr] != null)
            {
                String label = doExpand ? expandText(seriesLabels[sr])
                                        : seriesLabels[sr];
                s.append(QUOTE + label + QUOTE + BLANK);
            }
            else {
                s.append(QUOTE + QUOTE  + BLANK);
            }

            // Handle zero case.
            if (yvalues[sr] == null || yvalues[sr].length <= 0) {
                s.append(0 + NL);
                continue;
            }

            // Number of elements in series
            s.append(yvalues[sr].length + NL);

            // Now output the series itself
            StringBuffer x = new StringBuffer();
            StringBuffer y = new StringBuffer();
            for (int i = 0; i < yvalues[sr].length; i++) {
                double[] xVals = xvalues[sr];
                double xValue =
                    (xVals == null || i >= xVals.length) ? i : xVals[i];
                x.append(doubleToString(xValue, holeValue, cvtHoles) +
                         BLANK);
                y.append(doubleToString(yvalues[sr][i], holeValue, cvtHoles) +
                         BLANK);
            }
            s.append(x + NL);
            s.append(y + NL);
        }
    }
    return(s.toString());
}

/**
 * Check if there are any non-empty point labels.
 * @param pointLabels The point label list to check
 * @return true if there is at least one non-empty point label
 */
public static boolean checkPointLabels(String[] pointLabels)
{
    boolean pointLabelFound = false;
    if (pointLabels != null) {
        for (int i = 0; i < pointLabels.length; i++) {
            if (pointLabels[i] != null && pointLabels[i].length() > 0) {
                pointLabelFound = true;
                break;
            }
        }
    }
    return(pointLabelFound);
}

/**
 * Write point labels to the given string buffer in a format that can be read
 * in by JClass chart.
 * @param pointLabels The point labels as a list of strings
 * @param s The string buffer to write to.
 * @param writeLabelString If true, write a special string before writing
 *     the point labels.
 * @param doExpand Replace characters outside of the character bounds range with
 *     html character entities
 * @see #setCharacterBounds
 */
private static void writePointLabels(String[] pointLabels, StringBuffer s,
                                     boolean writeLabelString,
                                     boolean doExpand)
{
    if (checkPointLabels(pointLabels)) {
        if (writeLabelString) {
            s.append("LABELS ");
        }
        for (int i = 0; i < pointLabels.length; i++ ) {
            if (pointLabels[i] != null ) {
                // Point labels
                String label = doExpand ? expandText(pointLabels[i])
                                        : pointLabels[i];
                s.append(QUOTE + label + QUOTE + BLANK);
            }
            else {
                s.append(QUOTE + QUOTE  + BLANK);
            }
        }
        s.append(NL);
    }
}

/**
 * Writes the provided data source to a file.
 * @param src The data source
 * @param fileName The file name to write to.
 * @param doExpand Replace characters outside of the character bounds range with
 *     html character entities
 * @see #setCharacterBounds
 * @throws IOException
 */
public static void writeDataSource(ChartDataModel src, String fileName,
                                   boolean doExpand)
    throws IOException
{
    if (src == null || src.getNumSeries() <= 0 || fileName == null) {
        return;
    }

    FileOutputStream fos = new FileOutputStream(fileName);
    writeDataSource(src, fos, doExpand);
    fos.close();
}

/**
 * Writes the provided data source to the given output stream.
 * @param src The data source
 * @param os The output stream to write to.
 * @param doExpand Replace characters outside of the character bounds range with
 *     html character entities
 * @see #setCharacterBounds
 * @throws IOException
 */
public static void writeDataSource(ChartDataModel src, OutputStream os,
                                   boolean doExpand)
    throws IOException
{
    if (src == null || src.getNumSeries() <= 0 || os == null) {
        return;
    }

    DataOutputStream out = null;
    if (os instanceof DataOutputStream) {
        out = (DataOutputStream)os;
    }
    else {
        out = new DataOutputStream(os);
    }
    if (out == null) {
        return;
    }

    String s = buildDataSourceString(src, doExpand);
    if (s != null) {
        out.write(s.getBytes());
    }
}

/**
 * Writes the provided data source using the given OutputStreamWriter.
 * @param src The data source
 * @param writer The OutputStreamWriter with which to write
 * @param doExpand Replace characters outside of the character bounds range with
 *     html character entities
 * @see #setCharacterBounds
 * @throws IOException
 */
public static void writeDataSource(ChartDataModel src,
                                   OutputStreamWriter writer, boolean doExpand)
    throws IOException
{
    if (src == null || src.getNumSeries() <= 0 || writer == null) {
        return;
    }

    String s = buildDataSourceString(src, doExpand);
    if (s != null) {
        writer.write(s);
        writer.flush();
    }
}

/**
 * Expand text string to escape unprintable characters with character entities.
 * @param s The string to expand
 * @return The expanded string
 */
public static String expandText(String s)
{
    if (s == null) {
        return(null);
    }

    StringBuffer sb = new StringBuffer(s);
    // Replace unprintable characters with character entities (i.e. with
    // strings of the form "&#ddd;" where d's are digits)
    for (int i = sb.length() - 1; i >= 0; i--) {
        char ch = sb.charAt(i);

        // Escape double-quote
        if (ch == DOUBLE_QUOTE) {
            sb.replace(i, i + 1, "\\\"");
            continue;
        }

        int c = (int)ch;
        // Don't expand characters inside character bound range
        if (c >= lowCharacterBound && c <= highCharacterBound) {
            continue;
        }

        String intString = Integer.toString(c);
        StringBuffer sbNew = new StringBuffer("&#");
        sbNew.append(intString);
        sbNew.append(";");
        sb.replace(i, i + 1, sbNew.toString());
    }
    return(sb.toString());
}

/**
 * Remove all html character entities from the string and replace with
 * the actual unicode character.
 * @param s The input string
 * @return The processed string
 */
public static String processText(String s)
{
    if (s == null) {
        return(null);
    }

    // Look for substrings of the form "&#ddd;" where d's are digits.  There
    // can be one or more d's.
    StringBuffer sb = new StringBuffer(s);
    for (int i = 0; i < sb.length() - 1; i++) {
        if (sb.charAt(i) == '&' && sb.charAt(i+1) == '#') {
            boolean found = false;
            int j = i + 2;
            StringBuffer sbDigit = new StringBuffer();
            while (sb.length() > j) {
                char ch = sb.charAt(j);
                if (ch == ';') {
                    // Only declare found if at least one digit
                    if (sbDigit.length() > 0) {
                        found = true;
                    }
                    break;
                }
                else if (Character.isDigit(ch)) {
                    sbDigit.append(ch);
                }
                else {
                    // Not a digit or semi-colon.  Bad format.  Do not replace.
                    break;
                }
                j++;
            }
            if (found) {
                String digit = sb.substring(i + 2, j);
                int c = Integer.valueOf(digit).intValue();
                sb.delete(i + 1, j + 1);
                sb.setCharAt(i, (char)c);
            }
        }
    }
    return(sb.toString());
}

}
