/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartSwingDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:51 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;

import com.klg.jclass.chart.*;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;
import java.io.Serializable;

/**
 * <code>JCChartSwingDataSource</code> provides an intepretive layer between
 * <code>JCChart</code> and data existing in a Swing <code>TableModel</code>
 * format.<p>
 * The most notable differences between them is the inversion of rows and
 * columns - <code>TableModel</code>'s rows are <code>JCChart</code>'s columns,
 * and vice versa.  Also, <code>TableModel</code> provides no X values, and so
 * these are generated here.
 */

public class JCChartSwingDataSource extends ChartDataSupport 
                                    implements EditableChartDataModel,
                                               LabelledChartDataModel,
                                               HoleValueChartDataModel,
                                               TableModelListener,
                                               ChartDataManageable,
                                               Serializable
{

/**
 * TableModel instance (Swing data source instance).
 */
protected TableModel  tableModel = null;
protected String[]    pLabels = null;
protected String      dsName = "Legend";

/** 
 * The default constructor.  Makes an empty data source.
 */
public JCChartSwingDataSource() {}

/**
 * Constructor. Sets the table model to passed object.
 * @param model The table model on which to base this datasource
 */
public JCChartSwingDataSource(TableModel model)
{
    tableModel = model;
    tableModel.addTableModelListener(this);
}

/**
 * Constructor. Sets the table model to passed object, values for
 * point labels, and a data source name.
 * @param model The table model on which to base this datasource
 * @param pointLabels A list of point labels for this data source
 * @param dataSourceName The name of this data source
 */
public JCChartSwingDataSource(TableModel model, String[] pointLabels,
                              String dataSourceName)
{
    tableModel = model;
    tableModel.addTableModelListener(this);
    pLabels = pointLabels;
    dsName = dataSourceName;
}

/**
 * Sets the value of a single data item in the <code>TableModel</code>.
 * @param newValue The new data item value
 * @param seriesIndex The series (or column) to change
 * @param pointIndex The point (or row) to change
 * @return Whether change succeeded
 */
public synchronized boolean setDataItem(int seriesIndex, int pointIndex,
                                        double newValue)
{
    if (tableModel == null || seriesIndex < 0) {
        return(false);
    }

    tableModel.setValueAt(new Double(newValue), pointIndex, seriesIndex);
    return(true);
}

/**
 * Retrieves the specified x-value series.
 * @param index data series index
 * @return array of double values representing x-value data
 */
public double[] getXSeries(int index)
{
    if (tableModel == null) {
        return(null);
    }

    // X data is manufactured in the Swing data model
    int numPoints = tableModel.getRowCount();
    return ChartDataSourceUtil.generateSingleXSeries(numPoints)[0];
}

/**
 * Retrieves the specified y-value series.
 * @param index data series index
 * @return array of double values representing x-value data
 */
public double[] getYSeries(int index)
{
    if (tableModel == null) {
        return(null);
    }

    int numPoints = tableModel.getRowCount();
    if (numPoints == 0) {
        return(null);
    }

    // Generate y values from TableModel
    double[] y = new double[numPoints];
    for (int i = 0; i < numPoints; i++) {
        Object val = tableModel.getValueAt(i, index);
        if (val == null) {
            // If data is null, use HoleValue
            y[i] = getHoleValue();
        }
        else if (val instanceof Number) {
            // Easy - get the number as a double and put it in the array
            y[i] = ((Number)val).doubleValue();
        }
        else {
            try {
                if (val instanceof String) {
                    // Try to parse the string
                    y[i] = Double.parseDouble((String)val);
                }
                else {
                    // Desperation - convert to string and parse.
                    y[i] = Double.parseDouble(val.toString());
                }
            }
            catch (Exception e) {
                // failed, use Zero, because that's what Cafe does
                y[i] = 0.0;
            }
        }
    }
    return(y);
}

/**
 * Retrieves the number of data series.
 * @return Number of data series
 */
public int getNumSeries()
{
    if (tableModel == null) {
        return(0);
    }
    return(tableModel.getColumnCount());
}


/**
 * Point labels can be set manually if desired.  There are none in the
 * <code>JTable</code> itself.
 * @see com.klg.jclass.chart.LabelledChartDataModel#getPointLabels
 * @return The current list of point labels
 */
public String[] getPointLabels()
{
    return(pLabels);
}

/**
 * Point labels can be set manually if desired.  There are none in the
 * JTable itself.
 * @see com.klg.jclass.chart.LabelledChartDataModel#getPointLabels
 * @param labels A list of point labels
 */
public void setPointLabels(String[] labels)
{
    pLabels = labels;
}

/**
 * Gets series labels from <code>TabelModel</code>'s column headers.
 * @see com.klg.jclass.chart.LabelledChartDataModel#getSeriesLabels
 * @return A list of series labels gleaned from the column headers
 */
public String[] getSeriesLabels()
{
    if (tableModel == null) {
        return null;
    }

    int numSeries = tableModel.getColumnCount();
    if (numSeries == 0) return null;
    String[] seriesLabels = new String[numSeries];
    for (int i = 0; i < seriesLabels.length; i++) {
        String tmpString = tableModel.getColumnName(i);
        if (tmpString.trim().length() == 0) {
            seriesLabels[i] = null;
        }
        seriesLabels[i] = tmpString;
    }
    return(seriesLabels);
}

/**
 * Name of data source, used at the top of the legend. Can be set manually. It
 * is not found in the <code>JTable</code> itself.
 * @return The current data source name
 */
public String getDataSourceName()
{
    return(dsName);
}

/**
 * Name of data source, used at the top of the legend. Can be set manually. It
 * is not found in the <code>JTable</code> itself.
 * @param name The new data source name
 */
public void setDataSourceName(String name)
{
    dsName = name;
}

/**
 * Returns the hole value being used for this data source.
 * @return The current hole value
 */
public double getHoleValue()
{
    return(Double.MAX_VALUE);
}

/**
 * Implements <code>TableModelListener</code> for changes to
 * <code>TableModel</code>.
 * @param e The event generated by the change
 */
public void tableChanged(TableModelEvent e)
{
    int firstRow = e.getFirstRow();
    // new header row means a whole new table!
    if (firstRow == TableModelEvent.HEADER_ROW) {
        fireChartDataEvent(ChartDataEvent.RESET, 0, 0);
        return;
    }

    int type = e.getType();
    int column = e.getColumn();
    boolean allRows = false;
    if (e.getLastRow() == Integer.MAX_VALUE) {
        allRows = true;
    }

    if (type == TableModelEvent.UPDATE) {
        // all table columns updated -- means one or more rows of table data
        // were updated
        if (column == TableModelEvent.ALL_COLUMNS) {
            // Used to modify individual columns (points)-don't bother anymore
            fireChartDataEvent(ChartDataEvent.RESET, 0, 0);
        }
        else {
            // allRows? update entire column (our series)!
            if (allRows) {
                fireChartDataEvent(ChartDataEvent.RELOAD_SERIES, column, 0);
            }
            else {
                // update just this row, column value
                // I have to use RELOAD_SERIES instead of RELOAD_POINT
                // because RELOAD_POINT relies on changes having been made
                // to the shared value array.  RELOAD_SERIES forces a
                // requery of the entire array.  Since this data source
                // must build the series array by querying the source
                // TableModel, that's the way to go.
                fireChartDataEvent(ChartDataEvent.RELOAD_SERIES, column, 0);
            }
        }
    }
    else if (type == TableModelEvent.INSERT) {
        // reset entire data source from table so that 1) x values are
        // properly recalced and 2) we don't have to worry about matching
        // the correct number of rows and columns.
        fireChartDataEvent(ChartDataEvent.RESET, 0, 0);
    }
    else if (type == TableModelEvent.DELETE) {
        // must reset entire data source from table so that x values
        // are properly recalced
        fireChartDataEvent(ChartDataEvent.RESET, 0, 0);
    }
}

/**
 * Retrieves the <code>ChartDataManager</code> instance.  Required
 * by the <code>ChartDataManageable</code> interface.
 * @return
 */
public ChartDataManager getChartDataManager()
{
    return(this);
}

}
