/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCDefaultDataInterpreter.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:51 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;



import java.io.IOException;
import java.io.Reader;
import java.io.StreamTokenizer;

/**
 * Implements a interpreter for the default chart data format.  Relies on input
 * from some reader to populate the specified <code>BaseDataSource</code> class.
 * <p>
 * <H2>Format Information</H2><br>
 * <b>Array Format Examples:</b>
 * In array format, there is one set of
 * x-axis data, multiple sets of y-axis data.
 * The data may or may not be transposed, for example the two following
 * data files are equivalent:
 *
 * <PRE>
 * ARRAY 2 4  # 2 series (+x-data) w. 4 points each
 * 1 2 3 4    # x-data
 * 1 2 3 4    # y1
 * 1 4 9 16   # y2
 * </PRE>
 *
 * and
 *
 * <PRE>
 * ARRAY 2 4 T
 * 1 1 1
 * 2 2 4
 * 3 3 9
 * 4 4 16
 * </PRE>
 *
 * Or, with names and point labels,
 * <PRE>
 * ARRAY 'My Chart' 2 4
 *                 'Point Label 1' 'Point Label 2' 'Point Label 3' 'Point Label 4'
 * 'X Axis Data'    1               2               3               4
 * 'Y Axis #1 Data' 1               2               3               4
 * 'Y Axis #2 Data' 1               4               9               16
 * </PRE>
 *
 * or
 *
 * <PRE>
 * ARRAY 'MyChart' 2 4 T
 *                'X Data' 'Y Data' 'Y 2 Data'
 * 'Point Label 1' 1        1        1
 * 'Point Label 2' 2        2        4
 * 'Point Label 3' 3        3        9
 * 'Point Label 4' 4        4        16
 * </PRE>
 *
 * <p>
 *
 * <b>General format examples:</b>
 *  In general format, there are multiple x series.<br>
 * <PRE>
 * # 3 lines, max 5 points each
 * GENERAL 3 5
 * # line 1, 3 points
 * 3
 * 1 3 5    # x-values
 * 5 14 24  # y-values
 * # line 2, 5 points
 * 5
 * 1 2 3 4 5
 * 11 15 18 21 22
 * # line 3, 2 points
 * 2
 * 2 4
 * 16 15.2
 * </PRE>
 *
 * or, using the transposed format
 *
 * <PRE>
 * GENERAL 3 5 T
 * # line 1, 3 points
 * 3
 * # X  Y
 * 1   5
 * 3  14
 * 5  24
 * # line 2, 5 points
 * 5
 * # X  Y
 * 1   11
 * 2   15
 * 3   18
 * 4   21
 * 5   22
 * # line 3, 2 points
 * 2
 * # X  Y
 * 2   16
 * 4   15.2
 * </PRE>
 *
 * Or, with point labels and series names
 *
 * <PRE>
 * GENERAL 3 5
 * LABELS 'Mercury' 'Venus' 'Earth' 'Mars' 'Jupiter'
 * 'Series 1' 3
 * 1 3 5
 * 5 14 24
 * 'Series 2' 5
 * 1 2 3 4 5
 * 11 15 18 21 22
 * 'Series 3' 2
 * 2 4
 * 16 15.2
 * </PRE>
 *
 * <PRE>
 * GENERAL 3 5 T
 * LABELS 'Mercury' 'Venus' 'Earth' 'Mars' 'Jupiter'
 * # line 1, 3 points
 * 'Series 1' 3
 * # X  Y
 * 1   5
 * 3  14
 * 5  24
 * # line 2, 5 points
 * 'Series 2' 5
 * # X  Y
 * 1   11
 * 2   15
 * 3   18
 * 4   21
 * 5   22
 * # line 3, 2 points
 * 'Series 3' 2
 * # X  Y
 * 2   16
 * 4   15.2
 * </PRE>
 */
public class JCDefaultDataInterpreter extends JCDataInterpreter
									  implements java.io.Serializable {

private StreamTokenizer tokenizer;

private final char QUOTE = '\'';

/**
 * This constructor creates an instance of a data interpreted and associates it
 * with a data source. It does not yet do any interpreting.
 * @param bds the data source to eventually be populated with the interpreted data
 */
public JCDefaultDataInterpreter(BaseDataSource bds)
{
	super(bds);
}

/**
 * Default no-argument constructor.
 */
public JCDefaultDataInterpreter() {}

/**
 * Parses the specified input stream according to the default chart
 * data format and populates the previously set data source. Called
 * by the other <code>ctor()</code> method, and also used for String
 * input where you can't have an input stream.
 * @param r the Reader object used to read data
 * @throws IOException
 */
protected void ctor(Reader r) throws IOException
{
	tokenizer = new StreamTokenizer(r);
	tokenizer.commentChar('#');
	tokenizer.quoteChar(QUOTE);
	tokenizer.slashSlashComments(false);
	tokenizer.slashStarComments(false);
	tokenizer.eolIsSignificant(false);
	tokenizer.lowerCaseMode(true);
	tokenizer.parseNumbers();
	tokenizer.whitespaceChars('+', '+');
	init();
	tokenizer = null; // let go of it so it can be collected
}

/**
 * Private utility function that retreives the next double value
 * from the stream.
 * @return
 * @throws IOException
 */
private double getNextDouble() throws IOException
{
	tokenizer.nextToken();

	if (tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
		if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
			String s = tokenizer.sval;
            Double d = null;
			if (s.equalsIgnoreCase("HOLE")) {
			    d = new Double(baseDS.holeValue);
			}
            else if (s.equalsIgnoreCase("MAX")) {
                d = new Double(Double.MAX_VALUE);
            }
            else if (s.equalsIgnoreCase("MIN")) {
                d = new Double(Double.MIN_VALUE);
            }
            else if (s.equalsIgnoreCase("INFINITY")) {
                d = new Double(Double.POSITIVE_INFINITY);
            }
            else {
                try {
                    d = Double.valueOf(s);
                }
                catch (Exception e) {}
            }
            if (d != null) {
                return(d.doubleValue());
            }
		}
		displayErrorMessage(PARSE_ERR_HOLE_OR_VALUE);
		throw new IOException(ERRMSG_FILE_FORMAT + tokenizer);
	}

	double rval = tokenizer.nval;
	tokenizer.nextToken();
	if (tokenizer.ttype == StreamTokenizer.TT_WORD &&
		tokenizer.sval != null &&
		(tokenizer.sval.charAt(0) == 'e' ||
		 tokenizer.sval.charAt(0) == 'E'))
	{
		if (tokenizer.sval.length() == 1) {
			tokenizer.nextToken();
			if (tokenizer.ttype == StreamTokenizer.TT_NUMBER) {
				rval = rval * Math.pow(10.0, tokenizer.nval);
			}
		}
		else {
			String s = tokenizer.sval;
			// Scientific notation
			int mult = Integer.valueOf(s.substring(1)).intValue();
			rval = rval * Math.pow(10.0, mult);
		}
	}
	else {
		tokenizer.pushBack();
	}
	return(rval);
}

/**
 * A private method that displays parsing error message.
 * @param nErrorIndex
 */
private void displayErrorMessage(int nErrorIndex)
{
	boolean appendTval = true;
	String strError =  ERRMSG_DATA_LINE + tokenizer.lineno() + ": ";

	switch (nErrorIndex) {
	case PARSE_ERR_DATA_STYLE:
		strError +=  ERRMSG_MUST_START + " and not";
		break;
	case PARSE_ERR_HOLE_OR_VALUE:
		strError +=  ERRMSG_EXPECT_HOLE;
		break;
	case PARSE_ERR_POINT_LABEL:
		strError +=  ERRMSG_EXPECT_POINT_LABEL;
		break;
	default:
		return;
	}
	if (appendTval) {
		if (tokenizer.ttype == QUOTE) {
			strError += "\'" + tokenizer.sval + "\'";
		}
		else if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
			strError += tokenizer.sval;
		}
		else if (tokenizer.ttype == StreamTokenizer.TT_NUMBER) {
			strError += tokenizer.nval;
		}
		else {
			strError += tokenizer.toString();
		}
	}
	System.err.println(strError);
}

/**
 * A private method that does all of the work with respect to reading in and
 * interpreting the stream.
 * <p>
 * We are not sure if synchronized is really necessary, but would hate
 * for someone to be in the middle of init still when they decide
 * they want to start drawing the chart.
 * @throws IOException
 */
private synchronized void init() throws IOException
{
	int nsets, npoints;
	boolean isTransposed;
	boolean tokenError = false;

	if (tokenizer.nextToken() != StreamTokenizer.TT_WORD) {
		tokenError = true;
	}
	else if (tokenizer.sval.equals("debug")) {
		
		
		if (tokenizer.nextToken() != StreamTokenizer.TT_WORD) {
			tokenError = true;
		}
	}
	if (tokenError ||
	   !(tokenizer.sval.equals("array") ||
		 tokenizer.sval.equals("general")))
	{
		displayErrorMessage(PARSE_ERR_DATA_STYLE);
		throw new IOException(ERRMSG_MUST_START);
	}

	boolean isArray = false;
	isArray = (tokenizer.sval.equals("array"));

	if (tokenizer.nextToken() == QUOTE) {
		String name = new String(tokenizer.sval);
		baseDS.name = ChartDataSourceUtil.processText(name);
	}
	else {
		tokenizer.pushBack();
	}

	nsets = (int)getNextDouble();
	npoints = (int)getNextDouble();

	if (tokenizer.nextToken() == StreamTokenizer.TT_WORD &&
		tokenizer.sval.equals("t"))
	{
		isTransposed = true;
	}
	else {
		isTransposed = false;
		tokenizer.pushBack();
	}

	

	// Check to see if there's a hole value
	if (tokenizer.nextToken() == StreamTokenizer.TT_WORD &&
		tokenizer.sval.equals("hole"))
	{
		baseDS.holeValue = getNextDouble();
		
	}
	else {
		tokenizer.pushBack();
	}

	if (isArray) {
		readArrayData(nsets, npoints, isTransposed);
	}
	else {
		readGeneralData(nsets, npoints, isTransposed);
	}
}

/**
 * Reads the data from the stream as if it is in array format
 * <p>
 * In array format, there is one set of
 * x-axis data, multiple sets of y-axis data.
 * The data may or may not be transposed; for example the two following
 * data files are equivalent:
 *
 * <PRE>
 * ARRAY 2 4  # 2 series (+x-data) w. 4 points each
 * 1 2 3 4    # x-data
 * 1 2 3 4    # y1
 * 1 4 9 16   # y2
 * </PRE>
 *
 * and
 *
 * <PRE>
 * ARRAY 2 4 T
 * 1 1 1
 * 2 2 4
 * 3 3 9
 * 4 4 16
 * </PRE>
 *
 * Or, with names and point labels,
 * <PRE>
 * ARRAY 'My Chart' 2 4
 *                 'Point Label 1' 'Point Label 2' 'Point Label 3' 'Point Label 4'
 * 'X Axis Data'    1               2               3               4
 * 'Y Axis #1 Data' 1               2               3               4
 * 'Y Axis #2 Data' 1               4               9               16
 * </PRE>
 *
 * or
 *
 * <PRE>
 * ARRAY 'MyChart' 2 4 T
 *                'X Data' 'Y Data' 'Y 2 Data'
 * 'Point Label 1' 1        1        1
 * 'Point Label 2' 2        2        4
 * 'Point Label 3' 3        3        9
 * 'Point Label 4' 4        4        16
 * </PRE>
 * @param isTransposed
 * @param npoints
 * @param nsets
 * @throws IOException
 */
private void readArrayData(int nsets, int npoints, boolean isTransposed)
    throws IOException
{
	int row, col;

	// Array data has only one x series
	baseDS.xvalues      = new double[1][npoints];
	baseDS.yvalues      = new double[nsets][npoints];
	baseDS.seriesLabels = new String[nsets];
	baseDS.pointLabels  = new String[npoints];

	if (isTransposed) {
		// First, read in the series names
		for (col = 0; col < nsets + 1; col++) {
			// Read in the series name, if it is there
			if (tokenizer.nextToken() == QUOTE) {
				if (col > 0) {
				    String l = new String(tokenizer.sval);
					baseDS.seriesLabels[col - 1] =
					    ChartDataSourceUtil.processText(l);
					
				}
			}
			else {
				tokenizer.pushBack();
				break;
			}
		}

		// Next, read in the point labels and the data
		for (col = 0; col < npoints; col++) {
			if (tokenizer.nextToken() == QUOTE) {
				String l = new String(tokenizer.sval);
				baseDS.pointLabels[col] = ChartDataSourceUtil.processText(l);
				
			}
			else {
				// ATTN GV in 3.*, used to send back empty string.
				tokenizer.pushBack();
			}

			// X value is the first column
			baseDS.xvalues[0][col] = getNextDouble();

			// Get y values
			for (row = 0; row < nsets; row++) {
				baseDS.yvalues[row][col] = getNextDouble();
				
			}
		}
	}
	else {
		// First, read the point labels if they're there
		for (col = 0; col < npoints; col++) {
			if (tokenizer.nextToken() == QUOTE) {
				String l = new String(tokenizer.sval);
				baseDS.pointLabels[col] = ChartDataSourceUtil.processText(l);
				
			}
			else {
				// ran out of labels
				tokenizer.pushBack();
			}
		}

		// Now read in the series labels and data
		for (row = 0; row < nsets + 1; row++) {

			// Get the series label if it is there
			if (tokenizer.nextToken() == QUOTE) {
				if (row > 0) {
				    String l = new String(tokenizer.sval);
					baseDS.seriesLabels[row - 1] =
					    ChartDataSourceUtil.processText(l);
					
				}
			}
			else {
				// ATTN GV in 3.*, used to send back empty string.
				tokenizer.pushBack();
			}

			// Now get the data
			for (col = 0; col < npoints; col++) {
				double dd = getNextDouble();
				if (row == 0) {
					// First row is a row of x values
					baseDS.xvalues[0][col] = dd;
				}
				else {
					// Second and all following rows contain y values
					baseDS.yvalues[row-1][col] = dd;
					
				}
			}
		}
	}
}

/**
 * Reads in data in general format.<p>
 * In general format, there are multiple x series.
 *
 * <PRE>
 * # 3 lines, max 5 points each
 * GENERAL 3 5
 * # line 1, 3 points
 * 3
 * 1 3 5    # x-values
 * 5 14 24  # y-values
 * # line 2, 5 points
 * 5
 * 1 2 3 4 5
 * 11 15 18 21 22
 * # line 3, 2 points
 * 2
 * 2 4
 * 16 15.2
 * </PRE>
 *
 * or, using the transposed format
 *
 * <PRE>
 * GENERAL 3 5 T
 * # line 1, 3 points
 * 3
 * # X  Y
 * 1   5
 * 3  14
 * 5  24
 * # line 2, 5 points
 * 5
 * # X  Y
 * 1   11
 * 2   15
 * 3   18
 * 4   21
 * 5   22
 * # line 3, 2 points
 * 2
 * # X  Y
 * 2   16
 * 4   15.2
 * </PRE>
 *
 * Or, with point labels and series names
 *
 * <PRE>
 * GENERAL 3 5
 * LABELS 'Mercury' 'Venus' 'Earth' 'Mars' 'Jupiter'
 * 'Series 1' 3
 * 1 3 5
 * 5 14 24
 * 'Series 2' 5
 * 1 2 3 4 5
 * 11 15 18 21 22
 * 'Series 3' 2
 * 2 4
 * 16 15.2
 * </PRE>
 *
 * <PRE>
 * GENERAL 3 5 T
 * LABELS 'Mercury' 'Venus' 'Earth' 'Mars' 'Jupiter'
 * # line 1, 3 points
 * 'Series 1' 3
 * # X  Y
 * 1   5
 * 3  14
 * 5  24
 * # line 2, 5 points
 * 'Series 2' 5
 * # X  Y
 * 1   11
 * 2   15
 * 3   18
 * 4   21
 * 5   22
 * # line 3, 2 points
 * 'Series 3' 2
 * # X  Y
 * 2   16
 * 4   15.2
 * </PRE>
 * @param isTransposed
 * @param npoints
 * @param nsets
 * @throws IOException
 */
private void readGeneralData(int nsets, int npoints, boolean isTransposed)
    throws IOException
{
	int row, col;

	// General data has equal # of x and y series.  Don't
	// know # of points yet.
	baseDS.xvalues      = new double[nsets][];
	baseDS.yvalues      = new double[nsets][];
	baseDS.seriesLabels = new String[nsets];
	baseDS.pointLabels  = new String[npoints];

	// Read in the point labels, which are shared between all the series
	if (tokenizer.nextToken() == StreamTokenizer.TT_WORD &&
	    tokenizer.sval.equals("labels"))
	{
		for (int i = 0; i < npoints; i++) {
			if (tokenizer.nextToken() != QUOTE) {
				displayErrorMessage(PARSE_ERR_POINT_LABEL);
				throw new IOException(ERRMSG_READ_POINT_LABELS + tokenizer);
			}
			else {
			    String l = new String(tokenizer.sval);
				baseDS.pointLabels[i] = ChartDataSourceUtil.processText(l);
				
			}
		}
	}
	else {
		tokenizer.pushBack();
		baseDS.pointLabels = null;
	}

	if (isTransposed) {
		for (row = 0; row < nsets; row++) {
			// Read in series labels
			if (tokenizer.nextToken() == QUOTE) {
			    String l = new String(tokenizer.sval);
				baseDS.seriesLabels[row] = ChartDataSourceUtil.processText(l);
				
			}
			else {
				tokenizer.pushBack();
			}

			npoints = (int)getNextDouble();
			

			baseDS.xvalues[row] = new double[npoints];
			baseDS.yvalues[row] = new double[npoints];

			for (col = 0; col < npoints; col++) {
				baseDS.xvalues[row][col] = getNextDouble();
				baseDS.yvalues[row][col] = getNextDouble();
				
			}
		}
	}
	else {
		for (row = 0; row < nsets; row++) {
			if (tokenizer.nextToken() == QUOTE) {
			    String l = new String(tokenizer.sval);
				baseDS.seriesLabels[row] = ChartDataSourceUtil.processText(l);
				
			}
			else {
				tokenizer.pushBack();
			}

			npoints = (int)getNextDouble();
			

			baseDS.xvalues[row] = new double[npoints];
			baseDS.yvalues[row] = new double[npoints];

			for (int x = 0; x < 2; x++) {
				for (col = 0; col < npoints; col++) {
					double dd = getNextDouble();
					if (x == 0) {
						// Add to x values
						baseDS.xvalues[row][col] = dd;
					}
					else {
						// Add to y values
						baseDS.yvalues[row][col] = dd;
						
					}
				}
			}
		}
	}
}

}

