/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCInputStreamDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:52 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;

import com.klg.jclass.util.io.Properties;

import java.io.*;

/**
 * This class is the base for any data source that relies on input
 * from some stream, which includes files, urls, large Strings, etc.
 * Data is interpreted according by a interpreter classes that subclasses from
 * <code>JCDataInterpreter</code>.
 */
public class JCInputStreamDataSource extends BaseDataSource {

/** Interpreter being used to interpret data. */
protected JCDataInterpreter interpreter = null;

/** The name of a supported charset */
protected String charsetName = Properties.DEFAULT_CHARSET_NAME;

/**
 * This constructor takes the <code>InputStream</code>, reads in all the
 * appropriate data and interprets it according to the default interpreter
 * class.
 *
 * @param i <code>InputStream</code> object used to read the data
 * @exception IOException An <code>IOException</code> will be thrown if there
 * are any errors while reading in the data.
 */
public JCInputStreamDataSource(InputStream i)
    throws IOException
{
    ctor(i);
}

/**
 * This constructor takes the <code>InputStream</code>, reads in all the
 * appropriate data, and interprets it according to the specified interpreter
 * class.
 *
 * @param i <code>InputStream</code> object used to read the data
 * @param p <code>Interpreter</code> used to make sense of the data
 * @exception IOException An <code>IOException</code> will be thrown if there
 * are any errors while reading in the data.
 */
public JCInputStreamDataSource(InputStream i, JCDataInterpreter p)
    throws IOException
{
    interpreter = p;
    ctor(i);
}

/**
 * This constructor takes the <code>InputStream</code>, reads in all the
 * appropriate data based on the given charset, and interprets it according to
 * the specified interpreter class.
 *
 * @param i <code>InputStream</code> object used to read the data
 * @param p <code>Interpreter</code> used to make sense of the data
 * @param charsetName The name of a supported charset
 * @exception IOException An <code>IOException</code> will be thrown if there
 * are any errors while reading in the data.
 */
public JCInputStreamDataSource(InputStream i, JCDataInterpreter p,
                               String charsetName)
    throws IOException
{
    interpreter = p;
    if (charsetName == null) {
        throw new IllegalArgumentException("charsetName cannot be null");
    }
    this.charsetName = charsetName;
    ctor(i);
}

/**
 * This constructor sets up a <code>JCInputStreamDataSource</code> class to use
 * the default interpreter. It does not yet read or interpret any data.
 * @throws IOException An <code>IOException</code> will be thrown if there are
 * any errors while reading in the data.
 */
protected JCInputStreamDataSource() throws IOException {}

/**
 * This constructor sets up a <code>JCInputStreamDataSource</code> class to use
 * the specified interpreter. It does not yet read or interpret any data.
 * @param p the data interpreter to be used
 * @throws IOException An <code>IOException</code> will be thrown if there are
 * any errors while reading in the data.
 */
protected JCInputStreamDataSource(JCDataInterpreter p)
    throws IOException
{
    interpreter = p;
}

/**
 * This constructor sets up a <code>JCInputStreamDataSource</code> class to use
 * the specified interpreter and charset. It does not yet read or interpret any
 * data.
 * @param p the data interpreter to be used
 * @param charsetName The name of a supported charset
 * @throws IOException An <code>IOException</code> will be thrown if there are
 * any errors while reading in the data.
 */
protected JCInputStreamDataSource(JCDataInterpreter p, String charsetName)
    throws IOException
{
    interpreter = p;
    if (charsetName == null) {
        throw new IllegalArgumentException("charsetName cannot be null");
    }
    this.charsetName = charsetName;
}

/**
 * Gets the interpreter
 * @return The current interpreter
 */
public JCDataInterpreter getInterpreter()
{
    return(interpreter);
}

/**
 * Sets the interpreter
 * @param interpreter The new interpreter
 */
public void setInterpreter(JCDataInterpreter interpreter)
{
    this.interpreter = interpreter;
}

/**
 * Gets the name of the charset
 * @return The name of the current charset
 */
public String getCharsetName()
{
    return(charsetName);
}

/**
 * Sets the name of the charset
 * @param charsetName The name of a supported charset
 */
public void setCharsetName(String charsetName)
{
    if (charsetName == null) {
        throw new IllegalArgumentException("charsetName cannot be null");
    }
    this.charsetName = charsetName;
}

/**
 * Parses the specified input stream and populates this data source.
 * Provided because subclasses may need to set the stream up before
 * they can call <code>super()</code> (for example,
 * <code>AppletDataSource</code>).
 * @param i the <code>InputStream</code> object to read data from
 * @throws IOException An <code>IOException</code> will be thrown if there are
 * any errors while reading in the data.
 */
protected void ctor(InputStream i)
    throws IOException
{
    Reader r = null;
    try {
        r = new BufferedReader(new InputStreamReader(i, charsetName));
    }
    catch (UnsupportedEncodingException e) {
        if (charsetName.equals(Properties.DEFAULT_CHARSET_NAME)) {
            r = new BufferedReader(new InputStreamReader(i));
        }
        else {
            throw e;
        }
    }
    ctor(r);
}

/**
 * Version of <code>ctor()</code> that takes a <code>Reader</code>.
 * Called by the other <code>ctor()</code>
 * and for String input where you can't have an input stream.  If no
 * interpreter has been set, it determines the appropriate interpreter
 * to use from the data in the reader.
 * @param r the <code>Reader</code> object to read data from
 * @throws IOException An <code>IOException</code> will be thrown if there are
 * any errors while reading in the data.
 */
protected void ctor(Reader r)
    throws IOException
{
    if (interpreter == null) {
        try {
            String name = "com.klg.jclass.chart.data.JCDefaultDataInterpreter";
            interpreter = (JCDataInterpreter)Class.forName(name).newInstance();
        }
        catch (Exception e) {
            e.printStackTrace();
            interpreter = null;
            return;
        }
    }
    interpreter.setBaseDataSource(this);
    interpreter.ctor(r);
}

}
