/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCXMLDataInterpreter.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:52 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;

import com.klg.jclass.util.xml.JCXMLDTDResolver;
import com.klg.jclass.util.xml.JCXMLErrorPrinter;
import com.klg.jclass.util.io.JCParseException;
import com.klg.jclass.util.io.LoadProperties;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.ErrorHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.Reader;

/**
 * Implements a interpreter for the chart XML data format. Relies on input
 * from some reader to populate the specified <code>BaseDataSource</code> class.
 *
 * <H2>Format Information</H2><P>
 * Data can be loaded either by series or by point.
 * <code>data-point-label</code> and <code>data-series-label</code> tags are
 * optional with either method. To represent virtual hole values in an
 * <code>x-data</code> or <code>y-data</code> tag, use the word
 * <code>HOLE</code>.
 * <p>
 * Here is an example of an XML data file loading by series.
 * This can be used to represent both the array and general style
 * data sources that are used by the default chart data source.
 * If only one <code>data-series</code> tag contains <code>x-data</code> tags,
 * that x data will be used for every series. Otherwise, each series uses its
 * own <code>x-data</code> tags to define its x data.
 *
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;!DOCTYPE chart-data SYSTEM "chart-data.dtd"&gt;
 * &lt;chart-data name="My Chart" hole="MAX"&gt;
 *   &lt;data-point-label&gt; Point Label 1 &lt;/data-point-label&gt;
 *   &lt;data-point-label&gt; Point Label 2 &lt;/data-point-label&gt;
 *   &lt;data-point-label&gt; Point Label 3 &lt;/data-point-label&gt;
 *   &lt;data-point-label&gt; Point Label 4 &lt;/data-point-label&gt;
 *   &lt;data-series&gt;
 *     &lt;data-series-label&gt; Y Axis #1 Data &lt;/data-series-label&gt;
 *     &lt;x-data&gt;1&lt;/x-data&gt;
 *     &lt;x-data&gt;2&lt;/x-data&gt;
 *     &lt;x-data&gt;3&lt;/x-data&gt;
 *     &lt;x-data&gt;4&lt;/x-data&gt;
 *     &lt;y-data&gt;1&lt;/y-data&gt;
 *     &lt;y-data&gt;2&lt;/y-data&gt;
 *     &lt;y-data&gt;3&lt;/y-data&gt;
 *     &lt;y-data&gt;4&lt;/y-data&gt;
 *    &lt;/data-series&gt;

 *    &lt;data-series&gt;
 *      &lt;data-series-label&gt; Y Axis #2 Data &lt;/data-series-label&gt;
 *      &lt;y-data&gt;1&lt;/y-data&gt;
 *      &lt;y-data&gt;4&lt;/y-data&gt;
 *      &lt;y-data&gt;9&lt;/y-data&gt;
 *      &lt;y-data&gt;16&lt;/y-data&gt;
 *    &lt;/data-series&gt;
 *  &lt;/chart-data&gt;
 * </pre>
 *
 *
 * Here is an example of an XML data file loading by point.
 * This format is similar to the transposed array format of the
 * default chart data source. No more than one <code>x-data</code> tag can be
 * used per <code>data-point</code> tag. If no <code>x-data</code> tags are
 * present at all, x data will be generated automatically staring at 1 in
 * increments of 1. If some <code>data-point</code> tags contain an
 * <code>x-data</code> tags and others do not, the missing pieces of x data
 * will be filled with the hole value.
 *
 * <pre>
 *  &lt;?xml version="1.0"?&gt;
 *  &lt;!DOCTYPE chart-data SYSTEM "chart-data.dtd"&gt;
 *  &lt;chart-data name="MyChart"&gt;
 *    &lt;data-series-label&gt; Y Data &lt;/data-series-label&gt;
 *    &lt;data-series-label&gt; Y 2 Data &lt;/data-series-label&gt;
 *    &lt;data-point&gt;
 *      &lt;data-point-label&gt; Point Label 1 &lt;/data-point-label&gt;
 *      &lt;x-data&gt;1&lt;/x-data&gt;
 *      &lt;y-data&gt;1&lt;/y-data&gt;
 *      &lt;y-data&gt;1&lt;/y-data&gt;
 *    &lt;/data-point&gt;
 *    &lt;data-point&gt;
 *      &lt;data-point-label&gt; Point Label 2 &lt;/data-point-label&gt;
 *      &lt;x-data&gt;2&lt;/x-data&gt;
 *      &lt;y-data&gt;2&lt;/y-data&gt;
 *      &lt;y-data&gt;4&lt;/y-data&gt;
 *    &lt;/data-point&gt;
 *    &lt;data-point&gt;
 *      &lt;data-point-label&gt; Point Label 3 &lt;/data-point-label&gt;
 *      &lt;x-data&gt;3&lt;/x-data&gt;
 *      &lt;y-data&gt;3&lt;/y-data&gt;
 *      &lt;y-data&gt;9&lt;/y-data&gt;
 *    &lt;/data-point&gt;
 *    &lt;data-point&gt;
 *      &lt;data-point-label&gt; Point Label 4 &lt;/data-point-label&gt;
 *      &lt;x-data&gt;4&lt;/x-data&gt;
 *      &lt;y-data&gt;4&lt;/y-data&gt;
 *      &lt;y-data&gt;16&lt;/y-data&gt;
 *    &lt;/data-point&gt;
 *  &lt;/chart-data&gt;
 * </pre>
 */
public class JCXMLDataInterpreter extends JCDataInterpreter
							      implements java.io.Serializable {

static public String DATA_DTD_NAME = "JCChartData.dtd";

protected LoadProperties loadProperties;

/**
 * Default no-argument constructor.
 */
public JCXMLDataInterpreter()
{
	loadProperties = new LoadProperties();
}

/**
 * Constructor specifying LoadProperties class.
 */
public JCXMLDataInterpreter(LoadProperties lp)
{
	loadProperties = lp;
}

/**
 * This constructor creates an instance of a data interpreter and associates it
 * with a data source. It does not yet do any interpreting.
 * @param bds the data source to eventually populate with the interpreted data
 *
 * @throws IOException
 */
public JCXMLDataInterpreter(BaseDataSource bds) throws IOException
{
	super(bds);
}

/**
 * This constructor creates an instance of a data interpreter and associates it
 * with a data source. It does not yet do any interpreting.
 * @param bds the data source to eventually populate with the interpreted data
 * @param lp an instance of LoadProperties which contains attributes
 *            related to XML parsing of files.
 *
 * @throws IOException
 */
public JCXMLDataInterpreter(BaseDataSource bds, LoadProperties lp)
													 throws IOException
{
	this(bds);
	loadProperties = lp;
}

/**
 * Parses the specified input stream according to the XML chart
 * data format and populates the previously set data source. Called
 * by the other <code>ctor()</code> method, and also used for String
 * input where you cannot have an input stream.
 * @param r the <code>Reader</code> object used to read data
 * @throws IOException
 */
protected void ctor(Reader r) throws IOException
{
	loadData(new InputSource(r), true);
}

/**
 * Internal use only.
 * Method creates data objects from an XML InputSource.
 * <p>
 * We are not sure if synchronized is really necessary, but would hate
 * for someone to be in the middle of init still when they decide
 * they want to start drawing the chart.
 *
 * @param inSource the <code>InputSource</code> that contains the data file
 * @param validate <code>true</code> if validation should be turned on
 * @throws IOException
 * @throws JCParseException
 */
public synchronized void loadData(InputSource inSource, boolean validate)
    throws IOException, JCParseException
{
	// System id necessary of SAX parser will throw exceptions
	// if not set, use value from LoadProperties or empty string
	if (inSource.getSystemId() == null) {
		if (loadProperties != null) {
    		inSource.setSystemId(loadProperties.getLocalEntitySystemId());
		}
        else {
    		inSource.setSystemId("");
		}
	}

    SAXParser parser = null;
    XMLReader xmlReader = null;
    SAXParserFactory factory = SAXParserFactory.newInstance();
	factory.setValidating(validate);
	if (baseDS == null) {
		baseDS = new BaseDataSource();
	}
    XMLDataHandler handler = new XMLDataHandler(baseDS);
    try {
	    parser = factory.newSAXParser();
		xmlReader = parser.getXMLReader();
	    xmlReader.setContentHandler(handler);

        // Set the error handler.  Can be user defined.
        ErrorHandler errorHandler =
            (loadProperties == null) ? null
                                     : loadProperties.getXmlErrorHandler();
        if (errorHandler == null) {
            errorHandler = new JCXMLErrorPrinter();
        }
	    xmlReader.setErrorHandler(errorHandler);

	    xmlReader.setEntityResolver(new JCXMLDTDResolver(DATA_DTD_NAME,
												         handler.getClass()));
	    xmlReader.parse(inSource);
    }
    catch (SAXException se) {
        throw new JCParseException(se.getMessage(), se);
    }
    catch (ParserConfigurationException pce) {
        throw new JCParseException(pce.getMessage(), pce);
    }
}

}
