/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JDBCDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:52 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.data;

import com.klg.jclass.chart.ChartDataEvent;

import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Locale;
import java.util.Vector;

/**
 * JDBCDataSource extends JCDefaultDataSource to create
 * a data source for use with JDBC.
 */
public class JDBCDataSource extends JCDefaultDataSource {

/**
 * Default constructor.
 */
public JDBCDataSource()
{
    this(Locale.getDefault());
}

/**
 * Default constructor with locale.
 */
public JDBCDataSource(Locale locale)
{
    super(locale);
}

/**
 * The JDBCDataSource constructor takes a ResultSet as source data.
 * Default behavior is to use all numeric columns for Y data and automatically
 * generate x values.
 *
 * @param jdbcData the input data as a ResultSet
 */
public JDBCDataSource(ResultSet jdbcData)
{
    setJDBCData(jdbcData, null, null, null);
}

/**
 * The JDBCDataSource constructor takes a ResultSet as source data and
 * column names to specify the columns to be used for data.
 * @param jdbcData ResultSet object to get data from.
 * @param xColumnName the name of the column to use as x value data.
 *                    if null, x values are generated automatically.
 * @param pointLabelColumnName the name of the column to use for point
 *                             label names.  If null, point labels are
 *                             generated automatically.
 * @param activeColumnNames array of column names to use for y values.
 *                          if null, all numeric columns not associated
 *                          with the x column or the point labels column
 *                          are used.
 *
 * @param jdbcData the input data as a ResultSet
 */
public JDBCDataSource(ResultSet jdbcData, String xColumnName,
                      String pointLabelColumnName, String[] activeColumnNames)
{
    setJDBCData(jdbcData, xColumnName, pointLabelColumnName, activeColumnNames);
}

/**
 * Sets the data for this data source from the passed result set and according
 * to  the specification of columns in the passed name parameters.
 * @param jdbcData ResultSet object to get data from.
 * @param xColumnName the name of the column to use as x value data.
 *                    if null, x values are generated automatically.
 * @param pointLabelColumnName the name of the column to use for point
 *                             label names.  If null, point labels are
 *                             generated automatically.
 * @param activeColumnNames array of column names to use for y values.
 *                          if null, all numeric columns not associated
 *                          with the x column or the point labels column
 *                          are used.
 *
 * @param jdbcData the input data as a ResultSet
 */
public void setData(ResultSet jdbcData, String xColumnName,
                    String pointLabelColumnName, String[] activeColumnNames)
{
    setJDBCData(jdbcData, xColumnName, pointLabelColumnName, activeColumnNames);
    updateChart(ChartDataEvent.RESET, 0, 0);
}

/**
 * Internal method to Populate the datasource with data from the result set
 * according to the specification of columns in the passed name parameters.
 * Series labels are the column labels of the active columns.
 * @param jdbcData ResultSet object to get data from.
 * @param xColumnName the name of the column to use as x value data.
 *                    if null, x values are generated automatically.
 * @param pointLabelColumnName the name of the column to use for point
 *                             label names.  If null, point labels are
 *                             generated automatically.
 * @param activeColumnNames array of column names to use for y values.
 *                          if null, all numeric columns not associated
 *                          with the x column or the point labels column
 *                          are used.
 */
protected void setJDBCData(ResultSet jdbcData, String xColumnName,
                           String pointLabelColumnName,
                           String[] activeColumnNames)
{
    try {
        // No result set? No meta data?  No columns? Clear the data and get out.
        if (jdbcData == null) {
            clearData();
            return;
        }
        ResultSetMetaData md = jdbcData.getMetaData();
        if (md == null) {
            clearData();
            return;
        }
        int columns = md.getColumnCount();
        if (columns <= 0) {
            clearData();
            return;
        }

        // Initialize column numbers to default case where no columns
        // are explicitly specified
        int xColumnNumber = -1;
        int pointLabelColumnNumber = -1;
        boolean hasXColumn = false;
        boolean hasPointLabelColumn = false;
        Vector activeColumnNumbers = new Vector();

        // Root through columns looking for specified x, pointlabel, and
        // active y columns.  Note them if found.
        for (int i = 1; i <= columns; i++) {
            String columnName = md.getColumnName(i);
            if (xColumnName != null && columnName.equals(xColumnName)) {
                xColumnNumber = i;
                hasXColumn = true;
            }
            if (pointLabelColumnName != null &&
                columnName.equals(pointLabelColumnName))
            {
                pointLabelColumnNumber = i;
                hasPointLabelColumn = true;
            }
            if (activeColumnNames != null) {
                for (int j = 0; j < activeColumnNames.length; j++) {
                    if (activeColumnNames[j] != null &&
                        columnName.equals(activeColumnNames[j]))
                    {
                        activeColumnNumbers.addElement(new Integer(i));
                    }
                }
            }
        }

        // If point label column not specified, take the first column if
        // it contains text
        if (pointLabelColumnName == null) {
            int SQLType = md.getColumnType(1);
            if (SQLType == Types.CHAR    ||
                SQLType == Types.VARCHAR ||
                SQLType == Types.LONGVARCHAR)
            {
                pointLabelColumnNumber = 1;
                hasPointLabelColumn = true;
            }
        }

        // If we're dealing with default column selection, do a second pass
        // through the columns.  Pull out all columns containing numbers that
        // aren't already specified to be the x or point labels column.
        if (activeColumnNames == null) {
            for (int i = 1; i <= columns; i++) {
                if (i == xColumnNumber || i == pointLabelColumnNumber) {
                    continue;
                }

                int SQLType = md.getColumnType(i);
                if (SQLType == Types.DOUBLE   ||
                    SQLType == Types.FLOAT    ||
                    SQLType == Types.BIGINT   ||
                    SQLType == Types.INTEGER  ||
                    SQLType == Types.TINYINT  ||
                    SQLType == Types.SMALLINT ||
                    SQLType == Types.REAL     ||
                    SQLType == Types.DECIMAL  ||
                    SQLType == Types.NUMERIC)
                {
                    activeColumnNumbers.addElement(new Integer(i));
                }
            }
        }

        // Number of series in data source is equal to the number of columns
        // of data we'll pull out of the query
        int nSeries = activeColumnNumbers.size();

        // Convert active column vector back into array for easier access
        int[] activeColumnArray = new int[nSeries];
        for (int i = 0; i < nSeries; i++) {
            activeColumnArray[i] =
                ((Integer)activeColumnNumbers.elementAt(i)).intValue();
        }

        Vector vData = new Vector();
        Vector vLabels = new Vector();
        Vector xData = null;
        if (hasXColumn) {
            xData = new Vector();
        }
        seriesLabels = new String[nSeries];

        // Go through the result set row by row
        try {
            int rowIndex = -1;
            Object pointLabelObject = null;
            Object xObject = null;
            while (jdbcData.next()) {
                rowIndex++;

                // Get the data from the x column and the point label column
                if (hasPointLabelColumn) {
                    pointLabelObject =
                        jdbcData.getObject(pointLabelColumnNumber);
                    String label = "";
                    if (pointLabelObject != null) {
                        label = pointLabelObject.toString();
                    }
                    vLabels.addElement(label);
                }
                if (hasXColumn) {
                    // Be careful if x and point labels aren't sharing the same
                    // column
                    if (xColumnNumber == pointLabelColumnNumber) {
                        xObject = pointLabelObject;
                    }
                    else {
                        xObject = jdbcData.getObject(xColumnNumber);
                    }
                    xData.addElement(xObject);
                }

                // Get the data from the y columnd
                double[] rowData = new double[nSeries];
                for (int i = 0; i < nSeries; i++) {
                    int thisColumnNumber = activeColumnArray[i];

                    // Be careful if y values are sharing columns with x or
                    // point labels
                    Object dataObject;
                    if (thisColumnNumber == xColumnNumber) {
                        dataObject = xObject;
                    }
                    else if (thisColumnNumber == pointLabelColumnNumber) {
                        dataObject = pointLabelObject;
                    }
                    else {
                        dataObject = jdbcData.getObject(thisColumnNumber);
                    }

                    // If a number, use it
                    if (dataObject instanceof Number) {
                        rowData[i] = ((Number)dataObject).doubleValue();
                    }
                    // If a string, try to parse it
                    else if (dataObject instanceof String) {
                        try {
                            Double.valueOf((String)dataObject).doubleValue();
                        }
                        catch (NumberFormatException nfe) {
                            rowData[i] = Double.MAX_VALUE;
                        }
                    }
                    // Otherwise, we can't use it
                    else {
                        rowData[i] = Double.MAX_VALUE;
                    }

                    // First time through, get the series labels
                    if (rowIndex == 0) {
                        seriesLabels[i] = md.getColumnName(thisColumnNumber);
                    }
                }
                vData.addElement(rowData);
            }
        }
        catch (SQLException e) {
            e.printStackTrace();
        }

        int nDataPoints = vData.size();

        yvalues = new double[nSeries][nDataPoints];
        if (hasXColumn) {
            xvalues = new double[1][nDataPoints];
        }
        else {
            xvalues = null;
        }
        if (hasPointLabelColumn) {
            pointLabels = new String[nDataPoints];
        }
        else {
            pointLabels = null;
        }

        for (int j = 0; j < nDataPoints; j++) {
            double[] rowData = (double[])vData.elementAt(j);
            for (int i = 0; i < nSeries; i++) {
                yvalues[i][j] = rowData[i];
            }
            if (hasXColumn) {
                xvalues[0][j] = ((Number)xData.elementAt(j)).doubleValue();
            }
            if (hasPointLabelColumn) {
                pointLabels[j] = (String)vLabels.elementAt(j);
            }
        }

        // Generate x values if necessary
        check();
    }
    catch (SQLException e) {
        e.printStackTrace(System.out);
        clearData();
        return;
    }
}

/**
 * Internal method to clear the data
 */
void clearData()
{
    xvalues = null;
    yvalues = null;
}

}
