/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: ChartDataBinding.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:52 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.db;

import com.klg.jclass.datasource.DataModelException;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.datasource.util.DataBinding;


/**
 * This is the core data binding handling class for
 * LiveChart.  Most of the key functionality is in
 * DataBinding.
 */
class ChartDataBinding extends DataBinding {

/** Data-bound chart instance.  */
protected DataChart dataChart;

/** Data-bound chart data source */
protected DataBoundSource dataSource;

public static final int START_POINT = 0;
public static final int END_POINT = 100;

protected int startPoint = START_POINT;
protected int endPoint = END_POINT;
protected boolean allRows = false;

/**
 * Creates a data binding object give an DataChart and
 * its corresponding data source
 * @param chart data-bound chart instance
 * @param dataSource external data source for data-bound chart.
 */
ChartDataBinding(DataChart chart, DataBoundSource dataSource) {
	super(chart);
	this.dataSource = dataSource;
    dataChart = chart;
}


/**
 * Returns current zero-based row index for the DataChartModel instance.
 * @return -1 if current bookmark is invalid.
 */
public int getCurrentRowIndex() {
    return super.getCurrentRowIndex();
}

/**
 * Changes the current row in the component.
 * Does nothing here.
 */
protected void currentRowChanged( int rowIndex ) {}

/**
 * Checks if all listeners have accepted the move.
 * Does nothing here.
 */
protected boolean beforeCurrentRowChange() {
	return true;
}

/**
 * Process a move to row for any metadata level called from dataModelChanged()
 */
protected void tableChanged() {
	dataSource.refreshChart();
}

/**
 * Must be called after row data were changed.
 */
protected void rowDataChanged(int rowIndex) {
	dataSource.refreshChart();
}

/**
 * Must be called after a row was inserted.
 */
protected void rowInserted() {
	dataSource.refreshChart();
}

/**
 * Must be called after a row was deleted.
 */
protected void rowDeleted() {
	dataSource.refreshChart();
}

/**
 * Empty implementation of setDataAvailable.
 */
protected void setDataAvailable(boolean state) {}

/**
 * Must be called after a single cell value was changed.
 */
protected void cellDataChanged(int row, String columnName) {
	dataSource.refreshChart();
}

protected int getFirstRow() {
    return allRows ? 0 : startPoint;
}

protected int getLastRow() {
    int lastRow = binding.getRowCount()-1;
	return allRows ? lastRow : Math.min(lastRow, endPoint);
}

public String[] getLabels(String column) {
	if (!binding.isDataAvailable() || column == null || column.length() == 0) {
		return null;
    }

    int firstRow = getFirstRow();
    int lastRow = getLastRow();
    String[] labels = new String[lastRow-firstRow+1];
    for (int i=firstRow; i<=lastRow; i++) {
        try {
            labels[i-firstRow] = binding.getCell(this, i, column).toString();
        }
        catch(DataModelException e) {
            e.printStackTrace();
        }
    }
    return labels;
}

/**
 * @see DataBindable#getData
 */
public double[] getData(String column) {
	if (!binding.isDataAvailable()) {
		return null;
    }
    int firstRow = getFirstRow();
    int lastRow = getLastRow();
    double[] data = new double[lastRow-firstRow+1];
    Object value = null;
    for (int i=firstRow; i<=lastRow; i++) {
        try {
        	value = binding.getCell(this, i, column);
			if (value instanceof Number) {
		        data[i-firstRow] = ((Number)value).doubleValue();
            }
			else if (value instanceof String) {
		        data[i-firstRow] = (Double.valueOf((String)value)).doubleValue();
            }
        }
        catch(DataModelException e) {
	        processException(e);
        }
        catch (ClassCastException e) {
            System.out.println("Invalid double value for column '"+column+
            	"', row "+i+": "+value);
        }
        catch (NumberFormatException e) {
            System.out.println("Invalid double value for column '"+column+
            	"', row "+i+": "+value);
        }
    }
    return data;
}

/**
 * Is the given type numeric?
 * this includes: MetaDataModel.BIGINT or MetaDataModel.DECIMAL or
 * MetaDataModel.DOUBLE or MetaDataModel.FLOAT or MetaDataModel.INTEGER or 
 * MetaDataModel.NUMERIC or MetaDataModel.SMALLINT orMetaDataModel.TINYINT;
 * @param represents the MetaDataModel type
 * @return true is this type is numeric, false otherwise
 */
public static boolean isNumerical( int type ) {
	return type == MetaDataModel.TYPE_DOUBLE ||
    	type == MetaDataModel.TYPE_FLOAT ||
		type == MetaDataModel.TYPE_INTEGER ||
        type == MetaDataModel.TYPE_BIG_DECIMAL ||
		type == MetaDataModel.TYPE_LONG ||
        type == MetaDataModel.TYPE_BYTE ||
		type == MetaDataModel.TYPE_SHORT;
}

/**
 * Returns true if the column is numerical.
 */
public boolean isNumericalColumn(String column) {
    return isNumerical(getJavaColumnType(column));
}

/**
 * Returns true if the component is bound to the datasource.
 */
public boolean isEnabled() {
    return binding != null && binding.isDataAvailable();
}

/**
 * Returns true if the column name is valid.
 */
public boolean isValidColumn(String column) {
    return binding.getColumnIndex(column) != -1;
}

}	// ChartDataBinding

