/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: DataBoundSource.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:52 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.db;

import com.klg.jclass.chart.ChartDataEvent;
import com.klg.jclass.chart.data.JCDefaultDataSource;

import com.klg.jclass.datasource.DataModel;
import com.klg.jclass.datasource.DataTableModel;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.datasource.BindingModel;

/**
 * External data source for LiveChart data binding.  Uses
 * a ChartDataBinding instance to communicate with the
 * JClass DataSource.
 */
class DataBoundSource extends JCDefaultDataSource {

protected ChartDataBinding dataBinding;

protected String labelsColumn = null;		// Labels column
protected String[] seriesColumns = null; 	// Data column
protected int[] numericalColumns = null;
protected String[] allColumns = null;

private MetaDataModel lastMetaDataModel = null;	// Last MetaDataModel used 


/**
 * Creates a new DataBoundSource object which is bound to the root metadata level.
 */
DataBoundSource(DataChart chart, DataModel dataModel) {
	this(chart, dataModel, (MetaDataModel)null);
}

/**
 * Creates a new DataBoundSource object which is bound to the given MetaDataModel.
 */
DataBoundSource(DataChart chart, DataModel dataModel, MetaDataModel metaDataModel) {
    this(chart);
	dataBinding.setDataBinding(dataModel, metaDataModel);
}

/**
 * Creates a new DataBoundSource object for a DataModel and binds the MetaData to
 * the MetaDataModel represented by a "path" of MetaDataModel
 * descriptions separated by '|' (e.g. Orders|Customers).
 */
DataBoundSource(DataChart chart, DataModel dataModel, String metaDescription) {
	this(chart);
    dataBinding.setDataBinding(dataModel, metaDescription);
}

/**
 * Creates a new DataBoundSource object for the data binding hierarchy string, starting from a
 * TreeDataBean instance name and containing full sequence of MetaDataModel levels starting
 * from the root to the level the chart is bound with.
 */
DataBoundSource(DataChart chart, String s) {
	this(chart);
    dataBinding.setDataBinding(s);
}

/**
 * Bind using a Binding.
 */
DataBoundSource(DataChart chart, BindingModel binding) {
	this(chart);
    dataBinding.setDataBinding(binding);
}
	
/**
 * Creates a new DataBoundSource with ChartDataBinding not bound to a datasource.
 */
private DataBoundSource(DataChart chart) {
    dataBinding = new ChartDataBinding(chart, this);
}

/**
 * Returns the data binding hierarchy string, starting from a TreeDataBean instance name
 * and containing full sequence of MetaDataModel levels starting from the root to the level
 * the chart is bound with.
 *
 * @see #setDataBinding
 */
public String getDataBinding() {
	return dataBinding.getDataBinding();
}

/**
 * Sets the DataBinding property value to the chart.
 *
 * Sets a chart datasource for the data binding hierarchy string, starting from a TreeDataBean
 * instance name and containing full sequence of MetaDataModel levels starting from the root
 * to the level the chart is bound with.
 *
 * @see #getDataBinding
 */
public void setDataBinding(String s) {
	dataBinding.setDataBinding(s);
    updateConfig();
}

/**
 * Bind to a binding.
 */
public void setDataBinding(BindingModel binding) {
	dataBinding.setDataBinding(binding);
}
	
/**
 * Sets up parameters to bind the DataBoundSource to a MetaDataModel in the
 * given DataModel.
 */
protected void setDataBinding(DataModel dataModel, MetaDataModel metaDataModel) {
	dataBinding.setDataBinding(dataModel, metaDataModel);
    updateConfig();
}

/**
 * Allows the DataBoundSource to clean up any references.
 */
public void finalize() {
	dispose();
}

/**
 * If the DataBoundSource is attached to a TreeData object, remove it from
 * the TreeData's DataModelListener list.
 */
public void dispose() {
	dataBinding.dispose();
}

/**
 * Gets an array containing names of all table columns.
 * @return array containing all the columns for this data source.
 */
public void initColumns() {
    if (dataBinding == null) {
        return;
    }
    allColumns = new String [dataBinding.getNumColumns()];
    int numericalCount = 0;
	  for (int i = 0; i < dataBinding.getNumColumns(); i++) {
        String column = (String)dataBinding.getTableColumnLabel(i);
	  	  allColumns[i] = column;
        if (dataBinding.isNumericalColumn(column)) {
          numericalCount++;
        }
	  }
    if (allColumns.length > 0) {
	      labelsColumn = allColumns[0];
    }
	boolean fillSeriesColumns = false;
	if (seriesColumns == null)
		fillSeriesColumns = true;

	if (fillSeriesColumns)
    	seriesColumns = new String[numericalCount];
    numericalColumns = new int[numericalCount];
	for (int i=0,j=0; i<allColumns.length; i++) {
        String column = allColumns[i];
        if (dataBinding.isNumericalColumn(column)) {
			if (fillSeriesColumns)
            	seriesColumns[j] = column;
            numericalColumns[j] = i;
            j++;
        }
	}
}

//================== Helper methods ==================

/**
 * Returns the native MetaData.
 */
public Object getNativeMetaData() {
	return dataBinding.getNativeMetaData();
}

/**
 * Gets the value of the SeriesColumns indexed property.
 * @return array containing all the series columns for this data source.
 */
public String[] getDataColumns() {
	return seriesColumns;
}

/**
 * Sets the value of the SeriesColumns indexed property. Only columns that have
 * a numeric type can be potential series columns.
 * @param _seriesColumns containing all the series columns for this data source.
 * @see #isNumericalColumn
 */
public void setDataColumns(String[] seriesColumns) {
	this.seriesColumns = seriesColumns;
	refreshSeries();
}

/**
 * Sets the value of the PointLabelsColumn property.
 * @param _pointLabelsColumn containing the name of the column to be bound as
 * the point labels column
 */
public void setPointLabelsColumn(String labelsColumn) {
	if (!dataBinding.isValidColumn(labelsColumn)) {
   		throw new IllegalArgumentException("The column "+labelsColumn+
       		" doesn't exist in the table.");
   	}
	this.labelsColumn = labelsColumn;
	setPointLabels(dataBinding.getLabels(this.labelsColumn));
}

/**
 * Gets the value of the PointLabelsColumn property.
 * @return a String containing the name of the column to be bound as
 * the point labels column
 */
public String getPointLabelsColumn() {
	return labelsColumn;
}

/**
 * Sets the value of the name property. The name property defines
 * the name of the data object.
 * @param s the name string
 */
public void setName(String s) {
    super.setName(s);
    updateChart(ChartDataEvent.RESET, 0, 0);
}

/**
 * Returns the dataBindingConfig property value.
 * @return a DataBindingConfigWrapper containing the data binding
 * configuration information.
 */
public DataBindingConfigWrapper getDataBindingConfigWrapper() {
	return new DataBindingConfigWrapper(
        dataBinding.allRows, dataBinding.startPoint, dataBinding.endPoint,
        labelsColumn, seriesColumns, numericalColumns, allColumns);
}

/**
 * Sets the value of the dataBindingConfig property.
 * @param wrapper the property value
 */
public void setDataBindingConfigWrapper(DataBindingConfigWrapper wrapper) {
	dataBinding.allRows = wrapper.allRows;
    dataBinding.startPoint = wrapper.startPoint;
    dataBinding.endPoint = wrapper.endPoint;

    initColumns();
    if (isValidColumn(wrapper.labelsColumn)) {
	    labelsColumn = wrapper.labelsColumn;
    }
    if (isValidColumnSet(wrapper.seriesColumns)) {
	    seriesColumns = wrapper.seriesColumns;
    }

	refreshPointLabels();
	refreshSeries();
}

/**
 * Implements the DBUpdateListener interface. Listenes to DBUpdateEvents.
 */
public void refreshChart() {
	// save current labels columns and series columns in case they
	// have been set by the user to something specific
	String saveLabelsColumn = labelsColumn;
	String[] saveSeriesColumns = seriesColumns;

   	updateConfig();

	// if set by the user, revert back to set series and label columns
	if (saveSeriesColumns != null) {
		seriesColumns = saveSeriesColumns;
	}
	if (saveLabelsColumn != null) {
		labelsColumn = saveLabelsColumn;
	}
	refreshSeries();
	refreshPointLabels();

}

/**
 * Updates the data binding configuration information.
 */
protected void updateConfig() {
    initColumns();
}

protected boolean isValidColumn(String column) {
	for (int i=0; i<allColumns.length; i++) {
        if (column.equalsIgnoreCase(allColumns[i])) {
        	return true;
        }
	}
    return false;
}

protected boolean isValidColumnSet(String[] columns) {
	for (int i=0; i<columns.length; i++) {
        if (!isValidColumn(columns[i]) ||
        	!dataBinding.isNumericalColumn(columns[i])) {
        	return false;
        }
	}
    return true;
}

protected void refreshPointLabels() {
	if (labelsColumn != null && labelsColumn.compareTo("") != 0) {
		setPointLabels(dataBinding.getLabels(labelsColumn));
	}
}

protected void refreshSeries() {
	double[][] newData = new double[seriesColumns.length][];
	// Get y values
	for (int i=0; i < seriesColumns.length; i++) {
		newData[i] = dataBinding.getData(seriesColumns[i]);
	}
	synchronized(this) {
        yvalues = newData;
        // Set xvalues to null so that call to check() will know to regenerate them
        xvalues = null;
        if (yvalues[0] != null) {
          check();
        }
        updateChart(ChartDataEvent.RESET, 0, 0);
    }
	setSeriesLabels(seriesColumns);
}

}	// DataBoundSource
