/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: DataChart.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:52 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.db;

import com.klg.jclass.datasource.DataModel;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.datasource.BindingModel;

import com.klg.jclass.chart.ChartDataModel;
import com.klg.jclass.chart.beans.BaseChart;


/**
 * This is a data-aware subclass of <code>BaseChart</code>.
 * It displays data from the data source.
 * This component displays the result set of a data bean and can insert and delete
 * records, cancel, save and requery data in the data source.
 * The class is not a bean, but a common ancestor of all data-aware chart beans.
 */
public class DataChart extends BaseChart {

/**
 * Protected <code>DataChart</code> constructor. The constructor must not be called
 * directly, but only from it's subclasses.
 */
protected DataChart() {
	super();
}

/**
 * Sets a chart datasource.
 */
public void setDataSource(ChartDataModel dataSource) {
	super.getDataView(0).setDataSource(dataSource);
}

/**
 * Binds the chart to the given DataModel and to the
 * root <code>MetaDataModel</code>.
 */
public void setDataSource(DataModel dataModel) {
    initDataBinding(new DataBoundSource(this, dataModel));
}

/**
 * Binds the chart to the given
 * <code>DataModel</code> and Meta<code>DataModel</code>DataModel.
 */
public void setDataSource(DataModel dataModel, MetaDataModel metaDataModel) {
    initDataBinding(new DataBoundSource(this, dataModel, metaDataModel));
}

/**
 * Bind using a <code>Binding</code>.
 */
public void setDataSource(BindingModel binding) {
    initDataBinding(new DataBoundSource(this, binding));
}

/**
 * Sets the <code>DataBinding</code> property value to the chart.
 *
 * Sets a chart datasource for the data binding hierarchy string,
 * starting from a <code>JCTreeData</code>
 * instance name and containing full sequence of
 * <code>MetaDataModel</code> levels starting from the root
 * to the level the chart is bound with.
 *
 * @see com.klg.jclass.chart.db.datasource.DSdbChart#getDataBinding
 */
protected void setDataSource(String dataBinding) {
    initDataBinding(new DataBoundSource(this, dataBinding));
}

/**
 * Sets the data source to a <code>DataModel</code>
 * and binds the <code>MetaData</code> to
 * the <code>MetaDataModel</code>
 * represented by a "path" of <code>MetaDataModel</code>
 * descriptions separated by '|' (e.g. Orders|Customers).
 */
protected void setDataSource(DataModel dataModel, String metaDescription) {
    initDataBinding(new DataBoundSource(this, dataModel, metaDescription));
}

/**
 * Verifies that a new data source is valid and binds the chart.
 */
protected void initDataBinding(DataBoundSource dataSource) {
    if (dataSource == null || dataSource.dataBinding == null ||
    	!dataSource.dataBinding.isEnabled()) {
    	return;
	}
	ChartDataModel cdm = getDataView(0).getDataSource();
	if(cdm != null && cdm instanceof DataBoundSource) {
	    ((DataBoundSource)cdm).dispose();
	}
	setDataSource(dataSource);
}

/**
 * Returns the chart's value of the
 * <code>DataBinding</code> object property.
 * Returns the data binding hierarchy string,
 * starting from a <code>JCTreeData</code> instance name
 * and containing full sequence of <code>MetaDataModel</code>
 * levels starting from the root to the level
 * at which the chart is bound.
 *
 * The <code>DataBinding</code> property specifies a string
 * describing a <code>JCTreeData</code> instance and
 * metadata level at which the instance this chart is bound.
 * The format of the property is as follows:
 * <br>
 * &nbsp;&nbsp;&nbsp;&nbsp;&lt;JCTreeData name&gt;:&lt;MetaDataModel
 * description&gt;[|&lt;MetaDataModel description&gt;...]
 *
 * @see #setDataBinding
 */
protected String getDataBindingString() {
	DataBoundSource dbs = getDataBoundSource();
	return dbs != null ? dbs.getDataBinding() : "";
}

/**
 * Sets up parameters to bind the <code>DataBoundSource</code>
 * to a <code>MetaDataModel</code> in the
 * given <code>DataModel</code>.
 */
protected boolean setDataBinding(DataModel dataModel, MetaDataModel metaDataModel) {
	DataBoundSource dbs = getDataBoundSource();
	if (dbs != null) {
		dbs.dataBinding.setDataBinding(dataModel, metaDataModel);
        return true;
    }
    return false;
}

/**
 * Sets the meta property used to build a <code>DataBoundSource</code>.
 * @param wrapper the instance of <code>DataBindingConfigWrapper</code>
 */
public void setDataBindingConfig(DataBindingConfigWrapper wrapper) {
	DataBoundSource dbs = getDataBoundSource();
	if (dbs != null) {
    	dbs.setDataBindingConfigWrapper(wrapper);
    }
}

/**
 * Gets the <code>dataBindingConfig</code>
 * property used to build a <code>DataBoundSource</code>.
 * @return the <code>dataBindingConfig</code> property
 */
public DataBindingConfigWrapper getDataBindingConfig() {
	DataBoundSource dbs = getDataBoundSource();
//	return dbs == null ? null : dbs.getDataBindingConfigWrapper();
	return dbs == null ?
    	new DataBindingConfigWrapper() : dbs.getDataBindingConfigWrapper();
}

/**
 * Saves the value from the current editor to the data source before
 * the current row has been changed.
 */
protected boolean beforeCurrentRowChange() {
	return true;
}

/**
 * Returns <code>DataBoundSource</code> for the chart.
 */
protected DataBoundSource getDataBoundSource() {
	ChartDataModel dataSource = getDataView(0).getDataSource();
	return (dataSource != null && dataSource instanceof DataBoundSource) ?
		(DataBoundSource)dataSource : null;
}

/**
 * Returns the IDE specific metadata object bound to the chart.
 */
protected Object getNativeMetaData() {
	DataBoundSource dbs = getDataBoundSource();
	return dbs != null ? dbs.getNativeMetaData() : null;
}

/**
 * Sets the value of the <code>SeriesColumns</code> indexed property.
 * Only columns that have
 * a numeric type can be potential series columns.
 * @param _seriesColumns containing all the series columns for this data source.
 * @see com.klg.jclass.chart.db.DataBindingConfigEditor#isNumerical
 */
public void setDataColumns(String[] seriesColumns) {
	DataBoundSource dbs = getDataBoundSource();
    if (dbs != null) {
		dbs.setDataColumns(seriesColumns);
    }
}

/**
 * Sets the value of the <code>PointLabelsColumn</code> property.
 * @param _pointLabelsColumn containing the name of the column to be bound as
 * the point labels column
 */
public void setPointLabelsColumn(String pointLabelsColumn) {
	DataBoundSource dbs = getDataBoundSource();
    if (dbs != null) {
		dbs.setPointLabelsColumn(pointLabelsColumn);
    }
}

/**
 * Sets the value of the <code>name</code> property.
 * The <code>name</code> property defines
 * the name of the data object.
 * @param s the name string
 */
public void setName(String s) {
	DataBoundSource dbs = getDataBoundSource();
    if (dbs != null) {
		dbs.setName(s);
    }
}

}	// DataChart


