/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ChartDataViewPropertyLoad.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:53 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.*;
import com.klg.jclass.chart.data.*;
import com.klg.jclass.chart.property.html.JCAppletAccessor;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.io.LoadProperties;
import com.klg.jclass.util.io.OutputDataProperties;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.io.Properties;
import com.klg.jclass.util.swing.JCSwingTypeConverter;

import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

/**
 * Responsible for loading properties of ChartDataView
 * objects using a specified PropertyAcessModel.  Not
 * really intended for public consumption
 *
 * @see com.klg.jclass.chart.ChartDataView
 * @see com.klg.jclass.chart.property.ChartDataViewPropertySave
 */
public class ChartDataViewPropertyLoad implements PropertyLoadModel {

public static final String[] formatPrefix = {
	"Plot",       // PLOT
	"Plot",       // SCATTER_PLOT
	"PolarRadar", // POLAR
	"PolarRadar", // RADAR
	"PolarRadar", // AREA_RADAR
	"Hilo",       // HILO
	"HLOC",       // HILO_OPEN_CLOSE
	"Candle",     // CANDLE
	"Area",       // AREA
	"Bar",        // BAR
	"Bar",        // STACKING_BAR
	"Pie",        // PIE
	"Area"        // STACKING_AREA
};

protected ChartDataView  dataView = null;
protected Hashtable      uniqueFormatNames = null;

/**
 * Constructor that intializes the unique format prefix names hash table
 */
public ChartDataViewPropertyLoad()
{
	// By putting the format prefix names in a hash table we get a unique
	// list of format prefix names.
	uniqueFormatNames = new Hashtable();
	for (int i = 0; i < formatPrefix.length; i++) {
		uniqueFormatNames.put(formatPrefix[i], new Integer(i));
	}
}

/**
 * Sets the source object ie the object on which to set properties.
 *
 * @param o Object that is loading properties
 */
public void setSource(Object o)
{
	if (o instanceof ChartDataView) {
		this.dataView = (ChartDataView)o;
	}
}

/**
 * Loads properties using the given property accessor object
 *
 * @param prop Property access object.
 * @param name Name of object to be loaded.
 * @throws JCIOException
 */
public void loadProperties(PropertyAccessModel prop, String name)
    throws JCIOException
{
	loadDataSource(prop, name);

	String pval = prop.getProperty(name + "inverted");
	dataView.setInverted(JCTypeConverter.toBoolean(pval,
												   dataView.isInverted()));

	pval = prop.getProperty(name + "visible");
	dataView.setVisible(JCTypeConverter.toBoolean(pval,
												  dataView.isVisible()));

	// make sure VisibleInLegend is done *after* visible, since
	// the two properties are coupled
	pval = prop.getProperty(name + "visibleInLegend");
	if (pval != null) {
		dataView.setVisibleInLegend(JCTypeConverter.toBoolean(pval,
											  dataView.isVisibleInLegend()));
		// since the value is explicitly specified, ensure that
		// it is marked as set, even if it is the current value
		dataView.setVisibleInLegendChanged(true);
	}

	pval = prop.getProperty(name + "autoLabel");
	dataView.setAutoLabel(JCTypeConverter.toBoolean(pval,
													dataView.getAutoLabel()));

	pval = prop.getProperty(name + "bufferPlotData");
	dataView.setBufferPlotData(JCTypeConverter.toBoolean(pval,
											 dataView.getBufferPlotData()));

	pval = prop.getProperty(name + "fastUpdate");
	dataView.setFastUpdate(JCTypeConverter.toBoolean(pval,
												     dataView.getFastUpdate()));

	pval = prop.getProperty(name + "drawFrontPlane");
	dataView.setDrawFrontPlane(JCTypeConverter.toBoolean(pval,
											   dataView.getDrawFrontPlane()));

	// Note this is here for backwards compatibility only as this property is
	// now part of the outlineStyle (loaded below)
	pval = prop.getProperty(name + "outlineColor");
	dataView.setOutlineColor(JCSwingTypeConverter.toColor(pval,
											   dataView.getOutlineColor()));

	pval = prop.getProperty(name + "holeValue");
	dataView.setHoleValue(JCTypeConverter.toDouble(pval,
											       dataView.getHoleValue()));

	pval = prop.getProperty(name + "name");
	dataView.setName(pval != null ? pval : dataView.getName());

    // Load the dataview's imageMapInfo, if set.
    ImageMapInfo info = new ImageMapInfo();
    PropertyLoadFactory.load(prop, info, name + "imageMapInfo.");
    if (!info.isEmpty()) {
        dataView.setImageMapInfo(info);
    }

	Object o = prop.getPropertyObject(name + "pointLabels");
	if (o != null) {
	    List l = null;
		if (o instanceof String) {
			pval = (String)o;
		    String[] list = JCTypeConverter.toStringList(pval);
		    l = Arrays.asList(list);
		}
		else if (o instanceof List) {
			l = (List)o;
		}
		if (l != null) {
		    dataView.setPointLabels(l);
		}
	}

	// The default x and y axis are already set by JCChartArea
	// This will handle any explicit axis assignment requests
	// Also provided to support existing format
	// For example : chart.data.xaxis.isInverted
	PropertyLoadFactory.load(prop, dataView.getXAxis(), name + "xaxis.");
	PropertyLoadFactory.load(prop, dataView.getYAxis(), name + "yaxis.");

	if (dataView.getParent() != null &&
		dataView.getParent().getChartArea() != null)
	{
		JCChartArea ca = dataView.getParent().getChartArea();

		// Get xaxis value for this data view
		pval = prop.getProperty(name + "xaxis");
		if (pval != null) {
			JCAxis ax = ca.findAxisByName(pval);
			if (ax != null) {
				dataView.setXAxis(ax);
			}
		}

		// Get yaxis value for this data view
		pval = prop.getProperty(name + "yaxis");
		if (pval != null) {
			JCAxis ax = ca.findAxisByName(pval);
			if (ax != null) {
				dataView.setYAxis(ax);
			}
		}
	}

	// Get DataViewSeries

	int numSeriesInData = dataView.getNumSeries();
	int numSeriesToLoad = numSeriesInData;
	pval = prop.getProperty(name + "lastSeriesIndex");
	if (pval != null) {
		numSeriesToLoad = (new Integer(pval)).intValue();
	}
	// load properties for series that exist in data
	for (int i = 0; i < numSeriesInData; i++) {
		PropertyLoadFactory.load(prop, dataView.getSeries(i),
								 name + "series" + (i+1) + ".");
	}

	// load chart styles for all series in load file, not just in data
	// so that chart styles will still be set if data is to be loaded
	// later
	List styleList = dataView.getChartStyle();
	int numStylesInView = 0;
	if (styleList != null) {
		numStylesInView = styleList.size();
	}
	for (int i = 0; i < numSeriesToLoad; i++) {
		JCChartStyle styleToSet;
		// get existing style or create new one if needed
		if (i < numStylesInView) {
			styleToSet = dataView.getChartStyle(i);
		} else {
			styleToSet = new JCChartStyle();
			dataView.setChartStyle(i, styleToSet);
		}
		// load properties into style
    	PropertyLoadFactory.load(prop, styleToSet,
								 name + "series" + (i+1) + ".");
	}

	// Load the dataview's outlineStyle
	PropertyLoadFactory.load(prop, dataView.getOutlineStyle(), name + "line.");

	// Do this next to last so special values due to chart type
	// take effect and don't get erased
	boolean isXML    = prop.getType().equals("XML");
	String[] strings = isXML ? JCChartEnumMappings.chartType_xml_strings
		                     : JCChartEnumMappings.chartType_strings;
	int[] values     = JCChartEnumMappings.chartType_values;
	pval = prop.getProperty(name + "chartType");
	if (pval != null) {
		dataView.setChartType(JCTypeConverter.toEnum(pval, "chart type",
									                 strings, values,
									                 dataView.getChartType()));
	}

	// If the PropertyAccessModel object has stored properties in a hash table
	// then we want to load properties for all the format types found in the
	// hashtable.  We use our own hashtable 'uniqueFormatNames' to get the
	// format names.
	if (prop instanceof JCBaseAccessor) {
	    JCBaseAccessor accessor = (JCBaseAccessor)prop;
		Enumeration e = uniqueFormatNames.keys();
		while (e.hasMoreElements()) {
			String format = (String)e.nextElement();
			if (accessor.searchKeysForPrefix(name + format)) {
				int type = ((Integer)uniqueFormatNames.get(format)).intValue();
				PropertyLoadFactory.load(accessor,
				                         dataView.getChartFormat(type),
							             name + format + ".");
			}
		}
	}
	else {
		// Load chart format data for current chart type only
	    PropertyLoadFactory.load(prop, dataView.getChartFormat(),
							     name +
								 formatPrefix[dataView.getChartType()] + ".");
	}
}

/**
 * Load the data source.
 *
 * When the data is accessed via file name, the code checks to see if the
 * property accessor implements FilePropertyAccessModel.  If so, the code tries
 * to get the property through the getFilePropertyStream() method so that the
 * file name is changed to reflect the correct path (i.e. via a ServletContext)
 *
 * @param prop Property access object.
 * @param name Name of object to be loaded.
 * @throws JCIOException
 */
protected void loadDataSource(PropertyAccessModel prop, String name)
    throws JCIOException
{
	ChartDataModel dataSource = null;
	name = name.substring(0, name.lastIndexOf("."));
	Object o = prop.getPropertyObject(name);
    LoadProperties loadProperties = prop.getLoadProperties();
    OutputDataProperties outProps = new OutputDataProperties();
    String type = "None";
    try {
        if (o == null) {
            // No data set, look for a dataFile property
            String fileName = prop.getProperty(name + "File");
            if (fileName != null) {
                type = prop.getProperty(name + "FileType");
                JCDataInterpreter interp = null;
                if (type != null && type.equalsIgnoreCase("Xml")) {
                    interp = new JCXMLDataInterpreter(prop.getLoadProperties());
                }
                else {
                    // Null interpreter means load text data
                    type = "Text";
                }
                dataSource = getDataFromName(prop, name, interp, fileName,
                                             outProps);
            }
        }
        else {
            // Data has been set already during property processing.  Thus it
            // must have been embedded.
            if (o instanceof String) {
                String s = (String)o;
                try {
                    dataSource = new JCStringDataSource(s);
                }
                catch (Exception e) {
                    throw new JCIOException(e.getMessage(), e);
                }
            }
            else if (o instanceof ChartDataModel) {
                dataSource = (ChartDataModel)o;
            }
            type = "Embed";
        }
    }
    catch (JCIOException jce) {
        if (!loadProperties.ignoreExternalResourceExceptions()) {
            throw jce;
        }
    }
    outProps.setSaveType(type);

	if (dataSource != null) {
		dataView.setDataSource(dataSource);
        dataView.setOutputDataProperties(outProps);
	}
}

protected ChartDataModel getDataFromName(PropertyAccessModel prop, String name,
                                         JCDataInterpreter interp,
                                         String fileName,
                                         OutputDataProperties outProps)
    throws JCIOException
{
    String charset = prop.getProperty(name + "FileCharset");
    charset = (charset == null) ? Properties.DEFAULT_CHARSET_NAME : charset;
    outProps.setFileCharset(charset);

    ChartDataModel dataSource = null;
    if (prop instanceof FilePropertyAccessModel) {
        FilePropertyAccessModel fm = (FilePropertyAccessModel)prop;
        InputStream is = fm.getFilePropertyStream(name, outProps);
        if (is != null) {
            try {
                dataSource = new JCInputStreamDataSource(is, interp, charset);
            }
            catch (Exception e) {
                throw new JCIOException(e.getMessage(), e);
            }
        }
    }
    else {  // Use an absolute path
        try {
            dataSource = new JCFileDataSource(fileName, interp, charset);
        }
        catch (Exception e) {
            // Try as an url if we can find an applet
            if (prop instanceof JCAppletAccessor) {
                try {
                    JCAppletAccessor a = (JCAppletAccessor)prop;
                    URL url = a.getApplet().getDocumentBase();
                    dataSource = new JCURLDataSource(url, fileName, interp,
                                                     charset);
                }
                catch (Exception e2) {
                    throw new JCIOException(e2.getMessage(), e2);
                }
            }
            else {
                throw new JCIOException(e.getMessage(), e);
            }
        }
        outProps.setPropertyName(fileName);
        outProps.setFileAccess(OutputDataProperties.ABSOLUTE);
    }
    return(dataSource);
}

}
