/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ChartDataViewPropertySave.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:53 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.*;
import com.klg.jclass.chart.data.ChartDataSourceUtil;
import com.klg.jclass.chart.data.XMLDataHandler;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.io.OutputDataProperties;
import com.klg.jclass.util.io.Properties;

import java.io.*;
import java.util.*;

/**
 * Responsible for saving properties of ChartDataView
 * objects to a destiation specified by PropertyPersistorModel.
 * Not really intended for public consumption.
 *
 * @see com.klg.jclass.chart.ChartDataView
 * @see com.klg.jclass.chart.property.ChartDataViewPropertyLoad
 */
public abstract class ChartDataViewPropertySave implements PropertySaveModel {

public static final String[] formatPrefix = {
	"Plot",       // PLOT
	"Plot",       // SCATTER_PLOT
	"PolarRadar", // POLAR
	"PolarRadar", // RADAR
	"PolarRadar", // AREA_RADAR
	"Hilo",       // HILO
	"HLOC",       // HILO_OPEN_CLOSE
	"Candle",     // CANDLE
	"Area",       // AREA
	"Bar",        // BAR
	"Bar",        // STACKING_BAR
	"Pie",        // PIE
	"Area"        // STACKING_AREA
};

protected static int dataFileNumber = 0;

protected ChartDataView  dataView = null;
protected ChartDataView  defaultDataView = null;
protected TreeMap        uniqueFormatNames = null;
protected int            increment = PropertySaveFactory.tabIncrement;

public ChartDataViewPropertySave()
{
	// By putting the format prefix names in a hash table we get a unique
	// list of format prefix names.
	uniqueFormatNames = new TreeMap();
	for (int i = 0; i < formatPrefix.length; i++) {
		uniqueFormatNames.put(formatPrefix[i], new Integer(i));
	}
}

/**
 * Set the source component whose properties are to be saved
 * @param o The new source component
 */
public void setSource(Object o) {
	if (o instanceof ChartDataView) {
		this.dataView = (ChartDataView)o;
	}
}

/**
 * Sets default object instance.  This object is typically used by implementors
 * to determine whether a property really needs to be saved.
 * @param def The new default object
 */
public void setDefault(Object def) {
	if (def instanceof ChartDataView) {
		this.defaultDataView = (ChartDataView)def;
	}
}

/**
 * Check whether are any properties different from their default.  If so, then
 * then this saver will have something to write
 * @param prop Property persistor object
 * @return Whether any properties are different from their default.
 */
public boolean checkProperties(PropertyPersistorModel prop)
{
	return(true);
}


/**
 * Saves HTML parameters from the current param source;
 * click the method name to see which properties are saved.<p>
 * Saves the following properties:
 *   <pre>
 *   &lt;data&gt;
 *   inverted (boolean)
 *   chartType (enum)
 *   visibleInLegend (boolean)
 *   visible (boolean)
 *   autoLabel (boolean)
 *   bufferPlotData (boolean)
 *   fastUpdate (boolean)
 *   drawFrontPlane (boolean)
 *   outlineColor (Color)
 *   holeValue (double)
 *   xaxis (JCAxis, using axis name)
 *   yaxis (JCAxis, using axis name)
 *   series (ChartDataViewSeries, see
 *   <a href="ChartDataViewSeriesPropertySave.html">ChartDataViewSeriesPropertySave</a>)
 *   format (Depends on chart type)
 *   </pre>
 *
 * @param prop Property persistor object
 * @param name Name of object to be saved
 * @param tab The tab indent for this group of properties
 * @throws JCIOException
 */
public void saveProperties(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
	if (dataView == null || defaultDataView == null) {
		System.out.println("FAILURE: No ChartDataView set ");
		return;
	}

	int newTab = prop.writeBegin("chart-data-view", tab);

	// Inverted
	if (dataView.isInverted() != defaultDataView.isInverted()) {
		prop.writeProperty(name, "inverted", newTab,
						   new Boolean(dataView.isInverted()));
	}

	// ChartType
	boolean isXML    = prop.getType().equals("XML");
	String[] strings = isXML ? JCChartEnumMappings.chartType_xml_strings
		                     : JCChartEnumMappings.chartType_strings;
	int[] values     = JCChartEnumMappings.chartType_values;
	if (dataView.getChartType() != defaultDataView.getChartType()) {
		prop.writeProperty(name, "chartType", newTab,
						   JCTypeConverter.fromEnum(dataView.getChartType(),
										            strings, values));
	}

	// Visible
	if (dataView.isVisible() != defaultDataView.isVisible()) {
		prop.writeProperty(name, "visible", newTab,
		                   new Boolean(dataView.isVisible()));
	}

	// VisibleInLegend -- this property is coupled with Visible, and
	// should only be written out if it was explicitly set by the user.
	if (dataView.hasVisibleInLegendChanged()) {
		prop.writeProperty(name, "visibleInLegend", newTab,
						   new Boolean(dataView.isVisibleInLegend()));
	}

	// AutoLabel
/* Don't write out the auto label property.  All of the labels created
   by it will be written out on their own.

	if (dataView.getAutoLabel() != defaultDataView.getAutoLabel()) {
		prop.writeProperty(name, "autoLabel",
						   new Boolean(dataView.getAutoLabel()));
	}
*/

	// BufferPlotData
	if (dataView.getBufferPlotData() != defaultDataView.getBufferPlotData()) {
		prop.writeProperty(name, "bufferPlotData", newTab,
						   new Boolean(dataView.getBufferPlotData()));
	}

	// FastUpdate
	if (dataView.getFastUpdate() != defaultDataView.getFastUpdate()) {
		prop.writeProperty(name, "fastUpdate", newTab,
						   new Boolean(dataView.getFastUpdate()));
	}

	// DrawFrontPlane
	if (dataView.getDrawFrontPlane() != defaultDataView.getDrawFrontPlane()) {
		prop.writeProperty(name, "drawFrontPlane", newTab,
						   new Boolean(dataView.getDrawFrontPlane()));
	}

	// HoleValue
    double holeValue = dataView.getHoleValue();
    if (holeValue != defaultDataView.getHoleValue()) {
		prop.writeProperty(name, "holeValue", newTab,
						   XMLDataHandler.doubleToString(holeValue, holeValue,
                                                         false));
	}

	// Name
	if (dataView.getName() == null) {
	    dataView.setName(name.substring(0, name.lastIndexOf(".")));
	}
	prop.writeProperty(name, "name", newTab,
	                   prop.expandText(dataView.getName()));

	// Note: actual axis saving occurs in JCChartAreaPropertySave.

	// Save X axis name
	JCAxis xAxis = dataView.getXAxis();
	if (xAxis != null && !xAxis.getName().equals("xaxis")) {
		prop.writeProperty(name, "xaxis", newTab,
		                   prop.expandText(xAxis.getName()));
	}

	// Save Y axis name
	JCAxis yAxis = dataView.getYAxis();
	if (yAxis != null && !yAxis.getName().equals("yaxis")) {
		prop.writeProperty(name, "yaxis", newTab,
		                   prop.expandText(yAxis.getName()));
	}
	prop.writeEnd(null, tab, true, true);

	saveDataSource(prop, name, tab + increment);

    // Save ImageMapInfo properties
    PropertySaveFactory.save(prop, dataView.getImageMapInfo(),
                             new ImageMapInfo(), name + "imageMapInfo.", tab);

	// Save chart format data for all existing formats
	Set set = uniqueFormatNames.keySet();
	Iterator iterator = set.iterator();
	while (iterator.hasNext()) {
	    String format = (String)iterator.next();
		int type = ((Integer)uniqueFormatNames.get(format)).intValue();
		JCChartTypeFormat fmt = dataView.getExistingChartFormat(type);
	    if (fmt != null) {
		    Class cz = fmt.getClass();
		    Object dflt = defaultDataView.getChartFormat(type);
		    try {
			    dflt = cz.newInstance();
			}
		    catch (Exception ex) {}
		    PropertySaveFactory.save(prop, fmt, dflt, name + format + ".", tab);
	    }
	}

	// Save the outlineStyle
    PropertySaveFactory.save(prop, dataView.getOutlineStyle(),
	                         defaultDataView.getOutlineStyle(),
							 name + "line.", tab);

	// Save PointLabels.  Note that PointLabels may be saved out in the data.
	// If they are (and they are non-empty) then don't save them here.
	ChartDataModel data = dataView.getDataSource();
	String[] pointLabels = null;
	if (data instanceof LabelledChartDataModel) {
		LabelledChartDataModel lData = (LabelledChartDataModel)data;
		pointLabels = lData.getPointLabels();
	}
	List pointLabelsList = dataView.getPointLabels();
	if (!ChartDataSourceUtil.checkPointLabels(pointLabels) &&
	    pointLabelsList != null)
	{
		ListIterator li = pointLabelsList.listIterator();
		boolean pointLabelFound = false;
		while (li.hasNext()) {
			String label = (String)li.next();
			if (label != null && label.length() > 0) {
				pointLabelFound = true;
				break;
			}
		}
		if (pointLabelFound) {
			li = pointLabelsList.listIterator();
		    if (prop.getType().equals("HTML")) {
	            StringBuffer sb = new StringBuffer();
			    while (li.hasNext()) {
				    sb.append(prop.expandText((String)li.next()));
				    if (li.hasNext()) {
					    sb.append(",");
					}
				}
			    prop.writeProperty(name, "pointLabels", tab, new String(sb));
			}
		    else {
			    while (li.hasNext()) {
				    newTab = prop.writeBegin("point-label", tab + increment);
				    prop.writeEnd(null, tab, false, true);
				    prop.writeString(prop.expandText((String)li.next()), 0);
				    prop.writeEnd("point-label", 0, true, false);
				}
			}
		}
	}

	// Save DataviewSeries
	int numSeries = dataView.getNumSeries();
	for (int i = 0; i < numSeries; i++) {
		PropertySaveFactory.save(prop, dataView.getSeries(i),
								 new ChartDataViewSeries(),
								 name + "series" + (i+1) + ".", tab);
	}

	// handle any chart styles that exist beyond on the end of the
	// series list -- (chart styles are included in the series tags,
 	// but are not actually stored in the series themselves -- this
	// means that there can be more styles than series)
	List chartStyles = dataView.getChartStyle();
	if (chartStyles != null) {
		int numStyles = chartStyles.size();
		int seriesTab = tab + PropertySaveFactory.tabIncrement;
		for (int i = numSeries; i < numStyles; i++) {
    		prop.writeBegin("chart-data-view-series", seriesTab);
    		prop.writeEnd(null, seriesTab, true, true);
    		// Save chart style
    		PropertySaveFactory.save(prop, chartStyles.get(i),
							new JCChartStyle(),
							name + "series" + (i+1) + ".", seriesTab);
    		prop.writeEnd("chart-data-view-series", seriesTab, true, false);
		}
	}

	prop.writeEnd("chart-data-view", tab, true, false);
}

protected void saveDataSource(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
    OutputDataProperties outProps = dataView.getOutputDataProperties();
    if (outProps == null) {
        // By default, embed data (no files are generated)
        outProps = new OutputDataProperties(null, null, null,
                                            OutputDataProperties.DEFAULT_ACCESS);
    }
    int saveType = outProps.getSaveTypeAsEnum();
    ChartDataModel data = dataView.getDataSource();
    if (data == null || saveType == OutputDataProperties.NO_DATA) {
        return;
    }

    if (saveType == OutputDataProperties.EMBED_DATA) {
        saveEmbededData(prop, outProps, name, tab, data);
        return;
    }

    // Save out file if its name is non-null
	String outputFileName = outProps.getOutputFileName();
    String fileName = (outputFileName == null) ? null : outputFileName.trim();
    if (fileName != null && fileName.length() > 0) {
        String ext = (saveType == OutputDataProperties.DATA_FILE_TEXT) ? ".dat"
                                                                       : ".xml";
        // Allow user to use ".txt" endings for data files
        if (saveType == OutputDataProperties.DATA_FILE_TEXT &&
            fileName.endsWith(".txt"))
        {
            ext = ".txt";
        }

        if (!fileName.endsWith(ext)) {
            fileName = fileName + ext;
        }

        OutputStreamWriter writer = null;
        try {
            FileOutputStream fos = new FileOutputStream(fileName);
            DataOutputStream dos = new DataOutputStream(fos);
            String charSet = outProps.getFileCharset();
            if (charSet != null) {
                try {
                    writer = new OutputStreamWriter(dos, charSet);
                }
                catch (UnsupportedEncodingException uee) {
                    if (!charSet.equals(Properties.DEFAULT_CHARSET_NAME)) {
                        throw uee;
                    }
                }
            }
            if (writer == null) {
                writer = new OutputStreamWriter(dos);
            }

            switch (saveType) {
            case OutputDataProperties.DATA_FILE_TEXT:
                ChartDataSourceUtil.writeDataSource(data, writer, true);
                break;
            case OutputDataProperties.DATA_FILE_XML:
                XMLDataHandler handler = new XMLDataHandler(null);
                handler.writeDataToXML(data, writer, 0, increment, true);
            }
        }
        catch (IOException e) {
            if (!outProps.ignoreExternalResourceExceptions()) {
                throw new JCIOException(e.getMessage(), e);
            }
        }
        finally {
            try {
                writer.close();
            }
            catch (Exception e) {}
        }
    }
    saveDataFileProperties(prop, name, tab, outProps);
}

protected abstract void saveEmbededData(PropertyPersistorModel prop,
                                        OutputDataProperties outProps,
                                        String name, int tab,
										ChartDataModel data)
    throws JCIOException;

protected abstract void saveDataFileProperties(PropertyPersistorModel prop,
                                               String name, int tab,
											   OutputDataProperties outProps);

}
