/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ComponentPropertySave.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:53 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.JCChart;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.swing.JCSwingTypeConverter;

import javax.swing.*;
import javax.swing.border.*;
import java.awt.*;

/**
 * Responsible for saving properties of Component
 * objects to a destination specified by PropertyPersistorModel.
 * This is a base class extended by other chart property
 * savers, like JCChartAreaPropertySave, JCChartPropertySave,
 * JCLegendPropertySave and JLabelPropertySave.
 * Not really intended for public consumption.
 *
 * @see java.awt.Component
 * @see com.klg.jclass.chart.property.ComponentPropertyLoad
 */
public class ComponentPropertySave implements PropertySaveModel {

protected Component component        = null;
protected Component defaultComponent = null;

protected Color     compBackground;
protected Color     defBackground;
protected Color     compForeground;
protected Color     defForeground;
protected Font      compFont;
protected Font      defFont;
protected boolean   compIsOpaque;
protected boolean   defIsOpaque;
protected boolean   compIsVisible;
protected boolean   defIsVisible;
protected Border    compBorder;
protected Border    defBorder;

/**
 * Set the source component whose properties are to be saved
 * @param o The new source component
 */
public void setSource(Object o)
{
	if (o instanceof Component) {
		this.component = (Component)o;
	}
}

/**
 * Sets default object instance.  This object is typically used by implementors
 * to determine whether a property really needs to be saved.
 * @param def The new default object
 */
public void setDefault(Object def)
{
	if (def instanceof Component) {
		this.defaultComponent = (Component)def;
	}
}

/**
 * Check whether are any properties different from their default.  If so, then
 * then this saver will have something to write
 * @param prop Property persistor object
 * @return Whether any properties are different from their default.
 */
public boolean checkProperties(PropertyPersistorModel prop)
{
	if (component == null || defaultComponent == null) {
		return(false);
	}

	compBackground = component.getBackground();
	defBackground  = defaultComponent.getBackground();
	compForeground = component.getForeground();
	defForeground  = defaultComponent.getForeground();
	compFont       = component.getFont();
	defFont        = defaultComponent.getFont();
	compIsOpaque   = component.isOpaque();
	defIsOpaque    = defaultComponent.isOpaque();
	compIsVisible  = component.isVisible();
	defIsVisible   = defaultComponent.isVisible();
	compBorder     = ((JComponent)component).getBorder();
	defBorder      = ((JComponent)defaultComponent).getBorder();

	if (!compBackground.equals(defBackground) ||
	    !compForeground.equals(defForeground) ||
		!compFont.equals(defFont)             ||
		compIsOpaque != defIsOpaque           ||
		compIsVisible != defIsVisible         ||
		compBorder != defBorder)
	{
		return(true);
	}
	return(false);
}

/**
 * Saves object properties using the specified persistance object.
 *
 * @param prop Property persistor object
 * @param name Name of object to be saved
 * @param tab The tab indent for this group of properties
 */
public void saveProperties(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
	if (component == null || defaultComponent == null) {
		System.out.println("FAILURE: No component set ");
		return;
	}

	// Check if something to do.
	if (!checkProperties(prop)) {
		return;
	}

	int newTab = prop.writeBegin("component", tab);

	boolean writeBackground = false;
	if (!compBackground.equals(defBackground)) {
		writeBackground = true;
		writeUI(prop, name, "background", newTab,
				JCSwingTypeConverter.fromColor(compBackground));
	}

	// Write opaque if it has been changed by the user OR if the
	// background color has been changed but opaque has been set to false.
	if (compIsOpaque != defIsOpaque || (writeBackground && !compIsOpaque)) {
		prop.writeProperty(name, "opaque", newTab,
		                   new Boolean(compIsOpaque));
	}

	if (!compForeground.equals(defForeground)) {
		writeUI(prop, name, "foreground", newTab,
				JCSwingTypeConverter.fromColor(compForeground));
	}

	if (!compFont.equals(defFont)) {
		writeUI(prop, name, "font", newTab,
				JCSwingTypeConverter.fromFont(compFont));
	}

	if (compIsVisible != defIsVisible) {
		prop.writeProperty(name, "visible", newTab,
		                   new Boolean(compIsVisible));
	}

	if (compBorder != defBorder) {
		prop.writeEnd(null, tab, true, true);
		writeBorder(prop, name, tab + PropertySaveFactory.tabIncrement);
		prop.writeEnd("component", tab, true, false);
	}
	else {
	    prop.writeEnd(null, tab, true, false);
	}
}

public static void writeInsets(PropertyPersistorModel prop, String name,
                               int tab, Insets insets)
{
    int newTab = prop.writeBegin("insets", tab);
	prop.writeProperty(name, "left", newTab, new Integer(insets.left));
	prop.writeProperty(name, "right", newTab, new Integer(insets.right));
	prop.writeProperty(name, "top", newTab, new Integer(insets.top));
	prop.writeProperty(name, "bottom", newTab, new Integer(insets.bottom));
	prop.writeEnd(null, tab, true, false);
}

/**
 * Write the border property
 * @param prop The property persistance object
 * @param name The name of the object to be saved
 * @param tab The tab indent for this group of properties
 */
public void writeBorder(PropertyPersistorModel prop, String name, int tab)
{
	if (compBorder == null ||
	    compBorder instanceof javax.swing.plaf.UIResource)
	{
		return;
	}

	int     newTab;
	String  type;
	Color   color, sColor, hColor;

	if (prop.getType().equals("HTML")) {
		String border = JCSwingTypeConverter.fromBorder(compBorder, component);
		if (border != null) {
		    prop.writeProperty(name, "border", tab, border);
		}
	}
	else {
		if (compBorder instanceof MatteBorder) {
			MatteBorder matte = (MatteBorder)compBorder;
			newTab = prop.writeBegin("matte-border", tab);

			color = JCSwingTypeConverter.getMatteBorderColor(matte, component);
			if (color == null) {
				color = Color.black;
			}
			prop.writeProperty(name, "color", newTab,
			                   JCSwingTypeConverter.fromColor(color));

			prop.writeEnd(null, tab, true, true);
			writeInsets(prop, name, tab + PropertySaveFactory.tabIncrement,
			            matte.getBorderInsets(component));
			prop.writeEnd("matte-border", tab, true, false);
		}
        else if (compBorder instanceof EmptyBorder) {
			EmptyBorder empty = (EmptyBorder)compBorder;
			newTab = prop.writeBegin("empty-border", tab);
			prop.writeEnd(null, tab, true, true);
			writeInsets(prop, name, tab + PropertySaveFactory.tabIncrement,
			            empty.getBorderInsets(component));
			prop.writeEnd("empty-border", tab, true, false);
        }
		else if (compBorder instanceof BevelBorder ||
		         compBorder instanceof SoftBevelBorder)
		{
			BevelBorder bevel = (BevelBorder)compBorder;
			newTab = prop.writeBegin("bevel-border", tab);

			prop.writeProperty(name, "soft", newTab,
			                   new Boolean(bevel instanceof SoftBevelBorder));

			type = (bevel.getBevelType() == BevelBorder.RAISED) ? "Raised"
				                                                : "Lowered";
			prop.writeProperty(name, "type", newTab, type);

            // Don't write out highlight & shadow if they haven't been specified
			hColor = bevel.getHighlightInnerColor();
            sColor = bevel.getShadowOuterColor();
            if (hColor != null || sColor != null) {
                if (hColor == null) {
                    hColor = bevel.getHighlightInnerColor(component);
                }
                prop.writeProperty(name, "highlightColor", newTab,
                                   JCSwingTypeConverter.fromColor(hColor));

                if (sColor == null) {
                    sColor = bevel.getShadowOuterColor(component);
                }
                prop.writeProperty(name, "shadowColor", newTab,
                                   JCSwingTypeConverter.fromColor(sColor));

            }
			prop.writeEnd(null, tab, true, false);
		}
		else if (compBorder instanceof EtchedBorder) {
			EtchedBorder etched = (EtchedBorder)compBorder;
			newTab = prop.writeBegin("etched-border", tab);

			type = (etched.getEtchType() == EtchedBorder.RAISED) ? "Raised"
				                                                 : "Lowered";
			prop.writeProperty(name, "type", newTab, type);

            hColor = etched.getHighlightColor();
			if (hColor != null) {
			    prop.writeProperty(name, "highlightColor", newTab,
				                   JCSwingTypeConverter.fromColor(hColor));
			}
            sColor = etched.getShadowColor();
			if (sColor != null) {
			    prop.writeProperty(name, "shadowColor", newTab,
				                   JCSwingTypeConverter.fromColor(sColor));
			}
			prop.writeEnd(null, tab, true, false);
		}
		else if (compBorder instanceof LineBorder) {
			LineBorder line = (LineBorder)compBorder;
			newTab = prop.writeBegin("line-border", tab);

			color = line.getLineColor();
			if (color == null) {
				color = Color.black;
			}
			prop.writeProperty(name, "color", newTab,
			                   JCSwingTypeConverter.fromColor(color));

			if (line.getThickness() > 1) {
				prop.writeProperty(name, "thickness", newTab,
				                   new Integer(line.getThickness()));
			}
			prop.writeEnd(null, tab, true, false);
		}
		else if (compBorder instanceof TitledBorder) {
			TitledBorder titled = (TitledBorder)compBorder;
			String title = titled.getTitle();
			if (title == null || title.length() == 0) {
				return;
			}

			newTab = prop.writeBegin("titled-border", tab);
			prop.writeProperty(name, "title", newTab, prop.expandText(title));

			int position = titled.getTitlePosition();
            String[] strings = JCSwingTypeConverter.positionStrings;
            int [] values = JCSwingTypeConverter.positionValues;
			if (position != TitledBorder.DEFAULT_POSITION) {
				String posString = JCTypeConverter.fromEnum(position, strings,
                                                            values);
				prop.writeProperty(name, "titlePosition", newTab, posString);
			}

			int justification = titled.getTitleJustification();
            strings = JCSwingTypeConverter.justificationStrings;
            values = JCSwingTypeConverter.justificationValues;
			if (justification != TitledBorder.DEFAULT_JUSTIFICATION) {
				String justString = JCTypeConverter.fromEnum(justification,
                                                             strings, values);
				prop.writeProperty(name, "titleJustification", newTab,
				                   justString);
			}

			color = titled.getTitleColor();
			if (color != null) {
			    prop.writeProperty(name, "color", newTab,
			                       JCSwingTypeConverter.fromColor(color));
			}

			Font font = titled.getTitleFont();
			if (font != null) {
				prop.writeProperty(name, "font", newTab,
				                   JCSwingTypeConverter.fromFont(font));
			}
			prop.writeEnd(null, tab, true, false);
		}
	}
}

/**
 * Use the property persistance object to write the property
 * @param prop The property persistance object
 * @param name The name of the object to be saved
 * @param property The property string name
 * @param tab The tab indent for this group of properties
 * @param value Actual object to persist.
 */
protected static void writeUI(PropertyPersistorModel prop, String name,
					          String property, int tab, Object value)
{
	if (value == null || value instanceof javax.swing.plaf.UIResource) {
		return;
	}
	prop.writeProperty(name, property, tab, value);
}

public boolean checkLayoutHints(JCChart chart, JComponent child)
{
	Rectangle hints = chart.getLayoutHints(child);
	if (hints == null) {
		return(false);
	}

	// Check if anything to write
	if (hints.x == Integer.MAX_VALUE     &&
	    hints.y == Integer.MAX_VALUE     &&
	    hints.width == Integer.MAX_VALUE &&
		hints.height == Integer.MAX_VALUE)
	{
	    return(false);
	}
	return(true);
}

/**
 * Saves the layout hints for the given child using the PropertyPersistorModel
 * and the chart
 * @param prop Property persistor object
 * @param chart The chart from which to get the layout hints
 * @param child The child component
 * @param name Name of object to be saved
 * @param tab The tab indent for this group of properties
 */
public void saveLayoutHints(PropertyPersistorModel prop, JCChart chart,
                            JComponent child, String name, int tab)
{
	if (!checkLayoutHints(chart, child)) {
		return;
	}

	Rectangle hints = chart.getLayoutHints(child);
	if (hints != null) {
	    int newTab = prop.writeBegin("layout-hints", tab);
		if (hints.x != Integer.MAX_VALUE) {
			prop.writeProperty(name, "x", newTab,
			                   new Integer(hints.x));
		}
		if (hints.y != Integer.MAX_VALUE) {
			prop.writeProperty(name, "y", newTab,
			                   new Integer(hints.y));
		}
		if (hints.width != Integer.MAX_VALUE) {
			prop.writeProperty(name, "width", newTab,
			                   new Integer(hints.width));
		}
		if (hints.height != Integer.MAX_VALUE) {
			prop.writeProperty(name, "height", newTab,
			                   new Integer(hints.height));
		}
		prop.writeEnd(null, tab, true, false);
	}
}

}

