/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCAxisPropertySave.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:53 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.JCAxis;
import com.klg.jclass.chart.JCGrid;
import com.klg.jclass.chart.JCLineStyle;
import com.klg.jclass.chart.JCAnno;
import com.klg.jclass.chart.JCChartEnumMappings;
import com.klg.jclass.chart.JCChartUtil;
import com.klg.jclass.chart.JCValueLabel;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.io.JCIOException;

import java.util.Date;
import java.util.List;
import java.util.TimeZone;

/**
 * Responsible for saving properties of <code>JCAxis</code>
 * objects to a destiation specified by <code>PropertyPersistorModel</code>.
 * Not intended for public use.
 *
 * @see com.klg.jclass.chart.JCAxis
 * @see com.klg.jclass.chart.property.JCAxisPropertyLoad
 */
public abstract class JCAxisPropertySave extends ChartInteriorRegionPropertySave {

protected JCAxis axis = null;
protected JCAxis defaultAxis = null;

/**
 * Sets the source component whose properties are to be saved.
 * @param o the new source component
 */
public void setSource(Object o)
{
    super.setSource(o);
    if (o instanceof JCAxis) {
        this.axis = (JCAxis)o;
    }
}

/**
 * Sets default object instance.  This object is typically used by implementors
 * to determine whether a property really needs to be saved.
 * @param def the new default object
 */
public void setDefault(Object def)
{
    super.setDefault(def);
    if (def instanceof JCAxis) {
        this.defaultAxis = (JCAxis)def;
    }
}

/**
 * Checks whether there are any properties different from their default.  If so,
 * then this saver has information to write.
 * @return whether any properties are different from their default
 */
public boolean checkProperties()
{
    return(true);
}

/**
 * Returns true if all attributes of the JCGrid's JCLineStyle
 * have default values.
 */
protected boolean isDefaultGridLineStyle(JCGrid cg) {
    boolean fDefault = true;
    if (cg != null) {
        JCLineStyle ls = cg.getGridStyle().getLineStyle();
        fDefault = JCGrid.equivalentGridLineStyles(ls,
                        JCGrid.makeDefaultLineStyle(axis), axis);
    }
    return fDefault;
}

/**
 * For internal use only.
 * Returns true if any of the JCGrid properties are different from
 * their defaults.
 */
protected boolean checkGridProperties(JCGrid cg) {
    if (cg == null) {
        return false;
    }
    return (!cg.getStartValueIsDefault() ||
            !cg.getStopValueIsDefault()  ||
            !cg.getIncrementIsDefault()  ||
            !isDefaultGridLineStyle(cg));
}

/**
 * Returns true if any of the JCAnno properties are different
 * from their defaults.
 */
protected boolean checkAnnoProperties(JCAnno anno) {
    if (anno == null) {
        return false;
    }
    boolean fDefAxisIncrement = true;
    if (anno.getType() == JCAnno.DEFAULT_LABELS) {
        fDefAxisIncrement = axis.getNumSpacingIsDefault();
    }
    else if (anno.getType() == JCAnno.DEFAULT_TICKS) {
        fDefAxisIncrement = axis.getTickSpacingIsDefault();
    }
    return (!anno.getStartValueIsDefault()                                  ||
            !anno.getStopValueIsDefault()                                   ||
            !(anno.getIncrementIsDefault() && fDefAxisIncrement)            ||
            !(anno.getPrecisionIsDefault() && axis.getPrecisionIsDefault()) ||
            !anno.getLabelExtentIsDefault()                                 ||
            !anno.getInnerExtentIsDefault()                                 ||
            !anno.getOuterExtentIsDefault()                                 ||
            !anno.getDrawLabelsIsDefault()                                  ||
            !anno.getDrawTicks()                                            ||
            !anno.getTickColor().equals(axis.getForeground())               ||
            !anno.getLabelColor().equals(axis.getForeground()));
}

/**
 * Saves HTML parameters from the current param source;
 * click the method name to see which properties are saved.<p>
 * Saves the following properties:
 *   <pre>
 *   vertical (boolean)
 *   logarithmic (boolean)
 *   annotationRotation (enum)
 *   annotationMethod (enum)
 *   placement (enum)
 *   placementAxis (JCAxis, by string)
 *   reversed (boolean)
 *   placementLocation (double)
 *   gap (integer)
 *   timeBase (Date)
 *   timeUnit (enum)
 *   timeFormat (string)
 *   timeZone (string)
 *   min (double)
 *   max (double)
 *   precision (integer)
 *   tickSpacing (double)
 *   numSpacing (double)
 *   origin (double)
 *   originPlacement (enum)
 *   grid.visible (boolean)
 *   grid.spacing (double)
 *   grid.&lt;style&gt; (JCLineStyle, see JCLineStylePropertySave)
 *   valueLabels
 *   editable
 *   title (JCAxisTitle, see JCAxisTitlePropertySave)
 *   formula (JCAxisFormula, see JCAxisFormulaPropertySave)
 *   </pre>
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 * @throws JCIOException
 */
public void saveProperties(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
    if (axis == null || defaultAxis == null) {
        System.out.println("FAILURE: No axis set");
        return;
    }

    int newTab = prop.writeBegin("axis", tab);
    boolean isXAxis = name.startsWith("x");
    saveTypeAndName(prop, name, isXAxis, newTab);

    // Set vertical
    if (axis.isVertical() != defaultAxis.isVertical()) {
        prop.writeProperty(name, "vertical", newTab,
                           new Boolean(axis.isVertical()));
    }

    // Set logarithmic
    if (axis.isLogarithmic() != defaultAxis.isLogarithmic()) {
        prop.writeProperty(name, "logarithmic", newTab,
                           new Boolean(axis.isLogarithmic()));
    }

    // Set annotationRotation
    boolean isXML    = prop.getType().equals("XML");
    String[] strings = isXML ? JCChartEnumMappings.annRotation_xml_strings
                             : JCChartEnumMappings.annRotation_strings;
    int[] values     = JCChartEnumMappings.annRotation_values;
    if (axis.getAnnotationRotation() != defaultAxis.getAnnotationRotation()) {
        prop.writeProperty(name, "annotationRotation", newTab,
                           JCTypeConverter.fromEnum(axis.getAnnotationRotation(),
                                                    strings, values));
    }

    // Set annotationRotationAngle
    double angle = axis.getAnnotationRotationAngle(JCChartUtil.DEGREES);
    if (angle != defaultAxis.getAnnotationRotationAngle(JCChartUtil.DEGREES)) {
        prop.writeProperty(name, "annotationRotationAngle", newTab,
                           new Double(angle));
    }

    int annoMethod = axis.getAnnotationMethod();
    // if there is a label generator attached to the axis, get the
    // generated labels as fixed value labels
    if (axis.getLabelGenerator() != null) {
        annoMethod = JCAxis.VALUE_LABELS;
    }
    // Set annotationMethod
    strings = JCChartEnumMappings.annotationMethod_strings;
    values  = JCChartEnumMappings.annotationMethod_values;
    if (annoMethod != defaultAxis.getAnnotationMethod()) {
        prop.writeProperty(name, "annotationMethod", newTab,
                           JCTypeConverter.fromEnum(annoMethod, strings,
                                                    values));
    }

    // Set placement
    strings = JCChartEnumMappings.placement_strings;
    values  = JCChartEnumMappings.placement_values;
    if (!axis.getPlacementIsDefault()) {
        prop.writeProperty(name, "placement", newTab,
                           JCTypeConverter.fromEnum(axis.getPlacement(),
                                                    strings, values));
    }

    // Set placementAxis
    JCAxis pAxis = axis.getPlacementAxis();
    if (pAxis != null) {
        prop.writeProperty(name, "placementAxis", newTab,
                           prop.expandText(pAxis.getName()));
    }

    // Set reversed
    if (axis.isReversed() != defaultAxis.isReversed()) {
        prop.writeProperty(name, "reversed", newTab,
                           new Boolean(axis.isReversed()));
    }

    // Set placementLocation
    if (axis.getPlacementLocation() != defaultAxis.getPlacementLocation()) {
        prop.writeProperty(name, "placementLocation", newTab,
                           new Double(axis.getPlacementLocation()));
    }

    // Set gap
    if (axis.getGap() != defaultAxis.getGap()) {
        prop.writeProperty(name, "gap", newTab, new Integer(axis.getGap()));
    }

    // Set timeBase
    if (axis.isTimeBaseSet) {
        Date tb = axis.getTimeBase();
        String dateString = JCTypeConverter.fromDate(tb, "" + tb);
        prop.writeProperty(name, "timeBase", newTab,
                           prop.expandText(dateString));
    }

    // Set timeUnit
    strings        = JCChartEnumMappings.timeUnit_strings;
    long[] lValues = JCChartEnumMappings.timeUnit_values;
    if (axis.getTimeUnit() != defaultAxis.getTimeUnit()) {
        prop.writeProperty(name, "timeUnit", newTab,
                           JCTypeConverter.fromEnum(axis.getTimeUnit(),
                                                    strings, lValues));
    }

    // Set timeFormat
    if (!axis.getTimeFormatIsDefault()) {
        prop.writeProperty(name, "timeFormat", newTab,
                           prop.expandText(axis.getTimeFormat()));
    }

    // Set timeZone
    TimeZone tz = axis.getTimeZone();
   if (!tz.getID().equals(defaultAxis.getTimeZone().getID())) {
        prop.writeProperty(name, "timeZone", newTab, tz.getID());
    }

    // Set min
    if (!axis.getMinIsDefault()) {
        prop.writeProperty(name, "min", newTab, new Double(axis.getMin()));
    }

    // Set max
    if (!axis.getMaxIsDefault()) {
        prop.writeProperty(name, "max", newTab, new Double(axis.getMax()));
    }

    // Set precision
//    if (!axis.getPrecisionIsDefault()) {
//        prop.writeProperty(name, "precision", newTab,
//                           new Integer(axis.getPrecision()));
//    }

    // Set tickSpacing
//    if (!axis.getTickSpacingIsDefault()) {
//        prop.writeProperty(name, "tickSpacing", newTab,
//                           new Double(axis.getTickSpacing()));
//    }

    // Set numSpacing
//    if (!axis.getNumSpacingIsDefault()) {
//        prop.writeProperty(name, "numSpacing", newTab,
//                           new Double(axis.getNumSpacing()));
//    }

    // Set origin
    if (!axis.getOriginIsDefault()) {
        prop.writeProperty(name, "origin", newTab,
                           new Double(axis.getOrigin()));
    }

    // Set originPlacement
    strings = JCChartEnumMappings.originPlacement_strings;
    values  = JCChartEnumMappings.originPlacement_values;
    if (!axis.getOriginPlacementIsDefault()) {
        prop.writeProperty(name, "originPlacement", newTab,
                           JCTypeConverter.fromEnum(axis.getOriginPlacement(),
                                                    strings, values));
    }

    // Set editable
    if (axis.isEditable() != defaultAxis.isEditable()) {
        prop.writeProperty(name, "editable", newTab,
                           new Boolean(axis.isEditable()));
    }

    // Save JCAxis.gridVisible and JCAxis.gridDefault
    saveAxisGridProperties(prop, name, newTab);

    // Set useDefaultLabels and useDefaultTicks
    if (isNonDefaultAnnotation()) {
        // Write useDefaultLabels=false if a DEFAULT_LABELS tick doesn't
        // exist or useDefaultLabels=true if we have an unmodified
        // DEFAULT_LABELS tick.
        if (axis.getDefaultLabels() == null ||
            isDefaultAnno(axis.getDefaultLabels()))
        {
            boolean fDefaultLabels = (axis.getDefaultLabels() != null);
            prop.writeProperty(name, "useDefaultLabels", newTab,
                               new Boolean(fDefaultLabels));
        }
        // Write useDefaultTicks=false if a DEFAULT_TICKS JCAnno doesn't
        // exist or useDefaultTicks=true if we have an unmodified
        // DEFAULT_TICKS JCAnno.
        if (axis.getDefaultTicks() == null ||
            isDefaultAnno(axis.getDefaultTicks()))
        {
            boolean fDefaultTicks = (axis.getDefaultTicks() != null);
            prop.writeProperty(name, "useDefaultTicks", newTab,
                               new Boolean(fDefaultTicks));
        }
    }

    // Set useAnnoTicks
    if (axis.getAnnotationMethod() != JCAxis.VALUE) {
        boolean useAnnoTicks = !axis.getTickSpacingIsDefault();
        prop.writeProperty(name, "useAnnoTicks", newTab,
                               new Boolean(useAnnoTicks));
    }

    if (isNonDefaultGrid()) {
        // Write useDefaultGrid=false if a default grid doesn't
        // exist or useDefaultGrid=true if we have an unmodified
        // default grid in addition to a modified grid.
        boolean fDefaultGrid = (axis.getDefaultGrid() != null);
        prop.writeProperty(name, "useDefaultGrid", newTab,
                           new Boolean(fDefaultGrid));
    }

    prop.writeEnd(null, tab, true, true);

    // Check if ChartInteriorRegion properties need to be saved
    if (super.checkProperties(prop)) {
        super.saveProperties(prop, name,
                             tab + PropertySaveFactory.tabIncrement);
    }

    PropertySaveFactory.save(prop, axis.getTitle(), defaultAxis.getTitle(),
                             name + "title.", tab);
    PropertySaveFactory.save(prop, axis.getFormula(), defaultAxis.getFormula(),
                             name + "formula.", tab);

    // If there is a label generator attached to the axis, get the
    // generated labels as fixed value labels.  Otherwise, get user
    // set value labels as per usual.
    JCValueLabel[] vl;
    if (axis.getLabelGenerator() != null) {
        vl = axis.getGeneratedValueLabels();
    }
    else {
        vl = axis.getValueLabels();
    }
    saveValueLabels(prop, name, tab + PropertySaveFactory.tabIncrement, vl);

    saveGridProperties(prop, name, tab, newTab);

    saveAnnoProperties(prop, name, tab, newTab);

    prop.writeEnd("axis", tab, true, false);
}

private boolean isDefaultAnno(JCAnno tickObj) {
    return (!checkAnnoProperties(tickObj));
}

/**
 * Returns false if the axis annoList contains one unmodified DEFAULT_LABELS
 * anno object and one unmodified DEFAULT_TICKS anno object.  Returns true
 * in all other cases.
 */
private boolean isNonDefaultAnnotation() {
    List annoList = axis.getAnnotationList();
    return (annoList == null ||
            annoList.size() != 2 ||
            checkDefaultLabels() ||
            checkDefaultTicks());
}

/**
 * Returns true if there is no DEFAULT_LABELS anno object, or if a
 * DEFAULT_LABELS anno object has been modified.
 *
 */
private boolean checkDefaultLabels() {
    boolean fNonDefault = false;
    JCAnno annoObj = axis.getDefaultLabels();
    if (annoObj == null || !isDefaultAnno(annoObj)) {
        fNonDefault = true;
    }
    return fNonDefault;
}

/**
 * Returns true if there is no DEFAULT_TICKS anno object, or if a
 * DEFAULT_TICKS anno object has been modified.
 */
private boolean checkDefaultTicks() {
    boolean fNonDefault = false;
    JCAnno tickObj = axis.getDefaultTicks();
    if (tickObj == null || !isDefaultAnno(tickObj)) {
        fNonDefault = true;
    }
    return fNonDefault;
}

private boolean isNonDefaultGrid() {
    List gridList = axis.getGridList();
    if (gridList == null || gridList.size() != 1) {
        return true;
    }
    JCGrid gridObj = (JCGrid)gridList.get(0);
    return (gridObj == null || checkGridProperties(gridObj));
}

protected boolean isDefaultIncrement(JCAnno tickObj) {
    boolean fDefault = true;  // Don't save if the JCAnno is null
    if (tickObj != null) {
        fDefault = tickObj.getIncrementIsDefault();
        if (tickObj.getType() == JCAnno.DEFAULT_LABELS) {
            fDefault = fDefault && axis.getNumSpacingIsDefault();
        }
        else if (tickObj.getType() == JCAnno.DEFAULT_TICKS) {
            fDefault = fDefault && axis.getTickSpacingIsDefault();
        }
    }
    return fDefault;
}

/**
 * Returns false if the JCAnno precision or JCAxis precision
 * is non-default.
 */
protected boolean isDefaultPrecision(JCAnno tickObj) {
    boolean fDefault = true;  // Don't save if the JCAnno is null
    if (tickObj != null) {
        fDefault = (tickObj.getPrecisionIsDefault() &&
                    axis.getPrecisionIsDefault());
    }
    return fDefault;
}

protected void saveValueLabels(PropertyPersistorModel prop, String name,
                               int tab, JCValueLabel[] valueLabels)
{
    if (valueLabels == null || valueLabels.length == 0) {
        return;
    }

    if (prop.getType().equals("HTML")) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < valueLabels.length; i++) {
            if (valueLabels[i] == null) {
                continue;
            }
            sb.append("" + valueLabels[i].getValue());
            sb.append(";");
            sb.append(prop.expandText(valueLabels[i].getText()));
            if (i != (valueLabels.length - 1)) {
                sb.append(";");
            }
        }
        prop.writeProperty(name, "valueLabels", tab, new String(sb));
    }
    else {
        for (int i = 0; i < valueLabels.length; i++) {
            if (valueLabels[i] == null) {
                continue;
            }

            int newTab = prop.writeBegin("value-label", tab);
            prop.writeProperty(name, "value", newTab,
                               new Double(valueLabels[i].getValue()));
            prop.writeEnd(null, tab, false, true);
            prop.writeString(prop.expandText(valueLabels[i].getText()), 0);
            prop.writeEnd("value-label", 0, true, false);
        }
    }
}

protected abstract void saveAxisGridProperties(PropertyPersistorModel prop,
                                           String name, int tab);
protected abstract void saveGridProperties(PropertyPersistorModel prop,
                        String name, int tab, int newTab) throws JCIOException;
protected abstract void saveTypeAndName(PropertyPersistorModel prop,
                                        String name, boolean isXAxis, int tab);
protected abstract void saveAnnoProperties(PropertyPersistorModel prop,
                                           String name, int tab, int newTab);

}
