/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCBaseAccessor.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:53 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.data.ChartDataSourceUtil;

import com.klg.jclass.util.JCStringTokenizer;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.io.LoadProperties;
import com.klg.jclass.util.io.OutputProperties;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.io.JCParseException;

import java.awt.*;
import java.io.InputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * <code>JCBaseAccessor</code> is the abstract parent class of property accessor
 * classes such as <code>JCFileAccessor</code> and <code>JCStringAccessor</code>.
 * <p>
 * It provides a basic property storage Hashtable, and a
 * mechanism for retrieving properties from the Hashtable.
 *
 * @see com.klg.jclass.chart.property.PropertyAccessModel
 * @see com.klg.jclass.chart.property.FilePropertyAccessModel
 */
public abstract class JCBaseAccessor implements PropertyAccessModel,
									            FilePropertyAccessModel
{

protected Hashtable       source = null;
protected Component       component;
protected LoadProperties  loadProperties;

/**
 * Get the properties from the string, file, etc, parse them, and store them.
 * Later a loader, actually loads the parsed properties into the chart in the
 * order it needs.
 * @param o The object from which to get the properties (fileName, inputStream,
 * reader, ...)
 * @throws IOException
 * @throws JCParseException
 */
public abstract void getProperties(Object o)
    throws IOException, JCParseException;

/**
 * Gets the accessor type String.
 * @return the accessor type String
 */
public abstract String getType();

/**
 * Gets the accessor subDirectory (if any).
 * @return the accessor subDirectory; null if there is no subDirectory
 */
public abstract String getSubDirectory();

/**
 * Gets property value as a String based on specified key.
 *
 * @param key name of property to be retrieved
 * @return property value as a String
 */
public String getProperty(String key)
{
	Object rval = getPropertyObject(key);
	if (rval instanceof String) {
		return((String)rval);
	}
	return(null);
}

/**
 * Gets property value as an object based on specified key.
 *
 * @param key name of property to be retrieved
 * @return property value as an object
 */
public Object getPropertyObject(String key)
{
	if (source == null || key == null) {
		return(null);
	}
	String keylc = key.toLowerCase();
	Object rval = source.get(keylc);
	if (rval == null) {
		rval = source.get(component.getName() + "." + keylc);
	}
	return(rval);
}

/**
 * Store an instance of LoadProperties containing the information
 * that will be used during the load of component properties.
 * @param loadProperties instance of LoadProperties
 */
public void setLoadProperties(LoadProperties loadProperties)
{
	this.loadProperties = loadProperties;
}

/**
 * Get the instance of LoadProperties
 * that will be used during the load of component properties.
 * @return instance of LoadProperties
 */
public LoadProperties getLoadProperties()
{
	return(loadProperties);
}

/**
 * Asks if one or more of the keys in the hashtable starts with the given prefix.
 * @param prefix the prefix with which to search
 * @return <code>true</code> if one or more of the keys starts with the prefix
 */
public boolean searchKeysForPrefix(String prefix)
{
	if (source == null) {
		return(false);
	}
	if (prefix == null) {
		return(true);
	}
	String prefixlc = prefix.toLowerCase();
	Enumeration e = source.keys();
	while (e.hasMoreElements()) {
		String key = (String)e.nextElement();
		if (key.startsWith(prefixlc)) {
			return(true);
		}
	}
	return(false);
}

/**
 * Debug method that prints out the hashtable.
 */
public void printProperties()
{
    if (source == null) {
		return;
    }
	Enumeration e = source.keys();
	while (e.hasMoreElements()) {
		String key = (String)e.nextElement();
		System.out.println("key:" + key + " value:" + getPropertyObject(key));
	}
}

/**
 * Sets the <code>String</code> property to value based on a specified key.
 * @param key name of property to be set
 * @param value the new String value of the property
 */
public void setProperty(String key, String value)
{
	setPropertyObject(key, value);
}

/**
 * Sets the property to a value based on a specified key.
 * @param key name of property to be set
 * @param value the new value of the property
 */
public void setPropertyObject(String key, Object value)
{
    if (source == null || key == null) {
		return;
	}
	String keylc = key.toLowerCase();
	if (source.containsKey(keylc) && value == null) {
		source.remove(keylc);
	}
	if (value == null) {
		return;
	}
	source.put(keylc, value);
}

/**
 * Gets property value based on specified key. The property value is a file
 * name. The implementing accessor will provide the rules that determine
 * the formatting of the file name and will return an input stream based
 * on the file.
 *
 * @param key name of property to be retrieved
 * @return property value (a file name) as an input stream
 * @throws JCIOException
 */
public InputStream getFilePropertyStream(String key,
                                         OutputProperties outputProperties)
    throws JCIOException
{
	String fileName = getProperty(key + "File");
    if (fileName == null || fileName.length() == 0 || loadProperties == null) {
        return(null);
    }
    String fileAccess = getProperty(key + "FileAccess");
    int access = loadProperties.getFileAccessEnumFromString(fileAccess, false);
    if (outputProperties != null) {
        outputProperties.setPropertyName(fileName);
        outputProperties.setFileAccess(access);
    }
    InputStream inputStream = null;
    try {
        inputStream = loadProperties.getInputStreamByAccessType(fileName,
                                                                access);
    }
    catch (Exception e) {
        throw new JCIOException(e.getMessage(), e);
    }
    return(inputStream);
}

/**
 * Get URL based on specified key.
 * @param key Name of property to be retrieved as URL
 * @return Property value as URL.
 * @throws MalformedURLException
 */
public URL getPropertyURL(String key, OutputProperties outputProperties)
    throws MalformedURLException
{
    if (loadProperties == null) {
        return (null);
    }

    String fileName = getProperty(key + "image.fileName");
    if (fileName == null || fileName.length() == 0) {
        // For backwards compatibility look for the image attribute
        fileName = getProperty(key + "image");
        if (fileName == null || fileName.length() == 0) {
            return(null);
        }
    }
    String fileAccess = getProperty(key + "image.fileAccess");
    if (fileAccess == null && fileName.indexOf("http") >= 0) {
        fileAccess = "Url";
    }
    int access = loadProperties.getFileAccessEnumFromString(fileAccess, false);
    if (outputProperties != null) {
        outputProperties.setPropertyName(fileName);
        outputProperties.setFileAccess(access);
    }
    URL url = loadProperties.getURLByAccessType(fileName, access);
    return(url);
}

/**
 * Gets the properties from an HTML input String and store them in a hashtable.
 * @param src the input String
 * @return a hashtable that stores the param/value pairs in the String
 */
protected static Hashtable findProperties(String src)
{
    int p, m, k;

    Hashtable psrc = new Hashtable();
    if (src == null) {
        return(psrc);
    }

    // Handles one level of HTML tags inside HTML param tags
    while ((p = src.indexOf('<')) != -1) {
        src = src.substring(p + 1);
        JCStringTokenizer string = new JCStringTokenizer(src);
        // ATTN GV Possibly invalid conversion
        string.strip_esc = false;

        String line = "";
        boolean lineDone = false;
        while (!lineDone) {
            String lineSoFar = string.nextToken('>');
            line += lineSoFar;
            if (lineSoFar.indexOf('<') == -1) {
                lineDone = true;
            }
            else {
                line += ">";
            }
        }
        line = JCTypeConverter.trim(line);

        if (line == null || line.length() == 0) {
            return (psrc);
        }

        src = JCTypeConverter.trim(src.substring(string.getPosition()));
        if (line.charAt(0) == '!' ||
            line.length() < 20 ||
            Character.isWhitespace(line.charAt(11)) ||
            !line.regionMatches(true, 0, "param name=", 0, 11)) {
            continue;
        }
        line = line.substring(11);

        // Find next non-blank
        for (m = 1; m < line.length(); m++) {
            if (Character.isWhitespace(line.charAt(m))) {
                break;
            }
        }
        for (k = m + 1; k < line.length(); k++) {
            if (!Character.isWhitespace(line.charAt(k))) {
                break;
            }
        }
        if (k == line.length() ||
            !line.regionMatches(true, k, "value=", 0, 6)) {
            continue;
        }

        String param = removeQuotes(line.substring(0, m).toLowerCase());
        String value = removeQuotes(line.substring(k + 6));
        value = ChartDataSourceUtil.processText(value);
        
        if (param != null) {
            psrc.put(param, value);
        }
    }
    return(psrc);
}

/**
 * Removes the beginning and end quote from a String (if they exist).
 * @param src the input string
 * @return a substring with the quotes removed
 */
protected static String removeQuotes(String src)
{
    String rval = src;
    if (src.charAt(0) == '"') {
        if (src.length() < 2) {
            return("");
        }
        rval = src.substring(1);
        if (rval.charAt(rval.length() - 1) == '"') {
            rval = rval.substring(0, rval.length() - 1);
        }
    }
    return(rval);
}

}
