/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartPropertyLoad.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:54 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.*;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.io.LoadProperties;
import com.klg.jclass.util.legend.JCMultiColLegend;

import javax.swing.*;
import java.awt.*;
import java.util.Hashtable;
import java.util.List;
import java.util.Locale;

/**
 * Responsible for loading properties of <code>JCChart</code>
 * objects using a specified <code>PropertyAcessModel</code>. Not
 * intended for public use.
 *
 * @see com.klg.jclass.chart.JCChart
 * @see com.klg.jclass.chart.property.JCChartPropertySave
 */
public abstract class JCChartPropertyLoad extends ComponentPropertyLoad {

protected Hashtable indexHashtable = null;
protected String[]  dvNames = null;
protected JCChart   chart = null;

/**
 * Sets the source object (the object on which to set properties).
 *
 * @param o object that is loading properties
 */
public void setSource(Object o)
{
	super.setSource(o);
	if (o instanceof JCChart) {
		this.chart = (JCChart)o;
	}
}

/**
 * Finds the <code>JCDataIndex</code> corresponding to key <i>s</i> in the
 * <code>indexHashTable</code>, which was previously created.
 * @param s the key of the data index in the hash table
 * @param def the data index to return if nothing is found or an error occurs
 * @return the data index retrieved from the hash table
 */
public JCDataIndex toDataIndex(String s, JCDataIndex def)
{
	if (s == null || indexHashtable == null) {
		return(def);
	}

	if (indexHashtable.containsKey(s)) {
		return((JCDataIndex)indexHashtable.get(s));
	}
	return(def);
}

/**
 * Returns the ChartDataView based on its property name in the dvNames array.
 * @param propertyName The name of the property.
 * @return The ChartDataView that corresponds to the name.
 */
public ChartDataView getDataViewFromProperty(String propertyName)
{
    if (propertyName != null && dvNames != null) {
        for (int i = 0; i < dvNames.length; i++) {
            if (dvNames[i].equals(propertyName)) {
                return(chart.getDataView(i));
            }
        }
    }
    // If the user has specified an incorrect name but there's only one
    // dataView anyways, then just return it.
    if (propertyName != null && chart.getNumData() == 1) {
        return(chart.getDataView(0));
    }
    return(null);
}

/**
 * Loads HTML parameters from the current param source and create a chart.
 *
 * @param prop property access object.
 * @param name name of object to be loaded.
 * @throws JCIOException
 */
public void loadProperties(PropertyAccessModel prop, String name)
    throws JCIOException
{
	if (prop == null) {
		return;
	}

	indexHashtable = new Hashtable();
    chart.setBatched(true);
	try {
	    loadPropertiesFromFile(prop, name);
	}
	catch (JCIOException e) {
        if (!prop.getLoadProperties().ignoreExternalResourceExceptions()) {
		    throw e;
        }
	}

	// Normalize name
	if (name == null || name.equals("")) {
		name = "";
	}
	else {
		name = name + ".";
	}

	super.loadProperties(prop, name);

	loadEventTriggers(prop, name);

	// Set name property for XML.  We can't do it for HTML because the chart's
	// name could be used for loading in properties
	String pval = prop.getProperty(name + "name");
	if (pval != null && prop.getType().equals("XML")) {
		chart.setName(pval);
	}

    // Set allow user changes
	pval = prop.getProperty(name + "allowUserChanges");
    chart.setAllowUserChanges(JCTypeConverter.toBoolean(pval,
							  chart.getAllowUserChanges()));
	loadWidthHeight(prop, name);

    pval = prop.getProperty(name + "antiAliasing");
    if (pval != null && !pval.equals("Default")) {
        String[] strings = JCChartEnumMappings.anti_aliasing_strings;
        int[] values = JCChartEnumMappings.anti_aliasing_values;
        chart.setAntiAliasing(JCTypeConverter.toEnum(pval, "antiAliasing",
                                                     strings, values,
                                                     chart.getAntiAliasing()));
    }

    // Load the chart's imageMapInfo, if set.
    ImageMapInfo info = new ImageMapInfo();
    PropertyLoadFactory.load(prop, info, name + "imageMapInfo.");
    if (!info.isEmpty()) {
        chart.setImageMapInfo(info);
    }

	loadLocale(prop, name);

    // Load the footer
	PropertyLoadFactory.load(prop, chart.getFooter(), "footer.");
	loadLayoutHints(prop, chart, chart.getFooter(), "footer.");

    // Load the footer's imageMapInfo, if set.
    info = new ImageMapInfo();
    PropertyLoadFactory.load(prop, info, "footer.imageMapInfo.");
    if (!info.isEmpty()) {
        chart.setFooterImageMapInfo(info);
    }

    // Load the header
	PropertyLoadFactory.load(prop, chart.getHeader(), "header.");
	loadLayoutHints(prop, chart, chart.getHeader(), "header.");

    // Load the header's imageMapInfo, if set.
    info = new ImageMapInfo();
    PropertyLoadFactory.load(prop, info, "header.imageMapInfo.");
    if (!info.isEmpty()) {
        chart.setHeaderImageMapInfo(info);
    }

	// Check legend type and create new legend if necessary
	pval = prop.getProperty(name + "legend.type");
	if (pval != null && pval.equals("MultiCol")) {
		chart.setLegend(new JCMultiColLegend());
	}

	// Now that correct legend has been set, handle legend properties
	PropertyLoadFactory.load(prop, chart.getLegend(), "legend.");
	loadLayoutHints(prop, chart, chart.getLegend(), "legend.");

	// Load ChartArea properties, including axis names, but actual axis
	// property loading is done a little further down
	PropertyLoadFactory.load(prop, chart.getChartArea(), "chartArea.");
	loadLayoutHints(prop, chart, chart.getChartArea(), "chartArea.");

	dvNames = new String[100];
    // The first dataview MUST be named data
    // This is also to provide compatibility with previous version

	PropertyLoadFactory.load(prop, chart.getDataView(0), "data.");
	dvNames[0] = "data";

    // Check if there are multiple dataviews
    // There is a limit of 100 dataviews
    int viewNum = 1;
    while (viewNum < 100) {
        ChartDataView dataview = null;
        String dataName = prop.getProperty(name + "dataName" + viewNum);
        if (dataName != null) {
			// If we don't have a data view, create one
			if (viewNum + 1 > chart.getNumData()) {
            	dataview = chart.addDataView(chart.getNumData());
			}
			// Otherwise, use an existing data view
			else {
            	dataview = chart.getDataView(viewNum);
			}
			dvNames[viewNum] = dataName;
			PropertyLoadFactory.load(prop, dataview, dataName + ".");
		}
        viewNum++;
    }

	// Now load the axes properties.  This must be done AFTER the data so
	// that the appropriate axis annotations will be set
	List xaxes = chart.getChartArea().getXAxes();
	for (int i=0; i < xaxes.size(); i++) {
		JCAxis xaxis = (JCAxis) xaxes.get(i);
        PropertyLoadFactory.load(prop, xaxis, xaxis.getName() + ".");
	}
	List yaxes = chart.getChartArea().getYAxes();
	for (int i=0; i < yaxes.size(); i++) {
		JCAxis yaxis = (JCAxis) yaxes.get(i);
        PropertyLoadFactory.load(prop, yaxis, yaxis.getName() + ".");
	}

    // Get JCDataIndices for the labels
    // No default DataIndex set
    // Maximum DataIndices read from lastLabelIndex property -- if not
	//   present, 99 is default

	String lastLabelIndexString = prop.getProperty(name + "lastLabelIndex");
	int lastLabelIndex = 99;
    if (lastLabelIndexString != null) {
		lastLabelIndex = (new Integer(lastLabelIndexString)).intValue();
	}

    int indexNum = 1;
    while (indexNum <= lastLabelIndex) {
		String indexName = prop.getProperty("indexName" + indexNum);
        if (indexName != null) {
            JCDataIndex index = new JCDataIndex();
            indexHashtable.put(indexName, index);
			PropertyLoadModel plm = PropertyLoadFactory.makeLoader(index);
			if (plm != null) {
				// Have to get dataView value here because JCDataIndex
				// doesn't know about JCChart or ChartDataView
				pval = prop.getProperty(name + indexName + ".dataView");
                ChartDataView dataView = getDataViewFromProperty(pval);
                if (dataView != null) {
                    index.setDataView(dataView);
                }

				// Now load rest of properties so we get series as
				// well as seriesIndex
				plm.loadProperties(prop, indexName + ".");
			}
        }
        indexNum++;
    }

    // Get labels. No default label set
    int labelNum = 1;
    while (labelNum <= lastLabelIndex) {
		String labelName = prop.getProperty("labelName" + labelNum);
        if (labelName != null) {
            JCChartLabel label = new JCChartLabel();
            chart.getChartLabelManager().addChartLabel(label);
			PropertyLoadModel plm = PropertyLoadFactory.makeLoader(label);
			if (plm != null) {
				plm.loadProperties(prop, labelName + ".");

				// Have to get dataView and dataIndex values here
				// because JCChartLabel doesn't know about them.
				pval = prop.getProperty(name + labelName + ".dataView");
                ChartDataView dataView = getDataViewFromProperty(pval);
                if (dataView != null) {
                    label.setDataView(dataView);
                }

				pval = prop.getProperty(name + labelName + ".dataIndex");
				if (pval != null) {
					JCDataIndex di = toDataIndex(pval, label.getDataIndex());
					if (di != null) {
						label.setDataIndex(di);
					}
				}
			}
		}
        labelNum++;
    }

    loadImageMapRules(prop, name);

	LoadProperties loadProps = prop.getLoadProperties();
	if (loadProps != null              &&
        loadProps.getStoreUserObject() &&
        loadProps.getUserObject() != null)
    {
		chart.setUserObject(loadProps.getUserObject());
	}

    chart.setBatched(false);

    // Set batched mode
    chart.setBatched(JCTypeConverter.toBoolean(prop.getProperty("batched"),
											   chart.isBatched()));
}

/**
 * Loads the properties from a file.  Does nothing here. May be overridden in
 * subclasses to provide access to external property files.
 *
 * @param prop property access object.
 * @param name name of object to be loaded.
 * @throws JCIOException
 */
protected void loadPropertiesFromFile(PropertyAccessModel prop, String name)
    throws JCIOException
{
}

/**
 * Loads a list of image map rules.  Does nothing here as image maps are not
 * relevant for JCChart.  May be overridden in subclasses to provide image
 * map rules specification.
 *
 * @param prop property access object.
 * @param name name of object to be loaded.
 * @throws JCIOException
 */
protected void loadImageMapRules(PropertyAccessModel prop, String name)
    throws JCIOException
{
}

/**
 * Loads the width and height properties.
 * @param prop property access object
 * @param name name of object to be loaded
 */
protected void loadWidthHeight(PropertyAccessModel prop, String name)
{
	String widthString  = prop.getProperty(name + "width");
	String heightString = prop.getProperty(name + "height");
	if (widthString != null || heightString != null) {
		Dimension s = chart.getPreferredSize();
		Dimension d = new Dimension(JCTypeConverter.toInt(widthString,
		                                                  s.width),
    			                    JCTypeConverter.toInt(heightString,
									                      s.height));
		chart.setPreferredSize(d);
	}
}

/**
 * Loads the locale properties properties.
 * @param prop property access object
 * @param name name of object to be loaded
 */
protected void loadLocale(PropertyAccessModel prop, String name)
{
	Locale locale = null;

	String localeLanguage = prop.getProperty(name + "locale.language");

	// No language attr -- no locale -- return and use default
	if (localeLanguage == null) {
		return;
	}

	String localeCountry = prop.getProperty(name + "locale.country");

	// If no country specified, use emptry string
	if (localeCountry == null) {
		localeCountry = "";
	}

	String localeVariant = prop.getProperty(name + "locale.variant");
	if (localeVariant == null) {
		locale = new Locale(localeLanguage, localeCountry);
	}
    else {
		locale = new Locale(localeLanguage, localeCountry, localeVariant);
	}

	// Set locale
	if (locale != null) {
		chart.setLocale(locale);
	}
}


/**
 * Loads the layout hints (position of chart children).
 * @param prop property access object
 * @param chart the chart for which to get layout hints
 * @param child the child of the chart for which to get layout hints
 * @param name name of object to be loaded
 */
protected void loadLayoutHints(PropertyAccessModel prop, JCChart chart,
							   JComponent child, String name)
{
	Rectangle hints = chart.getLayoutHints(child);

	boolean changed = false;

	String pval = prop.getProperty(name + "x");
	if (pval != null) {
		hints.x = Integer.parseInt(pval);
		changed = true;
	}

	pval = prop.getProperty(name + "y");
	if (pval != null) {
		hints.y = Integer.parseInt(pval);
		changed = true;
	}

	pval = prop.getProperty(name + "width");
	if (pval != null) {
		hints.width = Integer.parseInt(pval);
		changed = true;
	}

	pval = prop.getProperty(name + "height");
	if (pval != null) {
		hints.height = Integer.parseInt(pval);
		changed = true;
	}

	if (changed) {
		chart.setLayoutHints(child, hints);
	}
}

/**
 * Loads all event triggers from the <code>PropertyAccessModel</code>.
 *
 * @param prop property access object
 * @param name name of object to be loaded
 */
protected void loadEventTriggers(PropertyAccessModel prop, String name)
{
    // Set EventTriggers
	for (int i = 1; i < JCChartEnumMappings.trigger_values.length; i++) {
		String triggerName = name + JCChartEnumMappings.trigger_strings[i] +
			                 "Trigger";
		int triggerEnum = JCChartEnumMappings.trigger_values[i];
		loadTrigger(prop, triggerName, triggerEnum);
	}

    // Set Keys
    if (prop.getProperty("resetKey") != null) {
		chart.setResetKey(Integer.parseInt(prop.getProperty("resetKey")));
	}
    if (prop.getProperty("cancelKey") != null) {
		chart.setCancelKey(Integer.parseInt(prop.getProperty("cancelKey")));
	}
}

/**
 * Loads an individual event trigger from the <code>PropertyAccessModel</code>.
 *
 * @param prop property access object
 * @param triggerName the name of the trigger
 * @param triggerEnum the trigger enum value
 */
protected void loadTrigger(PropertyAccessModel prop, String triggerName,
						   int triggerEnum)
{
    String mod = prop.getProperty(triggerName);
    if (mod != null) {
		int e = JCTypeConverter.toEnum(mod, triggerName,
								       JCChartEnumMappings.eventMod_strings,
								       JCChartEnumMappings.eventMod_values, 0);
        chart.addTrigger(e, triggerEnum);
    }
}

/**
 * Given an object o representing a list of points (stored as either a List or
 * a String), convert it to an array of Point objects.
 * @param o The object to convert
 * @return The newly created array of Point objects.
 */
public static Point[] convertObjectToPointList(Object o)
{
    Point[] pointList = null;
    String[] dataStrings = JCChartEnumMappings.data_format_strings;
    String allString = dataStrings[0];
    String otherString = dataStrings[1];
    if (o instanceof List) {
        // Data was loaded from XML as a List.
        List seriesPoints = (List)o;

        // Each item in the ArrayList is a 2-element String array.  The
        // first element represents the series value ("All", "Other_Slice",
        // or a number).  The second element represents the point value
        // ("All" or a number).
        pointList = new Point[seriesPoints.size()];
        for (int i = 0; i < seriesPoints.size(); i++) {
            String[] sp = (String[])seriesPoints.get(i);
            if (sp != null) {
                int series = -1;
                if (sp[0].equalsIgnoreCase(allString)) {
                    series = ChartDataModel.ALL;
                }
                else if (sp[0].equalsIgnoreCase(otherString)) {
                    series = JCPieChartFormat.OTHER_SLICE;
                }
                else {
                    series = JCTypeConverter.toInt(sp[0], Integer.MAX_VALUE);
                }

                int point = -1;
                if (sp[1].equalsIgnoreCase(allString)) {
                    point = ChartDataModel.ALL;
                }
                else {
                    point = JCTypeConverter.toInt(sp[1], Integer.MAX_VALUE);
                }
                pointList[i] = new Point(point, series);
            }
        }
    }
    else if (o instanceof String) {
        // Data was loaded from HTML.
        String str = (String)o;
        String all = Integer.toString(JCPieChartFormat.ALL);
        str = str.replaceAll(allString, all);
        str = str.replaceAll(allString.toUpperCase(), all);
        str = str.replaceAll(allString.toLowerCase(), all);
        String other = Integer.toString(JCPieChartFormat.OTHER_SLICE);
        str = str.replaceAll(otherString, other);
        str = str.replaceAll(otherString.toUpperCase(), other);
        str = str.replaceAll(otherString.toLowerCase(), other);
        pointList = JCTypeConverter.toPointList(str, '|');
    }
    return(pointList);
}

}
