/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChartPropertySave.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:54 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.ChartDataView;
import com.klg.jclass.chart.JCChart;
import com.klg.jclass.chart.JCChartEnumMappings;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.ImageMapInfo;

import java.awt.*;
import java.util.ListIterator;

/**
 * Responsible for saving properties of <code>JCChart</code>
 * objects to a destiation specified by <code>PropertyPersistorModel</code>.
 * Not intended for public use.
 *
 * @see com.klg.jclass.chart.JCChart
 * @see com.klg.jclass.chart.property.JCChartPropertyLoad
 */
public abstract class JCChartPropertySave extends ComponentPropertySave {

protected JCChart chart = null;
protected JCChart defaultChart = null;

/**
 * Sets the source component whose properties are to be saved.
 * @param o the new source component
 */
public void setSource(Object o)
{
	super.setSource(o);
	if (o instanceof JCChart) {
		this.chart = (JCChart)o;
	}
}

/**
 * Sets default object instance.  This object is typically used by implementors
 * to determine whether a property really needs to be saved.
 * @param def the new default object
 */
public void setDefault(Object def)
{
	super.setDefault(def);
	if (def instanceof JCChart) {
		this.defaultChart = (JCChart)def;
	}
}

/**
 * Checks whether there are any properties different from their default.  If so,
 * then this saver has information to write.
 * @return whether any properties are different from their default
 */
public boolean checkProperties()
{
	return(true);
}

/**
 * Saves HTML parameters from the current param source and creates a chart;
 * click the method name to see which properties are saved.<p>
 *
 * The following properties are saved:
 *   <pre>
 *   rotateTrigger (event)
 *   zoomTrigger (event)
 *   translateTrigger (event)
 *   editTrigger (event)
 *   pickTrigger (event)
 *   depthTrigger (event)
 *   customizeTrigger (event)
 *   allowUserChanges (boolean)
 *   footer (Component)
 *   header (Component)
 *   legend (JCLegend, see JCLegendPropertySave)
 *   chartArea (JCChartArea, see JCChartAreaPropertySave)
 *   data&lt;n&gt; (ChartDataView, see ChartDataViewPropertySave)
 *   </pre>
 *   eventually, data indices and chart labels...
 *
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 * @throws JCIOException
 */
public void saveProperties(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
	if (chart == null || defaultChart == null || prop == null) {
		return;
	}

	// Normalize name
	if (name == null || name.equals("")) {
		name = "";
	}
	else {
		name = name + ".";
	}

	saveHeader(prop, tab);

	int newTab = prop.writeBegin("chart", tab);

	if (chart.getName() != null && prop.getType().equals("XML")) {
		prop.writeProperty(name, "name", newTab, chart.getName());
	}

    // Set allow user changes
	if (chart.getAllowUserChanges() != defaultChart.getAllowUserChanges()) {
		prop.writeProperty(name, "allowUserChanges", newTab,
						   new Boolean(chart.getAllowUserChanges()));
	}
	saveWidthHeight(prop, name, newTab);
    if (chart.getAntiAliasing() != defaultChart.getAntiAliasing()) {
        String[] strings = JCChartEnumMappings.anti_aliasing_strings;
        int[] values = JCChartEnumMappings.anti_aliasing_values;
        prop.writeProperty(name, "antiAliasing", newTab,
                           JCTypeConverter.fromEnum(chart.getAntiAliasing(),
                                                    strings, values));
    }


	prop.writeEnd(null, tab, true, true);

	if (super.checkProperties(prop)) {
	    super.saveProperties(prop, name, tab + PropertySaveFactory.tabIncrement);
	}

    // Save ImageMapInfo properties
    PropertySaveFactory.save(prop, chart.getImageMapInfo(),
                             new ImageMapInfo(), name + "imageMapInfo.", tab);

	saveLocale(prop, name, tab + PropertySaveFactory.tabIncrement);
	saveEventTriggers(prop, name, tab + PropertySaveFactory.tabIncrement);

    prop.setHelperObject(chart);
	PropertySaveFactory.save(prop, chart.getHeader(),
							 defaultChart.getHeader(), "header.", tab);

	PropertySaveFactory.save(prop, chart.getFooter(),
							 defaultChart.getFooter(), "footer.", tab);
    prop.setHelperObject(null);

	PropertySaveFactory.save(prop, chart.getLegend(),
							 defaultChart.getLegend(), "legend.", tab);

	PropertySaveFactory.save(prop, chart.getChartArea(),
							 defaultChart.getChartArea(), "chartArea.", tab);

	int num = 0;
	ChartDataView defaultView = defaultChart.getDataView(0);
	for (ListIterator li = chart.getDataView().listIterator();
	     li.hasNext();
		 num++)
	{
		ChartDataView view = (ChartDataView)li.next();
		String numString = (num == 0 ? "" : "" + num);
		saveDataView(view, defaultView, numString, prop, name, tab);
	}

	saveChartLabels(prop, name, tab);

    saveImageMapRules(prop, name, tab);

	/*
    // Set Batched mode
    chart.setBatched(JCTypeConverter.toBoolean(prop.writeProperty("batched"),
											   chart.isBatched()));
											   */
	prop.writeEnd("chart", tab, true, false);
}

/**
 * Saves a header for the chart.
 * @param prop property persistor object
 * @param tab the tab indent
 */
protected abstract void saveHeader(PropertyPersistorModel prop, int tab);

/**
 * Saves the width and height parameters from the chart.
 * @param prop property persistor object
 * @param tab the tab indent
 * @param name name of object to be saved
 */
protected void saveWidthHeight(PropertyPersistorModel prop, String name,
                               int tab)
{
	Dimension size =  chart.getSize();
	Dimension defSize = defaultChart.getSize();
	if (size.width != defSize.width) {
		prop.writeProperty(name, "width", tab, new Integer(size.width));
	}
	if (size.height != defSize.height) {
		prop.writeProperty(name, "height", tab, new Integer(size.height));
	}
}

/**
 * Saves the properties for the given data view.
 * @param view the <code>ChartDataView</code> to save
 * @param defaultView the default <code>ChartDataView</code> to compare
 * @param numString the String which indicates the data view number
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 * @throws JCIOException
 */
protected void saveDataView(ChartDataView view, ChartDataView defaultView,
                            String numString, PropertyPersistorModel prop,
							String name, int tab)
    throws JCIOException
{
    String dvName = name + "data" + numString;
    PropertySaveFactory.save(prop, view, defaultView, dvName + ".", tab);
}

/**
 * Saves the chart labels for the chart.
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 */
protected abstract void saveChartLabels(PropertyPersistorModel prop,
                                        String name, int tab)
    throws JCIOException;

/**
 * Saves the image map rules for the chart.
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 */
protected abstract void saveImageMapRules(PropertyPersistorModel prop,
                                          String name, int tab)
    throws JCIOException;

/**
 * Saves all event triggers from the <code>PropertyPersistorModel</code>.
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 */
protected abstract void saveEventTriggers(PropertyPersistorModel prop,
                                          String name, int tab);

/**
 * Saves locale information from the <code>PropertyPersistorModel</code>.
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 */
protected abstract void saveLocale(PropertyPersistorModel prop,
                                          String name, int tab);

public static String getDataViewName(JCChart chart, ChartDataView dataView)
{
    String dvname = null;
    if (chart != null && dataView != null) {
        int num = 0;
        ListIterator li = chart.getDataView().listIterator();
        while (li.hasNext()) {
            ChartDataView view = (ChartDataView)li.next();
            if (view == dataView) {
                dvname = "data" + (num == 0 ? "" : "" + num);
                break;
            }
            num++;
        }
    }
    return((dvname == null) ? "data" : dvname);
}

}
