/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCPieChartFormatPropertySave.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:54 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.JCChartEnumMappings;
import com.klg.jclass.chart.JCPieChartFormat;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.io.JCIOException;

/**
 * Responsible for saving properties of <code>JCPieChartFormat</code>
 * objects to a destiation specified by <code>PropertyPersistorModel</code>.
 * Not intended for public use.
 *
 * @see com.klg.jclass.chart.JCPieChartFormat
 * @see com.klg.jclass.chart.property.JCPieChartFormatPropertyLoad
 */
public abstract class JCPieChartFormatPropertySave implements PropertySaveModel {

protected JCPieChartFormat format = null;
protected JCPieChartFormat defaultFormat = null;

/**
 * Sets the source component whose properties are to be saved.
 * @param o the new source component
 */
public void setSource(Object o)
{
    if (o instanceof JCPieChartFormat) {
        this.format = (JCPieChartFormat)o;
    }
}

/**
 * Sets default object instance.  This object is typically used by implementors
 * to determine whether a property really needs to be saved.
 * @param def the new default object
 */
public void setDefault(Object def)
{
    if (def instanceof JCPieChartFormat) {
        this.defaultFormat = (JCPieChartFormat)def;
    }
}

/**
 * Checks whether there are any properties different from their default.  If so,
 * then this saver has information to write.
 * @param prop property persistor object
 * @return whether any properties are different from their default
 */
public boolean checkProperties(PropertyPersistorModel prop)
{
    if (format == null || defaultFormat == null) {
        return(false);
    }

    if (format.getMinSlices() != defaultFormat.getMinSlices()             ||
        format.getExplodeOffset() != defaultFormat.getExplodeOffset()     ||
        (format.getExplodeList() != null &&
            format.getExplodeList().length > 0)                           ||
        format.getThresholdMethod() != defaultFormat.getThresholdMethod() ||
        format.getThresholdValue() != defaultFormat.getThresholdValue()   ||
        format.getSortOrder() != defaultFormat.getSortOrder()             ||
        format.getStartAngle() != defaultFormat.getStartAngle()           ||
        format.getOtherLabel() != defaultFormat.getOtherLabel())
    {
        return(true);
    }

    ImageMapInfo legendInfo = format.getOtherLegendImageMapInfo();
    if (!ImageMapInfo.isEmpty(format.getOtherImageMapInfo()) ||
        (legendInfo != null && !legendInfo.isEmpty()))
    {
        return(true);
    }

    return(PropertySaveFactory.check(prop, format.getOtherStyle(),
                                     defaultFormat.getOtherStyle()));
}

/**
 * Saves HTML parameters from the current param source;
 * click the method name to see which properties are saved.<p>
 * Saves the following properties
 *  <pre>
 *   minSlices (integer)
 *   explodeOffset (integer)
 *   thresholdMethod (enum)
 *   thresholdValue (double)
 *   sortOrder (enum)
 *   startAngle (double)
 *   other.label (string)
 *   other.&lt;style&gt; (JCChartStyle)
 *  </pre>
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 * @throws JCIOException
 */
public void saveProperties(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
    if (format == null || defaultFormat == null) {
        System.out.println("FAILURE: No pie format set");
        return;
    }

    // Check if something to do.
    if (!checkProperties(prop)) {
        return;
    }

    int newTab = prop.writeBegin("pie-format", tab);

    // Set minSlices
    if (format.getMinSlices() != defaultFormat.getMinSlices()) {
        prop.writeProperty(name, "minSlices", newTab,
                           new Integer(format.getMinSlices()));
    }

    // Set explodeOffset
    if (format.getExplodeOffset() != defaultFormat.getExplodeOffset()) {
        prop.writeProperty(name, "explodeOffset", newTab,
                           new Integer(format.getExplodeOffset()));
    }

    // Set thresholdMethod
    String[] strings = JCChartEnumMappings.thresholdMethod_strings;
    int[] values     = JCChartEnumMappings.thresholdMethod_values;
    if (format.getThresholdMethod() != defaultFormat.getThresholdMethod()) {
        prop.writeProperty(name, "thresholdMethod", newTab,
                           JCTypeConverter.fromEnum(format.getThresholdMethod(),
                                                    strings, values));
    }

    // Set thresholdValue
    if (format.getThresholdValue() != defaultFormat.getThresholdValue()) {
        prop.writeProperty(name, "thresholdValue", newTab,
                           new Double(format.getThresholdValue()));
    }

    // Set sortOrder
    strings = JCChartEnumMappings.sortOrder_strings;
    values  = JCChartEnumMappings.sortOrder_values;
    if (format.getSortOrder() != defaultFormat.getSortOrder()) {
        prop.writeProperty(name, "sortOrder", newTab,
                           JCTypeConverter.fromEnum(format.getSortOrder(),
                                                    strings, values));
    }

    // Set thresholdValue
    if (format.getStartAngle() != defaultFormat.getStartAngle()) {
        prop.writeProperty(name, "startAngle", newTab,
                           new Double(format.getStartAngle()));
    }

    saveOtherProperties(prop, name, tab, newTab);

    saveExplodedSliceProperties(prop, name,
                                tab + PropertySaveFactory.tabIncrement);

    prop.writeEnd("pie-format", tab, true, false);
}

public abstract void saveExplodedSliceProperties(PropertyPersistorModel prop,
                                                 String name, int tab);

public abstract void saveOtherProperties(PropertyPersistorModel prop,
                                         String name, int tab, int newTab)
    throws JCIOException;

}
