/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JLabelPropertySave.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:55 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.JCChart;
import com.klg.jclass.chart.JCChartLabel;
import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.io.JCIOException;
import com.klg.jclass.util.swing.JCSwingTypeConverter;

import javax.swing.*;

/**
 * Responsible for saving properties of <code>JLabel</code>
 * objects to a destiation specified by <code>PropertyPersistorModel</code>.
 * Not intended for public use.
 *
 * @see javax.swing.JLabel
 * @see com.klg.jclass.chart.property.JLabelPropertyLoad
 */
public class JLabelPropertySave extends ComponentPropertySave {

/**
 * Checks whether there are any properties different from their default.  If so,
 * then this saver has information to write
 * @param prop property persistor object
 * @return whether any properties are different from their default
 */
public boolean checkProperties(PropertyPersistorModel prop)
{
    boolean writeProperties = false;

    // For headers and footers, the helper object will be the parent chart.
    Object helper = prop.getHelperObject();
	if (helper != null && helper instanceof JCChart) {
		if (checkLayoutHints((JCChart)helper, (JComponent)component)) {
            writeProperties = true;
		}
	}

    JLabel label    = (JLabel)component;
    JLabel defLabel = (JLabel)defaultComponent;
	String text     = label.getText();
	if (!writeProperties &&
        (super.checkProperties(prop)                                         ||
         determineSaveText(prop, text) != null                               ||
         label.getHorizontalAlignment() != defLabel.getHorizontalAlignment() ||
         label.getVerticalAlignment() != defLabel.getVerticalAlignment()))
    {
		writeProperties = true;
	}

    if (!writeProperties && helper != null && helper instanceof JCChart) {
        JCChart chart = (JCChart)helper;
        ImageMapInfo info =
            (component == chart.getHeader()) ? chart.getHeaderImageMapInfo()
                                             : chart.getFooterImageMapInfo();
        if (info != null && !info.isEmpty()) {
            writeProperties = true;
        }
    }

	return(writeProperties);
}

/**
 * Determine the correct text to save.  If this JLabel is part of a
 * JCChartLabel, the text to save will depend on the defaults set in
 * the chart label.
 * @param prop the PropertyPersistorModel
 * @param text the text from the JLabel that is the candidate for saving
 * @return the text that should be saved.  If no text should be saved,
 *         this is null.
 */
protected String determineSaveText(PropertyPersistorModel prop, String text) {

	Object helper = prop.getHelperObject();

	// If JLabel is part of a JCChartLabel, only write out the text if
	// it's not the default text
	if (helper != null && helper instanceof JCChartLabel) {
		JCChartLabel cl = (JCChartLabel) helper;
		String defaultText = cl.getDefaultText();
		if (defaultText != null && text != null && defaultText.equals(text)) {
			text = null;
		}
	}

	return(text);
}

/**
 * Saves HTML parameters from the current param source;
 * click the method name to see which properties are saved.<p>
 * Saves the following properties:
 *  <pre>
 *   text (String)
 *  </pre>
 * @param prop property persistor object
 * @param name name of object to be saved
 * @param tab the tab indent for this group of properties
 * @throws JCIOException
 */
public void saveProperties(PropertyPersistorModel prop, String name, int tab)
    throws JCIOException
{
	// Check is something to do
	if (!checkProperties(prop)) {
		return;
	}

	String labelName = name.substring(0, name.indexOf("."));
	int newTab = prop.writeBegin(labelName, tab);

    JLabel label = (JLabel)component;
	String text = label.getText();
	text = determineSaveText(prop, text);

	// if no text (or default text), don't write out text
	// Note: it is important to write out empty text
	if (text != null) {
	    prop.writeProperty(name, "text", newTab, prop.expandText(text));
	}

    JLabel defLabel = (JLabel)defaultComponent;
    int horizontalAlignment = label.getHorizontalAlignment();
    if (horizontalAlignment != defLabel.getHorizontalAlignment()) {
        String hAlignString =
            JCTypeConverter.fromEnum(horizontalAlignment,
                                     JCSwingTypeConverter.horizontalAlignmentStrings,
                                     JCSwingTypeConverter.horizontalAlignmentValues);
        prop.writeProperty(name, "horizontalAlignment", newTab, hAlignString);
    }
    int verticalAlignment = label.getVerticalAlignment();
    if (verticalAlignment != defLabel.getVerticalAlignment()) {
        String vAlignString =
            JCTypeConverter.fromEnum(verticalAlignment,
                                     JCSwingTypeConverter.verticalAlignmentStrings,
                                     JCSwingTypeConverter.verticalAlignmentValues);
        prop.writeProperty(name, "verticalAlignment", newTab, vAlignString);
    }

	prop.writeEnd(null, tab, true, true);

	if (super.checkProperties(prop)) {
	    super.saveProperties(prop, name,
		                     tab + PropertySaveFactory.tabIncrement);
	}

    // For headers and footers, the helper object will be the parent chart.
    Object helper = prop.getHelperObject();

	if (helper != null && helper instanceof JCChart) {
	    saveLayoutHints(prop, (JCChart)helper, (JComponent)component,
		                name, tab + PropertySaveFactory.tabIncrement);
	}

    // Save ImageMapInfo properties
    if (helper != null && helper instanceof JCChart) {
        JCChart chart = (JCChart)helper;
        ImageMapInfo info =
            (component == chart.getHeader()) ? chart.getHeaderImageMapInfo()
                                             : chart.getFooterImageMapInfo();
        if (info != null && !info.isEmpty()) {
            PropertySaveFactory.save(prop, info, new ImageMapInfo(),
                                     name + "imageMapInfo.", tab);
        }
    }

	prop.writeEnd(labelName, tab, true, false);
}

}
