/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PropertyLoadFactory.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:55 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.property.html.JCHTMLStringAccessor;
import com.klg.jclass.util.io.JCIOException;

import java.lang.reflect.Modifier;

/**
 * Factory object for loading Chart property load
 * objects based on a simple naming convention.
 * The naming convention assumes the name of the
 * chart object is the first part of the class name,
 * and is followed by PropertyLoad.class eg:
 * &lt;Class&gt;PropertyLoad.class.
 *
 * @see com.klg.jclass.chart.property.PropertyLoadModel
 */
public class PropertyLoadFactory {

private final static String pkg = "com.klg.jclass.chart.property";

/**
 * Return the package name
 * @return The package name
 */
public static String getPackage()
{
	return(pkg);
}

/**
 *  Load with loaders from the default package.
 *  @param c The object from which to make the loader
 *  @return A new property load object
 */
public static PropertyLoadModel makeLoader(Object c)
{
	if (c == null) {
		return(null);
	}

	PropertyAccessModel pam = new JCHTMLStringAccessor();
	return(makeLoader(c, pam));
}

/**
 *  Load with loaders from the default package.
 *  @param c The object from which to make the loader
 *  @param pam The property access model
 *  @return A new property load object
 */
public static PropertyLoadModel makeLoader(Object c, PropertyAccessModel pam)
{
	if (c == null) {
		return(null);
	}

	return(makeLoader(c, pkg, pam.getSubDirectory(), pam.getType()));
}

/**
 * Load with loaders from package specified in the 'prefix' parameter.
 * PropertyLoad class loading is simple because we follow
 * a convention &lt;Class&gt;PropertyLoad.class.
 *  @param c The object from which to make the loader
 *  @param loaderPkg A prefix (namely the package name) to prepend to the
 *      classname
 *  @param loaderSub The subdirectory to first look for the loader
 *  @param loaderType The type of loader
 *  @return A new property load object
 */
public static PropertyLoadModel makeLoader(Object c, String loaderPkg,
                                           String loaderSub, String loaderType)
{
	if (c == null || loaderPkg == null) {
		return(null);
	}
	Class objectClass = c.getClass();

	PropertyLoadModel rval = null;
	try {
		while (objectClass != null && rval == null) {
			String name       = objectClass.getName();
			String className  = name.substring(name.lastIndexOf(".") + 1);
			Class clz         = null;

            // Try general loader
            String prefix = loaderPkg;
            String loaderName = className + "PropertyLoad";
            try {
                clz = Class.forName(prefix + "." + loaderName);
            }
            catch (Exception e) {
                clz = null;
            }

            if (clz == null ) {
                // No property loader yet - move up to superclass
                objectClass = objectClass.getSuperclass();
                continue;
            }

            // If class is abstract then we need to look in a subpackage
            int modifier = clz.getModifiers();
            if (Modifier.isAbstract(modifier)) {
                clz = null;
                if (loaderSub != null && loaderType != null) {
                    prefix     = loaderPkg + "." + loaderSub;
                    loaderName = className + loaderType + "PropertyLoad";
                    try {
                        clz = Class.forName(prefix + "." + loaderName);
                    }
                    catch (Exception e) {
                        clz = null;
                    }
                }
                if (clz == null) {
                    // This shouldn't happen, but react gracefully if it does
                    return(null);
                }
            }

            // Found a property loader - create it and set it up
            rval = (PropertyLoadModel)clz.newInstance();
            rval.setSource(c);
         }
	}
	catch (Exception e) {
		rval = null;
	}

	return(rval);
}

/**
 * Loads properties for the specific object c using the provided
 * property accessor prop.  The string tag is used to refer to the
 * properties to be loaded.
 *
 * For example:  load(prop, chart.getChartArea(), "chart.chartArea");
 *
 * @param prop The property accessor
 * @param c The object from which to load the properties
 * @param tag The tag
 * @throws JCIOException
 */
public static void load(PropertyAccessModel prop, Object c, String tag)
    throws JCIOException
{
	if (c == null) {
		return;
	}
	PropertyLoadModel plm = makeLoader(c, pkg, prop.getSubDirectory(),
                                       prop.getType());
	if (plm != null) {
		plm.loadProperties(prop, tag);
	}
}

}
