/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PropertySaveFactory.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:55 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart.property;

import com.klg.jclass.chart.JCChart;
import com.klg.jclass.chart.JCChartFactory;
import com.klg.jclass.chart.property.html.JCHTMLStringPersistor;
import com.klg.jclass.util.io.JCIOException;

import java.awt.*;
import java.lang.reflect.Modifier;

/**
 * Factory object for loading Chart property save
 * objects based on a simple naming convention.
 * The naming convention assumes the name of
 * the chart object is the first part of the class
 * name, and is followed by PropertySave.class eg:
 * &lt;Class&gt;PropertySave.class.
 *
 * @see com.klg.jclass.chart.property.PropertySaveModel
 */
public class PropertySaveFactory {

private final static String pkg = "com.klg.jclass.chart.property";

public static int tabIncrement = 2;

/**
 * Return the package name
 * @return The package name
 */
public static String getPackage()
{
	return(pkg);
}

/**
 * Load a Save class from the default package.
 * @param c The object for which to save
 * @param def This object is typically used by implementors to determine whether
 *     a property really needs to be saved.
 * @return The newly created property save model
 */
public static PropertySaveModel makeSaver(Object c, Object def)
{
	PropertyPersistorModel ppm = new JCHTMLStringPersistor(null);
	return(makeSaver(c, def, ppm));
}

/**
 * Load a Save class from the default package.
 * @param c The object for which to save
 * @param def This object is typically used by implementors to determine whether
 *     a property really needs to be saved.
 * @param prop The property persistor
 * @return The newly created property save model
 */
public static PropertySaveModel makeSaver(Object c, Object def,
                                          PropertyPersistorModel prop)
{
	if (c == null || def == null || prop == null) {
		return(null);
	}
	return(makeSaver(c, def, pkg, prop.getSubDirectory(), prop.getType()));
}

/**
 * Load a Save class from the package by the 'prefix' method.
 * PropertySave class saving is simple because we follow
 * a convention &lt;Class&gt;&lt;Type&gt;PropertySave.class.
 * @param c The object for which to make a saver
 * @param def This object is typically used by implementors to determine whether
 *     a property really needs to be saved.
 * @param saverPkg The prefix (usually the package name of the class)
 * @param saverSub The subdirectory of the type of saver
 * @param saverType The type string
 * @return s The newly created property save model
 */
public static PropertySaveModel makeSaver(Object c, Object def, String saverPkg,
                                          String saverSub, String saverType)
{
	if (c == null || def == null || saverPkg == null) {
		return(null);
	}
	Class objectClass = c.getClass();

	PropertySaveModel rval = null;
	try {
		while (objectClass != null && rval == null) {
			String name      = objectClass.getName();
			String className = name.substring(name.lastIndexOf(".") + 1);
			Class clz        = null;

            // Type specific saver not found.  Try general one.
            String prefix    = saverPkg;
            String saverName = className + "PropertySave";
            try {
                clz = Class.forName(prefix + "." + saverName);
            }
            catch (Exception e) {
                clz = null;
            }

            if (clz == null) {
                // No property loader yet - move up to superclass
                objectClass = objectClass.getSuperclass();
                continue;
            }

            // If class is abstract then we need to look in a subpackage
            int modifier = clz.getModifiers();
            if (Modifier.isAbstract(modifier)) {
                clz = null;
                if (saverSub != null && saverType != null) {
                    prefix = saverPkg + "." + saverSub;
                    saverName = className + saverType + "PropertySave";
                    try {
                        clz = Class.forName(prefix + "." + saverName);
                    }
                    catch (Exception e) {
                        clz = null;
                    }
                }
                if (clz == null) {
                    // This shouldn't happen, but react gracefully if it does
                    return (null);
                }
            }

            // Found a property Saver - create it and set it up
            rval = (PropertySaveModel)clz.newInstance();
            rval.setSource(c);
            rval.setDefault(def);
		}
	}
	catch (Exception e) {
		rval = null;
	}

	return(rval);
}

/**
 * Saves properties for the specific object c using the provided
 * property persistor prop.  The string tag is used to refer to the
 * properties to be saved.
 *
 * For example:  save(prop, chart.getChartArea(), defaultChartArea,
 *                    "chart.chartArea");
 *
 * @param prop The property persistor
 * @param c The object whose properties are to be saved
 * @param def This object is typically used by implementors to determine whether
 *     a property really needs to be saved.
 * @param tag The tag
 * @param tab The tab position to save the object
 */
public static void save(PropertyPersistorModel prop, Object c, Object def,
						String tag, int tab)
    throws JCIOException
{
	if (c == null) {
		return;
	}
	PropertySaveModel plm = makeSaver(c, def, pkg, prop.getSubDirectory(),
	                                  prop.getType());
	if (plm != null) {
		plm.saveProperties(prop, tag, tab + tabIncrement);
	}
}

/**
 * Checks whether properties for the specific object c need to be saved (i.e.
 * they are different from their default values).
 *
 * For example:  val = check(chart.getChartArea(), defaultChartArea);
 *
 * If val is true, then some properties are different from their default values
 * and need to be saved.
 *
 * @param prop The property persistor
 * @param c The object whose properties are to be saved
 * @param def This object is typically used by implementors to determine whether
 *     a property really needs to be saved.
 * @return Whether one or more properties of object c need to be saved.
 */
public static boolean check(PropertyPersistorModel prop, Object c, Object def)
{
	if (c == null || def == null) {
		return(false);
	}
	PropertySaveModel plm = makeSaver(c, def, pkg, prop.getSubDirectory(),
	                                  prop.getType());
	if (plm != null) {
		return(plm.checkProperties(prop));
	}
	return(false);
}

/**
 * Refreshes static properties in the save class for the specific object c
 * using the provided property persistor prop.  The string tag is used to
 * refer to the properties to be saved.
 *
 * For example:  refresh(prop, chart.getChartArea(), defaultChartArea,
 *                       "chart.chartArea");
 *
 * @param prop The property persistor
 * @param c The object whose properties are to be saved
 * @param def This object is typically used by implementors to determine whether
 *     a property really needs to be saved.
 * @param tag The tag
 */
public static void refresh(PropertyPersistorModel prop, Object c, Object def,
						   String tag)
{
	if (c == null) {
		return;
	}
	PropertySaveModel plm = makeSaver(c, def, pkg, prop.getSubDirectory(),
	                                  prop.getType());
	if (plm != null && plm instanceof PropertyRefreshModel) {
		((PropertyRefreshModel)plm).refreshProperties(prop, tag);
	}
}

/**
 * For testing purposes only
 * @param args Command line arguments
 */
public static void main(String args[])
 {
	JCChart chart = null;
	try {
		chart = JCChartFactory.makeChartFromFile("test.html", "chart",
		                                         JCChartFactory.HTML);
	}
	catch (Exception e) {
		e.printStackTrace();
	}
	chart.setPreferredSize(new Dimension(700, 700));
	com.klg.jclass.util.swing.JCExitFrame f =
	    new com.klg.jclass.util.swing.JCExitFrame("Test");
	f.getContentPane().add(chart);
	f.pack();
	f.show();

	try {
        JCChartFactory.setOutputDataPropertiesOnChart(chart,
                                                      JCChartFactory.EMBED_DATA,
                                                      null);
        JCChartFactory.saveChartToFile(chart, "test.new.html",
                                       JCChartFactory.HTML);
	}
	catch (Exception e) {
		e.printStackTrace();
	}
}

}
