/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Anno.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:55 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;


import com.klg.jclass.util.JCNumberUtil;

import javax.vecmath.Matrix4d;
import javax.vecmath.Point2d;
import javax.vecmath.Point4d;
import javax.vecmath.Vector3d;
import java.awt.*;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.PathIterator;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;

/**
 * Internal class for JClass Chart 3D.  This class calculates the annotation
 * extents and also draws the annotation
 */
public class Anno implements java.io.Serializable {

static final long serialVersionUID = -5542417106751358122L;

/**
 * The amount of space between the tick and the annotation as a fraction of the
 * annotation height
 */
public final static double TICK_PAD_FACTOR = 0.15;

/** Used to translate font measurements to "cube" space (absolute space) */
public final static double CUBE_SCALE      = (2.0 / 1000.0);

/** Number of points in internal u point list */
public final static int NUM_POINTS = 4;

/** Number of vectors in internal v vector list */
public final static int NUM_VECTORS = 3;

/**
 * Calculates the orientation matrix.  This matrix orients the annotation plane
 * so that it faces the viewer most.
 */
protected Matrix4d orientationMatrix = null;

/**
 * The edge matrix is responsible for translating the annotation to the proper
 * edge of the cube.
 */
protected Matrix4d edgeMatrix        = null;

/**
 * The justify matrix positions the annotation so that it will not conflict
 * with the tick-marks and is properly centered on the tick mark.
 */
protected Matrix4d justifyMatrix     = null;

/**
 * The plane matrix is used to help find the best annotation plane.  The
 * planeMatrixInit matrix is the same for each plane matrix and thus needs
 * to only be computed once.
 */
protected Matrix4d planeMatrixInit   = null;

/**
 * Used to compute the annotation orientation for an axis (vertical or
 * horizontal) based on whether the horizontal or vertical component of the
 * axis is dominant.
 */
protected Matrix4d compMatrix        = null;

/** A general utility matrix */
protected Matrix4d matrix            = null;

/**
 * A matrix that is computed once per edge for a given axis and used for each
 * label on the axis.
 */
protected Matrix4d axisMatrix        = null;

/** List of points used during calculations */
protected Point4d[] u = null;

/** List of vector objects used during calculations */
protected Vector3d[] v = null;

/** If true, label picking is done instead of drawing */
protected boolean doPicking = false;

/** Point used for label picking */
protected Point pickPoint = null;

/** The label that the pickPoint falls in (null if no label picked) */
protected JCValueLabel selectedLabel = null;

/** The parent JCChart3dArea object */
protected JCChart3dArea area = null;

/**
 * No args constructor.  The parent JCChart3dArea is null.  The Anno object
 * should not be used if its parent is null.
 */
public Anno() {}

/**
 * Contstructor which sets the parent JCChart3dArea.
 *
 * @param area The parent JCChart3dArea
 */
public Anno(JCChart3dArea area)
{
    this.area = area;
}

/**
 * Gets the parent chart3d area.
 *
 * @return The parent JCChart3dArea object
 */
public JCChart3dArea getChart3dArea()
{
	return(area);
}

/**
 * Sets the parent chart3d area.
 *
 * @param area The new JCChart3dArea
 */
public void setChart3dArea(JCChart3dArea area)
{
	this.area = area;
}

/**
 * Initialize internal point array
 */
protected void initializePoints()
{
	if (u == null) {
		u = new Point4d[NUM_POINTS];
		for (int i = 0; i < NUM_POINTS; i++) {
			u[i] = new Point4d();
		}
	}
	if (v == null) {
		v = new Vector3d[NUM_VECTORS];
		for (int i = 0; i < NUM_VECTORS; i++) {
			v[i] = new Vector3d();
		}
	}
}

/**
 * Given a value and an integer specifying the number of fractional digits
 * to be used, returns a formatted string based on the value.
 *
 * @param value The value to be formatted
 * @param nfrac The number of fractional digits
 * @return The formatted string
 */
protected String formatAnno(double value, int nfrac)
{
    String s = JCNumberUtil.format(value, Math.max(0, nfrac));
	return(s);
}

/**
 * Returns an edge matrix based on the given edge of the cube.  This matrix is
 * responsible for translating the annotation to the proper edge and rotating
 * from the x-axis to the y or z axis (all annotation is assumed to start out
 * on the x-axis).
 *
 * @param edge The edge of the cube for which the edge Matrix is being
 *     calculated
 * @return The computed edge matrix.  Note that this matrix is also stored
 *     in the edgeMatrix attribute of the current object.
 */
public Matrix4d getEdgeMatrix(int edge)
{
	if (edgeMatrix == null) {
	    edgeMatrix = new Matrix4d();
	}
	edgeMatrix.setIdentity();

	JCPlotCube plotCube   = area.getPlotCube();
	double     normXScale = plotCube.getXNormalizedScale();
	double     normYScale = plotCube.getYNormalizedScale();
	double     normZScale = plotCube.getZNormalizedScale();
	switch (edge) {
	case Cube.YNEG_ZNEG:
		Transform.translateMatrix(edgeMatrix, 0.0, -normYScale, -normZScale);
	    break;
	case Cube.YNEG_ZPOS:
	    Transform.translateMatrix(edgeMatrix, 0.0, -normYScale,  normZScale);
	    break;
	case Cube.YPOS_ZNEG:
	    Transform.translateMatrix(edgeMatrix, 0.0,  normYScale, -normZScale);
	    break;
	case Cube.YPOS_ZPOS:
	    Transform.translateMatrix(edgeMatrix, 0.0,  normYScale,  normZScale);
	    break;
	case Cube.XNEG_ZNEG:
		Transform.translateMatrix(edgeMatrix, -normXScale, 0.0, -normZScale);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_90, JCAxis.AXIS_Z);
	    break;
	case Cube.XNEG_ZPOS:
	    Transform.translateMatrix(edgeMatrix, -normXScale, 0.0,  normZScale);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_90, JCAxis.AXIS_Z);
	    break;
	case Cube.XPOS_ZNEG:
	    Transform.translateMatrix(edgeMatrix,  normXScale, 0.0, -normZScale);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_90, JCAxis.AXIS_Z);
	    break;
	case Cube.XPOS_ZPOS:
	    Transform.translateMatrix(edgeMatrix,  normXScale, 0.0,  normZScale);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_90, JCAxis.AXIS_Z);
	    break;
	case Cube.XNEG_YNEG:
		Transform.translateMatrix(edgeMatrix, -normXScale, -normYScale, 0.0);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_270, JCAxis.AXIS_Y);
	    break;
	case Cube.XNEG_YPOS:
	    Transform.translateMatrix(edgeMatrix, -normXScale,  normYScale, 0.0);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_270, JCAxis.AXIS_Y);
	    break;
	case Cube.XPOS_YNEG:
	    Transform.translateMatrix(edgeMatrix,  normXScale, -normYScale, 0.0);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_270, JCAxis.AXIS_Y);
	    break;
	case Cube.XPOS_YPOS:
	    Transform.translateMatrix(edgeMatrix,  normXScale,  normYScale, 0.0);
		Transform.rotateMatrix(edgeMatrix, Transform.ROTATE_270, JCAxis.AXIS_Y);
	    break;
	default:
	    
	    break;
	}
	return(edgeMatrix);
}

/**
 * Returns the delta in <code>x</code> and <code>y</code> for a given edge
 * (based on the current transformation).
 *
 * @param edge The edge of the cube for which the deltas are to be calculated
 * @return Return the x and y deltas in a Point2d object.
 */
protected Point2d findEdgeDxDy(int edge)
{
    Point4d[] cube = area.getCube().getCubePoints();   // Projected cube points

    double dx = cube[Cube.edges[edge][1]].x - cube[Cube.edges[edge][0]].x;
    double dy = cube[Cube.edges[edge][1]].y - cube[Cube.edges[edge][0]].y;
    return(new Point2d(dx, dy));
}

/**
 * Finds the delta in <code>x</code> and <code>y</code> for a given edge based
 * on an axis.
 *
 * @param edge The edge of the cube for which the deltas are to be calculated
 * @return Return the x and y deltas in a Point2d object.
 */
protected Point2d findAxisDxDy(int edge)
{
	if (compMatrix == null) {
        Transform transform = area.getTransform();
		compMatrix = new Matrix4d(transform.getPerspective());
        compMatrix.mul(transform.getRotation());
	}

    // Note that xrt->sm has not been applied; thus the Y values are negated.
    double dx = 0.0;
    double dy = 0.0;
	switch (edge) {
	case Cube.YNEG_ZNEG:		/* X axes */
	case Cube.YNEG_ZPOS:
	case Cube.YPOS_ZNEG:
	case Cube.YPOS_ZPOS:
		dx =  compMatrix.m00;
		dy = -compMatrix.m10;
		break;
	case Cube.XNEG_ZNEG:		/* Y axes */
	case Cube.XNEG_ZPOS:
	case Cube.XPOS_ZNEG:
	case Cube.XPOS_ZPOS:
		dx =  compMatrix.m01;
		dy = -compMatrix.m11;
		break;
	case Cube.XNEG_YNEG:		/* Z axes */
	case Cube.XNEG_YPOS:
	case Cube.XPOS_YNEG:
	case Cube.XPOS_YPOS:
		dx =  compMatrix.m02;
		dy = -compMatrix.m12;
		break;
	default:
	    
	    break;
	}
    return(new Point2d(dx, dy));
}

/**
 * Calculates the orientation matrix.  This matrix orients the annotation plane
 * so that it faces the viewer most.  This plane is identified by finding which
 * of the unit axes has the largest magnitude Z component after the perspective,
 * rotation, and <code>edgeMatrix</code> transformations have been applied.
 * Also computes, whether to rotate the annotation, whether its above or below
 * the axis, whether to flip the annotation in the Y direction, and whether to
 * spin the axis title.  All this is stored in the justify object.
 *
 * @param edge The edge of the cube for which the orientation plane needs to
 *     be calculated
 * @param justify The justify object where information computed by this method
 *     is stored.
 * @return The computed orientation matrix.  Note that this matrix is also
 *     stored in the orientationMatrix attribute of the current object.
 */
public Matrix4d getOrientationMatrix(int edge, Justify justify)
{
    Transform transform = area.getTransform();
    Matrix4d  rotation  = transform.getRotation();

	if (orientationMatrix == null) {
	    orientationMatrix = new Matrix4d();
	}
    orientationMatrix.setIdentity();

    justify.setFlipX(false);
    justify.setFlipY(false);
    justify.setSpinTitle(false);

	// Find best annotation plane: matrix = [-y][pm][rm][edgeMatrix]
    if (planeMatrixInit == null) {
        planeMatrixInit = new Matrix4d();
        planeMatrixInit.setIdentity();
        Transform.scaleMatrix(planeMatrixInit, 1.0, -1.0, 1.0);
        planeMatrixInit.mul(transform.getPerspective());
        planeMatrixInit.mul(rotation);
    }
    matrix.mul(planeMatrixInit, edgeMatrix);

	initializePoints();
    u[0].set(0.0, 0.0, 0.0, 1.0);
    u[1].set(1.0, 0.0, 0.0, 1.0);
    u[2].set(0.0, 1.0, 0.0, 1.0);
    u[3].set(0.0, 0.0, 1.0, 1.0);

	for (int i = 0; i < NUM_POINTS; i++) {
		Transform.transformPoint(matrix, u[i]);
	}

	// Compute unit vectors v (one for x, y, and z)
	//
	// It is essential to renormalize the unit vectors since the perspective
	// foreshortening effect is a shear in x and y which stretches/shrinks
	// the actual length of the unit vector
	for (int i = 0; i < NUM_VECTORS; i++) {
		v[i].set(u[i+1].x - u[0].x, u[i+1].y - u[0].y, u[i+1].z - u[0].z);
		v[i].normalize();
	}

    double dx = 0.0;
    double dy = 0.0;

	double uy_z = v[1].z;
	double uz_z = v[2].z;

	switch (edge) {
	case Cube.YNEG_ZNEG:			/* X back-bottom */
        justify.setAbove(false);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m00;
		dy = rotation.m10;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_90,
                                   JCAxis.AXIS_X);
			if (-uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.YNEG_ZPOS:			/* X front-bottom */
        justify.setAbove(false);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m00;
		dy = rotation.m10;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_270,
                                   JCAxis.AXIS_X);
			if (uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.YPOS_ZNEG:			/* X back-top */
        justify.setAbove(true);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m00;
		dy = rotation.m10;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_270,
                                   JCAxis.AXIS_X);
			if (uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.YPOS_ZPOS:			/* X front-top */
        justify.setAbove(true);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m00;
		dy = rotation.m10;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_90,
                                   JCAxis.AXIS_X);
			if (-uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XNEG_ZNEG:			/* Y back-left */
        justify.setAbove(true);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m01;
		dy = rotation.m11;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to yz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_270,
                                   JCAxis.AXIS_X);
			if (uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XNEG_ZPOS:
        justify.setAbove(true);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m01;
		dy = rotation.m11;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to yz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_90,
                                   JCAxis.AXIS_X);
			if (-uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XPOS_ZNEG:
        justify.setAbove(false);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m01;
		dy = rotation.m11;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to yz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_90,
                                   JCAxis.AXIS_X);
			if (-uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XPOS_ZPOS:
        justify.setAbove(false);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m01;
		dy = rotation.m11;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from xy plane to yz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_270,
                                   JCAxis.AXIS_X);
			if (uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XNEG_YNEG:			/* Z bottom-left */
        justify.setAbove(false);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m02;
		dy = rotation.m12;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from yz plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_270,
                                   JCAxis.AXIS_X);
			if (uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XNEG_YPOS:
        justify.setAbove(true);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m02;
		dy = rotation.m12;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from yz plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_90,
                                   JCAxis.AXIS_X);
			if (-uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XPOS_YNEG:
        justify.setAbove(false);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m02;
		dy = rotation.m12;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from yz plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_90,
                                   JCAxis.AXIS_X);
			if (-uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	case Cube.XPOS_YPOS:
        justify.setAbove(true);
		// Load up axis dx,dy for justify.rotate determination
		dx = rotation.m02;
		dy = rotation.m12;

		// Decide best plane for display
		if (Math.abs(uz_z) < Math.abs(uy_z)) {
			// Change from yz plane to xz plane
            Transform.rotateMatrix(orientationMatrix, Transform.ROTATE_270,
                                   JCAxis.AXIS_X);
			if (uy_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
		else {
			if (uz_z < 0.0) {
                justify.setFlipY(true);
				dx = -dx;
			}
		}
	    break;
	default:
	    
	    break;
	}

	// Find rotation for best horizontal bias when annotation is drawn.
	// The annotation value calculations will force either vertical or
    // horizontal orientation.  Choose the closest.
    int rotate = Transform.ROTATE_NONE;
	if (justify.isVertical()) {
        rotate = (dy >= 0.0) ? Transform.ROTATE_90 : Transform.ROTATE_270;
	}
    else {
        rotate = (dx >= 0.0) ? Transform.ROTATE_NONE : Transform.ROTATE_180;
    }
    justify.setRotate(rotate);

	Point2d point = findEdgeDxDy(edge);
    if (justify.doFlipY()) {
        point.x = -point.x;
    }
    if (point.x < 0.0) {
        justify.setSpinTitle(true);
    }
    else {
        justify.setSpinTitle(false);
    }
	return(orientationMatrix);
}

/**
 * This method creates a justify matrix which positions the annotation so that
 * it will not conflict with the tick-marks, etc.  The annotation has a height
 * and a width. This routine expects to be called from a frame-of-reference
 * where <i>y</i> is positive downwards (ie. screen system). The origin of the
 * annotation string is on the left side of the string and at the midpoint in
 * height.	ie.
 *     0 <= x <= width;     -height/2 <= y <= height/2;
 *
 * TICK_PAD_FACTOR is intended to specify the amount of space between the tick
 * and the annotation as a fraction of the annotation height.	Note that the
 * tick size is half of the annotation height.
 *
 * The easeway variable returns the scalar distance from the axis to the
 * maximum extent of the annotation string.
 *
 * @param height The annotation height
 * @param width The annotation width
 * @param justify The object which stores information like whether the
 *     annotation is rotated, the annotation is above the axis, and if the
 *     annotation is flipped in the y direction.  It also  stores the computed
 *     easeway.
 * @return The computed justify matrix.  Note that this matrix is also stored
 *     in the justifyMatrix attribute of the current object.
 */
public Matrix4d getJustifyMatrix(double height, double width, Justify justify)
{
	double dx      = 0.0;
	double dy      = 0.0;
	double easeway = 0.0;
	switch (justify.getRotate()) {
	case Transform.ROTATE_NONE:
		dx = -width / 2.0;
		dy = height * (1.0 + TICK_PAD_FACTOR);
		if (justify.isAbove()) {
			dy = -dy;
		}
		easeway = height * (1.5 + TICK_PAD_FACTOR);
		break;
	case Transform.ROTATE_90:
		dy = height * (0.5 + TICK_PAD_FACTOR);
		if (justify.isAbove()) {
			dy = - width - dy;
		}
		easeway = width + height * (0.5 + TICK_PAD_FACTOR);
		break;
	case Transform.ROTATE_180:
		dx = width / 2.0;
		dy = height * (1.0 + TICK_PAD_FACTOR);
		if (justify.isAbove()) {
			dy = -dy;
		}
		easeway = height*(1.5 + TICK_PAD_FACTOR);
		break;
	case Transform.ROTATE_270:
		dy = -height * (0.5 + TICK_PAD_FACTOR);
		if (!justify.isAbove()) {
			dy = width - dy;
		}
		easeway = width + height * (0.5 + TICK_PAD_FACTOR);
		break;
	default:
	    
	    break;
	}
	justify.setEaseway(easeway);

	if (justifyMatrix == null) {
	    justifyMatrix = new Matrix4d();
	}
    justifyMatrix.setIdentity();
	Transform.translateMatrix(justifyMatrix, dx, dy, 0.0);
	Transform.rotateMatrix(justifyMatrix, justify.getRotate(), JCAxis.AXIS_Z);

	if (justify.doFlipY()) {
	    Transform.translateMatrix(justifyMatrix, width, 0.0, 0.0);
	    Transform.rotateMatrix(justifyMatrix, Transform.ROTATE_180,
							   JCAxis.AXIS_Y);
	}
	return(justifyMatrix);
}

/**
 * Finds the text width relative to the cube size.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis the text is on
 * @param text The text whose width is to be computed
 * @param isTitle Is this text from an axis title?
 * @param haveAnno Have the annotations already been computed for this axis?
 * @return The text width
 */
public double cubeTextWidth(Graphics gc, JCAxis axis, String text,
                            boolean isTitle, boolean haveAnno)
{
    if (gc == null || axis == null || text == null || text.length() == 0) {
		return(0);
    }

	Font font = (isTitle) ? axis.getTitleFont() : axis.getAnnoFont();
	if (font == null) {
		return(0);
	}

    Graphics2D g2 = (Graphics2D)gc;
	g2.setFont(font);
    FontRenderContext frc = g2.getFontRenderContext();
    TextLayout layout = new TextLayout(text, font, frc);

	Chart3dAnnoHandler  annoHandler  = axis.getAnnoHandler();
    Rectangle2D         bounds       = layout.getBounds();
	int                 fontCubeSize = 0;
	double              height       = bounds.getHeight();
	if (isTitle) {
		fontCubeSize = axis.getTitleFontCubeSize();
	}
	else {
		fontCubeSize = axis.getAnnoFontCubeSize();
        if (haveAnno) {
            // If the annotations have already been calculated use the
            // maximum annotation height of all the annotation.  Otherwise,
            // we must just use the current text's height
		    height = annoHandler.getAnnoHeight();
        }
	}
	if (!haveAnno) {
		
		
		
	}

	return(bounds.getWidth() * (double)fontCubeSize / height);
}

/**
 * Transforms and draws string on the cube.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis with which this string is associated
 * @param m The matrix with which to transform the shape determined by the
 *     string
 * @param text The string to draw
 * @param isTitle Are we drawing the title?
 */
protected void cubeDrawString(Graphics gc, JCAxis axis, Matrix4d m, String text,
                              Color foreground, boolean isTitle)
{
    if (gc == null || axis == null || text == null || text.length() == 0) {
		return;
    }

	Font font = (isTitle) ? axis.getTitleFont() : axis.getAnnoFont();
	if (font == null) {
		return;
	}

    Graphics2D g2 = (Graphics2D)gc;
	g2.setColor(foreground == null ? area.getPlotCube().getDrawForeground()
                                   : foreground);
	g2.setFont(font);
    FontRenderContext frc = g2.getFontRenderContext();
    TextLayout layout = new TextLayout(text, font, frc);

	Chart3dAnnoHandler  annoHandler  = axis.getAnnoHandler();
    Rectangle2D         bounds       = layout.getBounds();
	int                 fontCubeSize = 0;
	double              height       = 0.0;
	if (isTitle) {
		fontCubeSize = axis.getTitleFontCubeSize();
		height       = bounds.getHeight();
	}
	else {
		fontCubeSize = axis.getAnnoFontCubeSize();
		height       = annoHandler.getAnnoHeight();
	}

    double ty = isTitle ? -bounds.getY() - height / 2.0
				        : annoHandler.getAnnoShift();

    AffineTransform at = AffineTransform.getTranslateInstance(-bounds.getX(),
                                                              ty);
    Shape shape = layout.getOutline(at);
    PathIterator pi = shape.getPathIterator(null);

    double scale = (double)fontCubeSize * CUBE_SCALE / height;

    double v0, v1, w, w0, w1, w2, w3, w4, w5;

    GeneralPath gp = new GeneralPath(pi.getWindingRule());
    while (!pi.isDone()) {
        double[] coordinates = new double[6];
        int type  = pi.currentSegment(coordinates);
        switch(type) {
        case PathIterator.SEG_MOVETO:
            v0 = coordinates[0] * scale;
            v1 = coordinates[1] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w0 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w1 =  w *(m.m10*v0 + m.m11*v1 + m.m13);

            gp.moveTo((float)w0, (float)w1);
//          
            break;
        case PathIterator.SEG_LINETO:
            v0 = coordinates[0] * scale;
            v1 = coordinates[1] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w0 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w1 =  w *(m.m10*v0 + m.m11*v1 + m.m13);

            gp.lineTo((float)w0, (float)w1);
//          
            break;
        case PathIterator.SEG_QUADTO:
            v0 = coordinates[0] * scale;
            v1 = coordinates[1] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w0 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w1 =  w *(m.m10*v0 + m.m11*v1 + m.m13);

            v0 = coordinates[2] * scale;
            v1 = coordinates[3] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w2 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w3 =  w *(m.m10*v0 + m.m11*v1 + m.m13);

            gp.quadTo((float)w0, (float)w1,
                      (float)w2, (float)w3);
//          
            break;
        case PathIterator.SEG_CUBICTO:
            v0 = coordinates[0] * scale;
            v1 = coordinates[1] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w0 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w1 =  w *(m.m10*v0 + m.m11*v1 + m.m13);

            v0 = coordinates[2] * scale;
            v1 = coordinates[3] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w2 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w3 =  w *(m.m10*v0 + m.m11*v1 + m.m13);

            v0 = coordinates[4] * scale;
            v1 = coordinates[5] * scale;

			w  = 1.0/(m.m30*v0 + m.m31*v1 + m.m33);
			w4 =  w *(m.m00*v0 + m.m01*v1 + m.m03);
			w5 =  w *(m.m10*v0 + m.m11*v1 + m.m13);
            gp.curveTo((float)w0, (float)w1,
                       (float)w2, (float)w3,
                       (float)w4, (float)w5);
//          
            break;
        case PathIterator.SEG_CLOSE:
            gp.closePath();
//          
            break;
        default:
	        
	        break;
        }
        pi.next();
    }
    g2.fill(gp);
}

/**
 * Given a string and a transformation matrix, draws the tick and/or annotation
 * based on information in the passed in value label.  It uses the justify
 * object to position the label.
 *
 * This same method is called during picking.  Instead of drawing the annotation
 * the code computes whether the pick point falls within the label's boundary.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis for which the annotation is to be drawn
 * @param ctm Current transformation matrix (used to position the annotation)
 * @param vLabel The JCValueLabel object which stores the label, its background
 *     and foreground colors, and other info on how to draw the label
 * @param justify Object which has information about label positioning
 * @param height The height of the annotation
 * @param width The width of the annotation
 */
protected void drawAnnotation(Graphics gc, JCAxis axis, Matrix4d ctm,
							  JCValueLabel vLabel, Justify justify,
							  double height, double width)
{
	initializePoints();

	// Draw the tick mark (without j_mat)
	if (vLabel.isTickDrawn() && !doPicking) {
	    u[0].set(0.0, 0.0, 0.0, 1.0);
	    Transform.transformPoint(ctm, u[0]);

	    u[1].set(0.0, justify.isAbove() ? - height / 2.0 : height / 2.0,
                 0.0, 1.0);
	    if (vLabel.isMinorTick()) {
		    u[1].y *= 0.6;
		}
	    Transform.transformPoint(ctm, u[1]);
	    gc.drawLine((int)Math.round(u[0].x), (int)Math.round(u[0].y),
		            (int)Math.round(u[1].x), (int)Math.round(u[1].y));
	}

	if (vLabel.isLabelDrawn()) {
		String drawString = vLabel.getLabel();
	    if (drawString == null || drawString.length() == 0) {
		    return;
	    }

	    // Compose the annotation-specific transformation matrix. This matrix
	    // depends on the height & width of the specific annotation to position
	    // it properly with respect to its tick mark
	    justifyMatrix = getJustifyMatrix(height, width, justify);

	    // Apply the justification transformation
	    matrix.mul(ctm, justifyMatrix);

	    // Draw the string background or, if picking, test whether the point
		// is within the label's background area
		if (vLabel.getBackground() != null || doPicking) {
	        u[0].set(0.0,   -height / 2.0, 0.0, 1.0);
	        u[1].set(0.0,    height / 2.0, 0.0, 1.0);
	        u[2].set(width,  height / 2.0, 0.0, 1.0);
	        u[3].set(width, -height / 2.0, 0.0, 1.0);

			int[] xPoints = new int[NUM_POINTS];
			int[] yPoints = new int[NUM_POINTS];
			for (int i = 0; i < NUM_POINTS; i++) {
			    Transform.transformPoint(matrix, u[i]);
				xPoints[i] = (int)Math.round(u[i].x);
				yPoints[i] = (int)Math.round(u[i].y);
			}
			if (doPicking) {
				Polygon p = new Polygon(xPoints, yPoints, NUM_POINTS);
				if (p.contains(pickPoint)) {
					selectedLabel = vLabel;
				}
				return;
			}
            gc.setColor(vLabel.getBackground());
            gc.fillPolygon(xPoints, yPoints, NUM_POINTS);

            // Draw a line around the label (extra space, looks better)
            gc.drawPolyline(xPoints, yPoints, NUM_POINTS);
		}

	    // Draw the string
	    cubeDrawString(gc, axis, matrix, drawString, vLabel.getForeground(),
                       false);
	}
}

/**
 * Finds the extents of the bounding box determined by the given height and
 * width after it is transformed.
 *
 * @param ctm The current transformation matrix
 * @param height The height of the original "box"
 * @param width The width of the original "box"
 * @param extents The Extents object that stores the current viewport extents
 */
protected void boundBox(Matrix4d ctm, double height, double width,
                        Extents extents)
{
	initializePoints();
	u[0].set(0.0,   -height / 2.0, 0.0, 1.0);
	u[1].set(0.0,    height / 2.0, 0.0, 1.0);
	u[2].set(width,  height / 2.0, 0.0, 1.0);
	u[3].set(width, -height / 2.0, 0.0, 1.0);

	for (int i = 0; i < NUM_POINTS; i++) {
		Transform.transformPoint(ctm, u[i]);
		extents.checkXYExtents(u[i].x, u[i].y);
	}
}

/**
 * Binds the given text relative to the cube based on the axis font
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis that the given text is associated with
 * @param ctm The current transformation matrix
 * @param text The text to bound
 * @param offset The position offset along the axis for this text
 * @param extents The Extents object that stores the current viewport extents
 * @param justify Used to compute the justification matrix for this text
 */
protected void boundText(Graphics gc, JCAxis axis, Matrix4d ctm, String text,
			             double offset, Extents extents, Justify justify)
{
	double w      = cubeTextWidth(gc, axis, text, false, true);
	double width  = w * CUBE_SCALE;
	double height = (double) axis.getAnnoFontCubeSize() * CUBE_SCALE;

	justifyMatrix = getJustifyMatrix(height, width, justify);

	// matrix = [ctm][offset][orientation][-y][justify]
	matrix.set(ctm);
	Transform.translateMatrix(matrix, offset, 0.0, 0.0);
	matrix.mul(orientationMatrix);
	Transform.scaleMatrix(matrix, 1.0, -1.0, 1.0);
	matrix.mul(justifyMatrix);

	boundBox(matrix, height, width, extents);
}

/**
 * Binds the given title text relative to the cube based on the title font.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis associated with this title
 * @param ctm The current transformation matrix
 * @param text The text for this title
 * @param extents The Extents object that stores the current viewport extents
 * @param justify The Justify object that store whether to flip the title in the
 *    X or Y direction.
 */
protected void boundTitle(Graphics gc, JCAxis axis, Matrix4d ctm, String text,
                          Extents extents, Justify justify)
{
	double w          = cubeTextWidth(gc, axis, text, true, true);
	double width      = w * CUBE_SCALE;
	double height     = (double) axis.getTitleFontCubeSize() * CUBE_SCALE;
	Transform.translateMatrix(ctm, -width / 2.0, 0.0, 0.0);

	// justify.spinTitle should have no effect on the bounding box
	if (justify.doFlipY()) {
		Transform.translateMatrix(ctm, width, 0.0, 0.0);
		Transform.rotateMatrix(ctm, Transform.ROTATE_180, JCAxis.AXIS_Y);
	}
	if (justify.doFlipX()) {
		Transform.rotateMatrix(ctm, Transform.ROTATE_180, JCAxis.AXIS_X);
	}

	boundBox(ctm, height, width, extents);
}

/**
 * Since the min/max values are not necessarily representative of the
 * intermediate annotations, we assume that the step size will introduce at
 * least 1 extra digit of precision. The spaces included in the format strings
 * basically reserve space for 2 extra characters.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis to which the annotation belongs.
 */
protected double guessLongestAnnotation(Graphics gc, JCAxis axis)
{
	double min = axis.getAnnoMin();
	double max = axis.getAnnoMax();

	// Since we don't have nFrac yet, set it temporarily to a pretty good
	// estimate of its value.  Later the actual value used will be set in
	// calcAxisBounds(),
	axis.setNFrac(JCChart3dUtil.nicePrecision(max - min));
	Chart3dAnnoHandler annoHandler = axis.getAnnoHandler();

    double range = JCChart3dUtil.niceNum(max - min, false);
    double value = JCChart3dUtil.niceNum(min, false) + range / 10.0;
	JCValueLabel vLabel = annoHandler.getLabelFromValue(value, true, true,
                                                        false, false);
	String s = "00" + vLabel.getLabel();
	double w1 = cubeTextWidth(gc, axis, s, false, false);

	value = JCChart3dUtil.niceNum(max, true) - range / 10.0;
	vLabel = annoHandler.getLabelFromValue(value, true, true, false, false);
	s = "00" + vLabel.getLabel();
	double w2 = cubeTextWidth(gc, axis, s, false, false);
	return(Math.max(w1, w2));
}

/**
 * Calculates the tick increment and axis min. and max., given the font size,
 * window size and range of data.
 *
 * @param axis The axis to which this annotation belongs
 * @param fontSize The font size (unit coord)
 * @param windowSize The window size (unit coord) in approp. dimension
 * @param scale A scale for determining how many annotations on an axis.  The
 *     bigger the value the less annotations.
 * @param tight We want tight bounds in the Z direction for bars
 * @param min The min value for this axis
 * @param max The max value for this axis
 */
protected void calcAxisBounds(JCAxis axis, double fontSize, double windowSize,
			                  double scale, boolean tight, double min, double max)
{
	// How many will fit?  Note that we wish to place ticks halfway between
	// these numbers, so be generous.  We need at least two labels, regardless
	// of whether they will fit or not.
	int nLabels = 2;
    if ((1.0 + scale * fontSize) != 1.0) {
		nLabels = (int) Math.round(JCChart3dUtil.ceil((windowSize + fontSize) /
						                              (scale * fontSize)));
	    if (nLabels < 2) {
		    nLabels = 2;
	    }
    }
    
    
    
    

    double range      = max - min;
	double microRange = range / 2000.0;	// Rounding factor of 1/2th pixel for
									    // 1000 line window
    
    

	double step  = Double.MAX_VALUE;
	int    nFrac = 0;
	if (1.0 + range == 1.0) {
	    // If min == max we have a flat surface.  Arbitrarily set to +-1
		step  = 2.0;
		min   -= 1.0;
		max   += 1.0;
	}
	else {
	    double mul = JCChart3dUtil.niceNum(range/((double)(nLabels - 1)), true);
	    if (mul > range) {
		    mul = JCChart3dUtil.niceNum(range/2.0, true);
		}
	    double hmul = mul / 2.0;

	    step = mul;
		double lo = min;
		double hi = max;
	    if (!tight) {
			lo = hmul * JCChart3dUtil.floor((min + microRange) / hmul);
			hi = hmul * JCChart3dUtil.ceil((max - microRange) / hmul);
		}
	    if ((min - lo) >= hmul / 2) {
		    lo = min;
		}
	    if ((hi - max) >= hmul / 2) {
		    hi = max;
		}
		min = lo;
		max = hi;

	    // Compute the display precision & format
	    nFrac = JCChart3dUtil.nDigits(mul);
	}
    
    
    
    
	axis.setAnnoMin(min);
	axis.setAnnoMax(max);
	axis.setAnnoStep(step);
	axis.setNFrac(nFrac);
}

/**
 * Do we draw annotation parallel to the axis (vertical == false) or
 * orthogonal to it (vertical == true)
 *
 * @param edge The edge/axis for which the annotation orientation is to be
 *     determined
 * @param justify Stores the results of orientation determination
 */
public void determineVertical(int edge, Justify justify)
{
	justify.setVertical(true);

	JCAxis axis = null;
	switch(edge) {
	case Cube.YNEG_ZNEG:
	    axis = area.getAxis(JCAxis.AXIS_X);
		break;
	case Cube.XNEG_ZNEG:
	    axis = area.getAxis(JCAxis.AXIS_Y);
		break;
	case Cube.XNEG_YNEG:
	    axis = area.getAxis(JCAxis.AXIS_Z);
		break;
	default:
	    
        return;
	}
    int rotation = axis.getAnnotationRotation();
    if (rotation != JCAxis.ANNOTATION_ROTATION_DEFAULT) {
        if (rotation == JCAxis.ANNOTATION_ROTATION_HORIZONTAL) {
            justify.setVertical(false);
        }
        return;
    }
	if (axis.getMaxAnnoLength() > 1  &&
		axis.getAnnotationMethod() != JCAxis.ANNOTATION_VALUES)
	{
		return;
	}

	boolean[] annotateEdge = area.getCube().getAnnotateEdge();
	int n = 0;
	for (int i = edge; i <= edge + 3; i++) {
		if (annotateEdge[i]) {
			n++;
		}
	}

	Point2d point = null;
	if (n == 1) {
		// Use the actual cube edge for better accuracy
		point = findEdgeDxDy(edge);
	}
	else {
		// Compromise and use the unit axis
		point = findAxisDxDy(edge);
	}

	if (point != null) {
	   justify.setVertical(Math.abs(point.y) > Math.abs(point.x));
	}
}

/**
 * If the annotation method is ANNOTATION_VALUES, the axis may need to get
 * extended if the annotation overhangs the end of the axis.  Adjust the
 * dataport to reflect this.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis whose dataport might be extend
 * @param dataport The object which stores the current dataport extents
 * @param scale The axis normalized scale
 * @param tight We want tight bounds in the Z direction for bars
 * @param vertical Is the annotation orientation vertical or horizontal?
 * @param axisMin The axis min value
 * @param axisMax The axis max value
 */
public void extendAxisDataport(Graphics gc, JCAxis axis, JCBox dataport,
                               double scale, boolean tight, boolean vertical,
							   double axisMin, double axisMax)
{
	if (axis.getAnnotationMethod() != JCAxis.ANNOTATION_VALUES) {
		return;
	}

    double projLength = 0.0; // Length of an annotation when projected
		                     // against the given axis (usually font height
							 // or text length)
	if (vertical) {
		projLength = axis.getAnnoFontCubeSize() * CUBE_SCALE;
		calcAxisBounds(axis, projLength, 2.0 * scale, 2.0, tight, axisMin, axisMax);
	}
	else {
		projLength = guessLongestAnnotation(gc, axis) * CUBE_SCALE;
		calcAxisBounds(axis, projLength, 2.0 * scale, 1.5, tight, axisMin, axisMax);
	}

	// Now that we have the annotation min and max, make the labels
	axis.getAnnoHandler().makeAnnotations(gc, this);

	// Axis min/max may have been adjusted, so refresh values.
    axisMin = axis.getAnnoMin();
    axisMax = axis.getAnnoMax();

    double dataportMin = axisMin;
    double dataportMax = axisMax;

    ArrayList valueLabels = axis.getAnnoHandler().getValueLabels();
    if (valueLabels == null || valueLabels.size() == 0) {
        return;
    }

    double minPos   = Double.MAX_VALUE;
    double maxPos   = Double.MIN_VALUE;
	String minLabel = null;
	String maxLabel = null;
    for (int i = 0; i < valueLabels.size(); i++) {
        JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
        double value = vLabel.getValue();
        String label = vLabel.getLabel();
        if (axisMin <= value && value <= axisMax && vLabel.isLabelDrawn()) {
            double pos = scale * (1.0 - 2.0 * (axisMax - value) / (axisMax - axisMin));
			if (pos < minPos) {
				minPos   = pos;
				minLabel = label;
			}
			if (pos > maxPos) {
				maxPos   = pos;
				maxLabel = label;
			}
        }
    }
    if (minPos != Double.MAX_VALUE && !tight) {
        // If any of the annotations would overhang the axis, then
	    // extend the axis in data space, so that they don't.  This
		// avoids annotation collisions.  Add an extra "." to the
		// actual label to leave some "breathing room".
		projLength =
            cubeTextWidth(gc, axis, minLabel + ".", false, true) * CUBE_SCALE;
		double projVal = minPos - projLength / 2.0;
		if (projVal < -scale) {
			double value = (projVal - scale) *
						   (axisMax - axisMin) /
						   (2.0 * scale) + axisMax;
			dataportMin = Math.min(dataportMin, value);
		}

		projLength =
            cubeTextWidth(gc, axis, maxLabel + ".", false, true) * CUBE_SCALE;
		projVal = maxPos + projLength / 2.0;
		if (projVal > scale) {
			double value = (projVal - scale) *
						   (axisMax - axisMin) /
						   (2.0 * scale) + axisMax;
			dataportMax = Math.max(dataportMax, value);
		}
	}

    switch (axis.getAxisId()) {
    case JCAxis.AXIS_X:
	    dataportMax    = Math.max(dataportMax, dataport.x + dataport.width);
        dataport.x     = Math.min(dataportMin, dataport.x);
		dataportMin    = dataport.x;
        dataport.width = dataportMax - dataport.x;
        break;
    case JCAxis.AXIS_Y:
	    dataportMax     = Math.max(dataportMax, dataport.y + dataport.height);
        dataport.y      = Math.min(dataportMin, dataport.y);
		dataportMin    = dataport.y;
        dataport.height = dataportMax - dataport.y;
        break;
    case JCAxis.AXIS_Z:
	    dataportMax    = Math.max(dataportMax, dataport.z + dataport.depth);
        dataport.z     = Math.min(dataportMin, dataport.z);
		dataportMin    = dataport.z;
        dataport.depth = dataportMax - dataport.z;
        break;
    default:
        
        break;
    }
}

/**
 * Adjust the dataport for a given axis for annotation only if one of the
 * four edges are annotated.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis whose dataport might be adjusted
 * @param dataport The object which stores the current dataport extents
 * @param startEdge The starting edge for this axis
 * @param scale The axis normalized scale
 * @param tight We want tight bounds in the Z direction for bars
 * @param justify Justify object that has the anno orientation information we need
 * @param axisMin The axis min value
 * @param axisMax The axis max value
 */
public boolean adjustDataport(Graphics gc, JCAxis axis, JCBox dataport,
                              int startEdge, double scale, boolean tight,
							  Justify justify, double axisMin, double axisMax)
{
	boolean[] annotateEdge = area.getCube().getAnnotateEdge();

	if (annotateEdge[startEdge]     ||
		annotateEdge[startEdge + 1] ||
		annotateEdge[startEdge + 2] ||
		annotateEdge[startEdge + 3])
	{
		extendAxisDataport(gc, axis, dataport, scale, tight,
						   justify.isVertical(), axisMin, axisMax);
	}
	return(false);
}

/**
 * Finds the extents of the given axis modifying the viewport bounds and
 * dataport bounds as required.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis to be bounded
 * @param ctm The current transformation matrix
 * @param extents The current viewport extents
 * @param dataport The current cube dataport
 */
protected void bound3dAxis(Graphics gc, JCAxis axis, Matrix4d ctm,
                           Extents extents, JCBox dataport)
{
	JCPlotCube       plotCube       = area.getPlotCube();
	boolean[]        annotateEdge   = area.getCube().getAnnotateEdge();
    Chart3dDataView  dataView       = area.getChart3d().getDataView(0);
	Justify          justify        = new Justify();

    int     startEdge = -1;;
    double  scale     = 1.0;
	boolean tight     = false;
    switch (axis.getAxisId()) {
    case JCAxis.AXIS_X:
        startEdge = Cube.YNEG_ZNEG;
        scale     = plotCube.getXNormalizedScale();
        break;
    case JCAxis.AXIS_Y:
        startEdge = Cube.XNEG_ZNEG;
        scale     = plotCube.getYNormalizedScale();
        break;
    case JCAxis.AXIS_Z:
        startEdge = Cube.XNEG_YNEG;
        scale     = plotCube.getZNormalizedScale();
		tight     = (dataView.getChartType() == JCChart3d.BAR);
        break;
    default:
	    
        return;
    }

	// Use previous computed annotation min/max (from data values)
	double axisMin = axis.getAnnoMin();
	double axisMax = axis.getAnnoMax();

	// Note that order of calculation is very important in the next section.
	// If the annotation type not values (i.e. data or value labels), the
	// maximum annotation length affects the vertical determination.  Thus we
	// must calculate the annotation first.  If the annotation type is values,
	// the vertical determination affects any dataport extensions so it must
	// be done first.
	boolean makeAnno = true;
	if (axis.getAnnotationMethod() != JCAxis.ANNOTATION_VALUES) {
	    axis.getAnnoHandler().makeAnnotations(gc, this);
		axis.calcMaxAnnoLength();
		makeAnno = false;
	}
	determineVertical(startEdge, justify);
	if (axis.getAnnotationMethod() == JCAxis.ANNOTATION_VALUES) {
	    makeAnno = adjustDataport(gc, axis, dataport, startEdge, scale,
					              tight, justify, axisMin, axisMax);
	}

	if (annotateEdge[startEdge]     ||
		annotateEdge[startEdge + 1] ||
		annotateEdge[startEdge + 2] ||
		annotateEdge[startEdge + 3])
	{
		if (makeAnno) {
            double  projLength = 0.0; // Length of an annotation when projected
								      // against the given axis (usually font
								      // height or text length)
			if (justify.isVertical()) {
			    projLength = axis.getAnnoFontCubeSize() * CUBE_SCALE;
				calcAxisBounds(axis, projLength, 2.0 * scale, 2.0, tight,
				               axisMin, axisMax);
			}
			else {
			    projLength = guessLongestAnnotation(gc, axis) * CUBE_SCALE;
			    calcAxisBounds(axis, projLength, 2.0 * scale, 1.5, tight,
				               axisMin, axisMax);
			}

			// Now that we have the annotation min and max, make the labels
		    axis.getAnnoHandler().makeAnnotations(gc, this);
		}

		// Axis min/max may have been adjusted, so refresh values.
        axisMin = axis.getAnnoMin();
        axisMax = axis.getAnnoMax();

		for (int edge = startEdge; edge <= startEdge + 3; edge++) {
			if (annotateEdge[edge]) {
                edgeMatrix = getEdgeMatrix(edge);
                orientationMatrix = getOrientationMatrix(edge, justify);
                axisMatrix.mul(ctm, edgeMatrix);
                double easeway = 0.0;

                ArrayList valueLabels = axis.getAnnoHandler().getValueLabels();
                if (valueLabels == null || valueLabels.size() == 0) {
                    continue;
                }
               for (int i = 0; i < valueLabels.size(); i++) {
                    JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
                    double value = vLabel.getValue();
                    String label = vLabel.getLabel();
                    if (axisMin <= value && value <= axisMax && vLabel.isLabelDrawn()) {
                        double pos = scale *
                            (1.0 - 2.0 * (axisMax - value) / (axisMax - axisMin));
                        boundText(gc, axis, axisMatrix, label, pos,
								  extents, justify);
                        easeway = Math.max(justify.getEaseway(), easeway);
                    }
                }
				if (axis.getTitle() != null) {
					easeway += axis.getTitleFontCubeSize() * CUBE_SCALE;
                    axis.setAnnoEaseway(easeway);
					if (justify.isAbove()) {
                        easeway = -easeway;
					}

					/* matrix = [axisMatrix][orientationMatrix][-y][easeway] */
                    matrix.mul(axisMatrix, orientationMatrix);
                    Transform.scaleMatrix(matrix, 1.0, -1.0, 1.0);
                    Transform.translateMatrix(matrix, 0.0, easeway, 0.0);
                    boundTitle(gc, axis, matrix, axis.getTitle(), extents,
                               justify);
				}
			}
		}
	}
}

/**
 * In 3d mode, draws the annotation for a given axis.  The internal value
 * labels, generated earlier, stores the annotation.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis to be drawn
 * @param ctm The current transformation matrix
 */
protected void draw3dAxis(Graphics gc, JCAxis axis, Matrix4d ctm)
{
	JCPlotCube       plotCube       = area.getPlotCube();
	boolean[]        annotateEdge   = area.getCube().getAnnotateEdge();
	Justify          justify        = new Justify();

    int     startEdge = -1;;
    double  scale     = 1.0;
    switch (axis.getAxisId()) {
    case JCAxis.AXIS_X:
        startEdge = Cube.YNEG_ZNEG;
        scale     = plotCube.getXNormalizedScale();
        break;
    case JCAxis.AXIS_Y:
        startEdge = Cube.XNEG_ZNEG;
        scale     = plotCube.getYNormalizedScale();
        break;
    case JCAxis.AXIS_Z:
        startEdge = Cube.XNEG_YNEG;
        scale     = plotCube.getZNormalizedScale();
        break;
    default:
	    
        return;
    }

	double annoHeight  = axis.getAnnoFontCubeSize() * CUBE_SCALE;
	determineVertical(startEdge, justify);
	for (int edge = startEdge; edge <= startEdge + 3; edge++) {
		if (annotateEdge[edge]) {
			// Init the annotation basis
			double annoMin = axis.getAnnoMin();
			double annoMax = axis.getAnnoMax();
			double dMin    = axis.getDataportMin();
			double dMax    = axis.getDataportMax();

			edgeMatrix = getEdgeMatrix(edge);
			orientationMatrix = getOrientationMatrix(edge, justify);
			axisMatrix.mul(ctm, edgeMatrix);

            ArrayList valueLabels = axis.getAnnoHandler().getValueLabels();
            if (valueLabels == null || valueLabels.size() == 0) {
                continue;
            }
            for (int i = 0; i < valueLabels.size(); i++) {
                JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
                double value = vLabel.getValue();
                if (annoMin <= value && value <= annoMax) {
                    String label = vLabel.getLabel();

                    // matrix = [ctm][edgeMatrix][pos][orientationMatrix][-y]
					matrix.set(axisMatrix);
					double pos =
                        (2.0 * (value - dMin) / (dMax - dMin) - 1.0) * scale;
					Transform.translateMatrix(matrix, pos, 0.0, 0.0);
					matrix.mul(orientationMatrix);
					Transform.scaleMatrix(matrix, 1.0, -1.0, 1.0);

                    double w = cubeTextWidth(gc, axis, label, false, true);
					double width = w * CUBE_SCALE;
					drawAnnotation(gc, axis, matrix, vLabel, justify,
					               annoHeight, width);
               }
            }
            if (axis.getTitle() != null && !doPicking) {
                // matrix = [ctm][edgeMatrix][orientationMatrix][-y]
				matrix.mul(axisMatrix, orientationMatrix);
				Transform.scaleMatrix(matrix, 1.0, -1.0, 1.0);

                double w = cubeTextWidth(gc, axis, axis.getTitle(), true, true);
                double width = w * CUBE_SCALE;
                double easeway = axis.getAnnoEaseway();
                if (justify.above) {
                    easeway = -easeway;
                }
                Transform.translateMatrix(matrix, -width/2.0, easeway, 0.0);

                if (justify.doSpinTitle()) {
                    Transform.translateMatrix(matrix, width, 0.0, 0.0);
                    Transform.rotateMatrix(matrix, Transform.ROTATE_180,
                                           JCAxis.AXIS_Z);
                }
                if (justify.doFlipY()) {
                    Transform.translateMatrix(matrix, width, 0.0, 0.0);
                    Transform.rotateMatrix(matrix, Transform.ROTATE_180,
                                           JCAxis.AXIS_Y);
                }
                if (justify.doFlipX()) {
                    Transform.rotateMatrix(matrix, Transform.ROTATE_180,
                                           JCAxis.AXIS_X);
                }
				cubeDrawString(gc, axis, matrix, axis.getTitle(), null, true);
            }
		}
	}
}

/**
 * In 2d mode, draws the annotation for a given axis.  The internal value
 * labels, generated earlier, stores the annotation.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param axis The axis to be drawn
 * @param ctm The current transformation matrix
 */
protected void draw2dAxis(Graphics gc, JCAxis axis, Matrix4d ctm)
{
	double           w, annoWidth, titleWidth, xTrans, yTrans, pos;
	boolean          isXAxis, vertical;
	JCPlotCube       plotCube = area.getPlotCube();
	double           xScale   = plotCube.getXNormalizedScale();
	double           yScale   = plotCube.getYNormalizedScale();
	Justify          justify  = new Justify();

	justify.setFlipX(false);
	justify.setFlipY(false);
    int rotation = axis.getAnnotationRotation();
    switch (axis.getAxisId()) {
    case JCAxis.AXIS_X:
	    isXAxis = true;
        vertical = (rotation == JCAxis.ANNOTATION_ROTATION_VERTICAL);
		justify.setAbove(false);
        break;
    case JCAxis.AXIS_Y:
	    isXAxis = false;
        vertical = (rotation == JCAxis.ANNOTATION_ROTATION_DEFAULT ||
                    rotation == JCAxis.ANNOTATION_ROTATION_VERTICAL);
		justify.setAbove(true);
        break;
    case JCAxis.AXIS_Z:
    default:
	    
        return;
    }
    justify.setRotate(vertical ? Transform.ROTATE_90 : Transform.ROTATE_NONE);

	double  annoHeight  = axis.getAnnoFontCubeSize()  * CUBE_SCALE;
	double  titleHeight = axis.getTitleFontCubeSize() * CUBE_SCALE;
	if (axis.isShowing()) {
        double axisMin = axis.getAnnoMin();
        double axisMax = axis.getAnnoMax();

        ArrayList valueLabels = axis.getAnnoHandler().getValueLabels();
        if (valueLabels != null && valueLabels.size() > 0) {
            for (int i = 0; i < valueLabels.size(); i++) {
                JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
                double value = vLabel.getValue();
                String label = vLabel.getLabel();
                if (axisMin <= value && value <= axisMax) {
					w = cubeTextWidth(gc, axis, label, false, true);
					annoWidth = w * CUBE_SCALE;

					axisMatrix.set(ctm);
					pos = 2.0 * (value - axisMin) / (axisMax - axisMin) - 1.0;
					if (isXAxis) {
					    xTrans = pos * xScale;
						yTrans = -yScale;
					}
					else {
						xTrans = -xScale;
						yTrans = pos * yScale;
					}
					Transform.translateMatrix(axisMatrix, xTrans, yTrans, 0.0);
					Transform.scaleMatrix(axisMatrix, 1.0, -1.0, 1.0);
					if (!isXAxis) {
					    Transform.rotateMatrix(axisMatrix, Transform.ROTATE_270,
										       JCAxis.AXIS_Z);
					}

					drawAnnotation(gc, axis, axisMatrix, vLabel, justify,
								   annoHeight, annoWidth);
                }
            }
        }

		if (axis.getTitle() != null && !doPicking) {
            w = cubeTextWidth(gc, axis, axis.getTitle(), true, true);
			titleWidth = w * CUBE_SCALE;

			axisMatrix.set(ctm);
			Transform.scaleMatrix(axisMatrix, 1.0, -1.0, 1.0);
			if (isXAxis) {
				xTrans = -titleWidth / 2.0;
				yTrans = yScale + axis.getAnnoEaseway() + titleHeight / 2.0;
			}
			else {
				xTrans = -xScale - axis.getAnnoEaseway() - titleHeight / 2.0;
				yTrans = titleWidth / 2.0;
			}
			Transform.translateMatrix(axisMatrix, xTrans, yTrans, 0.0);
			if (!isXAxis) {
			    Transform.rotateMatrix(axisMatrix, Transform.ROTATE_270,
								       JCAxis.AXIS_Z);
			}

			cubeDrawString(gc, axis, axisMatrix, axis.getTitle(), null, true);
		}
	}
}

/**
 * Resets often used matrices to null so that they don't contain any info
 * that might have come from previous use.
 */
protected void annoInit()
{
	orientationMatrix = null;
	edgeMatrix        = null;
	justifyMatrix     = null;
	planeMatrixInit   = null;
	compMatrix        = null;
	if (matrix == null || axisMatrix == null) {
		matrix     = new Matrix4d();
		axisMatrix = new Matrix4d();
	}
}

/**
 * Caculates annotation extents for either the 2D or 3D plots (drawn with the
 * Java 2D api).
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param viewport The rectangular viewport into which the graph is to be drawn
 * @param dataport The cube's data extents
 */
protected void calcAnno(Graphics gc, JCBox viewport, JCBox dataport)
{
    annoInit();
	if (area.getChart3d().isProjection()) {
		calc2dAnno(gc, viewport, dataport);
	}
	else {
		calc3dAnno(gc, viewport, dataport);
	}
}

/**
 * Calculates annotation extents for 3D plots (drawn with the Java 2D api).
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param viewport The rectangular viewport into which the graph is to be drawn
 * @param dataport The cube's data extents
 */
protected void calc3dAnno(Graphics gc, JCBox viewport, JCBox dataport)
{
    Cube cube = area.getCube();
	cube.calc3DAnnoEdges();
    Point4d[] cubePoints = cube.getCubePoints();

	Extents extents = new Extents();
	for (int i = 0; i < Cube.NUM_POINTS; i++) {
		extents.checkXYExtents(cubePoints[i].x, cubePoints[i].y);
	}

	Chart3dDataView dataView = area.getChart3d().getDataView(0);
	Chart3dData data = dataView.getElevationData();

	// Make a first pass at the anno limits.  These values will be over-written
	// by the calcAxisBounds method based on the annotation range chosen.
	// However, these limits are needed by other places before then and are
	// also used to caculate the axis bounds.
	data.calcAnnoLimits();

	JCAxis xAxis = area.getAxis(JCAxis.AXIS_X);
	JCAxis yAxis = area.getAxis(JCAxis.AXIS_Y);
	JCAxis zAxis = area.getAxis(JCAxis.AXIS_Z);

    // Initialize dataport to the original anno min & max (basically the data
	// min & max with a few modifications).
    dataport.x      = xAxis.getAnnoMin();
    dataport.y      = yAxis.getAnnoMin();
    dataport.z      = zAxis.getAnnoMin();
    dataport.width  = xAxis.getAnnoMax() - xAxis.getAnnoMin();
    dataport.height = yAxis.getAnnoMax() - yAxis.getAnnoMin();
    dataport.depth  = zAxis.getAnnoMax() - zAxis.getAnnoMin();

	// Create a 'text' viewing matrix without the final viewport transform.
	// textMatrix = [-y][pm][rm] -- Note: not equal to Transform.textView
	Transform transform = area.getTransform();
	Matrix4d textMatrix = new Matrix4d();
	textMatrix.setIdentity();
	Transform.scaleMatrix(textMatrix, 1.0, -1.0, 1.0);
	textMatrix.mul(transform.getPerspective());
	textMatrix.mul(transform.getRotation());

    // Extend the viewport and dataport bounds by considering each axis in turn
    bound3dAxis(gc, zAxis, textMatrix, extents, dataport);
    bound3dAxis(gc, xAxis, textMatrix, extents, dataport);
    bound3dAxis(gc, yAxis, textMatrix, extents, dataport);

	// Set dataport bounds
	xAxis.setDataportMin(dataport.x);
	yAxis.setDataportMin(dataport.y);
	zAxis.setDataportMin(dataport.z);
	xAxis.setDataportMax(dataport.x + dataport.width);
	yAxis.setDataportMax(dataport.y + dataport.height);
	zAxis.setDataportMax(dataport.z + dataport.depth);

//	JCChart3dUtil.printAxisExtents(xAxis);
//	JCChart3dUtil.printAxisExtents(yAxis);
//	JCChart3dUtil.printAxisExtents(zAxis);

	// Return the viewport's extents as a rectangle
	viewport.x		= extents.getXMin();
	viewport.y		= -extents.getYMax();
	viewport.width 	= extents.getXMax() - extents.getXMin();
	viewport.height = extents.getYMax() - extents.getYMin();
}

/**
 * Draws annotation for 3d plots (drawn with the Java 2D api).
 *
 * @param gc The graphics object to use for drawing and getting font info
 */
public void draw3dAnno(Graphics gc)
{
	if (gc == null) {
	    return;
	}
    gc.setColor(area.getPlotCube().getDrawForeground());

    Matrix4d textView = area.getTransform().getTextView();
	draw3dAxis(gc, area.getAxis(JCAxis.AXIS_Z), textView);
	draw3dAxis(gc, area.getAxis(JCAxis.AXIS_X), textView);
	draw3dAxis(gc, area.getAxis(JCAxis.AXIS_Y), textView);
}

/**
 * Calculates the extents of the annotation for 2D projection plots.
 *
 * @param gc The graphics object to use for drawing and getting font info
 * @param viewport The rectangular viewport into which the graph is to be drawn
 * @param dataport The cube's data extents
 */
protected void calc2dAnno(Graphics gc, JCBox viewport, JCBox dataport)
{
    int            rotation;
    double         axisMin, axisMax, newMin, newMax, annoWidth, easeway, wMax;
    boolean        vertical;
    ArrayList      valueLabels;

    Cube cube = area.getCube();
	cube.calc2DAnnoEdges();

	JCPlotCube  plotCube = area.getPlotCube();
	double      xScale   = plotCube.getXNormalizedScale();
	double      yScale   = plotCube.getYNormalizedScale();
	Extents     extents  = new Extents(-xScale, xScale, -yScale, yScale);

	Chart3dDataView dataView = area.getChart3d().getDataView(0);
	Chart3dData data = dataView.getElevationData();
	data.calcAnnoLimits();

	JCAxis xAxis = area.getAxis(JCAxis.AXIS_X);
	JCAxis yAxis = area.getAxis(JCAxis.AXIS_Y);
	JCAxis zAxis = area.getAxis(JCAxis.AXIS_Z);

    // Initialize dataport to the original anno min & max (basically the data
	// min & max with a few modifications).
    dataport.x      = xAxis.getAnnoMin();
    dataport.y      = yAxis.getAnnoMin();
    dataport.z      = 0.0;
    dataport.width  = xAxis.getAnnoMax() - xAxis.getAnnoMin();
    dataport.height = yAxis.getAnnoMax() - yAxis.getAnnoMin();
    dataport.depth  = 0.0;

	/**********/
	/* X AXIS */
	/**********/

    Justify justify     = new Justify();
	double  annoHeight  = xAxis.getAnnoFontCubeSize()  * CUBE_SCALE;
	double  titleHeight = xAxis.getTitleFontCubeSize() * CUBE_SCALE;
	if (xAxis.isShowing()) {

		// Compute extents of annotations in the x-direction
        rotation = xAxis.getAnnotationRotation();
        vertical = (rotation == JCAxis.ANNOTATION_ROTATION_VERTICAL);
        if (vertical) {
            annoWidth = xAxis.getAnnoFontCubeSize() * CUBE_SCALE;
            calcAxisBounds(xAxis, annoWidth, 2.0 * xScale, 2.0, false,
                           xAxis.getAnnoMin(), xAxis.getAnnoMax());
        }
        else {
            annoWidth = guessLongestAnnotation(gc, xAxis) * CUBE_SCALE;
            calcAxisBounds(xAxis, annoWidth, 2.0 * xScale, 1.5, false,
                           xAxis.getAnnoMin(), xAxis.getAnnoMax());
        }

		// Now that we have the annotation min and max, make the labels
		xAxis.getAnnoHandler().makeAnnotations(gc, this);

		// Axis min/max may have been adjusted, so get latest values
        axisMin = xAxis.getAnnoMin();
        axisMax = xAxis.getAnnoMax();

		// Update the dataport (data to cube transform info calculated in
        // calcAxisBounds())
        dataport.x     = axisMin;
        dataport.width = axisMax - axisMin;

		wMax = 0.0;
        valueLabels = xAxis.getAnnoHandler().getValueLabels();
        if (valueLabels != null && valueLabels.size() > 0) {
            for (int i = 0; i < valueLabels.size(); i++) {
                JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
                double value = vLabel.getValue();
                String label = vLabel.getLabel();
                if (axisMin <= value && value <= axisMax) {
                    double w = cubeTextWidth(gc, xAxis, label, false, true);
                    if (!vertical) {
                        double pos = xScale * (1.0 - 2.0 * (axisMax - value) /
                                               (axisMax - axisMin));
                        // Bound min
                        newMin = pos - w * CUBE_SCALE / 2.0;
                        extents.checkXExtents(newMin);
                        // Bound max
                        newMax = pos + w * CUBE_SCALE / 2.0;
                        extents.checkXExtents(newMax);
                    }
                    wMax = Math.max(w, wMax);
                }
            }
        }

		// Bound y (bottom)
		annoWidth = wMax * CUBE_SCALE;
        justify.setRotate(vertical ? Transform.ROTATE_90
                                   : Transform.ROTATE_NONE);
        justify.setAbove(false);                    // Below axis
        justifyMatrix = getJustifyMatrix(annoHeight, annoWidth, justify);

        easeway = 0.0;
		if (wMax > 0.0) {					// At least one label
            easeway = justify.getEaseway();
			newMin = -yScale - easeway;
            extents.checkYExtents(newMin);

            if (vertical) {
                // Bound xmin (left side)
                newMin = -xScale - annoHeight / 2.0;
                extents.checkXExtents(newMin);

                // Bound xmax (right side)
                newMax = xScale + annoHeight / 2.0;
                extents.checkXExtents(newMax);
            }
		}

		// Bound x title
		if (xAxis.getTitle() != null) {
            double w = cubeTextWidth(gc, xAxis, xAxis.getTitle(), true, true);
            newMin = - w * CUBE_SCALE / 2.0;
            extents.checkXExtents(newMin);
            newMax = - newMin;
            extents.checkXExtents(newMax);

            newMin = -yScale - easeway - titleHeight - annoHeight/4.0;
            extents.checkYExtents(newMin);
            xAxis.setAnnoEaseway(easeway +           // Basic space
                                 annoHeight/4.0);    // Padding to title
		}
	}

	/**********/
	/* Y AXIS */
	/**********/

	annoHeight  = yAxis.getAnnoFontCubeSize()  * CUBE_SCALE;
	titleHeight = yAxis.getTitleFontCubeSize() * CUBE_SCALE;
	if (yAxis.isShowing()) {
		// Compute extents of annotations in the y-direction
        rotation = yAxis.getAnnotationRotation();
        vertical = (rotation == JCAxis.ANNOTATION_ROTATION_DEFAULT ||
                    rotation == JCAxis.ANNOTATION_ROTATION_VERTICAL);
        if (vertical) {
            calcAxisBounds(yAxis, annoHeight, 2.0 * yScale, 2.0, false,
                           yAxis.getAnnoMin(), yAxis.getAnnoMax());
        }
        else {
            double longAnno = guessLongestAnnotation(gc, yAxis) * CUBE_SCALE;
            calcAxisBounds(yAxis, longAnno, 2.0 * yScale, 1.5, false,
                           yAxis.getAnnoMin(), yAxis.getAnnoMax());
        }

		// Now that we have the annotation min and max, make the labels
		yAxis.getAnnoHandler().makeAnnotations(gc, this);

		// Axis min/max may have been adjusted, so get latest values
        axisMin = yAxis.getAnnoMin();
        axisMax = yAxis.getAnnoMax();

		// Update the dataport (data to cube transform info calculated in
        // calcAxisBounds())
        dataport.y      = axisMin;
        dataport.height = axisMax - axisMin;

		wMax = 0.0;
        valueLabels = yAxis.getAnnoHandler().getValueLabels();
        if (valueLabels != null && valueLabels.size() > 0) {
            for (int i = 0; i < valueLabels.size(); i++) {
                JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
                double value = vLabel.getValue();
                String label = vLabel.getLabel();
                if (axisMin <= value && value <= axisMax) {
                    double w = cubeTextWidth(gc, yAxis, label, false, true);
                    if (!vertical) {
                        double pos = yScale * (1.0 - 2.0 * (axisMax - value) /
                                               (axisMax - axisMin));
                        // Bound min
                        newMin = pos - w * CUBE_SCALE / 2.0;
                        extents.checkYExtents(newMin);
                        // Bound max
                        newMax = pos + w * CUBE_SCALE / 2.0;
                        extents.checkYExtents(newMax);
                    }
                    wMax = Math.max(w, wMax);
                }
            }
        }

		// Bound x (left)
		annoWidth = wMax * CUBE_SCALE;
        justify.setRotate(vertical ? Transform.ROTATE_90
                                   : Transform.ROTATE_NONE);
        justify.setAbove(true);                   // Above axis
        justifyMatrix = getJustifyMatrix(annoHeight, annoWidth, justify);

        easeway = 0.0;
		if (wMax > 0) {					// At least one label
            easeway = justify.getEaseway();
            newMin = -xScale - easeway;
            extents.checkXExtents(newMin);

            if (vertical) {
                // Bound ymin (bottom side)
                newMin = -yScale - annoHeight/2.0;
                extents.checkYExtents(newMin);

                // Bound ymax (top side)
                newMax = yScale + annoHeight/2.0;
                extents.checkYExtents(newMax);
            }
		}

		// Bound y title
		if (yAxis.getTitle() != null) {
            double w = cubeTextWidth(gc, yAxis, yAxis.getTitle(), true, true);
            newMin = - w * CUBE_SCALE / 2.0;
            extents.checkYExtents(newMin);
            newMax = - newMin;
            extents.checkYExtents(newMax);

            newMin = -xScale - easeway - titleHeight - annoHeight/4.0;
            extents.checkXExtents(newMin);
            yAxis.setAnnoEaseway(easeway +           // Basic space
                                 annoHeight/4.0);    // Padding to title
		}
	}

	// Set dataport bounds
	xAxis.setDataportMin(dataport.x);
	yAxis.setDataportMin(dataport.y);
	zAxis.setDataportMin(dataport.z);
	xAxis.setDataportMax(dataport.x + dataport.width);
	yAxis.setDataportMax(dataport.y + dataport.height);
	zAxis.setDataportMax(dataport.z + dataport.depth);

	// Return the viewport's extents as a rectangle
	viewport.x		= extents.getXMin();
	viewport.y		= extents.getYMin();
	viewport.width 	= extents.getXMax() - extents.getXMin();
	viewport.height = extents.getYMax() - extents.getYMin();
}

/**
 * Draws annotation for 2d projection plots
 *
 * @param gc The graphics object to use for drawing and getting font info
 */
protected void draw2dAnno(Graphics gc)
{
    if (gc == null) {
	    return;
	}
    gc.setColor(area.getPlotCube().getDrawForeground());

    Matrix4d screen = area.getTransform().getScreen();
    draw2dAxis(gc, area.getAxis(JCAxis.AXIS_X), screen);
    draw2dAxis(gc, area.getAxis(JCAxis.AXIS_Y), screen);
}

/**
 * Check if a label has been selected by going thru the drawing mechanism
 *
 * @param p The point to check against the labels
 * @return The selected label (null if none was selected)
 */
protected JCValueLabel pickLabel(Graphics gc, Point p)
{
	pickPoint     = p;
	selectedLabel = null;
	if (p != null) {
	    doPicking = true;
	    if (area.getChart3d().isProjection()) {
		    draw2dAnno(gc);
	    }
	    else {
		    draw3dAnno(gc);
	    }
	    doPicking = false;
	}
	return(selectedLabel);
}

}
