/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BarFacet.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:55 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Graphics;
import java.awt.Color;
import java.awt.Rectangle;
import javax.vecmath.Point4d;



/**
 * Class responsible for drawing bars
 */
public class BarFacet extends FacetAbstract {

// Enumerate all six sides of a bar
public final static int FACE_NONE = -1;
public final static int FACE_A    = 0;
public final static int FACE_B    = 1;
public final static int FACE_C    = 2;
public final static int FACE_D    = 3;
public final static int FACE_H    = 4;     // Higher face
public final static int FACE_L    = 5;     // Lower face

// The corners on the given faces
protected final static int faceCorner[][] = {
	{ 1, 2 },		// FACE_A
	{ 2, 3 },		// FACE_B
	{ 3, 0 },		// FACE_C
	{ 0, 1 }		// FACE_D
};

// Bar drawing styles
public final static int BAR_MESH       = 0; // Draw outline around face
public final static int BAR_FILLED     = 1; // Fill bar face
public final static int BAR_FILL_MESH  = 2; // Draw outline and fill bar face
public final static int BAR_ERASE      = 3; // Fill bar face with background color
public final static int BAR_ERASE_MESH = 4; // First do BAR_ERASE then draw outline
public final static int BAR_SIDES      = 5; // Only vertical sides of a bar
public final static int BAR_CONTOUR    = 6; // Draw contour level on bar

protected transient Graphics gc = null;

/**
 * Default constructor.
 */
public BarFacet()
{
}

/**
 * Constructor which supplies the drawing object and the Graphics object needed
 * to draw
 * @param drawGrid The drawing object that controls which bar is being drawn
 * @param gc The Graphics object needed to draw
 */
public BarFacet(DrawGrid drawGrid, Graphics gc)
{
	super(drawGrid);
	this.gc = gc;
}

/**
 * Projects a plane in data space to screen coordinates.  The plane is defined
 * by four points, which represent the (x,y) position of a bar's four corners
 * in data space.  A zValue is supplied to defined the z position of the plane.
 * The four points are then transformed to screen space and these points are
 * returned.
 *
 * @param plane A plane consisting of four points that defines the x and y
 *    extents of the bar in data space
 * @param zValue The zValue for each of the four points of the plane
 * @return A list of four screen points corresponding to the projected plane
 */
protected ScreenPoint[] planeToScreen(Point4d[] plane, double zValue)
{
	if (plane.length < 4) {
		
		return(null);
	}

	ScreenPoint[] spts = new ScreenPoint[4];
	for (int i = 0; i < 4; i++) {
		plane[i].z = zValue;
		spts[i]    = getScreenPoint(plane[i].x, plane[i].y, plane[i].z);
	}
	return(spts);
}

/**
 * Constructs and draws a face of a bar in the given style.
 *
 * @param top The top plane of the bar in screen space
 * @param bottom The bottom plane of the bar in screen space
 * @param fCode The face code for this bar face.  Must be one of FACE_A, FACE_B,
 *     FACE_C, FACE_D, FACE_H, or FACE_L.  If value is FACE_NONE, then nothing
 *     is drawn.
 * @param style The bar drawing style.  Is one of BAR_MESH, BAR_FILLED,
 *     BAR_FILL_MESH, BAR_ERASE, BAR_ERASE_MESH, BAR_SIDES, BAR_CONTOUR.  See
 *     definitions above for explanation of styles.
 */
protected void drawBarFace(ScreenPoint[] top, ScreenPoint[] bottom,
			               int fCode, int style)
{
    Chart3dGridData     data = drawGrid.getGridData();
	Chart3dDataView     dataView = data.getDataView();
	JCElevation         elevation = dataView.getElevation();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
    JCPlotCube          plotCube = area.getPlotCube();
	ScreenPoint[]       face;

	if (fCode == FACE_NONE) {
		return;
	}

	// Determine the list of points that define the bar face
	int i = 0;
	int j = 0;
	if (fCode == FACE_H) {
		face = top;
		if (JCChart3dUtil.gridBackfacing(face[0], face[1], face[2])) {
			return;
		}
	}
	else if (fCode == FACE_L) {
		face = bottom;
		if (!JCChart3dUtil.gridBackfacing(face[0], face[1], face[2])) {
			return;
		}
	}
	else {
		face = new ScreenPoint[4];
		i = faceCorner[fCode][0];
		j = faceCorner[fCode][1];
		face[0] = bottom[i];
		face[1] = bottom[j];
		face[2] = top[j];
		face[3] = top[i];
	}

	// Draw the bar (or bar segment) in the correct style
	JCFillStyle fillStyle = new JCFillStyle(plotCube.getDrawBackground(),
									        JCFillStyle.SOLID);
	fillPointArrays(face, 4);
	switch(style) {
	case BAR_MESH:
		// Draw the outline of the bar face
		gc.drawPolygon(xPoints, yPoints, 4);
		break;
	case BAR_FILLED:
		// Fill the bar face with a preset shade
		gc.fillPolygon(xPoints, yPoints, 4);
		break;
	case BAR_FILL_MESH:
		// Fill the bar face and draw the outline
		gc.fillPolygon(xPoints, yPoints, 4);
		gc.setColor(area.getForeground());
		gc.drawPolygon(xPoints, yPoints, 4);
	case BAR_ERASE:
		// Like BAR_FILLED except the shade is always the background colour */
		fillStyle.updateGraphics(gc);
		gc.fillPolygon(xPoints, yPoints, 4);
		break;
	case BAR_ERASE_MESH:
		// Like BAR_ERASE with an outline
		fillStyle.updateGraphics(gc);
		gc.fillPolygon(xPoints, yPoints, 4);
		gc.setColor(elevation.getMeshTopColor());
		gc.drawPolygon(xPoints, yPoints, 4);
		break;
	case BAR_SIDES:
		// Only draw the vertical sides of the bar
		if (fCode == FACE_H || fCode == FACE_L) {
			
			return;
		}
		gc.drawLine(bottom[j].getXPixel(), bottom[j].getYPixel(),
		            top[j].getXPixel(), top[j].getYPixel());
		gc.drawLine(bottom[i].getXPixel(), bottom[i].getYPixel(),
		            top[i].getXPixel(), top[i].getYPixel());
		break;
	case BAR_CONTOUR:
		// Only draw a contour level on the bar should use drawBarContour()
		if (fCode == FACE_H || fCode == FACE_L) {
			
			return;
		}
		gc.drawLine(face[i].getXPixel(), face[i].getYPixel(),
		            face[j].getXPixel(), face[j].getYPixel());
		break;
	default:
		break;
	}
}

/**
 * Draw a contour line on the side faces defined by face codes f0 and f1.  The
 * spts array define four corner points of the bar in screen space for the
 * contour level to be drawn.  It is assumed that the gc has been previously
 * updated so it will draw the line in the correct color, width and line style.
 * Note that this method is faster than calling drawBarFace(..., BAR_CONTOUR)
 *
 * @param spts The four corner points of the bar for the given contour level
 *     in screen space.
 * @param f0 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     contour line is drawn.
 * @param f1 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     contour line is drawn.
 */
protected void drawBarContour(ScreenPoint[] spts, int f0, int f1)
{
	if (f0 != FACE_NONE) {
		int i = faceCorner[f0][0];
		int j = faceCorner[f0][1];
		gc.drawLine(spts[i].getXPixel(), spts[i].getYPixel(),
		            spts[j].getXPixel(), spts[j].getYPixel());
		if (f1 != FACE_NONE) {
			i = faceCorner[f1][0];
			j = faceCorner[f1][1];
		    gc.drawLine(spts[i].getXPixel(), spts[i].getYPixel(),
		                spts[j].getXPixel(), spts[j].getYPixel());
		}
	}
}

/**
 * Draw a bar or bar segement in a given style by drawing four bar faces (top,
 * bottom, and the two side faces that can be seen from the current eye
 * position) between the planes defined by top and bottom.  It is assumed
 * that the gc has previously been updated with the correct information to draw
 * the bar in the correct color, line style, fill style, etc.
 *
 * @param top The top plane of this bar (or bar segment) in screen space
 * @param bottom The bottom plane of this bar (or bar segment) in screen space
 * @param f0 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     side face is drawn.
 * @param f1 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     side face is drawn.
 */
protected void drawBarStyle(ScreenPoint[] top, ScreenPoint[] bottom,
			                int f0, int f1, int style)
{
	drawBarFace(top, bottom, FACE_H, style);
	drawBarFace(top, bottom, FACE_L, style);
	drawBarFace(top, bottom, f0, style);
	drawBarFace(top, bottom, f1, style);
}

/**
 * Draw an entire bar by drawing four bar faces (top, bottom, and the two side
 * faces that can be seen from the current eye position) between the planes
 * defined by top and bottom.  If a gridColor for this bar has been defined, use
 * it. Otherwise determine the color of the bar by shaded top or bottom color
 * depending on whether the bar is above or below the z origin.  It is assumed
 * that other than the color the gc has previously been updated with the correct
 * information to draw the bar in the correct line style, fill style, etc.
 *
 * @param top The top plane of this bar (or bar segment) in screen space
 * @param bottom The bottom plane of this bar (or bar segment) in screen space
 * @param f0 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     side face is drawn.
 * @param f1 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     side face is drawn.
 * @param above Is the bar above the z origin?
 * @param gridColor The grid color used for this bar.  Null if none has been
 *     defined for this bar.
 */
protected void drawShadedBar(ScreenPoint[] top, ScreenPoint[] bottom,
			                 int f0, int f1, boolean above, JCGridColor gridColor)
{
    Chart3dGridData     data = drawGrid.getGridData();
	JCElevation         elevation = data.getDataView().getElevation();
	Color               topColor, bottomColor;

	if (gridColor != null && gridColor.getColor() != null) {
		topColor = bottomColor = gridColor.getColor();
	}
	else {
		topColor    = elevation.getShadedTopColor();
		bottomColor = elevation.getShadedBottomColor();
		if (gridColor != null) {
			
		}
	}

	if (above) {
		gc.setColor(topColor);
		drawBarFace(top, bottom, FACE_L, BAR_FILLED);
		drawBarFace(top, bottom, FACE_H, BAR_FILLED);
	}
	else {
		gc.setColor(bottomColor);
		drawBarFace(top, bottom, FACE_H, BAR_FILLED);
		drawBarFace(top, bottom, FACE_L, BAR_FILLED);
	}
	drawBarFace(top, bottom, f0, BAR_FILLED);
	drawBarFace(top, bottom, f1, BAR_FILLED);
}

/**
 * Main bar rendering routine. <code>f0</code> and <code>f1</code> indicate
 * the sides of the bar (other than the top and bottom) which need to be drawn
 * based on the eye position.  The plane gives the <code>x</code>/<code>y</code>
 * extents of the bar. <code>top</code> and <code>bottom</code> are precomputed
 * screen position of the top and bottom faces.
 *
 * @param x The x grid index of the bar being drawn
 * @param y The y grid index of the bar being drawn
 * @param f0 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     side face is drawn.
 * @param f1 The face code of a side face to be drawn for this bar.  Must be
 *     one of FACE_A, FACE_B, FACE_C, FACE_D.  If value is FACE_NONE, then no
 *     side face is drawn.
 * @param plane A plane that defines the x and y extents of the bar in data space
 * @param top The top plane of the bar in screen space
 * @param bottom The bottom plane of the bar in screen space
 */
protected void drawBar(int x, int y, int f0, int f1, Point4d[] plane,
                       ScreenPoint[] top, ScreenPoint[] bottom)
{
    int                 i;
    double              minValue, maxValue;
    boolean             above;
    Chart3dGridData     data = drawGrid.getGridData();
	Chart3dDataView     dataView = data.getDataView();
    Chart3dGridData     zoneData = dataView.getZoneData();
	JCElevation         elevation = dataView.getElevation();
    JCContour           contour = dataView.getContour();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
	JCPlotCube          plotCube = area.getPlotCube();
    ScreenPoint[]       high, low, hi, lo;

    double value   = data.getZValue(x, y);
    double zOrigin = area.getAxis(JCAxis.AXIS_Z).getOrigin();
    if (value < zOrigin) {
        high = hi = bottom;
        low = lo  = top;
        minValue  = value;
        maxValue  = zOrigin;
        above     = false;
    }
    else {
        high = hi = top;
        low = lo  = bottom;
        minValue  = zOrigin;
        maxValue  = value;
        above     = true;
    }
    Color meshColor       = above ? elevation.getMeshTopColor()
								  : elevation.getMeshBottomColor();
	JCLineStyle lineStyle = new JCLineStyle(1, plotCube.getDrawForeground(),
									        JCLineStyle.SOLID);
	JCFillStyle fillStyle = new JCFillStyle(Color.white, JCFillStyle.SOLID);

    boolean zoneBar = (zoneData.isDataOK() && contour.isZoned());
    int drawCode    = dataView.getDrawCode();

    // For zoned data we need to reset the draw code
    double zValue = 0.0;
    if (zoneBar) {
        zValue = zoneData.getZValue(x, y);
        switch (dataView.getDrawCode()) {
        case Chart3dDataView.ZONED:
        case Chart3dDataView.ZONED + Chart3dDataView.SHADED:
        case Chart3dDataView.ZONED + Chart3dDataView.CONTOURED:  /* unusual case */
        case Chart3dDataView.ZONED + Chart3dDataView.CONTOURED + Chart3dDataView.SHADED:
            drawCode = Chart3dDataView.SHADED;
            break;
        case Chart3dDataView.ZONED + Chart3dDataView.MESHED:
        case Chart3dDataView.ZONED + Chart3dDataView.CONTOURED + Chart3dDataView.MESHED:
            drawCode = Chart3dDataView.MESHED;
            break;
        case Chart3dDataView.ZONED + Chart3dDataView.SHADED +
             Chart3dDataView.MESHED:
        case Chart3dDataView.ZONED + Chart3dDataView.CONTOURED +
             Chart3dDataView.SHADED + Chart3dDataView.MESHED:

            drawCode = Chart3dDataView.SHADED + Chart3dDataView.MESHED;
            break;
        default:
			
            break;
        }
    }

    // Use bar colors only if DrawShaded is true and DrawZones is false.
    // "gridColor != null" indicates that bar colors will be used.
    JCGridColor gridColor = null;
    if (elevation.isShaded() && !zoneBar) {
        gridColor = dataView.findGridColor(x, y);
    }

    boolean isTransparent         = elevation.isTransparent();
	JCContourLevels contourLevels = contour.getContourLevels();
    double[] levels               = contourLevels.getLevels();
    int nLevels                   = contourLevels.getNumLevelsInternal();

	lineStyle.resetGraphics(gc);
	fillStyle.resetGraphics(gc);
    switch(drawCode) {
    case Chart3dDataView.MESHED:
        if (!isTransparent) {
            drawBarStyle(hi, lo, f0, f1, BAR_ERASE);
        }
        if (zoneBar) {
			contour.setZLineColor(gc, contourLevels.getLevelFromValue(zValue));
        }
        else {
            gc.setColor(meshColor);
        }
        if (isTransparent) {
            fillPointArrays(hi, 4);
            gc.drawPolygon(xPoints, yPoints, 4);
            fillPointArrays(lo, 4);
            gc.drawPolygon(xPoints, yPoints, 4);
            gc.drawLine(lo[0].getXPixel(), lo[0].getYPixel(),
                        hi[0].getXPixel(), hi[0].getYPixel());
            gc.drawLine(lo[1].getXPixel(), lo[1].getYPixel(),
                        hi[1].getXPixel(), hi[1].getYPixel());
            gc.drawLine(lo[2].getXPixel(), lo[2].getYPixel(),
                        hi[2].getXPixel(), hi[2].getYPixel());
            gc.drawLine(lo[3].getXPixel(), lo[3].getYPixel(),
                        hi[3].getXPixel(), hi[3].getYPixel());
        }
        else {
            drawBarStyle(hi, lo, f0, f1, BAR_MESH);
        }
        break;
    case Chart3dDataView.SHADED:
        if (zoneBar) {
			contour.setFillStyle(gc, contourLevels.getLevelFromValue(zValue));
            drawBarStyle(hi, lo, f0, f1, BAR_FILLED);
        }
        else {
            drawShadedBar(hi, lo, f0, f1, above, gridColor);
        }
        break;
    case Chart3dDataView.CONTOURED:
		lineStyle.updateGraphics(gc);
        gc.setColor(meshColor);
        if (!isTransparent) {
            drawBarStyle(hi, lo, f0, f1, BAR_ERASE);
        }
        for (i = 0; i < nLevels; i++) {
            double lev = levels[i];
            if (minValue < lev && lev < maxValue) {
                hi = planeToScreen(plane, lev);
				contour.setLineStyle(gc, i);
                if (isTransparent) {
                    fillPointArrays(hi, 4);
                    gc.drawPolygon(xPoints, yPoints, 4);
                }
                else {
                    drawBarContour(hi, f0, f1);
                }
            }
        }
		lineStyle.resetGraphics(gc);
        break;
    case Chart3dDataView.SHADED + Chart3dDataView.MESHED:
        if (zoneBar) {
			contour.setFillStyle(gc, contourLevels.getLevelFromValue(zValue));
            drawBarStyle(hi, lo, f0, f1, BAR_FILLED);
        }
        else {
            drawShadedBar(hi, lo, f0, f1, above, gridColor);
        }
        gc.setColor(meshColor);
        drawBarStyle(hi, lo, f0, f1, BAR_MESH);
        break;
    case Chart3dDataView.CONTOURED + Chart3dDataView.MESHED:
        if (!isTransparent) {
            drawBarStyle(hi, lo, f0, f1, BAR_ERASE_MESH);
        }
        if (zoneBar) {
			contour.setZLineColor(gc, contourLevels.getLevelFromValue(zValue));
        }
        else {
            gc.setColor(meshColor);
        }
        drawBarStyle(hi, lo, f0, f1, BAR_MESH);

		lineStyle.updateGraphics(gc);
        for (i = 0; i < nLevels; i++) {
            double lev = levels[i];
            if (minValue < lev && lev < maxValue) {
                hi = planeToScreen(plane, lev);
				contour.setLineStyle(gc, i);
                if (isTransparent) {
                    fillPointArrays(hi, 4);
                    gc.drawPolygon(xPoints, yPoints, 4);
                }
                else {
                    drawBarContour(hi, f0, f1);
                }
            }
        }
		lineStyle.resetGraphics(gc);
        break;
    case Chart3dDataView.ZONED + Chart3dDataView.MESHED:
        if (isTransparent) {
            f0 = FACE_A;
            f1 = FACE_C;
        }
        else {
            drawBarStyle(hi, lo, f0, f1, BAR_ERASE);
        }

        for (i = 0; i < nLevels && minValue > levels[i]; i++) {
            /* just look for first section */
        }

		contour.setZLineColor(gc, i);
        if (isTransparent) {
            fillPointArrays(lo, 4);
            gc.drawPolygon(xPoints, yPoints, 4);
        }
        else {
            drawBarContour(low, f0, f1);
            drawBarFace(hi, lo, FACE_L, BAR_MESH);
        }

        for (; i < nLevels && levels[i] < maxValue; i++) {
            hi = planeToScreen(plane, levels[i]);

		    contour.setZLineColor(gc, i);
            drawBarFace(hi, lo, f0, BAR_SIDES);
            drawBarFace(hi, lo, f1, BAR_SIDES);
            lo = hi;
        }

		contour.setZLineColor(gc, i);
        drawBarFace(high, low, f0, BAR_SIDES);
        drawBarFace(high, low, f1, BAR_SIDES);
        if (isTransparent) {
            fillPointArrays(high, 4);
            gc.drawPolygon(xPoints, yPoints, 4);
        }
        else {
            drawBarContour(high, f0, f1);
            drawBarFace(high, low, FACE_H, BAR_MESH);
        }
        break;
    case Chart3dDataView.CONTOURED + Chart3dDataView.SHADED:
        if (zoneBar) {
			contour.setFillStyle(gc, contourLevels.getLevelFromValue(zValue));
            drawBarStyle(hi, lo, f0, f1, BAR_FILLED);
        }
        else {
            drawShadedBar(hi, lo, f0, f1, above, gridColor);
        }

        gc.setColor(elevation.getMeshTopColor());
        for (i = 0; i < nLevels; i++) {
            double lev = levels[i];
            if (minValue < lev && lev < maxValue) {
                hi = planeToScreen(plane, lev);
				contour.setLineStyle(gc, i);
                drawBarContour(hi, f0, f1);
            }
        }
		lineStyle.resetGraphics(gc);
        break;
    case Chart3dDataView.ZONED:
    case Chart3dDataView.ZONED + Chart3dDataView.SHADED:
        for (i = 0; i < nLevels && minValue > levels[i]; i++) {
            /* just look for first section */
        }

		contour.setFillStyle(gc, i);
        drawBarFace(hi, lo, FACE_L, BAR_FILLED);

        for (; i < nLevels && levels[i] < maxValue; i++) {
            hi = planeToScreen(plane, levels[i]);

		    contour.setFillStyle(gc, i);
            drawBarFace(hi, lo, f0, BAR_FILLED);
            drawBarFace(hi, lo, f1, BAR_FILLED);
            lo = hi;
        }

		contour.setFillStyle(gc, i);
        drawBarFace(high, lo, f0, BAR_FILLED);
        drawBarFace(high, lo, f1, BAR_FILLED);
        drawBarFace(high, lo, FACE_H, BAR_FILLED);
        break;
    case Chart3dDataView.ZONED + Chart3dDataView.SHADED + Chart3dDataView.MESHED:
        for (i = 0; i < nLevels && minValue > levels[i]; i++) {
            /* just look for first section */
        }

		contour.setFillStyle(gc, i);
        drawBarFace(hi, lo, f0, BAR_FILLED);
        drawBarFace(hi, lo, f1, BAR_FILLED);
        drawBarFace(hi, lo, FACE_L, BAR_FILLED);

        for (; i < nLevels && levels[i] < maxValue; i++) {
            hi = planeToScreen(plane, levels[i]);

		    contour.setFillStyle(gc, i);
            drawBarFace(hi, lo, f0, BAR_FILLED);
            drawBarFace(hi, lo, f1, BAR_FILLED);
            lo = hi;
        }

		contour.setFillStyle(gc, i);
        drawBarFace(high, lo, f0, BAR_FILLED);
        drawBarFace(high, lo, f1, BAR_FILLED);
        drawBarFace(high, lo, FACE_H, BAR_FILLED);

        gc.setColor(meshColor);
        drawBarStyle(high, low, f0, f1, BAR_MESH);
        break;
    case Chart3dDataView.CONTOURED + Chart3dDataView.SHADED + Chart3dDataView.MESHED:
        if (zoneBar) {
			contour.setFillStyle(gc, contourLevels.getLevelFromValue(zValue));
            drawBarStyle(hi, lo, f0, f1, BAR_FILLED);
        }
        else {
            drawShadedBar(hi, lo, f0, f1, above, gridColor);
        }

        for (i = 0; i < nLevels; i++) {
            double lev = levels[i];
            if (minValue < lev && lev < maxValue) {
                hi = planeToScreen(plane, lev);
				contour.setLineStyle(gc, i);
                drawBarContour(hi, f0, f1);
            }
        }
		lineStyle.resetGraphics(gc);

        gc.setColor(meshColor);
        drawBarStyle(high, low, f0, f1, BAR_MESH);
        break;
    case Chart3dDataView.CONTOURED + Chart3dDataView.ZONED + Chart3dDataView.MESHED:
        if (isTransparent) {
            f0 = FACE_A;
            f1 = FACE_C;
        }
        else {
            drawBarStyle(hi, lo, f0, f1, BAR_ERASE);
        }

        for (i = 0; i < nLevels && minValue > levels[i]; i++) {
            /* just look for first section */
        }

		contour.setZLineColor(gc, i);
        if (isTransparent) {
            fillPointArrays(lo, 4);
            gc.drawPolygon(xPoints, yPoints, 4);
        }
        else {
            drawBarContour(low, f0, f1);
            drawBarFace(hi, lo, FACE_L, BAR_MESH);
        }

        for (; i < nLevels && levels[i] < maxValue; i++) {
            hi = planeToScreen(plane, levels[i]);

		    contour.setZLineColor(gc, i);
            drawBarFace(hi, lo, f0, BAR_SIDES);
            drawBarFace(hi, lo, f1, BAR_SIDES);
			contour.setLineStyle(gc, i);
			if (isTransparent) {
                fillPointArrays(hi, 4);
                gc.drawPolygon(xPoints, yPoints, 4);
			}
			else {
                drawBarContour(hi, f0, f1);
			}
            lo = hi;
        }
		lineStyle.resetGraphics(gc);

		contour.setZLineColor(gc, i);
        drawBarFace(high, lo, f0, BAR_SIDES);
        drawBarFace(high, lo, f1, BAR_SIDES);
        if (isTransparent) {
            fillPointArrays(high, 4);
            gc.drawPolygon(xPoints, yPoints, 4);
        }
        else {
            drawBarContour(high, f0, f1);
            drawBarFace(high, low, FACE_H, BAR_MESH);
        }
        break;

    case Chart3dDataView.CONTOURED + Chart3dDataView.ZONED:
    case Chart3dDataView.CONTOURED + Chart3dDataView.ZONED + Chart3dDataView.SHADED:
        for (i = 0; i < nLevels && minValue > levels[i]; i++) {
            /* just look for first section */
        }

		contour.setZLineColor(gc, i);
        drawBarFace(hi, lo, FACE_L, BAR_FILLED);

        boolean first = true;
        for (; i < nLevels && levels[i] < maxValue; i++) {
            hi = planeToScreen(plane, levels[i]);

		    contour.setFillStyle(gc, i);
            drawBarFace(hi, lo, f0, BAR_FILLED);
            drawBarFace(hi, lo, f1, BAR_FILLED);
            if (i > 0 && !first) {
			    contour.setLineStyle(gc, i - 1);
                drawBarContour(lo, f0, f1);
            }
            first = false;
            lo = hi;
        }

		contour.setFillStyle(gc, i);
        drawBarFace(high, lo, f0, BAR_FILLED);
        drawBarFace(high, lo, f1, BAR_FILLED);
        drawBarFace(high, lo, FACE_H, BAR_FILLED);
        if (i > 0) {
			contour.setLineStyle(gc, i - 1);
            drawBarContour(lo, f0, f1);
        }
		lineStyle.resetGraphics(gc);
        break;
    case Chart3dDataView.CONTOURED + Chart3dDataView.ZONED +
	     Chart3dDataView.SHADED + Chart3dDataView.MESHED:
        for (i = 0; i < nLevels && minValue > levels[i]; i++) {
            /* just look for first section */
        }

		contour.setFillStyle(gc, i);
        drawBarFace(hi, lo, FACE_L, BAR_FILLED);

        first = true;
        for (; i < nLevels && levels[i] < maxValue; i++) {
            hi = planeToScreen(plane, levels[i]);

		    contour.setFillStyle(gc, i);
            drawBarFace(hi, lo, f0, BAR_FILLED);
            drawBarFace(hi, lo, f1, BAR_FILLED);

			contour.setLineStyle(gc, i);
            drawBarContour(hi, f0, f1);

            if (i > 0 && !first) {
			    contour.setLineStyle(gc, i - 1);
                drawBarContour(lo, f0, f1);
            }
            first = false;
            lo = hi;
        }

		contour.setFillStyle(gc, i);
        drawBarFace(high, lo, f0, BAR_FILLED);
        drawBarFace(high, lo, f1, BAR_FILLED);
        drawBarFace(high, lo, FACE_H, BAR_FILLED);

        if (i > 0) {
		    contour.setLineStyle(gc, i - 1);
        }
        drawBarContour(lo, f0, f1);
		lineStyle.resetGraphics(gc);

        gc.setColor(meshColor);
        drawBarStyle(high, low, f0, f1, BAR_MESH);
        break;
    default:
        break;
    }
}

/**
 * Process the facet at grid point (x, y) which for this class means drawing a
 * bar at (x,y).  Based on the grid region (@see GridRegion), determine which
 * faces of the bar are visible and then call drawBar to actually draw the bar.
 *
 * @param x The x grid index of the bar being drawn
 * @param y The y grid index of the bar being drawn
 * @param edges The grid region that the bar is in.
 */
public void processFacet(int x, int y, int edges)
{
    Chart3dGridData     data = drawGrid.getGridData();
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	GridLimits          gridLimits = data.getGridLimits();
	Chart3dDataView     dataView = drawGrid.getDataView();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();

	if (x < gridLimits.getXOffset() ||
	    x > gridLimits.getXLast()   ||
		y < gridLimits.getYOffset() ||
		y > gridLimits.getYLast())
	{
		// Should never get here
	    
		return;
	}

    if (screenPoints == null) {
        return;
    }

    ScreenPoint sp = screenPoints[x][y];
	if (sp.isDrawn() || sp.isHoleValue()) {
		return;
	}

    // Construct the extents of the bar in data space
    Point4d[] plane = JCChart3dUtil.constructBarPlane(data, x, y);

    // Construct the bar's top plane in screen space
    ScreenPoint[] top = planeToScreen(plane, data.getZValue(x, y));

    // Construct the bar's bottom plane in screen space
    ScreenPoint[] bottom = planeToScreen(plane, area.getAxis(JCAxis.AXIS_Z).getOrigin());

    //  Don't process any facets outside the chart area due to zooming
    if (!area.getViewport().isNormalized()) {
        
        
		int x1 = JCChart3dUtil.min4(top[0].getXPixel(), top[1].getXPixel(),
                                    top[2].getXPixel(), top[3].getXPixel());
        int y1 = JCChart3dUtil.min4(top[0].getYPixel(), top[1].getYPixel(),
                                    top[2].getYPixel(), top[3].getYPixel());
        int x2 = JCChart3dUtil.max4(top[0].getXPixel(), top[1].getXPixel(),
                                    top[2].getXPixel(), top[3].getXPixel());
        int y2 = JCChart3dUtil.max4(top[0].getYPixel(), top[1].getYPixel(),
                                    top[2].getYPixel(), top[3].getYPixel());
        Rectangle rect = new Rectangle(x1, y1, x2 - x1, y2 - y1);
        

        x1 = JCChart3dUtil.min4(bottom[0].getXPixel(), bottom[1].getXPixel(),
                                bottom[2].getXPixel(), bottom[3].getXPixel());
        y1 = JCChart3dUtil.min4(bottom[0].getYPixel(), bottom[1].getYPixel(),
                                bottom[2].getYPixel(), bottom[3].getYPixel());
        x2 = JCChart3dUtil.max4(bottom[0].getXPixel(), bottom[1].getXPixel(),
                                bottom[2].getXPixel(), bottom[3].getXPixel());
        y2 = JCChart3dUtil.max4(bottom[0].getYPixel(), bottom[1].getYPixel(),
                                bottom[2].getYPixel(), bottom[3].getYPixel());
        Rectangle zrect = new Rectangle(x1, y1, x2 - x1, y2 - y1);
        

        Rectangle brect = JCChart3dUtil.computeBoundingRectangle(rect, zrect);
        if (!brect.intersects(area.getDrawingArea())) {
            
            return;
        }
    }

    
    switch(edges) {
    case GridRegion.REGION_1:
        /*   _   */
        /*  |_|  */         /* no self-occluding possibilities */
        /*       */
        drawBar(x, y, FACE_NONE, FACE_NONE, plane, top, bottom);
        break;

    case GridRegion.REGION_2:
        /*   _   */
        /*  | |  */         /* show side D */
        /*       */
        drawBar(x, y, FACE_D, FACE_NONE, plane, top, bottom);
        break;

    case GridRegion.REGION_3:
        /*       */
        /*  |_|  */         /* show side B */
        /*       */
        drawBar(x, y, FACE_B, FACE_NONE, plane, top, bottom);
        break;

    case GridRegion.REGION_4:
        /*   _   */
        /*   _|  */         /* show side A */
        /*       */
        drawBar(x, y, FACE_A, FACE_NONE, plane, top, bottom);
        break;

    case GridRegion.REGION_5:
        /*   _   */
        /*  |_   */         /* show side C */
        /*       */
        drawBar(x, y, FACE_C, FACE_NONE, plane, top, bottom);
        break;

    case GridRegion.REGION_6:
        /*   _   */
        /*    |  */         /* show side A and D */
        /*       */
        drawBar(x, y, FACE_A, FACE_D, plane, top, bottom);
        break;

    case GridRegion.REGION_7:
        /*       */
        /*   _|  */         /* show side A and B */
        /*       */
        drawBar(x, y, FACE_A, FACE_B, plane, top, bottom);
        break;

    case GridRegion.REGION_8:
        /*   _   */
        /*  |    */         /* show side C and D */
        /*       */
        drawBar(x, y, FACE_C, FACE_D, plane, top, bottom);
        break;

    case GridRegion.REGION_9:
        /*       */
        /*  |_   */         /* show side B and C */
        /*       */
        drawBar(x, y, FACE_B, FACE_C, plane, top, bottom);
        break;

    default:
		
        break;
    }

    sp.setDrawn(true);
}

}


