/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Chart3dAnnoHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;



import java.awt.*;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * Class that handles annotations for a given JCAxis.
 */
public class Chart3dAnnoHandler implements Comparator, java.io.Serializable {

/** ArrayList of <code>JCValueLabel</code> instances. */
protected ArrayList valueLabels = new ArrayList();

/** User value labels list. */
protected ArrayList userValueLabels = new ArrayList();

/** The maximum annotation height for the current valueLabel list. */
protected double annoHeight = 0.0;

/** The annotation shift for the current valueLabel list */
protected double annoShift  = 0.0;

/** The parent JCAxis object */
protected JCAxis axis = null;

/**
 * Constructor that sets the parent JCAxis object.
 *
 * @param axis The parent JCAxis object
 */
public Chart3dAnnoHandler(JCAxis axis)
{
	this.axis = axis;
}

/**
 * Gets the parent JCAxis object.
 *
 * @return The JCAxis parent
 */
protected JCAxis getAxis()
{
	return(axis);
}

/**
 * Gets the internally generated value labels.  If the
 * <code>annotationMethod</code> of the parent axis is
 * <code>ANNOTATION_VALUE_LABELS</code>, this list will hold the same value
 * labels as the user value label list.
 *
 * @return The internally generated value labels.
 */
public ArrayList getValueLabels()
{
	return(valueLabels);
}

/**
 * Returns the user value label list.  This is an array list that holds the
 * user list of value labels, either passed in all at once, or added one at
 * a time.
 *
 * @return The user value label list.
 */
protected ArrayList getUserValueLabels()
{
	return(userValueLabels);
}

/**
 * Sets the user value label list.  Sorts them so they are in increasing order.
 *
 * @param userValueLabels The new userValueLabel list.
 */
protected void setUserValueLabels(ArrayList userValueLabels)
{
	this.userValueLabels = userValueLabels;
	sortValueLabels();
}

/**
 * Gets the maximum annotation height for the current <code>valueLabel</code>
 * list.
 *
 * @return The maximum annotation height.
 */
public double getAnnoHeight()
{
	return(annoHeight);
}

/**
 * Gets the annotation shift for the current <code>valueLabel</code> list.
 *
 * @return The annotation shift.
 */
public double getAnnoShift()
{
	return(annoShift);
}

/**
 * Sorts user-provided value labels so that they are in increasing order.
 *
 */
protected void sortValueLabels()
{
	if (userValueLabels.size() <= 1) {
		return;
	}
	Collections.sort(userValueLabels, this);
}

/**
 * <code>JCValueLabel</code> sorting routine.
 *
 * @param o1 The first value label
 * @param o2 The second value label
 * @return Return whether first label is less than (negative), greater than
 *     (positive), or equal to (zero) the second label.
 */
public int compare(Object o1, Object o2)
{
    if (o1 == null && o2 == null) {
        return (0);
    }
    else if (o1 == null) {
        return(-1);
    }
    else if (o2 == null) {
        return(1);
    }

	JCValueLabel valueLabelOne = (JCValueLabel)o1;
	JCValueLabel valueLabelTwo = (JCValueLabel)o2;

    int value = 0;
    if (valueLabelOne.getValue() < valueLabelTwo.getValue()) {
        value = -1;
    }
    else if (valueLabelOne.getValue() > valueLabelTwo.getValue()) {
        value = 1;
    }
    return(value);
}

/**
 * Calculates the maximum annotation height for this axis
 *
 * @param gc Graphics context needed for font information
 */
protected void calcAnnoHeight(Graphics gc)
{
	annoHeight = 0.0;
	annoShift  = 0.0;
	Font font = axis.getAnnoFont();
	if (gc == null              ||
	    valueLabels == null     ||
		valueLabels.size() == 0 ||
		font == null)
	{
		return;
	}

    Graphics2D g2 = (Graphics2D)gc;
	g2.setFont(font);
    FontRenderContext frc = g2.getFontRenderContext();

	double annoStart = Double.MAX_VALUE;
	double maxExtent = Double.MIN_VALUE;
	for (int i = 0; i < valueLabels.size(); i++) {
		JCValueLabel vl = (JCValueLabel)valueLabels.get(i);
		if (vl.getLabel() != null && vl.isLabelDrawn()) {
            TextLayout layout = new TextLayout(vl.getLabel(), font, frc);
			Rectangle2D bounds = layout.getBounds();
			annoStart = Math.min(annoStart, bounds.getY());
			maxExtent = Math.max(maxExtent, bounds.getY() + bounds.getHeight());
		}
	}
	annoHeight = maxExtent - annoStart;
	annoShift  = annoHeight / 2.0 - maxExtent;
	
}

/**
 * Creates the axis annotations for display. The axis annotations are stored
 * internally as an <code>ArrayList</code> of <code>JCValueLabel</code> objects.
 * Each <code>JCValueLabel</code> represents a label and/or tick at a particular
 * value along the axis.
 *
 * @param gc Graphics context needed to calculate annotation font information
 * @param anno The Anno object for which the labels are to be generated.
 */
protected void makeAnnotations(Graphics gc, Anno anno)
{
	switch(axis.getAnnotationMethod()) {
	case JCAxis.ANNOTATION_VALUE_LABELS:
		makeValueLabels();
	    break;
	case JCAxis.ANNOTATION_DATA_LABELS:
	    makeDataLabels(anno);
	    break;
	case JCAxis.ANNOTATION_VALUES:
	    makeValues(anno);
	    break;
	default:
	    
	    break;
	}

	calcAnnoHeight(gc);
}

/**
 * Makes the chart 3D value labels using the user value label list.
 */
private void makeValueLabels()
{
	valueLabels = new ArrayList();
	if (userValueLabels == null) {
		return;
	}
	for (int i = 0; i < userValueLabels.size(); i++) {
		JCValueLabel vLabel = (JCValueLabel)userValueLabels.get(i);
	    vLabel.setAxis(axis);
		vLabel.setIndex(i);
		valueLabels.add(vLabel);
	}
}

/**
 * Makes the chart 3D value labels using data labels when the annotation method
 * is <code>ANNOTATION_DATA_LABELS</code>.  These data labels have been
 * specified by the user via <code>XLabels</code> or <code>YLabels</code>.  A
 * value label is created for each data label that falls within the grid limits.
 *
 * @param anno The annotation object for which to generate data labels.
 */
private void makeDataLabels(Anno anno)
{
	valueLabels = new ArrayList();
	if (anno == null) {
		return;
	}

	JCChart3dArea   area     = anno.getChart3dArea();
	Chart3dDataView dataView = area.getChart3d().getDataView(0);
	Chart3dData     data     = dataView.getElevationData();

	if (!(data instanceof Chart3dGridData)) {
		return;
	}

	// Setup for X or Y data labels
	Chart3dGridData gridData  = (Chart3dGridData)data;
	GridLimits      limits    = gridData.getGridLimits();
	String[]        labels    = null;
    ArrayList       list      = null;
	double[]        grid      = null;
	int             numGrid   = 0;
	int             barFormat = JCBar.FIXED;
	int             offset    = 0;
	int             last      = 0;
	switch(axis.getAxisId()) {
	case JCAxis.AXIS_X:
	    labels    = gridData.getXLabels();
        list      = gridData.getXLabelsArrayList();
		grid      = gridData.getXGrid();
		numGrid   = gridData.getNumX();
		barFormat = area.getBar().getXFormat();
		offset    = limits.getXOffset();
		last      = limits.getXLast();
	    break;
	case JCAxis.AXIS_Y:
	    labels    = gridData.getYLabels();
        list      = gridData.getYLabelsArrayList();
		grid      = gridData.getYGrid();
		numGrid   = gridData.getNumY();
		barFormat = area.getBar().getYFormat();
		offset    = limits.getYOffset();
		last      = limits.getYLast();
	    break;
	case JCAxis.AXIS_Z:     // Data labels not defined on Z axis
	default:
	    return;
	}

	// Finally, create the value labels
	if (labels == null && list == null) {
		return;
	}
	int n = (list == null) ? labels.length : list.size();
    JCValueLabel vLabel = null;
	for (int i = offset; i <= last && i < n; i++) {
	    double value = grid[i];
		if (dataView.getChartType() == JCChart3d.BAR &&
		    barFormat == JCBar.HISTOGRAM)
		{
		    if (i >= numGrid - 1) {
			    continue;
			}
			value = (value + grid[i+1]) / 2.0;
		}

		vLabel = null;
        if (list == null) {
            vLabel = new JCValueLabel(value, labels[i]);
        }
		else {
            Object o = list.get(i);
            if (o instanceof JCValueLabel) {
				// Must make a copy (since list could have multiple copies
				// of the same value label and we will need a different value
				// for each one).
				vLabel       = (JCValueLabel)o;
				vLabel       = (JCValueLabel)(vLabel.clone());
				vLabel.value = value;
            }
            else {
                vLabel = new JCValueLabel(value, o.toString());
            }
        }
		if (vLabel == null) {
			continue;
		}
	    vLabel.setAxis(axis);
		vLabel.setIndex(i);
		valueLabels.add(vLabel);
	}

    if (axis.isCombineLabels()) {
        String    lastLabel = null;
        int       count     = 0;
        int       start     = -1;
        boolean   combo     = false;
        ArrayList newList   = new ArrayList();
        for (int i = 0; i < valueLabels.size(); i++) {
            vLabel = (JCValueLabel)valueLabels.get(i);
            if (lastLabel != null && vLabel.getLabel() == lastLabel) {
                if (!combo) {
                    start = i - 1;
                    count = 1;
                    newList.remove(newList.size() - 1);
                    combo = true;
                }
                count++;
				if (i < valueLabels.size() - 1) {
                    continue;
				}
            }
            if (combo) {
                boolean      even     = ((count & 1) == 0);
                JCValueLabel curLabel = null;
                double       sum      = 0.0;
				int          j        = 0;
                for (j = start; j < start + count; j++) {
                    curLabel = (JCValueLabel)valueLabels.get(j);
                    sum += curLabel.getValue();

                    // If odd middle label is drawn
					curLabel.allowUpdates = false;
					if (!even && j == start + count/2) {
						curLabel.setLabelDrawn(true);
						curLabel.setIndex(start);
					}
					else {
					    curLabel.setLabelDrawn(false);
					}
					curLabel.allowUpdates = true;

                    newList.add(curLabel);
                }
                if (even) {
                    // Even number of identical labels -- add one in the middle
					JCValueLabel newLabel = (JCValueLabel)(curLabel.clone());
					newLabel.value      = sum / count;
					newLabel.labelDrawn = true;
					newLabel.tickDrawn  = false;
					newLabel.minorTick  = false;
					newLabel.index      = start;
                    newList.add(newLabel);
                }
                combo = false;
				if (j == valueLabels.size()) {
					// We've already added the last label
					break;
				}
            }
            newList.add(vLabel);
            lastLabel = vLabel.getLabel();
        }
		Collections.sort(newList, this);
        valueLabels = newList;
    }
}

/**
 * Makes a formatted value label given a value.  If a label generator has been
 * specified and useLabelGenerator is true, allow it to modify the value label
 * before returning.
 *
 * @param value The value for which the label is to be generated
 * @param labelDrawn Is the label drawn for the returned value label?
 * @param tickDrawn Is the tick drawn for the returned value label?
 * @param minorTick Minor or Major tick for the returned value label?
 * @param useLabelGenerator Call the user supplied label generator for this
 *     value label.
 */
protected JCValueLabel getLabelFromValue(double value, boolean labelDrawn,
                                         boolean tickDrawn, boolean minorTick,
										 boolean useLabelGenerator)
{
    String s = JCChart3dUtil.format(value, axis.getNFrac());

	// Localize our number if we can
	JCChart3dNumberFormat formatter = axis.getNumberFormatter();
    String label = formatter.localize(s);

	JCValueLabel vLabel = new JCValueLabel(value, label, labelDrawn,
						                   tickDrawn, minorTick);
	vLabel.setAxis(axis);
	vLabel.allowUpdates = false;
	JCLabelGenerator labelGenerator = axis.getLabelGenerator();
	if (useLabelGenerator && labelGenerator != null) {
	    labelGenerator.makeLabel(vLabel);
        vLabel.setValue(value);  // Reset the value in case it was changed
	}
	vLabel.allowUpdates = true;
	return(vLabel);
}

/**
 * Makes the chart 3D value labels when the <code>annotationMethod</code> is
 * <code>ANNOTATION_VALUES</code>.  The annotation <i>min</i>, <i>max</i>, and
 * <i>step</i> (previously calculated) are used to generate labels. A label and
 * major tick will be generated at each step. A half tick is generated between
 * each label.  A value label is generated for a half tick but its
 * <code>labelDrawn</code> property is set to false and its
 * <code>minorTick</code> property set to true so a minor tick is drawn with
 * no label.
 *
 * @param anno The annotation object for which to generate ticks
 */
private void makeValues(Anno anno)
{
	valueLabels = new ArrayList();
	if (anno == null) {
		return;
	}

//    switch (axis.getAxisId()) {
//    case JCAxis.AXIS_X:
//	    
//        break;
//    case JCAxis.AXIS_Y:
//	    
//        break;
//    case JCAxis.AXIS_Z:
//	    
//       break;
//    default:
//	    
//        return;
//    }

	double min   = axis.getAnnoMin();
	double max   = axis.getAnnoMax();
	double step  = axis.getAnnoStep();
	double step2 = step / 2.0;

	JCValueLabel vLabel   = null;
	double       value    = step2 * JCChart3dUtil.ceil(min / step2);
	double       newValue = step * JCChart3dUtil.ceil(min / step);
	boolean      tickOnly = (newValue - value) > (step2 / 2.0);
	int          n        = (int) (JCChart3dUtil.floor(max/step2) -
							       value/step2 + 0.5);
	for (int i = 0; i <= n; i++) {
		double nonNegZeroValue = (value == -0.0) ? 0.0 : value;
		if (tickOnly) {
		    vLabel = getLabelFromValue(nonNegZeroValue, false, true, true, true);
		}
		else {
		    vLabel = getLabelFromValue(nonNegZeroValue, true, true, false, true);
		}
		vLabel.setIndex(i);
		
		
		
		
		valueLabels.add(vLabel);
		value += step2;
		tickOnly = !tickOnly;
	}
}

}
