/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Chart3dData.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;


import com.klg.jclass.chart3d.resources.LocaleBundle;

/**
 * Base object for <code>JCChart3d</code>'s internal data.  It stores data
 * passed into it from a data source which implements one the Chart3d data
 * models.
 */
public abstract class Chart3dData extends TrackChange {

/** The parent dataView that draws with this data object */
protected Chart3dDataView dataView    = null;

/** The "special" data value which determines where holes are drawn */
protected double holeValue = HoleValueChart3dDataModel.DEFAULT_HOLE_VALUE;

/** The (optional) name of this data source */
protected String name = null;

/** Is the data passed in thru the data source in a state in which it can be drawn? */
protected boolean dataOK = false;

/** The data source for this internal data object */
protected Chart3dDataModel dataSource = null;

/** A data handler used to handle data events and extract data from data sources */
protected Chart3dDataHandler dataHandler = null;

/** The current data's maximum Z value. Readonly. */
protected double dataMin = Double.MAX_VALUE;

/** The current data's minimum Z value. Readonly.*/
protected double dataMax = Double.MAX_VALUE;

/**
 * A list of screen points which represent the transformed data values and
 * stores their pixel locations on the screen as well as some other flags.
 */
protected ScreenPoint[][]  screenPoints = null;

/**
 * Basic constructor. Parent and data arrays are set to null.
 */
Chart3dData()
{
}

/**
 * Internal constructor specifying parent data view and arrays holding data.
 *
 * @param dataView The parent data view
 * @see com.klg.jclass.chart3d.Chart3dDataView
 */
Chart3dData(Chart3dDataView dataView)
{
	this.dataView = dataView;
}

/**
 * Gets the value of the <code>dataView</code> property.
 * The <code>dataView</code> property points to the containing
 * <code>Chart3dDataView</code> for this data object
 *
 * @return The current parent data view for this object
 * @see com.klg.jclass.chart3d.Chart3dDataView
 */
public Chart3dDataView getDataView()
{
	return(dataView);
}

/**
 * Sets the parent <code>dataView</code>.
 *
 * @param dataView The new parent data view
 * @see com.klg.jclass.chart3d.Chart3dDataView
 */
public void setDataView(Chart3dDataView dataView)
{
	this.dataView = dataView;
}

/**
 * Gets the value of the name property. The name property
 * represents the name of this data object. In JClass Chart 3D,
 * data objects can be named, and can be retrieved by name.
 *
 * @return Data object name.
 */
public String getName()
{
	return(name);
}

/**
 * Sets the value of the name property. The name property
 * represents the name of the data object. In JClass Chart 3D,
 * data objects can be named, and can be retrieved by name.
 *
 * @param name The new name of the data object
 */
public void setName(String name) {
	this.name = name;
}

/**
 * Is the data valid?  In other words, can it be used to draw a surface, bar,
 * or scatter plot.  This object checks all data passed in to see if it is
 * valid.
 */
public boolean isDataOK()
{
	return(dataOK);
}

/**
 * Sets whether the data is valid or not (internal use only).
 *
 * @param dataOK The new value of dataOK.
 */
protected void setDataOK(boolean dataOK)
{
	this.dataOK = dataOK;
}

/**
 * Retrieves the hole value for the data.
 *
 * @return The current hole value
 */
public double getHoleValue()
{
	return(holeValue);
}

/**
 * Sets the hole value for the data (internal use only).  Users should not
 * set the hole value thru this method but rather thru the
 * HoleValueChart3dDataModel interface.
 *
 * @param holeValue The new hole value
 */
protected void setHoleValue(double holeValue)
{
	this.holeValue = holeValue;
}

/**
 * Retrieves the data min value (computed in sub-class).  Readonly.
 */
public double getDataMin()
{
	return(dataMin);
}

/**
 * Sets the data min value (usually set by sub-class).  Internal use only!
 *
 * @param dataMin The new dataMin value.
 */
protected void setDataMin(double dataMin)
{
	this.dataMin = dataMin;
}

/**
 * Retrieves the data max value (computed in sub-class).
 */
public double getDataMax()
{
	return(dataMax);
}

/**
 * Sets the data max value (usually set by sub-class).  Internal use only!
 *
 * @param dataMax The new dataMax value
 */
protected void setDataMax(double dataMax)
{
	this.dataMax = dataMax;
}
/**
 * Returns the current data source for this data object.
 *
 * @return The current data source
 */
public Chart3dDataModel getDataSource()
{
	return(dataSource);
}

/**
 * Returns the data handler for this data object.  A data handler is used to
 * handle data events and extract data from data sources
 *
 * @return The current data handler
 */
public Chart3dDataHandler getDataHandler()
{
	return(dataHandler);
}

/**
 * Returns the list of precomputed screen points for this data set.
 *
 * @return The precomputed screen points
 */
public ScreenPoint[][] getScreenPoints()
{
	return(screenPoints);
}

/**
 * Returns string representation of this data object.  If it has a name, use that.
 */
public String toString()
{
	String s = getName();
    if (s == null) {
        s = LocaleBundle.string(LocaleBundle.UNNAMED_DATA_OBJECT_ERROR);
    }
	return(s);
}

/**
 * Finds the closest match to a value in an array of doubles.
 *
 * @param v The input value (for closest match)
 * @param arr The array to search
 * @param n The size of the array (could be less than arr.length).
 */
public int getClosestPoint(double v, double[] arr, int n)
{
	if (arr == null) {
		return(0);
	}
	double dist = Double.MAX_VALUE;
	int match = 0;
	for (int i = 0; i < Math.min(n, arr.length); i++) {
		double newDist = Math.abs(arr[i] - v);
		if (newDist < dist) {
			dist = newDist;
			match = i;
		}
	}
	return(match);
}

/**
 * Usees a binary search to look for the value 't' in the array 'arr'.  The
 * index with the value closest to 't' but less than 't' is returned.  The
 * index returned is in the range [0..n-1].
 *
 * @param t The value to find
 * @param arr The array to search
 * @param n The size of the array (could be less than arr.length).
 */
public int getPointIndex(double t, double[] arr, int n)
{
	int l, m, u;

	if (n <= 1) {
		return(0);
	}

	l = 0;
	u = n-1;

	while (l <= u) {
		m = (l + u) / 2;
		if (arr[m] < t) {
			l = m+1;
		}
		else if (arr[m] > t) {
			u = m-1;
		}
		else {
			return(m);
		}
	}

	if (l == n || (l > 0 && arr[l] > t)) {
		l--;
	}

	// Make sure its within bounds
	l = Math.max(Math.min(l, n-1), 0);
	return(l);
}

/**
 * Finds the minimum delta (or step) in an array of doubles.
 *
 * @param arr The array to search
 * @param n The size of the array (could be less than arr.length).
 */
public double getMinStep(double[] arr, int n)
{
	double  minStep = Double.MAX_VALUE;

	if (!dataOK || arr == null || n < 1) {
		return(minStep);
	}

	if (n == 1) {
		/* Make step equal to value; if zero, make it 1 */
		minStep = arr[0];
		if ((minStep + 1.0) == 1.0) {
			minStep = 1.0;
		}
	}
	else {
		minStep = arr[1] - arr[0];
		for (int i = 1; i < n - 1; i++) {
			double step = arr[i+1] - arr[i];
			if (minStep > step) {
				minStep = step;
			}
		}
	}

	return(minStep);
}

/**
 * Notes that user specified <i>Z</i> max/min values cannot be inside the data
 * <i>Z</i> max/min (otherwise we need to do surface clipping).
 */
protected void checkValidZExtents()
{
	JCAxis axis = dataView.getChart3d().getChart3dArea().getAxis(JCAxis.AXIS_Z);

	double axisValue = axis.getMin();
	if (!axis.min.setValidValue(dataMin, true)) {
		
	}

	axisValue = axis.getMax();
	if (!axis.max.setValidValue(dataMax, false)) {
		
	}
}

/**
 * Overrides <code>setChanged</code> so that the flags are updated, passed on to the parent
 * <code>dataView</code>, and then reset.
 *
 * @param n Has anything changed?
 * @param cf The change mask which tells the type of change
 */
public void setChanged(boolean n, int cf)
{
	updateFlags(n, cf);
	if (dataView != null) {
		dataView.setChanged(changed, changedFlag);
	}
	updateFlags(false, Changeable.NO_CHANGE);
}

/**
 * Precompute the screen points (to avoid repeatedly transforming points)
 */
public abstract void precomputePoints();

/**
 * Calulate the annotation limits for this data object.  This might be
 * different than the data limits, since fixed bars extend outside of the
 * data range and the zOrigin may be outside of the data range.
 */
protected abstract void calcAnnoLimits();

}
