/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Chart3dDataView.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:56 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.event.*;
import com.klg.jclass.chart3d.resources.LocaleBundle;


import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Point4d;
import javax.vecmath.Vector2d;
import java.awt.*;
import java.util.ArrayList;

/**
 * Class containing the internal representation of the chartable data.
 * Chart3d data is represented as elevation data and zone data (used for 4D
 * grid plots).  User pass in a data source which the dataView extracts into
 * an internal data object.  The internal data object is then used to plot
 * the data.
 *
 * The <code>Chart3dDataView</code> also manages the drawing of the data,
 * handles changes to the data through <code>Chart3dDataEvent</code> objects
 * that are passed to it, and handles mapping and picking.
 */
public class Chart3dDataView extends TrackChange
                             implements Chart3dDataListener {

static final long serialVersionUID = -5542417106751358122L;

// Internal values to form a drawCode mask
public final static int MESHED    = 1;
public final static int SHADED    = 2;
public final static int ZONED     = 4;
public final static int CONTOURED = 8;

/** The parent <code>JCChart3d</code> object. */
protected JCChart3d             chart3d = null;

/**
 * Elevation data source for this <code>dataView</code>.  The elevation data is
 * the primary data for the <code>dataView</code> and provides a surface plot,
 * bar plot, or scatter plot depending on the <code>chartType</code>.
 */
protected Chart3dDataModel      elevationDataSource = null;

/**
 * The internal elevation data object.  This object allows user to query
 * information about the data that was supplied by the
 * <code>elevationDataSource</code>.
 */
protected Chart3dData           elevationData =
							        new Chart3dGridData(this, null);
/**
 * Zone data source for this <code>dataView</code>.  Zone data allows color to
 * be a 4th dimension for the corresponding elevation data.  The underlying grid
 * needs to match the elevation data's underlying grid.
 */
protected Chart3dGridDataModel  zoneDataSource = null;

/**
 * The internal zone data object.  This object allows user to query information
 * about the data that was supplied by the <code>zoneDataSource</code>.
 */
protected Chart3dGridData       zoneData =
							        new Chart3dGridData(this, null);

/** The chart type of this <code>dataView</code>.  One of <code>SURFACE</code>,
 * <code>BAR</code>, or <code>SCATTER</code>.
 * @see com.klg.jclass.chart3d.JCChart3d
 */
protected int                   chartType = JCChart3d.SURFACE;

/**
 * The <code>JCElevation</code> object determines meshing, shading, and
 * transparency.
 */
protected JCElevation           elevation = null;

/** The <code>JCContour</code> object handles contouring and zoning. */
protected JCContour             contour   = null;

/**
 * The name of this <code>dataView</code>.  Used by the legend code and the
 * customizer.
 */
protected String                name      = null;

/** Allows certain facets or bars to have different colors. */
protected ArrayList             gridColors = null;

/** The object used by this <code>dataView</code> to draw the data. */
protected Draw                  draw      = null;

// Internal variables
protected Mapping               mapping            = null;
protected Chart3dDataHandler    dataHandler        = null;
protected boolean               dataDrawn          = false;
protected int                   drawCode           = 0;
protected boolean               defaultsCalculated = false;
protected int                   dataViewId         = -1;
protected boolean               batched            = false;

/**
 * In Chart3d Java3d, when editing a SURFACE chart using the mouse, data points
 * around the pick point are modified according to the expression
 *
 *   deltaZ = scaleFactor * mouseDragDistance
 *                 * exp(-distanceFromPickPoint/radiusOfInfluence)
 *
 * where deltaZ is the change in the z coordinate value of a given point,
 *    scaleFactor = (MaxDataValue - MinDataValue) * mouseEditScaleFactor,
 *    mouseDragDistance is the mouse drag distance along the adjustment axis,
 *    distanceFromPickPoint is the distance in the x,y plane of a
 *                          data point from the pick point,
 *    radiusOfInfluence is a user-specified parameter.
 */
protected double radiusOfInfluence = 0.000001;
protected double mouseEditScaleFactor = 0.01;

/**
 * Null args constructor.  This means that <code>chart3d</code> is null (which
 * is not a good thing because many child objects call the
 * <code>getChart3d()</code> method and expect a valid <code>JCChart3d</code>
 * object to be there).
 */
public Chart3dDataView()
{
}

/**
 * Constructor which provides a parent <code>chart3d</code> and a
 * <code>dataViewId</code>.
 *
 * @param chart3d The parent chart3d
 * @param dataViewId The dataView index into the master dataView array
 */
public Chart3dDataView(JCChart3d chart3d, int dataViewId)
{
	this.chart3d    = chart3d;
	this.dataViewId = dataViewId;
}

/**
 * Constructor which provides a parent <code>chart3d</code>,
 * a <code>dataViewId</code>, and a <code>chartType</code>.
 *
 * @param chart3d The parent chart3d
 * @param dataViewId The dataView index into the master dataView array
 * @param chartType
 */
public Chart3dDataView(JCChart3d chart3d, int dataViewId, int chartType)
{
	this.chart3d    = chart3d;
	this.dataViewId = dataViewId;
	this.chartType  = chartType;
}

/**
 * Returns the parent <code>JCChart3d</code> object.
 *
 * @return The parent JCChart3d object
 */
public JCChart3d getChart3d()
{
	return(chart3d);
}

/**
 * Sets the parent <code>JCChart3d</code> object.
 *
 * @param chart3d The new parent JCChart3d object
 */
public void setChart3d(JCChart3d chart3d)
{
    if (this.chart3d == chart3d) {
        return;
    }
	this.chart3d = chart3d;
	setChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Gets the <code>elevationDataSource</code> property for
 * <code>Chart3dDataView</code>. This <code>elevationDataSource</code> property
 * represents the class used to load data into the <code>Chart3dDataView</code>
 * class. The <code>elevationDataSource</code> property holds this populating
 * class if it implements either <code>EditableChart3dDataModel</code> or
 * <code>Chart3dDataManager</code>. If the <code>elevationDataSource</code>
 * property is null, that means that the populating data source class was
 * discarded because it was no longer needed.
 *
 * @return Object acting as a data source for Chart3dDataView; this object
 *     implements either <code>EditableChart3dDataModel</code> or
 *     <code>Chart3dDataManager</code>.
 */
public Chart3dDataModel getElevationDataSource()
{
	return(elevationDataSource);
}

/**
 * Sets the elevation data source for this <code>Chart3dDataView</code>. The
 * data source  class must implement either <code>Chart3dDataModel</code> or
 * <code>EditableChart3dDataModel</code>. JClass Chart 3D will do the
 * correct operation based on the object provided.
 * If this data source class implements <code>Chart3dDataManager</code>,
 * this data view will be registered as a listener for data changes.
 * The data source class will be saved in the <code>elevationDataSource</code>
 * propertyfor future reference if it implements either
 * <code>Chart3dDataManager</code> or <code>EditableChart3dDataModel</code>.
 * Otherwise, since it is only needed to load data once,
 * the <code>elevationDataSource</code> property will be set to null.
 *
 * @param elevationDataSource Object acting as a elevation data source for
 *     Chart3dDataView; this object must implement either
 *     <code>Chart3dDataModel</code> or <code>EditableChart3dDataModel</code>.
 */
public void setElevationDataSource(Chart3dDataModel elevationDataSource)
{
	// First, remove any listeners on the old data source
	if (this.elevationDataSource instanceof Chart3dDataManager) {
		((Chart3dDataManager)this.elevationDataSource).removeChart3dDataListener(this);
	}

	boolean oldBatched = isBatched();
	setBatched(true);

    // Even if the reference is the same as before we must re-extract the data
	this.elevationDataSource = elevationDataSource;
	if (elevationDataSource == null) {
		elevationData = new Chart3dGridData(this, null);
	}
	else {
		if (elevationDataSource instanceof Chart3dGridDataModel) {
			Chart3dGridDataModel gridDataModel =
                (Chart3dGridDataModel)elevationDataSource;
			if (elevationData == null ||
                (elevationData instanceof Chart3dPointData))
            {
			    elevationData = new Chart3dGridData(this, gridDataModel);
			}
			else {
				Chart3dGridData gridData = (Chart3dGridData)elevationData;
				gridData.setDataSource(gridDataModel);
			}
		}
		else if (elevationDataSource instanceof Chart3dPointDataModel) {
			Chart3dPointDataModel pointDataModel =
                (Chart3dPointDataModel)elevationDataSource;
			if (elevationData == null ||
                (elevationData instanceof Chart3dGridData))
            {
			    elevationData = new Chart3dPointData(this, pointDataModel);
			}
			else {
				Chart3dPointData pointData = (Chart3dPointData)elevationData;
				pointData.setDataSource(pointDataModel);
			}
		}
		else {
		    throw new IllegalArgumentException(
		        LocaleBundle.string(LocaleBundle.UNRECOGNIZED_DATA_MODEL_ERROR));
		}

	    // If this data source implements Chart3dDataManager, register as a
		// listener on data changes
	    if (elevationDataSource instanceof Chart3dDataManager) {
		    ((Chart3dDataManager)elevationDataSource).addChart3dDataListener(this);
		}
	}

	setChanged(true, Changeable.ALL | Changeable.CONTOURS, false);
	setBatched(oldBatched);
}

/**
 * Returns the internal elevation data object.  This object is used internally
 * but is provided as a way to query the current elevation data parameters and
 * this object is needed as input to some convenience methods.  Note that
 * manipulating data inside the internal object can be done, but this is
 * strongly discouraged.  The recommended way to manipulate the data is through
 * the data model or through <code>Chart3dDataEvents</code>.
 *
 * @return The current internal elevation data object
 */
public Chart3dData getElevationData()
{
	return(elevationData);
}

/**
 * Gets the <code>zoneDataSource</code> property for
 * <code>Chart3dDataView</code>. This property allows the user to add a 4th
 * dimension to the data based on color. This <code>zoneDataSource</code>
 * property represents the class used to load data into the
 * <code>Chart3dDataView</code> class. The <code>zoneDataSource</code> property
 * holds this populating class if it implements either
 * <code>EditableChart3dDataModel</code> or <code>Chart3dDataManager</code>.
 * If the <code>zoneDataSource</code> property is null, that  means that the
 * populating data source class was discarded because it was no longer needed.
 *
 * @return Object acting as a data source for <code>Chart3dDataView</code>;
 *     this object implements either <code>EditableChart3dDataModel</code> or
 *     <code>Chart3dDataManager</code>.
 */
public Chart3dGridDataModel getZoneDataSource()
{
	return(zoneDataSource);
}

/**
 * Sets the zone data source for this <code>Chart3dDataView</code>. The
 * data source class must implement either <code>Chart3dGridDataModel</code> or
 * <code>EditableChart3dGridDataModel</code>. JClass Chart 3D will do the
 * correct operation based on the object provided.
 * If this data source class implements <code>Chart3dGridDataManager</code>,
 * this data view will be registered as a listener for data changes.
 * The data source class will be saved in the <code>zoneDataSource</code>
 * property for future reference if it implements either
 * <code>Chart3dDataManager</code> or <code>EditableChart3dDataModel</code>.
 * Otherwise, since it is only needed to load data once,
 * the <code>zoneDataSource</code> property will be set to null.
 *
 * @param zoneDataSource Object acting as a zone data source for
 *     <code>Chart3dDataView</code>; this object must implement either
 *     <code>Chart3dGridDataModel</code> or
 *     <code>EditableChart3dGridDataModel</code>.
 */
public void setZoneDataSource(Chart3dGridDataModel zoneDataSource)
{
	// First, remove any listeners on the old data source
	if (this.zoneDataSource instanceof Chart3dDataManager) {
		((Chart3dDataManager)this.zoneDataSource).removeChart3dDataListener(this);
	}

	boolean oldBatched = isBatched();
	setBatched(true);

    // Even if the reference is the same as before we must re-extract the data
	this.zoneDataSource = zoneDataSource;
	if (zoneDataSource == null) {
		zoneData = new Chart3dGridData(this, null);
	}
	else {
		if (zoneData == null) {
		    zoneData = new Chart3dGridData(this, zoneDataSource);
		}
		else {
			zoneData.setDataSource(zoneDataSource);
		}

	    // If this data source implements Chart3dDataManager, register as a
		// listener on data changes
	    if (zoneDataSource instanceof Chart3dDataManager) {
		    ((Chart3dDataManager)zoneDataSource).addChart3dDataListener(this);
		}
	}
	setChanged(true, Changeable.ALL | Changeable.CONTOURS, false);
	setBatched(oldBatched);
}

/**
 * Returns the internal zone data object.  This object is used internally
 * but is provided as a way to query the current zone data parameters and
 * is needed as input to some convenience methods.  Note that
 * manipulating data inside the internal object can be done, but this is
 * strongly discouraged.  The recommended way to manipulate the data is through
 * the data model or through <code>Chart3dDataEvents</code>.
 *
 * @return The current internal elevation data object
 */
public Chart3dGridData getZoneData()
{
	return(zoneData);
}

/**
 * Based on user inputs, is there any data drawn?  If the user input data that
 * doesn't make sense (i.e. an object that implements the
 * <code>Chart3dPointDataModel</code> when the <code>chartType</code> is
 * <code>SURFACE,/code>) or invalid data (i.e. an object the implements the
 * <code>Chart3dGridDataModel</code> which has less than two <i>x</i> or
 * <i>y</i> grid values) this flag will return false and nothing will be drawn
 * in the <code>JCChart3dArea</code>.
 *
 * @return Is there any data drawn?
 */
public boolean isDataDrawn()
{
	return(dataDrawn);
}

/**
 * Returns the current draw code mask.  This is a logical <code>OR</code> of
 * the <code>MESHED</code>, <code>SHADED</code>, <code>CONTOURED</code>, and
 * <code>ZONED</code> masks which are set based on the corresponding properties
 * in the child <code>JCContour</code> and <code>JCElevation</code> objects.
 * This is meant to be used as an internal mask but is exposed if users want to
 * query it. Note that this mask gets set somewhere during the drawing process.
 *
 * @return The current draw code
 */
public int getDrawCode()
{
	return(drawCode);
}

/**
 * Gets the <code>dataView</code> ID number. This number is index into the
 * master <code>dataView</code> array stored in the parent
 * <code>JCChart3d</code> class.
 *
 * @return The current <code>dataView</code> ID number
 */
public int getDataViewId()
{
	return(dataViewId);
}

/**
 * Sets the <code>dataView</code> ID number. This number is index into the
 * master <code>dataView</code> array stored in the parent
 * <code>JCChart3d</code> class.  Note that currently only one dataView is
 * supported.
 *
 * @param dataViewId The new <code>dataView</code> ID number
 */
public void setDataViewId(int dataViewId)
{
    if (this.dataViewId == dataViewId) {
        return;
    }
	this.dataViewId = dataViewId;
	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Return the chart type for this dataView.  This is one of SURFACE, BAR, or
 * SCATTER (defined in the JCChart3d class).
 *
 * @return The chart type for this dataView
 * @see com.klg.jclass.chart3d.JCChart3d
 */
public int getChartType()
{
	return(chartType);
}

/**
 * Sets the chart type for this <code>dataView</code>.  This is one of
 * <code>SURFACE</code>, <code>BAR</code>, or <code>SCATTER</code> (defined in
 * the <code>JCChart3d</code> class).
 *
 * @param chartType
 * @see com.klg.jclass.chart3d.JCChart3d
 */
public void setChartType(int chartType)
{
    if (this.chartType == chartType) {
        return;
    }
	if (chartType >= JCChart3d.SURFACE &&
		chartType < JCChart3d.LAST_PREDEFINED_CHART_TYPE)
	{
		this.chartType = chartType;
	}
	else {
        throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_CHART_TYPE_ERROR));
	}
	setChanged(true, Changeable.ALL);
}

/**
 * Gets the <code>dataView</code>'s name.  The name is used in the legend and
 * customizer to distinguish between <code>dataViews</code>.
 *
 * @return The current <code>dataView</code> name
 */
public String getName()
{
	return(name);
}

/**
 * Sets the <code>dataView</code>'s name.  The name is used in the legend and
 * customizer to distinguish between <code>dataViews</code>.
 *
 * @param name The new <code>dataView</code> name
 */
public void setName(String name)
{
    if (this.name == name) {
        return;
    }
	this.name = name;
	setChanged(true, Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Returns the current list of grid colors.  This list allows the user to color
 * a particular bar or facet, row/column of bars or facets with a given color.
 *
 * @return The current grid colors list
 */
public ArrayList getGridColors()
{
	return(this.gridColors);
}

/**
 * Sets the grid color list.  This list allows the user to color a particular
 * bar or facet, row/column of bars or facets with a given color.
 *
 * @param gridColors The new grid color list
 */
public void setGridColors(ArrayList gridColors)
{
	// Even if they're equal, we need to redraw because the contents of the
	// array list could have changed.
	this.gridColors = gridColors;
	filterGridColors();
	setChanged(true, Changeable.REDRAW);
}

/**
 * Finds the first <code>JCGridColor</code> (starting at the end of the list) in
 * the <code>gridColors</code> list that matches the indices (<i>x</i>, <i>y</i>).
 *
 * @param x The x index to match
 * @param y The y index to match
 */
public JCGridColor findGridColor(int x, int y)
{
    if (gridColors == null || gridColors.size() == 0) {
        return(null);
    }
    JCGridColor returnGridColor = null;

    int i = gridColors.size() - 1;
    while (returnGridColor == null && i >= 0) {   /* Is there a match? */
		JCGridColor gridColor = (JCGridColor)gridColors.get(i);
		int curX            = gridColor.dataIndex.getX();
		int curY            = gridColor.dataIndex.getY();
        if ((curX == JCData3dIndex.ALL || curX == x)  &&
            (curY == JCData3dIndex.ALL || curY == y))
        {
            returnGridColor = gridColor;
			break;
        }
        i--;
    }
    return(returnGridColor);
}

/**
 * Removes the first matching entry from the <code>gridColors</code>
 * <code>ArrayList</code> (starting from the <i>n</i>th entry back to the
 * first).  The entry must match exactly (<i>x</i> ==
 * <code>JCData3dIndex.ALL</code> will only match if the
 * <code>gridColor</code>'s index is <code>JCData3dIndex.ALL</code>).
 * Internal use only.
 *
 * @param x The x index of the grid color to be removed
 * @param y The y index of the grid color to be removed
 * @param n The number of grid colors in the array to scan (could be less
 *   than gridColors.size())
 * @param callSetChanged If true, mark the dataView as changed and redraw.
 */
protected boolean removeGridColor(int x, int y, int n, boolean callSetChanged)
{
    if (gridColors == null || gridColors.size() == 0) {
        return(false);
    }

	n = Math.min(n, gridColors.size());
	for (int i = n - 1; i >= 0; i--) {
		JCGridColor gridColor = (JCGridColor)gridColors.get(i);
        JCData3dGridIndex dataIndex = gridColor.getDataIndex();
        if (dataIndex.getX() == x && dataIndex.getY() == y) {
			// A match has occured -- Remove it
			gridColors.remove(i);
			if (callSetChanged) {
	            setChanged(true, Changeable.REDRAW);
			}
			return(true);
		}
	}
	return(false);
}

/**
 * Removes the first matching entry from the <code>gridColors</code>
 * <code>ArrayList</code> (starting from the last to the first).  The entry must
 * match exactly (<i>x</i> == <code>JCData3dIndex.ALL</code> will only match if
 * the <code>gridColor</code>'s index is <code>JCData3dIndex.ALL</code>).
 *
 * @param x The x index of the grid color to be removed
 * @param y The y index of the grid color to be removed
 */
public boolean removeGridColor(int x, int y)
{
	if (gridColors == null) {
		return(false);
	}
	return(removeGridColor(x, y, gridColors.size(), true));
}

/**
 * Sets the color of the specified bar (or set of bars) by creating a
 * <code>JCGridColor</code> and adding it to the <code>gridColors</code>
 * <code>ArrayList</code>.  If the <i>x</i> and <i>y</i> of the
 * <code>gridColor</code> match an entry already on the list then that entry
 * will be replaced by the new one.  If the color of the <code>gridColor</code>
 * is null, then remove the matching entry without replacing it.
 *
 * @param x The x index of the grid color to be added
 * @param y The y index of the grid color to be added
 * @param color The color of the grid color to be added
 */
public void addGridColor(int x, int y, Color color)
{
	addGridColor(new JCGridColor(x, y, color));
}

/**
 * Sets the color of the specified bar (or set of bars) by adding the passed
 * in <code>gridColor</code> to the <code>gridColors</code>
 * <code>ArrayList</code>.   If the <i>x</i> and <i>y</i> of the
 * <code>gridColor</code> match an entry already on the list then that entry
 * will be replaced by the new one.  If the color of the <code>gridColor</code>
 * is null, then remove the matching entry without replacing it.
 *
 * @param gridColor The gridColor object to be added.
 */
public void addGridColor(JCGridColor gridColor)
{
	// Check for valid inputs
	if (gridColor == null) {
		return;
	}
	int x = gridColor.getX();
	int y = gridColor.getY();
	if ((x < 0 && x != JCData3dIndex.ALL) ||
        (y < 0 && y != JCData3dIndex.ALL))
    {
	    throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_X_OR_Y_INDEX_ERROR));
	}

	if (gridColors == null) {
		gridColors = new ArrayList();
	}
	removeGridColor(gridColor.getX(), gridColor.getY(), gridColors.size(),
                    false);
	if (gridColor.getColor() != null) {
		gridColors.add(gridColor);
		gridColor.parent = this;
	}
	setChanged(true, Changeable.REDRAW);
}

/**
 * Processes <code>gridColors</code>.  Starting at the first entry, basically
 * act like you are adding each entry to a new list.  Duplicate entries will
 * replace old ones and null colors will wipe out a matching entry.  Note that
 * any new entry always be placed at the end of the list.
 */
protected void filterGridColors()
{
	if (gridColors == null) {
		return;
	}

	int n = 0;  // The number of entries in the filtered list

	while (n < gridColors.size()) {
		JCGridColor       gridColor = (JCGridColor)gridColors.get(n);
        JCData3dGridIndex dataIndex = gridColor.getDataIndex();

		/* Remove duplicate entry if there is one */
		if (n == 0) {
			if (gridColor.getColor() == null) {
			   // The first entry can't have a null color -- remove it.
				gridColors.remove(n);
			}
			else {
				n++;
			}
		}
		else if (removeGridColor(dataIndex.getX(), dataIndex.getY(), n, false)) {
			if (gridColor.getColor() == null) {
				// We've wiped out a matching entry.  We need to remove this
				// entry as well.
				n--;
				gridColors.remove(n);
			}
		}
		else { // A new, non-duplicate entry
			n++;
		}
	}
	for (int i = 0; i < gridColors.size(); i++) {
		JCGridColor gridColor = (JCGridColor)gridColors.get(i);
		gridColor.parent = this;
	}
}

/**
 * Returns the object used to draw the data.
 *
 * @return The object used to draw the data
 */
public Draw getDraw()
{
	return(draw);
}

/**
 * Returns the current <code>JCElevation</code> object for this
 * <code>dataView</code>.
 *
 * @return The current elevation object.
 */
public JCElevation getElevation()
{
	return(elevation);
}

/**
 * Sets a <code>JCElevation</code> object for this <code>dataView</code>.
 * Setting the elevation property to null not allowed.
 *
 * @param elevation The new <code>JCElevation</code> objection
 */
public void setElevation(JCElevation elevation)
{
	if (this.elevation == elevation) {
		return;
	}
	this.elevation = elevation;
	if (this.elevation != null) {
	    this.elevation.addParent(this);
	}
	else {
        throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NULL_ERROR));
	}
	setChanged(true, Changeable.ALL);
}

/**
 * Returns the current <code>JCContour</code> object for this
 * <code>dataView</code>.
 *
 * @return The current contour object.
 */
public JCContour getContour()
{
	return(contour);
}

/**
 * Sets a <code>JCContour</code> object on this <code>dataView</code>.  Note
 * that since the <code>JCContour</code> object is heavily dependent on the
 * <code>Chart3dDataView</code>'s data (among other things), a contour object
 * can only have one <code>dataView</code> parent.  Also, setting the contour
 * property to null not allowed.
 *
 * @param contour The new <code>JCContour</code> object
 */
public void setContour(JCContour contour)
{
	if (this.contour == contour) {
		return;
	}
	this.contour = contour;
	if (this.contour != null) {
		// Note that setDataView calls contour's addParent method and setChanged
		this.contour.setDataView(this);
	}
	else {
        throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NULL_ERROR));
	}
}

/**
 * Returns the current mapping object.
 *
 * @return The current mapping object
 */
public Mapping getMapping()
{
	return(mapping);
}

/**
 * Sets the mapping object.
 *
 * @param mapping The new mapping object
 */
public void setMapping(Mapping mapping)
{
	this.mapping = mapping;
}

/**
 * Gets the <code>Batched</code> property. The <code>Batched</code> property
 * controls whether the <code>Chart3dDataView</code> is notified immediately
 * of data model changes, or if the changes are accumulated and sent at a
 * later date.
 *
 * @return true if changes are accumulated, false otherwise.
 */
public boolean isBatched()
{
    return(batched);
}

/**
 * Sets the <code>Batched</code> property. The <code>Batched</code> property
 * controls whether the <code>Chart3dDataView</code> is notified immediately of
 * data model changes, or if the changes are accumulated and sent at a later date.
 *
 * @param isBatched Are updates batched until later?
 */
public void setBatched(boolean isBatched)
{
	if (batched == isBatched) {
		return;
	}
	batched = isBatched;

    // Check batched because batched is not synchronized and we
    // want to call propagateChange() without a local lock.
    if (batched == false) {
		propagateChange();
    }
}

/**
 * <code>chart3dDataChange</code> is called whenever the chart 3D data has
 * changed.  Interested listeners should implement this method and register the
 * object with the appropriate data source.
 *
 * @param event The data event to process
 */
public void chart3dDataChange(Chart3dDataEvent event)
{
	if (event == null) {
		return;
	}

	Object src = event.getSource();
	if (src == null) {
		return;
	}

	// Find the internal data structure associated with the given source
	Chart3dGridData  gridData  = null;
	Chart3dPointData pointData = null;
	if (src instanceof Chart3dGridDataModel) {
	    Chart3dGridDataModel  gridModel = (Chart3dGridDataModel)src;
		if (gridModel == elevationDataSource &&
		    (elevationData instanceof Chart3dGridData))
		{
			gridData = (Chart3dGridData)elevationData;
		}
		else if (gridModel == zoneDataSource) {
			gridData = zoneData;
		}
	}
	else if (src instanceof Chart3dPointDataModel) {
	    Chart3dPointDataModel pointModel = (Chart3dPointDataModel)src;
		if (pointModel == elevationDataSource &&
		    (elevationData instanceof Chart3dPointData))
		{
			pointData = (Chart3dPointData)elevationData;
		}
	}

	// Make sure event types match data types
	Chart3dDataHandler dataHandler = null;
	if (gridData != null) {
	    if (event instanceof Chart3dPointDataEvent) {
		    return;
	    }
        dataHandler = gridData.getDataHandler();
	}
	if (pointData != null) {
		if (event instanceof Chart3dGridDataEvent) {
		    return;
		}
		dataHandler = pointData.getDataHandler();
	}

	if (dataHandler != null) {
        if (dataHandler.handleEvent((Chart3dDataModel)src, event)) {
	        setChanged(true, Changeable.ALL | Changeable.CONTOURS);
        }
	}
}

/**
 * Sets the value of the <code>Changed</code> property. The <code>Changed</code>
 * property manages whether the data view requires recalculation.  If set to
 * true, a recalculation may be triggered.
 *
 * @param b Has anything been changed?
 * @param cf The change mask the tells what type of changed has happened
 */
public void setChanged(boolean b, int cf)
{
	changed = b;

	// Only changed = true is interesting, as it can propagate elsewhere
	if (!changed) {
		changedFlag = cf;
		return;
	}
	else {
		changedFlag |= cf;
	}

    // Mark chart area, contour, and legend as changed.
    markAsChanged(cf);

    // Possibly force update
    propagateChange();
}

/**
 * Utility routine used to force an update as a result of a
 * change inside <code>Chart3dDataView</code>.
 *
 */
private void propagateChange()
{
	// Only send change if change is true
	if (!changed) {
        return;
	}

	// Don't send change if Chart3dDataView is batched
	if (isBatched()) {
        return;
	}

	// Don't send change if Chart is null or batched.
	if (chart3d == null || chart3d.isBatched()) {
        return;
	}

    // JCChart3d.update will invalidate, validate and repaint.
    chart3d.update();

    // Now that changes have been marked and propagated, clear the flags here
	// DAM unsure about this -- CA clears out data view changes flags
	// too.  Doing it here only ensures CDV isn't marked as changed when
	// a chart finally becomes visible.
    setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Utility routine used to mark appropriate objects as changed
 * as a result of a change in <code>Chart3dDataView</code>.  Note that the
 * changes are not allowed to propagate up to the parent.
 *
 * @param cf The change mask that tells what type of change has happened
 */
private void markAsChanged(int cf) {
    // Chart3dArea should be marked as changed.
    if (chart3d != null && chart3d.getChart3dArea() != null) {
        chart3d.getChart3dArea().setChanged(true, cf & BASIC_CHANGE_MASK, false);
    }

    // Check if contours need to be recalculated.  If so, destroy the old ones
    // as the chart will only calculate the contours if they are null.
    if (((cf & Changeable.CONTOURS) > 0) && contour != null) {
        contour.destroyContours();
    }

    // Mark legend as changed if necessary
    int legendCF = Changeable.NO_CHANGE;
    if ((cf & Changeable.NOTIFY_LEGEND_RECALC) > 0) {
        legendCF |= Changeable.RECALC;
    }
    else if ((cf & Changeable.NOTIFY_LEGEND_REDRAW) > 0) {
        legendCF |= Changeable.REDRAW;
    }
    if (legendCF != Changeable.NO_CHANGE) {
        if (chart3d != null && chart3d.getLegend() != null) {
            chart3d.getLegend().setChanged(true, legendCF, false);
        }
    }
}

/**
 * Sets the value of the <code>Changed</code> property, with option to prevent parent
 * updating via the <code>updateParent</code> parameter. This method overrides the
 * <code>TrackChange</code> method so that it can mark other appropriate objects as
 * changed.
 * Note that <code>updateParent</code> is independent of <code>Batched</code>.
 *
 * @param b Has a change happened?
 * @param cf The change mask that tells what type of change has happened
 * @param updateParent Tell our parent about the change (which will cause
 *   a repaint)?
 */
public void setChanged(boolean b, int cf, boolean updateParent)
{
    if ((changed == b) && ((changedFlag & cf) == cf)) {
        return;
    }
    if (updateParent) {
        setChanged(b, cf);
    }
    else {
        changed = b;
        if (changed)  {
            changedFlag |= cf;
            markAsChanged(cf);
        }
        else {
            changedFlag = cf;
        }
    }
}

/**
 * Computes the limits of the data (i.e. maximum and minimum) for all three
 * dimensions based on the data input.  These are used to calculate axis
 * limits, the viewport, and plot cube boundaries.
 */
public void computeDataLimits()
{
	if (elevationDataSource != null && !elevationData.isDataOK()) {
		
        return;
	}
	if (zoneDataSource != null && !zoneData.isDataOK()) {
		
        return;
	}

	defaultsCalculated = false;
	if (elevationData instanceof Chart3dPointData) {
		Chart3dPointData pointData = (Chart3dPointData)elevationData;
		pointData.calcLimits();
	}
	else {
		Chart3dGridData gridData = (Chart3dGridData)elevationData;
		gridData.calcLimits(zoneData);
	}
}

/**
 * Will we draw anything?  Computes a <code>drawCode</code> (for easier access
 * to the four booleans that control the type of surface, bar, or scatter plot).
 * Also computes another boolean <code>dataDrawn</code> that determines if
 * anything is to be drawn.
 */
protected void computeDataDrawn()
{
	drawCode = 0;
	if (elevation.isMeshed()) {
		drawCode |= MESHED;
	}
	if (elevation.isShaded()) {
		drawCode |= SHADED;
	}
	if (contour != null) {
		if (contour.isZoned()) {
			drawCode |= ZONED;
		}
		if (contour.isContoured()) {
			drawCode |= CONTOURED;
		}
	}

	if (chart3d.isProjection()) {
	   dataDrawn = (drawCode >= ZONED);  // ZONED or CONTOURED
	}
	else {
	   dataDrawn = (drawCode != 0);         // At least one is true
	}

    boolean isGridData = elevationData instanceof Chart3dGridData;
    if (isGridData) {
        Chart3dGridData gridData = (Chart3dGridData)elevationData;
        if (chartType == JCChart3d.SURFACE &&
            (gridData.numX <= 1 || gridData.numY <= 1))
        {
            // Must have at least 2 points in each direction to draw a surface
            dataDrawn = false;
        }
    }

	// We don't understand point data unless the chart type is scatter but we do
	// understand grid data no matter what the chart type
	dataDrawn = (dataDrawn &&
			     (chartType == JCChart3d.SCATTER || isGridData));
}

/**
 * Calculates default values for those values that were not provided by the user.
 * For instance, this calculates the data min and max and sets the axis min
 * and max if the "useDefaults" are true.  It also calculates the default
 * contour levels.  The user can call this method if he/she requires these
 * values before rendering the chart.
 */
public void calcDefaults()
{
	computeDataLimits();
	computeDataDrawn();

	// !!! Text areas -- adjust cube limits goes here

	if (elevationData.isDataOK()  &&
	    isDataDrawn()             &&
		elevationData instanceof Chart3dGridData)
	{
		contour.getContourLevels().calcLevels();
	}
}

/**
 * Creates the correct draw object based on the data type.
 */
public void createDraw()
{
	// Create a grid or point draw object depending on the data
    boolean isGridData = (elevationData instanceof Chart3dGridData);
    if (draw == null) {
        if (isGridData) {
            draw = new DrawGrid(this);
        }
        else {
            draw = new DrawPoint(this);
        }
    }
    else {
		// Already a draw object.  Only create a new one if the type changed.
        if (isGridData && (draw instanceof DrawPoint)) {
            draw = new DrawGrid(this);
        }
        if (!isGridData && (draw instanceof DrawGrid)) {
            draw = new DrawPoint(this);
        }
    }

	// Set the correct mapping object on the draw object
	if (draw.getMapping() == null ||
	    draw.getMapping().getData() != elevationData)
	{
		if (mapping == null) {
			
			return;
		}
		draw.setMapping(mapping.createMapping(elevationData));
	}
}

/**
 * Does preprocessing needed before drawing.  Based on the user input compute
 * the draw object needed, process text objects, and calculate processing order.
 * Also, precomputes screen points for both data and contours.
 */
protected void doPreprocessing()
{
	if (!isDataDrawn()) {
		return;
	}

	// Create the proper draw object
	createDraw();

	// Compute the processing order and the grid regions (if grid data)
	if (!chart3d.isProjection()) {
		chart3d.getChart3dArea().getTransform().calcProcessingOrder();
		if (elevationData instanceof Chart3dGridData) {
			DrawGrid drawGrid = (DrawGrid)draw;
			drawGrid.getGridRegion().calcRegions();
		}
	}

	// Preprocess text objects here

	if (chartType != JCChart3d.SURFACE || drawCode != CONTOURED) {
		// We always need screen points for scatter plots and bar
		// plots(because bartext objects in 3D rely on screen_points)
		// but do not need them for surfaces with only contours.
	   elevationData.precomputePoints();
	}

	// Preprocess contours
	if (chartType != JCChart3d.SCATTER &&
	    drawCode >= ZONED)  // Do we have ZONES or CONTOURS?
	{
		contour.precompContours();
	}
}

/**
 * Actually draws the graph.  The drawing mechanism is quite different for a
 * normal 3d graph vs a 2d projection (only contours and zones).  This method
 * goes thru each of the graph parts (plotCube, Annotation, Gridlines, graph,
 * etc) and draws them in the correct order.
 *
 * @param gc The graphics object used to draw
 */
protected void drawGraph(Graphics gc)
{
	boolean         zMinBackFacing = false;
	boolean         zMaxBackFacing = false;
	JCChart3dArea   area           = getChart3d().getChart3dArea();
	Anno            anno           = area.getAnno();
	Cube            cube           = area.getCube();
	JCPlotCube      plotCube       = area.getPlotCube();
    boolean         isGridData     = (elevationData instanceof Chart3dGridData);

	if (chart3d.isProjection()) {
		if (isDataDrawn() && (elevationData.isDataOK() || zoneData.isDataOK())) {
			doPreprocessing();
	        plotCube.drawBoundingBox(gc, true);
			boolean drawPoints = (!isGridData || chartType == JCChart3d.SCATTER);
			if (!drawPoints) {
				JCContour contour = getContour();
				contour.draw2dZones(gc);
				contour.drawContours(gc);
			}
			anno.draw2dAnno(gc);
			plotCube.drawGridLines(gc);
	        plotCube.drawBoundingBox(gc, false);
			if (drawPoints) {
				draw.draw2dPoints(gc);
			}
			// Draw 2d text objects here
		}
	}
	else {
		if (elevationData.isDataOK()) {
			if (chartType == JCChart3d.SURFACE && plotCube.hasProjections()) {
				zMinBackFacing = cube.faceBackfacing(Cube.FACE_FAR);
				zMaxBackFacing = cube.faceBackfacing(Cube.FACE_NEAR);
			}
		    doPreprocessing();
			anno.draw3dAnno(gc);
	        plotCube.drawBoundingBox(gc, true);
			if (chartType == JCChart3d.SURFACE && isGridData) {
				if (plotCube.hasFloorProjection() && zMinBackFacing) {
					plotCube.drawProjection(gc, true);
				}
				if (plotCube.hasCeilingProjection() && zMaxBackFacing) {
					plotCube.drawProjection(gc, false);
				}
			}
			plotCube.drawGridLines(gc);
			if (isDataDrawn()) {
                draw.draw(gc);
			}
			if (chartType == JCChart3d.SURFACE && isGridData) {
				if (plotCube.hasCeilingProjection() && !zMaxBackFacing) {
					plotCube.drawProjection(gc, false);
					plotCube.finishEdges(gc);
				}
				if (plotCube.hasFloorProjection() && !zMinBackFacing) {
					plotCube.drawProjection(gc, true);
					plotCube.finishEdges(gc);
				}
			}
		}
	}
}

/**
 * Performs a map operation by mapping the provided screen pixel
 * information to a point in 3d space that is on the surface (if the point
 * misses the surface a point with (<i>x</i>, <i>y</i>, <i>z</i>) values of
 * <code>Double.MAX_VALUE</code> is returned.
 *
 * @param x <i>x</i> value in screen pixels
 * @param y <i>y</i> value in screen pixels
 * @return The data point that maps to the inputted pixel point.
 */
public Point3d map(int x, int y)
{
    // Adjust co-ordinate to be within chart area drawing area.
    Rectangle r1 = chart3d.getChart3dArea().getDrawingArea();
	Point     l  = chart3d.getChart3dArea().getLocation();
    x -= (r1.x + l.x);
    y -= (r1.y + l.y);

	// Make sure that the dataView has a draw object
	createDraw();
	return(draw.getMapping().map(x, y));
}

/**
 * Performs an unmap operation by taking a point in 3d data space and
 * transforming it to screen space.
 *
 * @param point The point in 3d dat space to be transformed.
 * @return AWT Point object representing the location in screen pixels
 *     (relative to the Chart 3D component)
 */
public java.awt.Point unmap(Point3d point)
{
	Rectangle rl = chart3d.getChart3dArea().getDrawingArea();
	Point     l  = chart3d.getChart3dArea().getLocation();

	// Make sure that the dataView has a draw object
	createDraw();
	Point p = draw.getMapping().unmap(point);
	if (p != null) {
		p.x += rl.x + l.x;
	    p.y += rl.y + l.y;
	}
    return(p);
}

/**
 * Same as map().  Converts pixel coordinates to data space coordinates.
 *
 * @param x <i>x</i> value in screen pixels
 * @param y <i>y</i> value in screen pixels
 * @return <code>Point3d</code> instance
 */
public Point3d coordToDataCoord(int x, int y)
{
	return(map(x, y));
}

/**
 * Same as unmap().  Converts data coordinates to pixel coordinates.
 *
 * @param point The point in 3d data space to be transformed.
 * @return AWT Point object representing the location in screen pixels
 *     (relative to the Chart 3D component)
 */
public java.awt.Point dataCoordToCoord(Point3d point)
{
	return(unmap(point));
}

/**
 * Similar to unpick() for a specific this specific data view.  Converts a
 * <code>JCData3dIndex</code> instance (containing a data view and an point
 * index for either grid data or point data) to pixel values relative to the
 * Chart 3D component.
 *
 * @param index Object representing the index of the point to unpick. This is
 *     either a <code>JCData3dGridIndex</code> representing the (<i>x</i>,
 *     <i>y</i>) index of grid data point or a <code>JCData3dPointIndex</code>
 *     representing the (series, point) index of a point in a point data set.
 * @return AWT Point object representing the location is screen pixels
 *     relative to the Chart 3D component.
 */
public java.awt.Point dataIndexToCoord(JCData3dIndex index)
{
    return(chart3d.unpick(this, index));
}

/**
 * Similar to <code>pick()</code> for a specific data view.  Converts pixel values
 * relative to the Chart 3D component to a <code>JCData3dIndex</code> instance
 * representing the index of the picked point.  This is either a
 * <code>JCData3dGridIndex</code> representing the (<i>x</i>, <i>y</i>) index of
 * grid data point or a <code>JCData3dPointIndex</code> representing the (series,
 * point) index of a point in a point data set.
 *
 * @param x The <i>x</i> value of screen position
 * @param y The <i>y</i> value of screen position
 * @return The <code>JCData3dIndex</code> object representing the index of the
 *      picked point.
 */
public JCData3dIndex coordToDataIndex(int x, int y)
{
    Point l = chart3d.getChart3dArea().getLocation();
	return(chart3d.getChart3dArea().pick(new Point(x - l.x, y - l.y), this));
}

/**
 * Finds a new <i>Z</i> value for a given point based on a given pixel position.
 * Given a start point <i>A</i> (for grid data a point on the grid and for point
 * data one of the points in the list of series) and a point <i>P</i> on the
 * screen, project the line <i>AP</i> (in 3-space) onto the line through
 * <i>A</i> parallel to the <i>z</i> axis and find the <i>z</i> value that
 * corresponds to <i>P</i> on the projected line.
 *
 * @param data The data for which this operation is to take place.  It is
 *     either an instance of <code>Chart3dGridData</code> or
 *     <code>Chart3dPointData</code>.
 * @param index The data index of the point.  For grid data, this must be an
 *     instance of <code>JCData3dGridIndex</code>, which corresponds to an
 *     <i>x</i> and <i>y</i> grid position specification.  For point data, this
 *     must be an instance of <code>JCData3dPoint</code> index which corresponds
 *      to the series and point number of the point.
 * @param x The <i>x</i> value of screen position
 * @param y The <i>y</i> value of screen position
 * @return The new computed z-value
 */
public double dragZValue(Chart3dData data, JCData3dIndex index, int x, int y)
{
	if (data == null || !data.isDataOK() || index == null) {
	    throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.BAD_DATA_OR_INDEX_ERROR));
	}

	// Check input types
	boolean isGridData = (data instanceof Chart3dGridData);
	if ((isGridData  && !(index instanceof JCData3dGridIndex)) ||
	    (!isGridData &&  (index instanceof JCData3dGridIndex)))
	{
	    throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.DATA_AND_INDEX_DONT_MATCH_ERROR));
	}

    // Adjust co-ordinate to be within chart area drawing area.
	Point caLocation = chart3d.getChart3dArea().getLocation();
    x -= caLocation.x;
    y -= caLocation.y;

	// Check input values
	double xValue = Double.MAX_VALUE;
	double yValue = Double.MAX_VALUE;
	double zValue = data.getHoleValue();
	if (isGridData) {
		Chart3dGridData   gridData  = (Chart3dGridData)data;
		JCData3dGridIndex gridIndex = (JCData3dGridIndex)index;
		int xIndex = gridIndex.getX();
		int yIndex = gridIndex.getY();
		if (xIndex < 0 || xIndex >= gridData.getNumX() ||
		    yIndex < 0 || yIndex >= gridData.getNumY())
		{
	        throw new IllegalArgumentException(
                LocaleBundle.string(LocaleBundle.INVALID_X_OR_Y_INDEX_ERROR));
		}
		xValue = gridData.getX(xIndex);
		yValue = gridData.getY(yIndex);
		zValue = gridData.getZValue(xIndex, yIndex);
	}
	else {
		Chart3dPointData   pointData  = (Chart3dPointData)data;
		JCData3dPointIndex pointIndex = (JCData3dPointIndex)index;
		int series = pointIndex.getSeries();
		int point  = pointIndex.getPoint();
		if (series < 0 || series >= pointData.getNumSeries() ||
		    point  < 0 || point  >= pointData.getSeries(series).getNumPoints())
		{
	        throw new IllegalArgumentException(
                LocaleBundle.string(LocaleBundle.BAD_SERIES_OR_POINT_ERROR));
		}
		Point3d pointXYZ = pointData.getSeries(series).getPoint(point);
		xValue = pointXYZ.x;
		yValue = pointXYZ.y;
		zValue = pointXYZ.z;
	}
	if (zValue == data.getHoleValue()) {
		// Doesn't make sense to compute for hole value.
		return(zValue);
	}

	JCChart3dArea area      = data.getDataView().getChart3d().getChart3dArea();
	JCAxis        zAxis     = area.getAxis(JCAxis.AXIS_Z);
	Transform     transform = area.getTransform();

	// Get pixel values for the line parallel to the zaxis thru (xvalue, zvalue)
	Point4d pt = new Point4d(xValue, yValue, zAxis.getMin(), 1.0);
	transform.transformPoint(transform.getView(), pt);
	Point point0 = new Point((int) Math.round(pt.x), (int)Math.round(pt.y));

	pt.set(xValue, yValue, zAxis.getMax(), 1.0);
	transform.transformPoint(transform.getView(), pt);
	Point point1 = new Point((int) Math.round(pt.x), (int)Math.round(pt.y));

	Point point2 = new Point(x, y);

	//	Compute the projection of the given pixel value onto the line
	//	parallel to the z-axis found above.  Do this all in pixel-space (2D).
	Vector2d vec1 = new Vector2d(point1.x - point0.x, point1.y - point0.y);
	Vector2d vec2 = new Vector2d(point2.x - point0.x, point2.y - point0.y);

	double v1_dot_v2        = vec1.dot(vec2);
	double len_v1_sq        = vec1.x * vec1.x + vec1.y * vec1.y;
	if (1.0 + len_v1_sq == 1.0) {
		// If this is true, then the zAxis is very small (as projected on
		// the screen) and the old z-value is as good as any value.
		return(zValue);
	}
	double factor           = v1_dot_v2 / len_v1_sq;
	double[] projectedPoint = new double[2];
	projectedPoint[0]       = factor * vec1.x + point0.x;
	projectedPoint[1]       = factor * vec1.y + point0.y;

	//	Look for the largest element in absolute value in the first
	//	two rows of column 2 (the x and y entries of the z column) of
	//	the viewing transform matrix and use the corresponding row to
	//	compute the z value. This gives the most numerically stable
	//	way of computing this value.
	Matrix4d view     = transform.getView();
	double   maxValue = 0.0;
	int      maxIndex = -1;
	for (int i = 0; i < 2; i++) {
		double value = Math.abs(projectedPoint[i] * view.m32 - view.getElement(i, 2));
		if (maxValue < value) {
			maxValue = value;
			maxIndex = i;
		}
	}
	if (maxIndex < 0 || (maxValue + 1.0 == 1.0)) {
		// Must be looking straight down the z axis in which it
		// doesn't make sense to compute z value.  Return old value of z.
		return(zValue);
	}

	//	Use row r (max_index: chosen above) and the 4th row (corresponding
	//	to w) to solve for zvalue.  Given the viewing matrix A and the
	//	vectors:
	//
	//		x = <xValue, yValue, zValue, 1.0>
	//		b = <xp * w, yp * w, ?,      w>
	//
	//	where xp and yp are the x and y values of the projected point
	//	calculated above, then Ax = b  (where A is the view matrix) and the
	//  only unknowns are the zValue and w (note the question mark -- this
	//  value is not known and fortunately not needed).  If one solves for w in
	//  the 4th row and substitues in row r, then the zvalue is computed
	//     as follows ...
	double valProj = projectedPoint[maxIndex];
	double value0  = valProj * view.m30 - view.getElement(maxIndex, 0);
	double value1  = valProj * view.m31 - view.getElement(maxIndex, 1);
	double value2  = valProj * view.m32 - view.getElement(maxIndex, 2);
	double value3  = valProj * view.m33 - view.getElement(maxIndex, 3);

	zValue =  - (value0 * xValue + value1 * yValue + value3) / value2;
	return(zValue);
}

/**
 * Given grid data and an (<i>x</i>,<i>y</i>) point on the visible <i>xy</i>
 * plane within the grid, do bilinear interpolation using the 4 closest grid
 * points and return the corresponding <i>z</i> value.  Note that this does
 * error checking and then just calls <code>JChart3dUtil.biLinearInterp();</code>.
 *
 * @param data The internal grid data object.  This can be retrieved from a
 *     <code>Chart3dDataView</code> object via <code>getElevationData()</code>
 *     or <code>getZoneData()</code>.
 * @param x The <i>x</i> data-space value
 * @param y The <i>y</i> data-space value
 * @return The interpolated <i>z</i> value.  Return's the data's hole value if
 *      an error occurs.
 * @see #getElevationData
 * @see #getZoneData
 * @see com.klg.jclass.chart3d.JCChart3dUtil#biLinearInterp
 */
public double gridValue(Chart3dGridData data, double x, double y)
{
	if (data == null || !data.isDataOK()) {
	    throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.BAD_DATA_OR_INDEX_ERROR));
	}

	// Make sure (x, y) lies within the grid
	int numX = data.getNumX();
	int numY = data.getNumY();
	double z = data.getHoleValue();
	if (x < data.getX(0) || x > data.getX(numX - 1) ||
	    y < data.getY(0) || y > data.getY(numY - 1))
	{
		return(z);
	}

	// Find the grid index of the facet in which (x, y) lies.
    int xIndex = data.getPointIndex(x, data.getXGrid(), numX);
    int yIndex = data.getPointIndex(y, data.getYGrid(), numY);
	if (xIndex + 1 == numX) {
		xIndex--;
	}
	if (yIndex + 1 == numY) {
		yIndex--;
	}
	if (!JCChart3dUtil.holeInFacet(data, xIndex, yIndex)) {
		// Do bilinear interpolation on the data.values array
		z = JCChart3dUtil.biLinearInterp(data, xIndex, yIndex, x, y);
	}
	return(z);
}

/**
 * This method sets the radiusOfInfluence parameter used in mouse editing
 * of a SURFACE chart.  Applies only to the Java3d version of Chart3d.
 */
public void setRadiusOfInfluence(double radius) {
	radiusOfInfluence = radius;
}

/**
 * This method returns the value of the radiusOfInfluence parameter used
 * in mouse editing of a SURFACE chart.  Applies only to the Java3d version
 * of Chart3d.
 */
public double getRadiusOfInfluence() {
	return radiusOfInfluence;
}

/**
 * This method sets the mouseEditScaleFactor used in mouse editing of
 * a SURFACE, BAR, or SCATTER chart.  Applies only to the Java3d version
 * of Chart3d.
 */
public void setMouseEditScaleFactor(double factor) {
	mouseEditScaleFactor = factor;
}

/**
 * This method returns the mouseEditScaleFactor used in mouse editing
 * of a SURFACE, BAR, or SCATTER chart.  Applies only to the Java3d version
 * of Chart3d.
 */
public double getMouseEditScaleFactor() {
	return mouseEditScaleFactor;
}

}
