/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Chart3dGridData.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;


import javax.vecmath.Matrix4d;
import java.awt.*;
import java.util.ArrayList;

/**
 * Stores information that represents a grid of points.  The user sets
 * information on this object by setting the grid data model attribute of the
 * parent <code>Chart3dDataView</code> object.  The user can then retrieve
 * information via this object.
 */
public class Chart3dGridData extends Chart3dData {

/** The array of <i>X</i> grid values. */
protected double[]          xGrid = null;

/** The array of <i>Y</i> grid values. */
protected double[]          yGrid = null;

/** The grid of <i>Z</i> values. */
protected double[][]        zValues = null;

/** The number of <i>X</i> grid values used. */
protected int               numX = 0;

/** The number of <i>Y</i> grid values used */
protected int               numY = 0;

/** The <code>chartStyle</code> used if the chart type is a scatter plot. */
protected JCChart3dStyle    chartStyle = null;

/** The <i>X</i> data labels for this <code>gridData</code> object. */
protected String[]          xLabels = null;

/** The <i>Y</i> data labels for this <code>gridData</code> object. */
protected String[]          yLabels = null;

/**
 * The <i>X</i> data labels for this <code>gridData</code> object in an array
 * list.  If this list is non-null, the xLabels property is ignored.
 */
protected ArrayList         xLabelsArrayList = null;

/**
 * The <i>Y</i> data labels for this <code>gridData</code> object in an array
 * list.  If this list is non-null, the yLabels property is ignored.
 */
protected ArrayList         yLabelsArrayList = null;

// Internal variables
protected GridLimits        gridLimits = new GridLimits(this, false);
protected double            minXStep = 0.0;
protected double            minYStep = 0.0;

/**
 * Basic constructor. Data arrays are set to null.
 */
public Chart3dGridData()
{
	super();
	dataHandler = new Chart3dGridDataHandler(this);
	chartStyle  =
        new JCChart3dStyle(new JCLineStyle(1, Color.blue, JCLineStyle.SOLID),
                           new JCSymbolStyle(JCSymbolStyle.getDefaultShape(null),
                                             Color.orange,
                                             JCSymbolStyle.DEFAULT_SIZE));
	processData();
}

/**
 * Internal constructor specifying parent data view and arrays holding data.
 *
 * @param dataView The parent data view
 * @param data The data model from which to extra data
 */
public Chart3dGridData(Chart3dDataView dataView, Chart3dGridDataModel data)
{
	super(dataView);
	dataHandler = new Chart3dGridDataHandler(this);
	chartStyle  =
        new JCChart3dStyle(new JCLineStyle(1, Color.blue, JCLineStyle.SOLID),
                           new JCSymbolStyle(JCSymbolStyle.getDefaultShape(null),
                                             Color.orange,
                                             JCSymbolStyle.DEFAULT_SIZE));
	chartStyle.addParent(dataView);
	setDataSource(data);
}

/**
 * This is the number of <i>X</i> points that is actually used (this may not
 * match <code>xGrid.length</code>)
 *
 * @return The number of <i>X</i> points used
 */
public int getNumX()
{
	return(numX);
}

/**
 * This is the number of <i>Y</i> points that is actually used (this may not
 * match <code>yGrid.length</code>)
 *
 * @return The number of <i>Y</i> points used
 */
public int getNumY()
{
	return(numY);
}

/**
 * Returns the <i>X</i> value at the specified index.
 *
 * @param i The index of the <i>X</i> grid value to be retrieved
 * @return The<i>X</i> value at the specified index.
 */
public double getX(int i)
{
	if (xGrid == null || i >= xGrid.length) {
		return(Double.MAX_VALUE);
	}
	return(xGrid[i]);
}

/**
 * Returns the <i>Y</i> value at the specified index.
 *
 * @param i The index of the <i>Y</i> grid value to be retrieved
 * @return The<i>Y</i> value at the specified index.
 */
public double getY(int i)
{
	if (yGrid == null || i >= yGrid.length) {
		return(Double.MAX_VALUE);
	}
	return(yGrid[i]);
}

/**
 * Returns the <i>X</i> grid array.
 *
 * @return The current <i>X</i> grid array
 */
public double[] getXGrid()
{
	return(xGrid);
}

/**
 * Sets the X grid array.  Internal use only! This method should never be called
 * by any object other than the dataHandler and thus should never be made
 * public. Use the data model mechanism to set the X grid array.
 *
 * @param xGrid The new xGrid array
 */
protected void setXGrid(double[] xGrid)
{
	this.xGrid = xGrid;
	processData();
}

/**
 * Returns the <i>Y</i> grid array.
 *
 * @return The current <i>X</i> grid array
 */
public double[] getYGrid()
{
	return(yGrid);
}

/**
 * Sets the Y grid array.  Internal use only! This method should never be called
 * by any object other than the dataHandler and thus should never be made
 * public.  Use the data model mechanism to set the Y grid array.
 *
 * @param yGrid The new yGrid array
 */
protected void setYGrid(double[] yGrid)
{
	this.yGrid = yGrid;
	processData();
}

/**
 * Returns the data value (or <i>z</i> value) that corresponds to grid point
 * (<i>x</i>, <i>y</i>).
 *
 * @param x The x index of the data value
 * @param y The y index of the data value
 * @return The data value at (<i>x</i>, <i>y</i>)
 */
public double getZValue(int x, int y)
{
	if (zValues == null || zValues[x] == null) {
		return(Double.MAX_VALUE);
	}
	return(zValues[x][y]);
}

/**
 * Returns the entire zValues array.
 *
 * @return The doubly index z values array
 */
public double[][] getZValues()
{
	return(zValues);
}

/**
 * Sets the Z values array.  Internal use only! This method should never be
 * called be any object other than the dataHandler and thus should never be
 * made public.  Use the data model mechanism to set the Z values array.
 *
 * @param zValues The new zValues array
 */
protected void setZValues(double[][] zValues)
{
	this.zValues = zValues;
	processData();
}

/**
 * Returns the <i>X</i> data labels for this <code>Chart3dGridData</code> object
 *
 * @return The current X data labels
 */
public String[] getXLabels()
{
	return(xLabels);
}

/**
 * Sets the <i>X</i> data labels for this <code>Chart3dGridData</code> object.
 *
 * @param xLabels The new X data labels
 */
public void setXLabels(String[] xLabels)
{
	this.xLabels = xLabels;
	setChanged(true, Changeable.RECALC);
}

/**
 * Get the <i>X</i> data labels for this <code>Chart3dGridData</code> object
 * in an array list.  If this property is non-null, the xLabels property is
 * ignored.
 *
 * @return The current X data labels ArrayList
 */
public ArrayList getXLabelsArrayList()
{
    return(xLabelsArrayList);
}

/**
 * Set the <i>X</i> data labels for this <code>Chart3dGridData</code> object
 * as an array list.  If this property is non-null, the xLabels property is
 * ignored.
 *
 * In this array list you can specify xLabels as a list of Strings,
 * JCValueLabels, or other objects (or a mixture of these types).   For each
 * data label, if the object is a JCValueLabel, a copy of it is used internally
 * (the value is ignored).  If it is a String, it is used as is.  If it is any
 * other object, the value of is "toString()" method is used as the label.
 *
 * @param xLabelsArrayList The new X data labels ArrayList
 * @see #setXLabelsArrayList
 */
public void setXLabelsArrayList(ArrayList xLabelsArrayList)
{
    this.xLabelsArrayList = xLabelsArrayList;
	setChanged(true, Changeable.RECALC);
}

/**
 * Returns the <i>Y</i> data labels.
 *
 * @return The current Y data labels for this <code>Chart3dGridData</code>
 */
public String[] getYLabels()
{
	return(yLabels);
}

/**
 * Sets the <i>Y</i> data labels for this <code>Chart3dGridData</code> object.
 *
 * @param yLabels The new Y data labels
 */
public void setYLabels(String[] yLabels)
{
	this.yLabels = yLabels;
	setChanged(true, Changeable.RECALC);
}

/**
 * Get the <i>Y</i> data labels for this <code>Chart3dGridData</code> object
 * in an array list.  If this property is non-null, the yLabels property is
 * ignored.
 *
 * @return The current Y data labels ArrayList
 * @see #setYLabelsArrayList
 */
public ArrayList getYLabelsArrayList()
{
    return(yLabelsArrayList);
}

/**
 * Set the <i>Y</i> data labels for this <code>Chart3dGridData</code> object
 * as an array list.  If this property is non-null, the yLabels property is
 * ignored.
 *
 * In this array list you can specify yLabels as a list of Strings,
 * JCValueLabels, or other objects (or a mixture of these types).   For each
 * data label, if the object is a JCValueLabel, a copy of it is used internally
 * (the value is ignored).  If it is a String, it is used as is.  If it is any
 * other object, the value of is "toString()" method is used as the label.
 *
 * @param yLabelsArrayList The new Y data labels ArrayList
 */
public void setYLabelsArrayList(ArrayList yLabelsArrayList)
{
    this.yLabelsArrayList = yLabelsArrayList;
	setChanged(true, Changeable.RECALC);
}

/**
 * Sets the chart style for this grid data object.  This is used for rendering
 * the line style and symbol style of the grid points when the chart type is
 * <code>JCChart3d.SCATTER</code> and no contours or zones are being drawn.
 *
 * @param chartStyle The new JCChartStyle for this object
 */
public void setChartStyle(JCChart3dStyle chartStyle)
{
	if (this.chartStyle == chartStyle) {
		return;
	}
	this.chartStyle = chartStyle;
	if (this.chartStyle != null) {
	    this.chartStyle.addParent(dataView);
	}
	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the chart style for this grid data object.  This is used for rendering
 * the line style and symbol style of the grid points when the chart type is
 * <code>JCChart3d.SCATTER</code> and no contours or zones are being drawn.
 *
 * @return The chart style for this grid data object
 */
public JCChart3dStyle getChartStyle()
{
	return(chartStyle);
}

/**
 * Sets the X grid, Y grid, and Z values arrays.  Internal use only! This method
 * should never be called be any object other than the dataHandler and thus
 * should never be made public.  Use the data model mechanism to set the Z
 * values array.
 *
 * @param xGrid The new xGrid array
 * @param yGrid The new yGrid array
 * @param zValues The new z values array
 */
protected void setAllXYZValues(double[] xGrid, double[] yGrid, double[][] zValues)
{
	this.xGrid  = xGrid;
	this.yGrid  = yGrid;
	this.zValues = zValues;
	processData();
}

/**
 * Sets the data values from a data source and processes the data.  Internal
 * use only!  Use the com.klg.jclass.chart3d.Chart3dDataView#setElevationDataSource
 * or com.klg.jclass.chart3d.Chart3dDataView#setZoneDataSource to set the data
 * source on this object.  Note that <code>setChanged</code> does not get called
 * because the only place this should get called from is the parent
 * <code>dataView</code> (which will handle <code>setChanged</code>).
 *
 * @param dataSource The new data source from which to extract data
 */
protected void setDataSource(Chart3dGridDataModel dataSource)
{
	this.dataSource = dataSource;
	if (dataSource == null) {
		this.xGrid     = null;
		this.yGrid     = null;
		this.zValues    = null;
		this.holeValue = HoleValueChart3dDataModel.DEFAULT_HOLE_VALUE;
		this.name      = null;
		processData();
		return;
	}
	if (dataHandler == null) {
		
		return;
	}
	dataHandler.reset(dataSource);
    if (dataSource instanceof LabelledChart3dGridDataModel) {
        LabelledChart3dGridDataModel lModel = (LabelledChart3dGridDataModel)dataSource;
        this.xLabels = lModel.getXLabels();
        this.yLabels = lModel.getYLabels();
    }
}

/**
 * Returns the current <code>GridLimits</code> object.  This object stores the
 * <i>x</i> and <i>y</i> start and stop indices that the chart actually uses
 * based on the current <i>x</i> and <i>y</i> axis min and max (if these are
 * within their default values, only part of the grid will be drawn).
 * @return The current <code>GridLimits</code> object
 */
public GridLimits getGridLimits()
{
	return(gridLimits);
}

/**
 * Retrieves the minimum difference between <i>x</i> grid values.
 *
 * @return The minimum <i>x</i> step value
 */
public double getMinXStep()
{
	return(minXStep);
}

/**
 * Retrieves the minimum difference between <i>y</i> grid values.
 *
 * @return The minimum <i>y</i> step value
 */
public double getMinYStep()
{
	return(minYStep);
}

/**
 * Returns the index that contains the <i>x</i> value closest to the specified
 * value.
 *
 * @param x The value for which the closest index should be found
 * @return The closest <i>x</i> index
 */
public int getXClosest(double x)
{
	return(getClosestPoint(x, xGrid, numX));
}

/**
 * Returns the index that contains the <i>y</i> value closest to the specified
 * value.
 *
 * @param y The value for which the closest index should be found
 * @return The closest <i>y</i> index
 */
public int getYClosest(double y)
{
	return(getClosestPoint(y, yGrid, numY));
}

/**
 * Calculates the data array lengths that can be used and value of
 * <code>dataOK</code>.
 */
protected void processData()
{
	numX = numY = 0;
	if (zValues != null) {
		if (xGrid != null) {
			numX = xGrid.length;
		}
	    if (yGrid != null) {
			numY = yGrid.length;
		}
	    numX = Math.min(numX, zValues.length);
	    for (int i = 0; i < numX; i++) {
			numY = Math.min(numY, zValues[i].length);
		}
	}

	dataOK = true;
	// Must have at least one point in both the X and Y direction
	if (numX < 1 || numY < 1) {
		dataOK = false;
	}
	// XGrid values must be strictly increasing
	for (int i = 1; dataOK && i < numX; i++) {
		if (xGrid[i-1] >= xGrid[i]) {
			dataOK = false;
		}
	}
	// YGrid values must be strictly increasing
	for (int i = 1; dataOK && i < numY; i++) {
		if (yGrid[i-1] >= yGrid[i]) {
			dataOK = false;
		}
	}

	minXStep = getMinStep(xGrid, numX);
	minYStep = getMinStep(yGrid, numY);
}

/**
 * Checks to see if the grid from the passed in data matches the grid from
 * this data.  Needed to see if surface and zone data are compatible.
 *
 * @param data The data object whose data needs to match the data of this object
 * @return Whether the passed in data matches the data from this object
 */
protected boolean matchData(Chart3dGridData data)
{
	if (!dataOK || data == null || !data.isDataOK()) {
		return(false);
	}

	if (numX != data.getNumX() || numY != data.getNumY()) {
		return(false);
	}
	for (int i = 0; i < numX; i++) {
		if (xGrid[i] != data.getX(i)) {
			return(false);
		}
	}
	for (int i = 0; i < numY; i++) {
		if (yGrid[i] != data.getY(i)) {
			return(false);
		}
	}
	return(true);
}

/**
 * Computes the <i>Z</i> min and max data values and store these in
 * <code>dataMin</code> and <code>dataMax</code>.
 *
 * @param data If not null, use as a zone data object
 * @param oldMin The old data min (if this one can't be computed for some reason)
 * @param oldMax The old data max (if this one can't be computed for some reason)
 */
protected void computeZLimits(Chart3dGridData data, double oldMin, double oldMax)
{
	if (!dataOK || (data != null && !data.isDataOK())) {
		dataMin = Double.MAX_VALUE;
		dataMax = Double.MAX_VALUE;
		return;
	}

	int iStart = gridLimits.xOffset;
	int iEnd   = gridLimits.xLast;
	int jStart = gridLimits.yOffset;
	int jEnd   = gridLimits.yLast;

	double zMin = Double.MAX_VALUE;
	double zMax = Double.MAX_VALUE;
	for (int i = iStart; i <= iEnd; i++) {
		for (int j = jStart; j <= jEnd; j++) {
			boolean considerValue = (zValues[i][j] != getHoleValue());
			if (considerValue && data != null) {
				considerValue = (data.zValues[i][j] != data.getHoleValue());
			}
			if (considerValue) {
				if (zMin == Double.MAX_VALUE) {
					zMin = zMax = zValues[i][j];
				}
				else if (zValues[i][j] > zMax) {
					zMax = zValues[i][j];
				}
				else if (zValues[i][j] < zMin) {
					zMin = zValues[i][j];
				}
			}
		}
	}

	// Check if data is all holes
	if (zMin == Double.MAX_VALUE) {
		
		if (oldMin == Double.MAX_VALUE) {
			// No previous value -- pick arbitrary range
			zMin = 0.0;
			zMax = 1.0;
		}
		else {
			// Restore previous values
			zMin = oldMin;
			zMax = oldMax;
		}
	}
	dataMin = zMin;
	dataMax = zMax;
}

/**
 * Calculates and sets the axis limits, grid limits, and the contour limits.
 *
 * @param zoneData A zone data object that may have valid data
 */
protected void calcLimits(Chart3dGridData zoneData)
{
	if (!dataOK || zoneData == null) {
		return;
	}

	// Make sure zone data matches this data, if used
	if (zoneData.isDataOK()) {
		if (!matchData(zoneData)) {
			zoneData.setDataOK(false);
		    throw new IllegalArgumentException(
                LocaleBundle.string(
                    LocaleBundle.ELEVATION_AND_ZONE_DATA_MISMATCH_ERROR));
		}
	}

	// Set default X axis min/max values
	JCChart3dArea area = dataView.getChart3d().getChart3dArea();
	JCAxis axis = area.getAxis(JCAxis.AXIS_X);
	axis.min.setValue(xGrid[0]);
	axis.max.setValue(xGrid[numX - 1]);
	
	

	// Set default Y axis min/max values
	axis = area.getAxis(JCAxis.AXIS_Y);
	axis.min.setValue(yGrid[0]);
	axis.max.setValue(yGrid[numY - 1]);
	
	

	// Compute grid limits based on axis min/max values
	gridLimits.assignGridValues(this);
	if (zoneData.isDataOK()) {
		// Make sure zone data has same grid limits as elevation data
		GridLimits zoneGridLimits = zoneData.getGridLimits();
		zoneGridLimits.setXOffset(gridLimits.getXOffset());
		zoneGridLimits.setYOffset(gridLimits.getYOffset());
		zoneGridLimits.setXLast(gridLimits.getXLast());
		zoneGridLimits.setYLast(gridLimits.getYLast());
	}

	// Compute the Z data limits and axis min/max values
	axis = area.getAxis(JCAxis.AXIS_Z);
	computeZLimits(null, axis.getMin(), axis.getMax());
	checkValidZExtents();
	axis.checkLimits(dataMin, dataMax);
	
	

	// Compute the contour limits
	dataView.getContour().getContourLevels().calcLimits(dataView, this, zoneData);
}

/**
 * Precomputes the screen points for each of the grid values
 */
public void precomputePoints()
{
	JCChart3dArea area = dataView.getChart3d().getChart3dArea();

	// Try to use old screenPoints if possible
	boolean allocate = false;
	if (screenPoints == null            ||
	    screenPoints.length != numX     ||
		screenPoints[0].length != numY)
	{
		// Something about the data is different.  Need new screenPoints.
	    screenPoints = new ScreenPoint[numX][numY];
		allocate = true;
	}
	for (int i = gridLimits.getXOffset(); i <= gridLimits.getXLast(); i++) {
		for (int j = gridLimits.getYOffset(); j <= gridLimits.getYLast(); j++) {
			double      x  = xGrid[i];
			double      y  = yGrid[j];
			double      z  = zValues[i][j];
	        ScreenPoint sp = null;
			if (allocate) {
				sp = new ScreenPoint();
			}
			else {
				sp = screenPoints[i][j];
				sp.reset();
			}
			sp.setHoleValue(z == getHoleValue());
			JCChart3dUtil.getScreenPoint(dataView.getChart3d(), x, y, z,
			                             dataMin, sp);
			screenPoints[i][j] = sp;
		}
	}

	area.getSurface().computeMeshFiltering(dataView);
}

/**
 * Makes a first rough pass at the annotation limits, extending the axis limits
 * out to include the origin and making sure that bars fit within the axis.
 */
protected void calcAnnoLimits()
{
	if (!dataOK) {
		return;
	}

	JCChart3dArea   area  = dataView.getChart3d().getChart3dArea();
	JCAxis          xAxis = area.getAxis(JCAxis.AXIS_X);
	JCAxis          yAxis = area.getAxis(JCAxis.AXIS_Y);
	JCAxis          zAxis = area.getAxis(JCAxis.AXIS_Z);

	double xMin = xGrid[gridLimits.getXOffset()];
	double yMin = yGrid[gridLimits.getYOffset()];
	double xMax = xGrid[gridLimits.getXLast()];
	double yMax = yGrid[gridLimits.getYLast()];
	double zMin = zAxis.getMin();
	double zMax = zAxis.getMax();

	if (!area.getChart3d().isProjection()) {
		JCBar bar = area.getBar();
	    if (dataView.getChartType() == JCChart3d.BAR) {
			// Expand the [xy]axis extents 10% to leave a gap around bars
		    double val = (xMax - xMin) * 0.05;
		    if (bar.getXFormat() == JCBar.FIXED) {
			    val += minXStep / 2.0;
			}
		    xMin -= val;
		    xMax += val;

		    val = (yMax - yMin) * 0.05;
		    if (bar.getYFormat() == JCBar.FIXED) {
			    val += minYStep / 2.0;
			}
		    yMin -= val;
		    yMax += val;

		    // Adjust z extents to account for z origin
		    zMin = Math.min(zMin, zAxis.getOrigin());
		    zMax = Math.max(zMax, zAxis.getOrigin());
		}
	    else if (dataView.getChartType() == JCChart3d.SCATTER &&
		         area.getScatter().hasDropLines())
		{
		    // Adjust z extents to account for z origin
		    zMin = Math.min(zMin, zAxis.getOrigin());
		    zMax = Math.max(zMax, zAxis.getOrigin());
		}
	}

	// Let the user force the axis bounds out
	xMin = Math.min(xAxis.getMin(), xMin);
	yMin = Math.min(yAxis.getMin(), yMin);
	xMax = Math.max(xAxis.getMax(), xMax);
	yMax = Math.max(yAxis.getMax(), yMax);

	// Set the anno min and max.  Reset the step to zero (its actual value
	// is calculated later).
	xAxis.setAnnoMin(xMin);
	xAxis.setAnnoMax(xMax);
	xAxis.setAnnoStep(0.0);
	yAxis.setAnnoMin(yMin);
	yAxis.setAnnoMax(yMax);
	yAxis.setAnnoStep(0.0);
	zAxis.setAnnoMin(zMin);
	zAxis.setAnnoMax(zMax);
	zAxis.setAnnoStep(0.0);
}

/**
 * Returns the largest <i>x</i> or <i>y</i> grid index that the value
 * '<i>t</i>' is greater than. This function is used for determining the facet
 * on the grid within which a point is contained.  For fixed bars and scatter
 * plots the closest grid line index is returned.
 *
 * @param t The value for which the index needs to be found
 * @param isXArray Is this an <i>x</i> array or <i>y</i> array
 * @param transform Should the point be transformed before searching
 * @return The largest <i>x</i> or <i>y</i> grid number that the value
 *     '<i>t</i>' is greater than.
 */
public int getGridIndex(double t, boolean isXArray, boolean transform)
{
	int             chartType = dataView.getChartType();
	int             n         = isXArray ? numX : numY;
	double          grid[]    = isXArray ? xGrid : yGrid;
	JCChart3dArea   area      = dataView.getChart3d().getChart3dArea();
	JCBar           bar       = area.getBar();

	if (transform) {
		Matrix4d absolute = area.getTransform().getAbsolute();
		if (isXArray) {
		   t = (t - absolute.m03) / absolute.m00;
		}
		else {
		   t = (t - absolute.m13) / absolute.m11;
		}
	}
	int index = getPointIndex(t, grid, n);

	int offset    = isXArray ? gridLimits.getXOffset() : gridLimits.getYOffset();
	int last      = isXArray ? gridLimits.getXLast() : gridLimits.getYLast();
	int barFormat = isXArray ? bar.getXFormat() : bar.getYFormat();
	if (chartType == JCChart3d.SCATTER ||
		(chartType == JCChart3d.BAR && barFormat == JCBar.FIXED))
	{
		if (index > n - 2) {
			index = n - 1;
		}
		else if (t + t > (grid[index] + grid[index+1])) {
			index = index + 1;	/* its closer to index + 1 */
		}
		index = JCChart3dUtil.clamp(index, offset, last);
	}
	else {
		if (!transform && chartType != JCChart3d.BAR) {
			index = JCChart3dUtil.clamp(index, offset, last);
		}
		else {
			index = JCChart3dUtil.clamp(index, offset, last - 1);
		}
	}
	return(Math.max(0, index));
}

}


