/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Chart3dGridDataHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.event.*;
import com.klg.jclass.chart3d.resources.LocaleBundle;

/**
 * Data interpretation class (this class is used internally
 * by JClass Chart 3D and should not be used by applications).
 * Ambitious users could attempt to create and plug
 * in their own data handlers for specific handling,
 * but this is not common use.
 */
public class Chart3dGridDataHandler extends Chart3dDataHandler {

/**
 * Constructor that sets the containing grid data object for this handler
 *
 * @param data The containing grid data object
 */
public Chart3dGridDataHandler(Chart3dGridData data)
{
    super(data);
}

/**
 * Given an input source and an data event, updates the parent internal data
 * object appropriately.
 *
 * @param source The data source
 * @param event The event to be handled
 */
public boolean handleEvent(Chart3dDataModel source, Chart3dDataEvent event)
{
    boolean changed = false;
    if (source == null || event == null) {
        return(changed);
    }

	Chart3dGridDataModel gridSource = (Chart3dGridDataModel)source;
	Chart3dGridDataEvent gridEvent = null;
	if (event instanceof Chart3dGridDataEvent) {
        gridEvent = (Chart3dGridDataEvent)event;
	}
	if (gridEvent == null && event.getType() > Chart3dDataEvent.RELOAD_HOLE_VALUE) {
		return(changed);
	}
    switch(gridEvent.getType()) {
	case Chart3dDataEvent.RELOAD_DATA_SOURCE_NAME:
		changed = reloadDataSourceName(source);
		break;
	case Chart3dDataEvent.RESET:
		changed = reset(source);
		break;
	case Chart3dDataEvent.RELOAD:
		changed = reload(source);
		break;
	case Chart3dDataEvent.RELOAD_HOLE_VALUE:
		changed = reloadHoleValue(source);
		break;
	case Chart3dGridDataEvent.RELOAD_XVALUE:
	    changed = reloadXValue(gridSource, gridEvent.getX());
		break;
	case Chart3dGridDataEvent.RELOAD_XGRID:
	    changed = reloadXGrid(gridSource);
		break;
	case Chart3dGridDataEvent.RELOAD_YVALUE:
	    changed = reloadYValue(gridSource, gridEvent.getY());
		break;
	case Chart3dGridDataEvent.RELOAD_YGRID:
	    changed = reloadYGrid(gridSource);
		break;
	case Chart3dGridDataEvent.RELOAD_ZVALUE:
	    changed = reloadZValue(gridSource, gridEvent.getX(), gridEvent.getY());
		break;
	case Chart3dGridDataEvent.RELOAD_ZARRAY:
	    changed = reloadZArray(gridSource, gridEvent.getX());
		break;
	case Chart3dGridDataEvent.RELOAD_ZALL:
	    changed = reloadZAll(gridSource);
		break;
	case Chart3dGridDataEvent.RELOAD_XLABEL:
	    changed = reloadXLabel(gridSource, gridEvent.getX());
		break;
	case Chart3dGridDataEvent.RELOAD_ALL_XLABELS:
	    changed = reloadAllXLabels(gridSource);
		break;
	case Chart3dGridDataEvent.RELOAD_YLABEL:
	    changed = reloadXLabel(gridSource, gridEvent.getY());
		break;
	case Chart3dGridDataEvent.RELOAD_ALL_YLABELS:
	    changed = reloadAllYLabels(gridSource);
		break;
	default:
        throw new RuntimeException(
            LocaleBundle.string(LocaleBundle.UNKNOWN_EVENT_ERROR) +
            gridEvent.getType());
	}
    return(changed);
}

/**
 * Implements <code>Chart3dDataEvent.RESET</code>.
 *
 * @param source The data source from which the data is to be reset
 * @see Chart3dDataEvent
 */
public boolean reset(Chart3dDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData      gridData   = (Chart3dGridData)data;
	Chart3dGridDataModel gridSource = (Chart3dGridDataModel)source;

	gridData.setAllXYZValues(gridSource.getXGrid(), gridSource.getYGrid(),
	                         gridSource.getZValues());
	reloadAllXLabels(gridSource);
	reloadAllYLabels(gridSource);
	reloadHoleValue(source);
	reloadDataSourceName(source);
    return(true);
}

/**
 * Implements <code>Chart3dDataEvent.RELOAD</code>.
 *
 * @param source The data source from which the data is to be reloaded
 * @see Chart3dDataEvent
 */
public boolean reload(Chart3dDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData      gridData   = (Chart3dGridData)data;
	Chart3dGridDataModel gridSource = (Chart3dGridDataModel)source;
	gridData.setAllXYZValues(gridSource.getXGrid(), gridSource.getYGrid(),
	                         gridSource.getZValues());
	reloadHoleValue(source);
    return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_XVALUE</code>.
 *
 * @param source The data source from which the x value is to be reloaded
 * @param xIndex The index of the x value
 * @see Chart3dGridDataEvent
 */
public boolean reloadXValue(Chart3dGridDataModel source, int xIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	double[] xGrid = gridData.getXGrid();
	if (xIndex < 0 || xGrid == null || xIndex >= xGrid.length) {
        return(false);
	}
	// Reset X grid in case the user duplicated the array
	gridData.setXGrid(source.getXGrid());
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_XGRID</code>.
 *
 * @param source The data source from which the x grid is to be reloaded
 * @see Chart3dGridDataEvent
 */
public boolean reloadXGrid(Chart3dGridDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	gridData.setXGrid(source.getXGrid());
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_YVALUE</code>.
 *
 * @param source The data source from which the y value is to be reloaded
 * @param yIndex The index of the y value
 * @see Chart3dGridDataEvent
 */
public boolean reloadYValue(Chart3dGridDataModel source, int yIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	double[] yGrid = gridData.getYGrid();
	if (yIndex < 0 || yGrid == null || yIndex >= yGrid.length) {
        return(false);
	}
	// Reset Y grid in case the user duplicated the array
	gridData.setYGrid(source.getYGrid());
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_YGRID</code>.
 *
 * @param source The data source from which the y grid is to be reloaded
 * @see Chart3dGridDataEvent
 */
public boolean reloadYGrid(Chart3dGridDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	gridData.setYGrid(source.getYGrid());
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_ZVALUE</code>.
 *
 * @param source The data source from which the z value is to be reloaded
 * @param xIndex The x index of the z value
 * @param yIndex The y index of the z value
 * @see Chart3dGridDataEvent
 */
public boolean reloadZValue(Chart3dGridDataModel source, int xIndex, int yIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	double[][] values = gridData.getZValues();
	if (values == null || xIndex < 0 || xIndex >= values.length) {
		return(false);
	}
	if (values[xIndex] == null || yIndex < 0 ||
        yIndex > values[xIndex].length)
    {
		return(false);
	}
	// Reset Z values in case user duplicated array
	gridData.setZValues(source.getZValues());
	return(true);
}

/**
 * Implements <code>Chart3dDataEvent.RELOAD_ZARRAY
 * </code>.
 *
 * @param source The data source from which the z array is to be reloaded
 * @param xIndex The index of the z array
 * @see Chart3dGridDataEvent
 */
public boolean reloadZArray(Chart3dGridDataModel source, int xIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	double[][] values = gridData.getZValues();
	if (values == null || xIndex < 0 || xIndex >= values.length) {
		return(false);
	}
	// Reset Z values in case user duplicated array
	gridData.setZValues(source.getZValues());
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_ZALL</code>.
 *
 * @param source The data source from which all the z data is to be reloaded
 * @see Chart3dGridDataEvent
 */
public boolean reloadZAll(Chart3dGridDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	gridData.setZValues(source.getZValues());
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_XLABEL</code>.
 *
 * @param source The data source from which the x label is to be reloaded
 * @param xIndex The index of the x label
 * @see Chart3dGridDataEvent
 */
public boolean reloadXLabel(Chart3dGridDataModel source, int xIndex)
{
	if (source == null || !(source instanceof LabelledChart3dGridDataModel)) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	LabelledChart3dGridDataModel lSource = (LabelledChart3dGridDataModel)source;

	String[] xLabels = gridData.getXLabels();
	if (xLabels == null || xIndex < 0 || xIndex >= xLabels.length) {
		return(false);
	}
	// Reset the XLabels (user could have made a duplicate array)
	gridData.xLabels = lSource.getXLabels();
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_ALL_XLABELS</code>.
 *
 * @param source The data source from which the x labels are to be reloaded
 * @see Chart3dGridDataEvent
 */
public boolean reloadAllXLabels(Chart3dGridDataModel source)
{
	if (source == null || !(source instanceof LabelledChart3dGridDataModel)) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	LabelledChart3dGridDataModel lSource = (LabelledChart3dGridDataModel)source;
	gridData.xLabels = lSource.getXLabels();
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_YLABEL
 * </code>.
 *
 * @param source The data source from which the y label is to be reloaded
 * @param yIndex The y index of the label
 * @see Chart3dGridDataEvent
 */
public boolean reloadYLabel(Chart3dGridDataModel source, int yIndex)
{
	if (source == null || !(source instanceof LabelledChart3dGridDataModel)) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	LabelledChart3dGridDataModel lSource = (LabelledChart3dGridDataModel)source;

	String[] yLabels = gridData.getYLabels();
	if (yLabels == null || yIndex < 0 || yIndex >= yLabels.length) {
		return(false);
	}
	// Reset the YLabels (user could have made a duplicate array)
	gridData.yLabels = lSource.getYLabels();
	return(true);
}

/**
 * Implements <code>Chart3dGridDataEvent.RELOAD_ALL_YLABELS</code>.
 *
 * @param source The data source from which the y labels are to be reloaded
 * @see Chart3dGridDataEvent
 */
public boolean reloadAllYLabels(Chart3dGridDataModel source)
{
	if (source == null || !(source instanceof LabelledChart3dGridDataModel)) {
        return(false);
	}
	Chart3dGridData gridData = (Chart3dGridData)data;
	LabelledChart3dGridDataModel lSource = (LabelledChart3dGridDataModel)source;
	gridData.yLabels = lSource.getYLabels();
	return(true);
}

/**
 * Modifies z value of data point in an <code>EditableChart3dDataModel</code>
 * instance.
 *
 * @param index A JCData3dGridIndex which indcates the index of the point
 *    being edited.
 * @param value The new z value
 * @return Whether the edit succeeded
 */
public boolean editZValue(JCData3dIndex index, double value)
{
    if (index == null || !(index instanceof JCData3dGridIndex)) {
        return(false);
    }
    JCData3dGridIndex gridIndex = (JCData3dGridIndex)index;
	Chart3dDataModel tmpSrc = data.getDataSource();
	if (tmpSrc == null || !(tmpSrc instanceof EditableChart3dDataModel)) {
		return false;
	}

	EditableChart3dDataModel editSrc = (EditableChart3dDataModel)tmpSrc;
	return(editSrc.setZValue(gridIndex, value));
}

}
