/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Chart3dPointData.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;


import javax.vecmath.Point3d;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

/**
 * Stores information that represents a list of series of points.  The user sets
 * information on this object by setting the point data model attribute of the
 * parent <code>Chart3dDataView</code> object.  The user can then retrieve
 * information via this object.
 */
public class Chart3dPointData extends Chart3dData implements Comparator {

/** The list of series for this point data object. */
protected ArrayList              series = null;

/** The number of series. */
protected int                    numSeries;

/** A sorted list of all the points -- sorted by distance to eye. */
protected ArrayList              sortPoints = null;

/** Has the sorted list been constructed yet? */
protected boolean                sorted = false;

/**
 * Basic constructor. Parent object and data arrays are set to null.
 */
public Chart3dPointData()
{
	super();
	dataHandler = new Chart3dPointDataHandler(this);
	setDataSource(null);
}

/**
 * Internal constructor specifying parent object and point series list.
 *
 * @param dataView The parent data view
 * @param data The data source from which to extract data
 */
public Chart3dPointData(Chart3dDataView dataView, Chart3dPointDataModel data)
{
	super(dataView);
	dataHandler = new Chart3dPointDataHandler(this);
	setDataSource(data);
}

/**
 * Returns the point indexed by <code>pointNum</code> in the series indexed by
 * <code>seriesNum</code>.
 *
 * @param seriesNum The series index of the point wanted
 * @param pointNum  The point index of the point wanted
 * @return The point indexed by series and point.
 */
public Point3d getPoint(int seriesNum, int pointNum)
{
	if (seriesNum < 0 || seriesNum > numSeries - 1) {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.SERIES_OUT_OF_RANGE_ERROR));
	}
	Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(seriesNum);
	return(pointSeries.getPoint(pointNum));
}

/**
 * Returns the <code>Chart3dPointSeries</code> object indexed by
 * <code>seriesNum</code>.
 *
 * @param seriesNum The series index of the Chart3dPointSeries wanted.
 * @return The <code>Chart3dPointSeries</code> object indexed by series.
 */
public Chart3dPointSeries getSeries(int seriesNum)
{
	if (seriesNum < 0 || seriesNum > numSeries - 1) {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.SERIES_OUT_OF_RANGE_ERROR));
	}
	Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(seriesNum);
	return(pointSeries);
}

/**
 * Returns the series array list.
 *
 * @return The <code>ArrayList</code> that holds the list of
 *     <code>Chart3dPointSeries</code>.
 */
public ArrayList getSeries()
{
	return(series);
}

/**
 * Gets the current number of series.  This basically the same as the size of
 * the series ArrayList.
 *
 * @return The current number of series.
 */
public int getNumSeries()
{
	return(numSeries);
}

/**
 * Appends a new series with the given <code>pointList</code> to the end of
 * the series list.
 *
 * @param pointList The point list for the series to be added
 */
protected void addSeries(Point3d[] pointList)
{
	if (pointList == null) {
		return;
	}
	if (series == null) {
		series = new ArrayList();
	}
	series.add(new Chart3dPointSeries(this, pointList));
	numSeries = series.size();
	processData();
}

/**
 * Adds a series at the given index from to the list of series.
 *
 * @param pointList The point list for the series to be added
 * @param index The index at which the new series is to be inserted
 */
protected void addSeries(Point3d[] pointList, int index)
{
	if (series == null) {
		series = new ArrayList();
	}
	if (index < 0 || index > series.size() || pointList == null) {
		return;
	}
	series.add(index, new Chart3dPointSeries(this, pointList));
	numSeries = series.size();
	processData();
}

/**
 * Removes a series at the given index from the list of series.
 *
 * @param index The index of the series to be removed.
 */
protected void removeSeries(int index)
{
	if (series == null || index < 0 || index > series.size()) {
		return;
	}
	series.remove(index);
	numSeries = series.size();
	processData();
}

/**
 * Makes a list of <code>Chart3dPointSeries</code> based on a doubly linked
 * list of <code>Point3d</code> points.  Internal use only!  Use a point
 * data source to set data on this object.
 *
 * @param seriesList The list of points to make a new series list from.
 */
protected void makePointSeriesList(Point3d[][] seriesList)
{
	series    = null;
	numSeries = 0;
	if (seriesList != null) {
		numSeries = seriesList.length;
		series = new ArrayList();
		for (int i = 0; i < numSeries; i++) {
			series.add(new Chart3dPointSeries(this, seriesList[i]));
		}
	}
	processData();
}

/**
 * Refreshes the list of <code>Chart3dPointSeries</code> with possibly new data.
 * This means that the number of series remains the same (so no new
 * <code>Chart3dPointSeries</code> objects need to be created).  Internal use
 * only!  Use a point data source to set data on this object.
 *
 * @param seriesList The list of points from which to refresh the series list
 */
protected void refreshPointSeriesList(Point3d[][] seriesList)
{
	if (seriesList == null ||
	    series == null     ||
		seriesList.length != numSeries)
	{
		return;
	}
	for (int i = 0; i < numSeries; i++) {
	    Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(i);
		Point3d[] points = seriesList[i];
		pointSeries.setPoints(points);
	}
	processData();
}

/**
 * Calculates the data array lengths that can be used and value of
 * <code>dataOK</code>.
 */
protected void processData()
{
	dataOK = true;
	// Must have at least one series
	if (numSeries < 1) {
		dataOK = false;
	}
	// Must be at least one point in each series
	for (int i = 0; i < numSeries; i++) {
	    Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(i);
		if (pointSeries == null || pointSeries.getNumPoints() < 1) {
			dataOK = false;
		}
	}
}

/**
 * Sets a new series list of points on this point data object.  Internal use
 * only!  Use the com.klg.jclass.chart3d.Chart3dDataView#setElevationDataSource
 * to set the data source on this object.
 *
 * @param dataSource The data source from which to extract data
 */
protected void setDataSource(Chart3dPointDataModel dataSource)
{
	this.dataSource = dataSource;
    if (dataSource == null) {
		makePointSeriesList(null);
        return;
    }
	if (dataHandler == null) {
		
		return;
	}
	dataHandler.reset(dataSource);
}

/**
 * Gets the array of sorted points.
 */
public ArrayList getSortPoints()
{
	return(sortPoints);
}

/**
 * Computes the min and max values of all points in the <i>x</i>, <i>y</i>,
 * and <i>z</i> direction and uses these values to set the axis limits.
 */
protected void computePointLimits()
{
	Extents extents = new Extents();
	for (int i = 0; i < numSeries; i++) {
	    Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(i);
		for (int j = 0; j < pointSeries.getNumPoints(); j++) {
			Point3d point = pointSeries.getPoint(j);

			if (point.z == getHoleValue()) {
				continue;
			}
			extents.checkXYZExtents(point.x, point.y, point.z);
		}
	}

	// Check if all holes
	if (extents.getXMin() == Double.MAX_VALUE) {
		
		extents.setXMin(0.0);
		extents.setXMax(1.0);
		extents.setYMin(0.0);
		extents.setYMax(1.0);
		extents.setZMin(0.0);
		extents.setZMax(1.0);
	}

	JCChart3dArea area = dataView.getChart3d().getChart3dArea();
	JCAxis axis = area.getAxis(JCAxis.AXIS_X);
	axis.setExtents(extents.getXMin(), extents.getXMax());

	axis = area.getAxis(JCAxis.AXIS_Y);
	axis.setExtents(extents.getYMin(), extents.getYMax());

	axis = area.getAxis(JCAxis.AXIS_Z);
	axis.setExtents(extents.getZMin(), extents.getZMax());
	dataMin = extents.getZMin();
	dataMax = extents.getZMax();
}

/**
 * Calculates the data limits and checks for valid <i>Z</i> extents (the user
 * can only change the <i>z</i> limits so that they are outside of the data limits).
 */
protected void calcLimits()
{
	if (!dataOK) {
		return;
	}

	computePointLimits();
	checkValidZExtents();
}

/**
 * Makes a first rough pass at the annotation limits, extending the axis limits
 * out to include the origin.  Resets the step to zero (its calculated later).
 */
protected void calcAnnoLimits()
{
	JCChart3dArea   area  = dataView.getChart3d().getChart3dArea();
	JCAxis          xAxis = area.getAxis(JCAxis.AXIS_X);
	JCAxis          yAxis = area.getAxis(JCAxis.AXIS_Y);
	JCAxis          zAxis = area.getAxis(JCAxis.AXIS_Z);

	xAxis.setAnnoMin(xAxis.getMin());
	xAxis.setAnnoMax(xAxis.getMax());
	xAxis.setAnnoStep(0.0);
	yAxis.setAnnoMin(yAxis.getMin());
	yAxis.setAnnoMax(yAxis.getMax());
	yAxis.setAnnoStep(0.0);

	double zMin = zAxis.getMin();
	double zMax = zAxis.getMax();
	if (!area.getChart3d().isProjection() && area.getScatter().hasDropLines()) {
		// Adjust z extents to account for z origin
		zMin = Math.min(zMin, zAxis.getOrigin());
		zMax = Math.max(zMax, zAxis.getOrigin());
	}
	zAxis.setAnnoMin(zMin);
	zAxis.setAnnoMax(zMax);
	zAxis.setAnnoStep(0.0);
}

/**
 * Precomputes the screen points for each of the data points.
 */
public void precomputePoints()
{
	JCChart3dArea area = dataView.getChart3d().getChart3dArea();

	// Try to use old screenPoints if possible
	boolean allocate = false;
	if (screenPoints == null             ||
	    screenPoints.length != numSeries)
	{
		// Something about the data is different.  Need new screenPoints.
	    screenPoints = new ScreenPoint[numSeries][];
		allocate = true;
	}
	for (int i = 0; i < numSeries; i++) {
	    Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(i);
		int numPoints = pointSeries.getNumPoints();

		// Again, try to use old screenPoints for this series
		boolean allocateSeries = false;
		if (allocate || screenPoints[i].length != numPoints) {
		    screenPoints[i] = new ScreenPoint[numPoints];
			allocateSeries = true;
		}
		for (int j = 0; j < numPoints; j++) {
			Point3d point = pointSeries.getPoint(j);
			ScreenPoint sp = null;
			if (allocateSeries) {
			   sp = new ScreenPoint();
			}
			else {
				sp = screenPoints[i][j];
				sp.reset();
			}
			if (point.x == holeValue ||
			    point.y == holeValue ||
				point.z == holeValue)
			{
				sp.setHoleValue(true);
			}

	        JCAxis xAxis = area.getAxis(JCAxis.AXIS_X);
	        JCAxis yAxis = area.getAxis(JCAxis.AXIS_Y);
	        JCAxis zAxis = area.getAxis(JCAxis.AXIS_Z);

			if (point.x > xAxis.getMax() ||
			    point.x < xAxis.getMin() ||
				point.y > yAxis.getMax() ||
				point.y < yAxis.getMin() ||
				point.z > zAxis.getMax() ||
				point.z < zAxis.getMin())
			{
				sp.setHidden(true);
			}
			JCChart3dUtil.getScreenPoint(dataView.getChart3d(),
										 point.x, point.y, point.z,
			                             zAxis.getMin(), sp);
			screenPoints[i][j] = sp;
		}
	}
	sorted = false;
}

/**
 * Creates a <code>ArrayList</code> of <code>SortPoints</code> and then sorts
 * them based on the distance to the eye.  This is used to later draw the
 * points from back to front.
 */
public void sortPoints()
{
	// Already done?
	if (sorted) {
		return;
	}
	if (!isDataOK()) {
		return;
	}

	if (sortPoints == null) {
		sortPoints = new ArrayList();
	}
	else {
		sortPoints.clear();
	}

	for (int i = 0; i < numSeries; i++) {
	    Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(i);
		for (int j = 0; j < pointSeries.getNumPoints(); j++) {
			Point3d point = pointSeries.getPoint(j);
			double distanceToEye =
			    JCChart3dUtil.distanceToEye(dataView.getChart3d(),
											point.x, point.y, point.z);
			SortPoint sortPoint = new SortPoint(i, j, distanceToEye);
			sortPoints.add(sortPoint);
		}
	}
	Collections.sort(sortPoints, this);
	sorted = true;
}

/**
 * <code>SortPoint</code> sorting routine.
 *
 * @param o1 The first SortPoint to compare
 * @param o2 The second SortPoint to compare
 */
public int compare(Object o1, Object o2)
{
	SortPoint sortPoint1 = (SortPoint)o1;
	SortPoint sortPoint2 = (SortPoint)o2;

	int value = 0;
	if (sortPoint1.getDistanceToEye() > sortPoint2.getDistanceToEye()) {
		value = -1;
	}
	else if (sortPoint1.getDistanceToEye() < sortPoint2.getDistanceToEye()) {
		value = 1;
	}
	return(value);
}

}
