/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Chart3dPointDataHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.util.ArrayList;
import javax.vecmath.Point3d;

import com.klg.jclass.chart3d.event.*;
import com.klg.jclass.chart3d.resources.LocaleBundle;

/**
 * Data interpretation class (this class is used internally
 * by JClass Chart and should not be used by applications).
 * Ambitious users could attempt to create and plug
 * in their own data handlers for specific handling,
 * but this is not common use.
 */
public class Chart3dPointDataHandler extends Chart3dDataHandler {

/**
 * Constructor that sets the containing point data object for this handler
 *
 * @param data The containing point data object
 */
public Chart3dPointDataHandler(Chart3dPointData data)
{
    super(data);
}

/**
 * Given an input source and an data event, updates the parent internal data
 * object appropriately.
 *
 * @param source The data source from which the update happens
 * @param event The event to handle
 */
public boolean handleEvent(Chart3dDataModel source, Chart3dDataEvent event)
{
    boolean changed = false;
    if (source == null || event == null) {
        return(changed);
    }

	Chart3dPointDataModel pointSource = (Chart3dPointDataModel)source;
	Chart3dPointDataEvent pointEvent = null;
	if (event instanceof Chart3dPointDataEvent) {
        pointEvent = (Chart3dPointDataEvent)event;
	}
	if (pointEvent == null &&
        event.getType() > Chart3dDataEvent.RELOAD_HOLE_VALUE)
    {
		return(changed);
	}
    switch(pointEvent.getType()) {
	case Chart3dDataEvent.RELOAD_DATA_SOURCE_NAME:
		changed = reloadDataSourceName(source);
		break;
	case Chart3dDataEvent.RESET:
		changed = reset(source);
		break;
	case Chart3dDataEvent.RELOAD:
		changed = reload(source);
		break;
	case Chart3dDataEvent.RELOAD_HOLE_VALUE:
		changed = reloadHoleValue(source);
		break;
	case Chart3dPointDataEvent.RELOAD_POINT:
	    changed = reloadPoint(pointSource, pointEvent.getSeries(),
							  pointEvent.getPoint());
		break;
	case Chart3dPointDataEvent.RELOAD_SERIES:
	    changed = reloadSeries(pointSource, pointEvent.getSeries());
		break;
	case Chart3dPointDataEvent.ADD_SERIES:
	    changed = addSeries(pointSource);
		break;
	case Chart3dPointDataEvent.INSERT_SERIES:
	    changed = insertSeries(pointSource, pointEvent.getSeries());
		break;
	case Chart3dPointDataEvent.REMOVE_SERIES:
	    changed = removeSeries(pointSource, pointEvent.getSeries());
		break;
	case Chart3dPointDataEvent.RELOAD_SERIES_LABEL:
	    changed = reloadSeriesLabel(pointSource, pointEvent.getSeries());
		break;
	case Chart3dPointDataEvent.RELOAD_ALL_SERIES_LABELS:
	    changed = reloadAllSeriesLabels(pointSource);
		break;
	default:
        throw new RuntimeException(
            LocaleBundle.string(LocaleBundle.UNKNOWN_EVENT_ERROR) +
            pointEvent.getType());
	}
    return(changed);
}

/**
 * Implements <code>Chart3dDataEvent.RESET</code>.
 *
 * @param source  The data source from which to reset
 * @see Chart3dDataEvent
 */
public boolean reset(Chart3dDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData      pointData   = (Chart3dPointData)data;
	Chart3dPointDataModel pointSource = (Chart3dPointDataModel)source;

	pointData.makePointSeriesList(pointSource.getPoints());
	reloadAllSeriesLabels(pointSource);
	reloadHoleValue(source);
	reloadDataSourceName(source);
    return(true);
}

/**
 * Implements <code>Chart3dDataEvent.RELOAD</code>.
 *
 * @param source  The data source from which to reload
 * @see Chart3dDataEvent
 */
public boolean reload(Chart3dDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData      pointData   = (Chart3dPointData)data;
	Chart3dPointDataModel pointSource = (Chart3dPointDataModel)source;

	pointData.refreshPointSeriesList(pointSource.getPoints());
	reloadAllSeriesLabels(pointSource);
	reloadHoleValue(source);
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.RELOAD_POINT</code>.
 *
 * @param source  The data source from which a point is reloaded
 * @param seriesIndex The series index of the point
 * @param pointIndex The point index of the point
 * @see Chart3dPointDataEvent
 */
public boolean reloadPoint(Chart3dPointDataModel source, int seriesIndex,
			               int pointIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
	if (seriesIndex < 0 || seriesIndex >= pointData.getNumSeries()) {
		return(false);
	}
	ArrayList series = pointData.getSeries();
	Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(seriesIndex);
	if (pointIndex < 0 || pointIndex >= pointSeries.getNumPoints()) {
		return(false);
	}
	Point3d[][] points = source.getPoints();
	if (points == null || seriesIndex >= points.length) {
		return(false);
	}
	// Reset the point array just in case the user allocated a new array
	pointSeries.setPoints(points[seriesIndex]);
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.RELOAD_SERIES</code>.
 *
 * @param source  The data source from which the series is reloaded
 * @param seriesIndex The index of the series
 * @see Chart3dPointDataEvent
 */
public boolean reloadSeries(Chart3dPointDataModel source, int seriesIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
	if (seriesIndex < 0 || seriesIndex >= pointData.getNumSeries()) {
		return(false);
	}
	ArrayList series = pointData.getSeries();
	Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(seriesIndex);
	Point3d[][] points = source.getPoints();
	if (points == null || seriesIndex >= points.length) {
		return(false);
	}
	pointSeries.setPoints(points[seriesIndex]);
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.ADD_SERIES</code>.
 *
 * @param source  The data source from which the addedseries is taken
 * @see Chart3dPointDataEvent
 */
public boolean addSeries(Chart3dPointDataModel source)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
	int seriesIndex = pointData.getNumSeries();
	Point3d[][] points = source.getPoints();
	if (points == null || seriesIndex >= points.length) {
		return(false);
	}
	pointData.addSeries(points[seriesIndex]);
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.INSERT_SERIES</code>.
 *
 * @param source  The data source from which the inserted series is taken
 * @param seriesIndex The index of the series to be inserted
 * @see Chart3dPointDataEvent
 */
public boolean insertSeries(Chart3dPointDataModel source, int seriesIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
	if (seriesIndex < 0 || seriesIndex >= pointData.getNumSeries()) {
		return(false);
	}
	Point3d[][] points = source.getPoints();
	if (points == null || seriesIndex >= points.length) {
		return(false);
	}
	pointData.addSeries(points[seriesIndex], seriesIndex);
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.REMOVE_SERIES</code>.
 *
 * @param source The data source
 * @param seriesIndex The index of the series to be removed
 * @see Chart3dPointDataEvent
 */
public boolean removeSeries(Chart3dPointDataModel source, int seriesIndex)
{
	if (source == null) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
	if (seriesIndex < 0 || seriesIndex >= pointData.getNumSeries()) {
		return(false);
	}
	pointData.removeSeries(seriesIndex);
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.RELOAD_SERIES_LABEL</code>.
 *
 * @param source  The data source from which the series label is reloaded
 * @param seriesIndex The index of the series
 * @see Chart3dPointDataEvent
 */
public boolean reloadSeriesLabel(Chart3dPointDataModel source, int seriesIndex)
{
	if (source == null || !(source instanceof LabelledChart3dPointDataModel)) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
    LabelledChart3dPointDataModel lModel =
        (LabelledChart3dPointDataModel)source;

	String[] labels = lModel.getSeriesLabels();
	if (labels == null                          ||
	    seriesIndex < 0                         ||
		seriesIndex >= pointData.getNumSeries() ||
		seriesIndex >= labels.length)
	{
		return(false);
	}
	ArrayList series = pointData.getSeries();
	if (series == null) {
		return(false);
	}
    Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(seriesIndex);
    pointSeries.label = labels[seriesIndex];
    return(true);
}

/**
 * Implements <code>Chart3dPointDataEvent.RELOAD_ALL_SERIES_LABELS</code>.
 *
 * @param source The data source from which the series labels are reloaded
 * @see Chart3dPointDataEvent
 */
public boolean reloadAllSeriesLabels(Chart3dPointDataModel source)
{
	if (source == null || !(source instanceof LabelledChart3dPointDataModel)) {
        return(false);
	}
	Chart3dPointData pointData = (Chart3dPointData)data;
    LabelledChart3dPointDataModel lModel =
        (LabelledChart3dPointDataModel)source;

	String[] labels = lModel.getSeriesLabels();
	if (labels == null || pointData.getNumSeries() == 0) {
		return(false);
	}
	ArrayList series = pointData.getSeries();
	if (series == null) {
		return(false);
	}
    for (int i = 0; i < pointData.getNumSeries() && i < labels.length; i++) {
        Chart3dPointSeries pointSeries = (Chart3dPointSeries)series.get(i);
        pointSeries.label = labels[i];
    }
    return(true);
}

/**
 * Modifies z value of data point in an <code>EditableChart3dDataModel</code>
 * instance.
 *
 * @param index A JCData3dPointIndex which indcates the index of the point
 *    being edited.
 * @param value The new z value
 * @return Whether the edit succeeded
 */
public boolean editZValue(JCData3dIndex index, double value)
{
    if (index == null || !(index instanceof JCData3dPointIndex)) {
        return(false);
    }
    JCData3dPointIndex pointIndex = (JCData3dPointIndex)index;
	Chart3dDataModel tmpSrc = data.getDataSource();
	if (tmpSrc == null || !(tmpSrc instanceof EditableChart3dDataModel)) {
		return(false);
	}

	EditableChart3dDataModel editSrc = (EditableChart3dDataModel)tmpSrc;
	return(editSrc.setZValue(pointIndex, value));
}

}
