/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ContourCell.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:56 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import javax.vecmath.Point3d;



/** 
 * Data structure internal to JClass Chart 3D.  Stores contour information on a per
 * cell basis.
 * 
 */
public class ContourCell implements Comparator, java.io.Serializable {

/**  * 
 */
public final static int SIDE = 4;

/** Does this cell correspond to a hole?
 * 
 */
protected boolean         hole = false;

/** The number of contour intersections for this cell.
 * 
 */
protected int             nIntersections = 0;

/** A contour intersection list for each corner of the cell.
 * 
 */
protected ArrayList[]     corners = new ArrayList[4];

/** A contour intersection list for the side of the cell (either vertical or
 * horizontal, it can't be both).
 * 
 */
protected ArrayList       side = null;

/** A place to temporarily store new Intersections.
 * 
 */
protected Intersection    newIntersection = null;

/**  * 
 */
ContourCell()
{
}

/** 
 * Sets whether this cell corresponds to a hole or not.
 * 
 * @param hole 
 */
protected void setHole(boolean hole)
{
	this.hole = hole;
}

/** 
 * Does this cell correspond to a hole?
 * 
 */
protected boolean isHole()
{
	return(hole);
}

/** 
 * Adds an intersection to one of the four corner lists or the side list.
 * 
 * @param intersection 
 * @param place 
 */
protected void addIntersection(Intersection intersection, int place)
{
	if (place < 0 || place > SIDE) {
		
		return;
	}

	if (place == SIDE) {
		if (side == null) {
			side = new ArrayList();
		}
		side.add(intersection);
	}
	else {
		if (corners[place] == null) {
			corners[place] = new ArrayList();
		}
		corners[place].add(intersection);
	}
	nIntersections++;
}

/** 
 * Retrieves the number of contour intersections for this cell.
 * 
 */
public int getNIntersections()
{
	return(nIntersections);
}

/** 
 * Gets the newest intersection (possibly still under construction).
 * 
 */
protected Intersection getNewIntersection()
{
	return(newIntersection);
}

/** 
 * Sets the newest intersection.
 * 
 * @param newIntersection 
 */
protected void setNewIntersection(Intersection newIntersection)
{
	this.newIntersection = newIntersection;
}

/**  * 
 * @param place 
 */
public ArrayList getIntersectionList(int place)
{
	if (place < 0 || place > SIDE) {
		
		return(null);
	}

	if (place == SIDE) {
		return(side);
	}
	else {
		return(corners[place]);
	}
}

/** 
 * Sorts the contour intersection lists of each of the four corners and
 * the side list.
 * 
 */
protected void sortIntersectionLists()
{
	for (int i = 0; i < corners.length; i++) {
		if (corners[i] != null && corners[i].size() > 1) {
			Collections.sort(corners[i], this);
		}
	}
	if (side != null && side.size() > 1) {
		Collections.sort(side, this);
	}
}

/** 
 * Generic segment sorting routine, for all four corners, and sides
 * (left-to-right or top-to-bottom).  Each of these has a list of intersections
 * that need to be sorted.  We know that all edges have the same side1 and
 * side2, so the sorting is easy.
 * 
 * @param o1 
 * @param o2 
 */
public int compare(Object o1, Object o2)
{
	Intersection intersect1 = (Intersection)o1;
	Intersection intersect2 = (Intersection)o2;

	Point3d point1 = intersect1.getSegment2().getPoint();
	Point3d point2 = intersect2.getSegment2().getPoint();

	int value = 0;
	switch(intersect2.side2) {
	case 0:
	    // Sort by increasing y value
	    value = (point1.y < point2.y) ? -1 : (point1.y > point2.y ? 1 : 0);
	    break;
	case 1:
	    // Sort by decreasing x value
	    value = (point1.x > point2.x) ? -1 : (point1.x < point2.x ? 1 : 0);
	    break;
	case 2:
	    // Sort by decreasing y value
	    value = (point1.y > point2.y) ? -1 : (point1.y < point2.y ? 1 : 0);
	    break;
	case 3:
	    // Sort by increasing x value
	    value = (point1.x < point2.x) ? -1 : (point1.x > point2.x ? 1 : 0);
	    break;
	default:
	    
	    break;
	}
	return(value);
}

}
