/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DrawFacetAbstract.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:57 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Graphics;



/**
 * Abstract class that forms a basis for the DrawFacet??? classes.  It contains
 * polygon methods that draw, erase, or fill facets in various ways.  The
 * subclasses implement the abstract draw method and call the polygon methods
 * in various ways to implement the look required (determined by the meshed,
 * shaded, contoured, and zoned attributes).
 *
 */
public abstract class DrawFacetAbstract implements DrawFacet, java.io.Serializable {

/** Parent </code>DrawGrid</code> object.
 *
 */
protected DrawGrid            drawGrid = null;

/** The graphics context used.
 *
 */
protected transient Graphics  gc = null;

/** If this facet has been triangularized, draw the mesh line for this facet?
 *
 */
protected boolean             drawMeshLine = false;

/**  */
DrawFacetAbstract()
{
}

/**  */
DrawFacetAbstract(DrawGrid drawGrid, Graphics gc)
{
	this.drawGrid  = drawGrid;
	this.gc        = gc;
}

/**
 * Sets the <code>drawGrid</code> parent.
 *
 * @param drawGrid
 */
public void setDrawGrid(DrawGrid drawGrid)
{
	this.drawGrid = drawGrid;
}

/**
 * Gets the <code>drawGrid</code> parent.
 *
 */
public DrawGrid getDrawGrid()
{
	return(drawGrid);
}

/**
 * Sets the graphics context used to draw this facet.
 *
 * @param gc
 */
public void setGC(Graphics gc)
{
	this.gc = gc;
}

/**
 * Gets the graphics context used to draw this facet.
 *
 */
public Graphics getGC()
{
	return(gc);
}

/**
 * Sets whether the mesh line gets drawn for this facet.
 *
 * @param drawMeshLine
 */
public void setDrawMeshLine(boolean drawMeshLine)
{
	this.drawMeshLine = drawMeshLine;
}

/**
 * Returns whether the mesh line get drawn for this facet.
 *
 */
public boolean getDrawMeshLine()
{
	return(drawMeshLine);
}

/**
 * Draws the outline of the polygon represented by <code>xPoints</code> and <code>yPoints</code>.  This
 * will either be a 3 or 4 sided polygon depending on whether the original
 * facet has been triangularized.
 *
 * @param xPoints
 * @param yPoints
 * @param n
 * @param holeCode
 * @param x
 * @param y
 */
public void polyOutline(int[] xPoints, int[] yPoints, int n,
			            int holeCode, int x, int y)
{
	Chart3dDataView dataView  = drawGrid.getDataView();
	JCElevation     elevation = dataView.getElevation();
	JCChart3dArea   area      = dataView.getChart3d().getChart3dArea();
	JCSurface       surface   = area.getSurface();

	if (holeCode == 0) {
		if (n == 3) {
		   // If n = 3 then the original 4-sided facet has been chopped up into
	       // triangles. In this case, we only want to draw the base of the
		   // triangle if the corresponding mesh line is supposed to be drawn.
			if (drawMeshLine) {
			    if (JCChart3dUtil.backfacing(xPoints, yPoints)) {
				    gc.setColor(elevation.getMeshBottomColor());
				}
		        else {
				    gc.setColor(elevation.getMeshTopColor());
				}
                gc.drawLine(xPoints[1], yPoints[1], xPoints[2], yPoints[2]);
			}
		}
		else {
			// Determine color from backfacing information
			if (JCChart3dUtil.backfacing(xPoints, yPoints)) {
			    gc.setColor(elevation.getMeshBottomColor());
			}
		    else {
			    gc.setColor(elevation.getMeshTopColor());
			}

			boolean[] xMesh = surface.getXMesh();
			boolean[] yMesh = surface.getYMesh();
			if (xMesh[y+1]) {
                gc.drawLine(xPoints[1], yPoints[1], xPoints[2], yPoints[2]);
			}
			if (xMesh[y]) {
                gc.drawLine(xPoints[3], yPoints[3], xPoints[0], yPoints[0]);
			}
			if (yMesh[x]) {
                gc.drawLine(xPoints[0], yPoints[0], xPoints[1], yPoints[1]);
			}
			if (yMesh[x+1]) {
                gc.drawLine(xPoints[2], yPoints[2], xPoints[3], yPoints[3]);
			}
		}
	}
}

/**
 * Fills the polygon represented by <code>xPoints</code> and <code>yPoints</code>.  This will either be
 * a 3 or 4 sided polygon depending on whether the original facet has been
 * triangularized.
 *
 * @param xPoints
 * @param yPoints
 * @param n
 * @param holeCode
 */
public void polyFilled(int[] xPoints, int[] yPoints, int n, int holeCode)
{
	JCElevation     elevation = drawGrid.getDataView().getElevation();

	if (holeCode == 0) {
		// Determine color from backfacing information
		if (JCChart3dUtil.backfacing(xPoints, yPoints)) {
			gc.setColor(elevation.getShadedBottomColor());
		}
		else {
			gc.setColor(elevation.getShadedTopColor());
		}

        gc.fillPolygon(xPoints, yPoints, n);
	}
}

/**
 * Erases the polygon represented by <code>xPoints</code> and <code>yPoints</code>.  This will either be
 * a 3 or 4 sided polygon depending on whether the original facet has been
 * triangularized.  Erasing is needed so that the correct background will
 * show underneath the facet.
 *
 * @param xPoints
 * @param yPoints
 * @param n
 * @param holeCode
 */
public void polyErase(int[] xPoints, int[] yPoints, int n, int holeCode)
{
	JCChart3dArea   area = drawGrid.getDataView().getChart3d().getChart3dArea();
	JCPlotCube      plotCube = area.getPlotCube();
	JCFillStyle     fillStyle = new JCFillStyle(plotCube.getDrawBackground(),
									            JCFillStyle.SOLID);

	if (holeCode == 0) {
	    fillStyle.updateGraphics(gc);
        gc.fillPolygon(xPoints, yPoints, n);
	}
}

/**  *
 * @param x
 * @param y
 */
public boolean polyFilledZones(int x, int y)
{
	Chart3dGridData     data = drawGrid.getGridData();
	GridLimits          gridLimits = data.getGridLimits();
	JCContour           contour = data.getDataView().getContour();

	return(contour.fillCell(gc, x - gridLimits.getXOffset(),
								y - gridLimits.getYOffset()));
}

/**  *
 * @param x
 * @param y
 */
public void polyContours(int x, int y)
{
	Chart3dGridData     data = drawGrid.getGridData();
	GridLimits          gridLimits = data.getGridLimits();
	JCContour           contour = data.getDataView().getContour();

	contour.drawCellContours(gc, x - gridLimits.getXOffset(),
								 y - gridLimits.getYOffset());
}

/**  *
 * @param x
 * @param y
 */
public boolean polyOutlineZones(int x, int y)
{
	Chart3dGridData     data = drawGrid.getGridData();
	GridLimits          gridLimits = data.getGridLimits();
	JCContour           contour = data.getDataView().getContour();

	return(contour.zoneMeshCell(gc, x - gridLimits.getXOffset(),
									y - gridLimits.getYOffset()));
}

/**  *
 * @param xPoints
 * @param yPoints
 * @param n
 * @param holeCode
 * @param x
 * @param y
 */
public abstract void draw(int[] xPoints, int[] yPoints, int n, int holeCode,
				          int x, int y);

}
