/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DrawGrid.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:57 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Graphics;
import javax.vecmath.Vector3d;
import javax.vecmath.Point3d;
import javax.vecmath.Point4d;
import javax.vecmath.Matrix4d;



/**
 * The class that controls the drawing for grid data.  It also houses the
 * Ray-surface intersection algorithm for grid data.
 *
 */
public class DrawGrid extends Draw implements java.io.Serializable {

/**  */
public final static double AB_EPSILON = 1.0e-12;

/**  */
protected GridRegion        gridRegion = new GridRegion(this);
/**  */
protected ProcessGrid       processGrid = new ProcessGrid(this);
/**  */
protected DrawFacetFactory  drawFacetFactory = null;

/**  */
protected Vector3d		N    = null;
/**  */
protected Vector3d      V0V1 = null;
/**  */
protected Vector3d      V0V2 = null;
/**  */
protected Vector3d      V0   = null;
/**  */
protected Vector3d      V1   = null;
/**  */
protected Vector3d      V2   = null;

/**
 * Basic constructor.
 *
 */
public DrawGrid()
{
	super();
}

/**
 * Constructor specifying parent data view.
 *
 * @param dataView
 */
public DrawGrid(Chart3dDataView dataView)
{
	super(dataView);
}

/**
 * Data must be grid data.  This is a convenience method to get the data from
 * the <code>Chart3dDataView</code>.
 *
 */
public Chart3dGridData getGridData()
{
    if (dataView != null) {
        return((Chart3dGridData)dataView.getElevationData());
    }
    return(null);
}

/**
 * Returns the <code>GridRegion</code> object.  This object is used to define the nine
 * regions within a grid based on the current eye position
 *
 */
public GridRegion getGridRegion()
{
	return(gridRegion);
}

/**
 * Returns the <code>ProcessGrid</code> object.  This object determines the grid traversal
 * order based on the current eye position.
 *
 */
public ProcessGrid getProcessGrid()
{
	return(processGrid);
}

/**
 * Actually draws the graph based on the grid data.
 *
 * @param gc
 */
protected void draw(Graphics gc)
{
	int             type      = dataView.getChartType();
	JCChart3dArea   area      = dataView.getChart3d().getChart3dArea();
	JCElevation     elevation = dataView.getElevation();
	JCSurface       surface   = area.getSurface();
	JCBar           bar       = area.getBar();
    Chart3dGridData data      = getGridData();

    int numX = data.getNumX();
    int numY = data.getNumY();
	if (type == JCChart3d.SURFACE) {
		if (numX < 2 || numY < 2) {
			return;
		}
	}
	else if (type == JCChart3d.BAR) {
		if ((bar.getXFormat() == JCBar.HISTOGRAM && numX < 2) ||
		    (bar.getYFormat() == JCBar.HISTOGRAM && numY < 2))
		{
			return;
		}
	}
	else {	// JCChart3d.SCATTER_PLOT
		if (numX < 1 || numY < 1) {
			return;
		}
	}

	// Special cases, saves work
	if (type == JCChart3d.SURFACE) {
	    int drawCode = dataView.getDrawCode();
		if ((drawCode & ~dataView.ZONED) == dataView.MESHED) {
			if (!surface.isXMeshShowing() && !surface.isYMeshShowing()) {
				return;
			}
		}
		if (drawCode == dataView.MESHED &&
		    !surface.isSolid() &&
			elevation.isTransparent() &&
            // !xrt->text_list &&
			elevation.getMeshTopColor().equals(elevation.getMeshBottomColor()))
		{
			// The smoker!
			processGrid.drawFastMesh(gc);
			return;
		}
	}

	// Figure out which facet drawing method based on chart type.
	Facet  facet = null;
	switch (dataView.getChartType()) {
	case JCChart3d.SURFACE:
	    if (drawFacetFactory == null) {
			drawFacetFactory = new DrawFacetFactory();
	    }
	    facet = new SurfaceFacet(this, drawFacetFactory.createDrawFacet(this, gc));
		break;
	case JCChart3d.BAR:
	    facet = new BarFacet(this, gc);
		drawFacetFactory = null;
		break;
	case JCChart3d.SCATTER:
	    facet = new ScatterFacet(this, gc);
		drawFacetFactory = null;
		break;
	default:
		return;
	}
	processGrid.setFacet(facet);
	processGrid.processGridRegion(gridRegion);

	// Draw solid surfaces, if required
    if (dataView.getChartType() == JCChart3d.SURFACE &&
	    area.getSurface().isSolid())
	{
        SurfaceFacet surfaceFacet = (SurfaceFacet)facet;
        surfaceFacet.drawSolid(gridRegion);
    }
}

/**
 * For 2d grid data and the chart type is scatter, draws the points.
 *
 * @param gc
 */
protected void draw2dPoints(Graphics gc)
{
	Chart3dGridData    data          = getGridData();
	GridLimits         gridLimits    = data.getGridLimits();
	ScreenPoint[][]    screenPoints  = data.getScreenPoints();
	Chart3dDataView    dataView      = data.getDataView();
    Chart3dGridData    zoneData      = dataView.getZoneData();
	JCContour          contour       = dataView.getContour();
	JCContourLevels    contourLevels = contour.getContourLevels();
	JCContourStyle     contourStyle  = null;
	JCLineStyle        lineStyle     = null;
	JCSymbolStyle      symbolStyle   = null;

	if (!data.isDataOK()) {
		return;
	}

	JCChart3dStyle chartStyle = data.getChartStyle();
	for (int i = gridLimits.getXOffset(); i <= gridLimits.getXLast(); i++) {
		for (int j = gridLimits.getYOffset(); j <= gridLimits.getYLast(); j++) {
		    ScreenPoint pt = screenPoints[i][j];
		    if (!pt.isDrawn() && !pt.isHoleValue()) {
				if (contour.isContoured() || contour.isZoned()) {
			        double value = data.getZValue(i, j);
			        if (zoneData.isDataOK()) {
				        value = zoneData.getZValue(i, j);
					}
					int level = contourLevels.getLevelFromValue(value);
					contourStyle = contour.getContourStyleFromLevel(level);
					if (contourStyle == null) {
						continue;
					}
					symbolStyle = contourStyle.getSymbolStyle();
					lineStyle   = contourStyle.getLineStyle();
				}
				else {
					if (chartStyle == null) {
						return;
					}
					symbolStyle = chartStyle.getSymbolStyle();
					lineStyle   = chartStyle.getLineStyle();
				}
			    renderPoint(gc, symbolStyle, lineStyle, pt,
							data.getX(i), data.getY(j));
			}

		}
	}
}

// Ray-Surface Intersection algorithm for grid surfaces. (Used for mapping
// and picking)

/**
 * Given a grid index, maps its value to absolute (or Cube) space.
 *
 * @param absolute
 * @param data
 * @param gridIndex
 * @param isX
 */
protected double cubeSpace(Matrix4d absolute, Chart3dGridData data,
				           int gridIndex, boolean isX)
{
	double gridValue = Double.MAX_VALUE;

	if (isX) {
		gridValue = absolute.m00 * data.getX(gridIndex) + absolute.m03;
	}
	else {
		gridValue = absolute.m11 * data.getY(gridIndex) + absolute.m13;
	}
	return(gridValue);
}

/**
 * Maps a point to absolute space.
 *
 * @param data
 * @param absolute
 * @param x
 * @param y
 */
public Point4d findAbsolutePoint(Chart3dGridData data, Matrix4d absolute,
			                     int x, int y)
{
	Point4d p = new Point4d();
	p.x = cubeSpace(absolute, data, x, true);
	p.y = cubeSpace(absolute, data, y, false);
	p.z = data.getZValue(x, y) * absolute.m22 + absolute.m23;
	p.w = 1.0;
	return(p);
}

/**
 * Given a ray via its origin and direction and a triangle with its vertices
 * ,
 * computes the intersection point of the ray and the triangle.  This method
 * returns true if there is an intersection and if its closer than any other
 * ray.
 *
 * @param rOrigin
 * @param rDirection
 * @param vertex0
 * @param vertex1
 * @param vertex2
 * @param intersection
 * @param index
 */
protected boolean rayIntersectTriangle(Point3d rOrigin, Vector3d rDirection,
				                       Point4d vertex0, Point4d vertex1,
									   Point4d vertex2, Point3d intersection,
									   JCData3dGridIndex index)
{
	double          u0, u1, u2, v0, v1, v2;

	if (N == null) {
		N    = new Vector3d();			/* Normal to triangle */
		V0V1 = new Vector3d();
		V0V2 = new Vector3d();
		V0   = new Vector3d();
		V1   = new Vector3d();
		V2   = new Vector3d();
	}
	V0.set(vertex0.x, vertex0.y, vertex0.z);
	V1.set(vertex1.x, vertex1.y, vertex1.z);
	V2.set(vertex2.x, vertex2.y, vertex2.z);

	// Compute the triangle's normal vector -- the cross product of two
	// of the triangle sides
	V0V1.sub(V1, V0);
	V0V2.sub(V2, V0);
	N.cross(V0V1, V0V2);

	// Take dot product of normal and direction
	double ND = N.dot(rDirection);
	if (ND == 0.0) {
		// Ray is parallel to triangle
		return(false);
	}

	// Calculate ray parameter t at intersection point.
	// Then intersection = rOrigin + rDirection * t
	Vector3d vOrigin = new Vector3d(rOrigin);
	double t = (N.dot(V0) - N.dot(vOrigin)) / ND;

	if (t <= 0.0) {
		// Intersection is behind ray
		return(false);
	}

	// Compute intersection point (ray & plane)
	intersection.scaleAdd(t, rDirection, rOrigin);

	// Find dominant projection of triangle to axis planes by computing
	// absolute value of normal components
	N.absolute();

	// Compute parameters in projected plane
	if ((N.z > N.x) && (N.z > N.y)) {  // Z is dominant
	   u0 = intersection.y - V0.y;
	   u1 = V1.y - V0.y;
	   u2 = V2.y - V0.y;

	   v0 = intersection.x - V0.x;
	   v1 = V1.x - V0.x;
	   v2 = V2.x - V0.x;
	}
	else  if (N.y > N.x) {             // Y is dominant
	   u0 = intersection.x - V0.x;
	   u1 = V1.x - V0.x;
	   u2 = V2.x - V0.x;

	   v0 = intersection.z - V0.z;
	   v1 = V1.z - V0.z;
	   v2 = V2.z - V0.z;
	}
	else {                             // X is dominant
	   u0 = intersection.y - V0.y;
	   u1 = V1.y - V0.y;
	   u2 = V2.y - V0.y;

	   v0 = intersection.z - V0.z;
	   v1 = V1.z - V0.z;
	   v2 = V2.z - V0.z;
	}

	double alpha = (u0 * v2 - v0 * u2) / (u1 * v2 - v1 * u2);
	double beta  = (u1 * v0 - v1 * u0) / (u1 * v2 - v1 * u2);

	// If alpha, beta are inside triangle, return true
	boolean hit = (alpha 	     >= (0.0 - AB_EPSILON) &&
		           beta          >= (0.0 - AB_EPSILON) &&
		           alpha + beta  <= (1.0 + AB_EPSILON));

	// Return a hit only if the point is inside the triangle AND the
	// point is closer than any other ray
	if (hit && index.getDistance() > t) {
		index.setDistance(t);
		return(true);
	}
	return(false);
}

/**
 * Does the ray with given origin and direction intersect the bar at
 * the given grid point?
 *
 * @param data
 * @param rOrigin
 * @param rDirection
 * @param x
 * @param y
 * @param intersection
 * @param index
 */
protected boolean rayTestBar(Chart3dGridData data, Point3d rOrigin,
							 Vector3d rDirection, int x, int y,
							 Point3d intersection, JCData3dGridIndex index)
{
	Point4d[]		B         = new Point4d[4];        // Bottom face
	boolean			hit       = false;
	JCChart3dArea   area      = data.getDataView().getChart3d().getChart3dArea();
	Transform       transform = area.getTransform();
	Matrix4d        absolute  = transform.getAbsolute();

	// Contstruct top plane
	Point4d[] T = JCChart3dUtil.constructBarPlane(data, x, y);
	T[0].z = T[1].z = T[2].z = T[3].z = data.getZValue(x, y);
	transform.transformPoint(absolute, T[0]);
	transform.transformPoint(absolute, T[1]);
	transform.transformPoint(absolute, T[2]);
	transform.transformPoint(absolute, T[3]);

	// Construct bottom plane
	B[0] = new Point4d(T[0]);
	B[1] = new Point4d(T[1]);
	B[2] = new Point4d(T[2]);
	B[3] = new Point4d(T[3]);
	double zOrigin = area.getAxis(JCAxis.AXIS_Z).getOrigin();
	B[0].z = B[1].z = B[2].z = B[3].z = zOrigin * absolute.m22 + absolute.m23;

	// Test all 6 faces and return closest intersection if any
	hit |= rayIntersectTriangle(rOrigin, rDirection, T[0], T[1], T[2],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, T[0], T[3], T[2],
		                        intersection, index);

	hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], B[1], B[2],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], B[2], B[3],
		                        intersection, index);

	hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], B[1], T[1],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], T[0], T[1],
		                        intersection, index);

	hit |= rayIntersectTriangle(rOrigin, rDirection, B[1], B[2], T[2],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, B[1], T[1], T[2],
		                        intersection, index);

	hit |= rayIntersectTriangle(rOrigin, rDirection, B[2], B[3], T[3],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, B[2], T[2], T[3],
		                        intersection, index);

	hit |= rayIntersectTriangle(rOrigin, rDirection, B[3], B[0], T[0],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, B[3], T[3], T[0],
		                        intersection, index);
	return(hit);
}

/**
 * Does the ray with given origin and direction intersect the facet at
 * the given grid point?
 *
 * @param data
 * @param rOrigin
 * @param rDirection
 * @param x
 * @param y
 * @param intersection
 * @param index
 */
protected boolean rayTestFacet(Chart3dGridData data, Point3d rOrigin,
							   Vector3d rDirection, int x, int y,
							   Point3d intersection, JCData3dGridIndex index)
{
	JCChart3dArea   area      = data.getDataView().getChart3d().getChart3dArea();
	Transform       transform = area.getTransform();
	Matrix4d        absolute  = transform.getAbsolute();
	Point4d[]		T         = new Point4d[5];
	boolean			hit       = false;

	// Get the 4 corners of the grid in absolute space
	T[1] = findAbsolutePoint(data, absolute, x, y);
	T[2] = findAbsolutePoint(data, absolute, x, y + 1);
	T[3] = findAbsolutePoint(data, absolute, x + 1, y + 1);
	T[0] = findAbsolutePoint(data, absolute, x + 1, y);

	// Make center point (in absolute space)
	T[4] = new Point4d();
	T[4].x = (T[1].x + T[3].x) / 2.0;
	T[4].y = (T[1].y + T[3].y) / 2.0;
	T[4].z = (T[0].z + T[1].z + T[2].z + T[3].z) / 4.0;
	T[4].w = 1.0;

	// Split surface into 4 triangles and test for closest intersection
	hit |= rayIntersectTriangle(rOrigin, rDirection, T[0], T[1], T[4],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, T[1], T[2], T[4],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, T[2], T[3], T[4],
		                        intersection, index);
	hit |= rayIntersectTriangle(rOrigin, rDirection, T[3], T[0], T[4],
		                        intersection, index);

	// Must also test the sides and bottom of solid surfaces
	if (dataView.getChartType() == JCChart3d.SURFACE &&
	    area.getSurface().isSolid())
	{
		Point4d[] B = JCChart3dUtil.constructBarPlane(data, x, y);
		B[0].z = B[1].z = B[2].z = B[3].z = 0.0;

		transform.transformPoint(absolute, B[0]);
		transform.transformPoint(absolute, B[1]);
		transform.transformPoint(absolute, B[2]);
		transform.transformPoint(absolute, B[3]);
		double minValue = -area.getPlotCube().getZNormalizedScale();
		B[0].z = B[1].z = B[2].z = B[3].z = minValue;

		hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], B[1], B[2],
									intersection, index);
		hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], B[2], B[3],
									intersection, index);

		hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], B[1], T[1],
									intersection, index);
		hit |= rayIntersectTriangle(rOrigin, rDirection, B[0], T[0], T[1],
									intersection, index);

		hit |= rayIntersectTriangle(rOrigin, rDirection, B[1], B[2], T[2],
									intersection, index);
		hit |= rayIntersectTriangle(rOrigin, rDirection, B[1], T[1], T[2],
									intersection, index);

		hit |= rayIntersectTriangle(rOrigin, rDirection, B[2], B[3], T[3],
									intersection, index);
		hit |= rayIntersectTriangle(rOrigin, rDirection, B[2], T[2], T[3],
									intersection, index);

		hit |= rayIntersectTriangle(rOrigin, rDirection, B[3], B[0], T[0],
									intersection, index);
		hit |= rayIntersectTriangle(rOrigin, rDirection, B[3], T[3], T[0],
									intersection, index);
	}

	return(hit);
}

/**
 * Does the ray with given origin and direction intersect the surface, bar,
 * or scatter plot for the given grid data
 *
 * @param data
 * @param rOrigin
 * @param rDirection
 * @param x
 * @param y
 * @param intersection
 * @param index
 */
protected boolean testFacet(Chart3dGridData data, Point3d rOrigin,
							Vector3d rDirection, int x, int y,
							Point3d intersection, JCData3dGridIndex index)
{
	Chart3dDataView    dataView   = data.getDataView();
	GridLimits         gridLimits = data.getGridLimits();
	JCChart3dArea      area       = dataView.getChart3d().getChart3dArea();
	int                chartType  = dataView.getChartType();

	if (chartType == JCChart3d.BAR ||
	    (chartType == JCChart3d.SCATTER && area.getScatter().hasDropLines()))
	{
		if (data.getZValue(x, y) == data.getHoleValue() ||
			!rayTestBar(data, rOrigin, rDirection, x, y, intersection, index))
		{
			return(false);
		}
	}
	else if (chartType == JCChart3d.SURFACE ||
		     (chartType == JCChart3d.SCATTER && !area.getScatter().hasDropLines()))
	{
		if (chartType == JCChart3d.SCATTER) {
			x = JCChart3dUtil.clamp(x, gridLimits.getXOffset(),
			                        gridLimits.getXLast() - 1);
			y = JCChart3dUtil.clamp(y, gridLimits.getYOffset(),
			                        gridLimits.getYLast() - 1);
		}
		if (JCChart3dUtil.holeInFacet(data, x, y) ||
			!rayTestFacet(data, rOrigin, rDirection, x, y, intersection, index))
		{
			return(false);
		}
	}

	// We have a hit.  Calculate intersection point.
	intersection.scaleAdd(index.getDistance(), rDirection, rOrigin);
	return(true);
}

/**
 * Computes the intersection parameter (t) of a ray (p1->p2) and a line
 * segment (q1->q2) in 2D.
 *
 * @param p1
 * @param p2
 * @param q1
 * @param q2
 */
protected double lineIntersect2D(Point3d p1, Point3d p2, Point3d q1, Point3d q2)
{
	double Ax = p2.x - p1.x;
	double Ay = p2.y - p1.y;
	double Bx = q1.x - q2.x;
	double By = q1.y - q2.y;

	double t = Ay * Bx - Ax * By;

	if ((1.0 + t) == 1.0) {
		return(Double.MAX_VALUE);		// Parallel lines
	}

	double Cx = p1.x - q1.x;
	double Cy = p1.y - q1.y;

	double s = (Ax * Cy - Ay * Cx) / t;

	if (-0.01 > s || s > 1.01) {
		return(Double.MAX_VALUE);		// Intersection not on segment
	}
	t = (By * Cx - Bx * Cy) / t;

	return(t);
}

/**
 * Computes whether the ray defined by the given origin and point, intersects
 * the surface defined by the grid data associated with this object.
 *
 * @param rayOrigin
 * @param rayPoint
 * @param result
 * @param index
 */
public boolean gridSurfaceIntersect(Point4d rayOrigin, Point4d rayPoint,
			                        Point4d result, JCData3dGridIndex index)
{
	Chart3dGridData    data         = getGridData();
	Chart3dDataView    dataView     = data.getDataView();
	JCChart3dArea      area         = dataView.getChart3d().getChart3dArea();
	JCBar              bar          = area.getBar();
	GridLimits         gridLimits   = data.getGridLimits();
	Matrix4d           absolute     = area.getTransform().getAbsolute();
	Point3d            rOrigin      = new Point3d();
	Vector3d           rDirection   = new Vector3d();
	Point3d            beginPt      = new Point3d();
	Point3d            endPt        = new Point3d();
	Point3d            intersection = new Point3d();

	int xBegin = gridLimits.getXOffset();
	int yBegin = gridLimits.getYOffset();
	int xEnd   = gridLimits.getXLast();
	int yEnd   = gridLimits.getYLast();

	int    xHistogram = 1;
	int    yHistogram = 1;
	double xExtra     = 0.0;
	double yExtra     = 0.0;
	if (dataView.getChartType() == JCChart3d.BAR) {
		if (bar.getXFormat() == JCBar.FIXED) {
			xExtra = absolute.m00 * data.getMinXStep() / 2.0;
			xHistogram = 0;
		}
		if (bar.getYFormat() == JCBar.FIXED) {
			yExtra = absolute.m11 * data.getMinYStep() / 2.0;
			yHistogram = 0;
		}
	}

	rOrigin.project(rayOrigin);
	rDirection.set(0.0, 0.0, 0.0);
	beginPt.set(rOrigin);
	endPt.project(rayPoint);

	// 2D bounding region
	Point3d gridYMin = new Point3d();
	Point3d gridYMax = new Point3d();
	Point3d gridXMin = new Point3d();
	Point3d gridXMax = new Point3d();

	boolean first = true;
	int     x     = 0;
	int     y     = 0;
	while (true) {
		// Determine bounding region (in absolute space).  Surface and histogram
		// boundaries lie on grid lines whereas fixed bar boudaries lie in the
		// middle of grid lines.

		gridYMin.x = cubeSpace(absolute, data, xBegin, true);
		gridYMax.x = cubeSpace(absolute, data, xEnd,   true);
		gridXMin.y = cubeSpace(absolute, data, yBegin, false);
		gridXMax.y = cubeSpace(absolute, data, yEnd,   false);

		if (xHistogram == 0) {
			if (xBegin <= gridLimits.getXOffset()) {
				gridYMin.x -= xExtra;
			}
			else {
				gridYMin.x += cubeSpace(absolute, data, xBegin - 1, true);
				gridYMin.x /= 2.0;
			}

			if (xEnd >= gridLimits.getXLast()) {
				gridYMax.x += xExtra;
			}
			else {
				gridYMax.x += cubeSpace(absolute, data, xEnd + 1, true);
				gridYMax.x /= 2.0;
			}
		}

		if (yHistogram == 0) {
			if (yBegin <= gridLimits.getYOffset()) {
				gridXMin.y -= yExtra;
			}
			else {
				gridXMin.y += cubeSpace(absolute, data, yBegin - 1, false);
				gridXMin.y /= 2.0;
			}

			if (yEnd >= gridLimits.getYLast()) {
				gridXMax.y += yExtra;
			}
			else {
				gridXMax.y += cubeSpace(absolute, data, yEnd + 1, false);
				gridXMax.y /= 2.0;
			}
		}

		/*	Check each side of boundary for closest intersection
		 *  using the ray direction to eliminate some tests
		 */

		double smin = Double.MAX_VALUE;
		double s    = 0.0;
		int    side = 0;

		if (rDirection.x <= 0) {
			gridXMin.x = gridXMax.x = gridYMin.x;
			s = lineIntersect2D(beginPt, endPt, gridXMin, gridXMax);
			if (smin > s) {
				smin = s;
				side = 1;
			}
		}

		if (rDirection.x >= 0) {
			gridXMin.x = gridXMax.x = gridYMax.x;
			s = lineIntersect2D(beginPt, endPt, gridXMin, gridXMax);
			if (smin > s) {
				smin = s;
				side = 2;
			}
		}

		if (rDirection.y <= 0) {
			gridYMin.y = gridYMax.y = gridXMin.y;
			s = lineIntersect2D(beginPt, endPt, gridYMin, gridYMax);
			if (smin > s) {
				smin = s;
				side = 3;
			}
		}

		if (rDirection.y >= 0) {
			gridYMin.y = gridYMax.y = gridXMax.y;
			s = lineIntersect2D(beginPt, endPt, gridYMin, gridYMax);
			if (smin > s) {
				smin = s;
				side = 4;
			}
		}

		if (first) {
			// The first pass through this loop we intersect with the entire
			// bounding cube.  This intersection point is the start of the
			// grid traversal.
			first = false;
			rDirection.sub(endPt, beginPt);

			if (side != 0 &&
				!(gridYMin.x < beginPt.x &&  beginPt.x < gridYMax.x &&
				  gridXMin.y < beginPt.y &&  beginPt.y < gridXMax.y))
			{
				// Outside bounding box, calculate intersection point
				beginPt.x = beginPt.x + smin * rDirection.x;
				beginPt.y = beginPt.y + smin * rDirection.y;
			}

			xBegin = x = data.getGridIndex(beginPt.x, true, true);
			yBegin = y = data.getGridIndex(beginPt.y, false, true);
		}
		else {                      // Find next grid cell
			switch (side) {
			case 1: x--; break;
			case 2: x++; break;
			case 3: y--; break;
			case 4: y++; break;
			default: return(false);
			}
		}

		if (testFacet(data, rOrigin, rDirection, xBegin, yBegin,
		              intersection, index))
		{
			result.set(intersection);
			result.w = 1.0;
			index.setX(xBegin);
			index.setY(yBegin);
			return(true);
		}

		xBegin = x;
		yBegin = y;
		xEnd   = x + xHistogram;
		yEnd   = y + yHistogram;

		if (xBegin < gridLimits.getXOffset() || xEnd > gridLimits.getXLast() ||
			yBegin < gridLimits.getYOffset() || yEnd > gridLimits.getYLast())
		{
			// We have reached the edge of the grid
			return(false);
		}
	}
}

}
