/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCAxis.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Font;
import java.util.ArrayList;

import com.klg.jclass.chart3d.resources.LocaleBundle;

/**
 * The <code>JCAxis</code> class represents one of the three axes in 3D space.
 * It it used to store and control axis related properties like its min, max,
 * annotation method, title, gridlines, annotation, and title font.  It has
 * an annotation handler to handle the annotation generation.
 */
public class JCAxis extends ParentFormatter {

static final long serialVersionUID = -5542417106751358122L;

/**
 * The maximum axis font cube size. As this measure is in 1000th of the
 * of the cube, it stands to reason that 1000 is the maximum value.
 */
public final static int MAX_AXIS_FONT_CUBE_SIZE = 1000;

/** Enum used in the rotate actions where no particular axis is selected. */
public final static int AXIS_NONE = 0;

/** The X axis. */
public final static int AXIS_X    = 1;

/** The Y axis. */
public final static int AXIS_Y    = 2;

/** The Z axis. */
public final static int AXIS_Z    = 3;

/** Enum used in the rotate actions for rotation about the user's eye position. */
public final static int AXIS_EYE  = 4;

/** The number of axes on the plotcube. */
public final static int NUM_AXES  = 3;

/** Annotation method in which chart 3D computes the labels based on axis values. */
public final static int ANNOTATION_VALUES       = 1;

/** Annotation method in which the xLabels and yLabels are used (grid data only). */
public final static int ANNOTATION_DATA_LABELS  = 2;

/** Annotation method in which the user's value labels are used for annotation. */
public final static int ANNOTATION_VALUE_LABELS = 3;

/** Annotation rotation is caclulated based on type of annotation an spacing */
public final static int ANNOTATION_ROTATION_DEFAULT    = 0;

/** Annotation is horizontal or parallel to the axis */
public final static int ANNOTATION_ROTATION_HORIZONTAL = 1;

/** Annoation is vertical or orthogonal to the axis */
public final static int ANNOTATION_ROTATION_VERTICAL   = 2;


/** Asks if the axis showing. */
protected boolean            showing = true;

/**
 * The annotation method.  One of <code>ANNOTATION_VALUES</code>,
 * <code>ANNTOATION_DATA_LABELS</code>, or <code>ANNOTATION_VALUE_LABELS</code>.
 */
protected int                annotationMethod = ANNOTATION_VALUES;

/**
 * The annotation rotation.  One of <code>ANNOTATION_ROTATION_DEFAULT</code>,
 * <code>ANNTOATION_ROTATION_HORIZONTAL</code>, or
 * <code>ANNOTATION_ROTATION_VERTICAL</code>.
 */
protected int                annotationRotation = ANNOTATION_ROTATION_DEFAULT;

/**
 * The origin is used for <i>Z</i> axis only (for bars and scatter plot drop
 * lines).  It defaults to 0.0 (or <i>Z</i> axis min) but can be set to any
 * value within the <i>Z</i> axis range).
 * */
protected double             origin = 0.0;

/** The axis title. */
protected String             title = null;

/** The axis minimum value. */
protected DerivedDouble      min = new DerivedDouble(Double.MAX_VALUE, true);

/** The axis maximum value. */
protected DerivedDouble      max = new DerivedDouble(Double.MAX_VALUE, true);

/** Gridlines on a per plane basis for this axis. */
protected JCGridLines        gridLines = null;

/** The annotation font for this axis. */
protected Font               annoFont = null;

/**
 * The annotation font cube size for this axis (this size is measured in
 * thousandths of the unit cube size and must be between 0 and 1000).
 */
protected int                annoFontCubeSize = 80;

/** The title font for this axis. */
protected Font               titleFont = null;

/**
 * The title font cube size for this axis (this size is measured in
 * thousandths of the unit cube size and must be between 0 and 1000).
 */
protected int                titleFontCubeSize = 80;

/**
 * If <code>true</code>, consecutive data labels that are identical (in other
 * words, that refer to the same object) are combined into one label, and the
 * resulting label is positioned at the average (or middle) label position.
 * Ticks are still drawn at each data label position.
 */
protected boolean            combineLabels = false;

// The annotation extents
protected double             annoMin       = Double.MAX_VALUE;
protected double             annoMax       = Double.MAX_VALUE;
protected double             annoStep      = Double.MAX_VALUE;
protected double             annoEaseway   = 0.0;
protected int                maxAnnoLength = 0;

// The dataport extents
protected double             dataportMin = Double.MAX_VALUE;
protected double             dataportMax = Double.MAX_VALUE;

/** Handler for code specific to axis annotations. */
protected Chart3dAnnoHandler annoHandler = new Chart3dAnnoHandler(this);

/** Label generator which is used for custom annotations. */
protected JCLabelGenerator labelGenerator = null;

/**
 * The axis id number.  Usually one of <code>AXIS_X</code>, <code>AXIS_Y</code>,
 * and <code>AXIS_Z</code>.  This property can only be set on creation.
 */
protected int axisId = -1;

/**
 * No args constructor.  Axis id will be invalid.  Font will be null.
 */
public JCAxis()
{
}

/**
 * Constructor which supplies the <code>axisId</code>.  The </code>axisId</code>
 * can only be set on creation.
 *
 * @param axisId Which axis is it? Should be one of <code>AXIS_X</code>,
 * <code>AXIS_Y</code>, or <code>AXIS_Z</code>.
 * @param defaultFont What font should be used as the default annotation and
 * title font?
 */
public JCAxis(int axisId, Font defaultFont)
{
    this.axisId    = axisId;
	this.annoFont  = defaultFont;
	this.titleFont = defaultFont;
}

/**
 * Returns the <code>axisId</code>.  Should be one of <code>AXIS_X</code>,
 * <code>AXIS_Y</code>, or <code>AXIS_Z</code>.
 *
 * @return the current <code>axisId</code>
 */
public int getAxisId()
{
    return(axisId);
}

/**
 * Sets flag, which determines whether this axis is showing or not.
 *
 * @param showing The new value of the showing flag
 */
public void setShowing(boolean showing)
{
	if (this.showing == showing) {
		return;
	}
    this.showing = showing;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns whether this axis is showing or not.
 *
 * @return The value of the showing flag
 */
public boolean isShowing()
{
    return(showing);
}

/**
 * Sets the axis annotation method.  Must be one of
 * <code>ANNOTATION_VALUES</code>, <code>ANNTOATION_DATA_LABELS</code>, or
 * <code>ANNOTATION_VALUE_LABELS</code>.
 *
 * @param annotationMethod The new value of the annotation method
 */
public void setAnnotationMethod(int annotationMethod)
{
	if (this.annotationMethod == annotationMethod) {
		return;
	}
	if (annotationMethod < ANNOTATION_VALUES ||
	    annotationMethod > ANNOTATION_VALUE_LABELS)
	{
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_ERROR));
	}
    this.annotationMethod = annotationMethod;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the annotation method.
 *
 * @return The current annotation method
 */
public int getAnnotationMethod()
{
    return(annotationMethod);
}

/**
 * Sets the axis annotation rotation. Must be one of
 * <code>ANNOTATION_ROTATION_DEFAULT</code>,
 * <code>ANNTOATION_ROTATION_HORIZONTAL</code>, or
 * <code>ANNOTATION_ROTATION_VERTICAL</code>.
 *
 * @param annotationRotation The new value of the annotation rotation
 */
public void setAnnotationRotation(int annotationRotation)
{
    if (this.annotationRotation == annotationRotation) {
        return;
    }
    if (annotationRotation < ANNOTATION_ROTATION_DEFAULT ||
        annotationRotation > ANNOTATION_ROTATION_VERTICAL)
    {
        String s = LocaleBundle.string(LocaleBundle.INVALID_VALUE_ERROR);
        throw new IllegalArgumentException(s);
    }
    this.annotationRotation = annotationRotation;
    callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the annotation rotation.
 *
 * @return The current annotation rotation
 */
public int getAnnotationRotation()
{
    return(annotationRotation);
}

/**
 * Sets the axis origin.
 *
 * @param origin the new value of the axis origin
 */
public void setOrigin(double origin)
{
    if (this.origin == origin) {
		return;
    }
	this.origin = origin;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the axis origin.
 *
 * @return the axis origin (only valid for Z axis)
 */
public double getOrigin()
{
    return(origin);
}

/**
 * Sets annotation font for this axis. <BR><B>Note:</b> The annotation font is scaled
 * to be the size of the <code>annoFontCubeSize</code>.
 *
 * @param annoFont the new annotation font
 */
public void setAnnoFont(Font annoFont)
{
	if (this.annoFont == annoFont) {
		return;
	}
	this.annoFont = annoFont;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Gets the annotation font for this axis.  <BR><B>Note:</b> The annotation font is
 * scaled to be the size of the <code>annoFontCubeSize</code>.
 *
 * @return the current annotation font
 */
public Font getAnnoFont()
{
	return(annoFont);
}

/**
 * Sets the size of the annotation font.  The size is measured in
 * thousandths of the unit cube size and must be between 0 and 1000.
 *
 * @param annoFontCubeSize the new size of the annotation font
 */
public void setAnnoFontCubeSize(int annoFontCubeSize)
{
	if (this.annoFontCubeSize == annoFontCubeSize) {
		return;
	}
	if (annoFontCubeSize < 0 || annoFontCubeSize > MAX_AXIS_FONT_CUBE_SIZE) {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_OUT_OF_THOUSAND_ERROR));

	}
	this.annoFontCubeSize = annoFontCubeSize;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the size of the annotation font.  The size is measured in
 * thousandths of the unit cube size and must be between 0 and 1000.
 *
 * @return the current size of the annotation font.
 */
public int getAnnoFontCubeSize()
{
	return(annoFontCubeSize);
}

/**
 * Sets title font for this axis. <br><B>Note:</b> The title font is scaled
 * to be the size of the <code>titleFontCubeSize</code>.
 *
 * @param titleFont the new title font
 */
public void setTitleFont(Font titleFont)
{
	if (this.titleFont == titleFont) {
		return;
	}
	this.titleFont = titleFont;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Gets the title font for this axis. <br><B>Note:</b> The title font is scaled
 * to be the size of the <code>titleFontCubeSize</code>.
 *
 * @return the current title font
 */
public Font getTitleFont()
{
	return(titleFont);
}

/**
 * Sets the size of the title font.  The size is measured in
 * thousandths of the unit cube size and must be between 0 and 1000.
 *
 * @param titleFontCubeSize the new size of the title font
 */
public void setTitleFontCubeSize(int titleFontCubeSize)
{
	if (this.titleFontCubeSize == titleFontCubeSize) {
		return;
	}
	if (titleFontCubeSize < 0 || titleFontCubeSize > MAX_AXIS_FONT_CUBE_SIZE) {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_OUT_OF_THOUSAND_ERROR));

	}
	this.titleFontCubeSize = titleFontCubeSize;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the size of the title font.  The size is measured in
 * thousandths of the unit cube size and must be between 0 and 1000.
 *
 * @return the current size of the title font
 */
public int getTitleFontCubeSize()
{
	return(titleFontCubeSize);
}

/**
 * Sets the title for this axis.
 *
 * @param title the new axis title
 */
public void setTitle(String title)
{
	if (this.title == title) {
		return;
	}
	this.title = title;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the title for this axis.
 *
 * @return the current axis title
 */
public String getTitle()
{
	return(title);
}

/**
 * Sets the <code>min</code> property of the <code>JCAxis</code>, which
 * controls the minimum value shown on the axis.
 *
 * @param newMin the new minimum value
 */
public void setMin(double newMin)
{
	if (!min.isDefault && min.value == newMin) {
		return;
	}
	min.value     = newMin;
	min.isDefault = false;

	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Gets the <code>min</code> property of the <code>JCAxis</code>, which
 * controls the minimum value shown on the axis.
 *
 * @return the current minimum value
 */
public double getMin() {
    return (min.value);
}

/**
 * Sets the <code>minIsDefault</code> property. The <code>minIsDefault</code>
 * property determines whether Chart3d is responsible for calculating the
 * minimum axis value. If <code>true</code>, Chart3d will calculate the axis min. If <code>false</code>,
 * Chart3d will use the provided axis min.
 *
 * @param useDef if <code>true</code>, Chart3d calculates the <code>min</code> property; if <code>false</code>, the
 *     <code>min</code> property is assumed to be already specified
 */
public void setMinIsDefault(boolean useDef)
{
	if (min.isDefault == useDef) {
		return;
	}
	min.isDefault = useDef;

	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Gets the <code>minIsDefault</code> property. The <code>minIsDefault</code>
 * property determines whether Chart3d is responsible for calculating the
 * minimum axis value. If <code>true</code>, Chart3d will calculate the axis min. If <code>false</code>,
 * Chart3d will use the provided axis min.
 *
 * @return <code>true</code> Chart3d calculates the <code>min</code> property; if <code>false</code>,
 * the <code>min</code> property is assumed to be already specified
 */
public boolean getMinIsDefault()
{
	return(min.isDefault);
}

/**
 * Sets the <code>max</code> property of the axis, which controls the maximum
 * value shown on the axis.
 *
 * @param newMax the new maximum value
 */
public void setMax(double newMax)
{
	if (!max.isDefault && max.value == newMax) {
		return;
	}
	max.value     = newMax;
	max.isDefault = false;

	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Gets the <code>max</code> property of the axis, which controls the maximum
 * value shown on the axis.
 *
 * @return the current maximum value
 */
public double getMax()
{
    return(max.value);
}

/**
 * Sets the <code>naxIsDefault</code> property. The <code>naxIsDefault</code>
 * property determines whether Chart3d is responsible for calculating the
 * maximum axis value. If <code>true</code>, Chart3d calculates the axis max. If <code>false</code>,
 * Chart3d uses the provided axis max.
 *
 * @param useDef if <code>true</code>, Chart3d calculates the <code>max</code> property;
 * if <code>false</code>, the <code>max</code> property is assumed to be already specified
 */
public void setMaxIsDefault(boolean useDef)
{
	if (max.isDefault == useDef) {
		return;
	}
	max.isDefault = useDef;

	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Gets the <code>maxIsDefault</code> property. The <code>maxIsDefault</code>
 * property determines whether Chart3d is responsible for calculating the
 * maximum axis value. If <code>true</code>, Chart3d calculates the axis max. If <code>false</code>,
 * Chart3d uses the provided axis max.
 *
 * @return true Chart3d calculates the <code>max</code> property. If <code>false</code>, the <code>max</code>
 *     property is assumed to be already specified
 */
public boolean getMaxIsDefault()
{
	return(max.isDefault);
}

/**
 * Sets the gridlines object.
 *
 * @param gridLines the new <code>JCGridLines</code> object (cannot be null)
 */
public void setGridLines(JCGridLines gridLines)
{
	if (this.gridLines == gridLines) {
		return;
	}
	this.gridLines = gridLines;
	if (this.gridLines != null) {
		// Note that setAxis calls this object's callParentSetChanged method.
		this.gridLines.setAxis(this);
	}
	else {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NULL_ERROR));
	}
}

/**
 * Gets the gridlines object.
 *
 * @return the current <code>JCGridLines</code> object
 */
public JCGridLines getGridLines()
{
	return(gridLines);
}

/**
 * Sets the annotation minimum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @param annoMin the new annotation minimum
 */
protected void setAnnoMin(double annoMin)
{
	this.annoMin = annoMin;
}

/**
 * Gets the annotation minimum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @return the current annotation minimum
 */
public double getAnnoMin()
{
	return(annoMin);
}

/**
 * Sets the annotation maximum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @param annoMax the new annotation maximum
 */
protected void setAnnoMax(double annoMax)
{
	this.annoMax = annoMax;
}

/**
 * Gets the annotation maximum.  This value is computed by Chart3d and is
 * read-only.
 *
 * @return the current annotation maximum
 */
public double getAnnoMax()
{
	return(annoMax);
}

/**
 * Sets the annotation step.  This value is computed by Chart3d and is
 * read-only.
 *
 * @param annoStep the new annotation step
 */
protected void setAnnoStep(double annoStep)
{
	this.annoStep = annoStep;
}

/**
 * Gets the annotation step.  This value is computed by Chart3d and is
 * read-only.
 *
 * @return the current annotation step
 */
public double getAnnoStep()
{
	return(annoStep);
}

/**
 * Sets the annotation easeway.  This value is computed by Chart3d and is
 * read-only.
 *
 * @param annoEaseway The new annotation easeway
 */
protected void setAnnoEaseway(double annoEaseway)
{
	this.annoEaseway = annoEaseway;
}

/**
 * Gets the annotation easeway.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @return the current annotation easeway
 */
public double getAnnoEaseway()
{
	return(annoEaseway);
}

/**
 * Gets the maximum annotation length for this axis (previously computed).
 *
 * @return the maximum annotation length for this axis
 */
protected int getMaxAnnoLength()
{
	return(maxAnnoLength);
}

/**
 * Sets the dataport minimum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @param dataportMin dataport minimum
 */
protected void setDataportMin(double dataportMin)
{
	this.dataportMin = dataportMin;
}

/**
 * Gets the dataport minimum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @return the current dataport minimum
 */
public double getDataportMin()
{
	return(dataportMin);
}

/**
 * Sets the dataport maximum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @param dataportMax dataport maximum
 */
protected void setDataportMax(double dataportMax)
{
	this.dataportMax = dataportMax;
}

/**
 * Gets the dataport maximum.  This value is computed by Chart3d and
 * should be viewed as read-only.
 *
 * @return the current dataport maximum
 */
public double getDataportMax()
{
	return(dataportMax);
}

/**
 * Returns the annotation handler for this axis.
 *
 * @return the annotation handler
 */
public Chart3dAnnoHandler getAnnoHandler()
{
	return(annoHandler);
}

/**
 * Ensures that the max value is greater than the min value.
 *
 * @param dataMin the default data min
 * @param dataMax the default data max
 */
protected void checkLimits(double dataMin, double dataMax)
{
	// First check if the user specified silly values and force them to
	// default values if this is true.
	if (max.value <= min.value && !max.isDefault) {
		max.isDefault = true;
		max.value     = dataMax;
	}
	if (max.value <= min.value && !min.isDefault) {
		min.isDefault = true;
		min.value     = dataMin;
	}

	// If all data points are equivalent values then find a min and max
	// that specify an interval around the data.
	if (max.value <= min.value) {
		double tprec = 0;
		while (dataMax <= dataMin) {
			dataMax  = dataMin + Math.pow(10.0, tprec);
			dataMin -= dataMax - dataMin;
			tprec += 1.0;
		}
		min.value = dataMin;
		max.value = dataMax;
	}
}

/**
 * Sets the extents of this axis.  Do checking to see whether the user wants
 * to specifiy the value, and then see if the values are acceptable.
 *
 * @param minValue the calculated (or default) minimum
 * @param maxValue the calculated (or default) maximum
 */
protected void setExtents(double minValue, double maxValue)
{
	min.setValue(minValue);
	max.setValue(maxValue);
	checkLimits(minValue, maxValue);
}

/**
 * Calculates the length of the longest annotation.  Just goes through the list
 * of labels and computes the longest label length.
 */
protected void calcMaxAnnoLength()
{
	ArrayList valueLabels = annoHandler.getValueLabels();

	maxAnnoLength = 0;
	for (int i = 0; i < valueLabels.size(); i++) {
		JCValueLabel vl = (JCValueLabel)valueLabels.get(i);
		double value = vl.getValue();
		if (value < annoMin || value > annoMax) {
			continue;
		}
		String label = vl.getLabel();
		if (label == null) {
			continue;
		}
		int labelLength = label.length();
		if (labelLength > maxAnnoLength) {
			maxAnnoLength = labelLength;
		}
	}
}

/**
 * Sets the value of the <code>LabelGenerator</code> property.  The
 * <code>LabelGenerator</code> property holds a reference to an object that
 * implements the <code>JCLabelGenerator</code> interface.  This interface is
 * used to externally generate labels.
 *
 * @param labelGenerator reference to object instance used to generate labels
 */
public void setLabelGenerator(JCLabelGenerator labelGenerator)
{
	if (this.labelGenerator == labelGenerator) {
		return;
	}
	this.labelGenerator = labelGenerator;

	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>LabelGenerator</code> property.  The
 * <code>LabelGenerator</code> property holds a reference to an object that
 * implements the <code>JCLabelGenerator</code> interface.  This interface is
 * used to externally generate labels.
 *
 * @return reference to object instance used to generate labels
 */
public JCLabelGenerator getLabelGenerator() {
	return(labelGenerator);
}

/**
 * Sets the value of the <code>combineLabels</code> property. If <code>true</code>,
 * consecutive data labels that are identical (that refer to the same object),
 * are combined into one label and the resulting label is positioned at the
 * average (or middle) label position.  Ticks are still drawn at each data
 * label position.
 *
 * @param combineLabels the new value of combineLabels
 */
public void setCombineLabels(boolean combineLabels)
{
	if (this.combineLabels == combineLabels) {
		return;
	}
	this.combineLabels = combineLabels;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Are data labels combined?  If <code>true</code>, consecutive data labels that are
 * identical (that refer to the same object), are combined into one label and
 * the resulting label is positioned at the average (or middle) label position.
 * Ticks are still drawn at each data label position.
 *
 * @return the current value of combineLabels
 */
public boolean isCombineLabels()
{
	return(combineLabels);
}

/**
 * Retrieves the value label for the specified value from the list of
 * user-specified value labels.
 *
 * @param value Data value for which to find the value label
 * @return JCValueLabel instance
 */
public JCValueLabel getValueLabel(double value)
{
    JCValueLabel ret = null;
	ArrayList userValueLabels = annoHandler.getUserValueLabels();
    for (int i = 0; i < userValueLabels.size(); i++) {
		JCValueLabel vl = (JCValueLabel)userValueLabels.get(i);
		if (value == vl.value) {
			ret = vl;
			break;
		}
    }
    return(ret);
}

/**
 * Retrieves the value label for the specified value from the list of
 * value labels generated for this axis.  If the <code>AnnotationMethod</code> is
 * set to <code>ANNOTATION_VALUE_LABELS</code>, this list will be the same as the
 * user-specified list of value labels.
 *
 * @param value data value for which to find the value label
 * @return JCValueLabel instance
 */
public JCValueLabel getGeneratedValueLabel(double value)
{
    JCValueLabel ret = null;
	ArrayList valueLabels = annoHandler.getValueLabels();
    for (int i = 0; i < valueLabels.size(); i++) {
		JCValueLabel vl = (JCValueLabel)valueLabels.get(i);
		if (value == vl.value) {
			ret = vl;
			break;
		}
    }
    return(ret);
}

/**
 * Adds the provided value label to the list of value labels. If the
 * <code>AnnotationMethod</code> is <code>ANNOTATION_VALUE_LABELS</code>, then
 * the value label will appear at the appropriate point along the axis.
 *
 * @param vl <code>JCValueLabel</code> instance to add
 */
public void addValueLabel(JCValueLabel vl)
{
    int duplicate = -1;
    int try_insert = -1;
    // If there is an existing label with the same value, replace it
	ArrayList userValueLabels = annoHandler.getUserValueLabels();
    for (int i = 0; i < userValueLabels.size(); i++) {
		JCValueLabel lbl = (JCValueLabel)userValueLabels.get(i);
		// We assume the list is sorted
		if (lbl.value > vl.value && try_insert < 0) {
			try_insert = i;
			break;
		}
		if (lbl.value == vl.value) {
			duplicate = i;
			break;
		}
    }
    if (duplicate >= 0) {
		userValueLabels.set(duplicate, vl);
    }
	else if (try_insert >= 0) {
		userValueLabels.add(try_insert, vl);
    }
	else {
		userValueLabels.add(vl);
    }
	annoHandler.sortValueLabels();
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>ValueLabels</code> property.
 * The <code>ValueLabels</code> property is an indexed property containing
 * a list of all annotations specified by the user for an axis.
 *
 * @return an array of user specified value labels
 */
public JCValueLabel[] getValueLabels()
{
	ArrayList userValueLabels = annoHandler.getUserValueLabels();
	JCValueLabel[] valueLabels = null;
	if (userValueLabels.size() == 0) {
		return(valueLabels);
	}
	valueLabels = new JCValueLabel[userValueLabels.size()];
	for (int i = 0; i < valueLabels.length; i++) {
		valueLabels[i] = (JCValueLabel)(userValueLabels.get(i));
	}
	return(valueLabels);
}

/**
 * Gets the list of value labels generated for this axis.  If the
 * <code>AnnotationMethod</code> is set to <code>ANNOTATION_VALUE_LABELS</code>,
 * this list will be the same as the user-specified list of value labels.
 *
 * @return an array of the internally generated value labels
 */
public JCValueLabel[] getGeneratedValueLabels()
{
	ArrayList genValueLabels = annoHandler.getValueLabels();
	JCValueLabel[] valueLabels = null;
	if (genValueLabels.size() == 0) {
		return(valueLabels);
	}
	valueLabels = new JCValueLabel[genValueLabels.size()];
	for (int i = 0; i < valueLabels.length; i++) {
		valueLabels[i] = (JCValueLabel)(genValueLabels.get(i));
	}
	return(valueLabels);
}

/**
 * Sets the value of the <code>ValueLabels</code> property.
 * The <code>ValueLabels</code> property is an indexed property containing
 * a list of all annotations specified by the user for an axis. <br><b>Note:</b>
 * The <code>AnnotationMethod</code> must be set to
 * <code>ANNOTATION_VALUE_LABELS</code> for these labels to be used.
 *
 * @param vl a list of value labels to be used on this axis
 */
public void setValueLabels(JCValueLabel[] vl)
{
    ArrayList userValueLabels = new ArrayList();
	if (vl != null) {
	    for (int i = 0; i < vl.length; i++) {
		    userValueLabels.add(vl[i]);
	    }
	}
	annoHandler.setUserValueLabels(userValueLabels);
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>ValueLabels</code> property at a specified index.
 * The <code>ValueLabels</code> property is an indexed property containing
 * a list of all annotations specified by the user for an axis.
 *
 * @param index the index of the user specified value label to be retrieved
 * @return the value label at the specified index; will be null on error
 */
public JCValueLabel getValueLabels(int index)
{
	ArrayList userValueLabels = annoHandler.getUserValueLabels();
	if (userValueLabels.size() == 0) {
		return(null);
	}
	JCValueLabel vl = null;
	try {
		vl = (JCValueLabel)userValueLabels.get(index);
	}
	catch (Exception e) {
	}
	return(vl);
}

/**
 * Gets the value label at the specified index in the list of value labels
 * generated for this axis.  If the <code>AnnotationMethod</code> is set to
 * <code>ANNOTATION_VALUE_LABELS</code>, this list will be the same as the
 * user-specified list of value labels.
 *
 * @param index the index of the generated value label to be retrieved
 * @return the value label at the specified index; will be null on error
 */
public JCValueLabel getGeneratedValueLabels(int index)
{
	ArrayList valueLabels = annoHandler.getValueLabels();
	if (valueLabels.size() == 0) {
		return(null);
	}
	JCValueLabel vl = null;
	try {
		vl = (JCValueLabel)valueLabels.get(index);
	}
	catch (Exception e) {
	}
	return(vl);
}

/**
 * Sets the value of one one value label in the <code>ValueLabels</code> property.
 * The <code>ValueLabels</code> property is an indexed property containing
 * a list of all annotations specified by the user for an axis.
 * <br><B>Note:</b> This method is required by the BeanBox, but is somewhat
 * ill-formed. The index specified will not
 * necessarily be honoured because the <code>ValueLabel</code> list
 * is sorted by value. It is not recommended for use.
 *
 * @param index the index of the value label to be set
 * @param vl the new value label at the specified index
 */
public void setValueLabels(int index, JCValueLabel vl)
{
	ArrayList userValueLabels = annoHandler.getUserValueLabels();
	if (index < 0 || index > userValueLabels.size()) {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.BAD_DATA_OR_INDEX_ERROR));
	}
	if (index < userValueLabels.size()) {
		userValueLabels.set(index, vl);
	}
	else {
		userValueLabels.add(vl);
	}
	annoHandler.sortValueLabels();

	callParentSetChanged(true, Changeable.RECALC);
}

}
