/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChart3d.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.event.JCChart3dEvent;
import com.klg.jclass.chart3d.event.JCChart3dListener;
import com.klg.jclass.chart3d.event.JCPick3dEvent;
import com.klg.jclass.chart3d.event.JCPick3dListener;
import com.klg.jclass.chart3d.resources.LocaleBundle;
import com.klg.jclass.util.*;
import com.klg.jclass.util.legend.JCLegend;
import com.klg.jclass.util.legend.JCLegendItem;
import com.klg.jclass.util.legend.LegendComponentLayout;
import com.klg.jclass.util.legend.LegendComponentLayoutUser;

import javax.swing.*;
import javax.swing.plaf.ComponentUI;
import javax.vecmath.Point3d;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.util.*;
import java.util.List;

/**
 * The core charting class. <code>JCChart3d</code> allows the
 * retrieval of related objects, such as the chart area, the
 * legend, the header, and the footer. Bean classes are represented
 * by a charting class derived from <code>JCChart3d</code>.
 *
 */
public abstract class JCChart3d extends JComponent
                                implements Changeable, java.io.Serializable,
                                           LegendComponentLayoutUser,
                                           ServerRenderable,
	                                       KeyListener, MouseListener,
	                                       MouseMotionListener,
                                           Chart3dGridDataModel,
	                                       Chart3dPointDataModel
{

/**
 * The UI class ID.
 * @see #getUIClassID
 * @see #writeObject
 */
private static final String uiClassID = "JCChart3dUI";

/** The component that manages the area where the chart is drawn. */
protected JCChart3dArea             chart3dArea   = null;

/** A list of <code>dataViews</code> for this chart.
 * Currently only one <code>dataView</code> is supported.
 */
protected Vector                    dataViews     = new Vector();

/** The legend for this chart. */
protected JCLegend                  legend        = null;

/** The default implementation of the legend populator and renderer. */
protected JCChart3dLegendManager    legendManager = null;

/** The footer for this chart. Any <code>JComponent</code> can be used. */
protected JComponent                footer        = null;

/** The header for this chart.  Any <code>JComponent</code> can be used. */
protected JComponent                header        = null;

/** The current list of <code>JCChart3dListener</code>'s for this chart. */
protected JCListenerList            listeners     = null;

/** The current list of <code>JCPick3dListener</code>'s for this chart. */
protected JCListenerList            pickListeners = null;

// Ensures that chartArea responds to our actions/events
private int flip = 1;

private boolean  warningOn        = true;
private boolean  batched          = false;
private boolean  allowUserChanges = false;

/** Asks if a key press has been initiated without a corresponding release. */
protected boolean inKeyPress = false;

// Chart Types

/** Invalid chart type used for error checking. */
public final static int  NONE                       = -1;

/** Specifes that a surface should be drawn. */
public final static int  SURFACE                    =  0;

/** Specifies that an array of bars should be drawn. */
public final static int  BAR                        =  1;

/** Specifies that a scatter plot should be drawn. */
public final static int  SCATTER                    =  2;

/** Specifies the end of array so more chart can possibly be added later. */
public final static int  LAST_PREDEFINED_CHART_TYPE =  3;

/** Error message for customizer failure. */
public final static String ERRMSG_CUSTOMIZER_FAILURE =
	"Could not initialize Customizer.";

/**
 * An index into the color list.  Specifies that the next requested color
 * ia different.
 */
protected int lineColorIndex;

/**
 * An index into the symbol list.  Specifies that the next requested symbol
 * is different.
 */
protected int symbolColorIndex;

/**
 * An index into the color list.  Specifies that the next requested fill color
 * is different.
 */
protected int fillColorIndex;

/**
 * An index into the shape list.  Specifies that the next requested symbol
 * shape is different.
 */
protected int symbolShapeIndex;

private Dimension curSize = new Dimension(-1, -1);

/**
 * Gives the layout manager info about the position and size of the
 * chart area.
 */
protected Rectangle chart3dAreaLayoutHints = new Rectangle(
	Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);

/** Gives the layout manager info about the position and size of the legend. */
protected Rectangle legendLayoutHints = new Rectangle(
	Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);

/** Gives the layout manager info about the position and size of the header. */
protected Rectangle headerLayoutHints = new Rectangle(
	Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);

/** Give the layout manager info about the position and size of the footer. */
protected Rectangle footerLayoutHints = new Rectangle(
	Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);

// Has this chart been changed?
private boolean changed = false;

/** Temporary Graphics object that draws without showing to screen. */
private transient Graphics prevSetGraphics = null;

/** The ActionTable object that associates key and mouse triggers with actions. */
protected JCActionTable actionTable;

/** References the Customizer, if it exists. */
public transient JCChart3dCustomizer customizer = null;

/**
 * The locale object for this <code>chart3d</code>.
 */
protected Locale locale = Locale.getDefault();

/**
 * Initializes assorted chart member variables. Used as a
 * convenience when writing a new <code>JCChart3d</code> constructor.
 *
 * @param chartType the initial chart type of the chart
 */
protected void init(int chartType)
{
	setOpaque(true);

	initializeLayout();

	createHeader();
	createFooter();
	createLegend();

	resetDefaults();
	initializeUI();

	createActionTable();
}

/**
 * Creates an instance of the Java 2d version of JClass Chart 3D, that is
 * the subclass of <code>JCChart3d</code> that implements three-dimensional
 * chart functionality using the Java 2d API.
 * @return an instance of com.klg.jclass.chart3d.j2d.JCChart3dJava2d
 *         or null if there are any problems instantiating the class.
 */
public static JCChart3d createJava2dChart()
{
	JCChart3d chart3dInstance = null;
	try {
		Class java2dClass = Class.forName(
			"com.klg.jclass.chart3d.j2d.JCChart3dJava2d");
		chart3dInstance = (JCChart3d) java2dClass.newInstance();
	}
    catch (Exception e) {
		chart3dInstance = null;
	}
	return(chart3dInstance);
}

/**
 * Creates an instance of the Java 3d version of JClass Chart 3D, that is
 * the subclass of <code>JCChart3d</code> that implements three-dimensional
 * chart functionality using the Java 3d API.
 * @return an instance of com.klg.jclass.chart3d.j3d.JCChart3dJava3d
 *         or null if there are any problems instantiating the class.
 */
public static JCChart3d createJava3dChart()
{
	return(createJava3dChart(false));
}

/**
 * Creates an instance of the Java 3d version of JClass Chart 3D, that is
 * the subclass of <code>JCChart3d</code> that implements three-dimensional
 * chart functionality using the Java 3d API.  If there are any problems
 * creating this class (for example, Java 3d is not available) and the fallback
 * parameter is set to <code>true</code>, this method will revert to creating
 * an instance of the Java 2d version of JClass Chart 3d.
 * @param fallback if <code>true</code>, fall back to using the Java 2d version
 *        of JClass Chart 3D when there are problems using the Java 3d version.
 * @return an instance of com.klg.jclass.chart3d.j3d.JCChart3dJava3d, if
 *         possible.  Otherwise, either null or an instance of
 *         com.klg.jclass.chart3d.j2d.JCChart3dJava2d, depending on the
 *         value of the fallback parameter.
 */
public static JCChart3d createJava3dChart(boolean fallback)
{
	JCChart3d chart3dInstance = null;
	try {
		Class java3dClass = Class.forName(
			"com.klg.jclass.chart3d.j3d.JCChart3dJava3d");
		chart3dInstance = (JCChart3d) java3dClass.newInstance();
	}
    // Catch both errors and exceptions.
    // Catch both errors and exceptions.
    // Missing Java3d might throw a NoClassDefFoundError.
    // Missing the JCChart3dJava3d class might thros a ClassNotFoundException.
    catch (Throwable e) {
		// Fallback to java 2d if requested
		if (fallback) {
			chart3dInstance = createJava2dChart();
		}
        else {
			chart3dInstance = null;
		}
	}
	return(chart3dInstance);
}

/**
 * Creates the action table for the chart.  This method must be overridden
 * by subclasses.
 */
protected abstract void createActionTable();

/**
 * Sets the initial layout.
 */
protected void initializeLayout()
{
	super.setLayout(new LegendComponentLayout());
}

/**
 * Registers the user interface.
 */
protected void initializeUI()
{
	enableEvents(AWTEvent.KEY_EVENT_MASK);
	enableEvents(AWTEvent.MOUSE_EVENT_MASK);
	enableEvents(AWTEvent.MOUSE_MOTION_EVENT_MASK);

	registerUI();
	updateUI();
}

/**
 * Creates and adds a new <code>JLabel</code> header to the chart.
 */
protected void createHeader()
{
	header = new JLabel();
	header.setVisible(false);
	add(header);
}

/**
 * Creates and adds a new <code>JLabel</code> footer to the chart.
 */
protected void createFooter()
{
	footer = new JLabel();
	footer.setVisible(false);
	add(footer);
}

/**
 * Creates and adds a legend to the chart.
 */
protected void createLegend()
{
	legend = new JCChart3dLegend();
	legendManager = new JCChart3dLegendManager(this, legend);
	legend.setLegendPopulator(legendManager);
	legend.setLegendRenderer(legendManager);
	add(legend);
}

/**
 * Creates and adds a <code>JCChart3dArea</code> to chart.
 *
 */
protected void createChart3dArea()
{
	chart3dArea = new JCChart3dArea(this);
	add(chart3dArea);
}

/**
 * The <code>JCActionTable</code> object that associates key and mouse triggers
 * with actions.
 *
 * @return Chart3d's <code>JCActionTable</code>
 * @see com.klg.jclass.chart3d.JCActionTable
 */
public JCActionTable getActionTable() {
    return actionTable;
}

/**
 * Asks if the surface is represented by the first <code>dataView</code> is a
 * 3d view or a 2d projection.
 *
 * @return a boolean indicating whether the first dataView is a 2D projection
 *   or not
 */
public boolean isProjection()
{
	Chart3dDataView dv = (Chart3dDataView)(dataViews.elementAt(0));
	JCElevation elevation = dv.getElevation();
	return((dv.getChartType() != JCChart3d.BAR &&
	       (!elevation.isMeshed() && !elevation.isShaded())));
}

/**
 * Processes mouse click events on the chart.
 *
 * @param e the mouse event to be processed
 */
protected void processMouseEvent(MouseEvent e)
{
	super.processMouseEvent(e);

	if (e.isConsumed()) {
		return;
	}

	int id = e.getID();
	switch(id) {
	case MouseEvent.MOUSE_PRESSED:
		mousePressed(e);
		break;
	case MouseEvent.MOUSE_RELEASED:
		mouseReleased(e);
		break;
	case MouseEvent.MOUSE_CLICKED:
		mouseClicked(e);
		break;
	case MouseEvent.MOUSE_EXITED:
		mouseExited(e);
		break;
	case MouseEvent.MOUSE_ENTERED:
		mouseEntered(e);
		break;
	}
}

/**
 * Processes mouse motion events on the chart.
 *
 * @param e the mouse motion event to be processed
 */
protected void processMouseMotionEvent(MouseEvent e)
{
	super.processMouseMotionEvent(e);

	if(e.isConsumed()) {
		return;
	}

	int id = e.getID();
	switch(id) {
	case MouseEvent.MOUSE_MOVED:
		mouseMoved(e);
		break;
	case MouseEvent.MOUSE_DRAGGED:
		mouseDragged(e);
		break;
	}
}

/** Processes key events on the chart.
 *
 * @param e the key event to be processed
 */
protected void processKeyEvent(KeyEvent e)
{
	super.processKeyEvent(e);

	int id = e.getID();
	switch(id) {
	case KeyEvent.KEY_PRESSED:
		keyPressed(e);
		break;
	case KeyEvent.KEY_RELEASED:
		keyReleased(e);
		break;
	case KeyEvent.KEY_TYPED:
		keyTyped(e);
		break;
	}
}

public int convertXToChartAreaX(int x)
{
    return(x - chart3dArea.getLocation().x);
}

public int convertYToChartAreaY(int y)
{
    return(y - chart3dArea.getLocation().y);
}

public int convertChartAreaXToX(int x)
{
    return(x + chart3dArea.getLocation().x);
}

public int convertChartAreaYToY(int y)
{
    return(y + chart3dArea.getLocation().y);
}

/**
 * Retrieves the action handler that is associated with a given
 * <code>InputEvent</code>. This method must be overridden by subclasses.
 * @param event the <code>InputEvent</code>
 * @return the action handler associated with the event
 */
protected abstract JCAction getActionHandler(InputEvent event);

/**
 * Processes <code>mousePressed</code> events for the chart; requests focus and
 * performs associated actions.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     click
 */
public void mousePressed(MouseEvent event)
{
    // Need this, otherwise key actions will never work
    if (isFocusTraversable()) {
        requestFocus();
    }
    JCAction actionHandler = getActionHandler(event);
    if (actionHandler == null) {
        return;
    }
    chart3dArea.setActionHandler(actionHandler);
    chart3dArea.setInAction(true);
	chart3dArea.startAction(event, convertXToChartAreaX(event.getX()),
	                        convertYToChartAreaY(event.getY()));
	
}

/**
 * Processes <code>mouseClicked</code> events for the chart; does nothing.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     event
 */
public void mouseClicked(MouseEvent event) {}

/**
 * Processes <code>mouseReleased</code> events for the chart; finishes
 * associated actions.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     event
 */
public void mouseReleased(MouseEvent event)
{
	
	chart3dArea.endAction(event, convertXToChartAreaX(event.getX()),
	                      convertYToChartAreaY(event.getY()));
    chart3dArea.setInAction(false);
}

/**
 * Processes <code>mouseEntered</code> events for the chart; does nothing.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     event
 */
public void mouseEntered(MouseEvent event) {}

/**
 * Processes <code>mouseExited</code> events for the chart; does nothing.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     event
 */
public void mouseExited(MouseEvent event) {}

/**
 * Processes <code>mouseDragged</code> events for the chart; performs
 * associated actions.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     event
 */
public void mouseDragged(MouseEvent event)
{
	
	chart3dArea.animateAction(event, convertXToChartAreaX(event.getX()),
	                          convertYToChartAreaY(event.getY()));
}

/**
 * Processes <code>mouseMoved</code> events for the chart; does nothing.
 *
 * @param event the <code>MouseEvent</code> instance that represents this mouse
 *     event
 */
public void mouseMoved(MouseEvent event) {}

/**
 * Processes <code>keyTyped</code> events for the chart; does nothing.
 *
 * @param event the <code>KeyEvent</code> instance that represents this
 *     keyboard event
 */
public void keyTyped(KeyEvent event) {}

/**
 * Processes <code>keyPressed</code> events for the chart; resets chart or
 * cancels action as specified.
 *
 * @param event the <code>KeyEvent</code> instance that represents this
 *     keyboard event
 */
public void keyPressed(KeyEvent event)
{
    JCAction actionHandler = getActionHandler(event);
    if (actionHandler == null) {
        return;
    }
//  Fix?
//  if (chart3dArea.isInAction()) {
//      return;
//  }
    if (!inKeyPress) {
        inKeyPress = true;
        chart3dArea.setActionHandler(actionHandler);
	    
	    chart3dArea.startAction(event, 0, 0);
        flip = 1 - flip;
    }
    chart3dArea.setActionHandler(actionHandler);
    
    chart3dArea.animateAction(event, flip, flip);
    flip = 1 - flip;
}

/**
 * Processes <code>keyReleased</code> events for the chart; does nothing.
 *
 * @param event the <code>KeyEvent</code> instance that represents this
 *     keyboard event
 */
public void keyReleased(KeyEvent event)
{
	
//  Fix?
//  if (chart3dArea.isInAction()) {
//      return;
//  }
    inKeyPress = false;
	chart3dArea.endAction(event, flip, flip);
}

/**
 * Launches Customizer.
 *
 * @param p the point at which to launch the Customizer
 */
public void launchPropertyPage(Point p)
{
	
	if (!allowUserChanges) {
	    return;
	}

	if (customizer == null) {
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		try {
			Class newclass = Class.forName(getCustomizerName());
			customizer = (JCChart3dCustomizer)newclass.newInstance();
		}
		catch (Exception e) {
			e.printStackTrace(System.out);
			customizer = null;
			ErrorDialog.raise(this, JCChart3d.ERRMSG_CUSTOMIZER_FAILURE +
			                  "\n" + e);
		}
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}
	if (customizer != null) {
		customizer.setTarget(this);
		SwingUtilities.convertPointToScreen(p, this);
		customizer.showCustomizer(p);
	}
}

/**
 * Constructor that creates a <code>JCChart3d</code> of a particular type.
 * <br><b>Note:</b> The <code>ChartType</code> is actually a property of
 * <code>Chart3dDataView</code>. This constructor is provided as a convenience.
 *
 * @param tp the chart type
 * @see com.klg.jclass.chart3d.Chart3dDataView
 */
public JCChart3d(int tp)
{
	super();
	init(tp);
}

/**
 * Default constructor, required by JavaBeans.  Chart type defaults to
 * <code>SURFACE</code>.
 */
public JCChart3d()
{
	super();
	try {
		init(SURFACE);
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
}

/**
 * Static function to allow creation of a chart from a serialized file. The
 * serialized file can be created by using the JClass Chart 3D Customizer's
 * <i>Save As</i> option.
 *
 * @param fileName the name of the file from which to load
 * @return the new <code>JCChart3d</code> instance
 */
public static JCChart3d makeChart3d(String fileName)
{
	JCChart3d c = null;
	try {
		FileInputStream fis = new FileInputStream(fileName);
		ObjectInputStream ois = new ObjectInputStream(fis);
		c = (JCChart3d)ois.readObject();
		ois.close();
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
		c = null;
	}
	return(c);
}

/**
 * Gets the value of the <code>Changed</code> property, which determines whether
 * the chart requires a recalculation.
 * @return the current value of the <code>Changed</code> property
 */
final public boolean isChanged()
{
	return(getChanged());
}

/**
 * Checks the value of the <code>ChangedFlag</code> property for the specified
 * mask (see {@link Changeable} for a list of valid masks). The
 * <code>ChangedFlag</code> property determines the type of action that the
 * chart needs to take regarding the current change.
 *
 * @param mask the mask to check against
 * @return the current value of the <code>Changed</code> property, based on the
 *     mask
 */
final public boolean isChanged(int mask)
{
	return(getChanged(mask));
}

/**
 * Gets the value of the <code>Changed</code> property, which determines
 * whether the chart requires a recalculation.
 * @return Does the chart need a racalc?
 */
public boolean getChanged()
{
	// Chart3d is changed if one of its subcomponents has changed
	return(chart3dArea.isChanged()                  ||
		   legend.isChanged()                       ||
		   !chart3dArea.isValid()                   ||
		   !legend.isValid()                        ||
		   (header != null && !header.isValid())    ||
		   (footer != null && !footer.isValid()));
}

/**
 * Checks the value of the <code>ChangedFlag</code> property for the specified
 * mask (see {@link Changeable} for a list of valid masks). The
 * <code>ChangedFlag</code> property determines the type of action the chart
 * needs to take regarding the current change.
 *
 * @param mask the mask to check against
 * @return Does the chart need a racalc based on the mask?
 */
public boolean getChanged(int mask)
{
	// Chart3d is changed if one of its subcomponents has changed
	return(chart3dArea.isChanged(mask)              ||
		   legend.isChanged(mask)                   ||
		   !chart3dArea.isValid()                   ||
		   !legend.isValid()                        ||
		   (header != null && !header.isValid())    ||
		   (footer != null && !footer.isValid()));
}

/**
 * Recalculates the entire chart, if it has been marked for recalculation.
 */
public void recalc()
{
	if (!isChanged(Changeable.RECALC)) {
		return;
	}

	try {
		// Recalc chart area first so that any new drawables will be created
		// in case the legend recalc needs them.
		chart3dArea.recalc();

		// Recalc legend here in case no layout manager is being used.
		legend.recalc();
	}
	catch (Exception e) {
		// Don't fail as a result of child exceptions.
		e.printStackTrace(System.out);
	}
	setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Sets the value of the <code>Changed</code> property, which determines
 * whether the chart requires recalculation. This value is propagated to all
 * child components.
 *
 * @param n boolean value of the <code>Changed</code> property
 * @param cf the changed flag mask to check against
 */
public void setChanged(boolean n, int cf) {
	if (n == changed) {
		return;
	}
	changed = n;
	if (changed) {
		int legendCF = Changeable.NO_CHANGE;
		if ((cf & Changeable.NOTIFY_LEGEND_RECALC) > 0) {
			legendCF |= Changeable.RECALC;
		}
		if ((cf & Changeable.NOTIFY_LEGEND_REDRAW) > 0) {
			legendCF |= Changeable.REDRAW;
		}
		if (legendCF != Changeable.NO_CHANGE) {
			legend.setChanged(true, legendCF, false);
		}

		int caCF = Changeable.NO_CHANGE;
		if ((cf & Changeable.LAYOUT) > 0) {
			caCF |= Changeable.RECALC;
		}
		if ((cf & Changeable.NOTIFY_CA_RECALC) > 0) {
			caCF |= Changeable.RECALC;
		}
		if ((cf & Changeable.NOTIFY_CA_REDRAW) > 0) {
			caCF |= Changeable.REDRAW;
		}
		if (caCF != Changeable.NO_CHANGE) {
			chart3dArea.setChanged(true, caCF, false);
		}

		if (isBatched()) {
			invalidate();
		}
		else if (((cf & Changeable.RECALC) > 0) ||
				 ((cf & Changeable.LAYOUT) > 0))
		{
			update();
		}
		else if ((cf & Changeable.REDRAW) > 0) {
			repaint();
		}
	}
}

/**
 * Sets the value of the <code>Changed</code> property, which determines
 * whether the chart requires recalculation. This value is propagated to all
 * the child components. <br><b>Note:</b> This method is mandated by the
 * {@link Changeable} interface, but has no use in <code>JCChart3d</code>.
 * It is usually used for cases in which the <code>Changed</code> property must
 * be changed without forcing a recalculation.
 *
 * @param n boolean value of the <code>Changed</code> property
 * @param cf the changed flag mask to check against
 * @param updateParent since this is the final parent, call
 *     <code>setChanged</code> if <code>true</code> and store the fact that a
 *     change is needed if <code>false</code>
 */
public void setChanged(boolean n, int cf, boolean updateParent)
{
	if (updateParent) {
		setChanged(n, cf);
	}
	else {
		changed = n;
	}
}

/**
 * For internal use only.  Called by the <code>setChanged()</code> method of
 * <code>JCChart3dArea</code>. The Java3D version of <code>JCChart3d</code>
 * overrides this method to call <code>update()</code>.
 */
public void redraw()
{
	repaint();
}

/**
 * Overrides the superclass's <code>paint()</code> method so that the chart is
 * fully painted and flushed before the listeners are called.
 *
 * @param gc the graphics used for painting
 */
public void paint(java.awt.Graphics gc)
{
	super.paint(gc);
	firePaint();
}

/**
 * Fires a paint event to interested listeners.
 */
public void firePaint()
{
	Enumeration e = JCListenerList.elements(listeners);
	while (e.hasMoreElements()) {
		((JCChart3dListener)e.nextElement()).paintChart(this);
	}
}

/**
 * Gets the drawing area represented by this chart.
 *
 * @return rectangle object containing drawing area
 */
public Rectangle getDrawingArea()
{
	Rectangle rect = new Rectangle();
	getDrawingArea(rect);
	return(rect);
}

/**
 * Gets the drawing area represented by this chart and places it in the
 * provided rectangle.
 *
 * @param rect rectangle object to fill, with drawing area coordinates
 */
public void getDrawingArea(Rectangle rect)
{
	Insets insets = getInsets();
	if (insets == null) {
		insets = new Insets(0, 0, 0, 0);
	}
	rect.setBounds(insets.left, insets.top,
				   Math.max(0, getSize().width - (insets.left + insets.right)),
				   Math.max(0, getSize().height - (insets.top + insets.bottom)));
}

/**
 * Gets the height of the drawing area represented by this chart.
 *
 * @return the height of the drawing area
 */
public int getDrawingAreaHeight()
{
	Insets insets = getInsets();
	int k = 0;
	if (insets != null) {
		k = insets.top + insets.bottom;
	}
	return(Math.max(0, getSize().height - k));
}

/**
 * Gets the width of the drawing area represented by this chart.
 *
 * @return the width of the drawing area
 */
public int getDrawingAreaWidth()
{
	Insets insets = getInsets();
	int k = 0;
	if (insets != null) {
		k = insets.left + insets.right;
	}
	return(Math.max(0, getSize().width - k));
}

/**
 * The top-level paint routine in <code>JCChart3d</code>.
 *
 * @param gc the graphics object used to paint
 */
public void paintComponent(java.awt.Graphics gc)
{
	if (gc == null) {
		return;
	}
	Graphics newgc = gc.create();

	// Draw our own opaque backgrounds.  Shouldn't this be
	// done by the superclass?  That code appears to be in the UI, which
	// is null for this component it seems.
	if (isOpaque()) {
		if (!newgc.getColor().equals(getBackground())) {
			newgc.setColor(getBackground());
		}
		Rectangle r = getBounds();
		newgc.fillRect(0, 0, r.width, r.height);
	}

	if (newgc != null) {
		newgc.dispose();
	}

	setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Prints this component and all of its subcomponents.
 * Overridden from <code>java.awt.Component</code>, but should be used
 * in the same way.
 *
 * @param g the graphics object used to paint
 */
public void printAll(Graphics g)
{
	super.printAll(g);
}

/**
 * <code>JCChart3d</code> layout method. It lays out the header, footer, legend,
 * and chart area.
 */
public void doLayout()
{
	try {
		// Trap all layout errors
		performLayout();
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
}

/**
 * Actual layout method.
 */
private void performLayout()
{
	Dimension sz = new Dimension();

	sz.width = getDrawingAreaWidth();
	sz.height = getDrawingAreaHeight();

	if (sz.width <= 0 || sz.height <= 0) {
		return;
	}

    boolean sizeChanged = (sz.width != curSize.width ||
                           sz.height != curSize.height);

	// If the size of the chart has not changed, layout should only be
	// done if the header, footer, or legend has changed.  (This is because
	// the chart area always ends up with the leftover space.)  However if
	// the chart area has changed, we still need to call its recalc method to
	// force its redraw.
	if (!chart3dArea.isChanged(Changeable.NOTIFY_PARENT_LAYOUT) &&
		chart3dArea.isValid() &&
		!sizeChanged)
	{
		if (!legend.isChanged(Changeable.RECALC) &&
			!legend.isChanged(Changeable.NOTIFY_PARENT_LAYOUT) &&
			legend.isValid() &&
			(header != null && header.isValid()) &&
			(footer != null && footer.isValid()))
		{
			if (chart3dArea.isChanged(Changeable.RECALC) ||
				!chart3dArea.isValid())
			{
				recalc();
			}
			repaint();
			return;
		}
	}

    // If the size changes, recalc everything.
    if (sizeChanged) {
        setChanged(true, Changeable.ALL);
    }

	recalc();
	curSize = sz;

	// Now that we've decided to actually layout and all the objects have
	// been calced, do the layout.
	super.doLayout();

	repaint();
}

/**
 * Overrides <code>java.awt.Component setLayout()</code>.
 *
 * @param mgr the layout manager to be used for doing the layout
 */
public void setLayout(LayoutManager mgr)
{
	boolean doSet = true;

	if (doSet) {
		super.setLayout(mgr);
	}
}

/**
 * Sets layout hints for chart children. Hints are rectangle objects.
 * A value of <code>Integer.MAX_VALUE</code> in the rectangle's members
 * indicates to calculate default values during layout. Other values
 * indicate to the layout to use that value. For example, a rectangle
 * with members <i>x=5</i>, <i>y=10</i>, <i>width=MAX_VALUE</i>, and
 * <i>height=200</i>, would indicate to the layout mechanism that the chart
 * child should be placed at <i>(5,10)</i>, have a height of 200, and use the
 * default width.
 *
 * <P>Layout hints are only used by the <code>DefaultChartLayout</code>
 * layout manager.
 *
 * @param child chart child - either the <code>chart3dArea</code>, legend,
 *     header, or footer
 * @param layoutHints rectangle object containing the desired layout hints
 */
public void setLayoutHints(JComponent child, Rectangle layoutHints)
{
	if (child == null) {
		return;
	}

	if (child == chart3dArea) {
		chart3dAreaLayoutHints = new Rectangle(layoutHints.x, layoutHints.y,
											   layoutHints.width,
											   layoutHints.height);
	}
	else if (child == legend) {
		legendLayoutHints = new Rectangle(layoutHints.x, layoutHints.y,
										  layoutHints.width,
										  layoutHints.height);
	}
	else if (child == header) {
		headerLayoutHints = new Rectangle(layoutHints.x, layoutHints.y,
										  layoutHints.width,
										  layoutHints.height);
	}
	else if (child == footer) {
		footerLayoutHints = new Rectangle(layoutHints.x, layoutHints.y,
										  layoutHints.width,
										  layoutHints.height);
	}

	child.invalidate();
	validate();
}

/**
 * Gets layout hints for the specified chart child.
 *
 * @param child chart child - either the <code>chart3dArea</code>, legend,
 *     header, or footer
 * @return Rectangle object containing the desired layout hints.  Null
 *     child is not one of the valid children of chart.
 */
public Rectangle getLayoutHints(JComponent child)
{
	if (child == null) {
		return(new Rectangle(Integer.MAX_VALUE, Integer.MAX_VALUE,
							 Integer.MAX_VALUE, Integer.MAX_VALUE));
	}
	else if (child == chart3dArea) {
		return(chart3dAreaLayoutHints);
	}
	else if (child == legend) {
		return(legendLayoutHints);
	}
	else if (child == header) {
		return(headerLayoutHints);
	}
	else if (child == footer) {
		return(footerLayoutHints);
	}
	else {
		return(new Rectangle(Integer.MAX_VALUE, Integer.MAX_VALUE,
							 Integer.MAX_VALUE, Integer.MAX_VALUE));
	}
}

/**
 * Sets the value of the <code>Batched</code> property, which controls whether
 * chart updates are accumulated.
 *
 * @param bt if <code>true</code>, it will accumulate chart updates; if
 *     <code>false</code>, it forces the accumulated updates to be processed
 */
public void setBatched(boolean bt)
{
	if (batched == bt) {
		return;
	}
	batched = bt;

	if (!bt) {
		update();
	}
}

/**
 * Gets the value of the <code>Batched</code> property.
 *
 * @return <code>true</code> if updates are being accumulated; if
 *     <code>false</code>, updates are not accumulated
 */
public boolean isBatched()
{
	return(batched);
}

/**
 * Forces the chart to re-layout and recalculate.
 */
public void update()
{
	invalidate();
	if (!isShowing()) {
		return;
	}
	validate();
}

/**
 * Overrides <code>javax.swing.JComponent addNotify()</code>.
 */
public void addNotify()
{
	super.addNotify();

 
    doLayout();
}

/**
 * Overrides the superclass's method so that changes will generate redraws.
 *
 * @param c the new background color
 */
public void setBackground(Color c)
{
	super.setBackground(c);
	setChanged(true, Changeable.RECALC | Changeable.NOTIFY_CA_RECALC);
}

/**
 * Overrides the superclass's method so that changes will generate redraws.
 *
 * @param c the new foreground color
 */
public void setForeground(Color c)
{
	super.setForeground(c);
	setChanged(true, Changeable.RECALC |
					 Changeable.NOTIFY_CA_RECALC |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Overrides the superclass's method so that changes will generate redraws.
 *
 * @param f the new chart font
 */
public void setFont(Font f)
{
	super.setFont(f);
	setChanged(true, Changeable.RECALC |
					 Changeable.NOTIFY_CA_RECALC |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Overrides the superclass's method so that changes will generate redraws.
 *
 * @param op the new value of opaque
 */
public void setOpaque(boolean op)
{
	super.setOpaque(op);
	setChanged(true, Changeable.REDRAW);
}

/**
 * Sets the value of the <code>AllowUserChanges</code> property, which
 * determines whether the user viewing the chart can modify chart values. Used
 * to allow edits to values and changes to parameters via the Customizer.
 *
 * @param change if <code>true</code>, the user is allowed to modify graph
 *     values; if <code>false</code>, the user is not allowed to modify graph
 *     values
 */
public void setAllowUserChanges(boolean change)
{
	allowUserChanges = change;
}

/**
 * Gets the value of the <code>AllowUserChanges</code> property, which
 * determines whether the user viewing the chart can modify chart values. Used
 * to allow edits to values and changes to parameters via the Customizer.
 *
 * @return if <code>true</code>, the user is allowed to modify the graph values;
 *     <code>false</code> otherwise
 */
public boolean getAllowUserChanges()
{
	return(allowUserChanges);
}

/**
 * Sets the value of the <code>WarningDialog</code> property, which controls
 * whether a dialog will appear when the chart has warning messages.
 *
 * @param on if <code>true</code> (default), the warning dialog will appear; if
 *     <code>false</code>, the user will not be warned of anything
 */
public void setWarningDialog(boolean on)
{
	if (warningOn == on) {
		return;
	}
	warningOn = on;
}

/**
 * Gets the value of the <code>WarningDialog</code> property.
 *
 * @return if <code>true</code> (default), the warning dialog will appear;
 *     if <code>false</code>, the user will not be warned of anything
 */
public boolean isWarningDialog()
{
	return(warningOn);
}


/**
 * Gets the <code>Header</code> property.
 * The <code>Header</code> property controls the object which controls the
 * display of the header.
 *
 * @return The <code>JComponent</code> object associated with the header.  By
 *     default it is an instance of <code>JLabel</code>.
 */
public JComponent getHeader()
{
	return(header);
}

/**
 * Sets the <code>Header</code> property.
 * The <code>Header</code> property controls the object that controls the
 * display of the header.
 *
 * @param head the <code>JComponent</code> object associated with the header
 */
public void setHeader(JComponent head)
{
	if (head == null) {
		return;
	}
	if (header != null) {
		remove(header);
	}
	header = head;
	add(header);
	setChanged(true, Changeable.LAYOUT |
					 Changeable.NOTIFY_CA_RECALC |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Sets the <code>Footer</code> property.
 * The <code>Footer</code> property controls the object that controls the
 * display of the footer.
 *
 * @return The <code>JComponent</code> object associated with the footer.  By
 *     default it is an instance of <code>JLabel</code>.
 */
public JComponent getFooter()
{
	return(footer);
}

/**
 * Sets the <code>Footer</code> property.
 * The <code>Footer</code> property controls the object that controls the
 * display of the footer.
 *
 * @param foot the <code>JComponent</code> object associated with the footer
 */
public void setFooter(JComponent foot)
{
	if (foot == null) {
		return;
	}
	if (footer != null) {
		remove(footer);
	}
	footer = foot;
	add(footer);
	setChanged(true, Changeable.LAYOUT |
					 Changeable.NOTIFY_CA_RECALC |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the <code>Legend</code> property.
 * The <code>Legend</code> property controls the object that controls
 * the display of the legend.
 *
 * @return the <code>JCLegend</code> object associated with the legend
 */
public JCLegend getLegend()
{
	return(legend);
}

/**
 * Sets the <code>Legend</code> property.
 * The <code>Legend</code> property controls the object that controls
 * the display of the legend. If the legend has no populator
 * or renderer set, they will be set with JClass Chart 3D's default
 * legend populator and renderer.
 *
 * @param jl the <code>JCLegend</code> object associated with the legend
 */
public void setLegend(JCLegend jl)
{
	if (jl == null) {
        return;
    }
	if (legend != null) {
		remove(legend);
	}
	legend = jl;
	if (legend.getLegendPopulator() == null) {
		legend.setLegendPopulator(legendManager);
	}
	if (legend.getLegendRenderer() == null) {
		legend.setLegendRenderer(legendManager);
	}
	add(legend);
	setChanged(true, Changeable.LAYOUT |
					 Changeable.NOTIFY_CA_RECALC |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the value of the <code>Chart3dArea</code> property.
 * The <code>Chart3dArea</code> property controls the object that controls the
 * display of the graph.
 *
 * @return the <code>JCChart3dArea</code> object associated with this
 *     <code>JCChart3d</code>
 */
public JCChart3dArea getChart3dArea()
{
	return(chart3dArea);
}

/**
 * Used to implement the
 * <code>{@link com.klg.jclass.util.legend.LegendComponentLayoutUser}</code>
 * interface. Returns the component used to draw the actual chart.
 *
 * @return the subcomponent used to draw the actual chart
 */
public JComponent getComponentArea()
{
	return(getChart3dArea());
}

/**
 * Sets the value of the <code>Chart3dArea</code> property. The
 * <code>Chart3dArea</code> property controls the object that controls the
 * display of the graph.
 *
 * @param ca the new chart area
 */
public void setChart3dArea(JCChart3dArea ca)
{
	if (ca == null) {
		return;
	}
	if (chart3dArea != null) {
		remove(chart3dArea);
	}
	chart3dArea = ca;
	add(chart3dArea);
	chart3dArea.setChart3d(this);
	setChanged(true, Changeable.LAYOUT |
					 Changeable.NOTIFY_CA_RECALC |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Gets the value of the specified index within the <code>dataViews</code>
 * property. The <code>dataviews</code> property is an indexed property of
 * <code>JCChart3d</code> that contains one or more <code>Chart3dDataView</code>
 * objects. Each <code>Chart3dDataView</code> object controls the display of
 * data using a particular chart type.
 *
 * @param index index of the <code>dataView</code> object to return
 * @return <code>ChartDataView</code> object
 */
public Chart3dDataView getDataView(int index)
{
	Chart3dDataView d = null;
	try {
		d = (Chart3dDataView)dataViews.get(index);
	}
	catch (Exception e) {
	}
	return(d);
}

/**
 * Creates a new <code>Chart3dDataView</code> and adds it to the chart
 * at the specified index, <B>replacing</B> the existing
 * <code>ChartDataView</code> object at that specified index. If the index is
 * one past the final occupied index, the <code>ChartDataView</code> will be
 * added to the end of the list. Also, see {@link #setDataView setDataView}.
 *
 * @param index index of the <code>dataView</code> object to add
 * @param chartType the <code>chartType</code> of the added
 *     <code>dataView</code>
 * @return the newly creaated <code>ChartDataView</code>
 */
public Chart3dDataView addDataView(int index, int chartType)
{
	Chart3dDataView ndv = new Chart3dDataView(this, index, chartType);
	setDataView(index, ndv);
	return(ndv);
}

/**
 * Takes the specified <code>Chart3dDataView</code> and adds it to the chart
 * at the index indicated. The act of adding this new
 * <code>Chart3dDataView</code> will cause all existing
 * <code>Chart3dDataView</code> objects to re-order themselves. Also, see
 * {@link #addDataView addDataView}.
 *
 * @param index position at which to add the new object
 * @param ndv new data view object
 */
public void setDataView(int index, Chart3dDataView ndv)
{
	if (index < 0 || index > dataViews.size()) {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.BAD_DATA_OR_INDEX_ERROR));
	}
	if (ndv != null) {
		ndv.setChart3d(this);
	}
	if (index >= 0 && index < dataViews.size()) {
		dataViews.set(index, ndv);
	}
	else {
		dataViews.add(index, ndv);
	}
	reorderDataViews();

	//chart3dArea.clearDrawables();
	setChanged(true, Changeable.LAYOUT              |
					 Changeable.NOTIFY_CA_RECALC    |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Sets the value of the <code>dataViews</code> property, which is an indexed
 * property that contains all the data to be displayed in the 3d chart.
 *
 * @param views array of data views to be included in the
 *     <code>dataViews</code> property
 */
public void setDataView(List views)
{
	dataViews = new Vector();
	if (views == null) {
		return;
	}
    Iterator i = views.iterator();
	while (i.hasNext()) {
		Chart3dDataView cdv = (Chart3dDataView) i.next();
		if (cdv != null) {
			cdv.setChart3d(this);
		}
		dataViews.add(cdv);
	}

	//chart3dArea.clearDrawables();
	setChanged(true, Changeable.LAYOUT              |
					 Changeable.NOTIFY_CA_RECALC    |
					 Changeable.NOTIFY_LEGEND_RECALC);
}

/**
 * Finds a data view, given its name.
 *
 * @param name the name to use in the search
 * @return the desired <code>ChartDataView</code>; if not found, null is returned
 */
public Chart3dDataView findDataView(String name)
{
	if (name == null || name.length() == 0) {
		return(null);
	}
	if (dataViews == null || dataViews.size() == 0) {
		return(null);
	}
	for (int i = 0; i < dataViews.size(); i++) {
		Chart3dDataView dv = (Chart3dDataView)(dataViews.elementAt(i));
		if (dv.getName().equals(name)) {
			return(dv);
		}
	}
	return(null);
}

/**
 * Gets the value of the <code>DataViews</code> property, which is an indexed
 * property that contains all the data to be displayed in the chart.
 *
 * @return array of data views that are included in the <code>dataViews</code>
 *     property (currently only one is supported)
 */
public List getDataView()
{
	if (dataViews == null || dataViews.size() == 0) {
		return(null);
	}
	return(new Vector(dataViews));
}

/**
 * Gets the actual data view collection. Used within the chart instead of using
 * <code>getDataView()</code> in order to avoid copying.
 *
 * @return the list of <code>dataViews</code> (currently only one is supported)
 */
protected List getOriginalDataView()
{
	return(dataViews);
}

/**
 * Removes a <code>Chart3dDataView</code> from the index specified. The act of
 * removing a <code>Chart3dDataView</code> from the 3d chart will cause all
 * existing <code>Chart3dDataView</code> objects to re-order themselves.
 *
 * @param index position from which to remove the object
 */
public void removeDataView(int index)
{
	try {
		//chartArea.removeDrawable((ChartDataView)(data.elementAt(index)));
		dataViews.remove(index);
		reorderDataViews();
		setChanged(true, Changeable.LAYOUT              |
						 Changeable.NOTIFY_CA_RECALC    |
						 Changeable.NOTIFY_LEGEND_RECALC);
	}
	catch (Exception e) {}
}

/**
 * Gets the value of the <code>NumDataViews</code> property, which indicates
 * how many <code>Chart3dDataView</code> objects are stored in JClass Chart 3D.
 * It is a read-only property.
 *
 * @return number of <code>Chart3dDataView</code> objects inside this
 *     <code>JCChart3d</code> instance. Currently only one <code>dataView</code>
 *     is supported.
 */
public int getNumDataViews()
{
	int rval = 0;
	try {
		if (dataViews != null) {
			rval = dataViews.size();
		}
	}
	catch (Exception e) {
	}
	return(rval);
}

/**
 * Sets the order attribute of each attached <code>ChartDataView</code>
 * object to its index in the data array.
 */
void reorderDataViews()
{
	for (int i = 0; i < dataViews.size(); i++) {
		Chart3dDataView dv = (Chart3dDataView)dataViews.elementAt(i);
		dv.setDataViewId(i);
	}
}

/**
 * Performs a reset on the chart. Returns the default dataport to
 * <code>chart3d</code>.
 */
public void reset()
{
	chart3dArea.reset();
}

/**
 * Performs a cancel on the chart. Cancel will stop any current action.
 */
public void cancel()
{
	chart3dArea.cancel();
}

/**
 * Gets the value of the <code>About</code> property.
 * The <code>About</code> property displays contact information for
 * Quest Software in the BeanBox.
 * @return the about String
 */
public String getAbout()
{
	return(JCVersion.getVersionString());
}

/**
 * Sets the value of the <code>About</code> property.
 * The <code>About</code> property displays contact information for
 * Quest Software in the BeanBox.
 *
 * @param s the new about String
 */
public void setAbout(String s) {}

/**
 * Given a screen position in pixels, returns a <code>JCData3dIndex</code>
 * object that represents the index of the closest point in the elevation data
 * set of the specified <code>ChartData3dView</code> instance. If no data view
 * is supplied, all data views are considered when finding the closest point
 * (only one <code>dataView</code> is currently supported). If the data in a
 * data view is being updated when <code>pick()</code> is called, the result may
 * be incorrect.
 *
 * @param p pick point in pixels relative to the <code>JCChart3d</code> object
 * @param dataView data view on which to perform pick; if null, all data views
 *      are used (only one <code>dataView</code> is currently supported)
 * @return The <code>JCData3dIndex</code> object representing the index of the
 *     picked point. This is either a <code>JCData3dGridIndex</code>
 *     representing the <i>(x, y)</i> index of grid data point, or a
 *     <code>JCData3dPointIndex</code> representing the <i>(series, point)</i>
 *     index of a point in a point data set.
 */
public JCData3dIndex pick(Point p, Chart3dDataView dataView)
{
	Component comp = getComponentAt(p.x, p.y);
	if (comp == null) {
		return(null);
	}

	Point pp = new Point(p.x - comp.getLocation().x,
						 p.y - comp.getLocation().y);

    // Since there is only one possible dataView, get it.
    if (dataView == null) {
        dataView = getDataView(0);
    }

	if (comp instanceof JCLegend) {
		return(pickLegend(pp, dataView));
	}
	else if (comp instanceof JCChart3dArea) {
		return(chart3dArea.pick(pp, dataView));
	}
	else if (comp != null) {
		JCData3dIndex di = new JCData3dIndex(dataView);
		di.obj = comp;
		return(di);
	}
	else {
		return(null);
	}
}

/**
 * Internal pick method.
 * Chooses the contour level or data series represented by the legend item at
 * the provided position. If a data view is supplied, picking is done for
 * the provided data view only. Otherwise, it tries to do picking for any
 * data view (currently only one is supported).
 *
 * @param p pick point relative to legend
 * @param dataView data view on which to perform pick; if null, all data views
 *      are used (only one <code>dataView</code> is currently supported)
 * @return <code>JCData3dIndex</code> object representing the picked legend
 *     item.  This might be a <code>JCData3dContourIndex</code> that represents
 *     a contour range, or a <code>JCData3dContourPointIndex</code> that
 *     represents a Point Data series.
 */
protected JCData3dIndex pickLegend(Point p, Chart3dDataView dataView)
{
    JCLegendItem pickItem = legend.pick(p);

    JCData3dIndex di = null;
	boolean isGridData =
        (dataView.getElevationData() instanceof Chart3dGridData);
	di = (isGridData) ? (JCData3dIndex)(new JCData3dContourIndex(dataView))
					  : (JCData3dIndex)(new JCData3dPointIndex(dataView,
                                                               -1, -1));
    di.setObject(legend);

	if (pickItem == null || pickItem.itemInfo == null) {
		return(di);
	}

    JCData3dIndex pickIndex = null;
	if (pickItem.itemInfo instanceof LegendEntry) {
    	pickIndex = ((LegendEntry)pickItem.itemInfo).dataIndex;
	}
	else if (pickItem.itemInfo instanceof JCData3dIndex) {
    	pickIndex = (JCData3dIndex) pickItem.itemInfo;
	}
	else {
		return (di);
	}

    Chart3dDataView dv = pickIndex.getDataView();
    if ((dataView != null) && (dv != dataView)) {
        return(di);
    }
    di.setDataView(dataView);
	if (isGridData) {
		JCData3dContourIndex gridPickIndex =
				(JCData3dContourIndex)pickIndex;
		JCData3dContourIndex gridIndex = (JCData3dContourIndex)di;
		gridIndex.setContourStyleIndex(gridPickIndex.getContourStyleIndex());
		gridIndex.setUpperContourRangeValue(
				gridPickIndex.getUpperContourRangeValue());
		gridIndex.setLowerContourRangeValue(
				gridPickIndex.getLowerContourRangeValue());
	}
	else {
		JCData3dPointIndex pointPickIndex = (JCData3dPointIndex)pickIndex;
		JCData3dPointIndex pointIndex     = (JCData3dPointIndex)di;
		pointIndex.setSeries(pointPickIndex.getSeries());
		pointIndex.setPoint(pointPickIndex.getPoint());
	}
    return(di);
}

/**
 * Returns the position in screen pixels of a particular point in a
 * particular data set (grid data or point data).
 *
 * @param dataView the data view containing the specified series
 * @param index The data index of the point.  This is either a
 *     <code>JCData3dGridIndex</code> representing the <i>(x, y)</i> index of
 *     grid data point, or a <code>JCData3dPointIndex</code> representing the
 *     <i>(series, point)</i> index of a point in a point data set.
 * @return <code>AWT Point</code> object representing position in screen pixels
 *     relative to the <code>JCChart3d</code> object, or null if the point does
 *     not exist
 */
public Point unpick(Chart3dDataView dataView, JCData3dIndex index)
{
	if (index == null) {
		return(null);
	}

    // Since there is only one possible dataView, get it.
    if (dataView == null) {
        dataView = getDataView(0);
    }

	// Make sure that the dataView has a draw object
	dataView.createDraw();
	Point p = dataView.getDraw().getMapping().unpick(index);

	// Adjust unpicked point to screen pixels relative to JCChart object.
	if (p != null) {
		Point l = chart3dArea.getLocation();
		p.x += l.x;
		p.y += l.y;

		Rectangle r = chart3dArea.getDrawingArea();
		p.x += r.x;
		p.y += r.y;
	}
	return(p);
}

/**
 * Adds a listener to changes in JClass Chart 3D. Called after zoom, translate,
 * and scale (when used either through method calls or interactive triggers),
 * as well as following rotate, depth change, or edit (interactive only).
 *
 * @param l the listener to be added
 */
public void addChart3dListener(JCChart3dListener l)
{
	listeners = JCListenerList.add(listeners, l);
}

/**
 * Removes the listener for changes in JClass Chart 3D from a list of listeners.
 *
 * @param l the listener to be removed
 */
public void removeChart3dListener(JCChart3dListener l)
{
	listeners = JCListenerList.remove(listeners, l);
}

/**
 * Tells interested users that the chart is changed after an
 * action is performed.
 */
public void sendEvent()
{
	JCChart3dEvent  ev = new JCChart3dEvent(this);
	Enumeration     e  = JCListenerList.elements(listeners);

	while (e.hasMoreElements()) {
		((JCChart3dListener)e.nextElement()).changeChart(ev);
	}
}

/**
 * Adds listener to JClass Chart 3D pick events.
 *
 * @param l the new pick listener to be added
 */
public void addPick3dListener(JCPick3dListener l)
{
    pickListeners = JCListenerList.add(pickListeners,l);
}

/**
 * Removes the listener for JClass Chart 3D pick events from a list of
 * listeners.
 *
 * @param l the pick listener to be deleted
 */
public void removePick3dListener(JCPick3dListener l)
{
    pickListeners = JCListenerList.remove(pickListeners,l);
}

/**
 * Sends pick event to all objects listening for it.
 *
 * @param index The data index to send to the pick listeners.  This could be
 *     a <code>JCData3dGridIndex</code>, <code>JCData3dPointIndex</code>, or
 *     <code>JCData3dContourIndex</code>, depending on the location and type of
 *     picking that has happened.
 */
public void sendPick3dEvent(JCData3dIndex index)
{
    JCPick3dEvent ev = new JCPick3dEvent(this, index);
    Enumeration e = JCListenerList.elements(pickListeners);
    while(e.hasMoreElements()) {
        ((JCPick3dListener)e.nextElement()).pick(ev);
    }
}

/**
 * Requests the input focus. A <code>FocusEvent</code> will be generated
 * to indicate that the focus has been gained.
 *
 * @see java.awt.event.FocusEvent
 */
public void requestFocus()
{
	super.requestFocus();
	if (isFocusTraversable()) {
		processEvent(new FocusEvent(this, FocusEvent.FOCUS_GAINED));
	}
}

/**
 * Overrides the superclass's <code>isFocusTraversable</code> method to always
 * return <code>true</code>, which is required to get <code>keyPress</code>
 * events.
 * @return Is this component focus traversable?
 */
public boolean isFocusTraversable()
{
	return(true);
}


/**
 * Resets the line, symbol, and fill color and pattern
 * to the default index values.
 * The colors are defined in {@link JCStyle}.
 */
public void resetDefaults()
{
	lineColorIndex = 0;
	fillColorIndex = 0;
	symbolColorIndex = 0;
	symbolShapeIndex = 1;
}

/**
 * Sets the line color index.
 *
 * @param colorIndex the new line color index
 */
public void setLineColorIndex(int colorIndex)
{
	lineColorIndex = colorIndex;
}

/**
 * Sets the color index based on the style.
 *
 * @param colorIndex the new color index
 * @param style the fill style, line style, or symbol style color index
 */
protected void setColorIndex(int colorIndex, int style)
{
    switch (style) {
    case JCStyle.FILL_STYLE:
        setFillColorIndex(colorIndex);
        break;
    case JCStyle.LINE_STYLE:
        setLineColorIndex(colorIndex);
        break;
    case JCStyle.SYMBOL_STYLE:
        setSymbolColorIndex(colorIndex);
        break;
    }
}

/**
 * Sets the initial fill color index.
 *
 * @param colorIndex the new fill color index
 */
public void setFillColorIndex(int colorIndex)
{
	fillColorIndex = colorIndex;
}

/**
 * Sets the initial symbol color index.
 *
 * @param colorIndex the new symbol color index
 */
public void setSymbolColorIndex(int colorIndex)
{
	symbolColorIndex = colorIndex;
}

/**
 * Sets the initial symbol shape index.
 *
 * @param shapeIndex the new shape index
 */
public void setSymbolShapeIndex(int shapeIndex)
{
	symbolShapeIndex = shapeIndex;
}

/**
 * Gets the color index based on the style.
 *
 * @param style the fill style, line style, or symbol style color index
 * @return the color index based on the style
 */
protected int getColorIndex(int style)
{
    int index = 0;
    switch (style) {
    case JCStyle.FILL_STYLE:
        index = getFillColorIndex();
        break;
    case JCStyle.LINE_STYLE:
        index = getLineColorIndex();
        break;
    case JCStyle.SYMBOL_STYLE:
        index = getSymbolColorIndex();
        break;
    }
    return(index);
}

/**
 * Gets the current line color index.
 * @return the current line color index
 */
public int getLineColorIndex()
{
	return(lineColorIndex);
}

/**
 * Gets the current fill color index.
 * @return the current fill color index
 */
public int getFillColorIndex()
{
	return(fillColorIndex);
}

/**
 * Gets the current symbol color index.
 * @return the current symbol color index
 */
public int getSymbolColorIndex()
{
	return(symbolColorIndex);
}

/**
 * Gets the current symbol shape index.
 * @return the current symbol shape index
 */
public int getSymbolShapeIndex()
{
	return(symbolShapeIndex);
}

/**
 * Gets the value of the <code>CustomizerName</code> property.
 * The <code>CustomzierName</code> property controls the name of the customizer
 * used by JClass Chart 3D. It must be a full class name. The default
 * value is <code>com.klg.jclass.chart3d.customizer.Chart3dCustomizer</code>.
 *
 * @return the current customizer name
 */
public abstract String getCustomizerName();

/**
 * Sets the value of the <code>CustomizerName</code> property.
 * The <code>CustomzierName</code> property controls the name of the customizer
 * used by JClass Chart 3D. It must be a full class name. The default
 * value is <code>com.klg.jclass.chart3d.customizer.Chart3dCustomizer</code>.
 *
 * @param s the new customizer name
 */
protected abstract void setCustomizerName(String s);

/**
 * Returns the UI for <code>JCChart3d</code>.
 * @return the <code>JCChart3d</code> UI object
 */
public JCChart3dUI getUI()
{
	return((JCChart3dUI)ui);
}

/**
 * Sets the UI for <code>JCChart3d</code>.
 *
 * @param newUI the new user interface object
 */
public void setUI(ComponentUI newUI)
{
	super.setUI(newUI);
}

/**
 * Updates the UI for <code>JCChart3d</code>.
 */
public void updateUI()
{
    ComponentUI compUI = UIManager.getUI(this);
    // Because Swing's UI Manager caches not only our UI string, but also
    // the resulting UI class that it refers too, UIManager.getUI() can
    // return a UI class from a different copy of the chart jar rather
    // than the one this instance of class resides in.
    // To test for this case, do a simple cast of the retrieved UI.
    // If the cast fails, ask the UI class to create a new instance
    // of itself for use with this class.
    try {
        JCChart3dUI chartUI = (JCChart3dUI) compUI;
    }
    catch (Exception e) {
        compUI = JCChart3dUI.createUI(this);
    }
    setUI(compUI);
	//setUI((JCChart3dUI)UIManager.getUI(this));
}

/**
 * Returns the UIClass ID for <code>JCChart3d</code>.
 * @return the UIClass ID
 */
public String getUIClassID()
{
	return(uiClassID);
}

/**
 * Tells the <code>UIManager</code> class which class to use for
 * <code>JCChart3d</code>.
 */
public void registerUI()
{
	UIManager.getDefaults().put(uiClassID,
                                "com.klg.jclass.chart3d.JCChart3dUI");
}

// Hack for Visual Cafe event usage.  Cafe insists all object passed
// to listener methods are events that a getSource() can be done on.
// One of our listeners takes a chart, so we need this.
/**
 * Returns this chart as an object.
 * @return this chart as an object
 */
public Object getSource()
{
	return(this);
}

/**
 * Overrides the <code>Container</code> class's method to return only
 * components that are actually showing.
 *
 * @param x the <i>x</i> position of the component
 * @param y the <i>y</i> position of the component
 * @return the current component at the specified point
 */
public Component getComponentAt(int x, int y)
{
	if (!contains(x, y)) {
		return(null);
	}
	int ncomponents = getComponentCount();
	Component component[] = getComponents();
	for (int i = 0 ; i < ncomponents ; i++) {
		Component comp = component[i];
		if (comp != null) {
			Point p = comp.getLocation();
			if (comp.isShowing() && comp.contains(x - p.x, y - p.y)) {
				return(comp);
			}
		}
	}
	return(this);
}

/**
 * Takes a snapshot of the current chart.
 *
 * @return image object containing a snapshot of the chart
 */
public Image snapshot()
{
	return(snapshot(BufferedImage.TYPE_INT_ARGB));
}

/**
 * Takes a snapshot of the current chart and places it in an image of the
 * specified type.  The image types are as specified in the
 * <code>BufferedImage</code> class. <code>BufferedImage.TYPE_INT_ARGB</code>
 * is a good default for representing many possible colors.  If using less than
 * 256 colors, </code>BufferedImage.TYPE_BYTE_INDEXED</code> may prove to
 * generate faster and smaller images.
 *
 * @param imagetype the type of image to write to, as defined in the
 *     <code>java.awt.image.BufferedImage</code> class
 * @return image object containing snapshot of chart
 */
public Image snapshot(int imagetype)
{
	Image img;
	try {
		img = new BufferedImage(getSize().width,getSize().height, imagetype);
	}
	catch (Exception e) {
		img = createImage(getSize().width,getSize().height);
	}
	snapshot(img);
	return(img);
}

/**
 * Takes a snapshot of the current chart and places it in the supplied image.
 * @param img the image to put the snapshot in
 */
public void snapshot(Image img)
{
	if (img != null) {
		Graphics igc = img.getGraphics();
		//Gotta set the clip, or else paint throws an exception
		igc.setClip(0, 0, getSize().width, getSize().height);
		paint(igc);
	}
}

//ServerRenderable methods
/**
 * Overrides parent <code>getGraphics()</code>. If another
 * <code>Graphics</code> object has been set to be the current one, that one is
 * returned. Otherwise, call the superclass' <code>getGraphics()</code> method.
 *
 * @return the <code>Graphics</code> object to use to draw this chart
 */
public Graphics getGraphics()
{
	if (prevSetGraphics == null) {
		return(super.getGraphics());
	}
	return(prevSetGraphics);
}

/**
 * Provides a way for the chart to draw to a custom <code>Graphics</code>
 * object. A null call specifies that the default <code>Graphics</code> object
 * is used to draw to the screen.
 *
 * @param g <code>Graphics</code> object to use for drawing
 */
public void setGraphics(Graphics g)
{
	prevSetGraphics = g;
	getChart3dArea().setGraphics(g);
	getLegend().setGraphics(g);
}

/**
 * Internal use only.  Don't use.
 */
public Image serverSnapshot()
{
	return(ServerRenderer.snapshot(this));
}

/**
 * Internal use only.  Don't use.
 */
public void serverPaint(Graphics igc)
{
	ServerRenderer.paintComponent(igc, this);
}

/**
 * Overrides <code>readObject()</code> to first register UI class before deserialiaing.
 * Otherwise, the <code>UIManager</code> will throw a null pointer exception when
 * trying to find it during object reading.
 *
 * @exception java.io.IOException
 * @param in the input stream to read from
 */
private void readObject(java.io.ObjectInputStream in) throws java.io.IOException
{
	try {
		registerUI();
		in.defaultReadObject();
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
}

/**
 * Override write object to remove the customizer as a JCChart3dListener before
 * serializing.  Otherwise it will serialize the customizer.
 * @param out The output stream to which the object is to be written
 * @throws IOException
 */
private void writeObject(java.io.ObjectOutputStream out)
    throws IOException
{
    // Remove the customizer as a chart 3d listener.  We don't want to save it.
    if (customizer instanceof JCChart3dListener) {
        removeChart3dListener((JCChart3dListener)customizer);
    }
    out.defaultWriteObject();
    // Re-add the customizer as a chart 3d listener.
    if (customizer instanceof JCChart3dListener) {
        addChart3dListener((JCChart3dListener)customizer);
    }
}


 

/**
 * Retrieves the <i>x</i> grid values. This method is required to implement the
 * <code>Chart3dGridDataModel</code> interface.
 *
 * @return array of double values representing <i>x</i> grid points
 */
public double[] getXGrid()
{
	double  xarray[] = new double[11];
	for (int i = 0; i < xarray.length; i++) {
		xarray[i] = (double)i;
	}
	return(xarray);
}

/**
 * Retrieves the <i>y</i> grid values. This method is required to implement the
 * <code>Chart3dGridDataModel</code> interface.
 *
 * @return array of double values representing <i>y</i> grid points
 */
public double[] getYGrid()
{
	double  yarray[] = new double[11];
	for (int i = 0; i < yarray.length; i++) {
		yarray[i] = (double)i;
	}
	return(yarray);
}

/**
 * Retrieves the <i>z</i> values - one for each (<i>x</i>,<i>y</i>) grid point.
 * This method is required to implement <code>Chart3dGridDataModel</code>
 * interface.
 *
 * @return doubly subscripted array of double values representing <i>z</i>
 *     values
 */
public double[][] getZValues()
{
	double  zvals[][] = new double[11][11];
	for (int i = 0; i < zvals.length; i++) {
		for (int j = 0; j < zvals[i].length; j++) {
			double xval = (double)(i-5);
			double yval = (double)(j-5);
			zvals[i][j] = xval*xval + yval*yval;
		}
	}
	return(zvals);
}

/**
 * Retrieves a list of points for a scatter plot. This method is required to
 * implement the <code>Chart3dPointDataModel</code> interface.
 *
 * @return an array of <code>Point3d</code> arrays, each of which represents
 *     the points in a series
 */
public Point3d[][] getPoints()
{
	Point3d[][] series = new Point3d[5][];
	int sizes[] = {3, 5, 6, 4, 2};

	for (int i = 0; i < series.length; i++) {
		Point3d[] points = new Point3d[sizes[i]];
		for (int j = 0; j < points.length; j++) {
			points[j] = new Point3d((double)(20 + 2*i),
			                        (double)(30 + j), (double)(10*i + j));
		}
		series[i] = points;
	}
	return(series);
}

/**
 * Sets the locale for chart3d.
 *
 * @param locale the <code>Locale</code> object
 * @see java.util.Locale
 */
public void setLocale(Locale locale)
{
    if (this.locale == locale || locale == null) {
        return;
    }

	Locale oldValue = locale;
    try {
        LocaleBundle.setBundleLocale(locale);
        this.locale = locale;
    }
    catch ( MissingResourceException e) {
        this.locale = Locale.getDefault();
    }

	if (legend != null) {
        legend.setLocale(locale);
	}

	if (chart3dArea != null) {
        JCAxis[] axes = chart3dArea.getAxes();
		if (axes != null && axes.length > 0) {
            for (int i = 0; i < axes.length; i++ ) {
                axes[i].getNumberFormatter().setNumberFormatterLocale(locale);
            }
		}
	}

	// This is a bound property, so report the change to
	// any registered listeners.  (Cheap if there are none.)
	firePropertyChange("locale", oldValue, locale);

	// This could change the preferred size of the Component.
	if (isValid()) {
	    invalidate();
	}
}

/**
 * Returns the chart3d locale.
 *
 * @return Locale the chart3d locale
 */
public Locale getLocale()
{
    return(this.locale);
}

}
