/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChart3dBundle.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.util.Locale;
import java.util.ResourceBundle;
import java.text.MessageFormat;
import java.text.ChoiceFormat;

/** 
 * Class containing locale-specific objects and helpers to get a locale-specific
 * String. When a program needs a locale-specific String, it can load it
 * from the resource bundle that is appropriate for the current user's locale.
 * Those bundles are located in the <code>com.klg.jclass.chart\resources</code>
 * directory. This means that program code can be written so that the code is
 * largely independent of the user's locale, and thus isolates most of the
 * locale-specific information in resource bundles.
 */
public class JCChart3dBundle {

private static final String BUNDLE_PATH =
    "com.klg.jclass.chart3d.resources.LocaleInfo";

private static ResourceBundle chart3d_bundle =
    ResourceBundle.getBundle(BUNDLE_PATH, Locale.getDefault());

/**
 * Static function that localizes a String based on its key,
 * language, and country.
 *
 * @param language desired language
 * @param country desired country
 * @param key the key of the String that is to be retrieved
 * @return
 */
public static String string(String language, String country, String key)
{
    return(ResourceBundle.getBundle(BUNDLE_PATH,
                                    new Locale(language,
                                               country)).getString(key));
}

/**
 * Static function that retrives a String based on its key;
 * in JDK 1.1, the default locale is used.
 *
 * @param key the key of the String that is to be retrieved
 * @return
 */
public static String string(String key)
{
    return(chart3d_bundle.getString(key));
}

/**
 * Static function used to format a String based on its key.
 * It produces concatenated messages in a language-neutral way.
 *
 * @param key the key of the String that is to be retrieved
 * @param args set of objects used in formatting
 * @return
 */
public static String string(String key, Object[] args)
{
    return(MessageFormat.format(chart3d_bundle.getString(key), args));
}

/**
 * Static function that formats a String based on its key.
 * It produces concatenated messages in a language-neutral way.
 *
 * @param key the key of the String that is to be retrieved
 * @param formatter a formatter provided by the user
 * @param args set of objects used in formatting
 * @return
 */
public static String string(String key, MessageFormat formatter, Object args[])
{
    return(formatter.format(chart3d_bundle.getString(key), args));
}

/**
 * Static function used to change the default locale programatically.
 *
 * @param l
 */
public static void setBundleLocale(Locale l)
{
    chart3d_bundle = ResourceBundle.getBundle(BUNDLE_PATH, l);
}

/**
 * Static function used to format a String based on its key.
 * It produces concatenated messages in a language-neutral way and allows
 * the user to attach a format to a range of numbers. It is generally used for
 * handling plurals. The choice is specified with an ascending list of doubles,
 * where each item specifies a half-open interval up to the next item:
 * <i>X</i> matches <i>j</i> if <i>limit[j] <= X < limit[j+1]</i>.
 *
 * @param key the key of the String that is to be retrieved
 * @param args set of objects used in formatting
 * @param limits the list of doubles that specifies a half-open interval
 * @param formats it is an array of choices: one choice for each interval
 *     defined by the limits array
 * @param choiceVar it is used to identify which of the variables, from the
 *     String that contains the formatting information, uses the choice
 *     formatting feature<br><B>Note:</b> There is a limitation of only one
 *     choice variable per String.
 * @return
 */
public static String string(String key, Object[] args, double[] limits,
                            String[] formats, int choiceVar)
{
    MessageFormat formatter = new MessageFormat(chart3d_bundle.getString(key));
    ChoiceFormat choicer = new ChoiceFormat(limits, formats);
    formatter.setFormat(choiceVar, choicer);
    return(formatter.format(args));
}


// Key constants

/**  */
public static final String key1   = "com.klg.jclass.chart3d.Key1";
/**  */
public static final String key2   = "com.klg.jclass.chart3d.Key2";
/**  */
public static final String key3   = "com.klg.jclass.chart3d.Key3";
/**  */
public static final String key4   = "com.klg.jclass.chart3d.Key4";
/**  */
public static final String key5   = "com.klg.jclass.chart3d.Key5";
/**  */
public static final String key6   = "com.klg.jclass.chart3d.Key6";
/**  */
public static final String key7   = "com.klg.jclass.chart3d.Key7";
/**  */
public static final String key20  = "com.klg.jclass.chart3d.Key20";
/**  */
public static final String key26  = "com.klg.jclass.chart3d.Key26";
/**  */
public static final String key27  = "com.klg.jclass.chart3d.Key27";
/**  */
public static final String key28  = "com.klg.jclass.chart3d.Key28";
/**  */
public static final String key29  = "com.klg.jclass.chart3d.Key29";
/**  */
public static final String key30  = "com.klg.jclass.chart3d.Key30";
/**  */
public static final String key31  = "com.klg.jclass.chart3d.Key31";
/**  */
public static final String key44  = "com.klg.jclass.chart3d.Key44";
/**  */
public static final String key45  = "com.klg.jclass.chart3d.Key45";
/**  */
public static final String key87  = "com.klg.jclass.chart3d.Key87";
/**  */
public static final String key88  = "com.klg.jclass.chart3d.Key88";
/**  */
public static final String key89  = "com.klg.jclass.chart3d.Key89";
/**  */
public static final String key90  = "com.klg.jclass.chart3d.Key90";
/**  */
public static final String key100 = "com.klg.jclass.chart3d.Key100";

/**  */
public static final String key105 = "com.klg.jclass.chart3d.Key105";

/**  */
public static final String key118 = "com.klg.jclass.chart3d.Key118";
/**  */
public static final String key121 = "com.klg.jclass.chart3d.Key121";
/**  */
public static final String key123 = "com.klg.jclass.chart3d.Key123";
/**  */
public static final String key125 = "com.klg.jclass.chart3d.Key125";
/**  */
public static final String key127 = "com.klg.jclass.chart3d.Key127";
/**  */
public static final String key128 = "com.klg.jclass.chart3d.Key128";
/**  */
public static final String key129 = "com.klg.jclass.chart3d.Key129";
/**  */
public static final String key130 = "com.klg.jclass.chart3d.Key130";
/**  */
public static final String key132 = "com.klg.jclass.chart3d.Key132";
/**  */
public static final String key141 = "com.klg.jclass.chart3d.Key141";
/**  */
public static final String key142 = "com.klg.jclass.chart3d.Key142";
/**  */
public static final String key143 = "com.klg.jclass.chart3d.Key143";
/**  */
public static final String key146 = "com.klg.jclass.chart3d.Key146";
/**  */
public static final String key147 = "com.klg.jclass.chart3d.Key147";
/**  */
public static final String key163 = "com.klg.jclass.chart3d.Key163";
/**  */
public static final String key165 = "com.klg.jclass.chart3d.Key165";
/**  */
public static final String key166 = "com.klg.jclass.chart3d.Key166";
/**  */
public static final String key167 = "com.klg.jclass.chart3d.Key167";
/**  */
public static final String key168 = "com.klg.jclass.chart3d.Key168";
/**  */
public static final String key169 = "com.klg.jclass.chart3d.Key169";
/**  */
public static final String key170 = "com.klg.jclass.chart3d.Key170";
/**  */
public static final String key176 = "com.klg.jclass.chart3d.Key176";
/**  */
public static final String key177 = "com.klg.jclass.chart3d.Key177";
/**  */
public static final String key178 = "com.klg.jclass.chart3d.Key178";
/**  */
public static final String key179 = "com.klg.jclass.chart3d.Key179";
/**  */
public static final String key180 = "com.klg.jclass.chart3d.Key180";
/**  */
public static final String key181 = "com.klg.jclass.chart3d.Key181";
/**  */
public static final String key182 = "com.klg.jclass.chart3d.Key182";
/**  */
public static final String key183 = "com.klg.jclass.chart3d.Key183";
/**  */
public static final String key184 = "com.klg.jclass.chart3d.Key184";
/**  */
public static final String key185 = "com.klg.jclass.chart3d.Key185";
/**  */
public static final String key186 = "com.klg.jclass.chart3d.Key186";
/**  */
public static final String key187 = "com.klg.jclass.chart3d.Key187";
/**  */
public static final String key188 = "com.klg.jclass.chart3d.Key188";
/**  */
public static final String key189 = "com.klg.jclass.chart3d.Key189";
/**  */
public static final String key191 = "com.klg.jclass.chart3d.Key191";
/**  */
public static final String key192 = "com.klg.jclass.chart3d.Key192";
/**  */
public static final String key193 = "com.klg.jclass.chart3d.Key193";
/**  */
public static final String key194 = "com.klg.jclass.chart3d.Key194";
/**  */
public static final String key201 = "com.klg.jclass.chart3d.Key201";
/**  */
public static final String key202 = "com.klg.jclass.chart3d.Key202";
/**  */
public static final String key203 = "com.klg.jclass.chart3d.Key203";
/**  */
public static final String key208 = "com.klg.jclass.chart3d.Key208";
/**  */
public static final String key209 = "com.klg.jclass.chart3d.Key209";
/**  */
public static final String key210 = "com.klg.jclass.chart3d.Key210";
/**  */
public static final String key211 = "com.klg.jclass.chart3d.Key211";
/**  */
public static final String key212 = "com.klg.jclass.chart3d.Key212";
/**  */
public static final String key213 = "com.klg.jclass.chart3d.Key213";
/**  */
public static final String key214 = "com.klg.jclass.chart3d.Key214";
/**  */
public static final String key215 = "com.klg.jclass.chart3d.Key215";
/**  */
public static final String key216 = "com.klg.jclass.chart3d.Key216";
/**  */
public static final String key224 = "com.klg.jclass.chart3d.Key224";
/**  */
public static final String key225 = "com.klg.jclass.chart3d.Key225";
/**  */
public static final String key226 = "com.klg.jclass.chart3d.Key226";
/**  */
public static final String key227 = "com.klg.jclass.chart3d.Key227";
/**  */
public static final String key228 = "com.klg.jclass.chart3d.Key228";
/**  */
public static final String key229 = "com.klg.jclass.chart3d.Key229";
/**  */
public static final String key230 = "com.klg.jclass.chart3d.Key230";
/**  */
public static final String key231 = "com.klg.jclass.chart3d.Key231";
/**  */
public static final String key232 = "com.klg.jclass.chart3d.Key232";
/**  */
public static final String key233 = "com.klg.jclass.chart3d.Key233";
/**  */
public static final String key234 = "com.klg.jclass.chart3d.Key234";
/**  */
public static final String key235 = "com.klg.jclass.chart3d.Key235";
/**  */
public static final String key236 = "com.klg.jclass.chart3d.Key236";
/**  */
public static final String key237 = "com.klg.jclass.chart3d.Key237";
/**  */
public static final String key238 = "com.klg.jclass.chart3d.Key238";
/**  */
public static final String key239 = "com.klg.jclass.chart3d.Key239";
/**  */
public static final String key240 = "com.klg.jclass.chart3d.Key240";
/**  */
public static final String key244 = "com.klg.jclass.chart3d.Key244";
/**  */
public static final String key245 = "com.klg.jclass.chart3d.Key245";
/**  */
public static final String key246 = "com.klg.jclass.chart3d.Key246";
/**  */
public static final String key247 = "com.klg.jclass.chart3d.Key247";
/**  */
public static final String key248 = "com.klg.jclass.chart3d.Key248";
/**  */
public static final String key249 = "com.klg.jclass.chart3d.Key249";
/**  */
public static final String key250 = "com.klg.jclass.chart3d.Key250";
/**  */
public static final String key251 = "com.klg.jclass.chart3d.Key251";
/**  */
public static final String key252 = "com.klg.jclass.chart3d.Key252";
/**  */
public static final String key253 = "com.klg.jclass.chart3d.Key253";
/**  */
public static final String key254 = "com.klg.jclass.chart3d.Key254";
/**  */
public static final String key255 = "com.klg.jclass.chart3d.Key255";
/**  */
public static final String key256 = "com.klg.jclass.chart3d.Key256";
/**  */
public static final String key257 = "com.klg.jclass.chart3d.Key257";
/**  */
public static final String key258 = "com.klg.jclass.chart3d.Key258";
/**  */
public static final String key259 = "com.klg.jclass.chart3d.Key259";
/**  */
public static final String key260 = "com.klg.jclass.chart3d.Key260";
/**  */
public static final String key261 = "com.klg.jclass.chart3d.Key261";
/**  */
public static final String key262 = "com.klg.jclass.chart3d.Key262";
/**  */
public static final String key263 = "com.klg.jclass.chart3d.Key263";
/**  */
public static final String key264 = "com.klg.jclass.chart3d.Key264";
/**  */
public static final String key265 = "com.klg.jclass.chart3d.Key265";
/**  */
public static final String key266 = "com.klg.jclass.chart3d.Key266";
/**  */
public static final String key267 = "com.klg.jclass.chart3d.Key267";
/**  */
public static final String key268 = "com.klg.jclass.chart3d.Key268";
/**  */
public static final String key269 = "com.klg.jclass.chart3d.Key269";
/**  */
public static final String key275 = "com.klg.jclass.chart3d.Key275";
/**  */
public static final String key276 = "com.klg.jclass.chart3d.Key276";
/**  */
public static final String key277 = "com.klg.jclass.chart3d.Key277";
/**  */
public static final String key278 = "com.klg.jclass.chart3d.Key278";
/**  */
public static final String key279 = "com.klg.jclass.chart3d.Key279";
/**  */
public static final String key280 = "com.klg.jclass.chart3d.Key280";
/**  */
public static final String key281 = "com.klg.jclass.chart3d.Key281";
/**  */
public static final String key282 = "com.klg.jclass.chart3d.Key282";
/**  */
public static final String key283 = "com.klg.jclass.chart3d.Key283";
/**  */
public static final String key284 = "com.klg.jclass.chart3d.Key284";

/**  */
public static final String key298 = "com.klg.jclass.chart3d.Key298";
}
