/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChart3dLegend.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;
import com.klg.jclass.util.Changeable;
import com.klg.jclass.util.legend.JCLegend;
import com.klg.jclass.util.legend.JCLegendItem;

import java.awt.*;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;

/**
 * <code>JCChart3dLegend</code> manages and renders a grid-based legend.
 * It overrides <code>JCLegend</code>'s layout methods to position legend items
 * in a grid, where every row has the same height and every column the same
 * width.
 */
public class JCChart3dLegend extends JCLegend
							 implements JCChart3dLegendTemplate,
                                        java.io.Serializable
{

/** The style to use when laying out out ranges. */
protected int layoutStyle = CONTINUOUS;

/** The constraints to place upon the data ranges that appear in the legend. */
protected int distRange = RANGE_DATA;

/** Label generator class.  Overrides all other forms of label generation. */
protected JCChart3dLegendLabelGenerator labelGenerator = null;

/** List of user-specified legend labels. */
protected List userLabels = null;

/** Space between outside and inside of legend. */
protected int marginGap = 2;

/**
 * Space between groups in a legend (columns when legend vertical, rows when
 * horizontal).
 */
protected int groupGap = 5;

/** Horizontal space between items. */
protected int horizItemGap = 5;

/** Vertical space between items. */
protected int vertItemGap = 2;

/** Space between the symbol and String inside an item. */
protected int insideItemGap = 2;

/** Asks if the layout is really continuous. */
protected boolean continuousLayout = false;

/** Number of rows in the legend. */
protected int legendRows = 0;
/** Number of columns in the legend. */
protected int legendColumns = 0;

/**
 * Default constructor.
 */
public JCChart3dLegend()
{
	super();
}

/**
 * Sets the <code>Anchor</code> property, which determines the position of the
 * legend relative to the parent component's primary draw area.
 *
 * @param a The new anchor.  Must be one of <code>NORTH</code>,
 * <code>SOUTH</code>, <code>EAST</code>, <code>WEST</code>,
 * <code>NORTHWEST</code>, <code>SOUTHWEST</code>, <code>NORTHEAST</code>, or
 * <code>SOUTHEAST</code>
 */
public void setAnchor(int a)
{
    if (a == anchor) {
        return;
    }
    checkAnchor(a);

    setChanged(true, Changeable.RECALC);
}

/**
 * Sets the style to be used for displaying data ranges in the legend.  One
 * of </code>CONTINUOUS</code> or <code>STEPPED</code>.
 *
 * @param style The range style to use
 */
public void setLayoutStyle(int style)
{
    if (layoutStyle == style) {
		return;
    }
    switch (style) {
    case CONTINUOUS:
    case STEPPED:
        layoutStyle = style;
        break;
    default:
        throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_CONTINUOUS_OR_STEPPED_ERROR));
    }

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the style being used for displaying data ranges in the legend.  One
 * of <code>CONTINUOUS</code> or <code>STEPPED</code>.
 *
 * @return The current range style
 */
public int getLayoutStyle()
{
    return(layoutStyle);
}

/**
 * Sets the type of distribution range to use to display data in the legend.
 * One of <code>RANGE_DATA</code> or <code>RANGE_ALL</code>.
 * <code>RANGE_DATA</code> constrains the legend data ranges to the minimum and
 * maximum of the data in the chart. <code>RANGE_ALL</code> ensures that all
 * data ranges are included.
 *
 * @param range The new distribution range type
 */
public void setDistributionRange(int range) {
    if (distRange == range) {
		return;
    }
    switch (range) {
    case RANGE_DATA:
    case RANGE_ALL:
        distRange = range;
        break;
    default:
        throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_DATA_OR_ALL_ERROR));
    }

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the type of distribution range being used to display data in the legend
 * (one of <code>RANGE_DATA</code> or <code>RANGE_ALL</code>).
 * <code>RANGE_DATA</code> constrains the legend data ranges to the minimum and
 * maximum of the data in the chart.  </code>RANGE_ALL</code> ensures that all
 * data ranges are included.
 *
 * @return The current distribution range type
 */
public int getDistributionRange()
{
    return(distRange);
}

/**
 * Sets the class used to specify custom labels for the
 * legend.  Setting this property to a non-null value overrides
 * all other forms of setting legend labels.  A null value indicates
 * that another method will be used.  This property is ignored
 * for legends of <code>SCATTER</code> charts.
 *
 * @param generator A class implementing the
 *     <code>JCChart3dLegendLabelGenerator</code> interface
 */
public void setLabelGenerator(JCChart3dLegendLabelGenerator generator)
{
    labelGenerator = generator;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the class specifying custom labels for the
 * legend.  A null value indicates that another method of setting
 * legend labels is being used.
 *
 * @return A class implementing the <code>JCChart3dLegendLabelGenerator</code>
 *      interface that is being used to specify legend labels
 */
public JCChart3dLegendLabelGenerator getLabelGenerator()
{
    return(labelGenerator);
}

/** Sets the list of user specified label lists used to override
 * default label generation.  The passed List instance is
 * indexed by data view number.  Each item in the passed List
 * is itself a List instance representing the list of legend
 * labels (either <code>String</code> or <code>JCMultiFieldString</code>
 * objects) to represent the data view in the legend.
 *
 * <p>
 * A null value in the main List indicates that there are no
 * legend labels specified for the data view at that index.  A null
 * value for the property indicates that no legend labels have been
 * specified for any data view.
 * <p>
 * If the <code>LabelGenerator</code> property is set, this property is ignored.
 *
 * @param labels the list of user specified label lists
 */
public void setLabels(List labels)
{
    userLabels = labels;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the list of user specified label lists used to override
 * default label generation.  The value is null if no lists have been
 * specified.  The returned List instance is indexed by a data view
 * number.  Each item in the returned List is itself a List
 * instance representing the list of legend labels (either <code>String</code>
 * or <code>JCMultiFieldString</code> objects) to be used for the data view.
 *
 * @return the list of user specified label lists
 */
public List getLabels()
{
    return userLabels;
}

/**
 * Sets a list of user specified labels used to override default label
 * generation for the specified data view. Each item in the passed List instance
 * represents a legend label (either <code>String</code> or
 * <code>JCMultiFieldString</code> objects).  A null value indicates that no
 * override labels have been specified for the specified data view.
 *
 * If the <code>LabelGenerator</code> property is set, this property is ignored.
 *
 * @param dvIndex the data view labels are being attached to
 * @param labels the list of user specified labels
 */
public void setLabels(int dvIndex, List labels)
{
	if (userLabels == null) {
		userLabels = new Vector();
    }

	int currentLen = userLabels.size();
	if (dvIndex < currentLen) {
		userLabels.set(dvIndex, labels);
	}
	else {
		for (int i = currentLen; i < dvIndex; i++) {
			userLabels.add(i, null);
		}
		userLabels.add(dvIndex, labels);
	}

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the list of user specified labels being used to override
 * the default label generation for the specified data view.
 * Each item in the returned List instance represents a legend
 * label (either <code>String</code> or
 * <code>JCMultiFieldString</code> objects).  A null
 * value indicates that no override labels have been specified
 * for the specified data view.
 *
 * @param dvIndex the data view that the labels are being requested for
 * @return the list of user specified labels
 */
public List getLabels(int dvIndex)
{
	if (userLabels == null) {
		return(null);
	}

	int currentLen = userLabels.size();
	if (dvIndex < currentLen) {
		return((List) userLabels.get(dvIndex));
	}
	else {
		return(null);
	}
}

/**
 * Sets the gap between the edge of the legend and where the items start
 * being layed out.
 *
 * @param gap the margin gap
 */
public void setMarginGap(int gap)
{
    marginGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between the edge of the legend and where the items start
 * being layed out.
 *
 * @return the margin gap
 */
public int getMarginGap()
{
    return(marginGap);
}

/**
 * Sets the gap between groups of items in the legend (columns in a
 * vertical legend, rows in a horizontal legend).
 *
 * @param gap the group gap
 */
public void setGroupGap(int gap)
{
    groupGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between groups of items in the legend (columns in a
 * vertical legend, rows in a horizontal legend).
 *
 * @return the group gap
 */
public int getGroupGap()
{
    return(groupGap);
}

/**
 * Gets the gap between the legend items depending on the orientation.
 *
 * @return the item gap
 */
public int getItemGap()
{
	if (orientation == HORIZONTAL) {
    	return(getHorizontalItemGap());
	}
	else {
    	return(getVerticalItemGap());
	}
}

/**
 * Sets the gap between the legend items in a horizontal legend.
 *
 * @param gap the item gap
 */
public void setHorizontalItemGap(int gap)
{
    horizItemGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between the legend items in a horizontal legend.
 *
 * @return the item gap
 */
public int getHorizontalItemGap()
{
    return(horizItemGap);
}

/**
 * Sets the gap between the legend items in a vertical legend.
 *
 * @param gap the item gap
 */
public void setVerticalItemGap(int gap)
{
    vertItemGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between the legend items in a vertical legend.
 *
 * @return the item gap
 */
public int getVerticalItemGap()
{
    return(vertItemGap);
}

/**
 * Sets the gap between the symbol and text portions of a legend item.
 *
 * @param gap the inside item gap
 */
public void setInsideItemGap(int gap)
{
    insideItemGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between the symbol and text portions of a legend item.
 *
 * @return the inside item gap
 */
public int getInsideItemGap()
{
    return(insideItemGap);
}

/**
 * For internal use only.  Set by legend manager to determine if
 * continuous layout is actually continuous.  (In some cases,
 * <code>CONTINUOUS</code> layout is forced to a <code>STEPPED</code> layout.)
 *
 * @param continuous Is the layout actually continuous?
 */
public void setContinuousLayout(boolean continuous)
{
	continuousLayout = continuous;
}

/**
 * For internal use only.  Returns <code>true</code> if layout really is
 * continuous. (In some cases, <code>CONTINUOUS</code> layout is forced to
 * behave like a <code>STEPPED</code> layout.)
 * @return <code>true</code> if the layout is really continuous
 */
public boolean isContinuousLayout()
{
	return(continuousLayout);
}

/**
 * Overrides <code>JCLegend</code>'s <code>layoutLegend()</code> to layout the
 * legend items in a grid, calculating the overall size of the legend.  This is
 * accomplished by filling in the <code>dim</code>, <code>pos</code>,
 * <code>textPos</code>,  and <code>symPos</code> member of each
 * <code>JCLegendItem</code> instance in the legend vector.  The
 * <code>pickRectangle</code> member is also filled in.
 *
 * @param itemList The legend vector.  It contains a number of group vectors.
 * Each group vector contains the legend items (<code>JCLegendItem</code>
 * instances) for that group.
 * @param vertical <code>true</code> if a vertical legend; <code>false</code>
 * if horizontal
 * @param useFont the font for the legend
 * @return The size of the legend
 */
public Dimension layoutLegend(List itemList, boolean vertical, Font useFont)
{
    Dimension maxItemSize = new Dimension(0, 0);
    Dimension titleSize = new Dimension(0, 0);
    boolean hasTitle[] = findBiggestLegendItem(itemList, vertical,
                                               maxItemSize, titleSize);
    Dimension dim = calcLegendSize(itemList, vertical, maxItemSize,
                                   hasTitle, titleSize);
    if (isContinuousLayout()) {
        layoutContinuousLegendItems(itemList, vertical, maxItemSize,
                                    dim, hasTitle, titleSize);
    }
    else {
        layoutSteppedLegendItems(itemList, vertical, maxItemSize,
                                 dim, hasTitle, titleSize);
    }

    return(dim);
}

/**
 * Finds the biggest item in the legend.  The width of a legend item
 * is determined by adding the width of the symbol, the inside item gap
 * and the width of the text.  The height of a legend item is determined
 * by the maximum height of either the symbol or the text.
 *
 * @param itemList the list of items to layout
 * @param vertical <code>true</code> if the legend is vertical
 * @param maxItemSize the size of the largest legend item, to be filled in by
 *     this method
 * @param titleSize the size of the title item, to be filled in by this
 *     method
 * @return <code>true</code> if legend has a title; <code>false</code> otherwise
 */
protected boolean[] findBiggestLegendItem(List itemList, boolean vertical,
								          Dimension maxItemSize,
										  Dimension titleSize)
{
	int maxItemWidth = 0;
	int maxItemHeight = 0;
	int maxTitleItemWidth = 0;
	int maxTitleItemHeight = 0;

	int numViews = itemList.size();
	boolean[] hasTitle = new boolean[numViews];
	Arrays.fill(hasTitle, false);

    for (int i = 0; i < numViews; i++) {
        List viewItems = (List)itemList.get(i);

        for (int j = 0; j < viewItems.size(); j++) {
            JCLegendItem thisItem = (JCLegendItem)viewItems.get(j);
            int itemWidth = 0;
            int itemHeight = 0;

            if (isTitleItem(thisItem)) {
                hasTitle[i] = true;
                maxTitleItemWidth = Math.max(maxTitleItemWidth,
                                             thisItem.textDim.width);
                maxTitleItemHeight = Math.max(maxTitleItemHeight,
                                              thisItem.textDim.height);
            }
            else {
                if (isContinuousLayout() && vertical) {
                    itemWidth = thisItem.symbolDim.width +
                                getInsideItemGap() + thisItem.textDim.width;
                    itemHeight = thisItem.textDim.height;
                }
                else if (isContinuousLayout() && !vertical) {
                    itemWidth = thisItem.textDim.width;
                    itemHeight = thisItem.symbolDim.height +
                                 getInsideItemGap() + thisItem.textDim.height;
                }
                else {
                    itemWidth = thisItem.symbolDim.width +
                                getInsideItemGap() + thisItem.textDim.width;
                    itemHeight = Math.max(thisItem.symbolDim.height,
                                          thisItem.textDim.height);
                }
                maxItemWidth = Math.max(maxItemWidth, itemWidth);
                maxItemHeight = Math.max(maxItemHeight, itemHeight);
            }
        }
    }

	maxItemSize.width = maxItemWidth;
	maxItemSize.height = maxItemHeight;
	titleSize.width = maxTitleItemWidth;
	titleSize.height = maxTitleItemHeight;

	return(hasTitle);
}

/**
 * Calculates the size of a legend.  This includes calculating
 * the number of columns and rows.
 *
 * @param itemList the list of items to layout
 * @param vertical <code>true</code> if the legend is vertical
 * @param itemSize the size of the largest legend item
 * @param hasTitle <code>true</code> of this list of legend items contains a
 * title item
 * @param titleSize the size of the title item
 * @return the size of the legend
 */
protected Dimension calcLegendSize(List itemList, boolean vertical,
					               Dimension itemSize, boolean hasTitle[],
                                   Dimension titleSize)
{
    Dimension legendDim = null;
    int totalColumns = 0, totalRows = 0;

    int numViews = itemList.size();
    for (int i = 0; i < numViews; i++) {
        List viewItems = (List)itemList.get(i);

        int nlabels = viewItems.size();
        if (hasTitle[i]) {
            nlabels--;
        }

        int nrow = 0, ncol = 0;
        int minrow, mincol;
        if (!vertical) {
            minrow = 1;
            mincol = 2;
        }
        else {
            minrow = 3;
            mincol = 1;
        }

        // nlabels one smaller than actual number -- for use in the
        // row/column calculations below
        nlabels--;

        // Continuous legend -- don't consider last item (it's text only)
        if (isContinuousLayout()) {
            nlabels--;
        }

        // How many legend items can we fit?
        if ((getAnchor() & EASTWEST_MASK) != 0) {
            int space;
            if (isContinuousLayout() && vertical) {
                // Vertical continuous -- take into account extra
                // text item that is appended to the ends of rows
                space = getParent().getHeight() -
                    getInsets().top - getInsets().bottom -
                    itemSize.height -
                    (2 * getMarginGap());
            }
            else {
                space = getParent().getHeight() -
                    getInsets().top - getInsets().bottom -
                    (2 * getMarginGap());
            }
            if (hasTitle[i]) {
                space -= titleSize.height;
                space -= getGroupGap();
            }
            int itemHeight = itemSize.height + getItemGap();
            int temp = space / itemHeight;

            // Calc number of rows and columns
            nrow = Math.max(temp, minrow);
            ncol = Math.max(1 + nlabels / nrow, mincol);
            if (ncol != 0) {
                nrow = 1 + nlabels / ncol;
            }
            if (nrow != 0) {
                ncol = 1 + nlabels / nrow;
            }
        }
        else {
            // How many legend items can we fit?
            int space;
            if (isContinuousLayout() && !vertical) {
                // Horizontal continuous -- take into account extra
                // text item that is appended to the ends of columns
                space = getParent().getWidth() -
                    getInsets().left - getInsets().right -
                    itemSize.width -
                    (2 * getMarginGap());
            }
            else {
                space = getParent().getWidth() -
                    getInsets().left - getInsets().right -
                    (2 * getMarginGap());
            }
            int itemWidth = itemSize.width + getItemGap();
            int temp = space / itemWidth;

            // Calc number of rows and columns
            ncol = Math.max(temp, mincol);
            nrow = Math.max(1 + nlabels / ncol, minrow);
            if (nrow != 0) {
                ncol = 1 + nlabels / nrow;
            }
            if (ncol != 0) {
                nrow = 1 + nlabels / ncol;
            }
        }

        if (!vertical) {
            totalRows += nrow;
            totalColumns = ncol;
        }
        else {
            totalRows = nrow;
            totalColumns += ncol;
        }
    }
    // Use number of rows and columns to calculate legend size
    int legendWidth, legendHeight;
    if (isContinuousLayout() && !vertical) {
        // Horizontal continous legend -- need extra text item plus item gap
        legendWidth =
            (totalColumns * itemSize.width) + itemSize.width +
            ((totalColumns) * getItemGap()) + (getMarginGap() * 2);
    }
    else if (vertical) {
        // Vertical legends -- columns are groups
        legendWidth =
            (totalColumns * itemSize.width) +
            ((totalColumns - 1) * getGroupGap()) + (getMarginGap() * 2);
    }
    else {
        // Horizontal legends -- columns are collections of items
        legendWidth =
            (totalColumns * itemSize.width) +
            ((totalColumns - 1) * getItemGap()) + (getMarginGap() * 2);
    }

    if (isContinuousLayout() && vertical) {
        // Vertical continous legend -- need extra text item plus item gap
        legendHeight = (totalRows * itemSize.height) + itemSize.height +
                       ((totalRows) * getItemGap()) + (getMarginGap() * 2);
    }
    else if (!vertical) {
        // Horizontal legends -- rows are groups
        legendHeight = (totalRows * itemSize.height) +
                       ((totalRows - 1) * getGroupGap()) + (getMarginGap() * 2);
    }
    else {
        // Vertical legends -- rows are collections of items
        legendHeight = (totalRows * itemSize.height) +
                       ((totalRows - 1) * getItemGap()) + (getMarginGap() * 2);
    }

    // Check to see how many title items we have
    int numTitles = 0;
    for (int i = 0; i < numViews; i++) {
        if (hasTitle[i]) {
            numTitles++;
        }
    }

    // Add enough space to the ledgend to account for titles
    if (numTitles > 0) {
        if (vertical) {
            legendWidth = Math.max(legendWidth, (titleSize.width * numViews) +
                                                (getMarginGap() * 2));
            legendHeight += (titleSize.height + getGroupGap());
        }
        else {
            legendHeight += (numTitles * (titleSize.height + getGroupGap()));
            legendWidth = Math.max(legendWidth,
                                   titleSize.width + (getMarginGap() * 2));
        }
    }


    legendDim = new Dimension(legendWidth, legendHeight);
    legendColumns = totalColumns;
    legendRows = totalRows;
    return(legendDim);
}

/**
 * Controls to layout of the items in a stepped legend.
 *
 * @param itemList the list of items to layout
 * @param vertical <code>true</code> if the legend is vertical
 * @param maxItemSize the size of the largest legend item
 * @param legendSize the size of the legend
 * @param hasTitle <code>true</code> of this list of legend items contains a
 * title item
 * @param titleSize the size of the title item
 */
protected void layoutSteppedLegendItems(List itemList, boolean vertical,
			                            Dimension maxItemSize,
										Dimension legendSize,
	                                    boolean hasTitle[], Dimension titleSize)
{
	// Now layout the legend items, filling in the dim, pos, textPos, symPos
	// elements of the legend item

	int row = 0, col = 0;

	int itemx  = getMarginGap();
	int itemy  = getMarginGap();
	int titlex = getMarginGap();
	int titley = getMarginGap();

	int numViews = itemList.size();

	int titleSpace;
	if (vertical) {
		titleSpace = ((legendSize.width - (2 * getMarginGap()) -
                      ((numViews - 1) * getGroupGap())) / numViews);
	}
	else {
		titleSpace = legendSize.width - (2 * getMarginGap());
	}

	for (int i = 0; i < numViews; i++) {
		List viewItems = (List) itemList.get(i);

		// Horizontal layout
		if (!vertical) {
			if (hasTitle[i]) {
				itemy += titleSize.height + getGroupGap();
			}
			for (int j = 0; j < viewItems.size(); j++) {
				JCLegendItem thisItem = (JCLegendItem) viewItems.get(j);

				if (isTitleItem(thisItem)) {
					positionTitleItem(thisItem, titleSpace, titlex, titley);
				}
				else {
					Dimension itemDim = new Dimension(maxItemSize.width,
                                                      maxItemSize.height);
					if (col >= legendColumns) {
						col = 0;
						row++;
						itemy += maxItemSize.height + getGroupGap();
						itemx = getMarginGap();
					}
					if (col != 0) {
						itemx += getItemGap();
					}
					thisItem.pos = new Point(itemx, itemy);
					thisItem.dim = itemDim;
					thisItem.symbolPos = new Point(0, 0);
					thisItem.textPos = new Point(thisItem.symbolDim.width +
												 getInsideItemGap(), 0);
					thisItem.pickRectangle = new Rectangle(thisItem.pos.x,
                                                           thisItem.pos.y,
							                               thisItem.dim.width,
                                                           thisItem.dim.height);
					itemx += maxItemSize.width;
					col++;
				}
			}
			col = 0;
			row++;
			itemy += maxItemSize.height + getGroupGap();
			itemx = getMarginGap();
			titley = itemy;
		}
        // Vertical layout
		else {
			itemy = getMarginGap();
			if (hasTitle[i]) {
				itemy += titleSize.height + getGroupGap();
			}
			for (int j = 0; j < viewItems.size(); j++) {
				JCLegendItem thisItem = (JCLegendItem) viewItems.get(j);

				if (isTitleItem(thisItem)) {
					positionTitleItem(thisItem, titleSpace, titlex, titley);
				}
                else {
					Dimension itemDim = new Dimension(maxItemSize.width,
                                                      maxItemSize.height);
					if (row >= legendRows) {
						row = 0;
						col++;
						itemx += maxItemSize.width + getGroupGap();
						itemy = getMarginGap();
						if (hasTitle[i]) {
							itemy += titleSize.height + getGroupGap();
						}
					}
					if (row != 0) {
						itemy += getItemGap();
					}
					thisItem.pos = new Point(itemx, itemy);
					thisItem.dim = itemDim;
					thisItem.symbolPos = new Point(0, 0);
					thisItem.textPos = new Point(thisItem.symbolDim.width +
												 getInsideItemGap(), 0);
					thisItem.pickRectangle = new Rectangle(thisItem.pos.x,
                                                           thisItem.pos.y,
                                                           thisItem.dim.width,
                                                           thisItem.dim.height);
					itemy += maxItemSize.height;
					row++;
				}
			}
			row = 0;
			col++;
			itemx += maxItemSize.width + getGroupGap();
			itemy = getMarginGap();

			titlex += titleSize.width + getGroupGap();
			if (itemx < titlex) {
				itemx = titlex;
			}
			else if (titlex < itemx) {
				titlex = itemx;
			}
		}
	}
}

/**
 * Controls the layout of the items in a continuous legend.
 *
 * @param itemList the list of items to layout
 * @param vertical <code>true</code> if the legend is vertical
 * @param maxItemSize the size of the largest legend item
 * @param legendSize the size of the legend
 * @param hasTitle <code>true</code> if this list of legend items contains a
 * title item
 * @param titleSize the size of the title item
 */
public void layoutContinuousLegendItems(List itemList, boolean vertical,
							            Dimension maxItemSize,
										Dimension legendSize,
							            boolean hasTitle[], Dimension titleSize)
{
	// Now layout the legend items, filling in the dim, pos, textPos, symPos
	// elements of the legend item

	int row = 0, col = 0;

	int itemx  = getMarginGap();
	int itemy  = getMarginGap();
	int titlex = getMarginGap();
	int titley = getMarginGap();

	int numViews = itemList.size();

	int titleSpace;
	if (vertical) {
		titleSpace = ((legendSize.width - (2 * getMarginGap()) -
                       ((numViews - 1) * getGroupGap())) / numViews);
	}
	else {
		titleSpace = legendSize.width - (2 * getMarginGap());
	}

	for (int i = 0; i < numViews; i++) {
		List viewItems = (List) itemList.get(i);

		// Horizontal layout
		if (!vertical) {
			if (hasTitle[i]) {
				itemy += titleSize.height + getGroupGap();
			}
			for (int j = 0; j < viewItems.size(); j++) {
				JCLegendItem thisItem = (JCLegendItem) viewItems.get(j);

				if (isTitleItem(thisItem)) {
					positionTitleItem(thisItem, titleSpace, titlex, titley);
				}
				else {
					Dimension itemDim =
                        new Dimension(((int)(maxItemSize.width * 1.5) +
                                       getItemGap()),
						              maxItemSize.height);
					// At end of column -- create new text only legend item and
					// append to end of column before moving on to next column
					if (col >= legendColumns) {
						itemDim = new Dimension(maxItemSize.width,
                                                maxItemSize.height);
						// But don't create one in the last row as the last row
						// already has a symbol-less text item at the end
						if (row != legendRows - 1) {
							thisItem = new JCLegendItem(thisItem);
							thisItem.drawType = JCLegend.NONE;
							thisItem.pickRectangle = null;

							viewItems.add(j, thisItem);
						}
					}

					// Position the text and symbol for this legend item
					if (col != 0) {
						itemx += getItemGap();
					}
					thisItem.pos = new Point(itemx, itemy);
					thisItem.dim = itemDim;
					thisItem.textPos =
                        new Point((maxItemSize.width -
                                   thisItem.textDim.width) / 2,
						          thisItem.symbolDim.height +
                                  getInsideItemGap());
					thisItem.symbolPos = new Point(maxItemSize.width / 2, 0);
					if (thisItem.drawType != JCLegend.NONE) {
						thisItem.pickRectangle =
                            new Rectangle(thisItem.pos.x + thisItem.symbolPos.x,
							              thisItem.pos.y,
							              thisItem.symbolDim.width,
                                          thisItem.dim.height);
					}
					else {
						thisItem.pickRectangle = null;
					}

					// Move on to next column
					if (col >= legendColumns) {
						col = 0;
						row++;
						itemy += maxItemSize.height + getGroupGap();
						itemx = getMarginGap();
					}
                    // Move on to next item in column
					else {
						itemx += maxItemSize.width;
						col++;
					}
				}
			}
			col = 0;
			row++;
			itemy += maxItemSize.height + getGroupGap();
			itemx = getMarginGap();
			titley = itemy;
		}
        // Vertical layout
		else {
			if (hasTitle[i]) {
				itemy += titleSize.height + getGroupGap();
			}
			for (int j = 0; j < viewItems.size(); j++) {
				JCLegendItem thisItem = (JCLegendItem) viewItems.get(j);

				if (isTitleItem(thisItem)) {
					positionTitleItem(thisItem, titleSpace, titlex, titley);
				}
				else {
					Dimension itemDim =
                        new Dimension(maxItemSize.width,
							          (int) (maxItemSize.height * 1.5) +
                                      getItemGap());
					// At end of row -- create new text only legend item and
					// append to end of row before moving on to next row
					if (row >= legendRows) {
						itemDim = new Dimension(maxItemSize.width,
                                                maxItemSize.height);
						// But don't create one in the last column as the last
						// column already has a symbol-less text item at the end
						if (col != legendColumns - 1) {
							thisItem = new JCLegendItem(thisItem);
							thisItem.drawType = JCLegend.NONE;

							viewItems.add(j, thisItem);
						}
					}

					// Position the text and symbol for this legend item
					if (row != 0) {
						itemy += getItemGap();
					}
					thisItem.pos = new Point(itemx, itemy);
					thisItem.dim = itemDim;
					thisItem.textPos = new Point(thisItem.symbolDim.width +
											     getInsideItemGap(),
                                                 0);
					thisItem.symbolPos = new Point(0,
                                                   thisItem.textDim.height / 2);
					if (thisItem.drawType != JCLegend.NONE) {
						thisItem.pickRectangle =
                            new Rectangle(thisItem.pos.x,
                                          thisItem.pos.y + thisItem.symbolPos.y,
                                          thisItem.dim.width,
                                          thisItem.symbolDim.height);
					}
					else {
						thisItem.pickRectangle = null;
					}

					// Move on to next row
					if (row >= legendRows) {
						row = 0;
						col++;
						itemx += maxItemSize.width + getGroupGap();
						itemy = getMarginGap();
						if (hasTitle[i]) {
							itemy += titleSize.height + getGroupGap();
						}
					}
                    // Move on to next item in row
					else {
						itemy += maxItemSize.height;
						row++;
					}
				}
			}
			row = 0;
			col++;
			itemx += maxItemSize.width + getGroupGap();
			itemy = getMarginGap();

			titlex += titleSize.width + getGroupGap();
			if (itemx < titlex) {
				itemx = titlex;
			}
			else if (titlex < itemx) {
				titlex = itemx;
			}
		}
	}
}

/**
 * Positions the title item above the legend.
 *
 * @param thisItem the legend item containing the legend title
 * @param space the space available
 * @param titlex the <i>x</i> position of space available
 * @param titley the <i>y</i> position of space available
 */
protected void positionTitleItem(JCLegendItem thisItem, int space,
								 int titlex, int titley)
{
	thisItem.pos = new Point(titlex + ((space - thisItem.textDim.width) / 2),
		                     titley);
	thisItem.dim = new Dimension(thisItem.textDim.width,
								 thisItem.textDim.height);
	thisItem.textPos = new Point(0, 0);
}

/**
 * Overrides parent legend pick to change default behaviour.  By default,
 * the legend will create its own pick rectangle if one is absent.
 * For the purposes of 3d chart legends, a null pick rectangle should
 * prevent the pick operation.
 * @return The <code>JCLegendItem</code> instance representing the picked point.
 * @param p The point at which to pick
 */
public JCLegendItem pick(Point p)
{
	JCLegendItem pickItem = super.pick(p);
	if (pickItem != null) {
		if (pickItem.pickRectangle == null) {
			pickItem = null;
		}
	}

	return(pickItem);
}

}
