/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCChart3dLegendManager.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.util.JCNumberUtil;
import com.klg.jclass.util.legend.JCLegend;
import com.klg.jclass.util.legend.JCLegendItem;
import com.klg.jclass.util.legend.JCLegendPopulator;
import com.klg.jclass.util.legend.JCLegendRenderer;

import java.awt.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;

/**
 * This is the default implementation of both <code>JCLegendPopulator</code> and
 * <code>JCLegendRenderer</code>. These interfaces, respectively, control the
 * creation and display characteristics of the <code>JCLegendItem</code>s.
 *
 * @see com.klg.jclass.util.legend.JCLegend#setLegendPopulator
 * @see com.klg.jclass.util.legend.JCLegend#setLegendRenderer
 */
public class JCChart3dLegendManager implements JCLegendRenderer,
                                               JCLegendPopulator,
                                               java.io.Serializable
{

static final long serialVersionUID = 6132225179285554488L;

/** The 3d chart instance containing the legend. */
protected JCChart3d	chart3d;

/** The legend being managed by this class. */
protected JCLegend	legend;

/** User-specified outline color for legend items. */
protected Color		outlineColor = Color.black;

/** List of legend items to be passed to the legend. */
protected Vector    legendItems = null;

/** The gap between fields in <code>JCMultiColumnStrings</code>. */
protected int fieldGap = 2;

/**
 * The max value of each field used in a <code>JCMultiColumnString</code>
 * object.
 */
protected int[] maxFieldWidths = null;

//enums used for determining the legend style internally
static final protected int STYLE_NONE = 0;
static final protected int STYLE_BOX_RANGE = 1;
static final protected int STYLE_LINE_RANGE = 2;
static final protected int STYLE_SYMBOL_RANGE = 3;
static final protected int STYLE_CONTOUR = 4;
static final protected int STYLE_CNTN_H = 5;
static final protected int STYLE_CNTN_V = 6;

/**
 * Creates a <code>Populator</code> and <code>Renderer</code> for a
 * <code>JCLegend</code>.
 *
 * @param chart3d
 * @param legend the <code>JCLegend</code> that this class will populate
 *     and render <code>JCLegendItems</code> for
 */
public JCChart3dLegendManager(JCChart3d chart3d, JCLegend legend)
{
	this.chart3d = chart3d;
	this.legend	= legend;
}

/**
 * Sets the value to use as a gap between fields in
 * <code>JCMultiColumnString</code> objects.
 *
 * @param gap the field gap
 */
public void setFieldGap(int gap)
{
	fieldGap = gap;
}

/**
 * Gets the value being used as a gap between fields in
 * <code>JCMultiColumnString</code> objects.
 *
 * @return the field gap
 */
public int getFieldGap()
{
	return(fieldGap);
}

///////////// Implement the JCLegendPopulator interface

/**
 * Creates legend items for each range listed in the ranges list.
 * This uses the range name and range foreground color to create the
 * <code>JCLegendItem</code>s.
 * @return
 * @param fm FontMetrics - information about the font in which legend text will
 *     be written
 */
public List getLegendItems(FontMetrics fm) {

	legendItems = new Vector();
	List dataViewList = chart3d.getOriginalDataView();
	int numViews = dataViewList.size();
	for (int i = 0; i < numViews; i++) {
		legendItems.add(new Vector());
	}

	// If legend implements the 3d legend template, note it so that
	// we can make use of its features
	List userLabelList = null;
	JCChart3dLegendTemplate threeDLegend = null;
	if (legend instanceof JCChart3dLegendTemplate) {
		threeDLegend = (JCChart3dLegendTemplate) legend;
		userLabelList = threeDLegend.getLabels();
	}

	int numTextFields = -1; // Num of fields used by JCMultiFieldString objects
	int symbolSize	= legend.getSymbolSize();

	for (int i = 0; i < numViews; i++) {
        List viewItems = (List) legendItems.get(i);

		Chart3dDataView view = (Chart3dDataView) dataViewList.get(i);

		// Add the legend item representing the legend title
		String viewLabel = view.getName();
		JCLegendItem titleItem = new JCLegendItem();
		// DAM check for HTML
		if ((viewLabel != null) && (viewLabel.length() > 0)) {
			titleItem.textDim = new Dimension(fm.stringWidth(viewLabel),
                                              fm.getHeight());
			titleItem.contents = viewLabel;
		}
        else {
			titleItem.textDim = new Dimension(0, 0);
			titleItem.contents = null;
		}
		viewItems.add(titleItem);

		boolean drawScatter = (view.getChartType() == JCChart3d.SCATTER);

		// Get user set labels if there are any
		List userLabels = null;
		if (threeDLegend != null) {
			if (userLabelList != null && i < userLabelList.size()) {
				userLabels = (List) userLabelList.get(i);
			}
		}

		int legendStyle = STYLE_NONE;
		Chart3dData elevationData = view.getElevationData();
		if (drawScatter && elevationData instanceof Chart3dPointData) {
			legendStyle = STYLE_SYMBOL_RANGE;
			Chart3dPointData pointData = (Chart3dPointData) elevationData;

			int numSeries = pointData.getNumSeries();
			for (int j = 0; j < numSeries; j++) {
				Chart3dPointSeries series = pointData.getSeries(j);

				// Get label for this legend item.  If user labels are
                // specified, use them otherwise, get label from series
				Object label = null;
				if (userLabels != null && j < userLabels.size()) {
					label = userLabels.get(j);
				}
                else {
					label = series.getLabel();
				}
				if (label != null) {
					// Have label, create and populate legend item
					JCData3dPointIndex pointIndex =
                        new JCData3dPointIndex(view, j, -1);
					LegendEntry entry =
                        new LegendEntry(pointIndex,
                                        series.getChartStyle(),
                                        series.getChartStyle().getSymbolStyle());

					JCLegendItem item = new JCLegendItem();
					numTextFields =
                        populate3dLegendItem(item, fm, label,
                                             JCLegend.CUSTOM_SYMBOL,
                                             new Dimension(symbolSize,
                                                           symbolSize),
                                             entry, numTextFields);
    				viewItems.add(item);
				}
			}
		}
        else {
			Chart3dGridData gridData = null;
			boolean gridScatter = false;
			if (elevationData instanceof Chart3dGridData) {
 				gridData = (Chart3dGridData) elevationData;
                if (drawScatter) {
				    gridScatter = true;
                }
			}

			// Get the mappings between levels and data that determine
			// the legend items
			JCElevation elevation = view.getElevation();
			JCContour contour = view.getContour();

			JCChart3dArea area = view.getChart3d().getChart3dArea();
			JCProjection floor = area.getPlotCube().getFloor();
			JCProjection ceiling = area.getPlotCube().getCeiling();

			boolean projectContours = (floor.isContoured() ||
									ceiling.isContoured());
			boolean projectZones = (floor.isZoned() ||
									ceiling.isZoned());

			boolean drawContours = (contour.isContoured() || projectContours);
			boolean drawZones = (contour.isZoned() || projectZones);

			// Determine the legend style to use for this legend
			if ((contour.isZoned() &&
				 (!elevation.isMeshed() || elevation.isShaded())) ||
			    projectZones)
            {
				if (gridScatter) {
					legendStyle = STYLE_SYMBOL_RANGE;
                    if (threeDLegend != null) {
                        threeDLegend.setContinuousLayout(false);
                    }
				}
                else if (threeDLegend != null) {
					if (threeDLegend.getLayoutStyle() == threeDLegend.CONTINUOUS){
						if (legend.getOrientation() == JCLegend.HORIZONTAL) {
							legendStyle = STYLE_CNTN_H;
						}
                        else {
							legendStyle = STYLE_CNTN_V;
						}
						threeDLegend.setContinuousLayout(true);
					}
                    else {
						legendStyle = STYLE_BOX_RANGE;
						threeDLegend.setContinuousLayout(false);
					}
				}
                else {
					legendStyle = STYLE_BOX_RANGE;
				}
			}
            else if (contour.isZoned()    &&
                     elevation.isMeshed() &&
					 !elevation.isShaded())
            {
				if (gridScatter) {
					legendStyle = STYLE_SYMBOL_RANGE;
				}
                else {
					legendStyle = STYLE_LINE_RANGE;
				}
                if (threeDLegend != null) {
                    threeDLegend.setContinuousLayout(false);
                }
			}
            else if (!contour.isZoned() && drawContours) {
				if (gridScatter) {
					legendStyle = STYLE_SYMBOL_RANGE;
				}
                else {
					legendStyle = STYLE_CONTOUR;
				}
                if (threeDLegend != null) {
                    threeDLegend.setContinuousLayout(false);
                }
			}
            // Special case -- grid data scatter chart with no contours
            // or zones (but mesh or shading is still on).  In this case
            // the legend has a single item, derived from information in
            // the grid data object
            else if (gridScatter && !drawContours && !drawZones &&
					 (elevation.isMeshed() || elevation.isShaded()))
            {
				legendStyle = STYLE_SYMBOL_RANGE;
                if (threeDLegend != null) {
                    threeDLegend.setContinuousLayout(false);
                }
				// Get label for this legend item
				// if user labels are specified, use them
				// otherwise, get label from series
				Object label = null;
				if (userLabels != null && userLabels.size() > 0) {
					label = userLabels.get(i);
				}
                else {
					label = gridData.getName();
				}
                if (label != null) {
                    // Have label, create and populate legend item
                    JCData3dGridIndex gridIndex =
                        new JCData3dGridIndex(view,
                                              Integer.MAX_VALUE,
                                              Integer.MAX_VALUE);
                    JCChart3dStyle chartStyle = gridData.getChartStyle();
                    LegendEntry entry =
                        new LegendEntry(gridIndex,chartStyle,
                                        chartStyle.getSymbolStyle());

                    JCLegendItem item = new JCLegendItem();
                    numTextFields =
                        populate3dLegendItem(item, fm, label,
                                             JCLegend.CUSTOM_SYMBOL,
                                             new Dimension(symbolSize,
                                                           symbolSize),
                                             entry, numTextFields);
                    viewItems.add(item);
				}
				continue;
			}

			if (legendStyle == STYLE_NONE) {
				if (viewItems.size() > 0) {
					viewItems.clear();
                }
				continue;
			}

			JCChart3dLegendLabelGenerator labelGenerator = null;
			if (threeDLegend != null) {
				labelGenerator = threeDLegend.getLabelGenerator();
			}

			// Filter the levels into a series of numbers and contour
			// styles ready to create labels from
			Vector levels = filterLevels(view, legendStyle, threeDLegend,
                                         contour);

			JCContourLevels contourLevels = contour.getContourLevels();
			JCChart3dNumberFormat format = contourLevels.getNumberFormatter();
			int nfrac = contourLevels.getNFrac();

			// Create legend items depending on legend style and list of levels
			for (int j = levels.size() - 1; j >= 0; j--) {
				LegendEntry entry = (LegendEntry) levels.get(j);

				Object label = getNonScatterLabel(legendStyle, format, nfrac,
                                                  levels, userLabels, j,
                                                  labelGenerator);

				if (label == null) {
					continue;
                }

				JCLegendItem item = null;

				// Create & populate a legend item depending on style
                Dimension symbolDim = new Dimension(symbolSize * 2, symbolSize);
				if (legendStyle == STYLE_CNTN_H || legendStyle == STYLE_CNTN_V)
                {
					item = new JCLegendItem();
					// Continuous legend -- most items are associated with
					// a box derived from the fill style
					if (entry.parentStyle != null) {
						entry.drawStyle =
							((JCContourStyle)entry.parentStyle).getFillStyle();
						if (legendStyle == STYLE_CNTN_V) {
							int height = fm.getHeight() +
                                         threeDLegend.getItemGap();
                            symbolDim.setSize(symbolSize * 2, height);
                            numTextFields =
                                populate3dLegendItem(item, fm, label,
                                                     JCLegend.BOX_PLAIN,
                                                     symbolDim, entry,
                                                     numTextFields);
						}
                        else {
                            // Fill in symbol width later
                            symbolDim.setSize(0, symbolSize);
                            numTextFields =
                                populate3dLegendItem(item, fm, label,
                                                     JCLegend.BOX_PLAIN,
                                                     symbolDim, entry,
                                                     numTextFields);
						}
					}
                    // Continuous legend -- but the last item is text only,
                    // the passed style being nothing more than a place holder
                    else {
						entry.drawStyle = null;
                        numTextFields =
                            populate3dLegendItem(item, fm, label,
                                                 JCLegend.NONE, symbolDim,
                                                 entry, numTextFields);
					}
				}
                else {
					item = new JCLegendItem();

					// Box range legend -- legend item is box, colour derived
					// from contour fill style
                    JCContourStyle cStyle = (JCContourStyle)entry.parentStyle;
					if (legendStyle == STYLE_BOX_RANGE) {
                        entry.drawStyle = cStyle.getFillStyle();
						numTextFields =
                            populate3dLegendItem(item, fm, label,
                                                 JCLegend.BOX, symbolDim,
                                                 entry, numTextFields);

					}
                    // Line range legend -- legend item is line, colour derived
                    // from contour fill style
                    else if (legendStyle == STYLE_LINE_RANGE) {
						entry.drawStyle = cStyle.getFillStyle();
						numTextFields =
                            populate3dLegendItem(item, fm, label,
                                                 JCLegend.LINE, symbolDim,
                                                 entry, numTextFields);

					}
                    else if (legendStyle == STYLE_SYMBOL_RANGE) {
						entry.drawStyle = cStyle.getSymbolStyle();
                        symbolDim.setSize(symbolSize, symbolSize);
						numTextFields =
                            populate3dLegendItem(item, fm, label,
                                                 JCLegend.CUSTOM_SYMBOL,
                                                 symbolDim, entry,
                                                 numTextFields);
					}
                    // Contour legend -- legend item is line, colour derived
                    // from contour line style
                    else if (legendStyle == STYLE_CONTOUR) {
						entry.drawStyle = cStyle.getLineStyle();
                        numTextFields =
                            populate3dLegendItem(item, fm, label,
                                                 JCLegend.LINE, symbolDim,
                                                 entry, numTextFields);
					}
				}

				// Finally add it
				if (item != null) {
	    			viewItems.add(item);
                }
			}
		}
	}

	// Adjust text widths due to number of fields in text items
	calcTextWidthsFromFields(numTextFields, legendItems, threeDLegend, fm);

    return(legendItems);
}


/**
 * Filters the contour levels into a list that can be used to create legend
 * item labels.
 * @return
 * @param view
 * @param legendStyle style of this legend; one of <code>STYLE_CNTN_H</code>,
 *     <code>STYLE_CNTN_V</code>, <code>STYLE_CONTOUR</code>,
 *     <code>STYLE_BOX_RANGE</code>, or <code>STYLE_LINE_RANGE</code>
 * @param threeDLegend the legend class if it implements the
 *     <code>JCChart3DLegendTemplate</code> interface
 * @param contour the <code>JCContour</code> object representing the contour
 *     levels
 */
protected Vector filterLevels(Chart3dDataView view, int legendStyle,
		                      JCChart3dLegendTemplate threeDLegend,
							  JCContour contour)
{
	Vector filteredLevels = new Vector();

	JCContourLevels levels = contour.getContourLevels();
    int numLevels = levels.getNumLevelsInternal();
	double[] levelsArray = levels.getLevels();
	int nfrac = levels.getNFrac();

    ArrayList styles = contour.getContourStyles();
    JCContourMapping mapping = contour.getContourMapping();

	boolean constrainToData = false;
	if (threeDLegend != null) {
		constrainToData =
			(threeDLegend.getDistributionRange() == threeDLegend.RANGE_DATA);
	}


	// Match up each contour level with the correct style as defined by
	// mapping and create legend items.  Note: there's always one more
	// level than there are styles.
	double minData = levels.getMin();
	double maxData = levels.getMax();

	// If using default levels or constraining to data, continuous style
	// legends need an extra lower text only data item equal to the data min
    JCData3dContourIndex ci = null;
	if (levels.isDefault() || constrainToData) {
		if (legendStyle == STYLE_CNTN_H || legendStyle == STYLE_CNTN_V) {
            ci = new JCData3dContourIndex(view, -1, minData, minData);
            filteredLevels.add(new LegendEntry(ci, null));
		}
	}

	int lastLevelUsed = -1;
	for (int j = 0; j < numLevels; j++) {
		double levelValue = levelsArray[j];

		// If this value is outside the data constraints, continue
		// until we find data that is inside
		if (constrainToData && (levelValue < minData || levelValue > maxData)) {
			continue;
		}

        // map the levels array index to the proper contour style index
        int index = mapping.contourIndex(j);
        JCContourStyle style = (JCContourStyle) styles.get(index);

		if (legendStyle == STYLE_CONTOUR) {
			// Add level value item
			// Contour style legends use level values, not ranges
            ci = new JCData3dContourIndex(view, j, levelValue, levelValue);
            filteredLevels.add(new LegendEntry(ci, style));
		}
        else {
			// Add first level value item
			if (lastLevelUsed < 0) {
				// If using default levels or constraining levels to data
				// the first legend entry should go from the min to the
				// current contour level
				if (levels.isDefault() || constrainToData) {
					String formattedMin = JCNumberUtil.format(minData, nfrac);
					String formattedLevel = JCNumberUtil.format(levelValue,
                                                                nfrac);

					// But only add it if it is not a duplicate of the min
					// value that has already been added
					if (!formattedMin.equals(formattedLevel)) {
                        ci = new JCData3dContourIndex(view, j, minData,
                                                      levelValue);
                        filteredLevels.add(new LegendEntry(ci, style));
					}
				}
                // Otherwise, in the continuous case, the first entry
                // is text only and represents the lowest user-supplied level
                else if (legendStyle == STYLE_CNTN_H ||
						 legendStyle == STYLE_CNTN_V)
                {
                    ci = new JCData3dContourIndex(view, j, levelValue,
                                                  levelValue);
                    filteredLevels.add(new LegendEntry(ci, null));
				}
			}
            // Add any other level value item
            else {
                ci = new JCData3dContourIndex(view, j,
                                              levelsArray[lastLevelUsed],
                                              levelValue);
                filteredLevels.add(new LegendEntry(ci, style));
			}
		}
		lastLevelUsed = j;
	}

	// If using default levels or constrining to data, item is needed for the
    // maximum data value
	if (levels.isDefault() || constrainToData) {
	    if (legendStyle == STYLE_CNTN_H     ||
			legendStyle == STYLE_CNTN_V     ||
			legendStyle == STYLE_BOX_RANGE  ||
			legendStyle == STYLE_LINE_RANGE ||
            legendStyle == STYLE_SYMBOL_RANGE)
        {
			if (lastLevelUsed < 0) {
               	int index = mapping.contourIndex(numLevels);
               	JCContourStyle style = (JCContourStyle)styles.get(index);
                ci = new JCData3dContourIndex(view, numLevels, minData,
                                              maxData);
				filteredLevels.add(new LegendEntry(ci, style));
			}
            else {
				String formattedMax = JCNumberUtil.format(maxData, nfrac);
				String formattedLevel =
                    JCNumberUtil.format(levelsArray[lastLevelUsed], nfrac);
				// Ensure that the max item is not a duplicate of the highest
				// level value that was already added
				if (!formattedMax.equals(formattedLevel)) {
               		int index = mapping.contourIndex(lastLevelUsed + 1);
               		JCContourStyle style = (JCContourStyle) styles.get(index);
                    ci = new JCData3dContourIndex(view, lastLevelUsed + 1,
                                                  levelsArray[lastLevelUsed],
                                                  maxData);
					filteredLevels.add(new LegendEntry(ci, style));
				}
			}
		}
        // Special case, flat surface, where min = max so the levels array is
        // null and no contours were generated
        else if (legendStyle == STYLE_CONTOUR && numLevels == 0) {
            int index = mapping.contourIndex(numLevels);
            JCContourStyle style = (JCContourStyle)styles.get(index);
            ci = new JCData3dContourIndex(view, numLevels, maxData, maxData);
            filteredLevels.add(new LegendEntry(ci, style));
        }
	}

	return(filteredLevels);
}

/**
 * Gets the label for a legend item in a non-Scatter legend.
 * @param legendStyle style of this legend; one of <code>STYLE_CNTN_H</code>,
 *     <code>STYLE_CNTN_V</code>,  <code>STYLE_CONTOUR</code>,
 *     <code>STYLE_BOX_RANGE</code>, or <code>STYLE_LINE_RANGE</code>
 * @param formatter class used to format labels from numbers
 * @param nfrac the number of fractional digits in the formatted label
 * @param levels the array of range levels (used for default labels)
 * @param userLabels a List of user-set labels; possibly null
 * @param index the index of the current range level
 * @param labelGenerator user-set generator for custom labels; possibly null
 * @return the label to use for this legend item
 */
protected Object getNonScatterLabel(int legendStyle,
                                    JCChart3dNumberFormat formatter, int nfrac,
                                    Vector levels, List userLabels, int index,
                                    JCChart3dLegendLabelGenerator labelGenerator)
{
	Object label = null;
	LegendEntry entry = (LegendEntry) levels.get(index);
	JCData3dContourIndex dataIndex = (JCData3dContourIndex) entry.dataIndex;

	double maxValue = dataIndex.getUpperContourRangeValue();
	double minValue = dataIndex.getLowerContourRangeValue();

	String formattedMax = JCNumberUtil.format(maxValue, nfrac);
	String formattedMin = JCNumberUtil.format(minValue, nfrac);

	formattedMax = formatter.localize(formattedMax);
	formattedMin = formatter.localize(formattedMin);

	// If there's a user specified label, use that
	if (userLabels != null && userLabels.size() > index) {
		label = userLabels.get(index);
	}
    // If not, generate a default label depending on the legend style
    else {
		// Continuous horizontal -- center level array value
		if (legendStyle == STYLE_CNTN_H) {
			label = new JCMultiFieldString("\\c" + formattedMax);

			// Single value label, so ensure that min and max are
			// the same for the purposes of the label generator
			minValue = maxValue;
		}
        // Continuous vertical -- right justify level array value
        else if (legendStyle == STYLE_CNTN_V) {
			label = new JCMultiFieldString("\\r" + formattedMax);

			// Single value label, so ensure that min and max are
			// the same for the purposes of the label generator
			minValue = maxValue;
		}
        // Range types -- create label from range
        else if (legendStyle == STYLE_BOX_RANGE  ||
				 legendStyle == STYLE_LINE_RANGE ||
				 legendStyle == STYLE_SYMBOL_RANGE)
        {
			label = new JCMultiFieldString("\\r" + formattedMin +
							               "\\c" + " .. " +
							               "\\r" + formattedMax);
		}
        else if (legendStyle == STYLE_CONTOUR) {
			label = new JCMultiFieldString("\\r" + formattedMax);

			// Single value label, so ensure that min and max are
			// the same for the purposes of the label generator
			minValue = maxValue;
		}
	}

	// Defer to the legend generator is one has been set
	if (labelGenerator != null) {
        int ci = dataIndex.getContourStyleIndex();
        label = labelGenerator.generateLegendLabel(chart3d, ci, minValue,
                                                   maxValue, label);
	}

	return(label);
}

/**
 * Populates a legend item for the 3d legend.  If the label is a
 * <code>JCMultiFieldString</code>, it notes the number of fields and returns
 * the value.  Item text dimensions are determined from the label parameter with
 * passed fontmetrics.  Other parameters are passed to the legend item as is.
 *
 * @param item the legend item to populate
 * @param fm the font metrics to use for sizing the passed legend label
 * @param label the label to use for the contents of this legend item
 * @param drawType the draw type to use for this legend item
 * @param symbolDim the symbol dimensions to use for this legend item
 * @param itemInfo the <code>itemInfo</code> field to use for this legend item
 * @param numTextFields the maximum number of fields contained in any
 *     <code>JCMultiFieldString</code> objects used so far
 * @return if using <code>JCMultiFieldString</code> objects for labels, this
 *     value will be the maximum number of fields used so far; if not using
 *     such any such objects, the return value is <i>-1</i>
 */
protected int populate3dLegendItem(JCLegendItem item, FontMetrics fm,
                                   Object label, int drawType,
                                   Dimension symbolDim,
                                   Object itemInfo, int numTextFields)
{
	if (item  == null) {
		return(numTextFields);
    }

	int width = 0;
	int height = fm.getHeight();

	// Accumulate useful info, like max size of each field for later sizing
	if (label != null) {
		if (label instanceof JCMultiFieldString) {
			JCMultiFieldString mfsLabel = (JCMultiFieldString) label;
			List mfsContentList = mfsLabel.getContentList();
			numTextFields = Math.max(numTextFields,
								     mfsContentList.size() / 2);
			// We calc the width later
			width = 0;
		}
        else if (label instanceof String) {
			// DAM handle HTML
			width = fm.stringWidth((String)label);
		}
	}

	item.contents		= label;
	item.drawType 		= drawType;
	item.textDim		= new Dimension(width, height);
	item.symbolDim		= symbolDim;
	item.itemInfo		= itemInfo;
	return(numTextFields);
}


/**
 * Runs through the given list of legend items and adjusts text sizes for
 * the number of fields being used by <code>JCMultiFieldString</code> objects.
 *
 * @param numTextFields the maximum number of text fields being used by
 *     <code>JCMultiFieldString</code> objects in this legend
 * @param legendItems the list of legend items
 * @param threeDLegend the legend class if it implements the
 *     <code>JCChart3DLegendTemplate</code> interface
 * @param fm
 */
protected void calcTextWidthsFromFields(int numTextFields, List legendItems,
                                        JCChart3dLegendTemplate threeDLegend,
                                        FontMetrics fm)
{
	int numViews = legendItems.size();

	// If 1 or more fields being used, create array to hold
	// their maximum sizes
	if (numTextFields > 0) {
		// calculate maximum width of each field
		maxFieldWidths = new int[numTextFields];
		Arrays.fill(maxFieldWidths, 0);
	}

	// Cycle through each legend item to find the maximum size of each field
	// and (if present) the maximum non-fielded string
	int maxStringWidth = 0;
	for (int i = 0; i < numViews; i++) {
       	List viewItems = (List) legendItems.get(i);
		for (int j = 0; j < viewItems.size(); j++) {
       		JCLegendItem thisItem = (JCLegendItem) viewItems.get(j);
			if (isTitleItem(thisItem)) {
				continue;
            }

			if (thisItem.contents instanceof JCMultiFieldString) {
				JCMultiFieldString mfstring =
					(JCMultiFieldString) thisItem.contents;
				List mfsContentList = mfstring.getContentList();
				for (int k = 0; k < mfsContentList.size() / 2; k++) {
					// DAM check for HTML text
                    String str = (String)mfsContentList.get((k*2)+1);
                    maxFieldWidths[k] = Math.max(maxFieldWidths[k],
						                         fm.stringWidth(str));
				}
			}
            else if (thisItem.contents instanceof String) {
                String str = (String)thisItem.contents;
                maxStringWidth = Math.max(maxStringWidth, fm.stringWidth(str));
			}
		}
	}

	// Determine maximum size of legend item text portion that
	int maxContentWidth = 0;
	if (numTextFields > 0) {
		for (int i = 0; i < maxFieldWidths.length; i++) {
			maxContentWidth += maxFieldWidths[i];
			if (i < maxFieldWidths.length - 1) {
				maxContentWidth += fieldGap;
            }
		}
	}
	maxContentWidth = Math.max(maxContentWidth, maxStringWidth);

	// And use that width to adjust the text width of all legend items
	for (int i = 0; i < numViews; i++) {
       	List viewItems = (List) legendItems.get(i);
		for (int j = 0; j < viewItems.size(); j++) {
       		JCLegendItem thisItem = (JCLegendItem) viewItems.get(j);
			if (isTitleItem(thisItem)) {
				continue;
            }

			if (thisItem.contents instanceof JCMultiFieldString) {
				thisItem.textDim.width = maxContentWidth;
			}
			if (threeDLegend != null) {
				if (threeDLegend.isContinuousLayout() &&
					(legend.getOrientation() == JCLegend.HORIZONTAL))
                {
					thisItem.symbolDim.width = maxContentWidth +
										       threeDLegend.getItemGap();
				}
			}
		}
	}
}

/**
 * Indicates if an item is a title, creating a way to identify it from all the
 * other <code>JCLegendItem</code>s. In this case, a null <code>itemInfo</code>
 * indicates that this is a title.
 *
 * @param item
 * @return <code>true</code> if this item is the legend title
 */
public boolean isTitleItem(JCLegendItem item)
{
	return(item.itemInfo == null);
}

///////////// implement the JCLegendRenderer interface

/**
 * For implementing custom legend item draws.  For use with
 * <code>CUSTOM_ALL</code> draw type.
 *
 * @param gc tbe graphics object used to draw
 * @param useFont the font to use
 * @param thisItem the item to draw
 */
public void drawLegendItem(Graphics gc, Font useFont, JCLegendItem thisItem)
{
	// No implementation necessary
}

/**
 * For implementing custom legend item symbol draws.  For use with
 * <code>CUSTOM_SYMBOL</code> draw type.
 *
 * @param gc tbe graphics object used to draw
 * @param useFont the font to use
 * @param thisItem the item to draw
 */
public void drawLegendItemSymbol(Graphics gc, Font useFont,
                                 JCLegendItem thisItem)
{
	if (thisItem.itemInfo == null  ||
        thisItem.symbolDim == null ||
		thisItem.symbolPos == null)
    {
		return;
    }

	int symX = thisItem.pos.x + thisItem.symbolPos.x;
	int symY = thisItem.pos.y + thisItem.symbolPos.y;

	if (thisItem.itemInfo instanceof LegendEntry) {
		LegendEntry entry = (LegendEntry) thisItem.itemInfo;
		// scatter plot -- draw symbol
		if (entry.drawStyle != null &&
			entry.drawStyle instanceof JCSymbolStyle)
        {
			JCSymbolStyle style = (JCSymbolStyle) entry.drawStyle;

			// Draw with drop lines
			JCScatter scatter = chart3d.getChart3dArea().getScatter();
			if (scatter != null        &&
                scatter.hasDropLines() &&
				entry.parentStyle != null)
            {
				JCLineStyle lstyle = null;
				if (entry.parentStyle instanceof JCContourStyle) {
					lstyle = ((JCContourStyle)entry.parentStyle).getLineStyle();
				}
                else if (entry.parentStyle instanceof JCChart3dStyle) {
					lstyle = ((JCChart3dStyle)entry.parentStyle).getLineStyle();
				}
				if (lstyle != null) {
                    lstyle.draw(gc,
                                symX + ((thisItem.symbolDim.width -
                                         lstyle.getWidth())/2),
                                symY + 1,
                                symX + ((thisItem.symbolDim.width -
                                         lstyle.getWidth())/2),
                                symY + thisItem.symbolDim.height - 1);
				}
                style.draw(gc,
                           symX + (thisItem.symbolDim.width/2),
                           symY + (thisItem.symbolDim.height/3) + 1,
                           (2 * (thisItem.symbolDim.width / 3)) + 1);
            }
            // Draw symbol only
            else {
                style.draw(gc,
                           symX + (thisItem.symbolDim.width/2),
                           symY + (thisItem.symbolDim.height/2),
                           (2 * (thisItem.symbolDim.width / 3)) + 1);
			}
		}
	}
}

/**
 * For implementing custom legend item text draws.  For use when the
 * legend item's contents field holds an object other than a <code>String</code> object.
 *
 * @param gc tbe graphics object used to draw
 * @param useFont the font to use
 * @param thisItem the item to draw
 */
public void drawLegendItemText(Graphics gc, Font useFont,
                                 JCLegendItem thisItem)
{
	if (thisItem.contents instanceof JCMultiFieldString) {
		// DAM Can I rely on the global maxFieldWidths????
		// How best to save it otherwise?
        JCMultiFieldString multiString = (JCMultiFieldString)thisItem.contents;
        multiString.draw(gc, useFont,
                         thisItem.pos.x + thisItem.textPos.x,
                         thisItem.pos.y + thisItem.textPos.y,
                         maxFieldWidths, fieldGap);
	}
}

/**
 * Gets the outline color to use when drawing this legend item's symbol. If
 * null, the legend will use its own foreground color.
 * @return
 * @param thisItem which item to get outline color from
 */
public Color getOutlineColor(JCLegendItem thisItem)
{
	return(outlineColor);
}

/**
 * Sets the outline color to use when drawing this legend item's symbol. If
 * null, legend will use its own foreground color. The default is black.
 *
 * @param outlineColor the new outline color
 */
public void setOutlineColor(Color outlineColor)
{
	this.outlineColor = outlineColor;
}

/**
 * Sets the provided <code>Graphics</code> object with the appropriate settings to
 * use when drawing this legend item's symbol. In this case it sets
 * the <code>gc</code> color to the range's foreground color.
 *
 * @param gc the graphics object to set
 * @param thisItem the item to get information from
 */
public void setFillGraphics(Graphics gc, JCLegendItem thisItem)
{
	if (thisItem.itemInfo == null) {
		return;
    }

	if (thisItem.itemInfo instanceof LegendEntry) {
		LegendEntry entry = (LegendEntry) thisItem.itemInfo;
		if (entry.drawStyle instanceof JCSymbolStyle) {
			gc.setColor(((JCSymbolStyle)entry.drawStyle).getColor());
		}
		else if (entry.drawStyle instanceof JCFillStyle) {
			gc.setColor(((JCFillStyle)entry.drawStyle).getColor());
		}
		else if (entry.drawStyle instanceof JCLineStyle) {
			gc.setColor(((JCLineStyle)entry.drawStyle).getColor());
		}
	}
}

}
