/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCChart3dUtil.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;

import com.klg.jclass.util.JCNumberUtil;

import javax.vecmath.Matrix4d;
import javax.vecmath.Point4d;
import java.awt.*;

/**
 * <code>JCChartUtil</code> is a collection of static methods and static
 * variables that are used throughout JClass Chart 3D.
 */
public class JCChart3dUtil extends JCNumberUtil {

/** Value to add to values to combat round off error. */
public static final double roundOffEpsilon = 1.0e-13;

/**
 * Determines if the convex polygon is backfacing.
 *
 * @param xPoints the xPoints array of the polygon
 * @param yPoints the yPoints array of the polygon
 * @return
 */
public static boolean backfacing(int[] xPoints, int[] yPoints)
{
	long val = 0;
	int length = Math.min(xPoints.length, yPoints.length);
	for (int i = 0; i < length; i++) {
		int j = (i < length - 1) ? i + 1 : 0;
		val += (long)xPoints[i] * yPoints[j] - (long)yPoints[i] * xPoints[j];
	}

	return(val < 0);
}

/**
 * Computes whether the plane defined by the three screen points is backfacing.
 *
 * @param p1 the first point
 * @param p2 the second point
 * @param p3 the third point
 * @return
 */
public static boolean gridBackfacing(ScreenPoint p1, ScreenPoint p2,
			                         ScreenPoint p3)
{
    long    dx1, dy1, dx2, dy2;

	dx1 = p2.getXPixel() - p1.getXPixel();
	dy1 = p2.getYPixel() - p1.getYPixel();
	dx2 = p3.getXPixel() - p1.getXPixel();
	dy2 = p3.getYPixel() - p1.getYPixel();

	return( (dy1*dx2 - dx1*dy2) > 0 );
}

/**
 * Transforms the point (<i>x</i>, <i>y</i>, <i>z</i>) based on the current
 * transformation defined in the chart.  Stores the result in the screen point.
 * <br>If the <i>z</i> value is a hole, use the
 * <i>zAlternate</i> value to do the tranformation.
 *
 * @param chart3d the chart which has the transformation
 * @param x the <i>x</i> coordinate of the point
 * @param y the <i>y</i> coordinate of the point
 * @param z the <i>z</i> coordinate of the point
 * @param zAlternate the alternate <i>z</i> value to use if a hole is encountered
 * @param screenPoint the screen point to store the transformed value
 */
public static void getScreenPoint(JCChart3d chart3d, double x, double y,
                                  double z, double zAlternate,
                                  ScreenPoint screenPoint)
{
	if (screenPoint.isHoleValue()) {
		// Need a transformable value
		z = zAlternate;
	}
	Point4d pt = new Point4d(x, y, z, 1.0);

	Transform transform = chart3d.getChart3dArea().getTransform();
	transform.transformPoint(transform.getView(), pt);

	screenPoint.setXPixel((int) Math.round(pt.x));
	screenPoint.setYPixel((int) Math.round(pt.y));
}

/**
 * Finds the distance of the point (<i>x</i>, <i>y</i>, <i>z</i>) to the eye
 * position. <br><B>Note:</b> The eye position is in absolute space and first needs to
 * be tranformed to graph coordinates.
 *
 * @param chart3d the chart for which the distance to eye is needed
 * @param x the <i>x</i> coordinate of the point
 * @param y the <i>y</i> coordinate of the point
 * @param z the <i>z</i> coordinate of the point
 * @return
 */
public static double distanceToEye(JCChart3d chart3d, double x, double y,
                                   double z)
{
	Transform transform = chart3d.getChart3dArea().getTransform();
	Point4d pt = new Point4d(transform.getEyePosition());
	transform.transformPoint(transform.getInverseAbsolute(), pt);

	double dx = pt.x - x;
	double dy = pt.y - y;
	double dz = pt.z - z;

	return(Math.sqrt(dx*dx + dy*dy + dz*dz));
}

/**
 * Utility debugging function that prints a given 4D matrix.
 *
 * @param m the matrix to be printed
 */
public static void printMatrix(Matrix4d m)
{
	System.out.println("[" +  m.m00 + ", " + m.m01 + ", " + m.m02 + ", " + m.m02 + "]");
	System.out.println("[" +  m.m10 + ", " + m.m11 + ", " + m.m12 + ", " + m.m12 + "]");
	System.out.println("[" +  m.m20 + ", " + m.m21 + ", " + m.m22 + ", " + m.m22 + "]");
	System.out.println("[" +  m.m30 + ", " + m.m31 + ", " + m.m32 + ", " + m.m32 + "]");
}

/**
 * Prints the min/max, the anno min/max, and the dataport min/max for a given
 * axis.
 *
 * @param axis the axis for which the extents are to be printed
 */
protected static void printAxisExtents(JCAxis axis)
{
	if (axis == null) {
		return;
	}
    String s = null;
    switch (axis.getAxisId()) {
    case JCAxis.AXIS_X:
	    s = LocaleBundle.string(LocaleBundle.X_AXIS_LABEL) + ":";
        break;
    case JCAxis.AXIS_Y:
	    s = LocaleBundle.string(LocaleBundle.Y_AXIS_LABEL) + ":";
        break;
    case JCAxis.AXIS_Z:
	    s = LocaleBundle.string(LocaleBundle.Z_AXIS_LABEL) + ":";
       break;
    default:
	    
        return;
    }
    System.out.println(s);
    String sMin = LocaleBundle.string(LocaleBundle.MIN_LABEL);
    String sMax = LocaleBundle.string(LocaleBundle.MAX_LABEL);
	System.out.println(sMin + ": " + axis.getMin() + " " + sMax + ": " +
                       axis.getMax());
	System.out.println("Anno " + sMin + ":" + axis.getAnnoMin() + " Anno " +
					   sMax + ": " + axis.getAnnoMax());
	System.out.println("Dataport " + sMin + ": " + axis.getDataportMin() +
	                   " Dataport " + sMax + ": " + axis.getDataportMax());
}

/**
 * Uses the system method arraycopy to copy over larger blocks of values to
 * fill a boolean array efficiently.
 *
 * @param array the array to fill
 * @param value the value with which to fill it
 */
public static void booleanfill(boolean[] array, boolean value)
{
	int len = array.length;
    if (len > 0) {
       array[0] = value;
    }
    for (int i = 1; i < len; i += i) {
        System.arraycopy(array, 0, array, i, ((len - i) < i) ? (len - i) : i);
    }
}

/**
 * Finds the min of the four given integers.
 *
 * @param a the first integer
 * @param b the second integer
 * @param c the third integer
 * @param d the fourth integer
 * @return
 */
public static int min4(int a, int b, int c, int d)
{
    int v1 = Math.min(a, b);
    int v2 = Math.min(c, d);
    return(Math.min(v1, v2));
}


/**
 * Finds the max of the four given integers.
 *
 * @param a the first integer
 * @param b the second integer
 * @param c the third integer
 * @param d the fourth integer
 * @return
 */
public static int max4(int a, int b, int c, int d)
{
    int v1 = Math.max(a, b);
    int v2 = Math.max(c, d);
    return(Math.max(v1, v2));
}

/**
 * Returns the floor of this value.  Math library function floor does most of
 * the work, but we add <code>roundOffEpsilon</code> to it to combat round off
 * near integers.
 *
 * @param value the value for which to find the floor
 * @return
 */
public static double floor(double value)
{
	return(Math.floor(value + roundOffEpsilon));
}

/**
 * Returns the ceiling of this value.  Math library function ceil does most of
 * the work, but we subtract <code>roundOffEpsilon</code> from it to combat
 * round off near integers.
 *
 * @param value the value for which to find the ceil
 * @return
 */
public static double ceil(double value)
{
	return(Math.ceil(value - roundOffEpsilon));
}

/**
 * Returns the number of decimal places after the decimal point the significant
 * digit occurs. Negative numbers mean powers of 10.  Watch out for zero (return
 * Integer.MAX_VALUE/2), because we could use anything we want that would not
 * easily cause an overflow later on.
 *
 * @param value the value for which to find the significant digits
 * @return
 */
public static int nDigits(double value)
{
	if (value == 0.0) {
		return(Integer.MAX_VALUE/2);
	}
	return(- (int) floor(JCNumberUtil.log10(JCNumberUtil.abs(value))));
}

/**
 * The following two routines implement Heckbert's algorithm for finding 'nice'
 * numbers for labelling graphs (see Graphics Gems, "Nice Numbers for Graph
 * Labels" by Paul Heckbert). It has been implemented more or less directly.
 *
 * @param x the number for which a nice number is needed
 * @param round if <code>true</code>, do rounding
 * @return
 */
public static double niceNum(double x, boolean round)
{
	// Given 'x', push it towards a "nice" number.  Here nice means that the
	// significant digit is a 0, 1, 2 or 5.  Use logarithms to find the
	// significant digits, and then convert it.
	if (x == 0) {
		return(0.0);
	}
	else if (x < 0.0) {
		x = -x;
	}

	int    exp = (int) floor(JCNumberUtil.log10(x));
	double pow = JCNumberUtil.pow10(exp);
	double f   = x / pow;
	double nf  = 0.0;
	if (round) {
		if (f < 1.5) {
			nf = 1.0;
		}
		else if (f < 3.0) {
			nf = 2.0;
		}
		else if (f < 7.0) {
			nf = 5.0;
		}
		else {
			nf = 10.0;
		}
	}
	else {
		if (f <= 1.0) {
			nf = 1.0;
		}
		else if (f <= 2.0) {
			nf = 2.0;
		}
		else if (f <= 5.0) {
			nf = 5.0;
		}
		else {
			nf = 10.0;
		}
	}

	return(nf * pow);
}

/**
 * Calculates the <i>x</i>/<i>y</i> extents of the bar (in data space) constuct
 * a flat plane of these values.  This means that the <i>z</i> value is
 * arbitrary and should be set elsewhere.
 *
 * @param data the internal grid data source
 * @param x the <i>x</i> grid coordinate
 * @param y the <i>y</i> grid coordinate
 * @return
 */
public static Point4d[] constructBarPlane(Chart3dGridData data, int x, int y)
{
    Chart3dDataView dataView = data.getDataView();
	int             chartType = dataView.getChartType();
	double          val, gap;
	Point4d         plane[] = new Point4d[4];
	JCBar           bar = dataView.getChart3d().getChart3dArea().getBar();

    double[] xGrid = data.getXGrid();
    double[] yGrid = data.getYGrid();

	for (int i = 0; i < 4; i++) {
		plane[i] = new Point4d();
	}
	if (chartType == JCChart3d.SCATTER ||
		(chartType == JCChart3d.BAR && bar.getXFormat() == JCBar.FIXED))
	{
		val = xGrid[x];
		gap = data.getMinXStep() * bar.getXSpacing()/200.0;
		plane[1].x = val - gap;
		plane[3].x = val + gap;
	}
	else {
		plane[1].x = xGrid[x];
		plane[3].x = xGrid[x + 1];
	}

	if (chartType == JCChart3d.SCATTER ||
		(chartType == JCChart3d.BAR && bar.getYFormat() == JCBar.FIXED))
	{
		val = yGrid[y];
		gap = data.getMinYStep() * bar.getYSpacing()/200.0;
		plane[1].y = val - gap;
		plane[2].y = val + gap;
	}
	else {
		plane[1].y = yGrid[y];
		plane[2].y = yGrid[y + 1];
	}

	plane[0].x = plane[3].x;
	plane[0].y = plane[1].y;
	plane[2].x = plane[1].x;
	plane[3].y = plane[2].y;

	plane[0].w = plane[1].w = plane[2].w = plane[3].w = 1.0;
	return(plane);
}

/**
 * Is there a hole in the facet with bottom-left corner (<i>x</i>, <i>y</i>)?
 * There is a hole if any of the four corners (<i>x</i>,<i>y</i>),
 * (<i>x</i>,<i>y</i>+1), (<i>x</i>+1, <i>y</i>), (<i>x</i>+1,<i>y</i>+1), have
 * a <i>z</i> value which is the hole value.
 *
 * @param data the grid data source
 * @param x the <i>x</i> coordinate of the bottom left corner
 * @param y the <i>y</i> coordinate of the bottom left corner
 * @return
 */
public static boolean holeInFacet(Chart3dGridData data, int x, int y)
{
	if (data == null) {
		return(true);
	}

	ScreenPoint[][] screenPoints = data.getScreenPoints();
	if (screenPoints == null) {
		return(true);
	}

    int             numX         = data.getNumX();
    int             numY         = data.getNumY();
    if (x < 0 || y < 0 || x >= numX || y >= numY) {
        
        return(true);
    }

    boolean hole = screenPoints[x][y].isHoleValue();
    if (x + 1 < numX && !hole) {
        hole = screenPoints[x+1][y].isHoleValue();
    }
    if (y + 1 < numY && !hole) {
        hole = screenPoints[x][y+1].isHoleValue();
        if (x + 1 < numX && !hole) {
            hole = screenPoints[x+1][y+1].isHoleValue();
        }
    }

    return(hole);
}

/**
 * Performs bi-linear interpolation within a grid facet to return the <i>z</i>
 * value.
 *
 * @param data the internal grid data object
 * @param xIndex the index of the <i>x</i> grid value for this facet
 * @param yIndex the index of the <i>y</i> grid value for this facet
 * @param x the <i>x</i> location in data space
 * @param y the <i>y</i> location in data space
 * @return the interpolated <i>z</i> value for the given (<i>x</i>, <i>y</i>)
 */
public static double biLinearInterp(Chart3dGridData data,
                                    int xIndex, int yIndex,
                                    double x, double y)
{
    double x0 = data.getX(xIndex);
    double y0 = data.getY(yIndex);

    double dx = data.getX(xIndex + 1) - x0;
    double dy = data.getY(yIndex + 1) - y0;

    double value00 = data.getZValue(xIndex, yIndex);
    double value01 = data.getZValue(xIndex, yIndex + 1);
    double value10 = data.getZValue(xIndex + 1, yIndex);
    double value11 = data.getZValue(xIndex + 1, yIndex + 1);

    // Interpolate along y0 x axis
    double dz = value10 - value00;
    double z0 = dz * (x - x0) / dx + value00;

    // Interpolate along y1 x axis
    dz = value11 - value01;
    double z1 = dz * (x - x0) / dx + value01;

    // Interpolate in y
    return( (y-y0) * (z1-z0) / dy + z0 );
}

/**
 * Computes the bounding rectangle for two Rectangles.  Needed because
 * Rectangle.union does not work for rectangles whose position has negative
 * values.
 *
 * @param r1 the first rectangle
 * @param r2 the second rectangle
 * @return the bounding rectangle
 */
public static Rectangle computeBoundingRectangle(Rectangle r1, Rectangle r2)
{
	if (r1 == null) {
        return(r2);
	}
    if (r2 == null) {
        return(r1);
    }

    Rectangle r = new Rectangle();
    r.x      = Math.min(r1.x, r2.x);
    r.y      = Math.min(r1.y, r2.y);
    r.width  = Math.max(r1.x + r1.width,  r2.x + r2.width)  - r.x;
    r.height = Math.max(r1.y + r1.height, r2.y + r2.height) - r.y;
	return(r);
}

}
