/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCContour.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.data.JCDefault3dPointDataSource;
import com.klg.jclass.chart3d.resources.LocaleBundle;


import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Point4d;
import java.awt.*;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * Object used to deal with information about contours and zones.  It has
 * methods to compute and to draw the contours/zones.
 */
public class JCContour extends ParentTrackChange implements JCContourMapping {

/*
 * This diagram describes how we view each grid cell.  Note how
 * the even sides are vertical, the odd sides horizontal.  Adjacent
 * cells share a side, so that side 1 of one cell is the same as
 * side 3 of the cell above it.
 *
 *                               1
 *                        p2 ---------- p1
 *           ^              |          |
 *           |              |          |
 *     y increasing      2  |          |  0
 *                          |          |
 *                          |          |
 *                        p3 ---------- p0
 *                               3
 *
 *                         x increasing ->
 *
 * While generating the contours, we must know whether or
 * not a given side of a grid cell has been used or not.
 * We keep track of this in the CellSide object, which
 * stores whether or not a side of the cell is used, for
 * each contour level.
 *
 * These CellSide objects are stored in one array, called
 * 'sides'.  All vertical sides are stored first (left-to-right,
 * bottom-to-top), followed by the horizontal sides (bottom-to-top,
 * left-to-right).  In this manner, it is easy to get to either
 * side of a cell, simply by computing the index of the cell, and
 * either adding or subtracting one to get to the desired side.
 *
 * Cells themselves are indexed in a left-to-right, bottom-to-top
 * order.  This is done to index into the ContourCells array, which
 * contains a structure for each cell in the grid.  The structure
 * keeps track of the contour crossing information on a per-cell
 * basis, to make it easy to fill the contours at any time.
 *
 * Throughout the code you will see references to horizontal and
 * vertical sides, left side, top side, etc.  Always
 * refer to the above diagram, and this discussion, for proper
 * orientation.  Note that horizontal and vertical have nothing
 * to do with the orientation of the surface.
 */

// The zone method either uses contours or zones by the cell's average value

/** Instructs zoning to be done by contours. */
public final static int ZONE_CONTOURS = 1;

/** Instructs zoning to be done by cells using the average of the four corners of the cell. */
public final static int ZONE_CELLS_AVERAGE = 2;

/** Instructs zoning to be done by cells using the value at the bottom left corner of the cell. */
public final static int ZONE_CELLS_CORNER = 3;

// Contours are searched in either a clockwise or counter-clockwise direction.
protected final static int CYCLE_CLOCKWISE         = 0;
protected final static int CYCLE_COUNTER_CLOCKWISE = 1;

// Code that aids in determining which ContourCells should be labelled as holes
protected final static int LEFT_EDGE           = 1;
protected final static int RIGHT_EDGE          = 2;
protected final static int BOTTOM_EDGE         = 4;
protected final static int TOP_EDGE            = 8;
protected final static int BOTTOM_LEFT_CORNER  = 5;
protected final static int TOP_LEFT_CORNER     = 9;
protected final static int TOP_RIGHT_CORNER    = 10;
protected final static int BOTTOM_RIGHT_CORNER = 6;

/** Asks if the contour lines are drawn. */
protected boolean            contoured = false;

/** Asks if the contour zones are drawn. */
protected boolean            zoned = false;

/** Zoning filling method.  Fills according to contours or on a per cell basis. */
protected int                zoneMethod = ZONE_CONTOURS;

/** A list of contour levels. */
protected JCContourLevels    contourLevels = null;

/** List of contour styles which determine how zones and contour lines are drawn. */
protected ArrayList          contourStyles = null;

/** A mapping of contour levels to contour styles. */
protected JCContourMapping   contourMapping = this;

/** The parent <code>dataView</code> object. */
protected Chart3dDataView    dataView = null;

/** A list of default colors (for default contour styles generation). */
protected DefaultColors      defaultColors = new DefaultColors();

// Internal variables (for ease of use)
protected Chart3dGridData       data     = null;
protected Chart3dGridData       zoneData = null;
protected int                   xOffset = -1;
protected int                   yOffset = -1;
protected int                   numX = -1;
protected int                   numY = -1;
protected double                dataHoleValue = Double.MAX_VALUE;
protected double                zoneHoleValue = Double.MAX_VALUE;

// The guts of the contours, the list of contour lines, and contouring
// information by cell.  These are built in calcContours() and then
// used by various routines to render contours & zones.
/** A list of contour lines. */
protected ArrayList             contourLines  = null;
/** Contouring information by cell. */
protected ContourCell[]         contourCells = null;

/** An array of color bins for faster rendering of 2D projections. */
protected ColorBin[] bins = null;

// Polygon chains are constructed from a free-list of allocated Pchain's.
protected PolygonChain[]        fixedChain = null;
protected PolygonChain          freeList = null;
protected int                   pCount = 0;

// Each side, horizontal and vertical, has a CellSide object associated
// with it that stores a bit array.  This bit array stores whether a contour
// line has crossed the side for each contour level.
protected CellSide[]            sides = null;
protected int                   nHorizontalSides = 0;
protected int                   nVerticalSides = 0;

// Index for first line in contour
protected int                   firstIndex = 0;

// Most recent segment in contour line
protected ContourSegment        curSegment = null;

// Point arrays for drawing
protected int[] xPoints = null;
protected int[] yPoints = null;

/**
 * No args constructor.  Sets default contour styles.
 */
public JCContour()
{
    defaultContourStyles();
}

/**
 * Constructor which specifies the parent <code>dataView</code>.  There can be
 * only one <code>dataView</code> parent since the contours are heavily
 * dependent on the data.
 *
 * @param dataView the parent <code>dataView</code>
 */
public JCContour(Chart3dDataView dataView)
{
    this();
    setDataView(dataView);
}

/**
 * Sets the parent <code>dataView</code>.  There can be only one
 * <code>dataView</code> parent since the contours are heavily dependent on
 * the data.
 *
 * @param dataView the new parent dataView
 */
protected void setDataView(Chart3dDataView dataView)
{
	if (this.dataView == dataView) {
		return;
	}
	this.dataView = dataView;

    // Reset parenting on this object as well as child objects that need
    // to tell the dataView when it has updated
	clearParents();
	addParent(dataView);
    if (contourLevels != null) {
        contourLevels.clearParents();
        contourLevels.addParent(dataView);
    }
    for (int i = 0; i < contourStyles.size(); i++) {
        JCContourStyle contourStyle = (JCContourStyle)contourStyles.get(i);
        contourStyle.clearParents();
        contourStyle.addParent(dataView);
    }
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Returns the parent <code>dataView</code>.
 * @return parent <code>dataView</code>
 */
public Chart3dDataView getDataView()
{
	return(dataView);
}

/**
 * Sets flag which determine whether this data view is <code>contoured</code>.
 *
 * @param contoured the new value of the contoured flag
 */
public void setContoured(boolean contoured)
{
	if (this.contoured == contoured) {
		return;
	}
	this.contoured = contoured;
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Sets flag which determines whether this data view is contoured, but doesn't
 * update the <code>dataView</code>.  This is used in drawing projections where
 * the value of <code>contoured</code> needs to be set to a temporary value
 * while drawing the projection.
 *
 * @param contoured the new value of the contoured flag
 */
protected void setContouredNoUpdate(boolean contoured)
{
	this.contoured = contoured;
}

/**
 * Returns whether this data view is <code>contoured</code>.
 *
 * @return the value of the contoured flag
 */
public boolean isContoured()
{
	return(contoured);
}

/**
 * Sets flag which determines whether this data view is <code>zoned</code>.
 *
 * @param zoned the new value of the zoned flag
 */
public void setZoned(boolean zoned)
{
	if (this.zoned == zoned) {
		return;
	}
	this.zoned = zoned;
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Sets flag which determines whether this data view is zoned, but doesn't
 * update the <code>dataView</code>.  This is used in drawing projections where
 * the value of <code>zoned</code> needs to be set to a temporary value while
 * drawing the projection.
 *
 * @param zoned the new value of the zoned flag
 */
protected void setZonedNoUpdate(boolean zoned)
{
	this.zoned = zoned;
}

/**
 * Returns whether this data view is <code>zoned</code>.
 *
 * @return the value of the <code>zoned</code> flag
 */
public boolean isZoned()
{
	return(zoned);
}

/**
 * Sets the <code>zone</code> method - by <code>ZONE_CONTOURS</code> or
 * <code>ZONE_CELLS</code>.
 *
 * @param zoneMethod the new <code>zone</code> method
 */
public void setZoneMethod(int zoneMethod)
{
	if (this.zoneMethod == zoneMethod) {
		return;
	}
	if (zoneMethod >= ZONE_CONTOURS && zoneMethod <= ZONE_CELLS_CORNER) {
	   this.zoneMethod = zoneMethod;
	}
	else {
		throw new IllegalArgumentException(
		    LocaleBundle.string(LocaleBundle.INVALID_CONTOURS_OR_CELLS_ERROR));
	}
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Returns the <code>zone</code> method - by <code>ZONE_CONTOURS</code>,
 * <code>ZONE_CELLS_AVERAGE</code>, or <code>ZONE_CELLS_CORNER</code>.
 *
 * @return the current <code>zone</code> method
 */
public int getZoneMethod()
{
	return(zoneMethod);
}

/**
 * Sets the <code>contourLevels</code> object.
 *
 * @param contourLevels the new <code>contourLevels</code> object
 */
public void setContourLevels(JCContourLevels contourLevels)
{
	if (this.contourLevels == contourLevels) {
		return;
	}
	this.contourLevels = contourLevels;
	if (this.contourLevels == null) {
		// We can't have a null contourLevels reference
		this.contourLevels = new JCContourLevels();
	}
	this.contourLevels.clearParents(); // Can only have one dataView parent
	this.contourLevels.addParent(dataView);
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Returns the <code>contourLevels</code> object.
 *
 * @return the <code>contourLevels</code> object
 */
public JCContourLevels getContourLevels()
{
	return(contourLevels);
}

/**
 * Provides an object that implements the <code>JCContourMapping</code>
 * interface.  This object provides a mapping between contour levels and
 * contour styles.
 *
 * @param contourMapping the new contour mapping
 */
public void setContourMapping(JCContourMapping contourMapping)
{
	if (this.contourMapping == contourMapping) {
		return;
	}
	this.contourMapping = contourMapping;
	if (this.contourMapping == null) {
		this.contourMapping = this;
	}
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Returns the object which currently implements the mapping between contour
 * levels and contour styles (by default this is the
 * <code>JCContour</code> object).
 *
 * @return the current contour mapping
 */
public JCContourMapping getContourMapping()
{
	return(contourMapping);
}

/**
 * Sets the contour styles array.  Passing in null resets the contour styles
 * to their default. <br><B>Note:</b> It is assumed that all entries of the contour
 * styles array are non-null.
 *
 * @param contourStyles the new list of contour styles
 */
public void setContourStyles(ArrayList contourStyles)
{
	// We need contour styles so if the user sets it to null, the contour
	// styles are reset to their default value
	if (contourStyles == null || contourStyles.size() == 0) {
		defaultContourStyles();
	}
	else {
        for (int i = 0; i < contourStyles.size(); i++) {
            JCContourStyle contourStyle = (JCContourStyle)contourStyles.get(i);
            contourStyle.clearParents();
            contourStyle.addParent(dataView);
        }
        this.contourStyles = contourStyles;
	}
    // Note: Even if the user sets the same contourStyles, we should recalc
    // because the contents could have changed.
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Returns the current contour styles.  If no contour styles are currently
 * available, a new set of default contours styles is generated and returned.
 *
 * @return the current set of contour styles
 */
public ArrayList getContourStyles()
{
	if (contourStyles == null) {
		// This should never happen
		
		defaultContourStyles();
	    callParentSetChanged(true, Changeable.ALL);
	}
	return(contourStyles);
}

/**
 * Allows the user to reset the contour styles to their default. <br><B>Note:</B> This
 * is identical to <code>setContourStyles(null);</code>.
 */
public void resetContourStyles()
{
	defaultContourStyles();
	callParentSetChanged(true, Changeable.ALL);
}

/**
 * Retrieves the default colors object which stores the list of default contour
 * style colors.
 * @return default colors object
 */
protected DefaultColors getDefaultColors()
{
	return(defaultColors);
}

/**
 * Returns the list of contour lines (internal use only).
 * @return list of contour lines
 */
public ArrayList getContourLines()
{
	return(contourLines);
}

/**
 * Returns the contour cell array. <br><b>Note:</B> This is a linear array used for a cell
 * grid.  The <code>cellIndex</code>(<i>x</i>, <i>y</i>) method will return the
 * correct position in the array for a given (<i>x</i>, <i>y</i>).  For internal
 * use only.
 * @return the contour cell array
 */
public ContourCell[] getContourCells()
{
	return(contourCells);
}

// Utility routines for contour generation

/**
 * Is side <i>x</i> a vertical side?  It is if it is even.
 *
 * @param x the side in question
 * @return Is the side vertical?
 */
protected boolean verticalSide(int x)
{
	return((x % 2) == 0);
}

/**
 * Is side <i>x</i> a horizontal side?  It is if it is odd.
 *
 * @param x the side in question
 * @return Is the side horizontal?
 */
protected boolean horizontalSide(int x)
{
	return((x % 2) == 1);
}

/**
 * Gets the index of the left side of cell (<i>x</i>, <i>y</i>) in the sides array.
 *
 * @param x the <i>x</i> index of the cell
 * @param y the <i>y</i> index of the cell
 * @return the index of the left side of cell (<i>x</i>, <i>y</i>)
 */
protected int leftSide(int x, int y)
{
	return(y * numX + x);
}

/**
 * Gets the index of the right side of cell (<i>x</i>, <i>y</i>) in the sides array.
 *
 * @param x the <i>x</i> index of the cell
 * @param y the <i>y</i> index of the cell
 * @return the index of the right side of cell (<i>x</i>, <i>y</i>)
 */
protected int rightSide(int x, int y)
{
	return(leftSide(x, y) + 1);
}

/**
 * Gets the index of the bottom side of cell (<i>x</i>, <i>y</i>) in the sides array.
 *
 * @param x the <i>x</i> index of the cell
 * @param y the <i>y</i> index of the cell
 * @return the index of the bottom side of cell (<i>x</i>, <i>y</i>)
 */
protected int bottomSide(int x, int y)
{
	return(nVerticalSides + x * numY + y);
}

/**
 * Gets the index of the top side of cell (<i>x</i>, <i>y</i>) in the sides array.
 *
 * @param x the <i>x</i> index of the cell
 * @param y the <i>y</i> index of the cell
 * @return the index of the top side of cell (<i>x</i>, <i>y</i>)
 */
protected int topSide(int x, int y)
{
	return(bottomSide(x, y) + 1);
}

/**
 * Gets the index of a given side in the sides array for cell (<i>x</i>, <i>y</i>).
 *
 * @param x the <i>x</i> index of the cell
 * @param y the <i>y</i> index of the cell
 * @param side the side in question
 * @return the index of the passed in side of cell (<i>x</i>, <i>y</i>)
 */
protected int getSide(int x, int y, int side)
{
	int index;
	if (verticalSide(side)) {
		index = rightSide(x, y) - side / 2;
	}
	else {
		index = topSide(x, y) - side / 2;
	}
	return(index);
}

/**
 * Returns the index in the <code>ContourCells</code> array for this cell.
 *
 * @param x the <i>x</i> index of the cell
 * @param y the <i>y</i> index of the cell
 * @return the index in the <code>ContourCells</code> array for this cell
 */
protected int cellIndex(int x, int y)
{
	return(x * (numY - 1) + y);
}

// Contour generation routines

/**
 * Sets some internal values (for easier access).
 */
protected void setInternalValues()
{
	if (dataView == null) {
		
		return;
	}

	data     = (Chart3dGridData)dataView.getElevationData();
	zoneData = dataView.getZoneData();
	if (!zoneData.isDataOK()) {
		zoneData = data;
	}
	dataHoleValue = data.getHoleValue();
	zoneHoleValue = zoneData.getHoleValue();

	GridLimits gridLimits = data.getGridLimits();
	xOffset = gridLimits.getXOffset();
	yOffset = gridLimits.getYOffset();
	numX    = gridLimits.getNumX();
	numY    = gridLimits.getNumY();
}

/**
 * Is the side of the cell at (<i>x_index</i>, <i>y_index</i>) available to
 * cross a contour at this level?  Make sure no other contour of this level has
 * crossed; if not, see if a contour does cross it.
 *
 * @param xIndex <i>xIndex</i> of the cell
 * @param yIndex <i>yIndex</i> of the cell
 * @param side the side whose availability is in question
 * @param level the contour level who wants to cross the given side
 * @param right the <i>z</i> value on the "right" side of the contour level value
 * @param left the <i>z</i> value on the "left" side of the contour level value
 * @return Is side available?
 */
protected boolean avail(int xIndex, int yIndex, int side, int level,
		                double right, double left)
{
	int index = getSide(xIndex, yIndex, side);
	if (index == firstIndex) {
		return(true);	// Can always complete loop
	}

	double[] levels = contourLevels.getLevels();
	double c = levels[level];
	return(!sides[index].isUsed(level) && ((right - c) * (left - c) <= 0.0));
}

/**
 * Interpolates in 2D along the line running from (<i>a1</i>, <i>o1</i>) to
 * (<i>a2</i>, <i>o2</i>) and point <i>a</i> = <i>av</i>.
 *
 * @return
 * @param a1 <i>x</i> value for first point
 * @param a2 <i>x</i> value for second point
 * @param av
 * @param o1 <i>y</i> value for first point
 * @param o2 <i>y</i> value for second point
 */
protected double interp(double a1, double a2, double o1, double o2, double av)
{
	if (a1 == a2) {
		return(o1);
	}
	return(o1 + (av - a1) * (o2 - o1)/(a2 - a1));
}

/**
 * A contour line is entering this cell, this side. Records it for filling
 * purposes.
 *
 * @param xIndex <i>xIndex</i> of the cell
 * @param yIndex <i>yIndex</i> of the cell
 * @param level the contour level entering cell
 * @param side the side of the cell being entered
 */
protected void addContourCell(int xIndex, int yIndex, int level, int side)
{
	ContourCell cell = contourCells[cellIndex(xIndex, yIndex)];
	cell.setNewIntersection(new Intersection(level, side, curSegment));
}

/**
 * Adds the second crossing point for the current segment across this cell.
 * Makes sure the line is oriented so that the higher <i>y</i> value or higher
 * <i>x</i> value is in the rightmost point.
 *
 * @param xIndex the <i>xIndex</i> of the cell
 * @param yIndex the <i>yIndex</i> of the cell
 * @param side the side which the contour line is crossing
 */
protected void modifyContourCell(int xIndex, int yIndex, int side)
{
	int     corner;
	boolean flip;

	// Note that the first point (on side1) of the intersection is stored in the
	// segment1 field of the newIntersection field of the contour cell.  The
	// second, and latest, point is stored in curSegment.
	ContourCell cell = contourCells[cellIndex(xIndex, yIndex)];
	Intersection intersection = cell.getNewIntersection();
	if (intersection == null) {
		
	}

	// Find out if the line crosses a corner, and if so which one, or goes side-to-side.
	int side1               = intersection.getSide1();
	ContourSegment segment1 = intersection.getSegment1();
	Point3d segPoint        = segment1.getPoint();
	Point3d curPoint        = curSegment.getPoint();
	if (((side1 - side) % 2) == 0) {  // Is the difference even?
		// Side-to-side; sort by increasing x or y

		corner = 4;
		if (verticalSide(side)) {
			// Crosses left to right
			flip = (curPoint.x < segPoint.x);
		}
		else {
			// Crosses bottom to top
			flip = (curPoint.y < segPoint.y);
		}
	}
	else {
		corner = Math.max(side1, side);
		if (corner == 3) {
			// Special case mod 4
			if (side == 0 || side1 == 0) {
				corner = 0;
			}
		}

		// Orient edge so crosses in direction of side
		flip = (corner != side);
	}

	if (flip) {
		intersection.setSide1(side);
		intersection.setSegment1(curSegment);
		intersection.setSide2(side1);
		intersection.setSegment2(segment1);
	}
	else {
		// Side 1 and Segment 1 already in place
		intersection.setSide2(side);
		intersection.setSegment2(curSegment);
	}

	cell.addIntersection(intersection, corner);
	cell.setNewIntersection(null);
}

/**
 * Reverses an array of length <i>n</i>.
 *
 * @param array the array to reverse
 * @param n The number of elements of the array to reverse.  May not be equal
 *   to array.length.
 */
protected void reverseArray(Object[] array, int n)
{
	if (array.length < n) {
		
		return;
	}

	int half = n/2;
	n--;
	for (int i = 0; i < half; i++) {
		Object obj   = array[i];
		array[i]     = array[n - i];
		array[n - i] = obj;
	}
}

/**
 * Cycles the array's lower most four elements by <i>n</i>.
 *
 * @param array the array to cycle
 * @param n The number of elements to cycle.  May not be equal to array.length.
 */
protected void cycleArray(Object[] array, int n)
{
	if (array.length < 4) {
		
		return;
	}

	for (int i = 0; i < (n % 4); i++) {
		Object obj = array[3];
		array[3]   = array[2];
		array[2]   = array[1];
		array[1]   = array[0];
		array[0]   = obj;
	}
}

/**
 * Adds this point to the current contour line.
 *
 * @param xIndex the <i>x</i> grid index of the point
 * @param yIndex the <i>y</i> grid index of the point
 * @param x the <i>x</i> value of the point to add
 * @param y the <i>y</i> value of the point to add
 */
protected void addPoint(int xIndex, int yIndex, double x, double y)
{
	// Get the current contour line (the last one)
	int nLines = contourLines.size();
	ContourLine cLine = (ContourLine)contourLines.get(nLines - 1);

	// Create a new contour segment and add it to the contour line
	curSegment = new ContourSegment();
	cLine.addContourSegment(curSegment);

	// Create a new point
	double[] levels = contourLevels.getLevels();
	Point3d point = new Point3d();
	point.x = x;
	point.y = y;

	// Compute the z value
	if (zoneData == data) {
		// z value is same as contour level
		point.z = levels[cLine.getLevel()];
	}
	else {
		// 4D graph - interpolate from surface
		xIndex += xOffset;
		yIndex += yOffset;

		double x1 = data.getX(xIndex);
		double y1 = data.getY(yIndex);
		double z1 = data.getZValue(xIndex, yIndex);

		double x2 = x1;
		double y2 = data.getY(yIndex + 1);
		double z2 = data.getZValue(xIndex, yIndex + 1);

		double x3 = data.getX(xIndex + 1);
		double y3 = y2;
		double z3 = data.getZValue(xIndex + 1, yIndex + 1);

		double x4 = x3;
		double y4 = y1;
		double z4 = data.getZValue(xIndex + 1, yIndex);

		if (x == x1) {
			// On left side of cell
			if (z1 == dataHoleValue || z2 == dataHoleValue) {
				point.z = dataHoleValue;
			}
			else {
				point.z = interp(y1, y2, z1, z2, y);
			}
		}
		else if (x == x3) {
			// On right side of cell
			if (z3 == dataHoleValue || z4 == dataHoleValue) {
				point.z = dataHoleValue;
			}
			else {
				point.z = interp(y3, y4, z3, z4, y);
			}
		}
		else if (y == y1) {
			// On bottom of cell
			if (z1 == dataHoleValue || z4 == dataHoleValue) {
				point.z = dataHoleValue;
			}
			else {
				point.z = interp(x1, x4, z1, z4, x);
			}
		}
		else if (y == y2) {
			// On top of cell
			if (z2 == dataHoleValue || z3 == dataHoleValue) {
				point.z = dataHoleValue;
			}
			else {
				point.z = interp(x2, x3, z2, z3, x);
			}
		}
		else {      // Internal value
			if (z1 == dataHoleValue ||
			    z2 == dataHoleValue ||
				z3 == dataHoleValue ||
				z4 == dataHoleValue)
			{
				point.z = dataHoleValue;
			}
			else {
				// First interpolate in y and then in x
				double za  = interp(y1, y2, z1, z2, y);
				double zb  = interp(y3, y4, z3, z4, y);
				double zv1 = interp(x1, x4, za, zb, x);

				// First interpolate in x and then in y
				za = interp(x1, x4, z1, z4, x);
				zb = interp(x2, x3, z2, z3, x);
				double zv2 = interp(y1, y2, za, zb, y);

				// Take the average of the two interpolated values
				point.z = (zv1 + zv2) / 2.0;
			}
		}
	}
	curSegment.setPoint(point);
}

/**
 * Starts a new contour at this level.  The starting point crosses the line
 * joining <i>p1</i> and <i>p2</i>.
 *
 * @param xIndex the <i>x</i> grid index of the start point
 * @param yIndex the <i>y</i> grid index of the start point
 * @param level the contour level to start
 * @param p1 point 1 of the line the start point crosses
 * @param p2 point 2 of the line the start point crosses
 */
protected void startContour(int xIndex, int yIndex, int level,
                            Point3d p1, Point3d p2)
{
	if (contourLines == null) {
		contourLines = new ArrayList();
	}

	// Create a new contour line and add it (at the end)
	ContourLine cLine = new ContourLine(level);
	contourLines.add(cLine);

	// Use linear interpolation to calculate the place between the two points
	// where the value equals the contour value c = levels[level]
	double[] levels = contourLevels.getLevels();
	double c = levels[level];
	double xa =  p2.x * (c - p1.z) - p1.x * (c - p2.z);
	double ya =  p2.y * (c - p1.z) - p1.y * (c - p2.z);
	xa /= (p2.z - p1.z);
	ya /= (p2.z - p1.z);
	addPoint(xIndex, yIndex, xa, ya);
}

/**
 * Generates a complete contour that starts at (<i>xIndex</i>, <i>yIndex</i>).
 * We are coming into the 'side' of the cell, and are contouring at
 * the 'level'. (See diagram at top of file).
 * <p>
 * We reorient everything to make it look like we're coming in side 3.  The
 * order we search the sides is determined by 'cycle' - either clockwise or
 * counter-clockwise.
 *
 * @param cycle the order we search the sides, either equal to <code>CYCLE_CLOCKWISE</code>
 *     or <code>CYCLE_COUNTER_CLOCKWISE</code>
 * @param xIndex the <i>x</i> grid index of the start point
 * @param yIndex the <i>y</i> grid index of the start point
 * @param level the contour level
 * @param side the side of the cell to start contouring
 */
protected void contour(int cycle, int xIndex, int yIndex, int level, int side)
{
	int				i;
	Integer[]		order = new Integer[4];
	double			xa, ya;

	Point3d[] pts = new Point3d[4];
	for (i = 0; i < pts.length; i++) {
		pts[i] = new Point3d();
	}

	boolean done = false;
	boolean starting = true;

	while (!done) {
		if (xIndex < 0           ||
		    yIndex < 0           ||
		    xIndex >= (numX - 1) ||
			yIndex >= (numY - 1))
		{
			// Contouring off the end
			break;
		}

		//  Set up the four corner points.  This is for
		//  side 3, searching counter-clockwise; we might
		//  need to shuffle things around.

		int xi = xOffset + xIndex;
		int yi = yOffset + yIndex;

		pts[3].z = zoneData.getZValue(xi,     yi);
		pts[0].z = zoneData.getZValue(xi + 1, yi);
		pts[1].z = zoneData.getZValue(xi + 1, yi + 1);
		pts[2].z = zoneData.getZValue(xi,     yi + 1);

		if (pts[0].z == zoneHoleValue ||
		    pts[1].z == zoneHoleValue ||
			pts[2].z == zoneHoleValue ||
			pts[3].z == zoneHoleValue)
		{
			// Can't contour through a cell with a hole
			break;
		}

		pts[2].x = pts[3].x = zoneData.getX(xi);
		pts[0].y = pts[3].y = zoneData.getY(yi);
		pts[0].x = pts[1].x = zoneData.getX(xi + 1);
		pts[2].y = pts[1].y = zoneData.getY(yi + 1);

		for (i = 0; i < order.length; i++) {
			order[i] = new Integer(i);
		}

		cycleArray(pts,   3 - side);
		cycleArray(order, 3 - side);
		if (cycle == CYCLE_CLOCKWISE) {
			// Note we reverse the entire array "pts" while only reversing
			// the first three elements of array "order".  This is because
			// we always want to start with side 3 (stored in order[3])
			reverseArray(pts,   4);
			reverseArray(order, 3);
		}
		if (starting) {
			// Begin a contour
			startContour(xIndex, yIndex, level, pts[3], pts[0]);
			starting = false;
		}

		for (i = 0; i < 3; i++) {
			int orderValue = order[i].intValue();
			if (avail(xIndex, yIndex, orderValue, level, pts[i].z, pts[i+1].z)) {
				// Side order[i] can be used

				// Record the fact that contour line came in
				addContourCell(xIndex, yIndex, level, side);
				if (pts[i].z == pts[i+1].z) {
					// Just take the midpoint
					xa = (pts[i+1].x + pts[i].x) / 2.0;
					ya = (pts[i+1].y + pts[i].y) / 2.0;
				}
				else {
					// Interpolate along line
					double[] levels = contourLevels.getLevels();
					double c = levels[level];
					xa =  pts[i+1].x * (c - pts[i].z) -
                          pts[i].x * (c - pts[i+1].z);
					ya =  pts[i+1].y * (c - pts[i].z) -
                          pts[i].y * (c - pts[i+1].z);
					xa /= (pts[i+1].z - pts[i].z);
					ya /= (pts[i+1].z - pts[i].z);
				}
				addPoint(xIndex, yIndex, xa, ya);
				modifyContourCell(xIndex, yIndex, orderValue);

				int segIndex = getSide(xIndex, yIndex, orderValue);
				sides[segIndex].setUsed(level);

				if (segIndex != firstIndex) {
					// Follow to the next cell
					switch(orderValue) {
					case 0:
						xIndex++;
						side = 2;
						break;
					case 1:
						yIndex++;
						side = 3;
						break;
					case 2:
						xIndex--;
						side = 0;
						break;
					case 3:
						yIndex--;
						side = 1;
						break;
					}
				}
				else {
					// Close the contour
					done = true;
				}
				break;
			}
		}
		if (i == 3) {
			// No more sides left
			done = true;
		}
	}
	curSegment = null;
}

/**
 * A zone hole has been detected during the contour search at point
 * (<i>xIndex</i>, <i>yIndex</i>).  Mark all adjoining cells as having a hole.
 *
 * @param xIndex the <i>x</i> grid index of the contour cell
 * @param yIndex the <i>y</i> grid index of the contour cell
 */
protected void setContourCellHole(int xIndex, int yIndex)
{
	int code = 0;
	if (xIndex == 0) {
		code |= LEFT_EDGE;
	}
	else if (xIndex == numX - 1) {
		code |= RIGHT_EDGE;
	}
	if (yIndex == 0) {
		code |= BOTTOM_EDGE;
	}
	else if (yIndex == numY - 1) {
		code |= TOP_EDGE;
	}

	switch(code) {
	case LEFT_EDGE:
		contourCells[cellIndex(xIndex, yIndex    )].setHole(true);
		contourCells[cellIndex(xIndex, yIndex - 1)].setHole(true);
		break;
	case RIGHT_EDGE:
		contourCells[cellIndex(xIndex - 1, yIndex    )].setHole(true);
		contourCells[cellIndex(xIndex - 1, yIndex - 1)].setHole(true);
		break;
	case BOTTOM_EDGE:
		contourCells[cellIndex(xIndex - 1, yIndex)].setHole(true);
		contourCells[cellIndex(xIndex,     yIndex)].setHole(true);
		break;
	case TOP_EDGE:
		contourCells[cellIndex(xIndex - 1, yIndex - 1)].setHole(true);
		contourCells[cellIndex(xIndex,     yIndex - 1)].setHole(true);
		break;
	case BOTTOM_LEFT_CORNER:
		contourCells[cellIndex(xIndex, yIndex)].setHole(true);
		break;
	case TOP_LEFT_CORNER:
		contourCells[cellIndex(xIndex, yIndex - 1)].setHole(true);
		break;
	case BOTTOM_RIGHT_CORNER:
		contourCells[cellIndex(xIndex - 1, yIndex)].setHole(true);
		break;
	case TOP_RIGHT_CORNER:
		contourCells[cellIndex(xIndex - 1, yIndex - 1)].setHole(true);
		break;
	default:		// Middle
		contourCells[cellIndex(xIndex,     yIndex    )].setHole(true);
		contourCells[cellIndex(xIndex,     yIndex - 1)].setHole(true);
		contourCells[cellIndex(xIndex - 1, yIndex    )].setHole(true);
		contourCells[cellIndex(xIndex - 1, yIndex - 1)].setHole(true);
		break;
	}
}

/**
 * Calculates the contours for this data.  This is done by scanning the grid
 * from bottom-to-top, and at each step looking for crossings on the bottom
 * line of the cell, and the left line.
 * <p>
 * Once we have found a crossing, this generates a starting point for the
 * contouring.  The method <code>contour()</code> will either search for a new
 * cell in either a clockwise or counter-clockwise fashion (this is done to
 * ensure that no two contour lines intersect).
 * <p>
 * All crossings are searched either clockwise or counter-clockwise; this
 * maintains symmetry, and ensures the non-intersection rule (but see the
 * Known Bugs section at the top of this file).
 */
public void calcContours()
{
	int     k, cycle, lineIndex;
	double  zLeft, zRight, zLower, zUpper, z1, z2;

	// Destroy all old contours; since the zoning by cell methods don't
	// need the contours, switching from either ZONE_CELL_AVERAGE or
	// ZONE_CELL_CORNER to ZONE_CONTOURS requires contours to be calculated,
	// while an existing ContourCells array exists.
	destroyContours();

	if (dataView == null) {
		
		return;
	}

	setInternalValues();

	if (!data.isDataOK()) {
		return;
	}

	if (dataView.getChartType() != JCChart3d.SURFACE) {
		return;
	}

	JCChart3dArea area   = dataView.getChart3d().getChart3dArea();
	JCProjection floor   = area.getPlotCube().getFloor();
	JCProjection ceiling = area.getPlotCube().getCeiling();

	boolean drawContours = (contoured           ||
						    floor.isContoured() ||
							ceiling.isContoured());
	boolean drawZones    = (zoned           ||
			                floor.isZoned() ||
							ceiling.isZoned());
	boolean doContours    = (drawContours ||
			                 (drawZones && zoneMethod == ZONE_CONTOURS));

	int nLevels = contourLevels.getNumLevelsInternal();
	double[] levels = contourLevels.getLevels();

	if (doContours) {
		nVerticalSides   = (numY - 1) * numX;
		nHorizontalSides = (numX - 1) * numY;

		// Store verticals, then horizontals
		sides = new CellSide[nVerticalSides + nHorizontalSides];
		for (int i = 0; i < (nVerticalSides + nHorizontalSides); i++) {
			sides[i] = new CellSide(nLevels);
		}
	}

	contourCells = new ContourCell[(numX - 1) * (numY - 1)];
	for (int i = 0; i < contourCells.length; i++) {
		contourCells[i] = new ContourCell();
	}

	//  Start the contour search.  Scan the grid from bottom to
	//  top, first searching for crossings along the horizontal
	//  lines, then along the vertical lines.
	//
	//  If the cell is increasing from left-to-right from the
	//  point of view of the line entering the cell, then the search
	//  will be done in a counter-clockwise fashion; otherwise,
	//  search clockwise.  This preserves symmetry and ensures that
	//  no contour crossings are generated.
	//
	//  If on return from the contour search algorithm, the contour is
	//  not closed, search in the opposite direction, using the opposite
	//  cycle.  This handles short contour segments that only span
	//  grid cells between holes.

	int left   = 0;
	int right  = numX - 1;
	int bottom = 0;
	int top    = numY - 1;

	for (int j = bottom; j <= top; j++) {
		zLeft = zoneData.getZValue(xOffset + left, yOffset + j);
		zRight = zLeft;
		for (int i = left; i <= right; i++) {
			if (i != right) {
				zRight = zoneData.getZValue(xOffset + i + 1, yOffset + j);
			}
			if (zLeft != zoneHoleValue) {
				if (i != right && doContours) {
					if (zRight != zoneHoleValue) {
						//  Look for crossings on the bottom side
						if (zLeft <= zRight) {
							z1 = zLeft;
							z2 = zRight;
							cycle = CYCLE_COUNTER_CLOCKWISE;
						}
						else {          // if (zRight < zLeft)
							z1 = zRight;
							z2 = zLeft;
							cycle = CYCLE_CLOCKWISE;
						}
						lineIndex = bottomSide(i, j);

						//  Find the first level which is >= z1.  Then
						//  consider all levels which are < z2; these are
						//  the only levels we need to look at.

						for (k = 0; k < nLevels && levels[k] < z1; k++) {
							// EMPTY
						}
						for (; k < nLevels && levels[k] < z2; k++) {
							if (!sides[lineIndex].isUsed(k)) {
								firstIndex = lineIndex;
								contour(cycle, i, j, k, 3);
							    if (!sides[lineIndex].isUsed(k)) {
									// Not closed - go the other way
									contour(1-cycle, i, j-1, k, 1);
								}
							}
						}
					}
				}

				if (j != top && doContours) {
					// Check for left side crossings
					zLower = zLeft;
					zUpper = zoneData.getZValue(xOffset + i, yOffset + j + 1);
					if (zUpper != zoneHoleValue) {
						if (zLower <= zUpper) {
							z1 = zLower;
							z2 = zUpper;
							cycle = CYCLE_CLOCKWISE;
						}
						else {          // (zLower > zUpper)
							z1 = zUpper;
							z2 = zLower;
							cycle = CYCLE_COUNTER_CLOCKWISE;
						}
						lineIndex = leftSide(i, j);

						// Find the first level which is >= z1.  Then
						//  consider all levels which are < z2; these are
						// the only levels we need to look at.

						for (k = 0; k < nLevels && levels[k] < z1; k++) {
							// EMPTY
						}
						for (; k < nLevels && levels[k] < z2; k++) {
							if (!sides[lineIndex].isUsed(k)) {
								firstIndex = lineIndex;
								contour(cycle, i, j, k, 2);
								if (!sides[lineIndex].isUsed(k)) {
									// Not closed - go the other way
									contour(1-cycle, i-1, j, k, 0);
								}
							}
						}
					}
				}
			}
			else {
				setContourCellHole(i, j);
			}
			zLeft = zRight;
		}
	}

	//  Sort the cells by counter-clockwise intersection order.
	if (doContours) {
		for (int cellIndex = 0; cellIndex < contourCells.length; cellIndex++) {
			contourCells[cellIndex].sortIntersectionLists();
		}
	}

	nVerticalSides   = 0;
	nHorizontalSides = 0;
	sides            = null;
}

/**
 * Checks whether to calculate or destroy the contours based on the contour and
 * zone flags from this object, and the floor and ceiling projections.
 */
public void recalcContours()
{
	if (dataView == null) {
		
		return;
	}

	if (!dataView.isDataDrawn() ||
	    !(dataView.getElevationData() instanceof Chart3dGridData))
	{
		return;
	}

	JCChart3dArea area   = dataView.getChart3d().getChart3dArea();
	JCProjection floor   = area.getPlotCube().getFloor();
	JCProjection ceiling = area.getPlotCube().getCeiling();

	boolean drawContours = (contoured           ||
						    floor.isContoured() ||
							ceiling.isContoured());
	boolean drawZones    = (zoned           ||
			                floor.isZoned() ||
							ceiling.isZoned());

	// Do the contours, new or old, even need to be around?
	if (!drawContours && !drawZones) {
		destroyContours();
		return;
	}

	boolean recalcContours = (drawContours && (contourLines == null ||
							                   contourCells == null));
	if (!recalcContours) {
		if (drawZones) {
			if (zoneMethod == ZONE_CONTOURS) {
				recalcContours = (contourLines == null || contourCells == null);
			}
			else {
				recalcContours = (contourCells == null);
			}
		}
	}
	if (recalcContours) {
		calcContours();
	}
}

/**
 * Destroys the contour information.
 */
public void destroyContours()
{
	contourLines = null;
	contourCells = null;
}

/**
 * Goes through each contour point and transforms it.  This is precomputed so that
 * rendering the transformation is already done and there is no need to transform
 * a point more than once.
 */
protected void precompContours()
{
	if (contourLines == null || dataView == null) {
		return;
	}

	Point4d   transPoint = new Point4d();
	Transform transform = dataView.getChart3d().getChart3dArea().getTransform();

	ListIterator iteratorLines = contourLines.listIterator();
	while (iteratorLines.hasNext()) {
		ContourLine cLine = (ContourLine)iteratorLines.next();
		LinkedList segments = cLine.getSegments();
		ListIterator iteratorSegments = segments.listIterator();
		while(iteratorSegments.hasNext()) {
			ContourSegment segment = (ContourSegment)iteratorSegments.next();

			// Load up data point and transform from data space to screen space
			Point3d segPoint = segment.getPoint();
			transPoint.set(segPoint);
			Matrix4d absolute = transform.getAbsolute();
			if (absolute.m22 == 0.0) {
				transPoint.z = 0.0;
			}
			transform.transformPoint(transform.getView(), transPoint);

			// Extract and save the data for rendering
			segment.setXPixel((int)Math.round(transPoint.x));
			segment.setYPixel((int)Math.round(transPoint.y));
		}
	}
}

// These next routines have to do with rendering of the contours after they've
// been generated.

/**
 * Allocates points only when necessary.
 *
 * @param n the number of points to allocate
 */
protected void allocatePoints(int n)
{
	if (xPoints == null || xPoints.length < n) {
	    xPoints = new int[n];
		yPoints = new int[n];
	}
}

/**
 * Resets the contour styles to their default value.  Note that the fill
 * colors are intialized to null because only some of the contour styles
 * may be used.  The <code>initContourStyles</code> method will assign colors
 * later.
 */
protected void defaultContourStyles()
{
   contourStyles = new ArrayList();
    for (int i = 0; i < defaultColors.getNumColors(); i++) {
        JCContourStyle style = new JCContourStyle(defaultColors.getColor(i));
        style.addParent(dataView);
        contourStyles.add(style);
    }
}

/**
 * Allocates a color bin for each level.
 * @param gc the graphics object to use for drawing
 */
protected void allocColorBins(Graphics gc)
{
	int nbins = contourLevels.getNumLevelsInternal() + 1;
	bins = new ColorBin[nbins];
	for (int i = 0; i < nbins; i++) {
		bins[i] = new ColorBin(this, gc, i);
	}
}

/**
 * Flushes the color bins.  The objects in each bin are drawn.
 */
protected void flushColorBins()
{
	if (bins == null) {
		return;
	}
	for (int i = 0; i < bins.length; i++) {
		bins[i].flushBin();
	}
	bins = null;
}

/**
 * Returns the contour style index that corresponds to this level.  This mapping
 * is based on an even distribution of contour styles through the number of
 * levels.
 *
 * @param level the level for which the contour style index is desired
 * @return the index of the contour style for this level
 */
public int contourIndex(int level)
{
	int numLevels = contourLevels.getNumLevelsInternal();
	int index = 0;
	if (numLevels > 0) {
		double value = (double)(level * (contourStyles.size() - 1));
		value /= (double) numLevels;
		index = (int)Math.round(value);
	}
	return(index);
}

/**
 * Returns the contour style used for a given contour level.  This method uses
 * the current <code>contourMapping</code>'s <code>contourIndex</code> method to get the index of the
 * contour style in the <code>contourStyles</code> array.
 *
 * @param level the level for which the contour style index is desired
 * @return the contour style for the given index
 */
public JCContourStyle getContourStyleFromLevel(int level)
{
    if (contourStyles == null || contourStyles.size() == 0) {
        return(null);
    }
	int            index        = contourMapping.contourIndex(level);
	JCContourStyle contourStyle = (JCContourStyle)contourStyles.get(index);
	if (contourStyle == null) {
		return(null);
	}
	contourStyle.initialize(this, index);
	return(contourStyle);
}

/**
 * Sets the line style for this level.  If the width of the line is 0, or the
 * line pattern is none, then return <code>false</code> (nothing should be drawn);
 * otherwise, return <code>true</code>.
 *
 * @param gc the graphics object to use for drawing
 * @param level the level whose line style info you want to set on the gc
 * @return whether the gc was updated with the <code>lineStyle</code> info
 */
protected boolean setLineStyle(Graphics gc, int level)
{
	JCContourStyle contourStyle = getContourStyleFromLevel(level);
	if (contourStyle == null) {
		return(false);
	}
    JCLineStyle lineStyle = contourStyle.getLineStyle();
    return(lineStyle.updateGraphics(gc));
}

/**
 * Sets the fill style for this level.
 *
 * @param gc the graphics object to use for drawing
 * @param level the level whose fill style info you want to set on the gc
 */
protected void setFillStyle(Graphics gc, int level)
{
	JCContourStyle contourStyle = getContourStyleFromLevel(level);
	if (contourStyle == null) {
		return;
	}
	JCFillStyle fillStyle = contourStyle.getFillStyle();
	fillStyle.updateGraphics(gc);
}

/**
 * Sets the line color for this level based on the fill (zone) style color
 * (rather than the line style color).
 *
 * @param gc the graphics object to use for drawing
 * @param level the level whose fill color you want to set on the gc
 */
protected void setZLineColor(Graphics gc, int level)
{
	JCContourStyle contourStyle = getContourStyleFromLevel(level);
	if (contourStyle == null) {
		return;
	}
	gc.setColor(contourStyle.getFillStyle().getColor());
}

/**
 * Fills a zoned rectangle, using either the value at the given grid point or
 * the average of the four corner values of the cell.  The average is used
 * when <code>zoneMethod</code> is <code>ZONE_CELL_AVERAGE</code>.  This fills
 * one grid cell, at (<i>xIndex</i>, <i>yIndex</i>).
 *
 * @param xIndex the <i>x</i> grid index of the cell
 * @param yIndex the <i>y</i> grid index of the cell
 * @param useAverage use the average of the four corners of the cell to zone,
 *   or just the value at the bottom left corner
 */
protected void fillRectangle(int xIndex, int yIndex, boolean useAverage)
{
	int xi    = xIndex + xOffset;
	int yi    = yIndex + yOffset;

	double z0 = 0.0;
	if (useAverage) {
	    z0 = (zoneData.getZValue(xi, yi)     +
	          zoneData.getZValue(xi + 1, yi) +
	          zoneData.getZValue(xi, yi + 1) +
			  zoneData.getZValue(xi + 1, yi + 1)) / 4.0;
	}
	else {
		z0 = zoneData.getZValue(xi, yi);
	}

	ScreenPoint[][] screenPoints = data.getScreenPoints();
	int x      = screenPoints[xi][yi].getXPixel();
	int y      = screenPoints[xi][yi+1].getYPixel();
	int width  = screenPoints[xi+1][yi].getXPixel() - x;
	int height = screenPoints[xi][yi].getYPixel() - y;
	Rectangle r = new Rectangle(x, y, width, height);

	int level = contourLevels.getLevelFromValue(z0);
	bins[level].addObjectToBin(r);
}

/**
 * Makes a new chain, containing this single point. The <i>zValue</i> comes
 * from elevation data, <i>cValue</i> from zone data.
 *
 * @param xPixel the <i>x</i> pixel value for this point
 * @param yPixel the <i>y</i> pixel value for this point
 * @param zValue elevation data <i>z</i> value
 * @param cValue zone (or contour) data value
 * @return
 */
protected PolygonChain makePChain(int xPixel, int yPixel,
					              double zValue, double cValue)
{
	PolygonChain    pChain;

	// If its the first call, allocate fixed array of PolygonChain objects.
	if (fixedChain == null) {
		fixedChain = new PolygonChain[PolygonChain.MAX_CHAIN_SIZE];
		for (int i = 0; i < fixedChain.length; i++) {
			pChain = new PolygonChain();
			fixedChain[i] = pChain;
			if (i > 0) {
			   fixedChain[i-1].setNext(pChain);
			}
		}
		freeList = fixedChain[0];
	}

	// Pull next point off free-list
	pChain = freeList;
	freeList = freeList.getNext();
	pChain.setValues(xPixel, yPixel, zValue, cValue);
	pChain.setNext(null);
	pCount++;
	return(pChain);
}

/**
 * Copies an entire chain to a new one, and returns the new one.
 *
 * @param pChain the chain to be copied
 * @return
 */
protected PolygonChain copyChain(PolygonChain pChain)
{
	PolygonChain newChain = null;
	if (pChain != null) {
		newChain = makePChain(pChain.getXPixel(), pChain.getYPixel(),
		                      pChain.getZValue(), pChain.getCValue());
		PolygonChain current = newChain;
		PolygonChain next    = pChain.getNext();
		while(next != null) {
			current.setNext(makePChain(next.getXPixel(), next.getYPixel(),
		                               next.getZValue(), next.getCValue()));
			current = current.getNext();
			next = next.getNext();
		}
	}
	return(newChain);
}

/**
 * Appends chain <i>pchain2</i> to the end of <i>pchain1</i>. Returns a
 * reference to the start of the combined chain.
 *
 * @param pChain1 the initial chain
 * @param pChain2 the chain to be appended
 * @return
 */
protected PolygonChain appendChain(PolygonChain pChain1, PolygonChain pChain2)
{
	if (pChain1 == null) {
		return(pChain2);
	}
	if (pChain2 == null) {
		return(pChain1);
	}

	PolygonChain start = pChain1;
	while (pChain1.getNext() != null) {
		pChain1 = pChain1.getNext();
	}
	pChain1.setNext(pChain2);

	return(start);
}

/**
 * Reverses the order of points in this chain.
 *
 * @param pChain the chain to be reversed
 * @return
 */
protected PolygonChain reverseChain(PolygonChain pChain)
{
	PolygonChain previous = null;
	PolygonChain current  = pChain;
	while (current != null) {
		PolygonChain next = current.getNext();
		current.setNext(previous);
		previous = current;
		current = next;
	}

	return(previous);
}

/**
 * Gets rid of this chain.  Prepends it to the free-list.
 *
 * @param pChain the chain to be destroyed
 */
protected void destroyChain(PolygonChain pChain)
{
	if (pChain == null) {
		
	}

	PolygonChain first = pChain;
	PolygonChain last  = null;
	while (pChain != null) {
		last = pChain;
		pChain = last.getNext();
		pCount--;
	}

	last.setNext(freeList);
	freeList = first;
}

/**
 * Given a list of points and a contour level, process the polygon that the
 * list of points defines.  If drawing a 3D chart, draw the polygon.  If doing
 * a 2D projection, add the polygon to the bin for this level so it can be
 * drawn later.
 *
 * @param gc the graphics object to use for drawing
 * @param count the number of points in the list
 * @param level the contour level for this polygon
 */
public void processPolygon(Graphics gc, int count, int level)
{
	Transform transform = dataView.getChart3d().getChart3dArea().getTransform();
    Matrix4d absolute   = transform.getAbsolute();
	if (absolute.m22 != 0.0) {  // 3D
		setFillStyle(gc, level);
		gc.fillPolygon(xPoints, yPoints, count);
	}
	else {                      // 2D Projection
		bins[level].addObjectToBin(new Polygon(xPoints, yPoints, count));
	}
}

/**
 * Fills in a <code>PolygonChain</code>.  This chain represents a polygon, which
 * should be filled in one solid color.  We determine that value by taking the
 * average value of all the points, unless the <code>zoneMethod</code> is <code>ZONE_CELLS_CORNER</code>,
 * in which case we take the value of the bottom left corner of the cell.
 *
 * @param gc the graphics object to use for drawing
 * @param pChain the polygon chain to be filled
 */
protected void fillChain(Graphics gc, PolygonChain pChain)
{
	allocatePoints(9);  // Worst case is 9 points

	int          count   = 0;
	double       zSum    = 0;
	PolygonChain current = pChain;
	while (current != null) {
		// Extract and save the data for rendering
		xPoints[count] = current.getXPixel();
		yPoints[count] = current.getYPixel();
        if (zoneMethod == ZONE_CELLS_CORNER) {
            // Bottom left corner value is in 2nd element of chain
            if (count == 1) {
                zSum = current.getCValue();
            }
        }
        else {
		    zSum += current.getCValue();
        }
		count++;
		current = current.getNext();
	}

	if (count > 2) {
        double zValue = (zoneMethod == ZONE_CELLS_CORNER) ? zSum : zSum/((double)count);
		int level = contourLevels.getLevelFromValue(zValue);
		processPolygon(gc, count, level);
	}
}

/**
 * Fills the zoned cell at each of the four corners.
 *
 * @param gc the graphics object to use for drawing
 * @param chainArray the array of polygon chains to fill (one per corner)
 * @param index the index of the <code>contourCell</code> whose corners are to be filled
 */
protected void fillCellCorners(Graphics gc, PolygonChain[] chainArray, int index)
{
	double[]    levels = contourLevels.getLevels();
	ContourCell cell   = contourCells[index];

	for (int i = 0; i < 4; i++) {
		ArrayList intersectionList = cell.getIntersectionList(i);
		if (intersectionList == null) {
			continue;
		}
		for (int j = 0; j < intersectionList.size(); j++) {
			Intersection intersection = (Intersection)intersectionList.get(j);
			ContourSegment segment = intersection.getSegment1();
			PolygonChain pChain = makePChain(segment.getXPixel(),
										     segment.getYPixel(),
											 segment.getPoint().z,
											 levels[intersection.getLevel()]);
			segment = intersection.getSegment2();
			pChain.setNext(makePChain(segment.getXPixel(),
								      segment.getYPixel(),
									  segment.getPoint().z,
									  levels[intersection.getLevel()]));
			PolygonChain fill = appendChain(reverseChain(chainArray[i]),
                                            copyChain(pChain));
			fillChain(gc, fill);
			destroyChain(fill);
			chainArray[i] = pChain;
		}
	}
}

/**
 * Fills the portions of the zoned cell relating to the left-to-right or
 * bottom-to-top crossings.  When this is called, the corner polygons have
 * already been processed.  What is left for each corner is a line segment or
 * a point (if no crossings for that corner) that form a polygon with the
 * line crossing.
 *
 * @param gc the graphics object to use for drawing
 * @param chainArray the array of polygons chains to fill
 * @param corner1 the first corner
 * @param corner2 the second corner
 * @param index the index of the <code>contourCell</code> whose cell sides are to be filled
 */
protected void fillCellSide(Graphics gc, PolygonChain[] chainArray,
			                int corner1, int corner2, int index)
{
	double[]    levels = contourLevels.getLevels();
	ContourCell cell   = contourCells[index];

	ArrayList intersectionList = cell.getIntersectionList(ContourCell.SIDE);
	if (intersectionList == null) {
		return;
	}
	for (int j = 0; j < intersectionList.size(); j++) {
	    Intersection intersection = (Intersection)intersectionList.get(j);
		ContourSegment segment = intersection.getSegment1();
		PolygonChain pChain1 = makePChain(segment.getXPixel(),
									      segment.getYPixel(),
										  segment.getPoint().z,
										  levels[intersection.getLevel()]);
		segment = intersection.getSegment2();
		PolygonChain pChain2 = makePChain(segment.getXPixel(),
									      segment.getYPixel(),
										  segment.getPoint().z,
										  levels[intersection.getLevel()]);

		// Do shortest first, and construct in two parts to minimize work
		// done by appendChain

		PolygonChain  fill = appendChain(copyChain(pChain2),
                                         copyChain(pChain1));
		PolygonChain fill2 = appendChain(chainArray[corner1],
                                         chainArray[corner2]);
		fill = appendChain(fill, fill2);

		fillChain(gc, fill);
		destroyChain(fill);

		chainArray[corner1] = pChain1;
		chainArray[corner2] = pChain2;
	}
}

/**
 * Fills in the cell at (<i>xIndex</i>, <i>yIndex</i>) by breaking it into
 * smaller polygons that must be filled individually.
 * <P>
 * All the contour crossings have been computed for this cell, and are stored
 * in the <code>ContourCell</code> intersection lists.  There are 5 lists.
 * The first four represent those lines which cross the 4 corners and are
 * stored in the corners array.  The last represents those lines which cross
 * side-to-side or top-to-bottom (note that no cell can contain contour lines
 * which both cross side-to-side and top-to-bottom).
 * <P>
 * All those segments which cross the corners have been sorted (see
 * <code>sortIntersectionLists()</code> method in <code>ContourCells</code> for
 * the particulars).  We construct a polygon chain for each corner that consists
 * only of the corner value, and then possible extend the polygons after that,
 * by growing outward from the corner.  After that, do the side-to-side
 * crossings.
 * <P>
 * If this is a zone hole, return <code>false</code> and the caller will fill the cell
 * normally (4D).
 *
 * @param gc the graphics object to use for drawing
 * @param xIndex the <i>x</i> index of the cell to be filled
 * @param yIndex the <i>y</i> index of the cell to be filled
 * @return
 */
protected boolean fillCell(Graphics gc, int xIndex, int yIndex)
{
	int index = cellIndex(xIndex, yIndex);
	ContourCell cell = contourCells[index];

	if (cell.isHole()) {
		// Zone hole
		return(false);
	}

	int x0 = xOffset + xIndex;
	int y0 = yOffset + yIndex;
	int x1 = x0 + 1;
	int y1 = y0 + 1;

	//  Make a pchain for each corner
	ScreenPoint[][] screenPoints = data.getScreenPoints();
	PolygonChain[] chainArray = new PolygonChain[4];
	chainArray[0] = makePChain(screenPoints[x1][y0].getXPixel(),
							   screenPoints[x1][y0].getYPixel(),
		                       data.getZValue(x1, y0),
		                       zoneData.getZValue(x1, y0));
	chainArray[1] = makePChain(screenPoints[x1][y1].getXPixel(),
							   screenPoints[x1][y1].getYPixel(),
		                       data.getZValue(x1, y1),
		                       zoneData.getZValue(x1, y1));
	chainArray[2] = makePChain(screenPoints[x0][y1].getXPixel(),
							   screenPoints[x0][y1].getYPixel(),
		                       data.getZValue(x0, y1),
		                       zoneData.getZValue(x0, y1));
	chainArray[3] = makePChain(screenPoints[x0][y0].getXPixel(),
							   screenPoints[x0][y0].getYPixel(),
		                       data.getZValue(x0, y0),
		                       zoneData.getZValue(x0, y0));

	if (zoneMethod == ZONE_CONTOURS) {

		// Fill all the corners first, then worry about the middle.  This may
		// modify chain chainArray[i], which represents the polygon at corner i.

		fillCellCorners(gc, chainArray, index);

		// Now do the side-to-side, or top-to-bottom crossings (note that we
		// can't have both, otherwise contour lines would intersect!).  Use the
		// corner chains chainArray[i] to construct what is left on the inside;
		// whatever is left will be dealt with at the end.

		ArrayList intersectionList = cell.getIntersectionList(ContourCell.SIDE);
		if (intersectionList != null && intersectionList.size() > 0) {
			Intersection intersection = (Intersection)intersectionList.get(0);
			if (verticalSide(intersection.getSide1())) {
				// Left to right crossings, sorted bottom to top
				fillCellSide(gc, chainArray, 3, 0, index);
			}
			else {
				/* bottom to top crossings, sorted right to left */
				fillCellSide(gc, chainArray, 0, 1, index);
			}
		}
	}

	// This is the piece left over, after all the corners and all the middles
	// have been done.  Construct in two parts to minimize work done by
	// appendChain.

	PolygonChain fill  = appendChain(chainArray[2], chainArray[3]);
	PolygonChain fill2 = appendChain(chainArray[0], chainArray[1]);
	fill = appendChain(fill, fill2);

	fillChain(gc, fill);
	destroyChain(fill);

	if (pCount != 0) {
		
	}
	return(true);
}

/**
 * Draws all contours for the specified cell.  This should only be used in 3D;
 * otherwise, drawing the entire contour chain at once would be much quicker
 * (<code>drawContours()</code>).
 *
 * @param gc the graphics object to use for drawing
 * @param xIndex the <i>x</i> index of the cell to be drawn
 * @param yIndex the <i>y</i> index of the cell to be drawn
 */
protected void drawCellContours(Graphics gc, int xIndex, int yIndex)
{
	int index = cellIndex(xIndex, yIndex);
	ContourCell cell = contourCells[index];

	for (int i = 0; i < 5; i++) {
		ArrayList intersectionList = cell.getIntersectionList(i);
		if (intersectionList == null) {
			continue;
		}
		for (int j = 0; j < intersectionList.size(); j++) {
			Intersection intersection = (Intersection)intersectionList.get(j);
			setLineStyle(gc, intersection.getLevel());
			gc.drawLine(intersection.getSegment1().getXPixel(),
			            intersection.getSegment1().getYPixel(),
						intersection.getSegment2().getXPixel(),
						intersection.getSegment2().getYPixel());
		}
	}
}

/**
 * Draws the zoned mesh for this intersection list for a given mesh line. The
 * point <i>right</i> and value <i>zr</i> are passed in and updated as the intersection
 * list is traversed.  Go forward or backward through the list based on the
 * value of the increasing flag.
 *
 * @param gc the graphics object to use for drawing
 * @param intersectionList the intersection list to draw line segments from
 * @param right the starting point - updated as the list is traversed
 * @param zr the starting value - updated as the list is traversed
 * @param doSeg1 draw the points in segment1 or segment2
 * @param increasing draw forwards or backwards in list
 * @return the final <i>zr</i> value
 */
protected double zoneMeshIntersection(Graphics gc, ArrayList intersectionList,
				                      Point right, double zr, boolean doSeg1,
									  boolean increasing)
{
	double          zl;
	Intersection    intersection;
	ContourSegment  segment;
	Point           left = new Point();
    double[]        levels = contourLevels.getLevels();

    if (intersectionList != null) {
		int start     = increasing ? 0 : intersectionList.size() - 1;
		int stop      = increasing ? intersectionList.size() : 0;
		int increment = increasing ? 1 : -1;
		for (int i = start; (increasing ? i < stop : i >= stop); i+= increment) {
            intersection = (Intersection)intersectionList.get(i);
            segment      = doSeg1 ? intersection.getSegment1()
								  : intersection.getSegment2();
            left.x       = segment.getXPixel();
            left.y       = segment.getYPixel();
			zl           = levels[intersection.getLevel()];
			setZLineColor(gc, contourLevels.getLevelFromValue((zl + zr) / 2.0));
            gc.drawLine(left.x, left.y, right.x, right.y);
            right.setLocation(left);
			zr = zl;
		}
	}
	return(zr);
}

/**
 * Draws a zoned mesh for this cell.  If its a zone hole, return <code>false</code> and the
 * caller will draw a normal mesh in its place (4D).
 * <p>
 * Basic algorithm is to follow along each side of the mesh, from corner to
 * corner, taking each contour intersection as they occur.
 *
 * @param gc the graphics object to use for drawing
 * @param xIndex the <i>x</i> index of the cell to draw
 * @param yIndex the <i>y</i> index of the cell to draw
 * @return
 */
protected boolean zoneMeshCell(Graphics gc, int xIndex, int yIndex)
{
	double			zl, zr;
    ScreenPoint     sp;
    ArrayList       intersectionList;
    Intersection    intersection;
    JCChart3dArea   area = data.getDataView().getChart3d().getChart3dArea();
    JCSurface       surface = area.getSurface();

	int index = cellIndex(xIndex, yIndex);
    ContourCell cell = contourCells[index];
    if (cell.isHole()) {
        return(false);
    }

	int xi = xOffset + xIndex;
	int yi = yOffset + yIndex;

    Point            left         = new Point();
    Point            right        = new Point();
    ScreenPoint[][]  screenPoints = data.getScreenPoints();
    boolean[]        xMesh        = surface.getXMesh();
    boolean[]        yMesh        = surface.getYMesh();

	if (surface.isXMeshShowing()) {

        // Bottom of mesh. Start at corner 0, move to corner 3.

		if (xMesh[yi]) {
			// Corner 0 crossings
            sp = screenPoints[xi + 1][yi];
            right.setLocation(sp.getXPixel(), sp.getYPixel());
            zr = zoneData.getZValue(xi + 1, yi);
			if (zoneMethod == ZONE_CONTOURS) {
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(0),
				                          right, zr, true, true);

				// Bottom-to-top crossings, sorted right-to-left
				intersectionList = cell.getIntersectionList(ContourCell.SIDE);
                intersection = ((intersectionList == null)
                                ? null
                                : (Intersection)intersectionList.get(0));
                if (intersection != null &&
				    horizontalSide(intersection.getSide1()))
				{
					zr = zoneMeshIntersection(gc, intersectionList,
				                              right, zr, true, true);
                }

				// Corner 3 crossings
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(3),
				                          right, zr, false, false);
			}

			// Finally do corner 3
            sp = screenPoints[xi][yi];
            left.setLocation(sp.getXPixel(), sp.getYPixel());
            zl = zoneData.getZValue(xi, yi);
            setZLineColor(gc, contourLevels.getLevelFromValue((zl + zr) / 2.0));
            gc.drawLine(left.x, left.y, right.x, right.y);
		}

        //  Top of mesh.  Start at corner 1, move towards 2

		// Corner 1 crossings
		if (xMesh[yi + 1]) {
            sp = screenPoints[xi + 1][yi + 1];
            right.setLocation(sp.getXPixel(), sp.getYPixel());
            zr = zoneData.getZValue(xi + 1, yi + 1);
			if (zoneMethod == ZONE_CONTOURS) {
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(1),
				                          right, zr, false, true);

				// Bottom-to-top crossings, sorted right-to-left
				intersectionList = cell.getIntersectionList(ContourCell.SIDE);
                intersection = ((intersectionList == null)
                                ? null
                                : (Intersection)intersectionList.get(0));
                if (intersection != null &&
				    horizontalSide(intersection.getSide1()))
				{
					zr = zoneMeshIntersection(gc, intersectionList,
				                              right, zr, false, true);
				}

				// Corner 2 crossings
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(2),
				                          right, zr, true, false);
			}

			// Finally do corner 2
            sp = screenPoints[xi][yi + 1];
            left.setLocation(sp.getXPixel(), sp.getYPixel());
            zl = zoneData.getZValue(xi, yi + 1);
            setZLineColor(gc, contourLevels.getLevelFromValue((zl + zr) / 2.0));
            gc.drawLine(left.x, left.y, right.x, right.y);
		}
	}

	if (surface.isYMeshShowing()) {

        //  Right of mesh. Start at corner 0, move to corner 1.

		if (yMesh[xi + 1]) {
			// Corner 0 crossings
            sp = screenPoints[xi + 1][yi];
            right.setLocation(sp.getXPixel(), sp.getYPixel());
            zr = zoneData.getZValue(xi + 1, yi);
			if (zoneMethod == ZONE_CONTOURS) {
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(0),
				                          right, zr, false, true);

				// Left-to-right crossings, sorted bottom-to-top
				intersectionList = cell.getIntersectionList(ContourCell.SIDE);
                intersection = ((intersectionList == null)
                                ? null
                                : (Intersection)intersectionList.get(0));
                if (intersection != null &&
				    verticalSide(intersection.getSide1()))
				{
			        zr = zoneMeshIntersection(gc, intersectionList,
				                              right, zr, false, true);
				}

				// Corner 1 crossings
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(1),
				                          right, zr, true, false);
			}

			// Finally do corner 1
            sp = screenPoints[xi + 1][yi + 1];
            left.setLocation(sp.getXPixel(), sp.getYPixel());
            zl = zoneData.getZValue(xi + 1, yi + 1);
            setZLineColor(gc, contourLevels.getLevelFromValue((zl + zr) / 2.0));
            gc.drawLine(left.x, left.y, right.x, right.y);
		}

        // Left of mesh.  Start at corner 3, move towards 2

		if (yMesh[xi]) {
			// Corner 3 crossings
            sp = screenPoints[xi][yi];
            right.setLocation(sp.getXPixel(), sp.getYPixel());
            zr = zoneData.getZValue(xi, yi);
			if (zoneMethod == ZONE_CONTOURS) {
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(3),
				                          right, zr, true, true);

				// Left-to-right crossings, sorted bottom-to-top crossings
				intersectionList = cell.getIntersectionList(ContourCell.SIDE);
                intersection = ((intersectionList == null)
                                ? null
                                : (Intersection)intersectionList.get(0));
                if (intersection != null &&
				    verticalSide(intersection.getSide1()))
				{
					zr = zoneMeshIntersection(gc, intersectionList,
				                              right, zr, true, true);
				}

				// Corner 2 crossings
			    zr = zoneMeshIntersection(gc, cell.getIntersectionList(2),
				                          right, zr, false, false);
			}

			// Finally do corner 2
            sp = screenPoints[xi][yi + 1];
            left.setLocation(sp.getXPixel(), sp.getYPixel());
            zl = zoneData.getZValue(xi, yi + 1);
            setZLineColor(gc, contourLevels.getLevelFromValue((zl + zr) / 2.0));
            gc.drawLine(left.x, left.y, right.x, right.y);
		}
	}

	return(true);
}

/**
 * Draws flat zones in a 2D projection.  Everything should have been
 * precomputed so it only needs to be drawn.
 *
 * @param gc the graphics object to use for drawing
 */
protected void draw2dZones(Graphics gc)
{
	if (!zoned || contourCells == null || dataView == null) {
		return;
	}

	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	JCChart3dArea       area         = dataView.getChart3d().getChart3dArea();
	JCViewport          viewport     = area.getViewport();

	boolean[] xSkip = new boolean[numX - 1];
	for (int i = 0; i < numX - 1; i++) {
		xSkip[i] = (screenPoints[xOffset + i + 1][yOffset].getXPixel() ==
					screenPoints[xOffset + i][yOffset].getXPixel());
	}
	boolean[] ySkip = new boolean[numY - 1];
	for (int j = 0; j < numY - 1; j++) {
		ySkip[j] = (screenPoints[xOffset][yOffset + j + 1].getYPixel() ==
					screenPoints[xOffset][yOffset + j].getYPixel());
	}
	allocColorBins(gc);
	int index = 0;
	for (int i = 0; i < numX - 1; i++) {
	    if (xSkip[i]) {
		    index += numY - 1;
			continue;
		}
		for (int j = 0; j < numY - 1; j++, index++) {
			if (ySkip[j]) {
				continue;
			}
			if (!viewport.isNormalized()) {
			    //  Don't process any facets outside clip rectangle
				int xi = xOffset + i;
				int yj = yOffset + j;
				int x       = screenPoints[xi][yj].getXPixel();
				int y       = screenPoints[xi][yj + 1].getYPixel();
				int width   = screenPoints[xi + 1][yj].getXPixel() - x;
				int height  = screenPoints[xi][yj].getYPixel() - y;
				Rectangle rect = new Rectangle(x, y, width, height);
                if (!rect.intersects(area.getBounds())) {
                    continue;
                }
			}
			if (zoneMethod == ZONE_CONTOURS) {
				ContourCell cell = contourCells[index];
				if (cell.getNIntersections() == 0) {
					// No intersections, all four cell corners are the same zone
					if (!cell.isHole()) {
						fillRectangle(i, j, false);
					}
				}
				else {
					fillCell(gc, i, j);
				}
			}
			else { // zoneMethod = ZONE_CELLS_AVERAGE or ZONE_CELLS_CORNER
				fillRectangle(i, j, zoneMethod == ZONE_CELLS_AVERAGE);
			}
		}
	}
	flushColorBins();
	bins = null;
}

/**
 * Draws flat zones in a 3D projection ( projecting zones onto either
 * the top or bottom of the cube).  Everything should have been precomputed
 * so it only needs to be drawn.
 *
 * @param gc the graphics object to use for drawing
 */
protected void draw3dZones(Graphics gc)
{
	if (!zoned || contourCells == null || dataView == null) {
		return;
	}

	allocColorBins(gc);

	ScreenPoint[]       s            = new ScreenPoint[4];
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	JCChart3dArea       area         = dataView.getChart3d().getChart3dArea();
	JCViewport          viewport     = area.getViewport();

	for (int i = 0; i < numX - 1; i++) {
	    for (int j = 0; j < numY - 1; j++) {
			if (!viewport.isNormalized()) {
				int xi = xOffset + i;
				int yj = yOffset + j;
				s[0] = screenPoints[xi    ][yj];
				s[1] = screenPoints[xi    ][yj + 1];
				s[2] = screenPoints[xi + 1][yj + 1];
				s[3] = screenPoints[xi + 1][yj];
		        int x1 = JCChart3dUtil.min4(s[0].getXPixel(), s[1].getXPixel(),
                                            s[2].getXPixel(), s[3].getXPixel());
                int y1 = JCChart3dUtil.min4(s[0].getYPixel(), s[1].getYPixel(),
                                            s[2].getYPixel(), s[3].getYPixel());
                int x2 = JCChart3dUtil.max4(s[0].getXPixel(), s[1].getXPixel(),
                                            s[2].getXPixel(), s[3].getXPixel());
                int y2 = JCChart3dUtil.max4(s[0].getYPixel(), s[1].getYPixel(),
                                            s[2].getYPixel(), s[3].getYPixel());
                Rectangle rect = new Rectangle(x1, y1, x2 - x1, y2 - y1);
                if (!rect.intersects(area.getBounds())) {
                    continue;
                }
			}
			fillCell(gc, i, j);
	    }
	}
	flushColorBins();
	bins = null;
}

/**
 * Draws flat contours.  There is no surface to obscure them, so we can
 * optimize drawing by just looping through all the precompute vectors, and
 * drawing them.
 *
 * @param gc the graphics object to use for drawing
 */
protected void drawContours(Graphics gc)
{
	if (!contoured || contourLines == null) {
		return;
	}

	int maxPoints = 300;
	allocatePoints(maxPoints);

	ListIterator iteratorLines = contourLines.listIterator();
	while (iteratorLines.hasNext()) {
		ContourLine cLine = (ContourLine)iteratorLines.next();
		if (!setLineStyle(gc, cLine.getLevel())) {
			continue;
		}
		LinkedList segments = cLine.getSegments();
		ListIterator iteratorSegments = segments.listIterator();
		while(iteratorSegments.hasNext()) {
			int j = 0;
			while (j < maxPoints && iteratorSegments.hasNext()) {
				// Extract and save the data for rendering
			    ContourSegment segment =
                    (ContourSegment)iteratorSegments.next();
				xPoints[j] = segment.getXPixel();
				yPoints[j] = segment.getYPixel();
				j++;
			}
			gc.drawPolyline(xPoints, yPoints, j);
		}
	}

	xPoints = yPoints = null;
}

/**
 * Returns the contour lines in point data format.  Each contour level is
 * stored in a series.  The user can select a subset of the contour lines by
 * specifying the first and last contour.  Every <i>n</i>th contour line
 * (between first and last) can be chosen by setting the filter to be <i>n</i>.
 *
 * @param first the first contour line to be in the data set
 * @param last the last contour line to be in the data set (by setting this
 *     parameter to be <code>Integer.MAX_VALUE</code> you will get all the
 *     contours from first to the end of the contour line array).
 * @param filter by setting this value to be <i>n</i>, you get every
 *     <i>n</i>th contour line starting at first
 * @param combineLevels A contour level can be specified by more than one
 *     distinct line segment.  If <code>combineLevels</code> is <code>false</code>, each
 *     segment will be stored in its own series.  If <code>combineLevels</code>
 *     is <code>true</code>, all segements from the same level will be stored in one series.
 * @return A <code>Chart3dPointDataModel</code> (stored in a
 *     <code>JCDefault3dPointDataSource</code> object) that contains the point
 *     data.  If an error occurs, or no contour lines match the inputted values
 *     of first, last, and filter, null is returned.  <br><B>Note:</b> Either the
 *     contoured or zoned property of this <code>JCContour</code> object must
 *     be <code>true</code> for the contours to be generated.
 */
public Chart3dPointDataModel getDataContours(int first, int last, int filter,
											 boolean combineLevels)
{
    // Contour lines may not have been calculated yet
	if (contourLines == null) {
		recalcContours();
	}
    // If contourLines is still null (this is possible if they are not needed
    // for rendering), then we can't compute the contour data.
    if (contourLines == null) {
        return(null);
    }

    // Check that a valid range was provided
    if (first > last) {
        return(null);
    }
    filter = Math.max(1, filter);

    // Make sure we have valid ranges, discard out of bounds values
    int firstLevel = Math.max(0, first);
    int lastLevel  = Math.min(getContourLevels().getNumLevelsInternal() - 1, last);

    // First count the number of valid contour lines -- note that there can
    // be more than one contour line per level.  Keep track of levels
	// encountered using the levels array.
	int[] levels = new int[contourLines.size()];
	ListIterator iteratorLines = contourLines.listIterator();
    int count = 0;
	while (iteratorLines.hasNext()) {
		ContourLine cLine = (ContourLine)iteratorLines.next();
        int level = cLine.getLevel();

        // Ignore levels not requested
        if (level < firstLevel ||
            level > lastLevel  ||
            ((level - firstLevel) % filter) != 0)
        {
            continue;
        }

		boolean found = false;
		if (combineLevels) {
		    // Make sure it doesn't match one of the already stored levels
		    for (int index = 0; index < count; index++) {
			    if (levels[index] == level) {
				    found = true;
			    }
		    }
		}
		if (!found) {
			levels[count] = level;
            count++;
		}
	}
    if (count == 0) {
		// Nothing to generate
        return(null);
    }
    Point3d[][] points = new Point3d[count][];

    // Now build the individual series of the points array
	iteratorLines = contourLines.listIterator();
	int index = -1;
	while (iteratorLines.hasNext()) {
		ContourLine cLine = (ContourLine)iteratorLines.next();
        int level = cLine.getLevel();

        // Ignore levels not requested
        if (level < firstLevel ||
            level > lastLevel  ||
            ((level - firstLevel) % filter) != 0)
        {
            continue;
        }
        LinkedList segments = cLine.getSegments();

		// Compute the index for this level
		if (combineLevels) {
		    // Find the index of the level which matches this level
		    index = 0;
		    while (index < count) {
			    if (levels[index] == level) {
				    break;
			    }
			    index++;
		    }
		    if (index == count) {
			    // It didn't find it.  Something went very wrong
			    
			    return(null);
		    }
		}
		else {
			index++;
		}

		// Are we starting a new level or continuing an old one.
		int start = 0;
		if (points[index] == null || !combineLevels) {
			// New level
            points[index] = new Point3d[segments.size()];
		}
		else {
			// Old level -- Make new extended array, copy old points at beginning
			Point3d[] temp = points[index];
			points[index] = new Point3d[temp.length + segments.size()];
			for (int j = 0; j < temp.length; j++) {
				points[index][j] = temp[j];
			}
			start = temp.length;
		}

		// Copy this contour segment's points into array
		ListIterator iteratorSegments = segments.listIterator();
		for (int j = start; iteratorSegments.hasNext(); j++) {
            ContourSegment segment = (ContourSegment)iteratorSegments.next();
            points[index][j] = new Point3d(segment.getPoint());
		}
	}
    return(new JCDefault3dPointDataSource("dataContours", points));
}

}
