/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCContourLevels.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.util.JCNumberUtil;

import com.klg.jclass.chart3d.resources.LocaleBundle;

/**
 * This object deals with contour levels(distribution table).  It
 * calculates default levels (if <code>isDefault</code> is <code>true</code>), but also
 * allows the user to set their own levels.
 */
public class JCContourLevels extends ParentFormatter {

/** The default number of contour levels. */
public final static int DEFAULT_NUM_LEVELS = 10;

/**
 * Asks if linear contour levels are generated automatically (based on
 * <code>numLevels</code>).
 */
protected boolean            isDefault = true;

/** A strictly increasing array of contour levels. */
protected double []          levels    = null;

/** The number of contour levels.  Must be non-negative. */
protected int                numLevels = DEFAULT_NUM_LEVELS;

// Internal actual numLevels used.  Don't javadoc.  Initial value needs to
// match a null levels value.
protected int                numLevelsInternal = 0;

/** The contour minimum, calculated from the data. */
protected double             min = Double.MAX_VALUE;

/** The contour minimum, calculated from the data. */
protected double             max = Double.MAX_VALUE;

/**
 * No args constructor.  Values at default.
 */
public JCContourLevels() {}

/**
 * Constructor that sets the number of levels.
 *
 * @param numLevels the number of contour levels
 */
public JCContourLevels(int numLevels)
{
    if (numLevels < 0) {
        numLevels = 0;
        String s = LocaleBundle.string(LocaleBundle.INVALID_NUM_LEVELS);
        throw new IllegalArgumentException(s);
    }
    this.numLevels = numLevels;
    setNumLevelsInternal();
}

/**
 * Constructor that sets the contour levels array.  The <code>isDefault</code>
 * flag is set to <code>false</code>.
 *
 * @param levels a strictly increasing array of contour levels
 */
public JCContourLevels(double[] levels)
{
	this.levels = levels;
	isDefault = false;
    setNumLevelsInternal();
}

/**
 * Returns the current contour levels array (distribution table).
 *
 * @return an array of doubles the determine the contour levels
 */
public double[] getLevels()
{
	if (levels == null) {
		if (min == Double.MAX_VALUE) {
			// Defaults not calculated yet!
			Chart3dDataView dataView = (Chart3dDataView)parents.get(0);
			dataView.calcDefaults();
		}
	}
    return(levels);
}

/**
 * Sets a new <code>levels</code> array.  This has a side effect of turning
 * <code>isDefault</code> to <code>false</code>.  Also, <code>numLevels</code>
 * will be adjusted to be the size of the inputted array. <br><B>Note:</b> It is
 * assumed that the new levels array is sorted in increasing order without
 * duplicate values.
 *
 * @param levels the new set of contour levels
 */
public void setLevels(double[] levels)
{
	if (this.levels == levels) {
		return;
	}
	this.levels = levels;
    isDefault = false;
    setNumLevelsInternal();

	// Calculate nfrac from newly set levels
	int nfrac = 0;
	if (levels != null) {
		for (int i = 0; i < levels.length; i++) {
			nfrac = Math.max(JCChart3dUtil.nDigits(levels[i]), nfrac);
		}
	}
	setNFrac(Math.max(0, nfrac + 1));

	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Returns the actual number of levels used.  This may be different from the
 * number of levels specified by the user if <code>isDefault</code> is
 * <code>false</code>. In this case the actual number of levels used is the size
 * of the specified <code>levels</code> array.
 *
 * @return the current number of levels actually used
 */
public int getNumLevelsInternal()
{
    return(numLevelsInternal);
}

// Sets the internal number of levels value based on the current numLevels value
private void setNumLevelsInternal()
{
    if (isDefault) {
        numLevelsInternal = numLevels;
    }
    else {
        numLevelsInternal = (levels == null) ? 0 : levels.length;
    }
}

/**
 * Returns the current number of levels property as specified by the user.  This
 * value is ignored if <code>isDefault</code> is <code>false</code>.
 *
 * @return the current number of levels
 */
public int getNumLevels()
{
    return(numLevels);
}

/**
 * Sets the current number of levels. <br><b>Note:</b> <code>numLevels</code> is
 * ignored if <code>isDefault</code> is <code>false</code> (with actual value
 * used equal to the size of the <code>levels</code> array). Also note that
 * <code>numLevels</code> must be non-negative.
 *
 * @param numLevels the new number of contour levels
 */
public void setNumLevels(int numLevels)
{
	if (this.numLevels == numLevels) {
		return;
	}
    if (numLevels < 0) {
        numLevels = 0;
        String s = LocaleBundle.string(LocaleBundle.INVALID_NUM_LEVELS);
        throw new IllegalArgumentException(s);
    }
    this.numLevels = numLevels;
    setNumLevelsInternal();
	if (isDefault) {
	    callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
	}
}

/**
 * Are default distribution levels being used?  This value defaults to
 * <code>true</code>, but is automatically set to <code>false</code> if the user
 * sets the <code>levels</code> array, or calls methods <code>addLevel()</code>
 * or <code>removeLevel()</code>.
 *
 * @return Are default distribution levels being used?
 */
public boolean isDefault()
{
	return(isDefault);
}

/**
 * Sets the <code>isDefault</code> flag.  If set to <code>true</code>, JClass
 * Chart 3D will automatically calculate a default set of levels based on the
 * <code>numLevels</code> property.  If set to <code>false</code>, the current
 * levels array is used and <code>numLevels</code> becomes read-only.
 *
 * @param isDefault the new value os the <code>isDefault</code> flag
 */
public void setDefault(boolean isDefault)
{
	if (this.isDefault == isDefault) {
		return;
	}
	this.isDefault = isDefault;
    setNumLevelsInternal();
	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Returns the contour <code>min</code>.
 *
 * @return the current contour <code>min</code>
 */
public double getMin()
{
	return(min);
}

/**
 * Returns the contour <code>max</code>.
 *
 * @return the current contour <code>max</code>
 */
public double getMax()
{
	return(max);
}

/**
 * Adds a level at the given value.  The value will be added into the sorted
 * levels array.  Calling this method has the side effect of setting
 * <code>isDefault</code> to <code>false</code>.
 *
 * @param value the new contour level value
 */
public void addLevel(double value)
{
	int i, j;

	isDefault = false;
	int n = numLevelsInternal;
	double[] newLevels = new double[n + 1];
	for (i = 0; i < n && levels[i] < value; i++) {
		newLevels[i] = levels[i];
	}

	j = i;
	if (levels != null && levels[i] != value) {
		newLevels[j++] = value;
	}
	for (; i < n; i++) {
		newLevels[j++] = levels[i];
	}
	levels = newLevels;
	setNumLevelsInternal();
	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Removes the level at index <code>level</code>.  Calling this method has the
 * side effect of setting <code>isDefault</code> to <code>false</code>.
 *
 * @param level the level number to be removed
 */
public void removeLevel(int level)
{
	int n = numLevelsInternal;
	if (n == 0 || level < 0 || level > n - 1) {
		return;
	}

	isDefault = false;
	if (n <= 1)  {
		levels = null;
		setNumLevelsInternal();
		return;
	}

	double[] newLevels = new double[n - 1];
	for (int i = 0; i < n - 1; i++) {
		newLevels[i] = (i < level) ? levels[i] : levels[i+1];
	}
	levels = newLevels;
	setNumLevelsInternal();
	callParentSetChanged(true, Changeable.ALL | Changeable.CONTOURS);
}

/**
 * Removes the level closest to the give value.  Calling this method has the
 * side effect of setting <code>isDefault</code> to <code>false</code>.
 *
 * @param value the value that determines which level to remove
 */
public void removeLevel(double value)
{
	if (numLevelsInternal == 0) {
		return;
	}

	double smallestDiff   = Double.MAX_VALUE;
	int indexSmallestDiff = -1;
	for (int i = 0; i < numLevelsInternal; i++) {
		double diff = Math.abs(levels[i] - value);
		if (diff < smallestDiff) {
			smallestDiff      = diff;
			indexSmallestDiff = i;
		}
	}
	if (indexSmallestDiff >= 0) {
		removeLevel(indexSmallestDiff);
	}
}

/**
 * Calculates the contour level for this value.
 *
 * <p>
 * <b>Note:</b> A value between 0 and <code>numLevels</code> (inclusive - there
 * should be one more <code>contourStyle</code> than contour level) is returned.
 *
 * @param value the data value from which a contour level is calculated
 * @return
 */
public int getLevelFromValue(double value)
{
	int i;

	// Binary search would be faster!
	for (i = 0; i < numLevelsInternal; i++) {
		if (value <= levels[i]) {
			break;
		}
	}
    i = Math.min(i, numLevelsInternal - 1);

	return(i);
}

/**
 * Computes the contour limits (from the data).
 *
 * @param dataView the <code>dataView</code> for which to calculated the contour
 *     limits
 * @param elevationData the elevation data to use to calculate the contour
 *     limits
 * @param zoneData the zone data to use to calculate the contour limits
 */
protected void calcLimits(Chart3dDataView dataView,
			              Chart3dGridData elevationData,
						  Chart3dGridData zoneData)
{
	if (dataView == null || elevationData == null || zoneData == null) {
		
		return;
	}

	if (zoneData.isDataOK()) {
		zoneData.computeZLimits(elevationData, min, max);
		min = zoneData.getDataMin();
		max = zoneData.getDataMax();
	}
	else {
		min = elevationData.getDataMin();
		max = elevationData.getDataMax();

		// On bars, origin is adjustable, so take this into account in the
		// contour range.  We want zones to fill to origin.
		if (dataView.getChartType() == JCChart3d.BAR) {
			JCChart3dArea area = dataView.getChart3d().getChart3dArea();
			double origin      = area.getAxis(JCAxis.AXIS_Z).getOrigin();
			min                = Math.min(min, origin);
			max                = Math.max(max, origin);
		}
	}

	
	
}

/**
 * Cacluates default contour levels.
 */
protected void calcLevels()
{
    // Just to be safe
    setNumLevelsInternal();

    // User has supplied levels, no need to calculate.
	if (!isDefault) {
        
		return;
	}

    // Recalculate levels based on numLevelsInternal (which in this case
    // equals numLevels).
    levels = null;
    if (numLevelsInternal == 0) {
        
        return;
    }

	if (min == max) {
		// Special case, levels will be null, getNumLevelsInternal will return 0
		numLevelsInternal = 0;
        
		return;
	}

	levels = new double[numLevelsInternal];
	double interval = (max - min)/((double)(numLevelsInternal + 1));
	int nfrac = 0;
	if (1.0 + interval == 1.0) {
        if (min == 0.0) {
            // If min is 0, then we will get Integer.MAX_VALUE/2 back so try
            // max instead
			nfrac = JCChart3dUtil.nDigits(max);
        }
        else {
            nfrac = JCChart3dUtil.nDigits(min);
        }
	}
	else {
	    nfrac = JCChart3dUtil.nDigits(interval);
	}

	setNFrac(Math.max(0, nfrac + 1));
	for (int i = 0; i < numLevelsInternal; i++) {
		// Pick uniformly distributed levels, but keep them within the bounds
        // of the precision of the increment.  These numbers are chosen to be
        // within 1% of interval of their original value.

		levels[i] = min + (double)(i + 1) * interval;
		String s = JCNumberUtil.format(levels[i], getNFrac());
//		JCChart3dNumberFormat formatter = getNumberFormatter();
//      String label = formatter.localize(s);
		levels[i] = Double.valueOf(s).doubleValue();
	}
    
}

}
