/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCData3dContourIndex.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

/**
 * <code>JCData3dContourIndex</code> contains a unique index to a contour level
 * style and the minimum and maximum values of the range that it represents.
 * It is used by the chart's <code>pick()</code> methods, and also contains
 * information related to these operations.
 */
public class JCData3dContourIndex extends JCData3dIndex
                                  implements java.io.Serializable
{

/** The index of the contour style selected. */
protected int contourStyleIndex = -1;

/** The lower value of the range for the selected contour. */
protected double lowerContourRangeValue = -Double.MAX_VALUE;

/** The upper value of the range for the selected contour. */
protected double upperContourRangeValue = Double.MAX_VALUE;

/**
 * Simple constructor for <code>JCData3dContourIndex</code>.
 */
public JCData3dContourIndex()
{
}

/**
 * Constructor that creates a data index from a data view, leaving
 * other values to their default, undefined values.
 *
 * @param dataView the data view represented by this class
 */
public JCData3dContourIndex(Chart3dDataView dataView)
{
	super(dataView);
}

/**
 * Constructor that creates a data index from a data view, an index
 * to a contour level style, and the minimum and maximum values of
 * the range that the style specifies.  If the contour range values
 * are equal, then specific contour level is being referenced as opposed
 * to a range.
 *
 * @param dataView the data view that holds the contour indices
 * @param styleIndex the index of the contour style that is used to illustrate
 *      this range or contour
 * @param lowerVal The lower value of the contour range. If equal to
 *     <code>upperVal</code>, this object refers to a specific contour rather
 *     than a range.
 * @param upperVal The higher value of the contour range. If equal to
 *     <code>upperVal</code>, this object refers to a specific contour rather
 *     than a range.
 */
public JCData3dContourIndex(Chart3dDataView dataView, int styleIndex,
							double lowerVal, double upperVal)
{
	super(dataView);
	contourStyleIndex = styleIndex;
	upperContourRangeValue = upperVal;
	lowerContourRangeValue = lowerVal;
}

/**
 * Gets the value of the <code>ContourStyleIndex</code> property.  It contains
 * the index to the contour style used to illustrate the contour range specified
 * by this object.  If negative, then no contour style is associated with
 * this range.
 *
 * @return the index of the contour style used to illustrate this contour
 *     range
 */
public int getContourStyleIndex()
{
	return(contourStyleIndex);
}

/**
 * Sets the value of the <code>ContourStyleIndex</code> property.  It refers
 * the index to the contour style used to illustrate the contour range specified
 * by this object.  If negative, then no contour style is associated with
 * this range.
 *
 * @param index the index of the contour style used to illustrate this contour
 *      range
 */
public void setContourStyleIndex(int index)
{
	if (this.contourStyleIndex == index) {
		return;
	}
	this.contourStyleIndex = index;
	setChanged(true, Changeable.REDRAW);
}


/**
 * Gets the value of the <code>UpperContourRangeValue</code> property.  It
 * contains the upper value of the contour range specified by this object.
 * If this property is equal to <code>LowerContourRangeValue</code>, then this
 * object refers not to a contour range but to a single contour level.
 *
 * @return the upper value of the contour range
 */
public double getUpperContourRangeValue() {
	return(upperContourRangeValue);
}

/**
 * Sets the value of the <code>UpperContourRangeValue</code> property.  This
 * sets the upper value of the contour range specified by this object. If this
 * property is equal to <code>LowerContourRangeValue</code>, then this object
 * refers not to a contour range but to a single contour level.
 *
 * @param value the upper value of the contour range
 */
public void setUpperContourRangeValue(double value)
{
	if (this.upperContourRangeValue == value) {
		return;
	}
	this.upperContourRangeValue = value;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Gets the value of the <code>LowerContourRangeValue</code> property.  It
 * contains the upper value of the contour range specified by this object. If
 * this property is equal to <code>UpperContourRangeValue</code>, then this
 * object refers not to a contour range but to a single contour level.
 *
 * @return the upper value of the contour range
 */
public double getLowerContourRangeValue() {
	return(lowerContourRangeValue);
}

/**
 * Sets the value of the <code>LowerContourRangeValue</code> property.  This
 * sets the upper value of the contour range specified by this object. If this
 * property is equal to <code>UpperContourRangeValue</code>, then this object
 * refers not to a contour range but to a single contour level.
 *
 * @param value the upper value of the contour range
 */
public void setLowerContourRangeValue(double value)
{
	if (this.lowerContourRangeValue == value) {
		return;
	}
	this.lowerContourRangeValue = value;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Determines if the specified object refers to the same data point as this
 * <code>JCData3dContourIndex</code> instance.
 *
 * @param o object to compare
 * @return if the passed object is a <code>JCData3dContourIndex</code>
 *     instance and refers to the same data point as this instance
 */
public boolean equals(Object o) {
	if (o == null) {
		return(false);
	}

	if (o instanceof JCData3dContourIndex) {
		JCData3dContourIndex di = (JCData3dContourIndex)o;
		return(dataView == di.dataView &&
					contourStyleIndex == di.contourStyleIndex &&
					upperContourRangeValue == di.upperContourRangeValue &&
					lowerContourRangeValue == di.lowerContourRangeValue);
	}
	else {
		return(false);
	}
}


}
