/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCGridLines.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Color;
import java.awt.Graphics;
import java.util.ArrayList;
import javax.vecmath.Point3d;
import javax.vecmath.Point4d;
import javax.vecmath.Matrix4d;

/**
 * Specifies which gridlines are drawn for a given axis.  This information
 * is specified on a plane by plane basis using a <code>plane_mask</code>.
 */
public class JCGridLines implements java.io.Serializable {

/** No planes have been specified. */
public final static int NO_PLANE   = 0;

/** The <i>XY</i> plane mask. */
public final static int XY_PLANE   = 1;

/** The <i>XZ</i> plane mask. */
public final static int XZ_PLANE   = 2;

/** The <i>YZ</i> plane mask. */
public final static int YZ_PLANE   = 4;

/** All planes have been specified. */
public final static int ALL_PLANES = XY_PLANE | XZ_PLANE | YZ_PLANE;

/** The plane mask. A logic <code>OR</code> of the three possible planes. */
protected int planeMask = NO_PLANE;

/** Gridline color, width, and pattern. */
protected JCLineStyle lineStyle = null;

/** The parent axis. */
protected JCAxis axis = null;

// Local variables
protected Point4d data1 = null;
protected Point4d data2 = null;

/**
 * Null constructor.  Parent is null. Set mask to <code>NO_PLANES</code> and use
 * default line style.
 */
public JCGridLines()
{
    lineStyle = new JCLineStyle(1, Color.black, JCLineStyle.SOLID);
}

/**
 * Constructor that specifies the axis and <code>planeMask</code>.
 *
 * @param axis the parent axis
 * @param planeMask mask which determines on which plane(s) to draw gridlines
 */
public JCGridLines(JCAxis axis, int planeMask)
{
	this();
	this.lineStyle.parent = axis;
	this.axis             = axis;
	this.planeMask        = planeMask;
}

/**
 * Constructor which specifies axis, <code>planeMask</code>, and
 * <code>lineStyle</code>.
 *
 * @param axis the parent axis
 * @param planeMask mask which determines on which plane(s) to draw gridlines
 * @param lineStyle the <code>lineStyle</code> for this set of gridlines
 */
public JCGridLines(JCAxis axis, int planeMask, JCLineStyle lineStyle)
{
	this.axis             = axis;
	this.planeMask        = planeMask;
	this.lineStyle        = lineStyle;
	this.lineStyle.parent = axis;
}

/**
 * Returns the plane mask that determines on which plane(s) to draw gridlines.
 * It is a logical <code>OR</code> of <code>XY_PLANE</code>,
 * <code>XZ_PLANE</code>, and <code>YZ_PLANE</code>, two of which make sense for
 * a given axis.
 *
 * @return The current plane mask
 */
public int getPlaneMask()
{
	return(planeMask);
}

/**
 * Sets the plane mask which determines on which plane(s) to draw gridlines.
 * It is a logical <code>OR</code> of <code>XY_PLANE</code>,
 * <code>XZ_PLANE</code>, and <code>YZ_PLANE</code>, two of which make sense for
 * a given axis.
 *
 * @param planeMask The new plane mask
 */
public void setPlaneMask(int planeMask)
{
	if (this.planeMask == planeMask) {
		return;
	}
	this.planeMask = planeMask;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Gets the <code>JCLineStyle</code> property, which controls how gridlines
 * are drawn.
 *
 * @return <code>JCLineStyle</code> instance representing the gridline
 *     drawing style
 */
public JCLineStyle getLineStyle()
{
    return(lineStyle);
}

/**
 * Sets the <code>JCLineStyle</code> property, which controls how gridlines
 * are drawn.
 *
 * @param lineStyle The new grid line style.
 */
public void setLineStyle(JCLineStyle lineStyle)
{
	if (this.lineStyle == lineStyle) {
		return;
	}
	this.lineStyle        = lineStyle;
	this.lineStyle.parent = axis;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Gets the value of the <code>Axis</code> property. The <code>Axis</code>
 * property specifies the axis associated with this gridline.
 *
 * @return The axis associated with this gridline
 */
public JCAxis getAxis()
{
	return(axis);
}

/**
 * Sets the value of the <code>Axis</code> property.  The <code>Axis</code
 * property specifies the axis associated with this gridline.
 *
 * @param axis The new axis associated with this gridline
 */
public void setAxis(JCAxis axis)
{
	if (this.axis == axis) {
		return;
	}
	this.axis = axis;
	if (lineStyle != null) {
		lineStyle.parent = axis;
	}
	setChanged(true, Changeable.RECALC);
}

/**
 * Tells the parent (if there is one) that a change has happened.
 *
 * @param b Has anything changed?
 * @param cf the change mask which tells what type of change
 */
protected void setChanged(boolean b, int cf)
{
	if (axis != null) {
		axis.callParentSetChanged(b, cf);
	}
}

/**
 * Computes the end points of the gridline, and draws the gridline if the user
 * has specifically asked for it to be drawn.
 *
 * @param gc the graphics object used for drawing
 * @param ctm the current transform matrix (determined by caller)
 * @param value the value along the axis at which to draw
 * @param faceFlag which face to draw on for the parent axis
 * @param facePoint a representative point on the face which is being drawn
 */
protected void drawSingleGridLine(Graphics gc, Matrix4d ctm, double value,
								  int faceFlag, Point3d facePoint)
{
	int     axisId = axis.getAxisId();

	// First time through
	if (data1 == null) {
	   data1 = new Point4d();
	   data2 = new Point4d();
	}

	boolean draw = false;
	if ((axisId == JCAxis.AXIS_X)        &&     // X Grid lines & XY-Plane
	    ((faceFlag & Cube.VP_YMASK) > 0) &&
		((planeMask & XY_PLANE) > 0))
	{
		data1.set(value, -1.0, facePoint.z, 1.0);
		data2.set(value,  1.0, facePoint.z, 1.0);
		draw = true;
	}
	else if ((axisId == JCAxis.AXIS_X)      &&  // X Grid lines & XZ-Plane
		     ((faceFlag & Cube.VP_ZMASK) > 0) &&
			 ((planeMask & XZ_PLANE) > 0))
	{
		data1.set(value, facePoint.y, -1.0, 1.0);
		data2.set(value, facePoint.y,  1.0, 1.0);
		draw = true;
	}
	else if ((axisId == JCAxis.AXIS_Y)      &&  // Y Grid lines & XY-Plane
		     ((faceFlag & Cube.VP_XMASK) > 0) &&
			 ((planeMask & XY_PLANE) > 0))
	{
		data1.set(-1.0, value, facePoint.z, 1.0);
		data2.set( 1.0, value, facePoint.z, 1.0);
		draw = true;
	}
	else if ((axisId == JCAxis.AXIS_Y)      &&  // Y Grid lines & YZ-Plane
		     ((faceFlag & Cube.VP_ZMASK) > 0) &&
			 ((planeMask & YZ_PLANE) > 0))
	{
		data1.set(facePoint.x, value, -1.0, 1.0);
		data2.set(facePoint.x, value,  1.0, 1.0);
		draw = true;
	}
	else if ((axisId == JCAxis.AXIS_Z)      &&  // Z Grid lines & XZ-Plane
		     ((faceFlag & Cube.VP_XMASK) > 0) &&
			 ((planeMask & XZ_PLANE) > 0))
	{
		data1.set(-1.0, facePoint.y, value, 1.0);
		data2.set( 1.0, facePoint.y, value, 1.0);
		draw = true;
	}
	else if ((axisId == JCAxis.AXIS_Z)      &&  // Z Grid lines & YZ-Plane
		     ((faceFlag & Cube.VP_YMASK) > 0) &&
			 ((planeMask & YZ_PLANE) > 0))
	{
		data1.set(facePoint.x, -1.0, value, 1.0);
		data2.set(facePoint.x,  1.0, value, 1.0);
		draw = true;
	}


	if (draw) {					// Finally, transform and draw grid line
		Transform.transformPoint(ctm, data1);
		int x1 = (int) Math.round(data1.x);
		int y1 = (int) Math.round(data1.y);

		Transform.transformPoint(ctm, data2);
		int x2 = (int) Math.round(data2.x);
		int y2 = (int) Math.round(data2.y);

		gc.drawLine(x1, y1, x2, y2);
	}
}

/**
 * Draws gridlines on a particular grid face (indicated by <code>faceFlag</code>).
 *
 * @param gc the graphics object used for drawing
 * @param ctm the current transform matrix (determined by caller)
 * @param faceFlag which face to draw on for the parent axis
 * @param facePoint a representative point on the face which is being drawn
 */
protected void drawGridFace(Graphics gc, Matrix4d ctm, int faceFlag,
                            Point3d facePoint)
{
	// Are we drawing anything?
	if (lineStyle == null         ||
		lineStyle.getWidth() <= 0 ||
		lineStyle.getPattern() == JCLineStyle.NONE)
	{
		return;
	}

	// Dataport min & max too close together?
	double min = axis.getDataportMin();
	double max = axis.getDataportMax();
	if (1.0 + (max - min) == 1.0) {
		return;
	}

	lineStyle.updateGraphics(gc);
    ArrayList valueLabels = axis.getAnnoHandler().getValueLabels();
    if (valueLabels == null || valueLabels.size() == 0) {
        return;
    }
    for (int i = 0; i < valueLabels.size(); i++) {
        JCValueLabel vLabel = (JCValueLabel)valueLabels.get(i);
        double value = vLabel.getValue();

		// Don't draw grid lines for minor ticks and on the plot cube
		if (vLabel.isMinorTick()      ||
		    value <= min              ||
		    value >= max              ||
		    value < axis.getAnnoMin() ||
			value > axis.getAnnoMax())
		{
			continue;
		}
		if (axis.getAnnotationMethod() == JCAxis.ANNOTATION_DATA_LABELS &&
		    axis.isCombineLabels()                                      &&
		    !vLabel.isTickDrawn())
		{
			continue;
		}
		double newValue = 2.0 * (value - min) / (max - min) - 1.0;
		drawSingleGridLine(gc, ctm, newValue, faceFlag, facePoint);
	}
	lineStyle.resetGraphics(gc);
}

}
