/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCLineStyle.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;

import java.awt.*;

/**
 * <code>JCLineStyle</code> controls how a line is drawn. It stores the line
 * pattern, line color, and line width, and then draws lines accordingly.
 */
public class JCLineStyle extends JCStyle {

/** Sets the join style. */
protected int join = 0;

/** Sets the cap style. */
protected int cap = 0;

/** Sets the line pattern. */
protected int pattern = SOLID;

/** Sets the line width. */
protected int width = 1;

/** Sets the index into the default color array. */
protected int colorIndex = 0;

/** Sets the line color. */
protected Color color = null;

// Line patterns
public final static int  NONE       = 0;
public final static int  SOLID      = 1;
public final static int  LONG_DASH  = 2;
public final static int  SHORT_DASH = 3;
public final static int  LSL_DASH   = 4;
public final static int  DASH_DOT   = 5;
public final static int  CUSTOM     = 6;

// Line pattern arrays
public final static float[] LONG_DASH_ARRAY = {10.0F, 10.0F};
public final static float[] SHORT_DASH_ARRAY = {5.0F, 10.0F};
public final static float[] LSL_DASH_ARRAY = {10.0F, 10.0F, 5.0F, 10.0F};
public final static float[] DASH_DOT_ARRAY = {10.0F, 10.0F, 1.0F, 10.0F};

// Legend line pattern arrays
public final static float[] LEGEND_LONG_DASH_ARRAY = {5.0F, 5.0F};
public final static float[] LEGEND_SHORT_DASH_ARRAY = {2.5F, 5.0F};
public final static float[] LEGEND_LSL_DASH_ARRAY = {5.0F, 5.0F, 2.5F, 5.0F};
public final static float[] LEGEND_DASH_DOT_ARRAY = {5.0F, 5.0F, 1.0F, 5.0F};

private float[] dashArray = null;
private float[] legendDashArray = null;

/**
 * Constructor for line style objects.
 *
 * @param width line width
 * @param color line color
 * @param pattern line pattern; one of <code>NONE</code>, <code>SOLID</code>,
 * <code>LONG_DASH</code>, <code>SHORT_DASH</code>, <code>LSL_DASH</code>,
 * <code>DASH_DOT</code>, or <code>CUSTOM</code>
 */
public JCLineStyle(int width, Color color, int pattern)
{
	this.width   = width;
	this.color   = color;
	this.pattern = pattern;
	join = BasicStroke.JOIN_MITER;
	cap  = BasicStroke.CAP_BUTT;
	updateDashArray();
}

/**
 * Full Constructor for line style objects.
 *
 * @param width line width
 * @param color line color
 * @param pattern line pattern; one of <code>NONE</code>, <code>SOLID</code>,
 * <code>LONG_DASH</code>, <code>SHORT_DASH</code>, <code>LSL_DASH</code>,
 * <code>DASH_DOT</code>, or <code>CUSTOM</code>
 * @param cap
 * @param join
 */
public JCLineStyle(int width, Color color, int pattern, int cap, int join)
{
	this.width   = width;
	this.color   = color;
	this.pattern = pattern;
	this.cap     = cap;
	this.join    = join;
	updateDashArray();
}

/**
 * Full Constructor for line style objects.
 *
 * @param width line width
 * @param color line color
 * @param pattern line pattern
 * @param cap line cap style
 * @param join line join style
 * @param dashArray line dash array
 * @param legendDashArray legend dash array
 */
public JCLineStyle(int width, Color color, int pattern, int cap, int join,
                   float[] dashArray, float[] legendDashArray)
{
	this.width     = width;
	this.color     = color;
	this.pattern   = pattern;
	this.cap       = cap;
	this.join      = join;
	this.dashArray = dashArray;
	if (legendDashArray != null) {
		this.legendDashArray = legendDashArray;
	}
    else {
		this.legendDashArray = dashArray;
	}
	updateDashArray();
}

/**
 * Creates a default line style.
 *
 * @param c the chart for which to generate the default line style
 * @return <code>JCLineStyle</code> object with default width, color, and
 *     pattern.  The defaults are cycled, so consecutive calls will return
 *     different line styles.
 */
public static JCLineStyle makeDefault(JCChart3d c)
{
	JCLineStyle style = new JCLineStyle(1, getDefaultColor(c, LINE_STYLE),
                                        getDefaultPattern());
	int index = 0;
	if (c != null) {
		index = c.getLineColorIndex();
		if (index == 0) {
			index = defaultColors.length - 1;
		}
		else {
			index--;
		}
	}
	style.colorIndex = index;
	return(style);
}

/**
 * Updates the graphics object with the elements of this line style (strokes
 * and colors).
 *
 * @param gc <code>Graphics</code> object to change
 * @return <code>true</code> if you should now use this object to draw; <code>false</code>
 * if the style is such that no drawing should occur
 */
public boolean updateGraphics(Graphics gc)
{
	return updateGraphics(gc, -1);
}

/**
 * Updates the graphics object with the elements of this line style (strokes
 * and colors).  This method scales the dashed lines of the stroke down for
 * small line lengths, such as need to be drawn in the legend.
 *
 * @param gc <code>Graphics</code> object to change
 * @param drawLineLength Maximum length of line to be drawn in pixels. This is used
 *     to scale dashed lines down such that the line will be drawn with the
 *     appropriate pattern.
 * @return <code>true</code> if you should now use this object to draw; <code>false</code> if the style
 *     is such that no drawing should occur
 */
public boolean updateGraphics(Graphics gc, int drawLineLength)
{
	if (pattern == NONE || width <= 0) {
        return false;
	}

	if (!gc.getColor().equals(color)) {
		gc.setColor(color);
	}

	Graphics2D gc2D = (Graphics2D) gc;

	if (pattern == SOLID) {
		gc2D.setStroke(new BasicStroke(width, cap, join));
	}
    else {
		// If draw line length given, scale dashed array down to length.
		if (drawLineLength > 0) {

			// Calculate length of dashed line we want to draw from the
            // dashed line array.  Remember to repeat the first element
			// at the end to give that nice, all rounded look.
			float dashLineLength = 0.0F;
			for (int i = 0; i < dashArray.length; i++) {
				dashLineLength += dashArray[i];
			}
			dashLineLength += dashArray[0];

			float[] newDashArray = new float[dashArray.length + 1];

			// Make sure no dashed bit goes beneath 1.0 (won't draw) by
			// first counting the number of segments that dip below that
			// length
			int roundUps = 0;
			for (int i = 0; i < newDashArray.length; i++) {
				float value;
				if (i == newDashArray.length - 1) {
					value = dashArray[0];
				}
                else {
					value = dashArray[i];
				}
				float scaledValue = (float) Math.floor(
					((drawLineLength/dashLineLength) * value));
				if (scaledValue < 1.0) {
					roundUps++;
				}
			}

			// Remove the bits we have to leave at 1.0 from the lengths
			float modDashLineLength = dashLineLength - (float) roundUps;
			float modDrawLineLength = drawLineLength - (float) roundUps;

			// Now, really calculate the new length of the dashed segments
			for (int i = 0; i < newDashArray.length; i++) {
				float value;
				if (i == newDashArray.length - 1) {
					value = dashArray[0];
				}
                else {
					value = dashArray[i];
				}

				// First, leave the bits less than 1.0 at a length of 1.0
				float scaledValue = (float) Math.floor(
					((drawLineLength/dashLineLength) * value));
				if (scaledValue < 1.0) {
					scaledValue = 1.0F;
				}
                else {
					// Scale correctly, with the above bits left out of the
					// scaling.
					scaledValue =
                        ((modDrawLineLength/modDashLineLength) * value);
				}
				newDashArray[i] = scaledValue;
			}

			gc2D.setStroke(new BasicStroke(width, cap, join,
									       10.0F, newDashArray, 0.0F));
		}
        else {  // Otherwise, it's just a straight set of the dashed array.
			gc2D.setStroke(new BasicStroke(width, cap, join,
									       10.0F, dashArray, 0.0F));
		}
	}

	return(true);
}

/**
 * Resets the graphics object to a default drawing style.
 *
 * @param gc the graphics object to reset
 */
public void resetGraphics(Graphics gc)
{
	Graphics2D gc2D = (Graphics2D) gc;
	gc2D.setStroke(new BasicStroke(1));
}

protected int xp[] = null;
protected int yp[] = null;

/**
 * Draws a line between two points according to the properties of this
 * line style.
 *
 * @param gc graphics context to use for drawing
 * @param x1 <i>x</i> value of first point (in pixels)
 * @param y1 <i>y</i> value of first point (in pixels)
 * @param x2 <i>x</i> value of second point (in pixels)
 * @param y2 <i>y</i> value of second point (in pixels)
 */
public void draw(Graphics gc, int x1, int y1, int x2, int y2)
{
	draw(gc, x1, y1, x2, y2, null);
}

/**
 * Draws a line between two points according to the the properties of this
 * line style, except that the color is overridden with the specified color.
 *
 * @param gc graphics context to use for drawing
 * @param x1 <i>x</i> value of first point (in pixels)
 * @param y1 <i>y</i> value of first point (in pixels)
 * @param x2 <i>x</i> value of second point (in pixels)
 * @param y2 <i>y</i> value of second point (in pixels)
 * @param col overriding color
 */
public void draw(Graphics gc, int x1, int y1, int x2, int y2, Color col)
{
	if (!updateGraphics(gc)) {
        return;
	}

	if (col != null) {
		if (!(gc.getColor().equals(col))) {
			gc.setColor(col);
		}
	}

	drawLine(gc, x1, y1, x2, y2);

	resetGraphics(gc);
}

/**
 * Draws a line between two points according to the given <code>Graphics</code>
 * object (it must already be set with the appropriate drawing
 * attributes).
 *
 * @param gc graphics context to use for drawing
 * @param x1 <i>x</i> value of first point (in pixels)
 * @param y1 <i>y</i> value of first point (in pixels)
 * @param x2 <i>x</i> value of second point (in pixels)
 * @param y2 <i>y</i> value of second point (in pixels)
 */
public void drawLine(Graphics gc, int x1, int y1, int x2, int y2)
{
	gc.drawLine(x1, y1, x2, y2);
}

/**
 * Draws a squared line between two points according to the given <code>Graphics
 * </code>
 * object (it must already be set with the appropriate drawing
 * attributes).  A squared line is one that, when angled and having a width
 * greater than 1, will square off the edges to the appropriate vertical
 * or horizontal.
 *
 * @param gc graphics context to use for drawing
 * @param x1 <i>x</i> value of first point (in pixels)
 * @param y1 <i>y</i> value of first point (in pixels)
 * @param x2 <i>x</i> value of second point (in pixels)
 * @param y2 <i>y</i> value of second point (in pixels)
 * @param horizOrVert Is this a vertically oriented (<code>false</code>) or horizontally
 *     oriented (<code>true</code>) line?
 * @param delta the delta value for 3D calculations in the orientation
 *     direction
 * @param min the minimum value to clip against in the orientation direction
 * @param max the maximum value to clip against in the orientation direction
 */
public void drawSquaredLine(Graphics gc, int x1, int y1, int x2, int y2,
				            boolean horizOrVert, int delta, int min, int max)
{
	if (width == 1) {
		gc.drawLine(x1, y1, x2, y2);
	}
	else {
		// Attempt to draw rectangle to represent wide line.

		// Vertical line
		if (x1 == x2) {
			int top = Math.min(y1, y2);
			int h = Math.abs(y2 - y1);
			// Coming from a vertical line drawing perspective
			if (!horizOrVert) {
				gc.fillRect(x1-width/2, top, width, h+1);
			}
            // Coming from a horizontal line drawing perspective
            else {
				int fillwidth = width;
				int left = x1 - width/2;
				if (left < min) {
					fillwidth -= (min - left);
					left = min;
				}
                else if (left + fillwidth > max) {
					fillwidth -= (left + fillwidth - max);
				}
				gc.fillRect(left, top, fillwidth, h + 1);
			}
		}
		// Horizontal line
		else if (y1 == y2) {
			int left = Math.min(x1, x2);
			int w = Math.abs(x2 - x1);
			// Coming from a vertical line drawing perspective
			if (!horizOrVert) {
				int fillheight = width;
				int top = y1 - width/2;
				if (top < min) {
					fillheight -= (min-top);
					top = min;
				}
                else if (top + fillheight > max) {
					fillheight -= (top+fillheight-max);
				}
				gc.fillRect(left, top, w, fillheight);
			}
            // Coming from a horizontal line drawing perspective
            else {
				gc.fillRect(left, y1 - width/2, w, width);
			}
		}
		else {
			if (xp == null || yp == null) {
				xp = new int[4];
				yp = new int[4];
			}
			// Vertically oriented lines -- square off ends vertically
			if (!horizOrVert) {
				// Adjust to spread drawn y values on either side of
				// actual y value
				int yadj1 = width / 2;
				int yadj2 = (width + 1) / 2;

				// x values as given, y values adjusted
				xp[0] = x1;
				yp[0] = y1+yadj1;
				xp[1] = x2;
				yp[1] = y2+yadj1;
				xp[2] = x2;
				yp[2] = y2-yadj2;
				xp[3] = x1;
				yp[3] = y1-yadj2;

				// Clip to min and max -- positions 2 and 3 need added delta
				for (int i = 0; i < 4; i++) {
					int extra = 0;
					if (i == 1 || i == 2) {
						extra = delta;
					}
					if (yp[i] < min + extra) {
						yp[i] = min + extra;
					}
                    else if (yp[i] > max + extra) {
						yp[i] = max + extra;
					}
				}
			}
			// Horizontally oriented lines -- square off ends horizontally
            else {
				// Adjust to spread drawn x values on either side of
				// actual x value
				int xadj1 = width/2;
				int xadj2 = (width + 1)/2;
				// y values as given, x values adjusted
				xp[0] = x1 + xadj1;
				yp[0] = y1;
				xp[1] = x2 + xadj1;
				yp[1] = y2 + 1;
				xp[2] = x2 - xadj2;
				yp[2] = y2 + 1;
				xp[3] = x1 - xadj2;
				yp[3] = y1;

				// Clip to min and max -- positions 2 and 3 need added delta
				for (int i = 0; i < 4; i++) {
					int extra = 0;
					if (i == 1 || i == 2) {
						extra = delta;
					}
					if (xp[i] < min + extra) {
						xp[i] = min + extra;
					}
                    else if (xp[i] >= max + extra) {
						xp[i] = max + extra;
					}
				}
			}
			gc.fillPolygon(xp, yp, 4);
		}
	}
}

/**
 * Performs the requested draw rect given the values of the properties
 * of the <code>JCLineStyle</code> instance.
 *
 * @param gc graphics context to use for drawing
 * @param x <i>x</i> origin of rect region (in pixels)
 * @param y <i>y</i> origin of rect region (in pixels)
 * @param width width of rect region (in pixels)
 * @param height height of rect region (in pixels)
 */
public void drawRect(java.awt.Graphics gc, int x, int y, int width, int height)
{
	if (updateGraphics(gc)) {
		gc.drawRect(x, y, width, height);
		resetGraphics(gc);
	}
}

/**
 * Performs the requested fill given the values of the properties
 * of the <code>JCLineStyle</code> instance.
 *
 * @param gc graphics context to use for drawing
 * @param x <i>x</i> origin of fill region (in pixels)
 * @param y <i>y</i> origin of fill region (in pixels)
 * @param width width of fill region (in pixels)
 * @param height height of fill region (in pixels)
 */
public void fillRect(java.awt.Graphics gc, int x, int y, int width, int height)
{
	if (updateGraphics(gc)) {
		gc.fillRect(x, y, width, height);
		resetGraphics(gc);
	}
}

/**
 * Gets the <code>Pattern</code> property, which dictates the pattern
 * used to draw a line.
 *
 * @return one of <code>NONE</code>, <code>SOLID</code>, <code>LONG_DASH</code>,
 * <code>SHORT_DASH</code>, <code>LSL_DASH</code>, or <code>DASH_DOT</code>
 */
public int getPattern()
{
    return(pattern);
}

/**
 * Sets the <code>Pattern</code> property, which dictates the pattern
 * used to draw a line.
 *
 * @param p one of <code>NONE</code>, <code>SOLID</code>, <code>LONG_DASH</code>,
 * <code>SHORT_DASH</code>, <code>LSL_DASH</code>, or <code>DASH_DOT</code>
 */
public void setPattern(int p)
{
	if (p == pattern) return;
	switch (p) {
	case NONE:
	case SOLID:
	case LONG_DASH:
	case SHORT_DASH:
	case LSL_DASH:
	case DASH_DOT:
		pattern = p;
		break;
	default:
		throw new IllegalArgumentException(
				LocaleBundle.INVALID_LINESTYLE_ERROR
										   );
	}
	updateDashArray();

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Sets the <code>Pattern</code> property, which dictates the pattern used to
 * draw a line, with a custom user-defined pattern.
 * <BR>
 * The pattern type is automatically set to </code>CUSTOM</code>.
 *
 * @param patternArray an array of floats representing the pattern to use when
 *      drawing this line
 * @param legendPatternArray An array of floats representing the pattern to
 *     use when drawing this line in the legend. Usually, this means taking
 *     the pattern array and cutting the values down.  Half seems to work well,
 *     although if the values are too small nothing will be drawn.
 */
public void setPattern(float patternArray[], float legendPatternArray[])
{
	pattern = CUSTOM;
	dashArray = patternArray;
	if (legendPatternArray != null) {
		legendDashArray = legendPatternArray;
	}
	else {
		legendDashArray = patternArray;
	}

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Updates the dashed array according to the set pattern.
 */
private void updateDashArray()
{
	switch (pattern) {
		case NONE:
			dashArray = null;
			legendDashArray = null;
			break;
		case SOLID:
			dashArray = null;
			legendDashArray = null;
			break;
		case LONG_DASH:
			dashArray = LONG_DASH_ARRAY;
			legendDashArray = LONG_DASH_ARRAY;
			break;
		case SHORT_DASH:
			dashArray = SHORT_DASH_ARRAY;
			legendDashArray = SHORT_DASH_ARRAY;
			break;
		case LSL_DASH:
			dashArray = LSL_DASH_ARRAY;
			legendDashArray = LSL_DASH_ARRAY;
			break;
		case DASH_DOT:
			dashArray = DASH_DOT_ARRAY;
			legendDashArray = DASH_DOT_ARRAY;
			break;
		case CUSTOM:
			break;
		default:
			dashArray = null;
			legendDashArray = null;
			break;
	}
}

/**
 * Gets the <code>Width</code> property, which controls line width.
 *
 * @return positive integer representing line width
 */
public int getWidth()
{
    return(width);
}

/**
 * Sets the <code>Width</code> property, which controls line width.
 *
 * @param w positive integer representing line width
 */
public void setWidth(int w)
{
	if (width == w) {
		return;
	}
	width = w;
	// For JCChart, HLOC and Candle chart types need a RECALC
	setChanged(true, Changeable.RECALC | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Gets the <code>Color</code> property of <code>JCLineStyle</code>, which determines the
 * color used to draw the line.
 *
 * @return AWT Color class representing the color to be used to draw the
 *     lines. If null, the current color of the <code>Graphics</code> object is used.
 */
public Color getColor()
{
    return(color);
}

/**
 * Sets the <code>Color</code> property of <code>JCLineStyle</code>, which determines the
 * color used to draw the line.
 *
 * @param color AWT Color class representing the color to be used to draw the
 *     lines. If null, the current color of the <code>Graphics</code> object is used.
 */
public void setColor(Color color) {
	if (this.color == color) {
		return;
	}
	this.color = color;

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Returns the index used to generate the color.  Only set if default colors are used.
 *
 * @return the index into the default color array
 */
public int getColorIndex()
{
	return(colorIndex);
}

/**
 * The default line pattern.
 * @return
 */
static int getDefaultPattern()
{
	return(SOLID);
}

/**
 * Gets the <code>Join</code> property, which dictates the join style to use
 * when joining two lines.
 *
 * @return one of <code>BasicStroke.JOIN_MITER</code>, <code>BasicStroke.JOIN_BEVEL</code>, or
 *     <code>BasicStroke.JOIN_ROUND</code>.
 */
public int getJoin()
{
    return(join);
}

/**
 * Sets the <code>Join</code> property, which dictates the join style to use when
 * joining two lines.
 *
 * @param join the new join style
 */
public void setJoin(int join) {
	if (this.join == join) {
		return;
	}
	this.join = join;

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Gets the <code>Cap</code> property, which dictates the cap style to use at
 * the ends of a line.
 *
 * @return one of <code>BasicStroke.CAP_BUTT</code>, <code>BasicStroke.CAP_ROUND</code>, or
 *     <code>BasicStroke.CAP_SQUARE</code>.
 */
public int getCap()
{
    return(cap);
}

/**
 * Sets the <code>Cap</code> property, which dictates the cap style to use at
 * the ends of a line.
 *
 * @param cap
 */
public void setCap(int cap) {
	if (this.cap == cap) {
		return;
	}
	this.cap = cap;

	setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_LEGEND_REDRAW);
}

/**
 * Returns the dash array.
 * @return the current dash array
 */
public float[] getDashArray()
{
    return (dashArray);
}

/**
 * Returns the legend dash array.
 * @return the current legend dash array
 */
public float[] getLegendDashArray()
{
    return(legendDashArray);
}

}
