/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCMultiFieldString.java,v $ $Revision: 1.2 $
// $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.*;
import java.util.List;
import java.util.Vector;

/**
 * Encapsulates the notion of a multi-field String, where each field may
 * have a different alignment of text within its field.  Used by the
 * <code>JCChart3dLegend</code> class to manage specialized legend text.
 */
public class JCMultiFieldString implements java.io.Serializable {

public final static Integer LEFT_ALIGNMENT = new Integer(0);
public final static Integer CENTER_ALIGNMENT = new Integer(1);
public final static Integer RIGHT_ALIGNMENT = new Integer(2);

protected Vector contentList = new Vector();

/**
 * Empty constructor.
 */
public JCMultiFieldString() {}

/**
 * Constructor with input String.  The input String is parsed to see if
 * it contains any field alignment codes (\<i>c</i> for center alignment,
 * \<i>r</i> for right alignment, and \<i>l</i> for left alignment).
 *
 * @param inputString the input String
 */
public JCMultiFieldString(String inputString)
{
	setContents(inputString);
}

/**
 * Sets the contents of the multi field String.  The input String is parsed to
 * see if it contains any field alignment codes (\<i>c</i> for center alignment,
 * \<i>r</i> for right alignment, and \<i>l</i> for left alignment).
 *
 * @param inputString the input String
 */
public void setContents(String inputString)
{
	contentList = parseEncodedString(inputString);
}

/**
 * Gets the multi field String as a list of alignment items and Strings.
 * (Each field should be represented in the list as one alignment item
 * and one String.)
 */
public List getContentList()
{
	return(contentList);
}

/**
 * Parses a String encoded in the <code>JCMultiFieldString</code> format into a
 * series of field alignment attributes and field Strings.
 *
 * @param inputString the String to parse
 * @return a Vector of Strings and field alignment items
 *     (<code>RIGHT_ALIGNMENT</code>, <code>LEFT_ALIGNMENT</code>, or
 *     <code>CENTER_ALIGNMENT</code>)
 */
public Vector parseEncodedString(String inputString)
{
	Vector  v        = new Vector();
	int     pos      = 0;
	int     slashPos = pos;

	// Run through input string looking for slashes
    while ((slashPos = inputString.indexOf('\\', slashPos)) != -1) {
        if (slashPos + 1 < inputString.length()) {
            char c = inputString.charAt(slashPos + 1);
            // Dpecial alignment character found
            if (c == 'l' || c == 'r' || c == 'c') {
                // First, add the previous string to the list of fields
                if (pos != slashPos) {
                    // No alignment specified?  Default to left.
                    if (v.size() == 0) {
                        v.add(LEFT_ALIGNMENT);
                    }
                    v.add(inputString.substring(pos, slashPos));
                }
                // Second, add the appropriate alignment item to the list
                if (c == 'l') {
                    v.add(LEFT_ALIGNMENT);
                }
                else if (c == 'r') {
                    v.add(RIGHT_ALIGNMENT);
                }
                else if (c == 'c') {
                    v.add(CENTER_ALIGNMENT);
                }
                // move on to next characters in string
                slashPos += 2;
                pos = slashPos;
            }
            // An unrecognized slash, just include it in the string
            else {
                slashPos += 2;
            }
        }
        else {
            // Special case -- single slash at end of string -- just include it
            slashPos++;
        }
    }
    // If there is any remaining characters in the string, add a field for them
    if (pos < inputString.length()) {
        // No alignment specified?  Default to left.
        if (v.size() == 0) {
            v.add(LEFT_ALIGNMENT);
        }
        v.add(inputString.substring(pos));
    }
	return(v);
}

/**
 * Draws the multi-field String.
 *
 * @param gc the <code>Graphics</code> object to write the String to
 * @param useFont the font to write the String with
 * @param x the <i>x</i> position to write the String at
 * @param y the <i>y</i> position to write the String at
 * @param maxFieldWidths an array of maximum field widths that is used to justify
 *     the text in each field
 * @param fieldGap the spacing between fields
 */
public void draw(Graphics gc, Font useFont, int x, int y,
				 int[] maxFieldWidths, int fieldGap)
{
    int fieldx = x;
    FontMetrics fm = gc.getFontMetrics(useFont);
    int ascent = fm.getAscent();
    // cycle through the content list, drawing each field
    for (int i = 1; i < contentList.size(); i += 2) {
        // DAM check for HTML text???
        int posx = fieldx;
        String text = (String)contentList.get(i);
        int width = fm.stringWidth(text);
        Integer align = (Integer)contentList.get(i - 1);
        if (maxFieldWidths != null) {
            if (align.equals(LEFT_ALIGNMENT)) {
            }
            else if (align.equals(RIGHT_ALIGNMENT)) {
                posx += (maxFieldWidths[(i - 1) / 2] - width);
            }
            else if (align.equals(CENTER_ALIGNMENT)) {
                posx += (maxFieldWidths[(i - 1) / 2] - width) / 2;
            }
        }
        gc.drawString((String)contentList.get(i), posx, y + ascent);
        fieldx += maxFieldWidths[(i - 1) / 2];
        fieldx += fieldGap;
    }
}

/**
 * Converts this object back to a String representation of its contents.
 */
public String toString()
{
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < contentList.size(); i++) {
        Object contentItem = contentList.get(i);
        if (contentItem instanceof Integer) {
            Integer alignment = (Integer)contentItem;
            if (alignment.equals(RIGHT_ALIGNMENT)) {
                sb.append("\\r");
            }
            else if (alignment.equals(CENTER_ALIGNMENT)) {
                sb.append("\\c");
            }
            else if (alignment.equals(LEFT_ALIGNMENT)) {
                sb.append("\\l");
            }
        }
        else {
            sb.append((String)contentItem);
        }
    }
    return(sb.toString());
}

}
