/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCPlotCube.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;


import javax.vecmath.Matrix4d;
import javax.vecmath.Point3d;
import javax.vecmath.Point4d;
import javax.vecmath.Vector3d;
import java.awt.*;
import java.awt.image.BufferedImage;

/**
 * The <code>JCPlotCube</code> class deals with drawing the cube around the
 * surface, bar, or scatter plot.
 */
public class JCPlotCube extends ParentTrackChange {

static final long serialVersionUID = -5542417106751358122L;

/**
 * The scale in the <i>x</i> direction.  Measured relative to the other scale
 * values.
 */
protected double xScale = 1.0;

/**
 * The scale in the <i>y</i> direction.  Measured relative to the other scale
 * values.
 */
protected double yScale = 1.0;

/**
 * The scale in the <i>z</i> direction.  Measured relative to the other scale
 * values.
 */
protected double zScale = 1.0;

/** The <code>plotCube</code>'s ceiling projection. */
protected JCProjection ceiling = null;

/** The <code>plotCube</code>'s floor projection. */
protected JCProjection floor = null;

/** The <code>plotCube</code>'s background color. */
protected Color background = null;

/** The <code>plotCube</code>'s foreground color. */
protected Color foreground = null;

// Local variables
protected double xNormalizedScale = 1.0;
protected double yNormalizedScale = 1.0;
protected double zNormalizedScale = 1.0;

/** The parent <code>chart3dArea</code> object. */
protected JCChart3dArea area = null;

// Note that the extra 5 points are for the preview cube
private ScreenPoint[]	pts = new ScreenPoint[Cube.NUM_POINTS + 5];
private Point4d		    pt = new Point4d();

/**
 * No args constructor. <code>Chart3d</code> area is null. Other values at their default.
 */
public JCPlotCube()
{
}

/**
 * Contstructor given a parent <code>chart3d</code> area. All other values at default.
 *
 * @param area the parent <code>chart3d</code> area
 */
public JCPlotCube(JCChart3dArea area)
{
    this.area = area;
	addParent(area);
}

/**
 * Contstructor given a parent <code>chart3d</code> area, the <i>x</i>/<i>y</i>/<i>z</i>
 * scale values, and the background and foreground colors of the cube.
 *
 * @param area the parent <code>chart3d</code> area
 * @param xScale the scale in the <i>x</i> direction
 * @param yScale the scale in the<i> y</i> direction
 * @param zScale the scale in the <i>z</i> direction
 * @param background the plotcube's background (if null, the area's
 *     background is used)
 * @param foreground the plotcube's foreground (if null, the area's
 *     foreground is used)
 */
JCPlotCube(JCChart3dArea area, double xScale, double yScale, double zScale,
           Color background, Color foreground)
{
    this(area);
    this.xScale     = xScale;
    this.yScale     = yScale;
    this.zScale     = zScale;
    this.background = background;
    this.foreground = foreground;
    normalizeScales();
}

/**
 * Gets the parent <code>chart3d</code> area.
 *
 * @return the parent <code>JCChart3dArea</code> object
 */
public JCChart3dArea getChart3dArea()
{
	return(area);
}

/**
 * Sets the parent <code>chart3d</code> area. There can be only one <code>chart3d</code> area parent since
 * the plotcube is heavily dependent on things in the chart area (for example, the axes,
 * the transformation, the cube object, and so on).
 *
 * @param area the new parent <code>chart3dArea</code>
 */
public void setChart3dArea(JCChart3dArea area)
{
	if (this.area == area) {
		return;
	}
	this.area = area;
	clearParents();
	addParent(area);
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Sets the <code>PlotCube</code>'s <i>x</i> scale (default value is 1.0).
 *
 * @param xScale the new <i>x</i> scale
 */
public void setXScale(double xScale)
{
	if (this.xScale == xScale) {
		return;
	}
    if (xScale > 0.0) {
        this.xScale = xScale;
        normalizeScales();
	    callParentSetChanged(true, Changeable.RECALC);
    }
	else {
		throw new IllegalArgumentException(
		    LocaleBundle.string(LocaleBundle.INVALID_VALUE_NEGATIVE_ERROR));
	}
}

/**
 * Returns the current value of the <code>PlotCube</code>'s <i>x</i> scale.
 *
 * @return the <code>PlotCube</code>'s <i>x</i> scale
 */
public double getXScale()
{
    return(xScale);
}

/**
 * Returns the current value of the <code>PlotCube</code>'s normalized
 * <i>x</i> scale.
 *
 * @return the <code>PlotCube</code>'s normalized <i>x</i> scale
 */
public double getXNormalizedScale()
{
    return(xNormalizedScale);
}

/**
 * Sets the <code>PlotCube</code>'s <i>y</i> scale (default value is 1.0).
 *
 * @param yScale the new <i>y</i> scale
 */
public void setYScale(double yScale)
{
	if (this.yScale == yScale) {
		return;
	}
    if (yScale > 0.0) {
        this.yScale = yScale;
        normalizeScales();
	    callParentSetChanged(true, Changeable.RECALC);
    }
	else {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NEGATIVE_ERROR));
	}
}

/**
 * Returns the current value of the <code>PlotCube</code>'s <i>y</i> scale.
 *
 * @return the <code>PlotCube</code>'s <i>y</i> scale
 */
public double getYScale()
{
    return(yScale);
}

/**
 * Returns the current value of the <code>PlotCube</code>'s normalized
 * <i>y</i> scale.
 *
 * @return the <code>PlotCube</code>'s normalized <i>y</i> scale
 */
public double getYNormalizedScale()
{
    return(yNormalizedScale);
}

/**
 * Sets the <code>PlotCube</code>'s <i>z</i> scale (default value is 1.0).
 *
 * @param zScale the new <i>z</i> scale
 */
public void setZScale(double zScale)
{
	if (this.zScale == zScale) {
		return;
	}
    if (zScale > 0.0) {
        this.zScale = zScale;
        normalizeScales();
	    callParentSetChanged(true, Changeable.RECALC);
    }
	else {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NEGATIVE_ERROR));
	}
}

/**
 * Returns the current value of the <code>PlotCube</code>'s <i>z</i> scale.
 *
 * @return the <code>PlotCube</code>'s <i>z</i> scale
 */
public double getZScale()
{
    return(zScale);
}

/**
 * Returns the current value of the <code>PlotCube</code>'s normalized
 * <i>z</i> scale.
 *
 * @return the <code>PlotCube</code>'s normalized <i>z</i> scale
 */
public double getZNormalizedScale()
{
    return(zNormalizedScale);
}

/**
 * Normalizes the three scale values between 0 and 1.  Needed so that we always
 * know that the plotcube is between -1 and 1.
 */
protected void normalizeScales()
{
	boolean isProj = (area == null) ? false : area.getChart3d().isProjection();
    double  max  = Math.max(xScale, yScale);
	if (!isProj) {
       max = Math.max(max, zScale);
	}
    xNormalizedScale = xScale / max;
    yNormalizedScale = yScale / max;
    zNormalizedScale = isProj ? 1.0 : zScale / max;
}

/**
 * Sets the <code>plotCube</code> background.  This value is null by default,
 * which means that the <code>plotCube</code> color is the same as the parent
 * chart area's background, or seemingly transparent.
 *
 * @param background the new <code>plotCube</code> background
 */
public void setBackground(Color background)
{
	if (this.background == background) {
		return;
	}
    this.background = background;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Gets the <code>plotCube</code> background.  This value is null by default,
 * which means that the <code>plotCube</code> color is the same as the parent
 * chart area's background, or seemingly transparent.
 *
 * @return the current <code>plotCube</code> background
 */
public Color getBackground()
{
    return(background);
}

/**
 * Sets the <code>plotCube</code> foreground.  This value is null by default,
 * which means that the <code>plotCube</code> color is the same as the parent
 * chart area's foreground, or seemingly transparent.
 *
 * @param foreground the new <code>plotCube</code> foreground
 */
public void setForeground(Color foreground)
{
	if (this.foreground == foreground) {
		return;
	}
    this.foreground = foreground;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Gets the <code>plotCube</code> foreground.  This value is null by default,
 * which means that the <code>plotCube</code> color is the same as the parent
 * chart area's foreground, or seemingly transparent.
 *
 * @return the current <code>plotCube</code> foreground
 */
public Color getForeground()
{
    return(foreground);
}

/**
 * Returns the actual background color drawn in the <code>plotCube</code>.  If
 * the user has set a <code>plotCube</code> background color, use it.
 * Otherwise, use the parent <code>JCChart3dArea</code>'s background color.
 *
 * @return the actual plotcube background color
 */
public Color getDrawBackground()
{
	Color drawBackground = background;
	if (drawBackground == null && area != null) {
		drawBackground = area.getBackground();
	}
	return(drawBackground);
}

/**
 * Returns the foreground color used to draw items within the
 * <code>plotCube</code>.  Such items include the <code>plotCube</code> lines
 * and the default mesh line color.  If the user has set a <code>plotCube</code>
 * foreground color, use it.  Otherwise, use the parent
 * <code>JCChart3dArea</code>'s foreground color.
 *
 * @return the actual plotcube foreground color
 */
public Color getDrawForeground()
{
	Color drawForeground = foreground;
	if (drawForeground == null && area != null) {
		drawForeground = area.getForeground();
	}
	return(drawForeground);
}

/**
 * Sets the <code>PlotCube</code>'s ceiling projection (null parameters are not
 * allowed).
 *
 * @param ceiling the new projection for the ceiling of the cube
 */
public void setCeiling(JCProjection ceiling)
{
	if (this.ceiling == ceiling) {
		return;
	}
    if (ceiling != null) {
        this.ceiling = ceiling;
		this.ceiling.parent = this;
	    callParentSetChanged(true, Changeable.REDRAW);
    }
	else {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NULL_ERROR));
	}
}

/**
 * Returns the <code>PlotCube</code>'s ceiling projection.
 *
 * @return the <code>PlotCube</code>'s ceiling projection
 */
public JCProjection getCeiling()
{
    return(ceiling);
}

/**
 * Sets the <code>PlotCube</code>'s floor projection (null parameters are not
 * allowed).
 *
 * @param floor the new projection for the floor of the cube
 */
public void setFloor(JCProjection floor)
{
	if (this.floor == floor) {
		return;
	}
    if (floor != null) {
        this.floor = floor;
		this.floor.parent = this;
	    callParentSetChanged(true, Changeable.REDRAW);
    }
	else {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_VALUE_NULL_ERROR));
	}
}

/**
 * Returns the <code>PlotCube</code>'s floor projection.
 *
 * @return the <code>PlotCube</code>'s floor projection
 */
public JCProjection getFloor()
{
    return(floor);
}

/**
 * Asks whether this <code>PlotCube</code> has a ceiling projection.
 * @return
 */
public boolean hasCeilingProjection()
{
    return(ceiling != null && (ceiling.isContoured() || ceiling.isZoned()));
}

/**
 * Asks if this <code>PlotCube</code> has a floor projection.
 * @return
 */
public boolean hasFloorProjection()
{
    return(floor != null && (floor.isContoured() || floor.isZoned()));
}

/**
 * Asks if this <code>PlotCube</code> has any projections.
 * @return
 */
public boolean hasProjections()
{
    return(hasCeilingProjection() || hasFloorProjection());
}

/**
 * Draws the bounding box for the plotcube (either 2d or 3d).
 * @param gc
 * @param drawBackground
 */
protected void drawBoundingBox(Graphics gc, boolean drawBackground)
{
	if (area == null || !area.getChart3d().isProjection()) {
		if (drawBackground) {
		   draw3dBoundingBox(gc);
		}
	}
	else {
		draw2dBoundingBox(gc, drawBackground);
	}
}

/**
 * Sets the <i>i</i>th <code>ScreenPoint</code> to the transformed value of the
 * point (<i>x</i>,<i>y</i>,<i>z</i>).
 *
 * @param i the index of the screen point
 * @param ctm the current transformation matrix
 * @param x the <i>x</i> coordinate of the point
 * @param y the <i>y</i> coordinate of the point
 * @param z the <i>z</i> coordinate of the point
 */
protected void setPoint(int i, Matrix4d ctm, double x, double y, double z)
{
    pt.set(x, y, z, 1.0);
    Transform.transformPoint(ctm, pt);

	
    if (pts[i] == null) {
        pts[i] = new ScreenPoint();
	}
	pts[i].setXPixel((int)Math.round(pt.x));
	pts[i].setYPixel((int)Math.round(pt.y));
}

/**
 * Draws a line from point <i>i</i> to <i>j</i>.  <br><B>Note:</b> This method assumes
 * the line style has been set and that the <code>plotcube</code>
 * <code>ScreenPoint</code> array "pts" has been calculated for points
 * <i>i</i> and <i>j</i>.
 *
 * @param gc the graphics context with which to draw
 * @param i the index of the first point
 * @param j the index of the second point
 */
protected void drawLine(Graphics gc, int i, int j)
{
	if (pts[i] == null) {
		return;
	}
    int x1 = pts[i].getXPixel();
    int y1 = pts[i].getYPixel();

	if (pts[j] == null) {
		return;
	}
    int x2 = pts[j].getXPixel();
    int y2 = pts[j].getYPixel();

    gc.drawLine(x1, y1, x2, y2);
}

/**
 * Draws edge <i>e</i>. <br><b>Note:</b> This method assumes the line style has been
 * set and that the <code>plotcube</code> <code>ScreenPoint</code> array "pts"
 * has been calculated.
 *
 * @param gc the graphics context with which to draw
 * @param e the edge number to draw
 */
protected void drawEdge(Graphics gc, int e)
{
	drawLine(gc, Cube.edges[e][0], Cube.edges[e][1]);
}

/**
 * Draws the 3d bounding box.  This method computes the <code>plotcube</code>
 * <code>ScreenPoint</code> array "pts".
 *
 * @param gc the graphics context with which to draw
 */
protected void draw3dBoundingBox(Graphics gc)
{
	if (area == null) {
		return;
	}

    int             face_pts_x[] = new int[Cube.NUM_POINTS_IN_FACE];
    int             face_pts_y[] = new int[Cube.NUM_POINTS_IN_FACE];
    Transform       transform    = area.getTransform();

	// ctm = [sm][pm][rm][scales]
    Matrix4d ctm = new Matrix4d();
    ctm.mul(transform.getScreen(), transform.getPerspective());
    ctm.mul(transform.getRotation());
    Transform.scaleMatrix(ctm, getXNormalizedScale(), getYNormalizedScale(),
                          getZNormalizedScale());

	for (int i = 0; i < Cube.NUM_POINTS; i++) {
		setPoint(i, ctm, Cube.points[i][0], Cube.points[i][1],
                 Cube.points[i][2]);
	}

	// Make sure the fill style is solid and line pattern is a solid line
	JCFillStyle fillStyle = new JCFillStyle(Color.white, JCFillStyle.SOLID);
	fillStyle.resetGraphics(gc);
	JCLineStyle lineStyle = new JCLineStyle(1, Color.black, JCLineStyle.SOLID);
	lineStyle.resetGraphics(gc);

	// Draw the plot cube background if user has specified a value
    Cube cube = area.getCube();
	if (background != null) {
        gc.setColor(background);
		for (int i = 0; i < Cube.NUM_FACES; i++) {
			if (cube.faceVisible[i]) {
				for (int j = 0; j < Cube.NUM_POINTS_IN_FACE; j++) {
                    int point = Cube.faces[i][j];
					face_pts_x[j] = pts[point].getXPixel();
					face_pts_y[j] = pts[point].getYPixel();
				}
                gc.fillPolygon(face_pts_x, face_pts_y, Cube.NUM_POINTS_IN_FACE);
			}
		}
	}

	// Draw each edge of the bounding box
    gc.setColor(getDrawForeground());
	for (int i = 0; i < Cube.NUM_EDGES; i++) {
		if (cube.drawEdge[i]) {
		    drawEdge(gc, i);
		}
	}
}

/**
 * Draws the rest of the cube edges that haven't been drawn before.  This is
 * done when projections have been drawn.  This method assumes that the
 * <code>plotcube</code> <code>ScreenPoint</code> array "pts" has been
 * calculated.
 *
 * @param gc the graphics context used to draw
 */
protected void finishEdges(Graphics gc)
{
	if (area == null) {
		return;
	}

	boolean     drawEdge = false;
    Cube        cube = area.getCube();

	JCLineStyle lineStyle = new JCLineStyle(1, Color.black, JCLineStyle.SOLID);
	lineStyle.resetGraphics(gc);

    gc.setColor(getDrawForeground());
	for (int i = 0; i < Cube.NUM_EDGES; i++) {
		if (!cube.drawEdge[i]) {
			switch(i/4) {
			case 0:		// X anno edges
			    drawEdge = area.getAxis(JCAxis.AXIS_X).isShowing();
				break;
			case 1:		// Y anno edges
			    drawEdge = area.getAxis(JCAxis.AXIS_Y).isShowing();
				break;
			case 2:		// Z anno edges
			    drawEdge = area.getAxis(JCAxis.AXIS_Z).isShowing();
				break;
			}
			if (drawEdge) {
				drawEdge(gc, i);
			}
		}
	}
}

/**
 * Draws the 2d bounding box.  It is assumed that this routine first gets
 * called with </code>drawBackground = true</code> (to compute the
 * transformation and screen position of the bounding box and possibly the
 * background) and then with </code>drawBackground = false</code>.  This method
 * computes the <code>plotcube</code> <code>ScreenPoint </code> array "pts" the
 * first time through.
 *
 * @param gc the graphics context with which to draw
 * @param drawBackground whether to draw the background
 */
protected void draw2dBoundingBox(Graphics gc, boolean drawBackground)
{
	if (area == null) {
		return;
	}
    Transform transform = area.getTransform();

	// ctm = [sm][rm][scales]
	// No need to do this twice
	if (drawBackground) {
		Matrix4d ctm = new Matrix4d();
        ctm.mul(transform.getScreen(), transform.getRotation());
        Transform.scaleMatrix(ctm, getXNormalizedScale(), getYNormalizedScale(),
		                      getZNormalizedScale());
		for (int i = 0; i < Cube.NUM_POINTS_IN_FACE; i++) {
			int j = 2 * i;
		    setPoint(i, ctm, Cube.points[j][0], Cube.points[j][1], 0.0);
		}
		// Last point is the same as the first
		setPoint(Cube.NUM_POINTS_IN_FACE, ctm, Cube.points[0][0],
		         Cube.points[0][1], 0.0);
	}

	// Make sure the fill style is solid and line pattern is a solid line
	JCFillStyle fillStyle = new JCFillStyle(Color.white, JCFillStyle.SOLID);
	fillStyle.resetGraphics(gc);
	JCLineStyle lineStyle = new JCLineStyle(1, Color.black, JCLineStyle.SOLID);
	lineStyle.resetGraphics(gc);

	// Draw the plot cube background if user has specified a value
	if (drawBackground && background != null) {
        gc.setColor(background);
        int face_pts_x[] = new int[Cube.NUM_POINTS_IN_FACE];
        int face_pts_y[] = new int[Cube.NUM_POINTS_IN_FACE];
		for (int i = 0; i < Cube.NUM_POINTS_IN_FACE; i++) {
			face_pts_x[i] = pts[i].getXPixel();
			face_pts_y[i] = pts[i].getYPixel();
		}
        gc.fillPolygon(face_pts_x, face_pts_y, Cube.NUM_POINTS_IN_FACE);
	}

	// Draw each edge of the bounding box
	if (!drawBackground) {
        gc.setColor(getDrawForeground());
		for (int i = 0; i < Cube.NUM_POINTS_IN_FACE; i++) {
            gc.drawLine(pts[i].getXPixel(), pts[i].getYPixel(),
			            pts[i+1].getXPixel(), pts[i+1].getYPixel());
		}
	}
}

/**
 * Erases and then draws the bounding box preview cube.
 *
 * @param gc  the graphics context with which to draw
 * @param axisId the id of the axis about which the preview cube is rotating
 * @param x the <i>x</i> coordinate of the eye point for <code>JCAxis.AXIS_EYE</code>
 *     rotation
 * @param y the <i>y</i> coordinate of the eye point for <code>JCAxis.AXIS_EYE</code>
 *     rotation
 */
protected void updateBoundingBox(Graphics gc, int axisId, int x, int y)
{
	if (area == null) {
		return;
	}

	// Erase any previous bounding box by drawing over it with the background
	// color of the chart3darea
	JCLineStyle lineStyleSolid  = new JCLineStyle(1, area.getBackground(),
												  JCLineStyle.SOLID);
	JCLineStyle lineStyleDashed = new JCLineStyle(1, area.getBackground(),
												  JCLineStyle.SHORT_DASH);

	Cube cube = area.getCube();
	for (int i = 0; i < Cube.NUM_EDGES; i++) {
		if (cube.drawEdge[i] || area.getChart3d().isProjection()) {
		    lineStyleSolid.updateGraphics(gc);
		}
		else {
		    lineStyleDashed.updateGraphics(gc);
		}
		drawEdge(gc, i);
	}

	lineStyleSolid.updateGraphics(gc);
	drawLine(gc, Cube.NUM_POINTS, Cube.NUM_POINTS + 1);
	drawLine(gc, Cube.NUM_POINTS, Cube.NUM_POINTS + 2);
	drawLine(gc, Cube.NUM_POINTS + 3, Cube.NUM_POINTS + 4);

    Transform transform = area.getTransform();

	// ctm = [sm][pm][rm][scales]
    Matrix4d ctm = new Matrix4d();
    ctm.mul(transform.getScreen(), transform.getPerspective());
    ctm.mul(transform.getRotation());
    Transform.scaleMatrix(ctm, getXNormalizedScale(), getYNormalizedScale(),
                          getZNormalizedScale());

	for (int i = 0; i < Cube.NUM_POINTS; i++) {
		setPoint(i, ctm, Cube.points[i][0], Cube.points[i][1],
                 Cube.points[i][2]);
	}

	setPoint(Cube.NUM_POINTS,     ctm, -0.8, -0.8, -1.0);
	setPoint(Cube.NUM_POINTS + 1, ctm, -0.8, -1.0, -1.0);
	setPoint(Cube.NUM_POINTS + 2, ctm, -1.0, -0.8, -1.0);

	switch(axisId) {
	case JCAxis.AXIS_X:
	    setPoint(Cube.NUM_POINTS + 3, ctm, -1.5, 0.0, 0.0);
	    setPoint(Cube.NUM_POINTS + 4, ctm,  1.5, 0.0, 0.0);
	    break;
	case JCAxis.AXIS_Y:
	    setPoint(Cube.NUM_POINTS + 3, ctm, 0.0, -1.5, 0.0);
	    setPoint(Cube.NUM_POINTS + 4, ctm, 0.0,  1.5, 0.0);
	    break;
	case JCAxis.AXIS_Z:
	    setPoint(Cube.NUM_POINTS + 3, ctm, 0.0, 0.0, -1.5);
	    setPoint(Cube.NUM_POINTS + 4, ctm, 0.0, 0.0,  1.5);
	    break;
	case JCAxis.AXIS_EYE:
	    setPoint(Cube.NUM_POINTS + 3, ctm, 0.0, 0.0, 0.0);
		if (pts[Cube.NUM_POINTS + 4] == null) {
			pts[Cube.NUM_POINTS + 4] = new ScreenPoint();
		}
		pts[Cube.NUM_POINTS + 4].setXPixel(x);
		pts[Cube.NUM_POINTS + 4].setYPixel(y);
	    break;
	case JCAxis.AXIS_NONE:
	default:
	    pts[Cube.NUM_POINTS + 3] = null;
	    pts[Cube.NUM_POINTS + 4] = null;
	    break;
	}

	lineStyleSolid.setColor(area.getForeground());
	lineStyleDashed.setColor(area.getForeground());

	for (int i = 0; i < Cube.NUM_EDGES; i++) {
		if (cube.drawEdge[i] || area.getChart3d().isProjection()) {
		    lineStyleSolid.updateGraphics(gc);
		}
		else {
		    lineStyleDashed.updateGraphics(gc);
		}
		drawEdge(gc, i);
	}

	lineStyleSolid.updateGraphics(gc);
	drawLine(gc, Cube.NUM_POINTS, Cube.NUM_POINTS + 1);
	drawLine(gc, Cube.NUM_POINTS, Cube.NUM_POINTS + 2);
	drawLine(gc, Cube.NUM_POINTS + 3, Cube.NUM_POINTS + 4);
}

/**
 * Creates a parallel projection of the chart (without the annotation) and
 * stores it in a passed in image, filling the entire image with the projection.
 * If there is any gap between the data and the dataport, the plot cube
 * backgound will be used.
 *
 * @param dest the image to store the parallel projection in
 * @param projection the <code>JCProjection</code> object that tells whether
 *     the projection is contoured and/or zoned
 */
public void createParallelProjection(BufferedImage dest,
                                     JCProjection projection)
{
	createParallelProjection(dest, projection, getDrawBackground());
}

/**
 * Creates a parallel projection of the chart (without the annotation) and
 * stores it in a passed in image, filling the entire image with the projection.
 * If there is any gap between the data and the dataport, the specified
 * background color will be used.
 *
 * @param dest the image to store the parallel projection in
 * @param projection the <code>JCProjection</code> object that tells whether
 *     the projection is contoured and/or zoned
 * @param background
 */
public void createParallelProjection(BufferedImage dest,
                                     JCProjection projection, Color background)
{
	createParallelProjection(dest, projection, 0, 0, dest.getWidth(),
                             dest.getHeight(), background);
}


/**
 * Creates a parallel projection of the chart (without the annotation) and
 * stores it in a passed in image, at a given position, and with a given width
 * and height within the image.
 * <p>
 * If there is any gap between the data and the dataport, the specified
 * background color will be used.
 *
 * @param dest the image to store the parallel projection in
 * @param projection the <code>JCProjection</code> object that tells whether
 *     the projection is contoured and/or zoned
 * @param x the <i>x</i> position at which to place the projection in the
 *     image
 * @param y the <i>y</i> position at which to place the projection in the
 *     image
 * @param width the width (within the image) of the projection
 * @param height the height (within the image) of the projection
 * @param background
 */
public void createParallelProjection(BufferedImage dest,
									 JCProjection projection,
									 int x,
									 int y,
									 int width,
									 int height,
									 Color background)
{
	// Construct the transformation matrix which for a parallel projection of
	// the data surface onto the given image.  The first step is to construct
	// a pseudo absolute matrix which will map the data to a unit square (the
	// depth values are ignored).  The perspective matrix is not used - it's
	// a parallel projection we're using.  The rotation matrix is the identity
	// matrix, so it is ignored also.  The screen is renamed the image matrix,
	// and is computed so that the unit square maps perfectly onto it.

	Chart3dDataView dataView = area.getChart3d().getDataView(0);
	JCContour		contour = dataView.getContour();
	Chart3dGridData	data = (Chart3dGridData)dataView.getElevationData();

	JCAxis 			xAxis = area.getAxis(JCAxis.AXIS_X);
	JCAxis 			yAxis = area.getAxis(JCAxis.AXIS_Y);

    double 			dataX = xAxis.getMin();
    double 			dataY = yAxis.getMin();
    double	 		dataW = xAxis.getMax() - xAxis.getMin();

    double 			annoX = xAxis.getDataportMin();
    double 			annoY = yAxis.getDataportMin();
    double	 		annoW = xAxis.getDataportMax() - xAxis.getDataportMin();
    double 			annoH = yAxis.getDataportMax() - yAxis.getDataportMin();

	double			sx = 2.0 / annoW;
	double			sy = 2.0 / annoH;
	double			dx = -1.0 -annoX * sx;
	double			dy = -1.0 -annoY * sy;

	Matrix4d abs = new Matrix4d();
	abs.setIdentity();
	abs.m00 = sx; // x scale
	abs.m03 = dx; // x translation
	abs.m11 = sy; // y scale
	abs.m13 = dy; // y translation
	abs.m22 = 1;  // z scale

	Matrix4d image = new Matrix4d();
	image.setIdentity();
	image.m00 = width / 2;
	image.m03 = x + width / 2;
	image.m11 = -height / 2;
	image.m13 = height / 2 - y;

	Matrix4d view = new Matrix4d();
	view.mul(image, abs);

	// Set the absolute and view matrices.
	area.getTransform().absolute = abs;
	area.getTransform().view = view;
	area.getTransform().eyePosition = new Point4d(0, 0, 1, 1);

	// Before we set the projection type, store the current chart type.
	boolean wasContoured = contour.isContoured();
	boolean wasZoned = contour.isZoned();

	// Now set them.
	contour.setContouredNoUpdate(projection.isContoured());
	contour.setZonedNoUpdate(projection.isZoned());

	data.precomputePoints();
	data.calcAnnoLimits();
	contour.precompContours();

	// Create a graphics object to draw on the image.
	Graphics2D g = dest.createGraphics();

	// Clear the background.
	if (background == null) {
		background = Color.white;
	}
	g.setBackground(background);
	g.clearRect(x, y, width, height);

	double
		clipX = ((dataX - annoX) / annoW) * width,
		clipW = clipX + ((dataW / annoW) * width),
		clipY = 0.,
		clipH = (annoH - dataY + annoY) * (height / annoH);

	g.setClip((int)Math.rint(clipX),
			  (int)Math.rint(clipY),
			  (int)Math.rint(clipW),
			  (int)Math.rint(clipH));

	contour.draw3dZones(g);
	contour.drawContours(g);

	// Then set it free.
	area.getTransform().computeTransformations(g);
	g.dispose();

	// Now restore previous state.
	contour.setZonedNoUpdate(wasZoned);
	contour.setContouredNoUpdate(wasContoured);

	// Recompute screen points and contours??? Set points to null???
	data.precomputePoints();
	contour.precompContours();
}

/**
 * Draws the contoured and/or zoned surface projection on the floor or ceiling
 * of the the plot cube.  The idea is to alter the absolute tranformation so
 * that it maps onto the bottom or top of the cube, reconstruct the view
 * matrix, transform the data points and contour points, draw, set
 * the transformation back to what it was, and transform again.
 *
 * @param gc the graphics context with which to draw
 * @param isFloor if <code>true</code>, it is a floor projection; otherwise, it is a ceiling
 *     projection
 */
protected void drawProjection(Graphics gc, boolean isFloor)
{
	if (area == null) {
		return;
	}

	JCProjection    projection = isFloor ? floor : ceiling;
	Chart3dDataView dataView   = area.getChart3d().getDataView(0);
	JCContour       contour    = dataView.getContour();
	Chart3dGridData data       = (Chart3dGridData)dataView.getElevationData();
	Transform       transform  = area.getTransform();
	Matrix4d        absolute   = transform.getAbsolute();

	if (!dataView.isDataDrawn()) {
		return;
	}

	// Save real values
	double sz         = absolute.m22;
	double dz         = absolute.m23;
	boolean contoured = contour.isContoured();
	boolean zoned     = contour.isZoned();

	// Substitute fake ones
	absolute.m22 = 0.0;
	absolute.m23 = isFloor ? -zNormalizedScale : zNormalizedScale;
	contour.setContouredNoUpdate(projection.isContoured());
	contour.setZonedNoUpdate(projection.isZoned());

	// Reconstruct view matrix
	transform.computeViewMatrix();

	// Precompute and draw
	data.precomputePoints();
	contour.precompContours();
	contour.draw3dZones(gc);
	contour.drawContours(gc);

	// Put everything back
	absolute.m22 = sz;
	absolute.m23 = dz;
	contour.setContouredNoUpdate(contoured);
	contour.setZonedNoUpdate(zoned);

	// Reconstruct view matrix
	transform.computeViewMatrix();

	// Recalculate screen points and contour projections
	data.precomputePoints();
	contour.precompContours();
}

/**
 * This method computes and draws the <i>X</i>, <i>Y</i>, and <i>Z</i> gridlines.
 * It checks to see which faces are visible (precomputed during
 * annotation phase and stored in the <code>Cube</code> object) and draws gridlines
 * on the faces which the user has specifically requested.
 *
 * @param gc the graphics context with which to draw
 */
protected void drawGridLines(Graphics gc)
{
	if (area == null) {
		return;
	}

	JCAxis          xAxis     = area.getAxis(JCAxis.AXIS_X);
	JCAxis          yAxis     = area.getAxis(JCAxis.AXIS_Y);
	JCAxis          zAxis     = area.getAxis(JCAxis.AXIS_Z);
    Transform       transform = area.getTransform();
	Cube            cube      = area.getCube();

	// Is there anything to draw?
	boolean xGridFlag = (xAxis.getGridLines().getPlaneMask() > 0);
	boolean yGridFlag = (yAxis.getGridLines().getPlaneMask() > 0);
	boolean zGridFlag = (zAxis.getGridLines().getPlaneMask() > 0);
	boolean gridFlag  = xGridFlag || yGridFlag;
	if (!area.getChart3d().isProjection()) {
	    gridFlag = gridFlag || zGridFlag;
	}
	if (!gridFlag) {
		return;
	}

    Matrix4d ctm = new Matrix4d();
	if (area.getChart3d().isProjection()) {
	    // ctm = [sm][rm][scales]
        ctm.mul(transform.getScreen(), transform.getRotation());
	}
	else {
	    // ctm = [sm][pm][rm][scales]
        ctm.mul(transform.getScreen(), transform.getPerspective());
        ctm.mul(transform.getRotation());
	}
    Transform.scaleMatrix(ctm, getXNormalizedScale(), getYNormalizedScale(),
		                  getZNormalizedScale());

	Vector3d[] points = new Vector3d[3];
	for (int i = 0; i < 3; i++) {
		points[i] = new Vector3d();
	}
	Point3d facePoint = new Point3d();      // A representative point on a face
	for (int face = Cube.FACE_LEFT; face <= Cube.FACE_NEAR; face++) {
		if (!cube.faceVisible[face]) {
			continue;
		}

		// Determine which two of the 3 directions make sense for
	    // drawing gridlines.  The cross product of two vectors on
		// the face gives the direction which does not make sense.
		for (int i = 0; i < 3; i++) {
		    int pointNum = Cube.faces[face][i];
			double[] point = Cube.points[pointNum];
			points[i].set(point[0], point[1], point[2]);
			if (i == 0) {
				facePoint.set(points[i]);
			}
		}
		points[2].sub(points[1]);
		points[1].sub(points[0]);
		points[0].cross(points[1], points[2]);

		// One of these 3 has to be biggest
		points[0].absolute();
		int faceFlag = 0;
		if (points[0].x > Math.max(points[0].y, points[0].z)) {
			faceFlag = Cube.VP_YMASK | Cube.VP_ZMASK;    // YZ Plane
		}
		else if (points[0].y > Math.max(points[0].x, points[0].z)) {
			faceFlag = Cube.VP_XMASK | Cube.VP_ZMASK;    // XZ Plane
		}
		else {
			faceFlag = Cube.VP_XMASK | Cube.VP_YMASK;    // XY Plane
		}

		// X Grid Lines
		if ((faceFlag & Cube.VP_XMASK) > 0 && xAxis.isShowing() && xGridFlag) {
			xAxis.getGridLines().drawGridFace(gc, ctm, faceFlag, facePoint);
		}
		// Y Grid Lines
		if ((faceFlag & Cube.VP_YMASK) > 0 && yAxis.isShowing() && yGridFlag) {
			yAxis.getGridLines().drawGridFace(gc, ctm, faceFlag, facePoint);
		}
		// Z Grid Lines
		if ((faceFlag & Cube.VP_ZMASK) > 0 && zAxis.isShowing() && zGridFlag) {
			zAxis.getGridLines().drawGridFace(gc, ctm, faceFlag, facePoint);
		}
	}
}

}
