/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCSurface.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import javax.vecmath.Matrix4d;



/**
 * JClass Chart 3D class that contains properties that pertain to plots of
 * type surface.  These are properties such as whether to draw a solid skirt
 * and whether to show or filter mesh lines.
 */
public class JCSurface extends ParentTrackChange {

static final long serialVersionUID = -5542417106751358122L;

/** Pixel spread that determines the space used in automatic mesh filtering. */
public final static double PIXEL_SPREAD = 10.0;

/** Asks if solid "skirts" under the surface should be drawn. */
protected boolean solid = false;

/** Asks if <i>x</i> mesh lines are showing. */
protected boolean xMeshShowing = true;

/** Asks if <i>y</i> mesh lines are showing. */
protected boolean yMeshShowing = true;

/** Filter value for <i>x</i> mesh lines. */
protected int     xMeshFilter = 0;

/** Filter value for <i>y</i> mesh lines. */
protected int     yMeshFilter = 0;

/** Array that determines whether an <i>x</i> mesh line is drawn (based on grid). */
protected boolean[]   xMesh = null;

/** Array that determines whether a <i>y</i> mesh line is drawn (based on grid). */
protected boolean[]   yMesh = null;

/**
 * No args constructor.  Values at default values.
 */
public JCSurface()
{
}

/**
 * Constructor that provides values for solid, <i>x</i> and <i>y</i> showing
 * and filter.
 *
 * @param solid Should solid "skirts" under the surface be drawn?
 * @param xMeshShowing Are <i>x</i> mesh lines showing?
 * @param yMeshShowing Are <i>y</i> mesh lines showing?
 * @param xMeshFilter filter value for <i>x</i> mesh lines
 * @param yMeshFilter filter value for y mesh lines
 */
public JCSurface(boolean solid,
	             boolean xMeshShowing, boolean yMeshShowing,
				 int xMeshFilter, int yMeshFilter)
{
	this.solid        = solid;
	this.xMeshShowing = xMeshShowing;
	this.yMeshShowing = yMeshShowing;
	this.xMeshFilter  = xMeshFilter;
	this.yMeshFilter  = yMeshFilter;
}

/**
 * Asks if solid "skirts" under the surface are drawn.
 *
 * @return Is the surface a solid object?
 */
public boolean isSolid()
{
	return(solid);
}

/**
 * Sets whether solid "skirts" under the surface are drawn.
 *
 * @param solid sets whether the surface is a solid object
 */
public void setSolid(boolean solid)
{
	if (this.solid == solid) {
		return;
	}
	this.solid = solid;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Asks if the <i>x</i> mesh lines are showing.
 *
 * @return Are <i>x</i> mesh lines be drawn?
 */
public boolean isXMeshShowing()
{
	return(xMeshShowing);
}

/**
 * Sets whether <i>x</i> mesh lines are drawn.
 *
 * @param xMeshShowing Are the <i>x</i> mesh lines drawn?
 */
public void setXMeshShowing(boolean xMeshShowing)
{
	if (this.xMeshShowing == xMeshShowing) {
		return;
	}
	this.xMeshShowing = xMeshShowing;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Asks if the <i>y</i> mesh lines are showing.
 *
 * @return Are <i>y</i> mesh lines drawn?
 */
public boolean isYMeshShowing()
{
	return(yMeshShowing);
}

/**
 * Sets whether <i>y</i> mesh lines are drawn.
 *
 * @param yMeshShowing Are the <i>y</i> mesh lines drawn?
 */
public void setYMeshShowing(boolean yMeshShowing)
{
	if (this.yMeshShowing == yMeshShowing) {
		return;
	}
	this.yMeshShowing = yMeshShowing;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Returns the current <i>x</i> mesh filter.  A filter value of <i>n</i> means
 * every <i>n</i>th mesh line is drawn.  If the value is 1 or greater, the
 * filter value is honored.  If the value is zero, a "pleasing" value is
 * calculated so that mesh lines are not drawn too close together.
 *
 * @return the current <i>x</i> mesh filter
 */
public int getXMeshFilter()
{
	return(xMeshFilter);
}

/**
 * Sets the current <i>x</i> mesh filter.  A filter value of <i>n</i> means
 * every <i>n</i>th mesh line is drawn.  If the value is 1 or greater, the
 * filter value is honored.  If the value is zero, a "pleasing" value is
 * calculated so that mesh lines are not drawn too close together.
 *
 * @param xMeshFilter the new <i>x</i> mesh filter
 */
public void setXMeshFilter(int xMeshFilter)
{
	if (this.xMeshFilter == xMeshFilter) {
		return;
	}
	this.xMeshFilter = xMeshFilter;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Returns the current <i>y</i> mesh filter.  A filter value of <i>n</i> means
 * every <i>n</i>th mesh line is drawn.  If the value is 1 or greater, the
 * filter value is honored.  If the value is zero, a "pleasing" value is
 * calculated so that mesh lines are not drawn too close together.
 *
 * @return the current <i>y</i> mesh filter
 */
public int getYMeshFilter()
{
	return(yMeshFilter);
}

/**
 * Sets the current <i>y</i> mesh filter.  A filter value of <i>n</i> means
 * every <i>n</i>th mesh line is drawn.  If the value is 1 or greater, the
 * filter value is honored.  If the value is zero, a "pleasing" value is
 * calculated so that mesh lines are not drawn too close together.
 *
 * @param yMeshFilter the new <i>y</i> mesh filter
 */
public void setYMeshFilter(int yMeshFilter)
{
	if (this.yMeshFilter == yMeshFilter) {
		return;
	}
	this.yMeshFilter = yMeshFilter;
	callParentSetChanged(true, Changeable.REDRAW);
}

/**
 * Returns the current <i>x</i> mesh array, which tells whether a given mesh
 * line is drawn.
 *
 * @return the current <i>x</i> mesh array
 */
public boolean[] getXMesh()
{
	return(xMesh);
}

/**
 * Returns the current <i>y</i> mesh array, which tells whether a given mesh
 * line is drawn.
 *
 * @return the current <i>y</i> mesh array
 */
public boolean[] getYMesh()
{
	return(yMesh);
}

/**
 * Computes mesh filtering for both <i>x</i>s and <i>y</i>s.  If the user specified
 * filter is 1 or greater, it is honored. If it is zero, then a "pleasing"
 * mesh filtering, based on the size of the viewport, is computed.
 *
 * @param dataView the <code>Chart3dDataView</code> whose mesh lines are to be filtered
 */
public void computeMeshFiltering(Chart3dDataView dataView)
{
	int             filter = 0;
	Chart3dData     data = dataView.getElevationData();
	JCChart3dArea   area = dataView.getChart3d().getChart3dArea();

	xMesh = null;
	yMesh = null;
	if (!dataView.getElevation().isMeshed()          ||
	    dataView.getChartType() != JCChart3d.SURFACE ||
	    !(data instanceof Chart3dGridData))
	{
		return;
	}
	Chart3dGridData gridData   = (Chart3dGridData)data;
	GridLimits      gridLimits = gridData.getGridLimits();
	Transform       transform  = area.getTransform();
	JCBox           viewport   = transform.getCurrentViewport();
	Matrix4d        screen     = transform.getScreen();

	int x      = (int) (viewport.x * screen.m00 + screen.m03);
	int y      = (int) ((viewport.y + viewport.height) * screen.m11 +
				         screen.m13);
	int width  = (int) (viewport.width * screen.m00);
	int height = (int) (-viewport.height * screen.m11);

	// X mesh lines are drawn parallel to X axis, so the number is controlled
	// by the number of Y points; similarly with y mesh lines.
	xMesh = new boolean[gridData.getNumY()];
	if (xMeshShowing) {
		filter = xMeshFilter;
		if (filter == 0) {
			filter = (int) (PIXEL_SPREAD /
						    ((double) height / gridLimits.getNumY()));
			filter = Math.max(1, filter);
		}
		if (filter == 1) {
			// All True
			JCChart3dUtil.booleanfill(xMesh, true);
		}
		else {
			for (int i = gridLimits.getYOffset();
				 i <= gridLimits.getYLast();
				 i += filter)
			{
				xMesh[i] = true;
			}
			xMesh[gridLimits.getYLast()] = true;
		}
	}

	yMesh = new boolean[gridData.getNumX()];
	if (yMeshShowing) {
		filter = yMeshFilter;
		if (filter == 0) {
			filter = (int) (PIXEL_SPREAD /
						    ((double) width / gridLimits.getNumX()));
			filter = Math.max(1, filter);
		}
		if (filter == 1) {
			// All True
			JCChart3dUtil.booleanfill(yMesh, true);
		}
		else {
			for (int i = gridLimits.getXOffset();
				 i <= gridLimits.getXLast();
				 i += filter)
			{
				yMesh[i] = true;
			}
			yMesh[gridLimits.getXLast()] = true;
		}
	}
}

}
