/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCValueLabel.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Color;

/**
 * A <code>JCValueLabel</code> is an association between a floating-point axis
 * value and a label.
 */
public class JCValueLabel implements Cloneable, java.io.Serializable {

/** The value on the axis for this label. */
protected double value;

/** The text for this value label. */
protected String label;

/** Asks if the tick drawn is for this value label. */
protected boolean tickDrawn = true;

/** Asks if the label drawn is for this value label. */
protected boolean labelDrawn = true;

/** Only a minor tick is drawn for this label if <code>true</code> (if <code>tickDrawn = true</code>). */
protected boolean minorTick = false;

/** The value label's background color (null means don't draw a background). */
protected Color background = null;

/** The value label's foreground color (null means use plotcube's foreground). */
protected Color foreground = null;

/** The index in the value labels array.  User's may not have write access! */
protected int index = -1;

/**
 * The parent <code>JCAxis</code> who tracks changes in this object.  User's may not have
 * write access!
 */
protected JCAxis axis = null;

/** Allow updates (in other words, repaints) on user changes */
protected boolean allowUpdates = true;

/**
 * Default constructor required for Beans.
 */
public JCValueLabel()
{
	value    = 0.0;
	label    = null;
}

/**
 * Constructor that creates a value label given a value and a String.
 *
 * @param value floating-point axis value
 * @param label String to use as the value label
 */
public JCValueLabel(double value, String label)
{
	this.value = value;
	this.label = label;
}

/**
 * This constructor exists for backwards compatibility only.  It is better to
 * use the constrcutor that sets the <code>labelDrawn</code>, <code>tickDrawn</code>,
 * and <code>minorTick</code> properties explicitly.<p>
 * Constructor that creates a value label given a value, a String, and a boolean
 * <code>tickOnly</code> which, if <code>true</code>, dictates that no label and only a minor
 * tick is to be drawn for this value label (in other words, <code>labelDrawn</code> is set to
 * <code>false</code>). If <code>tickOnly</code> is false, a label with a major tick is drawn.
 *
 * @param value floating-point axis value
 * @param label String to use as the value label
 * @param tickOnly If <code>true</code>, only a minor tick is drawn at this value (if
 *     <code>tickDrawn</code> is <code>true</code>). If <code>false</code>, a label with
 * a major tick is drawn, depending on the values of <code>labelDrawn</code> and <code>tickDrawn</code>.
 */
public JCValueLabel(double value, String label, boolean tickOnly)
{
	this.value      = value;
	this.label      = label;
	this.minorTick  = tickOnly;
	this.labelDrawn = !tickOnly;
}

/**
 * Constructor that creates a value label given a value, a String, and a boolean
 * <code>minorTick</code> that dictates that no label is to be drawn for this
 * value label, only a minor tick (<code>labelDrawn</code> is set to <code>false</code>).
 * To get no label with a major tick, set <code>minorTick</code> to <code>false</code>.
 *
 * @param value floating-point axis value
 * @param label String to use as the value label
 * @param labelDrawn Is the label drawn for this value label?
 * @param tickDrawn Is the tick drawn for this value label?
 * @param minorTick only a minor tick is drawn at this value (if <code>tickDrawn</code>
 * is <code>true</code>)
 */
public JCValueLabel(double value, String label, boolean labelDrawn,
                    boolean tickDrawn, boolean minorTick)
{
	this.value      = value;
	this.label      = label;
	this.labelDrawn = labelDrawn;
	this.tickDrawn  = tickDrawn;
	this.minorTick  = minorTick;
}

/**
 * Returns the axis to which this value label belongs.
 *
 * @return the axis to which this value label belongs
 */
public JCAxis getAxis()
{
	return(axis);
}

/**
 * Sets the axis to which this value label belongs. <br><B>This method should not
 * be called by users.</b> Calling it sets an update relationship between the
 * value label and the axis.  This means that any change to the value label
 * from that point on will cause JClass Chart 3D to repaint (even though
 * the value label may not be drawn).
 *
 * @param axis the new axis to which this value label belongs
 */
protected void setAxis(JCAxis axis)
{
	this.axis = axis;
}

/**
 * Gets the value of the <code>Value</code> property.
 * The <code>Value</code> property controls the position of a label in
 * data space along a particular axis.
 *
 * @return data-space value
 */
public double getValue()
{
	return(value);
}

/**
 * Sets the value of the <code>Value</code> property.
 * The <code>Value</code> property controls the position of a label in
 * data space along a particular axis.
 *
 * @param value data-space value
 */
public void setValue(double value)
{
	if (this.value == value) {
		return;
	}
	this.value = value;
	setChanged(true, Changeable.RECALC);
}

/**
 * Gets the value of the <code>Label</code> property. The <code>Label</code>
 * property specifies the text displayed inside the value label.
 *
 * @return String representing the text inside the label
 */
public String getLabel()
{
	return(label);
}

/**
 * Sets the value of the <code>Label</code> property. The <code>Label</code>
 * property specifies the text displayed inside the value label.
 *
 * @param label String representing the text inside the label
 */
public void setLabel(String label)
{
	if (this.label == label) {
		return;
	}
	this.label = label;
	setChanged(true, Changeable.RECALC);
}

/**
 * Asks if the label drawn is for this value label.
 *
 * @return the current value of <code>labelDrawn</code>
 */
public boolean isLabelDrawn()
{
	return(labelDrawn);
}

/**
 * Sets whether the label for this value label should be drawn.
 *
 * @param labelDrawn the new value of <code>labelDrawn</code>
 */
public void setLabelDrawn(boolean labelDrawn)
{
	if (this.labelDrawn == labelDrawn) {
		return;
	}
	this.labelDrawn = labelDrawn;
	setChanged(true, Changeable.RECALC);
}

/**
 * Asks if the tick drawn is for this value label.
 *
 * @return the current value of <code>tickDrawn</code>
 */
public boolean isTickDrawn()
{
	return(tickDrawn);
}

/**
 * Sets whether the tick for this value label should be drawn.
 *
 * @param tickDrawn the new value of <code>tickDrawn</code>
 */
public void setTickDrawn(boolean tickDrawn)
{
	if (this.tickDrawn == tickDrawn) {
		return;
	}
	this.tickDrawn = tickDrawn;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Asks if a major or minor tick is drawn for this value label.
 *
 * @return The current value of <code>minorTick</code>. Given that the value of
 * <code>tickDrawn</code> is <code>true</code>: if the return value is <code>true</code>,
 * a minor tick is drawn; otherwise a major tick is drawn.
 */
public boolean isMinorTick()
{
	return(minorTick);
}

/**
 * Sets whether a major of minor tick is drawn for this value label.
 *
 * @param minorTick the new value of <code>minorTick</code>
 */
public void setMinorTick(boolean minorTick)
{
	if (this.minorTick == minorTick) {
		return;
	}
	this.minorTick = minorTick;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Returns the background color for this value label.  If null, no background
 * is drawn for this label.
 *
 * @return the current background color for this value label
 */
public Color getBackground()
{
	return(background);
}

/**
 * Sets the background color for this value label.  If null, no background
 * is drawn for this label.
 *
 * @param background the new background color for this value label
 */
public void setBackground(Color background)
{
	if (this.background == background) {
		return;
	}
	this.background = background;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Returns the foreground color for this value label.  If null, the plotcube
 * foreground color is used for this label (if this is null the <code>JCChart3dArea</code>'s
 * foreground color is used).
 *
 * @return the current foreground color for this value label
 */
public Color getForeground()
{
	return(foreground);
}

/**
 * Sets the foreground color for this value label.  If null, the plotcube
 * foreground color is used for this label (if this is null the <code>JCChart3dArea</code>'s
 * foreground color is used).
 *
 * @param foreground the new foreground color for this value label
 */
public void setForeground(Color foreground)
{
	if (this.foreground == foreground) {
		return;
	}
	this.foreground = foreground;
	setChanged(true, Changeable.REDRAW);
}

/**
 * Gets the index into the internal value labels array for this value label.
 *
 * @return the index into the internal value labels array for this value label
 */
public int getIndex()
{
	return(index);
}

/**
 * Sets the index into the internal value labels array for this value label.
 * <br><B>Users should not call this method.</b>
 *
 * @param index the new index into the internal value labels array for this
 *     value label
 */
protected void setIndex(int index)
{
	this.index = index;
}

/**
 * This method exists for backwards compatibility only.  It is better to
 * query the <code>labelDrawn</code>, <code>tickDrawn</code>, and
 * <code>minorTick</code> properties explicitly.<p>
 * If a <code>minorTick</code> with no label is drawn, returns <code>true</code>.
 * Returns <code>false</code> otherwise.
 * @return whether to just draw a minor tick for this value label or not
 * @deprecated As of JClass 6.0.  Use the <code>labelDrawn</code>, <code>tickDrawn</code>,
 * and <code>minorTick</code> properties instead.
 */
public boolean isTickOnly()
{
	return(minorTick && !labelDrawn);
}

/**
 * This method exists for backwards compatibility only.  It is better to
 * set the <code>labelDrawn</code>, <code>tickDrawn</code>, and
 * <code>minorTick</code> properties explicitly.<p>
 * Sets whether to just draw a minor tick for this label or not. If <code>true</code>,
 * a minor tick is drawn with no label.  If <code>false</code>, a major tick is drawn and, if
 * the label is non-null, a label is drawn as well.
 *
 * @param tickOnly <code>true</code> if only a minor tick is to be drawn for this label
 * @deprecated As of JClass 6.0.  Use the <code>labelDrawn</code>, <code>tickDrawn</code>,
 * and <code>minorTick</code> properties instead.
 */
public void setTickOnly(boolean tickOnly)
{
	if (this.minorTick == tickOnly) {
		return;
	}
	this.minorTick  = tickOnly;
	this.labelDrawn = !tickOnly;
	setChanged(true, Changeable.RECALC);
}

/**
 * Tells the parent (if there is one) that a change has happened.
 *
 * @param b Has a change happened?
 * @param cf change mask that determines the type of change
 */
protected void setChanged(boolean b, int cf)
{
	if (axis != null && allowUpdates) {
		axis.callParentSetChanged(b, cf);
	}
}

/**
 * Overrides the <code>toString</code> method.
 * @return
 */
public String toString()
{
	return("value:" + value + " label:" + label);
}

/**
 * Overrides clone in class Object.
 *
 * @see Object#clone().
 * @throws java.lang.OutOfMemoryError - if there is not enough memory.
 * @return a clone of this instance
 */
public Object clone()
{
	JCValueLabel clone = null;

	try {
		clone = (JCValueLabel)super.clone();

		clone.value        = value;
		clone.label        = label;
		clone.tickDrawn    = tickDrawn;
		clone.labelDrawn   = labelDrawn;
		clone.minorTick    = minorTick;
		clone.background   = background;
		clone.foreground   = foreground;
		clone.index        = index;
		clone.axis         = axis;
		clone.allowUpdates = false;
	}
	catch (CloneNotSupportedException e) {
		// This should never happen.
		throw new RuntimeException("JCValueLabel does not support cloneable.");
	}

	return(clone);
}

}


