/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCViewport.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import com.klg.jclass.chart3d.resources.LocaleBundle;

/**
 * The <code>JCViewport</code> class stores information concerning the viewport through which
 * the user views the plotcube.  The default viewport is determined by JClass
 * Chart 3D.  The user can modify this default by scaling it and translating
 * it within the <code>JCChart3dArea</code>.
 */
public class JCViewport extends ParentTrackChange {

/** Does rotation action by drawing a wireframe preview cube (fast drawing).
 * Applies only to the 2D version of JClass Chart 3D.
 */
public final static int PREVIEW_CUBE = 1;

/** Does rotation action by drawing the full 3D chart. */
public final static int PREVIEW_FULL = 2;

/** Maximum scale at which you can zoom the data. */
public final static double MAX_VIEW_SCALE = 30.0;

/** Minimum scale at which you can zoom the data. */
public final static double MIN_VIEW_SCALE = (1.0 / MAX_VIEW_SCALE);

/**
 * Queries if the viewport is normalized (for example, <code>scale = 1</code>,
 * and <code>shifts = 0</code>).
 */
protected boolean normalized = true;

/**
 * Zoom factor.
 */
protected double scale = 1.0;

/**
 * Horizontal shift as a multiple of the original viewport size.
 */
protected double horizontalShift = 0.0;

/**
 * Vertical shift as a multiple of the original viewport size.
 */
protected double verticalShift = 0.0;

/**
 * Preview method.  Either a wireframe cube or the complete chart (surface, bar, and so on).
 */
protected int previewMethod = PREVIEW_CUBE;

/**  */
static final long serialVersionUID = -5542417106751358122L;

/**
 * No args constructor.  Values at their default.
 */
public JCViewport() {}

/**
 * Constructor that supplies values for normalized, scale, horizontal, and
 * vertical shift. <br><b>Note:</b> If the supplied normalized value is <code>true</code>, the
 * other three values will not take effect.
 *
 * @param normalized Is the viewport normalized (for example, <code>scale = 1</code>, and
 * <code>shifts = 0</code>)
 * @param scale zoom factor
 * @param horizontalShift horizontal shift as a multiple of the original
 *     viewport size
 * @param verticalShift vertical shift as a multiple of the original viewport
 *      size
 */
public JCViewport(boolean normalized, double scale,
                  double horizontalShift, double verticalShift)
{
    this.normalized = normalized;
	if (this.normalized) {  // Ignore other values
	    this.scale           = 1.0;
        this.horizontalShift = 0.0;;
        this.verticalShift   = 0.0;
	}
	else {
        this.scale           = scale;
        this.horizontalShift = horizontalShift;
        this.verticalShift   = verticalShift;
	    this.normalized = (this.scale           == 1.0 &&
					       this.horizontalShift == 0.0 &&
					       this.verticalShift   == 0.0);
	}
}

/**
 * Sets the normalized value.  If normalized is <code>true</code>, the scale, horizontal
 * shift, and vertical shift will be set to their default values.  If <code>false</code>,
 * these values can be set to non-default values.
 *
 * @param normalized the new normalized value
 */
public void setNormalized(boolean normalized)
{
	if (this.normalized == normalized) {
		return;
	}
    if (normalized == true) {
        scale           = 1.0;
        horizontalShift = 0.0;
        verticalShift   = 0.0;
    }
    this.normalized = normalized;
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Asks if the viewport is normalized (for example, <code>scale = 1</code>, and
 * <code>shifts = 0</code>).
 * @return whether the viewport is normalized or not
 */
public boolean isNormalized()
{
    return(normalized);
}

/**
 * Sets the zoom factor.  A value less than one means a viewport smaller than
 * the default.  A value greater than one means a viewport greater than the
 * default (the graph image will be clipped by the <code>JCChart3dArea</code>).
 *
 * @param scale the new scale zoom factor
 */
public void setScale(double scale)
{
	if (this.scale == scale) {
		return;
	}
    this.scale = JCChart3dUtil.clamp(scale, MIN_VIEW_SCALE, MAX_VIEW_SCALE);
	this.normalized = (this.scale           == 1.0 &&
					   this.horizontalShift == 0.0 &&
					   this.verticalShift   == 0.0);
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Gets the scale zoom factor.
 *
 * @return the current scale zoom factor
 */
public double getScale()
{
    return(scale);
}

/**
 * Sets the horizontal shift as a multiple of the original viewport size.
 *
 * @param horizontalShift the new horizontal shift
 */
public void setHorizontalShift(double horizontalShift)
{
	if (this.horizontalShift == horizontalShift) {
		return;
	}
    this.horizontalShift = horizontalShift;
	this.normalized = (this.scale           == 1.0 &&
					   this.horizontalShift == 0.0 &&
					   this.verticalShift   == 0.0);
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the horizontal shift as a multiple of the original viewport size.
 *
 * @return the current horizontal shift
 */
public double getHorizontalShift()
{
    return(horizontalShift);
}

/**
 * Sets the vertical shift as a multiple of the original viewport size.
 *
 * @param verticalShift the new vertical shift
 */
public void setVerticalShift(double verticalShift)
{
	if (this.verticalShift == verticalShift) {
		return;
	}
    this.verticalShift = verticalShift;
	this.normalized = (this.scale           == 1.0 &&
					   this.horizontalShift == 0.0 &&
					   this.verticalShift   == 0.0);
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Returns the vertical shift as a multiple of the original viewport size.
 *
 * @return the current vertical shift
 */
public double getVerticalShift()
{
    return(verticalShift);
}
/**
 * Set the scale, <code>horizontalShift</code>, and <code>verticalShift</code> in one call.
 * @param scale the new scale
 * @param horizontalShift the new horizontal shift
 * @param verticalShift the new vertical shift
 */
public void setViewport(double scale, double horizontalShift,
                        double verticalShift)
{
	// Don't check for equality here as the actions may call this with the
	// same values as before.  We want an update.
    this.scale           = scale;
    this.horizontalShift = horizontalShift;
    this.verticalShift   = verticalShift;
	this.normalized      = (this.scale           == 1.0 &&
							this.horizontalShift == 0.0 &&
							this.verticalShift   == 0.0);
	callParentSetChanged(true, Changeable.RECALC);
}

/**
 * Sets the preview method.  This controls what is drawn as the user rotates
 * the cube interactively.  This can either be <code>PREVIEW_CUBE</code>
 * (display a wireframe cube) or <code>PREVIEW_FULL</code> (display the entire
 * surface).
 *
 * @param previewMethod the new preview method
 */
public void setPreviewMethod(int previewMethod)
{
	if (previewMethod >= PREVIEW_CUBE || previewMethod <= PREVIEW_FULL) {
		this.previewMethod = previewMethod;
	}
	else {
		throw new IllegalArgumentException(
            LocaleBundle.string(LocaleBundle.INVALID_CUBE_OR_FULL_ERROR));
	}
	// No need for "callParentSetChanged" here as this property is only in
	// effect during the time a user interactively rotates the cube.
}

/**
 * Returns the <code>preview</code> method.  This controls what is drawn as the
 * user rotates the cube interactively.  This can either be
 * <code>PREVIEW_CUBE</code> (display a wireframe cube) or
 * <code>PREVIEW_FULL</code> (display the entire surface).
 *
 * @return the current preview method
 */
public int getPreviewMethod()
{
	return(previewMethod);
}

}
