/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ProcessGrid.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Graphics;

/**
 * Class that controls the drawing order for a grid data set.
 */
public class ProcessGrid implements java.io.Serializable {

/** The max number of points to be drawn at one time during a fast mesh draw */
public final static int MAX_POINTS = 500;

/** Parent DrawGrid object */
protected DrawGrid drawGrid = null;

/** The facet processing object */
protected Facet facet = null;

/**
 * Null constructor.  Parent is null!
 */
public ProcessGrid() {}

/**
 * Constructor that specifies the parent DrawGrid object.
 *
 * @param drawGrid The parent DrawGrid object
 */
public ProcessGrid(DrawGrid drawGrid)
{
	this.drawGrid = drawGrid;
}

/**
 * Get the parent DrawGrid object
 * @return The current parent DrawGrid object
 */
protected DrawGrid getDrawGrid()
{
	return(drawGrid);
}

/**
 * Set the parent DrawGrid object
 * @param drawGrid The new parent DrawGrid object
 */
protected void setDrawGrid(DrawGrid drawGrid)
{
	this.drawGrid = drawGrid;
}

/**
 * Get the facet processing object
 * @return The current facet processing object
 */
protected Facet getFacet()
{
	return(facet);
}

/**
 * Set the facet processing object
 * @param facet The new facet processing object
 */
public void setFacet(Facet facet)
{
	this.facet = facet;
}

/**
 * Process facets in the grid based on the order specified by gridRegion (@see
 * GridRegion).  Process the four corners first, then the four side strips,
 * and then the center (the eye).  Note that regions can be null which results
 * in no facets being processed for that region.  In fact, the entire grid
 * could be one corner region.
 *
 * @param gridRegion The object which defines how the grid is split up into
 *     drawing regions.
 */
public void processGridRegion(GridRegion gridRegion)
{
	if (facet == null) {
		return;
	}

	int xBegin = gridRegion.getXBegin();
	int xLeft  = gridRegion.getXLeft();
	int xEye   = gridRegion.getXEye();
	int xRight = gridRegion.getXRight();
	int xEnd   = gridRegion.getXEnd();
	int yBegin = gridRegion.getYBegin();
	int yLeft  = gridRegion.getYLeft();
	int yEye   = gridRegion.getYEye();
	int yRight = gridRegion.getYRight();
	int yEnd   = gridRegion.getYEnd();

	// Process the four corner regions
	processGrid(xEnd,   xRight, -1, yEnd,   yRight, -1, GridRegion.REGION_6);
	processGrid(xEnd,   xRight, -1, yBegin, yLeft,   1, GridRegion.REGION_7);
	processGrid(xBegin, xLeft,   1, yEnd,   yRight, -1, GridRegion.REGION_8);
	processGrid(xBegin, xLeft,   1, yBegin, yLeft,   1, GridRegion.REGION_9);

	// Process the Y strips
	if (xRight == xLeft + 2) {
		processY(xEye, yEnd,   yRight, -1, GridRegion.REGION_2);
		processY(xEye, yBegin, yLeft,   1, GridRegion.REGION_3);
	}

	// Process the X strips and the center (the eye)
	if (yRight == yLeft + 2) {
		processX(yEye, xEnd,   xRight, -1, GridRegion.REGION_4);
		processX(yEye, xBegin, xLeft,   1, GridRegion.REGION_5);

		if (xRight == xLeft + 2) {
			facet.processFacet(xEye, yEye, GridRegion.REGION_1);
		}
	}
}

/**
 * Process the grid in X major order.
 *
 * @param iFirst The index of the first facet of the region in the X direction
 * @param iLast The index of the last facet of the region in the X direction
 * @param iDelta The X increment (could be 1 or -1)
 * @param jFirst The index of the first facet of the region in the Y direction
 * @param jLast The index of the last facet of the region in the Y direction
 * @param jDelta The Y increment (could be 1 or -1)
 * @param edges The grid region that the facet is in.
 */
protected void processXY(int iFirst, int iLast, int iDelta,
                         int jFirst, int jLast, int jDelta,
					     int edges)
{
	if (iDelta > 0) {					/* GridRegion.REGION_9 */
		if (jDelta > 0) {
			for (int i = iFirst; i <= iLast; i++) {
				for (int j = jFirst; j <= jLast; j++) {
					facet.processFacet(i, j, edges);

				}
			}
		}
		else {							/* GridRegion.REGION_8 */
			for (int i = iFirst; i <= iLast; i++) {
				for (int j = jFirst; j >= jLast; j--) {
					facet.processFacet(i, j, edges);
				}
			}
		}
	}
	else {								/* GridRegion.REGION_7 */
		if (jDelta > 0) {
			for (int i = iFirst; i >= iLast; i--) {
				for (int j = jFirst; j <= jLast; j++) {
					facet.processFacet(i, j, edges);
				}
			}
		}
		else {							/* GridRegion.REGION_6 */
			for (int i = iFirst; i >= iLast; i--) {
				for (int j = jFirst; j >= jLast; j--) {
					facet.processFacet(i, j, edges);
				}
			}
		}
	}
}

/**
 * Process the grid in Y major order.
 *
 * @param iFirst The index of the first facet of the region in the X direction
 * @param iLast The index of the last facet of the region in the X direction
 * @param iDelta The X increment (could be 1 or -1)
 * @param jFirst The index of the first facet of the region in the Y direction
 * @param jLast The index of the last facet of the region in the Y direction
 * @param jDelta The Y increment (could be 1 or -1)
 * @param edges The grid region that the facet is in.
 */
protected void processYX(int iFirst, int iLast, int iDelta,
                         int jFirst, int jLast, int jDelta,
					     int edges)
{
	if (jDelta > 0) {					/* GridRegion.REGION_9 */
		if (iDelta > 0) {
			for (int j = jFirst; j <= jLast; j++) {
				for (int i = iFirst; i <= iLast; i++) {
					facet.processFacet(i, j, edges);
				}
			}
		}
		else {							/* GridRegion.REGION_7 */
			for (int j = jFirst; j <= jLast; j++) {
				for (int i = iFirst; i >= iLast; i--) {
					facet.processFacet(i, j, edges);
				}
			}
		}
	}
	else {								/* GridRegion.REGION_8 */
		if (iDelta > 0) {
			for (int j = jFirst; j >= jLast; j--) {
				for (int i = iFirst; i <= iLast; i++) {
					facet.processFacet(i, j, edges);
				}
			}
		}
		else {							/* GridRegion.REGION_6 */
			for (int j = jFirst; j >= jLast; j--) {
				for (int i = iFirst; i >= iLast; i--) {
					facet.processFacet(i, j, edges);
				}
			}
		}
	}
}

/**
 * Process the rectangular portion of the grid region defined by the first
 * and last indices in either X major or Y major order.
 *
 * @param iFirst The index of the first facet of the region in the X direction
 * @param iLast The index of the last facet of the region in the X direction
 * @param iDelta The X increment (could be 1 or -1)
 * @param jFirst The index of the first facet of the region in the Y direction
 * @param jLast The index of the last facet of the region in the Y direction
 * @param jDelta The Y increment (could be 1 or -1)
 * @param edges The grid region that the facet is in.
 */
protected void processGrid(int iFirst, int iLast, int iDelta,
			               int jFirst, int jLast, int jDelta,
						   int edges)
{
    Chart3dGridData data = drawGrid.getGridData();
	JCChart3dArea   area = data.getDataView().getChart3d().getChart3dArea();
	int             order = area.getTransform().getProcessingOrder() & 15;

	switch(order) {
	case Transform.EDGE_YMIN:
	case Transform.EDGE_YMAX:
	    processXY(iFirst, iLast, iDelta, jFirst, jLast, jDelta, edges);
		break;
	default:
	    processYX(iFirst, iLast, iDelta, jFirst, jLast, jDelta, edges);
		break;
	}
}

/**
 * Process an X range for a fixed Y grid value
 *
 * @param j0 The fixed X grid value
 * @param iFirst The index of the first facet of the region in the X direction
 * @param iLast The index of the last facet of the region in the X direction
 * @param iDelta The X increment (could be 1 or -1)
 * @param edges The grid region that the facet is in.
 */
protected void processX(int j0, int iFirst, int iLast, int iDelta, int edges)
{
	if (iDelta > 0) {						/* GridRegion.REGION_5 */
		for (int i = iFirst; i <= iLast; i++) {
			facet.processFacet(i, j0, edges);
		}
	}
	else {									/* GridRegion.REGION_4 */
		for (int i = iFirst; i >= iLast; i--) {
			facet.processFacet(i, j0, edges);
		}
	}
}

/**
 * Process an Y range for a fixed X grid value
 *
 * @param i0 The fixed X grid value
 * @param jFirst The index of the first facet of the region in the Y direction
 * @param jLast The index of the last facet of the region in the Y direction
 * @param jDelta The Y increment (could be 1 or -1)
 * @param edges The grid region that the facet is in.
 */
protected void processY(int i0, int jFirst, int jLast, int jDelta, int edges)
{
	if (jDelta > 0) {						/* GridRegion.REGION_3 */
		for (int j = jFirst; j <= jLast; j++) {
			facet.processFacet(i0, j, edges);
		}
	}
	else {									/* GridRegion.REGION_2 */
		for (int j = jFirst; j >= jLast; j--) {
			facet.processFacet(i0, j, edges);
		}
	}
}

/**
 * Draws the list of lines represented by the given point list.
 *
 * @param pts The list of points
 * @param n The number of points within the list to draw
 * @param gc The graphics object used to draw
 */
protected void flushLines(ScreenPoint[] pts, int n, Graphics gc)
{
	for (int i = 0; i < n-1; i++) {
		gc.drawLine(pts[i].getXPixel(), pts[i].getYPixel(),
		            pts[i+1].getXPixel(), pts[i+1].getYPixel());
	}
}

/**
 * Draws the mesh as fast as possible.  The bottom and top mesh colors are
 * the same, and hidden lines are being drawn, so we can just blast 'em!
 *
 * @param gc The graphics object used to draw
 */
protected void drawFastMesh(Graphics gc)
{
	boolean				hole;
	ScreenPoint[]		pts = new ScreenPoint[MAX_POINTS];
    Chart3dGridData     data = drawGrid.getGridData();
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	GridLimits          gridLimits = data.getGridLimits();
	Chart3dDataView     dataView = data.getDataView();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
	JCSurface           surface = area.getSurface();
	boolean[]           xMesh = surface.getXMesh();
	boolean[]           yMesh = surface.getYMesh();

	gc.setColor(dataView.getElevation().getMeshBottomColor());

	int k = 0;
	if (surface.isYMeshShowing()) {
		for (int i = gridLimits.getXOffset(); i <= gridLimits.getXLast(); i++) {
			if (yMesh[i]) {
				for (int j = gridLimits.getYOffset();
                     j <= gridLimits.getYLast();
                     j++)
                {
					hole = (screenPoints[i][j].isHoleValue() ||
						    ((i == gridLimits.getXOffset() ||
							  screenPoints[i-1][j].isHoleValue()) &&
							 (i == gridLimits.getXLast() ||
							  screenPoints[i+1][j].isHoleValue())));
					if (hole) {
					    flushLines(pts, k, gc);
						k = 0;
					}
					else {
						if (k == MAX_POINTS) {
							flushLines(pts, k, gc);
							k = 0;
						}
						pts[k] = screenPoints[i][j];
						k++;
					}
				}
			}
			flushLines(pts, k, gc);
			k = 0;
		}
	}

	if (surface.isXMeshShowing()) {
		for (int j = gridLimits.getYOffset(); j <= gridLimits.getYLast(); j++) {
			if (xMesh[j]) {
				for (int i = gridLimits.getXOffset();
                     i <= gridLimits.getXLast();
                     i++)
                {
					hole = (screenPoints[i][j].isHoleValue() ||
						    ((j == gridLimits.getYOffset() ||
							  screenPoints[i][j-1].isHoleValue()) &&
							 (j == gridLimits.getYLast() ||
							  screenPoints[i][j+1].isHoleValue())));
					if (hole) {
					    flushLines(pts, k, gc);
						k = 0;
					}
					else {
						if (k == MAX_POINTS) {
							flushLines(pts, k, gc);
							k = 0;
						}
						pts[k] = screenPoints[i][j];
						k++;
					}
				}
			}
			flushLines(pts, k, gc);
			k = 0;
		}
	}
}

}


