/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SurfaceFacet.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:12:58 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;



import java.awt.*;

/**
 * Draws facets for charts of type surface.
 *
 */
public class SurfaceFacet extends FacetAbstract {

// The facet processing proceeds differently based on the region code.
// The region code contains information about the facet processing order
// (ie. the order that the grid is being traversed).  This information is
// important since it allows us to determine:
//   a) which poly is farther away when a quad must be split into triangles
//   b) which sides must be ignored when drawing the 'skirt' (ie making the
//      surface appear as a solid).  Only the far sides of the facet are
//      drawn since the near sides are either covered by nearer facets or
//      handled when the final x and/or y skirt is drawn.
//
// An encoding scheme is used to determine which sides are to be processed
// for skirts.  In a front-to-back traversal with hidden surface removal,
// these sides will be the only sides that need to be processed for drawing
// as well.  The facet is arranged as follows :
//
//                B
//           1_________2            Corner 0 is [x][y]
//           |         |            Corner 1 is [x][y+1]
//           |         |            Corner 2 is [x+1][y+1]
//         A |         | C          Corner 3 is [x+1][y]
//	         |         |
//           |_________|
//           0         3
//                D
//
// The encoding scheme is an 8 bit code with each corner contributing 2
// bits (1 to each side) as  follows:
//
//   0 3 2 1 3 2 1 0
//  |-|-|-|-|-|-|-|-|  Only the lower 4 bits are passed on as the hole code.
//   D C B A D C B A
//
// Note: This code uses negative logic.
// Code bits are added for missing sides and corners.  For a side to be
// active, both of its corner bits must be clear.  The missing sides are
// tagged by adding bits to the upper byte.
//
// Note that we are really overlaying two pieces of information :
//  - the lower byte marks the corners which are missing
//  - the upper byte marks the sides that we DO NOT WISH to process

/**  */
public final static int FP_0_CODE = 129;       // 1000 0001 or 0x81
/**  */
public final static int FP_1_CODE = 18;        // 0001 1000 or 0x12
/**  */
public final static int FP_2_CODE = 36;        // 0010 0100 or 0x24
/**  */
public final static int FP_3_CODE = 72;        // 0100 1000 or 0x48
/**  */
public final static int FP_H_CODE = 15;        // 0000 1111 or 0x0F -- all corner bits

/**  */
public final static int FP_A_CODE = 17;        // 0001 0001 or 0x11
/**  */
public final static int FP_B_CODE = 34;        // 0010 0010 or 0x22
/**  */
public final static int FP_C_CODE = 68;        // 0100 0100 or 0x44
/**  */
public final static int FP_D_CODE = 136;       // 1000 1000 or 0x88
/**  */
public final static int FP_S_CODE = 240;       // 1111 0000 or 0xF0 -- all side bits

/**  */
protected DrawFacet drawFacet = null;

/**
 * No args contstructor.  Nothing will be drawn as long as
 * <code>drawFacet</code> is null.
 */
SurfaceFacet() {}

/**
 * Constructor that stores the parent <code>DrawGrid</code> object and sets the
 * <code>drawFacet</code>.
 *
 * @param drawGrid The parent <code>DrawGrid</code> object
 * @param drawFacet
 */
public SurfaceFacet(DrawGrid drawGrid, DrawFacet drawFacet)
{
	super(drawGrid);
	this.drawFacet = drawFacet;
}

/**
 * Setup integer arrays for drawing from a <code>ScreenPoint</code> array
 *
 * @param spts
 * @param holeCode
 * @param x
 * @param y
 */
protected void drawFacetFromScreenPoints(ScreenPoint[] spts, int holeCode,
			                             int x, int y)
{
	if (drawFacet == null) {
		return;
	}
	fillPointArrays(spts, spts.length);
	drawFacet.draw(xPoints, yPoints, spts.length, holeCode, x, y);
}

/**
 * Draws the facet for (<i>x</i>, <i>y</i>) based on the region passed in thru
 * "edges".
 *
 * @param x
 * @param y
 * @param edges
 */
public void processFacet(int x, int y, int edges)
{
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	GridLimits          gridLimits = data.getGridLimits();
	Chart3dDataView     dataView = data.getDataView();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
	JCSurface           surface = area.getSurface();

	if (x < gridLimits.getXOffset() ||
	    x > gridLimits.getXLast()   ||
		y < gridLimits.getYOffset() ||
		y > gridLimits.getYLast())
	{
		// Should never get here
	    
		return;
	}

	if (screenPoints[x][y].isDrawn()) {
		return;
	}

	ScreenPoint[] s = new ScreenPoint[4];
	s[0] = screenPoints[x][y];
	s[1] = screenPoints[x][y+1];
	s[2] = screenPoints[x+1][y+1];
	s[3] = screenPoints[x+1][y];

	int fCode = 0;
	if (s[0].isHoleValue()) {
		fCode |= FP_0_CODE;
	}
	if (s[1].isHoleValue()) {
		fCode |= FP_1_CODE;
	}
	if (s[2].isHoleValue()) {
		fCode |= FP_2_CODE;
	}
	if (s[3].isHoleValue()) {
		fCode |= FP_3_CODE;
	}
	int holeCode = fCode & FP_H_CODE;   // Only the corner part

	//  Don't process any facets outside chart area due to zooming
	if (!area.getViewport().isNormalized()) {
		int x1 = JCChart3dUtil.min4(s[0].getXPixel(), s[1].getXPixel(),
                                    s[2].getXPixel(), s[3].getXPixel());
        int y1 = JCChart3dUtil.min4(s[0].getYPixel(), s[1].getYPixel(),
                                    s[2].getYPixel(), s[3].getYPixel());
        int x2 = JCChart3dUtil.max4(s[0].getXPixel(), s[1].getXPixel(),
                                    s[2].getXPixel(), s[3].getXPixel());
        int y2 = JCChart3dUtil.max4(s[0].getYPixel(), s[1].getYPixel(),
                                    s[2].getYPixel(), s[3].getYPixel());
        Rectangle rect = new Rectangle(x1, y1, x2 - x1, y2 - y1);

		if (surface.isSolid() && holeCode != 0) {
			ScreenPoint[] z = new ScreenPoint[4];

			z[0] = makeZMinPoint(x, y);
			z[1] = makeZMinPoint(x+1, y);
			z[2] = makeZMinPoint(x+1, y+1);
			z[3] = makeZMinPoint(x, y+1);
		    x1 = JCChart3dUtil.min4(z[0].getXPixel(), z[1].getXPixel(),
                                    z[2].getXPixel(), z[3].getXPixel());
            y1 = JCChart3dUtil.min4(z[0].getYPixel(), z[1].getYPixel(),
                                    z[2].getYPixel(), z[3].getYPixel());
            x2 = JCChart3dUtil.max4(z[0].getXPixel(), z[1].getXPixel(),
                                    z[2].getXPixel(), z[3].getXPixel());
            y2 = JCChart3dUtil.max4(z[0].getYPixel(), z[1].getYPixel(),
                                    z[2].getYPixel(), z[3].getYPixel());
            Rectangle zrect = new Rectangle(x1, y1, x2 - x1, y2 - y1);

            Rectangle brect = JCChart3dUtil.computeBoundingRectangle(rect,
                                                                     zrect);
            rect.setBounds(brect);
		}

        if (!rect.intersects(area.getDrawingArea())) {
            return;
        }
	}

	int sCode = 0;
	switch (edges) {
	case GridRegion.REGION_1:
		/*   _   */
		/*  |_|  */
		/*       */
		/* no self-occluding possibilities */

		if (holeCode == 0) {
		    drawFacetFromScreenPoints(s, holeCode, x, y);
		}
		break;

	case GridRegion.REGION_2:
		/*   _   */
		/*  | |  */
		/*       */
		sCode = FP_D_CODE;

		drawEdgeRegionFacet(s, 0, 1, 2, 3, holeCode, x, y);
		break;

	case GridRegion.REGION_3:
		/*       */
		/*  |_|  */
		/*       */
		sCode  = FP_B_CODE;

		drawEdgeRegionFacet(s, 2, 3, 0, 1, holeCode, x, y);
		break;

	case GridRegion.REGION_4:
		/*   _   */
		/*   _|  */
		/*       */
		sCode  = FP_A_CODE;

		drawEdgeRegionFacet(s, 1, 2, 3, 0, holeCode, x, y);
		break;

	case GridRegion.REGION_5:
		/*   _   */
		/*  |_   */
		/*       */
		sCode  = FP_C_CODE;

		drawEdgeRegionFacet(s, 3, 0, 1, 2, holeCode, x, y);
		break;

	case GridRegion.REGION_6:
		/*   _   */
		/*    |  */
		/*       */
		sCode  = FP_A_CODE | FP_D_CODE;

		drawCornerRegionFacet(s, 1, 2, 3, 0, holeCode, x, y);
		break;

	case GridRegion.REGION_7:
		/*       */
		/*   _|  */
		/*       */
		sCode  = FP_A_CODE |  FP_B_CODE;

		drawCornerRegionFacet(s, 2, 3, 0, 1, holeCode, x, y);
		break;

	case GridRegion.REGION_8:
		/*   _   */
		/*  |    */
		/*       */
		sCode  = FP_C_CODE | FP_D_CODE;

		drawCornerRegionFacet(s, 0, 1, 2, 3, holeCode, x, y);
		break;

	case GridRegion.REGION_9:
		/*       */
		/*  |_   */
		/*       */
		sCode  = FP_B_CODE | FP_C_CODE;

		drawCornerRegionFacet(s, 3, 0, 1, 2, holeCode, x, y);
		break;

	default:
		
		break;
	}

	// Handle drawing interior skirts where visible due to holes
	if (area.getSurface().isSolid()) {
		fCode = neighbourTest(fCode, x, y);
		fCode |= (sCode & FP_S_CODE);

		if ((fCode & FP_H_CODE) != 0) {
			if ((fCode & FP_A_CODE) == 0) {
				dropSkirt(x, y, x, y + 1);
			}

			if ((fCode & FP_B_CODE) == 0) {
				dropSkirt(x, y + 1, x + 1, y + 1);
			}

			if ((fCode & FP_C_CODE) == 0) {
				dropSkirt(x + 1, y, x + 1, y + 1);
			}

			if ((fCode & FP_D_CODE) == 0) {
				dropSkirt(x, y, x + 1, y);
			}
		}
	}
	s[0].setDrawn(true);
}

/**
 * Is the passed in parameter odd?
 *
 * @param i
 */
protected int isOdd(int i)
{
	return(i & 1);
}

/**
 * Draws facets that are part of the four edge regions.
 *
 * @param spts
 * @param i
 * @param j
 * @param k
 * @param l
 * @param holeCode
 * @param x
 * @param y
 */
protected void drawEdgeRegionFacet(ScreenPoint[] spts,
			                       int i, int j, int k, int l,
								   int holeCode, int x, int y)
{
	if (drawFacet == null) {
		return;
	}

	Chart3dDataView     dataView = data.getDataView();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
	JCSurface           surface = area.getSurface();
	boolean[]           xMesh = surface.getXMesh();
	boolean[]           yMesh = surface.getYMesh();

	// Only need to draw polygons if there are NO holes
	if (holeCode != 0) {
		return;
	}

	if (dataView.getContour().isZoned()) {
		// Zoned facets are not triangled
		drawFacetFromScreenPoints(spts, holeCode, x, y);
		return;
	}

	if (testSelfOcclude((long)spts[j].getXPixel(), (long)spts[j].getYPixel(),
	                    (long)spts[k].getXPixel(), (long)spts[k].getYPixel(),
	                    (long)spts[l].getXPixel(), (long)spts[l].getYPixel(),
	                    (long)spts[i].getXPixel(), (long)spts[i].getYPixel()))
	{
		ScreenPoint[] tri = new ScreenPoint[3];
		tri[0] = quadCenter(x, y);

		// Draw the four triangles in order based on region
		tri[1] = spts[j];
		tri[2] = spts[k];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(j) != 0)
				                   ? xMesh[y + isOdd(k/2)]
				                   : yMesh[x + isOdd(j/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);

		tri[1] = spts[i];
		tri[2] = spts[j];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(i) != 0)
				                   ? xMesh[y + isOdd(j/2)]
				                   : yMesh[x + isOdd(i/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);

		tri[1] = spts[k];
		tri[2] = spts[l];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(k) != 0)
				                   ? xMesh[y + isOdd(l/2)]
				                   : yMesh[x + isOdd(k/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);

		tri[1] = spts[l];
		tri[2] = spts[i];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(l) != 0)
				                   ? xMesh[y + isOdd(i/2)]
				                   : yMesh[x + isOdd(l/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);
	}
	else {
		drawFacetFromScreenPoints(spts, holeCode, x, y);
	}
}

/**
 * Draws a facet that is in one of the four corner regions.
 *
 * @param spts
 * @param i
 * @param j
 * @param k
 * @param l
 * @param holeCode
 * @param x
 * @param y
 */
protected void drawCornerRegionFacet(ScreenPoint[] spts,
			                         int i, int j, int k, int l,
								     int holeCode, int x, int y)
{
	if (drawFacet == null) {
		return;
	}

	Chart3dDataView     dataView = data.getDataView();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
	JCSurface           surface = area.getSurface();
	boolean[]           xMesh = surface.getXMesh();
	boolean[]           yMesh = surface.getYMesh();

	// Only need to draw polygons if there are NO holes
	if (holeCode != 0) {
		return;
	}

	if (dataView.getContour().isZoned()) {
		// Zoned facets are not triangled
		drawFacetFromScreenPoints(spts, holeCode, x, y);
		return;
	}

	if (testSelfOcclude((long)spts[i].getXPixel(), (long)spts[i].getYPixel(),
	                    (long)spts[j].getXPixel(), (long)spts[j].getYPixel(),
	                    (long)spts[k].getXPixel(), (long)spts[k].getYPixel(),
	                    (long)spts[l].getXPixel(), (long)spts[l].getYPixel()) ||
		testSelfOcclude((long)spts[j].getXPixel(), (long)spts[j].getYPixel(),
	                    (long)spts[k].getXPixel(), (long)spts[k].getYPixel(),
	                    (long)spts[l].getXPixel(), (long)spts[l].getYPixel(),
	                    (long)spts[i].getXPixel(), (long)spts[i].getYPixel()))
	{
		ScreenPoint[] tri = new ScreenPoint[3];
		tri[0] = quadCenter(x, y);

		// Draw the four triangles in order based on region
		tri[1] = spts[i];
		tri[2] = spts[j];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(i) != 0)
				                   ? xMesh[y + isOdd(j/2)]
				                   : yMesh[x + isOdd(i/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);

		tri[1] = spts[j];
		tri[2] = spts[k];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(j) != 0)
				                   ? xMesh[y + isOdd(k/2)]
				                   : yMesh[x + isOdd(j/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);

		tri[1] = spts[k];
		tri[2] = spts[l];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(k) != 0)
				                   ? xMesh[y + isOdd(l/2)]
				                   : yMesh[x + isOdd(k/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);

		tri[1] = spts[l];
		tri[2] = spts[i];
		drawFacet.setDrawMeshLine(dataView.getElevation().isMeshed() &&
		                          ((isOdd(l) != 0)
				                   ? xMesh[y + isOdd(i/2)]
				                   : yMesh[x + isOdd(l/2)]));
		drawFacetFromScreenPoints(tri, holeCode, x, y);
	}
	else {
		drawFacetFromScreenPoints(spts, holeCode, x, y);
	}
}

/**
 * Returns the screen point of the center point of the quad.
 *
 * @param x
 * @param y
 */
protected ScreenPoint quadCenter(int x, int y)
{
	// Approx the data value by averaging the four points
    return(getScreenPoint((data.getX(x) + data.getX(x+1)) / 2.0,
	                      (data.getY(y) + data.getY(y+1)) / 2.0,
		                  (data.getZValue(x, y) + data.getZValue(x, y+1) +
		                   data.getZValue(x+1,y+1) + data.getZValue(x+1, y)) / 4.0));
}

/**
 * Makes a point that is at the current gric point but has a <i>z</i> value that is
 * at the <i>z</i> axis min.  This is used for the bottom of the skirts.
 *
 * @param x
 * @param y
 */
protected ScreenPoint makeZMinPoint(int x, int y)
{
	JCChart3dArea   area = data.getDataView().getChart3d().getChart3dArea();
	JCAxis          zAxis = area.getAxis(JCAxis.AXIS_Z);

	return(getScreenPoint(data.getX(x), data.getY(y), zAxis.getMin()));
}

/**
 * Does this facet occlude itself?
 *
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 * @param x3
 * @param y3
 * @param x4
 * @param y4
 */
protected boolean testSelfOcclude(long x1, long y1,
				                  long x2, long y2,
				                  long x3, long y3,
				                  long x4, long y4)
{
	long	x1lo, x1hi;
	long	y1lo, y1hi;

	long Ax = x2 - x1;
	long Bx = x3 - x4;

	// X bound box test
	if (Ax < 0)	{
		x1lo = x2;
		x1hi = x1;
	}
	else {
		x1hi = x2;
		x1lo = x1;
	}

	if (Bx > 0) {
		if (x1hi < x4 || x3 < x1lo) {
			return(false);
		}
	}
	else {
		if (x1hi < x3 || x4 < x1lo) {
			return(false);
		}
	}

	long Ay = y2 - y1;
	long By = y3 - y4;

	// Y bound box test
	if (Ay < 0)	{
		y1lo = y2;
		y1hi = y1;
	}
	else {
		y1hi = y2;
		y1lo = y1;
	}

	if (By > 0) {
		if (y1hi < y4 || y3 < y1lo) {
			return(false);
		}
	}
	else {
		if (y1hi < y3 || y4 < y1lo) {
			return(false);
		}
	}

	// Solve system of equations for the intersection point
	long Cx = x1 - x3;
	long Cy = y1 - y3;
	long d  = By*Cx - Bx*Cy;
	long f  = Ay*Bx - Ax*By;
	if (f > 0) {
		if (d < 0 || d > f) {
			return(false);
		}
	}
	else {
		if (d > 0 || d < f) {
			return(false);
		}
	}

	long e = Ax*Cy - Ay*Cx;
	if (f > 0) {
		if (e < 0 || e > f) {
			return(false);
		}
	}
	else {
		if (e > 0 || e < f) {
			return(false);
		}
	}

	return(true);
}

/**
 * Tests the neighbouring cells for holes which might affect the sides of this
 * cell.  We don't want to draw cell walls (ha, ha) all by themselves.
 *
 * @param fCode
 * @param x
 * @param y
 */
protected int neighbourTest(int fCode, int x, int y)
{
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	GridLimits          gridLimits = data.getGridLimits();

	/* side A */
	if (x <= gridLimits.getXOffset()) {
		fCode |= (FP_A_CODE & FP_S_CODE);	/* remove A */
	}
	else if (screenPoints[x-1][y].isHoleValue() ||
		     screenPoints[x-1][y+1].isHoleValue())
	{
		fCode |= (FP_A_CODE & FP_S_CODE);	/* remove A */
	}

	/* side D */
	if (y <= gridLimits.getYOffset()) {
		fCode |= (FP_D_CODE & FP_S_CODE);	/* remove D */
	}
	else if (screenPoints[x][y-1].isHoleValue() ||
		     screenPoints[x+1][y-1].isHoleValue())
	{
		fCode |= (FP_D_CODE & FP_S_CODE);	/* remove D */
	}

	/* side C */
	if (x >= gridLimits.getXLast() - 1) {
		fCode |= (FP_C_CODE & FP_S_CODE);	/* remove C */
	}
	else if (screenPoints[x+2][y].isHoleValue() ||
		     screenPoints[x+2][y+1].isHoleValue())
	{
		fCode |= (FP_C_CODE & FP_S_CODE);	/* remove C */
	}

	/* side B */
	if (y >= gridLimits.getYLast() - 1) {
		fCode |= (FP_B_CODE & FP_S_CODE);	/* remove B */
	}
	else if (screenPoints[x][y+2].isHoleValue() ||
		     screenPoints[x+1][y+2].isHoleValue())
	{
		fCode |= (FP_B_CODE & FP_S_CODE);	/* remove B */
	}
	return(fCode);
}

/**
 * Draws interior skirt between (<i>x</i>1,<i>y</i>1) and (<i>x</i>2,<i>y</i>2)
 *
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 */
protected void dropSkirt(int x1, int y1, int x2, int y2)
{
	int					ta, tb, tc;
	boolean				draw1, draw2;
	ScreenPoint[] 		sp = new ScreenPoint[4];
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	GridLimits          gridLimits = data.getGridLimits();
	Chart3dDataView     dataView = data.getDataView();
	JCElevation         elevation = dataView.getElevation();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();
	JCSurface           surface = area.getSurface();

	if (drawFacet == null) {
		return;
	}

	sp[1] = screenPoints[x1][y1];
	sp[2] = screenPoints[x2][y2];

	// Drop point 1
	sp[0] = makeZMinPoint(x1, y1);
	// Drop point 2
	sp[3] = makeZMinPoint(x2, y2);

	// Time to actually draw the polygon
	fillPointArrays(sp, sp.length);
	Graphics gc = drawFacet.getGC();
	if (elevation.isShaded()) {
		gc.setColor(elevation.getShadedBottomColor());
		gc.fillPolygon(xPoints, yPoints, 4);
		//xrt3d_fillpolygon_hires(xrt, sp, 4, Nonconvex, CoordModeOrigin);
	}
	else if (!dataView.getElevation().isTransparent()) {
		drawFacet.polyErase(xPoints, yPoints, 4, 0);
	}

	/* Draw the horizontal lines */
	gc.setColor(elevation.getMeshBottomColor());
	gc.drawLine(sp[0].getXPixel(), sp[0].getYPixel(),
	            sp[3].getXPixel(), sp[3].getYPixel());
	gc.drawLine(sp[1].getXPixel(), sp[1].getYPixel(),
	            sp[2].getXPixel(), sp[2].getYPixel());

	/* Draw the vertical mesh lines */
	draw1 = draw2 = false;
	if (elevation.isMeshed()) {
		if (x1 == x2 && surface.isXMeshShowing()) {
			boolean[] xMesh = surface.getXMesh();
			if (xMesh[y1]) {
			    gc.drawLine(sp[0].getXPixel(), sp[0].getYPixel(),
	                        sp[1].getXPixel(), sp[1].getYPixel());
				draw1 = true;
			}
			if (xMesh[y2]) {
			    gc.drawLine(sp[3].getXPixel(), sp[3].getYPixel(),
	                        sp[2].getXPixel(), sp[2].getYPixel());
				draw2 = true;
			}
		}
		if (y1 == y2 && surface.isYMeshShowing()) {
			boolean[] yMesh = surface.getYMesh();
			if (yMesh[x1]) {
			    gc.drawLine(sp[0].getXPixel(), sp[0].getYPixel(),
	                        sp[1].getXPixel(), sp[1].getYPixel());
				draw1 = true;
			}
			if (yMesh[x2]) {
			    gc.drawLine(sp[3].getXPixel(), sp[3].getYPixel(),
	                        sp[2].getXPixel(), sp[2].getYPixel());
				draw2 = true;
			}
		}
	}

	if (draw1 && draw2) {
		return;
	}

	// Draw vertical lines where the holes stop (even if no mesh)
	//
	// Look for corners.  We're at a corner if either B is
	// a hole, or both A and C are not.  B is the next
	// point collinear to the line, and A and C are to the
	// left and right (A and C can't both be holes, since
	//  that would make this a widow (?) wall).
	//
	//  Note that x2 >= x1 and y2 >= y1.

	if (x1 == x2) {
		ta = x1 - 1;
		tc = x1 + 1;

		if (!draw1) {
			tb = y1 - 1;
			if ((tb < gridLimits.getYOffset()  ||  screenPoints[x1][tb].isHoleValue()) ||
				((ta < gridLimits.getXOffset() || !screenPoints[ta][y1].isHoleValue()) &&
				 (tc > gridLimits.getXLast()   || !screenPoints[tc][y1].isHoleValue())))
			{
			    gc.drawLine(sp[0].getXPixel(), sp[0].getYPixel(),
	                        sp[1].getXPixel(), sp[1].getYPixel());
			}
		}

		if (!draw2) {
			tb = y2 + 1;
			if ((tb > gridLimits.getYLast()    ||  screenPoints[x1][tb].isHoleValue()) ||
				((ta < gridLimits.getXOffset() || !screenPoints[ta][y2].isHoleValue()) &&
				 (tc > gridLimits.getXLast()   || !screenPoints[tc][y2].isHoleValue())))
			{
			    gc.drawLine(sp[3].getXPixel(), sp[3].getYPixel(),
	                        sp[2].getXPixel(), sp[2].getYPixel());
			}
		}
	}
	else {						// y1 == y2
		ta = y1 - 1;
		tc = y1 + 1;

		if (!draw1) {
			tb = x1 - 1;
			if ((tb < gridLimits.getXOffset()  ||  screenPoints[tb][y1].isHoleValue()) ||
				((ta < gridLimits.getYOffset() || !screenPoints[x1][ta].isHoleValue()) &&
				 (tc > gridLimits.getYLast()   || !screenPoints[x1][tc].isHoleValue())))
			{
			    gc.drawLine(sp[0].getXPixel(), sp[0].getYPixel(),
	                        sp[1].getXPixel(), sp[1].getYPixel());
			}
		}

		if (!draw2) {
			tb = x2 + 1;
			if ((tb > gridLimits.getXLast()    ||  screenPoints[tb][y1].isHoleValue()) ||
				((ta < gridLimits.getYOffset() || !screenPoints[x2][ta].isHoleValue()) &&
				 (tc > gridLimits.getYLast()   || !screenPoints[x2][tc].isHoleValue())))
			{
			    gc.drawLine(sp[3].getXPixel(), sp[3].getYPixel(),
	                        sp[2].getXPixel(), sp[2].getYPixel());
			}
		}
	}
}

/**
 * Draws a skirt along the <i>X</i> axis.
 *
 * @param y
 * @param yp
 * @param xBegin
 * @param xEnd
 */
protected void drawXSkirt(int y, int yp, int xBegin, int xEnd)
{
	ScreenPoint[] 		sp;
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	Chart3dDataView     dataView = data.getDataView();
	JCElevation         elevation = dataView.getElevation();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();

	if (drawFacet == null) {
		return;
	}
    Graphics gc = drawFacet.getGC();

    // 1: If necessary, create and fill a polygonal area
    if (elevation.isShaded() || elevation.isMeshed()) {

        // Allocate polygon ScreenPoint array.
        // Make it large enough for the entire side :
        // ie. xEnd-xBegin+1 for the points on the curve,
        //     plus 2 for points on axis,

        sp = new ScreenPoint[xEnd - xBegin + 3];

        boolean idle   = true;
        int firstPoint = xBegin - 1;    // Initialize to illegal value
        int lastPoint  = xBegin - 1;    // Initialize to illegal value
        int curPoint   = 0;
        for (int i = xBegin; i <= xEnd; i++) {
            if (idle) {
                // Look for a start point on the surface
                if ((i != xEnd) && !screenPoints[i][y].isHoleValue() &&
                                   !screenPoints[i][yp].isHoleValue())
                {
                    // Create the pt on the axis as the 1st pt of the polygon
                    firstPoint = i;
                    sp[curPoint++] = makeZMinPoint(i, y);
                    // The next point is the one on the surface
                    sp[curPoint++] = screenPoints[i][y];

                    idle = false;
                }
            }
            else {
                // Look for a stop point on the surface
                if ((i == xEnd) || screenPoints[i][y].isHoleValue() ||
                                   screenPoints[i][yp].isHoleValue())
                {
                    // Time to end this polygon
                    if (screenPoints[i][y].isHoleValue() ||
                        screenPoints[i][yp].isHoleValue())
                    {
                        lastPoint = i - 1;
                    }
                    else {
                        lastPoint = i;
                    }
                    // Don't bother with single line polygons
                    if (firstPoint != lastPoint) {
                        // The next point is the one on the surface
                        sp[curPoint++] = screenPoints[lastPoint][y];

                        // Now the point on the axis
                        sp[curPoint++] = makeZMinPoint(lastPoint, y);

                        /* Time to actually draw the polygon */
                        fillPointArrays(sp, curPoint);
                        if (elevation.isShaded()) {
                            gc.setColor(elevation.getShadedBottomColor());
                            gc.fillPolygon(xPoints, yPoints, curPoint);
                        }
                        else if (elevation.isMeshed()) {
                            if (!elevation.isTransparent()) {
                                drawFacet.polyErase(xPoints, yPoints,
                                                    curPoint, 0);
                            }
                        }
                        // Also draw outline
                        gc.setColor(elevation.getMeshBottomColor());
                        gc.drawPolygon(xPoints, yPoints, curPoint);
                    }
                    curPoint = 0;
                    idle = true;
                }
                else {
                    // Just add the point to the list
                    sp[curPoint++] = screenPoints[i][y];
                }
            }
        }
    }

    // 2: If necessary, draw the drop-down vectors
    if (elevation.isMeshed()) {
	    JCSurface surface = area.getSurface();
        for (int i = xBegin; i <= xEnd; i++) {
            boolean[] yMesh = surface.getYMesh();
            if (surface.isYMeshShowing()          &&
                yMesh[i]                          &&
                !screenPoints[i][y].isHoleValue() &&
                !screenPoints[i][yp].isHoleValue())
            {
                if (((i > xBegin) && !screenPoints[i-1][y].isHoleValue()) ||
                    ((i < xEnd)   && !screenPoints[i+1][y].isHoleValue()))
                {
                    ScreenPoint dp0 = makeZMinPoint(i, y);
                    // The next point is the one on the surface
                    ScreenPoint dp1 = screenPoints[i][y];

                    gc.setColor(elevation.getMeshBottomColor());
                    gc.drawLine(dp0.getXPixel(), dp0.getYPixel(),
                                dp1.getXPixel(), dp1.getYPixel());
                }
            }
        }
    }
}

/**
 * Draws a skirt along the <i>Y</i> axis.
 *
 * @param x
 * @param xp
 * @param yBegin
 * @param yEnd
 */
protected void drawYSkirt(int x, int xp, int yBegin, int yEnd)
{
	ScreenPoint[] 		sp;
	ScreenPoint[][]     screenPoints = data.getScreenPoints();
	Chart3dDataView     dataView = data.getDataView();
	JCElevation         elevation = dataView.getElevation();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();

	if (drawFacet == null) {
		return;
	}
    Graphics gc = drawFacet.getGC();

    // 1: If necessary, create and fill a polygonal area
    if (elevation.isShaded() || elevation.isMeshed()) {

        // Allocate polygon XPoint array.
        // Make it large enough for the entire side :
        // ie. yEnd-yBegin+1 for the points on the curve,
        //     plus 2 for points on axis,

        sp = new ScreenPoint[yEnd - yBegin + 3];

        boolean idle   = true;
        int firstPoint = yBegin - 1;    // Initialize to illegal value
        int lastPoint  = yBegin - 1;    // Initialize to illegal value
        int curPoint   = 0;
        for (int i = yBegin; i <= yEnd; i++) {
            if (idle) {
                // Look for a start point on the surface
                if ((i != yEnd) && !screenPoints[x][i].isHoleValue() &&
                                   !screenPoints[xp][i].isHoleValue())
                {
                    // Create the pt on the axis as the 1st pt of the polygon
                    firstPoint = i;
                    sp[curPoint++] = makeZMinPoint(x, i);
                    // The next point is the one on the surface
                    sp[curPoint++] = screenPoints[x][i];

                    idle = false;
                }
            }
            else {
                // Look for a stop point on the surface
                if ((i == yEnd) || screenPoints[x][i].isHoleValue() ||
                                   screenPoints[xp][i].isHoleValue())
                {
                    // Time to end this polygon
                    if (screenPoints[x][i].isHoleValue() ||
                        screenPoints[xp][i].isHoleValue())
                    {
                        lastPoint = i - 1;
                    }
                    else {
                        lastPoint = i;
                    }
                    // Don't bother with single line polygons
                    if (firstPoint != lastPoint) {
                        // The next point is the one on the surface
                        sp[curPoint++] = screenPoints[x][lastPoint];

                        // Now the point on the axis
                        sp[curPoint++] = makeZMinPoint(x, lastPoint);

                        // Time to actually draw the polygon
                        fillPointArrays(sp, curPoint);
                        if (elevation.isShaded()) {
                            gc.setColor(elevation.getShadedBottomColor());
                            gc.fillPolygon(xPoints, yPoints, curPoint);
                        }
                        else if (elevation.isMeshed()) {
                            if (!elevation.isTransparent()) {
                                drawFacet.polyErase(xPoints, yPoints,
                                                    curPoint, 0);
                            }
                        }
                        // Also draw outline
                        gc.setColor(elevation.getMeshBottomColor());
                        gc.drawPolygon(xPoints, yPoints, curPoint);
                    }
                    curPoint = 0;
                    idle = true;
                }
                else {
                    // Just add the point to the list
                    sp[curPoint++] = screenPoints[x][i];
                }
            }
        }
    }

    // 2: If necessary, draw the drop-down vectors
    if (elevation.isMeshed()) {
	    JCSurface surface = area.getSurface();
        for (int i = yBegin; i <= yEnd; i++) {
            boolean[] xMesh = surface.getXMesh();
            if (surface.isXMeshShowing()          &&
                xMesh[i]                          &&
                !screenPoints[x][i].isHoleValue() &&
                !screenPoints[xp][i].isHoleValue())
            {
                if (((i > yBegin) && !screenPoints[x][i-1].isHoleValue()) ||
                    ((i < yEnd)   && !screenPoints[x][i+1].isHoleValue()))
                {
                    ScreenPoint dp0 = makeZMinPoint(x, i);
                    // The next point is the one on the surface
                    ScreenPoint dp1 = screenPoints[x][i];

                    gc.setColor(elevation.getMeshBottomColor());
                    gc.drawLine(dp0.getXPixel(), dp0.getYPixel(),
                                dp1.getXPixel(), dp1.getYPixel());
                }
            }
        }
    }
}

/**
 * Draws the portion of the skirt that is on the bottom of the cube (i.e. at
 * the <i>z</i> axis min).
 *
 * @param xBegin
 * @param xEnd
 * @param yBegin
 * @param yEnd
 */
protected void drawBottomOfSkirt(int xBegin, int xEnd, int yBegin, int yEnd)
{
    ScreenPoint []      sp       = new ScreenPoint[4];
	Chart3dDataView     dataView = data.getDataView();
	JCChart3dArea       area     = dataView.getChart3d().getChart3dArea();

	if (drawFacet == null) {
		return;
	}

    for (int x = xBegin; x <= xEnd; x++) {
        for (int y = yBegin; y <= yEnd; y++) {
            if (!JCChart3dUtil.holeInFacet(data, x, y)) {

                sp[0] = makeZMinPoint(x, y);
                sp[1] = makeZMinPoint(x, y + 1);
                sp[2] = makeZMinPoint(x + 1, y + 1);
                sp[3] = makeZMinPoint(x + 1, y);

                fillPointArrays(sp, sp.length);
	            JCElevation elevation = dataView.getElevation();
                Graphics    gc        = drawFacet.getGC();
                if (elevation.isShaded()) {
                    gc.setColor(elevation.getShadedBottomColor());
                    gc.fillPolygon(xPoints, yPoints, 4);
                }
                if (elevation.isMeshed()) {
                    if (!elevation.isShaded() && !elevation.isTransparent()) {
                        // Erase this facet to provide occlusion
						gc.setColor(area.getPlotCube().getDrawBackground());
                        gc.fillPolygon(xPoints, yPoints, 4);
                     }

                    JCSurface surface = area.getSurface();
	                boolean[] xMesh   = surface.getXMesh();
	                boolean[] yMesh   = surface.getYMesh();

                    boolean drawX1 = surface.isXMeshShowing() && xMesh[y];
                    boolean drawX2 = surface.isXMeshShowing() && xMesh[y + 1];
                    boolean drawY1 = surface.isYMeshShowing() && yMesh[x];
                    boolean drawY2 = surface.isYMeshShowing() && yMesh[x + 1];

                    gc.setColor(elevation.getMeshBottomColor());
                    if (drawX1 && drawX2 && drawY1 && drawY2) {
                        gc.drawPolygon(xPoints, yPoints, 4);
                    }
                    else {
                        if (drawX1) {
                            gc.drawLine(sp[3].getXPixel(), sp[3].getYPixel(),
                                        sp[0].getXPixel(), sp[0].getYPixel());
                        }
                        if (drawX2) {
                            gc.drawLine(sp[1].getXPixel(), sp[1].getYPixel(),
                                        sp[2].getXPixel(), sp[2].getYPixel());
                        }
                        if (drawY1) {
                            gc.drawLine(sp[0].getXPixel(), sp[0].getYPixel(),
                                        sp[1].getXPixel(), sp[1].getYPixel());
                        }
                        if (drawY2) {
                            gc.drawLine(sp[2].getXPixel(), sp[2].getYPixel(),
                                        sp[3].getXPixel(), sp[3].getYPixel());
                        }
                    }
                }
            }
        }
    }
}

/**
 * Draws the solid portion of the graph (if specified) by drawing the
 * appropriate <i>x</i> and <i>y</i> skirt and finally the bottom of the skirt.  Note that
 * skirts for holes are drawn separately.
 *
 * @param gridRegion
 */
protected void drawSolid(GridRegion gridRegion)
{
	Chart3dDataView     dataView = data.getDataView();
	JCChart3dArea       area = dataView.getChart3d().getChart3dArea();

	if (!area.getSurface().isSolid() ||
	    dataView.getChartType() != JCChart3d.SURFACE)
	{
		return;
	}

	int xBegin = gridRegion.getXBegin();
	int xEnd   = gridRegion.getXEnd();
	int yBegin = gridRegion.getYBegin();
	int yEnd   = gridRegion.getYEnd();

	switch(area.getTransform().getProcessingOrder() & 15) {
	case Transform.EDGE_XMAX:
		drawYSkirt(xEnd + 1, xEnd, yBegin, yEnd + 1);
		break;
	case Transform.EDGE_XMIN:
		drawYSkirt(xBegin, xBegin + 1, yBegin, yEnd + 1);
		break;
	case Transform.EDGE_YMAX:
		drawXSkirt(yEnd + 1, yEnd, xBegin, xEnd + 1);
		break;
	case Transform.EDGE_YMIN:
		drawXSkirt(yBegin, yBegin + 1, xBegin, xEnd + 1);
		break;
	case Transform.CORNER_XMAX_YMAX:
		drawXSkirt(yEnd + 1, yEnd, xBegin, xEnd + 1);
		drawYSkirt(xEnd + 1, xEnd, yBegin, yEnd + 1);
		break;
	case Transform.CORNER_XMAX_YMIN:
		drawXSkirt(yBegin, yBegin + 1, xBegin, xEnd + 1);
		drawYSkirt(xEnd + 1, xEnd, yBegin, yEnd + 1);
		break;
	case Transform.CORNER_XMIN_YMAX:
		drawXSkirt(yEnd + 1, yEnd, xBegin, xEnd + 1);
		drawYSkirt(xBegin, xBegin + 1, yBegin, yEnd + 1);
		break;
	case Transform.CORNER_XMIN_YMIN:
		drawXSkirt(yBegin, yBegin + 1, xBegin, xEnd + 1);
		drawYSkirt(xBegin, xBegin + 1, yBegin, yEnd + 1);
		break;
	}

	// For upside-down viewing, draw the base plate
	if ((area.getTransform().getProcessingOrder() & Transform.EDGE_ZMIN) != 0) {
	    drawBottomOfSkirt(xBegin, xEnd, yBegin, yEnd);
	}
}

}
