/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Transform.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:12:58 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d;

import java.awt.Graphics;
import java.awt.Rectangle;
import javax.vecmath.Matrix4d;
import javax.vecmath.Point4d;



/*
 *	This file deals with coordinate space conversion.
 *	The following conceptual coordinate spaces are used:
 *		graph		-	as specified by the user
 *		absolute	-	mapped onto the base cube
 *						and scaled
 *		world		-	after rotation
 *		perspective	-	after perspective
 *		projection	-	projected onto the X-Y plane
 *		screen		-	mapped onto the available screen space
 *		text		-	like screen, but only concerned with ratios,
 *						i.e. position is unimportant
 *
 *	The following set of mappings occurs:
 *
 *            graph
 *              |
 *              |	Map onto the base cube and scale.
 *              |
 *              V
 *           absolute
 *              |
 *              |	Apply rotation.
 *              |
 *              V
 *            world
 *              |
 *              |	Apply perspective transformation.
 *              |
 *              V
 *          perspective
 *              |
 *              |	Drop the Z coordinate.
 *              |
 *              V
 *          projection
 *              |
 *              |	Map onto the available screen space.
 *              |
 *              V
 *            screen
 *
 */

public class Transform implements java.io.Serializable {

/**  */
public final static double EDGE_PAD = 8.0;

// Resulting cases for viewpoint positioning
//
//   6 | 4 | 5
//  -----------
//   2 | 0 | 1
//  -----------
//  10 | 8 | 9

/**  */
public final static int FACE_ON   = 0;
/**  */
public final static int EDGE_XMAX = 1;
/**  */
public final static int EDGE_XMIN = 2;
/**  */
public final static int EDGE_YMAX = 4;
/**  */
public final static int EDGE_YMIN = 8;
/**  */
public final static int EDGE_ZMAX = 16;
/**  */
public final static int EDGE_ZMIN = 32;

/**  */
public final static int CORNER_XMAX_YMAX = 5;
/**  */
public final static int CORNER_XMIN_YMAX = 6;
/**  */
public final static int CORNER_XMAX_YMIN = 9;
/**  */
public final static int CORNER_XMIN_YMIN = 10;

/**  */
public final static int ROTATE_NONE = 1;
/**  */
public final static int ROTATE_90   = 2;
/**  */
public final static int ROTATE_180  = 3;
/**  */
public final static int ROTATE_270  = 4;

/**  */
protected Matrix4d absolute      = new Matrix4d();    // Data to unit cube
/**  */
protected Matrix4d rotation      = new Matrix4d();    // World to viewer
/**  */
protected Matrix4d perspective   = new Matrix4d();    // Perspective transform
/**  */
protected Matrix4d screen        = new Matrix4d();    // To screen space
/**  */
protected Matrix4d view          = new Matrix4d();    // Combined viewing (v = a*r*p*s)
/**  */
protected Matrix4d textView      = new Matrix4d();    // Combined viewing (t = r*p*s)
/**  */
protected Matrix4d iAbsolute     = new Matrix4d();    // Inverse of absolute
/**  */
protected Matrix4d iRotation     = new Matrix4d();    // Inverse of rotation
/**  */
protected Matrix4d iScreen       = new Matrix4d();    // Inverse of screen

/**  */
protected Point4d eyePosition     = null;
/**  */
protected JCBox   currentViewport = null;
/**  */
protected int     processingOrder = 0;

/**  */
protected JCChart3dArea area = null;

/**  */
static final long serialVersionUID = -5542417106751358122L;

/**  */
public Transform()
{
	absolute.setIdentity();
	rotation.setIdentity();
	perspective.setIdentity();
	screen.setIdentity();
	view.setIdentity();
	textView.setIdentity();
	iAbsolute.setIdentity();
	iRotation.setIdentity();
	iScreen.setIdentity();
}

/**  */
public Transform(JCChart3dArea area)
{
    this();
	this.area = area;
}

/**
 * Gets the parent <code>chart3d</code> area.
 *
 * @return The parent <code>JCChart3dArea</code> object
 */
public JCChart3dArea getChart3dArea()
{
	return(area);
}

/**
 * Sets the parent <code>chart3d</code> area
 *
 * @param area
 */
public void setChart3dArea(JCChart3dArea area)
{
	this.area = area;
}

/**  */
public Matrix4d getAbsolute()
{
	return(absolute);
}

/**  */
public Matrix4d getRotation()
{
	return(rotation);
}

/**  */
public Matrix4d getPerspective()
{
	return(perspective);
}

/**  */
public Matrix4d getScreen()
{
	return(screen);
}

/**  */
public Matrix4d getView()
{
	return(view);
}

/**  */
public Matrix4d getTextView()
{
	return(textView);
}

/**  */
public Matrix4d getInverseAbsolute()
{
	return(iAbsolute);
}

/**  */
public Matrix4d getInverseRotation()
{
	return(iRotation);
}

/**  */
public Matrix4d getInverseScreen()
{
	return(iScreen);
}

/**  */
public Point4d getEyePosition()
{
    if (eyePosition == null) {
        return (new Point4d());
    }
	return(eyePosition);
}

/**  */
public JCBox getCurrentViewport()
{
	return(currentViewport);
}

/**  */
protected int getProcessingOrder()
{
	return(processingOrder);
}

/**
 * Rotates the input matrix by a given predefined angle enum
 * about the <code>X</code>, <code>Y</code>, or <code>Z</code> axis.
 *
 * @param m
 * @param rotate
 * @param axis
 */
public static void rotateMatrix(Matrix4d m, int rotate, int axis)
{
	

	if (rotate == ROTATE_NONE) {
		// Nothing to do
		return;
	}

	double t;
	switch(axis) {
	case JCAxis.AXIS_X:
		if (rotate == ROTATE_90) {
			t = -m.m01; m.m01 = m.m02; m.m02 = t;
			t = -m.m11; m.m11 = m.m12; m.m12 = t;
			t = -m.m21; m.m21 = m.m22; m.m22 = t;
			t = -m.m31; m.m31 = m.m32; m.m32 = t;
		}
		else if (rotate == ROTATE_270) {
			t = m.m01; m.m01 = -m.m02; m.m02 = t;
			t = m.m11; m.m11 = -m.m12; m.m12 = t;
			t = m.m21; m.m21 = -m.m22; m.m22 = t;
			t = m.m31; m.m31 = -m.m32; m.m32 = t;
		}
		else if (rotate == ROTATE_180) {
			m.m01 = -m.m01; m.m02 = -m.m02;
			m.m11 = -m.m11; m.m12 = -m.m12;
			m.m21 = -m.m21; m.m22 = -m.m22;
			m.m31 = -m.m31; m.m32 = -m.m32;
		}
		break;
	case JCAxis.AXIS_Y:
		if (rotate == ROTATE_90) {
			t = m.m00; m.m00 = -m.m02; m.m02 = t;
			t = m.m10; m.m10 = -m.m12; m.m12 = t;
			t = m.m20; m.m20 = -m.m22; m.m22 = t;
			t = m.m30; m.m30 = -m.m32; m.m32 = t;
		}
		else if (rotate == ROTATE_180) {
			m.m00 = -m.m00; m.m02 = -m.m02;
			m.m10 = -m.m10; m.m12 = -m.m12;
			m.m20 = -m.m20; m.m22 = -m.m22;
			m.m30 = -m.m30; m.m32 = -m.m32;
		}
		else if (rotate == ROTATE_270) {
			t = -m.m00; m.m00 = m.m02; m.m02 = t;
			t = -m.m10; m.m10 = m.m12; m.m12 = t;
			t = -m.m20; m.m20 = m.m22; m.m22 = t;
			t = -m.m30; m.m30 = m.m32; m.m32 = t;
		}
	    break;
	case JCAxis.AXIS_Z:
		if (rotate == ROTATE_90) {
			t = -m.m00; m.m00 = m.m01; m.m01 = t;
			t = -m.m10; m.m10 = m.m11; m.m11 = t;
			t = -m.m20; m.m20 = m.m21; m.m21 = t;
			t = -m.m30; m.m30 = m.m31; m.m31 = t;
		}
		else if (rotate == ROTATE_180) {
			m.m00 = -m.m00; m.m01 = -m.m01;
			m.m10 = -m.m10; m.m11 = -m.m11;
			m.m20 = -m.m20; m.m21 = -m.m21;
			m.m30 = -m.m30; m.m31 = -m.m31;
		}
		else if (rotate == ROTATE_270) {
			t = m.m00; m.m00 = -m.m01; m.m01 = t;
			t = m.m10; m.m10 = -m.m11; m.m11 = t;
			t = m.m20; m.m20 = -m.m21; m.m21 = t;
			t = m.m30; m.m30 = -m.m31; m.m31 = t;
		}
	    break;
	}
}

/**
 * Rotates the input matrix by a given angle about the <i>X</i>, <i></i>Y, or
 * <i>Z</i> axis.
 *
 * @param m
 * @param angle
 * @param axis
 * @param doFast
 */
public static void rotateMatrix(Matrix4d m, double angle, int axis,
				                   boolean doFast)
{
	

	if (angle == 0.0) {
		// Nothing to do
		return;
	}

	double radAngle = JCChart3dUtil.degToRad(angle);
	double s = Math.sin(radAngle);
	double c = Math.cos(radAngle);
	if (doFast) {
		double t;
	    switch(axis) {
	    case JCAxis.AXIS_X:
			t = m.m01;
			m.m01 = c * t + s * m.m02;
			m.m02 = c * m.m02 - s * t;
			t = m.m11;
			m.m11 = c * t + s * m.m12;
			m.m12 = c * m.m12 - s * t;
			t = m.m21;
			m.m21 = c * t + s * m.m22;
			m.m22 = c * m.m22 - s * t;
			t = m.m31;
			m.m31 = c * t + s * m.m32;
			m.m32 = c * m.m32 - s * t;
		    break;
	    case JCAxis.AXIS_Y:
			t = m.m00;
			m.m00 = c * t - s * m.m02;
			m.m02 = c * m.m02 + s * t;
			t = m.m10;
			m.m10 = c * t - s * m.m12;
			m.m12 = c * m.m12 + s * t;
			t = m.m20;
			m.m20 = c * t - s * m.m22;
			m.m22 = c * m.m22 + s * t;
			t = m.m30;
			m.m30 = c * t - s * m.m32;
			m.m32 = c * m.m32 + s * t;
		    break;
	    case JCAxis.AXIS_Z:
			t = m.m00;
			m.m00 = c * t + s * m.m01;
			m.m01 = c * m.m01 - s * t;
			t = m.m10;
			m.m10 = c * t + s * m.m11;
			m.m11 = c * m.m11 - s * t;
			t = m.m20;
			m.m20 = c * t + s * m.m21;
			m.m21 = c * m.m21 - s * t;
			t = m.m30;
			m.m30 = c * t + s * m.m31;
			m.m31 = c * m.m31 - s * t;
		    break;
		}
	}
	else {
	    Matrix4d rotationMatrix = new Matrix4d();
	    switch(axis) {
	    case JCAxis.AXIS_X:
		    rotationMatrix.rotX(radAngle);
		    break;
	    case JCAxis.AXIS_Y:
		    rotationMatrix.rotY(radAngle);
		    break;
	    case JCAxis.AXIS_Z:
		    rotationMatrix.rotZ(radAngle);
		    break;
		}
	    m.mul(rotationMatrix);
	}
}

/**
 * Scales the input matrix by the given scale factor in <i>x</i>, <i>y</i>, and
 * <i>z</i>.  If <code>doFast</code> is true, we assume that all but one of the
 * scale factors is equal to 1 and thus scaling can be optimized.
 *
 * @param m
 * @param sx
 * @param sy
 * @param sz
 */
public static void scaleMatrix(Matrix4d m, double sx, double sy, double sz)
{
	

	// Things can be optimized if two of the three scale factors are one
	if ((sx == 1.0 && sy == 1.0) ||
	    (sx == 1.0 && sz == 1.0) ||
		(sy == 1.0 && sz == 1.0))
	{
		if (sy == -1.0) {
		   // The y value is usually scaled the most (and usually by -1)
		   m.m01 = -m.m01;
		   m.m11 = -m.m11;
		   m.m21 = -m.m21;
		   m.m31 = -m.m31;
		}
	    else if (sy != 1.0) {
		    m.m01 *= sy;
		    m.m11 *= sy;
		    m.m21 *= sy;
		    m.m31 *= sy;
		}
	    else if (sx != 1.0) {
		    m.m00 *= sx;
		    m.m10 *= sx;
		    m.m20 *= sx;
		    m.m30 *= sx;
	    }
		else if (sz != 1.0) {
		    m.m02 *= sz;
		    m.m12 *= sz;
		    m.m22 *= sz;
		    m.m32 *= sz;
		}
	}
	else {
		Matrix4d scaleMatrix = new Matrix4d();
	    scaleMatrix.setIdentity();
	    scaleMatrix.m00 = sx;
	    scaleMatrix.m11 = sy;
	    scaleMatrix.m22 = sz;
	    m.mul(scaleMatrix);
	}
}

/**
 * Translates the input matrix by the given translation in <i>x</i>, <i>y</i>,
 * and <i>z</i>.
 *
 * @param m
 * @param tx
 * @param ty
 * @param tz
 */
public static void translateMatrix(Matrix4d m, double tx, double ty, double tz)
{
	

	if (tx != 0.0) {
		m.m03 += m.m00 * tx;
		m.m13 += m.m10 * tx;
		m.m23 += m.m20 * tx;
		m.m33 += m.m30 * tx;
	}
	if (ty != 0.0) {
		m.m03 += m.m01 * ty;
		m.m13 += m.m11 * ty;
		m.m23 += m.m21 * ty;
		m.m33 += m.m31 * ty;
	}
	if (tz != 0.0) {
		m.m03 += m.m02 * tz;
		m.m13 += m.m12 * tz;
		m.m23 += m.m22 * tz;
		m.m33 += m.m32 * tz;
	}
}

/**  */
public static void transformPoint(Matrix4d m, Point4d p)
{
	m.transform(p);
	p.project(p);   // Normalize back to (x, y, z, 1)
}

/**
 * Adjusts the viewport, based on the scale and translations.
 *
 * @param vp The current viewport
 * @param scale The scale to apply to the viewport
 * @param horizontalShift The horizontal shift to apply to the viewport
 * @param verticalShift The vertical shift to apply to the viewport
*/
public void adjustForUserViewport(JCBox vp, double scale,
			                      double horizontalShift,
								  double verticalShift)
{
	vp.x      += vp.width  * horizontalShift;
	vp.y      += vp.height * verticalShift;
	vp.width  /= scale;
	vp.height /= scale;
}

/**
 * Create viewpoint position vector with eye at least at double the unit cube
 * and do an inverse rotation to get it into absolute space.
 */
public Point4d computeEyePosition()
{
	// Create viewpoint position vector with eye at least at double the unit cube
	Point4d p = new Point4d(0.0, 0.0,
                            2.0 * area.getView3d().getPerspective(), 1.0);
	transformPoint(iRotation, p);
	return(p);
}

/**
 * Squishes data into a ( +/-1, +/-1, +/-1) cube.
 *
 * @param dataport
 */
protected void computeAbsoluteMatrix(JCBox dataport)
{
	boolean     isProj   = area.getChart3d().isProjection();
	JCPlotCube  plotcube = area.getPlotCube();
	double      xScale   = plotcube.getXNormalizedScale();
	double      yScale   = plotcube.getYNormalizedScale();
	double      zScale   = plotcube.getZNormalizedScale();

	double sx = 2.0 * xScale / dataport.width;
	double sy = 2.0 * yScale / dataport.height;
	double sz = (dataport.depth == 0.0) ? 1.0 : 2.0 * zScale / dataport.depth;

	double dx = -xScale - dataport.x * sx;
	double dy = -yScale - dataport.y * sy;
	double dz = -zScale - dataport.z * sz;

	absolute.setIdentity();
	absolute.m00 = sx;
	absolute.m11 = sy;
	absolute.m22 = isProj ? 0.0 : sz;
	absolute.m03 = dx;
	absolute.m13 = dy;
	absolute.m23 = isProj ? 0.0 : dz;

	iAbsolute.setIdentity();
	iAbsolute.m00 = 1.0/sx;
	iAbsolute.m11 = 1.0/sy;
	iAbsolute.m22 = isProj ? 0.0 : 1.0/sz;
	iAbsolute.m03 = -dx/sx;
	iAbsolute.m13 = -dy/sy;
	iAbsolute.m23 = isProj ? 0.0 : -dz/sz;
}

/**  */
public void computeRotationMatrix()
{
	boolean  isProj = area.getChart3d().isProjection();
	JCView3d view3d = area.getView3d();

	rotation.setIdentity();
	if (!isProj) {
	   rotateMatrix(rotation, -90.0, JCAxis.AXIS_X, true);
	   rotateMatrix(rotation, view3d.getXRotation(), JCAxis.AXIS_X, false);
	   rotateMatrix(rotation, view3d.getYRotation(), JCAxis.AXIS_Y, false);
	   rotateMatrix(rotation, view3d.getZRotation(), JCAxis.AXIS_Z, false);
	}

	iRotation.setIdentity();
	if (!isProj) {
	   rotateMatrix(iRotation, -view3d.getZRotation(), JCAxis.AXIS_Z, false);
	   rotateMatrix(iRotation, -view3d.getYRotation(), JCAxis.AXIS_Y, false);
	   rotateMatrix(iRotation, -view3d.getXRotation(), JCAxis.AXIS_X, false);
	   rotateMatrix(iRotation, 90.0, JCAxis.AXIS_X, true);
	}
}

/**  */
protected void computePerspectiveMatrix()
{
	perspective.setIdentity();
	if (!area.getChart3d().isProjection()) {
	   perspective.m32 = -1.0 / (2.0 * area.getView3d().getPerspective());
	}
}

/**
 * Adjusts projected image to show only what which lies within the viewport.
 * Note: To keep the aspect ratio fixed the viewport should be square.  So we
 * use the smallest dimension (width or height) to represent the square
 * viewport.
 *
 */
protected void computeScreenMatrix()
{
	double  sx, sy;
	double	dx, dy;
	double  length, vlength;

	JCBox viewport = new JCBox(currentViewport.x, currentViewport.y,
							   currentViewport.width, currentViewport.height);
	JCViewport userViewport = area.getViewport();
	adjustForUserViewport(viewport, userViewport.getScale(),
	                      userViewport.getHorizontalShift(),
						  userViewport.getVerticalShift());

	screen.setIdentity();
	iScreen.setIdentity();
	if (area == null) {
		return;
	}
    Rectangle rect = area.getDrawingArea();
	double width  = rect.getWidth();
    double height = rect.getHeight();

	if ((width / viewport.width) * viewport.height < height) {
		length  = width;
		vlength = viewport.width;
	}
	else {
		length  = height;
		vlength = viewport.height;
	}

	if (length <= EDGE_PAD) {
		sx = length / vlength;
	}
	else {
		sx = (length - EDGE_PAD) / vlength;
	}

	sy = -sx;

	dx = (rect.getX() - viewport.x * sx + (width - viewport.width   * sx) / 2.0);
	dy = (rect.getY() - viewport.y * sy + (height - viewport.height * sy) / 2.0);

	screen.m00 = sx;
	screen.m11 = sy;
	screen.m03 = dx;
	screen.m13 = dy;

	iScreen.m00 = 1.0/sx;
	iScreen.m11 = 1.0/sy;
	iScreen.m03 = -dx/sx;
	iScreen.m13 = -dy/sy;
}

/**  */
public void computeTransformations(Graphics gc)
{
	computeRotationMatrix();
	eyePosition = computeEyePosition();
	computePerspectiveMatrix();

	JCBox dataport  = new JCBox();
	currentViewport = new JCBox();
	area.getAnno().calcAnno(gc, currentViewport, dataport);
	computeScreenMatrix();
	computeAbsoluteMatrix(dataport);
	computeViewMatrix();
}

/**
 * Computes a combined xform matrix to take an absolute coordinate system
 * vector and xform it to screen coordinates in one step. Also computes text
 * viewing matrix: </i>textView = [screen][perspective][rotation]</i>.  This is a
 * special 'text version' of <i>xrt->vm</i> since the text does not need to be
 * transformed from data space first (ie. no absolute).
 *
 */
protected void computeViewMatrix()
{
	textView.mul(screen, perspective);
	textView.mul(rotation);
	view.mul(textView, absolute);
}

/**
 * Finds the processing order for the grid facets based on the relative position
 * of the viewpoint to the grid.
 *
 */
protected void calcProcessingOrder()
{
	processingOrder = 0;

	JCAxis xAxis = area.getAxis(JCAxis.AXIS_X);
	JCAxis yAxis = area.getAxis(JCAxis.AXIS_Y);
	JCAxis zAxis = area.getAxis(JCAxis.AXIS_Z);

	// Transform the data extents into the normalized cube space
	double normXMin = xAxis.getDataportMin() * absolute.m00 + absolute.m03;
	double normXMax = xAxis.getDataportMax() * absolute.m00 + absolute.m03;
	double normYMin = yAxis.getDataportMin() * absolute.m11 + absolute.m13;
	double normYMax = yAxis.getDataportMax() * absolute.m11 + absolute.m13;

	if (eyePosition.x >= normXMax ) {
		processingOrder |= EDGE_XMAX;
	}
	if (eyePosition.x <= normXMin ) {
		processingOrder |= EDGE_XMIN;
	}
	if (eyePosition.y >= normYMax ) {
		processingOrder |= EDGE_YMAX;
	}
	if (eyePosition.y <= normYMin ) {
		processingOrder |= EDGE_YMIN;
	}

	ScreenPoint zpts0 = new ScreenPoint();
	JCChart3dUtil.getScreenPoint(area.getChart3d(), xAxis.getDataportMax(),
	                             yAxis.getDataportMin(), zAxis.getDataportMin(),
								 zAxis.getMin(), zpts0);
	ScreenPoint zpts1 = new ScreenPoint();
	JCChart3dUtil.getScreenPoint(area.getChart3d(), xAxis.getDataportMin(),
	                             yAxis.getDataportMin(), zAxis.getDataportMin(),
								 zAxis.getMin(), zpts1);
	ScreenPoint zpts2 = new ScreenPoint();
	JCChart3dUtil.getScreenPoint(area.getChart3d(), xAxis.getDataportMin(),
	                             yAxis.getDataportMax(), zAxis.getDataportMin(),
								 zAxis.getMin(), zpts2);

	if (JCChart3dUtil.gridBackfacing(zpts0, zpts1, zpts2)) {
		processingOrder |= EDGE_ZMIN;
	}
	else {
		processingOrder |= EDGE_ZMAX;
	}
}

}
