/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Base3dPointDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:02 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.data;

import com.klg.jclass.chart3d.*;
import com.klg.jclass.chart3d.event.*;
import com.klg.jclass.util.JCListenerList;

import java.util.Enumeration;
import java.io.Serializable;
import javax.vecmath.Point3d;

/** 
 * This class is the base for any data source that chooses to
 * store data internally using an array of <code>Point3d</code> objects.
 * 
 */
public class Base3dPointDataSource extends Base3dDataSource
			                       implements Chart3dPointDataModel,
		                                      LabelledChart3dPointDataModel,
											  Cloneable,
											  Serializable {

/** Array of <code>Point3d</code> objects to be plotted on the chart.
 * 
 */
protected Point3d[][] points;

/** Name of this datasource.
 * 
 */
protected String dataSourceName;

/** List of interested listeners.
 * 
 */
protected JCListenerList chartDataListeners;

/** List of Series Labels.
 * 
 */
protected String[] seriesLabels;

/** 
 * Constructor for a <code>Base3dPointDataSource</code> where the data is already known,
 * includes an argument for the datasource name.
 * 
 * @param name Name <code>String</code> of this datasource (optional, can be 
 *     null).
 * @param points Array of Point3d objects to be plotted on the chart.
 */
public Base3dPointDataSource(String name, Point3d[][] points)
{
	super();
	if (name != null) {
		this.dataSourceName = name;
	}
	this.points = points;

	int type = Chart3dDataEvent.RESET;
	JCData3dPointIndex index = new JCData3dPointIndex(JCData3dPointIndex.ALL,
													  JCData3dPointIndex.ALL);
	fireChart3dDataEvent(new Chart3dPointDataEvent(this, type, index));
}

/** 
 * Sets the name of this datasource
 * .
 * 
 * @param name Name <code>String</code> of this datasource
 */
public void setDataSourceName(String name)
{
	this.dataSourceName = name;
	int type = Chart3dDataEvent.RELOAD_DATA_SOURCE_NAME;
	JCData3dPointIndex index = new JCData3dPointIndex(JCData3dPointIndex.ALL,
													  JCData3dPointIndex.ALL);
	fireChart3dDataEvent(new Chart3dPointDataEvent(this, type, index));
}//setName

/** 
 * Retrieves the name of this datasource
 * .
 * 
 * @return Name <code>String</code> of this datasource
 */
public String getDataSourceName()
{
	return(dataSourceName);
}//getName

/** 
 * Sets the data values for this chart in the form of a two-dimensional
 * array of <code>Point3d</code> objects
 * .
 * 
 * @param points Two-dimensional array of <code>Point3d</code> objects
 */
public void setPoints(Point3d[][] points)
{
	this.points = points;
	int type = Chart3dDataEvent.RELOAD;
	JCData3dPointIndex index = new JCData3dPointIndex(JCData3dPointIndex.ALL,
													  JCData3dPointIndex.ALL);
	fireChart3dDataEvent(new Chart3dPointDataEvent(this, type, index));
}//setPoints

/** 
 * Retrieves a doubly subscripted array of points (arranged by series).
 * 
 * @return A doubly subscripted array of points (arranged by series).
 */
public Point3d[][] getPoints()
{
	return(points);
}//getPoints

/** 
 * Retrieves the number of data series this datasource holds.
 * 
 * @return <code>int</code> the number of dataseries.
 */
public int getNumSeries() {
	if (points == null) {
		return(0);
	}
	return(points.length);
}//getNumSeries

/** 
 * Retrieves the list of series labels
 * .
 * 
 * @return <code>seriesLabels</code> a <code>String</code> array
 */
public String[] getSeriesLabels()
{
	return(this.seriesLabels);
}//getSeriesLabels


/** 
 * Sets the list of series labels
 * .
 * 
 * @param seriesLabels a <code>String</code> array
 */
public void setSeriesLabels(String[] seriesLabels)
{
	this.seriesLabels = seriesLabels;
	int type = Chart3dPointDataEvent.RELOAD_ALL_SERIES_LABELS;
	JCData3dPointIndex index = new JCData3dPointIndex(JCData3dPointIndex.ALL,
													  JCData3dPointIndex.ALL);
	fireChart3dDataEvent(new Chart3dPointDataEvent(this, type, index));
}//setSeriesLabels

/** 
 * Overrides clone in class Object. This clone method retuns a deep copy of
 * all Base3dPointDataSource fields, except for the <code>JCListenerList</code>, which
 * is set to null.
 * 
 * <p>
 * Overrides @see Object#clone().
 * 
 * @throws java.lang.OutOfMemoryError - if there is not enough memory.
 * @return a clone of this instance.
 */
public Object clone()
{
	Base3dPointDataSource clone = null;

	try {
		clone = (Base3dPointDataSource)super.clone();

		// Make a deep copy of the data points.
		clone.points = (Point3d[][])points.clone();
		if (points != null) {
			for (int i = 0; i < points.length; i++) {
				// If the array is non-null, copy.
				clone.points[i] = (Point3d[])points[i].clone();
				if (points[i] != null) {
					for (int j = 0; j < points[i].length; j++) {
						clone.points[i][j] = new Point3d(points[i][j]);
					}
				}
			}
		}

		// Make a deep copy of the series labels.
		if (seriesLabels != null) {
			clone.seriesLabels = (String[])seriesLabels.clone();
			// String objects are immutable; no need to clone each label.
		}

		// The clone should not share the same chart data listeners of
		// it's prototype, unless ofcourse this is desired. So, by default,
		// set the list of listeners to null.
		clone.chartDataListeners = null;

	}
	catch (CloneNotSupportedException e) {
		// This should never happen.
		throw new RuntimeException("Base3dPointDataSource does not support cloneable.");
	}

	return clone;
}//clone

}//Base3dPointDataSource
