/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCData3dUtil.java,v $ $Revision: 1.2 $
// $Date: 2006-03-17 21:13:02 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.data;

import java.awt.geom.RectangularShape;

import java.util.Arrays;

import javax.swing.SwingConstants;
import javax.vecmath.Point3d;
import javax.vecmath.Vector3d;

import com.klg.jclass.util.JCNumberUtil;
import com.klg.jclass.chart3d.Chart3dDataModel;
import com.klg.jclass.chart3d.Chart3dGridDataModel;
import com.klg.jclass.chart3d.Chart3dPointDataModel;
import com.klg.jclass.chart3d.HoleValueChart3dDataModel;

/**  *
 */
public class JCData3dUtil implements SwingConstants {

/**  */
public static final int DATA_INTERPOLATE_LINEAR = 0;

/**  */
public static final int DATA_INTERPOLATE_CUBIC = 1;


 /**
  * Constructs a copy of the <code>Chart3dGridDataModel</code> source and returns it.
  * Only the <i>x</i>Grid, <i>y</i>Grid, and <i>z</i>Values are copied.
  *
  * @param source the data model to copy
  * @return a copy of the source <i>x</i>Grid, <i>y</i>Grid, and
  *     <i>z</i>Values
  */
 public static Chart3dGridDataModel createDataCopy(Chart3dGridDataModel source) {

	// Construct a Base3dGridDataSource, then clone it!
	Base3dGridDataSource dest = new Base3dGridDataSource(
										null,
										source.getXGrid(),
										source.getYGrid(),
										source.getZValues());

	return (Chart3dGridDataModel)dest.clone();
}

 /**
  * Copies the <i>x</i>Grid, <i>y</i>Grid, and <i>z</i>Values of the source argument into the
  * dest argument.  The source and dest must be compatible for this method
  * to work (in other words, the <i>x</i>Grid, <i>y</i>Grid, and <i>z</i>Values arrays of source and dest must be
  * non-null and of equal size).
  *
  * @param source
  * @param dest
  */
 public static void dataCopy(Chart3dGridDataModel source, Chart3dGridDataModel dest) {

	int x = getSizeX(source),
		y = getSizeY(source);

	double[]
		xGrid = source.getXGrid(),
		xCopy = new double[x],
		yGrid = source.getYGrid(),
		yCopy = new double[y];

	double[][]
		zValues = source.getZValues(),
		zCopy = new double[x][];

	System.arraycopy(xGrid, 0, xCopy, 0, x);
	System.arraycopy(yGrid, 0, yCopy, 0, y);

	for (int each = 0; each < zValues.length; each++) {
		int len = zValues[each].length;
		zCopy[each] = new double[len];
		System.arraycopy(zValues[each], 0, zCopy[each], 0, len);
	}
}

/**
 * Constructs an array of <i>z</i> values for the specified <code>Chart3dGridDataModel</code>
 * source where each <i>z</i> value is a measurement of light on the closed range
 * [0,1].  The amount of light at a given grid point has two components:
 * ambient and reflected.
 * <p>
 * The ambient component is constant for all grid points, and is specified
 * by the ambient argument (which must be on the range [0,1]).
 * <p>
 * The reflected component is the dot product of two vectors.  The first
 * is the light normal - a unit vector defined by the <code>sweepAngle</code> and <code>riseAngle</code>
 * arguments.  The <code>sweepAngle</code> is the position of the light source measured in
 * degrees from the positive <i>x</i> axis in the <i>xy</i> plane.  The <code>riseAngle</code> is the
 * position of the light source measured in degrees above the <i>xy</i> plane.  The
 * second is the surface normal - a unit vector computed at each individual
 * grid point.
 * <p>
 * The light at a given grid point is given by:
 * <i>light(x,y)  = ambient + (1-ambient) * lightNormalsurfaceNormal(x,y)</i>
 * <p>
 * The total amount of light at a given grid point is then adjusted for
 * overall brightness by:
 * <i>lightAdjusted(x,y) = brightness * light(x,y)</i>
 * where brightness is on the closed range [0,1].
 * <p>
 * The source <code>Chart3dGridDataModel</code> is cloned, and the shaded values are then
 * put into the array of <i>z</i>-values belogning to the clone.
 * <i>z</i>-values belong to the <code>Chart3dGridDataModel</code> dest.
 *
 * @param source the non-null data model used to compute the shaded data
 * @param sweepAngle
 * @param riseAngle
 * @param brightness
 * @param ambient
 * @return - <code>Chart3dGridDataModel</code> containing the same <i>x</i>
 *     and <i>y</i> grid as source, and <i>z</i>-values corresponding to the
 *     array of shaded values at each grid point
 */
public static Chart3dGridDataModel createShadedDataModel(
		Chart3dGridDataModel source,
		double sweepAngle,
		double riseAngle,
		double brightness,
		double ambient) {

	Chart3dGridDataModel dest = null;

	if (source instanceof Base3dGridDataSource) {
		Object clone  = ((Base3dGridDataSource)source).clone();
		dest = (Chart3dGridDataModel)clone;
	}
	else {
		// We have to construct a Chart3dGridDataModel ourselves.
		dest = createDataCopy(source);
	}

	createShadedDataModel(
						source,
						dest,
						sweepAngle,
						riseAngle,
						brightness,
						ambient);
	return dest;
}

/**
 * Constructs an array of <i>z</i> values for the specified <code>Chart3dGridDataModel</code>
 * source where each <i>z</i> value is a measurement of light on the closed range
 * [0,1].  The amount of light at a given grid point has two components:
 * ambient and reflected.
 * <p>
 * The ambient component is constant for all grid points, and is specified
 * by the ambient argument (which must be on the range [0,1]).
 * <p>
 * The reflected component is the dot product of two vectors.  The first
 * is the light normal - a unit vector defined by the <code>sweepAngle</code> and <code>riseAngle</code>
 * arguments.  The <code>sweepAngle</code> is the position of the light source measured in
 * degrees from the positive <i>x</i> axis in the <i>xy</i> plane.  The <code>riseAngle</code> is the
 * position of the light source measured in degrees above the <i>xy</i> plane.  The
 * second is the surface normal - a unit vector computed at each individual
 * grid point.
 * <p>
 * The light at a given grid point is given by:
 * <i>light(x,y)  = ambient + (1-ambient) * lightNormalsurfaceNormal(x,y)</i>
 * <p>
 * The total amount of light at a given grid point is then adjusted for
 * overall brightness by:
 * <i>lightAdjusted(x,y) = brightness * light(x,y)</i>
 * where brightness is on the closed range [0,1].
 * <p>
 * The resultant light for each grid point is then stored in the array of
 * <i>z</i>-values belogning to the <code>Chart3dGridDataModel</code> dest.
 *
 * @param source the non-null data model used to compute the shaded data
 * @param dest the non-null data model used to store the resultant shaded
 *     data
 * @param sweepAngle
 * @param riseAngle
 * @param brightness
 * @param ambient
 */
public static void createShadedDataModel(
		Chart3dGridDataModel source,
		Chart3dGridDataModel dest,
		double sweepAngle,
		double riseAngle,
		double brightness,
		double ambient) {

	if (getSizeX(source) < 2 || getSizeY(source) < 2) {
		throw new IllegalArgumentException(
			"Grid size must be atleast 2x2 to construct a shaded model."
		);
	}

	riseAngle = Math.toRadians(riseAngle);
	sweepAngle = Math.toRadians(sweepAngle);

	brightness = JCNumberUtil.clamp(brightness, 0.0, 1.0);
	ambient = JCNumberUtil.clamp(ambient, 0.0, 1.0);

	int	i1,
		i2,
		j1,
		j2,
		xSize = getSizeX(source),
		ySize = getSizeY(source);

	double
		dzx,
		dzy,
		sxy = Math.cos(riseAngle),
		sz = Math.sin(riseAngle),
		sy = sxy * Math.cos(sweepAngle),
		sx = sxy * Math.sin(sweepAngle),
		oneMinusAmbient = 1.0 - ambient,
		shadedValue,
		srcHoleValue = getHoleValue(source),
		destHoleValue = getHoleValue(dest);

	double[]
		xStep = new double[xSize-1],
		yStep = new double[ySize-1];

	Vector3d
		lightNormal = new Vector3d(sx, sy, sz),
		surfaceNormal = new Vector3d();

	double[] sourceXGrid = source.getXGrid();
	for (int each = 0; each < xSize - 1; each++) {
		xStep[each] = sourceXGrid[each+1] - sourceXGrid[each];
	}

	double[] sourceYGrid = source.getYGrid();
	for (int each = 0; each < ySize - 1; each++) {
		yStep[each] = sourceYGrid[each+1] - sourceYGrid[each];
	}

	double[][]
		zSource = source.getZValues(),
		zDest = dest.getZValues();

	for (int i = 0; i < xSize; i++) {
		if (i < xSize - 1) {
			i1 = i;
			i2 = i + 1;
		}
		else {
			i1 = i - 1;
			i2 = i;
		}

		for (int j = 0; j < ySize; j++) {
			if (j < ySize - 1) {
				j1 = j;
				j2 = j + 1;
			}
			else {
				j1 = j - 1;
				j2 = j;
			}

			if (zSource[i][j1] == srcHoleValue || zSource[i][j2] == srcHoleValue) {
				dzy = srcHoleValue;
			}
			else {
				dzy = zSource[i][j2] - zSource[i][j1];
			}

			if (zSource[i1][j] == srcHoleValue || zSource[i2][j] == srcHoleValue) {
				dzx = srcHoleValue;
			}
			else {
				dzx = zSource[i2][j] - zSource[i1][j];
			}

			if (dzx == srcHoleValue || dzy == srcHoleValue) {
				zDest[i][j] = destHoleValue;
			}
			else {

				// Add the ambient component of light.
				shadedValue = ambient;

				if (ambient < 1.0) {
					/* There is - possibly - a direct light component. Construct
					 * a surface normal for the current grid point by taking the
					 * cross product of two vector contained by the surface - then
					 * normalize it.
					 *
					 * First surface vector is along x axis: (dx, 0, dzx). Second
					 * surface vector is along y axis (0, dy, dzy). We could use
					 * Vector3d.cross(v1, v2) to compute the normal. However, the
					 * above cross product simplifies to (-dy*dzx, - dx*dzy, dx*dy).
					 * You do the math!
					 */
					surfaceNormal.set(
						-yStep[j1] * dzx,
						-xStep[i1] * dzy,
						 xStep[i1] * yStep[j1]);

					if (surfaceNormal.dot(lightNormal) > 0) {
						/* Then the light source normal is in the half space of
						 * the surface, and there is in fact a component of
						 * direct light.  Now we normalize the surface normal,
						 * and project it onto the the light normal to find
						 * the contribution of direct light.
						 */
						surfaceNormal.scale(1.0 / surfaceNormal.length());
						shadedValue += oneMinusAmbient * surfaceNormal.dot(lightNormal);
					}
				}

				// Finally adjust for brightness.
				shadedValue *= brightness;

				zDest[i][j] = shadedValue;
			}
		}
	}
}

/**
 * Constructs a data model with a data surface with the same <i>x</i> and <i>y</i>
 * grid values of source, but with "smoothed" <i>z</i> values.
 * <p>
 * The smoothing
 * algorithm is a centre weighted average algorithm which functions as
 * follows:
 * <br>
 * Each smoothed <i>z</i> value is a combination of itself, as well as a contribution
 * from all immediate neighbouring <i>z</i> values (values to the West, Northwest,
 * North, NorthWest, and so on).  The contribution that the candidate itself makes is
 * proportional to the <code>centreWeight</code> argument, which can be anywhere from 0.0
 * (no contribution) to 1.0 (100% contribution).  Any remaining contribution
 * is divided amongst all other contributors and is inversely proportional
 * to the distance from the each neighbour to the candidate. <br><B>Note:</b> Any data
 * holes will remain data holes, and will not contribuite to the smoothed values
 * of its neighbours.
 * <p>
 * The source data model is not affected by this method.
 *
 * @param source the data model used to construct the smoothed values
 * @param centreWeight
 * @throws IllegalArgumentException if there are fewer than 2x2 data values
 * @return a data model with the same <i>x</i>Grid and <i>y</i>Grid as
 *     source, but with the resultant "smoothed" <i>z</i>Values in its
 *     <i>z</i>Values array
 */
public static Chart3dGridDataModel createSmoothedDataModel(
		Chart3dGridDataModel source,
		double centreWeight) {

	Chart3dGridDataModel dest = null;

	if (source instanceof Base3dGridDataSource) {
		Object clone  = ((Base3dGridDataSource)source).clone();
		dest = (Chart3dGridDataModel)clone;
	}
	else {
		// We have to construct a Chart3dGridDataModel ourselves.
		dest = createDataCopy(source);
	}

	createSmoothedDataModel(source, dest, centreWeight);
	return dest;
}

/**
 * Constructs a data model with a data surface with the same <i>x</i> and <i>y</i>
 * grid values of source, but with "smoothed" <i>z</i> values.
 * <p>
 * The smoothing
 * algorithm is a centre weighted average algorithm which functions as follows:
 * <p>
 * Each smoothed <i>z</i> value is a combination of itself, as well as a contribution
 * from all immediate neighbouring <i>z</i> values (values to the West, Northwest,
 * North, NorthWest, and so on).  The contribution that the candidate itself makes is
 * proportional to the <code>centreWeight</code> argument which can be anywhere from 0.0
 * (no contribution) to 1.0 (100% contribution).  Any remaining contribution
 * is divided amongst all other contributors and is inversely proportional
 * to the distance from the each neighbour to the candidate. <br><B>Note:</b> Any data
 * holes will remain data holes, and will not contribuite to the smoothed values
 * of its neighbours.
 *
 * <p>
 * The source data model is not affected by this method.
 *
 * @param source the data model used to construct the smoothed values
 * @param dest the data model used to store the smoothed values
 * @param centreWeight
 * @throws IllegalArgumentException An IllegalArgumentException if there are fewer than 2x2 data values.
 */
public static void createSmoothedDataModel(
				Chart3dGridDataModel source,
				Chart3dGridDataModel dest,
				double centreWeight) {

	int	xSize = getSizeX(source),
		ySize = getSizeY(source);

	if (xSize < 2 || ySize < 2) {
		throw new IllegalArgumentException(
			"Grid size must be atleast 2x2 to construct a weighted average model."
		);
	}

	centreWeight = JCNumberUtil.clamp(centreWeight, 0.0, 1.0);
	if (centreWeight == 1.0) {
		// Then no smoothing is performed.  Copy the zValues of the source
		// into the destination.
		dataCopy(source, dest);
		return;
	}

	double
		srcHoleValue = getHoleValue(source),
		destHoleValue = getHoleValue(dest);

	double[]
		xGrid = source.getXGrid(),
		yGrid = source.getYGrid();

	double[]
		dxSquared = new double[xSize-1],
		dySquared = new double[ySize-1],
		distance9 = new double[9],
		zValue9 = new double[9];

	double[][] zValues = source.getZValues(),
			   zSmoothed = dest.getZValues();

	// Populate dxSquared with the square distance between consecutive
	// x grid points.
	double dx;
	for (int each = 0; each < xSize-1; each++) {
		dx = xGrid[each+1] - xGrid[each];
		dxSquared[each] = dx * dx;
	}

	// Populate dySquared with the square distance between consecutive
	// y grid points.
	double dy;
	for (int each = 0; each < ySize-1; each++) {
		dy = xGrid[each+1] - yGrid[each];
		dySquared[each] = dy * dy;
	}

	int edgeCode;

	for (int i = 0; i < xSize; i++) {

		for (int j = 0; j < ySize; j++) {

			/*
			 * If the data model has hole values, and the current
			 * grid point is a hole value, do no smoothing.
			 */
			if (zValues[i][j] == srcHoleValue) {
				zSmoothed[i][j] = destHoleValue;
				continue;
			}

			/*
			 * Populate our distanceArray matrix with the squared distance
			 * to adjacent grid points to the NORTH, EAST, SOUTH, and
			 * WEST.  But before we do that, fill it with zeroes, and
			 * reset the edge code to SwingConstants.CENTER.
			 */

			Arrays.fill(distance9, 0);
			Arrays.fill(zValue9, 0);

			edgeCode = CENTER;

			/*
			 * Check edge cases and for holes simultaneously.
			 */

			if (i > 0) {
				// The current x grid point is not on the west edge
				// of the grid. Fetch the distance to the adjacent
				// grid point to the west.
				edgeCode = edgeCode | WEST;
				distance9[WEST] = dxSquared[i-1];
				zValue9[WEST] = zValues[i-1][j];
			}

			if (i < xSize-1) {
				// The current x grid point is not on the east edge
				// of the grid and is not a hole value. Fetch the distance
				// to the adjacent grid point to the east.
				edgeCode = edgeCode | EAST;
				distance9[EAST] = dxSquared[i+1];
				zValue9[EAST] = zValues[i+1][j];
			}

			if (j > 0) {
				// The current y grid point is not on the south edge
				// of the grid. Fetch the distance to the adjacent
				// grid point to the south.
				edgeCode = edgeCode | SOUTH;
				distance9[SOUTH] = dySquared[j-1];
				zValue9[SOUTH] = zValues[i][j-1];
			}

			if (j < ySize-1) {
				// The current y grid point is not on the north edge
				// of the grid. Fetch the distance to the adjacent
				// grid point to the north.
				edgeCode = edgeCode | NORTH;
				distance9[NORTH] = dxSquared[i+1];
				zValue9[NORTH] = zValues[i][j+1];
			}

			/*
			 * Check corner cases.
			 */
			if ((edgeCode | NORTH_EAST) == 0) {
				// The north east grid point contributes.
				distance9[NORTH_EAST] = distance9[NORTH] + distance9[EAST];
				zValue9[NORTH_EAST] = zValues[i+1][j+1];
			}

			if ((edgeCode | NORTH_WEST) == 0) {
				// The north west grid point contributes.
				distance9[NORTH_WEST] = distance9[NORTH] + distance9[WEST];
				zValue9[NORTH_WEST] = zValues[i-1][j+1];
			}

			if ((edgeCode | SOUTH_EAST) == 0) {
				// The south east grid point contributes.
				distance9[SOUTH_EAST] = distance9[SOUTH] + distance9[EAST];
				zValue9[SOUTH_EAST] = zValues[i+1][j-1];
			}

			if ((edgeCode | SOUTH_WEST) == 0) {
				// The south west grid point contributes.
				distance9[SOUTH_WEST] = distance9[SOUTH] + distance9[WEST];
				zValue9[SOUTH_WEST] = zValues[i-1][j-1];
			}

			zValue9[CENTER] = zValues[i][j];

			zSmoothed[i][j] = computeInverseWeightedAverage(
					zValue9,
					distance9,
					centreWeight,
					CENTER,
					srcHoleValue);
		}
	}
}

/**
 * Generates a weighted average value for a data point based on its
 * own value, as well as contributions from values contained in the
 * values array.  The contribution each value makes is based upon the
 * formula:<br>
 * <i>(weightSum - value[i])
 * contribution[i] = ---------------------- X (1 - centreWeight)
 * weightSum</i>
 * <br>
 * except <i>contribution[centreIndex]</i>, which is: <br>
 * <i>contribution[centreIndex] = value[centreIndex] X centreWeight
 * </i>
 *
 * <p>
 * <code>weightedSum</code> is the sum of all weights except for <i>weights[centreIndex]</i>.
 * The weighted average is essentially just the sum of every contribution.
 * If <code>hasHoles</code> is <code>true</code>, all values which are equal to <code>holeValue</code> will
 * be ignored and will not impact the weighted average computation.
 * @return
 * @param values
 * @param weights
 * @param centreWeight
 * @param centreIndex
 * @param holeValue
 */
public static double computeInverseWeightedAverage(
		double[] values,
		double[] weights,
		double centreWeight,
		int centreIndex,
		double holeValue) {

	double
		sum = 0,
		weightedAverage = 0,
		oneMinusCentreWeight = 1 - centreWeight;


	for (int each = values.length-1; each >= 0; each--) {
		if (each != centreIndex) {
			sum += values[each];
		}
	}

	// Might as well compute this next expression only once for the loop.
	double constant = oneMinusCentreWeight / sum;

	for (int each = values.length-1; each >= 0; each--) {
		if (each == centreIndex) {
			// Weigh the grid point itself.
			weightedAverage += centreWeight * values[centreIndex];
		}
		else {
			// Weight the contribution of any other
			// point other than the grid point, provided
			// it is not a hole value
			if (values[each] != holeValue) {
				weightedAverage += (sum - values[each])
						           * values[each]
    							   * constant;
			}
		}
	}

	return weightedAverage;
}

 /**
  * Returns the index whereby value would be inserted into the
  * specified array values to maintain ascending order.  The
  * array of values must be in ascending order for this method
  * to function properly.
  *
  * <p>
  * For example, if <code>value</code> is 2.8 and <code>values</code> is [0.0, 2.5, 3.1, 4.5], then
  * 2 is returned.  Note the following boundary conditions for the above
  * values: if <code>value</code> is 0.0, 1 is returned; if <code>value</code> is 4.5, 3 is returned.
  *
  * @param value
  * @param values
  * @return index at which value is inserted into the array values to maintain
  *      ascending order, or -1 if <code>value</code> is less than <i>values[0]</i> or greater than
  *     <i>values[length-1]</i>
  */
 public static int findInsertionIndex(double value, double[] values) {

	if (value <= values[0]) {
		if (value == values[0]) {
			return 1;
		}
		return -1;
	}

	int max = values.length-1;
	if (value >= values[max]) {
		if (value == values[max]) {
			return max;
		}
		return -1;
	}

	int index = Arrays.binarySearch(values, value);

	if (index < 0) {
		// index = -(insertionPoint) - 1 according to Arrays.binarySearch()
		index = -(index + 1);
	}

	return index;
 }

 /**
  * This method is identical to <code>findInsertionIndex(double, double[])</code>, except
  * that it restricts the valid range of values to <i>[values[0], values[n-1]]</i>.
  * @return
  * @param value
  * @param values
  * @param n
  */
 public static int findInsertionIndex(double value, double[] values, int n) {

	int index = findInsertionIndex(value, values);

	if (index >= n) {
		index = -1;
	}

	return index;
 }

/**
 * Fits a piecewise linear spline to the series of <i>x</i>,<i>y</i> points.
 * Each cubic spline is represented by:
 * <br>
 * <i>f     (u) = a[index] + b[index]u
 * index
 * </i>
 *
 * <br>
 * where
 * <i>index</i> is defined on the range [0, n-1], and
 * <i>u</i> is defined on the range [0, x[index+1] - x[index]].
 *
 * <p>
 * The <i>a</i> and <i>b</i> coefficients are placed into the <i>a</i> and <i>b</i>
 * arrays, respectively.
 *
 * @param x
 * @param y
 * @param a
 * @param b
 * @param n
 */
public static void createLinearSplineCoefficients(
		double[] x,
		double[] y,
		double[] a,
		double[] b,
		int n) {

	for (int i = 0; i < n-1; i++) {
		a[i] = y[i];
		b[i] = (y[i+1] - y[i]) / (x[i+1] - x[i]);
	}
}

/**
 * Fits a piecewise cubic spline to the series of x,y points.
 * Each cubic spline is represented by:<br>
 *
 * <i>f     (u) = a[index] + b[index]u + c[index]u^2 + d[index]u^3
 * index </i>
 * <br>
 * where <i>index</i> is defined on the range [0, n-1], and
 * <i>u</i> is defined on the range [0, x[index+1] - x[index]].
 *
 *
 * The slope at the beginning of the series and at the end
 * of the series is estimated using qadratic interpolation.
 * The curve fitting algorithm used is outlined in Conte and
 * deBoor, Elementary Numerical Analysis, An Algorithmic Approach,
 * 3rd edition, pp 284-292.
 *
 * @param x
 * @param y
 * @param a
 * @param b
 * @param c
 * @param d
 * @param n the number of intervals on which to generate spline coefficients
 */
public static void createCubicSplineCoefficients(
		double[] x,
		double[] y,
		double[] a,
		double[] b,
		double[] c,
		double[] d,
		int n) {

	int	i;

	double
		g,
		dx,
		div1,
		div2;

	n -= 1;

	a[0] = y[0];
	b[0] = estimateInitialSlopeOfCurve(x[0], y[0], x[1], y[1], x[2], y[2]);
	c[0] = 0.0;
	d[0] = 1.0;

	b[n] = estimateInitialSlopeOfCurve(x[n], y[n], x[n-1], y[n-1], x[n-2], y[n-2]);

	for (i = 1; i <= n; i++) {
		a[i] = y[i];
		c[i] = x[i] - x[i-1];
		d[i] = (y[i] - y[i-1]) / c[i];
	}

	for (i = 1; i < n; i++) {
		b[i] = 3.0 * (c[i]*d[i+1] + c[i+1]*d[i]);
		d[i] = 2.0 * (c[i] + c[i+1]);
	}

	for (i = 1; i < n; i++) {
		g = -c[i+1] / d[i-1];
		d[i] += g*c[i-1];
		b[i] += g*b[i-1];
	}

	for (i = n-1; i > 0; i--) {
		b[i] = (b[i] - c[i]*b[i+1]) / d[i];
	}

	for (i = 0; i < n; i++) {
		dx = x[i+1] - x[i];
		div1 = (y[i+1] - y[i]) / dx;
		div2 = b[i] + b[i+1] - 2*div1;
		c[i] = (div1 - b[i] - div2) / dx;
		d[i] = div2 / (dx*dx);
	}
}

/**
 * Evaluates the cubic spline defined by <i>f(x) = a + bx + c^2 + dx^3</i>
 * at the specified parameter <i>x</i> value and returns the result.
 * The <i>x</i> parameter must be on the range of the spline for the result to
 * be valid.
 *
 * @param x
 * @param a
 * @param b
 * @param c
 * @param d
 * @return the initial slope of a quadratic curve fitted to the points
 *     (x0,y0), (x1,y1), and (x2,y2)
 */
public static double evaluateCubicSpline(
		double x,
		double a,
		double b,
		double c,
		double d) {

	return a + x * (b + x * (c + x * d));
}

/**
 * Evaluates the linear spline defined by <i>f(x) = a + bx</i> at the
 * specified parameter <i>x</i> and returns the result. The <i>x</i> parameter must be
 * on the range of the spline for the result to be valid.
 * @return
 * @param x
 * @param a
 * @param b
 */
public static double evaluateLinearSpline(double x, double a, double b) {

	return a + b * x;
}

/**
 * Estimates the initial slope of the curve which contains the
 * points (x0, y0), (x1, y1), (x2, y2).  The slope is estimated
 * by fitting a quadratic curve to the specified points, then
 * computing the first derviative at point (x0, y0).  The <i>x</i> values
 * of the points must be in ascending or descending order for
 * this method to function properly.
 *
 * @param x0
 * @param y0
 * @param x1
 * @param y1
 * @param x2
 * @param y2
 * @return the initial slope of a quadratic curve fitted to the points
 *     (x0,y0), (x1,y1), and (x2,y2)
 */
public static double estimateInitialSlopeOfCurve(
		double x0,
		double y0,
		double x1,
		double y1,
		double x2,
		double y2) {

	// Divided differences?
	double
		f01 = (y1 - y0) / (x1 - x0), // slope between x0 and x1
		f12 = (y2 - y1) / (x2 - x1), // slope between x1 and x2
		f02 = (f12 - f01) / (x2 - x0);

	return f01 + f02 * (x0 - x1);
}

/**
 * Constructs a data model by taking <code>xSampleSize</code> and <code>ySampleSize</code> samples
 * along the <i>x</i> and <i>y</i> axes of the source.  The first <i>x</i> sample is taken
 * along the source <i>x</i> axis at <code>dataWindow.getMinX()</code>, and the last at
 * <code>dataWindow.getMaxX()</code>.  The first <i>y</i> sample is taken along
 * the source <i>y</i> axis at <code>dataWindow.getMinY()</code>, and the last at
 * <code>dataWindow.getMaxY()</code>.  All samples are equally spaced between the
 * first and last samples along each axis respectively.
 *
 * <p>
 * The <i>z</i>Values at each sample location are computed by interpolation of a
 * series of piecewise linear splines which are fitted to the data values
 * in both the <i>x</i> and <i>y</i> directions.
 *
 * @param source
 * @param dataWindow
 * @param xSampleSize
 * @param ySampleSize
 * @throws NullPointerException If any of source or dataWindow is null.
 * @throws IllegalArgumentException If the source data model contains fewer
 *     than a 2x2 grid of data values, or if <code>xSampleSize</code> or <code>ySampleSize</code> is less
 *     than 2.
 * @return a <code>Chart3dGridDataModel</code>, or null (if source is null)
 */
public static Chart3dGridDataModel createLinearSampledDataModel(
		Chart3dGridDataModel source,
		RectangularShape dataWindow,
		int xSampleSize,
		int ySampleSize) {

	return createSampledDataModel(source,
								  dataWindow,
								  xSampleSize,
								  ySampleSize,
								  DATA_INTERPOLATE_LINEAR);
}

/**
 * Constructs a data model by taking <code>xSamples.length</code> and <code>ySamples.length
 * </code>
 * samples along the <i>x</i> and <i>y</i> axes of the source.  The samples of the source
 * data will be taken at locations defined by a grid formed by the <code>xSamples
 * </code>
 * and <code>ySamples</code> arrays.  To function properly, both of these arrays
 * should be in ascending order.
 *
 * <p>
 * The <code>Values</code> at each sample location are computed by interpolation of a
 * series of piecewise linear splines which are fitted to the data values
 * in both the <i>x</i> and <i>y</i> directions.
 *
 * @param source
 * @param xSamples
 * @param ySamples
 * @throws NullPointerException If any of source, <code>xSamples</code>, or <code>ySamples</code> is
 *     null.
 * @throws IllegalArgumentException If the source data model contains fewer
 *     than a 2x2 grid of data values, or if <code>xSamples</code> or <code>ySamples</code> contains fewer
 *      than 2 samples.
 * @return - a <code>Chart3dGridDataModel</code>, or null (if source is null)
 */
public static Chart3dGridDataModel createLinearSampledDataModel(
		Chart3dGridDataModel source,
		double[] xSamples,
		double[] ySamples) {

	return createSampledDataModel(source,
								  xSamples,
								  ySamples,
								  DATA_INTERPOLATE_LINEAR);
}

/**
 * Constructs a data model by taking <code>xSamples.length</code> and <code>ySamples.length</code>
 * samples along the <i>x</i> and <i>y</i> axes of the source.  The samples of the source
 * data will be taken at locations defined by a grid formed by the <code>xSamples
 * </code>
 * and <code>ySamples</code> arrays.  To function properly, both of these arrays
 * should be in ascending order.
 *
 * <p>
 * The <code>zValues</code> at each sample location are computed by interpolation of a
 * series of piecewise cubic splines which are fitted to the data values
 * in both the <i>x</i> and <i>y</i> directions.  The cubic interpolation generally
 * results in a more authentic approximation to the data than a linear
 * interpolation. <br><B>Note:</b> If there are fewer than 3 available data
 * values along a given axis, linear interpolation will be used by default.
 *
 * @param source
 * @param xSamples
 * @param ySamples
 * @throws NullPointerException If any of source, <code>xSamples</code>, or <code>ySamples</code> is
 *     null.
 * @throws IllegalArgumentException If the source data model contains fewer
 *     than a 2x2 grid of data values, or if <code>xSamples</code> or <code>ySamples</code> contains fewer
 *      than 2 samples.
 * @return a <code>Chart3dGridDataModel</code>, or null (if source is null)
 */
public static Chart3dGridDataModel createCubicSampledDataModel(
		Chart3dGridDataModel source,
		double[] xSamples,
		double[] ySamples) {

	return createSampledDataModel(source,
								  xSamples,
								  ySamples,
								  DATA_INTERPOLATE_CUBIC);
}

/**
 * Constructs a data model by taking <code>xSampleSize</code> and <code>ySampleSize</code> samples
 * along the <i>x</i> and <i>y</i> axes of the source.  The first <i>x</i> sample is taken
 * along the source <i>x</i> axis at <code>dataWindow.getMinX()</code>, and the last at
 * <code>dataWindow.getMaxX()</code>.  The first <i>y</i> sample is taken along
 * the source <i>y</i> axis at <code>dataWindow.getMinY()</code>, and the last at
 * <code>dataWindow.getMaxY()</code>.  All samples are equally spaced between the
 * first and last samples along each axis respectively.
 *
 * <p>
 * The <code>zValues</code> at each sample location are computed by interpolation of a
 * series of piecewise cubic splines which are fitted to the data values
 * in both the <i>x</i> and <i>y</i> directions.  The cubic interpolation generally
 * results in a more authentic approximation to the data than a linear
 * interpolation.  Note however that if there are fewer than 3 available data
 * values along a given axis, linear interpolation will be used by default.
 *
 * @param source
 * @param dataWindow
 * @param xSampleSize
 * @param ySampleSize
 * @throws NullPointerException If any of source or <code>dataWindow</code> is null.
 * @throws IllegalArgumentException If the source data model contains fewer
 *     than a 2x2 grid of data values, or if <code>xSampleSize</code> or <code>ySampleSize</code> is less
 *     than 2.
 * @return a <code>Chart3dGridDataModel</code>, or null (if source is null)
 */
public static Chart3dGridDataModel createCubicSampledDataModel(
		Chart3dGridDataModel source,
		RectangularShape dataWindow,
		int xSampleSize,
		int ySampleSize) {

	return createSampledDataModel(source,
								  dataWindow,
								  xSampleSize,
								  ySampleSize,
								  DATA_INTERPOLATE_CUBIC);
}

/**  */
private static Chart3dGridDataModel createSampledDataModel(
		Chart3dGridDataModel source,
		RectangularShape dataWindow,
		int xSampleSize,
		int ySampleSize,
		int interpolationType) {

	double
		xMin = dataWindow.getMinX(),
		xMax = dataWindow.getMaxX(),
		yMin = dataWindow.getMinY(),
		yMax = dataWindow.getMaxY(),
		xStep = (xMax - xMin) / (xSampleSize - 1),
		yStep = (yMax - yMin) / (ySampleSize - 1);

	double[]
		xSamples = new double[xSampleSize],
		ySamples = new double[ySampleSize];

	// Populate the x sample array.
	for (int each = 0; each < xSampleSize; each++) {
		xSamples[each] = xMin + each*xStep;
	}

	// Populate the y sample array.
	for (int each = 0; each < ySampleSize; each++) {
		ySamples[each] = yMin + each*yStep;
	}

	return createSampledDataModel(
				source,
				xSamples,
				ySamples,
				interpolationType);
}

/**  */
private static Chart3dGridDataModel createSampledDataModel(
		Chart3dGridDataModel source,
		double[] xSamples,
		double[] ySamples,
		int interpolationType) {

	if (getSizeX(source) < 2 || getSizeY(source) < 2) {
		throw new IllegalArgumentException(
			"The source data must have atleast 2x2 data values to interpolate.");
	}

	if (xSamples.length < 2 || ySamples.length < 2) {
		throw new IllegalArgumentException(
			"Sample size must be atleast 2x2 to create a sampled data model.");
	}

	int	type;

	double
		holeValue = getHoleValue(source);

 	// If sample size is 2, use linear always.
	type = (getSizeX(source) == 2)
		? DATA_INTERPOLATE_LINEAR
		: interpolationType;

	// Sample along the x axis.
	double[][] tempZValues = createSampleData(
			source.getXGrid(),
			source.getYGrid(),
			source.getZValues(),
			holeValue,
			false,
			xSamples,
			type);

	// If sample size is 2, use linear always.
	type = (getSizeY(source) == 2)
		? DATA_INTERPOLATE_LINEAR
		: interpolationType;

	// Sample along the y axis.
	double[][] zSamples = createSampleData(
			xSamples,
			source.getYGrid(),
			tempZValues,
			holeValue,
			true,
			ySamples,
			type);

	Base3dGridDataSource dest;
	dest = new Base3dGridDataSource(null, xSamples, ySamples, zSamples);
	dest.setHoleValue(holeValue);

	return dest;
}

/**  */
private static double[][] createSampleData(
		double[] xGrid,
		double[] yGrid,
		double[][] zValues,
		double holeValue,
		boolean invert,
		double[] gridSamples,
		int interpType) {

	boolean
		useCubic = (interpType == DATA_INTERPOLATE_CUBIC);

	int	i = 0,
		j = 0,
		index,
		sampleSize = gridSamples.length,
		noHoleSampleSize;

	double
		sample,
		zSample;

	double[]
		nGrid = (invert) ? yGrid : xGrid,
		mGrid = (invert) ? xGrid : yGrid,
		noHoleGridSamples = new double[nGrid.length],
		noHoleZSamples = new double[nGrid.length],
		a = new double[nGrid.length],
		b = new double[nGrid.length],
		c = (useCubic) ? new double[nGrid.length] : null, // don't use if linear
		d = (useCubic) ? new double[nGrid.length] : null; // don't use if linear

	double[][]
		zSamples = (invert)
			? new double[xGrid.length][sampleSize]
			: new double[sampleSize][yGrid.length];

	int	nSize = nGrid.length,
		mSize = mGrid.length;

	for (int m = 0; m < mSize; m++) {

		noHoleSampleSize = 0;

		if (invert)
			i = m;
		else
			j = m;

		for (int n = 0; n < nSize; n++) {

			useCubic = (interpType == DATA_INTERPOLATE_CUBIC);

			if (invert)
				j = n;
			else
				i = n;

			zSample = zValues[i][j];

			if (zSample != holeValue) {
				noHoleGridSamples[noHoleSampleSize] = nGrid[n];
				noHoleZSamples[noHoleSampleSize] = zSample;
				noHoleSampleSize++;
			}
		}

		if (noHoleSampleSize > 2 && useCubic) {
			createCubicSplineCoefficients(
					noHoleGridSamples,
					noHoleZSamples,
					a,
					b,
					c,
					d,
					noHoleSampleSize);
		}
		else if ((!useCubic && noHoleSampleSize >= 2)
				 || noHoleSampleSize == 2) {

			useCubic = false;

			/*
			 * If there are two or more data samples, and linear has
			 * been specified, use linear.  If there are only two data
			 * however, use linear whether the user specified linear,
			 * cubic or whatever.
			 */
			createLinearSplineCoefficients(
				noHoleGridSamples,
				noHoleZSamples,
				a,
				b,
				noHoleSampleSize);
		}

		for (int n = 0; n < sampleSize; n++) {

			if (invert)
				j = n;
			else
				i = n;

			sample = gridSamples[n];

			/*
			 * Find the insertion index of the current gridValue in the
			 * original grid series.
			 */
			index = findInsertionIndex(sample, nGrid);
			if (index < 0) {
				/*
				 * If the insertion index is < 0, the gridValue is outside
				 * of the available data range; set the sample to the hole
				 * value.
				 */
				zSample = holeValue;
			}
			else {
				/*
				 * The gridValue is inside the available data range.  But,
				 * before evaluating the spline, check that the grid points
				 * both before and after the insertion index are not hole
				 * values.  If either is a hole value, then the sample will
				 * also be a hole value.
				 */
				if (invert &&
					(zValues[i][index-1] == holeValue || zValues[i][index] == holeValue)) {
					zSample = holeValue;
				}
				else if (!invert &&
					(zValues[index-1][j] == holeValue || zValues[index][j] == holeValue)) {
					zSample = holeValue;
				}
				else {
					/*
					 * This is what we've been waiting for.  Evaluate the
					 * spline at gridValue.
					 *
					 * First, find the insertion index sample in the sampled grid.
					 */

					index = findInsertionIndex(sample, noHoleGridSamples, noHoleSampleSize);

					if (index < 0) {
						zSample = holeValue;
					}
					else {
						index--;
						sample -= noHoleGridSamples[index]; // evaluate spline at x - x[index-1]
						if (useCubic) {
							zSample = evaluateCubicSpline(
									sample,
									a[index],
								 	b[index],
									c[index],
									d[index]);
						}
						else {
							zSample = evaluateLinearSpline(
									sample,
									a[index],
									b[index]);
						}
					}
				}
			}

			zSamples[i][j] = zSample;
		}
	}

	return zSamples;
}

/**
 * A convenience method which returns the hole value for the
 * specified <code>Chart3dDataModel</code>.  If the model implements the
 * <code>HoleValue3dDataModel</code> interface, returns
 * <code>HoleValue3dDataModel.getHoleValue()</code>; otherwise, returns
 * <code>HoleValue3dDataModel.DEFAULT_HOLE_VALUE</code>.
 * @return
 * @param model
 */
public static double getHoleValue(Chart3dDataModel model) {

	if (model instanceof HoleValueChart3dDataModel) {
		return ((HoleValueChart3dDataModel)model).getHoleValue();
	}

	return HoleValueChart3dDataModel.DEFAULT_HOLE_VALUE;
}

/**
 * A convenience method which returns the size of the specified
 * data model's <i>x</i> grid, or 0 if the model has no <i>x</i> data.
 *
 * @param model
 * @return the size of the of the data model argument's <i>x</i> grid, or 0
 *     if the data model has no <i>x</i> grid
 */
public static int getSizeX(Chart3dGridDataModel model) {

	if (model.getXGrid() != null) {
		return model.getXGrid().length;
	}
	return 0;
}

/**
 * A convenience method which returns the size of the specified
 * data model's <i>y</i> grid, or 0 if the model has no <i>y</i> data.
 *
 * @param model
 * @return the size of the of the data model argument's <i>y</i> grid, or 0
 *     if the data model has no <i>y</i> grid
 */
public static int getSizeY(Chart3dGridDataModel model) {

	if (model.getYGrid() != null) {
		return model.getYGrid().length;
	}
	return 0;
}

}
