/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCSwing3dDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:02 $  $Locker:  $  Quest Software Inc.
package com.klg.jclass.chart3d.data;

import com.klg.jclass.chart3d.*;
import com.klg.jclass.chart3d.event.*;


import javax.swing.table.TableModel;
import javax.swing.event.*;

/**
 * <code>3d DataSource</code> converts a Swing <code>TableModel</code> into a form useable
 * by the <code>3d Chart</code>.  <i>X</i> and <i>Y</i> values are generated.  <i>Z</i>Values are generated
 * in the form <i>z</i>Values[table columns][table rows].<br>
 * <b>Note:</b> This <code>DataSource</code> is only available in Grid format.
 *
 * <p>
 * This datasource implements <code>TableModelListener</code> for the benefit of those
 * who wish to update their chart through the <code>TableModel</code>. It is
 * recommended that those who wish to update their <code>TableModel</code> through the
 * chart should implement a <code>Chart3dDataListener</code> for this purpose.
 *
 */
public class JCSwing3dDataSource extends JCEditable3dGridDataSource
			                     implements TableModelListener {

/** Swing <code>TableModel</code> instance used to provide <i>z</i>Values for this datasource.
 *
 */
protected TableModel tableModel;

/**
 * Simple constructor which takes the <code>TableModel</code> as an
 * argument.
 *
 * @param tm
 * @see javax.swing.table.TableModel
 */
public JCSwing3dDataSource(TableModel tm)
{
	super();
	this.tableModel = tm;
	tableModel.addTableModelListener(this);
	this.setData();
} //constructor

/**
 * Makes <i>z</i>Values out of the <code>tableModel</code>.  It is called by
 * the 1-arg constructor, the <code>setTableModel</code> method, and anytime
 * changes to the <code>tableModel</code> are comprehensive enough to warrant
 * a <code>RESET</code> event (which it fires).
 *
 */
protected void setData()
{
	int cols, rows;

	rows = tableModel.getRowCount();
	cols = tableModel.getColumnCount();

	double[][] zvals = new double[rows][cols];
	for (int i = 0; i < rows; i++) {
		for (int j = 0; j < cols; j++) {
			Object val = tableModel.getValueAt(i, j);
			if (val == null) {
				zvals[i][j] = getHoleValue();
			}
			else if (val instanceof Number) {
				zvals[i][j] = ((Number)val).doubleValue();
			}
			else if (val instanceof String) {
				try {
					// Try to parse the string

					zvals[i][j] = Double.parseDouble((String)val);

				}
				catch (Exception e) {
					//We'll default to 0.0
					zvals[i][j] = 0.0;
				}//catch
			}
			else {
				zvals[i][j] = 0.0;
			}//if
		}//for j
	}//for i

	setZValues(zvals);

	//We need to force recalculation of X and Y Values.
	setXGrid(null);
	setYGrid(null);
	//Don't forget to notify any interested listeners.
	int type = Chart3dDataEvent.RESET;
	JCData3dGridIndex index = new JCData3dGridIndex(JCData3dGridIndex.ALL,
													JCData3dGridIndex.ALL);
	fireChart3dDataEvent(new Chart3dGridDataEvent(this, type, index));

}//setData

/**
 * This method is from the <code>TableModelListener</code>.  It updates the
 * chart data, when the <code>TableModel</code> Changes.
 *
 * @param ev <code>TableModelEvent</code>
 */
public void tableChanged(TableModelEvent ev)
{
	int row     = ev.getFirstRow();
	int lastRow = ev.getLastRow();
	int col     = ev.getColumn();

	double newval;
	if ((row == TableModelEvent.HEADER_ROW)         ||
	    (lastRow != row)                            ||
		(ev.getType() != TableModelEvent.UPDATE)    ||
		(col == TableModelEvent.ALL_COLUMNS))
	{
		setTableModel((TableModel)ev.getSource());
	}
	else {
		try {
			Object changeVal = tableModel.getValueAt(row, col);
			newval = ((Number)changeVal).doubleValue();
		}
		catch (Exception ex) {
			newval = 0.0;
		}

	    JCData3dGridIndex index = new JCData3dGridIndex(row, col);
		setZValue(index, newval);
	    int type = Chart3dGridDataEvent.RELOAD_ZVALUE;
		fireChart3dDataEvent(new Chart3dGridDataEvent(this, type, index));
	}

}//tableChanged

/**
 * Gets the <code>TableModel</code> object for this datasource.
 *
 * @return tableModel <code>TableModel</code>
 */
public TableModel getTableModel()
{
	return(this.tableModel);
}//getTableModel

/**
 * Sets the <code>tableModel</code> property, adds the <code>tableModelListener</code> to
 * it, and calls <code>setData;</code>.
 *
 * @param tm
 */
public void setTableModel(TableModel tm)
{
	this.tableModel = tm;
	tableModel.addTableModelListener(this);
	setData();
}//setTableModel

}
