/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: MapPickGrid.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:03 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d;

import java.awt.Point;
import java.awt.Rectangle;
import javax.vecmath.Point3d;
import javax.vecmath.Point4d;
import javax.vecmath.Matrix4d;

import com.klg.jclass.chart3d.*;


/**
 * A class that implements the Mapping interface.  This computes the map and
 * pick functions for the j2d version of Chart3d.
 */
public class MapPickGrid extends MapPick {

public final static double T_EPSILON = 1.0e-10;

/**
 * Default constructor, required by Java Beans.
 */
public MapPickGrid()
{
}

/**
 * Default constructor, required by Java Beans.
 */
public MapPickGrid(Chart3dData data)
{
    super(data);
}

/**
 * Find the closest point in pixel space to (xPixel, yPixel).
 */
protected boolean findClosestPoint(int xPixel, int yPixel, JCData3dGridIndex index)
{
    Chart3dGridData      gridData = (Chart3dGridData)data;
    GridLimits           limits = gridData.getGridLimits();

    if (gridData == null || !gridData.isDataOK() || index == null) {
        return(false);
    }

    ScreenPoint[][] screenPoints = getScreenPoints();
    Chart3dDataView dataView     = gridData.getDataView();
    JCChart3dArea   area         = dataView.getChart3d().getChart3dArea();
    Rectangle       rect         = area.getDrawingArea();

    double distance = Double.MAX_VALUE;
	for (int i = limits.getXOffset(); i <= limits.getXLast(); i++) {
		for (int j = limits.getYOffset(); j <= limits.getYLast(); j++) {
			ScreenPoint sp = screenPoints[i][j];
			if (!sp.isHoleValue() && !sp.isHidden()) {
				int x = sp.getXPixel();
				int y = sp.getYPixel();
                if (rect.contains(x, y)) {
					// Only include points inside graph
					double curDistance = calcDist(xPixel, yPixel, x, y);
					if (curDistance <= distance) {
                        distance = curDistance;
						index.setDistance(curDistance);
                        index.setX(i);
                        index.setY(j);
					}
				}
				else {
					sp.setHidden(true);
				}
			}
		}
	}
	return(distance != Double.MAX_VALUE);
}

/**
 * Performs a map operation by mapping the provided screen pixel
 * information to a value in data space.
 * @param xPixel x value in screen pixels
 * @param yPixel y value in screen pixels
 * @return The data point that maps to the given screen point
 */
public Point3d map(int xPixel, int yPixel)
{
	if (!(data instanceof Chart3dGridData)) {
		return(null);
	}
    Chart3dGridData gridData = (Chart3dGridData)data;
    if (gridData == null || !gridData.isDataOK()) {
        return(null);
    }
	Chart3dDataView   dataView  = gridData.getDataView();
	JCChart3dArea     area      = dataView.getChart3d().getChart3dArea();

	Point3d mapPoint = new Point3d(Double.MAX_VALUE, Double.MAX_VALUE,
                                   Double.MAX_VALUE);
	if (dataView.getChart3d().isProjection()) {
        GridLimits limits = gridData.getGridLimits();
        double xMin = gridData.getX(limits.getXOffset());
		double xMax = gridData.getX(limits.getXLast());
		double yMin = gridData.getY(limits.getYOffset());
		double yMax = gridData.getY(limits.getYLast());
		Rectangle rect = interpolateXY(xPixel, yPixel, xMin, yMin,
                                       xMax, yMax, mapPoint);

		if ((xPixel >= rect.x) && (xPixel <= rect.x + rect.width) &&
            (yPixel >= rect.y) && (yPixel <= rect.y + rect.height))
        {
            int xIndex = gridData.getPointIndex(mapPoint.x, gridData.getXGrid(),
                                                gridData.getNumX());
            int yIndex = gridData.getPointIndex(mapPoint.y, gridData.getYGrid(),
                                                gridData.getNumY());
			if (xIndex + 1 == gridData.getNumX()) {
				xIndex--;
			}
			if (yIndex + 1 == gridData.getNumY()) {
				yIndex--;
			}
			if (!JCChart3dUtil.holeInFacet(gridData, xIndex, yIndex)) {
				// Do bilinear interpolation on the gridData values array
				mapPoint.z = JCChart3dUtil.biLinearInterp(gridData, xIndex, yIndex,
                                                          mapPoint.x, mapPoint.y);
			}
			else {
				mapPoint.z = gridData.getHoleValue();
			}
		}
		else {
			// Outside of grid area
			mapPoint.set(Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
		}
	}
	else {
		Transform transform = area.getTransform();

		// Find ray from the eye position to xPixel, yPixel &
		// transform back to absolute space
		Point4d rayOrigin = transform.computeEyePosition();

		Point4d rayPoint = new Point4d(xPixel, yPixel, 0.0, 1.0);
		transform.transformPoint(transform.getInverseScreen(), rayPoint);
		transform.transformPoint(transform.getInverseRotation(), rayPoint);

		Point4d result = new Point4d();

		// Walk through facets (near->far) to find first intersection
		Draw draw = dataView.getDraw();
		if (!(draw instanceof DrawGrid)) {
			return(null);
		}
		DrawGrid drawGrid = (DrawGrid)draw;
	    JCData3dGridIndex gridIndex = new JCData3dGridIndex(dataView, -1, -1);
		boolean onSurface = drawGrid.gridSurfaceIntersect(rayOrigin, rayPoint,
							                              result, gridIndex);
		if (onSurface) {
			/* Transform from absolute to data space */
			transform.transformPoint(transform.getInverseAbsolute(), result);
			mapPoint.project(result);
		}
	}
	return(mapPoint);
}

/**
 *  Search the grid points for the closest point in screen-space to the pick
 *  point (cursor position).  Test the best candidate for visibility by using
 *  the gridSurfaceIntersect routine.  If the point is visible, the
 *  intersection will occur at t = 1.0; less than 1.0 means that the ray
 *  hit a surface in front of the grid point.
 *  If a point fails the visibility test, mark its hidden bit so that it is not
 *  tested in subsequent picks.
 */
protected boolean gridPointSelect(int xPixel, int yPixel, JCData3dGridIndex index)
{
    Chart3dGridData   gridData     = (Chart3dGridData)data;
	Chart3dDataView   dataView     = gridData.getDataView();
    GridLimits        limits       = gridData.getGridLimits();
	JCChart3dArea     area         = dataView.getChart3d().getChart3dArea();
	Transform         transform    = area.getTransform();
	ScreenPoint[][]   screenPoints = getScreenPoints();

	Draw draw = dataView.getDraw();
	if (!(draw instanceof DrawGrid)) {
		return(false);
	}
	DrawGrid drawGrid = (DrawGrid)draw;

    // Reset the hidden flags from a possible earlier call
	for (int i = limits.getXOffset(); i <= limits.getXLast(); i++) {
		for (int j = limits.getYOffset(); j <= limits.getYLast(); j++) {
			screenPoints[i][j].setHidden(false);
		}
	}

	// Build the ray start point at the eye point
	Point4d rayOrigin                = transform.computeEyePosition();
	Point4d result                   = new Point4d();
	JCData3dGridIndex intersectIndex = new JCData3dGridIndex();
	while (true) {
		if (!findClosestPoint(xPixel, yPixel, index)) {
			break;
		}
		int xIndex = index.getX();
		int yIndex = index.getY();

		// Probe ray from eye_point to grid_point
		Point4d rayPoint = drawGrid.findAbsolutePoint(gridData,
												      transform.getAbsolute(),
						                              xIndex, yIndex);
		intersectIndex.setDistance(Double.MAX_VALUE);
		boolean intersect = drawGrid.gridSurfaceIntersect(rayOrigin, rayPoint,
							                              result, intersectIndex);
		double t = intersectIndex.getDistance();
		if (intersect &&  t >= (1.0 - T_EPSILON)) {
			return(true);
		}

		// Point didn't intersect or surface intersection happened in front
		// of the point.  Mark as hidden.
		screenPoints[xIndex][yIndex].setHidden(true);
	}
    return(false);
}

/**
 * Performs a pick operation by returning the index of the closest point in
 * the grid data set to the given screen point.
 * @param xPixel x value in screen pixels
 * @param yPixel y value in screen pixels
 */
public JCData3dIndex pick(int xPixel, int yPixel)
{
	if (!(data instanceof Chart3dGridData)) {
		return(null);
	}
    Chart3dGridData gridData = (Chart3dGridData)data;
    if (gridData == null || !gridData.isDataOK()) {
        return(null);
    }
	Chart3dDataView   dataView  = gridData.getDataView();
	JCChart3dArea     area      = dataView.getChart3d().getChart3dArea();
	JCData3dGridIndex gridIndex = new JCData3dGridIndex(dataView, -1, -1);

	if (dataView.getChart3d().isProjection()) {
		findClosestPoint(xPixel, yPixel, gridIndex);
	}
	else {  // 3D
		if (dataView.getChartType() == JCChart3d.BAR) {
			Transform transform = area.getTransform();

			// Find ray from the eye position to xPixel, yPixel &
			// transform back to absolute space
			Point4d rayOrigin = transform.computeEyePosition();

			Point4d rayPoint = new Point4d(xPixel, yPixel, 0.0, 1.0);
			transform.transformPoint(transform.getInverseScreen(), rayPoint);
			transform.transformPoint(transform.getInverseRotation(), rayPoint);

			Point4d res = new Point4d();

			// Walk through facets (near->far) to find first intersection
			Draw draw = dataView.getDraw();
			if (!(draw instanceof DrawGrid)) {
				return(null);
			}
			DrawGrid drawGrid = (DrawGrid)draw;
			drawGrid.gridSurfaceIntersect(rayOrigin, rayPoint, res, gridIndex);
		}
		else {
			if (!gridPointSelect(xPixel, yPixel, gridIndex)) {
				gridIndex.setX(-1);
				gridIndex.setY(-1);
			}
		}
	}
	return(gridIndex);
}

/**
 * Performs an unpick operation by returning the screen point that corresponds
 * to the given data index.
 * @param data The data set that defines the data space
 * @param index The data index of the point whose screen point is required.
 */
public java.awt.Point unpick(JCData3dIndex index)
{
	Point pickPoint = new Point(-1, -1);
	if (!(data instanceof Chart3dGridData) ||
	    !(index instanceof JCData3dGridIndex))
	{
		return(pickPoint);
	}

    Chart3dGridData      gridData  = (Chart3dGridData)data;
	JCData3dGridIndex    gridIndex = (JCData3dGridIndex)index;
    if (gridData == null || !gridData.isDataOK() || gridIndex == null) {
        return(pickPoint);
    }

	ScreenPoint[][] screenPoints = getScreenPoints();

	// Validate the input data
	int xIndex = gridIndex.getX();
	int yIndex = gridIndex.getY();
	if ((xIndex >= 0) && (xIndex < gridData.getNumX()) &&
		(yIndex >= 0) && (yIndex < gridData.getNumY()))
	{
		// Data is in range. Return the screen location of the grid point.
		ScreenPoint sp = screenPoints[xIndex][yIndex];
		pickPoint.setLocation(sp.getXPixel(), sp.getYPixel());
	}

	// Default distance to 0 (its meaningless here)
	index.setDistance(0.0);
	return(pickPoint);
}

}
