/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: MapPickPoint.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:03 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d;

import java.awt.Point;
import java.awt.Rectangle;
import javax.vecmath.Point3d;
import java.util.ArrayList;

import com.klg.jclass.chart3d.*;


/**
 * A class that implements the Mapping interface.  This computes the map and
 * pick functions for the j2d version of Chart3d.
 */
public class MapPickPoint extends MapPick {

/**
 * Default constructor, required by Java Beans.
 */
public MapPickPoint()
{
}

/**
 * Default constructor, required by Java Beans.
 */
public MapPickPoint(Chart3dData data)
{
    super(data);
}


/**
 * Find the closest point in pixel space to (int xPixel, int yPixel).
 */
protected boolean findClosestPoint(int xPixel, int yPixel, JCData3dPointIndex index)
{
    Chart3dPointData pointData = (Chart3dPointData)data;
    if (pointData == null || !pointData.isDataOK() || index == null) {
        return(false);
    }

    ScreenPoint[][] screenPoints = getScreenPoints();
    Chart3dDataView dataView     = pointData.getDataView();
    JCChart3dArea   area         = dataView.getChart3d().getChart3dArea();
    Rectangle       rect         = area.getDrawingArea();

    pointData.sortPoints();
    ArrayList sortPoints = pointData.getSortPoints();
	if (sortPoints == null) {
		return(false);
	}

	for (int i = 0; i < sortPoints.size(); i++) {
        SortPoint sortPoint = (SortPoint)sortPoints.get(i);
		int series          = sortPoint.getSeries();
		int point           = sortPoint.getPoint();
		ScreenPoint sp      = screenPoints[series][point];
		if (!sp.isHoleValue() && !sp.isHidden()) {
			int x = sp.getXPixel();
			int y = sp.getYPixel();
            if (rect.contains(x, y)) {
				// Only include points inside graph
				double curDistance = calcDist(xPixel, yPixel, x, y);
				if (curDistance <= index.getDistance()) {
					index.setDistance(curDistance);
                    index.setSeries(series);
                    index.setPoint(point);
				}
			}
			else {
				sp.setHidden(true);
			}
		}
	}
	return(index.getDistance() != Double.MAX_VALUE);
}

/**
 * Performs a map operation by mapping the provided screen pixel
 * information to a value in data space.
 * @param xPixel x value in screen pixels
 * @param yPixel y value in screen pixels
 * @return The data point that maps to the given screen point
 */
public Point3d map(int xPixel, int yPixel)
{
    int series, point;

	if (!(data instanceof Chart3dPointData)) {
		return(null);
	}
    Chart3dPointData pointData = (Chart3dPointData)data;
    if (pointData == null || !pointData.isDataOK()) {
        return(null);
    }
	Chart3dDataView    dataView   = pointData.getDataView();
	JCChart3dArea      area       = dataView.getChart3d().getChart3dArea();
    JCData3dPointIndex pointIndex = new JCData3dPointIndex(dataView, -1, -1);

	Point3d mapPoint = new Point3d(Double.MAX_VALUE, Double.MAX_VALUE,
                                   Double.MAX_VALUE);
	if (dataView.getChart3d().isProjection()) {
        double xMin = area.getAxis(JCAxis.AXIS_X).getMin();
		double xMax = area.getAxis(JCAxis.AXIS_X).getMax();
		double yMin = area.getAxis(JCAxis.AXIS_Y).getMin();
		double yMax = area.getAxis(JCAxis.AXIS_Y).getMax();
		Rectangle rect = interpolateXY(xPixel, yPixel, xMin, yMin,
                                       xMax, yMax, mapPoint);
		if ((xPixel >= rect.x) && (xPixel <= rect.x + rect.width) &&
            (yPixel >= rect.y) && (yPixel <= rect.y + rect.height))
        {
            //  For lack of a better z value, pick the z value of the closest
            //  point in pixel space.
            if (findClosestPoint(xPixel, yPixel, pointIndex)) {
                series = pointIndex.getSeries();
                point  = pointIndex.getPoint();
                mapPoint.z = pointData.getSeries(series).getPoint(point).z;
            }
        }
		else {
			// Outside of data area
			mapPoint.set(Double.MAX_VALUE, Double.MAX_VALUE, Double.MAX_VALUE);
		}
	}
	else {
        //  For lack of a better reference point, pick the closest point
        //  in pixel space.
        if (findClosestPoint(xPixel, yPixel, pointIndex)) {
            series = pointIndex.getSeries();
            point  = pointIndex.getPoint();
            mapPoint.set(pointData.getSeries(series).getPoint(point));
        }
	}
	return(mapPoint);
}

/**
 * Performs a pick operation by returning the index of the closest point in
 * the point data set to the given screen point.
 * @param xPixel x value in screen pixels
 * @param yPixel y value in screen pixels
 */
public JCData3dIndex pick(int xPixel, int yPixel)
{
	if (!(data instanceof Chart3dPointData)) {
		return(null);
	}
    Chart3dPointData pointData = (Chart3dPointData)data;
    if (pointData == null || !pointData.isDataOK()) {
        return(null);
    }
	JCData3dPointIndex pointIndex = new JCData3dPointIndex(pointData.getDataView(),
														   -1, -1);
	findClosestPoint(xPixel, yPixel, pointIndex);
	return(pointIndex);
}

/**
 * Performs an unpick operation by returning the screen point that corresponds
 * to the given data index.
 * @param data The data set that defines the data space
 * @param index The data index of the point whose screen point is required.
 */
public java.awt.Point unpick(JCData3dIndex index)
{
	Point pickPoint = new Point(-1, -1);
	if (!(data instanceof Chart3dPointData) ||
	    !(index instanceof JCData3dPointIndex))
	{
		return(pickPoint);
	}

    Chart3dPointData   pointData  = (Chart3dPointData)data;
	JCData3dPointIndex pointIndex = (JCData3dPointIndex)index;
    if (pointData == null || !pointData.isDataOK() || pointIndex == null) {
        return(pickPoint);
    }

	ScreenPoint[][] screenPoints = getScreenPoints();

	// Validate the input data
	int series = pointIndex.getSeries();
	int	point  = pointIndex.getPoint();
	if (series >= 0				            &&
		series < pointData.getNumSeries()   &&
		point >= 0				            &&
		point < pointData.getSeries(series).getNumPoints())
	{
		// Data is in range. Return the screen location of the desired point.
		ScreenPoint sp = screenPoints[series][point];
		pickPoint.setLocation(sp.getXPixel(), sp.getYPixel());
	}

	// Default distance to 0 (its meaningless here)
	index.setDistance(0.0);
	return(pickPoint);
}

}
