/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:03 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d.actions;

import com.klg.jclass.chart3d.*;

import java.io.Serializable;
import java.awt.event.InputEvent;

/**
 * BaseAction is an abstract base class for the Java2D actions of JCChart 3D.
 * It implements the JCAction interface used for JCChart 3D actions. It defines
 * methods for starting, animating, ending and cancelling an action.
 *
 * <code>JCChart3dArea</code> will create an instance of the appropriate
 * <code>JCAction</code> when an action is started, and will call
 * <code>start()</code>, <code>animate()</code> and <code>end()</code> during
 * <code>mouseDown()</code>, <code>mouseDrag()</code> and
 * <code>mouseUp()</code> respectively.  The appropriate <code>JCAction</code>
 * is dependent on the event.  A mapping from event to action type is stored
 * the chart's JCActionTable
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 */
public abstract class BaseAction implements JCAction, Serializable {

public final static int NONE      = 0;
public final static int START     = 1;
public final static int ANIMATING = 2;

/** Parent object on which the action occurs. */
protected JCChart3dArea chart3dArea = null;

/** Is this action enabled */
protected boolean enabled = true;
protected int     status = NONE;

/**
 * Constructs an action handler.  Note that it nees to be connected to a
 * <code>JCChart3dArea</code> instance so this needs to be set later.
 */
public BaseAction() {}

/**
 * Sets the parent JCChart3dArea object
 *
 * @param chart3dArea The new JCChart3dArea object
 */
public void setChart3dArea(JCChart3dArea chart3dArea)
{
	this.chart3dArea = chart3dArea;
}


/**
 * Returns the parent JCChart3dArea object
 *
 * @return The current JCChart3dArea object
 */
public JCChart3dArea getChart3dArea()
{
	return(chart3dArea);
}

/**
 * Sets whether this action handler is enabled
 *
 * @param enabled Is this action handler enabled?
 */
public void setEnabled(boolean enabled)
{
	this.enabled = enabled;
}

/**
 * Is this action handler enabled?
 *
 * @return Is this action handler enabled?
 */
public boolean isEnabled()
{
	return(enabled);
}

/**
 * Draws either the preview cube or the full chart 3D depending on the value
 * of the viewport's preview method.
 *
 * @param scale The scale of the preview cube
 * @param horizShift The horizontal shift of the preview cube
 * @param vertShift The vertical shift of the preview cube
 * @param xRotation The x rotation angle to draw the preview cube
 * @param yRotation The y rotation angle to draw the preview cube
 * @param zRotation The z rotation angle to draw the preview cube
 * @param axisId The type of rotation
 * @param x The current mouse x position (needed for EYE rotations)
 * @param y The current mouse y position (needed for EYE rotations)
 * @param isRotation Is the current action a rotation?
 */
protected void drawPreview(double scale, double horizShift, double vertShift,
					       double xRotation, double yRotation, double zRotation,
                           int axisId, int x, int y, boolean isRotation)
{
	if (chart3dArea == null) {
		return;
	}
	switch(chart3dArea.getViewport().getPreviewMethod()) {
	case JCViewport.PREVIEW_CUBE:
	    if (status == START || status == ANIMATING) {
			chart3dArea.drawPreview(scale, horizShift, vertShift,
			                        xRotation, yRotation, zRotation,
									axisId, x, y);
	    }
	    break;
	case JCViewport.PREVIEW_FULL:
	    // Note that this change will cause a repaint
        if (isRotation) {
	        chart3dArea.getView3d().setXYZRotation(xRotation,
                                                   yRotation,
                                                   zRotation);
        }
        else {
            chart3dArea.getViewport().setViewport(scale,
                                                  horizShift,
                                                  vertShift);
        }
	    break;
	}
}

/**
 *
 *
 * @param object Initialization parameters
 */
public void initialize(Object object)
{
    if (object instanceof JCChart3dArea) {
        setChart3dArea((JCChart3dArea)object);
    }
}

/**
 * Called by <code>JCChart3dArea</code> when the action is started.
 *
 * @param event The event for the action
 * @param x The x position to start the action
 * @param y The y position to start the action
 */
public abstract void start(InputEvent event, int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the action is to be animated.
 *
 * @param event The event for the action
 * @param x The current x position to animate the action
 * @param y The current y position to animate the action
 */
public abstract void animate(InputEvent event, int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the chart area has been refreshed
 * and the animation must be redrawn.
 *
 * @param x The x position to re-animate the action
 * @param y The y position to re-animate the action
 */
public abstract void reanimate(int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the action is over.
 *
 * @param event The event for the action
 * @param x The x position to end the action
 * @param y The y position to end the action
 */
public abstract void end(InputEvent event, int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public abstract void cancel();

/**
 * Return the JClass Chart 3D object with which this action handler is
 * associated.
 * @return The parent JCChart3d object
 */
public JCChart3d getChart3d()
{
	if (chart3dArea == null) {
        return(null);
	}
	return(chart3dArea.getChart3d());
}

}
