/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: EditAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:03 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d.actions;

import java.awt.Frame;
import java.awt.Label;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import javax.vecmath.Point3d;

import com.klg.jclass.chart3d.*;

/**
 * Handler object for edit actions within JClass Chart 3D Java2D.  When
 * a edit action is started, <code>JCChart3dArea</code> will create an
 * instance of <code>EditAction</code> and call <code>start()</code>,
 * <code>animate()</code> and <code>end()</code> at appropriate points.  A
 * mapping from event to action type is stored the chart's JCActionTable.
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 * @see com.klg.jclass.chart3d.j2d.actions.BaseAction
 */
public class EditAction extends BaseAction {

/**
 * Default constructor.
 */
public EditAction() {}

/** The dataview whose data is being edited */
protected Chart3dDataView dataView = null;

/** Points to the data object whose data is being edited. */
protected Chart3dData data = null;

/**
 * Data index (either (<i>x</i>, <i>y</i>) or (<i>series</i>, <i>point</i>))
 * pointing to the data point being edited.
 */
protected JCData3dIndex index = null;

/** Current  point */
protected Point p = null;

/** Save in case cancel is called */
protected double origZValue = Double.MAX_VALUE;

/**
 * Called by <code>JCChart3dArea</code> when the action is started.  For
 * the edit action, this method picks a point to be edited.
 *
 * @param event The event for the action
 * @param x The x position of the event
 * @param y The y position of the event
 */
public void start(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled) {
        return;
	}

    index = null;
	JCChart3d chart3d = getChart3d();
	p = new Point(chart3d.convertChartAreaXToX(x),
			      chart3d.convertChartAreaYToY(y));
    index = getChart3d().pick(p, null);
	if (index == null || index.getDataView() == null) {
        return;
	}

    dataView = index.getDataView();
	Chart3dDataModel ds = dataView.getElevationDataSource();
	if (ds == null || !(ds instanceof EditableChart3dDataModel)) {
		return;
	}

    data = dataView.getElevationData();
    if ((index instanceof JCData3dGridIndex && !(data instanceof Chart3dGridData)) ||
        (index instanceof JCData3dPointIndex && !(data instanceof Chart3dPointData)))
    {
        return;
    }

    if (index instanceof JCData3dGridIndex) {
        JCData3dGridIndex gridIndex = (JCData3dGridIndex)index;
        Chart3dGridData gridData = (Chart3dGridData)data;
		int xIndex = gridIndex.getX();
		int yIndex = gridIndex.getY();
		if (xIndex < 0) {
			return;
		}
        origZValue = gridData.getZValue(xIndex, yIndex);
    }
    else if (index instanceof JCData3dPointIndex) {
		if (index.getObject() == chart3d.getLegend()) {
			return;
		}
        JCData3dPointIndex pointIndex = (JCData3dPointIndex)index;
        Chart3dPointData pointData = (Chart3dPointData)data;
        int series = pointIndex.getSeries();
        int point  = pointIndex.getPoint();
		if (series < 0) {
			return;
		}
        Point3d pt = pointData.getSeries(series).getPoint(point);
        origZValue =  pt.z;
    }
	else {  // Could have picked the legend, header, footer
		return;
	}
	status = START;
}

/**
 * Called by <code>JCChartArea</code> when the action is to be
 * animated.  For the edit action, this method calls the <code>showEdit()</code>
 * method on the drawable.  This technique allows bar charts and
 * line charts to show data editing differently.
 *
 * @param event The event for the action
 * @param x The x position of the event
 * @param y The y position of the event
 */
public void animate(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled || status == NONE ||
	    index == null || dataView == null || data == null)
	{
        return;
	}
	status = ANIMATING;

	JCChart3d chart3d = getChart3d();
	p.setLocation(chart3d.convertChartAreaXToX(x),
			      chart3d.convertChartAreaYToY(y));
    double zValue = dataView.dragZValue(data, index, p.x, p.y);
    data.getDataHandler().editZValue(index, zValue);
}


/**
 * Called by <code>JCChar3dtArea</code> when the chart area has been refreshed
 * and the animation must be redrawn.
 *
 * @param x The x position of the event
 * @param y The y position of the event
 */
public void reanimate(int x, int y)
{
	animate(null, x, y);
}

/**
 * Called by <code>JCChart3dArea</code> when the action is over.  For the edit
 * action, this method calls <code>endEdit()</code> on the drawable, then edits
 * the point.  Note that currently, points can only be edited in the <i>z</i>
 * direction.
 *
 * @param event The event for the action
 * @param x The x position of the event
 * @param y The y position of the event
 */
public void end(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled || status == NONE ||
	    index == null || dataView == null || data == null)
	{
        return;
	}
	JCChart3d chart3d = getChart3d();
	chart3d.sendEvent();
    double xSave = p.getX();
    double ySave = p.getY();
	p.setLocation(chart3d.convertChartAreaXToX(x),
			      chart3d.convertChartAreaYToY(y));
    if (xSave != p.getX() || ySave != p.getY()) {
        double zValue = dataView.dragZValue(data, index, p.x, p.y);
        data.getDataHandler().editZValue(index, zValue);
    }
	status = NONE;
    index    = null;
    dataView = null;
    data     = null;
}

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public void cancel()
{
	if (index == null || dataView == null || data == null) {
        return;
	}
    data.getDataHandler().editZValue(index, origZValue);
	status = NONE;
    index    = null;
    dataView = null;
    data     = null;
}

}
