/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: RotateAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:03 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d.actions;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.vecmath.Point4d;
import javax.vecmath.Matrix4d;

import com.klg.jclass.chart3d.*;
import com.klg.jclass.util.JCNumberUtil;


/**
 * Handler object for rotate actions within JClass Chart 3D Java2D.  When
 * a rotate action is started, <code>JCChart3dArea</code> will create an
 * instance of <code>RotateAction</code> and call <code>start()</code>,
 * <code>animate()</code> and <code>end()</code> at appropriate points.  A
 * mapping from event to action type is stored the chart's JCActionTable.
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 * @see com.klg.jclass.chart3d.j2d.actions.BaseAction
 */
public class RotateAction extends BaseAction {

/** Point in chart area at which the rotate action was started. */
protected Point first = null;

/** Last point used for rotation. */
protected Point last = null;

/** Original X rotation value in case this action is cancelled */
protected double origXRotation = 0.0;

/** Original Y rotation value in case this action is cancelled */
protected double origYRotation = 0.0;

/** Original Z rotation value in case this action is cancelled */
protected double origZRotation = 0.0;

/** The current xRotation for this action */
protected double xRotation = 0.0;

/** The current yRotation for this action */
protected double yRotation = 0.0;

/** The current zRotation for this action */
protected double zRotation = 0.0;

/** The id of the axis on which to base the rotation */
protected int axisId = JCAxis.AXIS_NONE;

/** The rotation matrix used to extract the new rotation angles */
protected Matrix4d rotate = null;

/**
 * Default constructor.
 *
 */
public RotateAction() {
}

/**
 * Constructor which sets the axis on which to rotate.  JCAxis.AXIS_NONE means
 * use the general rotation method.
 * @param axisId The id of the axis on which to base the rotation
 */
public RotateAction(int axisId)
{
	this.axisId = axisId;
}

/**
 * Get the current x rotation angle
 * @return The current x rotation angle
 */
public double getXRotation()
{
	return(xRotation);
}

/**
 * Get the current y rotation angle
 * @return The current y rotation angle
 */
public double getYRotation()
{
	return(yRotation);
}

/**
 * Get the current z rotation angle
 * @return The current z rotation angle
 */
public double getZRotation()
{
	return(zRotation);
}

/**
 * Extracts the rotation angles from a rotation matrix.  The matrix <i>m</i> is
 * equated with the general rotation matrix and the resulting system is solved.
 * This gives nine equations and three unknowns.
 *
 * @param m The matrix from which rotation angles are extracted
 */
public void extractAngles(Matrix4d m)
{
	yRotation = Math.asin(m.m02);

	if (1.0 + Math.cos(yRotation) == 1.0) {
		xRotation = Math.atan2(m.m10, m.m11);
		zRotation = 0.0;
	}
	else {
		xRotation = Math.atan2(-m.m12, m.m22);
		zRotation = Math.atan2(-m.m01, m.m00);
	}

	xRotation = JCNumberUtil.radToDeg(xRotation);
	yRotation = JCNumberUtil.radToDeg(yRotation);
	zRotation = JCNumberUtil.radToDeg(zRotation);
}

/**
 * Called by <code>JCChart3dArea</code> when the action is started.  For
 * the rotate action, this method sets a series of initial
 * values.
 *
 * @param event The event for the action
 * @param x The x pixel position where rotation is starting
 * @param y The y pixel position where rotation is starting
 */
public void start(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled) {
		return;
	}
	first = new Point(x, y);
	last  = new Point(x, y);
	status = START;
    

	JCView3d view3d = chart3dArea.getView3d();
	xRotation = origXRotation = view3d.getXRotation();
	yRotation = origYRotation = view3d.getYRotation();
	zRotation = origZRotation = view3d.getZRotation();

	if (rotate == null) {
		rotate = new Matrix4d();
	}
}

/**
 * Called by <code>JCChart3dArea</code> when the action is to be
 * animated.  For the rotate action, this method calculates
 * new rotation and elevation values for the chart area.
 *
 * @param event The event for the action
 * @param x The x pixel position where the animation is happening
 * @param y The y pixel position where the animation is happening
 */
public void animate(InputEvent event, int x, int y) {
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
	if (last == null) {
		return;
	}
    if (event instanceof KeyEvent) {
        switch (((KeyEvent)event).getKeyCode()) {
        case KeyEvent.VK_UP:
            x = last.x - 1;
            y = last.y - 1;
            break;
        case KeyEvent.VK_DOWN:
            x = last.x + 1;
            y = last.y + 1;
            break;
        case KeyEvent.VK_LEFT:
            x = last.x - 1;
            y = last.y;
            break;
        case KeyEvent.VK_RIGHT:
            x = last.x + 1;
            y = last.y;
            break;
        }
    }

	if ((y == last.y) && (x == last.x)) {
		return;
	}

	status = ANIMATING;
    
	int dx = last.x - x;
	int dy = last.y - y;
	last.x = x;
	last.y = y;

	rotate.setIdentity();

	Rectangle  bounds    = chart3dArea.getBounds();
	Transform  transform = chart3dArea.getTransform();
    JCPlotCube plotCube  = chart3dArea.getPlotCube();

	switch (axisId) {
	case JCAxis.AXIS_X:
	case JCAxis.AXIS_Y:
	case JCAxis.AXIS_Z:
	    //	Constrain the rotation to a given axis.  The tricky
	    //	part is determining the angle of rotation.  Here we
	    //	use the distance of the point (dx, dy) from the line
	    //	that is parallel the screen mapped axis which passes
	    //	through the origin.

		// First project the given axis to pixel coordinates
        Point4d p1 = new Point4d(0.0, 0.0, 0.0, 1.0);
        Point4d p2 = new Point4d(0.0, 0.0, 0.0, 1.0);
        if (axisId == JCAxis.AXIS_X) {
            p1.x = -1.0;
            p2.x =  1.0;
        }
        else if (axisId == JCAxis.AXIS_Y) {
            p1.y = -1.0;
            p2.y =  1.0;
        }
        else {
            p1.z = -1.0;
            p2.z =  1.0;
        }

        rotate.mul(transform.getScreen(), transform.getPerspective());
        rotate.mul(transform.getRotation());
        Transform.scaleMatrix(rotate, plotCube.getXNormalizedScale(),
                              plotCube.getYNormalizedScale(),
		                      plotCube.getZNormalizedScale());
        Transform.transformPoint(rotate, p1);
        Transform.transformPoint(rotate, p2);

		// Calculate the normal to the screen
		double nx = p2.y - p1.y;
		double ny = p1.x - p2.x;
		double dr = nx*nx + ny*ny;
		if (dr < 100.0) {
			// Axis is pointing towards us: use default axis
			nx = 10.0;
			ny = 0.0;
			dr = 100.0;
		}

		// Calculate the distance to the line and use it as the angle */
		double angle = 100.0 * (dx * nx + dy * ny) / dr;

        rotate.setIdentity();
        Transform.rotateMatrix(rotate, angle, axisId, true);
        rotate.mul(transform.getRotation(), rotate);
		break;
	case JCAxis.AXIS_EYE :
		// Calculate the screen center of the cube

        p1 = new Point4d(0.0, 0.0, 0.0, 1.0);
        rotate.mul(transform.getScreen(), transform.getPerspective());
        rotate.mul(transform.getRotation());
        Transform.scaleMatrix(rotate, plotCube.getXNormalizedScale(),
                              plotCube.getYNormalizedScale(),
		                      plotCube.getZNormalizedScale());
        Transform.transformPoint(rotate, p1);

        // Calculate the angle change relative to the center
		angle = Math.atan2(y  - p1.y, x  - p1.x) -
			    Math.atan2(y - dy - p1.y, x - dx  - p1.x);

        rotate.setIdentity();
        Transform.rotateMatrix(rotate, JCNumberUtil.radToDeg(angle),
                               JCAxis.AXIS_Z, true);
        rotate.mul(transform.getRotation());
		break;
	case JCAxis.AXIS_NONE :
	default :
		//  Rotate the current orientation proportionally to
		//	dx and dy about the screen y and x axis respectively.
		//  The following code is explained in the article
		//	"The Rolling Ball" from Graphics Gems III pg. 51

		dr    = Math.sqrt((double)dx*dx + (double)dy*dy);
		angle = 180.0 * dr / bounds.getWidth();
		Transform.rotateMatrix(rotate, -angle * dy / dr, JCAxis.AXIS_X, true);
		Transform.rotateMatrix(rotate, -angle * dx / dr, JCAxis.AXIS_Y, true);
		rotate.mul(transform.getRotation());
		break;
	}

	extractAngles(rotate);
	xRotation += 90.0;     // +90 due to JClass 3D quirkiness

	JCViewport viewport = chart3dArea.getViewport();
	drawPreview(viewport.getScale(), viewport.getHorizontalShift(),
	            viewport.getVerticalShift(), xRotation, yRotation, zRotation,
				axisId, x, y, true);
}

/**
 * Called by <code>JCChartArea</code> when the chart area has been
 * refreshed and the animation must be redrawn.
 *
 * @param x The x pixel position where reanimation is happening
 * @param y The y pixel position where reanimation is happening
 */
public void reanimate(int x, int y) {
	if (chart3dArea == null || !enabled || status != ANIMATING) {
		return;
	}
    
	JCViewport viewport = chart3dArea.getViewport();
	drawPreview(viewport.getScale(), viewport.getHorizontalShift(),
	            viewport.getVerticalShift(), xRotation, yRotation, zRotation,
				JCAxis.AXIS_NONE, x, y, true);
}

/**
 * Called by <code>JCChart3dArea</code> when the action is over.  Does nothing
 * for the rotate action.
 *
 * @param event The event for the action
 * @param x The x pixel position where the user finished rotating
 * @param y The y pixel position where the user finished rotating
 */
public void end(InputEvent event, int x, int y) {
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
    
	JCChart3d chart3d = chart3dArea.getChart3d();
	if (chart3d != null) {
		chart3d.sendEvent();
	}

	// Adjust values so that they lie within [0..360]
	if (xRotation < 0.0) {
		xRotation += 360.0;
	}
	if (yRotation < 0.0) {
		yRotation += 360.0;
	}
	if (zRotation < 0.0) {
		zRotation += 360.0;
	}

	status = NONE;
	last.x = last.y = first.x = first.y = -1;
    
	chart3dArea.getView3d().setXYZRotation(xRotation, yRotation, zRotation);
}

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public void cancel()
{
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
	status = NONE;
	last.x = last.y = first.x = first.y = -1;
    
	chart3dArea.getView3d().setXYZRotation(origXRotation,
                                           origYRotation,
                                           origZRotation);
}

}
