/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ScaleAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:04 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d.actions;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;

import com.klg.jclass.chart3d.*;
import com.klg.jclass.util.JCNumberUtil;


/**
 * Handler object for scale actions within JClass Chart 3D Java2D.  When
 * a scale action is started, <code>JCChart3dArea</code> will create an
 * instance of <code>ScaleAction</code> and call <code>start()</code>,
 * <code>animate()</code> and <code>end()</code> at appropriate points.  A
 * mapping from event to action type is stored the chart's JCActionTable.
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 * @see com.klg.jclass.chart3d.j3d.actions.BaseAction
 */
public class ScaleAction extends BaseAction {

/** Point at which the scale action started */
protected Point first = null;

/** Last point used to do scaling */
protected Point last = null;

/** The current scale value for this action */
protected double scale               = 1.0;

/** The current horizontal shift value for this action */
protected double horizontalShift     = 0.0;

/** The current vertical shift value for this action */
protected double verticalShift       = 0.0;

/** The original scale value in case this action is cancelled */
protected double origScale           = 1.0;

/** The original horizontal shift value in case this action is cancelled */
protected double origHorizontalShift = 0.0;

/** The original vertical shift in case this action is cancelled */
protected double origVerticalShift   = 0.0;

/**
 * Default constructor
 */
public ScaleAction() {}

/**
 * Called by JCChart3dArea when the action is started.
 * For the scale action, this method saves the start point
 * as the anchor for the scale rectangle.
 *
 * @param event The InputEvent that triggered this action
 * @param x The x pixel position where scaling is starting
 * @param y The y pixel position where scaling is starting
 */
public void start(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled) {
		return;
	}
	first = new Point(x, y);
	last  = new Point(x, y);
	status = START;
    

	JCViewport viewport = chart3dArea.getViewport();
	origScale           = viewport.getScale();
	origHorizontalShift = viewport.getHorizontalShift();
	origVerticalShift   = viewport.getVerticalShift();
}

/**
 * Called by JCChart3dArea when the action is to be animated.  For the scale
 * action, the user can control the size of the plotCube view mouse actions.
 *
 * @param event The InputEvent that triggered this action
 * @param x The x pixel position where animation is happening
 * @param y The y pixel position where animation is happening
 */
public void animate(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}

	if (last == null) {
		return;
	}
    if (event instanceof KeyEvent) {
        switch (((KeyEvent)event).getKeyCode()) {
        case KeyEvent.VK_PAGE_UP:
            x = last.x;
            y = last.y + 1;
        break;
        case KeyEvent.VK_PAGE_DOWN:
            x = last.x;
            y = last.y - 1;
        break;
        }
    }

	if ((y == last.y) && (x == last.x)) {
		return;
	}

	status = ANIMATING;
    
	int dx = last.x - x;
	int dy = last.y - y;
	last.x = x;
	last.y = y;

	/* Create the previous view, before calculating new scale */
	JCViewport viewport   = chart3dArea.getViewport();
	Transform transform   = chart3dArea.getTransform();
	JCBox currentViewport = transform.getCurrentViewport();
	JCBox viewNew1 = new JCBox(currentViewport.x, currentViewport.y,
							   currentViewport.width, currentViewport.height);
	transform.adjustForUserViewport(viewNew1, viewport.getScale(),
									viewport.getHorizontalShift(),
									viewport.getVerticalShift());

	Rectangle bounds = chart3dArea.getChart3d().getBounds();
	double length = -(double)dy / bounds.getHeight();
	double factor = Math.pow(10.0, length);
	scale = viewport.getScale() * factor;
    scale = JCChart3dUtil.clamp(scale,
								JCViewport.MIN_VIEW_SCALE,
								JCViewport.MAX_VIEW_SCALE);

	// Construct the new view, and center it
	JCBox viewNew2 = new JCBox(currentViewport.x, currentViewport.y,
							   currentViewport.width, currentViewport.height);
	transform.adjustForUserViewport(viewNew2, scale,
									viewport.getHorizontalShift(),
									viewport.getVerticalShift());
	viewNew2.x += (viewNew1.width - viewNew2.width) / 2.0;
	viewNew2.y += (viewNew1.height - viewNew2.height) / 2.0;

	//  Calculate translation factors
	horizontalShift = (viewNew2.x - currentViewport.x) / currentViewport.width;
	verticalShift   = (viewNew2.y - currentViewport.y) / currentViewport.height;

	JCView3d view3d = chart3dArea.getView3d();
	drawPreview(scale, horizontalShift, verticalShift, view3d.getXRotation(),
				view3d.getYRotation(), view3d.getZRotation(),
				JCAxis.AXIS_NONE, x, y, false);
}

/**
 * Called by JCChart3dArea when the chart area has been refreshed and the
 * animation must be redrawn.
 *
 * @param event The InputEvent that triggered this action
 * @param x The x pixel position where reanimation is happening
 * @param y The y pixel position where reanimation is happening
 */
public void reanimate(int x, int y)
{
	if (chart3dArea == null || !enabled || status != ANIMATING) {
		return;
	}
    
	JCView3d view3d = chart3dArea.getView3d();
	drawPreview(scale, horizontalShift, verticalShift, view3d.getXRotation(),
				view3d.getYRotation(), view3d.getZRotation(),
				JCAxis.AXIS_NONE, x, y, false);
}

/**
 * Called by JCChar3dArea when the action is over.
 * For the scale action, this method takes the start point
 * and the current point and uses them to scale along the
 * horizontal and vertical action axes.
 *
 * @param x The x pixel position where the user finished scaling
 * @param y The y pixel position where the user finished scaling
 */
public void end(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
    
	JCChart3d chart3d = chart3dArea.getChart3d();
	if (chart3d != null) {
		chart3d.sendEvent();
	}

	status = NONE;
    first  = null;
    last   = null;
    
	chart3dArea.getViewport().setViewport(scale, horizontalShift,
	                                      verticalShift);
}

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public void cancel()
{
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
	status = NONE;
    first  = null;
    last   = null;
    
	chart3dArea.getViewport().setViewport(origScale,
                                          origHorizontalShift,
                                          origVerticalShift);
}

}
