/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: TranslateAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:04 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d.actions;

import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.vecmath.Matrix4d;

import com.klg.jclass.chart3d.*;


/**
 * Handler object for translate actions within JClass Chart 3D Java2D.  When
 * a translate action is started, <code>JCChart3dArea</code> will create an
 * instance of <code>TranslateAction</code> and call <code>start()</code>,
 * <code>animate()</code> and <code>end()</code> at appropriate points.  A
 * mapping from event to action type is stored the chart's JCActionTable.
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 * @see com.klg.jclass.chart3d.j2d.actions.BaseAction
 */
public class TranslateAction extends BaseAction {

/** Point at which the translate action started. */
protected Point first = null;

/** Last point used to do translationg */
protected Point last = null;

/** The current horizontal shift for this action */
protected double horizontalShift = 0.0;
/** The current vertical shift for this action */
protected double verticalShift   = 0.0;

/** The original horizontal shift in case a cancel happens */
protected double origHorizontalShift = 0.0;
/** The original vertical shift in case a cancel happens */
protected double origVerticalShift   = 0.0;

/**
 * Default constructor.
 */
public TranslateAction() {}

/**
 * Called by <code>JCChart3dArea</code> when the action is started.
 * For the translate action, this method stores the initial point and the
 * original horizontal and vertical shift (in case of a cancel action).
 *
 * @param event The event for the action
 * @param x The x pixel position where translation is starting
 * @param y The y pixel position where translation is starting
 */
public void start(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled) {
		return;
	}
	first = new Point(x, y);
	last  = new Point(x, y);
	status = START;
    

	JCViewport viewport = chart3dArea.getViewport();
	origHorizontalShift = viewport.getHorizontalShift();
	origVerticalShift   = viewport.getVerticalShift();
}

/**
 * Called by <code>JCChartArea</code> when the action is to be animated.  For
 * the translate action, this method moves the horizontal and vertical shifts
 * relative to the distance from the last point.
 *
 * @param event The event for the action
 * @param x The x pixel position where animation is happening
 * @param y The y pixel position where animation is happening
 */
public void animate(InputEvent event, int x, int y)
{
    
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
	if (last == null) {
		return;
	}
    if (event instanceof KeyEvent) {
        switch (((KeyEvent)event).getKeyCode()) {
        case KeyEvent.VK_UP:
            x = last.x;
            y = last.y - 1;
        break;
        case KeyEvent.VK_DOWN:
            x = last.x;
            y = last.y + 1;
        break;
        case KeyEvent.VK_LEFT:
            x = last.x - 1;
            y = last.y;
        break;
        case KeyEvent.VK_RIGHT:
            x = last.x + 1;
            y = last.y;
        break;
        }
    }

	if ((y == last.y) && (x == last.x)) {
		return;
	}

	status = ANIMATING;
    
	int dx = last.x - x;
	int dy = last.y - y;
	last.x = x;
	last.y = y;

	/* Create the previous view, before calculating new translate */
	JCViewport viewport   = chart3dArea.getViewport();
	Transform transform   = chart3dArea.getTransform();
	JCBox currentViewport = transform.getCurrentViewport();
	Matrix4d iScreen = transform.getInverseScreen();
	horizontalShift = viewport.getHorizontalShift() +
	                  (double)dx * iScreen.m00 / currentViewport.width;
	verticalShift   = viewport.getVerticalShift() +
	                  (double)dy * iScreen.m11 / currentViewport.height;

	JCView3d view3d = chart3dArea.getView3d();
	drawPreview(viewport.getScale(), horizontalShift, verticalShift,
	            view3d.getXRotation(), view3d.getYRotation(),
				view3d.getZRotation(), JCAxis.AXIS_NONE, x, y, false);
}

/**
 * Called by <code>JCChart3dArea</code> when the chart area has been
 * refreshed and the animation must be redrawn.
 *
 * @param x The x pixel position where reanimation is happening
 * @param y The y pixel position where reanimation is happening
 */
public void reanimate(int x, int y)
{
	if (chart3dArea == null || !enabled || status != ANIMATING) {
		return;
	}
    
	JCViewport viewport = chart3dArea.getViewport();
	JCView3d   view3d    = chart3dArea.getView3d();
	drawPreview(viewport.getScale(), horizontalShift, verticalShift,
	            view3d.getXRotation(), view3d.getYRotation(),
				view3d.getZRotation(), JCAxis.AXIS_NONE, x, y, false);
}

/**
 * Called by <code>JCChartA3drea</code> when the action is over.  This
 * method does nothing for translate actions.
 *
 * @param event The event for the action
 * @param x The x pixel position where the user finished translating
 * @param y The y pixel position where the user finished translating
 */
public void end(InputEvent event, int x, int y)
{
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
    
	JCChart3d chart3d = chart3dArea.getChart3d();
	if (chart3d != null) {
		chart3d.sendEvent();
	}

	status = NONE;
    first  = null;
    last   = null;
    
    JCViewport viewport = chart3dArea.getViewport();
	viewport.setViewport(viewport.getScale(), horizontalShift, verticalShift);
}

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public void cancel()
{
	if (chart3dArea == null || !enabled || status == NONE) {
		return;
	}
	status = NONE;
    first  = null;
    last   = null;
    
    JCViewport viewport = chart3dArea.getViewport();
	viewport.setViewport(viewport.getScale(),
                         origHorizontalShift,
                         origVerticalShift);
}

}
