/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ZoomAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:04 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j2d.actions;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import javax.vecmath.Matrix4d;

import com.klg.jclass.chart3d.*;


/**
 * Handler object for zoom actions within JClass Chart 3D Java2D.  When
 * a zoom action is started, <code>JCChart3dArea</code> will create an
 * instance of <code>ZoomAction</code> and call <code>start()</code>,
 * <code>animate()</code> and <code>end()</code> at appropriate points.  A
 * mapping from event to action type is stored the chart's JCActionTable.
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 * @see com.klg.jclass.chart3d.j3d.actions.BaseAction
 */
public class ZoomAction extends BaseAction {

/** Point at which the zoom action started. */
Point first = null;

/** Last point used to draw the bounding box. */
Point last = null;

/**
 * Default constructor.
 */
public ZoomAction() {}

/**
 * Called by <code>JCChart3dArea</code> when the action is started.
 * For the zoom action, this method saves the start point
 * as the anchor for the zoom rectangle.
 *
 * @param event The event for the action
 * @param x The x pixel position where zooming is starting
 * @param y The y pixel position where zooming is starting
 */
public void start(InputEvent event, int x, int y)
{
	if (getChart3dArea() == null || !enabled) {
		return;
	}
	first = new Point(x, y);
	last  = null;
	status = START;
    
}

/**
 * Called by <code>JCChart3dArea</code> when the action is to be
 * animated.  For the zoom action, thid method
 * undraws the previous rubber-band rectangle and
 * draws a new rubber-band rectangle from the
 * start point to the current point.  The rectangle
 * represents the zoom area.
 *
 * @param event The event for the action
 * @param x The x pixel position where animation is happening
 * @param y The y pixel position where animation is happening
 */
public void animate(InputEvent event, int x, int y)
{
	if (getChart3dArea() == null || !enabled || status == NONE) {
		return;
	}

	if (first == null) {
		return;
	}
	status = ANIMATING;
    
	if (last == null) {
		last = new Point(x, y);
		getChart3dArea().drawZoomRectangle(last, first);
		return;
	}
    getChart3dArea().drawZoomRectangle(last, first);
	last.x = x;
	last.y = y;
    getChart3dArea().drawZoomRectangle(last, first);
}

/**
 * Called by <code>JCChart3dArea</code> when the chart area has been
 * refreshed and the animation must be redrawn.
 *
 * @param x The x pixel position where re-animation is happening
 * @param y The y pixel position where re-animation is happening
 */
public void reanimate(int x, int y)
{
	if (getChart3dArea() == null || !enabled || first == null || status != ANIMATING) {
		return;
	}
    
	last = new Point(x, y);
    getChart3dArea().drawZoomRectangle(last, first);
}

/**
 * Called by <code>JCChar3dArea</code> when the action is over.
 * For the zoom action, this method takes the start point
 * and the current point and uses them to zoom along the
 * horizontal and vertical action axes.
 *
 * @param event The event for the action
 * @param x The x pixel position where the user finished zooming
 * @param y The y pixel position where the user finished zooming
 */
public void end(InputEvent event, int x, int y)
{
	if (getChart3dArea() == null || !enabled || status == NONE) {
		return;
	}
    
	JCChart3d chart3d = getChart3dArea().getChart3d();
	if (chart3d != null) {
		chart3d.sendEvent();
	}
	status = NONE;

	JCViewport viewport   = getChart3dArea().getViewport();
	if (first != null && last != null &&
	    first.x != last.x && first.y != last.y)
	{
        getChart3dArea().drawZoomRectangle(last, first);

		Rectangle rect = new Rectangle();

		rect.x       = Math.min(first.x, last.x);
		rect.y       = Math.min(first.y, last.y);
		rect.width   = Math.max(first.x, last.x);
		rect.height  = Math.max(first.y, last.y);
		rect.width  -= rect.x;
		rect.height -= rect.y;

	    Transform transform       = getChart3dArea().getTransform();
	    JCBox     currentViewport = transform.getCurrentViewport();
		Matrix4d  iScreen         = transform.getInverseScreen();

		JCBox view = new JCBox();
		view.x = rect.getX() * iScreen.m00 + iScreen.m03;
		view.y = (rect.getY() + rect.getHeight()) * iScreen.m11 + iScreen.m13;
		view.width  = rect.getWidth() * iScreen.m00;
		view.height = rect.getHeight() * (-iScreen.m11);
		double viewXScale = currentViewport.width  / view.width;
		double viewYScale = currentViewport.height / view.height;

	    double scale           = Math.min(viewXScale, viewYScale);
	    double horizontalShift =
		    (view.x - currentViewport.x) / currentViewport.width;
	    double verticalShift   =
		    (view.y - currentViewport.y) / currentViewport.height;
		viewport.setViewport(scale, horizontalShift, verticalShift);
	}

	first = null;
	last  = null;
}

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public void cancel()
{
	if (getChart3dArea() == null || !enabled || status == NONE) {
		return;
	}

    
	if (first != null && last != null &&
	    first.x != last.x && first.y != last.y)
	{
        getChart3dArea().drawZoomRectangle(last, first);
	}

	status = NONE;
	first  = null;
	last   = null;
}

}
