/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:06 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.chart3d.j3d.actions;

import com.klg.jclass.chart3d.*;
import com.klg.jclass.chart3d.j3d.*;

import java.io.Serializable;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.Point;

import com.sun.j3d.utils.picking.*;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;

/**
 * BaseAction is an abstract base class for the Java3D actions of JCChart 3D.
 * It implements the JCAction interface used for JCChart 3D actions. It defines
 * methods for starting, animating, ending and cancelling an action.
 *
 * <code>JCChart3dArea</code> will create an instance of the appropriate
 * <code>JCAction</code> when an action is started, and will call
 * <code>start()</code>, <code>animate()</code> and <code>end()</code> during
 * <code>mouseDown()</code>, <code>mouseDrag()</code> and
 * <code>mouseUp()</code> respectively.  The appropriate <code>JCAction</code>
 * is dependent on the event.  A mapping from event to action type is stored
 * the chart's JCActionTable
 *
 * @see com.klg.jclass.chart3d.JCAction
 * @see com.klg.jclass.chart3d.JCActionTable
 */
public abstract class BaseAction implements JCAction, Serializable {

public final static int NONE      = 0;
public final static int START     = 1;
public final static int ANIMATING = 2;

protected PickCanvas            pickCanvas = null;
protected ActionTransformGroup  transformGroup = null;
protected Transform3D           transformX = null;
protected Transform3D           transformY = null;
protected Transform3D           currXform = null;
protected TransformGroup        ViewerTG = null;
protected Point                 last = null;
protected double                x_factor = 0;
protected double                y_factor = 0;
protected double                z_factor = 0;
protected boolean               invert = false;
protected Vector3d              translation = new Vector3d();
protected JC3DUniverse          virtualUniverse = null;
protected JCChart3dJava3d       chart3d = null;
protected JCPlotCubeJava3d      plotCube = null;

/** Parent object on which the action occurs. */
protected JCChart3dArea chart3dArea = null;

/** Is this action enabled */
protected boolean enabled = true;

protected int status = NONE;

/**
 * Constructs a BaseAction.  Note that it nees to be connected to a
 * <code>JCChart3dArea</code> instance so this needs to be set later.
 */
public BaseAction() {}

/**
 * Sets the parent JCChart3dArea object
 *
 * @param chart3dArea The new JCChart3dArea object
 */
public void setChart3dArea(JCChart3dArea chart3dArea)
{
	this.chart3dArea = chart3dArea;
}

/**
 * Returns the parent JCChart3dArea object
 *
 * @return The current JCChart3dArea object
 */
public JCChart3dArea getChart3dArea()
{
	return(chart3dArea);
}

/**
 * Sets whether this action handler is enabled
 *
 * @param enabled Is this action handler enabled?
 */
public void setEnabled(boolean enabled)
{
	this.enabled = enabled;
}

/**
 * Is this action handler enabled?
 *
 * @return Is this action handler enabled?
 */
public boolean isEnabled()
{
	return(enabled);
}

/**
 * Initialize the action with the objects need to carry out the action, such
 * as the chart area, virtualUniverse, plotCube, etc.  These are store in an
 * ActionHandlerParams object.
 *
 * @param object Initialization parameters
 */
public void initialize(Object object)
{
    if (object instanceof ActionHandlerParams) {
		ActionHandlerParams params = (ActionHandlerParams)object;
		ViewerTG = params.getViewPlatformTransformGroup();
		pickCanvas = new PickCanvas(params.getCanvas3D(),
									params.getObjectRoot());
		JCChart3dAreaJava3d area = params.getChart3dAreaJava3d();
        setChart3dArea(area);
		virtualUniverse = area.getChart3dUniverse().getVirtualUniverse();
		plotCube = (JCPlotCubeJava3d)area.getPlotCube();
    }
}

/**
 * Called by <code>JCChart3dArea</code> when the action is started.
 *
 * @param event The InputEvent that triggered this action
 * @param x The x position to start the action
 * @param y The y position to start the action
 */
public abstract void start(InputEvent event, int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the action is to be animated.
 *
 * @param event The InputEvent that triggered this action
 * @param x The current x position to animate the action
 * @param y The current y position to animate the action
 */
public abstract void animate(InputEvent event, int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the chart area has been refreshed
 * and the animation must be redrawn.
 *
 * @param x The x position to re-animate the action
 * @param y The y position to re-animate the action
 */
public abstract void reanimate(int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the action is over.
 *
 * @param event The InputEvent that triggered this action
 * @param x The x position to end the action
 * @param y The y position to end the action
 */
public abstract void end(InputEvent event, int x, int y);

/**
 * Called by <code>JCChart3dArea</code> when the action is cancelled
 */
public abstract void cancel();

/**
 * Return the JClass Chart 3D object witch which this action handler is
 * associated.
 * @return The part JCChart3d object
 */
public JCChart3d getChart3d()
{
	if (chart3dArea == null) {
        return(null);
	}
	return(chart3dArea.getChart3d());
}

public void setScaleFactorX(double scalefactor)
{
	x_factor = scalefactor;
}

public void setScaleFactorY(double scalefactor)
{
	y_factor = scalefactor;
}

public void setScaleFactorZ(double scalefactor)
{
	z_factor = scalefactor;
}

public double getScaleFactorX()
{
	return(x_factor);
}

public double getScaleFactorY()
{
	return(y_factor);
}

public double getScaleFactorZ()
{
	return(z_factor);
}

protected void doStart(InputEvent event, int x, int y)
{
    last = new Point(x, y);
    if (event instanceof KeyEvent) {
        transformGroup = plotCube.getTransformGroup();
    }
    else {
	    TransformGroup tg = null;
	    pickCanvas.setShapeLocation(x, y);
	    pickCanvas.setTolerance(4.0f);
	    PickResult[] pr = pickCanvas.pickAllSorted();
	    if (pr != null) {
		    for (int i = 0; i < pr.length; i++) {
				tg = (TransformGroup)pr[i].getNode(PickResult.TRANSFORM_GROUP);
			    if (tg != null && (tg instanceof ActionTransformGroup)) {
				    transformGroup = (ActionTransformGroup)tg;
				    break;
			    }
		    }
		}
	}
	if (transformGroup != null) {
		currXform  = new Transform3D();
		transformX = new Transform3D();
		transformY = new Transform3D();
	}
}

protected void doEnd(InputEvent event, int x, int y)
{
	//  We should recompute point of view, then recalculate axis annotations.

//	if (transformGroup != null ) {
//		transformGroup.actionEnd();
//	}
	transformGroup = null;
    last = null;
}

}
