/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseColumn.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:08 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.io.Serializable;
import java.sql.Types;

import com.klg.jclass.datasource.MetaDataModel;

/**
 * An implementation of ColumnModel.
 * Extend it and set each of its properties.
 * <P>
 * Calls to MetaDataModel.getJavaColumnType call getMetaColumnType() below.
 *<P>
 * ResultSetMetaData methods are mirrored here, see their
 * documentation for descriptions.
 * Some drivers do not support all ResultSetMetaData methods.
 * If they do not the following defaults are used:
 * <BR>
 * <TABLE BORDER>
 * <TR><TH>NAME</TH><TH>VALUE</TH></TR>
 * <TR><TD>catalog</TD><TD>null</TD></TR>
 * <TR><TD>displayWidth</TD><TD>25</TD></TR>
 * <TR><TD>name</TD><TD>null</TD></TR>
 * <TR><TD>label</TD><TD>null</TD></TR>
 * <TR><TD>type</TD><TD>java.sql.Types.OTHER</TD></TR>
 * <TR><TD>typeName</TD><TD>"OTHER"</TD></TR>
 * <TR><TD>precision</TD><TD>5</TD></TR>
 * <TR><TD>scale</TD><TD>5</TD></TR>
 * <TR><TD>schema</TD><TD>null</TD></TR>
 * <TR><TD>table</TD><TD>null</TD></TR>
 * <TR><TD>isAutoIncrement</TD><TD>false</TD></TR>
 * <TR><TD>isCaseSensitive</TD><TD>false</TD></TR>
 * <TR><TD>isCurrency</TD><TD>false</TD></TR>
 * <TR><TD>isDefinitelyWritable</TD><TD>true</TD></TR>
 * <TR><TD>isNullable</TD><TD>ResultSetMetaData.columnNullable</TD></TR>
 * <TR><TD>isReadOnly</TD><TD>false</TD></TR>
 * <TR><TD>isSearchable</TD><TD>true</TD></TR>
 * <TR><TD>isSigned</TD><TD>false</TD></TR>
 * <TR><TD>isWritable</TD><TD>true</TD></TR>
 * </TABLE>
 */

public class BaseColumn implements ColumnModel, Serializable {

static final long serialVersionUID = 757537678176189885L;

protected static final int NOT_SET = -999;
protected String name = null, table = null, catalog = null, schema = null;
protected String label = null, typeName = "OTHER", identifier, columnClassName;
protected int type = java.sql.Types.OTHER, displayWidth = 25, precision = 5, scale = 5, isNullable, metaType = NOT_SET;
protected boolean isCurrency = false, isReadOnly, isAutoIncrement = false, isCaseSensitive = false;
protected boolean isSearchable = true, isDefinitelyWritable = true, isSigned = false, isWritable = true;


public BaseColumn() {
}

/**
 * Constructor
 * @param name the name of this column
 * @param type this column's type, one of,
 * <ul>
 *     <li>TYPE_BOOLEAN</li>
 *     <li>TYPE_SQL_DATE</li>
 *     <li>TYPE_DOUBLE</li>
 *     <li>TYPE_FLOAT</li>
 *     <li>TYPE_INTEGER</li>
 *     <li>TYPE_STRING</li>
 *     <li>TYPE_BIG_DECIMAL</li>
 *     <li>TYPE_LONG</li>
 *     <li>TYPE_SQL_TIME </li>
 *     <li>TYPE_SQL_TIMESTAMP</li>
 *     <li>TYPE_OBJECT</li>
 *     <li>TYPE_BYTE</li>
 *     <li>TYPE_SHORT</li>
 *     <li>TYPE_BYTE_ARRAY</li>
 *     <li>TYPE_UTIL_DATE</li>
 * </ul>
 */
public BaseColumn(String name, int type) {
	setColumnName(name);
	setMetaColumnType(type);	
}

/**
 * Return one of MetaDataModel types.
 * Java object types and java.sql.Types types are mapped to
 * MetaDataModel types.
 * @return this column's type as one of the MetaDataModel type constants.
 * @see #getColumnType
 */
public int getMetaColumnType() {
    if (metaType == NOT_SET) {
        setMetaColumnTypeFromSqlType(getColumnType());
    }
    return metaType;
}

/**
 * Is this column's type date/time/timestamp?
 * @return true if this column's type is date/time/timestamp, false otherwise
 */
public boolean isTimeDate() {
    return type == Types.DATE || type == Types.TIME ||
    type == Types.TIMESTAMP ||
    type == 9 || type == 10 || type == 11;
//The last three checks represent a workaround for a jdbc_odbc_bridge bug
//in Sun's bridge driver; the bug already filed against Sun (Bug Id  4038518)
//bug synopsis: JdbcOdbcDriver does not bridge ANSI SQL DATE type
//JDBC only converts 9, 10, and 11 for ODBC 2.0 drivers;
//it inhibits the conversion for ODBC 3.0 drivers. A correct ODBC 3.0
//driver uses the 91, 92, and 93 values for data/time types.
//Since JDBC does not define the 9, 10 and 11 values the bridge must
//prevent their use regardless of the source. Until JDBC defines them for
//some other purpose there is no harm applying the same conversion for
//ODBC 3.0.

}

public String getCatalogName() {
    return catalog;
}
public int getColumnDisplaySize() {
    return displayWidth;
}
public String getColumnLabel() {
    return label != null ? label : name;
}
public String getColumnName() {
    return name;
}

public int getColumnType() {
    return type;
}
public String getColumnTypeName() {
    return typeName;
}
public int getPrecision() {
    return precision;
}
public int getScale() {
    return scale;
}
public String getSchemaName() {
    return schema;
}
public String getTableName() {
    return table;
}
public boolean isAutoIncrement() {
    return isAutoIncrement;
}
public boolean isCaseSensitive() {
    return isCaseSensitive;
}
public boolean isCurrency() {
    return isCurrency;
}
public boolean isDefinitelyWritable() {
    return isDefinitelyWritable;
}
public int isNullable() {
    return isNullable;
}
public boolean isReadOnly() {
    return isReadOnly;
}
public boolean isSearchable() {
    return isSearchable;
}
public boolean isSigned() {
    return isSigned;
}
public boolean isWritable() {
    return isWritable;
}

/**
 * Return a string which uniquely identifies this column.
 * That is, determine if it is identified by an alias, by
 * it name, or by fully qualified name.  For example,
 * colA A, or colA, or qualifier.colA.
 * If identifier is null, column name is returned.
 * @return unique identifier for this column
 */
public String getIdentifier() {
    return identifier == null ? getColumnName() : identifier;
}

public String getColumnClassName() {
    return columnClassName;
}

/**
 * Set the metaType for this column.
 * Maps the passed in java.sql.Types type to
 * a MetaDataModel type.
 * @param metaType a int which is the MetaDataModel type for the column
 * @see com.klg.jclass.datasource.MetaDataModel
 */
public void setMetaColumnType(int metaType) {
    this.metaType = metaType;
}

/**
 * Set the metaType for this column.
 * Maps the passed in java.sql.Types type to
 * a MetaDataModel type.
 * @param sqlType a int which is the java.sql.Types type for the column
 */
public void setMetaColumnTypeFromSqlType(int type) {
    switch (type) {
    case Types.CHAR:
    case Types.VARCHAR:
    case Types.LONGVARCHAR:
        metaType = MetaDataModel.TYPE_STRING;
        break;
    case Types.NUMERIC:
    case Types.DECIMAL:
        metaType = MetaDataModel.TYPE_BIG_DECIMAL;
        break;
    case Types.BIT:
        metaType = MetaDataModel.TYPE_BOOLEAN;
        break;
    case Types.TINYINT:
    case Types.SMALLINT:
    case Types.INTEGER:
        metaType = MetaDataModel.TYPE_INTEGER;
        break;
    case Types.BIGINT:
        metaType = MetaDataModel.TYPE_LONG;
        break;
    case Types.REAL:
        metaType = MetaDataModel.TYPE_FLOAT;
        break;
    case Types.FLOAT:
    case Types.DOUBLE:
        metaType =  MetaDataModel.TYPE_DOUBLE;
        break;
    case Types.DATE:
        metaType =  MetaDataModel.TYPE_SQL_DATE;
        break;
    case Types.TIME:
        metaType =  MetaDataModel.TYPE_SQL_TIME;
        break;
    case Types.TIMESTAMP:
        metaType =  MetaDataModel.TYPE_SQL_TIMESTAMP;
        break;
    case Types.BINARY:
    case Types.VARBINARY:
    case Types.LONGVARBINARY:
        metaType =  MetaDataModel.TYPE_BYTE_ARRAY;
        break;
    default:
        metaType =  MetaDataModel.TYPE_OBJECT;
        break;
    }
}

/* setters */
public void setCatalogName(String catalog) {
    this.catalog = catalog;
}
public void setColumnDisplaySize(int displayWidth) {
    this.displayWidth = displayWidth;
}
public void setColumnLabel(String label) {
    this.label = label;
}
public void setColumnName(String name) {
    this.name = name;
}

/**
 * Set the java.sql.Types type for this
 * column.
 */
public void setColumnType(int type) {
    this.type = type;
}
public void setColumnTypeName(String typeName) {
    this.typeName = typeName;
}
public void setPrecision(int precision) {
    this.precision = precision;
}
public void setScale(int scale) {
    this.scale = scale;
}
public void setSchemaName(String schema) {
    this.schema = schema;
}
public void setTableName(String table) {
    this.table = table;
}
public void setAutoIncrement(boolean isAutoIncrement) {
    this.isAutoIncrement = isAutoIncrement;
}
public void setCaseSensitive(boolean isCaseSensitive) {
    this.isCaseSensitive = isCaseSensitive;
}
public void setCurrency(boolean isCurrency) {
    this.isCurrency = isCurrency;
}
public void setDefinitelyWritable(boolean isDefinitelyWritable) {
    this.isDefinitelyWritable = isDefinitelyWritable;
}
public void setNullable(int isNullable) {
    this.isNullable = isNullable;
}
public void setReadOnly(boolean isReadOnly) {
    this.isReadOnly = isReadOnly;
}
public void setSearchable(boolean isSearchable) {
    this.isSearchable = isSearchable;
}
public void setSigned(boolean isSigned) {
    this.isSigned = isSigned;
}
public void setWritable(boolean isWritable) {
    this.isWritable = isWritable;
}

public void setColumnClassName(String columnClassName) {
    this.columnClassName = columnClassName;
}

public void setIdentifier(String identifier) {
    this.identifier = identifier;
}

}
