/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseDataTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:08 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * This class provides a default implementation of DataTableModel and
 * DataTableAbstractionLayer interfaces.
 * <P>
 * Instances of this class provide basic storage, retrieval and manipulation
 * operations on data rows.
 * <P>
 * This class can be used without extending it.  In this case you must create
 * and populate rows "manually".  For example,
 * <pre><code>
 *     BaseDataTable root = new BaseDataTable(rootMetaData);
 *     data_tree.setRoot((TreeNode) root);
 *     int row1 = root.addRow();
 *     root.updateCell(row1, column1, value1);
 *     root.updateCell(row1, column2, value2); // etc ...
 * </code></pre>
 * Extensions of this class, for example the JClass DataSource jdbc implementation,
 * "automatically" creates/populates data tables based on data returned from
 * datasource queries.
 * <P>
 * In the case of ide-specific implementaions, they extend this class and override
 * the data storage and retrieval mechanisms.
 * <P>
 * Users wishing to extend this class should look at overriding some or all of the
 * methods defined in the DataTableAbstractionLayer interface.  These are the
 * methods most likely to need overriding.
 * <P>
 * Each instance of this class has its own cursor which can be
 * navigated independently of the DataModel's global cursor.  Only the global
 * cursor (controlled by DataModel.moveToRow) causes commits to occur when the
 * current row is changed.  The navigation methods here do not cause the
 * global cursor to change, only this table's private cursor.
 * @see #first
 * @see #last
 * @see #next
 * @see #previous
 * @see #beforeFirst
 * @see #afterLast
 * @see #isAfterLast
 * @see #isBeforeFirst
 * @see #relative
 * @see #absolute
 * @see #getCurrentBookmark
 * @see com.klg.jclass.datasource.DataModel#moveToRow
 */
public class BaseDataTable extends DataSourceTreeNode implements
    DataTableModel, DataTableAbstractionLayer {

static final long serialVersionUID = 3364765541867247105L;

protected MetaDataModel metaData;               // data control for this  result
protected long parentRow;                       // this detail relates to this parent row
protected StoreModel store;                     // cached rows indexed by bookmark
protected int initialHashtableSize = 5;         // Sun's default is 101, which is too big
protected Hashtable updatedRows;                // updated rows
protected Hashtable insertedRows;               // inserted rows
protected Hashtable rowIndexBeforeEvent;        // record idx before event so event can get idx
protected Hashtable rowCaches[];                  // contain three caches for looping
protected int cursor = DataTableModel.CURSOR_BEFORE_FIRST;                     // current row position in this table,
// default is before first row
protected Hashtable logicallyDeletedRows;       // copy of logically deleted row index
                                                // by bookmark
protected Hashtable userData = null;            // reference to user object
protected Hashtable internalData = null;        // used by higrid
protected Vector eventQueue;                      // for deferred commitRow events
protected static long bookmark = 1;               // bookmark, unique across all DataTables
protected String currentColumn;                   // column being edited


/****************** DataTableModel implementation *****************************/

/**
 * Base constructor.
 */
private BaseDataTable() {
    super();

    // and one to track logically deleted rows
    logicallyDeletedRows = new Hashtable(initialHashtableSize);

    // table of rows that were updated
    updatedRows = new Hashtable(initialHashtableSize);

    // or inserted
    insertedRows = new Hashtable(initialHashtableSize);

    // group the tables so they can be iterated over
    rowCaches = new Hashtable[] { logicallyDeletedRows, updatedRows, insertedRows};

    // for deferred commitRow events, to allow rollback
    eventQueue = new Vector();

    rowIndexBeforeEvent = new Hashtable(initialHashtableSize);
}

/**
 * Root level constructor.
 * Should be called only once to set the data table which will be the
 * root of the data table tree.
 * Subsequent nodes should be created using the constructor below.
 * @param metaData the MetaDataModel which provides the column structure
 * for this data table.
 */
public BaseDataTable(MetaDataModel metaData) {
    this();
    this.metaData = metaData;
    this.parentRow = ROOT_BOOKMARK;
    createStore();
}

/**
 * Constructor to call after root-level BaseDataTable has been created.
 * These tables will be sub-root level nodes in the data table tree.
 * @param metaData the MetaDataModel which provides the column structure
 * for this data table.
 * @param parentRow an int indicating to which parent bookmark this
 * data table relates.
 */
public BaseDataTable(MetaDataModel metaData, long parentRow) {
    this();
    this.metaData = metaData;
    this.parentRow = parentRow;
    createStore();
}

/**
 * Returns the number of rows in this DataTable.
 * @return the number of rows in this table
 */
public int getRowCount() {
    return store.getRowCount();
}

/**
 * Returns the status of a row, and will be one of INSERTED, UPDATED,
 * DELETED, COMMITTED, NONEXISTENT.
 * @param bookmark a unique long indicating for which row
 * status should be retrieved
 * @return an int indicating one of INSERTED, UPDATED, DELETED,COMMITTED,
 * or NONEXISTENT.
 */
public int getRowStatus(long bookmark) {
    if (logicallyDeleted(bookmark)) {
        return DataTableModel.DELETED;
    }
    else if (updatedRows.containsKey(new Long(bookmark))) {
        return DataTableModel.UPDATED;
    }
    else if (insertedRows.containsKey(new Long(bookmark))) {
        return DataTableModel.INSERTED;
    }
    else if (store.getIndex(bookmark) != ROW_NOT_FOUND) {
        return DataTableModel.COMMITTED;
    }
    else {
        return DataTableModel.NONEXISTENT;
    }
}

/**
 * Returns an int which will uniquely identify this row.
 * Facilitates sorting: index changes, identifier remains constant.
 * @param index an int for the original index of this row
 * @return a bookmark a unique long to identify this row from now on
 */
public long getRowIdentifier(int row_index) {
    return store.getBookmark(row_index);
}

/**
 * Returns a value for a given row/column.
 * @param bookmark a unique long representing a row
 * @param col_id a unique String representing a column
 * @return Object which is the value at bookmark/column_identifier
 * @exception DataModelException If get fails.
 */
public Object getResultData(long bookmark,String col_id)
throws DataModelException {

 
    ColumnModel c = metaData.getColumnObject(col_id);
    return c instanceof VirtualColumnModel ? ((VirtualColumnModel) c).getResultData(this, bookmark) : getCell(bookmark, col_id);
}

/**
 * Creates and returns the DataTable which corresponds to the specifed
 * row of this parent for the indicated child MetaData object.
 * Always returns a non-null DataTable. Creates new bookmarks for all
 * returned rows.  Calls getTable and returns the existing table if one
 * exists, otherwise the table is created.
 * <P>
 * To provide data you can either subclass this class and override this
 * method to return a DataTableModel, or you can set the ProviderModel
 * for this table's MetaDataModel. The creation of data is then delegated
 * to this class. If set, this method calls that class.
 * <P>
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @param child the MetaDataModel object which will drive the next query
 * @return the child DataTableModel of results
 * @exception DataModelException If createTable() fails.
 * @see #getTable
 */
public DataTableModel createTable(DataModelListener originator, long bookmark, TreeNode child)
throws DataModelException {

	ProviderModel provider;

    // see first if it already exists, if it does, just return it
    DataTableModel dtm = getTable(bookmark, child);
    if (dtm != null) {
    	// no early exits!
    }
    // does not exist, so create it
	else {
		// see if the creation of data has been delegated to the provider first
		if ((provider = metaData.getProvider()) != null) {
			Object data[][] = provider.getData(originator, bookmark, (MetaDataModel) child);
			dtm = createDataTableModelFromProviderData((MetaDataModel) child, bookmark, data);
		}

		// else get it from subclass
		else if (bookmark != ROW_NOT_FOUND) {
        	dtm = materializeTable(originator, bookmark, (TreeNode) child);
    	}

    	// add it to the tree, if not null
		if (dtm != null) {
    		add((DataSourceTreeNode) dtm);
		}
	}

    return dtm;
}

/**
 * Build a DataTableModel from the matrix returned
 * the the ProviderModel.
 */
public static DataTableModel createDataTableModelFromProviderData(MetaDataModel metaData, long parentRow, Object[][] data) {
	BaseDataTable table;

	// root case
	if (metaData.getParent() == null) {
		table = new BaseDataTable(metaData);
	}
	else {
		table = new BaseDataTable(metaData, parentRow);
	}

	// now populate the table
	for (int row = 0; row < data.length; row++) {
		table.addInternalRow(data[row]);
	}
	return table;
}

/**
 * Searches the existing children of this node for
 * a child node which matches the input criteria.
 * Returns that node if it exists, null otherwise.
 * @param bookmark a long which uniquely identifies a row
 * @param child the MetaDataModel object
 * @return the DataTable with matching bookmark and MetaDataModel
 * otherwise, null.
 * @see #createTable
 */
public DataTableModel getTable(long bookmark, TreeNode node) {
    // does a linear search, but expected to be called infrequently.
    // called when og wants to add a row to a
    // data table which has zero rows.
    MetaDataModel md = (MetaDataModel) node;
    for (Enumeration e = children(); e.hasMoreElements();) {
        DataTableModel dtm = (DataTableModel) e.nextElement();
        if (dtm.getParentBookmark() == bookmark && dtm.getMetaData().equals(md)) {
            return dtm;
        }
    }
    return null;
}

/**
 * Returns the MetaData for this DataTable.
 * @return MetaData associated with this DataTable
 */
public MetaDataModel getMetaData() {
    return(MetaDataModel) metaData;
}

/**
 * Returns the bookmark of the parent row for this table.
 * This is the immediate parent of this table.
 * @return the bookmark of the parent row for this DataTable
 * @see #getRootAncestorBookmark
 * @see #getAncestors
 */
public long getParentBookmark() {
    return parentRow;
}

/**
 * Adds a row whose status will be INSERTED.
 * Adds a blank row into which the user can input column data.
 * The record is appended to this ResultSet.  The record is saved to datasource
 * when the user commits the surrounding transaction.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator) throws DataModelException {
    Object new_row = createNewRow();
    return addRow(originator, new_row);
}

/**
 * Adds a populated row whose status will be 'status'.
 * Adds a pre-populated row whose status is dictated by the
 * 'status' parameter.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @param status an int, one of,
 * <ul>
 * <li>DataTableModel.COMMITTED</li>
 * <li>DataTableModel.DELETED</li>
 * <li>DataTableModel.INSERTED<li>
 * </ul>
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row, int status) throws DataModelException {
    return addRow(originator, row, status, -1);
}

/**
 * Adds a populated row whose status will be 'status'.
 * Adds a pre-populated row whose status is dictated by the
 * 'status' parameter.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @param status an int, one of,
 * <ul>
 * <li>DataTableModel.COMMITTED</li>
 * <li>DataTableModel.DELETED</li>
 * <li>DataTableModel.INSERTED<li>
 * </ul>
 * @param ordinal the index position into which the row should be inserted
 *
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row, int status, int ordinal) throws DataModelException {
    // give the listeners the opportunity to cancel this insert
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEFORE_INSERT_ROW))) {
        return -1;
    }
    Object new_row[] = (Object[]) row;
    // add it to the result table
    long bookmark = (ordinal < 0) ? addInternalRow(new_row) : addInternalRow(new_row, ordinal);

    // add the row to the updatedRows table
    switch (status) {
    case DataTableModel.INSERTED:
        insertedRows.put(new Long(bookmark), new_row);
        break;
    case DataTableModel.DELETED:
        logicallyDeletedRows.put(new Long(bookmark), new_row);
        break;
    case DataTableModel.COMMITTED:
    case DataTableModel.NONEXISTENT:
        // don't insert into any dirty tables
        break;
    default:
        // System.out.println("unknown status");
        break;
    }

    // update the view
    fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_INSERT_ROW));
    return bookmark;
}

/**
 * Adds a populated row whose status will be INSERTED.
 * Adds a pre-populated row.
 * The record is appended to this ResultSet.  The record is saved to datasource
 * when the user commits the surrounding transaction.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row) throws DataModelException {
    return addRow(originator, row, DataTableModel.INSERTED);
}

/**
 * Deletes a row.
 * Depending upon the commit policy the row will either be immediately and
 * permanently deleted from the view and the datasource
 * (commit policy == COMMIT_LEAVING_RECORD), or will simply be logically deleted.
 * <P>
 * In the case of a logical delete, if the <code>showDeletedRows</code> has been
 * set to false, the row will disappear from the display. It will be deleted
 * permanently upon commit, or restored on cancel or requery. The default is
 * for logically deleted rows to be shown.
 * <P>
 * Any pending changes to the row are automatically undone before any deletion.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating the row to be deleted.
 * @exception DataModelException If the logical delete fails.
 * @see com.klg.jclass.datasource.MetaDataModel#getShowDeletedRows
 * @see com.klg.jclass.datasource.MetaDataModel#setShowDeletedRows
 * @see com.klg.jclass.datasource.MetaDataModel#setCommitPolicy
 * @see com.klg.jclass.datasource.MetaDataModel#getCommitPolicy
 */
public void deleteRow(DataModelListener originator, long bookmark) throws DataModelException {
    // give the listeners the opportunity to cancel this delete
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.BEFORE_DELETE_ROW))) {
        return;
    }
    // cancel row changes before deleting
    int rowStatus = getRowStatus(bookmark);
    if (rowStatus != COMMITTED) {
        cancelChanges(originator, bookmark);
        if (rowStatus == INSERTED) {
            // the cancel would remove the inserted row, so return
            return;
        }
    }

    // user didn't cancel the event so change row's status to deleted
    logicallyDeletedRows.put(new Long(bookmark), getRowFromCache(bookmark));

    // commit now or defer?
    if (metaData.getCommitPolicy() == MetaDataModel.COMMIT_LEAVING_RECORD) {
        try {
            commitRow2(originator, bookmark);
            commitTransactionAndProcessEvent(originator, bookmark);
        }
        catch (DataModelException dme) {
            // delete failed so remove the row from deleted table
            removeRowFromLogicallyDeletedTable(bookmark);
            throw dme;
        }
    }
    else {
        // logical delete: should we still display the row?
        if (metaData.getShowDeletedRows()) {
            fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
        }
        else {
            // if not simulate a "real" delete, can be undone
            Long bm = new Long(bookmark);
            rowIndexBeforeEvent.put(bm, new Integer(store.getIndex(bookmark)));
            removeRowFromOriginalRowTable(bookmark);
            fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_DELETE_ROW));
            fireOriginatorDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.ORIGINATOR_NAVIGATE_ROW));
            rowIndexBeforeEvent.remove(bm);
        }
    }
}

/**
 * Updates a cell.
 * User has modified a cell. Call this method to inform this
 * data source of that change.  This method will then fire a
 * DataTableEvent to inform listeners about this change.
 * getRowStatus() will report this row as UPDATED.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating in which row
 * the change was made
 * @param column_identifer a unique String indicating in which
 * column the change occurred.
 * @param value the Object to be placed in the cell
 * @exception DataModelException If a cell update fails.
 */
public void updateCell(DataModelListener originator, long bookmark, String column_id, Object value) throws DataModelException {
    // if there are no changes in new versus old cell value, return
    if (!hasValueChanged(value, getResultData(bookmark, column_id))) {
        return;
    }
    // otherwise see if permissions and so on are okay ...
    if (!prepareCellUpdate(originator, bookmark, column_id)) {
        return;
    }
    // clone the row so it can be restored and mark as updated
    setRowDirty(bookmark);

	setValueAt(bookmark, column_id, value);

	// notify listeners
    fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, column_id, DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
}

/**
 * Commits the row.
 * Writes the row (new and/or modified) to datasource. Calls
 * saveRow(), an implmentation specific method to save a row.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating which row
 * to commit to the datasource.
 * @return true if the commit succeeded, false if failed or was cancelled.
 * @exception DataModelException If commit fails.
 * @see #saveRow
 */
public boolean commitRow(DataModelListener originator, long bookmark) throws DataModelException {

    // Give the listeners the opportunity to cancel this commit.
    // We don't do the check for changes yet since someone could
    // be in the middle of a cell edit which they may commit
    // upon receiving this event.
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.BEFORE_COMMIT_ROW))) {
        return false;
    }

    // everyone has been warned and given the chance to
    // apply their changes, so if still none, return
    if (getRowStatus(bookmark) == DataTableModel.COMMITTED || getRowStatus(bookmark) == DataTableModel.NONEXISTENT) {
        return true;
    }

    commitRow2(originator, bookmark);
    commitTransactionAndProcessEvent(originator, bookmark);
    return true;
}

/**
 * Commits all uncommited rows.
 * Call commitRow for each new and/or modified row in this DataTable.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If commit of any row fails.
 */
public void commitAll(DataModelListener originator) throws DataModelException {
    commitThisLevel(originator);
    // recurse and save
    for (Enumeration e = children(); e.hasMoreElements();) {
        DataTableModel table = (DataTableModel) e.nextElement();
        table.commitAll(originator);
    }
}

/**
 * Recurse on this node and each of its children and commit
 * all changes in each if commitPolicy does not equal COMMIT_MANUALLY
 * commitAll does not check commit policy before committing.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If commit fails.
 */
public void conditionallyCommitYourselfAndChildren(DataModelListener originator) throws DataModelException {
    if (metaData.getCommitPolicy() != MetaDataModel.COMMIT_MANUALLY) {
        commitThisLevel(originator);
    }
    // recurse and save
    for (Enumeration e = children(); e.hasMoreElements();) {
        DataTableModel table = (DataTableModel) e.nextElement();
        table.conditionallyCommitYourselfAndChildren(originator);
    }
}

/**
 * Cancels all uncommited rows.
 * Restore original cell values for each uncommited row, undelete
 * logically deleted rows.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If the changes can't be canceled.
 */
public void cancelAllRowChanges(DataModelListener originator) throws DataModelException {
    // loop on deleted, update, inserted rows
    for (int i = 0; i < rowCaches.length; i++) {
        Hashtable ht = (Hashtable) rowCaches[i].clone();
        for (Enumeration e = ht.keys(); e.hasMoreElements();) {
            cancelRowChanges(originator, ((Long)e.nextElement()).longValue());
        }
    }
    // recurse
    for (Enumeration e = children(); e.hasMoreElements();) {
        DataTableModel table = (DataTableModel) e.nextElement();
        table.cancelAllRowChanges(originator);
    }
}

/**
 * Cancels uncommited row changes.
 * Restore original cell values for this row.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long which identifies this row
 * @exception DataModelException If the changes can't be canceled.
 */
public void cancelRowChanges(DataModelListener originator, long bookmark) throws DataModelException {
    // give the listeners the opportunity to cancel this cancel
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.BEFORE_CANCEL_ROW_CHANGES))) {
        return;
    }
    cancelChanges(originator, bookmark);
}

/**
 * Recurses up the tree and determines the row number of
 * the ultimate ancestor.
 * @return A long indicating the bookmark of this
 * DataTable's ultimate ancestor.
 */
public long getRootAncestorBookmark(long bookmark) {
    DataTableModel parentTable = (DataTableModel) getParent();
    if (parentTable == null) {
        return bookmark;
    }
    else if (((TreeNode) parentTable).getParent() == null) {
        return parentRow;
    }
    else {
        return parentTable.getRootAncestorBookmark(bookmark);
    }
}

/**
 * Returns a list of this DataTable's ancestor bookmarks.
 * The list is ordered from the root ancestor down to the node
 * above this one.
 * @param  ancestors the Vector to be populated with
 *         ancestor bookmarks
 * @return a Vector of ancestor bookmarks.
 */
public void getAncestors(Vector ancestors) {
    DataTableModel parentTable = (DataTableModel) getParent();
    if (parentTable != null && ((TreeNode) parentTable).getParent() != null) {
        parentTable.getAncestors(ancestors);
    }
    // don't add root level to ancestors
    if (parentRow != ROOT_BOOKMARK) {
        ancestors.addElement(new Long(parentRow));
    }
}

/**
 * Requeries a DataTable row and deletes all children from the cache.
 * Calls requeryRow().
 * Refetches all rows in this DataTable from the
 * originating data source and deletes all child nodes from the cache.
 *
 * @param originator the DataModelListener which initiated this action
 * @return this data table model
 * @exception DataModelException If requery from server fails.
 * @see #requeryRow
 */
public void requeryRowAndDetails(DataModelListener originator, long bookmark) throws DataModelException {
    // give the listeners the opportunity to cancel this requery
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.BEFORE_REQUERY))) {
        return;
    }
    requeryRow(originator, bookmark);
    removeAllChildren();
    metaData.getDataModel().clearCurrentPath(metaData);
    fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_REQUERY_ROW_AND_DETAILS));
}

/**
 * Requeries the data for the root level DataTable.
 * An implementation specific method to requery the root level node.
 * @exception DataModelException If requery from server fails.
 */
public DataTableModel requeryLevel() throws DataModelException {
    return this;
}

/**
 * Position this table's cursor on the first row.
 */
public boolean first() {
    if (getRowCount() > 0) {
        cursor = 0;
        return true;
    }
    else {
        return false;
    }
}

/**
 * Positions this table's cursor on the last row.
 */
public boolean last() {
    if (getRowCount() > 0) {
        cursor = store.getRowCount() - 1;
        return true;
    }
    else {
        return false;
    }
}

/**
 * Advances the cursor by one row.
 * Returns true if there was a row to move forward to.
 * Otherwise, remains on last row.
 * @return true if move to next row succeeded.
 */
public boolean next() {
    if (++cursor  < store.getRowCount()) {
        return true;
    }
    cursor--;
    return false;
}

/**
 * Positions this table's cursor to the previous row.
 * Returns true if the move succeeds, false otherwise
 * in which case it will stay on the first row.
 * @return true if successful, false otherwise.
 */
public boolean previous() {
    if (--cursor > DataTableModel.CURSOR_BEFORE_FIRST) {
        return true;
    }
    cursor++;
    return false;
}

/**
 * Position the cursor before the first row.
 */
public void beforeFirst() {
    cursor = DataTableModel.CURSOR_BEFORE_FIRST;
}

/**
 * Position the cursor after the last row.
 */
public void afterLast() {
    cursor = store.getRowCount();
}

/**
 * Return true if the cursor is positioned before first row.
 * @return true if cursor is positioned before first row, false otherwise.
 */
public boolean isBeforeFirst() {
    return cursor == DataTableModel.CURSOR_BEFORE_FIRST;
}

/**
 * Return true if the cursor is positioned afte the last row.
 * @return true if the cursor is positioned after the last row, false otherwise.
 */
public boolean isAfterLast() {
    return cursor >= store.getRowCount();
}

/**
 * Move the cursor forward or backward by "n" rows from the current position.
 * Positive numbers move forward, negative backwards.
 * @param number_of_rows an int indicating the number of rows to move.
 * @exception DataModelException If move fails.
 */
public void relative(int rows) throws DataModelException {
    int new_position = cursor + rows;
    if (new_position < store.getRowCount() + 1 && new_position > -2) {
        cursor = new_position;
    }
    else {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.relative_move_failed) + new_position);
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/**
 * Move the cursor to this ordinal position.
 * Ordinal positions start at 1.
 * @param ordinal_position an int indicating to which row to move
 * @exception DataModelException If the operation fails.
 */
public void absolute(int ordinal_position) throws DataModelException {
    if (ordinal_position > 0 && ordinal_position  <= store.getRowCount()) {
        cursor = ordinal_position - 1;
    }
    else {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.absolute_move_failed) + ordinal_position);
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/**
 * Returns the bookmark of the current row.
 * @return the bookmark of the current row.
 * @exception DataModelException If there are no rows or
 * or an error occurs.
 */
public long getCurrentBookmark() throws DataModelException {
    if (store.getRowCount() < 1) {
        return 0;
    }
    if (cursor >= store.getRowCount() || cursor < 0) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.current_bookmark_error));
        throw new DataModelException(DataModelEvent.NOT_APPLICABLE, "", this, DataModelEvent.INITIALIZE, e);
    }
    return store.getBookmark(cursor);
}

/**
 * Gets user data.
 * Returns the Object which was arbitrarily associated with
 * this table.
 * @param user an Object, a reference to the user.
 * @return the user Object associated with this table.
 */
public Object getUserData(Object user) {
    if (userData == null || user == null) {
        return null;
    }
    return userData.get(user);
}

/**
 * Sets user data.
 * Allows users to augment this table by associating an
 * arbitrary Object reference with it.
 * @param user an Object, a reference to the user.
 * @param data an Object, a reference to an Object.
 */
public void setUserData(Object user, Object data) {
    if (user == null) {
        return;
    }
    // lazy instantiation of hashtable
    if (userData == null) {
        userData = new Hashtable(initialHashtableSize);
    }
    if (data == null) {
        userData.remove(user);
    }
    else {
        userData.put(user, data);
    }
}

/**
 * For internal use only.
 * Gets internal data.
 * Return the Object which was arbitrarily associated with
 * this table.
 * @param user an Object, a reference to the user.
 * @return the internal Object associated with this table.
 */
public Object getInternalData(Object user) {
    if (internalData == null || user == null) {
        return null;
    }
    return internalData.get(user);
}

/**
 * For internal use only.
 * Sets internal data.
 * Allow component writers to augment this table by associating an
 * arbitrary Object reference with it.
 * @param user an Object, a reference to the user.
 * @param data an Object, an arbitrary reference to an Object.
 */
public void setInternalData(Object user, Object data) {
    if (user == null) {
        return;
    }
    // lazy instantiation of hashtable
    if (internalData == null) {
        internalData = new Hashtable(initialHashtableSize);
    }
    if (data == null) {
        internalData.remove(user);
    }
    else {
        internalData.put(user, data);
    }
}

/**
 * Recursive flushing of pending events from a transaction.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If an exception occurs while flushing events.
 */
public void processEventQueue(DataModelListener originator) throws DataModelException {
    for (Enumeration e = eventQueue.elements(); e.hasMoreElements();) {
        long bookmark = ((Long) e.nextElement()).longValue();
        processRowEvent(originator, bookmark);
    }
    eventQueue.removeAllElements();
    // recurse
    for (Enumeration e = children(); e.hasMoreElements();) {
        ((DataTableModel) e.nextElement()).processEventQueue(originator);
    }
}

/**
 * Recursively clear pending events from a transaction.
 * Called when an exception occurs. Events are not dispatched
 * they are simply deleted.
 */
public void clearEventQueue() {
    eventQueue.removeAllElements();
    // recurse
    for (Enumeration e = children(); e.hasMoreElements();) {
        ((DataTableModel) e.nextElement()).clearEventQueue();
    }
}

/**
 * Move the cursor to the row indicated by the bookmark.
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If cursor move fails.
 */
public void moveCursorToRow(long bookmark) throws DataModelException {
    int index = getRowIndex(bookmark);  // store.indexOf(new Integer(bookmark));
    if (index == ROW_NOT_FOUND) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.cursor_move_failed) + bookmark);
        throw new DataModelException(bookmark, "", this, DataModelEvent.INITIALIZE, e);
    }
    cursor = index;
}

/** Gets the cursor of the data table. */
public int getCursor() {
    return cursor;
}

/**
 * Given a bookmark, return the zero-relative row index in this table
 * @param bookmark a long which uniquely identifies a row
 * @return rowIndex a int which is the ordinal position of the row
 * in this table
 */
public int getRowIndex(long bookmark) {
    if (bookmark < 0) {
        return ROW_NOT_FOUND;
    }
    int row_index = store.getIndex(bookmark);
    if (row_index == ROW_NOT_FOUND) {
        Integer i = (Integer) rowIndexBeforeEvent.get(new Long(bookmark));
        if (i != null) {
            row_index = i.intValue();
        }
        else {
            row_index = ROW_NOT_FOUND;
        }
    }
    return row_index;
}

/**
 * Deletes the subtree of DataTables with this DataTable
 * as its root, as well as itself.
 * @param originator the DataModelListener which initiated this action
 * @see com.klg.jclass.datasource.MetaDataModel#getCacheChildren
 * @exception DataModelException If the subtree can't be deleted.
 */
public void deleteCacheChildren(DataModelListener originator) throws DataModelException {
    if (recursivelyDeleteChildren(this, originator)) {
        removeAllChildren();
    }
	BaseDataTable dt = (BaseDataTable) getParent();
	if (dt != null) {
	    // see if any listeners object to deleting this DataTable
	    if (fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEFORE_DELETE_TABLE))) {
	        dt.remove(this);
	    }
	}
}

/**
 * Checks recursively to determine if any listener objects to
 * the deletion of the passed subtree.
 * @param node the DataTable to identify the subtree's root
 * @param originator the DataModelListener which initiated this action
 * @return true if all listeners agreed that all children could be deleted
 * @exception DataModelException If the subtree can't be deleted.
 */
protected boolean recursivelyDeleteChildren(BaseDataTable node, DataModelListener originator) throws DataModelException {
    if (node == null) {
        return true;
    }
    for (Enumeration e = node.children(); e.hasMoreElements();) {
        node = (BaseDataTable)e.nextElement();
        if (node.getChildCount() > 0) {
            if (!recursivelyDeleteChildren(node, originator)) {
                return false;
            }
        }
        else {
            // see if any listeners objects to deleting this DataTable
            if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, node, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEFORE_DELETE_TABLE))) {
                return false;
            }
        }
    }
    return true;
}

/**
 * For internal use only.
 * Move the ide cursor to this row, must be overridden for ide's.
 * <P>
 * There is only one path from the root to a child node in some
 * subtree.  For each meta data level, this method sets the data tables
 * which comprise this path to this table as well as the bookmarks
 * at each level which, when expanded, ultimately lead to this table.
 *
 * @param bookmark a long which uniquely identifies a row.
 * @exception DataModeolException If move fails.
 * @see com.klg.jclass.datasource.MetaDataModel#getCurrentDataTable
 * @see com.klg.jclass.datasource.MetaDataModel#getCurrentBookmark
 */
public void moveToRow(long bookmark) throws DataModelException {
    // the exception is here so we can override it
    // in extensions of this class

    // set those above us to point to this data table
    setAncestorPath();

    // and set it for this level, too
    getMetaData().setCurrentDataTable(this);
    getMetaData().setCurrentBookmark(bookmark);
}

/**
 * Does the subtree rooted at subTreeRootBookmark contain
 * modified nodes?  Modified means this row or at least one of its child
 * nodes has pending insert, update or delete changes.
 * @param subTreeRootBookmark a long, the bookmark of the row which is
 * the root of the subtree to be checked.
 * @return true if this subtree contains at least one modified
 * node, false otherwise.
 */
public boolean isSubTreeModified(long subTreeRootBookmark) {
    // see if this row is modified, then its subtree
    int rowStatus = getRowStatus(subTreeRootBookmark);
    if (rowStatus != COMMITTED && rowStatus != NONEXISTENT) {
        return true;
    }
    // Loop on each of this nodes children and call isSubTreeModified
    // which recursively drills down on each child table's children
    // looking for changes.
    for (Enumeration e = children(); e.hasMoreElements();) {
        DataTableModel child = (DataTableModel) e.nextElement();
        if (child.getParentBookmark() == subTreeRootBookmark) {
            if (child.isTableOrChildrenModified()) {
                return true;
            }
        }
    }
    return false;
}
/**
 * Are there any modifications to this table or any of its child nodes?
 * @return true if there are modifications to this table or any of its
 * child nodes
 */
public boolean isTableOrChildrenModified() {
    boolean modified = false;
    if (isModified()) {
        modified = true;
    }
    else {
        for (Enumeration e = children(); e.hasMoreElements();) {
            DataTableModel child = (DataTableModel) e.nextElement();
            if (child.isTableOrChildrenModified()) {
                modified = true;
                break;
            }
        }
    }
    return modified;
}

/**
 * Commit all pending changes in this subtree including this row.
 *
 * @param originator the DataModelListener which initiated this action
 * @param subTreeRootBookmark a long, the bookmark of the row which is
 * the root of the subtree to be checked.
 * @exception DataModelException If commit fails.
 */
public void commitSubTree(DataModelListener originator, long subTreeRootBookmark) throws DataModelException {
    commitRow(originator, subTreeRootBookmark);
    // recurse and save
    for (Enumeration e = children(); e.hasMoreElements();) {
        DataTableModel table = (DataTableModel) e.nextElement();
        table.commitAll(originator);
    }
}

/**
 * Does this DataTableModel contain any pending inserts, updates or
 * deletes?
 * @return true if there are pending modifications, false otherwise.
 */
public boolean isModified() {
    return insertedRows.size() + updatedRows.size() + logicallyDeletedRows.size() > 0;
}

/**
 * Return a long array of updated row bookmarks.  These are the rows
 * which have updates pending.
 * @return a long[] of updated row bookmarks
 */
public long[] getUpdatedRowBookmarks() {
    return getBookmarksFromTable(updatedRows);
}

/**
 * Return a long array of inserted row bookmarks.  These are the rows
 * which are inserted into the data model, but not yet persisted.
 * @return a long[] of inserted row bookmarks
 */
public long[] getInsertedRowBookmarks() {
    return getBookmarksFromTable(insertedRows);
}

/**
 * Retrun a long array of logically deleted row bookmarks.  These are the
 * rows which are logically deleted, but not yet physically deleted from the
 * originating data source.
 * @return a long[] of logically deleted row bookmarks
 */
public long[] getDeletedRowBookmarks() {
    return getBookmarksFromTable(logicallyDeletedRows);
}

/**
 * Requery all rows in this table and collapse its child nodes. The default
 * behavior is to call cancelAllRowChanges. This method will need to
 * be overridden to provide custom functionality.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If requery fails.
 */
public void requeryTable(DataModelListener originator) throws DataModelException {
    // give the listeners the opportunity to cancel this requery
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(),
                                               originator, this, getParentBookmark(), "",
                                               DataModelEvent.BEFORE_REQUERY))) {
        return;
    }
    // do the requery of this table, for the unbound case, just fire the event
    // overridden in jdbc extension
    cancelAllRowChanges(originator);
    removeAllChildren();
    metaData.getDataModel().clearCurrentPath(metaData);
    cursor = 0;

    // fire after event
    fireDataModelEvent(new DataModelEvent(metaData.getDataModel(),
                                          originator, this, getParentBookmark(), "",
                                          DataModelEvent.AFTER_REQUERY_TABLE));
}

/*************** End of DataTableModel implementation *************************/


/******************* Default DataTableAbstractionLayer implementation ********/

/**
 * For internal use only.
 * Returns a copy of this row.
 * Used to restore the row when changes are cancelled,
 * or used in the where clause to update the row.
 * Called by updateCell
 * @param bookmark a long uniquely identifying this row.
 * @exception DataModelException If clone fails.
 */
public Object cloneRow(long bookmark) throws DataModelException {
    Object dtr[] = (Object[]) getRowFromCache(bookmark);
    return dtr.clone();
}

/**
 * For internal use only.
 * Call DataTableModel.updateCell(..) instead.
 * Changes the value of an existing cell.
 * Called by updateCell.
 * @param row_idx an int indicating the row
 * @param col_idx an int indicating the column
 * @exception DataModelException If setValueAt fails.
 * @see com.klg.jclass.datasource.DataTableModel#updateCell
 */
public void setValueAt(long bookmark, String col_id, Object value) throws DataModelException {
    ((Object[])getRowFromCache(bookmark))[metaData.getColumnIndex(col_id)] = value;
}

/**
 * For internal use only.
 * Call DataTableModel.cancelRowChanges(..) or DataModel.cancelAll() instead.
 * Restores a row's original values.
 * Called by cancelRowChanges(bookmark).
 * A default implementation which expects a row to be an
 * Object[] array.
 * @param bookmark a long which uniquely identifies a row
 * @see com.klg.jclass.datasource.DataTableModel#cancelRowChanges
 * @see com.klg.jclass.datasource.DataModel#cancelAll
 */
public void restoreRow(long bookmark) {
    Object original_row[] = (Object[]) getOriginalRow(bookmark);
    if (original_row == null) {
        return;
    }
    for (int i = 0; i < original_row.length; i++) {
        try {
            // only restore changed columns
            String column = metaData.getColumnIdentifier(i);
            if (columnModified(bookmark, column)) {
                setValueAt(bookmark, column, original_row[i]);
            }
        }
        catch (DataModelException x) {
            // System.out.println(x.toString());
        }
    }
}

/**
 * Creates a new row, called by addRow().
 * An implementation-specific method for creating a new (blank) row, it
 * must be implemented in the derived class of BaseDataTable.
 * @return an Object which is the new row just added.
 * @exception DataModelException If create fails.
 */
public Object createNewRow() throws DataModelException {
    // create a new empty row
	int number_of_columns = metaData.getColumnCount();
    Object row[] = new Object[number_of_columns];

    // loop on columns and add blanks values
    for (int i = 0; i < number_of_columns; i++) {
        row[i] = null;
    }
    return row;
}

/**
 * Re-reads a row from the originating data source.
 * Called by requeryRow().
 * An implementation specific method for refetching a row, it
 * must be implemented in the derived class of BaseDataTable.
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If requery from server fails.
 * @see com.klg.jclass.datasource.DataTableModel#requeryRow
 */
public void refreshRow(long bookmark) throws DataModelException {
    //
}

/**
 * Saves row changes to originating data source.
 * An implementation specific method for writing a row to the data source.
 * Called by commitRow().
 * @param bookmark a unique long indicating which row
 * to commit to the datasource.
 * @exception DataModelException If save fails.
 * @see com.klg.jclass.datasource.DataTableModel#commitRow
 */
public void saveRow(long bookmark) throws DataModelException {
    //
}

/**
 * For internal use only.
 * Call DataTableModel.getResultData(..) instead.
 * Returns a value for a given row/column.
 * @param bookmark a unique long representing a row
 * @param col_id a unique String representing a column
 * @return an Object which is the value at row/column
 * @exception DataModelException If get fails.
 * @see com.klg.jclass.datasource.DataTableModel#getResultData
 */
public Object getCell(long bookmark,String col_id)
throws DataModelException {
    int column_idx = metaData.getColumnIndex(col_id);
    Object dtr[] = (Object[]) getRowFromCache(bookmark);
    if (dtr != null) {
        return dtr[column_idx];
    }
    return null;
}

/**
 * Called by DataTableModel.createTable to create a new table
 * based on the input criteria.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @param childMetaDataModel a MetaDataModel, this tells this table
 * what structure the child table it is creating should have
 * @exception DataModelException If materializeTable fails.
 */
public DataTableModel materializeTable(DataModelListener originator, long bookmark, TreeNode node)
throws DataModelException {
    return new BaseDataTable((MetaDataModel) node, bookmark);
}

/***************** End of DataTableAbstractionLayer implementation ************/

/****************** Helper methods ********************************************/

/**
 * Lets listeners of the data model know this table has changed.
 * @return true if the action was accepted, false if it was aborted.
 */
public boolean fireDataModelEvent(DataModelEvent event) {
    return metaData.getDataModel().fireDataModelEvent(event);
}

/**
 * Fire this event only to the originator.
 * @return true if the action was accepted, false if it was aborted.
 */
public boolean fireOriginatorDataModelEvent(DataModelEvent event) {
    metaData.getDataModel().fireDataModelEvent(event, event.getOriginator());
    return true;
}

/**
 * Returns the next unique row identifier.
 * @return a long to uniquely identify a row
 * across all DataTable instances.
 */
protected long getNextBookmark() {
    return bookmark++;
}

/**
 * Adds an internal row appending it to the end of the
 * in-memory data store.
 *<P>
 * THIS METHOD IS NOT INTENDED FOR REGULAR USE. Use
 * <code>addRow()</code> instead.  This method is provided
 * as a "back-door" into the datasource and is intended
 * to allow fast population of a datasource during initialization.
 * It does not fire BEFORE_INSERT_ROW nor AFTER_INSERT_ROW nor
 * will the row be flagged as INSERTED.
 *
 * @param row an Object which is the original row.
 * @return the unique id for this row as a long.
 */
public long addInternalRow(Object row) {
    long bm = getNextBookmark();
    store.addRow(bm, row);
    return bm;
}

/**
 * Adds an internal row inserting it into <code>ordinal</code>
 * position in the in-memory data store.
 * <P>
 * THIS METHOD IS NOT INTENDED FOR REGULAR USE. Use
 * <code>addRow()</code> instead.  This method is provided
 * as a "back-door" into the datasource and is intended
 * to allow fast population of a datasource during initialization.
 * It does not fire BEFORE_INSERT_ROW nor AFTER_INSERT_ROW nor
 * will the row be flagged as INSERTED.
 *
 * @param row an Object which is the original row.
 * @param ordinal the index position into which the row should be inserted
 * @return the unique id for this row as a long.
 */
public long addInternalRow(Object row, int ordinal) {
    long bm = getNextBookmark();
    store.addRow(bm, row, ordinal);
    return bm;
}


/**
 * Returns a row Object from the store.
 * @param bookmark a long uniquely identifying a row.
 * @return the original row Object from the cache.
 * @exception DataModelException If row not found in cache.
 */
public Object getRowFromCache(long bookmark) throws DataModelException {
    Object o = store.getRow(bookmark);
    if (o == null) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.bookmark_not_found)+bookmark);
        throw new DataModelException(bookmark, "", this, DataModelEvent.INITIALIZE, e);
    }
    return o;
}

/**
 * Returns true if a row is logically but not yet physically deleted.
 * @param bookmark a unique long indicating the row to be deleted.
 * @return true if the row is logically deleted, but not physically
 * deleted.  Pending changes can still be cancelled.
 */
private boolean logicallyDeleted(long bookmark) {
    return(logicallyDeletedRows.size() > 0 && logicallyDeletedRows.containsKey(new Long(bookmark)));
}

/**
 * Removes a row from the logicallyDeletedRows table.
 * Row has been physically deleted or the delete has been
 * cancelled.
 * @param bookmark a long uniquely identifying a row
 */
protected void removeRowFromLogicallyDeletedTable(long bookmark) {
    if (logicallyDeleted(bookmark)) {
        logicallyDeletedRows.remove(new Long(bookmark));
    }
}

/**
 * Removes a row from the updatedRows table.
 * @param bookmark a long uniquely identifying a row
 */
public void removeRowFromUpdatedTable(long bookmark) {
    Long bm = new Long(bookmark);
    if (updatedRows.containsKey(bm)) {
        updatedRows.remove(bm);
    }
}

/**
 * Removes a row from the insertedRows table.
 * @param bookmark a long uniquely identifying a row
 */
public void removeRowFromInsertedTable(long bookmark) {
    Long bm = new Long(bookmark);
    if (insertedRows.containsKey(bm)) {
        insertedRows.remove(bm);
    }
}

/**
 * Removes a row from the originalRow cache table.
 * @param bookmark a long uniquely identifying a row
 */
protected void removeRowFromOriginalRowTable(long bookmark) {
    store.removeRow(bookmark);
}

/**
 * Loop over deleted, updated and inserted row and call
 * commitRow for each.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If commit fails.
 * @see #commitAll
 * @see #conditionallyCommitYourselfAndChildren
 */
private void commitThisLevel(DataModelListener originator) throws DataModelException {
    // iterate over deleted, updated and inserted rows
    for (int i = 0; i < rowCaches.length; i++) {
        Hashtable ht = (Hashtable) rowCaches[i].clone();
        for (Enumeration e = ht.keys(); e.hasMoreElements();) {
            long bookmark = ((Long) e.nextElement()).longValue();
            // if auto commit is false, we are inside a transaction,
            // so queue events, otherwise, fire them immediately
            if (metaData.getAutoCommit()) {
                commitRow(originator, bookmark);
            }
            else {
                commitRow2(originator, bookmark);
            }
        }
    }
}

/**
 * Internal method, do not use.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row.
 * @see #cancelRowChanges
 * @see #deleteRow
 * @see #getRowIdentifier
 * @exception DataModelException If the changes can't be canceled.
 */
public void cancelChanges(DataModelListener originator, long bookmark) throws DataModelException {
// This logic was moved out of cancelRowChanges so it can be
// called without generating its own BEFORE_* event.  Lets the calling
// methods generate their own BEFORE_* events.
    switch (getRowStatus(bookmark)) {
    case DataTableModel.UPDATED:
        restoreRow(bookmark);
        removeRowFromUpdatedTable(bookmark);
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
        break;
    case DataTableModel.DELETED:
        // Since only non-updated rows can be deleted, no need to call restoreRow()
        // if the row was hidden from display, need to put it back on active list
        if (!metaData.getShowDeletedRows()) {
            Object row = logicallyDeletedRows.get(new Long(bookmark));
            long new_bookmark = addInternalRow(row);
            removeRowFromLogicallyDeletedTable(bookmark);
            fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, new_bookmark, "", DataModelEvent.AFTER_INSERT_ROW));
        }
        else {
            // row was not hidden, so just change its status
            removeRowFromLogicallyDeletedTable(bookmark);
            fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
        }
        break;
    case DataTableModel.INSERTED:
        Long bm = new Long(bookmark);
        rowIndexBeforeEvent.put(bm, new Integer(store.getIndex(bookmark)));
        removeInsertedRow(bookmark);
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_DELETE_ROW));
        fireOriginatorDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.ORIGINATOR_NAVIGATE_ROW));
        rowIndexBeforeEvent.remove(bm);
        break;
    default:
        // just refresh the row
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
        break;
    }
}

/**
 * Called by commitRow and deleteRow
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If commit fails.
 */
private void commitTransactionAndProcessEvent(DataModelListener originator, long bookmark) throws DataModelException {
    try {
        metaData.commitTransaction();
    }
    catch (DataModelException e) {
        int action = (getRowStatus(bookmark) == DataModelEvent.AFTER_DELETE_ROW) ? DataModelEvent.AFTER_DELETE_ROW : DataModelEvent.AFTER_CHANGE_OF_ROW_DATA;
        throw new DataModelException(bookmark, "", this, action,e);
    }
    processRowEvent(originator, bookmark);
}

/* This is called by commitAll, conditionallyCommit.  That is it is
 * always called as part of a multi-record update, whereas commitRow
 * can be called directly in the api.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If commit fails.
 */
private void commitRow2(DataModelListener originator, long bookmark) throws DataModelException {
    // if no changes, return
    if (getRowStatus(bookmark) == DataTableModel.COMMITTED) {
        return;
    }
    try {
        // call the IDE specific method to save a row
        saveRow(bookmark);
    }
    catch (DataModelException e) {
        int action = getRowStatus(bookmark) == DataTableModel.DELETED ? DataModelEvent.AFTER_DELETE_ROW : DataModelEvent.AFTER_CHANGE_OF_ROW_DATA;
        throw new DataModelException(bookmark, "", this, action, e);
    }
    // if auto commit is false, we are inside a transaction,
    // so queue events, otherwise, fire them immediately
    if (!metaData.getAutoCommit()) {
        eventQueue.addElement(new Long(bookmark));
    }
    else {
        processRowEvent(originator, bookmark);
    }
}

/**
 * Process a "commit row event" by firing the appropriate
 * event and cleaning up intermediate caches.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long uniquely identifying a row.
 * @exception DataModelException If an exception occurs while removing rows.
 */
private void processRowEvent(DataModelListener originator, long bookmark) throws DataModelException {
    // clean up and fire events
    switch (getRowStatus(bookmark)) {
    case DataTableModel.UPDATED:
        removeRowFromUpdatedTable(bookmark);
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
        break;
    case DataTableModel.INSERTED:
        removeRowFromInsertedTable(bookmark);
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
        break;
    case DataTableModel.DELETED:
        Long bm = new Long(bookmark);
        rowIndexBeforeEvent.put(bm, new Integer(store.getIndex(bookmark)));
        removeRowFromLogicallyDeletedTable(bookmark);
        removeRowFromOriginalRowTable(bookmark);
        // only need to fire this event if logically deleted rows are not
        // hidden.  If they are then this is redundant.
        if (metaData.getShowDeletedRows()) {
            fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_DELETE_ROW));
            fireOriginatorDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.ORIGINATOR_NAVIGATE_ROW));
        }
        rowIndexBeforeEvent.remove(bm);
        break;
    default:
        break;
    }
}

/**
 * Re-reads a row from the originating data source.
 * Cancels pending modificatoins to the row, if any.
 * Calls refreshRow(bookmark), which is the implementation specific
 * method to refresh a row.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If requery from server fails.
 */
public void requeryRow(DataModelListener originator, long bookmark) throws DataModelException {
    // give the listeners the opportunity to cancel this requery
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.BEFORE_REQUERY))) {
        return;
    }
    int status = getRowStatus(bookmark);
    Long bm = new Long(bookmark);
    rowIndexBeforeEvent.put(bm, new Integer(store.getIndex(bookmark)));
    getRow(originator, bookmark);
    if (status == DataTableModel.INSERTED) {
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_DELETE_ROW));
        fireOriginatorDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.ORIGINATOR_NAVIGATE_ROW));
    }
    else {
        fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, "", DataModelEvent.AFTER_CHANGE_OF_ROW_DATA));
    }
    rowIndexBeforeEvent.remove(bm);
}

/**
 * Logic moved out from requeryRow so it can be
 * called by requeryRowAndDetails, too, without
 * firing the AFTER_CHANGE_OF_ROW_DATA event.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If requery from server fails.
 */
private void getRow(DataModelListener originator, long bookmark) throws DataModelException {
    // requerying an inserted (not committed) means
    // eliminating it
    if (insertedRows.containsKey(new Long(bookmark))) {
        removeInsertedRow(bookmark);
        return;
    }
    // restore the row's orginal values before requerying
    // only need to do UPDATED since INSERTED is handled above and
    // DELETED rows cannot have pending edits
    if (getRowStatus(bookmark) == DataTableModel.UPDATED) {
        cancelChanges(originator, bookmark);
    }
    refreshRow(bookmark);
    removeRowFromUpdatedTable(bookmark);
    removeRowFromLogicallyDeletedTable(bookmark);
}

/**
 * Removes a -- non committed -- inserted row.
 * @param bookmark a unique long which identifies this row
 */
private void removeInsertedRow(long bookmark) {
    removeRowFromInsertedTable(bookmark);
    removeRowFromOriginalRowTable(bookmark);
}

/**
 * Returns true if the original cell value is null.
 */
public boolean originalCellWasNull(int idx, Vector columns, Object[] original_row) {
    int col_idx = ((Integer) columns.elementAt(idx)).intValue();
    Object value = original_row[col_idx];
    return(value == null ? true : false);
}

/**
 * Returns true if a column value has been modified.
 * @param bookmark a long uniquely identifying a row
 * @param col_id a unique String representing a column
 * @return true if this column has been modified.
 * @exception DataModelException If the operation fails.
 */
public boolean columnModified(long bookmark, String col_id) throws DataModelException {
    // int row_idx = getRowIndex(bookmark);
    int col_idx = metaData.getColumnIndex(col_id);
    Object original_row[] = (Object[]) getOriginalRow(bookmark);
    if (original_row == null) {
        return false;
    }
    Object newValue = getResultData(bookmark,col_id);
    Object oldValue = original_row[col_idx];

    return hasValueChanged(newValue, oldValue);
}

/**
 * Return true if the newValue is different than the oldValue
 */
public boolean hasValueChanged(Object newValue, Object oldValue) {
    // check for nulls and whether values have changed
    if ((newValue == null && oldValue != null)
        || (oldValue == null && newValue != null)
        || ((oldValue != null && newValue != null) && (!newValue.equals(oldValue)))) {
        return true;
    }
    // otherwise both null, or both same
    return false;
}

/**
 * Given a bookmark, returns the original row as it was
 * before any changes where made.
 * @param bookmark a unique long indicating which row to uncommit
 * @return the original (unmodified) row
 */
public Object getOriginalRow(long bookmark) {
    return(Object[]) updatedRows.get(new Long(bookmark));
}

/**
 * Return the logically deleted row.
 * @param bookmark a long which uniquely identifies a row
 * @return row an Object which is the logically deleted row.
 */
public Object getDeletedRow(long bookmark) {
    return logicallyDeletedRows.get(new Long(bookmark));
}

/**
 * Given a DataTable, returns whether it is an ancestor of
 * (or equal to) itself
 * @param table a DataTable which may be an ancestor
 * @return true if the table is an ancestor
 */
public boolean tableIsAncestor(DataTableModel table) {
    TreeNode parent = (TreeNode)table;
    while (parent != null) {
        if (parent == this) {
            return true;
        }
        parent = (TreeNode)parent.getParent();
    }
    return false;
}

/**
 * Adjust all current data table and bookmark pointers above such that they
 * form a path to this data table.
 * Called by moveToRow
 * @see #moveToRow
 */
private void setAncestorPath() {
    DataTableModel parent_data_table = (DataTableModel) getParent();
    // stop condition
    if (parent_data_table == null) {
        return;
    }
    else {
        ((BaseDataTable)parent_data_table).setAncestorPath(); // keep going
    }
    // otherwise we've reached the top, now work our way down setting pointers
    // properly as we go!!!
    MetaDataModel parent_meta_data = parent_data_table.getMetaData();
    parent_meta_data.setCurrentDataTable(parent_data_table);
    parent_meta_data.setCurrentBookmark(getParentBookmark());
}

/**
 * If column is not in logically deleted row, and is not a virtual
 * column and permissions on its table don't disallow edits and no
 * listeners cancel this event, return true, false otherwise.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If prepare fails.
 */
public boolean prepareCellUpdate(DataModelListener originator, long bookmark, String column_id)
throws DataModelException {
    // cannot edit deleted row
    if (logicallyDeleted(bookmark)) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.cannot_modify_deleted));
        throw new DataModelException(bookmark, column_id, this, DataModelEvent.BEFORE_EDIT_CELL, e);
    }

    // see if this is a virtual column
    ColumnModel cm = metaData.getColumnObject(column_id);
    if (cm == null || cm instanceof VirtualColumnModel) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.cannot_update_virtual_column));
        throw new DataModelException(bookmark, column_id, this, DataModelEvent.BEFORE_EDIT_CELL, e);
    }

    // see if permissions have been explicitly turned off on the table to which this
    // column belongs
    if (!metaData.isUpdateAllowedOnColumn(column_id) || cm.isReadOnly()) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.ds_update_not_allowed) + ": " + column_id);
        throw new DataModelException(bookmark, column_id, this, DataModelEvent.BEFORE_EDIT_CELL, e);
    }

    // give the listeners the opportunity to cancel this cell update
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(), originator, this, bookmark, column_id, DataModelEvent.BEFORE_EDIT_CELL))) {
        return false;
    }

    return true;
}

private long[] getBookmarksFromTable(Hashtable t) {
    long rows[] = new long[t.size()];
    int cnt = 0;
    for (Enumeration e = t.keys(); e.hasMoreElements();) {
        rows[cnt++] = ((Long) e.nextElement()).longValue();
    }
    return rows;
}

private void createStore() {
    // create the storage and retrieval object
    try {
        store = (StoreModel) metaData.getStoreClass().newInstance();
    }
    catch (IllegalAccessException e) {
        e.printStackTrace();
    }
    catch (InstantiationException e) {
        e.printStackTrace();
    }
    catch (DataModelException e) {
        e.printStackTrace();
    }


}

/**
 * Clone the row and add it to the table of updated
 * rows so its status will now be UPDATED, but only
 * if the row is not already in the INSERTED table.
 * @param bookmark a long which uniquely identifies a row
 */
public void setRowDirty(long bookmark) {
    if (!insertedRows.containsKey(new Long(bookmark)) &&
        !updatedRows.containsKey(new Long(bookmark))) {
        try {
            updatedRows.put(new Long(bookmark), cloneRow(bookmark));
        }
        catch (DataModelException e) {
            System.out.println(e.toString());
        }
    }
}

/*************************** End of helper methods ****************************/

}
