/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseMetaData.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:08 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.io.InputStream;
import java.io.IOException;
import java.sql.Types;
import java.util.*;

import javax.swing.tree.*;

/**
 * Base implementation of MetaDataModel.
 */
public class BaseMetaData extends DataSourceTreeNode implements MetaDataModel {

static final long serialVersionUID = -5410672693378221172L;

protected DataModel dataModel;               // container of this
protected String description = "";           // an instance's name
protected Hashtable insertPermissions;       // associative array of table permissions
protected Hashtable updatePermissions;       // ditto
protected Hashtable deletePermissions;       // ditto
protected Hashtable idxToColumnMap;          // mapping between columns and indexes
protected Vector columns = new Vector();     // list of ColumnModel objects
protected int commitPolicy = COMMIT_LEAVING_RECORD;  // when should record be committed?
protected Class storeClass = null;           // class for storage and retrieval

protected int initialHashtableSize = 5; // Sun's default is 101, too big for here

protected int metaDataID;            // unique identifier of an instance
protected static int counter = 0;    // static counter for id's
protected int maxRows = 0;           // max rows a query can return
protected boolean cacheChildren = false;  // should children be cached?
protected boolean showDeletedRows = true;// show logically deleted rows?
protected transient DataTableModel currentDataTable;    // the data table currently in focus at this level
protected long currentBookmark = DataTableModel.ROW_NOT_FOUND;          // the bookmark currently in focus
protected int bufferSize = 5000;              // Buffer size in bytes. The default is 5000 bytes.
protected BindingModel binding;
protected ProviderModel	provider;

/**
 * Constructor
 */
public BaseMetaData(DataModel dataModel) {
    super();
    this.dataModel = dataModel;
    this.metaDataID = counter++;
    insertPermissions = new Hashtable(initialHashtableSize);
    updatePermissions = new Hashtable(initialHashtableSize);
    deletePermissions = new Hashtable(initialHashtableSize);
    idxToColumnMap    = new Hashtable(initialHashtableSize);
}

/**
 * Sets a boolean indicating whether this table allows
 * inserts.
 * By default, inserts are allowed.
 * @param table a String name of the table
 * @param allowInsert a boolean indicating if inserts are allowed
 */
public void setInsertAllowed(String table, boolean v) {
    if (insertPermissions.containsKey(table)) {
        insertPermissions.remove(table);
    }
    insertPermissions.put(table, new Boolean(v));
}

/**
 * Sets a boolean indicating whether updates are allowed
 * in the table mentioned in the first parameter.
 * The default is to allow updates.
 * @param table a String name of the table
 * @param allowUpdate a boolean indicating if updates are allowed
 */
public void setUpdateAllowed(String table, boolean v) {
    if (updatePermissions.containsKey(table)) {
        updatePermissions.remove(table);
    }
    updatePermissions.put(table, new Boolean(v));
}

/**
 * Sets a boolean indicating whether rows in this table may
 * be deleted.
 * The default is to permit the table to be deleted.
 * @param table a String name of the table
 * @param allowDelete a boolean indicating if deletes are allowed
 */
public void setDeleteAllowed(String table, boolean v) {
    if (deletePermissions.containsKey(table)) {
        deletePermissions.remove(table);
    }
    deletePermissions.put(table, new Boolean(v));
}

/**
 * Returns true if this table can be inserted into.
 * The default is to allow inserts.
 * @param table a String name of the table
 * @return true if table can be updated
 */
public boolean getInsertAllowed(String table) {
    if (table == null) {
        return true;
    }
    return insertPermissions.get(table) != null ? ((Boolean) insertPermissions.get(table)).booleanValue() : true;
}

/**
 * Returns true if this table can be updated.
 * The default is to allow this.
 * @param table a String name of the table
 * @return true if table can be updated
 */
public boolean getUpdateAllowed(String table) {
    if (table == null) {
        return true;
    }
    return updatePermissions.get(table) != null ? ((Boolean) updatePermissions.get(table)).booleanValue() : true;
}

/**
 * Returns true if this table can be deleted.
 * The default is to allow this.
 * @param table a String name of the table
 * @return true if rows in this table can be deleted.
 */
public boolean getDeleteAllowed(String table) {
    if (table == null) {
        return true;
    }
    return deletePermissions.get(table) != null ? ((Boolean) deletePermissions.get(table)).booleanValue() : true;
}

/**
 * Returns the <pre>DataModel</pre> which encapsulates this <pre>MetaDataModel</pre>.
 * @return the DataModel
 */
public DataModel getDataModel() {
    return dataModel;
}

/**
 * Returns an int that uniquely identifies this instance
 * of the <pre>MetaDataModel</pre>. This id is automatically generated.
 * @return an int which uniquely identifies this object.
 */
public int getMetaID() {
    return metaDataID;
}

/**
 * For internal use only.
 * Assigns a unique id to a meta data instance
 * @param id an int which is the unique id for this instance
 * @see #getMetaID
 */
public void setMetaID(int metaDataID) {
    // this is expected to only be called by the customizer
    // and when materializing from a serialized file
    this.metaDataID = metaDataID;
    // ensure counter is always larger
    counter = metaDataID + 1;
}

/**
 * Returns the user-assigned description of this instance, or ""
 * if the description has not been set.
 * @return a String which describes this instance.
 */
public String getDescription() {
    return description;
}

/**
 * Sets the user-assigned description of this instance.
 * @param description a String which describles this instance.
 */
public void setDescription(String description) {
    this.description = description;
}

/**
 * Default implementation: returns true.
 * Must be overridden in specific implementations if transactions
 * are supported and enabled.
 * Returns true if the transaction capabilities of the underlying
 * data source are to be used when saving modifications
 * (transactions are supported and enabled).
 * If the underlying data source supports transactions and this
 * capability has been enabled, data modifications will be
 * wrapped in a transaction which can be committed or rolled back.
 * @exception DataModelException If the operation fails.
 * @see #commitTransaction
 * @see #rollbackTransaction
 */
public boolean getAutoCommit() throws DataModelException {
    return true;
}

/**
 * Default implementation: does nothing.
 * Can be overriden in specific implementations, but is not
 * for jdbc/jbuilder since this value is derived from the
 * underlying objects, the <pre>DataTableConnection</pre> object in the case of
 * jdbc.
 * Sets the auto commit flag.
 * If the underlying data source supports transactions and this
 * flag is set to false, a transaction will wrap data modifications.
 * The default is to not attempt to wrap modifications in a transaction.
 * @param bool a boolean, true if transactions are to be used,
 * false otherwise.
 * @exception DataModelException If the operation fails.
 */
public void setAutoCommit(boolean value) throws DataModelException {
    // no-op
}

/**
 * Rolls back the transaction: does nothing here, so it.
 * must be overridden.
 * If the underlying data source supports transactions and
 * transactions have been enabled, calling this method will
 * roll back all pending transactions on the data source.
 * @exception DataModelException If rollback fails.
 * @see MetaDataModel#getAutoCommit
 */
public void rollbackTransaction() throws DataModelException {
    // no-op
}

/**
 * Commits transaction: does nothing here, so it
 * must be overridden.
 * If the underlying data source supports transactions and
 * transactions have been enabled, calling this method will
 * commit all pending transactions on the data source.
 * Otherwise it is a no-op.
 * @exception DataModelException If commit fails.
 */
public void commitTransaction() throws DataModelException {
    // no-op
}

/**
 * Sets the commit policy for this level.  Determines when
 * changes are permanently written to the originating data source.
 * Must be one of COMMIT_LEAVING_RECORD, COMMIT_LEAVING_PARENT,
 * or COMMIT_MANUALLY.
 * <P>
 * Commit policy cannot be set to COMMIT_LEAVING_RECORD if getShowDeltedRows()
 * returns false.
 * @param commit_policy an int, one of COMMIT_LEAVING_RECORD,
 * COMMIT_LEAVING_PARENT, or COMMIT_MANUALLY
 * @see #getShowDeletedRows
 * @exception DataModelException If set fails.
 */
public void setCommitPolicy(int commit_policy) throws DataModelException {
    if (getShowDeletedRows() == false && commit_policy == MetaDataModel.COMMIT_LEAVING_RECORD) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.policy_mismatch));
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
    commitPolicy = commit_policy;
}

/**
 * Returns the commit policy for this level.  Determines when
 * changes are permanently written to the originating data source.
 * Will be one of COMMIT_LEAVING_RECORD, COMMIT_LEAVING_PARENT,
 * or COMMIT_MANUALLY.
 * @return an int, one of COMMIT_LEAVING_RECORD,
 * COMMIT_LEAVING_PARENT, or COMMIT_MANUALLY
 */
public int getCommitPolicy() {
    return commitPolicy;
}

/**
 * Returns the list of ColumnModel objects.
 * @return a vector of ColumnModel objects.
 * @see com.klg.jclass.datasource.ColumnModel
 */
public Vector getColumnObjects() {
    return columns;
}

/**
 * Returns number of columns in this result set.
 * @return an int for the number of columns in this table
 */
public int getColumnCount() {
    return columns.size();
}

/**
 * Adds a column to the list of column objects.
 * @param columnObject ColumnModel to be added.
 * @exception DataModelException If duplicate column found.
 * @see com.klg.jclass.datasource.ColumnModel
 */
public void addColumn(ColumnModel column) throws DataModelException {
    Enumeration e = columns.elements();
    while (e.hasMoreElements()) {
        ColumnModel cm = (ColumnModel) e.nextElement();
        if (cm.getIdentifier() != null && cm.getIdentifier().equals(column.getIdentifier())) {
            String s = LocaleBundle.string(LocaleBundle.duplicate_column_found)+ cm.getIdentifier();
            throw new DataModelException(DataModelEvent.INITIALIZE, new Exception(s));
        }
    }
    columns.addElement(column);
    getColumnMap().put(column.getIdentifier(), new Integer(columns.size()-1));
}

/**
 * Returns a string that uniquely identifies this column.
 * Used to access data rather than column index which can
 * change when columns are sorted.
 * @param column_index an int indicating for which column
 * a unique identifier should be returned.
 * @return a String which uniquely identifies a column
 * @exception DataModelException If getColumnIdentifier fails.
 */
public String getColumnIdentifier(int column_index) throws DataModelException {
    return((ColumnModel) columns.elementAt(column_index)).getIdentifier();
}

/**
 * Returns the Java type for a column, one of MetaDataModel
 * type constants.
 * @param column_identifier an String uniquely identifying a column
 * @return an int representing the meta column type for this column, will be one of
 * <ul>
 *     <li>TYPE_BOOLEAN</li>
 *     <li>TYPE_SQL_DATE</li>
 *     <li>TYPE_DOUBLE</li>
 *     <li>TYPE_FLOAT</li>
 *     <li>TYPE_INTEGER</li>
 *     <li>TYPE_STRING</li>
 *     <li>TYPE_BIG_DECIMAL</li>
 *     <li>TYPE_LONG</li>
 *     <li>TYPE_SQL_TIME </li>
 *     <li>TYPE_SQL_TIMESTAMP</li>
 *     <li>TYPE_OBJECT</li>
 *     <li>TYPE_BYTE</li>
 *     <li>TYPE_SHORT</li>
 *     <li>TYPE_BYTE_ARRAY</li>
 *     <li>TYPE_UTIL_DATE</li>
 * </ul>
 * @see MetaDataModel
 * @exception DataModelException If getJavaColumnType fails.
 */
public int getJavaColumnType(String column_identifier) throws DataModelException {
    return getColumnObject(column_identifier).getMetaColumnType();
}


/**
 * Returns the <pre>ColumnModel</pre> for the given column.
 * @param column_identifier a String which uniquely identifies a column
 * @return a ColumnModel
 * @exception DataModelException If get fails.
 */
public ColumnModel getColumnObject(String col_id) throws DataModelException {
    return(ColumnModel) getColumnObjects().elementAt(getColumnIndex(col_id));
}

/**
 * Given a column name, gets its index.
 * @param column_identifier a String indicating the column's identifier
 * @return an int which is this column's ordinal position
 * in the original result data.
 * @exception DataModelException If the operation fails.
 */
public int getColumnIndex(String name) throws DataModelException {
    if (!getColumnMap().containsKey(name)) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.column_not_found)+ name);
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
    return((Integer) idxToColumnMap.get(name)).intValue();
}

/**
 * Returns the list which associates an ordinal position with a column name.
 */
public Hashtable getColumnMap() {
    return idxToColumnMap;
}

/**
 * Returns the native meta-data object: must be overriden.
 * Returns null in the base implementation.
 * This is a DataSet in JBuilder and a RelationView
 * in VisualCafe.
 * @return the native meta data object for this level.
 */
public Object getNativeMetaData() {
    return null;
}

/**
 * Do permissions on the model allow this column to be editied?
 * Returns true in the base implementation.
 * @param column_identifier a String which uniquely identifies this column
 * @return true if edits are allowed, false otherwise
 * @see #setUpdateAllowed
 * @see #getUpdateAllowed
 * @exception DataModelException If primary keys cannot be determined.
 */
public boolean isUpdateAllowedOnColumn(String column) throws DataModelException {
	// get the table name ...
	ColumnModel cm = getColumnObject(column);
	if (cm == null || cm.getTableName() == null || cm.getTableName().equals("")) {
	    return true;
	}
	// then see if permissions were turned off
	return getUpdateAllowed(cm.getTableName());
}

/**
 * For internal use only.
 * Changes the query for this data control.
 * For customizer use during design-time.
 */
public void clear() {
    idxToColumnMap.clear();         // mapping between columns and indexes
    columns.removeAllElements();    // list of ColumnModel objects
    setCurrentBookmark(DataTableModel.ROW_NOT_FOUND);   // needs to be reset
}

/**
 * Indicates whether logically deleted rows showld be shown.  This attribute
 * can only be set if the commitPolicy is COMMIT_MANUALLY or
 * COMMIT_LEAVING_ANCESTOR, otherwise an exception is thrown.
 * @param value a boolean: true to show rows, false otherwise
 * @exception DataModelException If set fails.
 */
public void setShowDeletedRows(boolean value) throws DataModelException {
    if (value == false && getCommitPolicy() == MetaDataModel.COMMIT_LEAVING_RECORD) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.policy_mismatch));
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
    showDeletedRows = value;
}

/**
 * Returns whether logically deleted rows should be shown.
 * @return show a boolean: true means show rows, false hide them.
 * @see #setShowDeletedRows
 * @see #setCommitPolicy
 */
public boolean getShowDeletedRows() {
    return showDeletedRows;
}

/**
 * Sets the maximum number of rows a query at this level can return.
 * Useful for design time where only a sampling of rows is required.
 * Set to 0 (the default) to have no limit.
 * <P>
 * Does not apply to ide-specific implmentations.  Set the max rows
 * directly on the ide data object instead.
 * <P>
 * @param maxRows an int indicating the max rows to be returned
 * @see #getMaxRows
 */
public void setMaxRows(int maxRows) {
    this.maxRows = maxRows;
}


/**
 * Returns the maximum number of rows which can be returned by a query.
 * <P>
 * Does not apply to ide-specific implmentations.  Set the max rows
 * directly on the ide data object instead.
 * <P>
 * @return the maximum number of rows which can be returned by a query
 * @see #setMaxRows
 */
public int getMaxRows() {
    return maxRows;
}

/**
 * Sets a boolean indicating whether subtrees of <pre>DataTables</pre> rooted at this
 * level are eligible for deletion.  The default is false, meaning
 * <pre>DataTableModel.deleteCacheChildren</pre> should be called when a subtree is
 * collapsed.
 * @param cacheChildren a boolean: true means children should be cached, false
 * means dateCacheChildren should be called when a tree is collapsed.
 * @see #getCacheChildren
 * @see DataTableModel#deleteCacheChildren
 */
public void setCacheChildren(boolean cacheChildren) {
    this.cacheChildren = cacheChildren;
}

/**
 * Returns a boolean flag indicating whether subtrees of DataTables rooted at this
 * level are eligible for deletion.
 * @return false if deleteCacheChildren should be called on this level when
 * it is collapsed. Return true if the children should be preserved, that is
 * no call should be made to deleteCacheChildren.
 * @see #setCacheChildren
 * @see DataTableModel#deleteCacheChildren
 */
public boolean getCacheChildren() {
    return cacheChildren;
}

/**
 * Returns the data table which currently has focus.
 * @return a DataTableModel, the one currently in focus.
 */
public DataTableModel getCurrentDataTable() {
    return currentDataTable;
}

/**
 * Returns the bookmark of the row currently in focus.
 * @return the bookmark of the row currently in focus.
 */
public long getCurrentBookmark() {
    try {
        if (currentBookmark == DataTableModel.ROW_NOT_FOUND) {
            DataTableModel dtm = getCurrentDataTable();
            if (dtm != null && dtm.getRowCount() > 0) {
                dtm.first();
                setCurrentBookmark(dtm.getCurrentBookmark());
            }
        }
    }
    catch (DataModelException e) {
   		e.printStackTrace();
    }
    return currentBookmark;
}

public void setCurrentDataTable(DataTableModel currentDataTable) {
    this.currentDataTable = currentDataTable;
}

public void setCurrentBookmark(long currentBookmark) {
    // System.out.println("setting current bookmark to " + currentBookmark);
    this.currentBookmark = currentBookmark;
}

/**
 * Returns a byte array created from an input stream.
 * @param inputStream the InputStream supplying the data
 * @return the byte array created from that data
 * @exception DataModelException If the operation fails.
 */
public byte[] createByteArrayFromBinaryStream(InputStream is) throws DataModelException {
    if (is == null) {
        return null;
    }
    /*
    * Since ResultSet.getObject() and ResultSet.getBytes()
    * don't seem to be able to handle blobs bigger than
    * 5 K an InputStream must be used.  But there is a problem
    * because InputStream.available() throws an exception when the
    * blob is > 5K.  Thus I have to incrementally grow the array
    * because I do not know the size of the data being returned.
    */
    byte buffer[]   = new byte[getBufferSize()]; // temp buffer
    byte blob[]     = new byte[getBufferSize()]; // start off at chunksize
    int totalRead = 0, numRead = 0;

    try {
	    if (is.markSupported()) {
	        // in jbuilder, their components seem to consume the stream
	        // and leave the pointer at the end, so need to reset
	        	is.reset();
	    }
	    while ((numRead = is.read(buffer)) > 0) {
	        byte oldData[] = blob;
	        blob = new byte[totalRead + numRead];
	        System.arraycopy(oldData, 0, blob, 0, totalRead);
	        System.arraycopy(buffer, 0, blob, totalRead, numRead);
	        totalRead += numRead;
	    }
  	}
   	catch (IOException ioe) {
       	throw new DataModelException(DataModelEvent.INITIALIZE, ioe);
   	}

    return blob;
}

/**
 * Binary data (BINARY, VARBINARY, LONGVARBINARY) will be read in
 * chunks of this size.  The default is 5K.
 * @param the chunk size in bytes.
 */
public void setBufferSize(int bufferSize) {
    this.bufferSize = bufferSize;
}

/**
 * Returns the buffer size used for reading binary data, in bytes.
 * @return the block size used to read binary data, in bytes.
 * @see #setBufferSize
 */
public int getBufferSize() {
    return bufferSize;
}

/**
 * Set an alternate class to handle data storage and retrieval
 * responsibilities.  A "Class.forName()" will be done to
 * load this alternate class.
 * <P>
 * If set, BaseDataTable instances will use this instances of
 * this class rather than the default class ("Store") to
 * store and retrieve data.
 * <P>
 * @param className a String, the name of the class which implements
 * the StoreModel interface
 * @exception DataModelException If an error occurs loading the class.
 * @see StoreModel
 * @see #getStoreClass
 */
public void setStoreClass(String storeClassName) throws DataModelException  {
    try {
        this.storeClass = Class.forName(storeClassName);
    }
    catch (ClassNotFoundException e) {
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/**
 * Return the class responsible for data storage and
 * retrieval (implements StoreModel).
 * @return the class used to store/retrieve data
 * @exception DataModelException If storeClass is not set
 * and an error occurs while trying to load the default class,
 * com.klg.jclass.datasource.Store.
 */
public Class getStoreClass() throws DataModelException {
    if (storeClass == null) {
        setStoreClass("com.klg.jclass.datasource.Store");
    }
    return storeClass;
}

/**
 * Set the Binding object associated with this MetaData level.
 * @param the Binding object for this level.
 */
public void setBinding(BindingModel binding) {
	this.binding = binding;
}

/**
 * Return the BindingModel for this level, creating
 * it if it doesn't already exist.
 * @return the BindingModel for this level.
 */
public BindingModel getBinding() {
	if (binding == null) {
		binding = new Binding(dataModel);
		binding.setMetaDataModel(this);
	}
	return binding;
}

/**
 * Get the provider (if any) for this level.
 * <P>
 * @return ProviderModel the class which will
 * provide matrix data for this level
 * <P>
 * @see ProviderModel#getData
 * @see DataTableModel#createTable
 * @see DataTableModel#getTable
 */
public ProviderModel getProvider() {
	return provider;
}

/**
 * Set the ProviderModel class name which will provide
 * data for this level. This class will be instantiated
 * and to provide data. If this is the root level this
 * provider's getData method will actually be called to
 * populate the data root.
 * <P>
 * @param the name of the provider class
 * @exception DataModelException if the provider fails
 * to be instantiated.
 */
public void setProviderClassName(String providerModelClassName) throws DataModelException {
	if (providerModelClassName != null && !providerModelClassName.equals("")) {
		Exception e = null;
		try {
			provider = (ProviderModel) Class.forName(providerModelClassName).newInstance();
		}
		catch (ClassNotFoundException cnfe) {
			e = cnfe;
		}
		catch (IllegalAccessException iae) {
			e = iae;
		}
		catch (InstantiationException ie) {
			e = ie;
		}

		// wrap and throw error if there was one
		if (e != null) {
			throw new DataModelException(DataModelEvent.INITIALIZE, e);
		}

		// if this is the root, populate the data root
		if (getParent() == null) {
			Object[][] data = provider.getData(null, DataTableModel.ROOT_BOOKMARK, this);
			DataTableModel root =
				BaseDataTable.createDataTableModelFromProviderData(this, DataTableModel.ROOT_BOOKMARK, data);
			((DataSourceTreeModel)dataModel.getDataTableTree()).setRoot((TreeNode) root);
		}
	}
}


}
