/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseVirtualColumn.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:08 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.math.BigDecimal;
import java.sql.Types;
import java.util.Enumeration;

import com.klg.jclass.datasource.VirtualColumnModel;

/**
 * Adds columns to rows which are not retrieved from the data source.
 * This class allows you to add columns which are derived by performing an
 * <code>operation</code> on one or more other columns in the row to arrive
 * at a new value.<P>
 * The basic supported operations are:
 * <ul>
 *  <li>MIN         </li>
 *  <li>MAX         </li>
 *  <li>AVERAGE     </li>
 *  <li>SUM         </li>
 *  <li>PRODUCT     </li>
 *  <li>QUOTIENT    </li>
 *  <li>DIFFERENCE  </li>
 * </ul>
 * Operations can be performed on one or more columns. Columns are evaluated
 * from left to right.  Virutal columns can be used in calculations for other
 * virtual columns in the row.
 */
public class BaseVirtualColumn extends BaseColumn implements
    VirtualColumnModel {

static final long serialVersionUID = 2448569635266074965L;

/* operation constants */
public final static int MIN          = 0;
public final static int MAX          = 1;
public final static int AVERAGE      = 2;
public final static int SUM          = 3;
public final static int DIFFERENCE   = 4;
public final static int PRODUCT      = 5;
public final static int QUOTIENT     = 6;

protected int operation;
protected String columns[];
protected int scale       = 5;                       // arbitrary defaul value
protected int roundMode   = BigDecimal.ROUND_CEILING; // ditto

/**
 * Creates a column which is calculated based on one or more
 * data source columns.
 * For example,
 * <pre><code>
 *     String name = "difference";
 *     int type = java.sql.Types.INTEGER;
 *     int operation = BaseVirtualColumn.DIFFERENCE;
 *     String columns[] = {"colA", "colB", "colC"}
 *     BaseVirtualColumn bvc = new BaseVirtualColumn(name, type, operation, columns);
 * </code></pre>
 * This example is equivalent to the difference calculation colA - colB - colC.
 * @param name the String name of the column to be displayed.
 * @param type the java.sql.Types type of the column, will be automatically
 * mapped to a MetaDataModel type.
 * @param operation an int constant one of MIN, MAX, AVERAGE, SUM, PRODUCT,
 * QUOTIENT, DIFFERENCE.
 * @param columns an array of column identifiers indicating which columns to
 * use in the operation.
 */
public BaseVirtualColumn(String name, int type, int operation, String[] columns) {
    setColumnName(name);
    setIdentifier(getColumnName());          // same in this case
    this. columns   = columns;
    setColumnType(type);
    this.operation  = operation;
    setMetaColumnTypeFromSqlType(type);      // set the meta column type, one of MetaDataModel
                                             // type constants

}

/**
 * Returns the value calculated by performing <code> operation </code>
 * on <code> columns </code>.
 * @param DataTable the DataTableModel, a row from which will be used in the calculation.
 * @param bookmark a long which uniquely identifies a row.
 * @return the value calculated by performing <code> operation </code>
 * on <code> columns</code>.
 * @exception DataModelException If the operation fails.
 */
public Object getResultData(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal value = new BigDecimal("0.0");
    switch (operation) {
    case MIN:       value = getMinimum(dtm, bookmark); break;
    case MAX:       value = getMaximum(dtm, bookmark); break;
    case AVERAGE:   value = getAverage(dtm, bookmark); break;
    case SUM:       value = getSum(dtm, bookmark); break;
    case DIFFERENCE:value = getDifference(dtm, bookmark); break;
    case PRODUCT:   value = getProduct(dtm, bookmark); break;
    case QUOTIENT:  value = getQuotient(dtm, bookmark); break;
    }
    return convertToTargetType(value);
}

/** Given a BigDecimal value, convert it to the target type */
private Object convertToTargetType(BigDecimal value) {
    switch (getMetaColumnType()) {
    case MetaDataModel.TYPE_DOUBLE:
    case MetaDataModel.TYPE_FLOAT:      return new Double(value.doubleValue());
    case MetaDataModel.TYPE_BIG_DECIMAL:return value;
    case MetaDataModel.TYPE_BYTE:
    case MetaDataModel.TYPE_INTEGER:    return new Integer(value.intValue());
    case MetaDataModel.TYPE_LONG:       return new Long(value.intValue());
    case MetaDataModel.TYPE_SHORT:      return new Short(value.shortValue());
    }
    return value;
}

/** Returns the smallest value from the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getMinimum(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal smallest = new BigDecimal(Double.MAX_VALUE);
    for (int i = 0; i < columns.length; i++) {
        BigDecimal new_value = getBigDecimalValue(dtm.getResultData(bookmark, columns[i]));
        smallest = new_value.min(smallest);
    }
    return smallest;
}

/** Returns the largest value from the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getMaximum(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal largest = new BigDecimal(Double.MIN_VALUE);
    for (int i = 0; i < columns.length; i++) {
        BigDecimal new_value = getBigDecimalValue(dtm.getResultData(bookmark, columns[i]));
        largest = new_value.max(largest);
    }
    return largest;
}

/** Returns the average of the values from the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getAverage(DataTableModel dtm, long bookmark) throws DataModelException {
    return getSum(dtm, bookmark).divide(new BigDecimal(columns.length), scale, roundMode);
}

/** Returns the sum of the values from the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getSum(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal sum = new BigDecimal("0.0");
    for (int i = 0; i < columns.length; i++) {
        sum = sum.add(getBigDecimalValue(dtm.getResultData(bookmark, columns[i])));
    }
    return sum;
}

/** Returns the difference of the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getDifference(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal difference = new BigDecimal("0.0");
    boolean first = true;
    for (int i = 0; i < columns.length; i++) {
        if (first) {
            difference = getBigDecimalValue(dtm.getResultData(bookmark, columns[i]));
            first = false;
        }
        else {
            difference = difference.subtract(getBigDecimalValue(dtm.getResultData(bookmark, columns[i])));
        }
    }
    return difference;
}

/** Returns the product of the values from the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getProduct(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal product = new BigDecimal("0.0");
    boolean first = true;
    for (int i = 0; i < columns.length; i++) {
        if (first) {
            product = getBigDecimalValue(dtm.getResultData(bookmark, columns[i]));
            first = false;
        }
        else {
            product = product.multiply(getBigDecimalValue(dtm.getResultData(bookmark, columns[i])));
        }
    }
    return product;
}

/** Returns the quotient of the values from the indicated columns.
 * @param table the DataTableModel from which values are to be derived
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If the operation fails.
 */
private BigDecimal getQuotient(DataTableModel dtm, long bookmark) throws DataModelException {
    BigDecimal quotient = new BigDecimal("0.0");
    boolean first = true;
    for (int i = 0; i < columns.length; i++) {
        if (first) {
            quotient = getBigDecimalValue(dtm.getResultData(bookmark, columns[i]));
            first = false;
        }
        else {
            quotient = quotient.divide(getBigDecimalValue(dtm.getResultData(bookmark, columns[i])), scale, roundMode);
        }
    }
    return quotient;
}

/**
 * Helper method to get a BigDecimal value from a Object.
 *
 * @param value A value in Object form.
 */
private BigDecimal getBigDecimalValue(Object value) {
    if (value == null) {
        return new BigDecimal("0.0");
    }
    return new BigDecimal(value.toString());
}

/**
 * Sets the scale to be used when dividing.
 * The default value is 5.
 * @param scale an int which indicates the scale.
 * @see #getScale
 * @see java.math.BigDecimal
 */
public void setScale(int scale) {
    this.scale = scale;
}

/**
 * Returns the scale which is used when dividing.
 * The default value is 5.
 * @return the scale.
 * @see #setScale
 * @see java.math.BigDecimal
 */
public int getScale(int scale) {
    return scale;
}

/**
 * Sets the roundMode to be used when dividing.
 * The default mode is BigDecimal.ROUND_CEILING.
 * @param roundMode an int on of the java.math.BigDecimal
 * roundMode constants.
 * @see java.math.BigDecimal
 */
public void setRoundMode(int roundMode) {
    this.roundMode = roundMode;
}

/**
 * Returns the roundMode.
 * The default mode is BigDecimal.ROUND_CEILING.
 * @return the method used to round numbers when dividing.
 * @see java.math.BigDecimal
 */
public int getRoundMode() {
    return roundMode;
}

/**
 * Returns the operation.
 * @return the operation to be performed, one of
 * <ul>
 *  <li>MIN         </li>
 *  <li>MAX         </li>
 *  <li>AVERAGE     </li>
 *  <li>SUM         </li>
 *  <li>PRODUCT     </li>
 *  <li>QUOTIENT    </li>
 *  <li>DIFFERENCE  </li>
 * </ul>
 */
public int getOperation() {
    return operation;
}

/**
 * Returns the list of names of the columns used to derive
 *	the virtual column.
 * @return a String[] array of the column names.
 */
public String[] getColumns() {
    return columns;
}

public boolean isReadOnly() {
    return true;
}

public void setReadOnly(boolean isReadOnly) {
    // this is a no-op
}

}

