/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataModel.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:08 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import javax.swing.tree.*;

/**
 * This is the data interface for JClass DataSource.
 * All data for JClass DataSource is maintained
 * and manipulated in this data model through its sub-interfaces.
 *
 * This data model requires the implementation of two tree models,
 * one for describing the relationships of the hierarchical data 
 * (MetaDataModel) and one for the actual data (DataTableModel).
 */
public interface DataModel {

/**
 * Returns a handle to a TreeModel containing meta-data.
 * @return a TreeModel which can be traversed to determine the
 * hierarchy of the MetaDataModels.
 */
public TreeModel getMetaDataTree();

/**
 * Returns a handle to the DataTree.
 * @return a TreeModel which can be traversed to the hierarchy 
 * of DataTableModels.
 */
public TreeModel getDataTableTree();

/**
 * Registers a listener for this DataModel.
 * @see com.klg.jclass.datasource.DataModelListener
 */
public void addDataModelListener(DataModelListener listener);

/**
 * Removes a listener on this DataModel.
 * @see com.klg.jclass.datasource.DataModelListener
 */
public void removeDataModelListener(DataModelListener listener);

/**
 * Called by observers to tell this data model to which row it
 * should move.  Depending on the commit policy,
 * moving the internal cursor may automatically commit changes
 * to the originating data source.
 *
 * @param originator the DataModelListener which initiated this action
 * @param  data_table the DataTable to which the cursor should be moved
 * @param  bookmark a long uniquely identify to which row the cursor
 *         should move
 * @return true if move succeeded, false otherwise
 * @exception DataModelException If move to row fails.
 */
public boolean moveToRow(DataModelListener originator, DataTableModel data_table, long bookmark) throws DataModelException;

/**
 * For internal use only.
 * Called by observers to tell this data model to which row it
 * should move.  Depending on the commit policy,
 * moving the internal cursor may automatically commit changes
 * to the originating data source.
 *
 * @param originator the DataModelListener which initiated this action
 * @param  data_table the DataTable to which the cursor should be moved
 * @param  bookmark a long uniquely identify to which row the cursor
 *         should move
 * @param moveIdeCursor a boolean indicating if this move should also cause
 * the ide's cursor to move
 * @return true if move succeeded, false otherwise
 * @exception DataModelException If move to row fails.
 */
public boolean moveToRow(DataModelListener originator, DataTableModel data_table, long bookmark, boolean moveIdeCursor) throws DataModelException;

/**
 * Writes all pending changes in all DataTables to the data source.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If update fails.
 */
public void updateAll(DataModelListener originator) throws DataModelException;

/**
 * Cancels all pending changes in all DataTables.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If changes can't be canceled.
 */
public void cancelAll(DataModelListener originator) throws DataModelException;

/**
 * Requeries all DataTables.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If requery fails.
 */
public void requeryAll(DataModelListener originator) throws DataModelException;

/**
 * Returns a list of listeners on this DataModel.
 * @return an Object which is a list of listeners
 */
public Object getListeners();

/**
 * Passes event on to listeners of this DataModel.
 * @param event a DataModelEvent
 * @return true if the event was accepted by all listeners, or false
 * if at least one listener vetoed the change.
 */
public boolean fireDataModelEvent(DataModelEvent event);

/**
 * Passes event on to a single specific listener of this DataModel.
 * @param event a DataModelEvent
 * @param listener the DataModelListener
 * @return true if the event was accepted by the listener, or false
 * if the listener vetoed the change.
 */
public boolean fireDataModelEvent(DataModelEvent event, DataModelListener listener);

/**
 * Enables or disables data model events.  When disabled, events are not
 * dispatched to registered listeners.  This is useful for making mass changes
 * without having listeners repainting on each change.  That is,
 * set enabled to false, make the changes, then set back true.
 * <P>
 * This is different from the DataModelEvent
 * BEGIN_EVENTS and END_EVENTS events where events are still sent
 * but the listener having received the BEGIN_EVENTS event knows it may choose
 * to disregard the events until the END_EVENTS event is received.  They are
 * given the choice.
 * <P>
 * Temporarily disabling events is more efficient than using the BEGIN/END_EVENTS
 * method since the events are not dispatched to all listeners.
 * @param enable a boolean, true if events are to be dispatched to listeners,
 * false otherwise.
 * @see com.klg.jclass.datasource.DataModelEvent
 * @see #getEventsEnabled
 */
public void enableDataModelEvents(boolean value);

/**
 * Returns true if events are enabled, false otherwise.
 * @return true if events are enabled, false otherwise.
 * @see #enableDataModelEvents
 */
public boolean getEventsEnabled();

/**
 * Sets this model's name.
 * @param modelName a String which is to be this model's name
 */
public void setModelName(String modelName);

/**
 * Gets this model's name
 * @return a String which is the name of this model.
 */
public String getModelName();

/**
 * Returns the DataTableModel currently in focus.
 * @return the DataTableModel which has the global
 * cursor's focus
 * @see #moveToRow
 * @see #getCurrentGlobalBookmark
 */
public DataTableModel getCurrentGlobalTable();

/**
 * Returns the bookmark of the row which has the
 * global cursor's focus.
 * @return the bookmark of the row on which the
 * global cursor is located
 * @see #moveToRow
 * @see #getCurrentGlobalTable
 */
public long getCurrentGlobalBookmark();

/**
 * Returns the DataTable at the specified MetaData level based on
 * the current bookmark.  Bookmark refers to what we're calling
 * the "bookmark path" ie the bookmarks for each parent MetaData
 * level down to the specified level.
 * <P>
 * Will create the table if it does not exist.  In this case it
 * will be the child of the current row in the parent, and so on.
 *
 * @param originator the DataModelListener which initiated this action
 * @param metaDataModel the MetaDataModel for which the current
 * DataTableModel should be returned
 * @return the DataTableModel currently in focus for this level
 * @exception DataModelException If get fails.
 */
public DataTableModel getCurrentDataTable(DataModelListener originator, MetaDataModel myLevel) throws
DataModelException;


/**
 * Returns the data item at the specified MetaData level and column for
 * the current row.
 * Equivalent to calling DataTableModel.getResultData() with the
 * current bookmark at the MetaData level.
 * <P>
 * Will create the table if it does not exist.  In this case it
 * will be the child of the current row in the parent, and so on.
 *
 * @param originator the DataModelListener which initiated this action
 * @param metaDataModel the MetaDataModel from which the column value
 * is to be retrieved
 * @param column_identifier a String which is the column from which
 * a value is to be retrieved
 * @exception DataModelException If get fails.
 */
public Object getCurrentDataItem(DataModelListener originator, MetaDataModel myLevel, String column) throws
DataModelException;

/**
 * Gets a MetaDataModel instance by name, or null if not found.
 * @param metaDataName a String which is the name of the MetaDataModel object
 * @return the MetaDataModel for which getDescription() == metaDataName, or
 * null if not found.
 * @see MetaDataModel#setDescription
 * @see MetaDataModel#getDescription
 */
public MetaDataModel getMetaData(String metaDataName);

/**
 * Are there any pending modifications in any rows in this entire
 * data model?
 * @return true if there is at least one change pending, false otherwise
 */
public boolean isModified();

/**
 * Clear the current path of all children of this MetaDataModel.
 * @param parentMeta the MetaDataModel whose children are to be
 * cleared.
 */
public void clearCurrentPath(MetaDataModel parentMeta);

/**
 * Clears all user data.
 * Clears all Objects which were arbitrarily associated with
 * the entire DataTableTree.
 * @param user an Object, a reference to the user.
 */
public void clearDataTableTreeUserData(Object user);

/**
 * For internal use only.
 * Clears all internal data.
 * Clears all Objects which were arbitrarily associated with
 * the entire DataTableTree.
 * @param user an Object, a reference to the user.
 */
public void clearDataTableTreeInternalData(Object user);

}
