/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataModelEvent.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:08 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.util.*;

/**
 * DataModelEvent describes changes to the data source.
 * Interested listeners can then query this data source
 * to reflect the changes in their display.
 * <P>
 * These "BEFORE_" events can be ignored.  They are simply to allow
 * users to cancel the event.
 * <ul>
 * <li>BEFORE_INSERT_ROW</li>
 * <li>BEFORE_RESET</li>
 * <li>BEFORE_REQUERY</li>
 * <li>BEFORE_EDIT_CELL</li>
 * <li>BEFORE_DELETE_ROW</li>
 * <li>BEFORE_COMMIT_ALL</li>
 * <li>BEFORE_CANCEL_ALL</li>
 * <li>BEFORE_CANCEL_ROW_CHANGES</li>
 * <li>BEFORE_COMMIT_CONDITIONAL</li>
 * <li>BEFORE_COMMIT_ROW</li>
 * <li>BEFORE_MOVE_TO_CURRENT_ROW</li>
 * <li>BEFORE_DELETE_TABLE</li>
 * </ul>
 * These events are the minimum to which a listener must respond.
 * <ul>
 * <li>AFTER_CHANGE_OF_ROW_DATA</li>
 * <li>AFTER_INSERT_ROW</li>
 * <li>AFTER_DELETE_ROW</li>
 * <li>AFTER_RESET</li>
 * <li>AFTER_REQUERY_ROW_AND_DETAILS</li>
 * <li>AFTER_REQUERY_TABLE</li>
 * <li>AFTER_MOVE_TO_CURRENT_ROW</li>
 * </ul>
 * Other events.
 * <ul>
 * <li>INITIALIZE</li>
 * <li>BEGIN_EVENTS</li>
 * <li>END_EVENTS</li>
 * </ul>
 */
public class DataModelEvent extends EventObject {

static final long serialVersionUID = -1946917255406654889L;

// THE BEGIN EVENTS, THESE ARE SIMPLY TO ALLOW THE EVENTS TO BE
// CANCELLED, CAN BE IGNORED

/**
 * Event fired before a row is inserted. Can be cancelled.
 * If not cancelled this event will be followed by an
 * AFTER_INSERT_ROW event.
 * @see DataTableModel#addRow
 */
public final static int BEFORE_INSERT_ROW = 1;

/**
 * Event fired before entire grid is reset. Can be cancelled.
 * If not cancelled this event will be followed by an
 * AFTER_RESET event.
 * This event will result from a call to DataModel.requeryAll().
 * @see DataModel#requeryAll
 */
public final static int BEFORE_RESET = 2;

/**
 * Event fired when either DataTableModel.requeryRowAndDetails() or
 * DataTableModel.requeryRow() is called.
 * If not cancelled this event will be followed by an,
 * <ul>
 * <li>AFTER_REQUERY_ROW_AND_DETAILS event or an</li>
 * <li>AFTER_DELETE_ROW event in the case getRowStatus() == INSERTED or a</li>
 * <li>AFTER_CHANGE_OF_ROW_DATA event in the case getRowStatus() == UPDATED or COMMITTED.</li>
 * </ul>
 * @see DataTableModel#requeryRow
 * @see DataTableModel#requeryRowAndDetails
 */
public final static int BEFORE_REQUERY = 3;

/**
 * Event fired before a cell is edited. Can be cancelled.
 * @see DataTableModel#updateCell
 */
public final static int BEFORE_EDIT_CELL = 4;

/**
 * Event fired before a row is [logically] deleted. Can be cancelled.
 * If not cancelled this event will be followed by an
 * AFTER_DELETE_ROW or a AFTER_CHANGE_OF_ROW_DATA message if the commit policy
 * is COMMIT_MANUALLY or COMMIT_LEAVING_ANCESTOR.
 * @see DataTableModel#deleteRow
 * @see MetaDataModel#getCommitPolicy
 */
public final static int BEFORE_DELETE_ROW = 5;

/**
 * Event fired before all changes are committed. Can be cancelled.
 * All modified, deleted and inserted rows at all levels are about
 * to be committed.  BEFORE_DELETE_ROW and AFTER_CHANGE_OF_ROW_DATA events
 * will follow depending on the operations performed on the modified rows
 * being saved.
 * Results from a call to DataModel.updateAll.
 * @see DataModel#updateAll
 */
public final static int BEFORE_COMMIT_ALL = 6;

/**
 * Event fired before all changes are cancelled. Can be cancelled.
 * AFTER_INSERT_ROW and AFTER_CHANGE_OR_ROW_DATA events can follow this
 * event.
 * @see DataModel#cancelAll
 */
public final static int BEFORE_CANCEL_ALL = 7;

/** Event fired before all edits to a row are undone. Can be cancelled
 * An AFTER_DELETE_ROW or AFTER_CHANGE_OF_ROW_DATA event will follow.
 * @see DataTableModel#cancelRowChanges
 */
public final static int BEFORE_CANCEL_ROW_CHANGES = 8;

/**
 * Called when the root-level bookmark for a subtree changes.
 * When this happens those nodes in the previous subtree which
 * are not COMMIT_MANUALLY are committed.
 * Can be cancelled.  If cancelled the cursor moves but the
 * changes are automatically committed.
 */
public final static int BEFORE_COMMIT_CONDITIONAL = 9;

/**
 * Called before single row is committed to data source.  Can be cancelled in
 * which case the row edits are not written to the datasource and the rows
 * status remains modified.  AFTER_DELETE_ROW or AFTER_CHANGE_OF_ROW_DATA events will
 * followed depending on the status of the row to be committed.
 * @see DataTableModel#commitRow
 */
public final static int BEFORE_COMMIT_ROW = 10;

/**
 * The global cursor will move to a new row.
 * Can be cancelled.
 */
public final static int BEFORE_MOVE_TO_CURRENT_ROW = 11;

/**
 * The indicated DataTable will be deleted and flushed from the cache.
 * Can be cancelled.
 */
public final static int BEFORE_DELETE_TABLE = 12;

// THE AFTER EVENTS, NOT CANCELLABLE, must be > 100 since this is how
// we tell if an event is cancellable or not.  Those < 100 are; those
// above are not.


/**
 * A row has changed, re-read all its cells and its status to reflect the new
 * values.  If this event is the result of a cell edit call DataModelEvent.getColumn()
 * to get the name of the column which changed. If getColumn() returns "", re-read
 * the entire row.
 * Called when
 * <ul>
 * <li>a row is deleted and getShowDeletedRows() == true</li>
 * <li>a cell was edited</li>
 * <li>row edits are cancelled and getRowStatus == UPDATED</li>
 * <li>row edits are cancelled and getRowStatus == DELETED and getShowDeletedRows() == true</li>
 * <li>row is commited and getRowStatus == UPDATED or INSERTED</li>
 * <li>row is requeried and getRowStatus != INSERTED</li>
 * </ul>
 */
public final static int AFTER_CHANGE_OF_ROW_DATA = 100;

/**
 * A new row has been added to the datasource. Listeners need to
 * display the row.  Rows are always added to the end of DataTableModels.
 */
public final static int AFTER_INSERT_ROW = 101;

/**
 * Removes the row from the display.
 * A row has been physically deleted and needs to be
 * removed from the display or has been logically deleted but the
 * showDeletedRows property has been set to false. Called when,
 * <ul>
 * <li>a row has been logically deleted and getShowDeletedRows() == false</li>
 * <li>row changes have been cancelled and getRowStatus == INSERTED</li>'
 * <li>a row is committed and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>a row has been requeried and getRowStatus() == INSERTED</li>
 * </ul>
 */
public final static int AFTER_DELETE_ROW = 102;

/**
 * Listeners must close all expanded views and reset/reread the root node.
 * The previous pointer to the root node is no longer valid.  Call
 * DataModel.getDataTableTree().getRoot() for the new root table.
 * Called when the datasource has been reset.
 * @see DataModel#requeryAll
 */
public final static int AFTER_RESET = 103;

/**
 * Notification that multiple events are coming.
 * Multiple events will be nested
 * between BEGIN_EVENTS and END_EVENTS events.
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints.
 */
public final static int BEGIN_EVENTS = 104;

/**
 * Notification that multiple events are complete.
 * Multiple events will be nested
 * between BEGIN_EVENTS and END_EVENTS events.
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints. Called when
 * <ul>
 * <li>DataModel.updateAll is called</li>
 * <li>DataModel.cancelAll is called</li>
 * </ul>
 */
public final static int END_EVENTS = 105;

/**
 * The global cursor has moved to a new row.
 * Listeners should position their cursor on
 * the indicated row.  In a master-detail relationship child levels
 * should refresh themselves to reflect data sets which correspond
 * to the new parent row by calling DataModel.getCurrentDataTable()
 * or for field controls, DataModel.getCurrentDataItem().
 */
public final static int AFTER_MOVE_TO_CURRENT_ROW = 106;

/**
 * For internal use only
 * Used in the JCData.
 */
public final static int INITIALIZE = 107;

/**
 * The current row has been deleted and the originator
 * of the deletion should now reposition the global
 * cursor to a new, valid row.
 */
public final static int ORIGINATOR_NAVIGATE_ROW = 108;

/**
 * Re-reads the indicated row and refresh all open children
 * under this row.
 */
public final static int AFTER_REQUERY_ROW_AND_DETAILS = 109;

/**
 * Re-read this table and refresh all open children in
 * this table.
 */
public final static int AFTER_REQUERY_TABLE = 110;

/** The rowIndex or bookmark does not apply for this event */
public final static int NOT_APPLICABLE = -2;

protected int        rowIndex, command;
protected long       bookmark;
protected String     column;
protected Vector     ancestors = new Vector();
protected boolean    cancelled = false;
protected DataTableModel table;
protected DataModelListener originator;

/**
 * Constructor for DataModelEvent.
 *
 * @param source the containing DataModel
 * @param originator the DataModelListener which initiated this action
 * @param table the DataTableModel in which the change occurred
 * @param row_id an int specifying which row changed
 * @param column an String specifying which column changed
 * @param command an int indicating the what need to be done
 * @exception DataModelException If data model exceptions occur
 * within the construction of the event.
 */
public DataModelEvent(Object source, DataModelListener originator,
                      DataTableModel table, long bookmark, String column, int command)
throws DataModelException {
    super(source);
    if (originator == null) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.null_originator));
        throw new DataModelException(bookmark, column, table, command, e);
    }
    this.table      = table;
    this.originator = originator;
    this.column     = column;
    this.command    = command;
    this.bookmark   = bookmark;
    if (table != null) {
        this.rowIndex   = table.getRowIndex(bookmark);
    }
    else {
        this.rowIndex = NOT_APPLICABLE;
    }
}

/**
 * Returns the DataTableModel related to this event.
 * @return the DataTableModel which is the source of this event.
 * @see com.klg.jclass.datasource.DataTableModel
 */
public DataTableModel getTable() {
    return table;
}

/**
 * Returns the row index of the changed row.
 * @return a int indicating the current ordinal position of the
 * row in the DataTable
 */
public int getRowIndex() {
    return rowIndex;
}

/**
 * Returns the bookmark of the changed row
 * @return a long which uniquely identifies a row
 */
public long getBookmark() {
    return bookmark;
}

/**
 * Returns a string indicating which column changed, or "" in the case where
 * the column is not applicable.
 * @return the column name of the changed column
 */
public String getColumn() {
    return column == null ? "" : column;
}

/**
 * Gets the command which indicates what action should be taken.
 * @return an int, which is one of the constants from this class,
 * indicating the required action.
 */
public int getCommand() {
    return command;
}

/**
 * Returns a list of the bookmarks which comprise the
 * path from the root to the event node.  The event
 * node's bookmark is the last in the list.
 * @return a Vector of bookmarks from the root to the
 * event node.
 */
public Vector getAncestorBookmarks() {
    if (ancestors.size() > 0) {
        return ancestors;
    }
    // otherwise create a vector of bookmarks
    // for the path from the root to the event node
    if (getTable() != null && getTable() instanceof DataTableModel) {
        DataTableModel dtm = (DataTableModel) getTable();
        dtm.getAncestors(ancestors);
        ancestors.addElement(new Long(getBookmark()));
    }
    return ancestors;
}

/**
 * Cancels the proposed action.
 * Call this method to veto the pending action.
 * The default is to accept the change (false).
 * If isCancelable() returns false the cancel request is ignored.
 * @see #isCancelable
 * @see #getCancelled
 */
public void cancelProposedAction() {
    if (isCancelable()) {
        cancelled = true;
    }
}

/**
 * Sees if proposed action was cancelled by listener.
 * See if a listener vetoed the proposed action.
 * @return a boolean: true if cancelled.
 * @see #cancelProposedAction()
 */
public boolean getCancelled() {
    return cancelled;
}

/**
 * Returns true if this event can be cancelled. Cancelling an event
 * cancels the operation which generated the event.
 * @return true if this event can be cancelled.
 */
public boolean isCancelable() {
    // all cancellable enums are less than 100
    return getCommand() < 100;
}

/**
 * Returns the DataModelListener which initiated this event.
 * This is useful when responding to events to know if additional
 * processing is necessary based on whether you were the originator
 * or not.
 *
 * @return the DataModelListener which is the object which initated this
 * event
 */
public DataModelListener getOriginator() {
    return originator;
}

}
