/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataTableAbstractionLayer.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:09 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import javax.swing.tree.*;

/**
 * This interface is only of interest to those who wish to implement their
 * own data source using some other implementation.  Its methods should never
 * be called directly.  They will be called by the implementations of the
 * DataTableModel.
 * <P>
 * If you do wish to implement your own data source, simply extend
 * BaseDataTable (which abstractly implements these methods) and
 * concretely implement them.
 * <P>
 * This layer consists of a set of methods that abstract the dependencies
 * between the DataTableModel and the underlying implementation platform
 * (such as JDBC, VCafe, JBuilder).
 */
public interface DataTableAbstractionLayer {

/**
 * Returns a copy of a row, called by updateCell().
 * An implementation specific method for cloning rows, it
 * must be implemented in the derived class of BaseDataTable.
 * Clones a row so that its original values can be
 * restored when changes are cancelled, or so its
 * values can be used in the WHERE clause upon update.
 * @param bookmark a long which uniquely identifies a row
 * @return a an Object copy of this row.
 * @exception DataModelException If clone fails.
 */
public Object cloneRow(long bookmark) throws DataModelException;

/**
 * Changes the value of an existing cell, called by updateCell()
 * An implementation specific method for setting cell values, it
 * must be implemented in the derived class of BaseDataTable.
 * @param bookmark a long indicating the row
 * @param col_id a String indicating the column
 * @exception Exception If update of cell fails.
 * @see com.klg.jclass.datasource.DataTableModel#updateCell
 */
public void setValueAt(long bookmark, String col_id, Object value)
throws Exception;

/**
 * Restores a row's original values, called by cancelRowChanges().
 * An implementation specific method for restoring cell values, it
 * must be implemented in the derived class of BaseDataTable.
 * Will restore values from cache rather than from the data source.
 * @see com.klg.jclass.datasource.DataTableModel#cancelRowChanges
 * @see com.klg.jclass.datasource.DataTableModel#requeryRow
 */
public void restoreRow(long bookmark);

/**
 * Creates a new row, called by addRow().
 * An implementation specific method for creating a new (blank) row, it
 * must be implemented in the derived class of BaseDataTable.
 * @return an Object which is the new row just added.
 * @exception Exception If error occurs creating a new row.
 * @see com.klg.jclass.datasource.DataTableModel#addRow
 */
public Object createNewRow() throws Exception;

/**
 * Re-reads a row from the originating data source.
 * Called by requeryRow().
 * An implementation specific method for refetching a row, it
 * must be implemented in the derived class of BaseDataTable.
 * @param bookmark a long which uniquely identifies a row
 * @exception Exception If requery from server fails.
 * @see com.klg.jclass.datasource.DataTableModel#requeryRow
 */
public void refreshRow(long bookmark) throws Exception;

/**
 * Saves row changes to originating data source.
 * An implementation specific method for writing a row to the data source.
 * Called by commitRow().
 * @param bookmark a unique long indicating which row
 * to commit to the datasource.
 * @exception Exception If save fails.
 * @see com.klg.jclass.datasource.DataTableModel#commitRow
 */
public void saveRow(long bookmark) throws Exception;

/**
 * Requeries the data for the root level DataTable.
 * An implementation specific method to requery the root level node.
 * @exception Exception If requery from server fails.
 */
public DataTableModel requeryLevel() throws Exception;

/**
 * Returns a value for given bookmark/column_identifier.
 * An implementation specific method for getting a cell value.
 * @param bookmark a long which uniquely identifies this row
 * @param column_identifier a String which uniquely identifies this
 * column
 * @return the Object in the specified cell
 * @exception Exception If getCell() fails.
 */
public Object getCell(long bookmark,String column_identifer)
throws Exception;

/**
 * Called by DataTableModel.createTable to create a new table
 * based on the input criteria.
 * @param bookmark a long which uniquely identifies a row
 * @param childMetaDataModel a MetaDataModel, this tells this table
 * what structure the child table it is creating should have
 * @exception DataModelException If materializeTable fails.
 */
public DataTableModel materializeTable(DataModelListener originator, long bookmark, TreeNode childMetaDataModel)
throws DataModelException;

}
