/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataTableModel.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:09 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.sql.*;
import java.util.*;
import java.util.Date;

import javax.swing.tree.*;

/** 
 * This is the interface for data storage for the JClass DataSource 
 * datamodel. This data model will request data from instances of 
 * this table and will manipulate that data through this interface.
 * That is, rows can then be added, deleted or updated through
 * instances of its implementation.<P>
 *
 * To allow sorting of rows and columns, all operations access 
 * cell data using unique identifiers for rows and columns, rather
 * than using row-column indexes.
 * <P>
 * Each instance of this class has its own cursor which can be
 * navigated independently of the DataModel's global cursor.  Only the global
 * cursor (controlled by DataModel.moveToRow) causes commits to occur when the
 * current row is changed.  These methods do not cause the global cursor to
 * change, only this table's private cursor.
 * @see #first
 * @see #last
 * @see #next
 * @see #previous
 * @see #beforeFirst
 * @see #afterLast
 * @see #isBeforeFirst
 * @see #isAfterLast
 * @see #relative
 * @see #absolute
 * @see #getCurrentBookmark
 * @see com.klg.jclass.datasource.DataModel#moveToRow
 */
public interface DataTableModel {

/**
 * One or more of this row's cells have been modified.
 */
public final static int UPDATED         = 1;

/**
 * This is a new (uncommitted) row.
 */
public final static int INSERTED        = 2;

/**
 * This row has been logically deleted, but the 
 * delete has not yet been committed to the data source.
 */
public final static int DELETED         = 3;

/**
 * This row has no changes pending and is identical
 * to the data source row.
 */
public final static int COMMITTED       = 4;

/**
 * This row does not exist in the model. Either the bookmark
 * is invalid or the row has been permanently deleted.
 */
public final static int NONEXISTENT     = 5;

/**
 * Root-level rows are a special case since they
 * do not have parent rows.  This constant is used
 * to indicate this.
 */
public final static long ROOT_BOOKMARK = -1;

/**
 * Table's private cursor is before the first
 * row
 */
public final static int CURSOR_BEFORE_FIRST = -1;

/**
 * ROW_NOT_FOUND during lookup
 */
public static int ROW_NOT_FOUND = -1;

/**
 *	Returns the number of rows in this DataTable.
 */
public int getRowCount();

/**
 * Returns the status of a row which will be one of INSERTED, UPDATED,
 * DELETED, COMMITTED, NONEXISTENT.
 * @param bookmark a long which uniquely identifies this row 
 * @return an int row status, will be one of INSERTED, UPDATED,
 * DELETED, COMMITTED, NONEXISTENT.
 */
public int getRowStatus(long bookmark);

/**
 * Returns a long which will uniquely identify this row.
 * Facilitates sorting: index changes, identifier remains constant.
 * @param index an int for the original index of this row
 * @return a unique long to identify this row from now on
 */
public long getRowIdentifier(int row);

/**
 * Returns the object specified by the given bookmark and column identifier name.
 * @param bookmark a long which uniquely identifies this row 
 * @param column_identifier a String which uniquely identifies this
 * column
 * @return the Object in the specified cell
 * @exception DataModelException If getResultData() fails.
 */
public Object getResultData(long bookmark, String column_identifer)
throws DataModelException;

/**
 * Creates and returns the DataTable which corresponds to the specifed
 * row of this parent for the indicated child MetaData object.
 * Always returns a non-null DataTable. Creates new bookmarks for all
 * returned rows.  To retrieve an existing DataTable (and preserve
 * bookmarks) call getTable() instead.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @param child the MetaDataModel object which will drive the next query
 * @return the child DataTableModel of results
 * @exception DataModelException If createTable() fails.
 * @see #getTable
 */
public DataTableModel createTable(DataModelListener originator, long bookmark, TreeNode node)
throws DataModelException;

/**
 * Searches the existing children of this node for
 * a child node which matches the input criteria.
 * Return that node if it exists, null otherwise.
 * @param bookmark a long which uniquely identifies a row
 * @param child the MetaDataModel object
 * @return the DataTable with matching bookmark and MetaDataModel
 * otherwise, null.
 * @see #createTable
 */
public DataTableModel getTable(long bookmark, TreeNode node);

/**
 * Returns the MetaDataModel for this table.
 * @return the MetaDataModel Object for this table
 */
public MetaDataModel getMetaData();

/**
 * Returns the parent bookmark for this table.
 * This is the immediate parent of this table.
 * @return the bookmark of the parent row for this DataTable
 * @see #getRootAncestorBookmark
 * @see #getAncestors
 */
public long getParentBookmark();

/**
 * Creates a new (blank) row that can be used as an
 * argument to one of the addRow() methods.
 * @return an Object which is the new row just added.
 * @exception DataModelException If create fails.
 */
public Object createNewRow() throws DataModelException;

/**
 * Adds a empty row whose status will be INSERTED.
 * Adds a blank row into which the user can input column data.
 * The record is appended to this ResultSet.  The record is saved to datasource
 * when the user commits the surrounding transaction.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator) throws DataModelException;

/**
 * Adds a populated row whose status will be 'status'.
 * Adds a pre-populated row whose status is dictated by the
 * 'status' parameter.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @param status an int, one of,
 * <ul>
 * <li>DataTableModel.COMMITTED</li>
 * <li>DataTableModel.DELETED</li>
 * <li>DataTableModel.INSERTED<li>
 * </ul>
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row, int status) throws DataModelException;

/**
 * Adds a populated row whose status will be 'status'.
 * Adds a pre-populated row whose status is dictated by the
 * 'status' parameter.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @param status an int, one of,
 * <ul>
 * <li>DataTableModel.COMMITTED</li>
 * <li>DataTableModel.DELETED</li>
 * <li>DataTableModel.INSERTED<li>
 * </ul>
 * @param ordinal the index position into which the row should be inserted
 *
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row, int status, int ordinal) throws DataModelException;

/**
 * Adds a populated row whose status will be INSERTED.
 * Adds a pre-populated row.
 * The record is appended to this ResultSet.  The record is saved to datasource
 * when the user commits the surrounding transaction.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row) throws DataModelException;

/**
 * Deletes a row.
 * Logically deletes a row by setting its status to DELETED.  Will
 * be deleted permanently upon commit, or restored on rollback.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating the row to be deleted.
 * @exception DataModelException If logical delete fails.
 */
public void deleteRow(DataModelListener originator, long bookmark) throws DataModelException;

/** 
 * Updates a cell.
 * Calls setValueAt().
 * User has modified a cell. Call this method to inform this
 * data source of that change.  This method will then fire a
 * DataTableEvent to inform listeners about this change.
 * Internally a copy of the original row will be maintained which can
 * be used if an update is cancelled.
 * This row's status is set to UPDATED.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating in which row
 * the change was made
 * @param column_identifer a unique String indicating in which
 * column the change occurred.
 * @param value the Object to be placed in the cell
 * @exception DataModelException If error occurs updating a cell.
 * @see com.klg.jclass.datasource.DataTableAbstractionLayer#setValueAt
 */
public void updateCell(DataModelListener originator, long bookmark, String column_identifer,
                       Object value) throws DataModelException;

/**
 * Commits the row.
 * Writes the row (new and/or modified) to the data source. Calls
 * saveRow(), an implmentation
 * specific method to save a row to the data source.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating which row
 * to commit to the datasource.
 * @return true if the commit succeeded, false if failed or was cancelled.
 * @exception DataModelException If commit fails.
 * @see com.klg.jclass.datasource.DataTableAbstractionLayer#saveRow
 */
public boolean commitRow(DataModelListener originator, long bookmark) throws DataModelException;

/**
 * Do not use: internal method, call DataModel.updateAll() instead.
 * Recursively commit all uncommited rows, regardless of commit policy.
 * Used internally; to commit all changes call DataModel.updateAll().
 * Call commitRow for each new and/or modified row in this DataTable.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If any row fails to commit.
 * @see DataModel#updateAll
 */
public void commitAll(DataModelListener originator) throws DataModelException;

/**
 * Do not use: internal method.
 * Recurse on this node and each of its children and commit
 * all changes in each if commitPolicy does not equal COMMIT_MANUALLY.
 * Used internally.  To commit all changes call DataModel.updateAll() instead.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If write to data source fails.
 */
public void conditionallyCommitYourselfAndChildren(DataModelListener originator) throws DataModelException;

/**
 * Do not use this internal method; call DataModel.cancelAll() instead.
 * Recursively cancels all uncommited rows.
 * Restores original cell values for each uncommited row, for new rows
 * remove them.  Calls cancelRow for each uncommited row.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If changes can't be canceled.
 */
public void cancelAllRowChanges(DataModelListener originator) throws DataModelException;

/**
 * Cancels uncommited row changes.
 * Calls restoreRow().
 * Restores original cell values for this row, or if a new row remove it.
 * Restores values from cache (rather than from originating data source)
 * if supported in your environment.  To re-read the row from the
 * data source see requeryRow().
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a unique long indicating which row to uncommit.
 * @see com.klg.jclass.datasource.DataTableAbstractionLayer#restoreRow
 * @see #requeryRow
 * @exception DataModelException If changes can't be canceled.
 */
public void cancelRowChanges(DataModelListener originator, long bookmark) throws DataModelException;

/**
 * Recurses up the tree and determine the row number of
 * its ultimate ancestor.
 * @param bookmark a unique long for which the ultimate
 * ancestor bookmark is to be determined
 * @return a long indicating the bookmark of this
 * DataTable's ultimate ancestor.
 * @see #getAncestors
 */
public long getRootAncestorBookmark(long bookmark);

/**
 * Returns a list of this DataTable's ancestor bookmarks.
 * The list is ordered from the root ancestor down to the node
 * above this one.
 * @param  ancestors the Vector to be populated with
 *         ancestor bookmarks
 * @return a Vector of ancestor bookmarks.
 */
public void getAncestors(Vector ancestors);

/**
 * Requeries a DataTable row and clear all its children.
 * Deletes all child nodes then calls requeryRow().
 *
 * @param originator the DataModelListener which initiated this action
 * @return this data table model
 * @exception DataModelException If requery from server fails.
 */
public void requeryRowAndDetails(DataModelListener originator, long bookmark) throws DataModelException;

/**
 * Re-reads a row from the originating data source.
 * Calls refreshRow(bookmark), the implementation specific
 * method to refresh a row.
 *
 * @param originator the DataModelListener which initiated this action
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If requery from server fails.
 */
public void requeryRow(DataModelListener originator, long bookmark) throws DataModelException;

/**
 * Do not use: internal method.
 * Requeries the root DataTable.
 * Refetches all rows in this DataTable from the
 * originating data source. Only to be called on Root.
 * @return this data table model
 * @exception DataModelException If requery fails.
 */
public DataTableModel requeryLevel() throws DataModelException;

/**
 * Positions this table's private cursor on the first row.
 */
public boolean first();

/**
 * Position this table's private cursor on the last row.
 */
public boolean last();

/**
 * Advances this table's private cursor by one row.
 * Returns true if there was a row to move forward to.
 * @return true if move to next row succeeded.
 */
public boolean next();

/**
 * Positions this table's private cursor to the previous row.
 * Return true if the move succeeds.
 */
public boolean previous();

/**
 * Positions this table's private cursor before the first row.
 */
public void beforeFirst();

/**
 * Positions this table's private cursor after the last row.
 */
public void afterLast();

/**
 * Returns true if this table's private cursor is positioned before first row.
 * @return true if cursor is positioned before first row, false otherwise.
 */
public boolean isBeforeFirst();

/**
 * Returns true if this table's private cursor is positioned
 * after the last row.
 * @return true if the cursor is positioned after the last row,
 * false otherwise.
 * @see #getCurrentBookmark
 */
public boolean isAfterLast();

/**
 * Moves this table's private cursor forward or
 * backward by "n" rows from the current position.
 * Positive numbers move forward, negative backwards.
 * @param number_of_rows an int indicating the number of rows to move.
 * @exception DataModelException If move fails.
 * @see #getCurrentBookmark
 */
public void relative(int rows) throws DataModelException;

/**
 * Moves this table's private cursor to this ordinal position.
 * Ordinal positions start at 1.
 * @param ordinal_position an int indicating to which row to move
 * @exception DataModelException If move fails.
 * @see #getCurrentBookmark
 */
public void absolute(int ordinal_position) throws DataModelException;

/**
 * Returns the bookmark of the row on which this table's private cursor
 * is located.
 * <P>
 * Each instance of this class has its own cursor which can be
 * navigated independently of the DataModel's global cursor.  Only the global
 * cursor (controlled by DataModel.moveToRow) causes commits to occur when the
 * current row is changed.
 * <P>
 * @return the bookmark of the current row.
 * @exception DataModelException If there are no rows or
 * or an error occurs.
 * @see #first
 * @see #last
 * @see #next
 * @see #previous
 * @see #beforeFirst
 * @see #afterLast
 * @see #isBeforeFirst
 * @see #isAfterLast
 * @see #relative
 * @see #absolute
 * @see com.klg.jclass.datasource.DataModel#moveToRow
 */
public long getCurrentBookmark() throws DataModelException;

/**
 * Gets user data.
 * Return the Object which was arbitrarily associated with
 * this table.
 * @param user an Object, a reference to the user.
 * @return the user Object associated with this table.
 */
public Object getUserData(Object user);

/**
 * Sets user data.
 * Allow users to augment this table by associating an
 * arbitrary Object reference with it.
 * @param user an Object, a reference to the user.
 * @param data an Object, an arbitrary reference to an Object.
 */
public void setUserData(Object user, Object data);


/**
 * For internal use only.
 * Gets internal data.
 * Return the Object which was arbitrarily associated with
 * this table.
 * @param user an Object, a reference to the user.
 * @return the internal Object associated with this table.
 */
public Object getInternalData(Object user);

/**
 * For internal use only
 * Sets internal data.
 * Allow component writers to augment this table by associating an
 * arbitrary Object reference with it.
 * @param user an Object, a reference to the user.
 * @param internalData an Object, an arbitrary reference to an Object.
 */
public void setInternalData(Object user, Object data);

/**
 * Recursively processes pending events from a transaction.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If changes can't be canceled.
 */
public void processEventQueue(DataModelListener originator) throws DataModelException;

/**
 * Recursively clears pending events from a transaction.
 * Called when an exception occurs. Events are not dispatched;
 * they are simply deleted.
 */
public void clearEventQueue();

/**
 * Moves the cursor to the row indicated by the bookmark.
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If move fails.
 */
public void moveCursorToRow(long bookmark) throws DataModelException;

/**
 * Gets the cursor of the data table model.  This is the current
 * ordinal position of the cursor for this table.
 */
public int getCursor();

/**
 * Given a bookmark, return the zero-relative row index in this table
 * @param bookmark a long which uniquely identifies a row
 * @return rowIndex a int which is the ordinal position of the row
 * in this table
 */
public int getRowIndex(long bookmark);

/**
 * Deletes the subtree of DataTables rooted with this DataTable
 * as its root.
 * @see com.klg.jclass.datasource.MetaDataModel#getCacheChildren
 * @exception DataModelException If the subtree can't be deleted.
 */
public void deleteCacheChildren(DataModelListener originator) throws DataModelException;

/**
 * Internal method, do not use.
 * Move the ide cursor to this row.
 * @param bookmark a long which uniquely identifies a row.
 * @exception DataModelException If move fails.
 */
public void moveToRow(long bookmark) throws DataModelException;

/**
 * Does the subtree rooted at subTreeRootBookmark contain
 * modified nodes?  Modified means the row at subTreeRootBookmark or
 * any of its child nodes has pending insert, update or delete changes.
 * @param subTreeRootBookmark a long, the bookmark of the row which is
 * the root of the subtree to be checked.
 * @return true if this subtree contains at least one modified
 * node, false otherwise.
 */
public boolean isSubTreeModified(long subTreeRootBookmark);

/**
 * Are there any modifications to this table or any of its child nodes?
 * @return true if there are modifications to this table or any of its
 * child nodes
 */
public boolean isTableOrChildrenModified();

/**
 * Commit all pending changes in this subtree.
 * Commits this row and all child nodes.
 *
 * @param originator the DataModelListener which initiated this action
 * @param subTreeRootBookmark a long, the bookmark of the row which is
 * the root of the subtree to be checked.
 * @exception DataModelException If commit fails.
 */
public void commitSubTree(DataModelListener originator, long subTreeRootBookmark) throws DataModelException;

/**
 * Does this DataTableModel contain any pending inserts, updates or
 * deletes?
 * @return true if there are pending modifications, false otherwise.
 */
public boolean isModified();

/**
 * Return a long array of updated row bookmarks.  These are the rows
 * which have updates pending.
 * @return a long[] of updated row bookmarks
 */
public long[] getUpdatedRowBookmarks();

/**
 * Return a long array of inserted row bookmarks.  These are the rows
 * which are inserted into the data model, but not yet persisted.
 * @return a long[] of inserted row bookmarks
 */
public long[] getInsertedRowBookmarks();

/**
 * Return a long array of logically deleted row bookmarks.  These are the
 * rows which are logically deleted, but not yet physically deleted from the
 * originating data source.
 * @return a long[] of logically deleted row bookmarks
 */
public long[] getDeletedRowBookmarks();

/**
 * Given a DataTable, returns whether it is an ancestor of
 * (or equal to) itself
 * @param table a DataTable which may be an ancestor
 * @return true if the table is an ancestor
 */
public boolean tableIsAncestor(DataTableModel table);

/**
 * Requery all rows in this table and collapse its child nodes.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If requery fails.
 */
public void requeryTable(DataModelListener originator) throws DataModelException;

}
