/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: MetaDataModel.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:09 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.sql.*;
import java.util.*;
import java.io.*;

import javax.swing.tree.*;

/**
 * This is the interface for objects that hold the meta-data
 * for data tables. 
 * <br><i>Please note that only the "get" methods are meant to be used 
 * by an application programmer.
 * All other methods are for DataSource's internal use only.</i><br>
 * There can be zero, one or more
 * data tables associated with a MetaDataModel, thus it is more efficient
 * to store the meta-data only once.
 * <P>
 * In terms of JClass DataSource, each MetaDataModel is a node of the
 * MetaTree.  The MetaTree describes the hierarchacal relations between
 * MetaDataModel objects.
 * <P>
 * Each MetaData has its own commit policy.  The default is COMMIT_LEAVING_RECORD.
 * A commit policy determines when changes will be written to the data source.
 * There are three policies:
 * <ul>
 * <li><code>COMMIT_LEAVING_RECORD</code></li>
 * <li><code>COMMIT_LEAVING_ANCESTOR</code></li>
 * <li><code>COMMIT_MANUALLY</code></li>
 * </ul>
 * When the policy is set to <code>COMMIT_LEAVING_RECORD</code> (the default) modifications
 * to a row will be automatically written to the originating data source when
 * the cursor moves to any other row (see DataModel.movetoRow(..)).
 * In the case of delete, the row is immediately deleted.
 * <P>
 * When the policy is <code>COMMIT_LEAVING_ANCESTOR</code> changes will be automatically
 * written to the originating data source when the cursor moves to a row which is not
 * in the same subtree as the current row.  For example, in a master-detail relationship
 * where the master table is Orders and the details table is OrderDetails, the cursor
 * can be moved to any of the OrderDetail lines which correspond to the current
 * Order.  As soon as the cursor is moved to a different Order or to an OrderDetail
 * line for a different Order, the previous edits will be automatically committed.
 * <P>
 * A policy of COMMIT_MANUALLY will never automatically commit changes.  Any row changes
 * will simply change the status of those rows (see DataTableModel.getRowStatus()).
 * You must then call DataTableModel.commitRow(bookmark) or DataModel.updateAll() to make the changes
 * permanent, or call DataTableModel.cancelRowChanges(bookmark) or DataModel.cancelAll()
 * to undo the changes.
 *
 * @see com.klg.jclass.datasource.DataModel#moveToRow
 * @see com.klg.jclass.datasource.DataTableModel#getRowStatus
 * @see com.klg.jclass.datasource.DataTableModel#commitRow
 * @see com.klg.jclass.datasource.DataModel#updateAll
 * @see com.klg.jclass.datasource.DataModel#cancelAll
 *
 */
public interface MetaDataModel extends TreeNode {

/**
 * Java data types used to map JDBC data types
 */
public static final int TYPE_BOOLEAN        = 1000;
public static final int TYPE_SQL_DATE       = 1001;
public static final int TYPE_DOUBLE         = 1002;
public static final int TYPE_FLOAT      	= 1003;
public static final int TYPE_INTEGER        = 1004;
public static final int TYPE_STRING         = 1005;
public static final int TYPE_BIG_DECIMAL   	= 1006;
public static final int TYPE_LONG           = 1007;
public static final int TYPE_SQL_TIME       = 1008;
public static final int TYPE_SQL_TIMESTAMP 	= 1009;
public static final int TYPE_OBJECT         = 1010;
public static final int TYPE_BYTE          	= 1011;
public static final int TYPE_SHORT         	= 1012;
public static final int TYPE_BYTE_ARRAY    	= 1013;
public static final int TYPE_UTIL_DATE     	= 1014;

/** Automatically save changes to originating data source when cursor bookmark changes */
public final static int COMMIT_LEAVING_RECORD      = 1;

/**
 * Automatically save changes to originating data source when the
 * new bookmark is not in the same subtree as the previous bookmark.
 */
public final static int COMMIT_LEAVING_ANCESTOR    = 2;

/** No automatic commit. */
public final static int COMMIT_MANUALLY            = 3;

/**
 * Change the query for this data control.
 * For customizer use during design-time.
 */
public void clear();

/**
 * Returns a string which uniquely identifies this column.
 * Used to access data rather than a column index which can
 * change when columns are sorted. 
 * @param column_index an int indicating for which column
 * a unique identifier should be returned.
 * @return a String which uniquely identifies a column
 * @exception DataModelException If a column identifier cannot obtained.
 */
public String getColumnIdentifier(int column_index) throws DataModelException;

/**
 * Returns the number of columns in this result set
 * @return an int indicating the number of columns in table
 * @exception DataModelException If getColumnCount() fails.
 */
public int getColumnCount() throws DataModelException;

/**
 * Returns the Java Object type for a column as an enum.
 * All data in the datasource are objects.  These are the
 * object types returned by this method:
 * <ul>
 *     <li>TYPE_BOOLEAN</li>
 *     <li>TYPE_SQL_DATE</li>
 *     <li>TYPE_DOUBLE</li>
 *     <li>TYPE_FLOAT</li>
 *     <li>TYPE_INTEGER</li>
 *     <li>TYPE_STRING</li>
 *     <li>TYPE_BIG_DECIMAL</li>
 *     <li>TYPE_LONG</li>
 *     <li>TYPE_SQL_TIME </li>
 *     <li>TYPE_SQL_TIMESTAMP</li>
 *     <li>TYPE_OBJECT</li>
 *     <li>TYPE_BYTE</li>
 *     <li>TYPE_SHORT</li>
 *     <li>TYPE_BYTE_ARRAY</li>
 *     <li>TYPE_UTIL_DATE</li>
 * </ul>
 * Thus if a call to this method returns TYPE_BOOLEAN, then the object
 * returned by a call to DataTableModel.getResultData(..) will be
 * a Boolean object. TYPE_BYTE_ARRAY returns a byte[] array which is a "quasi" object.
 * @param column_identifier an String uniquely identifying a column
 * @return an int representing the meta column type for this column, will be one of
 * <ul>
 *     <li>TYPE_BOOLEAN</li>
 *     <li>TYPE_SQL_DATE</li>
 *     <li>TYPE_DOUBLE</li>
 *     <li>TYPE_FLOAT</li>
 *     <li>TYPE_INTEGER</li>
 *     <li>TYPE_STRING</li>
 *     <li>TYPE_BIG_DECIMAL</li>
 *     <li>TYPE_LONG</li>
 *     <li>TYPE_SQL_TIME </li>
 *     <li>TYPE_SQL_TIMESTAMP</li>
 *     <li>TYPE_OBJECT</li>
 *     <li>TYPE_BYTE</li>
 *     <li>TYPE_SHORT</li>
 *     <li>TYPE_BYTE_ARRAY</li>
 *     <li>TYPE_UTIL_DATE</li>
 * </ul>
 * @exception DataModelException If the Java column type cannot be found.
 */
public int getJavaColumnType(String column_identifier) throws DataModelException;

/**
 * Sets the commit policy for this level.  Determines when
 * changes are permanently written to the originating data source.
 * <P>
 * When the policy is set to <code>COMMIT_LEAVING_RECORD</code> (the default) modifications
 * to a row will be automatically written to the originating data source when
 * the cursor moves to any other row (see DataModel.movetoRow(..)).
 * In the case of delete, the row is immediately deleted.
 * <P>
 * When the policy is <code>COMMIT_LEAVING_ANCESTOR</code> changes will be automatically
 * written to the originating data source when the cursor moves to a row which is not
 * in the same subtree as the current row.  For example, in a master-detail relationship
 * where the master table is Orders and the details table is OrderDetails, the cursor
 * can be moved to any of the OrderDetail lines which correspond to the current
 * Order.  As soon as the cursor is moved to a different Order or to an OrderDetail
 * line for a different Order, the previous edits will be automatically committed.
 * <P>
 * A policy of COMMIT_MANUALLY will never automatically commit changes.  Any row changes
 * will simply change the status of those rows (see DataTableModel.getRowStatus()).
 * You must then call DataTableModel.commitRow(bookmark) or DataModel.updateAll() to make the changes
 * permanent, or call DataTableModel.cancelRowChanges(bookmark) or DataModel.cancelAll()
 * to undo the changes.
 * @param commit_policy an int, one of COMMIT_LEAVING_RECORD,
 * COMMIT_LEAVING_ANCESTOR, or COMMIT_MANUALLY
 * @exception DataModelException If set fails.
 */
public void setCommitPolicy(int commit_policy) throws DataModelException;

/**
 * Returns the commit policy for this level.  Determines when
 * changes are permanently written to the originating data source.
 * Will be one of COMMIT_LEAVING_RECORD, COMMIT_LEAVING_ANCESTOR,
 * or COMMIT_MANUALLY.
 * @return an int, one of COMMIT_LEAVING_RECORD,
 * COMMIT_LEAVING_ANCESTOR, or COMMIT_MANUALLY
 */
public int getCommitPolicy();

/**
 * Return the DataModel which encapsulates this MetaDataModel.
 * @return the DataModel
 */
public DataModel getDataModel();

/**
 * Sets a boolean indicating if this table is allowed to
 * be inserted into.
 * The default is to allow this.
 * Care must be taken in the case where a row is composed of columns
 * from several tables joined in a one-to-one relationship.  The default
 * behaviour when a row is inserted is to insert into all the tables in the
 * row.  In many cases this is not the desired behaviour.  Use this method
 * to turn off insert of the tables you do not wish to be inserted. For example,
 * when entering a SalesOrder row which is a join between SalesOrder and
 * Customer, the default is to try to insert the SalesOrder information into
 * the SalesOrder table, and insert the Customer information into the Customer
 * Table.  In many cases, you would just want to the insert into SalesOrder
 * to occur.  To accomplish this you would setInsertAllowed for Customer to
 * false.
 * @param table a String name of the table
 * @param allowInsert a boolean indicating if inserts are allowed
 */
public void setInsertAllowed(String table, boolean v);

/**
 * Sets a boolean indicating if this table is allowed to
 * be updated.
 * The default is to allow this.
 * @param table a String name of the table
 * @param allowUpdate a boolean indicating if updates are allowed
 */
public void setUpdateAllowed(String table, boolean v);

/**
 * Sets a boolean indicating if rows in this table are allowed to
 * be deleted.
 * The default is to allow this.
 * Care must be taken in the case where a row is composed of columns
 * from several tables joined in a one-to-one relationship.  The default
 * behaviour when a row is deleted is to delete from all the tables in the
 * row.  In many cases this is not the desired behaviour.  Use this method
 * to turn off delete of the tables you do not wish to be deleted.
 * @param table a String name of the table
 * @param allowDelete a boolean indicating if deletes are allowed
 */
public void setDeleteAllowed(String table, boolean v);

/**
 * Returns true if this table can be inserted into,
 * false otherwise.
 * The default is to allow this.
 * @param table a String name of the table
 * @return true if table can be updated
 */
public boolean getInsertAllowed(String table);

/**
 * Returns true if this table can be updated.
 * The default is to allow updating.
 * @param table a String name of the table
 * @return true if table can be updated
 */
public boolean getUpdateAllowed(String table);

/**
 * Returns true if this table can be deleted, false otherwise.
 * The default is to allow deletions.
 * @param table a String name of the table
 * @return true if rows in this table can be deleted.
 */
public boolean getDeleteAllowed(String table);

/**
 * Returns an int which uniquely identifies this instance
 * of the MetaDataModel. This id is automatically generated.
 * @return an int which uniquely identifies this object.
 * @see #setMetaID
 */
public int getMetaID();

/**
 * For internal use only.
 * Assigns a unique id to a meta data instance
 * @param id an int which is the unique id for this instance
 * @see #getMetaID
 */
public void setMetaID(int id);

/**
 * Returns the user-assigned description of this instance, or ""
 * if the description has not been set.
 * @return a String which describes this instance.
 */
public String getDescription();

/**
 * Sets the user-assigned description of this instance.
 * @param description a String which describles this instance.
 */
public void setDescription(String description);

/**
 * Returns true if the transaction capabilities of the underlying
 * data source should be used when saving modifications
 * (transactions are supported and enabled).
 * If the underlying data source supports transactions and this
 * capability has been enabled, data modifications will be
 * wrapped in a transaction which can be committed or rolled back.
 * @exception DataModelException If an error occurs getting commit flag.
 * @see #commitTransaction
 * @see #rollbackTransaction
 */
public boolean getAutoCommit() throws DataModelException;

/**
 * Sets the auto commit flag.
 * If the underlying data source supports transactions and this
 * flag is set to true, a transaction will wrap data modifications.
 * @param value a boolean, true if transactions are to be used,
 * false otherwise.
 * @exception DataModelException If an error occurs setting commit flag.
 */
public void setAutoCommit(boolean value) throws DataModelException;

/**
 * Rolls back the transaction.
 * If the underlying data source supports transactions and
 * transactions have been enabled, calling this method will
 * rollback all pending transactions on the data source (on
 * this connection).
 * @exception DataModelException If rollback fails.
 * @see MetaDataModel#getAutoCommit
 */
public void rollbackTransaction() throws DataModelException;

/**
 * Commits transaction.
 * If the underlying data source supports transactions and
 * transactions have been enabled, calling this method will
 * commit all pending transactions on the data source (on
 * this connection).
 * @exception DataModelException If commit fails.
 */
public void commitTransaction() throws DataModelException;

/**
 * Returns the list of ColumnModel objects.
 * @return a vector of ColumnModel objects.
 * @see com.klg.jclass.datasource.ColumnModel
 */
public Vector getColumnObjects();

/**
 * Returns the ColumnModel for the given column.
 * @param column_identifier a String which uniquely identifies a column
 * @return a ColumnModel
 * @exception DataModelException If get fails.
 */
public ColumnModel getColumnObject(String column_identifier) throws DataModelException;

/**
 * Do permissions on the model allow this column to be editied?
 * @param column_identifier a String which uniquely identifies this column
 * @return true if edits are allowed, false otherwise
 * @see #setUpdateAllowed
 * @see #getUpdateAllowed
 * @exception DataModelException If the primary keys cannot be determined for
 * the update table.
 */
public boolean isUpdateAllowedOnColumn(String column) throws DataModelException;

/**
 * Add a column to the list of column objects.
 * @param columnObject ColumnModel to be added.
 * @exception DataModelException If duplicate column found.
 * @see com.klg.jclass.datasource.ColumnModel
 */
public void addColumn(ColumnModel column) throws DataModelException;

/**
 * Returns the native meta data object.
 * This is a DataSet in JBuilder and a RelationView
 * in VisualCafe.
 * @return the native meta data object for this level.
 */
public Object getNativeMetaData();

/**
 * Indicates if logically deleted rows showld be shown.  This attribute
 * can only be set if the commitPolicy is COMMIT_MANUALLY or
 * COMMIT_LEAVING_ANCESTOR, otherwise an exception is thrown.
 * @param value a boolean: true to show rows, false otherwise
 * @exception DataModelException If set fails.
 */
public void setShowDeletedRows(boolean value) throws DataModelException;

/**
 * Show logically deleted rows?
 * @return a boolean: true to show rows, false otherwise
 */
public boolean getShowDeletedRows();

/**
 * Given a column name, gets its index.
 * @param column_identifier a String indicating the column's identifier
 * @return an int which is this column's ordinal position
 * in the original result data.
 * @exception DataModelException If the operation fails.
 */
public int getColumnIndex(String name) throws DataModelException;

/**
 * Sets the maximum number of rows a query at this level can return.
 * Useful for design time where only a sampling of row is required.
 * Set to -1 to have no limit.
 * <P>
 * Does not apply to ide-specific implmentations.  Set the max rows
 * directly on the ide data object instead.
 * <P>
 * @param maxRows an int indicating the max rows to be returned
 * @see #getMaxRows
 */
public void setMaxRows(int maxRows);

/**
 * Returns the maximum number of rows which can be returned by a query.
 * <P>
 * Does not apply to ide-specific implmentations.  Set the max rows
 * directly on the ide data object instead.
 * <P>
 * @return the maximum number of rows which can be returned by a query
 * @see #setMaxRows
 */
public int getMaxRows();

/**
 * Sets a boolean indicating whether subtrees of DataTables rooted at this
 * level are eligible for deletion.  The defaul is false meaning
 * DataTableModel.deleteCacheChildren should be called when a subtree is
 * collapsed.
 * @param cacheChildren a boolean: true means children should be cached, false
 * means deleteCacheChildren should be called when a tree is collapsed.
 * @see #getCacheChildren
 * @see DataTableModel#deleteCacheChildren
 */
public void setCacheChildren(boolean cacheChildren);

/**
 * Returns a boolean flag indicating whether subtrees of DataTables rooted at this
 * level are eligible for deletion.
 * @return false if deleteCacheChildren should be called on this level when
 * it is collapsed. Return true if the children should be preserved, that is
 * no call should be made to deleteCacheChildren.
 * @see #setCacheChildren
 * @see DataTableModel#deleteCacheChildren
 */
public boolean getCacheChildren();

/**
 * Returns the data table which currently has focus.
 * @return a DataTableModel, the one currently in focus.
 */
public DataTableModel getCurrentDataTable();

/**
 * Returns the bookmark of the row currently in focus.
 * @return the bookmark of the row currently in focus.
 */
public long getCurrentBookmark();

public void setCurrentDataTable(DataTableModel currentDataTable);
public void setCurrentBookmark(long currentBookmark);

/**
 * Sets an alternate class to handle data storage and retrieval
 * responsibilities.  A "Class.forName()" will be done to
 * load this alternate class.
 * <P>
 * If set, BaseDataTable instances will use this instances of
 * this class rather than the default class ("Store") to
 * store and retrieve data.
 * <P>
 * @param className a String, the name of the class which implements
 * the StoreModel interface
 * @exception DataModelException If an error occurs loading the class.
 * @see StoreModel
 * @see #getStoreClass
 */
public void setStoreClass(String storeClassName) throws DataModelException;

/**
 * Returns the class responsible for data storage and
 * retrieval (implements StoreModel). If not set, the default class
 * (com.klg.jclass.datasource.Store) is loaded and used.
 * @return the class used to store/retrieve data
 * @exception DataModelException If get fails.
 */
public Class getStoreClass() throws DataModelException;

/**
 * Return the Binding object associated with this MetaData level.
 * @return the Binding object for this level.
 */
public BindingModel getBinding();
	
/**
 * Set the Binding object associated with this MetaData level.
 * @param the Binding object for this level.
 */
public void setBinding(BindingModel binding);

/**
 * Get the provider (if any) for this level.
 * <P>
 * @return ProviderModel the class which will
 * provide matrix data for this level
 * <P>
 * @see ProviderModel#getData
 * @see DataTableModel#createTable
 * @see DataTableModel#getTable 
 */
public ProviderModel getProvider();

/**
 * Set the ProviderModel class name which will provide
 * data for this level. This class will be instantiated
 * and to provide data.
 * <P>
 * @param the name of the provider class
 * @exception DataModelException if the provider fails
 * to be instantiated.
 */
public void setProviderClassName(String providerModelClassName) throws DataModelException;

}
