/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ReadOnlyBinding.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:09 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import com.klg.jclass.util.JCListenerList;

import java.io.Serializable;
import java.util.Enumeration;
import javax.swing.tree.TreeNode;

/**
 *
 * This class provides a single-level, 2-dimensional view
 * of a set of data. It groups all non-update methods and
 * handles read-only events. This class exists only to provide
 * a logical separation between read-only and non-read-only
 * methods and event handling.
 * @see Binding
 *
 */
public abstract class ReadOnlyBinding implements BindingModel, Serializable {

static final long serialVersionUID = 8241141894499940558L;

/**
 *
 * The meta data model for this level.
 *
 */
protected MetaDataModel metaDataModel;

/**
 *
 * The data model for this level.
 *
 */
public DataModel dataModel;

/**
 *
 * The parent of this level.
 *
 */
protected Binding parent;

/**
 *
 * Listeners to this level.
 *
 */

 protected JCListenerList listeners;


/**
 *
 * The data binding hierarchy string.
 *
 */
protected String dataBinding;

/**
 *
 * Used as the originator in cases where it's not really needed.
 *
 */
private DataModelListener eventAdapter = new DataModelAdapter();

/**
 *
 * Root-level constructor.
 * @param	dataModel the DataModel to which this binding
 *			will belong.
 *
 */
public ReadOnlyBinding(DataModel dataModel) {
	this.dataModel = dataModel;

	// assume the root mdm, users can override it later
	// by calling setMetaDataModel themselves
	this.setMetaDataModel((MetaDataModel)dataModel.getMetaDataTree().getRoot());
}

/**
 *
 * Constructor for child levels.
 * @param	parent the Binding which is the parent of this level.
 *
 */
public ReadOnlyBinding(Binding parent) {
	this(parent.getDataModel());
}

/**
 *
 * Get the table currently in scope.
 * @param 	originator the DataModelListener making this request.
 * @return	the DataTableModel currently in focus for this level.
 *
 */
public DataTableModel getCurrentDataTableModel(DataModelListener originator) throws DataModelException {
	return dataModel.getCurrentDataTable(originator, metaDataModel);
}

/**
 *
 * Get the meta data model for this level.
 * @return	the MetaDataModel for this level.
 *
 */
public MetaDataModel getMetaDataModel() {
	return metaDataModel;
}

/**
 *
 * Set the meta data model for this level.
 * @param	the MetaDataModel for this level.
 *
 */
public void setMetaDataModel(MetaDataModel metaDataModel) {
	this.metaDataModel = metaDataModel;
}

/**
 *
 * Get the parent Binding of this table.
 * @return	the Binding which is the parent of this level.
 *
 */
public Binding getParent() {
	return parent;
}

/**
 *
 * Get the data model containing this level.
 * @return	the DataModel which contains this level.
 *
 */
public DataModel getDataModel() {
	return this.dataModel;
}

/**
 *
 * Set the data model for this level.
 * @param	dataModel the data model for this level.
 *
 */
public void setDataModel(DataModel dataModel) {
	this.dataModel = dataModel;
}

/**
 *
 * Set the name for this level.
 * @param 	the name for this level.
 *
 */
public void setDescription(String name) {
	metaDataModel.setDescription(name);
}

/**
 *
 * Get the name for this level.
 * @return 	the name for this level.
 *
 */
public String getDescription() {
	return metaDataModel.getDescription();
}

/**
 *
 * Implements DataModel method addDataModelListener.
 * @param	the DataModelListener to be registered.
 *
 */
public void addBindingListener(DataModelListener listener) {
    listeners = JCListenerList.add(listeners, listener);
}

/**
 *
 * Implements DataModel method removeDataModelListener.
 * @param	the DataModelListener to be removed.
 *
 */
public void removeBindingListener(DataModelListener listener) {
    listeners = JCListenerList.remove(listeners, listener);
}

/**
 *
 * Returns a list of listeners
 * @return	the list of listeners.
 *
 */
public Object getListeners() {
    return listeners;
}

/**
 *
 * Return the number of rows in the table currently in scope.
 * @return	the number of rows in the table currently in scope.
 *
 */
public int getRowCount() {
	int rows = 0;
	try {
		rows = getCurrentDataTableModel(new DataModelAdapter()).getRowCount();
	} catch (DataModelException e) {
		//
	}
	return rows;
}

/**
 *
 * Return the number of columns in this table.
 * @return	the number of columns in this table.
 *
 */
public int getColumnCount() {
	int cols = 0;
	try {
		cols = metaDataModel.getColumnCount();
	} catch (DataModelException e) {
		//
	}
	return cols;
}

/**
 *
 * Move focus to a new row in the current table.
 * @param	originator the DataModelListener requesting the move.
 * @param	rowIndex the index of the row
 *
 */
public boolean moveToRowIndex(DataModelListener originator, int rowIndex) throws DataModelException {
	DataTableModel dtm = getCurrentDataTableModel(originator);
	long bookmark = dtm.getRowIdentifier(rowIndex);
	return dataModel.moveToRow(originator, dtm, bookmark);
}

/**
 *
 * Return true if data is available from this table.
 * @return	true if this table has data, false otherwise.
 *
 */
public boolean isDataAvailable() {
	boolean dataAvailable = false;
	try {
		DataTableModel dtm 	=  getCurrentDataTableModel(eventAdapter);
		long bookmark 		= metaDataModel.getCurrentBookmark();
        dataAvailable = dtm != null && bookmark != dtm.ROW_NOT_FOUND &&
                        dtm.getRowStatus(bookmark) != DataTableModel.NONEXISTENT &&
                        dtm.getRowStatus(bookmark) != DataTableModel.DELETED;
   	} catch (DataModelException e) {
		//
	}
	return dataAvailable;
}

/**
 *
 * Return the name of this column.
 * @param	colIndex the int column index.
 * @return	the String name of this column.
 *
 */
public String getColumnIdentifier(int colIndex) {
	String columnName = "";
	try {
		columnName = metaDataModel.getColumnIdentifier(colIndex);
	} catch (DataModelException e) {
		//
	}
	return columnName;
}

/**
 *
 * Return the MetaDataModel type for this column.
 * @param	columnName the name of the column.
 * @return	one of MetaDataModel.TYPE_*.
 * @see 	MetaDataModel
 *
 */
public int getJavaColumnType(String columnName) {
	int type = -1;
	try {
		type = metaDataModel.getJavaColumnType(columnName);
	} catch (DataModelException e) {
		//
	}
	return type;
}

/**
 *
 * Return the MetaDataModel type for this column.
 * @param	the columnIndex
 * @return	one of MetaDataModel.TYPE_*.
 * @see 	MetaDataModel
 *
 */
public int getJavaColumnType(int columnIndex) {
	return getJavaColumnType(getColumnIdentifier(columnIndex));
}

/**
 *
 * Get the value in a cell for a specified row.
 * @param	originator the DataModelListener making the commit request.
 * @param	rowIndex the row index.
 * @param	columnIndex the column index.
 *
 */
public Object getCell(DataModelListener originator, int rowIndex, int columnIndex) throws DataModelException {
	DataTableModel dataTableModel =	getCurrentDataTableModel(originator);
	return dataTableModel.getResultData(dataTableModel.getRowIdentifier(rowIndex),
                                            metaDataModel.getColumnIdentifier(columnIndex));
}

/**
 *
 * Get the value in a cell for a specified row.
 * @param	originator the DataModelListener making the commit request.
 * @param	rowIndex the row index.
 * @param	columnName the column name.
 *
 */
public Object getCell(DataModelListener originator, int rowIndex, String columnName) throws DataModelException {
	return getCell(originator, rowIndex, getColumnIndex(columnName));
}

/**
 *
 * Get the value in a cell for the row currently in focus.
 * @param 	originator the DataModelListener requesting the value.
 * @param	columnName the String column name for the column.
 * @return	the Object in this cell.
 * @exception DataModelException If get fails.
 *
 */
public Object getCurrentCell(DataModelListener originator, String columnName) throws DataModelException {
	Object o = dataModel.getCurrentDataItem(originator, metaDataModel, columnName);
	return o;
}

/**
 *
 * Return the index of the row currently in focus.
 * @return the index of the current row.
 *
 */
public int getCurrentRowIndex() {
	int rowIndex = -1;
    long bookmark = metaDataModel.getCurrentBookmark();
	try {
		DataTableModel dtm = getCurrentDataTableModel(new DataModelAdapter());
    	rowIndex = dtm != null ? dtm.getRowIndex(bookmark) : -1;
	} catch (DataModelException dme) { }
	return rowIndex;
}

/**
 *
 * Requery the specified row.
 * @param	originator the DataModelListener making the requery request.
 * @param	rowIndex the row index to be requeried.
 *
 */
public void requeryRow(DataModelListener originator, int rowIndex) throws DataModelException {
	DataTableModel dataTableModel =	getCurrentDataTableModel(originator);
	dataTableModel.requeryRow(originator, dataTableModel.getRowIdentifier(rowIndex));
}

/**
 *
 * Requery the row currently in focus.
 * @param	originator the DataModelListener making the requery request.
 *
 */
public void requeryCurrentRow(DataModelListener originator) throws DataModelException {
	requeryRow(originator, getCurrentRowIndex());
}

/**
 *
 * Requery a row and its details.
 * @param	originator the DataModelListener making the requery request.
 * @param	rowIndex the row index to be requeried.
 *
 */
public void requeryRowAndDetails(DataModelListener originator, int rowIndex) throws DataModelException {
	DataTableModel dataTableModel =	getCurrentDataTableModel(originator);
	dataTableModel.requeryRowAndDetails(originator, dataTableModel.getRowIdentifier(rowIndex));
}

/**
 *
 * Requery the row currently in focus and its details.
 * @param	originator the DataModelListener making the requery request.
 *
 */
public void requeryCurrentRowAndDetails(DataModelListener originator) throws DataModelException {
	requeryRowAndDetails(originator, getCurrentRowIndex());
}

/**
 *
 * Requery all levels.
 * @param	originator the DataModelListener making the requery request.
 *
 */
public void requeryAll(DataModelListener originator) throws DataModelException {
	dataModel.requeryAll(originator);
}

/**
 *
 * Return the index for this column name.
 * @return the index of this column.
 *
 */
public int getColumnIndex(String columnName) {
	try {
		return metaDataModel.getColumnIndex(columnName);
	}
	catch (DataModelException dme) {
		//
	}
	return -1;
}

/**
 *
 * Requery all rows in the table currently in focus.
 * @param	originator the DataModelListener making the requery request.
 *
 */
public void requeryLevel(DataModelListener originator) throws DataModelException {
	DataTableModel dataTableModel = getCurrentDataTableModel(originator);
	if (dataTableModel == null)
        dataModel.requeryAll(originator);
    else {
        long bm = dataTableModel.getParentBookmark();
        if (bm == -1) {
            dataModel.requeryAll(originator);
        }
        else {
            MetaDataModel parentLevel = (MetaDataModel) metaDataModel.getParent();
            DataTableModel parentData = getCurrentDataTableModel(originator);
            parentData.requeryRowAndDetails(originator, bm);
        }
    }
}

/**
 *
 * Is the event intend for this level?
 *
 */
public boolean isOurData(DataModelEvent e) {
    DataTableModel dtm = e.getTable();
	return dtm != null && dtm.getMetaData() == metaDataModel;
}

/**
 * Will these event impact this level or any of its
 * descendents?
 */
public boolean isOutDataOrOurAncestors(DataModelEvent e) {
	// listeners are only interested in events which are
	// upstream in the hierarchy, otherwise they have no
	// effect.
	return isOurData(e) || this.isDescendantOf(e.getTable().getMetaData());
}

public boolean isDescendantOf(MetaDataModel possibleParent) {
    TreeNode node = metaDataModel;
    while (node != null) {
        if (node.getParent() == possibleParent) {
            return true;
        }
        node = node.getParent();
    }
    return false;
}

public ColumnModel getColumnObject(String columnName) {
	ColumnModel columnModel = null;
	try {
		columnModel = metaDataModel.getColumnObject(columnName);
	}
	catch (DataModelException dme) {}
	return columnModel;
}

}
