/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ReadOnlyBindingListener.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:09 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

/**
 * Passive listeners -- objects interested in receiving only 
 * read-only DataModelEvents should implement this interface. 
 * These methods will be called by the data source AFTER its 
 * data has changed.
 * <P>
 * This interface is extended by the DataModelListener to add
 * events of interest to active listeners -- listeners who 
 * make changes to the data and need to know about changes made
 * by other components BEFORE they occur.
 * <P>
 * For convenience, override these methods in the DataModelAdapter
 * to satisfy a read-only implmentation of a DataModelListener.
 * 
 * @see DataModelEvent
 * @see DataTableModel
 * @see DataModelAdapter
 */
public interface ReadOnlyBindingListener extends java.util.EventListener {

/**
 * Listeners must close all expanded views and reset/reread the root node.
 * The previous pointer to the root node is no longer valid.  Call
 * DataModel.getDataTableTree().getRoot() for the new root table.
 * Called when the datasource has been reset.
 * @see DataModel#requeryAll
 */
public void afterReset(DataModelEvent e);

/**
 * The global cursor has moved to a new row.
 * Listeners should position their cursor on
 * the indicated row.  In a master-detail relationship child levels
 * should refresh themselves to reflect data sets which correspond
 * to the new parent row by calling DataModel.getCurrentDataTable()
 * or for field controls, DataModel.getCurrentDataItem().
 */
public void afterMoveToCurrentRow(DataModelEvent e);

/**
 * Re-reads the indicated row and refresh all open children
 * under this row.
 */
public void afterRequeryRowAndDetails(DataModelEvent e);

/**
 * Re-read this table and refresh all open children in
 * this table.
 */
public void afterRequeryTable(DataModelEvent e);
	
/**
 * A row has changed, re-read all its cells and its status to reflect the new
 * values.  If this event is the result of a cell edit call DataModelEvent.getColumn()
 * to get the name of the column which changed. If getColumn() returns "", re-read
 * the entire row.
 * Called when
 * <ul>
 * <li>a row is deleted and getShowDeletedRows() == true</li>
 * <li>a cell was edited</li>
 * <li>row edits are cancelled and getRowStatus() == UPDATED</li>
 * <li>row edits are cancelled and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>row is commited and getRowStatus() == UPDATED or INSERTED</li>
 * <li>row is requeried and getRowStatus() != INSERTED</li>
 * </ul>
 */
public void afterChangeOfRowData(DataModelEvent e);

/**
 * A new row has been added to the datasource. Listeners need to
 * display the row.  Rows are always added to the end of DataTableModels.
 */
public void afterInsertRow(DataModelEvent e);

/**
 * Removes the row from the display.
 * A row has been physically deleted and needs to be
 * removed from the display or has been logically deleted but the
 * showDeletedRows property has been set to false. Called when,
 * <ul>
 * <li>a row has been logically deleted and getShowDeletedRows() == false</li>
 * <li>row changes have been cancelled and getRowStatus() == INSERTED</li>'
 * <li>a row is committed and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>a row has been requeried and getRowStatus() == INSERTED</li>
 * </ul>
 */
public void afterDeleteRow(DataModelEvent e);
	
/**
 * Notification that multiple events are coming.
 * Multiple events will be nested
 * between beginEvents() and endEvents().
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints.
 */
public void beginEvents(DataModelEvent e);

/**
 * Notification that multiple events are complete.
 * Multiple events will be nested
 * between beginEvents() and endEvents().
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints. Called when
 * <ul>
 * <li>DataModel.updateAll is called</li>
 * <li>DataModel.cancelAll is called</li>
 * </ul>
 */
public void endEvents(DataModelEvent e);

/**
 * For internal use only
 * Used by the JCData bean.
 */
public void initialize(DataModelEvent e);

}
