/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: TreeData.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:09 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource;

import java.io.Serializable;
import java.sql.*;
import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.util.JCListenerList;

/**
 * An implementation of DataModel.
 * Implements trees and listener functions.
 */
public class TreeData implements DataModel, Serializable {

static final long serialVersionUID = -280385478977847078L;

/**
 * DataSourceTreeModel for metadata.
 */
protected DataSourceTreeModel meta_tree;

/**
 * DataSourceTreeModel for DataTables.
 */
protected transient DataSourceTreeModel data_tree;

/**
 * Listeners
 */
protected JCListenerList listeners;

/**
 * Previous data table, used for cursor
 */
protected transient DataTableModel currentDataTable;

/**
 * Previous row_identifier, used for cursor
 */
protected long currentBookmark = -1;

/**
 * enable events?
 */
protected boolean eventsEnabled = true;

/**
 * Static counter used to create default names for
 * instances of this class.
 */
protected static int nameCounter = 0;

/**
 * This model's name
 */
protected String modelName;

/**
 * A constructor for TreeData, an implementation of
 * DataModel which can be extened.
 */
public TreeData() {
    // instantiate the tree implementations
    meta_tree = new DataSourceTreeModel(DataSourceTreeModel.NullRoot);
    data_tree = new DataSourceTreeModel(DataSourceTreeModel.NullRoot);
    modelName = "TreeData" + nameCounter++;

 
}

/**
 * Implements DataModel method getMetaDataTree.
 */
public TreeModel getMetaDataTree() {
    return(TreeModel) meta_tree;
}

/**
 * Implements DataModel method getDataTree.
 */
public TreeModel getDataTableTree() {
    return(TreeModel) data_tree;
}

/**
 * Implements DataModel method addDataModelListener.
 */
public void addDataModelListener(DataModelListener listener) {
    listeners = JCListenerList.add(listeners, listener);
}

/**
 * Implements DataModel method removeDataModelListener.
 */
public void removeDataModelListener(DataModelListener listener) {
    listeners = JCListenerList.remove(listeners, listener);
}

/**
 * Returns a list of listeners
 */
public Object getListeners() {
    return listeners;
}

/**
 * Called by observers to tell this data model to which row it
 * should move.  Depending on the commit policy,
 * moving the internal cursor may automatically commit changes
 * to the originating data source.
 * <P>
 * Upon intialization of the data model the current data table is the
 * root data table and the current bookmark is the first row.
 *<P>
 * Fires a BEFORE_MOVE_TO_CURRENT_ROW which if cancelled, cancels the move and
 * an AFTER_MOVE_TO_CURRENT_ROW on a sucessful move.
 * <P>
 * During the move if the row or subtree which is being left is modified
 * a BEFORE_COMMIT_ROW or a BEFORE_COMMIT_CONDITIONAL message will be 
 * sent.  If either of these is cancelled, the move will fail.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @param data_table the DataTable to which the cursor should be moved
 * @param bookmark a long uniquely identify to which row the cursor
 * should move
 * @return true if the move succeeded, false otherwise (may have been cancelled)
 * @exception DataModelException If moveToRow fails.
 */
public boolean moveToRow(DataModelListener originator, DataTableModel data_table, long bookmark) throws DataModelException {
    return moveToRow(originator, data_table, bookmark, true);
}

/**
 * For internal use only.
 * Called by observers to tell this data model to which row it
 * should move.  Depending on the commit policy,
 * moving the internal cursor may automatically commit changes
 * to the originating data source.
 *
 * @param originator the DataModelListener which initiated this action
 * @param  data_table the DataTable to which the cursor should be moved
 * @param  bookmark a long uniquely identify to which row the cursor
 *         should move
 * @param moveIdeCursor a boolean indicating if this move should also cause
 * the ide's cursor to move
 * @return true if move succeeded, false otherwise
 * @exception DataModelException If move to row fails.
 */
public boolean moveToRow(DataModelListener originator, DataTableModel data_table,
                         long bookmark, boolean moveIdeCursor) throws DataModelException {
    // moving to the same previous row?
    if (currentDataTable != null && bookmark == currentBookmark) {
        return true;
    }

    // give the listeners the opportunity to cancel this operation
    if (!fireDataModelEvent(new DataModelEvent(this, originator, data_table, bookmark, "", DataModelEvent.BEFORE_MOVE_TO_CURRENT_ROW))) {
        return false;
    }

    // first time is a special case
    if (currentDataTable == null) {
        currentDataTable = (DataTableModel) getDataTableTree().getRoot();
        if (currentDataTable.first()) {
            currentBookmark = currentDataTable.getCurrentBookmark();
        }
        else {
            currentBookmark = bookmark;
        }
    }
    // since ide's cause a requery when moving, remember the position so we
    // can move back to it
    int rowIndex = data_table.getRowIndex(bookmark);

    // if still under same ancestor we need only look at previous row's policy
    // and this is not the first time (ie currentBookmark != bookmark)
    if (currentDataTable.getRootAncestorBookmark(currentBookmark)== data_table.getRootAncestorBookmark(bookmark)) {
        if (currentDataTable.getMetaData().getCommitPolicy() == MetaDataModel.COMMIT_LEAVING_RECORD &&
            currentDataTable.getRowStatus(currentBookmark) != DataTableModel.COMMITTED &&
            currentDataTable.getRowStatus(currentBookmark) != DataTableModel.NONEXISTENT &&
            currentBookmark != bookmark) {

            // will return false if someone cancels the commit row
            if (!currentDataTable.commitRow(originator, currentBookmark)) {
                return false;
            }
        }
    }
    else if (((DataTableModel)getDataTableTree().getRoot()).isSubTreeModified(currentDataTable.getRootAncestorBookmark(currentBookmark))) {
        // will return false if someone cancels the commit
        if (!conditionallyCommitSubTreeRootedAt(originator, currentDataTable.getRootAncestorBookmark(currentBookmark))) {
            return false;
        }
    }

    // This code is to fix a re-entrancy problem.  It will only execute
    // in the case of ides where a commit causes a requery all to be done.
    // The issue is the data table on which the move was called is now
    // stale.  We detect this by checking that the bookmark still exists
    // in data_table, for if data_table were refreshed, it won't.
    if (resetHappened(data_table, bookmark)) {
        bookmark = currentDataTable.getRowIdentifier(rowIndex);       
    }
    else {
        // for next time
        currentDataTable = data_table;
    }
    currentBookmark = bookmark;

    // move the ide's cursor
    if (moveIdeCursor) {
    	currentDataTable.moveToRow(bookmark);
    }
    fireDataModelEvent(new DataModelEvent(this, originator, currentDataTable, bookmark, "", DataModelEvent.AFTER_MOVE_TO_CURRENT_ROW));
    return true;


}

private boolean resetHappened(DataTableModel dataTable, long bookmark) {
    DataTableModel root = (DataTableModel) getDataTableTree().getRoot();
    // on a reset, we set the parent to null
    if (((TreeNode)dataTable).getParent() == null && dataTable != root) {
        return true;
    }
    // or root has changed
    if (root.getRowIndex(dataTable.getRootAncestorBookmark(bookmark)) == dataTable.ROW_NOT_FOUND) {
        return true;
    }
    return false;
}

/**
 * This method is called when the cursor has moved to a new ancestor.
 * Check the root's commit policy.  If it's COMMIT_LEAVING_RECORD 
 * commit the row at rootBookmark, otherwise leave it.  Then recurse
 * on child nodes of this bookmark.  If they are not COMMIT_MANUALLY,
 * then commit them, since both other cases (LEAVING_ ANCESTOR and RECORD are true).
 *
 * @param originator the DataModelListener which initiated this action
 * @param rootbookmark a long which uniquely identifies a row.
 * @return true if the commit succeeded, false if failed or was cancelled.
 * @exception DataModelException If the operation fails.
 */
private boolean conditionallyCommitSubTreeRootedAt(DataModelListener originator, long rootBookmark) throws DataModelException {
    DataTableModel dataRoot = (DataTableModel) data_tree.getRoot();
    MetaDataModel  metaRoot = (MetaDataModel) getMetaDataTree().getRoot();

    // give the listeners the opportunity to cancel this operation
    if (!fireDataModelEvent(new DataModelEvent(this, originator, dataRoot, rootBookmark, "", DataModelEvent.BEFORE_COMMIT_CONDITIONAL))) {
        return false;
    }

    try {
        // see if this root level row is modified
        if (dataRoot.getMetaData().getCommitPolicy() == MetaDataModel.COMMIT_LEAVING_RECORD  &&
            dataRoot.getRowStatus(rootBookmark) != DataTableModel.COMMITTED &&
            dataRoot.getRowStatus(rootBookmark) != DataTableModel.NONEXISTENT) {

            dataRoot.commitRow(originator, rootBookmark);
        }
        // now do its children
        for (Enumeration e = ((TreeNode)dataRoot).children(); e.hasMoreElements();) {
            DataTableModel dtm = (DataTableModel) e.nextElement();
            if (dtm.getParentBookmark() == rootBookmark) {
                dtm.conditionallyCommitYourselfAndChildren(originator);
            }
        }
        metaRoot.commitTransaction();
        dataRoot.processEventQueue(originator);
    }
    catch (DataModelException e) {
        // on exception, rollback all transactions that can be rolled back
        try {
            metaRoot.rollbackTransaction();
        }
        catch (DataModelException x) {
        }
        dataRoot.clearEventQueue();
        fireDataModelEvent(new DataModelEvent(this, originator, dataRoot, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.END_EVENTS));
    	throw e;
    }
    return true;
}

/**
 * Writes all pending changes in all DataTables to the data source.
 * Wraps recursive calls to commitAll() between
 * BEGING_EVENTS and END_EVENTS events so listeners can
 * choose to batch these events.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If updateAll fails.
 */
public void updateAll(DataModelListener originator) throws DataModelException {
    DataTableModel dataRoot = (DataTableModel)getDataTableTree().getRoot();
    MetaDataModel  metaRoot = (MetaDataModel) getMetaDataTree().getRoot();

    // give the listeners the opportunity to cancel this operation
    if (!fireDataModelEvent(new DataModelEvent(this, originator, dataRoot, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEFORE_COMMIT_ALL))) {
        return;
    }

    fireDataModelEvent(new DataModelEvent(this, originator, dataRoot, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEGIN_EVENTS));
    // Three step process:
    // 1. recursively write all changes to the database
    // 2. recursively commit those changes
    // 3. recursively flush queued events related to changes
    // If transactions are not enabled, steps 2 and 3 are no-ops
    //
    // The tricky bit occurs when multiple connections are used since
    // the commitTransaction can succeed on some and fail on others.
    // But any exception will cause processEventQueue to be skipped.
    // FIX.
    try {
        dataRoot.commitAll(originator);
        metaRoot.commitTransaction();
        dataRoot.processEventQueue(originator);
    }
    catch (DataModelException e) {
        // on exception, rollback all transactions that can be rolled back
        try {
            metaRoot.rollbackTransaction();
        }
        catch (DataModelException x) {
        }
        dataRoot.clearEventQueue();
        fireDataModelEvent(new DataModelEvent(this, originator, dataRoot, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.END_EVENTS));
    	throw e;
    }
    fireDataModelEvent(new DataModelEvent(this, originator, dataRoot, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.END_EVENTS));
}

/**
 * Cancels all pending changes in all DataTables.
 * Wraps recursive calls to cancelAllRowChanges() between
 * BEGIN_EVENTS and END_EVENTS events so listeners can
 * choose to batch these events.
 *
 * @param originator the DataModelListener which initiated this action.
 * @exception DataModelException If an exception occurs while canceling pending changes.
 */
public void cancelAll(DataModelListener originator) throws DataModelException {
    DataTableModel root = (DataTableModel)getDataTableTree().getRoot();

    // give the listeners the opportunity to cancel this operation
    if (!fireDataModelEvent(new DataModelEvent(this, originator, root, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEFORE_CANCEL_ALL))) {
        return;
    }

    fireDataModelEvent(new DataModelEvent(this, originator, root, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEGIN_EVENTS));
    // handle cancel of the current row, which may not have a modified status because the editor
    // is still active on a cell, yet we really want to cancel it
    if (currentDataTable != null && currentBookmark != DataTableModel.ROW_NOT_FOUND) {
        currentDataTable.cancelRowChanges(originator, currentBookmark);
    }
    // now do all rows with STATUS != COMMITTED
    root.cancelAllRowChanges(originator);
    fireDataModelEvent(new DataModelEvent(this, originator, root, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.END_EVENTS));
}

/**
 * Requeries all DataTables.
 * All previous DataTableModels are invalidated.
 * The root is reset in the DataTableTree.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If requeryAll fails.
 */
public void requeryAll(DataModelListener originator) throws DataModelException {
    DataTableModel root = (DataTableModel) data_tree.getRoot();

    // give the listeners the opportunity to cancel this operation
    if (!fireDataModelEvent(new DataModelEvent(this, originator, root, DataModelEvent.NOT_APPLICABLE, "", DataModelEvent.BEFORE_RESET))) {
        return;
    }

    ((DataSourceTreeNode) root).removeAllChildren();
    MetaDataModel metaRoot = (MetaDataModel)getMetaDataTree().getRoot();

    // clear out pointers to current DataTables and bookmarks
    clearCurrentPath(metaRoot); // must come before requeryLevel
    metaRoot.setCurrentDataTable(null);
    metaRoot.setCurrentBookmark(DataTableModel.ROW_NOT_FOUND);

    // reset the data table tree root
    DataTableModel newDataRoot = root.requeryLevel();
    data_tree.setRoot((TreeNode)newDataRoot);

    // reset the current path for the root
    metaRoot.setCurrentDataTable(newDataRoot);

    // set the bookmark
    newDataRoot.first();
    long newCurrentBookmark = newDataRoot.getCurrentBookmark();
    metaRoot.setCurrentBookmark(newCurrentBookmark);

    // update the global values
    currentDataTable = newDataRoot;
    currentBookmark = newCurrentBookmark;

    // copy the listeners and send the event
    DataModelEvent event = new DataModelEvent(this, originator, (DataTableModel) newDataRoot, newCurrentBookmark , "", DataModelEvent.AFTER_RESET);
    fireDataModelEvent(event);
}

/**
 * Clear the current path of all children of this MetaDataModel.
 * @param parentMeta the MetaDataModel whose children are to be
 * cleared.
 */
public void clearCurrentPath(MetaDataModel parentMeta) {
    if (parentMeta == null) {
        return;
    }
    Enumeration e = parentMeta.children();
    while (e.hasMoreElements()) {
        // work down to leaves
        MetaDataModel childMeta = (MetaDataModel) e.nextElement();
        clearCurrentPath2(childMeta);
    }
}

void clearCurrentPath2(MetaDataModel parentMeta) {
    if (parentMeta == null) {
        return;
    }
    Enumeration e = parentMeta.children();
    while (e.hasMoreElements()) {
        // work down to leaves
        MetaDataModel childMeta = (MetaDataModel) e.nextElement();
        clearCurrentPath2(childMeta);
    }
    // now on way back up clear
    parentMeta.setCurrentDataTable(null);
    parentMeta.setCurrentBookmark(DataTableModel.ROW_NOT_FOUND);
}

/**
 * Passes event on to a single specific listener of this DataModel.
 * @param event a DataModelEvent
 * @param listener the DataModelListener
 * @return true if the event was accepted by the listener, or false
 * if the listener vetoed the change.
 */
public boolean fireDataModelEvent(DataModelEvent event, DataModelListener listener) {
    return sendDataModelEvent(event, listener);
}

/**
 * Passes event on to listeners of this DataModel.
 * @param event a DataModelEvent
 * @return true if the event was accepted by all listeners, or false
 * if at least one listener vetoed the change.
 */
public boolean fireDataModelEvent(DataModelEvent event) {
    return sendDataModelEvent(event, null);
}

/**
 * Sends event on to listeners of this DataModel (internal version)
 * @param event a DataModelEvent
 * @param listener the DataModelListener
 * @return true if the event was accepted by the listener, or false
 * if the listener vetoed the change.
 */
protected boolean sendDataModelEvent(DataModelEvent event, DataModelListener listener) {
    boolean accepted = true;
    if (!eventsEnabled) {
        return accepted;
    }
    Enumeration e;
    if (listener == null) {
        // copy the listeners and send the event
        e = JCListenerList.elements(listeners);
    }
    else {
        Vector v = new Vector(1);

        v.addElement(listener);
        e = v.elements();
    }
    while (e.hasMoreElements()) {
        DataModelListener l = (DataModelListener)e.nextElement();
        fireDataModelEventInternal(event, l);
        // first listener to cancel event aborts the change for all
        if (event != null && event.getCancelled()) {
            accepted = false;
            break;
        }
    }
    return accepted;
}

/**
 * Sends event on to a single specific listener of this DataModel (internal version)
 * @param event a DataModelEvent
 * @param listener the DataModelListener
 * @return true if the event was accepted by the listener, or false
 * if the listener vetoed the change.
 */
protected void fireDataModelEventInternal(DataModelEvent event, DataModelListener listener) {
    switch (event.getCommand()) {
    case DataModelEvent.BEFORE_INSERT_ROW:
        listener.beforeInsertRow(event);
        break;
    case DataModelEvent.BEFORE_RESET:
        listener.beforeReset(event);
        break;
    case DataModelEvent.BEFORE_REQUERY:
        listener.beforeRequery(event);
        break;
    case DataModelEvent.BEFORE_EDIT_CELL:
        listener.beforeEditCell(event);
        break;
    case DataModelEvent.BEFORE_DELETE_ROW:
        listener.beforeDeleteRow(event);
        break;
    case DataModelEvent.BEFORE_COMMIT_ALL:
        listener.beforeCommitAll(event);
        break;
    case DataModelEvent.BEFORE_CANCEL_ALL:
        listener.beforeCancelAll(event);
        break;
    case DataModelEvent.BEFORE_CANCEL_ROW_CHANGES:
        listener.beforeCancelRowChanges(event);
        break;
    case DataModelEvent.BEFORE_COMMIT_CONDITIONAL:
        listener.beforeCommitConditional(event);
        break;
    case DataModelEvent.BEFORE_COMMIT_ROW:
        listener.beforeCommitRow(event);
        break;
    case DataModelEvent.BEFORE_MOVE_TO_CURRENT_ROW:
        listener.beforeMoveToCurrentRow(event);
        break;
    case DataModelEvent.BEFORE_DELETE_TABLE:
        listener.beforeDeleteTable(event);
        break;
    case DataModelEvent.AFTER_CHANGE_OF_ROW_DATA:
        listener.afterChangeOfRowData(event);
        break;
    case DataModelEvent.AFTER_INSERT_ROW:
        listener.afterInsertRow(event);
        break;
    case DataModelEvent.AFTER_DELETE_ROW:
        listener.afterDeleteRow(event);
        break;
    case DataModelEvent.AFTER_RESET:
        listener.afterReset(event);
        break;
    case DataModelEvent.AFTER_MOVE_TO_CURRENT_ROW:
        listener.afterMoveToCurrentRow(event);
        break;
    case DataModelEvent.AFTER_REQUERY_ROW_AND_DETAILS:
        listener.afterRequeryRowAndDetails(event);
        break;
    case DataModelEvent.AFTER_REQUERY_TABLE:
        listener.afterRequeryTable(event);
        break;
    case DataModelEvent.BEGIN_EVENTS:
        listener.beginEvents(event);
        break;
    case DataModelEvent.END_EVENTS:
        listener.endEvents(event);
        break;
    case DataModelEvent.INITIALIZE:
        listener.initialize(event);
        break;
    case DataModelEvent.ORIGINATOR_NAVIGATE_ROW:
        listener.originatorNavigateRow(event);
        break;
//FIX should throw an exception here            
    default:
        break;
    }

}   

/**
 * Enable or disable data model events.  When disabled, events are not
 * dispatched to registered listeners.  This is useful for making mass changes
 * without having listeners repainting on each change.  That is,
 * set enabled to false, make the changes, then set back true.
 * <P>
 * This is different from the DataModelEvent
 * BEGIN_EVENTS and END_EVENTS events where events are still sent
 * but the listener having received the BEGIN_EVENTS event knows it may choose
 * to disregard the events until the END_EVENTS event is received.  They are
 * given the choice.
 * <P>
 * Temporarily disabling events is more efficient than using the BEGIN/END_EVENTS
 * method since the events are not dispatched to all listeners.
 * @param enable a boolean, true if events are to be dispatched to listeners,
 * false otherwise.
 * @see com.klg.jclass.datasource.DataModelEvent
 * @see #getEventsEnabled
 */
public void enableDataModelEvents(boolean value) {
    eventsEnabled = value;
}

/**
 * Return true if events are enabled, false otherwise.
 * @return true if events are enabled, false otherwise.
 * @see #enableDataModelEvents
 */
public boolean getEventsEnabled() {
    return eventsEnabled;
}

/**
 * Set this model's name.
 * This is used to identify this model so it must be unique.
 * @param modelName a String which is to be this model's name
 */
public void setModelName(String modelName) {
    this.modelName = modelName;
}

/**
 * Get this model's name.
 * If no name is set, the default name of an instance of this
 * class will be the concatenation of this class's name
 * and a numeric digit. For example TreeData1.
 * @return a String which is the name of this model.
 */
public String getModelName() {
    return modelName;
}

/**
 * Return the DataTableModel currently in focus.
 * @return the DataTableModel which has the global
 * cursor's focus
 * @see #moveToRow
 * @see #getCurrentGlobalBookmark
 */
public DataTableModel getCurrentGlobalTable() {
    return currentDataTable;
}

/**
 * Return the bookmark of the row which has the
 * global cursor's focus.
 * @return the bookmark of the row on which the
 * global cursor is located
 * @see #moveToRow
 * @see #getCurrentGlobalTable
 */
public long getCurrentGlobalBookmark() {
    return currentBookmark;
}

/**
 * Returns the DataTable at the specified MetaData level based on
 * the current bookmark.  Bookmark refers to what we're calling
 * the "bookmark path" ie the bookmarks for each parent MetaData
 * level down to the specified level.
 * <P>
 * Will create the table if it does not exist.  In this case it
 * will be the child of the current row in the parent, and so on.
 *
 * @param originator the DataModelListener which initiated this action
 * @param metaDataModel the MetaDataModel for which the current
 * DataTableModel should be returned
 * @return the DataTableModel currently in focus for this level
 * @exception DataModelException If get fails.
 */
public DataTableModel getCurrentDataTable(DataModelListener originator, MetaDataModel myLevel) throws DataModelException {
    setOrCreatePathFromRootToMe(originator, myLevel);
    return myLevel.getCurrentDataTable();
}

/**
 * Walk up a tree and create parent data tables such that there will
 * be a current data table at the indicated level.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If path can't be created.
 */
private void setOrCreatePathFromRootToMe(DataModelListener originator, MetaDataModel myLevel) throws DataModelException {
    // get the parent meta data for this level
    MetaDataModel parentMetaData    = (MetaDataModel) myLevel.getParent();
    if (parentMetaData == null) {
        // if it was null we reached the root, set and return
        myLevel.setCurrentDataTable((DataTableModel) getDataTableTree().getRoot());
        return;
    }
    // otherwise, get the current table and bookmark for this level
    DataTableModel myLevelTable = myLevel.getCurrentDataTable();
    long myLevelBookmark        = myLevel.getCurrentBookmark();

    // if the current table was not null, see if it's the right one
    if (myLevelTable != null) {
        if (!ancestorPathPointsToMe(myLevelTable)) {
            // this is not the correct table for this level, so go up the tree
            setOrCreatePathFromRootToMe(originator, (MetaDataModel) myLevel.getParent());
        }
        else {
            // was the correct table and bookmark for this level, so return it
            return;
        }
    }
    else {
        // was null, so we have to go up and create it on the way back down
        setOrCreatePathFromRootToMe(originator, (MetaDataModel) myLevel.getParent());
    }

    // finished going up, go down now
    DataTableModel parentDataTable = parentMetaData.getCurrentDataTable();
    if (parentDataTable == null) {
        // or could be the case that a request was
        // made for a leaf and the intermediate nodes were not created yet,
        // so create them
        setOrCreatePathFromRootToMe(originator, parentMetaData);
        parentDataTable = parentMetaData.getCurrentDataTable(); // now it's set
    }

    DataTableModel current_table = null;
    long current_table_bookmark = DataTableModel.ROW_NOT_FOUND;

    // if parentDataTable is still null at this point,
    // then the intermediate nodes did not exist, otherwise create the current table
    if (parentDataTable != null) {
        current_table = parentDataTable.createTable(originator, parentMetaData.getCurrentBookmark(), (TreeNode) myLevel);
        if (current_table != null) {
            current_table.first();
            if (current_table.getRowCount() > 0) {
                current_table_bookmark = current_table.getCurrentBookmark();
            }
        }
    }
    myLevel.setCurrentDataTable(current_table);
    myLevel.setCurrentBookmark(current_table_bookmark);
}

private boolean ancestorPathPointsToMe(DataTableModel dtm) {
    Vector ancestors = new Vector();
    dtm.getAncestors(ancestors);

    for (int i = 0; i < ancestors.size(); i++) {

        DataTableModel parent 		= (DataTableModel)((TreeNode)dtm).getParent();

		if (parent == null) {
			return false;
		}

        MetaDataModel parentMeta 	= parent.getMetaData();

        if (parentMeta == null || !ancestors.contains(new Long(parentMeta.getCurrentBookmark()))) {
            return false;
        }
        dtm = (DataTableModel) ((TreeNode) dtm).getParent();
    }

    return true;
}

/**
 * Returns the data item at the specified MetaData level and column for
 * the current row.
 * Equivalent to calling DataTableModel.getResultData() with the
 * current bookmark at the MetaData level.
 * <P>
 * Will create the table if it does not exist.  In this case it
 * will be the child of the current row in the parent, and so on.
 *
 * @param originator the DataModelListener which initiated this action
 * @param metaDataModel the MetaDataModel from which the column value
 * is to be retrieved
 * @param column_identifier a String which is the column from which
 * a value is to be retrieved
 * @exception DataModelException If get fails.
 */
public Object getCurrentDataItem(DataModelListener originator, MetaDataModel myLevel, String column) throws
DataModelException {
    DataTableModel dtm = getCurrentDataTable(originator, myLevel);
    long bm = myLevel.getCurrentBookmark();
    if (dtm != null && bm != DataTableModel.ROW_NOT_FOUND &&
        dtm.getRowStatus(bm) != DataTableModel.NONEXISTENT &&
        dtm.getRowStatus(bm) != DataTableModel.DELETED) {
        return dtm.getResultData(bm, column);
    }
    return null;
}

/**
 * Get a MetaDataModel instance by name, or null if not found.
 * @param metaDataName a String which is the name of the MetaDataModel object
 * @return the MetaDataModel for which getDescription() == metaDataName, or
 * null if not found.
 * @see MetaDataModel#setDescription
 * @see MetaDataModel#getDescription
 */
public MetaDataModel getMetaData(String metaDataName) {
    MetaDataModel root = (MetaDataModel) getMetaDataTree().getRoot();
    return getMetaDataByName(root, metaDataName);
}

// recurse and find the meta data instance with this name
private MetaDataModel getMetaDataByName(MetaDataModel root, String name) {
    if (root == null || root.getDescription().equals(name)) {
        return root;
    }
    Enumeration e = root.children();
    MetaDataModel mdm = null;
    while (e.hasMoreElements()) {
        MetaDataModel child = (MetaDataModel) e.nextElement();
        mdm = getMetaDataByName(child, name);
        if (mdm != null) break;
    }
    return mdm;
}

/**
 * Are there any pending modifications in any rows in this entire
 * data model?
 * @return true if there is at least one change pending, false otherwise
 */
public boolean isModified() {
	boolean modified = false;
    TreeModel tree = getDataTableTree();
	if (tree != null) {
    	DataTableModel root = (DataTableModel) tree.getRoot();
		if (root != null) {
    		modified = 
				root.isModified() ? true : root.isTableOrChildrenModified();
		}
	}
	return modified;
}

/**
 * Clears all user data.
 * Clears all Objects which were arbitrarily associated with
 * the entire DataTableTree.
 * @param user an Object, a reference to the user.
 */
public void clearDataTableTreeUserData(Object user) {
    DataTableModel root = (DataTableModel) getDataTableTree().getRoot();
    for (Enumeration e = ((DataSourceTreeNode) root).breadthFirstEnumeration(); e.hasMoreElements();) {
        DataTableModel dataTable = (DataTableModel) e.nextElement();
        dataTable.setUserData(user, null);
    }
}

/**
 * For internal use only.
 * Clears all internal data.
 * Clears all Objects which were arbitrarily associated with
 * the entire DataTableTree.
 * @param user an Object, a reference to the user.
 */
public void clearDataTableTreeInternalData(Object user) {
    DataTableModel root = (DataTableModel) getDataTableTree().getRoot();
    for (Enumeration e = ((DataSourceTreeNode) root).breadthFirstEnumeration(); e.hasMoreElements();) {
        DataTableModel dataTable = (DataTableModel) e.nextElement();
        dataTable.setInternalData(user, null);
    }
}

} 
