/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: MetaDataCreator.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:10 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.beans;

import com.klg.jclass.datasource.*;


import java.awt.Frame;
import java.beans.Beans;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.util.*;
import javax.swing.tree.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.datasource.jdbc.*;
import com.klg.jclass.util.JCEnvironment;

/** This class creates a MetaData.
 *	@see com.klg.jclass.datasource.beans.NodeProperties
 */
public class MetaDataCreator implements Serializable {

public static final long serialVersionUID = 5993967531083794512L;

private final static boolean TRACE = false;

// singletons
private static MetaDataCreator jdbcCreator;
private static MetaDataCreator unboundCreator;

/** Creates a MetaDataModel with no parent.
 *	@param data_model data model encapsulates actual data and fires data model events
 *	@param comp data bean component contains data source specific information
 *	to create the MetaDataModel.<p>
 *	@exception DataModelException If the method fails to create a meta data.
 */
public MetaDataModel createMetaData(DataModel data_model,
                                    NodeProperties comp) throws DataModelException {
    return createMetaData(data_model, null, comp);
}


/** Creates a child MetaDataModel of a parent MetaDataModel.
 *	@param data_model data model encapsulates actual data and fires data model events
 *	@param meta_parent meta data model of parent data source
 *	@param comp data bean component contains data source specific information
 *	@exception DataModelException If the method fails to create a meta data.
 *	@see com.klg.jclass.datasource.jdbc.MetaData#getDesignTimeMaxRows
 */
public MetaDataModel createMetaData(DataModel data_model,
                                    MetaDataModel meta_parent, NodeProperties comp) throws DataModelException {

    MetaDataModel meta = null;

    if (comp == null) return meta;

    meta = create(data_model, meta_parent, comp);

    setMetaDataProperties(meta, comp);

    // if meta id is not defined, set it.
    if (meta != null && comp.getMetaID() == -1) {
        comp.setMetaID(meta.getMetaID());
        comp.fireJCDataEvent(new JCDataEvent((Object)comp));
    }

    return meta;
}

/**
 * Helper method to set common meta data properties.
 *@exception DataModelException If no JDBC connection is established.
 */
private void setMetaDataProperties(MetaDataModel meta,
                                   NodeProperties comp) throws DataModelException {
    if (meta == null) return;

    try {
        meta.setAutoCommit(comp.getAutoCommit());
    }
    catch (Exception e) {
        // Fails if no connection is established yet in JDBC.
        if (!JCEnvironment.isDesignTime()) {
            throw new DataModelException(DataModelEvent.INITIALIZE, e);
        }
    }
    meta.setCommitPolicy(comp.getCommitPolicy());
    meta.setCacheChildren(comp.getCacheChildren());
    meta.setShowDeletedRows(comp.getShowDeletedRows());
    meta.setDescription(comp.getDescription());
    // if meta id is defined, set it.
    if (comp.getMetaID() > -1) {
        meta.setMetaID(comp.getMetaID());
    }

    // set table access.
    for (Enumeration e = comp.getTableAccess().keys(); e.hasMoreElements();) {
        String table = (String) e.nextElement();
        Boolean[] access = (Boolean[]) comp.getTableAccess().get(table);
        meta.setInsertAllowed(table,
                              access[NodeProperties.INSERT_ALLOWED_INDEX].booleanValue());
        meta.setUpdateAllowed(table,
                              access[NodeProperties.UPDATE_ALLOWED_INDEX].booleanValue());
        meta.setDeleteAllowed(table,
                              access[NodeProperties.DELETE_ALLOWED_INDEX].booleanValue());
    }

    // Add virtual columns.
    for (Enumeration e = comp.getVirtualColumns().elements(); e.hasMoreElements();) {
        meta.addColumn((VirtualColumnModel) e.nextElement());
    }
}

/** Gets a meta data creator. */
public static MetaDataCreator getCreator(int type) {
	// create singleton instances
	switch(type) {
		case NodeProperties.JDBC:
			if (jdbcCreator == null) {
				jdbcCreator = new com.klg.jclass.datasource.jdbc.MetaDataCreator();
			}
			return jdbcCreator;
		case NodeProperties.UNBOUND:
		default:
			if (unboundCreator == null) {
				unboundCreator = new MetaDataCreator();
			}
			return unboundCreator;
	}
}

/** Overrides to create a meta data for different data source type.
 * @exception DataModelException If a MetaDataModel can't be created.
 */
protected MetaDataModel create(DataModel data_model, MetaDataModel meta_parent, NodeProperties comp)
	throws DataModelException {

	// create the meta data for this level
	BaseMetaData metaData = new BaseMetaData(data_model);

	// name it
	metaData.setDescription(comp.getDescription());

	// add columns
	try {
		Vector v = comp.getUnboundColumns();
		int numCols = v.size();
		for (int i = 0; i < numCols; i++) {
			ColumnModel c = (ColumnModel) v.elementAt(i);
			System.out.println("Adding column " + c.getColumnName() + " to " + metaData.getDescription());
			metaData.addColumn(c);
		}
	}
	catch (DataModelException dme) {
		dme.printStackTrace();
	}

	// this is the root
	if (meta_parent == null) {
		// make a tree out of them
		((DataSourceTreeModel)data_model.getMetaDataTree()).setRoot((TreeNode) metaData);
		//add an empty root for now
		((DataSourceTreeModel)data_model.getDataTableTree()).setRoot(new BaseDataTable(metaData));
	}
	else {
		((DataSourceTreeNode)meta_parent).add(metaData);
	}

	// Finally, delegate data creation to separate class
	// which implements the ProviderModel interface. For
	// simplicity use the same class for each level.
	metaData.setProviderClassName(comp.getProviderName());


    return metaData;
}

}
