/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: NodeProperties.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:10 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.beans;

import java.beans.Beans;
import java.util.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.datasource.jdbc.DataTableConnection;
import com.klg.jclass.util.JCEnvironment;

/**
 *	This class encapsulates information related to data source.
 *	For IDE data sources, variable name of the data source is stored.
 */
public class NodeProperties extends SerializedProperties {

static final long serialVersionUID = 3857956278383656657L;

/** JBuilder data source type */
public static final int JBUILDER = 0;
/** JDBC data source type */
public static final int JDBC = 1;
/** PowerJ data source type */
public static final int POWERJ = 2;
/** VisualCafe data source type */
public static final int VCAFE = 3;
/** Array data source type */
public static final int ARRAY = 4;
/** Unbound data source type */
public static final int UNBOUND = 5;
/** Unknown data source type */
public static final int UNKNOWN = -999;


/** index of INSERT_ALLOWED */
public static final int INSERT_ALLOWED_INDEX = 0;
/** index of UPDATE_ALLOWED */
public static final int UPDATE_ALLOWED_INDEX = 1;
/** index of DELETE_ALLOWED */
public static final int DELETE_ALLOWED_INDEX = 2;

// provider specific

/** If delegation is used, this will be the name of the class providing data */
protected String providerName;
protected Vector unboundColumns = new Vector();


/** ID for a meta data created from this NodeProperties. */
private int metaID = -1;

/** data model name used in data-binding editor. */
private String modelName;

/** variable name of IDE data source. */
private String  dataSourceName;

/** Actual specific IDE data source object. Some are not serializable.
 *	They may be in the future release.
 */
private transient Object dataSource;

/** VCAFE specific.  It is the full name of DataBinding. */
private String vcFullName;
/** VCAFE specific.  It is the class name of QueryNavigator. */
private String vcClassName;


/** User-assigned description. */
private String  description;

/** Data source type.  It must be one of the following:
 *	<p><ul>
 *	<li>JDBC (default)</li>
 *	<li>JBUILDER</li>
 *	<li>POWERJ</li>
 *	<li>VCAFE</li>
 *	<li>ARRAY</li>
 *	<li>UNBOUND</li>
 *	</ul><p>
 */
private int type;

/** Commits the policy property.
 *	@see com.klg.jclass.datasource.beans.JCData#setCommitPolicy
 */
private int commitPolicy = MetaDataModel.COMMIT_LEAVING_RECORD;
/** Auto commit a transaction. */
private boolean autoCommit = true;

/** Controls data access to a data table in a data source. */
private Hashtable tableAccess;

/** A list of virtual columns. */
private Vector virtualColumns;

/** Indicate if logically deleted rows showld be shown.  */
private boolean showDeletedRows = true;

/** Indicate if children should be cached.  */
private boolean cacheChildren = false;


//
//	Properties when used in TreeProperties.
//
/** Use parent JDBC connection. */
private boolean useParentConnection = false;
/** parent NodeProperties. */
private NodeProperties parentComp;
/** columns of current data source to join on parent. */
private Vector joinColumns;
/** columns of parent data source to join on current data source. */
private Vector parentJoinColumns;


//
//	JDBC specific fields
//
/** JDBC driver. */
private String  driver;
/** JDBC connection URL. */
private String      url;
/** User to establish connection. */
private String      user;
/** Password of the user. */
private String      password;
/** Database initially connected to. */
private String      database;
/** SQL Statement for JDBC data source. */
private String  sqlStatement;
/** SQL Statement is set using expert mode. */
private boolean expertMode = false;
/** Driver table in a multi-table select statement. */
private String  driverTable;
/** Primary key columns of driver table. */
private Vector driverTablePrimaryKeys;
/** Indicates if user is prompted for login. */
private boolean promptUserForLogin = false;

/** DataTableConnection created by the connection information. */
private transient DataTableConnection connection;

/** Serialized version of the a bean component currently being deserialized. */
private static transient String serializedVersion = new String("");

private static final String base = "jcdbcomp";
private static int counter = 0;
private static final boolean TRACE = false;


/** Constructs a NodeProperties.
 */
public NodeProperties() {
    this(JDBC, null);
}

/** Constructs a NodeProperties.
 *	@param type data source type
 *	@param data_source_name variable name of data source for the type
 */
public NodeProperties(int type, String data_source_name) {
    this(type, data_source_name, null, null);
}

/** Constructs a NodeProperties.
 *	@param type data source type
 *	@param data_source_name variable name of data source for the type
 *	@param description description of the data bean component.
 */
public NodeProperties(int type, String data_source_name, String description, String model_name) {
    this.type = type;
    this.dataSourceName = data_source_name;
    this.description = description;
    this.modelName = model_name;
    init();
}

/**
 * Constructs a NodeProperties from a resource relative to a class loader.
 * @exception DataModelException If the component can't be instantiated.
 */
public NodeProperties(String resource_name, ClassLoader cl)
throws DataModelException {
    this(resource_name, cl, null);
}

/** Constructs a NodeProperties from a resource relative to a class loader.
 * @exception DataModelException If the component can't be instantiated.
 */
public NodeProperties(String resource_name, ClassLoader cl, Object data_source)
throws DataModelException {
    try {
        if (resource_name != null) {
            NodeProperties comp = (NodeProperties)
                                  NodeProperties.fromResource(cl, resource_name);
            // copy properties
            comp.setDataSource(data_source);
            copyProperties(comp);
        }
        else {
            init();
        }
    }
    catch (Exception ex) {
        throw new DataModelException(DataModelEvent.INITIALIZE, ex);
    }
}

/** Constructs a NodeProperties from a resource. First,
 *	the package name of the caller class is prepended to the resource
 *	name to locate the resource.  If no resource is found, the passed-in
 *	resource name is used.
 *	@param resource_name resource name relative to a class loader
 *						to the bean component. It must have the right extension.
 *	@param caller_class class that uses the serialized bean component.
 *	@param sfile serialization file that is used at design-time in JBuilder to
 *			load a NodeProperties.
 *	@see java.lang.ClassLoader
 *      @exception DataModelException If the class can't be loaded.
 */
public NodeProperties(String resource_name, Class caller_class, String sfile)
throws DataModelException {
    this(resource_name, caller_class, sfile, null);
}

/** Constructs a NodeProperties from a resource. First,
 *	the package name of the caller class is prepended to the resource
 *	name to locate the resource.  If no resource is found, the passed-in
 *	resource name is used.
 *	@param resource_name resource name relative to a class loader
 *						to the bean component. It must have the right extension.
 *	@param caller_class class that uses the serialized bean component.
 *	@param sfile serialization file that is used at design-time in JBuilder to
 *			load a NodeProperties.
 *	@see java.lang.ClassLoader
 *      @exception DataModelException If the class can't be loaded.
 */
public NodeProperties(String resource_name, Class caller_class, String sfile, Object data_source)
throws DataModelException {
    try {
//FIXOBSOLETE
//		if (JCEnvironment.isJBuilder() && JCEnvironment.isDesignTime()) {
        if (false) {
            NodeProperties comp = (NodeProperties)
                                  NodeProperties.fromFile(sfile);
            // copy properties
            comp.setDataSource(data_source);
            copyProperties(comp);
        }
        else if (resource_name != null) {
            NodeProperties comp = (NodeProperties)
                                  NodeProperties.fromResource(caller_class, resource_name);
            // copy properties
            comp.setDataSource(data_source);
            copyProperties(comp);
        }
        else {
            init();
        }
    }
    catch (Exception ex) {
        throw new DataModelException(DataModelEvent.INITIALIZE, ex);
    }
}

/** Constructs a NodeProperties from a resource relative to a class loader
 *	of a caller class.
 * @exception DataModelException If the component can't be instantiated.
 */
public NodeProperties(String resource_name, Class caller_class)
throws DataModelException {
    this(resource_name, caller_class, null);
}

/** Constructs a NodeProperties from a resource relative to a class loader
 *	of a caller class.
 * @exception DataModelException If the component can't be instantiated.
 */
public NodeProperties(String resource_name, Class caller_class, Object data_source)
throws DataModelException {
    try {
        if (resource_name != null) {
            NodeProperties comp = (NodeProperties)
                                  NodeProperties.fromResource(caller_class, resource_name);
            // copy properties
            comp.setDataSource(data_source);
            copyProperties(comp);
        }
        else {
            init();
        }
    }
    catch (Exception ex) {
        throw new DataModelException(DataModelEvent.INITIALIZE, ex);
    }
}

private void copyProperties(NodeProperties comp) {
    type = comp.getDataSourceType();
    dataSource = comp.getDataSourceName();
    dataSourceName = comp.getDataSourceName();
    vcFullName = comp.getVCFullName();
    vcClassName = comp.getVCClassName();
    description = comp.getDescription();
    modelName = comp.getModelName();
    virtualColumns = comp.getVirtualColumns();
    tableAccess = comp.getTableAccess();
    if (comp.getJoins() != null) {
        for (int i=0; i<comp.getJoins().length; i++) {
            addJoin(comp.getJoins()[i][0], comp.getJoins()[i][1]);
        }
    }
    metaID = comp.getMetaID();
    autoCommit = comp.getAutoCommit();
    showDeletedRows = comp.getShowDeletedRows();
    cacheChildren = comp.getCacheChildren();
    useParentConnection = comp.getUseParentConnection();
    parentComp = comp.getParentNodeProperties();
    driver = comp.getDriver();
    url = comp.getURL();
    user = comp.getUser();
    password = comp.getPassword();
    database = comp.getDatabase();
    sqlStatement = comp.getStatement();
    promptUserForLogin = comp.getPromptUserForLogin();
    driverTable = comp.getDriverTable();
    driverTablePrimaryKeys = comp.getDriverTablePrimaryKeys();
    serializationFile = comp.getSerializationFile();
    resourceName = comp.getResourceName();
    providerName = comp.getProviderName();
}

private void init() {
    virtualColumns = new Vector();
    tableAccess = new Hashtable();
    joinColumns = new Vector();
    parentJoinColumns = new Vector();
    driverTablePrimaryKeys = new Vector();
    setDefaultSerializationFileBase(base);
    serializationFile = getDefaultSerializationFile(base, counter);
    // update counter
    try {
        String s = serializationFile.substring(serializationFile.indexOf(base)+base.length());
        s = s.substring(0, s.indexOf(LocaleBundle.serialization_extension));
        counter = Integer.parseInt(s);
    }
    catch (Exception ex) {
    }
}

/** Sets the model name. */
public void setModelName(String s) {
    if (isSameString(modelName, s)) return;

    this.modelName = s;
    setChanged();
}

/** Sets the model name. */
public String getModelName() {
    return modelName;
}


/**
 * Returns an int which uniquely identifies this instance
 * of the MetaDataModel.
 * @return an int which uniquely identifies a meta data model.
 * @see #setMetaID
 */
public int getMetaID() {
    return metaID;
}

/**
 * 	Assigns a unique id to a meta data instance.
 *	Default is -1.  Unless this meta id is set, meta data created
 *	will use its auto-generated number.
 *	an auto-generated number is used.
 * 	For internal use only.
 * 	@param id an int which is the unique id for this instance
 * 	@see #getMetaID
 */
public void setMetaID(int id) {
    if (metaID == id) return;

    metaID = id;
    setChanged();
}

/** Gets the description of data bean component. */
public String getDescription() {
    return description;
}

/** Sets the description of data bean component. */
public void setDescription(String description) {
    if (isSameString(this.description, description)) return;

    this.description = description;
    setChanged();
}

/** Sets data source type.  It must be one of the following:<p>
 *	<p><ul>
 *	<li>JDBC (default)</li>
 *	<li>JBUILDER</li>
 *	<li>POWERJ</li>
 *	<li>VCAFE</li>
 *	<li>ARRAY</li>
 *	<li>UNBOUND</li>
 *	</ul><p>
 */
public void setDataSourceType(int type) {
    if (this.type == type) return;

    this.type = type;
    setChanged();
}
/** Gets the data source type. */
public int getDataSourceType() {
    return type;
}


/** Sets the Visual Cafe QueryNavigator class name.
 *	It is used at design-time for JClass data-binding.
 */
public void setVCClassName(String s) {
    if (isSameString(vcClassName, s)) return;

    this.vcClassName = s;
    setChanged();
}
/** Gets the Visual Cafe QueryNavigator class name.
 *	It is used at design-time for JClass data-binding.
 */
public String getVCClassName() {
    return vcClassName;
}

/** Sets the Visual Cafe data binding full name.
 *	It is used at design-time for JClass data-binding.
 */
public void setVCFullName(String s) {
    if (isSameString(vcFullName, s)) return;

    this.vcFullName = s;
    setChanged();
}
/** Gets the Visual Cafe data binding full name.
 *	It is used at design-time for JClass data-binding.
 */
public String getVCFullName() {
    return vcFullName;
}

/** Sets the data source name. */
public void setDataSourceName(String s) {
    if (isSameString(dataSourceName, s)) return;

    this.dataSourceName = s;
    setChanged();
}
/** Gets the data source name. */
public String getDataSourceName() {
    return dataSourceName;
}


/** Sets the data source. */
public void setDataSource(Object o) {
    this.dataSource = o;
    setChanged();
}
/** Gets the data source. */
public Object getDataSource() {
    return dataSource;
}

/** Gets the commitPolicy property.
 *	@see com.klg.jclass.datasource.BaseMetaData#setCommitPolicy
 */
public int getCommitPolicy() {
    return commitPolicy;
}
/**
 *	Sets the commit policy for this level.  Determines when
 *	changes are permanently written to the originating data source.
 *	Must be one of COMMIT_LEAVING_RECORD, COMMIT_LEAVING_PARENT,
 *	or COMMIT_MANUALLY.
 *	@param commit_policy an int, one of MetaDataModel.COMMIT_LEAVING_RECORD,
 *	 MetaDataModel.COMMIT_LEAVING_PARENT, or MetaDataModel.COMMIT_MANUALLY
 */
public void setCommitPolicy(int commit_policy) {
    if (commitPolicy == commit_policy) return;

    JCDataConverter.checkCommitPolicy(commit_policy);
    this.commitPolicy = commit_policy;
    setChanged();
}

/** Gets the autoCommit property.
 *	@see #setAutoCommit
 */
public boolean getAutoCommit() {
    return autoCommit;
}
/** Sets the autoCommit property.
 *	@see com.klg.jclass.datasource.MetaDataModel#setAutoCommit
 */
public void setAutoCommit(boolean auto_commit) {
    if (autoCommit == auto_commit) return;

    this.autoCommit = auto_commit;
    setChanged();
}


/** Gets the showDeletedRows property.
 *	@see #setShowDeletedRows
 */
public boolean getShowDeletedRows() {
    return showDeletedRows;
}
/** Sets the showDeletedRows property.
 *	@see com.klg.jclass.datasource.MetaDataModel#setShowDeletedRows
 */
public void setShowDeletedRows(boolean v) {
    if (showDeletedRows == v) return;

    this.showDeletedRows = v;
    setChanged();
}

/** Gets the cacheChildren property.
 *	@see #setCacheChildren
 */
public boolean getCacheChildren() {
    return cacheChildren;
}
/** Sets the cacheChildren property.
 *	@see com.klg.jclass.datasource.MetaDataModel#setCacheChildren
 */
public void setCacheChildren(boolean v) {
    if (cacheChildren == v) return;

    this.cacheChildren = v;
    setChanged();
}

/** The data source of the data component is JBuilder, Visual Cafe or PowerJ. */
public static boolean isIDEDataSourceType(int type) {
    return(type == JBUILDER || type == VCAFE || type == POWERJ);
}

/** Sets the data access type for all the tables in the data source.
 *	@param alias alias or qualified name of a table
 */
public void setTableAccess(String alias, Boolean[] access) {
    if (access.length != 3) {
        throw new IllegalArgumentException();
    }

    if (tableAccess == null) tableAccess = new Hashtable();

    boolean changed = false;

    if (tableAccess.containsKey(alias)) {
        Boolean[] old = (Boolean[]) tableAccess.get(alias);
        if (!(old[INSERT_ALLOWED_INDEX].booleanValue() == access[INSERT_ALLOWED_INDEX].booleanValue() &&
              old[UPDATE_ALLOWED_INDEX].booleanValue() == access[UPDATE_ALLOWED_INDEX].booleanValue() &&
              old[DELETE_ALLOWED_INDEX].booleanValue() == access[DELETE_ALLOWED_INDEX].booleanValue())) {
            changed = true;
            tableAccess.put(alias, access);
        }
    }
    else {
        changed = true;
        tableAccess.put(alias, access);
    }

    if (changed) setChanged();
}

/** Gets the data access of all the tables in the data source.
 */
public Hashtable getTableAccess() {
    if (tableAccess == null) tableAccess = new Hashtable();
    return tableAccess;
}

/** Removes the data access of the table in the data source.
 *	@param alias alias or qualified name of a table
 */
public void removeTableAccess(String alias) {
    if (tableAccess == null || !tableAccess.containsKey(alias)) return;
    tableAccess.remove(alias);
    setChanged();
}

/** Adds a virtual column.
 *	@param vc virtual column to be added.
 *	@see com.klg.jclass.datasource.BaseVirtualColumn
 */
public void addVirtualColumn(BaseVirtualColumn vc) {
    virtualColumns.addElement(vc);
    setChanged();
}

/** Removes a virtual column.
 *	@param vc virtual column to be removed.
 *	@see com.klg.jclass.datasource.BaseVirtualColumn
 */
public void removeVirtualColumn(BaseVirtualColumn vc) {
    virtualColumns.removeElement(vc);
    setChanged();
}

/** Removes all virtual columns.
 *	@see com.klg.jclass.datasource.BaseVirtualColumn
 */
public void removeAllVirtualColumns() {
    virtualColumns.removeAllElements();
    setChanged();
}

/** Gets all virtual columns.
 */
public Vector getVirtualColumns() {
    return virtualColumns;
}

/** Gets a virtual column by column name.
 */
public BaseVirtualColumn getVirtualColumn(String column_name) {
    for (Enumeration e=virtualColumns.elements(); e.hasMoreElements();) {
        BaseVirtualColumn col = (BaseVirtualColumn) e.nextElement();
        if (col.getColumnName().equals(column_name)) {
            return col;
        }
    }
    return null;
}


//////////////////////////////////////////////////////
// Provider specific
//////////////////////////////////////////////////////

public void setProviderName(String providerName) {
	this.providerName = providerName;
    setChanged();
}

public String getProviderName() {
	return providerName;
}

public Vector getUnboundColumns() {
	return unboundColumns;
}

public void addUnboundColumn(ColumnModel c) {
	unboundColumns.addElement(c);
}

public void removeUnboundColumn(ColumnModel c) {
	unboundColumns.removeElement(c);
}


//////////////////////////////////////////////////////////////////////////
//
//	JDBC specific methods
//
//////////////////////////////////////////////////////////////////////////

/** Set if user is prompted for login. */
public void setPromptUserForLogin(boolean v) {
    if (promptUserForLogin == v) return;

    promptUserForLogin = v;
    setChanged();
}

/** Get boolean if user is prompted for login. */
public boolean getPromptUserForLogin() {
    return promptUserForLogin;
}

/**
 * Sets the JDBC information.
 * @exception Exception If setJDBCInfo fails.
 */
public void setJDBCInfo(String driver, String url,
                        String user, String password, String database) throws Exception {
    if (type != JDBC) {
        throw new Exception(LocaleBundle.string(LocaleBundle.cannot_set_jdbc_connection));
    }

    if (isSameString(this.driver, driver) &&
        isSameString(this.url, url) &&
        isSameString(this.user, user) &&
        isSameString(this.password, password) &&
        isSameString(this.database, database)) {
        return;
    }

    this.driver = driver;
    this.url = url;
    this.user = user;
    this.password = password;
    this.database = database;

    this.connection = null; // Reset JDBC connection

    setChanged();
}

/** Sets the driver.
 * @exception Exception If setDriver fails.
 */
public void setDriver(String s) throws Exception {
    setJDBCInfo(s, url, user, password, database);
}
/** Gets the driver. */
public String getDriver() {
    return driver;
}

/** Sets the connection url.
 * @exception Exception If setURL fails.
 */
public void setURL(String s) throws Exception {
    setJDBCInfo(driver, s, user, password, database);
}
/** Gets the connection url. */
public String getURL() {
    return url;
}

/** Sets the username parameter.
 * @exception Exception If setUser fails.
 */
public void setUser(String s) throws Exception {
    setJDBCInfo(driver, url, s, password, database);
}
/** Gets the username parameter. */
public String getUser() {
    return user;
}

/** Sets the password for the database.
 * @exception Exception If setPassword fails.
 */
public void setPassword(String s) throws Exception {
    setJDBCInfo(driver, url, user, s, database);
}
/** Gets the password for the database. */
public String getPassword() {
    return password;
}

/** Sets the database.
 * @exception Exception If setDatabase fails.
 */
public void setDatabase(String s) throws Exception {
    setJDBCInfo(driver, url, user, password, s);
}
/** Gets the database. */
public String getDatabase() {
    return database;
}

/** Sets the SQL statement.
 * @exception Exception If setStatement fails.
 */
public void setStatement(String s) throws Exception {
    if (type != JDBC) {
        throw new Exception(LocaleBundle.string(LocaleBundle.cannot_set_jdbc_connection));
    }

    s = stripWhitespace(s);
    if (isSameString(sqlStatement, s)) {
        return;
    }

    this.sqlStatement = s;

    this.connection = null; // Reset JDBC connection

    setChanged();
}

// strip all newlines, tabs, etc.
static String stripWhitespace(String text) {
    if (text == null) {
        return text;
    }
    char[] bytes = text.toCharArray();
    for (int i = 0; i < bytes.length; i++) {
        if (bytes[i] < ' ') {
            bytes[i] = ' ';
        }
    }
    String s = new String(bytes);
    if (s != null) {
        s = s.trim();
    }
    return s;
}

/** Gets the SQL statement. */
public String getStatement() {
    return sqlStatement;
}

/** Indicates if expert mode is used.
 *	No parsing is done in expert mode.
 */
public boolean isExpertMode() {
    return expertMode;
}

/** Sets to use expert mode or not.
 *	No parsing is done in expert mode.
 */
public void setExpertMode(boolean v) {
    if (expertMode == v) return;

    expertMode = v;
    setChanged();
}

/** Sets the driver table in a select statement.
 * @exception Exception If setDriverTable fails.
 */
public void setDriverTable(String s) throws Exception {
    if (type != JDBC) {
        throw new Exception(LocaleBundle.string(LocaleBundle.cannot_set_jdbc_connection));
    }
    if (isSameString(driverTable, s)) return;

    this.driverTable = s;

    setChanged();
}
/** Gets the driver table in a select statement. */
public String getDriverTable() {
    return driverTable;
}

/** Sets the primary key columns of the driver table.
 * @exception Exception If setDriverTablePrimaryKeys fails.
 */
public void setDriverTablePrimaryKeys(Vector v) throws Exception {
    if (type != JDBC) {
        throw new Exception(LocaleBundle.string(LocaleBundle.cannot_set_jdbc_connection));
    }
    this.driverTablePrimaryKeys = v;

    setChanged();
}
/** Gets the primary key columns of the driver table. */
public Vector getDriverTablePrimaryKeys() {
    return driverTablePrimaryKeys;
}

/** Gets the useParentConnection property.
 *	@see #setUseParentConnection
 */
public boolean getUseParentConnection() {
    return useParentConnection;
}
/**
 * Sets a boolean indicating if this data source uses the
 * parent JDBC connection.
 * The default is to use the parent connection.
 * @param v a boolean indicating if parent connection is used.
 */
public void setUseParentConnection(boolean v) {
    if (useParentConnection == v) return;

    this.useParentConnection = v;
    this.connection = null;
    setChanged();
}

/** Gets DataTableConnection and establishs a JDBC connection.
 * @exception DataModelException If the connection fails.
 */
public DataTableConnection getConnection() throws DataModelException {
    if (connection != null) return connection;

    if (getUseParentConnection()) {
        connection = getParentNodeProperties().getConnection();
    }
    else {
        // If we are at design time and URL is empty,
        // don't bother creating connection to avoid exception.
        if (JCEnvironment.isDesignTime()) {
            if (getURL() != null && getURL().trim().length() != 0) {
                connection = com.klg.jclass.datasource.jdbc.MetaDataCreator.getJdbcConnection(null, this);
            }
        }
        else {
            connection = new DataTableConnection();
            if (getDriver() != null && getDriver().trim().length() != 0) {
                connection.loadDriver(getDriver().trim());
            }
            connection.connect(getURL(), getUser(), getPassword(), getDatabase());
        }
    }
    return connection;
}

/**
 * Resets the connection to null.
 */
public void resetConnection() {
    this.connection = null;
    if (JCEnvironment.isDesignTime()) {
        if (getURL() != null && getURL().trim().length() != 0) {
            com.klg.jclass.datasource.jdbc.MetaDataCreator.resetJdbcConnection(this);
        }
    }
}

/** Gets the parent data bean component property.
 *	@see #setParentNodeProperties
 */
public NodeProperties getParentNodeProperties() {
    return parentComp;
}
/**
 * Sets the parent data bean component property.
 * @param parent parent of this data bean component.
 */
public void setParentNodeProperties(NodeProperties parent) {
    this.parentComp = parent;
    setChanged();
}

/** Adds a join column pair.
 *	@param parent_column column name of parent data source
 *	@param column column name of current data source
 */
public synchronized void addJoin(String parent_column, String column) {
    parentJoinColumns.addElement(parent_column);
    joinColumns.addElement(column);
    setChanged();
}

/** Removes a join column pair.
 *	@param parent_column column name of parent data source
 *	@param column column name of current data source
 */
public synchronized void removeJoin(String parent_column, String column) {
    for (int i=0; i<joinColumns.size(); i++) {
        String current_parent_column = (String) parentJoinColumns.elementAt(i);
        String current_column = (String) joinColumns.elementAt(i);
        if (current_parent_column.equals(parent_column) &&
            current_column.equals(column)) {
            parentJoinColumns.removeElement(parent_column);
            joinColumns.removeElement(column);
            break;
        }
    }
    setChanged();
}

/** Removes all the join columns. */
public synchronized void removeAllJoins() {
    parentJoinColumns.removeAllElements();
    joinColumns.removeAllElements();
    setChanged();
}

/** Gets all the join columns.
 *	@return an array of two columns.
 *	First column is column name of parent data source and
 *	second column is column name of data source
 */
public synchronized String[][] getJoins() {
	if (joinColumns != null) {

	    String[][] s = new String[joinColumns.size()][2];
	    for (int i=0; i<joinColumns.size(); i++) {
	        s[i] = new String[] { (String) parentJoinColumns.elementAt(i),
	            (String) joinColumns.elementAt(i)};
	    }
	    return s;
	}
	return null;
}

/** Gets the current version.  */
public static String getVersion() {
    return 	JCVersion.getVersionNumber();
}

/** Gets the serialized version of the bean component currently being de-serialized.  */
public static synchronized String getSerializedVersion() {
    return serializedVersion;
}



//////////////////////////////////////////////////////////////////////////

/** Compares all properties of NodeProperties to check equality.
 *	It is done in this way because of serialization.  After serialization,
 *	a NodeProperties is stored in different memory space than
 *	the NodeProperties before serialization.
 */
public boolean equals(Object o) {
    if (!(o instanceof NodeProperties)) return false;

    NodeProperties comp = (NodeProperties) o;
    boolean v = true;

    if (this.tableAccess != null && comp.tableAccess != null) {
        v = elementsEqual(this.tableAccess.elements(), comp.tableAccess.elements());
    }
    else {
        v = this.tableAccess == null && comp.tableAccess == null;
    }
    if (!v) return v;

    if (this.virtualColumns != null && comp.virtualColumns != null) {
        v = baseVirtualColumnsEqual(this.virtualColumns.elements(),
                                    comp.virtualColumns.elements());
    }
    else {
        v = this.virtualColumns == null && comp.virtualColumns == null;
    }
    if (!v) return v;

    if (this.joinColumns != null && comp.joinColumns != null) {
        v = elementsEqual(this.joinColumns.elements(), comp.joinColumns.elements());
    }
    else {
        v = this.joinColumns == null && comp.joinColumns == null;
    }
    if (!v) return v;

    if (this.parentJoinColumns != null && comp.parentJoinColumns != null) {
        v = elementsEqual(this.parentJoinColumns.elements(),
                          comp.parentJoinColumns.elements());
    }
    else {
        v = this.parentJoinColumns == null &&
            comp.parentJoinColumns == null;
    }
    if (!v) return v;

    if (this.driverTablePrimaryKeys != null && comp.driverTablePrimaryKeys != null) {
        v = elementsEqual(this.driverTablePrimaryKeys.elements(),
                          comp.driverTablePrimaryKeys.elements());
    }
    else {
        v = this.driverTablePrimaryKeys == null &&
            comp.driverTablePrimaryKeys == null;
    }
    if (!v) return v;

    return 	isSameString(this.dataSourceName, comp.dataSourceName) &&
    isSameString(this.resourceName, comp.resourceName) &&
    isSameString(this.serializationFile, comp.serializationFile) &&
    isSameString(this.description, comp.description) &&
    isSameString(this.modelName, comp.modelName) &&
    this.type == comp.type &&
    this.commitPolicy == comp.commitPolicy &&
    this.autoCommit == comp.autoCommit &&
    this.useParentConnection == comp.useParentConnection &&
    ((this.parentComp == null && comp.parentComp == null) ||
     (this.parentComp != null && comp.parentComp != null &&
      this.parentComp.equals(comp.parentComp))) &&
    // JDBC
    isSameString(this.driver, comp.driver) &&
    isSameString(this.url, comp.url) &&
    isSameString(this.user, comp.user) &&
    isSameString(this.password, comp.password) &&
    isSameString(this.database, comp.database) &&
    isSameString(this.sqlStatement, comp.sqlStatement) &&
    isSameString(this.driverTable, comp.driverTable) &&
    this.promptUserForLogin == comp.promptUserForLogin;
}

public String toString() {
    String str = super.toString();
    if (TRACE) {
        StringBuffer s = new StringBuffer();
        s.append("\nDescription: " + description);
        s.append("\nModel Name: " + modelName);
        s.append("\nMeta ID: " + metaID);
        s.append("\nSerialization file: " + serializationFile);
        s.append("\nResource Name: " + resourceName);
        s.append("\nType: ");
        if (type == JDBC) s.append("JDBC");
        else if (type == JBUILDER) s.append("JBUILDER");
        else if (type == POWERJ) s.append("POWERJ");
        else if (type == VCAFE) s.append("VCAFE");
        else if (type == ARRAY) s.append("ARRAY");
        else if (type == UNBOUND) s.append("UNBOUND");

        if (type == JDBC) {
            s.append("\n\tuse parent connection: " + useParentConnection);
            if (useParentConnection) {
                s.append("\n\tparent description: " + parentComp.getDescription());
            }
            else {
                s.append("\n\tdriver: " + driver);
                s.append("\n\turl: " + url);
                s.append("\n\tuser: " + user);
                s.append("\n\tdatabase: " + database);
            }
            s.append("\n\tSQL Statement:\n " + sqlStatement);
            s.append("\n\tJoin Columns:\tparent = child \n\t");
            for (int i=0; i<joinColumns.size(); i++) {
                s.append("\n\t\t" + parentJoinColumns.elementAt(i));
                s.append(" = " + joinColumns.elementAt(i));
            }
            s.append("\n\tDriver table: " + driverTable);
            if (driverTablePrimaryKeys != null) {
                s.append("\n\tDriver table primary keys:\n");
                for (int i=0; i<driverTablePrimaryKeys.size(); i++) {
                    if (driverTablePrimaryKeys.elementAt(i) != null) {
                        s.append("\n\t\t" + driverTablePrimaryKeys.elementAt(i));
                    }
                }
            }
        }
        else if (isIDEDataSourceType(type)) {
            s.append("\n\tdata source name: " + dataSourceName);
            s.append("\n\tdata source: " + dataSource);
        } /*else if (type == ARRAY) {
            s.append("\n\tColumns:\t\n\t");
            for (int i=0; i<arrayColumns.size(); i++) {
                s.append("\n\t\t" + arrayColumns.elementAt(i));
                s.append("\t" + arrayColumnTypes.elementAt(i));
            }
        } */

        s.append("\n\tAuto commit: " + autoCommit);
        s.append("\n\tCommit policy: " + commitPolicy);
        if (virtualColumns != null) {
            s.append("\n\tvirtualColumns:\n");
            for (int i=0; i<virtualColumns.size(); i++) {
                if (virtualColumns.elementAt(i) != null) {
                    s.append("\n\t\t" + virtualColumns.elementAt(i));
                }
            }
        }
        if (tableAccess.keys() != null) {
            s.append("\n\tTable Access:");
            s.append("\n\t\tTable\tInsert Allowed\tUpdate Allowed\tDelete Allowed");
            for (Enumeration e = tableAccess.keys(); e.hasMoreElements();) {
                String table = (String) e.nextElement();
                Boolean[] access = (Boolean[]) tableAccess.get(table);
                s.append("\n\t\t" + table);
                s.append("\t" + access[INSERT_ALLOWED_INDEX].booleanValue());
                s.append("\t" + access[NodeProperties.UPDATE_ALLOWED_INDEX].booleanValue());
                s.append("\t" + access[NodeProperties.DELETE_ALLOWED_INDEX].booleanValue());
            }
        }
        str = str + new String(s);
    }
    return str;
}

private boolean elementsEqual(Enumeration e1, Enumeration e2) {
    boolean v = true;
    for (; e1.hasMoreElements() && e2.hasMoreElements();) {
        if (!e1.nextElement().equals(e2.nextElement())) {
            v = false;
            break;
        }
    }
    if (v) {
        // if either enumeration still has more elements,
        // it means numbers of elements mismatch.
        // Thus, they are not equal.
        v = !(e1.hasMoreElements() || e2.hasMoreElements());
    }
    return v;
}

private boolean baseVirtualColumnsEqual(Enumeration e1, Enumeration e2) {
    boolean v = true;
    for (; e1.hasMoreElements() && e2.hasMoreElements();) {
        BaseVirtualColumn col1 = (BaseVirtualColumn) e1.nextElement();
        BaseVirtualColumn col2 = (BaseVirtualColumn) e2.nextElement();
        if (!isSameBaseVirtualColumn(col1, col2)) {
            v = false;
            break;
        }
    }
    if (v) {
        // if either enumeration still has more elements,
        // it means numbers of elements mismatch.
        // Thus, they are not equal.
        v = !(e1.hasMoreElements() || e2.hasMoreElements());
    }
    return v;
}

public void setChanged() {
    if (TRACE) {
        Thread.currentThread().dumpStack();
    }
    super.setChanged();
}

private boolean isSameBaseVirtualColumn(BaseVirtualColumn col1, BaseVirtualColumn col2) {
    boolean v = true;
    if (col1.getColumns() != null && col2.getColumns() != null) {
        v = (col1.getColumns().length == col2.getColumns().length);
        if (v) {
            for (int i=0; i<col1.getColumns().length; i++) {
                if (!isSameString(col1.getColumns()[i], col2.getColumns()[i])) {
                    v = false;
                    break;
                }
            }
        }
    }
    else {
        v = (col1.getColumns() == null && col2.getColumns() == null);
    }
    if (!v) return v;
    return 	col1.getColumnName().equals(col2.getColumnName()) &&
    col1.getOperation() == col2.getOperation() &&
    col1.getColumnType() == col2.getColumnType();
}

/** Override default write object to write the current version first. */
private void writeObject(java.io.ObjectOutputStream on) throws java.io.IOException {
    if (TRACE) {
        System.out.println("write version: " + getVersion());
    }
    // Write the current version
    on.writeObject(getVersion());
    try {
        on.defaultWriteObject();
    }
    catch (Exception e) {
        e.printStackTrace(System.out);
    }
}

/** Override default read object to read the serialized version first.
 *	Classes that changed between versions can compare the serialized
 *	version and the current version to see if compatibility needs to
 *	be handled.
 */
private void readObject(java.io.ObjectInputStream in) throws java.io.IOException {
    synchronized(serializedVersion) {
        try {
            // Get the version serialized.
            serializedVersion = (String) in.readObject();
            if (TRACE) {
                System.out.println("serialized version: " + serializedVersion);
            }
            in.defaultReadObject();
        }
        catch (Exception e) {
            e.printStackTrace(System.out);
        }
    }
}

}
