/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SerializedProperties.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:10 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.beans;

import com.klg.jclass.datasource.*;

import java.beans.Beans;
import java.io.*;
import java.text.MessageFormat;
import java.util.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.util.JCListenerList;
import com.klg.jclass.util.JCEnvironment;

/** Serializes and deserializes a bean component to/from a file. 
 *	Serialization file is a physical file that serialized a bean component.
 *	It is used at design-time.  When the bean component is deployed, there is
 *	no guarantee that the serialization file saved at design-time is located at a
 *	directory that is accessible at run-time.  Thus, a resource name relative to
 *	a class loader is used to specify where to get the serialized bean component.  
 *	<p>A resource is identified by a String that is /-separated sequence of 
 *	substrings.  e.g. resource/jclass/datasource/mydb.ser
 *	@see java.lang.ClassLoader
 */
public class SerializedProperties implements Serializable {

static final long serialVersionUID = 6152977937877513889L;

/** Serialization file for this bean component at design time. */
protected String    serializationFile;

/** resource name relative to a class loader.  
 *	It is used at run-time to get the bean component. 
 */
protected String    resourceName;

/** Default serialization file (base part) at design time. */
protected String    defaultSerializationFileBase;

/** Indicates if serialization is required.
 *	Serialization is required when a change to
 *	the serialized bean component is made.
 *	isSerializationRequired is reset after 
 *  the bean component has been saved to a file.
 */
private boolean isSerializationRequired = false;

/** Event listeners */
private transient JCListenerList JCDataListeners = null;
private transient boolean isFiringJCDataEvent = false;

private static final boolean TRACE = false;

/** Sets the bean component serialization file. */
public void setSerializationFile(String s) {
    if (isSameString(serializationFile, s)) return;

    this.serializationFile = s;
    setChanged();
}
/** Gets the bean component serialization file. */
public String getSerializationFile() {
    return serializationFile;
}


/** Sets resource name relative to a class loader.  
 *	It is used at run-time to get the bean component. 
 */
public void setResourceName(String s) {
    if (isSameString(resourceName, s)) return;

    this.resourceName = s;
    setChanged();
}

/** Gets resource name relative to a class loader.  
 *	It is used at run-time to get the bean component. 
 */
public String getResourceName() {
    return resourceName;
}

/** Indicates if resource name has been specified.
 *  @return True if resource name is present.
 */
public boolean isResourceNameSpecified() {
    if (resourceName != null && resourceName.length() != 0) {
        return true;
    }
    return false;
}

/** Indicates if serialization is required.
 *	Serialization is required when a change to
 *	the serialized bean component is made.
 *	isSerializationRequired is reset after 
 *	the bean component is saved to a file.
 */
public boolean isSerializationRequired() {
    return isSerializationRequired;
}

/** Instantiates a SerializedProperties from a resource. 
 *	@param cl class loader used to get the resource.
 *	@param resource_name resource name relative to a class loader 
 *						to the bean component. It must have the right extension.
 *	@see java.lang.ClassLoader
 *      @exception ClassNotFoundException If the class can't be loaded.
 *      @exception IOException If an IO exception occurs.
 */
public static SerializedProperties fromResource(ClassLoader cl, 
                                                String resource_name) throws ClassNotFoundException, IOException {
    if (resource_name == null || resource_name.length() == 0) return null;

    // convert resource name to a class name because class name
    // is expected in Beans.instantiate().
    String class_name = resource_name.substring(0, 
                                                resource_name.lastIndexOf(".")).replace('/', '.');
    return(SerializedProperties) Beans.instantiate(cl, class_name);
}

/** Instantiates a SerializedProperties from a resource.  First, 
 *	the package name of the caller class is prepended to the resource
 *	name to locate the resource.  If it fails, the resource name is
 *	used.
 *	@param caller_class class that uses the serialized bean component.
 *	@param resource_name resource name relative to a class loader 
 *						to the bean component. It must have right extension.
 *	@see java.lang.ClassLoader
 *      @exception ClassNotFoundException If the class can't be loaded.
 *      @exception IOException If an IO exception occurs.
 */
public static SerializedProperties fromResource(Class caller_class, 
                                                String resource_name) throws ClassNotFoundException, IOException {
    if (resource_name == null || resource_name.length() == 0) return null;

    if (TRACE) {
        System.out.println("caller_class: " + caller_class);
    }

    // convert resource name to a class name because class name
    // is expected in Beans.instantiate().
    String class_name = resource_name.substring(0, 
                                                resource_name.lastIndexOf(".")).replace('/', '.');

    ClassLoader cl = (caller_class == null) ? null : caller_class.getClassLoader();

    // Get the package name of the caller class
    String package_name = null;
    if (caller_class != null && caller_class.getName().indexOf(".") > -1) {
        package_name = caller_class.getName().substring(0, 
                                                        caller_class.getName().lastIndexOf("."));
    }

    SerializedProperties sb = null;
    // try packaged path to find the resource first.
    if (package_name != null) {
        String packaged_class_name = package_name + "." + class_name;
        try {
            sb = (SerializedProperties) Beans.instantiate(cl, packaged_class_name);
        }
        catch (Exception ex) {
            if (TRACE) {
                System.out.println("Can't find resource: " + packaged_class_name);
            }
        }
    }

    if (sb == null) {
        try {
            sb = (SerializedProperties) Beans.instantiate(cl, class_name);
        }
        catch (ClassNotFoundException cnfe) {
            if (TRACE) {
                System.out.println("Can't find resource: " + class_name);
            }
            throw cnfe;
        }
        catch (IOException ioe) {
            if (TRACE) {
                System.out.println("Can't find resource: " + class_name);
            }
            throw ioe;
        }
    }

    return sb;
}

/** Deserializes a serialization file to get a bean component. 
 *	@param file serialization file.  Itmust have the right extension.
 * 	@exception FileNotFoundException If the file find operation fails.
 * 	@exception ClassNotFoundException If the class cannot be found.
 * 	@exception IOException If an IO exception occurs.
 */
public static SerializedProperties fromFile(String file) throws 
FileNotFoundException, ClassNotFoundException, IOException {
    if (TRACE) {
        System.out.println("//////////////////////////////////////");
        System.out.println("Reading a bean component from file: " + file  + " ...");
        Thread.currentThread().dumpStack();
    }
    ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file));
    SerializedProperties comp = (SerializedProperties) ois.readObject();
    ois.close();
    if (TRACE) {
        System.out.println("Read bean component:" + comp);
        System.out.println("//////////////////////////////////////");
    }

    return comp;
}

/** Serializes the bean component to a file. 
 *	@param comp bean component to be serialized.
 *	@param file name of the file used to serialize the bean component.
 */
public static void toFile(SerializedProperties comp, String file) throws 
FileNotFoundException, IOException {
    String original_file = comp.getSerializationFile();
    try {
        File f = new File(file);
        if (!f.exists() || f.canWrite()) {
            ObjectOutputStream os = new ObjectOutputStream(new FileOutputStream(f));
            // Serialize component using the new file name.
            comp.setSerializationFile(file);
            if (TRACE) {
                System.out.println("\n//////////////////////////////////////\n");
                System.out.println("Writing the following bean component to " + 
                                   file + "\n" + comp);
                Thread.currentThread().dumpStack();
                System.out.println("\n//////////////////////////////////////\n");
            }
            os.writeObject(comp);
            os.flush();
            os.close();
            comp.isSerializationRequired = false;
        }
        else {
            Object[] args = { file};
            String msg = MessageFormat.format(
                                             LocaleBundle.string(LocaleBundle.file_readonly), args);
            throw new IOException(msg);
        }
    }
    catch (FileNotFoundException ex) {
        // Retain the original file name.
        comp.setSerializationFile(original_file);
        if (TRACE) {
            System.out.println("Writing bean component to " + file + " failed.");
            System.out.println("Exception: " + ex);
        }
        throw ex;
    }
    catch (IOException ex) {
        // Retain the original file name.
        comp.setSerializationFile(original_file);
        if (TRACE) {
            System.out.println("Writing bean component to " + file + " failed.");
            System.out.println("Exception: " + ex);
        }
        throw ex;
    }
    // serialization file may be changed.
    comp.fireJCDataEvent(new JCDataEvent((Object)comp));
}

/** Gets a default serialization file name. 
 *	The file name is in the form of base + number + the serialization extension.
 *	If a file already exists, the passed in counter is incremented
 *	unless there is no file for the derived name.
 */
protected static String getDefaultSerializationFile(String base, int counter) {
    String file = null;
    if (JCEnvironment.isDesignTime()) {
        for (file = base + counter + LocaleBundle.serialization_extension; 
            new File(file).exists(); file = base + ++counter + LocaleBundle.serialization_extension) {
        }
    }
    return file;
}

/** Sets the default serialization file base part. 
 */
protected void setDefaultSerializationFileBase(String base) {
    defaultSerializationFileBase = new String(base);
}

/** Gets a default serialization file name to open. 
 *	The file name is in the form of base + number + the serialization extension.
 *	If a file does not exist, null is returned.
 */
public String getDefaultSerializationFileToOpen() {
    int counter = 0;
    String prev_file = null;
    String file = defaultSerializationFileBase + counter + LocaleBundle.serialization_extension;;
    while (new File(file).exists()) {
        prev_file = file;           
        file = defaultSerializationFileBase + ++counter + LocaleBundle.serialization_extension;
    }
    return prev_file;
}

/** Returns if two strings are equal.  They are considered equal
 *	if both of them are null. 
 */
protected boolean isSameString(String s1, String s2) {
    return(s1 == null && s2 == null) ||
    (s1 != null && s2 != null && s1.equals(s2));
}

/** Override to set isSerializationRequired to true. */
protected void setChanged() {
    isSerializationRequired = true;
}

/**
 * Adds the specified listener to receive JCDataEvent objects.
 *
 * @param listener A listener.
 */
public synchronized void addJCDataListener(JCDataListener listener) {
    JCDataListeners = JCListenerList.add(JCDataListeners, listener);
}

/**
 * Removes the specified listener so it no longer receives JCDataEvent objects.
 *
 * @param listener A listener.
 */
public synchronized void removeJCDataListener(JCDataListener listener) {
    JCDataListeners = JCListenerList.remove(JCDataListeners, listener);
}

/**
 * Inform registered listeners of a JCDataEvent.
 *
 * @param event The JCDataEvent.
 *
 * @return true if the event was accepted by all listeners, or false
 * as soon as any listener cancels the change.
 */
public void fireJCDataEvent(JCDataEvent event) {
    if (isFiringJCDataEvent) {
        // avoid recursive calls
        return;
    }
    isFiringJCDataEvent = true;
    Enumeration e = JCListenerList.elements(JCDataListeners);
    while (e.hasMoreElements()) {
        JCDataListener listener = (JCDataListener)e.nextElement();
        listener.JCDataValueChanged(event);
    }
    isFiringJCDataEvent = false;
}

}
