/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:12 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.jdbc;

import java.sql.*;
import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.datasource.util.*;
import com.klg.jclass.datasource.LocaleBundle;

/**
 * A copy of the data returned in a JDBC result table will be
 * copied into one of these result tables so the data can be
 * cached. Rows can then be added, deleted or updated through
 * this DataTable.<P>
 * All operations can access data through row/column idxToBookmarkMap
 * rather than indexes.  This facilitates sorting of rows and
 * or columns.
 */
public class DataTable extends BaseDataTable {

static final long serialVersionUID = 6356453874717375331L;

protected final static int NO_ROWS_CHANGED = 0;  // sql statement failed
protected final static int ONE_ROW_CHANGED = 1;  // single row affected by statement
protected int columns = 0;                      // number of columns in row
protected transient PreparedStatement statement = null;    // pre-compile on server
protected MetaData metaData;

/**
 * Creates a table into which results from a JDBC result set will be
 * copied. Rows can be added, modified and deleted.
 * @param size an int for the number of columns to create
 * @param metaData the MetaData which will be parent
 */
DataTable(int size, MetaData metaData) {
    super(metaData);
    this.metaData = metaData;
    // set the number of columns in a row
    columns = size;
}

DataTable(int size, MetaData metaData, long parent_row) {
    super(metaData, parent_row);
    this.metaData = metaData;
    columns = size;
}

/**
 * For internal use only.
 * Call DataTableModel.getResultData(..) instead.
 * Returns a value for a given row/column idxToBookmarkMap.
 * @param bookmark a unique long representing a row
 * @param col_id a unique String representing a column
 * @return an Object which is the value at row/column
 * @exception DataModelException If get fails.
 * @see com.klg.jclass.datasource.DataTableModel#getResultData
 */
public Object getCell(long bookmark,String col_id)
throws DataModelException {
    int column_idx = metaData.getColumnIndex(col_id);
    Object dtr[] = (Object[]) getRowFromCache(bookmark);
    if (dtr != null) {
        return dtr[column_idx];
    }
    return null;
}

/**
 * Called by DataTableModel.createTable to create a new table
 * based on the input criteria.
 * @param bookmark a long which uniquely identifies a row
 * @param childMetaDataModel a MetaDataModel, this tells this table
 * what structure the child table it is creating should have
 * @exception DataModelException If materializeTable fails.
 */
public DataTableModel materializeTable(DataModelListener originator, long bookmark, TreeNode node)
throws DataModelException {
    // cast to MetaData
    MetaData child = (MetaData) node;
    if (!child.isOpen()) {
        String message = LocaleBundle.string(LocaleBundle.meta_data_not_open);
        message += metaData.getDescription() + ".createTable() or getTable()";
        throw new DataModelException(DataModelEvent.INITIALIZE, new Exception(message));
    }
    DataTable table = null;
    try {
        Enumeration e = child.join_columns.elements();
        // for each join column get its value and set it in the child's
        // query
        int cnt = 1; // parameters start at 1
        while (e.hasMoreElements()) {
            ColumnMap cm = (ColumnMap) e.nextElement();
            ColumnModel c = getMetaData().getColumnObject(cm.getParentColumn());
            int columnType = c == null ? Types.VARCHAR : c.getColumnType();
            child.setParameter(cnt++, getResultData(bookmark,cm.getParentColumn()), columnType);
        }
        table = child.executeQuery(bookmark);
    }
    catch (Exception e) {
        throw new DataModelException(bookmark, "", this, DataModelEvent.INITIALIZE, e);
    }
    return table;
}

/**
 * Creates a new row, called by addRow().
 * An implementation-specific method for creating a new (blank) row, it
 * must be implemented in the derived class of BaseDataTable.
 * @return an Object which is the new row just added.
 */
public Object createNewRow() {
    // create a new empty row
    Object row[] = new Object[columns];

    // loop on columns and add blanks values
    for (int i = 0; i < columns; i++) {
        row[i] = null;
    }
    return row;
}

/**
 * For internal use only.
 * Call DataTableModel.commitRow or DataModel.updateAll() instead.
 * Saves row changes to originating data source.
 * Called by BaseDataTable.commitRow().
 * @param bookmark a unique long indicating which row
 * to commit to the datasource.
 * @exception DataModelException If save fails.
 * @see com.klg.jclass.datasource.BaseDataTable#commitRow
 * @see com.klg.jclass.datasource.DataModel#updateAll
 */
public void saveRow(long bookmark) throws DataModelException {
    PersistorModel p = metaData.getPersistor();
    if (p != null) {
        p.saveRow(bookmark, metaData, this);
    }
}

/**
 * For internal use only.
 * Repopulates this DataTable by re-reading rows
 * from the originating data source.  Previous bookmarks
 * are invalid and need to be re-read.
 * Resets the cursor to the first row.
 * @exception DataModelException If root level requery fails.
 */
public DataTableModel requeryLevel() throws DataModelException {
    // remove old mappings
    store.clear();
    cursor = 0;

    // execute the query, root is ROOT_BOOKMARK
    if (parentRow < 0) {
        return metaData.executeQuery();
    }
    else {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.requery_root_only));
    	throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/**
 * For internal use only.
 * Call DataTableModel.cancelRowChanges(..) or DataModel.cancelAll() instead.
 * Restores a row's original values.
 * Called by cancelRowChanges(bookmark).
 * An implementation-specific method for restoring cell values, it
 * must be implemented in the derived class of BaseDataTable.
 * @param bookmark a long which uniquely identifies a row
 * @see com.klg.jclass.datasource.DataTableModel#cancelRowChanges
 * @see com.klg.jclass.datasource.DataModel#cancelAll
 */
public void restoreRow(long bookmark) {
    Object original_row[] = (Object[]) getOriginalRow(bookmark);
    int column = 0;
    for (int i = 0; i < original_row.length; i++) {
        try {
            setValueAt(bookmark, metaData.getColumnIdentifier(column++), original_row[i]);
        }
        catch (Exception x) {
            // System.out.println(x.toString());
        }
    }
}

/**
 * For internal use only.
 * Call DataTableModel.requeryRow(..) instead.
 * Re-reads a row from the originating data source.
 * Called by BaseDataTable.requeryRow().
 * An implementation-specific method for refetching a row, it
 * must be implemented in the derived class of BaseDataTable.
 * @param row_identifier an int which uniquely identifies a row
 * @exception DataModelException If requery from data source fails.
 * @see com.klg.jclass.datasource.DataTableModel#requeryRow
 */
public void refreshRow(long bookmark) throws DataModelException {
    boolean andNeeded = false;

    // break the sql statement into its component parts
    String query = metaData.getParser().getClause(SqlParser.SELECT);
    query += " " + metaData.getParser().getClause(SqlParser.FROM);
    query += " " + metaData.getParser().getClause(SqlParser.WHERE);

    // if there is no WHERE clause, add one
    if (metaData.getParser().getClause(SqlParser.WHERE).equals("")) {
        query += " WHERE ";
    }
    else {
        query += " AND ";
    }

    // loop on primary keys for the driver table
    // this is the table on which other tables join in one-to-one
    // relationships
    String table = (String) metaData.getTables().elementAt(metaData.getDriverTableIndex());
    Vector keys = metaData.getPrimaryKeys(table);
    SqlStatementTable sst = metaData.getParser().getSqlStatementTable(table);

    if (sst == null) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.cannot_requery_row_for_table) + table);
    	throw new DataModelException(bookmark, "", this, DataModelEvent.INITIALIZE, e);
    }

	// make sure we know the key so we can get the row ...
	if (keys.size() == 0) {
		String msg = "\n" + LocaleBundle.string(LocaleBundle.cannot_requery_row_key_unknown);
		msg += "\"" + table + "\"";
		Exception exc = new Exception(msg);
		throw new DataModelException(DataModelEvent.AFTER_REQUERY_ROW_AND_DETAILS, exc);
	}


    for (Enumeration e = keys.elements(); e.hasMoreElements();) {
        String col_id = (String) e.nextElement();
        col_id = col_id.substring(col_id.lastIndexOf(".") + 1, col_id.length());
        query += (andNeeded ? " AND " : " ");
        andNeeded = true;
        String alias = sst.getAlias() != null ? sst.getAlias() : sst.getTableName();
        query +=  alias + "." + col_id + " = ? ";
    }

    // if the was an additional clause re-add it (for example ORDER BY, etc)
    query += " " + metaData.getParser().getClause(SqlParser.REST);

    // get the row from the server
    Object row[] = (Object[]) getRowFromServer(table, query, bookmark);
    // do an in-place update
    int ordinal = store.getIndex(bookmark);
    store.removeRow(bookmark);
    store.addRow(bookmark, row, ordinal);
}

/**
 * Requery all rows in this table and collapse its child nodes.
 * Overrides BaseDataTable.requeryTable. If this is the root, this
 * method calls DataTableModel.requeryAll().
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If requery fails.
 * @see com.klg.jclass.datasource.TreeData#requeryAll
 */
public void requeryTable(DataModelListener originator) throws DataModelException {
    DataTable parentTable = (DataTable) this.getParent();
    if (parentTable == null) {
        // this is the root, so do requeryAll instead
        metaData.getDataModel().requeryAll(originator);
        return;
    }
    // give the listeners the opportunity to cancel this insert
    if (!fireDataModelEvent(new DataModelEvent(metaData.getDataModel(),
                                               originator, this, getParentBookmark(), "",
                                               DataModelEvent.BEFORE_REQUERY))) {
        return;
    }
    removeAllChildren();
    metaData.getDataModel().clearCurrentPath(metaData);
    try {
        DataTable newDt = (DataTable) parentTable.materializeTable(originator,
                                                                   getParentBookmark(), (TreeNode) getMetaData());
        this.store = newDt.store;
        cursor = 0;
    }
    catch (Exception e) {
        throw new DataModelException(DataModelEvent.AFTER_REQUERY_TABLE, e);
    }

    // fire after event
    fireDataModelEvent(new DataModelEvent(metaData.getDataModel(),
                                          originator, this, getParentBookmark(), "",
                                          DataModelEvent.AFTER_REQUERY_TABLE));
}

/**
 * Sends the query to the server, fetches and returns the row.
 * @exception DataModelException If requery from data source fails.
 */
protected Object[] getRowFromServer(String table, String query, long bookmark) throws DataModelException {
	Object row[];
	try {
	    if (statement == null) {
	        statement = metaData.getDataTableConnection().getConnection().prepareStatement(query);
	    }
	    else {
	        statement.clearParameters();
	    }
	    int cnt = 1;
	    for (Enumeration e = getCombinedKeys(table).elements(); e.hasMoreElements();) {
	        String col_id = (String) e.nextElement();
	        SqlStatementColumn ssc = metaData.getParser().getSqlStatementColumn(table, col_id);
	        String alias = null;
	        if (ssc != null) {
	            alias = metaData.getParser().getSqlStatementColumn(table, col_id).getAlias();
	        }
	        col_id = alias != null ? alias : col_id;
	        ColumnModel cm = getMetaData().getColumnObject(col_id);
	        int type = cm == null ? Types.VARCHAR : cm.getColumnType();
	        setParameter(cnt++, getResultData(bookmark, col_id), type);

	    }
	    ResultSet rs = statement.executeQuery();
	    // only expect one row

	    if (rs.next()) {
	        row = metaData.createRowFromResultSet(rs);
	    }
	    else {
	        rs.close();
	        String msg = LocaleBundle.string(LocaleBundle.requery_row_failed) + query;
	        throw new DataModelException(DataModelEvent.INITIALIZE, new Exception(msg));
	    }
	    rs.close();
	}
	catch (SQLException e) {
		throw new DataModelException(DataModelEvent.INITIALIZE, e);
	}
    return row;
}

/**
 * Sets parameters in the requery.
 * @param number the ordinal of the column in the statement
 * @param value the Object value to be set
 * @param columnType one of the java.sql.Types constants
 * @exception DataModelException If setParameter fails.
 */
protected void setParameter(int number, Object value, int columnType) throws DataModelException {
	try {
	    if (value == null) {
	        statement.setNull(number, columnType);
	    }
	    else {
	        statement.setObject(number, value);
	    }
	}
	catch (SQLException e) {
		throw new DataModelException(DataModelEvent.INITIALIZE, e);
	}
}

/**
 * Returns a Vector of column names which are the join columns
 * and the primary keys for the driver table.
 * @exception DataModelException If get fails.
 */
Vector getCombinedKeys(String table) throws DataModelException {
    Vector v = new Vector();

    for (Enumeration a = metaData.join_columns.elements(); a.hasMoreElements();) {
        ColumnMap cm = (ColumnMap) a.nextElement();
    	String col_id = cm.getChildColumn();
    	col_id = col_id.substring(col_id.lastIndexOf(".") + 1, col_id.length());
        v.addElement(col_id);
    }
    for (Enumeration x = metaData.getPrimaryKeys(table).elements(); x.hasMoreElements();) {
        String col_id = (String) x.nextElement();
        col_id = col_id.substring(col_id.lastIndexOf(".") + 1, col_id.length());
        v.addElement(col_id);
    }
    return v;
}

/**
 * Free any lingering jdbc resources.
 */
protected void finalize() {
	if (statement != null) {
		try {
			statement.close();
		}
		catch (SQLException e) {}
	}
}

}
