/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataTableConnection.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:12 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.jdbc;

import java.io.*;
import java.sql.*;
import java.util.*;

import com.klg.jclass.datasource.*;

/**
 * A connection to a JDBC datasource.  It can be shared between various
 * data controls.
 */
public class DataTableConnection implements Serializable {
	
static final long serialVersionUID = 5569509834849517005L;

protected transient Connection connection;   // connection to db for this ds
protected boolean autoCommitFlag = true;     // flag indicating autocommit mode

// to facilitate reconnection
private String url;
private String login;
private String password;
private String driver;
private Properties properties;

/**
 * DataTableConnection Constructor.
 * @param driver a String indicating which driver to load
 * @param url a String url for the connection
 * @param user a String for the user name
 * @param password a String for the user password
 * @param database a String for the data base name
 * @exception DataModelException If initialization fails.
 */
public DataTableConnection(String driver, String url, String user,
                           String password,String database) throws DataModelException {
    // load Driver
    loadDriver(driver);

    // now connect
    connect(url, user, password, database);
}

/**
 * DataTableConnection Constructor, empty, must load driver and
 * call connect after instantiating.
 */
public DataTableConnection() {
}

/**
 * Load this driver into memory.
 * @param driver a String indicating which driver to load
 * @exception DataModelException If driver fails to load.
 */
public void loadDriver(String driver) throws DataModelException {
    try {
        Class.forName(driver).newInstance();
		this.driver = driver;
    }
    catch (Exception e) {
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/**
 * Establish a connection to a data database using hostname and port.
 * @param url a String url for the connection
 * @param login a String for the user name
 * @param password a String for the user password
 * @param db a String for the data base name
 * @exception DataModelException If connection fails.
 */
public void connect(String url,String login,String password,String db)
throws DataModelException {

    // Use a particular db if supplied.
    if (db != null && !db.equals("")) {
        url += "/";
        url += db;
    }

    // Establish the connection.
    connect(url,login,password);
}

/**
 * Establish a connection to a data database using odbc style.
 * @param url a String url for the connection
 * @param login a String for the user name
 * @param password a String for the user password
 * @exception DataModelException If connection fails.
 */
public void connect(String url, String login, String password)
throws DataModelException {
    
    try {
        // Establish the connection
        connection = DriverManager.getConnection(url,login,password);

		// remember so we can reconnect
		this.url 		= url;
		this.login 		= login;
		this.password 	= password;			
    }
    catch (Exception e) {
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/**
 * Establish a connection using property object. For example,
 * <pre><code>
 *  Properties props = new Properties();
 *  props.put("user", "dba");
 *  props.put("password", "90210");
 *  c = new DataTableConnection();
 *  c.loadDriver("sun.jdbc.odbc.JdbcOdbcDriver");
 *  c.connect("jdbc:odbc:JClassDemoSQLAnywhere", props);
 * </code></pre>
 * @param url a String a database url of the form jdbc:subprotocol:subname
 * @param properties a list of arbitrary string tag/value pairs as
 * connection arguments; normally at least a "user" and "password"
 * property should be included
 * @exception DataModelException If connection fails.
 */
public void connect(String url, Properties properties)
throws DataModelException {
    try {
        // Establish the connection
        connection = DriverManager.getConnection(url, properties);
		this.url = url;
		this.properties = properties;
    }
    catch (Exception e) {
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}

/** 
 * Return this connection to the database.
 * @return a JDBC Connection 
 */
public Connection getConnection() {
    return connection;
}

/**
 * Close the connection to the database.
 */
public void disconnect() {
    try {
        connection.close();
    }
    catch (Exception e) {
        System.out.println(e.toString());
    }
}

/**
 * Reconnect using previous parameters.
 * @exception DataModelException If reconnect fails.
 */
public void reconnect() throws DataModelException{
	if (properties != null) {
		connect(url, properties);
	}
	else {
		connect(url, login, password);
	}
}

/**
 * Set auto commit value.
 * Determines whether to automatically commit uncommited changes to the  * data source when the current row changes (true), or wait until
 * commit is called manually (false).
 * @param value a boolean determining when the commit occurs.
 * @exception SQLException If data base access error occurs.
 */
public void setAutoCommit(boolean value) throws SQLException {
    if (autoCommitFlag != value) {
        connection.setAutoCommit(value);
        autoCommitFlag = value;
    }
}

/**
 * Get auto commit value.
 * Determines whether to automatically commit uncommited changes to the  * data source when the current row changes (true), or wait until
 * commit is called manually (false). Default is true.
 * @return a boolean determining when the commit occurs.
 * @exception SQLException If data base access error occurs.
 */
public boolean getAutoCommit() throws SQLException {
    return autoCommitFlag;
}

}
