/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Persistor.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:12 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.jdbc;

import java.io.*;
import java.sql.*;
import java.util.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.datasource.jdbc.*;
import com.klg.jclass.datasource.util.SqlStatementTable;

/**
 * Default mechanism for persisting (resolving) data to the
 * originating database.
 */
public class Persistor implements PersistorModel, Serializable {

static final long serialVersionUID = 922782684857133093L;

protected final static int ONE_ROW_CHANGED = 1;
protected transient PreparedStatement update_statement;

public void saveRow(long bookmark, MetaData metaDataModel,
                    DataTableModel dataTableModel)
	throws DataModelException
{
    // start by casting to jdbc objects
    MetaData metaData   = (MetaData) metaDataModel;
    DataTable dataTable = (DataTable) dataTableModel;

    // if there are no modified rows, return
    int row_status = dataTable.getRowStatus(bookmark);
    if (row_status == DataTableModel.COMMITTED) {
        return;
    }
    Object original_row[] = null;

    // get the tables associated with this level
    Vector tables = metaData.getTables();

    // loop on each table and ...
    int param = 1;
    for (Enumeration e = tables.elements(); e.hasMoreElements();) {
        String table = (String) e.nextElement();

        // get the positions of this table's columns in this row
        Vector columns = metaData.getColumnIndexesForTable(table);

        // update the row?
        if (row_status == DataTableModel.UPDATED) {
            if (original_row == null) {
                original_row = (Object[])dataTable.getOriginalRow(bookmark);
            }
            if (original_row == null) {
                Exception exc = new Exception(LocaleBundle.string(LocaleBundle.bookmark_not_found) + bookmark);
				throw new DataModelException(DataModelEvent.INITIALIZE, exc);
            }

            // if no columns modified for this table, skip it
            if (!tablesColumnsModified(bookmark,columns, metaData, dataTable)) {
                continue;
            }

            // otherwise, build the update statement
            buildUpdateStatement(table, columns, bookmark, metaData, dataTable);

            // loop on each column and set the new values
            Enumeration x = columns.elements();
            while (x.hasMoreElements()) {
                int col_idx = ((Integer)x.nextElement()).intValue();
                if (!dataTable.columnModified(bookmark, metaData.getColumnIdentifier(col_idx))) {
                    continue;
                }
                Object value = dataTable.getResultData(bookmark, metaData.getColumnIdentifier(col_idx));
                setUpdateParameter(param++,value,metaData.getColumnIdentifier(col_idx), metaData);
            }
        }
        else if (row_status == DataTableModel.DELETED) {
            if (!metaData.getDeleteAllowed(table)) {
                continue;
            }
            // delete the row
            original_row = (Object[]) dataTable.getDeletedRow(bookmark);
            buildDeleteStatement(table, columns, original_row, metaData, dataTable);
        }
        else {
            if (!metaData.getInsertAllowed(table)) {
                continue;
            }
            // insert this row, FYI "original_row" is the new row
            original_row = (Object[]) dataTable.getRowFromCache(bookmark);
            buildInsertStatement(table, columns, original_row, metaData, dataTable);
        }

        // loop on each column and set the where values using orignals
		// don't include nulls since they are handled by "is null"
		// in building of the where clause
        for (int i = 0; i < columns.size(); i++) {
            if (dataTable.originalCellWasNull(i, columns, original_row)) {
                continue;
            }
            int col_idx = ((Integer) columns.elementAt(i)).intValue();
            Object value = original_row[col_idx];
            setUpdateParameter(param++,value,metaData.getColumnIdentifier(col_idx), metaData);
        }
        commit(metaData);
        // reset param
        param = 1;
    }
}

/**
 * For internal use only.
 * Call DataTableModel.commitRow(..) or DataModel.updateAll() instead.
 * Actually commits this row to the server.
 * Called by commitRow.
 * @exception DataModelException If commit fails.
 */
protected void commit(MetaData metaData) throws DataModelException {
    int rows_changed = executeUpdate(metaData);
    if (rows_changed != ONE_ROW_CHANGED) {
        Exception e = new Exception(LocaleBundle.string(LocaleBundle.data_changed));
    	throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
}



/**
 * Returns true if at least one of this table's columns been modified.
 * @param bookmark a long which uniquely identifies a row
 * @param columns a Vector of indexes for this table's columns
 * @return boolean, true if one or more columns were modified
 */
protected boolean tablesColumnsModified(long bookmark, Vector columns,
                                        MetaData metaData, DataTable dataTable)
	throws DataModelException
{
    for (Enumeration e = columns.elements(); e.hasMoreElements();) {
        int col_idx = ((Integer)e.nextElement()).intValue();
        if (dataTable.columnModified(bookmark, metaData.getColumnIdentifier(col_idx))) {
            return true;
        }
    }
    return false;
}

/**
 * Builds the delete statement for a table.
 * @param table a String name for the table
 * @param columns a Vector of column indexes
 * @exception DataModelException If creation of delete statement fails.
 */
protected void buildDeleteStatement(String table, Vector columns,
                                    Object[] original_row, MetaData metaData,
									DataTable dataTable)
    throws DataModelException
{
    String update = " DELETE FROM ";
    update += getQualifiedTableName(table, false, metaData);
    update += buildWhereClause(columns, original_row, false, metaData, dataTable);
    setUpdateStatement(update, metaData);
}

/**
 * Default method for building update statement.
 * Loops on columns and builds a PreparedStatement.
 * @exception DataModelException If build fails.
 */
protected void buildUpdateStatement(String table, Vector columns, long bookmark,
                                    MetaData metaData, DataTable dataTable)
	throws DataModelException
{
    Object original_row[] = (Object[]) dataTable.getOriginalRow(bookmark);
    // start by setting "?" place holders for each column
    // which has changed
    String update = " UPDATE ";
    update += getQualifiedTableName(table, false, metaData);
    update += " SET ";
    boolean commaNeeded = false;
    for (int i = 0; i < columns.size(); i++) {
        int idx = ((Integer) columns.elementAt(i)).intValue();
        if (!dataTable.columnModified(bookmark,metaData.getColumnIdentifier(idx))) {
            continue;
        }
        update += (commaNeeded ? "," : " ");
        ColumnModel cm = metaData.getColumnObject(metaData.getColumnIdentifier(idx));
        String column = cm.getColumnName();
        update += column.substring(column.lastIndexOf(".") + 1, column.length());
        update += " = ? ";
        commaNeeded = true;
    }
    update += buildWhereClause(columns, original_row, false, metaData, dataTable);
    setUpdateStatement(update, metaData);
}

/**
 * Default method for building insert statement.
 * Loops on columns and builds a CallableStatement.
 * @exception DataModelException If creation of insert statement fails.
 */
private void buildInsertStatement(String table, Vector columns,
                                  Object[] original_row, MetaData metaData,
								  DataTable dataTable)
	throws DataModelException
{
    boolean commaNeeded = false;
    // explicitly tell which columns will be inserted
    String insert = " INSERT INTO ";
    insert += getQualifiedTableName(table, false, metaData);
    insert += " (";
    for (int i = 0; i < columns.size(); i++) {
        if (dataTable.originalCellWasNull(i, columns, original_row)) {
            continue;
        }
        insert += (commaNeeded ? ", " : " ");
        int idx = ((Integer) columns.elementAt(i)).intValue();
        ColumnModel cm = metaData.getColumnObject(metaData.getColumnIdentifier(idx));
        String column = cm.getColumnName();
        insert += column.substring(column.lastIndexOf(".") + 1,column.length());
        commaNeeded = true;
    }
    insert += " ) VALUES ( ";

    // now set placeholders for those values
    commaNeeded = false;
    for (int i = 0; i < columns.size(); i++) {
        if (dataTable.originalCellWasNull(i, columns, original_row)) {
            continue;
        }
        insert += (commaNeeded ? ", " : " ");
        insert += " ? ";
        commaNeeded = true;
    }
    insert += " ) ";

 	// if we're here and "commaNeed" still equals false, then the
 	// resulting sql statement will be malformed -- throw exception
	if (commaNeeded == false) {
		String msg = "\n" + LocaleBundle.string(LocaleBundle.malformed_sql_on_insert);
		msg += " level \"" + metaData.getDescription() + "\"";
		msg += " statement \"" + insert + "\"";
		Exception exc = new Exception(msg);
		throw new DataModelException(DataModelEvent.AFTER_INSERT_ROW, exc);
	}

    setUpdateStatement(insert, metaData);
}

/**
 * Builds a WHERE clause for update/delete statements.
 * Uses all non-null attributes to ensure the primary key is covered.
 * This also ensures the row has not
 * been changed by another user since if any column is
 * different in the database, the update/delete fails.
 * @param columns a Vector of column indexes for the table
 * to be deleted
 * @param original_row a DataTableRow which is the original
 * (unmodified) row from which where clause attributes are to
 * be gotten
 * @return the " WHERE column = ? [AND column = ?] ..." clause
 */
private String buildWhereClause(Vector columns, Object[] original_row,
                                boolean qualify, MetaData metaData,
								DataTable dataTable)
{
    String update = " WHERE ";
    boolean andNeeded = false;
    for (int i = 0; i < columns.size(); i++) {
        update += (andNeeded ? " AND " : " ");

        int idx = ((Integer) columns.elementAt(i)).intValue();
        try {
			ColumnModel cm = metaData.getColumnObject(metaData.getColumnIdentifier(idx));
            String column = cm.getColumnName();

            update += qualify ? column : column.substring(column.lastIndexOf(".") + 1,column.length());
        }
        catch (Exception e) { /* obviously open by now */
        }
		// handle null orignal values with "is null"
		if (dataTable.originalCellWasNull(i, columns, original_row)) {
			update += " IS NULL ";
		}
		else {
        	update += " = ? ";
		}
        andNeeded = true;
    }
    return update;
}

/**
 * Return the qualified version of the table name, i.e., can
 * be [catalog.][schema.]table [alias]
 */
protected String getQualifiedTableName(String table, boolean includeAlias,
                                       MetaData metaData)
{
    SqlStatementTable t = metaData.getParser().getSqlStatementTable(table);
    String qualified = t.getQualifier() != null && !t.getQualifier().equals("") ? t.getQualifier() + "." + table : table;
    if (includeAlias) {
        qualified += (t.getAlias() != null && !t.getAlias().equals(qualified)) ? " " + t.getAlias() : " ";
    }
    return qualified;
}

/**
 * Sets the SQL statement to update data.
 * Statement can be sql statement or a stored procedure.
 * @exception DataModelException If setUpdateStatement fails.
 */
protected void setUpdateStatement(String update, MetaData metaData) throws DataModelException {
    try {
        update_statement = metaData.getDataTableConnection().getConnection().prepareCall(update);
    }
    catch (Exception e) {
        try {
            // callable statements not supported, so upcast to prepared
            update_statement = metaData.getDataTableConnection().getConnection().prepareStatement(update);
        }
        catch (Exception e2) {
            throw new DataModelException(DataModelEvent.INITIALIZE, e2);
        }
    }
}

/**
 * Sets a particular parameter in the SQL update statement.
 * The update statement is automatically built.
 * @param number an int indicating the position of the parameter
 * @param value the Object which is to be set
 * @param col_id column_name the column alias or column name as it appears in the
 * select statement ([catalog.][schema.][table.]column)
 * @exception DataModelException If setUpdatdParameter fails.
 */
protected void setUpdateParameter(int number, Object value, String col_id,
                                  MetaData metaData)
    throws DataModelException
{
    int type = getJDBCColumnType(col_id, metaData);
	try {
	    if (value == null) {
	        try {
	            update_statement.setNull(number,type);
	        }
	        catch (java.sql.SQLException e) {
	            // see the java bug parade at
	            //      http://developer.java.sun.com/developer/bugParade/bugs/4038518.html
	            // where this bug is explained. Basically there are conversion
	            // problems between odbc2 which uses 9, 10 and 11 for date,
	            // time and timestamp, whereas odbc3 uses 91, 92, 93.
	            // So on exception we'll do the conversion ourselves.
	            switch (type) {
	            case java.sql.Types.DATE:
	                update_statement.setNull(number, 9); break;
	            case java.sql.Types.TIME:
	                update_statement.setNull(number, 10); break;
	            case java.sql.Types.TIMESTAMP:
	                update_statement.setNull(number, 11); break;
	            default:
	                throw new DataModelException(DataModelEvent.INITIALIZE, e);
	            }
	        }

	    }
	    else {
	        switch (type) {
	        case Types.BINARY:
	        case Types.VARBINARY:
	        case Types.LONGVARBINARY:
	            InputStream is = new ByteArrayInputStream((byte[]) value);
	            update_statement.setBinaryStream(number, is,
				                                 ((byte[])value).length);
	            break;
	        default:
	            update_statement.setObject(number, value);
	            break;
	        }
	    }
	}
	catch (SQLException sqle) {
		throw new DataModelException(DataModelEvent.INITIALIZE, sqle);
	}
}

/**
 * Commits the update to the database.
 * @return rows_affected an int indicating the number of rows affected
 * @exception DataModelException thrown if query fails
 */
protected int executeUpdate(MetaData metaData) throws DataModelException {
    if (!metaData.isOpen()) {
        throw new DataModelException(DataModelEvent.INITIALIZE,
		                             createNotOpenException("executeUpdate()",
									 metaData));
    }
    int rows_affected = 0;
    try {
        rows_affected = update_statement.executeUpdate();
    }
    catch (SQLException e) {
        throw new DataModelException(DataModelEvent.INITIALIZE, e);
    }
    return rows_affected;
}

/**
 * Creates the exception to be thrown when this instance has
 * not been opened, yet a get method has been called.
 */
private Exception createNotOpenException(String method, MetaData metaData)
{
    String message = LocaleBundle.string(LocaleBundle.meta_data_not_open);
    message += metaData.getDescription();
    message += " " + method;
    return new Exception(message);
}

/**
 * Gets the JDBC column type (for instance, java.sql.Types value) for this column.
 * @param column_identifier an String uniquely identifying a column.
 * @return an int representing the java.sql.Type type for this column.
 * @exception DataModelException If the operation fails.
 * @exception DataModelException If the operation fails.
 * @see java.sql.Types
 */
private int getJDBCColumnType(String column_identifier, MetaData metaData)
    throws DataModelException
{
    return((Column) metaData.getColumnObject(column_identifier)).getColumnType();
}

}
