/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DSdbJList.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:13 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.swing;

import java.awt.*;
import java.util.Vector;

import javax.swing.*;
import javax.swing.event.*;

import com.klg.jclass.datasource.BindingModel;
import com.klg.jclass.datasource.DataModel;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.datasource.DataModelException;

import com.klg.jclass.datasource.util.ExceptionManager;

/**
 * This class binds a JList to a column in the datasource. The list selection
 * responds to row changes in the datasource and can initiate one.
 */
public class DSdbJList extends JList implements ListSelectionListener {

protected ColumnDataBinding columnBinding;
protected boolean shouldMove = true;

/**
 * Creates a new DSdbJList component.
 */
public DSdbJList() {
    setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    columnBinding = new ColumnDataBinding();
    addListSelectionListener(this);
}

/**
 * Creates a new DSdbJList component and binds it to a particular column
 * in a MetaDataModel.
 * @see com.klg.jclass.datasource.beans.NodeProperties#setDataSource
 */
public DSdbJList(DataModel dataModel, MetaDataModel metaDataModel, String columnName) {
    this();
    columnBinding.setDataBinding(dataModel, metaDataModel, columnName);
}

/**
 * Create and bind to a Binding
 */
public DSdbJList(BindingModel binding, String columnName) {
    this();
    setDataBinding(binding, columnName);
}

/**
 * Only SINGLE_SELECTION selection mode is allowed.
 */
public void setSelectionMode(int selectionMode) {
    super.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
}

/**
 * Sets the parameters for binding the component to a particular column
 * in a MetaDataModel. If the MetaDataModel is the root in the DataModel's
 * MetaDataTree, the DataTableModel is automatically set to the root
 * DataTableModel, otherwise, the DataTableModel is set when an appropriate
 * table becomes available.
 */
public void setDataBinding(DataModel dataModel, MetaDataModel metaDataModel, String columnName) {
    columnBinding.setDataBinding(dataModel, metaDataModel, columnName);
}

/**
 * Sets the data source to a DataModel and binds the MetaData to the
 * MetaDataModel represented by a "path" of MetaDataModel descriptions
 * separated by '|' (e.g. Orders|Customers).
 */
public void setDataBinding(DataModel dataModel, String metaPath) {
    columnBinding.setDataBinding(dataModel, metaPath);
}

/**
 * Sets the data source to a DataModel and binds the MetaData to the
 * MetaDataModel represented by a "path" of MetaDataModel descriptions
 * separated by '|' (e.g. Orders|Customers).
 */
public void setDataBinding(BindingModel binding, String columnName) {
    columnBinding.setDataBinding(binding, columnName);
}

/**
 * Sets the data source to the path of DataModel, MetaDataModel descriptions
 * and column name.
 */
public void setDataBinding(String dataBinding) {
    columnBinding.setDataBinding(dataBinding);
}

/**
 * Returns the data binding hierarchy string, starting from a JCTreeData instance name
 * and containing full sequence of MetaDataModel levels starting from the root to the
 * level the table is bound with.
 */
public String getDataBinding() {
    return columnBinding.getDataBinding();
}

/*
 * ItemListener Implementation.
 *****************************************************************************/

/**
 * Called when an item's state has been changed.
 */
public void valueChanged(ListSelectionEvent event) {
    if (event.getValueIsAdjusting())
        return;

    // invoke to ensure that we are really bound
    if (!columnBinding.checkBindingSelected()) {
        return;
    }

    // shouldMove is a flag to indicate that we were not
    // the originator of this event
    if (columnBinding.isDataAvailable() && shouldMove) {
        columnBinding.moveToRow(getSelectedIndex());
    }

    shouldMove = true;
}

// the text to be displayed in the label initially
protected final static String nullDatabindingText = "DSdbJList";

// text margin to display nullDatabindingText
protected int textMargin = 2;

/**
 * Override to display default text if unbound.
 */
public void paint(Graphics gc) {
    if (columnBinding.isBindingSelected())
        super.paint(gc);
    else {
        FontMetrics fm = gc.getFontMetrics();
        Rectangle r = gc.getClipBounds();
        Dimension d = getPreferredSize();
        gc.setColor(getBackground());
        gc.fillRect(r.x, r.y, r.width, r.height);
        gc.setColor(getForeground());
        gc.drawString(nullDatabindingText, textMargin+(r.width-d.width)/2,
                      textMargin+fm.getLeading()+fm.getAscent()+(r.height-d.height)/2);
    }
}

public int getFixedCellWidth() {
    if (!columnBinding.isBindingSelected()) {
        FontMetrics fm = getFontMetrics(getFont());
        return fm.stringWidth(nullDatabindingText) + textMargin * 2;
    }
    return super.getFixedCellWidth();
}

public int getFixedCellHeight() {
    if (!columnBinding.isBindingSelected()) {
        FontMetrics fm = getFontMetrics(getFont());
        return fm.getHeight() + textMargin * 2;
    }
    return super.getFixedCellHeight();
}

public int getVisibleRowCount() {
    if (!columnBinding.isBindingSelected()) {
        return 1;
    }
    return super.getVisibleRowCount();
}

/**
 * Returns the preferred size of the component.
 */
public Dimension getPreferredSize() {
    Dimension d = super.getPreferredSize();
    if (!columnBinding.isBindingSelected()) {
        FontMetrics fm = getFontMetrics(getFont());
        d = new Dimension(fm.stringWidth(nullDatabindingText) + textMargin * 2,
                             fm.getHeight() + textMargin * 2);
    }
    return d;
}

/**
 * Returns the minimum size of the component.
 */
public Dimension getMinimumSize() {
    return getPreferredSize();
}

// END DSdbJList

/*
 * DataBinding Inner Class
 *****************************************************************************/

class ColumnDataBinding extends com.klg.jclass.datasource.util.ColumnDataBinding {

    public ColumnDataBinding() {
        super(DSdbJList.this);
        checkBindingSelected();
    }

    protected void moveToRow(int row) {
        moveToRowIndex(row);
    }

/**
 * Process a change in the data
 */
    protected void tableChanged() {
        readCurrentData();
    }

/**
 * Process a change in the data in the row
 */
    protected void rowDataChanged(int rowIndex) {
        cellDataChanged(rowIndex, column_name);
    }

/**
 * Process a single cell value change.
 */
    protected void cellDataChanged(int rowIndex, String columnName) {
        if (!(columnName.equals(column_name)))
            return;

        readCurrentData();
    }

/**
 * Called when the current row changes.
 */
    protected void currentRowChanged(int rowIndex) {

        // an event from the data model is telling us
        // to change rows, so set the shouldMove flag
        // so we don't, in turn, cause a move
        if (getSelectedIndex() != rowIndex) {
            shouldMove = false;
            setSelectedIndex(rowIndex);
        }

    }

/**
 * Checks if all listeners have accepted the move.
 */
    protected boolean beforeCurrentRowChange() {
        return true;
    }

/**
 * Must be called after a row was inserted.
 */
    protected void rowInserted() {
        readCurrentData();
    }

/**
 * Must be called after a row was deleted.
 */
    protected void rowDeleted() {
        readCurrentData();
    }

/**
 * Process a move to row for any metadata level.
 */
    protected void readCurrentData() {
        // reset the list
        if (!checkBindingSelected()) {
            return;
        }

        try {
            ListModel listModel = getModel();
            Vector elements = new Vector();
            Object value;
            boolean changed = false;
            for (int i=0; i<getNumRows(); i++) {
                value = binding.getCell(columnBinding, i, column_name);
            	if (i >= listModel.getSize()) {
            		changed = true;
            	}
            	else {
            		Object listValue = listModel.getElementAt(i);
            		if (value == null) {
            			value = "";
            		}
            		if (!value.toString().equals(listValue)) {
            			changed = true;
            		}
            	}

            	if (changed) {
            		elements.addElement(value == null ? " " : value.toString());
            	}
            }
            if (changed)
                setListData(elements);
        }
        catch (DataModelException dme) {
            ExceptionManager.getHandler().processException(getComponent(), dme);
        }
    }

    protected boolean checkBindingSelected() {
        if (!isBindingSelected()) {
            Vector v = new Vector();
            v.addElement(nullDatabindingText);
            setListData(v);
            return false;
        }
        return true;
    }

    protected void setDataAvailable(boolean state) {
        DSdbJList.this.setEnabled(state);
    }

    protected boolean isBindingSelected() {
        String s = this.getDataBinding();
        return isDataAvailable() && s != null && s.length() > 0;
    }

} // END ColumnDataBinding

}
