/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataBinding.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:13 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.util;

import java.awt.Component;

import java.io.Serializable;

import javax.swing.tree.TreeModel;
import javax.swing.tree.TreeNode;
import java.util.StringTokenizer;
import java.util.Vector;

import java.util.Enumeration;

import com.klg.jclass.datasource.DataModelEvent;
import com.klg.jclass.datasource.DataModelException;
import com.klg.jclass.datasource.DataModelListener;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.datasource.DataModel;
import com.klg.jclass.datasource.BindingModel;
import com.klg.jclass.datasource.DataTableModel;
import com.klg.jclass.util.JCEnvironment;

/**
 * Abstract class to aid in binding to a particular meta-data level in
 * a given data source
 */
public abstract class DataBinding implements DataModelListener, Serializable {

// we store the component that we are bound to so that we can popup
// a message dialog if necessary
protected BindingModel	binding;
protected Component 	component;
protected String 		dataBinding;

private static int DEBUG = -1;
private static final boolean TRACE = false;
private DataTableModel dataTableModel;

public static final String PATH_DELIMITER = "|";
public static final String SOURCE_DELIMITER = ":";

public DataBinding(Component component) {
    this.component = component;
}

public DataBinding(Component component, BindingModel binding) {
    this.component = component;
    setDataAvailable(binding.isDataAvailable());
    setDataBinding(binding);
}

/**
 * Allows the DataBinding to clean up any references.
 * @exception Throwable If an exception occurs.
 */
protected void finalize() throws Throwable {
    dispose();
}

/**
 * If the DataBinding is attached to a DataModel object, remove it from
 * the DataModel's DataModelListener list.
 */
public void dispose() {
    component = null;
    if (binding != null)
        binding.removeBindingListener(this);
}

/**
 * Returns the component this object is bound to
 */
public Component getComponent() {
    return component;
}

/**
 * Sets the data source to a DataModel and binds the MetaData to
 * the MetaDataModel represented by a "path" of MetaDataModel
 * descriptions separated by '|' (e.g. Orders|Customers).
 *
 */
 public void setDataBinding(DataModel dataModel, String metaDescription) {
    if (dataModel == null) {
        // unbind
        // setDataBinding(null);
        return;
    }

    MetaDataModel mdm = (MetaDataModel)((TreeModel)dataModel.getMetaDataTree()).getRoot();
    StringTokenizer st = new StringTokenizer(metaDescription, PATH_DELIMITER);
    if (st.hasMoreTokens()) {
        String s = st.nextToken();
        while (st.hasMoreTokens()) {
            s = st.nextToken();
            Enumeration e = mdm.children();
            while (e.hasMoreElements()) {
                MetaDataModel m = (MetaDataModel)e.nextElement();
                if (s.compareTo(m.getDescription()) == 0) {
                    mdm = m;
                    break;
                }
            }
        }
    }
    setDataBinding(mdm.getBinding());
}

/**
 * Returns the data binding hierarchy string, starting from a data model
 * instance name and containing full sequence of MetaDataModel levels
 * starting from the root to the level the table is bound with.
 *
 * @see com.klg.jclass.datasource.beans.NodeProperties#setDataSource
 * @see #setDataBinding
 *
 */
public void setDataBinding(String data_binding) {
    if (JCEnvironment.isDesignTime()) {
        dataBinding = data_binding;
    }

    if (data_binding == null || data_binding.length() == 0) {
        // unbind
    	// setDataBinding(null);
        return;
    }

    // find the name of the data source
    int i = data_binding.indexOf(SOURCE_DELIMITER);
    if (i == -1) {
        System.out.println("Invalid data binding hierarchy: \"" +
                           data_binding + "\"");
        return;
    }

    // find the root node of the data source
    String path = data_binding.substring(0, i);
    DataModel dm = null;
    Vector inst_list = com.klg.jclass.datasource.beans.DataSourceBeanInstance.getInstances();

    for (int j = 0; j < inst_list.size(); j++) {
        dm = (DataModel)(inst_list.elementAt(j));
        if (dm != null && dm.getModelName() != null &&
            path.compareTo(dm.getModelName()) == 0)
            break;
    }

    if (dm == null) {
        return;
    }

    // call the set method that infers the meta-level
    setDataBinding(dm, data_binding.substring(i + 1));
}

/**
 *
 * Return the hierarchy string for this level.
 * @return the string which indicates the hierarchy for this level.
 *
 */
public String getDataBinding() {
	if (binding == null || !binding.isDataAvailable()) {
        // At design-time, JBuilder creates non-GUI components
        // last.  It means JCData, and JCTreeData
        // are created after a DS component.  To make DataBinding
        // editor work, the data binding string is returned to
        // the editor.
        if (JCEnvironment.isDesignTime()) {
            return dataBinding == null ? "" : dataBinding;
        }
        else {
            return "";
        }
    }

    // work backwards through the meta-data levels to (lastly) the data source
    String s = "";
    MetaDataModel mdm = binding.getMetaDataModel();
    while (mdm != null) {
        s = mdm.getDescription() + (s.length() == 0 ? "" : PATH_DELIMITER) + s;
        mdm = (MetaDataModel)mdm.getParent();
    }
    s = binding.getDataModel().getModelName() + SOURCE_DELIMITER + s;
    return s;
}

/**
 * Return the BindingModel for this data binding object.
 * @return the BindingModel for this component.
 */
public BindingModel getBinding() {
	return binding;
}

/**
 * Sets up parameters to bind the DataBinding to a MetaDataMode in the
 * given DataModel.
 */
public void setDataBinding(BindingModel binding) {
    setDataBindingInternalBegin(binding);
    setDataBindingInternalEnd();
}

/**
 *
 */
public void setDataBinding(DataModel dataModel, MetaDataModel metaDataModel) {

	BindingModel bindingModel = null;
	if (dataModel != null) {
		// if no metaDataModel is specified, assume the root
		bindingModel = metaDataModel == null ?
			((MetaDataModel)dataModel.getMetaDataTree().getRoot()).getBinding() :
				metaDataModel.getBinding();
	}
	setDataBinding(bindingModel);
}

/**
 * When overriding setDataBinding() it is necessary to call this
 * method first.<p>
 *
 * Sets up parameters to bind the DataBinding to a MetaDataModel in the
 * given DataModel.
 */
public void setDataBindingInternalBegin(BindingModel binding) {

    // if the binding is not null, a new data model is replacing the
    // existing one, so the DataModelListener needs to be removed.
    if (this.binding != null) {
        binding.removeBindingListener(this);
    }

	this.binding = binding;

	if (binding == null) {
		return;
	}

    // Datasource allows us to hide deletes rows only
    // if commit policy is not COMMIT_LEAVING_RECORD
    try {
        if (binding.getMetaDataModel().getCommitPolicy() != MetaDataModel.COMMIT_LEAVING_RECORD) {
            binding.getMetaDataModel().setShowDeletedRows(false);
        }
    }
    catch (DataModelException exception) {
        processException(exception);
    }
}

/**
 * When overriding setDataBinding() it is necessary to call this
 * method just before returning
 */
public void setDataBindingInternalEnd() {
    if (binding == null) {
        toggleDataAvailableState();
        return;
    }
    binding.addBindingListener(this);
    // If binding to a DataSource bean then adding the listener is
    // sufficient but if binding to a non-bean data-source it is
    // necessary (because of bad design which should rectified in the
    // next release) to request the data table
    checkDataTableChanged();
}

public Object getNativeMetaData() {
	return binding.getMetaDataModel().getNativeMetaData();
}

/**
 * Returns the number of rows the data source table contains.
 * Must not be called if the data are not available.
 */
public int getNumRows() {
    // This must not be called if dataTableModel is null, so there is no
    // check for null dataTableModel.
    return binding.getRowCount();
}

/**
 * Returns the number of rows the data source table contains.
 * Must not be called if the data source is disabled.
 */
public int getNumColumns() {
	return binding.getColumnCount();
}

/*
 * JClass DataModelListener Implementation.
 *****************************************************************************/

/**
 * Called by the data control to inform this component data has changed.
 */

/**
 * Called before a row is inserted. Can be cancelled.
 * If not cancelled will be followed by afterInsertRow().
 * @see DataTableModel#addRow
 */
public void beforeInsertRow(DataModelEvent e) {
}

/**
 * Called before entire grid is reset. Can be cancelled.
 * If not cancelled will be followed by afterReset().
 * This will result from a call to DataModel.requeryAll().
 * @see DataModel#requeryAll
 */
public void beforeReset(DataModelEvent e) {
}

/**
 * Called when either DataTableModel.requeryRowAndDetails() or
 * DataTableModel.requeryRow() is called.
 * If not cancelled will be followed by an,
 * <ul>
 * <li>afterRequeryRowAndDetails() call or an</li>
 * <li>afterDeleteRow() call in the case getRowStatus() == INSERTED or a</li>
 * <li>afterChangeOfRowData() call in the case getRowStatus() == UPDATED or COMMITTED.</li>
 * </ul>
 * @see DataTableModel#requeryRow
 * @see DataTableModel#requeryRowAndDetails
 */
public void beforeRequery(DataModelEvent e) {
}

/**
 * Called before a cell is edited. Can be cancelled.
 * @see DataTableModel#updateCell
 */
public void beforeEditCell(DataModelEvent e) {
}

/**
 * Called before a row is [logically] deleted. Can be cancelled.
 * If not cancelled will be followed by an
 * afterDeleteRow90 or a afterChangeOfRowData() if the commit policy
 * is COMMIT_MANUALLY or COMMIT_LEAVING_ANCESTOR.
 * @see DataTableModel#deleteRow
 * @see MetaDataModel#getCommitPolicy
 */
public void beforeDeleteRow(DataModelEvent e) {
}

/**
 * Called before all changes are committed. Can be cancelled.
 * All modified, deleted and inserted rows at all levels are about
 * to be committed.  beforeDeleteRow() and afterChangeOfRowData()
 * will follow depending on the operations performed on the modified rows
 * being saved.
 * Results from a call to DataModel.updateAll.
 * @see DataModel#updateAll
 */
public void beforeCommitAll(DataModelEvent e) {
    handleCancellingBeforeEvents(e);
}

/**
 * Called before all changes are cancelled. Can be cancelled.
 * afterInsertRow() and afterChangeOfRowData() can follow this
 * event.
 * @see DataModel#cancelAll
 */
public void beforeCancelAll(DataModelEvent e) {
}

/** Called before all edits to a row are undone. Can be cancelled
 * An afterDeleteRow() or afterChangeOfRowData() will follow.
 * @see DataTableModel#cancelRowChanges
 */
public void beforeCancelRowChanges(DataModelEvent e) {
}

/**
 * Called when the root-level bookmark for a subtree changes.
 * When this happens those nodes in the previous subtree which
 * are not COMMIT_MANUALLY are committed.
 * Can be cancelled.  If cancelled the cursor moves but the
 * changes are automatically committed.
 */
public void beforeCommitConditional(DataModelEvent e) {
}

/**
 * Called before single row is committed to data source.  Can be cancelled in
 * which case the row edits are not written to the datasource and the rows
 * status remains modified.  afterDeleteRow() or afterChangeOfRowData() will
 * follow depending on the status of the row to be committed.
 * @see DataTableModel#commitRow
 */
public void beforeCommitRow(DataModelEvent e) {
    handleCancellingBeforeEvents(e);
}

/**
 * The global cursor will move to a new row.
 * Can be cancelled.
 */
public void beforeMoveToCurrentRow(DataModelEvent e) {
    handleCancellingBeforeEvents(e);
}

/**
 * The indicated DataTable will be deleted and flushed from the cache.
 * Can be cancelled.
 */
public void beforeDeleteTable(DataModelEvent e) {
}

/**
 * A row has changed, re-read all its cells and its status to reflect the new
 * values.  If this event is the result of a cell edit call DataModelEvent.getColumn()
 * to get the name of the column which changed. If getColumn() returns "", re-read
 * the entire row.
 * Called when
 * <ul>
 * <li>a row is deleted and getShowDeletedRows() == true</li>
 * <li>a cell was edited</li>
 * <li>row edits are cancelled and getRowStatus() == UPDATED</li>
 * <li>row edits are cancelled and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>row is commited and getRowStatus() == UPDATED or INSERTED</li>
 * <li>row is requeried and getRowStatus() != INSERTED</li>
 * </ul>
 */
public void afterChangeOfRowData(DataModelEvent e) {
    String colName = e.getColumn();
    int rowIndex = e.getRowIndex();
    if (colName == null || colName.length() == 0) {
        rowDataChanged(rowIndex);
    }
    else {
        cellDataChanged(rowIndex, colName);
    }
}

/**
 * A new row has been added to the datasource. Listeners need to
 * display the row.  Rows are always added to the end of DataTableModels.
 */
public void afterInsertRow(DataModelEvent e) {
    rowInserted();
}

/**
 * Removes the row from the display.
 * A row has been physically deleted and needs to be
 * removed from the display or has been logically deleted but the
 * showDeletedRows property has been set to false. Called when,
 * <ul>
 * <li>a row has been logically deleted and getShowDeletedRows() == false</li>
 * <li>row changes have been cancelled and getRowStatus() == INSERTED</li>'
 * <li>a row is committed and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>a row has been requeried and getRowStatus() == INSERTED</li>
 * </ul>
 */
public void afterDeleteRow(DataModelEvent e) {
    rowDeleted();
}

/**
 * Listeners must close all expanded views and reset/reread the root node.
 * The previous pointer to the root node is no longer valid.  Call
 * DataModel.getDataTableTree().getRoot() for the new root table.
 * Called when the datasource has been reset.
 * @see DataModel#requeryAll
 */
public void afterReset(DataModelEvent e) {
	checkDataTableChanged();
}

/**
 * The global cursor has moved to a new row.
 * Listeners should position their cursor on
 * the indicated row.  In a master-detail relationship child levels
 * should refresh themselves to reflect data sets which correspond
 * to the new parent row by calling DataModel.getCurrentDataTable()
 * or for field controls, DataModel.getCurrentDataItem().
 */
public void afterMoveToCurrentRow(DataModelEvent e) {
    checkDataTableChanged(e);
}

/**
 * Re-reads the indicated row and refresh all open children
 * under this row.
 */
public void afterRequeryRowAndDetails(DataModelEvent e) {
    handleCheckingAfterEvents(e);
}

/**
 * Re-read this table and refresh all open children in
 * this table.
 */
public void afterRequeryTable(DataModelEvent e) {
    handleCheckingAfterEvents(e);
}

/**
 * Notification that multiple events are coming.
 * Multiple events will be nested
 * between beginEvents() and endEvents().
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints.
 */
public void beginEvents(DataModelEvent e) {
}

/**
 * Notification that multiple events are complete.
 * Multiple events will be nested
 * between beginEvents() and endEvents().
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints. Called when
 * <ul>
 * <li>DataModel.updateAll is called</li>
 * <li>DataModel.cancelAll is called</li>
 * </ul>
 */
public void endEvents(DataModelEvent e) {
    handleCheckingAfterEvents(e);
}

/**
 * The current row has been deleted and the originator
 * of the deletion should now reposition the global
 * cursor to a new, valid row.
 */
public void originatorNavigateRow(DataModelEvent e) {
    // determine where to go to after a delete if this object is the originator
    // of an event that triggered a delete to commit (or a logical delete
    // if ShowDeletedRows is false.
    if (this == e.getOriginator()) {
        moveToBestRowAfterDelete(e);
    }
}

/**
 * For internal use only
 * Used by the JCData bean.
 */
public void initialize(DataModelEvent e) {
    handleCheckingAfterEvents(e);
}

protected void handleCancellingBeforeEvents(DataModelEvent e) {
    if (!beforeCurrentRowChange() && e.isCancelable()) {
        e.cancelProposedAction();
    }
}

protected void handleCheckingAfterEvents(DataModelEvent e) {
    checkDataTableChanged();
}

/**
 * Process a change in the data
 */
protected abstract void tableChanged();

/**
 * Must be called after row data were changed.
 */
protected abstract void rowDataChanged(int rowIndex);

/**
 * Must be called after a single cell value was changed.
 */
protected abstract void cellDataChanged(int rowIndex, String columnName);

/**
 * Changes the current row in the component.
 */
protected abstract void currentRowChanged(int rowIndex);

/**
 * Checks if all listeners have accepted the move.
 */
protected abstract boolean beforeCurrentRowChange();

/**
 * Must be called after a row was inserted.
 */
protected abstract void rowInserted();

/**
 * Must be called after a row was deleted.
 */
protected abstract void rowDeleted();

/**
 * Called whenever the DataTableModel changes to
 * an invalid or valid state. State is invalid when there is
 * no DataTableModel associated with the MetaDataModel.
 */
protected abstract void setDataAvailable(boolean state);

/**
 * Returns the state of the data binding. State will be enabled
 * when a valid DataTableModel exists.
 */
public boolean isDataAvailable() {
    return binding == null ? false : binding.isDataAvailable();
}

/**
 * Process a move to row for any metadata level.
 */
protected void checkDataTableChanged() {
    checkDataTableChanged(null, -1);
}

/**
 * Process a move to row for any metadata level
 */
protected void checkDataTableChanged(DataModelEvent event) {
    checkDataTableChanged(event.getTable().getMetaData(), event.getRowIndex());
}

/**
 * Determines whether the DataTableModel for this object's metadata level
 * has been replaced by a new DataTableModel and calls tableChanged to
 * notify the component. It also updates the row index and calls
 * current row changed.
 */
private void checkDataTableChanged(MetaDataModel mdm, int row) {
    DataTableModel dtm = null;
    try {
        // get the "current" table for this MetaDataModel
        dtm = binding.getCurrentDataTableModel(this);
        // check if the table is different from the current table
        if (dtm != dataTableModel) {
            dataTableModel = dtm;
            // set the state to disabled if the DataTableModel is null
            toggleDataAvailableState();
            tableChanged();
        }
        // if there is a DataTableModel, get the current row index
        if (binding.isDataAvailable()) {
            currentRowChanged(binding.getCurrentRowIndex());
        }
    }
    catch (DataModelException exception) {
        processException(exception);
    }
}

/**
 * Determines the best row to move to after a row delete has been committed (or logically
 * deleted with ShowDeletedRows false). This should only be called by the originator
 * of the action that forced the commit.
 *
 * For all rows but the last, the next row becomes current after a delete. If the deleted
 * row is the last row, the previous row becomes current. If there are no rows left
 * after the delete, the parent record of the table becomes current (if there is a parent).
 */
protected void moveToBestRowAfterDelete(DataModelEvent event) {
	try {
		binding.moveToBestRowAfterDelete(this, event);
	}
	catch (DataModelException e) {
		processException(e);
	}
}

/**
 * Returns current zero-based row index for the DataTableModel instance.
 * @return -1 if current bookmark is invalid.
 */
public int getCurrentRowIndex() {
	return binding == null ?
		com.klg.jclass.datasource.DataTableModel.ROW_NOT_FOUND : binding.getCurrentRowIndex();
}

/**
 * Sets DataModel current row.
 */
protected boolean moveToRowIndex(int rowIndex) {
    //check to make sure binding exists.
    if (binding != null) {
	
	try {
		return binding.moveToRowIndex(this, rowIndex);
	} catch (DataModelException dme) { }
    }
    return false;
}

/**
 * Checks the new data available state to determine whether a setDataAvailable call is made.
 */
protected void toggleDataAvailableState() {
    setDataAvailable(binding.isDataAvailable());
}

/**
 * Returns the value for the column label at the specified column.
 * @param column the column index.
 * @return true if the column can be edited; false otherwise.
 * Must not be called if the data source is disabled.
 */
public String getTableColumnLabel(int columnIndex) {
	return binding.getColumnIdentifier(columnIndex);
}

/**
 * Returns the JavaColumnType for the column.
 * Must not be called if the data source is disabled.
 */
public int getJavaColumnType(String columnName) {
	return binding.getJavaColumnType(columnName);
}

/**
 * Determines whether the column can be edited.
 * @param columnName the column name string.
 * @return true if the column can be edited; false otherwise.
 * Must not be called if the data source is disabled.
 */
public boolean isColumnEditable(String columnName) {
	return binding.isColumnEditable(columnName);
}

protected void processException(DataModelException exception) {
	ExceptionManager.getHandler().processException(getComponent(), exception);
}

}
