/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: NavigatorDataBinding.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:13 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.util;

import java.awt.Component;
import java.util.Vector;
import javax.swing.tree.TreeNode;

import com.klg.jclass.datasource.DataModelException;
import com.klg.jclass.datasource.Binding;

/**
 * An abstract class extending DataBinding with data
 * manipulation functionality.
 */
public abstract class NavigatorDataBinding extends DataBinding {

/**
 * Creates a new NavigatorDataBinding object for the
 * the given componenet.
 * @param component data-aware component instance
 */
public NavigatorDataBinding(Component component) {
    super(component);
}

/**
 * Creates a new NavigatorDataBinding object which has to bind the
 * componenet to the given data model and metadata level.
 * @param component data-aware component instance
 * @param dataModel data source data model instance
 * @param metaDataModel data source metadata model instance.
 */
public NavigatorDataBinding(Component component, Binding binding) {
    super(component);
    setDataBinding(binding);
}

/**
 * Inserts a new row into the data source and moved the data source
 * current row to the end of the table.
 * @return false if insert fails by any reason.
 * Must not be called if the data are not available.
 */
public boolean addRow() {
    try {
        ((Binding)binding).addRow(this);
        moveToRowIndex(getNumRows()-1);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Deletes current row from the data source.
 * @return false if the current row is undefined in the
 * data source or delete fails by any reason.
 */
public boolean deleteRow() {
    int row = getCurrentRowIndex();
    return row >= 0 ? deleteRows(row, 1) : false;
}

/**
 * Deletes specified number of rows starting with
 * given row index from the data source.
 * @return false if the set of rows is empty or delete fails by any reason.
 * Must not be called if the data are not available.
 */
public boolean deleteRows(int row, int number) {
	if (number <= 0) {
        return false;
	}
	try {	
		((Binding)binding).deleteRows(this, row, number);
    }
    catch (DataModelException exception) {
        processException(exception);
    	return false;
    }
	
	return true;
}

/**
 * Requeries the current row in the data source.
 * @return false if the current row is undefined in the
 * data source or requery fails by any reason.
 */
public boolean requeryRow() {
    int row = getCurrentRowIndex();
    return row >= 0 ? requeryRows(row, 1) : false;
}

/**
 * Requeries specified number of rows starting with
 * given row index in the data source.
 * @return false if the set of rows is empty or requery fails by any reason.
 * Must not be called if the data are not available.
 */
public boolean requeryRows(int row, int number) {
    if (number <= 0)
        return false;

    try {
        for (int i = row; i < row + number; i++) {
            binding.requeryRow(this, row);
        }
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Requeries the current row and detail tables in the data source.
 * Deletes all child nodes then calls requeryRow().
 * @return false if the current row is undefined in the
 * data source or requery fails by any reason.
 */
public boolean requeryRowAndDetails() {
    int row = getCurrentRowIndex();
    return row >= 0 ? requeryRowsAndDetails(row, 1) : false;
}

/**
 * Requeries rows and detail tables for specified number of rows starting with
 * given row index from the data source.
 * Deletes all child nodes then calls requeryRow().
 * @return false if the set of rows is empty or requery fails by any reason.
 * Must not be called if the data are not available.
 */
public boolean requeryRowsAndDetails(int row, int number) {
    if (number <= 0)
        return false;

    try {
        for (int i = row; i < row + number; i++)
            binding.requeryRowAndDetails(this, i);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Requeries all data tables in the data source.
 * @return false if requery fails by any reason.
 */
public boolean requeryAll() {
    try {
        binding.requeryAll(this);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Cancels all the changes made for the current row in the data source.
 * @return false if the current row is undefined in the
 * data source or cancel fails by any reason.
 */
public boolean cancelRow() {
    int row = getCurrentRowIndex();
    return row >= 0 ? cancelRows(row, 1) : false;
}

/**
 * Cancels all the changes made for the specified number of rows
 * starting with given row index in the data source.
 * @return false if the set of rows is empty or cancel fails by any reason.
 * Must not be called if the data are not available.
 */
public boolean cancelRows(int row, int number) {
    if (number <= 0)
        return false;

    try {  		
        for (int i = row; i < row + number; i++)
            ((Binding)binding).cancelRowChanges(this, i);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Recursively cancels all uncommited rows. Restores original cell values for each
 * uncommited row, for new rows remove them.
 * @return false if cancel fails by any reason.
 */
public boolean cancelAll() {
    try {
        ((Binding)binding).cancelAll(this);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Commits all the changes made for the current row in the data source.
 * @return false if the current row is undefined in the
 * data source or commit fails by any reason.
 */
public boolean commitRow() {
    int row = getCurrentRowIndex();
    return row >= 0 ? commitRows(row, 1) : false;
}

/**
 * Commits all the changes made for the specified number of rows
 * starting with given row index in the data source.
 * @return false if the set of rows is empty or commit fails by any reason.
 * Must not be called if the data are not available.
 */
public boolean commitRows(int row, int number) {
    if (number <= 0)
        return false;

    try {
        for (int i = row; i < row + number; i++)
            ((Binding)binding).commitRow(this, i);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Commits all the changes made in the data source.
 * @return false if commit fails by any reason.
 */
public boolean commitAll() {
    try {
        ((Binding)binding).commitAll(this);
        return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Returns the value for the row label at the specified row.
 * Labels use the CellData interface just like cells do.
 * Must not be called if the data are not available.
 */
public Object getTableRowLabel(int row) {
    // this shouldn't be called if there are no rows (so there is no
    // check for null DataTableModel

    /* FOR DEBUGGING PURPOSES
    try {
        return "[" + dataTableModel.getRowIdentifier(row) +  (isRowModified(row) ? "*]" : "]");
    } catch(Exception e) {}
    */

    return  row+(isRowModified(row) ? "*" : "  ");
}

/**
 * Returns the value for the specified location.
 * Must not be called if the data are not available.
 */
public Object getResultData(int row, int column) {
    // This shouldn't be called if there are no rows (so there is no
    // check for null DataTableModel
    try {
		return binding.getCell(this, row, column);
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return null;
}

/**
 * Set the value in the data source.
 * Must not be called if the data are not available.
 */
public boolean updateCell(Object value, int row, int column) {
    // This shouldn't be called if there are no rows (so there is no
    // check for null DataTableModel
    try {
		((Binding)binding).setCell(this, row, column, value);
    	return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Returns true if the row has been modified.
 * Must not be called if the data are not available.
 */
public boolean isRowModified(int row) {
    // This shouldn't be called if there are no rows (so there is no
    // check for null DataTableModel
	boolean modified = false;
	try {
		((Binding)binding).isRowModified(row);
	} 
	catch (DataModelException exception) {
        processException(exception);
    }
	return modified;
}

/**
 * Returns the status of the row.
 * Must not be called if the data are not available.
 */
public int getRowStatus(int row) {
    // This shouldn't be called if there are no rows (so there is no
    // check for null DataTableModel
	try {
		return ((Binding)binding).getRowStatus(row);
	} 
	catch (DataModelException exception) {
        processException(exception);
    }
	return -1;
}

/**
 * Requeries a level if the data source table is available or all otherwise.
 * @return false if requery fails by any reason.
 * Must not be called if the data source is disabled.
 */
public boolean requeryLevel() {
    try {
		binding.requeryLevel(this);
    	return true;
    }
    catch (DataModelException exception) {
        processException(exception);
    }
    return false;
}

/**
 * Called whenever the DataTableModel changes to
 * an invalid or valid state.
 */
protected void setDataAvailable(boolean state) {
    boolean enabled = component.isEnabled();
    state = isEnabled();
    if (state != enabled)
        component.setEnabled(state);
}

/**
 * Returns true if the component is bound to the datasource.
 */
public boolean isEnabled() {
	if (binding != null) {
    	return binding.getDataModel() != null && binding.getMetaDataModel() != null;
	}
	return false;
}

/**
 * Returns true if the component is bound to the datasource and data are available.
 */
public boolean isAvailable() {
    return isDataAvailable();
}

}
   
