/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SqlParser.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:13 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.util;

import java.io.Serializable;
import java.util.*;

/**
 * Parse a sql-92 select statement into its various clauses.
 */
public class SqlParser implements Serializable {
	
static final long serialVersionUID = 1389861757357973153L;

Vector columns = new Vector();
Vector tables = new Vector();
Vector joins = new Vector();
Vector parameterColumns = new Vector();
String non_join_where = null;

String clauses[]    = { "","","",""};
int currentClause   = 0;
int    parenCount   = 0;
String terminators[]={  "from", "where", "group", "having", "union",
    "intersect", "except", "order"};
public final static int SELECT     = 0;    // from SELECT to FROM
public final static int FROM       = 1;    // from FROM to WHERE or REST or end
public final static int WHERE      = 2;    // from WHERE to REST or end
public final static int REST       = 3;    // any of ORDER BY, GROUP BY, HAVING, etc


private final static boolean TRACE = false;

/**
 * Parse a sql 92 select statement into it component clauses: SELECT,
 * FROM, WHERE, REST.
 */
public SqlParser(String query) {
    StringTokenizer st = new StringTokenizer(query,"\n\t\r )(",true);
    while (st.hasMoreTokens()) {
        String token = st.nextToken();
        if (token.equals("(")) {
            parenCount++;
        }
        else if (token.equals(")")) {
            parenCount--;
        }
        else {
            if (!betweenParentheses() && isTerminator(token)) {
                nextClause(token);
            }
        }
        clauses[currentClause] += token;
    }

    parseSelectClause();
    parseFromClause();
    parseWhereClause();
}

/**
 * Move to the next clause SELECT -> FROM -> [WHERE] -> [REST]
 * Handle situation where there is no WHERE clause.
 * Also situation where there where there are
 * multiple REST's. That is, don't go past end of
 * array in that case.
 */
void nextClause(String token) {
    if (currentClause < REST) {
        if (!token.trim().toLowerCase().equals("where") && (currentClause == FROM)) {
            currentClause = REST;
        }
        else {
            currentClause++;
        }
    }
}

/**
 * Return true if current token is enclosed by at least one
 * set of parentheses.
 * @return true if the current token is enclosed in parentheses.
 */
boolean betweenParentheses() {
    return parenCount != 0;
}

/**
 * Return true if the current token is one of
 * SELECT, FROM, WHERE, ORDER, GROUP, HAVING etc.
 * @return true if current token is terminator.
 */
boolean isTerminator(String token) {
    token = token.trim().toLowerCase();
    for (int i = 0; i < terminators.length; i++) {
        if (token.equals(terminators[i])) {
            return true;
        }
    }
    return false;
}

/**
 * Return indicated clause.
 * @param clause an int one of SELECT, FROM, WHERE, REST.
 * @return the clause, or "" if non-existent.
 */
public String getClause(int clause) {
    return getClause(clause, true);
}


/**
 * Return indicated clause.
 * @param clause an int one of SELECT, FROM, WHERE, REST.
 * @param include_clause_word indicates if the word "SELECT", "FROM" or "WHERE"
 * 			is included in the returned string.
 * @return the clause, or "" if non-existent.
 */
public String getClause(int clause, boolean include_clause_word) {
    String temp = clauses[clause].trim();
    if ((clause != SELECT && clause != FROM && clause != WHERE) ||
        include_clause_word) {
        return temp;
    }

    String word = new String("");
    if (clause == SELECT) {
        word = "select";
    }
    else if (clause == FROM) {
        word = "from";
    }
    else if (clause == WHERE) {
        word = "where";
    }
    else {
        System.err.println("assertion failed.");
    }
    String s = new String("");
    if (temp.toLowerCase().indexOf(word) > -1) {
        s = temp.substring(word.length());
    }
    return s;
}

/** Gets a list of <code>SqlStatementColumn</code>s
 */
public Vector getColumns() {
    return columns;
}

/** Gets a list of <code>SqlStatementTable</code>s
 */
public Vector getTables() {
    return tables;
}

/** Gets a list of <code>ColumnMap</code>s
 */
public Vector getJoins() {
    return joins;
}

/** Gets a list of parameter columns
 */
public Vector getParameterColumns() {
    return parameterColumns;
}


/** Gets text of non-join where clause
 */
public String getNonJoinWhereClause() {
    return non_join_where;
}

/** Helper methods to parse the select clause
 *	into columns.
 */
void parseSelectClause() {
    String s = getClause(SELECT, false);
    if (TRACE) {
        System.out.println("parse select clause:\n" + s);
    }
    StringTokenizer token = new StringTokenizer(s, ",");
    while (token.hasMoreTokens()) {
        SqlStatementColumn col = parseColumn(token.nextToken());
        if (col != null) {
            columns.addElement(col);
        }
    }
}

/** Helper methods to parse the from clause
 *	into tables.
 */
void parseFromClause() {
    String s = getClause(FROM, false);
    if (TRACE) {
        System.out.println("parse from clause:\n" + s);
    }
    StringTokenizer token = new StringTokenizer(s, ",");
    while (token.hasMoreTokens()) {
        SqlStatementTable table = parseTable(token.nextToken());
        if (table != null) {
            tables.addElement(table);
        }
    }
}

/** Helper methods to parse the where clause
 *	into joins and parameter columns.
 */
void parseWhereClause() {
    String s = getClause(WHERE, false);
    if (TRACE) {
        System.out.println("parse where clause:\n" + s);
    }
    String upper = new String(s.toUpperCase());
    int current_index = 0;
    String token = new String("");
    String  nl = System.getProperty("line.separator");
    non_join_where = new String("");
    while (current_index < upper.length()) {
        int index = upper.indexOf("AND", current_index);

        if (index < 0) {
            token += s.substring(current_index);
            current_index = upper.length();
        }
        else {
            // Make sure AND is a word by itself, i.e. before and after
            // are white spaces.
            if (Character.isWhitespace(upper.charAt(index-1)) &&
                (upper.length() == index+3 ||
                 Character.isWhitespace(upper.charAt(index+3))
                )) {
                token = s.substring(current_index, index);
                current_index += token.length() + 3;
            }
            else {
                // AND is part of a word, continue the search.
                token += s.substring(current_index, index+3);
                current_index = index + 3;
                continue;
            }
        }
        // Find columns
        StringTokenizer st = new StringTokenizer(token, "=");
        if (st.countTokens() == 2) {
            String col1 = st.nextToken().trim();
            String col2 = st.nextToken().trim();
            if (isColumn(col1) && isColumn(col2)) {
                // Strings are columns.  Add join.
                joins.addElement(new ColumnMap(col1, col2));
            }
            else if (isParameter(col1) || isParameter(col2)) {
                parameterColumns.addElement(
                                           isParameter(col1) ? col2 : col1);
            }
            else {
                // Strings are not columns.  Append to where clause.
                if (non_join_where.length() > 0) {
                    non_join_where += nl + "AND\t";
                }
                non_join_where += token.trim();
            }
            token = new String("");
        }
        else {
            if (non_join_where.length() > 0) {
                non_join_where += nl + "AND\t";
            }
            non_join_where += token.trim();
        }
    }
}

/** Parse a string into a <code>SqlStatementTable</code>.
 *	@return SqlStatementTable
 */
public static SqlStatementTable parseTable(String text) {
    StringTokenizer st = new StringTokenizer(text);
    String  table = null,
    name = null,
    qualifier = null,
    alias = null;
    if (st.hasMoreTokens()) table = st.nextToken();
    if (st.hasMoreTokens()) {
        alias = st.nextToken();
        // Allowed syntax is TABLE [AS] ALIAS
        // Ignore AS.
        if (alias.equalsIgnoreCase("AS")) {
            alias = null;
            if (st.hasMoreTokens()) {
                alias = st.nextToken();
            }
        }
    }


    // Split table into qualifier and name
    if (table.lastIndexOf(".") > -1) {
        qualifier = table.substring(0, table.lastIndexOf("."));
        name = table.substring(table.lastIndexOf(".")+1);
    }
    else {
        name = table;
    }

    if (alias == null) {
        alias = name;
    }

    return new SqlStatementTable(name, qualifier, alias);
}

/** Parse a string into a <code>SqlStatementColumn</code>.
 *	@return SqlStatementColumn
 */
public static SqlStatementColumn parseColumn(String text) {
    StringTokenizer st = new StringTokenizer(text);
    String  column = null,
    name = null,
    qualifier = null,
    alias = null;
    if (st.hasMoreTokens()) column = st.nextToken();
    if (st.hasMoreTokens()) {
        alias = st.nextToken();
        // Allowed syntax is COLUMN [AS] ALIAS
        // Ignore AS.
        if (alias.equalsIgnoreCase("AS")) {
            alias = null;
            if (st.hasMoreTokens()) {
                alias = st.nextToken();
                // Remove double-quotes.
                if (alias.startsWith("\"")) {
                    alias = alias.substring(1);
                }
                if (alias.endsWith("\"")) {
                    alias = alias.substring(0, alias.length()-1);
                }
            }
        }
    }

    // Split column into qualifier and name
    if (column.lastIndexOf(".") > -1) {
        qualifier = column.substring(0, column.lastIndexOf("."));
        name = column.substring(column.lastIndexOf(".")+1);
    }
    else {
        name = column;
    }
    // Don't set alias if column name is "*"
    if (!name.equals("*") && alias == null) alias = name;

    return new SqlStatementColumn(name, qualifier, alias);
}

/** A string is a column if it is not a number and it is not in quotes. */
public static boolean isColumn(String s) {
    boolean v = false;

    try {
        // Not number
        Double d = Double.valueOf(s);
    }
    catch (Exception e) {
        // Not in quotes
        v = !(s.indexOf('"') > -1);
        if (v) v = !(s.indexOf("'") > -1);

        // is not parameter
        if (v) v = !isParameter(s);
    }

    return v;
}

/** A string is a parameter if it contains a question mark "?" (without quotes). */
public static boolean isParameter(String s) {
    return(s.indexOf("?") > -1);
}


/** Given an unqualified table name, return the SqlStatementTable object
 * @param name the String name of the table
 * @return SqlStatementTable, or null if not found
 */
public SqlStatementTable getSqlStatementTable(String table) {
    for (Enumeration e = tables.elements(); e.hasMoreElements();) {
        SqlStatementTable t = (SqlStatementTable) e.nextElement();
        if (t.getTableName().equals(table)) {
            return t;
        }
    }
    return null;
}

/**
 * Return the SqlStatementColumn object for this table
 * column combination. Returns null if not found.
 * @param name a String, the unqualified table name
 * @param column a String, the unqualified column name
 * @return the SqlStatementColumn for this table/column combination
 */
public SqlStatementColumn getSqlStatementColumn(String table, String column) {
    for (Enumeration e = getColumns().elements(); e.hasMoreElements();) {
        SqlStatementColumn ssc = (SqlStatementColumn) e.nextElement();
        SqlStatementTable sst = getSqlStatementTable(table);
        if (ssc.getColumnName().equals(column)) {
            if (ssc.getTableName() != null) {
                if (ssc.getTableName().equals(table) || ssc.getTableName().equals(sst.getAlias())) {
                    return ssc;
                }
            }
        }
    }
    return null;
}

/**
 * Return a Vector of SqlStatementColumns for a table
 * @param name a String, the table alias or name if alias does not exist
 * @return a Vector of SqlStatementColumns for this table
 */
public Vector getColumnsForTable(String table) {
    Vector v = new Vector();
    for (Enumeration e = getColumns().elements(); e.hasMoreElements();) {
        SqlStatementColumn ssc = (SqlStatementColumn) e.nextElement();
        if (ssc.getTableName() == null || ssc.getTableName().equals(table)) {
            v.addElement(ssc);
        }
    }
    return v;
}

}
