/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SqlStatement.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:13 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.util;

import java.io.Serializable;
import java.util.*;

/** 
 *	This is the model for an sql statement.
 *	<p>
 *	An sql statement has the form,
 *	<code>
 *	<P>
 *	SELECT 	qualifier.column(s) [AS] alias(es)<br>
 *	FROM 	qualifier.table(s) [AS] alias(es)<br>
 *	WHERE	joins between tables<br>
 *	AND		more joins between tables<br>
 *	AND		search arguments, such as table.col = 1<br>
 *	</code>
 *	<P>
 *	Columns in the select clause are automatically aliased.  
 *	Duplicate aliases are not allowed.
 */
public class SqlStatement extends Observable implements Serializable {
	
static final long serialVersionUID = -4759648418780454669L;

/** A list of <code>SqlStatementColumn</code> in the select clause in the order of
 *	the columns being added to the SQL statement.
 */
private Vector columnPositions = new Vector();
/** A hashtable of <code>SqlStatementColumn</code> key by alias.
 *	It is used in addition to columnPositions to provide fast
 *	column manipulation.
 */
private Hashtable columns = new Hashtable();
private int columnAliasCounter = 0;

/** A hashtable of <code>SqlStatementTable</code> key by alias.
 *	If alias is not defined, the table qualified name is used.
 */
private Hashtable tables = new Hashtable();
private Hashtable tableAliasCounter = new Hashtable();

private Hashtable joins = new Hashtable();  // joins between tables

private Vector parameterColumns = new Vector(); // column parameters

private String statement;               // the text of SQL statement 

/** If statement is set manually, these are the clauses. */
private String userDefinedClauses[] = new String[4];    

private SqlParser parser = null;

/** If expert mode is used.  If so, no parsing is done. */
private boolean expertMode = false;

private static final boolean TRACE = false;

public SqlStatement() {
    super();
}

public SqlStatement(String s) {
    super();
    setText(s);
}


/**
 *	Add a column name to the SELECT clause of an SQL statement.
 *	If the table of the column does not exist in the FROM clause, the table
 *	is added and the alias of the table is the table name itself.
 *	Column alias is set to the column name.  Alias is ensured to
 *	be unique.  Thus, if column name already exists in the select clause
 *	as an alias, a new alias in the format of "column" + counter is created.
 *	@param column name of the column to be added in the format of 
 *	[[schema.]table.]column or [qualifier.]column
 */
public void addColumn(String column) {
    addColumn(column, null);
}

/**
 *	Add a column name to the SELECT clause of an SQL statement.
 *	If the table of the column does not exist in the FROM clause, the table
 *	is added and the alias of the table is the table name itself.
 *	@param column name of the column to be added in the format of 
 *	[[schema.]table.]column or [qualifier.]column
 *	@param alias alias of a column
 *	Alias is ensured to be unique.  Thus, if the alias already 
 *	exists in the select clause, a new alias in the format 
 *	of "column_name" + counter is created.
 */
public SqlStatementColumn addColumn(String column, String column_alias) {
    String table = null;
    if (column.lastIndexOf(".") > -1) {
        table = column.substring(0, column.lastIndexOf("."));
        column = column.substring(column.lastIndexOf(".")+1);
    }
    return addColumn(column, column_alias, table, null);
}


/** Adds a column
 *	@param column column name
 * 	@param table qualified name of the table
 *	@param alias alias or qualifed name of a table
 */
public SqlStatementColumn addColumn(String column, String table, String alias) {
    return addColumn(column, null, table, alias);
}

/** Adds a column
 *	@param column column name
 *	@param column_alias alias of a column
 *	Alias is ensured to be unique.  Thus, if the alias already 
 *	exists in the select clause, a new alias in the format 
 *	of "column_name" + counter is created.
 *	@param table qualified name of the table
 *	@param alias alias or qualifed name of a table
 */
public SqlStatementColumn addColumn(String column, String column_alias, 
                                    String table, String alias) {
    if (column == null) return null;

    String s = (column_alias == null) ? column : 
               new String(column + " AS " + column_alias);
    SqlStatementColumn c = SqlParser.parseColumn(s);

    // if column name is not qualified, prefix it with the table
    // qualified name or table alias.
    if (alias == null) alias = table;
    if (alias != null && c.getQualifier() == null) {
        c.setQualifier(alias);
    }

    if (columns.containsKey(c.getAlias())) {
        // Alias already exists.  use another alias.
        c.setAlias(c.getAlias() + 
                   getNextColumnAliasCounter(c.getColumnName()));
    }
    if (TRACE) {
        System.out.println("Adding SQL column: " + c);
    }
    columns.put(c.getAlias(), c);
    columnPositions.addElement(c);

    if (alias != null && !tables.containsKey(alias)) {
        // If the alias is not in the FROM clause, add it.
        // Don't need to refreshText() here because addTable() will do so.
        addTable(table, alias);
    }
    else {
        refreshText();
    }

    return c;
}


/**
 *	Remove a column name from the SELECT clause of an SQL statement.
 *	@param column_alias alias name of the column
 */
public void removeColumn(String column_alias) {
    if (columns.get(column_alias) != null) {
        columnPositions.removeElement(columns.get(column_alias));
        columns.remove(column_alias);
        refreshText();
    }
}

/**
 *	Add a table name to the FROM clause of an SQL statement.
 *	Its alias equals the table name.
 *	@param table name of the table
 *	@return <code>SqlStatementTable</code> or null if table already exists.
 */
public SqlStatementTable addTable(String table) {
    return addTable(table, null);
}

/**
 *	Uniquely add a table name and its alias to the FROM clause of an SQL statement.
 *	If the alias already exists, table is not added.
 *	@param table name of the table
 *	@param table name of the table alias
 *	@return <code>SqlStatementTable</code> or null if table already exists.
 */
public SqlStatementTable addTable(String table, String alias) {
    if (table == null) return null;

    String s = (alias == null) ? table : new String(table + " " + alias);
    SqlStatementTable t = SqlParser.parseTable(s);

    return addTable(t);
}

/**
 *	Uniquely add a table name and its alias to the FROM clause of an SQL statement.
 *	If the alias already exists, table is not added.
 *	@param <code>SqlStatementTable</code> to be added.
 *	@return <code>SqlStatementTable</code> or null if table already exists.
 */
public SqlStatementTable addTable(SqlStatementTable t) {
    if (TRACE) {
        System.out.println("Adding SQL table: " + t);
    }

    if (!tables.containsKey(t.getAlias())) {
        tableAliasCounter.put(t.getQualifiedName(), new Integer(0));
    }
    // If alias is the table name, generate an alias for the table.
    else if (t.getAlias().equals(t.getTableName())) {
        // Table already exists.  use another alias.
        t.setAlias(t.getTableName() + 
                   getNextTableAliasCounter(t.getQualifiedName()));
    }
    else {
        // Table already exists
        t = null;
    }

    if (t != null) {
        if (TRACE) {
            System.out.println("Added SQL table: " + t);
        }
        tables.put(t.getAlias(), t);
        refreshText();
    }

    return t;
}

/**
 *	Remove a table name by its alias from the FROM clause of an SQL statement.
 *	Also remove any columns in the column list and where clause
 *	that reference this alias table.
 *	@param alias alias or qualified name of the table. 
 */
public void removeTable(String s) {
    SqlStatementTable table = (SqlStatementTable) tables.get(s);

    // Can't find the alias/qualified name, no table to remove.
    if (table == null) return;

    /*
     * Make a copy of the vectors to avoid halloween affect
     * while deleting elements.
     */
    Hashtable remove = (Hashtable) columns.clone();
    Hashtable delete_joins = (Hashtable) joins.clone();
    Vector delete_parameter_columns = (Vector) parameterColumns.clone();

    /*
     * remove columns associated with this table
     */
    for (Enumeration e = remove.elements(); e.hasMoreElements();) {
        SqlStatementColumn column = (SqlStatementColumn) e.nextElement();
        if (TRACE) {
            System.out.println("column: " + column);
        }
        if (column.getQualifier() != null && 
            column.getQualifier().endsWith(s)) {
            columns.remove(column.getAlias());
            columnPositions.removeElement(column);
        }
    }

    /*
     * remove parameter columns associated with this table
     */
    for (Enumeration e = delete_parameter_columns.elements(); e.hasMoreElements();) {
        String column = (String) e.nextElement();
        if (TRACE) {
            System.out.println("column: " + column);
        }
        if (column.startsWith(s)) {
            parameterColumns.removeElement(column);
        }
    }

    /*
     * delete joins associated with this table
     */
    for (Enumeration e = delete_joins.keys(); e.hasMoreElements();) {
        String key = (String) e.nextElement();
        ColumnMap cm = (ColumnMap) joins.get(key);
        if (cm.getParentColumn().startsWith(s + ".") ||
            cm.getChildColumn().startsWith(s + ".")) {
            joins.remove(key);
        }
    }

    tables.remove(s);

    /*
     * update the display
     */
    refreshText();
}

/** Add a join between two columns.  */
public void addJoin(String parent_column, String child_column) {
    // FIX
    // 1.	Only inner join is handled right now.
    // 2.	id = prod_id is considered different from 
    //		product.id = sales_order_items.id even though they are identical
    String s1 = parent_column + " = " + child_column;
    String s2 = child_column + " = " + parent_column;
    if (!(joins.containsKey(s1) || joins.containsKey(s2))) {
        joins.put(s1, new ColumnMap(parent_column, child_column));
        refreshText();
    }
}

/** Remove a join between two columns. */
public void removeJoin(String parent_column, String child_column) {
    String s1 = parent_column + " = " + child_column;
    String s2 = child_column + " = " + parent_column;
    if (joins.containsKey(s1)) {
        joins.remove(s1);
        refreshText();
    }
    else if (joins.containsKey(s2)) {
        joins.remove(s2);
        refreshText();
    }
}

/** Remove all joins of a table. 
 *	@param alias alias or qualifed name of the table
 */ 
public void removeJoin(String alias) {
    boolean removed = false;
    Enumeration e = joins.elements();
    while (e.hasMoreElements()) {
        ColumnMap cm = (ColumnMap) e.nextElement();
        if (cm.getParentColumn().startsWith(alias + ".") ||
            cm.getChildColumn().startsWith(alias + ".")) {
            joins.remove(cm.getParentColumn() + " = " + cm.getChildColumn());
            removed = true;
        }
    }
    if (removed) {
        refreshText();
    }
}

/**
 *	Add a parameter column to the WHERE clause of an SQL statement.
 *	@param column name of the column to be added in the format of 
 *	[[schema.]table.]column or [qualifier.]column.
 */
public void addParameterColumn(String col) {
    parameterColumns.addElement(col);
    refreshText();
}

/**
 *	Remove a parameter column from the WHERE clause of an SQL statement.
 *	@param column name of the column to be deleted in the format of 
 *	[[schema.]table.]column or [qualifier.]column.
 */
public void removeParameterColumn(String col) {
    if (parameterColumns.contains(col)) {
        parameterColumns.removeElement(col);
        refreshText();
    }
}

/**
 *	Remove all parameter columns from the WHERE clause of an SQL statement.
 */
public void removeAllParameterColumns() {
    if (parameterColumns.size() > 0) {
        parameterColumns.removeAllElements();
        refreshText();
    }
}

/**
 *	The SQL statement has changed, show the
 * 	statement with the new changes.
 */
public void refreshText() {
    if (isExpertMode()) {
        setChanged();
        return;
    }

    int cnt = 0;
    String  nl = System.getProperty("line.separator"),
                 select = new String(""), 
                          from = new String(""), 
                                 where = new String("");
    statement = "SELECT\t";
    for (Enumeration e = columnPositions.elements(); e.hasMoreElements();) {
        select += (cnt++ == 0 ? "" : "," + nl + "\t");
        SqlStatementColumn c = (SqlStatementColumn) e.nextElement();
        if (c.isTableColumn()) {
            select += c.getQualifiedName();
            // Do not add alias if alias = column qualified name.
            if (c.getAlias() != null && 
                !c.getAlias().equals(c.getQualifiedName())) {
                select += " AS \"" + c.getAlias() + "\"";
            }
        }
        else {
            // Complex column.  Append to select
            select += c.getText();
        }
    }
    if (cnt > 0) select = "SELECT\t" + select;

    cnt = 0;
    for (Enumeration e = tables.elements(); e.hasMoreElements();) {
        from += (cnt++ == 0 ? "" : "," + nl + "\t");
        SqlStatementTable t = (SqlStatementTable) e.nextElement();
        if (!t.isSubQuery()) {
            from += t.getQualifiedName();
            // Do not add alias if alias = table qualified name.
            if (t.getAlias() != null && 
                !t.getAlias().equals(t.getQualifiedName())) {
                from += " " + t.getAlias(); 
            }
        }
        else {
            // subquery.  Append
            from += t.getSubQuery();
        }
    }
    if (cnt > 0) from = "FROM\t" + from;

    cnt = 0;
    for (Enumeration e = joins.elements(); e.hasMoreElements();) {
        where += (cnt++ == 0 ? "" : nl + "AND\t");
        ColumnMap cm = (ColumnMap) e.nextElement();
        where += cm.getParentColumn() + " = " + cm.getChildColumn();
    }
    for (Enumeration e = parameterColumns.elements(); e.hasMoreElements();) {
        where += (cnt++ == 0 ? "" : nl + "AND\t");
        String col = (String) e.nextElement();
        where += col + " = ?";
    }
    if (cnt > 0) {
        where = "WHERE\t" + where;
        if (userDefinedClauses[SqlParser.WHERE].length() > 0) {
            where += nl + "AND\t" + userDefinedClauses[SqlParser.WHERE];
        }
    }
    else if (userDefinedClauses[SqlParser.WHERE].length() > 0) {
        where = "WHERE\t" + userDefinedClauses[SqlParser.WHERE];
    }

    statement = select + nl + from + nl + where + nl + userDefinedClauses[SqlParser.REST];

    // get rid of spurious newlines
    String trimmed = statement.trim();
    if (trimmed.length() == 0) {
        statement = trimmed;
    }

    if (TRACE) {
        System.out.println("\n\n\n=========================\n\n\n");
        System.out.println("SQL Statement:\n" + statement);
        System.out.println("\n\n\n=========================\n\n\n");
    }

    // Indicates statement is changed.
    setChanged();
}

/**
 * Return the sql statement for this level.
 */
public String getText() {
    return statement;
}

/** Sets the text of the actual SQL statement. 
 *	The SQL statement is not converted back to joins, columns and tables.
 */
public void setText(String s) {
    if (isExpertMode()) {
        statement = s;
        setChanged();
        return;
    }

    clear();

    if (s != null) {
        parser = new SqlParser(s);

        String where = parser.getClause(SqlParser.WHERE, false);
        userDefinedClauses[SqlParser.REST] = parser.getClause(SqlParser.REST);

        // Convert from clause before select to add tables first.
        for (Enumeration e = parser.getTables().elements(); e.hasMoreElements();) {
            SqlStatementTable table = (SqlStatementTable) e.nextElement();
            addTable(table);
        }
        for (Enumeration e = parser.getColumns().elements(); e.hasMoreElements();) {
            SqlStatementColumn column = (SqlStatementColumn) e.nextElement();
            columns.put(column.getAlias(), column);
            columnPositions.addElement(column);
        }
        for (Enumeration e = parser.getJoins().elements(); e.hasMoreElements();) {
            ColumnMap cm = (ColumnMap) e.nextElement();
            addJoin(cm.getParentColumn(), cm.getChildColumn());
        }
        for (Enumeration e = parser.getParameterColumns().elements(); e.hasMoreElements();) {
            String col = (String) e.nextElement();
            addParameterColumn(col);
        }       
        userDefinedClauses[SqlParser.WHERE] = parser.getNonJoinWhereClause();

        refreshText();
    }
}



/**
 * Return a list of tables associated with this statement.
 * @return a list of the tables this statement.
 */
public Enumeration getTables() {
    return tables.elements();
}

/**
 * Return list of columns associated with this statement.
 * @return a list of the columns in this statement.
 */
public Enumeration getColumns() {
    return columns.elements();
}

/**
 * Return list of joins associated with this statement.
 * @return a list of the joins in this statement.
 */
public Enumeration getJoins() {
    return joins.elements();
}

/** Clear various vectors and user-defined clauses.  */
public void clear() {
    columns = new Hashtable();
    columnPositions = new Vector();
    tables = new Hashtable();
    joins = new Hashtable();
    parameterColumns = new Vector();
    columnAliasCounter = 0;
    tableAliasCounter = new Hashtable();
    userDefinedClauses = new String[] { "", "", "", ""};
    refreshText();
}

/** Indicates if expert mode is used.  
 *	No parsing is done in expert mode.
 */
public boolean isExpertMode() {
    return expertMode;
}

/** Sets to use expert mode or not.  
 *	No parsing is done in expert mode.
 */
public void setExpertMode(boolean v) {
    if (expertMode == v) return;

    expertMode = v;
    setChanged();
}

private int getNextColumnAliasCounter(String column_name) {
    // alias may be generated before and thus it contains 
    // a counter at the end. Make sure the new alias created is 
    // unique.
    for (Enumeration e = columns.keys(); e.hasMoreElements();) {
        String s = (String) e.nextElement();
        if (s.startsWith(column_name)) {
            String count_str = s.substring(
                                          s.indexOf(column_name) +
                                          column_name.length());
            try {
                int count = (count_str.length() == 0) ? 0 : 
                            new Integer(count_str).intValue();
                columnAliasCounter = Math.max(count, columnAliasCounter);
            }
            catch (Exception ex) {
            }
        }
    }
    return ++columnAliasCounter;
}       

private int getNextTableAliasCounter(String table_qualified_name) {
    // alias may be generated before and thus it contains 
    // a counter at the end. Make sure the new alias created is 
    // unique.
    Integer counter = (Integer) tableAliasCounter.get(table_qualified_name);
    tableAliasCounter.put(table_qualified_name, new Integer(counter.intValue()+1));
    return counter.intValue()+1;
}       

}
