/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SqlStatementColumn.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:13 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.datasource.util;

import java.io.Serializable;
import java.util.StringTokenizer;

/**	Describes a column in the select clause of a SQL statement.
 *	It contains qualifier, alias and name of the column.
 *	It also indicates if the column is a table column, i.e.
 *	the column refers to a single column in a table.
 *	If it is not a table column, the text is stored.
 *	If alias is not specified, it is default to its qualified name.
 *
 */
public class SqlStatementColumn implements Serializable {
	
static final long serialVersionUID = 7083424415349465593L;

private String alias;
private String column;
private String qualifier;
private String catalog;
private String schema;
private String table;

/** Indicates this is a table column in the format of [qualifier.]column.
 *	If it is false, the column can be a subquery, a value expression,
 *	or an aggregate function.
 */
private boolean isTableColumn = true;
/** Text for a subquery, a value expression,
 *	or an aggregate function.
 */
private String text;

/** Constructs a column.
 */
public SqlStatementColumn(String name, String qualifier, String alias) {
    this.column = name;
    this.qualifier = qualifier;
    setAlias(alias);
    parseQualifier();
}

/** Gets the column name without qualifier. */
public String getColumnName() {
    return column;
}

/** Gets the column qualifier, i.e. prefix of column name. */
public String getQualifier() {
    return qualifier;
}

/** Sets the column qualifier, i.e. prefix of column name. */
public void setQualifier(String s) {
    String old_qualified_name = getQualifiedName();
    qualifier = s;
    if (alias == old_qualified_name) {
        setAlias(getQualifiedName());
    }
}

/** Gets the qualified column name, i.e. [qualifier].column. */
public String getQualifiedName() {
    return(qualifier == null) ? column :
    qualifier + "." + column;
}

/** Gets the column alias explicited defined after the keyword "AS" */
public String getAlias() {
    return alias;
}

/** Sets the column alias explicited defined after the keyword "AS" */
public void setAlias(String s) {
    if (column == "*") {
        alias = getQualifiedName();
    }
    else {
        alias = (s == null) ? getQualifiedName() : s;
    }
}

public boolean isTableColumn() {
    return isTableColumn;
}

/** Sets isTableColumn property.
 *	Indicates this is a table column in the format of [qualifier.]column.
 *	If it is false, the column can be a subquery, a value expression,
 *	or an aggregate function.
 */
public void setIsTableColumn(boolean v) {
    this.isTableColumn = v;
}

/** Sets the text for a subquery, a value expression,
 *	or an aggregate function.
 */
public void setText(String s) {
    this.text = s;
}

/** Gets the text for a subquery, a value expression,
 *	or an aggregate function.
 */
public String getText() {
    return text;
}

public String toString() {
    return new String("[column name: " + column +
                      " qualifier: " + qualifier +
                      " alias: " + alias +
                      " isTableColumn: " + isTableColumn +
                      " text: " + text + "]\n");
}

/**
 * Parse a qualifier into it parts which can be
 * catalog.schema.table or,
 * schema.table or
 * table/alias
 */
void parseQualifier() {
    String qualifier = getQualifier();
    if (qualifier == null) {
        return;
    }
    // reverse the tokens
    StringBuffer buffer = new StringBuffer();
    StringTokenizer st = new StringTokenizer(qualifier,".", true);
    while (st.hasMoreTokens()) {
        buffer.insert(0, (String) st.nextToken());
    }
    st = new StringTokenizer(buffer.toString(),".",false);
    if (st.hasMoreTokens()) table   = (String) st.nextToken();
    if (st.hasMoreTokens()) schema  = (String) st.nextToken();
    if (st.hasMoreTokens()) catalog = (String) st.nextToken();
}
/** Return the catalog name, or null if it doesn't exist */
public String getCatalogName() {
    return catalog;
}

/** Return the schema name, or null if it doesn't exist */
public String getSchemaName() {
    return schema;
}

/** Return the table name or alias, or null if it doesn't exist */
public String getTableName() {
    return table;
}

}
