/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: DataPropertiesRegistry.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:14 $ $Locker:  $ Quest Software Inc.

package com.klg.jclass.field;


import java.util.HashMap;



import com.klg.jclass.util.value.JCValueModel;
import com.klg.jclass.util.value.MutableValueModel;
import com.klg.jclass.field.validate.JCValidator;

/**
 * Given a data type, this class determines the default validator or value model.
 * If you create your own validator, you can register it using the AddEntry()
 * method.
 */
public class DataPropertiesRegistry {

// default entries
protected static String default_table[][] = {
	{ "com.klg.jclass.util.JCIPAddress", "com.klg.jclass.field.validate.JCIPAddressValidator", null },
	{ "java.lang.Byte", "com.klg.jclass.field.validate.JCByteValidator", null },
	{ "java.lang.Double", "com.klg.jclass.field.validate.JCDoubleValidator", null },
	{ "java.lang.Float", "com.klg.jclass.field.validate.JCFloatValidator", null },
	{ "java.lang.Integer", "com.klg.jclass.field.validate.JCIntegerValidator", null },
	{ "java.lang.Long", "com.klg.jclass.field.validate.JCLongValidator", null },
	{ "java.lang.Short", "com.klg.jclass.field.validate.JCShortValidator", null },
	{ "java.lang.String", "com.klg.jclass.field.validate.JCStringValidator", null },
	{ "java.lang.StringBuffer", "com.klg.jclass.field.validate.JCStringValidator", null },
	{ "java.math.BigDecimal","com.klg.jclass.field.validate.JCBigDecimalValidator", null},
	{ "java.util.Calendar", "com.klg.jclass.field.validate.JCDateTimeValidator", null },
	{ "java.util.GregorianCalendar", "com.klg.jclass.field.validate.JCDateTimeValidator", null },
	{ "java.util.Date", "com.klg.jclass.field.validate.JCDateValidator", null },
	{ "java.sql.Date", "com.klg.jclass.field.validate.JCDateValidator", null },
	{ "java.sql.Time", "com.klg.jclass.field.validate.JCTimeValidator", null },
	{ "java.sql.Timestamp", "com.klg.jclass.field.validate.JCDateTimeValidator", null }
};

protected static HashMap propertiesTable = new HashMap(default_table.length);;


static {
	// install class defaults when the class is loaded
	for (int i = 0; i < default_table.length; i++) {
		addEntry(default_table[i][0], default_table[i][1],
				 default_table[i][2]);
	}
}


/**
 * Returns true if the object is registered.
 */
public static boolean isRegistered(Class c) {
	if (c == null) {
		return false;
	}
	return propertiesTable.containsKey(c.getName());
}

/**
 * Adds a new entry to the DataProperties registry.
 * @param classname Full class name of the data type. i.e. (java.lang.String)
 * @param validator Full class name of the validator.
 * @param value_model Full class name of the data model.
 */
public static void addEntry(String classname, String validator, String value_model) {
	propertiesTable.remove(classname);
	propertiesTable.put(classname, new Info(validator, value_model));
}

/**
 * Returns an Info class given the class name. If no entry
 * can be found corresponding to the class name, null is returned.
 * @see com.klg.jclass.field.DataPropertiesRegistry.Info
 */
public static Info getEntry(String classname) {
	return (Info)propertiesTable.get(classname);
}

/**
 * Removes an entry from the DataPropertiesRegistry which corresponds to
 * the class name specified.
 */
public static void removeEntry(String classname) {
	propertiesTable.remove(classname);
}

/**
 * Creates a value mode based on the class passed in.
 */
public static JCValueModel createValueModel(Class value_class) {
	Info dinfo = getEntry(value_class.getName());

	if (dinfo == null) {
		return null;
	}

	String model_class_name = dinfo.getValueModel();
	if (model_class_name == null) {
		model_class_name = "com.klg.jclass.util.value.MutableValueModel";
	}

	JCValueModel model_instance = null;

	// create the instance of the validator and data model
	try {
		Class model_class = Class.forName(model_class_name);
		model_instance =
			(JCValueModel) model_class.newInstance();
		if (model_instance instanceof MutableValueModel) {
			((MutableValueModel) model_instance).setValueClass(value_class);
		}
	}
	catch (ClassNotFoundException cnfe) {
		throw new IllegalArgumentException("Cannot find class \"" +
										   model_class_name + "\"");
	}
	catch (InstantiationException e) {
		throw new IllegalArgumentException("Cannot instantiate validator class \"" + model_class_name + "\"");
	}
	catch (IllegalAccessException e) {
		throw new IllegalArgumentException("Cannot instantiate validator class \"" + model_class_name + "\"");
	}

	return (JCValueModel) model_instance;
}


/**
 *	Creates an instance of the default validator for the given data_type.
 */
public static JCValidator createValidator(Class value_class) {
	Info dinfo = getEntry(value_class.getName());

	if (dinfo == null) {
		return null;
	}

	String validator_class_name = dinfo.getValidator();
	JCValidator validator_instance = null;

	// create the instance of the validator and data model
	try {
		Class validator_class = Class.forName(validator_class_name);
		validator_instance =
			(JCValidator)validator_class.newInstance();
	}
	catch (ClassNotFoundException cnfe) {
		throw new IllegalArgumentException("Cannot find class \"" +
										   validator_class_name + "\"");
	}
	catch (InstantiationException e) {
		throw new IllegalArgumentException("Cannot instantiate validator class \"" + validator_class_name + "\"");
	}
	catch (IllegalAccessException e) {
		throw new IllegalArgumentException("Cannot instantiate validator class \"" + validator_class_name + "\"");
	}

	return (JCValidator)validator_instance;
}


/**
  * Convenience class used to hold a validator and data model.
  * Returned from getEntry() in the class DataPropertiesRegistry.
  */
public static class Info {

	protected String validator;
	protected String value_model;

	/**
	 * Creates a DataProperties entry with the specifed validator and
	 * data model.
	 */
	public Info(String validator, String value_model) {
		this.validator = validator;
		this.value_model = value_model;
	}

	/**
	 * Returns the validator class name.
	 */
	public String getValidator() {
		return validator;
	}

	/**
	 * Returns the data model class name.
	 */
	public String getValueModel() {
		return value_model;
	}

} // Info

}
