/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: AbstractValidator.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:17 $ $Locker:  $ Quest Software Inc.

package com.klg.jclass.field.validate;

import java.util.Locale;
import java.util.ResourceBundle;
import java.io.Serializable;
import javax.swing.ListModel;

/**
 * A base class for all validators.
 */
public abstract class AbstractValidator extends Object
                                        implements JCValidator, Serializable
{

// Enum For CasePolicy
/**
 * Leaves characters as entered in the CasePolicy property.
 */
public static final int AS_IS = 0;
/**
 * Converts characters to upper case in the CasePolicy property.
 */
public static final int UPPERCASE = 1;
/**
 * Converts characters to lower case in the CasePolicy property.
 */
public static final int LOWERCASE = 2;

// Enum for SpinPolicy
/**
 * Allows spinning up and down between the maximum and minimum
 * values. (Default for numeric validators.)
 */
public static final int SPIN_FIELD = 0;
/**
 * Allows context sensitive spinning if it is allowed. (Default for date and
 * time validators.)
 */
public static final int SPIN_SUBFIELD = 1;
/**
 * Like spin field but allows continuous spinning. A value wraps from its
 * maximum value to its minimum when spinning in the up direction, and
 * from minimum to maximum in the other direction. (Default for String
 * validators.)
 */
public static final int SPIN_WRAP = 2;

public static final int NOT_FOUND = -1;

// Private storage for properties
protected Object  max;
protected Object  min;
protected Locale  locale;
protected Object  increment;

protected boolean    allow_null = false;
protected int        case_policy = AS_IS;
protected String     invalid_chars = null;
protected boolean    match_picklist = true;
protected String     valid_chars = null;
protected ListModel  picklist = null;
protected int        spin_policy = SPIN_FIELD;
protected boolean    abstractInitializing = true;
protected Class[]    supportedClasses;

protected ResourceBundle li; // Our LocaleInfo bundle

public AbstractValidator()
{
    setLocale(Locale.getDefault());
    abstractInitializing = false;
}

/**
 * Changes the specified text from the start_position to the end_position by
 * inserting/replacing it by rge specified new_text.
 */
public void changeText(TextEvent e)
{
    String text = e.getText();
    if (text == null || text.length() == 0) {
        return;
    }

    StringBuffer newtext = new StringBuffer(text.length());

    for (int i = 0; i < text.length(); i++) {
        newtext.append(convertCase(text.charAt(i)));
    }

    e.setText(newtext.toString());
}

/**
 * Sets the cursor to the specified position.
 */
abstract public void setCursor(TextCursorEvent e);

/**
 * Validates the current display string and returns a new Object if
 * successful, or "null" if validation fails.
 * @param text text to be converted to an object
 */
abstract public Object parse(Class c, String text) throws FieldParseException;

/**
 * Converts the specified type (if possible) to the type supported by
 * the validator.  To reverse this operation use the convertFromSupported()
 * method
 * @see #convertFromSupported
 */
abstract public Object convertToSupported(Object obj);

/**
 * Converts the supported type to the type supported by the validator.
 * @see #convertToSupported
 */
abstract public Object convertFromSupported(Object from, Class class_type);

/**
 * Validates the specified internal object and returns the Object if
 * successful, or "null" if validation fails.
 * @param obj object to be checked
 */
public boolean validate(Object obj)
{
    boolean is_valid = true;
    if (picklist != null && picklist.getSize() > 0 &&
        match_picklist == true) {
        int list_index = getPickListIndex(obj);
        if (list_index == -1) {
            is_valid = false;
        }
        else {
            is_valid = true;
        }
    }

    if (allow_null != true) {
        if (obj == null) {
            is_valid = false;
        }
    }
    return(is_valid);
}

/**
 * Since the clone() method is protected in the Object class we cannot
 * actually clone a generic object without knowing what it is; therefore
 * the validator must provide this routine to do the copying for JCField.
 */
abstract public Object copyValue(Object value);

/**
 * Since we cannot always rely on an objects equals() method we must
 * define our own compare method.  This default implementation
 * calls equals().  If this is not sufficient then the subclass can
 * override this method without having to subclass the data object.
 */
public boolean compareValues(Object v1, Object v2)
{
    if (v1 == v2) {
        return(true);
    }
    if (v1 == null) {
        // We know that v2 != null since v1 != v2 from the previous
        // check
        return(false);
    }
    return(v1.equals(v2));
}

/**
 * Generates a String representation of the given object.
 */
abstract public String format(Object obj);

/**
 * Generates a String representation of the given object for use in edit mode.
 */
abstract public String formatForEdit(Object obj);

/**
 * Returns true if the validator has a different edit format.
 */
abstract public boolean hasEditFormat();

/**
 * Reports the first valid cursor position.
 */
abstract public int getFirstValidCursorPosition();

/**
 * Returns whether or not null is a valid value.
 */
public boolean getAllowNull()
{
    return(allow_null);
}

/**
 * Sets whether or not a null value is supplied.
 */
public void setAllowNull(boolean allow_null)
{
    this.allow_null = allow_null;
}

/**
 * Returns the spin policy.
 * @see #setSpinPolicy
 */
public int getSpinPolicy()
{
    return(spin_policy);
}

/**
 * Sets the value of SpinPolicy.
 * <table border>
 * <tr><th>Value</th>    <th>Meaning</th></tr><tr><td>
 * SPIN_FIELD </td><td> Default value. Allow spinning up and down between
 * the maximum and minimum</td></tr><tr><td>
 * SPIN_SUBFIELD </td><td> Allow context sensitive spinning if it is allowed
 * </td></tr><tr><td>
 * SPIN_WRAP </td><td> Like spin field but allows spinning from max to min
 *  and min to max </td></tr>
 * </table><p>
 *
 */
public void setSpinPolicy(int spin_policy)
{
    if (spin_policy != SPIN_SUBFIELD &&
        spin_policy != SPIN_FIELD &&
        spin_policy != SPIN_WRAP)
    {
        throw new IllegalArgumentException("invalid value for spinPolicy");
    }

    this.spin_policy = spin_policy;
}

/**
 * Infer the sub field we need to work on from the current selected region
 * If there is no selected region then the selected region is equal to the
 * cursor position we don't ever need to look at the cursor position
 *
 * This particular implementation is a stub meant to be overridden if a
 * particular validator supports subfields
 */
public void inferSubField(int s, int e) {}

/**
 * Returns the Locale used for formatting and parsing.
 */
public Locale getLocale()
{
    return(locale);
}

/**
 * Sets the Locale used for formatting and parsing.
 */
public void setLocale(Locale locale)
{
    if (this.locale == locale) {
        return;
    }

    // get default locale if no locale is specified
    if (locale == null) {
        locale = Locale.getDefault();
    }

    String bundleName = "com.klg.jclass.field.resources.LocaleInfo";
    try {
        li = ResourceBundle.getBundle(bundleName, locale);
    }
    catch (java.util.MissingResourceException e1) {
        // Some implementations of the JDK throw an exception if the exact
        // locale is not present so we need to handle search for a parent
        // locale ourselves
        locale = new Locale(locale.getLanguage(), locale.getCountry());
        try {
            li = ResourceBundle.getBundle(bundleName, locale);
        }
        catch (java.util.MissingResourceException e2) {
            locale = new Locale(locale.getLanguage(), "");
            try {
                li = ResourceBundle.getBundle(bundleName, locale);
            }
            catch (java.util.MissingResourceException e3) {
                locale = null;
                try {
                    li = ResourceBundle.getBundle(bundleName, locale);
                }
                catch (java.util.MissingResourceException e4) {
                    // We REALLY didn't find any locale information
                    throw e1;
                }
            }
        }
    }
    this.locale = locale;
}

/**
 * Returns the value of the casePolicy property.
 * @see #setCasePolicy
 */
public int getCasePolicy()
{
    return(case_policy);
}

/**
 * Sets the value of the casePolicy property.
 * @param v AS_IS, UPPERCASE, LOWERCASE
 */
public void setCasePolicy(int v)
{
    if (v != AS_IS && v != UPPERCASE && v != LOWERCASE) {
        throw new IllegalArgumentException("invalid value for casePolicy");
    }
    case_policy = v;
}

/**
 * Returns the value of the character as filtered by the casePolicy of the
 * validator.
 * @param c value whose case is to be checked
 */
public char convertCase(char c)
{
    switch (case_policy) {
    case AS_IS:
        break;
    case UPPERCASE:
        c = Character.toUpperCase(c);
        break;
    case LOWERCASE:
        c = Character.toLowerCase(c);
        break;
    }
    return(c);
}

/**
 * Gets the pick list.
 */
public ListModel getPickList()
{
    return(picklist);
}

/**
 * Sets the pick list.
 */
public void setPickList(ListModel picklist)
{
    this.picklist = picklist;
}

/**
 * Checks to see if a given object is in the pick list.
 * @return returns the index into the list or NOTFOUND on failure
 */
public int getPickListIndex(Object obj)
{
    if (picklist == null || obj == null) {
        return(NOT_FOUND);
    }

    for (int i = 0; i < picklist.getSize(); i++) {
        if (obj.equals(picklist.getElementAt(i))) {
            return (i);
        }
    }
    return(NOT_FOUND);
}

/**
 * Returns the value of matchPickList.
 * @see #setMatchPickList
 */
public boolean getMatchPickList()
{
    return(match_picklist);
}

/**
 * Sets the value of matchPickList, which controls whether values must match
 * those in the PickList of the validator or not.  The default is "true" but
 * this property is only applicable if there actually are elements stored in
 * the PickList
 */
public void setMatchPickList(boolean v)
{
    match_picklist = v;
}

/**
 * Returns the next value in the pickList or "null" if we are already at the
 * end of the list.
 * @param current_index index of the current item in the picklist
 */
protected Object getNextValue(int current_index)
{
    if ((current_index + 1) >= 0
        && (current_index + 1) >= picklist.getSize()) {
        // if current index is not in range
        return (null);
    }
    return(picklist.getElementAt(current_index + 1));
}

/**
 * Returns the previous value in the pickList or "null" if we are already at
 * the beginning of the list.
 * @param current_index index of the current item in the picklist
 */
protected Object getPreviousValue(int current_index)
{
    if ((current_index - 1) < 0) {
        return (null);
    }
    return(picklist.getElementAt(current_index - 1));
}

/**
 * Returns a String of characters that are allowed to be typed in the
 * current field.
 * @see #setValidChars
 * @see #isValidChar
 */
public String getValidChars()
{
    return(valid_chars);
}

/**
 * Sets a String of characters which are allowed to be input in the
 * current field.
 * @see #isValidChar
 */
public void setValidChars(String v)
{
    valid_chars = v;
}

/**
 * Returns a String of characters that are not allowed to be typed in the
 * current field.
 * @see #setInvalidChars
 * @see #isValidChar
 */
public String getInvalidChars()
{
    return(invalid_chars);
}

/**
 * Sets a String of characters that are not allowed to be input in the
 * current field.
 * @see #isValidChar
 */
public void setInvalidChars(String v)
{
    invalid_chars = v;
}

/**
 * Determines whether the character specified is allowed in the field.  It
 * does this by first searching the InvalidChars and then the ValidChars strings.
 * @see #setInvalidChars
 * @see #setValidChars
 */
public boolean isValidChar(char c)
{
    if (invalid_chars != null) {
        // Check against invalid chars
        for (int i = 0; i < invalid_chars.length(); i++) {
            if (c == invalid_chars.charAt(i)) {
                return(false);
            }
        }
    }

    if (valid_chars == null || valid_chars.length() == 0) {
        return(true);
    }

    // Check against valid chars
    for (int i = 0; i < valid_chars.length(); i++) {
        if (c == valid_chars.charAt(i)) {
            return(true);
        }
    }
    // Valid char search failed so we return false
    return(false);
}

//
// Range Specific Control Methods
//

/**
 * Returns the maximum possible value of the object checked by the validator.
 * @see #setMax
 */
public Object getMax()
{
    return(max);
}

/**
 * Sets the maximum possible value of the object checked by the validator.
 * @see #setRange
 * @see #setMin
 */
public void setMax(Object max)
{
    this.max = max;
}

/**
 * Returns the minimum possible value of the object checked by the validator.
 * @see #setMin
 */
public Object getMin()
{
    return(min);
}

/**
 * Sets the minimum possible value of the object checked by the validator.
 * @see #setRange
 * @see #setMax
 */
public void setMin(Object min)
{
    this.min = min;
}

/**
 * Sets the max and min properties simultaneously.
 */
public void setRange(Object min, Object max)
{
    this.min = min;
    this.max = max;
}

abstract protected boolean inRange(Object value);

//
// List Specific Control Methods
//

/**
 * Gets a String representation of the object at the specified position in
 * to pick list.  This is a function since subclasses may override the simple
 * toString() conversion that the base functionality specifies.
 */
protected String getItem(int i)
{
    return(picklist.getElementAt(i).toString());
}

protected abstract Object addIncrement(Object current);

protected abstract Object subtractIncrement(Object current);

/**
 * If the validator has the ability to "spin" then this returns the next
 * value.
 */
public Object spinUp(Object current)
{
    int position;

    if (picklist != null && picklist.getSize() > 0) {
        // Do the picklist thing
        position = getPickListIndex(current);

        if ((position + 1) < picklist.getSize()) {
            return(picklist.getElementAt(position + 1));
        }
        else if (spin_policy == SPIN_WRAP) {
            // Wrap to the first element
            return(picklist.getElementAt(0));
        }
        else {
            return(null);
        }
    }

    // There was no picklist so attempt the increment
    if (increment != null) {
        Object value;
        value = addIncrement(current);
        if (value != null) {
            return(value);
        }
        else if (spin_policy == SPIN_WRAP) {
            // Spin around to min; set value to max
            return(min);
        }
    }
    return(null);
}

/**
 * If the validator has the ability to "spin" then this gives us the previous
 * value.
 */
public Object spinDown(Object current)
{
    int position;

    if (picklist != null && picklist.getSize() > 0) {
        // Do the picklist thing
        position = getPickListIndex(current);

        if (position > 0) {
            return(picklist.getElementAt(position - 1));
        }
        else if (position == NOT_FOUND) {
            return(picklist.getElementAt(picklist.getSize() - 1));
        }
        else if (spin_policy == SPIN_WRAP) {
            // Wrap to the last element
            return(picklist.getElementAt(picklist.getSize() - 1));
        }
        else {
            return(null);
        }
    }

    // There was no picklist so attempt the increment
    if (increment != null) {
        Object value;
        value = subtractIncrement(current);
        if (value != null) {
            return(value);
        }
        else if (spin_policy == SPIN_WRAP) {
            // Spin around to max; set value to max
            return(max);
        }
    }
    return(null);
}

/**
 * Determines if the value in the field can be spun further.
 */
public int calculateSpinability(Object current)
{
    if (current != null && !isClassSupported(current.getClass())) {
        return(0);
    }
    int mask = 0;
    // do the picklist thing
    if (picklist != null && picklist.getSize() > 0) {
        int position = getPickListIndex(current);

        // If the current string is not found in the list then position is set
        // to NOT_FOUND == -1 and we use that we can spin up to the 0th item
        // and down to the n-1th item

        if (spin_policy == SPIN_WRAP) {
            // we want to wrap so any thing is allowed
            // as long as we have picklist items
            mask |= DOWN;
            mask |= UP;
        }
        else {
            if (position > 0 || position == NOT_FOUND) {
                mask |= DOWN;
            }

            if (position < (picklist.getSize() - 1)) {
                mask |= UP;
            }
        }

        return(mask);
    }

    if (increment != null) {
        Object val = subtractIncrement(current);
        if (spin_policy == SPIN_WRAP) {
            // we want to wrap so anything is allowed
            // as long as we have a valid increment
            mask |= DOWN;
            mask |= UP;
        }
        else {
            if (val != null && !val.equals(current)) {
                mask |= DOWN;
            }

            val = addIncrement(current);
            if (val != null && !val.equals(current)) {
                mask |= UP;
            }
        }
        return(mask);
    }

    return(0);
}

/**
 * Given a partially complete string, this routine looks through the list and
 * constructs a String that matches as much as possible.
 * @return a more (or the same) complete string or null if it does not match
 * at all
 */
//abstract String complete(String partial);


protected StringBuffer checkValidInvalid(String text)
{
    StringBuffer new_text = new StringBuffer(text.length());

    for (int i = 0; i < text.length(); i++) {
        char c = text.charAt(i);
        if (isValidChar(c)) {
            new_text.append(c);
        }
    }
    return(new_text);
}

/**
 * Returns true if the class passed in is the native type that the
 * validator is designed to support.
 */
public boolean isClassSupported(Class c)
{
    if (supportedClasses == null) {
        return(false);
    }

    // check to see if the class is a supported by the validator
    for (int i = 0; i < supportedClasses.length; i++) {
        if (c == supportedClasses[i]) {
            return(true);
        }
    }

    return(false);
}

/**
 * Returns the popup component.
 */
public PopupFieldEditor createPopupComponent()
{
    return(null);
}

}
