/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCNumberValidator.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:17 $ $Locker:  $ Quest Software Inc. 

package com.klg.jclass.field.validate;

import com.klg.jclass.field.resources.LocaleInfo;

import java.util.*;
import java.text.*;

/**
 * This class is a base class for numeric validators.
 */
abstract public class JCNumberValidator extends AbstractValidator {

protected DecimalFormat         displayFormat;
protected DecimalFormat         editFormat;
protected DecimalFormatSymbols  symbols;
protected DecimalFormatSymbols  currencySymbols;
protected String                display_pattern = null;
protected String                edit_pattern    = null;
protected int                   spin_policy = SPIN_FIELD;
protected String                exponentialChar = null;
protected boolean               isCurrency = false;
protected Locale                currency_locale = null;
protected boolean               use_intl_currency_symbol = false;

public JCNumberValidator()
{
    super();

    // Invoke concrete class to specifiy the patterns
    setPatterns();

    displayFormat = createDisplayFormat(getLocale());
    if (display_pattern != null) {
        displayFormat.applyPattern(display_pattern);
    }
    editFormat = createEditFormat(getLocale());
    if (edit_pattern != null) {
        editFormat.applyPattern(edit_pattern);
    }

    String dtp[] = li.getStringArray(LocaleInfo.NUMBER_ELEMENTS);
    exponentialChar = dtp[LocaleInfo.EXPONENTIAL];
}

/**
 * Since the <code>clone()</code> method is protected in the <code>Object</code>
 * class we cannot actually clone a generic object without knowing what it is;
 * therefore the validator must provide this routine to do the copying for
 * <code>JCField</code>.
 * @param value The value to be copied.
 * @return The new copy of value.
 */
public Object copyValue(Object value)
{
    // Numbers are immutable so we don't need to copy them
    return(value);
}

protected abstract void setPatterns();

/**
 * Sets the Locale used for formatting and parsing.
 * @param locale
 */
public void setLocale(Locale locale)
{
    if (locale == null) {
        locale = Locale.getDefault();
    }

    super.setLocale(locale);
    symbols = new DecimalFormatSymbols(locale);

    if (abstractInitializing == false) {
        setPatterns();
        displayFormat = createDisplayFormat(getLocale());
        if (display_pattern != null) {
            displayFormat.applyPattern(display_pattern);
        }
        editFormat = createEditFormat(getLocale());
        if (edit_pattern != null) {
            editFormat.applyPattern(edit_pattern);
        }
    }

    String dtp[] = li.getStringArray(LocaleInfo.NUMBER_ELEMENTS);
    exponentialChar = dtp[LocaleInfo.EXPONENTIAL];
}

abstract protected DecimalFormat createDisplayFormat(Locale l);

abstract protected DecimalFormat createEditFormat(Locale l);

/**
 * Sets the cursor to the specified position.
 * @param e
 */
public void setCursor(TextCursorEvent e)
{
    // We allow all positions in number edits
    e.getNewPosition();
}

/**
 * Returns the <code>displayPattern</code>.
 * @see #setDisplayPattern
 * @return <code>displayPettern</code>
 */
public String getDisplayPattern()
{
    return(display_pattern);
}

/**
 * Sets a pattern on the <code>DecimalFormat</code> object used
 * by the validator.
 * @see java.text.DecimalFormat
 * @param pattern
 */
public void setDisplayPattern(String pattern)
{
    this.display_pattern = pattern;
    displayFormat.applyPattern(pattern);
}

/**
 * Returns the current <code>editPattern</code>.
 * @return current <code>editPattern</code>
 * @see #setEditPattern
 */
public String getEditPattern()
{
    return(edit_pattern);
}

/**
 * Sets a pattern on the <code>DecimalFormat</code> object used
 * by the validator.
 * @see java.text.DecimalFormat
 * @param pattern
 */
public void setEditPattern(String pattern)
{
    this.edit_pattern = pattern;
    editFormat.applyPattern(pattern);
}

/**
 * Gets the <code>DecimalFormat</code> object being used by the validator.
 * @return
 */
public DecimalFormat getDisplayFormat()
{
    return(displayFormat);
}

/**
 * Gets the <code>DecimalFormat</code> object being used by the validator.
 * @return
 */
public DecimalFormat getEditFormat()
{
    return(editFormat);
}

/**
 * Sets the <code>DecimalFormat</code> object being used to format the data.
 * @param formatter
 */
public void setDisplayFormat(DecimalFormat formatter)
{
    if (formatter != null) {
        displayFormat = formatter;
    }
    else {
        throw new IllegalArgumentException("null DecimalFormat object invalid");
    }
}

/**
 * Sets the <code>DecimalFormat</code> object being used to format the data.
 * @param formatter
 */
public void setEditFormat(DecimalFormat formatter)
{
    if (formatter != null) {
        editFormat = formatter;
    }
    else {
        throw new IllegalArgumentException("null DecimalFormat object invalid");
    }
}

/**
 * Reports the first valid cursor position.
 * @return
 */
public int getFirstValidCursorPosition()
{
    return (0);
}

/**
 * Returns a formatted String given an object.
 * @return a formatted String
 * @param obj
 */
public String format(Object obj)
{
    if (obj == null || !isClassSupported(obj.getClass())) {
        return ("");
    }

    String simple_format = displayFormat.format(obj);
    if (isCurrency && use_intl_currency_symbol == true ||
        !getLocale().equals(getCurrencyLocale()))
    {
        // We need to replace the current local's symbol with
        // the currency locale's symbol
        String oldsym = symbols.getCurrencySymbol();
        String newsym = getCurrencySymbol(getCurrencyLocale());
        StringBuffer newstr = new StringBuffer(simple_format.length() -
                                               oldsym.length() +
                                               newsym.length());
        // Get the position of the old symbol
        int pos = simple_format.indexOf(oldsym);
        if (pos > 0) {
            // we copy the bytes up the pos into the new string
            newstr.append(simple_format.substring(0, pos));
        }
        // now insert new symbol
        newstr.append(newsym);
        // now add the rest of the formatted line
        newstr.append(simple_format.substring(pos + oldsym.length()));
        return (newstr.toString());
    }
    else {
        return (simple_format);
    }
}

//
// Currency support methods
//

/**
 * Returns <code>true</code> if the formatted data is a currency.
 * @return <code>true</code> if the formatted data is a currency
 */
public boolean isCurrency()
{
    return isCurrency;
}

/**
 * Sets the currency property for the data.
 * @param is_currency
 */
public void setCurrency(boolean is_currency)
{
    this.isCurrency = is_currency;

    // reset patterns for currency
    setPatterns();

    // apply display pattern
    if (display_pattern != null) {
        displayFormat.applyPattern(display_pattern);
    }
}

/**
 * Returns the locale used to determine the currency symbol to use.  If this
 * has not been explicitly set, then the default locale of the validator is
 * returned.
 * @return locale used to determine the currency symbol to use
 */
public Locale getCurrencyLocale()
{
    if (currency_locale == null) {
        return(getLocale());
    }
    return(currency_locale);
}

/**
 * Sets the locale used to determine the currency symbol.  If the
 * <code>CurrencyLocale</code> is <code>null</code>, then the validator's
 * Locale is used.
 * @see AbstractValidator#setLocale
 * @param l
 */
public void setCurrencyLocale(Locale l)
{
    if (l == null) {
        return;
    }

    if (l != currency_locale) {
        currency_locale = l;
    }
    // reset patterns for currency
    setPatterns();

    currencySymbols = new DecimalFormatSymbols(currency_locale);
}

protected String getLocalCurrencySymbol(Locale l)
{
    return(currencySymbols.getCurrencySymbol());
}

protected String getIntlCurrencySymbol(Locale l)
{
    return(currencySymbols.getInternationalCurrencySymbol());
}

/**
 * Returns the appropriate currency symbol of the specified locale.
 */
public String getCurrencySymbol(Locale l)
{
    if (l == this.locale) {
        if (use_intl_currency_symbol == false) {
            return(getLocalCurrencySymbol(l));
        }
    }
    return(getIntlCurrencySymbol(l));
}

/**
 * Returns the appropriate currency symbol of the current locale.  If the
 * currency locale matches the display locale and the
 * <code>useIntlCurrencySymbol</code> flag is not set, then the local currency
 * symbol will be returned; otherwise the international currency symbol is
 * returned.
 * @see #setCurrencyLocale
 * @see AbstractValidator#setLocale
 * @see #setUseIntlCurrencySymbol
 */
public String getCurrencySymbol()
{
    return(getCurrencySymbol(getCurrencyLocale()));
}

/**
 * Returns the status of the flag that indicates whether the international or
 * local currency symbol is being used.
 * @see #getCurrencySymbol
 * @return status of the flag that indicates whether the international or local
 * currency symbol is being used
 */
public boolean getUseIntlCurrencySymbol()
{
    return(use_intl_currency_symbol);
}

/**
 * Sets the status of the flag that indicates whether the international or local
 * currency symbol is being used.
 * @param s
 */
public void setUseIntlCurrencySymbol(boolean s)
{
    if (s != use_intl_currency_symbol) {
        use_intl_currency_symbol = s;
    }
}

}
