/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseAggregate.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:18 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Hashtable;
import java.util.Vector;

import com.klg.jclass.datasource.MetaDataModel;

/**
 * Implements the base Aggregate type behavior. This class should be
 * extended to create useful types.
 */

public abstract class BaseAggregate implements Aggregate, Serializable {

static final long serialVersionUID = 8905431406164420225L;

/** The metaData ID associated with the given column */
protected int metaID;

/** The column to aggregate over */
protected SummaryColumn column;

/** Long version of aggregated value */
protected long longValue;

/** Double version of aggregated value */
protected double doubleValue;

/** BigDecimal version of aggregated value */
protected BigDecimal bigDecimal;

private Hashtable tableMetaID;
private boolean metaIDFound;

/**
 * Default Constructor
 */
public BaseAggregate() {
    init(null);
}

/**
 * Default Constructor
 */
public BaseAggregate(SummaryColumn column) {
    init(column);
}

/**
 * Helper method to set a value.
 *
 * @param value A value in Object form.
 */
public void setValue(Object value) {
    if (column == null || value == null) {
        return;
    }
    switch (column.getDataType()) {
    case MetaDataModel.TYPE_DOUBLE:
        doubleValue = ((Double)value).doubleValue();
        break;
    case MetaDataModel.TYPE_FLOAT:
        doubleValue = ((Float)value).doubleValue();
        break;
    case MetaDataModel.TYPE_BIG_DECIMAL:
        bigDecimal = (BigDecimal)value;
        break;
    case MetaDataModel.TYPE_INTEGER:
        longValue = ((Integer)value).longValue();
        break;
    case MetaDataModel.TYPE_LONG:
        longValue = ((Long)value).longValue();
        break;
    case MetaDataModel.TYPE_BYTE:
        longValue = ((Byte)value).longValue();
        break;
    case MetaDataModel.TYPE_SHORT:
        longValue = ((Short)value).longValue();
        break;
    default:
        break;
    }
}

/**
 * Helper method to add a value.
 *
 * @param value A value in Object form.
 */
public void addValue(Object value) {
    if (column == null || value == null) {
        return;
    }
    switch (column.getDataType()) {
    case MetaDataModel.TYPE_DOUBLE:
        doubleValue += ((Double)value).doubleValue();
        break;
    case MetaDataModel.TYPE_FLOAT:
        doubleValue += ((Float)value).doubleValue();
        break;
    case MetaDataModel.TYPE_BIG_DECIMAL:
        bigDecimal = bigDecimal.add((BigDecimal)value);
        break;
    case MetaDataModel.TYPE_INTEGER:
        longValue += ((Integer)value).longValue();
        break;
    case MetaDataModel.TYPE_LONG:
        longValue += ((Long)value).longValue();
        break;
    case MetaDataModel.TYPE_BYTE:
        longValue += ((Byte)value).longValue();
        break;
    case MetaDataModel.TYPE_SHORT:
        longValue += ((Short)value).longValue();
        break;
    default:
        break;
    }
}

/**
 * Helper method to divide the aggregate value.
 *
 * @param value The integer divisor.
 */
public void divideValue(int value) {
    if (column == null || value == 0) {
        return;
    }
    switch (column.getDataType()) {
    case MetaDataModel.TYPE_DOUBLE:
    case MetaDataModel.TYPE_FLOAT:
        doubleValue /= (double)value;
        break;
    case MetaDataModel.TYPE_BIG_DECIMAL:
        try {
            bigDecimal = bigDecimal.divide(new BigDecimal(value), BigDecimal.ROUND_DOWN);
        }
        catch (Exception e) {
            System.out.println(e.toString());
        }
        break;
    case MetaDataModel.TYPE_INTEGER:
    case MetaDataModel.TYPE_LONG:
    case MetaDataModel.TYPE_BYTE:
    case MetaDataModel.TYPE_SHORT:
        longValue /= (long)value;
        break;
    default:
        break;
    }
}

/**
 * Helper method to get a long value from an Object.
 *
 * @param value A value in Object form.
 */
public double getLongValue(Object value) {
    if (value == null) {
        return 0.0;
    }
    String string = value.toString();
    return(new Long(string)).longValue();
}

/**
 * Helper method to get a double value from an Object.
 *
 * @param value A value in Object form.
 */
public double getDoubleValue(Object value) {
    if (value == null) {
        return 0.0;
    }
    String string = value.toString();
    return(new Double(string)).doubleValue();
}

/**
 * Helper method to check if the row node metaID is the same as the summary column one.
 *
 * @param rowNode The row node.
 */
public boolean isSameMetaID(RowNode rowNode) {
    return(rowNode.getRowFormat().getMetaData().getMetaID() == metaID);
}

/**
 * Helper method to get the result data from the row node given the column identifier.
 *
 * @param rowNode The row node.
 * @param name The column identifier
 */
public Object getRowNodeResultData(RowNode rowNode, String name) {
    Object o = null;
    try {
        o = rowNode.getDataTableModel().getResultData(rowNode.getBookmark(), name);
    }
    catch (Exception e) { /* do nothing */
    }
    return o;
}

///////////////////////////////////////////////////////////////////////////
// Default implementation of the Aggregate interface

/**
 * Initialize the aggregate.
 *
 * @param column The summary column.
 */
public void init(SummaryColumn column) {
    this.metaID = (column == null) ? (-1) : column.getTable();
    this.column = column;
    longValue = 0;
    doubleValue = 0.0;
    bigDecimal = null;
    if (column != null && column.getDataType() == MetaDataModel.TYPE_BIG_DECIMAL) {
        bigDecimal = new BigDecimal(0.0);
    }
    if (metaID < 0) {
        tableMetaID = null;
    }
    else {
        tableMetaID = new Hashtable();
        metaIDFound = false;
    }
}

/**
 * Determine if aggregation is done.
 *
 * @return True if done.
 */
public boolean done() {
    // default is to traverse all nodes
    return false;
}

/**
 * Determine if the next row level should be traversed.
 *
 * @param rowNode The row node.
 *
 * @return True if the lower level is to be traversed.
 */
public boolean nextLevel(RowNode rowNode) {
    if (metaID == (-1)) {
        // no interested metaID, traverse all levels
        return true;
    }

    // maintain a list of all meta data levels traversed
    int currentMetaID = rowNode.getRowFormat().getMetaData().getMetaID();
    Integer intCurrentMetaID = new Integer(currentMetaID);

    if (!metaIDFound) {
        // interested metaID not found yet, keep building the list
        tableMetaID.put(intCurrentMetaID, intCurrentMetaID);
        if (currentMetaID == metaID) {
            // at interested metaID level, no need to do next level
            metaIDFound = true;
            return false;
        }
        return true;
    }

    // interested metaID in list, check list
    if (tableMetaID.get(intCurrentMetaID) != null) {
        if (currentMetaID == metaID) {
            // at interested metaID level, no need to do next level
            return false;
        }
        return true;
    }

    // not in list, no need to do next level
    return false;
}

/**
 * Perform the aggregation.
 *
 * @param rowNode The row node.
 */
abstract public void calculate(RowNode rowNode);

/**
 * Indicates whether the result data is to be cached for display.
 *
 * @return True if the result data is to be cached.
 */
public boolean cacheResultData() {
    // default is to cache for display
    return true;
}

/**
 * Returns the result data as an Object.
 *
 * @return The result data.
 */
public Object getResultData() {
    if (column == null) {
        return null;
    }
    switch (column.getDataType()) {
    case MetaDataModel.TYPE_DOUBLE:
        return(Object) new Double(doubleValue);
    case MetaDataModel.TYPE_FLOAT:
        return(Object) new Float(doubleValue);
    case MetaDataModel.TYPE_BIG_DECIMAL:
        return(Object)bigDecimal;
    case MetaDataModel.TYPE_INTEGER:
        return(Object) new Integer((int)longValue);
    case MetaDataModel.TYPE_LONG:
        return(Object) new Long(longValue);
    case MetaDataModel.TYPE_BYTE:
        return(Object) new Byte((byte)longValue);
    case MetaDataModel.TYPE_SHORT:
        return(Object) new Short((short)longValue);
    default:
        break;
    }
    return null;
}

///////////////////////////////////////////////////////////////////////////

}
