/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: CellFormat.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:18 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.io.Serializable;
import java.util.*;



import javax.swing.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.cell.*;

/**
 * The CellFormat class completely describes a given cell displayed within
 * HiGrid.  From the CellStyle class, it inherits basic formatting
 * information.  It implements the JCCellInfo interface, so that it can
 * render cells in a standard way.  It also adds other properties useful
 * for display.  Vectors of CellFormat objects are found within the
 * RowFormat class, which themselves are aggregated within the FormatNode
 * class.  The sum of all of these classes serve as a design-time template
 * for the run-time rendering of HiGrid.
 */

public class CellFormat extends CellStyle implements JCCellInfo, Serializable {

static final long serialVersionUID = -6948816648665420216L;

protected static final int MINIMUM_INITIAL_HEADER_WIDTH = 40;

// Properties
protected int height = 0;
protected int width = 0;
protected boolean editable = true;
protected String editorName = null;
protected String rendererName = null;
transient protected Class editor = null;
transient protected Class renderer = null;
protected RowFormat parent = null;
protected String text = null;
protected String name = null;
protected int type = MetaDataModel.TYPE_OBJECT;
protected Rectangle drawingArea = new Rectangle();
protected Rectangle totalArea = new Rectangle();
protected Rectangle preferredTotalArea = new Rectangle();
protected Point translationOffset = new Point();
protected boolean drawingAreaValid = false;
protected boolean totalAreaValid = false;
protected boolean preferredTotalAreaValid = false;
protected boolean translationOffsetValid = false;
protected boolean rowHeightIsValid = false;
protected boolean showing = true;
protected Comparator sortable = null;

/**
 *  Default Constructor.
 */
public CellFormat() {
}

/**
 * Constructor.
 *
 * @param cellStyle the CellStyle to initialize the CellFormat with.
 */
public CellFormat(CellStyleModel cellStyle) {
    copyFromCellStyle(cellStyle);
}

/**
 * Retrieves whether this object is showing.
 *
 * @return Whether object is currently showing.
 */
public boolean isShowing() {
    return showing;
}

/**
 * Sets whether this object is showing.
 *
 * @param showing Whether object is to be shown.
 */
public void setShowing(boolean showing) {
    this.showing = showing;
    CellFormat cellFormat = getOtherCellFormat();
    if (cellFormat != null) {
        cellFormat.setOtherShowing(showing);
    }
}

/**
 * Sets whether this object is showing
 * for the other header/record pair.
 *
 * @param showing Whether object is to be shown.
 */
public void setOtherShowing(boolean showing) {
    this.showing = showing;
}

/**
 * Sets the border insets for this object.
 *
 * @param size The new border insets.
 */
public void setBorderInsets(Insets insets) {
    super.setBorderInsets(insets);
    drawingAreaValid = false;
    totalAreaValid = false;
    preferredTotalAreaValid = false;
    translationOffsetValid = false;
}

/**
 * Retrieves the height of this object.
 *
 * @return The current height.
 */
public int getHeight() {
    return height;
}

/**
 * Sets the height of this object.
 *
 * @param height The new height.
 */
public void setHeight(int height) {
    this.height = height;
    preferredTotalAreaValid = false;
    parent.heightChanged();
}

/**
 * Retrieves the width of this object.
 *
 * @return The current width.
 */
public int getWidth() {
    return width;
}

/**
 * Sets the width of this object.
 *
 * @param width The new width.
 */
public void setWidth(int width) {
    this.width = width;
    CellFormat cellFormat = getOtherCellFormat();
    if (cellFormat != null) {
        cellFormat.setOtherWidth(width);
    }
    drawingAreaValid = false;
    totalAreaValid = false;
    preferredTotalAreaValid = false;
}

/**
 * Sets the width of the other header/record pair.
 *
 * @param width The new width.
 */
protected void setOtherWidth(int width) {
    this.width = width;
    drawingAreaValid = false;
    totalAreaValid = false;
    preferredTotalAreaValid = false;
}

/**
 * Sets the width of this object.
 *
 * @param width The new width.
 * @param checkAllow True if allowWidthSizing is to be checked.
 */
public void setWidth(int width, boolean checkAllow) {
    if ((checkAllow && allowWidthSizing) || !checkAllow) {
        setWidth(width);
    }
}

/**
 * Sets the margin insets for this object.
 *
 * @param insets The new margin insets.
 */
public void setMarginInsets(Insets insets) {
    super.setMarginInsets(insets);
    drawingAreaValid = false;
    totalAreaValid = false;
    preferredTotalAreaValid = false;
    translationOffsetValid = false;
}

/**
 * Sets the font for this object.
 *
 * @param font The new font.
 * @exception IllegalArgumentException If parameter is null.
 */
public void setFont(FontMetrics fm, Font font) {
    setFont(font);
    setHeight(getFontHeight(fm, font));
}

/**
 * Sets whether width sizing is currently allowed.
 *
 * @param allow Whether to allow width sizing.
 */
public void setAllowWidthSizing(boolean allow) {
    super.setAllowWidthSizing(allow);
    CellFormat cellFormat = getOtherCellFormat();
    if (cellFormat != null) {
        cellFormat.setOtherAllowWidthSizing(allow);
    }
}

/**
 * Retrieves the cell format of the other header/record pair.
 *
 * @return The cell format.
 */
public CellFormat getOtherCellFormat() {
    CellFormat cellFormat = null;
    if (parent != null && parent.getParent() != null) {
        FormatNode node = (FormatNode) parent.getParent();
        int index = parent.getDataFormats().indexOf(this);
        boolean isIndicator = false;
        if (index == (-1)) {
            index = parent.getIndicatorFormats().indexOf(this);
            if (index != (-1)) {
                isIndicator = true;
            }
        }
        if (parent instanceof HeaderFormat) {
            RecordFormat rf = node.getRecordFormat();
            if (index != (-1) && rf != null) {
                if (isIndicator) {
                    cellFormat = ((CellFormat)rf.getIndicatorFormats().elementAt(index));
                }
                else {
                    cellFormat = ((CellFormat)rf.getDataFormats().elementAt(index));
                }
            }
        }
        else if (parent instanceof RecordFormat) {
            HeaderFormat hf = node.getHeaderFormat();
            if (index != (-1) && hf != null) {
                if (isIndicator) {
                    cellFormat = ((CellFormat)hf.getIndicatorFormats().elementAt(index));
                }
                else {
                    cellFormat = ((CellFormat)hf.getDataFormats().elementAt(index));
                }
            }
        }
    }
    return cellFormat;
}

/**
 * Retrieves the cell's associated data type.
 */
public Class getDataType() {
    String className;
    switch (getType()) {
    case MetaDataModel.TYPE_BOOLEAN:
        className = "java.lang.Boolean";
        break;
    case MetaDataModel.TYPE_SQL_DATE:
        className = "java.sql.Date";
        break;
    case MetaDataModel.TYPE_DOUBLE:
        className = "java.lang.Double";
        break;
    case MetaDataModel.TYPE_FLOAT:
        className = "java.lang.Float";
        break;
    case MetaDataModel.TYPE_INTEGER:
        className = "java.lang.Integer";
        break;
    case MetaDataModel.TYPE_STRING:
        className = "java.lang.String";
        break;
    case MetaDataModel.TYPE_BIG_DECIMAL:
        className = "java.math.BigDecimal";
        break;
    case MetaDataModel.TYPE_LONG:
        className = "java.lang.Long";
        break;
    case MetaDataModel.TYPE_SQL_TIME:
        className = "java.sql.Time";
        break;
    case MetaDataModel.TYPE_SQL_TIMESTAMP:
        className = "java.sql.Timestamp";
        break;
    case MetaDataModel.TYPE_BYTE:
        className = "java.lang.Byte";
        break;
    case MetaDataModel.TYPE_SHORT:
        className = "java.lang.Short";
        break;
    case MetaDataModel.TYPE_BYTE_ARRAY:
        // not clear what to return here; it's really a byte[]
        className = "java.lang.String";
        break;
    case MetaDataModel.TYPE_UTIL_DATE:
        className = "java.util.Date";
        break;
    case MetaDataModel.TYPE_OBJECT:
    default:
        className = "java.lang.Object";
        break;
    }
    Class returnClass = null;
    try {
        returnClass = Class.forName(className);
    }
    catch (ClassNotFoundException e) {
        System.out.println(e);
    }
    return returnClass;
}

/**
 * Retrieves whether the cell is enabled.
 *
 * @return Whether the cell is enabled.
 */
// Note: this is just to implement the JCCellInfo interface
public boolean isEnabled() {
    return true;
}

/**
 * Retrieves whether input is allowed for this object.
 *
 * @return Whether input is currently allowed.
 */
public boolean isEditable() {
    return editable;
}

/**
 * Sets whether input is currently allowed.
 *
 * @param isEditable Whether to allow input.
 */
public void setEditable(boolean editable) {
    if (parent != null) {
        MetaDataModel metaData = parent.getMetaData();
        if (metaData != null) {
            try {
                ColumnModel cm = metaData.getColumnObject(name);
                if (cm != null && cm.isReadOnly()) {
                    editable = false;
                }
            }
            catch (Exception e) {
            } // do nothing
        }
    }
    this.editable = editable;
}

/**
 * Retrieves the Cell Editor for this object.
 *
 * @return The current Cell Editor.
 */
public Class getCellEditor() {
    if (editor == null && editorName != null) {
        setCellEditorName(editorName);
    }
    return editor;
}

/**
 * Sets the Cell Editor for this object.
 *
 * @param editor The new Cell Editor.
 */
public void setCellEditor(Class editor) {
    this.editor = editor;
    this.editorName = editor.getName();
}

/**
 * Retrieves the Cell Editor name for this object.
 *
 * @return The current Cell Editor name.
 */
public String getCellEditorName() {
    return editorName;
}

/**
 * Sets the Cell Editor for this object.
 *
 * @param editor The new Cell Editor name.
 */
public void setCellEditorName(String editorName) {
    this.editorName = new String(editorName);
    try {
        this.editor = editor.forName(editorName);
    }
    catch (Exception e) {
        e.printStackTrace();
    }
}

/**
 * Retrieves the Cell Renderer for this object.
 *
 * @return The current Cell Renderer.
 */
public Class getCellRenderer() {
    if (renderer == null && rendererName != null) {
        setCellRendererName(rendererName);
    }
    return renderer;
}

/**
 * Sets the Cell Renderer for this object.
 *
 * @param renderer The new Cell Renderer.
 */
public void setCellRenderer(Class renderer) {
    this.renderer = renderer;
    this.rendererName = renderer.getName();
}

/**
 * Retrieves the Cell Renderer name for this object.
 *
 * @return The current Cell Renderer name.
 */
public String getCellRendererName() {
    return rendererName;
}

/**
 * Sets the Cell Renderer for this object.
 *
 * @param renderer The new Cell Renderer name.
 */
public void setCellRendererName(String rendererName) {
    this.rendererName = new String(rendererName);
    try {
        this.renderer = renderer.forName(rendererName);
    }
    catch (Exception e) {
        e.printStackTrace();
    }
}

/**
 * Retrieves the parent of this object.
 *
 * @return The current parent.
 */
public RowFormat getParent() {
    return parent;
}

/**
 * Sets the parent of this object.
 *
 * @param parent The new parent.
 */
public void setParent(RowFormat parent) {
    this.parent = parent;
}

/**
 * Retrieves the text label of this object.
 *
 * @return The current text.
 */
public String getText() {
    return text;
}

/**
 * Sets the text label of this object.
 *
 * @param text The new text.
 */
public void setText(String text) {
    this.text = text;
}

/**
 * Retrieves the name of this object.
 *
 * @return The current name.
 */
public String getName() {
    return name;
}

/**
 * Sets the name of this object.
 *
 * @param name The new name.
 */
public void setName(String name) {
    this.name = name;
}

/**
 * Retrieves the type of this object.
 *
 * @return The current type.
 */
public int getType() {
    return type;
}

/**
 * Sets the type of this object.
 *
 * @param type The new type.
 */
public void setType(int type) {
    this.type = type;
}

/**
 * Retrieves the sortable.
 *
 * @return The current sortable.
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J, 
 * which is based on JDK 1.1, this field's or method's return type 
 * is based on the corresponding <code>com.sun.java.util</code> class. 
 */
public Comparator getSortable() {
    return sortable;
}

/**
 * Sets the sortable.
 *
 * @param sortable The new sortable.
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J, 
 * which is based on JDK 1.1, this method's parameter must be 
 * a <code>com.sun.java.util.collections.Comparator</code>. 
 */
public void setSortable(Comparator sortable) {
    this.sortable = sortable;
}

/**
 * Returns the drawing area for the cell being displayed or
 * edited.  This drawing area is already adjusted for BorderSize
 * and MarginSize
 */
public Rectangle getDrawingArea() {
    if (!drawingAreaValid) {
        calculateDrawingArea();
    }
    return drawingArea;
}

/**
 * Returns the total area for the cell being displayed or
 * edited.
 */
public Rectangle getTotalArea() {
    if (!totalAreaValid) {
        calculateTotalArea();
    }
    return totalArea;
}

/**
 * Returns the preferred area for the cell being displayed or
 * edited.  This is the area as determined by the design time information.
 */
public Rectangle getPreferredTotalArea() {
    if (!preferredTotalAreaValid) {
        calculatePreferredTotalArea();
    }
    return preferredTotalArea;
}

protected Point getTranslationOffset() {
    if (!translationOffsetValid) {
        calculateTranslationOffset();
    }
    return translationOffset;
}

// Ripped from BWTUtil.java
/** Accounts for bug in AWT which does not account for italic fonts. */
protected int getStringWidth(FontMetrics fm, Font font, String s) {
    return fm.stringWidth(s) + (font.isItalic() ? font.getSize()/3+1 : 0);
}

protected int getFontHeight(FontMetrics fm, Font font) {
    return fm.getLeading() + fm.getMaxAscent() + fm.getMaxDescent();
}

protected int getFontHeight(Component comp, Font font) {
    return getFontHeight(comp.getFontMetrics(font), font);
}

/**
 *  Sets the height and width of this cell based on its font and text.
 *  Used mainly to initialize headers.
 *
 * @param comp Component to get FontMetrics from.
 * @param addSortLabel true if space is to be reserved for the sort icon.
 */

public void setDefaultSize(Component comp, boolean addSortLabel) {
    if (font == null) {
        return;
    }
    FontMetrics fm = comp.getFontMetrics(font);
    int fontHeight = getFontHeight(fm, font);
    int iconHeight = addSortLabel ? SortLabelCellRenderer.getAscendingImage().getHeight(null) : 0;
    setHeight(Math.max(fontHeight, iconHeight));

    int extra = addSortLabel ? SortLabelCellRenderer.getAscendingImage().getWidth(null) : 0;
    int defaultWidth = getStringWidth(fm, font, name) +
                       getMarginInsets().left + getMarginInsets().right +
                       getBorderInsets().left + getBorderInsets().right +
                       extra;
    setWidth(Math.max(defaultWidth, MINIMUM_INITIAL_HEADER_WIDTH));
}

/**
 * Called by parent RowFormat when its height changes.
 */
protected void rowHeightChanged() {
    calculateDrawingArea();
    calculateTotalArea();
}

/**
 * Called by parent RowFormat when its width changes.
 */
protected void widthChanged() {
    calculateDrawingArea();
    calculateTotalArea();
    calculatePreferredTotalArea();
}

// The next calculate*() methods calculate and cache the
// indicated items for higher performance when rendering

protected void calculateDrawingArea() {
    // the row's height overrides the cell's actual height
    // however, the height must be adjusted for border and margin
    drawingArea.setBounds(0, 0, getWidth(), getParent().getHeight());
    drawingArea.height -= (getMarginInsets().top + getMarginInsets().bottom +
                           getBorderInsets().top + getBorderInsets().bottom);
    drawingAreaValid = true;
}

protected void calculateTotalArea() {
    // the row's height overrides the cell's actual height
    // however, the width must be adjusted for border and margin
    totalArea.setBounds(0, 0, getWidth(), getParent().getHeight());
    totalArea.width += (getMarginInsets().left + getMarginInsets().right +
                        getBorderInsets().left + getBorderInsets().right);
    totalAreaValid = true;
}

protected void calculatePreferredTotalArea() {
    // the cell's width and height must be adjusted for border and margin
    preferredTotalArea.setBounds(0, 0, getWidth(), getHeight());
    preferredTotalArea.width += (getMarginInsets().left + getMarginInsets().right +
                                 getBorderInsets().left + getBorderInsets().right);
    preferredTotalArea.height += (getMarginInsets().top + getMarginInsets().bottom +
                                  getBorderInsets().top + getBorderInsets().bottom);
    preferredTotalAreaValid = true;
}

protected void calculateTranslationOffset() {
    // the translation offset must be adjusted for margins and borders
    translationOffset.setLocation(getMarginInsets().left + getBorderInsets().left,
                                  getMarginInsets().top + getBorderInsets().top);
    translationOffsetValid = true;
}

}
