/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DefaultDataModelListener.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:18 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.event.KeyEvent;
import java.io.Serializable;
import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * The <code>DefaultDataModelListener</code> class handles DataModelEvents
 * from HiGrid's DataModel.  It ensures that the runtime display
 * of the HiGrid instance matches the state of the DataModel.
 */

public class DefaultDataModelListener extends DataModelAdapter implements Serializable {

static final long serialVersionUID = 6484051278069816163L;

private HiGrid grid = null;
protected boolean refreshGrid = true;
protected RowNode cellRowNode = null;
protected RowNode lastOpenedRowNode = null;

protected DefaultDataModelListener(HiGrid grid) {
    this.grid = grid;
    refreshGrid = true;
    setLastOpenedRowNode(null);
}

//
// Implement DataModelListener interface
//

/**
 * Called before a row is inserted. Can be cancelled.
 * If not cancelled will be followed by afterInsertRow().
 * @see DataTableModel#addRow
 */
public void beforeInsertRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called before entire grid is reset. Can be cancelled.
 * If not cancelled will be followed by afterReset().
 * This will result from a call to DataModel.requeryAll().
 * @see DataModel#requeryAll
 */
public void beforeReset(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called when either DataTableModel.requeryRowAndDetails() or
 * DataTableModel.requeryRow() is called.
 * If not cancelled will be followed by an,
 * <ul>
 * <li>afterRequeryRowAndDetails() call or an</li>
 * <li>afterDeleteRow() call in the case getRowStatus() == INSERTED or a</li>
 * <li>afterChangeOfRowData() call in the case getRowStatus() == UPDATED or COMMITTED.</li>
 * </ul>
 * @see DataTableModel#requeryRow
 * @see DataTableModel#requeryRowAndDetails
 */
public void beforeRequery(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called before a cell is edited. Can be cancelled.
 * @see DataTableModel#updateCell
 */
public void beforeEditCell(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called before a row is [logically] deleted. Can be cancelled.
 * If not cancelled will be followed by an
 * afterDeleteRow90 or a afterChangeOfRowData() if the commit policy
 * is COMMIT_MANUALLY or COMMIT_LEAVING_ANCESTOR.
 * @see DataTableModel#deleteRow
 * @see MetaDataModel#getCommitPolicy
 */
public void beforeDeleteRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called before all changes are committed. Can be cancelled.
 * All modified, deleted and inserted rows at all levels are about
 * to be committed.  beforeDeleteRow() and afterChangeOfRowData() 
 * will follow depending on the operations performed on the modified rows
 * being saved.
 * Results from a call to DataModel.updateAll.
 * @see DataModel#updateAll
 */
public void beforeCommitAll(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called before all changes are cancelled. Can be cancelled.
 * afterInsertRow() and afterChangeOfRowData() can follow this
 * event.
 * @see DataModel#cancelAll
 */
public void beforeCancelAll(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/** Called before all edits to a row are undone. Can be cancelled
 * An afterDeleteRow() or afterChangeOfRowData() will follow.
 * @see DataTableModel#cancelRowChanges
 */
public void beforeCancelRowChanges(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

/**
 * Called when the root-level bookmark for a subtree changes.
 * When this happens those nodes in the previous subtree which
 * are not COMMIT_MANUALLY are committed.
 * Can be cancelled.  If cancelled the cursor moves but the
 * changes are automatically committed.
 */
public void beforeCommitConditional(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    RowNode rowNode = grid.getCurrentRowNode();
    if (rowNode != null && isAncestor(rowNode, row)) {
        if (! grid.getController().commitAndConvert()) {
            // commit failed; don't need to move anywhere
            event.cancelProposedAction();
        }
    }
}

/**
 * Called before single row is committed to data source.  Can be cancelled in
 * which case the row edits are not written to the datasource and the rows
 * status remains modified.  afterDeleteRow() or afterChangeOfRowData() will
 * follow depending on the status of the row to be committed.
 * @see DataTableModel#commitRow
 */
public void beforeCommitRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    RowNode rowNode = grid.getCurrentRowNode();
    if (rowNode != null && rowNode.getBookmark() == row) {
        if (! grid.getController().commitAndConvert()) {
            // commit failed; don't need to move anywhere
            event.cancelProposedAction();
        }
    }
}

/**
 * The global cursor will move to a new row.
 * Can be cancelled.
 */
public void beforeMoveToCurrentRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
    // ask Controller to validate itself before allowing move
    if (!grid.getController().validateCell(null, false)) {
        event.cancelProposedAction();
    }
}

/**
 * The indicated DataTable will be deleted and flushed from the cache.
 * Can be cancelled.
 */
public void beforeDeleteTable(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    if (!grid.getRowTree().parentFolderClosed((DataTableModel) event.getTable())) {
        event.cancelProposedAction();
    }
}

/**
 * A row has changed, re-read all its cells and its status to reflect the new
 * values.  If this event is the result of a cell edit call DataModelEvent.getColumn()
 * to get the name of the column which changed. If getColumn() returns "", re-read
 * the entire row.
 * Called when
 * <ul>
 * <li>a row is deleted and getShowDeletedRows() == true</li>
 * <li>a cell was edited</li>
 * <li>row edits are cancelled and getRowStatus() == UPDATED</li>
 * <li>row edits are cancelled and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>row is commited and getRowStatus() == UPDATED or INSERTED</li>
 * <li>row is requeried and getRowStatus() != INSERTED</li>
 * </ul>
 */
public void afterChangeOfRowData(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    grid.resetSortIndications(row, false);

    RowTree rt = grid.getRowTree();

    // get the row node for the given bookmark
    RowNode rowNode = rt.findRecordRowNode(null, row);

    // update the row and children - do the following:
    if (rowNode != null) {
        // the row node after an insertion may now have children
        if (rowNode.getState() == NodeStatus.FOLDER_NONE) {
            if ((rowNode.getFormatNode()).hasChildren()) {
                rowNode.setState(NodeStatus.FOLDER_CLOSED);
            }
        }

        // redraw this row
        refreshRow(rowNode);
    }
}

/**
 * A new row has been added to the datasource. Listeners need to
 * display the row.
 */
public void afterInsertRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
    long row = event.getBookmark();
    grid.resetSortIndications(row, false);

    boolean status = false;
    DataTableModel dtm = event.getTable();
    RowTree rt = grid.getRowTree();
    RowNode rnOpenedFolder = getLastOpenedRowNode();
    setLastOpenedRowNode(null);

    // is there a non empty node level for this row
    RowNode rnFirst = rt.findFirstRowNode(null, dtm);
    RowNode rowNode = null;
    CellPosition cell = grid.getCurrentCell();
    if (cell == null) {
        RowNode rnParent = (RowNode) rt.getRoot();
        FormatNode formatNode = (FormatNode)grid.getFormatTree().getRoot();
        if (formatNode.getHeaderFormat().isShowing()) {
            RowNode header = rt.createRowNode(formatNode, formatNode.getHeaderFormat(), dtm, 0);
            rt.addRowNode(rnParent, header, null);
            rt.addTotalNumberOfScrollableRows(header, formatNode);
        }
        rowNode = rt.createRowNode(formatNode, formatNode.getRecordFormat(), dtm, row);
        rt.addRowNode(rnParent, rowNode, null);
        rowNode.setState(NodeStatus.FOLDER_NONE);
        rt.addTotalNumberOfScrollableRows(rowNode, formatNode);
        if (formatNode.getFooterFormat().isShowing()) {
            RowNode footer = rt.createRowNode(formatNode, formatNode.getFooterFormat(), dtm, 0);
            rt.addRowNode(rnParent, footer, null);
            rt.addTotalNumberOfScrollableRows(footer, formatNode);
        }
        status = true;
    }
    else if (rnFirst != null) {
        // get rowIndex of new row to find right place to insert it
        int rowIndex = event.getRowIndex();
        RowNode firstRowNode = rt.findFirstRowNode(null, dtm);
        RowNode rnParent = (RowNode)firstRowNode.getParent();
        RowNode rnBefore = firstRowNode;
        // initial position may be incorrect due to existence of headers and other non-record rowNodes
        while (rowIndex >= dtm.getRowIndex(rnBefore.getBookmark())) {
            RowNode next = (RowNode)rnBefore.getNextSibling();
            if (next != null && next.getDataTableModel() == dtm) {
                rnBefore = (RowNode)rnBefore.getNextSibling();
            }
            else {
                rnBefore = (RowNode)rnBefore.getNextSibling();
                break;
            }
        }
        // create and add the rowNode
        FormatNode formatNode = rnFirst.getFormatNode();
        rowNode = rt.createRowNode(formatNode, formatNode.getRecordFormat(), dtm, row);
        rt.addRowNode(rnParent, rowNode, rnBefore);
        rt.addTotalNumberOfScrollableRows(rowNode, formatNode);
        rowNode.setState(NodeStatus.FOLDER_NONE);
        status = true;
        // check whether the table should now be re-sorted
        HiGridInternalData idata = (HiGridInternalData)dtm.getInternalData(grid);
        if (idata != null) {
            SortData sortData = idata.getSortData();
            if (sortData != null) {
                grid.sortDataTable(rowNode, sortData.getColumnName(), sortData.getDirection(), false);
            }
        }
    }
    // at empty node level, need to insert an entire node level
    else {
        RowNode rnParent = rt.findRecordRowNode(null, dtm.getParentBookmark());
        if (rnParent != null) {
            // simply case of a single child
            if (! rnParent.hasChildren()) {
                // similiar to opening a folder with no previous children
                try {
                    rt.setSortTable(false);
                    rt.createRuntimeChildren(rnParent);
                    rt.setSortTable(true);
                }
                catch (Exception e) {
                    grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
                }
                rt.addTotalNumberOfScrollableRows(rt.getNumberOfScrollableRows(rnParent, true));
                rt.addRepeatHeader(rnParent);
                // change the folder state from FOLDER_NONE
                rnParent.setState(NodeStatus.FOLDER_OPEN);
            }
            // case of parent row node has other children already
            else {
                rt.setSortTable(false);
                FormatNode formatNode = rt.createRuntimeChildren(rnParent, dtm);
                rt.setSortTable(true);
                if (formatNode != null) {
                    rt.addTotalNumberOfScrollableRows(rt.getNumberOfScrollableRows(rnParent, formatNode, true));
                }
            }
            rowNode = rt.findRecordRowNode(rnParent, row);
            if (rowNode != null) {
                rowNode.setState(NodeStatus.FOLDER_NONE);
            }
            status = true;
        }
    }

    // move to the newly inserted row, if triggered by user
    // this may trigger commit row operation(s)
    if (status && event.getOriginator() == grid.getDataModelListener()) {
        rt.setGridAreaRows();
        CellPosition newPosition = new CellPosition(rowNode, rowNode.getRowFormat().getFirstVisibleCell());
        grid.getHorizontalScrollbar().setValue(0);
        if (!grid.getController().validateAndTraverse(newPosition, newPosition.columnIndex, null)) {
            // cancel the insert
            try {
                dtm.cancelRowChanges(this, row);
            }
            catch (Exception e) {
                grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
            }
            if (rnOpenedFolder != null) {
                rt.closeFolder(rnOpenedFolder);
            }
            else {
                rt.setGridAreaRows();
            }
        }
    }

    if (status) {
        refresh();
    }
}

/**
 * Removes the row from the display.
 * A row has been physically deleted and needs to be
 * removed from the display or has been logically deleted but the
 * showDeletedRows property has been set to false. Called when,
 * <ul>
 * <li>a row has been logically deleted and getShowDeletedRows() == false</li>
 * <li>row changes have been cancelled and getRowStatus() == INSERTED</li>'
 * <li>a row is committed and getRowStatus() == DELETED and getShowDeletedRows() == true</li>
 * <li>a row has been requeried and getRowStatus() == INSERTED</li>
 * </ul>
 */
public void afterDeleteRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    RowTree rt = grid.getRowTree();

    // get the row node for the given bookmark
    RowNode rowNode = rt.findRecordRowNode(null, row);

    // delete the row - do the following:
    if (rowNode != null) {
        // possibly move the current cell position
        cellRowNode = moveCurrentCell(rowNode, false);

        // collapse the row if opened
        if (rowNode.getState() == NodeStatus.FOLDER_OPEN) {
            rt.closeFolder(rowNode);
        }

        // remove its repeat header
        // only if there is a previous record row node
        // and no next record row node for the same level
        FormatNode formatNode = rowNode.getFormatNode();
        RowNode rnPrev = rt.findPreviousRecordRowNode(rowNode, formatNode);
        RowNode rnNext = rt.findNextRecordRowNode(rowNode, formatNode);
        if (rnPrev != null && rnNext == null) {
            rt.removeRepeatHeader(rnPrev);
        }

        RowNode rnParent = (RowNode) rowNode.getParent();
        // remove the row and its children
        rt.subtractTotalNumberOfScrollableRows(rowNode, formatNode);
        rnParent.remove(rowNode);

        boolean otherLevel = false;
        if (rt.findRecordRowNode(rnParent, formatNode) == null) {
            // last record row for format node, remove all non record rows
            TreeIterator ti = rnParent.getIterator();
            while (ti.hasMoreElements()) {
                RowNode rn = (RowNode)ti.get();
                if (rn.getFormatNode() == formatNode) {
                    rt.subtractTotalNumberOfScrollableRows(rn, formatNode);
                    rnParent.remove(rn);
                }
                else {
                    otherLevel = true;
                }
                ti.nextElement();
            }
            // collapse level if no more rows
            if (!otherLevel) {
                if (rnParent.getState() != NodeStatus.FOLDER_NONE) {
                    rnParent.setState(NodeStatus.FOLDER_NONE);
                }
                // remove any possible repeat header
                rt.removeRepeatHeader(rnParent);
            }
        }

        // not sure, but the top row node may have been deleted
        rt.setTopRow(grid.getVerticalScrollbar().getValue());

        // redraw the grid after resetting the summary data
        ((HiGridData)grid.getDataModel()).resetSummaryDataTables(grid, true);
        refresh();
    }
}

/**
 * Listeners must close all expanded views and reset/reread the root node.
 * The previous pointer to the root node is no longer valid.  Call
 * DataModel.getDataTableTree().getRoot() for the new root table.
 * Called when the datasource has been reset.
 * @see DataModel#requeryAll
 */
public void afterReset(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    grid.resetRunTimeGrid();
    refresh();
}

/**
 * The global cursor has moved to a new row.
 * Listeners should position their cursor on
 * the indicated row.  In a master-detail relationship child levels
 * should refresh themselves to reflect data sets which correspond
 * to the new parent row by calling DataModel.getCurrentDataTable()
 * or for field controls, DataModel.getCurrentDataItem().
 */
public void afterMoveToCurrentRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    CellPosition cell = grid.getCurrentCell();
    if (cell != null && cell.row != null && cell.row.getBookmark() != row) {
        RowNode rowNode = grid.getRowTree().findRecordRowNode(null, row);
        if (rowNode != null && rowNode != cell.row) {
            if (!grid.getController().isTraversing()) {
                grid.getController().repositionCurrentCell(rowNode, true, false);
            }
        }
    }
    // Scrolling into view (making the editor visible) is optional.
    if (grid.isScrollOnMoveEvent()) {
    	grid.getController().makeCellEditorVisible();
    }
}

/**
 * Re-reads the indicated row and refresh all open children
 * under this row.
 */
public void afterRequeryRowAndDetails(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    grid.resetSortIndications(row, true);

    RowTree rt = grid.getRowTree();
    RowNode cellRowNode = null;

    // get the row node for the given bookmark
    RowNode rowNode = rt.findRecordRowNode(null, row);

    // update the row and children - do the following:
    if (rowNode != null) {
        boolean folderOpened = rowNode.getState() == NodeStatus.FOLDER_OPEN;
        if (folderOpened) {
            // possibly move the current cell position
            cellRowNode = moveCurrentCell(rowNode, true);

            // collapse the row if opened
            rt.closeFolder(rowNode);
        }

        // remove any children
        rowNode.removeAllChildren();

        // re-open folder if opened previously
        if (folderOpened) {
            rt.openFolder(rowNode);
        }

        // not sure, but the top row node may have been deleted
        rt.setTopRow(grid.getVerticalScrollbar().getValue());

        // redraw the grid
        refresh();
    }
}

/**
 * Re-read this table and refresh all open children in
 * this table.
 */
public void afterRequeryTable(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    grid.resetSortIndications(row, false);

    RowNode rowNode = grid.getCurrentRowNode();
    DataTableModel dataTable = event.getTable();

    if ((rowNode == null) || (dataTable == null)) {
        return;
    }

    RowTree rt = grid.getRowTree();
    RowNode originalRowNode = rowNode;

    // find the first row of the table to be deleted
    RowNode firstRowNode = rowNode;
    FormatNode currentFormatNode = rowNode.getFormatNode();
    while (true) {
        rowNode = (RowNode)rowNode.getPreviousSibling();
        if (rowNode == null) {
            break;
        }
        if (rowNode.getFormatNode() == currentFormatNode) {
            firstRowNode = rowNode;
        }
        else {
            break;
        }
    }

    // insert the new dataTable rowNodes in front of the old table's first row
    int scrollableRows = rt.createRuntimeChildren((RowNode) originalRowNode.getParent(), 
                                                  currentFormatNode, dataTable, true, firstRowNode);

    // must move the current cell position
    if (scrollableRows == 0) {
        // table has vanished, move to the prior record
        cellRowNode = rt.findPreviousRecordRowNode(firstRowNode, null);
    }
    else {
        // move to the first record of the refreshed table
        cellRowNode = rt.findPreviousRecordRowNode(firstRowNode, currentFormatNode);
        CellPosition cellPosition = grid.getController().goLevelHome(cellRowNode);
        cellRowNode = cellPosition.row;
    }
    if (cellRowNode != null) {
        grid.getController().repositionCurrentCell(cellRowNode, true, true);
        cellRowNode = null;
    }
    else {
        grid.getController().setCurrentCell(null);
    }

    // delete all of the old rowNodes
    rowNode = firstRowNode;
    while (rowNode != null) {
        if (rowNode != null) {
            boolean folderOpened = rowNode.getState() == NodeStatus.FOLDER_OPEN;
            if (folderOpened) {
                // collapse the row if opened
                rt.closeFolder(rowNode);
            }
            // remove any children
            rowNode.removeAllChildren();
        }
        RowNode nodeToRemove = rowNode;
        rowNode = (RowNode)rowNode.getNextSibling();
        ((RowNode)nodeToRemove.getParent()).remove(nodeToRemove);
        if ((rowNode == null) || (rowNode.getFormatNode() != currentFormatNode)) {
            break;
        }
    }

    // not sure, but the top row node may have been deleted
    rt.setTopRow(grid.getVerticalScrollbar().getValue());

    // redraw the grid
    refresh();
}

/**
 * Notification that multiple events are coming.
 * Multiple events will be nested
 * between beginEvents() and endEvents().
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints.
 */
public void beginEvents(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    refreshGrid = false;
}

/**
 * Notification that multiple events are complete.
 * Multiple events will be nested
 * between beginEvents() and endEvents().
 * Allows listeners to treat the events as a batch, to,
 * for example, reduce repaints. Called when
 * <ul>
 * <li>DataModel.updateAll is called</li>
 * <li>DataModel.cancelAll is called</li>
 * </ul>
 */
public void endEvents(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    long row = event.getBookmark();
    refreshGrid = true;
    refresh();
}

/**
 * The current row has been deleted and the originator
 * of the deletion should now reposition the global
 * cursor to a new, valid row.
 */
public void originatorNavigateRow(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }

    if (cellRowNode != null) {
        grid.getController().repositionCurrentCell(cellRowNode, true, true);
        cellRowNode = null;
    }
}

/**
 * For internal use only
 * Used by the JCData bean.
 */
public void initialize(DataModelEvent event) {
    // sanity check
    if (grid == null || event == null) {
        return;
    }
}

//
// End of DataModelListener interface implementation
//

///////////////////////////////////

/**
 * Retrieves the rowNode last opened by the UI.
 * @return the rowNode last opened by the UI.
 */
protected RowNode getLastOpenedRowNode() {
    return lastOpenedRowNode;
}

/**
 * Sets the rowNode last opened by the UI.
 * @param the rowNode last opened by the UI.
 */
protected void setLastOpenedRowNode(RowNode rowNode) {
    this.lastOpenedRowNode = rowNode;
}

/**
 * Refresh the indicated row
 *
 * @param rowNode The row node to refresh.
 */
protected void refreshRow(RowNode rowNode) {
    // refresh the indicated row
    grid.getController().initializeEditor();
    grid.refreshRowWithoutEditor(rowNode);

    // refresh if necessary all summary rows that are visible
    ((HiGridData)grid.getDataModel()).resetSummaryDataTables(grid, false);
    SummaryRefreshWalk walkable = new SummaryRefreshWalk(grid, grid.getGridArea().getSize().height);
    grid.getRowTree().walk(grid.getRowTree().getTopRowNode(), walkable);
}

/**
 * Refresh higrid.
 */
protected void refresh() {
    if (refreshGrid) {
        ((HiGridData)grid.getDataModel()).resetSummaryDataTables(grid, false);
        grid.refresh();
    }
}


/**
 * Update the edit status for an existing row node for a given bookmark
 *
 * @param row The bookmark of the row node
 *
 * @return True if row node is updated successfully, else false.
 */
protected boolean updateEditStatus(long row) {
    RowTree rt = grid.getRowTree();

    // get the row node for the given bookmark
    RowNode rowNode = rt.findRecordRowNode(null, row);

    // update the row and children - do the following:
    if (rowNode != null) {
        // update the edit status for this row
        grid.refreshEditStatus(rowNode);
        return true;
    }
    return false;
}

/**
 * Possibly move the current cell position out of the given row node
 * and its children.
 *
 * @param rowNode The row node.
 * @param excludeRow True if the row node is excluded from the cell move.
 *
 * @return The row node for the new cell position
 */
protected RowNode moveCurrentCell(RowNode rowNode, boolean excludeRow) {
    CellPosition cell = grid.getCurrentCell();
    RowNode newRowNode = null;
    if (cell == null || cell.row == null) {
        return newRowNode;
    }

    // for the case of possibly moving the current cell out of the children
    if (excludeRow) {
        if (cell.row.isChildOf(rowNode)) {
            // just move it up to the parent (i.e. the row node)
            newRowNode = rowNode;
        }
        return newRowNode;
    }

    // for the case of possibly moving the current cell out of the children
    // and the row node itself
    if (cell.row != rowNode && !(cell.row.isChildOf(rowNode))) {
        return newRowNode;
    }
    // current cell position found to be in row node or its
    // children that is to be deleted

    // move current cell position to next sibling (same format node)
    // if no next sibling, move to its previous sibling (same format node)
    // if no previous sibling, move to next sibling (same branch)
    // if no next sibling, move to its previous sibling (same branch)
    // if no previous sibling, move up to its parent
    RowTree rt = grid.getRowTree();
    FormatNode formatNode = rowNode.getFormatNode();
    RowNode rn = rt.findNextRecordRowNode(rowNode, formatNode);
    if (rn == null) {
        rn = rt.findPreviousRecordRowNode(rowNode, formatNode);
    }
    if (rn == null) {
        rn = rt.findNextRecordRowNode(rowNode, null);
        if (rn == null) {
            rn = rt.findPreviousRecordRowNode(rowNode, null);
        }
    }
    if (rn == null) {
        rn = (RowNode) rowNode.getParent();
        if (rn == (RowNode) rt.getRoot()) {
            rn = null;
        }
    }
    if (rn != null) {
        // move current cell postion to new row node
        newRowNode = rn;
    }
    if (newRowNode == null) {
        grid.getController().setCurrentCell(null);
    }

    return newRowNode;
}

protected boolean isAncestor(RowNode rowNode, long row) {
    if (rowNode.getBookmark() == row) {
        return true;
    }
    Vector ancestors = new Vector();
    rowNode.getDataTableModel().getAncestors(ancestors);
    for (int index = 0; index < ancestors.size(); index++) {
        if (row == ((Long)ancestors.elementAt(index)).longValue()) {
            return true;
        }
    }
    return false;
}

}
