/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: FormatNode.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:19 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * The FormatNode class aggregates all of the information necessary to render
 * a given table.  It contains RowFormat objects for each of the five basic
 * RowFormat subclasses: Record, Header, Footer, BeforeDetails, AfterDetails.
 * FormatNodes are stored in a tree structure called the {@link FormatTree}
 * The FormatTree serves as a design-time template for the run-time rendering of 
 * an instance of HiGrid.
 */
public class FormatNode extends DataSourceTreeNode {

static final long serialVersionUID = -6170167103182514528L;

/** Specifies any RowFormat */
public static final int ANY_FORMAT = 0;

/** Specifies a RecordFormat */
public static final int RECORD_FORMAT = 1;

/** Specifies a HeaderFormat */
public static final int HEADER_FORMAT = 2;

/** Specifies a FooterFormat */
public static final int FOOTER_FORMAT = 3;

/** Specifies a BeforeDetailsFormat */
public static final int BEFORE_DETAILS_FORMAT = 4;

/** Specifies a AfterDetailsFormat */
public static final int AFTER_DETAILS_FORMAT = 5;

/** Specifies the first RowFormat */
public static final int FIRST_FORMAT = RECORD_FORMAT;

/** Specifies the last RowFormat */
public static final int LAST_FORMAT = AFTER_DETAILS_FORMAT;

/** Never expand any of this row's children */
public static final int AUTO_TRAVERSE_NONE = 0;

/** Always expand any of this row's children */
public static final int AUTO_TRAVERSE_EXPAND = 1;

/** Expand and collapse any of this row's children */
public static final int AUTO_TRAVERSE_DETECT_CHILDREN = 2;

/** Rely on callback to specify if this row has children */
public static final int AUTO_TRAVERSE_CALLBACK = 3;

// Properties
private RecordFormat recordFormat = null;
private HeaderFormat headerFormat = null;
private FooterFormat footerFormat = null;
private BeforeDetailsFormat beforeDetailsFormat = null;
private AfterDetailsFormat afterDetailsFormat = null;
private int autoTraverse = AUTO_TRAVERSE_NONE;
private SortData defaultSortData = null;
private int indent = 0;
transient private int numberOfScrollableRows = 0;

// Constructor
public FormatNode() {
    super(null);
}

public FormatNode(int indent) {
    super(null);
    setIndent(indent);
}

/**
 * Retrieves the indexed row format for this object.
 *
 * @param index the enumerated row format
 *
 * @return The indexed row format.
 */
public RowFormat getRowFormat(int index) {
    RowFormat rowFormat;
    switch (index) {
    case RECORD_FORMAT:
        rowFormat = getRecordFormat();
        break;
    case HEADER_FORMAT:
        rowFormat = getHeaderFormat();
        break;
    case FOOTER_FORMAT:
        rowFormat = getFooterFormat();
        break;
    case BEFORE_DETAILS_FORMAT:
        rowFormat = getBeforeDetailsFormat();
        break;
    case AFTER_DETAILS_FORMAT:
        rowFormat = getAfterDetailsFormat();
        break;
    default:
        rowFormat = null;
        break;
    }
    return rowFormat;
}

/**
 * Sets the indexed row format for this object.
 *
 * @param format The new row format.
 * @param index the enumerated row format
 */
public void setRowFormat(RowFormat format, int index) {
    switch (index) {
    case RECORD_FORMAT:
        setRecordFormat((RecordFormat)format);
        break;
    case HEADER_FORMAT:
        setHeaderFormat((HeaderFormat)format);
        break;
    case FOOTER_FORMAT:
        setFooterFormat((FooterFormat)format);
        break;
    case BEFORE_DETAILS_FORMAT:
        setBeforeDetailsFormat((BeforeDetailsFormat)format);
        break;
    case AFTER_DETAILS_FORMAT:
        setAfterDetailsFormat((AfterDetailsFormat)format);
        break;
    default:
        //FIX
        // Could/should throw exception here
        break;
    }
}

/**
 * Retrieves the record format for this object.
 *
 * @return The current record format.
 */
public RecordFormat getRecordFormat() {
    return recordFormat;
}

/**
 * Sets the record format for this object.
 *
 * @param format The new record format.
 */
public void setRecordFormat(RecordFormat format) {
    recordFormat = format;
    recordFormat.setParent(this);
}

/**
 * Retrieves the header format for this object.
 *
 * @return The current header format.
 */
public HeaderFormat getHeaderFormat() {
    return headerFormat;
}

/**
 * Sets the header format for this object.
 *
 * @param format The new header format.
 */
public void setHeaderFormat(HeaderFormat format) {
    headerFormat = format;
    headerFormat.setParent(this);
}

/**
 * Retrieves the footer format for this object.
 *
 * @return The current footer format.
 */
public FooterFormat getFooterFormat() {
    return footerFormat;
}

/**
 * Sets the footer format for this object.
 *
 * @param format The new footer format.
 */
public void setFooterFormat(FooterFormat format) {
    footerFormat = format;
    footerFormat.setParent(this);
}

/**
 * Retrieves the before details format for this object.
 *
 * @return The current before details format.
 */
public BeforeDetailsFormat getBeforeDetailsFormat() {
    return beforeDetailsFormat;
}

/**
 * Sets the before details format for this object.
 *
 * @param format The new before details format.
 */
public void setBeforeDetailsFormat(BeforeDetailsFormat format) {
    beforeDetailsFormat = format;
    beforeDetailsFormat.setParent(this);
}

/**
 * Retrieves the after details format for this object.
 *
 * @return The current after details format.
 */
public AfterDetailsFormat getAfterDetailsFormat() {
    return afterDetailsFormat;
}

/**
 * Sets the after details format for this object.
 *
 * @param format The new after details format.
 */
public void setAfterDetailsFormat(AfterDetailsFormat format) {
    afterDetailsFormat = format;
    afterDetailsFormat.setParent(this);
}

/**
 * Retrieves the name for this format node.
 *
 * @return The current name.
 */
public String getName() {
    if (getRecordFormat() == null) {
        return null;
    }
    String string = getRecordFormat().getMetaData().getDescription();
    if (string == null || string.length() == 0) {
        int level = getRecordFormat().getMetaData().getMetaID() + 1;
        string = new String(com.klg.jclass.higrid.LocaleBundle.string(com.klg.jclass.higrid.LocaleBundle.level)+" "+level);
    }
    return string;
}

/** 
 * Create a level name given a level value.
 *
 * @param level An arbitrary level value.
 *
 * @return The level name.
 */

protected String makeLevelName(int level) {
    return(new String(com.klg.jclass.higrid.LocaleBundle.string(
                                                               com.klg.jclass.higrid.LocaleBundle.level)+" "+level));
}

/**
 * Retrieves whether to automatically traverse this level during
 * runtime table creation.
 *
 * @return Whether level is automatically traversed during runtime table creation.
 */
public int getAutoTraverse() {
    return autoTraverse;
}

/**
 * Sets whether to automatically traverse this level during
 * runtime table creation.
 *
 * @param traverse Whether level is automatically traversed during runtime table creation.
 */
public void setAutoTraverse(int traverse) {
    switch (traverse) {
    case AUTO_TRAVERSE_NONE:
    case AUTO_TRAVERSE_EXPAND:
    case AUTO_TRAVERSE_DETECT_CHILDREN:
    case AUTO_TRAVERSE_CALLBACK:
        autoTraverse = traverse;
        break;
    }
}

/**
 * Retrieves the indent level.
 *
 * @return The indent level.
 */
public int getIndent() {
    return indent;
}

/**
 * Sets the indent level.
 *
 * @param indent The indent level.
 */
public void setIndent(int indent) {
    this.indent = indent;
}

/**
 * Retrieves the default sort data.
 *
 * @return The default sort data.
 */
public SortData getDefaultSortData() {
    return defaultSortData;
}

/**
 * Sets the default sort data.
 *
 * @param sort The default sort data.
 */
public void setDefaultSortData(SortData sortData) {
    defaultSortData = sortData;
}

/**
 * Retrieves the total number of scrollable rows for this FormatNode.
 *
 * @return The total number of scrollable rows for this FormatNode.
 */
protected int getNumberOfScrollableRows() {
    return numberOfScrollableRows;
}

/**
 * Sets the total number of scrollable rows for this FormatNode.
 *
 * @param rows The total number of scrollable rows for this FormatNode.
 */
protected void setNumberOfScrollableRows(int rows) {
    numberOfScrollableRows = rows;
    if (numberOfScrollableRows < 0) {
        numberOfScrollableRows = 0;
    }
}

/**
 * Adds to the total number of scrollable rows for this FormatNode.
 *
 * @param rows The number of scrollable rows to for this FormatNode.
 */
protected void addNumberOfScrollableRows(int rows) {
    setNumberOfScrollableRows(numberOfScrollableRows + rows);
}

/**
 * Subtracts from the total number of scrollable rows for this FormatNode.
 *
 * @param rows The number of scrollable rows to subtract for this FormatNode.
 */
protected void subtractNumberOfScrollableRows(int rows) {
    setNumberOfScrollableRows(numberOfScrollableRows - rows);
}

/**
 * Reset the runtime info for this format node.
 */
protected void resetRuntimeInfo() {
    numberOfScrollableRows = 0;
}

/**
 * Retrieves the total width of this level.
 *
 * @return The total width of this level.
 */
protected int getTotalWidth(HiGrid grid) {
    int width = 0;

    if (getNumberOfScrollableRows() == 0) {
        return width;
    }

    if (recordFormat != null) {
        width = Math.max(width, recordFormat.getTotalWidth(grid, this));
    }
    if (headerFormat != null) {
        width = Math.max(width, headerFormat.getTotalWidth(grid, this));
    }
    if (footerFormat != null) {
        width = Math.max(width, footerFormat.getTotalWidth(grid, this));
    }
    // before and after details formats are for the previous level
    FormatNode lnParent = (FormatNode) getParent();
    if (lnParent != null) {
        if (lnParent.getBeforeDetailsFormat() != null) {
            width = Math.max(width, lnParent.getBeforeDetailsFormat().getTotalWidth(grid, this));
        }
        if (lnParent.getAfterDetailsFormat() != null) {
            width = Math.max(width, lnParent.getAfterDetailsFormat().getTotalWidth(grid, this));
        }
    }

    return width;
}

}
