/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: FormatTree.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:19 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.cell.JCCellInfo;
import com.klg.jclass.datasource.*;

/** 
 * The <code>FormatTree</code> holds the design-time visual formatting information
 * for all of the levels of HiGrid. It is composed of {@link FormatNode}
 * objects that aggregate all of the information necessary to render
 * a given table.
 */

public class FormatTree extends DataSourceTreeModel {

static final long serialVersionUID = 8410535555861981597L;

// array indices 
protected final static int INDEX_RECORD           = 0;
protected final static int INDEX_FOOTER           = 1;
protected final static int INDEX_BEFORE_DETAILS   = 2;
protected final static int INDEX_AFTER_DETAILS    = 3;
protected final static int METADATA_ARRAY_SIZE    = 4;

transient private HiGrid grid;

/** 
 * Default constructor.
 */
public FormatTree(HiGrid grid) {
//    super(new DataSourceTreeNode());
    super(DataSourceTreeModel.NullRoot);
    this.grid = grid;
}

/**
 * (Re-)creates a default FormatTree from the passed DataModel's MetaDataModel.
 * @param dataModel the DataModel from which the Format Tree is created.
 */
public void createFormatTreeFromDataModel(DataModel dataModel) {
    // create accompanying summary meta data models
    ((HiGridData) dataModel).createSummaryMetaTree();

    // get the root meta data 
    MetaDataModel[] root = new MetaDataModel[METADATA_ARRAY_SIZE];
    root[INDEX_RECORD] = (MetaDataModel) dataModel.getMetaDataTree().getRoot();
    root[INDEX_FOOTER] = (MetaDataModel) ((HiGridData) dataModel).getFooterMetaTree().getRoot();
    root[INDEX_BEFORE_DETAILS] = (MetaDataModel) ((HiGridData) dataModel).getBeforeDetailsMetaTree().getRoot();
    root[INDEX_AFTER_DETAILS] = (MetaDataModel) ((HiGridData) dataModel).getAfterDetailsMetaTree().getRoot();

    // create the root format node
    FormatNode formatNode = new FormatNode(0);
    setRoot(formatNode);
    createFormats(formatNode, root);

    // create the rest of the format tree
    createFormatTreeFromNode(formatNode, root, formatNode.getIndent());

    // inform the interested parties 
    grid.fireHiGridUpdateEvent(new HiGridUpdateEvent(grid,
                                                     HiGridUpdateEvent.AFTER_RESET_FORMAT_DATA));
}

/**
 * Recursive helper routine for createFormatTreeFromDataModel() above
 *
 * @param node The FormatNode which describes the visual format of the created rows.
 * @param rootMetaData The array of root meta data for the format node.
 * @param indent The nesting level of the passed node.
 */
protected void createFormatTreeFromNode(FormatNode node, MetaDataModel[] rootMetaData, int indent) {
    MetaDataModel[] arrayMetaData = new MetaDataModel[METADATA_ARRAY_SIZE];
    indent++;
    TreeIterator tiRecord = ((DataSourceTreeNode) rootMetaData[INDEX_RECORD]).getIterator();
    TreeIterator tiFooter = ((DataSourceTreeNode) rootMetaData[INDEX_FOOTER]).getIterator();
    TreeIterator tiBeforeDetails = ((DataSourceTreeNode) rootMetaData[INDEX_BEFORE_DETAILS]).getIterator();
    TreeIterator tiAfterDetails = ((DataSourceTreeNode) rootMetaData[INDEX_AFTER_DETAILS]).getIterator();
    // assume the same hierarchy for all the meta data trees
    while (tiRecord.hasMoreElements()) {
        // get the meta data 
        arrayMetaData[INDEX_RECORD] = (MetaDataModel) tiRecord.get();
        arrayMetaData[INDEX_FOOTER] = (MetaDataModel) tiFooter.get();
        arrayMetaData[INDEX_BEFORE_DETAILS] = (MetaDataModel) tiBeforeDetails.get();
        arrayMetaData[INDEX_AFTER_DETAILS] = (MetaDataModel) tiAfterDetails.get();

        // create the root format node
        FormatNode nextNode = new FormatNode(indent);
        createFormats(nextNode, arrayMetaData);
        node.add(nextNode);

        createFormatTreeFromNode(nextNode, arrayMetaData, indent);
        tiRecord.nextElement();
        tiFooter.nextElement();
        tiBeforeDetails.nextElement();
        tiAfterDetails.nextElement();
    }
}

/**
 * Reset the runtime info from the format tree
 */
protected void resetRuntimeInfo() {
    resetRuntimeInfo((FormatNode) getRoot());
}

/**
 * Reset the runtime info for a given format node and its children
 *
 * @param node The format node.
 */
protected void resetRuntimeInfo(FormatNode node) {
    if (node == null) {
        return;
    }
    node.resetRuntimeInfo();
    TreeIterator i = node.getIterator();
    while (i.hasMoreElements()) {
        FormatNode formatNode = (FormatNode)i.get();
        resetRuntimeInfo(formatNode);
        i.nextElement();
    }
}

// Ripped from BWTUtil.java
/** Accounts for bug in AWT which does not account for italic fonts. */
private int getStringWidth(FontMetrics fm, Font font, String s) {
    return fm.stringWidth(s) + (font.isItalic() ? font.getSize()/3+1 : 0);
}

//FIX createFormats() and setSummaryFormat() should be unified

/**
 * Creates the specified RowFormat objects described by the passed FormatNode,
 *
 * @param node The FormatNode which describes the visual format of its created rows.
 * @param arrayMetaData The array of meta data for the current format node.
 */
protected void createFormats(FormatNode node, MetaDataModel[] arrayMetaData) {
    MetaDataModel metaData = arrayMetaData[INDEX_RECORD];
    grid.fireHiGridFormatNodeEvent(new HiGridFormatNodeEvent(grid,
                                                     HiGridFormatNodeEvent.BEFORE_CREATE_FORMAT_NODE_CONTENTS, node,
                                                     metaData));

    // create the Header and Record Formats by default
    // and associate them with the passed FormatNode
    int numberOfColumns = 0;
    try {
        numberOfColumns = metaData.getColumnCount();
    }
    catch (Exception e) {
        grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
    }
    Vector cellFormatVector;

    // first, the Header Format
    HeaderFormat headerFormat = new HeaderFormat();

    // create the indicator columns
    headerFormat.createIndicatorFormats(grid);

    cellFormatVector = new Vector(numberOfColumns);
    // create the columns from the datasource
    for (int i = 0; i < numberOfColumns; i++) {
        CellFormat cellFormat = new CellFormat(grid.getHeaderCellStyle());
        cellFormat.setParent(headerFormat);
        String name = null;
        try {
            name = metaData.getColumnIdentifier(i);
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        cellFormat.setName(name);
        String text = null;
        try {
            text = metaData.getColumnObject(name).getColumnLabel();
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        cellFormat.setText(text);
        cellFormat.setDefaultSize(grid, true);
        cellFormat.setCellRenderer(grid.lookupTable.getHeaderRenderer().getClass());
        cellFormatVector.addElement(cellFormat);
    }
    headerFormat.setMetaData(metaData);
    headerFormat.setDataFormats(cellFormatVector);
    node.setHeaderFormat(headerFormat);

    // now, the Record Format
    RecordFormat recordFormat = new RecordFormat();

    // create the indicator columns
    recordFormat.createIndicatorFormats(grid);

    cellFormatVector = new Vector(numberOfColumns);
    // create the columns from the datasource
    for (int i = 0; i < numberOfColumns; i++) {
        CellFormat cellFormat = new CellFormat(grid.getRecordCellStyle());
        cellFormat.setParent(recordFormat);
        String name = null;
        try {
            name = metaData.getColumnIdentifier(i);
            ColumnModel cm = metaData.getColumnObject(name);
            // CellFormat will check against datasource for editability 
            cellFormat.setEditable(true);
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        cellFormat.setName(name);
        cellFormat.setWidth(((CellFormat)headerFormat.getDataFormats().elementAt(i)).getWidth());
        Font font = cellFormat.getFont();
        FontMetrics fm = grid.getFontMetrics(font);
        cellFormat.setFont(fm, font);
        try {
            cellFormat.setType(metaData.getJavaColumnType(name));
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        cellFormat.setCellRenderer(grid.lookupTable.getDefaultCellRenderer(cellFormat.getType()));
        cellFormat.setCellEditor(grid.lookupTable.getDefaultCellEditor(cellFormat.getType()));
        cellFormatVector.addElement(cellFormat);
    }
    recordFormat.setMetaData(metaData);
    recordFormat.setDataFormats(cellFormatVector);
    node.setRecordFormat(recordFormat);

    // create the initial Footer Format
    FooterFormat footerFormat = new FooterFormat();
    footerFormat.setMetaData(arrayMetaData[INDEX_FOOTER]);
    node.setFooterFormat(footerFormat);

    // create the initial BeforeDetails Format
    BeforeDetailsFormat beforeDetailsFormat = new BeforeDetailsFormat();
    beforeDetailsFormat.setMetaData(arrayMetaData[INDEX_BEFORE_DETAILS]);
    node.setBeforeDetailsFormat(beforeDetailsFormat);

    // create the initial AfterDetails Format
    AfterDetailsFormat afterDetailsFormat = new AfterDetailsFormat();
    afterDetailsFormat.setMetaData(arrayMetaData[INDEX_AFTER_DETAILS]);
    node.setAfterDetailsFormat(afterDetailsFormat);

    grid.fireHiGridFormatNodeEvent(new HiGridFormatNodeEvent(grid,
                                                     HiGridFormatNodeEvent.AFTER_CREATE_FORMAT_NODE_CONTENTS, node,
                                                     metaData));
}

/**
 * Creates summary format information for the passed rowFormat
 * based on the columns in the passed metaData.
 * @param rowFormat the summary format for which to create the cell information.
 * @param metaData the meta data that contains the column information.
*/

public void setSummaryFormat(RowFormat rowFormat, MetaDataModel metaData) {
    setSummaryFormat(rowFormat, metaData, grid.getRecordCellStyle());
}

/**
 * Creates summary format information for the passed rowFormat
 * based on the columns in the passed metaData.
 * @param rowFormat the summary format for which to create the cell information.
 * @param metaData the meta data that contains the column information.
 * @param cellStyle the cellStyle to use when creating the cellFormats.
*/

public void setSummaryFormat(RowFormat rowFormat, MetaDataModel metaData,
                             CellStyleModel cellStyle) {
    int numberOfColumns = 0;
    try {
        numberOfColumns = metaData.getColumnCount();
    }
    catch (Exception e) {
        grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
    }

    // create the indicator columns
    rowFormat.createIndicatorFormats(grid);

    Vector cellFormatVector = new Vector(numberOfColumns);
    // create the columns from the datasource
    for (int i = 0; i < numberOfColumns; i++) {
        CellFormat cellFormat = new CellFormat(cellStyle);
        cellFormat.setParent(rowFormat);
        Font font = cellFormat.getFont();
        FontMetrics fm = grid.getFontMetrics(font);
        cellFormat.setFont(fm, font);
        String name = null;
        try {
            name = metaData.getColumnIdentifier(i);
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        cellFormat.setName(name);
        cellFormat.setWidth(getStringWidth(fm, font, name) + 
                            cellFormat.getMarginInsets().left + cellFormat.getMarginInsets().right +
                            cellFormat.getBorderInsets().left + cellFormat.getBorderInsets().right );
        try {
            cellFormat.setType(metaData.getJavaColumnType(name));
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        cellFormat.setEditable(false);
        cellFormat.setCellRenderer(grid.lookupTable.getDefaultCellRenderer(cellFormat.getType()));
        cellFormatVector.addElement(cellFormat);
    }
    rowFormat.setDataFormats(cellFormatVector);
}

/**
 * Searches for the CellFormat with the given name in the passed Vector.
 * @param name the name of the cell to search for.
 * @param cellFormats a Vector of CellFormats to search.
 *
 * @return If found, CellFormat with given name, null otherwise.
 */
protected CellFormat findCellFormat(String name, Vector cellFormats) {
    for (Enumeration e = cellFormats.elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (cellFormat.getName().equals(name)) {
            return cellFormat;
        }
    }
    return null;    
}

/**
 * Searches for the CellFormat with the given name in the passed node.
 *
 * @param name the name of the cell to search for.
 * @param node the FormatNode to search within.
 * @param index the type of RowFormat to search. Constants are defined in FormatNode.
 *
 * @return If found, CellFormat with given name, null otherwise.
 */
public CellFormat searchRowFormat(String name, FormatNode node, int index) {
    if (node == null) {
        return null;
    }
    CellFormat foundItem = null;
    if (index == FormatNode.ANY_FORMAT) {
        for (int i = FormatNode.FIRST_FORMAT; i <= FormatNode.LAST_FORMAT; i++) {
            foundItem = findCellFormat(name, node.getRowFormat(i).getDataFormats());
            if (foundItem != null) {
                break;
            }
        }
    }
    else {
        foundItem = findCellFormat(name, node.getRowFormat(index).getDataFormats());
    }
    return foundItem;
}


/**
 * Searches for the CellFormat with the given name in the subtree of FormatNodes
 * rooted at startNode. If startNode is null, it starts at the root of the
 * this FormatTree.
 *
 * @param name the name of the cell to search for.
 * @param startNode the FormatNode to start searching at (null == start at tree root).
 * @param index the type of RowFormat to search. Constants are defined in FormatNode.
 *
 * @return If found, CellFormat with given name, null otherwise.
 */
public CellFormat findCellFormat(String name, FormatNode startNode, int index) {
    if (startNode == null) {
        startNode = (FormatNode)getRoot();
        if (startNode == null || 
            ((index < FormatNode.FIRST_FORMAT || index > FormatNode.LAST_FORMAT) &&
             (index != FormatNode.ANY_FORMAT))) {
            return null;
        }
    }
    for (Enumeration e = startNode.breadthFirstEnumeration(); e.hasMoreElements();) {
        FormatNode formatNode = (FormatNode) e.nextElement();
        CellFormat foundItem = searchRowFormat(name, formatNode, index);
        if (foundItem != null) {
            return foundItem;
        }
    }
    return null;
}

/*
public CellFormat findCellFormat(String name, FormatNode startNode, int index) {
CellFormat foundItem = null;
if (startNode == null) {
    startNode = (FormatNode)getRoot();
    if (startNode == null || 
        ((index < FormatNode.FIRST_FORMAT || index > FormatNode.LAST_FORMAT) &&
         (index != FormatNode.ANY_FORMAT))) {
        return null;
    }
}
foundItem = searchRowFormat(name, startNode, index);
if (foundItem != null) {
    return foundItem;
}
TreeIterator i = startNode.getIterator();
while (i.hasMoreElements()) {
    FormatNode formatNode = (FormatNode)i.get();
    if (formatNode == null) {
        continue;
    }
    foundItem = findCellFormat(name, formatNode, index);
    if (foundItem != null) {
        return foundItem;
    }
    i.nextElement();
}
return null;
}
*/

/**
 *  Traverses the entire FormatTree and re-initializes the renderer for all
 *  cells of type NodeCellFormat with the passed nodeRenderer.
 *
 * @param nodeRenderer the new HiGridNodeRenderer to use everywhere.
 */
public void resetAllNodeRenderers(HiGridNodeRenderer nodeRenderer) {
    resetAllRenderers(nodeRenderer.getClass(), grid.NODE_INDICATOR_CELL);
}

/**
 *  Traverses the entire FormatTree and re-initializes the renderer for all
 *  cells of type EditStatusCellFormat with the passed editStatusRenderer.
 *
 * @param editStatusRenderer the new HiGridEditStatusRenderer to use everywhere.
 */
public void resetAllEditStatusRenderers(HiGridEditStatusRenderer editStatusRenderer) {
    resetAllRenderers(editStatusRenderer.getClass(), grid.EDITSTATUS_INDICATOR_CELL);
}

/**
 * Helper method that traverses the entire FormatTree and re-initializes the 
 * renderer for all cells of the given indicatorFormats renderer type
 *
 * @param renderer the class of renderer to reset.
 * @param rendererIndex the indicatorFormats index.
 */
protected void resetAllRenderers(Class renderer, int rendererIndex) {
    int index;
    CellFormat nodeFormat;
    for (Enumeration e = ((FormatNode)getRoot()).breadthFirstEnumeration(); e.hasMoreElements();) {
        FormatNode formatNode = (FormatNode) e.nextElement();
        for (int i = FormatNode.FIRST_FORMAT; i <= FormatNode.LAST_FORMAT; i++) {
            index = grid.getIndicatorIndex(formatNode.getRowFormat(i).getIndicatorFormats(), rendererIndex);
            if (index >= 0) {
                nodeFormat = (CellFormat)formatNode.getRowFormat(i).getIndicatorFormats().elementAt(index);
                nodeFormat.setCellRenderer(renderer);
            }
        }
    }
}

/**
 *  Traverses the entire FormatTree and re-initializes the renderer for all
 *  cells of a row with type HeaderFormat with the passed headerRenderer.
 *
 * @param headerRenderer the new HiGridHeaderRenderer to use everywhere.
 */
public void resetAllHeaderRenderers(HiGridHeaderRenderer headerRenderer) {
    CellFormat cellFormat;
    Class renderer = headerRenderer.getClass();
    for (Enumeration e = ((FormatNode)getRoot()).breadthFirstEnumeration(); e.hasMoreElements();) {
        FormatNode formatNode = (FormatNode) e.nextElement();
        for (Enumeration e1 = formatNode.getHeaderFormat().getDataFormats().elements(); e1.hasMoreElements();) {
            cellFormat = (CellFormat) e1.nextElement();
            cellFormat.setCellRenderer(renderer);
        }
    }
}

/**
 * Traverses the entire FormatTree and re-initializes the PLAF for all cells.
 *
 * @param headerRenderer the new HiGridHeaderRenderer to use everywhere.
 */
public void resetPlaf() {
    if (getRoot() == null || grid == null) {
        return;
    }
    CellFormat cellFormat;
    for (Enumeration e = ((FormatNode)getRoot()).breadthFirstEnumeration(); e.hasMoreElements();) {
        FormatNode formatNode = (FormatNode) e.nextElement();
        for (int i = FormatNode.FIRST_FORMAT; i <= FormatNode.LAST_FORMAT; i++) {
            RowFormat rowFormat = formatNode.getRowFormat(i);
            CellStyleModel cellStyle = grid.getCellStyle(i);
            for (Enumeration e1 = rowFormat.getIndicatorFormats().elements(); e1.hasMoreElements();) {
                cellFormat = (CellFormat) e1.nextElement();
                cellFormat.copyPlafFromCellStyle(grid.getHeaderCellStyle());
                cellFormat.setSelectedBackground(cellFormat.getBackground());
                cellFormat.setSelectedForeground(cellFormat.getForeground());
            }
            for (Enumeration e2 = rowFormat.getDataFormats().elements(); e2.hasMoreElements();) {
                cellFormat = (CellFormat) e2.nextElement();
                cellFormat.copyPlafFromCellStyle(cellStyle);
            }
        }
    }
}

} 
