/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: GridArea.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:19 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.JComponent;
import javax.swing.JToolTip;

import com.klg.jclass.util.JCEnvironment;


 

/**
 * The <code>GridArea</code> is the component onto which the HiGrid
 * nodes are rendered. The GridArea, along with the vertical and
 * horizontal scrollbars exist within the HiGrid container.
 */
public class GridArea extends JComponent implements FocusListener {

static final long serialVersionUID = 8474639562992639548L;

protected static final int MINIMUM_WIDTH = 200;
protected static final int MINIMUM_HEIGHT = 200;

private HiGrid grid = null;
transient protected HiGridTip toolTip = null;
protected boolean isDoubleBuffer = true;
transient private Image doubleBufferImage;

protected FocusModel focusManager = null;

/**
 * Default Constructor
 */
public GridArea(HiGrid grid) {
    this.grid = grid;
    setLayout(null);
    setOpaque(true);
    // this minimizes redraws of the GridArea
    setDoubleBuffered(true);

    recreateToolTip();

	boolean fNewFocusModel =
				(JCEnvironment.getJavaVersion() >= 140);
	String focusManagerString = (fNewFocusModel) ?
	    				"com.klg.jclass.higrid.FocusManagerNew"
						: "com.klg.jclass.higrid.FocusManagerOld";
	try {
	    Class c = Class.forName(focusManagerString);
	    focusManager = (FocusModel)c.newInstance();
	    focusManager.setGridArea(this);
	}
	catch (Exception e) {
		System.out.println("FocusManager initialization error:" +
		                   e.getMessage());
	}

	if (fNewFocusModel) {
    	setFocusTraversalKeysEnabled(false);
	}
//    addFocusListener(this);   //  For debugging
}

/**
 * Computes the point at which the HiGridTip must be placed
 * for correct alignment.
 * @param event MouseEvent that stores current mouse position
 * @return the point at which the HiGridTip must be placed
 * for correct alignment.
 */
public Point getToolTipLocation(MouseEvent event) {
    recreateToolTip();
    if (grid.isHeaderTipVisible()) {
        return toolTip.getTip(new Point(event.getX(), event.getY()));
    }
    else {
        return super.getToolTipLocation(event);
    }
}

/**
 * Retrieves the HiGridTip.
 * @return the HiGridTip.
 */
public JToolTip createToolTip() {
    recreateToolTip();
    if (grid.isHeaderTipVisible()) {
        return toolTip;
    }
    else {
        return super.createToolTip();
    }
}

/**
 * Allows lazy creation of tooltips and permits
 * cut and paste serialization.
 */
protected void recreateToolTip() {
    if (toolTip == null) {
        toolTip = new HiGridTip(grid);
        toolTip.setComponent(this);
    }
}

/**
 * Returns the object used to lock layout and repaint.
 */
protected Object getAWTLock() {
    return getTreeLock();
}

/**
 * Returns the double-buffer image previously created in a call to
 * getDoubleBufferGraphics.
 * @see #getDoubleBufferGraphics
 */
protected Image getDoubleBufferImage() {
    return doubleBufferImage;
}

/**
 * Sets the image to be used as the double-buffer image.
 * @param image new double-buffer to use
 */
protected Image setDoubleBufferImage(Image image) {
    if (doubleBufferImage != null) {
        doubleBufferImage.flush();
    }
    doubleBufferImage = image;
    return doubleBufferImage;
}

/**
 * Checks whether the double-buffer needs to be reset/recreated,
 * based on HiGrid's RowTree.
 * @return true if image has been recreated
 */
protected boolean doubleBufferSizeChanged() {
    int width = Math.max(grid.getRowTree().getGridAreaWidth(), getSize().width);
    int height = getSize().height;
    if ((width <= 0) || (height <= 0)) {
        setDoubleBufferImage(null);
        return false;
    }
    Image image = getDoubleBufferImage();
    if (image == null ||
        image.getWidth(null) != width ||
        image.getHeight(null) != getSize().height) {
        image = setDoubleBufferImage(createImage(width, getSize().height));
        if (image != null) {
            return true;
        }
    }
    return false;
}

/** Creates a double-buffer graphics for drawing into the doubleBufferImage
 * @see #getDoubleBufferImage
 * @return the Graphics instance for the image
 */
protected Graphics getDoubleBufferGraphics() {
    synchronized (getAWTLock()) {
        Image image = getDoubleBufferImage();
        Graphics doubleBufferImageGc;
        if (image == null)
            doubleBufferImageGc = null;
        else
            doubleBufferImageGc = image.getGraphics();
        return doubleBufferImageGc;
    }
}

/**
 * Call back to HiGrid to render the grid.
 * @param gc Graphics Context to use for rendering.
 */
protected void paintGrid(Graphics gc) {
//System.out.println("paintGrid: Clip rect is " + gc.getClipBounds());
    grid.draw(gc);
}

/**
 * Repaints the entire grid
 */
protected void refresh() {
    refresh(0, 0, Math.max(grid.getRowTree().getGridAreaWidth(), getSize().width), getSize().height);
}

/**
 * Repaints the grid based on the passed rectangle coordinates
 */
protected void refresh(int x, int y, int width, int height) {
    if (isDoubleBuffer) {
        if (doubleBufferSizeChanged()) {
            refresh();
            return;
        }
        // need to redraw the double buffer
        Graphics gc = getDoubleBufferGraphics();
        if (gc == null) {
            // System.out.println("Bad news!");
            return;
        }
        gc.setClip(x, y, width, height);
        paintGrid(gc);
    }
}

/**
 * Repaints the grid based on the passed number of pixels.
 * @param pixels Based on its sign, will copy the given number of
 * pixels up or down within the double-buffer and expose the remainder
 * to be repainted as usual.
 */
protected void refresh(int pixels) {
    synchronized (getAWTLock()) {
        // sanity checks first
        if (pixels == 0) {
            return;
        }
        // this may be slightly more optimal than going through the code below,
        // depending on how efficient copyArea is
        if (Math.abs(pixels) >= getSize().height) {
            refresh();
            return;
        }
        if (isDoubleBuffer) {
            if (getDoubleBufferImage() != null) {
                if (doubleBufferSizeChanged()) {
                    refresh();
                    return;
                }
                Graphics gc = getDoubleBufferGraphics();
                Rectangle paintRect = new Rectangle();

                paintRect.setBounds(0, 0, Math.max(grid.getRowTree().getGridAreaWidth(), getSize().width), getSize().height);
                gc.setClip(paintRect);
//FIXOBSOLETE
/*
            if (false) {
//                if (JCEnvironment.isVisualAge()) {
                Graphics extraGc = getExtraDoubleBufferGraphics();
                extraGc.drawImage(getDoubleBufferImage(), 0, 0, null);
                gc.drawImage(extraDoubleBufferImage, 0, -pixels, null);
            }
            else {
System.out.println("drawing image");
                gc.drawImage(getDoubleBufferImage(), 0, -pixels, null);
            }
*/
                // move the image up within the double buffer
                gc.copyArea(0, pixels > 0 ? pixels : 0,
                            grid.getRowTree().getGridAreaWidth(), getSize().height - Math.abs(pixels), 0, -pixels);
                paintRect.setBounds(0, (pixels > 0) ? getSize().height - pixels : 0, Math.max(grid.getRowTree().getGridAreaWidth(),
                                                                                              getSize().width), Math.abs(pixels));
                gc.setClip(paintRect);
                // repaint the newly exposed area
                paintGrid(gc);

                gc.dispose();
            }
        }
    }
}

/**
 * Override for Swing painting model.
 * @param gc Graphics Context to paint into.
 */
public void paintComponent(Graphics gc) {
//System.out.println("paintComponent: Clip rect is " + gc.getClipBounds());
    synchronized (getAWTLock()) {
        if (isDoubleBuffer) {
            if (doubleBufferSizeChanged()) {
                refresh();
            }
            if (getDoubleBufferImage() != null) {
                //DEBUG this code gives some JFC DebugGraphics capability
/*
// debug start
            Rectangle clipRect = gc.getClipBounds();
            gc.setColor(java.awt.Color.cyan);
            for (int i = 0; i < 5; i++) {
                gc.fillRect(clipRect.x, clipRect.y, clipRect.width, clipRect.height);
            }
            // waste some time
            try {
                Thread.currentThread().sleep(100);
            }
            catch (InterruptedException e) { }
// debug end
*/
/*
//  Comment out the above code to remove the debugging behavior
*/
//System.out.println("drawImage: Clip rect is " + gc.getClipBounds());
                gc.drawImage(getDoubleBufferImage(), -grid.getHorizontalScrollbar().getValue(), 0, null);
                // now paint the part of the grid area that is that is not data and is
                // now exposed on the right hand side
                Rectangle r = gc.getClipBounds();
                gc.setColor(getBackground());
                int rightEdge = getDoubleBufferImage().getWidth(null) - grid.getHorizontalScrollbar().getValue();
                gc.fillRect(rightEdge, r.y, getSize().width - rightEdge, r.height);
            }
        }
        else {
            gc.translate(-grid.getHorizontalScrollbar().getValue(), 0);
            paintGrid(gc);
            gc.translate(+grid.getHorizontalScrollbar().getValue(), 0);
        }
    }

    // paint any children
    super.paintComponent(gc);


 
}


 

/**
 * Gets the preferred size of this component.
 * @return A dimension object indicating this component's preferred size.
 */
public Dimension getPreferredSize() {
    int width = grid.getRowTree().getFirstLevelGridAreaWidth();
    return new Dimension(width <= 0 ? MINIMUM_WIDTH : width, MINIMUM_HEIGHT);
}

/**
 * Moves and resizes this component. The new location of the top-left
 * corner is specified by <code>x</code> and <code>y</code>, and the
 * new size is specified by <code>width</code> and <code>height</code>.
 * @param x The new <i>x</i>-coordinate of this component.
 * @param y The new <i>y</i>-coordinate of this component.
 * @param width The new <code>width</code> of this component.
 * @param height The new <code>height</code> of this component.
 */
public void setBounds(int x, int y, int width, int height) {
    Rectangle oldRect = getBounds();
    super.setBounds(x,y,width,height);
    // Sometimes there is a race condition during startup and the grid is not drawn properly.
    // Try to detect a transition between an empty and non-empty gridArea size
    // and only refresh when the gridArea becomes a non-zero size and thus visible.
    if (oldRect.isEmpty() && width > 0 && height > 0) {
        grid.refresh();
    }
}

/**
 * Disposes of any native resources.
 */
public void dispose() {
    setDoubleBufferImage(null);
}

/**
 * Returns the current FocusManager.
 * @see #setFocusManager
 * @return The current FocusManager
 */
public FocusModel getFocusManager()
{
	return(focusManager);
}

/**
 * Sets a new FocusManager
 * @param focusManager The new focus manager.  Null values are ignored.
 */
public void setFocusManager(FocusModel focusManager)
{
	if (focusManager != null) {
		this.focusManager = focusManager;
	}
}

/*
public void requestFocus()
{
	Component oldFocusOwner =
		KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
	System.out.println("gridArea.requestFocus: current focus owner="
							+ oldFocusOwner);
	super.requestFocus();
}
*/

public void focusGained(FocusEvent event)
{
//    System.out.println("gridArea.focusGained()   event=" + event);
}

public void focusLost(FocusEvent event)
{
//    System.out.println("gridArea.focusLost()   event=" + event);
}

}
