/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: GridScrollbar.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:19 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import com.klg.jclass.datasource.DataTableModel;

import javax.swing.*;
import java.awt.*;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.io.Serializable;


/**
 * <CODE>GridScrollbar</CODE> manages the interaction between an actual
 * scrollbar (extends <CODE>Component</CODE>, implements
 * <CODE>Adjustable</CODE>) and the <CODE>HiGrid</CODE> object.
 *
 * @see Adjustable
 */
public class GridScrollbar implements Serializable, AdjustmentListener,
    GridScrollbarCallback {

static final long serialVersionUID = 2256552432959303071L;

/**
 * The horizontal orientation of the scrollbar.
 */
public final static int HORIZONTAL = java.awt.Scrollbar.HORIZONTAL;
/**
 * The vertical orientation of the scrollbar.
 */
public final static int VERTICAL = java.awt.Scrollbar.VERTICAL;

/**
 * The scrollbar displays as needed.
 */
public final static int DISPLAY_AS_NEEDED   = 0;
/**
 * The scrollbar always is displayed.
 */
public final static int DISPLAY_ALWAYS      = 1;
/**
 * The scrollbar is never displayed.
 */
public final static int DISPLAY_NEVER       = 2;

private final static int INITIAL_ROWTIP_WIDTH = 100;
protected final static int MAGIC_BLOCK_INCREMENT = 2;

private Adjustable adjustable;
private Component component;
private HiGrid grid;
private int display = DISPLAY_AS_NEEDED;
private int trackValue = -1;
// if set to -1, extra initial repaint when clicking on scrollbar
private int startValue = 0;
private int draggingValue = 0;
private RowNode oldTopRowNode = null;
private boolean waitingForTrack = false;
private boolean settingVisibility = false;
private JLabel rowtip = null;
private GridScrollbarCallback rowtipCallback = null;
private boolean rowtipVisible = true;

public GridScrollbar(HiGrid grid, int orientation) {
    this.grid = grid;
    JScrollBar ds = new JScrollBar(orientation);
    adjustable = (Adjustable)ds;
    component = ds;
    adjustable.addAdjustmentListener(this);
    if (orientation == VERTICAL) {
        rowtip = new JLabel();
        rowtip.setHorizontalAlignment(SwingConstants.CENTER);
        rowtip.setOpaque(true);
        rowtip.setFont(new Font("Serif", Font.PLAIN, 12));
    //  standard MS-like rowtip colors
        rowtip.setForeground(Color.black);
        rowtip.setBackground(new Color(255, 255, 225));
        setRowtipCallback(this);
    }
}

public Component getComponent() {
    return component;
}

public void setComponent(Component comp) {
    component = comp;
    grid.resetScrollbarConstraints();
    grid.layoutGrid();
//FIX - not sure if a refresh is needed here
}

public void setAdjustable(Adjustable adj) {
    adjustable.removeAdjustmentListener(this);
    adjustable = adj;
    adjustable.addAdjustmentListener(this);
}

public Adjustable getAdjustable() {
    return adjustable;
}

public void adjustmentValueChanged(java.awt.event.AdjustmentEvent event) {
    if (adjustable == null || component == null ||
        adjustable != event.getAdjustable()) {
//        System.out.println("Scrollbar in an odd state");
        return;
    }

    if (settingVisibility) {
        return;
    }

    int value = event.getValue();
    int type = event.getAdjustmentType();
    int orientation = adjustable.getOrientation();

/*
//System.out.println("Entering trackValue = " + trackValue + " ");

// debug: see type
switch (type) {
    case AdjustmentEvent.TRACK:
        System.out.println(value+" - TRACK");
        break;
    case AdjustmentEvent.BLOCK_DECREMENT:
        System.out.println(value+" - BLOCK_DECREMENT");
        break;
    case AdjustmentEvent.BLOCK_INCREMENT:
        System.out.println(value+" - BLOCK_INCREMENT");
        break;
    case AdjustmentEvent.UNIT_DECREMENT:
        System.out.println(value+" - UNIT_DECREMENT");
        break;
    case AdjustmentEvent.UNIT_INCREMENT:
        System.out.println(value+" - UNIT_INCREMENT");
        break;
    default:
        System.out.println(value+" - UNKNOWN");
        break;
}
*/

    // do jump scrolling by rows
    if (orientation == Adjustable.VERTICAL) {
        if (oldTopRowNode == null) {
            oldTopRowNode = grid.getRowTree().getTopRowNode();
        }
        boolean adjust = false;
        int increment = value - draggingValue;
        int action = grid.getController().getCurrentAction();
        if ((startValue == draggingValue) &&
            ((action == HiGridAction.NO_ACTION) ||
             (action == HiGridAction.MOVE_PAGE_DOWN_ACTION) ||
             (action == HiGridAction.MOVE_PAGE_UP_ACTION)) &&
            (Math.abs(increment) == getBlockIncrement())) {
            adjust = true;
            if (increment < 0) {
                // recover original value then page up
                value += MAGIC_BLOCK_INCREMENT;
                value = grid.getRowTree().doPageUp(value);
            }
            else if (increment > 0) {
                // recover original value then page down
                value -= MAGIC_BLOCK_INCREMENT;
                value = grid.getRowTree().doPageDown(value);
            }
        }
        value = getAdjustedValue(value);
        draggingValue = value;
        if (adjustable instanceof JScrollBar) {
            if (((JScrollBar)adjustable).getValueIsAdjusting()) {
                if (rowtip.getParent() == null) {
//System.out.println("Adding tip " + value);
                    addRowtip(value);
                }
                else {
//System.out.println("Updating tip " + value);
                    trackValue = value;
                    updateRowtip(value);
                    placeRowtip();
                }
            }
            else {
//System.out.println("Removing tip " + value);
                adjust = value != startValue;
                trackValue = -1;
                removeRowtip();
                if (!adjust) {
                    grid.refresh(false);
                }
            }
        }
        else {
            if (trackValue == -1 && !waitingForTrack) {
                // user just clicked on scrollbar
                addRowtip(value);
            }
            if (trackValue == value) {
                adjust = value != startValue;
                trackValue = -1;
                grid.getGridArea().remove(rowtip);
            }
            else {
                trackValue = value;
                updateRowtip(value);
                placeRowtip();
            }
        }
        if (adjust) {
            removeRowtip();
            grid.getRowTree().setTopRow(value);
            boolean downward = value > startValue;
            grid.refresh(oldTopRowNode, grid.getRowTree().getTopRowNode(), downward);
            Controller controller = grid.getController();
            controller.showEditor(false);
            controller.positionEditor();
            controller.showEditor(true);
            controller.acquireFocusForCellEditor();
            waitingForTrack = false;
            // must also reset *after* refresh(), since refresh() will call us recursively
            trackValue = -1;
            startValue = value;
            draggingValue = value;
            oldTopRowNode = grid.getRowTree().getTopRowNode();
        }
    }

    // do smooth scrolling by pixels
    else if (orientation == Adjustable.HORIZONTAL) {
        if (trackValue != value) {
            trackValue = value;
            adjustable.setValue(value);
            grid.refresh(false);
            grid.getController().acquireFocusForCellEditor();
        }
    }
//System.out.println("Leaving trackValue = " + trackValue + " ");
}

public int getDisplay() {
    return display;
}

public void setDisplay(int display) {
    this.display = display;
    grid.layoutGrid();
}

public void setVisible(boolean b) {
    component.setVisible(b);
}

public boolean isVisible() {
    return component.isVisible();
}

public void setSize(Dimension d) {
    component.setSize(d);
//    try{
//        if (component instanceof JScrollBar) {
//            JScrollBar sb = (JScrollBar) component;
//            if (sb.getUI() instanceof LayoutManager){
//                LayoutManager sbLayoutManager = (LayoutManager) sb.getUI();
//                sbLayoutManager.layoutContainer(sb);
//            }
//        }
//    }
//    catch (Exception e){
//        Debug.print("Failed trying to layoutContainer after setting size: \n" + e.getMessage());
//        Debug.print("Trying to update componentUI instead");
//        ComponentUI cui = UIManager.getUI((JComponent)component);
//        cui.update(component.getGraphics(), (JComponent)component);
//    }
}

public void setUnitIncrement(int inc) {
    // override to prevent unit increment being set
}

public int getUnitIncrement() {
    return adjustable.getUnitIncrement();
}

public void setBlockIncrement(int inc) {
    // override to prevent block increment being set
}

public int getBlockIncrement() {
    return adjustable.getBlockIncrement();
}

public void setValues(int value, int visible, int min, int max) {
    adjustable.setMinimum(min);
    adjustable.setMaximum(max);
    adjustable.setVisibleAmount(visible);
    adjustable.setValue(value);
}

public void setValue(int value) {
    if (value == adjustable.getValue()) {
        return;
    }
    adjustable.setValue(value);
    if (adjustable.getOrientation() == HORIZONTAL) {
        adjustmentValueChanged(new AdjustmentEvent(adjustable,
                                                   AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, AdjustmentEvent.TRACK, value));
    }
    // this is to solve a RowTip bug, where keyboard positioning screws up the
    // startValue variable
    // this code is tricky because it gets called recursively
//FIXTRICKYTEST
/*
if (adjustable.getOrientation() == VERTICAL) {
    startValue = value;
}
*/
}

public int getValue() {
    return adjustable.getValue();
}

public int getOrientation() {
    return adjustable.getOrientation();
}

public void setMinimum(int min) {
    adjustable.setMinimum(min);
}

public int getMinimum() {
    return adjustable.getMinimum();
}

public void setMaximum(int max) {
    adjustable.setMaximum(max);
}

public int getMaximum() {
    return adjustable.getMaximum();
}

public boolean atEnd() {
    int value = adjustable.getValue();
    int max = adjustable.getMaximum();
    int slider = adjustable.getVisibleAmount();
    return(value == 0 || value >= max - slider);
}

public void scrollLineDown() {
    int value = getValue() + getUnitIncrement();
    int type = AdjustmentEvent.UNIT_INCREMENT;
    adjustmentValueChanged(new AdjustmentEvent(getAdjustable(),
                                               AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, type, value));
}

public void scrollLineUp() {
    int value = getValue() - getUnitIncrement();
    int type = AdjustmentEvent.UNIT_DECREMENT;
    adjustmentValueChanged(new AdjustmentEvent(getAdjustable(),
                                               AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, type, value));
}

public void scrollPageDown() {
    int value = getValue() + getBlockIncrement();
    int type = AdjustmentEvent.BLOCK_INCREMENT;
    adjustmentValueChanged(new AdjustmentEvent(getAdjustable(),
                                               AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, type, value));
}

public void scrollPageUp() {
    int value = getValue();
    if (value == 0) {
        // we're home already, nothing to do
        return;
    }
    value -= getBlockIncrement();
    if (value <= 0) {
        setValue(0);
        return;
    }
    int type = AdjustmentEvent.BLOCK_DECREMENT;
    adjustmentValueChanged(new AdjustmentEvent(getAdjustable(),
                                               AdjustmentEvent.ADJUSTMENT_VALUE_CHANGED, type, value));
}

public void initValues() {
    adjustable.setMinimum(0);
    adjustable.setMaximum(100);
    adjustable.setValue(0);
    adjustable.setUnitIncrement((adjustable.getOrientation()
                                 == Adjustable.HORIZONTAL) ? 10 : 1);
//    adjustable.setBlockIncrement(1);
    adjustable.setBlockIncrement(MAGIC_BLOCK_INCREMENT);
    adjustable.setVisibleAmount(10);
}

public void adjustValues(int maximum, int visible) {
//System.out.println("max = " + maximum + " old max = " + adjustable.getMaximum() + " vis = " + visible + " old vis = " + adjustable.getVisibleAmount());
    if (visible >= maximum ||
        (maximum == adjustable.getMaximum() &&
         visible == adjustable.getVisibleAmount())) {
        return;
    }

    adjustable.setMaximum(maximum);
    settingVisibility = true;
    adjustable.setVisibleAmount(visible);
    settingVisibility = false;
    if (adjustable.getOrientation() == Adjustable.HORIZONTAL) {
        adjustable.setBlockIncrement(visible);
    }

/*
// debug: see type
System.out.println("max: "+adjustable.getMaximum());
System.out.println("vis: "+adjustable.getVisibleAmount());
System.out.println("blk: "+adjustable.getBlockIncrement());
*/
}

protected int getAdjustedValue(int value) {
    if (value < adjustable.getMinimum()) {
        value = adjustable.getMinimum();
    }
    if (value > adjustable.getMaximum() - adjustable.getVisibleAmount()) {
        value = adjustable.getMaximum() - adjustable.getVisibleAmount();
    }
    return value;
}

protected RowNode getLabelRowNode(int value) {
    RowTree rowTree = grid.getRowTree();
    if (rowTree == null || rowTree.getRoot() == null) {
        return null;
    }
    RowNodeWalk walkable = new RowNodeWalk(value);
    rowTree.walk((RowNode)rowTree.getRootNode().getFirstChild(), walkable);
    return walkable.getFoundNode();
}

public GridScrollbarCallback getRowtipCallback() {
    return rowtipCallback;
}

public void setRowtipCallback(GridScrollbarCallback callback) {
    rowtipCallback = callback;
}

public String getRowtipText() {
    return rowtip.getText();
}

// access to Component for overriding font, color, etc.
public Component getRowtip() {
    return rowtip;
}

protected boolean isRowtipVisible() {
    return rowtipVisible;
}

protected void setRowtipVisible(boolean visible) {
    rowtipVisible = visible;
}

protected void removeRowtip() {
    grid.getGridArea().remove(rowtip);
}

protected void addRowtip(int value) {
    if (rowtipCallback != null && isRowtipVisible()) {
        // clear the text in case it was set previously
        rowtip.setText("");
        rowtip.setVisible(false);
        updateRowtip(value);
        placeRowtip();
        // make sure the rowtip is on top; use 0 = topmost for index
        grid.getGridArea().add(rowtip, 0);
        rowtip.setVisible(true);
    }
}

protected void placeRowtip() {
    if (rowtipCallback != null && isRowtipVisible()) {
        rowtipCallback.positionRowtip(grid.getGridArea(), rowtip);
    }
}

protected void updateRowtip(int currentValue) {
    if (rowtipCallback != null && isRowtipVisible()) {
        RowNode rowNode = getLabelRowNode(currentValue);
        if (rowNode == null) {
            return;
        }
        String s = rowtipCallback.setRowtipText(rowNode, currentValue, adjustable);
        // keep the previous value, if a null is returned
        if (s != null) {
            rowtip.setText(s);
        }
    }
}

// default implementation of GridScrollbarCallback interface
public String setRowtipText(RowNode rowNode, int currentValue, Adjustable adjustable) {
/*
    // this code simply returns the row index value
    return Integer.toString(currentValue);
*/
    // this code returns the value within the first cell, if it is a row format
    String result = null;
    DataTableModel dataTable = rowNode.getDataTableModel();
    RowFormat rowFormat = rowNode.getRowFormat();
    if (rowNode.getRowFormat().getDataFormats().isEmpty()) {
        return "";
    }
    CellFormat cellFormat = (CellFormat) rowNode.getRowFormat().getDataFormats().elementAt(0);
    String name = cellFormat.getName();
    if (rowFormat instanceof RecordFormat) {
        Object o = null;
        try {
            o = dataTable.getResultData(rowNode.getBookmark(), name);
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        if (o != null) {
            result = o.toString();
            // could optionally setBounds here based on string width and height
//            getRowtip().setBounds(0, 0, getStringWidth(), getFontHeight());
        }
    }

    if (result == null) {
        // just use the column name
        result = name;
    }
    return result;
}

public void positionRowtip(Container parent, Component rowtip) {
    ((JLabel)rowtip).setHorizontalAlignment(SwingConstants.CENTER);
    ((JLabel)rowtip).setOpaque(true);
    rowtip.setFont(new Font("Serif", Font.PLAIN, 12));
//  standard MS-like rowtip colors
    rowtip.setForeground(Color.black);
    rowtip.setBackground(new Color(255, 255, 225));
//  these colors conflict with normal row selection colors
//    rowtip.setForeground(SystemColor.textHighlightText);
//    rowtip.setBackground(SystemColor.textHighlight);
    // start with the top left corner; this is always visible
    int x = 0;
    int y = 0;
    if (parent != null) {
        Dimension containerDimension = parent.getSize();
        int height = getFontHeight(rowtip.getFont());
        int width = getStringWidth(rowtip.getFont(), getRowtipText());
        width = Math.max(width, INITIAL_ROWTIP_WIDTH);
        width = Math.min(width, containerDimension.width);
        rowtip.setSize(width, height);
        // rightmost justified, with 1 pixel spacing
        x = containerDimension.width - width - 1;
        // centered vertically
        y = (containerDimension.height - height) / 2;
        rowtip.setLocation(x, y);
        rowtip.setVisible(true);
    }
}

/**
 * Remove references to old DataModels held
 * by oldTopRowNode.
 */
protected void removeOldReferences() {
	oldTopRowNode = null;
}

// Ripped from BWTUtil.java
/** Accounts for bug in AWT which does not account for italic fonts. */
private int getStringWidth(Font font, String s) {
    FontMetrics fm = getComponent().getFontMetrics(font);
    return fm.stringWidth(s) + (font.isItalic() ? font.getSize()/3+1 : 0);
}

private int getFontHeight(Font font) {
    FontMetrics fm = getComponent().getFontMetrics(font);
    return fm.getLeading() + fm.getMaxAscent() + fm.getMaxDescent();
}

}
