/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: HiGridAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:19 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.awt.event.*;
import java.io.Serializable;

/**
 * <CODE>HiGridAction</CODE> is responsible for defining a
 * mapping from a user event to an action performed on HiGrid.
 * For example, <CODE>HiGridAction</CODE> can be used to map
 * a <STRONG>Control LeftClick</STRONG> to a selection event.
 */
public class HiGridAction implements Serializable {

static final long serialVersionUID = 2539170986365960859L;

/** initiator property. */
protected ActionInitiator initiator;

/** action property. */
protected int action;

/** Action constant for no matching action. */
public static final int NO_ACTION                = 0;
/** Action constant for moving the editor left one cell. */
public static final int MOVE_LEFT_ACTION         = 1;
/** Action constant for moving the editor right one cell. */
public static final int MOVE_RIGHT_ACTION        = 2;
/** Action constant for moving the editor up one cell. */
public static final int MOVE_UP_ACTION           = 3;
/** Action constant for moving the editor down one cell. */
public static final int MOVE_DOWN_ACTION         = 4;
/** Action constant for moving the editor one page up. */
public static final int MOVE_PAGE_UP_ACTION      = 5;
/** Action constant for moving the editor one page down. */
public static final int MOVE_PAGE_DOWN_ACTION    = 6;
/** Action constant for scrolling the grid one page left. */
public static final int MOVE_PAGE_LEFT_ACTION    = 7;
/** Action constant for scrolling the grid one page right. */
public static final int MOVE_PAGE_RIGHT_ACTION   = 8;
/** Action constant for moving the editor to the first column on the current row. */
public static final int MOVE_HOME_ACTION         = 9;
/** Action constant for moving the editor to the last column on the current row. */
public static final int MOVE_END_ACTION          = 10;
/** Action constant for moving the editor to the first row and column of the grid. */
public static final int MOVE_GRID_HOME_ACTION    = 11;
/** Action constant for moving the editor to the last row and column of the grid. */
public static final int MOVE_GRID_END_ACTION     = 12;
/** Action constant for moving the editor to the next cell. */
public static final int MOVE_NEXT_ACTION         = 13;
/** Action constant for moving the editor to the previous cell. */
public static final int MOVE_PREVIOUS_ACTION     = 14;
/** Action constant for expanding the current row. */
public static final int OPEN_TABLE_ACTION        = 15;
/** Action constant for collapsing the current row. */
public static final int CLOSE_TABLE_ACTION       = 16;
/** Action constant for saving all edits in the current editor. */
public static final int STOP_EDIT_ACTION         = 17;
/** Action constant for cancelling all edits in the current editor. */
public static final int CANCEL_EDIT_ACTION       = 18;
/** Action constant for invoking the popup menu. */
public static final int POPUP_MENU_ACTION        = 19;
/** Action constant for column selection. */
public static final int COLUMN_SELECT_ACTION     = 20;
/** Action constant for row selection/deselection. */
public static final int SINGLE_ROW_SELECT_ACTION = 21;
/** Action constant for contiguous group row selection. */
public static final int GROUP_ROW_SELECT_ACTION  = 22;
/** Action constant to ignore keys. */
public static final int IGNORE_ACTION            = 23;
/** Action constant for copy rows. */
public static final int COPY_ACTION              = 24;
/** Action constant for paste range. */
public static final int PASTE_ACTION             = 25;

/**
 * A public constructor that creates a mapping from a series of
 * modifiers to a action.
 */
public HiGridAction(ActionInitiator initiator, int action) {
    setInitiator(initiator);
    setAction(action);
}

/**
 * Gets the <CODE>ActionInitiator</CODE> for this
 * <CODE>HiGridAction</CODE>.
 * @see #setInitiator
 */
public ActionInitiator getInitiator() {
    return initiator;
}

/**
 * Sets the <CODE>ActionInitiator</CODE> for this
 * <CODE>HiGridAction</CODE>.
 * @see KeyActionInitiator
 * @see MouseActionInitiator
 */
public void setInitiator(ActionInitiator initiator) {
    this.initiator = initiator;
}

/**
 * Gets the value of the <CODE>Action</CODE> property for this
 * <CODE>ActionInitiator</CODE> instance.
 * The <CODE>Action</CODE> property determines what action will be
 * performed when the specified AWT event occurs.
 */
public int getAction() {
    return action;
}

/**
 * Sets the value of the <code>Action</code> property for this <CODE>HiGridAction</CODE>
 * instance. The <CODE>Action</CODE> property determines what action will
 * be performed when the specified AWT event occurs.
 */
public void setAction(int action) {
    this.action = action;
}

/**
 * Returns true if the event has a matching action set.
 */
public boolean isMatch(HiGrid grid, AWTEvent event) {
    boolean match = initiator.isMatch(event);
    // return false if there is no match
    if (!match) {
        return false;
    }
    switch (action) {
    case COLUMN_SELECT_ACTION:
    case SINGLE_ROW_SELECT_ACTION:
    case GROUP_ROW_SELECT_ACTION:
        CellPosition cell = null;
        if (event instanceof MouseEvent) {
            MouseEvent e = (MouseEvent)event;
            cell = grid.getHorizontalScrollableCellPositionFromPoint(
											new Point(e.getX(), e.getY()));
        }
        else if (event instanceof KeyEvent) {
            cell = grid.getCurrentCellPosition();
        }
        if (cell == null) {
            return false;
        }
        RowNode rowNode = cell.getRowNode();
        if (rowNode == null) {
            return false;
        }
        else if (action == COLUMN_SELECT_ACTION) {
            return (rowNode.getRowFormat() instanceof HeaderFormat);
        }
        else if (action == SINGLE_ROW_SELECT_ACTION || action == GROUP_ROW_SELECT_ACTION) {
            return (rowNode.getRowFormat() instanceof RecordFormat);
        }
        else {
            return false;
        }
    default:
        return true;
    }
}

/**
 * Convert to a string.
 */
public String toString() {
    return "Initiator: " + initiator + " Action=" + action;
}

}
