/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: HiGridTip.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.io.Serializable;
import java.util.*;

import javax.swing.*;

/**
 * The HiGridTip class is a custom implementation of a JToolTip.
 * It pops up an exact image of a properly aligned Header when the
 * mouse cursor pauses for a moment over a HiGrid Record.
 */

public class HiGridTip extends JToolTip implements Serializable {

static final long serialVersionUID = -1935142391221415679L;

/*
    Implementation details:  The HiGridTip must extend JToolTip, or
    else it will not be recognized as a tooltip by the ToolTipManager.
    We want to display an Image, so we create a JLabel, draw an Image
    and associate the Image with the JLabel.  Then, the JLabel is placed
    within the HiGridTip (since it is a Container) and we are done.
*/

protected HiGrid grid;
protected JLabel label = null;
transient protected Image panelImage = null;
protected ImageIcon panelIcon = null;
protected Dimension preferredSize = null;
protected RowNode headerRowNode = null;

public HiGridTip(HiGrid grid) {
    this.grid = grid;
    preferredSize = new Dimension(1, 1);
    label = new JLabel();
    label.setOpaque(true);
    setLayout(new BorderLayout());
}

protected RowNode getHeaderRowNode(CellPosition cell) {
    if (headerRowNode == null) {
        headerRowNode = new RowNode();
    }
    headerRowNode.setFormatNode(cell.getRowNode().getFormatNode());
    headerRowNode.setRowFormat(cell.getRowNode().getFormatNode().getHeaderFormat());
    return headerRowNode;
}

protected boolean drawTip(Point point, CellPosition cell, Point pointToPosition) {
    // only Records have Headers, so nothing else is valid
    if (!(cell.getRowNode().getRowFormat() instanceof RecordFormat) ||
         cell.isIndicator()) {
        return false;
    }
    // check whether we really need to show a tip
    if (!(grid.getTopRowNode().equals(cell.getRowNode()))) {
        HeaderTipWalk walkable = new HeaderTipWalk(grid.getTopRowNode(), cell.getRowNode(),
                                                   grid.getHeaderTipDisplay());
        grid.getRowTree().walk(cell.getRowNode(), walkable);
        if (!walkable.shouldDisplay()) {
            return false;
        }
    }

    // construct a parent Header for this Record
    RowNode rowNode = getHeaderRowNode(cell);
    // compute the (x,y) positioning for the tooltip
    int x = 0;
    int y = 0;
    x -= grid.getHorizontalScrollbar().getValue();
    x += rowNode.getRowFormat().getIndicatorIndent(grid, rowNode.getFormatNode());
    Point pointAbove = grid.getPointFromCellPosition(cell);
    y = pointAbove.y - rowNode.getHeight();
    // compute the actual size of the tooltip
    // if it is too big, then the ToolTipManager will not place it properly
    HeaderFormat headerFormat = rowNode.getFormatNode().getHeaderFormat();

    int gridWidth = grid.getGridArea().getSize().width;
    int imageWidth = headerFormat.getDataWidth();
    if (x > 0) {
        imageWidth = Math.min(gridWidth - x, imageWidth);
    }
    else {
        imageWidth = Math.min(gridWidth, imageWidth + x);
    }

    int height = headerFormat.getHeight();
    // create an image of exactly the right size
    if (panelImage == null ||
        panelImage.getWidth(null) != imageWidth ||
        panelImage.getHeight(null) != height) {
        if (panelImage != null) {
            panelImage.flush();
        }
        panelImage = grid.createImage(imageWidth, height);
    }
    if (panelImage != null) {
        Graphics gc = panelImage.getGraphics();
        Rectangle rect = new Rectangle(0, 0, 0, 0);
        Rectangle rectIntersect = new Rectangle(0, 0, headerFormat.getDataWidth(), height);
        if (x < 0) {
            gc.translate(x, 0);
        }
        // use the grid's drawing method to exactly replicate the header
        headerFormat.drawCells(gc, grid, rowNode, rowNode.getRowFormat().getDataFormats(), rect, rectIntersect);
        gc.translate(-rect.x, 0);
        if (x < 0) {
            gc.translate(-x, 0);
        }
        gc.setClip(0, 0, rect.x, height);
        // draw a border around it, to make it stand out a bit better as a tooltip
        gc.setColor(getBackground());
        gc.drawRect(0, 0, panelImage.getWidth(null) - 1, panelImage.getHeight(null) - 1);
        gc.dispose();
        panelIcon = new ImageIcon(panelImage);
        label.setIcon(panelIcon);
    }
    pointToPosition.x = Math.max(x, 0);
    pointToPosition.y = y;
    return true;
}

/**
 * This method returns the point at which a tool tip should be drawn.
 * @param point An x,y pixel co-ordinate.  This method identifies the
 * grid cell that contains the specified point.  If the grid is scrolled
 * horizontally, this method automatically adds the horizontal scroll offset
 * to the Point parameter.
 * @return The Point at which a tool tip should be drawn.
 */
public Point getTip(Point point) {
    // find out which cell the mouse is pointing to
    CellPosition cell = grid.getHorizontalScrollableCellPositionFromPoint(point);
    if (cell == null ||
        cell.getRowNode() == null ||
        cell.getCellFormat() == null) {
        grid.getGridArea().setToolTipText(null);
        return null;
    }
    // clean up
    preferredSize.setSize(0, 0);
    removeAll();
    // attempt to draw the tip
    Point pointToPosition = new Point(0, 0);
    if (drawTip(point, cell, pointToPosition)) {
        // force the tooltip to appear
        String toolTipText = grid.getGridArea().getToolTipText();
        grid.getGridArea().setToolTipText((toolTipText == "1") ? "2" : "1");
        // pointing to a Record, so add the label to the Container
        label.setSize(panelImage.getWidth(null), panelImage.getHeight(null));
        preferredSize.setSize(label.getSize());
        add(label);
    }
    else {
        // force the tooltip to disappear instantly
        grid.getGridArea().setToolTipText(null);
    }
    return pointToPosition;
}

public Dimension getPreferredSize() {
    return preferredSize;
}

/*
 * Remove references for proper garbage collection.
 */
public void dispose() {
    if (panelImage != null) {
        panelImage.flush();
    }
    panelImage = null;
    panelIcon = null;
    headerRowNode = null;
    label = null;
    removeAll();
    setLayout(null);
    setComponent(null);
    grid = null;
}

}



