/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: LookupTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.io.*;
import java.util.Hashtable;
import java.lang.reflect.Field;

import javax.swing.*;
import javax.swing.border.*;

import com.klg.jclass.cell.*;
import com.klg.jclass.cell.editors.JCStringCellEditor;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.cell.EditorRendererRegistry;

/**
 * Manages editors and renderers, and uses the
 * <code>com.klg.jclass.cell.EditorRendererRegistry</code>.
 */
public class LookupTable implements Serializable {

static final long serialVersionUID = -1273774710302779797L;

//FIX - newly public methods need to be documented!!!

/*
Hashtable documentation:
There are 4 hashtables used; two for rendering and two for editing.
Also, two of the hashtables are used for setting defaults,
while the other two are used to keep track of the runtime cached instances
of running editors/renderers, to minimize object creation.

The defaultRenderers hashtable has a mapping from an integer index based
on the datatype to a String describing the absolute package name of the
renderer to use for this data type.  The defaultEditors hashtable has
a similar mapping for editors.

The runtimeRenderers hashtable has a mapping between a Class and an
actual running object instance.  Initially, the hashtable is populated
with Class objects for all of the different data types (as well as the
special renderers for Nodes, EditStatus, HeaderRenderer, etc), based on
the defaultRenderers table.  When an object is requested from the
runtimeRenderers table, if it is still a Class object, a new instance
is created, stored in the hashtable, and a reference returned.  If the
object is the hashtable is a running object instance, it is simply
returned.  The runtimeEditors hashtable has a similar mapping for editors.
 */

protected HiGrid grid;
protected Hashtable defaultRenderers = null;
protected Hashtable defaultEditors = null;
protected Hashtable runtimeRenderers = null;
protected Hashtable runtimeEditors = null;
protected HiGridEditStatusRenderer editStatusRenderer = null;
protected HiGridNodeRenderer nodeRenderer = null;
protected HiGridHeaderRenderer headerRenderer = null;


public LookupTable(HiGrid grid) {
    this.grid = grid;

    initDefaultRendererHashtable();
    initRuntimeRendererHashtable();

    initDefaultEditorHashtable();
    initRuntimeEditorHashtable();
}

/**
 * Retrieves the default HiGridNodeRenderer.
 *
 * @return The default HiGridNodeRenderer.
 */
public static Class getDefaultNodeRenderer() {
    try {
        return Class.forName("com.klg.jclass.higrid.DefaultNodeRenderer");
    }
    catch (Exception e) {
        e.printStackTrace();
    }
    return null;
}

/**
 * Retrieves the default HiGridEditStatusRenderer.
 *
 * @return The default HiGridEditStatusRenderer.
 */
public static Class getDefaultEditStatusRenderer() {
    try {
        return Class.forName("com.klg.jclass.higrid.DefaultEditStatusRenderer");
    }
    catch (Exception e) {
        e.printStackTrace();
    }
    return null;
}

/**
 * Retrieves the default HiGridHeaderRenderer.
 *
 * @return The default HiGridHeaderRenderer.
 */
public static Class getDefaultHeaderRenderer() {
    try {
        return Class.forName("com.klg.jclass.higrid.SortLabelCellRenderer");
    }
    catch (Exception e) {
        e.printStackTrace();
    }
    return null;
}

/**
 * Retrieves the current object that implements the HiGridNodeRenderer
 * interface.
 *
 * @return The current HiGridNodeRenderer.
 */
public HiGridNodeRenderer getNodeRenderer() {
    return nodeRenderer;
}

/**
 * Retrieves the current object that implements the HiGridEditStatusRenderer
 * interface.
 *
 * @return The current HiGridEditStatusRenderer.
 */
public HiGridEditStatusRenderer getEditStatusRenderer() {
    return editStatusRenderer;
}

/**
 * Retrieves the current object that implements the HiGridHeaderRenderer
 * interface.
 *
 * @return The current HiGridHeaderRenderer.
 */
public HiGridHeaderRenderer getHeaderRenderer() {
    return headerRenderer;
}

/**
 * Sets the current object that implements the HiGridNodeRenderer
 * interface.
 *
 * @param renderer The new HiGridNodeRenderer.
 */
public void setNodeRenderer(HiGridNodeRenderer renderer) {
    nodeRenderer = renderer;
    runtimeRenderers.put(renderer.getClass(), renderer);
}

/**
 * Sets the current object that implements the HiGridEditStatusRenderer
 * interface.
 *
 * @param renderer The new HiGridEditStatusRenderer.
 */
public void setEditStatusRenderer(HiGridEditStatusRenderer renderer) {
    editStatusRenderer = renderer;
    runtimeRenderers.put(renderer.getClass(), renderer);
}

/**
 * Sets the current object that implements the HiGridHeaderRenderer
 * interface.
 *
 * @param renderer The new HiGridHeaderRenderer.
 */
public void setHeaderRenderer(HiGridHeaderRenderer renderer) {
    headerRenderer = renderer;
    runtimeRenderers.put(renderer.getClass(), renderer);
}

/*
 * Initializes the defaultRenderers table based on all supported
 * datatypes.
 */
protected void initDefaultRendererHashtable() {
    defaultRenderers = new Hashtable();

    // set up the renderer translation table
    // all values in the translation table must implement JCCellRenderer
    setDefaultRenderer(MetaDataModel.TYPE_BOOLEAN);
    setDefaultRenderer(MetaDataModel.TYPE_SQL_DATE);
    setDefaultRenderer(MetaDataModel.TYPE_DOUBLE);
    setDefaultRenderer(MetaDataModel.TYPE_FLOAT);
    setDefaultRenderer(MetaDataModel.TYPE_INTEGER);
    setDefaultRenderer(MetaDataModel.TYPE_STRING);
    setDefaultRenderer(MetaDataModel.TYPE_BIG_DECIMAL);
    setDefaultRenderer(MetaDataModel.TYPE_LONG);
    setDefaultRenderer(MetaDataModel.TYPE_SQL_TIME);
    setDefaultRenderer(MetaDataModel.TYPE_SQL_TIMESTAMP);
    setDefaultRenderer(MetaDataModel.TYPE_OBJECT);
    setDefaultRenderer(MetaDataModel.TYPE_BYTE);
    setDefaultRenderer(MetaDataModel.TYPE_SHORT);
    setDefaultRenderer(MetaDataModel.TYPE_BYTE_ARRAY);
    setDefaultRenderer(MetaDataModel.TYPE_UTIL_DATE);
}

/**
 * Sets the default renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 */
protected void setDefaultRenderer(int type) {
    try {
        defaultRenderers.put(new Integer(type), Class.forName(getDefaultCellRendererName(type)));
    }
    catch (Exception e) {
        try {
            defaultRenderers.put(new Integer(type), Class.forName(getAlternateCellRendererName(type)));
        }
        catch (Exception e2) {
            e2.printStackTrace();
        }
    }
}

/*
 * Initializes the runtimeRenderers table with instances of objects
 * that we know we will need.
 */
protected void initRuntimeRendererHashtable() {
    runtimeRenderers = new Hashtable();

    // set up the renderer translation table
    // all values in the translation table must implement JCCellRenderer
    try {
        setNodeRenderer((HiGridNodeRenderer)getDefaultNodeRenderer().newInstance());
        setEditStatusRenderer((HiGridEditStatusRenderer)getDefaultEditStatusRenderer().newInstance());
        setHeaderRenderer((HiGridHeaderRenderer)getDefaultHeaderRenderer().newInstance());
    }
    catch (Exception e) {
        e.printStackTrace();
    }
}

/**
 * Retrieves the defaultRenderers hashtable.
 *
 * @return The defaultRenderers hashtable.
 */
public  Hashtable getDefaultRenderersTable() {
    return defaultRenderers;
}

/**
 * Retrieves the runtimeRenderers hashtable.
 *
 * @return The runtimeRenderers hashtable.
 */
public Hashtable getRuntimeRenderersTable() {
    return runtimeRenderers;
}

/**
 * Retrieves the default renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The renderer associated with the passed datatype.
 */
public Class getDefaultCellRenderer(int type) {
    return(Class)defaultRenderers.get(new Integer(type));
}

/**
 * Retrieves the running renderer instance for the given class.
 *
 * @param className A class associated with one of the MetaDataModel
 * TYPE_* datatypes.
 *
 * @return The object renderer instance of the passed class.
 */
protected JCCellRenderer getCellRendererInstance(Class className) {
    JCCellRenderer cellRenderer = null;
    try {
        if (className == null) {
            className = Class.forName("com.klg.jclass.cell.renderers.JCStringCellRenderer");
        }
        Object object = runtimeRenderers.get(className);
        if ((object == null) || (object instanceof Class)) {
            cellRenderer = (JCCellRenderer)className.newInstance();
            runtimeRenderers.put(className, cellRenderer);
        }
        else {
            cellRenderer = (JCCellRenderer)object;
        }
    }
    catch (Exception e) {
        e.printStackTrace();
    }

    return cellRenderer;
}

/*
 * Editor lookup
 */

/*
 * Initializes the defaultEditors table based on all supported
 * datatypes.
 */
protected void initDefaultEditorHashtable() {
    defaultEditors = new Hashtable();

    // set up the editor translation table
    // all values in the translation table must implement JCCellEditor
    setDefaultEditor(MetaDataModel.TYPE_BOOLEAN);
    setDefaultEditor(MetaDataModel.TYPE_SQL_DATE);
    setDefaultEditor(MetaDataModel.TYPE_DOUBLE);
    setDefaultEditor(MetaDataModel.TYPE_FLOAT);
    setDefaultEditor(MetaDataModel.TYPE_INTEGER);
    setDefaultEditor(MetaDataModel.TYPE_STRING);
    setDefaultEditor(MetaDataModel.TYPE_BIG_DECIMAL);
    setDefaultEditor(MetaDataModel.TYPE_LONG);
    setDefaultEditor(MetaDataModel.TYPE_SQL_TIME);
    setDefaultEditor(MetaDataModel.TYPE_SQL_TIMESTAMP);
    setDefaultEditor(MetaDataModel.TYPE_OBJECT);
    setDefaultEditor(MetaDataModel.TYPE_BYTE);
    setDefaultEditor(MetaDataModel.TYPE_SHORT);
    setDefaultEditor(MetaDataModel.TYPE_BYTE_ARRAY);
    setDefaultEditor(MetaDataModel.TYPE_UTIL_DATE);
}

/**
 * Sets the default editor for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 */
protected void setDefaultEditor(int type) {
    try {
        defaultEditors.put(new Integer(type), Class.forName(getDefaultCellEditorName(type)));
    }
    catch (Exception e) {
        try {
            defaultEditors.put(new Integer(type), Class.forName(getAlternateCellEditorName(type)));
        }
        catch (Exception e2) {
            e2.printStackTrace();
        }
    }
}

/*
 * Initializes the runtimeEditors table with instances of objects
 * that we know we will need.
 */
protected void initRuntimeEditorHashtable() {
    runtimeEditors = new Hashtable();

    // set up the editor translation table
    // all values in the translation table must implement JCCellEditor
    try {
        runtimeEditors.put(Class.forName("com.klg.jclass.cell.editors.JCStringCellEditor"),
                           new JCStringCellEditor());
    }
    catch (Exception e) {
        e.printStackTrace();
    }
}

/**
 * Retrieves the defaultEditors hashtable.
 *
 * @return The defaultEditors hashtable.
 */
public Hashtable getDefaultEditorsTable() {
    return defaultEditors;
}

/**
 * Retrieves the runtimeEditors hashtable.
 *
 * @return The runtimeEditors hashtable.
 */
public Hashtable getRuntimeEditorsTable() {
    return runtimeEditors;
}

/**
 * Retrieves the default editor for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The editor associated with the passed datatype.
 */
public Class getDefaultCellEditor(int type) {
    return(Class)defaultEditors.get(new Integer(type));
}

/**
 * Retrieves the running editor instance for the given class.
 *
 * @param className A class associated with one of the MetaDataModel
 * TYPE_* datatypes.
 *
 * @return The object editor instance of the passed class.
 */
protected JCCellEditor getCellEditorInstance(Class className) {
    JCCellEditor cellEditor = null;
    try {
        if (className == null) {
            className = Class.forName("com.klg.jclass.cell.editors.JCStringCellEditor");
        }
        Object object = runtimeEditors.get(className);
        if ((object == null) || (object instanceof Class)) {
            cellEditor = (JCCellEditor)className.newInstance();
            runtimeEditors.put(className, cellEditor);
        }
        else {
            cellEditor = (JCCellEditor)object;
        }
    }
    catch (Exception e) {
        e.printStackTrace();
    }

    // make sure that we are using the current HiGrid border for the cell editor
    if (cellEditor instanceof JComponent) {
        JComponent e = (JComponent)cellEditor;
        if (e.getBorder() != grid.getEditorBorder()) {
            e.setBorder(grid.getEditorBorder());
        }
    }
    return cellEditor;
}

/**
 * Retrieves the String name of the renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The String name of the renderer associated with the passed datatype.
 */
public static String getCellRendererName(int type) {
    String s = getDefaultCellRendererName(type);
    if (s == null) {
        s = getAlternateCellRendererName(type);
    }
    return s;
}

/**
 * Retrieves the Alternate String name of the renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The Alternate String name of the renderer associated with the passed datatype.
 */
public static String getAlternateCellRendererName(int type) {
    String s = null;
//FIX
//        s = new String("com.klg.jclass.cell.renderers.JCLabelCellRenderer");
//        if (true) return s;
    switch (type) {
    case MetaDataModel.TYPE_SQL_DATE:
    case MetaDataModel.TYPE_SQL_TIME:
    case MetaDataModel.TYPE_SQL_TIMESTAMP:
    case MetaDataModel.TYPE_UTIL_DATE:
        s = new String("com.klg.jclass.cell.renderers.JCStringCellRenderer");
        break;
    case MetaDataModel.TYPE_BYTE_ARRAY:
        s = new String("com.klg.jclass.cell.renderers.JCRawImageCellRenderer");
        break;
    case MetaDataModel.TYPE_DOUBLE:
    case MetaDataModel.TYPE_FLOAT:
        s = new String("com.klg.jclass.cell.renderers.JCStringCellRenderer");
        break;
    case MetaDataModel.TYPE_INTEGER:
    case MetaDataModel.TYPE_SHORT:
    case MetaDataModel.TYPE_BYTE:
        s = new String("com.klg.jclass.cell.renderers.JCStringCellRenderer");
        break;
    case MetaDataModel.TYPE_BOOLEAN:
        s = new String("com.klg.jclass.cell.renderers.JCCheckBoxCellRenderer");
        break;
    case MetaDataModel.TYPE_STRING:
    case MetaDataModel.TYPE_BIG_DECIMAL:
    case MetaDataModel.TYPE_LONG:
    case MetaDataModel.TYPE_OBJECT:
        s = new String("com.klg.jclass.cell.renderers.JCStringCellRenderer");
        break;
    default:
        s = new String("com.klg.jclass.cell.renderers.JCStringCellRenderer");
        break;
    }
    return s;
}

/**
 * Retrieves the Default String name of the renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The Default String name of the renderer associated with the passed datatype.
 */
public static String getDefaultCellRendererName(int type) {
    String s = null;

    try {
        switch (type) {
            case MetaDataModel.TYPE_BOOLEAN:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Boolean"),null);
                break;
            case MetaDataModel.TYPE_SQL_DATE:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.sql.Date"),null);
                break;
            case MetaDataModel.TYPE_DOUBLE:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Double"),null);
                break;
            case MetaDataModel.TYPE_FLOAT:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Float"),null);
                break;
            case MetaDataModel.TYPE_INTEGER:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Integer"),null);
                break;
            case MetaDataModel.TYPE_STRING:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.String"),null);
                break;
            case MetaDataModel.TYPE_BIG_DECIMAL:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.math.BigDecimal"),null);
                break;
            case MetaDataModel.TYPE_LONG:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Long"),null);
                break;
            case MetaDataModel.TYPE_SQL_TIME:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.sql.Time"),null);
                break;
            case MetaDataModel.TYPE_SQL_TIMESTAMP:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.sql.Timestamp"),null);
                break;
            case MetaDataModel.TYPE_OBJECT:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Object"),null);
                break;
            case MetaDataModel.TYPE_BYTE:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Byte"),null);
                break;
            case MetaDataModel.TYPE_SHORT:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.Short"),null);
                break;
            case MetaDataModel.TYPE_BYTE_ARRAY:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(byte[].class,null);
                break;
            case MetaDataModel.TYPE_UTIL_DATE:
                s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.util.Date"),null);
                break;
        }
    }
    catch (ClassNotFoundException e) {
        s = null;
    }
    return s;
}

/**
 * Retrieves the Remaining String name of the renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The Remaining String name of the renderer associated with the passed datatype.
 */
public static String getRemainingCellRendererName(int index) {
    String s = null;
    switch (index) {
    case 0:
        // this renderer does not exist ???
        s = new String("com.klg.jclass.cell.renderers.JCEllipsisCellRenderer");
        break;
    case 1:
        try {
            s = EditorRendererRegistry.getCentralRegistry().getCellRendererString(Class.forName("java.lang.String"),"WordWrap");
        }
        catch (ClassNotFoundException e){
            s = null;
        }
        if(s == null) {
            s = new String("com.klg.jclass.cell.renderers.JCWordWrapCellRenderer");
        }
        break;
    default:
        break;
    }
    return s;
}

/**
 * Retrieves the String name of the renderer for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The String name of the renderer associated with the passed datatype.
 */
public static String getCellEditorName(int type) {
    String s = getDefaultCellEditorName(type);
    if (s == null) {
        s = getAlternateCellEditorName(type);
    }
    return s;
}

/**
 * Retrieves the Alternate String name of the editor for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The Alternate String name of the editor associated with the passed datatype.
 */
public static String getAlternateCellEditorName(int type) {
    String s = null;
    switch (type) {
    case MetaDataModel.TYPE_BOOLEAN:
        s = new String("com.klg.jclass.cell.editors.JCCheckBoxCellEditor");
        break;
    case MetaDataModel.TYPE_DOUBLE:
        s = new String("com.klg.jclass.cell.editors.JCDoubleCellEditor");
        break;
    case MetaDataModel.TYPE_FLOAT:
        s = new String("com.klg.jclass.cell.editors.JCFloatCellEditor");
        break;
    case MetaDataModel.TYPE_INTEGER:
        s = new String("com.klg.jclass.cell.editors.JCIntegerCellEditor");
        break;
    case MetaDataModel.TYPE_STRING:
        s = new String("com.klg.jclass.cell.editors.JCStringCellEditor");
        break;
    case MetaDataModel.TYPE_BIG_DECIMAL:
        s = new String("com.klg.jclass.cell.editors.JCBigDecimalCellEditor");
        break;
    case MetaDataModel.TYPE_LONG:
        s = new String("com.klg.jclass.cell.editors.JCLongCellEditor");
        break;
    case MetaDataModel.TYPE_OBJECT:
        s = new String("com.klg.jclass.cell.editors.JCStringCellEditor");
        break;
    case MetaDataModel.TYPE_BYTE:
        s = new String("com.klg.jclass.cell.editors.JCByteCellEditor");
        break;
    case MetaDataModel.TYPE_SHORT:
        s = new String("com.klg.jclass.cell.editors.JCShortCellEditor");
        break;
    case MetaDataModel.TYPE_BYTE_ARRAY:
        s = new String("com.klg.jclass.cell.editors.JCImageCellEditor");
        break;
    case MetaDataModel.TYPE_UTIL_DATE:
        s = new String("com.klg.jclass.cell.editors.JCDateCellEditor");
        break;
    case MetaDataModel.TYPE_SQL_DATE:
        s = new String("com.klg.jclass.cell.editors.JCSqlDateCellEditor");
        break;
    case MetaDataModel.TYPE_SQL_TIME:
        s = new String("com.klg.jclass.cell.editors.JCSqlTimeCellEditor");
        break;
    case MetaDataModel.TYPE_SQL_TIMESTAMP:
        s = new String("com.klg.jclass.cell.editors.JCSqlTimestampCellEditor");
        break;
    default:
        s = new String("com.klg.jclass.cell.editors.JCStringCellEditor");
        break;
    }
    return s;
}

/**
 * Retrieves the Default String name of the editor for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The Default String name of the editor associated with the passed datatype.
 */
public static String getDefaultCellEditorName(int type) {
    String s = null;

    try {
        switch (type) {
            case MetaDataModel.TYPE_BOOLEAN:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Boolean"),null);
                break;
            case MetaDataModel.TYPE_SQL_DATE:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.sql.Date"),null);
                break;
            case MetaDataModel.TYPE_DOUBLE:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Double"),null);
                break;
            case MetaDataModel.TYPE_FLOAT:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Float"),null);
                break;
            case MetaDataModel.TYPE_INTEGER:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Integer"),null);
                break;
            case MetaDataModel.TYPE_STRING:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.String"),null);
                break;
            case MetaDataModel.TYPE_BIG_DECIMAL:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.math.BigDecimal"),null);
                break;
            case MetaDataModel.TYPE_LONG:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Long"),null);
                break;
            case MetaDataModel.TYPE_SQL_TIME:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.sql.Time"),null);
                break;
            case MetaDataModel.TYPE_SQL_TIMESTAMP:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.sql.Timestamp"),null);
                break;
            case MetaDataModel.TYPE_OBJECT:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Object"),null);
                break;
            case MetaDataModel.TYPE_BYTE:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Byte"),null);
                break;
            case MetaDataModel.TYPE_SHORT:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.Short"),null);
                break;
            case MetaDataModel.TYPE_BYTE_ARRAY:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(byte[].class,null);
                break;
            case MetaDataModel.TYPE_UTIL_DATE:
                s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.util.Date"),null);
                break;
        }
    }
    catch (ClassNotFoundException e) {
        s = null;
    }
    return s;
}

/**
 * Retrieves the Remaining String name of the editor for the given datatype.
 *
 * @param type One of the MetaDataModel TYPE_* datatypes.
 *
 * @return The Remaining String name of the editor associated with the passed datatype.
 */
public static String getRemainingCellEditorName(int index) {
    String s = null;
    switch (index) {
    case 0:
        try {
            s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.String"),"MultiLine");
        }
        catch (ClassNotFoundException e) {
            s = null;
        }
        if(s == null) {
            s = new String("com.klg.jclass.cell.editors.JCMultilineCellEditor");
        }
        break;
    case 1:
        try {
            s = EditorRendererRegistry.getCentralRegistry().getCellEditorString(Class.forName("java.lang.String"),"WordWrap");
        }
        catch (ClassNotFoundException e) {
            s = null;
        }
        if(s == null) {
            s = new String("com.klg.jclass.cell.editors.JCWordWrapCellEditor");
        }
        break;
    default:
        break;
    }
    return s;
}

}
