/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PrintGrid12.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.awt.print.*;
import java.io.Serializable;
import java.util.Vector;

import javax.swing.tree.*;

import com.klg.jclass.datasource.DataModelException;

/**
 * Controls printing of pages with header and footer.
 * Margin insets are for the grid area only.  Header
 * and footer are part of the margin area.
 */

public class PrintGrid12 extends PrintGrid implements Printable, Pageable, Serializable {

static final long serialVersionUID = 671101792413160098L;

public static final int DEFAULT_PAGE_MARGIN_IN_72TH_INCHES = 72;
protected static final int MARGIN_ADJUSTMENT_IN_72TH_INCHES = 15;

transient protected HiGrid grid;
transient protected PageFormat currentPageFormat;

transient protected Insets defaultMargin = new Insets(
                                                     DEFAULT_PAGE_MARGIN_IN_72TH_INCHES, 
                                                     DEFAULT_PAGE_MARGIN_IN_72TH_INCHES, 
                                                     DEFAULT_PAGE_MARGIN_IN_72TH_INCHES, 
                                                     DEFAULT_PAGE_MARGIN_IN_72TH_INCHES);
transient protected Insets margin;

transient protected int totalNumPages;
transient protected int currentPage;
transient protected int rowPageIndex;
transient protected int columnPageIndex;
transient protected int printFormat = HiGrid.PRINT_AS_DISPLAYED;

transient protected RowTree printRowTree;
transient protected Vector rowNodeList;
transient protected int rowPages = 0;
transient protected int columnPages = 0;

/**
 * Constructor.
 * grid the parent HiGrid instance
 */
public PrintGrid12(HiGrid grid) {
    this.grid = grid;
    setMargin(getDefaultMargin());
    printRowTree = null;
}

/** 
 * Computes the total number of pages for this job.
 * @param pageFormat describes the page orientation for all pages to be printed.
 */
public void initialize(PageFormat pageFormat) {
    setPageFormat(pageFormat);

    // get the print time row tree
    printRowTree = getPrintTimeRowTree();
    if (printRowTree == null) {
        return;
    }

    // do the prelimary print walk to determine some printing parameters
    PrePrintWalk prePrintWalk = new PrePrintWalk(this, grid, pageFormat);
    printRowTree.walk((RowNode)(printRowTree.getRootNode().getFirstChild()), prePrintWalk);
    totalNumPages = prePrintWalk.getPrintPages();
    rowPages = prePrintWalk.getRowPages();
    columnPages = prePrintWalk.getColumnPages();
    rowNodeList = prePrintWalk.getRowNodeList();
}

/**
 * Cleanup the print time row tree.
 */
protected void disposePrintTimeRowTree() {
    printRowTree = null;
    rowNodeList = null;
}

/**
 * Setup the print time row tree.
 */
protected RowTree getPrintTimeRowTree() {
    RowTree runtimeRowTree = grid.getRowTree();
    printFormat = grid.getPrintFormat();

    // for print as displayed format:
    if (printFormat == HiGrid.PRINT_AS_DISPLAYED) {
        // use the runtime display row tree
        return runtimeRowTree;
    }

    // for print as expanded format:
    if (grid.getDataModel() == null || grid.getFormatTree() == null) {
        return null;
    }
    // get the list of meta id's of the expanded row nodes
    PrePrintRowTreeWalk walk = new PrePrintRowTreeWalk();
    runtimeRowTree.walk((RowNode)(runtimeRowTree.getRootNode().getFirstChild()), walk);
    int metaIDArray[] = walk.getMetaIDArray();
    if (metaIDArray == null) {
        return runtimeRowTree;
    }

    // create the print time row tree
    RowTree rowTree = new RowTree(grid, false);
    rowTree.createPrintTimeTree(grid.getFormatTree(), metaIDArray);
    return rowTree;
}

/**
 * Gets the width of a page.
 *
 * @return The page width.
 */
public int getPageWidth() {
    return(int)getPageFormat().getWidth();
}

/**
 * Gets the height of a page.
 *
 * @return The page height.
 */
public int getPageHeight() {
    return(int)getPageFormat().getHeight();
}

/**
 * Gets the margin insets.
 *
 * @return The margin insets.
 */
public Insets getMargin() {
    return margin;
}

/**
 * Sets the page margins. Passing null sets the margins to the default 1".
 * @param margin the new margin.
 */
public void setMargin(Insets margin) {
    if (margin == null) {
        margin = getDefaultMargin();
    }
    else {
        this.margin = margin;
    }
}

/**
 * Gets the default margin.
 *
 * @return The default margin insets.
 */
public Insets getDefaultMargin() {
    return defaultMargin;
}

/**
 * Gets the current pageFormat.
 *
 * @return The pageFormat.
 */
public PageFormat getPageFormat() {
    return currentPageFormat;
}

/**
 * Sets the current pageFormat.
 * @param format the new pageFormat.
 */
public void setPageFormat(PageFormat format) {
    this.currentPageFormat = format;
}

/**
 * Gets the printable height of a page.
 *
 * @return The printable height.
 */
public int getPrintableHeight() {
    return(int)getPageFormat().getImageableHeight();
}

/**
 * Gets the printable width of a page.
 *
 * @return The printable width.
 */
public int getPrintableWidth() {
    return(int)getPageFormat().getImageableWidth();
}

/**
 * Gets the top/left corner of the printable area on the page.
 *
 * @return the top/left corner of the printable area on the page.
 */
protected Point getPrintOffset(PageFormat pageFormat) {
    return new Point(margin.left, margin.top);
    // Note: this should really return the results of the getImageable[XY] methods
    //       from the PageFormat class, except there isn't any way to SET them
    //       programmatically.  Only the user can set them via the pageDialog()
    //       method.  Thus, we override and use internal margin calculations.
//    return new Point((int)pageFormat.getImageableX(), (int)pageFormat.getImageableY());
}


/**
 * Gets the row page index value.
 *
 * @return the row page index value.
 */
public int getRowPageIndex() {
    return rowPageIndex;
}

/**
 * Gets the column page index value.
 *
 * @return the column page index value.
 */
public int getColumnPageIndex() {
    return columnPageIndex;
}

/**
 * Gets the current page index.
 *
 * @return the current page index.
 */
public int getCurrentPageIndex() {
    return currentPage;
}

/**
 * Gets the total number of pages to print.
 *
 * @return the total number of pages to print.
 */
public int getTotalNumPages() {
    return totalNumPages;
}

///////////////////////////////////////////////////////////////////////
// Printable interface implementation

/**
 * Print/Paint the indicated page.
 *
 * @param gc the context into which the page is drawn.
 * @param pageFormat - the size and orientation of the page being drawn
 * @param pageIndex - the zero based index of the page to be drawn
 */
public int print(Graphics gc, PageFormat pageFormat, int pageIndex) {
    if (printRowTree == null || gc == null || 
        rowNodeList == null || rowNodeList.isEmpty() ||
        pageIndex >=  getTotalNumPages() || pageIndex < 0) {
        return Printable.NO_SUCH_PAGE;
    }

    setPageFormat(pageFormat);    
    currentPage = Math.max(0, Math.min(pageIndex, getTotalNumPages()-1));
    rowPageIndex = pageIndex / columnPages;
    columnPageIndex = pageIndex - (rowPageIndex * columnPages);

    // print the page
    PrintWalk printWalk = new PrintWalk(this, grid, pageFormat);
    RowNode topRowNode = (RowNode)(rowNodeList.elementAt(rowPageIndex));
    Point offset = getPrintOffset(pageFormat);
    gc.setClip(0, 0, (int)pageFormat.getWidth(), (int)pageFormat.getHeight());
    doPrintPageHeader(gc, pageFormat);
    Graphics gcGrid = gc.create();
    gcGrid.clipRect(offset.x, offset.y, (int)pageFormat.getImageableWidth(), 
                    (int)pageFormat.getImageableHeight());
    gcGrid.translate(offset.x, offset.y);
    printWalk.init(gcGrid, currentPage);
    printRowTree.walk(topRowNode, printWalk);
    gcGrid.dispose();
    if (rowPageIndex == (rowPages - 1)) {
        // for the set of column pages
        doPrintEnd(gc, pageFormat, offset.y + printWalk.getHeight());                        
    }
    doPrintPageFooter(gc, pageFormat);

    return Printable.PAGE_EXISTS;
}

///////////////////////////////////////////////////////////////////////
// Pageable interface implementation

/**
 * Gets the number of pages in the set.
 * @return the number of pages in this <code>Pageable</code>.
 */
public int getNumberOfPages() {
    return getTotalNumPages();
}

/**
 * Always returns the <code>PageFormat</code> of the page specified by
 * <code>pageIndex</code>.
 * @param pageIndex ignored
 * @return the <code>PageFormat</code> describing the size and
 *        orientation.
 * @exception java.lang.IndexOutOfBoundsException
 *          the <code>Pageable</code> does not contain the requested
 *        page.
 */
public PageFormat getPageFormat(int pageIndex) throws IndexOutOfBoundsException {
    return getPageFormat();
}

/**
 * Gets the <code>Printable</code> instance responsible for
 * rendering the page specified by <code>pageIndex</code>.
 * @param pageIndex ignored
 * @return the <code>Printable</code> that renders the page.
 * @exception java.lang.IndexOutOfBoundsException
 *            the <code>Pageable</code> does not contain the requested
 *          page.
 */
public Printable getPrintable(int pageIndex) throws IndexOutOfBoundsException {
    return this;
}

/**
 * Notifies print listeners that the page header is being printed.
 * @param graphicContext the context into which the page is drawn.
 * @param pageFormat - the size and orientation of the page being drawn
 */
protected void doPrintPageHeader(Graphics graphicContext, PageFormat pageFormat) {
    Graphics gc = graphicContext.create();
    gc.setColor(Color.black);
    Insets margin = getMargin();
    gc.setClip(margin.left, 0, getPrintableWidth(), margin.top);
    gc.translate(margin.left, 0);
    grid.fireHiGridPrintEvent(new HiGridPrintEvent(grid, 
                                                   HiGridPrintEvent.PRINT_HEADER, 
                                                   this, 
                                                   gc,
                                                   pageFormat,
                                                   getCurrentPageIndex(), 
                                                   getTotalNumPages()));        
    gc.dispose();
}

/**
 * Notifies print listeners that the page footer is being printed.
 * @param graphicContext the context into which the page is drawn.
 * @param pageFormat - the size and orientation of the page being drawn
 */
protected void doPrintPageFooter(Graphics graphicContext, PageFormat pageFormat) {
    Graphics gc = graphicContext.create();
    gc.setColor(Color.black);
    Insets margin = getMargin();
    gc.setClip(margin.left, getPageHeight() - margin.bottom,
               getPrintableWidth(), margin.bottom);
    gc.translate(margin.left, getPageHeight() - margin.bottom);
    grid.fireHiGridPrintEvent(new HiGridPrintEvent(grid, 
                                                   HiGridPrintEvent.PRINT_FOOTER, 
                                                   this, 
                                                   gc,
                                                   pageFormat,
                                                   getCurrentPageIndex(), 
                                                   getTotalNumPages()));
    gc.dispose();
}

/**
 * Notifies print listeners that the end of the last page is being printed.
 * @param graphicContext the context into which the page is drawn.
 * @param pageFormat - the size and orientation of the page being drawn
 */
protected void doPrintEnd(Graphics graphicContext, PageFormat pageFormat, int top) {
    Graphics gc = graphicContext.create();
    gc.setColor(Color.black);
    Insets margin = getMargin();
    gc.clipRect(margin.left, top,
                getPrintableWidth(), getPageHeight() - top - margin.bottom);
    gc.translate(margin.left, top);
    grid.fireHiGridPrintEvent(new HiGridPrintEvent(grid, 
                                                   HiGridPrintEvent.PRINT_END, 
                                                   this, 
                                                   gc, 
                                                   pageFormat,
                                                   getCurrentPageIndex(), 
                                                   getTotalNumPages()));
    gc.dispose();
}

/**
 * Prints default footer.
 * @param event the print event sent when rendering of the footer is needed.
 */
public void printPageFooter(HiGridPrintEvent event) {
    Graphics gc = event.getGraphics();    
    Rectangle r = gc.getClipBounds();
    Font font = grid.getFont();
    if (font != null) {
        gc.setFont(font);
        gc.setColor(Color.black);
        String page = LocaleBundle.string(LocaleBundle.printTextPage) + " " + (event.getPageIndex()+1) + " " +
                      LocaleBundle.string(LocaleBundle.printTextOf) + " " + event.getNumPages();
        gc.drawString(page, 0, r.height/2);
    }
}

}
