/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: RowFormat.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;
import java.io.Serializable;
import java.util.*;

import javax.swing.CellRendererPane;

import com.klg.jclass.cell.*;
import com.klg.jclass.datasource.*;
import com.klg.jclass.util.JCEnvironment;

/**
 * The <code>RowFormat</code> class contains the design-time and run-time
 * information associated with a given {@link RowNode}.
 */

public abstract class RowFormat implements Serializable {

static final long serialVersionUID = 7179966897087512288L;

/** use default height indication value */
private static final int USE_DEFAULT_HEIGHT_VALUE = -1;

/** indicates not all the cell formats' border styles are the same */
public static final int MIXED_BORDER_STYLES = -1;

// mouse y has to be at most this many pixels away for row sizing
protected static final int ROW_SIZING_TOLERANCE = 8;

// mouse x has to be at most this many pixels away on either side for column sizing
protected static final int COLUMN_SIZING_TOLERANCE = 4;

/** indicates not all the font styles are the same */
public static final int MIXED_STYLE = (-1);

/** indicates not all the font sizes are the same */
public static final int MIXED_SIZE = (-1);

// Properties
private int height = USE_DEFAULT_HEIGHT_VALUE;
private int width = 0;
private boolean allowHeightSizing = true;
private FormatNode parent = null;
private Vector indicatorFormats = new Vector();
private Vector dataFormats = new Vector();

/** Meta data is set every time RowFormat is used. */
private transient MetaDataModel metaData = null;

// working variable
private int printRectOffset = 0;

// preferred height optimizations
private int cachedPreferredHeight = USE_DEFAULT_HEIGHT_VALUE;

// only one CellRendererPane instance ever needed for JCComponentCellRenderer
private static CellRendererPane rendererPane = new CellRendererPane();

/**
 * Constructor
 */
public RowFormat() {
    // internal check
    if (HiGridBorder.NUM_BORDER_STYLES != com.klg.jclass.cell.Utilities.NUM_BORDER_STYLES) {
        System.out.println("HiGrid.RowFormat: Border styles not in sync.");
    }
}

/**
 * Retrieves the parent of this object.
 *
 * @return The current parent.
 */
protected FormatNode getParent() {
    return parent;
}

/**
 * Sets the parent of this object.
 *
 * @param parent The new parent.
 */
protected void setParent(FormatNode parent) {
    this.parent = parent;
}

/**
 * Retrieves the height of this object.
 *
 * @return The current height.
 */
public int getHeight() {
    if (height == USE_DEFAULT_HEIGHT_VALUE) {
        return getPreferredHeight();
    }
    return height;
}

/**
 * Sets the height of this object.
 *
 * @param height The new height.
 */
public void setHeight(int height) {
    this.height = height;
    // tell the cell formats about this change
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.rowHeightChanged();
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.rowHeightChanged();
    }
}

/**
 * Retrieves the width of this object.
 *
 * @return The current width.
 */
public int getWidth() {
    return width;
}

/**
 * Sets the width of this object.
 *
 * @param width The new width.
 */
protected void setWidth(int width) {
    this.width = width;
}

/**
 * Retrieves the metadata for this row format.
 *
 * @return The current metadata.
 */
public MetaDataModel getMetaData() {
    return metaData;
}

/**
 * Sets the metadata for this row format.
 *
 * @param metaData The new metadata.
 */
public void setMetaData(MetaDataModel metaData) {
    this.metaData = metaData;
}

/**
 * Retrieves whether this object is showing.
 * Should be overridden by its subclass.
 *
 * @return Whether object is currently showing.
 */
public boolean isShowing() {
    return true;
}

/**
 * Sets whether this object is showing.
 * Should be overridden by its subclass.
 *
 * @param showing Whether object is to be shown.
 */
public void setShowing(boolean showing) {
}

/**
 * Returns the first available editable position
 */
protected int getFirstVisibleCell() {
    if (this instanceof RecordFormat && !dataFormats.isEmpty()) {
        int index = HiGrid.FIRST_EDITABLE_COLUMN;
        int limit = dataFormats.size() - 1;
        while (index <= limit) {
            if (((CellFormat)dataFormats.elementAt(index)).isShowing()) {
                return index;
            }
            index++;
        }
    }
    return -1;
}

/**
 * Returns the last available editable position
 */
protected int getLastVisibleCell() {
    if (this instanceof RecordFormat && !dataFormats.isEmpty()) {
        int index = dataFormats.size() - 1;
        while (index >= HiGrid.FIRST_EDITABLE_COLUMN) {
            if (((CellFormat)dataFormats.elementAt(index)).isShowing()) {
                return index;
            }
            index--;
        }
    }
    return -1;
}

/**
 * Retrieves whether height sizing is allowed for this object.
 *
 * @return Whether height sizing is currently allowed.
 */
public boolean getAllowHeightSizing() {
    return allowHeightSizing;
}

/**
 * Sets whether height sizing is allowed for this object.
 *
 * @param allow Whether to allow height sizing.
 */
public void setAllowHeightSizing(boolean allow) {
    allowHeightSizing = allow;
}

/**
 * Sets the CellStyle style for this object (all its CellFormats).
 *
 * @param style The new CellStyle.
 */
public void setCellStyle(CellStyleModel style) {
    setIndicatorCellStyle(style);
    setDataCellStyle(style);
}

/**
 * Sets the CellStyle style for this object's IndicatorFormats).
 *
 * @param style The new CellStyle.
 */
public void setIndicatorCellStyle(CellStyleModel style) {
    setCellStyle(style, getIndicatorFormats());
}

/**
 * Sets the CellStyle style for this object's DataFormats).
 *
 * @param style The new CellStyle.
 */
public void setDataCellStyle(CellStyleModel style) {
    setCellStyle(style, getDataFormats());
}

/**
 * Sets the CellStyle style for this object's CellFormats).
 *
 * @param style The new CellStyle.
 */
protected void setCellStyle(CellStyleModel style, Vector v) {
    for (Enumeration e = v.elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.copyFromCellStyle(style);
    }
}

/**
 * Retrieves the border style for this object. All cell formats
 * must have the same style else the current border style returned
 * is MIXED_BORDER_STYLES.
 *
 * @return The current border style.
 */
public int getBorderStyle() {
    int style = MIXED_BORDER_STYLES;
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (style == MIXED_BORDER_STYLES) {
            style = cellFormat.getBorderStyle();
        }
        else if (style != cellFormat.getBorderStyle()) {
            return MIXED_BORDER_STYLES;
        }
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (style == MIXED_BORDER_STYLES) {
            style = cellFormat.getBorderStyle();
        }
        else if (style != cellFormat.getBorderStyle()) {
            return MIXED_BORDER_STYLES;
        }
    }
    return style;
}

/**
 * Sets the border style for this object (all its cell formats).
 *
 * @param style The new border style.
 */
public void setBorderStyle(int style) {
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.setBorderStyle(style);
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.setBorderStyle(style);
    }
}

/**
 * Retrieves the border insets for this object.
 *
 * @return The current border insets.
 */
public Insets getBorderInsets() {
    Insets insets = null;
    Insets defaultInsets = new Insets(2, 2, 2, 2);
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (insets == null) {
            insets = cellFormat.getBorderInsets();
        }
        else if (! insets.equals((Object) cellFormat.getBorderInsets())) {
            return defaultInsets;
        }
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (insets == null) {
            insets = cellFormat.getBorderInsets();
        }
        else if (! insets.equals((Object) cellFormat.getBorderInsets())) {
            return defaultInsets;
        }
    }
    if (insets == null) {
        insets = defaultInsets;
    }
    return insets;
}

/**
 * Sets the border insets for this object (all its cell formats).
 *
 * @param size The new border insets.
 */
public void setBorderInsets(Insets insets) {
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.setBorderInsets(insets);
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        cellFormat.setBorderInsets(insets);
    }
}

/**
 * Retrieves the foreground color for this object.  All data formats
 * must have the same color, else a null value is returned.
 *
 * @return The current foreground color.
 */
public Color getForeground() {
    Color color = null;
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (color == null) {
            color = cellFormat.getForeground();
        }
        else if (! color.equals((Object) cellFormat.getForeground())) {
            return null;
        }
    }
    return color;
}

/**
 * Sets the foreground color for this object (all its data formats).
 *
 * @param color The new foreground color.
 */
public void setForeground(Color color) {
    if (color != null) {
        for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
            CellFormat cellFormat = (CellFormat) e.nextElement();
            cellFormat.setForeground(color);
        }
    }
}

/**
 * Retrieves the background color for this object.  All data formats
 * must have the same color, else a null value is returned.
 *
 * @return The current background color.
 */
public Color getBackground() {
    Color color = null;
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (color == null) {
            color = cellFormat.getBackground();
        }
        else if (! color.equals((Object) cellFormat.getBackground())) {
            return null;
        }
    }
    return color;
}

/**
 * Sets the background color for this object (all its data formats).
 *
 * @param color The new background color.
 */
public void setBackground(Color color) {
    if (color != null) {
        for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
            CellFormat cellFormat = (CellFormat) e.nextElement();
            cellFormat.setBackground(color);
        }
    }
}

/**
 * Retrieves the font name for this object.  All data formats
 * must have the same font name, else a special value is returned.
 *
 * @return The current font name.
 */
public String getFontName() {
    String name = null;
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (name == null) {
            name = cellFormat.getFont().getName();
        }
        else if (! name.equals(cellFormat.getFont().getName())) {
            return null;
        }
    }
    return name;
}

/**
 * Retrieves the font style for this object.  All data formats
 * must have the same font style, else a special value is returned.
 *
 * @return The current font style.
 */
public int getFontStyle() {
    int style = MIXED_STYLE;
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (style == MIXED_STYLE) {
            style = cellFormat.getFont().getStyle();
        }
        else if (style != cellFormat.getFont().getStyle()) {
            return MIXED_STYLE;
        }
    }
    return style;
}

/**
 * Retrieves the font size for this object.  All data formats
 * must have the same font size, else a special value is returned.
 *
 * @return The current font size.
 */
public int getFontSize() {
    int size = MIXED_SIZE;
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (size == MIXED_SIZE) {
            size = cellFormat.getFont().getSize();
        }
        else if (size != cellFormat.getFont().getSize()) {
            return MIXED_SIZE;
        }
    }
    return size;
}

/**
 * Sets the font for this object (all its data formats).
 *
 * @param name The new font name.
 * @param style The new font style.
 * @param size The new font size.
 */
public void setFont(String name, int style, int size, FontMetrics fm) {
    if (name == null && style == MIXED_STYLE && size == MIXED_SIZE) {
        return;
    }
    Font oneFont = null;
    if (name != null && style != MIXED_STYLE && size != MIXED_SIZE) {
        oneFont = new Font(name, style, size);
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (oneFont != null) {
            cellFormat.setFont(fm , oneFont);
        }
        else {
            cellFormat.setFont(fm, new Font((name == null) ? cellFormat.getFont().getName() : name,
                                            (style == MIXED_STYLE) ? cellFormat.getFont().getStyle() : style,
                                            (size == MIXED_SIZE) ? cellFormat.getFont().getSize() : size));
        }
    }
}

/**
 * Retrieves the indicator formats for this level.
 *
 * @return Indicator formats for this level.
 */
public Vector getIndicatorFormats() {
    return indicatorFormats;
}

/**
 * Sets the indicator formats for this level.
 *
 * @param formats Indicator formats for this level.
 */
public void setIndicatorFormats(Vector formats) {
    indicatorFormats = formats;
    computeSize(indicatorFormats);
    heightChanged();
}

/**
 * Retrieves the data formats for this level.
 *
 * @return Data formats for this level.
 */
public Vector getDataFormats() {
    return dataFormats;
}

/**
 * Sets the data formats for this level.
 *
 * @param formats Data formats for this level.
 */
public void setDataFormats(Vector formats) {
    dataFormats = formats;
    computeSize(dataFormats);
    heightChanged();
}

protected void setCellWidths(int newWidth) {
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cell = (CellFormat) e.nextElement();
        cell.setWidth(newWidth, true);
    }
}

protected void heightChanged() {
    cachedPreferredHeight = USE_DEFAULT_HEIGHT_VALUE;
    setHeight(height);
}

/**
 * Get the preferred height of this row format.
 * It is based on the highest of the cells.
 *
 * @return The preferred height value.
 */
public int getPreferredHeight() {
    if (cachedPreferredHeight == USE_DEFAULT_HEIGHT_VALUE) {
        cachedPreferredHeight = calculatePreferredHeight();
    }
    return cachedPreferredHeight;
}

/**
 * Calculate the preferred height of this row format.
 * It is based on the highest of the cells.
 *
 * @return The preferred height value.
 */
protected int calculatePreferredHeight() {

    // the height is the largest of the cell heights
    int height = 0;
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        Rectangle cell = cellFormat.getPreferredTotalArea();
        height = Math.max(height, cell.height);
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (cellFormat.isShowing()) {
            Rectangle cell = cellFormat.getPreferredTotalArea();
            height = Math.max(height, cell.height);
        }
    }
    return height;
}

/**
 * Get the total width of all of the visible data cells.
 *
 * @return The total width of all of the visible data cells.
 */
protected int getDataWidth() {
    // the width is the total of the cell widths
    int width = 0;
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (cellFormat.isShowing()) {
            width += cellFormat.getPreferredTotalArea().width;
        }
    }
    return width;
}

protected void computeWidth() {
    // the width is the total of the cell widths
    int width = 0;
    for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        width += cellFormat.getPreferredTotalArea().width;
    }
    width += getDataWidth();
    setWidth(width);
}

protected void computeSize(Vector cells) {
    getHeight();
    computeWidth();
}

protected int getTotalWidth(HiGrid grid, FormatNode node) {
    // someone may have changed a cell width via mouse resizing
    computeWidth();
    return getInitialLevelIndent(grid, node) + getWidth();
}

protected int getTotalPrintWidth(HiGrid grid, FormatNode node, int printableWidth) {
    int printWidth = getInitialLevelIndent(grid, node);
    int totalPrintWidth = 0;
    int width = 0;
    if (grid.isDrawingFoldersAndConnections()) {
        for (Enumeration e = getIndicatorFormats().elements(); e.hasMoreElements();) {
            CellFormat cellFormat = (CellFormat) e.nextElement();
            width = cellFormat.getPreferredTotalArea().width;
            if ((printWidth + width) > printableWidth) {
                totalPrintWidth += printableWidth;
                printWidth = width;
            }
            else {
                printWidth += width;
            }
        }
    }
    for (Enumeration e = getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (cellFormat.isShowing()) {
            width = cellFormat.getPreferredTotalArea().width;
            if ((printWidth + width) > printableWidth) {
                totalPrintWidth += printableWidth;
                printWidth = width;
            }
            else {
                printWidth += width;
            }
        }
    }
    totalPrintWidth += printWidth;
    return totalPrintWidth;
}

protected int getNodeIndent(HiGrid grid, Vector indicatorFormats) {
    int index = grid.getIndicatorIndex(indicatorFormats, grid.NODE_INDICATOR_CELL);
    CellFormat nodeFormat = (CellFormat)indicatorFormats.elementAt(index);
    int nodeIndent = nodeFormat.getWidth() +
                     nodeFormat.getBorderInsets().left + nodeFormat.getBorderInsets().right +
                     nodeFormat.getMarginInsets().left + nodeFormat.getMarginInsets().right;
    return nodeIndent;
}

/**
 * Get the initial pixel indent for the given format level.
 *
 * @param grid the current instance of HiGrid.
 * @param node the FormatNode to examine.
 *
 * @return The number of pixels to indent before drawing a node.
 */
public int getInitialLevelIndent(HiGrid grid, FormatNode node) {
    if (grid.isLevelIndented()) {
        int nodeWidth = getNodeIndent(grid, node.getRecordFormat().getIndicatorFormats());
        return node.getIndent() * nodeWidth;
    }
    else {
        return 0;
    }
}

/**
 * Get the initial pixel indent for the given format level's indicator formats.
 *
 * @param grid the current instance of HiGrid.
 * @param node the FormatNode to examine.
 *
 * @return The number of pixels to indent before drawing the indicator formats.
 */
public int getIndicatorIndent(HiGrid grid, FormatNode node) {
    int nodeWidth = getNodeIndent(grid, node.getRecordFormat().getIndicatorFormats());
    int indentCount = grid.isLevelIndented() ? (node.getIndent()) : 0;
    int indent = (indentCount + 1) * nodeWidth;
    int index = grid.getIndicatorIndex(indicatorFormats, grid.EDITSTATUS_INDICATOR_CELL);
    if (index != -1) {
        CellFormat cellFormat = (CellFormat)indicatorFormats.elementAt(index);
        if (cellFormat.isShowing()) {
            indent += cellFormat.getTotalArea().width;
        }
    }
    return indent;
}

/**
 * Move down to prepare to draw the next row.
 *
 * @param gc the Graphics Context.
 * @param grid the current instance of HiGrid.
 * @param rowNode the current RowNode to move beyond.
 */
protected void nextRow(Graphics gc, HiGrid grid, RowNode rowNode) {
    if (gc instanceof NullGraphics) {
        NullGraphics nullgc = (NullGraphics)gc;
        if (nullgc.cell != null) {
            nullgc.cell.point.y += rowNode.getHeight();
        }
        nullgc.gridAreaRowsHeight += rowNode.getHeight();
    }
    else {
        gc.translate(0, rowNode.getHeight());
    }
}

/**
 * Determine whether this row can be visually connected to another row.
 *
 * @param rowNode the RowNode to examine.
 *
 * @return true if the row can be visually connected to another row.
 */
protected boolean shouldConnect(RowNode rowNode) {
    if (rowNode == null) {
        return false;
    }
    // HeaderFormat nodes are the only ones not connected by a line
    return(! (rowNode.getRowFormat() instanceof HeaderFormat));
}

/**
 * Draw a vector of cells within the given row.
 * This is a lower level routine that can optionally draw either
 * indicatorFormats or dataFormats.
 *
 * @param gc the Graphics Context.
 * @param grid the current instance of HiGrid.
 * @param rowNode the current RowNode to draw.
 * @param cells the cells to draw.
 * @param rect rectangle used for clipping.
 * @param intersectRect rectangle used for clipping.
 */
public void drawCells(Graphics gc, HiGrid grid, RowNode rowNode, Vector cells, Rectangle rect,
                      Rectangle intersectRect) {
    DataTableModel dataTable = rowNode.getDataTableModel();

    int xOffset = 0;
    boolean drawCell = false;
    int startColumn = 0;
    int columnIndex = 0;
    // inits for printing only
    if (grid.isPrinting()) {
        columnIndex = grid.getPrintColumnPageIndex();
    }
    // loop over all of the columns for this row object and draw each column
    for (Enumeration e = cells.elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (!(cellFormat.isShowing())) {
            continue;
        }
        String name = cellFormat.getName();
        Rectangle totalArea = cellFormat.getTotalArea();
        xOffset = totalArea.width;
        gc.setClip(totalArea);

        if (grid.isPrinting()) {
            int printableWidth = grid.getPrintGrid().getPrintableWidth();
            int startWidth = startColumn * printableWidth;
            int endWidth = startWidth + printableWidth;
            printRectOffset += xOffset;
            drawCell = false;
            if (printRectOffset > startWidth && printRectOffset <= endWidth) {
                if (startColumn == columnIndex) {
                    drawCell = true;
                }
            }
            else if (printRectOffset > endWidth) {
                printRectOffset = endWidth;
                if (startColumn == columnIndex) {
                    break;
                }
                else {
                    startColumn++;
                    printRectOffset += xOffset;
                    if (startColumn == columnIndex) {
                        drawCell = true;
                    }
                }
            }
            if (!drawCell) {
                continue;
            }
        }
        else {
            drawCell = !(((rect.x + xOffset) < intersectRect.x) ||
                         (rect.x > (intersectRect.x + intersectRect.width)));
        }
        // now do clipping based on cells
        if (drawCell) {
            JCCellRenderer renderer = grid.lookupTable.getCellRendererInstance(cellFormat.getCellRenderer());
            if (rowNode.getRowFormat() instanceof RecordFormat) {
                if (name != null && !(cellFormat instanceof IndicatorCellFormat)) {
                    int type = MetaDataModel.TYPE_STRING;
                    MetaDataModel metaData = rowNode.getFormatNode().getRecordFormat().getMetaData();
                    if (metaData != null) {
                        try {
                            type = metaData.getJavaColumnType(name);
                        }
                        catch (Exception exception) {
                            if (JCEnvironment.isDesignTime()) {
                                grid.adviseClearAllFormats = true;
                            }
                            else {
                                grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, exception));
                            }
                        }
                    }
                }
            }
            Object drawable = null;
            if (cellFormat instanceof NodeCellFormat) {
                // this case usually only happens when HiGrid is used as a bean
                if ((renderer == null) || !(renderer instanceof HiGridNodeRenderer)) {
                    cellFormat.setCellRenderer(grid.lookupTable.getNodeRenderer().getClass());
                    renderer = grid.lookupTable.getCellRendererInstance(cellFormat.getCellRenderer());
                }
                if (renderer instanceof HiGridNodeRenderer) {
                    // draw the folder icon beside the record
                    int state = rowNode.getState();
                    drawable = new NodeStatus(grid, state, grid.getFolderIcon(state));
                }
            }
            else if (cellFormat instanceof EditStatusCellFormat) {
                // this case usually only happens when HiGrid is used as a bean
                if ((renderer == null) || !(renderer instanceof HiGridEditStatusRenderer)) {
                    cellFormat.setCellRenderer(grid.lookupTable.getEditStatusRenderer().getClass());
                    renderer = grid.lookupTable.getCellRendererInstance(cellFormat.getCellRenderer());
                }
                if (renderer instanceof HiGridEditStatusRenderer) {
                    if (this instanceof RecordFormat) {
                        CellPosition currentCell = grid.getCurrentCell();
                        boolean isCurrentRow = (currentCell == null) ? false : currentCell.row == rowNode;
                        drawable = new EditStatus(grid, dataTable.getRowStatus(rowNode.getBookmark()), isCurrentRow, rowNode.getBookmark());
                    }
                }
            }
            else {
                if (this instanceof RecordFormat) {
                    try {
                        drawable = dataTable.getResultData(rowNode.getBookmark(), name);
                    }
                    catch (Exception exception) {
                        if (JCEnvironment.isDesignTime()) {
                            grid.adviseClearAllFormats = true;
                        }
                        else {
                            // we must now catch and ignore this exception, since
                            // it is probably caused by a race condition with multiple
                            // instances of HiGrid and deletion of the current row
//                            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, exception));
                        }
                    }
                }
                else if (this instanceof HeaderFormat) {
                    drawable = getHeaderLabel(grid, rowNode, cellFormat);
                }
                else {
                    // summary types
                    drawable = ((SummaryDataTable)((HiGridInternalData)dataTable.getInternalData(grid)).getDataTable()).getResultData(grid, dataTable, this, rowNode, name);
                }
            }
            if (! (renderer instanceof HiGridNodeRenderer)) {
                prepareDrawingArea(gc, cellFormat, totalArea, rowNode.isSelected());
            }
            Point p = cellFormat.getTranslationOffset();
            Rectangle drawingArea = cellFormat.getDrawingArea();
            if (renderer instanceof JCLightCellRenderer) {
                gc.translate(p.x, p.y);
                gc.setClip(drawingArea);
                ((JCLightCellRenderer)renderer).draw(gc, cellFormat, drawable, rowNode.isSelected());
                gc.translate(-p.x, -p.y);
            }
            else if (renderer instanceof JCComponentCellRenderer) {
                Component component = ((JCComponentCellRenderer)renderer).getRendererComponent(cellFormat, drawable, rowNode.isSelected());
                rendererPane.paintComponent(gc, component, grid, p.x, p.y, drawingArea.width, drawingArea.height, true);
            }
            else {
                // shouldn't happen
                throw new HiGridException("Bad cell renderer: "+renderer);
            }
        }
        rect.x += xOffset;
        gc.translate(xOffset, 0);
    }
    gc.setClip(-rect.x, 0, rect.width, rect.height);
}

/**
 * Draw an entire RowNode.
 *
 * @param gc the Graphics Context.
 * @param grid the current instance of HiGrid.
 * @param rowNode the RowNode to draw.
 *
 * @return true if the row can be drawn within the current clip bounds.
 */
public boolean draw(Graphics gc, HiGrid grid, RowNode rowNode) {
    if (gc instanceof NullGraphics) {
        return nullDraw((NullGraphics)gc, grid, rowNode);
    }

    // get original clipping rectangle
    Rectangle originalClipRect = gc.getClipBounds();
    if (-originalClipRect.y > originalClipRect.height) {
        // we've passed the end of the window
        return false;
    }

    int totalLevelIndent = getInitialLevelIndent(grid, rowNode.getFormatNode());
    // create logical clipping rectangle for this row
    Rectangle rowRect = new Rectangle(0, 0,
                                      totalLevelIndent + rowNode.getWidth(), rowNode.getHeight());

    Rectangle intersectRect = rowRect.intersection(originalClipRect);
    if (!(intersectRect.isEmpty())) {
        grid.fireHiGridRepaintEvent(new HiGridRepaintEvent(grid,
                                                           HiGridRepaintEvent.BEFORE_REPAINT_ROW, rowNode, intersectRect));

        Rectangle rect = new Rectangle(0, 0, 0, rowNode.getHeight());
        int nodeIndent = getNodeIndent(grid, rowNode.getRowFormat().getIndicatorFormats());
        RowNode node = rowNode;
        RowTree rowTree = grid.getRowTree();
        boolean drawIndicators = grid.isDrawingFoldersAndConnections();
        boolean drawConnections = (grid.isConnectionsVisible() && drawIndicators &&
                                   !grid.isPrintingSubsequentColumnPages());
        boolean indentRow = !grid.isPrintingSubsequentColumnPages();

        // draw the initial connecting lines for all previous levels
        rect.x += totalLevelIndent;
        int levelIndent = grid.isLevelIndented() ? nodeIndent : 0;
        // now do clipping based on connecting lines
        if ((rect.x >= intersectRect.x) &&
            (rect.x <= intersectRect.x + intersectRect.width)) {
            for (int i = 0; i < rowNode.getFormatNode().getIndent(); i++) {
                rect.x -= levelIndent;
                gc.translate(rect.x, 0);
                if (rowTree.getNextUncle(node) != null && drawConnections) {
                    Vector indicators = ((RowNode)(node.getParent())).getRowFormat().getIndicatorFormats();
                    gc.setColor(((CellFormat)indicators.elementAt(grid.getIndicatorIndex(indicators, grid.NODE_INDICATOR_CELL))).getForeground());
                    gc.drawLine(nodeIndent/2, 0, nodeIndent/2, rowNode.getHeight());
                }
                node = (RowNode)node.getParent();
                gc.translate(-rect.x, 0);
            }

            rect.x += totalLevelIndent;
        }

        int initialOffset = 0;
        if (indentRow) {
            gc.translate(rect.x, 0);
        }
        else {
            initialOffset = rect.x;
        }

        int verticalMiddle = (rowNode.getHeight() - 1) / 2;

        // do not connect the very first rows in the entire tree
/*
    if (! ((rowNode.getParent() == root) &&
        ((previousSibling == null) ||
         (previousSibling.getRowFormat() instanceof HeaderFormat) &&
          (previousSibling.getPreviousSibling() == null)))) {
        boolean connectNext = rowNode.getNextSibling() != null;
        gc.drawLine(nodeIndent/2 - 1, 0, nodeIndent/2 - 1, connectNext ? rowNode.getHeight() : verticalMiddle);
    }
*/
        if (drawConnections) {
            boolean connectNext = rowNode.getNextSibling() != null;
            Vector indicators = rowNode.getRowFormat().getIndicatorFormats();
            gc.setColor(((CellFormat)indicators.elementAt(grid.getIndicatorIndex(indicators, grid.NODE_INDICATOR_CELL))).getForeground());
            gc.drawLine(nodeIndent/2, 0, nodeIndent/2, connectNext ? rowNode.getHeight() : verticalMiddle);

            // draw the horizontal connecting lines
            if (shouldConnect(rowNode)) {
                // draw the horizontal connecting line
                gc.drawLine(nodeIndent/2, verticalMiddle, nodeIndent, verticalMiddle);
            }
        }
        if (grid.isPrinting()) {
            printRectOffset = rect.x;
        }
        if (drawIndicators) {
            drawCells(gc, grid, rowNode, getIndicatorFormats(), rect, intersectRect);
        }
        drawCells(gc, grid, rowNode, getDataFormats(), rect, intersectRect);

        gc.translate((-rect.x) + initialOffset, 0);
        gc.setClip(originalClipRect);
    }

    nextRow(gc, grid, rowNode);

    return true;
}

/**
 * "Pretend" to draw an entire RowNode.  This method is used
 * to accomplish hit detection within the grid.
 *
 * @param gc the Graphics Context.
 * @param grid the current instance of HiGrid.
 * @param rowNode the RowNode to draw.
 *
 * @return true if hit detection succeeded.
 */
protected boolean nullDraw(NullGraphics nullgc, HiGrid grid, RowNode rowNode) {
    if (nullgc.gridAreaRowsHeight > nullgc.gridAreaHeight) {
        // we can stop the operation now
        // the row is no longer visible in the gridArea
        return false;
    }
    if (nullgc.finishedOperation) {
        return false;
    }
    nullgc.fullyVisible = (nullgc.gridAreaRowsHeight + rowNode.getHeight()) <= nullgc.gridAreaHeight;
    switch (nullgc.operation) {
    case NullGraphics.OPERATION_HITTEST:
        // start with row hit-testing
        //FIX - assumption that height and next row are the same value
        int heightRowNode = rowNode.getHeight();
        if (nullgc.mouse.y >= heightRowNode) {
            nullgc.mouse.y -= heightRowNode;
        }
        else {
            // found it
            nullgc.cell.row = rowNode;
            hitTest(nullgc, grid, nullgc.cell.row);
            nullgc.finishedOperation = true;
        }
        break;

    case NullGraphics.OPERATION_GET_CELL:
    case NullGraphics.OPERATION_GET_INDICATOR:
        if (nullgc.cell.row == rowNode) {
            getXPosition(nullgc, grid, rowNode, nullgc.operation);
            nullgc.cell.row = rowNode;
            nullgc.finishedOperation = true;
        }
        break;

    case NullGraphics.OPERATION_GRIDAREA_ROWS:
        if (nullgc.fullyVisible) {
            nullgc.numberOfGridAreaRows++;
        }
        else {
            // finished, the row is not fully visible in the gridArea
            nullgc.finishedOperation = true;
        }
        break;

    case NullGraphics.OPERATION_NONE:
        break;
    }

    if (! nullgc.finishedOperation) {
        nextRow(nullgc, grid, rowNode);
    }
    return true;
}

protected boolean findFromFormat(NullGraphics nullgc, Enumeration e, int x) {
    for (int i = 0; e.hasMoreElements(); i++) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (cellFormat == nullgc.cell.format) {
            nullgc.cell.point.x = x;
            nullgc.cell.columnIndex = i;
            return true;
        }
        if (cellFormat.isShowing()) {
            x += cellFormat.getTotalArea().width;
        }
    }
    return false;
}

protected void getXPosition(NullGraphics nullgc, HiGrid grid, RowNode rowNode, int operation) {
    nullgc.cell.isIndicator = false;
    if (rowNode == null) {
        return;
    }

    // account for the indented level
    int x = getInitialLevelIndent(grid, rowNode.getFormatNode());

    Enumeration e = getIndicatorFormats().elements();
    if (operation == NullGraphics.OPERATION_GET_INDICATOR) {
        if (findFromFormat(nullgc, e, x)) {
            nullgc.cell.isIndicator = true;
        }
    }
    else if (operation == NullGraphics.OPERATION_GET_CELL) {
        // loop past indicatorFormats
        for (int i = 0; e.hasMoreElements(); i++) {
            CellFormat cellFormat = (CellFormat)e.nextElement();
            if (cellFormat.isShowing()) {
                x += cellFormat.getTotalArea().width;
            }
        }
        e = getDataFormats().elements();
        findFromFormat(nullgc, e, x);
    }
}

protected int hitTestWithinRow(NullGraphics nullgc, HiGrid grid, RowNode rowNode, Vector cells, int x) {
    boolean found = false;
    // while there are more cell format objects
    Enumeration e = cells.elements();
    for (int i = 0; e.hasMoreElements() && (nullgc.mouse.x > x) && (!found); i++) {
        CellFormat cell = (CellFormat) e.nextElement();
        if (!cell.isShowing()) {
            continue;
        }
        int columnX = x + cell.getTotalArea().width;
        // did we click within this cell?
        if ((nullgc.mouse.x >= x) && (nullgc.mouse.x < columnX)) {
            // FIX
            if (! nullgc.overColumnDivider) {
            }
            nullgc.cell.format = cell;
            nullgc.cell.point.x = x;
            nullgc.cell.columnIndex = i;
            found = true;
        }
        // are we hovering over a sizing column?
        if ((nullgc.mouse.x >= columnX - COLUMN_SIZING_TOLERANCE) &&
            (nullgc.mouse.x <= columnX + COLUMN_SIZING_TOLERANCE)) {
            nullgc.overColumnDivider = true;
            nullgc.columnX = columnX;
            nullgc.columnIndex = i;
            nullgc.resizeFormat = cell;
            nullgc.cell.format = cell;
            nullgc.cell.point.x = x;
            found = true;
        }
        x += cell.getTotalArea().width;
    }
    if (found) {
        x = -1;
    }
    return x;
}

protected void checkForRowResizing(NullGraphics nullgc, RowNode rowNode) {
    int rowYOffset = rowNode.getHeight() - nullgc.mouse.y;
    if (rowYOffset < ROW_SIZING_TOLERANCE) {
        nullgc.overRowDivider = true;
        nullgc.rowYOffset = rowYOffset;
    }
}

/**
 * Compute the number of visible dataFormat columns for this rowFormat.
 *
 * @return the number of visible dataFormat columns for this rowFormat.
 */
public int getVisibleColumnCount() {
    int visibleCount = 0;
    for (int i = 0; i < dataFormats.size(); i++) {
        if (((CellFormat)dataFormats.elementAt(i)).isShowing()) {
            visibleCount++;
        }
    }
    return visibleCount;
}

/**
 * Compute the visible column index from the actual column index.
 * The maximum visible column index is always less than or equal
 * to the maximum actual column index, since some columns can
 * be marked as invisible at runtime.
 *
 * @param actualColumnIndex the actual index into the dataFormat vector
 *
 * @return the visible column index.
 */
public int getVisibleColumnIndex(int actualColumnIndex) {
    int visibleIndex = -1;
    for (int i = 0; i < dataFormats.size(); i++) {
        if (((CellFormat)dataFormats.elementAt(i)).isShowing()) {
            visibleIndex++;
        }
        if (i >= actualColumnIndex) {
            break;
        }
    }
    return visibleIndex;
}

/**
 * Compute the actual column index from the visible column index.
 * The maximum visible column index is always less than or equal
 * to the maximum actual column index, since some columns can
 * be marked as invisible at runtime.
 *
 * @param visibleColumnIndex the visible index into the dataFormat vector
 *
 * @return the actual column index.
 */
public int getActualColumnIndex(int visibleColumnIndex) {
    int actualIndex = -1;
    for (int i = 0; i < dataFormats.size(); i++) {
        actualIndex++;
        if (((CellFormat)dataFormats.elementAt(i)).isShowing()) {
            visibleColumnIndex--;
        }
        if (visibleColumnIndex < 0) {
            break;
        }
    }
    return actualIndex;
}

protected void hitTest(NullGraphics nullgc, HiGrid grid, RowNode rowNode) {
    nullgc.cellLocation = NullGraphics.IN_NO_FORMAT;
    nullgc.cell.isIndicator = false;
    if (rowNode == null) {
        return;
    }
    int x = getInitialLevelIndent(grid, rowNode.getFormatNode());

    x = hitTestWithinRow(nullgc, grid, rowNode, getIndicatorFormats(), x);
    if (x == -1) {
        nullgc.cellLocation = NullGraphics.IN_INDICATOR_FORMAT;
        nullgc.cell.isIndicator = true;
        // handle special cases
        // first, the Node icons
        if (nullgc.cell.columnIndex ==
            grid.getIndicatorIndex(rowNode.getFormatNode(), grid.NODE_INDICATOR_CELL)) {
            if (rowNode.getRowFormat() instanceof RecordFormat) {
                nullgc.overFolderIcon = true;
            }
        }
        // second, the edit status indicator
        else if (nullgc.cell.columnIndex ==
                 grid.getIndicatorIndex(rowNode.getFormatNode(), grid.EDITSTATUS_INDICATOR_CELL)) {
            checkForRowResizing(nullgc, rowNode);
        }
    }
    else {
        x = hitTestWithinRow(nullgc, grid, rowNode, getDataFormats(), x);
        if (x == -1) {
            nullgc.cellLocation = NullGraphics.IN_CELL_FORMAT;
            if (grid.isRowHeightResizingAll()) {
                checkForRowResizing(nullgc, rowNode);
            }
        }
    }
}

/**
 * Paints the background and the border for the given cell.
 *
 * @param gc the Graphics Context.
 * @param cellFormat The cell to paint.
 * @param totalArea the rectangle to paint.
 * @param selected true if the cell should be painted as selected.
 */
public static void prepareDrawingArea(Graphics gc, CellFormat cellFormat,
                                      Rectangle totalArea, boolean selected) {
    Color background = selected ? cellFormat.getSelectedBackground() : cellFormat.getBackground();
    gc.setColor(background);
    gc.fillRect(0, 0, totalArea.width, totalArea.height);
    gc.setColor(cellFormat.getForeground());
    HiGridBorder.draw(gc, cellFormat.getBorderStyle(), cellFormat.getBorderInsets().left,
                      0, 0, totalArea.width, totalArea.height, background, cellFormat.getForeground());
}

/**
 * Setup the header label for the indicated cell.
 *
 * @param grid The higrid.
 * @param rowNode The row node.
 * @param cellFormat The cell format.
 *
 * @return The sort label object.
 */
public Object getHeaderLabel(HiGrid grid, RowNode rowNode, CellFormat cellFormat) {
    SortLabel sortLabel = null;
    String columnName = cellFormat.getName();
    String columnText = cellFormat.getText();
    int direction = SortGrid.UNKNOWN;
    if (rowNode.getDataTableModel() != null) {
        HiGridInternalData internalData = (HiGridInternalData)rowNode.getDataTableModel().getInternalData(grid);
        if (internalData != null) {
            SortData sortData = internalData.getSortData();
            if (sortData != null) {
                if (sortData.getColumnName().compareTo(columnName) == 0) {
                    direction = sortData.getDirection();
                }
            }
        }
    }
    sortLabel = new SortLabel(columnText, direction);
    return(Object) sortLabel;
}

/**
 * Setup the indicator formats.
 */
public void createIndicatorFormats(HiGrid grid) {
    Vector cellFormatVector = new Vector(HiGrid.NUMBER_OF_INDICATOR_CELLS);
    cellFormatVector.addElement(createNodeFormat(grid));
    cellFormatVector.addElement(createEditStatus(grid));
    setIndicatorFormats(cellFormatVector);
}

/**
 * Creates the edit status indicator column.
 *
 * @return The CellFormat for the edit status indicator column.
 */
protected CellFormat createEditStatus(HiGrid grid) {
    EditStatusCellFormat cellFormat = new EditStatusCellFormat(grid);
    cellFormat.setParent(this);
    if (grid != null) {
        cellFormat.setFont(grid.getFontMetrics(cellFormat.getFont()), cellFormat.getFont());
        cellFormat.setCellRenderer(grid.lookupTable.getEditStatusRenderer().getClass());
        cellFormat.setWidth(grid.getEditStatusWidth());
    }
    else {
        cellFormat.setWidth(HiGrid.INITIAL_EDIT_STATUS_WIDTH);
    }
    return cellFormat;
}

/**
 * Creates the node column.
 *
 * @return The CellFormat for the nodes.
 */
protected CellFormat createNodeFormat(HiGrid grid) {
    NodeCellFormat cellFormat = new NodeCellFormat(grid);
    cellFormat.setParent(this);
    if (grid != null) {
        cellFormat.setFont(grid.getFontMetrics(cellFormat.getFont()), cellFormat.getFont());
        cellFormat.setCellRenderer(grid.lookupTable.getNodeRenderer().getClass());
        cellFormat.setWidth(grid.getNodeWidth());
    }
    else {
        cellFormat.setWidth(HiGrid.INITIAL_NODE_WIDTH);
    }
    return cellFormat;
}

}
