/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: RowNode.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * The RowNode class aggregates all of the information necessary to render
 * a specific row within a given table.  A RowNode is constructed from one
 * of the five basic RowFormat subclasses: Record, Header, Footer, BeforeDetails, AfterDetails.
 * Thus, a RowNode can either be based entirely on information from a
 * FormatNode (e.g. Header), or from information from the datasource (e.g.
 * Record), or from some computed information (Footer, BeforeDetails, AfterDetails).
 * An individual instance of the RowNode class can usually be found as an element
 * in the (@ link RowTree}.
 */
public class RowNode extends DataSourceTreeNode {

static final long serialVersionUID = 6907178340899502322L;

private static final byte TRUE = 1;
private static final byte FALSE = 0;

// Properties
private FormatNode formatNode = null;
private RowFormat rowFormat = null;
private DataTableModel dataTableModel = null;
private long bookmark = 0;
private byte state = NodeStatus.FOLDER_NONE;
private byte selected = FALSE;

/** Default constructor. */
public RowNode() {
    super(null);
}

/**
 * Constructor for RowNode.
 * @param node The FormatNode to associate with this object.
 * @param format The RowFormat to associate with this object.
 * @param dataTableModel The DataTableModel to associate with this object.
 */
public RowNode(FormatNode node, RowFormat format, DataTableModel dataTableModel, long bookmark) {
    this();
    setFormatNode(node);
    setRowFormat(format);
    if (format instanceof RecordFormat) {
        setState(NodeStatus.FOLDER_CLOSED);
    }
    this.dataTableModel = dataTableModel;
    this.bookmark = bookmark;
}

/**
 * Retrieves the bookmark for this object.
 *
 * @return The object's bookmark.
 */
public long getBookmark() {
    return bookmark;
}

/**
 * Retrieves the DataTableModel for this object.
 *
 * @return The current DataTableModel.
 */
public DataTableModel getDataTableModel() {
    return dataTableModel;
}

/**
 * Sets the DataTableModel for this object.
 *
 * @param dataTableModel The new DataTableModel for this object.
 */
protected void setDataTableModel(DataTableModel dataTableModel) {
    this.dataTableModel = dataTableModel;
}

/**
 * Retrieves the format node for this object.
 *
 * @return The current format node.
 */
public FormatNode getFormatNode() {
    return formatNode;
}

/**
 * Sets the format node for this object.
 *
 * @param node The new format node.
 */
protected void setFormatNode(FormatNode node) {
    formatNode = node;
}

/**
 * Retrieves the row format for this object.
 *
 * @return The current row format.
 */
public RowFormat getRowFormat() {
    return rowFormat;
}

/**
 * Sets the row format for this object.
 *
 * @param format The new row format.
 */
protected void setRowFormat(RowFormat format) {
    rowFormat = format;
}

/**
 * Retrieves the state of the row.
 *
 * @return The state of the row.
 */
public byte getState() {
    return state;
}

/**
 * Sets the state of the row
 *
 * @param state The state of the row.
 */
protected void setState(byte state) {
    this.state = state;
}

/**
 * Retrieves the selected state of the row.
 *
 * @return The selected state of the row.
 */
public boolean isSelected() {
    return selected == TRUE;
}

/**
 * Sets the selected state of the row
 *
 * @param selected The selected state of the row.
 */
protected void setSelected(boolean selected) {
    this.selected = selected ? TRUE : FALSE;
}

/**
 * Delegates to the rowFormat object to draw itself.
 *
 * @param gc The graphics context to draw into.
 * @param grid The instance of HiGrid this row belongs to.
 */
protected boolean draw(Graphics gc, HiGrid grid) {
    return rowFormat.draw(gc, grid, this);
}

/**
 * Retrieves the height of this object.
 *
 * @return The current height.
 */
public int getHeight() {
    return rowFormat.getHeight();
}

/**
 * Sets the height of this object.
 *
 * @param height The new height.
 */
protected void setHeight(int height) {
    rowFormat.setHeight(height);
}

/**
 * Retrieves the width of this object.
 *
 * @return The current width.
 */
public int getWidth() {
    return rowFormat.getWidth();
}

/**
 * Sets the width of this object.
 *
 * @param height The new width.
 */
protected void setWidth(int width) {
    rowFormat.setWidth(width);
}

/**
 * Returns a string representation of the object.
 *
 * @return a string representation of the object.
 */
public String toString() {
    String s = rowFormat== null ? "" : "RowFormat=" + rowFormat.toString();
    if (rowFormat instanceof RecordFormat) {
        s += " Bookmark=" + bookmark;
    }
    s += " State=" + state + " Selected=" + (selected == TRUE);
    return s;
}

}
