/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: RowTree.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.awt.Graphics;
import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * The <code>RowTree</code> holds the run-time visual information for the
 * levels of an instance of HiGrid. It is composed entirely of {&#64; link RowNode}
 * objects.
 */

public class RowTree extends DataSourceTreeModel {

static final long serialVersionUID = -1732819294164988029L;

private HiGrid grid;
private boolean runtime;
private RowNode topRowNode = null;
private int gridAreaWidth = 0;
private int gridAreaRows = 0;
private int totalNumberOfScrollableRows = 0;
private int scrollableRowsHeight = 0;
private int lastPageTopRow = 0;
private RowNode lastPageTopRowNode = null;
private int metaIDArray[] = null;
private int firstLevelGridAreaWidth = 0;

// work variables
private FormatNode lastHeaderFormatNode;
private boolean sortTable = true;

/**
 * Constructor for the RowTree object.
 *
 * @param grid The HiGrid parent.
 * @param runtime True if created at run-time, false if at design-time.
*/
public RowTree(HiGrid grid, boolean runtime) {
    super(DataSourceTreeModel.NullRoot);
    RowNode rootRowNode = createRowNode(null, null, null, 0);
    setRoot(rootRowNode);
    this.grid = grid;
    this.runtime = runtime;
}

/**
 * Constructor for the RowTree object.
 *
 * @param grid The HiGrid parent.
*/
public RowTree(HiGrid grid) {
    this(grid, true);
}

/**
 * Set the indication that setup of the row tree,
 * sorting may be performed.
 *
 * @param sort True if sorting is to be performed if required.
 */
protected void setSortTable(boolean sort) {
    sortTable = sort;
}

/**
 * Creates the runtime RowTree.  This routine is called whenever the
 * tree needs to created (when the grid is initialized and first
 * displayed) or whenever the grids needs to be regenerated.
  *
 * @param formatTree The FormatTree for the higrid.
*/
protected void createRuntimeTree(FormatTree formatTree) {
    if (formatTree == null || grid.getDataModel() == null) {
        return;
    }
    FormatNode formatNode = (FormatNode)formatTree.getRoot();
    if (formatNode == null) {
        return;
    }
    // create the invisible root of the RowTree
    RowNode rootRowNode = createRowNode(formatNode, null, null, 0);
    setRoot(rootRowNode);
    lastHeaderFormatNode = null;
    if (grid.getDataModel() == null) {
        return;
    }
    DataTableModel rootDataTable = (DataTableModel) grid.getDataModel().getDataTableTree().getRoot();
    if (runtime) {
        ((HiGridData) grid.getDataModel()).setSummaryDataRoot(grid, rootDataTable);
    }
    sortTable = true;
    createRuntimeChildren(rootRowNode, formatNode, rootDataTable, false, null);

    // update the scrollable rows info
    scrollableRowsHeight = 0;
    setTotalNumberOfScrollableRows(0);
    addTotalNumberOfScrollableRows(getNumberOfScrollableRows(rootRowNode, true));

    // update the grid area info
    if (runtime) {
        updateGridState();
    }

    // reset the summary data tables
    ((HiGridData)grid.getDataModel()).resetSummaryDataTables(grid, true);

    // inform the interested parties
    if (runtime) {
        grid.fireHiGridUpdateEvent(new HiGridUpdateEvent(grid,
                                                         HiGridUpdateEvent.AFTER_RESET_HIGRID_DATA));
    }
}

/**
 * Creates the print time row tree.
 *
 * @param formatTree The FormatTree for the higrid.
*/
protected void createPrintTimeTree(FormatTree formatTree, int metaIDArray[]) {
    if (metaIDArray == null) {
        return;
    }
    this.metaIDArray = metaIDArray;
    createRuntimeTree(formatTree);
}

/**
 * A helper routine that is called by createRuntimeTree.
 * It recursively calls other versions of itself to create row objects,
 * based on whether the children should be automatically expanded.
 *
 * @param rootObject The row which will be the parent of the created rows.
 * @param node The FormatNode which describes the visual format of the created rows.
 * @param dataModel The data model for the created rows.
 * @param checkBeforeDetails True if possible BeforeDetails row node is to be created
 * @param rnBefore The row node to insert before (if any)
 *
 * @return The number of scrollable rows for this node/dataModel
 */
protected int createRuntimeChildren(RowNode rootObject, FormatNode node, DataTableModel dataModel,
                                    boolean checkBeforeDetails, RowNode rnBefore) {
    int numberOfScrollableRows = 0;
    try {
        numberOfScrollableRows = dataModel.getRowCount();
    }
    catch (Exception e) {
        grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
    }
    // finished if no scrollable rows
    if (numberOfScrollableRows == 0) {
        return 0;
    }

    RowNode rowNode = null;

    // add possible before details for the rootObject
    FormatNode rootLevel = rootObject.getFormatNode();
    if (checkBeforeDetails && rootLevel.getBeforeDetailsFormat().isShowing()) {
        rowNode = createRowNode(node, rootLevel.getBeforeDetailsFormat(),
                                rootObject.getDataTableModel(), 0);
        addRowNode(rootObject, rowNode, rnBefore);
    }

    long bookmark = 0;
    HeaderFormat headerFormat = node.getHeaderFormat();
    for (int i = 0; i < numberOfScrollableRows; i++) {
        // add possible first and repeat headers
        if (headerFormat.isShowing() && ((i == 0) ||
                                         (headerFormat.isRepeatHeader() && lastHeaderFormatNode != node))) {
            rowNode = createRowNode(node, headerFormat, dataModel, 0);
            addRowNode(rootObject, rowNode, rnBefore);
            lastHeaderFormatNode = node;
        }

        // add record
        bookmark = dataModel.getRowIdentifier(i);
        rowNode = createRowNode(node, node.getRecordFormat(), dataModel, bookmark);
        addRowNode(rootObject, rowNode, rnBefore);
        if (node.hasChildren()) {
            boolean createChildren = false;
            int autoTraverse = node.getAutoTraverse();
            if (runtime) {
                switch (autoTraverse) {
                case FormatNode.AUTO_TRAVERSE_NONE:
                    break;
                case FormatNode.AUTO_TRAVERSE_DETECT_CHILDREN:
                    createChildren = true;
                    break;
                case FormatNode.AUTO_TRAVERSE_EXPAND:
                    createChildren = grid.fireHiGridExpansionEvent(new HiGridExpansionEvent(grid,
                                                                                            HiGridExpansionEvent.BEFORE_EXPAND_ROW, rowNode));
                    break;
                case FormatNode.AUTO_TRAVERSE_CALLBACK:
                    if (grid.getAutoTraverseCallback() != null) {
                        byte rowState = grid.getAutoTraverseCallback().getState(grid, rowNode);
                        if (rowState == NodeStatus.FOLDER_OPEN) {
                            createChildren = grid.fireHiGridExpansionEvent(new HiGridExpansionEvent(grid,
                                                                                                    HiGridExpansionEvent.BEFORE_EXPAND_ROW, rowNode));
                        }
                        else {
                            rowNode.setState(rowState);
                            createChildren = false;
                        }
                    }
                    break;
                }
            }
            else {
                createChildren = true;
            }
            if (createChildren) {
                rowNode.setState(NodeStatus.FOLDER_CLOSED);
                openFolder(rowNode);
                if (rowNode.hasChildren()) {
                    if (runtime && (autoTraverse == FormatNode.AUTO_TRAVERSE_DETECT_CHILDREN)) {
                        closeFolder(rowNode);
                        lastHeaderFormatNode = node;
                    }
                }
            }
        }
        else {
            rowNode.setState(NodeStatus.FOLDER_NONE);
        }
    }

    // add possible footer
    if (node.getFooterFormat().isShowing()) {
        rowNode = createRowNode(node, node.getFooterFormat(), dataModel, 0);
        addRowNode(rootObject, rowNode, rnBefore);
    }

    // possibly sort the table
    if (sortTable && (dataModel != null)) {
        SortData sortData = node.getDefaultSortData();
        if (sortData != null) {
            SortGrid sortGrid = new SortGrid(grid, false);
            sortGrid.sortColumn(node, findFirstRowNode(rootObject, dataModel), sortData);
        }
		else{
			//  If a default SortData object wasn't already added, add it now.
			sortData = new SortData("");
			sortData.setDirection(SortGrid.UNKNOWN);
			node.setDefaultSortData(sortData);
		}
    }

    return numberOfScrollableRows;
}

/**
 * Helper method to create a row node.
 *
 * @param node The format node.
 * @param format The row format.
 * @param dataTableModel The data table model.
 * @param bookmark The row bookmark.
 *
 * @return The new row node.
 */
protected RowNode createRowNode(FormatNode node, RowFormat format, DataTableModel dataTableModel, long bookmark) {
    RowNode rowNode = new RowNode(node, format, dataTableModel, bookmark);
    if (runtime) {
        grid.fireHiGridUpdateEvent(new HiGridUpdateEvent(grid,
                                                         HiGridUpdateEvent.AFTER_CREATE_ROW, rowNode));
    }
    return rowNode;
}

/**
 * Add row node to the row tree.
 *
 * @param rnParent The parent row node.
 * @param rn The row node to insert.
 * @param rnBefore The row node to insert before
 */
protected void addRowNode(RowNode rnParent, RowNode rn, RowNode rnBefore) {
    if (rnBefore == null) {
        rnParent.add(rn);
    }
    else {
        rnParent.insert(rnBefore, rn);
    }
}

/**
 * Determine if the given format node is to be printed.
 *
 * @param node The format node.
 *
 * @return True if it is to be printed.
 */
protected boolean isPrintNode(FormatNode node) {
    if (grid.getPrintFormat() == HiGrid.PRINT_EXPAND_ALL) {
        return true;
    }
    for (int i = 0; i < metaIDArray.length; i++) {
        if (metaIDArray[i] == node.getRecordFormat().getMetaData().getMetaID()) {
            return true;
        }
    }
    return false;
}

/**
 * A helper routine that is called by createRuntimeChildren.
 * It recursively calls itself to create row objects, based on whether
 * the children should be automatically expanded
 *
 * @param rootObject The row which will be the parent of the created rows.
 * @exception DataModelException If an error occurs creating a new table.
 */
protected void createRuntimeChildren(RowNode rootObject) throws DataModelException {
    int numberOfScrollableRows = 0;

    // loop over all subLevels of this rowNode
    TreeIterator ti = rootObject.getFormatNode().getIterator();
    FormatNode node = null;
    boolean doSorting = false;
    while (ti.hasMoreElements()) {
        node = (FormatNode)ti.get();
        if (!runtime && !isPrintNode(node)) {
            ti.nextElement();
            continue;
        }
        doSorting = false;
        DataTableModel dataModel = rootObject.getDataTableModel().getTable(rootObject.getBookmark(),
                                                                           (DataSourceTreeNode)node.getRecordFormat().getMetaData());
        if (dataModel == null) {
            try {
                dataModel = rootObject.getDataTableModel().createTable(
                                                                      grid.getDataModelListener(), rootObject.getBookmark(),
                                                                      (DataSourceTreeNode)node.getRecordFormat().getMetaData());
            }
            catch (Exception e) {
                if (e instanceof DataModelException) {
                    throw (DataModelException) e;
                }
                else {
                    grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
                }
            }
        }
        if (dataModel == null) {
            // there *could* be rows here, but there aren't any right now
            ti.nextElement();
            continue;
        }
        HiGridInternalData internalData = (HiGridInternalData)dataModel.getInternalData(grid);
        if ((internalData == null) || (internalData.getDataTable() == null)) {
            ((HiGridData)grid.getDataModel()).createSummaryDataTable(grid, dataModel, node);
        }

        numberOfScrollableRows += createRuntimeChildren(rootObject, node, dataModel,
                                                        (numberOfScrollableRows == 0), null);
        ti.nextElement();
    }

    // add possible after details for the rootObject
    FormatNode rootLevel = rootObject.getFormatNode();
    if (numberOfScrollableRows > 0 && rootLevel.getAfterDetailsFormat().isShowing()) {
        RowNode rowNode = createRowNode(node, rootLevel.getAfterDetailsFormat(),
                                        rootObject.getDataTableModel(), 0);
        rootObject.add(rowNode);
    }
}

/**
 * Add a row node for a given data table model to the specified row node parent.
 *
 * @param rootObject The row which will be the parent of the created row.
 * @param dataTableModel The data table model of the new row node.
 *
 * @return The format node of the row node to be inserted.
 */
protected FormatNode createRuntimeChildren(RowNode rootObject, DataTableModel dataTableModel) {
    int numberOfScrollableRows = 0;
    FormatNode formatNode = null;

    // loop over all subLevels of this rowNode
    TreeIterator ti = rootObject.getFormatNode().getIterator();
    FormatNode node = null;
    while (ti.hasMoreElements()) {
        node = (FormatNode)ti.get();
        DataTableModel dataModel = rootObject.getDataTableModel().getTable(rootObject.getBookmark(),
                                                                           (DataSourceTreeNode)node.getRecordFormat().getMetaData());
        if (dataModel == null) {
            // there *could* be rows here, but there aren't any right now
            ti.nextElement();
            continue;
        }

        if (dataModel == dataTableModel) {
            // found the spot to insert the row node
            formatNode = node;
            RowNode rnBefore = null;
            ti.nextElement();
            if (ti.hasMoreElements()) {
                FormatNode nextNode = (FormatNode)ti.get();
                rnBefore = findInsertBeforeFormatNode(rootObject, nextNode);
            }
            numberOfScrollableRows += createRuntimeChildren(rootObject, node, dataModel,
                                                            (numberOfScrollableRows == 0),
                                                            rnBefore);
            continue;
        }
        else {
            // these are the other existing children
            try {
                numberOfScrollableRows += dataModel.getRowCount();
            }
            catch (Exception e) {
                grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
            }
        }
        ti.nextElement();
    }
    return formatNode;
}

/**
 * Add possible repeat header after current row node
 *
 * @param rowNode The row node before the possible repeat header.
 */
protected void addRepeatHeader(RowNode rowNode) {
    FormatNode node = rowNode.getFormatNode();
    HeaderFormat headerFormat = node.getHeaderFormat();
    if (headerFormat.isShowing() && headerFormat.isRepeatHeader()) {
        RowNode rn = findNextNode(rowNode, node);
        if (rn != null && rn.getRowFormat() instanceof RecordFormat) {
            RowNode hn = createRowNode(node, headerFormat,
                                       rn.getDataTableModel(), 0);
            ((RowNode)rowNode.getParent()).insert(rn, hn);
            addTotalNumberOfScrollableRows(hn, node);
        }
    }
}

/**
 * Remove possible repeat header after current row node
 *
 * @param rowNode The row node before the possible repeat header.
 */
protected void removeRepeatHeader(RowNode rowNode) {
    FormatNode node = rowNode.getFormatNode();
    HeaderFormat headerFormat = node.getHeaderFormat();
    if (headerFormat.isShowing() && headerFormat.isRepeatHeader()) {
        RowNode rn = findNextNode(rowNode, node);
        if (rn != null && rn.getRowFormat() instanceof HeaderFormat) {
            subtractTotalNumberOfScrollableRows(rn, node);
            ((RowNode)rowNode.getParent()).remove(rn);
        }
    }
}

/**
 * Retrieves the total number of scrollable rows in the grid.
 *
 * @return The total number of scrollable rows in the grid.
 */
public int getTotalNumberOfScrollableRows() {
    return totalNumberOfScrollableRows;
}

/**
 * Sets the total number of scrollable rows in the grid.
 *
 * @param rows The total number of scrollable rows in the grid.
 */
protected void setTotalNumberOfScrollableRows(int rows) {
    totalNumberOfScrollableRows = rows;
    if (totalNumberOfScrollableRows < 0) {
        totalNumberOfScrollableRows = 0;
    }
}

/**
 * Adds to the total number of scrollable rows in the grid.
 *
 * @param rows The number of scrollable rows to add to the grid.
 */
protected void addTotalNumberOfScrollableRows(int rows) {
    setTotalNumberOfScrollableRows(totalNumberOfScrollableRows + rows);
}

/**
 * Subtracts from the total number of scrollable rows in the grid.
 *
 * @param rows The number of scrollable rows to subtract from the grid.
 */
protected void subtractTotalNumberOfScrollableRows(int rows) {
    setTotalNumberOfScrollableRows(totalNumberOfScrollableRows - rows);
}

/**
 * Adds to the total number of scrollable rows in the grid
 * for the given row.
 *
 * @param rowNode The row node to add.
 * @param formatNode The format node of the row node.
 */
protected void addTotalNumberOfScrollableRows(RowNode rowNode, FormatNode formatNode) {
    addTotalNumberOfScrollableRows(1);
    formatNode.addNumberOfScrollableRows(1);
    scrollableRowsHeight += rowNode.getHeight();
}

/**
 * Subtracts from the total number of scrollable rows in the grid
 * for the given row.
 *
 * @param rowNode The row node to subtract.
 * @param formatNode The format node of the row node.
 */
protected void subtractTotalNumberOfScrollableRows(RowNode rowNode, FormatNode formatNode) {
    subtractTotalNumberOfScrollableRows(1);
    formatNode.subtractNumberOfScrollableRows(1);
    scrollableRowsHeight -= rowNode.getHeight();
    if (scrollableRowsHeight < 0) {
        scrollableRowsHeight = 0;
        System.out.println("scrollableRowsHeight less than zero");
    }
}

/**
 * Retrieves the number of scrollable rows beneath this row.
 * Also as it traverses the nodes, add/subtract the row count in the
 * FormatNodes.
 *
 * @param rootObject The parent row.
 * @param add Indicates to add or subtract the row count.
 *
 * @return The number of scrollable rows beneath this node.
 */
protected int getNumberOfScrollableRows(RowNode rootObject, boolean add) {
    if (rootObject == null) {
        return 0;
    }
    int numberOfChildrenRows = 0;
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getState() == NodeStatus.FOLDER_OPEN) {
            numberOfChildrenRows += getNumberOfScrollableRows(rowNode, add);
        }
        numberOfChildrenRows++;
        FormatNode formatNode = rowNode.getFormatNode();
        if (formatNode != null) {
            if (add) {
                formatNode.addNumberOfScrollableRows(1);
                scrollableRowsHeight += rowNode.getHeight();
            }
            else {
                formatNode.subtractNumberOfScrollableRows(1);
                scrollableRowsHeight -= rowNode.getHeight();
                if (scrollableRowsHeight < 0) {
                    scrollableRowsHeight = 0;
                    System.out.println("scrollableRowsHeight less than zero");
                }
            }
        }
        ti.nextElement();
    }
    return numberOfChildrenRows;
}

/**
 * Retrieves the number of scrollable rows beneath this row
 * of a given format node.
 * It add/subtract the row count in the FormatNode's.
 *
 * @param rootObject The parent row.
 * @param node The format node.
 * @param add Indicates the add or subtract the row count.
 *
 * @return The number of scrollable rows found.
 */
protected int getNumberOfScrollableRows(RowNode rootObject, FormatNode node, boolean add) {
    if (rootObject == null) {
        return 0;
    }
    int numberOfChildrenRows = 0;
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        FormatNode formatNode = rowNode.getFormatNode();
        if (formatNode != node) {
            ti.nextElement();
            continue;
        }
        numberOfChildrenRows++;
        if (add) {
            formatNode.addNumberOfScrollableRows(1);
            scrollableRowsHeight += rowNode.getHeight();
        }
        else {
            formatNode.subtractNumberOfScrollableRows(1);
            scrollableRowsHeight -= rowNode.getHeight();
            if (scrollableRowsHeight < 0) {
                scrollableRowsHeight = 0;
                System.out.println("scrollableRowsHeight less than zero");
            }
        }
        ti.nextElement();
    }
    return numberOfChildrenRows;
}

/**
 * Retrieves the ScrollableRows height in pixels.
 *
 * @return The ScrollableRows height.
 */
protected int getScrollableRowsHeight() {
    return scrollableRowsHeight;
}

/**
 * Sets the ScrollableRows height in pixels.
 *
 * @param height The ScrollableRows height.
 */
protected void setScrollableRowsHeight(int height) {
    scrollableRowsHeight = height;
}

/**
 * Retrieves the gridArea width.
 *
 * @return The gridArea width.
 */
public int getGridAreaWidth() {
    return gridAreaWidth;
}

/**
 * Sets the gridArea width.
 *
 * @param width The gridArea width.
 */
protected void setGridAreaWidth(int width) {
    gridAreaWidth = width + grid.getExtraWidth();;
//    Rectangle rect = gridArea.getBounds();
//    gridArea.setBounds(rect.x, rect.y, width, rect.height);
}

/**
 * Retrieves the first format node level width.
 *
 * @return The first level width.
 */
protected int getFirstLevelGridAreaWidth() {
    return firstLevelGridAreaWidth;
}

/**
 * Sets the first format node level width.
 *
 * @param The first level width.
 */
protected void setFirstLevelGridAreaWidth(int width) {
    firstLevelGridAreaWidth = width + grid.getExtraWidth();;
}

/**
 * Retrieves the number of rows fully visible in the gridArea
 *
 * @return The number of rows in the gridArea
 */
protected int getGridAreaRows() {
    return gridAreaRows;
}
/**
 * Sets the number of rows fully visible in the gridArea
 * and calculate the last top row.
 *
 * @param rowIndex The row index of the top row in the grid area.
 * @param rowNode The row node for the row index
 *
 * @return The true row index of the top row in the grid area.
 */
protected int setGridAreaRows() {
    // sanity check
    if (grid.getDataModel() == null) {
        return 0;
    }

    // should only have to be done here
    findLastPageTopRow();

    RowNode rowNode = getTopRowNode();
    int rowIndex = grid.getVerticalScrollbar().getValue();
    if (rowNode == null) {
        rowNode = (RowNode)getRootNode().getFirstChild();
        RowWalk walkable = new RowWalk(rowIndex, true);
        walk(rowNode, walkable);
        rowNode = walkable.getFoundNode();
    }
    return setGridAreaRows(rowIndex, rowNode);
}

protected int setGridAreaRows(int rowIndex, RowNode rowNode) {
    // change row index if it is in last page
    if (rowIndex > lastPageTopRow) {
        rowIndex = lastPageTopRow;
        rowNode = lastPageTopRowNode;
    }

    PageDownWalk walkable = new PageDownWalk(rowIndex, grid.getGridArea().getSize().height);
    walk(rowNode, walkable);
    gridAreaRows = walkable.getRowIndex() - rowIndex + 1;
    setTopRowNode(rowNode);
    grid.getVerticalScrollbar().setValue(rowIndex);
    return rowIndex;
}

/**
 * Update the visible grid area info.
 *
 * @param rowIndex The row index of the top row in the grid area.
 * @param rowNode The row node for the row index
 *
 * @return The true row index of the top row in the grid area.
 */
protected int updateGridArea(int rowIndex, RowNode rowNode) {
    rowIndex = setGridAreaRows(rowIndex, rowNode);
    grid.layoutGrid();
    return rowIndex;
}

protected void updateGridArea() {
    updateGridState();
    grid.layoutGrid();
}

/**
 * Compute the maximum width of scrollable rows for the grid area
 */
protected void computeTotalWidth() {
    // allow some extra width to resize the last column
    setGridAreaWidth(computeTotalWidth((FormatNode)grid.getFormatTree().getRoot(), true));
}

/**
 * Compute the maximum width of scrollable rows for this format node and its children
 */
protected int computeTotalWidth(FormatNode node, boolean root) {
    if (node == null) {
        return 0;
    }
    int totalWidth = node.getTotalWidth(grid);
    if (root) {
        setFirstLevelGridAreaWidth(totalWidth);
    }
    TreeIterator i = node.getIterator();
    while (i.hasMoreElements()) {
        FormatNode nextNode = (FormatNode)i.get();
        totalWidth = Math.max(totalWidth, computeTotalWidth(nextNode, false));
        i.nextElement();
    }
    return totalWidth;
}

/**
 * Whenever the visible grid has changed, the state has to be updated
 */
protected void updateGridState() {
    computeTotalWidth();
    setGridAreaRows();
}

/**
 * Draws the higrid into the specified Graphics context.
 * This routine will normally be called by a child to display
 * the grid in a sub-window, or may be called by mouse hit-testing
 * routines.
 *
 * @param gc The Graphics Context to use.  If gc is null, the tree
 * is traversed for visible rows in the Grid Area.
 */
protected void displayRuntimeTree(Graphics gc) {
    // sanity check
    if (grid.getDataModel() == null) {
        return;
    }

    // may be the first time
    grid.createDefaultTrees();
    displayThisRowAndBelow(gc, getTopRowNode());
}

/**
 * A recursive helper routine for displayRuntimeTree().
 * It draws the passed row, and any of its direct descendants that
 * are within open nodes.  It then continues to draw nodes that
 * are visibly below it, but are not direct descendants.
 *
 * @param gc The Graphics Context to use.  If gc is null, the tree
 * is traversed for visible rows in the Grid Area.
 * @param firstRowNode The first row to draw.
 *
 * @return False if there is no need to draw anymore rows else true
 */
protected boolean displayThisRowAndBelow(Graphics gc, RowNode node) {
    boolean hasUncle = true;
    while (node != null && node.getParent() != null) {
        if (hasUncle) {
            if (!displayRowChildren(gc, node)) {
                return false;
            }
        }
        RowNode next = (RowNode)getNextUncle(node);
        hasUncle = (next != null);
        node = (hasUncle) ? next : (RowNode)node.getParent();
    }
    return true;
}

/**
 * A recursive helper routine for displayThisRowAndBelow().
 * It draws the passed row, and any of its direct descendants that
 * are within open nodes.
 *
 * @param gc The Graphics Context to use.
 * @param firstRowNode The first row to draw.
 *
 * @return False if there is no need to draw anymore rows else true
 */
protected boolean displayRowChildren(Graphics gc, RowNode firstNode) {
    if (firstNode == null) {
        return true;
    }
    TreeIterator ti = ((DataSourceTreeNode)firstNode.getParent()).getIterator(firstNode);
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (!rowNode.draw(gc, grid)) {
            return false;
        }
        if (rowNode.getState() == NodeStatus.FOLDER_OPEN) {
            if (!displayRowChildren(gc, (RowNode)rowNode.getFirstChild())) {
                return false;
            }
        }
        ti.nextElement();
    }
    return true;
}

/**
 * A helper routine that returns the "uncle" (or "aunt") of
 * the passed row.
 *
 * @param rootObject The row whose uncle is desired..
 */
protected RowNode getNextUncle(RowNode rootObject) {
    RowNode parent = (RowNode)rootObject.getParent();
    if (parent != null) {
        return(RowNode)parent.getNextSibling();
    }
    return null;
}

/**
 * A helper routine that returns the "uncle" (or "aunt") of
 * the passed row.
 *
 * @param rootObject The row whose uncle is desired..
 */
protected RowNode getPreviousUncle(RowNode rootObject) {
    RowNode parent = (RowNode)rootObject.getParent();
    if (parent != null) {
        return(RowNode)parent.getPreviousSibling();
    }
    return null;
}

/**
 * Open the current RowNode folder.
 *
 * @param rowNode The record row to be opened.
 */
protected void openFolder(RowNode rowNode) {
    // sanity check
    if ((rowNode == null) || (rowNode.getState() == NodeStatus.FOLDER_OPEN)) {
        return;
    }

    // toggle the state to open
    rowNode.setState(NodeStatus.FOLDER_OPEN);

    // check for children
    if (! rowNode.hasChildren()) {
        // this might be the first time, so try to create some children
        try {
            createRuntimeChildren(rowNode);
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
    }
    // still no children
    if (! rowNode.hasChildren()) {
        // draw a blank icon
        rowNode.setState(NodeStatus.FOLDER_NONE);
    }
    else {
        // recursively add the number of scrollable rows
        addTotalNumberOfScrollableRows(getNumberOfScrollableRows(rowNode, true));
        // add possible repeat header for next record
        addRepeatHeader(rowNode);
    }
    updateGridArea();
}

/**
 * Close the current RowNode folder.
 *
 * @param rowNode The record row to be closed.
 */
protected void closeFolder(RowNode rowNode) {
    closeFolder(rowNode, true);
}

/**
 * Close the current RowNode folder.
 *
 * @param rowNode The record row to be closed.
 * @param updateRowCount true if numberOfScrollableRows should be updated.
 */
protected void closeFolder(RowNode rowNode, boolean updateRowCount) {
    // sanity check
    if ((rowNode == null) || (rowNode.getState() != NodeStatus.FOLDER_OPEN)) {
        return;
    }

    // toggle the state to closed
    rowNode.setState(NodeStatus.FOLDER_CLOSED);

    if (updateRowCount) {
        // recursively subtract the number of scrollable rows
        subtractTotalNumberOfScrollableRows(getNumberOfScrollableRows(rowNode, false));
    }

    // remove possible repeat header for next record
    removeRepeatHeader(rowNode);

    // remove its children if no caching
    if (!rowNode.getFormatNode().getRecordFormat().getMetaData().getCacheChildren()) {
        Vector uniqueChildren = getDataTableChildren(rowNode);
        try {
            for (Enumeration e = uniqueChildren.elements(); e.hasMoreElements();) {
                DataTableModel table = (DataTableModel) e.nextElement();
                table.deleteCacheChildren(grid.getDataModelListener());
            }
        }
        catch (Exception e) {
            grid.fireHiGridErrorEvent(new HiGridErrorEvent(grid, e));
        }
        rowNode.removeAllChildren();
    }

    updateGridArea();
}

protected static void addUnique(Vector v, Object object) {
    if (!v.contains(object)) {
        v.addElement(object);
    }
}

/**
 * Creates a list of DataTables that exist as immediate children of this row
 *
 * @param rowNode the parent rowNode.
 *
 * @return The Vector of unique DataTable children.
 */
protected Vector getDataTableChildren(RowNode rowNode) {
    Vector uniqueChildren = new Vector();
    TreeIterator ti = rowNode.getIterator();
    while (ti.hasMoreElements()) {
        RowNode node = (RowNode)ti.get();
        if (node.getDataTableModel() != null) {
            addUnique(uniqueChildren, node.getDataTableModel());
        }
        ti.nextElement();
    }
    return uniqueChildren;
}

/**
 * Converts an index to a RowNode
 *
 * @param index The index to convert.
 *
 * @return The RowNode in the index'th position in the RowTree.
 */
protected RowNode getRowNode(int index) {
    RowNodeWalk walkable = new RowNodeWalk(index);
    walk((RowNode)getRootNode().getFirstChild(), walkable);
    return walkable.getFoundNode();
}

/**
 * Converts a RowNode to its index in the RowTree
 *
 * @param rowNode The row to convert.
 *
 * @return The index of the RowNode in the RowTree.
 */
protected int getRowIndex(RowNode rowNode) {
    RowValueWalk walkable = new RowValueWalk(rowNode);
    walk((RowNode)getRootNode().getFirstChild(), walkable);
    return walkable.getCount();
}

/**
 * Retrieves the topmost visible RowNode.
 *
 * @return The topmost visible RowNode.
 */
protected RowNode getTopRowNode() {
    return topRowNode;
}

/**
 * Sets the topmost visible RowNode.
 *
 * @param rowNode The new topmost visible RowNode.
 */
protected void setTopRowNode(RowNode rowNode) {
    // don't set this to null during runtime, while a RowTree exists
    topRowNode = rowNode;
}

/**
 * Sets the topmost visible row in the higrid based
 * on the given row node.
 *
 * @param rowNode The row node.
 */
protected void setTopRow(RowNode rowNode) {
    if (getRoot() == null) return;

    RowValueWalk walkable = new RowValueWalk(rowNode);
    walk((RowNode)getRootNode().getFirstChild(), walkable);
    setTopRowNode(rowNode);
    int value = updateGridArea(walkable.getCount(), rowNode);
    grid.getVerticalScrollbar().setValue(value);
}

/**
 * Sets the topmost visible row in the higrid based
 * on the passed value and sets the vertical scrollbar position
 * to this value.
 *
 * @param value The new vertical scrollbar position.
 */
protected void setTopRow(int value) {
    if (getRoot() == null) return;

    RowNodeWalk walkable = new RowNodeWalk(value);
    walk((RowNode)getRootNode().getFirstChild(), walkable);
    setTopRowNode(walkable.getFoundNode());
    int newValue = updateGridArea(value, walkable.getFoundNode());
    grid.getVerticalScrollbar().setValue(newValue);
}

/**
 * Find the row node of the topmost visible row when the
 * row at the bottom of the window is scrolled into view.
 *
 * @param row The row index of the row node at the bottom of the window.
 *
 * @return The new top row node.
 */
protected RowNode getTopRowNodeFromLastFullyVisibleRow(RowNode lastVisibleRow) {
    LastVisibleRowWalk walkable = new LastVisibleRowWalk(grid.getGridArea().getSize().height - lastVisibleRow.getHeight());
    walk(lastVisibleRow, walkable);
    return walkable.getFoundNode();
}

/**
 * Does a quick search for the first record format node.
 *
 * @return The first record format node.
 */
protected RowNode getFirstRecordRowNode() {
    if (getRoot() == null) {
        return null;
    }
    RowNode first = (RowNode)getRootNode().getFirstChild();
    if (first == null) {
        return null;
    }
    if (first.getRowFormat() instanceof RecordFormat) {
        return first;
    }
    // now check the next sibling
    first = (RowNode)first.getNextSibling();
    if (first == null) {
        return null;
    }
    if (first.getRowFormat() instanceof RecordFormat) {
        return first;
    }
    return null;
}


/**
 * Find the row index of the topmost visible row on a page up
 * and update the grid area.
 *
 * @param row The row index of the current top row node
 *
 * @return The row index of new page up row node.
 */
protected int doPageUp(int row) {
    RowNode rowNode = getTopRowNode();
    PageUpWalk walkable = new PageUpWalk(row, grid.getGridArea().getSize().height, rowNode);
    walk(rowNode, walkable);
    int value = updateGridArea(walkable.getRowIndex(), walkable.getFoundNode());
    return value;
}

/**
 * Find the row index of the topmost visible row on a page down
 * and update the grid area.
 *
 * @param row The row index of the current top row node
 *
 * @return The row index of new page down row node.
 */
protected int doPageDown(int row) {
    RowNode rowNode = getTopRowNode();

    if (row > lastPageTopRow) {
        row = lastPageTopRow;
        rowNode = lastPageTopRowNode;
    }
    else {
        PageDownWalk walkable = new PageDownWalk(row, grid.getGridArea().getSize().height);
        walk(rowNode, walkable);
        RowWalk rowWalkable = new RowWalk(1, true);
        walk(walkable.getFoundNode(), rowWalkable);
        rowNode = rowWalkable.getFoundNode();
        row = walkable.getRowIndex() + 1;
    }

    int value = updateGridArea(row, rowNode);
    return value;
}

/**
 * Find the row index and the row node of the topmost visible row for
 * the last page.
 */
protected void findLastPageTopRow() {
    // go to the last row
    LastRowWalk lrWalkable = new LastRowWalk();
    RowNode row = ((RowNode)grid.getRowTree().getRootNode().getLastChild());
    walk(row, lrWalkable);
    // now go back up a page with the last row as the last row in grid area
    // assume this method is invoked after number of scrollable rows is
    // recalculated
    int rowIndex = getTotalNumberOfScrollableRows() - 1;
    RowNode rowNode = lrWalkable.getFoundNode();
    if (rowNode != null) {
        PageUpWalk puWalkable = new PageUpWalk(rowIndex, grid.getGridArea().getSize().height - lrWalkable.getFoundNode().getHeight(), rowNode);
        walk(rowNode, puWalkable);
        rowIndex = puWalkable.getRowIndex();
        rowNode = puWalkable.getFoundNode();
    }
    lastPageTopRow = rowIndex;
    lastPageTopRowNode = rowNode;
}

/*
 * Get a list of format nodes for a given row node.
 * This list is for the current row node and one level below.
 *
 * @param rowNode The row node for the current level.
 *
 * @return The list of format nodes.
 */
protected Vector getNodeLevelList(RowNode rowNode) {
    Vector list = new Vector();
    FormatNode node = null;
    if (rowNode == null) {
        node = (FormatNode)grid.getFormatTree().getRoot();
        if (node != null) {
            list.addElement(node);
        }
        return list;
    }
    list.addElement(rowNode.getFormatNode());
    TreeIterator ti = rowNode.getFormatNode().getIterator();
    while (ti.hasMoreElements()) {
        node = (FormatNode)ti.get();
        list.addElement(node);
        ti.nextElement();
    }

    return list;
}

/**
 * Retrieves the DataTableModel for a given format node
 * below a given row node (one level deep only).
 *
 * @param formatNode The given format node.
 * @param rootObject The parent row node.
 *
 * @return The current DataTableModel.
 */
protected DataTableModel getDataTableModel(FormatNode formatNode, RowNode rootObject) {
    DataTableModel dtm = null;
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getFormatNode() == formatNode &&
            rowNode.getRowFormat() instanceof RecordFormat) {
            dtm = rowNode.getDataTableModel();
            break;
        }
        ti.nextElement();
    }
    return dtm;
}

////////////////////////////////////////////////////////////////////////////////////////

/*
 * Find the record row node for the specified bookmark.
 *
 * @param rootObject The possible starting row node
 * @param bookmark The bookmark of the row node to find
 *
 * @return The record row node found else null.
 */
public RowNode findRecordRowNode(RowNode rootObject, long bookmark) {
    if (bookmark == DataTableModel.ROW_NOT_FOUND) {
        return null;
    }
    if (rootObject == null) {
        rootObject = (RowNode)getRoot();
    }
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getRowFormat() instanceof RecordFormat) {
            if (bookmark == rowNode.getBookmark()) {
                return rowNode;
            }
            RowNode resultRowNode = findRecordRowNode(rowNode, bookmark);
            if (resultRowNode != null) {
                return resultRowNode;
            }
        }
        ti.nextElement();
    }
    return null;
}

/*
 * Find a record row node for the same format node for a given parent row node.
 *
 * @param rootObject The parent row node
 * @param formatNode The format node of the interested record row node
 *
 * @return The record row node found else null.
 */
public RowNode findRecordRowNode(RowNode rootObject, FormatNode formatNode) {
    if (rootObject == null) {
        rootObject = (RowNode)getRoot();
    }
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getFormatNode() == formatNode) {
            if (rowNode.getRowFormat() instanceof RecordFormat) {
                return rowNode;
            }
        }
        ti.nextElement();
    }
    return null;
}

/*
 * Find a header row node for the same format node for a given parent row node.
 *
 * @param rootObject The parent row node
 * @param formatNode The format node of the interested record row node
 *
 * @return The header row node found else null.
 */
public RowNode findHeaderRowNode(RowNode rootObject, FormatNode formatNode) {
    if (rootObject == null) {
        rootObject = (RowNode)getRoot();
    }
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getFormatNode() == formatNode) {
            if (rowNode.getRowFormat() instanceof HeaderFormat) {
                return rowNode;
            }
        }
        ti.nextElement();
    }
    return null;
}

/*
 * Find a row node for the same format node for a given parent row node.
 *
 * @param rootObject The parent row node
 * @param formatNode The format node of the interested record row node
 * @param index the type of RowFormat to search. Constants are defined in FormatNode.
 *
 * @return The header row node found else null.
 */
public RowNode findRowNode(RowNode rootObject, FormatNode formatNode, int index) {
    if (rootObject == null) {
        rootObject = (RowNode)getRoot();
    }
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getFormatNode() == formatNode) {
            if (rowNode.getRowFormat() instanceof HeaderFormat) {
                return rowNode;
            }
        }
        ti.nextElement();
    }
    return null;
}

protected boolean parentFolderClosed(DataTableModel table) {
    RowNode rowNode = findFirstRowNode(null, table);
    if (rowNode == null) {
        return true;
    }
    RowNode rootObject = (RowNode)getRoot();
    RowNode parent = (RowNode)rowNode.getParent();
    while (parent != null && parent != rootObject) {
        if (parent.getState() == NodeStatus.FOLDER_CLOSED) {
            return true;
        }
        parent = (RowNode)parent.getParent();
    }
    return false;
}

/*
 * Find the first row node of a given data table model.
 *
 * @param rootObject The possible starting row node
 * @param dtm The data table model of the row node to find
 *
 * @return The row node found.
 */
public RowNode findFirstRowNode(RowNode rootObject, DataTableModel dtm) {
    if (rootObject == null) {
        rootObject = (RowNode)getRoot();
    }
    TreeIterator ti = rootObject.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (dtm == rowNode.getDataTableModel()) {
            return rowNode;
        }
        RowNode nextNode = findFirstRowNode(rowNode, dtm);
        if (nextNode != null) {
            return nextNode;
        }
        ti.nextElement();
    }
    return null;
}

/*
 * Find the previous record row node for the same format tree branch
 * matching possibly a given format node.
 *
 * @param rowNode The current row node
 * @param formatNode The current format node
 *
 * @return The previous record row node if found else null
 */
protected RowNode findPreviousRecordRowNode(RowNode rowNode, FormatNode formatNode) {
    RowNode rnParent = (RowNode) rowNode.getParent();
    RowNode rn = (RowNode) rnParent.getPreviousChild(rowNode);
    while (rn != null) {
        if (formatNode != null && rn.getFormatNode() != formatNode) {
            return null;
        }
        if (rn.getRowFormat() instanceof RecordFormat) {
            return rn;
        }
        rn = (RowNode) rnParent.getPreviousChild(rn);
    }
    return null;
}

/*
 * Find the next record row node for the same format tree branch
 * matching possibly a given format node.
 *
 * @param rowNode The current row node
 * @param formatNode The current format node
 *
 * @return The next record row node if found else null
 */
protected RowNode findNextRecordRowNode(RowNode rowNode, FormatNode formatNode) {
    RowNode rnParent = (RowNode) rowNode.getParent();
    RowNode rn = (RowNode) rnParent.getNextChild(rowNode);
    while (rn != null) {
        if (formatNode != null && rn.getFormatNode() != formatNode) {
            return null;
        }
        if (rn.getRowFormat() instanceof RecordFormat) {
            return rn;
        }
        rn = (RowNode) rnParent.getNextChild(rn);
    }
    return null;
}

/*
 * Find the next row node for the same format tree branch
 * matching possibly a given format node.
 *
 * @param rowNode The current row node
 * @param formatNode The current format node
 *
 * @return The next row node if found else null
 */
protected RowNode findNextNode(RowNode rowNode, FormatNode formatNode) {
    RowNode rn = (RowNode) rowNode.getNextSibling();
    if (rn != null) {
        if (formatNode != null && rn.getFormatNode() != formatNode) {
            return null;
        }
    }
    return rn;
}

/*
 * Find the row node to insert before for the row node with
 * the given format node.
 *
 * @param rootObject The parent row node
 * @param formatNode The format node of the row node to insert
 *
 * @return The before row node if found else null
 */
protected RowNode findInsertBeforeNode(RowNode rootObject, FormatNode formatNode) {
    RowNode rn = (RowNode) rootObject.getLastChild();
    RowNode rnBefore = null;
    while (rn != null) {
        if (rn.getFormatNode() == formatNode &&
            (rn.getRowFormat() instanceof RecordFormat ||
             rn.getRowFormat() instanceof HeaderFormat)) {
            break;
        }
        rnBefore = rn;
        rn = (RowNode) rootObject.getPreviousChild(rnBefore);
    }
    return rnBefore;
}

/*
 * Find the row node to insert before the row node(s)
 * of a given format node.
 *
 * @param rootObject The parent row node
 * @param formatNode The format node of the row node(s) after the insert point
 *
 * @return The before row node if found else null
 */
protected RowNode findInsertBeforeFormatNode(RowNode rootObject, FormatNode formatNode) {
    RowNode rn = (RowNode) rootObject.getLastChild();
    RowNode rnBefore = null;
    // assume the formatNode is present
    while (rn != null) {
        if (rn.getFormatNode() == formatNode) {
            rnBefore = rn;
        }
        else if (rnBefore != null) {
            break;
        }
        rn = (RowNode) rootObject.getPreviousChild(rn);
    }
    return rnBefore;
}

////////////////////////////////////////////////////////////////////////////////////////

// The following "walk" routines are mainly used for positioning
// the JCCellEditor via the keyboard
public void walk(RowNode node, Walkable walkable) {
    if (walkable.getDownward()) {
        walkDown(node, walkable);
    }
    else {
        walkUp(node, walkable);
    }
}

private void walkDown(RowNode node, Walkable walkable) {
    // we'll count the start node as one of the nodes traversed
    walkable.previous(node);
    boolean hasUncle = true;
    while (node != null && node.getParent() != null) {
        if (hasUncle) {
            walkDownMore(node, walkable);
            if (walkable.done()) {
                return;
            }
        }
        RowNode next = (RowNode)getNextUncle(node);
        hasUncle = (next != null);
        node = (hasUncle) ? next : (RowNode)node.getParent();
    }
}

private void walkDownMore(RowNode firstNode, Walkable walkable) {
    if (firstNode == null) {
        return;
    }
    TreeIterator ti = ((DataSourceTreeNode)firstNode.getParent()).getIterator(firstNode);
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        walkable.next(rowNode);
        if (walkable.find(rowNode)) {
            return;
        }
        if (rowNode.getState() == NodeStatus.FOLDER_OPEN) {
            walkDownMore((RowNode)rowNode.getFirstChild(), walkable);
            if (walkable.done()) {
                return;
            }
        }
        ti.nextElement();
    }
}

private void walkUp(RowNode node, Walkable walkable) {
    while (node != null && node.getParent() != null) {
        walkUpMore((RowNode)node.getPreviousSibling(), walkable);
        if (walkable.done()) {
            return;
        }
        RowNode next = (RowNode)node.getParent();
        if (next.getParent() == null) {
            return;
        }
        walkable.next(next);
        if (walkable.find(next)) {
            return;
        }
        node = (next != null) ? next : (RowNode)node.getParent();
    }
}

private void walkUpMore(RowNode firstNode, Walkable walkable) {
    if (firstNode == null) {
        return;
    }
    TreeIterator ti = ((DataSourceTreeNode)firstNode.getParent()).getIterator(firstNode, false);
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode.getState() == NodeStatus.FOLDER_OPEN) {
            walkUpMore((RowNode)rowNode.getLastChild(), walkable);
            if (walkable.done()) {
                return;
            }
        }
        walkable.next(rowNode);
        if (walkable.find(rowNode)) {
            return;
        }
        ti.nextElement();
    }
}

}
