/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SortGrid.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.io.Serializable;
import java.util.Enumeration;

import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;




import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * Sort the grid rows by specified columns.
 */
public class SortGrid implements Serializable {

static final long serialVersionUID = -7147369157833545298L;

public static final int UNKNOWN     = 0;
public static final int ASCENDING   = 1;
public static final int DESCENDING  = 2;

protected HiGrid grid = null;
protected boolean useDefaultSortData = false;

/**
 * Constructor
 * @param grid the HiGrid
 * @param useDefaultSortData true if the default SortData is to be used
 */
public SortGrid(HiGrid grid, boolean useDefaultSortData) {
    this.grid = grid;
    this.useDefaultSortData = useDefaultSortData;
}

/**
 * Returns the current Comparator for the specified column.
 * @param node The format node.
 * @param columnName The column name.
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this field's or method's return type
 * is based on the corresponding <code>com.sun.java.util.collections.Comparator</code> class.
 */
public Comparator getSortable(FormatNode node, String columnName) {
    Comparator sortable = grid.getSortable();
    if (node == null || columnName == null || columnName.length() == 0) {
        return sortable;
    }
    RowFormat rowFormat = (RowFormat) node.getRecordFormat();
    for (Enumeration e = rowFormat.getDataFormats().elements(); e.hasMoreElements();) {
        CellFormat cellFormat = (CellFormat) e.nextElement();
        if (columnName.compareTo(cellFormat.getName()) == 0) {
            if (cellFormat.getSortable() != null) {
                sortable = cellFormat.getSortable();
            }
            break;
        }
    }
    return sortable;
}

protected void setInternalData(DataTableModel dataModel, SortData sortData) {
    HiGridInternalData internalData = (HiGridInternalData)dataModel.getInternalData(grid);
    if (internalData == null) {
        internalData = new HiGridInternalData();
        dataModel.setInternalData(grid, internalData);
    }
    internalData.setSortData(new SortData(sortData));
}

/**
 * Perform the sorting for a column.
 * Assume using a default sort data and only a single data table is sorted.
 *
 * @param node The format node.
 * @param rowNode The row node within the data table.
 */
protected void sortColumn(FormatNode node, RowNode rowNode) {
    if (grid == null || !useDefaultSortData) {
        return;
    }

    SortData sortData = node.getDefaultSortData();
    if (sortData == null) {
        return;
    }
    RowNode parentRow = (RowNode)rowNode.getParent();
    RowTree rowTree = grid.getRowTree();
    if (!sortDataTable(rowTree, parentRow, node,
                  sortData.columnName, sortData.direction,
                  getSortable(node, sortData.columnName))) {
        sortData.setDirection(sortData.getPreviousDirection());
    }
    setInternalData(rowNode.getDataTableModel(), sortData);
}

/**
 * Perform the sorting for a column.
 *
 * @param cell The cell position.
 * @param singleTableSort Indication to sort a data table only.
 */
public void sortColumn(CellPosition cell, boolean singleTableSort) {
    if (grid == null) {
        return;
    }

    FormatNode node = cell.row.getFormatNode();
    SortData sortData = null;
    if (useDefaultSortData) {
        // case where it is not a singleTableSort
        setDefaultSortData(node, cell.format.getName());
        sortData = node.getDefaultSortData();
    }
    else {
        sortData = getDataTableSortData(node, cell);
    }
    Comparator sortable = getSortable(node, sortData.columnName);

    RowTree rowTree = grid.getRowTree();
    if (singleTableSort) {
        // sort the one required table
        if (!sortDataTable(rowTree, (RowNode) cell.row.getParent(), node,
                      sortData.columnName, sortData.direction, sortable)) {
            sortData.setDirection(sortData.getPreviousDirection());
        }
        setInternalData(cell.getRowNode().getDataTableModel(), sortData);
    }
    else {
        // sort all open tables in the RowTree that match the FormatNode of the cell's row
        SortGridWalk sortGridWalk = new SortGridWalk(node);
        RowNode topRowNode = (RowNode)(rowTree.getRootNode().getFirstChild());
        RowNode parentRow = null;
        while (topRowNode != null) {
            sortGridWalk.init();
            rowTree.walk(topRowNode, sortGridWalk);
            topRowNode = sortGridWalk.getLastRowNode();
            if (topRowNode != null) {
                parentRow = (RowNode) topRowNode.getParent();
                if (!sortDataTable(rowTree, parentRow, node,
                              sortData.columnName, sortData.direction, sortable)) {
                    sortData.setDirection(sortData.getPreviousDirection());
                }
                setInternalData(topRowNode.getDataTableModel(), sortData);
            }
        }
    }
}

/**
 * Perform the sorting for a column given the sortData.
 *
 * @param node The format node.
 * @param rowNode The row node within the data table.
 * @param sortData The sortData.
 */
protected void sortColumn(FormatNode node, RowNode rowNode, SortData sortData) {
    if (grid == null) {
        return;
    }
    RowNode parentRow = (RowNode)rowNode.getParent();
    if (!sortDataTable(grid.getRowTree(), parentRow, node,
                  sortData.columnName, sortData.direction,
                  getSortable(node, sortData.columnName))) {
        sortData.setDirection(sortData.getPreviousDirection());
    }
    setInternalData(rowNode.getDataTableModel(), sortData);
}

/**
 * Set the default sort data based on what column is to be sorted.
 *
 * @param node The format node.
 * @param columnName The name of the selected column.
 */
static public void setDefaultSortData(FormatNode node, String columnName) {
    SortData defaultSortData = node.getDefaultSortData();
    if (defaultSortData == null) {
        // new default sort data
        defaultSortData = new SortData(columnName);
        node.setDefaultSortData(defaultSortData);
        return;
    }

    if (defaultSortData.columnName.compareTo(columnName) == 0) {
        // change direction for default sort data
        defaultSortData.setNextDirection();
    }
    else {
        // change column for default sort data
        defaultSortData.columnName = columnName;
        defaultSortData.direction = SortGrid.ASCENDING;
    }
}

/**
 * Get the sort data for a single data table based on what
 * is on the sort list and what column is to be sorted.
 *
 * @param node The format node.
 * @param cell The cell of the selected column.
 *
 * @return The sort data.
 */
public SortData getDataTableSortData(FormatNode node, CellPosition cell) {
    SortData sortData = null;
    DataTableModel dataTable = cell.getRowNode().getDataTableModel();
    if (dataTable != null) {
        HiGridInternalData internalData = (HiGridInternalData)dataTable.getInternalData(grid);
        if (internalData == null) {
            internalData = new HiGridInternalData();
            dataTable.setInternalData(grid, internalData);
        }
        sortData = internalData.getSortData();
        if ((sortData != null) && (sortData.getColumnName().compareTo(cell.getCellFormat().getName()) == 0)) {
            sortData.setNextDirection();
        }
        else {
            // choose ASCENDING as default, since the user has clicked on it for the first time
            sortData = new SortData(cell.format.getName(), SortGrid.ASCENDING);
            internalData.setSortData(sortData);
        }
    }

    return sortData;
}

/**
 * Sorts a given data table for a single column and direction.
 *
 * @param rowTree The row tree.
 * @param parentNode The parent row node.
 * @param formatNode The format node.
 * @param column The column id.
 * @param direction The sorting direction.
 * @param sortable The sortable interface.
 * @return true if sort was not cancelled
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this method's parameter must be
 * a <code>com.sun.java.util.collections.Comparator</code>.
 */
public boolean sortDataTable(RowTree rowTree, RowNode parentNode, FormatNode formatNode,
                          String column, int direction, Comparator sortable) {
    String columns[] = {
        column
    };
    int directions[] = {
        direction
    };
    return sortDataTable(rowTree, parentNode, formatNode, columns, directions, sortable);
}

/**
 * Sorts a given data table by multiple columns.
 *
 * @param rowTree The row tree.
 * @param parentNode The parent row node.
 * @param formatNode The format node.
 * @param columns Array of column ids.
 * @param directions Array of sorting direction.
 * @param sortable The sortable interface.
 * @return true if sort was not cancelled
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this method's parameter must be
 * a <code>com.sun.java.util.collections.Comparator</code>.
 */
public boolean sortDataTable(RowTree rowTree, RowNode parentNode, FormatNode formatNode,
                          String columns[], int directions[], Comparator sortable) {
    if (sortable == null) {
        sortable = grid.getSortable();
        if (sortable == null) {
            return false;
        }
    }

    Vector rowData = setupRowData(rowTree, parentNode, formatNode, columns);
    if (rowData == null || rowData.size() == 0) {
        return false;
    }

    RowNode rowNode = ((SortRowData)rowData.elementAt(0)).rowNode;
    if (grid.fireHiGridSortTableEvent(new HiGridSortTableEvent(grid,
                                      HiGridSortTableEvent.BEFORE_SORT_TABLE,
                                      rowNode, columns[0]))) {

        Collections.sort(rowData, new ColumnSortable(sortable, directions));

        resetRowData(rowTree, parentNode, formatNode, rowData);
        addRepeatHeaders(rowTree, parentNode, formatNode);

		//  Update the direction in the FormatNode's default SortData object.
		SortData sortData = formatNode.getDefaultSortData();
		if (sortData != null) {
			sortData.setDirection(directions[0]);
		}

        grid.fireHiGridSortTableEvent(new HiGridSortTableEvent(grid,
                                      HiGridSortTableEvent.AFTER_SORT_TABLE,
                                      rowNode, columns[0]));
        return true;
    }
    return false;
}

/**
 * Setup a Vector of row data for a given parent and
 * format node.  For performance reasons, possible
 * repeat headers are removed as well.
 *
 * @param rowTree The row tree.
 * @param parentNode The parent row node.
 * @param formatNode The format node.
 * @param columns Array of column ids.
 *
 * @return The array of row data.
 */
protected Vector setupRowData(RowTree rowTree, RowNode parentNode, FormatNode formatNode, String columns[]) {
    // get a count of number of record rows
    int count = 0;
    TreeIterator ti = parentNode.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode != null && rowNode.getRowFormat() instanceof RecordFormat &&
            rowNode.getFormatNode() == formatNode) {
            rowTree.removeRepeatHeader(rowNode);
            count++;
        }
        ti.nextElement();
    }
    if (count == 0) {
        return null;
    }
    Vector rowData = new Vector(count);

    // put row node into the list
    int index = 0;
    ti = parentNode.getIterator();
    while (ti.hasMoreElements() && index < count) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode != null && rowNode.getRowFormat() instanceof RecordFormat &&
            rowNode.getFormatNode() == formatNode) {
            SortRowData sortRowData = new SortRowData(rowNode, columns.length);
            rowData.insertElementAt(sortRowData, index);
            DataTableModel dataTable = rowNode.getDataTableModel();
            if (dataTable != null) {
                for (int i = 0; i < columns.length; i++) {
                    try {
                        sortRowData.resultData[i] = dataTable.getResultData(rowNode.getBookmark(), columns[i]);
                    }
                    catch (Exception exception) {
                        sortRowData.resultData[i] = null;
                    }
                    if (sortRowData.resultData[i] == null) {
                        // null value found, probably due to uncommitted insert
                        // change it to an empty string for proper sorting
                        sortRowData.resultData[i] = "";
                    }
                }
            }
            index++;
        }
        ti.nextElement();
    }
    return rowData;
}

/**
 * Given a sorted list of row data, reset the row tree.
 *
 * @param rowTree The row tree.
 * @param parentNode The parent row node.
 * @param formatNode The format node.
 * @return rowData The list of sort row data.
 */
protected void resetRowData(RowTree rowTree, RowNode parentNode, FormatNode formatNode, Vector rowData) {
    // remove the previous set of row data
    for (Enumeration e = rowData.elements(); e.hasMoreElements();) {
        SortRowData sortRowData = (SortRowData) e.nextElement();
        parentNode.remove(sortRowData.rowNode);
    }
    RowNode insertBefore = rowTree.findInsertBeforeNode(parentNode, formatNode);
    for (Enumeration e = rowData.elements(); e.hasMoreElements();) {
        SortRowData sortRowData = (SortRowData) e.nextElement();
        if (insertBefore == null) {
            parentNode.add(sortRowData.rowNode);
        }
        else {
            parentNode.insert(insertBefore, sortRowData.rowNode);
        }
    }
}

/**
 * Add possible repeat headers in data table.
 *
 * @param rowTree The row tree.
 * @param parentNode The parent row node.
 * @param formatNode The format node.
 */
protected void addRepeatHeaders(RowTree rowTree, RowNode parentNode, FormatNode formatNode) {
    HeaderFormat headerFormat = formatNode.getHeaderFormat();
    if (!headerFormat.isShowing() || !headerFormat.isRepeatHeader()) {
        return;
    }
    TreeIterator ti = parentNode.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        if (rowNode != null && rowNode.getRowFormat() instanceof RecordFormat &&
            rowNode.getFormatNode() == formatNode &&
            rowNode.getState() == NodeStatus.FOLDER_OPEN &&
            rowNode.hasChildren()) {
            rowTree.addRepeatHeader(rowNode);
        }
        ti.nextElement();
    }
}

}
