/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SummaryColumn.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.io.Serializable;

import com.klg.jclass.datasource.*;

/**
 * The column object for the summary record types.
 */

public class SummaryColumn implements Serializable {

static final long serialVersionUID = -4348474341309558899L;

private int table;
private String name;
private int columnType;
private int dataType;
private int aggregateType;

private String columnIdentifier;

public final static int COLUMN_TYPE_UNKNOWN     = 0;
public final static int COLUMN_TYPE_LABEL       = 1;
public final static int COLUMN_TYPE_DATASOURCE  = 2;
public final static int COLUMN_TYPE_AGGREGATE   = 3;
public final static int COLUMN_TYPE_UNBOUND     = 4;

public final static int AGGREGATE_TYPE_NONE     = 0;
public final static int AGGREGATE_TYPE_COUNT    = 1;
public final static int AGGREGATE_TYPE_SUM      = 2;
public final static int AGGREGATE_TYPE_AVERAGE  = 3;
public final static int AGGREGATE_TYPE_MIN      = 4;
public final static int AGGREGATE_TYPE_MAX      = 5;
public final static int AGGREGATE_TYPE_FIRST    = 6;
public final static int AGGREGATE_TYPE_LAST     = 7;

/**
 * Constructors
 */
public SummaryColumn(String name) {
    // assume to be a static label
    init(0, name, COLUMN_TYPE_LABEL, AGGREGATE_TYPE_NONE, MetaDataModel.TYPE_STRING);
}

public SummaryColumn(MetaDataModel dataSourceMetaData, String name, int column) {
    int type = MetaDataModel.TYPE_STRING;
    try {
        type = dataSourceMetaData.getJavaColumnType(name);
    }
    catch (Exception e) {
        init(0, name, COLUMN_TYPE_LABEL, AGGREGATE_TYPE_NONE, MetaDataModel.TYPE_STRING);
        return;
    }

    if (column == COLUMN_TYPE_DATASOURCE) {
        // convenience constuctor for data source columns
        init(dataSourceMetaData.getMetaID(), name, column, AGGREGATE_TYPE_NONE, type);
    }
    else if (column == COLUMN_TYPE_AGGREGATE) {
        // aggregate type default to count for non numeric types
        // and sum for numeric types
        int aggregate = getDefaultAggregateType(type);
        init(dataSourceMetaData.getMetaID(), name, column, aggregate, type);
    }
    else {
        init(0, name, COLUMN_TYPE_LABEL, AGGREGATE_TYPE_NONE, MetaDataModel.TYPE_STRING);
    }
}

public SummaryColumn(MetaDataModel dataSourceMetaData, String name, int column, int aggregate) {
    if (column != COLUMN_TYPE_AGGREGATE) {
        init(0, name, COLUMN_TYPE_LABEL, AGGREGATE_TYPE_NONE, MetaDataModel.TYPE_STRING);
        return;
    }

    // convenience constuctor for aggregrate columns
    int type = MetaDataModel.TYPE_INTEGER;
    if (aggregate != AGGREGATE_TYPE_COUNT) {
        try {
            type = dataSourceMetaData.getJavaColumnType(name);
        }
        catch (Exception e) {
            init(0, name, COLUMN_TYPE_LABEL, AGGREGATE_TYPE_NONE, MetaDataModel.TYPE_STRING);
            return;
        }
    }
    init(dataSourceMetaData.getMetaID(), name, column, aggregate, type);
}

public SummaryColumn(MetaDataModel dataSourceMetaData, String name, int column, int aggregate, int type) {
    init(dataSourceMetaData.getMetaID(), name, column, aggregate, type);
}

public SummaryColumn(int table, String name, int column, int aggregate, int type) {
    init(table, name, column, aggregate, type);
}

/**
 * Initialize the summary column and redo the columnIdentifier.
 */
public void init(int table, String name, int column, int aggregate, int type) {
    this.name = name;
    this.table = table;
    columnType = column;
    aggregateType = aggregate;
    dataType = type;
    setIdentifier();
}

/**
 * Determine the default aggregate type based on the data type.
 */
public int getDefaultAggregateType(int type) {
    int defaultAggregate = AGGREGATE_TYPE_NONE;
    switch (type) {
    case MetaDataModel.TYPE_DOUBLE:
    case MetaDataModel.TYPE_FLOAT:
    case MetaDataModel.TYPE_INTEGER:
    case MetaDataModel.TYPE_BIG_DECIMAL:
    case MetaDataModel.TYPE_LONG:
    case MetaDataModel.TYPE_BYTE:
    case MetaDataModel.TYPE_SHORT:
        defaultAggregate = AGGREGATE_TYPE_SUM;
        break;
    default:
        defaultAggregate = AGGREGATE_TYPE_COUNT;
        break;
    }
    return defaultAggregate;
}

/**
 * Return a string identifier for this column.
 *
 * @return The column identifier.
 */
public String getIdentifier() {
    return columnIdentifier;
}

/**
 * Sets the string identifier for this column.
 */
public void setIdentifier() {
    // a combination of name, table, column type, and aggregrate type
    // but no quarantee that this is unique
    // Note: Must not be changed after this summary column is added to the
    //       SummaryMetaData.
    columnIdentifier = new String(name + "." + table + "." + columnType + "." + aggregateType);
}

/**
 * Returns the summary column name.
 *
 * @return The summary column name.
 */
public String getName() {
    return name;
}

/**
 * Sets the summary column name.
 *
 * @param name The summary column name.
 */
public void setName(String name) {
    this.name = new String(name);
}

/**
 * Returns the summary column table source.
 *
 * @return The summary column table source.
 */
public int getTable() {
    return table;
}

/**
 * Sets the summary column table source.
 *
 * @param The summary column table source.
 */
public void setTable(int table) {
    this.table = table;
}

/**
 * Returns the summary column type.
 *
 * @return The summary column type.
 */
public int getColumnType() {
    return columnType;
}

/**
 * Sets the summary column type.
 *
 * @param type The summary column type.
 */
public void setColumnType(int type) {
    columnType = type;
}

/**
 * Returns the summary column data type.
 *
 * @return The summary column data type.
 */
public int getDataType() {
    return dataType;
}

/**
 * Sets the summary column data type.
 *
 * @param type The summary column data type.
 */
public void setDataType(int type) {
    dataType = type;
}

/**
 * Returns the summary column aggregate type.
 * Applies to aggregated data source column type only.
 *
 * @return The summary column aggregate type.
 */
public int getAggregateType() {
    return aggregateType;
}

/**
 * Sets the summary column aggregate type.
 * Applies to aggregated data source column type only.
 *
 * @param type The summary column aggregate type.
 */
public void setAggregateType(int type) {
    aggregateType = type;
}

}
