/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SummaryDataTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;
import com.klg.jclass.util.JCEnvironment;

/**
 * The data table for a summary row type.
 */
public class SummaryDataTable extends DataSourceTreeNode implements
    DataTableModel {

static final long serialVersionUID = -3066187169744256335L;

private MetaDataModel metaData[];
private Vector row[];
private Hashtable identifiers[];
private int numColumns[];
protected Hashtable internalData;
protected int initialHashtableSize = 5;         // Sun's default is 101, which is too big

/**
 * SummaryDataTable Constructor.
 *
 * @param footerMetaData The Footer MetaData for this data table.
 * @param beforeDetailsMetaData The BeforeDetails MetaData for this data table.
 * @param afterDetailsMetaData The AfterDetails MetaData for this data table.
 */
public SummaryDataTable(MetaDataModel footerMetaData,
                        MetaDataModel beforeDetailsMetaData,
                        MetaDataModel afterDetailsMetaData) {
    super(null);

    metaData = new MetaDataModel[HiGridData.SUMMARY_ROWS];
    metaData[HiGridData.FOOTER] = footerMetaData;
    metaData[HiGridData.BEFORE_DETAILS] = beforeDetailsMetaData;
    metaData[HiGridData.AFTER_DETAILS] = afterDetailsMetaData;

    resetSummaryDataTables(true);
}

/**
 * Reset the summary data tables to remove all cached data.
 *
 * @param clearCachedData True if the cached data are to be removed.
 */
protected void resetSummaryDataTables(boolean clearCachedData) {
    if (clearCachedData) {
        row = new Vector[HiGridData.SUMMARY_ROWS];
        identifiers = new Hashtable[HiGridData.SUMMARY_ROWS];
        numColumns = new int[HiGridData.SUMMARY_ROWS];
        numColumns[HiGridData.FOOTER] = 0;
        numColumns[HiGridData.BEFORE_DETAILS] = 0;
        numColumns[HiGridData.AFTER_DETAILS] = 0;
    }
    else {
        resetSummaryRow(row[HiGridData.FOOTER]);
        resetSummaryRow(row[HiGridData.BEFORE_DETAILS]);
        resetSummaryRow(row[HiGridData.AFTER_DETAILS]);
    }
}

protected void resetSummaryRow(Vector summaryRow) {
    if (summaryRow == null) {
        return;
    }
    for (Enumeration e = summaryRow.elements(); e.hasMoreElements();) {
        SummaryData summaryData = (SummaryData) e.nextElement();
        if (summaryData != null) {
            summaryData.setUseCachedValue(false);
        }
    }
}

/**
 * Determine if the summary row data has changed in value.
 *
 * @param grid The HiGrid.
 * @param rowFormat The row format for this row.
 * @param dataTable The record type data table.
 * @param rowNode The row node of this row.
 * @param name The summary column identifier.
 *
 * @return True if a value has changed in the summary row.
 */
protected boolean isResultDataChanged(HiGrid grid, DataTableModel dataTable, RowFormat rowFormat, RowNode rowNode, String name) {
    return(getResultData(grid, dataTable, rowFormat, rowNode, name, true) != null);
}

/**
 * Return a value for given summary column identifier.
 *
 * @param grid The HiGrid.
 * @param rowFormat The row format for this row.
 * @param dataTable The record type data table.
 * @param rowNode The row node of this row.
 * @param name The summary column identifier.
 *
 * @return The Object in the specified cell.
 */
public Object getResultData(HiGrid grid, DataTableModel dataTable, RowFormat rowFormat, RowNode rowNode, String name) {
    return getResultData(grid, dataTable, rowFormat, rowNode, name, false);
}

/**
 * Helper method to support both isResultDataChanged() and getResultData().
 *
 * @param grid The HiGrid.
 * @param rowFormat The row format for this row.
 * @param dataTable The record type data table.
 * @param rowNode The row node of this row.
 * @param name The summary column identifier.
 * @param checkChanged True if result data changed is to be check.
 *
 * @return If checkChanged, a null object indicates no changes. If not checkChanged,
 *         returns the actual data object.
 */
protected Object getResultData(HiGrid grid, DataTableModel dataTable, RowFormat rowFormat, RowNode rowNode, String name, boolean checkChanged) {
    int rowIndex = getSummaryRow(rowFormat);
    SummaryMetaData summaryMetaData = (SummaryMetaData) metaData[rowIndex];
    SummaryColumn column = summaryMetaData.getColumn(name);

    boolean cacheValue = false;
    boolean useCachedValue = false;
    SummaryData summaryData = null;
    Object o = null;
    Aggregate aggregate = null;
    boolean dataChanged = false;
    try {
        switch (column.getColumnType()) {

        case SummaryColumn.COLUMN_TYPE_LABEL:
            // assume label is static
            o = (Object) column.getName();
            break;

        case SummaryColumn.COLUMN_TYPE_DATASOURCE:
            // check if cache value is to be used
            String id = getCachedDataID(rowIndex, name, rowNode);
            summaryData = getCachedData(rowIndex, id);
            if (summaryData != null) {
                useCachedValue = summaryData.getUseCachedValue();
            }
            if (useCachedValue) {
                // use cached value, no data changed
                o = summaryData.getValue();
                break;
            }
            // data source should be from current or one of the level above
            RowNode rnSource = getDataSourceRowNode(column.getTable(), dataTable, rowNode);
            if (rnSource == null) {
                break;
            }
            DataTableModel dataTableSource = rnSource.getDataTableModel();
            if (dataTableSource != null) {
                o = dataTableSource.getResultData(rnSource.getBookmark(), column.getName());
                if (summaryData == null || !summaryData.equals(o, column.getDataType())) {
                    dataChanged = true;
                    setCachedData(rowIndex, id, o);
                }
                else {
                    summaryData.setUseCachedValue(true);
                }
            }
            break;

        case SummaryColumn.COLUMN_TYPE_AGGREGATE:
            switch (column.getAggregateType()) {
            case SummaryColumn.AGGREGATE_TYPE_COUNT:
                aggregate = new AggregateCount(column);
                break;
            case SummaryColumn.AGGREGATE_TYPE_SUM:
                aggregate = new AggregateSum(column);
                break;
            case SummaryColumn.AGGREGATE_TYPE_AVERAGE:
                aggregate = new AggregateAverage(column);
                break;
            case SummaryColumn.AGGREGATE_TYPE_MIN:
                aggregate = new AggregateMin(column);
                break;
            case SummaryColumn.AGGREGATE_TYPE_MAX:
                aggregate = new AggregateMax(column);
                break;
            case SummaryColumn.AGGREGATE_TYPE_FIRST:
                aggregate = new AggregateFirst(column);
                break;
            case SummaryColumn.AGGREGATE_TYPE_LAST:
                aggregate = new AggregateLast(column);
                break;
            default:
                break;
            }
            // do the aggregate further down
            break;

        case SummaryColumn.COLUMN_TYPE_UNBOUND:
            // do the unbound data further down
            Class c = Class.forName(column.getName());
            if (c != null) {
                aggregate = (Aggregate) c.newInstance();
                if (aggregate != null) {
                    aggregate.init(column);
                }
            }
            break;

        default:
            break;
        }
        // do the aggregate/unbound data here
        if (aggregate != null) {
            // check if cache value is to be used
            String dataID = getCachedDataID(rowIndex, name, rowNode);
            summaryData = getCachedData(rowIndex, dataID);
            if (summaryData != null) {
                useCachedValue = summaryData.getUseCachedValue();
            }
            if (aggregate.cacheResultData() && useCachedValue) {
                // use cached value, no data changed
                o = summaryData.getValue();
            }
            if (o == null) {
                aggregateRowNode((RowNode) rowNode.getParent(), aggregate, grid);
                o = aggregate.getResultData();
                if (summaryData == null || !summaryData.equals(o, column.getDataType())) {
                    dataChanged = true;
                    setCachedData(rowIndex, dataID, o);
                }
                else {
                    summaryData.setUseCachedValue(true);
                }
            }
        }
    }
    catch (Exception e) {
        if (JCEnvironment.isDesignTime()) {
            grid.adviseClearAllFormats = true;
        }
        /* do nothing for runtime */
    }

    if (checkChanged && !dataChanged) {
        return null;
    }
    return o;
}

/**
 * Return the summary row index given the row format type.
 *
 * @param rowFormat The row format.
 *
 * @return The summary row index.
 */
protected int getSummaryRow(RowFormat rowFormat) {
    // row should have a value range of HiGridData.FOOTER
    // to HiGridData.AFTER_DETAILS
    if (rowFormat instanceof FooterFormat) {
        return HiGridData.FOOTER;
    }
    else if (rowFormat instanceof BeforeDetailsFormat) {
        return HiGridData.BEFORE_DETAILS;
    }
    return HiGridData.AFTER_DETAILS;
}

/**
 * Retrieves the cached data id.
 *
 * @param rowIndex The row index for the summary data table.
 * @param name The column identifier.
 * @param rowNode The summary row node.
 *
 * @return The cached data id.
 */
protected String getCachedDataID(int rowIndex, String name, RowNode rowNode) {
    if (rowIndex == HiGridData.FOOTER) {
        return name;
    }
    // BeforeDetails and AfterDetails need bookmark to be unique in data table
    return(new String(name+"."+((RowNode)rowNode.getParent()).getBookmark()));
}

/**
 * Retrieves the cached data for display.
 *
 * @param rowIndex The row index for the summary data table.
 * @param name The cached data identifier.
 *
 * @return The cached summary data.
 */
protected SummaryData getCachedData(int rowIndex, String name) {
    if (identifiers[rowIndex] == null) {
        return null;
    }
    Integer intValue  = ((Integer) identifiers[rowIndex].get(name));
    if (intValue != null) {
        return((SummaryData) row[rowIndex].elementAt(intValue.intValue()));
    }
    return null;
}

/**
 * Stores the cached data for display.
 *
 * @param rowIndex The row index for the summary data table.
 * @param name The cached data identifier.
 * @param object The data object.
 */
protected void setCachedData(int rowIndex, String name, Object object) {
    if (identifiers[rowIndex] == null) {
        identifiers[rowIndex] = new Hashtable();
        row[rowIndex] = new Vector();
    }
    Integer intValue  = ((Integer) identifiers[rowIndex].get(name));
    if (intValue == null) {
        row[rowIndex].addElement((Object) new SummaryData(object));
        identifiers[rowIndex].put(name, new Integer(numColumns[rowIndex]));
        numColumns[rowIndex]++;
    }
    else {
        row[rowIndex].setElementAt((Object) new SummaryData(object), intValue.intValue());
    }
}

/**
 * Retrieves the record row node with a data table for the given meta id.
 * This is for the non aggregated data source type.
 *
 * @param metaID The meta id of the meta table associated with the data table
 * @param dataTable The current level data table.
 * @param rowNode The row node of the current level.
 *
 * @return The row node found.
 */
protected RowNode getDataSourceRowNode(int metaID, DataTableModel dataTable, RowNode rowNode) {
    RowNode rnFound = null;

    // try the current level
    if (dataTable.getMetaData().getMetaID() == metaID) {
        rnFound = rowNode;
        if (!(rnFound.getRowFormat() instanceof RecordFormat)) {
            // the record row node is in the row tree level above
            rnFound = (RowNode) rnFound.getParent();
        }
        return rnFound;
    }

    // try the levels above
    RowNode rnCurrent = rowNode;
    while (rnCurrent != null) {
        rnCurrent = (RowNode) rowNode.getParent();
        if (rnCurrent != null) {
            if (rnCurrent.getDataTableModel().getMetaData().getMetaID() == metaID) {
                rnFound = rnCurrent;
                break;
            }
        }
    }

    return rnFound;
}

/**
 * Perform the aggregation as the data tables are traversed.
 *
 * @param rootNode The starting root row node.
 * @param aggregate The aggregate interface.
 */
protected void aggregateRowNode(RowNode rootNode, Aggregate aggregate, HiGrid grid) {
    // traverse each level, creating data table as necessary
    TreeIterator ti = rootNode.getIterator();
    while (ti.hasMoreElements()) {
        RowNode rowNode = (RowNode)ti.get();
        // for each record format row node:
        if (rowNode.getRowFormat() instanceof RecordFormat) {
            // perform the aggregation
            aggregate.calculate(rowNode);
            // traverse down next level if required
            if (aggregate.nextLevel(rowNode)) {
                if (rowNode.getState() == NodeStatus.FOLDER_OPEN) {
                    aggregateRowNode((RowNode)rowNode, aggregate, grid);
                }
                else {
                    // check for children
                    if (! rowNode.hasChildren()) {
                        // this might be the first time, so try to create some children
                        try {
                            grid.getRowTree().createRuntimeChildren(rowNode);
                        }
                        catch (Exception e) {
                            // this can be with a query for a newly inserted row
                            /* do nothing */
                        }
                    }
                    if (! rowNode.hasChildren()) {
                        // still no children
                        rowNode.setState(NodeStatus.FOLDER_NONE);
                    }
                    else {
                        aggregateRowNode(rowNode, aggregate, grid);
                    }
                }
            }
            if (aggregate.done()) {
                return;
            }
        }
        ti.nextElement();
    }
}

////////////////////////////////////////////////////////////////////
// DataTableModel interface implementation

// Note: The row indentifier (bookmark) is not unique and has a
//       value range of HiGridData.FOOTER to
//       HiGridData.AFTER_DETAILS.

/**
 * Return the number of rows in this DataTable.
 */
public int getRowCount() {
    return HiGridData.SUMMARY_ROWS;
}

/**
 * Return the status of a row, will be one of INSERTED, UPDATED,
 * DELETED, COMMITTED.
 * @param bookmark a long which uniquely identifies this row
 * @return an int row status, will be one of INSERTED, UPDATED,
 * DELETED, COMMITTED.
 */
public int getRowStatus(long bookmark) {
    // return something
    return DataTableModel.COMMITTED;
}

/**
 * Return an int which will uniquely identify this row.
 * Facilitates sorting: index changes, identifier remains constant.
 * @param index an int for the original index of this row
 * @return a unique long to identify this row from now on
 */
public long getRowIdentifier(int row) {
    // row should have a value range of HiGridData.FOOTER
    // to HiGridData.AFTER_DETAILS
    return new Long(row).longValue();
}

/**
 * Return a value for given row_identifier/column_identifier.
 * @param bookmark a long which uniquely identifies this row
 * @param column_identifier a String which uniquely identifies this
 * column
 * @return the Object in the specified cell
 * @exception DataModelException If getResultData() fails.
 */
public Object getResultData(long bookmark,String column_identifer) throws DataModelException {
    // not need
    return null;
}

/**
 * Create and return the DataTable which corresponds to the specifed
 * row of this parent for the indicated child MetaData object.
 * Always returns a non-null DataTable.
 * @see #getTable
 * @param bookmark a long which uniquely identifies a row
 * @param child the MetaDataModel object which will drive the next query
 * @return the child DataTableModel of results
 * @exception DataModelException If createTable() fails.
 */
public DataTableModel createTable(DataModelListener originator, long bookmark, TreeNode node) throws DataModelException {
    // not need
    return null;
}

/**
 * Search the existing children of this node for
 * a child node which matches the input criteria.
 * Return that node if it exists, null otherwise.
 * @see #createTable
 * @param bookmark a long which uniquely identifies a row
 * @param child the MetaDataModel object
 * @return the DataTable with matching bookmark and MetaDataModel
 * otherwise, null.
 */
public DataTableModel getTable(long bookmark, TreeNode node) {
    // not need
    return null;
}

/**
 * Return the MetaDataModel for this table.
 * @return the MetaDataModel Object for this table
 */
public MetaDataModel getMetaData() {
    // return the footer one
    return(MetaDataModel) metaData[HiGridData.FOOTER];
}

/**
 * Return the parent row for this table.
 * This is the immediate parent of this table.
 * @return the bookmark of the parent row for this DataTable
 * @see #getRootAncestorBookmark
 * @see #getAncestors
 */
public long getParentBookmark() {
    // not needed
    return(-1);
}

/**
 * Delete a row.
 * Logically deletes a row by setting its status to DELETED.  Will
 * be deleted permanently upon commit, or restored on rollback.
 * @param bookmark a unique long indicating the row to be deleted.
 * @exception DataModelException If logical delete fails.
 */
public void deleteRow(DataModelListener originator, long bookmark) throws DataModelException {
    // not needed
}

/**
 * Update a cell.
 * User has modified a cell. Call this method to inform this
 * data source of that change.  This method will then fire a
 * DataTableEvent to inform listeners about this change.
 * Internally a copy of the original row will be maintained which can
 * be used if an update is cancelled.
 * This row's status is set to UPDATED.
 * @param bookmark a unique long indicating in which row
 * the change was made
 * @param column_identifer a unique String indicating in which
 * column the change occurred.
 * @param value the Object to be placed in the cell
 * @exception DataModelException If error occurs updating a cell.
 */
public void updateCell(DataModelListener originator, long bookmark, String column_identifer, Object value) throws DataModelException {
    // not needed
}

/**
 * Commit row.
 * Write row (new and/or modified) to datasource.
 * @param bookmark a unique long indicating which row
 * to commit to the datasource.
 * @exception DataModelException If commit fails.
 */
public boolean commitRow(DataModelListener originator, long bookmark) throws DataModelException {
    return true; // not implemented
}

/**
 * Commit all uncommitted rows.
 * Call commitRow for each new and/or modified row in this DataTable.
 * @exception DataModelException If any row fails to commit.
 */
public void commitAll(DataModelListener originator) throws DataModelException {
    // not needed
}

/**
 * Recurse on this node and each of its children and commit
 * all changes in each if commitPolicy does not equal COMMIT_MANUALLY
 */
public void conditionallyCommitYourselfAndChildren(DataModelListener originator) {
    // not needed
}

/**
 * Cancel uncommitted row changes.
 * Restore original cell values for this row, or if a new row remove it.
 * @param bookmark a unique long indicating which row to uncommit.
 * @exception DataModelException If an exception occurs while canceling uncommitted rows.
 */
public void cancelRowChanges(DataModelListener originator, long bookmark) throws DataModelException {
    // not needed
}

/**
 * Cancel all uncommited rows.
 * Restore original cell values for each uncommited row, for new rows
 * remove them.  Calls cancelRow for each uncommited row.
 * @exception DataModelException If an exception occurs while canceling uncommitted rows.
 */
public void cancelAllRowChanges(DataModelListener originator) throws DataModelException {
    // not needed
}

/**
 * Rollback transaction.
 * Request rollback of the transaction surrounding these changes.
 * This method will be called by cancelRowChanges if the commit policy
 * is on a per row basis, or by cancelAllRowChanges if the policy is
 * on multiple rows or manual.
 * The transaction will only be rolled back if there are no other
 * pending changes in the transaction.
 * No explicit "begin" is required since the SQL Standard
 * implicitly starts a "begin" when any change is made.
 */
public void rollbackTransaction() {
    // not needed
}

/**
 * Recurse up the tree and determine the row number of
 * your ultimate ancestor.
 * @return a long indicating the bookmark of this
 * DataTable's ultimate ancestor.
 * @see #getAncestors
 */
public long getRootAncestorBookmark(long bookmark) {
    // not needed
    return(-1);
}

/**
 * Return a list of this DataTable's ancestor bookmarks.
 * The list is ordered from the root ancestor down to the node
 * above this one.
 * @param  ancestors the Vector to be populated with
 *         ancestor bookmarks
 * @return a Vector of ancestor bookmarks.
 */
public void getAncestors(Vector ancestors) {
    // not needed
}

/**
 * Requery a DataTable row and all children.
 * Calls requeryRow() then requeryLevel.
 * Refetches all rows in this DataTable from the
 * originating data source.  Then does the same for
 * each of this row's children.
 * @return this data table model
 * @exception DataModelException If requery from server fails.
 */
public void requeryRowAndDetails(DataModelListener originator, long bookmark) throws DataModelException {
    // not needed
}

/*
 * Re-read a row from the originating data source.
 * Calls refreshRow(bookmark) which is the implementation specific
 * method to refresh a row.
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If requery from server fails.
 */
public void requeryRow(DataModelListener originator, long bookmark) throws DataModelException {
    // not needed
}

/**
 * Requery a DataTable.
 * An implementation specific method for cloning rows,
 * must be implemented in the derived class of BaseDataTable.
 * Refetches all rows in this DataTable from the
 * originating data source.
 * @return this data table model
 */
public DataTableModel requeryLevel() {
    // not needed
    return null;
}

/**
 * Position this table's cursor on the first row.
 */
public boolean first() {
    // not needed
    return false;
}

/**
 * Position this table's cursor on the last row.
 */
public boolean last() {
    // not needed
    return false;
}

/**
 * Advance the cursor by one row.
 * Return true if there was a row to move forward to.
 * @return true if move to next row succeeded.
 */
public boolean next() {
    // not needed
    return false;
}

/**
 * Position this table's cursor to the previous row.
 * Return true if the move succeeds.
 */
public boolean previous() {
    // not needed
    return false;
}

/**
 * Position the cursor before the first row.
 */
public void beforeFirst() {
    // not needed
}

/**
 * Position the cursor after the last row.
 */
public void afterLast() {
    // not needed
}

/**
 * Return true if the cursor is positioned before first row.
 * @return true if cursor is positioned before first row, false otherwise.
 */
public boolean isBeforeFirst() {
    // not needed
    return false;
}

/**
 * Return true if the cursor is positioned afte the last row.
 * @return true if the cursor is positioned after the last row, false otherwise.
 */
public boolean isAfterLast() {
    // not needed
    return false;
}

/**
 * Move the cursor forward or backward by "n" rows from the current position.
 * Positive numbers move forward, negative backwards.
 * @param number_of_rows an int indicating the number of rows to move.
 * @exception DataModelException If move fails.
 */
public void relative(int rows) throws DataModelException {
    // not needed
}

/**
 * Move the cursor to this ordinal position.
 * Ordinal positions start at 1.
 * @param ordinal_position an int indicating to which row to move.
 * @exception DataModelException If the operation fails.
 */
public void absolute(int ordinal_position) throws DataModelException {
    // not needed
}

/**
 * Return the bookmark of the current row.
 * @return the bookmark of the current row.
 */
public long getCurrentBookmark() {
    // not needed
    return(-1);
}

/**
 * Gets user data.
 * Return the Object which was arbitrarily associated with
 * this table.
 * @param user an Object, a reference to the user.
 * @return the user Object associated with this table.
 */
public Object getUserData(Object user) {
    // not needed
    return null;
}

/**
 * Sets user data.
 * Allow users to augment this table by associating an
 * arbitrary Object reference with it.
 * @param user an Object, a reference to the user.
 * @param data an Object, an arbitrary reference to an Object.
 */
public void setUserData(Object user, Object data) {
    // not needed
}

/**
 * Recursively process pending events from a transaction
 * @exception DataModelException If an exception occurs while processing.
 */
public void processEventQueue(DataModelListener originator) throws DataModelException {
    // not needed
}

/**
 * Recursively clear pending events from a transaction.
 * Called when an exception occurs. Events are not dispatched
 * they are simply deleted.
 */
public void clearEventQueue() {
    // not needed
}

/**
 * Move the cursor to the row indicated by the bookmark.
 * @param bookmark a long which uniquely identifies a row
 * @exception DataModelException If move fails.
 */
public void moveCursorToRow(long bookmark) throws DataModelException {
    // not needed
}

/** Gets the cursor of the data table model. */
public int getCursor() {
    // not needed
    return(-1);
}

/**
 * Given a bookmark, return the zero-relative row index in this table
 * @param bookmark a long which uniquely identifies a row
 * @return rowIndex a int which is the ordinal position of the row
 * in this table
 */
public int getRowIndex(long bookmark) {
    // not needed
    return(-1);
}

/**
 * Deletes the subtree of DataTables rooted with this DataTable
 * as its root.
 * @see com.klg.jclass.datasource.MetaDataModel#getCacheChildren
 * @exception DataModelException If an exception occurs while deleting the subtree.
 */
public void deleteCacheChildren(DataModelListener originator) throws DataModelException {
    // not needed
}

/**
 * For internal use only.
 * Gets internal data.
 * Return the Object which was arbitrarily associated with
 * this table.
 * @param user an Object, a reference to the user.
 * @return the internal Object associated with this table.
 */
public Object getInternalData(Object user) {
    if (internalData == null || user == null) {
        return null;
    }
    return internalData.get(user);
}

/**
 * For internal use only.
 * Sets internal data.
 * Allow component writers to augment this table by associating an
 * arbitrary Object reference with it.
 * @param user an Object, a reference to the user.
 * @param data an Object, an arbitrary reference to an Object.
 */
public void setInternalData(Object user, Object data) {
    if (user == null) {
        return;
    }
    // lazy instantiation of hashtable
    if (internalData == null) {
        internalData = new Hashtable(initialHashtableSize);
    }
    if (data == null) {
        internalData.remove(user);
    }
    else {
        internalData.put(user, data);
    }
}

public void moveToRow(long bookmark) {
    // not needed
}

/**
 * Does the subtree rooted at subTreeRootBookmark contain
 * modified nodes?  Modified means it has pending insert, update or
 * delete changes.
 * @param subTreeRootBookmark a long, the bookmark of the row which is
 * the root of the subtree to be checked.
 * @return true if this subtree contains at least one modified
 * node, false otherwise.
 */
public boolean isSubTreeModified(long subTreeRootBookmark) {
    return false;
}

/**
 * Commit all pending changes in this subtree.
 * @param subTreeRootBookmark a long, the bookmark of the row which is
 * the root of the subtree to be checked.
 */
public void commitSubTree(DataModelListener originator, long subTreeRootBookmark) {
    // not needed
}

/**
 * Does this DataTableModel contain any pending inserts, updates or
 * deletes?
 * @return true if there are pending modifications, false otherwise.
 */
public boolean isModified() {
    // not needed
    return false;
}

/**
 * Return a long array of updated row bookmarks.  These are the rows
 * which have updates pending.
 * @return a long[] of updated row bookmarks
 */
public long[] getUpdatedRowBookmarks() {
    return null;
}

/**
 * Return a long array of inserted row bookmarks.  These are the rows
 * which are inserted into the data model, but not yet persisted.
 * @return a long[] of inserted row bookmarks
 */
public long[] getInsertedRowBookmarks() {
    return null;
}

/**
 * Return a long array of logically deleted row bookmarks.  These are the
 * rows which are logically deleted, but not yet physically deleted from the
 * originating data source.
 * @return a long[] of logically deleted row bookmarks
 */
public long[] getDeletedRowBookmarks() {
    return null;
}

/**
 * Are there any modifications to this table or any of its child nodes?
 * @return true if there are modifications to this table or any of its
 * child nodes
 */
public boolean isTableOrChildrenModified() {
    return false;
}

/**
 * Given a DataTable, returns whether it is an ancestor of
 * (or equal to) itself
 * @param table a DataTable which may be an ancestor
 * @return true if the table is an ancestor
 */
public boolean tableIsAncestor(DataTableModel table2) {
    return false;
}

/**
 * Adds a row whose status will be INSERTED.
 * Adds a blank row into which the user can input column data.
 * The record is appended to this ResultSet.  The record is saved to datasource
 * when the user commits the surrounding transaction.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator) throws DataModelException {
    // not needed
    return -1;
}

/**
 * Adds a populated row whose status will be 'status'.
 * Adds a pre-populated row whose status is dictated by the
 * 'status' parameter.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @param status an int, one of,
 * <ul>
 * <li>DataTableModel.COMMITTED</li>
 * <li>DataTableModel.DELETED</li>
 * <li>DataTableModel.INSERTED<li>
 * </ul>
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row, int status) throws DataModelException {
    // not needed
    return -1;
}

/**
 * Adds a populated row whose status will be INSERTED.
 * Adds a pre-populated row.
 * The record is appended to this ResultSet.  The record is saved to datasource
 * when the user commits the surrounding transaction.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 *
 * @param originator the DataModelListener which initiated this action
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row) throws DataModelException {
    // not needed
    return -1;
}

/**
 * Adds a populated row whose status will be 'status'.
 * Adds a pre-populated row whose status is dictated by the
 * 'status' parameter.
 * <P>
 * Fires BEFORE_INSERT_ROW and upon a successful row add AFTER_INSERT_ROW.
 * <P>
 * @param row an Object as defined by the interface, but used as an array
 * by this implementation.
 * @param status an int, one of,
 * <ul>
 * <li>DataTableModel.COMMITTED</li>
 * <li>DataTableModel.DELETED</li>
 * <li>DataTableModel.INSERTED<li>
 * </ul>
 * @param ordinal the index position into which the row should be inserted
 *
 * @exception DataModelException If row fails to be added.
 */
public long addRow(DataModelListener originator, Object row, int status, int ordinal) throws DataModelException {
    // not needed
    return -1;
}

/**
 * Creates a new (blank) row that can be used as an
 * argument to one of the addRow() methods.
 * @return an Object which is the new row just added.
 */
public Object createNewRow() {
    // not needed
    return null;
}

/**
 * Requery all rows in this table and collapse its child nodes.
 *
 * @param originator the DataModelListener which initiated this action
 * @exception DataModelException If requery fails.
 */
public void requeryTable(DataModelListener originator) throws DataModelException {
    // not needed
}

////////////////////////////////////////////////////////////////////

}
