/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SummaryMetaData.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:20 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.higrid;

import java.io.Serializable;
import java.util.*;

import javax.swing.tree.*;

import com.klg.jclass.datasource.*;

/**
 * <code>SummaryMetaData</code> is the base class for the
 * meta data for a summary record type.
 */

public class SummaryMetaData extends DataSourceTreeNode implements
    MetaDataModel, Serializable {

static final long serialVersionUID = -332831412824408705L;

/**
 * The number of columns for this record type
 */
private int numColumns;

/**
 * A vector of summary column objects
 */
private Vector columns;

/**
 * A hashtable mapping column identifiers to indices
 */
private Hashtable identifiers;

/**
 * meta id
 */
private int metaID; 

/**
 * SummaryMetaData Constructor.
 */
public SummaryMetaData() {
    super(null);
    init();
}

void init() {
    numColumns = 0;
    columns = new Vector();
    identifiers = new Hashtable();
    metaID = -1;
}

/**
 * Set an int which uniquely identifies this instance
 * of the MetaDataModel. This id is the same one as the 
 * corresponding record meta data one.
 * @param metaID An int which uniquely identifies this object.
 */
public void setMetaID(int metaID) {
    // should be the same one as the corresponding record meta data
    this.metaID = metaID;
}

/**
 * Append a column to the end of this table.
 *
 * @param column The summary column object.
 */
public void appendColumn(SummaryColumn column) {
    columns.addElement(column);
    identifiers.put(column.getIdentifier(), new Integer(numColumns));
    numColumns++;
}

/**
 * Insert a column to this table.
 *
 * @param column The summary column object.
 * @param index The index position to insert the object.
 */
public void insertColumn(SummaryColumn column, int index) {
    columns.insertElementAt(column, index);
    numColumns++;
    setupIdentifiers();
}

/**
 * Remove a column from this table.
 *
 * @param column The summary column object.
 */
public void removeColumn(SummaryColumn column) {
    columns.removeElement(column);
    numColumns--;
    setupIdentifiers();
}

/**
 * Remove a column from this table.
 *
 * @param index The index position to insert the object.
 */
public void removeColumn(int index) {
    columns.removeElementAt(index);
    numColumns--;
    setupIdentifiers();
}

/**
 * Return the summary column given the column identifier.
 *
 * @param name A String indicating the column's identifier
 *
 * @return The summary column.
 */
public SummaryColumn getColumn(String name) {
    return(SummaryColumn) columns.elementAt(getColumnIndex(name));
}

/**
 * Setup the summary columns.
 *
 * @param columns The summary column vectors.
 */
public void setupSumaryColumns(Vector columns) {
    this.columns = columns;
    numColumns = columns.size();
    setupIdentifiers();
}

/**
 * Setup the identifiers hashtable.
 */
protected void setupIdentifiers() {
    identifiers.clear();
    for (int i = 0; i < numColumns; i++) {
        identifiers.put(((SummaryColumn) columns.elementAt(i)).getIdentifier(), new Integer(i));
    }
}

/**
 * Given a column identifier get its index.
 *
 * @param name A String indicating the column's identifier
 * @return An int which is this column's ordinal position.
 */
public int getColumnIndex(String name) {
    return((Integer) identifiers.get(name)).intValue();
}

////////////////////////////////////////////////////////////////////
// MetaDataModel interface implementation

/**
 * Change the query for this data control.
 * Reset the summary column data.
 */
public void clear() {
    init();
}

/**
 * Return a string which identifies this column.
 *
 * @param column_index An int indicating the column for which
 * an identifier should be returned.
 * @return A String which identifies a column
 * @exception DataModelException If the operation fails.
 */
public String getColumnIdentifier(int column_index) throws DataModelException {
    return((SummaryColumn) columns.elementAt(column_index)).getIdentifier();
}

/**
 * Return number of columns in this result set
 *
 * @return An int indicating the number of columns in table
 */
public int getColumnCount() {
    return numColumns;
}

/**
 * Return the Java type for a column.
 *
 * @param column_identifier An String uniquely identifying a column
 * @return An int representing the Java type for this column
 * @see MetaDataModel
 * @exception DataModelException If the operation fails.
 */
public int getJavaColumnType(String column_identifier) throws DataModelException {
    if (numColumns == 0) {
        return TYPE_OBJECT;
    }
    return((SummaryColumn) columns.elementAt(getColumnIndex(column_identifier))).getDataType();
}

/**
 * Set the commit policy for this level.  Determines when
 * changes are permanently written to the originating data source.
 * Must be one of COMMIT_LEAVING_RECORD, COMMIT_LEAVING_PARENT,
 * or COMMIT_MANUALLY.
 *
 * @param commit_policy An int, one of COMMIT_LEAVING_RECORD,
 * COMMIT_LEAVING_PARENT, or COMMIT_MANUALLY
 */
public void setCommitPolicy(int commit_policy) {
    // a summary row is not editable
}

/**
 * Return the commit policy for this level.  Determines when
 * changes are permanently written to the originating data source.
 * Will be one of COMMIT_LEAVING_RECORD, COMMIT_LEAVING_PARENT,
 * or COMMIT_MANUALLY.
 *
 * @return An int, one of COMMIT_LEAVING_RECORD,
 * COMMIT_LEAVING_PARENT, or COMMIT_MANUALLY
 */
public int getCommitPolicy() {
    // the default
    return MetaDataModel.COMMIT_LEAVING_RECORD;
}

/**
 * Return the DataModel which encapsulates this MetaDataModel.
 * @return the DataModel
 */
public DataModel getDataModel() {
    return null;
}

/**
 * Set a boolean indicating if this table is allowed to
 * be inserted into.
 * The default is to allow this.
 * @param table a String name of the table
 * @param allowInsert a boolean indicating if inserts are allowed
 */
public void setInsertAllowed(String table, boolean v) {
}

/**
 * Set a boolean indicating if this table is allowed to
 * be updated.
 * The default is to allow this.
 * @param table a String name of the table
 * @param allowUpdate a boolean indicating if updates are allowed
 */
public void setUpdateAllowed(String table, boolean v) {
}

/**
 * Set a boolean indicating if rows in this table is allowed to
 * be deleted.
 * The default is to allow this.
 * @param table a String name of the table
 * @param allowDelete a boolean indicating if deletes are allowed
 */
public void setDeleteAllowed(String table, boolean v) {
}

/**
 * Return true if this table can be inserted into,
 * false otherwise.
 * The default is to allow this.
 * @param table a String name of the table
 * @return true if table can be updated
 */
public boolean getInsertAllowed(String table) {
    return false;
}

/**
 * Return true if this table can be updated, false otherwise.
 * The default is to allow this.
 * @param table a String name of the table
 * @return true if table can be updated
 */
public boolean getUpdateAllowed(String table) {
    return false;
}

/**
 * Return true if this table can be deleted, false otherwise.
 * The default is to allow this.
 * @param table a String name of the table
 * @return true if rows in this table can be deleted.
 */
public boolean getDeleteAllowed(String table) {
    return false;
}

/**
 * Return an int which uniquely identifies this instance
 * of the MetaDataModel. This id is automatically generated.
 * @return an int which uniquely identifies this object.
 */
public int getMetaID() {
    // should be the same one as the corresponding record meta data
    return metaID;
}

/**
 * Return the user-assigned description of this instance, or ""
 * if the description has not been set.
 * @return a String which describes this instance.
 */
public String getDescription() {
    return null;
}

/**
 * Set the user-assigned description of this instance.
 * @param description a String which describles this instance.
 */
public void setDescription(String description) {
}

/**
 * Returns true if the transaction capabilities of the underlying
 * data source should be used when saving modifications
 * (transactions are supported and enabled).
 * If the underlying data source supports transactions and this
 * capability has been enabled, data modifications will be
 * wrapped in a transaction which can be committed or rolled back.
 * @exception DataModelException If the operation fails.
 * @see #commitTransaction
 * @see com.klg.jclass.datasource.BaseMetaData#rollbackTransaction
 */
public boolean getAutoCommit() throws DataModelException {
    return false;
}

/**
 * Sets the auto commit flag.
 * If the underlying data source supports transactions and this
 * flag is set to true, a transaction will wrap data modifications.
 * @param value a boolean, true if transactions are to be used,
 * false otherwise.
 * @exception DataModelException If the operation fails.
 */
public void setAutoCommit(boolean value) throws DataModelException {
}

/**
 * Rolls back the transaction.
 * If the underlying data source supports transactions and
 * transactions have been enabled, calling this method will
 * rollback all pending transactions on the data source (on
 * this connection).
 * @exception DataModelException If the operation fails.
 * @see #getAutoCommit
 */
public void rollbackTransaction() throws DataModelException {
}

/**
 * Commits the transaction.
 * If the underlying data source supports transactions and
 * transactions have been enabled, calling this method will
 * commit all pending transactions on the data source (on
 * this connection).
 * @exception DataModelException If the operation fails.
 */
public void commitTransaction() throws DataModelException {
}

/**
 * Return the list of ColumnModel objects.
 * @return a vector of ColumnModel objects.
 * @see com.klg.jclass.datasource.ColumnModel
 */
public Vector getColumnObjects() {
    return null;
}

/**
 * Do permissions on the model allow this column to be editied?
 * @param column_identifier a String which uniquely identifies this column
 * @return true if edits are allowed, false otherwise
 * @see #setUpdateAllowed
 * @see #getUpdateAllowed
 */
public boolean isUpdateAllowedOnColumn(String column) {
    return false;
}

/**
 * Add a column to the list of column objects.
 * @param columnObject ColumnModel to be added.
 */
public void addColumn(ColumnModel column) {
    // unimplemented
}

/**
 * Return the ColumnModel for the given column.
 * @param column_identifier a String which uniquely identifies a column
 * @return a ColumnModel
 * @exception DataModelException If get fails.
 */
public ColumnModel getColumnObject(String column_identifier) throws DataModelException {
    return null;
}

/**
 * Return the native meta data object.
 * This is a DataSet in JBuilder and a RelationView
 * in VisualCafe.
 * @return the native meta data object for this level.
 */
public Object getNativeMetaData() {
    return null;
}

/**
 * Indicate if logically deleted rows showld be shown
 * @param show a boolean: true means show rows, false hide
 */
public void setShowDeletedRows(boolean value) {
    // unimplemented
}

/**
 * Show logically deleted rows?
 * @return show a boolean: true means show rows, false hide
 */
public boolean getShowDeletedRows() {
    return true;
}

/**
 * Set the maximum number of rows a query at this level can return.
 * Useful for design time where only a sampling of rows is required.
 * Set to -1 to have no limit.
 * <P>
 * Does not apply to ide-specific implmentations.  Set the max rows
 * directly on the ide data object instead.
 * <P>
 * @param maxRows an int indicating the max rows to be returned
 * @see #getMaxRows
 */
public void setMaxRows(int maxRows) {
    // unimplemented
}


/**
 * Return the maximum number of rows which can be returned by a query.
 * <P>
 * Does not apply to ide-specific implmentations.  Set the max rows
 * directly on the ide data object instead.
 * <P>
 * @return the maximum number of rows which can be returned by a query
 * @see #setMaxRows
 */
public int getMaxRows() {
    return -1;
}

/**
 * Sets a boolean indicating whether subtrees of DataTables rooted at this
 * level are eligable for deletion.
 * @param cacheChildren a boolean: true if deleteCacheChildren
 * should be called on this level when it is collapsed, false otherwise.
 * @see #getCacheChildren
 * @see com.klg.jclass.datasource.DataTableModel#deleteCacheChildren
 */
public void setCacheChildren(boolean cacheChildren) {
    // not needed
}

/**
 * Return a boolean flag indicating whether subtrees of DataTables rooted at this
 * level are eligable for deletion.
 * @return true if deleteCacheChildren should be called on this level when
 * it is collapsed, false otherwise.
 * @see #setCacheChildren
 * @see com.klg.jclass.datasource.DataTableModel#deleteCacheChildren
 */
public boolean getCacheChildren() {
    return true;
}

/**
 * Return the data table which currently has focus.
 * @return a DataTableModel, the one currently in focus.
 */
public DataTableModel getCurrentDataTable() {
    return null; // not needed
}
/**
 * Return the bookmark of the row currently in focus.
 * @return the bookmark of the row currently in focus.
 */
public long getCurrentBookmark() {
    return -1;  // not needed
}
public void setCurrentDataTable(DataTableModel currentDataTable) {
    // not needed
}
public void setCurrentBookmark(long currentBookmark) {
    // not needed
}

/**
 * Set an alternate class to handle data storage and retrieval
 * responsibilities.  A "Class.forName()" will be done to
 * load this alternate class.
 * <P>
 * If set, BaseDataTable instances will use this instances of
 * this class rather than the default class ("Store") to
 * store and retrieve data.
 * <P>
 * @param className a String, the name of the class which implements
 * the StoreModel interface
 * @exception DataModelException If an error occurs loading the class.
 * @see StoreModel
 * @see #getStoreClass
 */
public void setStoreClass(String storeClassName) throws DataModelException {
    // not needed
}

/**
 * Return the class responsible for data storage and
 * retrieval (implements StoreModel). If not set, the default class
 * (com.klg.jclass.datasource.Store) is loaded and used.
 * @return the class used to store/retrieve data
 * @exception DataModelException If get fails.
 */
public Class getStoreClass() throws DataModelException {
    return null; // not needed
}


/**
 * Return the Binding object associated with this MetaData level.
 * @return the Binding object for this level.
 */
public BindingModel getBinding() {
	return null; // not needed
}

/**
 * Set the Binding object associated with this MetaData level.
 * @param the Binding object for this level.
 */
public void setBinding(BindingModel binding) {
	// not needed
}

/**
 * Get the provider (if any) for this level.
 * <P>
 * @return ProviderModel the class which will
 * provide matrix data for this level
 * <P>
 * @see ProviderModel#getData
 * @see DataTableModel#createTable
 * @see DataTableModel#getTable 
 */
public ProviderModel getProvider() {
	// not needed
	return null;
}

/**
 * Set the ProviderModel class name which will provide
 * data for this level. This class will be instantiated
 * and to provide data.
 * <P>
 * @param the name of the provider class
 * @exception DataModelException if the provider fails
 * to be instantiated.
 */
public void setProviderClassName(String providerModelClassName) throws DataModelException {
	// not needed
}

////////////////////////////////////////////////////////////////////

}
