/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: EndOfFrameException.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:22 $ $Locker:  $  Quest Software Inc. 
package com.klg.jclass.page;




import java.util.List;


/**
 * PageLayout throws an EndOfFrameException when it runs out of space
 * in a frame while rendering user input.
 */
public class EndOfFrameException extends Exception {

/** A list of render objects comprising the current (partially complete) line */
protected List              currentLine;
/** The list of floating objects yet to be placed */
protected List              floatList;
/** The style that was last used to flow to the frame */
protected JCTextStyle       currentTextStyle;
/** Indicates if the current line is the first line of a new paragraph */
protected boolean           firstLine;

/** A table which was partially printed in the current frame is stored here */
protected JCPageTable       table;
/** The row of the table at which to start the next block of cells */
protected int               next_table_row;
/** The column of the table at which to start the next block of cells */
protected int               next_table_column;
/** The column of the table at which to end the next block of cells */
protected int               last_table_column;

/**
 * Create a blank end of frame exception
 */
public EndOfFrameException() {
	super();

	currentLine = null;
	floatList = null;
	currentTextStyle = null;
	firstLine = false;

	table = null;
	next_table_row = -1;
	next_table_column = -1;
}

/**
 * Create an end of frame exception reporting the given message
 * @param message the message to report in the exception
 */
public EndOfFrameException(String message) {
	super(message);

	currentLine = null;
	floatList = null;
	currentTextStyle = null;
	firstLine = false;

	table = null;
	next_table_row = -1;
	next_table_column = -1;
}

/**
 * Create an end of frame exception based on the frame flow status provided
 * by info and the given message
 * @param info The flow status of the current frame (generating the exception)
 * @param message the message to report in the exception
 */
public EndOfFrameException(JCFlowInfo info, String message) {
	super(message);

	currentLine = (List) info.currentLine.clone();
	floatList = (List) info.floatList.clone();
	currentTextStyle = info.currentTextStyle;
	firstLine = info.firstLine;

	table = null;
	next_table_row = -1;
	next_table_column = -1;
}

/**
 * Get the current line data
 * @return the list of elements in the frame's current line
 */
public List getCurrentLine() {
	return currentLine;
}

/**
 * Get the style last used in the frame's flow
 * @return the style used to flow the last object in the current line
 */
public JCTextStyle getCurrentStyle() {
	return currentTextStyle;
}

/**
 * Get the list of objects waiting to be floated to the document
 * @return the floating object list
 */
public List getFloatList() {
	return floatList;
}

/**
 * Determine if the current line is the first line of a paragraph
 * @return True if the current line is the first line of a new paragraph
 */
public boolean isFirstLine() {
	return firstLine;
}

/**
 * Return the start row for resuming the current table
 * @return the first row of the table as yet unprinted
 */
public int getTableStartRow() {
	return next_table_row;
}

/**
 * Return the start column for resuming the current table
 * @return the first column of the table as yet unprinted
 */
public int getTableStartColumn() {
	return next_table_column;
}

/**
 * Return the end column for resuming the current table section
 * @return the end column of the current section of the table as yet unprinted
 */
public int getTableEndColumn() {
	return last_table_column;
}

/**
 * Return the table which we are in the process of flowing to the document
 * @return the table currently being printed
 */
public JCPageTable getTable() {
	return table;
}

/**
 * Set the parameters for resuming a table in the next frame
 * @param table The table which is currently being printed
 * @param row The first row of the table to print in the new frame
 * @param column The first column of the table to print in the new frame
 */
public void setTableFlow(JCPageTable table, int row, int column) {
	this.table = table;
	next_table_row = row;
	next_table_column = column;
	last_table_column = -1;
}

/**
 * Set the parameters for resuming a table section in the next frame
 * @param table The table which is currently being printed
 * @param row The first row of the table to print in the new frame
 * @param column The first column of the table to print in the new frame
 * @param last_column The last column of the table section to print in the new frame
 */
public void setTableFlow(JCPageTable table, int row, int column, int last_column) {
	this.table = table;
	next_table_row = row;
	next_table_column = column;
	last_table_column = last_column;
}

}
