/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: FontFileLookup.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:22 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.page;

import com.klg.jclass.page.ttf.TTFFontNames;

import java.net.URL;
import java.net.URLConnection;
import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;
import java.util.Vector;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.PropertyResourceBundle;

/**
 * Internal use only.
 * Encapsulates the necessary information related to looking up a particular
 * file.  Currently supports opening files by absolute path, path relative
 * to classpath, path relative to servlet context and full URL.  Also supports
 * the creation of resource bundles from property files specified by
 * absolute path, path relative to classpath, path relative to servlet
 * context, full URL and URL representing a file.
 */
public class FontFileLookup
{

// type of file lookup encapsulated by this class, stored
// in the lookupType member below
/** No type. */
final public static int NONE = -1;
/** File with full (absolute) path. */
final public static int FILE_FULL = 0;
/** URL representing a file (note: this does not have to be an URL that
    begins with "file:", just that the last part of the URL represents
    a concrete file). */
final public static int URL_FILE = 1;
/** URL representing a the location of something that returns a byte stream,
    such as a servlet. */
final public static int URL_STREAM = 2;
/** File relative to classpath. */
final public static int FILE_REL_CLASSPATH = 3;


// type of file instance of this class is pointing to, stored in the fileType
// member below
/** AFM font file. */
final public static int AFM_FILE = 0;
/** TrueType font file. */
final public static int TTF_FILE = 1;
/** Resource Bundle properties file. */
final public static int PROPERTIES_FILE = 2;


/** Type of file lookup encapsulated by this class.  One of NONE,
 *  FILE_FULL, URL_FILE, URL_STREAM, FILE_REL_CLASSPATH or FILE_REL_SERVLET. */
public int    lookupType = NONE;

/* Location of file.  Used for FILE_FULL, FILE_REL_CLASSPATH and
   FILE_REL_SERVLET lookup types. */
public String location = null;

/* Name of properties file.  Used for all lookup types, but only for
   a file type of PROPERTIES_FILE. */
public String propertiesFileName = null;

/* Full URL specifiying file.  Used for URL_FILE and URL_STREAM lookup types. */
public URL    fileURL = null;

/* Type of file.  One of AFM_FILE, TTF_FILE or PROPERTIES_FILE. */
public int    fileType = NONE;

/** Length of the file encapsulated by this object.  This is only set once
    the file has been opened.  A length of -1 indicates that either the
    length has not yet been determined or, if the file has already been
    opened, that it can't be determined.
 */
protected long	fileLength = -1;

/** Stores optional TTFFontNames object containing possible names for
 TrueType font file. */
protected TTFFontNames ttfFontNames = null;


/**
 * Empty constructor.  Don't use is unless you know exactly what you're doing.
 */
public FontFileLookup() {
}

/**
 * Constructor for FILE_FULL and FILE_REL_CLASSPATH lookup types for
 * .afm and .ttf files.
 * @param lookupType one of FILE_FULL or FILE_REL_CLASSPATH.
 * @param location if FILE_FULL, the absolute path location of the
 *                 desired file; if FILE_REL_CLASSPATH, the location
 *                 of the desired file relative to the classpath.
 * @param fileType one of AFM_FILE or TTF_FILE.
 */
public FontFileLookup(int lookupType, String location, int fileType)
{

	if (lookupType != FILE_FULL && lookupType != FILE_REL_CLASSPATH) {
		throw new IllegalArgumentException("this constructor may only be used for lookup types of FILE_FULL and FILE_REL_CLASSPATH");
	}

	if (fileType != AFM_FILE && fileType != TTF_FILE) {
		throw new IllegalArgumentException("this constructor may only be used for file types of AFM_FILE and TTF_FILE");
	}

    this.lookupType = lookupType;
    this.location = location;
	this.fileType = fileType;
}


/**
 * Constructor for URL_FILE or URL_STREAM lookup types for .afm and .ttf files.
 * @param lookupType URL_FILE or URL_STREAM.
 * @param fileURL the location of the desired file specified as an URL object.
 * @param fileType one of AFM_FILE or TTF_FILE.
 */
public FontFileLookup(int lookupType, URL fileURL, int fileType)
{

	if (lookupType != URL_FILE && lookupType != URL_STREAM) {
		throw new IllegalArgumentException("this constructor may only be used for lookup types of URL_FILE and URL_STREAM");
	}

	if (fileType != AFM_FILE && fileType != TTF_FILE) {
		throw new IllegalArgumentException("this constructor may only be used for file types of AFM_FILE and TTF_FILE");
	}

    this.lookupType = lookupType;
    this.fileURL = fileURL;
	this.fileType = fileType;
}

/**
 * Constructor for FILE_FULL and FILE_REL_CLASSPATH lookup types for
 * resource bundle .properties files.
 * @param lookupType one of FILE_FULL or FILE_REL_CLASSPATH.
 * @param location if FILE_FULL, the absolute path location of the
 *                 directory containing the .propertiues file; if
 *                 FILE_REL_CLASSPATH, the location of this directory
 *                 relative to the classpath.
 * @param name the name portion of the .properties file, the bit before
 *             the locale information or .propeties extension.  For example,
 *             a file called "user_en_US.properties" would have a name of
 *             "user".
 * @param fileType PROPERTIES_FILE.
 */
public FontFileLookup(int lookupType, String location,
									String name, int fileType)
{

	if (lookupType != FILE_FULL && lookupType != FILE_REL_CLASSPATH) {
		throw new IllegalArgumentException("this constructor may only be used for lookup types of FILE_FULL and FILE_REL_CLASSPATH");
	}

	if (fileType != PROPERTIES_FILE) {
		throw new IllegalArgumentException("this constructor may only be used for file types of PROPERTIES_FILE");
	}

    this.lookupType = lookupType;
    this.location = location;
    this.propertiesFileName = name;
	this.fileType = fileType;
}

/**
 * Constructor for URL_FILE and URL_STREAM lookup types for
 * resource bundle .properties files.
 * @param lookupType one of URL_FILE or URL_STREAM.
 * @param fileURL if lookupType is URL_FILE, this is an URL specifying the
 *                directory the desired properties file will be found in.
 *                If lookupType is URL_STREAM, this is an URL specifying the
 *                exact location of an entity that returns a byte stream
 *                containing a properties file.
 * @param name if lookupType is URL_FILE, this is the name portion of the
 *             .properties file, the bit before the locale information or
 *             .propeties extension.  For example, a file called
 *             "user_en_US.properties" would have a name of "user".
 *             If lookupType is URL_STREAM, this parameter is ignored.
 * @param fileType PROPERTIES_FILE.
 */
public FontFileLookup(int lookupType, URL fileURL,
										String name, int fileType)
{

	if (lookupType != URL_FILE && lookupType != URL_STREAM) {
		throw new IllegalArgumentException("this constructor may only be used for lookup types of URL_FILE and URL_STREAM");
	}

	if (fileType != PROPERTIES_FILE) {
		throw new IllegalArgumentException("this constructor may only be used for file types of PROPERTIES_FILE");
	}

    this.lookupType = lookupType;
    this.fileURL = fileURL;
    this.propertiesFileName = name;
	this.fileType = fileType;
}

 /**
   * Store optional TTFFontNames object containing possible names
   * for TrueType file.
   */
public void setTTFFontNames(TTFFontNames ttfFontNames){
       this.ttfFontNames = ttfFontNames;
}
 
 /**
   * Return optional TTFFontNames object containing possible names
   * for TrueType file.
   */
 public TTFFontNames getTTFFontNames(){
       return ttfFontNames;
}
    
/**
 * Open an input stream on the file encapsulated by this class.
 * Should not be used for .properties resource bundle files.
 * @return an open InputStream object representing the file
 *         encapsulated by this class.  If the stream could not
 *         be opened, this will be null.
 */
public InputStream openInputStream() throws IOException
{
	InputStream inStream = null;

	// Absolute file -- just open a stream
	if (lookupType == FILE_FULL) {

		File file = new File(location);
		fileLength = file.length();
		inStream = new FileInputStream(file);

	// Absolute URL -- just open a stream
	} else if (lookupType == URL_FILE || lookupType == URL_STREAM) {

		URLConnection urlConn = fileURL.openConnection();
		fileLength = urlConn.getContentLength();
		inStream = urlConn.getInputStream();

	// File relative to classpath -- open stream using default class loader
	} else if (lookupType == FILE_REL_CLASSPATH) {
		// file relative to classpath -- ensure that path begins with
		// the classpath entries, not the directory of the current class,
		// by enforcing a leading slash

		String modLocation =
                (location.charAt(0) != '/' ? "/" : "") + location;

		URL url = getClass().getResource(modLocation);
		URLConnection urlConn = url.openConnection();
		fileLength = urlConn.getContentLength();
		inStream = urlConn.getInputStream();

	}

	return inStream;
}

/**
 * Get the length of the file encapsulated by this object.  Note:
 * this value is only set once the file has been opened with a call
 * to openInputStream(), and even then it may not be determinable.
 * A length of -1 indicates that the length of the file is unknown.
 */
public long getFileLength()
{
	return fileLength;
}

/**
 * Open a resource bundle .properties file based on the information
 * encapsulated by this class.  Unless otherwise specified, the default
 * locale should be taken into account when deriving the .properties file
 * name.  For example, for a bundle named "user" in the North American
 * English locale, this method would first search for a file named
 * "user_en_US.properties", then "user_en.properties" and finally
 * "user.properties".
 * @return a ResourceBundle derived from the .properties file
 *         encapsulated by this class
 */
public ResourceBundle getResourceBundle() throws IOException
{
	if (fileType != PROPERTIES_FILE)
		return null;

	ResourceBundle bundle = null;

	// Absolute file -- get a directory listing and look through it
	//                  for the most appropriate properties file
	if (lookupType == FILE_FULL) {

    	File filepath = new File(location);

    	// get the list of property files in this directory
    	File[] propFileList = filepath.listFiles(new FilenameFilter() {
        	public boolean accept(File dir, String name) {
            	if (name.trim().endsWith(".properties")) {
                	return true;
            	} else {
                	return false;
            	}
        	}
    	});

    	// apply the list of property bundle suffixes for this locale
		// to the bundle name
    	List bundleSuffixes =
        	calculateBundleSuffixes(propertiesFileName);

    	// find the most specific property file that matches this locale
    	File propFile = null;
    	for (int i = bundleSuffixes.size() - 1;  i >= 0 ; i--) {
        	for (int j = 0; propFileList != null && j < propFileList.length; j++) {
            	if (propFileList[j].getName().equals(
                    	(String) bundleSuffixes.get(i) + ".properties")) {
                	propFile = propFileList[j];
                	break;
            	}
        	}
        	if (propFile != null)
            	break;
    	}

    	if (propFile != null) {
        	// load the resource bundle
			// Note: we catch and report this FileNotFoundException because
			// if it occurs, it is being caused by our creation of the file
			// name and we want to know about it
			FileInputStream bundleStream = null;
			try {
           		bundleStream = new FileInputStream(propFile);
        	} catch (FileNotFoundException fnfe) {
            	fnfe.printStackTrace();
        	}

			if (bundleStream != null) {
          		bundle = new PropertyResourceBundle(bundleStream);
				bundleStream.close();
			}
    	}

	// Absolute URL  -- just open the specified stream and make a resource
 	//                  bundle out of it
	} else if (lookupType == URL_STREAM) {
		InputStream bundleStream = fileURL.openStream();
       	bundle = new PropertyResourceBundle(bundleStream);
		bundleStream.close();

	// The others -- try various possible properties files until one is found
	} else if (lookupType == FILE_REL_CLASSPATH ||
				lookupType == URL_FILE) {

		String modLocation = null;
		if (lookupType == FILE_REL_CLASSPATH) {
			// file relative to classpath -- ensure that path begins with
			// the classpath entries, not the directory of the current class,
			// by enforcing a leading slash

			modLocation =
                	(location.charAt(0) != '/' ? "/" : "") + location;
		} else if (lookupType == URL_FILE) {
			modLocation = fileURL.toString();
		}

		// enforce a trailing slash representing the division between
		// the directory location and the file name
		modLocation =
               ((modLocation.endsWith("/") || modLocation.endsWith("\\")) ?
				modLocation : modLocation + "/");

    	// apply the list of property bundle suffixes for this locale
		// to the bundle name
    	List bundleSuffixes =
        	calculateBundleSuffixes(propertiesFileName);

		InputStream bundleStream = null;

		// look for the existence of the possible bundle names,
		// starting with the most explicit and working to the most general
    	for (int i = bundleSuffixes.size() - 1;  i >= 0 ; i--) {
			if (lookupType == FILE_REL_CLASSPATH) {
				// file relative to classpath -- try to open stream from class
				bundleStream =
					getClass().getResourceAsStream(
						modLocation +
                   		(String) bundleSuffixes.get(i) +
						".properties");
			} else if (lookupType == URL_FILE) {

				// file url -- try to open stream at URL

				// Note: need to catch and eat this exception because it will
				// happen as result of checking to see if certain URLs
				// exist.  We only want to report exceptions if they're
				// caused by the opening of files explicitly requested by
				// the user.
        		try {
					URL modURL = new URL(modLocation +
                   						(String) bundleSuffixes.get(i) +
										".properties");
					//java.net.URLConnection urlcon = modURL.openConnection();
					//bundleStream = urlcon.getInputStream();
					bundleStream = modURL.openStream();
        		} catch (IOException ioe) {
        		}
			}
        	if (bundleStream != null)
            	break;
    	}

		// if input stream representing bundle file was found,
		// create an actual resource bundle from it

    	if (bundleStream != null) {
           	bundle = new PropertyResourceBundle(bundleStream);
			bundleStream.close();
    	}
	}

	return bundle;
}

/**
 * Apply a list of potential suffixes for use in searching for
 * resource bundles (derived from the default locale) to the
 * passed base name.
 * @param baseName name of resource bundle before locale information
 *                 is added.
 */
protected static List calculateBundleSuffixes(String baseName)
{
    // this is pretty much copied right out of Sun's ResouceBundle class
    Vector bundleSuffixes = new Vector();
    Locale thisLocale = Locale.getDefault();
    String language = thisLocale.getLanguage();
    int languageLength = language.length();
    String country = thisLocale.getCountry();
    int countryLength = country.length();
    String variant = thisLocale.getVariant();
    int variantLength = variant.length();

    bundleSuffixes.addElement(baseName);

    if (languageLength + countryLength + variantLength == 0) {
        //The locale is "", "", "".
        return bundleSuffixes;
    }

    final StringBuffer temp = new StringBuffer(baseName);
    temp.append('_');
    temp.append(language);
    bundleSuffixes.addElement(temp.toString());

    if (countryLength + variantLength == 0) {
        return bundleSuffixes;
    }
    temp.append('_');
    temp.append(country);
    bundleSuffixes.addElement(temp.toString());

    if (variantLength == 0) {
        return bundleSuffixes;
    }
    temp.append('_');
    temp.append(variant);
    bundleSuffixes.addElement(temp.toString());

    return bundleSuffixes;
}

public String toString()
{
	StringBuffer sb = new StringBuffer();
	if (fileType == AFM_FILE) {
		sb.append("AFM ");
	} else if (fileType == TTF_FILE) {
		sb.append("TTF ");
	} else if (fileType == PROPERTIES_FILE) {
		sb.append("PROP ");
	} else {
		sb.append("Unknown ");
	}

	if (lookupType == FILE_FULL) {
		sb.append("File: ");
		sb.append(location);
	} else if (lookupType == URL_FILE || lookupType == URL_STREAM) {
		sb.append("URL: ");
		sb.append(fileURL);
	} else if (lookupType == FILE_REL_CLASSPATH) {
		sb.append("Rel (CP) File: ");
		sb.append(location);
	}

	if (fileType == PROPERTIES_FILE) {
		sb.append(" ");
		sb.append(propertiesFileName);
	}

	return sb.toString();
}

/**
 * Test equality.
 */
public boolean equals(Object o)
{
	boolean ret = false;
	if (o instanceof FontFileLookup) {
		FontFileLookup ffl = (FontFileLookup) o;
		if (ffl.fileType != this.fileType) {
			return false;
		}
		if (ffl.lookupType != this.lookupType) {
			return false;
		}

		if (ffl.lookupType == NONE) {
			ret = true;
		} else if (ffl.lookupType == FILE_FULL) {
			if ((ffl.location == null && this.location == null) ||
							ffl.location.equals(this.location)) {
				ret = true;
			} else {
				return false;
			}
		} else if (ffl.lookupType == URL_FILE) {
			if ((ffl.fileURL == null && this.fileURL == null) ||
							ffl.fileURL.equals(this.fileURL)) {
				ret = true;
			} else {
				return false;
			}
		} else if (ffl.lookupType == URL_STREAM) {
			if ((ffl.fileURL == null && this.fileURL == null) ||
							ffl.fileURL.equals(this.fileURL)) {
				ret = true;
			} else {
				return false;
			}
		} else if (ffl.lookupType == FILE_REL_CLASSPATH) {
			if ((ffl.location == null && this.location == null) ||
							ffl.location.equals(this.location)) {
				ret = true;
			} else {
				return false;
			}
		}

		if (ffl.fileType == PROPERTIES_FILE) {
			if ((ffl.propertiesFileName == null &&
							this.propertiesFileName == null) ||
					ffl.propertiesFileName.equals(this.propertiesFileName)) {
				ret = true;
			} else {
				return false;
			}
		}
	}
	return ret;
}

}
