/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: FontLibrary.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:22 $ $Locker:  $  Quest Software Inc.


 

package com.klg.jclass.page;

import com.klg.jclass.page.resources.LocaleBundle;
import com.klg.jclass.page.adobe.postscript.TTFParser;
import com.klg.jclass.page.ttf.TTFFontNames;
import com.klg.jclass.util.SettableListResourceBundle;

import java.io.*;
import java.net.URL;
import java.util.*;

/**
 * Stores lists of available fonts in Hashtables for easy lookup,
 * sometimes depending on output type.
 * The list of stock fonts for PDF output and the files that contain them is
 * currently found in the font.map file in the 
 * com.klg.jclass.sreport.adobe.fonts package.
 * The list of stock fonts for RTF output is currently found in the
 * font.map file in the com.klg.jclass.sreport.rtf package.
 * The list of user font names that logically map to these fonts is
 * currently found in com.klg.jclass.sreport.adobe.JCAdobeFontMap class
 * for PDF output and the com.klg.jclass.sreport.rtf.JCRTFFontMap class
 * for RTF output.
 * Default directories are searched for additional TrueType fonts
 * that are added to the list of available fonts.
 * Users may specify additional fonts by directory or by font file name.
 * User-added directories are searched for font name map resource bundle
 * files named user.properties, but such files can also be added manually.
 * Automatic font mapping is also attempted for TrueType fonts.
 * User-specified font name mapping takes precedence over built-in mapping
 * of stock fonts, which itself takes precedence over automatic font name
 * mapping.
 */
public class FontLibrary {

// type of font 
/** Type 1 Font (backed with .afm file) */
final public static int AFM = FontFileLookup.AFM_FILE;
/** TrueType Font */
final public static int TTF = FontFileLookup.TTF_FILE;
/** Special case: all file types! */
final public static int ALL_FILE_TYPES = -100;

// Output types -- indexes into the following arrays
/** Output Type: PDF.  Used to determine which stock font map
 to use during font lookup.  */
public static final int PDF = 0;
/** Number of output types supported. Currently 1.*/
public static final int NUM_OUTPUT_TYPES = 1;

/** Array of Hashtables containing available fonts have been read from the
 default font directories or have been specified by a user.  Indexed by
 output type (PDF or RTF).  (Stock fonts used by the various output
 formats are stored in a separate variable.) Keys are actual font names 
 (often derived from looking up the user's font alias name in the
 font name map resource bundles).  Values are FontFileLookup classes
 encapsulating the location of the afm or ttf file backing the font.
 */
public static Hashtable availableFonts[] =
{
    new Hashtable(),
    new Hashtable()
};

/** Array of Hashtables containing available stock fonts.  Indexed by
 output type (PDF or RTF).  Keys are actual font names (often derived from
 looking up the user's font alias name in the font name map resource
 bundles).  Values are FontFileLookup classes encapsulating the location
 of the afm or ttf file backing the font.  This array is populated
 in the constructor.
 */
public static Hashtable availableStockFonts[] =
{
    new Hashtable(),
    new Hashtable()
};

/** Array of stock font name map bundles containing default mappings for
 stock fonts.  Indexed by output type.  These mappings take precedence
 over the automatically generated mappings but are overridden by any
 user-specified mappings.  Null indicates no map exists
 for the corresponding output type.  This array is populated
 in the constructor of the respective output types. */
public static ResourceBundle[] stockFontNameMapBundles =
    new ResourceBundle[NUM_OUTPUT_TYPES];

/** Name before ".properties" in user font name map bundle name. */
protected final static String defaultUserFontBundleName = "user";

/** Array of ResourceBundle lists representing the font name
 resource bundles for user-added font packages.  Indexed by
 output type. */
protected static List[] userFontBundles =
    {
        new Vector(),
        new Vector()
    };

/** Corresponding List of FontFileLookup objects that were used to
 find the font bundles.  Used to avoid loading bundles that
 have already been loaded.
 */
protected static List userFontBundleLookups = new Vector();

/** Table of automatically generate font name mappings from
 *  possible user font names to the real font name as used
 *  by ServerReport internals.  These mappings are overridden
 *  by the stock font name mappings and any user-specified mappings.
 */
protected static Hashtable automaticFontNameMapTable = new Hashtable();
    
/** Global hashtable storing user properties for use of true type fonts.
 *  Keys are the same font names used in the availableFonts hashtable.
 *  Values are TrueTypeFontProperties objects.
 */
public static Hashtable trueTypeFontProperties = new Hashtable();

/** True if default font directories should be searched for fonts.
    Default is false. Default font directories include, specifically,
    <BR>C:\WINNT\fonts<BR>C:\WINDOWS\Fonts<BR>/usr/share/fonts<BR>
    /usr/lib/X11/fonts<BR>. These directories can be loaded 
    automatically by setting this property to true. To load other
    directories, leave false and use
    <CODE><A HREF="#addFontPackage(java.lang.String)>
    addFontPackage(String)</A> to add the directories you want.
 */
protected static boolean autoLoad = false;

/** Actual instance of FontLibrary.  Currently only used so that
 the constructor is never called twice.
 */
protected static FontLibrary fontLibraryInstance = null;

/**
 * True if font-related error messages should be suppressed.
 * Default is true.
 */
protected static boolean suppressErrorMessages = true;

/**
 * Initialize FontLibrary class by creating an instance and calling its
 * constructor (if this hasn't been done yet).
 */
public static void initFontLibrary()
{
    if (fontLibraryInstance == null) {
        fontLibraryInstance = new FontLibrary();
        if (autoLoad) {
            addFontPackage("C:\\WINNT\\Fonts");    // Windows NT, 2K
            addFontPackage("C:\\WINDOWS\\Fonts");  // Windows XP
            addFontPackage("/usr/share/fonts");  // Unix
            addFontPackage("/usr/lib/X11/fonts");// Unix, X Windows
        } 
    }
}

/**
 * Singleton class with empty private constructor - all useful methods are 
 * static.
 */
private FontLibrary()
{
    // Stock font loading and resource bundle loading deferred to output-type 
    // specific classes (ie. FontPS). After loading stock fonts, FontLibrary needs 
    // to be informed to maintain accurate indices of available fonts. 
}//FontLibrary

/**
 * Set the AutoLoad property, which determines if default directories
 * are automatically searched.  Default is true.
 * @param al the new value of the AutoLoad property.
 */
public static void setAutoLoad(boolean al)
{
    autoLoad = al;
}

/**
 * Get the AutoLoad property, which determines if default directories
 * are automatically searched.  Default is true.
 * @return the value of the AutoLoad property.
 */
public static boolean isAutoLoad()
{
    return autoLoad;
}

/**
 * Set the SuppressErrorMessages property, which suppresses font-related
 * error messages if set to true.  Default is false.
 * @param suppress the new value of the SuppressErrorMessages property.
 */
public static void setSuppressErrorMessages(boolean suppress)
{
    suppressErrorMessages = suppress;
}

/**
 * Get the SuppressErrorMessages property, which determines if font-related
 * error messages are being suppressed.  Default is false.
 * @return the value of the SuppressErrorMessages property.
 */
public static boolean getSuppressErrorMessages()
{
    return suppressErrorMessages;
}

/**
 * Add any fonts represented by .afm or .ttf files in the provided directory
 * (and all of its subdirectories) to the hashtable of available fonts.
 * These directories are also stored for use in reading font files when
 * they become needed.  If a font name map resource bundle property
 * file of name "user" and corresponding to the default locale exists
 * in the directory, it is also loaded and its contents added to
 * ServerReport's global font name map.  (For example, for the North
 * American/English locale, the directory is first searched for the
 * file named "user_en_US.properties", then "user_en.properties" and finally
 * "user.properties".  Font name map files outline the names the user will
 * refer to fonts by the and the names by which they're referred inside the
 * corresponding .afm or .ttf file.)
 * This method may be called via the convenience method in JCDocument.
 * @param packagePath a String representing the absolute location of the 
 *                    directory which contains both font files (.afm
 *                    or .ttf) and a font name map resource bundle
 *                    property file (e.g. "user.properties").
 */
public static void addFontPackage(String packagePath)
{
    FontLibrary.initFontLibrary();
    addFontPackage(packagePath, true);
}

/**
 * Add any fonts represented by .afm or .ttf files in the provided 
 * directory (and all of its subdirectories, if recurse is set to
 * true) to the hashtable of available fonts.  These directories
 * are also stored for use in reading font files when they become
 * needed.  If a font name map resource bundle property file
 * of name "user" and corresponding to the default locale exists
 * in the directory, it is also loaded and its contents added to
 * ServerReport's global font name map.  (For example, for the North
 * American/English locale, the directory is first searched for the
 * file named "user_en_US.properties", then "user_en.properties" and finally
 * "user.properties".  Font name map files outline the names the user will
 * refer to fonts by the and the names by which they're referred inside the
 * corresponding .afm or .ttf file.)
 * This method may be called via the convenience method in JCDocument.
 * @param packagePath a String representing the absolute location of the 
 *                    directory which contains both font files (.afm
 *                    or .ttf) and a font name map resource bundle
 *                    property file (e.g. "user.properties").
 * @param recurse if true, all subdirectories under provided directory
 *                will also be parsed for font files and font name map files
 */
public static void addFontPackage(String packagePath, boolean recurse)
{
    FontLibrary.initFontLibrary();

    if (packagePath == null) {
        throw new IllegalArgumentException("package path may not be null");
    }

    File filepath = new File(packagePath);

    // first, load the afm and ttf files found in this directory
    File[] afmFileList = filepath.listFiles(new FilenameFilter() {
        public boolean accept(File dir, String name) {
            if (isAFMFile(name.trim())) {
                return true;
            }
            else {
                return false;
            }
        }
    });
    // keep list of added afm fonts for reconciling against the
    // font name map that may be in this directory
    Vector addedAFMFonts = new Vector();
    for (int i = 0; afmFileList != null && i < afmFileList.length; i++) {
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.FILE_FULL,
            afmFileList[i].getPath(),
            FontFileLookup.AFM_FILE);

        // We catch the exception here because the user has asked us
        // to add an entire package.  Problems with any single file
        // should be noted, but should not be cause for the whole
        // method to fail.
        try {
            String afmFontName = addAFMToAvailableFonts(flook);
            if (afmFontName != null) {
                addedAFMFonts.add(afmFontName);
            }
        } catch (IOException ioe) {
            if (!getSuppressErrorMessages()) {
                System.err.println(
                    LocaleBundle.string(LocaleBundle.ERROR_LOADING_AFM) +
                    afmFileList[i].getPath());
                System.err.println("    (" +
                                   LocaleBundle.string(LocaleBundle.EXCEPTION_COLON) +
                                   " " + ioe.getClass().getName() + " " +
                                   ioe.getMessage() + ")");
            }
        }
    }

    File[] ttfFileList = filepath.listFiles(new FilenameFilter() {
        public boolean accept(File dir, String name) {
            if (isTrueTypeFile(name.trim())) {
                return true;
            }
            else {
                return false;
            }
        }
    });
    // keep list of added ttf fonts for reconciling against the
    // font name map that may be in this directory
    Vector addedTTFFonts = new Vector();
    for (int i = 0; ttfFileList != null && i < ttfFileList.length; i++) {
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.FILE_FULL,
            ttfFileList[i].getPath(),
            FontFileLookup.TTF_FILE);

        // We catch the exception here because the user has asked us
        // to add an entire package.  Problems with any single file
        // should be noted, but should not be cause for the whole
        // method to fail.
        try {
            List ttfFontNames = addTTFToAvailableFonts(flook);
            if (ttfFontNames != null) {
                addedTTFFonts.addAll(ttfFontNames);
            }
        } catch (IOException ioe) {
            if (!getSuppressErrorMessages()) {
                System.err.println(
                    LocaleBundle.string(LocaleBundle.ERROR_LOADING_TTF) +
                    ttfFileList[i].getPath());
                System.err.println("    (" +
                                   LocaleBundle.string(LocaleBundle.EXCEPTION_COLON) +
                                   " " + ioe.getClass().getName() + " " +
                                   ioe.getMessage() + ")");
            }
        }
    }

    // Now, load the appropriate user.properties resource bundle from this
    // directory

    // We catch the exception here because the user has asked us
    // to add an entire package.  Problems with any single file
    // should be noted, but should not be cause for the whole
    // method to fail.
    // set up the file lookup for an absolute path properties file
    FontFileLookup propLook = new FontFileLookup(
        FontFileLookup.FILE_FULL,
        packagePath, defaultUserFontBundleName,
        FontFileLookup.PROPERTIES_FILE);

    // if we've loaded this bundle before, no need to load it again
    if (userFontBundleLookups.contains(propLook)) {
        return;
    }

    try {
        // search for the resource bundle using the file lookup object
        ResourceBundle thisBundle = propLook.getResourceBundle();

        if (thisBundle == null) {
            return;
        }

        // if found, add the bundle to the global list of font name map bundles
        // add bundle to lookup list so that we don't load it a second time
        userFontBundleLookups.add(propLook);

        // since this resource bundle refers to fonts that can be in
        // both afm and ttf files (and since resource bundles need
        // to be stored by output type), tear apart this resource bundles
        // and construct new ones for each output type

        // create new resource bundles, one per output type
        SettableListResourceBundle newBundles[] =
            new SettableListResourceBundle[NUM_OUTPUT_TYPES];
        for (int i = 0; i < NUM_OUTPUT_TYPES; i++) {
            newBundles[i] = new SettableListResourceBundle();
        }

        // for each key in the loaded bundle, check the type of file
        // the key is referring to and use that to determine which
        // output type-specific bundle the key should be added to
        Enumeration bundleKeys = thisBundle.getKeys();
        while (bundleKeys.hasMoreElements()) {
            String thisKey = (String)bundleKeys.nextElement();
            String thisValue = thisBundle.getString(thisKey);

            // afm fonts -- used by pdf only
            if (addedAFMFonts.contains(thisValue)) {
                newBundles[PDF].add(thisKey, thisValue);
            }
            // ttf fonts -- used by both pdf and rtf
            if (addedTTFFonts.contains(thisValue)) {
                newBundles[PDF].add(thisKey, thisValue);
            }
        }

        // add these new bundles (not the loaded bundle) to the
        // list of user-added bundles
        for (int i = 0; i < NUM_OUTPUT_TYPES; i++) {
            userFontBundles[i].add(newBundles[i]);
        }
    } catch (IOException ioe) {
        if (!getSuppressErrorMessages()) {
            System.err.println(
                LocaleBundle.string(LocaleBundle.ERROR_LOADING_MAP) +
                packagePath + "/" + defaultUserFontBundleName + ".properties");
            System.err.println("    (" +
                               LocaleBundle.string(LocaleBundle.EXCEPTION_COLON) +
                               " " + ioe.getClass().getName() + " " +
                               ioe.getMessage() + ")");
        }
    }//try-catch

    // recurse sub-directories if requested
    if (recurse) {
        File[] dirList = filepath.listFiles(new FileFilter() {
            public boolean accept(File name) {
                if (name.isDirectory()) {
                    return true;
                }
                else {
                    return false;
                }
            }
        });
        for (int i = 0; dirList != null && i < dirList.length; i++) {
            addFontPackage(dirList[i].getAbsolutePath(), true);
        }
    }

}//addFontPackage

/**
 * Add the font represented by the specified .afm or .ttf file to the hashtable
 * of available fonts.  The specified font file location must be the
 * absolute location of the desired file and will be stored for use in
 * reading the font file when it is needed.
 * @param fileLocation a String representing the absolute location of a
 *                     .afm or .ttf font file.
 */
public static void addFont(String fileLocation) throws IOException
{
    FontLibrary.initFontLibrary();

    if (fileLocation == null) {
        throw new IllegalArgumentException("file location may not be null");
    }

    File fullFile = new File(fileLocation);

    // if passed file is an .afm file ...
    if (isAFMFile(fileLocation)) {

        // set up font lookup class with absolute file path
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.FILE_FULL,
            fullFile.getPath(),
            FontFileLookup.AFM_FILE);

        //add .afm file to list of available fonts
        addAFMToAvailableFonts(flook);

        // if passed file is an .ttf (TrueType) file ...
    }
    else if (isTrueTypeFile(fileLocation)) {

        // set up font lookup class with absolute file path
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.FILE_FULL,
            fullFile.getPath(),
            FontFileLookup.TTF_FILE);

        //add .afm file to list of available fonts
        addTTFToAvailableFonts(flook);

    }

}//addFont

/**
 * Add the font represented by the specified AFM or TrueType file to the
 * hashtable of available fonts.  The specified font file location must be
 * the absolute location of the desired file and will be stored for use in
 * reading the font file when it is needed.
 * @param fontURL an URL representing the absolute location of a font file
 *                in either .afm or .ttf format
 * @param fontType either AFM or TTF, depending on the contents located
 *                 at the destination represented by fontURL
 */
public static void addFont(URL fontURL, int fontType) throws IOException
{
    FontLibrary.initFontLibrary();

    if (fontURL == null) {
        throw new IllegalArgumentException("font URL may not be null");
    }

    if (fontType != AFM && fontType != TTF) {
        throw new IllegalArgumentException("font type must be either AFM or TTF");
    }

    // afm file
    if (fontType == AFM) {

        // set up font lookup class with absolute url path
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.URL_STREAM,
            fontURL,
            FontFileLookup.AFM_FILE);

        //add .afm file to list of available fonts
        addAFMToAvailableFonts(flook);

        // true type file
    }
    else if (fontType == TTF) {

        // set up font lookup class with absolute url path
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.URL_STREAM,
            fontURL,
            FontFileLookup.TTF_FILE);

        //add .afm file to list of available fonts
        addTTFToAvailableFonts(flook);

    }

}//addFont

/**
 * Add the font represented by the specified AFM or TrueType file to the
 * hashtable of available fonts.  The specified font file location must be
 * a location relative to the classpath.  It will be stored for use in
 * reading the font file when it is needed.
 * @param fileLocation a String representing the relative location of a
 *                     .afm or .ttf font file.   It is assumed that the
 *                     location begins at the tops of the classpath
 *                     entries.  (In other words, if the location does
 *                     not begin with a '/', one is added.)
 */
public static void addRelativeFont(String fileLocation) throws IOException
{
    FontLibrary.initFontLibrary();

    if (fileLocation == null) {
        throw new IllegalArgumentException("file location may not be null");
    }

    // afm file
    if (isAFMFile(fileLocation)) {

        // set up font lookup class with a relative-to-classpath file path
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.FILE_REL_CLASSPATH,
            fileLocation,
            FontFileLookup.AFM_FILE);

        //add .afm file to list of available fonts
        addAFMToAvailableFonts(flook);

        // TrueType file
    }
    else if (isTrueTypeFile(fileLocation)) {

        // set up font lookup class with a relative-to-classpath file path
        FontFileLookup flook = new FontFileLookup(
            FontFileLookup.FILE_REL_CLASSPATH,
            fileLocation,
            FontFileLookup.TTF_FILE);

        //add .afm file to list of available fonts
        addTTFToAvailableFonts(flook);

    }

}//addRelativeFont

/**
 * Add the font inside the AFM file represented by the passed FontFileLookup
 * object to the list of available fonts by doing a quick parse of the afm file
 * to determine the name of the font that is contained within.
 * AFM fonts are used for PDF output only.
 * @param afmFile a FontFileLookup object referencing an .afm file
 * @return name of AFM font added.  May be null if no file was found
 *         at the provided location.
 */
protected static String addAFMToAvailableFonts(FontFileLookup afmFile)
    throws IOException
{
    if (availableFonts[PDF].contains(afmFile)) {
        return null;
    }

    InputStream inStream = afmFile.openInputStream();
    StreamTokenizer st = new StreamTokenizer(
        new InputStreamReader(inStream));
    st.eolIsSignificant(true);
    st.commentChar('#');
    st.nextToken();

    String afmFontName = "";

    while (st.ttype != StreamTokenizer.TT_EOF) {

        if (st.sval != null && st.sval.equals("FontName")) {
            st.nextToken();

            afmFontName = st.sval;
            availableFonts[PDF].put(afmFontName, afmFile);
            break;
        }
        st.nextToken();
    }

    inStream.close();
    return afmFontName;
}

/**
 * Add the font inside the TTF file represented by the passed FontFileLookup
 * object to the list of available fonts by doing a quick parse of the ttf file
 * to determine the name of the font that is contained within.
 * AFM fonts are used for both PDF and RTF output.
 * @param ttfFile a FontFileLookup object referencing an .afm file
 * @return List of names of TTF fonts added.  May be null if no file was found
 *         at the provided location.
 */
protected static List addTTFToAvailableFonts(FontFileLookup ttfFile)
    throws IOException
{
    if (availableFonts[PDF].contains(ttfFile)) {
        return null;
    }

    TTFParser ttfParser = new TTFParser();
    List fontNames = ttfParser.parseAllFontNames(ttfFile);
    Vector returnFontNames = new Vector(fontNames.size());
    if (fontNames != null) {
        for (int i = 0; i < fontNames.size(); i++) {
            TTFFontNames allNames = (TTFFontNames)fontNames.get(i);
            ttfFile.setTTFFontNames(allNames);
            returnFontNames.add(allNames.fontName);
            // TTF fontName is name used by ServeReport internals
            availableFonts[PDF].put(allNames.fontName, ttfFile);
        
            // try to guess some possible user font names that
            // might be used by Java or other programs, map these to the 
            // real font name as used by ServerReport internals
        
            // map TTF family name plus style to TTF fontName
            if (allNames.subFamilyName.equals("Bold")) {
                automaticFontNameMapTable.put(
                    allNames.familyName + "-Bold", allNames.fontName);
            }
            else if (allNames.subFamilyName.equals("Bold Italic") ||
                allNames.subFamilyName.equals("Bold Oblique")) {
                automaticFontNameMapTable.put(
                    allNames.familyName + "-BoldItalic", allNames.fontName);
            }
            else if (allNames.subFamilyName.equals("Italic") ||
                allNames.subFamilyName.equals("Oblique")) {
                automaticFontNameMapTable.put(
                    allNames.familyName + "-Italic", allNames.fontName);
            }
            else {
                // for no style, map both the family name and the full name
                // sometimes they're the same, sometimes different
                automaticFontNameMapTable.put(
                    allNames.familyName, allNames.fontName);
                automaticFontNameMapTable.put(
                    allNames.fullName, allNames.fontName);
            }
        }
    }
    return returnFontNames;
}

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map regardless of the type of the font file.  
 * The font name map must be located in a resource
 * bundle .properties file and is subject to standard resource bundle
 * localization processing.  In other words, the resource bundle that is
 * eventually read by this method is the one that exists at the specified
 * location and whose file name represents the most appropriate bundle for
 * the default locale.  For example, for the specified name of "user" in
 * the North American/English locale, this method will first look for the
 * a resource bundle in the file "user_en_US.properties", then 
 * "user_en.properties" and finally "user.properties".
 * @param location the absolute path of the directory containing the font
 *                 name map resource bundle file(s).
 * @param name the name of the resource bundle.  This name should not
 *             contain any locale-specific information or a file extension,
 *             as this information will be provided by the method.  For
 *             example, names should be of the name "file", not 
 *             "file.properties" or "file_en_US.properties".
 */
public static void addFontNameMap(String location, String name)
    throws IOException
{
    addFontNameMap(location, name, ALL_FILE_TYPES);
}

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map for fonts of the specified file type (TTF or AFM).
 * The font name map must be located in a resource
 * bundle .properties file and is subject to standard resource bundle
 * localization processing.  In other words, the resource bundle that is
 * eventually read by this method is the one that exists at the specified
 * location and whose file name represents the most appropriate bundle for
 * the default locale.  For example, for the specified name of "user" in
 * the North American/English locale, this method will first look for the
 * a resource bundle in the file "user_en_US.properties", then 
 * "user_en.properties" and finally "user.properties".
 * @param location the absolute path of the directory containing the font
 *                 name map resource bundle file(s).
 * @param name the name of the resource bundle.  This name should not
 *             contain any locale-specific information or a file extension,
 *             as this information will be provided by the method.  For
 *             example, names should be of the name "file", not 
 *             "file.properties" or "file_en_US.properties".
 * @param fileType type of font the values in this map should be used
 *                 with.  One of TTF, AFM or NAMED.
 */
public static void addFontNameMap(String location, String name, int fileType)
    throws IOException
{
    FontLibrary.initFontLibrary();

    if (location == null) {
        throw new IllegalArgumentException("location may not be null");
    }

    if (name == null) {
        throw new IllegalArgumentException("name may not be null");
    }

    // set up the file lookup for an absolute path properties file
    FontFileLookup propLook = new FontFileLookup(
        FontFileLookup.FILE_FULL,
        location, name, FontFileLookup.PROPERTIES_FILE);

    addFontNameMapBundle(propLook, fileType);

}//addFontNameMap

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map for fonts of the specified file type (TTF, AFM
 * or NAMED).
 * The font name map must be located in a resource
 * bundle .properties file and is subject to standard resource bundle
 * localization processing.  In other words, the resource bundle that is
 * eventually read by this method is the one that exists at the location
 * specified by the FontFileLookup object and whose file name represents
 * the most appropriate bundle for the default locale.  For example, for
 * a FontFileLookup object referencing the specified name of "user" in
 * the North American/English locale, this method will first look for the
 * a resource bundle in the file "user_en_US.properties", then 
 * "user_en.properties" and finally "user.properties".
 * @param propLook FontFileLookup object representing location of
 *                 resource properties file
 * @param fileType type of font the values in this map should be used
 *                 with.  One of TTF, AFM or NAMED.
 */
protected static void addFontNameMapBundle(
    FontFileLookup propLook, int fileType) throws IOException
{
    // if we've loaded this bundle before, no need to load it again
    if (userFontBundleLookups.contains(propLook)) {
        return;
    }

    // search for the resource bundle using the file lookup object
    ResourceBundle thisBundle = propLook.getResourceBundle();

    if (thisBundle == null) {
        return;
    }

    // if found, add the bundle to the global list of font name map bundles

    // add bundle to lookup list so that we don't load it a second time
    userFontBundleLookups.add(propLook);
    addFontNameMap(thisBundle,fileType);
}

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map regardless of the type of the font file.  
 * The font name map must be located in a resource
 * bundle .properties file and is subject to standard resource bundle
 * localization processing.  In other words, the resource bundle that is
 * eventually read by this method is the one that exists at the specified
 * location and whose file name represents the most appropriate bundle for
 * the default locale.  For example, for the specified name of "user" in
 * the North American/English locale, this method will first look for the
 * a resource bundle in the file "user_en_US.properties", then 
 * "user_en.properties" and finally "user.properties".
 * @param location the absolute path of the directory containing the font
 *                 name map resource bundle file(s) specified as an URL.
 * @param name the name of the resource bundle.  This name should not
 *             contain any locale-specific information or a file extension,
 *             as this information will be provided by the method.  For
 *             example, names should be of the name "file", not 
 *             "file.properties" or "file_en_US.properties".
 */
public static void addFontNameMap(URL location, String name)
    throws IOException
{
    addFontNameMap(location, name, ALL_FILE_TYPES);
}

 /**
  * Add the values contained in the specified font name map to PageLayout's
  * global font name map. Use of the other methods is recommended for better
  * performance, as they try to reduce the number of ResourceBundles stored.
  * @param fontBundle An existing font name map bundle.
  * @param fileType type of font the values in this map should be used
  *                 with.  One of TTF or AFM.
  *  
  */
 public static void addFontNameMap(ResourceBundle fontBundle, int fileType) {
     if (fontBundle != null) {
         // bundle referring to fonts in afm files means bundle is
         // specific to PDF output
         if (fileType == AFM) {
             userFontBundles[PDF].add(fontBundle);
         }
         // bundle referring to fonts in ttf files means bundle is for
         // all output types
         else if (fileType == TTF || fileType == ALL_FILE_TYPES) {
             userFontBundles[PDF].add(fontBundle);
         }
     }
 }
    
/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map for fonts of the specified file type (TTF, AFM
 * or NAMED).
 * The font name map must be located in a resource
 * bundle .properties file and is subject to standard resource bundle
 * localization processing.  In other words, the resource bundle that is
 * eventually read by this method is the one that exists at the specified
 * location and whose file name represents the most appropriate bundle for
 * the default locale.  For example, for the specified name of "user" in
 * the North American/English locale, this method will first look for the
 * a resource bundle in the file "user_en_US.properties", then 
 * "user_en.properties" and finally "user.properties".
 * @param location the absolute path of the directory containing the font
 *                 name map resource bundle file(s) specified as an URL.
 * @param name the name of the resource bundle.  This name should not
 *             contain any locale-specific information or a file extension,
 *             as this information will be provided by the method.  For
 *             example, names should be of the name "file", not 
 *             "file.properties" or "file_en_US.properties".
 * @param fileType type of font the values in this map should be used
 *                 with.  One of TTF, AFM or NAMED.
 */
public static void addFontNameMap(URL location, String name, int fileType)
    throws IOException {
    FontLibrary.initFontLibrary();

    if (location == null) {
        throw new IllegalArgumentException("location may not be null");
    }

    if (name == null) {
        throw new IllegalArgumentException("name may not be null");
    }


    // set up the file lookup for an file-oriented URL properties file
    FontFileLookup propLook = new FontFileLookup(
        FontFileLookup.URL_FILE,
        location, name,
        FontFileLookup.PROPERTIES_FILE);

    // if found, add the bundle to the global list of font name map bundles
    addFontNameMapBundle(propLook, fileType);

}//addFontNameMap

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map regardless of the type of the font file.  
 * The font name map located at the specified URL
 * must be a resource bundle .properties-style stream, but this particular
 * method does not apply any of the standard resource bundle localization
 * processing to the specified URL.  In other words, the resource bundle
 * that is eventually read by this method is exactly the one specified in
 * the passed URL.
 * @param location the absolute path of the desired font name map resource
 *                 bundle file specified as an URL.
 */
public static void addFontNameMap(URL location) throws IOException
{
    addFontNameMap(location, ALL_FILE_TYPES);
}

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map for fonts of the specified file type (TTF, AFM
 * or NAMED).
 * The font name map located at the specified URL
 * must be a resource bundle .properties-style stream, but this particular
 * method does not apply any of the standard resource bundle localization
 * processing to the specified URL.  In other words, the resource bundle
 * that is eventually read by this method is exactly the one specified in
 * the passed URL.
 * @param location the absolute path of the desired font name map resource
 *                 bundle file specified as an URL.
 * @param fileType type of font the values in this map should be used
 *                 with.  One of TTF, AFM or NAMED.
 */
public static void addFontNameMap(URL location, int fileType) throws IOException
{
    FontLibrary.initFontLibrary();

    if (location == null) {
        throw new IllegalArgumentException("location URL may not be null");
    }

    // set up the file lookup for an stream-oriented URL properties file
    FontFileLookup propLook = new FontFileLookup(
        FontFileLookup.URL_STREAM,
        location, "",
        FontFileLookup.PROPERTIES_FILE);

    // if found, add the bundle to the global list of font name map bundles
    addFontNameMapBundle(propLook, fileType);

}//addFontNameMap

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map regardless of the type of the font file.  
 * The font name map must be located in a resource
 * bundle .properties file and is subject to standard resource bundle
 * localization processing.  In other words, the resource bundle that is
 * eventually read by this method is the one that exists at the specified
 * location and whose file name represents the most appropriate bundle for
 * the default locale.  For example, for the specified name of "user" in
 * the North American/English locale, this method will first look for the
 * a resource bundle in the file "user_en_US.properties", then 
 * "user_en.properties" and finally "user.properties".
 * @param location the path of the directory containing the font name map
 *                 resource bundle file(s), specified as a directory
 *                 relative to the classpath.  It is assumed that the
 *                 location begins at the tops of the classpath
 *                 entries.  (In other words, if the location does
 *                 not begin with a '/', one is added.)
 * @param name the name of the resource bundle.  This name should not
 *             contain any locale-specific information or a file extension,
 *             as this information will be provided by the method.  For
 *             example, names should be of the name "file", not 
 *             "file.properties" or "file_en_US.properties".
 */
public static void addRelativeFontNameMap(String location, String name)
    throws IOException 
{
    addRelativeFontNameMap(location, name, ALL_FILE_TYPES);
}

/**
 * Add the values contained in the specified font name map to ServerReport's
 * global font name map for fonts of the specified file type (TTF, AFM
 * or NAMED).
 * The font name map located at the specified URL
 * must be a resource bundle .properties-style stream, but this particular
 * method does not apply any of the standard resource bundle localization
 * processing to the specified URL.  In other words, the resource bundle
 * that is eventually read by this method is exactly the one specified in
 * the passed URL.
 * @param location the absolute path of the desired font name map resource
 *                 bundle file specified as an URL.
 * @param fileType type of font the values in this map should be used
 *                 with.  One of TTF, AFM or NAMED.
 */
public static void addRelativeFontNameMap(
    String location, String name, int fileType) throws IOException 
{
    FontLibrary.initFontLibrary();

    if (location == null) {
        throw new IllegalArgumentException("location may not be null");
    }

    if (name == null) {
        throw new IllegalArgumentException("name may not be null");
    }

    // set up the file lookup for an relative-to-classpath file
    FontFileLookup propLook = new FontFileLookup(
        FontFileLookup.FILE_REL_CLASSPATH,
        location, name, FontFileLookup.PROPERTIES_FILE);

    // if found, add the bundle to the global list of font name map bundles
    addFontNameMapBundle(propLook, fileType);

}//addRelativeFontNameMap


/**
 * Find the actual name of the desired font by looking up the given
 * user alias in the various font name map resource bundles.  The bundles
 * are searched in the following order: user specified font name maps,
 * stock font name maps for the passed output type, automatically
 * font name maps generated from true type files found on the system.
 * If not found in any of the font name maps, see if the name maps
 * directly to a font.  If so, the font name is its own alias.
 * @param key the name used by the user to reference the font
 * @param  outputType the output type currently being used, which
 *                    determines which stock font packages will be searched.
 *                    One of PDF or RTF.
 * @return the actual name of the font as found in its .ttf or .afm file
 */
public static String lookupFontNameByUserAlias(String key, int outputType)
{
    // actual font name
    String fontName = null;

    // try user font name map bundle
    fontName = FontLibrary.lookupUserFontName(key, outputType);

    if (fontName == null) {

        // otherwise try the default bundle
        fontName = FontLibrary.lookupStockFontName(key, outputType);

        if (fontName == null) {

            // otherwise, try the automatic bundles
            fontName = FontLibrary.lookupAutomaticFontName(key);

            if (fontName == null) {
                // font not found in font name map resource bundles

                // first possibility -- no font name map exists, so check to see
                // if a font exists with the name of the key itself (usually
                // only works for fonts with a style (e.g. bold, italic)

                if (availableFonts[outputType].containsKey(key) ||
                    availableStockFonts[outputType].containsKey(key)) {

                    fontName = key;

                    // or check to see if a font exists with
                    // the name of the key plus "-Regular" (a common designation
                    // for a font with a plain style)

                }
                else if (availableFonts[outputType].containsKey(key + "-Regular") ||
                    availableStockFonts[outputType].containsKey(key + "-Regular")) {
                    fontName = key + "-Regular";

                    // if we can't find an appropriate font, just fall back
                    // to the default font, which we know exists
                }
                else {
                    String subFontName = "Times-Roman";
                    if (!getSuppressErrorMessages()) {
                        System.err.println("Warning: font " + key +
                                           " not mapped, " + "substituting " + subFontName);
                    }
                    fontName = FontLibrary.lookupUserFontName(subFontName,
                                                              outputType);
                    if (fontName == null) {
                        fontName = FontLibrary.lookupStockFontName(
                            subFontName, outputType);
                        if (fontName == null) {
                            fontName = new String(subFontName);
                        }
                    }
                }
            }
        }
    }

    return fontName.trim();
}

/**
 * Look up the specified font name in the stock font name map resource bundle
 * associated with the passed output type.  If it is found, return the
 * corresponding font name alias.  If not, return null.
 * @param  fontName the user font name used for lookup
 * @param  outputType the output type currently being used, which
 *                    determines which stock font packages will be searched.
 *                    One of PDF or RTF.
 * @return the alias for the specified font name from the stock font
 *         name map resource bundle
 */
public static String lookupStockFontName(String fontName, int outputType)
{
    FontLibrary.initFontLibrary();

    String alias = null;

    if (stockFontNameMapBundles[outputType] != null) {
        try {
            alias = stockFontNameMapBundles[outputType].getString(fontName);
        } catch (MissingResourceException e2) {
            // font name not found
            // if key has space in it, it may be specified in the
            // bundle with an underscore standing in for the space
            // -- try again
            if (fontName.indexOf(' ') >= 0) {
                try {
                    alias = stockFontNameMapBundles[outputType].
                        getString(fontName.replace(' ', '_'));
                } catch (MissingResourceException e3) {
                    // font name not found -- return null so that the caller
                    // can determine the correct course of action
                    alias = null;
                }
            }
            else {
                // return null so that the caller
                // can determine the correct course of action
                alias = null;
            }
        }
    }

    return alias;
}

/**
 * Look up the specified font name in the automatic font name map.
 * The automatic font name map is a list of possible mappings generated
 * automatically when TrueType fonts are loaded into the FontLibrary.
 * These mappings map possible user aliases to the font's real name,
 * as is used by the internals of ServerReport.
 * @param  fontName the user font name used for lookup
 * @return the alias for the specified font name from the automatic font
 *         name map table
 */
public static String lookupAutomaticFontName(String fontName)
{
    FontLibrary.initFontLibrary();

    String alias = null;

    if (automaticFontNameMapTable != null) {
        alias = (String)automaticFontNameMapTable.get(fontName);
    }

    return alias;
}

/**
 * Look up the specified font name alias in the list of user-supplied font name
 * map resource bundles associated with the specified output type.  If it is
 * found, return the font name that is being aliased.  If not, return null.
 * @param fontName user font name alias
 * @param outputType One of PDF or RTF.  Determines which user
 *                   font bundles are searched for the alias.
 * @return the font name associated with the specified font name alias in
 *         the user-supplied font name map resource bundles for the
 *         specified output type.
 */
public static String lookupUserFontName(String fontName, int outputType)
{
    FontLibrary.initFontLibrary();

    String alias = null;

    if (userFontBundles[outputType] != null) {
        for (int i = 0; i < userFontBundles[outputType].size(); i++) {
            ResourceBundle thisUserFontBundle =
                (ResourceBundle)userFontBundles[outputType].get(i);

            // look up user font name in user bundle
            try {
                alias = thisUserFontBundle.getString(fontName);
            } catch (MissingResourceException e2) {
                // user font name not found
                // if key has space in it, it may be specified in the
                // bundle with an underscore standing in for the space
                // -- try again
                if (fontName.indexOf(' ') >= 0) {
                    try {
                        alias = thisUserFontBundle.getString(
                            fontName.replace(' ', '_'));
                    } catch (MissingResourceException e3) {
                        // alias not found -- return null so that the caller
                        // can determine the correct course of action
                        alias = null;
                    }
                }
                else {
                    // alias not found -- return null so that the caller
                    // can determine the correct course of action
                    alias = null;
                }
            }
            if (alias != null)
                break;
        }
    }

    return alias;
}

/** 
 * Return list of all font name aliases that are used across
 * all output types.
 * @return array of font name aliases used across all output types.
 */
public static String[] getAvailableFontNames()
{
    FontLibrary.initFontLibrary();

    // available name set -- use a TreeSet to keep them in alphabetical order
    TreeSet fontNameSet = new TreeSet();

    // go through stock bundles -- only add font names that are the same
    // across all bundles to the available name set
    int baseIndex = PDF;
    ResourceBundle baseBundle = stockFontNameMapBundles[baseIndex];
    Enumeration baseEnumeration = baseBundle.getKeys();
    while (baseEnumeration.hasMoreElements()) {
        String baseKey = (String)baseEnumeration.nextElement();
        boolean foundInAll = true;
        for (int i = baseIndex + 1; i < NUM_OUTPUT_TYPES; i++) {
            ResourceBundle searchBundle = stockFontNameMapBundles[i];
            try {
                Object value = searchBundle.getObject(baseKey);
            } catch (MissingResourceException mre) {
                foundInAll = false;
                break;
            }
        }
        if (foundInAll) {
            String userKey = baseKey.replace('_', ' ');
            fontNameSet.add(userKey);
        }
    }

    // add in all info from auto font mapping (this gives us the
    // TTF stuff, basically)
    Enumeration autoEnumeration = automaticFontNameMapTable.keys();
    while (autoEnumeration.hasMoreElements()) {
        String autoName = (String)autoEnumeration.nextElement();
        fontNameSet.add(autoName);
    }

    // add in common fonts from output-specific user-added fonts
    List baseBundleList = userFontBundles[baseIndex];
    for (int i = 0; i < baseBundleList.size(); i++) {
        baseBundle =
            (ResourceBundle)baseBundleList.get(i);
        baseEnumeration = baseBundle.getKeys();
        while (baseEnumeration.hasMoreElements()) {
            // for each entry in the base set of bundles, see if it exists
            // in at least one bundle for all of the other output types
            String baseKey = (String)baseEnumeration.nextElement();
            boolean foundInAll = true;
            for (int j = baseIndex + 1; j < NUM_OUTPUT_TYPES && foundInAll; j++) {
                List searchBundleList = userFontBundles[j];
                boolean foundHere = false;
                for (int k = 0; k < searchBundleList.size() && !foundHere; k++) {
                    ResourceBundle searchBundle =
                        (ResourceBundle)searchBundleList.get(k);
                    try {
                        Object value = searchBundle.getObject(baseKey);
                        foundHere = true;
                    } catch (MissingResourceException mre) {
                    }
                }
                if (!foundHere) {
                    foundInAll = false;
                }
            }
            if (foundInAll) {
                String userKey = baseKey.replace('_', ' ');
                fontNameSet.add(userKey);
            }
        }
    }

    // finally, turn the name set into an alphabetical list of strings
    String[] names = new String[fontNameSet.size()];
    Iterator fontNameIterator = fontNameSet.iterator();
    int i = 0;
    while (fontNameIterator.hasNext()) {
        names[i] = (String)fontNameIterator.next();
        i++;
    }

    return names;
}

/** 
 * Return list of all font name aliases specified with respect
 * to the specified output type.
 * @param outputType - only PDF type supported in this release.
 * @return array of font name aliases used with respect to the
 *         specified output type.
 */
public static String[] getAvailableFontNames(int outputType)
{
    FontLibrary.initFontLibrary();

    // available name set -- use a TreeSet to keep them in alphabetical order
    TreeSet fontNameSet = new TreeSet();

    // go through stock bundles -- only add font names that are the same
    // across all bundles to the available name set
    ResourceBundle baseBundle = stockFontNameMapBundles[outputType];
    Enumeration baseEnumeration = baseBundle.getKeys();
    while (baseEnumeration.hasMoreElements()) {
        String baseKey = (String)baseEnumeration.nextElement();
        String userKey = baseKey.replace('_', ' ');
        fontNameSet.add(userKey);
    }

    // add in all info from auto font mapping (this gives us the
    // TTF stuff, basically)
    Enumeration autoEnumeration = automaticFontNameMapTable.keys();
    while (autoEnumeration.hasMoreElements()) {
        String autoName = (String)autoEnumeration.nextElement();
        fontNameSet.add(autoName);
    }

    // add in common fonts from output-specific user-added fonts
    List baseBundleList = userFontBundles[outputType];
    for (int i = 0; i < baseBundleList.size(); i++) {
        baseBundle =
            (ResourceBundle)baseBundleList.get(i);
        baseEnumeration = baseBundle.getKeys();
        while (baseEnumeration.hasMoreElements()) {
            String baseKey = (String)baseEnumeration.nextElement();
            String userKey = baseKey.replace('_', ' ');
            fontNameSet.add(userKey);
        }
    }

    // finally, turn the name set into an alphabetical list of strings
    String[] names = new String[fontNameSet.size()];
    Iterator fontNameIterator = fontNameSet.iterator();
    int i = 0;
    while (fontNameIterator.hasNext()) {
        names[i] = (String)fontNameIterator.next();
        i++;
    }

    return names;
}

/**
 * Lookup the font file (as encapsulated by the FontFileLookup object)
 * by the actual font name, depending on the output type.
 */
public static FontFileLookup lookupFontFile(String fontName, int outputType)
{
    FontFileLookup flook = null;
    flook = (FontFileLookup)availableFonts[outputType].get(fontName);
    if (flook == null) {
        flook = (FontFileLookup)availableStockFonts[outputType].get(fontName);
    }

    return flook;
}

/**
 *  Get the TrueTypeFontProperties class associated with the given user
 *  font name.  If one does not exist, one is created with default values.
 *  The given user font name is mapped to the real font name as defined
 *  in the backing font file, meaning that even if several user aliases
 *  refer to the same font, only a single TrueTypeFontProperties class
 *  will be associated eith the font.
 *  @param userFontName the user font name.  This will be mapped to the
 *                      real font name.
 *  @param outputType   the output type being used.  One of PDF or
 *                      RTF.  This is used to determine which bundle
 *                      stock fonts to look up the font name.
 *  @return the instance of TrueTypeFontProperties associated with the
 *          font name.
 */
public static TrueTypeFontProperties getTrueTypeFontProperties(
    String userFontName, int outputType) 
{
    FontLibrary.initFontLibrary();

    String fontName = lookupFontNameByUserAlias(userFontName, outputType);

    TrueTypeFontProperties ttfProps =
        (TrueTypeFontProperties)trueTypeFontProperties.get(fontName);
    if (ttfProps == null) {
        ttfProps = new TrueTypeFontProperties();
        trueTypeFontProperties.put(fontName, ttfProps);
    }

    return ttfProps;
}

/**
 *  Get the TrueTypeFontProperties class associated with the given real
 *  font name.  If one does not exist, one is created with default values.
 *  @param realFontName the real name of the font this properties class
 *                      is associated with.
 *  @return the instance of TrueTypeFontProperties associated with the
 *          font name.
 */
public static TrueTypeFontProperties lookupTrueTypeFontProperties(
    String realFontName) {
    FontLibrary.initFontLibrary();

    TrueTypeFontProperties ttfProps =
        (TrueTypeFontProperties)trueTypeFontProperties.get(realFontName);
    if (ttfProps == null) {
        ttfProps = new TrueTypeFontProperties();
        trueTypeFontProperties.put(realFontName, ttfProps);
    }

    return ttfProps;
}

/**
 * Check to see if passed file name is an AFM file by examining its extension.
 * @return true if the passed file name is an AFM file;  false otherwise.
 */
public static boolean isAFMFile(String fileName)
{
    if (fileName.endsWith(".afm") || fileName.endsWith(".AFM")) {
        return true;
    }
    else {
        return false;
    }
}

/**
 * Check to see if passed file name is a TrueType file (.ttf, .ttc) by
 * examining its extension.
 * @return true if the passed file name is a TrueType file;  false otherwise.
 */
public static boolean isTrueTypeFile(String fileName)
{
    if (fileName.endsWith(".ttf") || fileName.endsWith(".TTF") ||
        fileName.endsWith(".ttc") || fileName.endsWith(".TTC")) {
        return true;
    }
    else {
        return false;
    }
}

/**
 * Returns the list of font names contained with a TTF file.
 * @param fileLocation the absolute location of the TTF file on
 *                     the filesystem
 * @return a List containing the names (as Strings) for each
 *         font contained within the specified TTF file.  These
 *         names are the ones used internally within ServerReport
 *         to identify the font.
 */
public static List getTTFFontNames(String fileLocation) throws IOException
{
    if (fileLocation == null) {
        return new Vector();
    }

    File fullFile = new File(fileLocation);
    FontFileLookup flook = new FontFileLookup(
        FontFileLookup.FILE_FULL,
        fullFile.getPath(),
        FontFileLookup.TTF_FILE);

    TTFParser ttfParser = new TTFParser();
    List fontNames = ttfParser.parseFontNames(flook);

    return fontNames;
}

/**
 * Returns the list of font names contained with a TTF file.
 * @param fontURL the absolute location of the TTF file as an URL
 * @return a List containing the names (as Strings) for each
 *         font contained within the specified TTF file.  These
 *         names are the ones used internally within ServerReport
 *         to identify the font.
 */
public static List getTTFFontNames(URL fontURL) throws IOException
{
    if (fontURL == null) {
        return new Vector();
    }

    FontFileLookup flook = new FontFileLookup(
        FontFileLookup.URL_STREAM,
        fontURL,
        FontFileLookup.TTF_FILE);

    TTFParser ttfParser = new TTFParser();
    List fontNames = ttfParser.parseFontNames(flook);

    return fontNames;
}

}
