/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCDocument.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:22 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.page;

import java.awt.Graphics;

import java.util.NoSuchElementException;

import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import com.klg.jclass.page.awt.JCAWTPrinter;

/**
 * <code>JCDocument</code> contains a list of pages which are governed by a flow.
 */
public class JCDocument {

/** The name of the document. */
protected String        name = null;

/** The printer for which this document is being prepared. */
protected JCPrinter     printer = null;

/** The set of page templates available to copy for document pages. */
protected List          templatePages;
/** The list of printable pages created in the document. */
protected List          pageList;
/** The number of printable pages created. */
protected int           pageCount = 0;
/** A list of tables which this document is responsible for. */
protected List          tableList;

/** The flow which governs the addition of content to the document. */
protected JCFlow        flow;

/** Indicates whether pages are to be discarded once printed.
 * Default is <code>FLUSH_POLICY_ALWAYS_SAVE</code> (saves all pages).
 */
protected int           flushPolicy;


/** Indicates whether pages are to be held for printing.
 * Default is <code>OUTPUT_POLICY_ON_REQUEST</code> (holds all completed pages
 * in memory until the document is printed).
 */
protected int           outputPolicy;

/** The page number at which the flow should start counting. */
protected int           startPage = 1;

/** For immediate-mode printing, track the last page we successfully printed. */
protected int           lastPage;

// Possible flush policies
/** Always saves (does not flush) all pages (default). */
public static final int FLUSH_POLICY_ALWAYS_SAVE = 1;
/** Flushes pages as they are completed and printed. */
public static final int FLUSH_POLICY_ON_OUTPUT = 2;

// Supported document output modes
/** Holds all completed pages in memory until the document is printed (default). */
public static final int OUTPUT_POLICY_ON_REQUEST = 1;
/** Outputs each page as it is completed (if all predecessors are complete). */
public static final int OUTPUT_POLICY_IMMEDIATE = 2;

/** Blank (no headers or footers) 8.5 x 11 inch page. */
public static final List BLANK_8p5X11 = new BlankTemplate(JCUnit.INCHES,
                                                         8.5, 11.0, 0.75);
/** Blank (no headers or footers) 8.5 x 14 inch page. */
public static final List BLANK_8p5X14 = new BlankTemplate(JCUnit.INCHES,
                                                         8.5, 14, 0.75);
/** Blank (no headers or footers) 11 x 17 inch page. */
public static final List BLANK_11X17 = new BlankTemplate(JCUnit.INCHES,
                                                         11.0, 17, 0.75);
/** Blank (no headers or footers) A3 page. */
public static final List BLANK_A3 = new BlankTemplate(JCUnit.CM,
                                                     29.7, 42, 1.5);
/** Blank (no headers or footers) A4 page. */
public static final List BLANK_A4 = new BlankTemplate(JCUnit.CM,
                                                     21.0, 29.7, 1.5);
/** Blank (no headers or footers) A5 page. */
public static final List BLANK_A5 = new BlankTemplate(JCUnit.CM,
                                                     14.8, 21.0, 1.5);
/** Blank (no headers or footers) 11 x 8.5 inch (landscape) page. */
public static final List BLANK_11X8p5 = new BlankTemplate(JCUnit.INCHES,
                                                          11.0, 8.5, 0.75);

/**
 * Creates a named document.
 * @param name the name to assign to the document
 */
public JCDocument(String name) {
    this(name, null, null);
}

/**
 * Creates a document tied to a given printer.
 * @param printer the printer to use to format to the document
 */
public JCDocument(JCPrinter printer) {
    this(null, printer, null);
}

/**
 * Creates a document tied to a given printer.
 * @param printer the printer to use to format to the document
 * @param templates list of templates to use
 */
public JCDocument(JCPrinter printer, List templates) {
    this(null, printer, templates);
}

/**
 * Creates a document with the specified templates.
 * @param templates a list of templates to be used in the document
 */
public JCDocument(List templates) {
    this(null, null, templates);
}

/**
 * Creates a named document tied to a given printer.
 * @param name the name of the document
 * @param printer the printer to use to format to the document
 * @param templates list of templates to use
 */
private JCDocument(String name, JCPrinter printer, List templates) {

    
 

    
 

    if (name != null) {
        this.name = new String(name);
    }

    // Create the page list
    pageList = new Vector();
    pageCount = 0;

    // Define the default document state
    flushPolicy = FLUSH_POLICY_ALWAYS_SAVE;
    outputPolicy = OUTPUT_POLICY_ON_REQUEST;

    if (templates != null) {
        setTemplates(templates);
    }
    else {
        templatePages = new ArrayList();
    }

    tableList = new Vector();
    flow = null;

    this.printer = printer;
}

/**
 * Gets the name associated with the document.
 * @return the name of the document
 */
public String getName() {
    return name;
}

/**
 * Gets the printer for which this document is being prepared.
 * @return the printer object the document is using to generate the flow
 */
public JCPrinter getPrinter() {
    return printer;
}

/**
 * Sets the printer whose attributes will be used to format the document.
 * @param printer a printer object to associate with this document
 */
public void setPrinter(JCPrinter printer) {
    this.printer = printer;
}

/**
 * Gets the printer for which this document is being prepared.
 * @return the printer object the document is using to generate the flow
 */
public Graphics getPrinterGraphics() {
    return printer.getGraphics();
}

/**
 * Gets the name-to-template mapping the document is using.
 * @return a list of template pages
 */
public List getTemplates() {
    return templatePages;
}

/**
 * Sets the list of templates.
 * @param templates a list of templates
 */
public void setTemplates(List templates) {
    if (templates == null || templates.size() == 0) {
        throw new IllegalArgumentException("templates list may not be null");
    }
    templatePages = templates;
    if (templates != null) {
    // loop through templates and set document ownership
    int count = templates.size();
    for (int i = 0; i < count; i++) {
        JCPage template = (JCPage) templates.get(i);
        template.setDocument(this);
    }
    }
}

/**
 * Cleans up internal objects so <code>JCDocument</code> will not linger.
 * Call before setting document instance to null.
 */
public void dispose() {
    if (templatePages != null) {
    //loop through and destroy
    int count = templatePages.size();
    for (int i = 0; i < count; i++) {
        JCPage template = (JCPage) templatePages.get(i);
        template.setDocument(null);
    }
    templatePages = null;
    }

}



/**
 * Finds the page template that was marked for the first page of the document.
 * @return a template page to use for the document's first page
 */
public JCPage findFirstTemplate() {
    JCPage template = null;
    for (int i = 0; i < templatePages.size(); i++) {
        template = (JCPage) templatePages.get(i);
        if (template.isFirstTemplate() == true) {
            return template;
        }
    }
    return (JCPage) templatePages.get(0);
}

/**
 * Set the page number at which the flow should start counting.  This
 * method must be called after the JCDocument is created but before the
 * JCFlow is created from the document.
 * @param startPage the page number of the first page in the flow
 */
public void setStartingPageNumber(int startPage)
{
    this.startPage = startPage;
}

/**
 * Get the page number at which the flow will start counting.
 * @return the page number of the first page in the flow
 */
public int getStartingPageNumber()
{
    return startPage;
}

/**
 * Adds a page to the document's page list and updates the page numbers
 * being tracked by the flow.
 * @param page a new page to append to the document
 */
public void addPage(JCPage page) {
    addPage(page, true, false);
}

/**
 * Adds a page to the document's page list, optionally updating the page
 * numbers being tracked by the flow.
 * @param page a new page to append to the document
 * @param updateFlowPageNumbers if <code>true</code>, the page numbers being tracked
 *                              by the flow (for example, the current page
 *                              number, the number of pages in a section,
 *                              the number of pages in the document)
 *                              will include the page being added
 *                              with this method; if <code>false</code>, they will not
 * @param isFlowPage <code>true</code> if page is part of current flow (added by methods
 *                   in JCFlow); <code>false</code> if page is not part of flow (added by
 *                   user call). This information is needed to determine
 *                   whether to update page numbers when adding this page:
 *                   now, because new page is the next in the flow, or after
 *                   the next flow page has been created if this new page
 *                   is a non-flow page.
 */
public void addPage(JCPage page, boolean updateFlowPageNumbers,
                                                    boolean isFlowPage) {
    pageList.add(page);
    if (flow != null) {
        // requested to add new page to page count
        if (updateFlowPageNumbers) {
            // user-added, non-flow page -- defer counting it until later
            if (!isFlowPage) {
                flow.deferCountingNewPage(page);
            // new flow page, add it in to the flow-tracked page count
            } else {
                flow.updatePageNumbers(page);
            }
        }
    }
    pageCount++;
    page.setDocument(this);
}

/**
 * Convenience method which adds a package of <i>AFM</i> or <i>TTF</i> fonts to the
 * list of available fonts in the PageLayout <code>FontLibrary</code>.  All <i>.afm</i>, <i>.ttf</i>,
 * and <i>.ttc</i> files found in the specified directory and all of its
 * subdirectories will be added.  The specified directory will
 * be stored for use in reading font files when they become needed.
 * If a font name map resource bundle property file of name "user"
 * and corresponding to the default locale exists in the directory
 * (or its subdirectories), it is also loaded and its contents added
 * to <code>FontLibrary</code>'s global font name map.  (For example, for the
 * North American/English locale, the directory is first searched
 * for the file named "user_en_US.properties", then "user_en.properties"
 * and finally "user.properties".  Font name map files outline the names
 * the user will refer to fonts by the and the names by which they're
 * referred inside the corresponding <i>.afm</i> or <i>.ttf</i> file.) See the
 * <code>FontLibrary</code> class for methods allowing the loading of specific
 * font files and font name map resources bundles.
 * @param packagePath a String representing the absolute location of the
 *                    directory which contains both font files (<i>.afm</i>
 *                    or <i>.ttf</i>) and an optional font name map resource bundle
 *                    property file (for example, "user.properties")
 */
public void addFontPackage(String packagePath) {
    FontLibrary.addFontPackage(packagePath);
}


/**
 * Gets the number of pages currently created within the document.
 * @return the number of document pages created
 */
public int getPageCount() {
    return pageCount;
}

/**
 * Gets the List of pages currently created within the document.
 * @return the List of document pages
 */
public List getPages() {
    return((List) pageList);
}

/**
 * Gets the flow object which is printing to the document.
 * @return the document's flow
 */
public JCFlow getFlow() {
    return flow;
}

/**
 * Sets the flow object which is printing to the document.
 * This method is for internal use only. End users associate
 * a document with a flow when the flow is instantiated.
 * @param flow a flow object which will render to the document
 */
public void setFlow(JCFlow flow) {
    this.flow = flow;
}

/**
 * Gets the document's current policy on flushing completed (printed) pages.
 * @return the document's flush policy
 */
public int getFlushPolicy() {
    return flushPolicy;
}

/**
 * Sets the document's behavior for flushing printed pages.
 * @param policy the policy to apply to completed (output) pages
 */
public void setFlushPolicy(int policy) {

    if (policy <= 0 || policy > FLUSH_POLICY_ON_OUTPUT) {
        throw new IllegalArgumentException("Unrecognised document flush policy");
    }

    flushPolicy = policy;
}

/**
 * Gets the document's current policy on outputting completed (rendered) pages.
 * @return the document's output policy
 */
public int getOutputPolicy() {
    return outputPolicy;
}

/**
 * Sets the document's behavior for outputting printed pages.
 * @param policy the policy to apply to completed (rendered) pages
 */
public void setOutputPolicy(int policy) {

    if (policy <= 0 || policy > OUTPUT_POLICY_IMMEDIATE) {
        throw new IllegalArgumentException("Unrecognised document output policy");
    }

    outputPolicy = policy;
}

/**
 * Returns the named template associated with this document.
 * @param name the name of a page template
 * @return the corresponding template page for this document
 */
public JCPage stringToTemplate(String name) {

    // Look up the name in the set of page templates
    for (Iterator i = templatePages.iterator(); i.hasNext();) {
        JCPage page = (JCPage) i.next();
        if (page.getName().equals(name)) {
            return page;
        }
    }

    // If not found, throw an exception
    throw new NoSuchElementException("Page Template not Found: " + name);
}

/**
 * Gets the list of this document's tables.
 * @return the list of tables owned by this document
 */
public List getTables() {
    return((List) tableList);
}

/**
 * Adds a table to this document's list of owned tables.
 * @param table the table to add to the list
 */
public void addTable(JCPageTable table) {

    tableList.add(table);
}

/**
 * Prints the pages of the document to the current printer.
 */
public void print() {
    print(this.printer);
}

/**
 * Prints the pages of the document.
 * @param printer the printer to output the document to
 */
public void print(JCPrinter printer) {

    int     start_page, end_page, page_count;

    // Compute the bounds of the page range to print
    start_page = printer.getOutputPageStart();
    if (start_page < 0) {
        start_page = 0;
    }

    page_count = pageList.size();
    end_page = printer.getOutputPageEnd();

    // Detect the cases of an unset last page or last page too large
    if (end_page == -1 || end_page >= page_count) {
        // Use the actual last page of the document
        end_page = page_count - 1;
    }

    if (start_page > end_page) {
        return;
    }

    print(printer, start_page, end_page);
}

/**
 * Prints a range of pages from the document to the current printer.
 * @param first the start page of the range of pages to print
 * @param last the end page of the range of pages to print
 */
public void print(int first, int last) {
    print(this.printer, first, last);
}

/**
 * Prints a range of pages from the document.
 * @param printer the printer to output the document to
 * @param first the start page of the range of pages to print
 * @param last the end page of the range of pages to print
 */
public void print(JCPrinter printer, int first, int last) {

    int         current, page_count, pCpy, dCpy, step = 1, copies;
    boolean     collate, reverse;
    //Iterator    pages;
    JCPage      page;
    Rectangle2D pageSize;

    // Complete the document
    if (flow.isFlowActive()) {
        flow.endFlow();
    }

    // if we've been outputting pages in immediate mode
    if (outputPolicy == OUTPUT_POLICY_IMMEDIATE && printer.isDocumentOpen()) {

        // start off with the page size from the last page so we close
        // later with the correct page size
        pageSize = ((JCPage) pageList.get(lastPage < 0 ? 0 : lastPage)).
                                                            getPageSize();

        // output the remaining pages of the document
        while (++lastPage < pageCount) {

            // Get each page of the document in turn
            page = (JCPage) pageList.get(lastPage);
            pageSize = page.getPageSize();

            // Output an appropriate number of copies of the page
            for (pCpy = printer.getNumCopies(); pCpy > 0; pCpy--) {

                // Print the page
                // need to get the gc on each iteration because
                // awt11 printing it changes for each page
                printer.beginPage(pageSize, page.getPageNumber());
                page.draw((Graphics2D) printer.getGraphics());
                printer.endPage(pageSize);
            }

            // if we are flushing pages on output, remove the page
            checkFlushPage(page);
        }

        // close the document and exit
        printer.closeDocument(pageSize);
        return;
    }

    // sanity-check the range of pages
    page_count = pageList.size();
    if (first < 0) {
        first = 0;
    }
    if (last >= page_count) {
        last = page_count - 1;
    }
    if (last < first && page_count > 0) {
        return;
    }

    // Get the setup defined by the printer
    collate = printer.getCollate();
    reverse = printer.getReverseOrder();
    copies = printer.getNumCopies();

    // if the output order is reversed, switch the direction of the range
    if (reverse) {
        int temp = first;
        first = last;
        last = temp;
        step = -1;
    }

    // NOTE:  we may need to do a frame.newLine() or frame.complete() on all
    //        non-flow frames in the document

    // if we use an AWT printer, it needs to have a reference to this document
    if (printer instanceof JCAWTPrinter) {
        ((JCAWTPrinter) printer).registerDocument(this);
        ((JCAWTPrinter) printer).registerFirstLast(first, last);
    }

    // Get the size of the first page
    JCPage firstPage = (JCPage) pageList.get(first);
    pageSize = firstPage.getPageSize();

    // Initialise the document in the printer and get the graphics
    printer.setOrientation(firstPage.getOrientation());
    printer.openDocument(pageSize);

    // special case -- if AWT printer, printing was done while openDocument()
    // was in a wait state, so don't go throught the printing loop.
    if (printer instanceof JCAWTPrinter) {
        copies = -1;
    }

    // NOTE:  This currently doesn't take advantage of any printer
    //        functions for producing multiple copies!

    // Loop through as many collated copies of the document are required
    for (dCpy = 0; ((collate || dCpy == 0) && dCpy < copies); dCpy++) {

        // Loop through the range of pages specified for printing
        for (current = first;
                (step > 0 ? current <= last : current >= last);
                current += step) {

            page = (JCPage) pageList.get(current);
            pageSize = page.getPageSize();
            printer.setOrientation(page.getOrientation());

            // Loop through as many per-page copies (non-collated) as requested
            for (pCpy = 0; (pCpy == 0 || !collate) && pCpy < copies; pCpy++) {

                // Print the page
                // need to get the gc on each iteration because
                // awt11 printing it changes for each page
                printer.beginPage(pageSize, page.getPageNumber());
                page.draw((Graphics2D) printer.getGraphics());
                printer.endPage(pageSize);
            }

            // If this is the last copy of the page, check if we flush it
            if (!collate || (dCpy + 1 == copies)) {
                checkFlushPage(page);
            }
        }
    }

    // Complete the document
    printer.closeDocument(pageSize);
}

/**
 * (internal use only)
 * This class handles immediate-mode printing and page flushing for
 * completed pages, if those options are selected.
 * @param page
 */
public void checkOutputPage(JCPage page) {

    // We don't do anything unless we're outputting in immediate mode
    if (outputPolicy != OUTPUT_POLICY_IMMEDIATE) {
        return;
    }

    // Get the size of the page we are printing
    Rectangle2D pageSize = page.getPageSize();

    // if the printer isn't yet ready to take the document
    if (!printer.isDocumentOpen()) {

        // Initialise the document in the printer
        printer.openDocument(pageSize);

        // Note that we haven't printed any pages of the document yet
        lastPage = -1;
    }

    // Make sure we haven't been blocked by previously uncompleted pages
    int pageIndex = pageList.indexOf(page);
    if (pageIndex != lastPage + 1) {
        return;
    }

    // Output an appropriate number of copies of the page
    for (int pCpy = printer.getNumCopies(); pCpy > 0; pCpy--) {

        // Print the page
        // need to get the gc on each iteration because
        // awt11 printing it changes for each page
        printer.beginPage(pageSize, page.getPageNumber());
        page.draw((Graphics2D) printer.getGraphics());
        printer.endPage(pageSize);
    }

    // if we are flushing pages on output, remove the page
    checkFlushPage(page);

    // update the last printed page
    lastPage = pageIndex;
}

// Check if the page should be deleted.
private void checkFlushPage(JCPage page) {
    if (flushPolicy == FLUSH_POLICY_ON_OUTPUT) {
        page.delete();
    }
}

static private class BlankTemplate extends ArrayList {
    public BlankTemplate(JCUnit unit, double width, double height,
                         double margin) {
        // create a blank template page according to the parameters
        // passed in
        JCPage template = new JCPage("blank", new JCUnit.Point(unit, 0, 0),
                                     new JCUnit.Dimension(unit, width, height));
        template.setFlowSectionTemplate(template);
        template.setFlowPageTemplate(template);

        // don't count this as a real page
        template.setPageType(JCPage.PAGE_TYPE_TEMPLATE);

        // create a frame inset on the page the margin width all around
        JCFrame frame = new JCFrame("blank",
                new JCUnit.Point(unit, margin, margin),
                new JCUnit.Dimension(unit, width - (2 * margin), height - (2 * margin)));

        Vector frames = new Vector(1);
        frames.add(frame);
        template.setFrameList(frames);
        template.setFlowFrameList(frames);
        add(template);
    }
} // end BlankTemplate inner class

}
