/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCPage.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:22 $ $Locker:  $  Quest Software Inc.
package com.klg.jclass.page;

import java.awt.Color;



import com.klg.jclass.page.render.BoxRender;




import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;

import java.util.Vector;
import java.util.List;
import java.util.Iterator;


/**
 * <code>JCPage</code> describes a page in a <code>JCDocument</code>.
 */
public class JCPage extends PageArea {

/** User-controllable page numbering (not an absolute number). */
protected int               pageNumber;
/** List of frames on the page (in order) belonging to the document's flow. */
protected List              flowFrameList;
/** Page template from which to generate the next page (flowing). */
protected JCPage            flowPageTemplate;
/** Next page's template on a section break. */
protected JCPage            flowSectionTemplate;
/** Distinguishes template and document pages. */
protected int               pageType = PAGE_TYPE_DOCUMENT;
/** Name of the template this page was copied from. */
protected String            templateName;
/** Indicates if a template page is marked to be used as the first page. */
protected boolean           firstTemplate = false;
/** Indicates the orientation of the page -- automatic, portrait or landscape. */
protected int orientation = ORIENTATION_AUTOMATIC;

/** A <code>JCPage</code> used as a template page (not counted or printed). */
public static final int PAGE_TYPE_TEMPLATE = 1;
/** A <code>JCPage</code> which is a normal page in a <code>JCDocument</code>. */
public static final int PAGE_TYPE_DOCUMENT = 2;
/** A <code>JCPage</code> which appears in a document, but is not numbered or counted. */
public static final int PAGE_TYPE_TEAR_OFF = 3;
// Need a deleted page marker for flush-on-output
/** A <code>JCPage</code> left as a marker in place of a deleted (flushed) page. */
public static final int PAGE_TYPE_DELETED = 4;
// public static final int PAGE_TYPE_REFERENCE = 5;

/** The orientation of this <code>JCPage</code> is to be determined by its dimensions. */
public static final int ORIENTATION_AUTOMATIC = 0;
/** The orientation of this <code>JCPage</code> is portrait. */
public static final int ORIENTATION_PORTRAIT = 1;
/** The orientation of this <code>JCPage</code> is landscape. */
public static final int ORIENTATION_LANDSCAPE = 2;

/**
 * Creates a blank page with unspecified size and no contents.
 */
public JCPage() {

	super();

	pageNumber = 0;
	flowFrameList = new Vector();
}

/**
 * Creates a blank page in a given document with unspecified size and no contents.
 * @param doc the document to which this page belongs
 */
public JCPage(JCDocument doc) {

	super(doc);

	pageNumber = 0;
	flowFrameList = new Vector();
}

/**
 * Creates a named page with unspecified size and no contents.
 * @param name the name to assign to the page
 */
public JCPage(String name) {

	super(name);

	pageNumber = 0;
	flowFrameList = new Vector();
}

/**
 * Creates in a document a named page with unspecified size and no contents.
 * @param name the name to assign to the page
 * @param doc the document to which this page belongs
 */
public JCPage(String name, JCDocument doc) {

	super(name, doc);

	pageNumber = 0;
	flowFrameList = new Vector();
}

/**
 * Creates a named page with the given size and media offset.
 * @param name the name to assign to the page
 * @param location the offset of the printable page relative to the media
 * @param size the dimensions of the page
 */
public JCPage(String name, JCUnit.Point location, JCUnit.Dimension size) {

	super(name, location, size);

	pageNumber = 0;
	flowFrameList = new Vector();
}

/**
 * Creates in a document a named page with the given size and media offset.
 * @param name the name to assign to the page
 * @param doc the document to which this page belongs
 * @param location the offset of the printable page relative to the media
 * @param size the dimensions of the page
 */
public JCPage(String name, JCDocument doc, JCUnit.Point location,
			  JCUnit.Dimension size) {

	super(name, location, size);

	pageNumber = 0;
	flowFrameList = new Vector();
}

/**
 * Creates a page with the given name and contents copied from a template.
 * @param name the name to assign to the page
 * @param template_page a template page for initial page layout and contents
 */
public JCPage(String name, JCPage template_page) {

	super(name, template_page.location, template_page.size);

	JCFrame     frame, new_frame;
	Iterator	frames;

	pageNumber = 0;
	templateName = new String(template_page.name.toString());
	orientation = template_page.orientation;

	flowPageTemplate = template_page.flowPageTemplate;
	flowSectionTemplate = template_page.flowSectionTemplate;

	// if color is set, copy it
	if (template_page.color != null) {
		this.setColor(template_page.color);
	}

	

	// Copy (deep copy) the frame list
	frames = template_page.frameList.iterator();
	while (frames.hasNext()) {

		frame = (JCFrame) frames.next();

		
		

		// Copy the frame (and subframes)
		new_frame = (JCFrame) frame.clone();
		new_frame.setPage(this);
		this.frameList.add(new_frame);

		
		
	}

	// Replicate (equivalent references into frame list) flowFrameList
	
	frames = template_page.flowFrameList.iterator();
	this.flowFrameList = new Vector();
	while (frames.hasNext()) {
		frame = (JCFrame) frames.next();
		
		new_frame = this.stringToFrame(frame.getName());
		
		this.flowFrameList.add(new_frame);
	}

}

/**
 * Creates a page with the given name and contents copied from a template,
 * and makes the new page owned by the given document.
 * @param name the name to assign to the page
 * @param doc the document which is to own this page
 * @param template_page a template page for initial page layout and contents
 */
public JCPage(String name, JCDocument doc, JCPage template_page) {

	super(name, doc, template_page.location, template_page.size);

	JCFrame     frame, new_frame;
	Iterator	frames;

	pageNumber = 0;
	templateName = new String(template_page.name.toString());
	orientation = template_page.orientation;

	// Copy the flow page and flow section template names
	flowPageTemplate = template_page.flowPageTemplate;
	flowSectionTemplate = template_page.flowSectionTemplate;

	// If color is set, copy it
	if (template_page.color != null) {
		this.setColor(template_page.color);
	}

	

	// Copy (deep copy) the frame list
	frames = template_page.frameList.iterator();
	while (frames.hasNext()) {

		frame = (JCFrame) frames.next();

		
		

		// Copy the frame (and subframes) and assign them to this document
		new_frame = (JCFrame) frame.clone();
		new_frame.setDocument(doc);
		new_frame.setPage(this);
		this.frameList.add(new_frame);

		
		
	}

	// Replicate (equivalent references into frame list) flowFrameList
	
	frames = template_page.flowFrameList.iterator();
	this.flowFrameList = new Vector();
	while (frames.hasNext()) {
		frame = (JCFrame) frames.next();
		
		new_frame = this.stringToFrame(frame.getName());
		
		this.flowFrameList.add(new_frame);
	}

}

/**
 * Gets the number which is to be used to evaluate the pagenumber macro.
 * @return Number which is to be used to label the page
 */
public int getPageNumber() {
	return pageNumber;
}

/**
 * Sets the number which is to be used to evaluate the pagenumber macro.
 * @param page_number number which is to be used to label the page
 */
public void setPageNumber(int page_number) {
	pageNumber = page_number;
}

/**
 * Gets the name of the page template from which the next page is to be cloned.
 * @return name of the template page this page was copied from
 */
public JCPage getFlowPageTemplate() {
	return flowPageTemplate;
}

/**
 * Sets the page template from which the next page is to be cloned.
 * @param template template page
 */
public void setFlowPageTemplate(JCPage template) {
	// NOTE:  perhaps we should verify that the page template exists
	flowPageTemplate = template;
}

/**
 * Gets the page template from which a new page is to be cloned.
 * If the page is terminated with a flow section break,
 * gets the page template from which this page was cloned.
 * @return the template page this page was copied from
 */
public JCPage getFlowSectionTemplate() {
	return flowSectionTemplate;
}

/**
 * Sets the page template from which a new page is to be cloned
 * if the page is terminated with a flow section break.
 * @param template the template page
 */
public void setFlowSectionTemplate(JCPage template) {
	// NOTE:  perhaps we should verify that the page template exists
	flowSectionTemplate = template;
}

/**
 * Gets the list of flow frames on this page.
 * @return ordered list of flow frames this page contains
 */
public List getFlowFrameList() {
	return flowFrameList;
}

/**
 * Sets the frame flow list.
 * @param flow the frames (appearing on this page) which are used by the flow
 */
public void setFlowFrameList(List flow) {

	// ensure that all the frames in the flow are in the page
	Iterator i = flow.iterator();
	while (i.hasNext()) {
		JCFrame frame = (JCFrame) i.next();
		if (frameList.contains(frame) == false) {
			throw new IllegalArgumentException("Frame specified in flow List not in this page");
		}
	}
	this.flowFrameList = flow;
}

/**
 * Finds out if this page template is the first one or not.
 * @return <code>true</code> if this template page is marked to be used first
 */
public boolean isFirstTemplate() {
	return firstTemplate;
}

/**
 * Specifies if this page template is to be used for the first flow page.
 * @param is_first <code>true</code> if this page will be the first template used in the flow
 */
public void setFirstTemplate(boolean is_first) {
	firstTemplate = is_first;
}

/**
 * Gets the name of the page template from which this page was cloned.
 * @return name of the template page this page was copied from
 */
protected String getTemplateName() {
	return templateName;
}

/**
 * Sets the name of the page template from which this page was cloned.
 * @param template_name the unique name of a template page
 */
protected void setTemplateName(String template_name) {
	templateName = template_name;
}


/**
 *  Internal use only.
 *  Gets the page orientation.  Valid values are <code>AUTOMATIC</code> (to be determined
 *  by the page size), <code>PORTRAIT</code>, or <code>LANDSCAPE</code>.
 *  @return the orientation of this page
 */
public int getOrientation() {
	return orientation;
}

/**
 *  Internal use only.
 *  Sets the page orientation.  Valid values are <code>AUTOMATIC</code> (to be determined
 *  by the page size), <code>PORTRAIT</code>, or <code>LANDSCAPE</code>.
 *  @param orient the orientation of this page
 */
protected void setOrientation(int orient) {
	orientation = orient;
}

/**
 * Gets the page type.
 * @return the type of page this is (document, template, tear-off)
 */
public int getPageType() {
	return(pageType);
}

/**
 * Sets the page type.
 * Possible values are currently <code>JCPage.PAGE_TYPE_TEMPLATE</code>
 * and <code>JCPage.PAGE_TYPE_DOCUMENT</code>.
 * @param page_type indicates if this is a document or template page
 */
public void setPageType(int page_type) {

	if (page_type <= 0 || page_type > PAGE_TYPE_TEAR_OFF) {
		throw new IllegalArgumentException("Unrecognised page type");
	}

	// Don't count non-document pages
	if (page_type != PAGE_TYPE_DOCUMENT) {
		pageNumber = 0;
	}

	pageType = page_type;
}

/**
 * Gets the page's size and position and returns
 * the information in the form of a rectangle.
 * @return a <code>rectangle2D</code> specifying the page offset and size
 */
public Rectangle2D getPageSize() {

	double              x, y, width, height;

	// Convert the size and position of the page to points
	x = JCUnit.getAsPoints(location.units, location.x);
	y = JCUnit.getAsPoints(location.units, location.y);
	width = JCUnit.getAsPoints(size.units, size.width);
	height = JCUnit.getAsPoints(size.units, size.height);

	return(new Rectangle2D.Double(x, y, width, height));
}

/**
 * Deletes the contents of the page, leaving a minimal page object.
 * This enables us to flush pages as they are printed, conserving memory.
 */
public void delete() {

	// Discard the frames on the page
	frameList = null;
	flowFrameList = null;

	// We preserve the rest of the page (size, colour, etc.)

	// Mark the page as deleted
	pageType = PAGE_TYPE_DELETED;
}

/**
 * Draws the page.
 * @param gc the graphics object into which we are printing
 */
public void draw(Graphics2D gc) {

	double				x, y, width, height;
	Iterator            frames;
	JCFrame             frame;
	JCDrawStyle         style;
	Graphics2D          gc1;
	JCUnit.Point        location;
	JCUnit.Dimension    frame_size;
	JCUnit.Margins      clip_mask;

	

	// if there is a background colour set, clear the page with it
	if (color != null) {
		

		// Create a GC to draw the background in
		gc1 = (Graphics2D) gc.create();

		if (gc1 instanceof Graphics2DBase) {
			((Graphics2DBase) gc1).setFrameRoot();
		}

		// Set the clip rect on the new graphics
		double w = size.getWidth().getAs(JCUnit.POINTS);
		double h = size.getHeight().getAs(JCUnit.POINTS);
		Rectangle2D.Double rect = new Rectangle2D.Double(0.0, 0.0, w, h);
		gc1.setClip(rect);

		// Create a style for filling in the area's background
		style = new JCDrawStyle();
		style.setFillForegroundColor(color);
		style.makeImmutable();

		// Create a background object and draw it
		if (background == null) {
			background = new BoxRender(new JCUnit.Point(), size, style);
			((BoxRender) background).setFilled(true);
		}
		background.draw(gc1);
		gc1.dispose();
	}

	// Loop through all contained frames and draw them
	

	// We only draw flow frames to markup output devices
	if (gc instanceof Markup) {
		frames = flowFrameList.iterator();
	}
	// otherwise we draw all frames
	else {
		frames = frameList.iterator();
	}

	while(frames.hasNext()) {
		frame = (JCFrame) frames.next();
		

		// Create the gc for the new frame
		location = frame.getLocation();
		x = location.getX().getAs(JCUnit.POINTS);
		y = location.getY().getAs(JCUnit.POINTS);
		frame_size = frame.getSize();
		width = frame_size.getWidth().getAs(JCUnit.POINTS);
		height = frame_size.getHeight().getAs(JCUnit.POINTS);

		// graphics api does not have a create method which takes doubles, so
		// we create a gc then set the clip on it, then translate in doubles
		/*
		 * DO NOT change the way this frame's Graphics is created!!
		 */
		gc1 = (Graphics2D) gc.create();

		/*
		 * We set a marker on Graphics which correspond to document JCFrames
		 * so that they are permitted to nest within one another.
		 */
		if (gc1 instanceof Graphics2DBase) {
			((Graphics2DBase) gc1).setFrameRoot();
		}

		// create the clip rect in doubles
		Rectangle2D.Double r2d = new Rectangle2D.Double(x, y, width, height);

		// set it on the gc
		gc1.setClip(r2d);

		// finally, translate it
		gc1.translate(x,y);

		// Set the clip mask for the new gc
		clip_mask = frame.getClipMask();
		if (clip_mask != null) {
			x = clip_mask.getLeft().getAs(JCUnit.POINTS);
			y = clip_mask.getTop().getAs(JCUnit.POINTS);
			width = clip_mask.getRight().getAs(JCUnit.POINTS);
			height = clip_mask.getBottom().getAs(JCUnit.POINTS);
			Rectangle2D.Double r = new Rectangle2D.Double(x, y, width, height);
			gc1.setClip(r);
		}

		// draw the frame and dispose of the gc
		frame.draw(gc1);
		gc1.dispose();
	}

	
	

    
 
}

}
