/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCTab.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:23 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.page;

/**
 * <code>JCTab</code> objects define tab positions in flowed text.
 */
public class JCTab implements Cloneable {

/** The tab position from the left margin of the frame. */
protected JCUnit.Measure    position;
/** Alignment direction of text to the tab position. */
protected int               tabAlignment;
/** Fill pattern to the left of the tab (tab leader). */
protected int               tabFill;

/*
 * Alignment of text to the tab position.
 */
public static final int TAB_ALIGNMENT_LEFT = 1;
public static final int TAB_ALIGNMENT_RIGHT = 2;
public static final int TAB_ALIGNMENT_CENTER = 3;
public static final int TAB_ALIGNMENT_DECIMAL = 4;

/*
 * Tab fill patterns (tab leader types).
 */
public static final int TAB_FILL_NONE = 1;
public static final int TAB_FILL_UNDERLINE = 2;
public static final int TAB_FILL_DOTS = 3;

/**
 * Creates an unset (left-aligned, no fill, left margin) tab.
 */
public JCTab() {

	this(new JCUnit.Measure(), TAB_ALIGNMENT_LEFT);
}

/**
 * Creates a (left-aligned) tab at the given location.
 * @param location the position for the new tab
 */
public JCTab(JCUnit.Measure location) {

	this(location, TAB_ALIGNMENT_LEFT);
}

/**
 * Creates a tab at the given location with the specified alignment.
 * @param location the position for the new tab
 * @param alignment the alignment of text to the tab
 */
public JCTab(JCUnit.Measure location, int alignment) {

	// Ensure that the tab position is valid
	if (location.isNegative()) {
		throw new IllegalArgumentException("Tab position must be >= 0");
	}

	// Copy the location
	position = (JCUnit.Measure) location.clone();
	tabAlignment = alignment;

	// Assign default values
	tabFill = TAB_FILL_NONE;
}

/**
 * Gets the position of the tab measured from the left margin.
 * @return the horizontal position of the tab
 */
public JCUnit.Measure getPosition() {
	return position;
}

/**
 * Specifies the position of the tab measured from the left margin.
 * @param position the horizontal position of the tab
 */
public void setPosition(JCUnit.Measure position) {

	// Verify the position is >= 0
	if (position.isNegative()) {
		throw new IllegalArgumentException("Tab position must be >= 0");
	}

	this.position = position;
}

/**
 * Gets the alignment of text to the tab.
 * @return mode of alignment of text to this tab
 */
public int getTabAlignment() {
	return tabAlignment;
}

/**
 * Specifies how text is aligned to the tab position (right, left, center, or decimal).
 * @param alignment method for aligning text to the tab position
 */
public void setTabAlignment(int alignment) {

	if (alignment <= 0 || alignment > TAB_ALIGNMENT_DECIMAL) {
		throw new IllegalArgumentException("Unknown tab alignment");
	}

	this.tabAlignment = alignment;
}

/**
 * Gets the tab leader fill pattern.
 * @return the pattern used to draw the leader line of a tab
 */
public int getTabFill() {
	return tabFill;
}

/**
 * Sets the tab leader fill pattern (none, underline, dots).
 * @param fill_type the pattern used to draw the leader line of a tab
 */
public void setTabFill(int fill_type) {

	if (fill_type <= 0 || fill_type > TAB_FILL_DOTS) {
		throw new IllegalArgumentException("Unknown tab fill type");
	}

	this.tabFill = fill_type;
}

/**
 * Compares two tab positions.
 * @param new_tab a tab to compare to this tab
 * @return negative, zero, or positive corresponding to less, equal, or greater
 */
public int comparePosition(JCTab new_tab) {

	double  difference;

	difference = position.internal() - new_tab.position.internal();
	if (difference < 0) {
		return(-1);
	}
	else if (difference > 0) {
		return(1);
	}
	else {
		return(0);
	}
}

/**
 * Ensure that two <code>JCTab</code> objects with identical parameters returns <code>true</code>
 * when tested for equality.
 * @return
 * @param o
 */
public boolean equals(Object o)
{
    if (this == o) {
        return true;
    }
    if (!(o instanceof JCTab)) {
        return false;
    }

    final JCTab jcTab = (JCTab) o;

    if (tabAlignment != jcTab.tabAlignment) {
        return false;
    }
    if (tabFill != jcTab.tabFill) {
        return false;
    }
    if (position != null ? !position.equals(jcTab.position) : jcTab.position != null) {
        return false;
    }

    return true;
}

/**
 * Ensure that two <code>JCTab</code> objects with identical parameters have the same
 * hash code.
 * @return
 */
public int hashCode()
{
    int result;
    result = (position != null ? position.hashCode() : 0);
    result = 29 * result + tabAlignment;
    result = 29 * result + tabFill;
    return result;
}

/**
 * Creates a new tab which is a copy of this one.
 * @return an identical copy of this tab
 */
public Object clone() {

	JCTab   new_tab;

	// Clone the tab
	try {
		new_tab = (JCTab) super.clone();
	}
	catch (CloneNotSupportedException e) {
		return(null);
	}

	// Clone the position of the tab
	new_tab.position = (JCUnit.Measure) this.position.clone();

	// Return the newly created tab copy
	return(new_tab);
}

}
