/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCUnit.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:23 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.page;

/**
 * <code>JCUnit</code> defines the supported linear units of measurement:
 * inches, centimeters, and points.
 */
public final class JCUnit {

public static final JCUnit      INCHES = new JCUnit("inches");
public static final JCUnit      CM = new JCUnit("centimetres");
public static final JCUnit      POINTS = new JCUnit("points");
public static final JCUnit      INTERNAL = INCHES;
private static JCUnit           defaultUnit = INCHES;

public static final double INCHES_TO_CM_FACTOR = 2.54;
public static final double POINTS_TO_CM_FACTOR = 72.0 / INCHES_TO_CM_FACTOR;

/** List of available built-in units. */
public static final JCUnit UNITS[] = {
	INCHES, CM, POINTS
};

protected String name;

/**
 * A convenience method that creates an array of measures from the specified
 * array of doubles.
 * @param doubles
 * @param unit
 * @return
 */
public static final Measure [] doublesToMeasures(JCUnit unit, double doubles[]) {
	Measure measures [] = new Measure[doubles.length];
	for (int i = 0; i < doubles.length; i++) {
		measures[i] = new Measure(unit, doubles[i]);
	}
	return measures;
}

/**
 * Creates a new unit instance.
 * @param name
 */
public JCUnit(String name) {
	this.name = name;
}

public String toString() {
	return name;
}

/**
 * Creates a hash that generates the same value depending on the
 * property values.
 * @return
 */
public int hashCode() {
    return (name != null ? name.hashCode() : 0);
}

public boolean equals(Object o) {
    if (this == o) {
        return true;
    }
    if (!(o instanceof JCUnit)) {
        return false;
    }

    final JCUnit jcUnit = (JCUnit) o;

    if (name != null ? !name.equals(jcUnit.name) : jcUnit.name != null) {
        return false;
    }

    return true;
}

/**
 * Derives a hash value for a measure that will return the same hash value
 * when <code>areProbablyEqual()</code> returns <code>true</code>.
 * @return
 * @param baseMeasure
 */
protected static int hashMeasure(double baseMeasure )
{
    int hashcode = (int) baseMeasure;
    double piece = baseMeasure - (double) hashcode;
    hashcode += (int) (piece * 100000000.0);
    return hashcode;
}

/**
 * Returns <code>true</code> if two numbers are within an accepted tolerance.
 * <br><B>Note:</b> This tolerance must correspond to the value used in
 * <code>hashMeasure</code>, otherwise two objects that are equal will not return the same hashcode.
 * @return <code>true</code> if two numbers are within an accepted tolerance
 * @param value1
 * @param value2
 */
public static boolean areProbablyEqual(double value1, double value2) {
	if (Math.abs(value1 - value2) < 0.00000001) {
		return true;
	}
	else {
		return false;
	}
}

/**
 * Gets the <code>JCUnit</code> which specifies the interpretation of all numeric distances.
 * @return current global units setting
 */
public static final JCUnit getDefaultUnits() {
	return defaultUnit;
}

/**
 * Sets the <code>JCUnit</code> which specifies the interpretation of all numeric distances.
 * @param new_default unit of measure to be applied globally
 */
public static final void setDefaultUnit(JCUnit new_default) {
	defaultUnit = new_default;
}

/**
 * Converts a given distance in the given units to a distance in centimeters.
 * @return the equivalent distance in centimeters
 * @param units
 * @param measure
 */
public static final double getAsCentimeters(JCUnit units, double measure) {
	if (units == CM) {
		return measure;
	}
	else if (units == INCHES) {
		return measure * INCHES_TO_CM_FACTOR;
	}
	else if (units == POINTS) {
		return measure / POINTS_TO_CM_FACTOR;
	}
	else {
		/* Throw an exception */
		throw new IllegalArgumentException("Unrecognised Unit type");
	}
}

/**
 * Converts a given distance, given as a <code>JCUnit</code>, to a distance in internal units.
 * @return the equivalent distance in internal units
 * @param measure
 * @param units
 */
public static final double getIU(JCUnit units, double measure) {
	if (units == CM) {
		return measure / INCHES_TO_CM_FACTOR;
	}
	else if (units == INCHES) {
		return measure;
	}
	else if (units == POINTS) {
		return measure / 72.0;
	}
	else {
		/* Throw an exception */
		throw new IllegalArgumentException("Unrecognised Unit type");
	}
}

/**
 * Converts a given distance in the given units to a distance in inches.
 * @return the equivalent distance in inches
 * @param units
 * @param measure
 */
public static final double getAsInches(JCUnit units, double measure) {
	return JCUnit.getIU(units, measure);
}

/**
 * Converts a given distance in the given units to a distance in points.
 * @return the equivalent distance in points
 * @param units
 * @param measure
 */
public static final double getAsPoints(JCUnit units, double measure) {
	if (units == CM) {
		return measure * POINTS_TO_CM_FACTOR;
	}
	else if (units == INCHES) {
		return measure * 72.0;
	}
	else if (units == POINTS) {
		return measure;
	}
	else {
		/* Throw an exception */
		throw new IllegalArgumentException("Unrecognised Unit type");
	}
}

public static final double getAs(JCUnit to, JCUnit from, double measure) {

	if (to == from) {
		return measure;
	}

	if (to == CM) {
		return getAsCentimeters(from, measure);
	}
	else if (to == INCHES) {
		return getAsInches(from, measure);
	}
	else if (to == POINTS) {
		return getAsPoints(from, measure);
	}
	else {
		/* Throw an exception */
		throw new IllegalArgumentException("Unrecognised Unit type");
	}
}


/**
 * Measure encapsulates a linear distance with its units of measurement.
 */
public static class Measure implements Cloneable {

	/** <code>JCUnit</code> object representing the units of linear measurement. */
	public JCUnit    units;
	/** The magnitude of the distance being represented. */
	public double    distance;

	/**
	 * Creates a new measure using the current default units (magnitude 0).
	 */
	public Measure() {
		this.units = JCUnit.getDefaultUnits();
	}

	/**
	 * Creates a new measure using the given units (magnitude 0).
	 * @param units the units which define the distance represented
	 */
	public Measure(JCUnit units) {
		this.units = units;
	}

	/**
	 * Creates a new measure given a distance in the default units.
	 * @param distance a length measured in the current default units
	 */
	public Measure(double distance) {
		this.units = JCUnit.getDefaultUnits();
		this.distance = distance;
	}

	/**
	 * Creates a new measure given a distance and its units of measurement.
	 * @param units the units in which the length is computed
	 * @param distance a length measured in the given units
	 */
	public Measure(JCUnit units, double distance) {
		this.units = units;
		this.distance = distance;
	}

	/** Returns the distance and units as a String.
     * @return distance and units
     */
	public String toString() {
		return distance + " " + units;
	}

	/**
	 * Creates a clone of this Measure.
	 * @return a new Measure, which is a clone
	 */
	public Object clone() {
		try {
			return(super.clone());
		}
		catch (CloneNotSupportedException e) {
			return(null);
		}
	}

	/**
	 * Gets the represented distance in different units.
	 * @param new_units units in which to represent the distance
	 * @return distance converted to the new units
	 */
	public double getAs(JCUnit new_units) {
		if (new_units == units) {
			return distance;
		}
		else {
			if (new_units == JCUnit.CM) {
				return JCUnit.getAsCentimeters(units, distance);
			}
			else if (new_units == JCUnit.INCHES) {
				return JCUnit.getAsInches(units, distance);
			}
			else if (new_units == JCUnit.POINTS) {
				return JCUnit.getAsPoints(units, distance);
			}
			else {
				throw new IllegalArgumentException("Unrecognised Unit type.");
			}
		}
	}

	/**
	 * Gets the distance specified by this measure in internal units.
	 * @return the distance in internal units
	 */
	public double internal() {
		return(JCUnit.getIU(units, distance));
	}

	/**
	 * Sets this measure equal to the given one.
	 * @param newValue the value to assign to this measure
	 */
	public void assign(JCUnit.Measure newValue) {
		this.units = newValue.units;
		this.distance = newValue.distance;
	}

	/**
	 * Adds a given measure to the current measure.
	 * @param dist a measure whose distance is to be added to this one
	 */
	public void add(JCUnit.Measure dist) {
		this.distance += dist.getAs(this.units);
	}

	/**
	 * Subtracts a given measure from the current measure.
	 * @param dist a measure to be subtracted from this one
	 */
	public void subtract(JCUnit.Measure dist) {
		this.distance -= dist.getAs(this.units);
	}

	/**
	 * Multiplies this measure by a given factor.
	 * @param factor the amount by which to multiply this measure
	 */
	public void multiply(double factor) {
		this.distance *= factor;
	}

	/**
	 * Divides this measure by a given factor.
	 * @param factor the amount by which to divide this measure
	 */
	public void divide(double factor) {
		this.distance /= factor;
	}

    /**
     * Creates a hash that generates the same value depending on the
     * property values.
     * @return
     */
    public int hashCode() {
        return hashMeasure(getAsPoints(units, distance));
    }

	/**
	 * Tests the equality of an object to this Measure.
	 * @param obj an object to compare with this Measure
	 * @return <code>true</code> if both objects are Measures and equal in magnitude
	 */
	public boolean equals(Object obj) {
        //safe optimization
        if (this == obj) {
            return true;
        }

		// We can only successfully compare to another measure
		if (!(obj instanceof JCUnit.Measure)) {
			return(false);
		}

        // Compare magnitude of the Measures after converting to a common base
        // Use POINTS for common base, since it is the finest unit of measure
        // we have.  (All comparisons must use same base units; otherwise
        // in a case where two measure objects have different base units,
        // one could equal the other but not vice versa.)
        JCUnit.Measure m = (JCUnit.Measure) obj;
        return(areProbablyEqual(getAsPoints(units, distance),
                                 getAsPoints(m.units, m.distance)));
	}

	/**
	 * Compares another measure to this one.
	 * @param measure a measure to compare for size with this one
	 * @return <code>true</code> if this measure is larger than the given one
	 */
	public boolean greaterThan(Measure measure) {
        // Use common base for comparisons
        if (areProbablyEqual(getAsPoints(units, distance),
                                getAsPoints(measure.units, measure.distance))) {
            return false;
        }
        else {
            return (getAsPoints(units, distance) >
                                getAsPoints(measure.units, measure.distance));
		}
	}

	/**
	 * Compares another measure to this one.
	 * @param measure a measure to compare for size with this one
	 * @return <code>true</code> if this measure is smaller than the given one
	 */
	public boolean lessThan(Measure measure) {
        // Use common base for comparisons
        if (areProbablyEqual(getAsPoints(units, distance),
                                getAsPoints(measure.units, measure.distance))) {
            return false;
        }
        else {
            return(getAsPoints(units, distance) <
                        getAsPoints(measure.units, measure.distance));
        }
	}

	/**
	 * Determines if this measure is equal to zero.
	 * @return <code>true</code> if this measure has zero magnitude
	 */
	public boolean isZero() {
		return(areProbablyEqual(this.distance, 0.0));
	}

	/**
	 * Determines if this measure is less than zero.
	 * @return <code>true</code> if this measure has negative magnitude
	 */
	public boolean isNegative() {
		if (areProbablyEqual(this.distance, 0.0)) {
			return false;
		}
		else {
			return (this.distance < 0.0);
		}
	}
} // Measure


/**
 * Point is roughly equivalent to <i>java.awt.Point</i>, but the coordinates are
 * stored as double-precision numbers and the units of measurement are
 * recorded as well.
 */
public static class Point implements Cloneable {

	/** <code>JCUnit</code> object representing the units of linear measurement. */
	public JCUnit    units;
	/** The coordinates of the point being represented. */
	public double    x, y;

	/**
	 * Creates a point at (0, 0) using the default units.
	 */
	public Point() {
		this.units = JCUnit.getDefaultUnits();
	}

	/**
	 * Creates a point at (0, 0) using the given units.
	 * @param units the units in which the point is represented
	 */
	public Point(JCUnit units) {
		this.units = units;
	}

	/**
	 * Creates a point at (<i>x</i>, <i>y</i>) using the current units.
	 * @param x the <i>x</i>-coordinate of the point (measured in units)
	 * @param y the <i>y</i>-coordinate of the point (measured in units)
	 */
	public Point(double x, double y) {
		this.units = JCUnit.getDefaultUnits();
		this.x = x;
		this.y = y;
	}

	/**
	 * Creates a point at (<i>x</i>, <i>y</i>) using the given units.
	 * @param units the units in which the point is represented
	 * @param x the <i>x</i>-coordinate of the point (measured in units)
	 * @param y the <i>y</i>-coordinate of the point (measured in units)
	 */
	public Point(JCUnit units, double x, double y) {
		this.units = units;
		this.x = x;
		this.y = y;
	}

	/**
	 * Creates a clone of this Point.
	 * @return a new Point, which is a clone
	 */
	public Object clone() {

		try {
			return(super.clone());
		}
		catch (CloneNotSupportedException e) {
			return(null);
		}
	}

	/**
	 * Returns the <i>x</i> position as a measure.
	 * @return the <i>x</i>-coordinate in the units used by the Point
	 */
	public Measure getX() {
		return(new JCUnit.Measure(units, x));
	}

	/**
	 * Sets the value of the <i>x</i> coordinate based on a measure.
	 * @param newX a Measure specifying the new <i>x</i> position of the Point
	 */
	public void setX(Measure newX) {
		x = newX.getAs(units);
	}

	/**
	 * Gets the <i>x</i> position as a distance in internal units.
	 * @return the <i>x</i>-coordinate measured in internal units
	 */
	public double xInt() {
		return(JCUnit.getIU(units, x));
	}

	/**
	 * Returns the <i>y</i> position as a measure.
	 * @return the <i>y</i>-coordinate in the units used by the Point
	 */
	public Measure getY() {
		return(new JCUnit.Measure(units, y));
	}

	/**
	 * Sets the value of the <I>y</I> coordinate based on a measure.
	 * @param newY a Measure specifying the new <i>y</i> position of the Point
	 */
	public void setY(Measure newY) {
		y = newY.getAs(units);
	}

	/**
	 * Gets the <i>y</i> position as a distance in internal units.
	 * @return the <i>y</i>-coordinate measured in internal units
	 */
	public double yInt() {
		return(JCUnit.getIU(units, y));
	}

	/**
	 * Sets this point equal to the given one.
	 * @param newValue the value to assign to this point
	 */
	public void assign(JCUnit.Point newValue) {
		this.units = newValue.units;
		this.x = newValue.x;
		this.y = newValue.y;
	}

    /**
     * Creates a hash that generates the same value depending on the
     * property values.
     * @return
     */
    public int hashCode() {
        // convert to common base and iron out "probably equals" inconsistencies
        int hashx = hashMeasure(getAsPoints(units, x));
        int hashy = hashMeasure(getAsPoints(units, y));

        // same algorithm as java.lang.awt.Point
        long bits = java.lang.Double.doubleToLongBits(hashx);
        bits ^= java.lang.Double.doubleToLongBits(hashy) * 31;
        return (((int) bits) ^ ((int) (bits >> 32)));
    }

	/**
	 * Compares an object to this Point.
	 * @param obj an Object to test for equality with this Point
	 * @return <code>true</code> if both objects are points and they are congruent
	 * 				(independent of units)
	 */
	public boolean equals(Object obj) {
        //safe optimization
        if (this == obj) {
            return true;
        }

		// Both objects must be Point objects
		if (!(obj instanceof JCUnit.Point)) {
			return(false);
		}

        // compare the co-ordinates of the Points in a common base
        JCUnit.Point p = (JCUnit.Point) obj;

        return(areProbablyEqual(getAsPoints(units, x),
                                getAsPoints(p.units, p.x)) &&
               areProbablyEqual(getAsPoints(units, y),
                                getAsPoints(p.units, p.y)));
	}

	/** Returns the point co-ordinates and units as a String.
     * @return point co-ordinates and units
     */
	public String toString() {
		return("(" + x + ", " + y + ") " + units);
	}

} // Point


/**
 * Dimension is similar in concept to <i>java.awt.Dimension</i>, but the values are
 * stored as double-precision numbers and the units of measurement are
 * recorded as well.
 */
public static class Dimension implements Cloneable {

	/** <code>JCUnit</code> object representing the units of linear measurement. */
	public JCUnit    units;
	/** The magnitude of the width being represented. */
	public double    width;
	/** The magnitude of the height being represented. */
	public double    height;

	/**
	 * Creates zero dimensions based on the current units.
	 */
	public Dimension() {
		this.units = JCUnit.getDefaultUnits();
	}

	/**
	 * Creates zero dimensions based on the given units.
	 * @param units the units to apply to the new dimensions
	 */
	public Dimension(JCUnit units) {
		this.units = units;
	}

	/**
	 * Creates width and height dimensions measured in the current units.
	 * @param width the magnitude of the horizontal dimension in current units
	 * @param height the magnitude of the vertical dimension in current units
	 */
	public Dimension(double width, double height) {
		this.units = JCUnit.getDefaultUnits();
		this.width = width;
		this.height = height;
	}

	/**
	 * Creates width and height dimensions measured in the given units.
	 * @param units the units to apply to the new dimensions
	 * @param width the magnitude of the horizontal dimension in units
	 * @param height the magnitude of the vertical dimension in units
	 */
	public Dimension(JCUnit units, double width, double height) {
		this.units = units;
		this.width = width;
		this.height = height;
	}

	/**
	 * Creates a clone of this Dimension.
	 * @return a new Dimension, which is a clone
	 */
	public Object clone() {

		try {
			return(super.clone());
		}
		catch (CloneNotSupportedException e) {
			return(null);
		}
	}

	/**
	 * Returns the width of the dimension as a measure.
	 * @return the width in the units used by the Dimension
	 */
	public Measure getWidth() {
		return(new JCUnit.Measure(units, width));
	}

	/**
	 * Sets the width of the dimension from a measure.
	 * @param newWidth the measure giving the Dimension's new width
	 */
	public void setWidth(Measure newWidth) {
		width = newWidth.getAs(units);
	}

	/**
	 * Gets the width defined by the object in internal units.
	 * @return the width converted to internal units
	 */
	public double widthInt() {
		return(JCUnit.getIU(units, width));
	}

	/**
	 * Returns the height of the dimension as a measure.
	 * @return the height in the units used by the Dimension
	 */
	public Measure getHeight() {
		return(new JCUnit.Measure(units, height));
	}

	/**
	 * Sets the height of the dimension from a measure.
	 * @param newHeight the measure giving the Dimension's new height
	 */
	public void setHeight(Measure newHeight) {
		height = newHeight.getAs(units);
	}

	/**
	 * Gets the height defined by the object in internal units.
	 * @return the height converted to internal units
	 */
	public double heightInt() {
		return(JCUnit.getIU(units, height));
	}

	/**
	 * Sets this dimension equal to the given one.
	 * @param newValue the value to assign to this dimension
	 */
	public void assign(JCUnit.Dimension newValue) {
		this.units = newValue.units;
		this.width = newValue.width;
		this.height = newValue.height;
	}

    /**
     * Creates a hash that generates the same value depending on the
     * property values.
     * @return
     */
    public int hashCode() {
        // convert to common base and iron out "probably equals" inconsistencies
        int hashWidth = hashMeasure(getAsPoints(units, width));
        int hashHeight = hashMeasure(getAsPoints(units, height));

        // same algorithm as java.lang.awt.Dimension
        int sum = hashWidth + hashHeight;
        return sum * (sum + 1)/2 + hashWidth;
    }

	/**
	 * Compares an Object to this Dimension.
	 * @param obj an Object to compare for equality with this Dimension
	 * @return <code>true</code> if both objects are Dimensions and the two Dimensions
	 * 				are identical (independent of units)
	 */
	public boolean equals(Object obj) {
        //safe optimization
        if (this == obj) {
            return true;
        }

		// The objects can't be equal unless they are both Dimensions
		if (!(obj instanceof JCUnit.Dimension)) {
			return(false);
		}

        // compare the co-ordinates of the Dimensions in a common base
        JCUnit.Dimension d = (JCUnit.Dimension) obj;
        return(areProbablyEqual(getAsPoints(units, width),
                                getAsPoints(d.units, d.width)) &&
               areProbablyEqual(getAsPoints(units, height),
                                getAsPoints(d.units, d.height)));
	}

	/** Returns the dimension values and units as a String.
     * @return dimension values
     */
	public String toString() {
		return(width + "x" + height + " " + units);
	}

} // Dimension



/**
 * Stores the sizes of the margins around a figure or area.
 * Margin widths are stored as double-precision numbers along with the
 * units of measurement in which they are specified.
 */
public static class Margins implements Cloneable {

	/** <code>JCUnit</code> object representing the units of linear measurement. */
	public JCUnit    units;
	/** The widths of the top margin in the given units. */
	public double    top;
	/** The widths of the bottom margin in the given units. */
	public double    bottom;
	/** The widths of the left margin in the given units. */
	public double    left;
	/** The widths of the right margin in the given units. */
	public double    right;

	/**
	 * Creates a set of zero margins using the current units.
	 */
	public Margins() {
		this.units = JCUnit.getDefaultUnits();
	}

	/**
	 * Creates a set of zero margins using the specified units.
	 * @param units the units the margin distances are measured in
	 */
	public Margins(JCUnit units) {
		this.units = units;
	}

	/**
	 * Creates the margins for a region given the distances in current units.
	 * @param top the width of the top margin
	 * @param bottom the width of the bottom margin
	 * @param left the width of the left margin
	 * @param right the width of the right margin
	 */
	public Margins(double top, double bottom, double left, double right) {

		this.units = JCUnit.getDefaultUnits();
		this.top = top;
		this.bottom = bottom;
		this.left = left;
		this.right = right;
	}

	/**
	 * Creates the margins for a region given the distances and the units.
	 * @param units the units the margin distances are measured in
	 * @param top the width of the top margin
	 * @param bottom the width of the bottom margin
	 * @param left the width of the left margin
	 * @param right the width of the right margin
	 */
	public Margins(JCUnit units, double top, double bottom,
					 double left, double right) {

		this.units = units;
		this.top = top;
		this.bottom = bottom;
		this.left = left;
		this.right = right;
	}

	/**
	 * Creates a clone of this Margins.
	 * @return a new Margins, which is a clone
	 */
	public Object clone() {
		try {
			return(super.clone());
		}
		catch (CloneNotSupportedException e) {
			return(null);
		}
	}

	/**
	 * Returns the top margin width as a measure.
	 * @return the width of the top margin in the units used by these Margins
	 */
	public Measure getTop() {
		return(new JCUnit.Measure(units, top));
	}

	/**
	 * Sets the top margin width given a measure.
	 * @param newTop the width to set for the top margin
	 */
	public void setTop(Measure newTop) {
		top = newTop.getAs(units);
	}

	/**
	 * Returns the bottom margin width as a measure.
	 * @return the width of the bottom margin in the units used by these Margins
	 */
	public Measure getBottom() {
		return(new JCUnit.Measure(units, bottom));
	}

	/**
	 * Sets the bottom margin width given a measure.
	 * @param newBottom the width to set for the bottom margin
	 */
	public void setBottom(Measure newBottom) {
		bottom = newBottom.getAs(units);
	}

	/**
	 * Returns the left margin width as a measure.
	 * @return the width of the left margin in the units used by these Margins
	 */
	public Measure getLeft() {
		return(new JCUnit.Measure(units, left));
	}

	/**
	 * Sets the left margin width given a measure.
	 * @param newLeft the width to set for the left margin
	 */
	public void setLeft(Measure newLeft) {
		left = newLeft.getAs(units);
	}

	/**
	 * Returns the right margin width as a measure.
	 * @return the width of the right margin in the units used by these Margins
	 */
	public Measure getRight() {
		return(new JCUnit.Measure(units, right));
	}

	/**
	 * Sets the right margin width given a measure.
	 * @param newRight the width to set for the right margin
	 */
	public void setRight(Measure newRight) {
		right = newRight.getAs(units);
	}

	/**
	 * Sets these margins equal to the given ones.
	 * @param newValue the values to assign to these margins
	 */
	public void assign(JCUnit.Margins newValue) {
		this.units = newValue.units;
		this.left = newValue.left;
		this.right = newValue.right;
		this.top = newValue.top;
		this.bottom = newValue.bottom;
	}

    /**
     * Creates a hash that generates the same value depending on the
     * property values.
     * @return
     */
    public int hashCode() {
        // convert to common base and iron out "probably equals" inconsistencies
        int hashLeft = hashMeasure(getAsPoints(units, left));
        int hashRight = hashMeasure(getAsPoints(units, right));
        int hashTop = hashMeasure(getAsPoints(units, top));
        int hashBottom = hashMeasure(getAsPoints(units, bottom));

        // same algorithm as java.lang.awt.Insets
        int sum1 = hashLeft + hashBottom;
        int sum2 = hashRight + hashTop;
        int val1 = sum1 * (sum1 + 1)/2 + hashLeft;
        int val2 = sum2 * (sum2 + 1)/2 + hashTop;
        int sum3 = val1 + val2;
        return sum3 * (sum3 + 1)/2 + val2;
    }

	/**
	 * Compares another Object to this Margins.
	 * @param obj an object to compare to this Margins object for equality
	 * @return <code>true</code> if both objects are Margins and the size of the Margins
	 * 				are identical (independent of units)
	 */
	public boolean equals(Object obj) {
        //safe optimization
        if (this == obj) {
            return true;
        }

		// Make sure we're comparing to another Margins object
		if (!(obj instanceof JCUnit.Margins)) {
			return(false);
		}

		// compare the size of the margin sides in a common base
		JCUnit.Margins m = (JCUnit.Margins) obj;
        return(areProbablyEqual(getAsPoints(units, left),
                                getAsPoints(m.units, m.left)) &&
               areProbablyEqual(getAsPoints(units, right),
                                getAsPoints(m.units, m.right)) &&
               areProbablyEqual(getAsPoints(units, top),
                                getAsPoints(m.units, m.top)) &&
               areProbablyEqual(getAsPoints(units, bottom),
                                getAsPoints(m.units, m.bottom)));
	}

	/** Returns the margin positions and units as a String.
     * @return margin positions and units
     */
	public String toString() {
		return("(l, r, t, b) = (" + left + ", " + right + ", " + top + ", " +
				bottom + ") " + units);
	}

} // Margins

}
