/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PageTableCellInfo.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:23 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.page;

import java.awt.Color;
import java.awt.Font;
import java.awt.Insets;
import java.awt.Rectangle;

import com.klg.jclass.cell.JCCellInfo;

/**
 * Creates a JClass PageLayout-specific implementation of the JCCellInfo
 * interface so we can pass information from a JCPageTable to a
 * JCCellRenderer to draw table cells.
 */
public class PageTableCellInfo implements JCCellInfo {

/** The table to which the cell belongs */
protected JCPageTable		table;
/** The row of the cell in the table */
protected int				row;
/** The column of the cell in the table */
protected int				column;
/** The drawing area represented by this cell -- set externally */
protected Rectangle         drawingArea = null;

/**
 * Creates a TableCellInfo object for a given table cell.
 * @param table The table the cell belongs to
 * @param row The row index of the cell in the table
 * @param column The column index of the cell in the table
 */
public PageTableCellInfo(JCPageTable table, int row, int column) {
	this.table = table;
	this.row = row;
	this.column = column;
}

/**
 * Returns the background colour of the cell being displayed.
 * @return The colour to draw in the cell's background
 */
public Color getBackground() {

	// Get the background colour from the cell
	return table.getCell(row, column).getBackgroundColor();
}

/**
 * Returns the foreground colour of the cell being displayed.
 * @return The colour to draw the cell's foreground in
 */
public Color getForeground() {

	// Get the drawing colour of the cell's style
	return(table.getCell(row, column).getStyle().getColor());
}

/**
 * Returns the background color of the cell being displayed when
 * the cell is selected.
 * @return The colour to draw in the cell's background
 */
public Color getSelectedBackground() {

	// We don't have any selected cells in a pagelayout table
	return getBackground();
}

/**
 * Returns the foreground color of the cell being displayed when
 * the cell is selected.
 * @return The color to draw the cell's foreground in
 */
public Color getSelectedForeground() {

	// We don't have any selected cells in a pagelayout table
	return getForeground();
}

/**
 * Returns the font of the cell being displayed.
 * @return The font used to draw the cell
 */
public Font getFont() {

	// Get the font from the cell's style
	return(table.getCell(row, column).getStyle().getFont());
}

/**
 * Returns the horizontal alignment for the cell being displayed.
 * @return The position of the cell contents relative to the column
 */
public int getHorizontalAlignment() {

	// Get the alignment of cell contents from the cell's style
	int align = table.getCell(row, column).getStyle().getAlignment();

	// Convert the alignment to a CellInfo alignment
	if (align == JCTextStyle.ALIGNMENT_LEFT ||
		align == JCTextStyle.ALIGNMENT_NONE) {

		return(JCCellInfo.LEFT);
	}
	else if (align == JCTextStyle.ALIGNMENT_CENTER ||
			 align == JCTextStyle.ALIGNMENT_JUSTIFY) {

		return(JCCellInfo.CENTER);
	}
	else {		// align == JCTextStyle.ALIGNMENT_RIGHT

		return(JCCellInfo.RIGHT);
	}
}

/**
 * Returns the vertical alignment for the cell being displayed.
 * @return The position of the cell contents relative to the row space
 */
public int getVerticalAlignment() {

	// get the cell's vertical alignment
	int align = table.getCell(row, column).getCellAlignment();

	// Convert the alignment and return it
	if (align == JCPageTable.CELL_ALIGNMENT_BOTTOM) {
		return(JCCellInfo.BOTTOM);
	}
	else if (align == JCPageTable.CELL_ALIGNMENT_CENTER) {
		return(JCCellInfo.CENTER);
	}
	else {
		return(JCCellInfo.TOP);
	}
}

/**
 * Returns the Insets object representing the margins for the cell.
 * <p> This method should be applied to editor components but not to
 * renderer components since the renderer has already been mapped inside
 * the area bounded by the margins and borders.
 */
public Insets getMarginInsets() {
	// ???
	return(new Insets(0, 0, 1000, 1000));
}

/**
 * Returns the Insets object representing the borders for the cell.
 * <p> This method should be ignored by editor and renderer components
 * since the editor is deliberately made bigger than the renderer by this
 * amount.
 */
public Insets getBorderInsets() {
	// ???
	return(new Insets(0, 0, 1000, 1000));
}

/**
 * Returns the border style for the cell.  Renderers and Editors should
 * ignore this value.
 */
public int getBorderStyle() {
	// ???
	return(0);
}

/**
 * Returns the dimensions of the drawing area for the cell being displayed.
 * When the renderer's draw() method is called, the gc is translated
 * to the upper left-hand corner of the drawing area for the cell.  As
 * a result, the (x,y) of the Rectangle are always (0,0).
 * @return size of the drawing area.
 */
public Rectangle getDrawingArea() {
	if (drawingArea == null)
		return(new Rectangle());
	else
		return drawingArea;
}

public void setDrawingArea(Rectangle rect) {
	drawingArea = rect;
}

/**
 * Determines whether the cell is editable.
 * @return Always returns false for PageLayout tables
 */
public boolean isEditable() {
	// Cells in a pagelayout table are not editable
	return(false);
}

/**
 * Determines whether the cell is enabled.
 */
public boolean isEnabled() {
	// Cells in a pagelayout table are ???
	// ??? return(false);
	return(false);
}

/**
 * Determines whether the editor should select the contents before editing.
 * @return Always returns false for PageLayout tables
 */
public boolean getSelectAll() {
	// Cells in a pagelayout table are not editable
	return(false);
}

/**
 * Determines whether the renderer should draw some indication if the
 * entire contents of the cell cannot be rendered within the given area.
 */
public int getClipHints() {

	// We don't show visual indications of clipping
	// (We might change this later, but for now we don't)
	return JCCellInfo.SHOW_NONE;
}

/**
 * Returns the cell's associated data type.
 */
public Class getDataType() {
	// ???
	return(table.getTableData().getTableDataItem(row, column).getClass());
}

}
