/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: TextMacro.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:23 $ $Locker:  $  Quest Software Inc.
package com.klg.jclass.page;

import com.klg.jclass.page.render.Render;
import com.klg.jclass.page.render.PageNumberMacro;

import com.klg.jclass.page.render.RomanPageMacro;
import com.klg.jclass.page.render.SectionNumberMacro;
import com.klg.jclass.page.render.SectionPageNumberMacro;
import com.klg.jclass.page.render.SectionPageCountMacro;
import com.klg.jclass.page.render.DocumentPageCountMacro;


/**
 *
 * Interface which defines an embeddable text macro.
 * When evaluated in a context, the macro produces the text
 * to be substituted in its place.
 *
 * Several standard macros are included below, including:
 *   PAGE_NUMBER -- current page number
 *   ROMAN_NUMBER -- current page number in roman numerals
 *   SECTION_NUMBER -- current section number
 *   SECTION_PAGE_NUMBER -- current page number within current section
 *   SECTION_PAGE_TOTAL  -- total number of pages in current section
 *   PAGE_TOTAL -- total number of pages in document
 *
 * Users wanting to create their own macros should implement the methods
 * in this interface.  These methods will be called by PageLayout during
 * its regular operation.  They should behave as follows:
 *
 * If evaluate() has not yet been called by PageLayout, getStatus() should
 * return MACRO_INITIALIZED and getText() should return some non-null
 * initial value.
 *
 * When evaluate() is called by PageLayout, if the macro can be evaluated with
 * the given information, the value MACRO_EVALUATED should be returned.
 * getStatus() should henceforth return this value and getText() should
 * return the evaluated macro text.
 *
 * If the macro cannot be evaluated with the information available
 * when evaluate() is called, the method should return
 * MACRO_NOT_YET_EVALUATED.  getStatus() should henceforth return this
 * same value and future calls to getText() should return the same
 * non-null placeholder text returned during the initialize phase.
 *
 * Note that evaluate() may be called many several times per macro
 * instance with different, potentially null values for flow and page.
 * Once MACRO_EVALUATED has been returned by evaluate(), the method
 * will not be called again for the instance of the macro with the
 * current instance of the frame.  When a new instance of the frame
 * (such as a static header frame or the header frame of a table on
 * a brand new new page) is created, the cycle will restart and
 * evaluate() will be called again.  This allows the macro to react
 * to changes in the document between pages modify the text it
 * represents accordingly.
 */
public interface TextMacro {

/**
 * Computes the text to which this macro evaluates given the current context.
 * @param flow This parameter may be null when macro is initially added to
 *             a frame, but it usually contains the flow governing the
 *             document in which the marco appears, even if the macros
 *             has been added to a non-flow frame.
 * @param page The page on which the macro is displayed.  May be null
 *             when macro is initially added to frame.
 * @return a status indicating whether the macro could be successfully
 *         resolved.  One of MACRO_NOT_YET_EVALUATED or MACRO_EVALUATED.
 */
public int evaluate(JCFlow flow, JCPage page);

/**
 * Gets the text to which this macro evaluates given the current context.
 * @return The text which this macro resolves to.  Even if the macro has not
 *         yet succesfully evaluated, this should NOT return null.
 */
public String getText();

/**
 * Gets the status of the last evaluation of the macro.
 * @return An indication of the degree of success of the last evaluation.
 *         One of MACRO_NOT_YET_EVALUATED or MACRO_EVALUATED.
 */
public int getStatus();

/** Macro status -- initialized, but evaluate has not yet been called. */
public static final int   MACRO_INITIALIZED = Render.LAYOUT_PASS_0;
/** Macro status -- evaluate() has been called, but macro cannot yet
    be evaluated. */
public static final int   MACRO_NOT_YET_EVALUATED = Render.LAYOUT_PASS_1;
/** Macro status -- macro has been evaluated.  Value will be returned by
    future calls to getText(). */
public static final int   MACRO_EVALUATED = Render.LAYOUT_COMPLETE;

public static final TextMacro   PAGE_NUMBER = new PageNumberMacro();

public static final TextMacro   ROMAN_NUMBER = new RomanPageMacro();
public static final TextMacro   SECTION_NUMBER = new SectionNumberMacro();
public static final TextMacro   SECTION_PAGE_NUMBER = new SectionPageNumberMacro();
public static final TextMacro   SECTION_PAGE_TOTAL = new SectionPageCountMacro();
public static final TextMacro   PAGE_TOTAL   = new DocumentPageCountMacro();


}
