/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Ascii85OutputStream.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:23 $ $Locker:  $  Quest Software Inc. 

package com.klg.jclass.page.adobe.pdf;

import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * This class implements an output stream filter for applying ASCII-85
 * armouring to a stream of bytes.
 */
public class Ascii85OutputStream extends FilterOutputStream {

/** Working buffer (needed to accumulate groups of bytes to write) */
private byte[]			buffer = new byte[4];
/** Working buffer (needed to store converted 5-tuples) */
private byte[]			tuple = new byte[5];
/** Number of bytes in the working buffer */
private int				bufCount = 0;

/** The End-Of-Data marker for an ASCII-85 stream */
private static String	EOD = "~>";

/**
 * Create a filter to apply ASCII-85 armouring to the given output stream
 * @param out The output stream to pass filtered data to
 */
public Ascii85OutputStream(OutputStream out) {

	super(out);
}

/**
 * Implement the FilterOutputStream method
 */
public void close() throws IOException {

	// Flush the buffer and close the underlying stream
	flush();
	out.close();
}

/**
 * Implement the FilterOutputStream method
 */
public void flush() throws IOException {

	// Write out any remaining bytes
	if (bufCount > 0) {
		encode(buffer, 0, bufCount);
	}

	// Write out the end-of-data marker
	out.write(EOD.getBytes());
}

/**
 * Implement the FilterOutputStream method
 */
public void write(byte[] b, int off, int len) throws IOException {

	int b_index = 0;

	// Make sure we have enough bytes to start encoding
	if (len + bufCount >= 4) {

		// if there are saved bytes, consume them first
		if (bufCount > 0) {

			// Add bytes to the working buffer
			while (bufCount < 4) {
				buffer[bufCount++] = b[off + b_index];
				b_index++;
			}

			// Encode the 4-tuple
			encode(buffer, 0);
			bufCount = 0;
		}

		// While there are more 4-tuples in the input
		while (b_index + 3 < len) {

			// Encode each and advance
			encode(b, off + b_index);
			b_index += 4;
		}
	}

	// Store remaining bytes in the storage buffer
	for (int i = b_index; i < len; i++) {
		buffer[bufCount++] = b[off + i];
	}
}

/**
 * Implement the FilterOutputStream method
 */
public void write(byte[] b) throws IOException {
	write(b, 0, b.length);
}

/**
 * Implement the FilterOutputStream method
 */
public void write(int b) throws IOException {

	// Store the byte
	buffer[bufCount++] = (byte) b;

	// if we have a 4-tuple, encode it
	if (bufCount == 4) {
		encode(buffer, 0);
		bufCount = 0;
	}
}

protected final static char hd[] = { '0', '1', '2', '3', '4', '5', '6', '7',
									 '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
private static char[] char_b = new char[2];
private static char[] char_l = new char[16];

private static String hexVal(byte b) {

	char_b[0] = hd[(b & 0xF0) >> 4];
	char_b[1] = hd[(b & 0xF)];
	return(new String(char_b));
}

private static String hexVal(long l) {

	char_l[0] = hd[(int) ((l >> 60) & 0xF)];
	char_l[1] = hd[(int) ((l & 0xF00000000000000l) >> 56)];
	char_l[2] = hd[(int) ((l & 0xF0000000000000l) >> 52)];
	char_l[3] = hd[(int) ((l & 0xF000000000000l) >> 48)];
	char_l[4] = hd[(int) ((l & 0xF00000000000l) >> 44)];
	char_l[5] = hd[(int) ((l & 0xF0000000000l) >> 40)];
	char_l[6] = hd[(int) ((l & 0xF000000000l) >> 36)];
	char_l[7] = hd[(int) ((l & 0xF00000000l) >> 32)];
	char_l[8] = hd[(int) ((l & 0xF0000000l) >> 28)];
	char_l[9] = hd[(int) ((l & 0xF000000l) >> 24)];
	char_l[10] = hd[(int) ((l & 0xF00000) >> 20)];
	char_l[11] = hd[(int) ((l & 0xF0000) >> 16)];
	char_l[12] = hd[(int) ((l & 0xF000) >> 12)];
	char_l[13] = hd[(int) ((l & 0xF00) >> 8)];
	char_l[14] = hd[(int) ((l & 0xF0) >> 4)];
	char_l[15] = hd[(int) ((l & 0xF))];

	return(new String(char_l));
}

/**
 * Write the encoding of a 4-tuple
 */
private void encode(byte[] b, int offset) throws IOException {

	long			quad;

	quad = (((long) b[offset] & 0xFF) << 24) | ((b[offset + 1] & 0xFF) << 16) |
		   ((b[offset + 2] & 0xFF) << 8) | (b[offset + 3] & 0xFF);

	if (quad == 0) {

		// write 'z'
		out.write((byte) 122);
	}

	else {
		// convert the 4-tuple to an ASCII-85 5-tuple and write it
		tuple = getTuple(quad, tuple);
		out.write(tuple);
	}
}

/**
 * Write the encoding of a partial 4-tuple
 */
private void encode(byte[] b, int offset, int bytes) throws IOException {

	long			quad = 0;

	// Add the existing bytes into the working group
	for (int i = 0; i < bytes; i++) {
		quad = (quad << 8) | (b[offset + i] & 0xFF);
	}

	// Pad the remaining bytes
	for (int i = bytes; i < 4; i++) {
		quad = quad << 8;
	}

	// convert the 4-tuple to an ASCII-85 5-tuple
	tuple = getTuple(quad, tuple);

	// Write out the bytes that have data
	out.write(tuple, 0, bytes + 1);
}

private byte[] getTuple(long quad, byte[] outTuple) {

	// 33 is '!'
	outTuple[4] = (byte) (quad % 85 + 33);
	quad /= 85;
	outTuple[3] = (byte) (quad % 85 + 33);
	quad /= 85;
	outTuple[2] = (byte) (quad % 85 + 33);
	quad /= 85;
	outTuple[1] = (byte) (quad % 85 + 33);
	quad /= 85;
	outTuple[0] = (byte) (quad + 33);

	return(outTuple);
}

}
