/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: AFMParser.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:24 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.page.adobe.postscript;

import java.io.File;
import java.io.Reader;
import java.io.IOException;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;
import java.util.StringTokenizer;
import java.awt.Font;

import com.klg.jclass.page.FontParser;

import java.util.Vector;

/**
 * Reads in a file corresponding to the Adobe Font Metrics
 * Format Specification (.afm files) and builds an in-memory
 * representation of the specific font's information.
 * <P>
 * Parsing details: Each afm file contains information for one
 * PostScript font program. Each begins with Global information about
 * the font in general, then sections on individual character metrics.
 * <P>
 * The file format is line-oriented, each line beginning with a property
 * (key) followed by values for that property. Keys and values are separated
 * by one or more white spaces. Key names are case-sensitive.
 *
 */
public class AFMParser implements FontParser {

// AFM keys
// stored in an array so we can do binary search
// on them. Sorted lexigraphically.
public static final int ASCENDER			= 0;
public static final int CHARBBOX			= 1;
public static final int CODE				= 2;
public static final int COMPCHAR			= 3;
public static final int CAPHEIGHT			= 4;
public static final int COMMENT				= 5;
public static final int DESCENDER			= 6;
public static final int ENCODINGSCHEME		= 7;
public static final int ENDCHARMETRICS		= 8;
public static final int ENDCOMPOSITES		= 9;
public static final int ENDFONTMETRICS		= 10;
public static final int ENDKERNDATA			= 11;
public static final int ENDKERNPAIRS		= 12;
public static final int ENDTRACKKERN		= 13;
public static final int FAMILYNAME			= 14;
public static final int FONTBBOX			= 15;
public static final int FONTNAME			= 16;
public static final int FULLNAME			= 17;
public static final int ISFIXEDPITCH		= 18;
public static final int ITALICANGLE			= 19;
public static final int KERNPAIR			= 20;
public static final int KERNPAIRXAMT		= 21;
public static final int LIGATURE			= 22;
public static final int CHARNAME			= 23;
public static final int NOTICE				= 24;
public static final int COMPCHARPIECE		= 25;
public static final int STARTCHARMETRICS	= 26;
public static final int STARTCOMPOSITES		= 27;
public static final int STARTFONTMETRICS	= 28;
public static final int STARTKERNDATA		= 29;
public static final int STARTKERNPAIRS		= 30;
public static final int STARTTRACKKERN		= 31;
public static final int STDHW				= 32;
public static final int STDVW				= 33;
public static final int TRACKKERN			= 34;
public static final int UNDERLINEPOSITION	= 35;
public static final int UNDERLINETHICKNESS	= 36;
public static final int VERSION				= 37;
public static final int XYWIDTH				= 38;
public static final int XWIDTH				= 39;
public static final int WEIGHT				= 40;
public static final int XHEIGHT				= 41;
public static final int NOPE				= 42;


public static String keyStrings[] = {
    "Ascender", "B", "C", "CC", "CapHeight", "Comment",
    "Descender", "EncodingScheme", "EndCharMetrics", "EndComposites",
    "EndFontMetrics", "EndKernData", "EndKernPairs", "EndTrackKern",
    "FamilyName", "FontBBox", "FontName", "FullName", "IsFixedPitch",
    "ItalicAngle", "KP", "KPX", "L", "N",
    "Notice", "PCC", "StartCharMetrics", "StartComposites",
    "StartFontMetrics", "StartKernData", "StartKernPairs",
    "StartTrackKern", "StdHW", "StdVW", "TrackKern", "UnderlinePosition",
    "UnderlineThickness", "Version", "W", "WX", "Weight", "XHeight",
    null
};


protected String currentLine;
protected static final int EOF = -1;
protected static final int NOT_FOUND = -1;
protected BufferedReader reader;
protected String DELIMITERS = "    ;";
protected StringTokenizer tokenizer = new StringTokenizer("", DELIMITERS);

/**
 * Constructor
 * Call parse() to parse the file.
 */
public AFMParser() {
}

/**
 * Constructor
 * Calls parse() for you.
 *
 * @param afmFileName a String file name.
 * @throws FileNotFoundException If a file is not found.
 * @throws IOException If there is an error.
 */
public AFMParser(String afmFileName) throws FileNotFoundException,
		IOException {
	FileInputStream fis = new FileInputStream(afmFileName);
	parse(fis);
}

/**
 * Parse the file and build the in-memory image.
 *
 * @param reader a BufferedReader
 * @return the FontPS object created by reading this file.
 * @throws IOException If there is an error.
 */
public Vector parse(InputStream inputStream) throws IOException {

	if (inputStream == null) {
		throw new NullPointerException("parse(): Input stream is null");
	}

	// AFM files are ASCII-only by definition.
	// Pass in their encoding so that non-ascii machines (IBM 360, for example)
	// will interpret the bytes properly.
	reader = new BufferedReader(new InputStreamReader(inputStream, "ASCII"));

	// the default font size for fonts retrieved using Java's
	// getAllFonts method is a point size of one
	FontPS font = new FontPS("uninitialized", Font.PLAIN, 1);

	String keyword;
	boolean cont = true;

	// the afm file starts with the global font information
	parseGlobals(font);

	// parse the individual character metrics
	parseCharMetrics(font);

	/*
	 * The remaining sections of the AFM are optional. This code will
     * look at the next keyword in the file to determine what section
	 * is next, and then allocate the appropriate amount of storage
	 * for the data and call the appropriate parsing routine to parse
	 * the section.
	 */

	while (cont) {
		keyword = token();
		if (keyword == null) {
			/* Have reached an early and unexpected EOF. */
			/* Set flag and stop parsing */
			cont = false;
			break; /* get out of loop */
		}
		switch(recognize(keyword)) {
			case STARTKERNDATA:
				break;
			case ENDKERNDATA:
				break;
			case STARTTRACKKERN:
				parseTrackKernData(font);
				break;
			case STARTKERNPAIRS:
				parsePairKernData(font);
				break;
			case STARTCOMPOSITES:
				parseCompCharData();
				break;
			case ENDFONTMETRICS:
				cont = false;
				break;
			case NOPE:
			default:
				break;
		} /* switch */
	} /* while */
	// even though there is only one font per afm file we return it in
	// a vector for compatiblity with pcl fonts which can have
	// multiple fonts per tfm file
	Vector v = new Vector(1);
	v.addElement(font);
	return v;
} /* parseFile */

/**
 * Return the next token from the file.
 *
 * @return the next String token from the file or
 * null if the end of file has been reached.
 */
protected String token() throws IOException {
	String token = null;
	if (tokenizer != null && reader != null) {
		if (tokenizer.hasMoreTokens()) {
			token = tokenizer.nextToken();
		}
		else {
            getLine();
			if (tokenizer.hasMoreTokens()) {
				token = tokenizer.nextToken();
			}
            else if (currentLine != null) {
                token = "";
            }
		}
	}
	return token;
}

/**
 * Return the rest of the current line and move the pointer to
 * the next line. And sets up the next line.
 *
 * @return the currentLine
 */
protected String getLine() throws IOException {
    String line ="";
    while(tokenizer.hasMoreTokens()) {
        line += tokenizer.nextToken() + " ";
    }
    // move to the next line
    if (tokenizer != null && reader != null) {
        currentLine = reader.readLine();
		if (currentLine != null) {
        	tokenizer   = new StringTokenizer(currentLine, DELIMITERS);
		}
    }
    return line.trim();
}

/*  This function is called by "parseFile". It will parse the AFM File
 *  up to the "EndComposites" keyword. It will save the composite
 *  character data in a pointer to the PSFontInfo record.
 *
 *  This function will create the appropriate amount of storage for
 *  the composite character data and store a pointer to the storage
 *  in the PSFontInfo record.
 *
 */

protected void parseCompCharData() throws IOException {
	boolean cont = true;
	int ccount = 0, pcount = 0, numComposities;
	String keyword;
	CompCharPS data = null;

	numComposities = atoi(token());

	while (cont) {
		keyword = token();
		if (keyword == null) {
			/* Have reached an early and unexpected EOF. */
			throw new RuntimeException("");
		}
		if (ccount > numComposities) {
			throw new RuntimeException("");
		}
		/* parse Composite Character info section, saving the data */
		switch(recognize(keyword)) {
			case COMMENT:
				keyword = getLine();
				break;
			case COMPCHAR:
				if (ccount < numComposities) {
					/*
					 * Check for the correct number of pieces in the previous
					 * composite character definition
					 */
					if (ccount > 0 && pcount != data.numPieces) {
						throw new RuntimeException("");
					}
					pcount = 0;

					data 			= new CompCharPS();
					data.name 		= token();
					data.numPieces	= atoi(token());
					data.pieces 	= new CompPiecePS[data.numPieces];
					ccount++;
				}
				else {
					throw new RuntimeException("");
				}
				break;
			case COMPCHARPIECE:
				if (pcount < data.numPieces) {
					CompPiecePS c 	= new CompPiecePS();
					c.name 			= token();
					c.deltaX	 	= atoi(token());
					c.deltaY	 	= atoi(token());

					// add comp piece to the array
					data.pieces[pcount] = c;
					pcount++;
				}
				else {
					throw new RuntimeException("");
				}
				break;
			case ENDCOMPOSITES:
				cont = false;
				break;
			case ENDFONTMETRICS:
				cont = false;
				break;
			case NOPE:
			default:
				break;
		} /* switch */
	} /* while */
} /* parseCompCharData */

/*  This function is called by "parseFile". It will parse the AFM File
 *  up to the "EndKernPairs" or "EndKernData" keywords and save the
 *  pair kerning data.
 *
 *  parsePairKernData is passed in a pointer to the PSFontInfo record.
 *  If data is to be saved, the PSFontInfo record will already contain
 *  a valid pointer to storage for the pair kerning data.
 *
 */

protected void parsePairKernData(FontPS font) throws IOException {
	int pcount = 0, numKernPairs;
	String keyword;
	boolean cont = true;
	PairKern pairData;

	numKernPairs = atoi(token());

	while (cont) {
		keyword = token();

		if (keyword == null) {
			// throw new RuntimeException("");
		}
		/* parse entire Pair Kerning Data section, saving the data */
		switch(recognize(keyword)) {
			case COMMENT:
				getLine();
				break;
			case KERNPAIR:
				if (pcount < numKernPairs) {
					pairData 		= new PairKern();
					pairData.name1 	= token();
					pairData.name2 	= token();
					pairData.xAmount = atoi(token());
					pairData.yAmount = atoi(token());
					pcount++;
					// add this kern pair to char name1's list,
                    // indexed by name2
                    CharMetricPS c = font.getCharByName(pairData.name1);
					c.getKernPairs().put(pairData.name2, pairData);
				}
				else {
					throw new RuntimeException("");
				}
				break;
			case KERNPAIRXAMT:
				if (pcount < numKernPairs) {
					pairData 		= new PairKern();
					pairData.name1 	= token();
					pairData.name2 	= token();
					pairData.xAmount = atoi(token());
					pairData.yAmount = 0;
					pcount++;
					// add this kern pair to char name1's list,
                    // indexed by name2
                    CharMetricPS c = font.getCharByName(pairData.name1);
					c.getKernPairs().put(pairData.name2, pairData);
				}
				else {
					throw new RuntimeException("");
				}
				break;
			case ENDKERNPAIRS:
			case ENDKERNDATA:
				cont = false;
				break;
			case ENDFONTMETRICS:
				cont = false;
				break;
			case NOPE:
				default:
				break;
		} /* switch */
	} /* while */
} /* parsePairKernData */

/**
 *  This function is called by "parseFile". It will parse the AFM File
 *  up to the "EndTrackKern" or "EndKernData" keywords and save the
 *  track kerning data.
 *
 *  parseTrackKernData is passed in a pointer to the PSFontInfo record.
 *  If data is to be saved, the PSFontInfo record will already contain
 *  a valid pointer to storage for the track kerning data.
 *
 */

protected void parseTrackKernData(FontPS font) throws IOException {
	boolean cont = true;
	int tcount = 0, numTrackkern;
	String keyword;

	numTrackkern = atoi(token());

	while (cont) {
		keyword = token();

		if (keyword == null) {
			// throw exception
			break; /* get out of loop */
		}
		/* parse entire Track Kerning Data section, saving the data */
		switch(recognize(keyword)) {
			case COMMENT:
				getLine();
				break;
			case TRACKKERN:
				if (tcount < numTrackkern) {
					TrackKern trackData		= new TrackKern();
					trackData.degree		= atoi(token());
					trackData.minPtSize		= atof(token());
					trackData.minKernAmount	= atof(token());
					trackData.maxPtSize		= atof(token());
					trackData.maxKernAmount	= atof(token());
					// add it to the list
					font.getTrackKerning().put(new Integer(trackData.degree),
						trackData);
					tcount++;
				}
				else {
					// throw exception
					cont = false;
				}
				break;
			case ENDTRACKKERN:
			case ENDKERNDATA:
				cont = false;
				break;
			case ENDFONTMETRICS:
				cont = false;
				break;
			case NOPE:
			default:
				break;
		} /* switch */
	} /* while */
} /* parseTrackKernData */

/**
 *  This function is called by parseFile to save all character
 *  metric information.
 *
 *  parseCharMetrics is passed in a pointer to an array of records
 *  to hold information on a per character basis. This function
 *  parses the Character Metrics section storing all character
 *  metric information for the ALL characters (mapped and unmapped)
 *  into the array.
 */

protected void parseCharMetrics(FontPS font) throws IOException {
	int count = 0, numChar, cnt = 0;
	String keyword;
	boolean cont = true;

	numChar = atoi(token());

	// allocate memory for the array
	font.charMetricList = new CharMetricPS[numChar];

	CharMetricPS charInfo = null;

	while (cont) {
		keyword = token();
		if (keyword == null) {
			// error = earlyEOF;
			// throw an exception
			break; /* get out of loop */
		}
		// switch(java.util.Arrays.binarySearch(keyStrings, keyword)) {
        switch(recognize(keyword)) {
			case COMMENT:
				getLine(); // eat the line, not interested in it
				break;
			case CODE:
				if (count < numChar) {
					charInfo = new CharMetricPS();
					charInfo.setWx(0);
					charInfo.setWy(0);
					charInfo.getCharBBox().llx = 0;
					charInfo.getCharBBox().lly = 0;
					charInfo.getCharBBox().urx = 0;
					charInfo.getCharBBox().ury = 0;
					charInfo.setCode(atoi(token()));
					count++;
				}
				else {
					// error = parseError;
					// throw an exception
					cont = false;
				}
				break;
			case XYWIDTH:
				charInfo.setWx(atoi(token()));
				charInfo.setWy(atoi(token()));
				break;
			case XWIDTH:
				charInfo.setWx(atoi(token()));
				break;
			case CHARNAME:
				charInfo.setName(token());
				// add this new char to the list
				// we maintain two lists for fast lookup on
				// name and position
                font.getCharMetrics().put(charInfo.getName(), charInfo);
				font.charMetricList[cnt++] = charInfo;
				break;
			case CHARBBOX:
				charInfo.getCharBBox().llx = atoi(token());
				charInfo.getCharBBox().lly = atoi(token());
				charInfo.getCharBBox().urx = atoi(token());
				charInfo.getCharBBox().ury = atoi(token());
				break;
			case LIGATURE: {
				LigaturePS ligature = new LigaturePS();
				ligature.setSuccessor(token());
				ligature.setLigature(token());
				charInfo.getLigatures().put(ligature.getLigature(),
					ligature.getLigature());
				break;
			}
			case ENDCHARMETRICS:
				cont = false;;
				break;
			case ENDFONTMETRICS:
				cont = false;
				break;
			case NOPE:
			default:
				// not recognized, continue
				break;
		} /* switch */
	} /* while */
} /* parseCharMetrics */


/**
 *  This function is called by "parseFile". It will parse the AFM File
 *  up to the "StartCharMetrics" keyword, which essentially marks the
 *  end of the Global Font Information and the beginning of the character
 *  metrics information.
 *
 *  If the caller of "parseFile" specified that it wanted the Global
 *  Font Information (as defined by the "AFM File Specification"
 *  document), then that information will be stored in the returned
 *  data structure.
 *
 *  Any Global Font Information entries that are not found in a
 *  given file, will have the usual default initialization value
 *  for its type (i.e. entries of type int will be 0, etc).
 *
 */

protected void parseGlobals(FontPS font) throws IOException {
    String keyword;
    boolean cont = true;
	while (cont) {
		keyword = token();
		if (keyword == null) {
			// reached unexpected eof throw an exception?
			break;
		}
		// parse the entire global font info section, saving the data
        // switch(java.util.Arrays.binarySearch(keyStrings, keyword)) {
        switch(recognize(keyword)) {
			case STARTFONTMETRICS:
                font.setAfmVersion(token());
                break;
			case COMMENT:
				// consume the line since it's just a comment
                getLine();
				break;
			case FONTNAME:
				font.setFontName(token());
				break;
			case ENCODINGSCHEME:
                font.setEncodingScheme(token());
				break;
			case FULLNAME:
                font.setFullName(getLine());
				break;
			case FAMILYNAME:
                font.setFamilyName(getLine());
				break;
			case WEIGHT:
                font.setWeight(token());
				break;
			case ITALICANGLE:
				font.setItalicAngle(Float.valueOf(token()).floatValue());
				// CHECKERROR(ERANGE, parseError);
				break;
			case ISFIXEDPITCH:
				if (token().equals("false")) {
			        font.setFixedPitch(false);
				}
				else  {
				    font.setFixedPitch(true);
				}
				break;
			case UNDERLINEPOSITION:
				font.setUnderlinePosition(atoi(token()));
				break;
			case UNDERLINETHICKNESS:
				font.setUnderlineThickness(atoi(token()));
				break;
			case VERSION:
                font.setVersion(token());
				break;
			case NOTICE:
				font.setNotice(getLine());
				break;
			case FONTBBOX:
				font.getFontBBox().llx  = atoi(token());
				font.getFontBBox().lly  = atoi(token());
				font.getFontBBox().urx  = atoi(token());
				font.getFontBBox().ury  = atoi(token());
				break;
			case CAPHEIGHT:
				font.setCapHeight(atoi(token()));
				break;
			case XHEIGHT:
				font.setXHeight(atoi(token()));
				break;
			case DESCENDER:
				font.setDescender(atoi(token()));
				break;
			case ASCENDER:
				font.setAscender(atoi(token()));
				break;
			case STDHW:
				font.setStdHW(atoi(token()));
				break;
			case STDVW:
				font.setStdVW(atoi(token()));
				break;
			case STARTCHARMETRICS:
				cont = false;
				break;
			case ENDFONTMETRICS:
				cont = false;
				// error = normalEOF;
				break;
			case NOPE:
			default:
                // consume lines for keys we are not
                // interested in or don't recognize
				// error = parseError;
                getLine();
				break;
		} /* switch */

	}
}

protected int atoi(String a) {
    return Integer.valueOf(a).intValue();
}
protected float atof(String a) {
	return Float.valueOf(a).floatValue();
}

// binary search on keys
// can be replaced by Arrays.binarySearch when there is an environment
// which supports jdk1.2 debugging
protected int recognize(String key) {
	int lower = 0, upper = NOPE, midpoint = 0, cmpvalue;
	boolean found = false;

	while ((upper >= lower) && !found) {
		midpoint = (lower + upper)/2;
		if (keyStrings[midpoint] == null) {
			break;
		}
		cmpvalue = key.compareTo(keyStrings[midpoint]);
		if (cmpvalue == 0) {
			found = true;
		}
		else if (cmpvalue < 0) {
			upper = midpoint - 1;
		}
		else {
			lower = midpoint + 1;
		}
	} /* while */

	if (found) {
		return midpoint;
	}
	else {
		return NOPE;
	}
} /* recognize */


/**
 * Driver
 */
public static void main(String args[]) {
	if (args.length > 0) {
		try {
			AFMParser parser = new AFMParser(args[0]);
        }
		catch (Exception fnf) {
			System.out.println(fnf.toString());
		}
	}
	else {
	 	System.out.println("Usage: AFMParser <fileName>");
	}
	System.exit(0);
}
} // end AFMParser
