/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: ImageRender.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:26 $ $Locker:  $  Quest Software Inc. 

package com.klg.jclass.page.render;

import com.klg.jclass.page.*;
import com.klg.jclass.util.ServerRenderable;

import javax.swing.*;
import java.awt.*;
import java.awt.geom.Rectangle2D;
import java.awt.image.ImageObserver;

/**
 * ImageRender encapsulates various forms of images with their formatting
 * information.
 */
public class ImageRender extends Render implements Floatable, Embedable,
			 									   ImageObserver, Cloneable {

/** The image object which is to be displayed */
protected Object            image;

/** The horizontal alignment of the image within the frame */
protected int               horizontalAlignment;
/** The vertical alignment of the image within its line */
protected int               verticalAlignment;
/** Indicates if the image background is to be filled or not */
protected boolean           fillBackground;
/** Indicates if the image should be manually scaled */
protected boolean           scaleImage;

/** An id number used by the MediaTracker to watch this image */
private int                 imageID;

/** This component is used by MediaTracker to check the loading of images */
private static ImComponent	imageComponent = new ImComponent();
/** MediaTracker waits for images to load before printing them */
private static MediaTracker tracker = new MediaTracker(imageComponent);
/** Used to assign a unique ID to each image in MediaTracker */
private static int          nextImageID = 0;

/*
public void init() {
	// Create the MediaTracker
	tracker = new MediaTracker(imageComponent);
}
*/

/**
 * Create an Image render object
 */
public ImageRender() {

	this(null, null, JCTextStyle.ALIGNMENT_CENTER);
}

/**
 * Create an image render object for a given image
 * @param image The image (in a variety of forms) to be displayed
 */
public ImageRender(Object image) {

	this(image, null, JCTextStyle.ALIGNMENT_CENTER);
}

/**
 * Create an image render object for a given image and size
 * @param image The image (in a variety of forms) to be displayed
 * @param size The size the image is to be scaled to
 */
public ImageRender(Object image, JCUnit.Dimension size) {

	this(image, size, JCTextStyle.ALIGNMENT_CENTER);
}

/**
 * Create an image render object for a given image and size
 * @param image The image (in a variety of forms) to be displayed
 * @param alignment The horizontal alignment of the image within its frame
 */
public ImageRender(Object image, int alignment) {

	this(image, null, alignment);
}

/**
 * Create an image render object for a given image and size
 * @param image The image (in a variety of forms) to be displayed
 * @param size The size the image is to be scaled to
 * @param alignment The horizontal alignment of the image within its frame
 */
public ImageRender(Object image, JCUnit.Dimension size, int alignment) {

	// create a basic render object
	super();

	// set the image, size and alignment
	this.image = image;
	this.bounds.assign(size);
	this.horizontalAlignment = alignment;
	this.verticalAlignment = JCDrawStyle.CENTER_ON_BASELINE;
	this.scaleImage = false;

	// Assign this image an imageID
	imageID = nextImageID++;
}

/**
 * Get the image
 * @return The image object this render object encapsulates
 */
public Object getImage() {
	return image;
}

/**
 * Set the image
 * @param image The image object this render object encapsulates
 */
public void setImage(Object image) {

	this.image = image;
}

/**
 * Get the alignment mode of the image in the frame
 * @return Alignment of the image to the frame
 */
public int getAlignment() {
	return horizontalAlignment;
}

/**
 * Set the alignment mode of the image in the frame
 * @param alignment Alignment of the image to the frame
 */
public void setAlignment(int alignment) {

	if (alignment <= 0 || alignment > JCTextStyle.ALIGNMENT_JUSTIFY) {
		throw new IllegalArgumentException("Unrecognised horizontal alignment");
	}

	this.horizontalAlignment = alignment;
}

/**
 * Get the vertical alignment of the image in the current line
 * @return Alignment of the image to the current line
 */
public int getVerticalAlignment() {
	return verticalAlignment;
}

/**
 * Set the vertical alignment of the image in the current line
 * @param alignment Alignment of the image to the current line
 */
public void setVerticalAlignment(int alignment) {

	if (alignment <= 0 || alignment > JCDrawStyle.ALIGN_TO_BOTTOM) {
		throw new IllegalArgumentException("Unrecognised vertical alignment");
	}

	this.verticalAlignment = alignment;
}

/**
 * Check if the image background is to be filled
 * @return True if the background is to be filled
 */
public boolean isFilled() {
	return fillBackground;
}

/**
 * Set whether this image's background is to be filled
 * @param filled True if the image background should be filled in
 */
public void setFilled(boolean filled) {

	this.fillBackground = filled;
}

/**
 * Check if the image should be manually scaled.  This is for
 * internal use only.
 * @return True if the image should be manually scaled.  False if
 * the printer graphics class will do the scaling itself.
 */
public boolean isScaleImage() {
	return scaleImage;
}

/**
 * Set whether this image should be manually scaled.  This is for
 * internal use only.
 * @param scale True if the image should be manually scaled.  False if
 * the printer graphics class will do the scaling itself.
 */
public void setScaleImage(boolean scale) {

	this.scaleImage = scale;
}

/**
 * Returns an indication if this object can be split into components
 * @return True indicating the line can't be split
 */
public boolean isAtomic() {
	return(true);
}

/**
 * Draw the image described by this object
 * @param gc The graphics context which will draw this object
 */
public void draw(Graphics2D gc) {

	// Images are now handled by ImageMarkers for Markup output, so ignore this
	if (gc instanceof Markup) {
		return;
	}

	double x = JCUnit.getAsPoints(origin.units, origin.x);
	double y = JCUnit.getAsPoints(origin.units, origin.y);
	double width = JCUnit.getAsPoints(bounds.units, bounds.width);
	double height = JCUnit.getAsPoints(bounds.units, bounds.height);

	// Handle EPSImages separately from other Image types
	if (image instanceof EPSImage) {

		// We don't have to worry about rescaling, since that happens
		// automatically when the image is drawn

		// draw the image
		gc.drawImage((Image) image, (int) x, (int) y,
					 (int) width, (int) height, this);
	}

	// If this is an Image or ImageIcon object, use the graphic's drawImage()
	else if (image instanceof Image || image instanceof Icon) {

		Image outImage = null;

		// get the original (unscaled) image
		if (image instanceof Image) {
			outImage = (Image) image;
		}
		else if (image instanceof ImageIcon) {
			outImage = ((ImageIcon) image).getImage();
		}
		else {
			// Of Icons, we only know how to deal with ImageIcons
			throw new IllegalArgumentException("Unknown Icon type");
		}

		// Determine the pixel dimensions of the provided image
		int imgWidth = outImage.getWidth(this);
		int imgHeight = outImage.getHeight(this);

		// if the printer doesn't supports image scaling, scale the image
		// here if the sizes don't match
		if (scaleImage && 
				(imgWidth != (int) width || imgHeight != (int) height)) {
			outImage = outImage.getScaledInstance((int) width, (int) height,
												  Image.SCALE_DEFAULT);
		}

		// Set MediaTracker to watch the image until it's complete
		tracker.addImage(outImage, imageID);

		try {
			tracker.waitForID(imageID);
		}
		catch (InterruptedException e) {}

		// draw the image
		gc.drawImage(outImage, (int) x, (int) y, (int) width,
					 (int) height, this);

		// Remove the image from MediaTracker
		tracker.removeImage(outImage);
	}

	else if (image instanceof Component) {
		Component  c = (Component) image;

		// Get the size of the component
		

		
		int imgWidth = c.getWidth();
		int imgHeight = c.getHeight();


		// Create a graphics into which to draw the component
		Graphics2D gc1 = (Graphics2D) gc.create();

		// Note: order is important here.  Make sure to translate BEFORE
		// doing the scale.

		// Adjust the gc so the component is drawn correctly
		gc1.translate(x, y);

		// if the component size is ont the output size, scale the graphics
		if (width > 0 && height > 0 &&
			(imgWidth != (int) width || imgHeight != (int) height)) {

			gc1.scale(width / ((double) imgWidth),
					  height / ((double) imgHeight));

			// clip is relative to image region
			gc1.setClip(new Rectangle2D.Double(0.0, 0.0, imgWidth, imgHeight));
		} else {
			// Clip to the size of the component as intersected with the
			// current clip region
			Rectangle currentClip = gc1.getClipBounds();
			gc1.setClip(new Rectangle2D.Double(
				0.0, 0.0, 
				(double) Math.min(currentClip.width, imgWidth),
		    	(double) Math.min(currentClip.height, imgHeight)));
		}

		// reset the stroke to the default
		// otherwise, component may use last drawn stroke to draw with
		gc1.setStroke(new BasicStroke());

		// Tell the component to draw itself
		if (!c.isShowing() && image instanceof ServerRenderable) {
			// Use the serverPaint() method of any ServerRenderable objects
			ServerRenderable  sr = (ServerRenderable) image;
			sr.serverPaint(gc1);
		} else {
			// for any other component use its paint() method
			c.paint(gc1);
		}

		// dispose of the temporary graphics
		gc1.dispose();
	}
	else {
		throw new IllegalArgumentException("Bad image type");
	}
}

/**
 * This method satisfies the ImageObserver interface and is used to get
 * asynchronous notification of the known information about the image
 * @param img The image being updated
 * @param flags Flags indicating what information is now available
 * @param x X location of the image
 * @param y Y location of the image
 * @param width The width of the image
 * @param height The height of the image
 * @return True if more information is needed, otherwise false
 */
public boolean imageUpdate(Image img, int flags, int x, int y, int width,
						   int height) {

	// wait until at least the dimensions are available
	if ((flags & ImageObserver.WIDTH) == 0 ||
		(flags & ImageObserver.HEIGHT) == 0) {
		return(true);
	}

	// Use the width if we've been waiting for it
	if (bounds.getWidth().isZero()) {
		bounds.setWidth(new JCUnit.Measure(JCUnit.POINTS, (double) width));
	}

	// Use the height if we've been waiting for it
	if (bounds.getHeight().isZero()) {
		bounds.setHeight(new JCUnit.Measure(JCUnit.POINTS, (double) height));
	}

	// if we're drawing the image we want it to be completed before proceding
	if ((flags & ImageObserver.ALLBITS) == 0) {
		return(true);
	}

	// We're happy now
	return(false);
}

/**
 * Create an image which is a copy of this one
 * @return a clone of this image
 */
public Object clone() {

	ImageRender     new_image;

	// Clone the image
	new_image = (ImageRender) super.clone();

	// Copy dependent objects

	return((Object) new_image);
}

protected static class ImComponent extends Component {

	protected ImComponent() {
		super();
	}
}

}
