/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: LineRender.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:26 $ $Locker:  $  Quest Software Inc. 
package com.klg.jclass.page.render;

import java.awt.Color;

import com.klg.jclass.page.JCDrawStyle;
import com.klg.jclass.page.JCUnit;




import java.awt.BasicStroke;
import java.awt.Graphics2D;
import java.awt.geom.GeneralPath;


/**
 * LineRender describes a line (polyline) object
 */
public class LineRender extends Render implements Cloneable {

/** The list of points to connect with a line */
protected JCUnit.Point[]    points;

/** The draw style which specifies drawing parameters */
protected JCDrawStyle       style;
/** Indicates if the line is to be filled or not */
protected boolean           fill;

/**
 * Create a Line render object
 */
public LineRender() {

	// Create a degenerate (zero-point) line
	bounds = new JCUnit.Dimension(JCUnit.POINTS);
	origin = new JCUnit.Point(JCUnit.POINTS);
	points = null;
	style = null;
}

/**
 * Create a line render object given a pair of points and a style
 * @param point1 The first point the line is drawn from
 * @param point2 The point the line is drawn to
 * @param style the draw style to use for painting the line
 */
public LineRender(JCUnit.Point point1, JCUnit.Point point2, JCDrawStyle style) {

	JCUnit.Point[]  points = new JCUnit.Point[2];

	this.style = style;

	// copy points and compute bounds
	points[0] = point1;
	points[1] = point2;
	setPoints(points);
}

/**
 * Create a line render object given a point list and a style
 * @param points the list of points for this render object to display
 * @param style the draw style to use for painting the line
 */
public LineRender(JCUnit.Point[] points, JCDrawStyle style) {

	this.style = style;

	// copy points and compute bounds
	setPoints(points);
}

/**
 * Get the list of points
 * @return An array containing the points of the line in order
 */
public JCUnit.Point[] getPoints() {
	return points;
}

/**
 * Set the points to be connected by a line
 * @param points A list of points to be connected with a line
 */
public void setPoints(JCUnit.Point[] points) {

	int             i;
	double          x, y, x_min, y_min, x_max, y_max;
	JCUnit.Point    point;

	x_max = y_max = 0.0;
	x_min = y_min = 100000.0; // Big num - larger than will appear on any page

	this.points = new JCUnit.Point[points.length];

	// Copy the points to this.points
	for (i = 0; i < points.length; i++) {
		point = (JCUnit.Point) points[i];

		x = point.xInt();
		y = point.yInt();

		// Determine the bounding box of the line
		if (x < x_min) {
			x_min = x;
		}
		if (y < y_min) {
			y_min = y;
		}

		if (x < x_max) {
			x_max = x;
		}
		if (y < y_max) {
			y_max = y;
		}

		this.points[i] = (JCUnit.Point) point.clone();
	}

	// Store the bounds of the line
	this.origin = new JCUnit.Point(JCUnit.INTERNAL, x_min, y_min);
	this.bounds = new JCUnit.Dimension(JCUnit.INTERNAL, x_max - x_min,
									   y_max - y_min);
}

/**
 * Get the style used to draw this line
 * @return the draw style assigned to this line
 */
public JCDrawStyle getStyle() {
	return style;
}

/**
 * Set the style used to draw this line
 * @param style The draw style used to create this line
 */
public void setStyle(JCDrawStyle style) {

	this.style = style;
}

/**
 * Check if the figure described by the line is to be filled
 * @return True if the line is to be filled
 */
public boolean isFilled() {
	return fill;
}

/**
 * Set whether this line is to be filled
 * @param filled True if the line is to be used as the outline of a fill area
 */
public void setFilled(boolean filled) {

	this.fill = filled;
}

/**
 * Returns an indication if this object can be split into components
 * @returns True indicating the line can't be split
 */
public boolean isAtomic() {
	return(true);
}

/**
 * Draw the line described by this object
 * @param gc The graphics context which will render this object
 */
public void draw(Graphics2D gc) {

	int             i, line_type;
	double          line_width, dash, spacing;
	Color           colour;
	BasicStroke     stroke;
	GeneralPath     path;
	JCUnit.Point    p;

	// Determine the colour to use to draw/fill the figure
	if (fill) {
		colour = style.getFillForegroundColor();

		// if a fill colour is not given, use the draw colour
		if (colour == null) {
			colour = style.getForegroundColor();
		}
	}
	else {
		colour = style.getForegroundColor();
	}

	// if there is no colour set, don't draw anything
	if (colour == null) {
		return;
	}

	// Set the colour to use
	gc.setColor(colour);

	// Create a stroke to handle line width, etc.
	line_width = style.getLineWidth().getAs(JCUnit.POINTS);

	line_type = style.getLineType();
	if (line_type == JCDrawStyle.LINE_TYPE_BROKEN) {
		float[]     dashes = new float[2];

		// Get the dash length and create the dash sequence array
		dash = style.getDashLength().getAs(JCUnit.POINTS);
		dashes[0] = (float) dash; dashes[1] = (float) dash;

		// Create a dashed stroke
		stroke = new BasicStroke((float) line_width, BasicStroke.CAP_BUTT,
								 BasicStroke.JOIN_MITER, (float) 10.0, dashes,
								 (float) 0.0);
	}
	else if (line_type == JCDrawStyle.LINE_TYPE_DOUBLE) {
		spacing = style.getLineSpacing().getAs(JCUnit.POINTS);

		// If this is not a vertical or horizontal 2-point line
		if (points.length != 2 ||
			(!points[0].getX().equals(points[1].getX()) &&
			 !points[0].getY().equals(points[1].getY()))) {

			// Draw a line the thickness of the double line
			line_width = line_width * 2 + spacing;
			stroke = new BasicStroke((float) line_width, BasicStroke.CAP_BUTT,
									 BasicStroke.JOIN_MITER);
		}

		// Otherwise draw parallel lines
		else {
			double  x1, y1, x2, y2;
			double  x_delta, y_delta;

			// Compute the change in point position
			if (points[0].getY().equals(points[1].getY())) {
				x_delta = 0.0;
				y_delta = (spacing + line_width) / 2;
			}
			else {
				x_delta = (spacing + line_width) / 2;
				y_delta = 0.0;
			}

			// Extract the endpoints of the line
			x1 = JCUnit.getAsPoints(points[0].units, points[0].x);
			y1 = JCUnit.getAsPoints(points[0].units, points[0].y);
			x2 = JCUnit.getAsPoints(points[1].units, points[1].x);
			y2 = JCUnit.getAsPoints(points[1].units, points[1].y);

			// Create the stroke for the line and set it
			stroke = new BasicStroke((float) line_width, BasicStroke.CAP_BUTT,
									 BasicStroke.JOIN_MITER);
			gc.setStroke(stroke);

			// Create a new path for the lines
			path = new GeneralPath(style.getFillRule(), 2);

			// Draw the first line
			path.moveTo((float) (x1 + x_delta), (float) (y1 + y_delta));
			path.lineTo((float) (x2 + x_delta), (float) (y2 + y_delta));
			gc.draw(path);

			// Draw the second line
			path.reset();
			path.moveTo((float) (x1 - x_delta), (float) (y1 - y_delta));
			path.lineTo((float) (x2 - x_delta), (float) (y2 - y_delta));
			gc.draw(path);

			// Exit in the 2-point double line case
			return;
		}
	}
	else if (line_type == JCDrawStyle.LINE_TYPE_PLAIN ||
			 line_type == JCDrawStyle.LINE_TYPE_SINGLE) {
		// Create the standard stroke
		stroke = new BasicStroke((float) line_width, BasicStroke.CAP_BUTT,
								 BasicStroke.JOIN_MITER);
	} 
	else {
		// line type = JCDrawStyle.LINE_TYPE_NONE
		// so don't draw it
		return;
	}

	// Create a new path for the polyline
	path = new GeneralPath(style.getFillRule(), points.length);

	// Move to the first point of the polyline
	path.moveTo((float) JCUnit.getAsPoints(points[0].units, points[0].x),
				(float) JCUnit.getAsPoints(points[0].units, points[0].y));

	// Loop, adding all following points to the polyline
	for (i = 1; i < points.length; i++) {
		path.lineTo((float) JCUnit.getAsPoints(points[i].units, points[i].x),
					(float) JCUnit.getAsPoints(points[i].units, points[i].y));
	}

	// Use the given graphics to draw the line
	if (fill) {
		gc.fill(path);
	}
	else {
		gc.setStroke(stroke);
		gc.draw(path);
	}
}

/**
 * Create a line which is a copy of this one
 * @return a clone of this line
 */
public Object clone() {

	LineRender  new_line;

	// Clone the line
	new_line = (LineRender) super.clone();

	// Copy dependent objects
	new_line.points = new JCUnit.Point[this.points.length];
	System.arraycopy(this.points, 0, new_line.points, 0, this.points.length);

	return((Object) new_line);
}

}
