/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: MacroRender.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:26 $ $Locker:  $  Quest Software Inc. 
package com.klg.jclass.page.render;

import java.awt.Font;
import java.awt.FontMetrics;



import com.klg.jclass.page.FontMetricsBase;
import com.klg.jclass.page.TextMacro;
import com.klg.jclass.page.JCFlow;
import com.klg.jclass.page.JCPage;
import com.klg.jclass.page.JCTextStyle;
import com.klg.jclass.page.JCUnit;




import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;

import java.awt.font.TextAttribute;

import java.text.AttributedString;


/**
 * MacroRender describes a marker for a text flow action (e.g. tab, line feed)
 */
public class MacroRender extends Render implements Cloneable {

/** The macro which is to be evaluated */
protected TextMacro         macro;
/** The text to which this macro has been evaluated */
protected String            text;

/** The style which describes the text attributes at the symbol's location */
protected JCTextStyle       style;
/** The graphics in which the resolved text is to be drawn */
protected Graphics2D        saved_gc;

/** An indication of whether the macro has been evaluated or not */
protected int               status;

/**
 * Create a macro render object
 */
public MacroRender() {

	macro = null;
	style = null;
	text = null;
	saved_gc = null;
	status = Render.LAYOUT_PASS_0;

	origin = new JCUnit.Point(JCUnit.POINTS);
	bounds = new JCUnit.Dimension(JCUnit.POINTS);
}

/**
 * Create a macro render object given its type and a style
 * @param gc The graphics context with respect to which the text is drawn
 * @param macro the macro this must evaluate
 * @param style the text style to use for formatting the macro text
 */
public MacroRender(Graphics2D gc, TextMacro macro, JCTextStyle style) {

	FontMetrics     fm;
	Rectangle2D     extent;

	// Store the macro and the style
	this.macro = macro;
	this.style = style;

	// Store the gc
	saved_gc = gc;
	fm = gc.getFontMetrics(style.getFont());

	// Get the current evaluation of the macro
	text = macro.getText();
	status = macro.getStatus();

	// compute the size of the string
	extent = FontMetricsBase.getStringBounds(fm, text, gc);
	

	bounds = new JCUnit.Dimension(JCUnit.POINTS, extent.getWidth(),
								  extent.getHeight());
}

/**
 * Get the status of the latest resolution of the macro
 * @return The success of the last attempt at resolving this macro
 */
public int getStatus() {
	return(status);
}

/**
 * Get the style assigned to this symbol
 * @return the text style assigned to this text symbol
 */
public JCTextStyle getStyle() {
	return style;
}

/**
 * Set the style used to determine attributes of this symbol
 * @param style The text style used to format this symbol
 */
public void setStyle(JCTextStyle style) {

	this.style = style;
}

/**
 * Get the text to which this macro currently resolves
 * @return the current text value of the macro
 */
public String getText() {
	return(text);
}

/**
 * Update the evaluation of this macro (new information may have been provided)
 * @param flow the flow governing this document
 * @param page The page this macro appears upon
 * @param An indication of the degree of completeness/success
 */
public int evaluateMacro(JCFlow flow, JCPage page) {

	String          new_text;
	FontMetrics     fm;
	Rectangle2D     extent;

	
	// Evaluate the macro given the current information
	status = macro.evaluate(flow, page);

	// Get the evaluated value and check for a change
	new_text = macro.getText();
	if (new_text.equals(text)) {
		//JLA:  fixes Bunka JCFlow bugs 003 001 and 003 003
		//superclass's status was not being updated
		setLayoutStatus(status);

		return(status);
	}

	// Otherwise the text has changed, so update it
	text = new_text;
	fm = saved_gc.getFontMetrics(style.getFont());

	// compute the size of the string
	extent = FontMetricsBase.getStringBounds(fm, text, saved_gc);
	bounds = new JCUnit.Dimension(JCUnit.POINTS, extent.getWidth(),
								  extent.getHeight());
	//JLA:  fixes Bunka JCFlow bugs 003 001 and 003 003
	//superclass's status was not being updated
	setLayoutStatus(status);


	return(status);
}

/**
 * Returns an indication if this object can be split into components
 * @returns True indicating that a macro object can't be split
 */
public boolean isAtomic() {
	return(true);
}

/**
 * The symbol has no visual representation
 * @param gc The graphics context which will draw this object
 */
public void draw(Graphics2D gc) {

	AttributedString    attrib;
	float               x, y;
	int                 last;
	Font                font;

	// ignore unset or zero-length strings
	if (text == null || text.length() == 0) {
		
		return;
	}

	
	

	// Compute the location of the macro in points
	x = (float) JCUnit.getAsPoints(origin.units, origin.x);
	y = (float) JCUnit.getAsPoints(origin.units, origin.y);

	// Convert the text into an attributed string
	attrib = new AttributedString(text);

	last = text.length();
	font = style.getFont();

	// Set attributes on the string
	attrib.addAttribute(TextAttribute.SIZE, new Float(style.getPointSize()),
						0, last);

	if (style.getBaselineOffset() == JCTextStyle.OFFSET_SUBSCRIPT) {
		attrib.addAttribute(TextAttribute.SUPERSCRIPT,
							TextAttribute.SUPERSCRIPT_SUB, 0, last);
	}
	else if (style.getBaselineOffset() == JCTextStyle.OFFSET_SUPERSCRIPT) {
		attrib.addAttribute(TextAttribute.SUPERSCRIPT,
							TextAttribute.SUPERSCRIPT_SUPER, 0, last);
	}

	attrib.addAttribute(TextAttribute.FONT, font, 0, last);

	if (style.getUnderlining() == JCTextStyle.LINEMODE_UNDERLINE) {
		attrib.addAttribute(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON,
							0, last);
	}

	// draw the macro
	gc.setColor(style.getColor());
	gc.drawString(attrib.getIterator(), x, y);
}

/**
 * Create a Macro object which is a copy of this one
 * @return a clone of this text symbol
 */
public Object clone() {

	MacroRender     new_macro;

	// Clone the macro object
	new_macro = (MacroRender) super.clone();

	return((Object) new_macro);
}

}
