/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Render.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:26 $ $Locker:  $  Quest Software Inc. 
package com.klg.jclass.page.render;

import java.awt.Color;

import com.klg.jclass.page.JCUnit;




import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;


/**
 * Render describes a general render object
 */
public abstract class Render implements Cloneable {

/** The point at which the render object is located */
protected JCUnit.Point      origin;
/** The width and height of the render object */
protected JCUnit.Dimension  bounds;
/** Tracks the presence of unresolved macros in text render objects */
protected int               layoutStatus;

/*
 * Layout status (completion levels) enumeration
 */
public static final int LAYOUT_PASS_0 = 1;
public static final int LAYOUT_PASS_1 = 2;
public static final int LAYOUT_PASS_2 = 3;
public static final int LAYOUT_COMPLETE = 4;

/**
 * Create a render object
 */
public Render() {
	origin = new JCUnit.Point();
	bounds = new JCUnit.Dimension();
	layoutStatus = LAYOUT_PASS_0;
}

/**
 * Get the origin of the render object
 * @return Coordinates of the upper-left corner of the object
 */
public JCUnit.Point getLocation() {
	return origin;
}

/**
 * Set the origin of the render object
 * @param location Coordinates of the upper-left corner of the object
 */
public void setLocation(JCUnit.Point location) {
	origin = (JCUnit.Point) location.clone();
}

/**
 * Get the dimensions of the render object
 * @return Width and height of the rectangular region enclosing the object
 */
public JCUnit.Dimension getSize() {
	return bounds;
}

/**
 * Set the dimensions of the render object
 * @param size Width and height of the rectangular region enclosing the object
 */
public void setSize(JCUnit.Dimension size) {
	bounds = (JCUnit.Dimension) size.clone();
}

/**
 * Get the flag which indicates if more passes are necessary to resolve
 * any macros or print-time variables contained in the render object
 * @return Status of the render object
 */
public int getLayoutStatus() {
	return layoutStatus;
}

/**
 * Change the layout status flag to reflect manual changes to the
 * completeness of any macros contained in the render object
 * @param new_status New status to enforce on the object
 */
public void setLayoutStatus(int new_status) {

	if (new_status <= 0 || new_status > LAYOUT_COMPLETE) {
		throw new IllegalArgumentException("Layout status not valid");
	}

	layoutStatus = new_status;
}

public abstract boolean isAtomic();

/**
 * Return a copy of this render object
 * @return a clone of this object
 */
public Object clone() {

	Render  new_render;

	try {
		new_render = (Render) super.clone();
	}
	catch (CloneNotSupportedException e) {
		return(null);
	}

	new_render.bounds = (JCUnit.Dimension) this.bounds.clone();
	new_render.origin = (JCUnit.Point) this.origin.clone();

	return((Object) new_render);
}

/**
 * Compare two layout levels and return if the second is lower than the first
 * @param layoutBase base layout level for the comparison
 * @param layoutTest layout value to compare with the base
 * @return If the second is lower than the first
 */
public static boolean isLayoutLower(int layout_base, int layout_test) {

	if (layout_base <= 0 || layout_base > LAYOUT_COMPLETE ||
		layout_test <= 0 || layout_test > LAYOUT_COMPLETE) {
		throw new IllegalArgumentException("Layout status not valid");
	}

	if (layout_base > layout_test) {
		return(true);
	}
	return(false);
}

/**
 * Draw the object
 */
public void draw(Graphics2D gc) {

	// Default draw creates a box as a fallback
	double              x, y, width, height;
	Rectangle2D.Double  rect;

	// convert the dimensions of the box to points
	x = JCUnit.getAsPoints(origin.units, origin.x);
	y = JCUnit.getAsPoints(origin.units, origin.y);
	width = JCUnit.getAsPoints(bounds.units, bounds.width);
	height = JCUnit.getAsPoints(bounds.units, bounds.height);

	// Create a new rectangle
	rect = new Rectangle2D.Double(x, y, width, height);

	// Use the given graphics to draw (fill) the box
	gc.setColor(Color.gray);
	gc.fill(rect);
}

}
