/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: SymbolRender.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:27 $ $Locker:  $  Quest Software Inc. 
package com.klg.jclass.page.render;

import java.awt.Font;
import java.awt.FontMetrics;

import com.klg.jclass.page.FontMetricsBase;
import com.klg.jclass.page.JCTab;
import com.klg.jclass.page.JCTextStyle;
import com.klg.jclass.page.JCUnit;
import com.klg.jclass.page.Markup;




import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;

import java.text.AttributedString;

import java.awt.font.TextAttribute;


/**
 * SymbolRender describes a marker for a text flow action (e.g. tab, line feed)
 */
public class SymbolRender extends Render implements Cloneable {

/** The type of action this object is representing in the flow */
protected int               symbolType;

/** The style which describes the text attributes at the symbol's location */
protected JCTextStyle       style;

/** The alignment of a tab symbol */
protected int               alignment;
/** The start of the leader space for a tab symbol */
protected JCUnit.Measure    leaderStart;
/** The end of the leader space for a tab symbol */
protected JCUnit.Measure    leaderEnd;
/** The tab fill which is to be used for the leader */
protected int               leaderFill;

// The types of symbol renders
public static final int TAB_SYMBOL = 1;
public static final int NEWLINE_SYMBOL = 2;
public static final int NEWPARAGRAPH_SYMBOL = 3;
// NOTE:  We will need markers for floating images to accomplish full reflow
// public static final int FLOAT_MARKER_SYMBOL = 4;

/**
 * Create a symbol render object
 */
public SymbolRender() {

	symbolType = 0;
	style = null;
}

/**
 * Create a symbol render object given its type and a style
 * @param type the type of symbol this describes
 * @param style the text style to use for formatting the symbol
 */
public SymbolRender(int type, JCTextStyle style) {

	symbolType = type;
	this.style = style;
}

/**
 * Get the style assigned to this symbol
 * @return the text style assigned to this text symbol
 */
public JCTextStyle getStyle() {
	return style;
}

/**
 * Set the style used to determine attributes of this symbol
 * @param style The text style used to format this symbol
 */
public void setStyle(JCTextStyle style) {

	this.style = style;
}

/**
 * Get the type of the symbol object
 * @return the type of marker this symbol describes
 */
public int getType() {
	return symbolType;
}

/**
 * Set the type of the symbol
 * @param type The type of symbol this describes
 */
public void setType(int type) {

	if (type <= 0 || type > NEWLINE_SYMBOL) {
		throw new IllegalArgumentException("Unrecognised symbol type");
	}

	symbolType = type;
}

/**
 * Get the alignment of a symbol
 * @return The alignment the object this symbol represents has
 */
public int getAlignment() {
	return(alignment);
}

/**
 * Set the alignment of a symbol
 * @param alignment The alignment the object this symbol represents has
 */
public void setAlignment(int alignment) {
	this.alignment = alignment;
}

/**
 * Get the start of the leader space available to a tab symbol
 * @return The start position for a tab symbol's leader space
 */
public JCUnit.Measure getLeaderStart() {
	return(leaderStart);
}

/**
 * Set the start of the leader space available to a tab symbol
 * @param leader The start position for a tab symbol's leader space
 */
public void setLeaderStart(JCUnit.Measure leader) {
	leaderStart = leader;
}

/**
 * Get the end of the leader space available to a tab symbol
 * @return The end position for a tab symbol's leader space
 */
public JCUnit.Measure getLeaderEnd() {
	return(leaderEnd);
}

/**
 * Set the end of the leader space available to a tab symbol
 * @param leader The end position for a tab symbol's leader space
 */
public void setLeaderEnd(JCUnit.Measure leader) {
	leaderEnd = leader;
}

/**
 * Get the type of fill applied to a tab symbol
 * @return The type of fill applied to this tab
 */
public int getLeaderFill() {
	return(leaderFill);
}

/**
 * Set the type of fill applied to a tab symbol
 * @param fillType The type of fill applied to this tab
 */
public void setLeaderFill(int fillType) {
	leaderFill = fillType;
}

/**
 * Returns an indication if this object can be split into components
 * @returns True indicating that a symbol object (zero width) can't be split
 */
public boolean isAtomic() {
	return(true);
}

/**
 * The symbol has no visual representation
 * @param gc The graphics context which will draw this object
 */
public void draw(Graphics2D gc) {

	int                 i, count, last;
	float               x, y;
	double              space, width, basic_length, add_length;
	String              basic_pattern, augmented_pattern, atomic_unit;
	String              fill_string;
	StringBuffer        fill_text;
	FontMetrics         fm;
	AttributedString    attrib;
	Rectangle2D         extent1, extent2;
	Font                font;
	JCUnit.Measure      leader_length;

	// Draw a tab leader if it is visible
	if (symbolType == TAB_SYMBOL && leaderFill != JCTab.TAB_FILL_NONE) {

		// Compute the amount of space to be filled
		leader_length = (JCUnit.Measure) leaderEnd.clone();
		leader_length.subtract(leaderStart);
		space = leader_length.getAs(JCUnit.POINTS);

		// Determine the pattern to use to create the leader
		if (leaderFill == JCTab.TAB_FILL_UNDERLINE) {
			basic_pattern = "  ";
			augmented_pattern = "   ";
			atomic_unit = " ";
		}
		else if (leaderFill == JCTab.TAB_FILL_DOTS) {
			basic_pattern = ". .";
			augmented_pattern = ". . .";
			atomic_unit = " .";
		}
		else {
			// unknown fill type, so just return
			return;
		}

		// Compute the length of the strings used to fill the tab
		font = style.getFont();
		fm = gc.getFontMetrics(font);
		extent1 = FontMetricsBase.getStringBounds(fm, basic_pattern, gc);
		extent2 = FontMetricsBase.getStringBounds(fm, augmented_pattern, gc);

		basic_length = extent1.getWidth();
		add_length = extent2.getWidth() - basic_length;

		// if there is insufficient space for the basic unit, ignore it
		if (space < basic_length) {
			return;
		}

		// determine the number of additions needed to make the leader
		count = (int) ((space - basic_length) / add_length);

		// compose the string
		fill_text = new StringBuffer(basic_pattern);
		for (i = 0; i < count; i++) {
			fill_text.append(atomic_unit);
		}
		fill_string = new String(fill_text);

		// Compute the location of the string
		extent1 = FontMetricsBase.getStringBounds(fm, fill_string, gc);
		width = extent1.getWidth();
		x = (float) leaderStart.getAs(JCUnit.POINTS);
		x += (float) (space - width) / 2.0;
		y = (float) JCUnit.getAsPoints(origin.units, origin.y);

		// Create an attributed string to render the leader and set its size
		attrib = new AttributedString(fill_string);
		last = fill_string.length();

		attrib.addAttribute(TextAttribute.SIZE, new Float(style.getPointSize()),
							0, last);

		// Turn underlining on for underline leaders
		if (leaderFill == JCTab.TAB_FILL_UNDERLINE) {
			attrib.addAttribute(TextAttribute.UNDERLINE,
								TextAttribute.UNDERLINE_ON, 0, last);
		}

		// Draw the leader
		gc.setColor(style.getColor());
		gc.drawString(attrib.getIterator(), x, y);
	}

	else if (symbolType == NEWLINE_SYMBOL && gc instanceof Markup) {
		((Markup) gc).newLine();
	}

	else if (symbolType == NEWPARAGRAPH_SYMBOL && gc instanceof Markup) {
		((Markup) gc).newParagraph();
	}
}

/**
 * Create a Symbol object which is a copy of this one
 * @return a clone of this text symbol
 */
public Object clone() {

	SymbolRender    new_symbol;

	// Clone the symbol object
	new_symbol = (SymbolRender) super.clone();

	// Copy dependent objects
	if (this.leaderStart != null) {
		new_symbol.leaderStart = (JCUnit.Measure) this.leaderStart.clone();
	}
	if (this.leaderEnd != null) {
		new_symbol.leaderEnd = (JCUnit.Measure) this.leaderEnd.clone();
	}

	return((Object) new_symbol);
}

}
