/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

/*
 * $Id: FontFileReader.java,v 1.2 2006-03-17 21:13:27 lrdinuki Exp $
 * Copyright (C) 2001 The Apache Software Foundation. All rights reserved.
 * For details on use and redistribution please refer to the
 * LICENSE file included with these sources.
 */

package com.klg.jclass.page.ttf;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.File;
import java.io.IOException;

/**
 * This class was originally part of Apache FOP but has been modified
 * by Quest Software to fix some bugs and make it more useful to our
 * code.
 *
 * Reads a file into an array and
 * provides file like functions for array access.
 */
public class FontFileReader {
    private int fsize;      // file size
    private int current;    // current position in file
    private byte[] file;

//ToDo: we probably want a smarter font file reader class
//  that only reads what it knows it needs, or reads and returns bytes
//  but doesn't store them in an array -- useful for .ttc files which
//  hold multiple fonts and for the initial parsing when all we want is
//  the name table

    /**
     * Initializes class and reads stream. Init does not close stream
     * @param stream InputStream to read from
     * @param start initial size av array to read to
     * @param inc if initial size isn't enough, create
     * new array with size + inc
     */
    private void init(InputStream stream, int start,
                      int inc) throws java.io.IOException {
        fsize = 0;
        current = 0;

		// initialize the remiander -- this isn't the actual value yet,
		// that will be calculated below
		int remainder = start;

        file = new byte[start];

		// attempt to read the first batch of bytes from the input stream
        int l = stream.read(file, 0, start);
		if (l > 0)
        	fsize += l;

		// while there are still bytes remaining in the imput stream
        while (l > 0) {
			// if number of bytes read equals the remainder, expand the
			// byte array and read some more of it
        	if (l == remainder) {
            	byte[] tmpbuf;

            	tmpbuf = new byte[file.length + inc];
                System.arraycopy(file, 0, tmpbuf, 0, file.length);
				remainder = inc;
                l = stream.read(tmpbuf, file.length, remainder);
				if (l > 0) {
                	fsize += l;
				}
                file = tmpbuf;

			// if number of bytes read is less than what was requested,
			// calculate the remaining number of bytes in the request and
			// try again
        	} else if (l < remainder) {
				remainder = remainder - l;
        		l = stream.read(file, fsize, remainder);
				if (l > 0) {
                	fsize += l;
				}
			}
        }
    }

    /**
     * Constructor.  Read specified font file, using its length as the initial
     * size of the byte array.
     * @param fileName filename to read
     */
    public FontFileReader(String fileName) throws java.io.IOException {

        // Get estimates for file size and increment
        File f = new File(fileName);
        FileInputStream ins = new FileInputStream(fileName);
        init(ins, (int)(f.length() + 1), (int)(f.length() / 10));
        ins.close();
    }

    /**
     * Constructor.  Read font file from InputStream, using default
     * size for initial size of byte array.
     * @param ins InputStream to read
     */
    public FontFileReader(InputStream ins) throws java.io.IOException {

        // Get estimates for file size and increment
        this(ins, -1);
    }

    /**
     * Constructor.  Read font file from InputStream, using passed file
     * size for initial size of byte array.
     * @param fileSize length of InputStream
     * @param ins InputStream to read
     */
    public FontFileReader(InputStream ins, int fileSize) throws java.io.IOException {
		if (fileSize < 0)
			fileSize = 50000;

        // Get estimates for file size and increment
        init(ins, fileSize + 1, fileSize / 10);
    }

	/**
     * Constructor.  Don't read any files, just pretend that the pass
     * byte array is a file.  Useful for testing.
     */
    public FontFileReader(byte[] bytes) throws java.io.IOException {

		file = bytes;
		fsize = bytes.length;
		current = 0;
    }

    /**
     * Set current file position to offset
     */
    public void seek_set(long offset) throws IOException {
        if (offset > fsize || offset < 0)
            throw new java.io.EOFException("Reached EOF, file size=" + fsize
                                           + " offset=" + offset);
        current = (int)offset;
    }

    /**
     * Set current file position to offset
     */
    public void seek_add(long add) throws IOException {
        seek_set(current + add);
    }

	/**
     * Skip the requested number of bytes.
     */
    public void skip(long add) throws IOException {
        seek_add(add);
    }

    /**
     * return current file position
     */
    public int getCurrentPos() {
        return current;
    }

	/**
     * Get the size of the file.
     */
    public int getFileSize() {
        return fsize;
    }

    /**
     * Read 1 byte, throws EOFException on end of file
     */
    public byte read() throws IOException {
        if (current > fsize)
            throw new java.io.EOFException("Reached EOF, file size=" + fsize);

        byte ret = file[current++];
        return ret;
    }

    /**
     * Read 1 signed byte from InputStream
     */
    public final byte readTTFByte() throws IOException {
        return read();
    }

    /**
     * Read 1 unsigned byte from InputStream
     */
    public final int readTTFUByte() throws IOException {
        byte buf = read();

        if (buf < 0)
            return (int)(256 + buf);
        else
            return (int)buf;
    }

    /**
     * Read 2 bytes signed from InputStream
     */
    public final short readTTFShort() throws IOException {
        int ret = (readTTFUByte() << 8) + readTTFUByte();
        short sret = (short)ret;

        return sret;
    }

    /**
     * Read 2 bytes unsigned from InputStream
     */
    public final int readTTFUShort() throws IOException {
        int ret = (readTTFUByte() << 8) + readTTFUByte();

        return (int)ret;
    }

    /**
     * Write a USHort at a given position
     */
    public final void writeTTFUShort(int pos, int val) throws IOException {
        if ((pos + 2) > fsize)
            throw new java.io.EOFException("Reached EOF");
        byte b1 = (byte)((val >> 8) & 0xff);
        byte b2 = (byte)(val & 0xff);
        file[pos] = b1;
        file[pos + 1] = b2;
    }

    /**
     * Read 2 bytes signed from InputStream at position pos
     * without changing current position
     */
    public final short readTTFShort(long pos) throws IOException {
        long cp = getCurrentPos();
        seek_set(pos);
        short ret = readTTFShort();
        seek_set(cp);
        return ret;
    }

    /**
     * Read 2 bytes unsigned from InputStream at position pos
     * without changing current position
     */
    public final int readTTFUShort(long pos) throws IOException {
        long cp = getCurrentPos();
        seek_set(pos);
        int ret = readTTFUShort();
        seek_set(cp);
        return ret;
    }

    /**
     * Read 4 bytes from InputStream
     */
    public final int readTTFLong() throws IOException {
        long ret = readTTFUByte();    // << 8;
        ret = (ret << 8) + readTTFUByte();
        ret = (ret << 8) + readTTFUByte();
        ret = (ret << 8) + readTTFUByte();

        return (int)ret;
    }

    /**
     * Read 4 bytes from InputStream
     */
    public final long readTTFULong() throws IOException {
        long ret = readTTFUByte();
        ret = (ret << 8) + readTTFUByte();
        ret = (ret << 8) + readTTFUByte();
        ret = (ret << 8) + readTTFUByte();

        return ret;
    }

    /**
     * Read 4 bytes from InputStream in Fixed Decimal format
     */
    public final double readTTFFixedDecimal() throws IOException {
        int intVal = (int) readTTFShort();
        int decVal = (int) readTTFUShort();

		double ret = ((double) intVal) + ((double) decVal) / 65536.0;

        return ret;
    }

    /**
     * Read a 0 terminated ISO-8859-1 string
     */
    public final String readTTFString() throws IOException {
		return readTTFString("ISO-8859-1");
    }

    public final String readTTFString(String encoding) throws IOException {
        int i = current;
        while (file[i++] != 0) {
            if (i > fsize)
                throw new java.io.EOFException("Reached EOF, file size="
                                               + fsize);
        }

        byte[] tmp = new byte[i - current];
        System.arraycopy(file, current, tmp, 0, i - current);
        return new String(tmp, encoding);
    }


    /**
     * Read an ISO-8859-1 string of len bytes
     */
    public final String readTTFString(int len) throws IOException {
        return readTTFString(len, "ISO-8859-1");
	}

    public final String readTTFString(int len, String encoding) throws IOException {
        if ((len + current) > fsize)
            throw new java.io.EOFException("Reached EOF, file size=" + fsize);

        byte[] tmp = new byte[len];
        System.arraycopy(file, current, tmp, 0, len);
        current += len;
        return new String(tmp, encoding);
    }

    /**
     * Read the requested number of bytes.
     * @throws IOException if out of bounds
     */
    public byte[] readBytes(int length) throws java.io.IOException {
        if ((current + length) > fsize)
            throw new java.io.IOException("Reached EOF");

        byte[] ret = new byte[length];
        System.arraycopy(file, current, ret, 0, length);
        current += length;
        return ret;
    }

    /**
     * Return a copy of the internal array
     * @throws IOException if out of bounds
     */
    public byte[] getBytes(int offset,
                           int length) throws java.io.IOException {
        if ((offset + length) > fsize)
            throw new java.io.IOException("Reached EOF");

        byte[] ret = new byte[length];
        System.arraycopy(file, offset, ret, 0, length);
        return ret;
    }

}
