/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

/*
 * $Id: TTFFile.java,v 1.2 2006-03-17 21:13:27 lrdinuki Exp $
 * Copyright (C) 2001 The Apache Software Foundation. All rights reserved.
 * For details on use and redistribution please refer to the
 * LICENSE file included with these sources.
 */

package com.klg.jclass.page.ttf;

import com.klg.jclass.page.resources.*;

import java.io.*;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.Vector;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.TreeMap;
import java.util.Set;

/**
 * This class was originally part of Apache FOP but has been modified
 * by Quest Software to fix some bugs and make it more useful to our
 * product.
 *
 * Reads a TrueType file or a TrueType Collection.
 * The TrueType spec can be found at the Microsoft
 * Typography site: http://www.microsoft.com/truetype/
 */
public class TTFFile {

// status enums
public static final int UNPARSED = 0;
public static final int PARSE_OK = 1;
public static final int PARSE_WARNING = 2;
public static final int PARSE_FATAL = 3;

static final byte NTABS = 24;
static final int NMACGLYPHS = 258;
static final int MAX_CHAR_CODE = 255;
static final int ENC_BUF_SIZE = 1024;

static String encoding = "WinAnsiEncoding";    // Deafult encoding
short firstChar = 0;
boolean is_embeddable = true;
boolean hasSerifs = true;
Hashtable dirTabs;                      // Table directory
Hashtable kerningTab;                   // for CIDs
Hashtable ansiKerningTab;               // For winAnsiEncoding
Vector cmaps;
Vector unicodeMapping;                  //

/* table used to convert unicode values to glyph codes */
Hashtable unicodeToGlyphTable;

/* table used to story composit glyphs */
Hashtable compositeGlyphs;

double version;							// version from "head" table
int upem;                               // unitsPerEm from "head" table
int nhmtx;                              // Number of horizontal metrics
int post_format;
int loca_format;
long lastLoca = 0;                      // offset to last loca
int nglyphs;                            // Number of glyphs in font
// (read from "maxp" table)
int nmglyphs;                           // Used in fixWidths - remove?

TTFMtxEntry mtx_tab[];                  // Contains glyph data
int[] mtx_encoded = null;

String fontName = "";
String fullName = "";
String notice = "";
String familyName = "";
String subFamilyName = "";

double italicAngle = 0.0;
long isFixedPitch = 0;
int fontBBox1 = 0;
int fontBBox2 = 0;
int fontBBox3 = 0;
int fontBBox4 = 0;
int capHeight = 0;
int underlinePosition = 0;
int underlineThickness = 0;
int xHeight = 0;
int ascender = 0;
int descender = 0;

short lastChar = 0;

int ansiWidth[];
Hashtable ansiIndex;

long dirTabOffset = 0;

// panose value from OS/2 table
byte[] panose = null;
// IBM family class and subclass from OS/2 table
int ibmFamilyClass = 0;
int ibmFamilySubclass = 0;

int status = UNPARSED;
StringBuffer statusMessage = null;
String fontFileName = null;


/**
 * Position inputstream to position indicated
 * in the dirtab offset + offset
 */
void seek_tab(FontFileReader in, String name,
              long offset) throws IOException {
    TTFDirTabEntry dt = (TTFDirTabEntry)dirTabs.get(name);
    if (dt == null) {
        System.err.println("Warning: TTF: Dirtab " + name + " not found.");
        return;
    }

    in.seek_set(dt.offset + offset);
}

/**
 * Return status of the TTFFile parsing.  Possible values are
 * UNPARSED, PARSE_OK, PARSE_WARNING, PARSE_FATAL.
 */
public int getStatus()
{
    return status;
}

/**
 * Set status of the TTFFile parsing.  Status is only
 * recorded if it is worse than the current status.  In other
 * words, don't use this method to reset the status to a lesser
 * level.
 */
public void setStatus(int newStatus)
{
    if (newStatus > status) {
        status = newStatus;
    }
}

/**
 * Add a string to the current status message.
 */
public void addToStatusMessage(String message)
{
    if (message == null) {
        return;
    }

    if (statusMessage == null) {
        statusMessage = new StringBuffer();
    }

    boolean first = false;
    if (statusMessage.length() == 0) {
        first = true;
    }

    if (!first) {
        statusMessage.append('\n');
    }

    statusMessage.append(message);
}

/**
 * Retrive the current status message.
 */
public String getStatusMessage()
{
    if (statusMessage == null) {
        return "";
    }
    return statusMessage.toString();
}

/**
 * Record the font file name description associted with this TTF file.
 * Used for more descriptive error messages.
 */
public void setFontFileName(String fontFileName)
{
    this.fontFileName = fontFileName;
}

/**
 * Add a warning that the current font is not a Unicode font
 * to the current status message.
 */
protected void addNoUnicodeStatusMessage()
{
    StringBuffer newMessage = new StringBuffer();
    newMessage.append(LocaleBundle.string(LocaleBundle.NO_UNICODE_CMAP));
    String fileNameHolder = LocaleBundle.FILE_NAME_HOLDER;
    int fileNameIndex = newMessage.toString().indexOf(fileNameHolder);
    if (fileNameIndex >= 0) {
        newMessage.replace(fileNameIndex,
             fileNameIndex + fileNameHolder.length(),
             fullName);
    }
    addToStatusMessage(newMessage.toString());
    if (fontFileName != null) {
        newMessage = new StringBuffer();
        newMessage.append("    (");
        newMessage.append(fontFileName);
        newMessage.append(")");
        addToStatusMessage(newMessage.toString());
    }
}

/**
 * Add a warning that the current font is not a Unicode font
 * to the current status message.
 */
protected void addInvalidFileStatusMessage()
{
    StringBuffer newMessage = new StringBuffer();
    newMessage.append(LocaleBundle.string(LocaleBundle.INVALID_TTF));
    addToStatusMessage(newMessage.toString());
    if (fontFileName != null) {
        newMessage = new StringBuffer();
        newMessage.append("    (");
        newMessage.append(fontFileName);
        newMessage.append(")");
        addToStatusMessage(newMessage.toString());
    }
}

/**
 * Convert from truetype unit to pdf unit based on the
 * unitsPerEm field in the "head" table
 * @param n truetype unit
 * @return pdf unit
 */
public int get_ttf_funit(int n) {
    int ret;
    if (n < 0) {
        long rest1 = n % upem;
        long storrest = 1000 * rest1;
        long ledd2 = rest1 / storrest;
        ret = -((-1000 * n) / upem - (int)ledd2);
    }
    else {
        ret = (n / upem) * 1000 + ((n % upem) * 1000) / upem;
    }

    return ret;
}

/**
 * Read the cmap table,
 * return false if the table is not present or only unsupported
 * tables are present. Currently only unicode cmaps are supported.
 * Set the unicodeIndex in the TTFMtxEntries and fills in the
 * cmaps vector.
 */
private boolean readCMAP(FontFileReader in) throws IOException {

    unicodeMapping = new Vector();
    unicodeToGlyphTable = new Hashtable();

    /**
     * Read CMAP table and correct mtx_tab.index
     */
    int mtxPtr = 0;

    seek_tab(in, "cmap", 2);
    int num_cmap = in.readTTFUShort();    // Number of cmap subtables
    long cmap_unioffset = 0;


 
    
    /*
    * Read offset for all tables
    * We are only interested in the unicode table
    */
    for (int i = 0; i < num_cmap; i++) {
        int cmap_pid = in.readTTFUShort();
        int cmap_eid = in.readTTFUShort();
        long cmap_offset = in.readTTFULong();


     

        if (cmap_pid == 3 && cmap_eid == 1)
            cmap_unioffset = cmap_offset;
    }

    if (cmap_unioffset <= 0) {
        addNoUnicodeStatusMessage();
        setStatus(PARSE_WARNING);
        return false;
    }

    // Read unicode cmap
    seek_tab(in, "cmap", cmap_unioffset);
    int cmap_format = in.readTTFUShort();
    int cmap_length = in.readTTFUShort();


     

    if (cmap_format == 4) {

          //int cmap_lang = in.readTTFUShort();          
          //System.err.println("     table lang " + cmap_lang);
        in.skip(2);    // Skip version number

        int cmap_segCountX2 = in.readTTFUShort();
        int cmap_searchRange = in.readTTFUShort();
        int cmap_entrySelector = in.readTTFUShort();
        int cmap_rangeShift = in.readTTFUShort();


 

        int cmap_endCounts[] = new int[cmap_segCountX2 / 2];
        int cmap_startCounts[] = new int[cmap_segCountX2 / 2];
        int cmap_deltas[] = new int[cmap_segCountX2 / 2];
        int cmap_rangeOffsets[] = new int[cmap_segCountX2 / 2];

        for (int i = 0; i < (cmap_segCountX2 / 2); i++) {
            cmap_endCounts[i] = in.readTTFUShort();
        }

        in.skip(2);    // Skip reservedPad

        for (int i = 0; i < (cmap_segCountX2 / 2); i++) {
            cmap_startCounts[i] = in.readTTFUShort();
        }

        for (int i = 0; i < (cmap_segCountX2 / 2); i++) {
            cmap_deltas[i] = in.readTTFShort();
        }

        int startRangeOffset = in.getCurrentPos();

        for (int i = 0; i < (cmap_segCountX2 / 2); i++) {
            cmap_rangeOffsets[i] = in.readTTFUShort();
        }

        int glyphIdArrayOffset = in.getCurrentPos();

        // Insert the unicode id for the glyphs in mtx_tab
        // and fill in the cmaps Vector

        for (int i = 0;
             i < cmap_startCounts.length && cmap_startCounts[i] != 0xffff;
             i++) {

 
            for (int j = cmap_startCounts[i]; j <= cmap_endCounts[i];
                 j++) {

                // Update lastChar
                if (j < 256 && j > lastChar)
                    lastChar = (short)j;

                if (mtxPtr < mtx_tab.length) {
                    int glyphIdx;
                    if (cmap_rangeOffsets[i] != 0) {
                        int glyphOffset =
                            glyphIdArrayOffset
                            + ((cmap_rangeOffsets[i] / 2) + (j - cmap_startCounts[i]) + (i) - cmap_segCountX2 / 2)
                            * 2;
                        in.seek_set(glyphOffset);
                        glyphIdx = (in.readTTFUShort() + cmap_deltas[i])
                            & 0xffff;

                        unicodeMapping.addElement(new UnicodeMapping(glyphIdx,
                                                                     j));
                        unicodeToGlyphTable.put(new Integer(j),
                                                new Integer(glyphIdx));

                        mtx_tab[glyphIdx].unicodeIndex.addElement(new Integer(j));

                        // Also add winAnsiWidth
                        Vector v =
                            (Vector)ansiIndex.get(new Integer(j));
                        if (v != null) {
                            for (Enumeration e = v.elements();
                                 e.hasMoreElements();) {
                                Integer aIdx =
                                    (Integer)e.nextElement();
                                ansiWidth[aIdx.intValue()] =
                                    mtx_tab[glyphIdx].wx;

 
                            }
                        }

 

                    }
                    else {

                        glyphIdx = (j + cmap_deltas[i]) & 0xffff;

                        if (glyphIdx < mtx_tab.length) {
                            mtx_tab[glyphIdx].unicodeIndex.addElement(new Integer(j));
                        }
                        else {
                            System.err.println("Glyph " + glyphIdx
                                               + " out of range: "
                                               + mtx_tab.length);
                        }

                        unicodeMapping.addElement(new UnicodeMapping(glyphIdx,
                                                                     j));
                        unicodeToGlyphTable.put(new Integer(j),
                                                new Integer(glyphIdx));


                        // Also add winAnsiWidth
                        Vector v =
                            (Vector)ansiIndex.get(new Integer(j));
                        if (v != null) {
                            for (Enumeration e = v.elements();
                                 e.hasMoreElements();) {
                                Integer aIdx =
                                    (Integer)e.nextElement();
                                ansiWidth[aIdx.intValue()] =
                                    mtx_tab[glyphIdx].wx;
                            }
                        }


 
                    }
                    if (glyphIdx < mtx_tab.length) {
                        if (mtx_tab[glyphIdx].unicodeIndex.size() < 2) {
                            mtxPtr++;
                        }
                    }
                }
            }
        }
    }
    return true;
}

/**
 * Test whether this font is a Unicode TTF by looking at the CMAP table.
 */
private boolean isUnicode(FontFileReader in) throws IOException {

    /**
     * Read CMAP table and correct mtx_tab.index
     */
    int mtxPtr = 0;

    seek_tab(in, "cmap", 2);
    int num_cmap = in.readTTFUShort();    // Number of cmap subtables
    long cmap_unioffset = 0;

    // System.err.println(num_cmap+" cmap tables");

    /*
     * Read offset for all tables
     * We are only interested in the unicode table
     */
    for (int i = 0; i < num_cmap; i++) {
        int cmap_pid = in.readTTFUShort();
        int cmap_eid = in.readTTFUShort();
        long cmap_offset = in.readTTFULong();

        // System.err.println("Platform ID: "+cmap_pid+
        // " Encoding: "+cmap_eid);

        if (cmap_pid == 3 && cmap_eid == 1)
            cmap_unioffset = cmap_offset;
    }

    if (cmap_unioffset <= 0) {
        return false;
    }
    return true;
}

/**
 * Print first char/last char
 */
private void print_max_min() {
    int min = 255;
    int max = 0;
    for (int i = 0; i < mtx_tab.length; i++) {
        if (mtx_tab[i].index < min)
            min = mtx_tab[i].index;
        if (mtx_tab[i].index > max)
            max = mtx_tab[i].index;
    }
    System.err.println("Min: " + min);
    System.err.println("Max: " + max);
}

public void readFont(FontFileReader in) throws IOException {
    readFont(in, (String)null);
}

/**
 * initialize the ansiWidths array (for winAnsiEncoding)
 * and fill with the missingwidth
 */
private void initAnsiWidths() {
    ansiWidth = new int[256];
    for (int i = 0; i < 256; i++)
        ansiWidth[i] = mtx_tab[0].wx;

    // Create an index hash to the ansiWidth
    // Can't just index the winAnsiEncoding when inserting widths
    // same char (eg bullet) is repeated more than one place
    ansiIndex = new Hashtable();
    for (int i = 32; i < Glyphs.winAnsiEncoding.length; i++) {
        Integer ansi = new Integer(i);
        Integer uni = new Integer((int)Glyphs.winAnsiEncoding[i]);

        Vector v = (Vector)ansiIndex.get(uni);
        if (v == null) {
            v = new Vector();
            ansiIndex.put(uni, v);
        }
        v.addElement(ansi);
    }
}


/**
 * Read the font data
 * If the fontfile is a TrueType Collection (.ttc file)
 * The name of the font to read data for must be supplied,
 * else the name is ignored
 */
public void readFont(FontFileReader in, String name) throws IOException {

	setStatus(PARSE_OK);

    /*
    * Check if TrueType collection, and that the name
    * exists in the collection
    */
    if (!checkTTC(in, name, false))
        throw new IOException("Failed to read font");

    if (!readDirTabs(in)) {
        addInvalidFileStatusMessage();
        setStatus(PARSE_FATAL);
        return;
    }

    readName(in);
    if (!isUnicode(in)) {
        addNoUnicodeStatusMessage();
        setStatus(PARSE_FATAL);
        return;
    }
    readFontHeader(in);
    getNumGlyphs(in);
    //System.err.println("Number of glyphs in font: " + nglyphs);
    readHorizontalHeader(in);
    readHorizontalMetrics(in);
    initAnsiWidths();
    readPostscript(in);
    readOS2(in);
    readIndexToLocation(in);
    readGlyf(in);
    readPCLT(in);
    if (readCMAP(in)) {     // Read cmap table and fill in ansiwidths
        createCMaps();    // Create cmaps for bfentries
        readKerning(in);
    } else {
        setStatus(PARSE_FATAL);
        return;
    }
    // print_max_min();
}

/**
 * Write the font out, reduced by unnecessary
 * font tables and unused glyphs.
 */
public void writeReducedFont(OutputStream os, FontFileReader in,
                             SortedSet ansiCharactersUsed, SortedSet unicodeGlyphsUsed)
    throws IOException {
    // combine glyphs used by the unicode portion of the font and
    // the glyphs corresponding to the characters used in the ansi
    // portion of the font into one big table of all glyphs used from
    // this font

    SortedSet glyphsUsed = null;
    // Use glyphs derived from unicode as starting point for glyph list
    if (unicodeGlyphsUsed != null) {
        glyphsUsed = new TreeSet(unicodeGlyphsUsed);

        // include all glyphs making up glyphs that happen to be composite
        Iterator glyphIterator = unicodeGlyphsUsed.iterator();
        while (glyphIterator.hasNext()) {
            Integer glyph = (Integer)glyphIterator.next();
            List compositeList = (List)compositeGlyphs.get(glyph);
            if (compositeList != null) {
                for (int i = 0; i < compositeList.size(); i++) {
                    glyphsUsed.add(compositeList.get(i));
                }
            }
        }
        // no unicode being used -- start with empty set
    }
    else {
        glyphsUsed = new TreeSet();
    }

    // now convert all of the ansi characters being used to 
    // glyphs and add these into the list of glyphs used
    if (ansiCharactersUsed != null) {
        Iterator ansiIterator = ansiCharactersUsed.iterator();
        while (ansiIterator.hasNext()) {
            Integer ansiChar = (Integer)ansiIterator.next();

            // special case -- if WinAnsiEncoding value for the
            // Euro character found in Ansi characters, 
            // replace it with the actual unicode value for it
            if (ansiChar.intValue() == 128) {
                ansiChar = new Integer(0x20ac);
            }

            // add glyph to list of glyphs used
            Integer ansiGlyph = (Integer)unicodeToGlyphTable.get(ansiChar);
            if (ansiGlyph != null) {
                glyphsUsed.add(ansiGlyph);

                // if the glyph is composite, add the glyphs that it
                // references as well
                List compositeList = (List)compositeGlyphs.get(ansiGlyph);
                if (compositeList != null) {
                    for (int i = 0; i < compositeList.size(); i++) {
                        glyphsUsed.add(compositeList.get(i));
                    }
                }
            }
        }
    }

    // position input, store number of directory tables
    in.seek_set(dirTabOffset);
    int numDirTabs = dirTabs.size();

    // create new 'glyf' table with only the glyphs we need
    // also create new 'loca' table indexing into the new 'glyf' table

    ByteArrayOutputStream glyfOS = new ByteArrayOutputStream();
    ByteArrayOutputStream locaOS = new ByteArrayOutputStream();

    reduceGLYF(in, glyfOS, locaOS, glyphsUsed);

    byte[] locaBytes = locaOS.toByteArray();
    byte[] glyfBytes = glyfOS.toByteArray();

    // if file has EBLC/EBDT tables, create new reduced versions of these
    // tables with only the glyphs currently being used
    
    in.seek_set(dirTabOffset);

    byte[] eblcBytes = null;
    byte[] ebdtBytes = null;

    if (dirTabs.get("EBLC") != null && dirTabs.get("EBDT") != null) {
        TTFDirTabEntry eblcEntry = (TTFDirTabEntry)dirTabs.get("EBLC");
        TTFDirTabEntry ebdtEntry = (TTFDirTabEntry)dirTabs.get("EBDT");

        ByteArrayOutputStream eblcOS = new ByteArrayOutputStream();
        ByteArrayOutputStream ebdtOS = new ByteArrayOutputStream();

        reduceEBLC(in, eblcOS, ebdtOS, glyphsUsed);

        eblcBytes = eblcOS.toByteArray();
        ebdtBytes = ebdtOS.toByteArray();

        // if the tables have reduced down to nothing, don't include
        // them at all
        if (eblcBytes.length == 0) {
            numDirTabs--;
            eblcBytes = null;
        }
        if (ebdtBytes.length == 0) {
            numDirTabs--;
            ebdtBytes = null;
        }
    }

    // don't include vhea and vmtx tables -- PDF doesn't use them
    if (dirTabs.get("vhea") != null) {
        numDirTabs--;
    }
    if (dirTabs.get("vmtx") != null) {
        numDirTabs--;
    }
    // don't include DSIG table -- bound to be wrong if I just copy
    // it rather than run through the actual digital signature algorithm
    if (dirTabs.get("DSIG") != null) {
        numDirTabs--;
    }

    // remove cmap table from output if not using ansi characters
    boolean removeCmap = false;
    if (ansiCharactersUsed == null || ansiCharactersUsed.size() == 0) {
        if (dirTabs.get("cmap") != null) {
            numDirTabs--;
            removeCmap = true;
        }
    }

    in.seek_set(dirTabOffset);

    // Note: it's possible that we can get away with just the following
    // tables (that's what the fop class TTFFileSubSet says, although
    // I'm not sure they're entirely corret): "cmap" (ansi fonts only),
    // "cvt ", "fpgm", "glyf", "head", "hhea", "hmtx", "loca", "maxp"
    // and "prep".

    // keep track of offset to tables -- begins at end of directory table
    long runningOffset = 4 + (4 * 2) + (numDirTabs * 16);
    // keep track of checksum for entire font
    long fullCheckSum = 0;

    // begin output with directory header
    ByteArrayOutputStream dirOS = new ByteArrayOutputStream();
    dirOS.write(in.readBytes(4));
    dirOS.write(shortToBytes(numDirTabs));
    dirOS.write(in.readBytes(6));

    // have to write table entries alphabetically in order of keys!
    TreeMap sortedTabKeys = new TreeMap(dirTabs);
    Set tabKeyNames = sortedTabKeys.keySet();

    // write the directory entry for each table to the output
    Iterator iter = tabKeyNames.iterator();
    while (iter.hasNext()) {
        String tabKey = (String)iter.next();

        TTFDirTabEntry tabEntry = (TTFDirTabEntry)dirTabs.get(tabKey);

        byte[] tableBytes = null;
        // used modified/reduced tables if necessary
        if (tabKey.equals("glyf")) {
            tableBytes = glyfBytes;
        }
        else if (tabKey.equals("loca")) {
            tableBytes = locaBytes;
        }
        else if (tabKey.equals("EBLC")) {
            if (eblcBytes == null)
                continue;
            tableBytes = eblcBytes;
        }
        else if (tabKey.equals("EBDT")) {
            if (ebdtBytes == null)
                continue;
            tableBytes = ebdtBytes;
            // never include vertical header or vertical metrics
            // PDF doesn't recognize vertical information unless its
            // in its own format
        }
        else if (tabKey.equals("vhea") || tabKey.equals("vmtx") ||
            tabKey.equals("DSIG")) {
            continue;
            // if possible, remove the cmap table
        }
        else if (removeCmap && tabKey.equals("cmap")) {
            continue;
            // otherwise, just use the table as it exists in the input
        }
        else {
            tableBytes = in.getBytes((int)tabEntry.offset,
                                     (int)tabEntry.length);
        }

        // checkSumAdjustment field in 'head' table must be zero
        // when checksum for table is calculated
        
        if (tabKey.equals("head")) {
            tableBytes[8] = (byte)0;
            tableBytes[9] = (byte)0;
            tableBytes[10] = (byte)0;
            tableBytes[11] = (byte)0;
        }

        // output directory entries -- tag, checksum, offset of table
        // and table length
        dirOS.write(tabEntry.tag);
        long checksum = calcCheckSum(tableBytes);
        dirOS.write(longToBytes(checksum));
        dirOS.write(longToBytes(runningOffset));
        dirOS.write(longToBytes(tableBytes.length));

        // add table length to running offset
        runningOffset += tableBytes.length;

        // pad table length to 4-byte boundary
        int leftover = (int)(tableBytes.length % 4);
        if (leftover > 0) {
            int padding = 4 - leftover;
            runningOffset += padding;
        }

        // add table checksum to full font checksum
        fullCheckSum += checksum;
        if (fullCheckSum >= 0x100000000L)
            fullCheckSum -= 0x100000000L;
    }

    // calculate checksum for directory header and directory entries,
    // add it to the full font checksum -- full font checksum now complete
    fullCheckSum += calcCheckSum(dirOS.toByteArray());
    if (fullCheckSum >= 0x100000000L)
        fullCheckSum -= 0x100000000L;

    // write each table in the directory to the output
    int headPos = 0;
    iter = tabKeyNames.iterator();
    while (iter.hasNext()) {
        String tabKey = (String)iter.next();

        TTFDirTabEntry tabEntry = (TTFDirTabEntry)dirTabs.get(tabKey);

        byte[] tableBytes = null;
        // used modified/reduced tables if necessary
        if (tabKey.equals("glyf")) {
            tableBytes = glyfBytes;
        }
        else if (tabKey.equals("loca")) {
            tableBytes = locaBytes;
        }
        else if (tabKey.equals("EBLC")) {
            if (eblcBytes == null)
                continue;
            tableBytes = eblcBytes;
        }
        else if (tabKey.equals("EBDT")) {
            if (ebdtBytes == null)
                continue;
            tableBytes = ebdtBytes;
            // never include vertical header or vertical metrics
            // PDF doesn't recognize vertical information unless its
            // in its own format
        }
        else if (tabKey.equals("vhea") || tabKey.equals("vmtx") ||
            tabKey.equals("DSIG")) {
            continue;
            // if possible, remove the cmap table
        }
        else if (removeCmap && tabKey.equals("cmap")) {
            continue;
            // otherwise, just use the table as it exists in the input
        }
        else {
            tableBytes = in.getBytes((int)tabEntry.offset,
                                     (int)tabEntry.length);
        }

        // set checkSumAdjustment field in 'head' table to zero for now
        // but not its position so we can change it later
        if (tabKey.equals("head")) {
            headPos = dirOS.size();
            tableBytes[8] = (byte)0;
            tableBytes[9] = (byte)0;
            tableBytes[10] = (byte)0;
            tableBytes[11] = (byte)0;
        }

        // output table
        dirOS.write(tableBytes);

        // pad table output to 4-byte boundary
        int leftover = tableBytes.length % 4;
        if (leftover > 0) {
            int padding = 4 - leftover;
            for (int i = 0; i < padding; i++) {
                dirOS.write((byte)0);
            }
        }
    }

    // calculate the CheckSumAdjustment value from the font checksum
    long csa = 0xB1B0AFBAL - fullCheckSum;
    if (csa < 0)
        csa += 0x100000000L;
    byte[] csaBytes = longToBytes(csa);

    // get the full font output as a byte array
    byte[] dirBytes = dirOS.toByteArray();
    // add in the CheckSumAdjustment value to the appropriate field
    // in the 'head' table
    dirBytes[headPos + 8] = csaBytes[0];
    dirBytes[headPos + 9] = csaBytes[1];
    dirBytes[headPos + 10] = csaBytes[2];
    dirBytes[headPos + 11] = csaBytes[3];

    // write out new font to output stream
    os.write(dirBytes);
}

/**
 * Convert a 4-byte unsigned int value to a corresponding byte array.
 */
protected byte[] longToBytes(long longVal) {
    byte[] longBytes = new byte[4];
    for (int i = 3; i >= 0; i--) {
        longBytes[i] = (byte)(longVal & 0xff);
        longVal = longVal >> 8;
    }
    return longBytes;
}

/**
 * Convert a 2-byte unsigned int value to a corresponding byte array.
 */
protected byte[] shortToBytes(int shortVal) {
    byte[] shortBytes = new byte[2];
    for (int i = 1; i >= 0; i--) {
        shortBytes[i] = (byte)(shortVal & 0xff);
        shortVal = shortVal >> 8;
    }
    return shortBytes;
}

/**
 * Convert a byte array to a 4-byte unsigned int.
 */
protected long bytesToLong(byte[] bytes) throws IOException {
    long ret = bytes[0];
    ret = (ret << 8) + bytes[1];
    ret = (ret << 8) + bytes[2];
    ret = (ret << 8) + bytes[3];

    return ret;
}

/**
 * Convert a byte array to a 2-byte unsigned int.
 */
protected int bytesToShort(byte[] bytes) throws IOException {
    int ret = bytes[0];
    ret = (ret << 8) + bytes[1];

    return ret;
}

/** 
 * Calculate a TrueType table checksum, which is basically the unsigned
 * sum of all the unsigned longs in the table.
 */
public long calcCheckSum(byte[] tableBytes) {
    long sum = 0;
    long nLongs = (tableBytes.length + 3) / 4;
    int longIndex = 0;

    // consider the table as an array of unsigned 32 bit ints
    while (nLongs-- > 0) {
        // convert each 4 byte sequence to an unsigned 32 bit int
        long thisLong = 0;
        for (int i = 0; i < 4; i++) {
            // make sure we're still inside byte array
            if (longIndex + i < tableBytes.length) {
                // shift previous byte up
                if (i != 0) {
                    thisLong = (thisLong << 8);
                }

                byte thisByte = tableBytes[longIndex + i];

                // we need unsigned bytes to do the math correctly,
                // so adjust accordingly if byte is negative
                if (thisByte < 0) {
                    thisLong += (int)(thisByte + 256);
                }
                else {
                    thisLong += thisByte;
                }
                // outside byte array?  Pad with zeroes.
            }
            else {
                thisLong = (thisLong << 8);
            }
        }

        // add 32 bit unsigned int to sum -- make sure sum stays
        // within range of 32 bit int, wrapping around when necessary
        sum += thisLong;
        if (sum >= 0x100000000L) {
            sum -= 0x100000000L;
        }

        // next 4 bytes
        longIndex += 4;
    }

    return sum;
}

/**
 * Determine the name of the font contained within this file.
 * If the fontfile is a TrueType Collection (.ttc file), determine
 * the name of all fonts contained within.  The names returned
 * are derived from the 'fontName' field of the TTF name table
 * and may be slightly different from names other programs use
 * to refer to the font.
 * @return List containing the font names contained within this file.
 */
public List readFontNames(FontFileReader in) throws IOException {

	setStatus(PARSE_OK);

    Vector fontNames = new Vector();
    /*
    * If TrueType collection, see what fonts it contains and
    * return the resulting list.
    */
    if (checkTTCForFontNames(in, fontNames, false, false))
        return fontNames;

    /**
     * Only one font -- return its name
     */
    if (!readDirTabs(in)) {
        addInvalidFileStatusMessage();
        setStatus(PARSE_FATAL);
        return(fontNames);
    }
    readName(in);
    // only add names of unicode fonts
    if (isUnicode(in)) {
        fontNames.add(fontName);
    } else {
        addNoUnicodeStatusMessage();
        setStatus(PARSE_WARNING);
    }
    return (fontNames);
}

/**
 * Determine all possible names of the font contained within this file.
 * If the fontfile is a TrueType Collection (.ttc file), determine
 * all possible names of all fonts contained within.
 * @return List of TTFFontNames objects containing the various
 *         font names that can be used to refer to the font in question.
 *         There is one TTFFontNames object per font in the file.
 */
public List readAllFontNames(FontFileReader in) throws IOException {
	setStatus(PARSE_OK);

    Vector fontNames = new Vector();
    /*
    * If TrueType collection, see what fonts it contains and
    * return the resulting list.
    */
    if (checkTTCForFontNames(in, fontNames, false, true))
        return fontNames;

    /**
     * Only one font -- return its name
     */
    if (!readDirTabs(in)) {
        addInvalidFileStatusMessage();
        setStatus(PARSE_FATAL);
        return(fontNames);
    }
    readName(in);
    // only add names of unicode fonts
    if (isUnicode(in)) {
        TTFFontNames fn = new TTFFontNames();
        fn.fontName = fontName;
        fn.fullName = fullName;
        fn.familyName = familyName;
        fn.subFamilyName = subFamilyName;
        fontNames.add(fn);
    } else {
        addNoUnicodeStatusMessage();
        setStatus(PARSE_WARNING);
    }
    return(fontNames);
}

private void createCMaps() {
    cmaps = new Vector();
    TTFCmapEntry tce = new TTFCmapEntry();

    Enumeration e = unicodeMapping.elements();
    UnicodeMapping um = (UnicodeMapping)e.nextElement();
    UnicodeMapping lastMapping = um;

    tce.unicodeStart = um.uIdx;
    tce.glyphStartIndex = um.gIdx;

    while (e.hasMoreElements()) {
        um = (UnicodeMapping)e.nextElement();
        if (((lastMapping.uIdx + 1) != um.uIdx)
            || ((lastMapping.gIdx + 1) != um.gIdx)) {
            tce.unicodeEnd = lastMapping.uIdx;
            cmaps.addElement(tce);

            tce = new TTFCmapEntry();
            tce.unicodeStart = um.uIdx;
            tce.glyphStartIndex = um.gIdx;
        }
        lastMapping = um;
    }

    tce.unicodeEnd = um.uIdx;
    cmaps.addElement(tce);
}

public void printStuff() {
    System.err.println("Font name: " + fontName);
    System.err.println("Full name: " + fullName);
    System.err.println("Family name: " + familyName);
    System.err.println("Subfamily name: " + subFamilyName);
    System.err.println("Notice:    " + notice);
    System.err.println("xHeight:   " + (int)get_ttf_funit(xHeight));
    System.err.println("capheight: " + (int)get_ttf_funit(capHeight));
    System.err.println("ItalicAngle: " + italicAngle);
    System.err.println();
    System.err.println("Ascender:    " + get_ttf_funit(ascender));
    System.err.println("Descender:   " + get_ttf_funit(descender));
    System.err.println("FontBBox:    [" + (int)get_ttf_funit(fontBBox1)
                       + " " + (int)get_ttf_funit(fontBBox2) + " "
                       + (int)get_ttf_funit(fontBBox3) + " "
                       + (int)get_ttf_funit(fontBBox4) + "]");
}

public static void main(String[] args) {
    try {
        TTFFile ttfFile = new TTFFile();
        FontFileReader reader = new FontFileReader(args[0]);

        String name = null;
        if (args.length >= 2)
            name = args[1];

        ttfFile.readFont(reader, name);
        ttfFile.printStuff();

    } catch (IOException ioe) {
        System.err.println(ioe.toString());
    }
}

public String getWindowsName() {
    return new String(familyName + "," + subFamilyName);
}

public String getPostscriptName() {
    if ("Regular".equals(subFamilyName) || "Roman".equals(subFamilyName))
        return familyName;
    else
        return familyName + "," + subFamilyName;
}

public String getFamilyName() {
    return familyName;
}

public String getCharSetName() {
    return encoding;
}

public int getCapHeight() {
    return (int)get_ttf_funit(capHeight);
}

public int getXHeight() {
    return (int)get_ttf_funit(xHeight);
}

public boolean isSerif() {
    return hasSerifs;
}

public int getFlags() {
    int flags = 32;    // Use Adobe Standard charset
    if (italicAngle != 0.0)
        flags = flags | 64;
    if (isFixedPitch != 0)
        flags = flags | 2;
    if (hasSerifs)
        flags = flags | 1;
    return flags;
}

public String getStemV() {
    return "0";
}

public String getItalicAngle() {
    String ia = Double.toString(italicAngle);

    return ia;
}

public double getItalicAngleValue() {
    return italicAngle;
}

public int[] getFontBBox() {
    int[] fbb = new int[4];
    fbb[0] = (int)get_ttf_funit(fontBBox1);
    fbb[1] = (int)get_ttf_funit(fontBBox2);
    fbb[2] = (int)get_ttf_funit(fontBBox3);
    fbb[3] = (int)get_ttf_funit(fontBBox4);

    return fbb;
}

public int getLowerCaseAscent() {
    return (int)get_ttf_funit(ascender);
}

public int getLowerCaseDescent() {
    return (int)get_ttf_funit(descender);
}

// This is only for WinAnsiEncoding, so the last char is
// the last char < 256
public short getLastChar() {
    return lastChar;
}

public short getFirstChar() {
    return firstChar;
}

public int[] getWidths() {
    int[] wx = new int[mtx_tab.length];
    for (int i = 0; i < wx.length; i++)
        wx[i] = (int)get_ttf_funit(mtx_tab[i].wx);

    return wx;
}

public int getCharWidth(int idx) {
    return (int)get_ttf_funit(ansiWidth[idx]);
}

public Hashtable getKerning() {
    return kerningTab;
}

public Hashtable getAnsiKerning() {
    return ansiKerningTab;
}

public boolean isEmbeddable() {
    return is_embeddable;
}

/**
 * Get glyph metrics given unicode value.  May return null if unicode
 * value has no corresponding glyph in this font.
 */
public TTFMtxEntry getGlyphMetricsByUnicode(int unicodeValue) {
    int glyphIdx = unicodeToGlyphCode(unicodeValue);
    if (glyphIdx == -1) {
        return null;
    }
    return mtx_tab[glyphIdx];
}

/**
 * Get glyph metrics given glyph index.
 */
public TTFMtxEntry getGlyphMetricsByGlyphCode(int glyphCode) {
    if (glyphCode < mtx_tab.length) {
        return mtx_tab[glyphCode];
    }
    else {
        return null;
    }
}

/**
 * Given glyph code, return list of unicode values it represents.
 */
public List glyphCodeToUnicodes(int glyphCode) {
    return mtx_tab[glyphCode].unicodeIndex;
}

/**
 * Get glyph code corresponding to given unicode value.  May return -1
 * if unicode value has no corresponding glyph in this font.
 */
public int unicodeToGlyphCode(int unicodeValue) {
    Integer glyphObj = (Integer)unicodeToGlyphTable.get(
        new Integer(unicodeValue));
    if (glyphObj == null) {
        return -1;
    }
    return glyphObj.intValue();
}

/**
 * Get glyph width given unicode value.  May return 0 if unicode
 * value has no corresponding glyph in this font.
 */
public int getGlyphWidthByUnicode(int unicodeValue) {
    int glyphIdx = unicodeToGlyphCode(unicodeValue);
    if (glyphIdx == -1) {
        return 0;
    }
    return getGlyphWidthByGlyphCode(glyphIdx);
}

/**
 * Get glyph width given glyph index. 
 */
public int getGlyphWidthByGlyphCode(int glyphCode) {
    return get_ttf_funit(mtx_tab[glyphCode].getWx());
}

public byte[] getPanose() {
    return panose;
}

public int getIBMFamilyClass() {
    return ibmFamilyClass;
}

public int getIBMFamilySubclass() {
    return ibmFamilySubclass;
}

/**
 * Read Table Directory from the current position in the
 * FontFileReader and fill the global Hashtable dirTabs
 * with the table name (String) as key and a TTFDirTabEntry
 * as value.
 */
protected boolean readDirTabs(FontFileReader in) throws IOException {
    // first 4 bytes identify file as truetype
    // 0x74727565 for Mac, 0x00010000 for Windows or Mac
    long scaler = in.readTTFULong();
    if (!(scaler == 0x74727565 || scaler == 0x00010000)) {
        return false;
    }

    int ntabs = in.readTTFUShort();
    in.skip(6);    // 3xTTF_USHORT_SIZE

    dirTabs = new Hashtable();
    TTFDirTabEntry[] pd = new TTFDirTabEntry[ntabs];
    // System.err.println("Reading " + ntabs + " dir tables");
    for (int i = 0; i < ntabs; i++) {
        pd[i] = new TTFDirTabEntry();
        String tag = pd[i].read(in);
        dirTabs.put(tag, pd[i]);
    }
	return true;
}

/**
 * Read the "head" table, this reads the bounding box and
 * sets the upem (unitsPerEM) variable
 */
protected void readFontHeader(FontFileReader in) throws IOException {
    seek_tab(in, "head", 0);

    version = in.readTTFFixedDecimal();

    in.skip(4);
    in.skip(4); // skip checksum
    in.skip(4 + 2);

    upem = in.readTTFUShort();

    in.skip(16);

    fontBBox1 = in.readTTFShort();
    fontBBox2 = in.readTTFShort();
    fontBBox3 = in.readTTFShort();
    fontBBox4 = in.readTTFShort();

    in.skip(2 + 2 + 2);

    loca_format = in.readTTFShort();
}

/**
 * Read the number of glyphs from the "maxp" table
 */
protected void getNumGlyphs(FontFileReader in) throws IOException {
    seek_tab(in, "maxp", 4);
    nglyphs = in.readTTFUShort();
}


/**
 * Read the "hhea" table to find the ascender and descender and
 * size of "hmtx" table, i.e. a fixed size font might have only
 * one width
 */
protected void readHorizontalHeader(FontFileReader in)
    throws IOException {
    seek_tab(in, "hhea", 4);
    ascender = in.readTTFShort();    // Use sTypoAscender in "OS/2" table?
    descender = in.readTTFShort();    // Use sTypoDescender in "OS/2" table?

    in.skip(2 + 2 + 3 * 2 + 8 * 2);
    nhmtx = in.readTTFUShort();
    // System.err.println("Number of horizontal metrics: " + nhmtx);
}

/**
 * Read "hmtx" table and put the horizontal metrics
 * in the mtx_tab array. If the number of metrics is less
 * than the number of glyphs (eg fixed size fonts), extend
 * the mtx_tab array and fill in the missing widths
 */
protected void readHorizontalMetrics(FontFileReader in)
    throws IOException {
    seek_tab(in, "hmtx", 0);

    int mtx_size = (nglyphs > nhmtx) ? nglyphs : nhmtx;
    mtx_tab = new TTFMtxEntry[mtx_size];

    // System.err.println("*** Widths array: \n");
    for (int i = 0; i < mtx_size; i++)
        mtx_tab[i] = new TTFMtxEntry();
    for (int i = 0; i < nhmtx; i++) {
        mtx_tab[i].wx = in.readTTFUShort();
        mtx_tab[i].lsb = in.readTTFShort();
        /*
        * System.err.println("   width["+i+"] = "+
        * get_ttf_funit(mtx_tab[i].wx)+";");
        */
    }

    if (nhmtx < mtx_size) {
        // Fill in the missing widths
        int lastWidth = mtx_tab[nhmtx - 1].wx;
        for (int i = nhmtx; i < mtx_size; i++) {
            mtx_tab[i].wx = lastWidth;
            mtx_tab[i].lsb = in.readTTFShort();
        }
    }
}


/**
 * Read the "post" table
 * containing the postscript names of the glyphs.
 */
private final void readPostscript(FontFileReader in) throws IOException {
    String[] ps_glyphs_buf;
    int i, k, l;

    seek_tab(in, "post", 0);
    post_format = in.readTTFLong();
    italicAngle = in.readTTFFixedDecimal();
    underlinePosition = in.readTTFShort();
    underlineThickness = in.readTTFShort();
    isFixedPitch = in.readTTFULong();

    in.skip(4 * 4);

    // System.err.println("Post format: "+post_format);
    switch (post_format) {
        case 0x00010000:
            //System.err.println("Postscript format 1");
            for (i = 0; i < Glyphs.mac_glyph_names.length &&
            			i < mtx_tab.length; i++) {
                mtx_tab[i].name = Glyphs.mac_glyph_names[i];
            }
            break;
        case 0x00020000:
            //System.err.println("Postscript format 2");
            int numGlyphStrings = 0;
            l = in.readTTFUShort();      // Num Glyphs
            // short minIndex=256;
            for (i = 0; i < l; i++) {    // Read indexes
                mtx_tab[i].index = in.readTTFUShort();
                // if (minIndex > mtx_tab[i].index)
                // minIndex=(short)mtx_tab[i].index;

                if (mtx_tab[i].index > 257)
                    numGlyphStrings++;

                // System.err.println("Post index: "+mtx_tab[i].index);
            }
            // firstChar=minIndex;
            ps_glyphs_buf = new String[numGlyphStrings];
            // System.err.println("Reading " + numGlyphStrings +
            // " glyphnames"+
            // ", was n num glyphs="+l);
            for (i = 0; i < ps_glyphs_buf.length; i++) {
                ps_glyphs_buf[i] = in.readTTFString(in.readTTFUByte());
            }

            for (i = 0; i < l; i++) {
                if (mtx_tab[i].index < NMACGLYPHS) {
                    mtx_tab[i].name =
                        Glyphs.mac_glyph_names[mtx_tab[i].index];
                }
                else {
                    k = mtx_tab[i].index - NMACGLYPHS;
                    /*
                    * System.err.println(k+" i="+i+" mtx="+mtx_tab.length+
                    * " ps="+ps_glyphs_buf.length);
                    */
                    mtx_tab[i].name = ps_glyphs_buf[k];
                }
            }

            break;
        case 0x00030000:
            //System.err.println("Postscript format 3");
            //Postscript format 3 contains no glyph names
            break;
        default:
            System.err.println("Unknown Postscript format : " + post_format);
    }
}


/**
 * Read the "OS/2" table
 */
private final void readOS2(FontFileReader in) throws IOException {
    // Check if font is embeddable
    if (dirTabs.get("OS/2") != null) {
        seek_tab(in, "OS/2", 2 * 4);
        int fsType = in.readTTFUShort();
        if (fsType == 2) {
            is_embeddable = false;
        }
        else {
            is_embeddable = true;
        }

        in.readBytes(2 * 10);
        int famClass = in.readTTFUShort();
        ibmFamilyClass = famClass >> 8;
        ibmFamilySubclass = famClass & 0x000f;
        panose = in.readBytes(10);

    }
    else
        is_embeddable = true;
}

/**
 * Read the "loca" table
 */
protected final void readIndexToLocation(FontFileReader in)
    throws IOException {
    seek_tab(in, "loca", 0);
    for (int i = 0; i < nglyphs; i++) {
        mtx_tab[i].offset = (loca_format == 1 ? in.readTTFULong()
            : (in.readTTFUShort() << 1));
    }
    lastLoca = (loca_format == 1 ? in.readTTFULong()
        : (in.readTTFUShort() << 1));
}

/**
 * Read the "glyf" table to find the bounding boxes
 */
private final void readGlyf(FontFileReader in) throws IOException {
    compositeGlyphs = new Hashtable();

    TTFDirTabEntry dirTab = (TTFDirTabEntry)dirTabs.get("glyf");
    for (int i = 0; i < nglyphs; i++) {
        // offset different than next means glyph is in this font
        long offsetDiff = 0;
        if (i < nglyphs - 1)
            offsetDiff = mtx_tab[i + 1].offset - mtx_tab[i].offset;
        else
            offsetDiff = lastLoca - mtx_tab[i].offset;
        if (offsetDiff > 0) {
            in.seek_set(dirTab.offset + mtx_tab[i].offset);
            int contours = in.readTTFShort();
            mtx_tab[i].bbox[0] = in.readTTFShort();
            mtx_tab[i].bbox[1] = in.readTTFShort();
            mtx_tab[i].bbox[2] = in.readTTFShort();
            mtx_tab[i].bbox[3] = in.readTTFShort();

            // keep track of composite glyphs
            if (contours == -1) {
                Vector compositeVector = new Vector();
                Integer compositeIdx = null;
                int flags = 0;
                boolean moreComposites = true;
                while (moreComposites) {
                    flags = in.readTTFUShort();
                    compositeIdx = new Integer(in.readTTFUShort());
                    compositeVector.addElement(compositeIdx);

                    if ((flags & 1) > 0) {
                        // ARG_1_AND_ARG_2_ARE_WORDS
                        in.skip(4);
                    }
                    else {
                        in.skip(2);
                    }

                    if ((flags & 8) > 0)
                        in.skip(2);  // WE_HAVE_A_SCALE
                    else if ((flags & 64) > 0)
                        in.skip(4);  // WE_HAVE_AN_X_AND_Y_SCALE
                    else if ((flags & 128) > 0)
                        in.skip(8);  // WE_HAVE_A_TWO_BY_TWO

                    if ((flags & 32) > 0)
                        moreComposites = true;
                    else
                        moreComposites = false;
                }
                compositeGlyphs.put(new Integer(i), compositeVector);
            }
            // offset same as next means glyph is not in this font!
            // use glyph 0, the missing glyph
        }
        else {
            mtx_tab[i].bbox[0] = mtx_tab[0].bbox[0];
            mtx_tab[i].bbox[1] = mtx_tab[0].bbox[1];
            mtx_tab[i].bbox[2] = mtx_tab[0].bbox[2];
            mtx_tab[i].bbox[3] = mtx_tab[0].bbox[3];
        }
    }
}

/**
 * Read the "name" table
 */
private final void readName(FontFileReader in) throws IOException {
    int platform_id = -1;
    int encoding_id = -1;
    int language_id = -1;

    seek_tab(in, "name", 2);
    int i = in.getCurrentPos();
    int n = in.readTTFUShort();
    int j = in.readTTFUShort() + i - 2;
    i += 2 * 2;
    //System.err.println("     numrecords " + n);

    while (n-- > 0) {
        //System.err.println("Iteration: "+n);
        in.seek_set(i);
        platform_id = in.readTTFUShort();
        encoding_id = in.readTTFUShort();
        language_id = in.readTTFUShort();
        //System.err.println("     platform id " + platform_id);
        //System.err.println("     encoding id " + encoding_id);
        //System.err.println("     language id " + language_id);

        int k = in.readTTFUShort();
        //System.err.println("     name id " + k);
        int l = in.readTTFUShort();

        // platform id 1 is Mac, Mac encoding 0 is Roman character set
        // platform id 3 is Windows, Windows encoding 1 is Unicode font
        // ??Should we handle platform id 0, which is straight Unicode?
        if (((platform_id == 1 && encoding_id == 0) ||
            (platform_id == 3 && encoding_id == 1)) &&
            (k == 1 || k == 2 || k == 0 || k == 4 || k == 6)) {

            in.seek_set(j + in.readTTFUShort());
            String txt;
            // For Microsoft fonts, text appears to be in Unicode
            if (platform_id == 3) {
                txt = in.readTTFString(l, "UTF-16BE");
            }
            else {
                txt = in.readTTFString(l);
            }
            //System.err.println(platform_id+" "+encoding_id+" "+k+" "+txt);
            switch (k) {
                case 0:
                    notice = txt;
                    //System.err.println("     notice " + notice);
                    break;
                case 1:
                    familyName = txt;
                    //System.err.println("     familyName " + familyName);
                    break;
                case 2:
                    subFamilyName = txt;
                    //System.err.println("     subFamilyName " + subFamilyName);
                    break;
                case 4:
                    fullName = txt;
                    //System.err.println("     fullName " + fullName);
                    break;
                case 6:
                    fontName = txt;
                    //System.err.println("     fontName " + fontName);
                    break;
            }
            //System.err.println("");
            if (!notice.equals("") && !fullName.equals("")
                && !fontName.equals("") && !familyName.equals("")
                && !subFamilyName.equals("")) {
                break;
            }
        }
        i += 6 * 2;
    }
    //System.err.println("");
/*
*/
}

/**
 * Read the "PCLT" table to find xHeight and capHeight
 */
private final void readPCLT(FontFileReader in) throws IOException {
    TTFDirTabEntry dirTab = (TTFDirTabEntry)dirTabs.get("PCLT");
    if (dirTab != null) {
        in.seek_set(dirTab.offset + 4 + 4 + 2);
        xHeight = in.readTTFUShort();
        in.skip(2 * 2);
        capHeight = in.readTTFUShort();
        in.skip(2 + 16 + 8 + 6 + 1 + 1);

        int serifStyle = in.readTTFUByte();
        serifStyle = serifStyle >> 6;
        serifStyle = serifStyle & 3;
        if (serifStyle == 1)
            hasSerifs = false;
        else
            hasSerifs = true;

    }
    else {
        // Approximate capHeight from height of "H"
        // It's most unlikly that a font misses the PCLT table
        // This also assumes that psocriptnames exists ("H")
        // Should look it up int the cmap (that wouldn't help
        // for charsets without H anyway...)
        for (int i = 0; i < mtx_tab.length; i++) {
            if ("H".equals(mtx_tab[i].name))
                capHeight = mtx_tab[i].bbox[3] - mtx_tab[i].bbox[1];
        }
    }
}

/**
 * Reduce the "GLYF" and "LOCA" tables by only including necessary
 * glyphs.
 */
protected final void reduceGLYF(FontFileReader in,
                                ByteArrayOutputStream glyfOS, ByteArrayOutputStream locaOS,
                                SortedSet glyphsUsed) throws IOException {

    TTFDirTabEntry glyfEntry = (TTFDirTabEntry)dirTabs.get("glyf");

    int runningGlyfOffset = 0;
    for (int i = 0; i < nglyphs; i++) {
        // loca table is nothing but offsets for all possible glyphs
        // if no glyph in file, offset is same as previous glyph
        if (loca_format == 1) {
            locaOS.write(longToBytes(runningGlyfOffset));
        }
        else {
            locaOS.write(shortToBytes(runningGlyfOffset >> 1));
        }

        // include glyph in new table if A) it is the missing character
        // glyph (glyph 0) B) it is in the table of glyphs used by
        // the unicode font or C) the characters it represents are
        // used by the ansi font

        // if we're including it, copy the glyph from the original table
        // to the new glyf table, making note of the offset in the
        // loca table
        
        if (i == 0 || glyphsUsed.contains(new Integer(i))) {
            long nextOffset = (i + 1 < nglyphs) ?
                mtx_tab[i + 1].offset : lastLoca;
            long length = nextOffset - mtx_tab[i].offset;
            if (length > 0) {
                in.seek_set(glyfEntry.offset + mtx_tab[i].offset);
                glyfOS.write(in.readBytes((int)length));
                runningGlyfOffset += length;
            }
        }
    }

    // write last 'loca' entry
    if (loca_format == 1) {
        locaOS.write(longToBytes(runningGlyfOffset));
    }
    else {
        locaOS.write(shortToBytes(runningGlyfOffset >> 1));
    }
}

/**
 * Reduce the "EBLC" and "EBDT" tables by only including ranges
 * that contain only glyphs that are being used.
 */
protected final void reduceEBLC(FontFileReader in,
                                ByteArrayOutputStream eblcOS, ByteArrayOutputStream ebdtOS,
                                SortedSet glyphsUsed) throws IOException {
    // Note: there could certainly be some improvements on compression,
    // namely reducing the size of glyph ranges to include only used
    // glyphs, but for the sake of ease and speed, currently an
    // entire glyph range is kept as long as it contains at least one
    // glyph from the used list

    // initial offset into ebdt table starts after table header
    long ebdtOffset = 4;

    // size of one bitmap entry in the eblc table
    int bitmapTableSize = 48;

    TTFDirTabEntry eblcTab = (TTFDirTabEntry)dirTabs.get("EBLC");
    TTFDirTabEntry ebdtTab = (TTFDirTabEntry)dirTabs.get("EBDT");

    // write out the header for the new ebdt table, copied from the old one
    in.seek_set(ebdtTab.offset);
    ebdtOS.write(in.readBytes(4));

    // write out the header for the new eblc table, copied from the old one
    in.seek_set(eblcTab.offset);
    eblcOS.write(in.readBytes(4));

    // read through the EBLC bitmap tables, storing pertinent info
    // and checking to see which ones contain glyph ranges that
    // will mean they have to be kept in our new reduced table

    int eblcNumBitmapTables = (int)in.readTTFULong();
    long[] eblcISTAOffsets = new long[eblcNumBitmapTables];
    long[] eblcISTSizes = new long[eblcNumBitmapTables];
    long[] eblcNumISTs = new long[eblcNumBitmapTables];
    Vector reducedBitmapTableIndices = new Vector();
    for (int i = 0; i < eblcNumBitmapTables; i++) {
        eblcISTAOffsets[i] = in.readTTFULong();
        eblcISTSizes[i] = in.readTTFULong();
        eblcNumISTs[i] = in.readTTFULong();
        in.skip(4 + (2 * 12));
        int startGlyph = in.readTTFUShort();
        int endGlyph = in.readTTFUShort();
        SortedSet subset = glyphsUsed.subSet(new Integer(startGlyph),
                                             new Integer(endGlyph + 1));
        if (subset.size() > 0) {
            reducedBitmapTableIndices.add(new Integer(i));
        }
        in.skip(1);
        in.skip(1);
        in.skip(1);
        in.skip(1);
    }

    // Nothing here that interests us -- reset output streams and return
    if (reducedBitmapTableIndices.size() == 0) {
        eblcOS.reset();
        ebdtOS.reset();
        return;
    }

    // write out number of bitmap tables in new reduced eblc table
    eblcOS.write(longToBytes(reducedBitmapTableIndices.size()));

    // offset for first Index SubTableArray in reduced table
    // is right after the bitmap tables
    long reducedISTAOffset = 8 +
        (reducedBitmapTableIndices.size() * 48);

    // sub-output stream for Index SubTable Arrays and
    // corresponding Index SubTables
    ByteArrayOutputStream istaOS = new ByteArrayOutputStream();

    // for each of the bitmap tables we're going to keep, look
    // at its Index SubTable Arrays and SubTables
    for (int r = 0; r < reducedBitmapTableIndices.size(); r++) {
        int i = ((Integer)reducedBitmapTableIndices.get(r)).intValue();

        // position at start of IndexSubTableArray for this Bitmap
        long eblcISTAOffset = eblcTab.offset + eblcISTAOffsets[i];
        in.seek_set(eblcISTAOffset);

        Vector reducedISTIndices = new Vector();

        // traverse IndexSubTableArray, storing values
        // and noting which glyph ranges we'll be keeping for our
        // new reduced table

        int[] firstGlyphs = new int[(int)eblcNumISTs[i]];
        int[] lastGlyphs = new int[(int)eblcNumISTs[i]];
        long[] addOffsets = new long[(int)eblcNumISTs[i]];
        for (int j = 0; j < eblcNumISTs[i]; j++) {
            firstGlyphs[j] = in.readTTFUShort();
            lastGlyphs[j] = in.readTTFUShort();
            addOffsets[j] = in.readTTFULong();

            SortedSet subset = glyphsUsed.subSet(
                new Integer(firstGlyphs[j]), new Integer(lastGlyphs[j]));
            if (subset.size() != 0) {
                reducedISTIndices.add(new Integer(j));
            }
        }

        // offset for first Index SubTable in reduced Table is
        // at end of Index SubTable Array
        long reducedISTOffset = (reducedISTIndices.size() * 8);

        // sub-output streams for Index SubTables
        ByteArrayOutputStream[] istOS =
            new ByteArrayOutputStream[reducedISTIndices.size()];

        for (int s = 0; s < reducedISTIndices.size(); s++) {
            int j = ((Integer)reducedISTIndices.get(s)).intValue();

            istaOS.write(shortToBytes(firstGlyphs[j])); // first gl
            istaOS.write(shortToBytes(lastGlyphs[j]));  // last gl
            istaOS.write(longToBytes(reducedISTOffset));  // offset

            // position input at correct Index SubTable
            in.seek_set(eblcISTAOffset + addOffsets[j]);

            // read from input Index SubTable
            int indexFormat = in.readTTFUShort();
            int imageFormat = in.readTTFUShort();
            long imageOffset = in.readTTFULong();

            // and write to Index SubTable output stream
            istOS[s] = new ByteArrayOutputStream();
            istOS[s].write(shortToBytes(indexFormat));
            istOS[s].write(shortToBytes(imageFormat));
            istOS[s].write(longToBytes(ebdtOffset));

            // handle each Index SubTable format
            // grab any referenced ebdt data and write it to the new
            // reduced ebdt table
            // write the Index SubTable to the new reduced eblc table,
            // changing ebdt offsets as needed

            if (indexFormat == 1) {
                long offset = in.readTTFULong();
                for (int k = 1;
                     k < ((lastGlyphs[j] - firstGlyphs[j] + 1) + 1);
                     k++) {
                    long nextOffset = in.readTTFULong();
                    long imageSize = nextOffset - offset;

                    byte[] imageBytes = in.getBytes(
                        (int)(ebdtTab.offset + imageOffset + offset),
                        (int)imageSize);

                    istOS[s].write(longToBytes((int)ebdtOffset));

                    ebdtOS.write(imageBytes);
                    ebdtOffset += imageBytes.length;

                    offset = nextOffset;
                }
                istOS[s].write(longToBytes((int)ebdtOffset));
            }
            else if (indexFormat == 2) {
                long imageSize = in.readTTFULong();
                istOS[s].write(longToBytes(imageSize));
                istOS[s].write(in.readBytes(8));

                byte[] imageBytes = in.getBytes(
                    (int)(ebdtTab.offset + imageOffset), (int)imageSize);
                ebdtOS.write(imageBytes);
                ebdtOffset += imageBytes.length;
            }
            else if (indexFormat == 3) {
                int offset = in.readTTFUShort();
                int tsize = 0;
                for (int k = 1;
                     k < ((lastGlyphs[j] - firstGlyphs[j] + 1) + 1);
                     k++) {
                    int nextOffset = in.readTTFUShort();
                    int imageSize = nextOffset - offset;
                    tsize += imageSize;

                    byte[] imageBytes = in.getBytes(
                        (int)(ebdtTab.offset + imageOffset + offset),
                        imageSize);

                    istOS[s].write(shortToBytes((int)ebdtOffset));

                    ebdtOS.write(imageBytes);
                    ebdtOffset += imageBytes.length;

                    offset = nextOffset;
                }
                istOS[s].write(shortToBytes((int)ebdtOffset));
                while (in.getCurrentPos() % 4 != 0) {
                    in.skip(1);
                    istOS[s].write((byte)0);
                }
            }
            else if (indexFormat == 4) {
                long numGlyphs = in.readTTFULong();
                istOS[s].write(longToBytes(numGlyphs));

                int glyphCode = in.readTTFUShort();
                int offset = in.readTTFUShort();
                for (int k = 1;
                     k < numGlyphs + 1;
                     k++) {
                    //in.skip(2);
                    //in.skip(2);
                    int nextGlyphCode = in.readTTFUShort();
                    int nextOffset = in.readTTFUShort();

                    int imageSize = nextOffset - offset;

                    byte[] imageBytes = in.getBytes(
                        (int)(ebdtTab.offset + offset),
                        imageSize);

                    istOS[s].write(shortToBytes(glyphCode));
                    istOS[s].write(shortToBytes((int)ebdtOffset));

                    ebdtOS.write(imageBytes);
                    ebdtOffset += imageBytes.length;

                    glyphCode = nextGlyphCode;
                    offset = nextOffset;
                }

                istOS[s].write(shortToBytes(glyphCode));
                istOS[s].write(shortToBytes((int)ebdtOffset));

            }
            else if (indexFormat == 5) {
                long imageSize = in.readTTFULong();
                istOS[s].write(longToBytes(imageSize));
                istOS[s].write(in.readBytes(8));
                long numGlyphs = in.readTTFULong();
                istOS[s].write(longToBytes(numGlyphs));
                for (int k = 0;
                     k < numGlyphs;
                     k++) {
                    int glyphCode = in.readTTFUShort();
                    istOS[s].write(shortToBytes(glyphCode));
                }
                while (in.getCurrentPos() % 4 != 0) {
                    in.skip(1);
                    istOS[s].write((byte)0);
                }

                byte[] imageBytes = in.getBytes(
                    (int)(ebdtTab.offset + imageOffset), (int)imageSize);
                ebdtOS.write(imageBytes);
                ebdtOffset += imageBytes.length;
            }

            // increase offset for IndexSubTable by its size
            reducedISTOffset += istOS[s].size();
        }

        // write Index SubTable streams to Index SubTable Array
        // stream following end of array
        
        for (int s = 0; s < reducedISTIndices.size(); s++) {
            istaOS.write(istOS[s].toByteArray());
        }

        // write out bitmap tables
        long eblcBitmapOffset = eblcTab.offset + 8 + (i * bitmapTableSize);
        in.seek_set(eblcBitmapOffset);

        eblcOS.write(longToBytes(reducedISTAOffset)); // ISTAOffset
        eblcOS.write(longToBytes(istaOS.size())); // ISTSize
        eblcOS.write(longToBytes(reducedISTIndices.size())); // numISTs
        eblcOS.write(longToBytes(0)); // color ref -- always zero

        in.skip(16);

        eblcOS.write(in.readBytes(12)); // horizontal line metrics
        eblcOS.write(in.readBytes(12)); // vertical line metrics

        eblcOS.write(in.readBytes(2)); // start glyph index
        eblcOS.write(in.readBytes(2)); // end glyph index

        eblcOS.write(in.readBytes(4)); // ppemx, ppemy, bitdepth, flags

        reducedISTAOffset = 8 +
            (reducedBitmapTableIndices.size() * 48) +
            istaOS.size();
    }

    // write out Index SubTable Arrays and Index SubTables
    eblcOS.write(istaOS.toByteArray());
}

/* Method used for testing my reduced EBLC tables.
private final void checkEBLC(FontFileReader in, FontFileReader inEbdt)
throws IOException 
{
int bitmapTableSize = 48;

in.seek_set(0);
inEbdt.seek_set(0);
in.skip(4);
inEbdt.skip(4);

int eblcNumBitmapTables = (int) in.readTTFULong();
System.err.println("EBLC: num bitmap tables " + eblcNumBitmapTables);
long[] eblcISTAOffsets = new long[eblcNumBitmapTables];
long[] eblcISTSizes = new long[eblcNumBitmapTables];
long[] eblcNumISTs = new long[eblcNumBitmapTables];
for (int i = 0; i < eblcNumBitmapTables; i++) {
eblcISTAOffsets[i] = in.readTTFULong();
eblcISTSizes[i] = in.readTTFULong();
eblcNumISTs[i] = in.readTTFULong();
in.skip(4 + (2 * 12));
int startGlyph = in.readTTFUShort();
int endGlyph = in.readTTFUShort();
System.err.println("EBLC: bitmap table " + i + 
" start gl " + startGlyph + " end gl " + endGlyph);
in.skip(1);
in.skip(1);
in.skip(1);
in.skip(1);
}

// for each of the bitmap tables we're going to keep, look
// at its Index SubTable Arrays and SubTables
for (int i = 0; i < eblcNumBitmapTables; i++) {
System.err.println("EBLC: for bitmap " + i);

// position at start of IndexSubTableArray for this Bitmap
long eblcISTAOffset = eblcISTAOffsets[i];
//in.seek_set(eblcISTAOffset);

System.err.println("EBLC: outer " + i +
" current pos " + in.getCurrentPos() + " offset " +
(eblcISTAOffsets[i]));

// traverse IndexSubTableArray, storing values
// and noting which glyph ranges we'll be keeping for our
// new reduced table

int[] firstGlyphs = new int[(int)eblcNumISTs[i]];
int[] lastGlyphs = new int[(int)eblcNumISTs[i]];
long[] addOffsets = new long[(int)eblcNumISTs[i]];
for (int j = 0; j < eblcNumISTs[i]; j++) {
firstGlyphs[j] = in.readTTFUShort();
lastGlyphs[j] = in.readTTFUShort();
addOffsets[j] = in.readTTFULong();

System.err.println("EBLC:    ISTA " + j + " fg " + 
firstGlyphs[j] +
" lg " + lastGlyphs[j] + " add offset " +
addOffsets[j] + " full offset " +
(eblcISTAOffsets[i] + addOffsets[j]));
}

for (int j = 0; j < eblcNumISTs[i]; j++) {

// position input at correct Index SubTable
//in.seek_set(eblcISTAOffset + addOffsets[j]);
System.err.println("EBLC:    IST " + j + 
" currentPos " + in.getCurrentPos());

// read from input Index SubTable
int indexFormat = in.readTTFUShort();
int imageFormat = in.readTTFUShort();
long imageOffset = in.readTTFULong();

System.err.println("EBLC:    IST " + j + 
" index format " + 
indexFormat + " imageFormat " + imageFormat);

System.err.println("EBLC: imageOffset " + imageOffset);

if (indexFormat == 1) {
for (int k = 0; 
k < ((lastGlyphs[j] - firstGlyphs[j] + 1) + 1);
k++) 
{
in.skip(4);
}
} else if (indexFormat == 2) {
in.skip(4);
in.skip(8);
} else if (indexFormat == 3) {
for (int k = 0; 
k < ((lastGlyphs[j] - firstGlyphs[j] + 1) + 1);
k++) 
{
in.readBytes(2);
}
while (in.getCurrentPos() % 4 != 0) {
in.skip(1);
}
} else if (indexFormat == 4) {
long numGlyphs = in.readTTFULong();
for (int k = 0; 
k < numGlyphs + 1;
k++) 
{
in.skip(2);
in.skip(2);
}
} else if (indexFormat == 5) {
in.skip(4);
in.skip(8);
long numGlyphs = in.readTTFULong();
for (int k = 0; 
k < numGlyphs;
k++) 
{
int glyphCode = in.readTTFUShort();
//System.err.println("EBLC     gc " + glyphCode);
}
while (in.getCurrentPos() % 4 != 0) {
in.skip(1);
}
}

}

}
}
*/

/**
 * Read the kerning table, create a table for both CIDs and
 * winAnsiEncoding
 */
private final void readKerning(FontFileReader in) throws IOException {
    // Read kerning
    kerningTab = new Hashtable();
    ansiKerningTab = new Hashtable();
    TTFDirTabEntry dirTab = (TTFDirTabEntry)dirTabs.get("kern");
    if (dirTab != null) {
        seek_tab(in, "kern", 2);
        for (int n = in.readTTFUShort(); n > 0; n--) {
            in.skip(2 * 2);
            int k = in.readTTFUShort();
            if (!((k & 1) != 0) || (k & 2) != 0 || (k & 4) != 0) {
                System.err.println("Warning: TTF: only Kern format 1 is supported.  This is format " + (k & 3));
                return;
            }
            if ((k >> 8) != 0) {
                System.err.println("Warning: TTF: Ignoring this kerning table.");
                continue;
            }

            k = in.readTTFUShort();
            in.skip(3 * 2);
            while (k-- > 0) {
                int i = in.readTTFUShort();
                int j = in.readTTFUShort();
                int kpx = in.readTTFShort();
                if (kpx != 0) {
                    // CID table
                    Integer iObj = new Integer(i);
                    Hashtable adjTab = (Hashtable)kerningTab.get(iObj);
                    if (adjTab == null)
                        adjTab = new java.util.Hashtable();
                    adjTab.put(new Integer(j),
                               new Integer((int)get_ttf_funit(kpx)));
                    kerningTab.put(iObj, adjTab);
                }
            }
        }
        // System.err.println(kerningTab.toString());

        // Create winAnsiEncoded kerning table

        for (Enumeration ae = kerningTab.keys(); ae.hasMoreElements();) {
            Integer cidKey = (Integer)ae.nextElement();
            Hashtable akpx = new Hashtable();
            Hashtable ckpx = (Hashtable)kerningTab.get(cidKey);

            for (Enumeration aee = ckpx.keys(); aee.hasMoreElements();) {
                Integer cidKey2 = (Integer)aee.nextElement();
                Integer kern = (Integer)ckpx.get(cidKey2);

                for (Enumeration uniMap = mtx_tab[cidKey2.intValue()].unicodeIndex.elements();
                     uniMap.hasMoreElements();) {
                    Integer unicodeKey = (Integer)uniMap.nextElement();
                    Integer[] ansiKeys =
                        unicodeToWinAnsi(unicodeKey.intValue());
                    for (int u = 0; u < ansiKeys.length; u++) {
                        akpx.put(ansiKeys[u], kern);
                    }
                }
            }

            if (akpx.size() > 0)
                for (Enumeration uniMap = mtx_tab[cidKey.intValue()].unicodeIndex.elements();
                     uniMap.hasMoreElements();) {
                    Integer unicodeKey = (Integer)uniMap.nextElement();
                    Integer[] ansiKeys =
                        unicodeToWinAnsi(unicodeKey.intValue());
                    for (int u = 0; u < ansiKeys.length; u++) {
                        ansiKerningTab.put(ansiKeys[u], akpx);
                    }
                }
        }
    }
}

/**
 * Return a vector with TTFCmapEntry
 */
public Vector getCMaps() {
    return cmaps;
}

/**
 * Check if this is a TrueType collection and that the given
 * name exists in the collection.
 * If it does, set offset in fontfile to the beginning of
 * the Table Directory for that font
 * @ return true if not collection or font name present, false
 * otherwise
 */
protected final boolean checkTTC(FontFileReader in, String name,
                                 boolean verbose) throws IOException {
    String tag = in.readTTFString(4);

    if ("ttcf".equals(tag)) {
        // This is a TrueType Collection
        in.skip(4);

        // Read directory offsets
        int numDirectories = (int)in.readTTFULong();
        // int numDirectories=in.readTTFUShort();
        long[] dirOffsets = new long[numDirectories];
        for (int i = 0; i < numDirectories; i++) {
            dirOffsets[i] = in.readTTFULong();
        }

        if (verbose) {
            System.err.println("This is a TrueType collection file with"
                               + numDirectories + " fonts");
            System.err.println("Containing the following fonts: ");
        }
        // Read all the directories and name tables to check
        // If the font exists - this is a bit ugly, but...
        boolean found = false;

        // Iterate through all name tables even if font
        // Is found, just to show all the names
        dirTabOffset = 0;
        for (int i = 0; (i < numDirectories); i++) {
            in.seek_set(dirOffsets[i]);
            if (!readDirTabs(in)) {
                addInvalidFileStatusMessage();
                setStatus(PARSE_WARNING);
            }

            readName(in);

            if (fontName.equals(name)) {
                if (verbose)
                    System.err.println("* " + fullName);
                if (isUnicode(in)) {
                    found = true;
                    dirTabOffset = dirOffsets[i];
                } else {
                    addNoUnicodeStatusMessage();
                    setStatus(PARSE_WARNING);
                }
            }
            else {
                if (verbose)
                    System.err.println(fullName);
            }

            // Reset names
            notice = "";
            fullName = "";
            familyName = "";
            fontName = "";
            subFamilyName = "";
        }

        in.seek_set(dirTabOffset);
        return found;
    }
    else {
        in.seek_set(0);
        return true;
    }
}

/**
 * Parse a TrueType collection for the names of the fonts it contains.
 * Add these names to the passed List object.
 * @param all if true, fontNames List should be filled with
 *            TTFFontNames objects containing all font names for each font.
 *            If false, fontNames List should be filled with
 *            strings containing name from TTF font name field.
 * @return true if file is a TTC file, false otherwise
 */
protected final boolean checkTTCForFontNames(FontFileReader in,
                                             List fontNames, boolean verbose, boolean all) throws IOException {
    String tag = in.readTTFString(4);

    if ("ttcf".equals(tag)) {
        // This is a TrueType Collection
        in.skip(4);

        // Read directory offsets
        int numDirectories = (int)in.readTTFULong();
        // int numDirectories=in.readTTFUShort();
        long[] dirOffsets = new long[numDirectories];
        for (int i = 0; i < numDirectories; i++) {
            dirOffsets[i] = in.readTTFULong();
        }

        if (verbose) {
            System.err.println("This is a TrueType collection file with"
                               + numDirectories + " fonts");
            System.err.println("Containing the following fonts: ");
        }
        // Read all the directories and name tables to check
        // If the font exists - this is a bit ugly, but...
        boolean found = false;

        // Iterate through all name tables even if font
        // Is found, just to show all the names
        for (int i = 0; (i < numDirectories); i++) {
            in.seek_set(dirOffsets[i]);
            if (!readDirTabs(in)) {
                addInvalidFileStatusMessage();
                setStatus(PARSE_WARNING);
            }

            readName(in);

            if (verbose)
                System.err.println(fullName);

			// only add names of Unicode fonts
			if (isUnicode(in)) {
            	if (all) {
                	TTFFontNames fn = new TTFFontNames();
                	fn.fontName = fontName;
                	fn.fullName = fullName;
                	fn.familyName = familyName;
                	fn.subFamilyName = subFamilyName;
                	fontNames.add(fn);
            	}
            	else {
                	fontNames.add(fontName);
            	}
			} else {
				addNoUnicodeStatusMessage();
				setStatus(PARSE_WARNING);
			}

            // Reset names
            notice = "";
            fullName = "";
            familyName = "";
            fontName = "";
            subFamilyName = "";
        }

        in.seek_set(0);
        return true;
    }
    else {
        in.seek_set(0);
        return false;
    }

}

/**
 * Get the name of the font represented by this instance of the TTFFile
 * class.  This should only be called after readFont() has been called
 * and successfully executed.
 */
public String getFontName() {
    return fontName;
}

/**
 * Get the full name of the font represented by this instance of the TTFFile
 * class.  This should only be called after readFont() has been called
 * and successfully executed.
 */
public String getFullName() {
    return fullName;
}

/**
 * Get the notice section of the font represented by this instance of the
 * TTFFile class.  This should only be called after readFont() has been
 * called and successfully executed.
 */
public String getNotice() {
    return notice;
}

/**
 * Get the fixed pitch value of the font represented by this instance of the
 * TTFFile class.  This should only be called after readFont() has been
 * called and successfully executed.
 */
public boolean isFixedPitch() {
    if (isFixedPitch != 0)
        return true;
    else
        return false;
}

/**
 * Get the version of the font as font in the font "head" table.
 */
public double getVersion() {
    return version;
}

/**
 * Get the underling position of the font represented by this instance of
 * the TTFFile class.  This should only be called after readFont() has been
 * called and successfully executed.
 */
public int getUnderlinePosition() {
    return (int)get_ttf_funit(underlinePosition);
}

/**
 * Get the underline thickness of the font represented by this instance of 
 * the TTFFile class.  This should only be called after readFont() has been
 * called and successfully executed.
 */
public int getUnderlineThickness() {
    return (int)get_ttf_funit(underlineThickness);
}

/*
* Helper classes, they are not very efficient, but that really
* doesn't matter...
*/
private Integer[] unicodeToWinAnsi(int unicode) {
    Vector ret = new Vector();
    for (int i = 32; i < Glyphs.winAnsiEncoding.length; i++)
        if (unicode == Glyphs.winAnsiEncoding[i])
            ret.addElement(new Integer(i));
    Integer[] itg = new Integer[ret.size()];
    ret.copyInto(itg);
    return itg;
}
}


/**
 * Key-value helper class
 */
class UnicodeMapping {
    int uIdx;
    int gIdx;

    UnicodeMapping(int gIdx, int uIdx) {
        this.uIdx = uIdx;
        this.gIdx = gIdx;
    }

}

