/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTreeExplorer.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:28 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing;

import com.klg.jclass.cell.EditorRendererRegistry;
import com.klg.jclass.cell.JCCellEditor;
import com.klg.jclass.cell.JCCellRenderer;
import com.klg.jclass.cell.swing.JCCellEditorToJTree;
import com.klg.jclass.cell.swing.JCCellRendererToJTree;

import com.klg.jclass.util.JCEnvironment;
import com.klg.jclass.util.swing.JCSortableTable;
import com.klg.jclass.util.treetable.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.TableUI;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableModel;
import javax.swing.table.TableColumn;
import javax.swing.tree.*;
import java.awt.*;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.EventObject;

/**
 * A component that gives a visual representation of the hierarchical
 * nature of its data, presenting a tree view in its left-hand pane and
 * a table view in its right-hand pane.
 * <B>See</B>
 * <A HREF="http://www.javasoft.com/products/jdk/1.2/docs/api/index.html">
 * The Java 2 API</A>
 * to compare this component to <code>JTree</code> and <code>JTable</code>.
 */
public class JCTreeExplorer extends JSplitPane {

protected JCTreeTableModel    treetableModel;
protected TreeModel           treeModel;
protected NodeChildrenTable   tableModel;
protected JTree               tree;
protected JCSortableTable     table;
protected TableCellRenderer   firstColumnRenderer;
protected TableCellEditor     firstColumnEditor;
protected JCTreeIconRenderer  iconRenderer = new DefaultTreeIconRenderer();

//
// PLAF controlled variables
//
protected Icon   openIcon;        // "Tree.openIcon"
protected Icon   closedIcon;      // "Tree.closedIcon"
protected Icon   leafIcon;        // "Tree.leafIcon"
protected int    iconGap;         // Gap between the icon and text fixed at 2
protected int    iconWidth;       // Max width of above icons
protected int    cellHeight;      // Max height of above icons
protected Color  tableBackground; // "Table.background"

public JCTreeExplorer()
{
    this(new EmptyTreeTableModel(EmptyTreeTableModel.makeDefaultTree()));
}

public JCTreeExplorer(JCTreeTableModel treetable_model,
                      Dimension intercellSpacing)
{
    setModel(treetable_model);
    table.setIntercellSpacing(intercellSpacing);
}

public JCTreeExplorer(JCTreeTableModel treetable_model)
{
    this(treetable_model, new Dimension(0, 0));
}

public JCTreeTableModel getModel()
{
    return(treetableModel);
}

public void setModel(JCTreeTableModel treetable_model)
{
    if (!(treetable_model instanceof TreeModel) &&
        !(treetable_model instanceof TableModel))
    {
        throw new IllegalArgumentException("TreeTableModel MUST implement " +
                                           "both javax.swing.tree.TreeModel " +
                                           "and\n\tjavax.swing.table.TableModel");
    }

    this.inferLookAndFeelSetVars();

    this.treetableModel = treetable_model;

    SelectionListener selectionListen = new SelectionListener();

    // Create TreeModel
    treeModel = new BranchTree(treetableModel);

    // Create empty TableModel
    tableModel = new NodeChildrenTable(treetableModel);
    tableModel.addTableModelListener(selectionListen);

    // Create ScrollPane for tree component
    JScrollPane left_scrollpane = new JScrollPane();

    // Create Tree Component
    tree = new TreeIconRenderTree(treeModel);
    tree.addTreeSelectionListener(selectionListen);

    // Place the tree on the left hand side of the split pane
    left_scrollpane.setViewportView(tree);
    setLeftComponent(left_scrollpane);

    // Create ScrollPane for table component
    JScrollPane right_scrollpane = new JScrollPane();

    // Create Table Component
    table = new TreeIconRenderTable(tableModel);
    firstColumnRenderer = new FirstColumnRenderer();
    firstColumnEditor = new FirstColumnEditor();
    table.addMouseListener(new ClickListener());

    // Place the table on the Right hand side of the split pane
    right_scrollpane.setViewportView(table);
    setRightComponent(right_scrollpane);

    // Make sure the selection mode is set to single
    TreeSelectionModel tsm = tree.getSelectionModel();
    if (tsm != null) {
        tsm.setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
    }

    // PR11004: folder names don't paint properly in TreeExplorer when splitpane
    // is resized don't reset the cell renderers and editors
    // This is what happens:
    //    We use the defaults from the central registry
    //    These may be "LightCellRenderers" which require a graphics context
    //    that we can't supply, when they calculate the preferred size
    //    Thus, the size is always wrong, or we need to use deprecated methods
    //    So, just use the default JTree renderers and editors
    //
    //tree.setCellRenderer(new TreeRenderer());
    //tree.setCellEditor(new TreeEditor());

    if (JCEnvironment.isDesignTime()) {
        // make the split pane left side a little larger
        setDividerLocation(50);

        // expand the root
        tree.expandRow(0);
        tree.addSelectionRow(0);
    }

    this.inferLookAndFeelSetupControls();
}

public Dimension getPreferredSize()
{
    // Use a nice, compact size for design-time
    if (JCEnvironment.isDesignTime()) {
        return(new Dimension(300, 200));
    }
    else {
        return(super.getPreferredSize());
    }
}

/**
 * Returns the <code>TreePath</code> of the first selected row inside the table
 * view or, if no rows are selected, the selected node within the tree view.
 * @return <code>TreePath</code> of the first selected row inside the table
 * view or, if no rows are selected, the selected node within the tree view
 */
public TreePath getSelectionPath()
{
    TreePath paths[] = getSelectionPaths();
    if (paths == null || paths.length == 0) {
        return(tree.getSelectionPath());
    }
    else {
        return(paths[0]);
    }
}

/**
 * Returns the <code>TreePath</code>'s of the selected rows inside the table view.
 * @return the <code>TreePath</code>'s of the selected rows
 */
public TreePath[] getSelectionPaths()
{
    TreePath parent_path = tree.getSelectionPath();
    int rows[] = table.getSelectedRows();
    //PR33626 parent_path might also be null
    if (parent_path == null || rows == null || rows.length == 0) {
        return(null);
    }
    TreePath paths[] = new TreePath[rows.length];
    for (int i = 0; i < rows.length; i++) {
        Object child =
            treetableModel.getChild(parent_path.getLastPathComponent(),
                                    rows[i]);
        paths[i] = parent_path.pathByAddingChild(child);
    }
    return(paths);
}

/**
 * Provided for Beans property editors - has no effect.
 * @param s
 */
public void setAbout(String s) {}

/**
 * Returns vendor information.
 * @return vendor information
 */
public String getAbout()
{
    return(JCVersion.getVersionString());
}

/**
 * Returns the <code>JTree</code> component used.
 * @return <code>JTree</code> component used
 */
public JTree getTree()
{
    return(tree);
}

/**
 * Returns the <code>JTable</code> component used. <br><b>Note:</b> This is an
 * instance of <code>JCSortableTable</code>, and you can access the sortable
 * table functionality with an appropriate cast.
 * @see com.klg.jclass.util.swing.JCSortableTable
 * @return <code>JTable</code> component used
 */
public JTable getTable()
{
    return(table);
}




public void inferLookAndFeel()
{
    this.inferLookAndFeelSetVars();
    this.inferLookAndFeelSetupControls();
}

protected void inferLookAndFeelSetVars()
{
    openIcon = (Icon)UIManager.get("Tree.openIcon");
    closedIcon = (Icon)UIManager.get("Tree.closedIcon");
    leafIcon = (Icon)UIManager.get("Tree.leafIcon");
    iconGap = 2;
    iconWidth = Math.max(openIcon.getIconWidth(), closedIcon.getIconWidth());
    iconWidth = Math.max(iconWidth, leafIcon.getIconWidth());
    tableBackground = (Color)UIManager.get("Table.background");
    cellHeight = Math.max(openIcon.getIconHeight(), closedIcon.getIconHeight());
    cellHeight = Math.max(cellHeight, leafIcon.getIconHeight());
}

protected void inferLookAndFeelSetupControls()
{
    if (table.getRowHeight() < cellHeight) {
        table.setRowHeight(cellHeight);
    }
}

public void setUI(TableUI ui)
{
    inferLookAndFeel();
    super.setUI(ui);
}

/**
 * Class that wraps a component and makes sure that it draws everything.
 */
protected class WhiteOutContainer extends JComponent
{
    boolean   isLeaf = true;
    TreePath  path;
    Object    value;
    Class     valueClass;

    public void doLayout()
    {
        if (this.getComponentCount() > 0) {
            Component c = getComponent(0);
            if (c == null) {
                
                return;
            }
            Dimension s = getSize();
            c.setBounds(iconWidth + iconGap, 0, s.width - iconWidth - iconGap,
                        s.height);
        }
        else {
            super.doLayout();
        }
    }

    public void paint(Graphics gc)
    {
        Dimension size = getSize();
        // do a clear rect since returning isOpaque() == false
        // only does half the job
        gc.setColor(tableBackground);
        gc.fillRect(0, 0, size.width, size.height);
        super.paint(gc);
        int height = getSize().height;

        //
        // Draw the icon if the icon renderer is not null
        //
        if (iconRenderer == null) {
            return;
        }

        Icon plaf_icon = null;
        if (isLeaf) {
            plaf_icon = leafIcon;
        }
        else {
            plaf_icon = closedIcon;
        }

        Icon icon = iconRenderer.getNodeIcon(treeModel, path,
                                             value, valueClass,
                                             isLeaf, false,
                                             plaf_icon);

        if (icon != null) {
            // Center the icon on the row height
            icon.paintIcon(this, gc, 0,
                           (height - leafIcon.getIconHeight()) / 2);
        }
    }

    public Dimension getPreferredSize()
    {
        Component c = getComponent(0);
        if (c == null) {
            
            return(new Dimension(10, 10));
        }
        Dimension size = c.getPreferredSize();
        size.setSize(size.width + iconGap + iconWidth, size.height);
        return(size);
    }

    public Dimension getMaximumSize()
    {
        Component c = getComponent(0);
        if (c == null) {
            
            return(new Dimension(10, 10));
        }
        return(c.getMaximumSize());
    }

    public Dimension getMinimumSize()
    {
        Component c = getComponent(0);
        if (c == null) {
            
            return(new Dimension(10, 10));
        }
        return(c.getMinimumSize());
    }
}

protected class TreeRenderer extends WhiteOutContainer
                             implements TreeCellRenderer
{
    TreeCellRenderer renderer = null;

    public Component getTreeCellRendererComponent(JTree tree,
                                                  Object value,
                                                  boolean selected,
                                                  boolean expanded,
                                                  boolean leaf,
                                                  int row,
                                                  boolean hasFocus)
    {
        this.value = value;
        if (value != null) {
            valueClass = value.getClass();
        }
        else {
            valueClass = table.getColumnModel().getColumn(0).getClass();
        }

        if (renderer == null) {
            EditorRendererRegistry registry =
                EditorRendererRegistry.getCentralRegistry();
            JCCellRenderer jcRenderer =
                registry.getCellRendererInstance(valueClass, null);
            if (jcRenderer != null) {
                renderer = new JCCellRendererToJTree(jcRenderer);
            }
        }

        Component c = null;
        if (getComponentCount() != 0) {
            c = getComponent(0);
        }

        isLeaf = treeModel.isLeaf(value);
        Component rendererC =
            renderer.getTreeCellRendererComponent(tree, value, selected,
                                                  expanded, leaf, row,
                                                  hasFocus);
        if (c != rendererC) {
            add(rendererC);
        }
        return(this);
    }
}

protected class TreeEditor extends WhiteOutContainer implements TreeCellEditor {
    TreeCellEditor editor = null;

    public Component getTreeCellEditorComponent(JTree tree,
                                                Object value,
                                                boolean isSelected,
                                                boolean expanded,
                                                boolean leaf,
                                                int row)
    {
        this.value = value;
        if (value != null) {
            valueClass = value.getClass();
        }
        else {
            valueClass = table.getColumnModel().getColumn(0).getClass();
        }

        if (editor == null) {
            EditorRendererRegistry registry =
                EditorRendererRegistry.getCentralRegistry();
            JCCellEditor jcEditor = registry.getCellEditorInstance(valueClass,
                                                                   null);
            if (jcEditor != null) {
                editor = new JCCellEditorToJTree(jcEditor);
            }
        }

        Component c = null;
        if (getComponentCount() != 0) {
            c = getComponent(0);
        }

        isLeaf = treeModel.isLeaf(value);
        Component editorC = editor.getTreeCellEditorComponent(tree, value,
                                                              isSelected,
                                                              expanded, leaf,
                                                              row);
        if (c != editorC) {
            add(editorC);
        }
        return(this);

    }

    public TreeCellEditor getEditor()
    {
        if (editor == null) {
            EditorRendererRegistry registry =
                EditorRendererRegistry.getCentralRegistry();
            TableColumn column = table.getColumnModel().getColumn(0);
            JCCellEditor jcEditor =
                registry.getCellEditorInstance(column.getClass(), null);
            if (jcEditor != null) {
                return(new JCCellEditorToJTree(jcEditor));
            }
        }
        return(editor);
    }

    public Object getCellEditorValue()
    {
        return(getEditor().getCellEditorValue());
    }

    public boolean isCellEditable(EventObject anEvent)
    {
        return(getEditor().isCellEditable(anEvent));
    }

    public boolean shouldSelectCell(EventObject anEvent)
    {
        return(getEditor().shouldSelectCell(anEvent));
    }

    public boolean stopCellEditing()
    {
        return(getEditor().stopCellEditing());
    }

    public void cancelCellEditing()
    {
        getEditor().cancelCellEditing();
    }

    public void addCellEditorListener(CellEditorListener l)
    {
        getEditor().addCellEditorListener(l);
    }

    public void removeCellEditorListener(CellEditorListener l)
    {
        getEditor().removeCellEditorListener(l);
    }
}

/**
 * Class that wraps the tree renderer to use it to render the first column of
 * the table.
 */
protected class FirstColumnRenderer extends WhiteOutContainer
                                    implements TableCellRenderer
{
    /**
     * Map TreeCellRenderer to TableCellRenderer.
     * @return
     * @param column
     * @param hasFocus
     * @param isSelected
     * @param row
     * @param table
     * @param value
     */
    public Component getTableCellRendererComponent(JTable table,
                                                   Object value,
                                                   boolean isSelected,
                                                   boolean hasFocus,
                                                   int row, int column)
    {
        TableCellRenderer tcr =
            ((TreeIconRenderTable)table).getRealCellRenderer(row, column);
        // transform row to unsorted row mapping
        row = JCTreeExplorer.this.table.getUnsortedRow(row);

        // build path of row
        TreePath tPath = tableModel.getParentPath();
        Object clicked_node =
            treetableModel.getChild(tPath.getLastPathComponent(), row);
        path = tPath.pathByAddingChild(clicked_node);
        this.value = value;
        if (value != null) {
            valueClass = value.getClass();
        }
        else {
            valueClass = table.getColumnModel().getColumn(column).getClass();
        }

        // Notes on Mapping
        // value, row, focus, and selection parameters are passed through
        //     as is.
        // The tree is used as the invoking parameter but we assume that
        //     the Renderer does not actually need any information
        //     contained therein
        // We always specify any branch nodes as not being expanded in
        //     the table view.
        isLeaf = tableModel.isLeaf(row);
        removeAll();
        add(tcr.getTableCellRendererComponent(table, value, isSelected,
                                              hasFocus, row, column));
        return(this);
    }
}

/**
 * Class that wraps the tree editor to use it to edit the first column of the
 * table.
 */
protected class FirstColumnEditor extends WhiteOutContainer
                                  implements TableCellEditor
{
    protected TableCellEditor editor = null;

    public Component getTableCellEditorComponent(JTable table,
                                                 Object value,
                                                 boolean isSelected,
                                                 int row, int column)
    {
        // also transform row to unsorted row mapping
        row = JCTreeExplorer.this.table.getUnsortedRow(row);
        TableCellEditor ed =
            ((TreeIconRenderTable)table).getRealCellEditor(row, column);
        // See FirstColumnRenderer code for notes on the following mapping
        // build path of row
        TreePath path = tableModel.getParentPath();
        Object clicked_node =
            treetableModel.getChild(path.getLastPathComponent(), row);
        path = path.pathByAddingChild(clicked_node);
        this.value = value;
        if (value != null) {
            valueClass = value.getClass();
        }
        else {
            valueClass = table.getColumnModel().getColumn(column).getClass();
        }

        isLeaf = tableModel.isLeaf(row);
        removeAll();
        add(ed.getTableCellEditorComponent(table, value, isSelected, row,
                                           column));
        return(this);
    }

    public TableCellEditor getEditor()
    {
        if (editor == null) {
            TreeIconRenderTable treeIconRenderTable =
                (TreeIconRenderTable)JCTreeExplorer.this.table;
            editor = treeIconRenderTable.getRealCellEditor(0, 0);
        }
        return(editor);
    }

    public Object getCellEditorValue()
    {
        return(getEditor().getCellEditorValue());
    }

    public boolean isCellEditable(EventObject anEvent)
    {
        return(getEditor().isCellEditable(anEvent));
    }

    public boolean shouldSelectCell(EventObject anEvent)
    {
        return(getEditor().shouldSelectCell(anEvent));
    }

    public boolean stopCellEditing()
    {
        boolean ret = getEditor().stopCellEditing();
        return(ret);
    }

    public void cancelCellEditing()
    {
        getEditor().cancelCellEditing();
    }

    public void addCellEditorListener(CellEditorListener l)
    {
        getEditor().addCellEditorListener(l);
    }

    public void removeCellEditorListener(CellEditorListener l)
    {
        getEditor().removeCellEditorListener(l);
    }
}

/**
 * Sets the icon renderer.  If null, no icon will be drawn.
 * @param renderer
 */
public void setTreeIconRenderer(JCTreeIconRenderer renderer)
{
    this.iconRenderer = renderer;
}

/**
 * Returns the icon renderer being used. If null, no icons are being
 * drawn.
 * @return icon renderer being used
 */
public JCTreeIconRenderer getTreeIconRenderer()
{
    return(iconRenderer);
}

/**
 * Class that listens for selections on the tree component and changes the source
 * of the <code>TableModel</code>.
 */
protected class SelectionListener implements TreeSelectionListener,
                                             TableModelListener
{
    public void valueChanged(TreeSelectionEvent event)
    {
        TreePath path = event.getPath();
        if (path != null && path.equals(tableModel.getParentPath())) {
            // path hasn't changed so don't do anything
            return;
        }
        table.clearSelection();
        tableModel.setParentPath(path);
    }

    // Cancel editing if the model changes.  Otherwise the editor will remain
    // open and corrupt the data.
    public void tableChanged(TableModelEvent e)
    {
        table.editingCanceled(null);
    }
} // End of SelectionListener

protected class ClickListener extends MouseAdapter
{
    public void mousePressed(MouseEvent e)
    {
        if (e.getClickCount() != 2) {
            // if the click count is not 2 then do nothing
            return;
        }

        // retrieve the path at this point
        int row = table.rowAtPoint(e.getPoint());
        if (row == -1) {
            // we have click on somewhere that is not a row
            return;
        }

        TreePath parent = tableModel.getParentPath();
        // Handle any sorting of the table
        row = JCTreeExplorer.this.table.getUnsortedRow(row);

        Object clicked_node =
            treetableModel.getChild(parent.getLastPathComponent(), row);
        if (treetableModel.isLeaf(clicked_node)) {
            // Node is a leaf so we do nothing
            return;
        }

        TreePath new_path = parent.pathByAddingChild(clicked_node);
        // cause the tree view to select the new parent
        tree.setSelectionPath(new_path);
    }
} // End of ClickListener

/**
 * Inner class that extends <code>JCSortableTable</code> and adds a modification
 * of the render/editors, which allows overiding of column 1 to store the icon
 * render/editor functionality.
 */
protected class TreeIconRenderTable extends JCSortableTable
{
    public TreeIconRenderTable(TableModel model)
    {
        super(model);
    }

    /**
     * Overridden to return the appropriate data render for the first column if
     * the treetable is in tree display mode.
     * @return appropriate data render
     * @param row       the row of the cell to render, where 0 is the first
     * @param column    the column of the cell to render, where 0 is the first
     */
    public TableCellRenderer getCellRenderer(int row, int column)
    {
        if (column == 0) {
            return(firstColumnRenderer);
        }
        // Return the normal editor
        return(super.getCellRenderer(row, column));
    }

    /**
     * Overridden to return the appropriate data render for the first column if
     * the treetable is in tree display mode.
     * @return appropriate data render
     * @param row       the row of the cell to edit, where 0 is the first
     * @param column    the column of the cell to edit, where 0 is the first
     */
    public TableCellEditor getCellEditor(int row, int column)
    {
        if (column == 0) {
            return(firstColumnEditor);
        }
        // Return the normal editor
        return(super.getCellEditor(row, column));
    }

    /**
     * Supplied to figure out the real renderer needed. Simply calls the
     * unmodified super class.
     * @return
     * @param row       the row of the cell to render, where 0 is the first
     * @param column    the column of the cell to render, where 0 is the first
     */
    public TableCellRenderer getRealCellRenderer(int row, int column)
    {
        return(super.getCellRenderer(row, column));
    }

    /**
     * Supplied to figure out the real editor needed. Simply calls the
     * unmodified super class.
     * @return
     * @param row       the row of the cell to edit, where 0 is the first
     * @param column    the column of the cell to edit, where 0 is the first
     */
    public TableCellEditor getRealCellEditor(int row, int column)
    {
        return(super.getCellEditor(row, column));
    }
} // end sortabletable

/**
 * PENDING
 * Inner class that extends <code>JTree</code> and adds modification of
 * the <code>JTree</code>; this allows overides to the default render and
 * editor to enable "plug-in" of a <code>JCTreeIconRenderer</code> to the
 * <code>JTree</code>.
 */
protected class TreeIconRenderTree extends JTree {
    public TreeIconRenderTree(TreeModel tree)
    {
        super(tree);
    }
}

}
