/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: GaugeUtil.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Polygon;
import java.awt.Color;

import com.klg.jclass.util.legend.JCLegend;

/**
 * This class contains common methods for performing linear transformations.
 *
 */
public class GaugeUtil {

/**
 * Clamp value between min and max
 *
 * @param value The value to be clamped
 * @param min The min the value can be
 * @param max The max the value can be
 *
 * @return The clamped value
 */
public static double clamp(double value, double min, double max)
{
    if (value < min) {
        value = min;
    }
    if (value > max) {
        value = max;
    }
    return(value);
}

/**
 * Converts a double value to an angle in degrees. start_value and
 * stop_value are the min and max values for a given circular range (very
 * often these are the min and max values of a circular scale). start_angle and
 * stop_angle are the starting and stopping angle for the range.
 *
 * @param value The value to be transformed
 * @param start_value The start value of the circular range
 * @param stop_value The stop value of the circular range
 * @param start_angle The start angle of the circular range
 * @param stop_angle The stop angle of the circular range
 */
public static double valueToAngle(double value,
                                  double start_value, double stop_value,
                                  double start_angle, double stop_angle)
{
	double		degrees_per_value, angle;
	double		angle_diff, value_diff;

	if (start_angle >= stop_angle) {
		stop_angle += 360.0;
	}

	angle_diff = stop_angle - start_angle;
	value_diff = (Math.max(stop_value, start_value) -
                  Math.min(start_value, stop_value));

	degrees_per_value = angle_diff / value_diff;

	angle = degrees_per_value * (value - start_value) + start_angle;
	angle = normalizeAngle(angle);
    return(angle);
}

/**
 * Converts a double value to an position in pixels. start_value and
 * stop_value are the min and max values for a given linear range (very
 * often these are the min and max values of a linear scale). start_pos and
 * stop_pos are the starting and stopping position for the range.
 *
 * @param value The value to be transformed
 * @param start_value The start value of the linear range
 * @param stop_value The stop value of the linear range
 * @param start_pos The start angle of the linear range
 * @param stop_pos The stop angle of the linear range
 */
public static int valueToPosition(double value,
                                  double start_value, double stop_value,
                                  int start_pos, int stop_pos)
{
    int         pos;
	double		pixels_per_value;
	double		value_diff, length_diff;

	length_diff = (Math.max(stop_pos, start_pos) -
                   Math.min(start_pos, stop_pos));
	value_diff  = (Math.max(stop_value, start_value) -
                   Math.min(start_value, stop_value));

    pixels_per_value = length_diff / value_diff;

	pos = (int)(pixels_per_value * (value - start_value)) + start_pos;
    return(pos);
}

/**
 * Returns an angle given a point and the origin.
 *
 * @param x The x-coordinate of the point
 * @param y The y-coordinate of the point
 * @param x_orig The x-coordinate of the origin
 * @param y_orig The x-coordinate of the origin
 */
public static double pointToAngle(double x, double y,
                                  double x_orig, double y_orig)
{
    double  x_dist, y_dist, dval, angle;

    x_dist = x - x_orig;
    y_dist = y - y_orig;

    // Some implemenations of atan2 cannot handle x == 0
    dval  = (y_dist < 0.0) ? (Math.PI/2.0) : (-Math.PI/2.0);
    angle = (Math.abs(x_dist) > 0.0) ? Math.atan2(-y_dist, x_dist) : dval;

    return(fromRadians(angle));
}

/**
 * Given an angle, returns the closest circular scale value.
 *
 * @param angle An angle
 * @param startValue The circular scale's start value
 * @param stopValue The circular scale's stop value
 * @param startAngle The circular scale's start angle
 * @param stopAngle The circular scale's stop angle
 * @return The closest scale value corresponding to the given angle
 */
public static double angleToValue(double angle,
                                  double startValue, double stopValue,
				                  double startAngle, double stopAngle)
{
	if (startAngle >= stopAngle) {
		stopAngle += 360.0;
	}

	double angleDiff = stopAngle - startAngle;
	double valueDiff = (Math.max(stopValue, startValue) -
                        Math.min(startValue, stopValue));

	double valuePerDegrees = valueDiff / angleDiff;

	double value = (valuePerDegrees * (angle - startAngle) +
                    Math.min(startValue, stopValue));

	return(value);
}

/**
 * Given a position, returns the closest linear scale value.
 *
 * @param pos A position
 * @param startValue The linear scale's start value
 * @param stopValue The liner scale's stop value
 * @param startPos The linear scale's starting position
 * @param stopPos The linear scale's stoping position
 * @return The closest scale value corresponding to the given position
 */
public static double positionToValue(int pos,
                                     double startValue, double stopValue,
				                     int startPos, int stopPos)
{
	double lengthDiff = (Math.max(stopPos, startPos) -
                         Math.min(startPos, stopPos));
	double valueDiff  = (Math.max(stopValue, startValue) -
                         Math.min(startValue, stopValue));

	double valuePerPixel = valueDiff / lengthDiff;

	double value =  valuePerPixel * (double)(pos - startPos) + startValue;
	return(value);
}

/**
 * Returns the difference of two angles.
 *
 * @param a1 An angle
 * @param a2 An angle
 * @return The difference of the two angles
 */
public static double distanceBetweenAngles(double a1, double a2)
{
	double distance   = Math.abs(a1 - a2);

	if (Math.abs(360.0 - distance) < distance) {
		distance = Math.abs(360.0 - distance);
	}

	return(distance);
}

/**
 * Java uses the notion of <code>startAngle</code> and <code>arcAngle</code>
 * to draw arcs, so this method returns the <code>arcAngle</code> from
 * <code>startAngle</code> and <code>stopAngle</code>.
 *
 * @param startAngle
 * @param stopAngle
 */
public static double arcAngle(double startAngle, double stopAngle)
{
	double arcAngle = 0;

    // Full circle
    if (startAngle == stopAngle) {
    	arcAngle = 360;
    }
	else if (startAngle < stopAngle) {
    	arcAngle = stopAngle - startAngle;
    }
    else {
    	arcAngle = stopAngle + 180;
    }

    return(arcAngle);
}

/**
 * Returns the angle between 0&#176; and 360&#176; corresponding to the
 * inputted angle
 *
 * @param angle Any angle
 * @return An angle between 0&#176; and 360&#176;
 */
public static double normalizeAngle(double angle)
{
	while (angle >= 360.0) {
		angle -= 360.0;
	}
	while (angle < 0.0) {
		angle += 360.0;
	}
	return(angle);
}

/**
 * Converts degrees to radians.
 *
 * @param degrees
 * @return The input angle converted to radians
 */
public static double toRadians(double degrees)
{
	return(degrees * (Math.PI / 180.0));
}

/**
 * Converts radians to degrees.
 *
 * @param radians The radians to be converted to degrees
 * @return The input angle converted to degrees
 */
public static double fromRadians(double radians)
{
	return(radians * (180.0 / Math.PI));
}

/**
 * Scales an object. Amounts greater than one increase its size,
 * amounts less than one decrease
 * its size.
 *
 * @param p the input polygon
 * @param xAmount the amount by which the width is to be scaled
 * @param yAmount the amount by which the height is to be scaled
 * @return A new, scaled polygon
 */
public static Polygon scale(Polygon p, double xAmount, double yAmount)
{
	Polygon poly	= new Polygon();
	int xpoints[]	= p.xpoints;
	int ypoints[]	= p.ypoints;

	for (int v = 0; v < xpoints.length; v++) {
		// Scale and add to new polygon
		poly.addPoint((int) (xpoints[v] * xAmount),
                      (int) (ypoints[v] * yAmount));
     }
    return(poly);
}

/**
 * 2D rotation of points.
 *
 * @param p The input polygon
 * @param degrees The number of degrees to rotate this polygon
 * @return A new, rotated polygon
 */
public static Polygon rotate(Polygon p, double degrees)
{
    if (degrees == 0) {
        return(p);
    }

	Polygon poly	= new Polygon();
	int xpoints[]	= p.xpoints;
	int ypoints[]	= p.ypoints;

	// Convert degrees to radians
	double radians	= toRadians(degrees);

	// Pre-compute sin, cos
	double sin = Math.sin(radians);
	double cos = Math.cos(radians);

	for (int v = 0; v < xpoints.length; v++) {
		// Get the old point
		int old_x = xpoints[v];
		int old_y = ypoints[v];

		// Compute the new point
		int new_x = (int) (old_x * cos + old_y * sin);
		int new_y = (int) (old_y * cos - old_x * sin);

		poly.addPoint(new_x, new_y);
     }
     return(poly);
}

/**
 * Translate a 2D polygon.
 *
 * @param p The input polygon
 * @param xamount The amount by which the polygon will be shifted
 *     horizontally
 * @param yamount The amount by which the polygon will be shifted vertically
 * @return a new, translated polygon
 */
public static Polygon translate(Polygon p, int xamount, int yamount)
{
	Polygon poly = new Polygon();
	for (int v = 0; v < p.xpoints.length; v++) {
         int newX = p.xpoints[v] + xamount;
         int newY = p.ypoints[v] + yamount;
         poly.addPoint(newX, newY);
    }
	return(poly);
}

/**
 * Get a polygon position and contour given a circular scale, inner and 
 * outer extent, its width, a value on the scale at which to draw, and 
 * various other attributes.
 *
 * @param cscale The circular scale used to base the polygon drawing on
 * @param inner_extent The inner extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param outer_extent The outer extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param width The width of the polygon (in pixels)
 * @param value The value along the scale at which to draw (it must be
 *              between the scale's min and max)
 * @param is_reversed Is the polygon reversed?
 * @param polygonIsCentered If the polygon is centered, place it midway between
 *                          the inner and outer extent.  Otherwise, place the
 *                          polygon origin at left extent.
 * @param polygon The polygon to draw.  The JCPolygon class is a class that
 *                extends Polygon.  It contains a method to get a bounding
 *                rectangle around a polygon.
 * @return Polygon
 */
public static Polygon getCircularPolygon(JCCircularScale cscale,
                                         double inner_extent,
                                         double outer_extent,
                                         double width,
                                         double value,
                                         boolean is_reversed,
                                         boolean polygonIsCentered,
                                         JCPolygon polygon)
{
    // Get the length as fraction of the scale's radius
    double radius_size = cscale.getRadius();
    inner_extent *= radius_size;
    outer_extent *= radius_size;
    double drawLength = (outer_extent - inner_extent);

	// Use it to compute scale factors
    Rectangle extrema = polygon.getExtrema();
	double xScale	  = drawLength / (extrema.x - extrema.width);
	double yScale	  = width / (extrema.y - extrema.height);

	// Scale the indicator
	Polygon scaled = scale(polygon, Math.abs(xScale), Math.abs(yScale));

    // mMssage the value based on direction
    double circ_min = cscale.getMin();
    double circ_max = cscale.getMax();
    double tmpValue = clamp(value, circ_min, circ_max);
    if (cscale.getDirection().equals(JCAbstractScale.Direction.BACKWARD)) {
        tmpValue = circ_max - tmpValue + circ_min;
    }

	// Using the value compute the angle
    double start_angle = normalizeAngle(cscale.getStartAngle());
    double stop_angle  = normalizeAngle(cscale.getStopAngle());
    double angle = valueToAngle(tmpValue, circ_min, circ_max,
       		                    start_angle, stop_angle);
    if (is_reversed) {
        angle = normalizeAngle(angle + 180.0);
    }

    // Use it to get the radians
    double radians = toRadians(angle);

    // Rotate to angle
	Polygon rotated	= rotate(scaled, angle);

    // If the polygon is centered, place it midway between inner and outer
    // extent.  Otherwise, place polygon origin at left extent.
    double midway     = inner_extent + drawLength / 2.0;
    double extent_val = polygonIsCentered ? midway : inner_extent;

  	// get the center of the circle
    Rectangle bounds   = cscale.getCircularGauge().getArcBounds();
    double originX     = bounds.x + radius_size;
 	double originY     = bounds.y + radius_size;
    int x_inner, y_inner;
    if (is_reversed) {
        if (!polygonIsCentered) {
            extent_val += drawLength;
        }
	    x_inner = (int)originX - (int)(extent_val * Math.cos(radians));
        y_inner = (int)originY + (int)(extent_val * Math.sin(radians));
    }
    else {
	    x_inner = (int)originX + (int)(extent_val * Math.cos(radians));
        y_inner = (int)originY - (int)(extent_val * Math.sin(radians));
     }
    
    int length = rotated.npoints + 1;
    Polygon contour = new Polygon(new int[length], new int[length], length);
    contour.xpoints[0] = x_inner;   // polygon x position
    contour.ypoints[0] = y_inner;   // polygon y position
    for (int i = 1; i < contour.npoints; i++){ // polygon contour
        contour.xpoints[i] = rotated.xpoints[i - 1];
        contour.ypoints[i] = rotated.ypoints[i - 1];
    }
    return(contour);
}

/**
 * Draw a polygon given a circular scale, inner and outer extent,
 * its width, a value on the scale at which to draw, and various other
 * attributes.
 *
 * @param g The graphics context on which to draw
 * @param cscale The circular scale used to base the polygon drawing on
 * @param foreground The foreground color of the polygon
 * @param inner_extent The inner extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param outer_extent The outer extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param width The width of the polygon (in pixels)
 * @param value The value along the scale at which to draw (it must be
 *              between the scale's min and max)
 * @param is_reversed Is the polygon reversed?
 * @param polygonIsCentered If the polygon is centered, place it midway between
 *                          the inner and outer extent.  Otherwise, place the
 *                          polygon origin at left extent.
 * @param polygon The polygon to draw.  The JCPolygon class is a class that
 *                extends Polygon.  It contains a method to get a bounding
 *                rectangle around a polygon.
 */
public static void drawCircularPolygon(Graphics g,
                                       JCCircularScale cscale,
                                       Color foreground,
                                       double inner_extent,
                                       double outer_extent,
                                       double width,
                                       double value,
                                       boolean is_reversed,
                                       boolean polygonIsCentered,
                                       JCPolygon polygon)
{
    Polygon contour = getCircularPolygon(cscale, inner_extent, outer_extent,
                                         width, value, is_reversed,
                                         polygonIsCentered, polygon);
    // Define the polygon to fill in
    int length = contour.npoints;
    Polygon drawPolygon = new Polygon(new int[length - 1], new int[length - 1],
                                      length - 1);
    for(int i = 1; i < length; i++){
        drawPolygon.xpoints[i - 1] = contour.xpoints[i];
        drawPolygon.ypoints[i - 1] = contour.ypoints[i];
    }       

   	// Create a temporary gc onto which to draw
	Graphics gc = g.create();

    // Set the foreground
    gc.setColor(foreground);

	// Move (0,0) to the x, y positon of polygon
    gc.translate(contour.xpoints[0], contour.ypoints[0]);

	// Finally, draw
	gc.fillPolygon(drawPolygon);

	// Dispose
	gc.dispose();
}

/**
 * Get the contour of a polygon and its position given a linear scale, inner 
 * and outer extent, its width, a value on the scale at which to draw, and 
 * various other attributes.The xpoints[0], ypoints[0] from Polygon return 
 * the position of polygon. The points xpoints[1], ypoints[1]... return the
 * contour of the polygon.
 * 
 * @param lscale The linear scale used to base the polygon drawing on
 * @param inner_extent The inner extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param outer_extent The outer extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param width The width of the polygon (in pixels)
 * @param value The value along the scale at which to draw (it must be
 *              between the scale's min and max)
 * @param is_reversed Is the polygon reversed?
 * @param polygonIsCentered If the polygon is centered, place it midway between
 *                          the inner and outer extent.  Otherwise, place the
 *                          polygon origin at left extent.
 * @param polygon The polygon to draw.  The JCPolygon class is a class that
 *                extends Polygon.  It contains a method to get a bounding
 *                rectangle around a polygon.
 * @param bounds The reference rectangle for contour. For drawing used
 * lscale.getScaleSize(). For image mapping use lscale.getBounds()
 * @return Polygon The absolute position and contour of the polygon  
 */
public static Polygon getLinearPolygon(JCLinearScale lscale,
                                       double inner_extent,
                                       double outer_extent,
                                       double width,
                                       double value,
                                       boolean is_reversed,
                                       boolean polygonIsCentered,
                                       JCPolygon polygon,
                                       Rectangle bounds)
{
    JCLinearScale.Orientation orientation = lscale.getOrientation();
    JCAbstractScale.Direction direction = lscale.getDirection();

    // Get the various values
	double lin_min = lscale.getMin();
	double lin_max = lscale.getMax();

    // Adjust for direction and orientation
    value = clamp(value, lin_min, lin_max);
    if (direction.equals(JCAbstractScale.Direction.BACKWARD) ^
        (orientation == JCLinearScale.Orientation.VERTICAL))
    {
		value  = lin_max - value + lin_min;
	}

	//	Determine the scaled inner and outer extents (in pixels).
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
	    inner_extent *= (double)bounds.width;
	    outer_extent *= (double)bounds.width;
    }
    else {
	    inner_extent *= (double)bounds.height;
	    outer_extent *= (double)bounds.height;
    }
    double drawLength  = Math.abs(outer_extent  - inner_extent);

	// Use it to compute scale factors
    Rectangle extrema = polygon.getExtrema();
	double xScale	  = drawLength / (extrema.x - extrema.width);
	double yScale	  = width / (extrema.y - extrema.height);

	// Scale the polygon
	Polygon scaled = scale(polygon, Math.abs(xScale), Math.abs(yScale));

    // Rotate the polygon in one of the four possible directions
    double angle;
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
        angle = is_reversed ? 180 : 0;
    }
    else {
        angle = is_reversed ? 90 : 270;
    }

    // Rotate to angle
	Polygon rotated	= rotate(scaled, angle);

    // If the polygon is centered, place it midway between inner and outer
    // extent.  Otherwise, place polygon origin at left extent.
    double midway     = inner_extent + drawLength / 2.0;
    double extent_val = polygonIsCentered ? midway : inner_extent;
    if (is_reversed && !polygonIsCentered) {
        extent_val += drawLength;
    }

    // Determine the position to place the rotated polygon
    int x_inner, y_inner;
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
        y_inner = valueToPosition(value, lin_min, lin_max,
                                  bounds.y, bounds.y + bounds.height);
        x_inner = bounds.x + (int)extent_val;
    }
    else {
        x_inner = valueToPosition(value, lin_min, lin_max,
                                  bounds.x, bounds.x + bounds.width);
        y_inner = bounds.y + (int)extent_val;
    }
   
    int length = rotated.npoints + 1;
    Polygon contour = new Polygon(new int[length], new int[length], length);
    contour.xpoints[0] = x_inner;   // polygon x position
    contour.ypoints[0] = y_inner;   // polygon y position
    for (int i = 1; i < contour.npoints; i++) { // polygon contour
        contour.xpoints[i] = rotated.xpoints[i - 1];
        contour.ypoints[i] = rotated.ypoints[i - 1];
    }
    return(contour);
}

/**
 * Draw a polygon given a linear scale, inner and outer extent,
 * its width, a value on the scale at which to draw, and various other
 * attributes.
 *
 * @param g The graphics context on which to draw
 * @param lscale The linear scale used to base the polygon drawing on
 * @param foreground The foreground color of the polygon
 * @param inner_extent The inner extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param outer_extent The outer extent of the polygon (based on the radius
 *                     radius of the circular scale).
 * @param width The width of the polygon (in pixels)
 * @param value The value along the scale at which to draw (it must be
 *              between the scale's min and max)
 * @param is_reversed Is the polygon reversed?
 * @param polygonIsCentered If the polygon is centered, place it midway between
 *                          the inner and outer extent.  Otherwise, place the
 *                          polygon origin at left extent.
 * @param polygon The polygon to draw.  The JCPolygon class is a class that
 *                extends Polygon.  It contains a method to get a bounding
 *                rectangle around a polygon.
 */
public static void drawLinearPolygon(Graphics g,
                                     JCLinearScale lscale,
                                     Color foreground,
                                     double inner_extent,
                                     double outer_extent,
                                     double width,
                                     double value,
                                     boolean is_reversed,
                                     boolean polygonIsCentered,
                                     JCPolygon polygon)
{
    Polygon contour = getLinearPolygon(lscale, inner_extent, outer_extent,
                                       width,  value, is_reversed,
                                       polygonIsCentered, polygon,
                                       lscale.getScaleSize());
    // define the polygon to fill in
    int length = contour.npoints;
    Polygon drawPolygon = new Polygon(new int[length - 1], new int[length - 1],
                                      length - 1);
    for(int i = 1; i < length; i++){
        drawPolygon.xpoints[i - 1] = contour.xpoints[i];
        drawPolygon.ypoints[i - 1] = contour.ypoints[i];
    }        

	// Create a temporary gc onto which to draw
	Graphics gc = g.create();

    // Set the foreground
    gc.setColor(foreground);

	// Move (0,0) to the x, y positon of polygon
    gc.translate(contour.xpoints[0], contour.ypoints[0]);

	// Finally, draw
	gc.fillPolygon(drawPolygon);

	// Dispose
	gc.dispose();
}

/**
 * Get a circle based on a circular scale based on the value,
 * inner_extent, and outer_extent.
 *
 * @param cscale The circular scale used to determine radius & position
 * @param value The value on the scale at which this circle will be drawn
 * @param inner_extent The inner extent as a fraction of the radius
 * @param outer_extent The outer extent as a fraction of the radius
 * @return int[] The absolute(x,y) position of center and radius
 */
public static int[] getCircleForCircularScale(JCCircularScale cscale,
                                              double value,
		                                      double inner_extent,
                                              double outer_extent)
{
    Rectangle bounds = cscale.getCircularGauge().getArcBounds();

    // Get the center of the scale
    double radius_size = cscale.getRadius();
	double x_orig      = bounds.x + radius_size;
	double y_orig      = bounds.y + radius_size;

    // Get the scaled inner and outer extent
    inner_extent *= radius_size;
    outer_extent *= radius_size;

	// Get the distance this circle's center is from the center of the scale
	double length      = outer_extent - inner_extent;
	double half_length = length / 2.0;
	double midway      = inner_extent + half_length;

	// Massage the value based on direction
    double circ_min = cscale.getMin();
    double circ_max = cscale.getMax();
    double tmpValue = clamp(value, circ_min, circ_max);
    if (cscale.getDirection().equals(JCAbstractScale.Direction.BACKWARD)) {
        tmpValue = circ_max - tmpValue + circ_min;
    }

	// Using the value compute the angle
    double start_angle = normalizeAngle(cscale.getStartAngle());
    double stop_angle  = normalizeAngle(cscale.getStopAngle());
    double angle = valueToAngle(tmpValue, circ_min, circ_max,
       			                start_angle, stop_angle);
    // Use it to get the radians
    double radians = toRadians(angle);

	// Compute center of the circle
	int x = (int)x_orig + (int)(midway * Math.cos(radians));
	int y = (int)y_orig - (int)(midway * Math.sin(radians));

    int[] coords = new int[3];
    coords[0] = x;  // x position
    coords[1] = y;  // y position
    coords[2] = (int)half_length; // radius
    return(coords);
}

/**
 * Draws a filled circle based on a circular scale based on the value,
 * inner_extent, and outer_extent.
 *
 * @param g The graphics context on which to draw
 * @param cscale The circular scale used to determine radius & position
 * @param foreground The foreground color of the circle
 * @param value The value on the scale at which this circle will be drawn
 * @param inner_extent The inner extent as a fraction of the radius
 * @param outer_extent The outer extent as a fraction of the radius
 */
public static void drawCircleForCircularScale(Graphics g,
                                              JCCircularScale cscale,
                                              Color foreground,
                                              double value,
		                                      double inner_extent,
                                              double outer_extent)
{
	int[] circle = getCircleForCircularScale(cscale, value, inner_extent, 
                                             outer_extent);
    int x = circle[0];
    int y = circle[1];
    int radius = circle[2];
	
    // Create a temporary graphics context
	Graphics gc = g.create();

    // Set its color
	gc.setColor(foreground);

 	//	For drawing we need to be at the upper left of the square
	//	drawn around the circle
	x -= radius;
	y -= radius;
    
    int diameter = 2 * radius;
	gc.fillArc(x, y, diameter, diameter, 0, 360);

	// Clean up
	gc.dispose();
}

/**
 * Get the center position and the radius of a circle based on a linear scale 
 * based on the value, inner_extent, and outer_extent.
 *
 * @param lscale The linear scale used to determine radius & position
 * @param value The value on the scale at which this circle will be drawn
 * @param inner_extent The inner extent as a fraction of the radius
 * @param outer_extent The outer extent as a fraction of the radius
 * @param bounds Rectangle The reference rectangle for contour. For drawing used
 * lscale.getScaleSize(). For image mapping use lscale.getBounds()
 * @return int[] The returned array contains the absolute center position (x, y)
 * and the radius of the circle
 */
public static int[] getCircleForLinearScale(JCLinearScale lscale, double value,
                                            double inner_extent,
                                            double outer_extent,
                                            Rectangle bounds)
{
    JCLinearScale.Orientation orientation = lscale.getOrientation();
    JCAbstractScale.Direction direction   = lscale.getDirection();

    // Get the various values
	double lin_min = lscale.getMin();
	double lin_max = lscale.getMax();

    // Adjust for direction and orientation
    value = clamp(value, lin_min, lin_max);
    if (direction.equals(JCAbstractScale.Direction.BACKWARD) ^
        (orientation == JCLinearScale.Orientation.VERTICAL))
    {
		value  = lin_max - value + lin_min;
	}

	//	Determine the scaled inner and outer extents (in pixels).
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
	    inner_extent *= (double)bounds.width;
	    outer_extent *= (double)bounds.width;
    }
    else {
	    inner_extent *= (double)bounds.height;
	    outer_extent *= (double)bounds.height;
    }

	// Calculate the extent the circle's center is away from the top or left
    // of the scale (depending on orientation)
	double length      = outer_extent - inner_extent;
	double half_length = length / 2.0;
	double midway      = inner_extent + half_length;

    // Get the position of the value between the min and max of the scale
    int position;
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
        position = valueToPosition(value, lin_min, lin_max,
                                   bounds.y, bounds.y + bounds.height);
    }
    else {
        position = valueToPosition(value, lin_min, lin_max,
                                   bounds.x, bounds.x + bounds.width);
    }

	//	Determine the position center of the circle
    int x, y;
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
        x = bounds.x + (int)midway;
        y = position;
    }
    else {
        x = position;
        y = bounds.y + (int)midway;
    }
    int[] coords = new int[3];
    coords[0] = x;  // x position
    coords[1] = y;  // y position
    coords[2] = (int)half_length; // radius
    return(coords);
}

/**
 * Draws a filled circle based on a linear scale based on the value,
 * inner_extent, and outer_extent.
 *
 * @param g The graphics context on which to draw
 * @param lscale The linear scale used to determine position and width/height
 * @param foreground The foreground color of the circle
 * @param value The value on the scale at which this circle will be drawn
 * @param inner_extent The inner extent as a fraction of the width/height
 * @param outer_extent The outer extent as a fraction of the width/height
 */
public static void drawCircleForLinearScale(Graphics g,
                                            JCLinearScale lscale,
                                            Color foreground,
                                            double value,
		                                    double inner_extent,
                                            double outer_extent)
{
    int[] circle = getCircleForLinearScale(lscale, value, inner_extent, 
                                           outer_extent, lscale.getScaleSize());
    int x = circle[0];
    int y = circle[1];
    int radius = circle[2];
    
    // Create a temporary graphics context
	Graphics gc = g.create();

    // Set its color
	gc.setColor(foreground);

 	//	For drawing we need to be at the upper left of the square
	//	drawn around the circle
	x -= radius;
	y -= radius;
    
    int diameter = 2 * radius;
	gc.fillArc(x, y, diameter, diameter, 0, 360);

	// Clean up
	gc.dispose();
}

/**
 * Creates a LegendPopulatorRenderer for a given gauge and scale.
 *
 * @param gauge The gauge whose legend needs a PopulatorRenderer
 * @param scale The scale whose ranges need to be part of a legend.
 */
public static void createLegendPopulatorRenderer(JCGauge gauge, JCScale scale)
{
    JCLegend legend = gauge.getLegend();

	// add the default legend populator/renderer
    DefaultLegendPopulatorRenderer pop =
    	new DefaultLegendPopulatorRenderer(legend, scale.getRanges());
    legend.setLegendPopulator(pop);
    legend.setLegendRenderer(pop);
}

/**
* 
* @param input Polygon to be translated
* @return a Polygon that includes all the points of the input polygon excepts the
* first point, the pairs (x,y) beeing translated to the that point  
*/
public static Polygon translateContour(Polygon input)
{
    int length = input.npoints;
    Polygon contour = new Polygon(new int[length - 1], new int[length - 1],
                                  length - 1);
    for (int i = 0; i < contour.npoints; i++) {
        contour.xpoints[i] = input.xpoints[i + 1] + input.xpoints[0];
        contour.ypoints[i] = input.ypoints[i + 1] + input.ypoints[0];
    }
    return(contour);
}

}

