/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCAbstractIndicator.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import javax.swing.*;
import java.awt.*;

/**
 * An indicator points to a particular value on its associated scale.
 * There may be multiple indicators associated with a scale.
 *
 * @see com.klg.jclass.swing.gauge.JCIndicatorStyle
 */
public abstract class JCAbstractIndicator extends JComponent
                                          implements JCIndicator
{

static final long serialVersionUID = 7215270223197984024L;

/**
 *  The indicator's associated scale.
 */
protected JCScale scale;

/**
 * The indicator's width in pixels.
 */
protected double indicatorWidth = 15;

/**
 * Asks: "Is the indicator reversed?".
 */
protected boolean reversed = false;

/**
 * The inner extent of the indicator as a ratio of the scale's radius.
 */
protected double inner_extent = 0.0;

/**
 * The outer extent of the indicator as a ratio of the scale's radius.
 */
protected double outer_extent = 1.0;

/**
 * The indicator style, one of:
 * <ul>
 * <li><code>JCIndicatorStyle.ARROW</code></li>
 * <li><code>JCIndicatorStyle.CIRCLE</code></li>
 * <li><code>JCIndicatorStyle.POINTER</code></li>
 * <li><code>JCIndicatorStyle.RECTANGLE</code></li>
 * <li><code>JCIndicatorStyle.TRIANGLE</code></li>
 * <li><code>JCIndicatorStyle.TAILED_ARROW</code></li>
 * <li><code>JCIndicatorStyle.TAILED_POINTER</code></li>
 * </ul>
 * or a user-defined instance of the <code>JCIndicatorStyle</code> class.
 */
protected JCIndicatorStyle indicatorStyle = JCIndicatorStyle.RECTANGLE;

/**
 * The scale value at which this indicator should point.
 */
protected double value;

/**
 * Creates a indicator given a scale.  All other properties are at
 * their default values (the value property takes on the scale's min value by
 * default).
 *
 * @param scale the associated circular scale
 */
public JCAbstractIndicator(JCScale scale)
{
    this.scale = scale;

    // default the value to scale's min value
	this.value = scale.getMin();
}

/**
 * Creates an indicator.
 *
 * @param foreground the indicator's color
 * @param indicatorWidth the indicator's width in pixels
 * @param scale the associated circular scale
 * @param visible a Boolean that controls the visibility of the indicator
 * @param inner_extent the indicator's <code>inner_extent</code>, expressed as
 *        a ratio of the indicator's inner position to the top or center of the
 *        associated scale
 * @param outer_extent the indicator's <code>outer_extent</code>, expressed as
 *        a ratio of the indicator's outer position to the top or center of the
 *        associated scale
 * @param indicatorStyle indicator styles are kept in
 *        <code>JCIndicatorStyle</code>; user-defined indicator styles may be
 *        passed to <code>JCIndicatorStyle</code>'s constructor
 * @param value the indicator's value; the indicator points to this value on its
 *        associated scale
 */
public JCAbstractIndicator(
        Color						foreground,
        double						indicatorWidth,
        JCScale				        scale,
        boolean						visible,
        double						inner_extent,
        double						outer_extent,
        JCIndicatorStyle			indicatorStyle,
        double						value)
{
    this.indicatorWidth  = indicatorWidth;
    this.scale           = scale;
    this.inner_extent    = inner_extent;
    this.outer_extent    = outer_extent;
    this.indicatorStyle  = indicatorStyle;
    this.value    	     = value;

	setForeground(foreground);
	setVisible(visible);
}

/**
 * Returns the preferred size for this component.
 *
 * @return the preferred size
 */
public Dimension getPreferredSize()
{
	return(scale.getGauge().getGaugeArea().getPreferredSize());
}

/**
 * This routine actually draws the indicator object on the screen.
 *
 * @param g
 */
public abstract void paint(Graphics g);

//////////// setters and getters

/**
 * Returns the scale associated with this indicator.
 *
 * @return the scale for the <code>JCAbstractIndicator</code> component
 */
public JCScale getScale()
{
	return(scale);
}

/**
 * Sets the indicator's scale to the supplied value.
 *
 * @param scale the new scale
 */
protected void setScale(JCScale scale)
{
    this.scale = scale;

    value = GaugeUtil.clamp(value, scale.getMin(), scale.getMax());
}

/**
 * Returns the scale value at which this indicator is currently
 * pointing.
 *
 * @return the current value pointed to by this indicator
 */
public double getValue()
{
	return(value);
}

/**
 * Sets the scale value to which this indicator should point.
 *
 * @param value the value at which to point
 */
public void setValue(double value)
{
	// Only interested in value changes.
    // Make sure new value is in bounds.
    value = GaugeUtil.clamp(value, scale.getMin(), scale.getMax());
	if (this.value != value) {

		// Assign the new value
        this.value = value;

        redraw();
	}
}

/**
 * Returns the inner extent of this indicator as a ratio of the
 * scale's width, height, or radius (depending on the scale's type and
 * orientation). For example, in the circular case, 0.75 means draw out
 * to 75% of the scale's radius.
 *
 * @return The <code>inner_exent</code> of this indicator as a ratio of the
 *         scale's width, height, or radius (depending on the scale's type and
 *         orientation)
 */
public double getInnerExtent()
{
	return(this.inner_extent);
}

/**
 * Sets the inner extent of this indicator as a ratio of the
 * scale's width, height, or radius (depending on the scale's type and
 * orientation). For example, in the circular case, 0.75 means draw out
 * to 75% of the scale's radius.
 *
 * @param inner_extent the inner extent of this indicator as a ratio of the
 *        scale's width, height, or radius (depending on the scale's type and
 *        orientation)
 */
public void setInnerExtent(double inner_extent)
{
	this.inner_extent = inner_extent;
    redraw();
}

/**
 * Returns the outer extent of this indicator as a ratio of the
 * scale's width, height, or radius (depending on the scale's type and
 * orientation). For example, in the circular case, 0.75 means draw out
 * to 75% of the scale's radius.
 *
 * @return The <code>outer_exent</code> of this indicator as a ratio of the
 *         scale's width, height, or radius (depending on the scale's type and
 *         orientation)
 */
public double getOuterExtent()
{
	return(this.outer_extent);
}

/**
 * Sets the outer extent of this indicator as a ratio of the
 * scale's width, height, or radius (depending on the scale's type and
 * orientation). For example, in the circular case, 0.75 means draw out
 * to 75% of the scale's radius.
 *
 * @param outer_extent the outer extent of this indicator as a ratio of the
 *        scale's width, height, or radius (depending on the scale's type and
 *        orientation)
 */
public void setOuterExtent(double outer_extent)
{
	this.outer_extent = outer_extent;
    redraw();
}

/**
 * Returns the <code>JCIndicatorStyle</code> for this indicator.
 *
 * @return the <code>JCIndicatorStyle</code> of this indicator
 */
public JCIndicatorStyle getIndicatorStyle()
{
	return(this.indicatorStyle);
}

/**
 * Sets the <code>JCIndicatorStyle</code> for this indicator.
 *
 * @param indicatorStyle the <code>JCIndicatorStyle</code> for this indicator
 */
public void setIndicatorStyle(JCIndicatorStyle indicatorStyle)
{
	this.indicatorStyle = indicatorStyle;
    redraw();
}

/**
 * Returns the indicator width in pixels.
 *
 * @return the indicator width in pixels
 */
public double getIndicatorWidth()
{
	return(indicatorWidth);
}

/**
 * Sets the indicator width in pixels.
 *
 * @param indicatorWidth the indicator width in pixels
 */
public void setIndicatorWidth(double indicatorWidth)
{
	this.indicatorWidth = indicatorWidth;
    redraw();
}

/**
 * Asks :"Is the indicator reversed?".
 *
 * @return whether the indicator is reversed or not
 */
public boolean isReversed()
{
	return(reversed);
}

/**
 * Sets the reversed property of the indicator.  If the indicator is
 * reversed it draws in the opposite direction (for example, for linear
 * scales it points left or up instead of right or down, for circular
 * scales it points towards the center instead of away from the center).
 *
 * @param reversed is the indicator's direction reversed
 */
public void setReversed(boolean reversed)
{
	this.reversed = reversed;
    redraw();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
    if (scale.getGauge().getRepaintEnabled()) {
        repaint();
    }
}

} // JCAbstractIndicator
