/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCAbstractNeedle.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import com.klg.jclass.util.JCListenerList;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.util.Enumeration;

/**
 * A needle points to a current value on its associated scale.  The end user may
 * choose to interact with a needle.  There may be multiple needles
 * associated with a scale.
 *
 * @see com.klg.jclass.swing.gauge.JCAbstractNeedle
 */
public abstract class JCAbstractNeedle extends JCAbstractIndicator
                                       implements JCNeedle
{

static final long serialVersionUID = 3489006543322509139L;

/**
 * List of listeners interested in knowing when a needle's value has changed.
 */
protected JCListenerList 	changeListeners = null;

/**
 * Reuse the same event for notifications. The source will be this needle.
 */
protected ChangeEvent		changeEvent		= new ChangeEvent(this);

/**
 * Controls whether to send change events or not.
 */
protected boolean			sendEvents		= true;

/**
 * The type of interaction, one of
 * <ul>
 * <li><code>InteractionType.NONE</code></li>
 * <li><code>InteractionType.DRAG</code></li>
 * <li><code>InteractionType.CLICK</code></li>
 * <li><code>InteractionType CLICK_DRAG</code></li>
 * </ul>
 *
 */
protected InteractionType	interactionType	= InteractionType.NONE;

/**
 * Creates a needle.
 *
 * @param scale the associated circular or linear scale
 */
public JCAbstractNeedle(JCScale scale)
{
    super(scale);

    // default value to scale's min value
	this.value          = scale.getMin();
    this.indicatorStyle = JCIndicatorStyle.ARROW;
}

/**
 * Creates a needle.
 *
 * @param foreground the needle's color
 * @param needleWidth the needle's width in pixels
 * @param scale the associated <code>JCScale</code>
 * @param visible a Boolean that controls the visibility of the needle
 * @param inner_extent the needle's inner extent, expressed as a ratio of its
 *     length to the relevant dimension (radius, width/length) of the associated
 *     scale
 * @param outer_extent the needle's outer extent, expressed as a ratio of its
 *     length to the relevant dimension of the associated scale
 * @param needleStyle needle styles are kept in <code>JCIndicatorStyle</code>;
 *     user-defined needle styles may be passed to the
 *     <code>JCIndicatorStyle</code>'s constructor
 * @param value the needle's value; the needle points to this value on its
 *     associated scale
 */
public JCAbstractNeedle(Color						foreground,
                        double						needleWidth,
                        JCScale				        scale,
                        boolean						visible,
                        double						inner_extent,
                        double                      outer_extent,
                        JCIndicatorStyle			needleStyle,
                        double						value)
{
    super(foreground, needleWidth, scale, visible, inner_extent,
          outer_extent, needleStyle, value);
}

/**
 * Creates a needle.
 *
 * @param foreground the needle's color
 * @param needleWidth the needle's width in pixels
 * @param scale the associated <code>JCScale</code>
 * @param interactionType the interaction type,
 * <code>CLICK</code>, <code>CLICK_DRAG</code>, or <code>NONE</code>
 * @param visible a Boolean that controls the visibility of the needle
 * @param length the needle's length, expressed as a ratio of its length to
 *     the relevant dimension (radius, width/length) of the associated scale
 * @param needleStyle needle styles are kept in <code>JCIndicatorStyle</code>;
 *     user-defined needle styles may be passed to the
 *     <code>JCIndicatorStyle</code>'s constructor
 * @param value the needle's value; the needle points to this value on its
 *     associated scale
 * @deprecated Beginning with JClass 5, the <code>length</code> parameter has
 *  been replaced with <code>inner_extent</code> and <code>outer_extent</code>.
 */
public JCAbstractNeedle(Color						foreground,
                        double						needleWidth,
                        JCScale				        scale,
                        InteractionType				interactionType,
                        boolean						visible,
                        double						length,
                        JCIndicatorStyle			needleStyle,
                        double						value)
{
    super(foreground, needleWidth, scale, visible, 0.0, length,
          needleStyle, value);

    this.interactionType = interactionType;
}

/**
 * Returns the preferred size for this component.
 *
 * @return the preferred size
 */
public Dimension getPreferredSize()
{
	return(scale.getGauge().getGaugeArea().getPreferredSize());
}

/**
 * This routine actually draws the needle object on the screen.
 *
 * @param g
 */
public abstract void paint(Graphics g);

/**
 * Adds listener interested in needle movements.
 *
 * @param l the listener to be added
 */
public void addChangeListener(ChangeListener l)
{
    changeListeners = JCListenerList.add(changeListeners,l);
}

/**
 * Removes listener interested in needle movements.
 *
 * @param l the listener to be removed
 */
public void removeChangeListener(ChangeListener l)
{
    changeListeners = JCListenerList.remove(changeListeners,l);
}

/**
 * Broadcasts the move event to interested listeners.
 *
 * @param ev the <code>ChangeEvent</code> of interest
 */
public void sendChangeEvent(ChangeEvent ev)
{
	Enumeration e = JCListenerList.elements(changeListeners);

    while (e.hasMoreElements()) {
        ((ChangeListener)e.nextElement()).stateChanged(ev);
    }
}

/**
 * Sets the scale value to which this needle should point.  Also
 * notifies listeners if <code>sendEvents</code> is <code>true</code>.
 *
 * @param value the value at which to point
 */
public void setValue(double value)
{
	// only interested in value changes
    value = GaugeUtil.clamp(value, scale.getMin(),scale.getMax());
	if (this.value != value) {

		// assign the new value
		this.value = value;

    	// notify listeners
		if (sendEvents) {
    		sendChangeEvent(changeEvent);
		}

        redraw();
	}
}

/**
 * Returns the length of this needle as a ratio of the scale's radius, width,
 * or height (depending on the scale's type and orientation). <br><B>Note:</b>
 * The length is really the outer extent.
 *
 * @return the length of this needle as a ratio of the scale's radius, width,
 *         or height (depending on the scale's type and orientation)
 */
public double getLength()
{
	return(this.outer_extent);
}

/**
 * Sets the length of this needle as a ratio of the
 * scale's radius, width, or height (depending on the scale's type and
 * orientation). For example, for a circular scale, 0.75 means draw out to
 * 75% of the scale's radius.
 *
 * @param length the length of this needle as a ratio of the scale's radius,
 *      width, or height (depending on the scale's type and orientation)
 */
public void setLength(double length)
{
    setOuterExtent(length);
}

/**
 * Returns the <code>JCIndicatorStyle</code> for this needle.
 *
 * @return the <code>JCIndicatorStyle</code> of this needle
 */
public JCIndicatorStyle getNeedleStyle()
{
	return(this.indicatorStyle);
}

/**
 * Sets the <code>JCIndicatorStyle</code> for this needle.
 *
 * @param needleStyle the <code>JCIndicatorStyle</code> for this needle
 */
public void setNeedleStyle(JCIndicatorStyle needleStyle)
{
	this.indicatorStyle = needleStyle;
    redraw();
}

/**
 * Returns the width for this needle (this is actually stored in the inherited
 * <code>indicatorWidth</code> property).
 *
 * @return the width of this needle
 */
public double getNeedleWidth()
{
	return(getIndicatorWidth());
}

/**
 * Sets the width of needle (this is actually stored in the inherited
 * <code>indicatorWidth</code> property).
 *
 * @param needleWidth the width of this needle
 */
public void setNeedleWidth(double needleWidth)
{
	setIndicatorWidth(needleWidth);
}

/**
 * Returns the interation type for this needle.
 *
 * @return the <code>JCAbstractNeedle.InteractionType</code> for this needle
 */
public InteractionType getInteractionType()
{
	return(interactionType);
}

/**
 * Sets the interation type for this needle.
 *
 * @param interactionType the <code>JCNeedle.InteractionType</code> for this
 *     needle, one of: <ul> <li><code>InteractionType.NONE</code></li>
 *     <li><code>InteractionType.DRAG</code></li>
 *     <li><code>InteractionType.CLICK</code>
 *     </li> <li><code>InteractionType.CLICK_DRAG</code></li> </ul>
 */
public void setInteractionType(InteractionType interactionType)
{
	this.interactionType = interactionType;
}

/**
 * Returns a <code>sendEvents</code> flag; <code>true</code> means events will
 * be sent when the needle's value changes, <code>false</code> means do not
 * send the events.
 *
 * @return the <code>sendEvents</code> boolean
 */
public boolean getSendEvents()
{
	return(sendEvents);
}

/**
 * Sets the <code>sendEvents</code> flag; <code>true</code> means events will
 * be sent when the needle's value changes, <code>false</code> means do not
 * send the events.
 *
 * @param sendEvents <code>true</code> means send needle's value change events
 *     to listeners, <code>false</code> means do not send the events
 */
public void setSendEvents(boolean sendEvents)
{
	this.sendEvents = sendEvents;
}

////// inner class

/**
 * Needle interaction types. These control how the needle will
 * respond to clicks and drags.
 */
public static class InteractionType implements java.io.Serializable {

    /** Neither clicking nor dragging affects the needle.
     *
     */
    public static final InteractionType NONE 	= new InteractionType();

    /** The needle follows a mouse drag action.
     *
     */
    public static final InteractionType DRAG 	= new InteractionType();

    /** The needle snaps to a mouse click.
     *
     */
    public static final InteractionType CLICK	= new InteractionType();

    /** The needle responds to both mouse click and mouse drag actions.
     *
     */
    public static final InteractionType CLICK_DRAG = new InteractionType();

    // prevent bogus types
    /**      *
     */
    private InteractionType() {}
}

}
