/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCAbstractRange.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import javax.swing.*;
import java.awt.*;

/**
 * A range is also used to highlight some sub-portion of the scale by
 * the use of color or an image.
 *
 * For example, on a tachometer gauge you could have a range
 * showing the "red line" region of excessive revolutions per minute (rpm).
 * A gauge can have zero or more ranges associated with it.
 * <br>
 * By default, the items in a gauge's legend are ranges.
 */
public abstract class JCAbstractRange extends JComponent
                                      implements JCRange, java.io.Serializable
{

static final long serialVersionUID = -2632092271480949157L;

/**
 * The scale associated with this range object.
 */
protected JCScale 	scale;

/**
 * The place to start drawing this range. Values are ratios based on the
 * scale's radius, width, or height (depending on the scale's type and orientation).
 * For example for a circular scale, 0.8 means start drawing this range 80% of
 * the radius' size away from the center.
 *
 */
protected double	innerExtent;

/**
 * The place to stop drawing this range. Values are ratios based on the
 * scale's radius, width, or height (depending on the scale's type and orientation).
 * For example for a linear scale with horizontal oriention, 0.9 means stop
 * drawing this range 90% of the scale's height away from the top of the scale.
 *
 */
protected double	outerExtent;

/**
 * The scale value (not angle or position) at which to begin the range.
 */
protected double	startValue;

/**
 * The scale value (not angle or position) at which to stop the range.
 */
protected double	stopValue;

/**
 * The name of the range which will appear in the legend.
 */
protected String 	rangeName;

/**
 * Static counter for generated range names.
 */
protected static int rangeCounter = 0;

/**
* An image that may be used for range.
*/
protected Image image;

/**
* Boolean that determines if the image should be scaled when
* the range is resized. Default is <code>true</code>.
*/
protected boolean scaleImage = true;

/**
 * Creates a range with the properties specified in the parameters.
 *
 * @param foreground the color of this range
 * @param scale the associated scale
 * @param innerExtent The place to start drawing this range;
 *      values are ratios based on the scale's radius, width, or height
 *      (depending on the scale's type and orientation). For example, for a
 *      circular scale, 0.8 means start drawing this range 80% of the radius'
 *      size away from the center.
 * @param outerExtent The place to stop drawing this range. Values are ratios
 *      based on the scale's radius, width, or height (depending on the scale's
 *      type and orientation). For example for vertical linear scale, 0.9
 *      means stop drawing this range 90% of the scale's width away from the
 *      left of the scale.
 * @param startValue the scale value (not angle or position) at which to
 *      begin the range
 * @param stopValue the scale value (not angle or position) at which to
 *      end the range
 */
public JCAbstractRange(Color				foreground,
                       JCScale              scale,
                       double				innerExtent,
                       double				outerExtent,
                       double				startValue,
                       double				stopValue)
{
	setForeground(foreground);

    this.scale          = scale;
    this.innerExtent	= innerExtent;
    this.outerExtent	= outerExtent;
    this.startValue		= startValue;
    this.stopValue		= stopValue;
}

/**
 * Returns the scale associated with this range.
 *
 * @return the scale for the <code>JCAbstractRange</code> component
 */
public JCScale getScale()
{
	return(scale);
}

/**
 * Specifies where the inner boundary of the range is.  Values are ratios based
 * on the scale's radius, width, or height (depending on the scale's type and
 * orientation). For example, for circular scale, 0.8 means start drawing this
 * range 80% of the radius' size away from the center.
 *
 * @return the inner extent
 */
public double getInnerExtent()
{
	return(innerExtent);
}

/**
 * Sets where the inner boundary of the range is. Values are ratios based on
 * the scale's radius, width, or height (depending on the scale's type and
 * orientation). For example for horizontal linear scale, 0.9 means start
 * drawing this range 90% of the scale's height away from the top of the scale.
 *
 * @param innerExtent the inner extent of the range
 */
public void setInnerExtent(double innerExtent)
{
	this.innerExtent = innerExtent;
    redraw();
}

/**
 * Returns where the outer boundary of the range is.  Values are ratios based
 * on the scale's radius, width, or height (depending on the scale's type and
 * orientation). For example, for circular scale, 0.8 means stop drawing this
 * range 80% of the radius' size away from the center.
 *
 * @return the outer extent
 */
public double getOuterExtent()
{
	return(outerExtent);
}

/**
 * Sets where the outer boundary of the range is. Values are ratios based on
 * the scale's radius, width, or height (depending on the scale's type and
 * orientation). For example for vertical linear scale, 0.9 means stop drawing
 * this range 90% of the scale's width away from the left of the scale.
 *
 * @param outerExtent the outer extent of the range
 */
public void setOuterExtent(double outerExtent)
{
	this.outerExtent = outerExtent;
    redraw();
}

/**
 * Returns the scale value (not angle or position) at which to begin the range.
 *
 * @return the range's start value
 */
public double getStartValue()
{
	return(startValue);
}

/**
 * Sets the scale value (not angle or position) at which to begin the range.
 *
 * @param startValue the start value for the range
 */
public void setStartValue(double startValue)
{
	this.startValue = startValue;
    redraw();
}

/**
 * Returns the scale value (not angle or position) at which to stop the range.
 *
 * @return the range's stop value
 */
public double getStopValue()
{
	return(stopValue);
}

/**
 * Sets the scale value (not angle or position) at which to stop the range.
 *
 * @param stopValue the stop value for the range
 */
public void setStopValue(double stopValue)
{
	this.stopValue = stopValue;
    redraw();
}

/**
 * Gets the name of this range. This is the name which will appear in the
 * legend. If no name is supplied, a name of the form "Range &lt;x&gt;" will be
 * automatically generated.
 *
 * @return the range's name
 */
public String getRangeName()
{
	if (rangeName == null) {
		setRangeName("Range " + rangeCounter++);
	}
	return(rangeName);
}

/**
 * Sets the range's name. This is the name which will appear in the legend.
 *
 * @param rangeName the name to appear in the legend
 */
public void setRangeName(String rangeName)
{
	this.rangeName = rangeName;
    redraw();
}

/**
 * Gets the <code>Image</code> to be used to mark the range.
 *
 * @return the image that is used to mark the range
 */
public Image getImage()
{
	return(image);
}

/**
 * Sets the <code>Image</code> to be used to mark the range.
 *
 * @param image the <code>Image</code> to be used to mark the range
 */
public void setImage(Image image)
{
	this.image = image;
    redraw();
}

/**
 * Scales the image in relation to the range dimensions.
 *
 * @return <code>true</code> if the image is to be scaled to the size of the
 *     bounding box of the range; otherwise, center the image based on its
 *     length and width
 */
public boolean getScaleImage()
{
	return(scaleImage);
}

/**
 * Determines whether to scale the image according to the range dimensions.
 *
 * @param scaleImage <code>true</code> if the image is to be scaled to the size
 *     of the bounding box of the range; otherwise, center the image based on
 *     its length and width
 */
public void setScaleImage(boolean scaleImage)
{
	this.scaleImage = scaleImage;
    redraw();
}

/**
 * Does nothing, overridden in <code>JCLinearRange</code> or
 * <code>JCCircularRange</code>.
 */
public void redraw() {}

}
