/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCAbstractScale.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import javax.swing.*;
import java.awt.*;
import java.util.Vector;

/**
 * A graduated scale. A scale has a minimum value, a maximum value, and a
 * direction. It also is associated with other objects like <code>JCTick</code>,
 * <code>JCIndicator</code>, <code>JCNeedle</code>, and <code>JCRange</code>.
 * These associated objects use the scale to get information that
 * they need to render themselves.
 */
public abstract class JCAbstractScale extends JComponent implements JCScale {

static final long serialVersionUID = -1015210021570456430L;

/** The parent gauge. */
protected JCGauge gauge;

/**
 * A <code>JCCircularScale.Direction</code> enum, <code>FORWARD</code> or
 * <code>BACKWARD</code>, giving the direction in which scale values increase.
 * <br>The default value is <code>FORWARD</code>.
 */
protected Direction direction = Direction.FORWARD;

/** The maximum value this scale can have. */
protected double min = 0;

/** The minumum value this scale can have. */
protected double max = 100;

/**
 * Paints a scale background, covering this ratio from the origin for circular
 * scales and the top or left for linear scales.
 * See <A HREF="../../../../../../elements/jcelements-4.html#aNoteOnExtents">A
 * note on the zoomFactor property</A> in the manual.
 * The default is <code>1.0</code>.
 */
protected double zoomFactor	= 1.0;

/** A list of tick objects for this scale. */
protected Vector ticks = new Vector();

/** The list of ranges for this scale. */
protected Vector ranges = new Vector();

/** An image may be used for this scale. */
protected Image image;

/**
 * Boolean that determines if the image should be matched when
 * the scale is resized. The default is <code>true</code>.
 */
protected boolean scaleImage = true;

/**
 * Creates a graduated scale.
 *
 * @param gauge the parent gauge
 */
public JCAbstractScale(JCGauge gauge)
{
	setLayout(new java.awt.GridLayout(1,1));

    this.gauge = gauge;
}

/**
 * Creates a graduated scale.
 *
 * @param gauge the parent gauge
 * @param direction the scale's direction
 * @param min the scale's minimum value
 * @param max the scale's maximum value
 * @param foreground the scale's foreground color
 */
public JCAbstractScale(JCGauge     gauge,
                       Direction 	direction,
                       double		min,
                       double		max,
                       Color		foreground)
{
	setLayout(new java.awt.GridLayout(1,1));

    // Set the properties
    this.gauge      = gauge;
    this.direction	= direction;
    this.min        = min;
    this.max        = max;

	setForeground(foreground);
}

/**
 * Returns the gauge associated with this scale.
 *
 * @return the gauge associated with this scale
 */
public JCGauge getGauge()
{
	return(gauge);
}

/**
 * Returns the <code>JCCircularScale.Direction</code> for this scale,
 * <code>FORWARD</code> or <code>BACKWARD</code>, giving the direction in
 * which scale values increase.
 *
 * @return the <code>JCCircularScale.Direction</code> for this scale
 */
public Direction getDirection()
{
	return(this.direction);
}

/**
 * Sets the <code>JCAbstractScale.Direction</code> for this scale,
 * <code>FORWARD</code> or <code>BACKWARD</code>, giving the direction in
 * which scale values increase.
 *
 * @param direction the <code>JCCircularScale.Direction</code>
 */
public void setDirection(Direction direction)
{
	this.direction = direction;
    redraw();
}

/**
 * Returns the maximum value for this scale.
 *
 * @return the maximum value for this scale
 */
public double getMax()
{
	return(this.max);
}

/**
 * Sets the maximum value for this scale.
 *
 * @param max the maxiumum value for this scale
 */
public void setMax(double max)
{
	this.max = max;
    redraw();
}

/**
 * Returns the minimum value for this scale.
 *
 * @return the minimum value for this scale
 */
public double getMin()
{
	return(this.min);
}

/**
 * Sets the minimum value for this scale.
 *
 * @param min the minimum value for this scale
 */
public void setMin(double min)
{
	this.min = min;
    redraw();
}

/**
 * Returns the zoom factor for this scale.
 * @deprecated As of JClass version 4.5.1
 * replaced by <code>getZoomFactor()</code>.
 *
 * @return the <code>zoomFactor</code> of this scale
 */
public double getExtent()
{
	return(getZoomFactor());
}

/**
 * Sets the zoom factor for this scale.
 * @deprecated As of JClass version 4.5.1
 * replaced by <code>setZoomFactor()</code>.
 *
 * @param zoomFactor the new <code>zoomFactor</code>
 */

public void setExtent(double zoomFactor)
{
    setZoomFactor(zoomFactor);
}

/**
 * Returns the zoom factor for this scale.  The scale is normally
 * sized so it fits exactly within the component.  This may not always be what
 * the user wants as ticks, labels, and other components may need to be drawn
 * outside the scale.  Setting <code>zoomFactor</code> allows the scale to be
 * smaller so things outside the scale can be seen.
 *
 * @return the <code>zoomFactor</code> of this scale
 */
public double getZoomFactor()
{
	return(zoomFactor);
}

/**
 * Sets the zoom factor for this scale.  The scale is normally
 * sized so it fits exactly within the component. This may not always be what
 * the user wants as ticks, labels, and other components may need to be
 * drawn outside the scale.
 *
 * @param zoomFactor the new <code>zoomFactor</code>
 */
public void setZoomFactor(double zoomFactor)
{
	if (zoomFactor > 1.0) {
    	zoomFactor = 1.0;
    }
    else if (zoomFactor < 0) {
    	zoomFactor = 0;
    }
	this.zoomFactor = zoomFactor;
    redraw();
}

/**
 * Returns the <code>Vector</code> of tick objects for this scale.
 *
 * @return a <code>Vector</code> of ticks
 */
public Vector getTicks()
{
	return(ticks);
}

/**
 * Adds a tick object to a scale.
 *
 * @param tick the tick object to be added
 */
public void addTick(JCTick tick)
{
	addTick(tick, 0);
}

/**
 * Adds a tick object to a scale.
 *
 * @param tick the <code>JCTick</code> to be added to the scale
 * @param index the <i>Z</i>-order index
 */
public void addTick(JCTick tick, int index)
{
    // add it to the list
	ticks.addElement(tick);

	// add it to the display
	add((Component)tick, index);
}

/**
 * Removes a tick from a scale.
 *
 * @param tick the <code>JCCircularTick</code> to be removed
 */
public void removeTick(JCTick tick)
{
	if (ticks.contains(tick)) {
    	ticks.removeElement(tick);
    	remove((Component)tick);
    }
}

/**
 * Returns the <code>Vector</code> of tick objects for this scale.
 *
 * @return a <code>Vector</code> of ranges
 */
public Vector getRanges()
{
	return(ranges);
}

/**
 * Adds a range object to a scale.
 *
 * @param range the range object to be added
 */
public void addRange(JCRange range)
{
	addRange(range, 0);
}

/**
 * Adds a range object to a scale.
 *
 * @param range the <code>JCRange</code> to be added to the scale
 * @param index the <i>Z</i>-order index
 */
public void addRange(JCRange range, int index)
{
    // add it to the list
	ranges.addElement(range);

	// add it to the display
	add((Component)range, index);
}

/**
 * Removes a range from a scale.
 *
 * @param range the <code>JCRange</code> to be removed
 */
public void removeRange(JCRange range)
{
	if (ranges.contains(range)) {
    	ranges.removeElement(range);
    	remove((Component)range);
    }
}

/**
 * Returns <code>true</code> if the value is within the scale's minimum and
 * maximum.
 *
 * @param value the value to be tested
 * @return <code>true</code> if the value is within the scale's minimum and
 *     maximum
 */
public boolean inBounds(double value)
{
	return(value >= getMin() && value <= getMax());
}

/**
 * Gets the <code>Image</code> used to mark the scale.
 *
 * @return the <code>Image</code> used to mark the scale
 */
public Image getImage()
{
	return(image);
}

/**
 * Sets the <code>Image</code> used to mark the scale.
 *
 * @param image the <code>Image</code> used to mark the scale
 */
public void setImage(Image image)
{
	this.image = image;
    redraw();
}

/**
 * Scales the image in relation to the scale dimensions.
 *
 * @return <code>true</code> if the image is to be matched to the size of the
 *     bounding box of the scale; otherwise, center the image based on its
 *     length and width
 */
public boolean getScaleImage()
{
	return(scaleImage);
}

/**
 * Determines whether to adapt the image according to the scale dimensions.
 *
 * @param scaleImage <code>true</code> if the image is to be matched to the
 *     size of the bounding box of the scale; otherwise, center the image based
 *     on its length and width
 */
public void setScaleImage(boolean scaleImage)
{
	this.scaleImage = scaleImage;
    redraw();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw() {
}

/**
 * Given a screen position in pixels, returns the closest scale value.
 * This method is abstract and must be overridden.
 *
 * @param p the point representing the <i>X</i>, <i>Y</i> coordinates of the
 *     click
 * @return the value closest to the point
 */
public abstract double pick(Point p);

/////////////////////// inner class


/**
 * Enum's for the direction parameter. Valid values are:
 * <ul>
 * <li><code>BACKWARD</code></li>
 * <li><code>FORWARD</code></li>
 * <li><code>CLOCKWISE</code></li>
 * <li><code>COUNTERCLOCKWISE</code></li>
 * </ul>
 *
 * <b>Note:</b> <code>CLOCKWISE</code> and <code>COUNTERCLOCKWISE</code>
 * are old values that correspond to <code>BACKWARD</code> and
 * <code>FORWARD</code>, respectively. These values can still be used, but only
 * make sense for circular scales.
 */
public static class Direction implements java.io.Serializable {
    /**
	 * Scale values increase in a backward (or clockwise) direction.
     * For example, a circular scale whose start angle is 0&#176;
     * and whose stop angle is 180&#176;, and whose max
     * and min values are 0 and 100, places its zeroth tick mark
     * at 180&#176; and its 100th tick mark at 0&#176;.
     *
     * @see com.klg.jclass.swing.gauge.JCTick
     */

    public static final Direction BACKWARD        = new Direction("backward");
    public static final Direction CLOCKWISE       = new Direction("clockwise");

    /**
	 * Scale values increase in a forward (or counterclockwise) direction.
     * For example, a circular scale whose start angle is 0&#176;
     * and whose stop angle is 180&#176;, and whose max
     * and min values are 0 and 100 places its zeroth tick mark
     * at 0&#176; and its 100th tick mark at 180&#176;.
     *
     * @see com.klg.jclass.swing.gauge.JCTick
     */
    public static final Direction FORWARD          = new Direction("forward");
    public static final Direction COUNTERCLOCKWISE = new Direction("counterclockwise");

    String string;

    /**
	 * Private so no bogus directions can be created.
     * @param string
     */
    private Direction(String string)
    {
        this.string = string;
    }

    /**
     * Compares objects based on internal String.
     * @return
     * @param direction
     */
    public boolean equals(Direction direction)
    {
        boolean is_equal = direction.string.equals(this.string);

        if (is_equal == false) {
            if (direction.string.equals(Direction.BACKWARD.string) &&
                this.string.equals(Direction.CLOCKWISE.string))
            {
                is_equal = true;
            }
            else if (direction.string.equals(Direction.CLOCKWISE.string) &&
                     this.string.equals(Direction.BACKWARD.string))
            {
                is_equal = true;
            }
            else if (direction.string.equals(Direction.FORWARD.string) &&
                     this.string.equals(Direction.COUNTERCLOCKWISE.string))
            {
                is_equal = true;
            }
            else if (direction.string.equals(Direction.COUNTERCLOCKWISE.string) &&
                     this.string.equals(Direction.FORWARD.string))
            {
                is_equal = true;
            }
        }
        return(is_equal);
    }

} // Direction

}
