/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCAbstractTick.java,v $ $Revision: 1.2 $
//          $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import com.klg.jclass.util.JCNumberUtil;

import javax.swing.*;
import java.awt.*;

/**
 * A graduated scale marker object. It comprises a set of uniformly spaced
 * tick marks.
 *
 * @see com.klg.jclass.swing.gauge.JCAbstractTick
 */
public abstract class JCAbstractTick extends JComponent
                                     implements JCTick, java.io.Serializable
{

static final long serialVersionUID = 6997480289846291634L;

/**
 * The scale associated with this tick object.
 */
protected JCScale 	        scale;

/**
 * The <code>automatic</code> boolean determines whether tick increments are
 * calculated automatically or set by the <code>incrementValue</code> property.
 * If <code>true</code>, tick increments are calculated by the gauge.
 * If <code>false</code>, the <code>incrementValue</code> property is used to
 * set the tick spacing.
 */
protected boolean			automatic			= true;

/**
 * The <code>startValue</code> is the scale value at which to begin this set of
 * tick marks.
 */
protected double			startValue;

/**
 * The <code>stopValue</code> is the scale value at which this set of tick
 * marks ends.
 */
protected double			stopValue;

/**
 * If not automatic, the scale values at which tick marks are to be placed.
 */
protected double			incrementValue		= 10;

/**
 * If <code>true</code>, use gauge-determined precision. If <code>false</code>,
 * use the precision property to determine the number of decimal places on tick
 * labels.
 */
protected boolean			precisionUseDefault	= true;

/**
 * If <code>precisionUseDefault</code> is <code>false</code>, use this
 * precision.  If it is <code>true</code>, the value of this property will be
 * automatically calculated and the calculated value will overwrite any
 * previous value this property might have.
 */
protected int				precision			= 0;

/**
 * The width of a tick in pixels.
 */
protected double			tickWidth			= 2;

/**
 * Asks: "Is the tick reversed?".
 */
protected boolean           reversed            = false;

/**
 * If <code>true</code>, draw tick labels; otherwise, no tick labels are drawn.
 */
protected boolean			drawLabels			= true;

/**
 * If <code>true</code>, draw tick marks; otherwise, no tick marks are drawn.
 */
protected boolean			drawTicks			= true;

/**
 * The percentage, expressed as a decimal fraction of the scale's radius,
 * at which to place tick labels.
 *
 */
protected double			labelExtent			= 0.8;

/**
 * The percentage of the scale's radius (as a decimal) at
 * at which to begin drawing ticks radially outward.
 */
protected double			innerExtent			= 0.85;

/**
 * The percentage of the scale's radius or width/height (as a decimal) at
 * which to stop drawing the ticks.
 */
protected double			outerExtent			= 1.0;

/**
 * The color in which to draw ticks.
 */
protected Color				tickColor			= Color.black;

/**
 * The <code>JCTickStyle</code> controlling the shape of the tick.
 */
protected JCTickStyle		tickStyle			= JCTickStyle.LINE;

/**
 * The color of the font labeling ticks.
 */
protected Color				fontColor			= Color.black;

/**
 * A custom label generator.
 */
protected JCLabelGenerator	labelGenerator;

/**
 * Creates ticks. A scale, <code>startValue</code>, and <code>stopValue</code>
 * need to be supplied.  The other properties take on their default values.
 *
 * @param scale the associated circular scale
 * @param startValue where the tick marks are to begin
 * @param stopValue where the tick marks are to end
 */
public JCAbstractTick(JCScale scale, double startValue, double stopValue)
{
    this.scale      = scale;
    this.startValue = startValue;
    this.stopValue	= stopValue;
}

/**
 * Creates a tick object and sets its properties.
 *
 * @param scale the associated circular or linear scale
 * @param automatic whether to automatically calculate the tick start, stop,
 *      and increment values from the scale's min and max, or to use the
 *      provided values
 * @param startValue where the tick marks are to begin
 * @param stopValue where the tick marks are to end
 * @param incrementValue the tick interval
 * @param precisionUseDefault sets whether tick labels use a default
 *      precision, or use <code>precision</code>
 * @param precision The format of the tick label; positive values denote the
 *      number of places after the decimal point (for example, 3 means multiples
 *      of 0.001), and negative values indicate the positive number of zeros to
 *      be used before the decimal place (for example, -3 means numbering will
 *      be in multiples of 1000). Overridden if <code>precisionUseDefault</code>
 *      is set to <code>true</code>.
 * @param tickWidth the width of a tick mark
 * @param drawLabels whether tick labels are to be drawn
 * @param drawTicks whether tick marks are to be drawn
 * @param labelExtent the radial or linear distance to the center of a tick
 *      label, expressed as a decimal fraction of the scale's radius or
 *      width/height
 * @param innerExtent the place where the tick mark starts, expressed as a
 *      decimal fraction of the scale's radius or width/height
 * @param outerExtent the place where the tick mark ends, expressed as a
 *      decimal fraction of the scale's radius or width/height
 * @param tickColor the color of the tick mark
 * @param tickStyle a <code>JCTickStyle</code> enum
 * @param font the font for tick labels
 * @param fontColor the tick label's color
 */
public JCAbstractTick(JCScale           scale,
                      boolean			automatic,
                      double			startValue,
                      double			stopValue,
                      double			incrementValue,
                      boolean			precisionUseDefault,
                      int				precision,
                      double			tickWidth,
                      boolean			drawLabels,
                      boolean			drawTicks,
                      double			labelExtent,
                      double			innerExtent,
                      double			outerExtent,
                      Color				tickColor,
                      JCTickStyle		tickStyle,
                      Font				font,
                      Color				fontColor)
{
    this.scale                  = scale;
    this.automatic				= automatic;
    this.startValue				= startValue;
    this.stopValue				= stopValue;
    this.incrementValue			= incrementValue;
    this.precisionUseDefault	= precisionUseDefault;
    this.precision				= precision;
    this.tickWidth				= tickWidth;
    this.drawLabels				= drawLabels;
    this.drawTicks				= drawTicks;
    this.labelExtent			= labelExtent;
    this.innerExtent			= innerExtent;
    this.outerExtent			= outerExtent;
    this.tickColor				= tickColor;
    this.tickStyle				= tickStyle;
    this.fontColor				= fontColor;
    this.setFont(font);
}

/**
 * Controls whether labels should be drawn according to
 * the <code>drawLabels</code> boolean.
 *
 * @param g
 */
public void paintChildren(Graphics g)
{
	if (drawLabels) {
		super.paintChildren(g);
    }
}

/**
 * If <code>automatic</code>, this method is called to determine and
 * set the start/stop and increment values for laying out ticks.
 * The goal is to have as many ticks/labels as will fit without
 * the ticks or labels overlapping.
 *
 * @param g the graphics used to get font metric information
 * @param width width in pixels to lay out ticks
 */
protected void calcTicks(Graphics g, int width)
{
	// Get the scale values
	double min = scale.getMin();
	double max = scale.getMax();

    // Calculate the default precision (if necessary)
	calcPrecision(min, max);

	// The axis bounds must be outside the actual bounds.
	min = JCNumberUtil.precFloor(precision, min);
	max = JCNumberUtil.precCeil(precision, max);

    int tprec;
	if (max <= min) {
		// Move the max. and min. values based on
		// the precision.	Keep adjusting the
		// increment until we get some movement
		// (floating point error correction).

		tprec = precision;
		do {
			max = (min + Math.pow(10.0, (double) -tprec));
			tprec--;
		}
        while (max <= min);
	}

	double min_real = min;
	double max_real = max;

	int retry_count = 0;
	double tik = 0.0;

    // set a default font to avoid a NullPointerException
    Font font = getFont();
    if(font == null){
        font = new Font("Dialog", Font.PLAIN, 10);
    }
	FontMetrics fontMetrics = g.getFontMetrics(font);

	while (true) {
		//	The axis bounds must be outside the actual bounds.
		min = JCNumberUtil.precFloor(precision, min);
		max = JCNumberUtil.precCeil(precision, max);

		double min_used = min;
		double max_used = max;

	    double error = (0.5 * Math.pow(10.0, (double)(- precision - 1)));
		double size = max - min;

		// Determine the tick increment to use

		// Determine the widest label
		String sMax = JCNumberUtil.format(max, precision);
		int lWidth = fontMetrics.stringWidth(sMax);

		String sMin = JCNumberUtil.format(min, precision);
		int w = fontMetrics.stringWidth(sMin);
		if (w > lWidth) {
			lWidth = w;
		}

		// The goal here is to draw as many ticks/labels as possible
		// without overlapping of ticks or labels. If both are drawn,
		// choose the widest.
        int anno_size;
		if (drawTicks && drawLabels) {
			anno_size = Math.max((int) tickWidth, lWidth);
		}
		else if (drawTicks) {
			anno_size = (int) tickWidth;
		}
		else if (drawLabels) {
			anno_size = lWidth;
		}
		else {
			anno_size = 0;
		}

        int num_tiks = (anno_size == 0) ? 0 : width / anno_size;
		if (num_tiks <= 0) {
			tik = 0.0;
		}
		else {
			double tmp_val = size / (double) num_tiks;
			tprec = JCNumberUtil.nicePrecision(max - min);
			tprec = Math.min(tprec, precision);
			tik   = niceNum(tmp_val, -tprec, false);
		}

		// Adjust the axis, if necessary
		if (tik != 0) {
			min = tik * Math.floor(min/tik);
			max = tik * Math.ceil(max/tik);
		}
		else {
			if (Math.abs(max - min) < error) {
				max = min + 1;
			}
		}

		// Try this once ... if either of the axis min-max have
		// been moved, it might be too crowded to add these
		// extra labels, so try to determine these values again.
		if (Math.abs(min - min_used) > error ||
			Math.abs(max - max_used) > error)
        {
			int pass = 1;
			if ((max - tik) >= max_real) {
				max -= tik;
			}
			else {
				pass = 0;
			}

			if ((min + tik) <= min_real) {
				min += tik;
			}
			else {
				pass = 0;
			}

			if (pass == 0 && retry_count < 2) {
				retry_count++;
			}
			else {
				break;
			}
		}
        else {
        	break;
        }
	}

	startValue     = min;
	stopValue      = max;
	incrementValue = tik;
}

/**
 * Determines and sets a "nice" precision to use if
 * <code>useDefaultPrecision</code> is set to <code>true</code>.
 *
 * @param min
 * @param max
 */
protected void calcPrecision(double min, double max)
{
	if (precisionUseDefault) {
		if (max <= min) {
			// range is zero, so use value itself
			precision = JCNumberUtil.nicePrecision(Math.abs(max));
		}
		else {
			precision = JCNumberUtil.nicePrecision(max - min);
		}
	}

//	tick->precision = GaugeUtil.clamp(tick->precision,
//							          -JCAbstractTick.MAX_PRECISION,
//                                    JCAbstractTick.MAX_PRECISION);
}

/**
 * Given <i>x</i>, nudges it towards a "nice" number.  Here,
 * "nice" means that the least significant digit is a 0, 1,
 * 2, or 5.  The desired exponent to use is provided,
 * so that there is a consistency with <code>default_precision</code>.
 * @param x
 * @param exp
 * @param round
 * @return
 */
public static double niceNum(double x, int exp, boolean round)
{
    if (x == 0) {
        return(0.0);
    }
    else if (x < 0.0) {
        x = -x;
    }

    double nf = 1.0;
    double f  = x / Math.pow(10.0, (double) exp);
    if (round) {
        if (f < 1.5) {
            nf = 1.0;
        }
        else if (f < 3.0) {
            nf = 2.0;
        }
        else if (f < 7.0) {
            nf = 5.0;
        }
        else    {
            nf = 10.0;
        }
    }
    else    {
        if (f <= 1.0) {
            nf = 1.0;
        }
        else if (f <= 2.0) {
            nf = 2.0;
        }
        else if (f <= 5.0) {
            nf = 5.0;
        }
        else {
            nf = 10.0;
        }
    }
    return(nf * Math.pow(10.0, (double) exp));
}

//////////////////////////////////////////////////////
////////////////// setters and getters
//////////////////////////////////////////////////////

/**
 * Returns the scale associated with this tick.
 *
 * @return the scale for the <code>JCAbstractTick</code> component
 */
public JCScale getScale()
{
	return(scale);
}

/**
 * Returns the label generator associated with this tick.
 * If non-null, it will be called, and expected to
 * return a <code>JComponent</code> which will be the label for a tick mark.
 *
 * @return the <code>labelGenerator</code>
 */
public JCLabelGenerator getLabelGenerator()
{
	return(labelGenerator);
}

/**
 * Sets the label generator associated with this tick.
 * If non-null, it will be called, and expected to
 * return a <code>JComponent</code> which will be the label for a tick mark.
 * See the code in examples.elements.StopWatchExample for an example.
 *
 * @param labelGenerator the <code>labelGenerator</code> which implements the
 *     <code>JCLabelGenerator</code> interface
 */
public void setLabelGenerator(JCLabelGenerator labelGenerator)
{
	this.labelGenerator = labelGenerator;
}

/**
 * Returns the <code>automatic</code> Boolean for this tick object.
 *
 * @return the automatic setting for this tick, where <code>true</code> means
 *     ticks and increments are dynamically determined by the gauge; otherwise,
 *     <code>startValue</code>, <code>stopValue</code>, and
 *     <code>incrementValue</code> are used to lay out ticks
 */
public boolean getAutomatic()
{
	return(automatic);
}

/**
 * Sets the <code>automatic</code> Boolean for this tick object.
 *
 * @param automatic the automatic setting for this tick, where <code>true</code>
 *     means ticks and increments are dynamically determined by the gauge;
 *     otherwise, <code>startValue</code>, <code>stopValue</code>, and
 *     <code>incrementValue</code> are used to lay out ticks.
 */
public void setAutomatic(boolean automatic)
{
	this.automatic = automatic;
}

/**
 * In non-automatic mode, use this value as the start value.  In automatic
 * mode, the calculated start value will overwrite this value.
 *
 * @return the <code>startValue</code>
 */
public double getStartValue()
{
	return(startValue);
}

/**
 * In non-automatic mode, use this value as the start value.  In automatic
 * mode, the calculated start value will overwrite this value.
 *
 * @param startValue the start value
 */
public void setStartValue(double startValue)
{
	this.startValue = startValue;
    redraw(true);
}

/**
 * In non-automatic mode, use this value as the stop value.  In automatic
 * mode, the calculated stop value will overwrite this value.
 *
 * @return the <code>stopValue</code>
 */
public double getStopValue()
{
	return(stopValue);
}

/**
 * In non-automatic mode, use this value as the stop value.  In automatic
 * mode, the calculated stop value will overwrite this value.
 *
 * @param stopValue the stop value
 */
public void setStopValue(double stopValue)
{
	this.stopValue = stopValue;
    redraw(true);
}

/**
 * In non-automatic mode, use this value as the increment value. That
 * is, it places ticks at every <code>incrementValue</code> along the scale.
 * In automatic mode, the calculated increment value will overwrite this value.
 *
 * @return the spacing between adjacent tick marks
 */
public double getIncrementValue()
{
	return(incrementValue);
}

/**
 * In non-automatic mode, use this value as the increment value. That
 * is, it places ticks at every <code>incrementValue</code> along the scale.
 * In automatic mode, the calculated increment value will overwrite this value.
 *
 * @param incrementValue the spacing between adjacent tick marks
 */
public void setIncrementValue(double incrementValue)
{
	this.incrementValue = incrementValue;
    redraw(true);
}

/**
 * If <code>true</code>, uses gauge-determined precision for tick labels;
 * otherwise, uses the current value of <code>precision</code> property to
 * label tick marks.
 *
 * @return a Boolean indicating whether or not a label's precision is
 *      calculated by the gauge or gotten from the <code>precision</code>
 *      property
 */
public boolean getPrecisionUseDefault()
{
	return(precisionUseDefault);
}

/**
 * If <code>true</code>, uses the gauge-determined precision for tick labels;
 * otherwise, uses the current value of <code>precision</code> property to
 * label tick marks.
 *
 * @param precisionUseDefault
 */
public void setPrecisionUseDefault(boolean precisionUseDefault)
{
	this.precisionUseDefault = precisionUseDefault;
    redraw();
}

/**
 * Returns the format for numbers in a tick label: positive values denote the
 * number of places after the decimal point (for example, 3 means multiples
 * of 0.001), negative values indicate the positive number of zeros are
 * used before the decimal place (for example, -3 means numbering will be in
 * multiples of 1000). Overridden if <code>precisionUseDefault</code> is set
 * to <code>true</code>.
 *
 * @return the precision code for formatting numbers in a tick label
 */
public int getPrecision()
{
	return(precision);
}

/**
 * Sets the format for numbers in a tick label: positive values denote the
 * number of places after the decimal point (for example, 3 means multiples
 * of 0.001), negative values indicate the positive number of zeros are
 * used before the decimal place (for example, -3 means numbering will be in
 * multiples of 1000). Overridden if <code>precisionUseDefault</code> is set
 * to <code>true</code>.
 *
 * @param precision an integer
 */
public void setPrecision(int precision)
{
	this.precision = precision;
    redraw();
}

/**
 * Returns the width in pixels of the tick mark.
 *
 * @return the width of the tick to be drawn
 */
public double getTickWidth()
{
	return(tickWidth);
}

/**
 * Sets the width in pixels of the tick mark.
 *
 * @param tickWidth the width in pixels of the tick to be drawn
 */
public void setTickWidth(double tickWidth)
{
	this.tickWidth = tickWidth;
    redraw();
}

/**
 * Asks: "Is the tick reversed?".
 *
 * @return whether the tick is reversed or not
 */
public boolean isReversed()
{
	return(reversed);
}

/**
 * Sets the reversed property of the tick.  If the tick is
 * reversed it draws in the opposite direction (for example, for linear
 * scales it points left or up instead of right or down, for circular
 * scales it points towards the center instead of away from the center).
 *
 * @param reversed is the tick's direction reversed?
 */
public void setReversed(boolean reversed)
{
	this.reversed = reversed;
    redraw();
}

/**
 * Returns the Boolean controlling the drawing of labels. <code>True</code>
 * means draw the labels, <code>false</code> means hide the labels.
 *
 * @return the boolean controlling the drawing of labels
 */
public boolean getDrawLabels()
{
	return(drawLabels);
}

/**
 * Sets the Boolean controlling the drawing of labels. <code>True</code>
 * means draw the labels, <code>false</code> means hide the labels.
 *
 * @param drawLabels the boolean controlling the drawing of labels
 */
public void setDrawLabels(boolean drawLabels)
{
	this.drawLabels = drawLabels;
    redraw();
}

/**
 * Returns the Boolean controlling the drawing of ticks. <code>True</code>
 * means draw the ticks, <code>false</code> means hide the ticks.
 *
 * @return the boolean controlling the drawing of ticks
 */
public boolean getDrawTicks()
{
	return(drawTicks);
}

/**
 * Sets the Boolean controlling the drawing of ticks. <code>True</code>
 * means draw the ticks, <code>false</code> means hide the ticks.
 *
 * @param drawTicks the boolean controlling the drawing of ticks
 */
public void setDrawTicks(boolean drawTicks)
{
	this.drawTicks = drawTicks;
    redraw();
}

/**
 * Returns the position, expressed as a decimal fraction of the scale's size,
 * where tick labels are placed (the size is the scale's radius or its
 * width/height, depending on the linear scale's orientation).
 *
 * @return the fraction representing the position at which labels are placed
 */
public double getLabelExtent()
{
	return(labelExtent);
}
/**
 * Sets the percentage, expressed as a decimal fraction of the scale's size,
 * at which to place the center of the tick label objects.
 *
 * @param labelExtent
 */
public void setLabelExtent(double labelExtent)
{
	this.labelExtent = labelExtent;
    redraw(true);
}

/**
 * Returns the percentage of the scale's radius (as a decimal)
 * at which to begin drawing ticks radially outward.
 *
 * @return the inner extent of the tick mark
 */
public double getInnerExtent()
{
	return(innerExtent);
}

/**
 * Sets the percentage of the scale's radius (as a decimal) at
 * at which to begin drawing ticks radially outward.
 *
 * @param innerExtent the inner extent of the tick mark
 */
public void setInnerExtent(double innerExtent)
{
	this.innerExtent = innerExtent;
    redraw();
}

/**
 * Returns the percentage of the scale's radius (as a decimal) at
 * which to stop drawing the ticks.
 *
 * @return the <code>outerExtent</code> of the tick mark
 */
public double getOuterExtent()
{
	return(outerExtent);
}
/**
 * Sets the percentage of the scale's radius (as a decimal) at
 * which to stop drawing the ticks.
 *
 * @param outerExtent the <code>outerExtent</code> of the tick mark
 */
public void setOuterExtent(double outerExtent)
{
	this.outerExtent = outerExtent;
    redraw();
}

/**
 * Returns the color in which to draw ticks.
 *
 * @return the color of the tick mark
 */
public Color getTickColor()
{
	return(tickColor);
}

/**
 * Sets the color in which to draw ticks.
 *
 * @param tickColor the color of the tick mark
 */
public void setTickColor(Color tickColor)
{
	this.tickColor = tickColor;
    redraw();
}

/**
 * Returns the <code>JCTickStyle</code> controlling the shape of the tick.
 *
 * @return the tick style
 * @see com.klg.jclass.swing.gauge.JCTickStyle
 */
public JCTickStyle getTickStyle()
{
	return(tickStyle);
}

/**
 * Sets the <code>JCTickStyle</code> controlling the shape of the tick.
 * Users may define their own tick style, if desired.
 *
 * @param tickStyle one of the built-in tick styles, or a user-supplied style
 * @see com.klg.jclass.swing.gauge.JCTickStyle
 */
public void setTickStyle(JCTickStyle tickStyle)
{
	this.tickStyle = tickStyle;
    redraw();
}

/**
 * Returns the color of the font labeling ticks.
 *
 * @return the font color
 */
public Color getFontColor()
{
	return(fontColor);
}

/**
 * Sets the color of the font labeling ticks.
 *
 * @param fontColor the font color
 */
public void setFontColor(Color fontColor)
{
	this.fontColor = fontColor;
    redraw(true);
}

/**
 * Does nothing. Overridden in <code>JCCircularTick</code> or
 * <code>JCLinearTick</code>.
 * @param redrawLabels
 */
public void redraw(boolean redrawLabels) {}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
    redraw(false);
}

/**
 * Does nothing.  Overridden in <code>JCCircularTick</code> or
 * <code>JCLinearTick</code>.
 */
public void redrawLabels() {}

}
