/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCCenter.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import java.awt.Color;
import java.awt.Image;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Dimension;


import java.awt.geom.*;
import java.awt.Shape;
import java.awt.Polygon;


import javax.swing.JComponent;

/**
 * A center object is used to mark the position of the center of a circular
 * scale in a <code>JCCircularScale</code> widget. It can be a disk or an image.
 *
 */
public class JCCenter extends JComponent implements java.io.Serializable
{

static final long serialVersionUID = -1082954384206749087L;

/**
 * The associated circular scale.
 */
protected JCCircularScale scale;

/**
 * An <code>Image</code> may be used to mark the center of the gauge.
 */
protected Image image;

/**
 * The scale's foreground color. The default is black.
 */
protected Color foreground = Color.black;

/**
 * Boolean that determines whether the image should be scaled when the gauge
 * is resized. The default is <code>true</code>.
 */
protected boolean scaleImage = true;

/**
 * The size of the disk, expressed as a ratio
 * based on the radius of the associated scale.
 */
protected double radius = 0.1;

/**
 * Creates a black disk at the center of the scale whose default radius is 10%
 * of the scale's radius.
 *
 * @param scale the associated circular scale
 */
public JCCenter(JCCircularScale scale)
{
	this.scale = scale;
    setForeground(foreground);
}

/**
 * Creates a black disk at the center of the scale whose radius and
 * foreground color are parameters.
 *
 * @param scale the associated circular scale
 * @param foreground the Color of the foreground
 * @param radius The radius of the center disc as a fractional proportion of
 *      the associated scale's radius, from 0.0-1.0. For example, 0.5 means
 *      the center will have a radius half the size of the scale's radius.
 */
public JCCenter(JCCircularScale 	scale,
                Color		        foreground,
                double		        radius)
{
	this.scale 		= scale;
    this.radius		= radius;

    setForeground(foreground);
}

/**
 * Creates a center using the provided image with no scaling. The size of the
 * center will be based on the size of the image.
 *
 * @param scale the associated circular scale
 * @param foreground the Color of the foreground
 * @param image the image used as the center
 */
public JCCenter(JCCircularScale     scale,
                Color		    	foreground,
                Image		        image)
{
   	this.scale 		= scale;
    this.image 		= image;

	setForeground(foreground);
}

/**
 * This routine actually draws the center object on the screen. If there is
 * no associated circular scale, or if the radius is 0.0, nothing is drawn.
 * Also note that the background is not drawn if an image is used.
 *
 * @param g
 */
public void paintComponent(Graphics g)
{
    // Do we need to draw anything?
    if (!isVisible()                        ||
        scale == null                       ||
        scale.getMin() >= scale.getMax()    ||
        (radius <= 0.0 && image == null))
    {
        return;
    }

	// Draw an image if there is one
    Graphics gc = g.create();
    Rectangle bounds = getBounds();
    gc.translate(-bounds.x, -bounds.y);
    // save old clip
    Shape oldClip = g.getClip();
    Shape shape = getCenterShape();
    gc.setClip(shape);
    if (image != null) {
        if (scaleImage) {
            // draw image in the clip region
            drawImage(gc, image, bounds);
        }
        else {
            drawImage(gc, image, bounds);
        }
    }
    else {  // otherwise draw the disc
    	gc.fillArc(bounds.x, bounds.y, getWidth(), getHeight(), 0, 360);
    }
    // reset the previous clip region
    g.setClip(oldClip);
    
    gc.dispose();
}

/**
 * Overridden so that the center can scale itself proportionately with the scale.
 *
 * @return a <code>Dimension</code> object indicating the preferred size of
 *     the component
 */
public Dimension getPreferredSize()
{
	// get the diameter of this disc as a fraction of the scale
    int diameter = (int) (scale.getRadius() * radius) * 2;
	return(new Dimension(diameter, diameter));
}

/*
 * Draws an image as the center.
 *
 * @param g the Graphics on which to draw
 */
protected void drawImage(Graphics g, Image image, Rectangle bounds)
{
	// create an image observer
    java.awt.image.ImageObserver observer = new javax.swing.JLabel();
    
    // draw the image
    if (scaleImage) {
	    g.drawImage(image, bounds.x, bounds.y, bounds.width, bounds.height, 
	                getForeground(), observer);
    }
    else {
        g.drawImage(image, 0, 0, getForeground(), observer);
    }
}

/**
 * Draws a disk at the center of the scale.
 *
 * @param g the graphics on which to draw
 */
protected void drawDisc(Graphics g)
{
    // startAngle - the beginning angle.
    int startAngle = 0;

    // arcAngle - the angular extent of the arc, relative to the start angle
    int arcAngle = 360;

    g.fillArc(0, 0, getWidth(), getHeight(), startAngle, arcAngle);
}

///////////// setters and getters

/**
 * Returns the scale associated with this center.
 *
 * @return the circular scale for the <code>JCCenter</code> component
 */
public JCCircularScale getScale()
{
	return(scale);
}

/**
 * Sets the radius of the disk. The <code>radius</code> parameter is a decimal
 * fraction of the scale's radius.
 *
 * @param radius the radius of the center as a proportion of the circular
 *     scale's radius
 */
public void setRadius(double radius)
{
    if (radius <= 0.0) {
        radius = 0.0;
    }
	this.radius = radius;
    revalidate();
}

/**
 * Gets the radius of the center. The value is a decimal fraction of the
 * scale's radius.
 * @return
 */
public double getRadius()
{
	return(radius);
}

/**
 * Gets the <code>Image</code> to be used to mark the center of the circular
 * scale.
 *
 * @return the image that is used to mark the center of the circular scale
 */
public Image getImage()
{
	return(image);
}

/**
 * Sets the <code>Image</code> to be used to mark the center of the circular
 * scale.
 *
 * @param image the <code>Image</code> to be used to mark the center of the
 *     circular scale
 */
public void setImage(Image image)
{
	this.image = image;
    redraw();
}

/**
 * Scales the image in relation to the radius.
 *
 * @return <code>true</code> if the image is to be scaled to the size of the
 *     bounding box with length and width equal to "radius"; otherwise, center
 *     the image based on its length and width
 */
public boolean getScaleImage()
{
	return(scaleImage);
}

/**
 * Determines whether to scale the image according to the radius.
 *
 * @param scaleImage <code>true</code> if the image is to be scaled to the size
 *     of the bounding box with length and width equal to "radius"; otherwise,
 *     center the image based on its length and width
 */
public void setScaleImage(boolean scaleImage)
{
	this.scaleImage = scaleImage;
    redraw();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked.   No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
	if (scale.getGauge().getRepaintEnabled()) {
    	repaint();
	}
}

/*
* Returns the shape of the center.
* @return shape the shape of center
*/
protected Shape getCenterShape()
{
    Shape shape = null;
    JCCircularGauge cgauge = scale.getCircularGauge();
	double start_angle = GaugeUtil.normalizeAngle(scale.getStartAngle());
    double stop_angle = GaugeUtil.normalizeAngle(scale.getStopAngle());

    // Compute the start and stop angle
    if (scale.getPaintCompleteBackground()) {
        start_angle = cgauge.getGaugeType().getStartAngle();
    }
    else {  
        if (start_angle >= stop_angle) {
            stop_angle += 360.0;
        }
    }
      
    // Compute the center shape
    Rectangle bounds = getBounds();
    if ((cgauge.getGaugeType() == JCCircularGauge.GaugeType.FULL_CIRCLE)||
        (cgauge.getGaugeType() == JCCircularGauge.GaugeType.TOP_HALF_CIRCLE) ||
        (cgauge.getGaugeType() == JCCircularGauge.GaugeType.LEFT_HALF_CIRCLE))
    {
        Ellipse2D.Double ellipse = new Ellipse2D.Double();
        ellipse.setFrame(bounds.x, bounds.y, (double)getWidth(),
                         (double)getHeight());
        shape = ellipse; 
    }
    else { 
	    // Make a polar coordinates polygon of the center contour
        Polygon op = new Polygon();
        double df = 3;
        double centerRadius = radius * scale.getRadius();
        double angle = start_angle;
        while (angle < stop_angle) {
            op.addPoint((int)centerRadius, (int)angle);
            angle += df;
        }
        op.addPoint((int)centerRadius, (int)stop_angle);

        // Get the max width and height for center contour
        int maxWidth = cgauge.getGaugeArea().getBounds().width;
        int maxHeight = cgauge.getGaugeArea().getBounds().height;
        
        // Get the origin of the contour
        int xOrigin = bounds.x + getWidth() / 2;
        int yOrigin = bounds.y + getHeight() / 2;
        
        // Transform the polar polygon to an (x,y) Polygon
        Polygon xyContour = new Polygon();
        xyContour.addPoint(xOrigin, yOrigin);
        for (int i = 0; i < op.npoints; i++) {
            int xpoint = op.xpoints[i];
            double radians = GaugeUtil.toRadians((double)op.ypoints[i]);
            int x = xOrigin + (int)(xpoint * Math.cos(radians));
            int y = yOrigin - (int)(xpoint * Math.sin(radians));
            if (x > maxWidth) {
                x = maxWidth;
            }
            if (y > maxHeight) {
                y = maxHeight;
            }
            xyContour.addPoint(x, y);
        }
        xyContour.addPoint(xOrigin, yOrigin);
     
        shape = xyContour;
    }
    return(shape);
}

}
