/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCCircularGauge.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import com.klg.jclass.util.ServerRenderable;
import com.klg.jclass.util.ServerRenderer;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.util.Enumeration;
import java.util.Vector;

/**
 * <code>JCCircularGauge</code> models a real-world gauge, such as you might
 * see on a car dashboard. This class encapsulates the available types of
 * circular gauges: full circle, semicircle, and quadrant.
 */
public class JCCircularGauge extends JCGauge
                             implements java.io.Serializable,
                                        ServerRenderable
{

static final long serialVersionUID = -5471322510947631802L;

/**
 * The type of gauge; full circle, semicircle, or quadrant.
 */
protected GaugeType gaugeType = GaugeType.FULL_CIRCLE;

/**
 * The center object.
 */
protected JCCenter center;

/**
 * The list of labels for this gauge.
 *
 * @see com.klg.jclass.swing.gauge.JCLabelGenerator
 */
protected Vector labels = new Vector();

/** Temporary Graphics object used to draw without showing to screen.*/
private Graphics prevSetGraphics = null;

/**
 * No-argument constructor for <code>JCCircularGauge</code>.
 * Creates a default center, needle, and tick marks.
 * @see #setScale(JCCircularScale)
 * @see #setCenter(JCCenter)
 * @see #addNeedle(JCAbstractNeedle)
 * @see #addTick(JCTick)
 * @see #getScale()
 * @see #getCenter()
 * @see #removeNeedle(JCAbstractNeedle)
 * @see #removeTick(JCTick)
 */
public JCCircularGauge()
{
    // create a scale for a full circle
    JCCircularScale cscale = new JCCircularScale(this);
    setScale(cscale, false);

    // create the default center
    JCCenter center = new JCCenter(cscale);
    setCenter(center);

    // create a needle
    JCCircularNeedle needle = new JCCircularNeedle(cscale);
    addNeedle(needle);

    // default ticks
    JCCircularTick tick = new JCCircularTick(cscale);
    addTick(tick);

    // Now add scale to gauge area (so its drawn first)
    gaugeArea.add(cscale);

    // set the size
    setPreferredSize(new Dimension(75, 75));
}

/**
 * Constructs a <code>JCCircularGauge</code> with no default objects except
 * possibly a circular scale.
 *<p>
 * Use the following methods to add needles, indicators, ticks, and ranges.
 * @see #setScale(JCCircularScale)
 * @see #setCenter(JCCenter)
 * @see #addIndicator(JCAbstractIndicator)
 * @see #addNeedle(JCAbstractNeedle)
 * @see com.klg.jclass.swing.gauge.JCAbstractScale#addRange(JCRange)
 * @see #addTick(JCTick)
 * @see #getScale()
 * @see #getCenter()
 * @see #removeIndicator(JCAbstractIndicator)
 * @see #removeNeedle(JCAbstractNeedle)
 * @see com.klg.jclass.swing.gauge.JCAbstractScale#removeRange(JCRange)
 * @see #removeTick(JCTick)
 * @param createScale
 */
public JCCircularGauge(boolean createScale)
{
    if (createScale) {
        // create a linear scale
        JCCircularScale cscale = new JCCircularScale(this);
        setScale(cscale);
    }
}

/**
 * Constructs a gauge of the specified gauge type.
 * The scale, center, needles, ranges and ticks must be added using the
 * following methods.
 *
 * @param gaugeType one of the constants defined in this class
 * @see #setScale(JCCircularScale)
 * @see #setCenter(JCCenter)
 * @see #addIndicator(JCAbstractIndicator)
 * @see #addNeedle(JCAbstractNeedle)
 * @see com.klg.jclass.swing.gauge.JCAbstractScale#addRange(JCRange)
 * @see #addTick(JCTick)
 * @see #getScale()
 * @see #getCenter()
 * @see #removeIndicator(JCAbstractIndicator)
 * @see #removeNeedle(JCAbstractNeedle)
 * @see com.klg.jclass.swing.gauge.JCAbstractScale#removeRange(JCRange)
 * @see #removeTick(JCTick)
 */
public JCCircularGauge(GaugeType gaugeType)
{
    this.gaugeType = gaugeType;
}

/**
 * Paints the gauge.
 *
 * @param g the <code>Graphics</code> with which to paint
 */
public void paintComponent(Graphics g)
{
    // Draw a rectangle with gauge background
    Rectangle bounds = getBounds();
    if (isOpaque()) {
        g.setColor(getBackground());
        g.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);
    }

    // Draw a background image if there is one
    if (image != null) {
        drawImage(g, image, getBounds(), getBackground());
    }

    // Force the center to reposition itself
    gaugeArea.doLayout();

    // continue with normal painting
    super.paintComponent(g);




}

/**
 * Returns the circular gauge's circular scale (appropriately cast).
 *
 * @return the circular gauge's circular scale
 */
public JCCircularScale getCircularScale()
{
    return ((JCCircularScale)getScale());
}

/**
 * Adds a circular scale to the circular gauge.
 *
 * @param scale the circular scale to be added
 * @param addToGaugeArea add the scale to the gauge's <code>gaugeArea</code>?
 */
public void setScale(JCCircularScale scale, boolean addToGaugeArea)
{
    // Out with the old
    if (this.scale != null) {
        gaugeArea.remove((Component)this.scale);
    }
    // In with the new
    this.scale = scale;
    if (addToGaugeArea == true) {
        gaugeArea.add(scale);
    }
    setIndicatorScales(scale);
    setNeedleScales(scale);
}

/**
 * Adds a circular scale to the circular gauge.
 *
 * @param scale the circular scale to be added
 */
public void setScale(JCCircularScale scale)
{
    setScale(scale, true);
}

/**
 * Adds a label to the gauge.
 *
 * @param label the <code>JComponent</code> to be added as a label
 * @param constraint the <code>RadialConstraint</code> specifying where to
 *     place the label
 */
public void addLabel(JComponent label, RadialConstraint constraint)
{
    labels.addElement(label);
    getGaugeArea().add(label, constraint, 0);
}

/**
 * Adds a label to the gauge.
 *
 * @param label the <code>JComponent</code> to be added as a label
 * @param constraint the <code>RadialConstraint</code> specifying where to
 *     place the label
 * @param index the <i>z</i>-order index
 */
public void addLabel(JComponent label, RadialConstraint constraint, int index)
{
    labels.addElement(label);
    getGaugeArea().add(label, constraint, index);
}

/**
 * Removes a label.
 *
 * @param label the <code>JComponent</code> to be removed
 */
public void removeLabel(JComponent label)
{
    if (labels.contains(label)) {
        labels.removeElement(label);
        getGaugeArea().remove(label);
    }
}

/**
 * Returns the center for this gauge.
 *
 * @return the <code>JCCenter</code> object associated with this gauge
 */
public JCCenter getCenter()
{
    return(center);
}

/**
 * Sets the center object for this gauge.
 *
 * @param center the <code>JCCenter</code> for this gauge
 * @param index the <i>z</i>-order index at which to add the center
 */
public void setCenter(JCCenter center, int index)
{
    // Remove the old
    if (this.center != null) {
        gaugeArea.remove(this.center);
    }

    // Add the new
    this.center = center;
    if (center != null) {
        getGaugeArea().add(center, new RadialConstraint(this, 0, 0), index);
    }
}

/**
 * Sets the center for this gauge.
 *
 * @param center the <code>JCCenter</code> for this gauge
 */
public void setCenter(JCCenter center)
{
    setCenter(center, 0);
}

/**
 * Gets the gauge type, one of the
 * <code>JCCircularGauge.GaugeType</code> enums.
 *
 * @return the gauge type
 */
public GaugeType getGaugeType()
{
    return(gaugeType);
}

/**
 * Sets the type for this gauge, one of the
 * <code>JCCircularGauge.GaugeType</code> enums.
 *
 * @param gaugeType the gauge type, full circle, semicircle, or quadrant.
 */
public void setGaugeType(GaugeType gaugeType)
{
    this.gaugeType = gaugeType;

    if (gaugeArea != null) {
        gaugeArea.doLayout();
    }

    JCCircularScale cscale = getCircularScale();
    if (cscale != null) {
        // make the scale fit the type

        // Full circle -- use width for width and height
        if (gaugeType == JCCircularGauge.GaugeType.FULL_CIRCLE) {
            cscale.setStartAngle(0);
            cscale.setStopAngle(360);
        }

        // Top semi -- use width and half height
        else if (gaugeType == JCCircularGauge.GaugeType.TOP_HALF_CIRCLE) {
            cscale.setStartAngle(0);
            cscale.setStopAngle(180);
        }

        // Bottom semi -- use width and half height
        else if (gaugeType == JCCircularGauge.GaugeType.BOTTOM_HALF_CIRCLE) {
            cscale.setStartAngle(180);
            cscale.setStopAngle(360);
        }

        // LEFT semi -- use half width and height
        else if (gaugeType == JCCircularGauge.GaugeType.LEFT_HALF_CIRCLE) {
            cscale.setStartAngle(90);
            cscale.setStopAngle(270);
        }

        // RIGHT semi -- use half width and height
        else if (gaugeType == JCCircularGauge.GaugeType.RIGHT_HALF_CIRCLE) {
            cscale.setStartAngle(270);
            cscale.setStopAngle(90);
        }

        // LOWER_LEFT_QUARTER_CIRCLE
        else if (gaugeType == JCCircularGauge.GaugeType.LOWER_LEFT_QUARTER_CIRCLE) {
            cscale.setStartAngle(180);
            cscale.setStopAngle(270);
        }

        // LOWER_RIGHT_QUARTER_CIRCLE
        else if (gaugeType == JCCircularGauge.GaugeType.LOWER_RIGHT_QUARTER_CIRCLE) {
            cscale.setStartAngle(270);
            cscale.setStopAngle(360);
        }

        // UPPER_RIGHT_QUARTER_CIRCLE
        else if (gaugeType == JCCircularGauge.GaugeType.UPPER_RIGHT_QUARTER_CIRCLE) {
            cscale.setStartAngle(0);
            cscale.setStopAngle(90);
        }

        // UPPER_LEFT_QUARTER_CIRCLE
        else if (gaugeType == JCCircularGauge.GaugeType.UPPER_LEFT_QUARTER_CIRCLE) {
            cscale.setStartAngle(90);
            cscale.setStopAngle(180);
        }

        cscale.doLayout();
        for (int i = 0; i < cscale.getTicks().size(); i++) {
            JCCircularTick t = (JCCircularTick)cscale.getTicks().elementAt(i);
            t.doLayout();
        }
    }
    redraw();
}

/**
 * Determines the <i>x</i>, <i>y</i>, width and height of an arc's bounding box.
 *
 * @return the <code>Rectangle</code> bounding the arc
 */
public Rectangle getArcBounds()
{
    // There are nine cases that must be handled. These correspond
    // to the nine gauge types
    //
    // 1. full circle (0-360 degrees)
    // 2. top semi circle (0-180 degrees)
    // 3. bottom semi circle (180-360 degrees)
    // 4. left semi circle (270-450 degrees)
    // 5. right semi circle (90-270 degrees)
    // 6. upper right quater pie (0-90 degrees)
    // 7. upper left quarter pie (90-180 degrees)
    // 8. lower right quater pie (180-270 degrees)
    // 9. lower left quarter pie (270-360 degrees)

    Rectangle bounds = null;
    JCCircularScale cscale = getCircularScale();
    int radius = (int)cscale.getRadius();
    int diameter = (int)(cscale.getRadius() * 2);
    Insets insets = cscale.getScaleInsets();
    int x = insets.left;
    int y = insets.top;
    int width = gaugeArea.getBounds().width;
    int height = gaugeArea.getBounds().height;

    width -= insets.left + insets.right;
    height -= insets.top + insets.bottom;

    Rectangle gab = new Rectangle(x, y, width, height);

    // Full circle -- use width for width and height
    if (gaugeType.equals(JCCircularGauge.GaugeType.FULL_CIRCLE)) {
        // center wrt width and height
        bounds = new Rectangle(x + (gab.width - diameter) / 2,
                               y + (gab.height - diameter) / 2,
                               diameter, diameter);
    }

    // Top semi -- use width and half height
    else if (gaugeType.equals(JCCircularGauge.GaugeType.TOP_HALF_CIRCLE)) {
        // center wrt width
        bounds = new Rectangle(x + (gab.width - diameter) / 2, y,
                               diameter, diameter);
    }

    // Bottom semi -- use width and half height
    else if (gaugeType.equals(JCCircularGauge.GaugeType.BOTTOM_HALF_CIRCLE)) {
        // center wrt width
        bounds = new Rectangle(x + (gab.width - diameter) / 2, -(radius - y),
                               diameter, diameter);
    }

    // LEFT semi -- use half width and height
    else if (gaugeType.equals(JCCircularGauge.GaugeType.LEFT_HALF_CIRCLE)) {
        // center wrt height
        bounds = new Rectangle(x, y + (gab.height - diameter) / 2,
                               diameter, diameter);
    }

    // RIGHT semi -- use half width and height
    else if (gaugeType.equals(JCCircularGauge.GaugeType.RIGHT_HALF_CIRCLE)) {
        // center wrt height
        bounds = new Rectangle(-(radius - x), y + (gab.height - diameter) / 2,
                               diameter, diameter);
    }

    // LOWER_LEFT_QUARTER_CIRCLE
    else if (gaugeType.equals(JCCircularGauge.GaugeType.LOWER_LEFT_QUARTER_CIRCLE)) {
        bounds = new Rectangle(x, -(radius - y), diameter, diameter);
    }

    // LOWER_RIGHT_QUARTER_CIRCLE
    else if (gaugeType.equals(JCCircularGauge.GaugeType.LOWER_RIGHT_QUARTER_CIRCLE)) {
        bounds = new Rectangle(-(radius - x), -(radius - y), diameter, diameter);
    }

    // UPPER_RIGHT_QUARTER_CIRCLE
    else if (gaugeType.equals(JCCircularGauge.GaugeType.UPPER_RIGHT_QUARTER_CIRCLE)) {
        bounds = new Rectangle(-(radius - x), y, diameter, diameter);
    }

    // UPPER_LEFT_QUARTER_CIRCLE
    else if (gaugeType.equals(JCCircularGauge.GaugeType.UPPER_LEFT_QUARTER_CIRCLE)) {
        bounds = new Rectangle(x, y, diameter, diameter);
    }

    return (bounds);
}

/**
 * Sends pick events to listeners and moves the closest
 * needle with a <code>CLICK</code> interaction enabled to the value indicated
 * by the mouse click.
 *
 * @param event a <code>MouseEvent</code>
 */
public void mouseClicked(MouseEvent event)
{
    // propagate event to listeners
    super.mouseClicked(event);

    JCGaugePickEvent pe = pick(event.getPoint());

    if (pe != null) {
        // see if there are any needles interested in snap movements
        // and move the closest one
        JCCircularNeedle closest;

        // if there is such a needle, move it
        closest = getClosestNeedle(pe, JCAbstractNeedle.InteractionType.CLICK);
        if (closest != null) {
            closest.setValue(pe.getValue());
            redraw();
        }
    }
}

/**
 * Drags the closest needle.
 *
 * @param event a <code>MouseEvent</code>
 */
public void mouseDragged(MouseEvent event)
{
    // propagate event to listeners
    super.mouseClicked(event);

    JCGaugePickEvent pe = pick(event.getPoint());

    if (pe != null) {
        // if we are not already in the process of dragging a needle ...
        // @see mouseReleased where it is nulled out when released

        if (draggedNeedle == null) {
            // find the closest dragable needle
            draggedNeedle =
                getClosestNeedle(pe, JCAbstractNeedle.InteractionType.DRAG);
        }

        // if there was such a needle, move it
        if (draggedNeedle != null) {
            draggedNeedle.setValue(pe.getValue());
            redraw();
        }
    }
}

/**
 * Returns the closest needle to the clicked/dragged point.
 * Returns closest needle of type
 * <code>CLICK</code>/<code>DRAG</code>/<code>CLICK_DRAG</code>,
 * or if there are none of these, the closest needle of type <code>NONE</code>.
 *
 * @param pe the pick event with the click inforamation
 * @param interactionType the needle must be
 *     <code>JCNeedle.InteractionType</code>
 * @return the closest needle of the indicated type
 */
public JCCircularNeedle getClosestNeedle(
    JCGaugePickEvent pe,
    JCAbstractNeedle.InteractionType interactionType)
{
    JCCircularNeedle closestNeedle = null;

    // get the angle of the clicked point
    double pointAngle = valueToAngle(pe.getValue());

    // Find needle to drag/click
    // by comparing the angles between the point and the needles
    // and choosing the needle with the closest angle to the point
    Enumeration e = needles.elements();
    while (e.hasMoreElements()) {
        JCCircularNeedle needle = (JCCircularNeedle)e.nextElement();

        // Needle must be of the interactive sort
        JCAbstractNeedle.InteractionType it = needle.getInteractionType();
        if (it == JCAbstractNeedle.InteractionType.CLICK_DRAG ||
            it == interactionType)
        {
            if (closestNeedle == null) {
                closestNeedle = needle;
            }
            else {
                double newAngle = valueToAngle(needle.getValue());
                double oldAngle = valueToAngle(closestNeedle.getValue());
                double newDistance = GaugeUtil.distanceBetweenAngles(pointAngle,
                                                                     newAngle);
                double oldDistance = GaugeUtil.distanceBetweenAngles(pointAngle,
                                                                     oldAngle);

                // If the old distance was father away, select the new,
                // closest needle
                if (oldDistance > newDistance) {
                    closestNeedle = needle;
                }
            }
        }

    }
    return(closestNeedle);
}

/**
 * A convenience method that returns the angle corresponding to a circular
 * scale value.
 * @param value a circular scale value
 * @return The angle based on the value.
 */
protected double valueToAngle(double value)
{
    JCCircularScale cscale = getCircularScale();

    return GaugeUtil.valueToAngle(value,
                                  cscale.getMin(), cscale.getMax(),
                                  cscale.getStartAngle(),
                                  cscale.getStopAngle());
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
    if (getRepaintEnabled()) {
        repaint();
    }
}

/**
 * Draws an image as the gauge background.
 *
 * @param g the Graphics on which to draw
 * @param bounds
 * @param color
 * @param image
 */
protected void drawImage(Graphics g, Image image, Rectangle bounds, Color color)
{
    // create an image observer
    java.awt.image.ImageObserver observer = new javax.swing.JLabel();

    // draw the image
    if (super.getScaleImage()) {
        g.drawImage(image, bounds.x, bounds.y, bounds.width, bounds.height,
                    color, observer);
    }
    else {
        g.drawImage(image, bounds.x, bounds.y, color, observer);
    }
}

////////////////// inner class

/**
 * Enums for gauge types.
 * There are nine gauge types defined by start and sweep angles:
 *
 * <ol>
 * <li> <code>FULL_CIRCLE</code> (0-360 degrees)</li>
 * <li> <code>TOP_HALF_CIRCLE</code> (0-180 degrees)</li>
 * <li> <code>BOTTOM_HALF_CIRCLE</code> (180-180 degrees)</li>
 * <li> <code>LEFT_HALF_CIRCLE</code> (90-180 degrees)</li>
 * <li> <code>RIGHT_HALF_CIRCLE</code> (270-180 degrees)</li>
 * <li> <code>UPPER_RIGHT_QUARTER_CIRCLE</code> (0-90 degrees)</li>
 * <li> <code>LOWER_RIGHT_QUARTER_CIRCLE</code> (270-90 degrees)</li>
 * <li> <code>UPPER_LEFT_QUARTER_CIRCLE</code> (90-90 degrees)</li>
 * <li> <code>LOWER_LEFT_QUARTER_CIRCLE</code> (180-90 degrees)</li>
 * </ol>
 */
public static class GaugeType implements java.io.Serializable {

    /**
     * The scale is a full circle.
     */
    public static final GaugeType FULL_CIRCLE = new GaugeType(0, 360);

    /**
     * The scale is a semicircle oriented so that
     * its diameter is horizontal and its
     * circumference is above its base.
     */
    public static final GaugeType TOP_HALF_CIRCLE = new GaugeType(0, 180);

    /**
     * The scale is a semicircle oriented so that
     * its diameter is horizontal and its
     * circumference is below its base.
     */
    public static final GaugeType BOTTOM_HALF_CIRCLE = new GaugeType(180, 180);

    /**
     * The scale is a semicircle oriented so that
     * its diameter is vertical and its
     * circumference is to the left of its base.
     */
    public static final GaugeType LEFT_HALF_CIRCLE = new GaugeType(90, 180);

    /**
     * The scale is a semicircle oriented so that
     * its diameter is vertical and its
     * circumference is to the right of its base.
     */
    public static final GaugeType RIGHT_HALF_CIRCLE = new GaugeType(270, 180);

    /**
     * A quadrant whose arc lies northeast.
     */
    public static final GaugeType UPPER_RIGHT_QUARTER_CIRCLE = new GaugeType(0, 90);

    /**
     * A quadrant whose arc lies southeast.
     */
    public static final GaugeType LOWER_RIGHT_QUARTER_CIRCLE = new GaugeType(270, 90);

    /**
     * A quadrant whose arc lies northwest.
     */
    public static final GaugeType UPPER_LEFT_QUARTER_CIRCLE = new GaugeType(90, 90);

    /**
     * A quadrant whose arc lies southwest.
     */
    public static final GaugeType LOWER_LEFT_QUARTER_CIRCLE = new GaugeType(180, 90);

    /**      */
    int startAngle;
    /**      */
    int sweepAngle;

    /**
     * Only allow predefined types.
     *
     * @param startAngle
     * @param sweepAngle
     */
    private GaugeType(int startAngle, int sweepAngle)
    {
        this.startAngle = startAngle;
        this.sweepAngle = sweepAngle;
    }

    /**
     * Returns the start angle for this gauge.
     * @return the start angle
     */
    public int getStartAngle()
    {
        return(startAngle);
    }

    /**
     * Returns the sweep angle. The sweep angle is the angle between the gauge's
     * start and stop angles.
     * @return the sweep angle
     */
    public int getSweepAngle()
    {
        return(sweepAngle);
    }

    /**
     * Return <code>true</code> if the objects have the same start and sweep angles.
     * @return <code>true</code> if the objects have the same start and sweep angles
     * @param type
     */
    public boolean equals(GaugeType type)
    {
        return(this.startAngle == type.startAngle &&
               this.sweepAngle == type.sweepAngle);
    }

}

//ServerRenderable methods
/**
 * Overrides parent <code>getGraphics()</code>. If another
 * <code>Graphics</code> object has been set to be the current one, it is
 * returned. Otherwise, call the superclass' <code>getGraphics()</code> method.
 * @return the <code>Graphics</code> object to use to draw this chart
 */
public Graphics getGraphics()
{
    if (prevSetGraphics == null) {
        return(super.getGraphics());
    }
    return(prevSetGraphics);
}

/**
 * Provides a way for the circular gauge to draw to a custom
 * <code>Graphics</code> object. Call with a null to specify use of the default
 * <code>Graphics</code> object used to draw to the screen.
 * @param g <code>Graphics</code> object to use for drawing
 */
public void setGraphics(Graphics g)
{
    prevSetGraphics = g;
    getLegend().setGraphics(g);
}

/**
 * Internal use only.  Don't use.
 */
public Image serverSnapshot()
{
    return(ServerRenderer.snapshot(this));
}

/**
 * Internal use only.  Don't use.
 */
public void serverPaint(Graphics igc)
{
    ServerRenderer.paintContainer(igc, this);
}

}
