/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCCircularRange.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import java.awt.*;
import java.awt.geom.Arc2D;
import java.awt.geom.GeneralPath;

/**
 * A circular range is used to demarcate some sub-portion of the scale.
 * For example, on a tachometer gauge you could have a range
 * to delimit the "red line" range of revolutions per minute (rpm).
 * A gauge can have zero or more ranges associated with it.
 */
public class JCCircularRange extends JCAbstractRange
                             implements java.io.Serializable
{

static final long serialVersionUID = -3498488049576531605L;

protected int innerArcStartCompensator = 0;
protected int innerArcStopCompensator = 0;

/**
 * Creates a circular range.
 *
 * @param foreground the range's color
 * @param scale the scale associated with this range
 * @param innerExtent the ratio of the radius's length to start drawing this
 * range from the center (for example, 0.8 means start
 * drawing this range 80% of the radius's size from the center)
 * @param outerExtent  the ratio of the radius's length to end drawing this
 * range from the center (for example, 0.9 means stop
 * drawing this range 90% of the radius's size from the center)
 * @param startValue the scale value (not angle) at which to begin the range
 * @param stopValue the scale value (not angle) at which to stop the range
 */
public JCCircularRange(Color				foreground,
                       JCCircularScale		scale,
                       double				innerExtent,
                       double				outerExtent,
                       double				startValue,
                       double				stopValue)
{
	super(foreground,
          scale,
          innerExtent,
          outerExtent,
          startValue,
          stopValue);
}

/**
 * Returns the circular scale associated with this range.
 *
 * @return the circular scale associated with this range
 */
public JCCircularScale getCircularScale()
{
    return((JCCircularScale)getScale());
}

/**
 * Returns the preferred size of this range, which will be
 * the preferred size of the scale.
 *
 * @return the preferred size
 */
public Dimension getPreferredSize()
{
    return(getCircularScale().getPreferredSize());
}

/**
 * Compensates for awt <code>drawArc</code> precision
 * problems when drawing the inner arc
 * of the range. This applies to the J version only.
 *
 * @return the amount by which to overdraw in the inner arc
 * start angle
 */
public int getInnerArcStartCompensator()
{
    return(innerArcStartCompensator);
}

/**
 * Compensates for awt <code>drawArc</code> precision
 * problems when drawing the inner arc
 * of the range. This applies to the J version only.
 *
 * @param innerArcStartCompensator the amount by which to overdraw in the
 * inner arc start angle
 */
public void setInnerArcStartCompensator(int innerArcStartCompensator)
{
    this.innerArcStartCompensator = innerArcStartCompensator;
    redraw();
}

/**
 * Compensates for awt <code>drawArc</code> precision
 * problems when drawing the inner arc
 * of the range. This applies to the J version only.
 *
 * @return the amount by which to overdraw in the inner arc
 * stop angle
 */
public int getInnerArcStopCompensator()
{
    return(innerArcStopCompensator);
}

/**
 * Compensates for awt <code>drawArc</code> precision
 * problems when drawing the inner arc
 * of the range. This applies to the J version only.
 *
 * @param innerArcStopCompensator the amount by which to overdraw in the
 * inner arc stop angle
 */
public void setInnerArcStopCompensator(int innerArcStopCompensator)
{
    this.innerArcStopCompensator = innerArcStopCompensator;
    redraw();
}

/**
 * This routine will draw a circular range.
 * @param g
 */
public void paintComponent(Graphics g)
{
   JCCircularScale cscale = getCircularScale();

    // Get the various values
    double circ_min = cscale.getMin();
    double circ_max = cscale.getMax();
    double radius_size = cscale.getRadius();
    double circ_start = GaugeUtil.normalizeAngle(cscale.getStartAngle());
    double circ_stop = GaugeUtil.normalizeAngle(cscale.getStopAngle());

    // Make a local copy so they can be modified
    double start_value = GaugeUtil.clamp(startValue, circ_min, circ_max);
    double stop_value = GaugeUtil.clamp(stopValue, circ_min, circ_max);

    // Do we need to draw anything?
    if (!isVisible() ||
        circ_min >= circ_max ||
        innerExtent >= outerExtent ||
        start_value >= stop_value)
    {
        return;
    }

    // Adjust for direction
    if (cscale.getDirection().equals(JCAbstractScale.Direction.BACKWARD)) {
        double temp = start_value;
        start_value = stop_value;
        stop_value = temp;

        stop_value = circ_max - stop_value + circ_min;
        start_value = circ_max - start_value + circ_min;
    }

    // Convert our start and stop values to start and stop angles.
    double start_angle = GaugeUtil.valueToAngle(start_value, circ_min, circ_max,
                                                circ_start, circ_stop);
    double stop_angle = GaugeUtil.valueToAngle(stop_value, circ_min, circ_max,
                                               circ_start, circ_stop);

    // The line width is determined by the inner and outer extents (in pixels).
    double inner_extent = innerExtent * radius_size;
    double outer_extent = outerExtent * radius_size;

    // Since drawarc stop_angle is relative to start angle, we compensate
    if (start_angle >= stop_angle) {
        stop_angle += 360.0;
    }

    // java draws from start to arc, not start to stop
    double arcAngle = stop_angle - start_angle;

    Graphics gc = g.create();

    // compute the dimensions of the outer bound (outer arc)

    // get the center of the circle
    Rectangle bounds = cscale.getCircularGauge().getArcBounds();
    double xOrigin = bounds.x + radius_size;
    double yOrigin = bounds.y + radius_size;

    // compute the size of the larger arc
    int x = (int)(xOrigin - outer_extent);
    int y = (int)(yOrigin - outer_extent);
    int w = (int)(2.0 * outer_extent);
    int h = (int)(2.0 * outer_extent);

    // now compute the inner arc

    int x1 = (int)(xOrigin - inner_extent);
    int y1 = (int)(yOrigin - inner_extent);
    int h1 = (int)(2.0 * inner_extent);
    Graphics2D g2d = (Graphics2D)gc;

    // Draw an image if there is one
    if (image != null) {
        if (scaleImage) {
            Shape oldClip = g2d.getClip();
            // set the scale clip region
            Shape clip = getCircularRangeShape(start_angle, stop_angle,
                                               radius_size, inner_extent,
                                               outer_extent);
            g2d.setClip(clip);
            // draw image in the clip region
            drawImage(g2d, image, bounds);
            // reset the previous clip region
            g2d.setClip(oldClip);
        }
        else {
            drawImage(g2d, image, bounds);
        }
    }
    else {
        // Here is what is going on ...
        // For g2d we draw a thick line as the range.
        // g2d draws a thick line centered on the circle. That is, it draws
        // inward from the circumfrence 1/2 the thickness of the line
        // and outside 1/2 the thickness of the line
        // Simply bending the thick line does not result in a pie shape, so
        // we compute the arc trace and clip to that to give us the pie shape.

        // determine the width of the line
        int width = Math.abs(y - y1);

        // set width on the gc
        g2d.setStroke(new BasicStroke(width));

        // give us the bounding arc so we can clip the fat line to a pie shape

        //NOTE: that traceArcPath is borrowed from pagelayout. It's origin
        // is at the lower left, opposed to java's upper left. Hence,
        // y stays the same, but x is subtracted from the width (xOrigin * 2)
        // and the angles are inverted
        GeneralPath pie =
            traceArcPath((xOrigin * 2) - xOrigin,   // the inverted x
                         yOrigin,                   // regular y
                         w / 2,                     // radius
                         h / 2,                     // radius
                         -start_angle, -arcAngle);	// inverted angles
        pie.lineTo((int)((xOrigin * 2) - xOrigin), (int)yOrigin);
        pie.closePath();

        // Clip the line to a pie shape
        g2d.clip(pie);

        // Draw the fat line; handle fact line is centered on circumference
        g2d.draw(new Arc2D.Double(x + (x1 - x) / 2,
                                  y + (y1 - y) / 2,
                                  h - (h - h1) / 2,
                                  h - (h - h1) / 2,
                                  start_angle,
                                  stop_angle - start_angle,
                                  Arc2D.OPEN));
    }
    gc.dispose();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
    if (getCircularScale().getGauge().getRepaintEnabled()) {
        repaint();
    }
}

/**
 *
 * This code borrowed from JClass PageLayout's pdf printer.
 *
 */

private GeneralPath traceArcPath(double x, double y, double semiMajorX,
                                 double semiMajorY, double startAngle,
                                 double arcAngle)
{
    GeneralPath generalPath = new GeneralPath();
    double theta, rad, endAngle, cp_factor, angle;
    double x0 = 0, x1 = 0, x2 = 0, x3 = 0, y0 = 0, y1 = 0, y2 = 0, y3 = 0;


    // Find the start point of the arc and move there
    // NOTE: we compute radians because 1.1 doesn't provide a conversion
    if (arcAngle < 0) {
        rad = (startAngle + arcAngle) / 180.0 * Math.PI;
    }
    else {
        rad = startAngle / 180.0 * Math.PI;
    }
    x0 = x + semiMajorX * Math.cos(rad);
    y0 = y + semiMajorY * Math.sin(rad);

    generalPath.moveTo((float)x0, (float)y0);

    // If arc sweep < 0.001 degree, draw quick fake (straight line)
    if (Math.abs(arcAngle) < 0.001) {
        rad = (startAngle + arcAngle) / 180.0 * Math.PI;
        x0 = x + semiMajorX * Math.cos(rad);
        y0 = y + semiMajorY * Math.sin(rad);
        // output(truncateDecimalDigits(x0) + " " +
        // 	   truncateDecimalDigits(-y0) + " l");
    }

    // If arcAngle is negative, swap the endpoints
    if (arcAngle < 0) {
        startAngle += arcAngle;
        arcAngle = -arcAngle;
    }

    // Pre-compute the offset constants for 90-degree arcs
    double x_cp_offset = 0.5522848 * semiMajorX;
    double y_cp_offset = 0.5522848 * semiMajorY;

    // Compute the start point and end angle
    // NOTE: we compute radians because 1.1 doesn't provide a conversion
    rad = startAngle / 180.0 * Math.PI;
    x0 = x + semiMajorX * Math.cos(rad);
    y0 = y + semiMajorY * Math.sin(rad);

    // Compute the first cardinal point in the range
    theta = ((int)startAngle / 90) * 90;

    // If there is a partial (less than quarter) arc at the start of the range
    if (Math.abs(theta - startAngle) > 0.001) {

        // Adjust theta to the next cardinal point after the start
        if (startAngle > 0) {
            theta += 90;
        }

        // If the initial arc extends to (or past) a cardinal point, stop it
        endAngle = startAngle + arcAngle;
        if (endAngle > theta) {
            endAngle = theta;
        }

        // Draw an arc from startAngle to endAngle (less than 1/4 circle)
        // NOTE: we compute radians because 1.1 doesn't provide a conversion
        rad = endAngle / 180.0 * Math.PI;
        x3 = x + semiMajorX * Math.cos(rad);
        y3 = y + semiMajorY * Math.sin(rad);

        // Determine the actual arc swept
        angle = endAngle - startAngle;

        // Compute the bezier curve constant for this angle
        cp_factor = computeControlPointFactor(angle);

        // Compute the control points
        // x1 = x0 - cp_factor * semiMajorX * (y0 - y);
        // y1 = y0 + cp_factor * semiMajorY * (x0 - x);
        // x2 = x3 + cp_factor * semiMajorX * (y0 - y);
        // y2 = y3 - cp_factor * semiMajorY * (x0 - x);
        x1 = x0 - cp_factor * (y0 - y);
        y1 = y0 + cp_factor * (x0 - x);
        x2 = x3 + cp_factor * (y3 - y);
        y2 = y3 - cp_factor * (x3 - x);

        // Draw the arc segment
        // drawSegment(x1, y1, x2, y2, x3, y3);
        generalPath.curveTo((float)x1, (float)y1, (float)x2, (float)y2, (float)x3, (float)y3);

        // Advance to the start of the next arc segment
        x0 = x3;
        y0 = y3;
    }

    // While we can draw quarter circles within the range of the arc
    endAngle = startAngle + arcAngle;
    while (theta + 90 <= endAngle) {

        int modulus = ((int)theta) % 360;

        // Generate the control point offsets of the curve
        if (modulus == 0) {
            x1 = 0;
            y1 = y_cp_offset;
            x2 = x_cp_offset;
            y2 = 0;
        }
        else if (modulus == 90 || modulus == -270) {
            x1 = -x_cp_offset;
            y1 = 0;
            x2 = 0;
            y2 = y_cp_offset;
        }
        else if (modulus == 180 || modulus == -180) {
            x1 = 0;
            y1 = -y_cp_offset;
            x2 = -x_cp_offset;
            y2 = 0;
        }
        else if (modulus == 270 || modulus == -90) {
            x1 = x_cp_offset;
            y1 = 0;
            x2 = 0;
            y2 = -y_cp_offset;
        }

        // Compute the endpoint of this segment
        theta += 90;

        // NOTE: we compute radians because 1.1 doesn't provide a conversion
        rad = theta / 180.0 * Math.PI;
        x3 = x + semiMajorX * Math.cos(rad);
        y3 = y + semiMajorY * Math.sin(rad);

        // Compute the control points
        x1 += x0;
        y1 += y0;
        x2 += x3;
        y2 += y3;

        // Draw the curve for this segment
        // drawSegment(x1, y1, x2, y2, x3, y3);
        generalPath.curveTo((float)x1, (float)y1, (float)x2, (float)y2, (float)x3, (float)y3);

        // Advance to the start of the next segment
        x0 = x3;
        y0 = y3;
    }

    // If there is a segment of arc left to be drawn
    if (theta < endAngle && Math.abs(theta - endAngle) > 0.001) {

        // Draw an arc from theta to endAngle (less than 1/4 circle)
        // NOTE: we compute radians because 1.1 doesn't provide a conversion
        rad = endAngle / 180.0 * Math.PI;
        x3 = x + semiMajorX * Math.cos(rad);
        y3 = y + semiMajorY * Math.sin(rad);

        // Determine the actual arc swept
        angle = endAngle - theta;

        // Compute the bezier curve constant for this angle
        cp_factor = computeControlPointFactor(angle);

        // Compute the control points
        // x1 = x0 - cp_factor * semiMajorX * (y0 - y);
        // y1 = y0 + cp_factor * semiMajorY * (x0 - x);
        // x2 = x3 + cp_factor * semiMajorX * (y0 - y);
        // y2 = y3 - cp_factor * semiMajorY * (x0 - x);
        x1 = x0 - cp_factor * (y0 - y);
        y1 = y0 + cp_factor * (x0 - x);
        x2 = x3 + cp_factor * (y3 - y);
        y2 = y3 - cp_factor * (x3 - x);

        // Draw the arc segment
        // drawSegment(x1, y1, x2, y2, x3, y3);
        generalPath.curveTo((float)x1, (float)y1, (float)x2, (float)y2, (float)x3, (float)y3);

        // Advance to the start of the next arc segment
        x0 = x3;
        y0 = y3;
    }

    return generalPath;
}


private double computeControlPointFactor(double angle)
{
    double rad, factor;

    // NOTE: we compute radians because 1.1 doesn't provide a conversion
    rad = (angle / 2.0) / 180.0 * Math.PI;

    // r * (4/3) * (1 - cos(theta/2)) / sin(theta/2)
    factor = (4.0 / 3.0) * (1.0 - Math.cos(rad)) / Math.sin(rad);

    return (factor);
}

/*
* Gets the polygon that approximates the contour of the circular range.
* @param startAngle the start angle of the range (0-360)
* @param stopAngle the stop angle of the range (0-360)
* @param radiusSize the size of the scale radius
* @param innerRadius the inner radius of the range
* @param outerRadius the outer radius of the range
* @return Polygon a polygon that approximates the range contour
*/
protected Polygon getCircularRangeShape(double startAngle, double stopAngle,
                                        double radiusSize,
                                        double innerRadius, double outerRadius)
{
    // make polygons of polar coordinates
    // inner arc polygon
    Polygon ip = new Polygon();
    double df = 2;
    double angle = startAngle;
    while (angle < stopAngle) {
        ip.addPoint((int)innerRadius, (int)angle);
        angle += df;
    }
    ip.addPoint((int)innerRadius, (int)stopAngle);

    // outer arc polygon
    Polygon op = new Polygon();
    angle = stopAngle;
    while (angle > startAngle) {
        op.addPoint((int)outerRadius, (int)angle);
        angle -= df;
    }
    op.addPoint((int)outerRadius, (int)startAngle);

    // get the center of the circle
    JCCircularScale cscale = getCircularScale();
    Rectangle bounds = cscale.getCircularGauge().getArcBounds();
    double xOrigin = bounds.x + radiusSize;
    double yOrigin = bounds.y + radiusSize;

    // transform polar coordinates in scalar coordinates
    Polygon xyContour = new Polygon();
    int x, y;
    for (int i = 0; i < ip.npoints; i++) {
        double radians = GaugeUtil.toRadians((double)ip.ypoints[i]);
        x = (int)(xOrigin + ip.xpoints[i] * Math.cos(radians));
        y = (int)(yOrigin - ip.xpoints[i] * Math.sin(radians));
        xyContour.addPoint(x, y);
    }
    for (int i = 0; i < op.npoints; i++) {
        double radians = GaugeUtil.toRadians((double)op.ypoints[i]);
        x = (int)(xOrigin + op.xpoints[i] * Math.cos(radians));
        y = (int)(yOrigin - op.xpoints[i] * Math.sin(radians));
        xyContour.addPoint(x, y);
    }
    return(xyContour);
}

/*
* Draws an image as the range.
*
* @param g the Graphics on which to draw
*/
protected void drawImage(Graphics g, Image image, Rectangle bounds)
{
    // create an image observer
    java.awt.image.ImageObserver observer = new javax.swing.JLabel();

    // draw the image
    if (super.getScaleImage()) {
        g.drawImage(image, bounds.x, bounds.y, bounds.width, bounds.height,
                    getForeground(), observer);
    }
    else {
        g.drawImage(image, bounds.x, bounds.y, getForeground(), observer);
    }
}

}
