/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCCircularScale.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import java.awt.*;
import java.awt.geom.*;
import java.util.Enumeration;


 


 

/**
 * <code>JCCircularScale</code> provides a graduated scale.
 * Angular measurements follow this convention:
 * Zero degrees is at 3 o'clock, 90&#176; at 12 o'clock,
 * 180&#176; at 9 o'clock and 270&#176; at 6 o'clock.
 * <code>JCTicks</code>
 * are used to mark this scale so that readings may be taken.
 * See <A HREF="../../../../../../elements/jcelements-4.html#angularMeasurements">
 * the section on how angles are measured</A> in the manual.
 */
public class JCCircularScale extends JCAbstractScale
                             implements java.io.Serializable
{

static final long serialVersionUID = 6835510720826476109L;

/**
 * The start angle in degrees.
 */
protected double startAngle = 0;

/**
 * The stop angle in degrees.
 */
protected double stopAngle = 360;

/**
 * The minimum diameter for the scale.
 */
protected double minSize = 8;

/**
 * Paints the complete background defined by <code>JCCircularGauge.GaugeType</code>,
 * not just area swept by the scale. For instance,
 * a scale whose type is <code>FULL_CIRCLE</code> and whose angular sweep is from
 * 0&#176; to 270&#176; paints only that region.
 * The default is <code>false</code>.
 * <br>
 * <code>True</code> means paint the entire area specified in <code>GaugeType</code>,
 * <code>false</code> means just the area spanning the scale's sweep angle.
 */
protected boolean			paintCompleteBackground = false;

/**
 * Creates a scale given its parent gauge.
 *
 * @param gauge the <code>JCCircularGauge</code> associated with this scale.
 */
public JCCircularScale(JCCircularGauge gauge)
{
	super(gauge);

	// set the layout
    setLayout(new RadialLayout());

    setForeground(Color.white);
    GaugeUtil.createLegendPopulatorRenderer(gauge, this);
}

/**
 * Creates a graduated scale.
 *
 * @param gauge the <code>JCCircularGauge</code> associated with this scale.
 * @param direction a <code>JCAbstractScale.Direction</code> enum indicating
 *     <code>FORWARD</code> or <code>BACKWARD</code> direction
 * @param min the minumum value this scale can have
 * @param max the maximum value this scale can have
 * @param startAngle the start angle in degrees
 * @param stopAngle the stop angle in degrees
 * @param foreground the color of this scale's foreground
 */
public JCCircularScale(JCCircularGauge              gauge,
                       JCAbstractScale.Direction 	direction,
                       double		                min,
                       double		                max,
                       double		                startAngle,
                       double		                stopAngle,
                       Color		                foreground)
{
    // Set the properties
    super(gauge, direction, min, max, foreground);

    this.startAngle = startAngle;
    this.stopAngle  = stopAngle;

    // Set the layout
    setLayout(new RadialLayout());

    GaugeUtil.createLegendPopulatorRenderer(gauge, this);
}

/**
 * This is called by our widget parent when it wants to know our
 * preferrred size.
 *
 * @return the <code>Dimension</code> (size) we'd like to be based on our
 *     radius
 */
public Dimension getPreferredSize()
{
	return new Dimension(getParent().getBounds().width,
                         getParent().getBounds().height);
}

/**
 * Overrides add to pass in a radial constraint.
 * @return
 * @param c the component to be added
 */
public Component add(Component c)
{
	add(c, new RadialConstraint(getGauge(), Double.MAX_VALUE,
                                Double.MAX_VALUE));
    return(c);
}

/**
 * Overrides add to pass in a radial constraint.
 * @return
 * @param c the component to be added
 * @param index the order this component has in the list; components added
 *     last are drawn first
 */
public Component add(Component c, int index)
{
	add(c, new RadialConstraint(getGauge(), Double.MAX_VALUE,
                                Double.MAX_VALUE), index);
    return(c);
}

/**
 * Draws the background of this scale.
 *
 * @param g the <code>Graphics</code> object onto which we will draw our
 * background
 */
public void paintComponent(Graphics g)
{
    int		            arc_angle;
	double              start_angle	= GaugeUtil.normalizeAngle(startAngle);
    double 	            stop_angle	= GaugeUtil.normalizeAngle(stopAngle);
    JCCircularGauge     cgauge = getCircularGauge();
    Rectangle           bounds = cgauge.getArcBounds();

	super.paintComponent(g);


 

    // start painting
    Graphics gc = g.create();

 	// get the adjusted x, y, w, h
	int x = bounds.x;
    int y = bounds.y;
    int w = bounds.width;
    int h = bounds.height;

    // should we draw the entire gauge swept by GaugeType angles or
    // just by the scale?
    if (getPaintCompleteBackground()) {
    	// do it for complete background
        start_angle = cgauge.getGaugeType().getStartAngle();
        arc_angle	= cgauge.getGaugeType().getSweepAngle();
    }
    else {
    	// paint just area swept by scale

        // handle case where start > stop
        if (start_angle >= stop_angle) {
            stop_angle += 360.0;
        }

        // fill arc uses notion of start and sweep angles
        arc_angle = (int) GaugeUtil.arcAngle(start_angle, stop_angle);

    }

	// Draw an image if there is one
    if (image != null) {
        if (scaleImage) {
            Shape oldClip = gc.getClip();
            // set the scale clip region
            Shape clip = getCircularScaleShape(this);
            gc.setClip(clip);
            // draw image in the clip region
            drawImage(gc, image, bounds);
            // reset the previous clip region
            gc.setClip(oldClip);
        }
        else {
            drawImage(gc, image, bounds);
        }
    }
	else {  // draw the arc
        gc.fillArc(x, y, w, h, (int) start_angle, arc_angle);
    }


 





 




 

    // clean up
    gc.dispose();
}


 


 

/**
 * Given a screen position in pixels, returns the closest scale value.
 *
 * @param p the point representing the <i>X</i>, <i>Y</i> coordinates of the
 *     click
 * @return the value closest to the point
 */
public double pick(Point p)
{
	double              value 		    = 0;
    double              tmpStartAngle   = GaugeUtil.normalizeAngle(startAngle);
    double              tmpStopAngle    = GaugeUtil.normalizeAngle(stopAngle);
    boolean             reversed        = direction.equals(Direction.BACKWARD);
    JCCircularGauge     cgauge          = getCircularGauge();
    Rectangle           bounds          = cgauge.getArcBounds();

	// For the special case when start == stop
	if (tmpStartAngle == tmpStopAngle) {
		tmpStopAngle  += 360.0;
	}

	// get the adjusted x, y
	int x = p.x - bounds.x;
    int y = p.y - bounds.y;

   	// get the angle to the click
	double angle = GaugeUtil.pointToAngle(x, y, getRadius(), getRadius());

    // adjust if necessary
    double original_angle = angle;
	if (angle < tmpStartAngle) {
	 	angle += 360.0;
	}

    // get the value
    value = GaugeUtil.angleToValue(angle, min, max, tmpStartAngle, tmpStopAngle);

    // adjust value according to direction
    if (reversed) {
        value  = max - value + min;
    }

    // If out of bounds, attempt to set value nearest the bound
    // where its out.
    if (!inBounds(value)) {
        double distance_to_min = GaugeUtil.distanceBetweenAngles(original_angle,
                                                                 tmpStartAngle);
        double distance_to_max = GaugeUtil.distanceBetweenAngles(original_angle,
                                                                 tmpStopAngle);
        if (reversed) {
            value = (distance_to_min <= distance_to_max) ? max : min;
        }
        else {
            value = (distance_to_min <= distance_to_max) ? min : max;
       }
    }

	// finally, return the value
    return(value);
}

/**
 * Since the scale and the <code>gaugeArea</code> share the same real estate,
 * their insets may overlap (note that the scale's border will be drawn
 * overtop of the <code>gaugeArea</code>'s border).  Thus we need to return an
 * <code>Insets</code> object which is the union of both.
 *
 * @return the scale/gaugeArea insets
 */
protected Insets getScaleInsets()
{
	Insets      insets = getInsets();
    Container   parent = getParent();       // GaugeArea is parent

    int top    = insets.top;
    int left   = insets.left;
    int bottom = insets.bottom;
    int right  = insets.right;
    if (parent != null) {
        Insets parent_insets = parent.getInsets();
        top    = Math.max(parent_insets.top,    top);
        left   = Math.max(parent_insets.left,   left);
        bottom = Math.max(parent_insets.bottom, bottom);
        right  = Math.max(parent_insets.right,  right);
    }

    return(new Insets(top, left, bottom, right));
}

/**
 * Returns the current radius of this scale. The radius is calculated
 * as half the width of gauge area. <br><B>Note:</b> The radius will be "scaled"
 * by the scale's <code>zoomFactor</code> so the actual size may be smaller than the
 * gauge area.
 *
 * @return the radius of the circular scale
 */
public double getRadius()
{
	int width		= getGauge().getGaugeArea().getBounds().width;
	int height		= getGauge().getGaugeArea().getBounds().height;
	Insets insets	= getScaleInsets();

	width  -= insets.left + insets.right;
	height -= insets.top  + insets.bottom;

    // for all quarter circles
    double radius = Math.max(minSize, Math.min(width, height));

	// for the full circle, the radius is half the widest side
    JCCircularGauge.GaugeType gaugeType = getCircularGauge().getGaugeType();
	if (gaugeType.equals(JCCircularGauge.GaugeType.FULL_CIRCLE)) {
		radius =  radius / 2;
	}

    // for the top/bottom half circles, find the largest rectangle to fit
    // the width
    else if (	gaugeType.equals(JCCircularGauge.GaugeType.TOP_HALF_CIRCLE) ||
    			gaugeType.equals(JCCircularGauge.GaugeType.BOTTOM_HALF_CIRCLE))
    {
    	while (width / 2 > height) {
    		width--;
    	}
    	radius = width / 2;
    }

    // for the left/right half circles, find the largest rectangle to fit
    // the height
    else if (	gaugeType.equals(JCCircularGauge.GaugeType.LEFT_HALF_CIRCLE) ||
    			gaugeType.equals(JCCircularGauge.GaugeType.RIGHT_HALF_CIRCLE))
    {
    	while (height / 2 > width) {
    		height--;
    	}
    	radius = height / 2;
    }

    // Adjust for zoomFactor
    radius *= zoomFactor;

	return(radius);
}

/**
 * Returns the start angle for this scale.
 *
 * @return the start angle for this scale in degrees
 */
public double getStartAngle()
{
	return(this.startAngle);
}

/**
 * Sets the start angle for this scale.
 *
 * @param startAngle the start angle for this scale, in degrees
 */
public void setStartAngle(double startAngle)
{
	this.startAngle = startAngle;
    redraw();
}

/**
 * Returns the stop angle for this scale.
 *
 * @return the stop angle for this scale, in degrees
 */
public double getStopAngle()
{
	return(this.stopAngle);
}

/**
 * Sets the stop angle for this scale.
 *
 * @param stopAngle the stop angle for this scale, in degrees
 */
public void setStopAngle(double stopAngle)
{
	this.stopAngle = stopAngle;
    redraw();
}

/**
 * Returns the circular gauge associated with this scale.
 *
 * @return the circular gauge associated with this scale
 */
public JCCircularGauge getCircularGauge()
{
	return((JCCircularGauge)getGauge());
}

/**
 * Paints the complete background defined by the <code>GaugeType</code>,
 * not just area swept by scale. For instance, you can have a
 * scale defined from 0&#176; to 270&#176; for a <code>FULL_CIRCLE</code>, yet
 * have the full circle background painted, not just from 0&#176; to 270&#176;.
 * The default is <code>false</code>.
 *
 * @return <code>true</code> means paint all the background; <code>false</code>
 *     means just the area spanned by scale's sweep angle
 */
public boolean getPaintCompleteBackground()
{
	return(paintCompleteBackground);
}

/**
 * If <code>true</code>, paints the complete background for this
 * <code>GaugeType</code>, not just area between the scale's start and stop
 * angles. For instance, a scale defined from 0&#176; to 270&#176; and having
 * gauge type <code>FULL_CIRCLE</code> has the full circular background painted,
 * not just the part from 0&#176; to 270&#176;. The default is
 * <code>false</code>.
 *
 * @param paintCompleteBackground <code>true</code> means paint all of the
 *     background consistent with the gauge type
 */
public void setPaintCompleteBackground(boolean paintCompleteBackground)
{
	this.paintCompleteBackground = paintCompleteBackground;
    redraw();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
    if (gauge.getRepaintEnabled()) {
        Enumeration e = gauge.getTicks().elements();
        while(e.hasMoreElements()) {
            JCCircularTick tick = (JCCircularTick) e.nextElement();
            tick.redraw(true);
        }
        repaint();
    }
}

/*
* Returns the shape of the circular scale.
* @param cscale computes the shape of this <code>JCCircularScale</code>
* @return the shape of <code>cscale</code>
*/
protected Shape getCircularScaleShape(JCCircularScale cscale)
{
	double start_angle	= GaugeUtil.normalizeAngle(cscale.getStartAngle());
    double stop_angle	= GaugeUtil.normalizeAngle(cscale.getStopAngle());
    JCCircularGauge cgauge = cscale.getCircularGauge();

    // Compute the start and stop angle
    int arc_angle;
    if (cscale.getPaintCompleteBackground()) {
        start_angle = cgauge.getGaugeType().getStartAngle();
        arc_angle = cgauge.getGaugeType().getSweepAngle();
    }
    else {
        if (start_angle >= stop_angle) {
            stop_angle += 360.0;
        }
        arc_angle = (int)GaugeUtil.arcAngle(start_angle, stop_angle);
    }

    Shape shape = null;
    if ((cgauge.getGaugeType() == JCCircularGauge.GaugeType.FULL_CIRCLE) &&
        cscale.getPaintCompleteBackground()) {
        Rectangle r = cgauge.getArcBounds();
        shape = new Ellipse2D.Float(r.x, r.y, r.width, r.height);
    }
    else { // polygon contour
	    // make a polar coordinates polygon for the scale contour
	    double radius_size	= cscale.getRadius();
        Polygon op = new Polygon();
        double df = 5;
        double angle = start_angle;
        while (angle < stop_angle) {
            op.addPoint((int)radius_size, (int)angle);
            angle += df;
        }
        op.addPoint((int)radius_size, (int)stop_angle);

        // Transform to a (x,y) polygon
        // get the center of the circle
        Rectangle bounds = cgauge.getArcBounds();
        double xOrigin   = bounds.x + radius_size;
        double yOrigin   = bounds.y + radius_size;

        Polygon xyContour = new Polygon();
        xyContour.addPoint((int)xOrigin, (int)yOrigin);
        for (int i = 0; i < op.npoints; i++) {
            double radians = GaugeUtil.toRadians((double)op.ypoints[i]);
            int x = (int)(xOrigin + op.xpoints[i] * Math.cos(radians));
            int y = (int)(yOrigin - op.xpoints[i] * Math.sin(radians));
            xyContour.addPoint(x, y);
        }
        xyContour.addPoint((int)xOrigin, (int)yOrigin);

        shape = xyContour;
    }
    return(shape);
}

/*
* Draws an image as the scale.
*
* @param g the Graphics on which to draw
*/

protected void drawImage(Graphics g, Image image, Rectangle bounds)
{
	// create an image observer
    java.awt.image.ImageObserver observer = new javax.swing.JLabel();

    // draw the image
    if (super.getScaleImage()) {
	    g.drawImage(image, bounds.x, bounds.y, bounds.width, bounds.height,
	                getForeground(), observer);
    }
    else {
        g.drawImage(image, bounds.x, bounds.y, getForeground(), observer);
    }
}



 

}
