/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCCircularTick.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import com.klg.jclass.util.JCNumberUtil;

import javax.swing.*;
import java.awt.*;

/**
 * The class for tick marks that are drawn on a <code>JCCircularScale</code>.
 */
public class JCCircularTick extends JCAbstractTick
                            implements java.io.Serializable
{

static final long serialVersionUID = -5247737898082678005L;

/**
 * Creates a <code>JCCircularTick</code> that spans its circular scale. Other
 * tick properties have default values.
 *
 * @param scale the <code>JCCircularScale</code> for which these ticks are
 *     drawn
 */
public JCCircularTick(JCCircularScale scale)
{
	super(scale, scale.getMin(), scale.getMax());

    // set the layout
	setLayout(new RadialLayout());
}

/**
 * Creates a <code>JCCircularTick</code> by specifying its properties.
 *
 * @param scale the <code>JCCircularScale</code> for which these ticks are
 *     drawn
 * @param automatic whether to automatically calculate the tick start, stop,
 *      and increment values from the scale's min and max, or to use the
 *      provided values
 * @param startValue the value on the scale where these ticks start
 * @param stopValue the value on the scale where these ticks end
 * @param incrementValue the tick interval
 * @param precisionUseDefault if <code>true</code>, gauge will determine an
 *     appropriate precision automatically; if <code>false</code>, "precision"
 *     will be used
 * @param precision The format of the tick label: positive values denote the
 *     number of places after the decimal point (for example, 3 means multiples
 *     of 0.001); negative values indicate the positive number of zeros to be
 *     used before the decimal place (for example, -3 means numbering will be in
 *     multiples of 1000). Overridden if <code>precisionUseDefault</code> is
 *     set to <code>true</code>.
 * @param width the width of the tick to be drawn
 * @param drawLabels if <code>true</code>, draw labels for each tick; if
 *     <code>false</code>, do not draw labels
 * @param drawTicks if <code>true</code>, draw ticks; if <code>false</code>, do
 *     not draw ticks (you may want only the labels)
 * @param labelExtent a ratio of the radius at which the label will be drawn;
 *      for example, 0.8 means draw labels 80% of the radius size away from the
 *      center
 * @param innerExtent the inner extent of each tick drawn as a ratio of the
 *     radius size; for example, 0.8 means the tick will start at 80% of the
 *     radius size away from the center and be drawn to <code>outerExtent</code>
 * @param outerExtent the outer extent of each tick drawn as a ratio of the
 *     radius size; for example, 1.1 means start at inner extent and draw to a
 *     location 110% the size of the radius away from the center
 * @param tickColor the color the tick is drawn in
 * @param style The <code>JCTickStyle</code> that defines the shape of the tick;
 *      valid values are <code>DIAMOND</code>, <code>TRIANGLE</code>,
 *     <code>REVERSE_TRIANGLE</code>, <code>RECTANGLE</code>,
 *     <code>LINE</code>, and <code>CIRCLE</code>. You can also create your
 *     own <code>JCTickStyle</code>.
 * @param font the font used to draw tick labels
 * @param fontColor the color of the font for tick labels
 */
public JCCircularTick(JCCircularScale 	scale,
                      boolean			automatic,
                      double			startValue,
                      double			stopValue,
                      double			incrementValue,
                      boolean			precisionUseDefault,
                      int 				precision,
                      double			width,
                      boolean			drawLabels,
                      boolean			drawTicks,
                      double			labelExtent,
                      double			innerExtent,
                      double			outerExtent,
                      Color				tickColor,
                      JCTickStyle		style,
                      Font				font,
                      Color				fontColor)
{
	super(scale,
          automatic,
          startValue,
          stopValue,
          incrementValue,
          precisionUseDefault,
          precision,
          width,
          drawLabels,
          drawTicks,
          labelExtent,
          innerExtent,
          outerExtent,
          tickColor,
          style,
          font,
          fontColor);

	// set the layout
	setLayout(new RadialLayout());
}

/**
 * Returns the circular scale associated with this tick object.
 *
 * @return the circualr scale associated with this tick object
 */
public JCCircularScale getCircularScale()
{
	return((JCCircularScale)getScale());
}

/**  */
public Dimension getPreferredSize()
{
	return(getCircularScale().getPreferredSize());
}

/**
 * Overridden to preserve the layout of labels on a resize.
 * Moves and resizes this component.
 *
 * @param x  the new horizontal location
 * @param y  the new vertical location
 * @param w  the new width
 * @param h  the new height
 * @see java.awt.Component#setBounds
 */
public void reshape(int x, int y, int w, int h)
{
    super.reshape(x, y, w, h);
    redrawLabels();
}

/**
 * If <code>automatic</code>, this method is called to determine and
 * set the start/stop and increment values for laying out ticks.
 * The goal is to have as many ticks/labels as will fit without
 * the ticks or labels overlapping.
 *
 * @param g the graphics used to get font metric information
 */
public void calcCircularTicks(Graphics g)
{
    JCCircularScale     cscale = getCircularScale();

	// Initialize
	double start_angle	= GaugeUtil.normalizeAngle(cscale.getStartAngle());
	double stop_angle	= GaugeUtil.normalizeAngle(cscale.getStopAngle());
	double radius_size	= cscale.getRadius();

	// Handle case where start_angle > stop_angle
	if (start_angle >= stop_angle) {
		stop_angle += 360;
	}

	// Caculate width (in pixels) along the circumference
	double diff =
        (stop_angle == start_angle) ? 360.0
                                    : Math.abs(stop_angle - start_angle);
	int width = (int) (GaugeUtil.toRadians(diff) * radius_size);
    calcTicks(g, width);
}

/**
 *  Layout and redraw labels.
 */
public void redrawLabels()
{
    JCCircularScale             cscale    = getCircularScale();
	JCAbstractScale.Direction   direction = cscale.getDirection();

    // If we are drawing labels ...
    if (drawLabels) {

        // Remove all the old labels
        removeAll();

        // Get info from CircularScale
        double circ_min = cscale.getMin();
        double circ_max	= cscale.getMax();

        // Get the scale info
        double circ_start  = GaugeUtil.normalizeAngle(cscale.getStartAngle());
        double circ_stop   = GaugeUtil.normalizeAngle(cscale.getStopAngle());

		// Handle case where startAngle > stop angle
		if (circ_start >= circ_stop) {
			circ_stop += 360;
		}

        // Detect full circle case with start == stop
       	boolean start_equal_stop = (circ_start == circ_stop);

        // Calculate the ticks
        if (automatic) {
        	Graphics g = getGraphics();
        	if (g != null) {
            	calcCircularTicks(g);
        	}
        }

        // Get the label extent
        double label_extent = Math.max(0.0, labelExtent);

        // Loop and position the labels
        double value = startValue;
        while (value <= stopValue && value <= circ_max && incrementValue != 0) {
            if (value < circ_min) {
                value += incrementValue;
                continue;
            }

	        // Deal with full circle case with start == stop
            if (start_equal_stop && automatic &&
                (value == stopValue ||
                 (1.0 + Math.abs(stopValue - value) == 1.0)))
            {
                break;
            }
            double value_used = value;
            if (direction.equals(JCAbstractScale.Direction.BACKWARD)) {
        		value_used = circ_max - value + circ_min;
	        }

			// Assume the direction is forward and startAngle = 0. This will
			// be massaged at layout time.
            double angle =
                GaugeUtil.valueToAngle(value_used, circ_min, circ_max, 0,
                                       GaugeUtil.arcAngle(circ_start,
                                                          circ_stop));

            // Finally, do the label
            JComponent label = null;

            // Get the angle
            RadialConstraint constraint =
                new RadialConstraint(cscale.getCircularGauge(), label_extent,
                                     angle);

            // Allow the user to define a label
            if (labelGenerator != null) {
                label = labelGenerator.makeLabel(this, value, constraint);
            }

            // Otherwise create a JLabel for them
            else {
				// Format the value according to precision
				String sValue = JCNumberUtil.format(value, precision);
            	label = new JLabel(sValue);

	            // Set the labels attributes
    	    	label.setFont(getFont());
        	    label.setForeground(fontColor);
			}

            label.setSize(label.getPreferredSize());

            // and finally, add it to the layout
            add(label, constraint);

            value += incrementValue;
        }
    }
}

/**
 * Main tick draw routine.
 *
 * @param g The graphics object with which to draw
 */
public void paintComponent(Graphics g)
{
    boolean start_equal_stop = false;
    JCCircularScale cscale = getCircularScale();
    JCAbstractScale.Direction direction = cscale.getDirection();
    Rectangle bounds = cscale.getCircularGauge().getArcBounds();

	super.paintComponent(g);

	// calculate the ticks, this can potentially change start, stop
	// and increment values, so do it first
	if (automatic) {
		calcCircularTicks(g);
	}

	// get info from CircularScale
	double circ_min = cscale.getMin();
	double circ_max	= cscale.getMax();

	//	Determine the inner and outer extents (in pixels).  Note that
	//	for circular ticks, the extents must be non-negative.

	double inner_extent = Math.max(0.0, innerExtent);
	double outer_extent = Math.max(0.0, outerExtent);

    // Do we need to draw anything?
    if (!isVisible()                    ||
        circ_min >= circ_max            ||
        inner_extent >= outer_extent    ||
        startValue > stopValue          ||
        incrementValue <= 0.0           ||
        tickWidth <= 0)
    {
        return;
    }

    // get the center of the circle
    double radius_size = cscale.getRadius();
	double x_orig      = bounds.x + radius_size;
	double y_orig      = bounds.y + radius_size;

	double circ_start  = GaugeUtil.normalizeAngle(cscale.getStartAngle());
	double circ_stop   = GaugeUtil.normalizeAngle(cscale.getStopAngle());

	double scaled_inner_extent = inner_extent * radius_size;
	double scaled_outer_extent = outer_extent * radius_size;

	double value = startValue;
	while (value <= stopValue && value <= circ_max) {
		if (value < circ_min) {
			value += incrementValue;
			continue;
		}

		if (start_equal_stop && automatic &&
			(value == stopValue ||
			 (1.0 + Math.abs(stopValue - value) == 1.0)))
		{
			break;
		}

		// if there is something to draw ...
		if (drawTicks) {

			// draw a simple line
			if (tickStyle == JCTickStyle.LINE) {
                double value_used = value;
		        if (direction.equals(JCAbstractScale.Direction.BACKWARD)) {
			        value_used = circ_max - value + circ_min;
		        }

                // compute the angle
		        double angle = GaugeUtil.valueToAngle(value_used, circ_min,
                                                      circ_max, circ_start,
                                                      circ_stop);

                // use it to get the needed sine and cosine values
                double radians = GaugeUtil.toRadians(angle);
                double cosr = Math.cos(radians);
                double sinr = Math.sin(radians);

            	Graphics gc = g.create();
                gc.setColor(tickColor);

				int x_inner = (int)x_orig + (int)(scaled_inner_extent * cosr);
				int y_inner = (int)y_orig - (int)(scaled_inner_extent * sinr);
				int x_outer = (int)x_orig + (int)(scaled_outer_extent * cosr);
				int y_outer = (int)y_orig - (int)(scaled_outer_extent * sinr);

				gc.drawLine(x_inner, y_inner, x_outer, y_outer);
                gc.dispose();
			}
			// a circle
            else if (tickStyle == JCTickStyle.CIRCLE) {
                GaugeUtil.drawCircleForCircularScale(g, cscale, tickColor,
                                                     value, inner_extent,
                                                     outer_extent);
            }
			// or a polygon
			else {
                GaugeUtil.drawCircularPolygon(g, cscale, tickColor, inner_extent,
                                              outer_extent, tickWidth, value,
                                              reversed, true, tickStyle);
			}
		}

		value += incrementValue;
	}
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 *
 * @param redrawLabels <code>true</code> means reposition and redraw the labels; otherwise,
 *      just repaint the component
 */
public void redraw(boolean redrawLabels)
{
    if (getCircularScale().getGauge().getRepaintEnabled()) {
        if (redrawLabels) {
            redrawLabels();
        }

        revalidate();

        // revalidate is supposed to call repaint at some point, but doesn't
        // always, so .. ("Graphic Java 2", p 144)
        repaint();
    }
}

} // JCCircularTick
