/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCGauge.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import com.klg.jclass.swing.JCVersion;
import com.klg.jclass.util.JCListenerList;
import com.klg.jclass.util.legend.JCGridLegend;
import com.klg.jclass.util.legend.JCLegend;
import com.klg.jclass.util.legend.LegendComponentLayout;
import com.klg.jclass.util.legend.LegendComponentLayoutUser;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Vector;

/**
 * The parent class for linear and circular gauges. It contains methods for
 * setting many of the gauge's properties, such as a header, footer, and legend,
 * as well as for adding indicators, needles, ranges, and tick objects.
 *
 */
public abstract class JCGauge extends JComponent
                              implements MouseListener, MouseMotionListener,
                                         LegendComponentLayoutUser,
                                         java.io.Serializable
{

static final long serialVersionUID = 1763099717300034832L;

/**
 * The drawing area for the circular gauge.
 */
protected JCGaugeArea gaugeArea;

/**
 * The header component.
 */
protected JComponent header;

/**
 * The footer component.
 */
protected JComponent footer;

/**
 * The legend.
 *
 * @see com.klg.jclass.util.legend.JCLegend
 */
protected JCLegend legend;

/**
 * The scale associated with this gauge.
 */
protected JCScale scale;

/**
 * The list of indicators assocated with this gauge.
 */
protected Vector indicators;

/**
 * The list of needles assocated with this gauge.
 */
protected Vector needles;

/**
 * List of listeners interested in pick events.
 */
protected JCListenerList pickListeners = null;

/**
 * The needle currently being dragged.
 */
protected JCNeedle draggedNeedle = null;

/**
 * Snaps to closest integral scale value. The default is <code>false</code>.
 */
protected boolean snapToValue = false;

/**
 * Allows batching of changes by toggling repaint off/on.
 */
protected boolean repaintEnabled = true;

/**  */
protected Rectangle gaugeAreaLayoutHints = new Rectangle(Integer.MAX_VALUE,
                                                         Integer.MAX_VALUE,
                                                         Integer.MAX_VALUE,
                                                         Integer.MAX_VALUE);
/**  */
protected Rectangle legendLayoutHints = new Rectangle(Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE);
/**  */
protected Rectangle headerLayoutHints = new Rectangle(Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE);
/**  */
protected Rectangle footerLayoutHints = new Rectangle(Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE,
                                                      Integer.MAX_VALUE);

/**
* An image may be used for gauge.
*/
protected Image image = null;

/**
* Boolean that determines if the image should be matched when
* the gauge is resized. The default is <code>true</code>.
*/
protected boolean scaleImage = true;


/**
 * No-argument constructor for <code>JCCircularGauge</code>.
 */
public JCGauge()
{
	init();
}

/**
 * Initializes assorted gauge member variables. Used as a
 * convenience when writing a new <code>JCGauge</code> constructor.
 */
private void init()
{
	setOpaque(false);

    // set the default layout
	setLayout(new LegendComponentLayout());

    // create the header
    header = new JLabel();
	header.setVisible(false);
	add(header);

    // create the drawing area
    gaugeArea = new JCGaugeArea(this);
	add(gaugeArea);

    // and the footer
    footer = new JLabel();
	footer.setVisible(false);
	add(footer);

    // and finally the legend
    legend = new JCGridLegend();
    legend.setVisible(false);
	add(legend);

	// create the list for indicators
	indicators = new Vector();

	// create the list for needles
	needles = new Vector();

   	// listen for mouse events
   	addMouseListener(this);

   	// listen for mouse motion events
   	addMouseMotionListener(this);
}

/**
 * Returns the <code>snapToValue</code> property that controls whether the
 * needle should snap to the closest discrete integral scale value
 * (<code>true</code>) or to any scale value (<code>false</code>).
 *
 * @return <code>true</code> if discrete movements are enforced;
 *     <code>false</code> means continuous movement is allowed
 */
public boolean getSnapToValue()
{
	return(snapToValue);
}

/**
 * Sets the <code>snapToValue</code> property that controls whether the
 * needle should snap to the closest discrete integral scale value
 * (<code>true</code>) or to any scale value (<code>false</code>).
 *
 * @param snapToValue <code>true</code> for discrete movements;
 *     <code>false</code> means continuous movement is allowed
 */
public void setSnapToValue(boolean snapToValue)
{
	this.snapToValue = snapToValue;
}

/**
 * Adds a indicator to this gauge.
 *
 * @param indicator a <code>JCIndicator</code> to associate with this gauge
 */
public void addIndicator(JCAbstractIndicator indicator)
{
	indicators.addElement(indicator);
	gaugeArea.add(indicator);
}

/**
 * Adds a indicator to this gauge.
 *
 * @param indicator the indicator to be added
 * @param index the position in the list of indicators where this one
 *              is to be added
 */
public void addIndicator(JCAbstractIndicator indicator, int index)
{
	indicators.addElement(indicator);
	gaugeArea.add(indicator, index);
}

/**
 * Removes a indicator from this gauge.
 *
 * @param indicator the <code>JCIndicator</code> to be removed
 */
public void removeIndicator(JCAbstractIndicator indicator)
{
	if (indicators.contains(indicator)) {
    	indicators.removeElement(indicator);
    	gaugeArea.remove(indicator);
    }
}

/**
 * Returns the list of indicators associated with this gauge.
 *
 * @return the <code>Vector</code> of indicators for this gauge
 */
public Vector getIndicators()
{
    return(indicators);
}

/**
 * Sets the given scale on each of the gauge's indicators.
 *
 * @param scale the scale to be set on each indicator
 */
protected void setIndicatorScales(JCScale scale)
{
    if (indicators != null) {
        Enumeration e = indicators.elements();
        while (e.hasMoreElements()) {
            JCAbstractIndicator indicator =
                (JCAbstractIndicator)e.nextElement();
            indicator.setScale(scale);
        }
    }
}

/**
 * Adds a needle to this gauge.
 *
 * @param needle a <code>JCNeedle</code> to associate with this gauge
 */
public void addNeedle(JCAbstractNeedle needle)
{
	needles.addElement(needle);
	gaugeArea.add(needle);
}

/**
 * Adds a needle to this scale.
 *
 * @param needle the needle to be added
 * @param index the position in the list of needles where this one
 *              is to be added
 */
public void addNeedle(JCAbstractNeedle needle, int index)
{
	needles.addElement(needle);
	gaugeArea.add(needle, index);
}

/**
 * Removes a needle from this gauge.
 *
 * @param needle the <code>JCNeedle</code> to be removed
 */
public void removeNeedle(JCAbstractNeedle needle)
{
	if (needles.contains(needle)) {
    	needles.removeElement(needle);
    	gaugeArea.remove(needle);
    }
}

/**
 * Sets the given scale on each of the gauge's needles.
 *
 * @param scale the scale to be set on each needle
 */
protected void setNeedleScales(JCScale scale)
{
    if (needles != null) {
        Enumeration e = needles.elements();
        while (e.hasMoreElements()) {
            JCAbstractNeedle needle = (JCAbstractNeedle) e.nextElement();
            needle.setScale(scale);
        }
    }
}

/**
 * Returns the list of needles associated with this gauge.
 *
 * @return the <code>Vector</code> of needles for this gauge
 */
public Vector getNeedles()
{
    return(needles);
}

/**
 * Adds a tick to a gauge (actually added to the scale).
 *
 * @param tick the <code>JCTick</code> to be added
 */
public void addTick(JCTick tick)
{
	scale.addTick(tick);
}

/**
 * Adds a tick object to a gauge (actually added to the scale).
 *
 * @param tick the <code>JCTick</code> to be added
 * @param index the <i>z</i>-order index
 */
public void addTick(JCTick tick, int index)
{
	scale.addTick(tick, index);
}

/**
 * Removes a tick from a gauge (actually removed from the scale).
 *
 * @param tick the <code>JCTick</code> to be removed
 */
public void removeTick(JCTick tick)
{
	scale.removeTick(tick);
}

/**
 * Returns the tick objects for this gauge.
 *
 * @return a <code>Vector</code> of tick objects
 */
public Vector getTicks()
{
	return(scale.getTicks());
}

/**
 * Adds a range to the gauge (actually added to the scale).
 *
 * @param range the <code>JCRange</code> to be added
 */
public void addRange(JCRange range)
{
	scale.addRange(range);
}

/**
 * Adds a range to the gauge (actually added to the scale).
 *
 * @param range the <code>JCRange</code> to be added
 * @param index the <i>z</i>-order index
 */
public void addRange(JCRange range, int index)
{
	scale.addRange(range, index);
}

/**
 * Removes a range (actually removed from the scale).
 *
 * @param range the <code>JCRange</code> to be removed
 */
public void removeRange(JCRange range)
{
    scale.removeRange(range);
}

/**
 * Returns the list of ranges.
 *
 * @return the <code>Vector</code> of ranges for this gauge
 */
public Vector getRanges()
{
	return(scale.getRanges());
}

/**
 * Returns the scale associated with this gauge.
 *
 * @return the scale for the <code>JCCircularGauge</code> component.
 */
public JCScale getScale()
{
	return(scale);
}

/**
 * Returns the legend for this gauge.
 *
 * @return the legend for this gauge
 */
public JCLegend getLegend()
{
	return(legend);
}

/**
 * Sets the <code>Legend</code> property.
 * The <code>Legend</code> property controls the object that controls
 * the display of the legend.
 *
 * @param newLegend the legend for this gauge
 */
public void setLegend(JCLegend newLegend)
{
    if (newLegend == null) {
        return;
    }
	if (legend != null) {
		remove(legend);
    }
    legend = newLegend;
	add(legend);
}

/**
 * Returns the gauge area for this gauge.
 *
 * @return the <code>JCGaugeArea</code> for this gauge
 */
public JCGaugeArea getGaugeArea()
{
	return(gaugeArea);
}

/**
 * Sets the value of the <code>gaugeArea</code> property.  The
 * <code>gaugeArea</code> property controls the object that controls the
 * display of the gauge.
 *
 * @param newGaugeArea
 */
public void setGaugeArea(JCGaugeArea newGaugeArea)
{
    if (newGaugeArea == null) {
        return;
    }
	if (gaugeArea != null) {
		remove(gaugeArea);
    }
    gaugeArea = newGaugeArea;
	add(gaugeArea);
}

/**
 * Disable/enable repaints of gauge and its components. This allows for
 * batched updates. A repaint is issued when set to <code>true</code>.
 *
 * @param repaintEnabled <code>true</code> enables repaints,
 *     <code>false</code> disables them
 */
public void setRepaintEnabled(boolean repaintEnabled)
{
    this.repaintEnabled = repaintEnabled;

    // repaint when true
    if (repaintEnabled) {
        repaint();
    }
}

/**
 * Disable/enable repaints of gauge and its components. This allows for
 * batched updates.
 *
 * @return repaintEnabled <CODE>true</CODE> enables repaints;
 *     <code>false</code> disables them
 */
public boolean getRepaintEnabled()
{
    return(repaintEnabled);
}

/**
 * Given a screen location in pixels, returns the closest
 * scale value wrapped in a <code>JCGaugePickEvent</code>.
 *
 * @param p the <code>Point</code> at which the event occurred
 * @return a <code>JCGaugePickEvent</code> which wraps the closest value
 */
public JCGaugePickEvent pick(Point p)
{
	JCGaugePickEvent event = null;

	Component comp = getComponentAt(p.x, p.y);
	if (comp == null) {
		return(null);
	}

	// adjust for location
    Point pp = new Point(p.x - comp.getLocation().x,
    					 p.y - comp.getLocation().y);

	if (comp instanceof JCLegend) {
		// return legend.pick(pp, null);
	}
    else if (comp instanceof JCGaugeArea) {
    	double value = scale.pick(pp);

        // if discrete movements required, snap to integer value
        if (snapToValue) {
        	value = Math.round(value);
        }

		event = new JCGaugePickEvent(this, comp, value, pp);
        if (!scale.inBounds(event.getValue())) {
        	event = null;
        }
	}

    return(event);
}

/**
 * Adds a listener for gauge pick events.
 *
 * @param l
 */
public void addPickListener(JCGaugePickListener l)
{
    pickListeners = JCListenerList.add(pickListeners,l);
}

/**
 * Removes a listener to gauge pick events.
 *
 * @param l
 */
public void removePickListener(JCGaugePickListener l)
{
    pickListeners = JCListenerList.remove(pickListeners,l);
}

/**
 * Broadcasts the pick event to interested listeners.
 *
 * @param ev the <code>JCGaugePickEvent</code> of interest
 */
public void sendPickEvent(JCGaugePickEvent ev)
{
	Enumeration e = JCListenerList.elements(pickListeners);
    while (e.hasMoreElements()) {
        ((JCGaugePickListener)e.nextElement()).pick(ev);
    }
}

///////////// implement the LegendComponentLayoutUser interface

/**
 * Returns the header for this gauge.
 *
 * @return the header for this gauge
 */
public JComponent getHeader()
{
	return(header);
}

/**
 * Sets the <code>Header</code> property.
 * The <code>Header</code> property controls the object that controls the
 * display of the header.
 *
 * @param newHeader the <code>JComponent</code> that is used as a header for
 *     the gauge
 */
public void setHeader(JComponent newHeader)
{
    if (newHeader == null) {
        return;
    }
	if (header != null) {
		remove(header);
    }
    header = newHeader;
	add(header);
}

/**
 * Returns the footer for this gauge.
 *
 * @return the footer for this gauge
 */
public JComponent getFooter()
{
	return(footer);
}

/**
 * Sets the <code>Footer</code> property.
 * The <code>Footer</code> property controls the object that controls the
 * display of the footer.
 *
 * @param newFooter the <code>JComponent</code> that is used as a footer for
 *     the gauge
 */
public void setFooter(JComponent newFooter)
{
    if (newFooter == null) {
        return;
    }
	if (footer != null) {
		remove(footer);
    }
    footer = newFooter;
	add(footer);
}

/**
 * Gets the height of the drawing area represented by this gauge.
 * @return
 */
public int getDrawingAreaHeight()
{
    int     k      = 0;
	Insets  insets = getInsets();

	if (insets != null) {
    	k = insets.top + insets.bottom;
    }
    return(Math.max(0, getSize().height - k));
}

/**
 * Gets the width of the drawing area represented by this gauge.
 * @return
 */
public int getDrawingAreaWidth()
{
    int     k      = 0;
	Insets  insets = getInsets();

	if (insets != null) {
		k = insets.left + insets.right;
    }
    return(Math.max(0, getSize().width - k));
}

/**
 * Gets layout hints for the specified gauge child.
 *
 * @param child gauge child: either the <code>gaugeArea</code>, <code>legend</code>,
 * <code>header</code>, or <code>footer</code>
 * @return Rectangle object containing the desired layout hints; null is
 *     child, and is not one of the valid children of gauge
 */
public Rectangle getLayoutHints(JComponent child)
{
	if (child == gaugeArea) {
		return(gaugeAreaLayoutHints);
    }
	else if (child == legend) {
		return(legendLayoutHints);
    }
	else if (child == header) {
		return(headerLayoutHints);
    }
	else if (child == footer) {
		return(footerLayoutHints);
    }

	return(null);
}

/**
 * Returns the JClass component's sub-component on which the actual
 * drawing of the gauge will be performed.
 *
 * @return the <code>JCComponent</code> on which to draw
 */
public JComponent getComponentArea()
{
	return(getGaugeArea());
}


 
 
/**
 * Gets the <code>Image</code> to be used for the gauge background.
 *
 * @return the image that is used for the gauge background
 */
public Image getImage()
{
	return(image);
}

/**
 * Sets the <code>Image</code> to be used for the gauge background.
 *
 * @param image the <code>Image</code> to be used for the gauge background
 */
public void setImage(Image image)
{
	this.image = image;
    redraw();
}

/**
 * Scales the image in relation to the scale dimensions.
 *
 * @return <code>true</code> if the image is to be matched to the size of the bounding box
 *  of the gauge; otherwise, center the image based on its length and width
 */
public boolean getScaleImage()
{
	return(scaleImage);
}

/**
 * Determines whether to adapt the image according to the gauge dimensions.
 *
 * @param scaleImage <code>true</code> if the image is to be matched to the size of the bounding
 *	box of the gauge; otherwise, center the image based on its length and width
 */
public void setScaleImage(boolean scaleImage)
{
	this.scaleImage = scaleImage;
    redraw();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>. Method
 * must be overriden in subclasses.
 */
public abstract void redraw();
 
//////////// implement (empty) mouse listeners to be overridden

////// mouse events

/**  *
 * @param event
 */
public void mouseEntered(MouseEvent event) 	{}
/**  *
 * @param event
 */
public void mouseExited(MouseEvent event) 	{}
/**  *
 * @param event
 */
public void mouseReleased(MouseEvent event)
{
	// cannot be dragging if the mouse is released
    draggedNeedle = null;

}
/**  *
 * @param event
 */
public void mousePressed(MouseEvent event) 	{}

/**
 * Sends pick events to listeners.
 *
 * @param event
 */
public void mouseClicked(MouseEvent event)	{

	JCGaugePickEvent pe = pick(event.getPoint());
	sendPickEvent(pe);
}

////// mouse motion

/**  */
public void mouseMoved(MouseEvent event) 	{}
/**  */
public void mouseDragged(MouseEvent event)
{
	JCGaugePickEvent pe = pick(event.getPoint());
	sendPickEvent(pe);
}

/**
 * Gets the value of the <code>About</code> property.
 * The <code>About</code> property displays contact information for
 * Quest Software in the Bean box.
 * @return
 */
public String getAbout()
{
    return(JCVersion.getVersionString());
}

/**
 * Sets the value of the <code>About</code> property.
 * The <code>About</code> property displays contact information for
 * Quest Software in the Bean box.
 * @param s
 */
public void setAbout(String s) {}


 

}
