/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCLinearRange.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:29 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import java.awt.*;

public class JCLinearRange extends JCAbstractRange {

/**
* Constants used for the masking orientation of the background image.
*/
public static final int HORIZONTAL = 0;
public static final int VERTICAL = 1;

/**
* The foreground image used as range.
*/
protected Image foregroundImage;

/**
* The background image used as range.
*/
protected Image backgroundImage;

/**
* Percent of foreground image that covers background image.
*/
protected double foregroundCoverage = 0;

/**
* The coverage orientation.
*/
protected int maskOrientation = JCLinearRange.HORIZONTAL;

/**
 * Creates a linear range.
 *
 * @param foreground the range's color
 * @param scale the scale associated with this range
 * @param innerExtent the ratio of the scale's width/height to start drawing
 *      this range from the top/left of the scale (for example, for a
 *      horizontal scale, a value of 0.8 means start drawing this range 80% of
 *      the scale's height from the top of the scale)
 * @param outerExtent the ratio of the scale's width/height to stop drawing
 *      this range from the top/left of the scale (for example, for a
 *      vertical scale, a value of 0.9 means stop drawing this range 90% of
 *      the scale's width from the left of the scale)
 * @param startValue the scale value (not position) at which to begin the range
 * @param stopValue the scale value (not position) at which to stop the range
 */
public JCLinearRange(Color				foreground,
                     JCLinearScale		scale,
                     double				innerExtent,
                     double				outerExtent,
                     double				startValue,
                     double				stopValue)
{
	super(foreground,
          scale,
          innerExtent,
          outerExtent,
          startValue,
          stopValue);
}

/**
 * Returns the linear scale associated with this range.
 *
 * @return the linear scale associated with this range
 */
protected JCLinearScale getLinearScale()
{
	return((JCLinearScale)getScale());
}

/**
 * Returns the preferred size of this range, which will be
 * the preferred size of the scale.
 *
 * @return the preferred size
 */
public Dimension getPreferredSize()
{
	return(getLinearScale().getPreferredSize());
}

/**
* This routine computes the range rectangle to be drawn
* with respect to range's scale.
*
* @param lscale the <code>JCLinearScale</code> of this range
* @param bounds the linear scale bounds; use <code>lscale.getScaleSize()</code>
 *     for range painting and <code>lscale.getBounds()</code> for image mapping
* @return Rectangle defines the range position and dimensions
*/
public Rectangle getLinearRangeSize(JCLinearScale lscale, Rectangle bounds)
{
    JCLinearScale.Orientation orientation = lscale.getOrientation();
    JCAbstractScale.Direction direction = lscale.getDirection();

    // Get the various values
	double lin_min = lscale.getMin();
	double lin_max = lscale.getMax();

    // Make a local copy so they can be modified
    double start_value = GaugeUtil.clamp(startValue, lin_min, lin_max);
    double stop_value = GaugeUtil.clamp(stopValue,  lin_min, lin_max);

    // Adjust for direction and orientation
    if (direction.equals(JCAbstractScale.Direction.BACKWARD) ^
        (orientation == JCLinearScale.Orientation.VERTICAL)) {
            stop_value  = lin_max - stop_value + lin_min;
	    start_value = lin_max - start_value + lin_min;
	}

    // Limit range bounds to scale bounds for image drawing
    if ( backgroundImage != null || foregroundImage != null ) {
        if( innerExtent > 1) {
            innerExtent = 1;
        }
        if( outerExtent > 1) {
            outerExtent = 1;
        }
    }

    int x, x2, y, y2, w, h;
    if (orientation == JCLinearScale.Orientation.VERTICAL) {
        y  = GaugeUtil.valueToPosition(start_value, lin_min, lin_max,
                                       bounds.y, bounds.y + bounds.height);
        y2 = GaugeUtil.valueToPosition(stop_value, lin_min, lin_max,
                                       bounds.y, bounds.y + bounds.height);
        h = y2 - y;

        x  = bounds.x + (int)((double)bounds.width * innerExtent);
        x2 = bounds.x + (int)((double)bounds.width * outerExtent);
        w = x2 - x;

        if (direction.equals(JCAbstractScale.Direction.FORWARD)) {
            y = y2;
            h = -h;
        }
    }
    else {
        x  = GaugeUtil.valueToPosition(start_value, lin_min, lin_max,
                                       bounds.x, bounds.x + bounds.width);
        x2 = GaugeUtil.valueToPosition(stop_value, lin_min, lin_max,
                                       bounds.x, bounds.x + bounds.width);

        w = x2 - x;

        y  = bounds.y + (int)((double)bounds.height * innerExtent);
        y2 = bounds.y + (int)((double)bounds.height * outerExtent);
        h = y2 - y;

        if (direction.equals(JCAbstractScale.Direction.BACKWARD)) {
            x = x2;
            w = -w;
        }
    }
    return(new Rectangle(x, y, w, h));
}

/**
* Draws a linear range.
 * @param g The graphics object with which to draw
*/
public void paintComponent(Graphics g)
{
    JCLinearScale lscale = getLinearScale();

    // Get the various values
	double lin_min = lscale.getMin();
	double lin_max	= lscale.getMax();

    double start_value = GaugeUtil.clamp(startValue, lin_min, lin_max);
	double stop_value = GaugeUtil.clamp(stopValue, lin_min, lin_max);

    // Do we need to draw anything?
    if (!isVisible() ||
        lin_min >= lin_max ||
        innerExtent >= outerExtent ||
        start_value >= stop_value) {
        return;
    }

    Rectangle bounds = getLinearRangeSize(lscale, lscale.getScaleSize());
    Graphics gc = g.create();

    // Draw an image if there is one
    if ( backgroundImage != null || foregroundImage != null ) {
        drawImage(gc, bounds);
    }
    else {  // draw the range rectangle
        gc.fillRect(bounds.x, bounds.y, bounds.width, bounds.height);
    }

	gc.dispose();
}

/**
 * Call this instead of paint so that the <code>repaintEnabled</code> flag can
 * be checked. No repaint will occur if that flag is <code>false</code>.
 */
public void redraw()
{
	if (super.getScale().getGauge().getRepaintEnabled()) {
    	repaint();
	}
}

/**
* Sets the background image of the range. Overrides the superclass method.
* @param image the background image of the range
*/
public void setImage(Image image)
{
    super.image = image;
    backgroundImage = image;
    foregroundImage = null;
    redraw();
}

/**
* Sets the foreground image of the range.
* @param image the foreground image of the range
*/
public void setForegroundImage(Image image)
{
    foregroundImage = image;
    redraw();
}

/**
* Sets the background image of the range.
* @param image the background image of the range
*/
public void setBackgroundImage(Image image)
{
    this.backgroundImage = image;
    super.image = image;
    redraw();
}

/**
* Returns the background image of the range. Overrides the superclass method.
* @return image the bakcground image of the range
*/
public Image getImage()
{
    return(backgroundImage);
}

/**
* Returns the foreground image of the range.
* @return image the foreground image of the range
*/
public Image getForegroundImage()
{
    return(foregroundImage);
}

/**
* Returns the background image of the range.
* @return image the background image of the range
*/
public Image getBackgroundImage(){
    return backgroundImage;
}

/**
* Sets the foreground coverage of the background.
* @param coverage the coverage factor is a float between 0 and 1
*/
public void setForegroundCoverage(double coverage)
{
    foregroundCoverage = coverage;
    if (foregroundCoverage < 0){
        foregroundCoverage = 0;
    }
    if (foregroundCoverage > 1){
        foregroundCoverage = 1;
    }
    if (foregroundImage != null && backgroundImage != null) {
        redraw();
    }
}

/**
* Returns the foreground coverage of the background.
* @return float the foreground coverage of the range
*/
public double getForegroundCoverage()
{
    return(foregroundCoverage);
}

/**
 * Sets the mask orientation. There are  two possible values:
 * for <code>JCLinearRange.HORIZONTAL</code>, the foreground image covers the
 * left side of the background image;
 * for <code>JCLinearRange.VERTICAL</code>, the foreground image covers the
 * upper side of the background image.
 * @param orientation the mask orientation of the range
 */
public void setMaskOrientation(int orientation)
{
    if( orientation != JCLinearRange.HORIZONTAL &&
        orientation != JCLinearRange.VERTICAL)
    {
        return;
    }
    else {
        maskOrientation = orientation;
    }

    if (foregroundImage != null && backgroundImage != null) {
        redraw();
    }
}

/**
* Gets the mask orientation of the range.
* @return maskOrientation the mask orientation of the range
*/
public int getMaskOrientation()
{
    return(maskOrientation);
}

/**
* Draws an image as the range.
*
* @param g the Graphics on which to draw
* @param bounds the bounds of the image used for image scaling
*/
protected void drawImage(Graphics g, Rectangle bounds)
{
	// create an image observer
    java.awt.image.ImageObserver observer = new javax.swing.JLabel();

    // draw the image
    if( backgroundImage != null && foregroundImage != null) {
        if (scaleImage) {
            // draw background image
            g.drawImage(backgroundImage, bounds.x, bounds.y, bounds.width,
                        bounds.height, getForeground(), observer);

            Rectangle oldClip = (Rectangle)g.getClip();
            Rectangle newClip = null;
            switch (maskOrientation) {
            case HORIZONTAL:
                newClip = new Rectangle(bounds.x, bounds.y,
                                        (int)(bounds.width *
                                              foregroundCoverage),
                                        bounds.height);
                break;
            case VERTICAL:
                newClip = new Rectangle(bounds.x, bounds.y,
                                        bounds.width,
                                        (int)(bounds.height *
                                              foregroundCoverage));
                break;
            }

            g.setClip(newClip);

            // draw foreground image
            g.drawImage(foregroundImage, bounds.x, bounds.y, bounds.width,
                        bounds.height, getForeground(), observer);
            g.setClip(oldClip);

        }
        else {
            g.drawImage(foregroundImage, bounds.x, bounds.y, getForeground(),
                        observer);
        }

        return;
	}

    if (backgroundImage != null) {
        image = backgroundImage;
    }
    else if(foregroundImage != null) {
        image = foregroundImage;
    }

    if (scaleImage) {
        g.drawImage(image, bounds.x, bounds.y, bounds.width, bounds.height,
	                getForeground(), observer);
    }
    else {
        g.drawImage(image, bounds.x, bounds.y, getForeground(), observer);
    }
    return;
}

} // JCLinearRange


