/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCTick.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:30 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.gauge;

import java.awt.Color;

/**
 * The interface for tick objects.  Those wishing to use their own
 * tick objects should implement this interface.
 *
 * A graduated marks on a scale. A <code>JCTick</code> object
 * comprises a set of uniformly spaced tick marks.
 */
public interface JCTick {

/**
 * The <code>automatic</code> boolean determines whether tick increments are
 * calculated automatically or set by the <code>incrementValue</code> property.
 * If <code>true</code>, tick increments are calculated by the gauge.
 * If <code>false</code>, the <code>incrementValue</code> property is used to
 * set the tick spacing.
 *
 * @return <code>true</code> if tick increments are to be calculated
 *     automatically
 */
public boolean getAutomatic();

/**
 * Returns the Boolean controlling the drawing of labels. <code>true</code>
 * means draw the labels, <code>false</code> means hide the labels.
 *
 * @return the <code>boolean</code> controlling the drawing of labels
 */
public boolean getDrawLabels();

/**
 * Returns the Boolean controlling the drawing of ticks. <code>true</code>
 * means draw the ticks, <code>false</code> means hide the ticks.
 *
 * @return the <code>boolean</code> controlling the drawing of ticks
 */
public boolean getDrawTicks();

/**
 * Returns the color of the font labeling ticks.
 *
 * @return the font color
 */
public Color getFontColor();

/**
 * In non-automatic mode, use this value as the increment value; that
 * is, it places ticks at every <code>incrementValue</code> along the scale.
 * In automatic mode, the calculated increment value will overwrite this value.
 *
 * @return the spacing between adjacent tick marks
 */
public double getIncrementValue();

/**
 * Returns the percentage of the scale's radius (as a decimal)
 * at which to begin drawing ticks. The tick marks are drawn between
 * the inner extent and the outer extent.
 *
 * @return the inner extent of the tick mark
 */
public double getInnerExtent();

/**
 * Returns the position, expressed as a decimal fraction of the scale's size,
 * where tick labels are placed. The size parameter depends on what type of
 * scale is in use.
 *
 * @return the fraction representing the position at which labels are
 *     placed
 */
public double getLabelExtent();

/**
 * Returns the label generator associated with this tick.
 * If non-null, it will be called, and expected to
 * return a <code>JComponent</code> which will be the label for a tick mark.
 *
 * @return the <code>labelGenerator</code>
 */
public JCLabelGenerator getLabelGenerator();

/**
 * Returns the percentage of the scale's size parameter (as a decimal) at
 * which to stop drawing the ticks. The size depends on the type of scale.
 * For a circular scale it is its radius or for a linear scale, the width
 * or height, depending on its orientation.
 *
 * @return the outer extent of the tick mark
 */
public double getOuterExtent();

/**
 * Returns the format for numbers in a tick label. As implemented in
 * <code>JCAbstractTick</code>,  positive values denote the
 * number of places after the decimal point (for example, 3 means multiples
 * of 0.001); negative values indicate the positive number of zeros that are
 * used before the decimal place (for example, -3 means numbering will be in
 * multiples of 1000). Overridden if <code>precisionUseDefault</code> is set
 * to <code>true</code>.
 *
 * @return the precision code for formatting numbers in a tick label
 */
public int getPrecision();

/**
 * If <code>true</code>, uses the gauge-determined precision for tick labels;
 * otherwise, uses the current value of <code>precision</code> property to
 * label tick marks.
 *
 * @return a Boolean indicating whether or not a label's precision is
 *     calculated by the gauge or gotten from the <code>precision</code>
 *     property
 */
public boolean getPrecisionUseDefault();

/**
 * Returns the scale associated with this tick.
 *
 * @return the scale for the <code>JCTick</code> component
 */
public JCScale getScale();

/**
 * In non-automatic mode, usez this value as the start value.  In automatic
 * mode, the calculated start value will overwrite this value.
 *
 * @return the start value
 */
public double getStartValue();

/**
 * In non-automatic mode, usez this value as the stop value.  In automatic
 * mode, the calculated stop value will overwrite this value.
 *
 * @return the stop value
 */
public double getStopValue();

/**
 * Returns the color in which to draw ticks.
 *
 * @return the color of the tick mark
 */
public Color getTickColor();

/**
 * Returns the <code>JCTickStyle</code> controlling the shape of the tick.
 *
 * @return the tick style
 */
public JCTickStyle getTickStyle();

/**
 * Returns the width, in pixels, of the tick mark.
 *
 * @return the width of the tick to be drawn
 */
public double getTickWidth();

/**
 * Is the tick reversed?
 * By setting this property to <code>true</code>,
 * a mirror image of the tick shape is obtained.
 *
 * @return the flag indicating whether to use the mirror image
 */
public boolean isReversed();

/**
 * Sets the <code>automatic</code> Boolean for this tick object.
 * <code>true</code> means ticks and
 * increments are dynamically determined by the gauge. Otherwise,
 * <code>startValue</code>, <code>stopValue</code>, and
 * <code>incrementValue</code> are used to lay out ticks.
 *
 * @param automatic whether automatic or code-controlled tick spacing is
 *     used
 */
public void setAutomatic(boolean automatic);

/**
 * Returns the Boolean controlling the drawing of labels. <code>true</code>
 * means draw the labels; <code>false</code> means hide the labels.
 *
 * @param drawLabels the boolean controlling the drawing of
 *     labels
 */
public void setDrawLabels(boolean drawLabels);

/**
 * Sets the boolean controlling the drawing of ticks. <code>true</code>
 * means draw the ticks; <code>false</code> means hide the ticks.
 *
 * @param drawTicks the boolean controlling the drawing of
 *     ticks
 */
public void setDrawTicks(boolean drawTicks);

/**
 * Sets the color of the font used to label ticks.
 *
 * @param fontColor the font color
 */
public void setFontColor(Color fontColor);

/**
 * In non-automatic mode, use this value as the increment value; that
 * is, it places ticks at every <code>incrementValue</code> along the scale.
 * In automatic mode, the calculated increment value will overwrite this value.
 *
 * @param incrementValue the spacing between adjacent tick marks
 */
public void setIncrementValue(double incrementValue);

/**
 * Sets the percentage of the scale's size (as a decimal) at
 * at which to begin drawing ticks radially outward.
 * The size parameter is the radius in a circular scale,
 * or the width or height, depending on the orientation,
 * in a linear scale.
 *
 * @param innerExtent the inner extent of the tick mark
 */
public void setInnerExtent(double innerExtent);

/** Sets the percentage, expressed as a decimal fraction of the scale's size,
 * at which to place the center of the tick label objects.
 *
 * @param labelExtent the label extent
 */
public void setLabelExtent(double labelExtent);

/**
 * Sets the label generator associated with this tick.
 * If non-null, it will be called, and expected to
 * return a <code>JComponent</code>, which will be the label for a tick mark.
 * See the code in examples.elements.StopWatchExample for an example.
 *
 * @param labelGenerator the <code>labelGenerator</code> which
 *     implements the <code>JCLabelGenerator</code> interface
 */
public void setLabelGenerator(JCLabelGenerator labelGenerator);

/**
 * Sets the percentage of the scale's size (as a decimal) at
 * which to stop drawing the ticks.
 * The size parameter is the radius in a circular scale,
 * or the width or height, depending on the orientation,
 * in a linear scale.
 *
 * @param outerExtent the outer extent of the tick mark
 */
public void setOuterExtent(double outerExtent);

/**
 * Sets the format for numbers in a tick label; positive values denote the
 * number of places after the decimal point (for example, 3 means multiples
 * of 0.001), negative values indicate the positive number of zeros are
 * used before the decimal place (for example, -3 means numbering will be in
 * multiples of 1000). Overridden if <code>precisionUseDefault</code> is set
 * to <code>true</code>.
 *
 * @param precision an integer that specifies the number of decimal places
 *      for the values in tick labels
 */
public void setPrecision(int precision);

/**
 * If <code>true</code>, uses the gauge-determined precision for tick labels;
 * otherwise, uses the current value of <code>precision</code> property to
 * label tick marks.
 *
 * @param precisionUseDefault whether or not to use the gauge-determined
 *     precision
 */
public void setPrecisionUseDefault(boolean precisionUseDefault);

/**
 * Sets the <code>reversed</code> property of the tick.  If the tick is
 * reversed it draws in the opposite direction (for linear
 * scales it points left or up instead of right or down, for circular
 * scales it points towards the center instead of away from the center).
 *
 * @param reversed Is the tick's direction reversed?
 */
public void setReversed(boolean reversed);

/**
 * In non-automatic mode, use this value as the start value.  In automatic
 * mode, the calculated start value will overwrite this value.
 *
 * @param startValue the start value
 */
public void setStartValue(double startValue);

/**
 * In non-automatic mode, use this value as the stop value.  In automatic
 * mode, the calculated stop value will overwrite this value.
 *
 * @param stopValue the stop value
 */
public void setStopValue(double stopValue);

/**
 * The color to use to draw ticks.
 *
 * @param tickColor the color of the tick mark
 */
public void setTickColor(Color tickColor);

/**
 * Sets the <code>JCTickStyle</code> controlling the shape of the tick.
 * Users may define their own tick style, if desired.
 *
 * @see com.klg.jclass.swing.gauge.JCTickStyle
 *
 * @param tickStyle one of the built-in tick styles, or a user-supplied style
 */
public void setTickStyle(JCTickStyle tickStyle);

/**
 * Sets the width in pixels of the tick mark.
 *
 * @param tickWidth the width in pixels of the tick to be drawn
 */
public void setTickWidth(double tickWidth);

}
