/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: LayoutButton.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:32 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.swing.util;

import java.awt.*;
import javax.swing.plaf.*;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.JButton;
import javax.swing.UIManager;
import javax.swing.Icon;

/**
 * A <code>JButton</code> sub-class that creates the min/max, Iconify, and
 * Close buttons used in the JClass MDI interface.
 */
public class LayoutButton extends JButton {

// Maximize and minimize are the same since the button used toggles
public static final int MAXIMIZE = 0;
public static final int MINIMIZE = 1;
public static final int ICONIFY  = 2;
public static final int CLOSE    = 3;

public final static int MOTIF_BUTTON_SIZE = 19;  // 17 + 1 pixel border
public final static Dimension motifButtonDimension =
    new Dimension(MOTIF_BUTTON_SIZE, MOTIF_BUTTON_SIZE);

protected static final String basic_icon_names[] = {
    "InternalFrame.maximizeIcon", "InternalFrame.minimizeIcon",
    "InternalFrame.iconifyIcon", "InternalFrame.closeIcon"
};

protected static final String metal_icon_names[] = {
    "InternalFrame.maximizeIcon",
    "InternalFrame.minimizeIcon",
    "InternalFrame.iconizeIcon",
    "InternalFrame.closeIcon"
};

protected static final String action_commands[] = {
    "Maximize", "Minimize", "Iconify", "Close"
};

protected Color  motifColor;
protected Color  motifHighlight;
protected Color  motifShadow;
protected int    controlType = -1;
public boolean   isMotif = false;

public LayoutButton(int control_type)
{
    super();
    this.controlType = control_type;
    // call inferLookAndFeel since it was already called _before_
    // we were able to set the type
    inferLookAndFeel();
}

/**
 * Returns the type of this Layout control button.  Types are <code>
 * MAXIMIZE, MINIMIZE, ICONIFY, </CODE> or <CODE>CLOSE</CODE>
 */
public int getControlType()
{
    return(controlType);
}

public boolean isFocusTraversable()
{
    return(false);
}

public void requestFocus() {}

public boolean isOpaque()
{
    return(true);
}

public Dimension getMinimumSize()
{
    if (isMotif) {
        return(motifButtonDimension);
    }
    return super.getMinimumSize();
}

public Dimension getPreferredSize()
{
    if (isMotif) {
        return(motifButtonDimension);
    }
    return(super.getPreferredSize());
}

/**
 * Override to handle look at feel changes.
 */
protected void setUI(ComponentUI newUI)
{
    super.setUI(newUI);
    inferLookAndFeel();
}

/**
 * Override to handle UI initialization.
 */
public void updateUI()
{
    super.updateUI();
    inferLookAndFeel();
}

public void paint(Graphics g)
{
    if (!isMotif) {
        super.paint(g);
        return;
    }

    // Do Motif painting

    Dimension d = getSize();
    int maxX = d.width - 1;
    int maxY = d.height - 1;

    // draw background
    g.setColor(motifColor);
    g.fillRect(1, 1, d.width, d.height);

    // draw border
    boolean pressed = getModel().isPressed();
    g.setColor(pressed ? motifShadow : motifHighlight);
    g.drawLine(0, 0, maxX, 0);
    g.drawLine(0, 0, 0, maxY);
    g.setColor(pressed ? motifHighlight : motifShadow);
    g.drawLine(1, maxY, maxX, maxY);
    g.drawLine(maxX, 1, maxX, maxY);

    if (controlType == ICONIFY) {
        g.setColor(motifHighlight);
        g.drawLine(7, 8, 7, 11);
        g.drawLine(7, 8, 10, 8);
        g.setColor(motifShadow);
        g.drawLine(8, 11, 10, 11);
        g.drawLine(11, 9, 11, 11);
    }
    else if (controlType == MAXIMIZE || controlType == MINIMIZE) {
        int max = MOTIF_BUTTON_SIZE - 5;
        boolean isMaxed = controlType == MINIMIZE;
        g.setColor(isMaxed ? motifShadow : motifHighlight);
        g.drawLine(4, 4, 4, max);
        g.drawLine(4, 4, max, 4);
        g.setColor(isMaxed ? motifHighlight : motifShadow);
        g.drawLine(5, max, max, max);
        g.drawLine(max, 5, max, max);
    }
    else {
        // do nothing
    }
}

private static final Border handyEmptyBorder = new EmptyBorder(0, 0, 0, 0);

protected void inferLookAndFeel()
{
    if (controlType == -1) {
        // the type has NOT been set yet.  This routine is being called
        // (indirectly) by the super class BEFORE our contructor code has
        // a chance to execute
        return;
    }
    String laf = UIManager.getLookAndFeel().getClass().getName();
    setFocusPainted(false);
    if (laf.indexOf("MetalLookAndFeel") != -1) {
        Icon icon = UIManager.getIcon(metal_icon_names[this.controlType]);

        // fix pr 13072 -- the names of the icons changed between 1.2.2-1.3
        // so if we cannot find it, use the basic which is the correct one
        // under 1.3
        if (icon == null) {
            icon = UIManager.getIcon(basic_icon_names[this.controlType]);
        }

        setIcon(icon);
        // uncomment the following line if you wish the buttons to
        // be drawn with a purple background
        // putClientProperty("paintActive", Boolean.TRUE);
        setBorder(handyEmptyBorder);
        isMotif = false; //fix for 15377
        setOpaque(false);
    }
    else if (laf.indexOf("MotifLookAndFeel") != -1) {
        motifColor = getBackground();
        motifShadow = motifColor.darker().darker();
        motifHighlight = motifColor.brighter();
        isMotif = true;
        setBorderPainted(false);
    }
    else {
        Icon icon = UIManager.getIcon(basic_icon_names[this.controlType]);
        setIcon(icon);
        isMotif = false;
    }

    setActionCommand(action_commands[controlType]);
}

}
