/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: AdjustmentHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:32 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.awt.Adjustable;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.io.Serializable;
import javax.swing.JScrollBar;
import javax.swing.JLabel;

/**
 * <CODE>AdjustmentHandler</CODE> responds to <CODE>AdjustmentEvents</CODE> and
 * manages table scrolling.
 */
public class AdjustmentHandler implements AdjustmentListener, Serializable {

protected JCTable     table;
protected boolean     isTracking = false;
protected JLabel      trackLabel = new JLabel();

// We have to keep additional information to deal with the handicapped
// Swing scrolling.
private int vertLastRow       = -1;
private int vertLastValue     = -1;
private int vertLastRealValue = -1;
private int horLastColumn     = -1;
private int horLastValue      = -1;
private int horLastRealValue  = -1;

/**
 * Creates a new AdjustmentHandler and links it to table.
 * @param table The table to which this handler is linked
 */
public AdjustmentHandler(JCTable table)
{
	this.table = table;
	trackLabel.setOpaque(true);
	trackLabel.setVisible(false);
}

/**
 * Controls the origin of the scrollable cell region when
 * <CODE>AdjustmentEvent</CODE> objects are generated by the scrollbars.
 * AdjustmentHandler determines the origin based on the <CODE>JumpScroll</CODE>
 * setting.
 * @param e The event generated by the adjustment
 */
public void adjustmentValueChanged(AdjustmentEvent e)
{
	// Abort if repaint is disabled
	if (!table.isRepaintEnabled()) {
		table.setRecalcRequired(true);
		return;
	}

	Adjustable adjustable = e.getAdjustable();
	CellAreaHandler cellAreaHandler = table.getCellAreaHandler();
	EditTraverseHandler editHandler = table.getEditTraverseHandler();

	int value = e.getValue();
	JCScrollEvent se = null;
	int jumpMode = table.getJumpScroll();

	// Handle vertical adjustment changes
	if (adjustable.getOrientation() == Adjustable.VERTICAL) {
		int y = cellAreaHandler.getCellOrigin().y;
		int currentValue = -cellAreaHandler.getCellOrigin().y;

        if (value == currentValue && !isTracking) {
            return;
        }
		// Flag for deferred notification to scrollbar that value
		// needs to be adjusted.  Handy in jump scroll case.
		boolean notifyAdj = false;

		if ((jumpMode == JCTableEnum.JUMP_VERTICAL ||
        	 jumpMode == JCTableEnum.JUMP_ALL) && value != vertLastValue)
		{
			// Determine which cell contains the value
            synchronized (this) {
                int standardOffsetY =
                    cellAreaHandler.getCellArea().getStandardOffsetY();
                int row = table.getCellLayout().getRow(0, value +
                                                          standardOffsetY);
                if (row == vertLastRow) {
				    // They cannot be equal
                    row += value > vertLastValue ? 1 : -1;
                }
                if (adjustable instanceof JScrollBar &&
                    ((JScrollBar)adjustable).getValueIsAdjusting() &&
                    ((value > vertLastRealValue && row < vertLastRow) ||
                    (value < vertLastRealValue && row > vertLastRow)))
				{
                    row = vertLastRow;
                }
                vertLastRealValue = value;
                value = (table.getCellLayout().getRowPosition(row) -
                         standardOffsetY);
                int maxAmount = (adjustable.getMaximum() -
				                 adjustable.getVisibleAmount());
                // We cannot scroll below the bottom
                if (value > maxAmount) {
                    value = maxAmount;
                }
                vertLastRow = row;
                vertLastValue = value;
                if (value != vertLastRealValue) {
					// defer call to adjust scrollbar until after
					// we've notified a scroll will begin
					notifyAdj = true;
                }
            }
		}
		// Notify scroll listeners that a new scroll is beginning
		if (table.scrollListeners != null && value != currentValue) {
			se = new JCScrollEvent(this, JCScrollEvent.SCROLL, adjustable,
								   adjustable.getOrientation(), value, e);
			table.fireJCScrollEvent(se);
			if (se.getValue() != value) {
				adjustable.setValue(se.value);
				return;
			}
		}
		cellAreaHandler.cellOrigin.y = -value;
		// Adjust scroll value if necessary (jump scroll case)
		// this MUST be done AFTER the cell origin has been changes
		// so that the inevitably resulting event will be filtered out
		if (notifyAdj) {
            adjustable.setValue(value);
		}

		JScrollBar sb = (JScrollBar)e.getSource();
		if (sb.getValueIsAdjusting()) {
			if (table.getVertSBTrack() != JCTableEnum.TRACK_LIVE) {
				if (isTracking) {
					int top_row = table.getVisibleCells().start_row;
					String track_text = "";
					if (table.getVertSBTrack() == JCTableEnum.TRACK_ROW_NUMBER){
						track_text = Integer.toString(top_row + 1);
					}
					else {
						int track_column = table.getVertSBTrackColumn();
						Object track_object =
						    table.getDataView().getObject(top_row,
							                              track_column);
						if (track_object == null ||
						    track_object.toString() == null)
						{
							track_text = "";
						}
						else {
							track_text = track_object.toString();
						}
					}
					if (!trackLabel.isVisible()) {
						trackLabel.setVisible(true);
                    }
					trackLabel.setText(track_text);
				}
				else {
					CellArea area = getRightCellArea();
					area.add(trackLabel, 0);
					table.getFocusManager().requestFocus(trackLabel);
					int track_width = 50;
					int track_height = 20;
					Dimension d = table.getTrackSize();
					if (d != null) {
						track_width = d.width;
						track_height = d.height;
					}
					trackLabel.setBounds(area.getWidth() - track_width,
										 (area.getHeight() - track_height) / 2,
										 track_width, track_height);
					trackLabel.setHorizontalAlignment(JLabel.RIGHT);
					if (table.getTrackBackground() != null) {
						trackLabel.setBackground(table.getTrackBackground());
					}
					if (table.getTrackForeground() != null) {
						trackLabel.setForeground(table.getTrackForeground());
					}
					trackLabel.setText("");
					isTracking = true;
				}

				// Fire the AFTER_SCROLL event.
				se = new JCScrollEvent(this, JCScrollEvent.AFTER_SCROLL,
				                       adjustable, adjustable.getOrientation(),
									   value, e);
				table.fireJCScrollEvent(se);
				return;
			}
		}
		else {
			if (isTracking) {
				isTracking = false;
				trackLabel.setVisible(false);
				editHandler.setFocusOnEditor();
				// need one last after scroll event now that we're no
				// longer tracking, otherwise components in cells don't
				// know enough to say goodbye
				JCScrollEvent lastSE = null;
				lastSE = new JCScrollEvent(this, JCScrollEvent.AFTER_SCROLL,
							               adjustable,
										   adjustable.getOrientation(),
										   value, e);
				table.fireJCScrollEvent(lastSE);
                table.layoutAreas();
			}
		}

		if (editHandler.getCellEditor() != null) {
			editHandler.move(EditTraverseHandler.VERTICAL, y + value);
		}

		// Force a repaint if a resize isn't under way
		if (!table.mouseHandler.isResizing()) {
			Rectangle cellRect = table.cellAreaHandler.cellRect;
			if (cellRect != null) {
    			table.repaint(0, cellRect.y, table.getSize().width,
				              cellRect.height);
    		}
		}
	}
	// Handle horizontal adjustment changes
	else {
		int x = cellAreaHandler.getCellOrigin().x;
		int currentValue = -cellAreaHandler.getCellOrigin().x;

        if (value == currentValue && !isTracking) {
            return;
        }
		// Flag for deferred notification to scrollbar that value
		// needs to be adjusted.  Handy in jump scroll case.
		boolean notifyAdj = false;

		if ((jumpMode == JCTableEnum.JUMP_HORIZONTAL ||
        	 jumpMode == JCTableEnum.JUMP_ALL) && value != horLastValue)
		{
			// Determine which cell contains the value
            synchronized (this) {
                int standardOffsetX =
                    cellAreaHandler.getCellArea().getStandardOffsetX();
                int column =
				    table.getCellLayout().getColumn(value + standardOffsetX, 0);
                if (column == horLastColumn) {
				    // They cannot be equal
                    column += value>horLastValue ? 1 : -1;
                }
                if (adjustable instanceof JScrollBar &&
                    ((JScrollBar)adjustable).getValueIsAdjusting() &&
                    ((value > horLastRealValue && column < horLastColumn) ||
                    (value < horLastRealValue && column > horLastColumn)))
				{
                    column = horLastColumn;
                }
                horLastRealValue = value;
                value = (table.getCellLayout().getColumnPosition(column) -
                         standardOffsetX);
                int maxAmount = (adjustable.getMaximum()-
                                 adjustable.getVisibleAmount());
                // We cannot scroll below the bottom
                if (value > maxAmount) {
                    value = maxAmount;
                }
                horLastColumn = column;
                horLastValue = value;
                if (value != horLastRealValue) {
					// Defer call to adjust scrollbar until after
					// we've notified a scroll will begin
					notifyAdj = true;
                }
            }
		}
		// Notify scroll listeners that a new scroll is beginning
		if (table.scrollListeners != null && value != currentValue) {
			se = new JCScrollEvent(this, JCScrollEvent.SCROLL, adjustable,
								   adjustable.getOrientation(), value, e);
			table.fireJCScrollEvent(se);
			if (se.getValue() != value) {
				adjustable.setValue(se.value);
				return;
			}
		}
		cellAreaHandler.cellOrigin.x = -value;
		// Adjust scroll value if necessary (jump scroll case)
		// this MUST be done AFTER the cell origin has been changes
		// so that the inevitably resulting event will be filtered out
		if (notifyAdj) {
            adjustable.setValue(value);
		}

		JScrollBar sb = (JScrollBar)e.getSource();
		if (sb.getValueIsAdjusting()) {
			if (table.getHorizSBTrack() != JCTableEnum.TRACK_LIVE) {
				if (isTracking) {
					int left_column = table.getVisibleCells().start_column;
					String track_text = "";
					if (table.getHorizSBTrack() == JCTableEnum.TRACK_COLUMN_NUMBER) {
						track_text = Integer.toString(left_column + 1);
					}
					else {
						int track_row = table.getHorizSBTrackRow();
						track_row =
						    (track_row >= -1) ? track_row : JCTableEnum.LABEL;
						Object track_object =
						    table.getDataView().getObject(track_row,
							                              left_column);
						if (track_object == null ||
						    track_object.toString() == null)
						{
							track_text = "";
						}
						else {
							track_text = track_object.toString();
						}
					}
					if (!trackLabel.isVisible()) {

						trackLabel.setVisible(true);
                    }
					trackLabel.setText(track_text);
				}
				else {
					CellArea area = getBottomCellArea();
					area.add(trackLabel, 0);
					table.getFocusManager().requestFocus(trackLabel);
					int track_width = 50;
					int track_height = 20;
					Dimension d = table.getTrackSize();
					if(d != null) {
						track_width = d.width;
						track_height = d.height;
					}
					trackLabel.setBounds((area.getWidth() - track_width) / 2,
										 area.getHeight() - track_height,
										 track_width, track_height);
					trackLabel.setHorizontalAlignment(JLabel.CENTER);
					if (table.getTrackBackground() != null) {
						trackLabel.setBackground(table.getTrackBackground());
					}
					if (table.getTrackForeground() != null) {
						trackLabel.setForeground(table.getTrackForeground());
					}
					trackLabel.setText("");
					isTracking = true;
				}

				// Fire the AFTER_SCROLL event.
				se = new JCScrollEvent(this, JCScrollEvent.AFTER_SCROLL,
				                       adjustable, adjustable.getOrientation(),
									   value, e);
				table.fireJCScrollEvent(se);
				return;
			}
		}
		else {
			if (isTracking) {
				isTracking = false;
				trackLabel.setVisible(false);
				editHandler.setFocusOnEditor();
				// Need one last after scroll event now that we're no
				// longer tracking, otherwise components in cells don't
				// know enough to say goodbye
				JCScrollEvent lastSE = null;
				lastSE = new JCScrollEvent(this, JCScrollEvent.AFTER_SCROLL,
							               adjustable,
							               adjustable.getOrientation(),
										   value, e);
				table.fireJCScrollEvent(lastSE);
                table.layoutAreas();
			}
		}

		// Move the cell editor as required
		if (editHandler.getCellEditor() != null) {
			editHandler.move(EditTraverseHandler.HORIZONTAL, x + value);
		}

		// Force a repaint if a resize isn't under way
		if (!table.mouseHandler.isResizing()) {
			Rectangle cellRect = table.cellAreaHandler.cellRect;
			if (cellRect != null) {
			    table.repaint(cellRect.x, 0, cellRect.width,
				              table.getSize().height);
			}
		}
	}

	// Notify scroll listeners that scroll is complete
	if (table.scrollListeners != null && se != null) {
		se = new JCScrollEvent(this, JCScrollEvent.AFTER_SCROLL, adjustable,
							   adjustable.getOrientation(), value, e);
		table.fireJCScrollEvent(se);
	}
}

/**
 * Ensures that the cell origin is always in sync with the current scrollbar
 * values.
 */
protected void adjustOrigin()
{
	Point p = table.getCellAreaHandler().getCellOrigin();

	// Reset the x value if the horizontal scrollbar value differs
	JScrollBar hsb = table.getHorizSB();
	if (hsb != null && hsb.isVisible() && hsb.getValue() != -p.x) {
		p.x = -hsb.getValue();
	}
	hsb = null;

	// Reset the y value if the vertical scrollbar value differs
	JScrollBar vsb = table.getVertSB();
	if (vsb != null & vsb.isVisible() && vsb.getValue() != -p.y) {
		p.y = -vsb.getValue();
	}
	vsb = null;
}

/**
 * Returns bottom-most CellArea that will be used for placing the
 * horizontal track label when visible.
 * @return The bottom-most CellArea
 */
protected CellArea getBottomCellArea()
{
	if (table.isColumnLabelDisplay() &&
	    table.getColumnLabelPlacement() == JCTableEnum.PLACE_BOTTOM)
	{
		return(table.getCellAreaHandler().findByType(CellArea.COLUMNLABEL));
	}
	else if (table.getFrozenRows() > 0 &&
			 table.getFrozenRowPlacement() == JCTableEnum.PLACE_BOTTOM)
	{
		return(table.getCellAreaHandler().findByType(CellArea.FROZENROW));
	}
	return(table.getCellAreaHandler().findByType(CellArea.CELL));
}

/**
 * Returns the right-most CellArea that will be used for placing the
 * vertical track label when visible.
 * @return The right most cell area
 */
protected CellArea getRightCellArea()
{
	if (table.isRowLabelDisplay() &&
	    table.getRowLabelPlacement() == JCTableEnum.PLACE_RIGHT)
	{
		return(table.getCellAreaHandler().findByType(CellArea.ROWLABEL));
	}
	else if (table.getFrozenColumns() > 0 &&
			 table.getFrozenColumnPlacement() == JCTableEnum.PLACE_RIGHT)
	{
		return(table.getCellAreaHandler().findByType(CellArea.FROZENCOLUMN));
	}
	return(table.getCellAreaHandler().findByType(CellArea.CELL));
}

/**
 * Returns the scrollbar tracking state. During the tracking the table does
 * not adjust according to the scrollbar value, but redisplays after the
 * tracking completes.
 * @return The current scrollbar tracking state
 */
public boolean isTracking()
{
	return(isTracking);
}

/**
 * Ensures that AdjustmentHandler can be disposed.
 */
public void dispose()
{
	table = null;
	trackLabel = null;
}

}
