/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BaseDataView.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:32 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.cell.EditorRendererRegistry;
import com.klg.jclass.cell.JCCellEditor;
import com.klg.jclass.cell.JCCellRenderer;
import com.klg.jclass.cell.renderers.JCStringCellRenderer;
import java.io.Serializable;





import java.util.Hashtable;


/**
 * The <CODE>BaseDataView</CODE> class implements all <CODE>DataViewModel</CODE> methods that control
 * editors and renderers.
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, some of the return types for this class'
 * fields or methods are based on the corresponding
 * <code>com.sun.java.util</code> class.
 */
public abstract class BaseDataView implements DataViewModel, Serializable {

protected JCTable   table;
protected Hashtable renderers = null;
protected Hashtable editors = null;

/**
 * Creates a BaseDataView instance that initializes the editor and renderer
 * tables with defaults settings but does not link it to table.
 */
public BaseDataView()
{
	this(null);
}

/**
 * Creates a BaseDataView instance and initializes the editor and renderer
 * tables with default settings.
 * @param table The table controlled by this DataView.
 */
public BaseDataView(JCTable table)
{
	setTable(table);
	editors   = new Hashtable();
	renderers = new Hashtable();
}

/**
 * Returns the JCTable controlled by the DataView.
 * @return The current table controlled by this DataView.
 */
public JCTable getTable()
{
	return(table);
}

/**
 * Sets the JCTable controlled by the DataView.
 * @param table The new table for this DataView
 */
public void setTable(JCTable table)
{
	this.table = table;
}

/**
 * Returns the cell renderer for a row/column.
 * @param row The row index
 * @param column The column index
 * @return The cell renderer for the given row and column.
 */
public JCCellRenderer getCellRenderer(int row, int column)
{
	Object o = null;
	if (row != JCTableEnum.LABEL && column != JCTableEnum.LABEL) {
		o = getTableDataItem(row, column);
	}
	else if(row == JCTableEnum.LABEL) {
		o = getTableColumnLabel(column);
	}
	else {
		o = getTableRowLabel(row);
	}
	return(getCellRenderer(row, column, o));
}

/**
 * Returns the cell renderer for a row/column given the data for the cell.
 * @param row The row index
 * @param column The column index
 * @param o The data in the cell indexed by row and column
 * @return The cell renderer based on the data in the cell given the row and
 *    column indicies
 */
public JCCellRenderer getCellRenderer(int row, int column, Object o)
{
	CellStyleModel csm = table.getCellStyle(row, column);

	JCCellRenderer cr = csm.getCellRenderer();
	// If a renderer was set for the cell, return it
	if (cr != null) {
		return(cr);
	}

	if (o != null) {
		return(getCellRenderer(o));
	}
	else {
		return(getCellRenderer(csm.getDataType()));
	}
}

/**
 * Returns the cell renderer given a data type class.  This looks in the
 * EditorRenderRegistry if there isn't one already in the renderer cache.
 *
 * @param type The class type of the required renderer
 * @return The cell renderer based on the class type
 */
public JCCellRenderer getCellRenderer(Class type)
{
	if (type == null) {
		return(new JCStringCellRenderer());
	}
	JCCellRenderer cr = null;
	try {
		Class c = type;
		Class key = null;
		Object mapped = null;
		do {
			mapped = renderers.get(c);
			key = c;
			c = c.getSuperclass();
		}
		while ((mapped == null) && (c != null));

        if (c == null) {
            // Note there is nothing to use for the modifier type for the
            // EditorRendererRegistry so null is passed.  The Registry allows
            // multipule editors and renderers to registered for a particular
            // class by defining a different modifier type for each.  Table may
            // make use of this if it wishes.  The default value for the
			// modifier is always null.
			EditorRendererRegistry r = EditorRendererRegistry.getCentralRegistry();
            cr = r.getCellRendererInstance(type, null, false);

            // Place the renderer in tables renderer cache
			if (cr != null) {
                renderers.put(type, cr);
			}
        }
        else if (mapped instanceof Class) {
			cr = (JCCellRenderer) ((Class)mapped).newInstance();
			renderers.put(key, cr);
		}
        else if (mapped == null) {
            cr = null;
        }
		else {
			cr = (JCCellRenderer)mapped;
		}
	}
	catch(Exception e) {
		e.printStackTrace();
	}

	if (cr == null) {
		// no translation available
		throw new JCTableException("No renderer for "+ type.getName() +
		                           " available");
	}
	return(cr);
}

/**
 * Returns the cell renderer given an object.  If the object isn't null, it
 * just returns the renderer for the class type of the object.
 *
 * @param o The object the cell renderer needs to render
 * @return The cell renderer based on the object.
 */
public JCCellRenderer getCellRenderer(Object o)
{
	if (o == null) {
		return(new JCStringCellRenderer());
	}
    JCCellRenderer cr = getCellRenderer(o.getClass());
	return(cr);
}

/**
 * Returns a Hashtable of object to renderer mappings.
 * @return The current renderer cache
 */
public Hashtable getRenderersTable()
{
	return(renderers);
}

/**
 * Sets the Hashtable of object to renderer mappings.
 * @param ht The new renderer cache
 */
public void setRenderersTable(Hashtable ht)
{
	if (ht == null) {
		return;
	}
	renderers = ht;
}

/**
 * Returns the cell editor for a row/column.
 * @param row The row index
 * @param column The column index
 * @return The cell editor for the given row and column
 */
public JCCellEditor getCellEditor(int row, int column)
{
	Object o = null;
	if (row != JCTableEnum.LABEL && column != JCTableEnum.LABEL) {
		o = getTableDataItem(row, column);
	}
	else if(row == JCTableEnum.LABEL) {
		o = getTableColumnLabel(column);
	}
	else {
		o = getTableRowLabel(row);
	}

	CellStyleModel csm = table.getCellStyle(row, column);

	JCCellEditor ce = csm.getCellEditor();
	// If an editor was set for the cell, return it
	if (ce != null) {
		return(ce);
	}

	if (o != null) {
		return(getCellEditor(o));
	}
	else {
		return(getCellEditor(csm.getDataType()));
	}
}

/**
 * Returns the cell editor given a data type class.
 * @param type The class type of the required cell editor
 * @return The cell editor based on the give class type
 */
public JCCellEditor getCellEditor(Class type)
{
	if (type == null) {
		return(null);
	}
/*
	if (editors.size() == 0) {
		return(null);
	}
*/
	JCCellEditor ce = null;
	try {
		Class c = type;
		Class key = null;
		Object mapped = null;
		do {
			mapped = editors.get(c);
			key = c;
			c = c.getSuperclass();
		}
		while ((mapped == null) && (c != null));

        if (c == null) {
            // Note there is nothing to use for the modifier type for the
            // EditorRendererRegistry so null is passed.  The Registry allows
            // multipule editors and renderers to registered for a particular
            // class by defining a different modifier type for each.  Table may
            // make use of this if it wishes.  The default value for the
			// modifier is always null.
			EditorRendererRegistry r = EditorRendererRegistry.getCentralRegistry();
            ce = r.getCellEditorInstance(type, null, false);

            // Place the renderer in tables renderer cache
            if (ce != null) {
                editors.put(type, ce);
            }
        }
        else if (mapped instanceof Class) {
			ce = (JCCellEditor) ((Class)mapped).newInstance();
			editors.put(key, ce);
		}
        else if (mapped == null)
        {
            ce = null;
        }
		else {
			ce = (JCCellEditor)mapped;
		}
	}
	catch (Exception e) {
		e.printStackTrace();
	}
	return(ce);
}

/**
 * Returns the cell editor given an object.
 * @param o The object for which a cell editor is required
 * @return The cell editor based on the given object
 */
public JCCellEditor getCellEditor(Object o)
{
	if (o == null) {
		return(null);
	}

  	JCCellEditor ce = getCellEditor(o.getClass());
	return(ce);
}

/**
 * Returns a Hashtable of object to editor mappings.
 * @return The current editor cache
 */
public Hashtable getEditorsTable()
{
	return(editors);
}

/**
 * Sets the Hashtable for object to editor mappings.
 * @param ht The new eidtor cache
 */
public void setEditorsTable(Hashtable ht)
{
	if (ht == null) {
		return;
	}
	editors = ht;
}

/**
 * Loads the special field renderers designed for use with JCTable.  This
 * method no longer does anything as of JClass 4.0.  Its around for backwards
 * compatibility only.
 * @param ht The hashtable to which the field renderers are to be added
 */
protected void loadFieldRenderers(Hashtable ht)
{
	// ATTN GV Disabled pending central editor-renderer registry.
	// For 4.0 and later, we aren't including the field-based editors/renderers
	// with the product.
	/*
	try {
		ht.put(Class.forName("java.lang.Float"),
			   Class.forName("com.klg.jclass.field.cell.JCTextFieldCellEditorRenderer"));
		ht.put(Class.forName("java.lang.Double"),
			   Class.forName("com.klg.jclass.field.cell.JCTextFieldCellEditorRenderer"));
		ht.put(Class.forName("java.lang.Integer"),
			   Class.forName("com.klg.jclass.field.cell.JCTextFieldCellEditorRenderer"));
		ht.put(Class.forName("java.lang.Long"),
			   Class.forName("com.klg.jclass.field.cell.JCTextFieldCellEditorRenderer"));
		ht.put(Class.forName("java.lang.Byte"),
			   Class.forName("com.klg.jclass.field.cell.JCTextFieldCellEditorRenderer"));
		ht.put(Class.forName("java.lang.Long"),
			   Class.forName("com.klg.jclass.field.cell.JCTextFieldCellEditorRenderer"));
		ht.put(Class.forName("java.util.Date"),
			   Class.forName("com.klg.jclass.field.cell.JCPopupFieldCellEditorRenderer"));
		ht.put(Class.forName("java.util.Calendar"),
			   Class.forName("com.klg.jclass.field.cell.JCPopupFieldCellEditorRenderer"));
		ht.put(Class.forName("java.sql.Date"),
			   Class.forName("com.klg.jclass.field.cell.JCPopupFieldCellEditorRenderer"));
		ht.put(Class.forName("java.sql.Timestamp"),
			   Class.forName("com.klg.jclass.field.cell.JCPopupFieldCellEditorRenderer"));
		ht.put(Class.forName("java.sql.Time"),
			   Class.forName("com.klg.jclass.field.cell.JCPopupFieldCellEditorRenderer"));
	}
	catch (Exception e) {
		e.printStackTrace(System.out);
	}
	*/
}

/**
 * Loads the special field renderers designed for use with JCTable.  This
 * method no longer does anything as of JClass 4.0.  Its around for backwards
 * compatibility only.
 * @param ht The hashtable to which the field editors are to be added
 */
protected void loadFieldEditors(Hashtable ht)
{
	loadFieldRenderers(ht);
}

/**
 * Sets the bound data source.
 * @param tableData The new bound data source
 */
public abstract void setDataSource(TableDataModel tableData);

/**
 * Returns the bound data source.
 * @return The current bound data source
 */
public abstract TableDataModel getDataSource();

/**
 * Returns true if the data source implements EditableTableData.
 * @return Is the data source editable?
 */
public abstract boolean isEditable();

/**
 * Returns the object given a row or column. This convenience method handles
 * both cells and labels.
 * @param row The row index
 * @param column The column index
 * @return The data object based on the given row and column
 */
public abstract Object getObject(int row, int column);

/**
 * Sets the cell value for editable data sources.
 * @param o The new cell value
 * @param row The row index
 * @param column The column index
 * @return The status of the edit.  Note that if the datasource is not
 *    editable, no edit is done and true is returned.
 */
public abstract boolean setTableDataItem(Object o, int row, int column);

/**
 * Returns the value for the cell at the specified location.<P>
 * @param row The row index
 * @param column The column index
 * @return The data object based on the given row and column
 */
public abstract Object getTableDataItem(int row, int column);

/**
 * Returns the value for the row label at the specified row.<P>
 * @param row The row index
 * @return The label value for the given row
 */
public abstract Object getTableRowLabel(int row);

/**
 * Returns the value for the column label at the specified column.<P>
 * @param column The column index
 * @return The label value for the given column
 */
public abstract Object getTableColumnLabel(int column);

/**
 * Method to change the data based on the passed in event
 * @param e The event used to change the data
 */
public abstract void dataChanged(JCTableDataEvent e);

}
