/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: CellSize.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:32 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.io.Serializable;

/**
 * The <CODE>CellSize</CODE> class controls the dimensions for single row or
 * column. Four properties control how cell sizes are calculated: character
 * value which bases the size of a cell on a set number of characters or lines;
 * minimum value which controls the minimum size of a cell; maximum value which
 * controls the maximum size of a cell; and pixel value which sets a fixed
 * dimension, determines whether character based control is used or sets the
 * size to variable control. Querying any of the properties returns the setting
 * and not the actual value. Use the <CODE>Hidden</CODE> property to hide and
 * show rows and columns.
 */
public class CellSize implements Serializable, Cloneable {

protected int         char_value  = 0;
protected int         min_value   = 0;
protected int         max_value   = JCTableEnum.MAXINT;
protected int         pixel_value = JCTableEnum.NOVALUE;
protected boolean     hidden      = false;

/**
 * Creates a new CellSize object based on character dimensions.
 */
public CellSize(int char_value)
{
    this(char_value, 0, JCTableEnum.MAXINT, JCTableEnum.NOVALUE, false);
}

/**
 * Creates a new CellSize object based on pixel dimensions.
 */
public CellSize(int min_value, int max_value, int pixel_value)
{
    this(0, min_value, max_value, pixel_value, false);
}

/**
 * Creates a new CellSize object based on all sizing options.
 */
public CellSize(int char_value, int min_value, int max_value, int pixel_value)
{
    this(char_value, min_value, max_value, pixel_value, false);
}

/**
 * Creates a new CellSize objet based on all sizing options, including
 * visibility.
 */
public CellSize(int char_value, int min_value, int max_value, int pixel_value,
                boolean hidden)
{
    this.char_value = char_value;
    this.min_value = min_value;
    this.max_value = max_value;
    this.pixel_value = pixel_value;
    this.hidden = hidden;
}

/**
 * Returns the number of characters that a column can display or the number
 * of lines for a row.
 * @see #setCharValue
 */
public int getCharValue()
{
    return(char_value);
}

/**
 * For columns, sets the number of characters that a column can display
 * (default: 10), and for rows, the number of lines of text that a row can
 * display (default: 1). Cell values that do not fit in the cell or label are
 * clipped. This property controls the size if <CODE>PixelValue</CODE> is
 * <CODE>JCTableEnum.NOVALUE</CODE>.
 */
public void setCharValue(int value)
{
    this.char_value = value;
}

/**
 * Returns true if the row or column is hidden. The hidden property does not
 * determine whether the cell is currently in view.
 * @see #setHidden
 */
public boolean isHidden()
{
    return(hidden);
}

/**
 * Determines whether a row or column is hidden without invalidating the
 * settings that control sizing when visible.
 */
public void setHidden(boolean value)
{
    this.hidden = value;
}

/**
 * Returns the minimum size in pixels of a row or columns.
 * @see #setMinValue
 */
public int getMinValue()
{
    return(min_value);
}

/**
 * Sets the minimum size in pixels of a row or column.
 */
public void setMinValue(int value)
{
    this.min_value = value;
}

/**
 * Returns the maximum size in pixels of a row or columns.
 * @see #setMaxValue
 */
public int getMaxValue()
{
    return(max_value);
}

/**
 * Sets the maximum size in pixels of a row or column.
 */
public void setMaxValue(int value)
{
    this.max_value = value;
}

/**
 * Returns the pixel value for a row or column.
 * @see #setPixelValue
 */
public int getPixelValue()
{
    return(pixel_value);
}

/**
 * Sets the width of a column or height of a row in pixels. Use
 * <CODE>JCTableEnum.NOVALUE</CODE> to control width/height by character
 * value.<P>
 * The following special values are valid:
 * <pre>
 * JCTableEnum.NOVALUE            use character setting to control the size of a
 *                                row or column
 * JCTableEnum.AS_IS              do not change. Use to convert from VARIABLE to
 *                                fixed size
 * JCTableEnum.VARIABLE           set to the highest value in the row or column
 * JCTableEnum.VARIABLE_ESTIMATE  set to the highest value in the first
 *                                <CODE>VariableCount</CODE> cells
 * </pre>
 */
public void setPixelValue(int value)
{
    this.pixel_value = value;
}

/**
 * Returns a copy of the CellSize object.
 */
public Object clone()
{
    return(new CellSize(char_value, min_value, max_value, pixel_value, hidden));
}

/**
 * Returns a string representation of the CellSize object.
 */
public String toString()
{
    return(super.toString() + " [char=" + char_value + ",min=" + min_value +
           ",max=" + max_value + ",pixel=" + pixel_value +
           ",hidden=" + hidden + "]");
}

}


