/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: CellSizeHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:32 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.cell.JCCellRenderer;
import com.klg.jclass.cell.JCComponentCellRenderer;
import com.klg.jclass.cell.JCLightCellRenderer;

import java.awt.*;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

/**
 * CellSizeHandler controls the calculation of column widths and row heights
 * based on a combination of the char/pixel/min/max/hidden state of the row/column
 * and the data contained in the cell.
 */
class CellSizeHandler {

// Used to cache a mapping between Font instances and row heights.
// Reduces start-up time on large tables when using character heights.
// Hash table set to null in JCTable.dispose() in order to not
// leak memory, even though memory held by this hash table is minor.
// Why bother?  You never know what references are held by the native
// portion of the Font, FontMetrics or Graphics objects.
//
// Jan 7/2000 GV: This is probably not worth doing for column widths.
static Hashtable cachedFontHeights = null;

/**
 * Recalculates all row and column dimensions.
 */
protected static void resetDimensions(JCTable table)
{
	// Abort if repaint is not enabled
	if (!table.repaint_enabled) {
		table.setRecalcRequired(true);
		return;
	}

	// Calculate the column widths
	calcColumnWidths(table, JCTableEnum.LABEL, JCTableEnum.MAXINT,
                     JCTableEnum.LABEL, JCTableEnum.MAXINT);
	// Calculate the row heights
	calcRowHeights(table, JCTableEnum.LABEL, JCTableEnum.MAXINT,
                   JCTableEnum.LABEL, JCTableEnum.MAXINT);
}

/**
 * Calculates the height of the column labels.
 */
protected static boolean calcColumnLabelHeight(JCTable table, int start_column,
				                               int end_column)
{
	return(calcRowHeights(table, JCTableEnum.LABEL, JCTableEnum.LABEL,
	                      start_column, end_column));
}

/**
 * Cache the pixel height of a range of rows given a range of columns to
 * consider.
 */
protected static boolean calcRowHeights(JCTable table,
										int start_row, int end_row,
                                        int start_column, int end_column)
{
	boolean changed = false;

	if (!table.repaint_enabled) {
		table.setRecalcRequired(true);
		return(changed);
	}

    int numRows    = table.getNumRows();
    int numColumns = table.getNumColumns();

    int defRowEnd = (numRows == 0)    ? JCTableEnum.LABEL : numRows - 1;
    int defColEnd = (numColumns == 0) ? JCTableEnum.LABEL : numColumns - 1;

	// Check range
	if (start_row == JCTableEnum.ALL) {
		start_row = JCTableEnum.LABEL;
		end_row = defRowEnd;
	}
	else if(start_row == JCTableEnum.ALLCELLS) {
		start_row = 0;
		end_row = defRowEnd;
	}
	if (start_column == JCTableEnum.ALL) {
		start_column = JCTableEnum.LABEL;
		end_column = defColEnd;
	}
	else if (start_column == JCTableEnum.ALLCELLS) {
		start_column = 0;
		end_column = defColEnd;
	}

    // Check start points
    if (start_row == JCTableEnum.LABEL && numColumns == 0) {
        // If the label row size is fixed we still want to calculate it
        int p = table.getRowCellSize(JCTableEnum.LABEL).getPixelValue();
        if (p == JCTableEnum.VARIABLE || p == JCTableEnum.VARIABLE_ESTIMATE) {
            start_row = 0;
        }
    }
    if (start_column == JCTableEnum.LABEL && numRows == 0) {
        start_column = 0;
    }

	// Check end points
	if (end_row == JCTableEnum.MAXINT) {
		end_row = defRowEnd;
	}
	if (end_column == JCTableEnum.MAXINT) {
		end_column = defColEnd;
	}

    if (start_row > end_row || start_column > end_column) {
        return(changed);
    }

	// Handle a single row request
	if (start_row == end_row) {
        changed |= setPixelHeight(table, start_row, start_row, end_row,
                                  start_column, end_column);
		return(changed);
	}

    // Reset the label row
    CellLayoutModel cellLayout = table.getCellLayout();
    cellLayout.setHeight(JCTableEnum.LABEL, 0);

	// Handle the label row
	if (start_row == JCTableEnum.LABEL) {
		changed |= setPixelHeight(table, start_row, start_row, end_row,
				                  start_column, end_column);
	}

    // Reset the column sizes of the table
	cellLayout.resetRows();

	// Handle the entire table
	changed |= setPixelHeight(table, JCTableEnum.ALLCELLS, 0, end_row,
	                          start_column, end_column);

	// Now handle the individual rows
	List rowHeightList = table.getRowHeightValues();
	for (Iterator it = rowHeightList.iterator(); it.hasNext(); ) {
		PositionCellSize pcs = (PositionCellSize)it.next();
 		int row = pcs.position;
		if (row >= 0) { // Non-special row numbers
			changed |= setPixelHeight(table, row, row, row,
			                          start_column, end_column);
		}
		else if (row == JCTableEnum.ALL || row == JCTableEnum.ALLCELLS) {
			CellSize cell_size = table.getRowCellSize(row);
			int pixel_height = cell_size.getPixelValue();
			if (pixel_height == JCTableEnum.VARIABLE ||
			    pixel_height == JCTableEnum.VARIABLE_ESTIMATE)
			{
			    for (int i = start_row; i <= end_row; i++) {
				    changed |= setPixelHeight(table, i, i, i,
					                          start_column, end_column);
			    }
			}
		}
	}
	return(changed);
}

protected static boolean setPixelHeight(JCTable table, int row,
										int start_row, int end_row,
										int start_column, int end_column)
{
    boolean changed = false;
	int offset = table.getTextHeightOffset() * 2;
	CellLayoutModel cellLayout = table.getCellLayout();
	CellSize cell_size = table.getRowCellSize(row);
	int current_height = cellLayout.getHeight(row);

	// Get the pixel height policy for the row
	int pixel_height = cell_size.getPixelValue();

	if (cell_size.isHidden()) {
		pixel_height = 0;
	}
	else if(row == JCTableEnum.LABEL && !table.isColumnLabelDisplay()) {
		pixel_height = 0;
	}
	else {
		switch(pixel_height) {
		case JCTableEnum.NOVALUE:
        	// Use char height
			pixel_height = Math.max(0, cell_size.getCharValue() *
			                           getFontHeight(table, row,
                                                     start_column) +
									   offset);
			break;
		case JCTableEnum.AS_IS:
			if (current_height < 0) {
				// If a height hasn't already been determined, calculate one
                pixel_height = Math.max(0, cell_size.getCharValue() *
				                           getFontHeight(table, row,
                                                         start_column) +
                                           offset);
			}
			else {
				pixel_height = current_height;
			}
			break;
		case JCTableEnum.VARIABLE:
			if (row == JCTableEnum.ALL) {
				pixel_height = getVariableCellHeight(table,
										             JCTableEnum.LABEL, end_row,
													 start_column, end_column,
													 offset);
			}
			else if (row == JCTableEnum.ALLCELLS) {
				pixel_height = getVariableCellHeight(table, 0, end_row,
							                         start_column, end_column,
													 offset);
			}
			else {
				pixel_height = getVariableCellHeight(table, row, row,
							                         start_column, end_column,
													 offset);
			}
			break;
		case JCTableEnum.VARIABLE_ESTIMATE:
			if (row == JCTableEnum.ALL) {
				pixel_height = getVariableEstimateCellHeight(table,
												             JCTableEnum.LABEL,
															 end_row,
															 start_column,
															 end_column,
															 offset);
			}
			else if (row == JCTableEnum.ALLCELLS) {
				pixel_height = getVariableEstimateCellHeight(table, 0, end_row,
															 start_column,
															 end_column,
															 offset);
			}
			else {
				pixel_height = getVariableEstimateCellHeight(table, row, row,
															 start_column,
															 end_column,
															 offset);
			}
			break;
		default: // Fixed pixel value
			//pixel_height = Math.max(current_height, pixel_height);
			break;
		}
		// Bound pixel height by the min/max cell size settings
		pixel_height =  Math.min(cell_size.getMaxValue(),
					             Math.max(cell_size.getMinValue(),
								          pixel_height));
	}

	if (current_height != pixel_height) {
		changed = true;
	}

	setCellLayoutHeight(cellLayout, start_row, end_row, pixel_height);

	return(changed);
}

protected static int getVariableCellHeight(JCTable table, int start_row,
										   int end_row, int start_column,
										   int end_column, int offset)
{
	int pixel_height = 0;
	for (int i = start_row; i <= end_row; i++) {
		int height = getLargestCellHeight(table, start_column, end_column, i);
		pixel_height = Math.max(pixel_height, height + offset);
	}
	return(pixel_height);
}

protected static int getVariableEstimateCellHeight(JCTable table, int start_row,
												   int end_row,
                                                   int start_column,
												   int end_column, int offset)
{
	int final_height = 0;
	int pixel_height = 0;
    int numColumns = table.getNumColumns();
    int count = Math.min(table.variable_estimate_count, numColumns - 1);
	for (int i = start_row; i <= end_row; i++) {
        if (count == JCTableEnum.ALL) {
			pixel_height = getLargestCellHeight(table, start_column,
			                                    end_column, i);
		}
		else if (count == JCTableEnum.LABEL) {
			pixel_height = getLargestCellHeight(table, JCTableEnum.LABEL,
			                                    JCTableEnum.LABEL, i);
		}
		else if (count == JCTableEnum.ALLCELLS) {
			pixel_height = getLargestCellHeight(table, 0, numColumns - 1, i);
		}
		else if (count == JCTableEnum.VISIBLE_CELLS) {
            // Calculate label height
            pixel_height = getCellHeight(table, i, JCTableEnum.LABEL);

            // Calculate max height of visible cells.  If the range of visible
            // cells is not valid, assume no cells are visible.
            CellAreaHandler cellAreaHandler = table.getCellAreaHandler();
            JCCellRange span = cellAreaHandler.getVisibleCells();
            if (span.start_column != JCTableEnum.NOVALUE &&
                span.end_column != JCTableEnum.NOVALUE)
            {
                pixel_height = Math.max(pixel_height,
								        getLargestCellHeight(table,
                                                             span.start_column,
								                             span.end_column,
                                                             i));
            }

            // Calculate frozen columns cell height
            CellArea fr = cellAreaHandler.findByType(CellArea.FROZENCOLUMN);
            if (fr != null) {
                pixel_height = Math.max(pixel_height,
                                        getLargestCellHeight(table,
                                                             fr.start_column,
                                                             fr.end_column, i));
            }
		}
		else if (count >= 0) {
			pixel_height = getLargestCellHeight(table, JCTableEnum.LABEL,
                                                count, i);
		}
		pixel_height += offset;
		final_height = Math.max(final_height, pixel_height);
	}
	return(final_height);
}

protected static void setCellLayoutHeight(CellLayoutModel cellLayout,
				                          int start_row, int end_row,
										  int pixel_height)
{
	if (start_row == JCTableEnum.LABEL) {
		cellLayout.setHeight(start_row, pixel_height);
	}
	else {
		cellLayout.setHeight(start_row, end_row, pixel_height);
	}
}

/**
 * Returns the height of an individual cell.
 */
protected static int getCellHeight(JCTable table, int row, int column)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(0);
	}

	Component comp = table.getComponent(row, column);
	if (comp != null) {
		return(comp.getPreferredSize().height);
	}

	Object o = table.getDataView().getObject(row, column);
	JCCellRenderer cr = table.getDataView().getCellRenderer(row, column, o);
	return(getCellHeight(table, row, column, o, cr));
}

/**
 * Returns the height of an individual cell given its data and cell renderer.
 */
protected static int getCellHeight(JCTable table, int row, int column,
                                   Object data, JCCellRenderer cr)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(0);
	}

	TableCellInfoModel ci = table.getCellAreaHandler().getTableCellInfo();
	ci.initialize(table, data, row, column);

	Dimension d = null;
	if (cr instanceof JCLightCellRenderer) {
		JCLightCellRenderer lcr = (JCLightCellRenderer)cr;
		d = lcr.getPreferredSize(table.getGraphics(), ci, data);
	}
	else if (cr instanceof JCComponentCellRenderer) {
		JCComponentCellRenderer ccr = (JCComponentCellRenderer)cr;
        Component c = ccr.getRendererComponent(ci, data,
                                               table.isSelected(row, column));
        d = c.getPreferredSize();
	}
	if (d == null && table.getGraphics() != null &&
	    table.getGraphics().getFontMetrics() != null) {
		return(getHeight(table.getGraphics().getFontMetrics(ci.getFont()),
						 (data == null) ? "" : data.toString()));
	}
	else {
		return(d.height);
	}
}

/**
 * Returns the height of a string using the given font metrics and accounting
 * for newlines.
 */
protected static final int getHeight(FontMetrics fm, String s)
{
	if (s == null || s.length() == 0) {
		return(0);
	}
	int lines = 1;
	for (int i = 0; i < s.length(); i++) {
		if (s.charAt(i) == '\n') {
			lines++;
		}
	}
	return(fm.getHeight() * lines);
}

/**
 * Returns the largest height of a cell or label in the given row.
 */
protected static final int getLargestCellHeight(JCTable table, int start_column,
                                                int end_column, int row)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(0);
	}

	int max_height = 0;
    if (row < 0 && start_column < 0) {
    	start_column = 0;
    }
	int columns = table.getDataView().getNumColumns();
    if (end_column >= columns) {
    	end_column = columns - 1;
    }
	for (int c = start_column; c <= end_column; c++) {
		if (table.spanHandler != null) {
			JCCellRange cr = table.getSpannedRange(row, c);
			if (cr != null && cr.start_row != cr.end_row) {
				continue;
			}
		}
		max_height = Math.max(max_height, getCellHeight(table, row, c));
	}
	return(max_height);
}

/**
 * Calculates the width of the row labels.
 */
protected static boolean calcRowLabelWidth(JCTable table, int start_row,
                                           int end_row)
{
	return(calcColumnWidths(table, start_row, end_row,
	                        JCTableEnum.LABEL, JCTableEnum.LABEL));
}

/**
 * Cache the pixel width of a range of columns given a range of rows to consider.
 */
protected static boolean calcColumnWidths(JCTable table,
                                          int start_row, int end_row,
										  int start_column, int end_column)
{
    boolean changed = false;

	if (!table.repaint_enabled) {
		table.setRecalcRequired(true);
		return(false);
	}

    int numRows    = table.getNumRows();
    int numColumns = table.getNumColumns();

    int defRowEnd = (numRows == 0)    ? JCTableEnum.LABEL : numRows - 1;
    int defColEnd = (numColumns == 0) ? JCTableEnum.LABEL : numColumns - 1;

	// Check range
	if (start_row == JCTableEnum.ALL) {
		start_row = JCTableEnum.LABEL;
		end_row = defRowEnd;
	}
	else if (start_row == JCTableEnum.ALLCELLS) {
		start_row = 0;
        end_row = defRowEnd;
	}
	if (start_column == JCTableEnum.ALL) {
		start_column = JCTableEnum.LABEL;
        end_column = defColEnd;
	}
	else if (start_column == JCTableEnum.ALLCELLS) {
		start_column = 0;
        end_column = defColEnd;
	}

    // Check start points
    if (start_row == JCTableEnum.LABEL && numColumns == 0) {
        start_row = 0;
    }
    if (start_column == JCTableEnum.LABEL && numRows == 0) {
        // If the label column size is fixed we still want to calculate it
        int p = table.getColumnCellSize(JCTableEnum.LABEL).getPixelValue();
        if (p == JCTableEnum.VARIABLE || p == JCTableEnum.VARIABLE_ESTIMATE) {
            start_column = 0;
        }
    }

	// Check end points
	if (end_row == JCTableEnum.MAXINT) {
        end_row = defRowEnd;
	}
	if (end_column == JCTableEnum.MAXINT) {
        end_column = defColEnd;
	}

    if (start_row > end_row || start_column > end_column) {
        return(changed);
    }

	// Handle a single column request
	if (start_column == end_column) {
        changed |= setPixelWidth(table, start_column, start_row, end_row,
                                 start_column, end_column);
		return(changed);
	}

    // Reset the label column
    CellLayoutModel cellLayout = table.getCellLayout();
    cellLayout.setWidth(JCTableEnum.LABEL, 0);

	// Handle the label column
	if (start_column == JCTableEnum.LABEL) {
		changed |= setPixelWidth(table, start_column, start_row, end_row,
		                         start_column, end_column);
	}

    // Reset the column sizes of the table
    cellLayout.resetColumns();

	// Handle the entire table
	changed |= setPixelWidth(table, JCTableEnum.ALLCELLS, start_row, end_row,
	                         0, end_column);

	// Now handle the individual columns
	List columnWidthList = table.getColumnWidthValues();
	for (Iterator it = columnWidthList.iterator(); it.hasNext(); ) {
		PositionCellSize pcs = (PositionCellSize)it.next();
		int column = pcs.position;
		if (column >= 0) { // Non-special column numbers
			changed |= setPixelWidth(table, column, start_row, end_row,
			                         column, column);
		}
		else if (column == JCTableEnum.ALL || column == JCTableEnum.ALLCELLS) {
			CellSize cell_size = table.getColumnCellSize(column);
			int pixel_width = cell_size.getPixelValue();
			if (pixel_width == JCTableEnum.VARIABLE ||
			    pixel_width == JCTableEnum.VARIABLE_ESTIMATE)
			{
			    for (int i = start_column; i <= end_column; i++) {
				    changed |= setPixelWidth(table, i, start_row, end_row,
                                             i, i);
			    }
			}
		}
	}
	return(changed);
}

protected static boolean setPixelWidth(JCTable table, int column,
				                       int start_row, int end_row,
                                       int start_column, int end_column)
{
	boolean changed = false;
	int offset = table.getTextWidthOffset() * 2;
	CellLayoutModel cellLayout = table.getCellLayout();
	CellSize cell_size = table.getColumnCellSize(column);
	int current_width = table.getCellLayout().getWidth(column);

	// Get the pixel width policy for the column
	int pixel_width = cell_size.getPixelValue();

	if (cell_size.isHidden()) {
		pixel_width = 0;
	}
	else if (column == JCTableEnum.LABEL && !table.isRowLabelDisplay()) {
		pixel_width = 0;
	}
	else {
		switch(pixel_width) {
		case JCTableEnum.NOVALUE:  // Use char width
			pixel_width = Math.max(0, cell_size.getCharValue() *
						              getFontWidth(table, start_row, column) +
								      offset);
			break;
		case JCTableEnum.AS_IS:
			if (current_width < 0) {
				// If a width hasn't already been determined, calculate one
				pixel_width = Math.max(0, cell_size.getCharValue() *
										  getFontWidth(table, start_row, column) +
										  offset);
			}
			else {
				pixel_width = current_width;
			}
			break;
		case JCTableEnum.VARIABLE:
			if (column == JCTableEnum.ALL) {
				pixel_width = getVariableCellWidth(table, start_row, end_row,
												   JCTableEnum.LABEL,
												   end_column, offset);
			}
			else if (column == JCTableEnum.ALLCELLS) {
				pixel_width = getVariableCellWidth(table, start_row, end_row,
												   0, end_column, offset);
			}
			else {
				pixel_width = getVariableCellWidth(table, start_row, end_row,
							                       column, column, offset);
			}
			break;
		case JCTableEnum.VARIABLE_ESTIMATE:
			if (column == JCTableEnum.ALL) {
				pixel_width = getVariableEstimateCellWidth(table,
				                                           start_row, end_row,
														   JCTableEnum.LABEL,
														   end_column, offset);
			}
			else if (column == JCTableEnum.ALLCELLS) {
				pixel_width = getVariableEstimateCellWidth(table,
				                                           start_row, end_row,
														   0, end_column,
														   offset);
			}
			else {
				pixel_width = getVariableEstimateCellWidth(table,
				                                           start_row, end_row,
														   column, column,
														   offset);
			}
			break;
		default: // Fixed pixel value
			//pixel_width = Math.max(current_width, pixel_width);
			break;
		}

		// Bound pixel width by the min/max cell size settings
		pixel_width =  Math.min(cell_size.getMaxValue(),
		                        Math.max(cell_size.getMinValue(),
                                         pixel_width));
	}

	if (current_width != pixel_width) {
		changed = true;
	}

	setCellLayoutWidth(cellLayout, start_column, end_column, pixel_width);

	return(changed);
}

protected static int getVariableCellWidth(JCTable table, int start_row,
				                          int end_row, int start_column,
										  int end_column, int offset)
{
	int pixel_width = 0;
	for (int i = start_column; i <= end_column; i++) {
		int width = getLargestCellWidth(table, start_row, end_row, i);
		pixel_width = Math.max(pixel_width, width + offset);
	}
	return(pixel_width);
}

protected static int getVariableEstimateCellWidth(JCTable table, int start_row,
				                                  int end_row, int start_column,
												  int end_column, int offset)
{
	int final_width = 0;
	int pixel_width = 0;
    int numRows = table.getNumRows();
    int count = Math.min(table.variable_estimate_count, numRows - 1);;
	for (int i = start_column; i <= end_column; i++) {
		if (count == JCTableEnum.ALL) {
			pixel_width = getLargestCellWidth(table, start_row, end_row, i);
		}
		else if (count == JCTableEnum.LABEL) {
			pixel_width = getLargestCellWidth(table, JCTableEnum.LABEL,
											  JCTableEnum.LABEL, i);
		}
		else if (count == JCTableEnum.ALLCELLS) {
			pixel_width = getLargestCellWidth(table, 0, numRows - 1, i);
		}
		else if (count == JCTableEnum.VISIBLE_CELLS) {
            // Calculate label width
            int labelCellWidth = getCellWidth(table, JCTableEnum.LABEL, i);

            // Calculate max width of visible cells.  If the range of visible
            // cells is not valid, assume no cells are visible.
            CellAreaHandler cellAreaHandler = table.getCellAreaHandler();
            JCCellRange span = cellAreaHandler.getVisibleCells();
            if (span.start_row != JCTableEnum.NOVALUE &&
                span.end_row != JCTableEnum.NOVALUE)
            {
                pixel_width = Math.max(labelCellWidth,
                                       getLargestCellWidth(table,
                                                           span.start_row,
                                                           span.end_row, i));
            }

            // Calculate frozen rows cell width
            CellArea fr = cellAreaHandler.findByType(CellArea.FROZENROW);
            if (fr != null)  {
                pixel_width = Math.max(pixel_width,
                                       getLargestCellWidth(table, fr.start_row,
                                                           fr.end_row, i));
            }
		}
		else if (count >= 0) {
			pixel_width = getLargestCellWidth(table, JCTableEnum.LABEL,
                                              count, i);
		}
		pixel_width += offset;
		final_width = Math.max(final_width, pixel_width);
	}
	return(final_width);
}

protected static void setCellLayoutWidth(CellLayoutModel cellLayout,
					                     int start_column, int end_column,
										 int pixel_width)
{
	if (start_column == JCTableEnum.LABEL) {
		cellLayout.setWidth(start_column, pixel_width);
	}
	else {
		cellLayout.setWidth(start_column, end_column, pixel_width);
	}
}

/**
 * Returns the width of an individual cell.
 */
protected static int getCellWidth(JCTable table, int row, int column)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(0);
	}

	Component comp = table.getComponent(row, column);
	if (comp != null) {
		return(comp.getPreferredSize().width);
	}

	Object o = table.getDataView().getObject(row, column);
	JCCellRenderer cr = table.getDataView().getCellRenderer(row, column, o);
	return(getCellWidth(table, row, column, o, cr));
}

/**
 * Returns the width of an individual cell given its data and cell renderer.
 */
protected static int getCellWidth(JCTable table, int row, int column,
				                  Object data, JCCellRenderer cr)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(0);
	}

	TableCellInfoModel ci = table.getCellAreaHandler().getTableCellInfo();
	ci.initialize(table, data, row, column);

	Dimension d = null;
	if (cr instanceof JCLightCellRenderer) {
		JCLightCellRenderer lcr = (JCLightCellRenderer)cr;
		d = lcr.getPreferredSize(table.getGraphics(), ci, data);
	}
	else if (cr instanceof JCComponentCellRenderer) {
		JCComponentCellRenderer ccr = (JCComponentCellRenderer)cr;
        Component c = ccr.getRendererComponent(ci, data,
                                               table.isSelected(row, column));
        d = c.getPreferredSize();
	}
	if (d == null) {
		return(getWidth(table.getGraphics().getFontMetrics(ci.getFont()),
                        data.toString()));
	}
	else {
		return(d.width);
	}
}

/**
 * Returns the largest width of a cell or label in the given column.
 */
protected static final int getLargestCellWidth(JCTable table, int start_row,
						                       int end_row, int column)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(0);
	}

	int max_width = 0;
    if (column < 0 && start_row < 0) {
    	start_row = 0;
    }
	int rows = table.getDataView().getNumRows();
    if (end_row >= rows) {
    	end_row = rows - 1;
    }
	for (int r = start_row; r <= end_row; r++) {
		if (table.spanHandler != null) {
			JCCellRange cr = table.getSpannedRange(r, column);
			if (cr != null && cr.start_column != cr.end_column) {
				continue;
			}
		}
		max_width = Math.max(max_width, getCellWidth(table, r, column));
	}
	return(max_width);
}

/**
 * Returns the width of a string using the given font metrics and accounting
 * for newlines.
 */
protected static final int getWidth(FontMetrics fm, String s)
{
	if (s == null || s.length() == 0) {
		return(0);
	}
	if (s.indexOf('\n') != -1) {
		int start = 0, end, width = 0;
		for (; (end = s.indexOf('\n', start)) != -1; start = end + 1) {
			width = Math.max(width, fm.stringWidth(s.substring(start, end)));
		}
		return(Math.max(width, fm.stringWidth(s.substring(start, s.length()))));
	}
	else {
		return(fm.stringWidth(s));
	}
}

/**
 * Returns the dimensions of an individual cell.
 */
protected static Dimension getCellDimensions(JCTable table, int row, int column)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(new Dimension(0, 0));
	}

	Object o = table.getDataView().getObject(row, column);
	JCCellRenderer cr = table.getDataView().getCellRenderer(row, column, o);
	return(getCellDimensions(table, row, column, o, cr));
}

/**
 * Returns the dimensions of an individual cell given its data and cell
 * renderer.
 */
protected static Dimension getCellDimensions(JCTable table, int row, int column,
				                             Object data, JCCellRenderer cr)
{
	if (table.getGraphics() == null) {
		table.setRecalcRequired(true);
		return(new Dimension(0, 0));
	}

	TableCellInfoModel ci = table.getCellAreaHandler().getTableCellInfo();
	ci.initialize(table, data, row, column);

	Dimension d = null;
	if (cr instanceof JCLightCellRenderer) {
		JCLightCellRenderer lcr = (JCLightCellRenderer)cr;
		d = lcr.getPreferredSize(table.getGraphics(), ci, data);
	}
	else if (cr instanceof JCComponentCellRenderer) {
		JCComponentCellRenderer ccr = (JCComponentCellRenderer)cr;
        Component c = ccr.getRendererComponent(ci, data,
                                               table.isSelected(row, column));
        d = c.getPreferredSize();
	}
	if (d == null) {
		FontMetrics fm = table.getGraphics().getFontMetrics(ci.getFont());
		return(new Dimension(getWidth(fm, data.toString()),
		                     getHeight(fm, data.toString())));
	}
	else {
		return(d);
	}
}

/**
 * Returns the height of the font used in a cell.
 */
protected static int getFontHeight(JCTable table, int row, int column)
{
	Font font = table.getCellStyle(row, column).getFont();
	if (font == null) {
		return(0);
	}
	// It is expensive to get a FontMetrics and get the height
	// from it.  As a result, we cache the FontMetrics.getHeight()
	// result for Font values.  Works very well when character
	// width/height are used.
	if (cachedFontHeights == null) {
		// Created either the first time or after someone calls
		// JCTable.dispose()
		cachedFontHeights = new Hashtable();
	}
	Object fn = cachedFontHeights.get(font);
	if (fn != null && fn instanceof Integer) {
		return(((Integer)fn).intValue());
	}
	int rval = table.getFontMetrics(font).getHeight();
	cachedFontHeights.put(font, new Integer(rval));
	return(rval);
}

/**
 * Returns the width of the font used in a cell based on the character
 * <CODE>W</CODE>.
 */
protected static int getFontWidth(JCTable table, int row, int column)
{
	Font font = table.getCellStyle(row, column).getFont();
	if (font == null) {
		return(0);
	}
	return(table.getFontMetrics(font).charWidth('W'));
}

}
