/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: ComponentHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;


import com.klg.jclass.util.JCUtil;

import java.awt.*;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.Serializable;


/**
 * <CODE>ComponentHandler</CODE> controls the placement of components within
 * table and ensures that traverse from a component to table or table to
 * component flows across and down the cells or to the left and up for reverse
 * traversal.
 */
public abstract class ComponentHandler extends JCSeries
							           implements ComponentModel,
										          Serializable,
										          JCScrollListener,
										          FocusListener,
										          KeyListener
{

protected JCTable        table                = null;
protected boolean        adjustBoundsRequired = true;
protected boolean        needsSetParent       = true;
protected boolean        shiftDown            = false;
protected boolean        controlDown          = false;

/**
 * Creates a new ComponentHandler instance. <CODE>setTable</CODE> must be called to
 * connect the handler to table.
 */
public ComponentHandler()
{
	super();
}

/**
 * Creates new ComponentHandler instance linked to table.
 *
 * @param table The table for this component handler
 */
public ComponentHandler(JCTable table)
{
	setTable(table);
}

/**
 * Returns the JCTable.
 *
 * @return The table for this component handler
 */
public JCTable getTable()
{
	return(table);
}

/**
 * Sets the JCTable.
 *
 * @param table The new table for this component handler
 */
public void setTable(JCTable table)
{
	this.table = table;
	this.table.addScrollListener(this);
}

/**
 * Returns true if one or more components have not been added to table yet.
 *
 * @return Whether one or more components have not been added to table yet.
 * @see #setParentSetRequired
 */
public boolean isParentSetRequired()
{
	return(needsSetParent);
}

/**
 * Flags that one or more components do not have a parent yet. The appropriate
 * container for components will not be present before table is realized.
 *
 * @param value The new value of parentSetRequired which tells us whether one
 *     or more components do not have a parent yet.
 */
public void setParentSetRequired(boolean value)
{
	this.needsSetParent = value;
}

/**
 * Add key and focus listeners to component if it is focusable, otherwise add
 * them the first child which is focusable
 *
 * @param c The component to which to add the listeners
 * @param keyListener The key listener to add
 * @param focusListener  The focusListener to add
 */
protected void addKeyAndFocusListeners(Component c, KeyListener keyListener,
							           FocusListener focusListener)
{
	
	c = table.getEditTraverseHandler().getFocusableChild(c);
	if (c == null) {
		return;
	}

	// Remove first to ensure that we don't get duplicates
	c.removeKeyListener(keyListener);
	c.addKeyListener(keyListener);

	// Remove first to ensure that we don't get duplicates
	c.removeFocusListener(focusListener);
	c.addFocusListener(focusListener);

	
}

protected abstract void addListeners(Component c, KeyListener keyListener,
							         FocusListener focusListener);

/**
 * Overrides setDefault in order to add key and focus listeners.
 *
 * @param value The component value on which to add listeners.
 */
public void setDefault(Object value)
{
	Component component = (Component)value;
	

	// Add listeners to the component (or one of its children)
	addListeners(component, this, this);

	super.setDefault(value);
}

/**
 * Overrides setValue in order to add key and focus listeners.
 * Setting a null value removes an existing component at this
 * location.
 *
 * @param row The row at which the new component is to be added
 * @param column The column at which the new component is to be added
 * @param value The new component value
 *
 * @return Whether setValue succeeded.
 */
public boolean setValue(int row, int column, Object value)
{
	
	Component component			= (Component)value;
	Component existingComponent	= table.getComponent(row, column);
	CellArea  area 				= table.cellAreaHandler.find(row, column);
	boolean   returnValue		= true;

	// First, remove any existing component at this location.
	if (existingComponent != null) {
		Component focusComponent =
		    table.getEditTraverseHandler().getFocusableChild(existingComponent);
		focusComponent.removeKeyListener(this);
		focusComponent.removeFocusListener(this);
		super.remove(row, column);
		if (area != null) {
			area.remove(existingComponent);
		}
	}

	// Add the new component if it is not null.
	if (component != null) {
		// component already registered with another cell?  remove it!
		JCCellPosition findPos = null;
		if ((findPos = findPosition(component)) != null) {
			super.remove(findPos.row, findPos.column);
		}

	    // Add a listeners to the component (or one of its children)
	    addListeners(component, this, this);

		if (area != null) {
			area.add(component);
		}
		else {
			needsSetParent = true;
		}

		returnValue = super.setValue(row, column, value);
	}

    table.cellAreaHandler.getCellAreaRenderer().repaintRange(row, column,
															 row, column);
	return(returnValue);
}

/**
 * Ensures that all components are added to their appropriate CellArea.
 */
public void addComponentsToParent()
{
	
	for (int i = 0; i < size(); i++) {
		SeriesValue sv   = (SeriesValue)elementAt(i);
		CellArea    area = table.cellAreaHandler.find(sv.row, sv.column);
		area.add((Component)sv.value);
	}
}

/**
 * Remove components when a span is added. Any components not contained
 * in the span will re-added in CellAreaRenderer.
 */
public void removeCellComponents()
{
	
	// This removes all CellArea.CELL components when a span is added.
	// If the component was not contained in the span, it is
	// re-added in CellAreaRenderer. This fixes a bug where components
	// contained in a span which are not the controlling component were
	// still drawn. They were drawn because they were still managed by
	// the container, CellArea.CELL.

	for (int i = 0; i < size(); i++) {
		SeriesValue sv   = (SeriesValue)elementAt(i);
		CellArea    area = table.cellAreaHandler.find(sv.row, sv.column);
		if (area != null && area.getType() == CellArea.CELL) {
			Component components[] = area.getComponents();
			for (int j = 0; j < components.length; j++) {
				area.remove(components[j]);
			}
		}
	}

	needsSetParent = false;
}

/**
 * Determines the cell position of the given component.
 *
 * @param c The component whose position we want to find
 *
 * @return An object which contains the cell position of the component
 */
public JCCellPosition findPosition(Component c)
{
	
	SeriesValue sv;
	for (int i = 0; i < size(); i++) {
		sv = (SeriesValue)elementAt(i);
		if (c == (Component)sv.value) {
			return(new JCCellPosition(sv.row, sv.column));
		}
	}
	return(null);
}

/**
 * Required by JCScrollListener but not implemented.
 *
 * @param e The event which causes the scroll
 */
public void scroll(JCScrollEvent e) {}

/**
 * Controls the visibility of components when scrolled.
 *
 * @param e The event which caused the scroll
 */
public void afterScroll(JCScrollEvent e)
{
	reconfigureComponents();
}

/**
 * Reconfigure components such that components in visible cells are visible
 * and components in invisible cells are invisible.
 */
public void reconfigureComponents()
{
	
    // If table is tracking, ignore this event.  Otherwise, components will
	// try to draw when rest of table is trying not to draw.
    if (table.isTracking()) {
	    return;
    }

	JCCellRange cr = table.getVisibleCells();
	for (int i = 0; i < size(); i++) {
		SeriesValue sv        = (SeriesValue)elementAt(i);
		Component   component = (Component)sv.value;
		if (sv.row >= cr.start_row       &&
		    sv.row <= cr.end_row         &&
		    sv.column >= cr.start_column &&
			sv.column <= cr.end_column)
		{
			component.setVisible(true);
		}
		else {
			component.setVisible(false);
		}
	}
}

/**
 * Get the position of the component within the table.  If the component is
 * not found, trying querying the parent.  This is because the event which
 * caused this query could have originated from a child of a component which
 * is not itself focusable.
 *
 * @param component The component whose position we are interested in finding
 *
 * @return The position of the cell within the table
 * @see EditTraverseHandler#getFocusableChild
 */
protected JCCellPosition getTablePosition(Component component)
{
	JCCellPosition cp = table.getPosition(component);
	if (cp == null) {
		Container parent = component.getParent();
		if (parent instanceof Component) {
		    cp = table.getPosition(parent);
		}
	}
	int row = (cp == null) ? -1 : cp.row;
	int col = (cp == null) ? -1 : cp.column;
	
	return(cp);
}

/**
 * Adds a <CODE>KeyListener</CODE> to the component when table is managing
 * component focus.  Cancels any edit and cause a traversal to the underlying
 * cell.
 *
 * @param event The focus event which caused foucs to be gained.
 */
public abstract void focusGained(FocusEvent event);

/**
 * Removes TraverseHandler as a key listener on the component when it loses
 * focus.
 *
 * @param event The focus event which caused foucs to be lost.
 */
public abstract void focusLost(FocusEvent event);

/**
 * Update table to reflect that the passed in component in its cell is now the
 * new current cell (it has focus).  The new and old cells must be repainted to
 * reflect this change.
 *
 * @param component The component that has received the focus.
 * @return Whether the component is allowed to get focus.
 */
protected boolean updateTable(Component component)
{
	
	// If component in non-traversable cell, don't allow it to get focus
	JCCellPosition cp = getTablePosition(component);
    if (cp == null                     ||
	    cp.row == JCTableEnum.LABEL    ||
		cp.column == JCTableEnum.LABEL ||
		!table.isTraversable(cp.row, cp.column))
	{
		table.getFocusManager().requestFocus();
		
		return(false);
	}

	int old_row    = table.current_row;
	int old_column = table.current_column;

   	table.current_row    = cp.row;
   	table.current_column = cp.column;

	// Repaint both previous cell and this one so that the focus rectangle ends
	// up in the right spot
	CellAreaRenderer car = table.cellAreaHandler.cellAreaRenderer;
	car.repaintRange(old_row, old_column, old_row, old_column);
	car.repaintRange(table.current_row, table.current_column,
	                 table.current_row, table.current_column);
	
	return(true);
}

/**
 * Compute the direction from a KeyEvent.
 *
 * @param event The given key event
 * @return The computed direction.  One of JCTableEnum.TRAVERSE_LEFT,
 *     JCTableEnum.TRAVERSE_RIGHT, JCTableEnum.TRAVERSE_UP,
 *     JCTableEnum.TRAVERSE_DOWN or JCTableEnum.NOVALUE (on error).
 */
protected int getDirectionFromKeyEvent(KeyEvent event)
{
	
    int direction = JCTableEnum.NOVALUE;
	if (event == null) {
		return(direction);
	}

    // If this is an arrow key or TAB, move the focus
	String dirString = "None";
	controlDown = shiftDown = false;
    switch (event.getKeyCode()) {
	case KeyEvent.VK_TAB:
	    if (event.isControlDown()) {
			controlDown = true;
	    }
        if (event.isShiftDown()) {
			shiftDown = true;
			direction = JCTableEnum.TRAVERSE_LEFT;
			dirString = "Left";
		}
		else {
			direction = JCTableEnum.TRAVERSE_RIGHT;
			dirString = "Right";
		}
        break;
    case KeyEvent.VK_DOWN:
        direction = JCTableEnum.TRAVERSE_DOWN;
		dirString = "Down";
        break;
    case KeyEvent.VK_UP:
        direction = JCTableEnum.TRAVERSE_UP;
		dirString = "Up";
        break;
    case KeyEvent.VK_LEFT:
        direction = JCTableEnum.TRAVERSE_LEFT;
		dirString = "Left";
        break;
    case KeyEvent.VK_RIGHT:
        direction = JCTableEnum.TRAVERSE_RIGHT;
		dirString = "Right";
        break;
    }
	
	return(direction);
}

/**
 * Tracks TAB/SHIFT-TAB and keypad arrow keys to control navigation when a
 * component has focus.
 *
 * @param event The event from the key press
 */
public abstract void keyPressed(KeyEvent event);

/**
 * Required by KeyListener but not implemented.
 *
 * @param event The event from the key release
 */
public void keyReleased(KeyEvent event) {}

/**
 * Required by KeyListener but not implemented.
 *
 * @param event The event from the key being typed
 */
public void keyTyped(KeyEvent event) {}

/**
 * Traverses to the next cell when a TAB or SHIFT-TAB is detected on the
 * component.
 *
 * @param cur_pos The current cell position
 * @param type The desired direction in which to move, one of
 *     JCTableEnum.TRAVERSE_UP, JCTableEnum.TRAVERSE_DOWN,
 *     JCTableEnum.TRAVERSE_LEFT, JCTableEnum.TRAVERSE_RIGHT.
 */
public void traverseNext(JCCellPosition cur_pos, int type)
{
	
	JCCellPosition new_pos = null;

    // Find where to put the focus next
    new_pos = table.keyHandler.findTraversableCell(type, cur_pos.row,
												   cur_pos.column);

    // And move there
	if (new_pos != null) {
		table.getEditTraverseHandler().traverse(new_pos.row, new_pos.column,
												false, null, type, false);
	}
}

}
