/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DataViewSupport.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

/**
 * Support methods for data view move, delete, insert, sort and drag actions.
 */
class DataViewSupport {

/**
 * Drag columns calls moveColumns to redo the series.
 */
protected static boolean dragColumn(JCTable table, int source, int destination)
{
    return(moveColumns(table, source, 1, destination));
}

/**
 * Moves a range of columns. All attributes (such as CellStyle)  move along
 * with the columns.
 * @param source  the first column in the range
 * @param num_columns the number of columns to move
 * @param destination the column number before which to insert the moved columns
 * @return false if any argument is invalid
 */
protected static boolean moveColumns(JCTable table, int source, int num_columns,
                                     int destination)
{
    if (!validateMoveColumns(table, source, num_columns, destination)) {
        return(false);
    }
    if (source + num_columns == destination) {
        return(true);
    }
    moveColumnSeries(table, source, num_columns, destination);
    return(true);
}

/**
 * Moves column series.
 */
protected static void moveColumnSeries(JCTable table, int from, int amount,
                                       int to)
{
    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.moveColumns(from, amount, to);
        }
    }
    if (table.columnWidths != null) {
        table.columnWidths.move(from, amount, to);
        if (table.cellLayout != null) {
            table.cellLayout.moveColumns(from, amount, to);
        }
    }
}

/**
 * Swaps all columns given an original and new array.  All attributes (such as
 * the CellStyle) move along with the columns.
 * @param table the table
 * @param cur_columns  the current array of columns
 * @param new_columns the new array of columns
 * @return false if any argument is invalid
 */
static boolean swapColumnArrays(JCTable table, int cur_columns[],
                                int new_columns[])
{
    // Create an inverse list
    int ncolumns[] = new int[new_columns.length];
    int ccolumns[] = new int[new_columns.length];
    int scolumns[] = new int[new_columns.length];
    for (int i = 0; i < new_columns.length; i++) {
        ccolumns[cur_columns[i]] = i;
    }
    for (int i = 0; i < new_columns.length; i++) {
        ncolumns[new_columns[i]] = i;
    }
    for (int i = 0; i < new_columns.length; i++) {
        scolumns[i] = ncolumns[cur_columns[i]];
    }

    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.remapColumns(scolumns);
        }
    }
    if (table.columnWidths != null) {
        table.columnWidths.remap(scolumns);
        if (table.cellLayout != null) {
            table.cellLayout.remapColumns(ccolumns, new_columns);
        }
    }
    return(true);
}

/**
 * Swaps a pair of columns. All attributes (such as the CellStyle)
 * move along with the columns.
 * @param table the table
 * @param column1 the first column in the swap
 * @param column2 the second column in the swap
 * @return false if any argument is invalid
 */
protected static boolean swapColumns(JCTable table, int column1, int column2)
{
    // Move series values
    swapColumnSeries(table, column1, column2);

    if (table.cellLayout != null) {
        int p1 = table.cellLayout.getWidth(column1);
        int p2 = table.cellLayout.getWidth(column2);
        table.cellLayout.setWidth(column1, p2);
        table.cellLayout.setWidth(column2, p1);
    }

    return(true);
}

/**
 * Swaps column series
 */
protected static void swapColumnSeries(JCTable table, int column1, int column2)
{
    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.swapColumns(column1, column2);
        }
    }
    if (table.columnWidths != null) {
        table.columnWidths.swap(column1, column2);
    }
}

/**
 * Validate moveColumns arguments
 */
protected static boolean validateMoveColumns(JCTable table, int source,
                                             int num_columns, int destination)
{
    int columns = table.getNumColumns();

    if (num_columns <= 0) {
        return(false);
    }
    if (source < 0 || source + num_columns > columns) {
        return(false);
    }
    if (destination < 0 ||
        destination > columns ||
        (source <= destination && destination < source + num_columns))
    {
        return(false);
    }

    if (num_columns >= columns) {
        return(false);
    }
    return(true);
}

/**
 * Responds to datasource column addition by shifting any attributes
 * (such as background color) below the insertion point.
 * @param position of the column index where the the new column
 * should be added. If set to JCTblEnum.MAXINT, the column is added after
 * the last column, rather than before.
 */
protected static void addColumn(JCTable table, int position, int num_affected,
                                int num_rows, int num_columns)
{
    // Shift all series
    if (position < num_columns - num_affected) {
        shiftColumnSeries(table, position, num_affected, +1);
    }
    if (table.cellLayout != null) {
        table.cellLayout.addColumns(position, num_affected);
    }

    for (int c = position; c < position + num_rows; c++) {
        table.requestChange(TableChangeHandler.TABLE_RECALC_COLUMN_WIDTHS,
                            JCTableEnum.NOVALUE, c);
    }
}

/**
 * Responds to datasource column deletion by shifting any attributes (such as
 * background color).
 */
static void deleteColumn(JCTable table, int position, int num_affected,
                         int num_rows, int num_columns)
{
    shiftColumnSeries(table, position, num_affected, -1);
    if (table.cellLayout != null) {
        table.cellLayout.deleteColumns(position, num_affected);
    }
}

/**
 * Shifts column series starting at the given position by the number of column in
 * a positive or negative direction. This method is used for handling inserts
 * and deletes.
 */
protected static void shiftColumnSeries(JCTable table, int position,
                                        int num_affected, int direction)
{
    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.shiftColumn(position, num_affected, direction);
        }
    }
    if (table.columnWidths != null) {
        table.columnWidths.shift(position, num_affected, direction);
    }
}

/**
 * Drag rows calls moveRows to redo the series.
 */
protected static boolean dragRow(JCTable table, int source, int destination)
{
    return(moveRows(table, source, 1, destination));
}

/**
 * Moves a range of rows. All attributes (such as CellStyle)  move along
 * with the rows.
 * @param source  the first row in the range
 * @param num_rows the number of rows to move
 * @param destination the row number before which to insert the moved rows
 * @return false if any argument is invalid
 */
protected static boolean moveRows(JCTable table, int source, int num_rows,
                                  int destination)
{
    if (!validateMoveRows(table, source, num_rows, destination)) {
        return(false);
    }
    if (source + num_rows == destination) {
        return(true);
    }
    moveRowSeries(table, source, num_rows, destination);

    return(true);
}

/**
 * Moves row series.
 */
protected static void moveRowSeries(JCTable table, int from, int amount, int to)
{
    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.moveRows(from, amount, to);
        }
    }
    if (table.rowHeights != null) {
        table.rowHeights.move(from, amount, to);
        if (table.cellLayout != null) {
            table.cellLayout.moveRows(from, amount, to);
        }
    }
}

/**
 * Swaps a pair of rows. All attributes (such as the CellStyle)
 * move along with the rows.
 * @param table the table
 * @param row1 the first row in the swap
 * @param row2 the second row in the swap
 * @return false if any argument is invalid
 */
protected static boolean swapRows(JCTable table, int row1, int row2)
{
    // Move series values
    swapRowSeries(table, row1, row2);

    if (table.cellLayout != null) {
        int p1 = table.cellLayout.getHeight(row1);
        int p2 = table.cellLayout.getHeight(row2);
        table.cellLayout.setHeight(row1, p2);
        table.cellLayout.setHeight(row2, p1);
    }

    return(true);
}

/**
 * Swaps all rows given an original and new array.  All attributes (such as the
 * CellStyle) move along with the rows.
 * @param table the table
 * @param cur_rows  the current array of rows
 * @param new_rows the new array of rows
 * @return false if any argument is invalid
 */
protected static boolean swapRowArrays(JCTable table, int cur_rows[],
                                       int new_rows[])
{
    // Create an inverse list
    int nrows[] = new int[new_rows.length];
    int crows[] = new int[new_rows.length];
    int srows[] = new int[new_rows.length];
    for (int i = 0; i < new_rows.length; i++) {
        crows[cur_rows[i]] = i;
    }
    for (int i = 0; i < new_rows.length; i++) {
        nrows[new_rows[i]] = i;
    }
    for (int i = 0; i < new_rows.length; i++) {
        srows[i] = nrows[cur_rows[i]];
    }

    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.remapRows(srows);
        }
    }
    if (table.rowHeights != null) {
        table.rowHeights.remap(srows);
        if (table.cellLayout != null) {
            table.cellLayout.remapRows(crows, new_rows);
        }
    }

    return(true);
}

/**
 * Swaps column series
 */
protected static void swapRowSeries(JCTable table, int row1, int row2)
{
    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.swapRows(row1, row2);
        }
    }
    if (table.rowHeights != null) {
        table.rowHeights.swap(row1, row2);
    }
}

/**
 * Validate moveColumns arguments
 */
protected static boolean validateMoveRows(JCTable table, int source,
                                          int num_rows, int destination)
{
    int rows = table.getNumRows();

    if (num_rows <= 0) {
        return(false);
    }
    if (source < 0 || source + num_rows > rows) {
        return(false);
    }
    if (destination < 0 ||
        destination > rows ||
        (source <= destination && destination < source + num_rows))
    {
        return(false);
    }
    if (num_rows >= rows) {
        return(false);
    }
    return(true);
}

/**
 * Responds to datasource row addition by shifting any attributes (such as
 * background color) below the insertion point.
 * @param position of the index where the new row
 * should be added. If set to JCTblEnum.MAXINT, the row is added after
 * the last row, rather than before.
 */
protected static void addRow(JCTable table, int position, int num_affected,
                             int num_rows, int num_columns)
{
    // if rows were inserted, shift all series
    if (position < num_rows - num_affected) {
        shiftRowSeries(table, position, num_affected, +1);
    }
    if (table.cellLayout != null) {
        table.cellLayout.addRows(position, num_affected);
    }

    // force a layout of all rows under this row
    for (int r = position; r < num_rows - position; r++) {
        table.requestChange(TableChangeHandler.TABLE_RECALC_ROW_HEIGHTS, r,
                            JCTableEnum.NOVALUE);
    }
}

/**
 * Responds to datasource row deletion by shifting any attributes (such as
 * background color).
 */
static void deleteRow(JCTable table, int position, int num_affected,
                      int num_rows, int num_columns)
{
    shiftRowSeries(table, position, num_affected, -1);
    if (table.cellLayout != null) {
        table.cellLayout.deleteRows(position, num_affected);
    }
}

/**
 * Shifts row series starting at the given position by the number of row in
 * a positive or negative direction. This method is used for handling inserts
 * and deletes.
 */
protected static void shiftRowSeries(JCTable table, int position,
                                     int num_affected, int direction)
{
    if (table.moveables != null) {
        for (int i = 0; i < table.moveables.size(); i++) {
            Moveable moveable = (Moveable)table.moveables.elementAt(i);
            moveable.shiftRow(position, num_affected, direction);
        }
    }
    if (table.rowHeights != null) {
        table.rowHeights.shift(position, num_affected, direction);
    }
}

// Adjust selected cells after columns moved
/*
static void selectColumns(Table table, int source, int amount, int dest)
{
	if (table.selected_cells.size() == 0) {
		return;
    }
	for (int i=table.selected_cells.size()-1; i >= 0; i--) {
		JCCellRange r = (JCCellRange) table.selected_cells.elementAt(i);
		if (r.end_column == JCTblEnum.MAXINT) {
			continue;
        }
		int col1 = Math.min(r.start_column, r.end_column);
		int col2 = Math.max(r.start_column, r.end_column);

		// If range will be affected, delete it
		if ((col2 >= source && col1 <= source+amount-1) ||
            (col1 < dest && dest <= col2))
        {
			table.selected_cells.removeElementAt(i);
		}
		// Otherwise move range as a block
		else {
			r.start_column = adjustPosition(r.start_column, source, amount,
                                            dest);
			r.end_column = adjustPosition(r.end_column, source, amount, dest);
		}
	}
}
*/

// Adjust selected cells after rows moved
/*
static void selectRows(Table table, int source, int amount, int dest)
{
	if (table.selected_cells.size() == 0) {
		return;
    }
	for (int i=table.selected_cells.size()-1; i >= 0; i--) {
		JCCellRange r = (JCCellRange) table.selected_cells.elementAt(i);
		if (r.end_row == JCTblEnum.MAXINT) {
			continue;
        }
		int row1 = Math.min(r.start_row, r.end_row);
		int row2 = Math.max(r.start_row, r.end_row);

		// If range will be affected, delete it
		if ((row2 >= source && row1 <= source+amount-1)
				|| (row1 < dest && dest <= row2)) {
			table.selected_cells.removeElementAt(i);
		}
		// Otherwise move range as a block
		else {
			r.start_row = adjustPosition(r.start_row, source, amount, dest);
			r.end_row = adjustPosition(r.end_row, source, amount, dest);
		}
	}
}
*/

// Adjust spanned cells after columns moved
/*
static void spanColumns(Table table, int source, int amount, int dest)
{
	if (table.span.span_list_orig == null ||
        table.span.span_list_orig.size() == 0)
    {
		return;
    }
	for (int i = 0; i < table.span.span_list_orig.size(); i++) {
		JCCellRange r = (JCCellRange) table.span.span_list_orig.elementAt(i);
		int col1 = r.start_column, col2 = r.end_column;
		if (col1 == JCTblEnum.ALL ||
            col1 == JCTblEnum.ALLCELLS ||
            col2 == JCTblEnum.ALL ||
            col2 == JCTblEnum.ALLCELLS ||
            col2 == JCTblEnum.MAXINT)
        {
			continue;
        }

		// Move span source
		r.start_column = adjustPosition(r.start_column, source, amount, dest);
		r.end_column = r.start_column + col2 - col1;
	}
}
*/

// Adjust spanned cells after rows moved
/*
static void spanRows(Table table, int source, int amount, int dest)
{
	if (table.span.span_list_orig == null ||
        table.span.span_list_orig.size() == 0)
    {
		return;
    }
	for (int i=0; i < table.span.span_list_orig.size(); i++) {
		JCCellRange r = (JCCellRange) table.span.span_list_orig.elementAt(i);
		int row1 = r.start_row, row2 = r.end_row;
		if (row1 == JCTblEnum.ALL ||
            row1 == JCTblEnum.ALLCELLS ||
            row2 == JCTblEnum.ALL ||
            row2 == JCTblEnum.ALLCELLS ||
            row2 == JCTblEnum.MAXINT)
        {
			continue;
        }

		// Move span source
		r.start_row = adjustPosition(r.start_row, source, amount, dest);
		r.end_row = r.start_row + row2 - row1;
	}
}
*/

/**
 * Calculate a new row or column position after some rows or columns
 * have been moved.
 * @return the original value, shifted.
 */
/*
static int adjustPosition(int value, int source, int amount, int destination)
{
	if (source > destination) {
		if (value >= destination && value < source) {
			value += amount;
        }
		else if (value >= source && value < source + amount) {
			value -= source - destination;
        }
	}
	else {
		if (value >= source + amount && value < destination) {
			value -= amount;
        }
		else if (value >= source && value < source + amount) {
			value += destination - source - amount;
        }
	}
	return(value);
}
*/

}
