/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: EditTraverseHandlerAbstract.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.cell.JCCellEditor;
import com.klg.jclass.cell.JCCellEditorEvent;
import com.klg.jclass.cell.JCCellEditorListener;

import com.klg.jclass.util.JCUtil;

import javax.swing.*;
import java.awt.*;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.Serializable;

/**
 * <CODE>EditTraverseHandlerOld</CODE> controls the cell traversal and editor
 * display/commit/cancel logic for JDK's prior to 1.4. It also fires
 * <CODE>TraverseCell</CODE> and <CODE>EditCell</CODE> events as required.
 */
public abstract class EditTraverseHandlerAbstract implements EditTraverseHandler,
											                 FocusListener,
											                 JCCellEditorListener,
											                 Serializable {

protected JCTable          table;
protected JCCellEditor     editor;
protected Component        component;
protected int              edit_row;
protected int              edit_column;

/**
 * Constructor that requires a table
 *
 * @param table The table for which this EditTraverseHandler is used.
 */
public EditTraverseHandlerAbstract(JCTable table)
{
	this.table = table;
}

/**
 * Returns the current cell editor.
 *
 * @return The current cell editor.  Returns <CODE>null</CODE> if none.
 */
public JCCellEditor getCellEditor()
{
	return(editor);
}

/**
 * Add key and focus listeners to component if it is focusable, otherwise add
 * them the first child which is focusable
 *
 * @param c The component to which to add the listeners
 * @param keyListener The key listener to add
 * @param focusListener  The focusListener to add
 */
protected void addKeyAndFocusListeners(Component c, KeyListener keyListener,
							           FocusListener focusListener)
{
	

	if (c == null) {
		return;
	}

	// Remove first to ensure that we don't get duplicates
	c.removeKeyListener(keyListener);
	c.addKeyListener(keyListener);

	// Remove first to ensure that we don't get duplicates
	c.removeFocusListener(focusListener);
	c.addFocusListener(focusListener);

	
}

protected abstract void addListeners(Component c, KeyListener keyListener,
							         FocusListener focusListener);

/**
 * Sets the cell editor.
 *
 * @param ed The new cell editor
 */
public void setCellEditor(JCCellEditor ed)
{
	
	if (ed != null) {
		component = ed.getComponent();
		if (component != null) {
			addListeners(component, table.keyHandler, this);
		}
		ed.addCellEditorListener(this);
	}
	else {
		if (this.editor != null) {
			// Remove listener
			this.editor.removeCellEditorListener(this);
		}
		if (component != null) {
			// Note that the focus listener does not get removed here as one
			// might expect.  See the note in the focusLost() method below.
			component.removeKeyListener(table.keyHandler);
		}
		component = null;
	}
	this.editor      = ed;
	this.edit_row    = table.getCurrentRow();
	this.edit_column = table.getCurrentColumn();
}

/**
 * Get the row of the current cell being edited.
 *
 * @return The current edit row
 */
public int getEditRow()
{
	return(edit_row);
}

/**
 * Get the column of the current cell being edited.
 *
 * @return The current edit column
 */
public int getEditColumn()
{
	return(edit_column);
}

/**
 * Initialize the editor based on a cellInfo object and data.  Note that this
 * method is overridden in EditTraverseHandlerNew
 * @param e The event which caused the editor to be invoked
 * @param cellInfo The table cell information
 * @param o The data object
 */
public void initializeEditor(AWTEvent e, TableCellInfoModel cellInfo, Object o)
{
	
    editor.initialize(e, cellInfo, o);
}

/**
 * Traverse to the specified cell.
 *
 * @param row The row to which to traverse
 * @param column The column to which to traverse
 * @param select Should the cell be selected
 * @param event The event that caused this traverse (null if none)
 * @param traverse_type The type of traversal (right, left, down, up, home, ...)
 * @param show_editor Display the component, if called programatically
 *
 * @return false if the cell is not traversable
 */
public boolean traverse(int row, int column, boolean select, AWTEvent event,
						int traverse_type, boolean show_editor)
{
	
	if (row == JCTableEnum.NOVALUE || column == JCTableEnum.NOVALUE) {
		return(false);
	}

	// Don't want a traverse to start with a shift key
	if (event != null && event instanceof KeyEvent) {
		KeyEvent ke = (KeyEvent)event;
		if (ke.getKeyCode() == KeyEvent.VK_SHIFT) {
			return(false);
		}
	}

	// Get the origin of a span if the cell is contained in one.
	JCCellRange span = table.getSpannedRange(row, column);
	if (span != null) {
		row    = span.start_row;
		column = span.start_column;
	}

	if (!traverseInitial(row, column, select, traverse_type)) {
		return(false);
	}

	// NOTE: During the traverseInitial a listener can change the
	// row and column to be different than those passed in. So,
	// resynch row and column with the potentially new row and column.
	row 	= table.getCurrentRow();
	column	= table.getCurrentColumn();

	Component comp = table.getComponent(row, column);
	
	if (comp != null) {
		table.getFocusManager().requestFocus(comp);
		table.setCurrentCell(row, column);
		return(true);
	}

	if (event == null && show_editor) {
	    table.clearSelectedCells();
	}

	// Check for auto edit
	if (table.auto_edit) {
		show_editor = true;
	}

	// Check for the editable state
	if (!table.getDataView().isEditable() ||
	    !table.getCellStyle(row, column).isEditable())
	{
	   	show_editor = false;
	}

	if (show_editor) {
		JCEditCellEvent ev = null;
		if (table.editCellListeners != null) {
			ev = new JCEditCellEvent(table, JCEditCellEvent.BEFORE_EDIT_CELL,
									 row, column);
			table.fireJCEditCellEvent(ev);
			if (ev.isCancelled()) {
				ev.type = JCEditCellEvent.AFTER_EDIT_CELL;
				table.fireJCEditCellEvent(ev);
				return(false);
			}
		}

		// Unmap the component to avoid flashing
		JCCellEditor ce = table.getDataView().getCellEditor(row, column);
		if (editor != null && editor != ce) {
			cancel();
		}

		Object o = table.getDataView().getTableDataItem(row, column);
		setCellEditor(ce);
		CellAreaHandler car = table.getCellAreaHandler();
		TableCellInfoModel cellInfo = car.getTableCellInfo();
		cellInfo.initialize(table, o, row, column);

		if (editor != null) {
			Rectangle cell_size = car.getBounds(row, column);
			initializeEditor(event, cellInfo, o);

			if (component != null) {
				Color bg = table.getCellStyle(row, column).getBackground();
				if (bg instanceof SystemColor) {
					bg = new Color(bg.getRGB());
				}
				component.setVisible(false);
				CellArea area = car.find(row, column);
				if (component.getParent() != table) {
					if (component instanceof JComponent) {
						((JComponent)component).setBorder(null);
					}
					// Add to beginning of area, to make sure it will
					// supercede any components that happen to be in cells
					area.add(component, 0);
				}
				if (setValues(area, row, column, cell_size)) {
					if (component != null) {
						component.setVisible(true);
						setFocusOnEditor(component);
					}
				}
			}
		}

		if (ev != null) {
			ev.type = JCEditCellEvent.EDIT_CELL;
			ev.editComponent = component;
			table.fireJCEditCellEvent(ev);
		}
	}

	
	return(true);
}

/**
 * Traverses to the top-left cell.
 */
public void traverseInitial()
{
	
	int row    = 0;
	int column = 0;

	if (table.isTraversable(row, column)) {
		traverseInitial(row, column, false, JCTableEnum.TRAVERSE_TO_CELL);
	}
	else {
		JCCellPosition cp =
		    table.keyHandler.findTraversableCell(JCTableEnum.TRAVERSE_RIGHT,
										         row, column);
		if (cp == null) {
			return;
		}
		traverseInitial(cp.row, cp.column, false, JCTableEnum.TRAVERSE_TO_CELL);
	}
}

/**
 * Returns false if the cell is not traversable. This method posts
 * JCTraverseCell events to registered listeners.
 *
 * @param row The row to which to traverse
 * @param column The column to which to traverse
 * @param select Should the cell be selected
 * @param traverse_type The type of traversal (right, left, down, up, home, ...)
 *
 * @return Whether intial part of traverse succeeded
 */
protected boolean traverseInitial(int row, int column, boolean select,
				                  int traverse_type)
{
	

	// Check if the cell is traversable
	if (row < 0                   ||
	    row >= table.getNumRows() ||
	    column < 0                ||
	    column >= table.getNumColumns())
	{
		return(false);
	}

	if (!table.getCellStyle(row, column).isTraversable()) {
		return(false);
	}

	CellAreaRenderer car = table.getCellAreaHandler().getCellAreaRenderer();

	int old_current_row    = table.current_row;
	int old_current_column = table.current_column;

	table.current_row    = row;
	table.current_column = column;

	// Remove focus from prior cell
	if (isValidCell(old_current_row, old_current_column) &&
	    (old_current_row != row || old_current_column != column))
	{
        // If it is a component, make it let go of the focus
        Component comp = table.getComponent(old_current_row, old_current_column);

		// Send focus back to table so that we can traverse to the
		// appropriate cell
        if (comp != null) {
            table.getFocusManager().requestFocus();
        }

		car.repaintCell(old_current_row, old_current_column, true);
	}

	if (select) {
		table.setSelection(row, column, row, column);
	}

	// Scroll the cell onto the screen
	if (!table.makeVisible(row, column)) {
		return(false);
	}

	JCTraverseCellEvent ev = null;
	if (table.traverseListeners != null) {
		ev = new JCTraverseCellEvent(table, JCTraverseCellEvent.TRAVERSE_CELL,
									 old_current_row, old_current_column,
									 row, column, traverse_type);
		table.fireJCTraverseCellEvent(ev);
		if (ev.isCancelled()) {
			ev.next_row    = old_current_row;
			ev.next_column = old_current_column;
		}

		row    = ev.next_row;
		column = ev.next_column;

		table.current_row    = row;
		table.current_column = column;
	}

	// Paint the new cell
	car.repaintCell(row, column);

	if (ev != null) {
		ev.type = JCTraverseCellEvent.AFTER_TRAVERSE_CELL;
		table.fireJCTraverseCellEvent(ev);
	}

    // Don't move if the move has been cancelled by the listener
    if (ev != null && ev.isCancelled()) {
        return(false);
    }
    else {
	    return(true);
    }
}

/**
 * Returns true the given row and column is a valid cell.
 *
 * @param row The row of the cell
 * @param column The column of the cell
 *
 * @return Whether cell is valid or not
 */
protected boolean isValidCell(int row, int column)
{
	return(row >= JCTableEnum.LABEL && row < table.getNumRows() &&
	       column >= JCTableEnum.LABEL && column < table.getNumColumns());
}

/**
 * Set attributes, first determining cell's editability.
 *
 * @param row The row of the cell
 * @param column The column of the cell
 * @param size The size to which the component is set
 *
 * @return Whether setValues succeeded
 */
protected boolean setValues(int row, int column, Rectangle size)
{
	CellArea area = table.getCellAreaHandler().find(row, column);
	return(setValues(area, row, column, size));
}

/**
 * Sets the bounds of the editor for the given cell.
 *
 * @param area The CellArea of the cell to be resized
 * @param row The row of the cell
 * @param column The column of the cell
 * @param size The size to which the component is set
 *
 * @return Whether setValues succeeded
 */
protected boolean setValues(CellArea area, int row, int column, Rectangle size)
{
	
	if (area == null || editor == null) {
		return(false);
	}
	if (component == null) {
		return(true);
	}

	boolean editable = true; // table.getEditable(r,c) & table.dataView.getEditable()
	if (!editable) {
		return(false);
	}

	int edit_height_policy = table.getEditHeightPolicy();
	int edit_width_policy  = table.getEditWidthPolicy();

	int w = Math.max(size.width, 10);
	int h = Math.max(size.height, 10);

	int indent = (table.getCellBorderWidth() > 0) ? 1 : 0;
	// Need to adjust for potentially missing borders on bottom and right
	// (THIN border type)
	int thinSides      = table.cellAreaHandler.determineThinSides(row, column);
	boolean rightSide  = ((thinSides & JCCellBorder.BORDERSIDE_THIN_RIGHT) > 0);
	boolean bottomSide = ((thinSides & JCCellBorder.BORDERSIDE_THIN_BOTTOM) > 0);

	// Add 2 pixels to preferred/minimum width because that's what
	// gets subtracted at the bottom of this routine.  The subtraction is
	// all fine and dandy when restricting things to cell size, but it
	// makes the preferred/minimum size editors too small in the case of
	// certain fonts
	switch(edit_width_policy) {
	case JCTableEnum.EDIT_ENSURE_MINIMUM_SIZE:
		if (component.getMinimumSize().width > w) {
			w = component.getMinimumSize().width + (rightSide ? 2*indent
															  : 1*indent);
		}
		break;
	case JCTableEnum.EDIT_ENSURE_PREFERRED_SIZE:
		if (component.getPreferredSize().width > w) {
			w = component.getPreferredSize().width + (rightSide ? 2*indent
																: 1*indent);
		}
		break;
	}

	switch(edit_height_policy) {
	case JCTableEnum.EDIT_ENSURE_MINIMUM_SIZE:
		if (component.getMinimumSize().height > h) {
			h = component.getMinimumSize().height;
		}
		break;
	case JCTableEnum.EDIT_ENSURE_PREFERRED_SIZE:
		if (component.getPreferredSize().height > h) {
			h = component.getPreferredSize().height;
		}
		break;
	}

	// Adjust the coordinates by the type of area in order to determine the cell
	// that contains the given point
	if (area.isHorizontalScrollable()) {
		size.x += table.cellAreaHandler.cellOrigin.x;
	}
	if (area.isVerticalScrollable()) {
		size.y += table.cellAreaHandler.cellOrigin.y;
	}

	// Ensure the editor width does not exceed the width of the cell area,
	// otherwise the cursor can move past the end of the cell and the user
	// cannot see it.
	int compX = size.x - area.getX();
	if (w + compX > area.getWidth()) {
		w = area.getWidth() - compX;
	}

    component.setBounds(compX + indent, size.y + indent - area.getY(),
                        w - (rightSide ? 2*indent : 1*indent),
                        h - (bottomSide ? 2*indent : 1*indent));
	return(true);
}

/**
 * Check to see if editor component needs reparenting.  Reparenting is
 * needed if the CellArea objects have changed boundaries around the
 * cell the editor is in.
 */
public void checkAndReparentEditor()
{
	
    if (component != null) {
        CellArea area = table.getCellAreaHandler().find(edit_row, edit_column);
        if (area != component.getParent()) {
            area.add(component, 0);
        }
    }
}

/**
 * Sets the width of the edit component.
 *
 * @param value The new width
 */
public void setWidth(int value)
{
	
	Rectangle cell_size = table.getCellAreaHandler().getBounds(edit_row,
														       edit_column);
	cell_size.width = value;
	setValues(edit_row, edit_column, cell_size);
}

/**
 * Sets the height of the edit component.
 *
 * @param value The new height
 */
public void setHeight(int value)
{
	
	Rectangle cell_size = table.getCellAreaHandler().getBounds(edit_row,
														       edit_column);
	cell_size.height = value;
	setValues(edit_row, edit_column, cell_size);
}

/**
 * Repositions the editor.
 */
public void repositionEditor()
{
	
	if (editor == null || component == null) {
		return;
	}
	Rectangle cell_size = table.getCellAreaHandler().getBounds(edit_row,
														       edit_column);
	setValues(edit_row, edit_column, cell_size);
}

/**
 * Move the component.
 *
 * @param direction The direction of movement, horizontal or vertical
 * @param value The amount in pixels of movement
 */
public void move(int direction, int value)
{
	
	if (component == null) {
		return;
	}

	int x, y;

	// Need to find cell area for component to find out
	// if it is inside a frozen row or column
	CellArea area = table.getCellAreaHandler().find(edit_row, edit_column);
	if (area != null) {
		if (direction == HORIZONTAL) {
			// Handle horizontal scrolling special cases

			// Special case one: horizontal scrolling when editor
			// is in frozen column region.
			if (area.getType() == CellArea.FROZENCOLUMN) {
				return;
			}
			// Special case two: horizontal scrolling when
			// editor is in frozen cells region and the editor
			// is in the part of the frozen cells region that
			// is part of the frozen columns.  (Frozen cell
			// region occurs when there are frozen rows and columns,
			// and occurs in the intersection of the frozen row
			// and columns).
			if (area.getType() == CellArea.FROZENCELL &&
				edit_column < table.getFrozenColumns())
			{
				return;
			}
		}
		else if (direction == VERTICAL) {
			// Handle vertical scrolling special cases

			// Special case one: vertical scrolling when editor
			// is in frozen row region.
			if (area.getType() == CellArea.FROZENROW) {
				return;
			}
			// Special case two: vertical scrolling when
			// editor is in frozen cells region and the editor
			// is in the part of the frozen cells region that
			// is part of the frozen rows.  (Frozen cell
			// region occurs when there are frozen rows and columns,
			// and occurs in the intersection of the frozen row
			// and columns).
			if (area.getType() == CellArea.FROZENCELL &&
				edit_row < table.getFrozenRows())
			{
				return;
			}
		}
	}

	switch (direction) {
	case VERTICAL:
		x = component.getLocation().x;
		y = component.getLocation().y - value;
		component.setLocation(x, y);
		break;
	case HORIZONTAL:
		x = component.getLocation().x - value;
		y = component.getLocation().y;
		component.setLocation(x, y);
		break;
	}
}

/**
 * Executes a beep to signal an error.
 */
protected void signalError()
{
	if (component == null || !component.isVisible()) {
		Toolkit.getDefaultToolkit().beep();
		return;
	}
	Toolkit.getDefaultToolkit().beep();
}

/**
 * Cancels the current edit on the given table, if an edit is in progress
 *
 * @return Whether cancel succeeded
 */
protected boolean cancel()
{
	return(cancel(true));
}

/**
 * Cancels the current edit on the given table, if an edit is in progress
 * @param hide determines whether the editor is hidden after the commit.
 *
 * @return Whether cancel succeeded
 */
public boolean cancel(boolean hide)
{
	
	if (editor == null) {
		return(true);
	}

	if (table.editCellListeners != null) {
		JCEditCellEvent ev = new JCEditCellEvent(table,
							                     JCEditCellEvent.AFTER_EDIT_CELL,
								                 edit_row, edit_column);
		table.fireJCEditCellEvent(ev);
	}

	editor.cancelCellEditing();
	editor.removeCellEditorListener(this);
	if (hide) {
		removeEditor();
	}
	return(true);
}

/**
 * Commits the current edit on the given table, if an edit is in progress
 *
 * @return Whether commit succeeded
 */
protected boolean commit()
{
	return(commit(true));
}

/**
 * Commits the current edit on the table.
 * @param hide Determines whether the editor is hidden after the commit.
 *
 * @return Whether commit succeeded
 */
public boolean commit(boolean hide)
{
	return(commit(hide, true));
}

/**
 * Commits the current edit on the table.
 * @param hide Determines whether the editor is hidden after the commit.
 * @param stop Determines whether the editor should be told to stop editing.
 *     Set to false if the editor is originator for the commit to avoid round-trip.
 *
 * @return Whether commit succeeded
 */
protected boolean commit(boolean hide, boolean stop)
{
	
	if (editor == null) {
		return(true);
	}

	// First, stop editing.  Commit has been called
	if (stop) {
		boolean mod = editor.isModified();
		if (!editor.stopCellEditing()) {
			// Editor has refused the change, likely because the user-entered
			// data is invalid
			signalError();
			return(false);
		}
		// If nothing has been changed, go no further
		if (!mod) {
			if (hide) {
				removeEditor();
				if (table.editCellListeners != null) {
					JCEditCellEvent ev =
					    new JCEditCellEvent(table,
						                    JCEditCellEvent.AFTER_EDIT_CELL,
											edit_row, edit_column);
					table.fireJCEditCellEvent(ev);
				}
			}
			return(true);
		}
	}

	// Get the new cell value from the editor

    // PR11947 null should be allowed.  It's the validator's job to say
    // yes or no.
    Object o = editor.getCellEditorValue();
    // Set data directly in data source
    if (table.getDataView().setTableDataItem(o, edit_row, edit_column)) {
        // We're done editing in all cases.  Remove the editor
        if (hide) {
            removeEditor();
            if (table.editCellListeners != null) {
                JCEditCellEvent ev =
                    new JCEditCellEvent(table, JCEditCellEvent.AFTER_EDIT_CELL,
                                        edit_row, edit_column);
                table.fireJCEditCellEvent(ev);
            }
        }
        return(true);
    }

	// If the commit fails, we end up here.
	signalError();
	setFocusOnEditor(component);
	return false;
}

/**
 * Commits the edit and attempts a traversal in the given direction. If the
 * commit fails, the traversal is not performed.
 *
 * @param direction The direction which to traverse
 */
protected void commitAndTraverse(int direction)
{
	
	if (commit()) {
		table.getFocusManager().requestFocus();
		table.keyHandler.traverse(direction);
	}
}

/**
 * Removes the edit component from table.
 */
protected void removeEditor()
{
	
	if (component != null) {
		component.setVisible(false);
	}
	setCellEditor(null);
	table.getFocusManager().requestFocus();
}

/**
 * The <CODE>JCCellEditor</CODE> posts EditingStopped when it is done editing.
 * The container then will decide whether to call commit() in the editor.
 *
 * @param event The event which caused us to want editing to stop
 */
public void editingStopped(JCCellEditorEvent event)
{
	
	// Consume key events
	if (event.getEvent() != null &&
		event.getEvent() instanceof java.awt.event.KeyEvent)
	{
		((java.awt.event.KeyEvent)event.getEvent()).consume();
	}

	if (editor != null                     &&
	    table.getCurrentRow() == edit_row  &&
		table.getCurrentColumn() == edit_column)
	{
		// Non-traversal case
		// Note that we now force stopEdit in cell editors.  This
		// is a change from 4.0.0.  It is important that the cell
		// editor stopCellEditing() be called in both the pointer
		// and the mouse hit case.
		if (commit()) {
			table.getFocusManager().requestFocus();
		}
	}
}

/**
 * The <CODE>JCCellEditor</CODE> calls <CODE>cancelEdit()</CODE> to  cancel and undo an edit;
 * for example, when the user presses <STRONG>Esc</STRONG>.
 *
 * @param event The event which caused us to want editing to be canceled
 */
public void editingCanceled(JCCellEditorEvent event)
{
	
	if (editor != null                     &&
	    table.getCurrentRow() == edit_row  &&
		table.getCurrentColumn() == edit_column)
	{
		cancel(true);
		table.getFocusManager().requestFocus();
	}
}

/**
 * Return the component if it is focusable, otherwise the first kid which
 * is focusable.  If no focusable kids, return the component (we need to use
 * something).  Note that this method does not recurse into the children's
 * children.
 *
 * @param c The component whose focus or children's focus is being investigated
 *
 * @return The component or child component which is focusable.  Note that if
 *     no children are focusable, the original component is returned whether
 *     it is focusable or not
 */
public Component getFocusableChild(Component c)
{
	if (c != null && !table.getFocusManager().isFocusTraversable(c)) {
		Component kids[] = ((Container)c).getComponents();
		for (int p = 0; p < kids.length; p++) {
			if (table.getFocusManager().isFocusTraversable(kids[p])) {
				c = kids[p];
				break;
			}
		}
	}
	return(c);
}

/**
 * Set focus on the component if it is focusable or to the first focusable child.
 */
public void setFocusOnEditor()
{
	if (component != null) {
		setFocusOnEditor(component);
	}
}

/**
 * Set focus on the component if it is focusable or to the first focusable
 * child.
 *
 * @param editComponent The component or child of component on which to set
 *     the focus.
 */
protected void setFocusOnEditor(Component editComponent)
{
	if (editComponent == null) {
		return ;
	}
	

	// Set focus on the component if it is focusable
	FocusModel fm = table.getFocusManager();
	if (fm.isFocusTraversable(editComponent)) {
		fm.requestFocus(editComponent);
	}
	// Set focus on the first focusable child
	else {
	    Component focusComponent = getFocusableChild(editComponent);
		if (focusComponent != editComponent) {
		    fm.requestFocus(focusComponent);
		}
	}
}

/**
 * Reset the old focus owner for this focus period
 *
 * @param event The focus event which caused focus to be gained
 */
public abstract void focusGained(FocusEvent event);

/**
 * Sets the table as a focused component when the editor loses permanent focus.
 *
 * @param event The focus event which caused focus to be lost
 */
public abstract void focusLost(FocusEvent event);

}
