/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: InputHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.table.data.JCVectorDataSource;

import java.awt.*;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.io.Serializable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Collection;

/**
 * <CODE>InputHandler</CODE> provides copy/paste support for
 * <CODE>KeyInputHandler</CODE> and <CODE>MouseInputHandler</CODE>.
 */
public class InputHandler implements Serializable {

protected JCTable table;

// Used by KeyInputHandler and MouseInputHandler to flag whether
// selection is in progress
protected boolean SELECTING = false;

public InputHandler(JCTable table)
{
    this.table = table;
}

/**
 * Sets the given JCCellPosition with the cell found at the
 * x,y coordinates. JCTableEnum.NOVALUE will be set for
 * the row/column position if the coordinates don't
 * correspond to a cell location.
 */
public void getCellPosition(int x, int y, JCCellPosition cp)
{
    if (cp == null) {
        return;
    }

    CellAreaHandler ah = table.getCellAreaHandler();
    cp.row = ah.getRow(x, y);
    cp.column = ah.getColumn(x, y);
}

/**
 * Returns true if the modifier is a shift key.
 */
public boolean isShift(int modifiers)
{
    return((modifiers & Event.SHIFT_MASK) != 0);
}

/**
 * Returns true if the modifier contains a control key.
 */
public boolean isControl(int modifiers)
{
    return((modifiers & Event.CTRL_MASK) != 0);
}

/**
 * Returns true if the modifier contains an ALT key.
 */
public boolean isAlt(int modifiers)
{
    return((modifiers & Event.ALT_MASK) != 0);
}

/**
 * Returns true if the modifier contains a META key.
 */
public boolean isMeta(int modifiers)
{
    return((modifiers & Event.META_MASK) != 0);
}

/**
 * Returns true if the row/column is a cell.
 */
public boolean isCell(int row, int column)
{
    return(row > -1 && column > -1);
}

/**
 * Returns true if the row/column is a label.
 */
public boolean isLabel(int row, int column)
{
    return(row == JCTableEnum.LABEL || column == JCTableEnum.LABEL);
}

/**
 * Returns true if the given event is mapped to a copy action.
 */
public boolean isCopyAction(AWTEvent event, int row, int column)
{
    return(table.getAction(event, row, column) == JCTableEnum.COPY_ACTION);
}

/**
 * Returns true if the given event is mapped to a paste action.
 */
public boolean isPasteAction(AWTEvent event, int row, int column)
{
    return(table.getAction(event, row, column) == JCTableEnum.PASTE_ACTION);
}

/**
 * Executes a copy action using the system clipboard.
 */
public void performCopyAction()
{
    // Custom copy requires a local clipboard. This will not be
    // supported in the 4.0 release.
    JCCellRange cr;
    if (table.getSelectedCells() != null &&
        table.getSelectedCells().size() > 0)
    {
        JCCellRange temp_cr =
            (JCCellRange)(table.getSelectedCells().iterator().next());
        // Flip range to top-left, bottom-right order if selection was made
        // out of order
        cr = new JCCellRange(Math.min(temp_cr.start_row, temp_cr.end_row),
                             Math.min(temp_cr.start_column, temp_cr.end_column),
                             Math.max(temp_cr.start_row, temp_cr.end_row),
                             Math.max(temp_cr.start_column,
                                      temp_cr.end_column));
    }
    else {
        cr = new JCCellRange(table.current_row, table.current_column,
                             table.current_row, table.current_column);
    }
    StringCellRangeFlavor selection = new StringCellRangeFlavor(table, cr);
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(selection,
                                                                 selection);
}

/**
 * Executes a paste action using the system clipboard.
 */
public void performPasteAction()
{
    // Custom paste requires a local clipboard. This will not be
    // supported in the 4.0 release.
    Transferable tfr =
        Toolkit.getDefaultToolkit().getSystemClipboard().getContents(this);
    if (tfr != null) {
        try {
            DataFlavor flavors[] = tfr.getTransferDataFlavors();
            for (int i = 0; i < flavors.length; i++) {
                String name = flavors[i].getHumanPresentableName();
                if (name.equals("CellRange")) {
                    Vector cells = (Vector)(tfr.getTransferData(flavors[i]));
                    loadPasteData(cells);
                    return;
                }
                else if (name.equals("Unicode String")) {
                    String str = (String)tfr.getTransferData(flavors[i]);
                    StringTokenizer st = new StringTokenizer(str, "\n", false);
                    Vector cells = new Vector();
                    Vector row;
                    while (st.hasMoreElements()) {
                        String line = (String)st.nextElement();
                        if (line != null) {
                            StringTokenizer tabs = new StringTokenizer(line,
                                                                       "\t",
                                                                       false);
                            row = new Vector();
                            while (tabs.hasMoreElements()) {
                                row.addElement(tabs.nextToken());
                            }
                            cells.addElement(row);
                        }
                    }
                    loadPasteData(cells);
                    return;
                }
            }
        }
        catch (IOException ioe) {
            System.out.println("Error pasting data: " + ioe.getMessage());
        }
        catch (UnsupportedFlavorException ufe) {
            System.out.println("Error pasting data: " + ufe.getMessage());
        }
    }
}

/**
 * Pastes the given vector of cells into the current location. If
 * no cells are selected, the cells are pasted starting at the current
 * cell position for as many cells as required. If cells are selected,
 * the pasted values will not extend outside of the selected region. If
 * the vector of cells contains fewer cells than the current selection,
 * not all cells will receive a new value.
 *<P>All cells paste independently. As a result, individual values
 * may be rejected by the datasource.
 */
protected void loadPasteData(Vector cells)
{
    JCCellRange cr;
    Collection selectedCells = table.getSelectedCells();
    if (selectedCells != null && selectedCells.size() > 0) {
        JCCellRange temp_cr = (JCCellRange)(selectedCells.iterator().next());
        // Flip range to top-left, bottom-right order if selection was made
        // out of order
        cr = new JCCellRange(Math.min(temp_cr.start_row, temp_cr.end_row),
                             Math.min(temp_cr.start_column, temp_cr.end_column),
                             Math.max(temp_cr.start_row, temp_cr.end_row),
                             Math.max(temp_cr.start_column,
                                      temp_cr.end_column));
    }
    else {
        cr = new JCCellRange(table.current_row, table.current_column,
                             table.getNumRows() - 1, table.getNumColumns() - 1);
    }
    if (cr.end_row == JCTableEnum.MAXINT) {
        cr.end_row = table.getNumRows() - 1;
    }
    if (cr.end_column == JCTableEnum.MAXINT) {
        cr.end_column = table.getNumColumns() - 1;
    }
    if (cr != null) {
        int rows = cells.size();
        if (rows == 0) {
            return;
        }
        int columns = ((Vector)cells.elementAt(0)).size();
        for (int r = cr.start_row, rr = 0;
             r <= cr.end_row && rr < rows;
             r++, rr++)
        {
            Vector row = (Vector)cells.elementAt(rr);
            for (int c = cr.start_column, cc = 0;
                 c <= cr.end_column && cc < columns;
                 c++, cc++)
            {
                DataViewModel dataView = table.getDataView();
                if (c == JCTableEnum.LABEL || r == JCTableEnum.LABEL) {
                    TableDataModel ds = dataView.getDataSource();
                    if (ds instanceof JCVectorDataSource) {
                        JCVectorDataSource vds = (JCVectorDataSource)ds;
                        if (c == JCTableEnum.LABEL) {
                            vds.setRowLabel(r, row.elementAt(cc));
                        }
                        else if (r == JCTableEnum.LABEL) {
                            vds.setColumnLabel(c, row.elementAt(cc));
                        }
                    }
                }
                else {
                    dataView.setTableDataItem(row.elementAt(cc), r, c);
                }
            }
        }
    }
}

}
