/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCCellBorder.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.io.Serializable;

/**
 * The <CODE>JCCellBorder</CODE> class implements the <CODE>CellBorderModel</CODE>
 * interface, and provides border drawing for the standard border types.
<pre>
	BORDER_ETCHED_IN    double line; cell appears inset
	BORDER_ETCHED_OUT   double line; cell appears raised
	BORDER_FRAME_IN     1-pixel shadow-in at edge; cell appears framed
	BORDER_FRAME_OUT    1-pixel shadow-out at edge; cell appears framed
	BORDER_IN           cell appears inset
	BORDER_OUT          cell appears raised
	BORDER_PLAIN        shadow drawn in foreground color
	BORDER_THIN         thin border drawn in highlight color
	BORDER_NONE         no border
</pre>
 */
public class JCCellBorder implements CellBorderModel, Serializable {

/** A <code>JCCellBorder</code> <code>BorderSide</code> property constant indicating
 * that a thin border should draw an explicit right hand side.  Used internally.
 */
public static final int BORDERSIDE_THIN_RIGHT    = 16;
/** A <code>JCCellBorder</code> <code>BorderSide</code> property constant indicating
 * that a thin border should draw an explicit bottom hand side.  Used internally.
 */
public static final int BORDERSIDE_THIN_BOTTOM  = 32;

protected int borderType;

/**
 * Creates a new <code>JCCellBorder</code> object and initializes the type to
 * <CODE>JCTableEnum.BORDER_NONE</CODE>
 */
public JCCellBorder()
{
    this.borderType = JCTableEnum.BORDER_NONE;
}

/**
 * Creates a new <code>JCCellBorder</code> object.
 * @param borderType one of the following <code>JCTableEnum</code> members:
 <pre>
 BORDER_ETCHED_IN    double line; cell appears inset
 BORDER_ETCHED_OUT   double line; cell appears raised
 BORDER_FRAME_IN     1-pixel shadow-in at edge; cell appears framed
 BORDER_FRAME_OUT    1-pixel shadow-out at edge; cell appears framed
 BORDER_IN           cell appears inset
 BORDER_OUT          cell appears raised
 BORDER_PLAIN        shadow drawn in foreground color
 BORDER_THIN         thin border drawn in highlight color
 BORDER_NONE         no border
 </pre>
 */
public JCCellBorder(int borderType)
{
    this.borderType = borderType;
}

/**
 * Sets the border type.
 * @param v
 */
public void setBorderType(int v)
{
    borderType = v;
}

/**
 * Returns the border type.
 * @return border type
 */
public int getBorderType()
{
    return(borderType);
}

/**
 * Invokes the border drawing mechanism appropriate to the type of border.
 * @param bottom_color
 * @param gc
 * @param height
 * @param plain_color
 * @param sides
 * @param size
 * @param top_color
 * @param width
 * @param x
 * @param y
 */
public void drawBackground(Graphics gc, int size, int sides,
                           int x, int y, int width, int height, Color top_color,
                           Color bottom_color, Color plain_color)
{
    switch (borderType) {
    case JCTableEnum.BORDER_PLAIN:
        drawShadow(gc, size, sides, x, y, width, height, bottom_color,
                   bottom_color);
        break;
    case JCTableEnum.BORDER_THIN:
        //thin border -- a bit complicated
        //first draw top and left
        drawShadow(gc, size,
                   sides & (JCTableEnum.BORDERSIDE_LEFT |
                            JCTableEnum.BORDERSIDE_TOP),
                   x, y, width, height, plain_color, plain_color);
        // then, if told to, draw an explicit right side
        // otherwise, draw the normal right side, which involves overlapping
        //   the neighbouring cell if the size if > 1
        if ((sides & BORDERSIDE_THIN_RIGHT) > 0) {
            drawShadow(gc, size,
                       sides & JCTableEnum.BORDERSIDE_RIGHT,
                       x, y, width, height, plain_color, plain_color);
        }
        else {
            if (size - 1 > 0) {
                drawShadow(gc, size - 1,
                           sides & JCTableEnum.BORDERSIDE_RIGHT,
                           x, y, width, height, plain_color, plain_color);
            }
        }
        // then, if told to, draw an explicit bottom side
        // otherwise, draw the normal bottom side, which involves overlapping
        //   the neighbouring cell if the size if > 1
        if ((sides & BORDERSIDE_THIN_BOTTOM) > 0) {
            drawShadow(gc, size,
                       sides & JCTableEnum.BORDERSIDE_BOTTOM,
                       x, y, width, height, plain_color, plain_color);
        }
        else {
            if (size - 1 > 0) {
                drawShadow(gc, size - 1,
                           sides & JCTableEnum.BORDERSIDE_BOTTOM,
                           x, y, width, height, plain_color, plain_color);
            }
        }
        break;
    case JCTableEnum.BORDER_IN:
        drawShadow(gc, size, sides, x, y, width, height, bottom_color,
                   top_color);
        break;
    case JCTableEnum.BORDER_OUT:
        drawShadow(gc, size, sides, x, y, width, height, top_color,
                   bottom_color);
        break;
    case JCTableEnum.BORDER_ETCHED_IN:
        drawEtchedShadow(gc, size, sides, x, y, width, height, bottom_color,
                         top_color);
        break;
    case JCTableEnum.BORDER_ETCHED_OUT:
        drawEtchedShadow(gc, size, sides, x, y, width, height, top_color,
                         bottom_color);
        break;
    case JCTableEnum.BORDER_FRAME_OUT:
        // 1-pixel shadow
        drawShadow(gc, size, sides, x, y, width, height, top_color,
                   bottom_color);
        // Plain
        drawShadow(gc, size - 1, sides, x, y, width, height, plain_color,
                   plain_color);
        break;
    case JCTableEnum.BORDER_FRAME_IN:
        // 1-pixel shadow
        drawShadow(gc, size, sides, x, y, width, height, bottom_color,
                   top_color);
        // Plain
        drawShadow(gc, size - 1, sides, x, y, width, height, plain_color,
                   plain_color);
        break;
    default:
        break;
    }
}

/**
 * Draws the shadow border for FRAME border types.
 * @param bottom_color
 * @param gc
 * @param height
 * @param sides
 * @param size
 * @param top_color
 * @param width
 * @param x
 * @param y
 */
private void drawShadow(Graphics gc, int size, int sides,
                        int x, int y, int width, int height, Color top_color,
                        Color bottom_color)
{
    if (size <= 0 || width <= 0 || height <= 0) {
        return;
    }
    if (size > width / 2) {
        size = width / 2;
    }
    if (size > height / 2) {
        size = height / 2;
    }
    if (size <= 0) {
        return;
    }

    int i, offset;

    // top side
    gc.setColor(top_color);
    if ((sides & JCTableEnum.BORDERSIDE_TOP) != 0) {
        for (i = 0; i < size; i++) {
            offset = (sides & JCTableEnum.BORDERSIDE_RIGHT) != 0 ? i : 0;
            gc.drawLine(x, y + i, x + width - offset - 1, y + i);
        }
    }
    // left side
    if ((sides & JCTableEnum.BORDERSIDE_LEFT) != 0) {
        for (i = 0; i < size; i++) {
            offset = (sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0 ? i : 0;
            gc.drawLine(x + i, y, x + i, y + height - offset - 1);
        }
    }
    gc.setColor(bottom_color);
    // bottom side
    if ((sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0) {
        for (i = 0; i < size; i++) {
            offset = (sides & JCTableEnum.BORDERSIDE_LEFT) != 0 ? i : 0;
            gc.drawLine(x + offset, y + height - i - 1, x + width - offset - 1,
                        y + height - i - 1);
        }
    }
    // right side
    if ((sides & JCTableEnum.BORDERSIDE_RIGHT) != 0) {
        for (i = 0; i < size; i++) {
            offset = (sides & JCTableEnum.BORDERSIDE_TOP) != 0 ? i : 0;
            gc.drawLine(x + width - i - 1, y + offset, x + width - i - 1,
                        y + height - 1);
        }
    }
}

/**
 * Draws an <i>n</i> segment-wide etched shadow, using the provided rectangle.
 * @param bottom_color
 * @param gc
 * @param height
 * @param sides
 * @param size
 * @param top_color
 * @param width
 * @param x
 * @param y
 */
private void drawEtchedShadow(Graphics gc, int size, int sides,
                              int x, int y, int width, int height,
                              Color top_color, Color bottom_color)
{
    int k, half_size, size2, size3;
    int pos_top, pos_left, pos_bottom, pos_right;

    if (size <= 0 || width <= 0 || height <= 0)
        return;
    if (size == 1) {
        drawShadow(gc, size, sides, x, y, width, height, top_color,
                   bottom_color);
        return;
    }

    if (size > width / 2)
        size = width / 2;
    if (size > height / 2)
        size = height / 2;
    if (size <= 0)
        return;

    size = (size % 2) != 0 ? (size - 1) : (size);

    Rectangle rects[] = new Rectangle[size * 4];
    for (int i = 0; i < size * 4; i++) {
        rects[i] = new Rectangle();
    }

    half_size = size / 2;
    size2 = size + size;
    size3 = size2 + size;

    pos_top = 0;
    pos_left = half_size;
    pos_bottom = size2;
    pos_right = size2 + half_size;

    get_rects(half_size, 0, x, y, width, height,
              pos_top, pos_left, pos_bottom, pos_right, sides, rects);

    pos_top = size3;
    pos_left = size3 + half_size;
    pos_bottom = size;
    pos_right = size + half_size;

    get_rects(half_size, half_size, x, y, width, height,
              pos_top, pos_left, pos_bottom, pos_right, sides, rects);

    k = half_size;
    if ((sides & JCTableEnum.BORDERSIDE_TOP) != 0) {
        fillRectangles(gc, top_color, rects, 0, k);
        fillRectangles(gc, bottom_color, rects, size3, k);
    }
    if ((sides & JCTableEnum.BORDERSIDE_LEFT) != 0) {
        fillRectangles(gc, top_color, rects, k, k);
        fillRectangles(gc, bottom_color, rects, size3 + k, k);
    }
    if ((sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0) {
        fillRectangles(gc, top_color, rects, size, k);
        fillRectangles(gc, bottom_color, rects, size2, k);
    }
    if ((sides & JCTableEnum.BORDERSIDE_RIGHT) != 0) {
        fillRectangles(gc, top_color, rects, size + k, k);
        fillRectangles(gc, bottom_color, rects, size2 + k, k);
    }
}

/**
 * Calculates rectangles used by <CODE>drawEtchedShadow</CODE>.
 * @param max_i
 * @param height
 * @param offset
 * @param pos_bottom
 * @param pos_left
 * @param pos_right
 * @param pos_top
 * @param rects
 * @param sides
 * @param width
 * @param x
 * @param y
 */
private void get_rects(int max_i, int offset, int x, int y, int width,
                       int height, int pos_top, int pos_left, int pos_bottom,
                       int pos_right, int sides, Rectangle rects[])
{
    int i, k, offsetX2;

    if (rects == null)
        return;

    for (i = 0; i < max_i; i++, offset++) {
        offsetX2 = offset + offset;

        //  Top segments.
        if ((sides & JCTableEnum.BORDERSIDE_LEFT) != 0) {
            rects[pos_top + i].x = x + offset;
            k = (sides & JCTableEnum.BORDERSIDE_RIGHT) != 0 ? offsetX2 : 0;
        }
        else {
            rects[pos_top + i].x = x;
            k = (sides & JCTableEnum.BORDERSIDE_RIGHT) != 0 ? offset : 0;
        }
        rects[pos_top + i].y = y + offset;
        rects[pos_top + i].width = width - k;
        rects[pos_top + i].height = 1;

        //  Left segments.
        if ((sides & JCTableEnum.BORDERSIDE_TOP) != 0) {
            rects[pos_left + i].y = y + offset;
            k = (sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0 ? offsetX2 : 0;
        }
        else {
            rects[pos_left + i].y = y;
            k = (sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0 ? offset : 0;
        }
        rects[pos_left + i].x = x + offset;
        rects[pos_left + i].width = 1;
        rects[pos_left + i].height = height - k;

        //  Bottom segments.
        if ((sides & JCTableEnum.BORDERSIDE_LEFT) != 0) {
            rects[pos_bottom + i].x = x + offset;
            k = (sides & JCTableEnum.BORDERSIDE_RIGHT) != 0 ? offsetX2 : 0;
        }
        else {
            rects[pos_bottom + i].x = x;
            k = (sides & JCTableEnum.BORDERSIDE_RIGHT) != 0 ? offset : 0;
        }
        rects[pos_bottom + i].y = y + height - offset - 1;
        rects[pos_bottom + i].width = width - k;
        rects[pos_bottom + i].height = 1;

        //  Right segments./
        if ((sides & JCTableEnum.BORDERSIDE_TOP) != 0) {
            rects[pos_right + i].y = y + offset;
            k = (sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0 ? offsetX2 : 0;
        }
        else {
            rects[pos_right + i].y = y;
            k = (sides & JCTableEnum.BORDERSIDE_BOTTOM) != 0 ? offset : 0;
        }
        rects[pos_right + i].x = x + width - offset - 1;
        rects[pos_right + i].width = 1;
        rects[pos_right + i].height = height - k;
    }
}

private void fillRectangles(Graphics gc, Color color, Rectangle rects[],
                            int start, int num)
{
    gc.setColor(color);
    for (int i = start; i < start + num; i++) {
        gc.fillRect(rects[i].x, rects[i].y, rects[i].width, rects[i].height);
    }
}

}








