/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCCellLayout.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;



import java.awt.Rectangle;
import java.io.Serializable;

import java.util.ListIterator;
import java.util.List;
import java.util.Vector;
import java.util.ConcurrentModificationException;

/** 
 * Maintain row/column heights and widths in a "chain" or array of runs.
 * An individual chain keeps track of row heights or column widths.
 */
public class JCCellLayout implements CellLayoutModel, Serializable
{
	
protected JCTable         table;
protected int             column_label_height;
protected int             row_label_width;
protected Chain           widths;
protected Chain           heights;
	
public JCCellLayout()
{
	resetRows();
	resetColumns();
}
	
/**
 * Returns the <code>JCTable</code>.
 * @return <code>JCTable</code>
 */
public JCTable getTable()
{
	return(table);
}
	
/**
 * Sets the <code>JCTable</code>.
 * @param table
 */
public void setTable(JCTable table)
{
	this.table = table;
}

/**
 * Returns the height of a row.
 * @return height of a row
 * @param row
 */
public int getHeight(int row)
{
	if (row == JCTableEnum.LABEL) {
		return(column_label_height);
	}
	else {
		return(heights.getValue(row));
	}
}

/**
 * Sets the height of a row.
 * @param row
 * @param value
 */
public void setHeight(int row, int value)
{
	if (row == JCTableEnum.LABEL) {
		column_label_height = value;
	}
	else {
		heights.setValue(row, row, value);
	}
}

/**
 * Sets the height of a range of rows.
 * @param end_row
 * @param start_row
 * @param value
 */
public void setHeight(int start_row, int end_row, int value)
{
	heights.setValue(start_row, end_row, value);
}

/**
 * Returns the width of a column.
 * @return width of a column
 * @param column
 */
public int getWidth(int column)
{
	if (column == JCTableEnum.LABEL) {
		return(row_label_width);
	}
	else {
		return(widths.getValue(column));
	}
}

/**
 * Sets the width of a column.
 * @param column
 * @param value
 */
public void setWidth(int column, int value)
{
	if (column == JCTableEnum.LABEL) {
		row_label_width = value;
	}
	else {
		widths.setValue(column, column, value);
	}
}

/**
 * Sets the width of a range of columns.
 * @param end_column
 * @param start_column
 * @param value
 */
public void setWidth(int start_column, int end_column, int value)
{
	widths.setValue(start_column, end_column, value);
}

/**
 * Determines the column that contains the given (<i>x</i>,<i>y</i>) coordinate.
 * @param x
 * @param y
 * @return
 */
public int getColumn(int x, int y)
{
	if (x < row_label_width) {
		return(JCTableEnum.LABEL);
	}
	return(widths.getIndex(x - row_label_width));
}

/**
 * Determines the row that contains the given (<i>x</i>,<i>y</i>) coordinate.
 * @param x
 * @param y
 * @return
 */
public int getRow(int x, int y)
{
	if (y < column_label_height) {
		return(JCTableEnum.LABEL);
	}
	return(heights.getIndex(y - column_label_height));
}

/**
 * Returns the start position of a column.
 * @return start position of a column
 * @param column
 */
public int getColumnPosition(int column)
{
	if (column == JCTableEnum.LABEL) {
		return(0);
	}
	else {
		return(row_label_width + widths.getSum(column));
	}
}

/**
 * Returns the start position of a row.
 * @param row
 * @return start position of a row
 */
public int getRowPosition(int row)
{
	if (row == JCTableEnum.LABEL) {
		return(0);
	}
	else {
		return(column_label_height + heights.getSum(row));
	}
}
	
/**
 * Returns the bounds of the given cell or label.
 * @return bounds of the given cell or label
 * @param row
 * @param column
 */
public Rectangle getBounds(int row, int column)
{
	return(getBounds(row, column, new Rectangle()));
}

/**
 * Gets the bounds of the given cell or label using the given Rectangle.
 * @return bounds of the given cell or label
 * @param column
 * @param r
 * @param row
 */
public Rectangle getBounds(int row, int column, Rectangle r)
{
	r.setBounds(getColumnPosition(column), getRowPosition(row),
				getWidth(column), getHeight(row));
	return(r);
}


/**
 * Returns the bounds of a cell range.
 * @param end_column
 * @param end_row
 * @param start_column
 * @param start_row
 * @return bounds of a cell range
 */
public Rectangle getBounds(int start_row, int start_column, int end_row,
                           int end_column)
{
	return(getBounds(start_row, start_column, end_row, end_column,
					 new Rectangle()));
}

/**
 * Gets the bounds of a cell range using the given Rectangle.
 * @return bounds of a cell range
 * @param end_column
 * @param end_row
 * @param r
 * @param start_column
 * @param start_row
 */
public Rectangle getBounds(int start_row, int start_column,
						   int end_row, int end_column, Rectangle r)
{
	int x = getColumnPosition(start_column);
	int y = getRowPosition(start_row);

	r.setBounds(x,
				y,
				getColumnPosition(end_column) + getWidth(end_column) - x,
				getRowPosition(end_row) + getHeight(end_row) - y);
	return(r);
}

/**
 * Moves a set of rows to a position above the destination. To move
 * rows to the end, use <CODE>JCTableEnum.MAXINT</CODE>.
 * @param start_row the first row
 * @param num_rows the number of rows to move
 * @param destination the row before which the moved rows are inserted
 */
public void moveRows(int start_row, int num_rows, int destination)
{
	table.requestChange(TableChangeHandler.TABLE_RECALC_ROW_HEIGHTS);
}
	
/**
 * Moves a set of columns to a position left of the destination. To move
 * columns to the end, use <CODE>JCTableEnum.MAXINT</CODE>.
 * @param start_column the first column
 * @param num_columns the number of columns to move
 * @param destination the column before which the moved columns are inserted
 */
public void moveColumns(int start_column, int num_columns, int destination)
{
	table.requestChange(TableChangeHandler.TABLE_RECALC_COLUMN_WIDTHS);
}

/**
 * Inserts the given number of rows.
 * @param start_row the first row
 * @param num_rows the number of rows to insert
 */
public void addRows(int start_row, int num_rows)
{
	table.requestChange(TableChangeHandler.TABLE_RECALC_ROW_HEIGHTS);
}

/**
 * Inserts the given number of columns.
 * @param start_column the first column
 * @param num_columns the number of columns to insert
 */
public void addColumns(int start_column, int num_columns)
{
	table.requestChange(TableChangeHandler.TABLE_RECALC_COLUMN_WIDTHS);
}

/**
 * Deletes the given number of rows.
 * @param start_row the first row
 * @param num_rows the number of rows to delete
 */
public void deleteRows(int start_row, int num_rows)
{
	table.requestChange(TableChangeHandler.TABLE_RECALC_ROW_HEIGHTS);
}

/**
 * Deletes the given number of columns.
 * @param start_column the first column
 * @param num_columns the number of columns to delete
 */
public void deleteColumns(int start_column, int num_columns)
{
	table.requestChange(TableChangeHandler.TABLE_RECALC_COLUMN_WIDTHS);
}

/**
 * Remaps heights after the row map changes.
 * @param current_map
 * @param new_map
 */
public void remapRows(int current_map[], int new_map[])
{
	resetRows();
	table.requestChange(TableChangeHandler.TABLE_RECALC_ROW_HEIGHTS);
}

/**
 * Remaps widths after the column map changes.
 * @param current_map
 * @param new_map
 */
public void remapColumns(int current_map[], int new_map[]) {
	resetColumns();
	table.requestChange(TableChangeHandler.TABLE_RECALC_COLUMN_WIDTHS);
}

/**
 * Removes cached height information.
 */
public void resetRows()
{
	heights = new Chain();
}

/**
 * Removes cached width information.
 */
public void resetColumns()
{
	widths = new Chain();
}

/**
 * The <code>Chain</code> class encodes a List of row heights or column widths
 * as a sequence of Runs. A Run is a series of items with the same value.
 */
public class Chain implements Serializable
{

    protected List  list = new Vector();
    protected int   defaultValue = 0;

    public Chain()
    {
        list = new Vector();
    }

    /**
     * Returns the value at the index.
     * @param index The index at whose index the value is required
     * @return value at the index
     */
    public int getValue(int index)
    {
        ListIterator iterator = list.listIterator();
        int value = defaultValue;
        try {
            while (iterator.hasNext()) {
                Run currentRun = (Run)iterator.next();
                if (index <= currentRun.end) {
                    value = currentRun.value;
                    break;
                }
            }
        }
        catch (ConcurrentModificationException e) {
            return(defaultValue);
        }
        return(value);
    }

    /**
     * Returns the sum of all index's values, up to the index.
     * @param index The index up to which to sum
     * @return sum of all index's values
     */
    public int getSum(int index)
    {
        int sum = 0;
        ListIterator iterator = list.listIterator();
        try {
            while (iterator.hasNext()) {
                Run currentRun = (Run)iterator.next();
                if (index > currentRun.end) {
                    sum += (currentRun.value * (currentRun.end -
                        currentRun.start + 1));
                }
                else {
                    sum += (currentRun.value * (index - currentRun.start));
                    break;
                }
            }
        }
        catch (ConcurrentModificationException e) {
            return(0);
        }
        return(sum);
    }

    /**
     * Returns the index at the given position.
     * @param position The position at which to determine the index
     * @return index at the given position
     */
    public int getIndex(int position)
    {
        ListIterator iterator = list.listIterator();
        Run currentRun = null;
        try {
            while (iterator.hasNext()) {
                currentRun = (Run)iterator.next();
                if (currentRun.value <= 0) {
                    continue;
                }
                int sum = (currentRun.value * (currentRun.end -
                    currentRun.start + 1));
                if (sum <= position) {
                    position -= sum;
                }
                else {
                    int offset = position / currentRun.value;
                    int index = currentRun.start + offset;
                    return(index);
                }
            }
        }
        catch (ConcurrentModificationException e) {
            return(0);
        }
        // Shouldn't be here
        if (currentRun != null) {
            return(currentRun.end);
        }
        // Or here
        return(-1);
    }

    /**
     * Sets the value at the index.
     * @param index
     * @param value
     */
    public void setValue(int index, int value)
    {
        setValue(index, index, value);
    }

    /**
     * Sets the value for a range of indexes (inclusive).
     * @param startIndex
     * @param endIndex
     * @param value
     */
    public void setValue(int startIndex, int endIndex, int value)
    {
        
        ListIterator iterator = list.listIterator();
        Run currentRun = null;
        boolean found = false;
        while (iterator.hasNext()) {
            currentRun = (Run)iterator.next();

            // Found the run that encompasses the start position
            if (startIndex <= currentRun.end) {
                // New run fits entirely within this run
                if (endIndex <= currentRun.end) {
                    // Setting to the same value - nothing to do
                    if (value == currentRun.value) {
                        found = true;
                        break;
                    }
                    boolean splitLeft = (startIndex > currentRun.start);
                    boolean splitRight = (endIndex < currentRun.end);
                    // Breaking the current run into 3 pieces
                    if (splitLeft && splitRight) {
                        splitLeftAndRight(iterator, currentRun, currentRun,
                                          startIndex, endIndex, value);
                    }
                    else {
                        if (splitLeft) {
                            splitLeft(iterator, currentRun, currentRun,
                                      startIndex, endIndex, value);
                        }
                        else if (splitRight) {
                            splitRight(iterator, currentRun, currentRun,
                                       startIndex, endIndex, value);
                        }
                        // An exact match
                        else {
                            currentRun.value = value;
                        }
                    }
                    found = true;
                    break;
                }
                else {
                    // New run is spread over a number of runs
                    Run leftRun = currentRun;
                    boolean splitLeft = (startIndex > leftRun.start);
                    // Now look for all of the entries that encompass this run
                    while (iterator.hasNext()) {
                        currentRun = (Run)iterator.next();
                        if (currentRun.end < endIndex) {
                            // Don't need this one, since it will be replaced by
                            // a larger run
                            iterator.remove();
                            currentRun = leftRun; // In case there's no next run
                        }
                        else {
                            break;
                        }
                    }
                    boolean splitRight = (endIndex < currentRun.end);
                    if (splitLeft && splitRight) {
                        splitLeftAndRight(iterator, leftRun, currentRun,
                                          startIndex, endIndex, value);
                    }
                    else if (splitLeft) {
                        splitLeft(iterator, leftRun, currentRun,
                                  startIndex, endIndex, value);
                    }
                    else if (splitRight) {
                        splitRight(iterator, leftRun, currentRun,
                                   startIndex, endIndex, value);
                    }
                    else {
                        
                        
                        
                        if (leftRun != currentRun) {
                            leftRun.end = currentRun.start - 1;
                        }
                        leftRun.value    = value;
                        currentRun.value = value;
                    }
                    found = true;
                    break;
                }
            }
        }
        if (!found) {
            list.add(new Run(startIndex, endIndex, value));
        }
        mergeRuns();
    }

    /**
     * Splits a given run into three runs, one on either side.
     * @param currentRun
     * @param endIndex
     * @param iterator
     * @param leftRun
     * @param startIndex
     * @param value
     */
    protected void splitLeftAndRight(ListIterator iterator, Run leftRun,
                                     Run currentRun, int startIndex,
                                     int endIndex, int value)
    {
        
        
        

        // Create a new middle run with the new value
        Run middleRun = new Run(startIndex, endIndex, value);

        // If left run is the same as current run, then we need to create a
        // right run.  Otherwise, use the current run as the right run.
        Run rightRun = null;
        if (leftRun == currentRun) {
            rightRun = new Run(endIndex + 1, currentRun.end, currentRun.value);
        }
        else {
            // Modify current run in place
            currentRun.start = endIndex + 1;
        }
        // Modify left run in place (important, if leftRun == currentRun, this
        // needs to be done after right run is created, so don't do this first).
        leftRun.end = startIndex - 1;

        if (rightRun == null) {
            // If we're using the current run as the right run, we need to
            // insert the middle run before the current run
            iterator.previous();
        }
        iterator.add(middleRun);
        if (rightRun != null) {
            iterator.add(rightRun);
        }
    }

    /**
     * Splits a given run into two runs, and adds a run to the right side.
     * @param currentRun
     * @param endIndex
     * @param iterator
     * @param leftRun
     * @param startIndex
     * @param value
     */
    protected void splitLeft(ListIterator iterator, Run leftRun, Run currentRun,
                             int startIndex, int endIndex, int value)
    {
        
        
        

        // If left run is the same as current run, then we need to create a
        // right run.  Otherwise, use the current run as the right run.
        Run rightRun = null;
        if (leftRun == currentRun) {
            rightRun = new Run(startIndex, endIndex, value);
        }
        else {
            currentRun.start = startIndex;
            currentRun.value = value;
        }
        // Modify left run in place
        leftRun.end = startIndex - 1;
        if (rightRun != null) {
            iterator.add(rightRun);
        }
    }

    /**
     * Splits a given run into two runs, and adds a run to the right side.
     * @param currentRun
     * @param endIndex
     * @param iterator
     * @param leftRun
     * @param startIndex
     * @param value
     */
    protected void splitRight(ListIterator iterator, Run leftRun,
                              Run currentRun, int startIndex, int endIndex,
                              int value)
    {
        
        
        

        // If left run is the same as current run, then we need to create a
        // right run.  Otherwise, use the current run as the right run.
        Run rightRun = null;
        if (leftRun == currentRun) {
            rightRun = new Run(endIndex + 1, currentRun.end, currentRun.value);
        }
        else {
            currentRun.start = endIndex + 1;
        }
        // Modify left run in place
        leftRun.end   = endIndex;
        leftRun.value = value;
        if (rightRun != null) {
            iterator.add(rightRun);
        }
    }

    /**
     * Walks through the entire list of runs, and coalesces adjacent pairs of
     * runs with the same value into a single run.
     */
    protected void mergeRuns()
    {
        List mergeList = list;
        ListIterator mergeIterator = mergeList.listIterator();
        Run currentRun = null;
        Run previousRun = null;
        if (mergeIterator.hasNext()) {
            previousRun = (Run)mergeIterator.next();
        }
        while (mergeIterator.hasNext()) {
            currentRun = (Run)mergeIterator.next();
            if (previousRun.value == currentRun.value) {
                
                previousRun.end = currentRun.end;
                mergeIterator.remove();
            }
            else {
                previousRun = currentRun;
            }
        }
    }

    public String toString()
    {
        return(list.toString());
    }

    public boolean equals(Object o)
    {
        if (o == this) {
            return(true);
        }
        if (!(o instanceof Chain)) {
            return(false);
        }
        Chain chain = (Chain)o;
        return(chain.list.equals(list));
    }

    public int hashcode()
    {
        return(list.hashCode());
    }

} // END Chain

/**
 * A Run is a series of items with the same value.
 * The start and end items are indexes into the run, that are inclusive.
 */
public class Run implements Serializable
{
	protected int value;	// value of this run
	protected int start;	// first index in the run
	protected int end;	    // last index in the run
	
	public Run(int start, int end, int value)
    {
		this.start = start;
		this.end   = end;
		this.value = value;
	}
	
	public String toString()
    {
		return("Run(start=" + start + ", end=" + end +
               ", value=" + value + ")");
	}

    public boolean equals(Object o)
    {
        if (!(o instanceof Run)) {
            return(false);
        }
        Run r = (Run)o;
        return((start == r.start) && (end == r.end) && (value == r.value));
    }

    public int hashCode()
    {
        int result = 17;
        result = 37 * result + start;
        result = 37 * result + end;
        result = 37 * result + value;
        return(result);
    }
} // END Run

public static void main(String[] args)
{
    String tags[] = {"cellLayout"};
    
    
    JCCellLayout layout = new JCCellLayout();
    Chain w = layout.widths;
    w.setValue(0, 9, 97);
    w.setValue(10, 21, 98);
    System.out.println("layout:" + w.toString());
    w.setValue(22, 35, 99);
    System.out.println("layout:" + w.toString());
    w.setValue(2, 5, 100);
    System.out.println("layout:" + w.toString());

    w.setValue(2, 2, 101);
    System.out.println("layout:" + w.toString());
    w.setValue(3, 3, 101);
    System.out.println("layout:" + w.toString());
    w.setValue(9, 9, 98);
    System.out.println("layout:" + w.toString());
    w.setValue(14, 14, 103);
    System.out.println("layout:" + w.toString());

    w.setValue(1, 8, 104);
    System.out.println("layout:" + w.toString());
    w.setValue(9, 12, 104);
    System.out.println("layout:" + w.toString());
    w.setValue(1, 14, 97);
    System.out.println("layout:" + w.toString());
    w.setValue(3, 29, 98);
    System.out.println("layout:" + w.toString());

    layout.resetColumns();
    w = layout.widths;
    w.setValue(0, 9, 97);
    w.setValue(2, 5, 100);
    System.out.println("layout:" + w.toString());

    w.setValue(10, 21, 98);
    w.setValue(2, 2, 101);
    w.setValue(3, 3, 101);
    w.setValue(9, 9, 98);
    w.setValue(14, 14, 103);
    System.out.println("layout:" + w.toString());

    w.setValue(1, 8, 104);
    System.out.println("layout:" + w.toString());
    w.setValue(9, 17, 104);
    System.out.println("layout:" + w.toString());

    w.setValue(1, 14, 97);
    System.out.println("layout:" + w.toString());

    w.setValue(22, 28, 99);
    System.out.println("layout:" + w.toString());
    w.setValue(20, 35, 100);
    System.out.println("layout:" + w.toString());

    w.setValue(27, 40, 101);
    System.out.println("layout:" + w.toString());

    w.setValue(15, 26, 98);
    System.out.println("layout:" + w.toString());

    w.setValue(0, 40, 100);
    System.out.println("layout:" + w.toString());
}

}

