/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCListTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

/**
 * <CODE>JCListTable</CODE> is a subclass of <CODE>JCTable</CODE> that provides list-like look
 * and behavior. The following properties are set to values different from <CODE>JCTable</CODE>:
 * <UL><LI><code>AllowResizeBy</code>, uses <code>JCTableEnum.RESIZE_BY_LABELS</code>
 * so that cell-level resize is disabled</LI>
 * <LI><code>FocusIndicator</code>, uses <code>JCTableEnum.FOCUS_NONE</code>, disabling
 * the focus indicator because the current row is shown through selection</LI>
 * <LI><code>SelectIncludeLabels</code>, <code>false</code></LI>
 * <LI><code>SelectionPolicy</code>, uses <code>JCTableEnum.SELECT_RANGE</code></LI></UL>
 * <P>In addition, a specialized <CODE>SelectionModel</CODE> is set that selects
 * by row.
 */
public class JCListTable extends JCTable {

public static final int SELECT_ROW = 0;
public static final int SELECT_COLUMN = 1;

protected int           selection_type = SELECT_ROW;

public JCListTable()
{
    super();
    allow_resize_by = JCTableEnum.RESIZE_BY_LABELS;
    focus_indicator = JCTableEnum.FOCUS_NONE;
    select_include_labels = false;

    setSelectionModel(new ListSelection(this));
    setSelectionPolicy(JCTableEnum.SELECT_RANGE);
    addTraverseCellListener(new Listeners());
}

/**
 * Returns the selection type.
 * @return selection type
 */
public int getSelectionType()
{
    return(selection_type);
}

/**
 * Sets the type of selection. Valid types are
 * <CODE>JCListTable.SELECT_ROW</CODE> and
 * <CODE>JCListTable.SELECT_COLUMN</CODE>.
 * @param value
 */
public void setSelectionType(int value)
{
    clearSelection();
    selection_type = value;
}

/**
 * List-specific listeners.
 */
protected class Listeners implements JCTraverseCellListener {

    public void traverseCell(JCTraverseCellEvent e) {}

    /**
     * Selects the row of the current cell on each traversal.
     * @param e
     */
    public void afterTraverseCell(JCTraverseCellEvent e)
    {
        if (e.getTraverseType() == JCTableEnum.TRAVERSE_POINTER) {
            return;
        }
        int row = e.getNextRow();
        int column = e.getNextColumn();
        SelectionModel selectionModel = JCListTable.this.getSelectionModel();
        if (selectionModel instanceof ListSelection) {
            if (JCListTable.this.getSelectionType() == JCListTable.SELECT_ROW) {
                // Only select the current row, if nothing is currently selected
                // otherwise, leave the previously selected rows alone!
                if (JCListTable.this.getSelectedCells() == null) {
                    JCListTable.this.setSelection(row, JCTableEnum.LABEL, row,
                                                  JCTableEnum.MAXINT);
                    selectionModel.select(SelectionModel.END, row, column);
                }
            }
            else {
                JCListTable.this.setSelection(JCTableEnum.LABEL, column,
                                              JCTableEnum.MAXINT, column);
            }
        }
        else {
            JCListTable.this.setSelection(row, column, row, column);
        }
    }
} // End Listeners

/**
 * The <code>JCListSelection</code> class supports row and column level
 * selection.
 */
protected class ListSelection extends JCSelection {

    public ListSelection(JCTable table)
    {
        super(table);
    }

    /**
     * Adds a <code>JCCellRange</code> to the selection.
     * @return <code>true</code> if the selection is accepted
     * @param cr
     */
    public boolean addSelection(JCCellRange cr)
    {
        if (JCListTable.this.selection_type == SELECT_ROW &&
            !(cr.start_column == JCTableEnum.LABEL &&
              cr.end_column == JCTableEnum.MAXINT))
        {
            return(false);
        }
        else if (JCListTable.this.selection_type == SELECT_COLUMN &&
                 !(cr.start_row == JCTableEnum.LABEL &&
                   cr.end_row == JCTableEnum.MAXINT))
        {
            return(false);
        }

        return(super.addSelection(cr));
    }

    /**
     * Extends the current selection to the given row and column.
     * @return <code>true</code> if the selection is accepted
     * @param row
     * @param column
     */
    public boolean extendSelection(int row, int column)
    {
        if (selected_cells == null || selected_cells.size() == 0) {
            return(false);
        }

        if (JCListTable.this.selection_type == SELECT_ROW) {
            column = JCTableEnum.MAXINT;
        }
        else if (JCListTable.this.selection_type == SELECT_COLUMN) {
            row = JCTableEnum.MAXINT;
        }
        return(super.extendSelection(row, column));
    }

    /**
     * Sets the selection to a <code>JCCellRange</code>.
     * @return <code>true</code> if the selection is accepted
     * @param cr
     */
    public boolean setSelection(JCCellRange cr)
    {
        if (JCListTable.this.selection_type == SELECT_ROW &&
            !(cr.start_column == JCTableEnum.LABEL &&
              cr.end_column == JCTableEnum.MAXINT))
        {
            return(false);
        }
        else if (JCListTable.this.selection_type == SELECT_COLUMN &&
                 !(cr.start_row == JCTableEnum.LABEL &&
                   cr.end_row == JCTableEnum.MAXINT))
        {
            return(false);
        }
        return(super.setSelection(cr));
    }

    /**
     * Adds, extends, deselects, or ends a selection based on a row or column.
     * This method is called by the mouse and key process handler and should not
     * be used for programmatic manipulation of selected regions.
     *
     * @return <code>true</code> if the model supports the selection
     * @param column
     * @param row
     * @param state
     */
    public boolean select(int state, int row, int column)
    {
        // Ensure that the row and column is valid
        if (row == JCTableEnum.NOVALUE || column == JCTableEnum.NOVALUE) {
            // If we return false, interactive selection is terminated, so
            // we ignore the values but tell table that the selection process
            // can still continue
            return(true);
        }

        // Check for an attempt to reselect the same cell as previously selected
        if (row == last_selected_row &&
            column == last_selected_column &&
            state != END)
        {
            return(true);
        }
        else if (row == JCTableEnum.NOVALUE || column == JCTableEnum.NOVALUE) {
            return(true);
        }

        last_selected_row = row;
        last_selected_column = column;

        if (JCListTable.this.selection_type == SELECT_ROW &&
            row == JCTableEnum.LABEL)
        {
            return(false);
        }
        else if (JCListTable.this.selection_type == SELECT_COLUMN &&
                 column == JCTableEnum.LABEL)
        {
            return(false);
        }

        switch (state) {
        case SelectionModel.SELECT:
            if (JCListTable.this.selection_type == SELECT_ROW) {
                return setSelection(row, JCTableEnum.LABEL, row,
                                    JCTableEnum.MAXINT);
            }
            else {
                return setSelection(JCTableEnum.LABEL, column,
                                    JCTableEnum.MAXINT, column);
            }
        case SelectionModel.EXTEND:
            return extendSelection(row, column);
        case SelectionModel.ADD:
            if (JCListTable.this.selection_type == SELECT_ROW) {
                return addSelection(row, JCTableEnum.LABEL, row,
                                    JCTableEnum.MAXINT);
            }
            else {
                return addSelection(JCTableEnum.LABEL, column,
                                    JCTableEnum.MAXINT, column);
            }
        case SelectionModel.END:
            cleanup();
            if (table.selectListeners != null && selected_cells != null) {
                // notifies all listeners that a new range will be selected.
                JCSelectEvent event =
                    new JCSelectEvent(table, JCSelectEvent.AFTER_SELECTION,
                                      row, JCTableEnum.LABEL,
                                      row, JCTableEnum.MAXINT,
                                      JCSelectEvent.END);
                table.fireJCSelectEvent(event);
            }
            return(true);
        }
        return(false);
    }
}

}
