/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCPrintTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:33 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.table.resources.LocaleBundle;
import com.klg.jclass.util.JCListenerList;

import java.awt.*;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

/**
 * The <CODE>JCPrintTable</CODE> class controls paging, header and footers, and
 * print layout for JClass LiveTable.
 */
public class JCPrintTable extends JCTable implements Printable
{

public static final int   MARGIN_IN_INCHES = 0;
public static final int   MARGIN_IN_PIXELS = 1;

protected JCTable         table;
protected int             page_width = 0;
protected int             page_height = 0;
protected int             page_resolution = 72;

protected int             table_width = 0;
protected int             table_height = 0;

protected Insets          margin = new Insets(72, 72, 72, 72);
protected int             margin_unit = MARGIN_IN_PIXELS;

protected int             current_page;

public static final int   ALL_PAGES = -1;
protected int             print_page = ALL_PAGES;

protected PageFormat      page_format;

// listeners
protected JCListenerList  printListeners = null;

/**
 * Creates a new <CODE>JCPrintTable</CODE>. This class creates a
 * "copy" of the table's visible properties and links to the same data source.
 * Cell height and width is copied by pixel size. Scrollbars and components are
 * not part of the printed table and are not copied.
 * @param table
 */
public JCPrintTable(JCTable table)
{
    this.table = table;
    loadSettings(table);
}

/**
 * Displays a print preview screen with the specified caption.
 * @param caption
 */
public void showPrintPreview(String caption)
{
    JCPrintPreview preview = new JCPrintPreview(caption, this);
    preview.showPage(0);
}

/**
 * Prints all pages using the default settings.
 */
public void print()
{
    PrinterJob job = PrinterJob.getPrinterJob();
    if (job != null) {
        try {
            if (page_format == null) {
                setPageFormat(job.defaultPage());
            }
            job.setPrintable(this, page_format);
            setPageDimensions((int)page_format.getWidth(),
                              (int)page_format.getHeight());
            setPrintPage(ALL_PAGES);
            job.print();
        }
        catch (PrinterException pe) {
            System.out.println(pe.getMessage());
        }
    }
}

/**
 * Sets the page format. Table printing only supports a single
 * page format for all pages.
 */
public void setPageFormat(PageFormat format)
{
    this.page_format = format;
}

/**
 * Sets the page to print. Use <code>JCTableEnum.ALL_PAGES</code> to print
 * everything.
 * @param page
 */
public void setPrintPage(int page)
{
    this.print_page = page;
}

/**
 * Gets the page format. Table printing only supports a single
 * page format for all pages.
 */
public PageFormat getPageFormat()
{
    return(this.page_format);
}

/**
 * Prints the given page using the specified page format.
 * This method is used internally by JClass LiveTable and
 * should not be used by applications.
 */
public int print(Graphics gc, PageFormat pageFormat, int pageIndex)
{
    if (pageIndex >= getNumPages() ||
        (print_page != ALL_PAGES && pageIndex > 0))
    {
        return(Printable.NO_SUCH_PAGE);
    }

    if (print_page == ALL_PAGES) {
        this.printPage(gc, pageIndex);
    }
    else {
        this.printPage(gc, print_page);
    }
    return(Printable.PAGE_EXISTS);
}

/**
 * Loads/copies settings from the source table.
 * @param table
 */
protected void loadSettings(JCTable table)
{
    // Pass one --- just make a link to the table settings.
    columnWidths = table.columnWidths;
    cellStyles = table.cellStyles;
    rowHeights = table.rowHeights;

    // Take the tables cellRendererPane for our own -- it needs a visible
    // component to draw off of
    if (cellRendererPane != null && cellRendererPane.getParent() == this) {
        remove(cellRendererPane);
    }
    cellRendererPane = table.cellRendererPane;

    // Handlers
    cellAreaHandler = new PrintCellAreaHandler(this);
    cellAreaHandler.cellAreaRenderer =
        new PrintCellAreaRenderer(cellAreaHandler);
    cellAreaHandler.cellAreaRenderer.borderColor =
        table.cellAreaHandler.cellAreaRenderer.borderColor;
    cellAreaHandler.cellAreaRenderer.borderHighlightColor =
        table.cellAreaHandler.cellAreaRenderer.borderHighlightColor;
    cellAreaHandler.cellAreaRenderer.borderShadowColor =
        table.cellAreaHandler.cellAreaRenderer.borderShadowColor;

    cellAreaHandler.cellArea = table.cellAreaHandler.cellArea;
    cellAreaHandler.frozenRowArea = table.cellAreaHandler.frozenRowArea;
    cellAreaHandler.columnLabelArea = table.cellAreaHandler.columnLabelArea;
    cellAreaHandler.frozenColumnArea = table.cellAreaHandler.frozenColumnArea;
    cellAreaHandler.frozenCellArea = table.cellAreaHandler.frozenCellArea;
    cellAreaHandler.frozenColumnLabelArea =
        table.cellAreaHandler.frozenColumnLabelArea;
    cellAreaHandler.rowLabelArea = table.cellAreaHandler.rowLabelArea;
    cellAreaHandler.frozenRowLabelArea =
        table.cellAreaHandler.frozenRowLabelArea;

    cellAreaHandler.areas = table.cellAreaHandler.areas;
    cellAreaHandler.cellInfo = table.cellAreaHandler.cellInfo;
    cellAreaHandler.cellOrigin = table.cellAreaHandler.cellOrigin;

    dataView = table.dataView;

    int rows = getNumRows();
    int columns = getNumColumns();

    // column and row label offsets need to be set BEFORE creating the
    // cell layout!
    column_label_offset = table.column_label_offset;
    row_label_offset = table.row_label_offset;

    cellLayout = new PrintCellLayout(rows, columns);
    cellLayout.setTable(this);
    for (int r = JCTableEnum.LABEL; r < rows; r++) {
        cellLayout.setHeight(r, table.cellLayout.getHeight(r));
    }
    for (int c = JCTableEnum.LABEL; c < columns; c++) {
        cellLayout.setWidth(c, table.cellLayout.getWidth(c));
    }

    tableLayout = new TableLayout(cellAreaHandler);

    hsb_display = JCTableEnum.SCROLLBAR_NEVER;
    vsb_display = JCTableEnum.SCROLLBAR_NEVER;

    component_border_width = table.component_border_width;
    ComponentModel componentModel = table.getFocusManager().getComponentModel();
    getFocusManager().setComponentModel(componentModel);

    // Internal properties
    cell_border_width = table.cell_border_width;
    column_label_display = table.column_label_display;
    column_label_placement = table.column_label_placement;
    focus_color = table.focus_color;
    focus_indicator = table.focus_indicator;
    frame_border = table.frame_border;
    frame_border_width = table.frame_border_width;
    frozen_columns = table.frozen_columns;
    frozen_column_placement = table.frozen_column_placement;
    frozen_rows = table.frozen_rows;
    frozen_row_placement = table.frozen_row_placement;
    margin_height = table.margin_height;
    margin_width = table.margin_width;
    row_label_display = table.row_label_display;
    row_label_placement = table.row_label_placement;
    selected_background = table.selected_background;
    selected_foreground = table.selected_foreground;
}

//
// Main print methods
// ----------------------------------------------------------------------------
//

/**
 * Generates a new page.
 * @param gc
 * @param page
 */
public void printPage(Graphics gc, int page)
{
    current_page = page;
    JCCellRange cr;
    try {
        cr = getPageCellRange(page);
    }
    catch (IllegalSizeException e) {
        System.out.println(e.getMessage());
        return;
    }

    if (cr == null) {
        // No pages to print.
        return;
    }

    //Dimension d = getTablePageDimension(cr);
    printPage(gc, cr);
}

/**
 * Draws the table for a given cell range.
 * @param cr The cell range
 * @param gc The graphics context to use
 */
public void printPage(Graphics gc, JCCellRange cr)
{
    ((PrintCellAreaHandler)cellAreaHandler).page_range = cr;
    CellLayoutModel clm = table.getCellLayout();

    Dimension d = new Dimension(0, 0);
    d.width  = clm.getWidth(JCTableEnum.LABEL) + getRowLabelOffset() +
               getFrozenColumnWidth() + getColumnPosition(cr.end_column) -
               getColumnPosition(cr.start_column) + clm.getWidth(cr.end_column);
    d.height = clm.getHeight(JCTableEnum.LABEL) + getColumnLabelOffset() +
               getFrozenRowHeight() + getRowPosition(cr.end_row) -
               getRowPosition(cr.start_row) + clm.getHeight(cr.end_row);

    gc.setColor(Color.white);
    gc.fillRect(0, 0, page_width, page_height);

    Point p = getTableOffset();
    gc.setClip(p.x, p.y, table_width, table_height);

    // break spans that cross pages
    if (table.spanHandler != null) {

        // Get the orignal spans from table. This must be done for each page
        // otherwise the previous page spans are used.
        spanHandler = new SpanHandler();
        spanHandler.setTable(this);

        // copy over the ranges
        Collection spans = table.getSpannedRanges();
        try {
            if (spans != null) {
                Iterator it = spans.iterator();
                while (it.hasNext()) {
                    JCCellRange cr3 = (JCCellRange)it.next();
                    JCCellRange sr = new JCCellRange(cr3.start_row,
                                                     cr3.start_column,
                                                     cr3.end_row,
                                                     cr3.end_column);
                    spanHandler.addSpannedRange(sr);
                }
            }
        }
        catch (Exception e) {}

        // Then split them in our local spanHandler
        splitSpans(cr);
    }

    // Do frozen column headers
    int frozen_columns = getFrozenColumns();
    if (frozen_columns > 0) {
        for (int c = 0; c < frozen_columns; c++) {
            cellAreaHandler.cellAreaRenderer.paintCell(gc, JCTableEnum.LABEL,
                                                       c);
        }
    }
    // Do column headers
    for (int c = cr.start_column; c <= cr.end_column; c++) {
        cellAreaHandler.cellAreaRenderer.paintCell(gc, JCTableEnum.LABEL, c);
    }
    // do frozen row headers
    int frozen_rows = getFrozenRows();
    if (frozen_rows > 0) {
        for (int r = 0; r < frozen_rows; r++) {
            cellAreaHandler.cellAreaRenderer.paintCell(gc, r,
                                                       JCTableEnum.LABEL);
        }
    }
    // Do row headers
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        cellAreaHandler.cellAreaRenderer.paintCell(gc, r, JCTableEnum.LABEL);
    }
    // Draw frozen dead cells
    if (frozen_columns > 0 && frozen_rows > 0) {
        for (int r = 0; r < frozen_rows; r++) {
            for (int c = 0; c < frozen_columns; c++) {
                cellAreaHandler.cellAreaRenderer.paintCell(gc, r, c);
            }
        }
    }
    // Draw frozen columns
    if (frozen_columns > 0) {
        for (int r = cr.start_row; r <= cr.end_row; r++) {
            for (int c = 0; c < frozen_columns; c++) {
                cellAreaHandler.cellAreaRenderer.paintCell(gc, r, c);
            }
        }
    }
    // Draw frozen rows
    if (frozen_rows > 0) {
        for (int r = 0; r < frozen_rows; r++) {
            for (int c = cr.start_column; c <= cr.end_column; c++) {
                cellAreaHandler.cellAreaRenderer.paintCell(gc, r, c);
            }
        }
    }

    // Do cells
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            cellAreaHandler.cellAreaRenderer.paintCell(gc, r, c);
        }
    }

    if (getFrameBorderWidth() > 0) {
        // Calculate rectangles for frame border
        Rectangle startBounds = cellAreaHandler.getBounds(cr.start_row,
                                                          JCTableEnum.LABEL);
        Rectangle endBounds = cellAreaHandler.getBounds(cr.end_row,
                                                        JCTableEnum.LABEL);
        Rectangle rowLabelRect = new Rectangle(startBounds.x,
                                               startBounds.y,
                                               startBounds.width,
                                               endBounds.y +
                                               endBounds.height -
                                               startBounds.y);
        if (frozen_rows > 0) {
            Rectangle labelBounds =
                cellAreaHandler.getBounds(0, JCTableEnum.LABEL);
            Rectangle frozenLabelBounds =
                cellAreaHandler.getBounds(frozen_rows - 1, JCTableEnum.LABEL);
            Rectangle frRowLabelRect = new Rectangle(labelBounds.x,
                                                     labelBounds.y,
                                                     labelBounds.width,
                                                     frozenLabelBounds.y +
                                                     frozenLabelBounds.height -
                                                     labelBounds.y);
            rowLabelRect = new Rectangle(Math.min(rowLabelRect.x,
                                                  frRowLabelRect.x),
                                         Math.min(rowLabelRect.y,
                                                  frRowLabelRect.y),
                                         Math.max(rowLabelRect.width,
                                                  frRowLabelRect.width),
                                         rowLabelRect.height +
                                         frRowLabelRect.height);
        }
        startBounds = cellAreaHandler.getBounds(JCTableEnum.LABEL,
                                                cr.start_column);
        endBounds = cellAreaHandler.getBounds(JCTableEnum.LABEL, cr.end_column);
        Rectangle colLabelRect = new Rectangle(startBounds.x,
                                               startBounds.y,
                                               endBounds.x +
                                               endBounds.width -
                                               startBounds.x,
                                               startBounds.height);
        if (frozen_columns > 0) {
            Rectangle labelBounds =
                cellAreaHandler.getBounds(JCTableEnum.LABEL, 0);
            Rectangle frozenLabelBounds =
                cellAreaHandler.getBounds(JCTableEnum.LABEL,
                                          frozen_columns - 1);
            Rectangle frColLabelRect = new Rectangle(labelBounds.x,
                                                     labelBounds.y,
                                                     frozenLabelBounds.x +
                                                     frozenLabelBounds.width -
                                                     labelBounds.x,
                                                     labelBounds.height);
            colLabelRect = new Rectangle(Math.min(colLabelRect.x,
                                                  frColLabelRect.x),
                                         Math.min(colLabelRect.y,
                                                  frColLabelRect.y),
                                         colLabelRect.width +
                                         frColLabelRect.width,
                                         Math.max(colLabelRect.height,
                                                  frColLabelRect.height));
        }
        Rectangle cellRect = new Rectangle(colLabelRect.x,
                                           rowLabelRect.y,
                                           colLabelRect.width,
                                           rowLabelRect.height);

        // draw frame borders
        CellBorderModel cellBorder = table.getFrameBorder();
        Color bg = table.getBackground();

        cellBorder.drawBackground(gc, frame_border_width,
                                  JCTableEnum.BORDERSIDE_ALL,
                                  rowLabelRect.x - frame_border_width,
                                  rowLabelRect.y - frame_border_width,
                                  rowLabelRect.width + 2 * frame_border_width,
                                  rowLabelRect.height + 2 * frame_border_width,
                                  bg.brighter(), bg.darker(), bg);
        cellBorder.drawBackground(gc, frame_border_width,
                                  JCTableEnum.BORDERSIDE_ALL,
                                  colLabelRect.x - frame_border_width,
                                  colLabelRect.y - frame_border_width,
                                  colLabelRect.width + 2 * frame_border_width,
                                  colLabelRect.height + 2 * frame_border_width,
                                  bg.brighter(), bg.darker(), bg);
        cellBorder.drawBackground(gc, frame_border_width,
                                  JCTableEnum.BORDERSIDE_ALL,
                                  cellRect.x - frame_border_width,
                                  cellRect.y - frame_border_width,
                                  cellRect.width + 2 * frame_border_width,
                                  cellRect.height + 2 * frame_border_width,
                                  bg.brighter(), bg.darker(), bg);
    }

    doPrintPageHeader(gc, current_page);
    doPrintBody(gc, current_page);
    doPrintPageFooter(gc, current_page);
    ((PrintCellAreaHandler)cellAreaHandler).page_range = null;
}


//
// Printing public methods
// ----------------------------------------------------------------------------
//

/**
 * Returns the units for margins.
 * @return units for margins
 */
public int getMarginUnits()
{
    return(margin_unit);
}

/**
 * Sets the units for margin calculations.
 * @param mu
 */
public void setMarginUnits(int mu)
{
    this.margin_unit =
        (mu == MARGIN_IN_INCHES) ? MARGIN_IN_INCHES : MARGIN_IN_PIXELS;
}

/**
 * Returns the number of horizontal pages.
 * @return number of horizontal pages
 * @throws IllegalSizeException
 */
public int getNumHorizontalPages() throws IllegalSizeException
{
    return(getNumColumnPages());
}

/**
 * Returns the number of pages required to print the table.
 * @return number of pages required to print the table
 * @throws IllegalSizeException
 */
public int getNumPages() throws IllegalSizeException
{
    if (page_width * page_height == 0) {
        return(0);
    }
    return(getNumColumnPages() * getNumRowPages());
}

/**
 * Returns the number of vertical pages.
 * @return number of vertical pages
 */
public int getNumVerticalPages()
{
    return(getNumRowPages());
}

/**
 * Returns the cell range that defines the boundary of cells that are on
 * a page. The size of the cell range considers labels and frozen
 * cells, but does not include them in the range itself.
 * @return cell range
 * @throws IllegalSizeException
 * @param page
 */
public JCCellRange getPageCellRange(int page) throws IllegalSizeException
{
    if (getNumPages() == 0) {
        return(null);
    }
    JCCellPosition c1 = getPageStartCell(page);
    JCCellPosition c2 = getPageEndCell(page);

    return(new JCCellRange(c1.row, c1.column, c2.row, c2.column));
}

/**
 * Returns the page dimensions.
 * @return page dimensions
 */
public Dimension getPageDimensions()
{
    return(new Dimension(page_width, page_height));
}

/**
 * Sets the dimensions of a page.
 * @param page_height
 * @param page_width
 */
public void setPageDimensions(int page_width, int page_height)
{
    if (page_width == 0 || page_height == 0) {
        page_width = 612;
        page_height = 792;
        page_resolution = 72;
    }
    this.page_width = page_width;
    this.page_height = page_height;
    setInternalDimensions();
}

/**
 * Returns the height of a page.
 * @return height of a page
 */
public int getPageHeight()
{
    return(page_height);
}

/**
 * Sets the height of a page.
 * @param page_height
 */
public void setPageHeight(int page_height)
{
    this.page_height = page_height;
    setInternalDimensions();
}

/**
 * Returns the page margins.
 * @return page margins
 */
public Insets getPageMargins()
{
    return(margin);
}

/**
 * Returns the default margins.
 * @return default margins
 */
public Insets getDefaultPageMargins()
{
    this.margin_unit = MARGIN_IN_PIXELS;
    this.margin = new Insets(72, 72, 72, 72);
    return(margin);
}

/**
 * Sets the page margins. Passing null sets the margins to the default 1".
 * @param margin
 */
public void setPageMargins(Insets margin)
{
    if (margin == null) {
        margin = getDefaultPageMargins();
    }
    else {
        this.margin = margin;
    }
    setInternalDimensions();
}

/**
 * Returns the page resolution.
 * @return page resolution
 */
public int getPageResolution()
{
    return(page_resolution);
}

/**
 * Sets the page resolution in pixels/inch.
 * @param page_resolution
 */
public void setPageResolution(int page_resolution)
{
    this.page_resolution = page_resolution;
    setInternalDimensions();
}

/**
 * Returns the width of a page.
 * @return width of a page
 */
public int getPageWidth()
{
    return(page_width);
}

/**
 * Sets the width of a page.
 * @param page_width
 */
public void setPageWidth(int page_width)
{
    this.page_width = page_width;
    setInternalDimensions();
}

/**
 * Replaces <CODE>Table.getPosition</CODE> to provide for pixel placement of
 * any cell or label on a print page independent from the standard clip logic.
 * @param column
 * @param row
 * @return
 */
public Point getBounds(int row, int column)
{
    Point pos = new Point(0, 0);
    JCCellRange cr = getPageCellRange(current_page);

    // Column headers
    if (row == JCTableEnum.LABEL) {
        if (column < frozen_columns) {
            if (getRowLabelPlacement() == JCTableEnum.PLACE_LEFT) {
                if (getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT) {
                    pos.x = getRowLabelWidth() + getRowLabelOffset() +
                            3 * getFrameBorderWidth();
                }
                else {
                    pos.x = getTableDimensions(current_page).width -
                            getFrozenColumnWidth() - getFrameBorderWidth();
                }
            }
            else if (getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT) {
                pos.x = getFrameBorderWidth();
            }
            else {
                pos.x = getTableDimensions(current_page).width -
                        getRowLabelWidth() - getRowLabelOffset() -
                        getFrozenColumnWidth() - 3 * getFrameBorderWidth();
            }
            pos.x += getColumnPosition(column) - getColumnPosition(0);
        }
        else {
            if (getRowLabelPlacement() == JCTableEnum.PLACE_LEFT) {
                pos.x = getRowLabelWidth() + getRowLabelOffset() +
                        3 * getFrameBorderWidth();
            }
            else {
                pos.x = getFrameBorderWidth();
            }
            if (frozen_columns > 0 &&
                getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT)
            {
                pos.x += getFrozenColumnWidth();
            }
            pos.x += getColumnPosition(column) -
                     getColumnPosition(cr.start_column);
        }
        if (getColumnLabelPlacement() == JCTableEnum.PLACE_TOP) {
            pos.y = getFrameBorderWidth();
        }
        else {
            pos.y = getTableDimensions(current_page).height -
                    getColumnLabelHeight() - getColumnLabelOffset() -
                    2 * getFrameBorderWidth();
        }
    }
    // Row headers
    else if (column == JCTableEnum.LABEL) {
        if (getRowLabelPlacement() == JCTableEnum.PLACE_LEFT) {
            pos.x = getFrameBorderWidth();
        }
        else {
            pos.x = getTableDimensions(current_page).width -
                    getRowLabelWidth() - getRowLabelOffset() -
                    2 * getFrameBorderWidth();
        }
        if (row < frozen_rows) {
            if (getColumnLabelPlacement() == JCTableEnum.PLACE_TOP) {
                if (getFrozenRowPlacement() == JCTableEnum.PLACE_TOP)
                    pos.y = getColumnLabelHeight() + getColumnLabelOffset() +
                            3 * getFrameBorderWidth();
                else {
                    pos.y = getTableDimensions(current_page).height -
                            getFrozenRowHeight() - getFrameBorderWidth();
                }
            }
            else if (getFrozenRowPlacement() == JCTableEnum.PLACE_TOP) {
                pos.y = getFrameBorderWidth();
            }
            else {
                pos.y = getTableDimensions(current_page).height -
                        getColumnLabelHeight() - getColumnLabelOffset() -
                        getFrozenRowHeight() - 3 * getFrameBorderWidth();
            }
            pos.y += getRowPosition(row) - getRowPosition(0);
        }
        else {
            if (getColumnLabelPlacement() == JCTableEnum.PLACE_TOP) {
                pos.y = getColumnLabelHeight() + getColumnLabelOffset() +
                        3 * getFrameBorderWidth();
            }
            else {
                pos.y = getFrameBorderWidth();
            }
            if (frozen_rows > 0 &&
                getFrozenRowPlacement() == JCTableEnum.PLACE_TOP)
            {
                pos.y += getFrozenRowHeight();
            }
            pos.y += getRowPosition(row) - getRowPosition(cr.start_row);
        }
    }
    else {
        if (column < frozen_columns) {
            if (getRowLabelPlacement() == JCTableEnum.PLACE_LEFT) {
                if (getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT) {
                    pos.x = getRowLabelWidth() + getRowLabelOffset() +
                            3 * getFrameBorderWidth();
                }
                else {
                    pos.x = getTableDimensions(current_page).width -
                            getFrozenColumnWidth() - getFrameBorderWidth();
                }
            }
            else if (getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT) {
                pos.x = getFrameBorderWidth();
            }
            else {
                pos.x = getTableDimensions(current_page).width -
                        getRowLabelWidth() - getRowLabelOffset() -
                        getFrozenColumnWidth() - 3 * getFrameBorderWidth();
            }
            pos.x += getColumnPosition(column) - getColumnPosition(0);
        }
        else {
            if (getRowLabelPlacement() == JCTableEnum.PLACE_LEFT) {
                pos.x = getRowLabelWidth() + getRowLabelOffset() +
                        3 * getFrameBorderWidth();
            }
            else {
                pos.x = getFrameBorderWidth();
            }
            if (frozen_columns > 0 &&
                getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT)
            {
                pos.x += getFrozenColumnWidth();
            }
            pos.x += getColumnPosition(column) - getColumnPosition(cr.start_column);
        }
        if (row < frozen_rows) {
            if (getColumnLabelPlacement() == JCTableEnum.PLACE_TOP) {
                if (getFrozenRowPlacement() == JCTableEnum.PLACE_TOP) {
                    pos.y = getColumnLabelHeight() + getColumnLabelOffset() +
                            3 * getFrameBorderWidth();
                }
                else {
                    pos.y = getTableDimensions(current_page).height -
                            getFrozenRowHeight() - getFrameBorderWidth();
                }
            }
            else if (getFrozenRowPlacement() == JCTableEnum.PLACE_TOP) {
                pos.y = getFrameBorderWidth();
            }
            else {
                pos.y = getTableDimensions(current_page).height -
                        getColumnLabelHeight() - getColumnLabelOffset() -
                        getFrozenRowHeight() - 3 * getFrameBorderWidth();
            }
            pos.y += getRowPosition(row) - getRowPosition(0);
        }
        else {
            if (getColumnLabelPlacement() == JCTableEnum.PLACE_TOP) {
                pos.y = getColumnLabelHeight() + getColumnLabelOffset() +
                        3 * getFrameBorderWidth();
            }
            else {
                pos.y = getFrameBorderWidth();
            }
            if (frozen_rows > 0 &&
                getFrozenRowPlacement() == JCTableEnum.PLACE_TOP)
            {
                pos.y += getFrozenRowHeight();
            }
            pos.y += getRowPosition(row) - getRowPosition(cr.start_row);
        }
    }

    Point p = getTableOffset();
    pos.x += p.x;
    pos.y += p.y;
    return(pos);
}

/**
 * Returns the Dimensions that JClass LiveTable uses on a given page.
 * @return Dimensions used on a given page
 * @param page
 */
public Dimension getTableDimensions(int page)
{
    return(getTablePageDimension(getPageCellRange(page)));
}

//
// Helper methods
// ----------------------------------------------------------------------------
//

/**
 * Returns the bounds of a range of non-frozen cells.
 * @return bounds of a range of non-frozen cells
 * @param cr
 */
private Rectangle getCellRangeBounds(JCCellRange cr)
{
    return(getCellLayout().getBounds(cr.start_row, cr.start_column,
                                     cr.end_row, cr.end_column));
}

/**
 * Returns the column label height.
 * @return column label height
 */
private int getColumnLabelHeight()
{
    if (column_label_display) {
        return(getCellLayout().getHeight(JCTableEnum.LABEL));
    }
    return(0);
}

/**
 * Returns the start position of a column.
 * @return start position of a column
 * @param column
 */
private int getColumnPosition(int column)
{
    CellLayoutModel clm = getCellLayout();
    return(clm.getColumnPosition(column));
}

/**
 * Returns the width of frozen columns.
 * @return width of frozen columns
 */
private int getFrozenColumnWidth()
{
    CellLayoutModel clm = getCellLayout();

    int width = 0;
    for (int c = 0; c < getFrozenColumns(); c++) {
        width += clm.getWidth(c);
    }
    return(width);
}

/**
 * Returns the height of frozen rows.
 * @return height of frozen rows
 */
private int getFrozenRowHeight()
{
    CellLayoutModel clm = getCellLayout();

    int height = 0;
    for (int r = 0; r < getFrozenRows(); r++) {
        height += clm.getHeight(r);
    }
    return(height);
}

/**
 * Sets the internal table dimensions (page dimensions minus margins).
 */
private void setInternalDimensions()
{
    if (page_width == 0 || page_height == 0) {
        // Page dimensions not yet set, don't set internal dimensions yet
        return;
    }

    int unit_conversion =
        (margin_unit == MARGIN_IN_INCHES) ? page_resolution : 1;
    table_width = page_width - unit_conversion * (margin.left + margin.right);
    table_height = page_height - unit_conversion * (margin.top + margin.bottom);

    CellLayoutModel clm = getCellLayout();

    // Restrict the width to the largest available size
    int working_width = table_width - getRowLabelWidth() -
                        getRowLabelOffset() - getFrozenColumnWidth() -
                        4 * getFrameBorderWidth();
    for (int c = getFrozenColumns(); c < getNumColumns(); c++) {
        if (clm.getWidth(c) > working_width) {
            clm.setWidth(c, working_width);
        }
    }
    // Restrict the height to the largest available size
    int working_height = table_height - getColumnLabelHeight() -
                         getColumnLabelOffset() - getFrozenRowHeight() -
                         4 * getFrameBorderWidth();
    for (int r = getFrozenRows(); r < getNumRows(); r++) {
        if (clm.getHeight(r) > working_height) {
            clm.setHeight(r, working_height);
        }
    }
}

/**
 * Private method returns the number of horizontal pages.
 * @return number of horizontal pages
 * @throws IllegalSizeException
 */
private int getNumColumnPages() throws IllegalSizeException
{
    int current;
    int column_pages = 1;

    int row_label_width = getRowLabelWidth();
    int frozen_column_width = getFrozenColumnWidth();

    // Can't print table if the row labels and frozen columns are larger than
    // the page width
    if (row_label_width + row_label_offset + frozen_column_width +
        4 * frame_border_width > table_width)
    {
        throw new IllegalSizeException("The total width of row labels and " +
                                       "frozen columns is larger than the " +
                                       "page width.");
    }
    CellLayoutModel clm = getCellLayout();

    current = row_label_width + row_label_offset + frozen_column_width +
              4 * frame_border_width;
    for (int c = frozen_columns; c < getNumColumns(); c++) {
        current += clm.getWidth(c);
        if (current > table_width) {
            current = row_label_width + row_label_offset +
                      frozen_column_width + 4 * frame_border_width +
                      clm.getWidth(c);
            column_pages++;
        }
    }
    return(column_pages);
}

/**
 * Private method returns the number of vertical pages.
 * @return number of vertical pages
 * @throws IllegalSizeException
 */
private int getNumRowPages() throws IllegalSizeException
{
    int current;
    int row_pages = 1;

    int column_label_height = getColumnLabelHeight();
    int frozen_row_height = getFrozenRowHeight();

    // Can't print table if the column labels and frozen rows are higher than
    // the page height
    if (column_label_height + column_label_offset + frozen_row_height +
        4 * frame_border_width > table_height)
    {
        throw new IllegalSizeException("The total height of column labels " +
                                       "and frozen rows is larger than the " +
                                       "page height.");
    }

    CellLayoutModel clm = getCellLayout();

    current = column_label_height + column_label_offset + frozen_row_height +
              4 * frame_border_width;
    for (int r = getFrozenRows(); r < getNumRows(); r++) {
        current += clm.getHeight(r);
        if (current > table_height) {
            current = column_label_height + column_label_offset +
                      frozen_row_height + 4 * frame_border_width +
                      clm.getHeight(r);
            row_pages++;
        }
    }
    return(row_pages);
}

/**
 * Private method returns the first cell on a page.
 * @return
 * @param page
 * @throws IllegalSizeException
 */
private JCCellPosition getPageStartCell(int page) throws IllegalSizeException
{
    int column_pages = getNumColumnPages();

    int rp = page / column_pages + 1;
    int cp = page % column_pages + 1;

    int column = frozen_columns;
    if (cp == 1) {
        column = frozen_columns;
    }
    else {
        column = getPageEndCell(page - 1).column + 1;
    }

    int row = frozen_rows;
    if (rp == 1) {
        row = frozen_rows;
    }
    else if (cp == 1) {
        row = getPageEndCell(page - 1).row + 1;
    }
    else {
        row = getPageEndCell(column_pages * (rp - 1) - 1).row + 1;
    }
    return(new JCCellPosition(row, column));
}

/**
 * Private method returns the last cell on a page.
 * @return
 * @param page
 * @throws IllegalSizeException
 */
private JCCellPosition getPageEndCell(int page) throws IllegalSizeException
{
    int current;
    int column_pages = getNumColumnPages();
    int row_pages = getNumRowPages();

    int rp = page / column_pages + 1;
    int cp = page % column_pages + 1;

    CellLayoutModel clm = getCellLayout();

    int column = frozen_columns;
    if (cp == column_pages) {
        column = getNumColumns() - 1;
    }
    else {
        int start_column = frozen_columns;
        if (cp > 1) {
            start_column = getPageEndCell(page - 1).column + 1;
        }
        int xover_width = (table_width - getRowLabelWidth() -
                           getRowLabelOffset() - getFrozenColumnWidth() -
                           4 * getFrameBorderWidth());
        current = 0;
        for (int c = start_column; c < getNumColumns(); c++) {
            current += clm.getWidth(c);
            if (current <= xover_width) {
                column = c;
            }
        }
    }

    int row = frozen_rows;
    if (rp == row_pages) {
        row = getNumRows() - 1;
    }
    else {
        int start_row = frozen_rows;
        if (rp > 1 && cp == 1) {
            start_row = getPageEndCell(page - 1).row + 1;
        }
        else if (rp > 1 && cp != 1) {
            start_row = getPageEndCell(column_pages * (rp - 1) - 1).row + 1;
        }
        int xover_height = (table_height - getColumnLabelHeight() -
                            getColumnLabelOffset() - getFrozenRowHeight() -
                            4 * getFrameBorderWidth());
        current = 0;
        for (int r = start_row; r < getNumRows(); r++) {
            current += clm.getHeight(r);
            if (current <= xover_height) {
                row = r;
            }
        }
    }

    return(new JCCellPosition(row, column));
}

/**
 * Returns the row label width.
 * @return row label width
 */
private int getRowLabelWidth()
{
    if (row_label_display) {
        return(getCellLayout().getWidth(JCTableEnum.LABEL));
    }
    return(0);
}

/**
 * Returns the start position of a row.
 * @return start position of a row
 * @param row
 */
private int getRowPosition(int row)
{
    CellLayoutModel clm = getCellLayout();
    return(clm.getRowPosition(row));
}

/**
 * Returns the top/left corner of the table on the page.
 * @return top/left corner of the table on the page
 */
private Point getTableOffset()
{
    int unit_conversion = 1;
    if (margin_unit == MARGIN_IN_INCHES) {
        unit_conversion = page_resolution;
    }
    return(new Point(unit_conversion * margin.left,
                     unit_conversion * margin.top));
}

/**
 * Returns the dimensions of the table required to print the range.
 * @return dimensions of the table
 * @param cr
 */
private Dimension getTablePageDimension(JCCellRange cr)
{
    Rectangle bounds = getCellRangeBounds(cr);
    return(new Dimension(getRowLabelWidth() + getRowLabelOffset() +
                         4 * getFrameBorderWidth() + getFrozenColumnWidth() +
                         bounds.width,
                         getColumnLabelHeight() + getColumnLabelOffset() +
                         4 * getFrameBorderWidth() + getFrozenRowHeight() +
                         bounds.height));
}

/**
 * Private method splits spans that cross pages.
 * @param cr
 */
private void splitSpans(JCCellRange cr)
{
    if (spanHandler != null) {
        Vector new_spans = new Vector();

        JCCellRange cr1, cr2;

        Collection col = spanHandler.getSpannedRanges();
        if (col != null) {
            Iterator it = col.iterator();
            while (it.hasNext()) {
                cr1 = (JCCellRange)it.next();
                cr2 = new JCCellRange(cr1.start_row, cr1.start_column,
                                      cr1.end_row, cr1.end_column);
                // span starts above the print range
                if (cr1.start_row < cr.start_row && cr1.end_row >= cr.end_row) {
                    cr2.reshape(cr2.start_row, cr2.start_column,
                                cr.start_row - 1, cr2.end_column);
                }
                // span extends below the print range
                if (cr1.start_row <= cr.end_row && cr1.end_row > cr.end_row) {
                    cr2.reshape(cr2.start_row, cr2.start_column,
                                cr.end_row, cr2.end_column);
                }
                // span starts left of the print range
                if (cr1.start_column < cr.start_column &&
                    cr1.end_column >= cr.start_column)
                {
                    cr2.reshape(cr2.start_row, cr.start_column,
                                cr2.end_row, cr2.end_column);
                }
                // span extends right of the print range
                if (cr1.start_column <= cr.end_column &&
                    cr1.end_column > cr.end_column)
                {
                    cr2.reshape(cr2.start_row, cr2.start_column,
                                cr2.end_row, cr.end_column);
                }

                new_spans.addElement(cr2);
            }
        }

        try {
            spanHandler.setSpannedRanges(new_spans);
        }
        catch (Exception e) {}
    }
}

/**
 * Notifies print listeners that the page header is being printed.
 * @param gc
 * @param page
 */
private void doPrintPageHeader(Graphics gc, int page)
{
    if (printListeners != null) {
        Graphics gc_header = gc.create();

        gc_header.setColor(Color.black);
        gc_header.setFont(getFont());

        int unit_conversion =
            (margin_unit == MARGIN_IN_INCHES) ? page_resolution : 1;
        gc_header.setClip(unit_conversion * margin.left,
                          0,
                          table_width,
                          unit_conversion * margin.top);
        gc_header.translate(unit_conversion * margin.left, 0);

        JCPrintEvent ev = new JCPrintEvent(this, gc_header, page + 1,
                                           getNumPages(),
                                           JCPrintEvent.PRINT_HEADER, this);
        fireJCPrintEvent(ev);
        gc_header.dispose();
    }
}

/**
 * Notifies print listeners that the body of the page has been printed.
 * @param gc
 * @param page
 */
private void doPrintBody(Graphics gc, int page)
{
    if (printListeners != null) {
        Graphics gc_body = gc.create();

        gc_body.setColor(Color.black);
        gc_body.setFont(getFont());

        JCPrintEvent ev = new JCPrintEvent(this, gc_body, page + 1,
                                           getNumPages(),
                                           JCPrintEvent.PRINT_BODY, this);
        fireJCPrintEvent(ev);
        gc_body.dispose();
    }
}

/**
 * Notifies print listeners that the page footer is being printed.
 * @param gc
 * @param page
 */
private void doPrintPageFooter(Graphics gc, int page)
{
    Graphics gc_footer = gc.create();

    gc_footer.setColor(Color.black);
    gc_footer.setFont(getFont());

    int unit_conversion =
        (margin_unit == MARGIN_IN_INCHES) ? page_resolution : 1;
    gc_footer.setClip(unit_conversion * margin.left,
                      page_height - unit_conversion * margin.bottom,
                      table_width,
                      unit_conversion * margin.bottom);
    gc_footer.translate(unit_conversion * margin.left,
                        page_height - unit_conversion * margin.bottom);

    if (printListeners != null) {
        JCPrintEvent ev = new JCPrintEvent(this, gc_footer, page + 1,
                                           getNumPages(),
                                           JCPrintEvent.PRINT_FOOTER, this);
        fireJCPrintEvent(ev);
    }
    else {
        printPageFooter(gc, page + 1);
    }

    gc_footer.dispose();
}

/**
 * Prints default footer.
 * @param gc
 * @param page
 */
private void printPageFooter(Graphics gc, int page)
{
    Graphics gc_footer = gc.create();

    int unit_conversion =
        (margin_unit == MARGIN_IN_INCHES) ? page_resolution : 1;

    gc_footer.setClip(unit_conversion * margin.left,
                      page_height - unit_conversion * margin.bottom,
                      table_width,
                      unit_conversion * margin.bottom);
    gc_footer.translate(unit_conversion * margin.left,
                        page_height - unit_conversion * margin.bottom);

    Rectangle r = gc_footer.getClipBounds();

    gc_footer.setColor(Color.black);
    gc_footer.setFont(getFont());
    String pageText = LocaleBundle.string(LocaleBundle.PRINT_PAGE_FIRST) +
        page + LocaleBundle.string(LocaleBundle.PRINT_PAGE_SECOND) +
        getNumPages();
    gc_footer.drawString(pageText, 0, r.height / 2);
}

/**
 * Convenience method for firing <code>JCPrintEvents</code>.
 * @param event
 */
protected void fireJCPrintEvent(JCPrintEvent event)
{
    Enumeration e = JCListenerList.elements(printListeners);
    while (e.hasMoreElements()) {
        JCPrintListener listener = (JCPrintListener)e.nextElement();
        switch (event.getType()) {
        case JCPrintEvent.PRINT_HEADER:
            listener.printPageHeader(event);
            break;
        case JCPrintEvent.PRINT_BODY:
            listener.printPageBody(event);
            break;
        case JCPrintEvent.PRINT_FOOTER:
            listener.printPageFooter(event);
            break;
        }
    }
}

/**
 * Adds the specified listener to receive instances
 * of <CODE>JCPrintEvent</CODE> objects.
 * @see JCPrintEvent
 * @param l
 */
public void addPrintListener(JCPrintListener l)
{
    printListeners = JCListenerList.add(printListeners, l);
}

/**
 * Removes the specified listener so it no longer receives
 * <CODE>JCPrintEvent</CODE> objects.
 * @see #addPrintListener
 * @param l
 */
public void removePrintListener(JCPrintListener l)
{
    printListeners = JCListenerList.remove(printListeners, l);
}


class IllegalSizeException extends IllegalStateException {

    protected IllegalSizeException(String s)
    {
        super(s);
    }

} // IllegalSizeException

}
