/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCRowColumnSeries.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

/**
 * <CODE>RowColumnSeriesModel</CODE> defines the interface required by classes
 * that can  store row or column based properties (for example, row height and
 * column width).
 */
public class JCRowColumnSeries implements RowColumnSeriesModel, Serializable {

protected Hashtable            values;

private static final Integer   ALL_POSITION = new Integer(JCTableEnum.ALL);
private static final Integer   LABEL_POSITION = new Integer(JCTableEnum.LABEL);
private static final Integer   ALLCELLS_POSITION = new Integer(JCTableEnum.ALLCELLS);

/**
 * Creates an empty series with no default.
 */
public JCRowColumnSeries()
{
    values = new Hashtable();
}

/**
 * Creates a new series with the specified default.
 * @param value
 */
public JCRowColumnSeries(Object value)
{
    values = new Hashtable();
    setDefault(value);
}

/**
 * Gets the default member value. The default value is always equivalent to the
 * <code>ALL</code> position.
 * @return
 */
public Object getDefault()
{
    return values.get(ALL_POSITION);
}

/**
 * Creates or sets the default member to the specified value. The default value
 * is always equivalent to the <code>ALL</code> position.
 * @param value
 */
public void setDefault(Object value)
{
    values.put(ALL_POSITION, value);
}

/**
 * Returns the value at the specified position.
 * @return null if the series is empty
 * @param position
 */
public Object getValue(int position)
{
    Object value;

    // case: position is in the cell range. Change search to ALLCELLS
    // if no value is found
    if (position > -1) {
        value = values.get(new Integer(position));
        if (value == null) {
            position = JCTableEnum.ALLCELLS;
        }
        else {
            return value;
        }
    }
    // case: position is a label. Change search to ALL if no value is found
    else if (position == JCTableEnum.LABEL) {
        value = values.get(LABEL_POSITION);
        if (value == null) {
            position = JCTableEnum.ALL;
        }
        else {
            return value;
        }
    }
    if (position == JCTableEnum.ALLCELLS) {
        value = values.get(ALLCELLS_POSITION);
        if (value == null) {
            position = JCTableEnum.ALL;
        }
        else {
            return value;
        }
    }
    if (position == JCTableEnum.ALL) {
        return getDefault();
    }

    return null;
}

/**
 * Sets the specified cell to the value.
 * @return <code>true</code> if the series was changed after adding the value
 * @param position
 * @param value
 */
public boolean setValue(int position, Object value)
{
    // handle all
    if (position == JCTableEnum.ALL) {
        values = new Hashtable();
        values.put(ALL_POSITION, value);
    }
    // handle label
    else if (position == JCTableEnum.LABEL) {
        values.put(LABEL_POSITION, value);
    }
    // handle all cells
    else if (position == JCTableEnum.ALLCELLS) {
        // save the default and label values
        Object defaultValue = values.get(ALL_POSITION);
        Object labelValue = values.get(LABEL_POSITION);

        values = new Hashtable();
        values.put(ALL_POSITION, defaultValue);
        if (labelValue != null) {
            values.put(LABEL_POSITION, labelValue);
        }
        values.put(ALLCELLS_POSITION, value);
    }
    // handle a single position
    else {
        values.put(new Integer(position), value);
    }
    return(true);
}

/**
 * Sets a range of cells to the value.
 * @return <code>true</code> if the series was changed after adding the value
 * @param end_position
 * @param start_position
 * @param value
 */
public boolean setValue(int start_position, int end_position, Object value)
{
    for (int p = start_position; p <= end_position; p++) {
        setValue(p, value);
    }
    return(true);
}

/**
 * Returns an iterator for an ordered set of series values. The iterated
 * objects are <CODE>PositionCellSize</CODE> objects.
 * @return
 */
public List getSeriesValues()
{
    Vector v = new Vector();
    Enumeration e = values.keys();
    while (e.hasMoreElements()) {
        Integer position = (Integer)e.nextElement();
        Object value = values.get(position);

        v.addElement(new PositionCellSize(position.intValue(),
                                          (CellSize)value));
    }
    return(v);
}

/**
 * Adjusts the series after moving.
 * @param destination
 * @param num
 * @param start
 */
public void move(int start, int num, int destination)
{
    Hashtable new_values = new Hashtable();
    Enumeration e = values.keys();
    while (e.hasMoreElements()) {
        Integer position = (Integer)e.nextElement();
        Object value = values.get(position);

        int pos = position.intValue();
        int new_pos = adjust(pos, start, num, destination);

        if (pos == new_pos) {
            new_values.put(position, value);
        }
        else {
            new_values.put(new Integer(new_pos), value);
        }
    }
    values = new_values;
}

private int adjust(int value, int source, int amount, int destination)
{
    if (source > destination) {
        if (value >= destination && value < source) {
            value += amount;
        }
        else if (value >= source && value < source + amount) {
            value -= source - destination;
        }
    }
    else {
        if (value >= source + amount && value < destination) {
            value -= amount;
        }
        else if (value >= source && value < source + amount) {
            value += destination - source - amount;
        }
    }
    return(value);
}

/**
 * Reorders after the mapping changes.
 * @param new_positions
 */
public void remap(int new_positions[])
{
    Hashtable new_values = new Hashtable();
    Enumeration e = values.keys();
    while (e.hasMoreElements()) {
        Integer position = (Integer)e.nextElement();
        Object value = values.get(position);

        int pos = position.intValue();

        if (pos >= 0 && pos < new_positions.length) {
            position = new Integer(new_positions[pos]);
        }
        new_values.put(position, value);
    }
    values = new_values;
}

/**
 * Adjusts the series after adding or deleting.
 * @param start the start position
 * @param num the number of rows
 * @param direction use <CODE>+1</CODE> for insert and <CODE>-1</CODE>
 * for delete
 */
public void shift(int start, int num, int direction)
{
    Hashtable new_values = new Hashtable();
    Enumeration e = values.keys();
    while (e.hasMoreElements()) {
        Integer position = (Integer)e.nextElement();
        Object value = values.get(position);

        int pos = position.intValue();

        if (direction > 0) {
            if (pos >= start) {
                position = new Integer(pos + num);
            }
            new_values.put(position, value);
        }
        else {
            if (pos >= start + num) {
                position = new Integer(pos - num);
                new_values.put(position, value);
            }
            else if (pos < start) {
                new_values.put(position, value);
            }
        }
    }
    values = new_values;
}

/**
 * Adjusts the series based on two entries swap positions.
 * @param position1
 * @param position2
 */
public void swap(int position1, int position2)
{
    Object p1 = getValue(position1);
    Object p2 = getValue(position2);
    setValue(position2, p1);
    setValue(position1, p2);
}
	
}
