/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCSelection.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

/**
 * The <CODE>SelectionModel</CODE> interface represents the current cell
 * selection state in <CODE>Table</CODE>. This implementation supports all
 * selection policies. If the policy is changed, the current selection is
 * cleared.
 */
public class JCSelection implements SelectionModel, Serializable {

protected JCTable           table;
protected Vector            selected_cells = null;
protected int               policy = SelectionModel.SELECT_NONE;

protected int               last_selected_row = (-1);
protected int               last_selected_column = (-1);

public JCSelection() {
}

public JCSelection(JCTable table)
{
    this.table = table;
}

/**
 * Returns the Table controlled by the <code>SelectionModel</code>.
 * @return Table controlled by the <code>SelectionModel</code>
 */
public JCTable getTable()
{
    return table;
}

/**
 * Sets the Table controlled by the <code>SelectionModel</code>.
 * @param table
 */
public void setTable(JCTable table)
{
    this.table = table;
}

/**
 * Returns the policy of selection supported by the model.
 * Valid values are <CODE>SELECT_NONE</CODE>, <CODE>SELECT_SINGLE</CODE>,
 * <CODE>SELECT_RANGE</CODE>, and <CODE>SELECT_MULTIRANGE</CODE>.
 * @see #setSelectionPolicy
 * @return policy of selection supported by the model; one of
 * <CODE>SELECT_NONE</CODE>, <CODE>SELECT_SINGLE</CODE>,
 * <CODE>SELECT_RANGE</CODE>, or <CODE>SELECT_MULTIRANGE</CODE>
 */
public int getSelectionPolicy()
{
    return policy;
}

/**
 * Sets the policy of selection supported by the model. Valid policies
 * are <CODE>SELECT_NONE</CODE>, <CODE>SELECT_SINGLE</CODE>,
 * <CODE>SELECT_RANGE</CODE>, and <CODE>SELECT_MULTIRANGE</CODE>.
 * @param value
 */
public void setSelectionPolicy(int value)
{
    if (selected_cells != null) {
        clearSelection();
    }
    this.policy = value;
}


/**
 * Adds a range of columns to the selection.
 * @return <code>true</code> if the selection is accepted
 * @param start_column
 * @param end_column
 */
public boolean addColumnSelection(int start_column, int end_column)
{
    return addSelection(JCTableEnum.LABEL, start_column, JCTableEnum.MAXINT,
                        end_column);
}

/**
 * Adds a range of rows to the selection.
 * @return <code>true</code> if the selection is accepted
 * @param end_row
 * @param start_row
 */
public boolean addRowSelection(int start_row, int end_row)
{
    return addSelection(start_row, JCTableEnum.LABEL, end_row,
                        JCTableEnum.MAXINT);
}

/**
 * Adds a range of cells to the selection.
 * @return <code>true</code> if the selection is accepted
 * @param end_column
 * @param end_row
 * @param start_column
 * @param start_row
 */
public boolean addSelection(int start_row, int start_column, int end_row, int
    end_column)
{
    if (start_row == end_row && start_row == JCTableEnum.LABEL) {
        end_row = JCTableEnum.MAXINT;
    }

    if (start_column == end_column && start_column == JCTableEnum.LABEL) {
        end_column = JCTableEnum.MAXINT;
    }

    return addSelection(new JCCellRange(start_row, start_column, end_row,
                                        end_column));
}

/**
 * Adds a <code>JCCellRange</code> to the selection.
 * @return <code>true</code> if the selection is accepted
 * @param cr
 */
public boolean addSelection(JCCellRange cr)
{
    if (selected_cells == null) {
        return false;
    }

    // validate the new range against the selection policy
    if (policy != SelectionModel.SELECT_MULTIRANGE) {
        return false;
    }

    JCSelectEvent event;
    if (table.selectListeners != null) {
        // notifies all listeners of a change to the selection.
        event = new JCSelectEvent(table, JCSelectEvent.BEFORE_SELECTION,
                                  cr.start_row, cr.start_column,
                                  cr.end_row, cr.end_column, JCSelectEvent.ADD);
        table.fireJCSelectEvent(event);
        // aborts the selection change if a listener rejects it.
        if (event.isCancelled()) {
            return false;
        }
    }
    selected_cells.addElement(cr);

    if (table.selectListeners != null) {
        // notifies all listeners that the selection was changed.
        event = new JCSelectEvent(table, JCSelectEvent.SELECTION,
                                  cr.start_row, cr.start_column,
                                  cr.end_row, cr.end_column, JCSelectEvent.ADD);
        table.fireJCSelectEvent(event);
        // abort the selection (and back out the changes) if a listener
        // rejects it
        if (event.isCancelled()) {
            selected_cells.removeElement(cr);
            return false;
        }
    }

    table.getCellAreaHandler().getCellAreaRenderer().repaintRange(cr);

    return true;
}

/**
 * Ensures that the last range does not duplicate another.
 */
public void cleanup()
{
    int length;

    if (selected_cells == null) {
        return;
    }

    if ((length = selected_cells.size()) < 2) {
        return;
    }

    JCCellRange last_range = (JCCellRange)selected_cells.lastElement();

    for (int j = length - 2; j >= 0; j--) {
        if (last_range.equals(selected_cells.elementAt(j))) {
            selected_cells.removeElementAt(j);
        }
    }
}

/**
 * Clears all selected cells and generates required <CODE>JCSelectEvents</CODE>.
 */
public void clearSelection()
{
    clearSelection(true);
}

/**
 * Clears all selected cells. If <CODE>withAllEvents</CODE> is
 * <code>true</code>, before and after events are generated for the deselected
 * cell ranges.
 * @param withAllEvents
 */
public void clearSelection(boolean withAllEvents)
{
    if (selected_cells == null) {
        return;
    }

    Vector selection = new Vector();
    Vector current_selected = (Vector)selected_cells.clone();

    JCCellRange cr;
    JCSelectEvent event = null;

    for (int i = 0; i < current_selected.size(); i++) {
        cr = (JCCellRange)current_selected.elementAt(i);

        if (table.selectListeners != null) {
            // notifies all listeners that a cell range will be deselected.
            event = new JCSelectEvent(table, JCSelectEvent.BEFORE_SELECTION,
                                      cr.start_row, cr.start_column,
                                      cr.end_row, cr.end_column,
                                      JCSelectEvent.DESELECT);
            if (withAllEvents) {
                table.fireJCSelectEvent(event);
                // keep the range if a listeners rejects its deselection.
                if (event.isCancelled()) {
                    selection.addElement(cr);
                    // no more events for this range is BEFORE is rejected
                    continue;
                }
            }
            event.type = JCSelectEvent.SELECTION;
            table.fireJCSelectEvent(event);
            // abort the selection (and back out the changes) if a listener
            // rejects it
            if (event.isCancelled()) {
                selection.addElement(cr);
            }
        }
    }

    if (selection.size() == 0) {
        selected_cells = null;
        // repaint each range that was previously selected
        for (int i = 0; i < current_selected.size(); i++) {
            cr = (JCCellRange)current_selected.elementAt(i);
            CellAreaRenderer car =
                table.getCellAreaHandler().getCellAreaRenderer();
            // order of range appears to matter in repaintRange()
            int sr = Math.min(cr.start_row, cr.end_row);
            int er = Math.max(cr.start_row, cr.end_row);
            int sc = Math.min(cr.start_column, cr.end_column);
            int ec = Math.max(cr.start_column, cr.end_column);
            car.repaintRange(sr, sc, er, ec);
        }
    }
    else {
        selected_cells = (Vector)selection.clone();
        table.repaint();
    }

    if (withAllEvents && event != null) {
        event.type = JCSelectEvent.AFTER_SELECTION;
        table.fireJCSelectEvent(event);
    }

    selection = null;
    current_selected = null;

    last_selected_row = (-1);
    last_selected_column = (-1);
}

/**
 * Extends the current selection to the given row and column.
 * @return <code>true</code> if the selection is accepted
 * @param row
 * @param column
 */
public boolean extendSelection(int row, int column)
{
    if (selected_cells == null || selected_cells.size() == 0) {
        return false;
    }

    // validate the new range against the selection policy
    switch (policy) {
    case SelectionModel.SELECT_NONE:
    case SelectionModel.SELECT_SINGLE:
        return false;
    }

    JCCellRange cr = (JCCellRange)selected_cells.lastElement();

    if (isLabel(cr.start_row, cr.start_column) && !isLabel(row, column)) {
        if (cr.start_row == JCTableEnum.LABEL) {
            row = JCTableEnum.LABEL;
        }
        else if (cr.start_column == JCTableEnum.LABEL) {
            column = JCTableEnum.LABEL;
        }
        else {
            return false;
        }
    }

    JCSelectEvent event;
    if (table.selectListeners != null) {
        // notifies all listeners that a cell range will be extended.
        event = new JCSelectEvent(table, JCSelectEvent.SELECTION,
                                  cr.start_row, cr.start_column, row, column,
                                  JCSelectEvent.EXTEND);
        table.fireJCSelectEvent(event);
        // abort the extension if rejected by a listener
        if (event.isCancelled()) {
            return false;
        }
    }

    int sr = cr.start_row;
    int sc = cr.start_column;
    int er = cr.end_row;
    int ec = cr.end_column;

    if (er < sr) {
        int temp = sr;
        sr = er;
        er = temp;
    }
    if (ec < sc) {
        int temp = sc;
        sc = ec;
        ec = temp;
    }

    // adjust row if selecting a column label
    if (row == JCTableEnum.LABEL) {
        cr.end_column = column;
    }
    // adjust column if selecting a row label
    else if (column == JCTableEnum.LABEL) {
        cr.end_row = row;
    }
    else {
        cr.end_row = row;
        cr.end_column = column;
    }

    sr = Math.min(row, sr);
    er = Math.max(row, er);
    sc = Math.min(column, sc);
    ec = Math.max(column, ec);

    table.getCellAreaHandler().getCellAreaRenderer().repaintRange(sr, sc,
                                                                  er, ec);

    return true;
}

/**
 * Returns a Vector containing all of the selected regions as
 * <code>JCCellRange</code> objects.
 * @return Vector containing all of the selected regions as
 * <code>JCCellRange</code> objects
 */
public Collection getSelectedCells()
{
    return selected_cells;
}

/**
 * Returns <code>true</code> if the given row/column is selected.
 * @return <code>true</code> if the selection is accepted
 * @param column
 * @param row
 */
public boolean isSelected(int row, int column)
{
    if (selected_cells == null) {
        return false;
    }

    JCCellRange cr;
    for (int i = 0; i < selected_cells.size(); i++) {
        cr = (JCCellRange)selected_cells.elementAt(i);
        if (cr.inside(row, column)) {
            return true;
        }
    }
    return false;
}

/**
 * Selects all cells.
 * @return <code>true</code> if the selection is accepted
 */
public boolean selectAll()
{
    return setSelection(JCTableEnum.LABEL, JCTableEnum.LABEL,
                        JCTableEnum.MAXINT, JCTableEnum.MAXINT);
}

/**
 * Sets a range of columns as the current selection.
 * @return <code>true</code> if the selection is accepted
 * @param end_column
 * @param start_column
 */
public boolean setColumnSelection(int start_column, int end_column)
{
    return setSelection(JCTableEnum.LABEL, start_column, JCTableEnum.MAXINT,
                        end_column);
}

/**
 * Sets a range of rows as the current selection.
 * @return <code>true</code> if the selection is accepted
 * @param end_row
 * @param start_row
 */
public boolean setRowSelection(int start_row, int end_row)
{
    return setSelection(start_row, JCTableEnum.LABEL, end_row,
                        JCTableEnum.MAXINT);
}

/**
 * Sets a range of cells as the current selection.
 * @return <code>true</code> if the selection is accepted
 * @param end_column
 * @param end_row
 * @param start_column
 * @param start_row
 */
public boolean setSelection(int start_row, int start_column, int end_row,
                            int end_column)
{
    if (start_row == end_row && start_row == JCTableEnum.LABEL) {
        end_row = JCTableEnum.MAXINT;
    }

    if (start_column == end_column && start_column == JCTableEnum.LABEL) {
        end_column = JCTableEnum.MAXINT;
    }

    return setSelection(new JCCellRange(start_row, start_column, end_row,
                                        end_column));
}

/**
 * Sets the selection to a Vector of <code>JCCellRange</code>.
 * @return <code>true</code> if the selection is accepted
 * @param ranges
 */
public boolean setSelection(Collection ranges)
{
    if (selected_cells != null) {
        clearSelection(false);
    }

    // assume that all ranges will be accepted and check if
    // any reset the status to false
    boolean status = true;

    JCCellRange cr;

    Iterator it = ranges.iterator();
    boolean first = true;
    while (it.hasNext()) {
        cr = (JCCellRange)it.next();
        // for the first range, do a set
        if (first) {
            if (!setSelection(cr)) {
                status = false;
            }
            first = false;
        }
        // for all other ranges, do an add
        else {
            if (!addSelection(cr)) {
                status = false;
            }
        }
    }
    if (!status) {
        clearWithNoNotification();
    }
    return status;
}

/**
 * Sets the selection to a <code>JCCellRange</code>.
 * @return <code>true</code> if the selection is accepted
 * @param cr
 */
public boolean setSelection(JCCellRange cr)
{
    // validate the new range against the selection policy
    switch (policy) {
    case SelectionModel.SELECT_NONE:
        return false;
    case SelectionModel.SELECT_SINGLE:
        // reject the selection if it is not a single cell
        if (!(cr.start_row == cr.end_row && cr.start_column == cr.end_column)) {
            return false;
        }
    }

    JCSelectEvent event;
    if (table.selectListeners != null) {
        // notifies all listeners that a new range will be selected.
        event = new JCSelectEvent(table, JCSelectEvent.BEFORE_SELECTION,
                                  cr.start_row, cr.start_column,
                                  cr.end_row, cr.end_column,
                                  JCSelectEvent.SELECT);
        table.fireJCSelectEvent(event);
        // abort the extension if rejected by a listener
        if (event.isCancelled()) {
            return false;
        }
    }

    // clear the current selected cells
    if (selected_cells != null) {
        clearSelection(false);

        // Abort the selection if clearSelection didn't complete because of
        // a JCSelectListener rejection
        if (selected_cells != null) {
            return false;
        }
    }

    selected_cells = new Vector();
    selected_cells.addElement(cr);

    if (table.selectListeners != null) {
        // notifies all listeners that the selection was changed.
        event = new JCSelectEvent(table, JCSelectEvent.SELECTION,
                                  cr.start_row, cr.start_column,
                                  cr.end_row, cr.end_column,
                                  JCSelectEvent.SELECT);
        table.fireJCSelectEvent(event);
        // abort the selection (and back out the changes) if a listener
        // rejects it
        if (event.isCancelled()) {
            selected_cells.removeElement(cr);
            return false;
        }
    }

    table.repaintSelected();

    return true;
}

/**
 * Adds, extends, deselects, or ends a selection based on a row or column. This
 * method is called by the mouse and key process handler and should not be
 * used for programmatic manipulation of selected regions.
 *
 * @return <code>true</code> if the model supports the selection
 * @param column
 * @param row
 * @param state
 */
public boolean select(int state, int row, int column)
{
    // check if selection is allowed
    if (policy == SelectionModel.SELECT_NONE) {
        return false;
    }

    // ensure that the row and column is valid
    if (row == JCTableEnum.NOVALUE || column == JCTableEnum.NOVALUE) {
        // if we return false, interactive selection is terminated, so
        // we ignore the values but tell table that the selection process
        // can still continue
        return true;
    }

    // Check for an attempt to reselect the same cell as previously selected.
    if (row == last_selected_row &&
        column == last_selected_column &&
        state != END)
    {
        return true;
    }
    else if (row == JCTableEnum.NOVALUE || column == JCTableEnum.NOVALUE) {
        return true;
    }

    boolean accepted = false;

    int endRow = row;
    int endColumn = column;

    // Treat a span as a single cell so users can click in any of the
    // non-control cells to do the selection.
    JCCellRange span = table.getSpanHandler().getSpannedRange(row, column);
    if (span != null) {
        row = span.start_row;
        column = span.start_column;
        endRow = span.end_row;
        endColumn = span.end_column;
    }

    if (state == SelectionModel.END) {
        // reset last_selected upon END -- otherwise we won't be
        // able to reselect this last cell
        last_selected_row = JCTableEnum.NOVALUE;
        last_selected_column = JCTableEnum.NOVALUE;
    }
    else {
        // keep last_selected row and column to filter out unnecessary
        // duplicates
        last_selected_row = endRow;
        last_selected_column = endColumn;
    }

    switch (state) {
    case SelectionModel.SELECT:
        accepted = setSelection(row, column, endRow, endColumn);
        break;
    case SelectionModel.EXTEND:
        accepted = extendSelection(row, column);
        break;
    case SelectionModel.ADD:
        accepted = addSelection(row, column, endRow, endColumn);
        break;
    case SelectionModel.END:
        cleanup();

        if (table.selectListeners != null && selected_cells != null) {
            // notifies all listeners that a new range will be selected.
            JCCellRange cr = (JCCellRange)selected_cells.lastElement();
            JCSelectEvent event =
                new JCSelectEvent(table, JCSelectEvent.AFTER_SELECTION,
                                  cr.start_row, cr.start_column,
                                  row, column, JCSelectEvent.END);
            table.fireJCSelectEvent(event);
        }

        accepted = true;
        break;
    }

    if (!accepted) {
        last_selected_row = JCTableEnum.NOVALUE;
        last_selected_column = JCTableEnum.NOVALUE;
    }
    return accepted;
}

/**
 * Clears the current selection without notification. This is used when
 * processing a series of selections that must all succeed. If one fails,
 * rollback won't generate deselect events for changes already processed.
 */
protected void clearWithNoNotification()
{
    selected_cells = null;
}

/**
 * Returns <code>true</code> if the given row/column is a label.
 * @return <code>true</code> if the given row/column is a label
 * @param column
 * @param row
 */
protected boolean isLabel(int row, int column)
{
    return (row == JCTableEnum.LABEL || column == JCTableEnum.LABEL);
}

/**
 * Adjusts the selection after moving columns.
 * @param start_column the starting position
 * @param num_columns the number of columns
 * @param destination the column before which the source columns are
 * moved (use <CODE>JCTableEnum.MAXINT</CODE> to move the columns to
 * the end)
 */
public void moveColumns(int start_column, int num_columns, int destination) {}

/**
 * Adjusts the selection after moving rows.
 * @param start_row the starting position
 * @param num_rows the number of rows
 * @param destination the row before which the source rows are moved (use
 * <CODE>JCTableEnum.MAXINT</CODE> to move the rows to the end)
 */
public void moveRows(int start_row, int num_rows, int destination) {}

/**
 * Reorders selected columns after the column mapping changes.
 * @param new_positions
 */
public void remapColumns(int new_positions[]) {}

/**
 * Reorders selected rows after the row mapping changes.
 * @param new_positions
 */
public void remapRows(int new_positions[]) {}

/**
 * Adjusts the selection after adding or deleting columns.
 * @param start_column the start position
 * @param num_columns the number of columns
 * @param direction use <CODE>+1</CODE> for insert and <CODE>-1</CODE>
 * for delete
 */
public void shiftColumn(int start_column, int num_columns, int direction) {}

/**
 * Adjusts the selection after adding or deleting rows.
 * @param start_row the start position
 * @param num_rows the number of rows
 * @param direction use <CODE>+1</CODE> for insert and <CODE>-1</CODE>
 * for delete
 */
public void shiftRow(int start_row, int num_rows, int direction) {}

/**
 * Adjusts the selection based on two columns swapping position.
 * @param column1
 * @param column2
 */
public void swapColumns(int column1, int column2) {}

/**
 * Adjusts the selection based on two rows swapping position.
 * @param row1
 * @param row2
 */
public void swapRows(int row1, int row2) {}

}
