/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTableCellInfo.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.awt.*;
import java.io.Serializable;

/**
 * <CODE>JCTableCellInfo</CODE> is an object passed to <CODE>CellRenderer</CODE>
 * and <CODE>CellEditor</CODE> objects to supply information about the
 * type of cell being displayed or edited.
 */
public class JCTableCellInfo implements TableCellInfoModel, Serializable {

protected JCTable             table;
protected int                 row;
protected int                 column;
protected CellStyleModel      style;
protected Object              data;
	
protected Insets              margin_insets;
protected Insets              border_insets;

private Rectangle             drawing_area = new Rectangle();

public JCTableCellInfo() {}

/**
 * Initializes the <CODE>JCTableCellInfo</CODE> object with the cell
 * position and data value.
 * @param column
 * @param o
 * @param row
 * @param table
 */
public void initialize(JCTable table, Object o, int row, int column)
{
    initialize(table, o, row, column, null);
}

/**
 * Initializes the <CODE>JCTableCellInfo</CODE> object with the cell
 * position, data value, and <CODE>CellStyleModel</CODE>.
 * @param column
 * @param o
 * @param row
 * @param style
 * @param table
 */
public void initialize(JCTable table, Object o, int row, int column,
                       CellStyleModel style)
{
    this.table = table;
    this.row = row;
    this.column = column;

    if (style == null) {
        this.style = table.getCellStyle(row, column);
    }
    else {
        this.style = style;
    }

    data = o;
}

/**
 * Returns the <code>CellStyle</code> object that contains the property setting
 * for the cell.
 * @return <code>CellStyle</code> object
 */
public CellStyleModel getCellStyle()
{
    return style;
}

/**
 * Returns the background color of the cell being displayed or edited.
 * @return background color
 */
public Color getBackground()
{
    Color bg;
    int repeat = style.getRepeatBackground();
    if (repeat != JCTableEnum.REPEAT_NONE) {
        bg = getRepeatColor(row, column, repeat,
                            style.getRepeatBackgroundColors());
        if (bg != null) {
            return bg;
        }
    }
    return style.getBackground();
}

/**
 * Returns the foreground color of the cell being displayed or edited.
 * <CODE>Container</CODE> sets the
 * foreground color on the graphics object it passes to
 * <CODE>CellRenderer.draw()</CODE>.
 * @return foreground color
 */
public Color getForeground()
{
    Color fg;
    int repeat = style.getRepeatForeground();
    if (repeat != JCTableEnum.REPEAT_NONE) {
        fg = getRepeatColor(row, column, repeat,
                            style.getRepeatForegroundColors());
        if (fg != null) {
            return fg;
        }
    }
    return style.getForeground();
}

/**
 * Returns the background color of the cell being displayed or edited when
 * the cell is selected.
 * @return background color
 */
public Color getSelectedBackground()
{
    Color color = getSelectedColor(table.getSelectedBackgroundMode());
    return (color != null) ? color : getBackground();
}

/**
 * Returns the foreground color of the cell being displayed or edited when
 * the cell is selected.
 * @return foreground color
 */
public Color getSelectedForeground()
{
    Color color = getSelectedColor(table.getSelectedForegroundMode());
    return (color != null) ? color : getForeground();
}

/**
 * <CODE>SelectedBackgroundMode</CODE> and <CODE>SelectedForegroundMode</CODE>
 * determine how the selected color is obtained for a cell.
 * @return
 * @param mode
 */
protected Color getSelectedColor(int mode)
{
    if (mode == JCTableEnum.USE_SELECTED_BACKGROUND) {
        return table.getSelectedBackground();
    }
    else if (mode == JCTableEnum.USE_SELECTED_FOREGROUND) {
        return table.getSelectedForeground();
    }
    else if (mode == JCTableEnum.USE_CELL_BACKGROUND) {
        return getBackground();
    }
    else if (mode == JCTableEnum.USE_CELL_FOREGROUND) {
        return getForeground();
    }
    return null;
}

/**
 * Returns the font of the cell being displayed or edited.
 * <CODE>Container</CODE> sets the
 * font on the graphics object it passes to <CODE>CellRenderer.draw()</CODE>.
 * @return font of the cell being displayed or edited
 */
public Font getFont()
{
    return style.getFont();
}

/**
 * Returns the <CODE>FontMetrics</CODE> object associated with the font used in
 * the cell.
 * @return <code>FontMetrics</code> object
 */
public FontMetrics getFontMetrics()
{
    return table.getFontMetrics(style.getFont());
}

/**
 * Returns the horizontal alignment for the cell being displayed or edited.
 * Valid return values are <CODE>LEFT</CODE>, <CODE>CENTER</CODE>, and
 * <CODE>RIGHT</CODE>.
 * @return horizontal alignment; one of <code>LEFT</code>, <code>CENTER</code>,
 * or <code>RIGHT</code>
 */
public int getHorizontalAlignment()
{
    return style.getHorizontalAlignment();
}

/**
 * Returns the vertical alignment for the cell being displayed or edited.
 * Valid return values are <CODE>TOP</CODE>, <CODE>CENTER</CODE>, and
 * <CODE>BOTTOM</CODE>.
 * @return vertical alignment; one of <code>TOP</code>, <code>CENTER</code>,
 * or <code>BOTTOM</code>
 */
public int getVerticalAlignment()
{
    return style.getVerticalAlignment();
}

/**
 * Returns the <code>Insets</code> object representing the margins for the cell.
 * @return <code>Insets</code> object
 */
public Insets getMarginInsets()
{
    int mh = table.getMarginHeight();
    int mw = table.getMarginWidth();
    if (margin_insets == null ||
        (margin_insets.left != mw && margin_insets.top != mh))
    {
        margin_insets = new Insets(mh, mw, mh, mw);
    }
    return margin_insets;
}

/**
 * Returns the <code>Insets</code> object representing the borders for the cell.
 * @return <code>Insets</code> object
 */
public Insets getBorderInsets()
{
    int bw = table.getCellBorderWidth();
    if (border_insets == null || border_insets.left != bw) {
        border_insets = new Insets(bw, bw, bw, bw);
    }
    return border_insets;
}

/**
 * Returns the border style for the cell. The return result from this method is
 * meaningless for Table because of it's Border model. Other products that use
 * the Cell package may use this value.
 * @return border style
 */
public int getBorderStyle()
{
    return 0;
}

/**
 * Returns the dimensions of the drawing area for the cell being displayed
 * or edited. When the renderer's draw method is called, the gc is translated
 * to the upper left-hand corner of the drawing area for the cell.  As
 * a result, the (x,y) coordinates of the Rectangle are always (0,0).
 * @return size of the drawing area
 */
public Rectangle getDrawingArea()
{
    Rectangle r = table.getCellAreaHandler().getBounds(row, column,
                                                       drawing_area);

    Insets border_insets = getBorderInsets();
    Insets margin_insets = getMarginInsets();

    r.x = border_insets.left + margin_insets.left; // was +=
    r.y = border_insets.top + margin_insets.top; // was +=
    r.width -= border_insets.left + margin_insets.left +
               border_insets.right + margin_insets.right;
    r.height -= border_insets.top + margin_insets.top +
                border_insets.bottom + margin_insets.bottom;
    return r;
}

/**
 * Determines whether the cell is editable.
 * @return
 */
public boolean isEditable()
{
    return style.isEditable();
}

/**
 * Determines whether the cell is enabled.
 * @return
 */
public boolean isEnabled()
{
    return style.isTraversable();
}

/**
 * Determines whether the editor should select the contents before editing.
 * @return
 */
public boolean getSelectAll()
{
    return false;
}

/**
 * Determines whether the renderer should draw some indication if the
 * entire contents of the cell cannot be rendered within the given area.
 * @return
 */
public int getClipHints()
{
    return style.getClipHints();
}

/**
 * Returns the cell's associated data type. Three steps are taken to
 * obtain the data type. First, if the <CODE>CellStyleModel</CODE>
 * for the cell has a data type, which is returned. Second, the
 * <CODE>getColumnClass(int column)</CODE> method is called to
 * obtain the Class for the column. Finally, if the data type
 * is still null, the Class of the cell value is returned.
 * @return cell's associated data type
 */
public Class getDataType()
{
    // get the data type from the style
    Class cl = style.getDataType();
    if (cl != null) {
        return cl;
    }

    // get the ColumnClass from table
    cl = table.getColumnClass(column);
    if (cl != null) {
        return cl;
    }

    // get the Class based on the data
    if (data != null) {
        return data.getClass();
    }

    return null;
}

/**
 * Returns the table object associated with the cell.
 * @return table object associated with the cell
 */
public JCTable getTable()
{
    return table;
}

/**
 * Returns the row for the cell.
 * @return row for the cell
 */
public int getRow()
{
    return row;
}

/**
 * Returns the column for the cell.
 * @return column for the cell
 */
public int getColumn()
{
    return column;
}

/**
 * Returns the <code>CellBorderModel</code> for the cell.
 * @return <code>CellBorderModel</code> for the cell
 */
public CellBorderModel getCellBorder()
{
    return style.getCellBorder();
}

/**
 * Returns the <code>BorderSides</code> for the cell.
 * @return <code>BorderSides</code> for the cell
 */
public int getCellBorderSides()
{
    return style.getCellBorderSides();
}

/**
 * Gets the color of a cell from the repeated background or foreground list.
 * @return
 * @param column
 * @param list
 * @param repeat
 * @param row
 */
protected Color getRepeatColor(int row, int column, int repeat, Color[] list)
{
    if (list == null) {
        return(null);
    }
    int length = list.length;
    int mod_num = (repeat == JCTableEnum.REPEAT_ROW) ? row : column;
    int index = (mod_num < 0) ? (length - (((-1) * mod_num) % length)) % length
                              : mod_num % length;
    return list[index];
}

}
