/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTableDataEvent.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.util.EventObject;

/**
 * <code>TableDataEvent</code> objects describe changes to a
 * <code>TableDataModel</code> object.
 *
 * @see TableDataModel
 * @see JCTableDataListener
 */ 
public class JCTableDataEvent extends EventObject {

/** Indicates that a cell value has changed in the data source. */
public final static int CHANGE_VALUE		= 1;
/** Indicates that a row has changed one or more of its values in the data source. */
public final static int CHANGE_ROW			= 2;
/** Indicates that one or more rows have been added to the data source. */
public final static int ADD_ROW				= 3;
/** Indicates that one or more rows have been deleted from the data source. */
public final static int REMOVE_ROW			= 4;
/** Indicates that a column has changed one or more of its values in the data source. */
public final static int CHANGE_COLUMN		= 5;
/** Indicates that one or more columns have been added to the data source. */
public final static int ADD_COLUMN			= 6;
/** Indicates that one or more column have been deleted from the data source. */
public final static int REMOVE_COLUMN		= 7;
/** Indicates that a row label has changed in the data source. */
public final static int CHANGE_ROW_LABEL	= 8;
/** Indicates that a column label has changed in the data source. */
public final static int CHANGE_COLUMN_LABEL	= 9;
/** Indicates that one or more rows have moved in the data source. */
public final static int MOVE_ROW			= 10;
/** Indicates that one or more columns have moved in the data source. */
public final static int MOVE_COLUMN			= 11;
/** Indicates that the number of rows in the data source has changed. */
public final static int NUM_ROWS			= 12;
/** Indicates that the number of columns in the data source has changed. */
public final static int NUM_COLUMNS			= 13;
/** Indicates that table should reset all data values based on changes in the data source. */
public final static int RESET				= 14;

protected int row, column, num_affected, destination, command;

/** 
 * The event posted when a change is made to the data source.
 * @param source The data source
 * @param row the row number associated with changed data
 * @param column the column number associated with changed data
 * @param num_affected for <CODE>MOVE_ROW</CODE>, <CODE>MOVE_COLUMN</CODE>,
 * <CODE>REMOVE_ROW</CODE> and <CODE>REMOVE_COLUMN</CODE>, the number or rows
 * or columns affected
 * @param destination for <CODE>MOVE_ROW</CODE> and <CODE>MOVE_COLUMN</CODE>,
 * the destination row/column before which the insert occurs
 * @param command The type of change.  Possible values are:
*<TABLE BORDER="1">
*<TR><TD>TableDataEvent.CHANGE_VALUE</TD><TD>A single cell has changed</TD>
*<TR><TD>TableDataEvent.CHANGE_ROW</TD><TD>A single row has changed</TD>
*<TR><TD>TableDataEvent.ADD_ROW</TD><TD>One or more rows have been added</TD>
*<TR><TD>TableDataEvent.REMOVE_ROW</TD><TD>One or more rows have been deleted</TD>
*<TR><TD>TableDataEvent.CHANGE_COLUMN</TD><TD>A single column has changed</TD>
*<TR><TD>TableDataEvent.ADD_COLUMN</TD><TD>One or more columns have been added</TD>
*<TR><TD>TableDataEvent.REMOVE_COLUMN</TD><TD>One or more columns have been deleted</TD>
*<TR><TD>TableDataEvent.CHANGE_ROW_LABEL</TD><TD>A single row label has changed</TD>
*<TR><TD>TableDataEvent.CHANGE_COLUMN_LABEL</TD><TD>A single column label has changed</TD>
*<TR><TD>TableDataEvent.MOVE_ROW</TD><TD>One or more rows have been moved</TD>
*<TR><TD>TableDataEvent.MOVE_COLUMN</TD><TD>One or more columns have been moved</TD>
*<TR><TD>TableDataEvent.NUM_ROWS</TD><TD>The number of rows has changed</TD>
*<TR><TD>TableDataEvent.NUM_COLUMNS</TD><TD>The number of columns has changed</TD>
*<TR><TD>TableDataEvent.RESET</TD><TD>Recalculation and repaint of the table is required</TD>
*</TABLE>
 *
 * @see TableDataModel#addTableDataListener
 * @see JCTableDataAdapter
 * @see JCTableDataListener
 */
public JCTableDataEvent(Object source, int row, int column, int num_affected,
                        int destination, int command)
{
    super(source);
    this.row = row;
    this.column = column;
    this.num_affected = num_affected;
    this.destination = destination;
    this.command = command;
}

/**
 * Returns the row that was changed.
 * @return changed row
 */
public int getRow()
{
    return row;
}

/**
 * Returns the column that was changed.
 * @return changed column
 */
public int getColumn()
{
    return column;
}

/**
 * Returns the number of rows or columns affected.
 * @return number of rows or columns affected
 */
public int getNumAffected()
{
    return num_affected;
}

/**
 * For MOVE events, indicates the destination location.  Rows/columns are
 * inserted to the left (or above) the destination.
 * @return
 */
public int getDestination()
{
    return(destination);
}

/**
 * Returns the command that generated the event.  Valid commands include
 * <PRE> Commands:
 *TableDataEvent.CHANGE_VALUE
 *TableDataEvent.CHANGE_ROW
 *TableDataEvent.ADD_ROW
 *TableDataEvent.REMOVE_ROW
 *TableDataEvent.CHANGE_COLUMN
 *TableDataEvent.ADD_COLUMN
 *TableDataEvent.REMOVE_COLUMN
 *TableDataEvent.CHANGE_ROW_LABEL
 *TableDataEvent.CHANGE_COLUMN_LABEL
 *TableDataEvent.MOVE_ROW
 *TableDataEvent.MOVE_COLUMN
 *TableDataEvent.NUM_ROWS
 *TableDataEvent.NUM_COLUMNS
 *TableDataEvent.RESET
 * </PRE>
 * @return command that generated the event
 */
public int getCommand()
{
    return(command);
}

}
