/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTableUI.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.plaf.ColorUIResource;
import javax.swing.plaf.ComponentUI;
import java.awt.*;

/**
 * <CODE>JCTableUI</CODE> controls the pluggable look and feel (PLAF) changes
 * for <code>JCTable</code>. Affected properties are: label background,
 * foreground, font and border; cell background, foreground, font and border;
 * selection color; and focus indicator. Two special <CODE>CellStyleModel</CODE>
 * objects are modified for PLAF: <code>DefaultLabelStyle</code> and
 * <code>DefaultCellStyle</code>.
 * When modifying these two styles or basing another style on these,
 * all PLAF changes will continue for properties that weren't
 * modified. <CODE>JCTableUI</CODE> uses <code>UIResource</code> objects in
 * order to distinguish between PLAF-set properties (which change when the PLAF
 * changes) and user-set properties (which do not change).
 * @see JCTable#getDefaultLabelStyle
 * @see JCTable#getDefaultCellStyle
 */
public class JCTableUI extends ComponentUI {

/**
 * Installs the default settings for the current look and feel.
 * @param c
 */
public void installUI(JComponent c)
{
    c.setOpaque(false);
    LookAndFeel.installColorsAndFont(c, "Table.background",
                                     "Table.foreground",
                                     "Table.font");
    installDefaults(c);
}

/**
 * Changes the border, font, and colors (background, foreground, focus, and
 * selection) for table given the current look and feel.
 * @param c
 */
protected void installDefaults(JComponent c)
{
    JCTable table = (JCTable)c;

    // Get the current LookAndFeel
    String name = UIManager.getLookAndFeel().getName();

    // Background
    Color background = table.getBackground();
    if (background == null || background instanceof ColorUIResource) {
        table.setBackground(new ColorUIResource(
            UIManager.getColor("TableHeader.background")));
    }

    // All - only set if there are currently no styles on table
    if (table.cellStyles == null) {
        table.setCellStyle(JCTableEnum.ALL, JCTableEnum.ALL,
                           table.createCellStyle());
    }

    // Labels
    if (table.default_label_style == null) {
        table.default_label_style = table.createCellStyle();
        table.setCellStyle(JCTableEnum.LABEL, JCTableEnum.LABEL,
                           table.default_label_style);
    }
    CellStyleModel labelStyle = table.default_label_style;

    labelStyle.setBackground(UIManager.getColor("TableHeader.background"));
    labelStyle.setForeground(UIManager.getColor("TableHeader.foreground"));
    labelStyle.setFont(UIManager.getFont("TableHeader.font"));
    if (name.equals("CDE/Motif")) {
        BevelBorder border = new BevelBorder(BevelBorder.RAISED);
        labelStyle.setCellBorder(new SwingCellBorder(table, border));
    }
    else {
        labelStyle.setCellBorder(new JCCellBorder(JCTableEnum.BORDER_OUT));
    }

    // Cells
    if (table.default_cell_style == null) {
        table.default_cell_style = table.createCellStyle();
        table.setCellStyle(JCTableEnum.ALLCELLS, JCTableEnum.ALLCELLS,
                           table.default_cell_style);
    }
    CellStyleModel cellStyle = table.default_cell_style;

    cellStyle.setBackground(UIManager.getColor("Table.background"));
    cellStyle.setForeground(UIManager.getColor("Table.foreground"));
    cellStyle.setFont(UIManager.getFont("Table.font"));
    if (name.equals("Windows")) {
        cellStyle.setCellBorder(new JCCellBorder(JCTableEnum.BORDER_IN));
        cellStyle.setCellBorderColorMode(JCTableEnum.BASE_ON_BACKGROUND);
    }
    else {
        cellStyle.setCellBorder(new JCCellBorder(JCTableEnum.BORDER_THIN));
        cellStyle.setCellBorderColorMode(JCTableEnum.USE_CELL_BORDER_COLOR);
        cellStyle.setCellBorderColor(UIManager.getColor("Table.gridColor"));
    }

    // Selection
    Color selectedBackground = table.getSelectedBackground();
    if (selectedBackground == null ||
        selectedBackground instanceof ColorUIResource)
    {
        table.setSelectedBackgroundMode(JCTableEnum.USE_SELECTED_BACKGROUND);
        table.setSelectedBackground(new ColorUIResource(
            UIManager.getColor("Table.selectionBackground")));
    }
    Color selectedForeground = table.getSelectedForeground();
    if (selectedForeground == null ||
        selectedForeground instanceof ColorUIResource)
    {
        table.setSelectedForegroundMode(JCTableEnum.USE_SELECTED_FOREGROUND);
        table.setSelectedForeground(new ColorUIResource(
            UIManager.getColor("Table.selectionForeground")));
    }

    // Focus cell
    Color focus_color = table.getFocusColor();
    if (focus_color == null || focus_color instanceof ColorUIResource) {
        table.setFocusColor(new ColorUIResource(
            UIManager.getColor("Table.focusCellForeground")));
    }
}

/**
 * Returns a <CODE>JCTableUI</CODE> instance.
 * @return <code>JCTableUI</code> instance
 * @param c
 */
public static ComponentUI createUI(JComponent c)
{
    return(new JCTableUI());
}
	
public void uninstallUI(JComponent c) {}
	
}
