/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTraverseCellEvent.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.util.EventObject;

/**
 * This event is posted as a user traverses from one cell to another.<P>
 * The event is posted after a user has committed a cell edit, and before
 * moving to the next cell.
 * @see JCTable#addTraverseCellListener
 * @see JCTraverseCellListener
 */
public class JCTraverseCellEvent extends EventObject {

public static final int TRAVERSE_CELL = 1;
public static final int AFTER_TRAVERSE_CELL = 2;

protected int           type;
protected int           current_row;
protected int           current_column;
protected int           next_row;
protected int           next_column;
protected int           traverse_type;
protected boolean       cancelled = false;

public JCTraverseCellEvent(Object source, int type, int row, int column,
                           int next_row, int next_column, int traverse_type)
{
    super(source);
    this.type = type;

    this.current_row = row;
    this.current_column = column;

    this.next_row = next_row;
    this.next_column = next_column;

    this.traverse_type = traverse_type;
}

/**
 * Gets the row of the current cell.
 * @return
 */
public int getRow()
{
    return(current_row);
}

/**
 * Gets the column of the current cell.
 * @return
 */
public int getColumn()
{
    return(current_column);
}

/**
 * Gets the proposed row of the cell to traverse to.
 * @return
 */
public int getNextRow()
{
    return(next_row);
}

/**
 * Gets the proposed column of the cell to traverse to.
 * @return
 */
public int getNextColumn()
{
    return(next_column);
}

/**
 * Sets the row of the cell to traverse to.
 * @param value
 */
public void setNextRow(int value)
{
    if (type == AFTER_TRAVERSE_CELL) {
        throw new ImmutableEventException("setNextRow is not available " +
                                          "following AFTER_TRAVERSE_CELL");
    }
    if (value < 0) {
        throw new IllegalArgumentException("Row locations must be >= 0");
    }
    next_row = value;
}

/**
 * Sets the column of the cell to traverse to.
 * @param value
 */
public void setNextColumn(int value)
{
    if (type == AFTER_TRAVERSE_CELL) {
        throw new ImmutableEventException("setNextColumn is not available " +
                                          "following AFTER_TRAVERSE_CELL");
    }
    if (value < 0) {
        throw new IllegalArgumentException("Column locations must be >= 0");
    }
    next_column = value;
}

/**
 * Returns the action that caused the traverse. Valid actions are:
 *<table border="0" cellspacing="2">
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_POINTER</code></td><td>mouse click</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_DOWN</code></td><td>down</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_UP</code></td><td>up</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_LEFT</code></td><td> left</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_RIGHT</code></td><td>right</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_PAGEUP</code></td><td>page up</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_PAGEDOWN</code></td><td>page down</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_HOME</code></td><td>top/left</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_END</code></td><td>bottom/right</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_TOP</code></td><td>top of column</td>
 * </tr><tr>
 * <td><code>JCTableEnum.TRAVERSE_BOTTOM</code><td>bottom of column</td>
 * </tr>
 * <tr>
 * <td><code>JCTableEnum.TRAVERSE_TO_CELL</code><td>programmatic</td>
 * </tr>
 *</table>
 * @see JCTable#traverse
 * @return action that caused the traverse
 */
public int getTraverseType()
{
    return(traverse_type);
}

/**
 * Returns a String value for a traverse type.
 * @return String value
 */
public String getTraverseTypeString()
{
    switch (traverse_type) {
    case JCTableEnum.TRAVERSE_POINTER:
        return("POINTER");
    case JCTableEnum.TRAVERSE_DOWN:
        return("DOWN");
    case JCTableEnum.TRAVERSE_UP:
        return("UP");
    case JCTableEnum.TRAVERSE_LEFT:
        return("LEFT");
    case JCTableEnum.TRAVERSE_RIGHT:
        return("RIGHT");
    case JCTableEnum.TRAVERSE_HOME:
        return("HOME");
    case JCTableEnum.TRAVERSE_END:
        return("END");
    case JCTableEnum.TRAVERSE_TOP:
        return("TOP");
    case JCTableEnum.TRAVERSE_BOTTOM:
        return("BOTTOM");
    case JCTableEnum.TRAVERSE_TO_CELL:
        return("PROGRAMMATIC");
    }
    return("");
}

/**
 * Gets the <CODE>Cancelled</CODE> value.
 * @see #setCancelled
 * @return
 */
public boolean isCancelled()
{
    return(cancelled);
}

/**
 * Sets the <CODE>Cancelled</CODE> value.
 * @see #isCancelled
 * @param value
 */
public void setCancelled(boolean value)
{
    if (type == AFTER_TRAVERSE_CELL) {
        throw new ImmutableEventException("setCancelled is not available " +
                                          "in afterResize");
    }
    this.cancelled = value;
}

/**
 * Gets the event type.
 * @return
 */
public int getType()
{
    return(type);
}

/**
 * Returns a String representation of the event.
 * @return String representation of the event
 */
public String toString()
{
    return("TRAVERSE [" +
           (type == TRAVERSE_CELL ? "traverseCell: " : "afterTraverseCell: ") +
           "cancelled=" + isCancelled() + " from " + "R" + current_row +
           "C" + current_column + " to R" + next_row + "C" + next_column +
           ", traverse:" + getTraverseTypeString() + "]");
}

}



