/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: KeyInputHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.cell.JCCellEditor;
import com.klg.jclass.cell.JCKeyModifier;


import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.Serializable;

/**
 * <CODE>KeyInputHandler</CODE> processes all key events for
 * table during key press. It handles traversal, selection, and
 * editor start and stop.
 */
public class KeyInputHandler extends InputHandler
			                 implements KeyListener,
										Serializable
{

private static final int KEY_DOWN      = java.awt.event.KeyEvent.VK_DOWN;
private static final int KEY_UP        = java.awt.event.KeyEvent.VK_UP;
private static final int KEY_LEFT      = java.awt.event.KeyEvent.VK_LEFT;
private static final int KEY_RIGHT     = java.awt.event.KeyEvent.VK_RIGHT;
private static final int KEY_PAGE_UP   = java.awt.event.KeyEvent.VK_PAGE_UP;
private static final int KEY_PAGE_DOWN = java.awt.event.KeyEvent.VK_PAGE_DOWN;
private static final int KEY_TAB       = java.awt.event.KeyEvent.VK_TAB;
private static final int KEY_ESCAPE    = java.awt.event.KeyEvent.VK_ESCAPE;
private static final int KEY_HOME      = java.awt.event.KeyEvent.VK_HOME;
private static final int KEY_END       = java.awt.event.KeyEvent.VK_END;
private static final int KEY_ENTER     = java.awt.event.KeyEvent.VK_ENTER;

public KeyInputHandler(JCTable table)
{
	super(table);
}

/**
 * Processes key pressed events for table.
 */
public void keyPressed(KeyEvent e)
{
	

	int key = e.getKeyCode();
	int modifiers = e.getModifiers();

	if (e.isConsumed() && key == KEY_ENTER) {
		return;
	}

	// Keystrokes are ignored if repaint is disabled
	if (!table.repaint_enabled) {
		return;
	}

	// Checks if the key invokes copy or paste
	if (isCopyAction(e, table.current_row, table.current_column)) {
		performCopyAction();
		return;
	}
	else if (isPasteAction(e, table.current_row, table.current_column)) {
		performPasteAction();
		return;
	}

	// Allow editor to use the key unless reserved by table
	EditTraverseHandler editTraverseHandler = table.getEditTraverseHandler();
	if (editTraverseHandler.getCellEditor() != null) {
		JCKeyModifier[] keycodes =
		    editTraverseHandler.getCellEditor().getReservedKeys();
		if (keycodes != null && keycodes.length > 0) {
			for (int i = 0; i < keycodes.length; i++) {
				if (keycodes[i] == null) {
					continue;
				}
				// editor can reserve other keys for itself
				if (keycodes[i].match(key, modifiers)) {
					// table reserves keys
					if (isContainerReserved(keycodes[i])) {
						continue;
					}
					return;
				}
			}
		}
	}
	// If an editor isn't present, determine whether the cell's editor has
	// reserved the key/modifier keys for itself and the combination has
	// been flagged to initialize an edit. Edit will not be intialized
	// for keys reserved by the container.
	else if (table.current_row >= 0     &&
	         table.current_column >= 0  &&
			 !isContainerReserved(key, modifiers))
	{
		JCCellEditor ce = table.getDataView().getCellEditor(table.current_row,
														    table.current_column);
		if (ce != null) {
			JCKeyModifier[] keycodes = ce.getReservedKeys();
			if (keycodes != null && keycodes.length > 0) {
				for (int i = 0; i < keycodes.length; i++) {
					if (keycodes[i] == null) {
						continue;
					}
					if (keycodes[i].match(key, modifiers) &&
					    keycodes[i].canInitializeEdit)
					{
						// initialize an edit if the key isn't reserved
						if (isContainerReserved(keycodes[i])) {
							continue;
						}
						editTraverseHandler.traverse(table.current_row,
													 table.current_column,
													 false, e,
													 JCTableEnum.TRAVERSE_TO_CELL,
													 true);
						return;
					}
				}
			}
		}
	}

	// Consume tabs that will originate from editors
	if (key == KEY_TAB) {
		e.consume();
	}

	if (key == KEY_ENTER) {
		editTraverseHandler.traverse(table.getCurrentRow(),
									 table.getCurrentColumn(),
									 false, null,
									 JCTableEnum.TRAVERSE_POINTER,
									 true);
		return;
	}

	SELECTING = false;

	// Process the basic arrow keys
	if (key == KEY_TAB) {                                     // tab
		if (isControl(modifiers)) {
		    // Handle CTRL-TAB and CTRL-SHIFT-TAB.  We need to set the focus
		    // outside of table to either the next or previous component in the
		    // focus cycle above us.  Only works in JDK 1.4 and later.
		    table.getFocusManager().handleControlTab(isShift(modifiers), true);
		}
		else {
		    if (isShift(modifiers)) {
			    traverse(JCTableEnum.TRAVERSE_LEFT);
		    }
		    else {
			    traverse(JCTableEnum.TRAVERSE_RIGHT);
		    }
		}
	}
	else if (key == KEY_DOWN) {                              // down
		// If shift, extend selection, down, else traverse. Note that selection
		// does not cycle past the last row
		if (isShift(modifiers) && table.getSelectionModel() != null) {
			SELECTING = table.getSelectionModel().select(SelectionModel.SELECT,
														 table.current_row,
														 table.current_column);
			SELECTING = table.getSelectionModel().select(SelectionModel.EXTEND,
														 Math.min(table.getNumRows() - 1,
														          table.current_row + 1),
														 table.current_column);
		}
		if (!SELECTING) {
            table.clearSelectedCells();
		}
		traverse(JCTableEnum.TRAVERSE_DOWN);
	}
	else if (key == KEY_UP) {                                // up
		// If shift, extend selection, up, else traverse. Note that selection
		// does not cycle past row 0
		if (isShift(modifiers) && table.getSelectionModel() != null) {
			SELECTING = table.getSelectionModel().select(SelectionModel.SELECT,
														 table.current_row,
														 table.current_column);
			SELECTING = table.getSelectionModel().select(SelectionModel.EXTEND,
														 Math.max(table.current_row - 1,
														          0),
														 table.current_column);
		}
		if (!SELECTING) {
            table.clearSelectedCells();
		}
		traverse(JCTableEnum.TRAVERSE_UP);
	}
	else if (key == KEY_LEFT) {                              // left
		// If shift, extend selection, left, else traverse. Note that selection
		// does not cycle past column 0
		if (isShift(modifiers) && table.getSelectionModel() != null) {
			SELECTING = table.getSelectionModel().select(SelectionModel.SELECT,
														 table.current_row,
														 table.current_column);
			SELECTING = table.getSelectionModel().select(SelectionModel.EXTEND,
														 table.current_row,
														 Math.max(table.current_column - 1,
														          0));
		}
		if (!SELECTING) {
            table.clearSelectedCells();
		}
		traverse(JCTableEnum.TRAVERSE_LEFT);
	}
	else if (key == KEY_RIGHT) {                            // right
		// If shift, extend selection, right, else traverse. Note that selection
		// does not cycle past the right-most column
		if (isShift(modifiers) && table.getSelectionModel() != null) {
			SELECTING = table.getSelectionModel().select(SelectionModel.SELECT,
														 table.current_row,
														 table.current_column);
			SELECTING = table.getSelectionModel().select(SelectionModel.EXTEND,
														 table.current_row,
														 Math.min(table.getNumColumns() - 1,
														          table.current_column + 1));
		}
		if (!SELECTING) {
            table.clearSelectedCells();
		}
		traverse(JCTableEnum.TRAVERSE_RIGHT);
	}
	else if (key == KEY_HOME && isControl(modifiers)) {      // control home
		traverse(JCTableEnum.TRAVERSE_HOME);
    }
	else if (key == KEY_PAGE_DOWN && isControl(modifiers)) { // control page down
		traverse(JCTableEnum.TRAVERSE_BOTTOM);
	}
	else if (key == KEY_END && isControl(modifiers)) {       // control end
		traverse(JCTableEnum.TRAVERSE_END);
    }
	else if (key == KEY_PAGE_UP && isControl(modifiers)) {   // control page up
		traverse(JCTableEnum.TRAVERSE_TOP);
	}
	else if (key == KEY_PAGE_DOWN) {                         // page down
		traverse(JCTableEnum.TRAVERSE_PAGEDOWN);
	}
	else if (key == KEY_PAGE_UP) {                           // page up
		traverse(JCTableEnum.TRAVERSE_PAGEUP);
	}
	else if (key == KEY_ESCAPE) {                            // cancel
		table.clearSelection();
		SELECTING = false;
	}
	else if (isModifierKey(key)) {
		// Ignore
	}
	else if (editTraverseHandler.getCellEditor() == null) {
		editTraverseHandler.traverse(table.current_row,
									 table.current_column,
									 false, e,
									 JCTableEnum.TRAVERSE_TO_CELL,
									 true);
	}

    // Remove the selection if we are not selecting, but are traversing.
    if (!SELECTING) {
        switch(key) {
        case KEY_PAGE_UP:
        case KEY_PAGE_DOWN:
        case KEY_TAB:
            table.clearSelectedCells();
            break;
        default:
            break;
        }
    }
}

public void keyReleased(KeyEvent e)
{
	

	if (SELECTING) {
		if (e.getKeyCode() == KeyEvent.VK_SHIFT) {
			// Need to generate an AFTER_SELECTION JCSelectEvent message
		    table.getSelectionModel().select(SelectionModel.END,
		                                     table.current_row,
											 table.current_column);
		    SELECTING = false;
		}
	}
}

public void keyTyped(KeyEvent e)
{
	// ignore these events, only the key press event is used.
}

/**
 * Traverses from the current cell in the given direction.
 */
protected boolean traverse(int direction)
{
	

	int current_row = table.current_row;
	int current_column = table.current_column;

	// find the next traversable cell
	JCCellPosition cp;
	cp = findTraversableCell(direction, current_row, current_column);

	// return false if a cell isn't found
	if (cp == null) {
		return(false);
	}

	// check if the new position conflicts with the traverse cycle setting
	if (!table.isTraverseCycle()) {
		if (direction == JCTableEnum.TRAVERSE_RIGHT &&
		    (current_row != cp.row || cp.column < current_column))
		{
			return(false);
		}
		else if (direction == JCTableEnum.TRAVERSE_LEFT &&
				 (current_row != cp.row || cp.column > current_column))
		{
			return(false);
		}
		else if (direction == JCTableEnum.TRAVERSE_UP &&
				 (cp.row > current_row || current_column != cp.column))
		{
			return(false);
		}
		else if (direction == JCTableEnum.TRAVERSE_DOWN &&
				 (cp.row < current_row || current_column != cp.column))
		{
			return(false);
		}
	}

	EditTraverseHandler editTraverseHandler = table.getEditTraverseHandler();
	if (!editTraverseHandler.commit(true)) {
		return(false);
	}

	return(editTraverseHandler.traverse(cp.row, cp.column, false,
										null, direction, false));
}

/**
 * Finds the next traversable cell given a direction and current position.
 * Directions include:
 * <UL><LI><CODE>JCTableEnum.TRAVERSE_LEFT</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_RIGHT</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_UP</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_DOWN</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_PAGEUP</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_PAGEDOWN</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_TOP</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_BOTTOM</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_HOME</CODE>
 * <LI><CODE>JCTableEnum.TRAVERSE_END</CODE></UL>
 * <P>Null is returned if no traversable cell is found.
 */
protected JCCellPosition findTraversableCell(int direction, int row, int column)
{
	

	if (row < 0                     ||
	    column < 0                  ||
		table.getNumColumns() == 0  ||
        table.getNumRows() == 0)
	{
		return(null);
    }

	int i = row;
	int j = column;

	if (direction == JCTableEnum.TRAVERSE_RIGHT) {
		j = column + 1;
		while (i != row || j != column) {
			for (int cnt = table.getNumColumns(); j < cnt; j++) {
				// cycled back, return
				if (i == row && j == column) {
					return(null);
				}
				else if (table.isTraversable(i, j)) {
					return(new JCCellPosition(i, j));
				}
			}
			// no match, so increment the row and reset the column
			if (i == table.getNumRows() - 1) {
				i = 0;
			}
			else {
				i++;
			}
			j = 0;
		}
	}
	else if(direction == JCTableEnum.TRAVERSE_LEFT) {
		j = column - 1;
		while (i != row || j != column) {
			for (; j >= 0; j--) {
				// cycled back, return
				if (i == row && j == column) {
					return(null);
				}
				else if (table.isTraversable(i, j)) {
					return(new JCCellPosition(i, j));
				}
			}
			// no match, so decrement the row and reset the column
			if (i == 0) {
				i = table.getNumRows() - 1;
			}
			else {
				i--;
			}
			j = table.getNumColumns() - 1;
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_DOWN) {
		i = row + 1;
		while (i != row || j != column) {
			for (int cnt = table.getNumRows(); i < cnt; i++) {
				// cycled back, return
				if (i == row && j == column) {
					return(null);
				}
				else if (table.isTraversable(i, j)) {
					return(new JCCellPosition(i, j));
				}
			}
			// no match, so increment the column and reset the row
			if (j == table.getNumColumns() - 1) {
				j = 0;
			}
			else {
				j++;
			}
			i = 0;
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_UP) {
		i = row - 1;
		while (i != row || j != column) {
			for (; i >= 0; i--) {
				// cycled back, return
				if (i == row && j == column) {
					return(null);
				}
				else if (table.isTraversable(i, j)) {
					return(new JCCellPosition(i, j));
				}
			}
			// no match, so decrement the column and reset the row
			if(j == 0) {
				j = table.getNumColumns() - 1;
			}
			else {
				j--;
			}
			i = table.getNumRows() - 1;
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_HOME) {
		i = 0;
		j = 0;
		if (table.isTraversable(i,j)) {
			return(new JCCellPosition(i,j));
		}
		else {
			for(i = 0; i < table.getNumRows(); i++) {
				for(j = 0; j < table.getNumColumns(); j++) {
					if (table.isTraversable(i,j)) {
						return(new JCCellPosition(i,j));
					}
				}
			}
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_PAGEUP) {
		int rows = table.getNumRows();
		JCCellRange cr = table.getVisibleCells();
		i = Math.max(0, row - (cr.end_row - cr.start_row));
		j = column;
		if (table.isTraversable(i,j)) {
			return new JCCellPosition(i,j);
		}
		else {
			for (i = 0; i < rows; i++) {
				if (table.isTraversable(i,j)) {
					return(new JCCellPosition(i,j));
				}
			}
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_TOP) {
		i = 0;
		j = column;
		if (table.isTraversable(i,j)) {
			return(new JCCellPosition(i,j));
		}
		else {
			for (i = 0; i < table.getNumRows(); i++) {
				if (table.isTraversable(i,j)) {
					return(new JCCellPosition(i,j));
				}
			}
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_PAGEDOWN) {
		int rows = table.getNumRows();
		JCCellRange cr = table.getVisibleCells();
		i = Math.min(rows - 1, row + (cr.end_row - cr.start_row));
		j = column;
		if (table.isTraversable(i,j)) {
			return(new JCCellPosition(i,j));
		}
		else {
			for(i = 0; i < rows; i++) {
				if (table.isTraversable(i,j)) {
					return(new JCCellPosition(i,j));
				}
			}
			return(null);
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_END) {
		int rows = table.getNumRows();
		int columns = table.getNumColumns();
		i = rows - 1; j = columns - 1;
		if (table.isTraversable(i,j)) {
			return(new JCCellPosition(i,j));
		}
		else {
			for (i = rows - 1; i >= 0; i--) {
				for (j = columns - 1; j >= 0; j--) {
					if (table.isTraversable(i,j)) {
						return(new JCCellPosition(i,j));
					}
				}
			}
		}
	}
	else if (direction == JCTableEnum.TRAVERSE_BOTTOM) {
		int rows = table.getNumRows();
		i = rows - 1; j = column;
		if (table.isTraversable(i,j)) {
			return(new JCCellPosition(i,j));
		}
		else {
			for(i = rows - 1; i >= 0; i--) {
				if (table.isTraversable(i,j)) {
					return(new JCCellPosition(i,j));
				}
			}
		}
	}

	return(null);
}

/**
 * Returns true if the key is a modifier.
 */
public boolean isModifierKey(int key)
{
	return(key == KeyEvent.VK_CONTROL ||
	       key == KeyEvent.VK_SHIFT   ||
		   key == KeyEvent.VK_META    ||
		   key == KeyEvent.VK_ALT);
}

/**
 * Determines whether the specified key modifier represents an attempt
 * by an editor to reserve a key that table reserves for itself.
 * Table reserves tab and cancel for itself.
 */
private boolean isContainerReserved(JCKeyModifier km)
{
	return(isContainerReserved(km.key, km.modifier));
}

/**
 * Returns true if the given key and modifier are reserved by table.
 */
private boolean isContainerReserved(int key, int modifiers)
{
	return(key == KeyEvent.VK_ESCAPE || (key == KeyEvent.VK_TAB &&
										 (modifiers == 0 ||
                                          isShift(modifiers))));
}

}
