/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: MouseInputHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.awt.Cursor;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import javax.swing.Timer;

import com.klg.jclass.cell.JCCellEditor;


/**
 * <CODE>MouseInputHandler</CODE> processes all mouse events for
 * table. It handles traversal, selection, resize, editor start
 * and stop, and dragging.
 */
public class MouseInputHandler extends InputHandler
			                   implements java.io.Serializable,
										  ActionListener,
										  MouseListener,
										  MouseMotionListener
{

protected static final int  DEFAULT_CURSOR = Cursor.DEFAULT_CURSOR;
protected static final int  TRAV_CURSOR    = Cursor.TEXT_CURSOR;
protected static final int  NONTRAV_CURSOR = Cursor.CROSSHAIR_CURSOR;
protected static final int  CORNER_CURSOR  = Cursor.SE_RESIZE_CURSOR;
protected static final int  HORIZ_CURSOR   = Cursor.E_RESIZE_CURSOR;
protected static final int  VERT_CURSOR    = Cursor.S_RESIZE_CURSOR;
protected static final int  MOVE_CURSOR    = Cursor.HAND_CURSOR;
protected static final int  WAIT_CURSOR    = Cursor.WAIT_CURSOR;

protected static final int  NO_RESIZE     = 0;
protected static final int  HORIZ_RESIZE  = 1;
protected static final int  VERT_RESIZE   = 2;
protected static final int  CORNER_RESIZE = 3;

protected JCCellPosition    initial_cell = new JCCellPosition();
protected JCCellPosition    position     = new JCCellPosition();

protected boolean           multi_cell_drag = false;

// resize
protected int               resizeRow    = JCTableEnum.NOVALUE;
protected int               resizeColumn = JCTableEnum.NOVALUE;
protected int               resizeHeight;
protected int               resizeWidth;
protected CellSize          resizeRowCellSize;
protected CellSize          resizeColumnCellSize;
protected Point             resizePoint;

protected boolean           needs_resize_begin_event = false;
protected JCResizeCellEvent resize_motion_event      = null;

// autoscroll
protected static final int  AUTO_SCROLL_DEC    = -1;
protected static final int  AUTO_SCROLL_NONE   = 0;
protected static final int  AUTO_SCROLL_INC    = 1;
protected static final int  AUTO_SCROLL_AMOUNT = 20;

protected int               auto_scroll_horiz = AUTO_SCROLL_NONE;
protected int               auto_scroll_vert  = AUTO_SCROLL_NONE;
protected transient Timer   auto_scroll_timer;
protected JCCellPosition    auto_scroll_initial_cell;
protected int               auto_scroll_x;
protected int               auto_scroll_y;

protected boolean           AUTOSCROLLING = false;
protected boolean           DRAGGING      = false;
protected boolean           RESIZING      = false;
protected boolean           SORTING       = false;


public MouseInputHandler(JCTable table)
{
	super(table);
}

/**
 * Returns true if the mouse press is with the left mouse button.
 * @param modifiers The mouse event modifiers
 * @return Was the left button pressed?
 */
public final boolean isLeftButton(int modifiers)
{
	return((modifiers & MouseEvent.BUTTON1_MASK) != 0);
}

/**
 * Returns true if the mouse press is with the middle mouse button.
 * @param modifiers The mouse event modifiers
 * @return Was the middle button pressed?
 */
public final boolean isMiddleButton(int modifiers)
{
	return((modifiers & MouseEvent.BUTTON2_MASK) != 0);
}

/**
 * Returns true if the mouse press is with the right mouse button.
 * @param modifiers The mouse event modifiers
 * @return Was the right button pressed?
 */
public final boolean isRightButton(int modifiers)
{
	return((modifiers & MouseEvent.BUTTON3_MASK) != 0);
}

/**
 * Process mouse press.
 * @param e The event generated by the mouse press
 */
public void mousePressed(MouseEvent e)
{
	

	int x = e.getX();
	int y = e.getY();
	int modifiers = e.getModifiers();

	// Return if batching is enabled
	if (!table.isRepaintEnabled()) {
		return;
	}

	// Don't know where this will end up - just need it someplace to get
	// things going
	table.getFocusManager().requestFocus();

	getCellPosition(x, y, initial_cell);

	// Reset the flag that determines whether multiple cells have been "covered"
	// by mouse drag
	multi_cell_drag = false;

	// Initiate a resize if over a cell boundary
	if (resizeRow != JCTableEnum.NOVALUE ||
	    resizeColumn != JCTableEnum.NOVALUE)
	{
		// Get the CellSize objects for the cell being resized
		resizeRowCellSize = table.getRowCellSize(initial_cell.row);
		resizeColumnCellSize = table.getColumnCellSize(initial_cell.column);
		// Initiate resize
		resizeBegin(x, y);
		RESIZING = true;
		return;
	}

	// true if selection has been triggered by a method other than the default
	// way (left mouse button).
	boolean otherSelect = false;

	if (isValidCellPosition(initial_cell)) {
		int action = table.getAction(e, initial_cell.row, initial_cell.column);
		if (action == TableAction.COLUMN_DRAG_ACTION ||
		    action == TableAction.ROW_DRAG_ACTION)
		{
			if (isValidDragPosition(initial_cell)) {
				DRAGGING = true;
				table.cancelEdit(true);
				table.clearSelection();
				performDrag(initial_cell);
				return;
			}
		}
		else if (action == TableAction.COLUMN_SORT_ACTION &&
		         initial_cell.row == JCTableEnum.LABEL)
		{
			SORTING = true;
			table.cancelEdit(true);
			table.clearSelection();
			return;
		}
		else if ((action == TableAction.COLUMN_SELECT_ACTION &&
			      initial_cell.row == JCTableEnum.LABEL) ||
				 (action == TableAction.ROW_SELECT_ACTION &&
				  initial_cell.column == JCTableEnum.LABEL))
		{
			// A selection has been triggered through a non-standard way :
			// column/row select interactions on a particular trigger
			otherSelect = true;
		}
	}

	traverse(e, initial_cell, true);

	if (isLeftButton(modifiers) || otherSelect) {
		// Initiate a selection. The SelectionModel will return false if
		// selection is disallowed
		int select_mode = SelectionModel.SELECT;
		if (isShift(modifiers)) {
			select_mode = SelectionModel.EXTEND;
		}
		else if (isControl(modifiers)) {
			select_mode = SelectionModel.ADD;
			multi_cell_drag = true;
		}
		if (table.getSelectionModel() != null) {
			SELECTING = table.getSelectionModel().select(select_mode,
													     initial_cell.row,
														 initial_cell.column);
		}
		else {
			SELECTING = false;
		}
	}
    
}

/**
 * Returns true if the given position is valid for dragging. The drag process
 * is restricted to non-frozen cells.
 * @param cp The cell position (i.e. row/column)
 * @return Can you drag from the cell referenced by the given cell position
 */
protected boolean isValidDragPosition(JCCellPosition cp)
{
	if (cp.row == JCTableEnum.LABEL) {
		return(cp.column >= table.getFrozenColumns());
	}
	else if (cp.column == JCTableEnum.LABEL) {
		return(cp.row >= table.getFrozenRows());
	}
	return(false);
}

/**
 * Process mouse drag.
 * @param e The event generated by the mouse drag
 */
public void mouseDragged(MouseEvent e)
{
//	

	int x = e.getX();
	int y = e.getY();

	// return if batching is enabled
	if (!table.isRepaintEnabled()) {
		return;
	}

	JCCellPosition current_cell = new JCCellPosition();
	getCellPosition(x, y, current_cell);

	// set the multi_cell_drag flag true if the current cell doesn't match
	// the initial cell. This is used to determine whether selection or edit occurs
	// on mouse release
	if (!current_cell.equals(initial_cell)) {
		multi_cell_drag = true;
	}

	if (RESIZING) {
		resizeDrag(x, y);
	}
	else {
		if (isValidCellPosition(current_cell) &&
		    isValidCellPosition(initial_cell))
		{
			if (AUTOSCROLLING) {
				stopAutoScroll(true);
			}

			if (DRAGGING) {
				SELECTING = false;
				if(isValidDragPosition(current_cell)) {
					performDrag(current_cell);
				}
			}
		}
		// If the cell position is invalid, check for autoscroll
		else if (!isValidCellPosition(current_cell)) {
			CellArea cellArea = table.cellAreaHandler.cellArea;

			// Check for row autoscroll
			if (isAutoScrollRow()) {
				if (y < cellArea.getBounds().y) {
					auto_scroll_vert = AUTO_SCROLL_DEC;
				}
				else if (y > cellArea.getBounds().y + cellArea.getBounds().height) {
					auto_scroll_vert = AUTO_SCROLL_INC;
				}
				else {
					auto_scroll_vert = AUTO_SCROLL_NONE;
				}
				if (initial_cell.row == JCTableEnum.LABEL) {
					auto_scroll_vert = AUTO_SCROLL_NONE;
				}
			}
			// Check for column autoscroll
			if (isAutoScrollColumn()) {
				if (x < cellArea.getBounds().y) {
					auto_scroll_horiz = AUTO_SCROLL_DEC;
				}
				else if (x > cellArea.getBounds().x + cellArea.getBounds().width) {
					auto_scroll_horiz = AUTO_SCROLL_INC;
				}
				else {
					auto_scroll_horiz = AUTO_SCROLL_NONE;
				}
				if (initial_cell.column == JCTableEnum.LABEL) {
					auto_scroll_horiz = AUTO_SCROLL_NONE;
				}
			}
			// Start the autoscroll process -- autoscroll parameters are validated
			// in start autoscroll so there is no need to check beforehand
			startAutoScroll(x, y);
		}

		if (SELECTING) {
			getCellPosition(x, y, position);
			SELECTING = table.getSelectionModel().select(SelectionModel.EXTEND,
													     position.row,
														 position.column);
		}
	}
}

/**
 * Process mouse release.
 * @param e The event generated by the mouse release
 */
public void mouseReleased(MouseEvent e)
{
	

	int x = e.getX();
	int y = e.getY();

	// Return if batching is enabled
	if (!table.isRepaintEnabled()) {
		return;
	}

	int scroll_vert = auto_scroll_vert;
	int scroll_horiz = auto_scroll_horiz;

	if (AUTOSCROLLING) {
		stopAutoScroll(true);
	}

	JCCellPosition final_cell = new JCCellPosition();
	getCellPosition(x, y, final_cell);

	// End resize if in progress
	EditTraverseHandler editTraverseHandler = table.getEditTraverseHandler();
	if (RESIZING) {
		resizeEnd(x, y);
		RESIZING = false;
		return;
	}
	// End drag if in progress
	else if (DRAGGING) {
		DRAGGING = false;
		if (isValidDragPosition(final_cell)) {
			endDrag(final_cell);
		}
		else {
			JCCellPosition cp = new JCCellPosition();
			if (final_cell.row == JCTableEnum.LABEL) {
				cp.row = JCTableEnum.LABEL;
				cp.column = table.getFrozenColumns();
				endDrag(cp);
			}
			else if (final_cell.column == JCTableEnum.LABEL) {
				cp.row = table.getFrozenRows();
				cp.column = JCTableEnum.LABEL;
				endDrag(cp);
			}
			// If auto scroll was in progress when the mouse was released,
			// use the last position as the end point
			else if(scroll_vert != AUTO_SCROLL_NONE ||
			        scroll_horiz != AUTO_SCROLL_NONE)
			{
				cp.row = table.cellAreaHandler.drop_row;
				// The drag process increments the row counter for insertion
				if (scroll_vert == AUTO_SCROLL_INC &&
				    cp.row != JCTableEnum.LABEL)
				{
					cp.row--;
				}

				cp.column = table.cellAreaHandler.drop_column;
				// The drag process increments the column counter for insertion
				if (scroll_horiz == AUTO_SCROLL_INC &&
				    cp.column != JCTableEnum.LABEL)
				{
					cp.column--;
				}
				endDrag(cp);
			}
			else {
				cancelDrag();
			}
		}
		return;
	}
	// Sort the column
	else if (SORTING) {
		if (table.getDataView() instanceof SortableDataViewModel) {
			Cursor oldCursor = table.getCursor();
			table.setCursor(WAIT_CURSOR);
			SortableDataViewModel sdvm = (SortableDataViewModel)table.getDataView();
			int sort_column = sdvm.getDataColumn(final_cell.column);
			int sort_direction = Sort.ASCENDING;
			if(table.sort_column == sort_column) {
				if(table.sort_direction == Sort.ASCENDING) {
					sort_direction = Sort.DESCENDING;
				}
			}
			table.sortByColumn(sort_column, sort_direction);
			table.setCursor(oldCursor);
			SORTING = false;
		}
		return;
	}
	// Determine whether an editor should be invoked.  Note that if autoEdit
    // is true, the editor has already been invoked in the mousePress.
	else if (!table.isAutoEdit()                 &&
             table.getDataView().isEditable()    &&
			 initial_cell.row == final_cell.row  &&
			 initial_cell.column == final_cell.column)
	{
		// Edit if the cell is editable and has an editor
		if (!multi_cell_drag                                     &&
		    table.isEditable(final_cell.row, final_cell.column)  &&
		    table.getDataView().getCellEditor(final_cell.row,
                                              final_cell.column) != null)
		{
            JCCellEditor cellEditor = editTraverseHandler.getCellEditor();
            if (table.pointer_traverse_forces_edit || cellEditor == null) {
				editTraverseHandler.traverse(final_cell.row,
											 final_cell.column,
											 false, e,
											 JCTableEnum.TRAVERSE_POINTER,
											 true);
				return;
			}
			else {
				// Get the focus back on the editor that kept the edit
                if (cellEditor != null) {
					Component comp = cellEditor.getComponent();
					table.getFocusManager().requestFocus(comp);
				}
			}
		}
	}
	// End selection if in progress
	if (SELECTING) {
		table.getSelectionModel().select(SelectionModel.END,
		                                 final_cell.row, final_cell.column);
		SELECTING = false;
	}
    
}

/**
 * Process mouse enter.
 * @param e The event generated by the mouse enter
 */
public void mouseEntered(MouseEvent e)
{
	// ingore these
}

/**
 * Process mouse move.
 * @param e The event generated by the mouse move
 */
public void mouseMoved(MouseEvent e)
{
//	

    // If repaint and cursor tracking are enabled and a drag or resize isn't
    // in progress, change the cursor as required
	if (table.isRepaintEnabled() && !DRAGGING && !RESIZING && table.isTrackCursor()) {
		table.setCursor(trackResizeCursor(e.getX(), e.getY()));
	}
}

/**
 * Process mouse exit.
 * @param e The event generated by the mouse exit
 */
public void mouseExited(MouseEvent e)
{
//	

	if (!RESIZING && table.isTrackCursor()) {
		table.setCursor(DEFAULT_CURSOR);
	}
}

/**
 * Process mouse click.
 * @param e The event generated by the mouse click
 */
public void mouseClicked(MouseEvent e)
{
	// ignore these, it's just a press/release combo
}

///**
// * Returns a String representation of the cell at the given mouse coordinates.
// * @param x The x coordinate of the mouse
// * @param y The y coordinate of the mouse
// * @return A String representation of the cell at (x, y)
// */
//private String getCellAt(int x, int y)
//{
//	getCellPosition(x, y, position);
//	return("R" + position.row + "C" + position.column);
//}

/**
 * Returns the cursor type given a mouse position.
 * @param x The x coordinate of the mouse
 * @param y The y coordinate of the mouse
 * @return The cursor type based on the mouse being at (x, y)
 */
protected int trackResizeCursor(int x, int y)
{
	getCellPosition(x, y, position);
	int row    = position.row;
	int column = position.column;
	int cursor = NONTRAV_CURSOR;

	if (row == JCTableEnum.NOVALUE || column == JCTableEnum.NOVALUE) {
		return(cursor);
	}

	switch(getResizeDirection(x, y, row, column)) {
	case HORIZ_RESIZE:
		cursor = HORIZ_CURSOR;
		break;
	case VERT_RESIZE:
		cursor = VERT_CURSOR;
		break;
	case CORNER_RESIZE:
		cursor = CORNER_CURSOR;
		break;
	default:
		if (isTraversable(row, column)) {
			cursor = TRAV_CURSOR;
		}
		else {
			cursor = NONTRAV_CURSOR;
		}
	}
	return(cursor);
}

/**
 * Determines whether a resize is available at the current position.
 * @param x The x coordinate of the current position
 * @param y The y coordinate of the current position
 * @param row The row in which the current position is in
 * @param column The column in which the current position is in
 * @return The resize type available.  Valid return values are
 *     <CODE>NO_RESIZE</CODE>, <CODE>HORIZ_RESIZE</CODE>,
 *     <CODE>VERT_RESIZE</CODE> and <CODE>BORDER_RESIZE</CODE>.
 */
protected int getResizeDirection(int x, int y, int row, int column)
{
	int MAXDELTA = 4;

	resizeRow    = JCTableEnum.NOVALUE;
	resizeColumn = JCTableEnum.NOVALUE;

	Rectangle bounds = table.getCellAreaHandler().getBounds(row, column);

	CellArea area = table.getCellAreaHandler().find(row, column);
	if (area == null) {
		return(NO_RESIZE);
	}

	Point cellOrigin = table.getCellAreaHandler().getCellOrigin();

	// adjust the coordinates by the type of area in order to determine the cell
	// that contains the given point
	if (area.getType() == CellArea.CELL) {
		x += -cellOrigin.x;
		y += -cellOrigin.y;
	}
	else if (area.getType() == CellArea.FROZENCOLUMN ||
	         area.getType() == CellArea.ROWLABEL)
	{
		y += -cellOrigin.y;
	}
	else if (area.getType() == CellArea.FROZENROW ||
	         area.getType() == CellArea.COLUMNLABEL)
	{
		x += -cellOrigin.x;
	}

	int xPos = bounds.x;
	if (x - xPos < MAXDELTA) {
		column--;
	}

	int yPos = bounds.y;
	if (y - yPos < MAXDELTA) {
		row--;
	}

	// Return NO_RESIZE if resize is not available for the current position
	// based on resize properties
	if (!table.isResizable(row, column)) {
		return(NO_RESIZE);
	}

	int resize = NO_RESIZE;
	if (column < -1) {
		return(NO_RESIZE);
	}
	int checkX = bounds.x + bounds.width;
	if (Math.abs(checkX - x) < MAXDELTA) {
		resizeColumn = column;
		resize |= HORIZ_RESIZE;
	}
	if (row < -1) {
		return(NO_RESIZE);
	}
	int checkY = bounds.y + bounds.height;
	if (Math.abs(checkY - y) < MAXDELTA) {
		resizeRow = row;
		resize |= VERT_RESIZE;
	}

	// Return NO_RESIZE if resize is horizontal but column resize is disallowed
	if ((resize & HORIZ_RESIZE) != 0 &&
	    table.allow_cell_resize == JCTableEnum.RESIZE_ROW)
	{
		resizeRow = JCTableEnum.NOVALUE;
		resizeColumn = JCTableEnum.NOVALUE;
		return(NO_RESIZE);
	}
	// Return NO_RESIZE if resize is vertical but row resize is disallowed
	if ((resize & VERT_RESIZE) != 0 &&
	    table.allow_cell_resize == JCTableEnum.RESIZE_COLUMN)
	{
		resizeRow = JCTableEnum.NOVALUE;
		resizeColumn = JCTableEnum.NOVALUE;
		return(NO_RESIZE);
	}

	// Get the cell range that contains this cell if spanned
	if (table.spanHandler != null) {
		JCCellRange cr = table.getSpannedRange(row, column);
		if (cr != null) {
			if ((resize & VERT_RESIZE) != 0 && row != cr.end_row) {
				resize &= ~VERT_RESIZE;
			}
			if ((resize & HORIZ_RESIZE) != 0  && column != cr.end_column) {
				resize &= ~HORIZ_RESIZE;
			}
		}
	}

	return(resize);
}

/**
 * Initiates a resize action.
 * @param x The x position of the resize begin
 * @param y The y position of the resize begin
 */
protected void resizeBegin(int x, int y)
{
	resizePoint = new Point(x, y);
	CellLayoutModel cl = table.getCellLayout();

	resizeWidth  = (resizeColumn != JCTableEnum.NOVALUE) ? cl.getWidth(resizeColumn)
														 : JCTableEnum.NOVALUE;
	resizeHeight = (resizeRow    != JCTableEnum.NOVALUE) ? cl.getHeight(resizeRow)
													     : JCTableEnum.NOVALUE;

	needs_resize_begin_event = true;
}

/**
 * Handles a drag event during a resize action.
 * @param x The x position of the resize drag
 * @param y The y position of the resize drag
 */
protected void resizeDrag(int x, int y)
{
	// Invoke a resize begin event on the first drag
	if (needs_resize_begin_event) {
		// Sends a JCResizeCellEvent object to registered
		// JCResizeCellListener objects.
		if (table.resizeListeners != null) {
			if (resize_motion_event == null) {
				resize_motion_event =
				    new JCResizeCellEvent(table,
										  JCResizeCellEvent.BEFORE_RESIZE,
										  resizeRow, resizeColumn,
										  resizeWidth, resizeHeight);
			}
			else {
				resize_motion_event.type = JCResizeCellEvent.BEFORE_RESIZE;
				resize_motion_event.row = resizeRow;
				resize_motion_event.column = resizeColumn;
				resize_motion_event.current_row_height = resizeHeight;
				resize_motion_event.current_column_width = resizeWidth;
				resize_motion_event.new_row_height = JCTableEnum.NOVALUE;
				resize_motion_event.new_column_width = JCTableEnum.NOVALUE;
				resize_motion_event.cancelled = false;
			}
			table.fireJCResizeCellEvent(resize_motion_event);
			if (resize_motion_event.isCancelled()) {
				cancelResize();
				// fire an after event with the cancelled flag
				resize_motion_event.type = JCResizeCellEvent.AFTER_RESIZE;
				table.fireJCResizeCellEvent(resize_motion_event);
				return;
			}
		}
		needs_resize_begin_event = false;
	}

	int newHeight = JCTableEnum.NOVALUE;
	int originalNewHeight = JCTableEnum.NOVALUE;
	if (resizeRow != JCTableEnum.NOVALUE) {
		int adjustment = y - resizePoint.y;
		int height = table.getCellLayout().getHeight(resizeRow);

		if (height + adjustment < 0) {
			resizePoint.y = y - adjustment;
			adjustment = 0;
		}
		else {
			resizePoint.y = y;
		}

		newHeight = table.getCellLayout().getHeight(resizeRow) + adjustment;
		if (newHeight < 5) {
			newHeight = 5;
		}

		// Save the new height so that we can determine whether min/max logic
		// or a listener changes it and adjust the resizePoint accordingly
		originalNewHeight = newHeight;

		if (resizeRowCellSize != null) {
			newHeight = Math.min(Math.max(newHeight,
										  resizeRowCellSize.getMinValue()),
								 resizeRowCellSize.getMaxValue());
		}
	}

	int newWidth = JCTableEnum.NOVALUE;
	int originalNewWidth = JCTableEnum.NOVALUE;
	if (resizeColumn != JCTableEnum.NOVALUE) {
		int adjustment = x - resizePoint.x;
		int width = table.getCellLayout().getWidth(resizeColumn);

		if(width + adjustment < 0) {
			resizePoint.x  = x - adjustment;
			adjustment = 0;
		}
		else {
			resizePoint.x  = x;
		}

		newWidth = table.getCellLayout().getWidth(resizeColumn) + adjustment;
		if (newWidth < 5) {
			newWidth = 5;
		}

		// Save the new width so that we can determine whether min/max logic
		// or a listener changes it and adjust the resizePoint acccordingly
		originalNewWidth = newWidth;

		if (resizeColumnCellSize != null) {
			newWidth = Math.min(Math.max(newWidth,
										 resizeColumnCellSize.getMinValue()),
								resizeColumnCellSize.getMaxValue());
		}
	}

	// Sends a JCResizeCellEvent object to registered JCResizeCellListener objects.
	if (table.resizeMotionListeners != null) {
		JCResizeCellEvent ev =
		    new JCResizeCellEvent(table,
								  JCResizeCellEvent.RESIZE_DRAG,
			                      resizeRow, resizeColumn,
								  resizeWidth, resizeHeight);
		ev.setNewColumnWidth(newWidth);
		ev.setNewRowHeight(newHeight);
		table.fireJCResizeCellMotionEvent(ev);

		// Reset dimensions on abort
		if (ev.isCancelled()) {
			newHeight = resizeHeight;
			newWidth  = resizeWidth;
			ev.setNewColumnWidth(JCTableEnum.NOVALUE);
			ev.setNewRowHeight(JCTableEnum.NOVALUE);
		}
		else {
			newHeight = ev.getNewRowHeight();
			newWidth = ev.getNewColumnWidth();
		}
	}

	// Reset the resizePoint in case min/max or a listener reset the dimensions
	resizePoint.x += newWidth - originalNewWidth;
	resizePoint.y += newHeight - originalNewHeight;

	EditTraverseHandler editTraverseHandler = table.getEditTraverseHandler();
	CellAreaRenderer car = table.getCellAreaHandler().getCellAreaRenderer();
	if (resizeRow != JCTableEnum.NOVALUE) {
		table.getCellLayout().setHeight(resizeRow, newHeight);

		// Repaint the table if interactive resizing is enabled
		if (table.resize_interactive) {
			int last_row = table.getDataSource().getNumRows() - 1;
			int last_y = (table.getCellLayout().getRowPosition(last_row) +
                          table.getCellLayout().getHeight(last_row));
			if (last_y < table.getSize().height) {
				if (table.store_image_enabled) {
					table.store_image = null;
				}
				table.repaint(0, last_y, table.getSize().width,
							  table.getSize().height - last_y);
			}

			if (editTraverseHandler.getEditRow() == resizeRow) {
				editTraverseHandler.setHeight(newHeight);
			}
			else {
				editTraverseHandler.repositionEditor();
			}

			table.doLayout();

			JCCellRange cr = table.getVisibleCells();
			// if reducing height, not all rows are visible and table is scrolled fully
			// to the bottom, repaint the entire table
			if (newHeight - resizeHeight < 0          &&
			    cr.end_row == table.getNumRows() - 1  &&
			    cr.start_row > table.getFrozenRows())
			{
				table.repaint();
			}
			else {
				car.repaintRange(resizeRow,
								 JCTableEnum.LABEL,
								 table.getDataSource().getNumRows() - 1,
								 table.getDataSource().getNumColumns() - 1);
			}
		}
		else {
			int last_row = table.getDataSource().getNumRows() - 1;
			int last_y = (table.getCellLayout().getRowPosition(last_row) +
				          table.getCellLayout().getHeight(last_row));
			if (last_y < table.getSize().height) {
				// Get the bounds of the labels
				Rectangle r = table.getCellBounds(resizeRow, JCTableEnum.LABEL);
				if (table.store_image_enabled) {
					table.store_image = null;
				}
				table.repaint(r.x, last_y, r.width,
							  table.getSize().height - last_y);
			}
			car.repaintRange(resizeRow,
						     JCTableEnum.LABEL,
							 table.getNumRows() - 1,
							 JCTableEnum.LABEL);
		}
	}

	if (resizeColumn != JCTableEnum.NOVALUE) {
		table.getCellLayout().setWidth(resizeColumn, newWidth);

		// Repaint the table if interactive resizing is enabled
		if (table.resize_interactive) {
			int last_column = table.getDataSource().getNumColumns() - 1;
			int last_x = (table.getCellLayout().getColumnPosition(last_column) +
			              table.getCellLayout().getWidth(last_column));
			if (last_x < table.getSize().width) {
				if (table.store_image_enabled) {
					table.store_image = null;
				}
				table.repaint(last_x, 0, table.getSize().width - last_x,
				              table.getSize().height);
			}

			if (editTraverseHandler.getEditColumn() == resizeColumn) {
				editTraverseHandler.setWidth(newWidth);
			}
			else {
				editTraverseHandler.repositionEditor();
			}

			table.doLayout();

			JCCellRange cr = table.getVisibleCells();
			// If reducing width, not all columns are visible and table is
			// scrolled fully to the right, repaint the entire table.
			if (newWidth - resizeWidth < 0                  &&
			    cr.end_column == table.getNumColumns() - 1  &&
			    cr.start_column > table.getFrozenColumns())
			{
				table.repaint();
			}
			else {
				car.repaintRange(JCTableEnum.LABEL,
								 resizeColumn,
								 table.getDataSource().getNumRows() - 1,
								 table.getDataSource().getNumColumns() - 1);
			}
		}
		else {
			int last_column = table.getDataSource().getNumColumns() - 1;
			int last_x = (table.getCellLayout().getColumnPosition(last_column) +
				          table.getCellLayout().getWidth(last_column));
			if (last_x < table.getSize().width) {
				// Get the bounds of the labels
				Rectangle r = table.getCellBounds(JCTableEnum.LABEL, resizeColumn);
				if (table.store_image_enabled) {
					table.store_image = null;
				}
				table.repaint(last_x, r.y, table.getSize().width - last_x,
				              r.height);
			}
			car.repaintRange(JCTableEnum.LABEL,
			                 resizeColumn,
							 JCTableEnum.LABEL,
							 table.getNumColumns() - 1);
		}
	}
}

/**
 * Completes a resize action.
 * @param x The x position of the resize end
 * @param y The y position of the resize end
 */
protected void resizeEnd(int x, int y)
{
	needs_resize_begin_event = false;
	resize_motion_event = null;

	int newHeight = JCTableEnum.NOVALUE;
	int newWidth = JCTableEnum.NOVALUE;

	boolean isValidResize = false;

	if (resizeRow != JCTableEnum.NOVALUE) {
		int oldHeight = table.getCellLayout().getHeight(resizeRow);
		newHeight = Math.max(5, oldHeight + y - resizePoint.y);
		if (newHeight != resizeHeight) {
			isValidResize = true;
		}
	}
	if (resizeColumn != JCTableEnum.NOVALUE) {
		int oldWidth = table.getCellLayout().getWidth(resizeColumn);
		newWidth = Math.max(5, oldWidth + x - resizePoint.x);
		if (newWidth != resizeWidth) {
			isValidResize = true;
		}
	}

	// If neither the width nor height has changed, abort
	if (!isValidResize) {
		resizeRow    = JCTableEnum.NOVALUE;
		resizeColumn = JCTableEnum.NOVALUE;
		return;
	}

	JCResizeCellEvent ev = null;

	// Sends a JCResizeCellEvent object to registered JCResizeCellListener objects.
	if (table.resizeListeners != null) {
		ev = new JCResizeCellEvent(table, JCResizeCellEvent.RESIZE,
			                       resizeRow, resizeColumn,
								   resizeWidth, resizeHeight);
		ev.setNewColumnWidth(newWidth);
		ev.setNewRowHeight(newHeight);
		table.fireJCResizeCellEvent(ev);

		// reset dimensions on abort
		if (ev.isCancelled()) {
			newHeight = resizeHeight;
			newWidth = resizeWidth;
			ev.setNewColumnWidth(JCTableEnum.NOVALUE);
			ev.setNewRowHeight(JCTableEnum.NOVALUE);
		}
		else {
			newHeight = ev.getNewRowHeight();
			newWidth = ev.getNewColumnWidth();
		}
	}

	if (resizeRowCellSize != null) {
		newHeight = Math.min(Math.max(newHeight,
									  resizeRowCellSize.getMinValue()),
							 resizeRowCellSize.getMaxValue());
	}
	if (resizeColumnCellSize != null) {
		newWidth = Math.min(Math.max(newWidth,
									 resizeColumnCellSize.getMinValue()),
							resizeColumnCellSize.getMaxValue());
	}

	CellAreaRenderer car = table.getCellAreaHandler().getCellAreaRenderer();
	if (resizeRow != JCTableEnum.NOVALUE) {
		if (table.isResizeEven()) {
			table.setPixelHeight(JCTableEnum.ALL, newHeight);
		}
		else {
			table.setPixelHeight(resizeRow, newHeight);
		}

		int last_row = table.getDataSource().getNumRows() - 1;
		int last_y = table.getCellLayout().getRowPosition(last_row) +
                     table.getCellLayout().getHeight(last_row);
		if (last_y < table.getSize().height) {
			// Repaints the dead space if table's height is reduced
			table.repaint(0, last_y, table.getSize().width,
			              table.getSize().height - last_y);
		}
		table.doLayout();
		car.repaintRange(resizeRow, JCTableEnum.LABEL,
		                 table.getDataSource().getNumRows() - 1,
						 table.getDataSource().getNumColumns() - 1);
	}
	if (resizeColumn != JCTableEnum.NOVALUE) {
		if (table.isResizeEven()) {
			table.setPixelWidth(JCTableEnum.ALL, newWidth);
		}
		else {
			table.setPixelWidth(resizeColumn, newWidth);
		}

		int last_column = table.getDataSource().getNumColumns() - 1;
		int last_x = table.getCellLayout().getColumnPosition(last_column) +
			         table.getCellLayout().getWidth(last_column);
		if (last_x < table.getSize().width) {
			// repaints the dead space if table's width is reduced
			table.repaint(last_x, 0, table.getSize().width - last_x,
			              table.getSize().height);
		}
		table.doLayout();
		car.repaintRange(JCTableEnum.LABEL, resizeColumn,
		                 table.getDataSource().getNumRows() - 1,
						 table.getDataSource().getNumColumns() - 1);
	}

	if (table.resizeListeners != null && ev != null) {
		ev.type = JCResizeCellEvent.AFTER_RESIZE;
		table.fireJCResizeCellEvent(ev);
	}

	resizeRow = JCTableEnum.NOVALUE;
	resizeColumn = JCTableEnum.NOVALUE;
}

/**
 * Resets the resize variables.
 */
protected void cancelResize()
{
	RESIZING = false;
	resizeRow = JCTableEnum.NOVALUE;
	resizeColumn = JCTableEnum.NOVALUE;
}

/**
 * Returns true if row auto scroll is enabled.
 * @return Whether row auto scroll is enabled.
 */
protected final boolean isAutoScrollRow()
{
	int v = table.getAutoScroll();
	return(v == JCTableEnum.AUTO_SCROLL_ROW ||
	       v == JCTableEnum.AUTO_SCROLL_BOTH);
}

/**
 * Returns true if column auto scroll is enabled.
 * @return Whether column auto scroll is enabled.
 */
protected final boolean isAutoScrollColumn()
{
	int v = table.getAutoScroll();
	return(v == JCTableEnum.AUTO_SCROLL_COLUMN ||
	       v == JCTableEnum.AUTO_SCROLL_BOTH);
}

/**
 * Returns true if resize is under way.
 * @return Is a resize currently happening?
 */
protected final boolean isResizing()
{
	return(RESIZING);
}

/**
 * Returns true if the cell is potentially traversable.
 * @param row The row index of the cell
 * @param column The column index of the cell
 * @return Whether a cell is a candidate for traversal
 */
protected final boolean isTraversable(int row, int column)
{
	return(row != JCTableEnum.LABEL && column != JCTableEnum.LABEL);
}

/**
 * Returns true if the cell position is potentially valid.
 * This check do not validate against the number of rows or
 * columns in the datasource.
 * @param cell The cell position
 * @return Is the cell position potentially valid?
 */
protected final boolean isValidCellPosition(JCCellPosition cell)
{
	return(!(cell == null || (cell.row == JCTableEnum.NOVALUE ||
			                  cell.column == JCTableEnum.NOVALUE)));
}

/**
 * Traverse the current cell from one location to another, scrolling to make it
 * visible if necessary.
 * @param event The mouse event that caused traverse to be called
 * @param position The cell position of the mouse event
 * @param isPress Is this a button press
 * @return true if the traverse succeeded
 */
protected boolean traverse(MouseEvent event, JCCellPosition position,
                           boolean isPress)
{
	

	// Abort traverse if a resize is in progress
	if (isResizing()) {
		return(false);
	}

	EditTraverseHandler editTraverseHandler = table.getEditTraverseHandler();
	JCCellEditor cellEditor = editTraverseHandler.getCellEditor();
	JCTable curTable = null;
	if (cellEditor != null) {
	    Component component = cellEditor.getComponent();
		if (component != null) {
		    Container parent = component.getParent();
		    while (parent != null) {
			    if (parent instanceof JCTable) {
				    curTable = (JCTable)parent;
				    break;
				}
			    parent = parent.getParent();
			}
		}
	}
	// Don't want to do a commit if mouse click did not happen in the table
	// where the editing began.  Otherwise we are commiting on the other table.
	// Note that this should only happen if a user uses the same cell editor
	// across tables.  By default cell editors are cached per table (see
	// the BaseDataView class for details).
	if (table == curTable && !editTraverseHandler.commit(true)) {
		return(false);
	}

	if (position.row == JCTableEnum.LABEL ||
	    position.column == JCTableEnum.LABEL)
	{
		return(false);
	}

	if (position.row == JCTableEnum.NOVALUE ||
	    position.column == JCTableEnum.NOVALUE)
	{
		return(false);
	}

	return(editTraverseHandler.traverse(position.row, position.column,
		                                false, event,
										JCTableEnum.TRAVERSE_POINTER,
										false));
}

/**
 * Drags the invoking row or column to the new position. The invoking
 * row or column is determined at the time of the initial mouse press.
 * @param current_cell The current cell to which to drag
 */
protected void performDrag(JCCellPosition current_cell)
{
	int to_row    = initial_cell.row;
	int to_column = initial_cell.column;

	table.setCursor(MOVE_CURSOR);

	if (initial_cell.row == JCTableEnum.LABEL) { // column label
		to_column = current_cell.column;
		if (to_column == JCTableEnum.LABEL) {
			return;
		}
		if (current_cell.column > initial_cell.column) {
			to_column++;
		}
		table.cellAreaHandler.performDrag(initial_cell.row,
                                          initial_cell.column,
		                                  to_row, to_column);
	}
	else if (initial_cell.column == JCTableEnum.LABEL) { // row label
		to_row = current_cell.row;
		if (to_row == JCTableEnum.LABEL) {
			return;
		}
		if (current_cell.row > initial_cell.row) {
			to_row++;
		}
		table.cellAreaHandler.performDrag(initial_cell.row,
                                          initial_cell.column,
		                                  to_row, to_column);

	}
}

/**
 * Cancels the drag process if mouse release if over a frozen cell.
 */
protected void cancelDrag()
{
	table.cellAreaHandler.cancelDrag();
}

/**
 * Completes the drag process by actually performing the
 * manipulation. Before endDrag is called, all drag actions
 * are simulated by <CODE>CellAreaHandler</CODE>.
 * @param final_cell The final cell to which to drag
 */
protected void endDrag(JCCellPosition final_cell)
{
	table.cellAreaHandler.endDrag();
	int to_row = final_cell.row;
	if (final_cell.row > initial_cell.row) {
		to_row++;
	}
	int to_column = final_cell.column;
	if (final_cell.column > initial_cell.column) {
		to_column++;
	}

	if (initial_cell.row == JCTableEnum.LABEL) { // column drag
		if (initial_cell.column != final_cell.column) {
			table.dragColumn(initial_cell.column, to_column);
		}
	}
	else {
		if (initial_cell.row != final_cell.row) {
			table.dragRow(initial_cell.row, to_row);
		}
	}
}

/**
 * Starts the auto scroll thread.
 * @param x The x position to start the auto scroll
 * @param y The y position to start the auto scroll
 */
private void startAutoScroll(int x, int y)
{
	auto_scroll_x = x;
	auto_scroll_y = y;

	if ((auto_scroll_vert != AUTO_SCROLL_NONE) ||
	    (auto_scroll_horiz != AUTO_SCROLL_NONE))
	{
		if (auto_scroll_timer == null) {
			auto_scroll_timer = new Timer(10, this);
		}
		if (!auto_scroll_timer.isRunning()) {
			auto_scroll_timer.start();
			AUTOSCROLLING = true;
		}
	}
}

/**
 * Stops the auto scroll thread.
 * @param forced The stop is forced
 */
private void stopAutoScroll(boolean forced)
{
	if (forced || (auto_scroll_horiz == AUTO_SCROLL_NONE &&
	               auto_scroll_vert == AUTO_SCROLL_NONE))
	{
	    if (auto_scroll_timer != null) {
	    	auto_scroll_timer.stop();
	    }
	    AUTOSCROLLING = false;
		auto_scroll_horiz = AUTO_SCROLL_NONE;
		auto_scroll_vert = AUTO_SCROLL_NONE;
	}
}

/**
 * <CODE>ActionPerformed</CODE> is called by the AutoScroll timer in
 * order to make new rows/columns visible when the mouse is dragged
 * out of the cell/label area of table during selection or dragging.
 * @param event The action event generated by the action
 */
public void actionPerformed(ActionEvent event)
{
	int select_row = JCTableEnum.NOVALUE;
	int select_column = JCTableEnum.NOVALUE;

	if (AUTOSCROLLING && (auto_scroll_horiz != AUTO_SCROLL_NONE ||
	                      auto_scroll_vert != AUTO_SCROLL_NONE))
	{
		// if dragging down, make the next row visible
	   	if (auto_scroll_vert == AUTO_SCROLL_INC) {
			int row = table.getVisibleCells().end_row + 1;
			if (row >= table.getNumRows()) {
				auto_scroll_vert = AUTO_SCROLL_NONE;
			}
			else {
				table.makeRowVisible(row);
				if (SELECTING) {
					select_row = row;
					select_column = table.XtoColumn(auto_scroll_x);
				}
				else if(DRAGGING) {
					performDrag(new JCCellPosition(row, JCTableEnum.LABEL));
				}
			}
		}
		// if dragging up, make the row before the top row visible
		else if (auto_scroll_vert == AUTO_SCROLL_DEC) {
			int row = table.getVisibleCells().start_row - 1;
			if (row < 0) {
				auto_scroll_vert = AUTO_SCROLL_NONE;
			}
			else {
				table.makeRowVisible(row);
				if (SELECTING) {
					select_row = row;
					select_column = table.XtoColumn(auto_scroll_x);
				}
				else if(DRAGGING) {
					performDrag(new JCCellPosition(row, JCTableEnum.LABEL));
				}
			}
		}

		// if dragging to the right, make the next column visible
		if (auto_scroll_horiz == AUTO_SCROLL_INC) {
			int column = table.getVisibleCells().end_column + 1;
			if (column >= table.getNumColumns()) {
				auto_scroll_horiz = AUTO_SCROLL_NONE;
			}
			else {
				table.makeColumnVisible(column);
				if (SELECTING) {
					select_row = table.YtoRow(auto_scroll_y);
					select_column = column;
				}
				else if(DRAGGING) {
					performDrag(new JCCellPosition(JCTableEnum.LABEL, column));
				}
			}
		}
		// if dragging to the left, make the column before the leftmost
		// column visible
		else if (auto_scroll_horiz == AUTO_SCROLL_DEC) {
			int column = table.getVisibleCells().start_column - 1;
			if (column < 0) {
				auto_scroll_horiz = AUTO_SCROLL_NONE;
			}
			else {
				table.makeColumnVisible(column);
				if (SELECTING) {
					select_row = table.YtoRow(auto_scroll_y);
					select_column = column;
				}
				else if(DRAGGING) {
					performDrag(new JCCellPosition(JCTableEnum.LABEL, column));
				}
			}
		}

	   	// extend the selection, as required.
		if (SELECTING                          &&
		    table.getSelectionModel() != null  &&
		    select_row != JCTableEnum.NOVALUE  &&
			select_column != JCTableEnum.NOVALUE)
		{
			SELECTING = table.getSelectionModel().select(SelectionModel.EXTEND,
			                                             select_row,
														 select_column);
		}
	}
}

}
