/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: PrintCellLayout.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import com.klg.jclass.table.*;
import java.awt.Rectangle;

/**
 * <CODE>PrintCellLayout</CODE> is a simplified implementation of the 
 * <CODE>CellLayoutModel</CODE> that stores row heights and 
 * column widths during printing.
 */
public class PrintCellLayout implements CellLayoutModel {

protected JCTable          table;
protected int[]            columnWidth;
protected int[]            columnPosition;
protected int[]            rowHeight;
protected int[]            rowPosition;

protected int              labelWidth = JCTableEnum.NOVALUE;
protected int              labelHeight = JCTableEnum.NOVALUE;

protected int              rows = 0;
protected int              columns = 0;

/**
 * Maintains row/column heights and widths for printing.
 */
public PrintCellLayout(int rows, int columns)
{
    rows = Math.max(1, rows);
    this.rows = rows;
    this.columns = columns;

    rowHeight = new int[rows];
    rowPosition = new int[rows];

    for (int r = 0; r < rows; r++) {
        rowHeight[r] = 21;
    }
    calcPositions(rowPosition, labelHeight, rowHeight);

    columnWidth = new int[columns];
    columnPosition = new int[columns];

    for (int c = 0; c < columns; c++) {
        columnWidth[c] = JCTableEnum.NOVALUE;
    }
    calcPositions(columnPosition, labelWidth, columnWidth);
}

protected void calcPositions(int[] position, int label, int[] value)
{
    int length = value.length;

    if (length == 0) {
        return;
    }

    position[0] = label;
    for (int i = 1; i < length; i++) {
        position[i] = position[i - 1] + value[i - 1];
    }
}

/**
 * Returns the JCTable
 */
public JCTable getTable()
{
    return table;
}

/**
 * Sets the JCTable
 */
public void setTable(JCTable table)
{
    this.table = table;
}

/**
 * Returns the height of a row.
 */
public int getHeight(int row)
{
    if (row == JCTableEnum.LABEL) {
        return(labelHeight);
    }
    else {
        return(rowHeight[row]);
    }
}

/**
 * Sets the height of a row.
 */
public void setHeight(int row, int value)
{
    if (row == JCTableEnum.LABEL) {
        labelHeight = Math.max(0, value);
        int frame_border_width = 2 * table.getFrameBorderWidth();
        int offset =
            (table != null) ? table.getColumnLabelOffset() + frame_border_width
                            : 0;
        calcPositions(rowPosition, labelHeight + offset, rowHeight);
    }
    else {
        int adjust = value - rowHeight[row];
        rowHeight[row] = Math.max(0, value);
        for (int r = row + 1; r < rows; r++) {
            rowPosition[r] += adjust;
        }
    }
}

/**
 * Sets the height of a sequence of rows.
 */
public void setHeight(int start_row, int end_row, int value)
{
    for (int i = start_row; i <= end_row; i++) {
        setHeight(i, value);
    }
}

/**
 * Returns the width of a column.
 */
public int getWidth(int column)
{
    if (column == JCTableEnum.LABEL) {
        return(labelWidth);
    }
    else {
        return(columnWidth[column]);
    }
}

/**
 * Sets the width of a column.
 */
public void setWidth(int column, int value)
{
    if (column == JCTableEnum.LABEL) {
        labelWidth = Math.max(0, value);
        int frame_border_width = 2 * table.getFrameBorderWidth();
        int offset =
            (table != null) ? table.getRowLabelOffset() + frame_border_width
                            : 0;
        calcPositions(columnPosition, labelWidth + offset, columnWidth);
    }
    else {
        int adjust = value - columnWidth[column];
        columnWidth[column] = Math.max(0, value);
        for (int c = column + 1; c < columns; c++) {
            columnPosition[c] += adjust;
        }
    }
}

/**
 * Sets the height of a sequence of column.
 */
public void setWidth(int start_column, int end_column, int value)
{
    for (int i = start_column; i <= end_column; i++) {
        setWidth(i, value);
    }
}

/**
 * Reutrns the start position of a row.
 */
public int getRowPosition(int row)
{
    if (row == JCTableEnum.LABEL) {
        return(0);
    }
    else {
        return(rowPosition[row]);
    }
}

/**
 * Returns the start position of a column.
 */
public int getColumnPosition(int column)
{
    if (column == JCTableEnum.LABEL) {
        return(0);
    }
    else {
        return(columnPosition[column]);
    }
}

/**
 * Returns the bounds of the given cell or label.
 */
public Rectangle getBounds(int row, int column)
{
    return(getBounds(row, column, new Rectangle()));
}

/**
 * Gets the bounds of the given cell or label using the given Rectangle.
 */
public Rectangle getBounds(int row, int column, Rectangle r)
{
    r.setBounds(getColumnPosition(column), getRowPosition(row),
                getWidth(column), getHeight(row));
    return(r);
}

/**
 * Returns the bounds of a cell range.
 */
public Rectangle getBounds(int start_row, int start_column,
                           int end_row, int end_column)
{
    return(getBounds(start_row, start_column, end_row, end_column,
                     new Rectangle()));
}

/**
 * Gets the bounds of a cell range using the given Rectangle
 */
public Rectangle getBounds(int start_row, int start_column,
                           int end_row, int end_column, Rectangle r)
{
    int x = getColumnPosition(start_column);
    int y = getRowPosition(start_row);

    r.setBounds(x,
                y,
                getColumnPosition(end_column) + getWidth(end_column) - x,
                getRowPosition(end_row) + getHeight(end_row) - y);
    return(r);
}

/**
 * Returns the column at the given x,y coordinates.
 */
public int getColumn(int x, int y)
{
    if (x < labelWidth) {
        return(-1);
    }

    int c = columns - 1;
    for (; c > 0; c--) {
        if (x >= columnPosition[c]) {
            break;
        }
    }
    return(c);
}

/**
 * Returns the row at the given x,y coordinates.
 */
public int getRow(int x, int y)
{
    if (y < labelHeight) {
        return(-1);
    }

    int r = rows - 1;
    for (; r > 0; r--) {
        if (y >= rowPosition[r]) {
            break;
        }
    }
    return(r);
}

public void addRows(int start_row, int num_rows) {}
public void addColumns(int start_column, int num_columns) {}
public void moveRows(int start_row, int num_rows, int destination) {}
public void moveColumns(int start_column, int num_columns, int destination) {}
public void deleteRows(int start_row, int num_rows) {}
public void deleteColumns(int start_column, int num_columns) {}
public void remapRows(int current_map[], int new_map[]) {}
public void remapColumns(int current_map[], int new_map[]) {}
public void resetRows() {}
public void resetColumns() {}


}
