/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: SpanHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

/**
 * <CODE>SpanHandler</CODE> controls the creation, validation, removal and
 * repaint of spanned cell ranges.
 */
public class SpanHandler implements Serializable {

protected JCTable      table;
protected Vector       spans;

/**
 * Creates a new SpanHandler
 */
public SpanHandler() {}

public void setTable(JCTable table)
{
	this.table = table;
}

/**
 * Adds a new JCCellRange to the spanned list. Throws an
 * IllegalArgumentException if the cell range overlaps an existing spanned
 * cell range.
 * @exception java.lang.IllegalArgumentException thrown if a cell range overlaps
 * and existing cell range or crosses over scrollable/non-scrollable regions.
 */
public void addSpannedRange(JCCellRange cr) throws IllegalArgumentException
{
	// Verify that the cell range doesn't overlap an existing span
	// if there are no spanned cell ranges, accept the new range
	if (spans == null) {
		spans = new Vector();
	}

	// Verifies that no range overlaps or crosses scrollable regions
	validateRange(spans, cr);

	// Remove non-controlling components
	ComponentModel componentHandler = table.getComponentModel();
	if (componentHandler != null) {
		componentHandler.removeCellComponents();
	}

	// Add the JCCellRange to the vector of spans
	spans.addElement(cr);
}

/**
 * Validates a cell ranges against the current list of spanned
 * cell ranges. Validation will fail if the cell range overlaps
 * an existing cell range or crosses over scrollable/non-scrollabel
 * regions.
 * @exception java.lang.IllegalArgumentException thrown if a cell range overlaps
 * and existing cell range or crosses over scrollable/non-scrollable regions.
 */
protected void validateRange(Vector spans, JCCellRange cr)
    throws IllegalArgumentException
{
	// Spans must start at upper left
	if (cr.start_row > cr.end_row || cr.start_column > cr.end_column) {
		String s = "End row/column must be greater than or equal to start " +
                   "row/column. " + cr;
		throw new IllegalArgumentException(s);
	}

	// Spans cannot contain both labels and cells
	if ((cr.start_row == JCTableEnum.LABEL &&
         cr.end_row != JCTableEnum.LABEL) ||
	    (cr.start_column == JCTableEnum.LABEL &&
         cr.end_column != JCTableEnum.LABEL))
    {
			throw new IllegalArgumentException("cell range " + cr +
			                                   " contains labels and cells");
	}

	// Loops through each range and checks if the new range overlaps
	// or crosses over scrollable/non-scrollable regions
	JCCellRange range;
	for (int i = 0, cnt = spans.size(); i < cnt; i++) {
		range = (JCCellRange)spans.elementAt(i);
		if (range.overlaps(cr)) {
			throw new IllegalArgumentException("cell range " + cr +
			                                   " overlaps an existing span");
		}
	}

	int frow = table.getFrozenRows();
	int fcol = table.getFrozenColumns();

    String message = "cell range " + cr +
                     " contains frozen and non-frozen cells";
    if (frow > 0 && cr.start_row < frow && cr.end_row >= frow) {
		throw new IllegalArgumentException(message);
	}
	if (fcol > 0 && cr.start_column < fcol && cr.end_column >= fcol) {
		throw new IllegalArgumentException(message);
	}
}

/**
 * Ensures that the given cell range includes all cells that are
 * part of spans.
 * @return JCCellRange a complete cell range
 */
public JCCellRange getCompleteRange(int start_row, int start_column,
                                    int end_row, int end_column)
{
	if (spans != null && spans.size() > 0) {
		for (int i = 0; i < spans.size(); i++) {
			JCCellRange cr = (JCCellRange)spans.elementAt(i);

			if (cr.start_row < start_row) {
				start_row = cr.start_row;
			}
			if (cr.end_row > end_row) {
				end_row = cr.end_row;
			}
			if (cr.start_column < start_column) {
				start_column = cr.start_column;
			}
			if (cr.end_column > end_column) {
				end_column = cr.end_column;
			}
		}
	}

	return(new JCCellRange(start_row, start_column, end_row, end_column));
}

/**
 * Reprocesses all cell ranges to ensure that none overlap or cross
 * scrollable/non-scrollable regions. Validation may fail if
 * the number of frozen rows or columns changes.
 * @exception java.lang.IllegalArgumentException thrown if a cell range overlaps
 * and existing cell range or crosses over scrollable/non-scrollable regions.
 */
public void validateAllRanges() throws IllegalArgumentException
{
	// Abort if there are no spans
	if (spans == null || spans.size() == 0) {
		return;
	}

	boolean validated = true;

	Vector v = new Vector();
	for (int i = 0; i < spans.size(); i++) {
		JCCellRange cr = (JCCellRange)spans.elementAt(i);
		try {
			validateRange(v, cr);
			v.addElement(cr);
		}
		catch (IllegalArgumentException iae) {
			validated = false;
		}
	}
	spans = v;

	if (!validated) {
		throw new IllegalArgumentException("one or more spanned cell ranges " +
                                           "failed validation");
	}
}

/**
 * Removes all spanned cell ranges.
 */
public void clearSpannedRanges()
{
	// aborts if there are no spans
	if (spans == null || spans.size() == 0) {
		return;
	}

	// loop through each range and remove it
	while (spans.size() > 0) {
		removeSpannedRange((JCCellRange)spans.elementAt(0));
	}
}

/**
 * Returns a Vector containing all of the spanned cell ranges.
 */
public Collection getSpannedRanges()
{
	if (spans == null) {
		return(null);
	}

	return((Vector)spans.clone());
}

/**
 * Removes a cell range from the list of spanned cell ranges.
 */
public void removeSpannedRange(JCCellRange cr)
{
	if (spans == null) {
		return;
	}

	// loops through the spanned ranges and remove one that matches
	JCCellRange range;
	for (int i = 0, cnt = spans.size(); i < cnt; i++) {
		range = (JCCellRange)spans.elementAt(i);
		if (range.equals(cr)) {
			spans.removeElementAt(i);
			return;
		}
	}
}

/**
 * Sets the spanned cell ranges to a Collection of JCCellRanges. Ranges are
 * validated in order of iteration in the Collection.
 * @exception java.lang.IllegalArgumentException thrown if a cell range overlaps
 * and existing cell range or crosses over scrollable/non-scrollable regions.
 */
public void setSpannedRanges(Collection ranges) throws IllegalArgumentException
{
	spans = null;

	// loops through each range & adds it with validation to the vector of spans
	Iterator it = ranges.iterator();
	while (it.hasNext()) {
		addSpannedRange((JCCellRange)it.next());
	}
}

/**
 * Returns a JCCellRange containing the given row and column. If the cell is
 * not spanned, null is returned.
 */
public JCCellRange getSpannedRange(int row, int column)
{
	if (spans == null) {
		return(null);
	}

	// loops through each range and checks if the given row and column exist
	// in the range
	JCCellRange range;
	for (int i = 0, cnt = spans.size(); i < cnt; i++) {
		range = (JCCellRange)spans.elementAt(i);
		if (range.inside(row, column)) {
			return(range);
		}
	}
	return(null);
}

/**
 * Returns true if the given cell is the upper-left-most non-hidden cell
 * in a span.
 */
public boolean isControllingCell(JCCellRange cr, int row, int column)
{
	boolean isControlCell = false;

	// Cell has to be visible to be a candidate.
	if (cellIsNotHidden(row, column)) {

		// Assume it is the control cell.
		isControlCell = true;

		// Now look for a cell higher and/or more left in the span which is also
		// visible. If one is found, then this cell is not the control cell.
		// Note: non-contiguous rows/columns in a spanned range can be hidden.
		for (int i = cr.start_row; i < row + 1 && isControlCell; i++) {
			for (int j = cr.start_column; j < column + 1; j++) {
				if (cellIsNotHidden(i, j)) {
					if (i < row || j < column) {
						isControlCell = false;
						break;
					}
				}
			}
		}
	}

	return(isControlCell);
}

// Return true if this cell is not hidden.
private boolean cellIsNotHidden(int row, int column)
{
	return(!table.isRowHidden(row) && !table.isColumnHidden(column));
}

}
