/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: Table.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.awt.Color;
import java.awt.Component;
import java.awt.Font;

import com.klg.jclass.cell.JCCellRenderer;
import com.klg.jclass.cell.JCCellEditor;

/**
 * Compatability class. <p>
 * Table is the JClass LiveTable 3.* compatability class
 * for JClass LiveTable 4.*.  The easiest way to migrate JClass LiveTable 3.*
 * code to JClass LiveTable 4.* is to use this class; documentation is
 * provided in the JClass LiveTable Programmer's Guide.
 *<p>
 * Since JClass LiveTable 3.* and JClass LiveTable 4.* use different
 * property setting mechanisms, using the Table class
 * is inefficient.  Table tries to map 3.*-style
 * calls to LiveTable 4.* style-based property setting.
 * The result is typically a lot of extra styles.
 *<p>
 * <B>Note:</b> JClass LiveTable 2.* compatability is no longer
 * available. It simply was no longer practical to
 * support the container-based model of JClass LiveTable 2.*.
 */
public class Table extends JCTable {

/**
 * Default constructor.
 */
public Table()
{
    super();
}

/**
 * Retreives the alignment enumeration for
 * a particular cell.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return alignment enum; for example, <code>JCTableEnum.BOTTOMLEFT</code>
 */
public int getAlignment(int row, int column)
{
    int ha = getCellStyle(row, column).getHorizontalAlignment();
    int va = getCellStyle(row, column).getVerticalAlignment();

    if (ha == JCTableEnum.LEFT) {
        if (va == JCTableEnum.TOP) {
            return JCTableEnum.TOPLEFT;
        }
        else if (va == JCTableEnum.CENTER) {
            return JCTableEnum.MIDDLELEFT;
        }
        else {
            return JCTableEnum.BOTTOMLEFT;
        }
    }
    else if (ha == JCTableEnum.CENTER) {
        if (va == JCTableEnum.TOP) {
            return JCTableEnum.TOPCENTER;
        }
        else if (va == JCTableEnum.CENTER) {
            return JCTableEnum.MIDDLECENTER;
        }
        else {
            return JCTableEnum.BOTTOMCENTER;
        }
    }
    else {
        if (va == JCTableEnum.TOP) {
            return JCTableEnum.TOPRIGHT;
        }
        else if (va == JCTableEnum.CENTER) {
            return JCTableEnum.MIDDLERIGHT;
        }
        else {
            return JCTableEnum.BOTTOMRIGHT;
        }
    }
}

/**
 * Sets the alignment for a particular cell.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value alignment enum; for example, <code>JCTableEnum.BOTTOMLEFT</code>
 */
public void setAlignment(int row, int column, int value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    switch (value) {
    case JCTableEnum.TOPLEFT:
        csm.setHorizontalAlignment(JCTableEnum.LEFT);
        csm.setVerticalAlignment(JCTableEnum.TOP);
        break;
    case JCTableEnum.TOPCENTER:
        csm.setHorizontalAlignment(JCTableEnum.CENTER);
        csm.setVerticalAlignment(JCTableEnum.TOP);
        break;
    case JCTableEnum.TOPRIGHT:
        csm.setHorizontalAlignment(JCTableEnum.RIGHT);
        csm.setVerticalAlignment(JCTableEnum.TOP);
        break;
    case JCTableEnum.MIDDLELEFT:
        csm.setHorizontalAlignment(JCTableEnum.LEFT);
        csm.setVerticalAlignment(JCTableEnum.CENTER);
        break;
    case JCTableEnum.MIDDLECENTER:
        csm.setHorizontalAlignment(JCTableEnum.CENTER);
        csm.setVerticalAlignment(JCTableEnum.CENTER);
        break;
    case JCTableEnum.MIDDLERIGHT:
        csm.setHorizontalAlignment(JCTableEnum.RIGHT);
        csm.setVerticalAlignment(JCTableEnum.CENTER);
        break;
    case JCTableEnum.BOTTOMLEFT:
        csm.setHorizontalAlignment(JCTableEnum.LEFT);
        csm.setVerticalAlignment(JCTableEnum.BOTTOM);
        break;
    case JCTableEnum.BOTTOMCENTER:
        csm.setHorizontalAlignment(JCTableEnum.CENTER);
        csm.setVerticalAlignment(JCTableEnum.BOTTOM);
        break;
    case JCTableEnum.BOTTOMRIGHT:
        csm.setHorizontalAlignment(JCTableEnum.RIGHT);
        csm.setVerticalAlignment(JCTableEnum.BOTTOM);
        break;
    }
    setCellStyle(row, column, csm);
}

/**
 * Sets the alignment for a range of cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @param value alignment enum; for example, <code>JCTableEnum.BOTTOMLEFT</code>
 * @see Table#setAlignment
 */
public void setAlignment(JCCellRange cr, int value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setAlignment(r, c, value);
        }
    }
}

/**
 * Returns the background value for a cell or label.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return background color for desired cell
 * @see Table#setBackground
 */
public Color getBackground(int row, int column)
{
    return getCellStyle(row, column).getBackground();
}

/**
 * Sets the background color of cells and labels. When set to null,
 * the color is inherited from the Component's background value.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value new background color for specified cell
 * @see Table#getBackground
 */
public void setBackground(int row, int column, Color value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setBackground(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the background color for a range of cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setBackground
 * @param value
 */
public void setBackground(JCCellRange cr, Color value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setBackground(r, c, value);
        }
    }
}

/**
 * Gets the cell's <CODE>CellRenderer</CODE> value.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see #setCellRenderer
 * @return the cell's <code>CellRenderer</code> value
 */
public JCCellRenderer getCellRenderer(int row, int column)
{
    return getCellStyle(row, column).getCellRenderer();
}

/**
 * Sets the renderer for a cell.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param renderer
 */
public void setCellRenderer(int row, int column, JCCellRenderer renderer)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setCellRenderer(renderer);
    setCellStyle(row, column, csm);
}

/**
 * Sets the renderer for a range of cells.
 *
 * @param cr range of cells (<B>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setCellRenderer
 * @param renderer
 */
public void setCellRenderer(JCCellRange cr, JCCellRenderer renderer)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setCellRenderer(r, c, renderer);
        }
    }
}


/**
 * Gets the cell's <CODE>CellEditor</CODE> value.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see #setCellEditor
 * @return the cell's <code>CellEditor</code> value
 */
public JCCellEditor getCellEditor(int row, int column)
{
    return getCellStyle(row, column).getCellEditor();
}

/**
 * Sets the editor for a cell.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param editor
 */
public void setCellEditor(int row, int column, JCCellEditor editor)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setCellEditor(editor);
    setCellStyle(row, column, csm);
}

/**
 * Sets the editor for a range of cells.
 *
 * @param cr range of cells (<B>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setCellEditor
 * @param editor
 */
public void setCellEditor(JCCellRange cr, JCCellEditor editor)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setCellEditor(r, c, editor);
        }
    }
}


/**
 * Retrieves the cell border side enum for a particular cell.
 * This method calls <code>getCellBorderSizes()</code>.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return Cell border sides enum; for example,
 * <code>JCTableEnum.BORDERSIDE_ALL</code>
 * @see #getCellBorderSides
 */
public int getBorderSides(int row, int column)
{
    return getCellBorderSides(row, column);
}

/**
 * Sets the cell border side enum for a particular cell.
 * This method calls <code>setCellBorderSides()</code>.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value cell border sides enum; for example,
 * <code>JCTableEnum.BORDERSIDE_ALL</code>
 * @see #setCellBorderSides
 */
public void setBorderSides(int row, int column, int value)
{
    setCellBorderSides(row, column, value);
}

/**
 * Gets the broder type for a particular cell.
 * This methods calls <code>getCellBorderType()</code>.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return cell border type enum; for example,
 * <code>JCTableEnum.BORDER_ETCHED_IN</code>
 * @see #getCellBorderType
 */
public int getBorderType(int row, int column)
{
    JCCellBorder border = (JCCellBorder)getCellBorderType(row, column);
    return border.getBorderType();
}

/**
 * Sets the broder type for a particular cell.
 * This methods calls <code>setCellBorderType()</code>.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value cell border type enum; for example,
 * <code>JCTableEnum.BORDER_ETCHED_IN</code>
 * @see #setCellBorderType
 */
public void setBorderType(int row, int column, int value)
{
    setCellBorderType(row, column, value);
}

/**
 * Returns the <code>CellBorderSides</code> value for a cell or label.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return cell border sides enum; for example,
 * <code>JCTableEnum.BORDERSIDE_ALL</code>
 * @see Table#setCellBorderSides
 */
public int getCellBorderSides(int row, int column)
{
    return getCellStyle(row, column).getCellBorderSides();
}

/**
 * Sets the sides of a cell or label that display the border type in their
 * border area. The value is a bitwise OR of the following:
 * <UL><LI><CODE>JCTableEnum.BORDERSIDE_NONE</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDERSIDE_ALL</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDERSIDE_LEFT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDERSIDE_RIGHT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDERSIDE_TOP</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDERSIDE_BOTTOM</CODE></LI></UL><P>
 * The width of all cell/label borders is specified by
 * <CODE>CellBorderWidth</CODE>.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value cell border sides enum; for example,
 * <code>JCTableEnum.BORDERSIDE_ALL</code>
 * @see Table#getCellBorderSides
 * @see #setCellBorderWidth
 */
public void setCellBorderSides(int row, int column, int value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setCellBorderSides(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the sides of a cell or label for a range of cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @param value cell border sides enum; for example,
 * <code>JCTableEnum.BORDERSIDE_ALL</code>
 * @see Table#setCellBorderSides
 */
public void setCellBorderSides(JCCellRange cr, int value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setCellBorderSides(r, c, value);
        }
    }
}

/**
 * Returns the <code>cellBorderType</code> value for a cell or label.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return cell border type enum; for example, <code>JCTableEnum.BORDER_ETCHED_IN</code>
 * @see Table#setCellBorderType
 */
public CellBorderModel getCellBorderType(int row, int column)
{
    return getCellStyle(row, column).getCellBorder();
}

/**
 * Sets the custom style used for cell and label borders. Valid values are
 * objects that implement the <code>CellBorder</code> interface.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value object implementing <code>CellBorderModel</code> (in other words,
 * a real Swing border that LiveTable can understand)
 */
public void setCellBorderType(int row, int column, CellBorderModel value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setCellBorder(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the custom style used for cell and label borders as an
 * enum.  The enums are translated to an actual border
 * using the <code>JCCellBorder</code> object.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param type valid border type enum, one of
 * <UL><LI><CODE>JCTableEnum.BORDER_NONE</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_ETCHED_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_ETCHED_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_PLAIN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_FRAME_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_FRAME_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_THIN</CODE></LI>
 * </UL><P>
 */
public void setCellBorderType(int row, int column, int type)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setCellBorder(new JCCellBorder(type));
    setCellStyle(row, column, csm);
}

/**
 * Sets the custom style used for cell and label borders for a range of
 * cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @param value object implementing <code>CellBorderModel</code> (in other words,
 * a real Swing border that LiveTable can understand)
 */
public void setCellBorderType(JCCellRange cr, CellBorderModel value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setCellBorderType(r, c, value);
        }
    }
}

/**
 * Returns <code>true</code> if column label display is enabled,
 * <code>false</code> otherwise.  This method maps to
 * <code>JCTable.isColumnLabelDisplay()</code>.
 *
 * @see JCTable#isColumnLabelDisplay
 * @return <code>true</code> if column label display is enabled,
 * <code>false</code> otherwise
 */
public boolean getColumnLabelDisplay()
{
    return isColumnLabelDisplay();
}

/**
 * Returns <code>true</code> if double buffering is enabled,
 * <code>false</code> otherwise.  This method maps to
 * <code>JComponent.isDoubleBuffered()</code>.
 *
 * @see javax.swing.JComponent#isDoubleBuffered
 * @return <code>true</code> if double buffering is enabled,
 * <code>false</code> otherwise
 */
public boolean getDoubleBuffer()
{
    return isDoubleBuffered();
}

/**
 * Sets the double buffering value for table.
 * This method maps to <code>JComponent.setDoubleBuffer()</code>.
 *
 * @see javax.swing.JComponent#setDoubleBuffered
 * @param value
 */
public void setDoubleBuffer(boolean value)
{
    setDoubleBuffered(value);
}

/**
 * Returns the editable value for a cell.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see #setEditable
 * @return
 */
public boolean getEditable(int row, int column)
{
    return getCellStyle(row, column).isEditable();
}

/**
 * Sets whether the cell can be edited (default: <CODE>true</CODE>).
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value
 */
public void setEditable(int row, int column, boolean value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setEditable(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets whether a range of cells can be edited.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setEditable
 * @param value
 */
public void setEditable(JCCellRange cr, boolean value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setEditable(r, c, value);
        }
    }
}

/**
 * Gets the focus rectangle color.
 * This methods calls <code>JCTable.getFocusColor()</code>.
 *
 * @see JCTable#getFocusColor
 * @return the focus rectangle color
 */
public Color getFocusRectColor()
{
    return getFocusColor();
}

/**
 * Sets the focus rectangle color
 * This methods calls <code>JCTable.setFocusColor()</code>.
 *
 * @see JCTable#setFocusColor
 * @param value
 */
public void setFocusRectColor(Color value)
{
    setFocusColor(value);
}

/**
 * Returns the font value for a cell or label.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see Table#setFont
 * @return the font value for a cell or label
 */
public Font getFont(int row, int column)
{
    return getCellStyle(row, column).getFont();
}

/**
 * Sets the font to be used to display cell and/or label values.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value
 */
public void setFont(int row, int column, Font value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setFont(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the font for a range of cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setFont
 * @param value
 */
public void setFont(JCCellRange cr, Font value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setFont(r, c, value);
        }
    }
}

/**
 * Returns the foreground value for a cell or label.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see Table#setForeground
 * @return the foreground value for a cell or label
 */
public Color getForeground(int row, int column)
{
    return getCellStyle(row, column).getForeground();
}

/**
 * Sets the foreground color of cells and labels. When set to null
 * the color is inherited from Component's foreground value.
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see Table#getForeground
 * @param value
 */
public void setForeground(int row, int column, Color value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setForeground(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the foreground color for a range of cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setForeground
 * @param value
 */
public void setForeground(JCCellRange cr, Color value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setForeground(r, c, value);
        }
    }
}

/**
 * Gets the frame shadow thickness.  This
 * method calls <code>JCTable.getFrameBorderWidth()</code>.
 *
 * @see JCTable#getFrameBorderWidth
 * @return frame shadow thickness
 */
public int getFrameShadowThickness()
{
    return getFrameBorderWidth();
}

/**
 * Sets the frame shadow thickness.  This
 * method calls <code>JCTable.setFrameBorderWidth()</code>.
 *
 * @see JCTable#setFrameBorderWidth
 * @param value
 */
public void setFrameShadowThickness(int value)
{
    setFrameBorderWidth(value);
}

/**
 * Gets the frame border type as an enum value.
 *
 * @return Valid border type enum, one of
 * <UL><LI><CODE>JCTableEnum.BORDER_NONE</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_ETCHED_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_ETCHED_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_PLAIN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_FRAME_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_FRAME_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_THIN</CODE></LI>
 * </UL><P>
 */
public int getFrameBorderType()
{
    JCCellBorder border = (JCCellBorder)getFrameBorder();
    return border.getBorderType();
}

/**
 * Sets the frame border type as an enum value.
 *
 * @param value Valid border type enum, one of
 * <UL><LI><CODE>JCTableEnum.BORDER_NONE</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_ETCHED_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_ETCHED_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_PLAIN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_FRAME_IN</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_FRAME_OUT</CODE></LI>
 * <LI><CODE>JCTableEnum.BORDER_THIN</CODE></LI>
 * </UL><P>
 */
public void setFrameBorderType(int value)
{
    setFrameBorder(new JCCellBorder(value));
}

/**
 * Returns the horizontal alignment value for a cell or label.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see Table#setHorizontalAlignment
 * @return the horizontal alignment value for a cell or label
 */
public int getHorizontalAlignment(int row, int column)
{
    return getCellStyle(row, column).getHorizontalAlignment();
}

/**
 * Sets the horizontal alignment of cell/label text and/or image. Valid values are:
 * <UL><LI><CODE>JCTableEnum.LEFT</CODE></LI>
 * <LI><CODE>JCTableEnum.CENTER</CODE></LI>
 * <LI><CODE>JCTableEnum.RIGHT</CODE></LI></UL>
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value
 */
public void setHorizontalAlignment(int row, int column, int value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setHorizontalAlignment(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the horizontal alignment for a range of cells.
 *
 * @param cr range of cells (<b>Note:</b> if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setHorizontalAlignment
 * @param value
 */
public void setHorizontalAlignment(JCCellRange cr, int value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setHorizontalAlignment(r, c, value);
        }
    }
}

/**
 * Returns <code>true</code> if repaint is enabled, <code>false</code> otherwise.
 * This method calls <code>JCTable.isRepaintEnabled()</code>.
 *
 * @see JCTable#isRepaintEnabled
 * @return <code>true</code> if repaint is enabled, <code>false</code> otherwise
 */
public boolean getRepaint()
{
    return isRepaintEnabled();
}

/**
 * Disables repainting if set to <code>false</code>; otherwise,
 * enables table repainting.  Use this method to change
 * a lot of properties without forcing repainting.
 * This method calls <code>JCTable.setRepaintEnabled()</code>.
 *
 * @see JCTable#setRepaintEnabled
 * @param value
 */
public void setRepaint(boolean value)
{
    setRepaintEnabled(value);
}

/**
 * Returns <code>true</code> if cell resizing can happen only using
 * column or row labels, <code>false</code> otherwise.  This method
 * uses <code>JCTable.getAllowResizeBy()</code>.
 *
 * @see JCTable#getAllowResizeBy
 * @return <code>true</code> if cell resizing can happen only using
 * column or row labels, <code>false</code> otherwise
 */
public boolean getResizeByLabelsOnly()
{
    return(!(getAllowResizeBy() == JCTableEnum.RESIZE_BY_ALL));
}

/**
 * Determines whether cell resizing is limited to column
 * and row labels, or can happen on any cell.
 * This method uses <code>JCTable.setAllowResizeBy()</code>.
 *
 * @see JCTable#setAllowResizeBy
 * @param value
 */
public void setResizeByLabelsOnly(boolean value)
{
    if (value) {
        setAllowResizeBy(JCTableEnum.RESIZE_BY_LABELS);
    }
    else {
        setAllowResizeBy(JCTableEnum.RESIZE_BY_ALL);
    }
}

/**
 * Returns <code>true</code> if row label display is enabled,
 * <code>false</code> otherwise.  This method maps to
 * <code>JCTable.isRowLabelDisplay()</code>.
 *
 * @see JCTable#isRowLabelDisplay
 * @return <code>true</code> if row label display is enabled,
 * <code>false</code> otherwise
 */
public boolean getRowLabelDisplay()
{
    return isRowLabelDisplay();
}

/**
 * Gets the shadown thickness, which is the thickness
 * of all cell borders.  This method maps to
 * <code>JCTable.getCellBorderWidth()</code>.
 *
 * @see JCTable#getCellBorderWidth
 * @return shadow thickness
 */
public int getShadowThickness()
{
    return getCellBorderWidth();
}

/**
 * Sets the shadown thickness, which is the thickness
 * of all cell borders.  This method maps to
 * <code>JCTable.setCellBorderWidth()</code>.
 *
 * @see JCTable#setCellBorderWidth
 * @param value
 */
public void setShadowThickness(int value)
{
    setCellBorderWidth(value);
}

/**
 * Returns <code>true</code> if cursor tracking is enabled, <code>false</code>
 * otherwise.  Cursor tracking determines whether the
 * cursor changes as it moves over cell borders.  This
 * method maps to <code>JCTable.isTrackCursor()</code>.
 *
 * @see JCTable#isTrackCursor
 * @return <code>true</code> if cursor tracking is enabled, <code>false</code>
 * otherwise
 */
public boolean getTrackCursor()
{
    return isTrackCursor();
}

/**
 * Returns <code>true</code> if traversal will cycle on the
 * same line of cells (same row), <code>false</code> otherwise.
 * This method maps to <code>JCtable.isTraverseCycle()</code>.
 *
 * @see JCTable#isTraverseCycle
 * @return <code>true</code> if traversal will cycle on the
 * same line of cells (same row), <code>false</code> otherwise
 */
public boolean getTraverseCycle()
{
    return isTraverseCycle();
}

/**
 * Returns the traversable value for a cell.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see #setTraversable
 * @return the traversable value for a cell
 */
public boolean getTraversable(int row, int column)
{
    return getCellStyle(row, column).isTraversable();
}

/**
 * Sets whether the cell can be traversed to (default: <CODE>true</CODE>).
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value
 */
public void setTraversable(int row, int column, boolean value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setTraversable(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets whether the cell can be traversed to for a range of cells.
 *
 * @param cr range of cells (<B>Note:</b> that if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setTraversable
 * @param value
 */
public void setTraversable(JCCellRange cr, boolean value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setTraversable(r, c, value);
        }
    }
}

/**
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @return
 */
public Object getUserdata(int row, int column)
{
    return getUserData(row, column);
}

/**
 * Returns the vertical alignment value for a cell or label.
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @see Table#setVerticalAlignment
 * @return vertical alignment value for a cell or label
 */
public int getVerticalAlignment(int row, int column)
{
    return getCellStyle(row, column).getVerticalAlignment();
}

/**
 * Sets the vertical alignment of cell/label text and/or image. Valid values are:
 * <UL><LI><CODE>JCTableEnum.TOP</CODE></LI>
 * <LI><CODE>JCTableEnum.CENTER</CODE></LI>
 * <LI><CODE>JCTableEnum.BOTTOM</CODE></LI></UL>
 *
 * @param row row index for desired cell (starts at 0)
 * @param column column index for desired cell (starts at 0)
 * @param value
 */
public void setVerticalAlignment(int row, int column, int value)
{
    CellStyleModel csm = getUniqueCellStyle(row, column);
    csm.setVerticalAlignment(value);
    setCellStyle(row, column, csm);
}

/**
 * Sets the vertical alignment for a range of cells.
 *
 * @param cr range of cells (<B>Note:</b> that if ALL or
 * ALLCELLS is used for <code>end_row</code> or <code>end_column</code> in
 * cr, it will be translated to the actual number of
 * rows or columns in the table; this means that
 * you must attach the data source before using
 * ALL or ALLCELLS in order for the call to work)
 * @see Table#setVerticalAlignment
 * @param value
 */
public void setVerticalAlignment(JCCellRange cr, int value)
{
    adjustCellRange(cr);
    for (int r = cr.start_row; r <= cr.end_row; r++) {
        for (int c = cr.start_column; c <= cr.end_column; c++) {
            setVerticalAlignment(r, c, value);
        }
    }
}

/**
 * Gets the value of the <code>ClipArrows</code> property.
 * The <code>ClipArrows</code> property determines whether text clip arrows
 * are shown when the contents of the cell do not fit in the cell
 * frame.
 * @return One of <CODE>JCTableEnum.CLIP_ARROWS_BOTH</CODE>,
 * <CODE>JCTableEnum.CLIP_ARROWS_RIGHT</CODE>,
 * <CODE>JCTableEnum.CLIP_ARROWS_DOWN</CODE>, or
 * <CODE>JCTableEnum.CLIP_ARROWS_NONE</CODE>,
 * @see #setClipArrows
 */
public int getClipArrows()
{
    return getCellStyle(JCTableEnum.ALLCELLS,
                        JCTableEnum.ALLCELLS).getClipHints();
}

/**
 * Sets the value of the <code>ClipArrows</code> property.
 * The <code>ClipArrows</code> property determines whether text clip arrows
 * are shown when the contents of the cell do not fit in the cell
 * frame.
 * @param value one of <CODE>JCTableEnum.CLIP_ARROWS_BOTH</CODE>,
 * <CODE>JCTableEnum.CLIP_ARROWS_RIGHT</CODE>,
 * <CODE>JCTableEnum.CLIP_ARROWS_DOWN</CODE>, or
 * <CODE>JCTableEnum.CLIP_ARROWS_NONE</CODE>,
 */
public void setClipArrows(int value)
{
    CellStyleModel csm = getUniqueCellStyle(JCTableEnum.ALLCELLS,
                                            JCTableEnum.ALLCELLS);
    csm.setClipHints(value);
    setCellStyle(JCTableEnum.ALLCELLS, JCTableEnum.ALLCELLS, csm);
}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @return
 */
public int getDatatype(int row, int column)
{
    return(0);
}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @param value
 */
public void setDatatype(int row, int column, int value) {}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @return
 */
public int getMaxLength(int row, int column)
{
    return(JCTableEnum.MAXINT);
}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @param value
 */
public void setMaxLength(int row, int column, int value) {}

/**
 * This method is not yet supported.
 * @return
 */
public int getMode()
{
    return(0);
}

/**
 * This method is not yet supported.
 * @param value
 */
public void setMode(int value) {}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @return
 */
public boolean getMultiline(int row, int column)
{
    return(false);
}

/**
 * This method is not yet supported.
 * @param row
 * @param column
 * @param value
 */
public void setMultiline(int row, int column, boolean value) {}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @return
 */
public int getStringCase(int row, int column)
{
    return(0);
}

/**
 * This method is not yet supported.
 * @param column
 * @param row
 * @param value
 */
public void setStringCase(int row, int column, int value) {}

/**
 * This method is not yet supported.
 * @return
 */
public Component getTextComponent()
{
    return(null);
}

/**
 * This method is not yet supported.
 */
public void print() {}

/**
 * Adjusts a cell range to this table instance.
 * Used by all <code>set*(JCCellRange)</code> methods in Table to
 * make ALLCELLS/ALL values use the current table size.
 * <br><B>Note:</b> This means the table must be connected to
 * a data source at the time of the call.
 * @param cr
 */
protected void adjustCellRange(JCCellRange cr)
{
    if (cr.end_row == JCTableEnum.ALLCELLS ||
        cr.end_row == JCTableEnum.ALL)
    {
        cr.end_row = getNumRows();
    }
    if (cr.end_column == JCTableEnum.ALLCELLS ||
        cr.end_column == JCTableEnum.ALL)
    {
        cr.end_column = getNumColumns();
    }
}

}
