/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: TableChangeHandler.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;

/**
 * <CODE>TableChangeHandler</CODE> batches repaint/recalc/relayout requests in
 * order to minimize calculation and painting and to process requests
 * "off-line".
 */
public class TableChangeHandler implements Serializable,
                                           PropertyChangeListener
{

/** Recalculate, relayout and redisplay the table. */
public static final int    TABLE_RESET = 1 << 1;

/** Repaint the table. */
public static final int    TABLE_REPAINT = 1 << 2;

/** Recalculate the layout of table. */
public static final int    TABLE_LAYOUT = 1 << 3;

/** Recalculate all row dimensions. */
public static final int    TABLE_RECALC_ROW_HEIGHTS = 1 << 4;

/** Recalculate all column dimensions. */
public static final int    TABLE_RECALC_COLUMN_WIDTHS = 1 << 5;

/** Recalculates the row height based on a single cell change. */
public static final int    CELL_RECALC_HEIGHT = 1 << 6;

/** Recalculates the column width based on a single cell change. */
public static final int    CELL_RECALC_WIDTH = 1 << 7;

/** Repaints a single cell. */
public static final int    CELL_REPAINT = 1 << 8;

protected JCTable          table;


/**
 * Creates a new change handler for the table.
 */
public TableChangeHandler(JCTable table)
{
	this.table = table;
}

/**
 * Sets flags for non-row/column specific properties.
 */
protected void process(int flags)
{
	process(flags, JCTableEnum.NOVALUE, JCTableEnum.NOVALUE);
}

/**
 * Sets a property for a row/column.
 */
protected void process(int flags, int row, int column)
{
	// Do no calculations if table has not been realized
	if (!table.isDisplayable()) {
		return;
	}

	// process a reset request
	if (table.isRecalcRequired() || (flags & TABLE_RESET) != 0) {
		table.adjustmentHandler.adjustOrigin();

		CellSizeHandler.resetDimensions(table);
		if (table.isRepaintEnabled()) {
			table.setRecalcRequired(false);
		}
		flags = TABLE_LAYOUT | TABLE_REPAINT;
	}

	if ((flags & TABLE_RECALC_ROW_HEIGHTS) != 0) {
		if (row == JCTableEnum.NOVALUE || row == JCTableEnum.ALL) {
			CellSizeHandler.calcRowHeights(table, JCTableEnum.LABEL,
										   JCTableEnum.MAXINT,
			                               JCTableEnum.LABEL,
										   JCTableEnum.MAXINT);
		}
		else if (row == JCTableEnum.ALLCELLS) {
			CellSizeHandler.calcRowHeights(table, 0, JCTableEnum.MAXINT,
			                               JCTableEnum.LABEL,
										   JCTableEnum.MAXINT);
		}
		else {
			CellSizeHandler.calcRowHeights(table, row, row,
                                           JCTableEnum.LABEL,
										   JCTableEnum.MAXINT);
		}
		flags |= TABLE_LAYOUT | TABLE_REPAINT;
	}

	if ((flags & TABLE_RECALC_COLUMN_WIDTHS) != 0) {
		if (column == JCTableEnum.NOVALUE || column == JCTableEnum.ALL) {
			CellSizeHandler.calcColumnWidths(table, JCTableEnum.LABEL,
											 JCTableEnum.MAXINT,
                                             JCTableEnum.LABEL,
											 JCTableEnum.MAXINT);
		}
		else if (column == JCTableEnum.ALLCELLS) {
			CellSizeHandler.calcColumnWidths(table, JCTableEnum.LABEL,
											 JCTableEnum.MAXINT,
                                             0, JCTableEnum.MAXINT);
		}
		else {
			CellSizeHandler.calcColumnWidths(table, JCTableEnum.LABEL,
											 JCTableEnum.MAXINT,
                                             column, column);
		}
		flags |= TABLE_LAYOUT | TABLE_REPAINT;
	}

	if ((flags & CELL_RECALC_HEIGHT) != 0 && row != JCTableEnum.NOVALUE) {
		if (CellSizeHandler.calcRowHeights(table, row, row,
		                                   JCTableEnum.LABEL,
										   JCTableEnum.MAXINT))
		{
			flags |= TABLE_LAYOUT | TABLE_REPAINT;
		}
	}

	if ((flags & CELL_RECALC_WIDTH) != 0 && column != JCTableEnum.NOVALUE) {
		if (CellSizeHandler.calcColumnWidths(table, JCTableEnum.LABEL,
		                                     JCTableEnum.MAXINT,
		                                     column, column))
		{
			flags |= TABLE_LAYOUT | TABLE_REPAINT;
		}
	}

	if ((flags & CELL_REPAINT) != 0) {
		// make a valid cell range for repaint
		int sr = row, er = row;
		int sc = column, ec = column;
		if (sr == JCTableEnum.ALL) {
			sr = JCTableEnum.LABEL;
			er = JCTableEnum.MAXINT;
		}
		else if (sr == JCTableEnum.ALLCELLS) {
			sr = 0;
			er = JCTableEnum.MAXINT;
		}
		if (sc == JCTableEnum.ALL) {
			sr = JCTableEnum.LABEL;
			er = JCTableEnum.MAXINT;
		}
		else if (sc == JCTableEnum.ALLCELLS) {
			sr = 0;
			er = JCTableEnum.MAXINT;
		}

		if (!table.isRepaintEnabled()) {
			table.setRepaintRequired(true);
		}
		else {
		    CellAreaRenderer car = table.getCellAreaHandler().getCellAreaRenderer();
			car.repaintRange(sr, sc, er, ec);
		}
	}

	// Revalidate spans if there are frozen rows or columns
	if (table.spanHandler != null &&
	    (table.frozen_rows > 0 ||
		 table.frozen_columns > 0))
	{
		try {
			table.spanHandler.validateAllRanges();
		}
		catch(IllegalArgumentException ex) {
			System.out.println(ex.getMessage());
		}
	}

	// Process a layout request
	if ((flags & TABLE_LAYOUT) != 0) {
		if (table.isRepaintEnabled()) {
			table.layoutAreas();
			table.setRelayoutRequired(false);
		}
		else {
			table.setRelayoutRequired(true);
		}
	}
	else if (table.isRepaintEnabled() && table.isRelayoutRequired()) {
		table.layoutAreas();
		table.setRelayoutRequired(false);
	}

	// Check if traverse is required
	if (table.needs_initial_traverse) {
		table.getEditTraverseHandler().traverseInitial();
		table.needs_initial_traverse = false;

		// On the offchance that the top row or left column have been
		// explicitly set before now, we must reset them here (because
		// inital traversal will move the table to the first traversable
		// cell)
		table.needs_initial_set_top_check = false;
		if (table.set_initial_top &&
			table.current_row != table.getTopRow())
		{
			table.setTopRow(table.getTopRow());
			table.set_initial_top = false;
		}
		table.needs_initial_set_left_check = false;
		if (table.set_initial_left &&
		    table.current_column != table.getLeftColumn())
		{
			table.setLeftColumn(table.getLeftColumn());
			table.set_initial_left = false;
		}
	}

	// Process a repaint request
	ComponentModel componentHandler = table.getComponentModel();
	if (table.isRepaintEnabled() && (flags & TABLE_REPAINT) != 0) {
		if (componentHandler != null) {
			if (componentHandler.isParentSetRequired()) {
				componentHandler.addComponentsToParent();
				componentHandler.setParentSetRequired(false);
			}
			componentHandler.reconfigureComponents();
		}
		if (table.store_image_enabled) {
			table.store_image = null;
		}
		table.repaint();
		table.setRepaintRequired(false);
	}
}

/**
 * Listens for CellStyleModel property change.
 */
public void propertyChange(PropertyChangeEvent e)
{
	if (e.getSource() instanceof CellStyleModel) {
		CellStyleModel csm = (CellStyleModel)e.getSource();
		// Get the set of JCCellRanges that "use" this style
		Collection ranges = table.getCellStyleRanges(csm);

		if (e.getPropertyName().equals(CellStyleChangeSupport.STYLE_FONT)) {
			process(TABLE_RESET);
		}
		else {
			if (!table.isRepaintEnabled()) {
				table.setRepaintRequired(true);
			}
			else {
				// loop through each cell range and force request repaint
				Iterator it = ranges.iterator();
				CellAreaRenderer car =
                    table.getCellAreaHandler().getCellAreaRenderer();
				while (it.hasNext()) {
					CellRangeValue cr = (CellRangeValue)it.next();
					car.repaintRange(cr.start_row, cr.start_column,
					                 cr.end_row, cr.end_column);
				}

				// Grab the collection of styles and notify any children of
                // the change.
				Collection styles = table.getCellStyles();
				Iterator sit = styles.iterator();
				while (sit.hasNext()) {
					CellStyleModel model = (CellStyleModel)sit.next();
					if (model.getParentStyle() == csm) {
						PropertyChangeEvent pce =
						    new PropertyChangeEvent(model, e.getPropertyName(),
													e.getOldValue(),
													e.getNewValue());
						propertyChange(pce);
					}
				}
			}
		}
	}
}

}
