/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: TableLayout.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:34 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table;

import java.awt.Dimension;
import java.awt.Rectangle;
import javax.swing.JScrollBar;

/**
 * <CODE>TableLayout</CODE> controls the placement of <CODE>CellAreas</CODE>
 * and scrollbars in table based on their need and positionning policy.
 */
public class TableLayout implements java.io.Serializable {

protected JCTable              table;
protected CellAreaHandler      cah;

protected boolean              needs_layout = true;

/**
 * Constructor that requires a CellAreaHandler object.
 * @param cah The cell area handler
 */
public TableLayout(CellAreaHandler cah)
{
	this.cah = cah;
	this.table = cah.table;
}

/**
 * Lays out the areas of table and sets the scrollbar visibility
 * and placement.
 */
public void layoutAreas()
{
	Dimension d = table.getSize();
	layoutAreas(d.width, d.height);
}

/**
 * Lays out the areas of table and sets the scrollbar visibility
 * and placement based on a given width and height of the container.
 * @param container_width The width of the container
 * @param container_height The height of the container
 */
public void layoutAreas(int container_width, int container_height)
{
    if (table.isTracking()) {
		return;
    }

	int cell_height = 0;
	int cell_width = 0;
	int column_label_height = 0;
	int frozen_column_width = 0;
   	int frozen_row_height = 0;
	int row_label_width = 0;

	cah.createRequiredCellAreas();

	JScrollBar hsb = table.getHorizSB();
	int hsb_display = table.getHorizSBDisplay();

	JScrollBar vsb = table.getVertSB();
	int vsb_display = table.getVertSBDisplay();

	// Row label width
    row_label_width = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL,
                                   0, true);

    // Column_label_height
    column_label_height = getAreaSpace(cah.columnLabelArea,
                                       CellArea.COLUMNLABEL, 0, true);

	// Frozen row height
    frozen_row_height = getAreaSpace(cah.frozenRowArea, CellArea.FROZENROW,
                                     0, true);

    // Frozen column width
    frozen_column_width = getAreaSpace(cah.frozenColumnArea,
                                       CellArea.FROZENCOLUMN, 0, true);

	// Default need for scrollbar based on whether one is always required.
	boolean need_hsb = (hsb_display == JCTableEnum.SCROLLBAR_ALWAYS);
	boolean need_vsb = (vsb_display == JCTableEnum.SCROLLBAR_ALWAYS);

	int frame_border = table.getFrameBorderWidth();

	int remaining_width  = (container_width - row_label_width -
	                        frozen_column_width - 2 * frame_border);
	int remaining_height = (container_height - column_label_height -
	                        frozen_row_height - 2 * frame_border);
	// If the row label is visible, adjust the remaining area by the
	// row label offset
	if (row_label_width > 0) {
		remaining_width -= table.getRowLabelOffset() + 2 * frame_border;
	}
	// If the column label is visible, adjust the remaining area by the
	// column label offset
	if (column_label_height > 0) {
		remaining_height -= table.getColumnLabelOffset() + 2 * frame_border;
	}

	CellArea cellArea = cah.cellArea;
	int cellAreaPrefHeight = 0;
	int cellAreaPrefWidth = 0;
    if (cellArea != null && cellArea.isVisible()) {
        Dimension d = cellArea.getPreferredSize();
        cellAreaPrefHeight = d.height;
        cellAreaPrefWidth = d.width;
    }

	// Case: at least one of the scrollbars is "as_needed"
	if ((hsb_display == JCTableEnum.SCROLLBAR_AS_NEEDED) ||
	    (vsb_display == JCTableEnum.SCROLLBAR_AS_NEEDED))
	{
		// Get the preferred size of the cell area
		int pref_height = cellAreaPrefHeight;
		int pref_width = cellAreaPrefWidth;

		// Get the workable size of the cell area
		int work_height = remaining_height;
		int work_width = remaining_width;

		if (hsb_display == JCTableEnum.SCROLLBAR_NEVER) {
			need_hsb = false;
			if (pref_height > work_height) {
				need_vsb = true;
			}
			else {
				need_vsb = false;
			}
		}
		else if (hsb_display == JCTableEnum.SCROLLBAR_ALWAYS) {
			need_hsb = true;
			if (pref_height > (work_height - hsb.getPreferredSize().height -
			                   table.getHorizSBOffset()))
            {
				need_vsb = true;
			}
			else {
				need_vsb = false;
			}
		}
		else if (vsb_display == JCTableEnum.SCROLLBAR_NEVER) {
			need_vsb = false;
			if (pref_width > work_width) {
				need_hsb = true;
			}
			else {
				need_hsb = false;
			}
		}
		else if (vsb_display == JCTableEnum.SCROLLBAR_ALWAYS) {
			need_vsb = true;
			if (pref_width > (work_width - vsb.getPreferredSize().width -
			                  table.getVertSBOffset()))
			{
				need_hsb = true;
			}
			else {
				need_hsb = false;
			}
		}
		// Both as-needed case
		else if ((hsb_display == JCTableEnum.SCROLLBAR_AS_NEEDED) &&
				 (vsb_display == JCTableEnum.SCROLLBAR_AS_NEEDED))
		{
			// Case 1: both are too small
			if (pref_width < work_width && pref_height < work_height) {
				need_hsb = false;
				need_vsb = false;
			}
			// Case 2: width is too big... height is unknown
			else if (pref_width > work_width) {
				// Check that we have an area to scroll before committing
				// to having a scrollbar
				if (pref_width > 0) {
					need_hsb = true;
				}
				if (pref_height > 0 &&
					pref_height > (work_height - hsb.getPreferredSize().height -
								   table.getHorizSBOffset()))
				{
					need_vsb = true;
				}
				else {
					need_vsb = false;
				}
			}
			// Case 3: height is too big... width must be small
			else if (pref_height > work_height) {
				// Check that we have an area to scroll before committing
				// to having a scrollbar
				if (pref_height > 0) {
					need_vsb = true;
				}
				if (pref_width > 0 &&
				    pref_width > (work_width - vsb.getPreferredSize().width -
								  table.getVertSBOffset()))
				{
					need_hsb = true;
				}
				else {
					need_hsb = false;
				}
			}
		}
		if (need_hsb) {
			cell_height = work_height - hsb.getPreferredSize().height -
			              table.getHorizSBOffset();
		}
		else {
			cell_height = work_height;
		}

		if (need_vsb) {
			cell_width = work_width - vsb.getPreferredSize().width -
			             table.getVertSBOffset();
		}
		else {
			cell_width = work_width;
		}

	}
	// Case: horizontal is always present and the other is never
	else if ((hsb_display == JCTableEnum.SCROLLBAR_ALWAYS) &&
			 (vsb_display == JCTableEnum.SCROLLBAR_NEVER))
	{
		cell_height = remaining_height - hsb.getPreferredSize().height -
		              table.getHorizSBOffset();
		cell_width  = remaining_width;
		need_vsb    = false;
	}
	// Case: vertical is always present and the other is never
	else if ((hsb_display == JCTableEnum.SCROLLBAR_NEVER) &&
			 (vsb_display == JCTableEnum.SCROLLBAR_ALWAYS))
	{
		cell_height = remaining_height;
		cell_width  = remaining_width - vsb.getPreferredSize().width -
		              table.getVertSBOffset();
		need_hsb    = false;
	}
	// Case: both scrollbars are always present
	else if ((hsb_display == JCTableEnum.SCROLLBAR_ALWAYS) &&
			 (vsb_display == JCTableEnum.SCROLLBAR_ALWAYS))
	{
		cell_height = remaining_height - hsb.getPreferredSize().height -
		              table.getHorizSBOffset();
		cell_width  = remaining_width - vsb.getPreferredSize().width -
		              table.getVertSBOffset();
	}
	// Case: both scrollbars are never present
	else if ((hsb_display == JCTableEnum.SCROLLBAR_NEVER) &&
			 (vsb_display == JCTableEnum.SCROLLBAR_NEVER))
	{
		cell_height = remaining_height;
		cell_width = remaining_width;
		need_hsb = false;
		need_vsb = false;
	}

	// Cap the size of the cell area to the preferred size.
    int column_label_width = cell_width;
	cell_width = Math.min(cell_width, cellAreaPrefWidth);
	if (cell_width < 0) {
		cell_width = 0;
    }

	if (cell_width == 0) {
		need_hsb = false;
	}
    else {
        column_label_width = cell_width;
    }

    int row_label_height = cell_height;
	cell_height = Math.min(cell_height, cellAreaPrefHeight);
	if (cell_height < 0) {
		cell_height = 0;
    }

	if (cell_height == 0) {
		need_vsb = false;
	}
    else {
        row_label_height = cell_height;
    }

	// Determine the top-left position of each region
	int label_x = 0;    // the left position for the label
	int frozen_x = 0;   // the left position for the frozen column
	int cell_x = 0;     // the left position for the cell area
	int offset_x = 0;   // the adjustment for row label offset

	// If the row label is visible, account for the offset
	if (row_label_width > 0) {
		offset_x = table.getRowLabelOffset();
	}

	if (table.getRowLabelPlacement() == JCTableEnum.PLACE_LEFT) {
		label_x = frame_border;
		if (table.frozen_columns == 0 ||
		    table.getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT)
        {
			frozen_x = offset_x + row_label_width;
			cell_x = offset_x + row_label_width + frozen_column_width;
		}
		else if (table.getFrozenColumnPlacement() == JCTableEnum.PLACE_RIGHT) {
			cell_x = offset_x + row_label_width;
			frozen_x = offset_x + row_label_width + cell_width;
		}
		if (row_label_width > 0) {
			frozen_x += 3 * frame_border;
			cell_x += 3 * frame_border;
		}
	}
	else if (table.getRowLabelPlacement() == JCTableEnum.PLACE_RIGHT) {
		label_x = offset_x + frozen_column_width +
                  cell_width + 3 * frame_border;
		if (table.frozen_columns == 0 ||
		    table.getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT)
        {
			frozen_x = 0;
			cell_x = frozen_column_width;
		}
		else if (table.getFrozenColumnPlacement() == JCTableEnum.PLACE_RIGHT) {
			cell_x = 0;
			frozen_x = cell_width;
		}
		frozen_x += frame_border;
		cell_x += frame_border;
	}

	int label_y = 0;    // the top position for the label
	int frozen_y = 0;   // the top position for the frozen row
	int cell_y = 0;     // the top position for the cell area
	int offset_y = 0;   // the adjustment for column label offset

	// If the column label is visible, account for the offset
	if (column_label_height > 0) {
		offset_y = table.getColumnLabelOffset();
	}

	if (table.getColumnLabelPlacement() == JCTableEnum.PLACE_TOP) {
		label_y = frame_border;
		if (table.frozen_rows == 0 ||
		    table.getFrozenRowPlacement() == JCTableEnum.PLACE_TOP)
        {
			frozen_y = offset_y + column_label_height;
			cell_y = offset_y + column_label_height + frozen_row_height;
		}
		else if (table.getFrozenRowPlacement() == JCTableEnum.PLACE_BOTTOM) {
			cell_y = offset_y + column_label_height;
			frozen_y = offset_y + column_label_height + cell_height;
		}
		if (column_label_height > 0) {
			frozen_y += 3 * frame_border;
			cell_y += 3 * frame_border;
		}
	}
	else if (table.getColumnLabelPlacement() == JCTableEnum.PLACE_BOTTOM) {
		label_y = offset_y + frozen_row_height + cell_height + 3 * frame_border;
		if (table.frozen_rows == 0 ||
		    table.getFrozenRowPlacement() == JCTableEnum.PLACE_TOP)
        {
			frozen_y = 0;
			cell_y = frozen_row_height;
		}
		else if (table.getFrozenRowPlacement() == JCTableEnum.PLACE_BOTTOM) {
			cell_y = 0;
			frozen_y = cell_height;
		}
		frozen_y += frame_border;
		cell_y += frame_border;
	}

	cah.rowLabelRect = null;
	cah.columnLabelRect = null;
	cah.cellRect = null;

	// Set the bounds for the areas
	if (cellArea != null && cellArea.isVisible()) {
		cellArea.setBounds(cell_x, cell_y, cell_width, cell_height);
		cah.cellRect = cellArea.getBounds();
		setLayoutOffset(cellArea);
	}
	if (cah.rowLabelArea != null && cah.rowLabelArea.isVisible()) {
		// Top position and height determined by the cell area
		cah.rowLabelArea.setBounds(label_x, cell_y, row_label_width,
								   row_label_height);
		cah.rowLabelRect = cah.rowLabelArea.getBounds();
		setLayoutOffset(cah.rowLabelArea);
	}
	if (cah.columnLabelArea != null && cah.columnLabelArea.isVisible()) {
		// Left position and width determined by the cell area
		cah.columnLabelArea.setBounds(cell_x, label_y, column_label_width,
									  column_label_height);
		cah.columnLabelRect = cah.columnLabelArea.getBounds();
		setLayoutOffset(cah.columnLabelArea);
	}
	if (cah.frozenColumnArea != null && cah.frozenColumnArea.isVisible()) {
		// Top position and height determined by the cell area
		cah.frozenColumnArea.setBounds(frozen_x, cell_y, frozen_column_width,
		                               cell_height);
		setLayoutOffset(cah.frozenColumnArea);
	}
	if (cah.frozenRowArea != null && cah.frozenRowArea.isVisible()) {
		// Left position and width determined by the cell area
		cah.frozenRowArea.setBounds(cell_x, frozen_y, cell_width,
		                            frozen_row_height);
		setLayoutOffset(cah.frozenRowArea);
	}
	if (cah.frozenRowLabelArea != null && cah.frozenRowLabelArea.isVisible()) {
		// Position based on the row label and frozen row
		cah.frozenRowLabelArea.setBounds(label_x, frozen_y,
										 row_label_width, frozen_row_height);
		if (cah.rowLabelRect != null) {
			Rectangle rect = cah.rowLabelRect;
			rect.setBounds(Math.min(rect.x, label_x),
						   Math.min(rect.y, frozen_y),
						   Math.max(rect.width, row_label_width),
						   rect.height + frozen_row_height);
		}
		setLayoutOffset(cah.frozenRowLabelArea);
	}
	if (cah.frozenColumnLabelArea != null &&
	    cah.frozenColumnLabelArea.isVisible())
	{
		// Position based on the column label and frozen column
		cah.frozenColumnLabelArea.setBounds(frozen_x, label_y,
											frozen_column_width,
											column_label_height);
		if (cah.columnLabelRect != null) {
			Rectangle rect = cah.columnLabelRect;
			rect.setBounds(Math.min(rect.x, frozen_x),
						   Math.min(rect.y, label_y),
						   rect.width + frozen_column_width,
						   Math.max(rect.height, column_label_height));
		}
		setLayoutOffset(cah.frozenColumnLabelArea);
	}
	if (cah.frozenCellArea != null && cah.frozenCellArea.isVisible()) {
		// Position based on the frozen row and frozen column
		cah.frozenCellArea.setBounds(frozen_x, frozen_y,
									 frozen_column_width, frozen_row_height);
		setLayoutOffset(cah.frozenCellArea);
	}
	if (cah.cellRect != null) {
		Rectangle rect = cah.cellRect;
		if (cah.rowLabelRect != null) {
			rect.y = cah.rowLabelRect.y;
			rect.height = cah.rowLabelRect.height;
		}
		if (cah.columnLabelRect != null) {
			rect.x = cah.columnLabelRect.x;
			rect.width = cah.columnLabelRect.width;
		}
	}

	Rectangle vsb_bounds = null;
	if (need_vsb) {
		int sb_width = vsb.getPreferredSize().width;
		vsb_bounds = new Rectangle(0, cell_y, sb_width, cell_height);

		// Attachment: cells or side
		if (table.getVertSBAttachment() == JCTableEnum.SIZE_TO_TABLE) {
			vsb_bounds.height = container_height - cell_y;
		}

		// Position: cells or side
        int xPosMax = container_width - sb_width;
		if (table.getVertSBPosition() == JCTableEnum.POSITION_BY_CELLS) {
            int offset = row_label_width > 0 ? table.getRowLabelOffset() : 0;
            vsb_bounds.x = row_label_width + offset +
				           frozen_column_width + cell_width +
						   table.getVertSBOffset() + 2 * frame_border;
			if (row_label_width > 0) {
				vsb_bounds.x += 2 * frame_border;
			}
            if (vsb_bounds.x > xPosMax) {
                vsb_bounds.x = xPosMax;
            }
		}
		else {
			vsb_bounds.x = xPosMax;
		}
	}

	Rectangle hsb_bounds = null;
	if (need_hsb) {
		int sb_height = hsb.getPreferredSize().height;
		hsb_bounds = new Rectangle(cell_x, 0, cell_width, sb_height);

		// Attachment: cells or side
		if (table.getHorizSBAttachment() == JCTableEnum.SIZE_TO_TABLE) {
			hsb_bounds.width = container_width - cell_x;
		}

		// Position: cells or side
        int yPosMax = container_height - sb_height;
		if (table.getHorizSBPosition() == JCTableEnum.POSITION_BY_CELLS) {
            int offset =
                column_label_height > 0 ? table.getColumnLabelOffset() : 0;
            hsb_bounds.y = column_label_height + offset +
				           frozen_row_height + cell_height +
						   table.getHorizSBOffset() + 2 * frame_border;
			if (column_label_height > 0) {
				hsb_bounds.y += 2 * frame_border;
			}
            if (hsb_bounds.y > yPosMax) {
                hsb_bounds.y = yPosMax;
            }
		}
		else {
			hsb_bounds.y = yPosMax;
		}
	}

	if (need_hsb && need_vsb) {
		// Caps the width of the horizontal scrollbar so that it doesn't extend
		// past the left edge of the vertical scrollbar
		hsb_bounds.width = Math.min(vsb_bounds.x - hsb_bounds.x,
		                            hsb_bounds.width);
		// Caps the height of the vertical scrollbar so that it doesn't extend
		// past the top of the horizontal scrollbar
		vsb_bounds.height = Math.min(hsb_bounds.y - vsb_bounds.y,
						             vsb_bounds.height);
	}

	int vert_origin = cah.getCellOrigin().y;
	int act_height = cellAreaPrefHeight;

	if (table.set_top_row && cellArea != null) {
		vert_origin = -(cah.getBounds(table.getTopRow(), 0).y -
		                cellArea.getStandardOffsetY());
		if (cah.frozenRowArea != null && cah.frozenRowArea.isVisible() &&
		    table.getFrozenRowPlacement() == JCTableEnum.PLACE_BOTTOM)
        {
			vert_origin -= frozen_row_height;
		}
		table.set_top_row = false;
	}

	if (cell_height - vert_origin > act_height) {
		vert_origin = -Math.max(0, act_height - cell_height);
		boolean needs_top_row = (vert_origin != cah.cellOrigin.y);
		cah.cellOrigin.y = vert_origin;
		if (needs_top_row) {
			table.top_row = table.getVisibleCells().start_row;
			table.set_top_row = false;
		}
	}

	if (need_vsb && cellArea != null) {
		vsb.setBounds(vsb_bounds.x, vsb_bounds.y,
		              vsb_bounds.width, vsb_bounds.height);
	}
    int jump_mode = table.getJumpScroll();
    if (jump_mode == JCTableEnum.JUMP_VERTICAL ||
	    jump_mode == JCTableEnum.JUMP_ALL)
	{
        vsb.setUnitIncrement(1);
    }
    else {
        vsb.setUnitIncrement(21);
    }

    vsb.setBlockIncrement(cell_height);
    vsb.setValues(-vert_origin, cell_height, 0, cellAreaPrefHeight);

	int horiz_origin = cah.getCellOrigin().x;
	int act_width = cellAreaPrefWidth;

	if (table.set_left_column && cellArea != null) {
		horiz_origin = -(cah.getBounds(0, table.getLeftColumn()).x -
						 cellArea.getStandardOffsetX());
		if (cah.frozenColumnArea != null && cah.frozenColumnArea.isVisible() &&
		    table.getFrozenColumnPlacement() == JCTableEnum.PLACE_RIGHT)
		{
			horiz_origin -= frozen_column_width;
		}
		table.set_left_column = false;
	}

	if (cell_width - horiz_origin > act_width) {
		horiz_origin = -Math.max(0, act_width - cell_width);
		boolean needs_left_column = (horiz_origin != cah.cellOrigin.x);
		cah.cellOrigin.x = horiz_origin;
		if (needs_left_column) {
			table.left_column = table.getVisibleCells().start_column;
			table.set_left_column = false;
		}
	}

	if (need_hsb && cellArea != null) {
		hsb.setBounds(hsb_bounds.x, hsb_bounds.y,
		              hsb_bounds.width, hsb_bounds.height);
	}
    if (jump_mode == JCTableEnum.JUMP_HORIZONTAL ||
	    jump_mode == JCTableEnum.JUMP_ALL)
	{
        hsb.setUnitIncrement(1);
    }
    else {
    	hsb.setUnitIncrement(20);
    }
    hsb.setBlockIncrement(cell_height);
	hsb.setValues(-horiz_origin, cell_width, 0, cellAreaPrefWidth);

	if (hsb != null) {
		hsb.setVisible(need_hsb);
	}
	if (vsb != null) {
		vsb.setVisible(need_vsb);
	}

	if (table.getRowLabelPlacement()     == JCTableEnum.PLACE_LEFT &&
	    table.getFrozenColumnPlacement() == JCTableEnum.PLACE_LEFT &&
	    table.getColumnLabelPlacement()  == JCTableEnum.PLACE_TOP &&
	    table.getFrozenRowPlacement()    == JCTableEnum.PLACE_TOP)
	{
		cah.setStandardLayout(true);
	}
	else {
		cah.setStandardLayout(false);
	}

	// Ensure that the editor is correctly placed
    table.getEditTraverseHandler().repositionEditor();
}

/**
 * The the space used for a given cell area type.  This is used to calculate
 * positioning for other cell areas.
 * @param area The cell area for which the offset is required.
 * @param type The type of area.  We can't just get this from the cell area
 * because it might be null.
 * @param initialOffset An initial value to be added to the offset.
 * @param pref If true, use the preferred size, otherwise use the minimum size.
 * @return The offset of the CellArea.
 */
protected int getAreaSpace(CellArea area, int type, int initialOffset,
                           boolean pref)
{
    int offset = 0;
    Dimension size = null;
    if (area != null && area.isVisible()) {
        size = pref ? area.getPreferredSize() : area.getMinimumSize();
    }

    switch (type) {
    case CellArea.ROWLABEL:
        if (table.isRowLabelDisplay()) {
            if (area != null && area.isVisible()) {
                offset = size.width + initialOffset;
            }
            else if (table.getNumRows() == 0) {
                offset = table.getCellLayout().getWidth(JCTableEnum.LABEL) +
                         initialOffset;
            }
        }
        break;
    case CellArea.COLUMNLABEL:
        if (table.isColumnLabelDisplay()) {
            if (area != null && area.isVisible()) {
                offset = size.height + initialOffset;
            }
            else if (table.getNumColumns() == 0) {
                offset = table.getCellLayout().getHeight(JCTableEnum.LABEL) +
                         initialOffset;
            }
        }
        break;
    case CellArea.FROZENROW:
        if (area != null && area.isVisible()) {
            offset = size.height + initialOffset;
        }
        break;
    case CellArea.FROZENCOLUMN:
        if (area != null && area.isVisible()) {
            offset = size.width + initialOffset;
        }
        break;
    }
    return(offset);
}

/**
 * Sets the offset from standard position (LABEL, FROZEN, CELL)
 * based on the visibility and position of labels and frozen
 * cells.
 */
protected void setLayoutOffset(CellArea area)
{
	int x = 0;
	int y = 0;

    switch(area.getType()) {
    case CellArea.CELL:
        x = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL, 0, true) +
            getAreaSpace(cah.frozenColumnArea, CellArea.FROZENCOLUMN, 0, true);
        y = getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL, 0, true) +
            getAreaSpace(cah.frozenRowArea, CellArea.FROZENROW, 0, true);
        break;
    case CellArea.FROZENROW:
        x = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL, 0, true) +
            getAreaSpace(cah.frozenColumnArea, CellArea.FROZENCOLUMN, 0, true);
        y = getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL, 0, true);
        break;
    case CellArea.COLUMNLABEL:
        x = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL, 0, true) +
            getAreaSpace(cah.frozenColumnArea, CellArea.FROZENCOLUMN, 0, true);
        break;
    case CellArea.FROZENCOLUMN:
        x = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL, 0, true);
        y = getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL, 0, true) +
            getAreaSpace(cah.frozenRowArea, CellArea.FROZENROW, 0, true);
        break;
    case CellArea.FROZENCELL:
        x = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL, 0, true);
        y = getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL, 0, true);
        break;
    case CellArea.FROZENCOLUMNLABEL:
        x = getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL, 0, true);
        break;
    case CellArea.ROWLABEL:
        y = getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL, 0, true) +
            getAreaSpace(cah.frozenRowArea, CellArea.FROZENROW, 0, true);
        break;
    case CellArea.FROZENROWLABEL:
        y = getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL, 0, true);
        break;
    }
	area.setStandardOffset(x, y);
}

/**
 * Returns the minimum size of table based on the
 * number of rows and columns and scrollbar policy.
 */
public Dimension getMinimumSize()
{
	if (needs_layout) {
		needs_layout = false;
		layoutAreas();
	}

	JScrollBar hsb = table.getHorizSB();
	JScrollBar vsb = table.getVertSB();

	int width = 0;
	int height = 0;
	int frame_border = 2 * table.getFrameBorderWidth();

    // Row labels
    width += getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL,
                          table.getRowLabelOffset() + frame_border, false);

    // Column labels
    height += getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL,
                           table.getColumnLabelOffset() + frame_border, false);

    // Frozen columns
    width += getAreaSpace(cah.frozenColumnArea, CellArea.FROZENCOLUMN,
                          0, false);

    // Frozen rows
    height += getAreaSpace(cah.frozenRowArea, CellArea.FROZENROW, 0, false);

	// Cell area
	if (cah.cellArea != null && cah.cellArea.isVisible()) {
		width  += cah.cellArea.getMinimumSize().width  + frame_border;
		height += cah.cellArea.getMinimumSize().height + frame_border;
	}
	// Horizontal scrollbar
	if (vsb != null && vsb.isVisible() &&
	    table.getVertSBDisplay() == JCTableEnum.SCROLLBAR_ALWAYS)
	{
		width += vsb.getMinimumSize().width;
	}
	// Vertical scrollbar
	if (hsb != null && hsb.isVisible() &&
	    table.getHorizSBDisplay() == JCTableEnum.SCROLLBAR_ALWAYS)
	{
		height += hsb.getMinimumSize().height;
	}
	return(new Dimension(width, height));
}

/**
 * Returns the preferred size of table based on the
 * number of rows and columns and scrollbar policy.
 * Table will return the maximum possible size.
 */
public Dimension getPreferredSize()
{
	if (needs_layout) {
		needs_layout = false;
		layoutAreas(JCTableEnum.MAXINT, JCTableEnum.MAXINT);
	}

	JScrollBar hsb = table.getHorizSB();
	JScrollBar vsb = table.getVertSB();

	int width = 0;
	int height = 0;
	int frame_border = 2 * table.getFrameBorderWidth();

	// Row labels
    width += getAreaSpace(cah.rowLabelArea, CellArea.ROWLABEL,
                          table.getRowLabelOffset() + frame_border, true);

	// Column labels
    height += getAreaSpace(cah.columnLabelArea, CellArea.COLUMNLABEL,
                           table.getColumnLabelOffset() + frame_border, true);

	// Frozen columns
    width += getAreaSpace(cah.frozenColumnArea, CellArea.FROZENCOLUMN,
                          0, true);

	// Frozen rows
    height += getAreaSpace(cah.frozenRowArea, CellArea.FROZENROW, 0, true);

	// Cell area
	if (cah.cellArea != null && cah.cellArea.isVisible()) {
		if (table.getVisibleRows() > 0 || table.getVisibleColumns() > 0) {
			int sr = cah.cellArea.start_row;
			int sc = cah.cellArea.start_column;
			int er = cah.cellArea.end_row;
			int ec = cah.cellArea.end_column;
			if (table.getVisibleRows() > 0) {
				int visible_rows = table.getVisibleRows();
				er = Math.min(er, sr + visible_rows - 1);
			}
			if (table.getVisibleColumns() > 0) {
				int visible_columns = table.getVisibleColumns();
				ec = Math.min(ec, sc + visible_columns - 1);
			}

			CellLayoutModel clm = table.getCellLayout();
			Rectangle bounds = clm.getBounds(sr, sc, er, ec);

			width += bounds.width + frame_border;
			height += bounds.height + frame_border;
		}
		else {
			width  += cah.cellArea.getPreferredSize().width + frame_border;
			height += cah.cellArea.getPreferredSize().height + frame_border;
		}
	}
	// Horizontal scrollbar
	if (vsb != null && vsb.isVisible()) {
		width += vsb.getPreferredSize().width;
	}
	// Vertical scrollbar
	if (hsb != null && hsb.isVisible()) {
		height += hsb.getPreferredSize().height;
	}
	return(new Dimension(width, height));
}

}
