/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: AbstractDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:35 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table.data;

import com.klg.jclass.table.TableDataModel;
import com.klg.jclass.table.JCTableDataListener;
import com.klg.jclass.table.JCTableDataEvent;
import com.klg.jclass.util.JCListenerList;
import java.io.Serializable;
import java.util.Enumeration;

/**
 * The <CODE>AbstractDataSource</CODE> abstract class implements the
 * <CODE>TableDataModel</CODE> interface and defines methods for dealing with
 * JCTableDataListener processing.
 *
 * @see TableDataModel
 * @see com.klg.jclass.util.JCListenerList
 */
public abstract class AbstractDataSource implements TableDataModel,
                                                    Serializable
{

protected JCListenerList listeners;

/**
 * Adds a <CODE>JCTableDataListener</CODE> to this object.
 *
 * @see JCTableDataListener
 * @see JCTableDataEvent
 */
public synchronized void addTableDataListener(JCTableDataListener l)
{
    listeners = JCListenerList.add(listeners, l);
}

/**
 * Removes a <CODE>JCTableDataListener</CODE> from this object
 *
 * @see JCTableDataListener
 * @see JCTableDataEvent
 */
public synchronized void removeTableDataListener(JCTableDataListener l)
{
    listeners = JCListenerList.remove(listeners, l);
}

/**
 * Sends a JCTableDataEvent object to registered JCTableDataListener objects.
 */
public void fireTableDataEvent(JCTableDataEvent evt)
{
    Enumeration e = JCListenerList.elements(listeners);

    JCTableDataListener l;
    synchronized (this) {
        while (e.hasMoreElements()) {
            l = (JCTableDataListener)e.nextElement();
            l.dataChanged(evt);
        }
    }
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * a single cell value changes.
 */
public void fireValueChanged(int row, int column)
{
    fireTableDataEvent(new JCTableDataEvent(this, row, column, 0, 0,
                                            JCTableDataEvent.CHANGE_VALUE));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * a single row changes.
 */
public void fireRowChanged(int row)
{
    fireTableDataEvent(new JCTableDataEvent(this, row, 0, 0, 0,
                                            JCTableDataEvent.CHANGE_ROW));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * one or more rows are added.
 */
public void fireRowsAdded(int start_row, int num_rows)
{
    fireTableDataEvent(new JCTableDataEvent(this, start_row, 0, num_rows, 0,
                                            JCTableDataEvent.ADD_ROW));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * one or  more rows are deleted.
 */
public void fireRowDeleted(int start_row, int num_rows)
{
    fireTableDataEvent(new JCTableDataEvent(this, start_row, 0, num_rows, 0,
                                            JCTableDataEvent.REMOVE_ROW));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * a single column changes.
 */
public void fireColumnChanged(int column)
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, column, 0, 0,
                                            JCTableDataEvent.CHANGE_COLUMN));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * one or more columns are added.
 */
public void fireColumnsAdded(int start_column, int num_columns)
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, start_column, num_columns,
                                            0, JCTableDataEvent.ADD_COLUMN));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when one or
 * more columns are deleted.
 */
public void fireColumnsDeleted(int start_column, int num_columns)
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, start_column, num_columns,
                                            0, JCTableDataEvent.REMOVE_COLUMN));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * a single row label changes.
 */
public void fireRowLabelChanged(int row)
{
    fireTableDataEvent(new JCTableDataEvent(this, row, 0, 0, 0,
                                            JCTableDataEvent.CHANGE_ROW_LABEL));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * a single column label changes.
 */
public void fireColumnLabelChanged(int column)
{
    JCTableDataEvent event =
        new JCTableDataEvent(this, 0, column, 0, 0,
                             JCTableDataEvent.CHANGE_COLUMN_LABEL);
    fireTableDataEvent(event);
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * one or more rows are moved.
 * @param destination the row that will follow the moved rows.
 */
public void fireRowsMoved(int start_row, int num_rows, int destination)
{
    fireTableDataEvent(new JCTableDataEvent(this, start_row, 0, num_rows,
                                            destination,
                                            JCTableDataEvent.MOVE_ROW));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * one or more columns are moved.
 * @param destination the column that will follow the moved columns.
 */
public void fireColumnsMoved(int start_column, int num_columns, int destination)
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, start_column, num_columns,
                                            destination,
                                            JCTableDataEvent.MOVE_COLUMN));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * the number of rows changes.
 */
public void fireNumRowsChanged()
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, 0, 0, 0,
                                            JCTableDataEvent.NUM_ROWS));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * the number of columns changes.
 */
public void fireNumColumnsChanged()
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, 0, 0, 0,
                                            JCTableDataEvent.NUM_COLUMNS));
}

/**
 * Sends a JCTableDataEvent object to registered TableDataListener objects when
 * a reset is required because numerous changes have been made.
 */
public void fireDataReset()
{
    fireTableDataEvent(new JCTableDataEvent(this, 0, 0, 0, 0,
                                            JCTableDataEvent.RESET));
}

public boolean hasListeners()
{
    return(listeners != null);
}

public void dispose() {}

}
