/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCCachedDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:36 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table.data;

import com.klg.jclass.table.TableDataModel;
import com.klg.jclass.table.JCTableDataEvent;
import com.klg.jclass.table.JCTableDataListener;





import java.util.Vector;


/**
 * The <CODE>JCCachedDataSource</CODE> class stores data in a Vector of Vectors. 
 * Unlike other data sources, the source for label and cell data is another 
 * <CODE>TableData</CODE> class. If the object stored for the cell or label 
 * is null, the call passes through to the contained <CODE>TableData</CODE> 
 * class, and the value is fetched and stored in <CODE>JCCachedDataSource</CODE>.
 *
 * <CODE>JCCachedDataSource</CODE> should only be used when the actual
 * <code>TableData</code>'s <code>getTableDataItem</code>, <code>getTableRowLabel</code>,
 * or <code>getTableColumnLabel</code> are calculation-intensive or expensive to fetch.
 * 
 * Although <CODE>JCCachedDataSource</CODE> can bind to any data source, it 
 * implements <CODE>TableData</CODE> and not <CODE>EditableTableData</CODE>,
 * and is therefore not editable. Use <CODE>EditableCachedDataSource</CODE> 
 * to bind to an editable data source and have edit functionality. <br><B>Note:</b>
 * A non-editable data source bound to <CODE>JCEditableCachedDataSource</CODE>
 * will display an editor but reject all changes.
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, some of the return types for this class'
 * fields or methods are based on the corresponding
 * <code>com.sun.java.util</code> class.
 */
public class JCCachedDataSource extends AbstractVectorDataSource implements JCTableDataListener {
	
/** The actual source of data. */
protected TableDataModel dataSource;

protected Vector       cached_row_labels = createVector();
protected Vector       cached_column_labels = createVector();
protected Vector       cached_cells = createVector();

/**
 * Creates a new <CODE>TableDataModel</CODE> caching data source.
 * @param source
 */
public JCCachedDataSource(TableDataModel source) {
	this.dataSource = source;
	this.dataSource.addTableDataListener(this);
}

/**
 * Returns the number of rows in the contained data source.
 * Row count is not cached.
 * @return number of rows in the contained data source
 */
public int getNumRows() {
	return dataSource.getNumRows();
}

/**
 * Returns the number of columns in the contained data source.
 * Column count is not cached.
 * @return number of columns in the contained data source
 */
public int getNumColumns() {
	return dataSource.getNumColumns();
}

/**
 * Returns the value for the cell at the specified location. If the cache is null,
 * the contained data source is read and the value is stored in the cache.
 * @return value for the cell at the specified location
 * @param row
 * @param column
 */
public Object getTableDataItem(int row, int column) {
	if(row > cached_cells.size() - 1) {
		Object data = dataSource.getTableDataItem(row, column);
		
		Vector row_values = createVector();
		setElementAt(row_values, column, data);
		setElementAt(cached_cells, row, row_values);
		
		return data;
	} else {
		Vector row_values = (Vector)cached_cells.elementAt(row);
		if(row_values == null) {
			row_values = createVector();
		}
		if(column > row_values.size() - 1 || row_values.elementAt(column) == null) {
			Object data = dataSource.getTableDataItem(row, column);
			setElementAt(row_values, column, data);
			setElementAt(cached_cells, row, row_values);
			
			return data;
		}
		
	}
	return ((Vector)cached_cells.elementAt(row)).elementAt(column);
}

/**
 * Returns the value for the row label at the specified row. If the cache is null,
 * the contained data source is read and the value is stored in the cache.
 * @return value for the row label at the specified row
 * @param row
 */
public Object getTableRowLabel(int row) {
	if(row > cached_row_labels.size() - 1 || cached_row_labels.elementAt(row) == null) {
		Object data = dataSource.getTableRowLabel(row);
		setElementAt(cached_row_labels, row, data);
		return data;
	}
	return cached_row_labels.elementAt(row);
}

/**
 * Returns the value for the column label at the specified column. If the cache is
 * null, the contained data source is read and the value is stored in the cache.
 * @return value for the column label at the specified column
 * @param column
 */
public Object getTableColumnLabel(int column) {
	if(column > cached_column_labels.size() - 1 || cached_column_labels.elementAt(column) == null) {
		Object data = dataSource.getTableColumnLabel(column);
		setElementAt(cached_column_labels, column, data);
		return data;
	}
	return cached_column_labels.elementAt(column);
}

/**
 * Listens for <code>JCTableDataEvents</code> on the contained data source. A simplistic approach has
 * been taken to dirty the cache. For single cell changes, the element is set to 
 * null so that the next read forces a re-read of the contained data source. For
 * row specific changes, the row and cell cache is nulled; for column specific
 * changes, the column and cell cache is nulled. Reset nulls everything.
 * @param e
 */
public void dataChanged(JCTableDataEvent e) {
	int row = e.getRow();
	int column = e.getColumn();
	
	switch(e.getCommand()) {
	case JCTableDataEvent.CHANGE_VALUE:
		try {
			setElementAt(((Vector)cached_cells.elementAt(row)), column, null);
		}
		catch(Exception ex) {
			System.out.println("An error in JCCachedDataSource.dataChanged()");
		}
		break;
	case JCTableDataEvent.CHANGE_ROW:
	case JCTableDataEvent.ADD_ROW:
	case JCTableDataEvent.REMOVE_ROW:
	case JCTableDataEvent.MOVE_ROW:
	case JCTableDataEvent.NUM_ROWS:
		cached_row_labels = createVector();
		cached_cells = createVector();
		break;
	case JCTableDataEvent.CHANGE_COLUMN:
	case JCTableDataEvent.ADD_COLUMN:
	case JCTableDataEvent.REMOVE_COLUMN:
	case JCTableDataEvent.MOVE_COLUMN:
	case JCTableDataEvent.NUM_COLUMNS:
		cached_column_labels = createVector();
		cached_cells = createVector();
		break;
	case JCTableDataEvent.CHANGE_ROW_LABEL:
		setElementAt(cached_row_labels, row, null);
		break;
	case JCTableDataEvent.CHANGE_COLUMN_LABEL:
		setElementAt(cached_column_labels, column, null);
		break;
	case JCTableDataEvent.RESET:
		cached_row_labels = createVector();
		cached_column_labels = createVector();
		cached_cells = createVector();
		break;
	}
	fireTableDataEvent(e);
}

} // End JCCachedDataSource
