/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCInputStreamDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:36 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table.data;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Serializable;
import java.io.StreamTokenizer;

import com.klg.jclass.table.JCTableEnum;
import com.klg.jclass.util.JCFormatParserUtil;

/**
 *
 * This class is the base for any data source that relies on input
 * from some stream, which includes files, URLs, large Strings, etc.
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, some of the return types for this class'
 * fields or methods are based on the corresponding
 * <code>com.sun.java.util</code> class.
 *
 * @see com.klg.jclass.table.TableDataModel
 */
public class JCInputStreamDataSource extends JCVectorDataSource implements Serializable {

protected int dataFormat = JCTableEnum.DETECT;
protected JCFileFormatParser customParser = null;

/**
 * The constructor takes the <CODE>InputStream</CODE> and reads 
 * in all the appropriate data.
 *
 * @exception IOException An IOException will be thrown if there are any errors
 *  while reading in the data.
 * @param i
 */
public JCInputStreamDataSource(InputStream i) throws IOException {
	readFromStream(i);
}

/**
 * The constructor takes the <CODE>InputStream</CODE> and reads 
 * in all the appropriate data.
 *
 * @param i Open input stream
 * @param newFormat Data format; one of <code>STANDARD</code> or <code>CSV</code>
 *
 * @exception IOException An IOException will be thrown if there are any errors
 *  while reading in the data.
 */
public JCInputStreamDataSource(InputStream i, int newFormat) throws IOException {
	dataFormat = newFormat;
	readFromStream(i);
}

/**
 * The constructor takes the <CODE>InputStream</CODE> and reads 
 * in all the appropriate data according to the specified parser.
 *
 * @param i open input stream
 * @param prs custom parser for this data format
 *
 * @exception IOException An IOException will be thrown if there are any errors
 *  while reading in the data.
 */
public JCInputStreamDataSource(InputStream i, JCFileFormatParser prs) 
													throws IOException {
	dataFormat = JCTableEnum.CUSTOM;
	customParser = prs;
	readFromStream(i);
}

/**
 * Gets the value of the <code>DataFormat</code> property.
 * The <code>DataFormat</code> property determines how data will
 * be parsed by this data source.  Valid options include
 * <code>STANDARD</code> (for table-style format), <code>CSV</code> (for
 * comma-separated values), and <code>CUSTOM</code> (for custom
 * formats parsed by a user provided parser).
 *
 * @return one of <code>JCTableEnum.STANDARD</code>, <code>JCTableEnum.CSV</code>,
 *         or <code>JCTableEnum.CUSTOM</code>
 */
public int getDataFormat() {
	return dataFormat;
}

/**
 * Sets the value of the <code>DataFormat</code> property.
 * The <code>DataFormat</code> property determines how data will
 * be parsed by this data source.  Valid options include
 * <code>STANDARD</code> (for table-style format), <code>CSV</code> (for
 * comma-separated values), and <code>CUSTOM</code> (for custom
 * formats parsed by a user provided parser).
 *
 * @param newFormat one of <code>JCTableEnum.STANDARD</code>, <code>JCTableEnum.CSV</code>, or
 *         <code>JCTableEnum.CUSTOM</code>
 */
public void setDataFormat(int newFormat) {
	dataFormat = newFormat;
}

/**
 * Gets the value of the <code>CustomParser</code> property.
 * The <code>CustomParser</code> property identifies the parser to be
 * used when the <code>DataFormat</code> property has been set to <code>CUSTOM</code>.
 * The value of this property must be a class that has
 * implemented the <code>JCFileFormatParser</code> interface.
 *
 * @return the class implementing <code>JCFileFormatParser</code> that will be used
 *         to parse data
 */
public JCFileFormatParser getCustomParser() {
	return customParser;
}

/**
 * Sets the value of the <code>CustomParser</code> property.
 * The <code>CustomParser</code> property identifies the parser to be
 * used when the <code>DataFormat</code> property has been set to <code>CUSTOM</code>.
 * The value of this property must be a class that has
 * implemented the <code>JCFileFormatParser</code> interface.
 *
 * @param newParser the class implementing <code>JCFileFormatParser</code> that will
 *                  be used to parse data
 */
public void setCustomParser(JCFileFormatParser newParser) {
	customParser = newParser;
}

/**
 * Subclasses may need to set the stream up before they can call
 * <CODE>super()</CODE>, so provide a default <CODE>ctor</CODE> 
 * and the <CODE>ctor()</CODE> function 
 * to replace the <CODE>super()</CODE> call (for example,
 * <CODE>JCAppletDataSource</CODE>).
 */
protected JCInputStreamDataSource() {
}

/**
 * Reads data from an <code>InputStream</code>.
 * @param i
 * @throws IOException
 */
protected void readFromStream(InputStream i) throws IOException {
	Reader r = new BufferedReader(new InputStreamReader(i));
	read(r);
}

/**
 * Reads and parses data from the reader, and 
 * places it inside the current data source.
 * @param r
 * @throws IOException
 */
protected void read(Reader r) throws IOException {
	JCFileFormatParser p = null;
	if (dataFormat == JCTableEnum.DETECT) {
		// Mark/reset not available on all readers.
		try {
			r.mark(1024);
			dataFormat = detectFileFormat(r);
			r.reset();
		}
		catch (IOException ioe) {
			System.out.println("mark/reset not supported");
			dataFormat = JCTableEnum.STANDARD;
		}
	}
	switch (dataFormat) {
	case JCTableEnum.STANDARD:
		p = new JCStandardFormatParser();
		break;
	case JCTableEnum.CSV:
		p = new JCCSVFormatParser();
		break;
	case JCTableEnum.CUSTOM:
		p = customParser;
		break;
	}

	if (p != null)
		p.readAndParse(r, this);
}

/**
 * Global method for detecting the format of a
 * file.
 * @return
 * @param r
 */
public static int detectFileFormat(Reader r) {
	// Any kind of error will result in CSV
	int format = JCTableEnum.CSV;

	StreamTokenizer tokenizer = new StreamTokenizer(r);
	try {
		// Need to do this if you don't want to parse #s
		// Why?  Because java.io.StreamTokenizer does not
		// allow you to not parse numbers.
		tokenizer.resetSyntax();

		// Now manually re-do everything that StreamTokenizer
		// does in order to create a don't-parse-numbers
		// instance.
		// (Future generations: Please note the author of
		// the class.  Proves yet again that "visionary"
		// does not mean "good coder".  Grrr. )
		tokenizer.wordChars('a', 'z');
		tokenizer.wordChars('A', 'Z');
		tokenizer.wordChars(128 + 32, 255);
		tokenizer.whitespaceChars(0, ' ');
		tokenizer.commentChar('/');
		tokenizer.quoteChar('"');

		// Now do the stuff I want to do
		tokenizer.commentChar('#');
		tokenizer.quoteChar(JCFormatParserUtil.QUOTE);
		tokenizer.slashSlashComments(false);
		tokenizer.slashStarComments(false);
		tokenizer.eolIsSignificant(false);
		tokenizer.lowerCaseMode(false);
		tokenizer.whitespaceChars('+', '+');

		int rval = tokenizer.nextToken();
		if (rval == StreamTokenizer.TT_WORD &&
				tokenizer.sval != null &&
				tokenizer.sval.equalsIgnoreCase("table")) {
			format = JCTableEnum.STANDARD;
		}
	}
	catch (IOException ioe) {
	}

	return format;

}


}



