/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCStandardFormatParser.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:36 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table.data;

import java.io.Reader;
import java.io.IOException;
import java.io.StreamTokenizer;

import com.klg.jclass.util.JCFormatParserUtil;





import java.util.Vector;



/**
 * This class takes in a stream of data and parses
 * it under the assumption it is in "standard"
 * LiveTable format.  It then puts the data into
 * a <code>JCVectorDataSource</code> instance. <p>
 * The standard LiveTable format looks like this:
 * <pre>
 * TABLE 2 4 NOLABEL # 2 rows, 4 columns
 * 1 2 3 4           # row 1
 * 1 2 3 4           # row 2
 *</pre>
 * Or, with row and column labels,
 * <pre>
 * TABLE 2 4
 *         'Column 1' 'Column 2' 'Column 3' 'Column 4'
 * 'Row 1' 1          2          3          4
 * 'Row 2' 1          4          9          16
 * 'Row 3' 1          16         81         256
 * </pre>
 */
public class JCStandardFormatParser implements JCFileFormatParser {

/** Stream tokenizer used to parse input stream. */
protected StreamTokenizer		tokenizer;

/** Enum representing a parsing error involving data. */
public final static int PARSE_ERR_DATA_STYLE   = 1;
/** Enum representing a parsing error involving data values. */
public final static int PARSE_ERR_VALUE        = 2;
/** Enum representing a parsing error involving column labels. */
public final static int PARSE_ERR_COLUMN_LABEL = 3;
/** Enum representing a parsing error involving row labels. */
public final static int PARSE_ERR_ROW_LABEL    = 4;

/** Flag indicating whether parsing is in debug mode, which turns
 * on error messages on error out. */
protected boolean isDebug = false;
/** Flag indicating whether data source has labels. */
protected boolean isLabels = false;

/**
 * Creates and sets up a valid tokenizer active on the
 * provided stream.
 *
 * @param r reader open on an active stream
 */
protected void initTokenizer(Reader r) {
	tokenizer = new StreamTokenizer(r);
	tokenizer.commentChar('#');
	tokenizer.quoteChar(JCFormatParserUtil.QUOTE);
	tokenizer.slashSlashComments(false);
	tokenizer.slashStarComments(false);
	tokenizer.eolIsSignificant(false);
	tokenizer.lowerCaseMode(false);
	tokenizer.parseNumbers();
	tokenizer.whitespaceChars('+', '+');
}

/**
 * Reads and parses data in standard format and places it
 * inside a LiveTable data source.
 *
 * @param r reader open on an active stream
 * @param ds <code>JCVectorDataSource</code> instance to act as a
 * container for the data read in from r
 * @throws IOException
 *
 */
public void readAndParse(Reader r, JCVectorDataSource ds) throws IOException {
	boolean tokenError = false;
	int nrows = 0;
	int ncolumns= 0;

	initTokenizer(r);

	if(tokenizer.nextToken() != StreamTokenizer.TT_WORD) {
		tokenError = true;
	}
	else if(tokenizer.sval.equalsIgnoreCase("debug")) {
		isDebug = true;
		if(tokenizer.nextToken() != StreamTokenizer.TT_WORD) {
			tokenError = true;
		}
	}

	if (tokenError || !tokenizer.sval.equalsIgnoreCase("table")) {
		displayErrorMessage(tokenizer, PARSE_ERR_DATA_STYLE);
		throw new IOException("Data must start with TABLE");
	}

	nrows = (int)getNextDouble(tokenizer);
	ncolumns = (int)getNextDouble(tokenizer);

	if (tokenizer.nextToken() == StreamTokenizer.TT_WORD &&
	   tokenizer.sval.equalsIgnoreCase("nolabel")) {
		isLabels = false;
	}
	else {
		isLabels = true;
		tokenizer.pushBack();
	}

	if (isDebug) {
		System.out.println("Reading in TABLE format with "+nrows+
						   " rows and "+ncolumns+" columns with " +
						   (isLabels == true ? "labels" : "no labels"));
	}

	readArrayData(nrows,ncolumns, ds);

	tokenizer = null;
}

/**
 * Reads the data from the stream as if it is in table format.
 * <pre>
 * TABLE 2 4 NOLABEL # 2 rows, 4 columns
 * 1 2 3 4           # row 1
 * 1 2 3 4           # row 2
 * </pre>
 * Or, with row and column labels,
 * <pre>
 * TABLE 2 4
 *         'Column 1' 'Column 2' 'Column 3' 'Column 4'
 * 'Row 1' 1          2          3          4
 * 'Row 2' 1          4          9          16
 * 'Row 3' 1          16         81         256
 * </pre>
 * @param ds
 * @param ncolumns
 * @param nrows
 * @throws IOException
 */
private void readArrayData(int nrows, int ncolumns,
						   JCVectorDataSource ds) throws IOException {
	int row, col;

	Object value = null;
	Vector v = null;

	// if labels are set, read column labels
	if(isLabels) {
		for(col=0; col < ncolumns; col++) {
			value = JCFormatParserUtil.getNextToken(tokenizer);
			ds.getColumnLabels().addElement(value);
			if(isDebug) {
				System.out.println("Column Label "+col+" is "+value);
			}
		}
	}

	for(row=0; row < nrows; row++) {
		// if labels are set, read the row label
		if(isLabels) {
			value = JCFormatParserUtil.getNextToken(tokenizer);
			ds.getRowLabels().addElement(value);
			if(isDebug) {
				System.out.println("Row Label "+row+" is "+value);
			}
		}
		v = new Vector();
		for(col=0; col < ncolumns; col++) {
			value = JCFormatParserUtil.getNextToken(tokenizer);
			v.addElement(value);
			if(isDebug) {
				System.out.println("Value at row "+row+" col "+col+" is "+value);
			}
		}
		ds.getCells().addElement(v);
	}

	ds.setNumRows(nrows);
	ds.setNumColumns(ncolumns);
}

/**
 * Reads the next double value.
 * @param tokenizer
 * @throws IOException
 * @return
 */
protected static double getNextDouble(StreamTokenizer tokenizer) throws IOException {
	tokenizer.nextToken();

	if(tokenizer.ttype != StreamTokenizer.TT_NUMBER) {
		displayErrorMessage(tokenizer, PARSE_ERR_VALUE);
		throw new IOException("Error in data file format - "+tokenizer);
	}

	double rval = tokenizer.nval;
	tokenizer.nextToken();
	if(tokenizer.ttype == StreamTokenizer.TT_WORD &&
		tokenizer.sval != null &&
		(tokenizer.sval.charAt(0) == 'e' ||
		 tokenizer.sval.charAt(0) == 'E')) {
		if(tokenizer.sval.length() == 1) {
			tokenizer.nextToken();
			if(tokenizer.ttype == StreamTokenizer.TT_NUMBER) {
				rval = rval * Math.pow(10.0, tokenizer.nval);
			}
		}
		else {
			String s = tokenizer.sval;
			// Scientific notation
			int mult = Integer.valueOf(tokenizer.sval.substring(1)).intValue();
			rval = rval * Math.pow(10.0, mult);
		}
	}
	else {
		tokenizer.pushBack();
	}
	return rval;
}

/**
 * A method to display a parsing error message based on
 * an enum value.
 *
 * @see com.klg.jclass.util.JCFormatParserUtil#displayErrorMessage
 * @param tokenizer
 * @param nErrorIndex
 */
protected static void displayErrorMessage(StreamTokenizer tokenizer,
									   int nErrorIndex) {
	String strError = new String("");
	switch (nErrorIndex) {
	case PARSE_ERR_DATA_STYLE:
		strError += "Expect data type TABLE and not ";
		break;
	case PARSE_ERR_VALUE:
		strError += "Expect data value and not ";
		break;
	case PARSE_ERR_COLUMN_LABEL:
		strError += "Expect a column label and not ";
		break;
	case PARSE_ERR_ROW_LABEL:
		strError += "Expect a row label and not ";
		break;
	default:
		return;
	}
	JCFormatParserUtil.displayErrorMessage(tokenizer, strError);
}



}
