/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCVectorDataSource.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:36 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.table.data;

import com.klg.jclass.table.JCTableEnum;
import com.klg.jclass.table.JCTableDataEvent;
import java.io.Serializable;
import java.util.Vector;

/**
 * <CODE>JCVectorDataSource</CODE> is the main data source of
 * <code>JCTable</code> from which most other datasources are derived. It stores
 * label and cell values in a Vector of Vectors and provides a range of methods
 * for getting and setting the cell values.
 */
public class JCVectorDataSource extends AbstractVectorDataSource
                                implements Serializable
{

protected Vector			cells = createVector();
protected Vector			column_labels = createVector();
protected Vector			row_labels = createVector();

protected int				columns = 5;
protected int				rows = 10;

protected boolean			is_moving = false;

/**
 * Creates a new <CODE>JCVectorDataSource</CODE> instance.
 */
public JCVectorDataSource() {}

/**
 * Creates a new <CODE>JCVectorDataSource</CODE> instance.
 * @param c
 * @param columnLabels
 * @param data
 * @param r
 * @param rowLabels
 */
public JCVectorDataSource(int r, int c, Vector rowLabels, Vector columnLabels,
                          Vector data)
{
    columns = c;
    rows = r;
    cells = data;
    if (rowLabels != null) {
        row_labels = rowLabels;
    }
    if (columnLabels != null) {
        column_labels = columnLabels;
    }
}

/**
 * Gets the number of columns in the table from the data source's
 * <CODE>NumColumns</CODE> value.
 * @see #setNumColumns
 * @return
 */
public int getNumColumns()
{
    return(columns);
}

/**
 * Sets the current number of columns in the data source (default is 5).
 * This value does not affect the internal <CODE>CellValues</CODE> vector.
 * The <CODE>addColumn()</CODE> and <CODE>deleteColumns()</CODE> methods update
 * this property by the number of columns added/deleted.
 * If set to zero, no cells are displayed.
 * @param value
 */
public void setNumColumns(int value)
{
    columns = value;
    setDataChanged(0, 0, 0, 0, JCTableDataEvent.NUM_COLUMNS);
}

/**
 * Gets the number of rows in the table by retrieving the value of the data
 * source's <CODE>NumRows</CODE> property.
 * @see #setNumRows
 * @return
 */
public int getNumRows()
{
    return(rows);
}

/**
 * Sets the current number of rows in the data source (default is 10).
 * This value does not affect the internal <CODE>CellValues</CODE> vector.
 * The <CODE>addRow()</CODE> and <CODE>deleteRow()</CODE> methods update
 * this property by the number of rows added/deleted.
 * If set to zero, no cells are displayed.
 * @param value
 */
public void setNumRows(int value)
{
    rows = value;
    setDataChanged(0, 0, 0, 0, JCTableDataEvent.NUM_ROWS);
}

/**
 * Gets the label of a given column in the data source.
 * @see #setColumnLabel
 * @param column
 * @return
 */
public Object getTableColumnLabel(int column)
{
    if (column >= 0 && column < column_labels.size()) {
        return(column_labels.elementAt(column));
    }
    return(null);
}

/**
 * Gets the label of a given row in the data source.
 * @see #setRowLabel
 * @return
 * @param row
 */
public Object getTableRowLabel(int row)
{
    if (row >= 0 && row < row_labels.size()) {
        return(row_labels.elementAt(row));
    }
    return(null);
}

/**
 * Returns the value for the cell at the specified location.
 * @return value for the cell at the specified location
 * @param column
 * @param row
 */
public Object getTableDataItem(int row, int column)
{
    Object value =
        isValidCell(row, column) ? ((Vector)cells.elementAt(row)).elementAt(column) : null;
    return(value);
}

/**
 * Clears the values of all cells by setting them to null.
 * @see #setCells
 */
public void clearCells()
{
    setCells((Vector)null);
}

/**
 * Returns the internal cell values of the table.
 * @see #setCells
 * @return internal cell values
 */
public Vector getCells()
{
    return(cells);
}

/**
 * Gets the data source's <CODE>ColumnLabels</CODE> value.
 * @see #setColumnLabels
 * @return
 */
public Vector getColumnLabels()
{
    return(column_labels);
}

/**
 * Gets the data source's <CODE>RowLabels</CODE> value.
 * @see #setRowLabels
 * @return
 */
public Vector getRowLabels()
{
    return(row_labels);
}

/**
 * Returns <code>true</code> if the row and column constitute a valid cell.
 * @return <code>true</code> if the row and column constiture a valid cell
 * @param row The row index
 * @param column The column index
 */
public final static boolean isCell(int row, int column)
{
    return(row >= 0 && column >= 0);
}

/**
 * Returns <code>true</code> if the row and column constitute a valid column
 * label.
 * @return <code>true</code> if the row and column constitute a valid column
 * level
 * @param row The row index
 * @param column The column index
 */
public final static boolean isColumnLabel(int row, int column)
{
    return(column >= 0 && row == JCTableEnum.LABEL);
}

/**
 * Returns <code>true</code> if the row and column constitute a valid label.
 * @return <code>true</code> if the row and column constitute a valid label
 * @param row The row index
 * @param column The column index
 */
public final static boolean isLabel(int row, int column)
{
    return(row == JCTableEnum.LABEL ? column >= 0
                                    : (column == JCTableEnum.LABEL ? row >= 0
                                                                   : false));
}

/**
 * Returns <code>true</code> if the row and column constitute a valid row label.
 * @return <code>true</code> if the row and column constitute a valid row label
 * @param row The row index
 * @param column The column index
 */
public final static boolean isRowLabel(int row, int column)
{
    return(row >= 0 && column == JCTableEnum.LABEL);
}

/**
 * Returns <code>true</code> if the row and column constitute a valid cell.
 * @return <code>true</code> if the row and column constitute a valid cell
 * @param row The row index
 * @param column The column index
 */
public final boolean isValidCell(int row, int column)
{
    return(row >= 0 && row < cells.size() && cells.elementAt(row) != null &&
           column >= 0 && column < ((Vector)cells.elementAt(row)).size());
}

/**
 * Returns <code>true</code> if the row and column constitute a valid label.
 * @return <code>true</code> if the row and column constitute a valid label
 * @param r The row index
 * @param c The column index
 */
public final boolean isValidLabel(int r, int c)
{
    return(c == JCTableEnum.LABEL ? r < row_labels.size()
                                  : (r == JCTableEnum.LABEL ? c < column_labels.size()
                                                            : false));
}

/**
 * Specifies the contents of a column label in the data source.
 * @param column
 * @param value
 */
public void setColumnLabel(int column, Object value)
{
    setLabelValue(JCTableEnum.LABEL, column, value);
}

/**
 * Sets the column labels as an array of Strings.
 * @param labels
 */
public void setColumnLabels(String[] labels)
{
    setColumnLabels(createVector(labels));
}

/**
 * Sets the column labels as a Vector. Each element of the vector may be an
 * instance of a String, Image, Component, or other
 * Object. To clear the labels, call this method with a null argument.
 * @param labels
 */
public void setColumnLabels(Vector labels)
{
    if (labels == null) {
        column_labels = createVector();
    }
    else {
        column_labels = labels;
    }

    setDataChanged(JCTableEnum.ALL, JCTableEnum.ALL, 0, 0,
                   JCTableDataEvent.RESET);
}

/**
 * Specifies the contents a row label in the data source.
 * @param row
 * @param value
 */
public void setRowLabel(int row, Object value)
{
    setLabelValue(row, JCTableEnum.LABEL, value);
}

/**
 * Sets the row labels as an array of Strings.
 * @param labels
 */
public void setRowLabels(String[] labels)
{
    setRowLabels(createVector(labels));
}

/**
 * Sets the row labels as a Vector. Each element of the vector may be an
 * instance of a String, Image, Component, or other
 * Object. To clear the labels, call this method with a null argument.
 * @param labels
 */
public void setRowLabels(Vector labels)
{
    if (labels == null) {
        row_labels = createVector();
    }
    else {
        row_labels = labels;
    }

    setDataChanged(JCTableEnum.ALL, JCTableEnum.ALL, 0, 0,
                   JCTableDataEvent.RESET);
}

/**
 * Sets the value of a cell.
 * @param column
 * @param row
 * @param v
 */
public void setCell(int row, int column, Object v)
{
    if ((row == JCTableEnum.ALL ||
         row == JCTableEnum.ALLCELLS ||
         row >= 0) &&
        (column == JCTableEnum.ALL ||
         column == JCTableEnum.ALLCELLS ||
         column >= 0))
    {
        setCellValue(row, column, v);
    }
}

/**
 * Sets the cell values as a matrix of Strings.
 * To remove all values, call <CODE>clearCells()</CODE>.
 * @see #clearCells
 * @param v
 */
public void setCells(String[][] v)
{
    if (v == null) {
        setCells(createVector());
    }
    else {
        Vector cells = createVector(v.length);
        for (int i = 0; i < v.length; i++) {
            setElementAt(cells, i, createVector(v[i]));
        }
        setCells(cells);
    }
}

/**
 * Sets the cell values as a Vector of Vectors.
 * To remove all values, call <CODE>clearCells()</CODE>.
 * @param v
 */
public void setCells(Vector v)
{
    if (v == null) {
        cells = createVector();
    }
    else if (v instanceof Vector) {
        cells = v;
    }
    else {
        cells = createVector(v);
    }

    setDataChanged(JCTableEnum.ALL, JCTableEnum.ALL, 0, 0,
                   JCTableDataEvent.RESET);
}

/**
 * Private method gets the label value.
 * @return
 * @param column
 * @param row
 */
protected Object getLabelValue(int row, int column)
{
    if (row >= 0 && column == JCTableEnum.LABEL && row < row_labels.size()) {
        return(row_labels.elementAt(row));
    }

    if (column >= 0 &&
        row == JCTableEnum.LABEL &&
        column < column_labels.size())
    {
        return(column_labels.elementAt(column));
    }
    return(null);
}

/**
 * Private method sets the label value.
 * @param column
 * @param row
 * @param value
 */
protected void setLabelValue(int row, int column, Object value)
{
    if (!isLabel(row, column)) {
        return;
    }
    else if (row == JCTableEnum.LABEL) {
        if (setLabelValue(column_labels, column, value)) {
            setDataChanged(JCTableEnum.LABEL, column, 1, 0,
                           JCTableDataEvent.CHANGE_COLUMN_LABEL);
        }
    }
    else {
        if (setLabelValue(row_labels, row, value)) {
            setDataChanged(row, JCTableEnum.LABEL, 1, 0,
                           JCTableDataEvent.CHANGE_ROW_LABEL);
        }
    }
}

/**
 * Private method sets the label value. Returns <code>false</code> if the
 * original value is the same as the new value.
 * @param labels
 * @param position
 * @param value
 * @return <code>false</code> if the original value is the same as the new value
 */
private boolean setLabelValue(Vector labels, int position, Object value)
{
    if (position < 0) {
        return(false);
    }
    // Don't set if not different
    if (position < labels.size()) {
        Object string = labels.elementAt(position);
        if (string != null && value != null && string.equals(value)) {
            return(false);
        }
    }
    setElementAt(labels, position, value);
    // If setting the last label to null, discard excess
    if (value == null && position == labels.size() - 1) {
        strip(labels);
    }
    return(true);
}

///**
// * Private method that sets the number of rows and columns in the data
// * source.  Sends only one event to registered <CODE>TableDataListener</CODE>
// * objects.
// * @param columns
// * @param rows
// */
//private void setNumRowsColumns(int rows, int columns)
//{
//    this.rows = rows;
//    this.columns = columns;
//    setDataChanged(JCTableEnum.ALL, JCTableEnum.ALL, 0, 0,
//                   JCTableDataEvent.RESET);
//}

/**
 * Private method sets the cell values.
 * @param column
 * @param row
 * @param value
 */
private void setCellValue(int row, int column, Object value)
{
    Object cell_value;
    int i, j, i1 = 0, i2 = rows - 1, j1 = 0, j2 = columns - 1;

    if (row >= 0 && column >= 0) {
        i1 = i2 = row;
        j1 = j2 = column;
    }
    else {
        if (row == JCTableEnum.ALLCELLS) {
            row = JCTableEnum.ALL;
        }

        if (column == JCTableEnum.ALLCELLS) {
            column = JCTableEnum.ALL;
        }

        if (row == JCTableEnum.ALL) {
            if (column != JCTableEnum.ALL) {
                j1 = j2 = column;
            }
            i1 = 0;
            i2 = rows - 1;
        }

        if (column == JCTableEnum.ALL) {
            if (row != JCTableEnum.ALL) {
                i1 = i2 = row;
            }
            j1 = 0;
            j2 = columns - 1;
        }

        if (row != JCTableEnum.ALL && column != JCTableEnum.ALL) {
            i1 = i2 = row;
            j1 = j2 = column;
        }

        if (i1 < 0 || j1 < 0) {
            return;
        }
    }

    for (i = i1; i <= i2; i++) {
        for (j = j1; j <= j2; j++) {
            cell_value = isValidCell(row, column) ?
                ((Vector)cells.elementAt(row)).elementAt(column) : null;
            if (cell_value != null) {
                if (cell_value.equals(value)) {
                    continue;
                }
                if (isValidCell(i, j)) {
                    setElementAt(((Vector)cells.elementAt(i)), j, null);
                }
            }

            if (cells.size() < i + 1 || cells.elementAt(i) == null) {
                setElementAt(cells, i, createVector(j + 1));
            }
            setElementAt(((Vector)cells.elementAt(i)), j, value);

            setDataChanged(i, j, 1, 0, JCTableDataEvent.CHANGE_VALUE);
        }

        // if we are setting the last cell in row to null, discard excess values
        if (value == null && isValidCell(i, j) && j2 == cells.size() - 1) {
            strip(((Vector)cells.elementAt(i)));
            strip(cells);
        }
    }
}

/**
 *  Notifies all registered <CODE>TableDataListeners</CODE>
 *  of a data source change.
 * @param column
 * @param command
 * @param destination
 * @param num_affected
 * @param row
 */
public void setDataChanged(int row, int column, int num_affected,
                           int destination, int command)
{
    if (is_moving || !hasListeners()) {
        return;
    }
    JCTableDataEvent ev = new JCTableDataEvent(this, row, column, num_affected,
                                               destination, command);
    fireTableDataEvent(ev);
}

/**
 * Inserts a new column into the data source, shifting any cell values
 * to the right of the insertion.
 * @param position this is the column index before which to add the new column
 * (if set to <code>JCTableEnum.MAXINT</code>, the column is added after the
 * last column)
 * @param label	this is the column label (can be null)
 * @param values an array of objects to be inserted (can be null)
 * @return <code>false</code> if any argument is invalid
 */
public boolean addColumn(int position, Object label, Vector values)
{
    if (position == JCTableEnum.MAXINT) {
        position = columns;
    }
    else if (position < 0 || position > columns) {
        return(false);
    }

    return(addColumn(position, label, values, true));
}

protected boolean addColumn(int position, Object label, Vector values,
                            boolean shift_labels)
{
    // add the new column into the data structure
    if (values != null && cells.size() < values.size()) {
        setElementAt(cells, values.size() - 1, createVector());
    }

    for (int i = 0, j = 0; i < cells.size(); i++, j++) {
        Vector rows, old_rows = (Vector)cells.elementAt(i);
        if (old_rows == null) {
            setElementAt(cells, i, (rows = createVector()));
        }
        else {
            rows = old_rows;
        }

        if (position < rows.size()) {
            rows.insertElementAt(null, position);
        }
        Object value =
            (values != null && j < values.size()) ? values.elementAt(j) : null;
        setElementAt(rows, position, value);
    }

    // Shift the labels unless a JCLabelValueListener is registered
    if (shift_labels) {
        if (!(column_labels.size() == 0 || position >= column_labels.size())) {
            column_labels.insertElementAt(null, position);
        }
    }

    setNumColumns(columns + 1);
    if (label != null && shift_labels == true) {
        setColumnLabel(position, label);
    }

    setDataChanged(JCTableEnum.LABEL, position, 1, position,
                   JCTableDataEvent.ADD_COLUMN);
    return(true);
}

/**
 * Inserts a new row into the data source, shifting any cell values
 * below the insertion. The row labels will be shifted unless a
 * <CODE>JCLabelValueListener</CODE> has been registered.
 * @param position this is the row index before which to add the new row (if
 * set to <CODE>JCTableEnum.MAXINT</CODE>, the row is added after the last row)
 * @param label	this is the row label (can be null)
 * @param values an array of objects to be inserted (can be null)
 * @return <code>false</code> if any argument is invalid
 */
public boolean addRow(int position, Object label, Vector values)
{
    Vector lv = values;
    if (values != null && !(values instanceof Vector)) {
        lv = createVector(values);
    }

    if (position == JCTableEnum.MAXINT) {
        position = rows;
    }
    else if (position < 0 || position > rows) {
        return(false);
    }

    return(addRow(position, label, lv, true));
}

protected boolean addRow(int position, Object label, Vector values,
                         boolean shift_labels)
{
    // Add the new row into the data structure
    if (values != null) {
        if (position < cells.size()) {
            cells.insertElementAt(null, position);
        }
        setElementAt(cells, position, values);
    }

    // Shift the labels unless a JCLabelValueListener is registered
    if (shift_labels) {
        if (!(row_labels.size() == 0 || position >= row_labels.size())) {
            row_labels.insertElementAt(null, position);
        }
    }

    setNumRows(rows + 1);
    if (label != null && shift_labels == true) {
        setRowLabel(position, label);
    }

    setDataChanged(position, JCTableEnum.LABEL, 1, position,
                   JCTableDataEvent.ADD_ROW);

    return(true);
}

/**
 * Deletes columns from the data source. The column labels will be shifted
 * unless
 * a <CODE>JCLabelValueListener</CODE> has been registered.
 * @param position this is the first column number to delete
 * @param num_columns this is the number of columns to be deleted
 * @return <code>false</code> if any argument is invalid
 */
public boolean deleteColumns(int position, int num_columns)
{
    return(deleteColumns(position, num_columns, true));
}

protected boolean deleteColumns(int position, int num_columns,
                                boolean shift_labels)
{
    if (num_columns <= 0) {
        return(false);
    }

    if (position < 0 || position + num_columns > columns) {
        return(false);
    }

    // shift the labels unless a JCLabelValueListener is registered
    if (shift_labels) {
        if (!(column_labels.size() == 0 || position >= column_labels.size())) {
            removeElementsAt(column_labels, position, num_columns);
        }
    }

    // shift columns after ones being deleted
    Vector row_values;
    for (int i = 0; i < cells.size(); i++) {
        if ((row_values = (Vector)cells.elementAt(i)) != null) {
            removeElementsAt(row_values, position, num_columns);
        }
    }

    setNumColumns(columns - num_columns);
    setDataChanged(JCTableEnum.LABEL, position, num_columns, 0,
                   JCTableDataEvent.REMOVE_COLUMN);
    return(true);
}

/**
 * Deletes rows from the data source. The row labels will be shifted unless a
 * <CODE>JCLabelValueListener</CODE> has been registered.
 * @param position this is the first row number to delete
 * @param num_rows this is the number of rows to be deleted
 * @return <code>false</code> if any argument is invalid
 */
public boolean deleteRows(int position, int num_rows)
{
    return(deleteRows(position, num_rows, true));
}

protected boolean deleteRows(int position, int num_rows, boolean shift_labels)
{
    if (num_rows <= 0) {
        return(false);
    }

    if (position < 0 || position + num_rows > rows) {
        return(false);
    }

    // shift the labels unless a JCLabelValueListener is registered
    if (shift_labels) {
        if (!(row_labels.size() == 0 || position >= row_labels.size())) {
            removeElementsAt(row_labels, position, num_rows);
        }
    }

    // remove the rows
    removeElementsAt(cells, position, num_rows);

    setNumRows(rows - num_rows);
    setDataChanged(position, JCTableEnum.LABEL, num_rows, 0,
                   JCTableDataEvent.REMOVE_ROW);

    return(true);
}

/**
 * Moves a range of columns in the data source. The column labels will be moved
 * unless a
 * <CODE>JCLabelValueListener</CODE> has been registered.
 * @param source this is the first column in the range
 * @param num_columns this is the number of columns to move
 * @param destination this is the column number before which to insert the moved
 * columns
 * @return <code>false</code> if any argument is invalid
 */
public boolean moveColumns(int source, int num_columns, int destination)
{
    if (num_columns <= 0) {
        return(false);
    }

    if (source < 0 ||
        (source + num_columns > columns && destination != JCTableEnum.MAXINT))
    {
        return(false);
    }

    if (destination < 0 ||
        (destination > columns && destination != JCTableEnum.MAXINT) ||
        (source <= destination && destination < source + num_columns))
    {
        return(false);
    }

    if (source + num_columns == destination) {
        return(true);
    }

    return(moveColumns(source, num_columns, destination, true));
}

protected boolean moveColumns(int source, int num_columns, int destination,
                              boolean shift_labels)
{
    is_moving = true;

    if (destination == JCTableEnum.MAXINT) {
        destination = columns;
    }

    Vector column_copy = createVector(num_columns);
    Vector label_copy = createVector(num_columns);

    // Copy the columns
    for (int cnt = 0; cnt < num_columns; cnt++) {
        Vector values = createVector();
        for (int r = 0; r < rows; r++) {
            try {
                Vector v = ((Vector)(cells.elementAt(r)));
                values.addElement(v.elementAt(source + cnt));
            }
            catch (Exception e) {
                // No data, so make the data vector null
                // and get out of this loop
                values = null;
                r = rows;
            }
            //values.addElement(getTableDataItem(r,source+cnt));
        }
        column_copy.addElement(values);
        label_copy.addElement(getTableColumnLabel(source + cnt));
    }

    // Delete columns
    if (!deleteColumns(source, num_columns, shift_labels)) {
        return(false);
    }

    // insert into new location
    if (source < destination) {
        for (int cnt = 0; cnt < num_columns; cnt++) {
            if (!addColumn(destination + cnt - num_columns,
                           label_copy.elementAt(cnt),
                           (Vector)column_copy.elementAt(cnt), shift_labels))
            {
                return(false);
            }
        }
    }
    else {
        for (int cnt = 0; cnt < num_columns; cnt++) {
            if (!addColumn(destination + cnt, label_copy.elementAt(cnt),
                           (Vector)column_copy.elementAt(cnt), shift_labels))
            {
                return(false);
            }
        }
    }

    column_copy = null;
    label_copy = null;

    is_moving = false;

    setDataChanged(JCTableEnum.ALL, source, num_columns, destination,
                   JCTableDataEvent.MOVE_COLUMN);

    return(true);
}

/**
 * Moves a range of rows. The row labels will be shifted unless a
 * <CODE>JCLabelValueListener</CODE> has been registered.
 * @param source this is the first row in the range
 * @param num_rows this is the number of rows to move
 * @param destination this is the row number before which to insert the moved
 * rows
 * @return <code>false</code> if any argument is invalid
 */
public boolean moveRows(int source, int num_rows, int destination)
{
    if (num_rows <= 0) {
        return(false);
    }

    if (source < 0 || source + num_rows > rows) {
        return(false);
    }

    if (destination < 0 || destination > rows ||
        (source <= destination && destination < source + num_rows))
    {
        return(false);
    }

    if (source + num_rows == destination) {
        return(true);
    }

    return(moveRows(source, num_rows, destination, true));
}

protected boolean moveRows(int source, int num_rows, int destination,
                           boolean shift_labels)
{
    is_moving = true;

    Vector row_copy = createVector(num_rows);
    Vector label_copy = createVector(num_rows);

    // copy the rows
    boolean rowData = true;
    boolean labelData = true;
    for (int cnt = 0; cnt < num_rows; cnt++) {
        try {
            row_copy.addElement(cells.elementAt(source + cnt));
        }
        catch (Exception e) {
            rowData = false;
        }
        try {
            label_copy.addElement(getTableRowLabel(source + cnt));
        }
        catch (Exception e) {
            labelData = false;
        }
    }

    if (!rowData && !labelData) {
        return(false);
    }

    // delete rows
    if (!deleteRows(source, num_rows, shift_labels)) {
        return(false);
    }

    // insert into new location
    if (source < destination) {
        for (int cnt = 0; cnt < num_rows; cnt++) {
            Vector rowV = null;
            if (rowData) {
                rowV = (Vector)row_copy.elementAt(cnt);
            }
            if (!addRow(destination + cnt - num_rows, label_copy.elementAt(cnt),
                        rowV, shift_labels))
            {
                return(false);
            }
        }
    }
    else {
        for (int cnt = 0; cnt < num_rows; cnt++) {
            Vector rowV = null;
            if (rowData) {
                rowV = (Vector)row_copy.elementAt(cnt);
            }
            if (!addRow(destination + cnt, label_copy.elementAt(cnt),
                        rowV, shift_labels))
            {
                return(false);
            }
        }
    }

    row_copy = null;
    label_copy = null;

    is_moving = false;

    setDataChanged(source, JCTableEnum.ALL, num_rows, destination,
                   JCTableDataEvent.MOVE_ROW);
    return(true);
}

}
