/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: DataBoundSource.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:36 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.table.db;

import com.klg.jclass.table.data.AbstractDataSource;
import com.klg.jclass.table.EditableTableDataModel;

import com.klg.jclass.datasource.DataModel;
import com.klg.jclass.datasource.DataTableModel;
import com.klg.jclass.datasource.MetaDataModel;
import com.klg.jclass.datasource.BindingModel;

/**
 * External data source for <code>LiveTable</code> data binding.  Uses
 * a <code>TableDataBinding</code> instance to communicate with the
 * JClass DataSource.
 */
class DataBoundSource extends AbstractDataSource implements EditableTableDataModel {

TableDataBinding dataBinding;

/**
 * Creates a new <code>DataBoundSource</code> object,
 * which is bound to the root meta data level.
 */
DataBoundSource(DataTable table, DataModel dataModel) {
	this(table, dataModel, (MetaDataModel)null);
}

/**
 * Creates a new <code>DataBoundSource</code> object,
 * which is bound to the given <code>MetaDataModel</code>.
 */
DataBoundSource(DataTable table, DataModel dataModel, MetaDataModel metaDataModel) {
    this(table);
	dataBinding.setDataBinding(dataModel, metaDataModel);
}

/**
 * Creates a new <code>DataBoundSource</code> object for a
 * <code>DataModel</code> and binds the <code>MetaData</code> to
 * the <code>MetaDataModel</code> represented
 * by a "path" of <code>MetaDataModel</code>
 * descriptions separated by '|' (e.g. Orders|Customers).
 */
DataBoundSource(DataTable table, DataModel dataModel, String metaDescription) {
	this(table);
    dataBinding.setDataBinding(dataModel, metaDescription);
}

/**
 * Creates a new <code>DataBoundSource</code> object for
 * the data binding hierarchy string, starting from a
 * <code>JCTreeData</code> instance name and containing
 * full sequence of <code>MetaDataModel</code> levels starting
 * from the root to the level the table is bound with.
 */
DataBoundSource(DataTable table, String s) {
	this(table);
    dataBinding.setDataBinding(s);
}

/**
 * Creates a new <code>DataBoundSource</code> object
 * for the data binding hierarchy string, starting from a
 * <code>JCTreeData</code> instance name and containing
 * full sequence of <code>MetaDataModel</code> levels starting
 * from the root to the level the table is bound with.
 */
DataBoundSource(DataTable table, BindingModel binding) {
	this(table);
    dataBinding.setDataBinding(binding);
}

/**
 * Creates a new <code>DataBoundSource</code> with
 * <code>TableDataBinding</code> not bound to a datasource.
 */
private DataBoundSource(DataTable table) {
    dataBinding = new TableDataBinding(table, this);
}

/**
 * Returns the data binding hierarchy string, starting
 * from a <code>JCTreeData</code> instance name
 * and containing full sequence of <code>MetaDataModel</code>
 * levels starting from the root to the level
 * the table is bound with.
 *
 * @see #setDataBinding
 */
public String getDataBinding() {
	return dataBinding.getDataBinding();
}

/**
 * Sets the <code>DataBinding</code> property value to the table.
 *
 * Sets a table datasource for the data binding hierarchy string,
 * starting from a <code>JCTreeData</code>
 * instance name and containing full sequence of
 * <code>MetaDataModel</code> levels starting from the root
 * to the level the table is bound with.
 *
 * @see #getDataBinding
 */
public void setDataBinding(String s) {
	dataBinding.setDataBinding(s);
}

/**
 * Sets up parameters to bind the <code>DataBoundSource</code>
 * to a <code>MetaDataModel</code> in the
 * given <code>DataModel</code>.
 */
protected void setDataBinding(DataModel dataModel, MetaDataModel metaDataModel) {
	dataBinding.setDataBinding(metaDataModel.getBinding());
}

/**
 * Allows the DataBoundSource to clean up any references.
 */
public void finalize() {
	dispose();
}

/**
 * If the <code>DataBoundSource</code> is attached to a
 * <code>TreeData</code> object, remove it from
 * the <code>TreeData</code>'s <code>DataModelListener</code> list.
 */
public void dispose() {
	dataBinding.dispose();
}

/*
 * JClass TableData Implementation.
 * TableData methods are mapped to MetaDataModel/TableDataModel methods.
 *****************************************************************************/

/**
 * Returns the number of rows of data the data source contains.
 */
public int getNumRows() {
	return dataBinding.isDataAvailable() ? dataBinding.getNumRows() : 0;
}

/**
 * Returns the number of columns of data the data source contains.
 */
public int getNumColumns() {
	return dataBinding.isDataAvailable() ? dataBinding.getNumColumns() : 0;
}

/**
 * Returns the value for the row label at the specified row.
 * Labels use the <code>CellData</code> interface just like cells do.
 */
public Object getTableRowLabel(int row) {
	return dataBinding.isDataAvailable() ?
    	dataBinding.getTableRowLabel(row) : null;
}

/**
 * Returns the value for the column label at the specified column.
 * Labels use the <code>CellData</code> interface just like cells do.
 */
public Object getTableColumnLabel(int column) {
	return dataBinding.isDataAvailable() ?
		dataBinding.getTableColumnLabel(column) : null;
}

/**
 * Returns the value for the cell at the specified location.
 */
public Object getTableDataItem(int row, int column) {
	return dataBinding.isDataAvailable() ?
		dataBinding.getResultData(row, column) : null;
}

/*
 * JClass EditableTableData Implementation.
 *****************************************************************************/

/**
 * Set the cell value in the data source.
 */
public boolean setTableDataItem(Object value, int row, int column) {
	return dataBinding.isDataAvailable() ?
		dataBinding.updateCell(value, row, column) : false;
}

/*
 * Helper methods.
 *****************************************************************************/

/**
 * Returns the native <code>MetaData</code>.
 */
public Object getNativeMetaData() {
	return dataBinding.getNativeMetaData();
}

/**
 * Returns the <code>JavaColumnType</code> for the column.
 */
public int getJavaColumnType(String columnIdentifier) {
	return dataBinding.getJavaColumnType(columnIdentifier);
}

///**
// * Returns the bound <code>DataModel</code>.
// */
//public DataModel getDataModel() {
//	return dataBinding.getDataModel();
//}
//
///**
// * Returns the bound <code>MetaDataModel</code>.
// */
//public MetaDataModel getMetaDataModel() {
//	return dataBinding.getMetaDataModel();
//}
//
///**
// * Returns the <code>DataTableMgodel</code>.
// */
//public DataTableModel getDataTableModel() {
//	return dataBinding.getDataTableModel();
//}

}	// DataBoundSource
