/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCDebug.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:38 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util;

import java.io.PrintStream;

/**
 * A useful mechanism for managing assertions and debug prints.
 * A perl script is provided to remove references to this class
 * to produce production code.  See
 * <code>com.klg.jclass/util/scripts/rmdebug.pl</code>.<p>
 * The script removes any line containing
 * <code>JCDebug.&#60print|assert&#62</code> AND any
 * <code>import com.klg.jclass.util.JCDebug</code> statements that may exist.
 * Optionally you may call <code>setEnabled(false)</code> to globally
 * disable <code>JCDebug</code> printing.<p>
 * By calling <code>setEnabled(true)</code>, you can activate normal debug
 * output. If you wish to view code marked "mycode", then call
 * <code>setTag("mycode")</code>.<p>
 * Another useful construct is the concept of "levels".  The advantage of
 * this is that by calling, for example, <code>setLevel(2)</code> you will
 * avoid seeing any debug messages marked 3 or greater.  This is useful for
 * controlling the amount of detail you are viewing without removing the
 * debug information.<p>
 * <b>Note:</b> JClass distribution bytecode does <i>not</i> contain
 * any references to this class.
 */
public class JCDebug {

protected static boolean enabled = false;
protected static String tags[] = null;
protected static int level = Integer.MAX_VALUE;
protected static PrintStream stream = System.out;

/**
 * Controls global enable/disable of Debug printing.  By default this
 * is <code>false</code> and should be turned on to see the debug output.
 *
 * @param e Should Debug printing be enabled?
 */
public static void setEnabled(boolean e)
{
	enabled = e;
}

/**
 * Returns whether or not Debug output is enabled.
 *
 * @return whether Debug printing is enabled
 */
public static boolean isEnabled()
{
	return(enabled);
}

/**
 * Sets the output stream to use.  The default is <code>System.out</code>.
 *
 * @param pstream the new output stream
 */
public static void setPrintStream(PrintStream pstream)
{
	stream = pstream;
}

/**
 * Returns the <code>PrintStream</code> currently in use.
 *
 * @return the current output stream
 */
public static PrintStream getPrintStream()
{
	return(stream);
}

/**
 * Sets the level of debug information to print.  Output will only occur
 * for statements marked with the specified level or less.  Print statements
 * that do not specify a level are always printed.
 *
 * @param new_level the new level of debug information to print
 */
public static void setLevel(int new_level)
{
	level = new_level;
}

/**
 * Returns the level that determines what gets printed.
 *
 * @return the current level of debug information to print
 * @see #setLevel
 */
public static int getLevel()
{
	return(level);
}

/**
 * Sets the list of tags to match against when determining which debug
 * print statements to output.  If this list is null (the default), then
 * no tagged print statements will be output.
 *
 * @param new_tags the new list of tags
 */
public static void setTags(String new_tags[])
{
	tags = new_tags;
}

/**
 * Convienience method to allow setting of one tag.  This tag will replace any
 * old tags.
 *
 * @param tag the new tag
 */
public static void setTag(String tag)
{
	if (tag == null) {
		setTags(null);
	}
	else {
		String new_tags[] = new String[1];
		new_tags[0] = tag;
		setTags(new_tags);
	}
}

/**
 * Gets the tags currently used to filter output.
 *
 * @return the current list of tags to filter output
 */
public static String[] getTags()
{
	return(tags);
}

/**
 * Invoked to assert that a particular condition is <code>true</code>.  Useful
 * if you need a pre JDK 1.4 assert mechanism.
 *
 * @param condition the condition to assert <code>true</code> or
 * <code>false</code>
 */
public final static void assertThat(boolean condition)
{
	if (enabled == true && condition == false) {
		throw new AssertionException();
	}
}

/**
 * Invoked to assert that a particular condition is <code>true</code> and throw
 * an exception with the specified message if it is not.  Useful
 * if you need a pre JDK 1.4 assert mechanism.
 *
 * @param condition the condition to assert <code>true</code> or
 * <code>false</code>
 * @param message the message to print if the assertion fails
 */
public final static void assertThat(boolean condition, String message)
{
	if (enabled == true && condition == false) {
		throw new AssertionException(message);
	}
}

/**
 * Useful routine to force a stack trace at the current location by
 * generating and catching a mock <code>NullPointerException</code>.
 */
public static void printStackTrace()
{
    try {
        throw new NullPointerException("Forced Stack Trace");
    }
    catch (NullPointerException e) {
        e.printStackTrace(stream);
    }
}

/**
 * Useful routine to force a stack trace at the current location by
 * generating and catching a mock <code>NullPointerException</code>.
 *
 * @param s a String to display at the top of the trace
 */
public static void printStackTrace(String s)
{
	try {
        throw new NullPointerException(s);
    }
    catch (NullPointerException e) {
        e.printStackTrace(stream);
    }
}

/**
 * Prints a stack trace for the specified exception, but
 * only if the specified tag is currently enabled.
 *
 * @param ptag the "tag" to be used to determine whether to print the trace
 * @param t exception to display
 */
public static void printStackTrace(String ptag, Throwable t)
{
	if (!isPrintableTag(ptag)) {
		return;
	}
	t.printStackTrace(stream);
}

/**
 * Prints the text according to level and tag filter options.
 * @param plevel the "level" of the print command
 * @param ptag the "tag" associated with the print command (note that
 *     the tag must be in the tags list for the data to be output); a tag of
 *     null means that the line should always be printed
 * @param text the text to print
 */
public static void print(int plevel, String ptag, String text)
{
	if (isPrintableLevel(plevel) == false) {
		return;
	}

	if (isPrintableTag(ptag) == false) {
		return;
	}

	stream.print(text);
}

/**
 * Equivalent to calling <code>print(plevel, null, text)</code>.
 *
 * @param plevel the "level" of the print command
 * @param text the text to print
 */
public static void print(int plevel, String text)
{
	print(plevel, null, text);
}

/**
 * Equivalent to calling "print(0, ptag, text)".
 *
 * @param ptag the "tag" associated with the print command (note that
 *     the tag must be in the tags list for the data to be output); a tag of
 *     null means that the line should always be printed
 * @param text the text to print
 */
public static void print(String ptag, String text)
{
	print(0, ptag, text);
}

/**
 * Equivalent to calling <code>print(0, null, text)</code>.
 *
 * @param text the text to print
 */
public static void print(String text)
{
	print(0, null, text);
}

/**
 * Prints the text followed by a newline according to level and tag
 * filter options.
 *
 * @param plevel the "level" of the print command
 * @param ptag the "tag" associated with the print command (note that
 *     the tag must be in the tags list for the data to be output); a tag of
 *     null means that the line should always be printed
 * @param text the text to print
 */
public static void println(int plevel, String ptag, String text)
{
	if (isPrintableLevel(plevel) == false) {
		return;
	}

	if (isPrintableTag(ptag) == false) {
		return;
	}

	stream.println(text);
}

/**
 * Equivalent to calling <code>println(plevel, null, text)</code>.
 *
 * @param plevel the "level" of the print command
 * @param text the text to print
 */
public static void println(int plevel, String text)
{
	println(plevel, null, text);
}

/**
 * Equivalent to calling <code>println(0, ptag, text)</code>.
 *
 * @param ptag the "tag" associated with the print command (note that
 *     the tag must be in the tags list for the data to be output); a tag of
 *     null means that the line should always be printed
 * @param text the text to print
 */
public static void println(String ptag, String text)
{
	println(0, ptag, text);
}

/**
 * Equivalent to calling <code>println(0, null, text)</code>.
 * @param text the text to print
 */
public static void println(String text)
{
	println(0, null, text);
}

/**
 * If the passed-in level is less than or equal to the set level,
 * then we deem the level to be printable.
 *
 * @param plevel the level to be tested
 *
 * @return whether the level is printable
 */
protected static boolean isPrintableLevel(int plevel)
{
	if (enabled == false) {
		// Nothing is printable if we are not enabled
		return(false);
	}

	return(plevel <= level);
}

/**
 * If the passed-in tag is in the tags list or the tag is null, then
 * we deem the tag to be printable; otherwise it is not.
 *
 * @param tag the tag to be tested
 *
 * @return whether tag is printable
 */
protected static boolean isPrintableTag(String tag)
{
	if (enabled == false) {
		// Nothing is printable if we are not enabled
		return(false);
	}

	if (tag == null) {
		// Null tags are printable
		return(true);
	}

	if (tags == null) {
		// We have no tags and since we have no tags list the tag cannot
		// be printable
		return(false);
	}

	for (int i = 0; i < tags.length; i++) {
		// Compare the tag to a particular tag in the list
		if (tag != null && tag.equals(tags[i])) {
			// We have a match; so the tag is printable
			return(true);
		}
	}

	return(false);
}

}