/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCNumberUtil.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:38 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util;

import com.klg.jclass.util.swing.TextRenderer;

import java.awt.Color;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Dimension;

/**
 * <code>JCNumberUtil</code> is a collection of static methods and static
 * variables that are used throughout
 * <code>JCChart</code> and <code>JCGauge</code> components.
 */
public class JCNumberUtil {

/** Value of pi. */
public static final double M_PI    = 3.14159265358979323846;

/** Value of pi divided by 2. */

public static final double M_PI_2  = 1.57079632679489661923;

/** Value of pi multiplied by 2. */
public static final double M_2PI   = 6.28318530717958647692;

/** Square root of 2. */
public static final double M_SQRT2 = 1.41421356237309504880;

/** Angle unit enum value indicating angle unit is degrees. */
public final static int DEGREES = 1;

/** Angle unit enum value indicating angle unit is radians. */
public final static int RADIANS = 2;

/** Angle unit enum value indicating angle unit is gradians. */
public final static int GRADS = 3;

/** Default floating point value. */
public final static double DEFAULT_FLOAT = Double.MAX_VALUE;

/** Precalculated ln(10) is more accurate. */
public final static double LOG_TEN_BASE_E = 2.302585092994;

public static char DECIMAL = 0;

/** Gets the decimal point char.
 * @return
 */
public static char getDecimalPointChar()
{
	if (DECIMAL != 0) {
        return(DECIMAL);
    }
	return(DECIMAL = new Double(5.5).toString().charAt(1));
}

/**
 * Converts degrees to radians.
 * @param d angle value in degrees
 * @return converted value in radians
 */
public static final double degToRad(double d)
{
	return((d) * (M_PI / 180.0));
}

/**
 * Converts radians to degrees.
 * @param r angle value in radians
 * @return converted value in degrees
 */
public static final double radToDeg(double r)
{
	return((r) * (180.0 / M_PI));
}

/**
 * Checks a value to see if it is part of the angle unit
 * enum; <code>DEGREES</code>, <code>RADIANS</code>, or <code>GRADS</code>.
 * @param unit enum representing an angle unit
 * @return <code>true</code> if the unit enum is valid; <code>false</code>
 * otherwise
 */
public static boolean validUnit(int unit)
{
	return(unit == JCNumberUtil.DEGREES ||
           unit == JCNumberUtil.RADIANS ||
		   unit == JCNumberUtil.GRADS);
}

/**
 * Guarantees that an angle value lies within its valid range.
 * For example, all degree values should be between -360 and
 * +360.
 * @param unit unit for the angle value
 * @param angle angle value
 * @return
 */
public static double boundAngle(int unit, double angle)
{
    boolean negative = (angle < 0.0);
    double circum = 0.0;
    switch (unit) {
    case JCNumberUtil.DEGREES:
        circum = 360.0;
        break;

    case JCNumberUtil.GRADS:
        circum = 400.0;
        break;

    default:
        circum = M_2PI;
        break;
    }

    double boundAngle = fmod(Math.abs(angle), circum);
    if (negative && boundAngle != 0.0) {
        boundAngle = -boundAngle;
    }
    return(boundAngle);
}

/**
 * Java only provides a natural logarithm function, so this
 * method is necessary to perform log<sub>10</sub> operations.
 * @param a floating-point value
 * @return log-base-10 of the provided value
 */
public static double log10(double a)
{
	double baseE = Math.log(a);
	return(baseE / LOG_TEN_BASE_E);
}

/**
 * Returns log<sub>10</sub> of a value, rounded to the nearest integer
 * value.  Useful to the axis calculation code.
 * @return
 * @param a
 */
public static int intLog10(double a)
{
	return((int)precCorrect(1, log10(a)));
}

/**
 *	Returns a nice precision value for this range.
 *	It does not take into account font size, window
 *	size, etc. Instead, it uses log<sub>10</sub> of the range.
 * @return
 * @param range
 */
public static int nicePrecision(double range)
{
	if (range <= 0)	 {
		return 0;
	}

	int exp  = (int) Math.floor(log10(range));
	double f = range / pow10(exp);

	/* We need the extra digit near the lower end */
	int extra = (f < 3.0) ? 1 : 0;
	return (int)(-Math.floor(log10(range)) + extra);
}

/**
 * Returns a precision value based on the numspacing or tickspacing
 * value of an axis.  This method looks for the number of significant
 * (non-zero) digits either after or before a decimal place and uses
 * that as the default precision.
 * @return
 * @param numsp
 */
public static int tickBasedPrecision(double numsp)
{
    if (numsp <= 0)  {
        return(0);
    }

    String numspString = (new Double(numsp)).toString();
    int decPos = numspString.indexOf('.');
    // Num spacing has a decimal partition -- check it for non-zero digits
    if (decPos >= 0) {
        int numDecimals = numspString.length() - decPos - 1;
        for (int i = numspString.length() - 1; i > decPos; i--) {
            if (numspString.charAt(i) == '0') {
                numDecimals--;
            }
            else {
                i = decPos;
            }
        }
        if (numDecimals > 0) {
            return(numDecimals);
        }
   }
    // No decimal portion -- start at end of string
    else {
        decPos = numspString.length();
    }

    // No significant digits after decimal place, start looking in
    // main portion of the string
    int numZeroes = 0;
    for (int i = decPos - 1; i >= 0; i--) {
        if (numspString.charAt(i) == '0') {
            numZeroes++;
        }
        else {
            i = -1;
        }
    }
    return(-numZeroes);
}

/**
 * Ensures that a value is within a provided range.
 * @param x value that requires clamping
 * @param lo low value in the range
 * @param hi high value in the range
 * @return clamped value that lies between lo and hi
 */
public static double clamp(double x, double lo, double hi)
{
	return((x) > (hi) ? (hi) : ((x) < (lo) ? (lo) : (x)));
}

/**
 * Ensures that a value is within a provided range.
 * @param x value that requires clamping
 * @param lo low value in the range
 * @param hi high value in the range
 * @return clamped value that lies between lo and hi
 */
public static int clamp(int x, int lo, int hi)
{
	return((x) > (hi) ? (hi) : ((x) < (lo) ? (lo) : (x)));
}

/**
 * Ensures that a value is within a provided range.
 * @param x value that requires clamping
 * @param lo low value in the range
 * @param hi high value in the range
 * @return clamped value that lies between lo and hi
 */
public static long clamp(long x, long lo, long hi)
{
	return((x) > (hi) ? (hi) : ((x) < (lo) ? (lo) : (x)));
}

/**
 * Truncates a number to the specified precision.
 * If the number is negative, takes the ceiling of
 * its negated value instead.
 * @param prec precision value
 * @param value value to truncate
 * @return value truncated at the specified precision
 */
public static double precFloor(int prec, double value)
{
	double rval = value;
	if (rval < 0) {
		rval = -precCeil(prec, -rval);
	}
	else {
		if (prec >= 0) {
			String buffer = internalFormat(rval, prec+1);
			int sz = buffer.length();
			if (sz >= 1) {
				StringBuffer sb = new StringBuffer(buffer);
				sb.setCharAt(sz-1, '0');
				buffer = new String(sb);
			}
			double old_value = rval;
			rval = Double.valueOf(buffer).doubleValue();
			// Handle multiple rounding propagation
			// ex. xrt_format(0.95, 1, buffer) will produce 1.0 when 0 is
			// desired
			if (rval > old_value) {
				rval -= pow10(-prec);
			}
		}
		else {
			double tens;
			tens = pow10(-prec);
			rval = (tens * Math.floor(rval/tens));
		}
	}
	return(rval);
}


/**
 * Takes the ceiling of a number to the specified precision.
 * If the number is negative, takes the floor of
 * its negated value instead.
 * @param prec precision value
 * @param value value to truncate
 * @return ceiling of value at the specified precision
 */
public static double precCeil(int prec, double value)
{
	double rval = value;
	if (rval < 0) {
		rval = -precFloor(prec, -rval);
	}
	else {
		if (prec >= 0) {
			char[] buffer = internalFormat(rval, prec + 1).toCharArray();
			int i = buffer.length - 1;
			if (buffer[i] != '0') {
				buffer[i] = '0';
				for (i--; i >= 0; i--) {
					if (buffer[i] == getDecimalPointChar()) continue;
					if (buffer[i] != '9') {
						buffer[i] += 1;
						break;
					}
					else {
						buffer[i] = '0';
					}
				}
				if (i < 0) {
					char oldbuff[] = buffer;
					buffer = new char[oldbuff.length+1];
					i = oldbuff.length;
					for (; i > 0; i--) {
						buffer[i] = oldbuff[i-1];
					}
					buffer[0] = '1';
				}
			}
			double old_value = rval;
			rval = Double.valueOf(new String(buffer)).doubleValue();
			// Handle multiple rounding propagation
			// ex. xrt_format(1.01, 1, buffer) will produce 1.0 when 2 is
			// desired
			if (rval < old_value) {
				rval += pow10(-prec);
			}

		}
		else {
			double tens = pow10(-prec);
			rval = tens * (Math.ceil(rval/tens));
		}
	}
	return(rval);
}

/**
 * Formats a value for display on a logarithmic axis. <br><B>Note:</b>
 * This routine currently calls <code>JCNumberUtil.format()</code>.
 * @param exponent exponent of log value
 * @return formatted String
 */
public static String logFormat(int exponent) {
	double value = pow10(exponent);
	return format(value, -exponent);
}

/**
 * Formats a value for display.  Performs the equivalent of
 * <code>sprintf(buffer, "%.*f", ndigit, v);</code>
 * @param v value to format
 * @param ndigit number of digits to appear in the formatted
 * String
 * @return formatted String
 */
public static String format(double v, int ndigit)
{
	// First we have to round it
	double value = precCorrect(ndigit, v);

	String s = String.valueOf(value);
	if (s.indexOf('e') != -1 || s.indexOf('E') != -1) {
        return(s);
    }
	int decLocation = findDecimal(s);
	if (ndigit > 0) {
		if (decLocation < 0) {
			// No decimal point, but there should be.
			StringBuffer sb = new StringBuffer(s);
			sb.append(getDecimalPointChar());
			for (int i = 0; i < ndigit; i++) {
                sb.append('0');
            }
			s = new String(sb);
		}
		else {
			// Pad or clip?
			int numDec = s.length() - decLocation - 1;
			int extraZeroes = numDec-ndigit;
			if (extraZeroes < 0) {
				// Add zeroes
				StringBuffer sb = new StringBuffer(s);
				for (int i = 0; i > extraZeroes; i--) {
                    sb.append('0');
                }
				s = new String(sb);
			}
			else if (extraZeroes > 0) {
				// Remove zeroes
				s = s.substring(0, s.length() - extraZeroes);
			}
		}
	}
	else if (ndigit < 0) {
		int end = (decLocation < 0) ? s.length() : decLocation;
		int numCutoff = end + ndigit;
		if (numCutoff >= 0) {
			StringBuffer sb = new StringBuffer(s.substring(0, numCutoff));
			for (int i = 0; i < end-numCutoff; i++) {
                sb.append('0');
            }
			s = new String(sb);
		}
		else {
			s = new String("0");
		}
	}
	else {
		if (decLocation >= 0) s = s.substring(0, decLocation);
	}

	return(s);
}

public static String internalFormat(double v, int ndigit)
{
	String s = format(v, ndigit);
	if (s.indexOf('e') != -1 || s.indexOf('E') != -1) {
        s = convertFromSN(s);
    }
	return(s);
}

public static String convertFromSN(String s)
{
	int indx = s.indexOf('e');
	if (indx < 0) {
        indx = s.indexOf('E');
    }
	if (indx < 0) {
        return(s);
    }

	int tens = 0;
	if (indx + 1 < s.length()) {
		String tTens = s.substring(indx+1);
		if (tTens.charAt(0) == '+') {
            tens = 0;
            if (tTens.length() > 1) {
                tens = Integer.valueOf(tTens.substring(1)).intValue();
            }
        }
		else {
			tens = Integer.valueOf(tTens).intValue();
        }
	}
	String num = s.substring(0, indx);
	if (tens == 0) {
        return(num);
    }

	StringBuffer rval = new StringBuffer();
	char c = num.charAt(0);
	switch (c) {
	case '-':
		rval.append('-');
	case '+':
		num = num.substring(1);
	}

	if (tens < 0) {
		rval.append("0.");
		for (int i = -1; i > tens; i--) {
            rval.append('0');
        }
		rval.append(num.charAt(0));
		for (int j = 2; j < num.length(); j++) {
            rval.append(num.charAt(j));
        }
	}
	else {
		rval.append(num.charAt(0));
		int i = 0;
		for (; i < tens; i++) {
 			if ((i + 2) < num.length()) {
				rval.append(num.charAt(i + 2));
            }
			else {
				rval.append('0');
            }
        }

		if ((i + 2) < num.length()) {
			rval.append(getDecimalPointChar());
			for (; (i + 2) < num.length(); i++) {
                rval.append(num.charAt(i + 2));
            }
		}
	}
	return(new String(rval));
}

/**
 * Finds the decimal point in a String.
 * @param str String representation of a number
 * @return index in array of the decimal point; -1 if no
 * decimal point was found
 */
private static int findDecimal(String str)
{
	if (str == null) {
        return(-1);
    }
	return(str.indexOf(getDecimalPointChar()));
}


/**
 * Converts an angle value from one unit to another.
 * Units may be <code>DEGREES</code>, <code>RADIANS</code>, or
 * <code>GRADS</code>.
 * @param oldUnit unit of the provided angle value
 * @param newUnit unit of the new angle value
 * @param angle angle value
 * @return converted angle value
 */
public static double convertAngle(int oldUnit, int newUnit, double angle)
{
	double newAngle = angle;

	switch (newUnit) {
	case DEGREES:
		switch (oldUnit) {
        case RADIANS:
            newAngle = angle * 180.0 / M_PI;
            break;
        case GRADS:
            newAngle = angle * 0.9;
            break;
        default:
            newAngle = angle;
            break;
        }
        break;

    case GRADS:
        switch (oldUnit) {
        case RADIANS:
            newAngle = angle * 200.0 / M_PI;
            break;
        case DEGREES:
            newAngle = angle / 0.9;
            break;
        default:
            newAngle = angle;
            break;
        }
        break;

    default:
        switch (oldUnit) {
        case DEGREES:
            newAngle = angle * M_PI / 180.0;
            break;
        case GRADS:
            newAngle = angle * M_PI / 200.0;
            break;
        default:
            newAngle = angle;
            break;
        }
        break;
    }

    return(newAngle);
}

final static String tens_str[] = {
	"0","00","000","0000","00000","000000",
	"0000000", "00000000", "000000000", "0000000000", "00000000000",
	"000000000000"
};

static double tens_val[] = {
	1E-12, 1E-11, 1E-10, 1E-9, 1E-8, 1E-7, 1E-6, 1E-5, 1E-4, 1E-3, 1E-2, 1E-1,
	1E0, 1E1, 1E2, 1E3, 1E4, 1E5, 1E6, 1E7, 1E8, 1E9, 1E10, 1E11, 1E12
};

/**
 * 10<sup><i>x</i></sup> function that makes use of a predefined array
 * (<code>tens_val</code>).
 * @param x <i>x</i> in 10<sup><i>x</i></sup>
 * @return requested power of 10
 */
public static double pow10(int x)
{
	int index = 12 + x;
	if (index >= 0 && index < tens_val.length) {
		return(tens_val[index]);
    }
	return(Math.pow(10.0, (double)x));
}


/**
 * Rounds a number to the specified precision.
 * @param prec requested precision
 * @param value number to be rounded
 * @return rounded value
 */
public static double precCorrect(int prec, double value)
{
	// Keep the same behaviour as JDK 1.0.2
	if (value == Double.NEGATIVE_INFINITY) {
		return(0);
    }

    // I know this looks a bit funny, but it's necessary to prevent some
    // strange rounding errors that come in with division and negative
    // exponents
    if (prec >= 0) {
        double tens = pow10(prec);
        return(Math.floor(value * tens + 0.5) / tens);
    }
    else {
        double tens = pow10(-prec);
        return(Math.floor(value / tens + 0.5) * tens);
    }
}

/**
 * Returns the maximum of two floating point values.  Faster
 * than the naive <i>Math.max()</i>.
 * @param a first floating point value
 * @param b second floating point value
 * @return maximum of a and b
 */
public static double max(double a, double b)
{
	return(a > b ? a : b);
}

/**
 * Returns the minimum of two floating point values.  Faster
 * than the naive <i>Math.min()</i>.
 * @param a first floating point value
 * @param b second floating point value
 * @return minimum of a and b
 */
public static double min(double a, double b)
{
	return(a < b ? a : b);
}

/**
 * Returns the absolute value of a floating point number.
 * Faster than the naive <i>Math.abs()</i>.
 * @param a floating point value
 * @return absolute value of a
 */
public static double abs(double a)
{
	return((a > 0.0) ? a : -a);
}

/**
 * Returns the absolute value of an integer number.
 * Faster than the naive <i>Math.abs()</i>.
 * @param a floating point value
 * @return absolute value of a
 */
public static int abs(int a)
{
	return((a > 0) ? a : -a);
}

/**
 * Returns the floating-point remainder of the division of v1 by v2.
 * @param v1 floating-point dividend
 * @param v2 floating-point divisor
 * @return floating-point remainder, Integer.MAX_VALUE if the divisor is zero
 */
public static double fmod(double v1, double v2)
{
	double remainder = Integer.MAX_VALUE;
	if (v2 != 0.0) {
		int n = ifloor(v1 / v2);
		remainder = v1 - v2 * n;
	}
	return(remainder);
}

/**
 * Debug method that displays a call stack at a particular point.
 */
public static void trace()
{
 	try {
		throw new ArrayIndexOutOfBoundsException("");
 	}
     catch (Exception e) {
 	    e.printStackTrace(System.out);
 	}
}

/**
 * Calculates the distance between two points.
 * @param x1 first <i>x</i> point
 * @param y1 first <i>y</i> point
 * @param x2 second <i>x</i> point
 * @param y2 second <i>y</i> point
 * @return distance between the two points
 */
public static int distance(int x1, int y1, int x2, int y2)
{
	int srun  = x2 - x1;
	int srise = y2 - y1;
	return((int)Math.sqrt(srun * srun + srise * srise));
}

/**
 * Calculates the distance betwen a point and a rectangle.
 * @param x <i>x</i> point
 * @param y <i>y</i> point
 * @param rect rectangle
 * @return distance between the point and the rectangle
 */
public static int distance(int x, int y, Rectangle rect)
{
	int xdist = distance(x, rect.x, rect.width);
	int ydist = distance(y, rect.y, rect.height);
	int sqr = xdist * xdist + ydist * ydist;
	if (sqr <= 0) {
        return(0);
    }
	return((int)Math.sqrt(sqr));

}

/**
 * Calculates the separation between an integer
 * value and the nearest boundary of an integer range.
 * @param p integer value
 * @param start start point of range
 * @param width end point of range
 * @return separation between the integer value and
 * the specified range
 */
public static int distance(int p, int start, int width)
{
	if (p < start) {
        return(start - p);
    }
	if (p > start + width) {
        return(p - (start + width));
    }
	return(0);
}

/* Used for <code>distTrap()</code>. */
/** <code>pix_x</code> lies between (x1, x2). */
final public static int XINSIDE     = 1;

/** <code>pix_x</code> lies outside <i>x1</i>. */
final public static int XBYX1       = 2;

/** <code>pix_x</code> lies outside <i>x2</i>. */
final public static int XBYX2       = 3;

/** <code>pix_y</code> lies within top and bottom line. */
final public static int YINSIDE     = 10;

/** <code>pix_y</code> lies outside top line. */
final public static int YBYTOP      = 11;

/** <code>pix_y</code> lies by origin. */
final public static int YBYORIGIN   = 12;

/**
 *  Finds the distance from the passed-in point to the specified trapezoid.
 *  Determines the distance as follows:
 * <ul>
 * <li> if the point is inside, distance is zero</li>
 * <li>if inside the <i>x</i> range, but not the <i>y</i> range, distance is
 *  the vertical distance to the top line</li>
 * <li> if outside the <i>x</i> range, extrapolates the
 *  curve to that <i>x</i>-pixel - if within that extended trapezoid and the
 *  <i>y</i>-pixel is not above the corresponding side of the trapezoid,
 *  distance is to nearest trapezoid side; otherwise, distance is to
 *  nearest corner on the trapezoid</li>
 * </ul>
 * <p>
 * The trapezoid is defined by a top edge, two vertical sides, and a
 * horizontal line at the origin of the <i>y</i> axis.
 *
 *  @param pix_x the <i>x</i> value of the point
 *  @param pix_y the <i>y</i> value of the point
 *  @param yorigin the origin of the <i>y</i> axis
 *  @param x1 the <i>x</i> value of the first point that defines the
 *  trapezoid's top edge
 *  @param x2 the <i>x</i> value of the second point that defines the
 *  trapezoid's top edge
 *  @param y1 the <i>y</i> value of the first point that defines the
 *  trapezoid's top edge
 *  @param y2 the <i>y</i> value of the second point that defines the
 *  trapezoid's top edge
 * @return
 */
public static long distTrap(int pix_x, int pix_y, int yorigin,
                            double x1, double x2, double y1, double y2)
{
    long    yprime;
    long    xspot;
    long    yspot;

    /*
     *  Determines the <i>Y</I> coordinate where this x-pixel would
     *  intersect the top line.  Then, determines where the
     *  point falls with respect to the two bounding values
     *  on that axis.  Finally, handles all the cases.
     */

    if (x1 == x2) {
        /*
         *  Any large number will do; you can't
         *  pick inside a 0-pixel wide trapezoid.
         */
        yprime = 32000;
	}
    else {
        yprime = (long)(y1 + (pix_x - x1) * ((y1-y2)/(x1-x2)));
    }

    if ((x1 <= pix_x && pix_x <= x2) || (x2 <= pix_x && pix_x <= x1)) {
        xspot = XINSIDE;
    }
    else if ((pix_x <= x1 && x1 <= x2) || (pix_x >= x1 && x1 >= x2)) {
        xspot = XBYX1;
    }
    else {
        xspot = XBYX2;
    }

    if ((yprime <= pix_y && pix_y <= yorigin) ||
        (yorigin <= pix_y && pix_y <= yprime))
    {
        yspot = YINSIDE;
    }
    else if ((yprime <= yorigin && yorigin <= pix_y) ||
             (pix_y <= yorigin && yorigin <= yprime))
    {
        yspot = YBYORIGIN;
    }
    else {
        yspot = YBYTOP;
    }
    double dist = 0;
    switch((int)yspot) {
    case YINSIDE:
        switch((int)xspot) {
        case XINSIDE:           /* within the trapezoid */
            dist = 0;
            break;
        case XBYX1:             /* outside x1 vertical side */
            if ((yorigin <= pix_y && pix_y <= y1) ||
                (y1 <= pix_y && pix_y <= yorigin))
            {
                dist = Math.abs(pix_x - (int)x1);
            }
            else {
                dist = (pix_y - y1)*(pix_y - y1) + (pix_x - x1)*(pix_x - x1);
                dist = Math.sqrt(dist);
            }
            break;
        case XBYX2:             /* outside x2 vertical side */
            if ((yorigin <= pix_y && pix_y <= y2) ||
                (y2 <= pix_y && pix_y <= yorigin))
            {
                dist = Math.abs(pix_x - (int)x2);
            }
            else {
                dist = (pix_y - y2)*(pix_y - y2) + (pix_x - x2)*(pix_x - x2);
                dist = Math.sqrt(dist);
            }
            break;
        }
        break;
    case YBYTOP:
        switch((int)xspot) {
        case XINSIDE:           /* close to top side */
            dist = (double)Math.abs(pix_y - (int)yprime);
            break;
        case XBYX1:             /* closest to (x1, y1) */
            dist = (pix_y - y1)*(pix_y - y1) + (pix_x - x1)*(pix_x - x1);
            dist = Math.sqrt(dist);
            break;
        case XBYX2:             /* closest to (x2, y2) */
            dist = (pix_y - y2)*(pix_y - y2) + (pix_x - x2)*(pix_x - x2);
            dist = Math.sqrt(dist);
            break;
        }
        break;
    case YBYORIGIN:
        switch((int)xspot) {
        case XINSIDE:           /* outside origin (bottom) side */
            dist = Math.abs(pix_y - yorigin);
            break;
        case XBYX1:             /* closest to (x1, yorigin) */
            dist = (pix_y - yorigin)*(pix_y - yorigin) +
                   (pix_x - x1)*(pix_x - x1);
            dist = Math.sqrt(dist);
            break;
        case XBYX2:             /* closest to (x2, yorigin) */
            dist = (pix_y - yorigin)*(pix_y - yorigin) +
                   (pix_x - x2)*(pix_x - x2);
            dist = Math.sqrt(dist);
            break;
        }
    }

    return((long)dist);
}

/**
 * Takes an array of points and converts it into a <code>Polygon</code> object.
 * @param points array of <code>java.awt.Point</code> objects
 * @return polygon representing the points
 */
public static Polygon pointsToPolygon(Point[] points)
{
    Polygon poly = new Polygon();

    for (int i = 0; i < points.length; i++) {
        if (i == 0) {
            poly.addPoint(points[i].x, points[i].y);
        }
        else {
            poly.addPoint(poly.xpoints[i-1] + points[i].x,
						  poly.ypoints[i-1] + points[i].y);
        }
    }
	return(poly);
}

private static final Color WHITE_BRIGHTER = new Color(200, 200, 200);
private static final Color WHITE_DARKER = new Color(140, 140, 140);
private static final Color BLACK_BRIGHTER = new Color(125, 125, 125);
private static final Color BLACK_DARKER = new Color(75, 75, 75);
private static final double FACTOR = 0.5;

/**
 * Calculates the color even when it is saturated (for example,
 * when it is black or white).
 * @return
 * @param color
 */
public static Color brighter(Color color)
{
    if (color.equals(Color.white)) {
        return(WHITE_BRIGHTER);
    }
    else if (color.equals(Color.black)) {
        return(BLACK_BRIGHTER);
    }
    else {
		// Instead of calling Color.brighter() I do my own calculation
		// for a lighter color.  This calc brightens up dark colors very
		// quickly while lightening up bright colors only a bit.
        int red = color.getRed();
        red += (int) ((255 - red) * FACTOR);
        int blue = color.getBlue();
        blue += (int) ((255 - blue) * FACTOR);
        int green = color.getGreen();
        green += (int) ((255 - green) * FACTOR);

        return(new Color(Math.min(red, 255),
                         Math.min(green, 255),
                         Math.min(blue, 255)));
    }
}

/** Calculates the color even when it is saturated (for example,
 * when it is black or white).
 * @return
 * @param color
 */
public static Color darker(Color color)
{
    if (color.equals(Color.white)) {
        return(WHITE_DARKER);
    }
    else if (color.equals(Color.black)) {
        return(BLACK_DARKER);
    }
    else {
        return(color.darker());
    }
}

/** Determines whether two colors are close to the same color.  Used
 *  to determine whether two colors should be used as a foreground/background
 *  combination in default cases.
 *  @param c1 one color to check
 *  @param c2 the other color to check
 *  @return <code>true</code> if both colors are close; <code>false</code>
 *  otherwise.
 */
public static boolean colorsNear(Color c1, Color c2)
{
	if (c1 == null || c2 == null) {
		return(false);
    }

	if (Math.abs(c1.getRed() - c2.getRed()) <= 15      &&
		Math.abs(c1.getGreen() - c2.getGreen()) <= 15  &&
		Math.abs(c1.getBlue() - c2.getBlue()) <= 15)
    {
		return(true);
	}
	return(false);
}

/**
 * Takes the ceiling of a floating point value.
 * @return
 * @param val
 */
public static int iceil(double val)
{
	int i = 0;
	if (val < 0) {
		i = -ifloor(-val);
	}
	else {
		i = (int)Math.ceil(val);
	}
	return(i);
}

/**
 * Takes the floor of a floating point value.
 * @return
 * @param val
 */
public static int ifloor(double val)
{
	int i = 0;
	// Keep the same behaviour as JDK 1.0.2
	if (val == Double.NEGATIVE_INFINITY) {
		return(i);
    }
	if (val < 0) {
		i = -iceil(-val);
	}
	else {
		i = (int)(val);
	}
	return(i);
}

/**
 * Calculates an appropriate error based on a precision value.
 * @return
 * @param base
 */
public static double calcError(int base)
{
	return(0.5 * JCNumberUtil.pow10(-base-1));
}

/**
 * Locates an <i>x</i> value within given <i>x</i> bounds.
 * @param x  the value to locate
 * @param lx the lower bound to check against
 * @param ux the upper bound to check against
 * @return 1 if below lower bound, 2 if above upper bound, 0 if within bounds
 */
public static int locateX(double x, double lx, double ux)
{
    int first  = ((x > ux ? 1 : 0) << 1);
	int second = ((x < lx ? 1 : 0) << 0);
	return(first | second);
}

/**
 * Locates a <i>y</i> value within given <i>y</i> bounds.
 * @param y  the value to locate
 * @param ly the lower bound to check against
 * @param uy the upper bound to check against
 * @return 4 if below lower bound, 8 if above upper bound, 0 if within bounds
 */
public static int locateY(double y, double ly, double uy)
{
    int first  = ((y > uy ? 1 : 0) << 3);
	int second = ((y < ly ? 1 : 0) << 2);
	return(first | second);
}

/**
 * Locates a point within given <i>x</i> and <i>y</i> bounds.
 * @param x  the <i>x</i> value of the point to locate
 * @param y  the <i>y</i> value of the point to locate
 * @param lx the lower <i>x</i> bound to check against
 * @param ux the upper <i>x</i> bound to check against
 * @param ly the lower <i>y</i> bound to check against
 * @param uy the upper <i>y</i> bound to check against
 * @return 0 if within all bounds; otherwise, returns a combination of the
 * following values: 1 if below lower <i>x</i> bound, 2 if above upper
 * <i>x</i> bound, 4 is below lower <i>y</i> bound, 8 if above upper
 * <i>x</i> bound
 */
public static int locateXY(double x, double y, double lx, double ux,
                           double ly, double uy)
{
    return(locateY(y, ly, uy) | locateX(x, lx, ux));
}

/** Checks for HTML-encoded text.  The check is the same as that in
 * <code>javax.swing.plaf.basic.BasicHTML.isHTMLString()</code>.  That call
 * cannot be used because the class isn't public, so it is repeated here.
 * For a String to be HTML-encoded, it needs to begin with "&#60html&#62".
 * @param s the String to check for HTML encoding.
 * @deprecated Use com.klg.jclass.util.swing.TextRenderer.isHTML() instead
 * @return <code>true</code> if String is encoded in html; <code>false</code>
 * otherwise
 */
public static boolean isHTML(String s)
{
    return(TextRenderer.isHTML(s));
}

/**
 * Counts the occurrences of String token within the larger target String.
 * @return
 * @param target
 * @param token
 */
public static int countInString(String token, String target)
{
	int count = 0;

    int textIndex = 0;
    int tmpIndex = 0;
    while ((tmpIndex = target.indexOf(token, textIndex)) >= 0) {
        count++;
        textIndex = tmpIndex + 1;
    }
	return(count);
}

/**
 * Fudges the size of a label with italic fonts by making it a little bigger
 * so that all of the italicized font can be seen.  This is used to get around
 * a jdk bug where italics aren't taken into account when calculating a label
 * size.
 * @param dim
 * @param font
 */
public static void fudgeItalicLabel(Font font, Dimension dim)
{
	int style = font.getStyle();
	if ((style | Font.ITALIC) > 0) {
		int size = font.getSize();
		dim.width += Math.max(1, size / 6);
	}
}

}
