/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCStringTokenizer.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:38 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util;

/**
 * <code>JCStringTokenizer</code> controls simple linear tokenization
 * of a <code>String</code>. The set of delimiters, which defaults to common
 * whitespace characters, can be specified either during creation or on a
 * per-token basis.<p>
 * It is similar to <code>java.util.StringTokenizer</code>, but delimiters can
 * be included as literals by preceding them with a backslash character (the
 * default).<p>
 * It also fixes a known problem:
 * if one delimiter immediately follows another, a null String is returned
 * as the token instead of being skipped over.
 * <p>
 * Example usage:
 * <pre>
 *  String token, s = "this, is, a,, test";
 *  JCStringTokenizer st = new JCStringTokenizer(s);
 *  while (st.hasMoreTokens()) {
 *      token = st.nextToken(',');
 *      println(token);
 *  }
 * </pre>
 * This prints the following to the console:
 * <pre>
 *  this
 *  is
 *  a
 *
 *  test
 * </pre>
 */
public class JCStringTokenizer {

private int      index;               // last found position
private String   string;              // internal string
private int      length;              // string's length
private boolean  count = false;       // only count tokens?
public boolean   strip_esc = true;    // strip escape chars?
protected char   escape_char = '\\';  // escape char
private char[]   token;

/** Creates a tokenizer based on the specified String.
 * @param s
 */
public JCStringTokenizer(String s)
{
    if (s != null) {
        string = s.trim();
        length = string.length();
    }
}

public JCStringTokenizer(String s, boolean trim)
{
    if (s != null) {
        string = trim ? s.trim() : s;
        length = string.length();
    }
}

public JCStringTokenizer(String s, char delim)
{
    if (s != null) {
        if (Character.isWhitespace(delim)) {
            string = new String(s);
        }
        else {
            string = s.trim();
        }
        length = string.length();
    }
}

/** Parses the String using the specified delimiter.
 * @see #nextToken
 * @return
 * @param delim
 * @param s
 */
public static String[] parse(String s, char delim)
{
    return(parse(s, delim, '\\'));
}

/** Parses the String using the specified delimiter and escape char.
 * @see #setEscapeChar
 * @see #nextToken
 * @return
 * @param delim
 * @param escape_char
 * @param s
 */
public static String[] parse(String s, char delim, char escape_char)
{
    JCStringTokenizer t = new JCStringTokenizer(s);
    t.escape_char = escape_char;
    String[] list = new String[t.countTokens(delim)];
    for (int i = 0; i < list.length; i++) {
        list[i] = t.nextToken(delim);
        if (list[i] == null) {
            list[i] = "";
        }
    }
    return(list);
}

public static String[] parseWhitespaceDelim(String s, char delim,
                                            char escape_char)
{
    JCStringTokenizer t = new JCStringTokenizer(s, delim);
    t.escape_char = escape_char;
    String[] list = new String[t.countTokens(delim)];
    for (int i = 0; i < list.length; i++) {
        list[i] = t.nextToken(delim);
        if (list[i] == null) {
            list[i] = "";
        }
    }
    return(list);
}

/** Gets the escape char (default is \).
 * @return
 */
public char getEscapeChar()
{
    return(escape_char);
}

/** Sets the escape char (default is \). If 0, no escape char is used.
 * @param c
 */
public void setEscapeChar(char c)
{
    escape_char = c;
}

/**
 * Gets the next whitespace-delimited token.
 * @return
 */
public String nextToken()
{
    if (string == null || index >= length) {
        return null;
    }

    // Strip the leading whitespace.
    int i, k;
    for (i = index; i < length && Character.isWhitespace(string.charAt(i)); i++)
        ;
    index = i;
    if (i >= length) {
        return(null);
    }

    if (!count) {
        token = new char[length + 1];
    }

    for (i = index, k = 0; i < length; i++) {
        if (Character.isWhitespace(string.charAt(i))) {
            break;
        }
        else if (!count) {
            token[k++] = string.charAt(i);
        }
    }
    index = i + 1;
    return(!count ? new String(token).trim() : null);
}

/**
 * Gets the next token from delimited String.<p>
 * The delimeter can be "escaped" by a backslash character.<p>
 * To include a backslash character, precede it by another backslash
 * character.
 * @return
 * @param delim
 */
public String nextToken(char delim)
{
    if (string == null || index >= length) {
        return(null);
    }

    if (!count) {
        token = new char[length + 1];
    }

    int i, k;
    for (i = index, k = 0; i < length; i++, k++) {
        // Escaped char
        if (escape_char != 0 && i+1 < length &&
            string.charAt(i) == escape_char)
        {
            if (!strip_esc) {
                if (token != null) {
                    token[k++] = string.charAt(i);
                    token[k] = string.charAt(++i);
                }
                continue;
            }
            i++;
            if (!count) {
                if (string.charAt(i) == 'n') {
                    token[k] = '\n';
                }
                else {
                    token[k] = string.charAt(i);
                }
            }
        }
        // Delimiter found
        else if (string.charAt(i) == delim) {
            break;
        }
        // Otherwise copy the char
        else if (!count) {
            token[k] = string.charAt(i);
        }
    }

    // Strip token from string
    index = i + 1;
    return((!count && k > 0) ? new String(token, 0, k) : null);
}

/**
 * Returns the next number of tokens in the String using the
 * specified delimiter.
 * @return next number of tokens in the String using the specified delimiter
 * @param delim the delimiter
 */
public int countTokens(char delim)
{
    int num, old_index = index;
    count = true;
    for (num = 0; index < length; num++) {
        nextToken(delim);
    }
    index = old_index;
    count = false;
    return(num);
}

/**
 * Returns <code>true</code> if more tokens exist.
 * @return <code>true</code> if more tokens exist
 */
public boolean hasMoreTokens()
{
    return(index < length);
}

/**
 * Returns the current scan position within the String.
 * @return current scan position within the String
 */
public int getPosition()
{
    return(index);
}

}
