/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTypeConverter.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:38 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.awt.Point;

/**
 * <code>JCTypeConverter</code> contains static methods for retrieving
 * parameters from a source file or applet, and for converting parameters to
 * particular data types.
 */
public class JCTypeConverter {

public static boolean reportErrors = true;

/**
 * Writes a parse error message to stdout.
 * @param s the entire string being parsed
 * @param token the String which could not be parsed
 */
public static void error(String s, String token)
{
    error("Error parsing '" + token + "' in " + s);
}

/**
 * Writes a parse error message to stdout.
 * @param msg the String which could not be parsed
 */
public static void error(String msg)
{
    System.err.println(msg);
}

/**
 * Displays a parsing error.
 * @param token the token on which the error happened
 */
public static void parseError(String token)
{
    if (reportErrors) {
        error("Error parsing '" + token + "'");
    }
}

/**
 * Returns a trimmed String, with trailing nulls removed.
 * @param obj the object to trim
 * @return the trimmed String
 */
public static String trim(Object obj)
{
    String s;
    if (obj == null || (s = obj.toString()) == null) {
        return(null);
    }
    s = s.trim();
    int k = s.indexOf('\0');
    if (k != -1) {
        return(s.substring(0, k));
    }
    return(s);
}

/**
 * Converts a String to an integer.
 * @param s the String for conversion
 * @param def the default value to be used when a parsing error occurs
 * @return the integer resulting from the converson of the String
 */
public static int toInt(String s, int def)
{
    if (s == null) {
        return(def);
    }
    if (s.equalsIgnoreCase("max")) {
        return(Integer.MAX_VALUE);
    }
    if (s.equalsIgnoreCase("min")) {
        return(Integer.MIN_VALUE);
    }
    if (s.equalsIgnoreCase("maxint")) { // Backwards compatibility
        return(Integer.MAX_VALUE);
    }
    else if (s.equalsIgnoreCase("novalue")) {
        return(-999);
    }
    else if (s.equalsIgnoreCase("variable")) {
        return(-998);
    }
    else if (s.equalsIgnoreCase("default")) {
        return(-999);
    }

    try {
        return(Integer.parseInt(s));
    }
    // ATTN GV Violation
    catch (Exception e) {
        parseError(s);
    }
    return(def);
}

/**
 * Converts a String to a double.
 * @param s the String for conversion
 * @param def the default value to be used when a parsing error occurs
 * @return the double resulting from the converson of the String
 */
public static double toDouble(String s, double def)
{
    if (s == null) {
        return(def);
    }
    try {
        double d = def;
        if (s.equalsIgnoreCase("MAX")) {
            d = Double.MAX_VALUE;
        }
        else if (s.equalsIgnoreCase("MIN")) {
            d = Double.MIN_VALUE;
        }
        else if (s.equalsIgnoreCase("INFINITY")) {
            d = Double.POSITIVE_INFINITY;
        }
        else {
            d = Double.valueOf(s).doubleValue();
        }
        return(d);
    }
    // ATTN GV Violation
    catch(Exception e) {
        parseError(s);
    }
    return(def);
}

/**
 * Converts a String to a boolean.
 * @param s the String for conversion
 * @param def the default value to be used when a parsing error occurs
 * @return the boolean resulting from the converson of the String
 */
public static boolean toBoolean(String s, boolean def)
{
    if (s == null) {
        return(def);
    }
    try {
        return(Boolean.valueOf(s).booleanValue());
    }
    // ATTN GV Violation
    catch (Exception e) {
        parseError(s);
    }
    return(def);
}

/**
 * Converts a String to an array of Strings.
 * @param s the comma-separated Strings
 * @return the String list resulting from the converson of the String
 */
public static String[] toStringList(String s)
{
    return(toStringList(s, ','));
}

/**
 * Converts a String to an array of Strings.
 * Spaces are trimmed.
 * @param s the Strings separated by delimiter
 * @param delim the delimiter
 * @return the String list resulting from the converson of the String
 */
public static String[] toStringList(String s, char delim)
{
    return(toStringList(s, delim, true));
}


/**
 * Converts a String to an array of Strings.
 * @param s the Strings separated by delimiter
 * @param delim the delimiter
 * @param trim boolean to indicate if spaces are trimmed
 * @return the String list resulting from the converson of the String
 */
public static String[] toStringList(String s, char delim, boolean trim)
{
    if (s == null) {
        return(null);
    }
    JCStringTokenizer string = new JCStringTokenizer(s, false);
    String list[] = new String[string.countTokens(delim)];
    for (int i = 0; string.hasMoreTokens(); i++) {
        String next = string.nextToken(delim);
        if (next == null) {
            // next token is an empty string.
            next = "";
        }
        String trimmed = next.trim();
        if (trimmed == null) {
            trimmed = "";
        }
        list[i] = (trim) ? trimmed : next;
    }
    return(list);
}

/**
 * Converts a String to an array of Points.
 * @param s A String of x,y pairs
 * @param delim the delimiter used to separate the x,y pairs
 * @return the Point array resulting from the conversion of the String
 */
public static Point[] toPointList(String s, char delim)
{
    if (s == null) {
        return(null);
    }
    JCStringTokenizer strTok = new JCStringTokenizer(s);
    ArrayList list = new ArrayList();
    while (strTok.hasMoreTokens()) {
        String next = strTok.nextToken(delim);
        if (next != null) {
            JCStringTokenizer tok = new JCStringTokenizer(next);
            String sx = tok.nextToken(',');
            String sy = tok.nextToken(',');
            if (sx != null && sy != null) {
                try {
                    int x = Integer.parseInt(sx);
                    int y = Integer.parseInt(sy);
                    list.add(new Point(x, y));
                }
                catch(NumberFormatException e) {}
            }
        }
    }
    return((Point[])list.toArray(new Point[0]));
}

/**
 * Converts an array of Strings to a comma-separated String.
 * @param slist the array of Strings
 * @return the String resulting from the converson of the String list
 */
public static String fromStringList(String[] slist)
{
    return(fromStringList(slist, ','));
}

/**
 * Converts an array of Strings to a String in which values are separated
 * by the specified delimiter.
 * @param slist the array of Strings
 * @param delim the delimiter
 * @return the String resulting from the converson of the String list
 */
public static String fromStringList(String[] slist, char delim)
{
    if (slist == null) {
        return(null);
    }
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < slist.length; i++) {
         sb.append(slist[i]);
         if (i != slist.length - 1) {
            sb.append(delim);
         }
    }
    return(sb.toString());
}

/**
 * Converts a List of Strings to a comma-separated String.
 * @param slist the array of Strings
 * @return the String resulting from the converson of the String list
 */
public static String fromStringList(List slist)
{
    return(fromStringList(slist, ','));
}

/**
 * Converts a List of Strings to a String in which values are separated
 * by the specified delimiter.
 * @param slist the array of Strings
 * @param delim the delimiter
 * @return the String resulting from the converson of the List of Strings
 */
public static String fromStringList(List slist, char delim)
{
    if (slist == null) {
        return(null);
    }
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < slist.size(); i++) {
         sb.append(slist.get(i));
         if (i != slist.size() - 1) {
            sb.append(delim);
         }
    }
    return(sb.toString());
}

/**
 * Converts an array of Points to a String in which the x,y pairs are
 * separated by the specified delimiter.
 * @param parray the array of Points
 * @param delim the delimiter
 * @return the String resulting from the conversion of the Point array.
 */
public static String fromPointList(Point[] parray, char delim)
{
    if (parray == null) {
        return(null);
    }
    StringBuffer sb = new StringBuffer();
    for (int i = 0; i < parray.length; i++) {
        if (parray[i] != null) {
            sb.append(Integer.toString(parray[i].x));
            sb.append(',');
            sb.append(Integer.toString(parray[i].y));
            sb.append(delim);
        }
    }
    return(sb.toString());
}

/**
 * Converts a String to an array of integers based on the
 * provided delimiter.
 * @param s the String which will be converted to integers
 * @param delim the delimiter separating the integers in the String
 * @return the int list resulting from the converson of the String
 */
public static int[] toIntList(String s, char delim)
{
    if (s == null) {
        return(null);
    }
    JCStringTokenizer string = new JCStringTokenizer(s);
    int list[] = new int[string.countTokens(delim)];
    for (int i=0; string.hasMoreTokens(); i++) {
        String token = string.nextToken(delim).trim();
        list[i] = toInt(token, 0);
    }
    return(list);
}

/**
 * Converts a String to an array of Integer objects based on the
 * provided delimiter.
 * @param s the String which will be converted to Integers
 * @param delim the delimiter separating the Integers in the String
 * @return the Integer list resulting from the converson of the String
 */
public static Integer[] toIntegerList(String s, char delim)
{
    if (s == null) {
        return(null);
    }
    JCStringTokenizer string = new JCStringTokenizer(s);
    Integer list[] = new Integer[string.countTokens(delim)];
    for (int i=0; string.hasMoreTokens(); i++) {
        String token = string.nextToken(delim).trim();
        list[i] = new Integer(toInt(token, 0));
    }
    return list;
}

/**
 * Converts a String to an array of <code>Double</code> objects based on the
 * provided delimiter.
 * @param s the String which will be converted to <code>Doubles</code>
 * @param delim the delimiter separating the <code>Doubles</code> in the String
 * @return the <code>Double</code> list resulting from the converson of the
 * String
 */
public static Double[] toDoubleList(String s, char delim)
{
    if (s == null) {
        return(null);
    }
    JCStringTokenizer string = new JCStringTokenizer(s);
    Double list[] = new Double[string.countTokens(delim)];
    for (int i=0; string.hasMoreTokens(); i++) {
        String token = string.nextToken(delim).trim();
        list[i] = new Double(toDouble(token, 0));
    }
    return(list);
}

/**
 * Converts a String to an array of integers based on the
 * provided delimiter.
 * @param s the String which will be converted to integers
 * @param delim the delimiter separating the integers in the String
 * @param def the default value, returned if a parse error occurs
 * @return the int list resulting from the converson of the String
 */
public static int[] toIntList(String s, char delim, int[] def)
{
    int[] list = toIntList(s, delim);
    return(list != null ? list : def);
}

/**
 * Converts a String to an array of integers based on the
 * provided delimiter.
 * @param s the String which will be converted to integers
 * @param delim the delimiter separating the integers in the String
 * @param def the default value, returned if a parse error occurs
 * @return the Integer list resulting from the converson of the String
 */
public static Integer[] toIntegerList(String s, char delim, Integer[] def)
{
    Integer[] list = toIntegerList(s, delim);
    return(list != null ? list : def);
}

/**
 * Converts a String to an array of integers based on the
 * provided delimiter.
 * @param s the String which will be converted to integers
 * @param delim the delimiter separating the integers in the String
 * @param def the default value, returned if a parse error occurs
 * @return the <code>Double</code> list resulting from the converson of the
 * String
 */
public static Double[] toDoubleList(String s, char delim, Double[] def)
{
    Double[] list = toDoubleList(s, delim);
    return(list != null ? list : def);
}

/**
 * Converts all occurrences of "\n" in a String to newlines.
 * @param string the string to parse
 * @return the new String with the new lines replaced
 */
public static String toNewLine(String string)
{
    if (string == null || string.indexOf("\\n") == -1) {
        return(string);
    }
    char[] s = new char[string.length()];
    int k = 0;

    for (int i=0; i < s.length; i++, k++) {
        if (string.charAt(i) == '\\'
            && i < s.length-1
            && string.charAt(i+1) == 'n') {
            s[k] = '\n';
            i++;
        }
        else {
            s[k] = string.charAt(i);
        }
    }
    return(new String(s, 0, k));
}

/**
 * Converts all occurrences of char '\n' to String "\n".
 * @param s the String to parse
 * @return the new String with the char '\n' occurences replaced
 */
public static String fromNewLine(String s)
{
    if (s == null || s.indexOf('\n') == -1) {
        return(s);
    }
    int i, k, total = 0;
    for (i=0; i < s.length(); i++) {
        if (s.charAt(i) == '\n') total++;
    }
    char[] c = new char[s.length() + total];

    for (i=0, k=0; i < s.length(); i++, k++) {
        if (s.charAt(i) == '\n') {
            c[k++] = '\\';
            c[k] = 'n';
        }
        else {
            c[k] = s.charAt(i);
        }
    }
    return(new String(c, 0, k));
}

/** Converts a delimited list of tokens to a Vector.
 * @param s the String to convert
 * @param delim the delimiter that separates the tokens in the String
 * @return the Vector resulting from the conversion of the String
 */
public static Vector toVector(String s, char delim)
{
    if (s == null) {
        return(null);
    }
    JCStringTokenizer string = new JCStringTokenizer(s);
    Vector list = new Vector(string.countTokens(delim));

    string.strip_esc = false;

    while (string.hasMoreTokens()) {
        String token = string.nextToken(delim);
        list.addElement(token);
    }
    return(list);
}

/** Converts a delimited list of tokens to a Vector.
 * @param s the String to convert
 * @param delim the delimiter that separates the tokens in the String
 * @param def the default value, returned if a parse error occurs
 * @return the Vector resulting from the conversion of the String
 */
public static Vector toVector(String s, char delim, Vector def)
{
    Vector v = toVector(s, delim);
    return(v != null ? v : def);
}

/**
 * Converts a String using comma-separated value format.
 * @param s the String to be tokenized
 */
// ATTN GV Not yet - don't know what to do with JCCSVTokenizer
    /*
public static Vector toVectorFromCSV(String s)
{
    if (s == null) {
        return(null);
    }
    JCCSVTokenizer string = new JCCSVTokenizer(s);
    Vector list = new Vector(string.countTokens());

    for (int i=0; string.hasMoreTokens(); i++) {
        list.setElementAt(string.nextToken(), i);
    }
    return(list);
}
*/

/**
 * Converts a String using comma-separated value format.
 * @param s the String to be tokenized
 * @param def the default value, returned if a parse error occurs
 */
// ATTN GV Not yet - don't know what to do with JCCSVTokenizer
    /*
public static Vector toVectorFromCSV(String s, Vector def)
{
    Vector v = toVectorFromCSV(s);
    return(v != null ? v : def);
}
*/

/** Converts a String to an enum. If the String cannot be converted,
 * an error message is written to the console.
 * @param string the String to be converted
 * @param type the enum type
 * @param param the optional PARAM name for the enum, used for the error message
 * @param strings the first item in each list is the enum type;
 * each String has a corresponding value in the values array
 * @param values list of values corresponding to the Strings
 * @param def the default value, returned if the String could not be converted
 * @return either the converted value or the default
 */
public static int toEnum(String string, String type, String param,
                         String[][] strings, int[][] values, int def)
{
    if (param == null) {
        param = type;
    }
    int index;
    for (index = 0; index < strings.length; index++) {
        if (strings[index][0].equalsIgnoreCase(type)) {
            break;
        }
    }
    if (index == strings.length) {
        error("Error converting '"+string+"' to "+param);
        return(def);
    }

    string = string.trim();
    for (int i = 1; i < strings[index].length; i++) {
        if (string.equalsIgnoreCase(strings[index][i])) {
            return(values[index][i-1]);
        }
    }
    try {
        return(Integer.parseInt(string));
    }
    // ATTN GV Violation
    catch (Exception e) {
        error("Error converting '"+string+"' to "+param);
    }
    return(def);
}

/** Converts a String to an enum. If the String cannot be converted,
 * an error message is written to the console.
 * @param s the String to be converted
 * @param type the enum type (for the console message)
 * @param strings list of valid Strings
 * @param values list of values corresponding to each String
 * @param def the default value, returned if the String could not be converted
 * @return either the converted value or the default
 */
public static int toEnum(String s, String type, String[] strings,
                         int[] values, int def)
{
    if (s == null) {
        return(def);
    }
    s = s.trim();
    for (int i = 0; i < strings.length; i++) {
        if (s.equalsIgnoreCase(strings[i])) {
            return(values[i]);
        }
    }
    try {
        return(Integer.parseInt(s));
    }
    // ATTN GV Violation
    catch (Exception e) {
        error("Error converting '"+s+"' to "+type);
    }
    return(def);
}

/** Converts a String to an enum. If the String cannot be converted,
 * an error message is written to the console.
 * @param s the String to be converted
 * @param type the enum type (for the console message)
 * @param strings list of valid Strings
 * @param values list of values corresponding to each String
 * @param def the default value, returned if the String could not be converted
 * @return either the converted value or the default
 */
public static long toEnum(String s, String type, String[] strings,
                          long[] values, long def)
{
    if (s == null) {
        return(def);
    }
    s = s.trim();
    for (int i = 0; i < strings.length; i++) {
        if (s.equalsIgnoreCase(strings[i])) {
            return(values[i]);
        }
    }
    try {
        return(Long.parseLong(s));
    }
    // ATTN GV Violation
    catch (Exception e) {
        error("Error converting '"+s+"' to "+type);
    }
    return(def);
}

/** Converts a String to an enum.
 * @param s the String to be converted
 * @param strings list of valid Strings
 * @param values list of values corresponding to each String
 * @param def the default value, returned if the String could not be converted
 * @return the converted value or the default
 */
public static int toEnum(String s, String[] strings, int[] values, int def)
{
    if (s == null) {
        return(def);
    }
    s = s.trim();
    for (int i = 0; i < strings.length; i++) {
        if (s.equalsIgnoreCase(strings[i])) {
            return(values[i]);
        }
    }
    return(def);
}

/** Converts a String to a list of enums. If the String cannot be converted,
 * an error message is written to the console.
 * @param s the comma-separated list of enums
 * @param type the enum type
 * @param strings valid Strings
 * @param values list of values corresponding to each String
 * @param def the default value, returned if the String could not be converted
 * @return either the converted value or the default
 */
public static int[] toEnumList(String s, String type, String[] strings,
                               int[] values, int[] def)
{
    if (s == null) {
        return(def);
    }
    JCStringTokenizer string = new JCStringTokenizer(s);
    int list[] = new int[string.countTokens(',')];
    for (int i = 0; string.hasMoreTokens(); i++) {
        String token = string.nextToken(',').trim();
        int def_value = (def != null && i < def.length) ? def[i] : 0;
        list[i] = toEnum(token, type, strings, values, def_value);
    }
    return(list);
}

/** Converts an enum to a String.
 * @param value the enum to be converted
 * @param strings list of valid Strings
 * @param values list of values corresponding to each String
 * @return either the converted value or null
 */
public static String fromEnum(int value, String[] strings, int[] values)
{
    if (values == null || strings == null) {
        return(null);
    }
    for (int i = 0; i < values.length; i++) {
        if (value == values[i] && i < strings.length) {
            return strings[i];
        }
    }
    return(null);
}

/**
 * Converts an enum to a String where the enum index is a long.
 * @param value the enum to be converted
 * @param strings list of valid Strings
 * @param values list of values corresponding to each String
 * @return either the converted value or null
 */
public static String fromEnum(long value, String[] strings, long[] values)
{
    if (values == null || strings == null) {
        return(null);
    }
    for (int i = 0; i < values.length; i++) {
        if (value == values[i] && i < strings.length) {
            return(strings[i]);
        }
    }
    return(null);
}

/** Converts an enum to a String. If the enum cannot be converted,
 * an error message is written to the console.
 * @param value the enum to be converted
 * @param param the optional PARAM name for the enum, used for the error message
 * @param type the enum type
 * @param strings the first item in each list is the enum type;
 * each String has a corresponding value in the values array
 * @param values list of values corresponding to the Strings
 * @param def the default value, returned if the enum could not be converted
 * @return either the converted value or the default
 */
public static String fromEnum(int value, String param, String type,
                              String[][] strings, int[][] values, String def)
{
    int index;
    for (index = 0; index < strings.length; index++) {
        if (strings[index][0].equalsIgnoreCase(type)) {
            break;
        }
    }
    if (index == strings.length) {
        return(null);
    }

    for (int i = 0; i < values[index].length; i++) {
        if (values[index][i] == value) {
            return(strings[index][i+1]);
        }
    }
    try {
        return(String.valueOf(value));
    }
    // ATTN GV Violation
    catch (Exception e) {
        error("Error converting '" + value + "' to " + param);
    }
    return(def);
}

/** Finds an object within a match list of objects and returns the
 *  corresponding String from a synched-up String list.
 * @param value the object to be converted
 * @param strings list of Strings corresponding to
 *               each object in the values list
 * @param values list of objects being searched
 * @return either the converted value or null
 */
public static String fromMatchListObject(Object value, String[] strings,
                                         Object[] values)
{
    if (values == null || strings == null) {
        return(null);
    }
    for (int i = 0; i < values.length; i++) {
        if (value == values[i] && i < strings.length) {
            return(strings[i]);
        }
    }
    return(null);
}

/** Finds a String within a match list of Strings and returns the
 *  corresponding object from a synched-up object list.
 * @param value the String to be converted
 * @param strings list of Strings being searched
 * @param values list of objects corresponding to
 *               each String in the Strings list
 * @return either the converted value or null
 */
public static Object toMatchListObject(String value, String[] strings,
                                       Object[] values)
{
    if (values == null || strings == null) {
        return(null);
    }
    for (int i = 0; i < strings.length; i++) {
        if (value.equals(strings[i]) && i < values.length) {
            return(values[i]);
        }
    }
    return(null);
}

/** Checks the validity of an enum.
 *  @param value the enum value to check
 *  @param param enum name, used for error message
 *  @param values the array in which to check for the value
 *  @exception IllegalArgumentException If the value is invalid.
 */
public static void checkEnum(int value, String param, int[] values)
{
    for (int i = 0; i < values.length; i++) {
        if (values[i] == value) {
            return;
        }
    }
    throw new IllegalArgumentException("invalid " + param + ": " + value);
}

/** Converts an object to a String. If a String, newlines are replaced by "\n".
 *  If a Vector, it is converted to a comma-separated list.
 *  @param o the object to be converted to a String
 *  @return the String resulting from the conversion of the object
 */
public static String toString(Object o)
{
    if (!(o instanceof Vector)) {
        return(fromNewLine((o != null) ? o.toString() : ""));
    }

    Vector v = (Vector) o;
    StringBuffer s = new StringBuffer(v.size());
    for (int k = 0, size = v.size(); k < size; k++) {
        s.append(toString(v.elementAt(k)));
        if (k < size - 1) {
            s.append(',');
        }
    }
    return(new String(s));
}

/**
 * Converts a String to a Date.
 * @param s the String to convert
 * @param def the default value, returned if the String could not be converted
 * @return the Date resulting from the conversion of the String
 */
public static Date toDate(String s, Date def)
{
    if (s == null) {
        return(def);
    }

    try {
        SimpleDateFormat formatter =
            (SimpleDateFormat)DateFormat.getDateTimeInstance(DateFormat.LONG,
                                                             DateFormat.LONG);

        formatter.setCalendar(Calendar.getInstance());
        formatter.setTimeZone(TimeZone.getDefault());
        return(formatter.parse(s));
    }
    catch (ParseException pe) {
        return(def);
    }
}

/**
 * Converts a Date to a proper String that can be read back in parsed to a date.
 * @param d the date to convert
 * @param def the default String to return if the date could not be converted
 * @return the String resulting from the conversion of the date
 */
public static String fromDate(Date d, String def)
{
    if (d == null) {
        return(def);
    }

    try {
        SimpleDateFormat formatter =
            (SimpleDateFormat)DateFormat.getDateTimeInstance(DateFormat.LONG,
                                                             DateFormat.LONG);

        formatter.setCalendar(Calendar.getInstance());
        formatter.setTimeZone(TimeZone.getDefault());
        return(formatter.format(d));
    }
    catch (Exception e) {
        return(def);
    }
}

/**
 * Removes "escape" characters (backslashes) from the String.
 * @param s the String from which the "escape" characters are to be removed
 * @return the String which the "escape" characters removed
 */
public static String removeEscape(String s)
{
    if (s == null || s.indexOf('\\') == -1) {
        return(s);
    }
    int i, k, len = s.length();
    char[] token = new char[len];

    for (i = 0, k = 0; i < len; i++, k++) {
        char c = s.charAt(i);

        // Escaped char
        if (i + 1 < len && c == '\\') {
            i++;
            if (s.charAt(i) == 'n') {
                token[k] = '\n';
            }
            else {
                token[k] = s.charAt(i);
            }
        }
        // Otherwise copy the char
        else {
            token[k] = c;
        }
    }
    return((k > 0) ? new String(token, 0, k) : null);
}

}
