/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: SeqRangeCollector.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:39 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util;

import java.util.List;
import java.util.Vector;

/**
 * Collect added integer values into ranges defined by specified rules
 * (such as contiguous integers in a sequence, all integers of same value, etc).
 * Call addItem() to add items to ranges.  Call done() when done.
 * Call getRanges() to get list of ranges (after first calling done()).
 * Call reset() to begin anew.
 */
public class SeqRangeCollector
{

/** Range: contiguous integers, each member one greater than the previous */
public final static int RANGE_INCREASING_BY_1 = 0;
/** Range: all integers in a row that have same value.  Not yet implemented. */
public final static int RANGE_SAME_VALUE = 1;

/** Range Break: as specified by RangeType */
public final static int RANGE_BREAK_NORMAL = 0;
/** Range Break: as specified by RangeType and at 0xff boundaries. */
public final static int RANGE_BREAK_HEX_BOUNDARY = 1;

/** Type of range to look for */
protected int rangeType = RANGE_INCREASING_BY_1;
/** Extra rule for breaking a range */
protected int rangeBreak = RANGE_BREAK_NORMAL;

/** Value defining integer that is not part of the range.  Used as delimiter. */
protected int emptyValue = -1;

/** Start of current range.  If equal to empty value, no current range exists.*/
protected int startOfCurrentRange;
/** Last item in current range. */
protected int lastItemInCurrentRange;
/** Size of current range. */
protected int sizeOfCurrentRange;

/** List of ranges found so far. */
protected Vector ranges = null;

/** Simple constructor.  Will reset range list and other tracking values. */
public SeqRangeCollector()
{
	reset();
}

/**
 * Clear range list and reset range tracking variables for starting anew.
 */
public void reset()
{
	startOfCurrentRange = emptyValue;
	lastItemInCurrentRange = emptyValue;
	sizeOfCurrentRange = 0;
	ranges = new Vector();
}

/**
 * Add the given integer to a range as per the rules defined in the
 * RangeType property.  This item will either be added to the current
 * range, thereby extending it, or will force the creation of a new range.
 */
public void addItem(int item)
{
	// No current range -- create one
	if (startOfCurrentRange == emptyValue) {
		startOfCurrentRange = item;
		lastItemInCurrentRange = item;
		sizeOfCurrentRange++;
	}
    else if (rangeBreak == RANGE_BREAK_HEX_BOUNDARY &&
			 lastItemInCurrentRange % 0x100 == 0xff &&
			 item % 0x100 == 0)
    {
		flushRange(item);
	}
    // Increasing by one range -- if item is next in sequence, extend range
    else if (rangeType == RANGE_INCREASING_BY_1 &&
			 item == lastItemInCurrentRange + 1)
    {
		lastItemInCurrentRange = item;
		sizeOfCurrentRange++;
	}
    // Item can't be made part of current range, so end current range,
    // add it to list of ranges, and start new range with this item
    else {
		flushRange(item);
	}
}

/**
 * Finished adding items.  This must be called before calling getRanges(),
 * otherwise getRanges() will be missing final range.
 */
public void done()
{
	flushRange(emptyValue);
}

/**
 * End current range, add it to the list of ranges, and start new
 * range with given item.
 */
protected void flushRange(int item)
{
	if (startOfCurrentRange == emptyValue) {
		return;
    }

	Range range = new Range();
	if (rangeType == RANGE_INCREASING_BY_1) {
		range.start = startOfCurrentRange;
		range.end = lastItemInCurrentRange;
		range.size = sizeOfCurrentRange;
	}
	ranges.add(range);

	startOfCurrentRange = item;
	lastItemInCurrentRange = item;
	if (item == emptyValue) {
		sizeOfCurrentRange = 0;
	}
    else {
		sizeOfCurrentRange = 1;
	}
}

/**
 * Get list of ranges created from integer input.  Make sure to call done()
 * before calling this method, otherwise the returned list of ranges will
 * be missing the current (final) range.  Each item in the List is an
 * instance of the SeqRangeCollector.Range inner class.
 */
public List getRanges()
{
	return(ranges);
}

/**
 * Set the RangeType property, which determines the type of range that
 * is being tracked.  Currently, this value must be 
 * RANGE_INCREASING_BY_1.
 */
public void setRangeType(int rangeType)
{
	this.rangeType = rangeType;
}

/**
 * Get the RangeType property, which determines the type of range that
 * is being tracked.  Currently, this value must be 
 * RANGE_INCREASING_BY_1.
 */
public int getRangeType()
{
	return(rangeType);
}

/**
 * Set the RangeBreak property, which determines the type of range breaks
 * that will be forced in addition to the rule specified in RangeType;
 * Currently, this value must be one of RANGE_BREAK_NORMAL or
 * RANGE_BREAK_HEX_BOUNDARY.
 */
public void setRangeBreak(int rangeBreak)
{
	this.rangeBreak = rangeBreak;
}

/**
 * Set the RangeBreak property, which determines the type of range breaks
 * that will be forced in addition to the rule specified in RangeType;
 * Currently, this value must be one of RANGE_BREAK_NORMAL or
 * RANGE_BREAK_HEX_BOUNDARY.
 */
public int getRangeBreak()
{
	return(rangeBreak);
}


/**
 * Set the EmptyValue property, which is used to determine the integer
 * value used as an internal delimiter for ranges.  This value must never
 * be a part of the input added to ranges with the addItem() method.
 * By default, the value is -1, but users including negative numbers
 * in their ranges will want to change this something else, such as
 * Integer.MAX_VALUE.
 */
public void setEmptyValue(int emptyValue)
{
	this.emptyValue = emptyValue;
}

/**
 * Set the EmptyValue property, which is used to determine the integer
 * value used as an internal delimiter for ranges.  This value must never
 * be a part of the input added to ranges with the addItem() method.
 * By default, the value is -1.
 */
public int getEmptyValue()
{
	return(emptyValue);
}

/**
 * Class holding range.
 */
public class Range
{
	/** Start of range. */
	public int start;
	/** End of range. */
	public int end;
	/** Size of range. */
	public int size;

	public Range()
    {
		start = emptyValue;
		end   = emptyValue;
		size  = 0;
	}

	public Range(int start, int end, int size)
    {
		this.start = start;
		this.end   = end;
		this.size  = size;
	}

	public String toString()
    {
		return("Range: from " + Integer.toHexString(start) + " to " +
               Integer.toHexString(end) + " (" + size + " items)");
	}

    public boolean equals(Object o)
    {
        if (!(o instanceof Range)) {
            return(false);
        }
        Range r = (Range)o;
        return((start == r.start) && (end == r.end) && (size == r.size));
    }

    public int hashCode()
    {
        int result = 17;
        result = 37 * result + start;
        result = 37 * result + end;
        result = 37 * result + size;
        return(result);
    }
}

}
