/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BoundsValidator.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:39 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.calendar;

import com.klg.jclass.util.value.JCValueListener;
import com.klg.jclass.util.value.JCValueEvent;

import java.util.Calendar;
import java.awt.*;

/**
 * Class that given a minimum and a maximum date, checks that the date is
 * between the two bounds.
 */
public class BoundsValidator implements JCValueListener,
                                        java.io.Serializable
{

/** The Calendar object containing the minimum date. */
protected Calendar minimumDate;
/** The Calendar object containing the maximum date. */
protected Calendar maximumDate;
/** The parent component requiring the bounds check.  Can be null. */
protected Component parent;

/**
 * Default constructor.  All values are null.
 */
public BoundsValidator() {}

/**
 * Constructor specifying parent component.
 * @param c The parent component.
 */
public BoundsValidator(Component c)
{
    parent = c;
}

/**
 * Constructor specifying the minimum and maximum dates.
 * @param minDate The minimum date.
 * @param maxDate The maximum date.
 */
public BoundsValidator(Calendar minDate, Calendar maxDate)
{
    this(null, minDate, maxDate);
}

/**
 * Constructor specifying the parent component and the minimum and maximum
 * dates.
 * @param c The parent component.
 * @param minDate The minimum date.
 * @param maxDate The maximum date.
 */
public BoundsValidator(Component c, Calendar minDate, Calendar maxDate)
{
    parent = c;
    minimumDate = minDate;
    maximumDate = maxDate;
}

/**
 * Return the parent component.  This may be null.
 * @return The parent component.
 */
public Component getParent()
{
    return(parent);
}

/**
 * Set the parent component.
 * @param c The new parent component.
 */
public void setParent(Component c)
{
    parent = c;
}

protected boolean checkDates(Calendar min, Calendar max)
{
    boolean datesOK = true;
    if (min != null && max != null) {
        if (min.after(max)) {
            datesOK = false;
        }
    }
    return(datesOK);
}

/**
 * Returns the minimum roll date. The calendar cannot change before the date.
 * @return the minimum date
 * @see java.util.Calendar
 * @see #setMinimumDate
 * @see #getMaximumDate
 * @see #setMaximumDate
 */
public Calendar getMinimumDate()
{
    return(minimumDate);
}

/**
 * Sets the minimum date. The calendar cannot change to before this date.
 * @param minDate the new minimum date
 * @see java.util.Calendar
 * @see #getMinimumDate
 * @see #getMaximumDate
 * @see #setMaximumDate
 */
public void setMinimumDate(Calendar minDate)
{
    if (!checkDates(minDate, maximumDate)) {
        throw new IllegalArgumentException("Minimum date must be before " +
                                           "maximum date!");
    }
    minimumDate = minDate;
}

/**
 * Returns the maximum roll date. The calendar cannot change later than this
 * date.
 * @return the maximum date
 * @see java.util.Calendar
 * @see #setMinimumDate
 * @see #getMinimumDate
 * @see #setMaximumDate
 */
public Calendar getMaximumDate()
{
    return(maximumDate);
}

/**
 * Sets the maximum roll date. The calendar cannot change later than this date.
 * @param maxDate the new maximum date
 * @see java.util.Calendar
 * @see #setMinimumDate
 * @see #getMinimumDate
 * @see #getMaximumDate
 */
public void setMaximumDate(Calendar maxDate)
{
    if (!checkDates(minimumDate, maxDate)) {
        throw new IllegalArgumentException("Maximum date must be after " +
                                           "minimum date!");
    }
    maximumDate = maxDate;
}

/**
 * Given the minimum and the maximum date, check if the supplied date falls
 * between the dates.  If so, return true.
 * @param c The Calendar object containing the date to check
 * @return true if the date in the specified object is between the minimum and
 * maximum dates.
 */
public boolean isValid(Calendar c)
{
    if (c == null) {
        return(false);
    }

    // Check the bounds of the event
    if ((maximumDate != null && maximumDate.before(c)) ||
        (minimumDate != null && minimumDate.after(c)))
    {
        return(false);
    }
    return(true);
}

/**
 * Don't allow the change if it is before the minimum or after the maximum.
 * @param e the event generated by the value change
 */
public void valueChanging(JCValueEvent e)
{
    // Check the bounds of the event
    if (!isValid((Calendar)e.getNewValue())) {
        e.setAllowChange(false);
        if (parent != null) {
            Toolkit t = parent.getToolkit();
            if (t != null) {
                t.beep();
            }
        }
    }
}

/**
 * Nothing done here.
 * @param e the event generated by the value being changed
 */
public void valueChanged(JCValueEvent e) {}

}
