/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: CalendarContainer.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:39 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.calendar;

import com.klg.jclass.util.JCListenerList;
import com.klg.jclass.util.value.CalendarValueModel;
import com.klg.jclass.util.value.JCValueModel;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ContainerEvent;
import java.awt.event.ContainerListener;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Locale;

/**
 * Container to manage CalendarComponent children.  That is, any calls to
 * the calendar component interface methods are automatically passed to any
 * calendar component children
 */
public class CalendarContainer extends JPanel
			                   implements CalendarComponent, ContainerListener,
										  ActionListener
{

protected boolean         initialized = false;
protected JCValueModel    model;
protected JCCalendar      specialDates;
protected Locale          locale;
protected JCListenerList  actionListeners = null;
protected Font            userFont = null;

public CalendarContainer()
{
    initialized = true;
	setLocale(Locale.getDefault());
	addContainerListener(this);
}

/**
 * Set the current date being used by the calendar component
 * @param model The new calendar value model
 */
public void setCalendarModel(JCValueModel model)
{
	if (this.model == model) {
		return;
	}
	firePropertyChange("calendarmodel", this.model, model);

	this.model = model;
	Component kids[] = getComponents();
	for (int i = 0; i < kids.length; i++) {
		if (kids[i] instanceof CalendarComponent) {
			((CalendarComponent) kids[i]).setCalendarModel(model);
		}
	}
}

/**
 * Return the current value model which contains the date for this calendar
 * container.
 * @return The current calendar value model
 */
public JCValueModel getCalendarModel()
{
    return(model);
}

/**
 * Set the special dates being used by the calendar component
 * @param special_dates The new list of special dates stored within a
 *     JCCalendar object
 */
public void setSpecialDates(JCCalendar special_dates)
{
	if (this.specialDates == special_dates) {
		return;
	}

	firePropertyChange("specialdates", this.specialDates, special_dates);

	this.specialDates = special_dates;
	Component kids[] = getComponents();
	for (int i = 0; i < kids.length; i++) {
		if (kids[i] instanceof CalendarComponent) {
			((CalendarComponent) kids[i]).setSpecialDates(special_dates);
		}
	}
}

/**
 * Set the locale being used by the calendar component
 * @param locale The new locale to be used
 */
public void setLocale(Locale locale)
{
	if (this.locale == locale) {
		return;
	}
	if (locale == null) {
		throw new IllegalArgumentException("null is an invalid value " +
                                           "for locale");
	}

	firePropertyChange("locale", this.locale, locale);
	this.locale = locale;

	Component kids[] = getComponents();
	for (int i = 0; i < kids.length; i++) {
		if (kids[i] instanceof CalendarComponent) {
			kids[i].setLocale(locale);
		}
	}
}

/**
 * Return the locale used
 * @return The current locale used.
 */
public Locale getLocale()
{
	return(locale);
}

/**
 * Sets the <code>Font</code> of the container.  This font gets propagated to
 * the child components of type CalendarComponent
 *
 * @param f the new <code>Font</code> for the popup
 */
public void setFont(Font f)
{
    super.setFont(f);
    userFont = f;
    if (initialized) {
        Component kids[] = getComponents();
        for (int i = 0; i < kids.length; i++) {
            if (kids[i] instanceof CalendarComponent) {
                kids[i].setFont(f);
            }
        }
    }
}

/**
 * If a font has been set by the user, return it.
 * @return A user set font.  May be null, if no font was set by the user.
 */
public Font getUserFont()
{
    return(userFont);
}

/**
 * Add listener to receive action event when data is changed by the user.
 * @param l The action listener to be added
 */
public void addActionListener(ActionListener l)
{
	actionListeners = JCListenerList.add(actionListeners, l);
}

/**
 * Remove listener to receive action event when data is changed by the user.
 * @param l The action listener to be removed
 */
public void removeActionListener(ActionListener l)
{
	actionListeners = JCListenerList.remove(actionListeners, l);
}

//
// ContainerListener implementation
//

/**
 * Invoked when a component has been added to the container.
 * @param e The event generated when the component was added
 */
public void componentAdded(ContainerEvent e)
{
	if (e.getChild() instanceof CalendarComponent) {
		CalendarComponent cal = (CalendarComponent) e.getChild();
		cal.setLocale(locale);
		cal.addActionListener(this);

		if (model == null) {
			model = new CalendarValueModel(Calendar.getInstance(locale));
		}
		cal.setCalendarModel(model);
		cal.setSpecialDates(specialDates);
	}
}

/**
 * Invoked when a component has been removed from the container.
 * @param e The event generated when the component was removed
 */
public void componentRemoved(ContainerEvent e)
{
	if (e.getChild() instanceof CalendarComponent) {
		CalendarComponent cal = (CalendarComponent) e.getChild();
		cal.removeActionListener(this);
	}
}

//
// ActionListener implementation
//

/**
 * Method called when the action happens.
 * @param event The event generated by the action
 */
public void actionPerformed(ActionEvent event)
{
	// Repost action event
	ActionEvent actionEvent = new ActionEvent(this, event.getID(),
											  event.getActionCommand(),
											  event.getModifiers());
	Enumeration e = JCListenerList.elements(actionListeners);
	while (e.hasMoreElements()) {
		ActionListener listener = (ActionListener) e.nextElement();
		listener.actionPerformed(actionEvent);
	}
}

}
