/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCCalendar.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:39 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.calendar;

import java.text.DateFormatSymbols;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.util.Vector;

/**
 * For use in conjunction with a standard <i>java.util.Calendar</i>.
 * This component allows storage of special dates and then will
 * match a given <code>Calendar</code> object to those dates.
 */
public class JCCalendar {

protected Vector specialDates = new Vector(0);

static protected final byte daytable[][] = {
	{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
	{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};

/**
 * Checks to see if the year is a leap year.
 * For Gregorian Calendars, centennial years are leap
 * years only if they are multiples of 400. Otherwise,
 * any year which is a multiple of 4 is a leap year.
 * @param year the year to check
 * @return whether the given year is a leap year
 */
public static boolean isLeapYear(int year)
{
	boolean leap_year = false;

	// Check if centennial year
	if ((year % 100) == 0) {
		if ((year % 400) == 0) {
			leap_year = true;
		}
	}
	else if ((year % 4) == 0) {
		leap_year = true;
	}

	return(leap_year);
}

/**
 * Calculates the day of week given the date, month, and year.
 * (for example, Sunday, Monday, Tuesday, Wednesday, Thursday, Friday, or
 * Saturday)
 * @param date the day of the month
 * @param month the month (0 to 11)
 * @param year the year to use (for example, 1997)
 * @return the day of the week (0 = Sunday, ..., 6 = Saturday)
 */
public static int dayOfWeek(int date, int month, int year)
{
	int day;	// Week day (0 = Sunday, ..., 6 = Saturday)
	int M;		// M is month (1 = March,...,10 = December, 11 = Jan, 12 = Feb)
	int Y;		// Y is year (1997 has Y = 97 except Y = 96 for Jan & Feb)
	int C;		// C is century (1997 has C = 19)

	// Treat Jan and Feb as months of the preceding year
	if ((month == 0) || (month == 1)) {
		year = year - 1;
	}
	Y = year % 100;
	C = year / 100;

	// Convert the month so March is considered the first month
	M = ((month + 10) % 12) + 1;

	day = (int)(date + Math.floor(2.6 * M - 0.2) - 2 * C + Y +
					   Math.floor(Y/4) + Math.floor(C/4)) % 7;
	day = (day + 7) % 7;
	return(day);
}

/**
 * Calculates the number of days in a given month.
 * @param year the year to use (for example, 1997)
 * @param month the month (0 to 11)
 * @return the number of days in the given month and year
 */
public static int daysInMonth(int month, int year)
{
	int leap = isLeapYear(year) ? 1 : 0;
	return(daytable[leap][month]);
}

/**
 * Processes array for empty or null Strings and returns a new array with them
 * stripped out.
 * @param input the input array to be checked
 * @return the new output array with no empty or null Strings
 */
protected static String[] processArray(String[] input)
{
	if (input == null) {
		return(null);
	}

	// Count entries that are not null or empty strings.
	int count = 0;
	for (int i = 0; i < input.length; i++) {
		if (input[i] != null && input[i].length() > 0) {
			count++;
		}
	}

	// Store non-empty strings in a new array
	String[] output = new String[count];
	count = 0;
	for (int i = 0; i < input.length; i++) {
		if (input[i] != null && input[i].length() > 0) {
			output[count] = input[i];
			count++;
		}
	}
	return(output);
}

/**
 * Gets the list of short week day names for a given locale.
 * @param l the locale for which the day names are required
 * @return the list of short week day names (with nulls and empty Strings
 *         removed)
 */
public static String[] getDays(Locale l)
{
	if (l == null) {
		l = Locale.getDefault();
	}

	DateFormatSymbols symbols = new DateFormatSymbols(l);
	return(processArray(symbols.getShortWeekdays()));
}

/**
 * Gets the list of short or long month names for a given locale.
 * @param l the locale for which the month names are required
 * @param isShort if <code>true</code>, returns the short month names
 * @return the list of month names (with nulls and empty Strings removed)
 */
public static String[] getMonths(Locale l, boolean isShort)
{
	if (l == null) {
		l = Locale.getDefault();
	}

	DateFormatSymbols symbols = new DateFormatSymbols(l);
	String[] symbolMonths = null;
	if (isShort) {
	    symbolMonths = symbols.getShortMonths();
	}
	else {
	    symbolMonths = symbols.getMonths();
	}
	return(processArray(symbolMonths));
}

/**
 * Adds a special day given a date, month and year.
 * @param sd the <code>SpecialDate</code> to add
 * @see DateMonthYear
 * @see DayOfWeek
 * @see MonthDayOfMonth
 * @see MonthWeekDayOfWeek
 * @see #removeSpecialDate
 */
public void addSpecialDate(SpecialDate sd)
{
	specialDates.add(sd);
	// ??? post property changed
}

/**
 * Removes a special day given a date, month, and year.
 * @param sd the <code>SpecialDate</code> to remove
 * @see DateMonthYear
 * @see DayOfWeek
 * @see MonthDayOfMonth
 * @see MonthWeekDayOfWeek
 * @see #addSpecialDate
 */
public void removeSpecialDate(SpecialDate sd)
{
	specialDates.remove(sd);
	// ??? post property changed
}

/**
 * Determines if a specified date is a special date.
 * @param year the year to use (for example, 1997)
 * @param month the month (0 to 11)
 * @param date the day of the month
 * @param week the week of the month
 * @return <code>true</code> if the date specified has been designated as
 * special
 */
public boolean isSpecialDate(int year, int month, int date, int week)
{
	if (isInSpecialDateList(year, month, date, week) >= 0) {
		return(true);
	}
	return(false);
}

/**
 * Determines is a specified date is a special date.
 * @param c the <code>Calendar</code> object which contains the specified date
 * @return <code>true</code> if the date specified has been designated as
 * special
 */
public boolean isSpecialDate(Calendar c)
{
	return(isSpecialDate(c.get(c.YEAR), c.get(c.MONTH),
						 c.get(c.DAY_OF_MONTH), c.get(c.WEEK_OF_MONTH)));
}

/**
 * Returns the index in the <code>specialDays</code> vector if the object is
 * in the list. Otherwise, it returns -1.
 * @param year the year to use (for example, 1997)
 * @param month the month (0 to 11)
 * @param date the day of the month
 * @param week the week of the month
 * @return the index in the <code>specialDays</code> vector if the object is in
 * the list; otherwise, it returns -1
 */
protected int isInSpecialDateList(int year, int month, int date, int week)
{
	for (int i = 0; i < specialDates.size(); i++) {
		SpecialDate sd = (SpecialDate) specialDates.elementAt(i);
		if (sd.isSpecialDate(year, month, date, week)) {
			return (i);
		}
	}
	return(-1);
}

/**
 * Class to store a <code>DateMonthYear</code> value.<p>
 * For example:<br>  July 4th, 1776 --> new DateMonthYear(1776, 7, 4)
 */
public static class DateMonthYear implements SpecialDate
{
	public int year;
	public int month;
	public int date;

	/**
	 * Creates a <code>DateMonthYear</code> object with the specified date,
     * month, and year.
     * @param date the day of the month
	 * @param month 0 based: 0 = January ... 11 = December
     * @param year the year to use (for example, 1997)
	 */
	public DateMonthYear(int date, int month, int year)
	{
		this.year  = year;
		this.month = month;
		this.date  = date;
	}

	/**
	 * Creates a <code>DateMonthYear</code> object with a date, month, and year
     * stored in the given <code>Calendar</code> object.
	 * @param c the <code>Calendar</code> object which stores the specified date
	 */
	public DateMonthYear(Calendar c)
	{
		this(c.get(c.DATE), c.get(c.MONTH), c.get(c.YEAR));
	}

	/**
	 * Creates a <code>DateMonthYear</code> object with a date, month, and year
	 * stored in the given <code>Date</code> object.
	 * @param d the <code>Date</code> object which stores the specified date
	 */
	public DateMonthYear(Date d)
	{
		Calendar c = Calendar.getInstance();
		c.setTime(d);
		this.year  = c.get(c.YEAR);
		this.month = c.get(c.MONTH);
		this.date  = c.get(c.DATE);
	}

	/**
	 * Returns whether the given date is special.
     * @param year the year to use (for example, 1997)
     * @param month the month (0 to 11)
     * @param date the day of the month
     * @param week the week of the month
	 * @return Is the date specified special?
	 */
	public boolean isSpecialDate(int year, int month, int date, int week)
	{
		//System.out.println("isSpecialDate(" + year + ", " + month + ", " +
		//                   date + ", " + week + ")");
		//System.out.println("this values = " + this.year + ", " + this.month +
		//                   ", " + this.date + ")");

		if (this.year == year && this.month == month && this.date == date) {
			//System.out.println("returning true");
			return(true);
		}
		return(false);
	}

} // End of DateMonthYear

/**
 * Class to store a day of the week.<p>
 * For example:<br> Sunday --> new DayOfWeek(0)
 */
public static class DayOfWeek implements SpecialDate
{
	public int day_of_week;

	/**
	 * Creates a class with the specified day of week.
	 * @param day_of_week 0 based (0 - Sunday ... 6 - Saturday)
	 */
	public DayOfWeek(int day_of_week)
	{
		if (day_of_week < 0 || day_of_week > 6) {
			throw new IllegalArgumentException("invalid value for dayOfWeek");
		}
		this.day_of_week = day_of_week;
	}

	/**
	 * Returns whether the given date is special.
     * @param year the year to use (for example, 1997)
     * @param month the month (0 to 11)
     * @param date the day of the month
     * @param week the week of the month
	 * @return Is the date specified special?
	 */
	public boolean isSpecialDate(int year, int month, int date, int week)
	{
		int day = dayOfWeek(date, month, year);
		if (day_of_week == day) {
			return(true);
		}
		return(false);
	}

} // End of DayOfWeek

/**
 * Class to store Month and Day of Month. <p>
 * For example:<BR> November 11 --> new MonthDayOfMonth(11, 11)
 */
public static class MonthDayOfMonth implements SpecialDate
{
	public int month;
	public int day_of_month;

	/**
	 * Creates a class with the specified month and day of month.
	 * @param month 0 based: 0 = January ... 11 = December
	 * @param day_of_month the day within the month
	 */
	public MonthDayOfMonth(int month, int day_of_month)
	{
		this.month = month;
		this.day_of_month = day_of_month;
	}

	/**
	 * Returns whether the given date is special.
     * @param year the year to use (for example, 1997)
     * @param month the month (0 to 11)
     * @param date the day of the month
     * @param week the week of the month
	 * @return Is the date specified special?
	 */
	public boolean isSpecialDate(int year, int month, int date, int week)
	{
		if (this.month == month && this.day_of_month == date) {
			return(true);
		}
		return(false);
	}

} // End of MonthDayOfMonth

/**
 * Class to store a Month, Week, and Day of Week. <p>
 * For example:<BR> Labour Day (First Monday of September) -->
 *    new MonthWeekDayOfWeek(9, 1, 1)
 */
public class MonthWeekDayOfWeek implements SpecialDate {
	public int month;
	public int week;
	public int day_of_week;

	/**
	 * Creates a class with the specified month, week, and day of week.
	 * @param month 0 based: 0 - January ... 11 - December
	 * @param week 1 based
	 * @param day_of_week 0 based: 0 - Sunday ... 6 - Saturday
	 */
	public MonthWeekDayOfWeek(int month, int week, int day_of_week)
	{
		this.month = month;
		this.week = week;
		this.day_of_week = day_of_week;
	}

	/**
	 * Returns whether the given date is special.
     * @param year the year to use (for example, 1997)
     * @param month the month (0 to 11)
     * @param date the day of the month
     * @param week the week of the month
	 * @return Is the date specified special?
	 */
	public boolean isSpecialDate(int year, int month, int date, int week)
	{
		int day = dayOfWeek(date, month, year);
		if (this.month == month && this.day_of_week == day &&
			this.week == week) {
			return(true);
		}
		return(false);
	}

} // End of MonthWeekDayOfWeek


/**
 * Convenience method to clone a calendar.
 * @param cal the calendar object to clone
 * @return the new calendar clone
 */
static public Calendar copyCalendar(Calendar cal)
{
	if (cal == null) {
		return(null);
	}
	return((Calendar) cal.clone());
}

/**
 * Converts the specified type (if possible) to a supported type.  To reverse
 * this operation, use the <code>convertCalendarToSupported()</code> method.
 * The supported types are java.util.Date, java.util.Calendar,
 * java.util.GregorianCalendar, java.sql.Time, java.sql.Timestamp, and
 * java.sql.Date.
 * @see #convertCalendarToObject
 * @param obj The supported object to convert to Calendar.
 * @param l The locale to use
 * @return A Calendar object converted from the passed in supported object.  If
 * the object is not support, null is returned.
 */
static public Calendar convertObjectToCalendar(Object obj, Locale l)
{
    if (obj instanceof java.util.Calendar) {
        return((Calendar)obj);
    }
    else if (obj instanceof java.util.Date) {
        // Note since all SQL base time classes are subclasses of
        // java.util.Date we can use the same code to convert
        if (l == null) {
            l = Locale.getDefault();
        }
        Calendar cal = Calendar.getInstance(l);
        cal.setTime((java.util.Date)obj);
        return(cal);
    }
    return(null);
}

/**
 * Converts the Calendar object to an object designed by the class_type, if
 * its one of the supported class types. The supported types are java.util.Date,
 * java.util.Calendar, java.util.GregorianCalendar, java.sql.Time,
 * java.sql.Timestamp, and java.sql.Date.
 *
 * <br><B>Note:</B> The object passed in is not modified, but only used as a
 * reference.
 * @param from The Calendar object to convert.
 * @param classType The supported type to which the Calendar object is
 * converted.  If the type is unsupported, null will be returned.  The supported
 * types are java.util.Date, java.util.Calendar, java.util.GregorianCalendar,
 * java.sql.Time, java.sql.Timestamp, and java.sql.Date.
 * @return An object of the requested type converted from the passed in
 * Calendar object.
 * @see #convertObjectToCalendar
 */
static public Object convertCalendarToObject(Calendar from, Class classType)
{
    if (from == null || classType == null) {
        return (null);
    }

    String class_name = classType.getName();
    long milliseconds = from.getTime().getTime();

    if (class_name.equals("java.util.Date")) {
        return(new java.util.Date(milliseconds));
    }
    if (class_name.equals("java.sql.Time")) {
        return(new java.sql.Time(milliseconds));
    }
    if (class_name.equals("java.sql.Timestamp")) {
        return(new java.sql.Timestamp(milliseconds));
    }
    if (class_name.equals("java.sql.Date")) {
        return(new java.sql.Date(milliseconds));
    }
    if (class_name.equals("java.util.Calendar") ||
        class_name.equals("java.util.GregorianCalendar"))
    {
        return(from);
    }
    return(null);
}

}
