/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: TimeSpin.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:40 $ $Locker:  $ Quest Software Inc.

package com.klg.jclass.util.calendar;

import com.klg.jclass.util.calendar.CalendarComponent;
import com.klg.jclass.util.calendar.JCCalendar;
import com.klg.jclass.util.value.CalendarValueModel;
import com.klg.jclass.util.value.JCValueEvent;
import com.klg.jclass.util.value.JCValueListener;
import com.klg.jclass.util.value.JCValueModel;
import com.klg.jclass.util.JCEnvironment;

import javax.swing.*;
import javax.swing.text.DateFormatter;
import javax.swing.text.DefaultFormatterFactory;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import java.awt.event.ActionListener;
import java.awt.*;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import java.text.SimpleDateFormat;
import java.text.DateFormat;

/**
 * The TimeSpin class is used in the DateTimePopup to allow the user to
 * interactively modify the time portion of the current date.  In order to
 * get information passed on to it by the DateTimePopup, this class implements
 * the CalendarComponent interface.  Note that there are two models, the
 * normal SpinnerModel and the JCValueModel that need to be kept in sync.  One
 * holds a Date object and the other holds a Calendar object.
 */
public class TimeSpin extends JSpinner implements CalendarComponent,
                                                  JCValueListener,
                                                  ChangeListener
{

private static int defaultColumns = 10;

/** The calendar value model. */
protected JCValueModel valueModel;

/** A validator that checks for values between a min and max date. */
protected BoundsValidator boundsValidator = new BoundsValidator(this);

/** Ignore change in spinner if already changing spinner. */
protected boolean ignoreChange = false;

/**
 * No args constructor.  Use the current time.
 */
public TimeSpin()
{
    this(new Date());
}

/**
 * Constructor which supplies a starting time (in a Date object).
 * @param date The starting time
 */
public TimeSpin(Date date)
{
    this(date, Locale.getDefault());
}

/**
 * Constructor which supplies a starting time (in a Date object) and a locale.
 * @param date The starting time
 * @param l The locale to use
 */
public TimeSpin(Date date, Locale l)
{
    if (JCEnvironment.getJavaVersion() < 140) {
        throw new IllegalStateException("TimeSpin is incompatible " +
                                        "with JDKs prior to 1.4.0");
    }
    if (date == null) {
        throw new IllegalArgumentException("Null date in TimeSpin constructor");
    }
    if (l == null) {
        l = Locale.getDefault();
    }
    setLocale(l);
    valueModel = new CalendarValueModel();
    valueModel.addValueListener(this);
    SpinnerDateModel model = new SpinnerDateModel();
    model.setValue(date);
    setModel(model);
}

/**
 * Sets the current date being used by the calendar component.  Note that the
 * Calendar component part (which allows it to be part of the DateTimePopup)
 * uses a Calendar value for its model where the spinner uses a SpinnerDateModel
 * (based on a Date).  These models are kept in sync.
 * @param valueModel The new calendar value model.  This must be of type
 * CalendarValueModel.
 */
public void setCalendarModel(JCValueModel valueModel)
{
    if (!(valueModel instanceof CalendarValueModel)) {
        throw new IllegalArgumentException("Must use a CalendarValueModel");
    }
    if (this.valueModel == valueModel) {
        return;
    }

    if (this.valueModel != null) {
        this.valueModel.removeValueListener(this);
    }
    this.valueModel = valueModel;
    this.valueModel.addValueListener(this);

    Calendar c = (Calendar)valueModel.getValue();
    Date d = (Date)JCCalendar.convertCalendarToObject(c, Date.class);
    getModel().setValue(d);
}

/**
 * Overrides JSpinner.setModel(model) so that it can keep both of its models
 * in sync.  The Calendar component part (which allows it to be part of the
 * DateTimePopup) uses a Calendar value for its model where the spinner uses a
 * SpinnerDateModel (based on a Date).
 * @param model The new model. This must be a SpinnerDateModel.
 */
public void setModel(SpinnerModel model)
{
    if (model == null) {
        throw new IllegalArgumentException("Can't set null model");
    }
    if (!(model instanceof SpinnerDateModel)) {
        throw new IllegalArgumentException("Must set SpinnerDateModel");
    }
    SpinnerModel oldModel = getModel();
    if (oldModel == model) {
        return;
    }

    if (oldModel != null) {
        oldModel.removeChangeListener(this);
    }
    super.setModel(model);
    model.addChangeListener(this);

    if (valueModel != null) {
        Date date = ((SpinnerDateModel)model).getDate();
        Calendar cal = Calendar.getInstance(getLocale());
        cal.setTime(date);
        valueModel.setValue(cal);
    }
}

/**
 * Returns the minimum roll date. The calendar cannot change before the date.
 * @return the minimum date
 * @see java.util.Calendar
 * @see #setMinimumDate
 * @see #getMaximumDate
 * @see #setMaximumDate
 */
public Calendar getMinimumDate()
{
    return(boundsValidator.getMinimumDate());
}

/**
 * Sets the minimum date. The calendar cannot change to before this date.
 * @param minDate the new minimum date
 * @see java.util.Calendar
 * @see #getMinimumDate
 * @see #getMaximumDate
 * @see #setMaximumDate
 */
public void setMinimumDate(Calendar minDate)
{
    boundsValidator.setMinimumDate(minDate);
}

/**
 * Returns the maximum roll date. The calendar cannot change later than this
 * date.
 * @return the maximum date
 * @see java.util.Calendar
 * @see #setMinimumDate
 * @see #getMinimumDate
 * @see #setMaximumDate
 */
public Calendar getMaximumDate()
{
    return(boundsValidator.getMaximumDate());
}

/**
 * Sets the maximum roll date. The calendar cannot change later than this date.
 * @param maxDate the new maximum date
 * @see java.util.Calendar
 * @see #setMinimumDate
 * @see #getMinimumDate
 * @see #getMaximumDate
 */
public void setMaximumDate(Calendar maxDate)
{
    boundsValidator.setMaximumDate(maxDate);
}

/**
 * This is unused.
 * Sets the special dates being used by the calendar component.
 * @param special_dates The new list of special dates stored within a
 *     JCCalendar object
 */
public void setSpecialDates(JCCalendar special_dates) {}

/**
 * This is unused.
 * Adds an action listener to detect specific actions on this component.
 * @param l The action listener to be added
 */
public void addActionListener(ActionListener l) {}

/**
 * This is unused.
 * Removes an action listener on this component.
 * @param l The action listener to be removed
 */
public void removeActionListener(ActionListener l) {}

/**
 * The value is changing.  Check if the date falls in the min/max bounds.
 * @param event The event generated by the value changing
 */
public void valueChanging(JCValueEvent event)
{
    // Check the bounds of the event
    if (!boundsValidator.isValid((Calendar)event.getNewValue())) {
        event.setAllowChange(false);
        Toolkit t = getToolkit();
        if (t != null) {
            t.beep();
        }
    }
}

/**
 * The calendar value model has changed.  Update the spinner model.
 * @param event The event generated by the value changed
 */
public void valueChanged(JCValueEvent event)
{
    Calendar c = (Calendar)event.getNewValue();
    SpinnerDateModel model = (SpinnerDateModel)getModel();
    Date date = (Date)model.getValue();
    Date newDate = c.getTime();
    if (!date.equals(newDate)) {
        // Update the spinner model to the new date
        model.setValue(newDate);
    }
}

/**
 * The spinner model has changed.  Update the calendar value model.
 * @param e A ChangeEvent object
 */
public void stateChanged(ChangeEvent e)
{
    if (ignoreChange) {
        return;
    }
    Calendar cValueModel = (Calendar)valueModel.getValue();
    SpinnerDateModel model = (SpinnerDateModel)getModel();
    Date dateSpinnerModel = (Date)model.getValue();
    Date dateValueModel = cValueModel.getTime();
    if (!dateSpinnerModel.equals(dateValueModel)) {
        // Update the calendar value model to the new date
        Calendar cSpinnerModel = Calendar.getInstance(getLocale());
        cSpinnerModel.setTime(dateSpinnerModel);
        Calendar cNew = (Calendar)cValueModel.clone();
        cNew.set(Calendar.HOUR_OF_DAY, cSpinnerModel.get(Calendar.HOUR_OF_DAY));
        cNew.set(Calendar.MINUTE, cSpinnerModel.get(Calendar.MINUTE));
        cNew.set(Calendar.SECOND, cSpinnerModel.get(Calendar.SECOND));
        cNew.set(Calendar.AM_PM, cSpinnerModel.get(Calendar.AM_PM));
        // If the value is valid, set it on the calendar model.  Otherwise,
        // reset the spinner to the old value.
        if (boundsValidator.isValid(cNew)) {
            valueModel.setValue(cNew);
        }
        else {
            ignoreChange = true;
            model.setValue(dateValueModel);
            ignoreChange = false;
        }
    }
}

/**
 * Set the locale the TimeSpin object should use.  Also update the formatter to
 * be based on the new locale.
 * @param l The locale object to use.
 */
public void setLocale(Locale l)
{
    super.setLocale(l);

    // Rest the format to the medium time format for the new locale
    DateFormat df = DateFormat.getTimeInstance(DateFormat.MEDIUM, getLocale());
    setFormat(df);
    if (df instanceof SimpleDateFormat) {
        setColumns((SimpleDateFormat)df);
    }
    else {
        setColumns(defaultColumns);
    }
}

/**
 * Set the format to use in the JCFormattedTextField for this TimeSpin object.
 * @param df The new format
 */
public void setFormat(DateFormat df)
{
    if (df == null) {
        df = new SimpleDateFormat();
    }

    // Get the editor and set the format on the editor
    JSpinner.DefaultEditor de = (JSpinner.DefaultEditor)getEditor();
    if (de != null) {
        JFormattedTextField ftf = de.getTextField();
        DateFormatter formatter = new DateFormatter(df);
        DefaultFormatterFactory factory =
            new DefaultFormatterFactory(formatter);
        ftf.setFormatterFactory(factory);
    }
}

/**
 * Set the columns in the editor to be the pattern size of the passed in
 * SimpleDateFormat.
 * @param sdf The format from which to get the pattern size
 */
public void setColumns(SimpleDateFormat sdf)
{
    if (sdf == null) {
        return;
    }

    // Get the pattern
    String pattern = sdf.toPattern();
    setColumns((pattern == null ? defaultColumns : pattern.length()));
}

/**
 * Set the number of columns in the editor to a new value.
 * @param columns The new number of columns (must be non-negative)
 */
public void setColumns(int columns)
{
    if (columns < 0) {
        throw new IllegalArgumentException("Number of columns must " +
                                           "be non-negative");
    }
    JSpinner.DefaultEditor de = (JSpinner.DefaultEditor)getEditor();
    if (de != null) {
        JFormattedTextField ftf = de.getTextField();
        ftf.setColumns(columns);
    }
}

/**
 * Sets the font for this component.  Also passes the font on to internal
 * formatted text field.
 *
 * @param f The desired <code>Font</code> for this component
 * @see java.awt.Component#getFont
 */
public void setFont(Font f)
{
    super.setFont(f);

    // Set the font on internal formmatted text field
    JComponent curEditor = getEditor();
    if (curEditor instanceof JSpinner.DefaultEditor) {
        JSpinner.DefaultEditor de = (JSpinner.DefaultEditor)curEditor;
        JFormattedTextField ftf = de.getTextField();
        ftf.setFont(f);
    }
}

}
