/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Add.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:40 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;





/**
 * This is a concrete implementation for the addition "+" operation.
 * Addition is normally the sum of a pair of elements (unary identity),
 * *???but it can also be between two lists???*.
 */
public class Add extends Operation {

/**
 * Create an operation to add two Expressions
 */
public Add(Expression left, Expression right) {
	super(left, right);
}

/**
 * Create an operation to add an Expression and a Number
 */
public Add(Expression left, Number right) {
	super(left, right);
}

/**
 * Create an operation to add a Number and an Expression
 */
public Add(Number left, Expression right) {
	super(left, right);
}

/**
 * Create an operation to add two Numbers
 */
public Add(Number left, Number right) {
	super(left, right);
}

/**
 * Return the result of adding the operands
 * @return The result of the first operand plus the second
 */
public Result evaluate() {

	// Can't do anything with no data
	if (leftOperand == null && rightOperand == null) {
		throw new IllegalArgumentException("Addition is not defined on zero operands");
	}

	// Catch the unary case (no-op)
	if (leftOperand == null) {
		return(rightOperand.evaluate());
	}
	if (rightOperand == null) {
		return(leftOperand.evaluate());
	}

	return(getResult(leftOperand, rightOperand));
}

private Result getResult(Expression left, Expression right) {

	// Evaluate any nested operations
	if (left instanceof Operation || left instanceof ExpressionReference) {
		left = left.evaluate();
	}
	if (right instanceof Operation || right instanceof ExpressionReference) {
		right = right.evaluate();
	}

	// Compute the sum of two elements
	if (left instanceof MathValue && right instanceof MathValue) {

		MathValue mvLeft = (MathValue) left;
		MathValue mvRight = (MathValue) right;

		return(doAddition(mvLeft, mvRight));
	}

	// Compute the sum of elements of lists
	else if (left instanceof ExpressionList &&
			 right instanceof ExpressionList) {

		ExpressionList xlLeft = (ExpressionList) left;
		ExpressionList xlRight = (ExpressionList) right;

		// Check that we have the same number of elements in each list
		if (xlLeft.size() != xlRight.size()) {
			throw new OperandMismatchException("Cannot add lists of different sizes");
		}

		// get iterators over the left and right lists
		Iterator leftIter = xlLeft.iterator();
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Add all elements of the list pairwise and create a result list
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(leftElem, rightElem));
		}

		// return the result list
		return(result);

	}

	// Compute an offset of a set of numbers
	else if (left instanceof MathValue && right instanceof ExpressionList) {

		ExpressionList xlRight = (ExpressionList) right;

		// get an iterator over the list
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Add all elements of the list to the offset
		while (rightIter.hasNext()) {
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(left, rightElem));
		}

		// return the result list
		return(result);
	}

	// Compute an offset of a set of numbers
	else if (left instanceof ExpressionList && right instanceof MathValue) {

		ExpressionList xlLeft = (ExpressionList) left;

		// get an iterator over the list
		Iterator leftIter = xlLeft.iterator();
		ExpressionList result = new MathExpressionList();

		// Add the offset to all elements of the list
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();

			result.add(getResult(leftElem, right));
		}

		// return the result list
		return(result);
	}

	else {
		throw new OperandMismatchException("Cannot add given data types");
	}
}

/**
 * Add one value to another
 */
private MathValue doAddition(MathValue left, MathValue right) {

	// Add two scalar quantities
	if (left instanceof MathScalar && right instanceof MathScalar) {

		MathScalar sLeft = (MathScalar) left;
		MathScalar sRight = (MathScalar) right;

		Number result = addSimple(sLeft.realValue, sRight.realValue);
		return new MathScalar(result);
	}

	// Add a pair of vectors
	else if (left instanceof MathVector && right instanceof MathVector) {

		MathVector vLeft = (MathVector) left;
		MathVector vRight = (MathVector) right;

		// Ensure that the vectors have the same size
		if (vLeft.size != vRight.size) {
			throw new OperandMismatchException("Cannot add different-sized vectors");
		}

		// Compute the result for each entry of the vector
		Number[] result = new Number[vLeft.size];
		for (int i = 0; i < vLeft.size; i++) {
			result[i] = addSimple(vLeft.realValues[i], vRight.realValues[i]);
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Add a matrix to another matrix
	else if (left instanceof MathMatrix && right instanceof MathMatrix) {

		MathMatrix mLeft = (MathMatrix) left;
		MathMatrix mRight = (MathMatrix) right;

		// Ensure that the matrices have the same size
		if (mLeft.rows != mRight.rows || mLeft.columns != mRight.columns) {
			throw new OperandMismatchException("Cannot add different-sized matrices");
		}

		// Compute the result for each entry of the matrix
		Number[][] result = new Number[mLeft.rows][mLeft.columns];
		for (int i = 0; i < mLeft.rows; i++) {
			for (int j = 0; j < mLeft.columns; j++) {
				result[i][j] = addSimple(mLeft.realValues[i][j],
										 mRight.realValues[i][j]);
			}
		}

		// create and return a resultant matrix
		return new MathMatrix(result);
	}

	// otherwise the types don't match (e.g. adding a vector to a matrix)
	else {
		throw new OperandMismatchException("Cannot add differently-dimensioned data types");
	}
}

/**
 * Compute the sum of two Numbers
 */
private Number addSimple(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(left.doubleValue() + right.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(left.intValue() + right.intValue());
	}
}

}
