/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Divide.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:40 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;





/**
 * This is a concrete implementation for the division "/" operation.
 * Division is normally the ratio between a pair of elements, but we
 * also permit a unary form (returns 1 / x) *???and division of two lists???*.
 */
public class Divide extends Operation {

/**
 * Create an operation to divide two Expressions
 */
public Divide(Expression left, Expression right) {
	super(left, right);
}

/**
 * Create an operation to divide an Expression by a Number
 */
public Divide(Expression left, Number right) {
	super(left, right);
}

/**
 * Create an operation to divide a Number by an Expression
 */
public Divide(Number left, Expression right) {
	super(left, right);
}

/**
 * Create an operation to divide two Numbers
 */
public Divide(Number left, Number right) {
	super(left, right);
}

/**
 * Return the result of dividing the operands
 * @return The result of the first operand divided by the second
 */
public Result evaluate() {

	// Can't do anything with no data
	if (leftOperand == null && rightOperand == null) {
		throw new IllegalArgumentException("Division is not defined on zero operands");
	}

	// Catch the unary case (compute 1/x for all elements)
	if (leftOperand == null) {
		return(getResult(new MathScalar(new Integer(1)), rightOperand));
	}
	if (rightOperand == null) {
		return(getResult(new MathScalar(new Integer(1)), leftOperand));
	}

	return(getResult(leftOperand, rightOperand));
}

private Result getResult(Expression left, Expression right) {

	// Evaluate any nested operations
	if (left instanceof Operation || left instanceof ExpressionReference) {
		left = left.evaluate();
	}
	if (right instanceof Operation || right instanceof ExpressionReference) {
		right = right.evaluate();
	}

	// Compute the ratio between two elements
	if (left instanceof MathValue && right instanceof MathValue) {

		MathValue mvLeft = (MathValue) left;
		MathValue mvRight = (MathValue) right;

		return(doDivision(mvLeft, mvRight));
	}

	// Compute the ratio between elements of lists
	else if (left instanceof ExpressionList &&
			 right instanceof ExpressionList) {

		ExpressionList xlLeft = (ExpressionList) left;
		ExpressionList xlRight = (ExpressionList) right;

		// Check that we have the same number of elements in each list
		if (xlLeft.size() != xlRight.size()) {
			throw new OperandMismatchException("Cannot divide lists of different sizes");
		}

		// get iterators over the left and right lists
		Iterator leftIter = xlLeft.iterator();
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Divide all elements of the list pairwise and create a result list
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(leftElem, rightElem));
		}

		// return the result list
		return(result);

	}

	// Divide a set of numbers into a constant
	else if (left instanceof MathValue && right instanceof ExpressionList) {

		ExpressionList xlRight = (ExpressionList) right;

		// get an iterator over the list
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Divide a constant by the elements of a list
		while (rightIter.hasNext()) {
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(left, rightElem));
		}

		// return the result list
		return(result);
	}

	// Scale a set of numbers
	else if (left instanceof ExpressionList && right instanceof MathValue) {

		ExpressionList xlLeft = (ExpressionList) left;

		// get an iterator over the list
		Iterator leftIter = xlLeft.iterator();
		ExpressionList result = new MathExpressionList();

		// Divide all elements of the list by the scale factor
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();

			result.add(getResult(leftElem, right));
		}

		// return the result list
		return(result);
	}

	else {
		throw new OperandMismatchException("Cannot divide given data types");
	}
}

/**
 * Divide one value by another
 */
private MathValue doDivision(MathValue left, MathValue right) {

	// Divide two scalar quantities
	if (left instanceof MathScalar && right instanceof MathScalar) {

		MathScalar sLeft = (MathScalar) left;
		MathScalar sRight = (MathScalar) right;

		Number result = divideSimple(sLeft.realValue, sRight.realValue);
		return new MathScalar(result);
	}

	// Divide a vector by a scalar (scale the vector)
	else if (left instanceof MathVector && right instanceof MathScalar) {

		MathVector vLeft = (MathVector) left;
		MathScalar sRight = (MathScalar) right;

		// Compute the result for each entry of the vector
		Number[] result = new Number[vLeft.size];
		for (int i = 0; i < vLeft.size; i++) {
			result[i] = divideSimple(vLeft.realValues[i], sRight.realValue);
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Divide a matrix by a scalar (scale the matrix)
	else if (left instanceof MathMatrix && right instanceof MathScalar) {

		MathMatrix mLeft = (MathMatrix) left;
		MathScalar sRight = (MathScalar) right;

		// Compute the result for each entry of the matrix
		Number[][] result = new Number[mLeft.rows][mLeft.columns];
		for (int i = 0; i < mLeft.rows; i++) {
			for (int j = 0; j < mLeft.columns; j++) {
				result[i][j] = divideSimple(mLeft.realValues[i][j],
											sRight.realValue);
			}
		}

		// create and return a resultant matrix
		return new MathMatrix(result);
	}

	// otherwise the operation isn't defined (dividing by vector or matrix)
	else {
		throw new OperandMismatchException("Cannot divide by a vector or a matrix");
	}
}

/**
 * Compute the ratio of two Numbers
 */
private Number divideSimple(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(left.doubleValue() / right.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(left.intValue() / right.intValue());
	}
}

}
