/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: ExpressionVariable.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:40 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.HashMap;




/**
 * A variable is a named expression which can have its value changed
 * by name-keyed access to it.
 */
public class ExpressionVariable implements Expression, Cloneable {

/** The name by which the variable is known and referred to */
protected String		name;
/** The context in which the variable is defined */
protected Object		context;
/** The value which the variable is currently assigned */
protected Expression	value;

/** The variable map stores all named variables (disambiguated by context) */
protected static HashMap variableMap = new HashMap();

/**
 * Create a variable Expression given a context, name and value
 * @param context The object in which context the name is unique
 * @param name The name of the variable
 * @param value The current value of the variable
 */
private ExpressionVariable(Object context, String name, Expression value) {

	// Copy the values to the object
	this.context = context;
	this.name = new String(name);
	// ??? Should we clone the value?
	this.value = value;
}

/**
 * Create a variable Expression given a context and name
 * @param context The object in which context the name is unique
 * @param name The name of the variable
 */
public static ExpressionVariable createVariable(Object context, String name) {

	// Create a temporary expression to check for existence
	ExpressionVariable testVariable = new ExpressionVariable(context, name,
															 null);

	// Return the variable if it exists already
	if (variableMap.containsKey(testVariable)) {
		return((ExpressionVariable) variableMap.get(testVariable));
	}

	// otherwise add this variable to the map and return it
	else {
		variableMap.put(testVariable, testVariable);
		return(testVariable);
	}
}

/**
 * Create or set a variable Expression given a context, name and value
 * @param context The object in which context the name is unique
 * @param name The name of the variable
 * @param value The current value of the variable
 */
public static ExpressionVariable setVariable(Object context, String name,
											 Expression value) {

	// Create a temporary expression to check for existence
	ExpressionVariable testVariable = new ExpressionVariable(context, name,
															 null);

	// Check if this variable exists already
	if (variableMap.containsKey(testVariable)) {

		// The variable has been created already, so update its value
		ExpressionVariable ev = (ExpressionVariable) variableMap.get(testVariable);

		// ??? Should we clone the value?
		ev.value = value;

		// Return the instance of this variable
		return(ev);
	}
	else {

		// The variable doesn't exist, so the temporary variable gets used
		// ??? Should we clone the value?
		testVariable.value = value;

		// We store it in the lookup table, keyed by itself.
		variableMap.put(testVariable, testVariable);

		return(testVariable);
	}
}

/**
 * Retrieve a variable Expression given a context and name
 * @param context The object in which context the name is unique
 * @param name The name of the variable
 */
public static ExpressionVariable getVariable(Object context, String name) {

	// Create a temporary expression to check for existence
	ExpressionVariable testVariable = new ExpressionVariable(context, name,
															 null);

	// Return the variable if it exists already
	if (variableMap.containsKey(testVariable)) {
		return((ExpressionVariable) variableMap.get(testVariable));
	}

	// otherwise we create the variable so people don't have to always
	// make sure they call "create()" first
	else {
		variableMap.put(testVariable, testVariable);
		return(testVariable);
	}
}

/**
 * As an Expression, the evaluation of this variable is its stored value
 * (or the evaluation of a stored Expression).
 * @return The Result of the stored Expression
 */
public Result evaluate() {

	// if there is no value stored here, throw an exception
	if (value == null) {
		// ??? Is this the exception we should throw?
		throw new ArithmeticException("Undefined variable \"" + name +
									  "\" in context " + context);
	}

	// otherwise evaluate the stored value and return it.
	return(value.evaluate());
}

/**
 * Set the current value of a variable
 * @param value The new value the variable will have
 */
public void setValue(Expression value) {
	// ??? Should we clone the value?
	this.value = value;
}

/**
 * Retrieve the Expression which is stored by this Variable
 * @return The unevaluated Expression this variable is equal to
 */
public Expression getValue() {
	return(value);
}

/**
 * Compute the hashcode for this variable based on the context and name
 * @return A hash of the context and name values of this variable
 */
public int hashCode() {
	return(context.hashCode() * 31 + name.hashCode());
}

/**
 * Determine if this variable matches another
 * @param obj The object to test for equality
 * @return True if the variables have the same name and context
 */
public boolean equals(Object obj) {

	// Doesn't match null or a non-key object
	if (obj == null || !(obj instanceof ExpressionVariable)) {
		return(false);
	}

	// Equality only if context and name both match
	ExpressionVariable target = (ExpressionVariable) obj;
	if (target.context == this.context && this.name.equals(target.name)) {
		return(true);
	}

	// otherwise the keys don't match
	return(false);
}

/**
 * Clone this variable; since variables are unique to a context, we
 * return the variable itself.
 */
public Object clone() {
	// Return the variable itself
	return((Object) this);
}

}
