/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Max.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:40 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;




/**
 * This is a concrete implementation for the max operation.
 * Max can be determined for a pair of elements or across a list.
 */
public class Max extends Operation {

/**
 * Maximum across the given operand (usually a list of expressions)
 */
public Max(Expression operand) {
	super(operand);
}

/**
 * Max of a pair of Expressions
 */
public Max(Expression left, Expression right) {
	super(left, right);
}

/**
 * Max of an Expression and a Number
 */
public Max(Expression left, Number right) {
	super(left, right);
}

/**
 * Max of a Number and an Expression
 */
public Max(Number left, Expression right) {
	super(left, right);
}

/**
 * Max of two Numbers
 */
public Max(Number left, Number right) {
	super(left, right);
}

/**
 * Return the result of taking the max across all the operands
 * @return The maximum value of any of the operands
 */
public Result evaluate() {

	if (leftOperand == null && rightOperand == null) {
		throw new IllegalArgumentException("Max is not defined on zero operands");
	}

	MathValue leftOperandResult = null;
	MathValue rightOperandResult = null;

	if (leftOperand != null) {
		leftOperandResult = evaluateOperand(leftOperand);
	}
	if (rightOperand != null) {
		rightOperandResult = evaluateOperand(rightOperand);
	}

	// Don't create default (0) values 'cause we don't want to
	// muck with returned format or type
	if (leftOperand == null) {
		return(rightOperandResult);
	}
	if (rightOperand == null) {
		return(leftOperandResult);
	}

	// Take the min of the results from the left and right operand
	return(doMax(leftOperandResult, rightOperandResult));
}

/**
 * Extract a value from the given operand
 */
private MathValue evaluateOperand(Expression operand) {

	if (operand instanceof MathValue) {
		return((MathValue) operand);
	}

	else if (operand instanceof Operation ||
			 operand instanceof ExpressionReference) {
		return(evaluateOperand(operand.evaluate()));
	}

	else if (operand instanceof ExpressionList) {

		MathValue result = null, tmp_result;

		// Process the elements of the list
		Iterator eli = ((ExpressionList) operand).iterator();
		while (eli.hasNext()) {

			// Get the value of each element
			tmp_result = evaluateOperand((Expression) eli.next());

			// Accumulate the result
			if (result == null) {
				result = tmp_result;
			}
			else {
				result = doMax(result, tmp_result);
			}
		}
		return(result);
	}

	else {
		throw new IllegalArgumentException("Unrecognized operand type to Max");
	}
}

/**
 * Determine the maximum of two values
 */
private MathValue doMax(MathValue left, MathValue right) {

	// We can't apply the concept of "max" to a vector or matrix
	// (It may help to consider the magnitude of a vector, but we
	// need to define a metric for that first.)

	if (left instanceof MathVector || left instanceof MathMatrix ||
		right instanceof MathVector || right instanceof MathMatrix) {

			throw new IllegalArgumentException("Cannot determine max of a vector or matrix.");
	}

	// determine the max of two scalar quantities
	MathScalar sLeft = (MathScalar) left;
	MathScalar sRight = (MathScalar) right;

	Number result = maxSimple(sLeft.realValue, sRight.realValue);
	return new MathScalar(result);
}

/**
 * Find the max of two Numbers
 */
private Number maxSimple(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(Math.max(left.doubleValue(), right.doubleValue()));
	}
	// otherwise use integers
	else {
		return new Integer(Math.max(left.intValue(), right.intValue()));
	}
}

}
