/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Multiply.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;




/**
 * This is a concrete implementation for the multiplication "*" operation.
 * Multiplication is normally the product of a pair of elements,
 * *???but we also permit the multiplication of two lists???*.
 */
public class Multiply extends Operation {

/**
 * Create a multiply operation
 */
public Multiply() {
	super();
}

/**
 * Create an operation to multiply a given pair of Expressions
 */
public Multiply(Expression ex1, Expression ex2) {
	super(ex1, ex2);
}

/**
 * Create an operation to multiply an Expression and a Number
 */
public Multiply(Expression left, Number right) {
	super(left, right);
}

/**
 * Create an operation to multiply a Number and an Expression
 */
public Multiply(Number left, Expression right) {
	super(left, right);
}

/**
 * Create an operation to multiply two Numbers
 */
public Multiply(Number left, Number right) {
	super(left, right);
}

/**
 * Return the result of multiplying the operands
 * @return The result of the first operand multiplied by the second
 */
public Result evaluate() {

	// Can't do anything with no data
	if (leftOperand == null && rightOperand == null) {
		throw new IllegalArgumentException("Multiplication is not defined on zero operands");
	}

	// Catch the unary case (no-op)
	if (leftOperand == null) {
		return(rightOperand.evaluate());
	}
	if (rightOperand == null) {
		return(leftOperand.evaluate());
	}

	return(getResult(leftOperand, rightOperand));
}

private Result getResult(Expression left, Expression right) {

	// Evaluate any nested operations
	if (left instanceof Operation || left instanceof ExpressionReference) {
		left = left.evaluate();
	}
	if (right instanceof Operation || right instanceof ExpressionReference) {
		right = right.evaluate();
	}

	// Compute the product of two elements
	if (left instanceof MathValue && right instanceof MathValue) {

		MathValue mvLeft = (MathValue) left;
		MathValue mvRight = (MathValue) right;

		return(doMultiply(mvLeft, mvRight));
	}

	// Compute the product (pairwise) of elements of lists
	else if (left instanceof ExpressionList &&
			 right instanceof ExpressionList) {

		ExpressionList xlLeft = (ExpressionList) left;
		ExpressionList xlRight = (ExpressionList) right;

		// Check that we have the same number of elements in each list
		if (xlLeft.size() != xlRight.size()) {
			throw new OperandMismatchException("Cannot multiply lists of different sizes");
		}

		// get iterators over the left and right lists
		Iterator leftIter = xlLeft.iterator();
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Multiply all elements of the list pairwise and create a result list
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(leftElem, rightElem));
		}

		// return the result list
		return(result);

	}

	// Multiply a set of numbers into a constant
	else if (left instanceof MathValue && right instanceof ExpressionList) {

		ExpressionList xlRight = (ExpressionList) right;

		// get an iterator over the list
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Multiply a constant by the elements of a list
		while (rightIter.hasNext()) {
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(left, rightElem));
		}

		// return the result list
		return(result);
	}

	// Scale a set of numbers
	else if (left instanceof ExpressionList && right instanceof MathValue) {

		ExpressionList xlLeft = (ExpressionList) left;

		// get an iterator over the list
		Iterator leftIter = xlLeft.iterator();
		ExpressionList result = new MathExpressionList();

		// Multiply all elements of the list by the scale factor
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();

			result.add(getResult(leftElem, right));
		}

		// return the result list
		return(result);
	}

	else {
		throw new OperandMismatchException("Cannot multiply given data types");
	}
}

/**
 * Multiply one value by another
 */
private MathValue doMultiply(MathValue left, MathValue right) {

	// Multiply two scalar quantities
	if (left instanceof MathScalar && right instanceof MathScalar) {

		MathScalar sLeft = (MathScalar) left;
		MathScalar sRight = (MathScalar) right;

		Number result = multiplySimple(sLeft.realValue, sRight.realValue);
		return new MathScalar(result);
	}

	// Multiply a vector by a scalar (scale the vector)
	else if ((left instanceof MathVector && right instanceof MathScalar) ||
			 (left instanceof MathScalar && right instanceof MathVector)) {

		MathVector vector;
		MathScalar scalar;

		if (left instanceof MathVector) {
			scalar = (MathScalar) left;
			vector = (MathVector) right;
		}
		else {
			scalar = (MathScalar) right;
			vector = (MathVector) left;
		}

		// Compute the result for each entry of the vector
		Number[] result = new Number[vector.size];
		for (int i = 0; i < vector.size; i++) {
			result[i] = multiplySimple(vector.realValues[i], scalar.realValue);
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Multiply a matrix by a scalar (scale the matrix)
	else if (left instanceof MathMatrix && right instanceof MathScalar) {

		MathMatrix matrix;
		MathScalar scalar;

		if (left instanceof MathVector) {
			scalar = (MathScalar) left;
			matrix = (MathMatrix) right;
		}
		else {
			scalar = (MathScalar) right;
			matrix = (MathMatrix) left;
		}

		// Compute the result for each entry of the matrix
		Number[][] result = new Number[matrix.rows][matrix.columns];
		for (int i = 0; i < matrix.rows; i++) {
			for (int j = 0; j < matrix.columns; j++) {
				result[i][j] = multiplySimple(matrix.realValues[i][j],
											  scalar.realValue);
			}
		}

		// create and return a resultant matrix
		return new MathMatrix(result);
	}

	// Compute the dot product of a pair of vectors
	else if (left instanceof MathVector && right instanceof MathVector) {

		MathVector vLeft = (MathVector) left;
		MathVector vRight = (MathVector) right;

		// Ensure that the vectors have the same size
		if (vLeft.size != vRight.size) {
			throw new OperandMismatchException("Cannot compute the dot product of different-sized vectors");
		}

		// Sum the products of elements of the vectors
		Number result = new Integer(0);
		for (int i = 0; i < vLeft.size; i++) {
			result = sum(result, multiplySimple(vLeft.realValues[i],
											    vRight.realValues[i]));
		}

		// create and return a resultant vector
		return new MathScalar(result);
	}

	// Compute the product of a vector and a matrix
	else if (left instanceof MathVector && right instanceof MathMatrix) {

		MathVector vector = (MathVector) left;
		MathMatrix matrix = (MathMatrix) right;

		// Ensure that the data sizes match
		if (vector.size != matrix.rows) {
			throw new OperandMismatchException("Vector and matrix sizes do not match");
		}

		// Sum the product of the vector against each column of the matrix
		Number result[] = new Number[matrix.columns];
		for (int i = 0; i < matrix.columns; i++) {

			result[i] = new Integer(0);
			for (int j = 0; j < vector.size; j++) {
				result[i] = sum(result[i],
								multiplySimple(vector.realValues[j],
											   matrix.realValues[j][i]));
			}
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Compute the product of a matrix and a vector
	else if (left instanceof MathMatrix && right instanceof MathVector) {

		MathMatrix matrix = (MathMatrix) left;
		MathVector vector = (MathVector) right;

		// Ensure that the data sizes match
		if (vector.size != matrix.columns) {
			throw new OperandMismatchException("Vector and matrix sizes do not match");
		}

		// Sum the product of the vector against each row of the matrix
		Number result[] = new Number[matrix.rows];
		for (int i = 0; i < matrix.rows; i++) {

			result[i] = new Integer(0);
			for (int j = 0; j < vector.size; j++) {
				result[i] = sum(result[i],
								multiplySimple(matrix.realValues[i][j],
											   vector.realValues[j]));
												  

			}
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Multiply a matrix onto another matrix
	else if (left instanceof MathMatrix && right instanceof MathMatrix) {

		MathMatrix mLeft = (MathMatrix) left;
		MathMatrix mRight = (MathMatrix) right;

		// Ensure that the matrices have complementary size
		if (mLeft.columns != mRight.rows) {
			throw new OperandMismatchException("Cannot multiply matrices of non-complementary sizes");
		}

		// Compute the result for each entry of the matrix
		Number[][] result = new Number[mLeft.rows][mRight.columns];
		for (int i = 0; i < mLeft.rows; i++) {
			for (int j = 0; j < mRight.columns; j++) {
				result[i][j] = multiplySimple(mLeft.realValues[i][j],
											  mRight.realValues[i][j]);

				// Sum the product of a row and a column
				result[i][j] = new Integer(0);
				for (int k = 0; k < mLeft.columns; k++) {
					result[i][j] = sum(result[i][j],
									multiplySimple(mLeft.realValues[i][k],
												   mRight.realValues[k][j]));
				}
			}
		}

		// create and return a resultant matrix
		return new MathMatrix(result);
	}

	// otherwise the operation isn't defined
	else {
		throw new OperandMismatchException("Cannot multiply the given data types");
	}
}

/**
 * Add two Numbers
 */
private Number sum(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(left.doubleValue() + right.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(left.intValue() + right.intValue());
	}
}

/**
 * Compute the product of two Numbers
 */
private Number multiplySimple(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(left.doubleValue() * right.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(left.intValue() * right.intValue());
	}
}

}
