/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Operation.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;

/**
 * The Operation class defines the basic elements of an operator.
 * All operators have (at least potentially) a left and right operand;
 * this allows the correct ordering to be applied to matrix operations
 * and any other non-commutative operators.
 * NOTE:  The issue of cycle detection still needs to be addressed
 */
public abstract class Operation implements Expression, Cloneable {

/** The operand on the left side of the operator for binary operations;
	the only operand for unary operations */
public Expression leftOperand = null;
/** The operand on the right side of the operator for binary operations */
public Expression rightOperand = null;

/**
 * Constructor for a generic operator with no operands
 */
public Operation() {
}

/**
 * Constructor for a unary or List operator
 * @param operand An Expression to be transformed by a unary or
 *                aggregation operation
 */
public Operation(Expression operand) {
		leftOperand = operand;
}

/**
 * Constructor for a unary operator taking a Number
 * @param num A Number on which to perform a unary operation
 */
public Operation(Number num) {
		this(new MathScalar(num));
}

/**
 * Constructor for a binary operator
 * @param left The Expression on the left-hand side of the operation
 * @param right The Expression on the right-hand side of the operation
 */
public Operation(Expression left, Expression right) {
	leftOperand = left;
	rightOperand = right;
}

/**
 * Constructor for a binary operator given an Expression and a Number
 * @param left The Expression on the left-hand side of the operation
 * @param right A Number for the right-hand side of the operation
 */
public Operation(Expression left, Number right) {
	this(left, new MathScalar(right));
}

/**
 * Constructor for a binary operator given a Number and an Expression 
 * @param left A Number for the left-hand side of the operation
 * @param right The Expression on the right-hand side of the operation
 */
public Operation(Number left, Expression right) {
	this(new MathScalar(left), right);
}

/**
 * Constructor for a binary operator given a pair of Numbers
 * @param left A Number for the left-hand side of the operation
 * @param right A Number for the right-hand side of the operation
 */
public Operation(Number left, Number right) {
	this(new MathScalar(left), new MathScalar(right));
}

/**
 * Evaluate the Operation, returning the result.
 * @return The result of evaluating this operation on the current operands
 */
public abstract Result evaluate();

/**
 * Return an Expression which is a clone of this Operation
 * (all operands are also cloned)
 */
public Object clone() {

	Operation newOperation;

	try {
		newOperation = (Operation) super.clone();
	}
	catch (CloneNotSupportedException e) {
		return(null);
	}

	if (leftOperand != null) {
		try {
			newOperation.leftOperand = (Expression) leftOperand.clone();
		}
		catch (CloneNotSupportedException e) {
			newOperation.leftOperand = null;
		}
	}
	if (rightOperand != null) {
		try {
			newOperation.rightOperand = (Expression) rightOperand.clone();
		}
		catch (CloneNotSupportedException e) {
			newOperation.leftOperand = null;
		}
	}

	return(newOperation);
}

}
