/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Power.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;




/**
 * This is a concrete implementation for the exponentiation "^" operation.
 * Exponentiation is normally a binary operation,
 * *???but we also permit the use of one list as the exponents of another???*.
 */
public class Power extends Operation {

/**
 * Create a exponentiation operation
 */
public Power() {
	super();
}

/**
 * Create an operation to raise a given of operand to the power of another
 */
public Power(Expression base, Expression exp) {
	super(base, exp);
}

/**
 * Create an operation to raise an Expression to the power of a Number
 */
public Power(Expression base, Number exp) {
	super(base, exp);
}

/**
 * Create an operation to raise a Number to the power of an Expression
 */
public Power(Number base, Expression exp) {
	super(base, exp);
}

/**
 * Create an operation to raise a Number to the power of another
 */
public Power(Number base, Number exp) {
	super(base, exp);
}

/**
 * Return the result of raising the first operand to the power of the second
 * @return The result of first^second
 */
public Result evaluate() {

	// Can't do anything with no or only one operand
	if (leftOperand == null || rightOperand == null) {
		throw new IllegalArgumentException("Power is not defined on fewer than two operands");
	}

	return(getResult(leftOperand, rightOperand));
}

private Result getResult(Expression left, Expression right) {

	// Evaluate any nested operations
	if (left instanceof Operation || left instanceof ExpressionReference) {
		left = left.evaluate();
	}
	if (right instanceof Operation || right instanceof ExpressionReference) {
		right = right.evaluate();
	}

	// Compute the exponentiation of two elements
	if (left instanceof MathValue && right instanceof MathValue) {

		MathValue mvLeft = (MathValue) left;
		MathValue mvRight = (MathValue) right;

		return(doPower(mvLeft, mvRight));
	}

	// Compute the exponentiation (pairwise) of elements of lists
	else if (left instanceof ExpressionList &&
			 right instanceof ExpressionList) {

		ExpressionList xlLeft = (ExpressionList) left;
		ExpressionList xlRight = (ExpressionList) right;

		// Check that we have the same number of elements in each list
		if (xlLeft.size() != xlRight.size()) {
			throw new OperandMismatchException("Power cannot use lists of different sizes");
		}

		// get iterators over the left and right lists
		Iterator leftIter = xlLeft.iterator();
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Match all elements of the list pairwise and generate the results
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(leftElem, rightElem));
		}

		// return the result list
		return(result);

	}

	// Use a set of numbers as exponents for a constant
	else if (left instanceof MathValue && right instanceof ExpressionList) {

		ExpressionList xlRight = (ExpressionList) right;

		// get an iterator over the list
		Iterator rightIter = xlRight.iterator();
		ExpressionList result = new MathExpressionList();

		// Raise a constant by the elements of a list
		while (rightIter.hasNext()) {
			Expression rightElem = (Expression) rightIter.next();

			result.add(getResult(left, rightElem));
		}

		// return the result list
		return(result);
	}

	// Use a constant as a power for all elements of a list
	else if (left instanceof ExpressionList && right instanceof MathValue) {

		ExpressionList xlLeft = (ExpressionList) left;

		// get an iterator over the list
		Iterator leftIter = xlLeft.iterator();
		ExpressionList result = new MathExpressionList();

		// Raise all elements of the list by the given exponent
		while (leftIter.hasNext()) {
			Expression leftElem = (Expression) leftIter.next();

			result.add(getResult(leftElem, right));
		}

		// return the result list
		return(result);
	}

	else {
		throw new OperandMismatchException("Cannot perform exponentiation with the given data types");
	}
}

/**
 * Raise one value by another
 */
private MathValue doPower(MathValue left, MathValue right) {

	// Raise one scalar quantity by another
	if (left instanceof MathScalar && right instanceof MathScalar) {

		MathScalar sLeft = (MathScalar) left;
		MathScalar sRight = (MathScalar) right;

		Number result = powerSimple(sLeft.realValue, sRight.realValue);
		return new MathScalar(result);
	}

	// otherwise the operation isn't defined
	else {
		throw new OperandMismatchException("Cannot use vectors or matrices to compute exponents.");
	}
}

/**
 * Compute the value of one Number to the exponent of another
 */
private Number powerSimple(Number base, Number exponent) {

	// Compute the result
	double result = Math.pow(base.doubleValue(), exponent.doubleValue());

	// Promote to doubles
	if (base instanceof Double || exponent instanceof Double) {
		return new Double(result);
	}
	// otherwise use integers
	else {
		return new Integer((int) result);
	}
}

}
