/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Product.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;




/**
 * This is a concrete implementation for the product "*" operation.
 * A product can be performed on a pair of elements or across a list.
 */
public class Product extends Operation {

/**
 * Product of the given operand (usually a list of expressions)
 */
public Product(Expression operand) {
	super(operand);
}

/**
 * Product of a pair of Expressions
 */
public Product(Expression left, Expression right) {
	super(left, right);
}

/**
 * Product of an Expression and a Number
 */
public Product(Expression left, Number right) {
	super(left, right);
}

/**
 * Product of a Number and an Expression
 */
public Product(Number left, Expression right) {
	super(left, right);
}

/**
 * Product of two Numbers
 */
public Product(Number left, Number right) {
	super(left, right);
}

/**
 * Return the result of multiplying the operands
 * @return The result of this product on the operands
 */
public Result evaluate() {

	if (leftOperand == null && rightOperand == null) {
		throw new IllegalArgumentException("Product is not defined on zero operands");
	}

	MathValue leftOperandResult = null;
	MathValue rightOperandResult = null;

	if (leftOperand != null) {
		leftOperandResult = evaluateOperand(leftOperand);
	}
	if (rightOperand != null) {
		rightOperandResult = evaluateOperand(rightOperand);
	}

	// Don't create default (0) values 'cause we don't want to
	// muck with returned format or type
	if (leftOperand == null) {
		return(rightOperandResult);
	}
	if (rightOperand == null) {
		return(leftOperandResult);
	}

	// multiply the results from the left and right operand
	return(doProduct(leftOperandResult, rightOperandResult));
}

/**
 * Extract a value from the given operand
 */
private MathValue evaluateOperand(Expression operand) {

	if (operand instanceof MathValue) {
		return((MathValue) operand);
	}

	else if (operand instanceof Operation ||
			 operand instanceof ExpressionReference) {
		return(evaluateOperand(operand.evaluate()));
	}

	else if (operand instanceof ExpressionList) {

		MathValue result = null, tmp_result;

		// Process the elements of the list
		Iterator eli = ((ExpressionList) operand).iterator();
		while (eli.hasNext()) {

			// Get the value of each element
			tmp_result = evaluateOperand((Expression) eli.next());

			// Accumulate the result
			if (result == null) {
				result = tmp_result;
			}
			else {
				result = doProduct(result, tmp_result);
			}
		}
		return(result);
	}

	else {
		throw new IllegalArgumentException("Unrecognized operand type to Product");
	}
}

/**
 * Multiply two values
 */
private MathValue doProduct(MathValue left, MathValue right) {

	// Multiply two scalar quantities
	if (left instanceof MathScalar && right instanceof MathScalar) {

		MathScalar sLeft = (MathScalar) left;
		MathScalar sRight = (MathScalar) right;

		Number result = productSimple(sLeft.realValue, sRight.realValue);
		return new MathScalar(result);
	}

	// Scale a vector
	else if ((left instanceof MathScalar && right instanceof MathVector) ||
			 (left instanceof MathVector && right instanceof MathScalar)) {

		MathScalar scalar;
		MathVector vector;

		if (left instanceof MathScalar) {
			scalar = (MathScalar) left;
			vector = (MathVector) right;
		}
		else {
			vector = (MathVector) left;
			scalar = (MathScalar) right;
		}

		// Scale the elements of the vector
		Number[] result = new Number[vector.size];
		for (int i = 0; i < vector.size; i++) {
			result[i] = productSimple(scalar.realValue, vector.realValues[i]);
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Scale a matrix
	else if ((left instanceof MathScalar && right instanceof MathMatrix) ||
			 (left instanceof MathMatrix && right instanceof MathScalar)) {

		MathScalar scalar;
		MathMatrix matrix;

		if (left instanceof MathScalar) {
			scalar = (MathScalar) left;
			matrix = (MathMatrix) right;
		}
		else {
			matrix = (MathMatrix) left;
			scalar = (MathScalar) right;
		}

		// Scale the elements of the vector
		Number[][] result = new Number[matrix.rows][matrix.columns];
		for (int i = 0; i < matrix.rows; i++) {
			for (int j = 0; j < matrix.columns; j++) {
				result[i][j] = productSimple(scalar.realValue,
											 matrix.realValues[i][j]);
			}
		}

		// create and return a resultant Matrix
		return new MathMatrix(result);
	}

	// Compute the dot product of a pair of vectors
	else if (left instanceof MathVector && right instanceof MathVector) {

		MathVector vLeft = (MathVector) left;
		MathVector vRight = (MathVector) right;

		// Ensure that the vectors have the same size
		if (vLeft.size != vRight.size) {
			throw new OperandMismatchException("Cannot compute the dot product of different-sized vectors");
		}

		// Sum the products of elements of the vectors
		Number result = new Integer(0);
		for (int i = 0; i < vLeft.size; i++) {
			result = sum(result, productSimple(vLeft.realValues[i],
											   vRight.realValues[i]));
		}

		// create and return a resultant vector
		return new MathScalar(result);
	}

	// Compute the product of a vector and a matrix
	else if (left instanceof MathVector && right instanceof MathMatrix) {

		MathVector vector = (MathVector) left;
		MathMatrix matrix = (MathMatrix) right;

		// Ensure that the data sizes match
		if (vector.size != matrix.rows) {
			throw new OperandMismatchException("Vector and matrix sizes do not match");
		}

		// Sum the product of the vector against each column of the matrix
		Number[] result = new Number[matrix.columns];
		for (int i = 0; i < matrix.columns; i++) {

			result[i] = new Integer(0);
			for (int j = 0; j < vector.size; j++) {
				result[i] = sum(result[i],
								productSimple(vector.realValues[j],
											  matrix.realValues[j][i]));
			}
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Compute the product of a matrix and a vector
	else if (left instanceof MathMatrix && right instanceof MathVector) {

		MathMatrix matrix = (MathMatrix) left;
		MathVector vector = (MathVector) right;

		// Ensure that the data sizes match
		if (vector.size != matrix.columns) {
			throw new OperandMismatchException("Vector and matrix sizes do not match");
		}

		// Sum the product of the vector against each row of the matrix
		Number[] result = new Number[matrix.rows];
		for (int i = 0; i < matrix.rows; i++) {

			result[i] = new Integer(0);
			for (int j = 0; j < vector.size; j++) {
				result[i] = sum(result[i],
								productSimple(matrix.realValues[i][j],
											  vector.realValues[j]));
												  

			}
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Multiply a matrix onto another matrix
	else if (left instanceof MathMatrix && right instanceof MathMatrix) {

		MathMatrix mLeft = (MathMatrix) left;
		MathMatrix mRight = (MathMatrix) right;

		// Ensure that the matrices have complementary size
		if (mLeft.columns != mRight.rows) {
			throw new OperandMismatchException("Cannot multiply matrices of non-complementary sizes");
		}

		// Compute the result for each entry of the matrix
		Number[][] result = new Number[mLeft.rows][mRight.columns];
		for (int i = 0; i < mLeft.rows; i++) {
			for (int j = 0; j < mRight.columns; j++) {
				result[i][j] = productSimple(mLeft.realValues[i][j],
											 mRight.realValues[i][j]);

				// Sum the product of a row and a column
				result[i][j] = new Integer(0);
				for (int k = 0; k < mLeft.columns; k++) {
					result[i][j] = sum(result[i][j],
									productSimple(mLeft.realValues[i][k],
												  mRight.realValues[k][j]));
				}
			}
		}

		// create and return a resultant matrix
		return new MathMatrix(result);
	}

	// otherwise the types don't match
	else {
		throw new OperandMismatchException("Cannot multiply given operands");
	}
}

/**
 * Add a number to another
 */
private Number sum(Number accumulator, Number addend) {

	// Promote to doubles
	if (accumulator instanceof Double || addend instanceof Double) {
		return new Double(accumulator.doubleValue() + addend.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(accumulator.intValue() + addend.intValue());
	}
}

/**
 * Multiply two Numbers
 */
private Number productSimple(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(left.doubleValue() * right.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(left.intValue() * right.intValue());
	}
}

}
