/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Remainder.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;

import java.util.Iterator;

/**
 * This is a concrete implementation for the remained "%" operation.
 * Remainder is normally a binary operation, but we also permit
 * remainders to be taken against a list.
 */
public class Remainder extends Operation {

/**
 * Create an operation to take the reminder of one Expression WRT another
 */
public Remainder(Expression numerator, Expression divisor)
{
    super(numerator, divisor);
}

/**
 * Create an operation to take the remainder of an Expression WRT a Number
 */
public Remainder(Expression numerator, Number divisor)
{
    super(numerator, divisor);
}

/**
 * Create an operation to take the remainder of a Number WRT an Expression
 */
public Remainder(Number numerator, Expression divisor)
{
    super(numerator, divisor);
}

/**
 * Create an operation to take the remainder of a Number WRT another
 */
public Remainder(Number numerator, Number divisor)
{
    super(numerator, divisor);
}

/**
 * Return the remainder of the first operand divided by the second
 * @return The residue of the first operand after division by the second
 */
public Result evaluate()
{

    // Must have two operands to do anything
    if (leftOperand == null || rightOperand == null) {
        throw new IllegalArgumentException("Remainder must have two operands");
    }

    return(getResult(leftOperand, rightOperand));
}

private Result getResult(Expression left, Expression right)
{

    // Evaluate any nested operations
    if (left instanceof Operation || left instanceof ExpressionReference) {
        left = left.evaluate();
    }
    if (right instanceof Operation || right instanceof ExpressionReference) {
        right = right.evaluate();
    }

    // Compute the remainder from two elements
    if (left instanceof MathValue && right instanceof MathValue) {

        MathValue mvLeft = (MathValue)left;
        MathValue mvRight = (MathValue)right;

        return (doRemainder(mvLeft, mvRight));
    }

    // Compute remainders pairwise between elements of lists
    else if (left instanceof ExpressionList &&
             right instanceof ExpressionList)
    {

        ExpressionList xlLeft = (ExpressionList)left;
        ExpressionList xlRight = (ExpressionList)right;

        // Check that we have the same number of elements in each list
        if (xlLeft.size() != xlRight.size()) {
            throw new OperandMismatchException("Remainder:  Cannot use lists " +
                                               "of different sizes");
        }

        // get iterators over the left and right lists
        Iterator leftIter = xlLeft.iterator();
        Iterator rightIter = xlRight.iterator();
        ExpressionList result = new MathExpressionList();

        // Divide all elements of the list pairwise and create a result list
        while (leftIter.hasNext()) {
            Expression leftElem = (Expression)leftIter.next();
            Expression rightElem = (Expression)rightIter.next();

            result.add(getResult(leftElem, rightElem));
        }

        // return the result list
        return(result);
    }

    // Compute a set of remainders from a constant
    else if (left instanceof MathValue && right instanceof ExpressionList) {

        ExpressionList xlRight = (ExpressionList)right;

        // Get an iterator over the list
        Iterator rightIter = xlRight.iterator();
        ExpressionList result = new MathExpressionList();

        // Divide a constant by the elements of a list
        while (rightIter.hasNext()) {
            Expression rightElem = (Expression)rightIter.next();

            result.add(getResult(left, rightElem));
        }

        // Return the result list
        return(result);
    }

    // Compute remainders from a set relative to a constant
    else if (left instanceof ExpressionList && right instanceof MathValue) {

        ExpressionList xlLeft = (ExpressionList)left;

        // Get an iterator over the list
        Iterator leftIter = xlLeft.iterator();
        ExpressionList result = new MathExpressionList();

        // Divide all elements of the list by the scale factor
        while (leftIter.hasNext()) {
            Expression leftElem = (Expression)leftIter.next();

            result.add(getResult(leftElem, right));
        }

        // Return the result list
        return(result);
    }

    else {
        throw new OperandMismatchException("Remainder:  Cannot divide given " +
                                           "data types");
    }
}

/**
 * Compute the remainder of one value relative to another
 */
private MathValue doRemainder(MathValue left, MathValue right)
{

    // Divide two scalar quantities and return the remainder
    if (left instanceof MathScalar && right instanceof MathScalar) {

        MathScalar sLeft = (MathScalar)left;
        MathScalar sRight = (MathScalar)right;

        Number result = remainderSimple(sLeft.realValue, sRight.realValue);
        return(new MathScalar(result));
    }

    // Otherwise (both are not scalar) throw an exception (undefined operation)
    else {
        throw new OperandMismatchException("Cannot compute remainders from " +
                                           "a vector or a matrix");
    }
}

/**
 * Compute the remainder of two Numbers
 */
private Number remainderSimple(Number left, Number right)
{

    // Promote to doubles
    if (left instanceof Double || right instanceof Double) {

        // Take the integer part of the quotient (truncate to an int)
        int integer_part = (int)(left.doubleValue() / right.doubleValue());
        return(new Double(left.doubleValue() -
                          (right.doubleValue() * integer_part)));
    }
    // Otherwise use integers
    else {

        // Take the integer part of the quotient
        int integer_part = left.intValue() / right.intValue();
        return(new Integer(left.intValue() -
                           (right.intValue() * integer_part)));
    }
}

}
