/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: StdDeviation.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;




/**
 * This is a concrete implementation for the Standard Deviation operation.
 * We implement the "unbiased" estimator version of standard deviation,
 * since that is what's commonly understood by "Standard Deviation."
 * In essence, the implementation is:
 *     sd = root( (sum(1 to N)(element - average)^2) / (n - 1))
 */
public class StdDeviation extends Operation {

/**
 * The standard deviation of the given list of elements
 */
public StdDeviation(Expression operand) {
	super(operand);
}

/**
 * Standard Deviation of a pair of operands
 */
public StdDeviation(Expression operand1, Expression operand2) {
	super(operand1, operand2);
}

/**
 * Return the result of computing the population standard deviation over
 * the operands (all members have equal probability)
 * @return The standard deviation of all the operands
 */
public Result evaluate() {

	MathValue count, average, squares, result;
	MathValue left_squares = null, right_squares = null;
	Operation sumOp, countOp, divOp, subOp, resultOp;

	// Construct a sum operator to total the values
	sumOp = new Sum(leftOperand, rightOperand);

	// Construct a count operator to count the values
	countOp = new Count(leftOperand, rightOperand);
	count = (MathValue) countOp.evaluate();
	if (((MathScalar) count).realValue.intValue() < 2) {
		throw new IllegalArgumentException("Must have 2 or more elements to compute Standard Deviation");
	}

	// Divide the sum by the count
	divOp = new Divide(sumOp, count);

	// Compute the average (arithmetic mean) of the data
	average = (MathValue) divOp.evaluate();

	// Compute the square differences from the average
	if (leftOperand != null) {
		left_squares = getSquares(average, leftOperand);
	}
	if (rightOperand != null) {
		right_squares = getSquares(average, rightOperand);
	}

	// if there is only one operand use that value
	if (leftOperand == null) {
		squares = right_squares;
	}
	else if (rightOperand == null) {
		squares = left_squares;
	}

	// otherwise sum the operands
	else {
		sumOp = new Sum(left_squares, right_squares);
		squares = (MathValue) sumOp.evaluate();
	}

	// subtract one from the count
	subOp = new Subtract(count, new MathScalar(new Integer(1)));
	count = (MathValue) subOp.evaluate();

	// divide the sum of squares and take the square root
	resultOp = new Root(new Divide(squares, count));
	result = (MathValue) resultOp.evaluate();

	return(result);
}

/**
 * Count the elements in the given operand
 */
private MathValue getSquares(MathValue average, Expression operand) {

	// Values and operations are considered single elements
	if (operand instanceof MathValue ||
		operand instanceof ExpressionReference ||
		operand instanceof Operation) {

		// Compute the square of the difference
		Operation differenceOp = new Subtract(average, operand);
		Expression difference = differenceOp.evaluate();
		Operation multiplyOp = new Multiply(difference, difference);

		return((MathValue) multiplyOp.evaluate());
	}

	// Sum the square differences from the mean of elements in list
	else if (operand instanceof ExpressionList) {

		// Process the elements of the list
		Iterator eli = ((ExpressionList) operand).iterator();
		ExpressionList result = new MathExpressionList();

		while (eli.hasNext()) {

			// Add each square to the result array
			result.add(getSquares(average, (Expression) eli.next()));
		}

		// Total sum of all of the result elements
		Operation sumOp = new Sum(result);

		// return the total of square-difference of elements in this list
		return((MathValue) sumOp.evaluate());
	}

	else {
		throw new IllegalArgumentException("Unrecognized operand type to Standard Deviation");
	}
}

}
