/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: Sum.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;


import java.util.Iterator;




/**
 * This is a concrete implementation for the sum "+" operation.
 * A sum can be performed on a pair of elements or across a list.
 */
public class Sum extends Operation {

/**
 * Sum across the given operand (usually a list of expressions)
 */
public Sum(Expression operand) {
	super(operand);
}

/**
 * Sum over a pair of Expressions
 */
public Sum(Expression left, Expression right) {
	super(left, right);
}

/**
 * Sum over an Expression and a Number
 */
public Sum(Expression left, Number right) {
	super(left, right);
}

/**
 * Sum over a Number and an Expression
 */
public Sum(Number left, Expression right) {
	super(left, right);
}

/**
 * Sum of two Numbers
 */
public Sum(Number left, Number right) {
	super(left, right);
}

/**
 * Return the result of summing the operands
 * @return The result of this sum on the operands
 */
public Result evaluate() {

	if (leftOperand == null && rightOperand == null) {
		throw new IllegalArgumentException("Sum is not defined on zero operands");
	}

	MathValue leftOperandResult = null;
	MathValue rightOperandResult = null;

	if (leftOperand != null) {
		leftOperandResult = evaluateOperand(leftOperand);
	}
	if (rightOperand != null) {
		rightOperandResult = evaluateOperand(rightOperand);
	}

	// Don't create default (0) values 'cause we don't want to
	// muck with returned format or type
	if (leftOperand == null) {
		return(rightOperandResult);
	}
	if (rightOperand == null) {
		return(leftOperandResult);
	}

	// sum the results from the left and right operand
	return(doSum(leftOperandResult, rightOperandResult));
}

/**
 * Extract a value from the given operand
 */
private MathValue evaluateOperand(Expression operand) {

	if (operand instanceof MathValue) {
		return((MathValue) operand);
	}

	else if (operand instanceof Operation ||
			 operand instanceof ExpressionReference) {
		return(evaluateOperand(operand.evaluate()));
	}

	else if (operand instanceof ExpressionList) {

		MathValue result = null, tmp_result;

		// Process the elements of the list
		Iterator eli = ((ExpressionList) operand).iterator();
		while (eli.hasNext()) {

			// Get the value of each element
			tmp_result = evaluateOperand((Expression) eli.next());

			// Accumulate the result
			if (result == null) {
				result = tmp_result;
			}
			else {
				result = doSum(result, tmp_result);
			}
		}
		return(result);
	}

	else {
		throw new IllegalArgumentException("Unrecognized operand type to Sum");
	}
}

/**
 * Sum two values
 */
private MathValue doSum(MathValue left, MathValue right) {

	// Add two scalar quantities
	if (left instanceof MathScalar && right instanceof MathScalar) {

		MathScalar sLeft = (MathScalar) left;
		MathScalar sRight = (MathScalar) right;

		Number result = sumSimple(sLeft.realValue, sRight.realValue);
		return new MathScalar(result);
	}

	// Add a pair of vectors
	else if (left instanceof MathVector && right instanceof MathVector) {

		MathVector vLeft = (MathVector) left;
		MathVector vRight = (MathVector) right;

		// Ensure that the vectors have the same size
		if (vLeft.size != vRight.size) {
			throw new OperandMismatchException("Cannot add mixed data sizes");
		}

		// Compute the result for each entry of the vector
		Number[] result = new Number[vLeft.size];
		for (int i = 0; i < vLeft.size; i++) {
			result[i] = sumSimple(vLeft.realValues[i],
								  vRight.realValues[i]);
		}

		// create and return a resultant vector
		return new MathVector(result);
	}

	// Add a matrix to another matrix
	else if (left instanceof MathMatrix && right instanceof MathMatrix) {

		MathMatrix mLeft = (MathMatrix) left;
		MathMatrix mRight = (MathMatrix) right;

		// Ensure that the matrices have the same size
		if (mLeft.rows != mRight.rows || mLeft.columns != mRight.columns) {
			throw new OperandMismatchException("Cannot add mixed data sizes");
		}

		// Compute the result for each entry of the matrix
		Number[][] result = new Number[mLeft.rows][mLeft.columns];
		for (int i = 0; i < mLeft.rows; i++) {
			for (int j = 0; j < mLeft.columns; j++) {
				result[i][j] = sumSimple(mLeft.realValues[i][j],
										 mRight.realValues[i][j]);
			}
		}

		// create and return a resultant matrix
		return new MathMatrix(result);
	}

	// otherwise the types don't match (e.g. adding a vector to a matrix)
	else {
		throw new OperandMismatchException("Cannot add mixed data sizes");
	}
}

/**
 * Sum two Numbers
 */
private Number sumSimple(Number left, Number right) {

	// Promote to doubles
	if (left instanceof Double || right instanceof Double) {
		return new Double(left.doubleValue() + right.doubleValue());
	}
	// otherwise use integers
	else {
		return new Integer(left.intValue() + right.intValue());
	}
}

}
