/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: TableReference.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:41 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.formulae;

import java.awt.Point;

import com.klg.jclass.util.JCTableDataModel;

/**
 * TableReference selects a single cell from a JCTableDataModel
 */
public class TableReference implements ExpressionReference, Cloneable {

/** The table data from which the data is referenced */
protected JCTableDataModel		tableData;
/** The row location of the cell in the table */
protected Expression			row;
/** The column location of the cell in the table */
protected Expression			column;
/** A Point identifying the cell location, if it can be computed */
protected Point					location;

/**
 * Create a table value from a table data model and cell co-ordinates
 * @param dataModel The JCTableDataModel whose cell is referenced by this object
 * @param rowNumber The row the cell appears in
 * @param columnNumber The column the cell appears in
 */
public TableReference(JCTableDataModel tdm, int rowNum, int columnNum) {

	// Store the tabledata
	tableData = tdm;

	// Check the location for validity and store it
	location = new Point(columnNum, rowNum);
	checkLocation(location);

	// Store the coordinates of the location
	row = new MathScalar(rowNum);
	column = new MathScalar(columnNum);
}

/**
 * Create a table value from a table data model and cell location
 * @param dataModel The JCTableDataModel whose cell is referenced by this object
 * @param location The location of the cell
 */
public TableReference(JCTableDataModel tdm, Point location) {

	// Store the tabledata
	tableData = tdm;

	// Check the location for validity and store it
	checkLocation(location);
	this.location = location;

	// Store the coordinates of the location
	row = new MathScalar(location.y);
	column = new MathScalar(location.x);
}

/**
 * Create a table reference from a table data model and the row/column
 * location of a cell
 * @param dataModel The JCTableDataModel whose cell is referenced by this object
 * @param row The row in which the cell appears
 * @param column The column in which the cell appears
 */
public TableReference(JCTableDataModel tdm, Expression row, Expression column) {

	// Store the tabledata
	tableData = tdm;

	// Compute the location of the cell defined by the Expressions
	location = computeLocation(row, column);

	// Store the coordinates of the location (even if they're not defined yet)
	this.row = row;
	this.column = column;
}

/**
 * Satisfy the Expression interface
 * @return This value
 */
public Result evaluate() {

	// Get the data from the cell's location
	location = computeLocation(row, column);
	Object datum = tableData.getTableDataItem(location.y, location.x);

	// Return the correct data type
	if (datum instanceof Result) {
		// Result objects are returned directly
		return((Result) datum);
	}
	else if (datum instanceof Expression) {
		// Expressions are evaluated and the result returned
		return(((Expression) datum).evaluate());
	}
	else {
		// Numbers are encapsulated in MathValue objects
		return((Result) encapsulateValue(datum));
	}
}

/**
 * Return an Expression which is a clone of this TableReference
 * (NOTE:  The referenced table data is not cloned, but if the location
 *         contains any parametrized values they are substituted with
 *         their current values)
 */
public Object clone() {

	TableReference newReference;

	try {
		newReference = (TableReference) super.clone();
	}
	catch (CloneNotSupportedException e) {
		return(null);
	}

	if (row != null) {
		try {
			newReference.row = (Expression) row.evaluate().clone();
		}
		catch (CloneNotSupportedException e) {
			newReference.row = new MathScalar(0);
		}
	}
	if (column != null) {
		try {
			newReference.column = (Expression) column.evaluate().clone();
		}
		catch (CloneNotSupportedException e) {
			newReference.column = new MathScalar(0);
		}
	}
	if (location != null) {
		newReference.location = new Point(location);
	}

	return(newReference);
}

/**
 * Return the location of the table cell being used by this expression
 * @return A Point storing the coordinates of the referenced cell
 */
public Point getLocation() {

	// if the location is not computed yet, compute it
	if (location == null) {
		return(computeLocation(row, column));
	}

	return(new Point(location));
}

/**
 * Set the location of the cell accessed by this expression
 * @param newLocation The location of the cell in the table data
 */
public void setLocation(Point newLocation) {

	// Check that the location is permissible
	checkLocation(newLocation);

	// set the location
	this.location = new Point(newLocation);

	// Store the extents of the location
	this.row = new MathScalar(location.y);
	this.column = new MathScalar(location.x);
}

/**
 * Get the column coordinate of the referenced cell
 * @return The number of the cell's column in the table
 */
public Expression getColumn() {
	return(column);
}

/**
 * Set the column coordinate of the referenced cell
 * @param column The column location of the cell in the table
 */
public void setColumn(Expression column) {

	// Verify the location
	location = computeLocation(row, column);

	// ??? Should we clone this?
	this.column = column;
}

/**
 * Get the row coordinate of the referenced cell
 * @return The number of the cell's row in the table
 */
public Expression getRow() {
	return(row);
}

/**
 * Set the row coordinate of the referenced cell
 * @param row The row location of the cell in the table
 */
public void setRow(Expression row) {

	// Verify the location
	location = computeLocation(row, column);

	// ??? Should we clone this?
	this.row = row;
}

/**
 * Set a table data object
 */
/*
protected void setTableData(JCTableDataModel tdm) {

	// Store the tabledata
	tableData = tdm;

	// Check the location for validity against the new data
	if (location != null) {
		checkLocation(location);
	}
}
*/

/**
 * Verify a location as being within the range of the given table data
 */
private void checkLocation(Point loc) {

	// Check for loc < 0 or loc > end
	if (loc.x < 0 || loc.x > tableData.getNumColumns()) {
		throw new IllegalArgumentException("Column selection is out of range");
	}

	if (loc.y < 0 || loc.y > tableData.getNumRows()) {
		throw new IllegalArgumentException("Row selection is out of range");
	}
}

/**
 * Compute the location from the Expressions, if necessary and possible
 * @return The location of the cell defined by the coordinate Expressions
 */
private Point computeLocation(Expression r, Expression c) {

	// Attempt to evaluate the Expressions to get row and column numbers
	try {

		// Get the start and end row and columns from the Expressions
		int r_tmp = ((MathValue) r.evaluate()).numberValue().intValue();
		int c_tmp = ((MathValue) c.evaluate()).numberValue().intValue();

		// Verify the location and store it if it's valid
		Point pt = new Point(c_tmp, r_tmp);
		checkLocation(pt);
		return(pt);
	}

	// We catch exceptions thrown for undefined variables
	catch (ArithmeticException e) {
		return(null);
	}
}

/**
 * Make sure the location has been evaluated; throw an exception if it can
 * not yet be evaluted.
 */
private void updateLocation() {

	// If the location has been evaluated, we don't need to do anything
	if (location != null) {
		return;
	}

	// Try to evaluate the location
	location = computeLocation(row, column);

	// If it was not possible to evaluate the location, this will
	// throw an appropriate exception
	if (location == null) {
		Result temp_row = row.evaluate();
		Result temp_column = column.evaluate();
	}
}

/**
 * Compare the given object and this TableReference, given that equality
 * implies both objects reference the same entry in the same table.
 * @param obj the object whose equality with this TableReference is tested
 * @return True if the object refers to the same object as this TableReference
 */
public boolean equals(Object obj) {

	// Can only be equal to another TableReference
	if (!(obj instanceof TableReference)) {
		return(false);
	}

	TableReference tr = (TableReference) obj;

	// TableData objects must be equivalent (how to compare otherwise?)
	if (tr.tableData != tableData) {
		return(false);
	}

	// if location is defined (evaluated) then it must be equal
	if (location != null) {
		return(location.equals(tr.location));
	}

	// Here we have a case where the location contains undefined variables
	// The only way we have of testing equality is if the expressions are
	// identical, since we can't algebraically compare them.
	return(column == tr.column && row == tr.row);
}

/**
 * Create a copy of this table reference (not necessarily a clone, since we
 * may choose to use hard co-ordinates rather than parametric ones
 * @param fixLocation Indicates if the row and column positions are to be
 *  				  evaluated before being set
 * @return A new TableReference which is a copy of this one (it will point
 *  	   to the same table cell this one currently does
 */
protected Expression copy(boolean fixLocation) {

	// If we are fixing the location, evaluate the co-ordinates
	if (fixLocation) {
		location = computeLocation(row, column);
	}

	// if we're not fixing the location or there is no valid current location
	if (!fixLocation || location == null) {
		// create a copy using the same expressions
		return(new TableReference(tableData, row, column));
	}

	// otherwise use the computed location
	else {
		return(new TableReference(tableData, location.y, location.x));
	}
}

/**
 * Wrap a Number or other expression operand in a MathValue
 */
private Object encapsulateValue(Object obj) {

	// Expressions should be evaluated to produce a Result
	if (obj instanceof Expression) {
		return(((Expression) obj).evaluate());
	}

	// Numbers should be wrapped into a MathValue
	else if (obj instanceof Number) {
		return(new MathScalar((Number) obj));
	}
	else if (obj instanceof Number[]) {
		return(new MathVector((Number[]) obj));
	}
	else if (obj instanceof Number[][]) {
		return(new MathMatrix((Number[][]) obj));
	}

	// Try to convert strings to numbers
	else if (obj instanceof String) {
		String s = (String) obj;
		if (s.indexOf('.') >= -1) {
			return(new MathScalar(Double.valueOf(s)));
		}
		else {
			return(new MathScalar(Integer.valueOf(s)));
		}
	}

	// No other types can be converted (at least for now)
	else {
		throw new ClassCastException("Can't convert " + obj + " to type MathValue");
	}
}

}
