/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: LoadProperties.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:42 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.io;

import org.xml.sax.ErrorHandler;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * Encapsulate the user properties that may be needed during the creation of a
 * JClass object (such as JCChart, JCServerChart, or JCDocument) from loaded
 * files. These properties include classes used to resolve files and a
 * user-specified object that will be associated with the created JClass object.
 */
public class LoadProperties extends Properties
{

/** Class used to resolve files relative to the classpath. */
protected Class             resolvingClass;

/** Prefix pre-pended to relative URLs */
protected String            relativeURLPrefix;

/**
 * User-specified object. Will be set on the created JClass object if
 * storeUserObject is true.
 */
protected Object			userObject;

/** If true, user object will be stored on created JClass object. */
protected boolean			storeUserObject;

/** System ID used for local (to JClass) external entities, such as DTDs. */
protected String			localEntitySystemId = "";

/** Error handler for XML */
protected ErrorHandler      xmlErrorHandler = null;

/**
 * Default constructor.  Use default values for all properties.
 */
public LoadProperties()
{
    this(null, "", null, true);
}

public LoadProperties(Class resolvingClass, String relativeURLPrefix,
                      Object userObject, boolean storeUserObject)
{
    setResolvingClass(resolvingClass);
    setRelativeURLPrefix(relativeURLPrefix);
    setUserObject(userObject);
    setStoreUserObject(storeUserObject);
}

/**
 * Set a Class object that will be used to resolve files relative
 * to the classpath.  Such files are resolved using the
 * getResource() method of Class.
 * @param resolvingClass Class object to be used to resolve files
 */
public void setResolvingClass(Class resolvingClass)
{
	this.resolvingClass = resolvingClass;
}

/**
 * Get the Class object that will be used to resolve files relative
 * to the classpath.  Such files are resolved using the
 * getResource() method of Class.
 * @return The Class object to be used to resolve files
 */
public Class getResolvingClass()
{
	return(resolvingClass);
}

/**
 * Set the prefix that will be prepended to relative urls
 * @param prefix The prefix that will be prepended to any relative urls
 */
public void setRelativeURLPrefix(String prefix)
{
    if (prefix == null) {
        prefix = "";
    }
    this.relativeURLPrefix = prefix;
}

/**
 * Get the prefix that will be prepended to relative urls
 * @return The prefix that will be prepended to any relative urls
 */
public String getRelativeURLPrefix()
{
    return(relativeURLPrefix);
}

/**
 * Set a user-specified object.  If the StoreUserObject property is true,
 * the object will be stored in the created JClass object's UserObject property.
 * @param userObject Any user-specified object
 */
public void setUserObject(Object userObject)
{
	this.userObject = userObject;
}

/**
 * Get the user-specified object.  If the StoreUserObject property is true,
 * the object will be stored in the created JClass object's UserObject property.
 * @return The user-specified object
 */
public Object getUserObject()
{
	return(userObject);
}

/**
 * Note whether the object specified in the UserObject
 * property should be stored on the created JClass object.
 * The default value is true.
 * @param storeUserObject true if the UserObject property should be
 *                        stored on the created JClass object; false otherwise.
 */
public void setStoreUserObject(boolean storeUserObject)
{
	this.storeUserObject = storeUserObject;
}

/**
 * Determine whether the object specified in the UserObject
 * property will be stored on the created JClass object.
 * The default value is true.
 * @return true if the UserObject property should be
 *         stored on the created JClass object; false otherwise.
 */
public boolean getStoreUserObject()
{
	return(storeUserObject);
}

/**
 * Given a file name string and an access type, return an InputStream based
 * on these two values.
 * @param fileName The file name string
 * @param accessType The type of access to file name.  Either an absolute
 * filename, a url, a relative url to a url prefix, or a relatiave to a class.
 * @return
 * @throws java.net.MalformedURLException
 */
public URL getURLByAccessType(String fileName, int accessType)
    throws MalformedURLException
{
    URL possibleURL = null;
    String message = null;

    switch (accessType) {
    case ABSOLUTE:
    case DEFAULT_ACCESS:
        File file = new File(fileName);
        possibleURL = file.toURL();
        break;
    case RESOLVING_CLASS:
        if (resolvingClass != null) {
            // Create an url by resovling the file name against the
            // provided class
            possibleURL = resolvingClass.getResource(fileName);
            // Class.getResource() returns null when the file doesn't exist,
            // so we can just check for that case in determining if the
            // file exists
            if (possibleURL == null) {
                // If url is null, that means the file was not found
                message = "Cannot find file relative to resolving class";
            }
        }
        break;
    case URL:
        possibleURL = new URL(fileName);
        break;
    case RELATIVE_URL:
        String urlString = relativeURLPrefix;
        if (fileName.charAt(0) != '/') {
            urlString += '/';
        }
        urlString += fileName;
        possibleURL = new URL(urlString);
        break;
    default:
        throw new IllegalArgumentException("Bad access type: " + accessType +
                                           " for file: " + fileName);
    }

    if (possibleURL == null) {
        String fileAccess = getFileAccessStringFromEnum(accessType, true);
        String errString = (message == null) ? "" : " Error:" + message;
        throw new MalformedURLException("Cannot create URL from:" + fileName +
                                        " for access type:" + fileAccess +
                                        errString);
    }
    return(possibleURL);
}

/**
 * Given a file name string and an access type, return an InputStream based
 * on these two values.
 * @param fileName The file name string
 * @param accessType The type of access to file name.  Either an absolute
 * filename, a url, a relative url to a url prefix, or a relatiave to a class.
 * @return
 * @throws java.io.IOException
 * @throws java.net.MalformedURLException
 */
public InputStream getInputStreamByAccessType(String fileName, int accessType)
    throws IOException, MalformedURLException
{
    InputStream inStream = null;
    URL possibleURL = getURLByAccessType(fileName, accessType);
    if (possibleURL != null) {
        inStream = possibleURL.openStream();
    }
    return(inStream);
}

/**
 * Set System ID used for local (to jclass) external entities,
 * such as DTDs.  Default is empty string.  Other valid values
 * include "file:/".
 */
public void setLocalEntitySystemId(String id)
{
	localEntitySystemId = id;
}

/**
 * Get System ID used for local (to jclass) external entities,
 * such as DTDs.  Default is empty string.  Other valid values
 * include "file:/".
 */
public String getLocalEntitySystemId()
{
	return(localEntitySystemId);
}

/**
 * Return the current XML handler.  If null the default handler
 * (com.klg.jclass.util.xml.JCXMLErrorPrinter) will be used.
 * @return The current error handler.
 * @see com.klg.jclass.util.xml.JCXMLErrorPrinter
 */
public ErrorHandler getXmlErrorHandler()
{
    return(xmlErrorHandler);
}

/**
 * Set the current XML handler.  If null, the default handler
 * (com.klg.jclass.util.xml.JCXMLErrorPrinter) will be used.
 * @param xmlErrorHandler The new error handler.
 * @see com.klg.jclass.util.xml.JCXMLErrorPrinter
 */
public void setXmlErrorHandler(ErrorHandler xmlErrorHandler)
{
    this.xmlErrorHandler = xmlErrorHandler;
}

}
