/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: OutputDataProperties.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:42 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.io;

import com.klg.jclass.util.JCTypeConverter;
import com.klg.jclass.util.io.resources.LocaleBundle;

/**
 * Encapsulate the properties that may be needed to save a data file
 */
public class OutputDataProperties extends OutputProperties
{

public final static int NO_DATA        = 0;
public final static int EMBED_DATA     = 1;
public final static int DATA_FILE_TEXT = 2;
public final static int DATA_FILE_XML  = 3;

// Data save type values/strings
public final static int[] dataSave_values = {
    NO_DATA, EMBED_DATA, DATA_FILE_TEXT, DATA_FILE_XML
};
public final static String[] dataSave_strings = {
    "None", "Embed", "Text", "Xml"
};
public final static String[] dataSave_i18n_strings = {
    LocaleBundle.string(LocaleBundle.NO_DATA),
    LocaleBundle.string(LocaleBundle.EMBED_DATA),
    LocaleBundle.string(LocaleBundle.DATA_FILE_TEXT),
    LocaleBundle.string(LocaleBundle.DATA_FILE_XML)
};

protected String fileCharset = DEFAULT_CHARSET_NAME;

/**
 * Default constructor.  Properties set to defaults
 */
public OutputDataProperties()
{
    this(null, null, null, OutputProperties.DEFAULT_ACCESS);
}

/**
 * Constructor which specifies output fileName, propertyName, saveType,
 * and fileAccess
 * @param outputFileName Name of actual output data file to write to (if
 *     null, no data is written out)
 * @param propertyName Name to write to the property
 * @param saveType The type of file to save
 * @param fileAccess Type of access to file.  Must be one of
 *     <code>Properties.DEFAULT_ACCESS</code>,
 *     <code>Properties.ABSOLUTE</code>,
 *     <code>Properties.RESOLVING_CLASS</code>,
 *     <code>Properties.URL</code>,
 *     <code>Properties.RELATIVE_URL</code>,
 *     <code>Properties.SERVLET</code>
 */
public OutputDataProperties(String outputFileName, String propertyName,
                            String saveType, int fileAccess)
{
    super(outputFileName, propertyName,
          (saveType == null) ? dataSave_strings[1] : saveType, fileAccess);
}

/**
 * Gets the name of the charset to use when reading data
 * @return The name of the current charset
 */
public String getFileCharset()
{
    return(fileCharset);
}

/**
 * Sets the name of the charset to use when reading data
 * @param fileCharset The name of a supported charset
 */
public void setFileCharset(String fileCharset)
{
    this.fileCharset = fileCharset;
}

/**
 * Sets the type of file to save as a string.
 * @param saveType The new way to save data.  Must be one of "None", "Embed",
 *     "Text", or "Xml".
 */
public void setSaveType(String saveType)
{
    String newSaveType = null;
    if (saveType != null) {
        for (int i = 0; i < dataSave_strings.length; i++) {
            if (dataSave_strings[i].compareToIgnoreCase(saveType) == 0) {
                newSaveType = saveType;
            }
        }
    }
    if (newSaveType == null) {
        throw new IllegalArgumentException("Bad Save Type:" + saveType);
    }
    this.saveType = newSaveType;
}

/**
 * Set the type of file to save as an enum.
 * @param saveEnum The new save type as an enum.  Must be one of:
 *     <code>NO_DATA</code>, <code>EMDED_DATA</code>,
 *     <code>DATA_FILE_TEXT</code>, <code>DATA_FILE_XML</code>
 */
public void setSaveType(int saveEnum)
{
    setSaveType(getSaveTypeStringFromEnum(saveEnum, false));
}

/**
 * Return the save type as an enum.
 * @return The save type as an enum.  Will be one of:
 *     <code>NO_DATA</code>, <code>EMDED_DATA</code>,
 *     <code>DATA_FILE_TEXT</code>, <code>DATA_FILE_XML</code>
 */
public int getSaveTypeAsEnum()
{
    return(getSaveTypeEnumFromString(saveType, false));
}

/**
 * Get the file access enum from its string representation.  Default access is
 * absolute access. May be over-ridden to provide different behaviour for
 * sub-classes.
 * @param accessString The string version of the fileAccess
 * @param useI18n Use the internationalized strings
 * @return The enum corresponding to the string
 */
public static int getSaveTypeEnumFromString(String accessString,
                                            boolean useI18n)
{
    String[] strings = useI18n ? dataSave_i18n_strings : dataSave_strings;
    return(JCTypeConverter.toEnum(accessString, strings, dataSave_values,
                                  EMBED_DATA));
}

/**
 * Get the file access string representation from its enum value.
 * @param accessValue The enum version of the fileAccess
 * @param useI18n Use the internationalized strings
 * @return The String reprsentation of the fileAccess enum
 */
public static String getSaveTypeStringFromEnum(int accessValue,
                                               boolean useI18n)
{
    String[] strings = useI18n ? dataSave_i18n_strings : dataSave_strings;
    return(JCTypeConverter.fromEnum(accessValue, strings, dataSave_values));
}

}
