/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCGridLegend.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.legend;

import com.klg.jclass.util.Changeable;

import java.awt.*;
import java.util.List;

/**
 * <code>JCGridLegend</code> manages and renders a grid-based legend.
 * It overrides <code>JCLegend</code>'s layout methods to position legend items
 * in a grid, where every row has the same height and every column the same
 * width.
 */
public class JCGridLegend extends JCLegend {

// The dimensions of the various rows and columns
public int rowHeights[];
public int rowFirstLineHeights[]; // height of the first line of each row/column
public int colWidths[];
public int symWidths[];           // symbol widths per row/column
public int strWidths[];           // string widths per row/column

protected int marginGap     = 2;  // space between  utside and inside of legend
protected int itemGap       = 2;  // space between items
protected int insideItemGap = 2;  // space between symbol and string inside item
protected int groupGap      = 5;  // space between groupings of item (ie. by
                                  // data views)

public JCGridLegend()
{
    super();
}

/** Sets the gap between the edge of the legend and where the items start
 *  being layed out.
 *  @param gap the margin gap
 */
public void setMarginGap(int gap)
{
    marginGap = gap;

    setChanged(true, Changeable.RECALC);
}

/** Gets the gap between the edge of the legend and where the items start
 *  being layed out.
 *  @return the margin gap
 */
public int getMarginGap()
{
    return(marginGap);
}

/** Sets the gap between the legend items in the same group.
 *  @param gap the item gap
 */
public void setItemGap(int gap)
{
    itemGap = gap;

    setChanged(true, Changeable.RECALC);
}

/** Gets the gap between the legend items in the same group.
 *  @return the item gap
 */
public int getItemGap()
{
    return(itemGap);
}

/** Sets the gap between the symbol and text portions of a legend item.
 *  @param gap the inside item gap
 */
public void setInsideItemGap(int gap)
{
    insideItemGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between the symbol and text portions of a legend item.
 * @return the inside item gap
 */
public int getInsideItemGap()
{
    return(insideItemGap);
}

/**
 * Sets the gap between the groups of items in the legend (for example, the
 * columns/rows associated with a data view).
 * @param gap the group gap
 */
public void setGroupGap(int gap)
{
    groupGap = gap;

    setChanged(true, Changeable.RECALC);
}

/**
 * Gets the gap between the groups of items in the legend (for example, the
 * columns/rows associated with a data view).
 * @return the group gap
 */
public int getGroupGap()
{
    return(groupGap);
}

/**
 * Overrides <code>JCLegend</code>'s <code>layoutLegend()</code> to layout the
 * legend items in a grid, calculating the overall size of the legend.  This is
 * accomplished by filling in the <code>dim</code>, <code>pos</code>,
 * <code>textPos</code>, and <code>symPos</code> member of each
 * <code>JCLegendItem</code> instance in the legend vector.  The
 * <code>pickRectangle</code> member is also filled in.
 * @param itemList  the legend vector, which contains a number of group
 * vectors; each group vector contains the legend items
 * (<code>JCLegendItem</code> instances) for that group
 * @param vertical <code>true</code> if a vertical legend, <code>false</code>
 * if a horizontal legend
 * @param useFont the font for the legend
 * @return the size of the legend
 */
public Dimension layoutLegend(List itemList, boolean vertical, Font useFont)
{
    fillSizeArrays(itemList, vertical, useFont);
    Dimension dim = calcLegendSize(vertical);
    layoutLegendItems(itemList, vertical);
    return(dim);
}

/** Cacluates the size of each row and column.
 *  @param itemList the legend vector of groups and item
 *  @param vert <code>true</code> if a vertical legend, <code>false</code> if a
 *  horizontal legend
 *  @param useFont the font for the legend
 */
public void fillSizeArrays(List itemList, boolean vert, Font useFont)
{
    int numViews = itemList.size();
    int maxItemsPerView = 0;
    for (int i = 0; i < numViews; i++) {
        maxItemsPerView = Math.max(maxItemsPerView,
                                   ((List)itemList.get(i)).size());
    }

    rowHeights = new int[vert ? maxItemsPerView : numViews];
    for (int tr = 0; tr < rowHeights.length; tr++) {
        rowHeights[tr] = 0;
    }

    rowFirstLineHeights = new int[vert ? maxItemsPerView : numViews];
    for (int tr = 0; tr < rowFirstLineHeights.length; tr++) {
        rowFirstLineHeights[tr] = 0;
    }

    colWidths = new int[vert ? numViews : maxItemsPerView];
    for (int tc = 0; tc < colWidths.length; tc++) {
        colWidths[tc] = 0;
    }

    symWidths = new int[vert ? numViews : maxItemsPerView];
    for (int tc = 0; tc < symWidths.length; tc++) {
        symWidths[tc] = 0;
    }

    strWidths = new int[vert ? numViews : maxItemsPerView];
    for (int tc = 0; tc < strWidths.length; tc++) {
        strWidths[tc] = 0;
    }

    // Now calculate the preferred width and height
    for (int i = 0; i < numViews; i++) {
        List viewItems = (List)itemList.get(i);

        for (int j = 0; j < viewItems.size(); j++) {
            JCLegendItem thisItem = (JCLegendItem)viewItems.get(j);

            Dimension lineDim = null;
            if (thisItem.contents instanceof String) {
                if (thisItem.textDim != null) {
                    if (textRend.isHTML((String)thisItem.contents)) {
                        lineDim =
                            textRend.calcHTMLLineSize((String)thisItem.contents,
                                                      thisItem.textDim.width,
                                                      thisItem.textDim.height);
                    }
                    else {
                        lineDim = new Dimension(thisItem.textDim.width, ascent);
                    }
                }
            }
            else {
                if (thisItem.textDim != null) {
                    lineDim = new Dimension(thisItem.textDim.width,
                                            thisItem.textDim.height);
                }
            }

            if (vert) {
                // special case of just the title
                if (isTitleItem(thisItem)) {
                    if (thisItem.textDim != null &&
                        colWidths[i] < thisItem.textDim.width)
                    {
                        colWidths[i] = thisItem.textDim.width;
                    }
                }
                else {
                    if (thisItem.symbolDim != null &&
                        symWidths[i] < thisItem.symbolDim.width)
                    {
                        symWidths[i] = thisItem.symbolDim.width;
                    }
                    if (thisItem.textDim != null &&
                        strWidths[i] < thisItem.textDim.width)
                    {
                        strWidths[i] = thisItem.textDim.width;
                    }
                }

                if (thisItem.symbolDim != null &&
                    rowHeights[j] < thisItem.symbolDim.height)
                {
                    rowHeights[j] = thisItem.symbolDim.height;
                }

                if (thisItem.textDim != null &&
                    rowHeights[j] < thisItem.textDim.height)
                {
                    rowHeights[j] = thisItem.textDim.height;
                }

                if (lineDim != null &&
                    rowFirstLineHeights[j] < lineDim.height)
                {
                    rowFirstLineHeights[j] = lineDim.height;
                }

            }
            else {
                if (thisItem.symbolDim != null &&
                    rowHeights[i] < thisItem.symbolDim.height)
                {
                    rowHeights[i] = thisItem.symbolDim.height;
                }

                if (thisItem.textDim != null &&
                    rowHeights[i] < thisItem.textDim.height)
                {
                    rowHeights[i] = thisItem.textDim.height;
                }

                if (lineDim != null &&
                    rowFirstLineHeights[i] < lineDim.height)
                {
                    rowFirstLineHeights[i] = lineDim.height;
                }

                if (thisItem.symbolDim != null &&
                    symWidths[j] < thisItem.symbolDim.width)
                {
                    symWidths[j] = thisItem.symbolDim.width;
                }

                if (thisItem.textDim != null &&
                    strWidths[j] < thisItem.textDim.width)
                {
                    strWidths[j] = thisItem.textDim.width;
                }
            }
        }
    }

    // Firm up the column widths -- needed to calculate
    // all of the symbol and string widths first.
    for (int i = 0; i < colWidths.length; i++) {
        int effColWidth = symWidths[i] + strWidths[i];
        if (symWidths[i] != 0 && strWidths[i] != 0) {
            effColWidth += getInsideItemGap();
        }
        if (colWidths[i] < effColWidth) {
            colWidths[i] = effColWidth;
        }
    }
}

/** Calculates the size of the legend based on the height and width
 *  of rows and columns.
 *  @param vert <code>true</code> if a vertical legend, <code>false</code> if a
 *  horizontal legend
 * @return A Dimension object containing the calculated legend size
 */
public Dimension calcLegendSize(boolean vert)
{
    // Leave space for top and left gaps.  Other gaps are included
    // in row heights and column widths.
    Dimension dim = new Dimension(0, 0);
    for (int wl = 0; wl < colWidths.length; wl++) {
        dim.width += colWidths[wl];
    }
    for (int hl = 0; hl < rowHeights.length; hl++) {
        dim.height += rowHeights[hl];
    }

    // Extra space between data views and legend intems
    if (vert) {
        dim.width += ((colWidths.length - 1) * getGroupGap());
        dim.width += (2 * getMarginGap());
        dim.height += ((rowHeights.length - 1) * getItemGap());
        dim.height += (2 * getMarginGap());
    }
    else {
        dim.width += ((colWidths.length - 1) * getItemGap());
        dim.width += (2 * getMarginGap());
        dim.height += ((rowHeights.length - 1) * getGroupGap());
        dim.height += (2 * getMarginGap());
    }
    return(dim);
}

/**  Lays out the legend items from the legend vector by calculating their
 *   position within the legend and filling that in the
 *   <code>JCLegendItem</code> instances that define the legend item.
 *   @param itemList the legend vector of groups and items
 *   @param vert <code>true</code> if a vertical legend, <code>false</code> if
 *   a horizontal legend
 */
public void layoutLegendItems(List itemList, boolean vert)
{
    // Now layout the legend items, filling in the dim, pos, textPos, symPos
    // elements of the legend item

    int major = getMarginGap();
    int minor;

    for (int i = 0; i < itemList.size(); i++) {
        List viewItems = (List)itemList.get(i);

        minor = getMarginGap();

        for (int j = 0; j < viewItems.size(); j++) {
            JCLegendItem thisItem = (JCLegendItem)viewItems.get(j);
            thisItem.pos = new Point(vert ? major : minor,
                                     vert ? minor : major);
            thisItem.dim = new Dimension(vert ? colWidths[i] : colWidths[j],
                                         vert ? rowHeights[j] : rowHeights[i]);

            // special case -- title gets left aligned
            if (isTitleItem(thisItem)) {
                thisItem.textPos = positionItemText(vert, thisItem,
                                                    thisItem.dim,
                                                    new Point(0, 0));
                // normal case -- symbol and text component
            }
            else {
                int height = vert ? rowFirstLineHeights[j]
                                  : rowFirstLineHeights[i];
                thisItem.symbolPos =
                    positionItemSymbol(vert, thisItem, new Dimension(0, height),
                                       new Point(0, 0));
                thisItem.textPos =
                    positionItemText(vert, thisItem, new Dimension(0, height),
                                     new Point(getInsideItemGap() +
                                               (vert ? symWidths[i]
                                                     : symWidths[j]),
                                               0));
            }

            thisItem.pickRectangle = new Rectangle(thisItem.pos.x,
                                                   thisItem.pos.y,
                                                   thisItem.dim.width,
                                                   thisItem.dim.height);
            // if last item
            if (j != viewItems.size() - 1) {
                if (vert) {
                    thisItem.pickRectangle.height += getItemGap();
                }
                else {
                    thisItem.pickRectangle.width += getItemGap();
                }
            }
            if (i != itemList.size() - 1) {
                if (vert) {
                    thisItem.pickRectangle.width += getGroupGap();
                }
                else {
                    thisItem.pickRectangle.height += getGroupGap();
                }
            }

            minor += (vert ? rowHeights[j] : colWidths[j]) + getItemGap();
        }

        major += (vert ? colWidths[i] : rowHeights[i]) + getGroupGap();
    }
}

/**
 * Positions an legend item's symbol component.  It is centered vertically
 * for a horizontal legend.  It is centered within the given block (the
 * biggest first text line) for a vertical legend.
 * @param vertical <code>true</code> for a vertical legend; <code>false</code>
 * for a horizontal legend
 * @param thisItem the legend item containing the symbol to be positioned
 * @param boxDim the area to be positioned in
 * @param startPos the position within the item's area to start positioning at
 * @return the position within the item's area to position the symbol at
 */
public Point positionItemSymbol(boolean vertical, JCLegendItem thisItem,
                                Dimension boxDim, Point startPos)
{
    if (vertical) {
        if (thisItem.symbolDim.height < boxDim.height) {
            return(centerItemComponentVertically(vertical, thisItem.symbolDim,
                                                 boxDim.height, startPos));
        }
        else {
            return(startPos);
        }
    }
    else {
        return(centerItemComponentVertically(vertical, thisItem.symbolDim,
                                             thisItem.dim.height, startPos));
    }
}

/* Positions an legend item's text component.  It is centered vertically
* for a horizontal legend.  It is centered within the given block (the
* biggest first text line) for a vertical legend.
* @param vertical <code>true</code> for a vertical legend; <code>false</code>
for a horizontal legend
* @param thisItem the legend item containing the text to be positioned
* @param boxDim the area to be positioned in
* @param startPos the position within the item's area to start positioning at
* @return the position within the item's area to position the text at
*/
public Point positionItemText(boolean vertical, JCLegendItem thisItem,
                              Dimension boxDim, Point startPos)
{
    if (vertical) {
        if (thisItem.textDim.height < boxDim.height) {
            return(centerItemComponentVertically(vertical, thisItem.textDim,
                                                 boxDim.height, startPos));
        }
        else {
            return(startPos);
        }
    }
    else {
        return(centerItemComponentVertically(vertical, thisItem.textDim,
                                             thisItem.dim.height, startPos));
    }
}

/** Centers an item component (symbol, text) vertically withing the
 *  given box, starting at the given starting position.
 * @param vertical <code>true</code> for a vertical legend; <code>false</code>
 * for a horizontal legend
 * @param compDim the size of the item being positioned
 * @param boxHeight the height of the area to be centered vertically within
 * @param startPos the position within the item's area to start centering at
 * @return the position within the item's area to position the symbol at
 */
Point centerItemComponentVertically(boolean vertical, Dimension compDim,
                                    int boxHeight, Point startPos)
{
    Point p = new Point(startPos.x, startPos.y);

    if (boxHeight - startPos.y > compDim.height) {
        p.y += (int)Math.ceil((double)(boxHeight - startPos.y -
                                       compDim.height) / 2);
    }
    return(p);
}

}
