/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCLegend.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.legend;

import com.klg.jclass.util.Changeable;
import com.klg.jclass.util.ImageMapInfo;
import com.klg.jclass.util.swing.TextRenderer;

import javax.swing.*;
import java.awt.*;
import java.io.IOException;
import java.util.List;
import java.util.Locale;
import java.util.MissingResourceException;

/**
 * <code>JCLegend</code> manages and renders a legend for a component.  It is an
 * abstract class that requires a subclass to provide a specific
 * layout.  It requires a <code>JCLegendPopulator</code> instance to populate it
 * with legend items and a <code>JCLegendRenderer</code> instance to help in
 * rendering the items.  Each legend item is represented by a
 * <code>JCLegendItem</code> instance.
 *
 * <code>JCLegend</code> is a Swing-based component and should behave
 * accordingly.  It is not visible by default.  It is also transparent by
 * default, meaning that the <code>Opaque</code> property must be set to
 * <code>true</code> if any background color is to be drawn.
 *
 * @see JCGridLegend
 * @see JCMultiColLegend
 * @see JCLegendPopulator
 * @see JCLegendRenderer
 * @see JCLegendItem
 */
public abstract class JCLegend extends JComponent
	                           implements Changeable, java.io.Serializable
{

// legend anchor enums
public final static int NORTH           = 0x10;
public final static int SOUTH           = 0x20;
public final static int EAST            = 0x01;
public final static int WEST            = 0x02;
public final static int NORTHEAST       = 0x11;
public final static int NORTHWEST       = 0x12;
public final static int SOUTHEAST       = 0x21;
public final static int SOUTHWEST       = 0x22;
public final static int NORTHSOUTH_MASK = 0xf0;
public final static int EASTWEST_MASK   = 0x0f;

// legend orientation enums - values must not be changed.
public final static int  HORIZONTAL = 0;
public final static int  VERTICAL   = 1;

// draw type enums for legend symbol
public final static int  NONE           = 0;
public final static int  BOX            = 1;
public final static int  IMAGE          = 2;
public final static int  IMAGE_OUTLINED = 3;
public final static int  CUSTOM_SYMBOL  = 4;
public final static int  CUSTOM_ALL     = 5;
public final static int  BOX_PLAIN      = 6;
public final static int  LINE           = 7;

protected int anchor      = EAST;
protected int orientation = VERTICAL;

// text renderer used to draw HTML text
protected TextRenderer textRend = null;

// legend vector -- contains all items in the legend, located in subvectors
// by group (or view)
protected List legendItems = null;

protected Dimension ps = new Dimension();

// default size of symbol -- set to font ascent in recalc()
protected int symbolSize = 8;

// font ascent and descent
protected int ascent  = 0;
protected int descent = 0;

/** The image map information for the <code>JCLegend</code>. */
protected ImageMapInfo imageMapInfo = null;

// Cache data views, series, and related data used by draw() but that must be
// the same as used in recalc().
protected boolean cachedVertical = false;
protected transient Font cachedFont;

/** Used to turn off updates to the parent in certain cases. */
protected boolean updateParent = true;

/** Boolean representing the value of the <code>Changed</code> property. */
protected boolean changed = true;

/** A flag to record what has changed */
protected int changedFlag = Changeable.NO_CHANGE;

/** The preferred size of the legend. */
protected Dimension preferredSize = null;

/** <code>CellRendererPane</code> object used in rendering HTML text. */
protected CellRendererPane cellRendererPane = null;

/** <code>JCLegendPopulator</code> instance used to populate the legend.*/
protected JCLegendPopulator legendPopulator = null;

/** <code>JCLegendRenderer</code> instance used to help draw the legend. */
protected JCLegendRenderer legendRenderer = null;

/** Temp <code>Graphics</code> object used to draw without showing to screen.*/
private transient Graphics prevSetGraphics = null;

/** The locale field used for localization. */
protected Locale locale = Locale.getDefault();

/**
 * Internal constructor for legend.
 */
public JCLegend()
{
    super();
    setOpaque(false);
    setVisible(false);

    cellRendererPane = new CellRendererPane();
    add(cellRendererPane);
}

/**
 * Sets the legend populator class which will populate the legend.
 *
 * @param legPop The <code>JCLegendPopulator</code> instance used to populate
 *     this legend
 */
public void setLegendPopulator(JCLegendPopulator legPop)
{
    legendPopulator = legPop;
}

/**
 * Gets the <code>JCLegendPopulator</code> instance used to populate this
 * legend.
 *
 * @return The <code>JCLegendPopulator</code> instance used to populate this
 *     legend
 */
public JCLegendPopulator getLegendPopulator()
{
    return(legendPopulator);
}

/**
 * Sets the legend renderer class which will help draw the legend.
 *
 * @param legRend The <code>JCLegendRenderer</code> instance used to help draw
 *      this legend
 */
public void setLegendRenderer(JCLegendRenderer legRend)
{
    legendRenderer = legRend;
}

/**
 * Gets the legend renderer class which is used to help draw the legend.
 *
 * @return The <code>JCLegendRenderer</code> instance used to help draw this
 *     legend
 */
public JCLegendRenderer getLegendRenderer()
{
    return(legendRenderer);
}

/**
 * Gets the <code>Anchor</code> property, which determines the position of the
 * legend relative to the parent components primary draw area.
 *
 * @return One of <code>NORTH</code>, <code>SOUTH</code>, <code>EAST</code>,
 * <code>WEST</code>, <code>NORTHWEST</code>, <code>SOUTHWEST</code>,
 * <code>NORTHEAST</code>, or <code>SOUTHEAST</code>
 */
public int getAnchor()
{
    return(anchor);
}

/**
 * Sets the <code>Anchor</code> property, which determines the position of the
 * legend relative to the parent component's primary draw area.
 *
 * @param a The new anchor.  Must be one of <code>NORTH</code>,
 * <code>SOUTH</code>, <code>EAST</code>, <code>WEST</code>,
 * <code>NORTHWEST</code>, <code>SOUTHWEST</code>, <code>NORTHEAST</code>, or
 * <code>SOUTHEAST</code>
 */
public void setAnchor(int a)
{
    if (a == anchor) {
        return;
    }
    checkAnchor(a);

    setChanged(true, Changeable.REDRAW | Changeable.NOTIFY_PARENT_LAYOUT);
}

protected void checkAnchor(int a)
{
    switch (a) {
    case NORTH:
    case SOUTH:
    case EAST:
    case WEST:
    case NORTHEAST:
    case SOUTHEAST:
    case NORTHWEST:
    case SOUTHWEST:
        anchor = a;
        break;
    default:
        throw new IllegalArgumentException(
            "Invalid Anchor value in JCLegend.  Should be one of NORTH, " +
            "SOUTH, EAST, WEST, NORTHEAST, SOUTHEAST, NORTHWEST or SOUTHWEST"
        );
    }
}

    /**
 * Gets the <code>Orientation</code> property, which determines how the
 * legend information is laid out.
 *
 * @return One of <code>VERTICAL</code> or <code>HORIZONTAL</code>
 */
public int getOrientation()
{
    return(orientation);
}

/**
 * Sets the <code>Orientation</code> property, which determines how the
 * legend information is laid out.
 *
 * @param orient one of <code>VERTICAL</code> or <code>HORIZONTAL</code>
 */
public void setOrientation(int orient)
{
    if (orient == orientation) return;
    switch (orient) {
    case VERTICAL:
    case HORIZONTAL:
        orientation = orient;
        break;
    default:
        throw new IllegalArgumentException(
            "Invalid value for Orientation in JCLegend.  Should be one of " +
            "VERTICAL or HORIZONTAL"
        );
    }

    setChanged(true, Changeable.RECALC);
}

/**
 * Retrieve the image map information for this <code>JCLegend</code>.
 * @return The current image map information.
 */
public ImageMapInfo getImageMapInfo()
{
    return(imageMapInfo);
}

/**
 * Set the image map information for this <code>JCLegend</code>.
 * @param imageMapInfo The new image map information.
 */
public void setImageMapInfo(ImageMapInfo imageMapInfo)
{
    this.imageMapInfo = imageMapInfo;
}

/** Gets the default symbol size being used for legend item.  The default
 * value is the ascent of the font (set in <code>recalc()</code>).  It is overridable
 * by users who want to provide a different <code>symbolSize</code>.
 *
 * @return the default symbol size
 */
public int getSymbolSize()
{
    return(symbolSize);
}

/**
 * Recalculates extents of the legend.
 */
public void recalc()
{
    if (!isChanged(Changeable.RECALC) && isValid()) {
        return;
    }

    Insets insets = getInsets();
    ps = new Dimension(insets.left + insets.right, insets.top + insets.bottom);

    if (!isVisible()) {
        // no recalc, but need to reset the changed indication
        setChanged(false, Changeable.NO_CHANGE);
        return;
    }

	setUpLegendVariables();

    Dimension legendDim = layoutLegend(legendItems, cachedVertical, cachedFont);

    ps.width += legendDim.width;
    ps.height += legendDim.height;

    setChanged(false, Changeable.NO_CHANGE);
}

/**
 * Set up and store variables used by the legend, including the font,
 * font metrics and the list of legend items.
 */
protected void setUpLegendVariables()
{
    boolean vert = (orientation == VERTICAL);
    cachedVertical = vert;

    cachedFont = getFont();

    if (cachedFont == null) {
        cachedFont = new Font("dialog", Font.PLAIN, 12);
    }

    FontMetrics fm = null;
    try {
        Graphics useGraphics = getGraphics();
        if (useGraphics != null) {
            fm = useGraphics.getFontMetrics(cachedFont);
        }
        else {
            fm = getToolkit().getFontMetrics(cachedFont);
        }
		useGraphics.dispose();
    }
    catch (Exception e) {}

    if (fm == null) {
        return;
    }
    descent = fm.getDescent();
    ascent = fm.getAscent();

    symbolSize = ascent;

    if (legendPopulator != null) {
        legendItems = legendPopulator.getLegendItems(fm);
    }
}

/**
 * Returns the list of legend items being used by this legend.
 * @return The list of legend items.  It contains a vector for each group (data
 *     view).  Each group vector contains a number of <code>JCLegendItem</code>
 *     instances, each reflecting one item within the legend group.  The first
 *     <code>JCLegendItem</code> within each group is the title of the group.
 */
public List getLegendItems()
{
    return(legendItems);
}

/**
 * To be overridden by a legend implementation.  This method is used to layout
 * the items within the legend by filling in the <code>dim</code>,
 * <code>pos</code>, <code>textPos</code>, and <code>symPos</code> members of
 * each <code>JCLegendItem</code> insance, returning the size of the layed out
 * legend.
 *
* @param itemList The legend vector.  It contains a vector for each group
 *     (data view).  Each group vector contains a number of
 *     <code>JCLegendItem</code> instances, each reflecting one item within the
 *     legend group.  The first <code>JCLegendItem</code> within each group is
 *     the title of the group.
 * @param vertical <code>true</code> for vertical legend, <code>false</code>
 *     for horizontal legend
 * @param useFont font used for legend
 * @return The size of the legend
 */
public abstract Dimension layoutLegend(List itemList, boolean vertical,
                                       Font useFont);

/**
 * Paint method for the legend.
 *
 * @param gc graphics context to use for drawing
 */
public void paintComponent(Graphics gc)
{
    if (!isVisible() || gc == null) {
        return;
    }

    Container parent = getParent();
    if (parent == null) {
        return;
    }

    // Draw own opaque backgrounds.  Shouldn't this be
    // done by the superclass?  That code appears to be in the UI, which
    // is null for this component it seems.

    if (isOpaque()) {
        if (!gc.getColor().equals(getBackground())) {
            gc.setColor(getBackground());
        }
        Rectangle r = getBounds();
        gc.fillRect(0, 0, r.width, r.height);
    }

    boolean drawLegend = false;
    Rectangle rect = getDrawingArea();

    if (!drawLegend) {
        drawLegend = true;
        gc = gc.create();
        gc.translate(rect.x, rect.y);
        gc.clipRect(0, 0, rect.width + 1, rect.height + 1);
    }

    // If we have a legend renderer, draw each legend item
    if (legendRenderer != null) {
        for (int i = 0; i < legendItems.size(); i++) {
            List viewItems = (List)legendItems.get(i);
            for (int j = 0; j < viewItems.size(); j++) {
                JCLegendItem thisItem = (JCLegendItem)viewItems.get(j);

                // If entire legend item has custom drawing, call
                // legend renderer to do it
                if (thisItem.drawType == CUSTOM_ALL) {
                    legendRenderer.drawLegendItem(gc, cachedFont, thisItem);
                }
                // Otherwise, we draw by our default rules
                else {
                    drawLegendItem(gc, cachedFont, thisItem);
                }
            }
        }
    }

    if (drawLegend) {
        gc.dispose();
    }

    setChanged(false, Changeable.NO_CHANGE);
}

/** Draws one legend item based on the size and position given in its
 * <code>JCLegendItem</code> instance.  The symbol to be drawn is derived from
 * the legend item's draw type.  A user that wants to provide different
 * drawing behaviour should use either the <code>CUSTOM_SYMBOL</code> or
 * <code>CUSTOM_ALL</code> draw types.
 *
 * @param gc the graphics context
 * @param useFont the font used for this legend.
 * @param thisItem the legend item to draw; a <code>JCLegendItem</code> instance
 */
public void drawLegendItem(Graphics gc, Font useFont, JCLegendItem thisItem)
{
    Color outlineColor = null;
    if (gc.getColor() != getForeground()) {
        gc.setColor(getForeground());
    }
    if (gc.getFont() != useFont) {
        gc.setFont(useFont);
    }
    if (outlineColor == null && legendRenderer != null) {
        outlineColor = legendRenderer.getOutlineColor(thisItem);
    }
    if (outlineColor == null) {
        outlineColor = getForeground();
    }

    // Special case : draw title!
    if (isTitleItem(thisItem)) {
        if (thisItem.contents instanceof String) {
            String contents = (String)thisItem.contents;
            if (TextRenderer.isHTML(contents)) {
                if (textRend == null) {
                    textRend = new TextRenderer();
                }
                textRend.drawHTMLText(gc, getCellRendererPane(), this, contents,
                                      thisItem.pos.x + thisItem.textPos.x,
                                      thisItem.pos.y + thisItem.textPos.y,
                                      thisItem.textDim.width,
                                      thisItem.textDim.height);
            }
            else {
                // Don't draw empty strings -- they cause problems in printing
                if (contents.length() > 0) {
                    gc.drawString(contents,
                                  thisItem.pos.x + thisItem.textPos.x,
                                  thisItem.pos.y + thisItem.textPos.y + ascent);
                }
            }
        }
    }
    // Draw symbol/text pairs
    else {
        // Draw symbol
        int symX = thisItem.pos.x + thisItem.symbolPos.x;
        int symY = thisItem.pos.y + thisItem.symbolPos.y;

        // Draw symbols as per draw type
        if (thisItem.drawType == BOX || thisItem.drawType == BOX_PLAIN) {
            if (legendRenderer != null) {
                legendRenderer.setFillGraphics(gc, thisItem);
            }
            gc.fillRect(symX, symY, thisItem.symbolDim.width - 1,
                        thisItem.symbolDim.height - 1);
            if (thisItem.drawType == BOX) {
                gc.setColor(outlineColor);
            }
            gc.drawRect(symX, symY, thisItem.symbolDim.width - 1,
                        thisItem.symbolDim.height - 1);
        }
        else if (thisItem.drawType == LINE) {
            if (legendRenderer != null) {
                legendRenderer.setFillGraphics(gc, thisItem);
            }
            gc.drawLine(symX, symY + ((thisItem.symbolDim.height - 1) / 2),
                        symX + thisItem.symbolDim.width - 1,
                        symY + ((thisItem.symbolDim.height - 1) / 2));
        }
        else if (thisItem.drawType == IMAGE) {
            // If image fill, draw image.  Otherwise, draw swatch below.
            if (thisItem.symbol != null &&
                thisItem.symbol instanceof Image)
            {
                gc.drawImage((Image)thisItem.symbol, symX, symY, null);
            }
        }
        else if (thisItem.drawType == IMAGE_OUTLINED) {
            if (thisItem.symbol != null &&
                thisItem.symbol instanceof Image)
            {
                gc.drawImage((Image)thisItem.symbol, symX, symY, null);
                gc.setColor(outlineColor);
                gc.drawRect(symX, symY, thisItem.symbolDim.width - 1,
                            thisItem.symbolDim.height - 1);
            }
        }
        // Custom symbol -- call legend renderer to draw it
        else if (thisItem.drawType == CUSTOM_SYMBOL) {
            if (legendRenderer != null) {
                legendRenderer.drawLegendItemSymbol(gc, useFont, thisItem);
            }
        }

        if (gc.getColor() != getForeground()) {
            gc.setColor(getForeground());
        }
        if (gc.getFont() != useFont) {
            gc.setFont(useFont);
        }

        // Draw text
        if (thisItem.contents instanceof String) {
            String contents = (String)thisItem.contents;
            if (TextRenderer.isHTML(contents)) {
                if (textRend == null) {
                    textRend = new TextRenderer();
                }
                textRend.drawHTMLText(gc, getCellRendererPane(), this, contents,
                                      thisItem.pos.x + thisItem.textPos.x,
                                      thisItem.pos.y + thisItem.textPos.y,
                                      thisItem.textDim.width,
                                      thisItem.textDim.height);
            }
            else {
                // Don't draw empty strings -- they cause problems in printing
                if (contents.length() > 0) {
                    gc.drawString(contents,
                                  thisItem.pos.x + thisItem.textPos.x,
                                  thisItem.pos.y + thisItem.textPos.y + ascent);
                }
            }
        }
        else {
            if (legendRenderer != null) {
                legendRenderer.drawLegendItemText(gc, useFont, thisItem);
            }
        }
    }
}

/**
 * Defers the decision on whether an item is a title item or not to
 * the populator class.  It knows best.
 *
 * @param item the <code>JCLegendItem</code> in question
 * @return <code>true</code> if the item is a title in the legend;
 * <code>false</code> otherwise
 */
public boolean isTitleItem(JCLegendItem item)
{
    if (legendPopulator != null) {
        return(legendPopulator.isTitleItem(item));
    }
    return(false);
}

/**
 * Calculates and returns the preferred size for the legend, or
 * the manually set preferred size if it has been manually set.
 *
 * @return preferred size
 */
public Dimension getPreferredSize()
{
    Dimension prefSize = preferredSize;

    // if preferred size manually set, return it
    if (prefSize != null) {
        return(prefSize);
    }

    // if not manually set, calc it
    recalc();
    return(new Dimension(ps));
}

/**
 * Sets the preferred size manually.
 *
 * @param dim the preferred size
 */
public void setPreferredSize(Dimension dim)
{
    preferredSize = dim;
}

/**
 * Gets the <code>TextRenderer</code> class used to draw HTML text.
 * @return
 */
public TextRenderer getTextRenderer()
{
    if (textRend == null) {
        textRend = new TextRenderer();
    }
    return(textRend);
}

/**
 * Internal pick method.
 * Chooses the legend item closest to the provided position.
 *
 * @param p pick point relative to legend
 * @return the <code>JCLegendItem</code> instance representing the picked point.
 */
public JCLegendItem pick(Point p)
{
    Rectangle r = getDrawingArea();
    p.x -= r.x;
    p.y -= r.y;

    JCLegendItem pickItem = null;
    for (int i = 0; i < legendItems.size(); i++) {
        List viewItems = (List)legendItems.get(i);
        for (int j = 0; j < viewItems.size(); j++) {
            JCLegendItem thisItem = (JCLegendItem)viewItems.get(j);
            Rectangle pickRect = thisItem.pickRectangle;
            if (pickRect == null     &&
                thisItem.pos != null &&
                thisItem.dim != null)
            {
                pickRect = new Rectangle(thisItem.pos, thisItem.dim);
            }

            if (pickRect != null) {
                if (pickRect.contains(p.x, p.y)) {
                    pickItem = thisItem;
                    break;
                }
            }
        }
    }
    return(pickItem);
}

/** Gets the bounding rectangle of the component's drawing area,
 * (its area minus the shadows and insets).
 *
 * @see javax.swing.JComponent#setBorder
 * @return
 */
public Rectangle getDrawingArea()
{
    Rectangle rect = new Rectangle();
    getDrawingArea(rect);
    return(rect);
}

/** Gets the bounding rectangle of the component's drawing area,
 * (its area minus the insets).
 *
 * @param rect
 * @see javax.swing.JComponent#setBorder
 */
public void getDrawingArea(Rectangle rect)
{
    Insets insets = getInsets();
    Dimension sz = getSize();
    rect.setBounds(insets.left,
                   insets.top,
                   Math.max(0, sz.width - (insets.left + insets.right)),
                   Math.max(0, sz.height - (insets.top + insets.bottom)));
}

/**
 * Gets the cell renderer pane used by this component.  Used to render HTML
 * encoded text with a <code>JLabel</code>.
 * @return
 */
public CellRendererPane getCellRendererPane()
{
    return(cellRendererPane);
}

/**
 * Overrides parent <code>getGraphics()</code>.  If another
 * <code>Graphics</code> object has been set to be the current one, return that.
 * Otherwise, call the superclass' <code>getGraphics()</code> method.
 * @return the <code>Graphics</code> object to use to draw this legend
 */
public Graphics getGraphics()
{
    if (prevSetGraphics == null) {
        return(super.getGraphics());
    }
    // make sure to return a copy of the user-set graphics object
    // this is in keeping with the contract for getGraphics()
    // some Java code calls dispose() on the object returned from this method
    return(prevSetGraphics.create());
}

/**
 * Provides a way for the legend to draw to a custom <code>Graphics</code>
 * object. Call with a null to specify use of the default <code>Graphics</code>
 * object used to draw to the screen.
 * @param g <code>Graphics</code> object to use for drawing
 */
public void setGraphics(Graphics g)
{
    prevSetGraphics = g;
}

// Changeable implementation begins here

/**
 * Gets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property determines whether the object needs
 * to be recalculated.
 *
 * @return <code>true</code> if the object has been changed and needs
 * recalculation; <code>false</code> otherwise
 */
public boolean getChanged()
{
//  Container parent = getParent();
//  if (parent == null) {
//      return(changed);
//  }
    return(changed);
}


/**
 * Checks the value of the <code>ChangedFlag</code> property for the specified
 * mask (see <code>Changeable</code> for a list of valid masks).
 * The <code>ChangedFlag</code> property determines the action the object needs
 * to take in regards to the current change.
 *
 * @param mask
 * @return <code>true</code> if the specified mask exists in the
 * <code>ChangedFlag</code> property; <code>false</code> otherwise
 */
public boolean getChanged(int mask)
{
    return((changedFlag & mask) == mask);
}

/**
 * Gets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property determines whether the object needs
 * to be recalculated.
 *
 * @return <code>true</code> if the object has been changed and needs
 * recalculation; <code>false</code> otherwise
 */
final public boolean isChanged()
{
    return(getChanged());
}

/**
 * Checks the value of the <code>ChangedFlag</code> property for the specified
 * mask (see <code>Changeable</code> for a list of valid masks).
 * The <code>ChangedFlag</code> property determines the action the object needs
 * to take in regards to the current change.
 *
 * @param mask
 * @return <code>true</code> if the specified mask exists in the
 * <code>ChangedFlag</code> property, <code>false</code> otherwise
 */
final public boolean isChanged(int mask)
{
    return(getChanged(mask));
}

/**
 * Sets the value of the <code>Changed</code> property and updates the
 * parent object.
 * The <code>Changed</code> property determines whether the object needs
 * to be recalculated.
 *
 * @param c
 * @param cf
 */
public void setChanged(boolean c, int cf)
{
    if (c != getChanged() || !getChanged(cf)) {
        changed = c;
        if (c) {
            changedFlag |= cf;
        }
        else {
            changedFlag = cf;
        }
    }
    if (c && getParent() != null && updateParent) {
        updateParent(cf);
    }
}

/**
 * Sets the value of the <code>Changed</code> property.
 * The <code>Changed</code> property determines whether the object needs
 * to be recalculated.
 *
 * @param n new value for <code>Changed</code>
 * @param cf
 * @param updateParent if <code>true</code>, the parent is updated; otherwise,
 * the parent is not updated (the "update" semantics are up to the implementor)
 */
public void setChanged(boolean n, int cf, boolean updateParent)
{

    if (updateParent) {
        setChanged(n, cf);
    }
    else {
        changed = n;
        if (n) {
            changedFlag |= cf;
        }
        else {
            changedFlag = cf;
        }
    }
}

/**  *
 * @param cf
 */
public void updateParent(int cf)
{
    Container p = getParent();
    if (p != null) {
        Changeable ch = null;
        if (p instanceof Changeable) {
            ch = (Changeable)p;
        }

        if (((cf & Changeable.RECALC) > 0) ||
            ((cf & Changeable.NOTIFY_PARENT_LAYOUT) > 0))
        {
            p.invalidate();
            if ((ch != null && !ch.isBatched()) || ch == null) {
                p.validate();
                // This repaint may be redundant in some cases.
                p.repaint();
            }
        }
        else if ((cf & Changeable.REDRAW) > 0) {
            if ((ch != null && !ch.isBatched()) || ch == null) {
                p.repaint();
            }
        }
    }
}

/**
 * This method is not used by this implementation of <code>Changeable</code>, so
 * it always returns <code>false</code>.
 * @return
 */
public boolean isBatched()
{
    return(false);
}

/**
 * Sets the locale for legend.
 * @param locale
 */
public void setLocale(Locale locale)
{
    if (this.locale == locale || locale == null) {
        return;
    }

    try {
        this.locale = locale;
    }
    catch (MissingResourceException e) {
        this.locale = Locale.getDefault();
    }
}

/**
 * Overrides write object to change the state of the changed flags being
 * saved so that the component will actually recalc and draw when re-serialized.
 *
 * @exception IOException
 * @param out
 */
private void writeObject(java.io.ObjectOutputStream out)
    throws IOException
{
    // save current state first
    boolean oldChanged = changed;
    int oldChangedFlag = changedFlag;

    // save state that will generate a recalc
    changed = true;
    changedFlag = Changeable.RECALC;
    out.defaultWriteObject();

    // restore current state
    changed = oldChanged;
    changedFlag = oldChangedFlag;
}

}
