/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCLegendBundle.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.legend;

import java.util.Locale;
import java.util.Hashtable;
import java.util.ResourceBundle;
import java.text.MessageFormat;
import java.text.ChoiceFormat;
import java.util.MissingResourceException;

import com.klg.jclass.util.ServerLocaleBundleBase;

/**
 * JCLegendBundle contains locale-specific objects and helpers to get this
 * information. When a program needs a locale-specific string it can load
 * it from the resource bundle that is appropriate for the specified
 * locale. Those are located in the com.klg.jclass.util.legend.resources
 * directory.  In this way, it can be written program code that is largely
 * independent of the user's locale isolating most of the locale-specific
 * information in resource bundles.
 */
public class JCLegendBundle extends ServerLocaleBundleBase {

private static final String BUNDLE_PATH =
    "com.klg.jclass.util.legend.resources.LocaleInfo";

// Locale bundle cache for this bundle path, keyed by Locale.
// This hashtable is itself cached in the cacheCache Hashtable.
private static Hashtable bundleCache = getCache(BUNDLE_PATH);

/**
 * Static function that localizes a String based on its key,
 * language, and country.
 * @param language desired language
 * @param country desired country
 * @param key the key of the String that is to be retrieved
 * @return the localized string
 */
public static String string( String language, String country, String key )
{
    return(getBundle(bundleCache, language, country, BUNDLE_PATH).
                                                            getString(key));
}

/**
 * Static functiont that localized a String based on its key and locale.
 * @param key the key of the String that is to be retrieved.
 * @param locale the desired locale
 * @return the localized string
 */
public static String string(String key, Locale locale)
{
    return(getBundle(bundleCache, locale, BUNDLE_PATH).getString(key));
}

/**
 * Static function used to retrive a String based on its key for the
 * default locale.
 * @param key the key of the String that is to be retrieved
 * @return the localized string
 */
public static String string(String key)
{
    return(getBundle(bundleCache, Locale.getDefault(), BUNDLE_PATH).
                                                        getString(key));
}

/**
 * Static function used to format a String based on its key for the
 * default locale.
 * It produces concatenated messages in language-neutral way.
 * @param key the key of the String that is to be retrived
 * @param args set of objects used in formatting
 * @return the formatted string
 */
public static String string(String key, Object[] args)
{
	return string(key, args, Locale.getDefault());
}

/**
 * Static function used to format a String based on its key for the
 * default locale.
 * It produces concatenated messages in language-neutral way.
 * @param key the key of the String that is to be retrived
 * @param args set of objects used in formatting
 * @param locale the desired locale
 * @return the formatted string
 */
public static String string(String key, Object[] args, Locale locale)
{
    return(MessageFormat.format(
    	getBundle(bundleCache, locale, BUNDLE_PATH).getString(key),
		args));
}

/**
 * Static function used to format a String based on its key for the
 * default locale.
 * It produces concatenated messages in language-neutral way.
 * @param key the key of the String that is to be retrived
 * @param formatter a formatter provided by the user
 * @param args set of objects used in formatting
 * @return the formatted string
 */
public static String string(String key, MessageFormat formatter, Object args[])
{
	return string(key, formatter, args, Locale.getDefault());
}

/**
 * Static function used to format a String based on its key for the
 * default locale.
 * It produces concatenated messages in language-neutral way.
 * @param key the key of the String that is to be retrived
 * @param formatter a formatter provided by the user
 * @param args set of objects used in formatting
 * @param locale the desired locale
 * @return the formatted string
 */
public static String string(String key, MessageFormat formatter, 
										Object args[], Locale locale)
{
    return(formatter.format(
    	getBundle(bundleCache, locale, BUNDLE_PATH).getString(key),
		args));
}

/**
 * Static function used to format a String based on its key for the
 * default locale.
 * It produces concatenated messages in a language-neutral way.
 * Allows to attach a format to a range of numbers. It is generally used for
 * handling plurals. The choice is specified with an ascending list of doubles,
 * where each item specifies a half-open interval up to the next item:
 * <i>X</i> matches <i>j</i> if and only if <i>limit[j] <= X < limit[j+1]</i>.
 * @param key the key of the String that is to be retrived
 * @param args set of objects used in formatting
 * @param limits the list of doubles that specifies a half-open interval
 * @param formats it is an array of choices: one choice for each interval
 * defined by the limits array
 * @param choiceVar it is used to identify which of the variable from the
 * String that contains the formatting information uses the choice formatting
 * feature (note the limitation of only one choice variable per String)
 * @return the formatted string
 */
public static String string(String key, Object[] args, double[] limits,
                            String[] formats, int choiceVar)
{
	return string(key, args, limits, formats, choiceVar, Locale.getDefault());
}

/**
 * Static function used to format a String based on its key for the
 * default locale.
 * It produces concatenated messages in a language-neutral way.
 * Allows to attach a format to a range of numbers. It is generally used for
 * handling plurals. The choice is specified with an ascending list of doubles,
 * where each item specifies a half-open interval up to the next item:
 * <i>X</i> matches <i>j</i> if and only if <i>limit[j] <= X < limit[j+1]</i>.
 * @param key the key of the String that is to be retrived
 * @param args set of objects used in formatting
 * @param limits the list of doubles that specifies a half-open interval
 * @param formats it is an array of choices: one choice for each interval
 * defined by the limits array
 * @param choiceVar it is used to identify which of the variable from the
 * String that contains the formatting information uses the choice formatting
 * feature (note the limitation of only one choice variable per String)
 * @param locale the desired locale
 * @return the formatted string
 */
public static String string(String key, Object[] args, double[] limits,
                            String[] formats, int choiceVar, Locale locale)
{
    String s = null;
    MessageFormat formatter = 
		new MessageFormat(
    		getBundle(bundleCache, locale, BUNDLE_PATH).getString(key));
    ChoiceFormat choicer = new ChoiceFormat(limits, formats);

    formatter.setFormat(choiceVar, choicer);
    s = formatter.format(args);

    return(s);
}


// Key constants

public static final String key1   = "com.klg.jclass.util.legend.Key1";
public static final String key2   = "com.klg.jclass.util.legend.Key2";
public static final String key3   = "com.klg.jclass.util.legend.Key3";
public static final String key4   = "com.klg.jclass.util.legend.Key4";
public static final String key5   = "com.klg.jclass.util.legend.Key5";
public static final String key6   = "com.klg.jclass.util.legend.Key6";
public static final String key7   = "com.klg.jclass.util.legend.Key7";
public static final String key8   = "com.klg.jclass.util.legend.Key8";
public static final String key9   = "com.klg.jclass.util.legend.Key9";
public static final String key10  = "com.klg.jclass.util.legend.Key10";

}
