/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCMultiColLegend.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:42 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.legend;

import com.klg.jclass.util.Changeable;

import java.awt.*;
import java.util.List;
import java.util.Vector;

/**
 * <code>JCMultiColLegend</code> manages and renders a grid-based legend that
 * can have multiple columns. It extends <code>JCGridLegend</code>, rearranging
 * the position of legend items within the legend vector and relying on
 * <code>JCGridLegend</code>'s functionality to position and draw legend items.
 * <code>JCMultiColLegend</code> assumes that the legend vector contains a
 * number of group vectors, with the title of each group being the first item
 * in each group vector.
 */
public class JCMultiColLegend extends JCGridLegend {

protected int numRows    = -1;
protected int numColumns = -1;

protected int actualNumRows    = -1;
protected int actualNumColumns = -1;

public JCMultiColLegend()
{
	super();
}

/**
 * Sets the number of rows in this legend.  If less than or equal to 0
 * (the default), the number of rows will be set automatically.  Setting 1 row
 * in a vertical column will cause the title item (if existing) to be
 * included in the same row as the other legend items.
 * @param rows the number of rows to use
 */
public void setNumRows(int rows)
{
	numRows = rows;
	actualNumRows = rows;

	setChanged(true, Changeable.RECALC);
}

/**
 * Gets the number of rows set for this legend.
 * If less than or equal to 0 (the default), the
 * number of rows is set automatically.
 * @return  the value set to be the number of rows in the legend
 */
public int getNumRows()
{
	return(numRows);
}

/**
 * Sets the number of columns in this legend.  If less than or equal to 0
 * (the default), the number is set automatically.  Setting 1 column
 * in a horizontal column will cause the title item (if existing) to be
 * included in the same column as all of the other legend items.
 * @param columns the number of columns to use
 */
public void setNumColumns(int columns)
{
	numColumns = columns;
	actualNumColumns = columns;

	setChanged(true, Changeable.RECALC);
}

/**
 * Gets the number of columns set for this legend.  If  less than
 * or equal to zero (the default), the number of
 * columns is set automatically.
 * @return the value set to be the number of columns in the legend
 */
public int getNumColumns()
{
	return(numColumns);
}

/**
 * Gets the actual number of rows used in this legend.
 * @return the actual number of rows used in this legend
 */
public int getActualNumRows()
{
	// If set to default, need to recalc actual number of rows and columns
	if (actualNumRows < 0) {
		calcActualNumRowsAndColumns();
	}
	return(actualNumRows);
}

/**
 * Gets the actual number of columns used in this legend.
 * @return the actual number of columns used in this legend
 */
public int getActualNumColumns()
{
	// If set to default, need to recalc actual number of rows and columns
	if (actualNumColumns < 0) {
		calcActualNumRowsAndColumns();
	}
	return(actualNumColumns);
}

/**
 * <code>layoutLegend</code> overrides <code>JCGridLegend</code>'s
 * <code>layoutLegend()</code>. <code>layoutLegend</code> manipulates the legend
 * vector so that the items are in the order you specify.  It calls
 * <code>JCGridLegend</code>'s <code>layoutLegend()</code> to do the layout.
 * @param itemList the legend vector of groups and legend items
 * @param vertical <code>true</code> for a vertical legend; <code>false</code>
 * for a horizontal legend
 * @param useFont the font for the legend
 * @return size of legend returned from <code>JCGridLegend</code>'s
 * <code>layoutLegend()</code>
 */
public Dimension layoutLegend(List itemList, boolean vertical, Font useFont)
{
    if (itemList.size() == 0) {
        return(super.layoutLegend(itemList, vertical, useFont));
    }

    // The legend (and the itemList) is organized as a list of 'groups'
    // with a number of items inside each 'group'.  If vertical, groups
    // are columns; if horizontal, groups are rows.
    int numGroups = -1;
    int numLinesPerGroup = -1;
    if (vertical) {
        numGroups = numColumns;
        numLinesPerGroup = numRows;
    }
    else {
        numGroups = numRows;
        numLinesPerGroup = numColumns;
    }

   // Are there any actual titles in this legend ?
    boolean hasTitle = false;
    boolean hasItems = false;
    for (int i = 0; i < itemList.size(); i++) {
        List groupList = (List)itemList.get(i);
        JCLegendItem possibleTitle = null;
        if (groupList.size() > 0) {
            possibleTitle = (JCLegendItem)groupList.get(0);
            hasItems = true;
        }
        if (possibleTitle != null             &&
            isTitleItem(possibleTitle)        &&
            possibleTitle.textDim.height != 0 &&
            possibleTitle.textDim.width != 0)
        {
            hasTitle = true;
            break;
        }
    }

    // No items in vectors! return.
    if (!hasItems) {
        return(super.layoutLegend(itemList, vertical, useFont));
    }

    // If no actual titles, remove empty title blocks from group vectors
    if (!hasTitle) {
        for (int i = 0; i < itemList.size(); i++) {
            List groupList = (List)itemList.get(i);
            if (groupList.size() > 0) {
                groupList.remove(0);
            }
        }
    }

    // If there are titles, but numLinesPerGroup is 1, treat the title
    // line as just another member of the one line specified.
    if (hasTitle && numLinesPerGroup == 1) {
        hasTitle = false;
    }

    // Number of groups has been specified
    if (numGroups > 0) {
        // Expand the list of groups, moving existing items into the new groups
        if (numGroups > itemList.size()) {
            int subgroupsPer[] = new int[itemList.size()];
            int totalItems = 0;
            int totalGroups = 0;
            for (int i = 0; i < subgroupsPer.length; i++) {
                List groupList = (List)itemList.get(i);
                totalItems += (groupList.size() - (hasTitle ? 1 : 0));
            }
            for (int i = 0; i < subgroupsPer.length; i++) {
                List groupList = (List)itemList.get(i);
                int numItemsInGroup = groupList.size();
                if (hasTitle) {
                    numItemsInGroup--;
                }

                // How many subgroups will be split this group into ?
                // it depends on it's perecentage of number of items.
                subgroupsPer[i] = Math.round((numItemsInGroup
                                              * numGroups) / totalItems);
                if (subgroupsPer[i] == 0) {
                    subgroupsPer[i] = 1;
                }
                else if (subgroupsPer[i] + totalGroups > numGroups) {
                    subgroupsPer[i] = numGroups - totalGroups;
                }

                totalGroups += subgroupsPer[i];
                totalItems -= numItemsInGroup;
            }

            int m = 0;
            for (int i = 0; i < itemList.size(); i++) {
                List groupList = (List)itemList.get(i);
                int numItemsInGroup = groupList.size();
                if (hasTitle) {
                    numItemsInGroup--;
                }

                int numLinesPerThisGroup =
                    (int)(Math.ceil((float)numItemsInGroup /
                                    Math.max(subgroupsPer[m], 1))) + 1;
                int j = numLinesPerThisGroup;
                while (j < groupList.size()) {
                    List newGroupList = new Vector();
                    // Create blank place holder title for first entry
                    // of new subgroup
                    if (hasTitle) {
                        JCLegendItem blankTitle = new JCLegendItem();
                        blankTitle.contents = new String("");
                        blankTitle.textDim = new Dimension(0, 0);
                        newGroupList.add(blankTitle);
                    }
                    for (int k = (hasTitle ? 1 : 0);
                         k < numLinesPerThisGroup && j < groupList.size();
                         k++)
                    {
                        JCLegendItem item = (JCLegendItem)groupList.get(j);
                        groupList.remove(j);
                        newGroupList.add(item);
                    }
                    itemList.add(i + 1, newGroupList);
                    i++;
                }
                m++;
            }
        }
        // Shrink the list of groups, moving existing items into other groups
        else if (numGroups < itemList.size()) {
            int toContract = itemList.size() - numGroups;

            // Special case -- move everything to one column
            if (numGroups == 1) {
                while (itemList.size() > 1) {
                    contractItemList((List)itemList.get(0),
                                     (List)itemList.get(1), itemList, 1);
                }
            }
            else {
                // Take the biggest n columns (where n = numGroups)
                List activeGroups[] = new List[numGroups];
                for (int i = 0; i < numGroups; i++) {
                    activeGroups[i] = null;
                }
                for (int k = 0; k < numGroups; k++) {
                    int biggest = -1;
                    int biggestIndex = -1;
                    for (int i = 0; i < itemList.size(); i++) {
                        List groupList = (List)itemList.get(i);
                        if (groupList.size() > biggest) {
                            boolean fallOut = false;
                            for (int j = 0; j < k; j++) {
                                // Group already in active list?  move on
                                if (groupList == activeGroups[j]) {
                                    fallOut = true;
                                    break;
                                }
                            }
                            if (!fallOut) {
                                biggest = groupList.size();
                                biggestIndex = i;
                            }
                        }
                    }
                    activeGroups[k] = (List)itemList.get(biggestIndex);
                }

                // Now start adding the other columns to the active
                // columns, starting with the biggest and moving to
                // the smallest
                for (int k = 0; k < toContract; k++) {
                    // Find the next biggest group
                    int nextBiggest = -1;
                    int nextBiggestIndex = -1;
                    for (int i = 0; i < itemList.size(); i++) {
                        List groupList = (List)itemList.get(i);
                        if (groupList.size() > nextBiggest) {
                            boolean fallOut = false;
                            for (int j = 0; j < numGroups; j++) {
                                // Group in active list?  move on
                                if (groupList == activeGroups[j]) {
                                    fallOut = true;
                                    break;
                                }
                            }
                            if (!fallOut) {
                                nextBiggest = groupList.size();
                                nextBiggestIndex = i;
                            }
                        }
                    }

                    // Find the smallest active group
                    List smallestActiveGroup = null;
                    int smallestActiveSize = Integer.MAX_VALUE;
                    for (int j = 0; j < numGroups; j++) {
                        if (activeGroups[j].size() < smallestActiveSize) {
                            smallestActiveSize = activeGroups[j].size();
                            smallestActiveGroup = activeGroups[j];
                        }
                    }

                    // Add the next biggest group to the
                    // smallest active group -- attempt at balancing
                    contractItemList(smallestActiveGroup,
                                     (List)itemList.get(nextBiggestIndex),
                                     itemList, nextBiggestIndex);

                }
            }
        }
    }

    // Number of items per has been specified
    if (numLinesPerGroup > 0) {
        // Number of groups hasn't been specified -- calculate it and move
        // the items around
        if (numGroups <= 0) {
            for (int i = 0; i < itemList.size(); i++) {
                List groupList = (List)itemList.get(i);
                int j = numLinesPerGroup;
                while (j < groupList.size()) {
                    List newGroupList = new Vector();
                    // Should handle case with no titles
                    if (hasTitle) {
                        JCLegendItem blankTitle = new JCLegendItem();
                        blankTitle.contents = new String("");
                        blankTitle.textDim = new Dimension(0, 0);
                        newGroupList.add(blankTitle);
                    }
                    for (int k = (hasTitle ? 1 : 0);
                         k < numLinesPerGroup && j < groupList.size();
                         k++)
                    {
                        JCLegendItem item = (JCLegendItem)groupList.get(j);
                        groupList.remove(j);
                        newGroupList.add(item);
                    }
                    itemList.add(i + 1, newGroupList);
                    i++;
                }
            }
        }
        // Number of groups has been specified.  Just remove items after
        // the specified number of items per group.
        else {
            for (int i = 0; i < itemList.size(); i++) {
                List groupList = (List)itemList.get(i);
                int j = numLinesPerGroup;
                while (j < groupList.size()) {
                    groupList.remove(j);
                }
            }
        }
	}

	// Make sure the actual number of rows and columns is set
	calcActualNumRowsAndColumns(cachedVertical, numGroups, numLinesPerGroup);

    return(super.layoutLegend(itemList, vertical, useFont));
}

/**
 * Calculate the actual number of rows and columns from either the set
 * row and column variables or, if they've been set to default, from the
 * number of legend items associated with this legend.
 */
protected void calcActualNumRowsAndColumns()
{
	setUpLegendVariables();

    int numGroups = -1;
    int numLinesPerGroup = -1;
    if (cachedVertical) {
        numGroups = numColumns;
        numLinesPerGroup = numRows;
    }
    else {
        numGroups = numRows;
        numLinesPerGroup = numColumns;
    }

	calcActualNumRowsAndColumns(cachedVertical, numGroups, numLinesPerGroup);
}

/**
 * Calculate the actual number of rows and columns from either the passed
 * number of groups and lines per group or, if they're default values,
 * from the number of legend items associated with this legend.
 */
protected void calcActualNumRowsAndColumns(boolean vertical, int numGroups,
                                           int numLinesPerGroup)
{
	List itemList = getLegendItems();

	// if negative, figure out what the defaults being used are
	if (numGroups <= 0) {
        numGroups = itemList.size();
	}
	if (numLinesPerGroup <= 0) {
        for (int i = 0; i < itemList.size(); i++) {
            List groupList = (List)itemList.get(i);
            numLinesPerGroup = Math.max(numLinesPerGroup, groupList.size());
        }
	}

	// Set actual number of columns
    if (cachedVertical) {
        actualNumColumns = numGroups;
        actualNumRows = numLinesPerGroup;
    }
    else {
        actualNumRows = numGroups;
        actualNumColumns = numLinesPerGroup;
    }
}


/**
 * Removes a group from the legend vector, moving its contents to another group.
 * @param destList the destination group vector
 * @param sourceList the source group vector
 * @param itemList the legend vector
 * @param sourceIndex the index of the <code>sourceList</code> group within
 * the legend vector
 */
public void contractItemList(List destList, List sourceList,
                             List itemList, int sourceIndex)
{
    for (int j = 0; j < sourceList.size(); j++) {
        destList.add(sourceList.get(j));
    }
    itemList.remove(sourceIndex);
}

}
