/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DefaultRowComparator.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;



import java.util.Date;
import java.math.BigInteger;

/**
 * This class is to be used with <code>JCRowSortModel</code>.  It sorts rows by the
 * list of columns specified. By default it only sorts on the first column
 */
public class DefaultRowComparator implements JCRowComparator,
											 java.io.Serializable
{

/** Valid value for SortOrder */
static public final int FORWARD = 1;
/** Valid value for SortOrder */
static public final int REVERSE = -1;

protected int order = FORWARD;
protected int keyColumns[] = { 0 };

/**
 * Creates a <code>JCRowComparator</code> that sorts columns by the natural
 * order of the values in the first column.
 * @param order The sort order (FORWARD or REVERSE)
 */
public DefaultRowComparator(final int order)
{
	setSortOrder(order);
}

/**
 * Creates a <code>JCRowComparator</code> that sorts columns according with
 * the list of columns specified.
 * @param order The sort order (FORWARD or REVERSE)
 * @param keyColumns A list of columns by which to sort
 */
public DefaultRowComparator(final int order, int keyColumns[])
{
	setSortOrder(order);
	setKeyColumns(keyColumns);
}

/**
 * Sets a list of columns by which to sort.
 * @param keyColumns The new list of columns by which to sort
 */
public void setKeyColumns(int keyColumns[])
{
	if (keyColumns == null || keyColumns.length == 0) {
		throw new IllegalArgumentException("KeyColumns must by non-null and non-empty");
	}
	this.keyColumns = keyColumns;
}

/**
 * Gets a list of columns by which to sort.
 * @return The list of columns by which to sort
 */
public int[] getKeyColumns()
{
	return(keyColumns);
}

/**
 * Sets a sorting order.
 * @param order The new sort order (FORWARD or REVERSE)
 */
public void setSortOrder(final int order)
{
	if (order != FORWARD && order != REVERSE) {
		throw new IllegalArgumentException("SortOrder must be one of FORWARD or REVERSE");
	}
	this.order = order;
}

/**
 * Gets a sorting order.
 * @return The current sort order (FORWARD or REVERSE)
 */
public int getSortOrder()
{
	return(order);
}

/**
 * Compares its two arguments for order. Returns a negative integer, zero,
 * or a positive integer as the first argument is less than, equal to, or
 * greater than the second respectively.
 * @param row1 The row of the first argument in the compare
 * @param row2 The row of the second argument in the compare
 * @return A negative integer, zero, or a positive integer as the first
 *     argument is less than, equal to, or greater than the second respectively.
 */
public int compare(JCComparableRow row1, JCComparableRow row2)
{
    // Compare value; we start out assuming they are equal
	int compareValue = 0;
	for (int i = 0; i < keyColumns.length && compareValue == 0; i++) {
    	Object obj1 = row1.getValueAt(keyColumns[i]);
    	Object obj2 = row2.getValueAt(keyColumns[i]);

        if (obj1 != obj2) {
        	// null is less than everything
            if (obj1 == null) {
                compareValue = -1;
            }
            else if (obj2 == null) {
                compareValue = 1;
            }
            else {
				
                if (obj1 instanceof String && obj2 instanceof String) {
                    compareValue = ((String)obj1).compareTo((String)obj2);
                }
                else if (obj1 instanceof Boolean && obj2 instanceof Boolean) {
                    boolean v1 = ((Boolean)obj1).booleanValue();
                    boolean v2 = ((Boolean)obj2).booleanValue();
                    // assume that false is less than true
                    compareValue = (v1 == v2) ? 0 : (v2 ? -1: 1);
                }
                else if (obj1 instanceof Date && obj2 instanceof Date) {
                	long v1 = ((Date)obj1).getTime();
                    long v2 = ((Date)obj2).getTime();
                    compareValue = (v1 < v2) ? -1 : (v1 == v2 ? 0 : 1);
                }
                else if (obj1 instanceof Float && obj2 instanceof Float) {
                	float v1 = ((Float)obj1).floatValue();
                    float v2 = ((Float)obj2).floatValue();
                    compareValue = (v1 < v2) ? -1 : (v1 == v2 ? 0 : 1);
                }
                else if (obj1 instanceof Double && obj2 instanceof Double) {
                	double v1 = ((Double)obj1).doubleValue();
                    double v2 = ((Double)obj2).doubleValue();
                    compareValue = (v1 < v2) ? -1 : (v1 == v2 ? 0 : 1);
                }
                else if (obj1 instanceof Integer && obj2 instanceof Integer) {
                	int v1 = ((Integer)obj1).intValue();
                    int v2 = ((Integer)obj2).intValue();
                    compareValue = (v1 < v2) ? -1 : (v1 == v2 ? 0 : 1);
                }
                else if (obj1 instanceof Long && obj2 instanceof Long) {
                	long v1 = ((Long)obj1).longValue();
                    long v2 = ((Long)obj2).longValue();
                    compareValue = (v1 < v2) ? -1 : (v1 == v2 ? 0 : 1);
                }
                else if (obj1 instanceof BigInteger &&
				         obj2 instanceof BigInteger)
				{
                    compareValue =
                        ((BigInteger)obj1).compareTo((BigInteger)obj2);
                }
				else if (obj1 instanceof Number && obj2 instanceof Number) {
					// We could do somthing smarter here to handle strange
					// cases better (like comparing a really big BigInteger
					// to a really big BigDecimal)
                	double v1 = ((Number)obj1).doubleValue();
                    double v2 = ((Number)obj2).doubleValue();
                    compareValue = (v1 < v2) ? -1 : (v1 == v2 ? 0 : 1);
				}
                else if (obj1 instanceof Comparable &&
				         obj2 instanceof Comparable)
				{
					// This is done at the end since using default compare
					// methods will blow up in the mixed type case which is
					// handled immediately above.
                    compareValue = ((Comparable)obj1).compareTo(obj2);
                }
            }
        }
	}

    // Returns according with the compare value
    return(order*compareValue);
}

}
