/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: DefaultRowSortTableModel.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import com.klg.jclass.util.JCIntComparator;
import com.klg.jclass.util.JCMappingSort;
import com.klg.jclass.util.swing.JCComparableRow;

/**
 * Wrap a table model with a sorted table model
 * @see JCRowSortTableModel
 * @see TableModelListener
 */
public class DefaultRowSortTableModel extends AbstractTableModel
			                          implements JCRowSortTableModel,
												 TableModelListener
{

/** Valid value for SortOrder */
static public final int FORWARD = 1;
/** Valid value for SortOrder */
static public final int REVERSE = -1;

// Instance Variables
protected int                   order = FORWARD;
protected TableModel            tableModel = null;
protected int                   sortColumn = -1;
protected int                   indices[] = null;
protected int                   reverseIndices[] = null;
protected JCRowComparator       rowComparator = null;
protected int[]                 keyColumnsArray[];
protected RowIntComparator      rowIntComparator = null;
protected boolean               autoSort = false;
protected boolean               needsResorting = true;
protected boolean               needsReordering = true;

/**
 * Constructs a <code>JCRowSortModel</code> that wraps a given instance of a
 * <code>TableModel</code> and provides a sorted mapping of the
 * <code>TableModel</code> to the view. The sort order is defined by the
 * configurable <code>Comparator</code> property.
 * @param tableModel The TableModel to wrap
 * @param comparator The comparator to do the sorting
 */
public DefaultRowSortTableModel(TableModel tableModel,
								JCRowComparator comparator)
{
	this(tableModel);
	this.rowComparator = comparator;
}

/**
 * Constructs a <code>JCRowSortModel</code> that wraps a given instance of
 * a <code>TableModel</code> and provides a sorted mapping of the
 * <code>TableModel</code> to the view.
 * @param tableModel The TableModel to wrap
 */
public DefaultRowSortTableModel(TableModel tableModel)
{
	setModel(tableModel);
}

/**
 * Returns the table model this model wraps.
 * @return The TableModel which is wrapped my this model
 */
public TableModel getModel()
{
	return(tableModel);
}

/**
 * Sets the table model wrapped by this model.
 * @param newModel The new TableModel to wrap
 */
public void setModel(TableModel newModel)
{
	if (newModel == this.tableModel) {
		return;
	}
	if (this.tableModel != null) {
		this.tableModel.removeTableModelListener(this);
	}
	this.tableModel = newModel;
	newModel.addTableModelListener(this);
}

/**
 * Sets the key columns used to sort the table model when clicking
 * on a specified column.
 * @param column index of column in <code>TableModel</code>.
 * @param keyColumns columns to use as keys to sort when sorting
 *     on the specified column.
 */
public void setKeyColumns(int column, int keyColumns[])
{
    if (keyColumnsArray == null) {
		keyColumnsArray = new int[tableModel.getColumnCount()][];
    }
	keyColumnsArray[column] = keyColumns;
}

/**
 * Returns the key columns used to sort the table model when clicking
 * on the specified column.  If the user has not set the key columns,
 * the method returns an array of length 1 with the specified column
 * as the only member.
 * @param column index of column in <code>TableModel</code>.
 * @return The key columns used to sort the table for the given column
 */
public int[] getKeyColumns(int column)
{
	return(keyColumnsArray[column]);
}

/**
 * Returns whether or not the data set is currently sorted.
 * @return Is the data currently sorted?
 */
public boolean isDataSorted()
{
	if (indices == null || needsReordering == true || needsResorting == true) {
		return(false);
	}
	return(true);
}

/**
 * Returns whether the data is automatically sorted when it changes
 * according the current comparator.
 * @return Is the data automatically sorted?
 */
public boolean getAutoSort()
{
	return(autoSort);
}

 /**
  * Sets whether the data should be automatically sorted when it changes.
  * @param autoSort The new value of autoSort
  */
public void setAutoSort(boolean autoSort)
{
	if (this.autoSort == autoSort) {
		return;
	}
	this.autoSort = autoSort;
	if (autoSort == true) {
		setDataSorted(true);
	}
}

/**
 * Gets the sorting column.
 * @return The current sort column
 */
public int getSortColumn()
{
    return(sortColumn);
}

/**
 * Sets the sort column.
 * @param column The new sort column
 */
public void setSortColumn(int column)
{
	if (column < 0) {
        throw new IllegalArgumentException("Sorting column number is less than 0");
    }
	if (sortColumn == column) {
		return;
	}

    sortColumn = column;
	needsResorting = true;

	if (autoSort) {
		setDataSorted(true);
	}
}

/**
 * Gets the sorting order.
 * @return The current sorting order
 */
public int getSortOrder()
{
    return(order);
}

/**
 * Sets the sorting order.
 * @param order The new sorting order
 */
public void setSortOrder(int order)
{
	if (order == this.order) {
		return;
	}
    this.order = order;
	needsReordering = true;

	if (autoSort) {
		setDataSorted(true);
	}
}

/**
 * Toggles the sorting order.
 */
public void toggleSortOrder()
{
    order = (order == FORWARD) ? REVERSE : FORWARD;
	needsReordering = true;
}

/**
 * Returns the unsorted row index of specified sorted row.
 * @param sortedRow The sorted row index
 * @return The unsorted row index
 */
public int getUnsortedRow(int sortedRow)
{
	if (sortedRow < 0) {
		return(sortedRow);
	}
	if (indices == null) {
		return(sortedRow);
	}
	else if (order == FORWARD) {
		return(indices[sortedRow]);
	}
	else {
		// If order is reversed then we index from the end of the indices array
		return(indices[indices.length - sortedRow - 1]);
	}
}

/**
 * Returns the sorted row index of the specified unsorted row
 * @param row The unsorted row index
 * @return The sorted row index
 */
public int getSortedRow(int row)
{
	if (row < 0) {
		return(row);
	}
	if (reverseIndices == null) {
		return(row);
	}
	else if (order == FORWARD) {
		return(reverseIndices[row]);
	}
	else {
		// If order is reversed then we index from the end of the indices array
		return(reverseIndices.length - reverseIndices[row] - 1);
	}
}

/**
 * Sets the <code>Comparator</code> property defining the sort order to provide
 * a sorted mapping of the given <code>TableModel</code> to the view.
 * @param comparator The new comparator for sorting
 */
public void setComparator(JCRowComparator comparator)
{
	if (this.rowComparator == comparator) {
		return;
	}
	this.rowComparator = comparator;
	needsResorting = true;

	if (autoSort) {
		setDataSorted(true);
	}
}

/**
 * Returns the current comparator.
 * @return The current comparator for sorting
 */
public JCRowComparator getComparator()
{
	return(rowComparator);
}

/**
 * Sorts the indices of rows using the quicksort algorithm.
 * @param sorted If true and the data is not sorted, sorts the data.  If false,
 *     unsorts the data.
 */
public void setDataSorted(boolean sorted)
{
	if (sorted == false) {
		unsort();
		return;
	}

	if (sortColumn == -1) {
		// Column has not been set so we cannot sort
		return;
	}

	if (!needsReordering && !needsResorting) {
        return;
	}

	if (needsResorting) {
		if (rowComparator == null) {
			rowComparator = new DefaultRowComparator(JCRowSortModel.FORWARD);
		}
		if (keyColumnsArray == null) {
			keyColumnsArray = new int[tableModel.getColumnCount()][];
		}

		int keyColumns[] = { sortColumn };
		rowComparator.setKeyColumns(keyColumnsArray[sortColumn] != null ?
									keyColumnsArray[sortColumn] : keyColumns);

		setComparator(rowComparator);

		if (rowIntComparator == null) {
			rowIntComparator = new RowIntComparator();
		}

		if (indices == null) {
			indices = new int[tableModel.getRowCount()];
			reverseIndices = new int[tableModel.getRowCount()];
			for (int i = 0; i < indices.length; i++) {
				indices[i] = i;
			}
		}

		JCMappingSort.sort(rowIntComparator, indices);
		doReverseMapping();
		needsResorting = false;
	}
	needsReordering = false;
    fireTableDataChanged();
}

/**
 * Restores the unsorted order.
 */
public void unsort()
{
	if (indices != null) {
        indices = null;
		reverseIndices = null;
    	sortColumn = -1;
    }
	// Always fire a table changed since unsort should never be called for
	// no reason and in fact can be called when the underlying model has
	// changed and therefore the change needs to be propegated up
	fireTableDataChanged();
}

/**
 * Do the reverse mapping so we can get a sorted row from an unsorted row.
 */
protected void doReverseMapping()
{
	if (indices == null || reverseIndices == null) {
		return;
	}

	for (int i = 0; i < tableModel.getRowCount(); i++) {
        reverseIndices[indices[i]] = i;
	}
}

//================== TableModel interface methods ==================

/**
 *  Returns the number of columns in the <code>TableModel</code>.
 *  @return The current number of columns
 */
public int getColumnCount()
{
    return(tableModel.getColumnCount());
}

/**
 *  Returns the number of rows in the <code>TableModel</code>.
 *  @return The current number of rows
 */
public int getRowCount()
{
    return(tableModel.getRowCount());
}

/**
 *  Returns a <code>TableModel</code> default column name.
 *  @param column The column number whose name is required
 *  @return The name for the required column
 */
public String getColumnName(int column)
{
    return(tableModel.getColumnName(column));
}

/**
 *  Returns the lowest common denominator <code>Class</code> for the
 *  <code>TableModel</code> column.
 *  @param columnIndex The index of the column whose class is required
 *  @return The current class for the given column
 */
public Class getColumnClass(int columnIndex)
{
    return(tableModel.getColumnClass(columnIndex));
}

/**
 * Returns true if the <code>TableModel</code> cell at <code>rowIndex</code>
 * and <code>columnIndex</code> is editable.
 * @param rowIndex The row index of the cell
 * @param columnIndex The column index of the cell
 * @return Is the cell at the given row and column index editable?
 */
public boolean isCellEditable(int rowIndex, int columnIndex)
{
    return(tableModel.isCellEditable(rowIndex, columnIndex));
}

/**
 * Gets a value for the record in the <code>TableModel</code> cell at
 * <code>columnIndex</code> and <code>rowIndex</code>.
 * @param rowIndex The row index of the cell
 * @param columnIndex The column index of the cell
 * @return The value at the given row and column index
 */
public Object getValueAt(int rowIndex, int columnIndex)
{
	return(tableModel.getValueAt(getUnsortedRow(rowIndex), columnIndex));
}

/**
 * Sets a value for the record in the <code>TableModel</code> cell at
 * <code>columnIndex</code> and <code>rowIndex</code>.
 * @param aValue The new value for the cell with the given row and cell index
 * @param rowIndex The row index of the cell
 * @param columnIndex The column index of the cell
 */
public void setValueAt(Object aValue, int rowIndex, int columnIndex)
{
    needsResorting = true;
	tableModel.setValueAt(aValue, getUnsortedRow(rowIndex), columnIndex);
}


//================== TableModelListener interface method ==================

/**
 * The table model has changed.
 * @param e The event that generated this call
 */
public void tableChanged(TableModelEvent e)
{
    // Table has changed so we need to invalidate the
    // current mappings; we do this by setting indices = null
    indices = null;
    reverseIndices = null;
    needsResorting = true;
    
	// Both of the following cases will cause a tableChanged event to
	// be fired to any listeners on this model
	if (autoSort) {
		setDataSorted(true);
	}
    else {
        fireTableDataChanged();
    }
}

/**
 * Internal <code>ComparableRow</code> class.
 */
class ComparableRow implements JCComparableRow
{
	private int row = -1;

	//================== JCComparableRow interface method ==================

    /**
     * Returns the value of the cell specified by the column and the current row.
	 * @param column The column of the cell
	 * @return The value of the cell in the given column for this comparable row
     */
    public Object getValueAt(int column)
	{
	    return(DefaultRowSortTableModel.this.tableModel.getValueAt(row, column));
    }

    /**
     * Returns the natural (unsorted) index of the current row.
	 * @return The row index for this comparable row
     */
    public int getRowIndex()
	{
    	return(row);
    }

} // ComparableRow

/**
 * Internal <code>RowIntComparator</code> class.
 */
class RowIntComparator implements JCIntComparator
{
	private ComparableRow r1 = new ComparableRow();
    private ComparableRow r2 = new ComparableRow();

	//================== JCIntComparator interface method ==================

	/**
	 * The method which calls the row comparator to do the row comparison
	 * @param index1 The index of the first row
	 * @param index2 The index of the second row
	 * @return A value which tells whether row1 is less than (negative),
	 *     equal to (0), or greather than (positive) row2.
	 */
	public int compare(int index1, int index2)
	{
        r1.row = index1;
        r2.row = index2;
    	return(rowComparator.compare(r1, r2));
    }

} // RowIntComparator

}
