/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCAction.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import com.klg.jclass.util.JCLocaleManager;

import javax.swing.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.Vector;
import java.util.Iterator;

/**
 * Abstract implementation of <code>Action</code> interface, but also adds
 * i18n/l10n functionality.<p>
 * Key enhancements are implementations of <code>Accelerator</code> and
 * <code>Mnemonic</code> support. Does not implement
 * <code>actionPerformed()</code>.
 */
public abstract class JCAction implements Action, java.io.Serializable
{

//
// Action Properties
//
public static final String DEFAULT			  = Action.DEFAULT;
public static final String NAME				  = Action.NAME;
public static final String SHORT_DESCRIPTION  = Action.SHORT_DESCRIPTION;
public static final String LONG_DESCRIPTION	  = Action.LONG_DESCRIPTION;
public static final String SMALL_ICON		  = Action.SMALL_ICON;
public static final String ACCELERATOR		  = "Accelerator";
public static final String MNEMONIC			  = "Mnemonic";

protected Vector                 values = new Vector(5);
protected boolean                enabled = true;
protected PropertyChangeSupport  property;
protected String                 name;

public JCAction(String name)
{
	property = new PropertyChangeSupport(this);
	this.name = name;
}

/**
 * Creates an action.
 * @param label Label of the action used when creating buttons. On request of
 * the NAME, this value will be used as a key into the
 * <code>JCLocaleManager</code>'s default implementation.  If the key is not
 * found therein, it will be returned directly.
 * @param mnemonic mnemonic key to use when constructing a menu item
 * @param accelerator accelerator key to use when constructing a menu item
 * or button
 * @see com.klg.jclass.util.swing.JCMenu
 * @see com.klg.jclass.util.JCLocaleManager
 */
public JCAction(String label, Integer mnemonic, KeyStroke accelerator)
{
	this(null);
	putValue(NAME, label);
	putValue(MNEMONIC, mnemonic);
	putValue(ACCELERATOR, accelerator);;
}

public JCAction(String label, int mnemonic, KeyStroke accelerator)
{
	this(label, new Integer(mnemonic), accelerator);
}

public Object getValue(String key)
{
	// Certain values we get from the i18n bundle
	JCLocaleManager li = JCLocaleManager.getDefault();
	if (name != null && li != null) {
		if (key.equals(NAME)) {
			String label;
			try {
				label = li.getString(name);
			}
			catch (Exception e) {
				label = name;
			}
			if (label == null) {
				label = name;
			}
			return(label);
		}
		else if (key.equals(ACCELERATOR)) {
			KeyStroke k;
			try {
				k = li.getKeyStroke(name + "Accelerator");
			}
			catch (Exception e) {
				k = null;
			}
			return(k);
		}
		else if (key.equals(MNEMONIC)) {
			Integer mnemonic;
			try {
				mnemonic = (Integer) li.getObject(name + "Mnemonic");
			}
			catch (Exception e) {
				mnemonic = new Integer(-1);
			}
			return(mnemonic);
		}
	}

    Iterator iter = values.iterator();
	while (iter.hasNext()) {
		Value value = (Value)iter.next();
		if (value.key.equals(key)) {
			return(value.value);
		}
	}
	return(null);
}

public void putValue(String key, Object v)
{
	// Loop through to see if the key already exists
    Iterator iter = values.iterator();
    while (iter.hasNext()) {
        Value value = (Value)iter.next();
        if (value.key.equals(key)) {
            property.firePropertyChange(key, value.value, v);
            value.value = v;
            return;
        }
    }

	// Since we made it here the key doesn't exist so we create a new one
	property.firePropertyChange(key, null, v);
	Value value = new Value(key, v);
	values.addElement(value);
}

public void setEnabled(boolean b)
{
	if (enabled != b) {
		property.firePropertyChange("enabled",
									new Boolean(enabled),
									new Boolean(b));
		enabled = b;
	}
}

public boolean isEnabled()
{
	return(enabled);
}

public void addPropertyChangeListener(PropertyChangeListener listener)
{
	property.addPropertyChangeListener(listener);
}

public void removePropertyChangeListener(PropertyChangeListener listener)
{
	property.removePropertyChangeListener(listener);
}

public JButton createButton(boolean use_mnemonic)
{
	JButton button = new ActionButton((String) getValue(NAME));
	button.addActionListener(this);
	if (use_mnemonic) {
		try {
			button.setMnemonic(((Integer) getValue(MNEMONIC)).intValue());
		}
		catch (Exception e) {}
	}
	return button;
}

public JButton createButton()
{
	return(createButton(true));
}

public JMenu createMenu()
{
	JMenu menu = new JMenu((String) getValue(NAME));
	menu.addActionListener(this);
	try {
		menu.setMnemonic(((Integer) getValue(MNEMONIC)).intValue());
	}
	catch (Exception e) {}
	return(menu);
}

public JMenuItem createMenuItem()
{
	JMenuItem menuitem = new JMenuItem((String) getValue(NAME));
	menuitem.addActionListener(this);
	try {
		menuitem.setMnemonic(((Integer) getValue(MNEMONIC)).intValue());
	}
	catch (Exception e) {}
	try {
		menuitem.setAccelerator((KeyStroke) getValue(ACCELERATOR));
	}
	catch (Exception e) {}
	return(menuitem);
}

/**
 * Inner class to store key - value pairs in a list.
 */
protected class Value implements java.io.Serializable
{
    protected String key;
    protected Object value;

    public Value(String key, Object value)
    {
        this.key = key;
        this.value = value;
    }
}

protected class ActionButton extends JButton
                             implements PropertyChangeListener
{
	public ActionButton(String name)
    {
		super(name);
	}

	public void propertyChange(PropertyChangeEvent e)
    {
		String propertyName = e.getPropertyName();
		if (e.getPropertyName().equals(Action.NAME)) {
			String text = (String) e.getNewValue();
			setText(text);
			repaint();
		}
		else if (propertyName.equals("enabled")) {
			Boolean enabledState = (Boolean) e.getNewValue();
			this.setEnabled(enabledState.booleanValue());
			repaint();
		}
		else if (e.getPropertyName().equals(Action.SMALL_ICON)) {
			Icon icon = (Icon) e.getNewValue();
			setIcon(icon);
			invalidate();
			repaint();
		}
	}
}

protected class ActionMenu extends JMenu
                           implements PropertyChangeListener
{
	public ActionMenu(String name)
    {
		super(name);
	}

	public void propertyChange(PropertyChangeEvent e)
    {
		String propertyName = e.getPropertyName();
		if (e.getPropertyName().equals(Action.NAME)) {
			String text = (String) e.getNewValue();
			setText(text);
			repaint();
		}
		else if (propertyName.equals("enabled")) {
			Boolean enabledState = (Boolean) e.getNewValue();
			this.setEnabled(enabledState.booleanValue());
			repaint();
		}
		else if (e.getPropertyName().equals(Action.SMALL_ICON)) {
			Icon icon = (Icon) e.getNewValue();
			setIcon(icon);
			invalidate();
			repaint();
		}
	}
}

protected class ActionMenuItem extends JMenuItem
                               implements PropertyChangeListener
{
	public ActionMenuItem(String name)
    {
		super(name);
	}

	public void propertyChange(PropertyChangeEvent e)
    {
		String propertyName = e.getPropertyName();
		if (e.getPropertyName().equals(Action.NAME)) {
			String text = (String) e.getNewValue();
			setText(text);
			repaint();
		}
		else if (propertyName.equals("enabled")) {
			Boolean enabledState = (Boolean) e.getNewValue();
			this.setEnabled(enabledState.booleanValue());
			repaint();
		}
		else if (e.getPropertyName().equals(Action.SMALL_ICON)) {
			Icon icon = (Icon) e.getNewValue();
			setIcon(icon);
			invalidate();
			repaint();
		}
	}
}

}
