/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCAlignLayout.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:44 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import java.awt.Container;
import java.awt.Component;
import java.util.Hashtable;

/** A layout that provides a simple way to lay out a vertically arranged
 * group of control components, each with an associated label (or other 
 * component) placed to its left.<p>
 *
 * <strong>Properties</strong><p>
 * <table border="1">
 * <tr><th>Name</th>    <th>Method</th></tr>
 * <tr><td>LabelDefaultAlignment</td>
 * <td><a href=#setLabelDefaultAlignment>setLabelDefaultAlignment</a></td></tr>
 * <tr><td>LabelVerticalAlignment</td>
 * <td><a href=#setLabelVerticalAlignment>setLabelVerticalAlignment</a></td></tr>
 * <tr><td>ResizeHeight</td>
 * <td><a href=#setResizeHeight>setResizeHeight</a></td></tr>
 * <tr><td>ResizeWidth</td>
 * <td><a href=#setResizeWidth>setResizeWidth</a></td></tr>
 * </table>
 */
public class JCAlignLayout extends JCGridLayout {

public static final int TOP = 1;
public static final int MIDDLE = 4;
public static final int BOTTOM = 5;

protected Hashtable  alignment;
protected Hashtable  resize_width, resize_height;

/**
 * Creates an aligner layout with 2 columns, a variable number of rows,
 * and a gap of 5 pixels.
 */
public JCAlignLayout()
{
    this(2, 5, 5);
}

/**
 * Creates an aligner layout with the specified number of columns and gaps.
 * @param cols the number of columns (should be a multiple of 2)
 * @param hgap the horizontal gap variable
 * @param vgap the vertical gap variable
 * @exception IllegalArgumentException If the rows and columns are invalid.
 */
public JCAlignLayout(int cols, int hgap, int vgap)
{
    super(VARIABLE, cols, hgap, vgap);
}

private int get(Hashtable table, Component comp, int def)
{
    Object v = (table != null) ? table.get("" + comp.hashCode()) : null;
    return((v != null) ? ((Integer)v).intValue() : def);
}

private boolean get(Hashtable table, Component comp, boolean def)
{
    Object v = (table != null) ? table.get("" + comp.hashCode()) : null;
    return((v != null) ? ((Boolean)v).booleanValue() : def);
}

/**
 * Gets the vertical position of a label relative to its control.
 * @see #setLabelVerticalAlignment
 * @return vertical position of a label
 * @param child
 */
public int getLabelVerticalAlignment(Component child)
{
    return(get(alignment, child, MIDDLE));
}

/**
 * Sets the vertical position of a label relative to its control.
 * @param align <code>TOP</code>, <code>MIDDLE</code> (default), or
 * <code>BOTTOM</code>
 * @exception IllegalArgumentException If an invalid value is set.
 * @param child
 */
public void setLabelVerticalAlignment(Component child, int align)
{
    if (alignment == null) {
        alignment = new Hashtable(5);
    }
    alignment.put("" + child.hashCode(), new Integer(align));
}

/** Gets the component's <code>RezizeWidth</code> value.
 * @see #setResizeWidth
 * @param child
 * @return
 */
public boolean getResizeWidth(Component child)
{
    return(get(resize_width, child, false));
}

/** Sets whether the control should be resized horizontally to its parent's
 * right edge if it is in the last column (default is <code>false</code>).
 * @param child
 * @param v
 */
public void setResizeWidth(Component child, boolean v)
{
    if (resize_width == null) {
        resize_width = new Hashtable(5);
    }
    resize_width.put("" + child.hashCode(), new Boolean(v));
}

/** Gets the component's <code>RezizeHeight</code> value.
 * @see #setResizeHeight
 * @param child
 * @return
 */
public boolean getResizeHeight(Component child)
{
    return(get(resize_height, child, false));
}

/** Sets whether the control should be resized vertically to the height of the
 * largest component in its row (default is <code>false</code>). This value is
 * ignored for labels (the components in odd columns).
 * @param child
 * @param v
 */
public void setResizeHeight(Component child, boolean v)
{
    if (resize_height == null) {
        resize_height = new Hashtable(5);
    }
    resize_height.put("" + child.hashCode(), new Boolean(v));
}

protected boolean isLabel(int col)
{
    return((col % 2) == 0);
}

/**
 * Positions the component.
 * @param pos the component's index in its parents child list
 * @param row component's position
 * @param col component's position
 * @param col_width
 * @param comp
 * @param pos
 * @param row_height
 * @param x
 * @param y
 */
protected void setBounds(int pos, int row, int col, Component comp,
                         int x, int y, int col_width, int row_height)
{
    int comp_w = col_width;
    int comp_h = row_height;
    if (isLabel(col) || !getResizeHeight(comp)) {
        comp_h = comp.getPreferredSize().height;
    }

    /*
     * Resize a control to its parent's right edge if its resizeWidth value
     * is true, and it is in the last column
     */
    if (!isLabel(col)) {
        if (col < col_widths.length - 1) {
        }
        else if (getResizeWidth(comp)) {
            Container parent = comp.getParent();
            comp_w = parent.getSize().width - parent.getInsets().right - x;
        }
        else {
            comp_w = comp.getPreferredSize().width;
        }

        comp.setBounds(x, y, comp_w, comp_h);
        return;
    }

    int control_h = row_height;
    if (pos < comp.getParent().getComponentCount() - 1) {
        Component control = comp.getParent().getComponents()[pos + 1];
        if (control != null && !getResizeHeight(control)) {
            control_h = control.getPreferredSize().height;
        }
    }

    switch (getLabelVerticalAlignment(comp)) {
    case MIDDLE:
        y += (control_h - comp_h) / 2;
        break;
    case BOTTOM:
        y += control_h - comp_h;
        break;
    }
    comp.setBounds(x, y, comp_w, comp_h);
}

}
