/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCBorder.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:45 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;
import java.awt.*;

/**
 * The class contains static methods to draw shadows around a specified
 * area. The most useful methods contained here are the various
 * <i>draw()</i> methods.<p>
 * The key difference between this static class and the Swing border
 * classes is the ability to draw borders <i>anywhere</i>, not just
 * around the outside of a component.
 */
public class JCBorder {

	// Border Styles
/** No border. */
public static final int NONE        = 0;
/** Border drawn with double line; border appears inset. */
public static final int ETCHED_IN   = 1;
/** Border drawn with double line; border appears raised. */
public static final int ETCHED_OUT  = 2;
/** Border drawn appears inset. */
public static final int IN          = 3;
/** Border drawn appears raised. */
public static final int OUT         = 4;
/** Border drawn in foreground color. */
public static final int PLAIN       = 5;
/** Border drawn with 1-pixel shadow-in at edge; border appears framed. */
public static final int FRAME_IN    = 6;
/** Border drawn with 1-pixel shadow-out at edge; border appears framed. */
public static final int FRAME_OUT   = 7;
/** Border drawn with MS-Windows style control shadows. */
public static final int CONTROL_IN  = 8;
/** Border drawn with MS-Windows style control shadows. */
public static final int CONTROL_OUT = 9;
/** Number of different border styles. */
public static final int NUM_BORDER_STYLES = 10;

/** String names of the border styles. */
public final static String[] border_strings = {
    "None",
    "Etched_In", "Etched_Out",
    "In", "Out",
    "Plain",
    "Frame_In", "Frame_Out",
	"Control_In", "Control_Out",
};

/**
 * Style value that corresponds to the list of String names.
 * @see #border_strings
 */
public final static int[] border_values = {
	NONE,
	ETCHED_IN, ETCHED_OUT,
	IN, OUT,
	PLAIN,
	FRAME_IN, FRAME_OUT,
	CONTROL_IN, CONTROL_OUT
};

/**
 * Draws the top and left lines.
 * @param g The graphics context used to draw
 * @param size The number of lines to draw
 * @param x The x position to start drawing
 * @param y The y position to start drawing
 * @param w The width of the horizontal lines
 * @param h The height of the vertical lines
 * @param color The line color
 */
public static void drawTopLines(Graphics g, int size, int x, int y,
                                int w, int h, Color color)
{
    g.setColor(color);
    for (int i = 0; i < size; i++) {
        g.drawLine(x + i, y + i, x + w - (i + 1), y + i);
        g.drawLine(x + i, y + i + 1, x + i, y + h - (i + 1));
    }
}

/**
 * Draws the bottom and right lines.
 * @param g The graphics context used to draw
 * @param size The number of lines to draw
 * @param x The x position to start drawing
 * @param y The y position to start drawing
 * @param w The width of the horizontal lines
 * @param h The height of the vertical lines
 * @param color The line color
 */
public static void drawBottomLines(Graphics g, int size, int x, int y,
                                   int w, int h, Color color)
{
    g.setColor(color);
    for (int i = 1; i <= size; i++) {
        g.drawLine(x + i - 1, y + h - i, x + w - i, y + h - i);
        g.drawLine(x + w - i, y + i - 1, x + w - i, y + h - i);
    }
}

public static void drawNormal(Graphics g, int size, int x, int y, int w, int h,
                              Color top, Color bottom)
{
    drawTopLines(g, size, x, y, w, h, top);
    drawBottomLines(g, size, x, y, w, h, bottom);
}

private static final Color WHITE_BRIGHTER = new Color(200, 200, 200);
private static final Color WHITE_DARKER = new Color(140, 140, 140);
private static final Color BLACK_BRIGHTER = new Color(125, 125, 125);
private static final Color BLACK_DARKER = new Color(75, 75, 75);
private static final double FACTOR = 0.5;

/**
 * Calculates the color even when it is saturated, for example,
 * when it is black or white.
 * @param color The color to make brighter
 * @return The brighter color
 */
public static Color brighter(Color color)
{
    if (color.equals(Color.white)) {
        return(WHITE_BRIGHTER);
    }
    else if (color.equals(Color.black)) {
        return(BLACK_BRIGHTER);
    }
    else {
        // Instead of calling Color.brighter(), we do our own calculation
        // for a lighter color.  This calc brightens up dark colors very
        // quickly while lightening up bright colors only a bit.
        int red = color.getRed();
        red += (int)((255 - red) * FACTOR);
        int blue = color.getBlue();
        blue += (int)((255 - blue) * FACTOR);
        int green = color.getGreen();
        green += (int)((255 - green) * FACTOR);

        return(new Color(Math.min(red, 255),
                         Math.min(green, 255),
                         Math.min(blue, 255)));
    }
}

/**
 * Calculates the color even when it is saturated, for example,
 * when it is black or white.
 * @param color The color to make darker
 * @return The darker color
*/
public static Color darker(Color color)
{
    if (color.equals(Color.white)) {
        return(WHITE_DARKER);
    }
    else if (color.equals(Color.black)) {
        return(BLACK_DARKER);
    }
    else {
        return(color.darker());
    }
}

/**
 * Draws a border.
 * @param gc The graphics context used to draw
 * @param style One of the following members:
 <pre>
 ETCHED_IN    double line; border appears inset<br>
 ETCHED_OUT   double line; border appears raised<br>
 FRAME_IN     1-pixel shadow-in at edge; border appears framed<br>
 FRAME_OUT    1-pixel shadow-out at edge; border appears framed<br>
 IN           border appears inset<br>
 OUT          border appears raised<br>
 CONTROL_IN   MS-Windows control shadows<br>
 CONTROL_OUT  MS-Windows control shadows<br>
 PLAIN        border drawn in foreground color<br>
 NONE         no border drawn
 </pre>
 * @param size the shadow thickness
 * @param x The x position of the rectangle in which to draw
 * @param y The y position of the rectangle in which to draw
 * @param width The width of the rectangle in which to draw
 * @param height The height of the rectangle in which to draw
 * @param bg The background color
 * @param plain The color used to draw the sides of <code>PLAIN</code> shadows
 */
public static void draw(Graphics gc, int style, int size,
                        int x, int y, int width, int height,
                        Color bg, Color plain)
{
    draw(gc, style, size, x, y, width, height,
         brighter(bg), darker(bg), plain);
}

/**
 * Draws a border.
 * @param gc The graphics context used to draw
 * @param style One of the following members:
 <pre>
 ETCHED_IN    double line; border appears inset<br>
 ETCHED_OUT   double line; border appears raised<br>
 FRAME_IN     1-pixel shadow-in at edge; border appears framed<br>
 FRAME_OUT    1-pixel shadow-out at edge; border appears framed<br>
 IN           border appears inset<br>
 OUT          border appears raised<br>
 CONTROL_IN   MS-Windows control shadows<br>
 CONTROL_OUT  MS-Windows control shadows<br>
 PLAIN        border drawn in foreground color<br>
 NONE         no border drawn
 </pre>
 * @param size the shadow thickness
 * @param x The x position of the rectangle in which to draw
 * @param y The y position of the rectangle in which to draw
 * @param width The width of the rectangle in which to draw
 * @param height The height of the rectangle in which to draw
 * @param bright Bright shadow color
 * @param dark Dark shadow color
 * @param plain The color used to draw the sides of <code>PLAIN</code> shadows
 */
public static void draw(Graphics gc, int style, int size,
                        int x, int y, int width, int height,
                        Color bright, Color dark, Color plain)
{
    int r2 = size % 2;
    int s2 = (r2 == 0) ? size/2 : size/2 + 1;
    int s3 = size - s2;

    Color old_color = gc.getColor();
    switch (style) {
    case PLAIN:
        drawNormal(gc, size, x, y, width, height, plain, plain);
        break;
    case IN:
        drawNormal(gc, size, x, y, width, height, dark, bright);
        break;
    case OUT:
        drawNormal(gc, size, x, y, width, height, bright, dark);
        break;
    case CONTROL_IN:
        if (size == 2) {
            drawNormal(gc, 1, x + 1, y + 1, width - 2, height - 2, Color.black,
                       bright);
            drawNormal(gc, 1, x, y, width, height, dark, brighter(bright));
        }
        else {
            drawNormal(gc, size, x, y, width, height, dark, bright);
        }
        break;
    case CONTROL_OUT:
        if (size == 2) {
            drawNormal(gc, 1, x, y, width, height, brighter(bright), dark);
            drawNormal(gc, 1, x + 1, y + 1, width - 2, height - 2, bright,
                       Color.black);
        }
        else {
            drawNormal(gc, size, x, y, width, height, bright, dark);
        }
        break;
    case ETCHED_IN:
        drawNormal(gc, s2, x, y, width, height, dark, bright);
        drawNormal(gc, s3, x + s2, y + s2, width - size - r2,
                   height - size - r2, bright, dark);
        break;
    case ETCHED_OUT:
        drawNormal(gc, s2, x, y, width, height, bright, dark);
        drawNormal(gc, s3, x + s2, y + s2, width - size - r2,
                   height - size - r2, dark, bright);
        break;
    case FRAME_OUT:
        // 1-pixel shadow
        drawNormal(gc, size, x, y, width, height, bright, dark);
        // Plain
        drawNormal(gc, size - 1, x, y, width, height, plain, plain);
        break;
    case FRAME_IN:
        // 1-pixel shadow
        drawNormal(gc, size, x, y, width, height, dark, bright);
        // Plain
        drawNormal(gc, size - 1, x, y, width, height, plain, plain);
        break;
    default:
        break;
    }
    gc.setColor(old_color);
}

}
