/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCElasticLayout.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import java.awt.*;
import javax.swing.SwingConstants;
import java.util.HashMap;



/**
 * A layout manager that supports <code>JCElastic</code> components either
 * horizontally or vertically.  A component is considered <i>elastic</i> if it
 * either implements the <code>JCElastic</code> interface <i>or</i> it has a
 * constraint object the implements the <code>JCElastic</code> interface.<p>
 * Layout is performed in either a single row or column (depending on its
 * orientation at creation).  The preferred size is calculated in the
 * direction of orientation.  If the container is bigger than the preferred
 * size of all the components, then the extra space is divided up between the
 * components that are "elastic" in the direction of the orientation.<p>
 * The extra space is allocated to each of the components with respect to
 * their "elasticity".  If all the elastic components have the same
 * elasticity (in the direction of the orientation), then they are equally
 * stretched.  If there is an uneven number of pixels to allot, then the
 * first <i>n</i> units of elasticity are allocated the extra pixels; where
 * <i>n = total_elasticity % extra_pixels</i>.
 * @see JCElastic
 * @see JCSpring
 * @see JCBrace
 */
public class JCElasticLayout implements LayoutManager2, java.io.Serializable {

public static final int HORIZONTAL = 0;
public static final int VERTICAL = 1;

/** Constant to use when adding an elastic constraint to this object. */
public static final JCElastic HORIZONTALLY_ELASTIC_CONSTRAINT = new Elastic(1, 0);
/** Constant to use when adding an elastic constraint to this object. */
public static final JCElastic VERTICALLY_ELASTIC_CONSTRAINT = new Elastic(0, 1);
/** Constant to use when adding an elastic constraint to this object. */
public static final JCElastic COMPLETELY_ELASTIC_CONSTRAINT = new Elastic(1, 1);
/** Constant to use when adding an String constraint to this object. */
public static final String HORIZONTALLY_ELASTIC_TAG = "HorizontallyElastic";
/** Constant to use when adding an String constraint to this object. */
public static final String VERTICALLY_ELASTIC_TAG = "VerticallyElastic";
/** Constant to use when adding an String constraint to this object. */
public static final String COMPLETELY_ELASTIC_TAG = "CompletelyElastic";

/** Returned by <code>getElasticity()</code>. */
protected static final int NOT_ELASTIC = 0;
/** Returned by <code>getElasticity()</code>. */
protected static final int VERTICALLY_ELASTIC = 1;
/** Returned by <code>getElasticity()</code>. */
protected static final int HORIZONTALLY_ELASTIC = 2;

/** Alignment to use when laying out. */
protected int alignment;
/** Orientation to lay out in. <code>HORIZONTAL</code> or <code>VERTICAL</code>. */
protected int orientation;
/** Hashtable to store components added to the layout. */
protected HashMap constraintMap = new HashMap();

static protected HashMap constraintCache = null;

/**
 * Creates a <code>JCElastic</code> constraint with the elasticity specified.
 * @return
 * @param horizontal_elasticity
 * @param vertical_elasticity
 */
public static JCElastic createElasticConstraint(int horizontal_elasticity,
                                                int vertical_elasticity)
{
    if (constraintCache == null) {
        constraintCache = new HashMap();
    }
    // create a new Elastic object
    Elastic e = new Elastic(horizontal_elasticity, vertical_elasticity);

    if (constraintCache.containsKey(e)) {
        // found an object in the cache so we don't have to keep around the
        // new one
        return (JCElastic)constraintCache.get(e);
    }

    // Add the object to the cache using itself as a key
    constraintCache.put(e, e);
    return(e);
}

/**
 * Creates a row layout that by default aligns components to the left of
 * the row.
 * @param orientation
 */
public JCElasticLayout(int orientation)
{
    // Default should be based on orientation
    this(orientation,
         (orientation == VERTICAL) ? SwingConstants.LEFT : SwingConstants.TOP);
}

/**
 * Creates a column layout that aligns components according to the specified
 * alignment.
 * @param alignment <code>SwingConstants.LEFT</code>,
 * <code>SwingConstants.TOP</code>,
 * <code>SwingConstants.CENTER</code>,
 * <code>SwingConstants.BOTTOM</code>, or
 * <code>SwingConstants.RIGHT</code>
 * @param orientation
 */
public JCElasticLayout(int orientation, int alignment)
{
    this.orientation = orientation;
    if (alignment != SwingConstants.LEFT &&
        alignment != SwingConstants.TOP &&
        alignment != SwingConstants.CENTER &&
        alignment != SwingConstants.BOTTOM &&
        alignment != SwingConstants.RIGHT)
    {
        throw new IllegalArgumentException("Incorrect value; use one of: " +
                                           "SwingConstants.LEFT, " +
                                           "SwingConstants.TOP, " +
                                           "SwingConstants.CENTER, " +
                                           "SwingConstants.BOTTOM, " +
                                           "SwingConstants.RIGHT");
    }
    this.alignment = alignment;
}

//
// LayoutManager Implementation
//

/**
 * Adds a component to the layout.
 * @param name
 * @param comp
 */
public void addLayoutComponent(String name, Component comp)
{
    if (name.equals(HORIZONTALLY_ELASTIC_TAG)) {
        addLayoutComponent(comp, HORIZONTALLY_ELASTIC_CONSTRAINT);
    }
    else if (name.equals(VERTICALLY_ELASTIC_TAG)) {
        addLayoutComponent(comp, VERTICALLY_ELASTIC_CONSTRAINT);
    }
    else if (name.equals(COMPLETELY_ELASTIC_TAG)) {
        addLayoutComponent(comp, COMPLETELY_ELASTIC_CONSTRAINT);
    }
}

/**
 * Removes a component from the layout.
 * @param comp
 */
public void removeLayoutComponent(Component comp)
{
    if (constraintMap.containsKey(comp)) {
        constraintMap.remove(comp);
    }
}


/**
 * Returns the preferred layout size of the containers within their parent,
 * given this layout.
 * @param parent
 * @return preferred layout size
 */
public Dimension preferredLayoutSize(Container parent)
{
    Dimension size = new Dimension();
    getPreferredLayoutSize(parent, size);
    return size;
}

/**
 * Returns the preferred layout size of the containers within their parent
 * given this layout, and the total amount of "elasticity" in the child
 * components.
 * @return return the sum of elasticity of the child components in the
 * direction of orientation of the layout
 * @param parent container in which the layout is occuring
 * @param preferred_size object to return the preferred size in
 */
protected int getPreferredLayoutSize(Container parent,
                                     Dimension preferred_size)
{
    //
    // Loop through the children
    // requesting preferred sizes
    //
    int n = parent.getComponentCount();
    int width = 0;
    int height = 0;
    int elasticity = 0;
    int elasticity_total = 0;
    if (orientation == VERTICAL) {
        // COLUMN/HORIZONTAL CASE
        for (int i = 0; i < n; i++) {
            Component c = parent.getComponent(i);
            Dimension size = c.getPreferredSize();
            width = Math.max(width, size.width);
            height += size.height;
            elasticity = getElasticity(c);
            if (elasticity != 0) {
                elasticity_total += elasticity;
            }
        }
    }
    else {
        // ROW/HORIZONTAL CASE
        for (int i = 0; i < n; i++) {
            Component c = parent.getComponent(i);
            Dimension size = c.getPreferredSize();
            height = Math.max(height, size.height);
            width += size.width;
            elasticity = getElasticity(c);
            if (elasticity != 0) {
                elasticity_total += elasticity;
            }
        }
    }

    // add parent's insets
    Insets insets = parent.getInsets();
    width = width + insets.right + insets.left;
    height = height + insets.top + insets.bottom;

    if (preferred_size != null) {
        preferred_size.width = width;
        preferred_size.height = height;
    }

    return elasticity_total;
}

/**
 * Returns the minimum layout size of the containers within their parent
 * given this layout.
 * @param parent
 * @return minimum layout size
 */
public Dimension minimumLayoutSize(Container parent)
{
    //
    // Loop through the children
    // requesting maximum sizes
    //
    int n = parent.getComponentCount();
    int width = 0;
    int height = 0;
    if (orientation == VERTICAL) {
        // COLUMN/HORIZONTAL CASE
        for (int i = 0; i < n; i++) {
            Component c = parent.getComponent(i);
            Dimension size = c.getMinimumSize();
            width = Math.max(width, size.width);
            height += size.height;
        }
    }
    else {
        // ROW/HORIZONTAL CASE
        for (int i = 0; i < n; i++) {
            Component c = parent.getComponent(i);
            Dimension size = c.getMinimumSize();
            height = Math.max(height, size.height);
            width += size.width;
        }
    }

    return new Dimension(width, height);
}

/**
 * Lays out the components within the specified parent.
 * @param parent
 */
public void layoutContainer(Container parent)
{
    Dimension preferred_size = new Dimension();
    int elasticity_total = getPreferredLayoutSize(parent, preferred_size);
    Dimension parent_size = parent.getSize();
    // amount that the initial coords are offset by the container's insets
    int coord1_inset, coord2_inset;
    Insets insets = parent.getInsets();

    // shrink both the preferred size and actual size by the right and
    // bottom margins; we do this because we do not actually lay any thing
    // out in that area.  We have to consider the left and top margins
    // only because they are needed to calculate the initial coordinate
    // offsets
    parent_size.height -= insets.bottom;
    parent_size.width -= insets.right;
    preferred_size.height -= insets.bottom;
    preferred_size.width -= insets.right;

    // PENDING - current behaviour is that is there is not enough space
    // then elastic components are shrunk below their preferred size.
    // we need to take minimum size into account somehow

    int extra_space;

    if (orientation == VERTICAL) {
        extra_space = parent_size.height - preferred_size.height;
        coord1_inset = insets.left;
        coord2_inset = insets.top;
    }
    else {
        extra_space = parent_size.width - preferred_size.width;
        coord1_inset = insets.top;
        coord2_inset = insets.left;
    }

    int slice = 0,
        fatter_slice_count = 0;

    if (elasticity_total > 0) {
        slice = extra_space / elasticity_total;
        fatter_slice_count = extra_space % elasticity_total;
    }

    
    

    int n = parent.getComponentCount();
    int coord2 = coord2_inset;
    int current_slice = 0;

    for (int i = 0; i < n; i++) {
        Component c = parent.getComponent(i);
        int size_dim2;
        // VERTICAL:
        //    dim1 == width
        //    dim2 == height
        //    coord1 == x
        //    coord2 == y
        // HORIZONTAL:
        //    dim1 == height
        //    dim2 == width
        //    coord1 == y
        //    coord2 == x
        int dim1 = 0;
        int max_dim1;
        int parent_dim1, parent_dim2;
        if (orientation == VERTICAL) {
            parent_dim1 = parent_size.width - coord1_inset;
            parent_dim2 = parent_size.height - coord2_inset;
            max_dim1 = c.getMaximumSize().width;
            if (max_dim1 < 0) {
                // gridbag returns - as max width
                max_dim1 = c.getPreferredSize().width;
            }
            size_dim2 = c.getPreferredSize().height;
        }
        else {
            parent_dim1 = parent_size.height - coord1_inset;
            parent_dim2 = parent_size.width - coord2_inset;
            max_dim1 = c.getMaximumSize().height;
            if (max_dim1 < 0) {
                // gridbag returns - as max width
                max_dim1 = c.getPreferredSize().height;
            }
            size_dim2 = c.getPreferredSize().width;
        }

        dim1 = Math.min(parent_dim1, max_dim1);

        int coord1 = coord1_inset;
        if (dim1 != parent_dim1) {
            if (isDim1Elastic(c)) {
                // Component is elastic in a dimension opposite to the
                // orientation of the container so we stretch the component
                // to the dim1 available
                coord1 = 0;
                dim1 = parent_dim1;
            }
            else if (alignment == SwingConstants.LEFT ||
                alignment == SwingConstants.TOP) {
                // left/top justify the component
                coord1 = 0;
            }
            else if (alignment == SwingConstants.CENTER) {
                // center the component in the width available
                coord1 = (parent_dim1 - dim1) / 2;
            }
            else if (alignment == SwingConstants.RIGHT ||
                alignment == SwingConstants.BOTTOM) {
                // right justify the component
                coord1 = parent_dim1 - dim1;
            }
        }

        int elasticity = getElasticity(c);
        if (elasticity > 0) {
            // This block handles stretching out elastic components to
            // fill up remaining space in the component
            int stretch = calcStretch(slice, elasticity, current_slice,
                                      fatter_slice_count);
            if (orientation == VERTICAL) {
                c.setBounds(coord1, coord2, dim1, size_dim2 + stretch);
                
            }
            else {
                c.setBounds(coord2, coord1, size_dim2 + stretch, dim1);
                
            }
            current_slice += elasticity;
            coord2 += size_dim2 + stretch;

        }
        else {
            

            if (orientation == VERTICAL) {
                c.setBounds(coord1, coord2, dim1, size_dim2);
            }
            else {
                c.setBounds(coord2, coord1, size_dim2, dim1);
            }
            coord2 += size_dim2;
        }
    }
}

/**
 * Calculates the amount to stretch an elastic component based on the parameters
 * passed in.
 * @param slice the size of a "slice" of the extra space in the parent component
 * @param elasticity the amount of elasticity in the current component
 * @param current_slice the starting slice for the calculation
 * @param fatter_slice_count the number of slices that need to be "Fatter" by
 * a pixel because the extra space did not divide evenly into the number of
 * slices
 * @return
 */
protected int calcStretch(int slice, int elasticity,
                          int current_slice, int fatter_slice_count)
{
    int stretch = 0;
    // the current component chews up 1 slice per elasticity
    for (int i = current_slice; i < current_slice + elasticity; i++) {
        stretch += slice + ((i < fatter_slice_count) ? 1 : 0);
    }
    return stretch;
}

/**
 * Queries whether or not the dimension opposite to the current orientation is
 * elastic or not.
 * @return
 * @param c
 */
protected boolean isDim1Elastic(Component c)
{
    // Dim1 is defined as opposite the current orientation
    // so we need to figure out if the opposite bit is turned on
    if (orientation == VERTICAL) {
        return isHorizontallyElastic(c);
    }
    else {
        return isVerticallyElastic(c);
    }
}

/**
 * Returns the Elastic information for a particular component.  This information
 * is found either in the component itself or in the constraint associated
 * with the component.
 * @return Elastic information
 * @param c
 */
protected JCElastic getElastic(Component c)
{
    JCElastic e = null;

    if (c instanceof JCElastic) {
        e = (JCElastic)c;
    }
    else {
        Object constraints = constraintMap.get(c);
        if (constraints != null && constraints instanceof JCElastic) {
            e = (JCElastic)constraints;
        }
    }

    return(e);
}

/**
 * Queries whther the given component is vertically elastic or not.
 * @return
 * @param c
 */
protected boolean isVerticallyElastic(Component c)
{
    JCElastic e = getElastic(c);

    if (e != null) {
        if (e.getVerticalElasticity() != 0) {
            return true;
        }
    }
    return(false);
}

/**
 * Queries whther the given component is horizontally elastic or not.
 * @return
 * @param c
 */
protected boolean isHorizontallyElastic(Component c)
{
    JCElastic e = getElastic(c);

    if (e != null) {
        if (e.getHorizontalElasticity() != 0) {
            return(true);
        }
    }
    return(false);
}


/**
 * A component may be specified as elastic in two ways.
 * <ol><LI>It may implement the <code>JCElastic</code> interface.</li>
 * <li>It may have specified an elastic constraint when it was added.</li>
 * </ol>
 * This method checks a particular component for either in the orientataion
 * relevent to the current layout.
 * @return
 * @param c
 */
protected int getElasticity(Component c)
{
    int elasticity = 0;
    JCElastic e = getElastic(c);

    if (e != null) {
        if (orientation == HORIZONTAL) {
            elasticity = e.getHorizontalElasticity();
        }
        else if (orientation == VERTICAL) {
            elasticity = e.getVerticalElasticity();
        }
    }
    return(elasticity);
}


//
// LayoutManager2 Implementation
//

/**
 * Adds a component to the layout manager with the specified constraint.
 * @param comp
 * @param constraints
 */
public void addLayoutComponent(Component comp, Object constraints)
{
    if (constraints != null && !(constraints instanceof JCElastic)) {
        throw new IllegalArgumentException("constraint must be null or an " +
                                           "instance of JCElastic");
    }

    // Store hash of constraint with the component as the key
    constraintMap.put(comp, constraints);
}

/**
 * Returns the maximum layout size of the component.
 * @param parent
 * @return maximum layout size
 */
public Dimension maximumLayoutSize(Container parent)
{
    //
    // Loop through the children requesting maximum sizes
    //
    int n = parent.getComponentCount();
    int width = 0;
    int height = 0;
    for (int i = 0; i < n; i++) {
        Component c = parent.getComponent(i);
        Dimension size = c.getMaximumSize();
        if (orientation == VERTICAL) {
            width = Math.max(width, size.width);
            height += size.height;
        }
        else {
            height = Math.max(height, size.height);
            width += size.width;
        }
    }

    return(new Dimension(width, height));
}

/**
 * Returns the <i>X</i> layout alignment.
 * @return <i>X</i> layout alignment
 * @param target
 */
public float getLayoutAlignmentX(Container target)
{
    return((float)0.0);
}

/**
 * Returns the <i>Y</i> layout alignment.
 * @return <i>Y</i> layout alignment
 * @param target
 */
public float getLayoutAlignmentY(Container target)
{
    return((float)0.0);
}

/**
 * Returns the alignment for the layout.
 * @return alignment for the layout
 */
public int getAlignment()
{
    return(alignment);
}

/**
 * Sets the alignment for the layout.
 * @param alignment
 */
public void setAlignment(int alignment)
{
    this.alignment = alignment;
}

/**
 * Invalidates the current layout: deletes any information
 * that might be cached regarding the current layout.
 * @param target
 */
public void invalidateLayout(Container target)
{
    // PENDING - add layout caching
}

/**
 * Inner class that creates Elastic constraint objects.
 */
static class Elastic implements JCElastic, java.io.Serializable {
    protected int horizontal, vertical;

    public Elastic(int horizontal_elasticity, int vertical_elasticity)
    {
        this.horizontal = horizontal_elasticity;
        this.vertical = vertical_elasticity;
    }

    public int getHorizontalElasticity()
    {
        return(horizontal);
    }

    public int getVerticalElasticity()
    {
        return(vertical);
    }
} // end of innerclass Elastic

}
