/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCFileFilter.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import java.io.File;
import java.util.Hashtable;
import java.util.Enumeration;
import javax.swing.filechooser.*;

/**
 * A convenience implementation of <code>FileFilter</code> that filters out
 * all files except for those type extensions that it knows about.
 * <p>
 * Extensions are of the type <i>.foo</i>, which is typically found on
 * Windows and Unix boxes, but not on Macintosh. Case is ignored.
 * <P>
 * Example - create a new filter that filters out all files
 * but gif and jpg image files:
 *<pre>
 *     JFileChooser chooser = new JFileChooser();
 *     JCFileFilter filter = new JCFileFilter(
 *                   new String{"gif", "jpg"}, "JPEG & GIF Images")
 *     chooser.addChoosableFileFilter(filter);
 *     chooser.showOpenDialog(this);
 * </pre>
 */
public class JCFileFilter extends FileFilter implements java.io.FileFilter {

protected Hashtable  filters = null;
private String       description = null;
private String       fullDescription = null;
private boolean      useExtensionsInDescription = true;

/**
 * Creates a file filter. If no filters are added, then all
 * files are accepted.
 *
 * @see #addExtension
 */
public JCFileFilter()
{
    this.filters = new Hashtable();
}

/**
 * Creates a file filter that accepts files with the given extension.<br>
 * Example: <code>new JCFileFilter("jpg");</code>
 *
 * @see #addExtension
 * @param extension
 */
public JCFileFilter(String extension)
{
    this(extension, null);
}

/**
 * Creates a file filter that accepts the given file type.<br>
 * Example: <code>new JCFileFilter("jpg", "JPEG Image Images");</code>
 * <p>
 * <b>Note:</b> The "." before the extension is not needed. If
 * provided, it will be ignored.
 *
 * @see #addExtension
 * @param extension
 * @param description
 */
public JCFileFilter(String extension, String description)
{
    this();
    if (extension != null) {
        addExtension(extension);
    }
    if (description != null) {
        setDescription(description);
    }
}

/**
 * Creates a file filter from the given String array.<br>
 * Example: <code>new JCFileFilter(String {"gif", "jpg"});</code>
 * <p>
 * <b>Note:</b> The "." before the extension is not needed and
 * will be ignored.
 *
 * @see #addExtension
 * @param filters
 */
public JCFileFilter(String[] filters)
{
    this(filters, null);
}

/**
 * Creates a file filter from the given string array and description.<br>
 * Example: <code>new JCFileFilter(String {"gif", "jpg"}, "Gif and JPG
 * Images");</code>
 * <p>
 * <b>Note:</b> The "." before the extension is not needed and will be ignored.
 *
 * @see #addExtension
 * @param filters
 * @param description
 */
public JCFileFilter(String[] filters, String description)
{
    this();
    for (int i = 0; i < filters.length; i++) {
        // add filters one by one
        addExtension(filters[i]);
    }
    if (description != null) {
        setDescription(description);
    }
}

/**
 * Returns <code>true</code> if this file should be shown in the directory pane;
 * <code>false</code> if it shouldn't.
 * <br>
 * Files that begin with "." are ignored.
 *
 * @see #getExtension
 * @see FileFilter#accept
 * @return
 * @param f
 */
public boolean accept(File f)
{
    if (f != null) {
        if (f.isDirectory()) {
            return(true);
        }
        String extension = getExtension(f);
        if (extension != null && filters.get(getExtension(f)) != null) {
            return(true);
        }
    }
    return(false);
}

/**
 * Returns the extension portion of a String.
 * @return the extension portion of a String
 * @param s
 */
public String getExtension(String s)
{
    if (s != null) {
        int i = s.lastIndexOf('.');
        if ((i >= 0) && (i < s.length() - 1)) {
            return s.substring(i + 1);
        }
        ;
    }
    return(s);
}

/**
 * Returns the extension portion of the file's name.
 *
 * @see #getExtension
 * @see FileFilter#accept
 * @return extension portion of the file's name
 * @param f
 */
public String getExtension(File f)
{
    if (f != null) {
        String filename = f.getName();
        return getExtension(filename).toLowerCase();
    }
    return(null);
}

/**
 * Adds a filetype "dot" extension to filter against.
 * <br>
 * For example: The following code will create a filter that filters
 * out all files except those that end in ".jpg" and ".tif":
 * <pre>
 *   JCFileFilter filter = new JCFileFilter();
 *   filter.addExtension("jpg");
 *   filter.addExtension("tif");
 * </pre>
 * <br><B>Note:</b> The "." before the extension is not needed and will be ignored.
 * @param extension
 */
public void addExtension(String extension)
{
    if (filters == null) {
        filters = new Hashtable(5);
    }
    filters.put(getExtension(extension).toLowerCase(), this);
    fullDescription = null;
}

/**
 * Returns the human readable description of this filter. For
 * example: "JPEG and GIF Image Files (*.jpg, *.gif)"
 *
 * @see #setDescription
 * @see #setExtensionListInDescription
 * @see #isExtensionListInDescription
 * @see FileFilter#getDescription
 * @return
 */
public String getDescription()
{
    if (fullDescription == null) {
        if (description == null || isExtensionListInDescription()) {
            fullDescription = (description == null) ? "(" : description + " (";
            // Build the description from the extension list
            Enumeration extensions = filters.keys();
            if (extensions != null) {
                fullDescription += "*." + (String)extensions.nextElement();
                while (extensions.hasMoreElements()) {
                    fullDescription += ", *." +
                                       (String)extensions.nextElement();
                }
            }
            fullDescription += ")";
        }
        else {
            fullDescription = description;
        }
    }
    return(fullDescription);
}

/**
 * Sets the human readable description of this filter. For
 * example: <code>filter.setDescription("Gif and JPG Images");</code>
 *
 * @see #getDescription
 * @see #setExtensionListInDescription
 * @see #isExtensionListInDescription
 * @param description
 */
public void setDescription(String description)
{
    this.description = description;
    fullDescription = null;
}

/**
 * Determines whether the extension list (.jpg, .gif, etc) should
 * show up in the human readable description.
 * <br>
 * Only relevent if a description was provided in the constructor
 * or using <code>setDescription()</code>.
 *
 * @see #getDescription
 * @see #setDescription
 * @see #isExtensionListInDescription
 * @param b
 */
public void setExtensionListInDescription(boolean b)
{
    useExtensionsInDescription = b;
    fullDescription = null;
}

/**
 * Returns whether the extension list (.jpg, .gif, etc) should
 * show up in the human readable description.
 * <br>
 * Only relevant if a description was provided in the constructor
 * or using <code>setDescription()</code>.
 *
 * @see #getDescription
 * @see #setDescription
 * @see #setExtensionListInDescription
 * @return
 */
public boolean isExtensionListInDescription()
{
    return(useExtensionsInDescription);
}

}
