/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCIconCreator.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:45 $ $Locker:  $  Quest Software Inc.
package com.klg.jclass.util.swing;

import java.awt.*;
import java.awt.image.*;

import javax.swing.ImageIcon;

/**
 * <code>JCIconCreator</code> creates image icons from <code>String</code>
 * arrays. The format is similar to <i>XPixMap</i> (XPM) format files. For
 * example:
 * <pre>
    JCIconCreator im = new JCIconCreator(3, 2);
    im.setColor('G', Color.green);
    im.setColor('x', Color.orange);
    String[] pixels = { " .G", "G x" };
    im.setPixels(pixels);
    ImageIcon icon = im.getIcon();
* </pre>
*/
public class JCIconCreator {

protected ImageIcon  icon = null;
protected int        curRow;
protected int        width, height;
protected int        colorMap[] = new int[256];
protected int        pixels[];

/**
 * Creates an empty instance using the specified width and height.
 * @param w the initial image width
 * @param h the initial image height
 * @see java.awt.Component#createImage
 */
public JCIconCreator(int w, int h)
{
    setSize(w, h);
}

/**
 * Creates an empty instance.
 * @see java.awt.Component#createImage
 */
public JCIconCreator()
{
    // Set spaces to transparent
    colorMap[' '] = 0;
}

/** Clears the data. */
public synchronized void clear()
{
    if (width != 0 && height != 0) {
        pixels = new int[width * height];
    }
    icon = null;
    curRow = 0;
}

/**
 * Gets the <i>Icon</i> created by this instance of <code>JCIconCreator</code>.
 * If the icon does not exist, uses pixels and the <code>colorMap</code>
 * arrays to create it.
 * @see #setPixels
 * @return The icon created by this instance of  <code>JCIconCreator</code>
 */
public synchronized ImageIcon getIcon()
{
    if (pixels == null) {
        icon = null;
        return(null);
    }
    if (icon == null) {
        MemoryImageSource source =
            new MemoryImageSource(width, height, pixels, 0, width);
        Image im = Toolkit.getDefaultToolkit().createImage(source);
        icon = new ImageIcon(im);
    }
    return(icon);
}

/**
 * Sets the color (as an RGB int) corresponding to a character.
 * Spaces are transparent by default.
 * @param ch The character which represents the color
 * @param rgb The rgb value of the color
 */
public synchronized void setColor(char ch, int rgb)
{
    if (ch < 256) {
        colorMap[ch] = rgb;
        icon = null;
    }
}

/**
 * Sets the color corresponding to a character.
 * Spaces are transparent by default.
 * @param ch The character which represents the color
 * @param color The color object
 */
public synchronized void setColor(char ch, Color color)
{
    setColor(ch, (color == null ? 0 : color.getRGB()));
}

/**
 * Sets the size of the image to be created.
 * @param w The new width
 * @param h The new height
 */
public synchronized void setSize(int w, int h)
{
    if (w == width && h == height) {
        return;
    }
    width  = w;
    height = h;
    if (w == 0 && h == 0) {
        pixels = null;
    }
    else {
        pixels = new int[w * h];
    }
    icon = null;
}

/**
 * Sets the pixel data for all rows.
 * @param string A list of strings representing pixel data for all rows
 */
public synchronized void setPixels(String[] string)
{
    clear();
    for (int i = 0; i < string.length && i < height; i++) {
        setPixels(curRow++, string[i]);
    }
    icon = null;
}

/** Sets the pixel data at the specified row.
 * @param row The row index
 * @param string The new pixel data at that row
 */
public synchronized void setPixels(int row, String string)
{
    icon = null;
    if (string.length() != width) {
        return;
    }
    for (int j = 0, k = row * width; j < width; j++) {
        int c = colorMap[string.charAt(j)];
        pixels[j + k] = c;
    }
}

/**
 * This method takes a passed-in byte array (as might be obtained from a
 * database's image field) and attempts to convert it into an
 * <code>Image</code>.
 * @param bytes
 * @return
 */
public static ImageIcon getIcon(byte[] bytes)
{
    if (bytes == null) {
        return(null);
    }
    ImageIcon ii = new ImageIcon(bytes);
    return(ii);
}

}
