/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCProgressHelper.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import java.awt.*;
import javax.swing.*;
import java.beans.*;
import java.util.Enumeration;
import com.klg.jclass.util.progress.*;
import com.klg.jclass.util.JCListenerList;

import java.lang.reflect.InvocationTargetException;

/**
 * This is a class for creating and managing a progress dialog
 * in a Thread-safe manner.
 * An instance of this class monitors the progress of some operation.
 * <P>
 * Call <code>startProgress()</code>
 * when you want to start the progress meter.
 * The two parameters, <code>popupTime</code> and <code>decideTime</code>,
 * give you complete control over when the dialog appears.
 * Call <code>updateProgress()</code>
 * whenever you want to update the progress meter.
 * <P>
 * Normally, <code>decideTime</code> is less than
 * <code>popupTime</code>.
 * If the progress monitor is not yet being displayed,
 * a call to <code>updateProgress()</code>
 * causes the
 * progress meter to compare the
 * <code>decideTime</code> property (default 500 ms, or
 * set using <code>setTimeToDecideToPopup()</code>)
 * to the elapsesd time since the monitor was started.
 * If the elapsed time is greater than <code>decideTime</code>,
 * the progress monitor predicts how
 * long the entire operation will take.
 * If it is longer than <code>popupTime</code>
 * (default: 2 seconds,
 * or set with <code>setTimeToPopup()</code>), a
 * <code>JCProgressHelper</code> will be popped up.
 * <P>
 * To force the dialog to appear,
 * set <code>popupTime</code>to 0.
 * <P>
 * Before the progress is started the
 * <code>JCProgressHelper</code> should be given
 * a numeric range and a descriptive String,
 * either by setting them in the constructor
 * or by calling the appropriate methods.  As the operation progresses,
 * call the <code>updateProgress()</code> method periodically
 * to indicate how far along the [<i>min, max</i>] range
 * the operation is.
 *
 */
public class JCProgressHelper {

protected Component       parent;
protected String          staticMessage;
protected String          dynamicMessage = " ";
protected String          dialogTitle = "Progress...";
protected String          cancelString = "Cancel";
protected JProgressBar    progressbar;
protected JDialog         dialog;
protected String          staticMessageLabel;
protected JOptionPane     optionPane;
protected JLabel          dynamicMessageLabel;
protected boolean         isModal;
protected boolean         isDismissable;
protected boolean         showDynamicMessage;
protected int             max;
protected int             min;
protected int             value;
protected long            decideTime = 500;
protected long            popupTime = 2000;
protected long            startTime;
protected boolean         popupInitiated = false;
protected boolean         progressComplete = false;
protected boolean         firstPass = true;
protected JCListenerList  progressCancelledListeners = null;

/**
 * Creates a <code>JCProgressHelper</code> with default characteristics.
 * <P>
 * Defaults:
 * <pre>
 * static_message = null
 * min = 0
 * max = 0
 * show_dynamic_message = true
 * is_modal = false
 * is_dismissable = false
 * </pre>
 *
 * @param parent the <code>Component</code> that is the parent of the
 *     <code>JCProgressHelper</code>
 */
public JCProgressHelper(Component parent)
{
    this(parent, null, 0, 0, true, false, false);
}

/**
 * Creates a <code>JCProgressHelper</code> with default characteristics.
 * <P>
 * Defaults:
 * <pre>
 * show_dynamic_message = true
 * is_modal = false
 * is_dismissable = false
 * </pre>
 *
 * @param parent the <code>Component</code> that is the parent of the
 *     <code>JCProgressHelper</code>
 * @param static_message the <code>String</code> that is the text of the
 *     message shown
 * @param min the <code>int</code> denoting the minimum value of the
 *     ProgressBar
 * @param max the <code>int</code> denoting the maximum value of the
 *     ProgressBar
 */
public JCProgressHelper(Component parent, String static_message,
                        int min, int max)
{
    //this(parent, static_message, min, max, true, true, false);
    this(parent, static_message, min, max, true, false, false);
}

/**
 * Creates a <code>JCProgressHelper</code> with no default settings.
 * All settings are set using the arguments passed to this constructor.
 *
 * @param parent the <code>Component</code> that is the parent of the
 *     <code>JCProgressHelper</code>
 * @param static_message the <code>String</code> that is the text of the
 *     message shown in  the progress bar
 * @param min the <code>int</code> that is the minimum value of the
 *     ProgressBar
 * @param max the <code>int</code> that is the maximum value of the
 *     ProgressBar
 * @param show_dynamic_message the boolean that tells
 *     <code>JCProgressHelper</code> whether or not to display dynamic messages
 * @param is_modal the boolean that tells <code>JCProgressHelper</code>
 *     whether it is modal
 * @param is_dismissable the boolean that tells <code>JCProgressHelper</code>
 *      whether or  not the user can cancel it
 */
public JCProgressHelper(Component parent, String static_message,
                        int min, int max,
                        boolean show_dynamic_message,
                        boolean is_modal, boolean is_dismissable)
{
    this.parent = parent;
    this.staticMessage = static_message;
    this.max = max;
    this.min = min;
    this.value = min;
    this.showDynamicMessage = show_dynamic_message;
    this.isModal = is_modal;
    this.isDismissable = is_dismissable;
}

/**
 * Starts the progress meter.
 * The initial length of the bar, showing the amount completed,
 * is determined from the proportion
 * (<code>value</code> - <code>min</code>) /
 * (</code><code>max</code> - <code>min</code>).
 * <br>If <code>value</code> = <code>min</code>, the progress bar
 * shows 0% complete.
 */
public void startProgress()
{
    if (progressComplete == false && firstPass == false) {
        // Not the first pass, and the current progress is not complete.
        // the progress should have been ended by a call to processingEnd()
        // assume this was intended and end the current progress.
        completeProgress();
    }

    // Reset progress flags to start progress over.
    progressComplete = false;
    popupInitiated = false;

    startTime = System.currentTimeMillis();

    // Launch a thread that will popup a dialog after "popupTime" has
    // ellapsed if it has not already been popped up or completed
    // Sleeper is an instance of JCRunnable so it automatically launches
    // itself
    Runnable tmp = new Runnable() {
        public void run()
        {
            boolean interrupted = false;
            // Fix for bug 13336
            // Sleep will not be called if popupTime is not > 0
            if (popupTime > 0) {
                try {
                    Thread.sleep(popupTime);
                }
                catch (InterruptedException e) {
                    interrupted = true;
                }
            }

            if (interrupted == false &&
                popupInitiated == false && progressComplete == false) {
                // Only try to popup the dialog if it's not already
                // popped up (or on its way up) and if that progress
                // is not complete
                popupDialog();
            }
        }
    };
    Thread thread = new Thread(tmp);
    thread.start();
}

/**
 * Convenience method for adding a <code>JCProgressCancelledListener</code>.
 *
 * @param   l  <code>JCProgressCancelledListener</code>
 */
public void addProgressCancelledListener(JCProgressCancelledListener l)
{
    progressCancelledListeners =
        JCListenerList.add(progressCancelledListeners, l);
}

/**
 * Convenience method for removing a <code>JCProgressCancelledListener</code>.
 *
 * @param   l  <code>JCProgressCancelledListener</code>
 */
public void removeProgressCancelledListener(JCProgressCancelledListener l)
{
    progressCancelledListeners =
        JCListenerList.remove(progressCancelledListeners, l);
}

/**
 * Fires an event to tell interested listeners that the user has cancelled progress.
 *
 * @param   event  <code>JCProgressCancelledEvent</code>
 */
protected void fireUserCancelledProgressEvent(JCProgressCancelledEvent event)
{
    Enumeration e = JCListenerList.elements(progressCancelledListeners);
    while (e.hasMoreElements()) {
        JCProgressCancelledListener listener =
            (JCProgressCancelledListener)e.nextElement();
        listener.userCancelledProgress(event);
    }
}


/**
 * Called internally, creates and displays the dialog.
 */
protected void popupDialog()
{
    popupInitiated = true;
    if (progressComplete == true) {
        return;
    }
//	JCSwingRunnable tmp = new JCSwingRunnable() {
    Runnable tmp = new Runnable() {
        public void run()
        {
            // We lock the containing object to ensure that no one
            // complete's the operation while we are setting it up!
            synchronized (JCProgressHelper.this) {
                if (progressComplete == true) {
                    // progress is already completed so we bail!
                    return;
                }
                int count = 1;
                if (JCProgressHelper.this.max != -1) {
                    progressbar = new JProgressBar();
                    progressbar.setMaximum(JCProgressHelper.this.max);
                    progressbar.setMinimum(JCProgressHelper.this.min);
                    progressbar.setValue(JCProgressHelper.this.min);
                    count++;
                }
                if (JCProgressHelper.this.showDynamicMessage) {
                    dynamicMessageLabel = new JLabel(" ");
                    count++;
                }
                Object list[] = new Object[count];
                // reset count
                count = 0;
                list[count++] = JCProgressHelper.this.staticMessage;
                if (dynamicMessageLabel != null) {
                    list[count++] = dynamicMessageLabel;
                }
                if (progressbar != null) {
                    list[count++] = progressbar;
                }

                optionPane = new ProgressPane(list, cancelString);
                dialog = optionPane.createDialog(JCProgressHelper.this.parent,
                                                 dialogTitle);
                dialog.setModal(JCProgressHelper.this.isModal);
                //fire the userCancelled event when called.
                optionPane.addPropertyChangeListener(new PropertyChangeListener()
                {
                    boolean once = true;

                    public void propertyChange(PropertyChangeEvent e)
                    {
                        if (once) {
                            JCProgressCancelledEvent event =
                                new JCProgressCancelledEvent(e.getSource());
                            fireUserCancelledProgressEvent(event);
                        }
                        once = false;
                    }
                });

                dialog.show();
            }
        }
    };
    SwingUtilities.invokeLater(tmp);
}


/**
 * Sets the <code>decideTime</code> property.
 *
 * @param time the <code>long</code> value that is compared to the
 *     <code>popupTime</code> to determine whether or not to pop up the dialog
 */
public void setTimeToDecideToPopup(long time)
{
    decideTime = time;
}

/**
 * Sets the <code>popupTime</code> property. If this value is &#60;1, the
 * thread which pauses popping up the dialog in <code>startProgress()</code> is
 * never called, and the dialog pops up immediately.
 *
 * @param time the <code>long</code> value that gives the operation a metric
 *     for deciding whether or not the <code>JCProgressHelper</code> is
 *     required.
 */
public void setTimeToPopup(long time)
{
    popupTime = time;
}

/**
 * Sets the <code>staticMessage</code> property.
 * In the dialog, two areas are reserved for message lines.
 * The static message is on the upper line.
 *
 * @param message the <code>String</code> that is displayed throughout the
 *     life of the <code>JCProgressHelper</code>
 */
public void setStaticMessage(String message)
{
    staticMessage = message;
}


/**
 * Returns the <code>String</code> which displays in the dialog's title bar.
 *
 * @return dialogTitle the String that shows in the title bar of the dialog
 */
public String getDialogTitle()
{
    return(this.dialogTitle);
}


/**
 * Sets the <code>String</code> which displays in the dialog's title bar.
 * The default is "Progress..."
 *
 * @param dialogTitle the String that shows in the title bar of the dialog
 */
public void setDialogTitle(String dialogTitle)
{
    this.dialogTitle = dialogTitle;
}

/**
 * Returns the <code>String</code> which displays in the dialog's cancel button.
 *
 * @return dialogTitle the String that shows in the title bar of the dialog
 */
public String getCancelString()
{
    return(this.cancelString);
}

/**
 * Sets the <code>String</code> which displays in the dialog's cancel button.
 * The default is "Cancel".
 *
 * @param cancelString
 */
public void setCancelString(String cancelString)
{
    this.cancelString = cancelString;
}


/**
 * Sets the </code>max</code> property.
 * This variable represents the highest value
 * that may be passed to <code>updateProgess()</code>.
 *
 * @param max the <code>int</code> that is the maximum value of the progress
 *     bar
 */
public void setMaximum(int max)
{
    this.max = max;
//	JCSwingRunnable tmp = new JCSwingRunnable() {
    Runnable tmp = new Runnable() {
        public void run()
        {
            if (progressbar != null) {
                progressbar.setMaximum(JCProgressHelper.this.max);
            }
        }
    };
    SwingUtilities.invokeLater(tmp);
}

/**
 * Sets the <code>min</code> property.
 * This is the value used to indicate
 * the proportion of the task that has
 * been completed when the progress
 * meter first pops up.
 * Typically, this value is zero.
 *
 * @param min the <code>int</code> that is the minimum value of the progress
 *     bar
 */
public void setMinimum(int min)
{
    this.min = min;
//	JCSwingRunnable tmp = new JCSwingRunnable() {
    Runnable tmp = new Runnable() {
        public void run()
        {
            if (progressbar != null) {
                progressbar.setMinimum(JCProgressHelper.this.min);
            }
        }
    };
    SwingUtilities.invokeLater(tmp);
}

/**
 * Sets the <code>min</code> and <code>max</code> properties.
 *
 * @param min the <code>int</code> that is the minimum value of the progress
 *     bar
 * @param max the <code>int</code> that is the maximum value of the progress
 *     bar
 */
public void setRange(int min, int max)
{
    this.min = min;
    this.max = max;
//	JCSwingRunnable tmp = new JCSwingRunnable() {
    Runnable tmp = new Runnable() {
        public void run()
        {
            if (progressbar != null) {
                progressbar.setMinimum(JCProgressHelper.this.min);
                progressbar.setMaximum(JCProgressHelper.this.max);
            }
        }
    };
    SwingUtilities.invokeLater(tmp);
}

/**
 * Updates the value in the progress bar.
 *
 * @param new_value the <code>int</code> that is the new value displayed in
 *     the progress bar as a length
 */
public void updateProgress(int new_value)
{
    value = new_value;

    if (popupInitiated == false) {
        long elapsed_time = System.currentTimeMillis() - startTime;
        // if we are equal to or passed the decide time then we need to
        // guess at how long it's going to take to get to the end
        // If the expected time is longer than the popupTime we popup the
        // dialog
        if (elapsed_time >= decideTime) {
            long expected_time = Integer.MAX_VALUE; // as far as we know
            // its gonna take forever
            if (value != min) {
                // we have made some progress so extrapolate from that
                // to come at as good a guess as we can as to how long its
                // going to take
                expected_time = elapsed_time / (value - min) * (max - min);
            }
            if (expected_time > popupTime) {
                popupDialog();
            }
            else {
                return;
            }
        }
        else {
            return;
        }
    }

//	JCSwingRunnable tmp = new JCSwingRunnable(new Integer(new_value)) {
    Runnable tmp = new Runnable() {
        int val = value;
        public void run()
        {
            if (progressbar != null) {
                progressbar.setValue(val);
            }
        }
    };

    //Allow the event thread to update the progress bar before continuing
    try {
        SwingUtilities.invokeAndWait(tmp);
    }
    catch (InterruptedException ie) {}
    catch (InvocationTargetException ite) {}
}

/**
 * Closes and disposes the <code>JCProgressHelper</code>.
 */
public void completeProgress()
{
    progressComplete = true;
    popupInitiated = false;

//	JCSwingRunnable tmp = new JCSwingRunnable () {
    Runnable tmp = new Runnable() {
        public void run()
        {
            if (dialog != null) {
                // invoking setInputValue() causes the
                // dialog to popdown and get recycled in the
                // strange way that Swing does things (prior to jdk 1.4.0)
//				optionPane.setInputValue(new Integer(optionPane.OK_OPTION));
                // In jdk 1.4.0 and later, need to explicitly close the dialog.
                dialog.dispose();
                dialog = null;
                progressbar = null;
                dynamicMessageLabel = null;
            }
        }
    };
    SwingUtilities.invokeLater(tmp);
}

/**
 * Returns <code>true</code> as long as the user has not cancelled the
 * progress by hitting the cancel button in the <code>OptionPane</code>.
 *
 * @return the <code>boolean</code> representing whether or not the user has
 *     cancelled the dialog
 */
public boolean isOkayToContinue()
{
    if (optionPane == null) {
        return(true);
    }
    Object v = optionPane.getValue();
    return(v == optionPane.UNINITIALIZED_VALUE);
}

/**
 * Sets the <code>dynamicMessage</code> property.
 * Dynamic messages are the ones which
 * show up just above the progress bar
 * and just below the static message,
 * usually containing text information
 * on the current progress of the operation,
 * such as naming the sub-task currently being performed.
 *
 * @param dynamic_message the String that announces what the status of the
 *     operation is
 */
public void setDynamicMessage(String dynamic_message)
{
    if (dynamicMessage == dynamic_message) {
        return;
    }
    if (dynamic_message == null) {
        // never trust swing to handle a null
        this.dynamicMessage = "";
    }
    else {
        this.dynamicMessage = dynamic_message;
    }

    if (dynamicMessageLabel == null) {
        return;
    }

//	JCSwingRunnable tmp = new JCSwingRunnable () {
    Runnable tmp = new Runnable() {
       public void run()
        {
            if (dynamicMessageLabel != null) {
                dynamicMessageLabel.setText(dynamicMessage);
            }
        }
    };
    SwingUtilities.invokeLater(tmp);
}

/**
 * Inner class that extends <code>JOptionPane</code>, and creates
 * some default values for the dialog that <code>JCProgressHelper</code>
 * displays.
 */
class ProgressPane extends JOptionPane {
    /**          *
     * @param messageList
     * @param cancelString
     */
    ProgressPane(Object messageList, String cancelString)
    {
        super(messageList, JOptionPane.INFORMATION_MESSAGE,
              0,
              null,
              new Object[]{cancelString},
              null);
        setUI(new ProgressPaneUI());
    }

    /**          */
    public int getMaxCharactersPerLineCount()
    {
        return(60);
    }
}

/**
 * Inner class that extends <code>BasicOptionPaneUI</code>, and
 * overrides <code>getButtons()</code> if <code>!isDismissable</code>.
 */
class ProgressPaneUI extends javax.swing.plaf.basic.BasicOptionPaneUI {

    /**
     * Returns the buttons, if any, to the <code>ButtonAreaLayout</code>.
     * Overridden to handle the case where the <code>JCProgressHelper</code> is
     * not dismissable.
     * @return Object[] the buttons which will appear on the option pane.
     */
    protected Object[] getButtons()
    {
        if (JCProgressHelper.this.isDismissable) {
            return(super.getButtons());
        }
        else {
            return(null);
        }
    }
}

/**
 * Returns a listener that uses this <code>JCProgressHelper</code>
 * that is suitable for attaching to any Bean that implements an
 * <code>addProgressListener(JCProgressListener)</code> method.
 *
 * @see com.klg.jclass.util.progress.JCProgressListener
 * @return listener that uses this <code>JCProgressHelper</code>
 */
public JCProgressListener createProgressListener()
{
    return(new ProgressListener());
}

/**
 * Inner class that implements <code>JCProgressListener</code>, using the
 * containing <code>JCProgressHelper</code> object as the view.
 */
class ProgressListener implements JCProgressListener {
    /**
     * Invoked when a process has begun.
     *
     * @see com.klg.jclass.util.progress.JCProgressEvent
     * @param e
     */
    public void processingBegin(JCProgressEvent e)
    {
        setStaticMessage(e.getProcessName());
        setMinimum(0);
        setMaximum(e.getUnitCount());
        startProgress();
        
    }

    /**
     * Invoked when a process unit has been completed.
     *
     * @see com.klg.jclass.util.progress.JCProgressEvent
     * @param e
     */
    public void processingUnit(JCProgressEvent e)
    {
        if (isOkayToContinue() == false) {
            e.setAbort(true);
        }
        setDynamicMessage(e.getUnitName());
        updateProgress(e.getCurrentUnit());
        
    }

    /**
     * Invoked when the process has been compeleted.
     *
     * @see com.klg.jclass.util.progress.JCProgressEvent
     * @param e
     */
    public void processingEnd(JCProgressEvent e)
    {
        
        completeProgress();
    }

    /**
     * Invoked when a process encounters an error.
     *
     * @see com.klg.jclass.util.progress.JCProgressEvent
     * @param e
     */
    public void processingError(JCProgressEvent e)
    {
        
    }
}

}
