/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCRectUtil.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:45 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import java.awt.Rectangle;

import java.util.Vector;
import java.util.Collections;
import java.util.Comparator;

/**
 * <code>JCRectUtil</code> provides functions for computing the largest
 * available rectangle from a list of rectangles.<br>
 * Here is a brief summary of how to use it:<ol>
 *    <li> Initialize with a rectangle by calling <code>rlInit(r, v)</code>,
 *        where <i>r</i> is the initializing rectangle and <i>v</i> is a blank
 *        vector.
 *    <li> Call <code>rlRemove(r, v1, v2)</code> to remove sub-rectangle
 *        <i>r</i> from the rectangles in the <code>Vector</code> of rectangles
 *        <i>v1</i>, placing the new list of rectangles in <code>Vector</code>
 *        <i>v2</i>.  Repeat as necessary.
 *    <li> Call <code>largestRect(v)</code> to return the largest available
 *        rectangle from the passed-in<code>Vector</code> of rectangles.
 * </ol>
 */
public class JCRectUtil {

public final static int LEFTTORIGHT = 0;
public final static int TOPTOBOTTOM = 1;

/**
 * Removes any references to negative widths and/or heights, by
 * adjusting the <i>x</i>, and <i>y</i> positions of the rectangles.
 * @param rl the rectangle list, in vector form, to normalize
 */
static void normalize(Vector rl)
{
    for (int i = 0; i < rl.size(); i++) {
		Rectangle rect = (Rectangle) rl.elementAt(i);
        if (rect.width < 0) {
            rect.x += rect.width;
            rect.width = -rect.width;
        }

        if (rect.height < 0) {
            rect.y += rect.height;
            rect.height = -rect.height;
        }
    }
}

/**
 * Constructs the array of <code>Vector</code> objects, where the elements of
 * each <code>Vector</code> have the same <i>y</i> position.
 * @param rl the seed list, a vector
 * @return the resulting array of vectors
 */
static Vector[] make(Vector rl)
{
    Vector[] jc_rl = null;
    if (rl.size() == 0) {
        return(jc_rl);
    }

    int curr_y = ((Rectangle)rl.elementAt(0)).y;

    int i = 1;
    for (int j = 1; j < rl.size(); j++) {
        Rectangle rect = (Rectangle)rl.elementAt(j);
        if (rect.y != curr_y) {
            curr_y = rect.y;
            i++;
        }
    }

    int nrl = i + 1;
    jc_rl = new Vector[nrl];
    if (jc_rl == null) {
        return(jc_rl);
    }

    for (i = 0; i < nrl; i++) {
        jc_rl[i] = new Vector();
    }

    /*
     *  jc_rl is an array of Vectors where each Vector has
     *  nodes which have the same y value.
     */
    int rl_ind = 0;
    Vector jc_rln = jc_rl[rl_ind];
    Rectangle rect = (Rectangle)rl.elementAt(0);
    rlInit(rect, jc_rln);
    curr_y = rect.y;
    for (int j = 1; j < rl.size(); j++) {
        Rectangle thisRect = (Rectangle)rl.elementAt(j);
        if (thisRect.y == curr_y) {
            jc_rln.addElement(new Rectangle(thisRect));
        }
        else {
            Collections.sort(jc_rln, new RectSortLR());
            rl_ind++;
            jc_rln = jc_rl[rl_ind];
            curr_y = thisRect.y;
            rlInit(thisRect, jc_rln);
        }
    }
    Collections.sort(jc_rln, new RectSortLR());
    return(jc_rl);
}

/**
 *  The metric for the 'largest' rectangle is just the area.
 *  @param r a <code>PointRect</code> object containg the rectangle to calculate
 *  the metric for
 *  @return the metric (for example, the area of the supplied rectangle)
 */
static double lgMetric(PointRect r)
{
    return (((double)(r.x2 - r.x1)) * (r.y2 - r.y1));
}

/**
 * Checks to see if the union of the next rectangle in the
 * search path, <i>r</i>, and <i>tmp_rect</i> is larger than the largest
 * rectangle so far.
 * @param lg_rect the largest rectangle so far
 * @param tmp_rect the current union of rectangles given a starting rectangle
 * @param r the next rectangle in the search path
 */
static void calcLgRect(Rectangle lg_rect, Rectangle tmp_rect, Rectangle r)
{

    PointRect r1 = new PointRect();
    PointRect r2 = new PointRect();
    PointRect r3 = new PointRect();
    PointRect tmp_r = new PointRect();
    double r1_lg_metric, lg_rect_metric, tmp_r_lg_metric;

    r1.x1 = tmp_rect.x;
    r1.x2 = tmp_rect.width + tmp_rect.x - 1;
    r1.y1 = tmp_rect.y;
    r1.y2 = tmp_rect.height + tmp_rect.y - 1;

    r2.x1 = r.x;
    r2.x2 = r.width + r.x - 1;
    r2.y1 = r.y;
    r2.y2 = r.height + r.y - 1;

    r3.x1 = lg_rect.x;
    r3.x2 = lg_rect.width + lg_rect.x - 1;
    r3.y1 = lg_rect.y;
    r3.y2 = lg_rect.height + lg_rect.y - 1;

    lg_rect_metric = lgMetric(r3);

    if ((r1.x1 == r2.x1) &&
        (r1.x2 == r2.x2) &&
        (r1.y1 == r2.y1) &&
        (r1.y2 == r2.y2))
    {
        r1_lg_metric = lgMetric(r1);
        if (r1_lg_metric > lg_rect_metric) {
            lg_rect.x = tmp_rect.x;
            lg_rect.y = tmp_rect.y;
            lg_rect.width = tmp_rect.width;
            lg_rect.height = tmp_rect.height;
        }
    }
    else if (Math.abs(r1.y2 - r2.y1) == 1) {
        /*
         *  Case where the two rectangles are sitting one on top of the other.
         */
        tmp_r.x1 = Math.max(r1.x1, r2.x1);
        tmp_r.x2 = Math.min(r1.x2, r2.x2);
        tmp_r.y1 = r1.y1;
        tmp_r.y2 = r2.y2;

        if (tmp_r.x1 <= tmp_r.x2) {
            tmp_rect.x = tmp_r.x1;
            tmp_rect.y = tmp_r.y1;
            tmp_rect.width = tmp_r.x2 - tmp_r.x1 + 1;
            tmp_rect.height = tmp_r.y2 - tmp_r.y1 + 1;

            tmp_r_lg_metric = lgMetric(tmp_r);
            if (tmp_r_lg_metric > lg_rect_metric) {
                lg_rect.x = tmp_rect.x;
                lg_rect.y = tmp_rect.y;
                lg_rect.width = tmp_rect.width;
                lg_rect.height = tmp_rect.height;
            }
        }
    }
    else if (Math.abs(r1.x2 - r2.x1) == 1) {
        /*
         *  Case where the two rectangles are sitting side by side.
        */
        tmp_r.x1 = r1.x1;
        tmp_r.x2 = r2.x2;
        tmp_r.y1 = Math.max(r1.y1, r2.y1);
        tmp_r.y2 = Math.min(r1.y2, r2.y2);

        if (tmp_r.y1 <= tmp_r.y2) {
            tmp_rect.x = tmp_r.x1;
            tmp_rect.y = tmp_r.y1;
            tmp_rect.width = tmp_r.x2 - tmp_r.x1 + 1;
            tmp_rect.height = tmp_r.y2 - tmp_r.y1 + 1;

            tmp_r_lg_metric = lgMetric(tmp_r);
            if (tmp_r_lg_metric > lg_rect_metric) {
                lg_rect.x = tmp_rect.x;
                lg_rect.y = tmp_rect.y;
                lg_rect.width = tmp_rect.width;
                lg_rect.height = tmp_rect.height;
            }
        }
    }
}

/*
*  Performs a recursive search for the largest rectangle
*  given a start node in the <code>Vector</code>.  'Grow' the start
*  rectangle by taking its union with rectangles to the right and
*  below it.
*  @param jc_rl array of Vectors
*  @param ind index of current Vector in array
*  @param rn start node
*  @param lg_rect largest rect
*  @param tmp_rect
*/
static void searchLgRect(Vector[] jc_rl, int ind, Vector v, int rn,
                         Rectangle lg_rect, Rectangle tmp_rect)
{
    Vector jc_rln = jc_rl[ind];
    Rectangle local_tmp_rect;
    int rn_n;
    Vector jc_rln_n;

    if (jc_rln == null) {
        return;
    }

    for (rn_n = rn; rn_n < v.size(); rn_n++) {
        calcLgRect(lg_rect, tmp_rect, (Rectangle)v.elementAt(rn_n));
    }
    int i;
    for (i = ind + 1, jc_rln_n = jc_rl[i]; jc_rln_n.size() > 0;
         i++, jc_rln_n = jc_rl[i]) {
        for (int j = 0; j < jc_rln_n.size(); j++) {
            local_tmp_rect = new Rectangle(tmp_rect.x, tmp_rect.y,
                                           tmp_rect.width, tmp_rect.height);
            calcLgRect(lg_rect, local_tmp_rect,
                       (Rectangle)jc_rln_n.elementAt(j));
            searchLgRect(jc_rl, i, jc_rln_n, j, lg_rect, local_tmp_rect);
        }
    }
}


/**
 *  Appends rectangle <i>r2</i> to rectlist <i>rl</i>, after
 *  intersecting it with rectangle <i>r1</i>.
 *  @param r1 rectangle to intersect with
 *  @param r2 rectangle to append after intersection
 *  @param rl rectangle list, in vector form
 */
static void appendremaining(Rectangle r1, Rectangle r2, Vector rl)
{
    if (r1.contains(r2)) {
        return;
    }

    if (!r1.intersects(r2)) {
        rl.addElement(new Rectangle(r2));
        return;
    }

    /*
     *  <i>r2</i> intersects <i>r1</i>, but is not contained inside
     *  <i>r1</i>; depending on where it lies, add any new
     *  rectangle created to the rectlist.
     */
    Rectangle r    = new Rectangle(r2.x, r2.y, r2.width, r2.height);
    Rectangle newr = new Rectangle();

    int height = r1.y - r.y;
    if (height > 0) {
        /* Portion above r1 */
        newr.x = r.x;
        newr.y = r.y;
        newr.width = r.width;
        newr.height = height;
        rl.addElement(new Rectangle(newr));
        r.y += height;
        r.height -= height;
    }

    int top = r1.y + r1.height;
    height = r.y + r.height - top;
    if (height > 0) {
        /* Portion below r1 */
        newr.x = r.x;
        newr.y = top;
        newr.width = r.width;
        newr.height = height;
        rl.addElement(new Rectangle(newr));
        r.height -= height;
    }

    int width = r1.x - r.x;
    if (width > 0) {
        /* Portion left of r1 */
        newr.x = r.x;
        newr.y = r.y;
        newr.width = width;
        newr.height = r.height;
        rl.addElement(new Rectangle(newr));
        r.x += width;
        r.width -= width;
    }

    int left = r1.x + r1.width;
    width = r.x + r.width - left;
    if (width > 0) {
        /* Portion right of r1 */
        newr.x = left;
        newr.y = r.y;
        newr.width = width;
        newr.height = r.height;
        rl.addElement(new Rectangle(newr));
        r.width -= width;
    }
}

/*
*  Calculates the bounding rectangle of r1 and r2.
*/
static Rectangle bounding(Rectangle r1, Rectangle r2)
{
    Rectangle r = new Rectangle();
    if (r1.isEmpty()) {
        r = r2;
    }
    else if (r2.isEmpty()) {
        r = r1;
    }
    else {
        r.x = Math.min(r1.x, r2.x);
        r.y = Math.min(r1.y, r2.y);
        r.width = Math.max(r1.x + r1.width,
                           r2.x + r2.width) - r.x;
        r.height = Math.max(r1.y + r1.height,
                            r2.y + r2.height) - r.y;
    }
    return(r);
}

/**
 *  Initializes this rectangle list.
 *  @param r rectangle to intialize this list with
 *  @param rl rectangle list, a vector, to initialize
 * @since JClass 4.0K for Java 2. If you are using JClass 4.0J,
 * which is based on JDK 1.1, this method's parameter must be
 * a <code>com.sun.java.util.collections.Vector</code>.
 */
public static void rlInit(Rectangle r, Vector rl)
{
    rl.clear();
    rl.addElement(new Rectangle(r));
}

/**
 *  Removes rectangle r from rectlist rl1,
 *  storing the result in rl2.
 *  @param r rectangle to remove
 *  @param rl1 rectangle list, a vector, to remove rectangle from
 *  @param rl2 rectangle list, a vector, to place result in
 */
public static void rlRemove(Rectangle r, Vector rl1, Vector rl2)
{
    Vector rl;

    rl = new Vector();
    for (int j = 0; j < rl1.size(); j++) {
        appendremaining(r, (Rectangle)rl1.elementAt(j), rl);
    }
    rl2.clear();
    rl2.addAll(rl);
}

/**
 * Determines the largest rectangle in the space represented by
 * a list of non-overlapping rectangles.
 * @param rl the list of rectangles, in vector form,
 *           to use for the calculation
 * @return the largest rectangle in the list
 */
public static Rectangle largestRect(Vector rl)
{
    normalize(rl);
    Collections.sort(rl, new RectSortTB());
    Vector[] jc_rl = make(rl);

    /*
     * Traverse the Vector structure in a left-to-right
     * top-to-bottom order.  Use each node as the starting
     * node in the search algorithm to determine the largest
     * rectangle in the rectlist.
     */
    Rectangle headrect = (Rectangle)jc_rl[0].elementAt(0);
    Rectangle lg_rect = new Rectangle(headrect);
    int i;
    Vector jc_rln;
    for (i = 0, jc_rln = jc_rl[i];
         jc_rln.size() > 0;
         i++, jc_rln = jc_rl[i])
    {
        for (int j = 0; j < jc_rln.size(); j++) {
            Rectangle tmp_rect = new Rectangle((Rectangle)jc_rln.elementAt(j));
            searchLgRect(jc_rl, i, jc_rln, j, lg_rect, tmp_rect);
        }
    }
    return(lg_rect);
}


/**
 * Used by <code>JCRectUtil</code> to sort a list of rectangles in
 * left-to-right order.
 */
static class RectSortLR implements Comparator {

    /**
     * Compare Rectangle objects for left-to-right sort.
     * @param o1
     * @param o2
     * @return
     */
    public int compare(Object o1, Object o2)
    {
        /*
         *  Used by qsort to sort rectnodes in
         *  left-to-right order.
         */
        Rectangle rn1 = (Rectangle)o1;
        Rectangle rn2 = (Rectangle)o2;

        int diff = rn1.x - rn2.x;
        if (diff > 0) {
            return (1);
        }
        else if (diff < 0) {
            return (-1);
        }
        else {
            return (0);
        }
    }

}

/**
 * Used by <code>JCRectUtil</code> to sort a list of rectangles in
 * top-to-bottom order.
 */
static class RectSortTB implements Comparator
{
    /**
     * Compares <code>Rectangle</code> objects for a top-to-bottom sort.
     * @param o1
     * @param o2
     * @return
     */
    public int compare(Object o1, Object o2)
    {
        /*
         *  Used to sort rectnodes in top-to-bottom order.
         */
        Rectangle rn1 = (Rectangle)o1;
        Rectangle rn2 = (Rectangle)o2;

        int diff = rn1.y - rn2.y;
        if (diff > 0) {
            return(1);
        }
        else if (diff < 0) {
            return(-1);
        }
        else {
            return(0);
        }
    }
}

/**
 * Used by <code>JCRectUtil</code> to hold a point representation of a
 * rectangle.
 */
static class PointRect
{
    int x1 = 0; // Upper left-hand corner
    int y1 = 0;
    int x2 = 0; // Opposite corner
    int y2 = 0;
}

}
