/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCSpinNumberBox.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:46 $  $Locker:  $  Quest Software Inc.
package com.klg.jclass.util.swing;

import com.klg.jclass.util.JCEnvironment;
import com.klg.jclass.util.JCVersion;
import com.klg.jclass.util.value.AbstractValueModel;
import com.klg.jclass.util.value.JCValueEvent;
import com.klg.jclass.util.value.JCValueModel;

import javax.swing.*;
import java.awt.*;
import java.text.NumberFormat;
import java.text.ParseException;

/**
 * A spin box component whose spin arrows allow a user to increment or
 * decrement the contained <code>Number</code> value by the specified
 * <code>spinStep</code> amount.
 */
public class JCSpinNumberBox extends AbstractSpinBox {

public static final int INTEGER = 1;
public static final int FLOATING_POINT = 2;

protected int           operation = INTEGER;
protected NumberFormat  format = NumberFormat.getNumberInstance();
protected Number        step = new Integer(1);
protected Number        max = new Long(Long.MAX_VALUE);
protected Number        min = new Long(Long.MIN_VALUE);

public JCSpinNumberBox()
{
    super();
    setEditor(new NumberEditor());
    setValueModel(new NumberModel());
}

public void setOperation(int operation)
{
    if (operation != INTEGER && operation != FLOATING_POINT) {
        throw new IllegalArgumentException("operation must be one of INTEGER " +
                                           "or FLOATING_POINT");
    }

    if (this.operation == operation) {
        return;
    }
    int old_operation = this.operation;
    this.operation = operation;
    firePropertyChange("operation", old_operation, operation);
}

public int getOperation()
{
    return(operation);
}

/**
 * Provided for Beans property editors - has no effect.
 * @param s
 */
public void setAbout(String s) {}

/**
 * Returns vendor information.
 * @return vendor information
 */
public String getAbout()
{
    return(JCVersion.getVersionString());
}


public void setSpinStep(Number step)
{
    this.step = step;
}

public Number getSpinStep()
{
    return(step);
}

/**
 * Set the value displayed by the spin box.
 * @param o The new value
 */
public void setValue(Object o)
{
    if (!(o instanceof Number)) {
        throw new IllegalArgumentException("Value for JCSpinNumberBox must " +
                                           "be of type Number");
    }
    setValue((Number)o);
}

public void setValue(Number value)
{
    // Check range
    if (value != null &&
        (compare(value, max) > 0 || compare(value, min) < 0))
    {
        // Ignore invalid values
        return;
    }

    Number thisvalue = (Number)valueModel.getValue();
    if (thisvalue != null && compare(value, thisvalue) == 0) {
        // Values are equal so we don't try to change them
        return;
    }

    valueModel.setValue(value);
}

public Number getValue()
{
    return((Number)valueModel.getValue());
}

public void setValueRange(Number min, Number max)
{
    if (compare(min, max) > 0) {
        throw new IllegalArgumentException("Min must be less than Max");
    }
    this.max = max;
    this.min = min;
}

public void setValueRange(Range range)
{
    setValueRange(range.min, range.max);
}

public Number getMaximumValue()
{
    return(max);
}

public Number getMinimumValue()
{
    return(min);
}

/**
 * Sets the <code>NumberFormat</code> object used by the spinner to parse and
 * format numbers.
 * @param format
 */
public void setNumberFormat(NumberFormat format)
{
    if (format == null) {
        throw new NullPointerException("NumberFormat must be non-null");
    }
    this.format = format;
}

/**
 * Returns the <code>NumberFormat</code> object used by the spinner to parse
 * and format numbers.
 * @return <code>NumberFormat</code> object
 */
public NumberFormat getNumberFormat()
{
    return(format);
}

protected void checkArrowButtons()
{
    // If a spin up is not possible from this value then disable
    // the up arrow button else make sure that it is enabled
    Number next_value = (Number)spinUp();
    if (compare(next_value, max) > 0) {
        setArrowEnabled(upArrow, false);
    }
    else {
        setArrowEnabled(upArrow, true);
    }

    // If a spin down is not possible from this value then
    // disable the down arrow button else make sure that it is enabled
    next_value = (Number)spinDown();
    if (compare(next_value, min) < 0) {
        setArrowEnabled(downArrow, false);
    }
    else {
        setArrowEnabled(downArrow, true);
    }
}

/**
 * Returns the String format for the specified number, conditional on the
 * current operation.
 * @see #setOperation
 * @return String format for the specified number
 * @param number specified number
 */
protected String format(Number number)
{
    if (number == null) {
        return("");
    }

    String formatted_number = "";

    // Ideally we should use number format here
    if (operation == INTEGER) {
        formatted_number = format.format(number.longValue());
    }
    else {
        formatted_number = format.format(number.doubleValue());
    }
    return(formatted_number);
}

/**
 * Returns a parsed value for the specified String, conditional on the
 * current operation.
 * @see #setOperation
 * @return parsed value for the specified String
 * @param string
 */
protected Number parse(String string)
{
    Number parse_number = null;
    // Ideally we should use number format here
    if (operation == INTEGER) {
        format.setParseIntegerOnly(true);
    }
    else {
        format.setParseIntegerOnly(false);
    }
    try {
        parse_number = format.parse(string);
    }
    catch (ParseException e) {
        // no need to handle this eception since our parse value
        // is already initialized to null
    }

    return(parse_number);
}

/**
 * Overrides the preferred width to make sure the width is big enough for
 * the formatted max and min values.
 * @return
 */
protected int preferredWidthOfView()
{
    int width = 0;
    if (isEditable()) {
        Component ec = editor.getEditorComponent();
        // get width of max value
        editor.setItem(max);
        width = ec.getPreferredSize().width;

        // compare to width of min value
        editor.setItem(min);
        width = Math.max(width, ec.getPreferredSize().width);

        if (operation == FLOATING_POINT) {
            // the decimals and clean round-off of the formatting could
            // mess up our width calculations so we compare to the
            // width of "min + step" and "max - step" and hope this
            // covers all cases for most users
            editor.setItem(new Double(min.doubleValue() + step.doubleValue()));
            width = Math.max(width, ec.getPreferredSize().width);

            editor.setItem(new Double(max.doubleValue() - step.doubleValue()));
            width = Math.max(width, ec.getPreferredSize().width);
        }

        // make sure the editor is set back to its proper value
        editor.setItem(getValue());
    }
    else {
        // get width of max value
        Component ec = renderer.getComponent(this, max);
        width = ec.getPreferredSize().width;

        // compare to width of min value
        renderer.getComponent(this, min);
        width = Math.max(width, ec.getPreferredSize().width);

        if (operation == FLOATING_POINT) {
            // the decimals and clean round-off of the formatting could
            // mess up our width calculations so we compare to the
            // width of "min + step" and "max - step" and hope this
            // covers all cases for most users
            renderer.getComponent(this, new Double(min.doubleValue() +
                                                   step.doubleValue()));
            width = Math.max(width, ec.getPreferredSize().width);

            renderer.getComponent(this, new Double(max.doubleValue() -
                                                   step.doubleValue()));
            width = Math.max(width, ec.getPreferredSize().width);
        }

        // make sure the editor is set back to its proper value
        renderer.getComponent(this, getValue());
    }
    width += JCEnvironment.isWorking("MotifPreferedWidthAjustment");
    return width;
}

protected Object spinUp()
{
    Number value = (Number)valueModel.getValue();
    Number new_value = null;

    if (value == null) {
        new_value = step;
    }
    else if (operation == INTEGER) {
        // increment the value with integer math
        long inv = value.longValue() + step.longValue();
        new_value = new Long(inv);
    }
    else {
        // increment the value with floating point math
        double dnv = value.doubleValue() + step.doubleValue();
        new_value = new Double(dnv);
    }

    return(new_value);
}

protected Object spinDown()
{
    Number value = (Number)valueModel.getValue();
    Number new_value = null;

    if (value == null) {
        new_value = step;
    }
    else if (operation == INTEGER) {
        // decrement the value with integer math
        long inv = value.longValue() - step.longValue();
        new_value = new Long(inv);
    }
    else {
        // decrement the value with floating point math
        double dnv = value.doubleValue() - step.doubleValue();
        new_value = new Double(dnv);
    }

    return(new_value);
}

protected void processValueChanged(JCValueEvent event)
{
    editor.setItem(event.getNewValue());
    renderer.getComponent(this, event.getNewValue());
    super.processValueChanged(event);
}

/**
 * Compares two numbers according to the operation value.  <code>null</code> is
 * considered to be less than all non-null values.
 * @return
 * @param n1
 * @param n2
 */
protected int compare(Number n1, Number n2)
{
    if (n1 == null) {
        if (n2 == null) {
            return(0);
        }
        return(-1);
    }

    if (n2 == null) {
        return(1);
    }

    if (operation == INTEGER) {
        long n1Val = n1.longValue();
        long n2Val = n2.longValue();
        return(n1Val < n2Val ? -1 : (n1Val == n2Val ? 0 : 1));
    }
    else {
        double n1Val = n1.doubleValue();
        double n2Val = n2.doubleValue();
        return(n1Val < n2Val ? -1 : (n1Val == n2Val ? 0 : 1));
    }
}

/**
 * Public static class to pass in and out the allowable <code>Range</code> of
 * the value.
 */
static public class Range implements java.io.Serializable {
    public Number min;
    public Number max;

    public Range(Number min, Number max)
    {
        this.max = max;
        this.min = min;
    }
}

class NumberEditor extends JTextField implements JCSpinBoxEditor {
    /**
     * Returns the component that should be added to the tree hierarchy for
     * this editor.
     * @return component that should be added to the tree hierarchy
     */
    public Component getEditorComponent()
    {
        return(this);
    }

    /**
     * Sets the item that should be edited. Cancels any editing, if necessary.
     * @param anObject
     */
    public void setItem(Object anObject)
    {
        if (anObject == null) {
            setText("");
        }
        else {
            setText(format((Number)anObject));
        }
    }

    /**
     * Returns the edited item.
     * @return edited item
     */
    public Object getItem()
    {
        Number number = parse(getText());
        if (number == null) {
            // User typed in a non-numeric value. Restore old value.
           setItem(valueModel.getValue());
        }
        return(number);
    }
} // End of NumberEditor

static class NumberModel extends AbstractValueModel
                         implements JCValueModel,
                                    java.io.Serializable
{
    /**
     * Returns the type of the core value. It is necessary to store the
     * class of the value separately from the object since, if the value object
     * is null, then the class information is lost.
     * @return type of the core value
     */
    public Class getValueClass()
    {
        return(Number.class);
    }
} // End of NumberModel

}
