/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCSwingUtilities.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:46 $  $Locker:  $  Quest Software Inc.
package com.klg.jclass.util.swing;

import javax.swing.*;
import javax.swing.plaf.ButtonUI;
import javax.swing.plaf.basic.BasicButtonUI;
import java.awt.*;

/**
 * Utilities class that provides tools to perferm common tasks.
 */
public class JCSwingUtilities {

/**
 * Recursively calls <code>setEnabled()</code> on all child components of a
 * container. If the Component passed in is not a Container,
 * it is either enabled or disabled, and the method returns.
 *
 * @param container the Component to be recursively enabled
 * or disabled
 * @param value a boolean indicating if the container and its
 * children (recursively) should be enabled or disabled; <code>true</code> means
 * enabled, <code>false</code> disabled
 */
public static void setEnabled(Component container, boolean value)
{
	if (container instanceof Container) {
		// get all the components contained in this container
		Component components[] = ((Container) container).getComponents();
		for (int i = 0; i < components.length; i++) {
			setEnabled(components[i], value);
		}
	}
	container.setEnabled(value);
}

/**
 * Calculates the preferred size of Swing components from the font metrics in
 * the specified <code>Graphics</code> object, not from the default screen font
 * metrics as Swing usually does. <br><B>Note:</b> This method currently only
 * supports <code>JLabel</code> and <code>JButton</code>.
 * @param component The component to get the preferred size from (currently,
 *                  this must be a <code>JButton</code> or a
 *                  <code>JLabel</code>)
 * @param gc        The graphics object used to obtain font metrics
 * @param override  <code>true</code> if this overrides the use of
 *                  <code>getPreferredSize()</code> from the UI class specified
 *                  for the component all the time; <code>false</code> if
 *                  preferred size is to be calculated by this method only when
 *                  the component's UI class is the default one
 * @return
 */
public static Dimension getPreferredSizeFromGraphics(JComponent component,
                                                     Graphics gc,
                                                     boolean override)
{
	if (component == null) {
		return(new Dimension(0, 0));
	}

	if (gc == null) {
		return(component.getPreferredSize());
	}

    String uiClassID = component.getUIClassID();
    UIDefaults uiDefaults = UIManager.getDefaults();
    String className = (String)uiDefaults.get(uiClassID);
    Class uiDefaultClass = null;
    if (className != null) {
        uiDefaultClass = uiDefaults.getUIClass(uiClassID);
    }
   	Class actualUIClass = null;

	if (!override) {
    	if (component instanceof JLabel) {
			actualUIClass = ((JLabel)component).getUI().getClass();
    	}
    	else if (component instanceof JButton) {
			actualUIClass = ((JButton)component).getUI().getClass();
		}
    }

	if (actualUIClass == null || uiDefaultClass != actualUIClass) {
		return(component.getPreferredSize());
	}

    if (component instanceof JLabel) {
		return(getLabelPSFromGraphics((JLabel) component, gc));
    }
    else if (component instanceof JButton) {
		return(getButtonPSFromGraphics((JButton) component, gc));
    }

	return(component.getPreferredSize());
}

/**
 * Gets the preferred size of a swing <code>JLabel</code> from the font metrics
 * specified by the <code>Graphics</code> object.
 * @param label The JLabel whose preferred size is required
 * @param gc The graphics object used to obtain font metrics
 * @return The preferred size of the JLabel based on the gc's font metrics
 */
public static Dimension getLabelPSFromGraphics(JLabel label, Graphics gc)
{
    FontMetrics fm = gc.getFontMetrics(label.getFont());

    Insets insets = label.getInsets();
    int dx = insets.left + insets.right;
    int dy = insets.top + insets.bottom + 2;
    Rectangle textR = new Rectangle(0, 0, 0, 0);
    Rectangle iconR = new Rectangle(0, 0, 0, 0);
    Rectangle viewR = new Rectangle(dx, dy, Short.MAX_VALUE, Short.MAX_VALUE);

    SwingUtilities.layoutCompoundLabel(label, fm,
                                       label.getText(), label.getIcon(),
                                       label.getVerticalAlignment(),
                                       label.getHorizontalAlignment(),
                                       label.getVerticalTextPosition(),
                                       label.getHorizontalTextPosition(),
                                       viewR,
                                       iconR,
                                       textR,
                                       label.getIconTextGap());

    int x1 = Math.min(iconR.x, textR.x);
    int x2 = Math.max(iconR.x + iconR.width, textR.x + textR.width);
    int y1 = Math.min(iconR.y, textR.y);
    int y2 = Math.max(iconR.y + iconR.height, textR.y + textR.height);

    return(new Dimension(x2 - x1 + dx, y2 - y1 + dy));
}

/**
 * Gets the preferred size of a swing <code>JButton</code> from the font metrics
 * specified by the <code>Graphics</code> object.
 * @param button The JButton whose preferred size is required
 * @param gc The graphics object used to obtain font metrics
 * @return The preferred size of the JButton based on the gc's font metrics
 */
public static Dimension getButtonPSFromGraphics(JButton button, Graphics gc)
{
    ButtonUI ui = button.getUI();
    int gap = 0;
    if (button.getText() != null && ui instanceof BasicButtonUI) {
        gap = ((BasicButtonUI) ui).getDefaultTextIconGap(button);
    }

    FontMetrics fm = gc.getFontMetrics(button.getFont());

    Rectangle textR = new Rectangle(0, 0, 0, 0);
    Rectangle iconR = new Rectangle(0, 0, 0, 0);
    Rectangle viewR = new Rectangle(Short.MAX_VALUE, Short.MAX_VALUE);

    SwingUtilities.layoutCompoundLabel(button, fm,
                                       button.getText(), button.getIcon(),
                                       button.getVerticalAlignment(),
                                       button.getHorizontalAlignment(),
                                       button.getVerticalTextPosition(),
                                       button.getHorizontalTextPosition(),
                                       viewR,
                                       iconR,
                                       textR,
                                       gap);

    Rectangle r = iconR.union(textR);
    Insets insets = button.getInsets();
    int dx = insets.left + insets.right;
    int dy = insets.top + insets.bottom;
    r.width  += dx;
    r.height += dy;

    if (r.width % 2 == 0) {
        r.width += 1;
    }
    if (r.height % 2 == 0) {
        r.height += 1;
    }
    return(new Dimension(r.width, r.height));
}

}
