/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCSwingWorker.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:46 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.swing;

import javax.swing.SwingUtilities;

/**
 * An abstract class that you subclass to perform
 * GUI-related work in a dedicated thread.
 * To use this class you override the void implementations of
 * <code>start()</code>, <code>construct()</code>, and <code>finish()</code>.<p>
 * The <code>start()</code> method is called initially and
 * once this has completed, the <code>construct()</code> method is launched
 * in it's own thread.  When this thread has completed,
 * the <code>finish()</code> method is called from the Swing thread.<p>
 * It is guaranteed to be safe to alter GUI components in the
 * <code>start()</code> and <code>finish()</code> methods.<p>
 * To safely alter Swing components in the <code>construct()</code> method, you
 * must either create a separate <code>Runnable</code> class and use
 * <code>SwingUtilities().invokeLater()</code> on it, use the convienence object
 * <code>RunnableObjectWrapper</code>, or use a thread-safe helper class such as
 * </code>MessageHelper</code> or <code>ProgressHelper</code>.
 * @see com.klg.jclass.util.swing.JCSwingRunnable
 * @see com.klg.jclass.util.swing.JCMessageHelper
 * @see com.klg.jclass.util.swing.JCProgressHelper
 */
public abstract class JCSwingWorker
{

private Object value;
private Thread thread;

/**
 * Called on the event dispatching thread (not on the worker thread)
 * before the <code>construct</code> method is called.
 */
public void start() {}

/**
 * Computes the value to be returned by the <code>get</code> method.
 * @return
 */
public abstract Object construct();

/**
 * Called on the event dispatching thread (not on the worker thread)
 * after the <code>construct</code> method has returned.
 */
public void finished() {}

/**
 * A new method that interrupts the worker thread.  Call this method
 * to force the worker to abort what it's doing.
 */
public void interrupt()
{
    Thread t = thread;
    if (t != null) {
        t.interrupt();
    }
    thread = null;
}

/**
 * Returns the value created by the <code>construct</code> method.
 * @return value created by the <code>construct</code> method
 */
public Object get()
{
    while (true) {  // Keep trying if we're interrupted
        Thread t;
        synchronized (JCSwingWorker.this) {
            t = thread;
            if (t == null) {
                return value;
            }
        }
        try {
            t.join();
        }
        catch (InterruptedException e) {}
    }
}

final Runnable doFinished = new Runnable()
{
    public void run()
    {
        finished();
    }
};

final Runnable doConstruct = new Runnable()
{
    public void run() {
        synchronized(JCSwingWorker.this) {
            value = construct();
            thread = null;
        }
        SwingUtilities.invokeLater(doFinished);
        // yield the current thread to get the event thread to execute now
        Thread.currentThread().yield();
    }
};

/**
 * Starts a thread that will call the <code>construct</code> method
 * and then exit.
 */
public JCSwingWorker()
{
     
    // Currently it is NOT safe to practice the following since
    // a inner classes pointer to its containing class in not defined
    // until after its own constructor is called, hence it is not safe
    // to call the the abstract "start" method from the constructor as
    // we do here if we are in the EventDispatchThread
    /*
    if (SwingUtilities.isEventDispatchThread()) {
        // invoke start() work to do upfront work
        start();
        thread = new Thread(doConstruct);
        thread.start();
    }
    else {
        final Runnable doStart = new Runnable() {
            public void run() {
                start();
                thread = new Thread(doConstruct);
                thread.start();
            }
        };
        // this object _must_ be spawned from the event
        // dispatch thread to be swing thread safe
        SwingUtilities.invokeLater(doStart);
        // yield the current thread to get the event thread to execute now
        Thread.currentThread().yield();
    }
    */
        
    final Runnable doStart = new Runnable() {
        public void run() {
            start();
            thread = new Thread(doConstruct);
            thread.start();
        }
    };
    // This object _must_ be spawned from the event
    // dispatch thread to be swing thread safe
    SwingUtilities.invokeLater(doStart);
    // yield the current thread to get the event thread to execute now
    Thread.currentThread().yield();
    
}

}
