/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCEncodeComponent.java,v $ $Revision: 1.2 $
//          $Date: 2006-03-17 21:13:47 $ $Locker:  $ Quest Software Inc.

package com.klg.jclass.util.swing.encode;

import java.awt.Component;
import java.awt.Image;
import java.io.OutputStream;
import java.io.IOException;
import java.io.File;

import com.klg.jclass.util.JCLocaleManager;
import com.klg.jclass.util.swing.JCMessageHelper;
import com.klg.jclass.util.swing.encode.resources.LocaleInfo;

/**
 * This class allows for the creation of images of Java components and
 * the encoding of images in several graphics formats (such as PNG or JPEG).
 * It is used to output any AWT or Swing component (including JClass components)
 * or Image class to a file or output stream.
 * The image formats supported are:
 * <DL>
 * <DT>GIF
 * <DD><I>Only enabled if user has a GIF output license from Unisys.</I>
 * <DT>PNG
 * <DD>Portable Network Graphics format.
 * <DT>JPEG
 * <DD>Joint Photographic Experts Group format.
 * <DT>EPS
 * <DD>Encapsulated PostScript format. <b>Requires JClass PageLayout.</b>
 * <DT>PDF
 * <DD>Acrobat Portable Document format. <b>Requires JClass PageLayout.</b>
 * <DT>PS
 * <DD>PostScript Level 2 format. <b>Requires JClass PageLayout.</b>
 * <DT>PCL
 * <DD>Hewlett-Packard Page Control Language version 5. <b>Requires JClass PageLayout.</b>
 * </DL>
 */
public class JCEncodeComponent {

/** Specifies GIF encoding. */
public static final Encoding GIF = new Encoding("GIF",
	"CompuServe Graphic Interchange File",
	"com.klg.jclass.util.swing.encode.GIFEncoder",
	LocaleInfo.NEEDS_GIF_LICENSE);
/** Specifies PNG encoding. */
public static final Encoding PNG = new Encoding("PNG",
	"Portable Network Graphics",
	"com.klg.jclass.util.swing.encode.PNGEncoder",
	LocaleInfo.LOAD_ERROR);
/** Specifies SVG Encoding. */
public static final Encoding SVG = new Encoding("SVG",
	"Scalable Vector Graphics",
	"com.klg.jclass.util.swing.encode.SVGEncoder",
	LocaleInfo.LOAD_ERROR);
/** Specifies SWF Encoding. */
public static final Encoding SWF = new Encoding("SWF",
	"Shockwave Flash",
	"com.klg.jclass.util.swing.encode.swf.SWFEncoder",
	LocaleInfo.LOAD_ERROR);
/** Specifies JPEG encoding. */
public static final Encoding JPEG = new Encoding("JPEG",
	"Joint Photographic Experts Group",
	"com.klg.jclass.util.swing.encode.JPEGEncoder",
	LocaleInfo.LOAD_ERROR);
/** Specifies EPS encoding. */
public static final Encoding EPS = new Encoding("EPS",
	"Encapsulated PostScript",
	"com.klg.jclass.util.swing.encode.page.EPSEncoder",
	LocaleInfo.NEEDS_PAGELAYOUT);
/** Specifies PCL encoding. */
public static final Encoding PCL = new Encoding("PCL",
	"HP Page Control Language Version 5",
	"com.klg.jclass.util.swing.encode.page.PCLEncoder",
	LocaleInfo.NEEDS_PAGELAYOUT);
/** Specifies PDF encoding. */
public static final Encoding PDF = new Encoding("PDF",
	"Adobe Acrobat Portable Document File",
	"com.klg.jclass.util.swing.encode.page.PDFEncoder",
	LocaleInfo.NEEDS_PAGELAYOUT);
/** Specifies PS encoding. */
public static final Encoding PS = new Encoding("PS",
	"Adobe PostScript Level 2",
	"com.klg.jclass.util.swing.encode.page.PSEncoder",
	LocaleInfo.NEEDS_PAGELAYOUT);

/**
 *  Client side encodings.
 */
public static final Encoding ENCODINGS[] = {
	PNG, GIF, JPEG,
	EPS, PCL, PDF, PS
};

/**
 *  Server side encodings.
 */
public static final Encoding SERVER_ENCODINGS[] = {
	PNG, GIF, JPEG,
    SWF, SVG,
	PDF
};

/**
 *  Imagebased encodings.
 */
public static final Encoding IMAGE_ENCODINGS[] = {
	PNG, GIF, JPEG
};

protected static String bundleName =
    "com.klg.jclass.util.swing.encode.resources.LocaleInfo";

/**
 * Encodes the image of a Java component to an output stream.
 * Can be used with any AWT or Swing component (including
 * JClass components). See <A HREF="#navbar_top">above</A> for details
 * on supported image formats.
 *
 * @param encoding image format to encode to (GIF, PNG, JPEG, EPS, PS, PDF,
 * or PCL)
 * @param component component to encode
 * @param output stream to which to write the encoding
 * @throws IOException
 */
public static void encode(Encoding encoding, Component component,
                          OutputStream output)
    throws IOException, EncoderException
{

    Encoder encoder = findEncoder(encoding, component.isShowing());
    if (encoder != null) {
        encoder.encode(component, output);
    }
}

/**
 * Encodes the image of a Java component to a file.
 * Can be used with any AWT or Swing component (including
 * JClass components). See <A HREF="#navbar_top">above</A> for details
 * on supported image formats.
 *
 * @param encoding image format to encode to (GIF, PNG, JPEG, EPS, PS, PDF,
 * or PCL)
 * @param component component to encode
 * @param file file to which to write the encoding
 * @throws IOException
 */
public static void encode(Encoding encoding, Component component, File file)
    throws IOException, EncoderException
{
    OutputStream os = new java.io.FileOutputStream(file);
    encode(encoding, component, os);
    os.close();
}

/**
 * Encodes an image to an output stream.
 *
 * @param encoding image format to encode to (GIF, PNG, JPEG, EPS, PS, PDF,
 * or PCL)
 * @param image image to encode
 * @param output stream to which to write the encoding
 * @throws IOException
 */
public static void encode(Encoding encoding, Image image, OutputStream output)
    throws IOException, EncoderException
{
    Encoder encoder = findEncoder(encoding, false);
    if (encoder != null) {
        encoder.encode(image, output);
    }
}

/**
 * Encodes an image to a file.
 *
 * @param encoding image format to encode to (GIF, PNG, JPEG, EPS, PS, PDF,
 * or PCL)
 * @param image image to encode
 * @param file file to which to write the encoding
 * @throws IOException
 */
public static void encode(Encoding encoding, Image image, File file)
    throws IOException, EncoderException
{
    OutputStream os = new java.io.FileOutputStream(file);
    encode(encoding, image, os);
    os.close();
}

/**
 * Finds the encoder class for the specified encoding.
 * @param encoding image format to encode to (GIF, PNG, JPEG, EPS, PS, PDF,
 *                 or PCL)
 * @param throwDialog <code>true</code> is error dialog is to be shown in case
 *                    of error; <code>false</code> if no dialog is to be shown
 * @return encoder class for this encoding
 * @throws IOException
 */
public static Encoder findEncoder(Encoding encoding, boolean throwDialog)
    throws IOException, EncoderException
{
    Encoder encoder = encoding.getEncoder();
    if (encoder == null) {
        // Make sure the bundle has been added
        JCLocaleManager.getDefault().add(bundleName);
        String dialogTitle = LocaleInfo.ENCODER_DIALOG_TITLE;
        String title = JCLocaleManager.getDefault().getString(dialogTitle);
        if (title == null) {
            // Can't find text in resource bundles so use the key
            title = LocaleInfo.ENCODER_DIALOG_TITLE;
        }

        if (throwDialog) {
            JCMessageHelper.showError(title, encoding.getFailureMessage());
        }
        throw new EncoderException(encoding.getFailureMessage());
    }
    return(encoder);
}

/**
 * Checks if encoding type is image based. An image based encoding type
 * must encode from an image. Examples are PNG, GIF and JPEG.
 * @return
 * @param enc
 */
public static boolean isImageBasedEncoding(Encoding enc)
{
    for (int i = 0; i < IMAGE_ENCODINGS.length; i++) {
        if (IMAGE_ENCODINGS[i] == enc) {
            return(true);
        }
    }
    return(false);
}

/**
 * Class used to enumerate valid encoding formats.
 */
public static class Encoding {
    private String shortName;
    private String longName;
    private String encoderClass;
    private String failureMessage;

    /**
     * Constructs a new image encoder.
     * @param short_name
     * @param long_name
     * @param failure_message
     * @param encoder_class
     */
    public Encoding(String short_name, String long_name,
                    String encoder_class, String failure_message)
    {
        this.shortName = short_name;
        this.longName = long_name;
        this.encoderClass = encoder_class;
        this.failureMessage = failure_message;
    }

    /**
     * Returns the brief name of the supported encoding type.
     * @return brief name of the supported encoding type
     */
    public String getShortName()
    {
        return(shortName);
    }

    /**
     * Returns the long name of the supported encoding type.
     * @return long name of the supported encoding type
     */
    public String getLongName()
    {
        return(longName);
    }

    /**
     * Returns a String representation bassed on the long and short name of the
     * Encoder.
     * @return String representation based on the long and short name of the
     *         Encoder
     */
    public String toString()
    {
        return(getLongName() + " (" + getShortName() + ")");
    }

    /**
     * Returns message about possible reasons for encoder load
     * failure. (For example, when <code>getEncoder()</code> returns null.)
     * @return message about possible reasons for encoder load failure
     */
    public String getFailureMessage()
    {
        // Make sure the bundle has been added
        JCLocaleManager.getDefault().add(bundleName);
        String locale_message =
            JCLocaleManager.getDefault().getString(failureMessage);
        if (locale_message == null) {
            locale_message = failureMessage;
        }
        return(locale_message);
    }

    /**
     * Returns an encoder for this encoding type.
     * @return returns null if it cannot locate or load the encoder
     */
    public Encoder getEncoder()
    {
        Encoder encoder = null;
        // attempt to load encoder
        // we no longer cache the encoder because problems may arise
        // if the same encoder is used by multiple users at the same time
        // (such as in, say, a server-side environment)
        Class encoder_class = null;
        try {
            encoder_class = Class.forName(encoderClass);
        }
        catch (ClassNotFoundException cnfe) {}

        if (encoder_class != null) {
            try {
                encoder = (Encoder)encoder_class.newInstance();
            }
            catch (InstantiationException ie) {}
            catch (IllegalAccessException iae) {}
        }
        return(encoder);
    }

} // End of Encoding inner class

}
