/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: BranchTree.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:48 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.treetable;

import com.klg.jclass.util.JCListenerList;

import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import java.io.Serializable;
import java.util.Enumeration;

/**
 * Class to wrap a tree model so that it only exposes branches; that is, non-
 * Leaf nodes.  This is useful for Explorer-type views where the "tree view"
 * portion only displays the branch nodes.
 */
public class BranchTree implements TreeModel, Serializable
{

protected TreeModel       tree;
protected JCListenerList  listeners = null;

public BranchTree(TreeModel tree)
{
	this.tree = tree;
	// Install tree listeners on the tree
	tree.addTreeModelListener(new Listener());
}

//
// The following methods implement javax.swing.tree.TreeModel.
//

/**
 * Adds the specified treemodel listener.
 * @param l The treemodel listener to add
 */
public void addTreeModelListener(TreeModelListener l)
{
	listeners = JCListenerList.add(listeners, l);
}

/**
 * Removes the specified treemodel listener
 * @param l The treemodel listener to remove
 */
public void removeTreeModelListener(TreeModelListener l)
{
    listeners = JCListenerList.remove(listeners, l);
}

/**
 * Returns the "index"th non-leaf child of the parent node.
 * @param parent The parent node
 * @param index The index of the child to get
 * @return The "index"th child of the given parent
 */
public Object getChild(Object parent, int index)
{
	int num_kids = tree.getChildCount(parent);
	for (int i = 0; i < num_kids; i++) {
		Object child = tree.getChild(parent, i);
		if (tree.isLeaf(child) == false) {
			if (index == 0) {
				return(child);
			}
			index--;
		}
	}

	// ??? Perhaps throw an exception
	return(null);
}

/**
 * Returns the count of non-leaf children.
 * @param parent The parent node
 * @return The count of non-leaf children for the given parent
 */
public int getChildCount(Object parent)
{
	int count = 0;
	int num_kids = tree.getChildCount(parent);
	for (int i = 0; i < num_kids; i++) {
		Object child = tree.getChild(parent, i);
		if (tree.isLeaf(child) == false) {
			count++;
		}
	}
	return(count);
}

/**
 * Returns the non-leaf index of this child.
 * @param parent The parent node
 * @param child The child whose index is required
 * @return The non-leaf index of the child of the given parent
 */
public int getIndexOfChild(Object parent, Object child)
{
	int count = 0;
	int num_kids = tree.getChildCount(parent);
	for (int i = 0; i < num_kids; i++) {
		Object node = tree.getChild(parent, i);
		if (tree.isLeaf(node) == false) {
			count++;
			if (child == node) {
				return(count);
			}
		}
	}

	// Child was not found
	return(-1);
}

/**
 * Returns the root node of the tree.
 * @return The root node
 */
public Object getRoot()
{
	// Pass this call straight through to the underlying tree
	return(tree.getRoot());
}

/**
 * Returns true if the child is a leaf.
 * @param node The node to check
 * @return Is the given node a leaft?
 */
public boolean isLeaf(Object node)
{
	// All nodes are branches
	return(false);
}

/**
 * Called when the user has altered the value for the item identified
 * by <code>path</code> to <code>newValue</code>.
 * If <code>newValue</code> signifies a truly new value
 * the model should post a <code>treeNodesChanged</code> event.
 *
 * @param path path to the node that the user has altered
 * @param newValue the new value from the TreeCellEditor
 */
public void valueForPathChanged(TreePath path, Object newValue)
{
	// Pass this call straight through to the underlying tree
	tree.valueForPathChanged(path, newValue);
}

/**
 * Given a <code>TreePath</code>, this method returns the last component or
 * "node" of the <code>TreePath</code>.
 * @param path The path from which to get the node
 * @return The node (i.e. last path component) of the given path
 */
static protected Object getNode(TreePath path)
{
	return(path.getLastPathComponent());
}

/**
 * Inner class to handle events from the source tree.
 */
class Listener implements TreeModelListener
{
	protected TreeModelEvent convertTreeEvent(TreeModelEvent e)
    {
		Object node = getNode(e.getTreePath());
		int num_kids = tree.getChildCount(node);
		int child_indices[] = e.getChildIndices();
		int branch_indices[] = new int[child_indices.length];
		int branch_count = 0;
		for (int i = 0; i < child_indices.length; i++) {
            Object child = null;
			try {
                child = tree.getChild(node, child_indices[i]);
            }
            catch (ArrayIndexOutOfBoundsException e1) {
                child = null;
            }
			if (child == null || tree.isLeaf(child) == false) {
				branch_indices[branch_count++] = child_indices[i];
			}
		}
        
		if (branch_count == 0) {
			// None of the kids had changed so we do nothing
			return(null);
		}
        
        // build an array of ints from the vector
        int indices[] = new int[branch_count];
        for (int j = 0; j < branch_count; j++) {
            indices[j] = branch_indices[j];
        }

        // Build the new tree event
        return new TreeModelEvent(BranchTree.this, e.getTreePath(), indices,
                                 /* no document to be useful */ null);
	}


	public void treeNodesChanged(TreeModelEvent tme)
    {
		TreeModelEvent ntme = convertTreeEvent(tme);
		if (ntme == null) {
			return;
		}

		// repost treeNodesChanged
		Enumeration e = JCListenerList.elements(listeners);
		while (e.hasMoreElements()) {
			TreeModelListener listener = (TreeModelListener)e.nextElement();
			listener.treeNodesChanged(ntme);
		}
	}

	public void treeNodesInserted(TreeModelEvent tme)
    {
		TreeModelEvent ntme = convertTreeEvent(tme);
		if (ntme == null) {
			return;
		}

		// repost treeNodesInserted
		Enumeration e = JCListenerList.elements(listeners);
		while (e.hasMoreElements()) {
			TreeModelListener listener = (TreeModelListener)e.nextElement();
			listener.treeNodesInserted(ntme);
		}
	}

	public void treeNodesRemoved(TreeModelEvent tme)
    {
		TreeModelEvent ntme = convertTreeEvent(tme);
		if (ntme == null) {
			return;
		}
        
		// ??? May be tricky
		// Repost treeNodesRemoved
		Enumeration e = JCListenerList.elements(listeners);
		while (e.hasMoreElements()) {
			TreeModelListener listener = (TreeModelListener)e.nextElement();
            listener.treeNodesRemoved(ntme);
		}
	}

	public void treeStructureChanged(TreeModelEvent tme)
    {
		TreeModelEvent ntme = new TreeModelEvent(BranchTree.this,
                                                 tme.getTreePath());

		// Repost treeStructureChanged
		Enumeration e = JCListenerList.elements(listeners);
		while (e.hasMoreElements()) {
			TreeModelListener listener = (TreeModelListener)e.nextElement();
			listener.treeStructureChanged(ntme);
		}
	}

} // end of Listener

}
