/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: TableFromTreeTable.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:48 $  $Locker:  $  Quest Software Inc.
package com.klg.jclass.util.treetable;

import javax.swing.event.TableModelListener;
import javax.swing.table.TableModel;
import javax.swing.tree.TreePath;
import java.util.ArrayList;
import java.util.Hashtable;

/**
 * Class to wrap a treetable model so that the data in it can be viewed as a
 * table model. This is an internal class and not for public use.
 */
public class TableFromTreeTable implements TableModel {

// For efficiency we don't create a map of the rows (in an ArrayList) unless a
// row reference is explicitly requested
protected JCTreeTableModel     treetable = null;
protected ArrayList            rowMap    = null;
protected int                  hashInt   = 0;
protected Hashtable            treeMap   = null;

public TableFromTreeTable(JCTreeTableModel treetable)
{
	this.treetable = treetable;
}

/**
 * Add the kids of the parent whose is currently at the end of the parent_path,
 * to the row to treepath and treepath to row mappings.  If the kids have
 * children, recursively call this method on the kids.
 * @param parent_path The path from the root up to and including the parent
 * @param parent The parent whose kids we want to add
 */
protected void addKids(TreePath parent_path, Object parent)
{
	int num_kids = treetable.getChildCount(parent);
	for (int i = 0; i < num_kids; i++) {
		Object child = treetable.getChild(parent, i);
		TreePath child_path = parent_path.pathByAddingChild(child);
		rowMap.add(child_path);
	    treeMap.put(child_path, new Integer(hashInt++));

		if (!treetable.isLeaf(child)) {
			// Recursively descend the folder child
			addKids(child_path, child);
		}
	}
}

/**
 * Initialize the row to tree and tree to row mappings
 */
protected void initMappings()
{
	rowMap  = new ArrayList();
	treeMap = new Hashtable();
    hashInt = 0;

	// Traverse the tree adding the root first
	TreePath root_path = new TreePath(treetable.getRoot());
	rowMap.add(root_path);
	treeMap.put(root_path, new Integer(hashInt++));
	addKids(root_path, treetable.getRoot());
}

/**
 * If the treetable has changed it needs to call this to reset the mappings
 */
protected void resetMappings()
{
	rowMap  = null;
	treeMap = null;
    hashInt = 0;
}

/**
 * Get the node for the given row
 * @param row The row whose node we want
 * @return The node for the given row
 */
protected Object rowToNode(int row)
{
	return(rowToTreePath(row).getLastPathComponent());
}

/**
 * Get the tree path for the given row
 * @param row The row whose path we want
 * @return The tree path for the given row
 */
protected TreePath rowToTreePath(int row)
{
	if (rowMap == null) {
		initMappings();
	}

	if (row > rowMap.size()) {
		throw new IndexOutOfBoundsException("row = " + row + " > " +
											rowMap.size());
	}

	return((TreePath)rowMap.get(row));
}

/**
 * Get the row for the given tree path
 * @param path The path whose row we want
 * @return The row corresponding to the given tree path.  Returns a negative
 *     value on error.
 */
protected int treePathToRow(TreePath path)
{
    if (treeMap == null) {
        initMappings();
    }

	if (path != null) {
		Integer row = (Integer)treeMap.get(path);
		if (row != null) {
			return(row.intValue());
		}
	}
	return(-1);
}

/**
 * Returns the number of records managed by the data source object. A
 * <code>JTable</code> uses this method to determine how many rows it
 * should create and display.  This method should be quick, as it
 * is called by <code>JTable</code> quite frequently.
 *
 * @return the number or rows in the model
 * @see #getColumnCount
 */
public int getRowCount()
{
	if (rowMap == null) {
		initMappings();
	}
	return(rowMap.size());
}

/**
 * Returns the number of columns managed by the data source object. A
 * <code>JTable<</code> uses this method to determine how many columns it
 * should create and display upon initialization.
 *
 * @return the number or columns in the model
 * @see #getRowCount
 */
public int getColumnCount()
{
	return(treetable.getColumnCount());
}

/**
 * Returns the name of the column at <code>columnIndex</code>.  This is used
 * to initialize the table's column header name.  Note, this name does
 * not need to be unique.  Two columns in a table can have the same name.
 *
 * @param	columnIndex	the index of column
 * @return  the name of the column
 */
public String getColumnName(int columnIndex)
{
	return(treetable.getColumnName(columnIndex));
}

/**
 * Returns the lowest common denominator <code>Class</code> in the column.
 * This is used by the table to set up a default renderer and editor for the
 * column.
 *
 * @param  columnIndex the index of the column whose class has been requested
 * @return the common ancestor class of the object values in the model.
 */
public Class getColumnClass(int columnIndex)
{
	return(treetable.getColumnClass(columnIndex));
}

/**
 * Returns true if the cell at <code>rowIndex</code> and <code>columnIndex</code>
 * is editable.  Otherwise, <code>setValueAt()</code> on the cell will not change
 * the value of that cell.
 *
 * @param	rowIndex	the row of the cell whose editable value is requested
 * @param	columnIndex	the column of the cell whose editable value is requested
 * @return	true if the cell is editable.
 * @see #setValueAt
 */
public boolean isCellEditable(int rowIndex, int columnIndex)
{
	return(treetable.isCellEditable(rowToNode(rowIndex), columnIndex));
}

/**
 * Returns an attribute value for the cell at <code>columnIndex</code>
 * and <code>rowIndex</code>.
 *
 * @param	rowIndex	the row whose value is to be looked up
 * @param	columnIndex the column whose value is to be looked up
 * @return	the value Object at the specified cell
 */
public Object getValueAt(int rowIndex, int columnIndex)
{
	return(treetable.getValueAt(rowToNode(rowIndex), columnIndex));
}

/**
 * Sets an attribute value for the record in the cell at
 * <code>columnIndex</code> and <code>rowIndex</code>.  <code>aValue</code> is
 * the new value.
 *
 * @param	aValue		 the new value
 * @param	rowIndex	 the row whose value is to be changed
 * @param	columnIndex  the column whose value is to be changed
 * @see #getValueAt
 * @see #isCellEditable
 */
public void setValueAt(Object aValue, int rowIndex, int columnIndex)
{
	treetable.setValueAt(aValue, rowToNode(rowIndex), columnIndex);
}

/**
 * Adds a listener to the list that's notified each time a change
 * to the data model occurs.
 *
 * @param	l		the TableModelListener
 */
public void addTableModelListener(TableModelListener l) {}

/**
 * Removes a listener from the list that's notified each time a
 * change to the data model occurs.
 *
 * @param	l		the TableModelListener
 */
public void removeTableModelListener(TableModelListener l) {}

}
