/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: TreeTableSupport.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:48 $ $Locker:  $ Quest Software Inc.

package com.klg.jclass.util.treetable;

import com.klg.jclass.swing.JCTreeTable;
import com.klg.jclass.swing.JCTreeTableSelectionModel;
import com.klg.jclass.swing.DefaultTreeTableSelectionModel;

import com.klg.jclass.util.JCEnvironment;
import com.klg.jclass.util.JCListenerList;
import com.klg.jclass.util.swing.DefaultRowSortTableModel;
import com.klg.jclass.util.swing.JCAction;
import com.klg.jclass.util.swing.JCRowSortModel;
import com.klg.jclass.util.swing.JCRowSortTableModel;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.RowMapper;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Internal class.
 * <P> It is used by <code>JCTreeTable</code> to map tree functionality to a
 * table component.<p>
 * This class provides an implementation to handle a <i>TreeModel/Table
 * Model</i> combination for use within a Table component.
 * <p>
 * Specific functionality includes tracking expanded node counts as well as
 * mapping and posting expansion and selection events.
 * <p>
 * This class also provides a node "painter" object that can be wrapped into
 * a Swing <code>CellRenderer</code> or a JClass <code>CellRenderer</code>.
 * <p>
 * <B>NOTE:</B> the API for this class is not frozen.
 */
public class TreeTableSupport implements TableModel, JCRowSortTableModel,
										 TableModelListener, RowMapper,
										 java.io.Serializable
{

/** valid value for view */
public static final int TREE  = 1;
/** valid value for view */
public static final int TABLE = 2;

/** valid value for ShowNodeLines */
public static final int NODE_LINES_SHOW = 1;
/** valid value for ShowNodeLines */
public static final int NODE_LINES_HIDE = 2;
/** valid (default) value for ShowNodeLines */
public static final int NODE_LINES_USE_PLAF = 3;

protected boolean             sortable = true;
protected int                 sortOrder = JCRowSortModel.FORWARD;
protected int[]               keyColumnsArray[];
protected boolean             autoSort = true;

protected int                 showNodeLines = NODE_LINES_USE_PLAF;
protected JComponent          visualComponent;
protected int                 view = TREE;
protected JCTreeTableModel    treetable = null;
protected TableModel          table;
protected TreeModelListener   treeListener = new TreeListener();
protected JCListenerList      tableListeners = null;
protected int                 treeColumn = 0;
protected JCListenerList      treeExpansionListeners;
protected JCListenerList      treeSelectionListeners;
protected JCListenerList      treeWillExpandListeners;
protected boolean             rootVisible = true;
protected boolean             showsRootHandles = false;
protected JCTreeIconRenderer  iconRenderer = new DefaultTreeIconRenderer();

//
// PLAF controlled variables
//

protected boolean             scrollsOnExpand;   // "Tree.scrollsOnExpand"
protected Icon                expandedIcon;      // "Tree.expandedIcon"
protected Icon                collapsedIcon;     // "Tree.collapsedIcon"
protected int                 leftLevelSpacing;  // "Tree.leftChildIndent"
protected int                 rightLevelSpacing; // "Tree.rightChildIndent"
protected int                 levelIndent;       // calculated from spacing
protected Icon                openIcon;          // "Tree.openIcon"
protected Icon                closedIcon;        // "Tree.closedIcon"
protected Icon                leafIcon;          // "Tree.leafIcon"
protected Color               lineColor;         // "Tree.hash"
protected int                 lineThickness;
protected Stroke              lineStroke;
protected boolean             drawLines;
protected Color               tableBackground;   // "Table.background"
private boolean               isMotif = false;
protected int                 cellHeight;

/** Map of expanded nodes to "treetable" rows */
protected RemovableVector    rowMap = null;

protected Hashtable          expandedState = null;

public TreeTableSupport(JComponent visual_component, JCTreeTableModel model)
{
	this(visual_component);
	setTreeTableModel(model);
}

/**
 * construct a treetable support class
 * @param visual_component is the component we are supporting; this needed
 * for posting the correct event source.
 */
public TreeTableSupport(JComponent visual_component)
{
	this.visualComponent = visual_component;
	inferLookAndFeel();
}

public void inferLookAndFeel()
{
	expandedIcon = (Icon) UIManager.get("Tree.expandedIcon");
	if (expandedIcon == null) {
		expandedIcon = new DefaultIcon(false);
	}
	collapsedIcon = (Icon) UIManager.get("Tree.collapsedIcon");
	if (collapsedIcon == null) {
		collapsedIcon = new DefaultIcon(true);
	}
	leftLevelSpacing =
        ((Integer) UIManager.get("Tree.leftChildIndent")).intValue();
	rightLevelSpacing =
        ((Integer) UIManager.get("Tree.rightChildIndent")).intValue();

	openIcon   = (Icon) UIManager.get("Tree.openIcon");
	closedIcon = (Icon) UIManager.get("Tree.closedIcon");
	leafIcon   = (Icon) UIManager.get("Tree.leafIcon");

	lineColor = (Color) UIManager.get("Tree.hash");

	scrollsOnExpand =
        ((Boolean) UIManager.get("Tree.scrollsOnExpand")).booleanValue();

	levelIndent = leftLevelSpacing + rightLevelSpacing;
	tableBackground = (Color) UIManager.get("Table.background");

	cellHeight = Math.max(openIcon.getIconHeight(), closedIcon.getIconHeight());
	cellHeight = Math.max(cellHeight, leafIcon.getIconHeight());

	//
	// Here we determine hard coded behaviors
	//
	if (JCEnvironment.isMetal()) {
		drawLines = false;
	}
	else if (JCEnvironment.isMotif()) {
		drawLines = true;
		lineThickness = 2;
		isMotif = true;

		lineStroke = new BasicStroke((float) lineThickness,
		                             BasicStroke.CAP_BUTT,
									 BasicStroke.JOIN_BEVEL, 5, null, 0);
	}
	else {
		drawLines = true;
		lineThickness = 1;
		float dashes[] = { (float) lineThickness, (float) lineThickness };

		lineStroke = new BasicStroke((float) lineThickness,
		                             BasicStroke.CAP_BUTT,
									 BasicStroke.JOIN_BEVEL, 5, dashes, 0);
	}

	if (showNodeLines != NODE_LINES_USE_PLAF) {
		if (showNodeLines == NODE_LINES_SHOW) {
			drawLines = true;
		}
		else {
			drawLines = false;
		}
	}
}

//
// RowMapper interface implementaion
//

/**
 * Return the rows for the visible specified paths
 * @param paths The paths whose rows are needed
 * @return The list of rows for the given paths
 */
public int[] getRowsForPaths(TreePath paths[])
{
	if (paths == null) {
		return(null);
	}

	int rows[] = new int[paths.length];
	int i, j;
	for (i = 0, j = 0; i < paths.length; i++) {
		int row = getRowForPath(paths[i]);
		if (row >= 0) {
			// Only allow selection of visible paths
			rows[j++] = row;
		}
	}
	if (i != j) {
		int new_rows[] = new int[j];
		for (int k = 0; k < j; k++) {
			new_rows[k] = rows[k];
		}
		rows = new_rows;
	}
	return(rows);
}

/**
 * Allows you override the plaf specified behavior of drawing lines.
 * Particularly useful in the Metal LookAndFeel.
 *
 * @param lines The type of line drawing desired.  Valid values are
 *     NODE_LINES_USE_PLAF (the default), NODE_LINES_SHOW,
 *     NODE_LINES_HIDE.
 */
public void setShowNodeLines(int lines)
{
	if (showNodeLines == lines) {
		return;
	}
	if (lines != NODE_LINES_SHOW &&
		lines != NODE_LINES_HIDE &&
		lines != NODE_LINES_USE_PLAF)
	{
		throw new IllegalArgumentException("Invalid value for showNodeLines");
	}
	showNodeLines = lines;
	if (showNodeLines == NODE_LINES_USE_PLAF) {
		inferLookAndFeel();
	}

	if (showNodeLines == NODE_LINES_SHOW) {
		drawLines = true;
	}
	else {
		drawLines = false;
	}
}

/**
 * Return the state of ShowNodeLines.
 * @see #setShowNodeLines
 * @return The current style of line drawing used.  Will be one of
 *     NODE_LINES_USE_PLAF (the default), NODE_LINES_SHOW,
 *     NODE_LINES_HIDE.
 */
public int getShowNodeLines()
{
	return(showNodeLines);
}

/**
 * Return the maximum height of the icons used
 * @return The maximum height of the icons used
 */
public int getMaxIconHeight()
{
	return(cellHeight);
}

/**
 * Returns whether we allow sorting of the data or not
 * @return Is sorting of data allowed?
 * @see #setSortable
 */
public boolean isSortable()
{
	return(this.sortable);
}

/**
 * Sets whether we allow sorting or not.
 * @param sortable The new value of sortable.
 * @see #isSortable
 */
public void setSortable(boolean sortable)
{
    this.sortable = sortable;
}

protected void initTableViewModels()
{
	// Remove table listeners from any old model
	if (table != null) {
		table.removeTableModelListener(this);
	}

	/*
	 * Currently we do not support models supplying their own wrapping
	 * to properly handle this we need to create a different interface
	 * with a method that looks like getPathForRow()
	if (treetable instanceof TableModel) {
		table = (TableModel) treetable;
	}
	else {
	*/

	// Create a wrapper to map a treetablemodel into a table model
	table = new TableFromTreeTable(treetable);

	/*
	}
	*/

	if (!(table instanceof JCRowSortTableModel)) {
		// Table is not an instance of the JCRowSortTableModel so
		// we have to wrap it further
		table = new DefaultRowSortTableModel(table);
	}

	// Add table listeners to table model
	table.addTableModelListener(this);
}

/**
 * Set whether we are viewing this as a tree or a table.
 * @param view valid values are TREE and TABLE
 */
public void setView(int view)
{
	if (this.view == view) {
		return;
	}

	if (view != TABLE && view != TREE) {
		throw new IllegalArgumentException("view must be one of TREE or TABLE");
	}

	this.view = view;

	if (view == TABLE) {
		initTableViewModels();
	}

	// Post an event that indicates that the whole thing has changed
	postTableModelChanged(new TableModelEvent(this));
}

/**
 * Return the current view.
 * @return The current view.  Will be one TREE or TABLE.
 * @see #setView
 */
public int getView()
{
	return(view);
}

/**
 * Set the tree/table model that we are using
 * @param treetable The new treetable model
 */
public void setTreeTableModel(JCTreeTableModel treetable)
{
    // If the model is null create a simple default model
    if (treetable == null) {
        treetable =
            new EmptyTreeTableModel(EmptyTreeTableModel.makeDefaultTree());
    }

	if (sortable && view == TREE &&
		(!(treetable instanceof JCRowSortModel)))
	{
		// Sort siblings is turned on and the data source is not sortable so
		// we need to wrap existing data model in a sortable one
		treetable = new TreeWithSortableChildren(treetable);
	}

	if (this.treetable == treetable) {
		return;
	}

	this.treetable = treetable;

	// Install listeners
	treetable.addTreeModelListener(treeListener);

	if (view == TABLE) {
		initTableViewModels();
	}

	//
	// Setup the tree mappings
	//
	// The initial expanded state of the tree is that the
	// root is expanded and its immediate children are visible and
	// not expanded
	//
	// Create the expansionState to track:
	//   Which nodes have been expanded vs. which have not
	//   Which are expanded vs. which are not
	clearExpandedState(false);
}

/**
 * Resets the expanded state hashtable.  Has effect of setting the expanded
 * state back to the default.
 */
public void clearExpandedState()
{
    clearExpandedState(true);
}

/**
 * Resets the expanded state hashtable.  Has effect of setting the expanded
 * state back to the default.
 */
protected void clearExpandedState(boolean post_change_event)
{
    expandedState = new Hashtable();
    mapTreeTable(post_change_event);
    if (visualComponent instanceof JCTreeTable) {
        JCTreeTable tt = (JCTreeTable)visualComponent;
        tt.removeEditor();
        // We'd really like to call "tt.clearSelection()" here.  However,
        // clearing the list selection will not work in all cases,
        // especially if one has selected a node and then calls expandPath().
        // See PR31204.  Comment out the next four lines and then run the test
        // case.  The added nodes will be selected due to a JTable bug and
        // clearing the selection here will not help.
        JCTreeTableSelectionModel ttm = tt.getTreeSelectionModel();
        if (ttm != null) {
            tt.setTreeTableSelectionModel(new DefaultTreeTableSelectionModel());
        }
    }
}

protected void mapTreeTable(boolean post_change_event)
{
	Object root = treetable.getRoot();
	// Create row map for initial entries
	rowMap = new RemovableVector();
	TreePath root_path = new TreePath(root);
	if (rootVisible) {
		// Do not add root to row map if it's not visible
		rowMap.add(root_path);
	}
	else {
		// Root is NOT visible
		// So Mark the root as expanded, if it isn't a leaf.
		if (treetable.isLeaf(root) == false) {
			expandedState.put(root_path, Boolean.TRUE);
		}
	}

	// Get a list of the already expanded children and add them
	Vector expanded_kids = new Vector();
	getExpandedKids(root_path, expanded_kids);
	rowMap.addAll(expanded_kids);

	// Reset the TableModel based on the treetable
    if (table instanceof DefaultRowSortTableModel) {
		DefaultRowSortTableModel rowTable = (DefaultRowSortTableModel)table;
        TableModel rowTableModel = rowTable.getModel();
        if (rowTableModel instanceof TableFromTreeTable) {
            TableFromTreeTable tableFromTreeTable =
                (TableFromTreeTable)rowTableModel;
            tableFromTreeTable.resetMappings();
        }
	}

	if (post_change_event) {
		postTableModelChanged(new TableModelEvent(this));
	}
}

/**
 * Return the treetable model that is the source of all.
 * @return The current treetable model
 */
public JCTreeTableModel getTreeTableModel()
{
	return(treetable);
}

/**
 * Return the table model that corresponds to the treetable model.  Note that
 * this model is a wrapper around the tree table model of type
 * TableFromTreeTable which itself has been wrapped by a
 * DefaultRowSortTableModel for sorting purposes.  The returned model should be
 * viewed as readonly!
 * @see TableFromTreeTable
 * @see DefaultRowSortTableModel
 * @return The current table model
 */
public TableModel getTableModel()
{
    return(table);
}

/**
 * Returns true if the node identified by the path has ever been expanded (since
 * the last "clearExpandedState" call).
 * @param path The input path
 * @return Whether this path has been expanded
 */
public boolean hasBeenExpanded(TreePath path)
{
	if (path != null) {
		if (expandedState.get(path) != null) {
			// Found a key for this path therefore it has been traversed
			// to before
			return(true);
		}
	}
	return(false);
}

/**
 * Sent the given table model changed event to interested listeners.
 * @param event The event generated by the model change
 */
protected void postTableModelChanged(TableModelEvent event)
{
	if (event == null) {
		return;
	}

	Enumeration e = JCListenerList.elements(tableListeners);
	while (e.hasMoreElements()) {
		
		TableModelListener listener = (TableModelListener)e.nextElement();
		listener.tableChanged(event);
	}
}

/**
 * Set the icon renderer.  If set to null no icon will be drawn
 * @param renderer The new icon renderer
 */
public void setTreeIconRenderer(JCTreeIconRenderer renderer)
{
	this.iconRenderer = renderer;
}

/**
 * Return the icon renderer being used.  If null then no icons are being
 * drawn.
 * @return The current icon renderer
 */
public JCTreeIconRenderer getTreeIconRenderer()
{
	return(iconRenderer);
}

/**
 * Returns the path for the specified row.
 * <!-->If row is not visible null is returned.<-->
 *
 * @param row An int specifying a row
 * @return The TreePath to the specified node, null if row < 0 or
 *     row > getRowCount()
 */
public TreePath getPathForRow(int row)
{
	return(getPathForRow(view, row));
}

/**
 * Returns the path for the specified row.
 * <!-->If row is not visible null is returned.<-->
 *
 * @param viewType The type of view.  Possible values are TREE and TABLE.
 * @param row An int specifying a row
 * @return The TreePath to the specified node, null if row < 0 or
 *     row > getRowCount()
 */
protected TreePath getPathForRow(int viewType, int row)
{
	if (viewType == TREE) {
		return((TreePath) rowMap.elementAt(row));
	}
	else if (table instanceof DefaultRowSortTableModel) {
        DefaultRowSortTableModel rowTable = (DefaultRowSortTableModel)table;
        TableModel rowTableModel = rowTable.getModel();
        if (rowTableModel instanceof TableFromTreeTable) {
            TableFromTreeTable tableFromTreeTable =
                (TableFromTreeTable)rowTableModel;
            int mapped_row = rowTable.getUnsortedRow(row);
            return(tableFromTreeTable.rowToTreePath(mapped_row));
        }
	}
	else {
		// Currently we do not handle this case where the table model
		// is provided by a data source other than our wrapper
	}
	return(null);
}
/**
 * Returns the row that displays the node identified by the specified
 * path.
 *
 * @param path  the TreePath identifying a node
 * @return an int specifying the display row, where 0 is the first
 *         row in the display, or -1 if any of the elements in path
 *         are hidden under a collapsed parent.
 */
public int getRowForPath(TreePath path)
{
	return(getRowForPath(view, path));
}

/**
 * Returns the row that displays the node identified by the specified
 * path.
 *
 * @param viewType The type of view.  Possible values are TREE and TABLE.
 * @param path  the TreePath identifying a node
 * @return an int specifying the display row, where 0 is the first
 *         row in the display, or -1 if any of the elements in path
 *         are hidden under a collapsed parent.
 */
protected int getRowForPath(int viewType, TreePath path)
{
	if (viewType == TREE) {
	   // For now we brute force the lookup. This is very inefficient so we
	   // should avoid using the method in this case if possible
	    for (int i = 0; i < rowMap.size(); i++) {
		    if (rowMap.elementAt(i).equals(path)) {
			    return(i);
		    }
	    }
	}
	else if (table instanceof DefaultRowSortTableModel) {
        DefaultRowSortTableModel rowTable = (DefaultRowSortTableModel)table;
        TableModel rowTableModel = rowTable.getModel();
	    if (rowTableModel instanceof TableFromTreeTable) {
		    TableFromTreeTable tableFromTreeTable =
                (TableFromTreeTable)rowTableModel;
		    int row = tableFromTreeTable.treePathToRow(path);
		    return(rowTable.getSortedRow(row));
        }
	}
	else {
		// Currently we do not handle this case where the table model
		// is provided by a data source other than our wrapper
		// ???
	}
	return(-1);
}

/**
 * Returns true if the node identified by the path is currently expanded,
 *
 * @param path  the TreePath specifying the node to check
 * @return false if any of the nodes in the node's path are collapsed,
 *         true if all nodes in the path are expanded
 */
public boolean isExpanded(TreePath path)
{
	if (path == null) {
	    return(false);
	}

	// Is this node expanded?
	Object state = expandedState.get(path);

	if (state == null) {
		// The node has never been expanded
		return(false);
	}

	if (((Boolean)state).booleanValue() == false) {
		// The node is not expanded
	    return(false);
	}

	// The node is expanded but we need to make sure that all it's parents
	// are expanded as well
	TreePath parent_path = path.getParentPath();

	if (parent_path != null) {
		// Recursively invoke this method on the parent
	    return(isExpanded(parent_path));
	}

	return(true);
}

//
// JCRowSortModel implementation
//

/**
 * Returns whether the data is automatically sorted when it changes
 * according the current comparator. If sortable is false, this will also
 * return false
 *
 * @return The current value of autoSort
 */
public boolean getAutoSort()
{
	// If the treetable isn't sortable, it isn't autosortable either.
	if (!sortable) {
		autoSort = false;
	}
	return(autoSort);
}

/**
 * Sets whether the data should be automatically sorted when it changes
 * @param autoSort If true, the data will be automatically sorted.
 */
public void setAutoSort(boolean autoSort)
{
	if (this.autoSort == autoSort) {
		return;
	}
	this.autoSort = autoSort;
	JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
		rsm.setAutoSort(autoSort);
	}
}

/**
 * Set the key columns used to sort the table model when clicking
 * on specified column.
 * @param column index of column in TableModel
 * @param keyColumns columns to use as keys to sort when sorting
 *     on the specified column.
 */
public void setKeyColumns(int column, int keyColumns[])
{
    if (keyColumnsArray == null) {
		keyColumnsArray = new int[treetable.getColumnCount()][];
    }
	keyColumnsArray[column] = keyColumns;
	JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
		rsm.setKeyColumns(column, keyColumns);
	}
}

/**
 * Return the key columns used to sort the table model when clicking
 * on the specified column.  If the user has not set the key columns
 * the this returns an array of length 1 with the specified column
 * as the only member.
 * @param column The column for which the key columns are required
 * @return The key columns for the given column
 */
public int[] getKeyColumns(int column)
{
	return(keyColumnsArray[column]);
}

/**
 * Routine that returns a JCRowSortModel if the cirrcumstances warrant it.
 * @return The model for sorting rows (different for TREE and TABLE view)
 */
protected JCRowSortModel getRowSortModel()
{
	if (sortable && view == TREE && treetable instanceof JCRowSortModel) {
		return((JCRowSortModel) treetable);
	}
	else if (sortable && view == TABLE && table != null) {
		return((JCRowSortModel) table);
	}
	return(null);
}

/**
 * Returns whether or not the data set is currently sorted
 * @return Is the data currently sorted (this could be a different answer
 *    depending on the current view)?
 */
public boolean isDataSorted()
{
	JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
	   return(rsm.isDataSorted());
	}
	return(false);
}

/**
 * Sorts data by current column if set to true and the data needs sorting;
 * or if set to false it returns the data set to its native unsorted state
 * @param sorted Causes a sort (if needed) if true and an unsort if false.
 */
public void setDataSorted(boolean sorted)
{
	JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
		rsm.setDataSorted(sorted);
	}
}

/**
 * Gets the sorting column. This should return -1 if the data is unsorted
 * @return The current sorting column (-1 if the data is unsorted).
 */
public int getSortColumn()
{
    JCRowSortModel rsm = getRowSortModel();
    if (rsm != null) {
        return(rsm.getSortColumn());
    }
    return(-1);
}

/**
 * Set the column to use the next time a sort is performed
 * @param column only valid column indicies are allowed
 */
public void setSortColumn(int column)
{
	if (sortable == false) {
		// Cannot set the sort column if the sort policy is none
		return;
	}
    if (getSortColumn() == column) {
        return;
    }

    JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
		rsm.setSortColumn(column);
	}
}

/**
 * Gets the sorting order to use the next time a sort is performed
 * @return The current sorting order.  Will be one of JCRowSortModel.FORWARD or
 *     JCRowSortModel.BACKARD.
 */
public int getSortOrder()
{
	return(sortOrder);
}

/**
 * Sets the sorting order.
 * @param order Sets the sort order to JCRowSortModel.FORWARD or
 *     JCRowSortModel.BACKWARD.
 */
public void setSortOrder(int order)
{
	if (sortOrder == order) {
		return;
	}
	JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
		rsm.setSortOrder(order);
	}
    this.sortOrder = order;
}

/**
 * Toggles the sorting order.
 */
public void toggleSortOrder()
{
    sortOrder = (sortOrder == FORWARD) ? REVERSE : FORWARD;
	JCRowSortModel rsm = getRowSortModel();
	if (rsm != null) {
		rsm.setSortOrder(sortOrder);
	}
}

//
// JCRowSortTableModel Implmenatation
//

/**
 * Returns the index of the unsorted row for the table view
 * @param sortRow The sorted row
 * @return The unsorted row based on the sorted row.  Returns a negative value
 *     on error.
 */
public int getUnsortedRow(int sortRow)
{
    if (table == null) {
        return(-1);
    }
    JCRowSortTableModel sortedModel = (JCRowSortTableModel)table;
    return(sortedModel.getUnsortedRow(sortRow));
}

/**
 * Returns the index of the unsorted row for the table view
 * @param row The unsorted row
 * @return The sorted row based on the unsorted row.  Returns a negative value
 *     on error.
 */
public int getSortedRow(int row)
{
    if (table == null) {
        return(-1);
    }
    JCRowSortTableModel sortedModel = (JCRowSortTableModel)table;
    return(sortedModel.getSortedRow(row));
}

//
// TableModelListener to listen to sortable tables
//

/**
 * Called when the table's model is changed
 * @param e The event generated by the change
 */
public void tableChanged(TableModelEvent e)
{
	TableModelEvent ne = new TableModelEvent(this, e.getFirstRow(),
											 e.getLastRow(), e.getColumn(),
											 e.getType());
	// Repost event to anybody listening to us
	Enumeration enum = JCListenerList.elements(tableListeners);
	while (enum.hasMoreElements()) {
		TableModelListener listener = (TableModelListener)enum.nextElement();
		listener.tableChanged(ne);
	}
}


class TreeListener implements TreeModelListener, java.io.Serializable {
    /**
     * <p>Invoked after a node (or a set of siblings) has changed in some
     * way. The node(s) have not changed locations in the tree or
     * altered their children arrays, but other attributes have
     * changed and may affect presentation. Example: the name of a
     * file has changed, but it is in the same location in the file
     * system.</p>
     *
     * <p>e.path() returns the path the parent of the changed node(s).</p>
     *
     * <p>e.childIndices() returns the index(es) of the changed node(s).</p>
	 * @param e The event generated by the change
     */
    public void treeNodesChanged(TreeModelEvent e)
	{
		// PENDING
		// currently since we dont keep a hash of rows to nodes we
		// change everything if the underlying tree structure changes
		mapTreeTable(true);
	}

    /**
     * <p>Invoked after nodes have been inserted into the tree.</p>
     *
     * <p>e.path() returns the parent of the new nodes
     * <p>e.childIndices() returns the indices of the new nodes in
     * ascending order.
	 * @param e The event generated by the change
     */
    public void treeNodesInserted(TreeModelEvent e)
	{
		/*
		System.out.println("parent path = " + e.getTreePath().getParentPath());
		if (!hasBeenExpanded(e.getTreePath().getParentPath())) {
			// Parent node has never been expanded so we don't need to handle it
			return;
		}
		if (isExpanded(e.getTreePath().getParentPath())) {
			// PENDING
			// currently since we dont keep a hash of rows to nodes we
			// change everything if the uderlying tree structure changes
			mapTreeTable(true);
		}
		*/
		mapTreeTable(true);
	}

    /**
     * <p>Invoked after nodes have been removed from the tree.  Note that
     * if a subtree is removed from the tree, this method may only be
     * invoked once for the root of the removed subtree, not once for
     * each individual set of siblings removed.</p>
     *
     * <p>e.path() returns the former parent of the deleted nodes.</p>
     *
     * <p>e.childIndices() returns the indices the nodes had before they
	 * were deleted in ascending order.</p>
	 * @param e The event generated by the change
     */
    public void treeNodesRemoved(TreeModelEvent e)
	{
		/*
		if (!hasBeenExpanded(e.getTreePath().getParentPath())) {
			// parent node has never been expanded so we don't nee to handle it
			return;
		}
		if (isExpanded(e.getTreePath().getParentPath())) {
			// PENDING
			// currently since we dont keep a hash of rows to nodes we
			// change everything if the uderlying tree structure changes
			// We should probably remove
			mapTreeTable(true);
		}
		*/
		mapTreeTable(true);

		// PENDING
		// We should also probably clean up any removed nodes from the
		// expanded hash map.  The easiest way to do this
		// would be to go through the whole hash map and delete any elements
		// that have this path in them
	}

    /**
     * <p>Invoked after the tree has drastically changed structure from a
     * given node down.  If the path returned by e.getPath() is of length
     * one and the first element does not identify the current root node
     * the first element should become the new root of the tree.<p>
     *
     * <p>e.path() holds the path to the node.</p>
     * <p>e.childIndices() returns null.</p>
 	 * @param e The event generated by the change
    */
    public void treeStructureChanged(TreeModelEvent e)
	{
		// PENDING
		// We should also probably blow any information on the children
		// nodes away from the expanded hash map.  The easiest way to do this
		// would be to go through the whole hash map and delete any elements
		// that have this path in them
		mapTreeTable(true);
	}

} // TreeListener

class ToggleViewAction extends JCAction {

	public ToggleViewAction()
	{
		super("");
		// PENDING
		// specify small and large icons here!
	}

	public void actionPerformed(ActionEvent e)
	{
		if (view == TABLE) {
			setView(TREE);
		}
		else {
			setView(TABLE);
		}
	}

}

//
// The following methods are the implementation of the TableModel
// interface that is provided to a Table as the nodes we are showing
// to the user.
//

public void addTableModelListener(TableModelListener l)
{
	tableListeners = JCListenerList.add(tableListeners, l);
}

public void removeTableModelListener(TableModelListener l)
{
	tableListeners = JCListenerList.remove(tableListeners, l);
}

public Class getColumnClass(int column)
{
	Class column_class = null;
    if (view == TREE && column == getTreeColumn()) {
		column_class = TreePath.class;
	}
	else {
		column_class = treetable.getColumnClass(column);
	}

	return(column_class);
}

public int getColumnCount()
{
	return(treetable.getColumnCount());
}

public String getColumnName(int column)
{
	return(treetable.getColumnName(column));
}

public int getRowCount()
{
	if (view == TREE) {
		// return the number of visible nodes
		return(rowMap.size());
	}
	else if (table != null) {
		return(table.getRowCount());
	}
	else {
		return(0);
	}
}

public Object getValueAt(int row, int column)
{
	if (view == TREE) {
        if (column == getTreeColumn()) {
			// Return the TreePath stored at the specified row
			return(rowMap.elementAt(row));
		}
		else {
			return(treetable.getValueAt(getNode(row), column));
		}
	}
	else if (table != null) {
		return(table.getValueAt(row, column));
	}
	else {
		return(null);
	}
}

public boolean isCellEditable(int row, int column)
{
	if (view == TREE) {
		return(treetable.isCellEditable(getNode(row), column));
	}
	else if (table != null) {
		return(table.isCellEditable(row, column));
	}
	else {
		return(false);
	}
}

public void setValueAt(Object value, int row, int column)
{
	if (view == TREE) {
		treetable.setValueAt(value, getNode(row), column);
	}
	else if (table != null) {
		table.setValueAt(value, row, column);
	}
}

//
// Editor/Renderer Support methods
//

/**
 * Given a tree path, this method returns the parent object or null if
 * no parent exists.
 * @param path The path for which the parent is needed.
 * @return The parent object or null if no parent exists
 * @see #getNode
 */
static protected Object getParent(TreePath path)
{
	if (path.getPathCount() == 1) {
		return(null);
	}
	return(path.getPathComponent(path.getPathCount() - 2));
}

/**
 * Given a TreePath this method returns the last component or "node" of
 * the TreePath.
 * @param path The path from which to get the node or last component
 * @return The last component or node of the path
 * @see #getParent
 */
static protected Object getNode(TreePath path)
{
	return(path.getLastPathComponent());
}

/**
 * Returns the node that maps to the specified row of the treetable view.
 * @param row The row for which the node is needed
 * @return The last component of the path for the given row
 */
protected Object getNode(int row)
{
	TreePath path = (TreePath) rowMap.elementAt(row);
	if (path == null) {
		return(null);
	}
	return(path.getLastPathComponent());
}

/**
 * Get the level (how many nodes on the path) for the current row
 * @param row The row for which the level is needed
 * @return The level for the given row
 */
protected int getLevel(int row)
{
	TreePath path = getPathForRow(row);
	return(path.getPathCount());
}

/**
 * Toggle the expanded/collapsed state of a row.  Only TREE view is affected.
 * @param row The row on which the toggle is needed.
 */
protected void toggleState(int row)
{
	if (isExpanded(getPathForRow(TREE, row))) {
		// We need to collapse the tree since we have a click inside the box
		
		collapseRow(row);
	}
	else {
		// We need to expand the tree since we have a click inside the box
		
		expandRow(row);
	}
}

/**
 * Processes a click in the table.
 * @param ev The event from the mouse click
 * @param row The row the point is in
 * @param column The column the point is in
 * @param p The coordinates of the point clicked relative to the cell
 * @param cell_height The cell or row height
 * @return true if the click has been processed
 */
public boolean processClickInTreeColumn(MouseEvent ev, int row, int column,
			                            Point p, int cell_height)
{
	if (row < 0 || column < 0) {
		// Invalid row and column values
		return(false);
	}

	if (view != TREE) {
		// Ignore any events if we are not in tree mode
		return(false);
	}
	if (!SwingUtilities.isLeftMouseButton(ev)) {
		// Not MB1
		return(false);
	}
	if (treetable.isLeaf(getNode(row))) {
		// A leaf node
		return(false);
	}
    if (visualComponent instanceof JCTreeTable) {
        JCTreeTable jctreetable = (JCTreeTable)visualComponent;
        if (jctreetable.isExpandInTreeColumnOnly() &&
            jctreetable.getTreeColumn() != column)
        {
            // Not in the tree column and its required only there
           return(false);
        }
    }

    // In the control box or a multiple click anywhere in the tree column
    if (isPointInControlBox(p, row, column, cell_height, getLevel(row)) ||
        ev.getClickCount() > 1)
    {
		toggleState(row);
		ev.consume();
	}
	else {
		return(false);
	}

	return(true);
}

/**
 * Is the point within the dimensions of the control icon within the tree
 * column cell for a given row.
 * @param p The point to check
 * @param row The row of the cell
 * @param column The column of the cell (must be the tree column)
 * @param cell_height The cell or row height
 * @param level The level (how many nodes on the path) for the current row
 * @return True if the point is within the control icon, false otherwise.
 */
protected boolean isPointInControlBox(Point p, int row, int column,
                                      int cell_height, int level)
{
    level -= 1;
    if (showsRootHandles == false) {
        level -= 1;
    }
    if (rootVisible == false) {
        level -= 1;
    }

    boolean isLeft = true;
    int columnWidth = 0;
    TableColumnModel model = null;
    if (visualComponent instanceof JTable) {
        isLeft = visualComponent.getComponentOrientation().isLeftToRight();
        model = ((JTable)visualComponent).getColumnModel();
        columnWidth = model.getColumn(column).getWidth();
    }

    int iconWidth  = expandedIcon.getIconWidth();
    int iconHeight = expandedIcon.getIconHeight();

    int x = (levelIndent / 2) + (level * levelIndent);
    if (!isLeft) {
        x = columnWidth - x;
    }
    x -= (iconWidth / 2);
    int y = (cell_height / 2) - (iconHeight / 2);
    
    

    if (visualComponent instanceof JCTreeTable) {
        JCTreeTable jctreetable = (JCTreeTable)visualComponent;
        if (column != jctreetable.getTreeColumn()) {
            // Not in the tree column
            return(false);
        }

        int index = isLeft ? 0 : model.getColumnCount() - 1;

        // Need to offset the point.  If it's less than jdk1.3, account for
        // the intercell spacing. Changes in 1.3 make this unnecessary and
        // damaging.
        p.y -= (row * jctreetable.getRowHeight());
        if (JCEnvironment.getJavaVersion() < 130) {
            Dimension spacing = jctreetable.getIntercellSpacing();
            x += (spacing.width / 2);
            while (isLeft ? index < column : index > column) {
                x += model.getColumn(index).getWidth() + spacing.width;
                index = isLeft ? index + 1 : index - 1;
            }
            p.y -= (row * spacing.height);
        }
        else {
            while (isLeft ? index < column : index > column) {
                x += model.getColumn(index).getWidth();
                index = isLeft ? index + 1 : index - 1;
            }
        }
        
    }

    boolean statusX = (p.x >= x) && (p.x < (x + iconWidth));
    boolean statusY = (p.y >= y) && (p.y < (y + iconHeight));
    return(statusX && statusY);
}

/**
 * Add children of the parent whose path is in parent_path to the new_kids vector
 * @param parent The parent which is assumed to be the last component of the
 *    passed in parent_path
 * @param parent_path The path of the parent
 * @param new_kids The vector which holds the list of children
 */
protected void addKids(Object parent, TreePath parent_path, Vector new_kids)
{
	int num_kids = treetable.getChildCount(parent);
	for (int i = 0; i < num_kids; i++) {
		Object child = treetable.getChild(parent, i);
		TreePath child_path = parent_path.pathByAddingChild(child);
		new_kids.add(child_path);

		if (isExpanded(child_path)) {
			// Recursively descend expanded tree
			addKids(child, child_path, new_kids);
		}
	}
}

/**
 * Ensures that the node in the specified row is expanded and
 * viewable. <p> If <code>row</code> is < 0 or >= getRowCount this
 * will have no effect.
 *
 * @param row  an int specifying a display row, where 0 is the
 *             first row in the display
 */
public void expandRow(int row)
{
	expandPathRow(getPathForRow(TREE, row), row);
}

protected int countKidsToRemove(Object parent, TreePath parent_path)
{
	int num_kids = treetable.getChildCount(parent);
	int total_kids = num_kids;
	for (int i = 0; i < num_kids; i++) {
		Object child = treetable.getChild(parent, i);
		TreePath child_path = parent_path.pathByAddingChild(child);
		if (isExpanded(child_path)) {
			// Recursively descend expanded tree
			total_kids += countKidsToRemove(child, child_path);
		}
	}

	return(total_kids);
}

/**
 * Ensures that the node in the specified row is collapsed.
 * <p> If <code>row</code> is < 0 or >= getRowCount this
 * will have no effect.
 *
 * @param row  an int specifying a display row, where 0 is the
 *             first row in the display
 */
public void collapseRow(int row)
{
	collapsePathRow(getPathForRow(TREE, row), row);
}

public void collapsePathRow(TreePath path, int row)
{
	if (!isExpanded(path)) {
		return;
	}

	TreeExpansionEvent event = new TreeExpansionEvent(visualComponent,
													  path);
	// Post treeWillCollapse event and handle potential veto
	Enumeration e = JCListenerList.elements(treeWillExpandListeners);
	while (e.hasMoreElements()) {
		TreeWillExpandListener listener =
            (TreeWillExpandListener)e.nextElement();
		try {
			listener.treeWillCollapse(event);
		}
		catch (ExpandVetoException eve) {
			// if the listener vetos the collapse then we don't do it
			return;
		}
	}

	int deletion_count = countKidsToRemove(getNode(row), path);
	// We must collapse the row _after_ we count the children
	expandedState.put(path, Boolean.FALSE);

	rowMap.remove(row + 1, row + 1 + deletion_count);

	// Post table changed event
    TableModelEvent tme = new TableModelEvent(this,
                                              row + 1, row + deletion_count,
                                              TableModelEvent.ALL_COLUMNS,
                                              TableModelEvent.DELETE);
	postTableModelChanged(tme);

	// Post treeCollapsed event
	e = JCListenerList.elements(treeExpansionListeners);
	while (e.hasMoreElements()) {
		TreeExpansionListener listener = (TreeExpansionListener)e.nextElement();
		listener.treeCollapsed(event);
	}
}

/**
 * Figure out if the current node is the last of it's siblings
 * we need to know this so we can figure out whether to draw a
 * continuation line or not
 * @param treetable The tree tabl model to do the check
 * @param path The path of the child node to test
 * @return Whether the last component or node of the path is the last child
 *     of its parent
 */
static protected boolean isLastChild(JCTreeTableModel treetable,
									 TreePath path)
{
	Object parent = getParent(path);

	if (parent == null) {
		// The node passed in is the root node therefore it cannot
		// have any children
		return(true);
	}

	// Note that we don't check for child count because it must be
	// > 0 since we know that "node" is one of the children
	if (treetable.getIndexOfChild(parent, getNode(path)) ==
		(treetable.getChildCount(parent) - 1))
	{
		// Node is the last child
		return(true);
	}
	return(false);
}

/**
 * Set the color and fill an area of width by height with that color
 * @param gc The gc to use
 * @param width The width to fill
 * @param height The height to fill
 */
public void prepareCellForRender(Graphics gc, int width, int height)
{
	// Do a clear rect since returning isOpaque() == false
	// only does half the job
	gc.setColor(tableBackground);
	gc.fillRect(0, 0, width, height);
}

/**
 * Return the icon for the node of a given path.
 * @param path The path for which the node icon is required.
 * @param value The value at node
 * @param value_class The class of the value at the node
 * @param is_leaf Is the node a leaf?
 * @param is_expanded Has the node been expanded?
 * @return The icon a the node of a given path.
 */
protected Icon getNodeIcon(TreePath path, Object value, Class value_class,
                           boolean is_leaf, boolean is_expanded)
{
    if (iconRenderer == null) {
        return(null);
    }
    Icon plaf_icon = null;
    if (is_leaf) {
        plaf_icon = leafIcon;
    }
    else if (is_expanded) {
        plaf_icon = openIcon;
    }
    else {
        plaf_icon = closedIcon;
    }

    Icon node_icon = iconRenderer.getNodeIcon(treetable, getNode(path),
                                              value, value_class,
                                              is_leaf, is_expanded,
                                              plaf_icon);
    return(node_icon);
}

/**
 * Used to draw the cell. Note that the size of the cell comes from
 * drawing_area.
 *
 * @param gc Graphics context to use for drawing.
 * @param drawing_area The rectangle into which to draw
 * @param path The path to the node to draw
 * @param value The value of the cell
 * @param value_class The class of the value of the cell
 *
 * Render the visual hierarchy informaion
 */
public void renderHierarchy(Graphics gc, Rectangle drawing_area,
						    TreePath path, Object value, Class value_class)
{
    
	// Set line draw color
	gc.setColor(lineColor);
	Stroke saved_stroke = null;
	if (lineStroke != null && gc instanceof Graphics2D) {
		saved_stroke = ((Graphics2D) gc).getStroke();
		((Graphics2D) gc).setStroke(lineStroke);
	}

	int level = path.getPathCount() - 1; // hierarchical level of the node
	boolean is_expanded = isExpanded(path);
	boolean is_leaf = treetable.isLeaf(getNode(path));

	int finish_at = 0;
	if (rootVisible) {
		finish_at = 0;
	}
	else if (showsRootHandles == false) {
		// Root is not visible and we are not showingRootHandles
		// which if the root is not visible means that we don't
		// show handles on the children of the root
		finish_at = 1;
	}

    boolean isLeft = visualComponent.getComponentOrientation().isLeftToRight();

	// Draw a vertical line through the cell for each parent that
	// is not the _last_ of it's siblings.
	int top_y = drawing_area.y;
	int halfway_y = (drawing_area.height / 2) + top_y;
	int bottom_y = drawing_area.height + top_y;
    int width = drawing_area.width;
    int halfLevelIndent = levelIndent / 2;
	int x = -1;

	// We start at 1 because we draw the last node stuff later
	// and the rendered node later
	TreePath parent_path = path;
	int node_indent = getNodeIndent(path);
	for (int i = 1; i < level - finish_at; i++) {
		parent_path = parent_path.getParentPath();
		if (isLastChild(treetable, parent_path) == false) {
			// Draw line since this node has siblings that need connecting to
			x = node_indent -         // starting point
				(i * levelIndent) -   // minus levelIndent for each level
				halfLevelIndent;      // minus half a level
            if (!isLeft) {
                x = width - x;
            }
            
			if (drawLines) {
				gc.drawLine(x, top_y, x, bottom_y);
			}
		}
	}

	if (!(level == 0 && showsRootHandles == false) &&
		!(level == 1 && rootVisible == false && showsRootHandles == false))
	{
		x = node_indent - halfLevelIndent;
        if (!isLeft) {
            x = width - x;
        }
		// Don't draw the last control node if
		//    it is the root node and the root handles are not supposed
		//    to show
		//    OR it is a child of the root node and the root handles are
		//    not supposed to show
		if (drawLines) {
			//
			// Draw the line passed the current node
			//
			if (isLastChild(treetable, path)) {
				// This is the last child so we only draw the connecting line
				// halfway
				gc.drawLine(x, top_y, x, halfway_y);
			}
			else {
				// Child has siblings that follow so draw the line all the way
				// through the cell
				gc.drawLine(x, top_y, x, bottom_y);
			}
		}

		//
		//  Draw the folder control aka handle
		//

		if (drawLines) {
			//
			// Draw the horizontal line to the node
			//
            int offset = isLeft ? halfLevelIndent : -halfLevelIndent;
			gc.drawLine(x, halfway_y, x + offset, halfway_y);
		}

		if (lineStroke != null && gc instanceof Graphics2D) {
			// Restore original stroke
			((Graphics2D) gc).setStroke(saved_stroke);
		}

		if (is_leaf == false) {
			//
			// Draw the folder state control if the node is a folder
			//
            Icon icon = is_expanded ? expandedIcon : collapsedIcon;
            int startX = x - (icon.getIconWidth() / 2);
            int startY = halfway_y - (icon.getIconHeight() / 2);
		    icon.paintIcon(visualComponent, gc, startX, startY);
		}
	}

	//
	// Draw the icon if the icon renderer is not null
	//
	if (iconRenderer == null) {
		return;
	}

    Icon node_icon = getNodeIcon(path, value, value_class, is_leaf,
                                 is_expanded);
    if (node_icon != null) {
        x = isLeft ? node_indent
                   : width - node_indent - node_icon.getIconWidth();
		boolean icon_bug = false;
        String treeLeaf = JCEnvironment.MOTIF_TREE_LEAF_ICON;
		if (isMotif && is_leaf &&
			JCEnvironment.isWorking(treeLeaf) == JCEnvironment.WORKING_NO)
		{
			icon_bug = true;
			// Translate by x
			gc.translate(x, 0);
		}

		node_icon.paintIcon(visualComponent, gc, x, drawing_area.y);

		if (icon_bug) {
			// Translate back
			gc.translate(-x, 0);
		}
	}
}

    /**
 * Returns the <i>x</i> offset at which we can begin drawing the value of the
 * heirarchy column.
 * @param path The path for which the indent is required
 * @return The indent of the given path
 */
public int getNodeIndent(TreePath path)
{
	if (path == null) {
		return(0);
	}
	int levels = path.getPath().length;
	if (rootVisible == false) {
		// Subtract 1 because the root is not visible
		levels -= 1;
	}
	if (showsRootHandles == false) {
		// We are not drawing root handles so subtract an a additional "1"
		levels -= 1;
	}
	return(levels * getLevelIndent());
}

/**
 * Returns the <i>x</i> offset at which we need to place the renderer.
 * This is usually getNodeIndent + the width of the plaf icons if
 * TreeIconRenderer is not null.
 * @param path The path for which the renderer indent is required
 * @deprecated As of JClass DesktopViews 6.3.  Use
 * getRenderIndent(path, value, value_class) instead.
 * @return The render indent for the given path.
 */
public int getRendererIndent(TreePath path)
{
    return(getRendererIndent(path, null, null));
}

/**
 * Returns the <i>x</i> offset at which we need to place the renderer.
 * This is usually getNodeIndent + the width of the plaf icons if
 * TreeIconRenderer is not null.
 * @param path The path for which the renderer indent is required
 * @param value The value at node
 * @param value_class The class of the value at the node
 * @return The render indent for the given path
 */
public int getRendererIndent(TreePath path, Object value, Class value_class)
{
	if (path == null) {
		return(0);
	}
	int indent = getNodeIndent(path);
	if (iconRenderer != null) {
        Icon nodeIcon = getNodeIcon(path, value, value_class,
                                    treetable.isLeaf(getNode(path)),
                                    isExpanded(path));
        indent += nodeIcon.getIconWidth();
	}
	return(indent);
}

/**
 * Get the width of a single indent level.
 * @return The width of an indent level
 */
public int getLevelIndent()
{
	return(levelIndent);
}

/**
 * Returns the true class stored in the tree column.
 * @return The class of the tree column
 */
public Class getTreeColumnClass()
{
    return(treetable.getColumnClass(getTreeColumn()));
}

/**
 * Returns the true value stored in the tree column.
 * @param node The node for which the value is required
 * @return The value of the node stored in the tree column
 */
public Object getTreeColumnValue(Object node)
{
    return(treetable.getValueAt(node, getTreeColumn()));
}

/**
 * Get the tree column index
 * @return The current tree column index
 */
public int getTreeColumn()
{
    return(treeColumn);
}

class RemovableVector extends Vector
{
	public void remove(int from, int to)
    {
		removeRange(from, to);
	}
}

/**
 * Returns an Enumeration of the descendants of <code>path</code> that
 * are currently expanded. If <code>path</code> is not currently
 * expanded, this will return null. If you expand/collapse nodes while
 * iterating over the returned Enumeration this may not return all
 * the expanded paths, or may return paths that are no longer expanded.
 * @param parent The path of the parent whose descendants are required
 * @return An enumeration of the descendants of the given path that are expanded
 */
public Enumeration getExpandedDescendants(TreePath parent)
{
	if (!isExpanded(parent)) {
		return(null);
	}

	// Build a vector that stores the expanded descendents of the parent path
	Vector expanded_kids = new Vector();
	getExpandedKids(parent, expanded_kids);
	return(expanded_kids.elements());
}

/**
 * Store the paths of the children which are expanded for a given parent in a
 * vector
 * @param parent The path of the parent
 * @param kids The place in which to store the expanded children
 */
protected void getExpandedKids(TreePath parent, Vector kids)
{
	if (isExpanded(parent) == false) {
		return;
	}

	Object node = parent.getLastPathComponent();
	int nkids = treetable.getChildCount(node);
	for (int i = 0; i < nkids; i++) {
        Object childNode = treetable.getChild(node, i);
		TreePath child = parent.pathByAddingChild(childNode);
		// Add the child
		kids.add(child);
		if (!treetable.isLeaf(childNode)) {
			// Add the childs children by recursively calling this method
			getExpandedKids(child, kids);
		}
	}
}

//
// JTree Methods
//

/**
 * Adds a listener for TreeExpansion events.
 *
 * @param l a TreeExpansionListener that will be notified when
 *            a tree node is expanded or collapsed (a "negative
 *            expansion")
 */
public void addTreeExpansionListener(TreeExpansionListener l)
{
	treeExpansionListeners = JCListenerList.add(treeExpansionListeners, l);
}

/**
 * Removes a listener for TreeExpansion events.
 *
 * @param l the TreeExpansionListener to remove
 */
public void removeTreeExpansionListener(TreeExpansionListener l)
{
	treeExpansionListeners = JCListenerList.remove(treeExpansionListeners, l);
}

/**
 * Adds a listener for TreeWillExpand events.
 *
 * @param l a TreeWillExpandListener that will be notified when
 *            a tree node will be expanded or collapsed (a "negative
 *            expansion")
 */
public void addTreeWillExpandListener(TreeWillExpandListener l)
{
	treeWillExpandListeners = JCListenerList.add(treeWillExpandListeners, l);
}

/**
 * Removes a listener for TreeWillExpand events.
 *
 * @param l the TreeWillExpandListener to remove
 */
public void removeTreeWillExpandListener(TreeWillExpandListener l)
{
	treeWillExpandListeners = JCListenerList.remove(treeWillExpandListeners, l);
}

/**
 * Ensures that the node identified by the specified path is
 * collapsed and viewable.
 *
 * @param path The TreePath identifying a node
 */
public void collapsePath(TreePath path)
{
	collapsePathRow(path, getRowForPath(TREE, path));
}

/**
 * Ensures that the node identified by the specified path is expanded and
 * viewable.
 *
 * @param path The TreePath identifying a node
 */
public void expandPath(TreePath path)
{
	if (isExpanded(path)) {
		// Node is already expanded and viewable so return
	}
    int rowForPath = getRowForPath(TREE, path);
    expandPathRow(path, rowForPath);
}

/**
 * Ensures that the node identified by the specified path is expanded and
 * viewable.
 *
 * @param path The TreePath identifying a node
 * @param row The row this path is displayed at
 */
protected void expandPathRow(TreePath path, int row)
{
	if (isExpanded(path)) {
		return;
	}

	if (row == -1) {
		// This means that this path's parents need expanding as well
		// so we travel down from the root expanding the unexpanded paths.
		// We only travel down to the next to last path since we are already
		// processing the last path in this method
		// We start from the root only if it is visible; else we start from
		// the node that follows the root
		TreePath parent_path = null;
		int row_of_parent = -1;
		for (int i = rootVisible ? 0 : 1; i < path.getPathCount() - 1; i++) {
			Object child_node = path.getPathComponent(i);
			if (parent_path == null) {
				parent_path = new TreePath(path.getPathComponent(0));
				if (i != 0) {
					parent_path = parent_path.pathByAddingChild(child_node);
				}
				// The following will be either -1 or 0 depending on
				// whether the root node is visible or not
				row_of_parent = getRowForPath(TREE, parent_path);
			}
			else {
				parent_path =
                    parent_path.pathByAddingChild(path.getPathComponent(i));
				row_of_parent = getRowForPath(TREE, parent_path);
			}

			if (!isExpanded(parent_path)) {
				expandPathRow(parent_path, row_of_parent);
			}
		}

		// Quickly calculate index of the child
		// PENDING if we develop a better method for storing indexes
		// in the future then we should use that method
        int n = path.getPathCount();
        if (n > 1) {
    		row += (row_of_parent + 2 +
			        treetable.getIndexOfChild(path.getPathComponent(n - 2),
					                          path.getPathComponent(n - 1)));
        }
	}
    int nextRow = row + 1;

	TreeExpansionEvent event = new TreeExpansionEvent(visualComponent, path);
	// Post treeWillExpand event and handle potential veto
	Enumeration e = JCListenerList.elements(treeWillExpandListeners);
	while (e.hasMoreElements()) {
		TreeWillExpandListener listener =
            (TreeWillExpandListener)e.nextElement();
		try {
			listener.treeWillExpand(event);
		}
		catch (ExpandVetoException eve) {
			// If the listener vetos the collapse then we don't do it
			return;
		}
	}

	Object node = getNode(path);

	// Need to set this nodes expanded state to true
	expandedState.put(path, Boolean.TRUE);

	if (isExpanded(path) == false) {
		// Don't add the node to the expanded row map if one of its parents
		// is not expanded
		return;
	}

    // This section corrects a situation with JTable where adding rows in the
    // location of a selected row will cause all the added rows to become
    // selected. This happens even when the table is set to single selection.
    // <Bug 10243>
    TreePath            selectedPath = null;
    boolean             singleSelectionCorrection = false;
    boolean             multiSelectionCorrection = false;
	JTable              vcTable = null;
	ListSelectionModel  lsm = null;
    if (visualComponent instanceof JTable) {
	    vcTable = (JTable)visualComponent;
		lsm     = vcTable.getSelectionModel();
    }
	if (vcTable != null && view == TREE) {
		if (lsm.isSelectedIndex(nextRow)) {
            selectedPath = getPathForRow(nextRow);
            if (lsm.getSelectionMode() == ListSelectionModel.SINGLE_SELECTION) {
                singleSelectionCorrection = true;
            }
            else {
                multiSelectionCorrection = true;
            }
		}
    }
    // </Bug 10243>

	Vector new_kids = new Vector();
	addKids(node, path, new_kids);

	// Insert new kids into rowmap
	rowMap.addAll(nextRow, new_kids);

	// Post table changed event
	TableModelEvent tme = new TableModelEvent(this,
											  nextRow, row + new_kids.size(),
											  TableModelEvent.ALL_COLUMNS,
											  TableModelEvent.INSERT);
	postTableModelChanged(tme);

	// Post treeExpanded event
	e = JCListenerList.elements(treeExpansionListeners);
	while (e.hasMoreElements()) {
		TreeExpansionListener listener = (TreeExpansionListener)e.nextElement();
		listener.treeExpanded(event);
	}

    // <Bug 10243> This will reselect the currently selected row, removing all
	// the erroniously selected rows.
	if (vcTable != null && view == TREE) {
	    int selectedPathRow = getRowForPath(selectedPath);
        if (singleSelectionCorrection) {
            vcTable.setRowSelectionInterval(selectedPathRow, selectedPathRow);
        }
        else if (multiSelectionCorrection) {
            if (selectedPathRow - nextRow > 0) {   // Anything added?
                int anchor = lsm.getAnchorSelectionIndex();
                int lead   = lsm.getLeadSelectionIndex();
                // Only do correction if selection range is after inserted rows.
                // Must reset the anchor and lead after removing the selection
                // interval as this call modifies the anchor & lead.
                if (anchor >= selectedPathRow && lead >= selectedPathRow) {
                    lsm.removeSelectionInterval(nextRow, selectedPathRow - 1);
                    lsm.setAnchorSelectionIndex(anchor);
                    lsm.setLeadSelectionIndex(lead);
                }
            }
        }
	}
    // </Bug 10243>

    // The scrollsOnExpand feature is only enabled if we are within a JScrollPane
    if (getScrollsOnExpand() &&
	    vcTable != null      &&
        (visualComponent.getParent() instanceof JViewport))
	{
        JViewport viewport = (JViewport)(visualComponent.getParent());
        if (!(viewport.getParent() instanceof JScrollPane)) {
            return;
        }
        Point p = viewport.getViewPosition();
        int viewportHeight = viewport.getExtentSize().height;
        Rectangle lastChildRect = vcTable.getCellRect(row + new_kids.size(),
								                      0, true);
        if (lastChildRect.y + lastChildRect.height <= p.y + viewportHeight) {
            // No scrolling necessary
            return;
        }
        Rectangle originalRowRect = vcTable.getCellRect(row, 0, true);
        Rectangle rect = new Rectangle(lastChildRect);
        rect.height = viewportHeight;
        // TotalHeight is the union of the two rectangles' height
        int totalHeight = (lastChildRect.y - originalRowRect.y) +
		                  lastChildRect.height;
        // Will both of the rows fit into the current view simultaneously?
        if (totalHeight <= viewportHeight) {
            // Yup, scroll the original row just enough to show the last
            // row completely
            rect.y = lastChildRect.y + lastChildRect.height -
			         viewportHeight - p.y;
        }
        else {
            // Nope, place the original row as the top row
            rect.y = originalRowRect.y - p.y;
        }
        // Scroll it into view
        viewport.scrollRectToVisible(rect);
    }
}

/**
 * Returns true if the tree scrolls to show previously hidden children.
 * @return true if when a node is expanded as many of the descendants
 * as possible are scrolled to be visible.
 */
public boolean getScrollsOnExpand()
{
	return(scrollsOnExpand);
}

/**
 * Determines whether or not when a node is expanded, as many of
 * the descendants are scrolled to be inside the viewport as
 * possible. The default is true.
 * @param newValue The new value of scrollsOnExpand
 */
public void setScrollsOnExpand(boolean newValue)
{
	boolean old_value = scrollsOnExpand;
	scrollsOnExpand = newValue;
	visualComponent.firePropertyChange(JTree.SCROLLS_ON_EXPAND_PROPERTY,
									   old_value, newValue);
}

/**
 * Returns true if the root node of the tree is displayed.
 *
 * @return true if the root node of the tree is displayed
 * @see #rootVisible
 */
public boolean isRootVisible()
{
	return(rootVisible);
}

/**
 * Determines whether or not the root node from the TreeModel is visible.
 *
 * @param rootVisible true if the root node of the tree is to be displayed
 * @see #rootVisible
 */
public void setRootVisible(boolean rootVisible)
{
	boolean oldValue = this.rootVisible;

	this.rootVisible = rootVisible;
	visualComponent.firePropertyChange(JTree.ROOT_VISIBLE_PROPERTY,
									  oldValue, rootVisible);
	if (treetable != null) {
		// Recalculate row mappings
		mapTreeTable(true);
	}
}

/**
 * Returns true if handles for the root nodes are displayed.
 *
 * @return true if root handles are displayed
 * @see #showsRootHandles
 */
public boolean getShowsRootHandles()
{
	return(showsRootHandles);
}

/**
 * Determines whether the node handles are to be displayed.
 *
 * @param newValue true if root handles are to be displayed
 * @see #showsRootHandles
 */
public void setShowsRootHandles(boolean newValue)
{
	if (newValue == showsRootHandles) {
		return;
	}
	boolean oldValue = showsRootHandles;
	showsRootHandles = newValue;
	visualComponent.firePropertyChange(JTree.SHOWS_ROOT_HANDLES_PROPERTY,
									   oldValue, showsRootHandles);
	visualComponent.invalidate();
}

/**
 * Makes sure all the path components in path are expanded (except
 * for the last path component) and scrolls so that the
 * node identified by the path is displayed. Only works when this
 * JCTreeTable is contained in a JScrollPane.
 *
 * @param path  the TreePath identifying the node to bring into view
 */
public void scrollPathToVisible(TreePath path) {}

/**
 * Icon to use if plaf fails to provide one
 */
protected class DefaultIcon implements Icon
{
	protected final int size = 9;
	protected final int halfsize = size / 2;
	protected boolean collapsed;

	/**
	 * Constructor.  Sets whether icon will show collapsed or not intially
	 * @param collapsed  The intial value of collapsed.
	 */
	public DefaultIcon(boolean collapsed)
	{
		this.collapsed = collapsed;
	}

	/**
	 * Get the icon height
	 * @return The icon height
	 */
	public int getIconHeight()
	{
		return(size);
	}

	/**
	 * Get the icon width
	 * @return The icon width
	 */
	public int getIconWidth()
	{
		return(size);
	}

	/**
	 * Draws the folder state control (expand/collapse) icon centered on
	 * the specified point.
	 * @param c The component from which information can be gather for
	 *    rendering.  Not used.
	 * @param gc The graphics context to do the drawing
	 * @param x The x position to draw
	 * @param y The y position to draw
	 */
	public void paintIcon(Component c, Graphics gc, int x, int y)
	{
		// Draw a box width a plus or minus in it
		gc.clearRect(x - halfsize, y, size - 1, size - 1);
		gc.drawRect(x - halfsize, y, size - 1, size - 1);
		gc.drawLine(x - halfsize + 2, y + halfsize,
					x + halfsize - 2, y + halfsize);
		if (collapsed) {
			gc.drawLine(x, y + 2, x, y + size - 3);
		}
	}
}

}
