/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: TreeWithSortableChildren.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:48 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.treetable;

import com.klg.jclass.util.JCIntComparator;
import com.klg.jclass.util.JCListenerList;
import com.klg.jclass.util.JCMappingSort;
import com.klg.jclass.util.swing.DefaultRowComparator;
import com.klg.jclass.util.swing.JCComparableRow;
import com.klg.jclass.util.swing.JCRowComparator;
import com.klg.jclass.util.swing.JCRowSortModel;

import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.tree.TreePath;
import java.lang.ref.WeakReference;
import java.util.Enumeration;
import java.util.HashMap;

/**
 * Provides a sorted mapping for children of a given node.
 */
public class TreeWithSortableChildren implements JCTreeTableModel,
												 JCRowSortModel,
												 java.io.Serializable {

protected JCTreeTableModel  treetable = null;
protected int               sortColumn = -1;
protected int               sortOrder = JCRowSortModel.FORWARD;
protected JCRowComparator   rowComparator =
                                new DefaultRowComparator(JCRowSortModel.FORWARD);
protected int[]             keyColumnsArray[];
protected RowIntComparator  rowIntComparator = new RowIntComparator();
protected HashMap           nodeChildIndexMaps = null;
protected boolean           autoSort = true;
protected boolean           needsResorting = true;
protected boolean           needsReordering = false;
protected JCListenerList    listeners = null;
protected Listener          l = new Listener();

/**
 * Constructs a <code>JCTreeTableModel</code> that wraps a given instance of a
 * <code>JCTreeTableModel</code> and provides a sorted mapping of the children
 * for any given leaf node. The sort order is defined by the configurable
 * <code>Comparator</code> property.
 */
public TreeWithSortableChildren(JCTreeTableModel treetable,
								JCRowComparator comparator)
{
	this(treetable);
	this.rowComparator = comparator;
}

/**
 * Constructs a <code>JCTreeTableModel</code> that that wraps a given instance
 * of a <code>JCTreeTableModel</code> and provides a sorted mapping of the
 * children for any given leaf node. The sort order is defined by the
 * configurable <code>Comparator</code> property.
 */
public TreeWithSortableChildren(JCTreeTableModel treetable)
{
	setModel(treetable);
}

/**
 * Returns the <code>JCTreeTable</code> model this model wraps
 */
public JCTreeTableModel getModel()
{
	return(treetable);
}

/**
 * Sets the tree table model wrapped by this model.
 */
public void setModel(JCTreeTableModel newModel)
{
	if (newModel == treetable) {
		return;
	}
    if (treetable != null) {
        treetable.removeTreeModelListener(l);
    }
	treetable = newModel;
	treetable.addTreeModelListener(l);

    // Rest sorting information
    needsResorting = true;
    needsReordering = false;
    nodeChildIndexMaps = autoSort ? new HashMap() : null;
    sortColumn = -1;

    // Make sure key column array size matches the new model's column count
    // while trying to keep as much of the previous keyColumn information
    // as possible
    if (treetable == null) {
        keyColumnsArray = null;
    }
    else {
        int nCols    = treetable.getColumnCount();
        int nKeyCols = (keyColumnsArray == null) ? 0 : keyColumnsArray.length;
        if (nKeyCols != nCols) {
            int[][] newArray = new int[nCols][];
            for (int i = 0; i < Math.min(nCols, nKeyCols); i++) {
                newArray[i] = keyColumnsArray[i];
            }
            keyColumnsArray = newArray;
        }
    }
}

//
// Row Sortable Implementation
//

/**
 * Returns whether the data is automatically sorted when it changes
 * according the current comparator.
 */
public boolean getAutoSort()
{
	return(autoSort);
}

/**
 * Sets whether the data should be automatically sorted when it changes.
 */
public void setAutoSort(boolean autoSort)
{
	if (this.autoSort == autoSort) {
		return;
	}
	this.autoSort = autoSort;
	if (autoSort == true) {
		setDataSorted(true);
	}
}

/**
 * Sets the key columns used to sort the table model when clicking
 * on a specified column.
 * @param column index of column in TableModel
 * @param keyColumns columns to use as keys to sort when sorting
 *     on the specified column.
 */
public void setKeyColumns(int column, int keyColumns[])
{
    if (keyColumnsArray == null) {
		keyColumnsArray = new int[treetable.getColumnCount()][];
    }
	keyColumnsArray[column] = keyColumns;
	needsResorting = true;
}

/**
 * Returns the key columns used to sort the table model when clicking
 * on a specified column.  If the user has not set the key columns,
 * this returns an array of length 1 with the specified column
 * as the only member.
 */
public int[] getKeyColumns(int column)
{
	return(keyColumnsArray[column]);
}

/**
 * Returns whether or not the data set is currently sorted.
 */
public boolean isDataSorted()
{
	return(nodeChildIndexMaps != null);
}

/**
 * Sorts data by current column if set to true and the data needs sorting;
 * or, if set to false, it returns the data in its native unsorted state.
 */
public void setDataSorted(boolean sorted)
{
	if (sorted && sortColumn == -1) {
		// cannot sort
		return;
	}

    if (!sorted && isDataSorted()) {
        needsReordering = true;
        sortColumn = -1;
    }

	if (needsResorting || needsReordering) {
		if (needsResorting) {
			// Wipe out existing sort map
			nodeChildIndexMaps = new HashMap();
			needsReordering = false;
		}

		// Post a change event here
		postTreeStructureChanged(null, false);
		needsResorting = false;
	}
}

/**
 * Gets the sorting column. This should return -1 if the data is unsorted.
 */
public int getSortColumn()
{
    return(sortColumn);
}

/**
 * Sets the column to use the next time a sort is performed.
 * @param column only valid column indicies are allowed
 */
public void setSortColumn(int column)
{
	if (sortColumn == column) {
		return;
	}

	if (treetable != null && treetable.getColumnCount() <= column) {
		// We allow an arbitrary column to be set but only if the
		// treetable is not set.  If it is set we enforce it's bounds
		throw new IllegalArgumentException("index value greater than " +
                                           "number of columns");
	}
	sortColumn = column;
	needsResorting = true;
	if (autoSort) {
		setDataSorted(true);
	}
}

/**
 * Gets the sorting order to use the next time a sort is performed.
 */
public int getSortOrder()
{
	return(sortOrder);
}

/**
 * Sets the sorting order.
 */
public void setSortOrder(int order)
{
    this.sortOrder = order;
	needsReordering = true;
	if (autoSort) {
		setDataSorted(true);
	}
}

/**
 * Toggles the sorting order.
 */
public void toggleSortOrder()
{
    sortOrder = sortOrder == FORWARD ? REVERSE : FORWARD;
	needsReordering = true;
	if (autoSort) {
		setDataSorted(true);
	}
}


/**
 * Gets the mapped indices; returns null if unsorted.
 * Sorts children "on the fly" as they are requested
 * and stores the sorted results in a hash table
 * that is thrown away each time a new sort is invoked.
 */
protected int[] getIndices(Object parent)
{
    int num_kids = treetable.getChildCount(parent);
	if (sortColumn == -1 || num_kids == 0) {
		return(null);
	}

    // We are unsorted, so nothing to calculate
    if (nodeChildIndexMaps == null) {
        return(null);
    }
	HashObject ho = new HashObject(parent);
	int indices[] = (int[]) nodeChildIndexMaps.get(ho);

	if (indices == null) {
		// Create the index mapping
		indices = new int[num_kids];
        for (int i = 0; i < indices.length; i++) {
            indices[i] = i;
        }
		// Do sort here
		rowIntComparator.parent = parent;
		if (keyColumnsArray == null) {
			keyColumnsArray = new int[treetable.getColumnCount()][];
		}
		int keyColumns[] = { sortColumn };
		rowComparator.setKeyColumns(keyColumnsArray[sortColumn] != null ?
									keyColumnsArray[sortColumn] : keyColumns);
		JCMappingSort.sort(rowIntComparator, indices);
		// Store in hash table
		nodeChildIndexMaps.put(ho, indices);
	}

	return(indices);
}

//
// JCTreeTableModel Implementation
//

/**
 * Removes the specified treemodel listener.
 */
public void addTreeModelListener(TreeModelListener l)
{
	listeners = JCListenerList.add(listeners, l);
}

/**
 * Removes the specified treemodel listener.
 */
public void removeTreeModelListener(TreeModelListener l)
{
	listeners = JCListenerList.remove(listeners, l);
}

/**
 * Returns the "index"th non-leaf child of this node.
 */
public Object getChild(Object parent, int index)
{
	int indices[] = getIndices(parent);
	if (indices != null) {
		// Map sorted index to unsorted lookup
		if (sortOrder == FORWARD) {
			index = indices[index];
		}
		else {
			// Index from then end since we are doing reverse order
			index = indices[indices.length - index - 1];
		}
	}
	return(treetable.getChild(parent, index));
}

/**
 * Return the number of children.
 */
public int getChildCount(Object parent)
{
	return(treetable.getChildCount(parent));
}

/**
 * Returns the non-leaf index of this child.
 */
public int getIndexOfChild(Object parent, Object child)
{
    int unsortedIndex = treetable.getIndexOfChild(parent, child);
    if (unsortedIndex < 0) {
        return(unsortedIndex);
    }

	int indices[] = getIndices(parent);
	if (indices == null) {
		return(unsortedIndex);
	}

	// Loop for the unsorted index in the indices array.  Its current position
    // in the array is what we want.
	for (int i = 0; i < indices.length; i++) {
		if (indices[i] == unsortedIndex) {
            if (sortOrder == FORWARD) {
                return(i);
            }
            else {
                // Index from then end since we are doing reverse order
                return(indices.length - i - 1);
            }
		}
	}

	throw new IllegalArgumentException("Node not child of specified parent");
}

/**
 * Returns the root node of the tree.
 */
public Object getRoot()
{
	// Pass this call straight through to the underlying tree
	return(treetable.getRoot());
}

/**
 * Returns whether the child is a leaf.
 */
public boolean isLeaf(Object node)
{
	return(treetable.isLeaf(node));
}

/**
 * Calls the underling trees valueForPathChanged method.
 */
public void valueForPathChanged(TreePath path, Object newValue)
{
	// Pass this call straight through to the underlying tree
	treetable.valueForPathChanged(path, newValue);
}

/**
 * Returns the value of the specific node and column.
 */
public Object getValueAt(Object node, int column)
{
	return(treetable.getValueAt(node, column));
}

/**
 * Returns whether a particular cell is editable given the node and column.
 */
public boolean isCellEditable(Object node, int column)
{
	return(treetable.isCellEditable(node, column));
}

/**
 * Sets the value at a particular node and column.
 */
public void setValueAt(Object value, Object node, int column)
{
	treetable.setValueAt(value, node, column);
}

//
// The following methods as a subset of javax.swing.table.TableModel
//

/**
 * Returns the class of the given column index.
 */
public Class getColumnClass(int column)
{
	return(treetable.getColumnClass(column));
}

/**
 * Returns the number of columns.
 */
public int getColumnCount()
{
	return(treetable.getColumnCount());
}

/**
 * Returns the column name for the given column index.
 */
public String getColumnName(int column)
{
	return(treetable.getColumnName(column));
}


/**
 * Given a <code>TreePath</code>, this method returns the last component or
 * "node" of the <code>TreePath</code>.
 */
static protected Object getNode(TreePath path)
{
	return(path.getLastPathComponent());
}

/**
 * <p>Invoked after the tree has drastically changed structure from a
 * given node down.  If the path returned by e.getPath() is of length
 * one and the first element does not identify the current root node
 * the first element should become the new root of the tree.<p>
 * @param event The event generated by the change.  If a null event is passed,
 * an "all changed" event is generated.
 */
protected void postTreeStructureChanged(TreeModelEvent event)
{
    postTreeStructureChanged(event, true);
}

/**
 * <p>Invoked after the tree has drastically changed structure from a
 * given node down.  If the path returned by e.getPath() is of length
 * one and the first element does not identify the current root node
 * the first element should become the new root of the tree.<p>
 * @param event The event generated by the change.  If a null event is passed,
 * an "all changed" event is generated.
 * @param external If true, this method is called externally thru the Listener.
 * In this case we need to reset the internal index maps.  If false, this
 * method is called internally and is meant to inform listeners of an internal
 * change.  In this case the index maps are already updated and should not be
 * reset.
 */
protected void postTreeStructureChanged(TreeModelEvent event, boolean external)
{
	if (event == null) {
		// Post an everything has changed event
		event = new TreeModelEvent(this,
								   new TreePath(treetable.getRoot()),
								   null, null);
	}

    if (external) {
        nodeChildIndexMaps = autoSort ? new HashMap() : null;
        needsResorting = true;
    }

	// Repost treeStructureChanged
	Enumeration e = JCListenerList.elements(listeners);
	while (e.hasMoreElements()) {
		TreeModelListener listener = (TreeModelListener)e.nextElement();
		listener.treeStructureChanged(event);
	}
}

/**
 * Inner class to handle events from the source tree.
 */
class Listener implements TreeModelListener {

	protected TreeModelEvent convertTreeEvent(TreeModelEvent e)
	{
        if (e == null) {
            return(null);
        }

		// Build the new tree event
		return(new TreeModelEvent(TreeWithSortableChildren.this,
								  e.getTreePath(), e.getChildIndices(),
								  e.getChildren()));
	}

	public void treeNodesChanged(TreeModelEvent tme)
	{
		TreeModelEvent ntme = null;
		if (tme == null) {
			// Post an everything has changed event
			ntme = new TreeModelEvent(TreeWithSortableChildren.this,
									  new TreePath(treetable.getRoot()),
									  null, null);
		}

		if (nodeChildIndexMaps == null) {
			// We are unsorted so we can do an efficient repost treeNodesChanged
			ntme = convertTreeEvent(tme);
			if (ntme == null) {
				return;
			}
		}

		if (ntme != null) {
			// We have an event so we post it
			Enumeration e = JCListenerList.elements(listeners);
			while (e.hasMoreElements()) {
				TreeModelListener listener = (TreeModelListener)e.nextElement();
				listener.treeNodesChanged(ntme);
			}
		}
		else {
            nodeChildIndexMaps = autoSort ? new HashMap() : null;
            needsResorting = true;

			// Recursivly call outselves with a null to make the tree do
			// a complete recalc
			treeNodesChanged(null);
		}
	}

	public void treeNodesInserted(TreeModelEvent tme)
	{
		TreeModelEvent ntme = convertTreeEvent(tme);
		if (ntme == null) {
			return;
		}
        nodeChildIndexMaps = autoSort ? new HashMap() : null;
        needsResorting = true;

		// Repost treeNodesInserted
		Enumeration e = JCListenerList.elements(listeners);
		while (e.hasMoreElements()) {
			TreeModelListener listener = (TreeModelListener)e.nextElement();
			listener.treeNodesInserted(ntme);
		}
	}

	public void treeNodesRemoved(TreeModelEvent tme)
	{
		TreeModelEvent ntme = convertTreeEvent(tme);
		if (ntme == null) {
			return;
		}
        nodeChildIndexMaps = autoSort ? new HashMap() : null;
        needsResorting = true;

		// ??? may be tricky
		// repost treeNodesRemoved
		Enumeration e = JCListenerList.elements(listeners);
		while (e.hasMoreElements()) {
			TreeModelListener listener = (TreeModelListener)e.nextElement();
			listener.treeNodesRemoved(ntme);
		}
	}

	public void treeStructureChanged(TreeModelEvent tme)
	{
		postTreeStructureChanged(tme);
	}
} // end of Listener


/**
 * Guarantees that the object that you are hashing will compare uniquely
 * only to the object being stored therein. We use a weak reference to
 * make sure that we are never holding on to the object.
 */
static public class HashObject {
	WeakReference weak;

	public HashObject(Object obj)
	{
		weak = new WeakReference(obj);
	}

	public boolean equals(Object ho)
	{
		Object obj1 = weak.get();
		Object obj2 = null;
		if (ho != null) {
			obj2 = ((HashObject) ho).weak.get();
		}
		if (obj1 == null) {
			return(false);
		}
		if (obj1 == obj2) {
			return(true);
		}
		return(false);
	}

	public int hashCode()
	{
		Object obj = weak.get();
		if (obj == null) {
			return(0);
		}
		return(obj.hashCode());
	}
}

/**
 * Internal <code>ComparableRow</code> class.
 */
class ComparableRow implements JCComparableRow  {
	public Object node = null;
	public int row;

	//
	// JCComparableRow implementation
	//

    /**
     * Returns the value of the cell specified by the column and the current
	 * row.
     */
    public Object getValueAt(int column)
	{
	    return(treetable.getValueAt(node, column));
    }

    /**
     * Returns the natural (unsorted) index of the current row
     */
    public int getRowIndex()
	{
    	return(row);
    }

} // ComparableRow


/**
 * Internal <code>RowIntComparator</code> class.
 */
class RowIntComparator implements JCIntComparator {
	protected ComparableRow r1 = new ComparableRow();
    protected ComparableRow r2 = new ComparableRow();
	public Object parent = null;

	//
	// JCIntComparator interface method
	//

	public int compare(int index1, int index2)
	{
        r1.row = index1;
		r1.node = treetable.getChild(parent, index1);
        r2.row = index2;
		r2.node = treetable.getChild(parent, index2);
    	return(rowComparator.compare(r1, r2));
    }

} // RowIntComparator

}
