/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

//   RCSID -- $RCSfile: JCTableXMLParser.java,v $ $Revision: 1.2 $
//            $Date: 2006-03-17 21:13:49 $  $Locker:  $  Quest Software Inc.

package com.klg.jclass.util.xml;

import org.w3c.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.IOException;
import java.io.Reader;
import java.util.Vector;

import com.klg.jclass.util.io.JCParseException;

/**
 * This class takes in a stream of data and parses
 * it under the assumption that it is in the defined XML
 * format that LiveTable uses. It then passes the data to
 * the given <code>JCTablePopulator</code> instance.
 *
 * <P>The XML LiveTable format looks like this:
 * <PRE>
 * &lt;?xml version="1.0"?&gt;
 * &lt;!DOCTYPE JCTableData SYSTEM "JCTableData.dtd"&gt;
 * &lt;JCTableData&gt;
 *   &lt;Row&gt;
 *     &lt;Cell&gt;1&lt;/Cell&gt; &lt;Cell&gt;2&lt;/Cell&gt; &lt;Cell&gt;3&lt;/Cell&gt; &lt;Cell&gt;4&lt;/Cell&gt;
 *   &lt;/Row&gt;
 *   &lt;Row&gt;
 *     &lt;Cell&gt;1&lt;/Cell&gt; &lt;Cell&gt;2&lt;/Cell&gt; &lt;Cell&gt;3&lt;/Cell&gt; &lt;Cell&gt;4&lt;/Cell&gt;
 *   &lt;/Row&gt;
 * &lt;/JCTableData&gt;
 * </PRE>
 * Or, with row and column labels:
 * <pre>
 * &lt;?xml version="1.0"?&gt;
 * &lt;!DOCTYPE JCTableData SYSTEM "JCTableData.dtd"&gt;
 * &lt;JCTableData&gt;
 *   &lt;ColumnLabel&gt;Column 1&lt;/ColumnLabel&gt;
 *   &lt;ColumnLabel&gt;Column 2&lt;/ColumnLabel&gt;
 *   &lt;ColumnLabel&gt;Column 3&lt;/ColumnLabel&gt;
 *   &lt;ColumnLabel&gt;Column 4&lt;/ColumnLabel&gt;
 *   &lt;Row&gt;
 *     &lt;RowLabel&gt;Row 1&lt;/RowLabel&gt;
 *     &lt;Cell&gt;1&lt;/Cell&gt; &lt;Cell&gt;2&lt;/Cell&gt; &lt;Cell&gt;3&lt;/Cell&gt; &lt;Cell&gt;4&lt;/Cell&gt;
 *   &lt;/Row&gt;
 *   &lt;Row&gt;
 *     &lt;RowLabel&gt;Row 2&lt;/RowLabel&gt;
 *     &lt;Cell&gt;1&lt;/Cell&gt; &lt;Cell&gt;4&lt;/Cell&gt; &lt;Cell&gt;9&lt;/Cell&gt; &lt;Cell&gt;16&lt;/Cell&gt;
 *   &lt;/Row&gt;
 *   &lt;Row&gt;
 *     &lt;RowLabel&gt;Row 3&lt;/RowLabel&gt;
 *     &lt;Cell&gt;1&lt;/Cell&gt; &lt;Cell&gt;16&lt;/Cell&gt; &lt;Cell&gt;81&lt;/Cell&gt; &lt;Cell&gt;256&lt;/Cell&gt;
 *   &lt;/Row&gt;
 * &lt;/JCTableData&gt;
 * </pre>
 */
public class JCTableXMLParser implements java.io.Serializable {

public final static String XML_DS             = "JCTableData";
public final static String XML_DS_COLUMNLABEL = "ColumnLabel";
public final static String XML_DS_ROWLABEL    = "RowLabel";
public final static String XML_DS_ROW         = "Row";
public final static String XML_DS_ROW_CELL    = "Cell";

/** Flag indicating whether parsing is in debug mode, which turns
 * on error messages on error out. */
protected boolean isDebug = false;

/**
 * Reads and parses table data in the defined XML format.
 *
 * @param r reader open on an active stream
 * @param pop <code>JCTablePopulator</code> instance with which to set table
 *            data
 * @throws IOException
 * @throws JCParseException
 */
public void readAndParse(Reader r, JCTablePopulator pop)
    throws IOException, JCParseException
{
    InputSource input = new InputSource(r);

    Document xml = null;
    try {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(true);

        DocumentBuilder builder = factory.newDocumentBuilder();

        input.setSystemId("file:/template.xml");
        //
        // Configure the parser
        //
        builder.setErrorHandler(new JCXMLErrorPrinter());
        builder.setEntityResolver(new JCXMLDTDResolver("JCTableData.dtd",
                                                       pop.getClass()));
        xml = builder.parse(input);
    }
    catch (ParserConfigurationException e) {
        e.printStackTrace(System.err);
        throw new JCParseException("XML Parser Configuration Error", e);
    }
    catch (SAXParseException e) {
        // Already reported by error printer
        e.printStackTrace(System.err);
        throw new JCParseException("XML Parsing Error", e);
    }
    catch (SAXException e) {
        e.printStackTrace(System.err);
        throw new JCParseException("XML Error", e);
    }
    catch (IOException e) {
        e.printStackTrace(System.err);
        throw e;
    }

    if (xml == null) {
        //XML Document doesn't exist
        return;
    }

    readTableData(xml, pop);
}

/**
 * Reads the table data from the given xml document.
 * @param xml the xml document to parse
 * @param pop the <code>JCTablePopulator</code> instance with which to set
 *            the newly parsed data
 * @throws IOException
 */
protected void readTableData(Document xml, JCTablePopulator pop)
    throws IOException
{

    NodeList dataNodeList = xml.getElementsByTagName(XML_DS);
    Node tableDataNode = dataNodeList.item(0);
    tableDataNode.normalize();

    //First Pass -- Count number of rows and columns
    int nrows = 0;
    int ncolumns = 0;
    boolean hasRowLabels = false;
    boolean hasColumnLabels = false;

    // count number of column labels -- use to help determine
    // maximum number of columns
    NodeList colLabelNodeList =
        ((Element)tableDataNode).getElementsByTagName(XML_DS_COLUMNLABEL);
    int ncollabels = colLabelNodeList.getLength();
    if (ncollabels > 0) {
        hasColumnLabels = true;
    }
    ncolumns = Math.max(ncolumns, ncollabels);

    NodeList rowNodeList =
        ((Element)tableDataNode).getElementsByTagName(XML_DS_ROW);
    for (int i = 0; i < rowNodeList.getLength(); i++) {
        Node rowNode = rowNodeList.item(i);

        // count number of rows
        nrows++;

        // Use number of value tags in this row to help determine
        // maximum number of columsn
        NodeList rowCellNodeList =
            ((Element)rowNode).getElementsByTagName(XML_DS_ROW_CELL);
        int nvalues = rowCellNodeList.getLength();
        ncolumns = Math.max(ncolumns, nvalues);

        // Does this document have row labels?
        NodeList rowLabelNodeList =
            ((Element)rowNode).getElementsByTagName(XML_DS_ROWLABEL);
        if (!hasRowLabels && rowLabelNodeList.getLength() > 0) {
            hasRowLabels = true;
        }

    }

    // Second Pass -- Fill in the data!
    Object value = null;
    Vector rowV = new Vector();
    Vector rowLabelV = new Vector();
    Vector colLabelV = new Vector();
    Vector v = null;

    // Read every column label tag in this document
    if (hasColumnLabels) {
        int col = 0;
        for (int i = 0; i < colLabelNodeList.getLength(); i++) {
            Node colLabelNode = colLabelNodeList.item(i);
            value = null;
            if (colLabelNode != null) {
                colLabelNode = colLabelNode.getFirstChild();
                if (colLabelNode != null && colLabelNode instanceof Text) {
                    value = cvtToTableValue(colLabelNode.getNodeValue());
                }
                else {
                    value = "";
                }
            }
            colLabelV.addElement(value);
            if (isDebug) {
                System.out.println("Column Label " + col + " is " + value);
            }
            col++;
        }
    }

    // Parse each row
    for (int row = 0; row < rowNodeList.getLength(); row++) {
        Node rowNode = rowNodeList.item(row);
        if (rowNode == null) {
            break;
        }

        // if this datasource has row labels, read the row label
        // if there is no row label for this row, use null
        if (hasRowLabels) {
            value = null;
            NodeList rowLabelNodeList =
                ((Element)rowNode).getElementsByTagName(XML_DS_ROWLABEL);
            Node rowLabelNode = rowLabelNodeList.item(0);
            if (rowLabelNode != null) {
                rowLabelNode = rowLabelNode.getFirstChild();
                if (rowLabelNode != null && rowLabelNode instanceof Text) {
                    value = cvtToTableValue(rowLabelNode.getNodeValue());
                }
            }
            else {
                // has no tag -- default
                value = null;
            }
            rowLabelV.addElement(value);
            if (isDebug) {
                System.out.println("Row Label " + row + " is " + value);
            }
        }

        //Pull each value node out of this row that exists and place
        //the resulting text in the datasource row

        v = new Vector();
        Node valueNode = null;
        int col = 0;
        NodeList valueNodeList =
            ((Element)rowNode).getElementsByTagName(XML_DS_ROW_CELL);
        for (int j = 0; j < valueNodeList.getLength(); j++) {
            valueNode = valueNodeList.item(j);

            value = null;
            if (valueNode != null) {
                valueNode = valueNode.getFirstChild();
                if (valueNode != null && valueNode instanceof Text) {
                    value = cvtToTableValue(valueNode.getNodeValue());
                }
            }
            v.addElement(value);
            if (isDebug) {
                System.out.println("Value at row " + row + " col " + col +
                                   " is " + value);
            }
            col++;
        }
        rowV.addElement(v);
    }

    // set the table data with the JCTablePopulator class
    pop.setTableData(rowV, rowLabelV, colLabelV, nrows, ncolumns);
}

/**
 * Converts a String from the xml data into a table value.  If it will
 * convert to a double, do that.  Otherwise, keep it as a String.
 * @return
 * @param value
 */
protected Object cvtToTableValue(String value)
{
    try {
        Double d = new Double(value);
        return(d);
    }
    catch (NumberFormatException e) {
        return(value);
    }
}

}
