/* 
______________________________________________________________________
*
* Copyright (c) 1996-2004 QUEST SOFTWARE INC.  All Rights Reserved.
* http://java.quest.com
*
* This software is the confidential and proprietary information of
* Quest Software Inc. ("Confidential Information").  You shall not disclose
* such Confidential Information and shall use it only in accordance with the
* terms of the license agreement you entered into with Quest Software.
*
* QUEST SOFTWARE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
* OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
* TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE, OR NON-INFRINGEMENT. QUEST SOFTWARE SHALL NOT BE LIABLE FOR ANY
* DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
* DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
* ______________________________________________________________________
*/

// RCSID -- $RCSfile: JCPageTable.java,v $ $Revision: 1.2 $ $Date: 2006-03-17 21:13:22 $ $Locker:  $  Quest Software Inc.

package com.klg.jclass.page; 

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Graphics2D;
import java.awt.FontMetrics;




import java.util.NoSuchElementException;



import com.klg.jclass.cell.JCCellInfo;
import com.klg.jclass.cell.JCCellRenderer;
import com.klg.jclass.cell.JCComponentCellRenderer;
import com.klg.jclass.cell.JCLightCellRenderer;
import com.klg.jclass.cell.EditorRendererRegistry;

import com.klg.jclass.page.render.Render;
import com.klg.jclass.page.render.BoxRender;
import com.klg.jclass.page.render.LineRender;
import com.klg.jclass.page.render.TableCellRender;

import com.klg.jclass.table.JCTableEnum;
import com.klg.jclass.table.TableDataModel;
import com.klg.jclass.table.EditableTableDataModel;

import com.klg.jclass.util.formulae.Expression;
import com.klg.jclass.util.formulae.ExpressionVariable;
import com.klg.jclass.util.formulae.MathScalar;




import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;


/**
 * <code>JCPageTable</code> provides methods for creating and customizing tables
 * into which data is flowed.  Content may be added in the form of formatted,
 * flowed text, or it can be added as content Objects to the underlying
 * <code>TableDataModel</code>, in which case the appropriate <code>CellRenderers</code> are used
 * to render/layout the content.
 * <p>
 * A <code>JCPageTable</code> consists of an ordered set of <code>JCPageTable.Columns</code> which
 * define the columns of the table and their attributes and an ordered
 * set of <code>JCPageTable.Rows</code>.  A Row contains both the row attributes and
 * the list of <code>JCPageTable.Cell</code> objects defined in that row.  Headers are
 * implemented using a nested <code>JCPageTable</code> object.
 * <p>
 * Attributes such as background colour, cell margins, and border styles
 * can be set on any object in the table heirarchy (<code>JCPageTable</code>, Row, Column
 * or Cell).  The most specific setting available in a particular cell
 * is used when the table is drawn, according to the inheritance order.
 * The standard accessor methods (for example, <code>Cell.getBackgroundColor()</code>)
 * will return the setting that will be applied at that object, even if the
 * value is inherited from a higher-level object.  All <code>get()</code> methods for
 * inheritable attributes are shadowed by a method which determines if
 * the value returned by the get method is inherited (versus being
 * locally specified).  (for example, <code>Cell.isBackgroundColorInherited()</code>)
 * <p>
 * Because of the ambiguity created by inheriting Cell attributes from
 * both the Row and Column, the method <code>setRowColumnDominance(int)</code>
 * takes the constants <code>ROW_DOMINANCE</code> and
 * <code>COLUMN_DOMINANCE</code> to define whether or values are inherited
 * first from the Row or the Column, respectively.
 */
public class JCPageTable {

/** The document to which this table belongs. */
protected JCDocument        document;
/** The header table describes the table row displayed as a header. */
protected JCPageTable       header;
/** The main table is the body of the table. It is associated with a header table row.*/
protected JCPageTable       mainTable;

/** The row list is an ordered list of rows in the table. */
protected List              rowList;
/** The column list is an ordered list of columns in the table. */
protected List              columnList;

/** The predefined table style for this table. */
protected JCTableStyle        tableStyle;

/** A template cell to use as a source of default cell attributes. */
protected Cell              defaultCell;

/** Tables use the paragraph alignment options for horizontal alignment. */
protected int               alignment;
/** Controls the borders drawn on the perimeter of a table range. */
protected int               borderMode;
/** Specifies which of a conflicting row or column attribute has dominance. */
protected int               dominance;
/** Controls where extra columns in a too-wide table appear in the flow. */
protected int               overflowMode;

/** Number of columns in the table. */
protected int               columnCount;
/** Number of rows in the table. */
protected int               rowCount;
/** <code>True</code> if this table is created as the header of another table. */
protected boolean           isHeader;

/** <code>True</code> if this table has had all frames completed and heights calculated. */
protected boolean           isComplete;

/** This stores data associated with cells on the table. */
protected EditableTableDataModel        tableData;




/** Clips table columns to page. */
public static final int OVERFLOW_CLIP_COLUMNS = 1;
/** Wraps table columns to same page. */
public static final int OVERFLOW_WRAP_COLUMNS = 2;
/** Internal use only. */
public static final int OVERFLOW_TEAROFF_COLUMNS = 3;
/** Wraps table columns to next frame column or flow frame. */
public static final int OVERFLOW_WRAP_COLUMNS_NEXT = 4;

public static final int BORDER_USE_INTERNAL = 1;
public static final int BORDER_USE_EXTERNAL = 2;

public static final int ROW_DOMINANCE = 1;
public static final int COLUMN_DOMINANCE = 2;

public static final int CELL_ALIGNMENT_NONE = 1;
public static final int CELL_ALIGNMENT_TOP = 2;
public static final int CELL_ALIGNMENT_CENTER = 3;
public static final int CELL_ALIGNMENT_BOTTOM = 4;

// These variables represent uninitialised components
public static final Row     UNINITIALISED_ROW = null;
public static final Cell    UNINITIALISED_CELL = null;

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths a List of column widths of which the table is composed
 */
public JCPageTable(JCDocument doc, List columnWidths) {
    this(doc, columnWidths.toArray(), null, (JCTableStyle) JCTableStyle.STYLE_DEFAULT.clone());
}

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths a List of column widths of which the table is composed
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
public JCPageTable(JCDocument doc, List columnWidths, JCTableStyle tableStyle) {
    this(doc, columnWidths.toArray(), null, tableStyle);
}

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths an array of column widths of which the table is composed
 */
public JCPageTable(JCDocument doc, JCUnit.Measure columnWidths[]) {
    this(doc, columnWidths, null, (JCTableStyle) JCTableStyle.STYLE_DEFAULT.clone());
}

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths an array of column widths of which the table is composed
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
public JCPageTable(JCDocument doc, JCUnit.Measure columnWidths[],
        JCTableStyle tableStyle) {
    this(doc, columnWidths, null, tableStyle);
}

/**
 * Creates a new table with the number of columns specified, where each column
 * is a specified width.
 * @param doc the document to which this table belongs
 * @param columnCount the number of columns in this table
 * @param columnWidth a <code>JCUnit.Measure</code> for the width of the columns
 */
public JCPageTable(JCDocument doc, int columnCount,
                   JCUnit.Measure columnWidth) {
    this(doc, (JCPageTable) null, (JCTableStyle) JCTableStyle.STYLE_DEFAULT.clone());
    for (int i = 0; i < columnCount; i++) {
        createColumn(columnWidth, i);
    }
    this.columnCount = columnCount;
}

/**
 * Creates a new table with the number of columns specified where each column
 * is a specified width.
 * @param doc the document to which this table belongs
 * @param columnCount the number of columns in this table
 * @param columnWidth a <code>JCUnit.Measure</code> for the width of the columns
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
public JCPageTable(JCDocument doc, int columnCount,
                   JCUnit.Measure columnWidth, JCTableStyle tableStyle) {
    this(doc, (JCPageTable) null, tableStyle);
    for (int i = 0; i < columnCount; i++) {
        createColumn(columnWidth, i);
    }
    this.columnCount = columnCount;
}

/**
 * Creates a new table with the number of columns specified where each column
 * is the default width.
 * @param doc the document to which this table belongs
 * @param columnCount the number of columns in this table
 */
public JCPageTable(JCDocument doc, int columnCount) {
    this(doc, columnCount, new JCUnit.Measure(JCUnit.INCHES, 1.0));
}

/**
 * Creates a new table with the number of columns specified where each column
 * is the default width.
 * @param doc the document to which this table belongs
 * @param columnCount the number of columns in this table
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
public JCPageTable(JCDocument doc, int columnCount, JCTableStyle tableStyle) {
    this(doc, columnCount, new JCUnit.Measure(JCUnit.INCHES, 1.0),
        tableStyle);
}

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths a List of column widths of which the table is composed
 * @param mainTable null if the table being created is a full (main) table
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
protected JCPageTable(JCDocument doc, List columnWidths, JCPageTable mainTable,
        JCTableStyle tableStyle) {
    this(doc, columnWidths.toArray(), mainTable, tableStyle);
}

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths a List of column widths of which the table is composed
 * @param mainTable <code>true</code> if the table being created is a full (main) table
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
protected JCPageTable(JCDocument doc, JCUnit.Measure columnWidths[],
                      JCPageTable mainTable, JCTableStyle tableStyle) {
    this(doc, (Object []) columnWidths, mainTable, tableStyle);
}

/**
 * Creates a new table with columns as specified in the list of widths.
 * @param doc the document to which this table belongs
 * @param columnWidths a list of column widths of which the table is composed
 * @param mainTable <code>true</code> if the table being created is a full (main) table
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
protected JCPageTable(JCDocument doc, Object columnWidths[],
                      JCPageTable mainTable, JCTableStyle tableStyle) {

    this(doc, mainTable, tableStyle);

    

    // Create the columns of the table
    for (int column = 0; column < columnWidths.length; column++) {
        JCUnit.Measure width = (JCUnit.Measure) columnWidths[column];
        createColumn(width, column);
    }
    columnCount = columnWidths.length;

    
}



/**
 * Creates a new table with unset columns.
 * @param doc the document to which this table belongs
 * @param mainTable <code>true</code> if the table being created is a full (main) table
 * @param tableStyle a <code>JCTableStyle</code> which predefines the table style
 */
protected JCPageTable(JCDocument doc, JCPageTable mainTable,
                      JCTableStyle tableStyle) {

    if (mainTable != null) {
        isHeader = true;
        this.mainTable = mainTable;
    }
    else {
        isHeader = false;
        mainTable = null;
    }

    header = null;
    this.tableStyle = tableStyle;

    

    // Create the row and column lists for the table
    rowList = new Vector();
    rowCount = 0;
    columnList = new Vector();

    // Add this table to the document's list of tables
    document = doc;
    document.addTable(this);

    // show the border on the top of the first row?
    if (!tableStyle.getShowFirstRowTopBorder()) {
        Row row = getRow(0);
        row.setBorderStyle(JCDrawStyle.BLANK);
    }

    // Set the alignment from the main table if one is supplied
    if (mainTable != null) {
        alignment = mainTable.getAlignment();
    }
    else {
        alignment = JCTextStyle.ALIGNMENT_CENTER;
    }

    // Create a cell to hold the table-wide cell defaults
    defaultCell = new Cell();

    // In the main table, create minimal table-wide defaults
    if (mainTable == null) {
        defaultCell.setCellAlignment(CELL_ALIGNMENT_TOP);
        JCUnit.Measure width = new JCUnit.Measure(JCUnit.POINTS, 2);
        defaultCell.setLeftMargin(width);
        defaultCell.setRightMargin(width);
        defaultCell.setTopMargin(width);
        defaultCell.setBottomMargin(width);
    }

    // set the text style
    defaultCell.setStyle(tableStyle.getTextStyle());

    // Create the TableDataModel
    tableData = new PageTableVectorDataSource(this);

    borderMode = BORDER_USE_EXTERNAL;
    dominance = ROW_DOMINANCE;
    overflowMode = OVERFLOW_WRAP_COLUMNS;

    isComplete = false;
}

/**
 * Fits the table into the specified frame, scaling the
 * columns in proportion to their original widths.
 * <br><B>Note:</b> This method should be called before "printing" to any of
 * the cells or the header.
 * @param frame the frame into which the table must fit; if the frame
 * is multi-column, then the table will be sized to the width of a column
 * @param style style to be used to render the table (needed to get the
 * margins used so that the constructor can accurately calculate
 * the width of the table)
 */
public void fitToFrame(JCFrame frame, JCTextStyle style) {
    int i;

    // Compute the width of a column of the frame
    JCUnit.Measure frame_width = (JCUnit.Measure) frame.getColumnWidth().clone();
    frame_width.subtract(style.getLeftIndent());
    frame_width.subtract(style.getRightIndent());

    // Go through the columns and figure out the total space taken up
    // by the borders

    // Total space occupied by borders
    JCUnit.Measure border_total = new JCUnit.Measure();
    // Current total width occupied by the columns
    JCUnit.Measure column_total = new JCUnit.Measure();
    border_total.add(getLeftBorder().getDrawnWidth());
    border_total.add(getRightBorder().getDrawnWidth());

    for (i = 0; i < columnCount; i++) {
        Column c = getColumn(i);
        column_total.add(c.getWidth());

		// only need width of last column -- border is handled by
		// right border width above
        if (i == (columnCount - 1)) {
            continue;
        }

        Column hc = getHeadersColumn(i);
        JCUnit.Measure border_width;

        if (hc == null) {
            border_width = c.getBorderWidth();
        }
        else {
            double max = Math.max(c.getBorderWidth().getAs(JCUnit.POINTS),
                               hc.getBorderWidth().getAs(JCUnit.POINTS));
            border_width = new JCUnit.Measure(JCUnit.POINTS, max);
        }

        border_total.add(border_width);
    }

    // Subtract the border total from the frame size to figure out
    // the total amount of space we have to fiddle with the columns
    frame_width.subtract(border_total);

    // Loop through the columns and scale them
    // in proportion to their original widths
    double scale_ratio = frame_width.getAs(JCUnit.POINTS) /
        						column_total.getAs(JCUnit.POINTS);

    for (i = 0; i < columnCount; i++) {
        Column c = getColumn(i);
        Column hc = getHeadersColumn(i);

        // this changes the width of the column
        c.getWidth().multiply(scale_ratio);

        if (hc != null) {
            hc.getWidth().multiply(scale_ratio);
        }
    }
}

/**
 * Returns the header table's <code>columnIndex</code> column.
 * @return header table's <code>columnIndex</code>
 * @param columnIndex
 */
public Column getHeadersColumn(int columnIndex) {
    if (getHeaders() != null) {
        return getHeaders().getColumn(columnIndex);
    }
    return null;
}

/**
 * Sets the width of a particular column. <br><B>Note:</b> This method should be
 * called before "printing" to any of the cells or the header.
 * @param columnIndex
 * @param width
 */
public void setColumnWidth(int columnIndex, JCUnit.Measure width) {
    Column column = getColumn(columnIndex);
    column.setWidth(width);
}

/**
 * Sets the width of a particular column. <br><B>Note:</b> This method should be
 * called before "printing" to any of the cells or the header.
 * @param columnWidths
 */
public void setColumnWidths(List columnWidths) {
    setColumnWidths(columnWidths.toArray());
}

/**
 * Sets the width of a particular column. <br><B>Note:</b> This method should be
 * called before "printing" to any of the cells or the header.
 * @param columnWidths
 */
public void setColumnWidths(JCUnit.Measure columnWidths[]) {
    setColumnWidths((Object []) columnWidths);
}

protected void setColumnWidths(Object widths[]) {
    for (int i = 0; i < widths.length; i++) {
        setColumnWidth(i, (JCUnit.Measure) widths[i]);
    }
}

/**
 * Adds a row of cells (or as many cells as are provided) to the table.
 * @param cellText an array of Strings that contain the cell text
 */
public void addRow(String[] cellText) {

    int         c, count;
    Row         row;
    Cell        cell;
    JCFrame     frame;
    JCTextStyle style;

    // Add a new row to the table
    row = getRow(rowCount);

    // Determine the number of cells to be added
    count = cellText.length;
    if (count > columnCount) {
        count = columnCount;
    }

    // Get each cell and print the given text to it
    for (c = 0; c < count; c++) {

        // Ignore null entries
        if (cellText[c] == null || cellText[c].length() == 0) {
            continue;
        }

        // Get the frame from the cell
        cell = row.getCell(c);
        frame = cell.getFrame();

        // Get the effective style applied to the cell
        style = cell.getStyle();
        if (style == null) {
            style = JCTextStyle.stringToStyle("default text");
        }

        // Print text to the cell
        try {
            frame.print(style, cellText[c]);
        }
        catch (EndOfFrameException e) {}
    }
}

/**
 * Adds a row of cells (or as many cells as are provided) to the table.
 * @param style the style to use for the text
 * @param cellText an array of Strings that contain the cell text
 */
public void addRow(JCTextStyle style, String[] cellText) {

    int     c, count;
    Row     row;
    Cell    cell;
    JCFrame frame;

    // Add a new row to the table
    row = getRow(rowCount);

    // Determine the number of cells to be added
    count = cellText.length;
    if (count > columnCount) {
        count = columnCount;
    }

    // Get each cell and print the given text to it
    for (c = 0; c < count; c++) {
        cell = row.getCell(c);
        frame = cell.getFrame();

        try {
            frame.print(style, cellText[c]);
        }
        catch (EndOfFrameException e) {}
    }
}

/**
 * Adds a row of cells (or as many cells as are provided) to the table,
 * given the row number.
 * @param rowNumber the row the cells are to be added in
 * @param cellText an array of Strings that contain the cell text
 */
public void addRow(int rowNumber, String[] cellText) {

    int         c, count;
    Row         row;
    Cell        cell;
    JCFrame     frame;
    JCTextStyle style;

    // Get the specified row from the table
    row = getRow(rowNumber);

    // Determine the number of cells to be added
    count = cellText.length;
    if (count > columnCount) {
        count = columnCount;
    }

    // Get each cell and print the given text to it
    for (c = 0; c < count; c++) {

        // Ignore null entries
        if (cellText[c] == null || cellText[c].length() == 0) {
            continue;
        }

        // ??? Should we ignore existing cells?
        // if (row.getCellRaw(c) != null) {
        //    continue;
        // }

        // Get the frame from the cell
        cell = row.getCell(c);
        frame = cell.getFrame();

        // Get the effective style applied to the cell
        style = cell.getStyle();
        if (style == null) {
            style = JCTextStyle.stringToStyle("default text");
        }

        // Print text to the cell
        try {
            frame.print(style, cellText[c]);
        }
        catch (EndOfFrameException e) {}
    }
}

/**
 * Adds a row of cells (or as many cells as are provided) to the table
 * starting at a given position. All cells must be on the given line.
 * @param rowNumber the row in which to add the cells
 * @param columnNumber the first column in which to place cell text
 * @param cellText an array of Strings that contain the cell text
 */
public void addRow(int rowNumber, int columnNumber, String[] cellText) {

    int         c, count;
    Row         row;
    Cell        cell;
    JCFrame     frame;
    JCTextStyle style;

    // Get the specified row from the table
    row = getRow(rowNumber);

    // Determine the number of cells to be added
    count = cellText.length;
    if (count > columnCount - columnNumber) {
        count = columnCount - columnNumber;
    }

    // Get each cell and print the given text to it
    for (c = 0; c < count; c++) {

        // Ignore null entries
        if (cellText[c] == null || cellText[c].length() == 0) {
            continue;
        }

        // ??? Should we ignore existing cells?
        // if (row.getCellRaw(c + columnNumber) != null) {
        //    continue;
        // }

        // Get the frame from the cell
        cell = row.getCell(c + columnNumber);
        frame = cell.getFrame();

        // Get the effective style applied to the cell
        style = cell.getStyle();
        if (style == null) {
            style = JCTextStyle.stringToStyle("default text");
        }

        // Print text to the cell
        try {
            frame.print(style, cellText[c]);
        }
        catch (EndOfFrameException e) {}
    }
}

/**
 * Adds a row of cell values (or as many values as are provided on one line)
 * to the end of the table.
 * @param cellValues an array of Objects which contain the cell values
 */
public void addRow(Object[] cellValues) {

    int         c, count;

    // Add a new row to the table
    Row row = getRow(rowCount);

    // Determine the number of cells to be added
    count = cellValues.length;
    if (count > columnCount) {
        count = columnCount;
    }

    // Get each cell and store the given value in it
    for (c = 0; c < count; c++) {

        // Ignore null entries
        if (cellValues[c] == null) {
            continue;
        }

        // Store the value in the TableDataModel at the cell
        Cell cell = row.getCell(c);
        cell.setCellValue(cellValues[c]);
    }
}

/**
 * Adds a row of cell values (or as many values as are provided) to the
 * table in the given position.
 * @param rowNumber the row in which to add the cell values
 * @param cellValues an array of Objects which contain the cell values
 */
public void addRow(int rowNumber, Object[] cellValues) {

    int         c, count;

    // Add a new row to the table
    Row row = getRow(rowCount);

    // Determine the number of cells to be added
    count = cellValues.length;
    if (count > columnCount) {
        count = columnCount;
    }

    // Get each cell and store the given value in it
    for (c = 0; c < count; c++) {

        // Ignore null entries
        if (cellValues[c] == null) {
            continue;
        }

        // Store the value in the TableDataModel at the cell
        Cell cell = row.getCell(c);
        cell.setCellValue(cellValues[c]);
    }
}

/**
 * Adds a row of cell values (or as many values as are provided) to the table
 * starting at a given position. All cells must be on the given line.
 * @param rowNumber the row in which to add the cell values
 * @param columnNumber the first column in which to place a cell value
 * @param cellValues an array of Objects which contain the cell values
 */
public void addRow(int rowNumber, int columnNumber, Object[] cellValues) {

    int         c, count;

    // Add a new row to the table
    Row row = getRow(rowNumber);

    // Determine the number of cells to be added
    count = cellValues.length;
    if (count > columnCount - columnNumber) {
        count = columnCount - columnNumber;
    }

    // Get each cell and store the given value in it
    for (c = 0; c < count; c++) {

        // Ignore null entries
        if (cellValues[c] == null) {
            continue;
        }

        // Store the value in the TableDataModel at the cell
        Cell cell = row.getCell(c + columnNumber);
        cell.setCellValue(cellValues[c]);
    }
}

/**
 * Adds a row of cell values (or as many values as are provided) to the table
 * starting at a given position. All cells must be on the given line.
 * @param row the row in which to add the cell values
 * @param column the first column in which to place a cell value
 * @param cellValues a list of cell value Objects
 */
/*
public void addRow(int rowNumber, int columnNumber, List cellValues) {

    Iterator    cvi;
    Object        entry;

    // Add a new row to the table
    Row row = getRow(rowNumber);

    // Get each cell and store the given value in it
    cvi = cellValues.iterator();
    while (cvi.hasNext() && columnNumber < columnCount) {

        entry = cvi.next();

        // Ignore null entries
        if (entry == null) {
            continue;
        }

        // Store the value in the TableDataModel at the cell
        Cell cell = row.getCell(c + columnNumber);
        cell.setCellValue(entry);
    }
}
*/

/**
 * Gets the alignment of the table within its frame/column.
 * @return the horizontal alignment (justification) of the table
 */
public int getAlignment() {
    return(alignment);
}

/**
 * Sets the alignment of the table within its frame/column.
 * @param alignment the horizontal alignment (justification) of the table
 */
public void setAlignment(int alignment) {

    // Justified alignment to be the same as centred
    if (alignment <= 0 || alignment > JCTextStyle.ALIGNMENT_JUSTIFY) {
        throw new IllegalArgumentException("Unrecognised horizontal alignment");
    }

    // We interpret (alignment == NONE) => get alignment from current style
    this.alignment = alignment;

    // Pass the alignment change down to the header table
    if (header != null) {
        header.setAlignment(alignment);
    }
}

/**
 * Determines if the background color is inherited from a higher-level table.
 * @return <code>true</code> if the background color is inherited
 */
public boolean isBackgroundColorInherited() {

    return(isHeader && tableStyle.getBackground() == null);
}

/**
 * Gets the color specified for painting the table background.
 * @return the table's background color
 */
public Color getBackgroundColor() {

    if (isBackgroundColorInherited()) {
        return(mainTable.getBackgroundColor());
    }

    return(tableStyle.getBackground());
}

/**
 * Sets the color specified for painting the table background.
 * @param color the table's background color
 */
public void setBackgroundColor(Color color) {

    tableStyle.setBackground(color);
}

/**
 * Determines the mode of selecting perimeter borders for a range of cells.
 * @return the border mode (whether cell or table borders are used)
 */
public int getBorderMode() {
    return(borderMode);
}

/**
 * Specifies the mode of selecting perimeter borders for a range of cells.
 * @param mode the border mode (whether cell or table borders are used)
 */
public void setBorderMode(int mode) {

    if (mode <= 0 || mode > BORDER_USE_EXTERNAL) {
        throw new IllegalArgumentException("Unrecognised border mode");
    }

    // NOTE:  we may need to pass this to a header table
    borderMode = mode;
}

/**
 * Gets the draw style used for the top border.
 * @return the draw style which controls the appearance of the top border
 */
public JCDrawStyle getTopBorder() {

    if (isTopBorderInherited()) {
        return(mainTable.getTopBorder());
    }
    return(tableStyle.getTopBorder());
}

/**
 * Determines if the value reported for the top border is inherited.
 * @return <code>true</code> if the top border style is inherited from a higher level
 */
public boolean isTopBorderInherited() {

    return(isHeader && tableStyle.getTopBorder() == null);
}

/**
 * Sets the draw style used for the top border.
 * @param border the draw style to apply to the table's top border
 */
public void setTopBorder(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    tableStyle.setTopBorder(border);
}

/**
 * Gets the draw style used for the bottom border.
 * @return the draw style which controls the appearance of the bottom border
 */
public JCDrawStyle getBottomBorder() {

    if (isBottomBorderInherited()) {
        return(mainTable.getBottomBorder());
    }
    return(tableStyle.getBottomBorder());
}

/**
 * Determines if the value reported for the bottom border is inherited.
 * @return <code>true</code> if the bottom border style is inherited from a higher level
 */
public boolean isBottomBorderInherited() {

    return(isHeader && tableStyle.getBottomBorder() == null);
}

/**
 * Sets the draw style used for the bottom border.
 * @param border the draw style to apply to the table's bottom border
 */
public void setBottomBorder(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    tableStyle.setBottomBorder(border);
}

/**
 * Gets the draw style used for the left border.
 * @return the draw style which controls the appearance of the left border
 */
public JCDrawStyle getLeftBorder() {

    if (isLeftBorderInherited()) {
        return(mainTable.getLeftBorder());
    }
    return(tableStyle.getLeftBorder());
}

/**
 * Determines if the value reported for the left border is inherited.
 * @return <code>true</code> if the left border style is inherited from a higher level
 */
public boolean isLeftBorderInherited() {

    return(isHeader && tableStyle.getLeftBorder() == null);
}

/**
 * Sets the draw style used for the left border.
 * @param border the draw style to apply to the table's left border
 */
public void setLeftBorder(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    tableStyle.setLeftBorder(border);
}

/**
 * Gets the draw style used for the right border.
 * @return the draw style which controls the appearance of the right border
 */
public JCDrawStyle getRightBorder() {

    if (isRightBorderInherited()) {
        return(mainTable.getRightBorder());
    }
    return(tableStyle.getRightBorder());
}

/**
 * Determines if the value reported for the right border is inherited.
 * @return <code>true</code> if the right border style is inherited from a higher level
 */
public boolean isRightBorderInherited() {

    return(isHeader && tableStyle.getRightBorder() == null);
}

/**
 * Sets the draw style used for the right border.
 * @param border the draw style to apply to the table's right border
 */
public void setRightBorder(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    tableStyle.setRightBorder(border);
}

/**
 * Gets the draw style used for the row border.
 * @return the draw style which controls the appearance of the row border
 */
public JCDrawStyle getRowBorder() {

    if (isRowBorderInherited()) {
        return(mainTable.getRowBorder());
    }
    return(tableStyle.getRowBorder());
}

/**
 * Determines if the value being reported for the row border is inherited.
 * @return <code>true</code> if the row border style is inherited from a higher level
 */
public boolean isRowBorderInherited() {

    return(isHeader && tableStyle.getRowBorder() == null);
}

/**
 * Sets the draw style used for the row border.
 * @param border the draw style to apply to the table's row border
 */
public void setRowBorder(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    tableStyle.setRowBorder(border);
}

/**
 * Gets the draw style used for the column border.
 * @return the draw style which controls the appearance of the column border
 */
public JCDrawStyle getColumnBorder() {

    if (isColumnBorderInherited()) {
        return(mainTable.getColumnBorder());
    }
    return(tableStyle.getColumnBorder());
}

/**
 * Determines if the value reported for the column border is inherited.
 * @return <code>true</code> if the column border style is inherited from a higher level
 */
public boolean isColumnBorderInherited() {

    return(isHeader && tableStyle.getColumnBorder() == null);
}

/**
 * Sets the draw style used for the column border.
 * @param border the draw style to apply to the table's column border
 */
public void setColumnBorder(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    tableStyle.setColumnBorder(border);
}

/**
 * Gets the draw style used for the header border.
 * @return the draw style which controls the appearance of the header border
 */
public JCDrawStyle getHeaderBorder() {
    if (isHeader()) {
        return(mainTable.getHeaderBorder());
    }
    return(tableStyle.getHeaderBorder());
}

/**
 * Determines if the value being reported for the header border is inherited.
 * @return <code>true</code> if the header border style is inherited from a higher level
 */
public boolean isHeaderBorderInherited() {

    return(isHeader && tableStyle.getHeaderBorder() == null);
}

/**
 * Sets the draw style used for the header border.
 * @param border the draw style to apply to the table's header border
 */
public void setHeaderBorder(JCDrawStyle border) {
    if (isHeader()) {
        throw new IllegalArgumentException("Must set the header border on the main table");
    }

    // The style must be immutable
    border = border.getImmutable();

    tableStyle.setHeaderBorder(border);
}

/**
 * Sets the draw style used for all table borders.
 * @param border the draw style to apply to all the table's borders
 */
public void setAllBorders(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    setBottomBorder(border);
    setTopBorder(border);
    setLeftBorder(border);
    setRightBorder(border);
    setRowBorder(border);
    setColumnBorder(border);
    setHeaderBorder(border);
}

/**
 * Sets the draw style used for all borders in the table's perimeter.
 * @param border the draw style to apply to all the table's edge borders
 */
public void setEdgeBorders(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    setBottomBorder(border);
    setTopBorder(border);
    setLeftBorder(border);
    setRightBorder(border);
}

/**
 * Sets the draw style used for the table's horizontal borders.
 * @param border the draw style to apply to all horizontal borders
 */
public void setHorizontalBorders(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    setBottomBorder(border);
    setTopBorder(border);
    setHeaderBorder(border);
    setRowBorder(border);
}

/**
 * Sets the draw style used for the table's internal borders.
 * @param border the draw style to apply internally between rows and columns
 */
public void setInternalBorders(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    setRowBorder(border);
    setColumnBorder(border);
}

/**
 * Sets the draw style used for the table's vertical borders.
 * @param border the draw style to apply to all vertical borders
 */
public void setVerticalBorders(JCDrawStyle border) {

    // The style must be immutable!
    border = border.getImmutable();

    setLeftBorder(border);
    setRightBorder(border);
    setColumnBorder(border);
}

/**
 * Gets a particular cell from the table, instantiating it if necessary.
 * @param row the row in which to find or create the cell
 * @param column the column that holds the cell
 * @return the table cell that occupies or obscures (spans) this cell
 */
public Cell getCell(int row, int column) {

    Row         rowObj;
    Cell        cellObj;

    // Verify that the cell is in range
    if (row < 0 || column < 0 || column >= columnCount) {
        throw new IllegalArgumentException("cell position out of range");
    }

    // Get the row and then the cell
    rowObj = getRow(row);
    cellObj = rowObj.getCell(column);

    // If the cell is obscured by a span, get the root cell
    if (cellObj.isSpanned()) {
        Point origin = cellObj.getSpanOrigin();

        if (origin.x != column || origin.y != row) {
            cellObj = getCell(origin.y, origin.x);
        }
    }

    return(cellObj);
}

/**
 * Gets the frame that holds the contents of a given cell.
 * @param row the cell's row
 * @param column the cell's column
 * @return the frame object owned by the target cell
 */
public JCFrame getCellFrame(int row, int column) {

    Cell        cell;

    cell = getCell(row, column);
    return(cell.getFrame());
}

/**
 * Prints the given text to a specified cell, using the cell's style.
 * @param row the cell's row
 * @param column the cell's column
 * @param text the text to be printed in the cell
 */
public void printToCell(int row, int column, String text) {

    JCFrame         frame;
    JCTextStyle     style;
    Cell            cell;

    // Get the frame from the cell
    cell = getCell(row, column);
    frame = cell.getFrame();

    // Get the style applied to the cell
    style = cell.getStyle();
    if (style == null) {
        style = JCTextStyle.stringToStyle("default text");
    }

    // Print text to the cell
    try {
        frame.print(style, text);
    }
    catch (EndOfFrameException e) {}
}

/**
 * Prints the given text to a specified cell using the given style.
 * @param row the cell's row
 * @param column the cell's column
 * @param style the style to use to print the text
 * @param text the text to be printed in the cell
 */
public void printToCell(int row, int column, JCTextStyle style, String text) {

    JCFrame         frame;
    Cell            cell;

    // Get the frame from the cell
    cell = getCell(row, column);
    frame = cell.getFrame();

    // Print text to the cell
    try {
        frame.print(style, text);
    }
    catch (EndOfFrameException e) {}
}

/**
 * Determines if the vertical cell alignment is inherited.
 * @return <code>true</code> if the table is examined to get the cell alignment
 */
public boolean isDefaultCellAlignmentInherited() {
    return(isHeader && defaultCell.isCellAlignmentInherited());
}

/**
 * Gets the vertical alignment of the cells in this table.
 * @return the alignment of each cell with respect to its row
 */
public int getDefaultCellAlignment() {

    if (isDefaultCellAlignmentInherited()) {
        return(mainTable.getDefaultCellAlignment());
    }

    return(defaultCell.getCellAlignment());
}

/**
 * Sets the vertical alignment of the cells in this table.
 * @param alignment the alignment of each cell with respect to its row
 */
public void setDefaultCellAlignment(int alignment) {

    defaultCell.setCellAlignment(alignment);
    // NOTE:  we may need to pass the alignment change to a header table
}

/**
 * Gets a column object corresponding to a column number.
 * @param column the number of the column to retrieve
 * @return the column object that represents this column
 */
public Column getColumn(int column) {

    // Verify that the column index is in range
    if (column < 0 || column >= columnCount) {
        throw new IllegalArgumentException("column index out of range");
    }

    // Return the selected column
    return((Column) columnList.get(column));
}

/**
 * Gives the number of columns in the table.
 * @return a count of the number of columns in the table
 */
public int getColumnCount() {
    return(columnCount);
}

/**
 * Returns the list of columns.
 * @return an ordered list of the columns of the table
 */
public List getColumnList() {
    return((List) columnList);
}

/**
 * Determines if the default style is inherited.
 * @return <code>true</code> if there is a parent table from which to get the default style
 */
public boolean isDefaultStyleInherited() {

    return(defaultCell.isStyleInherited() && isHeader);
}

/**
 * Gets the default cell text style.
 * @return the default text style used in cell text
 */
public JCTextStyle getDefaultStyle() {

    // If this is a header inheriting a default, get it instead
    if (isDefaultStyleInherited()) {
        return(mainTable.getDefaultStyle());
    }

    // Get the style from the default cell
    return(defaultCell.getStyle());
}

/**
 * Sets the default cell text style.
 * @param style a text style to apply to printing in table cells
 */
public void setDefaultStyle(JCTextStyle style) {

    defaultCell.setStyle(style);
}

/**
 * Determines if the default bottom margin is inherited.
 * @return <code>true</code> if there is a parent to provide the default bottom margin
 */
public boolean isDefaultBottomMarginInherited() {

    return(defaultCell.isBottomMarginInherited() && isHeader);
}

/**
 * Gets the default width of the bottom margin.
 * @return a measure of the width of the default bottom margin for all cells
 */
public JCUnit.Measure getDefaultBottomMargin() {

    // If this table inherits a default bottom margin
    if (isDefaultBottomMarginInherited()) {

        // Try the parent's default cell
        return(mainTable.getDefaultBottomMargin());
    }
    return(defaultCell.getBottomMargin());
}

/**
 * Sets the default width of a cell's bottom margin.
 * @param width the size of the bottom margin for table cells
 */
public void setDefaultBottomMargin(JCUnit.Measure width) {

    defaultCell.setBottomMargin(width);
}

/**
 * Determines if the default top margin is inherited.
 * @return <code>true</code> if there is a parent to get the default top margin from
 */
public boolean isDefaultTopMarginInherited() {

    return(defaultCell.isTopMarginInherited() && isHeader);
}

/**
 * Gets the default width of the top margin.
 * @return a measure of the width of the default top margin for all cells
 */
public JCUnit.Measure getDefaultTopMargin() {

    // If this table inherits a default top margin
    if (isDefaultTopMarginInherited()) {

        // Try the parent's default cell
        return(mainTable.getDefaultTopMargin());
    }
    return(defaultCell.getTopMargin());
}

/**
 * Sets the default width of a cell's top margin.
 * @param width the size of the top margin for table cells
 */
public void setDefaultTopMargin(JCUnit.Measure width) {

    defaultCell.setTopMargin(width);
}

/**
 * Determines if the default left margin is inherited.
 * @return <code>true</code> if there is a parent to get the default left margin from
 */
public boolean isDefaultLeftMarginInherited() {

    return(defaultCell.isLeftMarginInherited() && isHeader);
}

/**
 * Gets the default width of the left margin.
 * @return a measure of the width of the default left margin for all cells
 */
public JCUnit.Measure getDefaultLeftMargin() {

    // If this table inherits a default left margin
    if (isDefaultLeftMarginInherited()) {

        // Try the parent's default cell
        return(mainTable.getDefaultLeftMargin());
    }
    return(defaultCell.getLeftMargin());
}

/**
 * Sets the default width of a cell's left margin.
 * @param width the size of the left margin for table cells
 */
public void setDefaultLeftMargin(JCUnit.Measure width) {

    defaultCell.setLeftMargin(width);
}

/**
 * Determines if the default right margin is inherited.
 * @return <code>true</code> if there is a parent to provide the default right margin
 */
public boolean isDefaultRightMarginInherited() {

    return(defaultCell.isRightMarginInherited() && isHeader);
}

/**
 * Gets the default width of the right margin.
 * @return a measure of the width of the default right margin for all cells
 */
public JCUnit.Measure getDefaultRightMargin() {

    // If this table inherits a default right margin
    if (isDefaultRightMarginInherited()) {

        // Try the parent's default cell
        return(mainTable.getDefaultRightMargin());
    }
    return(defaultCell.getRightMargin());
}

/**
 * Sets the default width of a cell's right margin.
 * @param width the size of the right margin for table cells
 */
public void setDefaultRightMargin(JCUnit.Measure width) {

    defaultCell.setRightMargin(width);
}

/**
 * Gets the table data which stores values associated with the table.
 * @return a <code>TableDataModel</code> which can store values and expressions
 */
public TableDataModel getTableData() {

    return(tableData);
}

/**
 * Gets the document to which this table belongs.
 * @return the document which owns this table
 */
public JCDocument getDocument() {
    return(document);
}

/**
 * Sets the document to which this table belongs.
 * @param doc the document which is to own this table
 */
public void setDocument(JCDocument doc) {

    int            r, c;
    Cell        cell;
    JCFrame     frame;

    // Set the document
    document = doc;

    // Find all frames in the table and set the document against them
    for (r = 0; r < rowCount; r++) {
        for (c = 0; c < columnCount; c++) {

            // Get whatever cell exists at this point
            cell = getCellRaw(r, c);

            // Ignore if no cell exists
            if (cell == UNINITIALISED_CELL) {
                continue;
            }

            // Get this cell's frame, if one exists
            frame = cell.getFrameRaw();
            if (frame != null) {

                // If the frame exists, set its document
                frame.setDocument(doc);
            }
        }
    }
}

/**
 * Gets the dominant direction for precedence of row and column attributes.
 * @return the direction to give dominance for attributes
 */
public int getRowColumnDominance() {
    return(dominance);
}

/**
 * Sets the dominant direction for precedence of row and column attributes.
 * @param dominance the direction to give dominance for attributes
 */
public void setRowColumnDominance(int dominance) {

    if (dominance <= 0 || dominance > COLUMN_DOMINANCE) {
        throw new IllegalArgumentException("Unrecognised row/column dominance");
    }

    this.dominance = dominance;
}

/**
 * Gets the table which defines the header of this table.
 * @return the table which contains the rows of this table's header
 */
public JCPageTable getHeaders() {

    // NOTE:  Unlike PageFormatter this doesn't create the table if it
    //        doesn't already exist.  (See below for the create method)

    return(header);
}

/**
 * Creates a header table for this table.
 * <br><B>NOTE:</B> This table can't be a header table already.
 * @return
 */
public JCPageTable createHeaders() {

    int     i;
    Column  column;
    List    column_widths;

    // Error if this table is a header table
    if (isHeader) {
        throw new IllegalArgumentException("Can't create headers for headers");
    }

    // Create a new table with the same columns as the current table
    column_widths = new Vector();
    for (i = 0; i < columnCount; i++) {
        column = (Column) columnList.get(i);
        column_widths.add(column.getWidth());
    }
    header = new JCPageTable(document, column_widths, this, tableStyle.getHeaderStyle());

    // return the new header table
    return(header);
}

/**
 * Determines if this table has a header.
 * @return <code>true</code> if this table is associated with a header table
 */
public boolean hasHeaders() {
    return(header != null);
}

/**
 * Determines if this table forms the header of another table.
 * @return <code>true</code> if this table's rows are the header for another table
 */
public boolean isHeader() {
    return(isHeader);
}

/**
 * Gets the mode switch which controls the placement of columns that do not fit.
 * @return the overflow mode of the table
 */
public int getOverflowMode() {
    return(overflowMode);
}

/**
 * Sets the mode switch which controls the placement of columns that do not fit.
 * @param overflow_mode the overflow mode of the table
 */
public void setOverflowMode(int overflow_mode) {

    if (overflow_mode <= 0 || 
        overflow_mode > OVERFLOW_WRAP_COLUMNS_NEXT ||
        overflow_mode == OVERFLOW_TEAROFF_COLUMNS) {

        throw new IllegalArgumentException("Unrecognised overflow mode");
    }

    this.overflowMode = overflow_mode;
}

/**
 * Gets a row from the table, creating it if it doesn't exist.
 * @param row the number of the row to retrieve
 * @return the selected row of the table
 */
public Row getRow(int row) {

    Row         rowObj;

    // Verify that the row is in range
    if (row < 0) {
        throw new IllegalArgumentException("row out of range");
    }

    // Determine if the row is inside the range of the table
    if (row < rowCount) {

        // Get the row
        rowObj = (Row) rowList.get(row);

        // Create the row if it does not yet exist
        if (rowObj == UNINITIALISED_ROW) {
            rowObj = new Row(row);
            rowList.set(row, rowObj);
        }

    }

    // Otherwise we need to extend the table for the new row
    else {

        // Create any intervening blank rows
        while (rowCount < row) {        // for (; rowCount < row; rowCount++)
            rowList.add(UNINITIALISED_ROW);
            rowCount++;
        }

        // Create the row and add it to the table
        rowObj = new Row(row);
        rowList.add(rowObj);
        rowCount = rowList.size();

    }

    // set the color on the row if it alternates
    if (tableStyle.getAlternate() != null &&
            tableStyle.getAlternate().rowOriented) {
        if (row % 2 == 0) {
            rowObj.setBackgroundColor(tableStyle.getAlternate().c1);
        }
        else {
            rowObj.setBackgroundColor(tableStyle.getAlternate().c2);
        }
    }

    // Return the row
    return(rowObj);
}

/**
 * Gives the number of rows in the table.
 * @return a count of the number of rows in the table
 */
public int getRowCount() {
    return(rowCount);
}

/**
 * Returns the list of rows.
 * @return an ordered list of table rows
 */
public List getRowList() {
    return((List) rowList);
}

/**
 * Creates a span across the given range of cells.
 * @param startRow the row index of the upper-left cell of the span
 * @param startColumn the column index of the upper-left cell of the span
 * @param rows the number of rows to span downward
 * @param columns the number of columns to span to the right
 */
public void spanCells(int startRow, int startColumn, int rows, int columns) {

    int             r, c, cellCount;
    Row             row;
    Cell            cell;
    List            cells;
    JCUnit.Measure  cell_width;

    // Ensure the range is valid
    if (startRow < 0 || startColumn < 0 || startColumn >= columnCount) {
        throw new IllegalArgumentException("Cell position out of range");
    }
    if (rows < 1 || columns < 1 || startColumn + columns > columnCount) {
        throw new IllegalArgumentException("Invalid span range");
    }

    // A degenerate span is one cell in size; ignore it
    if (rows == 1 && columns == 1) {
        return;
    }

    // Check that the range is clear of other spans
    for (r = 0; r < rows; r++) {

        // Ignore rows past the end of this table
        if (startRow + r >= rowCount) {
            continue;
        }

        row = (Row) rowList.get(startRow + r);

        // Ignore uninitialized rows, since we initialize rows to add a span
        if (row != UNINITIALISED_ROW) {
            cells = row.getCellList();
            cellCount = cells.size();

            for (c = 0; c < columns; c++) {

                // Ignore cells past the end of this row
                if (startColumn + c >= cellCount) {
                    continue;
                }

                cell = (Cell) cells.get(startColumn + c);

                // Uninitialized cells are not spanned (by definition)
                if (cell != UNINITIALISED_CELL && cell.isSpanned()) {
                    throw new IllegalArgumentException("Span range conflicts");
                }
            }
        }
    }

    // Mark the cells as spanned
    for (r = 0; r < rows; r++) {

        // Use "getRow()" to create rows that are uninitialized
        row = getRow(startRow + r);

        for (c = 0; c < columns; c++) {

            // Use "getCell()" to create cells that are uninitialized
            cell = row.getCell(startColumn + c);

            // Setting the span origin marks the cell as spanned
            cell.setSpanOrigin(startRow, startColumn);
        }
    }

    // Save the span parameters at the root cell
    cell = getCell(startRow, startColumn);
    cell.setSpanSize(rows, columns);

    // Adjust the cell size (width)
    cell_width = (JCUnit.Measure) getColumn(startColumn).getWidth().clone();
    for (c = 1; c < columns; c++) {

        // NOTE:  the cell width can _only_ take into account the
        //        _existing_ border spacing
        cell_width.add(getColumn(startColumn).getBorderWidth());

        cell_width.add(getColumn(startColumn + c).getWidth());
    }
    cell.setWidth(cell_width);

}

/**
 * Determines the height required for a given range of rows.
 * @param first_row the number of the first row in the range
 * @param last_row the number of the last row in the range
 * @return the total height of the given range of rows, including border space
 */
public JCUnit.Measure getRowRangeHeight(int first_row, int last_row) {

    int             r;
    Row             row;
    JCUnit.Measure  total_height;

    

    // Check the validity of the given range
    if (first_row < 0 || last_row >= rowCount) {
        throw new IllegalArgumentException("Invalid row range");
    }

    // Get the width of the top border of the range
    row = this.getRow(first_row);
    total_height = row.getTopBorderWidth(true);
    

    for (r = first_row + 1; r <= last_row; r++) {
        

        // Add the full height of the row and its border
        total_height.add(row.getHeight());
        
        total_height.add(row.getExtraHeight());
        
        total_height.add(row.getBorderWidth());
        

        // Move to the next row in the range
        row = this.getRow(r);
    }

    // Add the height of the last row and the bottom border
    
    total_height.add(row.getHeight());
    
    total_height.add(row.getExtraHeight());
    
    total_height.add(row.getBottomBorderWidth(true));
    

    
    return(total_height);
}

/**
 * Determines the width required for a given range of columns.
 * @param first_column the number of the first column in the range
 * @param last_column the number of the last column in the range
 * @return the total width of the given range of columns, including border space
 */
public JCUnit.Measure getColumnRangeWidth(int first_column, int last_column) {

    int             c;
    Column          column;
    JCUnit.Measure  total_width;

    // Check the validity of the given range
    if (first_column < 0 || last_column >= columnCount) {
        throw new IllegalArgumentException("Invalid column range");
    }

    // Get the width of the top border of the range
    column = this.getColumn(first_column);
    total_width = column.getLeftBorderWidth(true);

    for (c = first_column + 1; c <= last_column; c++) {

        // Add the full width of the column and its border
        total_width.add(column.getWidth());
        total_width.add(column.getBorderWidth());

        // Move to the next column in the range
        column = this.getColumn(c);
    }

    // Add the width of the last column and the right border
    total_width.add(column.getWidth());
    total_width.add(column.getRightBorderWidth(true));

    return(total_width);
}

/**
 * Determines the range of rows that fits in a certain space, given a start row.
 * @param startRow the first row of the range
 * @param available the space in which to fit the range of rows
 * @param consumed this is set to the amount of space actually used
 * @return the last row of the range (inclusive)
 */
public int getRowSet(int startRow, JCUnit.Measure available,
                     JCUnit.Measure consumed) {

    int             endRow;
    JCUnit.Measure  space, used;

    // Determine the height of the single-row range
    endRow = startRow;
    space = getRowRangeHeight(startRow, endRow);
    used = new JCUnit.Measure();

    // Add rows while there are space and rows to add
    while (available.greaterThan(space) && endRow < rowCount - 1) {
        used = space;
        space = getRowRangeHeight(startRow, ++endRow);
    }

    // If we ran out of space (rather than rows), we leave off the final
    // partial row, as long as at least one row is printed
    if (available.lessThan(space) && endRow > startRow) {
        endRow--;
    }
    // If we are keeping all the rows, update the space consumed
    else {
        used = space;
    }

    // Copy the used space to the space consumed variable
    consumed.assign(used);

    return(endRow);
}

/**
 * Determines the range of columns that fits in a space, given the start column.
 * @param startColumn the first column of the range
 * @param available the space in which to fit the range of columns
 * @param consumed this is set to the amount of space actually used
 * @return the last column of the range (inclusive)
 */
public int getColumnSet(int startColumn, JCUnit.Measure available,
                        JCUnit.Measure consumed) {

    int             endColumn;
    JCUnit.Measure  space, used;

    // Determine the height of the single-column range
    endColumn = startColumn;
    space = getColumnRangeWidth(startColumn, endColumn);
    used = new JCUnit.Measure();

    // Add columns while there are space and columns to add
    while (available.greaterThan(space) && endColumn < columnCount - 1) {
        used = space;
        space = getColumnRangeWidth(startColumn, ++endColumn);
    }

    // If we ran out of space (rather than columns), we leave off the final
    // partial column, as long as at least one column is printed
    if (available.lessThan(space) && endColumn > startColumn) {
        endColumn--;
    }
    // If we are keeping all the columns update the space consumed
    else {
        used = space;
    }

    // Copy the used space to the space consumed variable
    consumed.assign(used);

    return(endColumn);
}

/**
 * Completes the layout tasks for a table, such as computing frame and row heights.
 */
public void completeLayout() {

    if (!isComplete) {

        // if the table has headers, complete them
        if (hasHeaders()) {
            header.completeLayout();

            // if there are no header rows, mark them as null
            if (header.getRowCount() == 0) {
                header = null;
            }
        }

        // Compute the row heights of the table
        computeRowHeights();
        isComplete = true;
    }
    
    
}

/**
 * This method is used internally by JClass PageLayout and should not be
 * used by applications.
 *
 * <P>Draws a set (rectangular row/column range) of cells from a table.
 * @param frame the frame in which the table is being constructed
 * @param startRow the first row of the table from which to paste cells
 * @param startColumn the first column of the table from which to paste cells
 * @param endRow the last row of the table from which to paste cells
 * @param endColumn the last column of the table from which to paste cells
 * @param headerH the amount of vertical space required for headers
 * @param tableH the amount of vertical space computed for the table
 * @param tableW the amount of horizontal space allocated to the table part
 */
public void drawTablePart(JCFrame frame, int startRow, int startColumn,
                          int endRow, int endColumn, JCUnit.Measure headerH,
                          JCUnit.Measure tableH, JCUnit.Measure tableW) {

    int             table_align;
    Row             row;
    Column          column;
    JCFlowInfo      frame_info;
    JCUnit.Measure  x0, y0, x1, y1, x2, y2;
    JCUnit.Measure  tmp_offset;
    JCUnit.Measure  left_offset, right_offset, top_offset, bottom_offset;

    // get the alignment of the table (headers conform to their parent)
    if (isHeader) {
        table_align = mainTable.getAlignment();
    }
    else {
        table_align = alignment;
    }

    // Establish the origin of the table at the left margin of the frame
    frame_info = frame.getFlowInfo();
    x0 = (JCUnit.Measure) frame_info.leftOfColumn.clone();
    y0 = frame_info.insertion.getY();

    // adjust the horizontal position for the table's alignment
    if (alignment == JCTextStyle.ALIGNMENT_RIGHT) {
        x0.add(frame.getColumnWidth());
        x0.subtract(tableW);
    }
    else if (alignment == JCTextStyle.ALIGNMENT_CENTER ||
             alignment == JCTextStyle.ALIGNMENT_JUSTIFY) {

        JCUnit.Measure  temp = (JCUnit.Measure) frame.getColumnWidth().clone();
        temp.subtract(tableW);
        temp.divide(2.0);

        x0.add(temp);
    }
    else {    // (alignment == ALIGNMENT_NONE || alignment == ALIGNMENT_LEFT)
        // do nothing in this case
    }

    // the amount of space required for the external borders is half the
    // width of the border along that side of the table

    // compute the offset of the left column from the outside table edge
    if (startColumn == 0 || borderMode == BORDER_USE_EXTERNAL) {

        // Use the max of the table's and header's left border width
        left_offset = getLeftBorder().getDrawnWidth();
        if (isHeader) {
            tmp_offset = mainTable.getLeftBorder().getDrawnWidth();
            if (tmp_offset.greaterThan(left_offset)) {
                left_offset = tmp_offset;
            }
        }
        else if (hasHeaders()) {
            tmp_offset = header.getLeftBorder().getDrawnWidth();
            if (tmp_offset.greaterThan(left_offset)) {
                left_offset = tmp_offset;
            }
        }
    }
    // if not using an external border, use the previous column's
    else {
        left_offset = getColumn(startColumn - 1).getBorderWidth();
    }

    // compute half of the border width ("getDrawnWidth()" returns a copy of
    // the width, so it may be safely manipulated)
    left_offset.divide(2.0);

    // compute the offset of the right column from the outside table edge
    if (endColumn == columnCount - 1 || borderMode == BORDER_USE_EXTERNAL) {

        // Use the max of the table's and header's left border width
        right_offset = getRightBorder().getDrawnWidth();
        if (isHeader) {
            tmp_offset = mainTable.getRightBorder().getDrawnWidth();
            if (tmp_offset.greaterThan(right_offset)) {
                right_offset = tmp_offset;
            }
        }
        else if (hasHeaders()) {
            tmp_offset = header.getRightBorder().getDrawnWidth();
            if (tmp_offset.greaterThan(right_offset)) {
                right_offset = tmp_offset;
            }
        }
    }
    // if not using an external border, use the last column's
    else {
        right_offset = getColumn(endColumn).getBorderWidth();
    }

    // compute half of the border width ("getDrawnWidth()" returns a copy of
    // the width, so it may be safely manipulated)
    right_offset.divide(2.0);

    // compute the coordinates of the left and right edges
    x1 = (JCUnit.Measure) x0.clone();
    x1.add(left_offset);
    x2 = (JCUnit.Measure) x0.clone();
    x2.add(tableW);
    x2.subtract(right_offset);

    // Compute the width of the table's top boundary space
    if (hasHeaders()) {
        // if there are headers, offsets will be handled when the header
        // is drawn or in the individual draw routines
        top_offset = new JCUnit.Measure();
    }
    else if (startRow == 0 || borderMode == BORDER_USE_EXTERNAL) {
        // there are no headers and this is the top row
        top_offset = getTopBorder().getDrawnWidth();
    }
    else {
        // remaining cases:  rows after the first, no headers, internal borders
        // use the border of the row above this one
        top_offset = getRow(startRow - 1).getBorderWidth();
    }

    // compute half of the border width
    top_offset = (JCUnit.Measure) top_offset.clone();
    top_offset.divide(2.0);

    // Compute the boundary space at the bottom of the table
    if (isHeader) {
        // a header table is _always_ followed by the header border
        bottom_offset = getHeaderBorder().getDrawnWidth();
    }
    else if (endRow == rowCount - 1 || borderMode == BORDER_USE_EXTERNAL) {
        // an external border is required at the bottom of the table
        bottom_offset = getBottomBorder().getDrawnWidth();
    }
    else {
        // use the internal row width at the bottom
        bottom_offset = getRow(endRow).getBorderWidth();
    }

    // compute half of the border width
    bottom_offset = (JCUnit.Measure) bottom_offset.clone();
    bottom_offset.divide(2.0);

    // compute the coordinates of the top and bottom edges
    y1 = (JCUnit.Measure) y0.clone();
    y1.add(headerH);
    y1.add(top_offset);
    y2 = (JCUnit.Measure) y0.clone();
    y2.add(headerH);
    y2.add(tableH);
    y2.subtract(bottom_offset);

    // draw the table part's backgrounds
    drawTableBackgrounds(frame, startRow, startColumn, endRow, endColumn,
                         x1, y1, x2, y2);

    // if the table has headers, draw them
    if (hasHeaders()) {
        int hrows = header.getRowCount() - 1;

        header.drawTablePart(frame, 0, startColumn, hrows, endColumn,
                             new JCUnit.Measure(), headerH, tableW);
    }

    // draw the borders of the cells in this view
    drawTableBorders(frame, startRow, startColumn, endRow, endColumn, x1, y1);

    // Paste the frames from the cells into place
    x1.add(left_offset);
    y1.add(top_offset);
    pasteCells(frame, startRow, startColumn, endRow, endColumn, x1, y1);
}

/**
 * Creates a Column and adds it to the column list.
 * @param width the width for this column
 * @param ordinal the position of this column
 */
protected void createColumn(JCUnit.Measure width, int ordinal) {

         Column c = new Column(width, ordinal);
        columnList.add(c);

        // set the color on the column if it alternates
        if (tableStyle.getAlternate() != null &&
                !tableStyle.getAlternate().rowOriented) {
            if (ordinal % 2 == 0) {
                c.setBackgroundColor(tableStyle.getAlternate().c1);
            }
            else {
                c.setBackgroundColor(tableStyle.getAlternate().c2);
            }
        }

}

/**
 * Gets the width of the left side border based on consideration both of this
 * table and any header or parent table it may have.
 * @return the width required for the left side border
 */
protected JCUnit.Measure getLeftBorderWidth() {

    JCUnit.Measure  width, tmp_width;

    // Determine the width of this table's left border
    width = this.getLeftBorder().getDrawnWidth();

    // Get the width of the left border at a header table
    if (hasHeaders()) {
        tmp_width = header.getLeftBorder().getDrawnWidth();

        // Take the maximum of the widths
        if (tmp_width.greaterThan(width)) {
            width = tmp_width;
        }
    }

    // Get the width of the left border at a parent table
    if (isHeader) {
        tmp_width = mainTable.getLeftBorder().getDrawnWidth();

        // Take the maximum of the widths
        if (tmp_width.greaterThan(width)) {
            width = tmp_width;
        }
    }

    return(width);
}

/**
 * Gets the width of the right side border based on consideration both of this
 * table and any header or parent table it may have.
 * @return the width required for the right side border
 */
protected JCUnit.Measure getRightBorderWidth() {

    JCUnit.Measure  width, tmp_width;

    // Determine the width of this table's right border
    width = this.getRightBorder().getDrawnWidth();

    // Get the width of the right border at a header table
    if (hasHeaders()) {
        tmp_width = header.getRightBorder().getDrawnWidth();

        // Take the maximum of the widths
        if (tmp_width.greaterThan(width)) {
            width = tmp_width;
        }
    }

    // Get the width of the right border at a parent table
    if (isHeader) {
        tmp_width = mainTable.getRightBorder().getDrawnWidth();

        // Take the maximum of the widths
        if (tmp_width.greaterThan(width)) {
            width = tmp_width;
        }
    }

    return(width);
}

/**
 * Gets the width of the default column border based on consideration both of
 * this table and any header or parent table it may have.
 * @return the width required for the default column border
 */
protected JCUnit.Measure getColumnBorderWidth() {

    JCUnit.Measure  width, tmp_width;

    // Determine the width of this table's column border
    width = this.getColumnBorder().getDrawnWidth();

    // Get the width of the column border at a header table
    if (hasHeaders()) {
        tmp_width = header.getColumnBorder().getDrawnWidth();

        // Take the maximum of the widths
        if (tmp_width.greaterThan(width)) {
            width = tmp_width;
        }
    }

    // Get the width of the column border at a parent table
    if (isHeader) {
        tmp_width = mainTable.getColumnBorder().getDrawnWidth();

        // Take the maximum of the widths
        if (tmp_width.greaterThan(width)) {
            width = tmp_width;
        }
    }

    return(width);
}

/**
 * Computes the heights of the rows of the table, optimally allocating
 * the space required by vertically spanned cells.
 */
private void computeRowHeights()
{
    int             r, c, r1, c1;
    int             cell_count, cell_rows, cell_columns;
    Row             row, row1;
    Cell            cell;
    Column          column;
    List            cell_list;
    Point           origin;
    JCUnit.Measure  cell_height, extra_height, per_row;

    

    // Compute the height of each row ignoring spans
    for (r = 0; r < rowCount; r++) {
        row = getRow(r);
        row.basicHeight();
    }

    

    // Add the extra height required by spanned cells to each row
    for (r = 0; r < rowCount; r++) {
        row = (Row) rowList.get(r);

        

        // determine the number of cell entries in this row
        cell_list = row.getCellList();
        cell_count = cell_list.size();

        // examine all cells in the row
        for (c = 0; c < cell_count; c++) {
            cell = (Cell) cell_list.get(c);

            // ignore cells which are not spanned
            // (NOTE:  spanned cells are necessarily instantiated)
            if (cell == UNINITIALISED_CELL || !cell.isSpanned()) {
                continue;
            }

            // We only care about the root cell of a span
            origin = cell.getSpanOrigin();
            if (origin.x != c || origin.y != r) {
                continue;
            }

            // Mark columns for the cells which cross column boundaries
            cell_columns = cell.getSpanColumns();
            if (cell_columns > 1) {
                column = getColumn(c);
                for (c1 = 1; c1 < cell_columns; c1++) {
                    // Note the span leaves the old column and enters the next
                    column.addDepartingSpan();
                    column = getColumn(c + c1);
                    column.addArrivingSpan();
                }
            }

            // henceforth ignore cells which are only one row high
            cell_rows = cell.getSpanRows();
            if (cell_rows == 1) {
                continue;
            }

            // NOTE:  At this point we _only_ have root cells which span
            //        over multiple rows
            
            

            // Determine the amount of space required and the extra available
            cell_height = cell.fixHeight();
            extra_height = new JCUnit.Measure();

            int expandable_rows = 0;

            for (r1 = 0; r1 < cell_rows; r1++) {

                row1 = (Row) rowList.get(r1 + r);
                cell_height.subtract(row1.getHeight());
                extra_height.add(row1.getExtraHeight());

                
                

                // Count the number of rows to which extra height can be added
                if (!row1.hasFixedHeight()) {
                    expandable_rows++;
                }
            }

            

            // If there is already enough space, continue
            if (cell_height.lessThan(extra_height)) {
                
                continue;
            }

            // if there are no non-fixed rows in the range, give up
            if (expandable_rows == 0) {
                
                continue;
            }

            // Compute the amount of extra space to allot per row
            per_row = (JCUnit.Measure) cell_height.clone();
            per_row.divide(expandable_rows);

            JCUnit.Measure zero_measure = new JCUnit.Measure();

            // NOTE:  The amount of extra space available at following rows
            //        will _necessarily_ be a non-increasing amount (since
            //        space is allocated from the top down).  This allows us
            //        to considerably simplify the addition of new space

            for (r1 = 0; r1 < cell_rows; r1++) {

                row1 = (Row) rowList.get(r1 + r);

                // Ignore rows which have fixed height
                if (row1.hasFixedHeight()) {
                    continue;
                }

                extra_height = row1.getExtraHeight();
                expandable_rows--;

                // if there is already more space here than needed, use it all
                if (per_row.lessThan(extra_height)) {
                    cell_height.subtract(extra_height);

                    // If there is space left to assign
                    if (cell_height.greaterThan(zero_measure)) {
                        // recompute the space required per row
                        per_row.assign(cell_height);
                        per_row.divide(expandable_rows);
                    }
                }

                // Otherwise assign the required amount
                else {
                    row1.setExtraHeight(per_row);
                    cell_height.subtract(per_row);
                }
            }
        }
    }

    
    
}

/**
 * Gets the cell at specified table coordinates, returning null if the cell
 * does not exist (instead of creating the cell as <code>getCell()</code> does).
 * @param r the row the cell is in
 * @param c the column the cell is in
 * @return the cell at the specified location
 */
private Cell getCellRaw(int r, int c) {

    Row     row;
    List    cell_list;

    // sanity-check the location
    if (r < 0 || r >= rowCount || c < 0 || c >= columnCount) {
        // return null for out-of-bounds
        return(null);
    }

    // Get the row the cell is in
    row = (Row) rowList.get(r);

    // if the row is uninitialised, so is the cell
    if (row == UNINITIALISED_ROW) {
        return(UNINITIALISED_CELL);
    }

    // Get the row's cell list
    cell_list = row.getCellList();

    // if the cell is not instantiated in this row
    if (c >= cell_list.size()) {
        return(UNINITIALISED_CELL);
    }

    // Get the cell from the row's cell list
    return((Cell) cell_list.get(c));

}

/**
 * Gets the bottom border style of the specified cell position in the table.
 * @param r the row of the cell position whose border is required
 * @param c the column of the cell position whose border is required
 * @return the bottom border style of this position (null for spans)
 */
private JCDrawStyle getRawBottomBorder(int r, int c) {

    Cell    cell;

    // we return null for border requests outside table bounds
    // (this permits us to request the borders surrounding a cell without
    // worrying about detecting the edge of the table)
    if (r < 0 || r >= rowCount || c < 0 || c >= columnCount) {
        return(null);
    }

    // Get the cell that directly occupies the given location
    cell = getCellRaw(r, c);

    // if the cell is not instantiated, use the column border
    if (cell == UNINITIALISED_CELL) {
        return(getRow(r).getBorderStyle());
    }

    // Get the style applied to this cell
    return(cell.getBottomEdgeStyle());
}

/**
 * Gets the right border style of the specified cell position in the table.
 * @param r the row of the cell position whose border is required
 * @param c the column of the cell position whose border is required
 * @return the right border style of this position (null for spans)
 */
private JCDrawStyle getRawRightBorder(int r, int c) {

    Cell    cell;

    // we return null for border requests outside table bounds
    // (this permits us to request the borders surrounding a cell without
    // worrying about detecting the edge of the table)
    if (r < 0 || r >= rowCount || c < 0 || c >= columnCount) {
        return(null);
    }

    // Get the cell that directly occupies the given location
    cell = getCellRaw(r, c);

    // if the cell is not instantiated, use the column border
    if (cell == UNINITIALISED_CELL) {
        return(getColumn(c).getBorderStyle());
    }

    // Get the style applied to this cell
    return(cell.getRightEdgeStyle());
}

/**
 * Draws a piece of the background of the table.
 * @param target the frame to draw the table in
 * @param color the color to use for the background
 * @param x0 <i>x</i>-coordinate of the upper-left corner
 * @param y0 <i>y</i>-coordinate of the upper-left corner
 * @param x1 <i>x</i>-coordinate of the lower-right corner
 * @param y1 <i>y</i>-coordinate of the lower-right corner
 */
private void drawBackground(JCFrame target, Color color,
                            JCUnit.Measure x0, JCUnit.Measure y0,
                            JCUnit.Measure x1, JCUnit.Measure y1) {

    List            render_list;
    BoxRender       render;
    JCDrawStyle     fill_style;
    JCUnit.Point    ul, lr;

    // look for a style to use to fill the area
    fill_style = JCDrawStyle.stringToStyle("__tableFillStyle"
                                                + color.toString());

    // if the style doesn't exist yet, create it.
    if (fill_style == null) {
        fill_style = new JCDrawStyle("__tableFillStyle" + color.toString());
        fill_style.setFillForegroundColor(color);
        fill_style.makeImmutable();
    }

    // if the units agree, use them to create a point
    if (x0.units == y0.units) {
        ul = new JCUnit.Point(x0.units, x0.distance, y0.distance);
    }
    // Otherwise convert y0 to units of x0
    else {
        ul = new JCUnit.Point(x0.units, x0.distance, y0.getAs(x0.units));
    }

    // if the units agree, use them to create a point
    if (x1.units == y1.units) {
        lr = new JCUnit.Point(x1.units, x1.distance, y1.distance);
    }
    // Otherwise convert y1 to units of x1
    else {
        lr = new JCUnit.Point(x1.units, x1.distance, y1.getAs(x1.units));
    }

    

    // Add the background area to the frame's render list
    render_list = target.getRenderList();
    render = new BoxRender(ul, lr, fill_style);
    render.setFilled(true);
    render_list.add(render);
}

/**
 * Draws the backgrounds of the cells in a given table part.
 * <br><b>NOTE:</b> Only cells which locally specify a color are drawn.
 * @param target the frame into which the table is being drawn
 * @param startRow the first row of the table in the part being printed
 * @param startColumn the first column of the table in the part being printed
 * @param endRow the last row of the table in the part being printed
 * @param endColumn the last column of the table in the part being printed
 * @param x0 the <i>x</i>-coordinate of the table with respect to the frame
 * @param y0 the <i>y</i>-coordinate of the table with respect to the frame
 */
private void drawCellBackgrounds(JCFrame target, int startRow, int startColumn,
                                 int endRow, int endColumn,
                                 JCUnit.Measure x0, JCUnit.Measure y0) {

    int                 r, c;
    Cell                cell;
    Color               color;
    JCUnit.Measure[]    column_widths;
    JCUnit.Measure      x1, y1, x2, y2, top_offset;

    // sanity-check the range
    if (startRow < 0 || endRow >= rowCount || startRow > endRow) {
        throw new IllegalArgumentException("Invalid row range");
    }
    if (startColumn < 0 || endColumn >= columnCount || startColumn > endColumn) {
        throw new IllegalArgumentException("Invalid column range");
    }

    

    // Make local copies of the location coordinates so we can update them
    x1 = new JCUnit.Measure();
    y1 = (JCUnit.Measure) y0.clone();
    x2 = new JCUnit.Measure();
    y2 = new JCUnit.Measure();

    // Create an array to save the widths of columns
    column_widths = new JCUnit.Measure[endColumn - startColumn + 1];

    // compute the offset of the top row from the start of the space
    if (hasHeaders()) {
        // if there is a header border we need to paint under half of it
        top_offset = this.getHeaderBorder().getDrawnWidth();
        top_offset.divide(2.0);
    }
    else {
        top_offset = new JCUnit.Measure();
    }

    // output each row of the table in turn (note the range is inclusive)
    for (r = startRow; r <= endRow; r++) {

        

        // Compute the bottom of the row
        y2.assign(y1);
        y2.add(getRow(r).getDrawnHeight((r == startRow), (r == endRow)));

        // Adjust the start of the first row of the table
        if (r == startRow) {
            y1.subtract(top_offset);
        }

        x1.assign(x0);

        // Loop through the cells in this row (note the range is inclusive)
        for (c = startColumn; c <= endColumn; c++) {

            

            // if we are in the first row, save the column widths
            if (r == startRow) {
                column_widths[c - startColumn] =
                                getColumn(c).getDrawnWidth((c == startColumn),
                                                            (c == endColumn));
                
            }

            // Compute the end of this column
            x2.assign(x1);
            x2.add(column_widths[c - startColumn]);

            

            // get the cell at this position in the table
            cell = getCellRaw(r, c);

            // we only care about initialised cells; only they have backgrounds
            if (cell != UNINITIALISED_CELL) {

                // if the cell is spanned get the root cell
                if (cell.isSpanned()) {
                    cell = getCell(r, c);
                }

                // draw the backgrounds of all spanned cells, and any other
                // cells which specify their own background colour
                if (cell.isSpanned() || !cell.isBackgroundColorInherited()) {

                    // get the background color from the cell
                    color = cell.getBackgroundColor();

                    // if the cell does not inherit any colour, try to get the
                    // colour from the frame the cell is being drawn into.
                    if (color == null) {
                        color = target.getColor();

                        // NOTE: it is currently not possible to trace upward
                        //       through the frame heirarchy to the page to
                        //       find a background colour.  It is also possible
                        //       for the page to have no background colour
                        //       specified, so even that would not provide a
                        //       complete solution.
                    }

                    // Draw the background if possible
                    if (color != null) {
                        drawBackground(target, color, x1, y1, x2, y2);
                    }
                }
            }

            // Advance to the next cell in this row
            x1.assign(x2);
        }

        // Procede to the next row of the table
        y1.assign(y2);
    }
}

/**
 * Draws the backgrounds of the rows in a given table part.
 * <br><B>NOTE:</B>  only rows which locally specify a color are drawn.
 * @param target the frame into which the table is being drawn
 * @param startRow the first row of the table in the part being printed
 * @param endRow the last row of the table in the part being printed
 * @param x0 the <i>x</i>-coordinate of the table part with respect to the frame
 * @param y0 the <i>y</i>-coordinate of the table part with respect to the frame
 * @param x1 the <i>x</i>-coordinate of the end of the drawn row space
 * @param endColumn the last column of the table in the part being printed
 * @param startColumn the first column of the table in the part being printed
 */
private void drawRowBackgrounds(JCFrame target, int startRow, int endRow,
                                int startColumn, int endColumn,
                                JCUnit.Measure x0, JCUnit.Measure y0,
                                JCUnit.Measure x1) {

    int                 r;
    Row                 row;
    Color               color;
    JCUnit.Measure      x2, x3, y1, y2, top_offset;

    // sanity-check the range
    if (startRow < 0 || endRow >= rowCount || startRow > endRow) {
        throw new IllegalArgumentException("Invalid row range");
    }

    // Make local copies of the location coordinates so we can update them
    y1 = (JCUnit.Measure) y0.clone();
    y2 = new JCUnit.Measure();

    // compute the offset of the top row from the start of the space
    if (hasHeaders()) {
        // if there is a header border we need to paint under half of it
        top_offset = this.getHeaderBorder().getDrawnWidth();
        top_offset.divide(2.0);
    }
    else {
        top_offset = new JCUnit.Measure();
    }

    // draw each row of the table in turn (note the range is inclusive)
    for (r = startRow; r <= endRow; r++) {

        
        // get the next row and compute the bottom of it
        row = getRow(r);
        y2.assign(y1);
        y2.add(row.getDrawnHeight((r == startRow), (r == endRow)));

        // Adjust the start of the first row of the table
        if (r == startRow) {
            y1.subtract(top_offset);
        }

        // if the background isn't inherited, draw it
        if (!row.isBackgroundColorInherited()) {

            // get the background color from the row
            color = row.getBackgroundColor();

            // draw the background
            // collect the cells outside of spans and draw them
            x2 = (JCUnit.Measure) x0.clone();
            x3 = (JCUnit.Measure) x2.clone();
            int startDraw = startColumn;
            int endDraw = -1;
            for (int c = startColumn; c <= endColumn; c++) {
                Cell cell = getCellRaw(r, c);
                // if we hit a span, draw the backround up to that point
                if (cell != UNINITIALISED_CELL && cell.isSpanned()) {
                    
                    // if we have a collection of columns to draw ...
                    if (endDraw >= 0) {
                        drawBackground(target, color, x2, y1, x3, y2);
                        endDraw = -1;
                        //reset start of draw rectangle
                        x2 = (JCUnit.Measure) x3.clone();
                    }
                    // move right corner to end of this column
                    x3.add(getColumn(c).getDrawnWidth((c == startColumn),
                                                        (c == endColumn)));
                } else {
                    // no collection of columns?  shore up left hand
                    // side of draw area
                    if (endDraw < 0) {
                        x2 = (JCUnit.Measure) x3.clone();
                    }
                    // move right corner to end of this column
                    x3.add(getColumn(c).getDrawnWidth((c == startColumn),
                                                            (c == endColumn)));
                    // note this column for the collection
                    endDraw = c;
                }
            }
            // draw the reaminder of the background, if there is any to draw
            if (endDraw >= 0) {
                drawBackground(target, color, x2, y1, x3, y2);
            }
        }

        // Procede to the next row of the table
        y1.assign(y2);
    }
}

/**
 * Draws the backgrounds of the columns in a given table part.
 * <br><B>NOTE:</B>  only columns which locally specify a color are drawn
 * @param target the frame into which the table is being drawn
 * @param startColumn the first column of the table in the part being printed
 * @param endColumn the last column of the table in the part being printed
 * @param x0 the <i>x</i>-coordinate of the table part with respect to the frame
 * @param y0 the <i>y</i>-coordinate of the table part with respect to the frame
 * @param y1 the <i>y</i>-coordinate of the end of the drawn column space
 * @param startRow the first row of the table to draw
 * @param endRow the last row of the table to draw
 */
private void drawColumnBackgrounds(JCFrame target,
                                   int startColumn, int endColumn,
                                   int startRow, int endRow,
                                   JCUnit.Measure x0, JCUnit.Measure y0,
                                   JCUnit.Measure y1) {

    Column              column;
    Color                color;
    JCUnit.Measure      x1, x2, y2, y3;

    // sanity-check the range
    if (startColumn < 0 || endColumn >= columnCount || startColumn > endColumn) {
        throw new IllegalArgumentException("Invalid column range");
    }

    // Make local copies of the location coordinates so we can update them
    x1 = (JCUnit.Measure) x0.clone();
    x2 = new JCUnit.Measure();

    // if there is a header border we need to paint under half of it
    if (hasHeaders()) {
        JCUnit.Measure  y_tmp = y0;

        y0 = this.getHeaderBorder().getDrawnWidth();
        y0.divide(-2.0);    // want to *subtract* this from y_tmp, hence "-"
        y0.add(y_tmp);
    }

    // draw each column of the table in turn (note the range is inclusive)
    for (int c = startColumn; c <= endColumn; c++) {

        // get the next column and compute the right edge of it
        column = getColumn(c);
        x2.assign(x1);
        x2.add(column.getDrawnWidth((c == startColumn), (c == endColumn)));

        // if the background isn't inherited, draw it
        boolean inherited = column.isBackgroundColorInherited();
        if (inherited && isHeader()) {
            inherited = mainTable.getColumn(c).isBackgroundColorInherited();
        }

        if (!inherited) {
            // draw the background
            color = column.getBackgroundColor();

            // collect the cells outside of spans and draw them
            y2 = (JCUnit.Measure) y0.clone();
            y3 = (JCUnit.Measure) y2.clone();
            int startDraw = startRow;
            int endDraw = -1;
            for (int r = startRow; r <= endRow; r++) {
                Cell cell = getCellRaw(r, c);
                // if we hit a span, draw the backround up to that point
                if (cell != UNINITIALISED_CELL && cell.isSpanned()) {

                    // if we have a collection of rows to draw ...
                    if (endDraw >= 0) {
                        drawBackground(target, color, x1, y2, x2, y3);
                        endDraw = -1;
                        //reset start of draw rectangle
                        y2 = (JCUnit.Measure) y3.clone();
                    }
                    // move bottom corner to end of this row
                    y3.add(getRow(r).getDrawnHeight((r == startRow),
                                                        (r == endRow)));
                } else {
                    // no collection of rows?  shore up top
                    // side of draw area
                    if (endDraw < 0) {
                        y2 = (JCUnit.Measure) y3.clone();
                    }
                    // move right corner to end of this row
                    y3.add(getRow(r).getDrawnHeight((r == startRow),
                                                            (r == endRow)));
                    // note this row for the collection
                    endDraw = r;
                }
            }
            // draw the reaminder of the background, if there is any to draw
            if (endDraw >= 0) {
                drawBackground(target, color, x1, y2, x2, y3);
            }
        }

        // Advance to the next column
        x1.assign(x2);
    }
}

/**
 * Draws the backgrounds of the table components.
 * @param target the frame the table is being drawn in
 * @param startRow the first row of the table to draw
 * @param startColumn the first column of the table to draw
 * @param endRow the last row of the table to draw
 * @param endColumn the last column of the table to draw
 * @param x0 the upper-left corner of the area to draw the table into
 * @param y0 the upper-left corner of the area to draw the table into
 * @param x1 the lower-right corner of the area to draw the table into
 * @param y1 the lower-right corner of the area to draw the table into
 */
private void drawTableBackgrounds(JCFrame target, int startRow, int startColumn,
                                  int endRow, int endColumn,
                                  JCUnit.Measure x0, JCUnit.Measure y0,
                                  JCUnit.Measure x1, JCUnit.Measure y1) {

    Color            color;
    JCUnit.Measure   top_width;

    // NOTE:  The internal printable area is given by (x0, y0) (x1, y1)

    // If the table has headers, get the width of the header border
    if (hasHeaders()) {
        top_width = getHeaderBorder().getDrawnWidth();
        top_width.divide(2.0);
    }
    else {
        top_width = new JCUnit.Measure();
    }

    // Get the background color of the table
    color = getBackgroundColor();

    // if there is a color draw the background of the table
    if (color != null) {
        JCUnit.Measure  y_tmp = (JCUnit.Measure) y0.clone();

        y_tmp.subtract(top_width);
        drawBackground(target, color, x0, y_tmp, x1, y1);
    }

    // Draw the columns then rows or vice versa depending on relative priority
    if (dominance == ROW_DOMINANCE) {
        drawColumnBackgrounds(target, startColumn, endColumn, 
                                        startRow, endRow, x0, y0, y1);
        drawRowBackgrounds(target, startRow, endRow, 
                                        startColumn, endColumn, x0, y0, x1);
    }
    else {
        drawRowBackgrounds(target, startRow, endRow, 
                                        startColumn, endColumn, x0, y0, x1);
        drawColumnBackgrounds(target, startColumn, endColumn, 
                                        startRow, endRow, x0, y0, y1);
    }

    // draw the cell backgrounds that override previous settings
    drawCellBackgrounds(target, startRow, startColumn, endRow, endColumn,
                        x0, y0);
}

/**
 * This method chooses which of a pair of border styles is drawn on top.
 * @param style1 a draw style describing a table border
 * @param style2 a draw style describing a table border
 * @param border_width a measure of border width in which the selected width is set
 * @return the style which is given higher priority
 */
private JCDrawStyle selectDominantStyle(JCDrawStyle style1, JCDrawStyle style2,
                                        JCUnit.Measure border_width) {

    JCUnit.Measure  width1, width2;

    // Determine the widths of the borders
    if (style1 != null) {
        width1 = style1.getDrawnWidth();
    }
    else {
        // null (unset) borders have zero width
        width1 = new JCUnit.Measure();
    }

    if (style2 != null) {
        width2 = style2.getDrawnWidth();
    }
    else {
        width2 = new JCUnit.Measure();
    }

    // Select the wider of the two lines
    if (width1.greaterThan(width2)) {
        border_width.assign(width1);
        return(style1);
    }
    else if (width1.lessThan(width2)) {
        border_width.assign(width2);
        return(style2);
    }

    // If the borders are equal width use other characteristics
    else {

        // if non-zero width, we arbitrarily favour the first style
        if (!width1.isZero()) {
            border_width.assign(width1);
            return(style1);
        }

        // otherwise both are zero width, so we try to find a non-null border
        else {
            border_width.assign(width1);

            // favour the first border if possible (for consistency)
            if (style1 != null && (style2 == null ||
                       style1.getLineType() != JCDrawStyle.LINE_TYPE_NONE)) {
                return(style1);
            }
            // otherwise the second can't be worse
            else {
                return(style2);
            }
        }
    }
}

/**
 * Draws a piece of the border (internal or external) of the table.
 * @param target the frame to draw the table in
 * @param border the draw style which defines the border
 * @param x0 <i>x</i>-coordinate of the first point of the line
 * @param y0 <i>y</i>-coordinate of the first point of the line
 * @param x1 <i>x</i>-coordinate of the second point of the line
 * @param y1 <i>y</i>-coordinate of the second point of the line
 */
private void drawBorder(JCFrame target, JCDrawStyle border,
                        JCUnit.Measure x0, JCUnit.Measure y0,
                        JCUnit.Measure x1, JCUnit.Measure y1) {

    List            render_list;
    JCUnit.Point[]  point_list = new JCUnit.Point[2];

    // draw nothing if there is no color
    if (border.getForegroundColor() == null) {
        return;
    }

    // if the units agree, use them to create a point
    if (x0.units == y0.units) {
        point_list[0] = new JCUnit.Point(x0.units, x0.distance, y0.distance);
    }
    // Otherwise convert y0 to units of x0
    else {
        point_list[0] = new JCUnit.Point(x0.units, x0.distance,
                                         y0.getAs(x0.units));
    }

    // if the units agree, use them to create a point
    if (x1.units == y1.units) {
        point_list[1] = new JCUnit.Point(x1.units, x1.distance, y1.distance);
    }
    // Otherwise convert y1 to units of x1
    else {
        point_list[1] = new JCUnit.Point(x1.units, x1.distance,
                                         y1.getAs(x1.units));
    }

    

    // Add the border line to the frame's render list
    render_list = target.getRenderList();
    render_list.add(new LineRender(point_list, border));
}

//
// This is the beginning of breaking the drawTablePart() method into a
// more object-oriented approach.  In a future release, this can be
// revised slightly and merged into the Row object to do its own drawing.
// (We would also change it so it is demand-drawn into a Graphics - by
// passing a Graphics2D object to it - instead of drawing into a JCFrame.)
//
private void drawRowBorder(JCFrame frame, int row, int startRow,
                           int startColumn, int endRow, int endColumn,
                           JCUnit.Measure x, JCUnit.Measure y,
                           JCUnit.Measure[] column_widths) {

    JCDrawStyle        above, below;
    JCDrawStyle        currBorder, oldBorder;
    JCUnit.Measure    left_offset = new JCUnit.Measure();
    JCUnit.Measure    right_offset = new JCUnit.Measure();
    JCUnit.Measure    x_start, x_end, x_tmp = new JCUnit.Measure();
    JCUnit.Measure    tmp_width;

    // if we have an external border on our left
    if (startColumn == 0 || borderMode == BORDER_USE_EXTERNAL) {

        // compute the width of the table edge border
        left_offset = getLeftBorder().getDrawnWidth();
    }

    // otherwise we have an internal border on the left
    else {

        // Get the border width on the left side of the row
        left_offset = getIntersectionBorderOffset(row, startColumn - 1, true);
    }

    // if we have an external border on our right
    if (endColumn == columnCount - 1 || borderMode == BORDER_USE_EXTERNAL) {

        // compute the width of the table edge border
        right_offset = getRightBorder().getDrawnWidth();
    }

    // otherwise we have an internal border on the right
    else {

        // Get the border width on the right side of the row
        right_offset = getIntersectionBorderOffset(row, endColumn, true);
    }

    JCUnit.Measure single_measure = new JCUnit.Measure(JCUnit.POINTS, 1.0);
    JCUnit.Measure zero_measure = new JCUnit.Measure(JCUnit.POINTS, 0.0);

    // if we're dealing with an offset of less than a point, don't bother
    // it only causes rounding errors -- temporary fix for much bigger problem
    if (left_offset.lessThan(single_measure))
        left_offset.assign(zero_measure);
    if (right_offset.lessThan(single_measure))
        right_offset.assign(zero_measure);

    // Compute the external part of the edge border widths
    left_offset.divide(2.0);
    right_offset.divide(2.0);

    // Compute the initial state for traversing the row
    x_start = (JCUnit.Measure) x.clone();
    x_start.subtract(left_offset);
    oldBorder = getEffectiveRowBorder(row, startColumn, startRow, endRow);
    
    
    x_end = (JCUnit.Measure) x.clone();

    // Loop through the cells in the row
    for (int c = startColumn; c <= endColumn; c++) {

        // Get the bottom border of this cell
        currBorder = getEffectiveRowBorder(row, c, startRow, endRow);

        
        

        // If the border has changed
        if ((currBorder != null && !currBorder.equals(oldBorder)) ||
            (currBorder == null && oldBorder != null)) {

            // Get the width of the border we cross
            tmp_width = getIntersectionBorderOffset(row, c, true);
            // if we're dealing with an offset of less than a point, 
            // don't bother it only causes rounding errors -- temporary 
            // fix for much bigger problem
            if (tmp_width.lessThan(single_measure))
                tmp_width.assign(zero_measure);
            tmp_width.divide(2.0);

            // if the previous style was non-null, draw it
            if (oldBorder != null) {

                // Add the width to the end of the line to draw
                x_tmp.assign(x_end);
                x_tmp.add(tmp_width);

                // output the last contiguous piece of border
                drawBorder(frame, oldBorder, x_start, y, x_tmp, y);
            }

            // Start accumulating the next piece of border
            oldBorder = currBorder;
            x_start.assign(x_end);
            x_start.subtract(tmp_width);
        }

        // update the end position to the right side of this column
        x_end.add(column_widths[c]);
    }

    // Output the last piece of border (if it's not null)
    if (oldBorder != null) {
        x_end.add(right_offset);
        drawBorder(frame, oldBorder, x_start, y, x_end, y);
    }
}

private JCDrawStyle getEffectiveRowBorder(int row, int column,
                                          int firstRow, int lastRow) {

    JCDrawStyle border = null;

    // in the internal portion of the table, return the raw border
    if (row >= firstRow && row < lastRow) {
        return(getRawBottomBorder(row, column));
    }

    // return the raw border on the edges if not table edges or external mode
    if (borderMode != BORDER_USE_EXTERNAL && row >= 0 && row < rowCount) {

        return(getRawBottomBorder(row, column));
    }

    // otherwise we have either external borders on table edges or on internal
    // table divisions (table part edges)

    // Check for top edge:
    if (row < firstRow) {

        // We use the table's top edge border or header border as required
        if (hasHeaders()) {
            border = getHeaderBorder();
        }
        else {
            border = getTopBorder();
        }

        // return it for the top edge or below headers
        if (row == -1 || hasHeaders()) {
            return(border);
        }
    }

    // Check for bottom edge:
    else if (row + 1 >= lastRow) {

        // We use the table's bottom edge border or header border
        if (isHeader()) {
            border = getHeaderBorder();
        }
        else {
            border = getBottomBorder();
        }

        // return it for the right edge
        if (row + 1 == lastRow) {
            return(border);
        }
    }
    else {
        throw new IllegalArgumentException("Invalid row address to get border");
    }

    // internal table part edge
    // check table part edges for null and prefer that to the table edge border
    if (getRawBottomBorder(row, column) == null) {
        return(null);
    }
    else {
        return(border);
    }
}

//
// This is the beginning of breaking the drawTablePart() method into a
// more object-oriented approach.  In a future release, this can be
// revised slightly and merged into the Column object to do its own drawing.
// (We would also change it so it is demand-drawn into a Graphics - by
// passing a Graphics2D object to it - instead of drawing into a JCFrame.)
//
private void drawColumnBorder(JCFrame frame, int column, int startRow,
                              int startColumn, int endRow, int endColumn,
                              JCUnit.Measure x, JCUnit.Measure y,
                              JCUnit.Measure[] row_heights) {

    JCDrawStyle        left, right;
    JCDrawStyle        currBorder, oldBorder;
    JCUnit.Measure    top_offset = new JCUnit.Measure();
    JCUnit.Measure    bottom_offset = new JCUnit.Measure();
    JCUnit.Measure    y_start, y_end, y_tmp = new JCUnit.Measure();
    JCUnit.Measure    tmp_width;

    // if we have an external or header border on our top
    if (startRow == 0 || borderMode == BORDER_USE_EXTERNAL || hasHeaders()) {

        // compute the width of the table edge border
        if (hasHeaders()) {
            top_offset = getHeaderBorder().getDrawnWidth();
        }
        else {
            top_offset = getTopBorder().getDrawnWidth();
        }
    }

    // otherwise we have an internal border on the top
    else {

        // Get the border width on the top of the column
        top_offset = getIntersectionBorderOffset(startRow - 1, column, false);
    }

    // if we have an external border on our bottom
    if (endRow == rowCount - 1 || borderMode == BORDER_USE_EXTERNAL) {

        // compute the width of the table bottom edge border
        if (isHeader()) {
            bottom_offset = getHeaderBorder().getDrawnWidth();
        }
        else {
            bottom_offset = getBottomBorder().getDrawnWidth();
        }
    }

    // otherwise we have an internal border on the bottom
    else {

        // Get the border width on the bottom of the column
        bottom_offset = getIntersectionBorderOffset(endRow, column, false);
    }

    JCUnit.Measure single_measure = new JCUnit.Measure(JCUnit.POINTS, 1.0);
    JCUnit.Measure zero_measure = new JCUnit.Measure(JCUnit.POINTS, 0.0);

    // if we're dealing with an offset of less than a point, 
    // don't bother it only causes rounding errors -- temporary 
    // fix for much bigger problem
    if (top_offset.lessThan(single_measure))
        top_offset.assign(zero_measure);
    if (bottom_offset.lessThan(single_measure))
        bottom_offset.assign(zero_measure);

    // Compute the external part of the edge border widths
    top_offset.divide(2.0);
    bottom_offset.divide(2.0);

    // Compute the initial state for traversing the column
    y_start = (JCUnit.Measure) y.clone();
    y_start.subtract(top_offset);
    oldBorder = getEffectiveColumnBorder(startRow, column,
                                         startColumn, endColumn);
    y_end = (JCUnit.Measure) y.clone();

    // Loop through the cells in the row
    for (int r = startRow; r <= endRow; r++) {

        // Get the bottom border of this cell
        currBorder = getEffectiveColumnBorder(r, column,
                                              startColumn, endColumn);

        // If the border has changed
        if ((currBorder != null && !currBorder.equals(oldBorder)) ||
            (currBorder == null && oldBorder != null)) {

            // Get the width of the border we cross
            tmp_width = getIntersectionBorderOffset(r, column, false);
            // if we're dealing with an offset of less than a point, 
            // don't bother it only causes rounding errors -- temporary 
            // fix for much bigger problem
            if (tmp_width.lessThan(single_measure))
                tmp_width.assign(zero_measure);
            tmp_width.divide(2.0);

            // if the previous style was non-null, draw it
            if (oldBorder != null) {

                // Add the width to the end of the line to draw
                y_tmp.assign(y_end);
                y_tmp.add(tmp_width);

                // output the last contiguous piece of border
                drawBorder(frame, oldBorder, x, y_start, x, y_tmp);
            }

            // Start accumulating the next piece of border
            oldBorder = currBorder;
            y_start.assign(y_end);
            y_start.subtract(tmp_width);
        }

        // update the end position to the bottom side of this row
        y_end.add(row_heights[r]);
    }

    // Output the last piece of border (if it's not null)
    if (oldBorder != null) {
        y_end.add(bottom_offset);
        drawBorder(frame, oldBorder, x, y_start, x, y_end);
    }
}

private JCDrawStyle getEffectiveColumnBorder(int row, int column,
                                             int firstColumn, int lastColumn) {

    JCDrawStyle border = null;

    // in the internal portion of the table, return the raw border
    if (column >= firstColumn && column < lastColumn) {
        return(getRawRightBorder(row, column));
    }

    // return the raw border on the edges if not table edges or external mode
    if (borderMode != BORDER_USE_EXTERNAL &&
        column >= 0 && column < columnCount) {

        return(getRawRightBorder(row, column));
    }

    // otherwise we have either external borders on table edges or on internal
    // table divisions (table part edges)

    // Check for left side edge:
    if (column < firstColumn) {

        // We use the table's left edge border
        border = getLeftBorder();

        // return it for the left edge
        if (column == -1) {
            return(border);
        }
    }

    // Check for right side edge:
    else if (column + 1 >= lastColumn) {

        // We use the table's right edge border
        border = getRightBorder();

        // return it for the right edge
        if (column + 1 == lastColumn) {
            return(border);
        }
    }
    else {
        throw new IllegalArgumentException("Invalid column address to get border");
    }

    // internal table part edge
    // check table part edges for null and prefer that to the table edge border
    if (getRawRightBorder(row, column) == null) {
        return(null);
    }
    else {
        return(border);
    }
}

private JCUnit.Measure getIntersectionBorderOffset(int iRow, int iColumn,
                                                   boolean rowEdge) {

    JCUnit.Measure    offset, tmp_width;
    JCDrawStyle        inside = null, outside = null;

    // Get the near border on the inside (cell side) of the required edge
    if (rowEdge) {
        if (iRow >= 0) {
            inside = getRawRightBorder(iRow, iColumn);
        }
    }
    else {
        if (iColumn >= 0) {
            inside = getRawBottomBorder(iRow, iColumn);
        }
    }

    // Compute the width of the border if one is set
    if (inside != null) {
        offset = inside.getDrawnWidth();
    }
    else {
        offset = new JCUnit.Measure();
    }

    // Get the far border on the outside of the required edge
    if (rowEdge) {
        if (iRow + 1 < rowCount) {
            outside = getRawRightBorder(iRow + 1, iColumn);
        }
    }
    else {
        if (iColumn + 1 < columnCount) {
            outside = getRawBottomBorder(iRow, iColumn + 1);
        }
    }

    // Find the width of the wider border
    if (outside != null) {
        tmp_width = outside.getDrawnWidth();
        if (tmp_width.greaterThan(offset)) {
            offset.assign(tmp_width);
        }
    }

    // Return the width of the wider intersecting border
    return(offset);
}

/**
 * Draws the borders for part of the table.
 * @param frame the frame the table is being drawn into
 * @param startRow the first row of table being drawn
 * @param startColumn the first column of table being drawn
 * @param endRow the last row of table being drawn
 * @param endColumn the last column of table being drawn
 * @param x0 the origin of the table view in the given frame
 * @param y0 the origin of the table view in the given frame
 */
private void drawTableBorders(JCFrame frame, int startRow, int startColumn,
                              int endRow, int endColumn,
                              JCUnit.Measure x0, JCUnit.Measure y0) {

    int                 r, c, firstRow, lastRow, firstColumn, lastColumn;
    boolean                external = (borderMode == BORDER_USE_EXTERNAL);
    JCUnit.Measure        x1, y1;
    JCUnit.Measure[]    column_widths, row_heights;

    // Create arrays to save the widths of columns and heights of rows
    column_widths = new JCUnit.Measure[columnCount];
    row_heights = new JCUnit.Measure[rowCount];

    // store the coordinates of the table origin
    x1 = (JCUnit.Measure) x0.clone();
    y1 = (JCUnit.Measure) y0.clone();

    // Procede through each row in this view of the table, storing the heights
    for (r = startRow; r <= endRow; r++) {

        row_heights[r] =
                    getRow(r).getDrawnHeight((r == startRow), (r == endRow));
    }

    // loop through the columns of this view in order, storing the widths
    for (c = startColumn; c <= endColumn; c++) {

        column_widths[c] =
            getColumn(c).getDrawnWidth((c == startColumn), (c == endColumn));
    }

    // If we're using internal borders on the perimeter of the table
    if (!external) {

        // Assume standard values initially
        firstRow = startRow;
        lastRow = endRow;
        firstColumn = startColumn;
        lastColumn = endColumn;

        // Back up one column to draw the left edge border
        if (startColumn > 0) {
            firstColumn = startColumn - 1;
            column_widths[firstColumn] = new JCUnit.Measure();
        }

        // Don't draw the right edge of the overall table
        if (endColumn + 1 == columnCount) {
            lastColumn = endColumn - 1;
        }

        // Back up one row to draw the top edge border (if no headers)
        if (startRow > 0 && !hasHeaders()) {
            firstRow = startRow - 1;
            row_heights[firstRow] = new JCUnit.Measure();
        }

        // Don't draw the bottom edge of the overall table
        if (endRow + 1 == rowCount) {
            lastRow = endRow - 1;
        }
    }

    // otherwise the outside of the table uses external borders, so _don't_
    // draw the last row and column borders
    else {
        firstColumn = startColumn;
        firstRow = startRow;
        lastColumn = endColumn - 1;
        lastRow = endRow - 1;
    }

    // if rows are dominant over columns, draw the columns first, then rows
    if (dominance == ROW_DOMINANCE) {

        // Loop through the columns, drawing their borders
        for (c = firstColumn; c <= lastColumn; c++) {

            // compute the right side of the column
            x1.add(column_widths[c]);

            drawColumnBorder(frame, c, startRow, startColumn, endRow,
                             endColumn, x1, y0, row_heights);
        }

        // Loop through the rows, drawing their borders
        for (r = firstRow; r <= lastRow; r++) {

            // compute the bottom of the row
            y1.add(row_heights[r]);

            drawRowBorder(frame, r, startRow, startColumn, endRow, endColumn,
                          x0, y1, column_widths);
        }
    }

    // otherwise draw the rows first, then the columns
    else { // dominance == COLUMN_DOMINANCE

        // Loop through the rows, drawing their borders
        for (r = firstRow; r <= lastRow; r++) {

            // compute the bottom of the row
            y1.add(row_heights[r]);

            drawRowBorder(frame, r, startRow, startColumn, endRow, endColumn,
                          x0, y1, column_widths);
        }

        // Loop through the columns, drawing their borders
        for (c = firstColumn; c <= lastColumn; c++) {

            // compute the right side of the column
            x1.add(column_widths[c]);

            drawColumnBorder(frame, c, startRow, startColumn, endRow,
                             endColumn, x1, y0, row_heights);
        }
    }

    // Analyze any borders that need to be re-drawn because of incorrect
    // precedence (e.g. row dominant, but column border is continuous)

    y1.assign(y0);
    for (r = firstRow; r <= lastRow; r++) {

        x1.assign(x0);
        y1.add(row_heights[r]);

        for (c = firstColumn; c <= lastColumn; c++) {
            x1.add(column_widths[c]);
            checkBorderIntersection(frame, r, c, x1, y1);
        }
    }

    // Move to the position of the right/bottom edges of the table if
    // we're not there already
    if (lastRow < endRow) {
        y1.add(row_heights[endRow]);
    }
    if (lastColumn < endColumn) {
        x1.add(column_widths[endColumn]);
    }

    // Output the edge borders
    if (dominance == ROW_DOMINANCE) {

        // if we need an external border on the left, draw it
        if (startColumn == 0 || external) {
            drawColumnBorder(frame, startColumn - 1, startRow, startColumn,
                             endRow, endColumn, x0, y0, row_heights);
        }

        // if we need an external border on the right, draw it
        if (endColumn + 1 == columnCount || external) {
            drawColumnBorder(frame, endColumn, startRow, startColumn,
                             endRow, endColumn, x1, y0, row_heights);
        }

        // if we need an external border on top (no headers), draw it
        if (!hasHeaders() && (startRow == 0 || external)) {
            drawRowBorder(frame, startRow - 1, startRow, startColumn,
                          endRow, endColumn, x0, y0, column_widths);
        }

        // if we need an external border below, draw it
        if (endRow + 1 == rowCount || external) {
            drawRowBorder(frame, endRow, startRow, startColumn,
                          endRow, endColumn, x0, y1, column_widths);
        }
    }
    else { // dominance == COLUMN_DOMINANCE

        // if we need an external border on top (no headers), draw it
        if (!hasHeaders() && (startRow == 0 || external)) {
            drawRowBorder(frame, startRow - 1, startRow, startColumn,
                          endRow, endColumn, x0, y0, column_widths);
        }

        // if we need an external border below, draw it
        if (endRow + 1 == rowCount || external) {
            drawRowBorder(frame, endRow, startRow, startColumn,
                          endRow, endColumn, x0, y1, column_widths);
        }

        // if we need an external border on the left, draw it
        if (startColumn == 0 || external) {
            drawColumnBorder(frame, startColumn - 1, startRow, startColumn,
                             endRow, endColumn, x0, y0, row_heights);
        }

        // if we need an external border on the right, draw it
        if (endColumn + 1 == columnCount || external) {
            drawColumnBorder(frame, endColumn, startRow, startColumn,
                             endRow, endColumn, x1, y0, row_heights);
        }
    }
}

// Check the intersection of borders to determine if it needs to be redrawn
private void checkBorderIntersection(JCFrame frame, int currRow, int currColumn,
                                     JCUnit.Measure x, JCUnit.Measure y) {

    Color        domColour;
    JCDrawStyle sub1, sub2, dom1, dom2 = null;

    // get the dominantn and subordinate direction borders
    if (dominance == ROW_DOMINANCE) {

        // The bottom row of the table takes the outside border; ignore it
        if (currRow + 1 >= rowCount) {
            return;
        }

        // Get the column (subordinate) borders at the intersection
        sub1 = getRawRightBorder(currRow, currColumn);
        sub2 = getRawRightBorder(currRow + 1, currColumn);

        // Get the row (dominant) borders at the intersection
        dom1 = getRawBottomBorder(currRow, currColumn);
        if (currColumn + 1 < columnCount) {
            dom2 = getRawBottomBorder(currRow, currColumn + 1);
        }
    }

    else { // dominance == COLUMN_DOMINANCE

        // The right column of the table takes the outside border; ignore it
        if (currColumn + 1 >= columnCount) {
            return;
        }

        // Get the row (subordinate) borders at the intersection
        sub1 = getRawBottomBorder(currRow, currColumn);
        sub2 = getRawBottomBorder(currRow, currColumn + 1);

        // Get the column (dominant) borders at the intersection
        dom1 = getRawRightBorder(currRow, currColumn);
        if (currRow + 1 < rowCount) {
            dom2 = getRawRightBorder(currRow + 1, currColumn);
        }
    }

    // if the dominant borders are the same, we're fine
    if ((dom1 != null && dom1.equals(dom2)) || dom1 == dom2) {
        return;
    }
    else {
        // They can't both be null!
        domColour = (dom2 != null) ? dom2.getForegroundColor()
                                   : dom1.getForegroundColor();
    }


    // if the subordinate borders are null or different, we're also fine
    if (sub1 == null || !sub1.equals(sub2) ||
        sub1.getForegroundColor() == null) {

        return;
    }

    // Hereafter, all intersections have a continuous subordinate border
    // but different borders in the dominant direction.

    // if the subordinate borders are the same colour as the dominant one,
    // we're still fine (since the difference won't be visible)
    if (sub1.getForegroundColor().equals(domColour)) {
        return;
    }

    // Get the dominant border width
    JCUnit.Measure intersectionLength = new JCUnit.Measure(), tmpWidth;

    if (dom1 != null) {
        intersectionLength.assign(dom1.getDrawnWidth());
    }
    if (dom2 != null) {
        tmpWidth = dom2.getDrawnWidth();
        if (tmpWidth.greaterThan(intersectionLength)) {
            intersectionLength.assign(tmpWidth);
        }
    }

    // Compute the width of the border we will be overwriting
    intersectionLength.divide(2.0);

    if (dominance == ROW_DOMINANCE) {
        JCUnit.Measure y1 = (JCUnit.Measure) y.clone();
        JCUnit.Measure y2 = (JCUnit.Measure) y.clone();
        y1.subtract(intersectionLength);
        y2.add(intersectionLength);

        drawBorder(frame, sub1, x, y1, x, y2);
    }
    else { // dominance == COLUMN_DOMINANCE
        JCUnit.Measure x1 = (JCUnit.Measure) x.clone();
        JCUnit.Measure x2 = (JCUnit.Measure) x.clone();
        x1.subtract(intersectionLength);
        x2.add(intersectionLength);

        drawBorder(frame, sub1, x1, y, x2, y);
    }
}

/**
 * Pastes the cells in a particular table view.
 * @param target the frame into which the cells are to be pasted
 * @param startRow the first row of the table from which to paste cells
 * @param startColumn the first column of the table from which to paste cells
 * @param endRow the last row of the table from which to paste cells
 * @param endColumn the last column of the table from which to paste cells
 * @param x0 the <i>x</i>-coordinate of the origin (upper-left) of the table view
 * @param y0 the <i>y</i>-coordinate of the origin (upper-left) of the table view
 */
private void pasteCells(JCFrame target, int startRow, int startColumn,
                        int endRow, int endColumn,
                        JCUnit.Measure x0, JCUnit.Measure y0) {

    int             r, c;
    Row             row;
    Column          column;
    Cell            cell;
    JCUnit.Measure  x1, y1, row_height;

    // sanity-check inputs
    if (startRow < 0 || endRow >= rowCount || startRow > endRow) {
        throw new IllegalArgumentException("Invalid row range");
    }
    if (startColumn < 0 || endColumn >= columnCount ||
        startColumn > endColumn) {

        throw new IllegalArgumentException("Invalid column range");
    }

    // Create temporary variables to hold the current position
    x1 = new JCUnit.Measure();
    y1 = new JCUnit.Measure();
    row_height = new JCUnit.Measure();

    // start at the top of the given space
    y1.assign(y0);

    // For each row of the table which is in the given range
    for (r = startRow; r <= endRow; r++) {
        x1.assign(x0);

        // get the height of the row
        row = getRow(r);
        row_height.assign(row.getHeight());
        row_height.add(row.getExtraHeight());

        // for each cell in this row that is in the given range
        for (c = startColumn; c <= endColumn; c++) {

            // Get the cell at this position in the table
            cell = getCellRaw(r, c);

            if (isHeader()) {
                column = mainTable.getColumn(c);
            }
            else {
                column = getColumn(c);
            }

            // if the cell is initialised and has a contents frame
            if (cell != UNINITIALISED_CELL && cell.getFrameRaw() != null) {

                // If the cell is part of a span
                if (cell.isSpanned()) {
                    Point origin = cell.getSpanOrigin();

                    // consider all spans whose root we have found in
                    // the region to draw, or the cell which is in the
                    // upper-left corner of the portion of a span which
                    // partially appears in the displayed region
                    if ((origin.x == c || c == startColumn) &&
                        (origin.y == r || r == startRow)) {

                        

                        // paste the displayed portion of a span
                        cell.pasteSpan(target, r, c, endRow, endColumn,
                                       x1, y1);
                    }
                }

                // otherwise (not part of a span) paste the cell
                else {
                    
                    cell.paste(target, x1, y1, row_height);
                }
            }

            // if the cell is uninitialised or has no contents frame,
            // but corresponds to an element of TableData
            else if (tableData != null) {

                // Check if the cell has any contents
                Object contents = tableData.getTableDataItem(r, c);
                if (contents != null) {

                    // if the cell value is an inherited Expression,
                    // we make a local copy specific to this cell
                    if (isTableDataValueInherited(r, c) &&
                        contents instanceof Expression) {

                        // set the context to be the current cell
                        ExpressionVariable.setVariable(this, "currentRow",
                                                       new MathScalar(r));
                        ExpressionVariable.setVariable(this, "currentColumn",
                                                       new MathScalar(c));

                        // clone the expression and set it on this cell
                        Expression currExpr = (Expression) contents;
                        try {
                            tableData.setTableDataItem(currExpr.clone(), r, c);
                        }
                        catch (CloneNotSupportedException e) {}
                    }

                    // Create a TableCellRender object from the cell and
                    // its contents and add it to the parent frame
                    Render tcr = new TableCellRender(this, r, c);
                    JCUnit.Point location = new JCUnit.Point(x1.units);
                    location.setX(x1);
                    location.setY(y1);
                    tcr.setLocation(location);
                    JCUnit.Dimension dimension = new JCUnit.Dimension(row_height.units);
                    dimension.setWidth(column.getWidth());
                    dimension.setHeight(row_height);
                    tcr.setSize(dimension);
                    target.getRenderList().add(tcr);
                }
            }

            // Advance the cell origin to the next column
            x1.add(column.getWidth());
            x1.add(column.getBorderWidth());
        }

        // Move down to the next row of the table
        y1.add(row_height);
        y1.add(row.getBorderWidth());
    }
}

/**
 * Determines if the <code>tableData</code> value for a given row and column is inherited.
 * @return
 * @param r
 * @param c
 */
private boolean isTableDataValueInherited(int r, int c) {

    Vector dataRows = ((PageTableVectorDataSource) tableData).getCells();
    if (r >= dataRows.size()) {
        return(true);
    }

    Vector dataCells = (Vector) dataRows.get(r);
    if (c >= dataCells.size()) {
        return(true);
    }

    return(dataCells.get(c) == null);
}


/**
 * Creates a report given the contents of a <code>TableDataModel</code>.
 * @param dataModel the table data model whose data are used to build the report
 */
/*
public createReport(TableDataModel dataModel) {

    // ??? Temporarily here while we work out mechanisms
    Row[] summaryRows = new Row[];

    // make sure we have enough data
    if (dataModel.getNumRows() < 1) {
        // ??? throw new IllegalArgumentException("Cannot report on zero data");
    }

    // Set the number of columns of data to copy from the data model
    int dataColumns = dataModel.getNumColumns();
    if (dataColumns > this.columnCount) {
        dataColumns = this.columnCount;
    }

    // Set the start row for summaries
    int[] startRows = new int[summaryRows.length];
    int[][] summaryPositions = new int[summaryRows.length][];

    // Add the first row of data to the table
    copyTableDataRow(dataModel, 0);

    lastRow = dataModel.getNumRows();
    previousRow = 0;
    currentRow = 1;

    // Copy the rows of data from the data model
    while (currentRow < lastRow) {

        // If the row changes context from the previous row
        if (rowComparator.isDifferent(dataModel, lastRow, currentRow)) {

            // determine the level of summary to add
            int level = rowComparator.getSummaryLevel(dataModel, lastRow,
                                                      currentRow);

            // Add a summary of the appropriate level
            addSummaryRow(summaryRows[level], summaryPositions);
        }

        // Copy the next row of data to the table
        copyTableDataRow(dataModel, currentRow);
        previousRow = currentRow;
        currentRow++;
    }

    // Add a summary over the whole table
    addSummaryRow(summaryRows[0], summaryPositions);
}
*/

/**
 * Adds a summary row to the table at the current line position.
 * @param row a description of the summary row to add
 * @param startRow an array of start rows to summarise from
 * @param summaryRows a structure listing the summary rows already in the table
 */
/*
private addSummaryRow(Row row, int[] startRows, int[][] summaryPositions) {
}
*/

/**
 * The Row object describes a single row of a <code>JCPageTable</code>.
 */
public class Row {

    /** The cell list is an ordered list of cell in the row. */
    protected List              cellList;

    /** The bottom border is drawn at the bottom of the row. */
    protected JCDrawStyle       bottomBorder;

    /** The color used as a default background for table elements. */
    protected Color             background;
    /** A template cell to use as a source of default cell attributes. */
    protected Cell              defaultCell;

    /** The height of the row is based on the tallest cell. */
    protected JCUnit.Measure    rowHeight;
    /** Flag if the row height has been overriden. */
    protected boolean            fixedHeight = false;

    /** Extra height can be added to equalise vertical spans. */
    protected JCUnit.Measure    extraHeight;
    /** The amount of space required to accomodate the row's border. */
    protected JCUnit.Measure    borderWidth;
    /** A count of the number of spans which begin on this row. */
    protected int               arrivingSpans;
    /** The number of spans which began above, but which appear in this row. */
    protected int               departingSpans;
    /** The number of this row in the table. */
    protected int               rowNumber;

    /**
     * Creates an empty row object.
     */
    public Row() {

        // Set all attributes to inherit their values
        bottomBorder = null;
        background = null;

        // Create a cell to hold all the cell defaults
        defaultCell = new Cell();

        // Create the list of cells for the row
        cellList = new Vector();

        rowHeight = null;
        extraHeight = null;
        borderWidth = null;
        rowNumber = -1;

        arrivingSpans = 0;
        departingSpans = 0;

        
    }

    /**
     * Creates a row object as a child of this table.
     * @param position The position of the row in the table
     */
    public Row(int position) {

        this();
        rowNumber = position;

        
    }

    /**
     * Determines if the background color is inherited from the table.
     * @return true if the background color is inherited
     */
    public boolean isBackgroundColorInherited() {

        return(background == null);
    }

    /**
     * Gets the color specified for painting the row background.
     * @return The row's background color
     */
    public Color getBackgroundColor() {

        if (isBackgroundColorInherited()) {
            return(JCPageTable.this.getBackgroundColor());
        }
        return(background);
    }

    /**
     * Sets the color specified for painting the row's background.
     * @param color The row's background color
     */
    public void setBackgroundColor(Color color) {

        background = color;
    }

    /**
     * Determine if the value being reported for the row's border is inherited.
     * @return true if the row's border style is inherited from the table
     */
    public boolean isBorderStyleInherited() {

        return(bottomBorder == null);
    }

    /**
     * Gets the draw style used for the row's border.
     * @return The draw style which defines the row's (bottom) border
     */
    public JCDrawStyle getBorderStyle() {

        if (isBorderStyleInherited()) {
            return(JCPageTable.this.getRowBorder());
        }
        return(bottomBorder);
    }

    /**
     * Sets the border style for the row's border.
     * @param border The draw style that describes the border
     */
    public void setBorderStyle(JCDrawStyle border) {

        // The style must be immutable!
        border = border.getImmutable();

        bottomBorder = border;
    }

    /**
     * Gets a particular cell from the row (instantiating it if necessary).
     * @param column The column to find (create) the cell in
     * @return The table cell which occupies or obscures (spans) this cell
     */
    public Cell getCell(int column) {

        int         i, cell_count;
        Cell        cellObj;

        // Verify that the cell is in range
        if (column < 0 || column >= columnCount) {
            throw new IllegalArgumentException("cell position out of range");
        }

        // If the cell has not been added yet
        cell_count = cellList.size();
        if (column >= cell_count) {
            // Add any necessary blank cells
            for (i = cell_count; i <= column; i++) {
                cellList.add(UNINITIALISED_CELL);
            }
        }

        // Get the cell
        cellObj = (Cell) cellList.get(column);

        // Create the cell if it doesn't exist
        if (cellObj == UNINITIALISED_CELL) {
            cellObj = new Cell(this, column);
            cellList.set(column, cellObj);
        }

        // If the cell is obscured by a span, get the root cell
        if (cellObj.isSpanned()) {
            Point origin = cellObj.getSpanOrigin();

            if (origin.x != column || origin.y != rowNumber) {
                cellObj = JCPageTable.this.getCell(origin.y, origin.x);
            }
        }

        return(cellObj);
    }

    /**
     * Returns the list of cells.
     * @return An ordered list of table cells in this row
     */
    public List getCellList() {
        return((List) cellList);
    }

    /**
     * Determines if the vertical cell alignment is inherited.
     * @return true if the table is examined to get the cell alignment
     */
    public boolean isDefaultCellAlignmentInherited() {

        return(defaultCell.isCellAlignmentInherited());
    }

    /**
     * Gets the alignment (within each row) of the cells in this column.
     * @return The vertical alignment of each cell with respect to its row
     */
    public int getDefaultCellAlignment() {

        if (isDefaultCellAlignmentInherited()) {
            return(JCPageTable.this.getDefaultCellAlignment());
        }
        return(defaultCell.getCellAlignment());
    }

    /**
     * Sets the alignment (within each row) of the cells in this column.
     * @param alignment The vertical alignment of each cell versus its row
     */
    public void setDefaultCellAlignment(int alignment) {

        defaultCell.setCellAlignment(alignment);
    }

    /**
     * Determines if the default style is inherited.
     * @return true if the table is examined to get the default style
     */
    public boolean isDefaultStyleInherited() {

        return(defaultCell.isStyleInherited());
    }

    /**
     * Gets the default cell style.
     * @return The default text style used in cell text for cells in this row
     */
    public JCTextStyle getDefaultStyle() {

        // if this row is inheriting a default, get it instead
        if (isDefaultStyleInherited()) {
            return(JCPageTable.this.getDefaultStyle());
        }

        // Get the style from the default cell
        return(defaultCell.getStyle());
    }

    /**
     * Sets the default cell text style.
     * @param style A text style to apply to printing in table cells
     */
    public void setDefaultStyle(JCTextStyle style) {

        defaultCell.setStyle(style);
    }

    /**
     * Determines whether the default bottom margin is inherited.
     * @return true if we need to get the default bottom margin from the parent
     */
    public boolean isDefaultBottomMarginInherited() {

        return(defaultCell.isBottomMarginInherited());
    }

    /**
     * Gets the default width of the bottom margin.
     * @return The width of the default bottom margin for cells in this row
     */
    public JCUnit.Measure getDefaultBottomMargin() {

        // if this table inherits a default bottom margin
        if (isDefaultBottomMarginInherited()) {

            // try the parent's default cell
            return(JCPageTable.this.getDefaultBottomMargin());
        }
        return(defaultCell.getBottomMargin());
    }

    /**
     * Sets the default width of a cell's bottom margin.
     * @param width The size of the bottom margin for table cells
     */
    public void setDefaultBottomMargin(JCUnit.Measure width) {

        defaultCell.setBottomMargin(width);
    }

    /**
     * Determine if the default top margin is inherited.
     * @return true if we need to get the default top margin from the parent
     */
    public boolean isDefaultTopMarginInherited() {

        return(defaultCell.isTopMarginInherited());
    }

    /**
     * Gets the default width of the top margin.
     * @return The width of the default top margin for cells in this row
     */
    public JCUnit.Measure getDefaultTopMargin() {

        // if this table inherits a default top margin
        if (isDefaultTopMarginInherited()) {

            // try the parent's default cell
            return(JCPageTable.this.getDefaultTopMargin());
        }
        return(defaultCell.getTopMargin());
    }

    /**
     * Sets the default width of a cell's top margin.
     * @param width The size of the top margin for table cells
     */
    public void setDefaultTopMargin(JCUnit.Measure width) {
        defaultCell.setTopMargin(width);
    }

    /**
     * Determines if the default left margin is inherited.
     * @return true if we need to get the default left margin from the parent
     */
    public boolean isDefaultLeftMarginInherited() {

        return(defaultCell.isLeftMarginInherited());
    }

    /**
     * Gets the default width of the left margin.
     * @return The width of the default left margin for cells in this row
     */
    public JCUnit.Measure getDefaultLeftMargin() {

        // if this table inherits a default left margin
        if (isDefaultLeftMarginInherited()) {

            // try the parent's default cell
            return(JCPageTable.this.getDefaultLeftMargin());
        }
        return(defaultCell.getLeftMargin());
    }

    /**
     * Sets the default width of a cell's left margin.
     * @param width The size of the left margin for table cells
     */
    public void setDefaultLeftMargin(JCUnit.Measure width) {
        defaultCell.setLeftMargin(width);
    }

    /**
     * Determine if the default right margin is inherited.
     * @return true if we need to get the default right margin from the parent
     */
    public boolean isDefaultRightMarginInherited() {

        return(defaultCell.isRightMarginInherited());
    }

    /**
     * Gets the default width of the right margin.
     * @return The width of the default right margin for cells in this row
     */
    public JCUnit.Measure getDefaultRightMargin() {

        // if this table inherits a default right margin
        if (isDefaultRightMarginInherited()) {

            // try the parent's default cell
            return(JCPageTable.this.getDefaultRightMargin());
        }
        return(defaultCell.getRightMargin());
    }

    /**
     * Sets the default width of a cell's right margin.
     * @param width The size of the right margin for table cells
     */
    public void setDefaultRightMargin(JCUnit.Measure width) {

        defaultCell.setRightMargin(width);
    }

    /**
     * Determines if the default cell value is inherited.
     * @return true if the row does not specify a default cell value
     */
    public boolean isDefaultCellValueInherited() {

        return(tableData.getTableRowLabel(rowNumber) == null);
    }

    /**
     * Gets the default contents for cells in this row.
     * @return The object which is the default contents for this row
     */
    public Object getDefaultCellValue() {

        // We don't pass inheritance of cell contents to the table
        return(tableData.getTableRowLabel(rowNumber));
    }

    /**
     * Sets the default contents for cells in this row.
     * @param obj The object to be the default contents for this row
     */
    public void setDefaultCellValue(Object obj) {

        // Store the value in the TableDataModel row label
        // tableData.setTableDataItem(obj, JCTableEnum.LABEL, rowNumber);
        ((PageTableVectorDataSource) tableData).setRowLabel(rowNumber, obj);
    }

    /**
     * Determine if the height of the row has been fixed.
     * @return true If the row has been assigned a fixed height
     */
    protected boolean hasFixedHeight() {
        return(fixedHeight);
    }

    /**
     * Gets the height of the row.
     * @return The row's height
     */
    public JCUnit.Measure getHeight() {
        return((JCUnit.Measure) rowHeight.clone());
    }

    /**
     * Sets the row's height; this may override the dynamic sizing of the row.
     * @param height The desired height of the row
     */
    public void setHeight(JCUnit.Measure height) {

        if (height.isNegative()) {
            throw new IllegalArgumentException("Invalid row height");
        }

        rowHeight = (JCUnit.Measure) height.clone();

        // set the row height on any existing cells
        Cell cell = null;
        int cell_count = cellList.size();
        for (int c = 0; c < cell_count; c++) {
            cell = (Cell) cellList.get(c);
            cell.setHeight(rowHeight);
        }

        this.fixedHeight = true;
    }

    /**
     * Gets the number (position) of this row.
     * @return the index of this row in the table
     */
    public int getRowNumber() {
        return(rowNumber);
    }

    /**
     * Gets the amount of extra height added to the row for alignment purposes.
     * @return The row's flexible additional height
     */
    protected JCUnit.Measure getExtraHeight() {

        // Make sure we don't accidentally return null
        if (extraHeight == null) {
            extraHeight = new JCUnit.Measure();
        }

        return(extraHeight);
    }

    /**
     * Sets the amount of extra height added to the row for alignment purposes.
     * @param height The extra height to expand the row
     */
    protected void setExtraHeight(JCUnit.Measure height) {

        if (height.isNegative()) {
            throw new IllegalArgumentException("Invalid height");
        }

        if (fixedHeight) {
            throw new IllegalArgumentException("Can't adjust height of fixed row");
        }

        extraHeight = (JCUnit.Measure) height.clone();
    }

    /**
     * Gets the number of spanned cells which enter this row.
     * @return the number of spans in this row which begin above
     */
    protected int getArrivingSpans() {
        return(arrivingSpans);
    }

    /**
     * Gets the number of spanned cells which leave this row.
     * @return the number of spans in this row which continue below
     */
    protected int getDepartingSpans() {
        return(departingSpans);
    }

    /**
     * Computes the height of the row based on cell heights.
     * @return The height required by the tallest cell in the row
     */
    protected JCUnit.Measure basicHeight() {

        int             c, cell_count;
        Cell            cell;
        Point           origin;
        JCUnit.Measure  new_height, cell_height;

        

        // Initialise the minimum height to the margin space
        new_height = new JCUnit.Measure(JCUnit.POINTS, 0.0);

        cell_count = cellList.size();
        for (c = 0; c < cell_count; c++) {
            cell = (Cell) cellList.get(c);

            if (cell != UNINITIALISED_CELL) {

                // Count cells which span multiple rows
                if (cell.isSpanned()) {

                    // Get the origin of the cell to determine span size
                    origin = cell.getSpanOrigin();
                    Cell root_cell = JCPageTable.this.getCell(origin.y,
                                                              origin.x);

                    if (root_cell.getSpanRows() > 1) {
                        

                        // Use a height of zero for spanned cells (this pass)
                        cell_height = new JCUnit.Measure();

                        // Count the number of spans that arrive from above
                        if (origin.y < this.rowNumber) {
                            arrivingSpans++;
                        }

                        // Count the number of spans that go on to the next row
                        if (this.rowNumber - origin.y <
                                                    cell.getSpanRows() - 1) {
                            departingSpans++;
                        }
                    }

                    // The cell is a horizontal span, one row high
                    else {
                        
                        cell_height = cell.fixHeight();
                    }
                }

                // The cell is a simple cell and is not spanned
                else {
                    
                    cell_height = cell.fixHeight();
                }
            }

            // otherwise the cell is not initialised, use the row or
			// column margins depending on the value of the dominance property

            else if (JCPageTable.this.getRowColumnDominance() == 
													JCPageTable.ROW_DOMINANCE) {

                

        		cell_height = (JCUnit.Measure) getDefaultTopMargin().clone();
        		cell_height.add(getDefaultBottomMargin());
			}

            else {
                Column      column;

                

                column = JCPageTable.this.getColumn(c);
                cell_height = (JCUnit.Measure)
                                        column.getDefaultTopMargin().clone();
                cell_height.add(column.getDefaultBottomMargin());
            }

            // The row height is the max of the current height and the cell's
            if (new_height.lessThan(cell_height)) {
                new_height = (JCUnit.Measure) cell_height.clone();
            }
        }

        // if the row height is fixed, use that height
        if (fixedHeight) {
            
            return(rowHeight);
        }

        // Set the computed height of the row and return it
        
        rowHeight = new_height;
        return(new_height);
    }

    /**
     * Calculates the width of the row's border based on the cells.
     * @return the space required for the borders of all the cells
     */
    private JCUnit.Measure calculateBorderWidth() {

        int             c, cell_count;
        Cell            cell;
        JCUnit.Measure  width, tmp_width;

        // Get the default border width for this row
        width = getBorderStyle().getDrawnWidth();

        

        // Scan the cells of the row in turn
        cell_count = cellList.size();
        for (c = 0; c < cell_count; c++) {
            cell = (Cell) cellList.get(c);

            // Ignore uninitialised cells
            if (cell == UNINITIALISED_CELL) {
                continue;
            }

            // Ignore all spanned cells but the lower-left corner
            if (cell.isSpanned()) {
                Point origin = cell.getSpanOrigin();

                // Ignore the cell if we're not in the first column
                if (origin.x != c) {
                    continue;
                }

                // get the root cell of the span
                cell = JCPageTable.this.getCell(origin.y, origin.x);

                // Ignore the cell if we're not in the last row
                if (origin.y + cell.getSpanRows() - 1 != rowNumber) {
                    continue;
                }
            }

            // Ignore the cell if it inherits its border style
            if (cell.isBottomBorderInherited()) {
                continue;
            }

            

            // Get the width of the cell's border and accumulate the largest
            tmp_width = cell.getBottomBorderStyle().getDrawnWidth();

            

            if (tmp_width.greaterThan(width)) {
                width = tmp_width;
            }
        }
        

        return(width);
    }

    /**
     * Returns the width of the row's border, computing it if necessary.
     * @return the space required for the row's border
     */
    protected JCUnit.Measure getBorderWidth() {

        
        // if the border width hasn't been calculated yet, calculate it.
        if (borderWidth == null) {
            JCUnit.Measure  new_width;

            // The last row of the table uses the table's border regardless
            if (rowNumber + 1 == JCPageTable.this.getRowCount()) {
                new_width = JCPageTable.this.getBottomBorder().getDrawnWidth();
            }
            else {
                new_width = calculateBorderWidth();
            }
            
            borderWidth = (JCUnit.Measure) new_width.clone();
        }
        
        return((JCUnit.Measure) borderWidth.clone());
    }

    /**
     * Computes the space required for this row's bottom border.
     * @return The width needed if this row is at the bottom of the table part
     */
    protected JCUnit.Measure getBottomBorderWidth() {

        // Return the full space required by this border
        return(this.getBottomBorderWidth(true));
    }

    /**
     * Computes the space required for the border of this row at the bottom
     * of a table or table section.
     * @param fullWidth True if the total space to the next row is required;
     *                  false if the half-width in this row's domain is wanted
     * @return The space required for the widest border below this row
     */
    protected JCUnit.Measure getBottomBorderWidth(boolean fullWidth) {

        int             mode;
        JCUnit.Measure  bottom, border, width;

        // Get the border width the parent table specifies for the bottom border
        if (JCPageTable.this.isHeader()) {
            bottom = JCPageTable.this.getHeaderBorder().getDrawnWidth();
        }
        else {
            bottom = JCPageTable.this.getBottomBorder().getDrawnWidth();
        }

        // Get the border mode and the row's computed border width
        mode = JCPageTable.this.getBorderMode();
        border = this.getBorderWidth();

        // if no cells span downwards (simple case)
        if (departingSpans == 0) {

            // if this is the table's last row or external borders are used
            if (rowNumber == JCPageTable.this.getRowCount() - 1 ||
                mode == BORDER_USE_EXTERNAL) {

                // use the bottom border's width
                width = (JCUnit.Measure) bottom.clone();
            }

            // otherwise use the row's border width
            else {
                width = border;
            }
        }

        // otherwise one or more cells spans to the next row
        else {

            // if internal borders are used, use the row's border width
            // NOTE:  we can't be at the bottom of the table if there is
            //        a span to a following row, so this must be internal
            if (mode == BORDER_USE_INTERNAL) {
                width = border;
            }

            // otherwise (external borders) we need to accomodate the span
            else {

                // use the maximum of the row and bottom border
                if (border.greaterThan(bottom)) {
                    width = border;
                }
                else {
                    width = (JCUnit.Measure) bottom.clone();
                }


                // if we are including outside space we use the computed max
                // (above) for the inside half and the bottom border for the
                // outside half.  (We won't be dividing by two in this case)
                if (fullWidth) {
                    width.add(bottom);
                    width.divide(2.0);
                }
            }
        }

        // if only the row's share of the width is required, halve it
        if (!fullWidth) {
            width.divide(2.0);
        }

        return(width);
    }

    /**
     * Computes the space required for this row's border at the top.
     * @return The width needed if this row is at the top of the table part
     */
    protected JCUnit.Measure getTopBorderWidth() {

        // Return the full space required by this border
        return(this.getTopBorderWidth(true));
    }

    /**
     * Computes the space required for the border of this row at the top
     * of a table or table section.
     * @param fullWidth True if the total space between this row and the
     *                  previous one is required, otherwise false (half)
     * @return The space required for the widest border above this row
     */
    protected JCUnit.Measure getTopBorderWidth(boolean fullWidth) {

        int             mode;
        JCUnit.Measure  top, border, width;

        // if there is a preceding row, use its computed border width;
        if (rowNumber > 0) {
            border = JCPageTable.this.getRow(rowNumber - 1).getBorderWidth();
        }

        // otherwise there is no preceding space requirement
        else {
            border = new JCUnit.Measure();
        }

        // If the table has headers, they accomodate all but span spacing
        if (JCPageTable.this.hasHeaders()) {

            // only add space if there are spans extending into this row
            if (arrivingSpans != 0) {
                width = border;
            }
            else {
                width = new JCUnit.Measure();
            }

            // since "full width" is meaningless with headers, halve the width
            if (fullWidth) {
                width.divide(2.0);
            }
        }

        // Otherwise there are no headers
        else {

            // Get the border mode and the top border
            mode = JCPageTable.this.getBorderMode();
            top = JCPageTable.this.getTopBorder().getDrawnWidth();

            // in internal mode use the previous row's border
            // (except in the case of the first row, which hasn't a previous)
            if (mode == BORDER_USE_INTERNAL && rowNumber > 0) {
                width = border;
            }

            // otherwise if this row has no cells spanning in from above
            // (accounts for the case of the first row regardless of mode and
            // the case of external borders with no spans)
            else if (arrivingSpans == 0) {
                width = (JCUnit.Measure) top.clone();
            }

            // finally the case of external borders with spans from above
            else {
                // use the max of the top and previous border
                if (top.greaterThan(border)) {
                    width = (JCUnit.Measure) top.clone();
                }
                else {
                    width = border;
                }

                // if we are including outside space we use the computed max
                // (above) for the inside half and the top border for the
                // outside half.  (We won't be dividing by two in this case)
                if (fullWidth) {
                    width.add(top);
                    width.divide(2.0);
                }
            }
        }

        // if only the row's share of the width is required, halve it
        if (!fullWidth) {
            width.divide(2.0);
        }

        return(width);
    }

    /**
     * Computes the height of this row (including extra space and border width)
     * based on the given information about its position in the table part.
     * @param first True if the row is the first (top) row of the table part
     * @param last True if the row is the last (bottom) row of the table part
     * @return the height of this row in the specified position
     */
    protected JCUnit.Measure getDrawnHeight(boolean first, boolean last) {

        JCUnit.Measure  draw_height;

        // if this is the first row, use the special top row border method
        if (first) {
            draw_height = getTopBorderWidth(false);
        }
        // otherwise use half of the previous row's border
        // (NOTE:  the first row of the table _must_ be a "first" row)
        else {
            draw_height =
                        JCPageTable.this.getRow(rowNumber - 1).getBorderWidth();
            draw_height.divide(2.0);
        }

        

        // Add the height of the row's cells and the extra space
        draw_height.add(this.rowHeight);

        

        // Force the evaluation of extraHeight
        draw_height.add(this.getExtraHeight());

        

        // if this is the last row, use the special bottom border method
        if (last) {
            draw_height.add(getBottomBorderWidth(false));

            
        }
        // otherwise use half of this row's border
        else {
            JCUnit.Measure  tmp_height;

            tmp_height = this.getBorderWidth();
            tmp_height.divide(2.0);

            draw_height.add(tmp_height);

            
        }

        

        return(draw_height);
    }

} // JCPageTable.Row


/**
 * The Column object describes a single column of a <code>JCPageTable</code>.
 */
public class Column {

    /** The right border is drawn on the right side of the column. */
    protected JCDrawStyle       rightBorder;

    /** The color used as a default background for table elements. */
    protected Color             background;
    /** A template cell to use as a source of default cell attributes. */
    protected Cell              defaultCell;

    /** The width of the column. */
    protected JCUnit.Measure    width;

    /** The amount of space required to accomodate the column's border. */
    protected JCUnit.Measure    borderWidth;
    /** The number of spans which began left of and appear in this column. */
    protected int               arrivingSpans;
    /** The number of spans which leave this column to the right. */
    protected int               departingSpans;
    /** The number of this column in the table. */
    protected int               columnNumber;

    /**
     * Creates an empty column object.
     */
    public Column() {

        // Inherit these attributes by default
        rightBorder = null;
        background = null;

        // Create a cell to store the default cell settings
        defaultCell = new Cell();

        // Set the width and number to initial/unset values
        width = null;
        borderWidth = null;
        columnNumber = -1;

        arrivingSpans = 0;
        departingSpans = 0;
    }

    /**
     * Creates a column object as a child of the given table.
     * @param position The position of the column in the table
     */
    public Column(int position) {

        this();
        columnNumber = position;
    }

    /**
     * Creates a column object as a child of the given table.
     * @param width The width of the column
     * @param position The position of the column in the table
     */
    public Column(JCUnit.Measure width, int position) {

        this();
        setWidth(width);
        columnNumber = position;
    }

    /**
     * Determines if the background color is inherited from the table.
     * @return true if the background color is inherited
     */
    public boolean isBackgroundColorInherited() {

        return(background == null);
    }

    /**
     * Gets the color specified for painting the column background.
     * @return The column's background color
     */
    public Color getBackgroundColor() {

		// if inherited, grab inherited color
        if (isBackgroundColorInherited()) {
			// for header tables, defer to column of main table, but
			// only if there is a color set there.  If no color is
			// specified on the column of the main table, inherit
			// from background of the header table
            if (isHeader() && columnNumber != -1) {
                if (!JCPageTable.this.mainTable.
						getColumn(columnNumber).isBackgroundColorInherited()) {
                	return(JCPageTable.this.mainTable.
						getColumn(columnNumber).getBackgroundColor());
				}
				else {
                	return(JCPageTable.this.getBackgroundColor());
				}
            }
			// non-header tables -- just inherit from background of the table
            else {
                return(JCPageTable.this.getBackgroundColor());
            }
        }
        return(background);
    }

    /**
     * Sets the color specified for painting the column's background.
     * @param color The column's background color
     */
    public void setBackgroundColor(Color color) {

        background = color;
    }

    /**
     * Determines if the value being reported for the border is inherited.
     * @return true if the column's border style is inherited from the table
     */
    public boolean isBorderStyleInherited() {

        return(rightBorder == null);
    }

    /**
     * Gets the draw style used for the column's border.
     * @return The draw style which defines the column's (right) border
     */
    public JCDrawStyle getBorderStyle() {

        if (isBorderStyleInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getBorderStyle());
            }
            else {
                return(JCPageTable.this.getColumnBorder());
            }
        }
        return(rightBorder);
    }

    /**
     * Sets the border style for the column's border.
     * @param border The draw style that describes the border
     */
    public void setBorderStyle(JCDrawStyle border) {

        // The style must be immutable!
        border = border.getImmutable();

        rightBorder = border;
    }

    /**
     * Determines if the vertical cell alignment is inherited.
     * @return true if the table is examined to get the cell alignment
     */
    public boolean isDefaultCellAlignmentInherited() {

        return(defaultCell.isCellAlignmentInherited());
    }

    /**
     * Gets the alignment (within each row) of the cells in this column.
     * @return The vertical alignment of each cell with respect to its row
     */
    public int getDefaultCellAlignment() {

        if (isDefaultCellAlignmentInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getDefaultCellAlignment());
            }
            else {
                return(JCPageTable.this.getDefaultCellAlignment());
            }
        }
        if (!isHeader && defaultCell.getCellAlignment() == CELL_ALIGNMENT_NONE) {
            return CELL_ALIGNMENT_TOP;
        }
        return(defaultCell.getCellAlignment());
    }

    /**
     * Sets the alignment (within each row) of the cells in this column.
     * @param alignment The vertical alignment of each cell versus its row
     */
    public void setDefaultCellAlignment(int alignment) {

        defaultCell.setCellAlignment(alignment);
    }

    /**
     * Determines if the default style is inherited.
     * @return true if the table is examined to get the default style
     */
    public boolean isDefaultStyleInherited() {

        return(defaultCell.isStyleInherited());
    }

    /**
     * Gets the default cell style.
     * @return The default text style used in cell text for cells in this column
     */
    public JCTextStyle getDefaultStyle() {

        // if this column is inheriting a default, get it instead
        if (isDefaultStyleInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getDefaultStyle());
            }
            else {
                return(JCPageTable.this.getDefaultStyle());
            }
        }

        // Get the style from the default cell
        return(defaultCell.getStyle());
    }

    /**
     * Sets the default cell text style.
     * @param style A text style to apply to printing in cells in this column
     */
    public void setDefaultStyle(JCTextStyle style) {

        defaultCell.setStyle(style);
    }

    /**
     * Determines if the default bottom margin is inherited.
     * @return true if we need to get the default bottom margin from the parent
     */
    public boolean isDefaultBottomMarginInherited() {

        return(defaultCell.isBottomMarginInherited());
    }

    /**
     * Gets the default width of the bottom margin.
     * @return The width of the default bottom margin for cells in this column
     */
    public JCUnit.Measure getDefaultBottomMargin() {

        // if this table inherits a default bottom margin
        if (isDefaultBottomMarginInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getDefaultBottomMargin());
            }
            else {
                // try the parent's default cell
                return(JCPageTable.this.getDefaultBottomMargin());
            }
        }
        return(defaultCell.getBottomMargin());
    }

    /**
     * Sets the default width of a cell's bottom margin.
     * @param width The size of the bottom margin for table cells
     */
    public void setDefaultBottomMargin(JCUnit.Measure width) {

        defaultCell.setBottomMargin(width);
    }

    /**
     * Determines if the default top margin is inherited.
     * @return true if we need to get the default top margin from the parent
     */
    public boolean isDefaultTopMarginInherited() {

        return(defaultCell.isTopMarginInherited());
    }

    /**
     * Gets the default width of the top margin.
     * @return The width of the default top margin for cells in this column
     */
    public JCUnit.Measure getDefaultTopMargin() {

        // if this table inherits a default top margin
        if (isDefaultTopMarginInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getDefaultTopMargin());
            }
            else {
                // try the parent's default cell
                return(JCPageTable.this.getDefaultTopMargin());
            }
        }
        return(defaultCell.getTopMargin());
    }

    /**
     * Sets the default width of a cell's top margin.
     * @param width The size of the top margin for table cells
     */
    public void setDefaultTopMargin(JCUnit.Measure width) {
        defaultCell.setTopMargin(width);
    }

    /**
     * Determines if the default left margin is inherited.
     * @return true if we need to get the default left margin from the parent
     */
    public boolean isDefaultLeftMarginInherited() {

        return(defaultCell.isLeftMarginInherited());
    }

    /**
     * Gets the default width of the left margin.
     * @return The width of the default left margin for cells in this column
     */
    public JCUnit.Measure getDefaultLeftMargin() {

        // if this table inherits a default left margin
        if (isDefaultLeftMarginInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getDefaultLeftMargin());
            }
            else {
                // try the parent's default cell
                return(JCPageTable.this.getDefaultLeftMargin());
            }
        }
        return(defaultCell.getLeftMargin());
    }

    /**
     * Sets the default width of a cell's left margin.
     * @param width The size of the left margin for table cells
     */
    public void setDefaultLeftMargin(JCUnit.Measure width) {
        defaultCell.setLeftMargin(width);
    }

    /**
     * Determines if the default right margin is inherited.
     * @return true if we need to get the default right margin from the parent
     */
    public boolean isDefaultRightMarginInherited() {

        return(defaultCell.isRightMarginInherited());
    }

    /**
     * Gets the default width of the right margin.
     * @return The width of the default right margin for cells in this column
     */
    public JCUnit.Measure getDefaultRightMargin() {

        // if this table inherits a default right margin
        if (isDefaultRightMarginInherited()) {
            if (isHeader() && columnNumber != -1) {
                return(JCPageTable.this.mainTable.getColumn(columnNumber).getDefaultRightMargin());
            }
            else {
                // try the parent's default cell
                return(JCPageTable.this.getDefaultRightMargin());
            }
        }
        return(defaultCell.getRightMargin());
    }

    /**
     * Sets the default width of a cell's right margin.
     * @param width The size of the right margin for table cells
     */
    public void setDefaultRightMargin(JCUnit.Measure width) {

        defaultCell.setRightMargin(width);
    }

    /**
     * Determine if the default cell value is inherited.
     * @return true if the column does not specify a default cell value
     */
    public boolean isDefaultCellValueInherited() {

        return(tableData.getTableColumnLabel(columnNumber) == null);
    }

    /**
     * Gets the default contents for cells in this column.
     * @return The object which is the default contents for this column
     */
    public Object getDefaultCellValue() {

        // We don't pass inheritance of cell contents to the table
        return(tableData.getTableColumnLabel(columnNumber));
    }

    /**
     * Sets the default contents for cells in this column.
     * @param obj The object to be the default contents for this column
     */
    public void setDefaultCellValue(Object obj) {

        // Store the value in the TableDataModel column label
        // tableData.setTableDataItem(obj, JCTableEnum.LABEL, columnNumber);
        ((PageTableVectorDataSource) tableData).setColumnLabel(columnNumber, obj);
    }

    /**
     * Gets the width of the column.
     * @return The column's width
     */
    public JCUnit.Measure getWidth() {
        return(width);
    }

    /*
     * Sets the column's width; this may break the creation of cell frames.
     * @param width The desired width of the column
     */
    public void setWidth(JCUnit.Measure width) {

        if (width.isNegative()) {
            throw new IllegalArgumentException("Invalid column width");
        }

        this.width = (JCUnit.Measure) width.clone();
    }

    /**
     * Gets the position of the column in the table.
     * @return This column's index in the table
     */
    public int getColumnNumber() {
        return(columnNumber);
    }

    /**
     * Increments the number of spans in this column which begin to the left.
     */
    protected void addArrivingSpan() {
        arrivingSpans++;
    }

    /**
     * Gets the number of spans which begin left of this column.
     * @return the number of spanned cells which continue into this column
     */
    protected int getArrivingSpans() {
        return(arrivingSpans);
    }

    /**
     * Increments the number of spans in this column which continue to the right.
     */
    protected void addDepartingSpan() {
        departingSpans++;
    }

    /**
     * Gets the number of spans which contine right of this column.
     * @return the number of spanned cells which leave this column to the right
     */
    protected int getDepartingSpans() {
        return(departingSpans);
    }

    /**
     * Calculates the width of the column's border based on the cells.
     * @return space required for the borders of the cells in the column
     */
    private JCUnit.Measure calculateBorderWidth() {

        int             r, row_count;
        Row             row;
        Cell            cell;
        List            row_list, cell_list;
        JCUnit.Measure  width, tmp_width;

        // Get the width of the column's default border
        width = getBorderStyle().getDrawnWidth();

        // Get the table's row list
        row_list = JCPageTable.this.getRowList();
        row_count = row_list.size();

        if (!isHeader() && JCPageTable.this.header != null) {
            // Get the width of the cell's border and accumulate the largest
            tmp_width =  JCPageTable.this.header.getColumn(columnNumber).getBorderWidth();
            if (tmp_width.greaterThan(width)) {
                width = tmp_width;
            }
        }

        // Scan the cells in the column
        for (r = 0; r < row_count; r++) {
            row = (Row) row_list.get(r);

            // Ignore uninitialised rows
            if (row == UNINITIALISED_ROW) {
                continue;
            }

            // If the row doesn't have a cell in this position, ignore it
            cell_list = row.getCellList();
            if (cell_list.size() <= columnNumber) {
                continue;
            }

            // Ignore place-holder cells
            cell = (Cell) cell_list.get(columnNumber);
            if (cell == UNINITIALISED_CELL) {
                continue;
            }

            // Ignore all spanned cells but the upper-right corner
            if (cell.isSpanned()) {
                Point origin = cell.getSpanOrigin();

                // Ignore the cell if we're not in the first row
                if (origin.y != r) {
                    continue;
                }

                // get the root cell of the span
                cell = JCPageTable.this.getCell(origin.y, origin.x);

                // Ignore the cell if we're not in the last column
                if (origin.x + cell.getSpanColumns() - 1 != columnNumber) {
                    continue;
                }
            }

            // Ignore the cell if it inherits its border style
            if (cell.isRightBorderInherited()) {
                continue;
            }

            // Get the width of the cell's border and accumulate the largest
            tmp_width = cell.getRightBorderStyle().getDrawnWidth();
            if (tmp_width.greaterThan(width)) {
                width = tmp_width;
            }
        }

        return(width);
    }

    /**
     * Computes the space required for the border of this column at the left.
     * side of a table or table section
     * @return The space required for the widest border before this column
     */
    protected JCUnit.Measure getLeftBorderWidth() {

        // Return the full space required by this border
        return(this.getLeftBorderWidth(true));
    }

    /**
     * Returns whether or not there are arriving spans in the same column
     * in this or related table (ie header table or main table).
     * @return whether or not there are arriving spans in the same column
     */
    protected boolean noArrivingSpans() {
        boolean local = (arrivingSpans == 0);
        boolean remote = false;

        if (JCPageTable.this.isHeader()) {
            // This is a headers table so check if there are arriving spans
            // in the main table
            remote = (JCPageTable.this.mainTable.getColumn(columnNumber).getArrivingSpans() == 0);
        }
        else if (JCPageTable.this.hasHeaders()) {
            // This is a main table so check if there are arriving spans
            // in the headers table
            remote = (JCPageTable.this.header.getColumn(columnNumber).getArrivingSpans() == 0);
        }
        return local && remote;
    }

    /**
     * Returns whether or not there are departing spans in the same column
     * in this or a related table (ie header table or main table).
     * @return whether or not there are departing spans in the same column
     */
    protected boolean noDepartingSpans() {
        boolean local = (departingSpans == 0);
        boolean remote = false;

        if (JCPageTable.this.isHeader()) {
            // This is a headers table so check if there are arriving spans
            // in the main table
            remote = (JCPageTable.this.mainTable.getColumn(columnNumber).getDepartingSpans() == 0);
        }
        else if (JCPageTable.this.hasHeaders()) {
            // This is a main table so check if there are arriving spans
            // in the headers table
            remote = (JCPageTable.this.header.getColumn(columnNumber).getDepartingSpans() == 0);
        }
        return local && remote;
    }

    /**
     * Computes the space required for the border of this column at the left
     * side of a table or table section.
     * @param fullWidth True if the total space between this column and the
     *                  previous one is required, otherwise false (half)
     * @return The space required for the widest border before this column
     */
    protected JCUnit.Measure getLeftBorderWidth(boolean fullWidth) {

        int             mode;
        JCUnit.Measure  left, border, width;

        // if there is a preceding column, use its computed border width;
        if (columnNumber > 0) {
            if (JCPageTable.this.isHeader()) {
                border = JCPageTable.this.mainTable.getColumn(columnNumber - 1).getBorderWidth();
            }
            else {
                border =
                    JCPageTable.this.getColumn(columnNumber - 1).getBorderWidth();
            }
        }

        // otherwise there is no preceding space requirement
        else {
            border = new JCUnit.Measure();
        }

        // Get the border mode and left border width of the table
        mode = JCPageTable.this.getBorderMode();
        left = JCPageTable.this.getLeftBorderWidth();

        // in internal mode use the previous column's border
        // (except in the case of the first column, which hasn't a previous)
        if (mode == BORDER_USE_INTERNAL && columnNumber > 0) {
            width = border;
        }

        // otherwise if this column has no cells spanning in from the left
        // (accounts for the case of the first column regardless of mode
        // and the case of external borders with no spans)
        else if (noArrivingSpans()) {
            width = (JCUnit.Measure) left.clone();
        }

        // finally the case of external borders with spans from the left
        else {
            // use the max of the left and previous border
            if (left.greaterThan(border)) {
                width = (JCUnit.Measure) left.clone();
            }
            else {
                width = border;
            }

            // if we are including outside space we use the computed max
            // (above) for the inside half and the left border for the
            // outside half.  (We won't be dividing by two in this case)
            if (fullWidth) {
                width.add(left);
                width.divide(2.0);
            }
        }

        // if only the column's share of the width is required, halve it
        if (!fullWidth) {
            width.divide(2.0);
        }

        return(width);
    }

    /**
     * Computes the space required for the border of this column at the right
     * side of a table or table section.
     * @return The space required for the widest border following this column
     */
    protected JCUnit.Measure getRightBorderWidth() {

        // Return the full space required by this border
        return(this.getRightBorderWidth(true));
    }

    /**
     * Computes the space required for the border of this column at the right
     * side of a table or table section.
     * @param fullWidth True if the total space between this column and the
     *                  next one is required, otherwise false (half)
     * @return The space required for the widest border following this column
     */
    protected JCUnit.Measure getRightBorderWidth(boolean fullWidth) {

        int             mode;
        JCUnit.Measure  right, border, width;

        // Get the border mode and the table's and column's border width
        mode = JCPageTable.this.getBorderMode();
        right = JCPageTable.this.getRightBorderWidth();
        if (JCPageTable.this.isHeader()) {
            border = JCPageTable.this.mainTable.getColumn(columnNumber).getBorderWidth();
        }
        else {
            border = this.getBorderWidth();
        }

        // if no cells span onwards (to the right)
        if (noDepartingSpans()) {

            // if this is the table's last column or external borders are used
            if (columnNumber == JCPageTable.this.getColumnCount() - 1 ||
                mode == BORDER_USE_EXTERNAL) {

                // use the right border's width
                width = (JCUnit.Measure) right.clone();
            }

            // otherwise use the column's border width
            else {
                width = border;
            }
        }

        // otherwise one or more cells spans to the next column
        else {

            // if internal borders are used, use the column's border width
            // NOTE:  we can't be at the right side of the table if there is
            //        a span to the right, so this will not be the last column
            if (mode == BORDER_USE_INTERNAL) {
                width = border;
            }

            // otherwise (external borders) we need to accomodate the span
            else {

                // use the maximum of the column and right border
                if (border.greaterThan(right)) {
                    width = border;
                }
                else {
                    width = (JCUnit.Measure) right.clone();
                }

                // if we are including outside space we use the computed max
                // (above) for the inside half and the bottom border for the
                // outside half.  (We won't be dividing by two in this case)
                if (fullWidth) {
                    width.add(right);
                    width.divide(2.0);
                }
            }
        }

        // if only the column's share of the width is required, halve it
        if (!fullWidth) {
            width.divide(2.0);
        }

        return(width);
    }

    /**
     * Computes the width of this column (including border width) based on the
     * given information about its position in the table part.
     * @param first True if the column is the first column of the table part
     * @param last True if the column is the last column of the table part
     * @return the width of this column in the specified position
     */
    protected JCUnit.Measure getDrawnWidth(boolean first, boolean last) {

        JCUnit.Measure  draw_width;

        // if this is the first column, use the left column border method
        if (first) {
            draw_width = getLeftBorderWidth(false);
        }
        // otherwise use half of the previous column's border
        // (NOTE:  the first column of the table _must_ be a "first" column)
        else {
            if (JCPageTable.this.isHeader()) {
                draw_width =
                    JCPageTable.this.mainTable.getColumn(columnNumber - 1).getBorderWidth();
            }
            else {
                draw_width =
                    JCPageTable.this.getColumn(columnNumber - 1).getBorderWidth();
            }

            draw_width.divide(2.0);
        }

        // Add the width of the column's cells
        draw_width.add(this.width);

        // if this is the last column, use the right column border method
        if (last) {
            draw_width.add(getRightBorderWidth(false));
        }
        // otherwise use half of this column's border
        else {
            JCUnit.Measure  tmp_width;

            if (JCPageTable.this.isHeader()) {
                tmp_width = JCPageTable.this.mainTable.getColumn(columnNumber).getBorderWidth();
            }
            else {
                tmp_width = this.getBorderWidth();
            }

            tmp_width.divide(2.0);

            draw_width.add(tmp_width);
        }

        return(draw_width);
    }

    /**
     * Returns the width of the column's border, computing it if necessary.
     * @return the space required for the column's border
     */
    protected JCUnit.Measure getBorderWidth() {

        // if the border width hasn't been calculated yet, calculate it.
        if (borderWidth == null) {
            JCUnit.Measure  new_width;

            // The last column of the table uses the table's border regardless
            if (columnNumber + 1 == JCPageTable.this.getColumnCount()) {
                new_width = JCPageTable.this.getRightBorder().getDrawnWidth();
            }
            else {
                new_width = calculateBorderWidth();
            }
            
            borderWidth = (JCUnit.Measure) new_width.clone();

        }
        return((JCUnit.Measure) borderWidth.clone());
    }

} // JCPageTable.Column


/**
 * The Cell object describes a single cell of a JCPageTable.
 */
public class Cell {

    /** The color used as a default background for table elements. */
    protected Color             background;
    /** The bottom border is drawn at the bottom of the cell. */
    protected JCDrawStyle       bottomBorder;
    /** The right border is drawn at the right of the cell. */
    protected JCDrawStyle       rightBorder;
    /** The text style used to format the text in the cell. */
    protected JCTextStyle       textStyle;
    /** Cell alignment controls the vertical alignment of cells in this cell. */
    protected int               cellAlignment = CELL_ALIGNMENT_NONE;

    /** The width of the bottom margin between the border and the frame. */
    protected JCUnit.Measure    bottomMargin;
    /** The width of the left margin between the border and the frame. */
    protected JCUnit.Measure    leftMargin;
    /** The width of the right margin between the border and the frame. */
    protected JCUnit.Measure    rightMargin;
    /** The width of the top margin between the border and the frame. */
    protected JCUnit.Measure    topMargin;

    /** Row which owns this cell. */
    protected Row               rowParent;
    /** Column to which this cell belongs. */
    protected Column            columnParent;

    /** The frame that contains the contents of the cell. */
    protected JCFrame           frame;
    /** The height of the cell is based on the size of its contents. */
    protected JCUnit.Measure    cellHeight;
    /** The width of the cell is based on the width of its column. */
    protected JCUnit.Measure    width;

    /** The row of the table this cell appears in. */
    protected int               rowNumber;
    /** The column of the table this cell appears in. */
    protected int               columnNumber;

    /** Marker for cells which belong to (are obscured by) a span. */
    protected boolean           isSpanned;
    /** The root cell of the span. */
    protected Point             spanOrigin;
    /** The number of rows the cell spans downward. */
    protected int               spanRows;
    /** The number of columns the cell spans to the right. */
    protected int               spanColumns;

    /**
     * Creates an empty cell object.
     */
    public Cell() {

        // Inherit all attributes which control the appearance
        background = null;
        bottomBorder = null;
        rightBorder = null;
        textStyle = null;

        bottomMargin = null;
        leftMargin = null;
        rightMargin = null;
        topMargin = null;

        // Create a frame only when required
        frame = null;

        // These values are initially inherited/dynamic
        cellHeight = null;
        width = null;

        // New cells are not spanned
        isSpanned = false;
        spanRows = 0;
        spanColumns = 0;

        // don't initialise row and column
        rowParent = null;
        rowNumber = 0;
        columnParent = null;
        columnNumber = 0;
    }

    /**
     * Creates a cell object as a child of the given table.
     * @param parent the parent row for this cell
     * @param position the position of the cell in the table
     */
    public Cell(Row parent, int position) {

        // Initialise the defaults of the cell
        this();

        // define the parent and location of the cell
        rowParent = parent;
        rowNumber = rowParent.getRowNumber();
        columnNumber = position;
        columnParent = JCPageTable.this.getColumn(position);

        // set the cell width based on the column
        width = columnParent.getWidth();

        // if row parent has fixed height, propagate it to this cell
        if (rowParent.hasFixedHeight()) {
            cellHeight = (JCUnit.Measure) rowParent.getHeight().clone();
        }

        
    }

    /**
     * Determines if the cell's background color attribute is inherited.
     * @return <code>true</code> if the background color is specified at a higher level
     */
    public boolean isBackgroundColorInherited() {

        return(background == null);
    }

    /**
     * Gets the color that will be used to paint the cell's background.
     * @return the color which will be painted under the cell
     */
    public Color getBackgroundColor() {

        // If this cell doesn't have a background color set, look up the chain
        if (isBackgroundColorInherited()) {

            // NOTE:  Special inheritance in the default cell case
            // if this is a default cell (of a row or column)
            if (rowParent == null) {

                // since we paint the table, then rows and columns, this will
                // only be misleading if it is queried by an application

                return(background);  // background == null
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isBackgroundColorInherited()) {
                    return(columnParent.getBackgroundColor());
                }

                return(rowParent.getBackgroundColor());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isBackgroundColorInherited()) {
                    return(rowParent.getBackgroundColor());
                }

                return(columnParent.getBackgroundColor());
            }
        }
        return(background);
    }

    /**
     * Sets the color specified for painting the cell's background.
     * @param color the cell's background color
     */
    public void setBackgroundColor(Color color) {

        background = color;

        // Use this as the background color for the cell's frame
        if (frame != null) {
            frame.setColor(color);
        }
    }

    /**
     * Determines whether the style being returned for the bottom border is inherited.
     * @return <code>true</code> if the bottom border style is specified at a higher level
     */
    public boolean isBottomBorderInherited() {

        return(bottomBorder == null);
    }

    /**
     * Gets the border style applied to the bottom (row border) of the cell.
     * @return the draw style used for the bottom border of the cell
     */
    public JCDrawStyle getBottomBorderStyle() {

        // If the cell is inheriting, use the row's border style
        if (isBottomBorderInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's row border
                return(JCPageTable.this.getRowBorder());
            }

            // if the cell is not spanned, return the row's border style
            if (!isSpanned) {
                return(rowParent.getBorderStyle());
            }

            // if the cell is spanned, use the span's bottom row border style
            else {
                Cell    root_cell;
                Row     row;

                root_cell =
                        JCPageTable.this.getCell(spanOrigin.y, spanOrigin.x);
                row = JCPageTable.this.getRow(root_cell.getRowNumber()
                                          + root_cell.getSpanRows() - 1);
                return(row.getBorderStyle());
            }
        }

        return(bottomBorder);
    }

    /**
     * Sets the border style for the cell's bottom border.
     * @param border the draw style that describes the bottom border
     */
    public void setBottomBorderStyle(JCDrawStyle border) {

        // The style must be immutable!
        border = border.getImmutable();

        bottomBorder = border;
    }

    /**
     * Determines if the style being returned for the right border is inherited.
     * @return <code>true</code> if the right border style is specified at a higher level
     */
    public boolean isRightBorderInherited() {

        return(rightBorder == null);
    }

    /**
     * Gets the border style applied to the right (column border) of the cell.
     * @return the draw style used for the right border of the cell
     */
    public JCDrawStyle getRightBorderStyle() {

        // If the cell is inheriting, use the column's border style
        if (isRightBorderInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's column border
                return(JCPageTable.this.getColumnBorder());
            }

            // if the cell is not spanned, return the column's border style
            if (!isSpanned) {
                return(columnParent.getBorderStyle());
            }

            // if the cell is spanned, use the span's right column border style
            else {
                Cell    root_cell;
                Column  column;

                root_cell =
                        JCPageTable.this.getCell(spanOrigin.y, spanOrigin.x);
                column = JCPageTable.this.getColumn(root_cell.getColumnNumber()
                                            + root_cell.getSpanColumns() - 1);
                return(column.getBorderStyle());
            }
        }

        return(rightBorder);
    }

    /**
     * Sets the border style for the cell's right border.
     * @param border the draw style that describes the right border
     */
    public void setRightBorderStyle(JCDrawStyle border) {

        // The style must be immutable!
        border = border.getImmutable();

        rightBorder = border;
    }

    /**
     * Determines if the vertical cell alignment is inherited.
     * @return <code>true</code> if a higher level specifies the effective cell alignment
     */
    public boolean isCellAlignmentInherited() {
        // Valid cell alignments are [1, 3]
        return(cellAlignment == CELL_ALIGNMENT_NONE);
    }

    /**
     * Gets the alignment of the cell with respect to its row.
     * @return the vertical alignment of the cell in the row space
     */
    public int getCellAlignment() {

        if (isCellAlignmentInherited()) {
            if (rowParent == null) {
                // use the value set on this cell
                return(cellAlignment);
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isDefaultCellAlignmentInherited()) {
                    return(columnParent.getDefaultCellAlignment());
                }

                return(rowParent.getDefaultCellAlignment());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isDefaultCellAlignmentInherited()) {
                    return(rowParent.getDefaultCellAlignment());
                }

                return(columnParent.getDefaultCellAlignment());
            }
        }
        return(cellAlignment);
    }

    /**
     * Sets the alignment of the cell with respect to its row.
     * @param alignment the vertical alignment of the cell in the row space
     */
    public void setCellAlignment(int alignment) {

        if (alignment <= 0 || alignment > CELL_ALIGNMENT_BOTTOM) {
            throw new IllegalArgumentException("Unrecognised cell alignment");
        }

        cellAlignment = alignment;
    }

    /**
     * Determines if the text style used for this cell is inherited.
     * @return <code>true</code> if this cell gets its alignment from a higher level object
     */
    public boolean isStyleInherited() {

        return(textStyle == null);
    }

    /**
     * Gets the cell style.
     * @return the text style used in cell text
     */
    public JCTextStyle getStyle() {

        // if the style is unset, try the parent's default
        if (isStyleInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's value
                return(JCPageTable.this.getDefaultStyle());
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isDefaultStyleInherited()) {
                    return(columnParent.getDefaultStyle());
                }

                return(rowParent.getDefaultStyle());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isDefaultStyleInherited()) {
                    return(rowParent.getDefaultStyle());
                }

                return(columnParent.getDefaultStyle());
            }
        }
        return(textStyle);
    }

    /**
     * Sets the default cell text style.
     * @param style a text style to apply to text printed in this cell
     */
    public void setStyle(JCTextStyle style) {

        // Ensure the style is immutable
        style = style.getImmutable();

        textStyle = style;
    }

    /**
     * Determines if the size of the bottom margin is inherited.
     * @return <code>true</code> if a higher-level component specifies the bottom margin
     */
    public boolean isBottomMarginInherited() {

        return(bottomMargin == null);
    }

    /**
     * Gets the effective width of the bottom margin.
     * @return a measure of the width of the bottom margin for this cell
     */
    public JCUnit.Measure getBottomMargin() {

        JCUnit.Measure  width;

        if (isBottomMarginInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's value
                return(JCPageTable.this.getDefaultBottomMargin());
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isDefaultBottomMarginInherited()) {
                    return(columnParent.getDefaultBottomMargin());
                }

                return(rowParent.getDefaultBottomMargin());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isDefaultBottomMarginInherited()) {
                    return(rowParent.getDefaultBottomMargin());
                }

                return(columnParent.getDefaultBottomMargin());
            }
        }

        return(bottomMargin);
    }

    /**
     * Sets the width of a cell's bottom margin.
     * @param width the size of the bottom margin for this cell
     */
    public void setBottomMargin(JCUnit.Measure width) {

        // Sanity-check the new width
        if (width.isNegative()) {
            throw new IllegalArgumentException("Margin width must be non-negative");
        }

        bottomMargin = width;
    }

    /**
     * Determines if the size of the top margin is inherited.
     * @return <code>true</code> if a higher-level component specifies the top margin
     */
    public boolean isTopMarginInherited() {

        return(topMargin == null);
    }

    /**
     * Gets the effective width of the top margin.
     * @return a measure of the width of the top margin for this cell
     */
    public JCUnit.Measure getTopMargin() {

        JCUnit.Measure  width;

        if (isTopMarginInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's value
                return(JCPageTable.this.getDefaultTopMargin());
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isDefaultTopMarginInherited()) {
                    return(columnParent.getDefaultTopMargin());
                }

                return(rowParent.getDefaultTopMargin());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isDefaultTopMarginInherited()) {
                    return(rowParent.getDefaultTopMargin());
                }

                return(columnParent.getDefaultTopMargin());
            }
        }

        return(topMargin);
    }

    /**
     * Sets the width of a cell's top margin.
     * @param width the size of the top margin for this cell
     */
    public void setTopMargin(JCUnit.Measure width) {

        // Sanity-check the new width
        if (width.isNegative()) {
            throw new IllegalArgumentException("Margin width must be non-negative");
        }

        topMargin = width;
    }

    /**
     * Determines whether the size of the left margin is inherited.
     * @return <code>true</code> if a higher-level component specifies the left margin
     */
    public boolean isLeftMarginInherited() {

        return(leftMargin == null);
    }

    /**
     * Gets the effective width of the left margin.
     * @return a measure of the width of the left margin for this cell
     */
    public JCUnit.Measure getLeftMargin() {

        JCUnit.Measure  width;

        if (isLeftMarginInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's value
                return(JCPageTable.this.getDefaultLeftMargin());
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isDefaultLeftMarginInherited()) {
                    return(columnParent.getDefaultLeftMargin());
                }

                return(rowParent.getDefaultLeftMargin());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isDefaultLeftMarginInherited()) {
                    return(rowParent.getDefaultLeftMargin());
                }

                return(columnParent.getDefaultLeftMargin());
            }
        }

        return(leftMargin);
    }

    /**
     * Sets the width of a cell's left margin.
     * @param width the size of the left margin for this cell
     */
    public void setLeftMargin(JCUnit.Measure width) {

        JCUnit.Measure      cell_width;
        JCUnit.Dimension    frame_size;

        // Sanity-check the new width
        if (width.isNegative()) {
            throw new IllegalArgumentException("Margin width must be non-negative");
        }

        // Set the new width of the left margin
        leftMargin = width;

        // Don't bother computing width on default cells
        if (rowParent == null) {
            return;
        }

        // Compute the usable space between the margins
        cell_width = (JCUnit.Measure) this.width.clone();
        cell_width.subtract(getRightMargin());
        cell_width.subtract(width);

        // NOTE:  We don't throw an exception since we can't _know_
        //        this is being done in error
        /*
        if (cell_width.isNegative()) {
            throw new IllegalArgumentException("Margins are wider than cell");
        }
        */

        // Adjust the width of the cell's frame
        if (frame != null) {
            frame_size = frame.getSize();
            frame_size.setWidth(cell_width);
            frame.setSize(frame_size);
        }
    }

    /**
     * Determines whether the size of the right margin is inherited.
     * @return <code>true</code> if a higher-level component specifies the right margin
     */
    public boolean isRightMarginInherited() {

        return(rightMargin == null);
    }

    /**
     * Gets the effective width of the right margin.
     * @return a measure of the width of the right margin for this cell
     */
    public JCUnit.Measure getRightMargin() {

        JCUnit.Measure  width;

        if (isRightMarginInherited()) {

            // if this is a default cell (of a row or column)
            if (rowParent == null) {
                // use the table's value
                return(JCPageTable.this.getDefaultRightMargin());
            }

            // If row attributes are dominant over column, check the row first
            if (dominance == ROW_DOMINANCE) {

                // If not at the row, use whatever the column specifies
                if (rowParent.isDefaultRightMarginInherited()) {
                    return(columnParent.getDefaultRightMargin());
                }

                return(rowParent.getDefaultRightMargin());
            }

            else { // COLUMN_DOMINANCE

                // If not at the column, use whatever the row specifies
                if (columnParent.isDefaultRightMarginInherited()) {
                    return(rowParent.getDefaultRightMargin());
                }

                return(columnParent.getDefaultRightMargin());
            }
        }

        return(rightMargin);
    }

    /**
     * Sets the width of a cell's right margin.
     * @param width the size of the right margin for this cell
     */
    public void setRightMargin(JCUnit.Measure width) {

        JCUnit.Measure      cell_width;
        JCUnit.Dimension    frame_size;

        // Sanity-check the new width
        if (width.isNegative()) {
            throw new IllegalArgumentException("Margin width must be non-negative");
        }

        // Set the new width of the right margin
        rightMargin = width;

        // Don't bother computing width on default cells
        if (rowParent == null) {
            return;
        }

        // Compute the usable space between the margins
        cell_width = (JCUnit.Measure) this.width.clone();
        cell_width.subtract(getLeftMargin());
        cell_width.subtract(width);

        // NOTE:  We don't throw an exception since we can't _know_
        //        this is being done in error
        /*
        if (cell_width.isNegative()) {
            throw new IllegalArgumentException("Margins are wider than cell");
        }
        */

        // Adjust the width of the cell's frame
        if (frame != null) {
            frame_size = frame.getSize();
            frame_size.setWidth(cell_width);
            frame.setSize(frame_size);
        }
    }

    /**
     * Determines whether the table value for this cell is inherited.
     * @return <code>true</code> if a higher-level component specifies the cell's value
     */
    public boolean isCellValueInherited() {

        Vector dataRows = ((PageTableVectorDataSource) tableData).getCells();
        if (rowNumber >= dataRows.size()) {
            return(true);
        }

        Vector dataCells = (Vector) dataRows.get(rowNumber);
        if (columnNumber >= dataCells.size()) {
            return(true);
        }

        return(dataCells.get(columnNumber) == null);
    }

    /**
     * Gets the value assigned to this cell.
     * @return an object stored in the <code>TableDataModel</code> for use in this cell
     */
    public Object getCellValue() {

        // if this is a default cell (of a row or column) it has no value
        if (rowParent == null) {
            return(null);
        }

        // inheritance happens inside the table data...
        return(tableData.getTableDataItem(rowNumber, columnNumber));
    }

    /**
     * Sets the value stored at this cell (set in the table data).
     * @param value the object stored as the value of this cell
     */
    public void setCellValue(Object value) {

        // Don't permit values to be set on default cells
        if (rowParent == null) {
            return;
        }

        // Store the value in the TableDataModel
        tableData.setTableDataItem(value, rowNumber, columnNumber);
    }

    /**
     * Gets the height of the cell (including the height of the cell's frame).
     * @return the height of the cell
     */
    public JCUnit.Measure getHeight() {

        JCUnit.Measure      new_height, frame_height;

        // If the row specifies a height, use that
        if (rowParent.hasFixedHeight()) {
            return(rowParent.getHeight());
        }

        // if there is already a height set, honour it
        if (cellHeight != null) {
            return(cellHeight);
        }

        // Determine the amount of space needed for the cell margins
        new_height = (JCUnit.Measure) getTopMargin().clone();
        new_height.add(getBottomMargin());

        // If there is a frame in this cell, determine its size
        if (frame != null) {
            JCUnit.Measure existingFrameHeight = frame.getSize().getHeight();
            if (!existingFrameHeight.equals(JCFrame.FLEXIBLE_SIZE) &&
                existingFrameHeight.greaterThan(new JCUnit.Measure(0.0))) {
                frame_height = frame.getSize().getHeight();

            }
            else {

            // NOTE:  We return the _approximate_ height of the cell (if the
            //        cell frame has not been complete()ed) based the amount
            //        of space currently used by the frame contents

                // determine the height of the insertion point in the frame
                frame_height = frame.getInsertionPoint().getY();
            }
            // add the width of the bottom border
            frame_height.add(frame.getMargins().getBottom());

            // Add the frame's height to the cell's margin space
            new_height.add(frame_height);
        }

        // return the current height of the cell and its contents
        return(new_height);
    }

    /**
     * Sets the cell's height; this may override the dynamic sizing of the cell.
     * @param height the desired height of the cell
     */
    public void setHeight(JCUnit.Measure height) {

        JCUnit.Measure  tmp_height;

        

        if (height.isNegative()) {
            throw new IllegalArgumentException("Invalid cell height");
        }

        cellHeight = (JCUnit.Measure) height.clone();
        tmp_height = (JCUnit.Measure) height.clone();

        // Propogate the height setting to the frame
        if (frame != null) {
            tmp_height.subtract(getTopMargin());
            tmp_height.subtract(getBottomMargin());

            // Height can't be negative
            if (tmp_height.isNegative()) {
                tmp_height.distance = 0.0;
            }

            JCUnit.Dimension size = frame.getSize();
            size.setHeight(tmp_height);
            frame.setSize(size);
        }
    }

    /**
     * Gets the width of the cell.
     * @return the cell's width
     */
    public JCUnit.Measure getWidth() {

        // If the width is inherited, get it from the column
        if (width == null) {
            return(columnParent.getWidth());
        }

        return(width);
    }

    /**
     * Sets the cell's width; this needs to be done after spanning.
     * <br><B>NOTE:</b> Perhaps this shouldn't be part of the public API???
     * @param width the new width of the cell
     */
    public void setWidth(JCUnit.Measure width) {

        JCUnit.Measure  tmp_width;

        if (width.isNegative()) {
            throw new IllegalArgumentException("Invalid cell width");
        }

        this.width = (JCUnit.Measure) width.clone();
        tmp_width = (JCUnit.Measure) width.clone();

        // Recompute the frame width, etc.
        if (leftMargin != null) {
            tmp_width.subtract(leftMargin);
        }
        if (rightMargin != null) {
            tmp_width.subtract(rightMargin);
        }

        if (tmp_width.isNegative()) {
            // Negative width implies no contents, so use a zero width
            tmp_width.distance = 0.0;
        }

        // If the cell has a frame, pass the width setting to it
        if (frame != null) {
            JCUnit.Dimension size = frame.getSize();
            size.width = tmp_width.getAs(size.units);
            frame.setSize(size);
        }
    }

    /**
     * Gets the contents from from this cell.
     * @return the frame into which the cell contents are drawn
     */
    public JCFrame getFrame() {

        String              frame_name;
        JCUnit.Measure      frame_width, frame_height;
        JCUnit.Dimension    frame_size;

        // if this cell is part of a span but not the root cell
        if (isSpanned &&
                (spanOrigin.y != rowNumber || spanOrigin.x != columnNumber)) {

            // get the root cell of the span and return its frame
            return(JCPageTable.this.getCell(spanOrigin.y,
                                            spanOrigin.x).getFrame());
        }

        // Create the frame if it doesn't already exist
        if (frame == null) {

            
            // Compute the size of the frame
            frame_width = (JCUnit.Measure) width.clone();
            frame_width.subtract(getLeftMargin());
            frame_width.subtract(getRightMargin());
            if (frame_width.isNegative()) {
                frame_width.distance = 0.0;
            }
            frame_size = new JCUnit.Dimension(frame_width.units);

            // the width of the frame is the space between the margins
            frame_size.setWidth(frame_width);

            // if the cell height is set, use that for the frame height
            if (cellHeight != null) {
                frame_height = (JCUnit.Measure) cellHeight.clone();
                frame_height.subtract(getTopMargin());
                frame_height.subtract(getBottomMargin());

                if (frame_height.isNegative()) {
                    frame_height.distance = 0.0;
                }
            }
            // otherwise the height of the frame is flexible
            else {
                frame_height = (JCUnit.Measure) JCFrame.FLEXIBLE_SIZE.clone();
            }

            // set the height for the frame
            frame_size.setHeight(frame_height);
            

            // Create the new frame and set its size
            frame_name = "cell_frame_" + rowNumber + "_" + columnNumber;
            frame = new JCFrame(frame_name, JCPageTable.this.getDocument());
            frame.setSize(frame_size);
        }

        return(frame);
    }

    /**
     * Gets the frame (if any) which the cell has constructed.
     * @return the frame into which the cell contents are drawn
     */
    public JCFrame getFrameRaw() {
        return(frame);
    }

    /**
     * Returns an indication whether this cell belongs to a span or not.
     * @return <code>true</code> if this cell is part of (subsumed by) a span
     */
    public boolean isSpanned() {
        return(isSpanned);
    }

    /**
     * Returns the origin (root cell) of the spanning cell.
     * @return the co-ordinates of the root cell of the span
     */
    public Point getSpanOrigin() {

        // return an error condition if not spanned
        if (!isSpanned) {
            return(new Point(-1, -1));
        }

        return(spanOrigin);
    }

    /**
     * Sets the origin of a span.
     * @param origin the co-ordinates of the cell which is the root of the span
     */
    public void setSpanOrigin(Point origin) {
        spanOrigin = new Point(origin);
        isSpanned = true;
    }

    /**
     * Sets the origin of a span.
     * @param row the row co-ordinate of the root cell
     * @param column the column co-ordinate of the root cell
     */
    public void setSpanOrigin(int row, int column) {
        spanOrigin = new Point(column, row);
        isSpanned = true;
    }

    /**
     * Gets the number of rows this cell spans over.
     * @return the number of rows high that this span is
     */
    public int getSpanRows() {
        return(spanRows);
    }

    /**
     * Gets the number of columns this cell spans over.
     * @return the number of columns wide that this span is
     */
    public int getSpanColumns() {
        return(spanColumns);
    }

    /**
     * Sets the number of rows and columns in the span.
     * @param rows the number of rows high the cell is
     * @param columns the number of columns wide the cell is
     */
    public void setSpanSize(int rows, int columns) {

        if (rows < 1 || columns < 1 ||
            columns > JCPageTable.this.getColumnCount()) {
            throw new IllegalArgumentException("Invalid span size");
        }

        spanRows = rows;
        spanColumns = columns;
    }

    /**
     * Gets the column parent of the cell in the table.
     * @return the column to which this cell belongs
     */
    public Column getColumn() {
        return(columnParent);
    }

    /**
     * Gets the column co-ordinate of the cell in the table.
     * @return the number of the column to which this cell belongs
     */
    public int getColumnNumber() {
        return(columnNumber);
    }

    /**
     * Gets the row parent of the cell in the table.
     * @return the row to which this cell belongs
     */
    public Row getRow() {
        return(rowParent);
    }

    /**
     * Gets the row co-ordinate of the cell in the table.
     * @return the number of the row to which this cell belongs
     */
    public int getRowNumber() {
        return(rowNumber);
    }

    /**
     * Computes the height of the cell based on the height of the cell's frame.
     * @return the height of the cell
     */
    public JCUnit.Measure fixHeight() {

        JCUnit.Measure      new_height, frame_height;
        JCUnit.Dimension    frame_size;

        

        // Determine the amount of space needed for the cell margins
        new_height = (JCUnit.Measure) getTopMargin().clone();
        new_height.add(getBottomMargin());

        // If there is a frame in this cell, complete and size it
        if (frame != null) {

            // paste the frame's last (current) line

            frame.complete();
            
            // determine the height of the insertion point in the frame
            JCUnit.Measure existingFrameHeight = frame.getSize().getHeight();
            if (!existingFrameHeight.equals(JCFrame.FLEXIBLE_SIZE) &&
                existingFrameHeight.greaterThan(new JCUnit.Measure(0.0))) {
                frame_height = frame.getSize().getHeight();
            } 
            else {
                frame_height = frame.getInsertionPoint().getY();


                
            }
            // add the width of the bottom border
            frame_height.add(frame.getMargins().getBottom());

            // set this as the height of the frame
            frame_size = frame.getSize();
            frame_size.setHeight(frame_height);
            frame.setSize(frame_size);

            // Add the frame's height to the cell's margin space
            new_height.add(frame_height);
            
        }

        // if there isn't a frame in the cell, check for tableData
        else if (tableData != null) {
            Object cellData = tableData.getTableDataItem(rowNumber, columnNumber);

            // Ignore the cell if there is no stored data
            if (cellData != null) {

                // Get a renderer for the cell data
                Class type = cellData.getClass();
                JCCellRenderer cr = EditorRendererRegistry.getCentralRegistry().
                                        getCellRendererInstance(type, null);

                // Create a fake JCCellInfo object
                JCCellInfo cellInfo = new PageTableCellInfo(JCPageTable.this,
                                                    rowNumber, columnNumber);
                Dimension prefSize;

                if (cr instanceof JCComponentCellRenderer) {

                    // get the renderer component and get its preferred size
                    Component cellComponent = ((JCComponentCellRenderer)
                            cr).getRendererComponent(cellInfo, cellData, false);
                    prefSize = cellComponent.getPreferredSize();
                }
                else { // (cr instanceof JCLightCellRenderer)

                    // Get the preferred size from the cell renderer
                    Graphics gc = document.getPrinter().getGraphics();
                    prefSize = ((JCLightCellRenderer)
                                cr).getPreferredSize(gc, cellInfo, cellData);
                }

                // Add the cell height info extracted from the preferred size
                new_height.add(new JCUnit.Measure(JCUnit.POINTS,
                                                  prefSize.height));
            }
        }

        // If the row has a fixed height, use that
        if (rowParent.hasFixedHeight()) {
            return(rowParent.getHeight());
        }
        // else if there is already a height set, honour it
        else if (cellHeight != null) {
            
            return(cellHeight);
        }
        // otherwise set and return the computed height of the cell
        else {
            cellHeight = new_height;
            return(cellHeight);
        }
    }

    /**
     * Gets the bottom border style of this cell position (if there is no
     * bottom border because of a span, this will return null).
     * @return the border style of the bottom of this cell position
     */
    protected JCDrawStyle getBottomEdgeStyle() {

        JCDrawStyle border;

        // if the cell is part of a span we need to be on the bottom edge
        if (isSpanned) {
            Cell root_cell = JCPageTable.this.getCell(spanOrigin.y,
                                                      spanOrigin.x);

            // If we are not in the bottommost row of the span
            if (spanOrigin.y + root_cell.getSpanRows() - 1 > rowNumber) {
                return(null);
            }

            // In the bottom row use the span's bottom border
            else {
                border = root_cell.getBottomBorderStyle();
            }
        }

        // otherwise we are not part of a span, simply get the bottom border
        else {
            border = getBottomBorderStyle();
        }

        return(border);
    }

    /**
     * Gets the right border style of this cell position (if there is no
     * right border because of a span, this will return null).
     * @return the border style of the right of this cell position
     */
    protected JCDrawStyle getRightEdgeStyle() {

        JCDrawStyle border;

        // if the cell is part of a span we need to be on the right edge
        if (isSpanned) {
            Cell root_cell = JCPageTable.this.getCell(spanOrigin.y,
                                                      spanOrigin.x);

            // If we are not in the rightmost column of the span
            if (spanOrigin.x + root_cell.getSpanColumns() - 1 > columnNumber) {
                return(null);
            }

            // In the right column use the span's right border
            else {
                border = root_cell.getRightBorderStyle();
            }
        }

        // otherwise we are not part of a span, simply get the right border
        else {
            border = getRightBorderStyle();
        }

        return(border);
    }

    /**
     * Adds the contents of this cell (cell frame) to the given frame.
     * @param target the frame to draw paste the cell into
     * @param x0 the <i>x</i>-coordinate of the location to paste the cell at
     * @param y0 the <i>y</i>-coordinate of the location to paste the cell at
     */
    protected void paste(JCFrame target, JCUnit.Measure x0, JCUnit.Measure y0) {

        JCFrame         new_frame;
        JCUnit.Point    location;

        // Create a copy of the cell's frame
        new_frame = (JCFrame) frame.clone();

        // set the position of the frame
        location = new JCUnit.Point(x0.units, x0.distance, y0.distance);
        if (y0.units != x0.units) {
            location.y = y0.getAs(x0.units);
        }
        new_frame.setLocation(location);

        // add the copied frame to the target frame
        target.getFrameList().add(new_frame);
    }

    /**
     * Adds the contents of this cell (cell frame) to the given frame subject
     * to the given mask for clipping the cell.
     * @param target the frame to draw paste the cell into
     * @param x0 the <i>x</i>-coordinate of the location to paste the cell at
     * @param y0 the <i>y</i>-coordinate of the location to paste the cell at
     * @param mask the clip mask to apply to displaying the cell
     */
    protected void paste(JCFrame target, JCUnit.Measure x0, JCUnit.Measure y0,
                         JCUnit.Margins mask) {

        JCFrame         new_frame;
        JCUnit.Point    location;

        // Create a copy of the cell's frame
        new_frame = (JCFrame) frame.clone();

        // set the position of the frame
        location = new JCUnit.Point(x0.units, x0.distance, y0.distance);
        if (y0.units != x0.units) {
            location.y = y0.getAs(x0.units);
        }
        new_frame.setLocation(location);

        // Set the clip mask on the new frame
        new_frame.setClipMask(mask);

        // add the copied frame to the target frame
        target.getFrameList().add(new_frame);
    }

    /**
     * Locates the contents of this cell (cell frame) in the given frame subject
     * to the amount of vertical space available and the cell alignment.
     * @param target the frame to draw paste the cell into
     * @param x0 the <i>x</i>-coordinate of the base location for the cell
     * @param y0 the <i>y</i>-coordinate of the base location for the cell
     * @param space the amount of vertical space available for the cell
     */
    protected void paste(JCFrame target, JCUnit.Measure x0, JCUnit.Measure y0,
                         JCUnit.Measure space) {

        int             v_align;
        JCUnit.Measure    x_delta, y_delta;
        JCUnit.Measure    clip_x, clip_width, clip_y, clip_height;

        // Create working copies of the location variables
        JCUnit.Measure x1 = (JCUnit.Measure) x0.clone();
        JCUnit.Measure y1 = (JCUnit.Measure) y0.clone();

        // Get the margin widths
        JCUnit.Measure top_margin = getTopMargin();
        JCUnit.Measure bottom_margin = getBottomMargin();
        JCUnit.Measure left_margin = getLeftMargin();
        JCUnit.Measure right_margin = getRightMargin();

        // Get the size of the frame
        JCUnit.Measure frame_height = this.frame.getSize().getHeight();
        JCUnit.Measure frame_width = this.frame.getSize().getWidth();

        // compute the horizontal offset of the cell frame
        x_delta = (JCUnit.Measure) left_margin.clone();

        // Adjust the frame's location according to the vertical alignment
        v_align = getCellAlignment();
        if (v_align == CELL_ALIGNMENT_TOP) {
            y_delta = (JCUnit.Measure) top_margin.clone();
        }
        else if (v_align == CELL_ALIGNMENT_CENTER) {
            y_delta = (JCUnit.Measure) space.clone();
            y_delta.subtract(top_margin);
            y_delta.subtract(bottom_margin);
            y_delta.subtract(frame_height);
            y_delta.divide(2.0);

            y_delta.add(top_margin);
        }
        else if (v_align == CELL_ALIGNMENT_BOTTOM) {
            y_delta = (JCUnit.Measure) space.clone();
            y_delta.subtract(bottom_margin);
            y_delta.subtract(frame_height);
        }
        else {    // CELL_ALIGNMENT_NONE => CELL_ALIGNMENT_TOP
            y_delta = (JCUnit.Measure) top_margin.clone();
        }

        // -------  Compute the vertical clip start and extent ------
        // Initially we assume the whole frame will fit.
        clip_height = (JCUnit.Measure) frame_height.clone();

        clip_y = new JCUnit.Measure();

        // if the vertical offset of the frame origin is inside the top margin
        // space (or even - if negative - outside the cell)
        if (top_margin.greaterThan(y_delta)) {

            // clip to the edge of the margin space
            clip_y.assign(top_margin);
            clip_y.subtract(y_delta);

            // adjust the height of the displayed part of the frame
            clip_height.subtract(clip_y);
        }

        JCUnit.Measure frame_bottom, bottom_space;

        // Compute the bottom of the frame content and the bottom space
        frame_bottom = (JCUnit.Measure) frame_height.clone();
        frame_bottom.add(y_delta);
        bottom_space = (JCUnit.Measure) space.clone();
        bottom_space.subtract(bottom_margin);

        // If the bottom of the frame extends past the bottom of the
        // available space to draw it in
        if (frame_bottom.greaterThan(bottom_space)) {

            // reduce the clip height by (frame_bottom - bottom_space)
            clip_height.subtract(frame_bottom);
            clip_height.add(bottom_space);
        }
        // -------  end vertical clip --------

        // -------  Compute the horizontal clip start and extent ------
        // Initially we assume the whole frame will fit.
        clip_width = (JCUnit.Measure) frame_width.clone();

        // Compute the amount of the content frame left of the displayed region
        clip_x = new JCUnit.Measure();

        // if the horizontal offset of the frame origin falls inside the left
        // margin space
        if (left_margin.greaterThan(x_delta)) {

            // clip to the edge of the margin space
            clip_x.assign(left_margin);
            clip_x.subtract(x_delta);

            // adjust the width of the displayed part of the frame
            clip_width.subtract(clip_x);
        }

        JCUnit.Measure frame_right, right_space;

        // Compute the right extent of the frame content and the right space
        frame_right = (JCUnit.Measure) frame_width.clone();
        frame_right.add(x_delta);
        right_space = (JCUnit.Measure) getWidth().clone();
        right_space.subtract(right_margin);

        // If the right of the frame extends past the right of the
        // available space to draw it in
        if (frame_right.greaterThan(right_space)) {

            // reduce the clip width by (frame_right - right_space)
            clip_width.subtract(frame_right);
            clip_width.add(right_space);
        }
        // -------  end horizontal clip --------

        // Compute the origin of the frame on the page
        x1.add(x_delta);
        y1.add(y_delta);

        // generate the clip mask for the displayed part
        JCUnit.Margins clip_mask = new JCUnit.Margins(clip_x.units);
        clip_mask.left = clip_x.distance;
        clip_mask.top = clip_y.getAs(clip_x.units);
        clip_mask.right = clip_width.getAs(clip_x.units);
        clip_mask.bottom = clip_height.getAs(clip_x.units);

        paste(target, x1, y1, clip_mask);
    }

    /**
     * Pastes the portion of this spanned cell that lies within the given
     * range of rows and columns of the table.
     * @param target the frame to draw paste the cell into
     * @param startRow the first row of the table to display the span in
     * @param startColumn the first column of the table to display the span in
     * @param endRow the last row of the table to display the span in
     * @param endColumn the last column of the table to display the span in
     * @param x0 the <i>x</i>-coordinate of the base location for the cell
     * @param y0 the <i>y</i>-coordinate of the base location for the cell
     */
    protected void pasteSpan(JCFrame target, int startRow, int startColumn,
                             int endRow, int endColumn,
                             JCUnit.Measure x0, JCUnit.Measure y0) {

        int             r, c, v_align;
        Row             row;
        Column          column;
        JCUnit.Measure  span_height, before_height, after_height;
        JCUnit.Measure  span_width, before_width, after_width;
        JCUnit.Measure  left_margin, top_margin, bottom_margin;
        JCUnit.Measure  x_delta, y_delta, x1, y1;
        JCUnit.Measure  clip_x, clip_y, clip_width, clip_height;
        JCUnit.Measure  border, frame_height, frame_width;
        JCUnit.Margins  clip_mask;

        // NOTE:  This function should always to be called on the root
        //        cell of a spanned cell

        // Cope with calling on a non-spanned cell
        if (!isSpanned) {
            
            span_height = (JCUnit.Measure) rowParent.getHeight().clone();
            span_height.add(rowParent.getExtraHeight());
            paste(target, x0, y0, span_height);
            return;
        }

        // Cope with calling not on root cell of the span
        if (spanOrigin.y != rowNumber || spanOrigin.x != columnNumber) {
            
            
            
            Cell root = JCPageTable.this.getCell(spanOrigin.y, spanOrigin.x);
            root.pasteSpan(target, startRow, startColumn, endRow, endColumn,
                           x0, y0);
            return;
        }

        // Hereafter we definitely have the root of a span
        

        // Create variables to store the position of the displayed area
        // within the span
        before_height = new JCUnit.Measure();
        after_height = new JCUnit.Measure();
        before_width = new JCUnit.Measure();
        after_width = new JCUnit.Measure();

        // Sum the height of the span
        row = JCPageTable.this.getRow(spanOrigin.y);
        span_height = (JCUnit.Measure) row.getHeight().clone();
        span_height.add(row.getExtraHeight());

        for (r = spanOrigin.y + 1; r < spanOrigin.y + spanRows; r++) {
            border = row.getBorderWidth();

            // at the start of the range to display save the current position
            if (r == startRow) {
                before_height.assign(span_height);
                // include half the preceding border
                border.divide(2.0);
                before_height.add(border);
                border.multiply(2.0);
            }

            // at the end of the range, store the position
            if (r == endRow + 1) {
                after_height.assign(span_height);
                // include half the last border space
                border.divide(2.0);
                after_height.add(border);
                border.multiply(2.0);
            }

            // add the (internal) border width to the span
            span_height.add(border);

            // get the next row and add its height
            row = JCPageTable.this.getRow(r);
            span_height.add(row.getHeight());
            span_height.add(row.getExtraHeight());
        }

        // compute the height of the portion after the span
        if (!after_height.isZero()) {
            after_height.subtract(span_height);
            after_height.multiply(-1.0);
        }

        // Sum the width of the span (including column border spaces)
        column = JCPageTable.this.getColumn(spanOrigin.x);
        span_width = (JCUnit.Measure) column.getWidth().clone();

        for (c = spanOrigin.x + 1; c < spanOrigin.x + spanColumns; c++) {
            // border = column.getBorderWidth();

            // at the start of the range to display save the current position
            if (c == startColumn) {
                before_width.assign(span_width);
                // NOTE:  PF added _half_ the border width here.
            }

            // add the border width from the current column
            span_width.add(column.getBorderWidth());

            // at the end of the range to display store the position
            if (c == endColumn + 1) {
                after_width.assign(span_width);
            }

            // get the next column and add its width
            column = JCPageTable.this.getColumn(c);
            span_width.add(column.getWidth());
        }

        // compute the width of the portion following the displayed part
        if (!after_width.isZero()) {
            after_width.subtract(span_width);
            after_width.multiply(-1.0);
        }

        // if the spanned cell is entirely displayed (no cut-offs recorded)
        if (before_height.isZero() && after_height.isZero() &&
            before_width.isZero() && after_width.isZero()) {

            // paste the complete cell in the available space
            paste(target, x0, y0, span_height);
            return;
        }

        // at this point we're pasting part of a cell; compute the region

        // compute the frame offset based on the cell margins
        frame_height = this.frame.getSize().getHeight();
        frame_width = this.frame.getSize().getWidth();
        left_margin = getLeftMargin();
        top_margin = getTopMargin();
        bottom_margin = getBottomMargin();

        // absent any horizontal alignment modes the x offset is the margin
        x_delta = left_margin;

        // adjust the vertical offset based on the alignment
        v_align = getCellAlignment();
        if (v_align == CELL_ALIGNMENT_TOP) {
            y_delta = (JCUnit.Measure) top_margin.clone();
        }
        else if (v_align == CELL_ALIGNMENT_CENTER) {
            y_delta = (JCUnit.Measure) span_height.clone();
            y_delta.subtract(top_margin);
            y_delta.subtract(bottom_margin);
            y_delta.subtract(frame_height);
            y_delta.divide(2.0);
            y_delta.add(top_margin);
        }
        else if (v_align == CELL_ALIGNMENT_BOTTOM) {
            y_delta = (JCUnit.Measure) span_height.clone();
            y_delta.subtract(bottom_margin);
            y_delta.subtract(frame_height);
        }
        else {
            y_delta = (JCUnit.Measure) top_margin.clone();
        }

        // -------  Compute the vertical clip start and extent ------
        // Initially we assume the whole frame will fit.
        clip_height = (JCUnit.Measure) frame_height.clone();

        // Compute the amount of the content frame above the top of the
        // displayed region of the span
        clip_y = (JCUnit.Measure) before_height.clone();
        clip_y.subtract(y_delta);

        // if the amount above is negative, there is still unused vertical
        // space inside the displayed region
        if (clip_y.isNegative()) {
            // We don't clip any of the top of the frame
            clip_y.distance = 0.0;
        }
        else {
            // Remove the space clipped at the top from the clip height
            clip_height.subtract(clip_y);
        }

        JCUnit.Measure frame_bottom, bottom_space;

        // Compute the bottom of the frame content and the bottom space
        frame_bottom = (JCUnit.Measure) frame_height.clone();
        frame_bottom.add(y_delta);
        bottom_space = (JCUnit.Measure) span_height.clone();
        bottom_space.subtract(after_height);

        // If the bottom of the frame extends past the bottom of the
        // available space to draw it in
        if (frame_bottom.greaterThan(bottom_space)) {

            // reduce the clip height by (frame_bottom - bottom_space)
            clip_height.subtract(frame_bottom);
            clip_height.add(bottom_space);
        }
        // -------  end vertical clip --------

        // -------  Compute the horizontal clip start and extent ------
        // Initially we assume the whole frame will fit.
        clip_width = (JCUnit.Measure) frame_width.clone();

        // Compute the amount of the content frame left of the displayed region
        clip_x = (JCUnit.Measure) before_width.clone();
        clip_x.subtract(left_margin);

        // if the extent of the frame left of this region is negative, there
        // is still unused horizontal space inside the displayed region
        if (clip_x.isNegative()) {
            // We don't clip from the left side of the frame
            clip_x.distance = 0.0;
        }
        else {
            // Remove the clipped space from the clip width
            clip_width.subtract(clip_x);
        }

        JCUnit.Measure frame_right, right_space;

        // Compute the right extent of the frame content and the right space
        frame_right = (JCUnit.Measure) frame_width.clone();
        frame_right.add(left_margin);
        right_space = (JCUnit.Measure) span_width.clone();
        right_space.subtract(after_width);

        // If the right of the frame extends past the right of the
        // available space to draw it in
        if (frame_right.greaterThan(right_space)) {

            // reduce the clip width by (frame_right - right_space)
            clip_width.subtract(frame_right);
            clip_width.add(right_space);
        }
        // -------  end horizontal clip --------

        // generate the clip mask for the displayed part
        clip_mask = new JCUnit.Margins(clip_x.units);
        clip_mask.left = clip_x.distance;
        clip_mask.top = clip_y.getAs(clip_x.units);
        clip_mask.right = clip_width.getAs(clip_x.units);
        clip_mask.bottom = clip_height.getAs(clip_x.units);

        // adjust the origin of the frame if it lies left or above
        x1 = (JCUnit.Measure) x0.clone();
        x1.add(x_delta);
        x1.subtract(before_width);
        y1 = (JCUnit.Measure) y0.clone();
        y1.add(y_delta);
        y1.subtract(before_height);

        // paste the cell at its computed position
        paste(target, x1, y1, clip_mask);
    }

} // JCPageTable.Cell

} // JCPageTable
